import os.path
import sys

sys.path.append(os.path.dirname(__file__))  # let code under test load stubs
from device_test import DeviceTestCase

class TestCalixModels(DeviceTestCase):
    def test_calix_exa_E7_20_3_4_51_25_01(self):
        obj = self.run_data("calix.exa.E7-20.3.4.51.25.01")
        exp_info = {
         'model': 'E7-20', 'sta_count': 1644, 'sw_version': None, 'vendor': 'Calix'
        }
        exp_stats = {
         'serial_number': '71412913092', 'system_mac_address': None
        }
        exp_sta_stats = {
                '100AAF9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "6979 MILLBANK MAIN ST (ZEHR'S COUNTRY "
                                           'MARKET)',
                                   'serial': '100AAF9',
                                   'serial_number': '100AAF9',
                                   'system_mac_address': None},
                '10948B4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '55 WILLIAM ST Unit 1 (Rahman Habibur)',
                                   'serial': '10948B4',
                                   'serial_number': '10948B4',
                                   'system_mac_address': None},
                '10948D1': {       'dev_macs': ['5c:0f:fb:2f:c6:e4'],
                                   'ip_address': None,
                                   'name': '10 MAIN ST N (JACOBS CHAD)',
                                   'serial': '10948D1',
                                   'serial_number': '10948D1',
                                   'system_mac_address': None},
                '10948FD': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '55 WILLIAM STREET APT 207 (FRASER KATHY)',
                                   'serial': '10948FD',
                                   'serial_number': '10948FD',
                                   'system_mac_address': None},
                '1094906': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2 MAIN ST N (FOSTER ANNE)',
                                   'serial': '1094906',
                                   'serial_number': '1094906',
                                   'system_mac_address': None},
                '10A8635': {       'dev_macs': ['14:21:03:92:c7:19'],
                                   'ip_address': None,
                                   'name': '51 COULTER ST (MARCHESE ANTHONY)',
                                   'serial': '10A8635',
                                   'serial_number': '10A8635',
                                   'system_mac_address': None},
                '10A8641': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7088 PERTH ROAD 121 (SAUDER MAYNARD)',
                                   'serial': '10A8641',
                                   'serial_number': '10A8641',
                                   'system_mac_address': None},
                '10A8642': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4460 LINE 52 (ROPP RICHARD)',
                                   'serial': '10A8642',
                                   'serial_number': '10A8642',
                                   'system_mac_address': None},
                '10A8644': {       'dev_macs': ['14:21:03:a0:88:95'],
                                   'ip_address': None,
                                   'name': '2288 LINE 47 (BALFOUR WENDY F)',
                                   'serial': '10A8644',
                                   'serial_number': '10A8644',
                                   'system_mac_address': None},
                '10A86A1': {       'dev_macs': ['14:21:03:92:f0:f3'],
                                   'ip_address': None,
                                   'name': '6405 ROAD 116 (METZGER CHARLENE)',
                                   'serial': '10A86A1',
                                   'serial_number': '10A86A1',
                                   'system_mac_address': None},
                '10A86AB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7106 PERTH ROAD 121 (SCHULTZ DALE)',
                                   'serial': '10A86AB',
                                   'serial_number': '10A86AB',
                                   'system_mac_address': None},
                '10A86AD': {       'dev_macs': ['14:21:03:a0:8a:58'],
                                   'ip_address': None,
                                   'name': '4149 LINE 46 (HARWOOD PAUL D)',
                                   'serial': '10A86AD',
                                   'serial_number': '10A86AD',
                                   'system_mac_address': None},
                '10A86BD': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6943 ROAD 124 (ALBRECHT MURRAY & LORETTA)',
                                   'serial': '10A86BD',
                                   'serial_number': '10A86BD',
                                   'system_mac_address': None},
                '10A86E7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4213 LINE 67 (MARTIN PAUL M)',
                                   'serial': '10A86E7',
                                   'serial_number': '10A86E7',
                                   'system_mac_address': None},
                '10A86ED': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6409 PERTH ROAD 121 (SCHERING LUCAS)',
                                   'serial': '10A86ED',
                                   'serial_number': '10A86ED',
                                   'system_mac_address': None},
                '10A871D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2700 LINE 45 (BALFOUR WESLEY)',
                                   'serial': '10A871D',
                                   'serial_number': '10A871D',
                                   'system_mac_address': None},
                '10A872C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5719 PERTH LINE 55 (CHARMAINE VALDEZ '
                                           'JOYCE)',
                                   'serial': '10A872C',
                                   'serial_number': '10A872C',
                                   'system_mac_address': None},
                '111BC3': {       'dev_macs': [       '5c:0f:fb:3d:5a:32',
                                                      '5c:0f:fb:3d:5a:e4',
                                                      '5c:0f:fb:3d:57:d0',
                                                      '00:03:e6:a5:74:46'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST',
                                  'serial': '111BC3',
                                  'serial_number': '111BC3',
                                  'system_mac_address': None},
                '117D97D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (SMITH JESSICA)',
                                   'serial': '117D97D',
                                   'serial_number': '117D97D',
                                   'system_mac_address': None},
                '117D993': {       'dev_macs': ['e4:6c:d1:4e:ec:ca'],
                                   'ip_address': None,
                                   'name': '29 AVERY PL (MCDOWELL PATTI)',
                                   'serial': '117D993',
                                   'serial_number': '117D993',
                                   'system_mac_address': None},
                '117D9DE': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2520 LINE 45 (ACKERSVILLER BRADLEY)',
                                   'serial': '117D9DE',
                                   'serial_number': '117D9DE',
                                   'system_mac_address': None},
                '117DA11': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4454 PERTH LINE 72 (WILLIAMS JUSTIN)',
                                   'serial': '117DA11',
                                   'serial_number': '117DA11',
                                   'system_mac_address': None},
                '117DA1D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4538 LINE 46 (ROSE JOHN P)',
                                   'serial': '117DA1D',
                                   'serial_number': '117DA1D',
                                   'system_mac_address': None},
                '117DA3B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4460 COMMUNITY CENTRE DR (ALBRECHT GREG)',
                                   'serial': '117DA3B',
                                   'serial_number': '117DA3B',
                                   'system_mac_address': None},
                '117DA6D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6267 PERTH ROAD 121 (ROES TYRELL)',
                                   'serial': '117DA6D',
                                   'serial_number': '117DA6D',
                                   'system_mac_address': None},
                '117DA73': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '111 PUGH ST (BECKER JANE)',
                                   'serial': '117DA73',
                                   'serial_number': '117DA73',
                                   'system_mac_address': None},
                '117DAA8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4458 COMMUNITY CENTRE DR (HAWLEY ALAN)',
                                   'serial': '117DAA8',
                                   'serial_number': '117DAA8',
                                   'system_mac_address': None},
                '117DAB7': {       'dev_macs': ['14:21:03:8b:74:af'],
                                   'ip_address': None,
                                   'name': '5135 ROAD 122 (REIGER JOE)',
                                   'serial': '117DAB7',
                                   'serial_number': '117DAB7',
                                   'system_mac_address': None},
                '117DAB8': {       'dev_macs': ['14:21:03:92:e9:c6'],
                                   'ip_address': None,
                                   'name': '4290 LINE 46 (BRUBACHER PAUL & CATHY)',
                                   'serial': '117DAB8',
                                   'serial_number': '117DAB8',
                                   'system_mac_address': None},
                '117DAC1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4081 PERTH LINE 72 (BARTSCH DAVID)',
                                   'serial': '117DAC1',
                                   'serial_number': '117DAC1',
                                   'system_mac_address': None},
                '117DAC6': {       'dev_macs': ['14:21:03:92:e9:6e'],
                                   'ip_address': None,
                                   'name': '4179 LINE 46 (PLAYERDISE FARMS LTD)',
                                   'serial': '117DAC6',
                                   'serial_number': '117DAC6',
                                   'system_mac_address': None},
                '117DAD1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4458 PERTH LINE 72 (WINSTON MIKE & SHARON)',
                                   'serial': '117DAD1',
                                   'serial_number': '117DAD1',
                                   'system_mac_address': None},
                '117DAD2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '94 Mill St. E. Milverton',
                                   'serial': '117DAD2',
                                   'serial_number': '117DAD2',
                                   'system_mac_address': None},
                '117DAD4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4188 LINE 46 (VAN MEURS MARTEN)',
                                   'serial': '117DAD4',
                                   'serial_number': '117DAD4',
                                   'system_mac_address': None},
                '117DADC': {       'dev_macs': ['14:21:03:8b:6d:14'],
                                   'ip_address': None,
                                   'name': '4247 LINE 46 (THOMAS CHRISTIAN)',
                                   'serial': '117DADC',
                                   'serial_number': '117DADC',
                                   'system_mac_address': None},
                '117DAE3': {       'dev_macs': ['14:21:03:8b:6c:b1'],
                                   'ip_address': None,
                                   'name': '2493 LINE 45 (HYDE AMBER)',
                                   'serial': '117DAE3',
                                   'serial_number': '117DAE3',
                                   'system_mac_address': None},
                '117DAE8': {       'dev_macs': ['14:21:03:8b:72:e1'],
                                   'ip_address': None,
                                   'name': '2542 LINE 45 (ROSE JOHN & LOIS)',
                                   'serial': '117DAE8',
                                   'serial_number': '117DAE8',
                                   'system_mac_address': None},
                '117DAED': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6569 ROAD 129 (DIETRICH DIANE)',
                                   'serial': '117DAED',
                                   'serial_number': '117DAED',
                                   'system_mac_address': None},
                '117DAF8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2624 LINE 47 (PATCH DALE)',
                                   'serial': '117DAF8',
                                   'serial_number': '117DAF8',
                                   'system_mac_address': None},
                '117DAFC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6889 ROAD 124 (JANTZI REUBEN K)',
                                   'serial': '117DAFC',
                                   'serial_number': '117DAFC',
                                   'system_mac_address': None},
                '117DAFE': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4434 PERTH LINE 72 (SCHMIDT KENDRICK)',
                                   'serial': '117DAFE',
                                   'serial_number': '117DAFE',
                                   'system_mac_address': None},
                '117DB0B': {       'dev_macs': ['14:21:03:8b:75:49'],
                                   'ip_address': None,
                                   'name': '15 MILL ST W (ANSTETT ADAM)',
                                   'serial': '117DB0B',
                                   'serial_number': '117DB0B',
                                   'system_mac_address': None},
                '117DB0E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4452 COMMUNITY CENTRE DR (GERBER RONALD)',
                                   'serial': '117DB0E',
                                   'serial_number': '117DB0E',
                                   'system_mac_address': None},
                '117DB13': {       'dev_macs': ['14:21:03:92:f5:fa'],
                                   'ip_address': None,
                                   'name': '115 PUGH ST (BASHNIAK MYKOLA)',
                                   'serial': '117DB13',
                                   'serial_number': '117DB13',
                                   'system_mac_address': None},
                '117DB22': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4461 COMMUNITY CENTRE DR (WAGLER EDGAR)',
                                   'serial': '117DB22',
                                   'serial_number': '117DB22',
                                   'system_mac_address': None},
                '117DB2E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '1 SCHADE ST MONKTON (WALL GEORGE)',
                                   'serial': '117DB2E',
                                   'serial_number': '117DB2E',
                                   'system_mac_address': None},
                '117DB30': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4463 PERTH LINE 72 (WAGLER LANELLE)',
                                   'serial': '117DB30',
                                   'serial_number': '117DB30',
                                   'system_mac_address': None},
                '117DB34': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4455 COMMUNITY CENTRE DR (ROLFE ROSANNE)',
                                   'serial': '117DB34',
                                   'serial_number': '117DB34',
                                   'system_mac_address': None},
                '117DB39': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6951 ELGIN ST (POOLE SAMANTHA)',
                                   'serial': '117DB39',
                                   'serial_number': '117DB39',
                                   'system_mac_address': None},
                '117DB41': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4174 LINE 61 (KUEPFER SARAH)',
                                   'serial': '117DB41',
                                   'serial_number': '117DB41',
                                   'system_mac_address': None},
                '117DB45': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5122 ROAD 122 (VAN STRAATEN DAVE)',
                                   'serial': '117DB45',
                                   'serial_number': '117DB45',
                                   'system_mac_address': None},
                '117DB48': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4168 LINE 46 (LOUWAGIE DENNIS)',
                                   'serial': '117DB48',
                                   'serial_number': '117DB48',
                                   'system_mac_address': None},
                '117DB4C': {       'dev_macs': ['14:21:03:8b:6f:be'],
                                   'ip_address': None,
                                   'name': '2483 LINE 51 (JANTZI ROBERT)',
                                   'serial': '117DB4C',
                                   'serial_number': '117DB4C',
                                   'system_mac_address': None},
                '117DB4F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6571 ROAD 140 (EBERSOL AMANDA)',
                                   'serial': '117DB4F',
                                   'serial_number': '117DB4F',
                                   'system_mac_address': None},
                '117DB51': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4442 PERTH LINE 72 (HIEBERT JACOB)',
                                   'serial': '117DB51',
                                   'serial_number': '117DB51',
                                   'system_mac_address': None},
                '117DB58': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2460 LINE 47 (GERBER COLLIN)',
                                   'serial': '117DB58',
                                   'serial_number': '117DB58',
                                   'system_mac_address': None},
                '117DB60': {       'dev_macs': ['14:21:03:92:eb:52'],
                                   'ip_address': None,
                                   'name': '61 PUGH ST (WOODTHORPE PETER)',
                                   'serial': '117DB60',
                                   'serial_number': '117DB60',
                                   'system_mac_address': None},
                '117DB6E': {       'dev_macs': [       'd4:6e:0e:9f:e4:9b',
                                                       '48:ee:0c:28:e6:4f'],
                                   'ip_address': None,
                                   'name': '4467 PERTH LINE 72 (BUECKERT PETER & '
                                           'TRUDY)',
                                   'serial': '117DB6E',
                                   'serial_number': '117DB6E',
                                   'system_mac_address': None},
                '117DB71': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '14 CRUICKSHANK ST (BRAUN JACOB)',
                                   'serial': '117DB71',
                                   'serial_number': '117DB71',
                                   'system_mac_address': None},
                '117DB76': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4917 LINE 67 (EBERSOL ALVIN)',
                                   'serial': '117DB76',
                                   'serial_number': '117DB76',
                                   'system_mac_address': None},
                '117DB7A': {       'dev_macs': ['14:21:03:92:c6:a0'],
                                   'ip_address': None,
                                   'name': '4119 LINE 46 (QUAST CHERYL)',
                                   'serial': '117DB7A',
                                   'serial_number': '117DB7A',
                                   'system_mac_address': None},
                '117DB92': {       'dev_macs': ['14:21:03:92:c7:9d'],
                                   'ip_address': None,
                                   'name': '70 PUGH ST (ZINKEN JONATHAN)',
                                   'serial': '117DB92',
                                   'serial_number': '117DB92',
                                   'system_mac_address': None},
                '117DB98': {       'dev_macs': ['14:21:03:92:ed:83'],
                                   'ip_address': None,
                                   'name': '46 AVERY PL (BENDER BOB & JANINE)',
                                   'serial': '117DB98',
                                   'serial_number': '117DB98',
                                   'system_mac_address': None},
                '117DBB0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4464 COMMUNITY CENTRE DR (GINGRICH NATHAN)',
                                   'serial': '117DBB0',
                                   'serial_number': '117DBB0',
                                   'system_mac_address': None},
                '117DBB2': {       'dev_macs': ['14:21:03:92:ea:60'],
                                   'ip_address': None,
                                   'name': '4463 COMMUNITY CENTRE DR (CARTER RUTH ANN)',
                                   'serial': '117DBB2',
                                   'serial_number': '117DBB2',
                                   'system_mac_address': None},
                '117DBBE': {       'dev_macs': ['14:21:03:92:f0:dd'],
                                   'ip_address': None,
                                   'name': '140 WINSTANLEY ST (BROWN ADAM)',
                                   'serial': '117DBBE',
                                   'serial_number': '117DBBE',
                                   'system_mac_address': None},
                '117DBC4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2435 LINE 47 (GERBER JAMES)',
                                   'serial': '117DBC4',
                                   'serial_number': '117DBC4',
                                   'system_mac_address': None},
                '117DBCA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2620 LINE 45 (SATCHELL DOUGLAS & BONNIE)',
                                   'serial': '117DBCA',
                                   'serial_number': '117DBCA',
                                   'system_mac_address': None},
                '117DBCD': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6033 PERTH ROAD 119 (GASCHO KEVIN & '
                                           'NADINE)',
                                   'serial': '117DBCD',
                                   'serial_number': '117DBCD',
                                   'system_mac_address': None},
                '117DBDB': {       'dev_macs': ['14:21:03:92:c6:d7'],
                                   'ip_address': None,
                                   'name': '143 PUGH ST (COURCHESNE BRIAN)',
                                   'serial': '117DBDB',
                                   'serial_number': '117DBDB',
                                   'system_mac_address': None},
                '117DBDF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2499 LINE 47 (ROTH JOHN & INGRID)',
                                   'serial': '117DBDF',
                                   'serial_number': '117DBDF',
                                   'system_mac_address': None},
                '117DBE0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5261 ROAD 108 (NEEBVIEW ACRES LTD.)',
                                   'serial': '117DBE0',
                                   'serial_number': '117DBE0',
                                   'system_mac_address': None},
                '117DBE7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4540 LINE 46 (RITSMA JOAN)',
                                   'serial': '117DBE7',
                                   'serial_number': '117DBE7',
                                   'system_mac_address': None},
                '117DBF1': {       'dev_macs': ['14:21:03:a1:e5:81'],
                                   'ip_address': None,
                                   'name': '66 PUGH ST (PEARCE DOUG)',
                                   'serial': '117DBF1',
                                   'serial_number': '117DBF1',
                                   'system_mac_address': None},
                '117DBF4': {       'dev_macs': ['14:21:03:8b:70:e7'],
                                   'ip_address': None,
                                   'name': '6B MAIN ST S (NEW ORLEANS PIZZA)',
                                   'serial': '117DBF4',
                                   'serial_number': '117DBF4',
                                   'system_mac_address': None},
                '117DBF6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4433 PERTH LINE 72 (HAFEMANN CHRISTIAN)',
                                   'serial': '117DBF6',
                                   'serial_number': '117DBF6',
                                   'system_mac_address': None},
                '117DC00': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4466 COMMUNITY CENTRE DR (JANTZI EMMANUEL)',
                                   'serial': '117DC00',
                                   'serial_number': '117DC00',
                                   'system_mac_address': None},
                '117DC03': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '220 WINSTANLEY ST MONKTON (HEWITT ROBERT)',
                                   'serial': '117DC03',
                                   'serial_number': '117DC03',
                                   'system_mac_address': None},
                '117DC09': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7006 ROAD 129 (WAGLER BRIAN)',
                                   'serial': '117DC09',
                                   'serial_number': '117DC09',
                                   'system_mac_address': None},
                '117DC0C': {       'dev_macs': ['14:21:03:92:ea:8c'],
                                   'ip_address': None,
                                   'name': '121 PUGH ST (STONE STEPHEN)',
                                   'serial': '117DC0C',
                                   'serial_number': '117DC0C',
                                   'system_mac_address': None},
                '117DC11': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4619 LINE 61 (WAGLER JONATHAN)',
                                   'serial': '117DC11',
                                   'serial_number': '117DC11',
                                   'system_mac_address': None},
                '117DC19': {       'dev_macs': ['14:21:03:92:ea:ce'],
                                   'ip_address': None,
                                   'name': '5264 ROAD 108 (ROTH TIM)',
                                   'serial': '117DC19',
                                   'serial_number': '117DC19',
                                   'system_mac_address': None},
                '117DC1A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4438 PERTH LINE 72 (POOLE TYLER)',
                                   'serial': '117DC1A',
                                   'serial_number': '117DC1A',
                                   'system_mac_address': None},
                '117DC1E': {       'dev_macs': ['5c:0f:fb:33:31:e6'],
                                   'ip_address': None,
                                   'name': '4446 PERTH LINE 72 (WAGLER SHIRLEY)',
                                   'serial': '117DC1E',
                                   'serial_number': '117DC1E',
                                   'system_mac_address': None},
                '117DC21': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4987 LINE 61 (STREICHER RON)',
                                   'serial': '117DC21',
                                   'serial_number': '117DC21',
                                   'system_mac_address': None},
                '117DC27': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2580 LINE 45 (KINGSBURY STEVE)',
                                   'serial': '117DC27',
                                   'serial_number': '117DC27',
                                   'system_mac_address': None},
                '117DC2A': {       'dev_macs': ['14:21:03:92:f4:6e'],
                                   'ip_address': None,
                                   'name': '4847 ROAD 109 (HORNE ARTHUR)',
                                   'serial': '117DC2A',
                                   'serial_number': '117DC2A',
                                   'system_mac_address': None},
                '1192F9B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5961 PERTH ROAD 119 (GASCHO ANDRE)',
                                   'serial': '1192F9B',
                                   'serial_number': '1192F9B',
                                   'system_mac_address': None},
                '1192FE9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '3996A LINE 61 (BARTSCH DYCK PETER)',
                                   'serial': '1192FE9',
                                   'serial_number': '1192FE9',
                                   'system_mac_address': None},
                '1192FF5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6005 PERTH ROAD 119 (GASCHO BRIAN)',
                                   'serial': '1192FF5',
                                   'serial_number': '1192FF5',
                                   'system_mac_address': None},
                '1192FF6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7002 PERTH ROAD 121 (KUEPFER JEREMY)',
                                   'serial': '1192FF6',
                                   'serial_number': '1192FF6',
                                   'system_mac_address': None},
                '119300B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6529 ROAD 129 (KUEPFER EDWARD)',
                                   'serial': '119300B',
                                   'serial_number': '119300B',
                                   'system_mac_address': None},
                '1193042': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6013 PERTH ROAD 119 (GEERTSMA PETER)',
                                   'serial': '1193042',
                                   'serial_number': '1193042',
                                   'system_mac_address': None},
                '11930CA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5911 PERTH ROAD 119 (YANTZI AMY)',
                                   'serial': '11930CA',
                                   'serial_number': '11930CA',
                                   'system_mac_address': None},
                '11945D1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4822 LINE 59 (GRAVELRIDGE WI HALL)',
                                   'serial': '11945D1',
                                   'serial_number': '11945D1',
                                   'system_mac_address': None},
                '1194A97': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '202 SELBY AV (TOM GOODYEAR)',
                                   'serial': '1194A97',
                                   'serial_number': '1194A97',
                                   'system_mac_address': None},
                '1194B8B': {       'dev_macs': ['84:39:8f:7e:b0:30'],
                                   'ip_address': None,
                                   'name': '4 COBALT ST N (AVON MAITLAND DISTRICT '
                                           'SCHOOL)',
                                   'serial': '1194B8B',
                                   'serial_number': '1194B8B',
                                   'system_mac_address': None},
                '1194D7D': {       'dev_macs': ['04:bc:9f:46:8e:86'],
                                   'ip_address': None,
                                   'name': '5640 PERTH LINE 55 (RAYCRAFT RANDY)',
                                   'serial': '1194D7D',
                                   'serial_number': '1194D7D',
                                   'system_mac_address': None},
                '1194E72': {       'dev_macs': [       '5c:0f:fb:3d:3f:5e',
                                                       '5c:0f:fb:43:96:e0'],
                                   'ip_address': None,
                                   'name': '5619A PERTH LINE 55 (MCEWEN BRIAN)',
                                   'serial': '1194E72',
                                   'serial_number': '1194E72',
                                   'system_mac_address': None},
                '1194E87': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6986 MILLBANK MAIN ST (KLASSEN JAKE & '
                                           'ANNA)',
                                   'serial': '1194E87',
                                   'serial_number': '1194E87',
                                   'system_mac_address': None},
                '1194E8C': {       'dev_macs': ['b8:94:70:f3:dd:4a'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (ROTH CAROLYN)',
                                   'serial': '1194E8C',
                                   'serial_number': '1194E8C',
                                   'system_mac_address': None},
                '11950FA': {       'dev_macs': ['cc:be:59:7e:b8:d2'],
                                   'ip_address': None,
                                   'name': '194 WINSTANLEY ST (THOM KENNEDY)',
                                   'serial': '11950FA',
                                   'serial_number': '11950FA',
                                   'system_mac_address': None},
                '1195105': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6906 ELGIN ST (ZEHR ALMA)',
                                   'serial': '1195105',
                                   'serial_number': '1195105',
                                   'system_mac_address': None},
                '1195111': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '78 PUGH ST (ENSINGER JASON)',
                                   'serial': '1195111',
                                   'serial_number': '1195111',
                                   'system_mac_address': None},
                '119513E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4095 WYNDHAM ST (SCHEERER JOHN)',
                                   'serial': '119513E',
                                   'serial_number': '119513E',
                                   'system_mac_address': None},
                '119513F': {       'dev_macs': ['04:bc:9f:46:8d:3c'],
                                   'ip_address': None,
                                   'name': '6985 RUTHERFORD ST (NIEWINSKI DAVE)',
                                   'serial': '119513F',
                                   'serial_number': '119513F',
                                   'system_mac_address': None},
                '1195143': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5262 LINE 52 (BVH EXCAVATION & '
                                           'CONTRACTING)',
                                   'serial': '1195143',
                                   'serial_number': '1195143',
                                   'system_mac_address': None},
                '1195144': {       'dev_macs': ['04:bc:9f:46:8f:f1'],
                                   'ip_address': None,
                                   'name': '5201 LINE 49 (NOON JOSHUA)',
                                   'serial': '1195144',
                                   'serial_number': '1195144',
                                   'system_mac_address': None},
                '1195146': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6983 MILLBANK MAIN ST (KROEKER ISAAK)',
                                   'serial': '1195146',
                                   'serial_number': '1195146',
                                   'system_mac_address': None},
                '119514D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '207 SELBY AV (REDEEMER LUTHERAN CHURCH)',
                                   'serial': '119514D',
                                   'serial_number': '119514D',
                                   'system_mac_address': None},
                '119515C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5479 ROAD 144 (DEGROOT CHRIS & ANGELA)',
                                   'serial': '119515C',
                                   'serial_number': '119515C',
                                   'system_mac_address': None},
                '1195161': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6972 MILLBANK MAIN STREET (MARTIN ROGER)',
                                   'serial': '1195161',
                                   'serial_number': '1195161',
                                   'system_mac_address': None},
                '1195169': {       'dev_macs': ['b8:94:70:9e:cb:d0'],
                                   'ip_address': None,
                                   'name': '95 PUGH ST (ROOPRAM ANTHONY)',
                                   'serial': '1195169',
                                   'serial_number': '1195169',
                                   'system_mac_address': None},
                '119516F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6944 PERTH ROAD 121 (BARTSCH JOHN & '
                                           'SUSANA)',
                                   'serial': '119516F',
                                   'serial_number': '119516F',
                                   'system_mac_address': None},
                '1195172': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6906 ELGIN ST (FRIESEN HELENA & JACOB)',
                                   'serial': '1195172',
                                   'serial_number': '1195172',
                                   'system_mac_address': None},
                '1195173': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '212 WINSTANLEY ST (HULLEY LAWRENCE)',
                                   'serial': '1195173',
                                   'serial_number': '1195173',
                                   'system_mac_address': None},
                '119517E': {       'dev_macs': ['b8:94:70:f3:df:f4'],
                                   'ip_address': None,
                                   'name': '6935 HORIZON TL (MCARTHUR DOROTHY & DOUG)',
                                   'serial': '119517E',
                                   'serial_number': '119517E',
                                   'system_mac_address': None},
                '1195187': {       'dev_macs': ['04:bc:9f:46:8d:31'],
                                   'ip_address': None,
                                   'name': '55 WILLIAM ST (MURAWSKY PAUL)',
                                   'serial': '1195187',
                                   'serial_number': '1195187',
                                   'system_mac_address': None},
                '119518D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4095 HILLCREST ST (FRIESEN HEINRICH & EVA)',
                                   'serial': '119518D',
                                   'serial_number': '119518D',
                                   'system_mac_address': None},
                '1195191': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "22 WILLIAM ST (O'MALLEY CAITLIN)",
                                   'serial': '1195191',
                                   'serial_number': '1195191',
                                   'system_mac_address': None},
                '1195199': {       'dev_macs': ['b8:94:70:f3:e4:f0'],
                                   'ip_address': None,
                                   'name': '6963 MILLBANK MAIN STREET (GOMES JOHN & '
                                           'NELIA)',
                                   'serial': '1195199',
                                   'serial_number': '1195199',
                                   'system_mac_address': None},
                '11951A5': {       'dev_macs': ['04:bc:9f:46:90:d8'],
                                   'ip_address': None,
                                   'name': '5657 PERTH LINE 55 (HAHN BRUCE)',
                                   'serial': '11951A5',
                                   'serial_number': '11951A5',
                                   'system_mac_address': None},
                '11951A7': {       'dev_macs': ['84:d3:43:6d:eb:e2'],
                                   'ip_address': None,
                                   'name': '14 MADDISON ST E (THIESSEN GERHARD)',
                                   'serial': '11951A7',
                                   'serial_number': '11951A7',
                                   'system_mac_address': None},
                '11951B0': {       'dev_macs': ['e4:6c:d1:4e:f0:5b'],
                                   'ip_address': None,
                                   'name': '2 MAIN ST N (WITMER TIM)',
                                   'serial': '11951B0',
                                   'serial_number': '11951B0',
                                   'system_mac_address': None},
                '11951B6': {       'dev_macs': ['04:bc:9f:46:ad:08'],
                                   'ip_address': None,
                                   'name': '129 PUGH ST (ANTHONY DAVE)',
                                   'serial': '11951B6',
                                   'serial_number': '11951B6',
                                   'system_mac_address': None},
                '11B227A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4072 QUEEN ST (THOMAS JIM)',
                                   'serial': '11B227A',
                                   'serial_number': '11B227A',
                                   'system_mac_address': None},
                '11B22BB': {       'dev_macs': ['04:bc:9f:46:93:fb'],
                                   'ip_address': None,
                                   'name': '6906 ELGIN ST (ROES SHARON)',
                                   'serial': '11B22BB',
                                   'serial_number': '11B22BB',
                                   'system_mac_address': None},
                '11B24F7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4075 PRINCESS ST (MISENER SCOTT)',
                                   'serial': '11B24F7',
                                   'serial_number': '11B24F7',
                                   'system_mac_address': None},
                '11B25D4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4863 LINE 59 (DAVIDSON TIM)',
                                   'serial': '11B25D4',
                                   'serial_number': '11B25D4',
                                   'system_mac_address': None},
                '11B25E5': {       'dev_macs': ['b8:94:70:84:99:cd'],
                                   'ip_address': None,
                                   'name': '5592 DUKE ST (COUNTRY COUSINS)',
                                   'serial': '11B25E5',
                                   'serial_number': '11B25E5',
                                   'system_mac_address': None},
                '11B25F8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6906 ELGIN ST (VAN ALLEN BARBIE)',
                                   'serial': '11B25F8',
                                   'serial_number': '11B25F8',
                                   'system_mac_address': None},
                '11B2612': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6906 ELGIN ST (VAN ALLEN TAUNYA)',
                                   'serial': '11B2612',
                                   'serial_number': '11B2612',
                                   'system_mac_address': None},
                '11BDFF4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '15 MAIN ST S (HASTINGS MANDY)',
                                   'serial': '11BDFF4',
                                   'serial_number': '11BDFF4',
                                   'system_mac_address': None},
                '11BE141': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6837 ROAD 129 (JANTZI RAYMOND)',
                                   'serial': '11BE141',
                                   'serial_number': '11BE141',
                                   'system_mac_address': None},
                '11BE15A': {       'dev_macs': ['14:21:03:a9:75:59'],
                                   'ip_address': None,
                                   'name': '2621 LINE 47 (WILSON JENNIFER)',
                                   'serial': '11BE15A',
                                   'serial_number': '11BE15A',
                                   'system_mac_address': None},
                '11BE1DC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4398 LINE 52 (WEBER DAN & TAMMY)',
                                   'serial': '11BE1DC',
                                   'serial_number': '11BE1DC',
                                   'system_mac_address': None},
                '11BE1EA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4430 PERTH LINE 72 NEWTON (A&A '
                                           'CONTRACTING)',
                                   'serial': '11BE1EA',
                                   'serial_number': '11BE1EA',
                                   'system_mac_address': None},
                '11BE301': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6841 ROAD 129 (REIMER ABRAM)',
                                   'serial': '11BE301',
                                   'serial_number': '11BE301',
                                   'system_mac_address': None},
                '11BE369': {       'dev_macs': ['14:21:03:d1:8a:51'],
                                   'ip_address': None,
                                   'name': '4436 PERTH LINE 72 (KUEPFER RAYMOND)',
                                   'serial': '11BE369',
                                   'serial_number': '11BE369',
                                   'system_mac_address': None},
                '11BE395': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4230 LINE 67 (KUEPFER ROSEANA)',
                                   'serial': '11BE395',
                                   'serial_number': '11BE395',
                                   'system_mac_address': None},
                '11BE3A2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '19 FULTON ST (TOTH DON)',
                                   'serial': '11BE3A2',
                                   'serial_number': '11BE3A2',
                                   'system_mac_address': None},
                '11D018': {       'dev_macs': ['48:a9:8a:22:67:6e'],
                                  'ip_address': None,
                                  'name': 'Peaceful Meadows Retirement (Office ONT)',
                                  'serial': '11D018',
                                  'serial_number': '11D018',
                                  'system_mac_address': None},
                '12483F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '3 Cobalt St. N. Milverton (Rick Banks)',
                                  'serial': '12483F',
                                  'serial_number': '12483F',
                                  'system_mac_address': None},
                '12484E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4837 Line 61 Milverton (Grace Mooney)',
                                  'serial': '12484E',
                                  'serial_number': '12484E',
                                  'system_mac_address': None},
                '1270D51': {       'dev_macs': ['e4:6c:d1:4e:f0:2f'],
                                   'ip_address': None,
                                   'name': '10 SCHADE ST (JARMUTH KEN & NORA)',
                                   'serial': '1270D51',
                                   'serial_number': '1270D51',
                                   'system_mac_address': None},
                '1270D6B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5897 PERTH ROAD 119 (WAGLER SARAH)',
                                   'serial': '1270D6B',
                                   'serial_number': '1270D6B',
                                   'system_mac_address': None},
                '1270D6E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '14 TEMPERANCE ST (DEBUS DAVID)',
                                   'serial': '1270D6E',
                                   'serial_number': '1270D6E',
                                   'system_mac_address': None},
                '1270D70': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4022 PERTH LINE 72 (KIPFER SAM S)',
                                   'serial': '1270D70',
                                   'serial_number': '1270D70',
                                   'system_mac_address': None},
                '1270D8E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2 MAIN ST N (COOPER JAMES)',
                                   'serial': '1270D8E',
                                   'serial_number': '1270D8E',
                                   'system_mac_address': None},
                '1270D92': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6905 PERTH ROAD 121 (WAGLER HAROLD)',
                                   'serial': '1270D92',
                                   'serial_number': '1270D92',
                                   'system_mac_address': None},
                '1270D9D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6771 PERTH ROAD 121 (KUEPFER SAM A)',
                                   'serial': '1270D9D',
                                   'serial_number': '1270D9D',
                                   'system_mac_address': None},
                '1270DE6': {       'dev_macs': [       '5c:0f:fb:2a:03:ea',
                                                       '5c:0f:fb:2a:13:e2'],
                                   'ip_address': None,
                                   'name': '5 GROSCH ST (DOUPE DOLORES)',
                                   'serial': '1270DE6',
                                   'serial_number': '1270DE6',
                                   'system_mac_address': None},
                '1270E05': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6484 ROAD 136 (COXON MRS LLOYD)',
                                   'serial': '1270E05',
                                   'serial_number': '1270E05',
                                   'system_mac_address': None},
                '1270E47': {       'dev_macs': ['e4:6c:d1:4e:f1:4d'],
                                   'ip_address': None,
                                   'name': '61 WILLIAM ST (ROTHWELL SARAH)',
                                   'serial': '1270E47',
                                   'serial_number': '1270E47',
                                   'system_mac_address': None},
                '1273117': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6600 PERTH ROAD 121 (SAUVE JOE)',
                                   'serial': '1273117',
                                   'serial_number': '1273117',
                                   'system_mac_address': None},
                '1273473': {       'dev_macs': ['60:db:98:09:1d:22'],
                                   'ip_address': None,
                                   'name': '82 GERBER DR (ANANTHAN SUBATHA)',
                                   'serial': '1273473',
                                   'serial_number': '1273473',
                                   'system_mac_address': None},
                '12738F3': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '86 MAIN ST N (FRIESEN SHAWN)',
                                   'serial': '12738F3',
                                   'serial_number': '12738F3',
                                   'system_mac_address': None},
                '12739D9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '55 WILLIAM ST (METERGY SOLUTIONS INC)',
                                   'serial': '12739D9',
                                   'serial_number': '12739D9',
                                   'system_mac_address': None},
                '1273AA6': {       'dev_macs': ['8c:90:2d:44:30:a8'],
                                   'ip_address': None,
                                   'name': '2390 LINE 47 (SAUDER CANDICE)',
                                   'serial': '1273AA6',
                                   'serial_number': '1273AA6',
                                   'system_mac_address': None},
                '1273ADF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4622 PERTH LINE 55 (EBERSOL RONALD)',
                                   'serial': '1273ADF',
                                   'serial_number': '1273ADF',
                                   'system_mac_address': None},
                '1273F55': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6238 ROAD 116 (NAFZIGER POULTRY INC)',
                                   'serial': '1273F55',
                                   'serial_number': '1273F55',
                                   'system_mac_address': None},
                '1273F6F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7098 PERTH ROAD 121 (MARTIN JAMES)',
                                   'serial': '1273F6F',
                                   'serial_number': '1273F6F',
                                   'system_mac_address': None},
                '1274029': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4133 LINE 67 (KUEPFER TREVOR)',
                                   'serial': '1274029',
                                   'serial_number': '1274029',
                                   'system_mac_address': None},
                '127402B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4186 LINE 61 (FAIRHAVEN AMISH MENNONITE '
                                           'CH)',
                                   'serial': '127402B',
                                   'serial_number': '127402B',
                                   'system_mac_address': None},
                '127402D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4155 LINE 67 (ERB KENNETH & LAURIE)',
                                   'serial': '127402D',
                                   'serial_number': '127402D',
                                   'system_mac_address': None},
                '1274037': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '15 MAIN ST S APT 203 (Psi)',
                                   'serial': '1274037',
                                   'serial_number': '1274037',
                                   'system_mac_address': None},
                '127403F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4227 LINE 50 (BRUNK MELVIN)',
                                   'serial': '127403F',
                                   'serial_number': '127403F',
                                   'system_mac_address': None},
                '1274044': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4019 LINE 67 (KUEPFER JUDY)',
                                   'serial': '1274044',
                                   'serial_number': '1274044',
                                   'system_mac_address': None},
                '127404B': {       'dev_macs': ['b8:94:70:f3:e1:28'],
                                   'ip_address': None,
                                   'name': '4742 LINE 64 (BIEHN LOWELL)',
                                   'serial': '127404B',
                                   'serial_number': '127404B',
                                   'system_mac_address': None},
                '127404C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7034 PERTH ROAD 121 (BAUMAN MURRAY)',
                                   'serial': '127404C',
                                   'serial_number': '127404C',
                                   'system_mac_address': None},
                '1274050': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6455 ROAD 140 (STREICHER MERLIN)',
                                   'serial': '1274050',
                                   'serial_number': '1274050',
                                   'system_mac_address': None},
                '1274054': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '13 MAIN ST N (DYCK BENJAMIN)',
                                   'serial': '1274054',
                                   'serial_number': '1274054',
                                   'system_mac_address': None},
                '127405B': {       'dev_macs': [       '5c:0f:fb:43:96:ba',
                                                       '5c:0f:fb:3d:3e:b8'],
                                   'ip_address': None,
                                   'name': '5619B PERTH LINE 55 (MCEWEN R DOUGLAS)',
                                   'serial': '127405B',
                                   'serial_number': '127405B',
                                   'system_mac_address': None},
                '127405F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '69 PUGH ST MILVERTON (SHANTZ TIFFANY)',
                                   'serial': '127405F',
                                   'serial_number': '127405F',
                                   'system_mac_address': None},
                '1274061': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7020 PERTH ROAD 121 (BARBU THOMAS & ROSE)',
                                   'serial': '1274061',
                                   'serial_number': '1274061',
                                   'system_mac_address': None},
                '1274064': {       'dev_macs': ['7c:5a:1c:8b:fa:33'],
                                   'ip_address': None,
                                   'name': '4078 WATERLOO ST (PRECISION WOODCRAFT)',
                                   'serial': '1274064',
                                   'serial_number': '1274064',
                                   'system_mac_address': None},
                '1274067': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7035 PERTH ROAD 121 (SCHULTZ DELMAR & '
                                           'JANET)',
                                   'serial': '1274067',
                                   'serial_number': '1274067',
                                   'system_mac_address': None},
                '127406D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '70 MAIN ST S (1000669244ONTARIO)',
                                   'serial': '127406D',
                                   'serial_number': '127406D',
                                   'system_mac_address': None},
                '1274070': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '33 MURRAY CT (SOUTHSIDE ESTATES INC)',
                                   'serial': '1274070',
                                   'serial_number': '1274070',
                                   'system_mac_address': None},
                '1274079': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4067 LINE 61 (VAN DEN TEMPEL TED & LIZ)',
                                   'serial': '1274079',
                                   'serial_number': '1274079',
                                   'system_mac_address': None},
                '127407A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '55 WILLIAM ST (OWUSU SELINA)',
                                   'serial': '127407A',
                                   'serial_number': '127407A',
                                   'system_mac_address': None},
                '1274081': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4710 LINE 64 (MUELLER PAUL)',
                                   'serial': '1274081',
                                   'serial_number': '1274081',
                                   'system_mac_address': None},
                '1274090': {       'dev_macs': ['b8:94:70:f3:e8:76'],
                                   'ip_address': None,
                                   'name': '62 PUGH ST (LAUFER PETER)',
                                   'serial': '1274090',
                                   'serial_number': '1274090',
                                   'system_mac_address': None},
                '1274097': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6346 PERTH ROAD 121 (ZEHR GERALD & ALICE '
                                           'BARN)',
                                   'serial': '1274097',
                                   'serial_number': '1274097',
                                   'system_mac_address': None},
                '12740A0': {       'dev_macs': ['60:db:98:09:1c:d5'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (IONCE ANCA)',
                                   'serial': '12740A0',
                                   'serial_number': '12740A0',
                                   'system_mac_address': None},
                '12740A1': {       'dev_macs': ['14:21:03:8b:68:bd'],
                                   'ip_address': None,
                                   'name': '4575 PERTH LINE 72 (WYATT JUDY)',
                                   'serial': '12740A1',
                                   'serial_number': '12740A1',
                                   'system_mac_address': None},
                '12740A5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4360 LINE 67 (MARTIN AARON F)',
                                   'serial': '12740A5',
                                   'serial_number': '12740A5',
                                   'system_mac_address': None},
                '12740B4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4224 LINE 61 (CARTER ALVIN)',
                                   'serial': '12740B4',
                                   'serial_number': '12740B4',
                                   'system_mac_address': None},
                '12740B8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '13 MAIN ST N APT 6 MILVERTON',
                                   'serial': '12740B8',
                                   'serial_number': '12740B8',
                                   'system_mac_address': None},
                '12740BD': {       'dev_macs': ['60:db:98:09:39:c0'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (POLANIC ADRIANA)',
                                   'serial': '12740BD',
                                   'serial_number': '12740BD',
                                   'system_mac_address': None},
                '12740BF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6672 ROAD 129 (KUEPFER PAUL)',
                                   'serial': '12740BF',
                                   'serial_number': '12740BF',
                                   'system_mac_address': None},
                '12740C6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4913 ROAD 122 (PLAYERDISE FARMS LTD)',
                                   'serial': '12740C6',
                                   'serial_number': '12740C6',
                                   'system_mac_address': None},
                '12740C8': {       'dev_macs': ['00:03:e6:fa:10:50'],
                                   'ip_address': None,
                                   'name': '7110 PERTH ROAD 121 (ZAWADZKI MAREK)',
                                   'serial': '12740C8',
                                   'serial_number': '12740C8',
                                   'system_mac_address': None},
                '12740C9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4192 LINE 61 (RAYNER WILLIAM)',
                                   'serial': '12740C9',
                                   'serial_number': '12740C9',
                                   'system_mac_address': None},
                '12740DD': {       'dev_macs': ['44:65:7f:88:d2:67'],
                                   'ip_address': None,
                                   'name': '5225 ROAD 130 (STEPHENSON BOB)',
                                   'serial': '12740DD',
                                   'serial_number': '12740DD',
                                   'system_mac_address': None},
                '12740EB': {       'dev_macs': ['44:65:7f:72:7d:cd'],
                                   'ip_address': None,
                                   'name': '81 Reagan St Milverton (K Saysangkhy '
                                           '104355220)',
                                   'serial': '12740EB',
                                   'serial_number': '12740EB',
                                   'system_mac_address': None},
                '12740ED': {       'dev_macs': ['14:21:03:8b:70:6e'],
                                   'ip_address': None,
                                   'name': '2288 LINE 45 (MCMILLAN ROBERT)',
                                   'serial': '12740ED',
                                   'serial_number': '12740ED',
                                   'system_mac_address': None},
                '127411B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4773 LINE 64 (BYLER JAMES)',
                                   'serial': '127411B',
                                   'serial_number': '127411B',
                                   'system_mac_address': None},
                '1274125': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '17 WEST ST (ALLEN JIM)',
                                   'serial': '1274125',
                                   'serial_number': '1274125',
                                   'system_mac_address': None},
                '127418D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4784 LINE 64 (FABER JOHN)',
                                   'serial': '127418D',
                                   'serial_number': '127418D',
                                   'system_mac_address': None},
                '127418F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6940 HORIZON TRAIL (KUEPFER CALVIN R)',
                                   'serial': '127418F',
                                   'serial_number': '127418F',
                                   'system_mac_address': None},
                '1274190': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '78 MAIN ST N (KUEPFER JOHN & EILEEN)',
                                   'serial': '1274190',
                                   'serial_number': '1274190',
                                   'system_mac_address': None},
                '1274193': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4223 LINE 61 (DIETRICH JESSE & JENNA)',
                                   'serial': '1274193',
                                   'serial_number': '1274193',
                                   'system_mac_address': None},
                '1274196': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4165 LINE 50 (JANTZI ROSS)',
                                   'serial': '1274196',
                                   'serial_number': '1274196',
                                   'system_mac_address': None},
                '127419A': {       'dev_macs': [       '5c:0f:fb:2a:c5:00',
                                                       '5c:0f:fb:2a:c7:32'],
                                   'ip_address': None,
                                   'name': '148 MADDISON ST E MONKTON (WICK GARY)',
                                   'serial': '127419A',
                                   'serial_number': '127419A',
                                   'system_mac_address': None},
                '127419D': {       'dev_macs': ['b8:94:70:f3:e0:15'],
                                   'ip_address': None,
                                   'name': '105 PUGH ST (HEIDENREICH MIKE & JEANETTE)',
                                   'serial': '127419D',
                                   'serial_number': '127419D',
                                   'system_mac_address': None},
                '127419E': {       'dev_macs': ['14:21:03:8b:7a:24'],
                                   'ip_address': None,
                                   'name': '113 PUGH ST (MAYNE STEPH)',
                                   'serial': '127419E',
                                   'serial_number': '127419E',
                                   'system_mac_address': None},
                '127419F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6700 ROAD 129 (JANTZI DOUG)',
                                   'serial': '127419F',
                                   'serial_number': '127419F',
                                   'system_mac_address': None},
                '12741A0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6390 PERTH ROAD 121 (BUECKERT HENRY & '
                                           'TINA)',
                                   'serial': '12741A0',
                                   'serial_number': '12741A0',
                                   'system_mac_address': None},
                '12741A2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7285 PERTH ROAD 121 (ALBRECHT LLOYD)',
                                   'serial': '12741A2',
                                   'serial_number': '12741A2',
                                   'system_mac_address': None},
                '12741AF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6475 PERTH ROAD 121 (POOLE MENNONITE '
                                           'CHURCH)',
                                   'serial': '12741AF',
                                   'serial_number': '12741AF',
                                   'system_mac_address': None},
                '12741B0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4187 LINE 61 (ZEHR TIM & LILLIAN)',
                                   'serial': '12741B0',
                                   'serial_number': '12741B0',
                                   'system_mac_address': None},
                '12741BB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7401 PERTH ROAD 121 (SAUDER MERVIN)',
                                   'serial': '12741BB',
                                   'serial_number': '12741BB',
                                   'system_mac_address': None},
                '12741BC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7033 PERTH ROAD 121 (SCHELLENBERGER '
                                           'FRIEDA)',
                                   'serial': '12741BC',
                                   'serial_number': '12741BC',
                                   'system_mac_address': None},
                '12741BF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4410 LINE 61 (KUEPFER VERNON & DARLENE)',
                                   'serial': '12741BF',
                                   'serial_number': '12741BF',
                                   'system_mac_address': None},
                '12741CD': {       'dev_macs': ['6c:63:f8:39:1e:d7'],
                                   'ip_address': None,
                                   'name': '4199 LINE 61 MILVERTON (BOSHART '
                                           'INDUSTRIES)',
                                   'serial': '12741CD',
                                   'serial_number': '12741CD',
                                   'system_mac_address': None},
                '12741D0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5346 PERTH ROAD 135 (SANDY PEAK DAIRY LTD)',
                                   'serial': '12741D0',
                                   'serial_number': '12741D0',
                                   'system_mac_address': None},
                '12741D2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4560 PERTH LINE 72 (KUEPFER LESTER)',
                                   'serial': '12741D2',
                                   'serial_number': '12741D2',
                                   'system_mac_address': None},
                '1274214': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6345 PERTH ROAD 121 (VAN DEN TEMPEL '
                                           'LAUREN)',
                                   'serial': '1274214',
                                   'serial_number': '1274214',
                                   'system_mac_address': None},
                '1274217': {       'dev_macs': ['60:db:98:09:1f:f8'],
                                   'ip_address': None,
                                   'name': '4093 LINE 61 (YODER LYNFORD)',
                                   'serial': '1274217',
                                   'serial_number': '1274217',
                                   'system_mac_address': None},
                '1274220': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7446 PERTH ROAD 121 (BURKHART WAYNE)',
                                   'serial': '1274220',
                                   'serial_number': '1274220',
                                   'system_mac_address': None},
                '1274298': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7092 PERTH ROAD 121 (WEBER SARAH & EDNA)',
                                   'serial': '1274298',
                                   'serial_number': '1274298',
                                   'system_mac_address': None},
                '127429B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7212 PERTH ROAD 121 (BURKHART JAMES M)',
                                   'serial': '127429B',
                                   'serial_number': '127429B',
                                   'system_mac_address': None},
                '12742A5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4896 PERTH LINE 72 (CARTER MARY)',
                                   'serial': '12742A5',
                                   'serial_number': '12742A5',
                                   'system_mac_address': None},
                '12742A8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '155 WINSTANLEY ST MONKTON (SIMPSON DENNIS)',
                                   'serial': '12742A8',
                                   'serial_number': '12742A8',
                                   'system_mac_address': None},
                '12742D2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6346 PERTH ROAD 121 (ZEHR GERALD & ALICE '
                                           'HOUSE)',
                                   'serial': '12742D2',
                                   'serial_number': '12742D2',
                                   'system_mac_address': None},
                '12742D4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4010 LINE 67 (KUEPFER PAUL A & MARY)',
                                   'serial': '12742D4',
                                   'serial_number': '12742D4',
                                   'system_mac_address': None},
                '12742D9': {       'dev_macs': ['04:bc:9f:46:ab:f5'],
                                   'ip_address': None,
                                   'name': '4111 LINE 73 (MORALES HUGO)',
                                   'serial': '12742D9',
                                   'serial_number': '12742D9',
                                   'system_mac_address': None},
                '12742DB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4375 PERTH LINE 72 (FRIESEN ARNO)',
                                   'serial': '12742DB',
                                   'serial_number': '12742DB',
                                   'system_mac_address': None},
                '12742E0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '3994 LINE 61 (VAN NES JAMES)',
                                   'serial': '12742E0',
                                   'serial_number': '12742E0',
                                   'system_mac_address': None},
                '12742E4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7272 PERTH ROAD 121 (MARTIN BENJAMIN)',
                                   'serial': '12742E4',
                                   'serial_number': '12742E4',
                                   'system_mac_address': None},
                '12742E7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '80 MAIN ST N (GOFTON LAURA)',
                                   'serial': '12742E7',
                                   'serial_number': '12742E7',
                                   'system_mac_address': None},
                '12742ED': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4084 LINE 50 (MANTLE BRUCE)',
                                   'serial': '12742ED',
                                   'serial_number': '12742ED',
                                   'system_mac_address': None},
                '12742F5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7082 PERTH ROAD 121 (MORALES MAYDA)',
                                   'serial': '12742F5',
                                   'serial_number': '12742F5',
                                   'system_mac_address': None},
                '12742F6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7452 PERTH ROAD 121 (FERNBANK PAROCHIAL '
                                           'SCHOOL)',
                                   'serial': '12742F6',
                                   'serial_number': '12742F6',
                                   'system_mac_address': None},
                '12742F8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7007 ROAD 124 (LEIS ROBERT)',
                                   'serial': '12742F8',
                                   'serial_number': '12742F8',
                                   'system_mac_address': None},
                '1274304': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4229 LINE 61 (ZEHR SAWYER)',
                                   'serial': '1274304',
                                   'serial_number': '1274304',
                                   'system_mac_address': None},
                '1274309': {       'dev_macs': ['14:21:03:8b:6a:07'],
                                   'ip_address': None,
                                   'name': '6194 ROAD 116 (WAGLER RUTH)',
                                   'serial': '1274309',
                                   'serial_number': '1274309',
                                   'system_mac_address': None},
                '127430C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4144 LINE 61 (KUEPFER FLOYD)',
                                   'serial': '127430C',
                                   'serial_number': '127430C',
                                   'system_mac_address': None},
                '127430F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6799 ROAD 129 (JANTZI LEVI K)',
                                   'serial': '127430F',
                                   'serial_number': '127430F',
                                   'system_mac_address': None},
                '1274317': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7374 PERTH ROAD 121 (BAUMAN DEAN)',
                                   'serial': '1274317',
                                   'serial_number': '1274317',
                                   'system_mac_address': None},
                '1274320': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4234 LINE 46 (CATANIA JACKIE)',
                                   'serial': '1274320',
                                   'serial_number': '1274320',
                                   'system_mac_address': None},
                '1274323': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6394 PERTH ROAD 121 (YANTZI AMY)',
                                   'serial': '1274323',
                                   'serial_number': '1274323',
                                   'system_mac_address': None},
                '1274329': {       'dev_macs': ['d0:76:8f:1f:53:52'],
                                   'ip_address': None,
                                   'name': '2 EDWARD ST (HOLST NELSON)',
                                   'serial': '1274329',
                                   'serial_number': '1274329',
                                   'system_mac_address': None},
                '1274333': {       'dev_macs': ['14:21:03:8b:6f:66'],
                                   'ip_address': None,
                                   'name': '5964 PERTH ROAD 119 (GERBER JOSEPH)',
                                   'serial': '1274333',
                                   'serial_number': '1274333',
                                   'system_mac_address': None},
                '1274335': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '113 MILL ST E (KIPFER REX)',
                                   'serial': '1274335',
                                   'serial_number': '1274335',
                                   'system_mac_address': None},
                '1274336': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7027 PERTH ROAD 121 (TERRY-ROSS ROBIN)',
                                   'serial': '1274336',
                                   'serial_number': '1274336',
                                   'system_mac_address': None},
                '127433A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4184 LINE 61 (FAIRHAVEN CH DAY SCHOOL)',
                                   'serial': '127433A',
                                   'serial_number': '127433A',
                                   'system_mac_address': None},
                '1274341': {       'dev_macs': ['44:65:7f:89:0e:73'],
                                   'ip_address': None,
                                   'name': '5 MAIN ST S (GERBER CASSIDY)',
                                   'serial': '1274341',
                                   'serial_number': '1274341',
                                   'system_mac_address': None},
                '12743E4': {       'dev_macs': ['14:21:03:92:ec:de'],
                                   'ip_address': None,
                                   'name': '13 PIERSON ST (HENRY SHAWN)',
                                   'serial': '12743E4',
                                   'serial_number': '12743E4',
                                   'system_mac_address': None},
                '12744B2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2339 LINE 45 (HART DAWN)',
                                   'serial': '12744B2',
                                   'serial_number': '12744B2',
                                   'system_mac_address': None},
                '127458F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5083 LINE 60 (GINGRICH TONY)',
                                   'serial': '127458F',
                                   'serial_number': '127458F',
                                   'system_mac_address': None},
                '1274593': {       'dev_macs': ['14:21:03:92:ed:2b'],
                                   'ip_address': None,
                                   'name': '5454 ROAD 126 (NEFF ALLEN & IRMA)',
                                   'serial': '1274593',
                                   'serial_number': '1274593',
                                   'system_mac_address': None},
                '12745D9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4107 LINE 61 (THIESSEN JACOB)',
                                   'serial': '12745D9',
                                   'serial_number': '12745D9',
                                   'system_mac_address': None},
                '12745FC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4460B PERTH LINE 72',
                                   'serial': '12745FC',
                                   'serial_number': '12745FC',
                                   'system_mac_address': None},
                '127460E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7 ST MARYS ST N (GROPP KEN (ESTATE OF))',
                                   'serial': '127460E',
                                   'serial_number': '127460E',
                                   'system_mac_address': None},
                '1274681': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4460B PERTH LINE 72 (PETERS WILHELM & '
                                           'ELIZABETH)',
                                   'serial': '1274681',
                                   'serial_number': '1274681',
                                   'system_mac_address': None},
                '127469C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7477 PERTH ROAD 121 (SAUDER ORVIE)',
                                   'serial': '127469C',
                                   'serial_number': '127469C',
                                   'system_mac_address': None},
                '12746BB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '14 ST MARYS ST S (KERSTEN THEO)',
                                   'serial': '12746BB',
                                   'serial_number': '12746BB',
                                   'system_mac_address': None},
                '12746CA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4372 LINE 61 (ALBRECHT JOHN)',
                                   'serial': '12746CA',
                                   'serial_number': '12746CA',
                                   'system_mac_address': None},
                '12746CD': {       'dev_macs': ['14:21:03:d1:7e:33'],
                                   'ip_address': None,
                                   'name': '4578 LINE 46 (PETERSON MONTGOMERY)',
                                   'serial': '12746CD',
                                   'serial_number': '12746CD',
                                   'system_mac_address': None},
                '12746D0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4429 PERTH LINE 72 (TOMIUK TIM)',
                                   'serial': '12746D0',
                                   'serial_number': '12746D0',
                                   'system_mac_address': None},
                '12746E1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4113 LINE 61 (JANTZI GABRIEL)',
                                   'serial': '12746E1',
                                   'serial_number': '12746E1',
                                   'system_mac_address': None},
                '1274702': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '124 WINSTANLEY ST (COOK TAMMY)',
                                   'serial': '1274702',
                                   'serial_number': '1274702',
                                   'system_mac_address': None},
                '1274711': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4444 PERTH LINE 72 (MARTIN CHRIS)',
                                   'serial': '1274711',
                                   'serial_number': '1274711',
                                   'system_mac_address': None},
                '1274722': {       'dev_macs': ['b8:94:70:9e:cd:3b'],
                                   'ip_address': None,
                                   'name': '4056 LINE 46 (ROES MELVIN)',
                                   'serial': '1274722',
                                   'serial_number': '1274722',
                                   'system_mac_address': None},
                '127473A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5595 ROAD 126 (ROTH BRENT)',
                                   'serial': '127473A',
                                   'serial_number': '127473A',
                                   'system_mac_address': None},
                '127473D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6019 PERTH ROAD 119 (GERLACH CONNIE)',
                                   'serial': '127473D',
                                   'serial_number': '127473D',
                                   'system_mac_address': None},
                '1274744': {       'dev_macs': ['84:39:8f:ed:41:5e'],
                                   'ip_address': None,
                                   'name': '42 AVERY PL (VAN NIEROP DAVE)',
                                   'serial': '1274744',
                                   'serial_number': '1274744',
                                   'system_mac_address': None},
                '1274761': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7464 PERTH ROAD 121 (ALBRECHT EDWARD)',
                                   'serial': '1274761',
                                   'serial_number': '1274761',
                                   'system_mac_address': None},
                '127478A': {       'dev_macs': ['14:21:03:d1:8c:82'],
                                   'ip_address': None,
                                   'name': '20 ARENA DR (NICHOLSON JOHN)',
                                   'serial': '127478A',
                                   'serial_number': '127478A',
                                   'system_mac_address': None},
                '127478D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4006 LINE 61 (STREICHER LAVERNE)',
                                   'serial': '127478D',
                                   'serial_number': '127478D',
                                   'system_mac_address': None},
                '127478F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6477A PERTH ROAD 121 (KUEPFER CLAYTON J '
                                           'AND SHARON)',
                                   'serial': '127478F',
                                   'serial_number': '127478F',
                                   'system_mac_address': None},
                '1274798': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6752 PERTH ROAD 121 (NITHVALLEY FARMS)',
                                   'serial': '1274798',
                                   'serial_number': '1274798',
                                   'system_mac_address': None},
                '127479A': {       'dev_macs': ['14:21:03:d1:89:07'],
                                   'ip_address': None,
                                   'name': '57 WILLIAM ST (EVANS HEATHER)',
                                   'serial': '127479A',
                                   'serial_number': '127479A',
                                   'system_mac_address': None},
                '12747A1': {       'dev_macs': ['d0:76:8f:1f:58:85'],
                                   'ip_address': None,
                                   'name': '34 MAIN ST S MILVERTON (FOSTER ANNE)',
                                   'serial': '12747A1',
                                   'serial_number': '12747A1',
                                   'system_mac_address': None},
                '12747A6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '14 CRUICKSHANK ST (TIGANI JAMES)',
                                   'serial': '12747A6',
                                   'serial_number': '12747A6',
                                   'system_mac_address': None},
                '12747AF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7014 PERTH ROAD 121 (KUEPFER WAYNE)',
                                   'serial': '12747AF',
                                   'serial_number': '12747AF',
                                   'system_mac_address': None},
                '12747B8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5977 PERTH ROAD 119 (JANTZI BRIAN & BEV)',
                                   'serial': '12747B8',
                                   'serial_number': '12747B8',
                                   'system_mac_address': None},
                '12747B9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4346 LINE 64 (JANTZI JASON & ROCHELLE)',
                                   'serial': '12747B9',
                                   'serial_number': '12747B9',
                                   'system_mac_address': None},
                '12747C1': {       'dev_macs': ['60:db:98:57:b5:0e'],
                                   'ip_address': None,
                                   'name': '5001 ROAD 109 (ROSE JOANNE)',
                                   'serial': '12747C1',
                                   'serial_number': '12747C1',
                                   'system_mac_address': None},
                '12747C4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4785 LINE 67 (WAGLER WELDING)',
                                   'serial': '12747C4',
                                   'serial_number': '12747C4',
                                   'system_mac_address': None},
                '12747C6': {       'dev_macs': ['14:21:03:d1:8a:d5'],
                                   'ip_address': None,
                                   'name': '5250 LINE 60 (CARTER JORDAN)',
                                   'serial': '12747C6',
                                   'serial_number': '12747C6',
                                   'system_mac_address': None},
                '12747D6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7026 PERTH ROAD 121 (SCHMIDT KATHRYN)',
                                   'serial': '12747D6',
                                   'serial_number': '12747D6',
                                   'system_mac_address': None},
                '12747E1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4951 LINE 67 (KUEPFER MANUEL & IRENE)',
                                   'serial': '12747E1',
                                   'serial_number': '12747E1',
                                   'system_mac_address': None},
                '12747FB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4462 PERTH LINE 72',
                                   'serial': '12747FB',
                                   'serial_number': '12747FB',
                                   'system_mac_address': None},
                '12747FF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4513 LINE 49 (WAGLER MURRAY)',
                                   'serial': '12747FF',
                                   'serial_number': '12747FF',
                                   'system_mac_address': None},
                '1274800': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6023 PERTH ROAD 119 (ALBRECHT ARNOLD)',
                                   'serial': '1274800',
                                   'serial_number': '1274800',
                                   'system_mac_address': None},
                '1274801': {       'dev_macs': ['44:65:7f:3e:b2:36'],
                                   'ip_address': None,
                                   'name': '4 TEMPERANCE ST (BULMER STUART)',
                                   'serial': '1274801',
                                   'serial_number': '1274801',
                                   'system_mac_address': None},
                '127481E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '17 PIERSON ST (MORNINGTON COMMUNICATIONS)',
                                   'serial': '127481E',
                                   'serial_number': '127481E',
                                   'system_mac_address': None},
                '1274829': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6908 PERTH ROAD 121 (DJ KUEPFER & FAMILY '
                                           'INC)',
                                   'serial': '1274829',
                                   'serial_number': '1274829',
                                   'system_mac_address': None},
                '1274844': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4003 LINE 61 (JANTZI ALLAN)',
                                   'serial': '1274844',
                                   'serial_number': '1274844',
                                   'system_mac_address': None},
                '127484E': {       'dev_macs': ['14:21:03:d1:8c:a3'],
                                   'ip_address': None,
                                   'name': '4042 WATERLOO ST (PENNER JOHAN)',
                                   'serial': '127484E',
                                   'serial_number': '127484E',
                                   'system_mac_address': None},
                '127485C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7008 ROAD 129 (NEWTON IRON INC)',
                                   'serial': '127485C',
                                   'serial_number': '127485C',
                                   'system_mac_address': None},
                '127485D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6800 ROAD 129 (KUEPFER CARL A)',
                                   'serial': '127485D',
                                   'serial_number': '127485D',
                                   'system_mac_address': None},
                '1274862': {       'dev_macs': ['48:77:46:d2:79:8e'],
                                   'ip_address': None,
                                   'name': '38 AVERY PL (ALENCAR ISABEL)',
                                   'serial': '1274862',
                                   'serial_number': '1274862',
                                   'system_mac_address': None},
                '1274865': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4300 LINE 61 (KUEPFER MERVIN)',
                                   'serial': '1274865',
                                   'serial_number': '1274865',
                                   'system_mac_address': None},
                '1274869': {       'dev_macs': ['ec:4f:82:ac:ed:66'],
                                   'ip_address': None,
                                   'name': '10 SPENCER ST (AW MILLWRIGHTS)',
                                   'serial': '1274869',
                                   'serial_number': '1274869',
                                   'system_mac_address': None},
                '127486F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '8 GERBER DR (KNIPE DOROTHY)',
                                   'serial': '127486F',
                                   'serial_number': '127486F',
                                   'system_mac_address': None},
                '1274871': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4852 ROAD 109 (COOK CHRIS)',
                                   'serial': '1274871',
                                   'serial_number': '1274871',
                                   'system_mac_address': None},
                '1274874': {       'dev_macs': ['14:21:03:d1:8a:30'],
                                   'ip_address': None,
                                   'name': '4820 ROAD 108 (SCHULER KEVIN)',
                                   'serial': '1274874',
                                   'serial_number': '1274874',
                                   'system_mac_address': None},
                '1274875': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5445 ROAD 108 (LUCKHARDT SHELLEY)',
                                   'serial': '1274875',
                                   'serial_number': '1274875',
                                   'system_mac_address': None},
                '127488B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6476 PERTH ROAD 121 (ROPP WAYNE)',
                                   'serial': '127488B',
                                   'serial_number': '127488B',
                                   'system_mac_address': None},
                '1274893': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '141 PUGH ST (HULLEY CRYSTAL)',
                                   'serial': '1274893',
                                   'serial_number': '1274893',
                                   'system_mac_address': None},
                '1274895': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "7022 ROAD 129 (THIESSEN BROTHER'S CUSTOM "
                                           'WOOD)',
                                   'serial': '1274895',
                                   'serial_number': '1274895',
                                   'system_mac_address': None},
                '12748A1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4943 ROAD 108 (COOK FRANCES)',
                                   'serial': '12748A1',
                                   'serial_number': '12748A1',
                                   'system_mac_address': None},
                '12748B1': {       'dev_macs': ['14:21:03:a9:70:7e'],
                                   'ip_address': None,
                                   'name': '6475 ROAD 116 (SCHULTZ WILLIAM)',
                                   'serial': '12748B1',
                                   'serial_number': '12748B1',
                                   'system_mac_address': None},
                '1274913': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6662 ROAD 116 (SCHMIDT TIM)',
                                   'serial': '1274913',
                                   'serial_number': '1274913',
                                   'system_mac_address': None},
                '1274917': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5384 PERTH LINE 55 (PARTRIDGE RYAN)',
                                   'serial': '1274917',
                                   'serial_number': '1274917',
                                   'system_mac_address': None},
                '1274946': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4348 PERTH LINE 72 (BAUMAN JOSEPH M)',
                                   'serial': '1274946',
                                   'serial_number': '1274946',
                                   'system_mac_address': None},
                '127495E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4094A LINE 61 (STECKLY SHEILA)',
                                   'serial': '127495E',
                                   'serial_number': '127495E',
                                   'system_mac_address': None},
                '127496F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7 GEORGE ST (WAGLER CATHERINE)',
                                   'serial': '127496F',
                                   'serial_number': '127496F',
                                   'system_mac_address': None},
                '1274987': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '3984 LINE 67 (ZEHR TERRY & SUE)',
                                   'serial': '1274987',
                                   'serial_number': '1274987',
                                   'system_mac_address': None},
                '1274993': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4087 LINE 61 (GOOD DWIGHT)',
                                   'serial': '1274993',
                                   'serial_number': '1274993',
                                   'system_mac_address': None},
                '1274997': {       'dev_macs': ['b8:94:70:f3:e9:1b'],
                                   'ip_address': None,
                                   'name': '108 PUGH ST (SANGWA MARIE)',
                                   'serial': '1274997',
                                   'serial_number': '1274997',
                                   'system_mac_address': None},
                '127499C': {       'dev_macs': ['b8:94:70:f3:e2:7d'],
                                   'ip_address': None,
                                   'name': '4011 LINE 61 (BANKS BRETT)',
                                   'serial': '127499C',
                                   'serial_number': '127499C',
                                   'system_mac_address': None},
                '127499F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6143 ANA ST (NAFZIGER CHRIS)',
                                   'serial': '127499F',
                                   'serial_number': '127499F',
                                   'system_mac_address': None},
                '12749A5': {       'dev_macs': ['d0:76:8f:91:d5:a3'],
                                   'ip_address': None,
                                   'name': '101 PUGH ST MILVERTON (HENHOEFFER BONNIE)',
                                   'serial': '12749A5',
                                   'serial_number': '12749A5',
                                   'system_mac_address': None},
                '12749A7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6404 PERTH ROAD 121 (ROTH IRENE I)',
                                   'serial': '12749A7',
                                   'serial_number': '12749A7',
                                   'system_mac_address': None},
                '12749A8': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '13 MAIN ST N (DURAN PEREZ SERGIO TOMAS)',
                                   'serial': '12749A8',
                                   'serial_number': '12749A8',
                                   'system_mac_address': None},
                '12749B3': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6755 ROAD 116 (JANTZI GERALD W)',
                                   'serial': '12749B3',
                                   'serial_number': '12749B3',
                                   'system_mac_address': None},
                '12749B4': {       'dev_macs': ['b8:94:70:f3:e7:42'],
                                   'ip_address': None,
                                   'name': '5913 PERTH ROAD 119 (THOMPSON STACY)',
                                   'serial': '12749B4',
                                   'serial_number': '12749B4',
                                   'system_mac_address': None},
                '12749C1': {       'dev_macs': ['14:21:03:8b:69:6d'],
                                   'ip_address': None,
                                   'name': '7116 PERTH ROAD 121 (KUEPFER GARY & MARY '
                                           'LOU)',
                                   'serial': '12749C1',
                                   'serial_number': '12749C1',
                                   'system_mac_address': None},
                '12749C3': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6799 ROAD 129 (JANTZI PERRY & KRISTA)',
                                   'serial': '12749C3',
                                   'serial_number': '12749C3',
                                   'system_mac_address': None},
                '12749C5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7112 PERTH ROAD 121',
                                   'serial': '12749C5',
                                   'serial_number': '12749C5',
                                   'system_mac_address': None},
                '12749C6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4001 LINE 61 (GINGERICH CLARE)',
                                   'serial': '12749C6',
                                   'serial_number': '12749C6',
                                   'system_mac_address': None},
                '12749C7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': 'VICTOR POOLE (6314 RD 124)',
                                   'serial': '12749C7',
                                   'serial_number': '12749C7',
                                   'system_mac_address': None},
                '12749CC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4413 PERTH LINE 72 (BAUMAN TOBIAS)',
                                   'serial': '12749CC',
                                   'serial_number': '12749CC',
                                   'system_mac_address': None},
                '12749E1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6825 PERTH ROAD 121 (KUEPFER DAVID & ANNA '
                                           'MAE)',
                                   'serial': '12749E1',
                                   'serial_number': '12749E1',
                                   'system_mac_address': None},
                '12749EC': {       'dev_macs': ['b8:94:70:f3:e6:be'],
                                   'ip_address': None,
                                   'name': '57 MILL ST E MILVERTON (HUNTER BILL)',
                                   'serial': '12749EC',
                                   'serial_number': '12749EC',
                                   'system_mac_address': None},
                '12749F1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6376 PERTH ROAD 121 (WAGLER CHARLES)',
                                   'serial': '12749F1',
                                   'serial_number': '12749F1',
                                   'system_mac_address': None},
                '12749F2': {       'dev_macs': ['b8:94:70:f3:e6:03'],
                                   'ip_address': None,
                                   'name': 'ont 232558',
                                   'serial': '12749F2',
                                   'serial_number': '12749F2',
                                   'system_mac_address': None},
                '12749F4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '3948 LINE 67 (SCHMIDT JAMES)',
                                   'serial': '12749F4',
                                   'serial_number': '12749F4',
                                   'system_mac_address': None},
                '12749F5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4741 LINE 64 (SCHADE MARY)',
                                   'serial': '12749F5',
                                   'serial_number': '12749F5',
                                   'system_mac_address': None},
                '12749FA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4228 LINE 61 (GASCHO DAN)',
                                   'serial': '12749FA',
                                   'serial_number': '12749FA',
                                   'system_mac_address': None},
                '12749FC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6946B PERTH ROAD 121 (THIBIDEAU KALEB)',
                                   'serial': '12749FC',
                                   'serial_number': '12749FC',
                                   'system_mac_address': None},
                '12749FF': {       'dev_macs': ['00:03:e6:f9:bc:38'],
                                   'ip_address': None,
                                   'name': '57 WILLIAM ST (PLAGER JEFFREY)',
                                   'serial': '12749FF',
                                   'serial_number': '12749FF',
                                   'system_mac_address': None},
                '1274A00': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4007 LINE 61 (BINKLE MURRAY)',
                                   'serial': '1274A00',
                                   'serial_number': '1274A00',
                                   'system_mac_address': None},
                '1274A01': {       'dev_macs': ['74:ac:b9:1f:74:ed'],
                                   'ip_address': None,
                                   'name': '15 MAIN ST S (ELLIS CHARLES)',
                                   'serial': '1274A01',
                                   'serial_number': '1274A01',
                                   'system_mac_address': None},
                '1274A05': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4404 PERTH LINE 72 (BAUMAN JESSE)',
                                   'serial': '1274A05',
                                   'serial_number': '1274A05',
                                   'system_mac_address': None},
                '1274A07': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6653 PERTH ROAD 121 (KUEPFER MANUEL A & '
                                           'RUTH ANNE)',
                                   'serial': '1274A07',
                                   'serial_number': '1274A07',
                                   'system_mac_address': None},
                '1274A08': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7120 PERTH ROAD 121 (ZEHR LAVERNE)',
                                   'serial': '1274A08',
                                   'serial_number': '1274A08',
                                   'system_mac_address': None},
                '1274A0B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6765 PERTH ROAD 121 (ADAMS STEVE)',
                                   'serial': '1274A0B',
                                   'serial_number': '1274A0B',
                                   'system_mac_address': None},
                '1274A0D': {       'dev_macs': ['14:21:03:92:eb:31'],
                                   'ip_address': None,
                                   'name': '4091 LINE 61 (KUEPFER ROGER)',
                                   'serial': '1274A0D',
                                   'serial_number': '1274A0D',
                                   'system_mac_address': None},
                '1274A0E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4898 PERTH LINE 72 (CARTER CHERYL)',
                                   'serial': '1274A0E',
                                   'serial_number': '1274A0E',
                                   'system_mac_address': None},
                '1274A0F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4410 LINE 64 (EGLI URS)',
                                   'serial': '1274A0F',
                                   'serial_number': '1274A0F',
                                   'system_mac_address': None},
                '1274A11': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4133 LINE 73 (MARTIN MELVIN & SHARON)',
                                   'serial': '1274A11',
                                   'serial_number': '1274A11',
                                   'system_mac_address': None},
                '1274A12': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6535 PERTH ROAD 121 (BRINSTON BRET)',
                                   'serial': '1274A12',
                                   'serial_number': '1274A12',
                                   'system_mac_address': None},
                '1274A19': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4094B LINE 61 (NEUFELD HELENA)',
                                   'serial': '1274A19',
                                   'serial_number': '1274A19',
                                   'system_mac_address': None},
                '1274A1D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4976 PERTH LINE 72 (CARTER CATHERINE)',
                                   'serial': '1274A1D',
                                   'serial_number': '1274A1D',
                                   'system_mac_address': None},
                '1274A1F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6906 ELGIN ST (HERRFORT MARLENE)',
                                   'serial': '1274A1F',
                                   'serial_number': '1274A1F',
                                   'system_mac_address': None},
                '1274A22': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6396 PERTH ROAD 121 (DEGROOT STEVEN)',
                                   'serial': '1274A22',
                                   'serial_number': '1274A22',
                                   'system_mac_address': None},
                '1274A23': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6574 ROAD 116 (MCCOY CHRISTINE)',
                                   'serial': '1274A23',
                                   'serial_number': '1274A23',
                                   'system_mac_address': None},
                '1274A25': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7019 ROAD 129 (STREICHER CARL)',
                                   'serial': '1274A25',
                                   'serial_number': '1274A25',
                                   'system_mac_address': None},
                '1274A26': {       'dev_macs': ['cc:32:e5:b2:c9:cd'],
                                   'ip_address': None,
                                   'name': '2541 LINE 54 (ERB HAROLD)',
                                   'serial': '1274A26',
                                   'serial_number': '1274A26',
                                   'system_mac_address': None},
                '1274A28': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4788 LINE 67 (EGLI ROGER)',
                                   'serial': '1274A28',
                                   'serial_number': '1274A28',
                                   'system_mac_address': None},
                '1274A2E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4493 LINE 61 (ALBRECHT KEN K)',
                                   'serial': '1274A2E',
                                   'serial_number': '1274A2E',
                                   'system_mac_address': None},
                '1274A33': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6176 PERTH ROAD 121 (ALBRECHT SANDRA)',
                                   'serial': '1274A33',
                                   'serial_number': '1274A33',
                                   'system_mac_address': None},
                '1274A37': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4172 PERTH LINE 72 (FENTON EUGENE & '
                                           'LESLEY)',
                                   'serial': '1274A37',
                                   'serial_number': '1274A37',
                                   'system_mac_address': None},
                '1274A3C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4013 LINE 61 (KUEPFER ROSS D)',
                                   'serial': '1274A3C',
                                   'serial_number': '1274A3C',
                                   'system_mac_address': None},
                '1274A3F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4097 LINE 50 (NEEB RICHARD)',
                                   'serial': '1274A3F',
                                   'serial_number': '1274A3F',
                                   'system_mac_address': None},
                '1274A40': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7455 PERTH ROAD 121 (KUEPFER MURRAY & '
                                           'ELIZABETH)',
                                   'serial': '1274A40',
                                   'serial_number': '1274A40',
                                   'system_mac_address': None},
                '1274A42': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4164 LINE 61 (SIEGNER PAUL)',
                                   'serial': '1274A42',
                                   'serial_number': '1274A42',
                                   'system_mac_address': None},
                '1274A44': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '139 MADDISON ST E (WEIR TERESA & ROBIN)',
                                   'serial': '1274A44',
                                   'serial_number': '1274A44',
                                   'system_mac_address': None},
                '1274A4A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4112 LINE 61 (JANTZI NICK)',
                                   'serial': '1274A4A',
                                   'serial_number': '1274A4A',
                                   'system_mac_address': None},
                '1274A4C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '162 SMITH ST (SMITH BRENDA)',
                                   'serial': '1274A4C',
                                   'serial_number': '1274A4C',
                                   'system_mac_address': None},
                '1274A50': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6566 PERTH ROAD 121 (ZEHR BRYAN & HEIDI)',
                                   'serial': '1274A50',
                                   'serial_number': '1274A50',
                                   'system_mac_address': None},
                '1274A51': {       'dev_macs': [       '00:03:e6:da:bb:34',
                                                       '00:03:e6:db:30:70'],
                                   'ip_address': None,
                                   'name': '5333 Line 49 Gadshill (Gerald Pol '
                                           '10401245)',
                                   'serial': '1274A51',
                                   'serial_number': '1274A51',
                                   'system_mac_address': None},
                '1274A52': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6395 PERTH ROAD 121 (JANTZI EXCAVATING & '
                                           'BULLDOZING)',
                                   'serial': '1274A52',
                                   'serial_number': '1274A52',
                                   'system_mac_address': None},
                '1274A55': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6400 PERTH ROAD 121 (KUEPFER VICTOR)',
                                   'serial': '1274A55',
                                   'serial_number': '1274A55',
                                   'system_mac_address': None},
                '1274A5A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6401 PERTH ROAD 121 (MILES WILLIAM)',
                                   'serial': '1274A5A',
                                   'serial_number': '1274A5A',
                                   'system_mac_address': None},
                '1274A5B': {       'dev_macs': ['44:65:7f:89:0c:60'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (HOSSFELD KRYSTAL)',
                                   'serial': '1274A5B',
                                   'serial_number': '1274A5B',
                                   'system_mac_address': None},
                '1274A5D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5311 LINE 52 (ERNST ERIKA)',
                                   'serial': '1274A5D',
                                   'serial_number': '1274A5D',
                                   'system_mac_address': None},
                '1274A5E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6433 ROAD 124 (ANDERSON RYAN)',
                                   'serial': '1274A5E',
                                   'serial_number': '1274A5E',
                                   'system_mac_address': None},
                '1274A60': {       'dev_macs': ['14:21:03:d1:7c:02'],
                                   'ip_address': None,
                                   'name': '170 MARKET ST (REITZEL ETHAN)',
                                   'serial': '1274A60',
                                   'serial_number': '1274A60',
                                   'system_mac_address': None},
                '1274A68': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6621 ROAD 116 (ZEHR TERRY & SUE)',
                                   'serial': '1274A68',
                                   'serial_number': '1274A68',
                                   'system_mac_address': None},
                '1274A6D': {       'dev_macs': ['44:65:7f:89:0d:ad'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (BAUMAN ROSEANNE)',
                                   'serial': '1274A6D',
                                   'serial_number': '1274A6D',
                                   'system_mac_address': None},
                '1274A6F': {       'dev_macs': ['00:03:e6:f9:fb:fc'],
                                   'ip_address': None,
                                   'name': '6405 PERTH ROAD 121 (LEIS ALBERT)',
                                   'serial': '1274A6F',
                                   'serial_number': '1274A6F',
                                   'system_mac_address': None},
                '1274A70': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '16 TEMPERANCE ST (LEIS STANLEY)',
                                   'serial': '1274A70',
                                   'serial_number': '1274A70',
                                   'system_mac_address': None},
                '1274A75': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5325 ROAD 140 (BRODHAGEN ELVEY)',
                                   'serial': '1274A75',
                                   'serial_number': '1274A75',
                                   'system_mac_address': None},
                '1274A76': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4577 LINE 61 (VAN NES JAMES)',
                                   'serial': '1274A76',
                                   'serial_number': '1274A76',
                                   'system_mac_address': None},
                '1274A78': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5899 PERTH RD 119 (BRAUN AARON & '
                                           'MARGARETHA)',
                                   'serial': '1274A78',
                                   'serial_number': '1274A78',
                                   'system_mac_address': None},
                '1274A79': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4464 LINE 64 (MCNALLY SHANE)',
                                   'serial': '1274A79',
                                   'serial_number': '1274A79',
                                   'system_mac_address': None},
                '1274A7A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6535 PERTH ROAD 121 (ROES NORMAN)',
                                   'serial': '1274A7A',
                                   'serial_number': '1274A7A',
                                   'system_mac_address': None},
                '1274A7E': {       'dev_macs': ['b8:94:70:f3:d5:6d'],
                                   'ip_address': None,
                                   'name': '5198 LINE 49 (GRAUL NICOLE)',
                                   'serial': '1274A7E',
                                   'serial_number': '1274A7E',
                                   'system_mac_address': None},
                '1274A80': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "6878 ROAD 116 (JOHN'S TRANSIT)",
                                   'serial': '1274A80',
                                   'serial_number': '1274A80',
                                   'system_mac_address': None},
                '1274A89': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6946 ROAD 124 (JANTZI JONATHAN)',
                                   'serial': '1274A89',
                                   'serial_number': '1274A89',
                                   'system_mac_address': None},
                '1274A8A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5583 ROAD 164 MONKTON (FRANCIS CHRISTINE)',
                                   'serial': '1274A8A',
                                   'serial_number': '1274A8A',
                                   'system_mac_address': None},
                '1274A8C': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6573 ROAD 140 (EBERSOL ELI)',
                                   'serial': '1274A8C',
                                   'serial_number': '1274A8C',
                                   'system_mac_address': None},
                '1274A91': {       'dev_macs': ['44:65:7f:89:0d:1d'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (PRIALA GHEORGHE)',
                                   'serial': '1274A91',
                                   'serial_number': '1274A91',
                                   'system_mac_address': None},
                '1274A94': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4341 PERTH LINE 72 (JANTZI JIM)',
                                   'serial': '1274A94',
                                   'serial_number': '1274A94',
                                   'system_mac_address': None},
                '1274A96': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6886 ROAD 129 (COOK WILLIAM & JEAN)',
                                   'serial': '1274A96',
                                   'serial_number': '1274A96',
                                   'system_mac_address': None},
                '1274A9A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4059 LINE 76 (MILLBANK MFG INC)',
                                   'serial': '1274A9A',
                                   'serial_number': '1274A9A',
                                   'system_mac_address': None},
                '1274A9F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6602 PERTH ROAD 121 (HOOVER SAMUEL B)',
                                   'serial': '1274A9F',
                                   'serial_number': '1274A9F',
                                   'system_mac_address': None},
                '1274AA0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6997 ROAD 124 (KUEPFER DANNIE)',
                                   'serial': '1274AA0',
                                   'serial_number': '1274AA0',
                                   'system_mac_address': None},
                '1274AA5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6261 PERTH ROAD 121 (ROES SCOTT)',
                                   'serial': '1274AA5',
                                   'serial_number': '1274AA5',
                                   'system_mac_address': None},
                '1274AA7': {       'dev_macs': ['00:03:e6:f9:d4:bc'],
                                   'ip_address': None,
                                   'name': '77 MAIN ST N (RASMUSSEN FLORENCE)',
                                   'serial': '1274AA7',
                                   'serial_number': '1274AA7',
                                   'system_mac_address': None},
                '1274AAA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6497 ROAD 124 (WAGLER JEREMY)',
                                   'serial': '1274AAA',
                                   'serial_number': '1274AAA',
                                   'system_mac_address': None},
                '1274AAD': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '3 MILL ST W (BRAUN DAVID & MARGARETHA)',
                                   'serial': '1274AAD',
                                   'serial_number': '1274AAD',
                                   'system_mac_address': None},
                '1274AB0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4740 LINE 67 (5588269 ONTARIO INC)',
                                   'serial': '1274AB0',
                                   'serial_number': '1274AB0',
                                   'system_mac_address': None},
                '1274AB5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4892 PERTH LINE 55 (WAGLER MRS KATIE)',
                                   'serial': '1274AB5',
                                   'serial_number': '1274AB5',
                                   'system_mac_address': None},
                '1274AB8': {       'dev_macs': ['d4:ca:6d:32:23:73'],
                                   'ip_address': None,
                                   'name': '4392 PERTH LINE 72 (NEWTON FURNITURE INC)',
                                   'serial': '1274AB8',
                                   'serial_number': '1274AB8',
                                   'system_mac_address': None},
                '1274AC2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6193 PERTH ROAD 121 (HYATT PAUL & '
                                           'KIMBERLY)',
                                   'serial': '1274AC2',
                                   'serial_number': '1274AC2',
                                   'system_mac_address': None},
                '1274AC9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6502 ROAD 124 (JANTZI RONALD)',
                                   'serial': '1274AC9',
                                   'serial_number': '1274AC9',
                                   'system_mac_address': None},
                '1274ACF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6800 PERTH ROAD 121 (KUEPFER WAYNE S)',
                                   'serial': '1274ACF',
                                   'serial_number': '1274ACF',
                                   'system_mac_address': None},
                '1274AD0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4119 PERTH LINE 72 (LIVESTOCK NUTRITIONAL '
                                           'SERVICES)',
                                   'serial': '1274AD0',
                                   'serial_number': '1274AD0',
                                   'system_mac_address': None},
                '1274AD5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6551 ROAD 124 (MARTIN AARON F)',
                                   'serial': '1274AD5',
                                   'serial_number': '1274AD5',
                                   'system_mac_address': None},
                '1274AD7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4341 PERTH LINE 72 (JANTZI LEVI K)',
                                   'serial': '1274AD7',
                                   'serial_number': '1274AD7',
                                   'system_mac_address': None},
                '1274ADF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4434 LINE 67 (BANCROFT PETER)',
                                   'serial': '1274ADF',
                                   'serial_number': '1274ADF',
                                   'system_mac_address': None},
                '1274AE9': {       'dev_macs': ['44:65:7f:89:09:ab'],
                                   'ip_address': None,
                                   'name': '5 MAIN ST S UNIT 102 (GOETZ LORI)',
                                   'serial': '1274AE9',
                                   'serial_number': '1274AE9',
                                   'system_mac_address': None},
                '1274AED': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6720 PERTH ROAD 121 (KUEPFER KEN)',
                                   'serial': '1274AED',
                                   'serial_number': '1274AED',
                                   'system_mac_address': None},
                '1274AF5': {       'dev_macs': ['b8:94:70:f3:ec:54'],
                                   'ip_address': None,
                                   'name': '57 PUGH ST (HEALEY RYAN)',
                                   'serial': '1274AF5',
                                   'serial_number': '1274AF5',
                                   'system_mac_address': None},
                '1274AFB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6752 PERTH ROAD 121 (ZEHR CHERYL)',
                                   'serial': '1274AFB',
                                   'serial_number': '1274AFB',
                                   'system_mac_address': None},
                '1274AFC': {       'dev_macs': ['5c:0f:fb:34:ac:02'],
                                   'ip_address': None,
                                   'name': '3 MURRAY CT MILVERTON (WILLIAM HORST '
                                           '10400332)',
                                   'serial': '1274AFC',
                                   'serial_number': '1274AFC',
                                   'system_mac_address': None},
                '1274AFE': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4147 PERTH LINE 72 (GERBER SHARON)',
                                   'serial': '1274AFE',
                                   'serial_number': '1274AFE',
                                   'system_mac_address': None},
                '1274AFF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4976 PERTH LINE 72 (CARTER KEN & ANITA)',
                                   'serial': '1274AFF',
                                   'serial_number': '1274AFF',
                                   'system_mac_address': None},
                '1274B00': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4435 LINE 67 (KUEPFER ALMEDA & CLARE)',
                                   'serial': '1274B00',
                                   'serial_number': '1274B00',
                                   'system_mac_address': None},
                '1274B02': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (KAUR  SIMRAN)',
                                   'serial': '1274B02',
                                   'serial_number': '1274B02',
                                   'system_mac_address': None},
                '1274B0D': {       'dev_macs': ['44:65:7f:89:0c:c3'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (JONES LOGAN)',
                                   'serial': '1274B0D',
                                   'serial_number': '1274B0D',
                                   'system_mac_address': None},
                '1274B0F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4232 LINE 61 (ERB WAYNE)',
                                   'serial': '1274B0F',
                                   'serial_number': '1274B0F',
                                   'system_mac_address': None},
                '1274B10': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '54 PUGH ST UNIT 4 (GERBER BUD)',
                                   'serial': '1274B10',
                                   'serial_number': '1274B10',
                                   'system_mac_address': None},
                '1274B14': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6551 PERTH ROAD 121 (MARTIN MERVIN M J)',
                                   'serial': '1274B14',
                                   'serial_number': '1274B14',
                                   'system_mac_address': None},
                '1274B1D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4287 LINE 67 (KUEPFER CALVIN)',
                                   'serial': '1274B1D',
                                   'serial_number': '1274B1D',
                                   'system_mac_address': None},
                '1274B20': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4230 LINE 61 (SCHWARTZENTRUBER JOYCE)',
                                   'serial': '1274B20',
                                   'serial_number': '1274B20',
                                   'system_mac_address': None},
                '1274B26': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4138 PERTH LINE 72 (BEREAN CHRISTIAN '
                                           'FELLOWSHIP)',
                                   'serial': '1274B26',
                                   'serial_number': '1274B26',
                                   'system_mac_address': None},
                '1274B2F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4137 PERTH LINE 72 (KUEPFER JOHN K)',
                                   'serial': '1274B2F',
                                   'serial_number': '1274B2F',
                                   'system_mac_address': None},
                '1274B30': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4939 PERTH LINE 72 (KUEPFER ALVIN K & ANNA '
                                           'MAE)',
                                   'serial': '1274B30',
                                   'serial_number': '1274B30',
                                   'system_mac_address': None},
                '1274B38': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6563 ROAD 116 (JANTZI SHARON)',
                                   'serial': '1274B38',
                                   'serial_number': '1274B38',
                                   'system_mac_address': None},
                '1274B3B': {       'dev_macs': ['b8:94:70:f3:e7:8f'],
                                   'ip_address': None,
                                   'name': '4092 QUEEN ST (BAUMAN DARRELL)',
                                   'serial': '1274B3B',
                                   'serial_number': '1274B3B',
                                   'system_mac_address': None},
                '1274B44': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4712 LINE 67 (ALBRECHT ELDON & CHRISTENA)',
                                   'serial': '1274B44',
                                   'serial_number': '1274B44',
                                   'system_mac_address': None},
                '1274B4A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4899 PERTH LINE 72 (KUEPFER ROY)',
                                   'serial': '1274B4A',
                                   'serial_number': '1274B4A',
                                   'system_mac_address': None},
                '1274B5A': {       'dev_macs': ['b8:94:70:f3:df:c8'],
                                   'ip_address': None,
                                   'name': '4 EDWARD ST (CHALMERS JENNIFER)',
                                   'serial': '1274B5A',
                                   'serial_number': '1274B5A',
                                   'system_mac_address': None},
                '1274B6F': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4276 PERTH LINE 72 (BAUMAN AMSEY S)',
                                   'serial': '1274B6F',
                                   'serial_number': '1274B6F',
                                   'system_mac_address': None},
                '1274FA3': {       'dev_macs': ['00:90:7f:19:e4:a2'],
                                   'ip_address': None,
                                   'name': '4570 PERTH LINE 72 (BATEMAN EQUIPMENT '
                                           'CORPORATION)',
                                   'serial': '1274FA3',
                                   'serial_number': '1274FA3',
                                   'system_mac_address': None},
                '1274FF3': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4317 PERTH LINE 72 (KUEPFER SANFORD)',
                                   'serial': '1274FF3',
                                   'serial_number': '1274FF3',
                                   'system_mac_address': None},
                '1274FF9': {       'dev_macs': ['14:21:03:8b:71:d9'],
                                   'ip_address': None,
                                   'name': '4908 LINE 59 (DIEHL FRAZER FARMING LTD)',
                                   'serial': '1274FF9',
                                   'serial_number': '1274FF9',
                                   'system_mac_address': None},
                '1274FFB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4492 PERTH LINE 72 (ALBRECHT AMMON)',
                                   'serial': '1274FFB',
                                   'serial_number': '1274FFB',
                                   'system_mac_address': None},
                '1275008': {       'dev_macs': ['b4:fb:e4:8e:43:79'],
                                   'ip_address': None,
                                   'name': '3 ST MARYS ST N (DRAGOI CRISTIAN & '
                                           'ADRIANA)',
                                   'serial': '1275008',
                                   'serial_number': '1275008',
                                   'system_mac_address': None},
                '1275036': {       'dev_macs': ['b8:94:70:f3:e0:e6'],
                                   'ip_address': None,
                                   'name': '5804 PERTH ROAD 119 (CNOSSEN SIMON & '
                                           'ANDREA)',
                                   'serial': '1275036',
                                   'serial_number': '1275036',
                                   'system_mac_address': None},
                '127503A': {       'dev_macs': ['14:21:03:92:ea:4a'],
                                   'ip_address': None,
                                   'name': '4244 PERTH LINE 72 (ALBRECHT ALEX)',
                                   'serial': '127503A',
                                   'serial_number': '127503A',
                                   'system_mac_address': None},
                '1275043': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4619 PERTH LINE 72 (MCTAVISH COLLEEN)',
                                   'serial': '1275043',
                                   'serial_number': '1275043',
                                   'system_mac_address': None},
                '1275053': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4620 PERTH LINE 72 (MCKERTCHER JOYCE)',
                                   'serial': '1275053',
                                   'serial_number': '1275053',
                                   'system_mac_address': None},
                '1275072': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '160 FISHLEIGH ST MONKTON (MCGIRR KENNETH  '
                                           '10433784)',
                                   'serial': '1275072',
                                   'serial_number': '1275072',
                                   'system_mac_address': None},
                '1275076': {       'dev_macs': ['00:03:e6:f9:d7:82'],
                                   'ip_address': None,
                                   'name': '26 MAIN ST N (DULIES ALMUTH)',
                                   'serial': '1275076',
                                   'serial_number': '1275076',
                                   'system_mac_address': None},
                '1275078': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5005 PERTH LINE 55 (ALBRECHT LESLIE)',
                                   'serial': '1275078',
                                   'serial_number': '1275078',
                                   'system_mac_address': None},
                '1275082': {       'dev_macs': ['b8:94:70:f3:e8:1e'],
                                   'ip_address': None,
                                   'name': '120 PUGH ST (SCHMIDT KELSEY)',
                                   'serial': '1275082',
                                   'serial_number': '1275082',
                                   'system_mac_address': None},
                '12750CC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5134 PERTH LINE 55 (JANTZI MARTIN & '
                                           'PAULINE)',
                                   'serial': '12750CC',
                                   'serial_number': '12750CC',
                                   'system_mac_address': None},
                '12750CD': {       'dev_macs': ['b8:94:70:f3:e6:0e'],
                                   'ip_address': None,
                                   'name': '31 MAIN ST S (HOBBS KEN & GUDRUN)',
                                   'serial': '12750CD',
                                   'serial_number': '12750CD',
                                   'system_mac_address': None},
                '12750D7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5808 PERTH ROAD 119 (FORLER RACHELLE)',
                                   'serial': '12750D7',
                                   'serial_number': '12750D7',
                                   'system_mac_address': None},
                '1275114': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '69 MAIN ST N (PRAYER HOUSE OF MILVERTON)',
                                   'serial': '1275114',
                                   'serial_number': '1275114',
                                   'system_mac_address': None},
                '127515A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '27 ARENA DR (PUGH TIM & ROSE)',
                                   'serial': '127515A',
                                   'serial_number': '127515A',
                                   'system_mac_address': None},
                '127518B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4482 PERTH LINE 72 (HOERLE MARY)',
                                   'serial': '127518B',
                                   'serial_number': '127518B',
                                   'system_mac_address': None},
                '127518F': {       'dev_macs': ['b8:94:70:f3:e4:e5'],
                                   'ip_address': None,
                                   'name': '103 PUGH ST (SABOL SAMANTHA)',
                                   'serial': '127518F',
                                   'serial_number': '127518F',
                                   'system_mac_address': None},
                '1275197': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4870 LINE 49 (TAMAS ALEXANDRA)',
                                   'serial': '1275197',
                                   'serial_number': '1275197',
                                   'system_mac_address': None},
                '12751C1': {       'dev_macs': ['44:65:7f:41:94:87'],
                                   'ip_address': None,
                                   'name': '57 WILLIAM ST (OWUSU SELINA)',
                                   'serial': '12751C1',
                                   'serial_number': '12751C1',
                                   'system_mac_address': None},
                '12751D0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4516 PERTH LINE 72 (REIST STEVEN)',
                                   'serial': '12751D0',
                                   'serial_number': '12751D0',
                                   'system_mac_address': None},
                '12751D6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4528 PERTH LINE 72 (THIESSEN JACOB)',
                                   'serial': '12751D6',
                                   'serial_number': '12751D6',
                                   'system_mac_address': None},
                '12751DA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '57 WILLIAM ST APT 202 (HOWATT MITCHELL)',
                                   'serial': '12751DA',
                                   'serial_number': '12751DA',
                                   'system_mac_address': None},
                '12751DD': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6968 MILLBANK MAIN ST (FRANKLIN GARY)',
                                   'serial': '12751DD',
                                   'serial_number': '12751DD',
                                   'system_mac_address': None},
                '1295E37': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '54 AVERY PL (KUHL WADE)',
                                   'serial': '1295E37',
                                   'serial_number': '1295E37',
                                   'system_mac_address': None},
                '1295E49': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5883 PERTH ROAD 119 (JACOBS HERB K)',
                                   'serial': '1295E49',
                                   'serial_number': '1295E49',
                                   'system_mac_address': None},
                '1295E66': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4494 LINE 64 (ZEHR JACOB)',
                                   'serial': '1295E66',
                                   'serial_number': '1295E66',
                                   'system_mac_address': None},
                '1295E76': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2332 LINE 47 (GERBER PAUL)',
                                   'serial': '1295E76',
                                   'serial_number': '1295E76',
                                   'system_mac_address': None},
                '1295E95': {       'dev_macs': ['e4:6c:d1:01:b9:e1'],
                                   'ip_address': None,
                                   'name': '58 AVERY PL (KARDAS BRANT)',
                                   'serial': '1295E95',
                                   'serial_number': '1295E95',
                                   'system_mac_address': None},
                '1295E97': {       'dev_macs': ['e4:6c:d1:4e:e6:bb'],
                                   'ip_address': None,
                                   'name': '112 PUGH ST (WITEN STACEY & LAURENCE)',
                                   'serial': '1295E97',
                                   'serial_number': '1295E97',
                                   'system_mac_address': None},
                '1295E9B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2263 LINE 51 (LINDNER JANE)',
                                   'serial': '1295E9B',
                                   'serial_number': '1295E9B',
                                   'system_mac_address': None},
                '1295EA3': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5979 PERTH ROAD 119 (MARTIN DARRYK)',
                                   'serial': '1295EA3',
                                   'serial_number': '1295EA3',
                                   'system_mac_address': None},
                '1295EA6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4464 CRESCENT AV (ALBRECHT PAUL & MARILYN)',
                                   'serial': '1295EA6',
                                   'serial_number': '1295EA6',
                                   'system_mac_address': None},
                '1295EAA': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4432 PERTH LINE 72 (FABER SUSAN)',
                                   'serial': '1295EAA',
                                   'serial_number': '1295EAA',
                                   'system_mac_address': None},
                '1295EB8': {       'dev_macs': ['e4:6c:d1:4e:ec:a9'],
                                   'ip_address': None,
                                   'name': '33 MURRAY CT (SHAW DANIELLE)',
                                   'serial': '1295EB8',
                                   'serial_number': '1295EB8',
                                   'system_mac_address': None},
                '1295EBB': {       'dev_macs': ['e4:6c:d1:4f:bc:28'],
                                   'ip_address': None,
                                   'name': '5499 PERTH LINE 55 (MEULLER LINDA)',
                                   'serial': '1295EBB',
                                   'serial_number': '1295EBB',
                                   'system_mac_address': None},
                '1295EC1': {       'dev_macs': ['bc:fc:e7:22:8d:58'],
                                   'ip_address': None,
                                   'name': '4926 Line 61 Milverton (Markus Mueller)',
                                   'serial': '1295EC1',
                                   'serial_number': '1295EC1',
                                   'system_mac_address': None},
                '1295FB2': {       'dev_macs': ['e4:6c:d1:4f:b1:d8'],
                                   'ip_address': None,
                                   'name': '5915 SOMERSET DR (TOMBS GREG)',
                                   'serial': '1295FB2',
                                   'serial_number': '1295FB2',
                                   'system_mac_address': None},
                '129600E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '128 PUGH ST',
                                   'serial': '129600E',
                                   'serial_number': '129600E',
                                   'system_mac_address': None},
                '1296012': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4460C PERTH LINE 72 (WALL AARON)',
                                   'serial': '1296012',
                                   'serial_number': '1296012',
                                   'system_mac_address': None},
                '1296022': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4957 ROAD 109 (ROTH CHRIS)',
                                   'serial': '1296022',
                                   'serial_number': '1296022',
                                   'system_mac_address': None},
                '129602E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4072 WATERLOO ST (MILLBANK ENTERPRISES '
                                           'INC)',
                                   'serial': '129602E',
                                   'serial_number': '129602E',
                                   'system_mac_address': None},
                '1296041': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5930 LINE 57 (PELKMANS ELS)',
                                   'serial': '1296041',
                                   'serial_number': '1296041',
                                   'system_mac_address': None},
                '1296044': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6932 HORIZON TL (NICKEL AMANDA)',
                                   'serial': '1296044',
                                   'serial_number': '1296044',
                                   'system_mac_address': None},
                '129604E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4486 LINE 52 (DEBUS RON)',
                                   'serial': '129604E',
                                   'serial_number': '129604E',
                                   'system_mac_address': None},
                '129605A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6982 CHURCH ST (HORST LAVERNE)',
                                   'serial': '129605A',
                                   'serial_number': '129605A',
                                   'system_mac_address': None},
                '12960D8': {       'dev_macs': ['60:db:98:09:38:34'],
                                   'ip_address': None,
                                   'name': '208 WINSTANLEY ST (KAREN AND GORD BENDER)',
                                   'serial': '12960D8',
                                   'serial_number': '12960D8',
                                   'system_mac_address': None},
                '12960DB': {       'dev_macs': ['e4:6c:d1:01:b3:c7'],
                                   'ip_address': None,
                                   'name': '19 SCHADE ST MONKTON (HUBER DERRICK)',
                                   'serial': '12960DB',
                                   'serial_number': '12960DB',
                                   'system_mac_address': None},
                '12960E0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6931 PERTH ROAD 121 (ED ZEHR)',
                                   'serial': '12960E0',
                                   'serial_number': '12960E0',
                                   'system_mac_address': None},
                '12960EF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4374 LINE 64 (ALBRECHT DAVID Z)',
                                   'serial': '12960EF',
                                   'serial_number': '12960EF',
                                   'system_mac_address': None},
                '1296110': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6896 PERTH ROAD 121 (T & D RESTORATIONS '
                                           'INC)',
                                   'serial': '1296110',
                                   'serial_number': '1296110',
                                   'system_mac_address': None},
                '1296121': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6190 PERTH ROAD 131 (DAVIDSON JEREMY & '
                                           'MALINDA)',
                                   'serial': '1296121',
                                   'serial_number': '1296121',
                                   'system_mac_address': None},
                '1296122': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6478 ROAD 129 (KUEPFER ELDON R)',
                                   'serial': '1296122',
                                   'serial_number': '1296122',
                                   'system_mac_address': None},
                '1296127': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6015 PERTH ROAD 119 (KUEPFER JUDITH)',
                                   'serial': '1296127',
                                   'serial_number': '1296127',
                                   'system_mac_address': None},
                '129613B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '74 PUGH ST (SCOTCHMER WAYNE)',
                                   'serial': '129613B',
                                   'serial_number': '129613B',
                                   'system_mac_address': None},
                '1296DED': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7118 PERTH ROAD 121 (BAUMAN ABNER)',
                                   'serial': '1296DED',
                                   'serial_number': '1296DED',
                                   'system_mac_address': None},
                '1296FCB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4009 LINE 61 (CHALMERS KEVIN)',
                                   'serial': '1296FCB',
                                   'serial_number': '1296FCB',
                                   'system_mac_address': None},
                '1296FD4': {       'dev_macs': ['60:db:98:54:a4:67'],
                                   'ip_address': None,
                                   'name': '58 PUGH ST (PATHAMMAVONG LEE)',
                                   'serial': '1296FD4',
                                   'serial_number': '1296FD4',
                                   'system_mac_address': None},
                '1296FE0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4198 LINE 61 (CONRAD JOHN)',
                                   'serial': '1296FE0',
                                   'serial_number': '1296FE0',
                                   'system_mac_address': None},
                '1297066': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4104 LINE 61 (BAUMAN FRED)',
                                   'serial': '1297066',
                                   'serial_number': '1297066',
                                   'system_mac_address': None},
                '12970D4': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4457 COMMUNITY CENTRE DRIVE (SHANTZ '
                                           'MARIANNE)',
                                   'serial': '12970D4',
                                   'serial_number': '12970D4',
                                   'system_mac_address': None},
                '1297268': {       'dev_macs': ['14:21:03:a9:6e:e7'],
                                   'ip_address': None,
                                   'name': '4499 LINE 46 (VANSTRAATEN CATHY)',
                                   'serial': '1297268',
                                   'serial_number': '1297268',
                                   'system_mac_address': None},
                '1297287': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4535 LINE 46 (VANSTRAATEN AMY)',
                                   'serial': '1297287',
                                   'serial_number': '1297287',
                                   'system_mac_address': None},
                '129734C': {       'dev_macs': ['d8:b3:70:20:13:d1'],
                                   'ip_address': None,
                                   'name': '218 WINSTANLEY ST (PEARL INSPIRATIONS)',
                                   'serial': '129734C',
                                   'serial_number': '129734C',
                                   'system_mac_address': None},
                '1297351': {       'dev_macs': ['c4:ad:34:b1:22:bf'],
                                   'ip_address': None,
                                   'name': '4676 HAWTHORNE CT (BADEN WOODTURNING INC.)',
                                   'serial': '1297351',
                                   'serial_number': '1297351',
                                   'system_mac_address': None},
                '1297360': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4028 LINE 50 (YODER ARNOLD)',
                                   'serial': '1297360',
                                   'serial_number': '1297360',
                                   'system_mac_address': None},
                '1297375': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "4158 LINE 67 (ALBRECHT'S CHAIN SAW "
                                           'SERVICE)',
                                   'serial': '1297375',
                                   'serial_number': '1297375',
                                   'system_mac_address': None},
                '1297379': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2568 LINE 47 (SCHNEIDER LARRY)',
                                   'serial': '1297379',
                                   'serial_number': '1297379',
                                   'system_mac_address': None},
                '1297389': {       'dev_macs': ['14:21:03:a1:e5:c3'],
                                   'ip_address': None,
                                   'name': '72 AVERY PL (MCINTOSH MARGO)',
                                   'serial': '1297389',
                                   'serial_number': '1297389',
                                   'system_mac_address': None},
                '129738A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "4464 PERTH LINE 72 (JACOBS' FAMILY "
                                           'SERVICE)',
                                   'serial': '129738A',
                                   'serial_number': '129738A',
                                   'system_mac_address': None},
                '129738E': {       'dev_macs': ['a8:6e:84:00:aa:1d'],
                                   'ip_address': None,
                                   'name': '4399 LINE 61 (HAVERKAMP RALPH)',
                                   'serial': '129738E',
                                   'serial_number': '129738E',
                                   'system_mac_address': None},
                '1297390': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4223 LINE 50 (VAN LEEUWEN WILLIAM)',
                                   'serial': '1297390',
                                   'serial_number': '1297390',
                                   'system_mac_address': None},
                '1297394': {       'dev_macs': ['14:21:03:a9:70:89'],
                                   'ip_address': None,
                                   'name': '2606 LINE 51 (CIPP-FREIBURGER DEBBIE)',
                                   'serial': '1297394',
                                   'serial_number': '1297394',
                                   'system_mac_address': None},
                '129739E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6802 ROAD 124 (KUEPFER NELSON & DORIS)',
                                   'serial': '129739E',
                                   'serial_number': '129739E',
                                   'system_mac_address': None},
                '12973B0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '2460 LINE 45 (GERBER TIMOTHY)',
                                   'serial': '12973B0',
                                   'serial_number': '12973B0',
                                   'system_mac_address': None},
                '12973B5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '14 CRUICKSHANK APT 1 (REIS KEVIN)',
                                   'serial': '12973B5',
                                   'serial_number': '12973B5',
                                   'system_mac_address': None},
                '12973CF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4109 LINE 61 (HILDEBRANDT ROSANA)',
                                   'serial': '12973CF',
                                   'serial_number': '12973CF',
                                   'system_mac_address': None},
                '12973D7': {       'dev_macs': ['14:21:03:92:c7:3a'],
                                   'ip_address': None,
                                   'name': '5 MAIN ST S (KETA LISA)',
                                   'serial': '12973D7',
                                   'serial_number': '12973D7',
                                   'system_mac_address': None},
                '12973DE': {       'dev_macs': ['14:21:03:a9:70:c0'],
                                   'ip_address': None,
                                   'name': '2339 LINE 51 (WAGLER CHRISTINE)',
                                   'serial': '12973DE',
                                   'serial_number': '12973DE',
                                   'system_mac_address': None},
                '12973E1': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '85 MAIN ST N (THEDORFF SANDY)',
                                   'serial': '12973E1',
                                   'serial_number': '12973E1',
                                   'system_mac_address': None},
                '12973E5': {       'dev_macs': ['14:21:03:8b:73:d3'],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (WEBER JOSH)',
                                   'serial': '12973E5',
                                   'serial_number': '12973E5',
                                   'system_mac_address': None},
                '12973ED': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7022 ROAD 129 (NEWTON RUSTICS)',
                                   'serial': '12973ED',
                                   'serial_number': '12973ED',
                                   'system_mac_address': None},
                '12973F0': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4085 PRINCESS ST (BARTEL ALLISON)',
                                   'serial': '12973F0',
                                   'serial_number': '12973F0',
                                   'system_mac_address': None},
                '12973F3': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4219 LINE 46 (BRUULSEMA HAROLD)',
                                   'serial': '12973F3',
                                   'serial_number': '12973F3',
                                   'system_mac_address': None},
                '12973FB': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "6684 ROAD 129 (JANTZI'S SHOE BOX)",
                                   'serial': '12973FB',
                                   'serial_number': '12973FB',
                                   'system_mac_address': None},
                '12973FD': {       'dev_macs': ['14:21:03:8b:73:23'],
                                   'ip_address': None,
                                   'name': '6025 PERTH ROAD 119 (SCHEERER KEN & '
                                           'CONNIE)',
                                   'serial': '12973FD',
                                   'serial_number': '12973FD',
                                   'system_mac_address': None},
                '1297411': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5542 ROAD 126 (GASCHO KEITH & ANNETTE)',
                                   'serial': '1297411',
                                   'serial_number': '1297411',
                                   'system_mac_address': None},
                '129741B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4505 LINE 46 (ROTH GARY & KIM)',
                                   'serial': '129741B',
                                   'serial_number': '129741B',
                                   'system_mac_address': None},
                '129741D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4721 LINE 64 (ARRO CRUSHING LTD)',
                                   'serial': '129741D',
                                   'serial_number': '129741D',
                                   'system_mac_address': None},
                '1297421': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4443 PERTH LINE 72 (ALBRECHT DIANE)',
                                   'serial': '1297421',
                                   'serial_number': '1297421',
                                   'system_mac_address': None},
                '1297426': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "13 MILL ST E (KIPFER'S GARAGE LTD)",
                                   'serial': '1297426',
                                   'serial_number': '1297426',
                                   'system_mac_address': None},
                '129742A': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4424 PERTH LINE 72 (STECKLY ELSIE)',
                                   'serial': '129742A',
                                   'serial_number': '129742A',
                                   'system_mac_address': None},
                '129743D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4447 PERTH LINE 72 (PURCELL DOUG)',
                                   'serial': '129743D',
                                   'serial_number': '129743D',
                                   'system_mac_address': None},
                '1297457': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4878 ROAD 109 (REFRESHING WIND MINISTRIES)',
                                   'serial': '1297457',
                                   'serial_number': '1297457',
                                   'system_mac_address': None},
                '129745A': {       'dev_macs': ['14:21:03:a9:75:9b'],
                                   'ip_address': None,
                                   'name': '5205 LINE 58 (YOUNG JACOB)',
                                   'serial': '129745A',
                                   'serial_number': '129745A',
                                   'system_mac_address': None},
                '129745D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4188 LINE 61 (BULMER D)',
                                   'serial': '129745D',
                                   'serial_number': '129745D',
                                   'system_mac_address': None},
                '1297469': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4445 PERTH LINE 72 (A & A CONTRACTING)',
                                   'serial': '1297469',
                                   'serial_number': '1297469',
                                   'system_mac_address': None},
                '1297482': {       'dev_macs': ['14:21:03:a4:35:f3'],
                                   'ip_address': None,
                                   'name': '4 MAIN ST S (GARDNER ROBYN)',
                                   'serial': '1297482',
                                   'serial_number': '1297482',
                                   'system_mac_address': None},
                '129748B': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6034 PERTH ROAD 119 (SLITS PEDRO & '
                                           'JOLANDA)',
                                   'serial': '129748B',
                                   'serial_number': '129748B',
                                   'system_mac_address': None},
                '129748C': {       'dev_macs': ['14:21:03:a9:73:5f'],
                                   'ip_address': None,
                                   'name': '33 MURRAY CT (BRAUN HELENA)',
                                   'serial': '129748C',
                                   'serial_number': '129748C',
                                   'system_mac_address': None},
                '12974A9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6000 PERTH ROAD 119 (ROPP STEPHEN)',
                                   'serial': '12974A9',
                                   'serial_number': '12974A9',
                                   'system_mac_address': None},
                '12974AC': {       'dev_macs': ['14:21:03:a9:74:88'],
                                   'ip_address': None,
                                   'name': '2733 LINE 45 (ROSE STEVEN & KERRY)',
                                   'serial': '12974AC',
                                   'serial_number': '12974AC',
                                   'system_mac_address': None},
                '12974AF': {       'dev_macs': ['e4:6c:d1:4e:f3:47'],
                                   'ip_address': None,
                                   'name': '109 PUGH ST (DAVIDSON DIANN)',
                                   'serial': '12974AF',
                                   'serial_number': '12974AF',
                                   'system_mac_address': None},
                '12974B2': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7009 ROAD 129 (MARTIN HOWARD)',
                                   'serial': '12974B2',
                                   'serial_number': '12974B2',
                                   'system_mac_address': None},
                '12974D6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '7007 ROAD 129 (KLASSEN ANNA)',
                                   'serial': '12974D6',
                                   'serial_number': '12974D6',
                                   'system_mac_address': None},
                '12974D9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4373 LINE 61 (VIANA KYLA)',
                                   'serial': '12974D9',
                                   'serial_number': '12974D9',
                                   'system_mac_address': None},
                '12974DC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4555 PERTH LINE 72 (ALBRECHT NORMAN)',
                                   'serial': '12974DC',
                                   'serial_number': '12974DC',
                                   'system_mac_address': None},
                '12974E9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4441 PERTH LINE 72 (KLASSEN KATHARIN)',
                                   'serial': '12974E9',
                                   'serial_number': '12974E9',
                                   'system_mac_address': None},
                '12974EC': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '4094 WATERLOO ST (HERRFORT RALPH)',
                                   'serial': '12974EC',
                                   'serial_number': '12974EC',
                                   'system_mac_address': None},
                '12FFD4': {       'dev_macs': ['60:db:98:09:1b:8b'],
                                  'ip_address': None,
                                  'name': '84 Mill St. E. Milverton (Sandi Jacobs '
                                          '10403134)',
                                  'serial': '12FFD4',
                                  'serial_number': '12FFD4',
                                  'system_mac_address': None},
                '12FFD9': {       'dev_macs': ['6c:b0:ce:e2:f6:4a'],
                                  'ip_address': None,
                                  'name': '12 GERBER DR (HUBER ERNA & KEVIN)',
                                  'serial': '12FFD9',
                                  'serial_number': '12FFD9',
                                  'system_mac_address': None},
                '12FFE1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '169 WINSTANLEY ST (KORNELSEN DONOVEN)',
                                  'serial': '12FFE1',
                                  'serial_number': '12FFE1',
                                  'system_mac_address': None},
                '12FFE4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '170 Queen Ave. Monkton (Julia Cork)',
                                  'serial': '12FFE4',
                                  'serial_number': '12FFE4',
                                  'system_mac_address': None},
                '12FFF3': {       'dev_macs': ['44:65:7f:89:0f:93'],
                                  'ip_address': None,
                                  'name': '92 MILL ST E (GRAUL SCHERI)',
                                  'serial': '12FFF3',
                                  'serial_number': '12FFF3',
                                  'system_mac_address': None},
                '12FFFA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '18 Mill St. E. Milverton (Curvin Martin)',
                                  'serial': '12FFFA',
                                  'serial_number': '12FFFA',
                                  'system_mac_address': None},
                '130005': {       'dev_macs': ['d0:76:8f:02:e3:ec'],
                                  'ip_address': None,
                                  'name': '30 Mill St. E. Milverton (HABEL TOM)',
                                  'serial': '130005',
                                  'serial_number': '130005',
                                  'system_mac_address': None},
                '130006': {       'dev_macs': ['14:21:03:d1:7c:bd'],
                                  'ip_address': None,
                                  'name': '86A MILL ST E (CAMPBELL MIKE & TAMMY)',
                                  'serial': '130006',
                                  'serial_number': '130006',
                                  'system_mac_address': None},
                '130026': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '98 Mill St. E. Milverton (Krystina '
                                          'Stachelski 10401363)',
                                  'serial': '130026',
                                  'serial_number': '130026',
                                  'system_mac_address': None},
                '130053': {       'dev_macs': ['e8:9f:80:0b:99:69'],
                                  'ip_address': None,
                                  'name': '90 Mill St. E. Milverton (Debbie Tanner)',
                                  'serial': '130053',
                                  'serial_number': '130053',
                                  'system_mac_address': None},
                '130058': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '68 Mill St. E. Milverton (Nursery School)',
                                  'serial': '130058',
                                  'serial_number': '130058',
                                  'system_mac_address': None},
                '13008E': {       'dev_macs': ['14:21:03:d1:7d:d0'],
                                  'ip_address': None,
                                  'name': '96 MILL ST E (MORWOOD NATHAN)',
                                  'serial': '13008E',
                                  'serial_number': '13008E',
                                  'system_mac_address': None},
                '130091': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '125 Maddison St. E. Monkton (Robert '
                                          'McGuire)',
                                  'serial': '130091',
                                  'serial_number': '130091',
                                  'system_mac_address': None},
                '13009B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '28 MILL ST E (JANTZI PAUL)',
                                  'serial': '13009B',
                                  'serial_number': '13009B',
                                  'system_mac_address': None},
                '13009C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '174 Winstanley St. Monkton',
                                  'serial': '13009C',
                                  'serial_number': '13009C',
                                  'system_mac_address': None},
                '13009F': {       'dev_macs': ['44:65:7f:88:d4:f8'],
                                  'ip_address': None,
                                  'name': '82 Mill St. E. Milverton (Craig Kipfer)',
                                  'serial': '13009F',
                                  'serial_number': '13009F',
                                  'system_mac_address': None},
                '1300A1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '203 king St. Monkton (Joel Leis)',
                                  'serial': '1300A1',
                                  'serial_number': '1300A1',
                                  'system_mac_address': None},
                '1300A2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '21 Maddison Av. E. Monkton (Kyle '
                                          'Donkers10431767)',
                                  'serial': '1300A2',
                                  'serial_number': '1300A2',
                                  'system_mac_address': None},
                '1300C5': {       'dev_macs': ['60:db:98:82:0a:f8'],
                                  'ip_address': None,
                                  'name': '40 MILL ST E (WARING ALICIA)',
                                  'serial': '1300C5',
                                  'serial_number': '1300C5',
                                  'system_mac_address': None},
                '1300C6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '76 Mill St. E. Milverton (Helena Mulder)',
                                  'serial': '1300C6',
                                  'serial_number': '1300C6',
                                  'system_mac_address': None},
                '1300C9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4 Main St. S. Unit 1 Milverton (Huron view '
                                          'Insurance 10411009)',
                                  'serial': '1300C9',
                                  'serial_number': '1300C9',
                                  'system_mac_address': None},
                '1300CB': {       'dev_macs': ['ec:4f:82:ac:f1:56'],
                                  'ip_address': None,
                                  'name': '76 MILL ST E (RICHARDS ALEXIS)',
                                  'serial': '1300CB',
                                  'serial_number': '1300CB',
                                  'system_mac_address': None},
                '1300D6': {       'dev_macs': ['c0:56:27:b5:25:10'],
                                  'ip_address': None,
                                  'name': '4890 Line 61 Milverton (Eddie Jantzi '
                                          '10425951)',
                                  'serial': '1300D6',
                                  'serial_number': '1300D6',
                                  'system_mac_address': None},
                '13448F': {       'dev_macs': [       '5c:0f:fb:3d:80:2c',
                                                      '5c:0f:fb:3d:7c:5e',
                                                      '5c:0f:fb:3d:3f:3e',
                                                      '5c:0f:fb:3d:52:08'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST (POOLE MADALENA)',
                                  'serial': '13448F',
                                  'serial_number': '13448F',
                                  'system_mac_address': None},
                '1380D4': {       'dev_macs': [       'ec:4f:82:a5:04:d2',
                                                      'ec:4f:82:a5:04:9c',
                                                      '30:68:93:49:c2:5d',
                                                      'ec:4f:82:ac:f0:3f',
                                                      'a4:2a:95:3e:22:80',
                                                      'cc:be:59:7e:b6:e3'],
                                  'ip_address': None,
                                  'name': "2 Llana Heights Milverton (Apt #'s 21 thr "
                                          '34)',
                                  'serial': '1380D4',
                                  'serial_number': '1380D4',
                                  'system_mac_address': None},
                '138139': {       'dev_macs': [       '08:62:66:9a:51:64',
                                                      'cc:32:e5:17:f2:45',
                                                      '14:21:03:8b:71:76'],
                                  'ip_address': None,
                                  'name': '54 PUGH ST (WOODS RUTH A)',
                                  'serial': '138139',
                                  'serial_number': '138139',
                                  'system_mac_address': None},
                '13813B': {       'dev_macs': ['48:a9:8a:75:88:59'],
                                  'ip_address': None,
                                  'name': '4705 Line 64 Milverton (Wagler Electric '
                                          '10400140)',
                                  'serial': '13813B',
                                  'serial_number': '13813B',
                                  'system_mac_address': None},
                '13813C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '50 William St. Milverton (Knollcrest Data '
                                          'Rm. 3)',
                                  'serial': '13813C',
                                  'serial_number': '13813C',
                                  'system_mac_address': None},
                '1520286': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '55 WILLIAM ST AP 104 (FOX LORI)',
                                   'serial': '1520286',
                                   'serial_number': '1520286',
                                   'system_mac_address': None},
                '152029D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '57 WILLIAM ST APT 201 (BOBAN AISWARYA)',
                                   'serial': '152029D',
                                   'serial_number': '152029D',
                                   'system_mac_address': None},
                '15202A6': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': "16 MAIN ST N (O'HEARN DANIEL)",
                                   'serial': '15202A6',
                                   'serial_number': '15202A6',
                                   'system_mac_address': None},
                '15202C7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '55 WILLIAM ST (ASSELSTINE MAGDELINA)',
                                   'serial': '15202C7',
                                   'serial_number': '15202C7',
                                   'system_mac_address': None},
                '15202CF': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '13 MILL ST E (SOCKETT KAREN)',
                                   'serial': '15202CF',
                                   'serial_number': '15202CF',
                                   'system_mac_address': None},
                '15242B7': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '25 MURRAY CT (COLE DANIELLE)',
                                   'serial': '15242B7',
                                   'serial_number': '15242B7',
                                   'system_mac_address': None},
                '15242D5': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '55 WILLIAM ST (ALBRECHT MATTHEW)',
                                   'serial': '15242D5',
                                   'serial_number': '15242D5',
                                   'system_mac_address': None},
                '153D8': {       'dev_macs': ['44:65:7f:88:d2:0d', 'a8:42:a1:2c:71:d0'],
                                 'ip_address': None,
                                 'name': '50 William St. Milverton (Countryside '
                                         'Midwifery Services 10400769)',
                                 'serial': '153D8',
                                 'serial_number': '153D8',
                                 'system_mac_address': None},
                '153E639': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6888 ROAD 124 (KIPFER LORETTA)',
                                   'serial': '153E639',
                                   'serial_number': '153E639',
                                   'system_mac_address': None},
                '155322': {       'dev_macs': ['00:03:e6:d8:e4:4c'],
                                  'ip_address': None,
                                  'name': '60 William St. Suite 108 (Betty Brubacher '
                                          '10401029)',
                                  'serial': '155322',
                                  'serial_number': '155322',
                                  'system_mac_address': None},
                '1553BA': {       'dev_macs': ['60:db:98:09:39:10'],
                                  'ip_address': None,
                                  'name': '93 Maddison Ave. Monkton (Mike Duncan '
                                          '10432186)',
                                  'serial': '1553BA',
                                  'serial_number': '1553BA',
                                  'system_mac_address': None},
                '1553BF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '195 Winstanley St Monkton (Ron McKee '
                                          '10433696)',
                                  'serial': '1553BF',
                                  'serial_number': '1553BF',
                                  'system_mac_address': None},
                '1553C6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '75 Maddison St. W. Monkton (Megan Mclean)',
                                  'serial': '1553C6',
                                  'serial_number': '1553C6',
                                  'system_mac_address': None},
                '1553D4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 William St. Suite 112 (DEZELL '
                                          'BERNADETTE)',
                                  'serial': '1553D4',
                                  'serial_number': '1553D4',
                                  'system_mac_address': None},
                '155436': {       'dev_macs': ['14:21:03:8b:6a:8b'],
                                  'ip_address': None,
                                  'name': '60 William St. Suite 111 (ERB CONNIE)',
                                  'serial': '155436',
                                  'serial_number': '155436',
                                  'system_mac_address': None},
                '155460': {       'dev_macs': ['60:db:98:55:a3:1f'],
                                  'ip_address': None,
                                  'name': '164 WINSTANLEY ST (CARAPINHA KYLA)',
                                  'serial': '155460',
                                  'serial_number': '155460',
                                  'system_mac_address': None},
                '15549F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '40 Cobalt St. N. Milverton (Lorne Reis)',
                                  'serial': '15549F',
                                  'serial_number': '15549F',
                                  'system_mac_address': None},
                '158FA71': {       'dev_macs': ['00:30:44:3e:1a:55'],
                                   'ip_address': None,
                                   'name': '6433 PERTH ROAD 131 (GROWMARK,INC. D/B/A '
                                           'CENTRAL)',
                                   'serial': '158FA71',
                                   'serial_number': '158FA71',
                                   'system_mac_address': None},
                '158FA8F': {       'dev_macs': ['b4:fb:e4:d3:ff:15'],
                                   'ip_address': None,
                                   'name': '4478 PERTH LINE 72 (NEW STAR '
                                           'RENTALS&EQUIPMENT LTD)',
                                   'serial': '158FA8F',
                                   'serial_number': '158FA8F',
                                   'system_mac_address': None},
                '158FB6E': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '5592 DUKE STREET UNIT C (SEW-WEAR MFG CO '
                                           'LTD)',
                                   'serial': '158FB6E',
                                   'serial_number': '158FB6E',
                                   'system_mac_address': None},
                '159F7F9': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6751 ROAD 124 (KUEPFER ARNOLD)',
                                   'serial': '159F7F9',
                                   'serial_number': '159F7F9',
                                   'system_mac_address': None},
                '159F7FE': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '6780 PERTH ROAD 121 (ALBRECHT ALVIN K)',
                                   'serial': '159F7FE',
                                   'serial_number': '159F7FE',
                                   'system_mac_address': None},
                '1606428': {       'dev_macs': [       'c8:4f:86:fc:00:03',
                                                       '60:be:b4:13:b6:c0'],
                                   'ip_address': None,
                                   'name': '40 Temperance St. Milverton (Perth East '
                                           'Complex 10401202)',
                                   'serial': '1606428',
                                   'serial_number': '1606428',
                                   'system_mac_address': None},
                '160D15': {       'dev_macs': ['ec:4f:82:2c:75:23'],
                                  'ip_address': None,
                                  'name': '22 Mill St. W. Milverton (County Of Perth '
                                          'EMS Base 10410443)',
                                  'serial': '160D15',
                                  'serial_number': '160D15',
                                  'system_mac_address': None},
                '161027': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 MADDISON ST E',
                                  'serial': '161027',
                                  'serial_number': '161027',
                                  'system_mac_address': None},
                '175E39': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '3 Jo-Lene Ct. Milverton (Bill&June Coulter)',
                                  'serial': '175E39',
                                  'serial_number': '175E39',
                                  'system_mac_address': None},
                '175E6A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 William St. Suite 106 (Clara Looser '
                                          '10400400)',
                                  'serial': '175E6A',
                                  'serial_number': '175E6A',
                                  'system_mac_address': None},
                '175E70': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5340 Line 60 Milverton (C VanDerStraeten '
                                          '10434906)',
                                  'serial': '175E70',
                                  'serial_number': '175E70',
                                  'system_mac_address': None},
                '175E72': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 WILLIAM ST (ZEHR SALLY)',
                                  'serial': '175E72',
                                  'serial_number': '175E72',
                                  'system_mac_address': None},
                '175EBA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '1 Pacific Ave. Milverton',
                                  'serial': '175EBA',
                                  'serial_number': '175EBA',
                                  'system_mac_address': None},
                '176215': {       'dev_macs': ['a8:91:62:00:96:13'],
                                  'ip_address': None,
                                  'name': '4700 Line 61 (Perth East Shop 10401421)',
                                  'serial': '176215',
                                  'serial_number': '176215',
                                  'system_mac_address': None},
                '176239': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 William St. Suite 109',
                                  'serial': '176239',
                                  'serial_number': '176239',
                                  'system_mac_address': None},
                '17624E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 PACIFIC AV (SHOULDICE DARRYL & CHARIS)',
                                  'serial': '17624E',
                                  'serial_number': '17624E',
                                  'system_mac_address': None},
                '1999A7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '215 Nelson St. Monkton',
                                  'serial': '1999A7',
                                  'serial_number': '1999A7',
                                  'system_mac_address': None},
                '1999FF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4186 LINE 67 (JANTZI MAHLON K)',
                                  'serial': '1999FF',
                                  'serial_number': '1999FF',
                                  'system_mac_address': None},
                '199A2E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '190 Selby Ave. Monkton (Mandy Smith)',
                                  'serial': '199A2E',
                                  'serial_number': '199A2E',
                                  'system_mac_address': None},
                '199A3A': {       'dev_macs': ['7c:5a:1c:82:a3:57'],
                                  'ip_address': None,
                                  'name': '15 Davis St. Milverton (Gary Boshart)',
                                  'serial': '199A3A',
                                  'serial_number': '199A3A',
                                  'system_mac_address': None},
                '1C5BA6': {       'dev_macs': ['5c:0f:fb:2a:c7:aa'],
                                  'ip_address': None,
                                  'name': '40 Temperance St. Milverton (Mornington '
                                          'Communications)',
                                  'serial': '1C5BA6',
                                  'serial_number': '1C5BA6',
                                  'system_mac_address': None},
                '1C5CB7': {       'dev_macs': ['00:03:e6:f9:fb:7e'],
                                  'ip_address': None,
                                  'name': '60 WILLIAM ST (LUCKHARDT MARION)',
                                  'serial': '1C5CB7',
                                  'serial_number': '1C5CB7',
                                  'system_mac_address': None},
                '1C5CE9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 WILLIAM ST (SCHWEITZER DIANE)',
                                  'serial': '1C5CE9',
                                  'serial_number': '1C5CE9',
                                  'system_mac_address': None},
                '1C5D18': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '73 SCHADE ST (SCHMIDT RUTH)',
                                  'serial': '1C5D18',
                                  'serial_number': '1C5D18',
                                  'system_mac_address': None},
                '1CAB1B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5 SIPPEL ST (CHAPMAN MICHELE)',
                                  'serial': '1CAB1B',
                                  'serial_number': '1CAB1B',
                                  'system_mac_address': None},
                '1CAB4A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 TEMPERANCE ST (MCKONE TYLER R)',
                                  'serial': '1CAB4A',
                                  'serial_number': '1CAB4A',
                                  'system_mac_address': None},
                '1CAB56': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '34 William St. Milverton (Melanie Sippel)',
                                  'serial': '1CAB56',
                                  'serial_number': '1CAB56',
                                  'system_mac_address': None},
                '1CC464': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'Milverton CO (844G)',
                                  'serial': '1CC464',
                                  'serial_number': '1CC464',
                                  'system_mac_address': None},
                '1CC4B3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6 Macbeth St (Dan Gerber 10400548)',
                                  'serial': '1CC4B3',
                                  'serial_number': '1CC4B3',
                                  'system_mac_address': None},
                '1D0335': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '12 Pierson St. Milverton (Kyle Baillie)',
                                  'serial': '1D0335',
                                  'serial_number': '1D0335',
                                  'system_mac_address': None},
                '1D0337': {       'dev_macs': ['14:21:03:d1:7d:83'],
                                  'ip_address': None,
                                  'name': '16 MAIN ST N (PATTERSON REBEKKAH)',
                                  'serial': '1D0337',
                                  'serial_number': '1D0337',
                                  'system_mac_address': None},
                '1D033F': {       'dev_macs': ['d0:76:8f:02:ef:7b'],
                                  'ip_address': None,
                                  'name': '9 WILFRED ST (PURVES ALI JOY)',
                                  'serial': '1D033F',
                                  'serial_number': '1D033F',
                                  'system_mac_address': None},
                '1D0343': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '31 William St (Ken Lehmann 10400650)',
                                  'serial': '1D0343',
                                  'serial_number': '1D0343',
                                  'system_mac_address': None},
                '1D0345': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 WILLIAM ST (CHAINEY KEVIN)',
                                  'serial': '1D0345',
                                  'serial_number': '1D0345',
                                  'system_mac_address': None},
                '1D0347': {       'dev_macs': ['d0:76:8f:1f:57:30'],
                                  'ip_address': None,
                                  'name': '30 WILLIAM ST (HARRIS CASSIDY)',
                                  'serial': '1D0347',
                                  'serial_number': '1D0347',
                                  'system_mac_address': None},
                '1D04EC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '22 Mill St E Milverton (Becky '
                                          'Steckly10422799)',
                                  'serial': '1D04EC',
                                  'serial_number': '1D04EC',
                                  'system_mac_address': None},
                '1D04EE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '47 William St. Milverton (SABRI AKHMAD)',
                                  'serial': '1D04EE',
                                  'serial_number': '1D04EE',
                                  'system_mac_address': None},
                '1D04F0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '27 William St. Milverton (Donald Schneider)',
                                  'serial': '1D04F0',
                                  'serial_number': '1D04F0',
                                  'system_mac_address': None},
                '1D04F1': {       'dev_macs': ['14:21:03:a9:7f:72'],
                                  'ip_address': None,
                                  'name': '4 WHALEY AVE. (Rod Steckly)',
                                  'serial': '1D04F1',
                                  'serial_number': '1D04F1',
                                  'system_mac_address': None},
                '1D04F2': {       'dev_macs': ['d0:76:8f:02:e2:b8'],
                                  'ip_address': None,
                                  'name': '49 WILLIAM ST (BOWMAN LAURA)',
                                  'serial': '1D04F2',
                                  'serial_number': '1D04F2',
                                  'system_mac_address': None},
                '1D04F5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 MAIN ST N (SCHMITT TONY)',
                                  'serial': '1D04F5',
                                  'serial_number': '1D04F5',
                                  'system_mac_address': None},
                '1D04F6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '40 William St. Milverton (Merton Kuepfer)',
                                  'serial': '1D04F6',
                                  'serial_number': '1D04F6',
                                  'system_mac_address': None},
                '1D04F7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '25 Temperance St. Milverton (Cathy '
                                          'Wettlaufer)',
                                  'serial': '1D04F7',
                                  'serial_number': '1D04F7',
                                  'system_mac_address': None},
                '1D04F8': {       'dev_macs': ['60:a4:b7:dd:29:f3'],
                                  'ip_address': None,
                                  'name': '25 WILLIAM ST (ELROY WAGLER MASONRY)',
                                  'serial': '1D04F8',
                                  'serial_number': '1D04F8',
                                  'system_mac_address': None},
                '1D04FA': {       'dev_macs': ['d0:76:8f:02:ed:1e'],
                                  'ip_address': None,
                                  'name': '21 FULTON ST (BOWMAN RON & TRACEY)',
                                  'serial': '1D04FA',
                                  'serial_number': '1D04FA',
                                  'system_mac_address': None},
                '1D04FB': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 FULTON ST (DEVRIES MARTIN)',
                                  'serial': '1D04FB',
                                  'serial_number': '1D04FB',
                                  'system_mac_address': None},
                '1D04FC': {       'dev_macs': ['5c:0f:fb:2a:c7:1e'],
                                  'ip_address': None,
                                  'name': '3 John St. Milverton (Fred Jones 10401195)',
                                  'serial': '1D04FC',
                                  'serial_number': '1D04FC',
                                  'system_mac_address': None},
                '1D04FD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '10 JOHN ST (GOWING COLE)',
                                  'serial': '1D04FD',
                                  'serial_number': '1D04FD',
                                  'system_mac_address': None},
                '1D0500': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '10 Adelaide St. Milverton (Jeremy Sippel '
                                          '10420159)',
                                  'serial': '1D0500',
                                  'serial_number': '1D0500',
                                  'system_mac_address': None},
                '1D0501': {       'dev_macs': ['d0:76:8f:1f:6a:f4'],
                                  'ip_address': None,
                                  'name': '1 Grosch St. Milverton (Joyce Riddell '
                                          '10432778)',
                                  'serial': '1D0501',
                                  'serial_number': '1D0501',
                                  'system_mac_address': None},
                '1D0502': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 Pugh St Milverton (Greg Storey)',
                                  'serial': '1D0502',
                                  'serial_number': '1D0502',
                                  'system_mac_address': None},
                '1D0503': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 ADELAIDE ST',
                                  'serial': '1D0503',
                                  'serial_number': '1D0503',
                                  'system_mac_address': None},
                '1D0504': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '17 MacBeth St. (David Schultz)',
                                  'serial': '1D0504',
                                  'serial_number': '1D0504',
                                  'system_mac_address': None},
                '1D0505': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 Wilfred St. Milverton (Gwen Moffat '
                                          '10431370)',
                                  'serial': '1D0505',
                                  'serial_number': '1D0505',
                                  'system_mac_address': None},
                '1D0506': {       'dev_macs': ['60:db:98:09:1b:96'],
                                  'ip_address': None,
                                  'name': '25A Temperance St. Milverton (Vince '
                                          'Etchells)',
                                  'serial': '1D0506',
                                  'serial_number': '1D0506',
                                  'system_mac_address': None},
                '1D0508': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '14 Fulton St. Milverton (Nathan Matheson '
                                          '10426412)',
                                  'serial': '1D0508',
                                  'serial_number': '1D0508',
                                  'system_mac_address': None},
                '1D050A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '12 John St. Milverton (Jim Abraham)',
                                  'serial': '1D050A',
                                  'serial_number': '1D050A',
                                  'system_mac_address': None},
                '1D050C': {       'dev_macs': ['44:65:7f:89:0d:89'],
                                  'ip_address': None,
                                  'name': '15 Fulton St. Milverton (Becky Westman '
                                          '10427702)',
                                  'serial': '1D050C',
                                  'serial_number': '1D050C',
                                  'system_mac_address': None},
                '1D050D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 George St. Milverton',
                                  'serial': '1D050D',
                                  'serial_number': '1D050D',
                                  'system_mac_address': None},
                '1D050F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '19 Whaley Ave. Milverton (Monica&Joe Phelan '
                                          '10421673)',
                                  'serial': '1D050F',
                                  'serial_number': '1D050F',
                                  'system_mac_address': None},
                '1D0510': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '18 William St. Milverton (Landon Sellers)',
                                  'serial': '1D0510',
                                  'serial_number': '1D0510',
                                  'system_mac_address': None},
                '1D0512': {       'dev_macs': ['d0:76:8f:02:f1:54'],
                                  'ip_address': None,
                                  'name': '23 Fulton St. Milverton (Doug Egerdee)',
                                  'serial': '1D0512',
                                  'serial_number': '1D0512',
                                  'system_mac_address': None},
                '1D067B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'William Nahrgang (12 George)',
                                  'serial': '1D067B',
                                  'serial_number': '1D067B',
                                  'system_mac_address': None},
                '1D067F': {       'dev_macs': ['ec:4f:82:a5:05:aa'],
                                  'ip_address': None,
                                  'name': '201 Winstanley St. Monkton (Neil Ward)',
                                  'serial': '1D067F',
                                  'serial_number': '1D067F',
                                  'system_mac_address': None},
                '1D0683': {       'dev_macs': ['b8:94:70:9d:af:21'],
                                  'ip_address': None,
                                  'name': '174 Maddison St. E. Monkton (Roy Beuerman)',
                                  'serial': '1D0683',
                                  'serial_number': '1D0683',
                                  'system_mac_address': None},
                '1D0687': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '42 William St. Milverton (Robert Sippel)',
                                  'serial': '1D0687',
                                  'serial_number': '1D0687',
                                  'system_mac_address': None},
                '1D0688': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '22 Pacific Ave. Milverton (Kevin Rich)',
                                  'serial': '1D0688',
                                  'serial_number': '1D0688',
                                  'system_mac_address': None},
                '1D0689': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 William St. Apt # 105 (Amanda Bush)',
                                  'serial': '1D0689',
                                  'serial_number': '1D0689',
                                  'system_mac_address': None},
                '1D068A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '53 Temperance St. Milverton (Heather Arand '
                                          '10401553)',
                                  'serial': '1D068A',
                                  'serial_number': '1D068A',
                                  'system_mac_address': None},
                '1D068B': {       'dev_macs': ['84:d3:43:6d:ed:bb'],
                                  'ip_address': None,
                                  'name': '112 Mill St. E. Milverton (Earl Leis)',
                                  'serial': '1D068B',
                                  'serial_number': '1D068B',
                                  'system_mac_address': None},
                '1D068C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '22 Pacific Ave. Milverton (Robert Paton)',
                                  'serial': '1D068C',
                                  'serial_number': '1D068C',
                                  'system_mac_address': None},
                '1D068D': {       'dev_macs': ['b8:94:70:f3:e5:5e'],
                                  'ip_address': None,
                                  'name': '44 Gerber Dr. Milverton (Ryan Harloff)',
                                  'serial': '1D068D',
                                  'serial_number': '1D068D',
                                  'system_mac_address': None},
                '1D068F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '108 Mill St. E. Milverton (Larry Kropf)',
                                  'serial': '1D068F',
                                  'serial_number': '1D068F',
                                  'system_mac_address': None},
                '1D0691': {       'dev_macs': ['d0:76:8f:91:d3:0f'],
                                  'ip_address': None,
                                  'name': '46 William St. Milverton (BRUBACHER MARK)',
                                  'serial': '1D0691',
                                  'serial_number': '1D0691',
                                  'system_mac_address': None},
                '1D0692': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '9 WEST ST (TRAINOR CASSIDY)',
                                  'serial': '1D0692',
                                  'serial_number': '1D0692',
                                  'system_mac_address': None},
                '1D0694': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '9 John St. Milverton (Shannon Cairns '
                                          '10432740)',
                                  'serial': '1D0694',
                                  'serial_number': '1D0694',
                                  'system_mac_address': None},
                '1D0695': {       'dev_macs': ['44:65:7f:5d:bd:9f'],
                                  'ip_address': None,
                                  'name': '41 Temperance St. Milverton (Charlene '
                                          'Lehman 10433199)',
                                  'serial': '1D0695',
                                  'serial_number': '1D0695',
                                  'system_mac_address': None},
                '1D069A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '18 Fulton St. Milverton (Joe Schlueter)',
                                  'serial': '1D069A',
                                  'serial_number': '1D069A',
                                  'system_mac_address': None},
                '1D069B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '13 MacBeth St. Milverton (James '
                                          'Adams-Kipfer)',
                                  'serial': '1D069B',
                                  'serial_number': '1D069B',
                                  'system_mac_address': None},
                '1D069D': {       'dev_macs': ['84:d3:43:0f:62:99'],
                                  'ip_address': None,
                                  'name': '6 WHALEY AV (PHILLIPS JODI)',
                                  'serial': '1D069D',
                                  'serial_number': '1D069D',
                                  'system_mac_address': None},
                '1D069F': {       'dev_macs': ['14:21:03:d1:88:e6'],
                                  'ip_address': None,
                                  'name': '11 PACIFIC AV (WEBER SARAH)',
                                  'serial': '1D069F',
                                  'serial_number': '1D069F',
                                  'system_mac_address': None},
                '1D06A1': {       'dev_macs': ['04:bc:9f:46:95:45'],
                                  'ip_address': None,
                                  'name': '14 Whaley Ave. Milverton (MCARTHUR JIM)',
                                  'serial': '1D06A1',
                                  'serial_number': '1D06A1',
                                  'system_mac_address': None},
                '1D06B9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 Whaley Ave. Milverton (Sherri Gilbert)',
                                  'serial': '1D06B9',
                                  'serial_number': '1D06B9',
                                  'system_mac_address': None},
                '1D06CD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 Mill St. E. Milverton (William Jantzi '
                                          '10432843)',
                                  'serial': '1D06CD',
                                  'serial_number': '1D06CD',
                                  'system_mac_address': None},
                '1D06D1': {       'dev_macs': ['d0:76:8f:91:d4:bc'],
                                  'ip_address': None,
                                  'name': '26 MAIN ST S (HOWE NICOLE)',
                                  'serial': '1D06D1',
                                  'serial_number': '1D06D1',
                                  'system_mac_address': None},
                '1D06D7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 WHALEY AV (STECKLY DOUG)',
                                  'serial': '1D06D7',
                                  'serial_number': '1D06D7',
                                  'system_mac_address': None},
                '1D06E1': {       'dev_macs': ['14:21:03:d1:8a:e0'],
                                  'ip_address': None,
                                  'name': '217 Winstanley St. Monkton (Mike Chambers)',
                                  'serial': '1D06E1',
                                  'serial_number': '1D06E1',
                                  'system_mac_address': None},
                '1D0714': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '10 Whaley Ave. Milverton (Dave Dickson)',
                                  'serial': '1D0714',
                                  'serial_number': '1D0714',
                                  'system_mac_address': None},
                '1D0718': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '17 Mill St. W. (Peter Wall 10423210)',
                                  'serial': '1D0718',
                                  'serial_number': '1D0718',
                                  'system_mac_address': None},
                '1FD6B7': {       'dev_macs': ['70:4c:a5:6d:0c:b0'],
                                  'ip_address': None,
                                  'name': '68 Mill St. E. Milverton (Milverton Public '
                                          'School)',
                                  'serial': '1FD6B7',
                                  'serial_number': '1FD6B7',
                                  'system_mac_address': None},
                '1FD6DB': {       'dev_macs': [       '14:21:03:8b:71:55',
                                                      'ec:4f:82:a5:04:54',
                                                      'ec:4f:82:ac:eb:89',
                                                      '04:bc:9f:46:94:c1'],
                                  'ip_address': None,
                                  'name': "2 Llana Heights Milverton (Apt #'s "
                                          '11-12-14-15)',
                                  'serial': '1FD6DB',
                                  'serial_number': '1FD6DB',
                                  'system_mac_address': None},
                '205B38': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 MAIN ST S (COOK CONSTANCE)',
                                  'serial': '205B38',
                                  'serial_number': '205B38',
                                  'system_mac_address': None},
                '205B3C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 William St Apt 3',
                                  'serial': '205B3C',
                                  'serial_number': '205B3C',
                                  'system_mac_address': None},
                '205BDC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4 Winstanley St Monkton (Cherie Smith '
                                          '10430700)',
                                  'serial': '205BDC',
                                  'serial_number': '205BDC',
                                  'system_mac_address': None},
                '205D97': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 MAIN ST S (MCCUTCHEN WINNIE)',
                                  'serial': '205D97',
                                  'serial_number': '205D97',
                                  'system_mac_address': None},
                '205D9B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4 Mill St E Milverton Apt 2 (Maxine '
                                          'Kuepfer)',
                                  'serial': '205D9B',
                                  'serial_number': '205D9B',
                                  'system_mac_address': None},
                '205D9C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 William St Milverton (Jonathan Siderius '
                                          '10434882)',
                                  'serial': '205D9C',
                                  'serial_number': '205D9C',
                                  'system_mac_address': None},
                '205DA4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '43 Maddison St E Monkton (Christine Rose '
                                          '10432948)',
                                  'serial': '205DA4',
                                  'serial_number': '205DA4',
                                  'system_mac_address': None},
                '205DB1': {       'dev_macs': ['b8:94:70:bc:b9:1f'],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (ANDRE JOLY)',
                                  'serial': '205DB1',
                                  'serial_number': '205DB1',
                                  'system_mac_address': None},
                '205DB9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '104 Mill St. E. Milverton (Doug Fyfe '
                                          '10402350)',
                                  'serial': '205DB9',
                                  'serial_number': '205DB9',
                                  'system_mac_address': None},
                '205DBB': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 William St Apt 203 Milverton (Diane '
                                          'Playford 10435440)',
                                  'serial': '205DBB',
                                  'serial_number': '205DBB',
                                  'system_mac_address': None},
                '205DBC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (PAULNITZ CATHY)',
                                  'serial': '205DBC',
                                  'serial_number': '205DBC',
                                  'system_mac_address': None},
                '205DBD': {       'dev_macs': ['14:21:03:d1:64:8c'],
                                  'ip_address': None,
                                  'name': '23 PACIFIC AV (WILLIAMS TY)',
                                  'serial': '205DBD',
                                  'serial_number': '205DBD',
                                  'system_mac_address': None},
                '205DC0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '15 MAIN ST APT. 202 (GOETZ MATTHEW)',
                                  'serial': '205DC0',
                                  'serial_number': '205DC0',
                                  'system_mac_address': None},
                '205DC2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 MAIN ST N',
                                  'serial': '205DC2',
                                  'serial_number': '205DC2',
                                  'system_mac_address': None},
                '205DC3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '78 MILL ST E (CARTER ROGER & MICHELLE)',
                                  'serial': '205DC3',
                                  'serial_number': '205DC3',
                                  'system_mac_address': None},
                '205DC5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '26 MAIN ST N',
                                  'serial': '205DC5',
                                  'serial_number': '205DC5',
                                  'system_mac_address': None},
                '205DC8': {       'dev_macs': ['5c:0f:fb:3d:80:66'],
                                  'ip_address': None,
                                  'name': '60 WILLIAM ST STE 104 (COSSEY JEAN)',
                                  'serial': '205DC8',
                                  'serial_number': '205DC8',
                                  'system_mac_address': None},
                '205DCB': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '44 Main St. N. Milverton (David Weber)',
                                  'serial': '205DCB',
                                  'serial_number': '205DCB',
                                  'system_mac_address': None},
                '205DCF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '65 MAIN ST S (HAHN LEWIS)',
                                  'serial': '205DCF',
                                  'serial_number': '205DCF',
                                  'system_mac_address': None},
                '205DD1': {       'dev_macs': ['44:65:7f:3e:98:0e'],
                                  'ip_address': None,
                                  'name': '20 Pacific Ave. Milverton (Debbie '
                                          'Timmermans10401639)',
                                  'serial': '205DD1',
                                  'serial_number': '205DD1',
                                  'system_mac_address': None},
                '205DDC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4 Main St S Apt 2 Milverton (DUNBAR SARAH)',
                                  'serial': '205DDC',
                                  'serial_number': '205DDC',
                                  'system_mac_address': None},
                '205DDE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '12 PUGH ST W (PRISECARU TIMOTEI)',
                                  'serial': '205DDE',
                                  'serial_number': '205DDE',
                                  'system_mac_address': None},
                '205DE3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '34 Pacific Av Milverton (Jeremy Matheson)',
                                  'serial': '205DE3',
                                  'serial_number': '205DE3',
                                  'system_mac_address': None},
                '205DED': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '82 Reagan St. Milverton (Craig & Julie '
                                          'Purvis 10435983)',
                                  'serial': '205DED',
                                  'serial_number': '205DED',
                                  'system_mac_address': None},
                '205DEE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '76 MAIN ST S (UHC OF WESTERN ONTARIO MCO '
                                          '839)',
                                  'serial': '205DEE',
                                  'serial_number': '205DEE',
                                  'system_mac_address': None},
                '205DF1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM STREET APT 201 (HARVEY JANE)',
                                  'serial': '205DF1',
                                  'serial_number': '205DF1',
                                  'system_mac_address': None},
                '205DF4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4 MAIN ST S (CASKANETTE BENJAMEN)',
                                  'serial': '205DF4',
                                  'serial_number': '205DF4',
                                  'system_mac_address': None},
                '205DF6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '45 Main St. N. Milverton (Greg Guenther)',
                                  'serial': '205DF6',
                                  'serial_number': '205DF6',
                                  'system_mac_address': None},
                '205DF8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '42 Main St S Milverton (Cook Millwrighting '
                                          '10401478)',
                                  'serial': '205DF8',
                                  'serial_number': '205DF8',
                                  'system_mac_address': None},
                '205DFC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6480 Perth Rd. 131 (Fritz Ryter)',
                                  'serial': '205DFC',
                                  'serial_number': '205DFC',
                                  'system_mac_address': None},
                '205DFE': {       'dev_macs': ['14:21:03:92:ea:76'],
                                  'ip_address': None,
                                  'name': '4 George St. Milverton (Tom Broda)',
                                  'serial': '205DFE',
                                  'serial_number': '205DFE',
                                  'system_mac_address': None},
                '205E03': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '57 Temperance St. Milverton (Jim Schneuker '
                                          '10401435)',
                                  'serial': '205E03',
                                  'serial_number': '205E03',
                                  'system_mac_address': None},
                '205E05': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '62 MAIN STREET N (FOBEL RICHARD)',
                                  'serial': '205E05',
                                  'serial_number': '205E05',
                                  'system_mac_address': None},
                '205E06': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (BARONETTE ROGER)',
                                  'serial': '205E06',
                                  'serial_number': '205E06',
                                  'system_mac_address': None},
                '205E1E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (HUBERTS KIRK)',
                                  'serial': '205E1E',
                                  'serial_number': '205E1E',
                                  'system_mac_address': None},
                '205E20': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '13 Mill St.. W. Milverton (Trevor Woods)',
                                  'serial': '205E20',
                                  'serial_number': '205E20',
                                  'system_mac_address': None},
                '237A25': {       'dev_macs': ['cc:be:59:32:94:ce'],
                                  'ip_address': None,
                                  'name': '37 Mill St. W. Milverton (John Gerber '
                                          'Construction 10401482)',
                                  'serial': '237A25',
                                  'serial_number': '237A25',
                                  'system_mac_address': None},
                '237AE5': {       'dev_macs': [       'ec:4f:82:12:e4:cd',
                                                      '5c:0f:fb:2f:cc:9c'],
                                  'ip_address': None,
                                  'name': '14 Mill St. W. Milverton (Mill Wood Manor)',
                                  'serial': '237AE5',
                                  'serial_number': '237AE5',
                                  'system_mac_address': None},
                '248ACB': {       'dev_macs': ['ec:4f:82:a5:13:0f'],
                                  'ip_address': None,
                                  'name': '45A & 45B William St. (Norma& Neil Wilker)',
                                  'serial': '248ACB',
                                  'serial_number': '248ACB',
                                  'system_mac_address': None},
                '248AD4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'Derick Stolz 10430219 (52 schade st)',
                                  'serial': '248AD4',
                                  'serial_number': '248AD4',
                                  'system_mac_address': None},
                '25982A': {       'dev_macs': ['ec:4f:82:2d:32:6b'],
                                  'ip_address': None,
                                  'name': '36 Main St. N. Milverton (Milverton '
                                          'Foodtown)',
                                  'serial': '25982A',
                                  'serial_number': '25982A',
                                  'system_mac_address': None},
                '259837': {       'dev_macs': ['00:0d:b9:57:ae:71'],
                                  'ip_address': None,
                                  'name': '6440 Perth Rd. 131 (Leis Feed 10422648)',
                                  'serial': '259837',
                                  'serial_number': '259837',
                                  'system_mac_address': None},
                '25D92E': {       'dev_macs': ['ec:4f:82:ac:ed:f6'],
                                  'ip_address': None,
                                  'name': '35 Main St. S. Milverton (Randy Martin '
                                          '10434943)',
                                  'serial': '25D92E',
                                  'serial_number': '25D92E',
                                  'system_mac_address': None},
                '260E6A': {       'dev_macs': ['d4:76:a0:a3:53:9a'],
                                  'ip_address': None,
                                  'name': '14 MILL ST W (VERELLEN GREG R)',
                                  'serial': '260E6A',
                                  'serial_number': '260E6A',
                                  'system_mac_address': None},
                '260E96': {       'dev_macs': ['70:4c:a5:f1:71:7a'],
                                  'ip_address': None,
                                  'name': '1 Main St. S. Milverton (Zehr Insurance)',
                                  'serial': '260E96',
                                  'serial_number': '260E96',
                                  'system_mac_address': None},
                '260E97': {       'dev_macs': [       'cc:be:59:7e:ba:ee',
                                                      '5c:0f:fb:2a:0a:48',
                                                      '44:65:7f:88:47:41',
                                                      '58:d9:d5:1a:75:99',
                                                      '44:65:7f:89:0a:44'],
                                  'ip_address': None,
                                  'name': '14 MILL ST W',
                                  'serial': '260E97',
                                  'serial_number': '260E97',
                                  'system_mac_address': None},
                '2625E5': {       'dev_macs': ['00:f8:2c:7f:5e:70'],
                                  'ip_address': None,
                                  'name': '12 Main St. S. Milverton (Kindred Credit '
                                          'Union)',
                                  'serial': '2625E5',
                                  'serial_number': '2625E5',
                                  'system_mac_address': None},
                '2625E7': {       'dev_macs': [       '00:03:e6:a4:db:66',
                                                      '00:03:e6:dd:32:0c',
                                                      '00:03:e6:a5:73:18',
                                                      '5c:0f:fb:43:96:d2',
                                                      '5c:0f:fb:3d:58:16'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST',
                                  'serial': '2625E7',
                                  'serial_number': '2625E7',
                                  'system_mac_address': None},
                '2625EA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '48 PUGH ST',
                                  'serial': '2625EA',
                                  'serial_number': '2625EA',
                                  'system_mac_address': None},
                '263200': {       'dev_macs': [       '00:03:e6:db:46:da',
                                                      '00:03:e6:a5:6e:9a',
                                                      '5c:0f:fb:3d:5b:d8',
                                                      'cc:be:59:c3:48:5f',
                                                      '00:03:e6:dd:30:52',
                                                      '5c:0f:fb:3d:74:50',
                                                      '5c:0f:fb:42:65:20'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST (BRUCE WILLIAM)',
                                  'serial': '263200',
                                  'serial_number': '263200',
                                  'system_mac_address': None},
                '263201': {       'dev_macs': [       '5c:0f:fb:15:4f:0a',
                                                      '5c:0f:fb:3d:5b:e0',
                                                      '5c:0f:fb:3d:58:28'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST',
                                  'serial': '263201',
                                  'serial_number': '263201',
                                  'system_mac_address': None},
                '263222': {       'dev_macs': [       '00:03:e6:a4:da:68',
                                                      '5c:0f:fb:3d:7d:a2',
                                                      '00:03:e6:a4:db:58',
                                                      '00:03:e6:a5:73:9e',
                                                      '5c:0f:fb:3d:5a:b2'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST (ALBRECHT MISS CATHERINE)',
                                  'serial': '263222',
                                  'serial_number': '263222',
                                  'system_mac_address': None},
                '264414': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '54 Mill St. W. Milverton (Wesbell '
                                          'Technologies)',
                                  'serial': '264414',
                                  'serial_number': '264414',
                                  'system_mac_address': None},
                '26AC6F': {       'dev_macs': [       '5c:0f:fb:3d:5b:1c',
                                                      '5c:0f:fb:3d:7d:82',
                                                      '00:03:e6:dd:31:50',
                                                      '5c:0f:fb:3d:5b:0c',
                                                      '5c:0f:fb:3d:80:1c',
                                                      '5c:0f:fb:17:75:1a'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST',
                                  'serial': '26AC6F',
                                  'serial_number': '26AC6F',
                                  'system_mac_address': None},
                '26CC54': {       'dev_macs': [       'cc:be:59:2d:a5:26',
                                                      '34:60:f9:0a:f2:d1',
                                                      '5c:0f:fb:2a:0e:10'],
                                  'ip_address': None,
                                  'name': '214 Winstanley St. Monkton (8 Plex Apt)',
                                  'serial': '26CC54',
                                  'serial_number': '26CC54',
                                  'system_mac_address': None},
                '26D966': {       'dev_macs': [       '00:03:e6:a5:70:b2',
                                                      '5c:0f:fb:3d:58:de',
                                                      '5c:0f:fb:43:97:58',
                                                      '5c:0f:fb:3d:7d:9a',
                                                      '00:03:e6:a5:6f:ae'],
                                  'ip_address': None,
                                  'name': '50 William St. Milverton (Data Rm 2 '
                                          'Knollcrest)',
                                  'serial': '26D966',
                                  'serial_number': '26D966',
                                  'system_mac_address': None},
                '26D992': {       'dev_macs': [       'b4:b0:24:8b:b9:11',
                                                      '88:d7:f6:b8:30:24',
                                                      '14:21:03:92:f6:1b'],
                                  'ip_address': None,
                                  'name': '1 LLANA HEIGHTS DR (KNIPE JACQUELYN)',
                                  'serial': '26D992',
                                  'serial_number': '26D992',
                                  'system_mac_address': None},
                '28F653': {       'dev_macs': ['40:5d:82:e1:9b:c9'],
                                  'ip_address': None,
                                  'name': '27 Mill St. W. Milverton (Mary Francis '
                                          'Lambert-Hill)',
                                  'serial': '28F653',
                                  'serial_number': '28F653',
                                  'system_mac_address': None},
                '28F664': {       'dev_macs': ['50:d4:f7:80:62:e9'],
                                  'ip_address': None,
                                  'name': '8 Edward St. Milverton (Katrina MacPherson)',
                                  'serial': '28F664',
                                  'serial_number': '28F664',
                                  'system_mac_address': None},
                '296FB8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '2 Whaley Ave. Milverton (Dave Paradis '
                                          '10401473)',
                                  'serial': '296FB8',
                                  'serial_number': '296FB8',
                                  'system_mac_address': None},
                '296FC0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 Whaley Ave. Milverton',
                                  'serial': '296FC0',
                                  'serial_number': '296FC0',
                                  'system_mac_address': None},
                '296FC1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '13 Fulton St. Milverton (Richard Roulston '
                                          '10401073)',
                                  'serial': '296FC1',
                                  'serial_number': '296FC1',
                                  'system_mac_address': None},
                '296FC4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '23 Mill St. W. Milverton (James Wagler '
                                          '10401047)',
                                  'serial': '296FC4',
                                  'serial_number': '296FC4',
                                  'system_mac_address': None},
                '296FC5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '20 Fulton St. Milverton (Jeff Leblanc)',
                                  'serial': '296FC5',
                                  'serial_number': '296FC5',
                                  'system_mac_address': None},
                '296FC7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '18 TEMPERANCE ST (DALE BRIAN)',
                                  'serial': '296FC7',
                                  'serial_number': '296FC7',
                                  'system_mac_address': None},
                '296FC9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '17 Whaley Ave. Milverton (Darrel Roth)',
                                  'serial': '296FC9',
                                  'serial_number': '296FC9',
                                  'system_mac_address': None},
                '296FEB': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 EDWARD ST (WATSON BRENNAN)',
                                  'serial': '296FEB',
                                  'serial_number': '296FEB',
                                  'system_mac_address': None},
                '2981E1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '57 MAIN ST N (DZODAN DEJAN)',
                                  'serial': '2981E1',
                                  'serial_number': '2981E1',
                                  'system_mac_address': None},
                '2981E4': {       'dev_macs': [       '00:03:e6:a2:fd:72',
                                                      '00:03:e6:db:23:56'],
                                  'ip_address': None,
                                  'name': '19 West St. Milverton (Brennan Erb)',
                                  'serial': '2981E4',
                                  'serial_number': '2981E4',
                                  'system_mac_address': None},
                '2981E5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '29 West St. Milverton (Tim Ellis 10401774)',
                                  'serial': '2981E5',
                                  'serial_number': '2981E5',
                                  'system_mac_address': None},
                '2981E6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '3 PACIFIC AV',
                                  'serial': '2981E6',
                                  'serial_number': '2981E6',
                                  'system_mac_address': None},
                '2981FA': {       'dev_macs': ['04:bc:9f:46:8e:a7'],
                                  'ip_address': None,
                                  'name': '11 JOHN ST (HURAS JENNIFER)',
                                  'serial': '2981FA',
                                  'serial_number': '2981FA',
                                  'system_mac_address': None},
                '2981FF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '18 West St. Milverton (Brian & Amy Martin '
                                          '10434104)',
                                  'serial': '2981FF',
                                  'serial_number': '2981FF',
                                  'system_mac_address': None},
                '298200': {       'dev_macs': ['b8:94:70:f3:e1:33'],
                                  'ip_address': None,
                                  'name': '18 Pacific Ave. Milverton (Allissa '
                                          'Wettlaufer)',
                                  'serial': '298200',
                                  'serial_number': '298200',
                                  'system_mac_address': None},
                '298212': {       'dev_macs': ['d0:76:8f:1f:58:9b'],
                                  'ip_address': None,
                                  'name': '6 PACIFIC AV (ROULSTON WAYNE)',
                                  'serial': '298212',
                                  'serial_number': '298212',
                                  'system_mac_address': None},
                '29821E': {       'dev_macs': ['14:21:03:92:f2:c1'],
                                  'ip_address': None,
                                  'name': '24 PACIFIC AV (SMITH CHRISTINA)',
                                  'serial': '29821E',
                                  'serial_number': '29821E',
                                  'system_mac_address': None},
                '29821F': {       'dev_macs': ['d0:76:8f:02:f1:ee'],
                                  'ip_address': None,
                                  'name': '10 WEST ST (MURRAY JENNIFER)',
                                  'serial': '29821F',
                                  'serial_number': '29821F',
                                  'system_mac_address': None},
                '298225': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 Fulton St. Milverton (Jill Ross '
                                          '10430010)',
                                  'serial': '298225',
                                  'serial_number': '298225',
                                  'system_mac_address': None},
                '298232': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '29 MAIN ST S (SCHWARTZENBERGER ALLAN '
                                          '10438370)',
                                  'serial': '298232',
                                  'serial_number': '298232',
                                  'system_mac_address': None},
                '298248': {       'dev_macs': ['d0:76:8f:1f:5b:5b'],
                                  'ip_address': None,
                                  'name': '3 Whaley Av. Milverton (MATHESON TYLER)',
                                  'serial': '298248',
                                  'serial_number': '298248',
                                  'system_mac_address': None},
                '29824D': {       'dev_macs': ['60:db:98:09:37:a5'],
                                  'ip_address': None,
                                  'name': '162 FISHLEIGH ST (MATTHEWS GRANT)',
                                  'serial': '29824D',
                                  'serial_number': '29824D',
                                  'system_mac_address': None},
                '29824E': {       'dev_macs': ['84:d3:43:6d:e7:5f'],
                                  'ip_address': None,
                                  'name': '13 West St. Milverton (Richard Harrow)',
                                  'serial': '29824E',
                                  'serial_number': '29824E',
                                  'system_mac_address': None},
                '29824F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6467 Perth Rd. 131 (Kate Arand)',
                                  'serial': '29824F',
                                  'serial_number': '29824F',
                                  'system_mac_address': None},
                '298253': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '20 Whaley Ave (FINNIE EVELYN)',
                                  'serial': '298253',
                                  'serial_number': '298253',
                                  'system_mac_address': None},
                '298254': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '38 William St. Milverton (James Brubacher)',
                                  'serial': '298254',
                                  'serial_number': '298254',
                                  'system_mac_address': None},
                '29825F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '9 MacBeth St. Milverton (Don Gascho)',
                                  'serial': '29825F',
                                  'serial_number': '29825F',
                                  'system_mac_address': None},
                '298262': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 WEST ST (DAGG DEVIN)',
                                  'serial': '298262',
                                  'serial_number': '298262',
                                  'system_mac_address': None},
                '298268': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '19 William St. Milverton (ROPP ADAM)',
                                  'serial': '298268',
                                  'serial_number': '298268',
                                  'system_mac_address': None},
                '29827C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '15 Pierson St. Milverton (Mryna Schmidt '
                                          '10400672)',
                                  'serial': '29827C',
                                  'serial_number': '29827C',
                                  'system_mac_address': None},
                '298289': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '14 William St. Milverton (Bill Moore '
                                          '10404221)',
                                  'serial': '298289',
                                  'serial_number': '298289',
                                  'system_mac_address': None},
                '29828D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '15 Whaley Ave. Milverton (Ron Kipfer)',
                                  'serial': '29828D',
                                  'serial_number': '29828D',
                                  'system_mac_address': None},
                '2982B7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '12 Whaley Ave. Milverton (Adam Sage)',
                                  'serial': '2982B7',
                                  'serial_number': '2982B7',
                                  'system_mac_address': None},
                '2982C4': {       'dev_macs': ['d0:76:8f:02:f3:01'],
                                  'ip_address': None,
                                  'name': '59 MAIN ST N (HOLMES  BRANDON)',
                                  'serial': '2982C4',
                                  'serial_number': '2982C4',
                                  'system_mac_address': None},
                '2982C8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 William St. Apt#102 Milverton',
                                  'serial': '2982C8',
                                  'serial_number': '2982C8',
                                  'system_mac_address': None},
                '298301': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '26 Main St. N. Milverton (Innovapost Inc '
                                          '10431404)',
                                  'serial': '298301',
                                  'serial_number': '298301',
                                  'system_mac_address': None},
                '29834C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '33 Pacific Ave. Milverton (Jeffrey Plager '
                                          '10434465)',
                                  'serial': '29834C',
                                  'serial_number': '29834C',
                                  'system_mac_address': None},
                '298385': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6465 Perth Rd. 131 (Peggy Birmingham)',
                                  'serial': '298385',
                                  'serial_number': '298385',
                                  'system_mac_address': None},
                '2983ED': {       'dev_macs': ['d0:76:8f:02:f3:d2'],
                                  'ip_address': None,
                                  'name': '6469 Perth Rd. 131 (Brian Nafziger '
                                          '10427589)',
                                  'serial': '2983ED',
                                  'serial_number': '2983ED',
                                  'system_mac_address': None},
                '2983F8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6 MILL ST E APT 2 (STREAMLINE ENGINEERING '
                                          'INC)',
                                  'serial': '2983F8',
                                  'serial_number': '2983F8',
                                  'system_mac_address': None},
                '298439': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6 ADELAIDE ST (SANGUYO DAVID)',
                                  'serial': '298439',
                                  'serial_number': '298439',
                                  'system_mac_address': None},
                '298442': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '166 Winstanley St. Monkton (Danielle '
                                          'MacDonald)',
                                  'serial': '298442',
                                  'serial_number': '298442',
                                  'system_mac_address': None},
                '29845A': {       'dev_macs': [       '44:65:7f:89:0b:ac',
                                                      '5c:0f:fb:2a:06:98',
                                                      '5c:0f:fb:2a:05:66'],
                                  'ip_address': None,
                                  'name': '6464 Perth Rd. 131 (Ross Gerber)',
                                  'serial': '29845A',
                                  'serial_number': '29845A',
                                  'system_mac_address': None},
                '298576': {       'dev_macs': ['84:d3:43:6d:f0:4f'],
                                  'ip_address': None,
                                  'name': '21 PACIFIC AV (EISENHOFER MARJ)',
                                  'serial': '298576',
                                  'serial_number': '298576',
                                  'system_mac_address': None},
                '2A1AD3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '28 Main St. N. Milverton (Milverton Family '
                                          'Dental)',
                                  'serial': '2A1AD3',
                                  'serial_number': '2A1AD3',
                                  'system_mac_address': None},
                '2AC4B6': {       'dev_macs': ['cc:be:59:7e:ba:5e'],
                                  'ip_address': None,
                                  'name': '8 Temperance St. Milverton (Mike Crapper '
                                          '10425901)',
                                  'serial': '2AC4B6',
                                  'serial_number': '2AC4B6',
                                  'system_mac_address': None},
                '2AD0BC': {       'dev_macs': ['ec:4f:82:ac:e3:19'],
                                  'ip_address': None,
                                  'name': '49 Maddison St W (A & A McKenzie)',
                                  'serial': '2AD0BC',
                                  'serial_number': '2AD0BC',
                                  'system_mac_address': None},
                '2AD0F1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '47 MILL ST (COXON ROBERT)',
                                  'serial': '2AD0F1',
                                  'serial_number': '2AD0F1',
                                  'system_mac_address': None},
                '2AFB74': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '3 Spencer St. Milverton (Jason Yantzi)',
                                  'serial': '2AFB74',
                                  'serial_number': '2AFB74',
                                  'system_mac_address': None},
                '2B0150': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5734 LINE 57 (VAN NULAND MARIA (MELANIE))',
                                  'serial': '2B0150',
                                  'serial_number': '2B0150',
                                  'system_mac_address': None},
                '2B0176': {       'dev_macs': ['5c:0f:fb:2a:0a:d2'],
                                  'ip_address': None,
                                  'name': '6 Grosch St. Milverton (Jim Partridge)',
                                  'serial': '2B0176',
                                  'serial_number': '2B0176',
                                  'system_mac_address': None},
                '2B0181': {       'dev_macs': ['60:db:98:09:38:8c'],
                                  'ip_address': None,
                                  'name': '6472 Perth Road 131 (Brett Kipfer)',
                                  'serial': '2B0181',
                                  'serial_number': '2B0181',
                                  'system_mac_address': None},
                '2B0182': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '78 Mill St. E. Milverton (MARTIN CHADRICK)',
                                  'serial': '2B0182',
                                  'serial_number': '2B0182',
                                  'system_mac_address': None},
                '2B0186': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '32 Mill St. E. Milverton (Brendan Barrett)',
                                  'serial': '2B0186',
                                  'serial_number': '2B0186',
                                  'system_mac_address': None},
                '2B018A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '28 Pacific Ave. Milverton (Jacob Harman)',
                                  'serial': '2B018A',
                                  'serial_number': '2B018A',
                                  'system_mac_address': None},
                '2B018D': {       'dev_macs': ['b8:94:70:9d:d9:8a'],
                                  'ip_address': None,
                                  'name': '74 MAIN ST N (MCCORMICK PHIL & CAROL)',
                                  'serial': '2B018D',
                                  'serial_number': '2B018D',
                                  'system_mac_address': None},
                '2B018F': {       'dev_macs': ['d0:76:8f:91:d4:9b'],
                                  'ip_address': None,
                                  'name': '55 MAIN ST N (MACDONALD ALANNA)',
                                  'serial': '2B018F',
                                  'serial_number': '2B018F',
                                  'system_mac_address': None},
                '2B0194': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '68 MAIN ST N (YOST RACHEL)',
                                  'serial': '2B0194',
                                  'serial_number': '2B0194',
                                  'system_mac_address': None},
                '2B0195': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '67 Main St. N. (Michele Duskocy 10400255)',
                                  'serial': '2B0195',
                                  'serial_number': '2B0195',
                                  'system_mac_address': None},
                '2B019B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': "66 Main St. N. Milverton (Burn's "
                                          'Presbyterian Church)',
                                  'serial': '2B019B',
                                  'serial_number': '2B019B',
                                  'system_mac_address': None},
                '2B019C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '81 Main St. N. Milverton (Carrie Wagler)',
                                  'serial': '2B019C',
                                  'serial_number': '2B019C',
                                  'system_mac_address': None},
                '2B019E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '146 Brock St Monkton (Kelly '
                                          'Battams-Trussler 10430830)',
                                  'serial': '2B019E',
                                  'serial_number': '2B019E',
                                  'system_mac_address': None},
                '2B01B6': {       'dev_macs': ['5c:0f:fb:2a:0b:3a'],
                                  'ip_address': None,
                                  'name': '4 Spencer St. Milverton (Neil Schlueter)',
                                  'serial': '2B01B6',
                                  'serial_number': '2B01B6',
                                  'system_mac_address': None},
                '2B01B7': {       'dev_macs': ['84:d3:43:6d:e5:f4'],
                                  'ip_address': None,
                                  'name': '37 Temperance St. Milverton (Chuck Meyers '
                                          '10401071)',
                                  'serial': '2B01B7',
                                  'serial_number': '2B01B7',
                                  'system_mac_address': None},
                '2B01C6': {       'dev_macs': ['cc:be:59:7e:b7:df'],
                                  'ip_address': None,
                                  'name': '55 William St.Apt # 101 (Catharine Birch)',
                                  'serial': '2B01C6',
                                  'serial_number': '2B01C6',
                                  'system_mac_address': None},
                '2B01C8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '53 MAIN ST N (HESSON CHRISTIAN FELLOWSHIP)',
                                  'serial': '2B01C8',
                                  'serial_number': '2B01C8',
                                  'system_mac_address': None},
                '2B01DE': {       'dev_macs': ['ec:4f:82:a5:04:42'],
                                  'ip_address': None,
                                  'name': '46 Main St. N. Milverton (John Kilmer)',
                                  'serial': '2B01DE',
                                  'serial_number': '2B01DE',
                                  'system_mac_address': None},
                '2B0206': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6432 Perth Rd 131 (Morris Gerber Excavating '
                                          '10401316)',
                                  'serial': '2B0206',
                                  'serial_number': '2B0206',
                                  'system_mac_address': None},
                '2B025D': {       'dev_macs': ['60:db:98:09:1c:3b'],
                                  'ip_address': None,
                                  'name': '6458 PERTH ROAD 131 (WAGLER  EZRA)',
                                  'serial': '2B025D',
                                  'serial_number': '2B025D',
                                  'system_mac_address': None},
                '2B0264': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 William St. Ste. 105 (Joyce Westman '
                                          '10433383)',
                                  'serial': '2B0264',
                                  'serial_number': '2B0264',
                                  'system_mac_address': None},
                '2B0266': {       'dev_macs': ['ec:4f:82:2c:6b:78'],
                                  'ip_address': None,
                                  'name': '19 Maddison St. W. Monkton (Rachel '
                                          'Chiarello)',
                                  'serial': '2B0266',
                                  'serial_number': '2B0266',
                                  'system_mac_address': None},
                '2B026A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 George St. Milverton (Stanley Kuepfer '
                                          '10400973)',
                                  'serial': '2B026A',
                                  'serial_number': '2B026A',
                                  'system_mac_address': None},
                '2B0272': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6468 Perth Rd. 131 (Jason Lantz 10430328)',
                                  'serial': '2B0272',
                                  'serial_number': '2B0272',
                                  'system_mac_address': None},
                '2B0294': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '9 Edward St. Milverton (Ezra Streicher)',
                                  'serial': '2B0294',
                                  'serial_number': '2B0294',
                                  'system_mac_address': None},
                '2B0298': {       'dev_macs': ['d0:76:8f:1f:51:dc'],
                                  'ip_address': None,
                                  'name': '169 SMITH ST (DOLSON ANDREA)',
                                  'serial': '2B0298',
                                  'serial_number': '2B0298',
                                  'system_mac_address': None},
                '2B0368': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '33 Temperance St. Milverton (Sarah&Mary '
                                          'Martin 10402460)',
                                  'serial': '2B0368',
                                  'serial_number': '2B0368',
                                  'system_mac_address': None},
                '2B03FA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '39 Main St. S. Milverton (United Church '
                                          '10401217)',
                                  'serial': '2B03FA',
                                  'serial_number': '2B03FA',
                                  'system_mac_address': None},
                '2B0405': {       'dev_macs': ['ec:4f:82:a5:06:0d'],
                                  'ip_address': None,
                                  'name': '10 Llana Heights Milverton (Dirk Willemsen '
                                          '10401603)',
                                  'serial': '2B0405',
                                  'serial_number': '2B0405',
                                  'system_mac_address': None},
                '2B0421': {       'dev_macs': ['cc:be:59:7e:b7:fa'],
                                  'ip_address': None,
                                  'name': '19 Mill St. W. Milverton (Llori Nicholls)',
                                  'serial': '2B0421',
                                  'serial_number': '2B0421',
                                  'system_mac_address': None},
                '2B043C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '38 Main St. N. Milverton (Chris Donnelly)',
                                  'serial': '2B043C',
                                  'serial_number': '2B043C',
                                  'system_mac_address': None},
                '2B043F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 Grosch St. Milverton (Angela Schneider)',
                                  'serial': '2B043F',
                                  'serial_number': '2B043F',
                                  'system_mac_address': None},
                '2B0444': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '42 MAIN ST S (2495008 ONTARIO LTD.)',
                                  'serial': '2B0444',
                                  'serial_number': '2B0444',
                                  'system_mac_address': None},
                '2B044C': {       'dev_macs': ['ec:4f:82:a9:a6:41'],
                                  'ip_address': None,
                                  'name': '31 Temperance St. Milverton (Peter Ryter '
                                          '10425942)',
                                  'serial': '2B044C',
                                  'serial_number': '2B044C',
                                  'system_mac_address': None},
                '2B0450': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 EDWARD ST (HERD CATHY)',
                                  'serial': '2B0450',
                                  'serial_number': '2B0450',
                                  'system_mac_address': None},
                '2B0453': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '40 Main St. N. Milverton (Eric Sider)',
                                  'serial': '2B0453',
                                  'serial_number': '2B0453',
                                  'system_mac_address': None},
                '2B0454': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 Pugh St. W. Milverton (Don Hollinger)',
                                  'serial': '2B0454',
                                  'serial_number': '2B0454',
                                  'system_mac_address': None},
                '2B0462': {       'dev_macs': ['ec:4f:82:a9:a0:e9'],
                                  'ip_address': None,
                                  'name': '14 Pugh St. W. Milverton (Ken Zehr '
                                          '10429938)',
                                  'serial': '2B0462',
                                  'serial_number': '2B0462',
                                  'system_mac_address': None},
                '2B0471': {       'dev_macs': ['14:21:03:92:f0:2d'],
                                  'ip_address': None,
                                  'name': '31 PACIFIC AV (MCKAY ROBERT)',
                                  'serial': '2B0471',
                                  'serial_number': '2B0471',
                                  'system_mac_address': None},
                '2B05BE': {       'dev_macs': ['ec:4f:82:ac:e3:85'],
                                  'ip_address': None,
                                  'name': '68 GERBER DR',
                                  'serial': '2B05BE',
                                  'serial_number': '2B05BE',
                                  'system_mac_address': None},
                '2B05D7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '29 PACIFIC AV (JUGDAVE VISHAAL)',
                                  'serial': '2B05D7',
                                  'serial_number': '2B05D7',
                                  'system_mac_address': None},
                '2BD241': {       'dev_macs': ['34:0a:33:81:59:d5'],
                                  'ip_address': None,
                                  'name': '9 Pugh St. W. Milverton (Ashley Mayberry)',
                                  'serial': '2BD241',
                                  'serial_number': '2BD241',
                                  'system_mac_address': None},
                '2BD264': {       'dev_macs': ['e4:6f:13:40:a4:f3'],
                                  'ip_address': None,
                                  'name': '82 Main St N (William & Krista Kuepfer)',
                                  'serial': '2BD264',
                                  'serial_number': '2BD264',
                                  'system_mac_address': None},
                '2BD26F': {       'dev_macs': ['ec:4f:82:2a:b0:41'],
                                  'ip_address': None,
                                  'name': '49 Temperance St Milverton (D&J Luyten '
                                          '10433326)',
                                  'serial': '2BD26F',
                                  'serial_number': '2BD26F',
                                  'system_mac_address': None},
                '2BD360': {       'dev_macs': ['84:d8:1b:e1:8f:89'],
                                  'ip_address': None,
                                  'name': '23 West St. Milverton (Tyson Londry '
                                          '10431157)',
                                  'serial': '2BD360',
                                  'serial_number': '2BD360',
                                  'system_mac_address': None},
                '2BD474': {       'dev_macs': ['78:32:1b:49:bf:78'],
                                  'ip_address': None,
                                  'name': '29 MADDISON ST W (DOCKING MELANIE)',
                                  'serial': '2BD474',
                                  'serial_number': '2BD474',
                                  'system_mac_address': None},
                '2BD478': {       'dev_macs': ['44:65:7f:89:0f:ed'],
                                  'ip_address': None,
                                  'name': '80 SCHADE ST (GOODMAN-SMITH ERIN)',
                                  'serial': '2BD478',
                                  'serial_number': '2BD478',
                                  'system_mac_address': None},
                '2BD5CA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '171 Winstanley St. Monkton',
                                  'serial': '2BD5CA',
                                  'serial_number': '2BD5CA',
                                  'system_mac_address': None},
                '2BD66F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '12 Main St. S. Milverton (Execulink Credit '
                                          'union)',
                                  'serial': '2BD66F',
                                  'serial_number': '2BD66F',
                                  'system_mac_address': None},
                '2BD68D': {       'dev_macs': ['68:4a:76:d7:a4:20'],
                                  'ip_address': None,
                                  'name': '32 Pacific Ave. Milverton (Rachel Brennan)',
                                  'serial': '2BD68D',
                                  'serial_number': '2BD68D',
                                  'system_mac_address': None},
                '2BD693': {       'dev_macs': ['cc:be:59:2d:a5:da'],
                                  'ip_address': None,
                                  'name': '19 Pacific Av. Milverton (Wade Goodwin '
                                          '10426524)',
                                  'serial': '2BD693',
                                  'serial_number': '2BD693',
                                  'system_mac_address': None},
                '2BD696': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 Pacific Av. Milverton',
                                  'serial': '2BD696',
                                  'serial_number': '2BD696',
                                  'system_mac_address': None},
                '2BD69F': {       'dev_macs': ['cc:be:59:7e:b6:f5'],
                                  'ip_address': None,
                                  'name': '7 West St. Milverton (Ian Clark 10434646)',
                                  'serial': '2BD69F',
                                  'serial_number': '2BD69F',
                                  'system_mac_address': None},
                '2C9656': {       'dev_macs': ['ec:4f:82:2d:7a:35'],
                                  'ip_address': None,
                                  'name': '6611 Perth Rd. 131 Milverton (The '
                                          'Garage&R&J Wagler)',
                                  'serial': '2C9656',
                                  'serial_number': '2C9656',
                                  'system_mac_address': None},
                '2DCD18': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (FELIX JAMES)',
                                  'serial': '2DCD18',
                                  'serial_number': '2DCD18',
                                  'system_mac_address': None},
                '2E91A8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '187 Erskine St. Monkton (Robert Doell '
                                          '10432296)',
                                  'serial': '2E91A8',
                                  'serial_number': '2E91A8',
                                  'system_mac_address': None},
                '2ED67B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6374 ROAD 154 (VAN NULAND MARIA)',
                                  'serial': '2ED67B',
                                  'serial_number': '2ED67B',
                                  'system_mac_address': None},
                '2ED92E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '47 Maddison St. E. Monkton (Mane Attraction '
                                          '10431045)',
                                  'serial': '2ED92E',
                                  'serial_number': '2ED92E',
                                  'system_mac_address': None},
                '2ED9C4': {       'dev_macs': ['b8:94:70:f3:e6:66'],
                                  'ip_address': None,
                                  'name': '32 WILLIAM ST (LEHMANN DON & SHARON)',
                                  'serial': '2ED9C4',
                                  'serial_number': '2ED9C4',
                                  'system_mac_address': None},
                '2ED9CD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '147 Brock St. Monkton (Kevin Coulson)',
                                  'serial': '2ED9CD',
                                  'serial_number': '2ED9CD',
                                  'system_mac_address': None},
                '2F3B17': {       'dev_macs': ['c8:4f:86:fc:00:02'],
                                  'ip_address': None,
                                  'name': '7003 Raymond St (MIllbank Hardware)',
                                  'serial': '2F3B17',
                                  'serial_number': '2F3B17',
                                  'system_mac_address': None},
                '2FE9B2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '149 BROCK STREET MONKTON (BROUGHTON JOANNE)',
                                  'serial': '2FE9B2',
                                  'serial_number': '2FE9B2',
                                  'system_mac_address': None},
                '2FE9C9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '37 Maddison St. W. Monkton (Amanda Pfeifer)',
                                  'serial': '2FE9C9',
                                  'serial_number': '2FE9C9',
                                  'system_mac_address': None},
                '2FE9E5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '175 Winstanley St. Monkton (Vince '
                                          'Priestrap)',
                                  'serial': '2FE9E5',
                                  'serial_number': '2FE9E5',
                                  'system_mac_address': None},
                '2FEA05': {       'dev_macs': ['c0:06:c3:1c:9b:46'],
                                  'ip_address': None,
                                  'name': '16 SCHADE ST (MAYER MICHAEL)',
                                  'serial': '2FEA05',
                                  'serial_number': '2FEA05',
                                  'system_mac_address': None},
                '2FEA0D': {       'dev_macs': ['b8:94:70:bc:b4:4f'],
                                  'ip_address': None,
                                  'name': '163 BROCK STREET MONKTON (GREGORY DAWN)',
                                  'serial': '2FEA0D',
                                  'serial_number': '2FEA0D',
                                  'system_mac_address': None},
                '2FEA1D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5020 LINE 49 (ROCDALE FARMS)',
                                  'serial': '2FEA1D',
                                  'serial_number': '2FEA1D',
                                  'system_mac_address': None},
                '2FEAF3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6879A ROAD 140 (KUEPFER MARVIN & SARAH)',
                                  'serial': '2FEAF3',
                                  'serial_number': '2FEAF3',
                                  'system_mac_address': None},
                '2FEAFA': {       'dev_macs': ['14:21:03:92:e9:a5'],
                                  'ip_address': None,
                                  'name': '3 GROSCH ST MILVERTON (NAHRGANG GEORGE)',
                                  'serial': '2FEAFA',
                                  'serial_number': '2FEAFA',
                                  'system_mac_address': None},
                '2FEB5D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '25 Maddison St. W. Monkton (Dean Smith)',
                                  'serial': '2FEB5D',
                                  'serial_number': '2FEB5D',
                                  'system_mac_address': None},
                '2FEB62': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '138 Maddison St. E. Monkton',
                                  'serial': '2FEB62',
                                  'serial_number': '2FEB62',
                                  'system_mac_address': None},
                '2FEC08': {       'dev_macs': ['14:21:03:92:f6:31'],
                                  'ip_address': None,
                                  'name': '103 Maddison St. W. Monkton (Charlene '
                                          'Hamilton 10430268)',
                                  'serial': '2FEC08',
                                  'serial_number': '2FEC08',
                                  'system_mac_address': None},
                '30039F': {       'dev_macs': ['14:21:03:a1:e6:7e'],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (BRUCK SAMANTHA)',
                                  'serial': '30039F',
                                  'serial_number': '30039F',
                                  'system_mac_address': None},
                '30060A': {       'dev_macs': ['b4:75:0e:54:f6:19'],
                                  'ip_address': None,
                                  'name': '17 William St. Milverton (Barry Brick)',
                                  'serial': '30060A',
                                  'serial_number': '30060A',
                                  'system_mac_address': None},
                '300623': {       'dev_macs': ['84:d3:43:0f:bb:12'],
                                  'ip_address': None,
                                  'name': '36 PACIFIC AV (HORTON CHAD)',
                                  'serial': '300623',
                                  'serial_number': '300623',
                                  'system_mac_address': None},
                '30064C': {       'dev_macs': ['b8:94:70:f3:e6:24'],
                                  'ip_address': None,
                                  'name': '157 Smith St. Monkton (Judith Burwell)',
                                  'serial': '30064C',
                                  'serial_number': '30064C',
                                  'system_mac_address': None},
                '30079D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 William St apt 207 Milverton',
                                  'serial': '30079D',
                                  'serial_number': '30079D',
                                  'system_mac_address': None},
                '3007AE': {       'dev_macs': ['cc:32:e5:11:31:61'],
                                  'ip_address': None,
                                  'name': '38 Maddison St. E. Monkton (Bjorn Makinen)',
                                  'serial': '3007AE',
                                  'serial_number': '3007AE',
                                  'system_mac_address': None},
                '3007AF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5907 Perth Road 119 Brunner',
                                  'serial': '3007AF',
                                  'serial_number': '3007AF',
                                  'system_mac_address': None},
                '3007C5': {       'dev_macs': ['c4:e9:0a:56:56:57'],
                                  'ip_address': None,
                                  'name': '26 Mill St. E. Milverton (Tony Douglas)',
                                  'serial': '3007C5',
                                  'serial_number': '3007C5',
                                  'system_mac_address': None},
                '300D95': {       'dev_macs': ['60:63:4c:6d:fe:81'],
                                  'ip_address': None,
                                  'name': '5792 LINE 57 (VAN NULAND MARIA (MELANIE))',
                                  'serial': '300D95',
                                  'serial_number': '300D95',
                                  'system_mac_address': None},
                '300D9C': {       'dev_macs': ['44:65:7f:41:c5:2f'],
                                  'ip_address': None,
                                  'name': '87 Gerber Dr (CUNHA CARLOS)',
                                  'serial': '300D9C',
                                  'serial_number': '300D9C',
                                  'system_mac_address': None},
                '300DCB': {       'dev_macs': ['30:68:93:36:15:5d'],
                                  'ip_address': None,
                                  'name': '76 MAIN ST N (NAFZIGER COLE)',
                                  'serial': '300DCB',
                                  'serial_number': '300DCB',
                                  'system_mac_address': None},
                '300DDF': {       'dev_macs': ['60:db:98:08:d2:ed'],
                                  'ip_address': None,
                                  'name': '5 Whaley Ave. Milverton (DENOMME ZACH '
                                          '10438436)',
                                  'serial': '300DDF',
                                  'serial_number': '300DDF',
                                  'system_mac_address': None},
                '300DF1': {       'dev_macs': ['24:2f:d0:36:09:f1'],
                                  'ip_address': None,
                                  'name': '1 Whaley Ave. Milverton (Nick Beggs)',
                                  'serial': '300DF1',
                                  'serial_number': '300DF1',
                                  'system_mac_address': None},
                '300DFD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '9 St. Marys St. N. Milverton (Alex Campbell '
                                          '10435536)',
                                  'serial': '300DFD',
                                  'serial_number': '300DFD',
                                  'system_mac_address': None},
                '300E72': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'test',
                                  'serial': '300E72',
                                  'serial_number': '300E72',
                                  'system_mac_address': None},
                '300E75': {       'dev_macs': ['24:f5:a2:09:16:62'],
                                  'ip_address': None,
                                  'name': '25 MAIN ST S (HALL MICHAEL)',
                                  'serial': '300E75',
                                  'serial_number': '300E75',
                                  'system_mac_address': None},
                '300E88': {       'dev_macs': ['1c:7e:e5:37:f4:9d'],
                                  'ip_address': None,
                                  'name': '7 Llana Heights Main Fl (Robyn Yost)',
                                  'serial': '300E88',
                                  'serial_number': '300E88',
                                  'system_mac_address': None},
                '300EF8': {       'dev_macs': ['ec:4f:82:ac:d0:53'],
                                  'ip_address': None,
                                  'name': '44 Main St S Milverton (JL Hallahan Ltd '
                                          '10435438)',
                                  'serial': '300EF8',
                                  'serial_number': '300EF8',
                                  'system_mac_address': None},
                '300F00': {       'dev_macs': ['d0:76:8f:02:ed:97'],
                                  'ip_address': None,
                                  'name': '60 MAIN ST N (NEWMAN ASHLYNNE)',
                                  'serial': '300F00',
                                  'serial_number': '300F00',
                                  'system_mac_address': None},
                '3210DE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '76 GERBER DR',
                                  'serial': '3210DE',
                                  'serial_number': '3210DE',
                                  'system_mac_address': None},
                '3210E4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '21 Arena Dr. Milverton (Denise Smith)',
                                  'serial': '3210E4',
                                  'serial_number': '3210E4',
                                  'system_mac_address': None},
                '3212EF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '80 Mill St. E. Milverton (Lynda Kuepfer)',
                                  'serial': '3212EF',
                                  'serial_number': '3212EF',
                                  'system_mac_address': None},
                '32143A': {       'dev_macs': ['84:d3:43:6d:e6:fc'],
                                  'ip_address': None,
                                  'name': '80 Gerber Drive Milverton (Randal Leis '
                                          '10434200)',
                                  'serial': '32143A',
                                  'serial_number': '32143A',
                                  'system_mac_address': None},
                '32171B': {       'dev_macs': ['60:db:98:09:1d:e8'],
                                  'ip_address': None,
                                  'name': '5 PACIFIC AV (MILOVANCEV CHLOE AND NATHAN)',
                                  'serial': '32171B',
                                  'serial_number': '32171B',
                                  'system_mac_address': None},
                '321720': {       'dev_macs': ['b8:94:70:bc:b8:4e'],
                                  'ip_address': None,
                                  'name': '5842 Perth Road 119 Brunner (Don & Eva '
                                          'Anderson 10430386)',
                                  'serial': '321720',
                                  'serial_number': '321720',
                                  'system_mac_address': None},
                '321779': {       'dev_macs': ['cc:be:59:2f:2f:eb'],
                                  'ip_address': None,
                                  'name': '160 Smith St Monkton (Carl Steicher '
                                          '10434033)',
                                  'serial': '321779',
                                  'serial_number': '321779',
                                  'system_mac_address': None},
                '32177F': {       'dev_macs': ['14:21:03:92:ec:d3'],
                                  'ip_address': None,
                                  'name': '21 MILL ST W (SHANNON KEELIE)',
                                  'serial': '32177F',
                                  'serial_number': '32177F',
                                  'system_mac_address': None},
                '32178E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5752 Perth Rd. 119 (Micheal Weber 10400697)',
                                  'serial': '32178E',
                                  'serial_number': '32178E',
                                  'system_mac_address': None},
                '3217CC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '68 Schade St Monkton (Ross Rennick '
                                          '10434031)',
                                  'serial': '3217CC',
                                  'serial_number': '3217CC',
                                  'system_mac_address': None},
                '3217E6': {       'dev_macs': ['cc:be:59:2d:a8:e9'],
                                  'ip_address': None,
                                  'name': '4345 LINE 61 (ROPP DAVID & BONNIE)',
                                  'serial': '3217E6',
                                  'serial_number': '3217E6',
                                  'system_mac_address': None},
                '321808': {       'dev_macs': ['14:21:03:8b:7a:a8'],
                                  'ip_address': None,
                                  'name': '24 Arena Dr Milverton (Shirley Weber '
                                          '10401054)',
                                  'serial': '321808',
                                  'serial_number': '321808',
                                  'system_mac_address': None},
                '321825': {       'dev_macs': [       '00:03:e6:a5:70:de',
                                                      '00:03:e6:a5:73:90'],
                                  'ip_address': None,
                                  'name': '93 Schade St Monkton (Steve Smith 10434116)',
                                  'serial': '321825',
                                  'serial_number': '321825',
                                  'system_mac_address': None},
                '321856': {       'dev_macs': ['cc:be:59:32:94:a1'],
                                  'ip_address': None,
                                  'name': '41 MILL ST W (GROPP KEN JR & INGRID)',
                                  'serial': '321856',
                                  'serial_number': '321856',
                                  'system_mac_address': None},
                '321889': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '29A William St. (Fred Manser)',
                                  'serial': '321889',
                                  'serial_number': '321889',
                                  'system_mac_address': None},
                '3218A5': {       'dev_macs': ['d0:76:8f:02:ee:26'],
                                  'ip_address': None,
                                  'name': '91 GERBER DR (DE KLEER SARAH)',
                                  'serial': '3218A5',
                                  'serial_number': '3218A5',
                                  'system_mac_address': None},
                '3218E5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '54 MAIN ST N (GRANT ERIKA)',
                                  'serial': '3218E5',
                                  'serial_number': '3218E5',
                                  'system_mac_address': None},
                '32194B': {       'dev_macs': ['d0:76:8f:1f:69:89'],
                                  'ip_address': None,
                                  'name': '11 Schade St Monkton (STAHLKE KRISTEN)',
                                  'serial': '32194B',
                                  'serial_number': '32194B',
                                  'system_mac_address': None},
                '3219BA': {       'dev_macs': ['cc:be:59:2d:a8:50'],
                                  'ip_address': None,
                                  'name': '90 GERBER DR (LIVINGSTONE MATTHEW)',
                                  'serial': '3219BA',
                                  'serial_number': '3219BA',
                                  'system_mac_address': None},
                '3223CB': {       'dev_macs': ['cc:be:59:7e:b8:4b'],
                                  'ip_address': None,
                                  'name': '14 PIERSON ST (KUEPFER JEFF & SANDRA)',
                                  'serial': '3223CB',
                                  'serial_number': '3223CB',
                                  'system_mac_address': None},
                '322403': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '37 Main St. S. Milverton (Vicki Baier)',
                                  'serial': '322403',
                                  'serial_number': '322403',
                                  'system_mac_address': None},
                '322421': {       'dev_macs': ['e4:6c:d1:4e:ea:20'],
                                  'ip_address': None,
                                  'name': '6717 Rd 140 Milverton (Beat Looser '
                                          '10400601)',
                                  'serial': '322421',
                                  'serial_number': '322421',
                                  'system_mac_address': None},
                '322602': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '51 WILLIAM ST (ENNS JACOB)',
                                  'serial': '322602',
                                  'serial_number': '322602',
                                  'system_mac_address': None},
                '32273F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6680B Road 140 Milverton (Melanie Elg '
                                          '10427044)',
                                  'serial': '32273F',
                                  'serial_number': '32273F',
                                  'system_mac_address': None},
                '3227F1': {       'dev_macs': ['d0:76:8f:02:ef:86'],
                                  'ip_address': None,
                                  'name': '135 Brock St (MORRISON W & V)',
                                  'serial': '3227F1',
                                  'serial_number': '3227F1',
                                  'system_mac_address': None},
                '32281D': {       'dev_macs': ['d0:76:8f:1f:56:f9'],
                                  'ip_address': None,
                                  'name': '58 MAIN ST N (SYDNEY ROCKLAND)',
                                  'serial': '32281D',
                                  'serial_number': '32281D',
                                  'system_mac_address': None},
                '32281E': {       'dev_macs': ['d0:76:8f:1f:5f:c8'],
                                  'ip_address': None,
                                  'name': '86 GERBER DR (NAFZIGER MIKE)',
                                  'serial': '32281E',
                                  'serial_number': '32281E',
                                  'system_mac_address': None},
                '322825': {       'dev_macs': ['04:bc:9f:46:8a:7c'],
                                  'ip_address': None,
                                  'name': '93 GERBER DR (WITT BROOKLYNN)',
                                  'serial': '322825',
                                  'serial_number': '322825',
                                  'system_mac_address': None},
                '322841': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4952 Line 67 Milverton (Ivo Looser '
                                          '10400216)',
                                  'serial': '322841',
                                  'serial_number': '322841',
                                  'system_mac_address': None},
                '322849': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4994 LINE 67 (MCDONALD RICHARD)',
                                  'serial': '322849',
                                  'serial_number': '322849',
                                  'system_mac_address': None},
                '32284F': {       'dev_macs': ['84:d3:43:6d:ed:00'],
                                  'ip_address': None,
                                  'name': '12 Pacific Ave Milverton (Matthew Bancroft '
                                          '10436288)',
                                  'serial': '32284F',
                                  'serial_number': '32284F',
                                  'system_mac_address': None},
                '32288C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5903 Perth Rd 119 Brunner (Marie Lindner '
                                          '10431377)',
                                  'serial': '32288C',
                                  'serial_number': '32288C',
                                  'system_mac_address': None},
                '3228CE': {       'dev_macs': ['cc:be:59:32:95:55'],
                                  'ip_address': None,
                                  'name': '5678 Perth Rd 119 Brunner (Orval Zehr '
                                          '10400708)',
                                  'serial': '3228CE',
                                  'serial_number': '3228CE',
                                  'system_mac_address': None},
                '3229F2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6680A Road 140 Milverton (Bruce Elg '
                                          '10401527)',
                                  'serial': '3229F2',
                                  'serial_number': '3229F2',
                                  'system_mac_address': None},
                '322A47': {       'dev_macs': ['b8:94:70:f3:e7:bb'],
                                  'ip_address': None,
                                  'name': '4344 LINE 61 (JANTZI JARRELL)',
                                  'serial': '322A47',
                                  'serial_number': '322A47',
                                  'system_mac_address': None},
                '329635': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '150 Winstanley St. Monkton (Andre Shtovba '
                                          '10431113)',
                                  'serial': '329635',
                                  'serial_number': '329635',
                                  'system_mac_address': None},
                '329655': {       'dev_macs': ['b8:94:70:9e:cc:d8'],
                                  'ip_address': None,
                                  'name': '48 MADDISON ST E (VILLEMAIRE JACQUELINE)',
                                  'serial': '329655',
                                  'serial_number': '329655',
                                  'system_mac_address': None},
                '329667': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '211 Yonge St., Monkton (Terry & Leslie '
                                          'Brown)',
                                  'serial': '329667',
                                  'serial_number': '329667',
                                  'system_mac_address': None},
                '329678': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '92 MADDISON ST E (HOWES BEN 10430142)',
                                  'serial': '329678',
                                  'serial_number': '329678',
                                  'system_mac_address': None},
                '32967D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '34 MADDISON ST E MONKTON (FELLOWSHIP HAVEN '
                                          'MENNO)',
                                  'serial': '32967D',
                                  'serial_number': '32967D',
                                  'system_mac_address': None},
                '3303B1': {       'dev_macs': ['04:bc:9f:46:92:d2'],
                                  'ip_address': None,
                                  'name': '31 MADDISON ST E (DOHERTY KERRI)',
                                  'serial': '3303B1',
                                  'serial_number': '3303B1',
                                  'system_mac_address': None},
                '3480DD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST APT 1',
                                  'serial': '3480DD',
                                  'serial_number': '3480DD',
                                  'system_mac_address': None},
                '349E51': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 Main St S Ste 17 Milverton (Katherine '
                                          'Priestap 10434089)',
                                  'serial': '349E51',
                                  'serial_number': '349E51',
                                  'system_mac_address': None},
                '349E7D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 WEST ST (STEVENS NCHOLAS)',
                                  'serial': '349E7D',
                                  'serial_number': '349E7D',
                                  'system_mac_address': None},
                '349E85': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '27 Main St S Milverton (Richard Smallbone '
                                          '10433849)',
                                  'serial': '349E85',
                                  'serial_number': '349E85',
                                  'system_mac_address': None},
                '349E87': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 MAIN ST S STE 10 (GROPP BARRY)',
                                  'serial': '349E87',
                                  'serial_number': '349E87',
                                  'system_mac_address': None},
                '349E8A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '111 Mill St E (Adam Harloff)',
                                  'serial': '349E8A',
                                  'serial_number': '349E8A',
                                  'system_mac_address': None},
                '35E8E5': {       'dev_macs': ['84:d3:43:6d:fa:10'],
                                  'ip_address': None,
                                  'name': '5 Main St N Milverton (V & M Lemieux '
                                          '10400112)',
                                  'serial': '35E8E5',
                                  'serial_number': '35E8E5',
                                  'system_mac_address': None},
                '35E8FB': {       'dev_macs': ['a4:2b:8c:7c:8c:8d'],
                                  'ip_address': None,
                                  'name': '170 Winstanley St Monkton (STECKLY JEREMY)',
                                  'serial': '35E8FB',
                                  'serial_number': '35E8FB',
                                  'system_mac_address': None},
                '35E913': {       'dev_macs': ['cc:be:59:7e:b7:85'],
                                  'ip_address': None,
                                  'name': '5818 LINE 55 (ROPP KRISTI)',
                                  'serial': '35E913',
                                  'serial_number': '35E913',
                                  'system_mac_address': None},
                '35E936': {       'dev_macs': ['44:65:7f:89:0b:6d'],
                                  'ip_address': None,
                                  'name': '85 Schade St Monkton (Jesse Hamilton '
                                          '10434996)',
                                  'serial': '35E936',
                                  'serial_number': '35E936',
                                  'system_mac_address': None},
                '35E938': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '22 Whaley Av. Milverton (Roy Wolfe)',
                                  'serial': '35E938',
                                  'serial_number': '35E938',
                                  'system_mac_address': None},
                '35E94D': {       'dev_macs': ['84:d3:43:6d:eb:ab'],
                                  'ip_address': None,
                                  'name': '4316 Line 61 Milverton (E & P Hill '
                                          '10405285)',
                                  'serial': '35E94D',
                                  'serial_number': '35E94D',
                                  'system_mac_address': None},
                '35E94F': {       'dev_macs': ['14:21:03:a9:70:3c'],
                                  'ip_address': None,
                                  'name': '53 MILL ST W (PAULI ROGER)',
                                  'serial': '35E94F',
                                  'serial_number': '35E94F',
                                  'system_mac_address': None},
                '35E95B': {       'dev_macs': ['18:31:bf:e5:79:00'],
                                  'ip_address': None,
                                  'name': '5845 Perth Rd 119 (Matthew & Sheila Ropp '
                                          '10420100)',
                                  'serial': '35E95B',
                                  'serial_number': '35E95B',
                                  'system_mac_address': None},
                '35E95C': {       'dev_macs': ['d8:47:32:21:9a:59'],
                                  'ip_address': None,
                                  'name': 'Arnold Yantzi 10400583 (5734 Perth Rd 119)',
                                  'serial': '35E95C',
                                  'serial_number': '35E95C',
                                  'system_mac_address': None},
                '35E960': {       'dev_macs': ['3c:fa:30:a0:18:10'],
                                  'ip_address': None,
                                  'name': '50 William St. Milverton (Knollcrest Main '
                                          'Internet&Elevator phone)',
                                  'serial': '35E960',
                                  'serial_number': '35E960',
                                  'system_mac_address': None},
                '35E968': {       'dev_macs': ['90:98:38:54:67:72'],
                                  'ip_address': None,
                                  'name': '36 MILL ST E (NAFZIGER BRETT 10438396)',
                                  'serial': '35E968',
                                  'serial_number': '35E968',
                                  'system_mac_address': None},
                '35E96A': {       'dev_macs': ['98:25:4a:de:d1:91'],
                                  'ip_address': None,
                                  'name': '5 John St (Becky Hastings & Trevor Lines)',
                                  'serial': '35E96A',
                                  'serial_number': '35E96A',
                                  'system_mac_address': None},
                '35E96C': {       'dev_macs': ['ec:4f:82:2d:34:75'],
                                  'ip_address': None,
                                  'name': '24 temperance St. Milverton (Royal Can. '
                                          'Legion)',
                                  'serial': '35E96C',
                                  'serial_number': '35E96C',
                                  'system_mac_address': None},
                '35E972': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '1 Main St. N. Milverton (Mane Appeal '
                                          '(Phone))',
                                  'serial': '35E972',
                                  'serial_number': '35E972',
                                  'system_mac_address': None},
                '35E9C3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'ont 6',
                                  'serial': '35E9C3',
                                  'serial_number': '35E9C3',
                                  'system_mac_address': None},
                '35E9CD': {       'dev_macs': ['ec:4f:82:ac:db:db'],
                                  'ip_address': None,
                                  'name': '100 Maddison S W Monkton (Gillian Ward)',
                                  'serial': '35E9CD',
                                  'serial_number': '35E9CD',
                                  'system_mac_address': None},
                '35EA08': {       'dev_macs': ['e8:9f:80:73:c0:9b'],
                                  'ip_address': None,
                                  'name': '5673 Perth Road 119 (Donald & Nelda Erb)',
                                  'serial': '35EA08',
                                  'serial_number': '35EA08',
                                  'system_mac_address': None},
                '35EA28': {       'dev_macs': ['14:dd:a9:92:18:70'],
                                  'ip_address': None,
                                  'name': '5894 Perth Road 119 (Robt & Esther Bitzer '
                                          '10400237)',
                                  'serial': '35EA28',
                                  'serial_number': '35EA28',
                                  'system_mac_address': None},
                '35EA2C': {       'dev_macs': ['e4:6c:d1:4e:f2:ef'],
                                  'ip_address': None,
                                  'name': '5720 PERTH ROAD 119 (BROADHURST EVAN)',
                                  'serial': '35EA2C',
                                  'serial_number': '35EA2C',
                                  'system_mac_address': None},
                '35EA2D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'INVENTORY ROOM (IPTV RESET)',
                                  'serial': '35EA2D',
                                  'serial_number': '35EA2D',
                                  'system_mac_address': None},
                '35EA44': {       'dev_macs': ['ec:4f:82:a5:03:8e'],
                                  'ip_address': None,
                                  'name': '21 Whaley Av. Milverton (Murray Schneider)',
                                  'serial': '35EA44',
                                  'serial_number': '35EA44',
                                  'system_mac_address': None},
                '35EA87': {       'dev_macs': ['e4:6f:13:67:f3:3b'],
                                  'ip_address': None,
                                  'name': '39 Main St. N. Milverton (Guenthers Bakery)',
                                  'serial': '35EA87',
                                  'serial_number': '35EA87',
                                  'system_mac_address': None},
                '35EAA5': {       'dev_macs': ['bc:ee:7b:ec:b5:38'],
                                  'ip_address': None,
                                  'name': '17 Fulton St Milverton (Edward Wiebe '
                                          '10426867)',
                                  'serial': '35EAA5',
                                  'serial_number': '35EAA5',
                                  'system_mac_address': None},
                '35EB24': {       'dev_macs': ['c8:d3:a3:59:59:e3'],
                                  'ip_address': None,
                                  'name': '7 Colbalt St. N. Milverton (Judy Hamilton '
                                          '10422751)',
                                  'serial': '35EB24',
                                  'serial_number': '35EB24',
                                  'system_mac_address': None},
                '35F842': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 llana Heights Upper Apt (Dereth Kipfer)',
                                  'serial': '35F842',
                                  'serial_number': '35F842',
                                  'system_mac_address': None},
                '35F843': {       'dev_macs': ['78:32:1b:52:b3:c6'],
                                  'ip_address': None,
                                  'name': '5620 Perth Rd 119 Brunner (James & Donna '
                                          'Jantzi 10423020)',
                                  'serial': '35F843',
                                  'serial_number': '35F843',
                                  'system_mac_address': None},
                '35F84C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '170 SMITH ST (JANZEN EDDIE)',
                                  'serial': '35F84C',
                                  'serial_number': '35F84C',
                                  'system_mac_address': None},
                '35F851': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '165 Smith St. Monkton',
                                  'serial': '35F851',
                                  'serial_number': '35F851',
                                  'system_mac_address': None},
                '35F953': {       'dev_macs': ['3c:7c:3f:0b:45:d8'],
                                  'ip_address': None,
                                  'name': '82 MADDISON ST E (COCHRANE  DEREK)',
                                  'serial': '35F953',
                                  'serial_number': '35F953',
                                  'system_mac_address': None},
                '35F9A5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '41 Main St. N. Milverton (Ginnie Perry)',
                                  'serial': '35F9A5',
                                  'serial_number': '35F9A5',
                                  'system_mac_address': None},
                '35F9B0': {       'dev_macs': ['a4:2b:b0:ee:cc:2b'],
                                  'ip_address': None,
                                  'name': '50 Mill St E Milverton (Barb Chalmers '
                                          '10400152)',
                                  'serial': '35F9B0',
                                  'serial_number': '35F9B0',
                                  'system_mac_address': None},
                '35F9B1': {       'dev_macs': ['ec:4f:82:a9:a9:11'],
                                  'ip_address': None,
                                  'name': '36 William St (Evan Hargrave)',
                                  'serial': '35F9B1',
                                  'serial_number': '35F9B1',
                                  'system_mac_address': None},
                '35F9E0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '27 Temperance St. Milverton (Chris Potwarka '
                                          '10401128)',
                                  'serial': '35F9E0',
                                  'serial_number': '35F9E0',
                                  'system_mac_address': None},
                '35F9FD': {       'dev_macs': ['84:16:f9:42:b0:f6'],
                                  'ip_address': None,
                                  'name': '5843 Perth Road 119 (Larry P Janzti '
                                          '10400600)',
                                  'serial': '35F9FD',
                                  'serial_number': '35F9FD',
                                  'system_mac_address': None},
                '362811': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '53 MADDISON ST E (KUEPFER BRYCE)',
                                  'serial': '362811',
                                  'serial_number': '362811',
                                  'system_mac_address': None},
                '366C67': {       'dev_macs': [       '00:03:e6:a8:0e:56',
                                                      '00:03:e6:a8:0e:58'],
                                  'ip_address': None,
                                  'name': '23 Schade St., Monkton (Sherry Heibein)',
                                  'serial': '366C67',
                                  'serial_number': '366C67',
                                  'system_mac_address': None},
                '36CC2F': {       'dev_macs': [       'f0:09:0d:f4:23:fe',
                                                      'e8:9f:80:13:2f:f3'],
                                  'ip_address': None,
                                  'name': "34 MAIN ST N (KAREN'S KITCHEN)",
                                  'serial': '36CC2F',
                                  'serial_number': '36CC2F',
                                  'system_mac_address': None},
                '37882C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '9 Main St. N. Milverton (Milv. '
                                          'Accounting&Oak Fin)',
                                  'serial': '37882C',
                                  'serial_number': '37882C',
                                  'system_mac_address': None},
                '378832': {       'dev_macs': ['2c:b0:5d:26:9f:9d'],
                                  'ip_address': None,
                                  'name': '6 Spencer St Milverton (Parks Livestock '
                                          '10420629)',
                                  'serial': '378832',
                                  'serial_number': '378832',
                                  'system_mac_address': None},
                '3788B6': {       'dev_macs': [       '5c:0f:fb:3d:58:38',
                                                      '14:21:03:a1:8c:bb',
                                                      '5c:0f:fb:3d:60:16'],
                                  'ip_address': None,
                                  'name': '50 WILLIAM ST',
                                  'serial': '3788B6',
                                  'serial_number': '3788B6',
                                  'system_mac_address': None},
                '3788BC': {       'dev_macs': ['e0:23:ff:2c:38:ee'],
                                  'ip_address': None,
                                  'name': '20 Main St N Milverton (Ellis HHBC '
                                          '10433408)',
                                  'serial': '3788BC',
                                  'serial_number': '3788BC',
                                  'system_mac_address': None},
                '3788D0': {       'dev_macs': [       '00:15:65:d4:bb:53',
                                                      '00:21:b7:81:63:4b',
                                                      '00:23:24:73:26:76',
                                                      '00:23:24:73:26:f4',
                                                      '00:d8:61:1d:63:0c',
                                                      '00:d8:61:44:bf:14',
                                                      '00:d8:61:45:96:e0',
                                                      '00:d8:61:b5:97:bf',
                                                      '1c:ee:c9:20:75:c7',
                                                      '1c:ee:c9:2a:5b:83',
                                                      '24:9a:d8:2b:8b:69',
                                                      '24:9a:d8:6d:ec:18',
                                                      '30:9c:23:76:22:06',
                                                      '30:9c:23:76:22:35',
                                                      '3c:e3:6b:44:d2:e7',
                                                      '3c:e3:6b:e3:96:18',
                                                      '44:37:e6:7f:79:61',
                                                      '44:db:d2:22:8b:e2',
                                                      '44:db:d2:44:b6:da',
                                                      '44:db:d2:6b:ae:fc',
                                                      '44:db:d2:6b:b0:9d',
                                                      '4c:cc:6a:1c:7c:a8',
                                                      '6c:4b:90:ed:6c:04',
                                                      '6c:4b:90:ed:6c:e9',
                                                      '80:5e:0c:44:4c:85',
                                                      '80:5e:0c:cb:31:02',
                                                      '80:5e:0c:cd:ac:4b',
                                                      '80:5e:0c:f4:ab:9c',
                                                      '80:5e:c0:14:73:d4',
                                                      '80:5e:c0:1c:fe:22',
                                                      '80:5e:c0:1c:fe:c2',
                                                      '80:5e:c0:1d:00:03',
                                                      '80:5e:c0:1d:01:9e',
                                                      '80:5e:c0:1d:03:36',
                                                      '80:5e:c0:2f:63:44',
                                                      '80:5e:c0:2f:64:18',
                                                      '80:5e:c0:2f:69:51',
                                                      'cc:d5:39:1d:4d:82',
                                                      'cc:d5:39:1d:4d:c0',
                                                      'cc:d5:39:1d:4d:c1',
                                                      'd8:cb:8a:92:e1:bc',
                                                      'f4:93:9f:f0:69:cd',
                                                      'fc:4d:d4:d4:0a:a8',
                                                      'fc:4d:d4:f4:32:bb'],
                                  'ip_address': None,
                                  'name': '50 William St.  Milverton (Knollcrest Data '
                                          'Rm 3)',
                                  'serial': '3788D0',
                                  'serial_number': '3788D0',
                                  'system_mac_address': None},
                '38F22A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '15 Spencer St. Milverton (Township Perth '
                                          'East)',
                                  'serial': '38F22A',
                                  'serial_number': '38F22A',
                                  'system_mac_address': None},
                '38F285': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '205 Winstanley St. Monkton (TJ Choi)',
                                  'serial': '38F285',
                                  'serial_number': '38F285',
                                  'system_mac_address': None},
                '38FD7F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '91 MAIN ST N (KUEPFER TRAVIS)',
                                  'serial': '38FD7F',
                                  'serial_number': '38FD7F',
                                  'system_mac_address': None},
                '3901AA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '86 Mill St. E. Milverton (John Spencer)',
                                  'serial': '3901AA',
                                  'serial_number': '3901AA',
                                  'system_mac_address': None},
                '3908F9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '33 Maddison St. W. (Paul Ward)',
                                  'serial': '3908F9',
                                  'serial_number': '3908F9',
                                  'system_mac_address': None},
                '390C5E': {       'dev_macs': ['60:db:98:82:08:01'],
                                  'ip_address': None,
                                  'name': '5384 Line 60 (Don&Barb Boyne)',
                                  'serial': '390C5E',
                                  'serial_number': '390C5E',
                                  'system_mac_address': None},
                '390DC8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '61 SCHADE ST (BROUGHTON KELLY)',
                                  'serial': '390DC8',
                                  'serial_number': '390DC8',
                                  'system_mac_address': None},
                '3940DE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '159 Fishleigh St. Monkton (Henry Vander '
                                          'Hyden)',
                                  'serial': '3940DE',
                                  'serial_number': '3940DE',
                                  'system_mac_address': None},
                '3961A8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 SPENCER ST (MARK JUTZI FUNERAL HOME)',
                                  'serial': '3961A8',
                                  'serial_number': '3961A8',
                                  'system_mac_address': None},
                '3B61DE': {       'dev_macs': ['d0:76:8f:91:4b:52'],
                                  'ip_address': None,
                                  'name': '72 GERBER DR (MCCLUSKIE CODY)',
                                  'serial': '3B61DE',
                                  'serial_number': '3B61DE',
                                  'system_mac_address': None},
                '3B6200': {       'dev_macs': ['cc:be:59:2d:a6:2b'],
                                  'ip_address': None,
                                  'name': '15 Pacific Ave (Steven & Debbie Emerson, '
                                          '10404480)',
                                  'serial': '3B6200',
                                  'serial_number': '3B6200',
                                  'system_mac_address': None},
                '3CFA9': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '95 Mill St. E. Milverton (Pumphouse Perth '
                                         'East)',
                                 'serial': '3CFA9',
                                 'serial_number': '3CFA9',
                                 'system_mac_address': None},
                '3FA90D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5460 LINE 60 (HARNEST ALEXANDER)',
                                  'serial': '3FA90D',
                                  'serial_number': '3FA90D',
                                  'system_mac_address': None},
                '3FC845': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '202 WINSTANLEY ST (KIRKHAM DOUG & KELLI)',
                                  'serial': '3FC845',
                                  'serial_number': '3FC845',
                                  'system_mac_address': None},
                '418B0': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '146 Winstanley St. Monkton (Robert Straeten)',
                                 'serial': '418B0',
                                 'serial_number': '418B0',
                                 'system_mac_address': None},
                '4281BF': {       'dev_macs': ['5c:0f:fb:43:97:c6'],
                                  'ip_address': None,
                                  'name': '76 Schade St Monkton (Murray Crawford '
                                          '10434552)',
                                  'serial': '4281BF',
                                  'serial_number': '4281BF',
                                  'system_mac_address': None},
                '4281C6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '139 Brock St Monkton (Angela Pestell '
                                          '10412111)',
                                  'serial': '4281C6',
                                  'serial_number': '4281C6',
                                  'system_mac_address': None},
                '4281D3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '94 GERBER DR (BUDDELL WILLIAM)',
                                  'serial': '4281D3',
                                  'serial_number': '4281D3',
                                  'system_mac_address': None},
                '4281E9': {       'dev_macs': ['14:21:03:a9:70:10'],
                                  'ip_address': None,
                                  'name': '161 Brock St Monkton (Wilma Acheson '
                                          '10434491)',
                                  'serial': '4281E9',
                                  'serial_number': '4281E9',
                                  'system_mac_address': None},
                '42844C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '126 Brock St (Kevin Mclean, 10422928)',
                                  'serial': '42844C',
                                  'serial_number': '42844C',
                                  'system_mac_address': None},
                '42B9B6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '45 Temperance St (Lindsay Hammer 10423568)',
                                  'serial': '42B9B6',
                                  'serial_number': '42B9B6',
                                  'system_mac_address': None},
                '42B9C3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6322 Rd. 148 (Kurt Gsell)',
                                  'serial': '42B9C3',
                                  'serial_number': '42B9C3',
                                  'system_mac_address': None},
                '42B9D7': {       'dev_macs': ['44:65:7f:41:8f:0b'],
                                  'ip_address': None,
                                  'name': '75 Gerber Drive Milverton (Cezary Cybulski '
                                          '10434099)',
                                  'serial': '42B9D7',
                                  'serial_number': '42B9D7',
                                  'system_mac_address': None},
                '42B9ED': {       'dev_macs': ['44:65:7f:79:d0:b7'],
                                  'ip_address': None,
                                  'name': '11 Main St S Milverton (Milverton Mini Mart '
                                          '10420723)',
                                  'serial': '42B9ED',
                                  'serial_number': '42B9ED',
                                  'system_mac_address': None},
                '42BE9A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4895 Line 67 House (Ron and Deborah '
                                          'Riddell)',
                                  'serial': '42BE9A',
                                  'serial_number': '42BE9A',
                                  'system_mac_address': None},
                '42BEA0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4633 Line 61 Milverton (James Kuepfer '
                                          '10400294)',
                                  'serial': '42BEA0',
                                  'serial_number': '42BEA0',
                                  'system_mac_address': None},
                '42BEA2': {       'dev_macs': ['44:65:7f:88:d5:a3'],
                                  'ip_address': None,
                                  'name': '5126 Line 58 Milverton (J & B Crowley '
                                          '10404213)',
                                  'serial': '42BEA2',
                                  'serial_number': '42BEA2',
                                  'system_mac_address': None},
                '42BEA7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5088 Line 58 Milverton (Franz & Susy Arnold '
                                          '10400118)',
                                  'serial': '42BEA7',
                                  'serial_number': '42BEA7',
                                  'system_mac_address': None},
                '42BEC1': {       'dev_macs': ['60:db:98:81:fa:a4'],
                                  'ip_address': None,
                                  'name': '47 MAIN ST N (WOLF JONATHAN)',
                                  'serial': '42BEC1',
                                  'serial_number': '42BEC1',
                                  'system_mac_address': None},
                '42BEDF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 John St Milverton (Craig Lines 10422212)',
                                  'serial': '42BEDF',
                                  'serial_number': '42BEDF',
                                  'system_mac_address': None},
                '42BF03': {       'dev_macs': ['cc:be:59:2d:a5:e3'],
                                  'ip_address': None,
                                  'name': '88 GERBER DRIVE (DEWAR JAMIE & ANGIE)',
                                  'serial': '42BF03',
                                  'serial_number': '42BF03',
                                  'system_mac_address': None},
                '42BF08': {       'dev_macs': [       '00:03:e6:dd:2e:38',
                                                      '70:56:81:c5:c5:ba'],
                                  'ip_address': None,
                                  'name': '4900 LINE 52 (DEGROOT JIM & WENDY)',
                                  'serial': '42BF08',
                                  'serial_number': '42BF08',
                                  'system_mac_address': None},
                '42BF1B': {       'dev_macs': ['e4:6c:d1:4e:ea:83'],
                                  'ip_address': None,
                                  'name': '77A REAGAN ST (LARA MARCELLO)',
                                  'serial': '42BF1B',
                                  'serial_number': '42BF1B',
                                  'system_mac_address': None},
                '42BF1D': {       'dev_macs': ['ec:4f:82:ac:f0:12'],
                                  'ip_address': None,
                                  'name': '6238 Road 140 Milverton (Rick Diehl '
                                          '10404130)',
                                  'serial': '42BF1D',
                                  'serial_number': '42BF1D',
                                  'system_mac_address': None},
                '42C048': {       'dev_macs': ['44:65:7f:89:0f:39'],
                                  'ip_address': None,
                                  'name': '194 SELBY AV (SMITH LORI)',
                                  'serial': '42C048',
                                  'serial_number': '42C048',
                                  'system_mac_address': None},
                '42C0AF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '37 Main St. N. Milverton (Doris Coultes)',
                                  'serial': '42C0AF',
                                  'serial_number': '42C0AF',
                                  'system_mac_address': None},
                '42C2D3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5256 Line 49 Gadshill (M & K DeJong '
                                          '10427640)',
                                  'serial': '42C2D3',
                                  'serial_number': '42C2D3',
                                  'system_mac_address': None},
                '42C38C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5340B Line 49 Gadshill (Stu McDonald '
                                          '10422004)',
                                  'serial': '42C38C',
                                  'serial_number': '42C38C',
                                  'system_mac_address': None},
                '42C38E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5340A Line 49 (Malcolm McDonald 10400307)',
                                  'serial': '42C38E',
                                  'serial_number': '42C38E',
                                  'system_mac_address': None},
                '42C40E': {       'dev_macs': ['44:65:7f:41:8e:57'],
                                  'ip_address': None,
                                  'name': '36 Gerber Ct Milverton (Norman Bell '
                                          '10435005)',
                                  'serial': '42C40E',
                                  'serial_number': '42C40E',
                                  'system_mac_address': None},
                '42C467': {       'dev_macs': ['44:65:7f:88:d5:40'],
                                  'ip_address': None,
                                  'name': '51 MILL ST E (HIPPE AMANDA)',
                                  'serial': '42C467',
                                  'serial_number': '42C467',
                                  'system_mac_address': None},
                '42C4A9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '199 WINSTANLEY ST (WHETSTONE JOSH)',
                                  'serial': '42C4A9',
                                  'serial_number': '42C4A9',
                                  'system_mac_address': None},
                '42C4C9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '140 Brock St Monkton (R & B Buchanan '
                                          '10430719)',
                                  'serial': '42C4C9',
                                  'serial_number': '42C4C9',
                                  'system_mac_address': None},
                '42C523': {       'dev_macs': ['b8:94:70:f3:e3:64'],
                                  'ip_address': None,
                                  'name': '84 Reagan St Milverton (Martin Andrews '
                                          '10435676)',
                                  'serial': '42C523',
                                  'serial_number': '42C523',
                                  'system_mac_address': None},
                '42C595': {       'dev_macs': ['44:65:7f:88:d3:75'],
                                  'ip_address': None,
                                  'name': '136 Brock St Monkton (Dale MacCannell '
                                          '10435836)',
                                  'serial': '42C595',
                                  'serial_number': '42C595',
                                  'system_mac_address': None},
                '42C5B4': {       'dev_macs': ['44:65:7f:88:d5:0a'],
                                  'ip_address': None,
                                  'name': '67 Reagan St. Milverton (Joy Sheridan '
                                          '10435977)',
                                  'serial': '42C5B4',
                                  'serial_number': '42C5B4',
                                  'system_mac_address': None},
                '42C6D2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '42 MADDISON ST E (MARTIN GERALD)',
                                  'serial': '42C6D2',
                                  'serial_number': '42C6D2',
                                  'system_mac_address': None},
                '42C702': {       'dev_macs': ['ec:4f:82:2b:48:69'],
                                  'ip_address': None,
                                  'name': '191 Selby Av. Monkton (Roger Orth)',
                                  'serial': '42C702',
                                  'serial_number': '42C702',
                                  'system_mac_address': None},
                '42C70A': {       'dev_macs': ['5c:0f:fb:2a:05:8a'],
                                  'ip_address': None,
                                  'name': '11 MacBeth St (David Sloat 10401199)',
                                  'serial': '42C70A',
                                  'serial_number': '42C70A',
                                  'system_mac_address': None},
                '42C70B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5137 Line 49 Gadshill (MJD FARMS LTD)',
                                  'serial': '42C70B',
                                  'serial_number': '42C70B',
                                  'system_mac_address': None},
                '42C748': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4987 Line 59 (Prima Valley Farms Ltd '
                                          '10400278)',
                                  'serial': '42C748',
                                  'serial_number': '42C748',
                                  'system_mac_address': None},
                '42C89C': {       'dev_macs': ['b8:94:70:f3:cb:ac'],
                                  'ip_address': None,
                                  'name': '4660 LINE 61 (MILVERTON ROD & GUN CLUB INC)',
                                  'serial': '42C89C',
                                  'serial_number': '42C89C',
                                  'system_mac_address': None},
                '42C8D3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4775 LINE 61 (BAKER TYLER & JULIE)',
                                  'serial': '42C8D3',
                                  'serial_number': '42C8D3',
                                  'system_mac_address': None},
                '42C8EA': {       'dev_macs': ['a8:42:a1:29:f2:75'],
                                  'ip_address': None,
                                  'name': '5383 Perth Road 135 (Ryan & Mandy '
                                          'Kuntze10421167)',
                                  'serial': '42C8EA',
                                  'serial_number': '42C8EA',
                                  'system_mac_address': None},
                '42C919': {       'dev_macs': ['00:09:0f:09:01:00'],
                                  'ip_address': None,
                                  'name': '4743 Line 52 Brunner (Milverton Layer '
                                          'Farms)',
                                  'serial': '42C919',
                                  'serial_number': '42C919',
                                  'system_mac_address': None},
                '42CA30': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '19 MAIN ST N APT 2 MILVERTON (FRIESEN '
                                          'HENRY)',
                                  'serial': '42CA30',
                                  'serial_number': '42CA30',
                                  'system_mac_address': None},
                '42CD66': {       'dev_macs': ['44:65:7f:88:45:7f'],
                                  'ip_address': None,
                                  'name': '5316 Road 144 Brunner (W & J Paxton '
                                          '10435748)',
                                  'serial': '42CD66',
                                  'serial_number': '42CD66',
                                  'system_mac_address': None},
                '42CDDC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5099 Line 49 Gadshill (Andy Groenestege '
                                          '10400881)',
                                  'serial': '42CDDC',
                                  'serial_number': '42CDDC',
                                  'system_mac_address': None},
                '42CE32': {       'dev_macs': ['44:65:7f:89:0e:58'],
                                  'ip_address': None,
                                  'name': '4661 LINE 49 (ROES JASON & ANGELA)',
                                  'serial': '42CE32',
                                  'serial_number': '42CE32',
                                  'system_mac_address': None},
                '42CE4A': {       'dev_macs': ['44:65:7f:89:0c:45'],
                                  'ip_address': None,
                                  'name': '4860 Line 52 Brunner (Ronald Raycraft '
                                          '10400550)',
                                  'serial': '42CE4A',
                                  'serial_number': '42CE4A',
                                  'system_mac_address': None},
                '42D092': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4606 Line 49 Gadshill (J & A Debus '
                                          '10400821)',
                                  'serial': '42D092',
                                  'serial_number': '42D092',
                                  'system_mac_address': None},
                '42D0B4': {       'dev_macs': ['60:db:98:54:ab:1b'],
                                  'ip_address': None,
                                  'name': '9 Colbalt St. S. Milverton (Jamie Scott)',
                                  'serial': '42D0B4',
                                  'serial_number': '42D0B4',
                                  'system_mac_address': None},
                '42D154': {       'dev_macs': ['84:d3:43:6d:e7:e3'],
                                  'ip_address': None,
                                  'name': '4535 Line 49 (Helmut&Liz Mergl 10400543)',
                                  'serial': '42D154',
                                  'serial_number': '42D154',
                                  'system_mac_address': None},
                '42D196': {       'dev_macs': ['44:65:7f:89:0f:42'],
                                  'ip_address': None,
                                  'name': '4650 Line 49-3 (J & J Millwrighting '
                                          '10403280)',
                                  'serial': '42D196',
                                  'serial_number': '42D196',
                                  'system_mac_address': None},
                '42D1CE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5113 Perth Road 135 Rostock (B & D Debus '
                                          '10402399)',
                                  'serial': '42D1CE',
                                  'serial_number': '42D1CE',
                                  'system_mac_address': None},
                '42D2AE': {       'dev_macs': ['44:65:7f:41:9c:9d'],
                                  'ip_address': None,
                                  'name': '79 GERBER DR (DOHERTY LARRY & DONNA)',
                                  'serial': '42D2AE',
                                  'serial_number': '42D2AE',
                                  'system_mac_address': None},
                '42D2AF': {       'dev_macs': ['d0:76:8f:02:ee:73'],
                                  'ip_address': None,
                                  'name': '5216 Line 49 (MCDONNELL MAGGIE)',
                                  'serial': '42D2AF',
                                  'serial_number': '42D2AF',
                                  'system_mac_address': None},
                '42D361': {       'dev_macs': ['e4:6c:d1:4e:f0:c9'],
                                  'ip_address': None,
                                  'name': '85B Mill St E Milverton (Tricia Nahrgang '
                                          '10435488)',
                                  'serial': '42D361',
                                  'serial_number': '42D361',
                                  'system_mac_address': None},
                '43B154': {       'dev_macs': ['d0:76:8f:02:ee:c0'],
                                  'ip_address': None,
                                  'name': '14 Pacific Ave (Mitchell Medeiros10434078)',
                                  'serial': '43B154',
                                  'serial_number': '43B154',
                                  'system_mac_address': None},
                '43B2D7': {       'dev_macs': ['d8:50:e6:c3:5a:05'],
                                  'ip_address': None,
                                  'name': '5668 Perth Rd 119 Brunner (Simon Jantzi '
                                          '10400499)',
                                  'serial': '43B2D7',
                                  'serial_number': '43B2D7',
                                  'system_mac_address': None},
                '43B37D': {       'dev_macs': ['d8:b3:70:12:51:39'],
                                  'ip_address': None,
                                  'name': '19 TEMPERANCE ST (BOWEY JOHN)',
                                  'serial': '43B37D',
                                  'serial_number': '43B37D',
                                  'system_mac_address': None},
                '43B38E': {       'dev_macs': ['6c:cd:d6:eb:84:7d'],
                                  'ip_address': None,
                                  'name': '4978 Line 52 Brunner (Andrew Brodhagen '
                                          '10433259)',
                                  'serial': '43B38E',
                                  'serial_number': '43B38E',
                                  'system_mac_address': None},
                '43B3A2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 William St Apt 4 Milverton',
                                  'serial': '43B3A2',
                                  'serial_number': '43B3A2',
                                  'system_mac_address': None},
                '43B3B6': {       'dev_macs': ['38:d5:47:87:96:d0'],
                                  'ip_address': None,
                                  'name': '4934 LINE 49 (LEYBOURNE KEVIN)',
                                  'serial': '43B3B6',
                                  'serial_number': '43B3B6',
                                  'system_mac_address': None},
                '43B3F4': {       'dev_macs': ['28:3b:82:66:18:9e'],
                                  'ip_address': None,
                                  'name': '15 West St Milverton (Maggie Enns 10430560)',
                                  'serial': '43B3F4',
                                  'serial_number': '43B3F4',
                                  'system_mac_address': None},
                '43B508': {       'dev_macs': ['20:23:51:49:b5:a9'],
                                  'ip_address': None,
                                  'name': '73 REAGAN ST (MILLER PAUL)',
                                  'serial': '43B508',
                                  'serial_number': '43B508',
                                  'system_mac_address': None},
                '43B54C': {       'dev_macs': ['70:62:b8:2f:ef:61'],
                                  'ip_address': None,
                                  'name': '20 ST MARYS ST S (HIEBERT JUDITH)',
                                  'serial': '43B54C',
                                  'serial_number': '43B54C',
                                  'system_mac_address': None},
                '43B5B7': {       'dev_macs': ['d0:76:8f:1f:66:df'],
                                  'ip_address': None,
                                  'name': '6121 Ana St Brunner (Stephan Gingerich '
                                          '10435309)',
                                  'serial': '43B5B7',
                                  'serial_number': '43B5B7',
                                  'system_mac_address': None},
                '43C1B8': {       'dev_macs': ['0c:80:63:34:53:53'],
                                  'ip_address': None,
                                  'name': '78 GERBER DR (LAURIAULT NICOLE)',
                                  'serial': '43C1B8',
                                  'serial_number': '43C1B8',
                                  'system_mac_address': None},
                '43C24B': {       'dev_macs': ['a0:ab:1b:88:c0:af'],
                                  'ip_address': None,
                                  'name': '74 Gerber Drive Milverton (Shannon Stangen '
                                          '10434431)',
                                  'serial': '43C24B',
                                  'serial_number': '43C24B',
                                  'system_mac_address': None},
                '43C42C': {       'dev_macs': ['14:59:c0:bf:74:f6'],
                                  'ip_address': None,
                                  'name': '153 Winstanley St. Monkton (Ben Champion)',
                                  'serial': '43C42C',
                                  'serial_number': '43C42C',
                                  'system_mac_address': None},
                '44AA8F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 MAIN ST N MILVERTON (VILLAGE PHARMACY)',
                                  'serial': '44AA8F',
                                  'serial_number': '44AA8F',
                                  'system_mac_address': None},
                '44AAA8': {       'dev_macs': [       '00:03:e6:f9:e2:1e',
                                                      'f0:ad:4e:18:9f:1c',
                                                      '00:03:e6:f9:fb:6f'],
                                  'ip_address': None,
                                  'name': '200 Nelson St. Monkton (North Perth Elma '
                                          'Logan Rec Complex)',
                                  'serial': '44AAA8',
                                  'serial_number': '44AAA8',
                                  'system_mac_address': None},
                '459638': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '27 Main St. N. Milverton (Koertland Garage)',
                                  'serial': '459638',
                                  'serial_number': '459638',
                                  'system_mac_address': None},
                '45965D': {       'dev_macs': ['00:1e:e3:12:34:56'],
                                  'ip_address': None,
                                  'name': '111 Maddison St. W. Monkton (Steve Smith '
                                          'Auto Sales)',
                                  'serial': '45965D',
                                  'serial_number': '45965D',
                                  'system_mac_address': None},
                '45968C': {       'dev_macs': ['5c:0f:fb:3d:60:02'],
                                  'ip_address': None,
                                  'name': '50 William St. Milverton (Knollcrest Data '
                                          'Rm. 1)',
                                  'serial': '45968C',
                                  'serial_number': '45968C',
                                  'system_mac_address': None},
                '4596A5': {       'dev_macs': [       '00:5f:67:a4:50:5d',
                                                      '60:83:e7:7c:39:98',
                                                      '24:2f:d0:93:60:3b'],
                                  'ip_address': None,
                                  'name': '9 & 11 Maddison Street E (Multi-customer)',
                                  'serial': '4596A5',
                                  'serial_number': '4596A5',
                                  'system_mac_address': None},
                '471755': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5387 LINE 60 (COXON TYE M)',
                                  'serial': '471755',
                                  'serial_number': '471755',
                                  'system_mac_address': None},
                '47E37': {       'dev_macs': [       '5c:0f:fb:3d:52:86',
                                                     '5c:0f:fb:3d:80:08',
                                                     '00:03:e6:f9:7f:8a',
                                                     '5c:0f:fb:3d:58:aa'],
                                 'ip_address': None,
                                 'name': '50 WILLIAM ST (WOODS RUTH A)',
                                 'serial': '47E37',
                                 'serial_number': '47E37',
                                 'system_mac_address': None},
                '4D114C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '165 WINSTANLEY ST (KLOOSTRA ALEX)',
                                  'serial': '4D114C',
                                  'serial_number': '4D114C',
                                  'system_mac_address': None},
                '4D1152': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '206 Queen Ave, Monkton',
                                  'serial': '4D1152',
                                  'serial_number': '4D1152',
                                  'system_mac_address': None},
                '4D41F2': {       'dev_macs': ['00:03:e6:eb:7d:1e'],
                                  'ip_address': None,
                                  'name': '218 Selby Ave, Monkton (Steve & Krystle '
                                          'Pettapiece)',
                                  'serial': '4D41F2',
                                  'serial_number': '4D41F2',
                                  'system_mac_address': None},
                '4D4239': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '214 SELBY AVENUE (WARD KELLY-ANN)',
                                  'serial': '4D4239',
                                  'serial_number': '4D4239',
                                  'system_mac_address': None},
                '4D423B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '100 MADDISON ST E (BOYD RON & KATHY)',
                                  'serial': '4D423B',
                                  'serial_number': '4D423B',
                                  'system_mac_address': None},
                '4D423F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '189 WINSTANLEY ST',
                                  'serial': '4D423F',
                                  'serial_number': '4D423F',
                                  'system_mac_address': None},
                '4D4244': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '22 Maddison St. West, Monkton (George & Pat '
                                          'Worth)',
                                  'serial': '4D4244',
                                  'serial_number': '4D4244',
                                  'system_mac_address': None},
                '4D424C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '170 Fishleigh St., Monkton (Ben Fizell)',
                                  'serial': '4D424C',
                                  'serial_number': '4D424C',
                                  'system_mac_address': None},
                '4D424E': {       'dev_macs': ['00:03:e6:a4:d3:80'],
                                  'ip_address': None,
                                  'name': '22 Winstanley St., Monkton (Natasha Baxter '
                                          '10433874)',
                                  'serial': '4D424E',
                                  'serial_number': '4D424E',
                                  'system_mac_address': None},
                '4EE08': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '8 Pugh St. Milverton (MCCL Cell Tower)',
                                 'serial': '4EE08',
                                 'serial_number': '4EE08',
                                 'system_mac_address': None},
                '4F8C9B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '82 Gerber Drive Milverton',
                                  'serial': '4F8C9B',
                                  'serial_number': '4F8C9B',
                                  'system_mac_address': None},
                '4F91FF': {       'dev_macs': ['00:31:92:dc:5c:f9'],
                                  'ip_address': None,
                                  'name': '61 Reagan St (Philip & Rosemary '
                                          'Lebold,10411732)',
                                  'serial': '4F91FF',
                                  'serial_number': '4F91FF',
                                  'system_mac_address': None},
                '4F92CF': {       'dev_macs': ['78:32:1b:49:95:0f'],
                                  'ip_address': None,
                                  'name': '5713 Perth Rd 119 Brunner (Harry Walmsley '
                                          '10402219)',
                                  'serial': '4F92CF',
                                  'serial_number': '4F92CF',
                                  'system_mac_address': None},
                '4F9AE4': {       'dev_macs': ['b8:94:70:f3:ec:33'],
                                  'ip_address': None,
                                  'name': '8A WILLIAM ST (HOLST SAMUEL)',
                                  'serial': '4F9AE4',
                                  'serial_number': '4F9AE4',
                                  'system_mac_address': None},
                '4F9B37': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5 MacBeth St Milverton (Damion Bijman '
                                          '10434605)',
                                  'serial': '4F9B37',
                                  'serial_number': '4F9B37',
                                  'system_mac_address': None},
                '4F9B3C': {       'dev_macs': ['e8:fc:af:f9:6d:00'],
                                  'ip_address': None,
                                  'name': '6383 Road 124 Milverton (Brad Gascho '
                                          '10400853)',
                                  'serial': '4F9B3C',
                                  'serial_number': '4F9B3C',
                                  'system_mac_address': None},
                '4FA244': {       'dev_macs': ['cc:be:59:7e:b6:26'],
                                  'ip_address': None,
                                  'name': '3 West St Milverton (Joe Kuepfer 10426185)',
                                  'serial': '4FA244',
                                  'serial_number': '4FA244',
                                  'system_mac_address': None},
                '4FA249': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 MacBeth St Milverton (Abe Hildebrandt '
                                          '10435622)',
                                  'serial': '4FA249',
                                  'serial_number': '4FA249',
                                  'system_mac_address': None},
                '4FA25D': {       'dev_macs': ['04:bc:9f:46:94:11'],
                                  'ip_address': None,
                                  'name': '25 MILL ST W (ENNS ANTHONY)',
                                  'serial': '4FA25D',
                                  'serial_number': '4FA25D',
                                  'system_mac_address': None},
                '5124E1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '17 MAIN ST N',
                                  'serial': '5124E1',
                                  'serial_number': '5124E1',
                                  'system_mac_address': None},
                '5124E6': {       'dev_macs': ['68:ff:7b:33:69:73'],
                                  'ip_address': None,
                                  'name': '4621 Line 49 Brunner (Andrew Debus '
                                          '10435392)',
                                  'serial': '5124E6',
                                  'serial_number': '5124E6',
                                  'system_mac_address': None},
                '51250D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'Rueben Kuepfer 10431068 (85 Reagan St)',
                                  'serial': '51250D',
                                  'serial_number': '51250D',
                                  'system_mac_address': None},
                '512523': {       'dev_macs': ['44:65:7f:89:0e:97'],
                                  'ip_address': None,
                                  'name': '4538 Line 49 (Samuel Ebersol 10430136)',
                                  'serial': '512523',
                                  'serial_number': '512523',
                                  'system_mac_address': None},
                '512575': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4650 Line 49 (Linda Debus10403081)',
                                  'serial': '512575',
                                  'serial_number': '512575',
                                  'system_mac_address': None},
                '512576': {       'dev_macs': ['e8:8d:28:55:cd:f1'],
                                  'ip_address': None,
                                  'name': '89 Main St N Milverton (Doug & Anna Hodgins '
                                          '10403089)',
                                  'serial': '512576',
                                  'serial_number': '512576',
                                  'system_mac_address': None},
                '512588': {       'dev_macs': ['c4:e9:84:87:af:6d'],
                                  'ip_address': None,
                                  'name': '4660 Line 52 Brunner (Rick Reis 10400560)',
                                  'serial': '512588',
                                  'serial_number': '512588',
                                  'system_mac_address': None},
                '5125CA': {       'dev_macs': ['a0:ab:1b:7b:68:2d'],
                                  'ip_address': None,
                                  'name': '52 Main St N Milverton (Diane Schweitzer '
                                          '10401082)',
                                  'serial': '5125CA',
                                  'serial_number': '5125CA',
                                  'system_mac_address': None},
                '5125E0': {       'dev_macs': ['60:db:98:09:1b:3e'],
                                  'ip_address': None,
                                  'name': '5 GEORGE ST (KRAFT STEPHANIE)',
                                  'serial': '5125E0',
                                  'serial_number': '5125E0',
                                  'system_mac_address': None},
                '5125E3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '32 MAIN ST N',
                                  'serial': '5125E3',
                                  'serial_number': '5125E3',
                                  'system_mac_address': None},
                '5125E4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': "5183 ROAD 130 (ST PAUL'S LUTHERAN CHURCH)",
                                  'serial': '5125E4',
                                  'serial_number': '5125E4',
                                  'system_mac_address': None},
                '5125F0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4514 Line 49 (Isaak Janzen 10429928)',
                                  'serial': '5125F0',
                                  'serial_number': '5125F0',
                                  'system_mac_address': None},
                '555C9F': {       'dev_macs': ['78:9a:18:18:7f:b0'],
                                  'ip_address': None,
                                  'name': '6380 Perth Rd 131 (MASONAL STONE)',
                                  'serial': '555C9F',
                                  'serial_number': '555C9F',
                                  'system_mac_address': None},
                '556382': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '69 Schade St. Monkton (Jay Yundt)',
                                  'serial': '556382',
                                  'serial_number': '556382',
                                  'system_mac_address': None},
                '556393': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '72 Schade St, Monkton (Geoff & Jodie '
                                          'Tummers 10430143)',
                                  'serial': '556393',
                                  'serial_number': '556393',
                                  'system_mac_address': None},
                '5563A9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '130 Maddison St. E. Monkton (Don Hymers)',
                                  'serial': '5563A9',
                                  'serial_number': '5563A9',
                                  'system_mac_address': None},
                '5563D4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '59 Maddison St. West, Monkton (Kevin Cook '
                                          '10430147)',
                                  'serial': '5563D4',
                                  'serial_number': '5563D4',
                                  'system_mac_address': None},
                '5563DF': {       'dev_macs': ['cc:be:59:7e:b7:19'],
                                  'ip_address': None,
                                  'name': '197 King Ave, Monkton (Dana Verwey)',
                                  'serial': '5563DF',
                                  'serial_number': '5563DF',
                                  'system_mac_address': None},
                '5563E9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '71 REAGAN ST (CAO QUI PHUC)',
                                  'serial': '5563E9',
                                  'serial_number': '5563E9',
                                  'system_mac_address': None},
                '55F537': {       'dev_macs': [       '08:96:ad:ec:1f:2d',
                                                      '08:96:ad:ec:1f:40',
                                                      '00:50:56:a3:97:c5',
                                                      'f8:6b:d9:26:63:e1'],
                                  'ip_address': None,
                                  'name': '132 Maddison St. W. Monkton (WFS Great '
                                          'Lakes 10424552)',
                                  'serial': '55F537',
                                  'serial_number': '55F537',
                                  'system_mac_address': None},
                '55F9BC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6432 ROAD 154 (OHM PERRY & JENNIFER)',
                                  'serial': '55F9BC',
                                  'serial_number': '55F9BC',
                                  'system_mac_address': None},
                '55F9C4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '166 Fishleigh St. Monkton (Keith Kuepfer '
                                          '10403450)',
                                  'serial': '55F9C4',
                                  'serial_number': '55F9C4',
                                  'system_mac_address': None},
                '560258': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '185 Winstanley St. Monkton (Becky Beuerman)',
                                  'serial': '560258',
                                  'serial_number': '560258',
                                  'system_mac_address': None},
                '560296': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '210 Winstanley Ave, Monkton (Donna '
                                          'McConnell)',
                                  'serial': '560296',
                                  'serial_number': '560296',
                                  'system_mac_address': None},
                '560C10': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '164 Smith St. Monkton (Bob Mclean)',
                                  'serial': '560C10',
                                  'serial_number': '560C10',
                                  'system_mac_address': None},
                '561162': {       'dev_macs': ['cc:be:59:32:93:c9'],
                                  'ip_address': None,
                                  'name': '206 Winstanely St., Monkton (Nicole '
                                          'Willfong 10434242)',
                                  'serial': '561162',
                                  'serial_number': '561162',
                                  'system_mac_address': None},
                '56171C': {       'dev_macs': ['44:65:7f:79:de:01'],
                                  'ip_address': None,
                                  'name': '63 Reagan St Milverton (Megan Hines '
                                          '10435292)',
                                  'serial': '56171C',
                                  'serial_number': '56171C',
                                  'system_mac_address': None},
                '5A61E6': {       'dev_macs': ['44:65:7f:b5:29:55'],
                                  'ip_address': None,
                                  'name': '5 Murray Ct Milverton (Lorne Steckley '
                                          '10402394)',
                                  'serial': '5A61E6',
                                  'serial_number': '5A61E6',
                                  'system_mac_address': None},
                '5A621B': {       'dev_macs': ['a8:42:a1:5b:fc:71'],
                                  'ip_address': None,
                                  'name': '44 Mill St E Milverton (Douglas Fink '
                                          '10425982)',
                                  'serial': '5A621B',
                                  'serial_number': '5A621B',
                                  'system_mac_address': None},
                '5A6223': {       'dev_macs': ['b8:94:70:f3:e0:6d'],
                                  'ip_address': None,
                                  'name': '4725 LINE 61 (MUELLER STEFAN)',
                                  'serial': '5A6223',
                                  'serial_number': '5A6223',
                                  'system_mac_address': None},
                '5A6591': {       'dev_macs': ['78:d2:94:ad:58:1e'],
                                  'ip_address': None,
                                  'name': '5177 Line 49 Gadshill (John Mueller '
                                          '10400461)',
                                  'serial': '5A6591',
                                  'serial_number': '5A6591',
                                  'system_mac_address': None},
                '5A65BE': {       'dev_macs': ['60:22:32:43:8e:43'],
                                  'ip_address': None,
                                  'name': '4895 Line 67 Barn (Ron & Deb Riddell '
                                          '10426917)',
                                  'serial': '5A65BE',
                                  'serial_number': '5A65BE',
                                  'system_mac_address': None},
                '5A65F2': {       'dev_macs': ['14:21:03:92:ee:1d'],
                                  'ip_address': None,
                                  'name': '4910 Line 48 Gadshill (Dan Groenstege '
                                          '10421236)',
                                  'serial': '5A65F2',
                                  'serial_number': '5A65F2',
                                  'system_mac_address': None},
                '5A65F5': {       'dev_macs': ['44:65:7f:88:d2:28'],
                                  'ip_address': None,
                                  'name': '1 Murray Ct (Dawn Griffin 10435906)',
                                  'serial': '5A65F5',
                                  'serial_number': '5A65F5',
                                  'system_mac_address': None},
                '5A6607': {       'dev_macs': ['78:54:2e:a4:13:a9'],
                                  'ip_address': None,
                                  'name': '5368 Perth Rd 135B Brunner (Glen Diehl '
                                          '10400698)',
                                  'serial': '5A6607',
                                  'serial_number': '5A6607',
                                  'system_mac_address': None},
                '5A6616': {       'dev_macs': ['14:dd:a9:91:e4:90'],
                                  'ip_address': None,
                                  'name': '4860 Line 48 Gadshill (Bert Groenestege '
                                          '10400636)',
                                  'serial': '5A6616',
                                  'serial_number': '5A6616',
                                  'system_mac_address': None},
                '5A6627': {       'dev_macs': ['44:65:7f:88:d4:d4'],
                                  'ip_address': None,
                                  'name': '87 Min St N Milverton (George Raycraft '
                                          '10400878)',
                                  'serial': '5A6627',
                                  'serial_number': '5A6627',
                                  'system_mac_address': None},
                '5A662D': {       'dev_macs': ['1c:61:b4:5c:5c:94'],
                                  'ip_address': None,
                                  'name': '40 Tye St Milverton (B & P Huber 10403559)',
                                  'serial': '5A662D',
                                  'serial_number': '5A662D',
                                  'system_mac_address': None},
                '5A6632': {       'dev_macs': ['d8:07:b6:74:04:e1'],
                                  'ip_address': None,
                                  'name': '5909 PERTH ROAD 119 (FRIESEN KEVIN)',
                                  'serial': '5A6632',
                                  'serial_number': '5A6632',
                                  'system_mac_address': None},
                '5A6657': {       'dev_macs': ['b0:4e:26:7b:bf:a3'],
                                  'ip_address': None,
                                  'name': '6 Fulton St Milverton (P & M Bueckert '
                                          '10421013)',
                                  'serial': '5A6657',
                                  'serial_number': '5A6657',
                                  'system_mac_address': None},
                '5A6661': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4672 Line 49 Brunner (Stan Debus 10400507)',
                                  'serial': '5A6661',
                                  'serial_number': '5A6661',
                                  'system_mac_address': None},
                '5A6663': {       'dev_macs': ['a8:63:7d:7a:80:cb'],
                                  'ip_address': None,
                                  'name': '198 Queen St Monkton (Peter Enns 10431166)',
                                  'serial': '5A6663',
                                  'serial_number': '5A6663',
                                  'system_mac_address': None},
                '5A6665': {       'dev_macs': ['b8:62:1f:51:e9:39'],
                                  'ip_address': None,
                                  'name': '5087 Line 58 Milverton (Dale Zehr 10401364)',
                                  'serial': '5A6665',
                                  'serial_number': '5A6665',
                                  'system_mac_address': None},
                '5A6668': {       'dev_macs': ['2c:fd:a1:c2:80:48'],
                                  'ip_address': None,
                                  'name': '4911 Line 52 Brunner (M & L Brodhagen '
                                          '10400659)',
                                  'serial': '5A6668',
                                  'serial_number': '5A6668',
                                  'system_mac_address': None},
                '5A668C': {       'dev_macs': ['98:da:c4:63:4b:2a'],
                                  'ip_address': None,
                                  'name': '12 Main St N Milverton (Wicked Snips '
                                          '10402369)',
                                  'serial': '5A668C',
                                  'serial_number': '5A668C',
                                  'system_mac_address': None},
                '5A66BA': {       'dev_macs': ['3c:37:86:43:b0:d8'],
                                  'ip_address': None,
                                  'name': '6475 Perh Rd. 131 Milverton (Megan Ray)',
                                  'serial': '5A66BA',
                                  'serial_number': '5A66BA',
                                  'system_mac_address': None},
                '5A66C8': {       'dev_macs': ['08:36:c9:ed:65:1d'],
                                  'ip_address': None,
                                  'name': '4595 Line 49 (PEHLKE MARJORIE)',
                                  'serial': '5A66C8',
                                  'serial_number': '5A66C8',
                                  'system_mac_address': None},
                '5A69C9': {       'dev_macs': ['34:97:f6:a7:1a:ec'],
                                  'ip_address': None,
                                  'name': '4677 PERTH LINE 55 (KUNTZE MURRAY)',
                                  'serial': '5A69C9',
                                  'serial_number': '5A69C9',
                                  'system_mac_address': None},
                '5A6A13': {       'dev_macs': ['04:d9:f5:8c:e9:e0'],
                                  'ip_address': None,
                                  'name': '4723 LINE 52 (ZONDERVAN TONY & CYNTHIA)',
                                  'serial': '5A6A13',
                                  'serial_number': '5A6A13',
                                  'system_mac_address': None},
                '5A6A60': {       'dev_macs': ['60:db:98:09:3a:e9'],
                                  'ip_address': None,
                                  'name': '5735 PERTH ROAD 119 (NAFZIGER JAMES)',
                                  'serial': '5A6A60',
                                  'serial_number': '5A6A60',
                                  'system_mac_address': None},
                '5A6AEE': {       'dev_macs': ['44:65:7f:88:37:15'],
                                  'ip_address': None,
                                  'name': '5358 Road 140 Brunner (Martin Brodhagen '
                                          '10401375)',
                                  'serial': '5A6AEE',
                                  'serial_number': '5A6AEE',
                                  'system_mac_address': None},
                '5A6B63': {       'dev_macs': ['7a:23:3f:12:f3:92'],
                                  'ip_address': None,
                                  'name': '4 WEST ST (OJIGBO RUTE)',
                                  'serial': '5A6B63',
                                  'serial_number': '5A6B63',
                                  'system_mac_address': None},
                '5A6B6F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '67 MAIN ST (BOOSE ERIC)',
                                  'serial': '5A6B6F',
                                  'serial_number': '5A6B6F',
                                  'system_mac_address': None},
                '5A6B80': {       'dev_macs': ['c0:a0:bb:e7:a9:40'],
                                  'ip_address': None,
                                  'name': '5280 LINE 49 (YOUNGBLUT BILL & SARAH)',
                                  'serial': '5A6B80',
                                  'serial_number': '5A6B80',
                                  'system_mac_address': None},
                '5A6B8F': {       'dev_macs': ['3c:52:a1:96:ba:ed'],
                                  'ip_address': None,
                                  'name': '5133 Rd 140 (Paul and Carolyn Visser)',
                                  'serial': '5A6B8F',
                                  'serial_number': '5A6B8F',
                                  'system_mac_address': None},
                '5A6B93': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6314 ROAD 124 (POOLE VICTOR)',
                                  'serial': '5A6B93',
                                  'serial_number': '5A6B93',
                                  'system_mac_address': None},
                '5A6B9A': {       'dev_macs': ['24:2f:d0:d5:3c:61'],
                                  'ip_address': None,
                                  'name': '5088 Line 52 Brunner (Jim Van Herk '
                                          '10400809)',
                                  'serial': '5A6B9A',
                                  'serial_number': '5A6B9A',
                                  'system_mac_address': None},
                '5A6BA6': {       'dev_macs': ['00:ad:24:60:e2:2b'],
                                  'ip_address': None,
                                  'name': '5 West St. Milverton (Michael Gordon)',
                                  'serial': '5A6BA6',
                                  'serial_number': '5A6BA6',
                                  'system_mac_address': None},
                '5A6BAE': {       'dev_macs': ['c0:c9:e3:09:df:e5'],
                                  'ip_address': None,
                                  'name': '23 William St Milverton (C & H Lichti '
                                          '10424574)',
                                  'serial': '5A6BAE',
                                  'serial_number': '5A6BAE',
                                  'system_mac_address': None},
                '5A6BB2': {       'dev_macs': ['10:27:f5:63:45:ad'],
                                  'ip_address': None,
                                  'name': '5173 Road 150 Gadshill (NATYWARY TYLER)',
                                  'serial': '5A6BB2',
                                  'serial_number': '5A6BB2',
                                  'system_mac_address': None},
                '5A6BCD': {       'dev_macs': ['3c:84:6a:57:6c:f5'],
                                  'ip_address': None,
                                  'name': '5020 Line 52 Brunner (B & J Natywary '
                                          '10400723)',
                                  'serial': '5A6BCD',
                                  'serial_number': '5A6BCD',
                                  'system_mac_address': None},
                '5A6BF8': {       'dev_macs': ['24:f5:a2:8d:fb:8e'],
                                  'ip_address': None,
                                  'name': '14 WEST ST (SALZMAN BRUCE)',
                                  'serial': '5A6BF8',
                                  'serial_number': '5A6BF8',
                                  'system_mac_address': None},
                '5A6C39': {       'dev_macs': ['48:77:46:9a:50:9a'],
                                  'ip_address': None,
                                  'name': '22 WEST ST (STECKLY STEVE)',
                                  'serial': '5A6C39',
                                  'serial_number': '5A6C39',
                                  'system_mac_address': None},
                '5A6C68': {       'dev_macs': ['98:da:c4:30:8b:40'],
                                  'ip_address': None,
                                  'name': '6515 ROAD 154 (ERB MARK & RACHEL)',
                                  'serial': '5A6C68',
                                  'serial_number': '5A6C68',
                                  'system_mac_address': None},
                '5A6C89': {       'dev_macs': ['78:32:1b:63:f0:6a'],
                                  'ip_address': None,
                                  'name': '4575 Line 49 (Roosendale Theo)',
                                  'serial': '5A6C89',
                                  'serial_number': '5A6C89',
                                  'system_mac_address': None},
                '5A6C8A': {       'dev_macs': ['a8:63:7d:4f:07:30'],
                                  'ip_address': None,
                                  'name': '5357 ROAD 140 (ROOBROECK JEREMY)',
                                  'serial': '5A6C8A',
                                  'serial_number': '5A6C8A',
                                  'system_mac_address': None},
                '5A6C9D': {       'dev_macs': ['d0:76:8f:02:f4:f0'],
                                  'ip_address': None,
                                  'name': '24 MILL ST E (WEBER JUSTIN)',
                                  'serial': '5A6C9D',
                                  'serial_number': '5A6C9D',
                                  'system_mac_address': None},
                '64079A': {       'dev_macs': ['8c:90:2d:72:7d:85'],
                                  'ip_address': None,
                                  'name': '5154 Road 140 (Joseph White 10435270)',
                                  'serial': '64079A',
                                  'serial_number': '64079A',
                                  'system_mac_address': None},
                '640A0A': {       'dev_macs': ['24:f5:a2:00:96:a6'],
                                  'ip_address': None,
                                  'name': '29 William St (Addition) (Lorilee Streicher '
                                          '10401618)',
                                  'serial': '640A0A',
                                  'serial_number': '640A0A',
                                  'system_mac_address': None},
                '640A76': {       'dev_macs': ['c4:41:1e:77:6f:d0'],
                                  'ip_address': None,
                                  'name': '91 Mill St E Milverton (Carey Soehner '
                                          '10431225)',
                                  'serial': '640A76',
                                  'serial_number': '640A76',
                                  'system_mac_address': None},
                '640AB6': {       'dev_macs': ['14:59:c0:c1:50:28'],
                                  'ip_address': None,
                                  'name': '5 CObalt St (Matthew Yost 10436053)',
                                  'serial': '640AB6',
                                  'serial_number': '640AB6',
                                  'system_mac_address': None},
                '640ADD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '13 Main St N Apt 5 Milverton',
                                  'serial': '640ADD',
                                  'serial_number': '640ADD',
                                  'system_mac_address': None},
                '640AF8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 Main St N Ste 2 Milverton',
                                  'serial': '640AF8',
                                  'serial_number': '640AF8',
                                  'system_mac_address': None},
                '640AFB': {       'dev_macs': ['44:65:7f:38:35:e3'],
                                  'ip_address': None,
                                  'name': '94 REAGAN ST (DEWAR MELANIE)',
                                  'serial': '640AFB',
                                  'serial_number': '640AFB',
                                  'system_mac_address': None},
                '640AFF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 MAIN ST S',
                                  'serial': '640AFF',
                                  'serial_number': '640AFF',
                                  'system_mac_address': None},
                '640B2D': {       'dev_macs': ['c0:56:27:68:f2:88'],
                                  'ip_address': None,
                                  'name': '5375 Rd 144 Brunner (Lisa Paxton 10432593)',
                                  'serial': '640B2D',
                                  'serial_number': '640B2D',
                                  'system_mac_address': None},
                '640B35': {       'dev_macs': [       '00:0b:82:ed:dc:35',
                                                      '00:18:39:73:29:7c',
                                                      '1c:61:b4:03:cd:d5'],
                                  'ip_address': None,
                                  'name': '150 Brock St Monkton (Ryan Buchanan '
                                          '10430196)',
                                  'serial': '640B35',
                                  'serial_number': '640B35',
                                  'system_mac_address': None},
                '640B3B': {       'dev_macs': ['f8:7b:8c:14:90:8f'],
                                  'ip_address': None,
                                  'name': '69 Reagan St Milverton (Richard Battrick, '
                                          '10436036)',
                                  'serial': '640B3B',
                                  'serial_number': '640B3B',
                                  'system_mac_address': None},
                '640B46': {       'dev_macs': ['a0:04:60:ca:69:e9'],
                                  'ip_address': None,
                                  'name': '46 Mill St E (Kyle Woodrow 10430299)',
                                  'serial': '640B46',
                                  'serial_number': '640B46',
                                  'system_mac_address': None},
                '640B4A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6 Main St N',
                                  'serial': '640B4A',
                                  'serial_number': '640B4A',
                                  'system_mac_address': None},
                '640B51': {       'dev_macs': ['44:65:7f:89:0c:96'],
                                  'ip_address': None,
                                  'name': '87 Reagan St. Milverton (Emily Mills '
                                          '10435602)',
                                  'serial': '640B51',
                                  'serial_number': '640B51',
                                  'system_mac_address': None},
                '640B67': {       'dev_macs': ['cc:be:59:c3:48:ef'],
                                  'ip_address': None,
                                  'name': '2 Coulter St Milverton (Lynn Giles '
                                          '10434418)',
                                  'serial': '640B67',
                                  'serial_number': '640B67',
                                  'system_mac_address': None},
                '66D53': {       'dev_macs': ['00:09:0f:09:01:02'],
                                 'ip_address': None,
                                 'name': '25 WHALEY AVE MILVERTON (BOSHART INDUSTRIES)',
                                 'serial': '66D53',
                                 'serial_number': '66D53',
                                 'system_mac_address': None},
                '71A4CC': {       'dev_macs': ['14:21:03:92:f1:35'],
                                  'ip_address': None,
                                  'name': '5223 Line 60 (Keith Reibeling 10400225)',
                                  'serial': '71A4CC',
                                  'serial_number': '71A4CC',
                                  'system_mac_address': None},
                '71A4CF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5161 Line 60 (Jeanette Young 10431137)',
                                  'serial': '71A4CF',
                                  'serial_number': '71A4CF',
                                  'system_mac_address': None},
                '71A4F6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '186 Selby Ave. Monkton (YUNDT JUD)',
                                  'serial': '71A4F6',
                                  'serial_number': '71A4F6',
                                  'system_mac_address': None},
                '71A51A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '124 Maddison St. W. Monkton (Walsh Truck '
                                          '10430198)',
                                  'serial': '71A51A',
                                  'serial_number': '71A51A',
                                  'system_mac_address': None},
                '71D7AE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '159 Market St. Monkton (Delford Kuepfer)',
                                  'serial': '71D7AE',
                                  'serial_number': '71D7AE',
                                  'system_mac_address': None},
                '71D894': {       'dev_macs': [       '00:03:e6:db:4c:3c',
                                                      '00:03:e6:db:b7:34'],
                                  'ip_address': None,
                                  'name': '142 Queen St. Line 57 Monkton (Connie Bull)',
                                  'serial': '71D894',
                                  'serial_number': '71D894',
                                  'system_mac_address': None},
                '722207': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '168 Winstanley St. Monkton (Don Buchanan)',
                                  'serial': '722207',
                                  'serial_number': '722207',
                                  'system_mac_address': None},
                '7233EA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '20 Schade St. Monkton (Ruby Rennick)',
                                  'serial': '7233EA',
                                  'serial_number': '7233EA',
                                  'system_mac_address': None},
                '724AD0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6208 Rd. 164 Monkton (Ginny White)',
                                  'serial': '724AD0',
                                  'serial_number': '724AD0',
                                  'system_mac_address': None},
                '757AD': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '10A Temperance St. Milverton (Julie '
                                         'Gallinger)',
                                 'serial': '757AD',
                                 'serial_number': '757AD',
                                 'system_mac_address': None},
                '75C1DC': {       'dev_macs': ['18:fd:74:0b:c4:d1'],
                                  'ip_address': None,
                                  'name': '93 Main St. N. Milverton (Milverton Vet '
                                          'Clinic 10400849)',
                                  'serial': '75C1DC',
                                  'serial_number': '75C1DC',
                                  'system_mac_address': None},
                '769C1C': {       'dev_macs': ['60:db:98:81:fd:0c'],
                                  'ip_address': None,
                                  'name': '89 Schade St., Monkton (Gord &Helen Young)',
                                  'serial': '769C1C',
                                  'serial_number': '769C1C',
                                  'system_mac_address': None},
                '76B064': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '205 King Ave, Monkton (Mervin Subject)',
                                  'serial': '76B064',
                                  'serial_number': '76B064',
                                  'system_mac_address': None},
                '76D052': {       'dev_macs': ['d0:76:8f:02:f1:1d'],
                                  'ip_address': None,
                                  'name': '86 REAGAN ST (YANTZI MARCUS)',
                                  'serial': '76D052',
                                  'serial_number': '76D052',
                                  'system_mac_address': None},
                '76EDCD': {       'dev_macs': ['d0:76:8f:02:f5:27'],
                                  'ip_address': None,
                                  'name': '189 QUEEN AVENUE MONKTON (MORRISON COLTON)',
                                  'serial': '76EDCD',
                                  'serial_number': '76EDCD',
                                  'system_mac_address': None},
                '7E0E9': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '47 Schade St. Monkton (Fay Baker)',
                                 'serial': '7E0E9',
                                 'serial_number': '7E0E9',
                                 'system_mac_address': None},
                '7E0ED': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '51 SCHADE ST (JAKOBSEN SOREN & LEAH)',
                                 'serial': '7E0ED',
                                 'serial_number': '7E0ED',
                                 'system_mac_address': None},
                '7E210': {       'dev_macs': ['d0:76:8f:02:f6:0e'],
                                 'ip_address': None,
                                 'name': '125 BROCK STREET MONKTON (QUACKENBUSH PARKER '
                                         '10436791)',
                                 'serial': '7E210',
                                 'serial_number': '7E210',
                                 'system_mac_address': None},
                '800A9': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '221 Winstanley St. Monkton',
                                 'serial': '800A9',
                                 'serial_number': '800A9',
                                 'system_mac_address': None},
                '82A532': {       'dev_macs': ['d0:76:8f:02:f5:8a'],
                                  'ip_address': None,
                                  'name': '15 JO-LENE CT (HALE RICK)',
                                  'serial': '82A532',
                                  'serial_number': '82A532',
                                  'system_mac_address': None},
                '82A696': {       'dev_macs': ['d0:76:8f:02:f3:b1'],
                                  'ip_address': None,
                                  'name': '66 GERBER DR (SCHEERER JEFF & KAREN)',
                                  'serial': '82A696',
                                  'serial_number': '82A696',
                                  'system_mac_address': None},
                '82A6CB': {       'dev_macs': ['ec:4f:82:a4:92:12'],
                                  'ip_address': None,
                                  'name': '23 WINSTANLEY ST (HEIBEIN SHERRY)',
                                  'serial': '82A6CB',
                                  'serial_number': '82A6CB',
                                  'system_mac_address': None},
                '82A741': {       'dev_macs': ['f4:e2:c6:f2:3a:e9'],
                                  'ip_address': None,
                                  'name': '5960 PERTH RD 119 (715276 ONTARIO INC)',
                                  'serial': '82A741',
                                  'serial_number': '82A741',
                                  'system_mac_address': None},
                '82A750': {       'dev_macs': ['d0:76:8f:02:f3:22'],
                                  'ip_address': None,
                                  'name': '88 MILL ST E (GASCHO BRIAN & LIZ)',
                                  'serial': '82A750',
                                  'serial_number': '82A750',
                                  'system_mac_address': None},
                '844BBE': {       'dev_macs': ['60:db:98:55:bb:2f'],
                                  'ip_address': None,
                                  'name': '20 WEST ST (SCHNEIDER EVAN)',
                                  'serial': '844BBE',
                                  'serial_number': '844BBE',
                                  'system_mac_address': None},
                '844BC7': {       'dev_macs': ['d0:76:8f:02:f0:c5'],
                                  'ip_address': None,
                                  'name': '5163 ROAD 140 (WHITE JOSEPH)',
                                  'serial': '844BC7',
                                  'serial_number': '844BC7',
                                  'system_mac_address': None},
                '844BD6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6466 PERTH ROAD 131 (WALL FRANK)',
                                  'serial': '844BD6',
                                  'serial_number': '844BD6',
                                  'system_mac_address': None},
                '844BEE': {       'dev_macs': ['24:f5:a2:a3:14:e0'],
                                  'ip_address': None,
                                  'name': '5136 LINE 49 (BAUER ELONA)',
                                  'serial': '844BEE',
                                  'serial_number': '844BEE',
                                  'system_mac_address': None},
                '844BEF': {       'dev_macs': ['cc:f4:11:60:9e:a9'],
                                  'ip_address': None,
                                  'name': '83 REAGAN ST (RUAN ZHIHENG)',
                                  'serial': '844BEF',
                                  'serial_number': '844BEF',
                                  'system_mac_address': None},
                '844BF5': {       'dev_macs': ['d0:76:8f:02:ef:bd'],
                                  'ip_address': None,
                                  'name': '7 MURRAY CT (JOSE NAJMA)',
                                  'serial': '844BF5',
                                  'serial_number': '844BF5',
                                  'system_mac_address': None},
                '845909': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '211 SELBY AV (HENRY NATHAN)',
                                  'serial': '845909',
                                  'serial_number': '845909',
                                  'system_mac_address': None},
                '845BB4': {       'dev_macs': [       '5c:0f:fb:17:68:76',
                                                      '5c:0f:fb:17:72:46'],
                                  'ip_address': None,
                                  'name': '193 QUEEN AVE (EARL JEFF & SUE)',
                                  'serial': '845BB4',
                                  'serial_number': '845BB4',
                                  'system_mac_address': None},
                '8678A2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 MAIN ST S (KUEPFER RACHEL)',
                                  'serial': '8678A2',
                                  'serial_number': '8678A2',
                                  'system_mac_address': None},
                '869226': {       'dev_macs': ['d0:76:8f:02:ee:9f'],
                                  'ip_address': None,
                                  'name': '65 GERBER DR (NORRIS DEANNE)',
                                  'serial': '869226',
                                  'serial_number': '869226',
                                  'system_mac_address': None},
                '869271': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5317 ROAD 144 (GRAUL RODNEY)',
                                  'serial': '869271',
                                  'serial_number': '869271',
                                  'system_mac_address': None},
                '869376': {       'dev_macs': ['d0:76:8f:1f:58:c7'],
                                  'ip_address': None,
                                  'name': '85 GERBER DR (SNYDER ASHLEY)',
                                  'serial': '869376',
                                  'serial_number': '869376',
                                  'system_mac_address': None},
                '869384': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5922 SOMERSET DR (WAGLER CHRISTOPHER & '
                                          'ROSALIND)',
                                  'serial': '869384',
                                  'serial_number': '869384',
                                  'system_mac_address': None},
                '869385': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '208 Selby Av (Stephanie Priestap 10436138)',
                                  'serial': '869385',
                                  'serial_number': '869385',
                                  'system_mac_address': None},
                '86938A': {       'dev_macs': ['d0:76:8f:1f:5a:69'],
                                  'ip_address': None,
                                  'name': '21 MURRAY CT (ARMSTRONG SAMANTHA)',
                                  'serial': '86938A',
                                  'serial_number': '86938A',
                                  'system_mac_address': None},
                '8693A0': {       'dev_macs': ['d0:76:8f:1f:63:38'],
                                  'ip_address': None,
                                  'name': '162 Winstanley St. Monkton (Jessie '
                                          'Buchanan)',
                                  'serial': '8693A0',
                                  'serial_number': '8693A0',
                                  'system_mac_address': None},
                '8693A1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6250 PERTH ROAD 131 (RAYCRAFT KEN & '
                                          'DARLENE)',
                                  'serial': '8693A1',
                                  'serial_number': '8693A1',
                                  'system_mac_address': None},
                '8693A5': {       'dev_macs': ['60:db:98:09:1c:67'],
                                  'ip_address': None,
                                  'name': '5597 DUKE ST (HIGGS TERRY)',
                                  'serial': '8693A5',
                                  'serial_number': '8693A5',
                                  'system_mac_address': None},
                '8693A7': {       'dev_macs': ['cc:be:59:2f:2d:12'],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (CONLAN MANDY)',
                                  'serial': '8693A7',
                                  'serial_number': '8693A7',
                                  'system_mac_address': None},
                '8693AC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '10400848 (GERBER STAN & ERMA)',
                                  'serial': '8693AC',
                                  'serial_number': '8693AC',
                                  'system_mac_address': None},
                '8693AF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6146 ROAD 136 (KUEPFER GERALD & DARLENE)',
                                  'serial': '8693AF',
                                  'serial_number': '8693AF',
                                  'system_mac_address': None},
                '8693B2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5931 SOMERSET DR (LICHTY HOWARD)',
                                  'serial': '8693B2',
                                  'serial_number': '8693B2',
                                  'system_mac_address': None},
                '8693B5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5940 PERTH ROAD 119 (FRIJTERS MARIETTE)',
                                  'serial': '8693B5',
                                  'serial_number': '8693B5',
                                  'system_mac_address': None},
                '8693B9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5923 SOMERSET DR (NAFZIGER WAYNE)',
                                  'serial': '8693B9',
                                  'serial_number': '8693B9',
                                  'system_mac_address': None},
                '8693BC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4540 LINE 61 (BUCHAN SCOTT)',
                                  'serial': '8693BC',
                                  'serial_number': '8693BC',
                                  'system_mac_address': None},
                '8693BE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4994 PERTH LINE 55 (ZEHR SUSAN)',
                                  'serial': '8693BE',
                                  'serial_number': '8693BE',
                                  'system_mac_address': None},
                '8693C0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5592 ROAD 126 (KOBUS DAN & BECCA)',
                                  'serial': '8693C0',
                                  'serial_number': '8693C0',
                                  'system_mac_address': None},
                '8693C3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'Beside the Still Water',
                                  'serial': '8693C3',
                                  'serial_number': '8693C3',
                                  'system_mac_address': None},
                '8693C5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'THOMAS DAHM (1-3 MADDISON ST E MONKTON)',
                                  'serial': '8693C5',
                                  'serial_number': '8693C5',
                                  'system_mac_address': None},
                '8693C7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5940 PERTH ROAD 119 (HUBER RACHEL)',
                                  'serial': '8693C7',
                                  'serial_number': '8693C7',
                                  'system_mac_address': None},
                '8693CC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '7 Cruikshank St. Milverton',
                                  'serial': '8693CC',
                                  'serial_number': '8693CC',
                                  'system_mac_address': None},
                '8693CE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4868 LINE 49 (DEGROOT TY)',
                                  'serial': '8693CE',
                                  'serial_number': '8693CE',
                                  'system_mac_address': None},
                '8693D0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6112 ANA ST (CRESSMAN STEVE & JUDY)',
                                  'serial': '8693D0',
                                  'serial_number': '8693D0',
                                  'system_mac_address': None},
                '8693D2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4861 PERTH LINE 55 (ZOOK EPHRAIM)',
                                  'serial': '8693D2',
                                  'serial_number': '8693D2',
                                  'system_mac_address': None},
                '8693D8': {       'dev_macs': ['d0:76:8f:02:ef:a7'],
                                  'ip_address': None,
                                  'name': '5438 Line 60 (CARTER MIKE & BEV)',
                                  'serial': '8693D8',
                                  'serial_number': '8693D8',
                                  'system_mac_address': None},
                '8693D9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5934 PERTH ROAD 119 (YANTZI RHODA)',
                                  'serial': '8693D9',
                                  'serial_number': '8693D9',
                                  'system_mac_address': None},
                '8693DA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5922 PERTH ROAD 119 (NAFZIGER TOM & ALYCIA)',
                                  'serial': '8693DA',
                                  'serial_number': '8693DA',
                                  'system_mac_address': None},
                '8693DE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4977 PERTH LINE 55 (MOGK BROTHERS)',
                                  'serial': '8693DE',
                                  'serial_number': '8693DE',
                                  'system_mac_address': None},
                '8693E0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6127 ANA ST (NAFZIGER KEN)',
                                  'serial': '8693E0',
                                  'serial_number': '8693E0',
                                  'system_mac_address': None},
                '8693E1': {       'dev_macs': [       'cc:be:59:c3:52:9a',
                                                      '04:42:1a:7a:b7:88',
                                                      '38:c0:ea:44:be:46'],
                                  'ip_address': None,
                                  'name': '209 NELSON STREET (MUNICIPALITY OF NORTH '
                                          'PERTH)',
                                  'serial': '8693E1',
                                  'serial_number': '8693E1',
                                  'system_mac_address': None},
                '8693E3': {       'dev_macs': ['b8:94:70:f3:e9:26'],
                                  'ip_address': None,
                                  'name': '5940 PERTH ROAD 119 (BANNON JOANNE)',
                                  'serial': '8693E3',
                                  'serial_number': '8693E3',
                                  'system_mac_address': None},
                '8693E4': {       'dev_macs': ['d0:76:8f:02:f4:da'],
                                  'ip_address': None,
                                  'name': '75 REAGAN ST (MILLER PATRICK)',
                                  'serial': '8693E4',
                                  'serial_number': '8693E4',
                                  'system_mac_address': None},
                '8693EE': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6212 PERTH ROAD 131 (MCCORMACK KELLY)',
                                  'serial': '8693EE',
                                  'serial_number': '8693EE',
                                  'system_mac_address': None},
                '8693F1': {       'dev_macs': ['d0:76:8f:1f:65:27'],
                                  'ip_address': None,
                                  'name': '5927 SOMERSET DR (HALLAHAN JAMIE & LAURA)',
                                  'serial': '8693F1',
                                  'serial_number': '8693F1',
                                  'system_mac_address': None},
                '8693F4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4731 PERTH LINE 55 (SLITS PEDRO & JOLANDA)',
                                  'serial': '8693F4',
                                  'serial_number': '8693F4',
                                  'system_mac_address': None},
                '8693F5': {       'dev_macs': ['60:db:98:55:ba:c1'],
                                  'ip_address': None,
                                  'name': '6 West St. Milverton (MCLAUGHLIN CAMERON)',
                                  'serial': '8693F5',
                                  'serial_number': '8693F5',
                                  'system_mac_address': None},
                '8693F9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4728 PERTH LINE 55 (KUNTZE HOWARD)',
                                  'serial': '8693F9',
                                  'serial_number': '8693F9',
                                  'system_mac_address': None},
                '8693FA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '89 REAGAN ST MILVERTON (VINK JASON)',
                                  'serial': '8693FA',
                                  'serial_number': '8693FA',
                                  'system_mac_address': None},
                '8693FB': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6108 ANA ST (DUNN ROBIN & WENDY)',
                                  'serial': '8693FB',
                                  'serial_number': '8693FB',
                                  'system_mac_address': None},
                '8693FD': {       'dev_macs': ['ec:4f:82:ac:ed:c9'],
                                  'ip_address': None,
                                  'name': '81 GERBER DR (MARSHALL GLEN)',
                                  'serial': '8693FD',
                                  'serial_number': '8693FD',
                                  'system_mac_address': None},
                '869400': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4631 PERTH LINE 55 (ROPP TITUS)',
                                  'serial': '869400',
                                  'serial_number': '869400',
                                  'system_mac_address': None},
                '869402': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4561 LINE 61 (LUPTON MACKENZIE (MAC))',
                                  'serial': '869402',
                                  'serial_number': '869402',
                                  'system_mac_address': None},
                '869404': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5916 PERTH ROAD 119 (ROTH DERRICK)',
                                  'serial': '869404',
                                  'serial_number': '869404',
                                  'system_mac_address': None},
                '869407': {       'dev_macs': ['cc:be:59:32:95:5e'],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (GRAVES ANDREA)',
                                  'serial': '869407',
                                  'serial_number': '869407',
                                  'system_mac_address': None},
                '869408': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '154 SMITH ST (GINGERICH BRAEDEN)',
                                  'serial': '869408',
                                  'serial_number': '869408',
                                  'system_mac_address': None},
                '869409': {       'dev_macs': ['60:db:98:09:36:c9'],
                                  'ip_address': None,
                                  'name': '4509 LINE 61 (KUEPFER ALVIN)',
                                  'serial': '869409',
                                  'serial_number': '869409',
                                  'system_mac_address': None},
                '86940A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4821 PERTH LINE 55 (LOOSER CHRIS)',
                                  'serial': '86940A',
                                  'serial_number': '86940A',
                                  'system_mac_address': None},
                '86940E': {       'dev_macs': ['14:21:03:a9:72:8e'],
                                  'ip_address': None,
                                  'name': '5926 SOMERSET DR (HARRIS  STEVE)',
                                  'serial': '86940E',
                                  'serial_number': '86940E',
                                  'system_mac_address': None},
                '86940F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '12 WILLIAM ST',
                                  'serial': '86940F',
                                  'serial_number': '86940F',
                                  'system_mac_address': None},
                '869410': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5375 Road 144 Brunner (W & J Paxton '
                                          '10400532)',
                                  'serial': '869410',
                                  'serial_number': '869410',
                                  'system_mac_address': None},
                '869415': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5241 LINE 58 (BANNON NIKKI)',
                                  'serial': '869415',
                                  'serial_number': '869415',
                                  'system_mac_address': None},
                '869418': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '36 MAIN ST S (JOHNSON JAMIE & HEIDI)',
                                  'serial': '869418',
                                  'serial_number': '869418',
                                  'system_mac_address': None},
                '86941A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '70 GERBER DR (DARASCH SAMANTHA)',
                                  'serial': '86941A',
                                  'serial_number': '86941A',
                                  'system_mac_address': None},
                '869424': {       'dev_macs': ['60:db:98:09:1c:eb'],
                                  'ip_address': None,
                                  'name': '184 SELBY AV (BOYD DON)',
                                  'serial': '869424',
                                  'serial_number': '869424',
                                  'system_mac_address': None},
                '869427': {       'dev_macs': [       '00:03:e6:dd:46:4a',
                                                      '00:03:e6:eb:4b:14',
                                                      'd0:76:8f:02:ee:10'],
                                  'ip_address': None,
                                  'name': '79 REAGAN ST (ALVAREZ JOE)',
                                  'serial': '869427',
                                  'serial_number': '869427',
                                  'system_mac_address': None},
                '86942F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6109 ANA ST (WALL GEORGE (ESTATE OF))',
                                  'serial': '86942F',
                                  'serial_number': '86942F',
                                  'system_mac_address': None},
                '869431': {       'dev_macs': ['d0:76:8f:1f:65:8a'],
                                  'ip_address': None,
                                  'name': '90 REAGAN ST (GRAHAM DAVE)',
                                  'serial': '869431',
                                  'serial_number': '869431',
                                  'system_mac_address': None},
                '869432': {       'dev_macs': ['cc:be:59:2d:a3:be'],
                                  'ip_address': None,
                                  'name': '50 MAIN ST N (CAMERON JEFF)',
                                  'serial': '869432',
                                  'serial_number': '869432',
                                  'system_mac_address': None},
                '86946D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6137 PERTH ROAD 131 (HADDOW BRIAN & REID)',
                                  'serial': '86946D',
                                  'serial_number': '86946D',
                                  'system_mac_address': None},
                '869473': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5929 SOMERSET DR (GERBER TOM & SHARON)',
                                  'serial': '869473',
                                  'serial_number': '869473',
                                  'system_mac_address': None},
                '869474': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4643 PERTH LINE 55 (REIS COURTNEY)',
                                  'serial': '869474',
                                  'serial_number': '869474',
                                  'system_mac_address': None},
                '869484': {       'dev_macs': ['d0:76:8f:02:ee:e1'],
                                  'ip_address': None,
                                  'name': '9 MURRAY CT (GREER SANDRA)',
                                  'serial': '869484',
                                  'serial_number': '869484',
                                  'system_mac_address': None},
                '86948C': {       'dev_macs': ['60:db:98:09:37:e7'],
                                  'ip_address': None,
                                  'name': '93 Reagan St Milverton (VANT VOORT M & C)',
                                  'serial': '86948C',
                                  'serial_number': '86948C',
                                  'system_mac_address': None},
                '869492': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4509 LINE 61 (LEGACY STONE INC)',
                                  'serial': '869492',
                                  'serial_number': '869492',
                                  'system_mac_address': None},
                '8704F': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '71 Mill St. E. Milverton (Valerie Yost '
                                         '10430631)',
                                 'serial': '8704F',
                                 'serial_number': '8704F',
                                 'system_mac_address': None},
                '87241E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '15 Main St. N. Milverton (Brenda Robertson)',
                                  'serial': '87241E',
                                  'serial_number': '87241E',
                                  'system_mac_address': None},
                '872420': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': "4679 LINE 61 (GJ'S HARVEST CENTRE INC.)",
                                  'serial': '872420',
                                  'serial_number': '872420',
                                  'system_mac_address': None},
                '872431': {       'dev_macs': ['60:db:98:09:37:b0'],
                                  'ip_address': None,
                                  'name': '15 MURRAY CT (BOWMAN JOHN)',
                                  'serial': '872431',
                                  'serial_number': '872431',
                                  'system_mac_address': None},
                '872432': {       'dev_macs': ['ec:4f:82:a9:a8:81'],
                                  'ip_address': None,
                                  'name': '5227 Line 60 (Mike Schultz 10421360)',
                                  'serial': '872432',
                                  'serial_number': '872432',
                                  'system_mac_address': None},
                '872435': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6104 RD 140 (DYCK PETER)',
                                  'serial': '872435',
                                  'serial_number': '872435',
                                  'system_mac_address': None},
                '872445': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 JOHN ST (ALBRECHT MIKE)',
                                  'serial': '872445',
                                  'serial_number': '872445',
                                  'system_mac_address': None},
                '872483': {       'dev_macs': ['f4:92:bf:8d:a7:f8'],
                                  'ip_address': None,
                                  'name': 'PERCON NEW SHOP 10434163 (4679 LINE 61)',
                                  'serial': '872483',
                                  'serial_number': '872483',
                                  'system_mac_address': None},
                '872484': {       'dev_macs': ['60:db:98:0c:d1:09'],
                                  'ip_address': None,
                                  'name': '6113 ANA ST (ERB DARYL)',
                                  'serial': '872484',
                                  'serial_number': '872484',
                                  'system_mac_address': None},
                '8724DE': {       'dev_macs': ['44:65:7f:88:d6:7b'],
                                  'ip_address': None,
                                  'name': '5376 LINE 49 (RIDLEY RYAN)',
                                  'serial': '8724DE',
                                  'serial_number': '8724DE',
                                  'system_mac_address': None},
                '8724EA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '20 MILL ST E (FRESHNUP INC.)',
                                  'serial': '8724EA',
                                  'serial_number': '8724EA',
                                  'system_mac_address': None},
                '8724F3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6168 RD 124 MILVERTON (SCHULTZ LONNIE)',
                                  'serial': '8724F3',
                                  'serial_number': '8724F3',
                                  'system_mac_address': None},
                '872503': {       'dev_macs': ['14:21:03:92:f0:17'],
                                  'ip_address': None,
                                  'name': '7 FULTON ST (KOCH JEREMY)',
                                  'serial': '872503',
                                  'serial_number': '872503',
                                  'system_mac_address': None},
                '872506': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5368 PERTH ROAD 135 (DIEHL BRAD)',
                                  'serial': '872506',
                                  'serial_number': '872506',
                                  'system_mac_address': None},
                '872507': {       'dev_macs': ['04:bc:9f:46:ac:6e'],
                                  'ip_address': None,
                                  'name': '17 MURRAY CT (AUSTIN MICHELLE)',
                                  'serial': '872507',
                                  'serial_number': '872507',
                                  'system_mac_address': None},
                '87250A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5267 ROAD 144 (GRAUL STEVE)',
                                  'serial': '87250A',
                                  'serial_number': '87250A',
                                  'system_mac_address': None},
                '87250E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6283 ROAD 124 (BAUMAN GREG)',
                                  'serial': '87250E',
                                  'serial_number': '87250E',
                                  'system_mac_address': None},
                '87250F': {       'dev_macs': ['ec:4f:82:2d:2f:c8'],
                                  'ip_address': None,
                                  'name': '4 MAIN ST S (BURGER AELAURA)',
                                  'serial': '87250F',
                                  'serial_number': '87250F',
                                  'system_mac_address': None},
                '872514': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '19 MURRAY CT MILVERTON (RIDDELL VANESSA)',
                                  'serial': '872514',
                                  'serial_number': '872514',
                                  'system_mac_address': None},
                '872524': {       'dev_macs': ['34:56:fe:2c:cc:ab'],
                                  'ip_address': None,
                                  'name': '4481 PERTH LINE 72 NEWTON (MITCHELL MILL '
                                          'SYSTEMS)',
                                  'serial': '872524',
                                  'serial_number': '872524',
                                  'system_mac_address': None},
                '872544': {       'dev_macs': ['14:21:03:d1:7d:6d'],
                                  'ip_address': None,
                                  'name': '10 MAIN ST N APT 2 (STREICHER DUSTIN)',
                                  'serial': '872544',
                                  'serial_number': '872544',
                                  'system_mac_address': None},
                '87254B': {       'dev_macs': ['d0:76:8f:1f:54:de'],
                                  'ip_address': None,
                                  'name': '6161 ROAD 124 (JANTZI BEN)',
                                  'serial': '87254B',
                                  'serial_number': '87254B',
                                  'system_mac_address': None},
                '872553': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6202 ROAD 148 (JANTZI SHAWN)',
                                  'serial': '872553',
                                  'serial_number': '872553',
                                  'system_mac_address': None},
                '872558': {       'dev_macs': ['b8:94:70:f3:e4:c4'],
                                  'ip_address': None,
                                  'name': '13 MURRAY CT (DOST MATTHEW)',
                                  'serial': '872558',
                                  'serial_number': '872558',
                                  'system_mac_address': None},
                '87257E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 MURRAY CT (ALBRECHT NIKKETA)',
                                  'serial': '87257E',
                                  'serial_number': '87257E',
                                  'system_mac_address': None},
                '87414': {       'dev_macs': ['b8:94:70:99:4d:4b'],
                                 'ip_address': None,
                                 'name': '148 Winstanley St. Monkton (Stan & Glenda '
                                         'Stolz 10430578)',
                                 'serial': '87414',
                                 'serial_number': '87414',
                                 'system_mac_address': None},
                '87F8BD': {       'dev_macs': ['d0:76:8f:1f:63:b1'],
                                  'ip_address': None,
                                  'name': '85A MILL ST E (KNUDSEN KAREN)',
                                  'serial': '87F8BD',
                                  'serial_number': '87F8BD',
                                  'system_mac_address': None},
                '87F965': {       'dev_macs': ['d0:76:8f:1f:54:0d'],
                                  'ip_address': None,
                                  'name': '18 Mill St. W. (MCCL Shop)',
                                  'serial': '87F965',
                                  'serial_number': '87F965',
                                  'system_mac_address': None},
                '885E5': {       'dev_macs': ['b0:6e:bf:95:30:18'],
                                 'ip_address': None,
                                 'name': '73 Mill St. E. Milverton (Theresa Hopf)',
                                 'serial': '885E5',
                                 'serial_number': '885E5',
                                 'system_mac_address': None},
                '8888F': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '93 Mill St. E. Milverton (Dan Crowley)',
                                 'serial': '8888F',
                                 'serial_number': '8888F',
                                 'system_mac_address': None},
                '888A5': {       'dev_macs': ['d0:76:8f:02:ee:05'],
                                 'ip_address': None,
                                 'name': '103 Mill St. E. Milverton (Jim Harloff)',
                                 'serial': '888A5',
                                 'serial_number': '888A5',
                                 'system_mac_address': None},
                '888F2': {       'dev_macs': ['d0:76:8f:02:ef:18'],
                                 'ip_address': None,
                                 'name': '107 Mill St. E. Milverton (Alvin Streicher)',
                                 'serial': '888F2',
                                 'serial_number': '888F2',
                                 'system_mac_address': None},
                '888F5': {       'dev_macs': ['bc:22:28:df:fb:c7'],
                                 'ip_address': None,
                                 'name': '99 Mill St. E. Milverton (C. '
                                         'Winegarden-Bynsdrop 10430437)',
                                 'serial': '888F5',
                                 'serial_number': '888F5',
                                 'system_mac_address': None},
                '888FA': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '75 Mill St. E. Milverton (Louise Campbell)',
                                 'serial': '888FA',
                                 'serial_number': '888FA',
                                 'system_mac_address': None},
                '88925': {       'dev_macs': ['60:db:98:09:1d:bc'],
                                 'ip_address': None,
                                 'name': '105 Mill St. E. Milverton (Ted Newman)',
                                 'serial': '88925',
                                 'serial_number': '88925',
                                 'system_mac_address': None},
                '88929': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '31 Mill St. E. Milverton (Dave Brand)',
                                 'serial': '88929',
                                 'serial_number': '88929',
                                 'system_mac_address': None},
                '8894F': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '48 GERBER DR (BAILLIE TERRY)',
                                 'serial': '8894F',
                                 'serial_number': '8894F',
                                 'system_mac_address': None},
                '88971': {       'dev_macs': ['00:03:e6:a1:c2:26'],
                                 'ip_address': None,
                                 'name': '43 Mill St. E. Milverton (Micheal Luckhardt '
                                         '10423461)',
                                 'serial': '88971',
                                 'serial_number': '88971',
                                 'system_mac_address': None},
                '889A5': {       'dev_macs': ['cc:be:59:32:94:1a'],
                                 'ip_address': None,
                                 'name': '97 Mill St. E. Milverton (Scott Steckly '
                                         '10402469)',
                                 'serial': '889A5',
                                 'serial_number': '889A5',
                                 'system_mac_address': None},
                '889A6': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '47 Gerber Dr. Milverton (Ron Hanna)',
                                 'serial': '889A6',
                                 'serial_number': '889A6',
                                 'system_mac_address': None},
                '889B7': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '8 Gerber Ct. Milverton (Maynard Streicher)',
                                 'serial': '889B7',
                                 'serial_number': '889B7',
                                 'system_mac_address': None},
                '889BA': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '4495 LINE 64 (KUEPFER KEN & CHRISTENA)',
                                 'serial': '889BA',
                                 'serial_number': '889BA',
                                 'system_mac_address': None},
                '889BD': {       'dev_macs': ['60:db:98:82:08:0c'],
                                 'ip_address': None,
                                 'name': '24 Gerber Dr. Milverton (Mark Matheson)',
                                 'serial': '889BD',
                                 'serial_number': '889BD',
                                 'system_mac_address': None},
                '889C1': {       'dev_macs': ['c8:9e:43:d6:7a:0a'],
                                 'ip_address': None,
                                 'name': '55 Mill St. E. Milverton (Kelly Hargrave '
                                         '10434003)',
                                 'serial': '889C1',
                                 'serial_number': '889C1',
                                 'system_mac_address': None},
                '889D1': {       'dev_macs': ['48:77:46:9a:56:1f'],
                                 'ip_address': None,
                                 'name': '37 MILL ST E (HORTON LEXY)',
                                 'serial': '889D1',
                                 'serial_number': '889D1',
                                 'system_mac_address': None},
                '889DE': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '109 Mill St. E. Milverton (Murray Stewart)',
                                 'serial': '889DE',
                                 'serial_number': '889DE',
                                 'system_mac_address': None},
                '889E1': {       'dev_macs': ['ec:4f:82:2b:47:fd'],
                                 'ip_address': None,
                                 'name': '83 MILL ST E (PAJOT MIKE)',
                                 'serial': '889E1',
                                 'serial_number': '889E1',
                                 'system_mac_address': None},
                '889E2': {       'dev_macs': ['d0:76:8f:1f:5c:16'],
                                 'ip_address': None,
                                 'name': '36 GERBER DR (PARKER JIM & SHEILA)',
                                 'serial': '889E2',
                                 'serial_number': '889E2',
                                 'system_mac_address': None},
                '889E6': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '56 Mill St. E. Milverton (Mark Gingerich '
                                         '10432800)',
                                 'serial': '889E6',
                                 'serial_number': '889E6',
                                 'system_mac_address': None},
                '889E7': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '10B TEMPERANCE STREET (PATERSON JOE & '
                                         'SHARON)',
                                 'serial': '889E7',
                                 'serial_number': '889E7',
                                 'system_mac_address': None},
                '889E9': {       'dev_macs': ['70:8b:cd:60:08:0c'],
                                 'ip_address': None,
                                 'name': '2 St. Marys St. S. Milverton (Darcy Bauman '
                                         '10433545)',
                                 'serial': '889E9',
                                 'serial_number': '889E9',
                                 'system_mac_address': None},
                '889EA': {       'dev_macs': ['b8:94:70:9e:cc:54'],
                                 'ip_address': None,
                                 'name': '87 Mill St. E. Milverton (David Kuehn)',
                                 'serial': '889EA',
                                 'serial_number': '889EA',
                                 'system_mac_address': None},
                '889EC': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '65 Mill St. E. Milverton',
                                 'serial': '889EC',
                                 'serial_number': '889EC',
                                 'system_mac_address': None},
                '889EF': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '28 Gerber Dr. Milverton (Norman Steckly)',
                                 'serial': '889EF',
                                 'serial_number': '889EF',
                                 'system_mac_address': None},
                '889F1': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '52 Gerber Dr. Milverton (Urius Bowman)',
                                 'serial': '889F1',
                                 'serial_number': '889F1',
                                 'system_mac_address': None},
                '889FD': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '12 JO-LENE CT (WOODTHORPE STEPHEN)',
                                 'serial': '889FD',
                                 'serial_number': '889FD',
                                 'system_mac_address': None},
                '88A00': {       'dev_macs': ['5c:0f:fb:2a:0b:d2'],
                                 'ip_address': None,
                                 'name': '33 Mill st. E. Milverton (Frank Huras)',
                                 'serial': '88A00',
                                 'serial_number': '88A00',
                                 'system_mac_address': None},
                '88A05': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '49 Mill St. E. Milverton (Bill Dale)',
                                 'serial': '88A05',
                                 'serial_number': '88A05',
                                 'system_mac_address': None},
                '88A1E': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '8 Colbalt St. S. Milverton (Elroy Horst)',
                                 'serial': '88A1E',
                                 'serial_number': '88A1E',
                                 'system_mac_address': None},
                '88A1F': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '53 Mill St. E. Milverton (Clark Moore '
                                         '10401228)',
                                 'serial': '88A1F',
                                 'serial_number': '88A1F',
                                 'system_mac_address': None},
                '8A06D': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '4 Gerber Dr. Milverton (Sean McCurdy)',
                                 'serial': '8A06D',
                                 'serial_number': '8A06D',
                                 'system_mac_address': None},
                '8A076': {       'dev_macs': ['f0:79:59:c8:1f:de'],
                                 'ip_address': None,
                                 'name': '16 Gerber Dr. Milverton (Peter Flanagan)',
                                 'serial': '8A076',
                                 'serial_number': '8A076',
                                 'system_mac_address': None},
                '8A090': {       'dev_macs': ['ec:4f:82:2d:32:47'],
                                 'ip_address': None,
                                 'name': '7 Gerber Dr. Milverton (Murray Pfeifer)',
                                 'serial': '8A090',
                                 'serial_number': '8A090',
                                 'system_mac_address': None},
                '8A231': {       'dev_macs': [       '00:03:e6:db:2f:90',
                                                     '00:03:e6:db:30:de',
                                                     'cc:be:59:32:95:9d'],
                                 'ip_address': None,
                                 'name': '77 Mill St. E. Milverton (Josh Ludwig, '
                                         '10431732)',
                                 'serial': '8A231',
                                 'serial_number': '8A231',
                                 'system_mac_address': None},
                '8A28C': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '109B Mill St. E. Milverton (Paul Jantzi '
                                         '10403500)',
                                 'serial': '8A28C',
                                 'serial_number': '8A28C',
                                 'system_mac_address': None},
                '8BDC8': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '3 GERBER DR (SCHMIDT CALVIN)',
                                 'serial': '8BDC8',
                                 'serial_number': '8BDC8',
                                 'system_mac_address': None},
                '8C300': {       'dev_macs': ['60:db:98:09:2c:63'],
                                 'ip_address': None,
                                 'name': '61 Mill St. E. Milverton (Jeff Scott)',
                                 'serial': '8C300',
                                 'serial_number': '8C300',
                                 'system_mac_address': None},
                '8E109': {       'dev_macs': ['60:db:98:09:2e:73'],
                                 'ip_address': None,
                                 'name': '69 Mill St. E. Milverton (Lois McEwin '
                                         '10401294)',
                                 'serial': '8E109',
                                 'serial_number': '8E109',
                                 'system_mac_address': None},
                '956DA6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '213 Selby Ave, Monkton (Craig Burek '
                                          '10435498)',
                                  'serial': '956DA6',
                                  'serial_number': '956DA6',
                                  'system_mac_address': None},
                '956E58': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5095 PERTH RD 135 ROSTOCK (DEBUS JOSH)',
                                  'serial': '956E58',
                                  'serial_number': '956E58',
                                  'system_mac_address': None},
                '956E59': {       'dev_macs': ['84:d3:43:6d:e9:b1'],
                                  'ip_address': None,
                                  'name': '25 West St. Milverton (Tom Broughton '
                                          '10401292)',
                                  'serial': '956E59',
                                  'serial_number': '956E59',
                                  'system_mac_address': None},
                '956EF1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '19 MAIN ST N Apt 1',
                                  'serial': '956EF1',
                                  'serial_number': '956EF1',
                                  'system_mac_address': None},
                '956F4A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST (WICE DAVID)',
                                  'serial': '956F4A',
                                  'serial_number': '956F4A',
                                  'system_mac_address': None},
                '956F67': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': "5920 PERTH ROAD 119 (JAY'S GARDEN GAZEBOS)",
                                  'serial': '956F67',
                                  'serial_number': '956F67',
                                  'system_mac_address': None},
                '956F75': {       'dev_macs': ['d0:76:8f:91:d5:e5'],
                                  'ip_address': None,
                                  'name': '92 REAGAN ST (KROEKER JOHN)',
                                  'serial': '956F75',
                                  'serial_number': '956F75',
                                  'system_mac_address': None},
                '956FAE': {       'dev_macs': ['d0:76:8f:91:d2:a1'],
                                  'ip_address': None,
                                  'name': '21 TEMPERANCE ST (RICHARDS MARTY)',
                                  'serial': '956FAE',
                                  'serial_number': '956FAE',
                                  'system_mac_address': None},
                '9C2EFB': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '13 MAIN ST N (FRIESEN HENRY)',
                                  'serial': '9C2EFB',
                                  'serial_number': '9C2EFB',
                                  'system_mac_address': None},
                '9C2F08': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '163 SMITH ST (FRIESEN CORNELIUS)',
                                  'serial': '9C2F08',
                                  'serial_number': '9C2F08',
                                  'system_mac_address': None},
                '9C3022': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '10 MAIN ST N (RAHMAN HABIBUR)',
                                  'serial': '9C3022',
                                  'serial_number': '9C3022',
                                  'system_mac_address': None},
                '9C3023': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 MAIN ST N (MILLIGAN JACKIE)',
                                  'serial': '9C3023',
                                  'serial_number': '9C3023',
                                  'system_mac_address': None},
                '9CA0D': {       'dev_macs': ['c8:4f:86:a5:0f:14'],
                                 'ip_address': None,
                                 'name': '25 Mill St. E. Milverton (Perth East Office '
                                         '10411713)',
                                 'serial': '9CA0D',
                                 'serial_number': '9CA0D',
                                 'system_mac_address': None},
                'A9C29E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '10 MAIN ST S (GERBER CAITLYN)',
                                  'serial': 'A9C29E',
                                  'serial_number': 'A9C29E',
                                  'system_mac_address': None},
                'A9C2B8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '13 MAIN ST N',
                                  'serial': 'A9C2B8',
                                  'serial_number': 'A9C2B8',
                                  'system_mac_address': None},
                'A9C2BD': {       'dev_macs': ['d0:76:8f:02:f3:7a'],
                                  'ip_address': None,
                                  'name': '4676 Line 55 (MICHAEL GIBSON)',
                                  'serial': 'A9C2BD',
                                  'serial_number': 'A9C2BD',
                                  'system_mac_address': None},
                'A9C37D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6A Main St. S. Milverton (ALLARD PIERRE & '
                                          'PETA)',
                                  'serial': 'A9C37D',
                                  'serial_number': 'A9C37D',
                                  'system_mac_address': None},
                'A9C3BB': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '58 MADDISON ST E (REDECOP JOHN)',
                                  'serial': 'A9C3BB',
                                  'serial_number': 'A9C3BB',
                                  'system_mac_address': None},
                'A9C3C6': {       'dev_macs': ['d0:76:8f:91:d5:cf'],
                                  'ip_address': None,
                                  'name': '39 SCHADE ST (CHANSE JORDAN)',
                                  'serial': 'A9C3C6',
                                  'serial_number': 'A9C3C6',
                                  'system_mac_address': None},
                'A9C3D2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '128 WINSTANLEY ST (SWINKELS IAN)',
                                  'serial': 'A9C3D2',
                                  'serial_number': 'A9C3D2',
                                  'system_mac_address': None},
                'A9C3E8': {       'dev_macs': ['60:db:98:55:bb:a8'],
                                  'ip_address': None,
                                  'name': '21 WILLIAM ST (WAGNER ASHLEY)',
                                  'serial': 'A9C3E8',
                                  'serial_number': 'A9C3E8',
                                  'system_mac_address': None},
                'A9C40E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '63 GERBER DR (FALCONER DAVE)',
                                  'serial': 'A9C40E',
                                  'serial_number': 'A9C40E',
                                  'system_mac_address': None},
                'A9C41B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '59 WILLIAM ST (MARTIN RENITA)',
                                  'serial': 'A9C41B',
                                  'serial_number': 'A9C41B',
                                  'system_mac_address': None},
                'A9C675': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '29 MILL ST E (HASENPFLUG ED & LOUISE)',
                                  'serial': 'A9C675',
                                  'serial_number': 'A9C675',
                                  'system_mac_address': None},
                'A9C6FA': {       'dev_macs': ['d0:76:8f:91:d4:dd'],
                                  'ip_address': None,
                                  'name': '29 TEMPERENCE ST, MILVERTON (Melissa & '
                                          'Chris Stahlke)',
                                  'serial': 'A9C6FA',
                                  'serial_number': 'A9C6FA',
                                  'system_mac_address': None},
                'A9C76F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '163 MARKET ST (BOESE PETER)',
                                  'serial': 'A9C76F',
                                  'serial_number': 'A9C76F',
                                  'system_mac_address': None},
                'A9C77B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5948 PERTH ROAD 119 (ENGLER DON)',
                                  'serial': 'A9C77B',
                                  'serial_number': 'A9C77B',
                                  'system_mac_address': None},
                'A9C7F9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '13 MAIN ST N (KLASSEN CAMERON)',
                                  'serial': 'A9C7F9',
                                  'serial_number': 'A9C7F9',
                                  'system_mac_address': None},
                'A9C879': {       'dev_macs': ['d0:76:8f:91:d6:74'],
                                  'ip_address': None,
                                  'name': '20 Gerber Dr. Milverton (LUDWIG RYAN)',
                                  'serial': 'A9C879',
                                  'serial_number': 'A9C879',
                                  'system_mac_address': None},
                'A9C8AA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '101 MADISON ST E MONKTON (GREEN MARIA)',
                                  'serial': 'A9C8AA',
                                  'serial_number': 'A9C8AA',
                                  'system_mac_address': None},
                'A9C8B4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4249 LINE 61 (JANTZI NELSON)',
                                  'serial': 'A9C8B4',
                                  'serial_number': 'A9C8B4',
                                  'system_mac_address': None},
                'A9C8E5': {       'dev_macs': ['84:d3:43:0f:b7:ce'],
                                  'ip_address': None,
                                  'name': '77 SCHADE ST (VAN DYK HENRY)',
                                  'serial': 'A9C8E5',
                                  'serial_number': 'A9C8E5',
                                  'system_mac_address': None},
                'A9C8E6': {       'dev_macs': ['d0:76:8f:1f:54:ff'],
                                  'ip_address': None,
                                  'name': '195 SELBY AV (SMITH EDWARD)',
                                  'serial': 'A9C8E6',
                                  'serial_number': 'A9C8E6',
                                  'system_mac_address': None},
                'A9C90F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4924 PERTH LINE 55 (WESTMAN STEPHEN & '
                                          'LINDA)',
                                  'serial': 'A9C90F',
                                  'serial_number': 'A9C90F',
                                  'system_mac_address': None},
                'A9C94B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5331 ROAD 144 (DAVIDSON FRASER & STARLA)',
                                  'serial': 'A9C94B',
                                  'serial_number': 'A9C94B',
                                  'system_mac_address': None},
                'A9C986': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '16 Mill St. E. Milverton (Tech Office '
                                          'Basement)',
                                  'serial': 'A9C986',
                                  'serial_number': 'A9C986',
                                  'system_mac_address': None},
                'A9C9C0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '78 MADDISON ST E (HIEBERT ABRAHAM)',
                                  'serial': 'A9C9C0',
                                  'serial_number': 'A9C9C0',
                                  'system_mac_address': None},
                'A9F08': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '167 Fishleigh St. Monkton (Mike Danieli)',
                                 'serial': 'A9F08',
                                 'serial_number': 'A9F08',
                                 'system_mac_address': None},
                'AC44E': {       'dev_macs': [       '00:40:7f:96:4d:31',
                                                     '1e:2d:5d:56:58:9f',
                                                     '24:9a:d8:9b:4c:51',
                                                     '26:fe:73:5a:c2:78',
                                                     '28:d0:ea:d1:e1:10',
                                                     '3e:9a:82:71:b8:5d',
                                                     '44:db:d2:28:4a:7f',
                                                     '44:db:d2:4a:59:ca',
                                                     '4e:a7:31:1b:7e:2f',
                                                     '52:00:6c:aa:87:59',
                                                     '52:eb:16:84:bb:f0',
                                                     '58:1c:f8:6d:29:9b',
                                                     '5e:6f:dc:0f:75:4f',
                                                     '60:22:32:45:a4:38',
                                                     '64:00:6a:72:72:d6',
                                                     '74:86:e2:07:24:0c',
                                                     '74:86:e2:07:a1:9f',
                                                     '78:45:58:43:c3:d9',
                                                     'a8:3c:a5:20:a7:b5',
                                                     'c2:88:9e:d8:3f:a9',
                                                     'ca:51:32:b1:ca:30',
                                                     'ca:d9:b4:e9:b5:b2',
                                                     'cc:96:e5:20:76:47',
                                                     'cc:96:e5:20:77:d9',
                                                     'd0:8e:79:13:6d:c8',
                                                     'd0:8e:79:13:6d:d7',
                                                     'd4:f0:c9:0d:51:50'],
                                 'ip_address': None,
                                 'name': '19 Mill St. E. Milverton (Perth East '
                                         'Library)',
                                 'serial': 'AC44E',
                                 'serial_number': 'AC44E',
                                 'system_mac_address': None},
                'ADB03': {       'dev_macs': ['44:65:7f:88:48:10'],
                                 'ip_address': None,
                                 'name': '55 MADDISON ST W (REID MELISSA)',
                                 'serial': 'ADB03',
                                 'serial_number': 'ADB03',
                                 'system_mac_address': None},
                'B08EF6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6988 MILLBANK MAIN ST (WEBER KACIE)',
                                  'serial': 'B08EF6',
                                  'serial_number': 'B08EF6',
                                  'system_mac_address': None},
                'B09F39': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6619 PERTH ROAD 131 (MILVERTON CHRISTIAN '
                                          'FELLOWSHIP)',
                                  'serial': 'B09F39',
                                  'serial_number': 'B09F39',
                                  'system_mac_address': None},
                'B0A14E': {       'dev_macs': ['00:03:e6:fa:10:92'],
                                  'ip_address': None,
                                  'name': '5944 PERTH ROAD 119 (PRONG DOUGLAS & LISA)',
                                  'serial': 'B0A14E',
                                  'serial_number': 'B0A14E',
                                  'system_mac_address': None},
                'B0A212': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6114 ANA ST (KRAEMER RICHARD & FRANCES)',
                                  'serial': 'B0A212',
                                  'serial_number': 'B0A212',
                                  'system_mac_address': None},
                'B0A274': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6414 ROAD 140 (ALIGN PLUMBING)',
                                  'serial': 'B0A274',
                                  'serial_number': 'B0A274',
                                  'system_mac_address': None},
                'B0A288': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5307 ROAD 130 (BALL ANDREW)',
                                  'serial': 'B0A288',
                                  'serial_number': 'B0A288',
                                  'system_mac_address': None},
                'B0A2A6': {       'dev_macs': ['60:db:98:55:b5:8e'],
                                  'ip_address': None,
                                  'name': '167 MADDISON ST E (MACHAN TORIE)',
                                  'serial': 'B0A2A6',
                                  'serial_number': 'B0A2A6',
                                  'system_mac_address': None},
                'B0AAE8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4 MAIN ST S (MACEWAN ADAM)',
                                  'serial': 'B0AAE8',
                                  'serial_number': 'B0AAE8',
                                  'system_mac_address': None},
                'B0AAF1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '55 WILLIAM ST 207-2',
                                  'serial': 'B0AAF1',
                                  'serial_number': 'B0AAF1',
                                  'system_mac_address': None},
                'B50DC4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4699 Line 49 (Darryl Terpstra 10430074)',
                                  'serial': 'B50DC4',
                                  'serial_number': 'B50DC4',
                                  'system_mac_address': None},
                'B5114C': {       'dev_macs': ['b8:94:70:bc:b7:e0'],
                                  'ip_address': None,
                                  'name': '5593 DUKE ST (LEIS WARREN)',
                                  'serial': 'B5114C',
                                  'serial_number': 'B5114C',
                                  'system_mac_address': None},
                'B51169': {       'dev_macs': ['78:98:e8:5f:d5:94'],
                                  'ip_address': None,
                                  'name': '6313 ROAD 140 (FALCONER JOHN)',
                                  'serial': 'B51169',
                                  'serial_number': 'B51169',
                                  'system_mac_address': None},
                'B5116C': {       'dev_macs': ['10:27:f5:32:b8:bb'],
                                  'ip_address': None,
                                  'name': '5917 PERTH ROAD 119 (GERBER KATHRYN)',
                                  'serial': 'B5116C',
                                  'serial_number': 'B5116C',
                                  'system_mac_address': None},
                'B51178': {       'dev_macs': ['b0:4e:26:e9:95:54'],
                                  'ip_address': None,
                                  'name': '6248 ROAD 124 (CLAASSEN ELISABETH '
                                          '(ISABELLE))',
                                  'serial': 'B51178',
                                  'serial_number': 'B51178',
                                  'system_mac_address': None},
                'B511B7': {       'dev_macs': ['24:4b:fe:1f:59:d8'],
                                  'ip_address': None,
                                  'name': '56 MADDISON ST W (BOYD RYAN)',
                                  'serial': 'B511B7',
                                  'serial_number': 'B511B7',
                                  'system_mac_address': None},
                'B511B9': {       'dev_macs': ['b8:94:70:bc:b8:38'],
                                  'ip_address': None,
                                  'name': '6214 ROAD 124 (KUNTZE JOHN)',
                                  'serial': 'B511B9',
                                  'serial_number': 'B511B9',
                                  'system_mac_address': None},
                'B511E0': {       'dev_macs': ['e0:63:da:83:f9:f0'],
                                  'ip_address': None,
                                  'name': '64 GERBER DR (LOESCH JOSEPH)',
                                  'serial': 'B511E0',
                                  'serial_number': 'B511E0',
                                  'system_mac_address': None},
                'B512B7': {       'dev_macs': ['04:bc:9f:46:94:a0'],
                                  'ip_address': None,
                                  'name': '77 GERBER DR (SZCZEPANEK KRZYSZTOF (KRIS))',
                                  'serial': 'B512B7',
                                  'serial_number': 'B512B7',
                                  'system_mac_address': None},
                'B512DB': {       'dev_macs': ['dc:62:79:c7:74:71'],
                                  'ip_address': None,
                                  'name': '5956 PERTH RD 119-5 (ELROY WAGLER MASONRY)',
                                  'serial': 'B512DB',
                                  'serial_number': 'B512DB',
                                  'system_mac_address': None},
                'B512DC': {       'dev_macs': ['00:21:91:f8:32:4d'],
                                  'ip_address': None,
                                  'name': '14 MAIN ST N (NORTHERN DREAM INK)',
                                  'serial': 'B512DC',
                                  'serial_number': 'B512DC',
                                  'system_mac_address': None},
                'BA8C3': {       'dev_macs': [       '5c:0f:fb:3d:52:38',
                                                     '5c:0f:fb:17:75:30',
                                                     '00:03:e6:a5:6f:84',
                                                     '5c:0f:fb:43:96:ae',
                                                     '5c:0f:fb:3d:5a:a0'],
                                 'ip_address': None,
                                 'name': '50 WILLIAM ST',
                                 'serial': 'BA8C3',
                                 'serial_number': 'BA8C3',
                                 'system_mac_address': None},
                'BDD9F4': {       'dev_macs': ['60:32:b1:e9:de:bc'],
                                  'ip_address': None,
                                  'name': '4082 PERTH LINE 72 (GOSPEL ECHOES)',
                                  'serial': 'BDD9F4',
                                  'serial_number': 'BDD9F4',
                                  'system_mac_address': None},
                'BDD9FD': {       'dev_macs': ['f4:92:bf:84:2a:53'],
                                  'ip_address': None,
                                  'name': '4072 PERTH LINE 72 (MILESTONE EQUIPMENT '
                                          'LTD)',
                                  'serial': 'BDD9FD',
                                  'serial_number': 'BDD9FD',
                                  'system_mac_address': None},
                'BF2429': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '43 SCHADE ST (DUNCAN KELLY)',
                                  'serial': 'BF2429',
                                  'serial_number': 'BF2429',
                                  'system_mac_address': None},
                'BF250E': {       'dev_macs': ['14:eb:b6:58:a7:d5'],
                                  'ip_address': None,
                                  'name': '25 PACIFIC AV (DOBES SEBASTIAN)',
                                  'serial': 'BF250E',
                                  'serial_number': 'BF250E',
                                  'system_mac_address': None},
                'C4DD1': {       'dev_macs': ['84:d3:43:6d:e8:eb'],
                                 'ip_address': None,
                                 'name': '4 Jo-Lenne Ct. Milverton (Jay Schultz)',
                                 'serial': 'C4DD1',
                                 'serial_number': 'C4DD1',
                                 'system_mac_address': None},
                'C4DDF': {       'dev_macs': ['ec:4f:82:2c:70:52'],
                                 'ip_address': None,
                                 'name': '15 Gerber Dr. Milverton (Jeff Keller)',
                                 'serial': 'C4DDF',
                                 'serial_number': 'C4DDF',
                                 'system_mac_address': None},
                'C4DEA': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '19 St. Marys St. S. Milverton (KUEPFER '
                                         'MARTHA)',
                                 'serial': 'C4DEA',
                                 'serial_number': 'C4DEA',
                                 'system_mac_address': None},
                'C4DEC': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '5 St. Marys St. S. Milverton (Wayne Poole '
                                         '10421583)',
                                 'serial': 'C4DEC',
                                 'serial_number': 'C4DEC',
                                 'system_mac_address': None},
                'C4DF6': {       'dev_macs': ['d0:76:8f:02:f2:0f'],
                                 'ip_address': None,
                                 'name': '24 Gerber Ct. Milverton (John Rasmussen)',
                                 'serial': 'C4DF6',
                                 'serial_number': 'C4DF6',
                                 'system_mac_address': None},
                'C4DFB': {       'dev_macs': ['cc:be:59:32:94:74'],
                                 'ip_address': None,
                                 'name': '16 ST MARYS ST S (SCHMIDT DAVID & KAREN)',
                                 'serial': 'C4DFB',
                                 'serial_number': 'C4DFB',
                                 'system_mac_address': None},
                'C4DFD': {       'dev_macs': ['14:21:03:d1:7e:12'],
                                 'ip_address': None,
                                 'name': '8 JO-LENE CT (FARIES JACINTA)',
                                 'serial': 'C4DFD',
                                 'serial_number': 'C4DFD',
                                 'system_mac_address': None},
                'C4F9B': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '17 ST MARYS ST S (Kristin Brooks 10438137)',
                                 'serial': 'C4F9B',
                                 'serial_number': 'C4F9B',
                                 'system_mac_address': None},
                'C4FE7': {       'dev_macs': ['b8:94:70:f3:e2:b4'],
                                 'ip_address': None,
                                 'name': '185 ERSKINE ST (IRELAND MORRIS)',
                                 'serial': 'C4FE7',
                                 'serial_number': 'C4FE7',
                                 'system_mac_address': None},
                'C4FE9': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '14 Pugh St. Milverton (COOK GERALD)',
                                 'serial': 'C4FE9',
                                 'serial_number': 'C4FE9',
                                 'system_mac_address': None},
                'C4FEA': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '8 ST MARYS ST S (WADDELL DALE)',
                                 'serial': 'C4FEA',
                                 'serial_number': 'C4FEA',
                                 'system_mac_address': None},
                'C4FF8': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '4 St. Marys St. S. Milverton (Rick Baier)',
                                 'serial': 'C4FF8',
                                 'serial_number': 'C4FF8',
                                 'system_mac_address': None},
                'C4FFA': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '15 ST MARYS ST S (MOHR RON)',
                                 'serial': 'C4FFA',
                                 'serial_number': 'C4FFA',
                                 'system_mac_address': None},
                'C4FFB': {       'dev_macs': ['a0:ab:1b:48:c4:02'],
                                 'ip_address': None,
                                 'name': '12 Gerber Ct. Milverton (William Westman)',
                                 'serial': 'C4FFB',
                                 'serial_number': 'C4FFB',
                                 'system_mac_address': None},
                'C4FFC': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '39 Gerber Dr. Milverton (Allan Steckley)',
                                 'serial': 'C4FFC',
                                 'serial_number': 'C4FFC',
                                 'system_mac_address': None},
                'C5002': {       'dev_macs': ['44:65:7f:88:d5:13'],
                                 'ip_address': None,
                                 'name': '31 Gerber Dr. Milverton (Nadine Hanna)',
                                 'serial': 'C5002',
                                 'serial_number': 'C5002',
                                 'system_mac_address': None},
                'C5004': {       'dev_macs': ['d0:76:8f:02:f1:12'],
                                 'ip_address': None,
                                 'name': '23 Gerber Dr. Milverton (Joe Streicher)',
                                 'serial': 'C5004',
                                 'serial_number': 'C5004',
                                 'system_mac_address': None},
                'C5039': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '81 Mill St. E. Milverton (Brian Hall)',
                                 'serial': 'C5039',
                                 'serial_number': 'C5039',
                                 'system_mac_address': None},
                'C5703': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '19 Gerber Dr. Milverton (Rob McEwen)',
                                 'serial': 'C5703',
                                 'serial_number': 'C5703',
                                 'system_mac_address': None},
                'C5713': {       'dev_macs': ['d0:76:8f:02:f5:32'],
                                 'ip_address': None,
                                 'name': '16 GERBER CT (STOKER BRANDON)',
                                 'serial': 'C5713',
                                 'serial_number': 'C5713',
                                 'system_mac_address': None},
                'C573E': {       'dev_macs': ['b8:94:70:f3:e2:9e'],
                                 'ip_address': None,
                                 'name': '63 MILL ST E (GRAY JACKIE)',
                                 'serial': 'C573E',
                                 'serial_number': 'C573E',
                                 'system_mac_address': None},
                'C5745': {       'dev_macs': ['e4:6c:d1:4f:b7:4d'],
                                 'ip_address': None,
                                 'name': '40 GERBER DR (AMSTEIN DEBORAH)',
                                 'serial': 'C5745',
                                 'serial_number': 'C5745',
                                 'system_mac_address': None},
                'C574C': {       'dev_macs': ['ac:9e:17:b2:fb:f8'],
                                 'ip_address': None,
                                 'name': '4 GERBER CT (KIPFER COLBY)',
                                 'serial': 'C574C',
                                 'serial_number': 'C574C',
                                 'system_mac_address': None},
                'C5778': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '3 Gerber Ct. Milverton (Durell Brubacher)',
                                 'serial': 'C5778',
                                 'serial_number': 'C5778',
                                 'system_mac_address': None},
                'C578C': {       'dev_macs': ['50:91:e3:71:ac:20'],
                                 'ip_address': None,
                                 'name': '101 MILL ST E (BAIER ADAM)',
                                 'serial': 'C578C',
                                 'serial_number': 'C578C',
                                 'system_mac_address': None},
                'C57E8': {       'dev_macs': ['40:b0:76:33:aa:00'],
                                 'ip_address': None,
                                 'name': '13 St. Marys St. S. Milverton (Heather '
                                         'Westman)',
                                 'serial': 'C57E8',
                                 'serial_number': 'C57E8',
                                 'system_mac_address': None},
                'C57F1': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '20 GERBER CT. (SCHWARTZENTRUBER GRANT)',
                                 'serial': 'C57F1',
                                 'serial_number': 'C57F1',
                                 'system_mac_address': None},
                'C57FA': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '7 Pugh St. Milverton (Shawn Gilbert)',
                                 'serial': 'C57FA',
                                 'serial_number': 'C57FA',
                                 'system_mac_address': None},
                'C5849': {       'dev_macs': ['ac:8b:a9:11:38:c1'],
                                 'ip_address': None,
                                 'name': '52 MILL ST E (RICHARD MITCHELL)',
                                 'serial': 'C5849',
                                 'serial_number': 'C5849',
                                 'system_mac_address': None},
                'C5852': {       'dev_macs': ['70:4c:a5:35:93:c2'],
                                 'ip_address': None,
                                 'name': '54 Mill St. W. Milverton (Wesbell '
                                         'Technologies)',
                                 'serial': 'C5852',
                                 'serial_number': 'C5852',
                                 'system_mac_address': None},
                'C585F': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '28 Gerber Ct. Milverton (George Tanner)',
                                 'serial': 'C585F',
                                 'serial_number': 'C585F',
                                 'system_mac_address': None},
                'C811E2': {       'dev_macs': [       'e0:55:3d:89:d9:c5',
                                                      '00:0d:b9:55:21:19'],
                                  'ip_address': None,
                                  'name': '4060 PERTH LINE 72 (MILLBANK HOME BAKERY & '
                                          'COUNTRY)',
                                  'serial': 'C811E2',
                                  'serial_number': 'C811E2',
                                  'system_mac_address': None},
                'C811FE': {       'dev_macs': ['44:65:7f:88:d3:09'],
                                  'ip_address': None,
                                  'name': '65 PUGH ST (REKTOR SHANE)',
                                  'serial': 'C811FE',
                                  'serial_number': 'C811FE',
                                  'system_mac_address': None},
                'C81206': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '145 PUGH ST (HAMMER PAM)',
                                  'serial': 'C81206',
                                  'serial_number': 'C81206',
                                  'system_mac_address': None},
                'C81227': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6313 ROAD 124 (STREICHER MURRAY & JOANNE)',
                                  'serial': 'C81227',
                                  'serial_number': 'C81227',
                                  'system_mac_address': None},
                'C815C8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '11 COLBALT ST (MCKONE STEPHANIE)',
                                  'serial': 'C815C8',
                                  'serial_number': 'C815C8',
                                  'system_mac_address': None},
                'C81609': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4097 QUEEN ST (ALBRECHT NATHAN)',
                                  'serial': 'C81609',
                                  'serial_number': 'C81609',
                                  'system_mac_address': None},
                'C816E6': {       'dev_macs': ['04:bc:9f:46:a7:bf'],
                                  'ip_address': None,
                                  'name': '5125 LINE 60 (EBERSOL ISRAEL & CHRISTENA)',
                                  'serial': 'C816E6',
                                  'serial_number': 'C816E6',
                                  'system_mac_address': None},
                'C8196A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6975 CHURCH ST (STREICHER FANNIE)',
                                  'serial': 'C8196A',
                                  'serial_number': 'C8196A',
                                  'system_mac_address': None},
                'C819B8': {       'dev_macs': ['60:db:98:09:1e:8d'],
                                  'ip_address': None,
                                  'name': '132 PUGH ST (GRAHAM JENNIFER)',
                                  'serial': 'C819B8',
                                  'serial_number': 'C819B8',
                                  'system_mac_address': None},
                'C819BA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6914 PERTH ROAD 121 (BARTSCH JOHAN)',
                                  'serial': 'C819BA',
                                  'serial_number': 'C819BA',
                                  'system_mac_address': None},
                'C819DD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6974 CHURCH ST (OXFORD MILKWAY - MILLBANK '
                                          'DEPO)',
                                  'serial': 'C819DD',
                                  'serial_number': 'C819DD',
                                  'system_mac_address': None},
                'C81BB5': {       'dev_macs': [       '04:eb:40:35:aa:f0',
                                                      '04:d4:c4:d6:79:5c'],
                                  'ip_address': None,
                                  'name': '4044 PERTH LINE 72 (MILLBANK FAMILY '
                                          'FURNITURE)',
                                  'serial': 'C81BB5',
                                  'serial_number': 'C81BB5',
                                  'system_mac_address': None},
                'C81C52': {       'dev_macs': [       '00:03:e6:eb:71:ce',
                                                      'd0:76:8f:1f:64:14'],
                                  'ip_address': None,
                                  'name': '169 MARKET ST (HOWARD MARDLING 10431857)',
                                  'serial': 'C81C52',
                                  'serial_number': 'C81C52',
                                  'system_mac_address': None},
                'C81C73': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6974 CHURCH ST (MILLBANK CHEESE AND BUTTER)',
                                  'serial': 'C81C73',
                                  'serial_number': 'C81C73',
                                  'system_mac_address': None},
                'C81C8F': {       'dev_macs': ['b8:94:70:9e:ce:dd'],
                                  'ip_address': None,
                                  'name': '100 PUGH ST (BOIVIN JASON)',
                                  'serial': 'C81C8F',
                                  'serial_number': 'C81C8F',
                                  'system_mac_address': None},
                'C81C91': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6906 ELGIN ST (WILLIS PAULINE)',
                                  'serial': 'C81C91',
                                  'serial_number': 'C81C91',
                                  'system_mac_address': None},
                'C81D2A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6939 ROAD 140 (LANNIN DON)',
                                  'serial': 'C81D2A',
                                  'serial_number': 'C81D2A',
                                  'system_mac_address': None},
                'C81D54': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '60 SCHADE ST (ARMSTRONG GERALD & BARB)',
                                  'serial': 'C81D54',
                                  'serial_number': 'C81D54',
                                  'system_mac_address': None},
                'C81D62': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6945 NICKLIN ST (ROES IVAN & RETA)',
                                  'serial': 'C81D62',
                                  'serial_number': 'C81D62',
                                  'system_mac_address': None},
                'C81D72': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6611D PERTH ROAD 131 (ACRE75)',
                                  'serial': 'C81D72',
                                  'serial_number': 'C81D72',
                                  'system_mac_address': None},
                'C81D7C': {       'dev_macs': ['b8:94:70:bc:c0:a4'],
                                  'ip_address': None,
                                  'name': '93 PUGH ST (SAMUEL MARK & GILLIAN)',
                                  'serial': 'C81D7C',
                                  'serial_number': 'C81D7C',
                                  'system_mac_address': None},
                'C81D80': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '25 MURRAY CT (SOUTH SIDE ESTATES INC)',
                                  'serial': 'C81D80',
                                  'serial_number': 'C81D80',
                                  'system_mac_address': None},
                'C81D85': {       'dev_macs': ['b8:94:70:84:92:69'],
                                  'ip_address': None,
                                  'name': '5 EDWARD ST MILVERTON (METZGER TREVOR & '
                                          'BRANDY)',
                                  'serial': 'C81D85',
                                  'serial_number': 'C81D85',
                                  'system_mac_address': None},
                'C81D86': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5 MADDISON ST E (SPEARS BARBARA)',
                                  'serial': 'C81D86',
                                  'serial_number': 'C81D86',
                                  'system_mac_address': None},
                'C81D87': {       'dev_macs': ['b8:94:70:9e:cd:b4'],
                                  'ip_address': None,
                                  'name': '104 PUGH ST (MOHAMED SOMSACK)',
                                  'serial': 'C81D87',
                                  'serial_number': 'C81D87',
                                  'system_mac_address': None},
                'C81D8A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6987 MILLBANK MAIN ST (FRIESEN JAKE)',
                                  'serial': 'C81D8A',
                                  'serial_number': 'C81D8A',
                                  'system_mac_address': None},
                'C81D8E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5905 PERTH ROAD 119 (JANTZI MARY J)',
                                  'serial': 'C81D8E',
                                  'serial_number': 'C81D8E',
                                  'system_mac_address': None},
                'C81DA5': {       'dev_macs': [       '00:03:e6:a8:0d:a6',
                                                      '5c:0f:fb:17:64:16',
                                                      '5c:0f:fb:17:72:10'],
                                  'ip_address': None,
                                  'name': '6983 CHURCH ST (KUEPFER WILLIAM R)',
                                  'serial': 'C81DA5',
                                  'serial_number': 'C81DA5',
                                  'system_mac_address': None},
                'C81DAA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4104 WYNDHAM ST (TAM KEITH & JACQUI)',
                                  'serial': 'C81DAA',
                                  'serial_number': 'C81DAA',
                                  'system_mac_address': None},
                'C81DC1': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6928 HORIZON TL (GREEN DARRYL)',
                                  'serial': 'C81DC1',
                                  'serial_number': 'C81DC1',
                                  'system_mac_address': None},
                'C81DC5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4069 NORFOLK ST (HORST PAULINE)',
                                  'serial': 'C81DC5',
                                  'serial_number': 'C81DC5',
                                  'system_mac_address': None},
                'C81DD1': {       'dev_macs': ['b8:94:70:bc:ae:ae'],
                                  'ip_address': None,
                                  'name': '152 PUGH ST (BURKE AMY)',
                                  'serial': 'C81DD1',
                                  'serial_number': 'C81DD1',
                                  'system_mac_address': None},
                'C81DD4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4102 WYNDHAM ST (SOULLIERE LINDA)',
                                  'serial': 'C81DD4',
                                  'serial_number': 'C81DD4',
                                  'system_mac_address': None},
                'C81DD5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4056 NORFOLK ST (ZACHARIAS JAKE)',
                                  'serial': 'C81DD5',
                                  'serial_number': 'C81DD5',
                                  'system_mac_address': None},
                'C81DDA': {       'dev_macs': ['b8:94:70:9e:ca:e9'],
                                  'ip_address': None,
                                  'name': '148 PUGH ST (CARREON ANNE)',
                                  'serial': 'C81DDA',
                                  'serial_number': 'C81DDA',
                                  'system_mac_address': None},
                'C81DEC': {       'dev_macs': ['b8:94:70:bc:ac:eb'],
                                  'ip_address': None,
                                  'name': "99 PUGH ST (D'HULSTER EMIEL)",
                                  'serial': 'C81DEC',
                                  'serial_number': 'C81DEC',
                                  'system_mac_address': None},
                'C81DF0': {       'dev_macs': ['60:db:98:09:1d:85'],
                                  'ip_address': None,
                                  'name': '6960 PERTH ROAD 121 (KUEPFER JAMES & SARAH)',
                                  'serial': 'C81DF0',
                                  'serial_number': 'C81DF0',
                                  'system_mac_address': None},
                'C81DF3': {       'dev_macs': [       '5c:0f:fb:17:6e:84',
                                                      '60:db:98:09:3c:49'],
                                  'ip_address': None,
                                  'name': '6540 ROAD 154 (SATCHELL SCOTT)',
                                  'serial': 'C81DF3',
                                  'serial_number': 'C81DF3',
                                  'system_mac_address': None},
                'C81DF4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4062 NORFOLK ST (GUENTHER DAVID)',
                                  'serial': 'C81DF4',
                                  'serial_number': 'C81DF4',
                                  'system_mac_address': None},
                'C81E10': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4885 LINE 59 (WAGNER KYLE & CRYSTAL)',
                                  'serial': 'C81E10',
                                  'serial_number': 'C81E10',
                                  'system_mac_address': None},
                'C81E11': {       'dev_macs': ['14:21:03:cf:b8:4b'],
                                  'ip_address': None,
                                  'name': '6984 CHURCH ST (BRUBACHER DARREN & DARLENE)',
                                  'serial': 'C81E11',
                                  'serial_number': 'C81E11',
                                  'system_mac_address': None},
                'C81E23': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4091 QUEEN ST (WAGLER EARL)',
                                  'serial': 'C81E23',
                                  'serial_number': 'C81E23',
                                  'system_mac_address': None},
                'C81E2D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4626 PERTH LINE 55 (JANTZI MARVIN A)',
                                  'serial': 'C81E2D',
                                  'serial_number': 'C81E2D',
                                  'system_mac_address': None},
                'C81E2E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6950 PERTH ROAD 121 (KUEPFER IVAN & '
                                          'ANNAMAE)',
                                  'serial': 'C81E2E',
                                  'serial_number': 'C81E2E',
                                  'system_mac_address': None},
                'C81E36': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4106 WYNDHAM ST MILLBANK (WAGLER WAYNE & '
                                          'DIANNE)',
                                  'serial': 'C81E36',
                                  'serial_number': 'C81E36',
                                  'system_mac_address': None},
                'C81E3E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6932 PERTH ST (DERKSEN ANDREAS)',
                                  'serial': 'C81E3E',
                                  'serial_number': 'C81E3E',
                                  'system_mac_address': None},
                'C81E40': {       'dev_macs': ['b8:94:70:f3:e1:ac'],
                                  'ip_address': None,
                                  'name': '92 GERBER DR (MEYER JACKIE)',
                                  'serial': 'C81E40',
                                  'serial_number': 'C81E40',
                                  'system_mac_address': None},
                'C81E50': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4092 PRINCESS ST (DELONG JOHN & JOCELYN)',
                                  'serial': 'C81E50',
                                  'serial_number': 'C81E50',
                                  'system_mac_address': None},
                'C81E56': {       'dev_macs': ['b8:94:70:bc:b7:1a'],
                                  'ip_address': None,
                                  'name': '11 COBALT STREET S (JUTZI TREVOR)',
                                  'serial': 'C81E56',
                                  'serial_number': 'C81E56',
                                  'system_mac_address': None},
                'C81E57': {       'dev_macs': ['b8:94:70:9d:ad:ab'],
                                  'ip_address': None,
                                  'name': '7201 ROAD 140 (SCOTT KRISTINE)',
                                  'serial': 'C81E57',
                                  'serial_number': 'C81E57',
                                  'system_mac_address': None},
                'C81E5C': {       'dev_macs': ['b8:94:70:bc:40:fb'],
                                  'ip_address': None,
                                  'name': '5610 PERTH LINE 55 (DAVIDSON KELLY)',
                                  'serial': 'C81E5C',
                                  'serial_number': 'C81E5C',
                                  'system_mac_address': None},
                'C81E5D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6936 HORIZON TL (KEEPSAKES BY KLASSEN)',
                                  'serial': 'C81E5D',
                                  'serial_number': 'C81E5D',
                                  'system_mac_address': None},
                'C81E5E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6946 NICKLIN ST (JANSEN SHERI)',
                                  'serial': 'C81E5E',
                                  'serial_number': 'C81E5E',
                                  'system_mac_address': None},
                'C81E5F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6906 ELGIN ST (BAKER MARNIE)',
                                  'serial': 'C81E5F',
                                  'serial_number': 'C81E5F',
                                  'system_mac_address': None},
                'C81E60': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4081 QUEEN ST (JANTZI MURRAY & TRACILEE)',
                                  'serial': 'C81E60',
                                  'serial_number': 'C81E60',
                                  'system_mac_address': None},
                'C81E61': {       'dev_macs': [       '14:21:03:8b:73:5a',
                                                      '5c:0f:fb:2a:0e:26'],
                                  'ip_address': None,
                                  'name': '6927 HORIZON TL (KROPF AUDREY)',
                                  'serial': 'C81E61',
                                  'serial_number': 'C81E61',
                                  'system_mac_address': None},
                'C81E63': {       'dev_macs': [       '5c:0f:fb:3d:72:c8',
                                                      '5c:0f:fb:3d:61:34'],
                                  'ip_address': None,
                                  'name': '12 GROSCH ST (WETTLAUFER DWIGHT)',
                                  'serial': 'C81E63',
                                  'serial_number': 'C81E63',
                                  'system_mac_address': None},
                'C81E6F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '30 MAIN ST S (OAK TREE FINANCIAL)',
                                  'serial': 'C81E6F',
                                  'serial_number': 'C81E6F',
                                  'system_mac_address': None},
                'C81E71': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4461 PERTH LINE 72 (TAKACH DON)',
                                  'serial': 'C81E71',
                                  'serial_number': 'C81E71',
                                  'system_mac_address': None},
                'C81E75': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '8 LLANA HEIGHTS DR (STECKLY JAMES)',
                                  'serial': 'C81E75',
                                  'serial_number': 'C81E75',
                                  'system_mac_address': None},
                'C81E79': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4806 LINE 48 GADSHILL (LAURA DEWEERD)',
                                  'serial': 'C81E79',
                                  'serial_number': 'C81E79',
                                  'system_mac_address': None},
                'C81E7C': {       'dev_macs': ['84:d3:43:6d:e9:d2'],
                                  'ip_address': None,
                                  'name': '140 PUGH ST (WEST JAMES)',
                                  'serial': 'C81E7C',
                                  'serial_number': 'C81E7C',
                                  'system_mac_address': None},
                'C81E82': {       'dev_macs': ['60:db:98:09:39:aa'],
                                  'ip_address': None,
                                  'name': '136 PUGH ST (ROBARTS CHARLOTTE)',
                                  'serial': 'C81E82',
                                  'serial_number': 'C81E82',
                                  'system_mac_address': None},
                'C81E8D': {       'dev_macs': ['14:21:03:d1:8c:4b'],
                                  'ip_address': None,
                                  'name': '4866 LINE 59 (MILLER WILLIAM A)',
                                  'serial': 'C81E8D',
                                  'serial_number': 'C81E8D',
                                  'system_mac_address': None},
                'C81E96': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4071 WATERLOO ST (STREICHER ELLA & ERVIN)',
                                  'serial': 'C81E96',
                                  'serial_number': 'C81E96',
                                  'system_mac_address': None},
                'C81E9E': {       'dev_macs': ['84:d3:43:0f:b7:a2'],
                                  'ip_address': None,
                                  'name': '147 PUGH ST (BLOUSE ASHLEE)',
                                  'serial': 'C81E9E',
                                  'serial_number': 'C81E9E',
                                  'system_mac_address': None},
                'C81E9F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4074 WYNDHAM ST (CRUMMER TOM)',
                                  'serial': 'C81E9F',
                                  'serial_number': 'C81E9F',
                                  'system_mac_address': None},
                'C81EAC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4092 WATERLOO ST (ELGIE CHAD)',
                                  'serial': 'C81EAC',
                                  'serial_number': 'C81EAC',
                                  'system_mac_address': None},
                'C81EAD': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '2 MAIN ST N (1262174 ONTARIO INC)',
                                  'serial': 'C81EAD',
                                  'serial_number': 'C81EAD',
                                  'system_mac_address': None},
                'C81EB0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6237 ROAD 136 (DIEHL ERIC & SUE)',
                                  'serial': 'C81EB0',
                                  'serial_number': 'C81EB0',
                                  'system_mac_address': None},
                'C81EB7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4088 PRINCESS ST (LICHTI NANCY)',
                                  'serial': 'C81EB7',
                                  'serial_number': 'C81EB7',
                                  'system_mac_address': None},
                'C81EB8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6958 PERTH ROAD 121 (WOELK JOHN)',
                                  'serial': 'C81EB8',
                                  'serial_number': 'C81EB8',
                                  'system_mac_address': None},
                'C81EC3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6927 PERTH ROAD 121 (SCHEERER MARK & SANDY)',
                                  'serial': 'C81EC3',
                                  'serial_number': 'C81EC3',
                                  'system_mac_address': None},
                'C81EC9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6948 ELGIN ST (PETERS ISAAK & SUSANA)',
                                  'serial': 'C81EC9',
                                  'serial_number': 'C81EC9',
                                  'system_mac_address': None},
                'C81ED9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '3 Sippel St. Milverton (Gerald Bain '
                                          '10430059)',
                                  'serial': 'C81ED9',
                                  'serial_number': 'C81ED9',
                                  'system_mac_address': None},
                'C81F02': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4105 CONNIE ST (GERBER DARLENE)',
                                  'serial': 'C81F02',
                                  'serial_number': 'C81F02',
                                  'system_mac_address': None},
                'C81F17': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4095 QUEEN ST (HOBDEN ROBERT & DIANE)',
                                  'serial': 'C81F17',
                                  'serial_number': 'C81F17',
                                  'system_mac_address': None},
                'C81F1A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4076 QUEEN ST (WATKINS JOANNA)',
                                  'serial': 'C81F1A',
                                  'serial_number': 'C81F1A',
                                  'system_mac_address': None},
                'C81F52': {       'dev_macs': ['cc:be:59:7e:b5:cc'],
                                  'ip_address': None,
                                  'name': '102 MILL ST E MILVERTON (JANTZI STEPHEN)',
                                  'serial': 'C81F52',
                                  'serial_number': 'C81F52',
                                  'system_mac_address': None},
                'C81F71': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '73 Maddison St. E. Monkton (Nancy '
                                          'Mcgoldrick)',
                                  'serial': 'C81F71',
                                  'serial_number': 'C81F71',
                                  'system_mac_address': None},
                'C820AC': {       'dev_macs': ['84:d3:43:0f:b5:71'],
                                  'ip_address': None,
                                  'name': '65 REAGAN ST (JORDAO JOSH)',
                                  'serial': 'C820AC',
                                  'serial_number': 'C820AC',
                                  'system_mac_address': None},
                'C820D2': {       'dev_macs': ['84:d3:43:0f:b5:5b'],
                                  'ip_address': None,
                                  'name': '160 QUEEN AVENUE MONKTON (GERBER MICHAEL)',
                                  'serial': 'C820D2',
                                  'serial_number': 'C820D2',
                                  'system_mac_address': None},
                'C83070': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4838 LINE 59 (WAGLER LEN)',
                                  'serial': 'C83070',
                                  'serial_number': 'C83070',
                                  'system_mac_address': None},
                'C8307E': {       'dev_macs': ['60:db:98:09:3c:f9'],
                                  'ip_address': None,
                                  'name': '125 PUGH ST (KIRCH STEVE)',
                                  'serial': 'C8307E',
                                  'serial_number': 'C8307E',
                                  'system_mac_address': None},
                'C83088': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4053 NORFOLK ST (WAGLER RICHARD)',
                                  'serial': 'C83088',
                                  'serial_number': 'C83088',
                                  'system_mac_address': None},
                'C830BA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4083 HILLCREST ST (ERB BARBARA)',
                                  'serial': 'C830BA',
                                  'serial_number': 'C830BA',
                                  'system_mac_address': None},
                'C830CE': {       'dev_macs': ['14:21:03:d1:8a:72'],
                                  'ip_address': None,
                                  'name': '4084 WYNDHAM ST (ROPP DAVID & JANET)',
                                  'serial': 'C830CE',
                                  'serial_number': 'C830CE',
                                  'system_mac_address': None},
                'C830D4': {       'dev_macs': ['84:d3:43:0f:64:04'],
                                  'ip_address': None,
                                  'name': '5339 ROAD 130 (STADELMANN MARK)',
                                  'serial': 'C830D4',
                                  'serial_number': 'C830D4',
                                  'system_mac_address': None},
                'C830D9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5323 PERTH LINE 55 (MUELLER GEORGE)',
                                  'serial': 'C830D9',
                                  'serial_number': 'C830D9',
                                  'system_mac_address': None},
                'C830E8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6990 RUTHERFORD ST (BURNS BLAKE)',
                                  'serial': 'C830E8',
                                  'serial_number': 'C830E8',
                                  'system_mac_address': None},
                'C830F5': {       'dev_macs': ['00:03:e6:dd:30:e8'],
                                  'ip_address': None,
                                  'name': '6963 PERTH ST MILLBANK (BRICK MARIE)',
                                  'serial': 'C830F5',
                                  'serial_number': 'C830F5',
                                  'system_mac_address': None},
                'C830F9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6989 CHURCH ST (ZWART JANELLE)',
                                  'serial': 'C830F9',
                                  'serial_number': 'C830F9',
                                  'system_mac_address': None},
                'C83108': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6939 ELGIN ST (GERBER RAY)',
                                  'serial': 'C83108',
                                  'serial_number': 'C83108',
                                  'system_mac_address': None},
                'C83114': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6972 NICKLIN ST MILLBANK (WAGLER ARLENE)',
                                  'serial': 'C83114',
                                  'serial_number': 'C83114',
                                  'system_mac_address': None},
                'C83116': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '155 PUGH ST (ZIERL MATTHEW)',
                                  'serial': 'C83116',
                                  'serial_number': 'C83116',
                                  'system_mac_address': None},
                'C8312C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5578 ROAD 144 (GINGRICH MAYNARD)',
                                  'serial': 'C8312C',
                                  'serial_number': 'C8312C',
                                  'system_mac_address': None},
                'C8313F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6998 MILLBANK MAIN ST (JANZEN GERHARD)',
                                  'serial': 'C8313F',
                                  'serial_number': 'C8313F',
                                  'system_mac_address': None},
                'C83146': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5921 SOMERSET DR (HART BEN)',
                                  'serial': 'C83146',
                                  'serial_number': 'C83146',
                                  'system_mac_address': None},
                'C8314A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '73 PUGH ST (SZYMIKOWSKI DARIUSZ)',
                                  'serial': 'C8314A',
                                  'serial_number': 'C8314A',
                                  'system_mac_address': None},
                'C83150': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5174 PERTH LINE 55 (BYLER JOHN)',
                                  'serial': 'C83150',
                                  'serial_number': 'C83150',
                                  'system_mac_address': None},
                'C83163': {       'dev_macs': ['84:d3:43:6d:e9:43'],
                                  'ip_address': None,
                                  'name': '165 MARKET ST (SUBJECT HELENA & FRED)',
                                  'serial': 'C83163',
                                  'serial_number': 'C83163',
                                  'system_mac_address': None},
                'C83169': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6927 NICKLIN ST MILLBANK (BAUMAN LAVERNE)',
                                  'serial': 'C83169',
                                  'serial_number': 'C83169',
                                  'system_mac_address': None},
                'C83174': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6994 MILLBANK MAIN STREET (LICHTI MATTHEW & '
                                          'SARAH)',
                                  'serial': 'C83174',
                                  'serial_number': 'C83174',
                                  'system_mac_address': None},
                'C8318D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '110 MILL ST E (WAGLER ALVIN)',
                                  'serial': 'C8318D',
                                  'serial_number': 'C8318D',
                                  'system_mac_address': None},
                'C83193': {       'dev_macs': [       '5c:0f:fb:2f:cc:0f',
                                                      '5c:0f:fb:2f:ce:bb',
                                                      '84:d3:43:6d:ea:40'],
                                  'ip_address': None,
                                  'name': '6998 FREEBORN ST (RICHARDS BILL)',
                                  'serial': 'C83193',
                                  'serial_number': 'C83193',
                                  'system_mac_address': None},
                'C831EA': {       'dev_macs': ['84:d3:43:6d:e8:25'],
                                  'ip_address': None,
                                  'name': '31 WEST ST (TOPHAM-LINTHOME TYLER)',
                                  'serial': 'C831EA',
                                  'serial_number': 'C831EA',
                                  'system_mac_address': None},
                'C831F3': {       'dev_macs': ['14:21:03:d1:52:96'],
                                  'ip_address': None,
                                  'name': '6995 MILLBANK MAIN ST (KROPF BRUCE)',
                                  'serial': 'C831F3',
                                  'serial_number': 'C831F3',
                                  'system_mac_address': None},
                'C8321B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4087 QUEEN ST (MARTIN MERLIN)',
                                  'serial': 'C8321B',
                                  'serial_number': 'C8321B',
                                  'system_mac_address': None},
                'C83246': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6420 ROAD 140 (STREICHER SAM & ESTHER)',
                                  'serial': 'C83246',
                                  'serial_number': 'C83246',
                                  'system_mac_address': None},
                'C83586': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4060 PRINCESS ST (SCHMITT GEOFF)',
                                  'serial': 'C83586',
                                  'serial_number': 'C83586',
                                  'system_mac_address': None},
                'C835A3': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5590 ROAD 126 (ZACH TREVOR)',
                                  'serial': 'C835A3',
                                  'serial_number': 'C835A3',
                                  'system_mac_address': None},
                'C83610': {       'dev_macs': ['34:0a:33:89:92:79'],
                                  'ip_address': None,
                                  'name': '4082B WATERLOO ST (ORANGE BARN)',
                                  'serial': 'C83610',
                                  'serial_number': 'C83610',
                                  'system_mac_address': None},
                'C83691': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6985 PARK ST (ORANGE BARN)',
                                  'serial': 'C83691',
                                  'serial_number': 'C83691',
                                  'system_mac_address': None},
                'C836C4': {       'dev_macs': ['e4:6c:d1:4e:ef:95'],
                                  'ip_address': None,
                                  'name': '6991 MILLBANK MAIN ST (WIEBE EVA)',
                                  'serial': 'C836C4',
                                  'serial_number': 'C836C4',
                                  'system_mac_address': None},
                'C83708': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4050 NORFOLK ST (WAGLER KRISTA)',
                                  'serial': 'C83708',
                                  'serial_number': 'C83708',
                                  'system_mac_address': None},
                'C83821': {       'dev_macs': ['44:65:7f:89:0d:bf'],
                                  'ip_address': None,
                                  'name': '2 MAIN ST N (KAUR LOVEPREET)',
                                  'serial': 'C83821',
                                  'serial_number': 'C83821',
                                  'system_mac_address': None},
                'C8392E': {       'dev_macs': ['14:21:03:92:ee:d8'],
                                  'ip_address': None,
                                  'name': '4 MILL ST E APT 1 (MANE APPEAL)',
                                  'serial': 'C8392E',
                                  'serial_number': 'C8392E',
                                  'system_mac_address': None},
                'C83954': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6213 Rd 124 (VAN NES MICHELLE)',
                                  'serial': 'C83954',
                                  'serial_number': 'C83954',
                                  'system_mac_address': None},
                'C83B0E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6280 ROAD 140',
                                  'serial': 'C83B0E',
                                  'serial_number': 'C83B0E',
                                  'system_mac_address': None},
                'C83B1E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6985 FREEBORN ST (GERBER BETTY)',
                                  'serial': 'C83B1E',
                                  'serial_number': 'C83B1E',
                                  'system_mac_address': None},
                'C83B33': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '20 COBALT STREET S (LEHMANN TRAVIS)',
                                  'serial': 'C83B33',
                                  'serial_number': 'C83B33',
                                  'system_mac_address': None},
                'C83B64': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4939 LINE 49 (BENNEWIES TREVOR)',
                                  'serial': 'C83B64',
                                  'serial_number': 'C83B64',
                                  'system_mac_address': None},
                'C83B6C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6990 CHURCH ST (ZEHR DYLAN)',
                                  'serial': 'C83B6C',
                                  'serial_number': 'C83B6C',
                                  'system_mac_address': None},
                'C83B73': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '9 PACIFIC AV (SCHAEFER ROBERT)',
                                  'serial': 'C83B73',
                                  'serial_number': 'C83B73',
                                  'system_mac_address': None},
                'C83B74': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5920 SOMERSET DR (WIEBE GERHARD)',
                                  'serial': 'C83B74',
                                  'serial_number': 'C83B74',
                                  'system_mac_address': None},
                'C83B7D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '38 PACIFIC AVE MILVERTON (VAN ROOSMALEN '
                                          'KRISTIE)',
                                  'serial': 'C83B7D',
                                  'serial_number': 'C83B7D',
                                  'system_mac_address': None},
                'C83B88': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4051 NORFOLK ST MILLBANK (HORST BILL '
                                          'ACUTIONEER)',
                                  'serial': 'C83B88',
                                  'serial_number': 'C83B88',
                                  'system_mac_address': None},
                'C83B8B': {       'dev_macs': ['60:db:98:09:14:d7'],
                                  'ip_address': None,
                                  'name': '6947 NICKLIN ST (EDWARDS ERNEST & DONNA)',
                                  'serial': 'C83B8B',
                                  'serial_number': 'C83B8B',
                                  'system_mac_address': None},
                'C83B8F': {       'dev_macs': ['14:21:03:d1:8c:77'],
                                  'ip_address': None,
                                  'name': '6935 NICKLIN ST MILLBANK (LICHTY DEL)',
                                  'serial': 'C83B8F',
                                  'serial_number': 'C83B8F',
                                  'system_mac_address': None},
                'C83BA6': {       'dev_macs': ['84:d3:43:0f:b8:cb'],
                                  'ip_address': None,
                                  'name': '16 Jo-Lene Ct. Milverton (Doug Biesinger '
                                          '10402290)',
                                  'serial': 'C83BA6',
                                  'serial_number': 'C83BA6',
                                  'system_mac_address': None},
                'C83BAF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4082 WATERLOO ST (WEBERS FOAM INSULATION)',
                                  'serial': 'C83BAF',
                                  'serial_number': 'C83BAF',
                                  'system_mac_address': None},
                'C83BB4': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4086 PRINCESS ST MILLBANK (ZEHR DAVID)',
                                  'serial': 'C83BB4',
                                  'serial_number': 'C83BB4',
                                  'system_mac_address': None},
                'C83BB8': {       'dev_macs': ['b8:94:70:9e:ca:a7'],
                                  'ip_address': None,
                                  'name': '81 PUGH ST (HEAD JOSH)',
                                  'serial': 'C83BB8',
                                  'serial_number': 'C83BB8',
                                  'system_mac_address': None},
                'C83BBC': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4049 PRINCESS ST (BIEHN RUTH ANNE)',
                                  'serial': 'C83BBC',
                                  'serial_number': 'C83BBC',
                                  'system_mac_address': None},
                'C83BC5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '162 FISHLEIGH ST',
                                  'serial': 'C83BC5',
                                  'serial_number': 'C83BC5',
                                  'system_mac_address': None},
                'C83BC6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6933 PERTH ROAD 121 (STREICHER ADAM)',
                                  'serial': 'C83BC6',
                                  'serial_number': 'C83BC6',
                                  'system_mac_address': None},
                'C83BC8': {       'dev_macs': ['60:db:98:09:3a:0d'],
                                  'ip_address': None,
                                  'name': '4059 NORFOLK ST (POOLE BRIAN)',
                                  'serial': 'C83BC8',
                                  'serial_number': 'C83BC8',
                                  'system_mac_address': None},
                'C83BC9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6976 MILLBANK MAIN ST (WYE CROSS '
                                          'PROPERTIES)',
                                  'serial': 'C83BC9',
                                  'serial_number': 'C83BC9',
                                  'system_mac_address': None},
                'C83BCF': {       'dev_macs': ['14:21:03:d1:7d:62'],
                                  'ip_address': None,
                                  'name': '188 ERSKINE ST (RAMIREZ CARLOS)',
                                  'serial': 'C83BCF',
                                  'serial_number': 'C83BCF',
                                  'system_mac_address': None},
                'C83BD5': {       'dev_macs': ['14:21:03:92:e9:8f'],
                                  'ip_address': None,
                                  'name': '6999 MILLBANK MAIN ST (JANTZI EZRA & '
                                          'ELIZABETH)',
                                  'serial': 'C83BD5',
                                  'serial_number': 'C83BD5',
                                  'system_mac_address': None},
                'C83BD8': {       'dev_macs': ['84:d3:43:6d:ed:0b'],
                                  'ip_address': None,
                                  'name': '11 ST MARYS ST S (FRIESEN PETER)',
                                  'serial': 'C83BD8',
                                  'serial_number': 'C83BD8',
                                  'system_mac_address': None},
                'C83BE6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6476 PERTH ROAD 131 (DYCK ABRAM W)',
                                  'serial': 'C83BE6',
                                  'serial_number': 'C83BE6',
                                  'system_mac_address': None},
                'C83BED': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '85 PUGH ST (CARTER ASHLEY)',
                                  'serial': 'C83BED',
                                  'serial_number': 'C83BED',
                                  'system_mac_address': None},
                'C83BF0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6996 MILLBANK MAIN ST (GERBER FLOYD)',
                                  'serial': 'C83BF0',
                                  'serial_number': 'C83BF0',
                                  'system_mac_address': None},
                'C83BF5': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '220 SELBY AV (GREENWALD ROBERT)',
                                  'serial': 'C83BF5',
                                  'serial_number': 'C83BF5',
                                  'system_mac_address': None},
                'C83BFA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6989 FREEBORN ST (JANTZI JOHN & MALLORY)',
                                  'serial': 'C83BFA',
                                  'serial_number': 'C83BFA',
                                  'system_mac_address': None},
                'C83BFF': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6990 FREEBORN ST (ROBERTS RICHARD & JUDY)',
                                  'serial': 'C83BFF',
                                  'serial_number': 'C83BFF',
                                  'system_mac_address': None},
                'C83C02': {       'dev_macs': ['60:db:98:09:3b:6d'],
                                  'ip_address': None,
                                  'name': '61 MAIN ST N (VATERS CARLIE)',
                                  'serial': 'C83C02',
                                  'serial_number': 'C83C02',
                                  'system_mac_address': None},
                'C83C16': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4038 PERTH LINE 72 (KUEPFER JOE)',
                                  'serial': 'C83C16',
                                  'serial_number': 'C83C16',
                                  'system_mac_address': None},
                'C83C1D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '77 MADDISON ST W (MCEWEN COLLEEN)',
                                  'serial': 'C83C1D',
                                  'serial_number': 'C83C1D',
                                  'system_mac_address': None},
                'C83C21': {       'dev_macs': ['60:db:98:82:01:8f'],
                                  'ip_address': None,
                                  'name': '64 SCHADE ST (YUNDT D DAWN)',
                                  'serial': 'C83C21',
                                  'serial_number': 'C83C21',
                                  'system_mac_address': None},
                'C83E3A': {       'dev_macs': ['84:d3:43:0f:bc:9e'],
                                  'ip_address': None,
                                  'name': '83 GERBER DR (JAMES COLLEEN)',
                                  'serial': 'C83E3A',
                                  'serial_number': 'C83E3A',
                                  'system_mac_address': None},
                'CBD160': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6964 MILLBANK MAIN ST (DORSCHT RANDY)',
                                  'serial': 'CBD160',
                                  'serial_number': 'CBD160',
                                  'system_mac_address': None},
                'CBD16C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '149 PUGH ST (WRIGHT BRITTANY)',
                                  'serial': 'CBD16C',
                                  'serial_number': 'CBD16C',
                                  'system_mac_address': None},
                'CBD16D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '151 PUGH ST (PICKERING RICHARD)',
                                  'serial': 'CBD16D',
                                  'serial_number': 'CBD16D',
                                  'system_mac_address': None},
                'CBD189': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6975 FREEBORN ST (CHAPMAN GORDON R)',
                                  'serial': 'CBD189',
                                  'serial_number': 'CBD189',
                                  'system_mac_address': None},
                'CBD194': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6984 FREEBORN ST (CUNNINGHAM ELAINE)',
                                  'serial': 'CBD194',
                                  'serial_number': 'CBD194',
                                  'system_mac_address': None},
                'CBD19C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6994 FREEBORN ST (ERB LLOYD & LAURENE)',
                                  'serial': 'CBD19C',
                                  'serial_number': 'CBD19C',
                                  'system_mac_address': None},
                'CBD19D': {       'dev_macs': ['d0:76:8f:91:d6:5e'],
                                  'ip_address': None,
                                  'name': '77 PUGH ST (LUDWIG VICTOR)',
                                  'serial': 'CBD19D',
                                  'serial_number': 'CBD19D',
                                  'system_mac_address': None},
                'CBD1BC': {       'dev_macs': ['60:db:98:09:39:b5'],
                                  'ip_address': None,
                                  'name': '6980 MILLBANK MAIN ST (ELIZABETH WATSON '
                                          'FLORAL)',
                                  'serial': 'CBD1BC',
                                  'serial_number': 'CBD1BC',
                                  'system_mac_address': None},
                'CBD1D6': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '22 Maddison St., East, Monkton (Greg & '
                                          'Kathy Leung)',
                                  'serial': 'CBD1D6',
                                  'serial_number': 'CBD1D6',
                                  'system_mac_address': None},
                'CBD1E8': {       'dev_macs': ['84:d3:43:6d:fa:9f'],
                                  'ip_address': None,
                                  'name': '48 MAIN ST N (SCHLOSSER KYLE)',
                                  'serial': 'CBD1E8',
                                  'serial_number': 'CBD1E8',
                                  'system_mac_address': None},
                'CBD1F3': {       'dev_macs': ['b8:94:70:9e:cf:e5'],
                                  'ip_address': None,
                                  'name': '6996 FREEBORN ST (BIGAM RONALD)',
                                  'serial': 'CBD1F3',
                                  'serial_number': 'CBD1F3',
                                  'system_mac_address': None},
                'CBD212': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5908 PERTH ROAD 119 (JANTZI MYRON)',
                                  'serial': 'CBD212',
                                  'serial_number': 'CBD212',
                                  'system_mac_address': None},
                'CBD222': {       'dev_macs': [       '5c:0f:fb:14:3c:4a',
                                                      '5c:0f:fb:17:66:c8',
                                                      'd0:76:8f:02:f7:00'],
                                  'ip_address': None,
                                  'name': '89 GERBER DR (ALVES JORGE)',
                                  'serial': 'CBD222',
                                  'serial_number': 'CBD222',
                                  'system_mac_address': None},
                'CBD224': {       'dev_macs': ['d0:76:8f:91:d6:69'],
                                  'ip_address': None,
                                  'name': '91 REAGAN ST (CASTRO FATIMA  D)',
                                  'serial': 'CBD224',
                                  'serial_number': 'CBD224',
                                  'system_mac_address': None},
                'CBD233': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '32 GERBER DRIVE (NAFZIGER BARB)',
                                  'serial': 'CBD233',
                                  'serial_number': 'CBD233',
                                  'system_mac_address': None},
                'CBD27E': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4541 LINE 61 (NAFZIGER STEVE)',
                                  'serial': 'CBD27E',
                                  'serial_number': 'CBD27E',
                                  'system_mac_address': None},
                'CBD29D': {       'dev_macs': ['54:af:97:39:64:4d'],
                                  'ip_address': None,
                                  'name': '34 Mill St. E. Milverton (Dale Brubacher '
                                          '10431281)',
                                  'serial': 'CBD29D',
                                  'serial_number': 'CBD29D',
                                  'system_mac_address': None},
                'CBE02B': {       'dev_macs': ['d0:76:8f:02:f1:75'],
                                  'ip_address': None,
                                  'name': '153 PUGH ST (ZURBRIGG CHANTELLE)',
                                  'serial': 'CBE02B',
                                  'serial_number': 'CBE02B',
                                  'system_mac_address': None},
                'CBE05B': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '54 MILL ST E (STOREY GRANT)',
                                  'serial': 'CBE05B',
                                  'serial_number': 'CBE05B',
                                  'system_mac_address': None},
                'CBFF4': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '5250 Line 60 (Theo Kersten)',
                                 'serial': 'CBFF4',
                                 'serial_number': 'CBFF4',
                                 'system_mac_address': None},
                'CBFF5': {       'dev_macs': ['ec:4f:82:2d:48:f4'],
                                 'ip_address': None,
                                 'name': '5433 Line 60 (Joanne Ohm)',
                                 'serial': 'CBFF5',
                                 'serial_number': 'CBFF5',
                                 'system_mac_address': None},
                'CC074': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '4713 Line 64 (Scott Steckly 10421864)',
                                 'serial': 'CC074',
                                 'serial_number': 'CC074',
                                 'system_mac_address': None},
                'CC075': {       'dev_macs': ['e8:9f:80:75:ac:4f'],
                                 'ip_address': None,
                                 'name': '48 Mill St. E. Milverton (Michelle Corbett)',
                                 'serial': 'CC075',
                                 'serial_number': 'CC075',
                                 'system_mac_address': None},
                'CC07F': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '3 Llana Heights Milverton (Deewar Randy)',
                                 'serial': 'CC07F',
                                 'serial_number': 'CC07F',
                                 'system_mac_address': None},
                'CC08C': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '45 Maddison St. W. Monkton (Darryl Smith)',
                                 'serial': 'CC08C',
                                 'serial_number': 'CC08C',
                                 'system_mac_address': None},
                'CC095': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '41 Mill St. E. Milverton (Rob Woodrow)',
                                 'serial': 'CC095',
                                 'serial_number': 'CC095',
                                 'system_mac_address': None},
                'CC09A': {       'dev_macs': ['04:bc:9f:46:92:a6'],
                                 'ip_address': None,
                                 'name': '62 MILL ST E (RICHARDSON CLAIRE)',
                                 'serial': 'CC09A',
                                 'serial_number': 'CC09A',
                                 'system_mac_address': None},
                'CC0B4': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '30 Maddison St. W. Monkton (Nancy Diehl)',
                                 'serial': 'CC0B4',
                                 'serial_number': 'CC0B4',
                                 'system_mac_address': None},
                'CC0B8': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '66 Maddison St. E. Monkton (Dan Brown)',
                                 'serial': 'CC0B8',
                                 'serial_number': 'CC0B8',
                                 'system_mac_address': None},
                'CC0BF': {       'dev_macs': ['14:21:03:a9:6f:d9'],
                                 'ip_address': None,
                                 'name': '7 JO-LENE CT (ALEXANDER ADRIAN)',
                                 'serial': 'CC0BF',
                                 'serial_number': 'CC0BF',
                                 'system_mac_address': None},
                'CC0C1': {       'dev_macs': ['cc:be:59:2d:a4:f0'],
                                 'ip_address': None,
                                 'name': '11 JO-LENE CT (BARKER PATRICIA)',
                                 'serial': 'CC0C1',
                                 'serial_number': 'CC0C1',
                                 'system_mac_address': None},
                'CC0C2': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '22 Arena Dr. Milverton (Hairstyling '
                                         'Headquarters 10401026)',
                                 'serial': 'CC0C2',
                                 'serial_number': 'CC0C2',
                                 'system_mac_address': None},
                'CC0C5': {       'dev_macs': ['14:21:03:8b:71:6b'],
                                 'ip_address': None,
                                 'name': '60 WILLIAM ST (JANTZI AARON J/MARTHA)',
                                 'serial': 'CC0C5',
                                 'serial_number': 'CC0C5',
                                 'system_mac_address': None},
                'CC0C9': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '10 St. Marys St. S. Milverton (Ron '
                                         'Schweitzer)',
                                 'serial': 'CC0C9',
                                 'serial_number': 'CC0C9',
                                 'system_mac_address': None},
                'CC0CC': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '5 LLANA HEIGHTS DR (PRESCOTT BRENDA)',
                                 'serial': 'CC0CC',
                                 'serial_number': 'CC0CC',
                                 'system_mac_address': None},
                'CC0D4': {       'dev_macs': [       '00:03:4f:09:c6:0a',
                                                     '00:80:a3:d7:b0:51',
                                                     '24:0f:9b:14:2e:5d',
                                                     'd0:67:e5:ab:f7:82',
                                                     'f0:9f:c2:d0:dc:59'],
                                 'ip_address': None,
                                 'name': '6372 PERTH ROAD 131 (COUNTY OF PERTH)',
                                 'serial': 'CC0D4',
                                 'serial_number': 'CC0D4',
                                 'system_mac_address': None},
                'CC0D5': {       'dev_macs': ['10:c3:7b:a3:51:38'],
                                 'ip_address': None,
                                 'name': '32 Gerber Ct. Milverton (Ken Leduc)',
                                 'serial': 'CC0D5',
                                 'serial_number': 'CC0D5',
                                 'system_mac_address': None},
                'D3E32': {       'dev_macs': ['c8:4f:86:c6:42:10'],
                                 'ip_address': None,
                                 'name': '50 William St. Milverton (Dr. Hodes & '
                                         'Anderson 10401327)',
                                 'serial': 'D3E32',
                                 'serial_number': 'D3E32',
                                 'system_mac_address': None},
                'D3E4E': {       'dev_macs': [],
                                 'ip_address': None,
                                 'name': '111 Mill St. E. Milverton',
                                 'serial': 'D3E4E',
                                 'serial_number': 'D3E4E',
                                 'system_mac_address': None},
                'DFFB40': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4084 PRINCESS ST (REINDL MELISSA)',
                                  'serial': 'DFFB40',
                                  'serial_number': 'DFFB40',
                                  'system_mac_address': None},
                'DFFB59': {       'dev_macs': [       '5c:0f:fb:17:6b:86',
                                                      '60:db:98:09:3c:e3'],
                                  'ip_address': None,
                                  'name': '10 MACBETH ST, Milverton (SCHMIDT JOSH)',
                                  'serial': 'DFFB59',
                                  'serial_number': 'DFFB59',
                                  'system_mac_address': None},
                'DFFB8B': {       'dev_macs': ['14:21:03:92:ec:2e'],
                                  'ip_address': None,
                                  'name': '4063 NORFOLK ST (SADIWNYK CATHERINE)',
                                  'serial': 'DFFB8B',
                                  'serial_number': 'DFFB8B',
                                  'system_mac_address': None},
                'DFFB92': {       'dev_macs': ['84:d3:43:6d:ee:a2'],
                                  'ip_address': None,
                                  'name': '11 Wilfred St. Milverton (Steve Court '
                                          '10403171)',
                                  'serial': 'DFFB92',
                                  'serial_number': 'DFFB92',
                                  'system_mac_address': None},
                'DFFC02': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4058 NORFOLK ST (KUEPFER RAYMOND K)',
                                  'serial': 'DFFC02',
                                  'serial_number': 'DFFC02',
                                  'system_mac_address': None},
                'DFFC32': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4851 LINE 67 (KUEPFER BRIAN)',
                                  'serial': 'DFFC32',
                                  'serial_number': 'DFFC32',
                                  'system_mac_address': None},
                'DFFC70': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4056 WATERLOO ST (GIESBRECHT ABRAM & '
                                          'KATHARINA)',
                                  'serial': 'DFFC70',
                                  'serial_number': 'DFFC70',
                                  'system_mac_address': None},
                'DFFD38': {       'dev_macs': ['84:d3:43:6d:ec:2f'],
                                  'ip_address': None,
                                  'name': '4066 WATERLOO ST (BARRY ALEXANDER)',
                                  'serial': 'DFFD38',
                                  'serial_number': 'DFFD38',
                                  'system_mac_address': None},
                'DFFD90': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4079 PRINCESS ST (MAGDANZ VERONIKA)',
                                  'serial': 'DFFD90',
                                  'serial_number': 'DFFD90',
                                  'system_mac_address': None},
                'DFFD9C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4094 PRINCESS ST (DISTEFAN RIA)',
                                  'serial': 'DFFD9C',
                                  'serial_number': 'DFFD9C',
                                  'system_mac_address': None},
                'DFFDA2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6991 CHURCH ST (TOMBLIN PAUL)',
                                  'serial': 'DFFDA2',
                                  'serial_number': 'DFFDA2',
                                  'system_mac_address': None},
                'DFFDB2': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6992 MILLBANK MAIN ST (BULMAR WYATT)',
                                  'serial': 'DFFDB2',
                                  'serial_number': 'DFFDB2',
                                  'system_mac_address': None},
                'DFFDB9': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4088 PERTH LINE 72 (MORNINGSTAR CHRISTIAN '
                                          'SCHOOL)',
                                  'serial': 'DFFDB9',
                                  'serial_number': 'DFFDB9',
                                  'system_mac_address': None},
                'DFFDD7': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4090 HILLCREST ST (WAGLER A M)',
                                  'serial': 'DFFDD7',
                                  'serial_number': 'DFFDD7',
                                  'system_mac_address': None},
                'DFFDE0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4094 HILLCREST ST MILLBANK (LINGELBACH '
                                          'BRITTANY)',
                                  'serial': 'DFFDE0',
                                  'serial_number': 'DFFDE0',
                                  'system_mac_address': None},
                'DFFDE8': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '97 PUGH ST (SKOMRA MARGARET)',
                                  'serial': 'DFFDE8',
                                  'serial_number': 'DFFDE8',
                                  'system_mac_address': None},
                'DFFDFA': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '2 MAIN ST N (PHYSIOTHERAPY ALLIANCE)',
                                  'serial': 'DFFDFA',
                                  'serial_number': 'DFFDFA',
                                  'system_mac_address': None},
                'DFFDFF': {       'dev_macs': [       'e4:6c:d1:4e:eb:33',
                                                      'e4:6c:d1:4e:e8:52',
                                                      'e4:6c:d1:4e:ee:a3',
                                                      'e4:6c:d1:4e:f1:63'],
                                  'ip_address': None,
                                  'name': 'Inventory (Room)',
                                  'serial': 'DFFDFF',
                                  'serial_number': 'DFFDFF',
                                  'system_mac_address': None},
                'DFFE01': {       'dev_macs': ['84:d3:43:0f:bc:30'],
                                  'ip_address': None,
                                  'name': '7 SCHADE ST (BAST SANDI)',
                                  'serial': 'DFFE01',
                                  'serial_number': 'DFFE01',
                                  'system_mac_address': None},
                'DFFE0C': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6967 PERTH ST (BRENNEMAN JEREMY & BEVERLEY)',
                                  'serial': 'DFFE0C',
                                  'serial_number': 'DFFE0C',
                                  'system_mac_address': None},
                'DFFE10': {       'dev_macs': ['18:fd:74:1f:86:f6'],
                                  'ip_address': None,
                                  'name': '5208 LINE 58 (GINGRICH MIKE)',
                                  'serial': 'DFFE10',
                                  'serial_number': 'DFFE10',
                                  'system_mac_address': None},
                'DFFE12': {       'dev_macs': ['b8:94:70:bc:b5:41'],
                                  'ip_address': None,
                                  'name': '5518 ROAD 144 (PAXTON WAYNE & JUDY)',
                                  'serial': 'DFFE12',
                                  'serial_number': 'DFFE12',
                                  'system_mac_address': None},
                'DFFE27': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6973 CHURCH ST (HANDORF BEN)',
                                  'serial': 'DFFE27',
                                  'serial_number': 'DFFE27',
                                  'system_mac_address': None},
                'DFFE2D': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '6969 PERTH ST (WAGLER PAUL & DARLENE)',
                                  'serial': 'DFFE2D',
                                  'serial_number': 'DFFE2D',
                                  'system_mac_address': None},
                'DFFE31': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4052 NORFOLK ST (KUEPFER NELSON)',
                                  'serial': 'DFFE31',
                                  'serial_number': 'DFFE31',
                                  'system_mac_address': None},
                'DFFE3F': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4091 PRINCESS ST (REIMER DARREN & LISA)',
                                  'serial': 'DFFE3F',
                                  'serial_number': 'DFFE3F',
                                  'system_mac_address': None},
                'DFFE41': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4059 PRINCESS ST (BUECKERT HENRY & ANNA)',
                                  'serial': 'DFFE41',
                                  'serial_number': 'DFFE41',
                                  'system_mac_address': None},
                'DFFE42': {       'dev_macs': ['44:65:7f:89:0e:2b'],
                                  'ip_address': None,
                                  'name': '124 Maddison St. East, Monkton (Vanessa '
                                          'Schiestel 10435517)',
                                  'serial': 'DFFE42',
                                  'serial_number': 'DFFE42',
                                  'system_mac_address': None},
                'DFFE44': {       'dev_macs': ['ec:4f:82:2c:71:b1'],
                                  'ip_address': None,
                                  'name': '2 MAIN ST N (SUTHERLAND RAVEN)',
                                  'serial': 'DFFE44',
                                  'serial_number': 'DFFE44',
                                  'system_mac_address': None},
                'DFFE4A': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '4053 PRINCESS ST (VAN BAKEL BRYAN)',
                                  'serial': 'DFFE4A',
                                  'serial_number': 'DFFE4A',
                                  'system_mac_address': None},
                'EBBDB0': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': '5101 PERTH LINE 55 (VAN DER WAL JOHN)',
                                  'serial': 'EBBDB0',
                                  'serial_number': 'EBBDB0',
                                  'system_mac_address': None},
                'FD873C': {       'dev_macs': ['ec:4f:82:2b:48:d5'],
                                  'ip_address': None,
                                  'name': '5208 Line 60 (Louis Byler)',
                                  'serial': 'FD873C',
                                  'serial_number': 'FD873C',
                                  'system_mac_address': None}
        }
        exp_ne_msg = {
                'data': {       'interfaces': [       {       'admin_up': True,
                                                              'id': 'front',
                                                              'mac_address': '00:02:5d:db:26:8a',
                                                              'name': 'Craft-FE: front',
                                                              'oper_up': True,
                                                              'poller_hash': 'NzE0MTI5MTMwOTJmcm9udA=='},
                                                      {       'admin_up': False,
                                                              'id': 'rear-A',
                                                              'name': 'Craft-FE: '
                                                                      'rear-A',
                                                              'oper_up': False,
                                                              'poller_hash': 'NzE0MTI5MTMwOTJyZWFyLUE='},
                                                      {       'admin_up': False,
                                                              'id': 'rear-B',
                                                              'name': 'Craft-FE: '
                                                                      'rear-B',
                                                              'oper_up': False,
                                                              'poller_hash': 'NzE0MTI5MTMwOTJyZWFyLUI='}],
                                'manufacturer': 'Calix',
                                'model': 'E7-20',
                                'modules': [       {       'id': '1',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '1/1',
                                                                                         'in_discards': '1886',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2747107',
                                                                                         'in_octets': '3654748878252',
                                                                                         'in_ucast_pkts': '16421940592',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '43543160581',
                                                                                                                  'in_ucast_pkts': '4590173464',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '25',
                                                                                                                  'out_octets': '1147333703828',
                                                                                                                  'out_ucast_pkts': '888658367',
                                                                                                                  'poller_hash': 'NDJDQTMwcGVlcmludGZwb24gNTU1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '811310843311',
                                                                                                                  'in_ucast_pkts': '10258512801',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '8150663018680',
                                                                                                                  'out_ucast_pkts': '6113749652',
                                                                                                                  'poller_hash': 'ODcyNEVBcGVlcmludGZwb24gMTQzMTQ3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '89852096030',
                                                                                                                  'in_ucast_pkts': '121120392',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '316429571559',
                                                                                                                  'out_ucast_pkts': '305085255',
                                                                                                                  'poller_hash': 'QzgxRUFEcGVlcmludGZwb24gMTU2ODcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '13429344474',
                                                                                                                  'in_ucast_pkts': '59313160',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '582538260235',
                                                                                                                  'out_ucast_pkts': '474729193',
                                                                                                                  'poller_hash': 'MTI3NDA1NHBlZXJpbnRmcG9uIDI1MTU2OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '23941028851',
                                                                                                                  'in_ucast_pkts': '229513627',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1195536599962',
                                                                                                                  'out_ucast_pkts': '823461452',
                                                                                                                  'poller_hash': 'MTI5NzNFMXBlZXJpbnRmcG9uIDI2Mzg5Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3MzlEOXBlZXJpbnRmcG9uIDI3NzM2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '5940634675',
                                                                                                                  'in_ucast_pkts': '16764845',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '44381762322',
                                                                                                                  'out_ucast_pkts': '40081754',
                                                                                                                  'poller_hash': 'MTI3MzhGM3BlZXJpbnRmcG9uIDI3OTk4NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-00000',
                                                                                         'name': 'gpon '
                                                                                                 '1/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '33879560977',
                                                                                         'out_octets': '118977625294158',
                                                                                         'out_ucast_pkts': '53297262333',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/1',
                                                                                                                 'id': '1/1',
                                                                                                                 'links': [       {       'connected_time': '688333',
                                                                                                                                          'fiber_length': 1340,
                                                                                                                                          'poller_hash': 'NDJDQTMwcGVlcnBvbnBvbiA1NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 2.118},
                                                                                                                                  {       'connected_time': '688334',
                                                                                                                                          'fiber_length': 1122,
                                                                                                                                          'poller_hash': 'ODcyNEVBcGVlcnBvbnBvbiAxNDMxNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 2.118},
                                                                                                                                  {       'connected_time': '24119266',
                                                                                                                                          'fiber_length': 622,
                                                                                                                                          'poller_hash': 'QzgxRUFEcGVlcnBvbnBvbiAxNTY4NzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 2.118},
                                                                                                                                  {       'connected_time': '688325',
                                                                                                                                          'fiber_length': 1370,
                                                                                                                                          'poller_hash': 'MTI3NDA1NHBlZXJwb25wb24gMjUxNTY5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 2.118},
                                                                                                                                  {       'connected_time': '688333',
                                                                                                                                          'fiber_length': 1336,
                                                                                                                                          'poller_hash': 'MTI5NzNFMXBlZXJwb25wb24gMjYzODk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 2.118},
                                                                                                                                  {       'connected_time': '9698186',
                                                                                                                                          'fiber_length': 1639,
                                                                                                                                          'poller_hash': 'MTI3MzlEOXBlZXJwb25wb24gMjc3MzY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 2.118},
                                                                                                                                  {       'connected_time': '688343',
                                                                                                                                          'fiber_length': 1599,
                                                                                                                                          'poller_hash': 'MTI3MzhGM3BlZXJwb25wb24gMjc5OTg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 2.118}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/1',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/2',
                                                                                         'in_discards': '10030',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3187421',
                                                                                         'in_octets': '3874726021771',
                                                                                         'in_ucast_pkts': '13768047395',
                                                                                         'links': [       {       'in_nucast_pkts': '670571',
                                                                                                                  'in_octets': '1099345285121',
                                                                                                                  'in_ucast_pkts': '3136493142',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '89055775',
                                                                                                                  'out_octets': '12246326613504',
                                                                                                                  'out_ucast_pkts': '9561568069',
                                                                                                                  'poller_hash': 'MTI3NDBFRHBlZXJpbnRmcG9uIDI0NjU3MQ=='},
                                                                                                          {       'in_nucast_pkts': '593380',
                                                                                                                  'in_octets': '254211694243',
                                                                                                                  'in_ucast_pkts': '2110349588',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77346422',
                                                                                                                  'out_octets': '12963656033255',
                                                                                                                  'out_ucast_pkts': '8774499208',
                                                                                                                  'poller_hash': 'MTE3REFFOHBlZXJpbnRmcG9uIDI1MDA0NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '141415249998',
                                                                                                                  'in_ucast_pkts': '34710156580',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '2271014758427',
                                                                                                                  'out_ucast_pkts': '1719259161',
                                                                                                                  'poller_hash': 'MTE3RDlERXBlZXJpbnRmcG9uIDI1MDA1NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '51862057841',
                                                                                                                  'in_ucast_pkts': '299859186',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4',
                                                                                                                  'out_octets': '1609948040095',
                                                                                                                  'out_ucast_pkts': '1191846291',
                                                                                                                  'poller_hash': 'MTBBODcxRHBlZXJpbnRmcG9uIDI1MDA5NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '286214671837',
                                                                                                                  'in_ucast_pkts': '813104367',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '3280063497982',
                                                                                                                  'out_ucast_pkts': '2595178367',
                                                                                                                  'poller_hash': 'MTE3REMyN3BlZXJpbnRmcG9uIDI1MTgwNQ=='},
                                                                                                          {       'in_nucast_pkts': '4543',
                                                                                                                  'in_octets': '587983643',
                                                                                                                  'in_ucast_pkts': '1870851',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '541495',
                                                                                                                  'out_octets': '10391722720',
                                                                                                                  'out_ucast_pkts': '7816061',
                                                                                                                  'poller_hash': 'MTE3REMyQXBlZXJpbnRmcG9uIDI1MzQ5Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '12955731561',
                                                                                                                  'in_ucast_pkts': '61565163',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '396613973719',
                                                                                                                  'out_ucast_pkts': '313195627',
                                                                                                                  'poller_hash': 'MTE3REJDQXBlZXJpbnRmcG9uIDI1MzUzNw=='},
                                                                                                          {       'in_nucast_pkts': '5370',
                                                                                                                  'in_octets': '3753327020',
                                                                                                                  'in_ucast_pkts': '7535005',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '541504',
                                                                                                                  'out_octets': '31709746042',
                                                                                                                  'out_ucast_pkts': '25903738',
                                                                                                                  'poller_hash': 'MTE3REFFM3BlZXJpbnRmcG9uIDI1NDU3MQ=='},
                                                                                                          {       'in_nucast_pkts': '9601',
                                                                                                                  'in_octets': '3179935927',
                                                                                                                  'in_ucast_pkts': '16421375',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151535',
                                                                                                                  'out_octets': '112012841839',
                                                                                                                  'out_ucast_pkts': '79426269',
                                                                                                                  'poller_hash': 'MTI5NzRBQ3BlZXJpbnRmcG9uIDI1NjQ0Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '49632428656',
                                                                                                                  'in_ucast_pkts': '62216055',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '100000718199',
                                                                                                                  'out_ucast_pkts': '126863025',
                                                                                                                  'poller_hash': 'MTI5NzQ1N3BlZXJpbnRmcG9uIDI1NjY2Ng=='},
                                                                                                          {       'in_nucast_pkts': '4585',
                                                                                                                  'in_octets': '2566691564',
                                                                                                                  'in_ucast_pkts': '11276635',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '541542',
                                                                                                                  'out_octets': '34862678993',
                                                                                                                  'out_ucast_pkts': '26565068',
                                                                                                                  'poller_hash': 'MTI3NDg3NHBlZXJpbnRmcG9uIDI2NTY0Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '95544726483',
                                                                                                                  'in_ucast_pkts': '179516480',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '976682089103',
                                                                                                                  'out_ucast_pkts': '726155109',
                                                                                                                  'poller_hash': 'MTI3NDRCMnBlZXJpbnRmcG9uIDI2NjMxMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzNCMHBlZXJpbnRmcG9uIDI3MjAzOQ=='},
                                                                                                          {       'in_nucast_pkts': '5196',
                                                                                                                  'in_octets': '4835430372',
                                                                                                                  'in_ucast_pkts': '13887485',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '541539',
                                                                                                                  'out_octets': '95566663649',
                                                                                                                  'out_ucast_pkts': '71523342',
                                                                                                                  'poller_hash': 'MTI3NDdDMXBlZXJpbnRmcG9uIDI3MjI3Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '69931149231',
                                                                                                                  'in_ucast_pkts': '353607186',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1332045539730',
                                                                                                                  'out_ucast_pkts': '978914800',
                                                                                                                  'poller_hash': 'MTI3NDg3MXBlZXJpbnRmcG9uIDI3MjU1Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '28775729668',
                                                                                                                  'in_ucast_pkts': '257200647',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1987845766194',
                                                                                                                  'out_ucast_pkts': '1354257246',
                                                                                                                  'poller_hash': 'MTI3NDhBMXBlZXJpbnRmcG9uIDI3MjU3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '5421787',
                                                                                                                  'in_ucast_pkts': '21112',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '221033891',
                                                                                                                  'out_ucast_pkts': '175166',
                                                                                                                  'poller_hash': 'MTI5NjAyMnBlZXJpbnRmcG9uIDI4NjMzNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '1/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '146550636',
                                                                                         'out_octets': '70261794628565',
                                                                                         'out_ucast_pkts': '51410431473',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/2',
                                                                                                                 'id': '1/2',
                                                                                                                 'links': [       {       'connected_time': '68517793',
                                                                                                                                          'fiber_length': 20771,
                                                                                                                                          'poller_hash': 'MTI3NDBFRHBlZXJwb25wb24gMjQ2NTcx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '64566130',
                                                                                                                                          'fiber_length': 20621,
                                                                                                                                          'poller_hash': 'MTE3REFFOHBlZXJwb25wb24gMjUwMDQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262226',
                                                                                                                                          'fiber_length': 20346,
                                                                                                                                          'poller_hash': 'MTE3RDlERXBlZXJwb25wb24gMjUwMDU0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '538347',
                                                                                                                                          'fiber_length': 22266,
                                                                                                                                          'poller_hash': 'MTBBODcxRHBlZXJwb25wb24gMjUwMDk0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '13204769',
                                                                                                                                          'fiber_length': 20919,
                                                                                                                                          'poller_hash': 'MTE3REMyN3BlZXJwb25wb24gMjUxODA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262228',
                                                                                                                                          'fiber_length': 22072,
                                                                                                                                          'poller_hash': 'MTE3REMyQXBlZXJwb25wb24gMjUzNDk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '28787924',
                                                                                                                                          'fiber_length': 21660,
                                                                                                                                          'poller_hash': 'MTE3REJDQXBlZXJwb25wb24gMjUzNTM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262229',
                                                                                                                                          'fiber_length': 20133,
                                                                                                                                          'poller_hash': 'MTE3REFFM3BlZXJwb25wb24gMjU0NTcx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '538350',
                                                                                                                                          'fiber_length': 22718,
                                                                                                                                          'poller_hash': 'MTI5NzRBQ3BlZXJwb25wb24gMjU2NDQy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262227',
                                                                                                                                          'fiber_length': 21562,
                                                                                                                                          'poller_hash': 'MTI5NzQ1N3BlZXJwb25wb24gMjU2NjY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262226',
                                                                                                                                          'fiber_length': 20065,
                                                                                                                                          'poller_hash': 'MTI3NDg3NHBlZXJwb25wb24gMjY1NjQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262226',
                                                                                                                                          'fiber_length': 19850,
                                                                                                                                          'poller_hash': 'MTI3NDRCMnBlZXJwb25wb24gMjY2MzEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '16333063',
                                                                                                                                          'fiber_length': 19984,
                                                                                                                                          'poller_hash': 'MTI5NzNCMHBlZXJwb25wb24gMjcyMDM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262228',
                                                                                                                                          'fiber_length': 22256,
                                                                                                                                          'poller_hash': 'MTI3NDdDMXBlZXJwb25wb24gMjcyMjcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '262228',
                                                                                                                                          'fiber_length': 21896,
                                                                                                                                          'poller_hash': 'MTI3NDg3MXBlZXJwb25wb24gMjcyNTU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '15479602',
                                                                                                                                          'fiber_length': 18809,
                                                                                                                                          'poller_hash': 'MTI3NDhBMXBlZXJwb25wb24gMjcyNTc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372},
                                                                                                                                  {       'connected_time': '8527',
                                                                                                                                          'fiber_length': 21602,
                                                                                                                                          'poller_hash': 'MTI5NjAyMnBlZXJwb25wb24gMjg2MzM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.372}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/2',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/3',
                                                                                         'in_discards': '6153',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '60236835',
                                                                                         'in_octets': '146642525803087',
                                                                                         'in_ucast_pkts': '153861008447',
                                                                                         'links': [       {       'in_nucast_pkts': '11976',
                                                                                                                  'in_octets': '6354362074',
                                                                                                                  'in_ucast_pkts': '46961627',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464148',
                                                                                                                  'out_octets': '155316913847',
                                                                                                                  'out_ucast_pkts': '111019602',
                                                                                                                  'poller_hash': 'MkIwMUI3cGVlcmludGZwb24gNTE='},
                                                                                                          {       'in_nucast_pkts': '927303',
                                                                                                                  'in_octets': '914699400952',
                                                                                                                  'in_ucast_pkts': '2464535569',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149411972',
                                                                                                                  'out_octets': '12316123095237',
                                                                                                                  'out_ucast_pkts': '9285011437',
                                                                                                                  'poller_hash': 'REZGQjkycGVlcmludGZwb24gNTM='},
                                                                                                          {       'in_nucast_pkts': '2741903',
                                                                                                                  'in_octets': '2416585084',
                                                                                                                  'in_ucast_pkts': '18653020',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2224499440',
                                                                                                                  'out_octets': '2868922608599',
                                                                                                                  'out_ucast_pkts': '73599187',
                                                                                                                  'poller_hash': 'MUM1QkE2cGVlcmludGZwb24gNTg='},
                                                                                                          {       'in_nucast_pkts': '1541',
                                                                                                                  'in_octets': '105610075866366',
                                                                                                                  'in_ucast_pkts': '87108680799',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '107535489',
                                                                                                                  'out_octets': '12032421437057',
                                                                                                                  'out_ucast_pkts': '52934690405',
                                                                                                                  'poller_hash': 'MTYwNjQyOHBlZXJpbnRmcG9uIDU5'},
                                                                                                          {       'in_nucast_pkts': '917679',
                                                                                                                  'in_octets': '364800707911',
                                                                                                                  'in_ucast_pkts': '2052409027',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149410495',
                                                                                                                  'out_octets': '8664543806430',
                                                                                                                  'out_ucast_pkts': '6432738827',
                                                                                                                  'poller_hash': 'MUQwMzNGcGVlcmludGZwb24gNjU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '452346816936',
                                                                                                                  'in_ucast_pkts': '74539523191',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '4407432749238',
                                                                                                                  'out_ucast_pkts': '3288162902',
                                                                                                                  'poller_hash': 'MUQwNTA1cGVlcmludGZwb24gNjY='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '73943621278',
                                                                                                                  'in_ucast_pkts': '603235882',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '4145346202955',
                                                                                                                  'out_ucast_pkts': '2841598088',
                                                                                                                  'poller_hash': 'MUQwNjhBcGVlcmludGZwb24gNzE='},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '2880364493',
                                                                                                                  'in_ucast_pkts': '15934851',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464249',
                                                                                                                  'out_octets': '78110485530',
                                                                                                                  'out_ucast_pkts': '57736170',
                                                                                                                  'poller_hash': 'MkJEMjZGcGVlcmludGZwb24gMjUz'},
                                                                                                          {       'in_nucast_pkts': '12',
                                                                                                                  'in_octets': '31223212088',
                                                                                                                  'in_ucast_pkts': '159275601',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464145',
                                                                                                                  'out_octets': '692089196995',
                                                                                                                  'out_ucast_pkts': '488716889',
                                                                                                                  'poller_hash': 'MUQwNjk1cGVlcmludGZwb24gMjg0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '56113053889',
                                                                                                                  'in_ucast_pkts': '12966358882',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '214',
                                                                                                                  'out_octets': '346587674878',
                                                                                                                  'out_ucast_pkts': '377951399',
                                                                                                                  'poller_hash': 'MkIwMzY4cGVlcmludGZwb24gNDIy'},
                                                                                                          {       'in_nucast_pkts': '9028',
                                                                                                                  'in_octets': '1278341985153',
                                                                                                                  'in_ucast_pkts': '5311681210',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149457820',
                                                                                                                  'out_octets': '17748755111674',
                                                                                                                  'out_ucast_pkts': '13097536764',
                                                                                                                  'poller_hash': 'MkIwNDRDcGVlcmludGZwb24gNDI5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '99755286517',
                                                                                                                  'in_ucast_pkts': '488590493',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '3041233838786',
                                                                                                                  'out_ucast_pkts': '10023435811',
                                                                                                                  'poller_hash': 'MjA1RTAzcGVlcmludGZwb24gNDQ2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '471054396504',
                                                                                                                  'in_ucast_pkts': '4207853903',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '14346508048285',
                                                                                                                  'out_ucast_pkts': '10087272070',
                                                                                                                  'poller_hash': 'NDJCOUI2cGVlcmludGZwb24gNDkx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzVGOUUwcGVlcmludGZwb24gNTE2'},
                                                                                                          {       'in_nucast_pkts': '14882',
                                                                                                                  'in_octets': '6881749883',
                                                                                                                  'in_ucast_pkts': '44945106',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464417',
                                                                                                                  'out_octets': '197250513106',
                                                                                                                  'out_ucast_pkts': '146960407',
                                                                                                                  'poller_hash': 'QTlDNkZBcGVlcmludGZwb24gMTQ3ODU0'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '1/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '77975867050',
                                                                                         'out_octets': '262614727929200',
                                                                                         'out_ucast_pkts': '175425730371',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/3',
                                                                                                                 'id': '1/3',
                                                                                                                 'links': [       {       'connected_time': '688339',
                                                                                                                                          'fiber_length': 1214,
                                                                                                                                          'poller_hash': 'MkIwMUI3cGVlcnBvbnBvbiA1MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '21815840',
                                                                                                                                          'fiber_length': 1282,
                                                                                                                                          'poller_hash': 'REZGQjkycGVlcnBvbnBvbiA1Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '17711637',
                                                                                                                                          'fiber_length': 1245,
                                                                                                                                          'poller_hash': 'MUM1QkE2cGVlcnBvbnBvbiA1OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '28789464',
                                                                                                                                          'fiber_length': 1245,
                                                                                                                                          'poller_hash': 'MTYwNjQyOHBlZXJwb25wb24gNTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '21816621',
                                                                                                                                          'fiber_length': 1304,
                                                                                                                                          'poller_hash': 'MUQwMzNGcGVlcnBvbnBvbiA2NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '21541485',
                                                                                                                                          'fiber_length': 1303,
                                                                                                                                          'poller_hash': 'MUQwNTA1cGVlcnBvbnBvbiA2Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '30482265',
                                                                                                                                          'fiber_length': 1370,
                                                                                                                                          'poller_hash': 'MUQwNjhBcGVlcnBvbnBvbiA3MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '688340',
                                                                                                                                          'fiber_length': 1354,
                                                                                                                                          'poller_hash': 'MkJEMjZGcGVlcnBvbnBvbiAyNTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '688359',
                                                                                                                                          'fiber_length': 1301,
                                                                                                                                          'poller_hash': 'MUQwNjk1cGVlcnBvbnBvbiAyODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '688338',
                                                                                                                                          'fiber_length': 1205,
                                                                                                                                          'poller_hash': 'MkIwMzY4cGVlcnBvbnBvbiA0MjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '21815431',
                                                                                                                                          'fiber_length': 1163,
                                                                                                                                          'poller_hash': 'MkIwNDRDcGVlcnBvbnBvbiA0Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '9448326',
                                                                                                                                          'fiber_length': 1377,
                                                                                                                                          'poller_hash': 'MjA1RTAzcGVlcnBvbnBvbiA0NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '2146718',
                                                                                                                                          'fiber_length': 1328,
                                                                                                                                          'poller_hash': 'NDJCOUI2cGVlcnBvbnBvbiA0OTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '67790133',
                                                                                                                                          'fiber_length': 1193,
                                                                                                                                          'poller_hash': 'MzVGOUUwcGVlcnBvbnBvbiA1MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739},
                                                                                                                                  {       'connected_time': '688351',
                                                                                                                                          'fiber_length': 1183,
                                                                                                                                          'poller_hash': 'QTlDNkZBcGVlcnBvbnBvbiAxNDc4NTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.739}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/3',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/4',
                                                                                         'in_discards': '9473',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '12281782',
                                                                                         'in_octets': '4515442372812',
                                                                                         'in_ucast_pkts': '24085688123',
                                                                                         'links': [       {       'in_nucast_pkts': '463446',
                                                                                                                  'in_octets': '158186346784',
                                                                                                                  'in_ucast_pkts': '1063847162',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58266095',
                                                                                                                  'out_octets': '6525792929733',
                                                                                                                  'out_ucast_pkts': '4552215267',
                                                                                                                  'poller_hash': 'MUQwNTA2cGVlcmludGZwb24gNjA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '179741728032',
                                                                                                                  'in_ucast_pkts': '1222960041',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '4341598309688',
                                                                                                                  'out_ucast_pkts': '3169524762',
                                                                                                                  'poller_hash': 'MUQwNEY3cGVlcmludGZwb24gNjE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '78129381355',
                                                                                                                  'in_ucast_pkts': '599263128',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '2311647974021',
                                                                                                                  'out_ucast_pkts': '1683981456',
                                                                                                                  'poller_hash': 'Mjk2RkM3cGVlcmludGZwb24gMTA5'},
                                                                                                          {       'in_nucast_pkts': '289819',
                                                                                                                  'in_octets': '63887590398',
                                                                                                                  'in_ucast_pkts': '734079880',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '36517218',
                                                                                                                  'out_octets': '3781335881931',
                                                                                                                  'out_ucast_pkts': '2525156997',
                                                                                                                  'poller_hash': 'MUQwNTAxcGVlcmludGZwb24gMTYw'},
                                                                                                          {       'in_nucast_pkts': '14728',
                                                                                                                  'in_octets': '6870349557',
                                                                                                                  'in_ucast_pkts': '52466482',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464137',
                                                                                                                  'out_octets': '186854302522',
                                                                                                                  'out_ucast_pkts': '136376174',
                                                                                                                  'poller_hash': 'OTU2RkFFcGVlcmludGZwb24gMjA0'},
                                                                                                          {       'in_nucast_pkts': '2861729',
                                                                                                                  'in_octets': '191671062003',
                                                                                                                  'in_ucast_pkts': '2613210116',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '199404850',
                                                                                                                  'out_octets': '14602589163063',
                                                                                                                  'out_ucast_pkts': '10340161682',
                                                                                                                  'poller_hash': 'MkIwMTc2cGVlcmludGZwb24gNDA5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '129576721655',
                                                                                                                  'in_ucast_pkts': '545239413',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '24',
                                                                                                                  'out_octets': '2981215854333',
                                                                                                                  'out_ucast_pkts': '2131144503',
                                                                                                                  'poller_hash': 'MkIwNDNGcGVlcmludGZwb24gNDI1'},
                                                                                                          {       'in_nucast_pkts': '75',
                                                                                                                  'in_octets': '153963567826',
                                                                                                                  'in_ucast_pkts': '409357579',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15193058',
                                                                                                                  'out_octets': '1844170042210',
                                                                                                                  'out_ucast_pkts': '1287954961',
                                                                                                                  'poller_hash': 'MkIwNDYycGVlcmludGZwb24gNDI4'},
                                                                                                          {       'in_nucast_pkts': '231',
                                                                                                                  'in_octets': '62444615293',
                                                                                                                  'in_ucast_pkts': '520375644',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '69117245',
                                                                                                                  'out_octets': '2865613830770',
                                                                                                                  'out_ucast_pkts': '1951612121',
                                                                                                                  'poller_hash': 'MzVFOTZDcGVlcmludGZwb24gNDMy'},
                                                                                                          {       'in_nucast_pkts': '42',
                                                                                                                  'in_octets': '16115801888',
                                                                                                                  'in_ucast_pkts': '60127040',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120144',
                                                                                                                  'out_octets': '313170434269',
                                                                                                                  'out_ucast_pkts': '244097670',
                                                                                                                  'poller_hash': 'NDNCMzdEcGVlcmludGZwb24gNzY3'},
                                                                                                          {       'in_nucast_pkts': '218524',
                                                                                                                  'in_octets': '187964415910',
                                                                                                                  'in_ucast_pkts': '1072731480',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19805635',
                                                                                                                  'out_octets': '4215902741878',
                                                                                                                  'out_ucast_pkts': '2985392652',
                                                                                                                  'poller_hash': 'MkZFQUZBcGVlcmludGZwb24gMjA1NDA2'},
                                                                                                          {       'in_nucast_pkts': '41780097',
                                                                                                                  'in_octets': '388122268558',
                                                                                                                  'in_ucast_pkts': '1272309578',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22291979323',
                                                                                                                  'out_octets': '33434616137731',
                                                                                                                  'out_ucast_pkts': '3101326650',
                                                                                                                  'poller_hash': 'QzgxRTYzcGVlcmludGZwb24gMjA5MTM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEE3MHBlZXJpbnRmcG9uIDIzMzY5Mw=='},
                                                                                                          {       'in_nucast_pkts': '749762',
                                                                                                                  'in_octets': '36223637146',
                                                                                                                  'in_ucast_pkts': '471787326',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '163492831',
                                                                                                                  'out_octets': '2765176744157',
                                                                                                                  'out_ucast_pkts': '1926630150',
                                                                                                                  'poller_hash': 'MTI3MERFNnBlZXJpbnRmcG9uIDI3OTE1Nw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '1/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '68536073872',
                                                                                         'out_octets': '196651151840767',
                                                                                         'out_ucast_pkts': '75047245471',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/4',
                                                                                                                 'id': '1/4',
                                                                                                                 'links': [       {       'connected_time': '24544077',
                                                                                                                                          'fiber_length': 1251,
                                                                                                                                          'poller_hash': 'MUQwNTA2cGVlcnBvbnBvbiA2MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '21816522',
                                                                                                                                          'fiber_length': 1250,
                                                                                                                                          'poller_hash': 'MUQwNEY3cGVlcnBvbnBvbiA2MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '15498467',
                                                                                                                                          'fiber_length': 994,
                                                                                                                                          'poller_hash': 'Mjk2RkM3cGVlcnBvbnBvbiAxMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '16507268',
                                                                                                                                          'fiber_length': 1172,
                                                                                                                                          'poller_hash': 'MUQwNTAxcGVlcnBvbnBvbiAxNjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '688343',
                                                                                                                                          'fiber_length': 955,
                                                                                                                                          'poller_hash': 'OTU2RkFFcGVlcnBvbnBvbiAyMDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '30076165',
                                                                                                                                          'fiber_length': 1147,
                                                                                                                                          'poller_hash': 'MkIwMTc2cGVlcnBvbnBvbiA0MDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '688343',
                                                                                                                                          'fiber_length': 1157,
                                                                                                                                          'poller_hash': 'MkIwNDNGcGVlcnBvbnBvbiA0MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '6487380',
                                                                                                                                          'fiber_length': 851,
                                                                                                                                          'poller_hash': 'MkIwNDYycGVlcnBvbnBvbiA0Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '29381420',
                                                                                                                                          'fiber_length': 1030,
                                                                                                                                          'poller_hash': 'MzVFOTZDcGVlcnBvbnBvbiA0MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '688334',
                                                                                                                                          'fiber_length': 847,
                                                                                                                                          'poller_hash': 'NDNCMzdEcGVlcnBvbnBvbiA3Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '10996353',
                                                                                                                                          'fiber_length': 1019,
                                                                                                                                          'poller_hash': 'MkZFQUZBcGVlcnBvbnBvbiAyMDU0MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '21816835',
                                                                                                                                          'fiber_length': 1162,
                                                                                                                                          'poller_hash': 'QzgxRTYzcGVlcnBvbnBvbiAyMDkxMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '4678524',
                                                                                                                                          'fiber_length': 1006,
                                                                                                                                          'poller_hash': 'MTI3NEE3MHBlZXJwb25wb24gMjMzNjkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647},
                                                                                                                                  {       'connected_time': '7351722',
                                                                                                                                          'fiber_length': 1076,
                                                                                                                                          'poller_hash': 'MTI3MERFNnBlZXJwb25wb24gMjc5MTU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.647}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/4',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/5',
                                                                                         'in_discards': '2488',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1021424',
                                                                                         'in_octets': '5220947861444',
                                                                                         'in_ucast_pkts': '30260621318',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '47297301923',
                                                                                                                  'in_ucast_pkts': '111946135',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '111',
                                                                                                                  'out_octets': '300968323037',
                                                                                                                  'out_ucast_pkts': '310366662',
                                                                                                                  'poller_hash': 'MjA1REY4cGVlcmludGZwb24gMTI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '421453026879',
                                                                                                                  'in_ucast_pkts': '1500968058',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9935722168654',
                                                                                                                  'out_ucast_pkts': '7265312128',
                                                                                                                  'poller_hash': 'ODg5RTdwZWVyaW50ZnBvbiA2OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '542369992698',
                                                                                                                  'in_ucast_pkts': '106235913943',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '10041968279789',
                                                                                                                  'out_ucast_pkts': '7367503586',
                                                                                                                  'poller_hash': 'MUNBQjRBcGVlcmludGZwb24gNzM='},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '1475408762',
                                                                                                                  'in_ucast_pkts': '13443559',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464199',
                                                                                                                  'out_octets': '62259095924',
                                                                                                                  'out_ucast_pkts': '43307763',
                                                                                                                  'poller_hash': 'MkFDNEI2cGVlcmludGZwb24gNzQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '279622632223',
                                                                                                                  'in_ucast_pkts': '1031731518',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '5441823747175',
                                                                                                                  'out_ucast_pkts': '3904607707',
                                                                                                                  'poller_hash': 'MUQwNTAycGVlcmludGZwb24gODA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '37307404101',
                                                                                                                  'in_ucast_pkts': '475499357',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '3040616867677',
                                                                                                                  'out_ucast_pkts': '2050355436',
                                                                                                                  'poller_hash': 'Mjk4MjdDcGVlcmludGZwb24gMTM5'},
                                                                                                          {       'in_nucast_pkts': '41967',
                                                                                                                  'in_octets': '9412763854',
                                                                                                                  'in_ucast_pkts': '49563886',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465360',
                                                                                                                  'out_octets': '278502084128',
                                                                                                                  'out_ucast_pkts': '208066315',
                                                                                                                  'poller_hash': 'MkJEMjQxcGVlcmludGZwb24gMTQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '151366931829',
                                                                                                                  'in_ucast_pkts': '26613710912',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '4269273357183',
                                                                                                                  'out_ucast_pkts': '3111817644',
                                                                                                                  'poller_hash': 'MkIwNDU0cGVlcmludGZwb24gNDI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3207379064',
                                                                                                                  'in_ucast_pkts': '18501115',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '70509937817',
                                                                                                                  'out_ucast_pkts': '64210620',
                                                                                                                  'poller_hash': 'MkIwNDQ0cGVlcmludGZwb24gNDMx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '76837579929',
                                                                                                                  'in_ucast_pkts': '557865808',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '50',
                                                                                                                  'out_octets': '2004026278759',
                                                                                                                  'out_ucast_pkts': '1562216345',
                                                                                                                  'poller_hash': 'MjA1RERFcGVlcmludGZwb24gNDcz'},
                                                                                                          {       'in_nucast_pkts': '256768',
                                                                                                                  'in_octets': '209580692829',
                                                                                                                  'in_ucast_pkts': '257107761',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31045783',
                                                                                                                  'out_octets': '415548364182',
                                                                                                                  'out_ucast_pkts': '446691296',
                                                                                                                  'poller_hash': 'MTE3REJGNHBlZXJpbnRmcG9uIDI1MzUyNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDgxRXBlZXJpbnRmcG9uIDI3Mzg4MQ=='},
                                                                                                          {       'in_nucast_pkts': '60',
                                                                                                                  'in_octets': '13410363296',
                                                                                                                  'in_ucast_pkts': '30653618',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1123734',
                                                                                                                  'out_octets': '99462387213',
                                                                                                                  'out_ucast_pkts': '78667256',
                                                                                                                  'poller_hash': 'MTI3NDgwMXBlZXJpbnRmcG9uIDI3NDQ5NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '33245879400',
                                                                                                                  'in_ucast_pkts': '192362671',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '843574328311',
                                                                                                                  'out_ucast_pkts': '648273917',
                                                                                                                  'poller_hash': 'MTI3MEQ2RXBlZXJpbnRmcG9uIDI3NzAzNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '1/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '3923586126',
                                                                                         'out_octets': '116644258947493',
                                                                                         'out_ucast_pkts': '80223613563',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/5',
                                                                                                                 'id': '1/5',
                                                                                                                 'links': [       {       'connected_time': '688344',
                                                                                                                                          'fiber_length': 1130,
                                                                                                                                          'poller_hash': 'MjA1REY4cGVlcnBvbnBvbiAxMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '688336',
                                                                                                                                          'fiber_length': 888,
                                                                                                                                          'poller_hash': 'ODg5RTdwZWVycG9ucG9uIDY4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '11231777',
                                                                                                                                          'fiber_length': 988,
                                                                                                                                          'poller_hash': 'MUNBQjRBcGVlcnBvbnBvbiA3Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '688348',
                                                                                                                                          'fiber_length': 962,
                                                                                                                                          'poller_hash': 'MkFDNEI2cGVlcnBvbnBvbiA3NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '688349',
                                                                                                                                          'fiber_length': 841,
                                                                                                                                          'poller_hash': 'MUQwNTAycGVlcnBvbnBvbiA4MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '24902851',
                                                                                                                                          'fiber_length': 1361,
                                                                                                                                          'poller_hash': 'Mjk4MjdDcGVlcnBvbnBvbiAxMzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '688345',
                                                                                                                                          'fiber_length': 838,
                                                                                                                                          'poller_hash': 'MkJEMjQxcGVlcnBvbnBvbiAxNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '21815758',
                                                                                                                                          'fiber_length': 848,
                                                                                                                                          'poller_hash': 'MkIwNDU0cGVlcnBvbnBvbiA0MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '21815373',
                                                                                                                                          'fiber_length': 1137,
                                                                                                                                          'poller_hash': 'MkIwNDQ0cGVlcnBvbnBvbiA0MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '688346',
                                                                                                                                          'fiber_length': 843,
                                                                                                                                          'poller_hash': 'MjA1RERFcGVlcnBvbnBvbiA0NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '13642123',
                                                                                                                                          'fiber_length': 1020,
                                                                                                                                          'poller_hash': 'MTE3REJGNHBlZXJwb25wb24gMjUzNTI3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '4587986',
                                                                                                                                          'fiber_length': 1503,
                                                                                                                                          'poller_hash': 'MTI3NDgxRXBlZXJwb25wb24gMjczODgx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '688348',
                                                                                                                                          'fiber_length': 997,
                                                                                                                                          'poller_hash': 'MTI3NDgwMXBlZXJwb25wb24gMjc0NDk0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609},
                                                                                                                                  {       'connected_time': '688343',
                                                                                                                                          'fiber_length': 975,
                                                                                                                                          'poller_hash': 'MTI3MEQ2RXBlZXJwb25wb24gMjc3MDM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.609}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/5',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/6',
                                                                                         'in_discards': '3623',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1375974',
                                                                                         'in_octets': '4800456291403',
                                                                                         'in_ucast_pkts': '18572151691',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '129837250126',
                                                                                                                  'in_ucast_pkts': '294748857',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '678522684122',
                                                                                                                  'out_ucast_pkts': '635337906',
                                                                                                                  'poller_hash': 'QzgzQkM5cGVlcmludGZwb24gMTcxOTI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '392537914599',
                                                                                                                  'in_ucast_pkts': '108641936525',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '74',
                                                                                                                  'out_octets': '4948946299963',
                                                                                                                  'out_ucast_pkts': '3725195165',
                                                                                                                  'poller_hash': 'QzgzMTc0cGVlcmludGZwb24gMTcyMjM5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '88180927156',
                                                                                                                  'in_ucast_pkts': '486466288',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '1678403168165',
                                                                                                                  'out_ucast_pkts': '1259393553',
                                                                                                                  'poller_hash': 'QjA4RUY2cGVlcmludGZwb24gMTc1MTQ2'},
                                                                                                          {       'in_nucast_pkts': '452292',
                                                                                                                  'in_octets': '183323476460',
                                                                                                                  'in_ucast_pkts': '464803530',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '44806989',
                                                                                                                  'out_octets': '1220034012146',
                                                                                                                  'out_ucast_pkts': '1068628776',
                                                                                                                  'poller_hash': 'Q0JEMUJDcGVlcmludGZwb24gMTc1MzYx'},
                                                                                                          {       'in_nucast_pkts': '199457',
                                                                                                                  'in_octets': '17872795505',
                                                                                                                  'in_ucast_pkts': '46241494',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23990096',
                                                                                                                  'out_octets': '165540549854',
                                                                                                                  'out_ucast_pkts': '144515071',
                                                                                                                  'poller_hash': 'QzgzNkM0cGVlcmludGZwb24gMTgyMjcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '28316392064',
                                                                                                                  'in_ucast_pkts': '123791302',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '850946041316',
                                                                                                                  'out_ucast_pkts': '722498072',
                                                                                                                  'poller_hash': 'QzgzQkYwcGVlcmludGZwb24gMTgyMzAy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '86417134426',
                                                                                                                  'in_ucast_pkts': '21921868883',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '2016692764010',
                                                                                                                  'out_ucast_pkts': '1496282246',
                                                                                                                  'poller_hash': 'QzgxRDhBcGVlcmludGZwb24gMTk2OTA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '186214273425',
                                                                                                                  'in_ucast_pkts': '665263535',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1839243489862',
                                                                                                                  'out_ucast_pkts': '1647351579',
                                                                                                                  'poller_hash': 'MTAwQUFGOXBlZXJpbnRmcG9uIDIwMTM1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '114189082198',
                                                                                                                  'in_ucast_pkts': '364936847',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '1834558033898',
                                                                                                                  'out_ucast_pkts': '1488736010',
                                                                                                                  'poller_hash': 'MTE5NEU4N3BlZXJpbnRmcG9uIDIxMjcyMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '201574888377',
                                                                                                                  'in_ucast_pkts': '56876949483',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '5857435651419',
                                                                                                                  'out_ucast_pkts': '4131900180',
                                                                                                                  'poller_hash': 'Q0JEMTYwcGVlcmludGZwb24gMjEzNTE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '488343378505',
                                                                                                                  'in_ucast_pkts': '69583600077',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '3065258784324',
                                                                                                                  'out_ucast_pkts': '2395717967',
                                                                                                                  'poller_hash': 'MTE5NTE0NnBlZXJpbnRmcG9uIDIxNTAyMQ=='},
                                                                                                          {       'in_nucast_pkts': '925310',
                                                                                                                  'in_octets': '1145224937597',
                                                                                                                  'in_ucast_pkts': '4822153098',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '137784697',
                                                                                                                  'out_octets': '15441594055366',
                                                                                                                  'out_ucast_pkts': '11455853966',
                                                                                                                  'poller_hash': 'MTE5NTE5OXBlZXJpbnRmcG9uIDIxNTAzMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '422035260559',
                                                                                                                  'in_ucast_pkts': '32142966629',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '4512030014842',
                                                                                                                  'out_ucast_pkts': '4308826404',
                                                                                                                  'poller_hash': 'MTE5NTE2MXBlZXJpbnRmcG9uIDIxNTY2NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '112332165902',
                                                                                                                  'in_ucast_pkts': '889776891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '72',
                                                                                                                  'out_octets': '5184818483767',
                                                                                                                  'out_ucast_pkts': '3602206490',
                                                                                                                  'poller_hash': 'MTI3NTFERHBlZXJpbnRmcG9uIDIyNDg3Mg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '26496198028',
                                                                                         'out_octets': '103168349479835',
                                                                                         'out_ucast_pkts': '52261162366',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/6',
                                                                                                                 'id': '1/6',
                                                                                                                 'links': [       {       'connected_time': '15498459',
                                                                                                                                          'fiber_length': 12050,
                                                                                                                                          'poller_hash': 'QzgzQkM5cGVlcnBvbnBvbiAxNzE5Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '688348',
                                                                                                                                          'fiber_length': 11718,
                                                                                                                                          'poller_hash': 'QzgzMTc0cGVlcnBvbnBvbiAxNzIyMzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '687625',
                                                                                                                                          'fiber_length': 11694,
                                                                                                                                          'poller_hash': 'QjA4RUY2cGVlcnBvbnBvbiAxNzUxNDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '69480770',
                                                                                                                                          'fiber_length': 11800,
                                                                                                                                          'poller_hash': 'Q0JEMUJDcGVlcnBvbnBvbiAxNzUzNjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '21816857',
                                                                                                                                          'fiber_length': 11627,
                                                                                                                                          'poller_hash': 'QzgzNkM0cGVlcnBvbnBvbiAxODIyNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '21816094',
                                                                                                                                          'fiber_length': 11750,
                                                                                                                                          'poller_hash': 'QzgzQkYwcGVlcnBvbnBvbiAxODIzMDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '688348',
                                                                                                                                          'fiber_length': 11613,
                                                                                                                                          'poller_hash': 'QzgxRDhBcGVlcnBvbnBvbiAxOTY5MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '47827',
                                                                                                                                          'fiber_length': 11660,
                                                                                                                                          'poller_hash': 'MTAwQUFGOXBlZXJwb25wb24gMjAxMzU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '21792971',
                                                                                                                                          'fiber_length': 11745,
                                                                                                                                          'poller_hash': 'MTE5NEU4N3BlZXJwb25wb24gMjEyNzIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '28863',
                                                                                                                                          'fiber_length': 12028,
                                                                                                                                          'poller_hash': 'Q0JEMTYwcGVlcnBvbnBvbiAyMTM1MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '12765140',
                                                                                                                                          'fiber_length': 11644,
                                                                                                                                          'poller_hash': 'MTE5NTE0NnBlZXJwb25wb24gMjE1MDIx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '19970731',
                                                                                                                                          'fiber_length': 11916,
                                                                                                                                          'poller_hash': 'MTE5NTE5OXBlZXJwb25wb24gMjE1MDMx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '19801590',
                                                                                                                                          'fiber_length': 11966,
                                                                                                                                          'poller_hash': 'MTE5NTE2MXBlZXJwb25wb24gMjE1NjY0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163},
                                                                                                                                  {       'connected_time': '68521034',
                                                                                                                                          'fiber_length': 11981,
                                                                                                                                          'poller_hash': 'MTI3NTFERHBlZXJwb25wb24gMjI0ODcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.163}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/6',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/7',
                                                                                         'in_discards': '77011',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1983851',
                                                                                         'in_octets': '29412544604915',
                                                                                         'in_ucast_pkts': '49651173250',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '27406028917',
                                                                                                                  'in_ucast_pkts': '114944376',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '152',
                                                                                                                  'out_octets': '226121164658',
                                                                                                                  'out_ucast_pkts': '252549702',
                                                                                                                  'poller_hash': 'QzgzQkFGcGVlcmludGZwb24gMTcxOTM4'},
                                                                                                          {       'in_nucast_pkts': '981641',
                                                                                                                  'in_octets': '2258603719014',
                                                                                                                  'in_ucast_pkts': '5963475083',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149352519',
                                                                                                                  'out_octets': '13585090105034',
                                                                                                                  'out_ucast_pkts': '11883156656',
                                                                                                                  'poller_hash': 'REZGRDM4cGVlcmludGZwb24gMTcyNjEw'},
                                                                                                          {       'in_nucast_pkts': '4129',
                                                                                                                  'in_octets': '57441885',
                                                                                                                  'in_ucast_pkts': '176802',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465893',
                                                                                                                  'out_octets': '353484219',
                                                                                                                  'out_ucast_pkts': '346087',
                                                                                                                  'poller_hash': 'QzgzNjEwcGVlcmludGZwb24gMTc3MDI1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '64636598501',
                                                                                                                  'in_ucast_pkts': '248477589',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '536529126999',
                                                                                                                  'out_ucast_pkts': '551210987',
                                                                                                                  'poller_hash': 'QzgzNjkxcGVlcmludGZwb24gMTgwMjgy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '66181739690',
                                                                                                                  'in_ucast_pkts': '17302872467',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '322163000643',
                                                                                                                  'out_ucast_pkts': '321808779',
                                                                                                                  'poller_hash': 'QzgzQjZDcGVlcmludGZwb24gMTgxMDAx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '222688948643',
                                                                                                                  'in_ucast_pkts': '26549638688',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '74',
                                                                                                                  'out_octets': '2692933599529',
                                                                                                                  'out_ucast_pkts': '2120661577',
                                                                                                                  'poller_hash': 'QzgzMEY5cGVlcmludGZwb24gMTgzNjU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '505822873053',
                                                                                                                  'in_ucast_pkts': '14951399657',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '7370525443020',
                                                                                                                  'out_ucast_pkts': '5640560103',
                                                                                                                  'poller_hash': 'REZGREEycGVlcmludGZwb24gMTgzNjY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '113765594256',
                                                                                                                  'in_ucast_pkts': '560177583',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '2316304197789',
                                                                                                                  'out_ucast_pkts': '1908363399',
                                                                                                                  'poller_hash': 'REZGQzcwcGVlcmludGZwb24gMTg1MjU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1534004368891',
                                                                                                                  'in_ucast_pkts': '44851822037',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148',
                                                                                                                  'out_octets': '1721015254856',
                                                                                                                  'out_ucast_pkts': '1628818068',
                                                                                                                  'poller_hash': 'QzgxQzczcGVlcmludGZwb24gMTg1NDk3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '169112998667',
                                                                                                                  'in_ucast_pkts': '168574049',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '60741735729',
                                                                                                                  'out_ucast_pkts': '127997431',
                                                                                                                  'poller_hash': 'QzgxOUREcGVlcmludGZwb24gMTg1NzAx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '371241092760',
                                                                                                                  'in_ucast_pkts': '100735757227',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '74',
                                                                                                                  'out_octets': '14243865456975',
                                                                                                                  'out_ucast_pkts': '9773719799',
                                                                                                                  'poller_hash': 'REZGRTI3cGVlcmludGZwb24gMTg2NjA3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '407593910183',
                                                                                                                  'in_ucast_pkts': '78875691509',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '40638813257',
                                                                                                                  'out_octets': '60001615027638',
                                                                                                                  'out_ucast_pkts': '4038466742',
                                                                                                                  'poller_hash': 'QzgxREE1cGVlcmludGZwb24gMTkyODY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '100566926582',
                                                                                                                  'in_ucast_pkts': '1287234984',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144',
                                                                                                                  'out_octets': '5346872929929',
                                                                                                                  'out_ucast_pkts': '3571137124',
                                                                                                                  'poller_hash': 'QzgxRTk2cGVlcmludGZwb24gMjAwMzg1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '577135536638',
                                                                                                                  'in_ucast_pkts': '1383781709',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '74',
                                                                                                                  'out_octets': '5211627569264',
                                                                                                                  'out_ucast_pkts': '3813577860',
                                                                                                                  'poller_hash': 'QzgxRUFDcGVlcmludGZwb24gMjAxNjU0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '246737583297',
                                                                                                                  'in_ucast_pkts': '129189300924',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '99',
                                                                                                                  'out_octets': '995607284861',
                                                                                                                  'out_ucast_pkts': '872432682',
                                                                                                                  'poller_hash': 'QzgxOTZBcGVlcmludGZwb24gMjAyNTk1'},
                                                                                                          {       'in_nucast_pkts': '378844',
                                                                                                                  'in_octets': '141974769016',
                                                                                                                  'in_ucast_pkts': '841270511',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '46232990',
                                                                                                                  'out_octets': '2225715288393',
                                                                                                                  'out_ucast_pkts': '1698640027',
                                                                                                                  'poller_hash': 'QzgxRTExcGVlcmludGZwb24gMjAyNjc0'},
                                                                                                          {       'in_nucast_pkts': '12117',
                                                                                                                  'in_octets': '13949976607',
                                                                                                                  'in_ucast_pkts': '49923038',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464236',
                                                                                                                  'out_octets': '157425846690',
                                                                                                                  'out_ucast_pkts': '129190380',
                                                                                                                  'poller_hash': 'MTE5NTEzRnBlZXJpbnRmcG9uIDIxMjQxNg=='},
                                                                                                          {       'in_nucast_pkts': '6',
                                                                                                                  'in_octets': '13055530373067',
                                                                                                                  'in_ucast_pkts': '13609361751',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '35094860',
                                                                                                                  'out_octets': '5799129896248',
                                                                                                                  'out_ucast_pkts': '8737277792',
                                                                                                                  'poller_hash': 'MTI3NDA2NHBlZXJpbnRmcG9uIDI0NjkwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REFDMXBlZXJpbnRmcG9uIDI1MTAyNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzRFQ3BlZXJpbnRmcG9uIDI1NjgzNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '92409522',
                                                                                                                  'in_ucast_pkts': '200582',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1433884571',
                                                                                                                  'out_ucast_pkts': '5623640',
                                                                                                                  'poller_hash': 'MTI5NjAyRXBlZXJpbnRmcG9uIDI4MDQwNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2104009264',
                                                                                                                  'in_ucast_pkts': '6686474',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '38985987894',
                                                                                                                  'out_ucast_pkts': '31969022',
                                                                                                                  'poller_hash': 'MTI5NjA1QXBlZXJpbnRmcG9uIDI4MjMwMw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '72602890608',
                                                                                         'out_octets': '208605481237532',
                                                                                         'out_ucast_pkts': '91740747261',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/7',
                                                                                                                 'id': '1/7',
                                                                                                                 'links': [       {       'connected_time': '3969452',
                                                                                                                                          'fiber_length': 11832,
                                                                                                                                          'poller_hash': 'QzgzQkFGcGVlcnBvbnBvbiAxNzE5Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '21816175',
                                                                                                                                          'fiber_length': 11627,
                                                                                                                                          'poller_hash': 'REZGRDM4cGVlcnBvbnBvbiAxNzI2MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '688348',
                                                                                                                                          'fiber_length': 11846,
                                                                                                                                          'poller_hash': 'QzgzNjEwcGVlcnBvbnBvbiAxNzcwMjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '27666955',
                                                                                                                                          'fiber_length': 11784,
                                                                                                                                          'poller_hash': 'QzgzNjkxcGVlcnBvbnBvbiAxODAyODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '21816725',
                                                                                                                                          'fiber_length': 11426,
                                                                                                                                          'poller_hash': 'QzgzQjZDcGVlcnBvbnBvbiAxODEwMDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '688348',
                                                                                                                                          'fiber_length': 11445,
                                                                                                                                          'poller_hash': 'QzgzMEY5cGVlcnBvbnBvbiAxODM2NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '74480379',
                                                                                                                                          'fiber_length': 11424,
                                                                                                                                          'poller_hash': 'REZGREEycGVlcnBvbnBvbiAxODM2NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '69274060',
                                                                                                                                          'fiber_length': 11600,
                                                                                                                                          'poller_hash': 'REZGQzcwcGVlcnBvbnBvbiAxODUyNTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '21816941',
                                                                                                                                          'fiber_length': 11543,
                                                                                                                                          'poller_hash': 'QzgxQzczcGVlcnBvbnBvbiAxODU0OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '6774594',
                                                                                                                                          'fiber_length': 11600,
                                                                                                                                          'poller_hash': 'QzgxOUREcGVlcnBvbnBvbiAxODU3MDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '688351',
                                                                                                                                          'fiber_length': 11602,
                                                                                                                                          'poller_hash': 'REZGRTI3cGVlcnBvbnBvbiAxODY2MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '1026215',
                                                                                                                                          'fiber_length': 11451,
                                                                                                                                          'poller_hash': 'QzgxREE1cGVlcnBvbnBvbiAxOTI4NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '21815653',
                                                                                                                                          'fiber_length': 11637,
                                                                                                                                          'poller_hash': 'QzgxRTk2cGVlcnBvbnBvbiAyMDAzODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '523084',
                                                                                                                                          'fiber_length': 11803,
                                                                                                                                          'poller_hash': 'QzgxRUFDcGVlcnBvbnBvbiAyMDE2NTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '688353',
                                                                                                                                          'fiber_length': 11573,
                                                                                                                                          'poller_hash': 'QzgxOTZBcGVlcnBvbnBvbiAyMDI1OTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '21815627',
                                                                                                                                          'fiber_length': 11546,
                                                                                                                                          'poller_hash': 'QzgxRTExcGVlcnBvbnBvbiAyMDI2NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '688354',
                                                                                                                                          'fiber_length': 11624,
                                                                                                                                          'poller_hash': 'MTE5NTEzRnBlZXJwb25wb24gMjEyNDE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '15498714',
                                                                                                                                          'fiber_length': 11849,
                                                                                                                                          'poller_hash': 'MTI3NDA2NHBlZXJwb25wb24gMjQ2OTAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '63443605',
                                                                                                                                          'fiber_length': 11556,
                                                                                                                                          'poller_hash': 'MTE3REFDMXBlZXJwb25wb24gMjUxMDI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '24110006',
                                                                                                                                          'fiber_length': 11854,
                                                                                                                                          'poller_hash': 'MTI5NzRFQ3BlZXJwb25wb24gMjU2ODM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '4257088',
                                                                                                                                          'fiber_length': 11703,
                                                                                                                                          'poller_hash': 'MTI5NjAyRXBlZXJwb25wb24gMjgwNDA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002},
                                                                                                                                  {       'connected_time': '688352',
                                                                                                                                          'fiber_length': 11508,
                                                                                                                                          'poller_hash': 'MTI5NjA1QXBlZXJwb25wb24gMjgyMzAz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.002}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/7',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/8',
                                                                                         'in_discards': '2',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '45891',
                                                                                         'in_octets': '840837418843',
                                                                                         'in_ucast_pkts': '8165828988',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '44766556446',
                                                                                                                  'in_ucast_pkts': '587943962',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '52',
                                                                                                                  'out_octets': '3742026782464',
                                                                                                                  'out_ucast_pkts': '2492691468',
                                                                                                                  'poller_hash': 'MjA1RDk3cGVlcmludGZwb24gNTI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96633036714',
                                                                                                                  'in_ucast_pkts': '5329082604',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144',
                                                                                                                  'out_octets': '5208321421769',
                                                                                                                  'out_ucast_pkts': '7807743112',
                                                                                                                  'poller_hash': 'MzQ5RTg3cGVlcmludGZwb24gNjkw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MjA1QjM4cGVlcmludGZwb24gNjk0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '102496007989',
                                                                                                                  'in_ucast_pkts': '22318210147',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '98',
                                                                                                                  'out_octets': '4909918565667',
                                                                                                                  'out_ucast_pkts': '3354070159',
                                                                                                                  'poller_hash': 'MzQ5RTUxcGVlcmludGZwb24gNzE1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NjQwQUZGcGVlcmludGZwb24gMTQxNzc0'},
                                                                                                          {       'in_nucast_pkts': '767296',
                                                                                                                  'in_octets': '118391865171',
                                                                                                                  'in_ucast_pkts': '872760014',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4815944432',
                                                                                                                  'out_octets': '10601189274562',
                                                                                                                  'out_ucast_pkts': '2865522368',
                                                                                                                  'poller_hash': 'ODY3OEEycGVlcmludGZwb24gMTcyNTk5'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '1/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '31433102625',
                                                                                         'out_octets': '86132965392919',
                                                                                         'out_ucast_pkts': '29597722807',
                                                                                         'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/8',
                                                                                                                 'id': '1/8',
                                                                                                                 'links': [       {       'connected_time': '21319494',
                                                                                                                                          'fiber_length': 1116,
                                                                                                                                          'poller_hash': 'MjA1RDk3cGVlcnBvbnBvbiA1MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.981},
                                                                                                                                  {       'connected_time': '20839243',
                                                                                                                                          'fiber_length': 1046,
                                                                                                                                          'poller_hash': 'MzQ5RTg3cGVlcnBvbnBvbiA2OTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.981},
                                                                                                                                  {       'connected_time': '30595765',
                                                                                                                                          'fiber_length': 1115,
                                                                                                                                          'poller_hash': 'MjA1QjM4cGVlcnBvbnBvbiA2OTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.981},
                                                                                                                                  {       'connected_time': '68802632',
                                                                                                                                          'fiber_length': 1119,
                                                                                                                                          'poller_hash': 'MzQ5RTUxcGVlcnBvbnBvbiA3MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.981},
                                                                                                                                  {       'connected_time': '688346',
                                                                                                                                          'fiber_length': 901,
                                                                                                                                          'poller_hash': 'NjQwQUZGcGVlcnBvbnBvbiAxNDE3NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.981},
                                                                                                                                  {       'connected_time': '688356',
                                                                                                                                          'fiber_length': 1115,
                                                                                                                                          'poller_hash': 'ODY3OEEycGVlcnBvbnBvbiAxNzI1OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.981}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/8',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNDEwNjAxMjg5',
                                                           'serial_number': '211410601289',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991355'},
                                                   {       'id': '3',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '3/1',
                                                                                         'in_discards': '1146',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '236769',
                                                                                         'in_octets': '838847003493',
                                                                                         'in_ucast_pkts': '3297859328',
                                                                                         'links': [       {       'in_nucast_pkts': '20569',
                                                                                                                  'in_octets': '197253958927',
                                                                                                                  'in_ucast_pkts': '606034054',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '32346044',
                                                                                                                  'out_octets': '1830384241804',
                                                                                                                  'out_ucast_pkts': '1545447936',
                                                                                                                  'poller_hash': 'QjUxMkRCcGVlcmludGZwb24gOTEw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '19639727250',
                                                                                                                  'in_ucast_pkts': '138626751',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '797490008257',
                                                                                                                  'out_ucast_pkts': '576831432',
                                                                                                                  'poller_hash': 'ODY5M0I1cGVlcmludGZwb24gMTQyMjQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '58183189697',
                                                                                                                  'in_ucast_pkts': '259479199',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1088240746343',
                                                                                                                  'out_ucast_pkts': '845403095',
                                                                                                                  'poller_hash': 'ODY5M0M3cGVlcmludGZwb24gMTQyMjU0'},
                                                                                                          {       'in_nucast_pkts': '752945',
                                                                                                                  'in_octets': '119660496513',
                                                                                                                  'in_ucast_pkts': '1083768890',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '95001851',
                                                                                                                  'out_octets': '6167832674671',
                                                                                                                  'out_ucast_pkts': '4226557460',
                                                                                                                  'poller_hash': 'ODY5M0UzcGVlcmludGZwb24gMTQyMjY0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '85128767875',
                                                                                                                  'in_ucast_pkts': '343233681',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1538863338824',
                                                                                                                  'out_ucast_pkts': '1167879766',
                                                                                                                  'poller_hash': 'ODY5M0UwcGVlcmludGZwb24gMTQyMjg2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '55847101573',
                                                                                                                  'in_ucast_pkts': '231915260',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1065560722477',
                                                                                                                  'out_ucast_pkts': '811391259',
                                                                                                                  'poller_hash': 'ODY5M0QwcGVlcmludGZwb24gMTQyMjk2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '7035734623',
                                                                                                                  'in_ucast_pkts': '41494881',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '224869884836',
                                                                                                                  'out_ucast_pkts': '169078705',
                                                                                                                  'poller_hash': 'ODY5NDA0cGVlcmludGZwb24gMTQyMzE2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '41142730647',
                                                                                                                  'in_ucast_pkts': '161134276',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '583938080550',
                                                                                                                  'out_ucast_pkts': '435761469',
                                                                                                                  'poller_hash': 'ODY5M0RBcGVlcmludGZwb24gMTQyMzM2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '111949369665',
                                                                                                                  'in_ucast_pkts': '47381067021',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '264410448598',
                                                                                                                  'out_ucast_pkts': '255515672',
                                                                                                                  'poller_hash': 'ODY5NDczcGVlcmludGZwb24gMTQyMzY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '98126934260',
                                                                                                                  'in_ucast_pkts': '277737030',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1176580628118',
                                                                                                                  'out_ucast_pkts': '854856246',
                                                                                                                  'poller_hash': 'ODY5M0ZCcGVlcmludGZwb24gMTQyMzk2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '10015870208',
                                                                                                                  'in_ucast_pkts': '4312390751',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '92979438877',
                                                                                                                  'out_ucast_pkts': '87314502',
                                                                                                                  'poller_hash': 'ODY5Mzg0cGVlcmludGZwb24gMTQyNDA2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '7362389747',
                                                                                                                  'in_ucast_pkts': '36232489',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '223759019455',
                                                                                                                  'out_ucast_pkts': '175045468',
                                                                                                                  'poller_hash': 'ODY5M0FDcGVlcmludGZwb24gMTQyNDE2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '35198371045',
                                                                                                                  'in_ucast_pkts': '82263663',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '358224715890',
                                                                                                                  'out_ucast_pkts': '299747892',
                                                                                                                  'poller_hash': 'ODY5M0IycGVlcmludGZwb24gMTQyNDY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODY5NDJGcGVlcmludGZwb24gMTQyNDc2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '101535944997',
                                                                                                                  'in_ucast_pkts': '9168315573',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3609851534314',
                                                                                                                  'out_ucast_pkts': '2545108758',
                                                                                                                  'poller_hash': 'MTI3NDk5RnBlZXJpbnRmcG9uIDE0MjQ5Ng=='},
                                                                                                          {       'in_nucast_pkts': '916665',
                                                                                                                  'in_octets': '4572262513',
                                                                                                                  'in_ucast_pkts': '14000535',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149217273',
                                                                                                                  'out_octets': '50650310535',
                                                                                                                  'out_ucast_pkts': '27712829',
                                                                                                                  'poller_hash': 'MTFCMjVFNXBlZXJpbnRmcG9uIDIxMDA4NQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '3/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '24238246',
                                                                                         'out_octets': '15205689565891',
                                                                                         'out_ucast_pkts': '11272334665',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/1',
                                                                                                                 'id': '3/1',
                                                                                                                 'links': [       {       'connected_time': '14251264',
                                                                                                                                          'fiber_length': 10790,
                                                                                                                                          'poller_hash': 'QjUxMkRCcGVlcnBvbnBvbiA5MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '18864638',
                                                                                                                                          'fiber_length': 10604,
                                                                                                                                          'poller_hash': 'ODY5M0I1cGVlcnBvbnBvbiAxNDIyNDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688346',
                                                                                                                                          'fiber_length': 10585,
                                                                                                                                          'poller_hash': 'ODY5M0M3cGVlcnBvbnBvbiAxNDIyNTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '71368108',
                                                                                                                                          'fiber_length': 10574,
                                                                                                                                          'poller_hash': 'ODY5M0UzcGVlcnBvbnBvbiAxNDIyNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688349',
                                                                                                                                          'fiber_length': 10461,
                                                                                                                                          'poller_hash': 'ODY5M0UwcGVlcnBvbnBvbiAxNDIyODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '6552385',
                                                                                                                                          'fiber_length': 10442,
                                                                                                                                          'poller_hash': 'ODY5M0QwcGVlcnBvbnBvbiAxNDIyOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688380',
                                                                                                                                          'fiber_length': 10495,
                                                                                                                                          'poller_hash': 'ODY5NDA0cGVlcnBvbnBvbiAxNDIzMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688353',
                                                                                                                                          'fiber_length': 10377,
                                                                                                                                          'poller_hash': 'ODY5M0RBcGVlcnBvbnBvbiAxNDIzMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '13582346',
                                                                                                                                          'fiber_length': 10555,
                                                                                                                                          'poller_hash': 'ODY5NDczcGVlcnBvbnBvbiAxNDIzNjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688346',
                                                                                                                                          'fiber_length': 10471,
                                                                                                                                          'poller_hash': 'ODY5M0ZCcGVlcnBvbnBvbiAxNDIzOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '18660841',
                                                                                                                                          'fiber_length': 10582,
                                                                                                                                          'poller_hash': 'ODY5Mzg0cGVlcnBvbnBvbiAxNDI0MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '21815921',
                                                                                                                                          'fiber_length': 10536,
                                                                                                                                          'poller_hash': 'ODY5M0FDcGVlcnBvbnBvbiAxNDI0MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688353',
                                                                                                                                          'fiber_length': 10593,
                                                                                                                                          'poller_hash': 'ODY5M0IycGVlcnBvbnBvbiAxNDI0NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688354',
                                                                                                                                          'fiber_length': 10373,
                                                                                                                                          'poller_hash': 'ODY5NDJGcGVlcnBvbnBvbiAxNDI0NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '4060597',
                                                                                                                                          'fiber_length': 10552,
                                                                                                                                          'poller_hash': 'MTI3NDk5RnBlZXJwb25wb24gMTQyNDk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '21816086',
                                                                                                                                          'fiber_length': 10633,
                                                                                                                                          'poller_hash': 'MTFCMjVFNXBlZXJwb25wb24gMjEwMDg1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/1',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '3/2',
                                                                                         'in_discards': '7185',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '330525',
                                                                                         'in_octets': '588038622418',
                                                                                         'in_ucast_pkts': '3112787849',
                                                                                         'links': [       {       'in_nucast_pkts': '14763',
                                                                                                                  'in_octets': '6407478930',
                                                                                                                  'in_ucast_pkts': '24295667',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464403',
                                                                                                                  'out_octets': '91063223357',
                                                                                                                  'out_ucast_pkts': '69457078',
                                                                                                                  'poller_hash': 'ODY5M0E1cGVlcmludGZwb24gMTQyNTA2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '31620576089',
                                                                                                                  'in_ucast_pkts': '166688457',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '814883807358',
                                                                                                                  'out_ucast_pkts': '600510609',
                                                                                                                  'poller_hash': 'ODY5M0Q5cGVlcmludGZwb24gMTQyNTU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '12139013374',
                                                                                                                  'in_ucast_pkts': '131607262',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '387120254706',
                                                                                                                  'out_ucast_pkts': '269441326',
                                                                                                                  'poller_hash': 'ODY5M0I5cGVlcmludGZwb24gMTQyNjA1'},
                                                                                                          {       'in_nucast_pkts': '952122',
                                                                                                                  'in_octets': '515850706480',
                                                                                                                  'in_ucast_pkts': '3792982580',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149539546',
                                                                                                                  'out_octets': '14737857952500',
                                                                                                                  'out_ucast_pkts': '11090321461',
                                                                                                                  'poller_hash': 'ODcyNDg0cGVlcmludGZwb24gMTQyODAz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '22117500414',
                                                                                                                  'in_ucast_pkts': '163054916',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '522818871620',
                                                                                                                  'out_ucast_pkts': '380055566',
                                                                                                                  'poller_hash': 'OTU2RjY3cGVlcmludGZwb24gMTQ1NTA4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45291303846',
                                                                                                                  'in_ucast_pkts': '295219888',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '880087551014',
                                                                                                                  'out_ucast_pkts': '628923100',
                                                                                                                  'poller_hash': 'QTlDNzdCcGVlcmludGZwb24gMTQ3MzY2'},
                                                                                                          {       'in_nucast_pkts': '119253',
                                                                                                                  'in_octets': '1012519937',
                                                                                                                  'in_ucast_pkts': '11282223',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '49519180',
                                                                                                                  'out_octets': '82644586229',
                                                                                                                  'out_ucast_pkts': '44686831',
                                                                                                                  'poller_hash': 'QjBBMTRFcGVlcmludGZwb24gMTQ5NjU1'},
                                                                                                          {       'in_nucast_pkts': '31819',
                                                                                                                  'in_octets': '54864245648',
                                                                                                                  'in_ucast_pkts': '95136356',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '85703255',
                                                                                                                  'out_octets': '182649242925',
                                                                                                                  'out_ucast_pkts': '162762136',
                                                                                                                  'poller_hash': 'QjUxMTZDcGVlcmludGZwb24gMTUxMjI3'},
                                                                                                          {       'in_nucast_pkts': '132634',
                                                                                                                  'in_octets': '10852915982',
                                                                                                                  'in_ucast_pkts': '21194325',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14580570',
                                                                                                                  'out_octets': '127108458661',
                                                                                                                  'out_ucast_pkts': '96700608',
                                                                                                                  'poller_hash': 'QjUxMTRDcGVlcmludGZwb24gMTUxNjI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '61840548718',
                                                                                                                  'in_ucast_pkts': '300762560',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1369935464804',
                                                                                                                  'out_ucast_pkts': '951651726',
                                                                                                                  'poller_hash': 'QjBBMjEycGVlcmludGZwb24gMTUyMzc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxRDhFcGVlcmludGZwb24gMTU3NDQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgzQjc0cGVlcmludGZwb24gMTYyNDcz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '67348657846',
                                                                                                                  'in_ucast_pkts': '274571166',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1631398275052',
                                                                                                                  'out_ucast_pkts': '1167797243',
                                                                                                                  'poller_hash': 'QzgzNUEzcGVlcmludGZwb24gMTYzMzA5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '39249919677',
                                                                                                                  'in_ucast_pkts': '72036623',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '164944559804',
                                                                                                                  'out_ucast_pkts': '152153486',
                                                                                                                  'poller_hash': 'Q0JEMjEycGVlcmludGZwb24gMTY0ODA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '28257824664',
                                                                                                                  'in_ucast_pkts': '136492831',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '582785009363',
                                                                                                                  'out_ucast_pkts': '447523324',
                                                                                                                  'poller_hash': 'QzgzMTQ2cGVlcmludGZwb24gMTcyNTIx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '50829907764',
                                                                                                                  'in_ucast_pkts': '374318656',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1795952028331',
                                                                                                                  'out_ucast_pkts': '1251365948',
                                                                                                                  'poller_hash': 'MTE5MzBDQXBlZXJpbnRmcG9uIDI3NjQ4Mg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '3/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '223034323',
                                                                                         'out_octets': '13173962265924',
                                                                                         'out_ucast_pkts': '9397297049',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/2',
                                                                                                                 'id': '3/2',
                                                                                                                 'links': [       {       'connected_time': '688356',
                                                                                                                                          'fiber_length': 10370,
                                                                                                                                          'poller_hash': 'ODY5M0E1cGVlcnBvbnBvbiAxNDI1MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '688349',
                                                                                                                                          'fiber_length': 10566,
                                                                                                                                          'poller_hash': 'ODY5M0Q5cGVlcnBvbnBvbiAxNDI1NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '71555921',
                                                                                                                                          'fiber_length': 10498,
                                                                                                                                          'poller_hash': 'ODY5M0I5cGVlcnBvbnBvbiAxNDI2MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '21816583',
                                                                                                                                          'fiber_length': 10332,
                                                                                                                                          'poller_hash': 'ODcyNDg0cGVlcnBvbnBvbiAxNDI4MDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '688356',
                                                                                                                                          'fiber_length': 10397,
                                                                                                                                          'poller_hash': 'OTU2RjY3cGVlcnBvbnBvbiAxNDU1MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '9084563',
                                                                                                                                          'fiber_length': 10866,
                                                                                                                                          'poller_hash': 'QTlDNzdCcGVlcnBvbnBvbiAxNDczNjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '2138232',
                                                                                                                                          'fiber_length': 10885,
                                                                                                                                          'poller_hash': 'QjBBMTRFcGVlcnBvbnBvbiAxNDk2NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '67790643',
                                                                                                                                          'fiber_length': 10181,
                                                                                                                                          'poller_hash': 'QjUxMTZDcGVlcnBvbnBvbiAxNTEyMjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '6230566',
                                                                                                                                          'fiber_length': 10606,
                                                                                                                                          'poller_hash': 'QjUxMTRDcGVlcnBvbnBvbiAxNTE2Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '21816063',
                                                                                                                                          'fiber_length': 10354,
                                                                                                                                          'poller_hash': 'QjBBMjEycGVlcnBvbnBvbiAxNTIzNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '6377864',
                                                                                                                                          'fiber_length': 10142,
                                                                                                                                          'poller_hash': 'QzgxRDhFcGVlcnBvbnBvbiAxNTc0NDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '21816543',
                                                                                                                                          'fiber_length': 10575,
                                                                                                                                          'poller_hash': 'QzgzQjc0cGVlcnBvbnBvbiAxNjI0NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '688387',
                                                                                                                                          'fiber_length': 10436,
                                                                                                                                          'poller_hash': 'QzgzNUEzcGVlcnBvbnBvbiAxNjMzMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '21815666',
                                                                                                                                          'fiber_length': 10161,
                                                                                                                                          'poller_hash': 'Q0JEMjEycGVlcnBvbnBvbiAxNjQ4MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '596470',
                                                                                                                                          'fiber_length': 10555,
                                                                                                                                          'poller_hash': 'QzgzMTQ2cGVlcnBvbnBvbiAxNzI1MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842},
                                                                                                                                  {       'connected_time': '688354',
                                                                                                                                          'fiber_length': 10132,
                                                                                                                                          'poller_hash': 'MTE5MzBDQXBlZXJwb25wb24gMjc2NDgy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.842}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/2',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '3/3',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '14',
                                                                                         'in_octets': '5505422196007',
                                                                                         'in_ucast_pkts': '4608761809',
                                                                                         'links': [       {       'in_nucast_pkts': '19',
                                                                                                                  'in_octets': '470604128300',
                                                                                                                  'in_ucast_pkts': '486686968',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464543',
                                                                                                                  'out_octets': '421370797828',
                                                                                                                  'out_ucast_pkts': '434516754',
                                                                                                                  'poller_hash': 'ODcyNDgzcGVlcmludGZwb24gOTA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '66589636596',
                                                                                                                  'in_ucast_pkts': '164878656',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '376110368002',
                                                                                                                  'out_ucast_pkts': '340422356',
                                                                                                                  'poller_hash': 'ODcyNDIwcGVlcmludGZwb24gMTQzMjA3'},
                                                                                                          {       'in_nucast_pkts': '26',
                                                                                                                  'in_octets': '29356322786',
                                                                                                                  'in_ucast_pkts': '29115592',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465446',
                                                                                                                  'out_octets': '24720525771',
                                                                                                                  'out_ucast_pkts': '29283774',
                                                                                                                  'poller_hash': 'MTI5NzM1MXBlZXJpbnRmcG9uIDI1NjY5Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '3/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '23913468',
                                                                                         'out_octets': '2106009366917',
                                                                                         'out_ucast_pkts': '3209911186',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/3',
                                                                                                                 'id': '3/3',
                                                                                                                 'links': [       {       'connected_time': '688368',
                                                                                                                                          'fiber_length': 2355,
                                                                                                                                          'poller_hash': 'ODcyNDgzcGVlcnBvbnBvbiA5MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.545},
                                                                                                                                  {       'connected_time': '21816353',
                                                                                                                                          'fiber_length': 2416,
                                                                                                                                          'poller_hash': 'ODcyNDIwcGVlcnBvbnBvbiAxNDMyMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.545},
                                                                                                                                  {       'connected_time': '688351',
                                                                                                                                          'fiber_length': 2787,
                                                                                                                                          'poller_hash': 'MTI5NzM1MXBlZXJwb25wb24gMjU2Njk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.545}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/3',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '3/4',
                                                                                         'in_discards': '12106',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '363742',
                                                                                         'in_octets': '1106546637591',
                                                                                         'in_ucast_pkts': '5061443330',
                                                                                         'links': [       {       'in_nucast_pkts': '19772',
                                                                                                                  'in_octets': '86192202848',
                                                                                                                  'in_ucast_pkts': '705114869',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '72095703',
                                                                                                                  'out_octets': '4467410172946',
                                                                                                                  'out_ucast_pkts': '3026447457',
                                                                                                                  'poller_hash': 'Mjk4MjRFcGVlcmludGZwb24gMTQx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '21247329189',
                                                                                                                  'in_ucast_pkts': '121713897',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '430559497447',
                                                                                                                  'out_ucast_pkts': '329800327',
                                                                                                                  'poller_hash': 'Mjk4MUZGcGVlcmludGZwb24gMTQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '63457684090',
                                                                                                                  'in_ucast_pkts': '13092406543',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '945931530684',
                                                                                                                  'out_ucast_pkts': '711698087',
                                                                                                                  'poller_hash': 'Mjk4MUU1cGVlcmludGZwb24gMTUw'},
                                                                                                          {       'in_nucast_pkts': '305922',
                                                                                                                  'in_octets': '9076408813',
                                                                                                                  'in_ucast_pkts': '42794659',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464242',
                                                                                                                  'out_octets': '256162167272',
                                                                                                                  'out_ucast_pkts': '196008874',
                                                                                                                  'poller_hash': 'MkJEMzYwcGVlcmludGZwb24gMTU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '64733845993',
                                                                                                                  'in_ucast_pkts': '393277218',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1666065753599',
                                                                                                                  'out_ucast_pkts': '1217642818',
                                                                                                                  'poller_hash': 'MUQwMzM1cGVlcmludGZwb24gMTU3'},
                                                                                                          {       'in_nucast_pkts': '129023',
                                                                                                                  'in_octets': '99400652464',
                                                                                                                  'in_ucast_pkts': '384629769',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14573150',
                                                                                                                  'out_octets': '2144788445160',
                                                                                                                  'out_ucast_pkts': '1511790438',
                                                                                                                  'poller_hash': 'OTU2RTU5cGVlcmludGZwb24gMTU5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '58069279695',
                                                                                                                  'in_ucast_pkts': '264049830',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '114903503',
                                                                                                                  'out_octets': '1693754715094',
                                                                                                                  'out_ucast_pkts': '1132697876',
                                                                                                                  'poller_hash': 'Mjk4MUU0cGVlcmludGZwb24gMTgw'},
                                                                                                          {       'in_nucast_pkts': '8864',
                                                                                                                  'in_octets': '228833052864',
                                                                                                                  'in_ucast_pkts': '1297857919',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45832228',
                                                                                                                  'out_octets': '5291943696178',
                                                                                                                  'out_ucast_pkts': '3820761452',
                                                                                                                  'poller_hash': 'MzIyM0NCcGVlcmludGZwb24gMTky'},
                                                                                                          {       'in_nucast_pkts': '743',
                                                                                                                  'in_octets': '15978955369',
                                                                                                                  'in_ucast_pkts': '76340151',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1121595',
                                                                                                                  'out_octets': '453956222853',
                                                                                                                  'out_ucast_pkts': '346817938',
                                                                                                                  'poller_hash': 'NDNCM0Y0cGVlcmludGZwb24gNzcx'},
                                                                                                          {       'in_nucast_pkts': '14069',
                                                                                                                  'in_octets': '35272410729',
                                                                                                                  'in_ucast_pkts': '149430233',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '85888364',
                                                                                                                  'out_octets': '611268189356',
                                                                                                                  'out_ucast_pkts': '503434866',
                                                                                                                  'poller_hash': 'NUE2QzM5cGVlcmludGZwb24gMTM3Njk3'},
                                                                                                          {       'in_nucast_pkts': '12466',
                                                                                                                  'in_octets': '24536744342',
                                                                                                                  'in_ucast_pkts': '88184358',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464287',
                                                                                                                  'out_octets': '461410088118',
                                                                                                                  'out_ucast_pkts': '336262861',
                                                                                                                  'poller_hash': 'ODQ0QkJFcGVlcmludGZwb24gMTUwMTE4'},
                                                                                                          {       'in_nucast_pkts': '1837',
                                                                                                                  'in_octets': '7063643270',
                                                                                                                  'in_ucast_pkts': '37803317',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '380390',
                                                                                                                  'out_octets': '360137241448',
                                                                                                                  'out_ucast_pkts': '250424212',
                                                                                                                  'poller_hash': 'QzgzMUVBcGVlcmludGZwb24gMTcyNTY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '23441823793',
                                                                                                                  'in_ucast_pkts': '120617855',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '568993365585',
                                                                                                                  'out_ucast_pkts': '415045839',
                                                                                                                  'poller_hash': 'MTI3NDEyNXBlZXJpbnRmcG9uIDIzMDc3NA=='},
                                                                                                          {       'in_nucast_pkts': '15547',
                                                                                                                  'in_octets': '7174213108',
                                                                                                                  'in_ucast_pkts': '50551816',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1499048',
                                                                                                                  'out_octets': '175273632415',
                                                                                                                  'out_ucast_pkts': '121398423',
                                                                                                                  'poller_hash': 'MTI3NDNFNHBlZXJpbnRmcG9uIDI3NDQxNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '3/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '1136266780',
                                                                                         'out_octets': '25983542658223',
                                                                                         'out_ucast_pkts': '17856020047',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/4',
                                                                                                                 'id': '3/4',
                                                                                                                 'links': [       {       'connected_time': '31460818',
                                                                                                                                          'fiber_length': 1185,
                                                                                                                                          'poller_hash': 'Mjk4MjRFcGVlcnBvbnBvbiAxNDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '1024852',
                                                                                                                                          'fiber_length': 1128,
                                                                                                                                          'poller_hash': 'Mjk4MUZGcGVlcnBvbnBvbiAxNDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '6229370',
                                                                                                                                          'fiber_length': 1269,
                                                                                                                                          'poller_hash': 'Mjk4MUU1cGVlcnBvbnBvbiAxNTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '688352',
                                                                                                                                          'fiber_length': 1321,
                                                                                                                                          'poller_hash': 'MkJEMzYwcGVlcnBvbnBvbiAxNTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '6231300',
                                                                                                                                          'fiber_length': 1323,
                                                                                                                                          'poller_hash': 'MUQwMzM1cGVlcnBvbnBvbiAxNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '6220910',
                                                                                                                                          'fiber_length': 1234,
                                                                                                                                          'poller_hash': 'OTU2RTU5cGVlcnBvbnBvbiAxNTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '688358',
                                                                                                                                          'fiber_length': 1304,
                                                                                                                                          'poller_hash': 'Mjk4MUU0cGVlcnBvbnBvbiAxODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '19868105',
                                                                                                                                          'fiber_length': 1319,
                                                                                                                                          'poller_hash': 'MzIyM0NCcGVlcnBvbnBvbiAxOTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '688344',
                                                                                                                                          'fiber_length': 1220,
                                                                                                                                          'poller_hash': 'NDNCM0Y0cGVlcnBvbnBvbiA3NzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '67790150',
                                                                                                                                          'fiber_length': 1188,
                                                                                                                                          'poller_hash': 'NUE2QzM5cGVlcnBvbnBvbiAxMzc2OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '688372',
                                                                                                                                          'fiber_length': 1211,
                                                                                                                                          'poller_hash': 'ODQ0QkJFcGVlcnBvbnBvbiAxNTAxMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '688357',
                                                                                                                                          'fiber_length': 1279,
                                                                                                                                          'poller_hash': 'QzgzMUVBcGVlcnBvbnBvbiAxNzI1NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '688352',
                                                                                                                                          'fiber_length': 1332,
                                                                                                                                          'poller_hash': 'MTI3NDEyNXBlZXJwb25wb24gMjMwNzc0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691},
                                                                                                                                  {       'connected_time': '688341',
                                                                                                                                          'fiber_length': 1348,
                                                                                                                                          'poller_hash': 'MTI3NDNFNHBlZXJwb25wb24gMjc0NDE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.691}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/4',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '3/5',
                                                                                         'in_discards': '104976',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '552969',
                                                                                         'in_octets': '2190562298299',
                                                                                         'in_ucast_pkts': '5337244235',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3047270292',
                                                                                                                  'in_ucast_pkts': '15927235',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '100003059487',
                                                                                                                  'out_ucast_pkts': '83334930',
                                                                                                                  'poller_hash': 'MzQ5RTdEcGVlcmludGZwb24gODY='},
                                                                                                          {       'in_nucast_pkts': '5714666',
                                                                                                                  'in_octets': '1720670962447',
                                                                                                                  'in_ucast_pkts': '8300401002',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149404501',
                                                                                                                  'out_octets': '20375364295764',
                                                                                                                  'out_ucast_pkts': '16539027239',
                                                                                                                  'poller_hash': 'Mjk4MjFGcGVlcmludGZwb24gMTA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '15815488351',
                                                                                                                  'in_ucast_pkts': '103363825',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '451030515582',
                                                                                                                  'out_ucast_pkts': '337854339',
                                                                                                                  'poller_hash': 'MUQwNjkycGVlcmludGZwb24gMTQ2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '138930132649',
                                                                                                                  'in_ucast_pkts': '321945455',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '751411398454',
                                                                                                                  'out_ucast_pkts': '677258796',
                                                                                                                  'poller_hash': 'Mjk4MjYycGVlcmludGZwb24gMTQ3'},
                                                                                                          {       'in_nucast_pkts': '11952',
                                                                                                                  'in_octets': '4432356627',
                                                                                                                  'in_ucast_pkts': '10982805',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464011',
                                                                                                                  'out_octets': '36995223884',
                                                                                                                  'out_ucast_pkts': '27779625',
                                                                                                                  'poller_hash': 'MUQwNkQxcGVlcmludGZwb24gMTU4'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '6069429757',
                                                                                                                  'in_ucast_pkts': '21835788',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464254',
                                                                                                                  'out_octets': '95047624277',
                                                                                                                  'out_ucast_pkts': '81917393',
                                                                                                                  'poller_hash': 'MkJENjlGcGVlcmludGZwb24gMTcy'},
                                                                                                          {       'in_nucast_pkts': '2527412',
                                                                                                                  'in_octets': '1202356911933',
                                                                                                                  'in_ucast_pkts': '6338618707',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '146310514',
                                                                                                                  'out_octets': '27279713419672',
                                                                                                                  'out_ucast_pkts': '19073082024',
                                                                                                                  'poller_hash': 'ODY5M0Y1cGVlcmludGZwb24gMjQw'},
                                                                                                          {       'in_nucast_pkts': '321295',
                                                                                                                  'in_octets': '431830854033',
                                                                                                                  'in_ucast_pkts': '1256073690',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11384417',
                                                                                                                  'out_octets': '810425602509',
                                                                                                                  'out_ucast_pkts': '1258994929',
                                                                                                                  'poller_hash': 'MjYyNUU1cGVlcmludGZwb24gMjcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '32291728406',
                                                                                                                  'in_ucast_pkts': '4414529026',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1254672908408',
                                                                                                                  'out_ucast_pkts': '898636499',
                                                                                                                  'poller_hash': 'QTlDMjlFcGVlcmludGZwb24gNDEy'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '601970829',
                                                                                                                  'in_ucast_pkts': '1300130',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120044',
                                                                                                                  'out_octets': '4476123378',
                                                                                                                  'out_ucast_pkts': '3646325',
                                                                                                                  'poller_hash': 'NEZBMjQ0cGVlcmludGZwb24gNDk3'},
                                                                                                          {       'in_nucast_pkts': '53',
                                                                                                                  'in_octets': '4986174164',
                                                                                                                  'in_ucast_pkts': '43187695',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1462308',
                                                                                                                  'out_octets': '186686989602',
                                                                                                                  'out_ucast_pkts': '129976292',
                                                                                                                  'poller_hash': 'NUE2QkE2cGVlcmludGZwb24gODA2'},
                                                                                                          {       'in_nucast_pkts': '1464023',
                                                                                                                  'in_octets': '146829894050',
                                                                                                                  'in_ucast_pkts': '467800737',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '85951116',
                                                                                                                  'out_octets': '1235358423734',
                                                                                                                  'out_ucast_pkts': '899020712',
                                                                                                                  'poller_hash': 'NUE2QkY4cGVlcmludGZwb24gMTM3ODEz'},
                                                                                                          {       'in_nucast_pkts': '628',
                                                                                                                  'in_octets': '43009546343',
                                                                                                                  'in_ucast_pkts': '69419158',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463508',
                                                                                                                  'out_octets': '169246511663',
                                                                                                                  'out_ucast_pkts': '137830063',
                                                                                                                  'poller_hash': 'NUE2QjYzcGVlcmludGZwb24gMTM5MTk0'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '3/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '24239741',
                                                                                         'out_octets': '14812045578241',
                                                                                         'out_ucast_pkts': '12073877382',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/5',
                                                                                                                 'id': '3/5',
                                                                                                                 'links': [       {       'connected_time': '688357',
                                                                                                                                          'fiber_length': 1024,
                                                                                                                                          'poller_hash': 'MzQ5RTdEcGVlcnBvbnBvbiA4Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '21815281',
                                                                                                                                          'fiber_length': 1166,
                                                                                                                                          'poller_hash': 'Mjk4MjFGcGVlcnBvbnBvbiAxMDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '16308318',
                                                                                                                                          'fiber_length': 1000,
                                                                                                                                          'poller_hash': 'MUQwNjkycGVlcnBvbnBvbiAxNDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '21816177',
                                                                                                                                          'fiber_length': 1205,
                                                                                                                                          'poller_hash': 'Mjk4MjYycGVlcnBvbnBvbiAxNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '688368',
                                                                                                                                          'fiber_length': 962,
                                                                                                                                          'poller_hash': 'MUQwNkQxcGVlcnBvbnBvbiAxNTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '688353',
                                                                                                                                          'fiber_length': 975,
                                                                                                                                          'poller_hash': 'MkJENjlGcGVlcnBvbnBvbiAxNzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '21816210',
                                                                                                                                          'fiber_length': 972,
                                                                                                                                          'poller_hash': 'ODY5M0Y1cGVlcnBvbnBvbiAyNDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '6175813',
                                                                                                                                          'fiber_length': 1317,
                                                                                                                                          'poller_hash': 'MjYyNUU1cGVlcnBvbnBvbiAyNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '688352',
                                                                                                                                          'fiber_length': 1008,
                                                                                                                                          'poller_hash': 'QTlDMjlFcGVlcnBvbnBvbiA0MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '688345',
                                                                                                                                          'fiber_length': 862,
                                                                                                                                          'poller_hash': 'NEZBMjQ0cGVlcnBvbnBvbiA0OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '688366',
                                                                                                                                          'fiber_length': 961,
                                                                                                                                          'poller_hash': 'NUE2QkE2cGVlcnBvbnBvbiA4MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '67789945',
                                                                                                                                          'fiber_length': 1118,
                                                                                                                                          'poller_hash': 'NUE2QkY4cGVlcnBvbnBvbiAxMzc4MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634},
                                                                                                                                  {       'connected_time': '688367',
                                                                                                                                          'fiber_length': 931,
                                                                                                                                          'poller_hash': 'NUE2QjYzcGVlcnBvbnBvbiAxMzkxOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.634}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/5',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '3/6',
                                                                                         'in_discards': '2120',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '351257',
                                                                                         'in_octets': '1581495373094',
                                                                                         'in_ucast_pkts': '3907366844',
                                                                                         'links': [       {       'in_nucast_pkts': '1769',
                                                                                                                  'in_octets': '129283058005',
                                                                                                                  'in_ucast_pkts': '287184325',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19475831',
                                                                                                                  'out_octets': '1310866529577',
                                                                                                                  'out_ucast_pkts': '976655744',
                                                                                                                  'poller_hash': 'MjVEOTJFcGVlcmludGZwb24gMTM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '305534140760',
                                                                                                                  'in_ucast_pkts': '456200196',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '790925644586',
                                                                                                                  'out_ucast_pkts': '815214787',
                                                                                                                  'poller_hash': 'QTlDMzdEcGVlcmludGZwb24gMTYy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '382116362577',
                                                                                                                  'in_ucast_pkts': '572266006',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7993589942051',
                                                                                                                  'out_ucast_pkts': '5552480472',
                                                                                                                  'poller_hash': 'Mjk4MjMycGVlcmludGZwb24gMzA4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '5704110957',
                                                                                                                  'in_ucast_pkts': '13473315',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '30733253407',
                                                                                                                  'out_ucast_pkts': '39683778',
                                                                                                                  'poller_hash': 'MkIwM0ZBcGVlcmludGZwb24gNDUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '111407538674',
                                                                                                                  'in_ucast_pkts': '419282533',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2109549162142',
                                                                                                                  'out_ucast_pkts': '1642253180',
                                                                                                                  'poller_hash': 'MzQ5RTg1cGVlcmludGZwb24gNTAz'},
                                                                                                          {       'in_nucast_pkts': '29196',
                                                                                                                  'in_octets': '2695353400',
                                                                                                                  'in_ucast_pkts': '12193942',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464824',
                                                                                                                  'out_octets': '70923633368',
                                                                                                                  'out_ucast_pkts': '51505076',
                                                                                                                  'poller_hash': 'MzAwRTc1cGVlcmludGZwb24gNTA2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '35958964854',
                                                                                                                  'in_ucast_pkts': '136828569',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '573100907147',
                                                                                                                  'out_ucast_pkts': '419335398',
                                                                                                                  'poller_hash': 'MzIyNDAzcGVlcmludGZwb24gNzA4'},
                                                                                                          {       'in_nucast_pkts': '15',
                                                                                                                  'in_octets': '6561547602',
                                                                                                                  'in_ucast_pkts': '6941177',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464380',
                                                                                                                  'out_octets': '3102915082',
                                                                                                                  'out_ucast_pkts': '4531554',
                                                                                                                  'poller_hash': 'MzAwRUY4cGVlcmludGZwb24gODIz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '52528117846',
                                                                                                                  'in_ucast_pkts': '226983349',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1041748484237',
                                                                                                                  'out_ucast_pkts': '758268984',
                                                                                                                  'poller_hash': 'ODY5NDE4cGVlcmludGZwb24gMTQwNjY5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '132812031037',
                                                                                                                  'in_ucast_pkts': '43141185377',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '454316814274',
                                                                                                                  'out_ucast_pkts': '430736318',
                                                                                                                  'poller_hash': 'QzgxRTZGcGVlcmludGZwb24gMjA0OTI1'},
                                                                                                          {       'in_nucast_pkts': '105949',
                                                                                                                  'in_octets': '20920149465',
                                                                                                                  'in_ucast_pkts': '265895097',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14237494',
                                                                                                                  'out_octets': '1425464949951',
                                                                                                                  'out_ucast_pkts': '947457510',
                                                                                                                  'poller_hash': 'MTI3NTBDRHBlZXJpbnRmcG9uIDIxODA3Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '21496092414',
                                                                                                                  'in_ucast_pkts': '41520178',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '234110108756',
                                                                                                                  'out_ucast_pkts': '182312687',
                                                                                                                  'poller_hash': 'MTI5NzNCNXBlZXJpbnRmcG9uIDI1NjU5Ng=='},
                                                                                                          {       'in_nucast_pkts': '14500',
                                                                                                                  'in_octets': '9592759743',
                                                                                                                  'in_ucast_pkts': '75891941',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465411',
                                                                                                                  'out_octets': '392652009078',
                                                                                                                  'out_ucast_pkts': '290165301',
                                                                                                                  'poller_hash': 'MTI3NDdBMXBlZXJpbnRmcG9uIDI3MjIyOQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '3/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '48093119',
                                                                                         'out_octets': '21486290127075',
                                                                                         'out_ucast_pkts': '15813604237',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/6',
                                                                                                                 'id': '3/6',
                                                                                                                 'links': [       {       'connected_time': '10784823',
                                                                                                                                          'fiber_length': 653,
                                                                                                                                          'poller_hash': 'MjVEOTJFcGVlcnBvbnBvbiAxMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '20037947',
                                                                                                                                          'fiber_length': 1017,
                                                                                                                                          'poller_hash': 'QTlDMzdEcGVlcnBvbnBvbiAxNjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '14582749',
                                                                                                                                          'fiber_length': 778,
                                                                                                                                          'poller_hash': 'Mjk4MjMycGVlcnBvbnBvbiAzMDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '72961741',
                                                                                                                                          'fiber_length': 813,
                                                                                                                                          'poller_hash': 'MkIwM0ZBcGVlcnBvbnBvbiA0NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '21536360',
                                                                                                                                          'fiber_length': 835,
                                                                                                                                          'poller_hash': 'MzQ5RTg1cGVlcnBvbnBvbiA1MDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '688359',
                                                                                                                                          'fiber_length': 822,
                                                                                                                                          'poller_hash': 'MzAwRTc1cGVlcnBvbnBvbiA1MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '29747800',
                                                                                                                                          'fiber_length': 766,
                                                                                                                                          'poller_hash': 'MzIyNDAzcGVlcnBvbnBvbiA3MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '688358',
                                                                                                                                          'fiber_length': 889,
                                                                                                                                          'poller_hash': 'MzAwRUY4cGVlcnBvbnBvbiA4MjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '6050414',
                                                                                                                                          'fiber_length': 1088,
                                                                                                                                          'poller_hash': 'ODY5NDE4cGVlcnBvbnBvbiAxNDA2Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '21816227',
                                                                                                                                          'fiber_length': 1127,
                                                                                                                                          'poller_hash': 'QzgxRTZGcGVlcnBvbnBvbiAyMDQ5MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '6075955',
                                                                                                                                          'fiber_length': 779,
                                                                                                                                          'poller_hash': 'MTI3NTBDRHBlZXJwb25wb24gMjE4MDcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '688365',
                                                                                                                                          'fiber_length': 990,
                                                                                                                                          'poller_hash': 'MTI5NzNCNXBlZXJwb25wb24gMjU2NTk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '688362',
                                                                                                                                          'fiber_length': 1106,
                                                                                                                                          'poller_hash': 'MTI3NDdBMXBlZXJwb25wb24gMjcyMjI5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/6',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '3/7',
                                                                                         'in_discards': '1296',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1525133',
                                                                                         'in_octets': '2120835427403',
                                                                                         'in_ucast_pkts': '4474127750',
                                                                                         'links': [       {       'in_nucast_pkts': '459',
                                                                                                                  'in_octets': '62749926320',
                                                                                                                  'in_ucast_pkts': '466444616',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '35084250',
                                                                                                                  'out_octets': '2332622550828',
                                                                                                                  'out_ucast_pkts': '1617325733',
                                                                                                                  'poller_hash': 'NDJCOUVEcGVlcmludGZwb24gNTUw'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '30435597033',
                                                                                                                  'in_ucast_pkts': '130995778',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465097',
                                                                                                                  'out_octets': '394459439029',
                                                                                                                  'out_ucast_pkts': '321474202',
                                                                                                                  'poller_hash': 'MTI3NEFFOXBlZXJpbnRmcG9uIDIzMzgzOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '38496656814',
                                                                                                                  'in_ucast_pkts': '128804503',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '592686974138',
                                                                                                                  'out_ucast_pkts': '482546907',
                                                                                                                  'poller_hash': 'MTI3NEFBRHBlZXJpbnRmcG9uIDIzNDM2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '64049020388',
                                                                                                                  'in_ucast_pkts': '4636480439',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1364415834006',
                                                                                                                  'out_ucast_pkts': '1023611975',
                                                                                                                  'poller_hash': 'MjA1REMwcGVlcmludGZwb24gMjQxNzI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2048199528',
                                                                                                                  'in_ucast_pkts': '14379354',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '44483617115',
                                                                                                                  'out_ucast_pkts': '32052460',
                                                                                                                  'poller_hash': 'MTI3NDAzN3BlZXJpbnRmcG9uIDI0MTg1NQ=='},
                                                                                                          {       'in_nucast_pkts': '133',
                                                                                                                  'in_octets': '45309852481',
                                                                                                                  'in_ucast_pkts': '77839351',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1439353',
                                                                                                                  'out_octets': '293127104363',
                                                                                                                  'out_ucast_pkts': '213013907',
                                                                                                                  'poller_hash': 'MTI3NDM0MXBlZXJpbnRmcG9uIDI0MTg3NQ=='},
                                                                                                          {       'in_nucast_pkts': '156945',
                                                                                                                  'in_octets': '218890006536',
                                                                                                                  'in_ucast_pkts': '488588224',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465059',
                                                                                                                  'out_octets': '454904259724',
                                                                                                                  'out_ucast_pkts': '457122834',
                                                                                                                  'poller_hash': 'MTI3NEEwMXBlZXJpbnRmcG9uIDI0NTU3Mg=='},
                                                                                                          {       'in_nucast_pkts': '12577',
                                                                                                                  'in_octets': '15721190108',
                                                                                                                  'in_ucast_pkts': '18517931',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465330',
                                                                                                                  'out_octets': '43993550596',
                                                                                                                  'out_ucast_pkts': '37731321',
                                                                                                                  'poller_hash': 'MTI5NzNEN3BlZXJpbnRmcG9uIDI1NTg4OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '12508235323',
                                                                                                                  'in_ucast_pkts': '71406581',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '710239396238',
                                                                                                                  'out_ucast_pkts': '507261354',
                                                                                                                  'poller_hash': 'MTFCREZGNHBlZXJpbnRmcG9uIDI1NzY3OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '57193564731',
                                                                                                                  'in_ucast_pkts': '227235849',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '922816005341',
                                                                                                                  'out_ucast_pkts': '705505821',
                                                                                                                  'poller_hash': 'MTI3NDdBNnBlZXJpbnRmcG9uIDI3NDY0Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '3/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '24239744',
                                                                                         'out_octets': '13384421289142',
                                                                                         'out_ucast_pkts': '10115744445',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/7',
                                                                                                                 'id': '3/7',
                                                                                                                 'links': [       {       'connected_time': '15498714',
                                                                                                                                          'fiber_length': 1133,
                                                                                                                                          'poller_hash': 'NDJCOUVEcGVlcnBvbnBvbiA1NTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '688355',
                                                                                                                                          'fiber_length': 1139,
                                                                                                                                          'poller_hash': 'MTI3NEFFOXBlZXJwb25wb24gMjMzODM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '3971929',
                                                                                                                                          'fiber_length': 1011,
                                                                                                                                          'poller_hash': 'MTI3NEFBRHBlZXJwb25wb24gMjM0MzY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '688357',
                                                                                                                                          'fiber_length': 1131,
                                                                                                                                          'poller_hash': 'MjA1REMwcGVlcnBvbnBvbiAyNDE3MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '449345',
                                                                                                                                          'fiber_length': 1121,
                                                                                                                                          'poller_hash': 'MTI3NDAzN3BlZXJwb25wb24gMjQxODU1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '676588',
                                                                                                                                          'fiber_length': 1146,
                                                                                                                                          'poller_hash': 'MTI3NDM0MXBlZXJwb25wb24gMjQxODc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '688358',
                                                                                                                                          'fiber_length': 1123,
                                                                                                                                          'poller_hash': 'MTI3NEEwMXBlZXJwb25wb24gMjQ1NTcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '688357',
                                                                                                                                          'fiber_length': 1141,
                                                                                                                                          'poller_hash': 'MTI5NzNEN3BlZXJwb25wb24gMjU1ODg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '688358',
                                                                                                                                          'fiber_length': 1143,
                                                                                                                                          'poller_hash': 'MTFCREZGNHBlZXJwb25wb24gMjU3Njc4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769},
                                                                                                                                  {       'connected_time': '12443958',
                                                                                                                                          'fiber_length': 988,
                                                                                                                                          'poller_hash': 'MTI3NDdBNnBlZXJwb25wb24gMjc0NjQ2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.769}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/7',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '3/8',
                                                                                         'in_discards': '1433',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '224385',
                                                                                         'in_octets': '2243432079419',
                                                                                         'in_ucast_pkts': '7505787045',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '67126724320',
                                                                                                                  'in_ucast_pkts': '336828729',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1382342806805',
                                                                                                                  'out_ucast_pkts': '1067039743',
                                                                                                                  'poller_hash': 'REZGRTQxcGVlcmludGZwb24gMTgzNjc2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '11207505808',
                                                                                                                  'in_ucast_pkts': '26958749',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '49571635091',
                                                                                                                  'out_ucast_pkts': '58244569',
                                                                                                                  'poller_hash': 'REZGRTNGcGVlcmludGZwb24gMTg1MjM2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '114423833352',
                                                                                                                  'in_ucast_pkts': '447369076',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2575618660266',
                                                                                                                  'out_ucast_pkts': '1794520043',
                                                                                                                  'poller_hash': 'REZGRDlDcGVlcmludGZwb24gMTg1Mzg4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '14520245020',
                                                                                                                  'in_ucast_pkts': '43236942',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '193097894521',
                                                                                                                  'out_ucast_pkts': '156950931',
                                                                                                                  'poller_hash': 'QzgzMTA4cGVlcmludGZwb24gMTg2MzM1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '154521852554',
                                                                                                                  'in_ucast_pkts': '223167394',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1182343710051',
                                                                                                                  'out_ucast_pkts': '894853727',
                                                                                                                  'poller_hash': 'QzgzNTg2cGVlcmludGZwb24gMTk4MjM0'},
                                                                                                          {       'in_nucast_pkts': '3579702',
                                                                                                                  'in_octets': '90204660881',
                                                                                                                  'in_ucast_pkts': '964877457',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '277198017',
                                                                                                                  'out_octets': '5922838420254',
                                                                                                                  'out_ucast_pkts': '4010254831',
                                                                                                                  'poller_hash': 'QzgxRTYxcGVlcmludGZwb24gMTk4Mzg0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '12193300315',
                                                                                                                  'in_ucast_pkts': '34311446',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '85416465331',
                                                                                                                  'out_ucast_pkts': '83360488',
                                                                                                                  'poller_hash': 'QzgxREQ0cGVlcmludGZwb24gMTk4NTc5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '242963451651',
                                                                                                                  'in_ucast_pkts': '86734570812',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2845741619342',
                                                                                                                  'out_ucast_pkts': '2066748858',
                                                                                                                  'poller_hash': 'QzgxREFBcGVlcmludGZwb24gMjAwMzU1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '13780343502',
                                                                                                                  'in_ucast_pkts': '45384089',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '633474524415',
                                                                                                                  'out_ucast_pkts': '443512410',
                                                                                                                  'poller_hash': 'QzgxREMxcGVlcmludGZwb24gMjAwNDMw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '254313639796',
                                                                                                                  'in_ucast_pkts': '4534362706',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '352422748404',
                                                                                                                  'out_ucast_pkts': '333696555',
                                                                                                                  'poller_hash': 'QzgxRTVEcGVlcmludGZwb24gMjAwODY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '370390449562',
                                                                                                                  'in_ucast_pkts': '635659650',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1709963262316',
                                                                                                                  'out_ucast_pkts': '1305378022',
                                                                                                                  'poller_hash': 'QzgxRjFBcGVlcmludGZwb24gMjAyODI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '20407853315',
                                                                                                                  'in_ucast_pkts': '113961296',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '315884441967',
                                                                                                                  'out_ucast_pkts': '253315771',
                                                                                                                  'poller_hash': 'QzgxRUI3cGVlcmludGZwb24gMjAyODc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '22928764110',
                                                                                                                  'in_ucast_pkts': '36826673',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '67876652101',
                                                                                                                  'out_ucast_pkts': '70173301',
                                                                                                                  'poller_hash': 'QzgxRTM2cGVlcmludGZwb24gMjA0OTcw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '272194299947',
                                                                                                                  'in_ucast_pkts': '897610253',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3533233105464',
                                                                                                                  'out_ucast_pkts': '2745436998',
                                                                                                                  'poller_hash': 'QzgxRTUwcGVlcmludGZwb24gMjA1MDcw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '12756133463',
                                                                                                                  'in_ucast_pkts': '47585746',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '173997181358',
                                                                                                                  'out_ucast_pkts': '153966442',
                                                                                                                  'poller_hash': 'MTFCMjRGN3BlZXJpbnRmcG9uIDIxMjIwNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '64411257135',
                                                                                                                  'in_ucast_pkts': '272338851',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1284120723732',
                                                                                                                  'out_ucast_pkts': '898851951',
                                                                                                                  'poller_hash': 'MTFCMjI3QXBlZXJpbnRmcG9uIDIxMjQ1Nw=='},
                                                                                                          {       'in_nucast_pkts': '14799',
                                                                                                                  'in_octets': '9565012653',
                                                                                                                  'in_ucast_pkts': '76138734',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464742',
                                                                                                                  'out_octets': '289507400617',
                                                                                                                  'out_ucast_pkts': '206715935',
                                                                                                                  'poller_hash': 'MTE5NTE3RXBlZXJpbnRmcG9uIDIxMzc5MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '6912990883',
                                                                                                                  'in_ucast_pkts': '23463037',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '123408950684',
                                                                                                                  'out_ucast_pkts': '107319026',
                                                                                                                  'poller_hash': 'MTI3NDE4RnBlZXJpbnRmcG9uIDI0NDg3MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '119202698810',
                                                                                                                  'in_ucast_pkts': '917440662',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2987252149522',
                                                                                                                  'out_ucast_pkts': '2159398981',
                                                                                                                  'poller_hash': 'MTE3REIzOXBlZXJpbnRmcG9uIDI1MDAzNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '15961784857',
                                                                                                                  'in_ucast_pkts': '28934145',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '74653534510',
                                                                                                                  'out_ucast_pkts': '75092596',
                                                                                                                  'poller_hash': 'MTI5NzNGMHBlZXJpbnRmcG9uIDI2NjMzMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '85367984980',
                                                                                                                  'in_ucast_pkts': '567973699',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1494757532031',
                                                                                                                  'out_ucast_pkts': '1062950934',
                                                                                                                  'poller_hash': 'MTI5NjA0NHBlZXJpbnRmcG9uIDI4MDIyNA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '3/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '217833586',
                                                                                         'out_octets': '27956031697036',
                                                                                         'out_ucast_pkts': '20512748151',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '3/8',
                                                                                                                 'id': '3/8',
                                                                                                                 'links': [       {       'connected_time': '21815991',
                                                                                                                                          'fiber_length': 11481,
                                                                                                                                          'poller_hash': 'REZGRTQxcGVlcnBvbnBvbiAxODM2NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '14746251',
                                                                                                                                          'fiber_length': 11066,
                                                                                                                                          'poller_hash': 'REZGRTNGcGVlcnBvbnBvbiAxODUyMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '68525743',
                                                                                                                                          'fiber_length': 11150,
                                                                                                                                          'poller_hash': 'REZGRDlDcGVlcnBvbnBvbiAxODUzODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '21815895',
                                                                                                                                          'fiber_length': 11417,
                                                                                                                                          'poller_hash': 'QzgzMTA4cGVlcnBvbnBvbiAxODYzMzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '523090',
                                                                                                                                          'fiber_length': 11461,
                                                                                                                                          'poller_hash': 'QzgzNTg2cGVlcnBvbnBvbiAxOTgyMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '29739904',
                                                                                                                                          'fiber_length': 11720,
                                                                                                                                          'poller_hash': 'QzgxRTYxcGVlcnBvbnBvbiAxOTgzODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '21816077',
                                                                                                                                          'fiber_length': 11905,
                                                                                                                                          'poller_hash': 'QzgxREQ0cGVlcnBvbnBvbiAxOTg1Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '21815971',
                                                                                                                                          'fiber_length': 11896,
                                                                                                                                          'poller_hash': 'QzgxREFBcGVlcnBvbnBvbiAyMDAzNTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '688360',
                                                                                                                                          'fiber_length': 11708,
                                                                                                                                          'poller_hash': 'QzgxREMxcGVlcnBvbnBvbiAyMDA0MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '1913956',
                                                                                                                                          'fiber_length': 11818,
                                                                                                                                          'poller_hash': 'QzgxRTVEcGVlcnBvbnBvbiAyMDA4NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '688354',
                                                                                                                                          'fiber_length': 11518,
                                                                                                                                          'poller_hash': 'QzgxRjFBcGVlcnBvbnBvbiAyMDI4Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '21815859',
                                                                                                                                          'fiber_length': 11039,
                                                                                                                                          'poller_hash': 'QzgxRUI3cGVlcnBvbnBvbiAyMDI4Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '21815545',
                                                                                                                                          'fiber_length': 11872,
                                                                                                                                          'poller_hash': 'QzgxRTM2cGVlcnBvbnBvbiAyMDQ5NzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '10014597',
                                                                                                                                          'fiber_length': 11100,
                                                                                                                                          'poller_hash': 'QzgxRTUwcGVlcnBvbnBvbiAyMDUwNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '511297',
                                                                                                                                          'fiber_length': 11246,
                                                                                                                                          'poller_hash': 'MTFCMjRGN3BlZXJwb25wb24gMjEyMjA2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '16596261',
                                                                                                                                          'fiber_length': 11480,
                                                                                                                                          'poller_hash': 'MTFCMjI3QXBlZXJwb25wb24gMjEyNDU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '688393',
                                                                                                                                          'fiber_length': 11818,
                                                                                                                                          'poller_hash': 'MTE5NTE3RXBlZXJwb25wb24gMjEzNzkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '14067777',
                                                                                                                                          'fiber_length': 11906,
                                                                                                                                          'poller_hash': 'MTI3NDE4RnBlZXJwb25wb24gMjQ0ODcw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '523101',
                                                                                                                                          'fiber_length': 11498,
                                                                                                                                          'poller_hash': 'MTE3REIzOXBlZXJwb25wb24gMjUwMDM0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '523099',
                                                                                                                                          'fiber_length': 11041,
                                                                                                                                          'poller_hash': 'MTI5NzNGMHBlZXJwb25wb24gMjY2MzMw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094},
                                                                                                                                  {       'connected_time': '688355',
                                                                                                                                          'fiber_length': 11792,
                                                                                                                                          'poller_hash': 'MTI5NjA0NHBlZXJwb25wb24gMjgwMjI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.094}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '3/8',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNTEyNjAzNzE3',
                                                           'serial_number': '211512603717',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '6176064'},
                                                   {       'id': '4',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '4/1',
                                                                                         'in_discards': '273766',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '27674929',
                                                                                         'in_octets': '26835709783078',
                                                                                         'in_ucast_pkts': '93216235563',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTYxMDI3cGVlcmludGZwb24gMTY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '235229225999',
                                                                                                                  'in_ucast_pkts': '2211419928',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11867083853474',
                                                                                                                  'out_ucast_pkts': '8220990337',
                                                                                                                  'poller_hash': 'MTU1M0M2cGVlcmludGZwb24gMTY2'},
                                                                                                          {       'in_nucast_pkts': '52021',
                                                                                                                  'in_octets': '23344697215',
                                                                                                                  'in_ucast_pkts': '138998116',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5062359',
                                                                                                                  'out_octets': '505698862636',
                                                                                                                  'out_ucast_pkts': '361813972',
                                                                                                                  'poller_hash': 'MTU1NDYwcGVlcmludGZwb24gMTY3'},
                                                                                                          {       'in_nucast_pkts': '495801',
                                                                                                                  'in_octets': '841834746815',
                                                                                                                  'in_ucast_pkts': '1543522898',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62167719',
                                                                                                                  'out_octets': '7804345623434',
                                                                                                                  'out_ucast_pkts': '6034653188',
                                                                                                                  'poller_hash': 'MkIwMjk4cGVlcmludGZwb24gMTY4'},
                                                                                                          {       'in_nucast_pkts': '693220',
                                                                                                                  'in_octets': '285064104106',
                                                                                                                  'in_ucast_pkts': '2243418535',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71945945',
                                                                                                                  'out_octets': '7889520996466',
                                                                                                                  'out_ucast_pkts': '5566846451',
                                                                                                                  'poller_hash': 'MTU1M0JBcGVlcmludGZwb24gMTY5'},
                                                                                                          {       'in_nucast_pkts': '4515864',
                                                                                                                  'in_octets': '5390449621639',
                                                                                                                  'in_ucast_pkts': '10820628211',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '171940664',
                                                                                                                  'out_octets': '46089137978319',
                                                                                                                  'out_ucast_pkts': '33412871825',
                                                                                                                  'poller_hash': 'NTVGNTM3cGVlcmludGZwb24gMTcw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '121219855624',
                                                                                                                  'in_ucast_pkts': '538310119',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2034673089853',
                                                                                                                  'out_ucast_pkts': '1623148562',
                                                                                                                  'poller_hash': 'MUM1RDE4cGVlcmludGZwb24gMTcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '224166089019',
                                                                                                                  'in_ucast_pkts': '1804314409',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4833385123037',
                                                                                                                  'out_ucast_pkts': '1072283128',
                                                                                                                  'poller_hash': 'MTU1M0JGcGVlcmludGZwb24gNDkw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1068491653712',
                                                                                                                  'in_ucast_pkts': '2387367482',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '22448255535000',
                                                                                                                  'out_ucast_pkts': '16014211482',
                                                                                                                  'poller_hash': 'NDI4MUM2cGVlcmludGZwb24gNTM1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '670633589940',
                                                                                                                  'in_ucast_pkts': '149252515727',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '13311437601478',
                                                                                                                  'out_ucast_pkts': '10445343543',
                                                                                                                  'poller_hash': 'NDI4NDRDcGVlcmludGZwb24gNTM3'},
                                                                                                          {       'in_nucast_pkts': '54011',
                                                                                                                  'in_octets': '23194716933',
                                                                                                                  'in_ucast_pkts': '109200478',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6782021',
                                                                                                                  'out_octets': '398968497561',
                                                                                                                  'out_ucast_pkts': '314532842',
                                                                                                                  'poller_hash': 'ODY5NDI0cGVlcmludGZwb24gNTc0'},
                                                                                                          {       'in_nucast_pkts': '31',
                                                                                                                  'in_octets': '80418833084',
                                                                                                                  'in_ucast_pkts': '187845635',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6782058',
                                                                                                                  'out_octets': '588321884380',
                                                                                                                  'out_ucast_pkts': '464457459',
                                                                                                                  'poller_hash': 'MzVFOTM2cGVlcmludGZwb24gNzgy'},
                                                                                                          {       'in_nucast_pkts': '2754',
                                                                                                                  'in_octets': '13109530874',
                                                                                                                  'in_ucast_pkts': '25826247',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6443563',
                                                                                                                  'out_octets': '81088845913',
                                                                                                                  'out_ucast_pkts': '67198824',
                                                                                                                  'poller_hash': 'NUE2NjYzcGVlcmludGZwb24gODI0'},
                                                                                                          {       'in_nucast_pkts': '11204',
                                                                                                                  'in_octets': '206490459383',
                                                                                                                  'in_ucast_pkts': '1443954117',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '150650055',
                                                                                                                  'out_octets': '8121686598315',
                                                                                                                  'out_ucast_pkts': '5625215652',
                                                                                                                  'poller_hash': 'NDJDMDQ4cGVlcmludGZwb24gODI1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '499475773766',
                                                                                                                  'in_ucast_pkts': '4263014785',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '140',
                                                                                                                  'out_octets': '13974742213691',
                                                                                                                  'out_ucast_pkts': '9725663706',
                                                                                                                  'poller_hash': 'NDJDNEM5cGVlcmludGZwb24gODI5'},
                                                                                                          {       'in_nucast_pkts': '86584',
                                                                                                                  'in_octets': '87193062249',
                                                                                                                  'in_ucast_pkts': '532479831',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5224046',
                                                                                                                  'out_octets': '1677918461929',
                                                                                                                  'out_ucast_pkts': '1238787631',
                                                                                                                  'poller_hash': 'NjQwQjM1cGVlcmludGZwb24gODM4'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '4649414274',
                                                                                                                  'in_ucast_pkts': '17097234',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3187206',
                                                                                                                  'out_octets': '91802601251',
                                                                                                                  'out_ucast_pkts': '65272219',
                                                                                                                  'poller_hash': 'NDJDNTk1cGVlcmludGZwb24gODQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '151263615691',
                                                                                                                  'in_ucast_pkts': '387275792',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '168',
                                                                                                                  'out_octets': '1506341880046',
                                                                                                                  'out_ucast_pkts': '1241717014',
                                                                                                                  'poller_hash': 'ODY5Mzg1cGVlcmludGZwb24gODcw'},
                                                                                                          {       'in_nucast_pkts': '835',
                                                                                                                  'in_octets': '22464188834',
                                                                                                                  'in_ucast_pkts': '203897243',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4959035',
                                                                                                                  'out_octets': '1361173533313',
                                                                                                                  'out_ucast_pkts': '984511694',
                                                                                                                  'poller_hash': 'NDNDNDJDcGVlcmludGZwb24gODc0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '429341258370',
                                                                                                                  'in_ucast_pkts': '26928436449',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '119',
                                                                                                                  'out_octets': '3330760003758',
                                                                                                                  'out_ucast_pkts': '2619571145',
                                                                                                                  'poller_hash': 'ODY5NDA4cGVlcmludGZwb24gMTQxMDQx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '124029822675',
                                                                                                                  'in_ucast_pkts': '571355331',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1',
                                                                                                                  'out_octets': '4859318172377',
                                                                                                                  'out_ucast_pkts': '3488857195',
                                                                                                                  'poller_hash': 'QTlDNzZGcGVlcmludGZwb24gMTQ2ODE1'},
                                                                                                          {       'in_nucast_pkts': '67732',
                                                                                                                  'in_octets': '6826278755',
                                                                                                                  'in_ucast_pkts': '39386224',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6782122',
                                                                                                                  'out_octets': '122484660432',
                                                                                                                  'out_ucast_pkts': '94322323',
                                                                                                                  'poller_hash': 'QTlDOEU2cGVlcmludGZwb24gMTQ2OTM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '507166049595',
                                                                                                                  'in_ucast_pkts': '36619831759',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '47',
                                                                                                                  'out_octets': '9333350595378',
                                                                                                                  'out_ucast_pkts': '6810867559',
                                                                                                                  'poller_hash': 'QTlDM0QycGVlcmludGZwb24gMTQ3MDQ3'},
                                                                                                          {       'in_nucast_pkts': '68322',
                                                                                                                  'in_octets': '41314882320',
                                                                                                                  'in_ucast_pkts': '105183159',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6782051',
                                                                                                                  'out_octets': '975689852294',
                                                                                                                  'out_ucast_pkts': '695763092',
                                                                                                                  'poller_hash': 'QTlDM0M2cGVlcmludGZwb24gMTQ3MDU3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '430726391380',
                                                                                                                  'in_ucast_pkts': '82335425455',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '147',
                                                                                                                  'out_octets': '2753426688244',
                                                                                                                  'out_ucast_pkts': '2242940688',
                                                                                                                  'poller_hash': 'QTlDOUMwcGVlcmludGZwb24gMTQ3MDg3'},
                                                                                                          {       'in_nucast_pkts': '56291',
                                                                                                                  'in_octets': '185719339003',
                                                                                                                  'in_ucast_pkts': '161646148',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6781892',
                                                                                                                  'out_octets': '196033823123',
                                                                                                                  'out_ucast_pkts': '211262605',
                                                                                                                  'poller_hash': 'QTlDOEU1cGVlcmludGZwb24gMTQ5MTI3'},
                                                                                                          {       'in_nucast_pkts': '66835',
                                                                                                                  'in_octets': '61977279961',
                                                                                                                  'in_ucast_pkts': '321198641',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6781920',
                                                                                                                  'out_octets': '1536095483149',
                                                                                                                  'out_ucast_pkts': '1168103753',
                                                                                                                  'poller_hash': 'Mjk4MjREcGVlcmludGZwb24gMTc3MjY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '82759220394',
                                                                                                                  'in_ucast_pkts': '401896319',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '168',
                                                                                                                  'out_octets': '2541203125015',
                                                                                                                  'out_ucast_pkts': '2062066654',
                                                                                                                  'poller_hash': 'QzgzQkY1cGVlcmludGZwb24gMTgwMTY3'},
                                                                                                          {       'in_nucast_pkts': '66951',
                                                                                                                  'in_octets': '19630593708',
                                                                                                                  'in_ucast_pkts': '112356840',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6781939',
                                                                                                                  'out_octets': '523511833952',
                                                                                                                  'out_ucast_pkts': '378514507',
                                                                                                                  'poller_hash': 'REZGRTAxcGVlcmludGZwb24gMTgyMzQ1'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '4/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '152812096363',
                                                                                         'out_octets': '583047520357293',
                                                                                         'out_ucast_pkts': '274419474740',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/1',
                                                                                                                 'id': '4/1',
                                                                                                                 'links': [       {       'connected_time': '24669181',
                                                                                                                                          'fiber_length': 18088,
                                                                                                                                          'poller_hash': 'MTYxMDI3cGVlcnBvbnBvbiAxNjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '47673',
                                                                                                                                          'fiber_length': 18000,
                                                                                                                                          'poller_hash': 'MTU1M0M2cGVlcnBvbnBvbiAxNjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '2969248',
                                                                                                                                          'fiber_length': 17601,
                                                                                                                                          'poller_hash': 'MTU1NDYwcGVlcnBvbnBvbiAxNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '25995562',
                                                                                                                                          'fiber_length': 17964,
                                                                                                                                          'poller_hash': 'MkIwMjk4cGVlcnBvbnBvbiAxNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '69226272',
                                                                                                                                          'fiber_length': 18015,
                                                                                                                                          'poller_hash': 'MTU1M0JBcGVlcnBvbnBvbiAxNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '18082129',
                                                                                                                                          'fiber_length': 18437,
                                                                                                                                          'poller_hash': 'NTVGNTM3cGVlcnBvbnBvbiAxNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '19619401',
                                                                                                                                          'fiber_length': 18461,
                                                                                                                                          'poller_hash': 'MUM1RDE4cGVlcnBvbnBvbiAxNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '47253',
                                                                                                                                          'fiber_length': 17584,
                                                                                                                                          'poller_hash': 'MTU1M0JGcGVlcnBvbnBvbiA0OTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '1131830',
                                                                                                                                          'fiber_length': 18593,
                                                                                                                                          'poller_hash': 'NDI4MUM2cGVlcnBvbnBvbiA1MzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '13124654',
                                                                                                                                          'fiber_length': 18297,
                                                                                                                                          'poller_hash': 'NDI4NDRDcGVlcnBvbnBvbiA1Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113880',
                                                                                                                                          'fiber_length': 17483,
                                                                                                                                          'poller_hash': 'ODY5NDI0cGVlcnBvbnBvbiA1NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113915',
                                                                                                                                          'fiber_length': 18800,
                                                                                                                                          'poller_hash': 'MzVFOTM2cGVlcnBvbnBvbiA3ODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '2969325',
                                                                                                                                          'fiber_length': 18200,
                                                                                                                                          'poller_hash': 'NUE2NjYzcGVlcnBvbnBvbiA4MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '21816174',
                                                                                                                                          'fiber_length': 17534,
                                                                                                                                          'poller_hash': 'NDJDMDQ4cGVlcnBvbnBvbiA4MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '11726986',
                                                                                                                                          'fiber_length': 18483,
                                                                                                                                          'poller_hash': 'NDJDNEM5cGVlcnBvbnBvbiA4Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113778',
                                                                                                                                          'fiber_length': 18444,
                                                                                                                                          'poller_hash': 'NjQwQjM1cGVlcnBvbnBvbiA4Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '1478640',
                                                                                                                                          'fiber_length': 18505,
                                                                                                                                          'poller_hash': 'NDJDNTk1cGVlcnBvbnBvbiA4NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113822',
                                                                                                                                          'fiber_length': 17582,
                                                                                                                                          'poller_hash': 'ODY5Mzg1cGVlcnBvbnBvbiA4NzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '2969198',
                                                                                                                                          'fiber_length': 17691,
                                                                                                                                          'poller_hash': 'NDNDNDJDcGVlcnBvbnBvbiA4NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '72955401',
                                                                                                                                          'fiber_length': 17977,
                                                                                                                                          'poller_hash': 'ODY5NDA4cGVlcnBvbnBvbiAxNDEwNDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '2969243',
                                                                                                                                          'fiber_length': 17827,
                                                                                                                                          'poller_hash': 'QTlDNzZGcGVlcnBvbnBvbiAxNDY4MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113850',
                                                                                                                                          'fiber_length': 17505,
                                                                                                                                          'poller_hash': 'QTlDOEU2cGVlcnBvbnBvbiAxNDY5Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '2015390',
                                                                                                                                          'fiber_length': 17958,
                                                                                                                                          'poller_hash': 'QTlDM0QycGVlcnBvbnBvbiAxNDcwNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113826',
                                                                                                                                          'fiber_length': 18318,
                                                                                                                                          'poller_hash': 'QTlDM0M2cGVlcnBvbnBvbiAxNDcwNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '72955601',
                                                                                                                                          'fiber_length': 18167,
                                                                                                                                          'poller_hash': 'QTlDOUMwcGVlcnBvbnBvbiAxNDcwODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113845',
                                                                                                                                          'fiber_length': 18838,
                                                                                                                                          'poller_hash': 'QTlDOEU1cGVlcnBvbnBvbiAxNDkxMjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113836',
                                                                                                                                          'fiber_length': 17559,
                                                                                                                                          'poller_hash': 'Mjk4MjREcGVlcnBvbnBvbiAxNzcyNjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113802',
                                                                                                                                          'fiber_length': 17705,
                                                                                                                                          'poller_hash': 'QzgzQkY1cGVlcnBvbnBvbiAxODAxNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133},
                                                                                                                                  {       'connected_time': '3113810',
                                                                                                                                          'fiber_length': 18301,
                                                                                                                                          'poller_hash': 'REZGRTAxcGVlcnBvbnBvbiAxODIzNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.133}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/1',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '4/2',
                                                                                         'in_discards': '13312',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3612021',
                                                                                         'in_octets': '5800894165215',
                                                                                         'in_ucast_pkts': '23091367002',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '142550186599',
                                                                                                                  'in_ucast_pkts': '973956642',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '29',
                                                                                                                  'out_octets': '4200341330739',
                                                                                                                  'out_ucast_pkts': '3035075336',
                                                                                                                  'poller_hash': 'MTI3NDMyMHBlZXJpbnRmcG9uIDI0ODE0OA=='},
                                                                                                          {       'in_nucast_pkts': '9760',
                                                                                                                  'in_octets': '4458023950',
                                                                                                                  'in_ucast_pkts': '20010700',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151530',
                                                                                                                  'out_octets': '95094135465',
                                                                                                                  'out_ucast_pkts': '68339439',
                                                                                                                  'poller_hash': 'MTE3REFCOHBlZXJpbnRmcG9uIDI0OTIwMw=='},
                                                                                                          {       'in_nucast_pkts': '11590',
                                                                                                                  'in_octets': '4799404396',
                                                                                                                  'in_ucast_pkts': '29984667',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151348',
                                                                                                                  'out_octets': '193278073945',
                                                                                                                  'out_ucast_pkts': '133354499',
                                                                                                                  'poller_hash': 'MTBBODZBRHBlZXJpbnRmcG9uIDI0OTIyMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '220598880140',
                                                                                                                  'in_ucast_pkts': '951749113',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '57',
                                                                                                                  'out_octets': '10298633928438',
                                                                                                                  'out_ucast_pkts': '2914650736',
                                                                                                                  'poller_hash': 'MTE3REI0OHBlZXJpbnRmcG9uIDI1MDE0NA=='},
                                                                                                          {       'in_nucast_pkts': '11485',
                                                                                                                  'in_octets': '7134496001',
                                                                                                                  'in_ucast_pkts': '12722342',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151478',
                                                                                                                  'out_octets': '53813366910',
                                                                                                                  'out_ucast_pkts': '41525180',
                                                                                                                  'poller_hash': 'MTE3REFDNnBlZXJpbnRmcG9uIDI1MDI0Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDBDNnBlZXJpbnRmcG9uIDI1MDI1Nw=='},
                                                                                                          {       'in_nucast_pkts': '556573',
                                                                                                                  'in_octets': '1522043202725',
                                                                                                                  'in_ucast_pkts': '4382952624',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70727695',
                                                                                                                  'out_octets': '16993377415206',
                                                                                                                  'out_ucast_pkts': '12347163019',
                                                                                                                  'poller_hash': 'MTE3REI3QXBlZXJpbnRmcG9uIDI1MTgxNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '199656288604',
                                                                                                                  'in_ucast_pkts': '1511768662',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '7224961585690',
                                                                                                                  'out_ucast_pkts': '5136990995',
                                                                                                                  'poller_hash': 'MTE3REExRHBlZXJpbnRmcG9uIDI1MzM5Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '268386125620',
                                                                                                                  'in_ucast_pkts': '34888193610',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '1944950496886',
                                                                                                                  'out_ucast_pkts': '1588880632',
                                                                                                                  'poller_hash': 'MTE3REFENHBlZXJpbnRmcG9uIDI1NDAyMg=='},
                                                                                                          {       'in_nucast_pkts': '6020',
                                                                                                                  'in_octets': '2110822379',
                                                                                                                  'in_ucast_pkts': '8734020',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '690126',
                                                                                                                  'out_octets': '63154248071',
                                                                                                                  'out_ucast_pkts': '43433248',
                                                                                                                  'poller_hash': 'MTE3REFEQ3BlZXJpbnRmcG9uIDI1NDU5Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '56351041477',
                                                                                                                  'in_ucast_pkts': '243847697',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '1037449028476',
                                                                                                                  'out_ucast_pkts': '815697476',
                                                                                                                  'poller_hash': 'MTE3REJFN3BlZXJpbnRmcG9uIDI1NDczNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '85807266371',
                                                                                                                  'in_ucast_pkts': '372378688',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '2274862873802',
                                                                                                                  'out_ucast_pkts': '1687085022',
                                                                                                                  'poller_hash': 'MTI5NzNGM3BlZXJpbnRmcG9uIDI1NjI5Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '253829124417',
                                                                                                                  'in_ucast_pkts': '593917159',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '2421897280533',
                                                                                                                  'out_ucast_pkts': '1920057416',
                                                                                                                  'poller_hash': 'MTI5NzQxQnBlZXJpbnRmcG9uIDI1Nzc2OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '156947775074',
                                                                                                                  'in_ucast_pkts': '452910833',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '2947640596519',
                                                                                                                  'out_ucast_pkts': '2164368684',
                                                                                                                  'poller_hash': 'MTI5NzI4N3BlZXJpbnRmcG9uIDI1NzkyOQ=='},
                                                                                                          {       'in_nucast_pkts': '9630',
                                                                                                                  'in_octets': '833120305',
                                                                                                                  'in_ucast_pkts': '3024575',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151535',
                                                                                                                  'out_octets': '5911150371',
                                                                                                                  'out_ucast_pkts': '5637661',
                                                                                                                  'poller_hash': 'MTI5NzI2OHBlZXJpbnRmcG9uIDI1ODI2Mw=='},
                                                                                                          {       'in_nucast_pkts': '308643',
                                                                                                                  'in_octets': '35578357865',
                                                                                                                  'in_ucast_pkts': '113373992',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39970992',
                                                                                                                  'out_octets': '524069379735',
                                                                                                                  'out_ucast_pkts': '441328923',
                                                                                                                  'poller_hash': 'MTI3NDcyMnBlZXJpbnRmcG9uIDI2NTY5Nw=='},
                                                                                                          {       'in_nucast_pkts': '9557',
                                                                                                                  'in_octets': '11427437360',
                                                                                                                  'in_ucast_pkts': '55988585',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151496',
                                                                                                                  'out_octets': '296804632211',
                                                                                                                  'out_ucast_pkts': '212321408',
                                                                                                                  'poller_hash': 'MTI3NDZDRHBlZXJpbnRmcG9uIDI2NjEwNw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '4/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '99390899',
                                                                                         'out_octets': '123264853639400',
                                                                                         'out_ucast_pkts': '88430867638',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/2',
                                                                                                                 'id': '4/2',
                                                                                                                 'links': [       {       'connected_time': '21472',
                                                                                                                                          'fiber_length': 21244,
                                                                                                                                          'poller_hash': 'MTI3NDMyMHBlZXJwb25wb24gMjQ4MTQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538374',
                                                                                                                                          'fiber_length': 22126,
                                                                                                                                          'poller_hash': 'MTE3REFCOHBlZXJwb25wb24gMjQ5MjAz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538379',
                                                                                                                                          'fiber_length': 20363,
                                                                                                                                          'poller_hash': 'MTBBODZBRHBlZXJwb25wb24gMjQ5MjIz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '64132371',
                                                                                                                                          'fiber_length': 20505,
                                                                                                                                          'poller_hash': 'MTE3REI0OHBlZXJwb25wb24gMjUwMTQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538379',
                                                                                                                                          'fiber_length': 20723,
                                                                                                                                          'poller_hash': 'MTE3REFDNnBlZXJwb25wb24gMjUwMjQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '64044936',
                                                                                                                                          'fiber_length': 22154,
                                                                                                                                          'poller_hash': 'MTI3NDBDNnBlZXJwb25wb24gMjUwMjU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '30330383',
                                                                                                                                          'fiber_length': 20155,
                                                                                                                                          'poller_hash': 'MTE3REI3QXBlZXJwb25wb24gMjUxODE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538378',
                                                                                                                                          'fiber_length': 24757,
                                                                                                                                          'poller_hash': 'MTE3REExRHBlZXJwb25wb24gMjUzMzk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538376',
                                                                                                                                          'fiber_length': 20673,
                                                                                                                                          'poller_hash': 'MTE3REFENHBlZXJwb25wb24gMjU0MDIy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '334508',
                                                                                                                                          'fiber_length': 21416,
                                                                                                                                          'poller_hash': 'MTE3REFEQ3BlZXJwb25wb24gMjU0NTkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538376',
                                                                                                                                          'fiber_length': 24561,
                                                                                                                                          'poller_hash': 'MTE3REJFN3BlZXJwb25wb24gMjU0NzM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538368',
                                                                                                                                          'fiber_length': 21078,
                                                                                                                                          'poller_hash': 'MTI5NzNGM3BlZXJwb25wb24gMjU2Mjky',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538373',
                                                                                                                                          'fiber_length': 24373,
                                                                                                                                          'poller_hash': 'MTI5NzQxQnBlZXJwb25wb24gMjU3NzY4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538371',
                                                                                                                                          'fiber_length': 24681,
                                                                                                                                          'poller_hash': 'MTI5NzI4N3BlZXJwb25wb24gMjU3OTI5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538377',
                                                                                                                                          'fiber_length': 24193,
                                                                                                                                          'poller_hash': 'MTI5NzI2OHBlZXJwb25wb24gMjU4MjYz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '17471855',
                                                                                                                                          'fiber_length': 19400,
                                                                                                                                          'poller_hash': 'MTI3NDcyMnBlZXJwb25wb24gMjY1Njk3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061},
                                                                                                                                  {       'connected_time': '538377',
                                                                                                                                          'fiber_length': 25365,
                                                                                                                                          'poller_hash': 'MTI3NDZDRHBlZXJwb25wb24gMjY2MTA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.061}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/2',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '4/3',
                                                                                         'in_discards': '78620',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '9439937',
                                                                                         'in_octets': '13576764488350',
                                                                                         'in_ucast_pkts': '59810669289',
                                                                                         'links': [       {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '19082349991',
                                                                                                                  'in_ucast_pkts': '60054411',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464401',
                                                                                                                  'out_octets': '301368232894',
                                                                                                                  'out_ucast_pkts': '210551882',
                                                                                                                  'poller_hash': 'QzgxMUZFcGVlcmludGZwb24gMTU2OTUx'},
                                                                                                          {       'in_nucast_pkts': '1003',
                                                                                                                  'in_octets': '224107518321',
                                                                                                                  'in_ucast_pkts': '2146511638',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '34480913',
                                                                                                                  'out_octets': '14172092545986',
                                                                                                                  'out_ucast_pkts': '9472786524',
                                                                                                                  'poller_hash': 'Q0JEMTlEcGVlcmludGZwb24gMTY4OTQx'},
                                                                                                          {       'in_nucast_pkts': '6574',
                                                                                                                  'in_octets': '567362730',
                                                                                                                  'in_ucast_pkts': '2455557',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '772208',
                                                                                                                  'out_octets': '23442932646',
                                                                                                                  'out_ucast_pkts': '18067540',
                                                                                                                  'poller_hash': 'QzgzQkI4cGVlcmludGZwb24gMTcwNzQy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '354652191154',
                                                                                                                  'in_ucast_pkts': '1328745384',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '5078768878050',
                                                                                                                  'out_ucast_pkts': '3825240824',
                                                                                                                  'poller_hash': 'QzgzMTRBcGVlcmludGZwb24gMTcyMjY5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '516305446523',
                                                                                                                  'in_ucast_pkts': '79724217018',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '86',
                                                                                                                  'out_octets': '8098216499401',
                                                                                                                  'out_ucast_pkts': '5704130253',
                                                                                                                  'poller_hash': 'QzgzQkVEcGVlcmludGZwb24gMTc1MjE0'},
                                                                                                          {       'in_nucast_pkts': '11966',
                                                                                                                  'in_octets': '8257788770',
                                                                                                                  'in_ucast_pkts': '47271769',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464468',
                                                                                                                  'out_octets': '297421827931',
                                                                                                                  'out_ucast_pkts': '215697682',
                                                                                                                  'poller_hash': 'QzgzMDdFcGVlcmludGZwb24gMTc1ODU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1063976737495',
                                                                                                                  'in_ucast_pkts': '2223679454',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70',
                                                                                                                  'out_octets': '7054263226137',
                                                                                                                  'out_ucast_pkts': '5189420177',
                                                                                                                  'poller_hash': 'REZGREU4cGVlcmludGZwb24gMTgwNzY0'},
                                                                                                          {       'in_nucast_pkts': '12768',
                                                                                                                  'in_octets': '14287321011',
                                                                                                                  'in_ucast_pkts': '74465891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464414',
                                                                                                                  'out_octets': '508182822283',
                                                                                                                  'out_ucast_pkts': '361073827',
                                                                                                                  'poller_hash': 'QzgxREVDcGVlcmludGZwb24gMjA1Mzc2'},
                                                                                                          {       'in_nucast_pkts': '12176',
                                                                                                                  'in_octets': '13874742607',
                                                                                                                  'in_ucast_pkts': '69030630',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464456',
                                                                                                                  'out_octets': '254688242062',
                                                                                                                  'out_ucast_pkts': '195684526',
                                                                                                                  'poller_hash': 'QzgxRDdDcGVlcmludGZwb24gMjA1NTM2'},
                                                                                                          {       'in_nucast_pkts': '13859',
                                                                                                                  'in_octets': '13202304341',
                                                                                                                  'in_ucast_pkts': '35863505',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464368',
                                                                                                                  'out_octets': '106459896686',
                                                                                                                  'out_ucast_pkts': '83688118',
                                                                                                                  'poller_hash': 'MTE5NTE2OXBlZXJpbnRmcG9uIDIxMTgwNQ=='},
                                                                                                          {       'in_nucast_pkts': '14795',
                                                                                                                  'in_octets': '6666357485',
                                                                                                                  'in_ucast_pkts': '28796125',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464398',
                                                                                                                  'out_octets': '106255835577',
                                                                                                                  'out_ucast_pkts': '80422526',
                                                                                                                  'poller_hash': 'MTE5NTFCNnBlZXJpbnRmcG9uIDIxNjczMg=='},
                                                                                                          {       'in_nucast_pkts': '12066',
                                                                                                                  'in_octets': '6957140903',
                                                                                                                  'in_ucast_pkts': '24910329',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464322',
                                                                                                                  'out_octets': '244623327780',
                                                                                                                  'out_ucast_pkts': '171797600',
                                                                                                                  'poller_hash': 'MTI3NTE4RnBlZXJpbnRmcG9uIDIxODQ2Ng=='},
                                                                                                          {       'in_nucast_pkts': '11996',
                                                                                                                  'in_octets': '10996664098',
                                                                                                                  'in_ucast_pkts': '23695548',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464635',
                                                                                                                  'out_octets': '125443874257',
                                                                                                                  'out_ucast_pkts': '94061558',
                                                                                                                  'poller_hash': 'MTI3NEFGNXBlZXJpbnRmcG9uIDIzMDAxOQ=='},
                                                                                                          {       'in_nucast_pkts': '14957',
                                                                                                                  'in_octets': '3567755315',
                                                                                                                  'in_ucast_pkts': '20607622',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464888',
                                                                                                                  'out_octets': '83165836258',
                                                                                                                  'out_ucast_pkts': '62055901',
                                                                                                                  'poller_hash': 'MTI3NDlBNXBlZXJpbnRmcG9uIDIzNTEyMg=='},
                                                                                                          {       'in_nucast_pkts': '12253',
                                                                                                                  'in_octets': '14746860714',
                                                                                                                  'in_ucast_pkts': '87296061',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465166',
                                                                                                                  'out_octets': '542406624599',
                                                                                                                  'out_ucast_pkts': '399673695',
                                                                                                                  'poller_hash': 'MTI3NDE5RHBlZXJpbnRmcG9uIDI0MzIzOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '81699210715',
                                                                                                                  'in_ucast_pkts': '597133113',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1',
                                                                                                                  'out_octets': '2894855343354',
                                                                                                                  'out_ucast_pkts': '2031580405',
                                                                                                                  'poller_hash': 'MTI3NDA1RnBlZXJpbnRmcG9uIDI0NzExNQ=='},
                                                                                                          {       'in_nucast_pkts': '310230',
                                                                                                                  'in_octets': '278465944530',
                                                                                                                  'in_ucast_pkts': '1237054984',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '35084752',
                                                                                                                  'out_octets': '8017314721521',
                                                                                                                  'out_ucast_pkts': '5795179460',
                                                                                                                  'poller_hash': 'MTI3NDE5RXBlZXJpbnRmcG9uIDI0Nzc1NA=='},
                                                                                                          {       'in_nucast_pkts': '12111',
                                                                                                                  'in_octets': '6524050749',
                                                                                                                  'in_ucast_pkts': '28516334',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465060',
                                                                                                                  'out_octets': '89590013163',
                                                                                                                  'out_ucast_pkts': '69032541',
                                                                                                                  'poller_hash': 'MTE3REI2MHBlZXJpbnRmcG9uIDI1MDA2NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '143690653886',
                                                                                                                  'in_ucast_pkts': '1061493372',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1',
                                                                                                                  'out_octets': '3481876808386',
                                                                                                                  'out_ucast_pkts': '2410879825',
                                                                                                                  'poller_hash': 'MTE3REE3M3BlZXJpbnRmcG9uIDI1MTYxOQ=='},
                                                                                                          {       'in_nucast_pkts': '14966',
                                                                                                                  'in_octets': '9485851040',
                                                                                                                  'in_ucast_pkts': '22626189',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465159',
                                                                                                                  'out_octets': '105645393221',
                                                                                                                  'out_ucast_pkts': '75527400',
                                                                                                                  'poller_hash': 'MTE3REMwQ3BlZXJpbnRmcG9uIDI1MjYwMA=='},
                                                                                                          {       'in_nucast_pkts': '14524',
                                                                                                                  'in_octets': '5118339443',
                                                                                                                  'in_ucast_pkts': '19347928',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465204',
                                                                                                                  'out_octets': '98243891135',
                                                                                                                  'out_ucast_pkts': '76570423',
                                                                                                                  'poller_hash': 'MTE3REIxM3BlZXJpbnRmcG9uIDI1NDA2Mg=='},
                                                                                                          {       'in_nucast_pkts': '12023',
                                                                                                                  'in_octets': '1890947956',
                                                                                                                  'in_ucast_pkts': '12918320',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465244',
                                                                                                                  'out_octets': '66652945507',
                                                                                                                  'out_ucast_pkts': '46663131',
                                                                                                                  'poller_hash': 'MTI5NzRBRnBlZXJpbnRmcG9uIDI1NjcyNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '98559074013',
                                                                                                                  'in_ucast_pkts': '587784435',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2754079757779',
                                                                                                                  'out_ucast_pkts': '1923240785',
                                                                                                                  'poller_hash': 'MTI3NDg5M3BlZXJpbnRmcG9uIDI3MzkwMQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '4/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '34930422431',
                                                                                         'out_octets': '310425650252203',
                                                                                         'out_ucast_pkts': '189446404442',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/3',
                                                                                                                 'id': '4/3',
                                                                                                                 'links': [       {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 2777,
                                                                                                                                          'poller_hash': 'QzgxMUZFcGVlcnBvbnBvbiAxNTY5NTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '15218523',
                                                                                                                                          'fiber_length': 2785,
                                                                                                                                          'poller_hash': 'Q0JEMTlEcGVlcnBvbnBvbiAxNjg5NDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '373561',
                                                                                                                                          'fiber_length': 2808,
                                                                                                                                          'poller_hash': 'QzgzQkI4cGVlcnBvbnBvbiAxNzA3NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 2773,
                                                                                                                                          'poller_hash': 'QzgzMTRBcGVlcnBvbnBvbiAxNzIyNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688362',
                                                                                                                                          'fiber_length': 2801,
                                                                                                                                          'poller_hash': 'QzgzQkVEcGVlcnBvbnBvbiAxNzUyMTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688389',
                                                                                                                                          'fiber_length': 2721,
                                                                                                                                          'poller_hash': 'QzgzMDdFcGVlcnBvbnBvbiAxNzU4NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688364',
                                                                                                                                          'fiber_length': 2578,
                                                                                                                                          'poller_hash': 'REZGREU4cGVlcnBvbnBvbiAxODA3NjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688367',
                                                                                                                                          'fiber_length': 2563,
                                                                                                                                          'poller_hash': 'QzgxREVDcGVlcnBvbnBvbiAyMDUzNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688376',
                                                                                                                                          'fiber_length': 2597,
                                                                                                                                          'poller_hash': 'QzgxRDdDcGVlcnBvbnBvbiAyMDU1MzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688371',
                                                                                                                                          'fiber_length': 2582,
                                                                                                                                          'poller_hash': 'MTE5NTE2OXBlZXJwb25wb24gMjExODA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688378',
                                                                                                                                          'fiber_length': 2720,
                                                                                                                                          'poller_hash': 'MTE5NTFCNnBlZXJwb25wb24gMjE2NzMy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 2543,
                                                                                                                                          'poller_hash': 'MTI3NTE4RnBlZXJwb25wb24gMjE4NDY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688367',
                                                                                                                                          'fiber_length': 2797,
                                                                                                                                          'poller_hash': 'MTI3NEFGNXBlZXJwb25wb24gMjMwMDE5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688358',
                                                                                                                                          'fiber_length': 2555,
                                                                                                                                          'poller_hash': 'MTI3NDlBNXBlZXJwb25wb24gMjM1MTIy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688363',
                                                                                                                                          'fiber_length': 2549,
                                                                                                                                          'poller_hash': 'MTI3NDE5RHBlZXJwb25wb24gMjQzMjM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688361',
                                                                                                                                          'fiber_length': 2752,
                                                                                                                                          'poller_hash': 'MTI3NDA1RnBlZXJwb25wb24gMjQ3MTE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '15498571',
                                                                                                                                          'fiber_length': 2592,
                                                                                                                                          'poller_hash': 'MTI3NDE5RXBlZXJwb25wb24gMjQ3NzU0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688374',
                                                                                                                                          'fiber_length': 2784,
                                                                                                                                          'poller_hash': 'MTE3REI2MHBlZXJwb25wb24gMjUwMDY0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 2573,
                                                                                                                                          'poller_hash': 'MTE3REE3M3BlZXJwb25wb24gMjUxNjE5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 2756,
                                                                                                                                          'poller_hash': 'MTE3REMwQ3BlZXJwb25wb24gMjUyNjAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688368',
                                                                                                                                          'fiber_length': 2610,
                                                                                                                                          'poller_hash': 'MTE3REIxM3BlZXJwb25wb24gMjU0MDYy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 2561,
                                                                                                                                          'poller_hash': 'MTI5NzRBRnBlZXJwb25wb24gMjU2NzI2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102},
                                                                                                                                  {       'connected_time': '688363',
                                                                                                                                          'fiber_length': 2734,
                                                                                                                                          'poller_hash': 'MTI3NDg5M3BlZXJwb25wb24gMjczOTAx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.102}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/3',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '4/4',
                                                                                         'in_discards': '208521',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '11368916',
                                                                                         'in_octets': '38480567411292',
                                                                                         'in_ucast_pkts': '86847203796',
                                                                                         'links': [       {       'in_nucast_pkts': '14880',
                                                                                                                  'in_octets': '4248064986',
                                                                                                                  'in_ucast_pkts': '9326994',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464287',
                                                                                                                  'out_octets': '78166586253',
                                                                                                                  'out_ucast_pkts': '56106253',
                                                                                                                  'poller_hash': 'QzgxRTlFcGVlcmludGZwb24gMTU3NzQz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '165706073113',
                                                                                                                  'in_ucast_pkts': '668258293',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '3953326007530',
                                                                                                                  'out_ucast_pkts': '3112066813',
                                                                                                                  'poller_hash': 'QzgxMjA2cGVlcmludGZwb24gMTU4MDAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '366565236459',
                                                                                                                  'in_ucast_pkts': '32096242129',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '16071240654554',
                                                                                                                  'out_ucast_pkts': '11187686402',
                                                                                                                  'poller_hash': 'Q0JEMTZDcGVlcmludGZwb24gMTY1MTAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '246013802141',
                                                                                                                  'in_ucast_pkts': '770623819',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '47',
                                                                                                                  'out_octets': '3002657171821',
                                                                                                                  'out_ucast_pkts': '2454583627',
                                                                                                                  'poller_hash': 'Q0JEMTZEcGVlcmludGZwb24gMTY3MTc3'},
                                                                                                          {       'in_nucast_pkts': '13426',
                                                                                                                  'in_octets': '9677610564',
                                                                                                                  'in_ucast_pkts': '35643509',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464282',
                                                                                                                  'out_octets': '126398124933',
                                                                                                                  'out_ucast_pkts': '96983254',
                                                                                                                  'poller_hash': 'Q0JFMDJCcGVlcmludGZwb24gMTY4OTcw'},
                                                                                                          {       'in_nucast_pkts': '11973',
                                                                                                                  'in_octets': '4430497054',
                                                                                                                  'in_ucast_pkts': '10505042',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464319',
                                                                                                                  'out_octets': '142284415759',
                                                                                                                  'out_ucast_pkts': '106930872',
                                                                                                                  'poller_hash': 'QzgxOUI4cGVlcmludGZwb24gMTk2MDg1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2581066627733',
                                                                                                                  'in_ucast_pkts': '5676346867',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '17420649061494',
                                                                                                                  'out_ucast_pkts': '12915323556',
                                                                                                                  'poller_hash': 'QzgzMTE2cGVlcmludGZwb24gMTk2MjQ0'},
                                                                                                          {       'in_nucast_pkts': '14897',
                                                                                                                  'in_octets': '7756183021',
                                                                                                                  'in_ucast_pkts': '25714983',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464381',
                                                                                                                  'out_octets': '198411442477',
                                                                                                                  'out_ucast_pkts': '148388707',
                                                                                                                  'poller_hash': 'QzgxRTgycGVlcmludGZwb24gMjAwMTk2'},
                                                                                                          {       'in_nucast_pkts': '12710',
                                                                                                                  'in_octets': '9726280181',
                                                                                                                  'in_ucast_pkts': '53205514',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464375',
                                                                                                                  'out_octets': '202018200868',
                                                                                                                  'out_ucast_pkts': '143471998',
                                                                                                                  'poller_hash': 'QzgxRTdDcGVlcmludGZwb24gMjAyNjg0'},
                                                                                                          {       'in_nucast_pkts': '7501',
                                                                                                                  'in_octets': '10094452733',
                                                                                                                  'in_ucast_pkts': '46248377',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '701966',
                                                                                                                  'out_octets': '293623966425',
                                                                                                                  'out_ucast_pkts': '200498549',
                                                                                                                  'poller_hash': 'QzgxREQxcGVlcmludGZwb24gMjA2MDYw'},
                                                                                                          {       'in_nucast_pkts': '14435',
                                                                                                                  'in_octets': '5332988487',
                                                                                                                  'in_ucast_pkts': '53935805',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464442',
                                                                                                                  'out_octets': '164111846735',
                                                                                                                  'out_ucast_pkts': '113068850',
                                                                                                                  'poller_hash': 'QzgxRERBcGVlcmludGZwb24gMjA5MTc0'},
                                                                                                          {       'in_nucast_pkts': '13656',
                                                                                                                  'in_octets': '12328811283',
                                                                                                                  'in_ucast_pkts': '38714282',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464342',
                                                                                                                  'out_octets': '214429032528',
                                                                                                                  'out_ucast_pkts': '156795547',
                                                                                                                  'poller_hash': 'QzgxQzhGcGVlcmludGZwb24gMjExNjIz'},
                                                                                                          {       'in_nucast_pkts': '14783',
                                                                                                                  'in_octets': '182813237532',
                                                                                                                  'in_ucast_pkts': '173591980',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463272',
                                                                                                                  'out_octets': '178903361866',
                                                                                                                  'out_ucast_pkts': '193076187',
                                                                                                                  'poller_hash': 'QzgxRDg3cGVlcmludGZwb24gMjExNzc1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '228518053151',
                                                                                                                  'in_ucast_pkts': '1171483603',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '47',
                                                                                                                  'out_octets': '3525560860788',
                                                                                                                  'out_ucast_pkts': '2662007248',
                                                                                                                  'poller_hash': 'MTE5NTExMXBlZXJpbnRmcG9uIDIxNzAxNg=='},
                                                                                                          {       'in_nucast_pkts': '11963',
                                                                                                                  'in_octets': '1585080934',
                                                                                                                  'in_ucast_pkts': '6250030',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464439',
                                                                                                                  'out_octets': '34913218510',
                                                                                                                  'out_ucast_pkts': '25571942',
                                                                                                                  'poller_hash': 'MTI3NTA4MnBlZXJpbnRmcG9uIDIxODIyNw=='},
                                                                                                          {       'in_nucast_pkts': '903366',
                                                                                                                  'in_octets': '847867046049',
                                                                                                                  'in_ucast_pkts': '4626636566',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '125470665',
                                                                                                                  'out_octets': '16677848189884',
                                                                                                                  'out_ucast_pkts': '11814447761',
                                                                                                                  'poller_hash': 'MTI3NDA5MHBlZXJpbnRmcG9uIDIzMTAxOA=='},
                                                                                                          {       'in_nucast_pkts': '14588',
                                                                                                                  'in_octets': '21117233674',
                                                                                                                  'in_ucast_pkts': '68319815',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464731',
                                                                                                                  'out_octets': '160563655881',
                                                                                                                  'out_ucast_pkts': '133673872',
                                                                                                                  'poller_hash': 'MTI3NDk5N3BlZXJpbnRmcG9uIDIzMjMwOQ=='},
                                                                                                          {       'in_nucast_pkts': '12019',
                                                                                                                  'in_octets': '4046134495',
                                                                                                                  'in_ucast_pkts': '17186622',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465086',
                                                                                                                  'out_octets': '65723722795',
                                                                                                                  'out_ucast_pkts': '52617894',
                                                                                                                  'poller_hash': 'MTI5NkZENHBlZXJpbnRmcG9uIDI0MjE4NQ=='},
                                                                                                          {       'in_nucast_pkts': '12425',
                                                                                                                  'in_octets': '10688741581',
                                                                                                                  'in_ucast_pkts': '25277908',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465153',
                                                                                                                  'out_octets': '134931551825',
                                                                                                                  'out_ucast_pkts': '99815627',
                                                                                                                  'poller_hash': 'MTE3REI5MnBlZXJpbnRmcG9uIDI1MTE2Mw=='},
                                                                                                          {       'in_nucast_pkts': '12625',
                                                                                                                  'in_octets': '53159130509',
                                                                                                                  'in_ucast_pkts': '182844077',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465170',
                                                                                                                  'out_octets': '738229245613',
                                                                                                                  'out_ucast_pkts': '541294983',
                                                                                                                  'poller_hash': 'MTE3REJEQnBlZXJpbnRmcG9uIDI1MTUzOQ=='},
                                                                                                          {       'in_nucast_pkts': '12010',
                                                                                                                  'in_octets': '14075602334',
                                                                                                                  'in_ucast_pkts': '33480663',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465217',
                                                                                                                  'out_octets': '154977066180',
                                                                                                                  'out_ucast_pkts': '119402844',
                                                                                                                  'poller_hash': 'MTE3REJGMXBlZXJpbnRmcG9uIDI1Mjc1MA=='},
                                                                                                          {       'in_nucast_pkts': '14388',
                                                                                                                  'in_octets': '14434489786',
                                                                                                                  'in_ucast_pkts': '91303309',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465060',
                                                                                                                  'out_octets': '829966742160',
                                                                                                                  'out_ucast_pkts': '565926629',
                                                                                                                  'poller_hash': 'MTI5NUU5N3BlZXJpbnRmcG9uIDI3NzQzNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '111558773650',
                                                                                                                  'in_ucast_pkts': '163495298',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '634962669695',
                                                                                                                  'out_ucast_pkts': '480373355',
                                                                                                                  'poller_hash': 'MTI5NjEzQnBlZXJpbnRmcG9uIDI4MjAzNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NjAwRXBlZXJpbnRmcG9uIDI4MjA4MA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '4/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '125941089',
                                                                                         'out_octets': '329803191061483',
                                                                                         'out_ucast_pkts': '246660077676',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/4',
                                                                                                                 'id': '4/4',
                                                                                                                 'links': [       {       'connected_time': '688365',
                                                                                                                                          'fiber_length': 2845,
                                                                                                                                          'poller_hash': 'QzgxRTlFcGVlcnBvbnBvbiAxNTc3NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688363',
                                                                                                                                          'fiber_length': 2865,
                                                                                                                                          'poller_hash': 'QzgxMjA2cGVlcnBvbnBvbiAxNTgwMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 2840,
                                                                                                                                          'poller_hash': 'Q0JEMTZDcGVlcnBvbnBvbiAxNjUxMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688364',
                                                                                                                                          'fiber_length': 2823,
                                                                                                                                          'poller_hash': 'Q0JEMTZEcGVlcnBvbnBvbiAxNjcxNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688375',
                                                                                                                                          'fiber_length': 2828,
                                                                                                                                          'poller_hash': 'Q0JFMDJCcGVlcnBvbnBvbiAxNjg5NzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688361',
                                                                                                                                          'fiber_length': 3088,
                                                                                                                                          'poller_hash': 'QzgxOUI4cGVlcnBvbnBvbiAxOTYwODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688360',
                                                                                                                                          'fiber_length': 2839,
                                                                                                                                          'poller_hash': 'QzgzMTE2cGVlcnBvbnBvbiAxOTYyNDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688379',
                                                                                                                                          'fiber_length': 3117,
                                                                                                                                          'poller_hash': 'QzgxRTgycGVlcnBvbnBvbiAyMDAxOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 2930,
                                                                                                                                          'poller_hash': 'QzgxRTdDcGVlcnBvbnBvbiAyMDI2ODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '340396',
                                                                                                                                          'fiber_length': 2882,
                                                                                                                                          'poller_hash': 'QzgxREQxcGVlcnBvbnBvbiAyMDYwNjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688367',
                                                                                                                                          'fiber_length': 2899,
                                                                                                                                          'poller_hash': 'QzgxRERBcGVlcnBvbnBvbiAyMDkxNzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 3109,
                                                                                                                                          'poller_hash': 'QzgxQzhGcGVlcnBvbnBvbiAyMTE2MjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688364',
                                                                                                                                          'fiber_length': 3098,
                                                                                                                                          'poller_hash': 'QzgxRDg3cGVlcnBvbnBvbiAyMTE3NzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688364',
                                                                                                                                          'fiber_length': 3351,
                                                                                                                                          'poller_hash': 'MTE5NTExMXBlZXJwb25wb24gMjE3MDE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 3044,
                                                                                                                                          'poller_hash': 'MTI3NTA4MnBlZXJwb25wb24gMjE4MjI3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '8458080',
                                                                                                                                          'fiber_length': 3328,
                                                                                                                                          'poller_hash': 'MTI3NDA5MHBlZXJwb25wb24gMjMxMDE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688364',
                                                                                                                                          'fiber_length': 3081,
                                                                                                                                          'poller_hash': 'MTI3NDk5N3BlZXJwb25wb24gMjMyMzA5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688364',
                                                                                                                                          'fiber_length': 3350,
                                                                                                                                          'poller_hash': 'MTI5NkZENHBlZXJwb25wb24gMjQyMTg1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 3318,
                                                                                                                                          'poller_hash': 'MTE3REI5MnBlZXJwb25wb24gMjUxMTYz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688376',
                                                                                                                                          'fiber_length': 2857,
                                                                                                                                          'poller_hash': 'MTE3REJEQnBlZXJwb25wb24gMjUxNTM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 3309,
                                                                                                                                          'poller_hash': 'MTE3REJGMXBlZXJwb25wb24gMjUyNzUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688366',
                                                                                                                                          'fiber_length': 3062,
                                                                                                                                          'poller_hash': 'MTI5NUU5N3BlZXJwb25wb24gMjc3NDM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 3334,
                                                                                                                                          'poller_hash': 'MTI5NjEzQnBlZXJwb25wb24gMjgyMDM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092},
                                                                                                                                  {       'connected_time': '688374',
                                                                                                                                          'fiber_length': 3074,
                                                                                                                                          'poller_hash': 'MTI5NjAwRXBlZXJwb25wb24gMjgyMDgw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.092}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/4',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '4/5',
                                                                                         'in_discards': '7568',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2253488',
                                                                                         'in_octets': '4121549153247',
                                                                                         'in_ucast_pkts': '10247674407',
                                                                                         'links': [       {       'in_nucast_pkts': '15022',
                                                                                                                  'in_octets': '214672509125',
                                                                                                                  'in_ucast_pkts': '310021559',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465249',
                                                                                                                  'out_octets': '571727594377',
                                                                                                                  'out_ucast_pkts': '453592433',
                                                                                                                  'poller_hash': 'MTBBODYzNXBlZXJpbnRmcG9uIDI0OTYyMg=='},
                                                                                                          {       'in_nucast_pkts': '11789',
                                                                                                                  'in_octets': '5915718053',
                                                                                                                  'in_ucast_pkts': '35114400',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465199',
                                                                                                                  'out_octets': '115650603427',
                                                                                                                  'out_ucast_pkts': '86134650',
                                                                                                                  'poller_hash': 'MTE3REI5OHBlZXJpbnRmcG9uIDI1Mjg1NQ=='},
                                                                                                          {       'in_nucast_pkts': '5899',
                                                                                                                  'in_octets': '30736083744',
                                                                                                                  'in_ucast_pkts': '39212158',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '647464',
                                                                                                                  'out_octets': '121256067482',
                                                                                                                  'out_ucast_pkts': '93443142',
                                                                                                                  'poller_hash': 'MTE3RDk5M3BlZXJpbnRmcG9uIDI1NDU0MQ=='},
                                                                                                          {       'in_nucast_pkts': '502469',
                                                                                                                  'in_octets': '181157846893',
                                                                                                                  'in_ucast_pkts': '1056279218',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '64693803',
                                                                                                                  'out_octets': '4018052321968',
                                                                                                                  'out_ucast_pkts': '2952573173',
                                                                                                                  'poller_hash': 'MTI5NzM4OXBlZXJpbnRmcG9uIDI1NTcwMg=='},
                                                                                                          {       'in_nucast_pkts': '11979',
                                                                                                                  'in_octets': '7334671967',
                                                                                                                  'in_ucast_pkts': '19306036',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465241',
                                                                                                                  'out_octets': '85041597432',
                                                                                                                  'out_ucast_pkts': '70415645',
                                                                                                                  'poller_hash': 'MTI3NDg2MnBlZXJpbnRmcG9uIDI3NDM5Ng=='},
                                                                                                          {       'in_nucast_pkts': '958',
                                                                                                                  'in_octets': '24619132911',
                                                                                                                  'in_ucast_pkts': '76842851',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465249',
                                                                                                                  'out_octets': '250852823793',
                                                                                                                  'out_ucast_pkts': '199799491',
                                                                                                                  'poller_hash': 'MTI3NDc0NHBlZXJpbnRmcG9uIDI3NDc0Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '25826325862',
                                                                                                                  'in_ucast_pkts': '97153165',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '297251118819',
                                                                                                                  'out_ucast_pkts': '247648327',
                                                                                                                  'poller_hash': 'MTI5NUUzN3BlZXJpbnRmcG9uIDI4MjA3MA=='},
                                                                                                          {       'in_nucast_pkts': '7349',
                                                                                                                  'in_octets': '14918303722',
                                                                                                                  'in_ucast_pkts': '32145046',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '723590',
                                                                                                                  'out_octets': '89931862425',
                                                                                                                  'out_ucast_pkts': '74636195',
                                                                                                                  'poller_hash': 'MTI5NUU5NXBlZXJpbnRmcG9uIDI4NjI4NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '4/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '88502250',
                                                                                         'out_octets': '37444731434653',
                                                                                         'out_ucast_pkts': '28342256772',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/5',
                                                                                                                 'id': '4/5',
                                                                                                                 'links': [       {       'connected_time': '688372',
                                                                                                                                          'fiber_length': 2949,
                                                                                                                                          'poller_hash': 'MTBBODYzNXBlZXJwb25wb24gMjQ5NjIy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824},
                                                                                                                                  {       'connected_time': '688375',
                                                                                                                                          'fiber_length': 3184,
                                                                                                                                          'poller_hash': 'MTE3REI5OHBlZXJwb25wb24gMjUyODU1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824},
                                                                                                                                  {       'connected_time': '688375',
                                                                                                                                          'fiber_length': 3236,
                                                                                                                                          'poller_hash': 'MTE3RDk5M3BlZXJwb25wb24gMjU0NTQx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824},
                                                                                                                                  {       'connected_time': '26960047',
                                                                                                                                          'fiber_length': 3100,
                                                                                                                                          'poller_hash': 'MTI5NzM4OXBlZXJwb25wb24gMjU1NzAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 3168,
                                                                                                                                          'poller_hash': 'MTI3NDg2MnBlZXJwb25wb24gMjc0Mzk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 3182,
                                                                                                                                          'poller_hash': 'MTI3NDc0NHBlZXJwb25wb24gMjc0NzQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824},
                                                                                                                                  {       'connected_time': '688362',
                                                                                                                                          'fiber_length': 3096,
                                                                                                                                          'poller_hash': 'MTI5NUUzN3BlZXJwb25wb24gMjgyMDcw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824},
                                                                                                                                  {       'connected_time': '351587',
                                                                                                                                          'fiber_length': 3064,
                                                                                                                                          'poller_hash': 'MTI5NUU5NXBlZXJwb25wb24gMjg2Mjg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.824}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/5',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '4/6',
                                                                                         'in_discards': '680',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '162896',
                                                                                         'in_octets': '3552794077794',
                                                                                         'in_ucast_pkts': '11533743957',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '486259273444',
                                                                                                                  'in_ucast_pkts': '130161343401',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '3271780468616',
                                                                                                                  'out_ucast_pkts': '2673734730',
                                                                                                                  'poller_hash': 'REZGQjQwcGVlcmludGZwb24gMTg0NzE1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '123510130289',
                                                                                                                  'in_ucast_pkts': '197162462',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '343160165537',
                                                                                                                  'out_ucast_pkts': '361522999',
                                                                                                                  'poller_hash': 'QzgzQkI0cGVlcmludGZwb24gMTg0Nzg1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '458965789610',
                                                                                                                  'in_ucast_pkts': '26998100584',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '3136034658086',
                                                                                                                  'out_ucast_pkts': '2824624561',
                                                                                                                  'poller_hash': 'REZGRDkwcGVlcmludGZwb24gMTg1NzEx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '252664499640',
                                                                                                                  'in_ucast_pkts': '69637711361',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '147',
                                                                                                                  'out_octets': '4297810388964',
                                                                                                                  'out_ucast_pkts': '3258849809',
                                                                                                                  'poller_hash': 'QzgxRUM5cGVlcmludGZwb24gMTk4MjU0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '7956582242',
                                                                                                                  'in_ucast_pkts': '34454319',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '120',
                                                                                                                  'out_octets': '95739241939',
                                                                                                                  'out_ucast_pkts': '123491827',
                                                                                                                  'poller_hash': 'QzgxRjAycGVlcmludGZwb24gMjAwNDQw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '36883163884',
                                                                                                                  'in_ucast_pkts': '155391641',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '584212465730',
                                                                                                                  'out_ucast_pkts': '503315994',
                                                                                                                  'poller_hash': 'QzgxQzkxcGVlcmludGZwb24gMjA4NjY4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '118608755568',
                                                                                                                  'in_ucast_pkts': '26089801693',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '147',
                                                                                                                  'out_octets': '1177131489725',
                                                                                                                  'out_ucast_pkts': '990390808',
                                                                                                                  'poller_hash': 'MTFCMjVGOHBlZXJpbnRmcG9uIDIwODk0Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '183258536936',
                                                                                                                  'in_ucast_pkts': '945791834',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '3929788040697',
                                                                                                                  'out_ucast_pkts': '2842901744',
                                                                                                                  'poller_hash': 'MTFCMjYxMnBlZXJpbnRmcG9uIDIwOTAyNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '191991797699',
                                                                                                                  'in_ucast_pkts': '964950917',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '120',
                                                                                                                  'out_octets': '4168810036943',
                                                                                                                  'out_ucast_pkts': '3153060804',
                                                                                                                  'poller_hash': 'MTE5NTE3MnBlZXJpbnRmcG9uIDIxMjUyMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE5NTEwNXBlZXJpbnRmcG9uIDIxMzkzNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '48905771558',
                                                                                                                  'in_ucast_pkts': '142565640',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '704846907500',
                                                                                                                  'out_ucast_pkts': '600982353',
                                                                                                                  'poller_hash': 'MTI3NEExRnBlZXJpbnRmcG9uIDIzNDM0Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '15510286359',
                                                                                                                  'in_ucast_pkts': '35786330',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '120409960686',
                                                                                                                  'out_ucast_pkts': '116169530',
                                                                                                                  'poller_hash': 'MTI3NDgyOXBlZXJpbnRmcG9uIDI3NDM4Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '14758921129',
                                                                                                                  'in_ucast_pkts': '29363363',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '51767495671',
                                                                                                                  'out_ucast_pkts': '54994029',
                                                                                                                  'poller_hash': 'MTI3MEQ5MnBlZXJpbnRmcG9uIDI3ODE5NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1751979952',
                                                                                                                  'in_ucast_pkts': '7627007',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '19026514314',
                                                                                                                  'out_ucast_pkts': '19149689',
                                                                                                                  'poller_hash': 'MTI5NjExMHBlZXJpbnRmcG9uIDI4MjA0Nw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '4/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '37529015221',
                                                                                         'out_octets': '89991535955136',
                                                                                         'out_ucast_pkts': '31144176158',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/6',
                                                                                                                 'id': '4/6',
                                                                                                                 'links': [       {       'connected_time': '523096',
                                                                                                                                          'fiber_length': 11102,
                                                                                                                                          'poller_hash': 'REZGQjQwcGVlcnBvbnBvbiAxODQ3MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '21815790',
                                                                                                                                          'fiber_length': 11112,
                                                                                                                                          'poller_hash': 'QzgzQkI0cGVlcnBvbnBvbiAxODQ3ODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '9398572',
                                                                                                                                          'fiber_length': 11196,
                                                                                                                                          'poller_hash': 'REZGRDkwcGVlcnBvbnBvbiAxODU3MTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '21816523',
                                                                                                                                          'fiber_length': 11407,
                                                                                                                                          'poller_hash': 'QzgxRUM5cGVlcnBvbnBvbiAxOTgyNTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '1130993',
                                                                                                                                          'fiber_length': 11683,
                                                                                                                                          'poller_hash': 'QzgxRjAycGVlcnBvbnBvbiAyMDA0NDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '6576700',
                                                                                                                                          'fiber_length': 11779,
                                                                                                                                          'poller_hash': 'QzgxQzkxcGVlcnBvbnBvbiAyMDg2Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '72951357',
                                                                                                                                          'fiber_length': 11763,
                                                                                                                                          'poller_hash': 'MTFCMjVGOHBlZXJwb25wb24gMjA4OTQz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '688377',
                                                                                                                                          'fiber_length': 11816,
                                                                                                                                          'poller_hash': 'MTFCMjYxMnBlZXJwb25wb24gMjA5MDI2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '22282804',
                                                                                                                                          'fiber_length': 11756,
                                                                                                                                          'poller_hash': 'MTE5NTE3MnBlZXJwb25wb24gMjEyNTIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '20916304',
                                                                                                                                          'fiber_length': 11791,
                                                                                                                                          'poller_hash': 'MTE5NTEwNXBlZXJwb25wb24gMjEzOTM0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '4055332',
                                                                                                                                          'fiber_length': 11790,
                                                                                                                                          'poller_hash': 'MTI3NEExRnBlZXJwb25wb24gMjM0MzQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '13405471',
                                                                                                                                          'fiber_length': 11794,
                                                                                                                                          'poller_hash': 'MTI3NDgyOXBlZXJwb25wb24gMjc0Mzg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '8387845',
                                                                                                                                          'fiber_length': 11732,
                                                                                                                                          'poller_hash': 'MTI3MEQ5MnBlZXJwb25wb24gMjc4MTk1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087},
                                                                                                                                  {       'connected_time': '3564005',
                                                                                                                                          'fiber_length': 12013,
                                                                                                                                          'poller_hash': 'MTI5NjExMHBlZXJwb25wb24gMjgyMDQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.087}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/6',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '4/7',
                                                                                         'in_discards': '6394',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2874195',
                                                                                         'in_octets': '8664546816150',
                                                                                         'in_ucast_pkts': '33517929334',
                                                                                         'links': [       {       'in_nucast_pkts': '444104',
                                                                                                                  'in_octets': '142814072481',
                                                                                                                  'in_ucast_pkts': '499266109',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '60233653',
                                                                                                                  'out_octets': '3499755642020',
                                                                                                                  'out_ucast_pkts': '2733245579',
                                                                                                                  'poller_hash': 'REZGQjhCcGVlcmludGZwb24gMTgyMjky'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '503125359973',
                                                                                                                  'in_ucast_pkts': '31683836879',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123',
                                                                                                                  'out_octets': '4458013344774',
                                                                                                                  'out_ucast_pkts': '3488774014',
                                                                                                                  'poller_hash': 'QzgzQjg4cGVlcmludGZwb24gMTg0NzQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '92675074820',
                                                                                                                  'in_ucast_pkts': '229543532',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '147',
                                                                                                                  'out_octets': '677396310975',
                                                                                                                  'out_ucast_pkts': '602545077',
                                                                                                                  'poller_hash': 'QzgzQkJDcGVlcmludGZwb24gMTg0NzU1'},
                                                                                                          {       'in_nucast_pkts': '262925',
                                                                                                                  'in_octets': '90496395418',
                                                                                                                  'in_ucast_pkts': '1163163512',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '30390529',
                                                                                                                  'out_octets': '3504477759681',
                                                                                                                  'out_ucast_pkts': '2355516950',
                                                                                                                  'poller_hash': 'QzgzQjhGcGVlcmludGZwb24gMTg0NzY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '179998021223',
                                                                                                                  'in_ucast_pkts': '1110679394',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '7554978269371',
                                                                                                                  'out_ucast_pkts': '5646401367',
                                                                                                                  'poller_hash': 'QzgzNzA4cGVlcmludGZwb24gMTg0Nzc1'},
                                                                                                          {       'in_nucast_pkts': '513285',
                                                                                                                  'in_octets': '104516592021',
                                                                                                                  'in_ucast_pkts': '930580447',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '67509329',
                                                                                                                  'out_octets': '5607462190608',
                                                                                                                  'out_ucast_pkts': '3829093714',
                                                                                                                  'poller_hash': 'QzgzQjhCcGVlcmludGZwb24gMTg1MTcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '8531467426',
                                                                                                                  'in_ucast_pkts': '35041838',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '146',
                                                                                                                  'out_octets': '118875295991',
                                                                                                                  'out_ucast_pkts': '137745264',
                                                                                                                  'poller_hash': 'QzgzMDg4cGVlcmludGZwb24gMTg1MTgx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'REZGQzAycGVlcmludGZwb24gMTg1NDA4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '159373492039',
                                                                                                                  'in_ucast_pkts': '490358623',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '2362233602308',
                                                                                                                  'out_ucast_pkts': '1831573515',
                                                                                                                  'poller_hash': 'REZGRTRBcGVlcmludGZwb24gMTg1NDE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '24568458634',
                                                                                                                  'in_ucast_pkts': '4394564576',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '120',
                                                                                                                  'out_octets': '689099899600',
                                                                                                                  'out_ucast_pkts': '559536211',
                                                                                                                  'poller_hash': 'REZGRTMxcGVlcmludGZwb24gMTg1NDI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxREQ1cGVlcmludGZwb24gMTg1NDM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '211003831310',
                                                                                                                  'in_ucast_pkts': '2351554557',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '14147598260495',
                                                                                                                  'out_ucast_pkts': '9460227816',
                                                                                                                  'poller_hash': 'QzgxRDYycGVlcmludGZwb24gMTg1NzIx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1586174610',
                                                                                                                  'in_ucast_pkts': '6765755',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '147',
                                                                                                                  'out_octets': '23210325710',
                                                                                                                  'out_ucast_pkts': '59501367',
                                                                                                                  'poller_hash': 'QzgzMTY5cGVlcmludGZwb24gMTg2NDc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '668448129871',
                                                                                                                  'in_ucast_pkts': '61752969470',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70',
                                                                                                                  'out_octets': '4358435713162',
                                                                                                                  'out_ucast_pkts': '3666681058',
                                                                                                                  'poller_hash': 'QzgzMTE0cGVlcmludGZwb24gMTg2NDg3'},
                                                                                                          {       'in_nucast_pkts': '6509970',
                                                                                                                  'in_octets': '334390036668',
                                                                                                                  'in_ucast_pkts': '2007918184',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8551117092',
                                                                                                                  'out_octets': '18670267170392',
                                                                                                                  'out_ucast_pkts': '5226717142',
                                                                                                                  'poller_hash': 'QzgzQkM4cGVlcmludGZwb24gMTg5ODA3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '332008495016',
                                                                                                                  'in_ucast_pkts': '628618286',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '1798814245244',
                                                                                                                  'out_ucast_pkts': '1639611895',
                                                                                                                  'poller_hash': 'QzgzMjFCcGVlcmludGZwb24gMTk4MjQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxRTIzcGVlcmludGZwb24gMTk4MjY0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '93759621423',
                                                                                                                  'in_ucast_pkts': '8998288164',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '1700146081646',
                                                                                                                  'out_ucast_pkts': '1289461895',
                                                                                                                  'poller_hash': 'QzgxRUMzcGVlcmludGZwb24gMjAwMTc2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '873312804834',
                                                                                                                  'in_ucast_pkts': '61385302694',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '3057680426785',
                                                                                                                  'out_ucast_pkts': '2622263255',
                                                                                                                  'poller_hash': 'QzgxRjE3cGVlcmludGZwb24gMjAwMTg2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '383854024099',
                                                                                                                  'in_ucast_pkts': '2248804416',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '9917490085363',
                                                                                                                  'out_ucast_pkts': '7039437168',
                                                                                                                  'poller_hash': 'QzgxRTVFcGVlcmludGZwb24gMjAwODU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxNjA5cGVlcmludGZwb24gMjAxMjg1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '85208868407',
                                                                                                                  'in_ucast_pkts': '554079434',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '117',
                                                                                                                  'out_octets': '4193121677740',
                                                                                                                  'out_ucast_pkts': '3066135511',
                                                                                                                  'poller_hash': 'QzgxRTYwcGVlcmludGZwb24gMjAxMzE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '417406299765',
                                                                                                                  'in_ucast_pkts': '951380939',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '4157617099043',
                                                                                                                  'out_ucast_pkts': '3004911992',
                                                                                                                  'poller_hash': 'QzgxREM1cGVlcmludGZwb24gMjA4NjU4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxREY0cGVlcmludGZwb24gMjEzNzgw'},
                                                                                                          {       'in_nucast_pkts': '12922',
                                                                                                                  'in_octets': '3508305069',
                                                                                                                  'in_ucast_pkts': '9929327',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464498',
                                                                                                                  'out_octets': '19793819238',
                                                                                                                  'out_ucast_pkts': '18225532',
                                                                                                                  'poller_hash': 'MTI3NEIzQnBlZXJpbnRmcG9uIDIzMjExNA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '4/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '37361340576',
                                                                                         'out_octets': '187066034369564',
                                                                                         'out_ucast_pkts': '100339774569',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/7',
                                                                                                                 'id': '4/7',
                                                                                                                 'links': [       {       'connected_time': '25305799',
                                                                                                                                          'fiber_length': 12067,
                                                                                                                                          'poller_hash': 'REZGQjhCcGVlcnBvbnBvbiAxODIyOTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '21816369',
                                                                                                                                          'fiber_length': 11986,
                                                                                                                                          'poller_hash': 'QzgzQjg4cGVlcnBvbnBvbiAxODQ3NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '21816264',
                                                                                                                                          'fiber_length': 11595,
                                                                                                                                          'poller_hash': 'QzgzQkJDcGVlcnBvbnBvbiAxODQ3NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '15498586',
                                                                                                                                          'fiber_length': 12020,
                                                                                                                                          'poller_hash': 'QzgzQjhGcGVlcnBvbnBvbiAxODQ3NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688362',
                                                                                                                                          'fiber_length': 11858,
                                                                                                                                          'poller_hash': 'QzgzNzA4cGVlcnBvbnBvbiAxODQ3NzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '28453801',
                                                                                                                                          'fiber_length': 12118,
                                                                                                                                          'poller_hash': 'QzgzQjhCcGVlcnBvbnBvbiAxODUxNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '2041884',
                                                                                                                                          'fiber_length': 11982,
                                                                                                                                          'poller_hash': 'QzgzMDg4cGVlcnBvbnBvbiAxODUxODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '21815953',
                                                                                                                                          'fiber_length': 12003,
                                                                                                                                          'poller_hash': 'REZGQzAycGVlcnBvbnBvbiAxODU0MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '523102',
                                                                                                                                          'fiber_length': 11587,
                                                                                                                                          'poller_hash': 'REZGRTRBcGVlcnBvbnBvbiAxODU0MTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '22882706',
                                                                                                                                          'fiber_length': 12024,
                                                                                                                                          'poller_hash': 'REZGRTMxcGVlcnBvbnBvbiAxODU0Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '6029064',
                                                                                                                                          'fiber_length': 12028,
                                                                                                                                          'poller_hash': 'QzgxREQ1cGVlcnBvbnBvbiAxODU0Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '69480697',
                                                                                                                                          'fiber_length': 12027,
                                                                                                                                          'poller_hash': 'QzgxRDYycGVlcnBvbnBvbiAxODU3MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '2139849',
                                                                                                                                          'fiber_length': 12256,
                                                                                                                                          'poller_hash': 'QzgzMTY5cGVlcnBvbnBvbiAxODY0Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688368',
                                                                                                                                          'fiber_length': 12284,
                                                                                                                                          'poller_hash': 'QzgzMTE0cGVlcnBvbnBvbiAxODY0ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '69798267',
                                                                                                                                          'fiber_length': 12028,
                                                                                                                                          'poller_hash': 'QzgzQkM4cGVlcnBvbnBvbiAxODk4MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '14513940',
                                                                                                                                          'fiber_length': 11426,
                                                                                                                                          'poller_hash': 'QzgzMjFCcGVlcnBvbnBvbiAxOTgyNDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '21816305',
                                                                                                                                          'fiber_length': 11387,
                                                                                                                                          'poller_hash': 'QzgxRTIzcGVlcnBvbnBvbiAxOTgyNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688377',
                                                                                                                                          'fiber_length': 11424,
                                                                                                                                          'poller_hash': 'QzgxRUMzcGVlcnBvbnBvbiAyMDAxNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688370',
                                                                                                                                          'fiber_length': 11378,
                                                                                                                                          'poller_hash': 'QzgxRjE3cGVlcnBvbnBvbiAyMDAxODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688379',
                                                                                                                                          'fiber_length': 11923,
                                                                                                                                          'poller_hash': 'QzgxRTVFcGVlcnBvbnBvbiAyMDA4NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '19095402',
                                                                                                                                          'fiber_length': 11526,
                                                                                                                                          'poller_hash': 'QzgxNjA5cGVlcnBvbnBvbiAyMDEyODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '10654124',
                                                                                                                                          'fiber_length': 11453,
                                                                                                                                          'poller_hash': 'QzgxRTYwcGVlcnBvbnBvbiAyMDEzMTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 12065,
                                                                                                                                          'poller_hash': 'QzgxREM1cGVlcnBvbnBvbiAyMDg2NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '20994506',
                                                                                                                                          'fiber_length': 11969,
                                                                                                                                          'poller_hash': 'QzgxREY0cGVlcnBvbnBvbiAyMTM3ODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 11269,
                                                                                                                                          'poller_hash': 'MTI3NEIzQnBlZXJwb25wb24gMjMyMTE0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/7',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '4/8',
                                                                                         'in_discards': '8199',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5099242',
                                                                                         'in_octets': '7065702052336',
                                                                                         'in_ucast_pkts': '28935514453',
                                                                                         'links': [       {       'in_nucast_pkts': '11930',
                                                                                                                  'in_octets': '1177823599',
                                                                                                                  'in_ucast_pkts': '3807716',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464152',
                                                                                                                  'out_octets': '28392467417',
                                                                                                                  'out_ucast_pkts': '22541139',
                                                                                                                  'poller_hash': 'MkIwMjVEcGVlcmludGZwb24gMjY4'},
                                                                                                          {       'in_nucast_pkts': '13017',
                                                                                                                  'in_octets': '65333049689',
                                                                                                                  'in_ucast_pkts': '316905604',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27344767',
                                                                                                                  'out_octets': '364581333822',
                                                                                                                  'out_ucast_pkts': '439917403',
                                                                                                                  'poller_hash': 'MjU5ODM3cGVlcmludGZwb24gMjky'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1301916765984',
                                                                                                                  'in_ucast_pkts': '6395314093',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '100',
                                                                                                                  'out_octets': '4048851407208',
                                                                                                                  'out_ucast_pkts': '3571713323',
                                                                                                                  'poller_hash': 'MkIwMjA2cGVlcmludGZwb24gMjkz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96840905530',
                                                                                                                  'in_ucast_pkts': '1143435300',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '4639166987305',
                                                                                                                  'out_ucast_pkts': '3167564379',
                                                                                                                  'poller_hash': 'MkIwMjcycGVlcmludGZwb24gMjk0'},
                                                                                                          {       'in_nucast_pkts': '523834',
                                                                                                                  'in_octets': '223822807392',
                                                                                                                  'in_ucast_pkts': '410406249',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '68938959',
                                                                                                                  'out_octets': '1098796975886',
                                                                                                                  'out_ucast_pkts': '958152557',
                                                                                                                  'poller_hash': 'Mjk4M0VEcGVlcmludGZwb24gMjk2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '113482170920',
                                                                                                                  'in_ucast_pkts': '1147660593',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66',
                                                                                                                  'out_octets': '6991124881989',
                                                                                                                  'out_ucast_pkts': '4676447276',
                                                                                                                  'poller_hash': 'MjA1REZDcGVlcmludGZwb24gMjk3'},
                                                                                                          {       'in_nucast_pkts': '20144086',
                                                                                                                  'in_octets': '798217996037',
                                                                                                                  'in_ucast_pkts': '3299301029',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45795903079',
                                                                                                                  'out_octets': '76373260104837',
                                                                                                                  'out_ucast_pkts': '12042743311',
                                                                                                                  'poller_hash': 'Mjk4NDVBcGVlcmludGZwb24gMjk4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '340434104675',
                                                                                                                  'in_ucast_pkts': '53739258918',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '147',
                                                                                                                  'out_octets': '6922817845843',
                                                                                                                  'out_ucast_pkts': '5128182132',
                                                                                                                  'poller_hash': 'Mjk4MjRGcGVlcmludGZwb24gMjk5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '211580395865',
                                                                                                                  'in_ucast_pkts': '5087828380',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '3358233147162',
                                                                                                                  'out_ucast_pkts': '2714752917',
                                                                                                                  'poller_hash': 'MjA1RENGcGVlcmludGZwb24gMzAy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '144324544593',
                                                                                                                  'in_ucast_pkts': '1092180976',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '4278007633853',
                                                                                                                  'out_ucast_pkts': '3043065068',
                                                                                                                  'poller_hash': 'Mjk4Mzg1cGVlcmludGZwb24gMzA3'},
                                                                                                          {       'in_nucast_pkts': '11670',
                                                                                                                  'in_octets': '7912780094',
                                                                                                                  'in_ucast_pkts': '41275628',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464172',
                                                                                                                  'out_octets': '153163083184',
                                                                                                                  'out_ucast_pkts': '112900608',
                                                                                                                  'poller_hash': 'NUE2NkJBcGVlcmludGZwb24gNDIz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '752047738474',
                                                                                                                  'in_ucast_pkts': '260199755',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '150',
                                                                                                                  'out_octets': '1970273090760',
                                                                                                                  'out_ucast_pkts': '1930719927',
                                                                                                                  'poller_hash': 'MjA1REVFcGVlcmludGZwb24gNDQ3'},
                                                                                                          {       'in_nucast_pkts': '14466',
                                                                                                                  'in_octets': '2249402765',
                                                                                                                  'in_ucast_pkts': '22153661',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464215',
                                                                                                                  'out_octets': '185160171225',
                                                                                                                  'out_ucast_pkts': '124934828',
                                                                                                                  'poller_hash': 'MkIwMTgxcGVlcmludGZwb24gNTAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODQ0QkQ2cGVlcmludGZwb24gMTQwNTI3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgzQkU2cGVlcmludGZwb24gMTcyMzY5'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '4/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '16710455250',
                                                                                         'out_octets': '130885485579631',
                                                                                         'out_ucast_pkts': '80243649554',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '4/8',
                                                                                                                 'id': '4/8',
                                                                                                                 'links': [       {       'connected_time': '688369',
                                                                                                                                          'fiber_length': 1342,
                                                                                                                                          'poller_hash': 'MkIwMjVEcGVlcnBvbnBvbiAyNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '15498562',
                                                                                                                                          'fiber_length': 1677,
                                                                                                                                          'poller_hash': 'MjU5ODM3cGVlcnBvbnBvbiAyOTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '10799098',
                                                                                                                                          'fiber_length': 1761,
                                                                                                                                          'poller_hash': 'MkIwMjA2cGVlcnBvbnBvbiAyOTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '688371',
                                                                                                                                          'fiber_length': 1407,
                                                                                                                                          'poller_hash': 'MkIwMjcycGVlcnBvbnBvbiAyOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '29301849',
                                                                                                                                          'fiber_length': 1271,
                                                                                                                                          'poller_hash': 'Mjk4M0VEcGVlcnBvbnBvbiAyOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '72824819',
                                                                                                                                          'fiber_length': 1496,
                                                                                                                                          'poller_hash': 'MjA1REZDcGVlcnBvbnBvbiAyOTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '21816413',
                                                                                                                                          'fiber_length': 1345,
                                                                                                                                          'poller_hash': 'Mjk4NDVBcGVlcnBvbnBvbiAyOTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '28453948',
                                                                                                                                          'fiber_length': 1272,
                                                                                                                                          'poller_hash': 'Mjk4MjRGcGVlcnBvbnBvbiAyOTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '2151357',
                                                                                                                                          'fiber_length': 1351,
                                                                                                                                          'poller_hash': 'MjA1RENGcGVlcnBvbnBvbiAzMDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '21816166',
                                                                                                                                          'fiber_length': 1329,
                                                                                                                                          'poller_hash': 'Mjk4Mzg1cGVlcnBvbnBvbiAzMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '688392',
                                                                                                                                          'fiber_length': 1243,
                                                                                                                                          'poller_hash': 'NUE2NkJBcGVlcnBvbnBvbiA0MjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '21816546',
                                                                                                                                          'fiber_length': 1560,
                                                                                                                                          'poller_hash': 'MjA1REVFcGVlcnBvbnBvbiA0NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '688380',
                                                                                                                                          'fiber_length': 1426,
                                                                                                                                          'poller_hash': 'MkIwMTgxcGVlcnBvbnBvbiA1MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '67790238',
                                                                                                                                          'fiber_length': 1397,
                                                                                                                                          'poller_hash': 'ODQ0QkQ2cGVlcnBvbnBvbiAxNDA1Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68},
                                                                                                                                  {       'connected_time': '21816677',
                                                                                                                                          'fiber_length': 1485,
                                                                                                                                          'poller_hash': 'QzgzQkU2cGVlcnBvbnBvbiAxNzIzNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.68}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '4/8',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNTEyNjAzNjgx',
                                                           'serial_number': '211512603681',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42990418'},
                                                   {       'id': '5',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '5/1',
                                                                                         'in_discards': '16167',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1716082',
                                                                                         'in_octets': '7737243812738',
                                                                                         'in_ucast_pkts': '33355304797',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '58790081170',
                                                                                                                  'in_ucast_pkts': '463629882',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '2527133650555',
                                                                                                                  'out_ucast_pkts': '1762970241',
                                                                                                                  'poller_hash': 'NDJCRUE3cGVlcmludGZwb24gODEz'},
                                                                                                          {       'in_nucast_pkts': '66618778',
                                                                                                                  'in_octets': '381633014468',
                                                                                                                  'in_ucast_pkts': '2717653620',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63485708336',
                                                                                                                  'out_octets': '93626962161218',
                                                                                                                  'out_ucast_pkts': '5748029905',
                                                                                                                  'poller_hash': 'NDJCRUEycGVlcmludGZwb24gODE1'},
                                                                                                          {       'in_nucast_pkts': '1370',
                                                                                                                  'in_octets': '845195844',
                                                                                                                  'in_ucast_pkts': '3428462',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447177',
                                                                                                                  'out_octets': '12405562249',
                                                                                                                  'out_ucast_pkts': '9708271',
                                                                                                                  'poller_hash': 'NUE2NjY1cGVlcmludGZwb24gODUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '501769043660',
                                                                                                                  'in_ucast_pkts': '2019428890',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '8901787679893',
                                                                                                                  'out_ucast_pkts': '6462736051',
                                                                                                                  'poller_hash': 'ODY5NDE1cGVlcmludGZwb24gMTQxOTE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1338142167095',
                                                                                                                  'in_ucast_pkts': '3253097975',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '7856821298234',
                                                                                                                  'out_ucast_pkts': '6339414335',
                                                                                                                  'poller_hash': 'ODY5NDBBcGVlcmludGZwb24gMTQxOTI3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '97241683566',
                                                                                                                  'in_ucast_pkts': '331057998',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '1613851457507',
                                                                                                                  'out_ucast_pkts': '1243777444',
                                                                                                                  'poller_hash': 'ODY5M0FGcGVlcmludGZwb24gMTQxOTQ3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '8524470669',
                                                                                                                  'in_ucast_pkts': '85189087',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '263731270759',
                                                                                                                  'out_ucast_pkts': '233840961',
                                                                                                                  'poller_hash': 'ODY5M0RFcGVlcmludGZwb24gMTQxOTU3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '139093459877',
                                                                                                                  'in_ucast_pkts': '5164304904',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '4451293077739',
                                                                                                                  'out_ucast_pkts': '3384813338',
                                                                                                                  'poller_hash': 'ODY5M0QycGVlcmludGZwb24gMTQyMDc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODcyNTUzcGVlcmludGZwb24gMTQyNzI1'},
                                                                                                          {       'in_nucast_pkts': '408',
                                                                                                                  'in_octets': '181887662914',
                                                                                                                  'in_ucast_pkts': '1166043982',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '44367200',
                                                                                                                  'out_octets': '4279798656659',
                                                                                                                  'out_ucast_pkts': '3126969233',
                                                                                                                  'poller_hash': 'REZGRTEwcGVlcmludGZwb24gMTcwOTQ5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '152109406897',
                                                                                                                  'in_ucast_pkts': '1263685400',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '6164660759485',
                                                                                                                  'out_ucast_pkts': '4448146166',
                                                                                                                  'poller_hash': 'QTlDOTBGcGVlcmludGZwb24gMTkzMzE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '441195473245',
                                                                                                                  'in_ucast_pkts': '5591397608',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '7970225170103',
                                                                                                                  'out_ucast_pkts': '6104260530',
                                                                                                                  'poller_hash': 'QzgxRUIwcGVlcmludGZwb24gMjAyODE4'},
                                                                                                          {       'in_nucast_pkts': '337236',
                                                                                                                  'in_octets': '3789793684',
                                                                                                                  'in_ucast_pkts': '12179639',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447542',
                                                                                                                  'out_octets': '52025578820',
                                                                                                                  'out_ucast_pkts': '39170411',
                                                                                                                  'poller_hash': 'QzgxRThEcGVlcmludGZwb24gMjAzODEz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '258293253090',
                                                                                                                  'in_ucast_pkts': '1362058557',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62',
                                                                                                                  'out_octets': '3806740244785',
                                                                                                                  'out_ucast_pkts': '3200725239',
                                                                                                                  'poller_hash': 'QzgxRTEwcGVlcmludGZwb24gMjA0MjIz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '32453739266',
                                                                                                                  'in_ucast_pkts': '74319388',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '312007839850',
                                                                                                                  'out_ucast_pkts': '301622566',
                                                                                                                  'poller_hash': 'QzgzMDcwcGVlcmludGZwb24gMjExNzg1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTFCMjVENHBlZXJpbnRmcG9uIDIxNTc3Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE5NDVEMXBlZXJpbnRmcG9uIDIxNTgzNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEFCNXBlZXJpbnRmcG9uIDIzMzI4OQ=='},
                                                                                                          {       'in_nucast_pkts': '123838',
                                                                                                                  'in_octets': '30398753323',
                                                                                                                  'in_ucast_pkts': '374665557',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '13466096',
                                                                                                                  'out_octets': '1956672456503',
                                                                                                                  'out_ucast_pkts': '1309326951',
                                                                                                                  'poller_hash': 'MTI3NEZGOXBlZXJpbnRmcG9uIDIzNDkxNA=='},
                                                                                                          {       'in_nucast_pkts': '234760',
                                                                                                                  'in_octets': '291979630890',
                                                                                                                  'in_ucast_pkts': '634793017',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27881900',
                                                                                                                  'out_octets': '3249671507104',
                                                                                                                  'out_ucast_pkts': '2501565286',
                                                                                                                  'poller_hash': 'MTI5NzQ1QXBlZXJpbnRmcG9uIDI1NjQ2Mg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '42738883454',
                                                                                         'out_octets': '183613670188187',
                                                                                         'out_ucast_pkts': '93699903017',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/1',
                                                                                                                 'id': '5/1',
                                                                                                                 'links': [       {       'connected_time': '73034707',
                                                                                                                                          'fiber_length': 8792,
                                                                                                                                          'poller_hash': 'NDJCRUE3cGVlcnBvbnBvbiA4MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '21815978',
                                                                                                                                          'fiber_length': 9078,
                                                                                                                                          'poller_hash': 'NDJCRUEycGVlcnBvbnBvbiA4MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '688391',
                                                                                                                                          'fiber_length': 8704,
                                                                                                                                          'poller_hash': 'NUE2NjY1cGVlcnBvbnBvbiA4NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '688374',
                                                                                                                                          'fiber_length': 10296,
                                                                                                                                          'poller_hash': 'ODY5NDE1cGVlcnBvbnBvbiAxNDE5MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '13121918',
                                                                                                                                          'fiber_length': 11030,
                                                                                                                                          'poller_hash': 'ODY5NDBBcGVlcnBvbnBvbiAxNDE5Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '2082693',
                                                                                                                                          'fiber_length': 11201,
                                                                                                                                          'poller_hash': 'ODY5M0FGcGVlcnBvbnBvbiAxNDE5NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '21817170',
                                                                                                                                          'fiber_length': 9364,
                                                                                                                                          'poller_hash': 'ODY5M0RFcGVlcnBvbnBvbiAxNDE5NTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '21816880',
                                                                                                                                          'fiber_length': 10386,
                                                                                                                                          'poller_hash': 'ODY5M0QycGVlcnBvbnBvbiAxNDIwNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '19808706',
                                                                                                                                          'fiber_length': 10716,
                                                                                                                                          'poller_hash': 'ODcyNTUzcGVlcnBvbnBvbiAxNDI3MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '19523274',
                                                                                                                                          'fiber_length': 10488,
                                                                                                                                          'poller_hash': 'REZGRTEwcGVlcnBvbnBvbiAxNzA5NDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '21817023',
                                                                                                                                          'fiber_length': 9662,
                                                                                                                                          'poller_hash': 'QTlDOTBGcGVlcnBvbnBvbiAxOTMzMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '73426872',
                                                                                                                                          'fiber_length': 9385,
                                                                                                                                          'poller_hash': 'QzgxRUIwcGVlcnBvbnBvbiAyMDI4MTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '688414',
                                                                                                                                          'fiber_length': 8409,
                                                                                                                                          'poller_hash': 'QzgxRThEcGVlcnBvbnBvbiAyMDM4MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '21816474',
                                                                                                                                          'fiber_length': 8168,
                                                                                                                                          'poller_hash': 'QzgxRTEwcGVlcnBvbnBvbiAyMDQyMjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '21816130',
                                                                                                                                          'fiber_length': 8705,
                                                                                                                                          'poller_hash': 'QzgzMDcwcGVlcnBvbnBvbiAyMTE3ODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '19363903',
                                                                                                                                          'fiber_length': 8450,
                                                                                                                                          'poller_hash': 'MTFCMjVENHBlZXJwb25wb24gMjE1Nzcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '19197442',
                                                                                                                                          'fiber_length': 8831,
                                                                                                                                          'poller_hash': 'MTE5NDVEMXBlZXJwb25wb24gMjE1ODM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '5354267',
                                                                                                                                          'fiber_length': 9878,
                                                                                                                                          'poller_hash': 'MTI3NEFCNXBlZXJwb25wb24gMjMzMjg5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '5868607',
                                                                                                                                          'fiber_length': 7931,
                                                                                                                                          'poller_hash': 'MTI3NEZGOXBlZXJwb25wb24gMjM0OTE0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05},
                                                                                                                                  {       'connected_time': '12377084',
                                                                                                                                          'fiber_length': 10048,
                                                                                                                                          'poller_hash': 'MTI5NzQ1QXBlZXJwb25wb24gMjU2NDYy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.05}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/1',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/2',
                                                                                         'in_discards': '183094',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '9305332',
                                                                                         'in_octets': '16997171666897',
                                                                                         'in_ucast_pkts': '68258153669',
                                                                                         'links': [       {       'in_nucast_pkts': '294734',
                                                                                                                  'in_octets': '216883787476',
                                                                                                                  'in_ucast_pkts': '594040066',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26865254',
                                                                                                                  'out_octets': '3170836455542',
                                                                                                                  'out_ucast_pkts': '2328487646',
                                                                                                                  'poller_hash': 'NzFBNENDcGVlcmludGZwb24gMjU5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '33105229337',
                                                                                                                  'in_ucast_pkts': '288787776',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1304370696857',
                                                                                                                  'out_ucast_pkts': '974275016',
                                                                                                                  'poller_hash': 'NzFBNENGcGVlcmludGZwb24gMjYw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '230812226566',
                                                                                                                  'in_ucast_pkts': '2217385669',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6608325717704',
                                                                                                                  'out_ucast_pkts': '6569871492',
                                                                                                                  'poller_hash': 'M0ZBOTBEcGVlcmludGZwb24gMjYx'},
                                                                                                          {       'in_nucast_pkts': '23',
                                                                                                                  'in_octets': '23898763152',
                                                                                                                  'in_ucast_pkts': '92316564',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447114',
                                                                                                                  'out_octets': '278476396992',
                                                                                                                  'out_ucast_pkts': '213910594',
                                                                                                                  'poller_hash': 'ODcyNDMycGVlcmludGZwb24gMjYy'},
                                                                                                          {       'in_nucast_pkts': '66701',
                                                                                                                  'in_octets': '25826350596',
                                                                                                                  'in_ucast_pkts': '95915845',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6672535',
                                                                                                                  'out_octets': '352756150426',
                                                                                                                  'out_ucast_pkts': '262049688',
                                                                                                                  'poller_hash': 'ODY5M0Q4cGVlcmludGZwb24gMjY0'},
                                                                                                          {       'in_nucast_pkts': '14613',
                                                                                                                  'in_octets': '45525902578',
                                                                                                                  'in_ucast_pkts': '315847210',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6705620',
                                                                                                                  'out_octets': '1359056735249',
                                                                                                                  'out_ucast_pkts': '1042813540',
                                                                                                                  'poller_hash': 'MTI5NUVDMXBlZXJpbnRmcG9uIDI3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2562795536683',
                                                                                                                  'in_ucast_pkts': '9243630750',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '42158244888375',
                                                                                                                  'out_ucast_pkts': '31181519067',
                                                                                                                  'poller_hash': 'MTI0ODRFcGVlcmludGZwb24gMjc4'},
                                                                                                          {       'in_nucast_pkts': '72',
                                                                                                                  'in_octets': '66328677332',
                                                                                                                  'in_ucast_pkts': '668128851',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53663600',
                                                                                                                  'out_octets': '3909817787367',
                                                                                                                  'out_ucast_pkts': '2620191485',
                                                                                                                  'poller_hash': 'Q0JGRjVwZWVyaW50ZnBvbiAyNzk='},
                                                                                                          {       'in_nucast_pkts': '517194',
                                                                                                                  'in_octets': '120563533209',
                                                                                                                  'in_ucast_pkts': '456670317',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27081373',
                                                                                                                  'out_octets': '2428181878168',
                                                                                                                  'out_ucast_pkts': '1801769523',
                                                                                                                  'poller_hash': 'MTMwMEQ2cGVlcmludGZwb24gMjgx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '567337392130',
                                                                                                                  'in_ucast_pkts': '4382959852',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '23129267556870',
                                                                                                                  'out_ucast_pkts': '16115199838',
                                                                                                                  'poller_hash': 'MTc1RTcwcGVlcmludGZwb24gMjgy'},
                                                                                                          {       'in_nucast_pkts': '368',
                                                                                                                  'in_octets': '131456322460',
                                                                                                                  'in_ucast_pkts': '256298285',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77874752',
                                                                                                                  'out_octets': '671222814870',
                                                                                                                  'out_ucast_pkts': '582706493',
                                                                                                                  'poller_hash': 'RkQ4NzNDcGVlcmludGZwb24gMjgz'},
                                                                                                          {       'in_nucast_pkts': '797853',
                                                                                                                  'in_octets': '1227044624553',
                                                                                                                  'in_ucast_pkts': '5147749106',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77791438',
                                                                                                                  'out_octets': '18002206058634',
                                                                                                                  'out_ucast_pkts': '13502299906',
                                                                                                                  'poller_hash': 'MzkwQzVFcGVlcmludGZwb24gNDgx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '145959188740',
                                                                                                                  'in_ucast_pkts': '522187134',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '1758110118038',
                                                                                                                  'out_ucast_pkts': '1439168946',
                                                                                                                  'poller_hash': 'NDJCOUMzcGVlcmludGZwb24gNTQ2'},
                                                                                                          {       'in_nucast_pkts': '18',
                                                                                                                  'in_octets': '6477136642',
                                                                                                                  'in_ucast_pkts': '31125299',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447193',
                                                                                                                  'out_octets': '238612823926',
                                                                                                                  'out_ucast_pkts': '172371251',
                                                                                                                  'poller_hash': 'NDJCRjFEcGVlcmludGZwb24gNjUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '141418925013',
                                                                                                                  'in_ucast_pkts': '914908673',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '4064085761154',
                                                                                                                  'out_ucast_pkts': '2986767287',
                                                                                                                  'poller_hash': 'NDJDNzQ4cGVlcmludGZwb24gNjkz'},
                                                                                                          {       'in_nucast_pkts': '4760',
                                                                                                                  'in_octets': '13090094586',
                                                                                                                  'in_ucast_pkts': '39974737',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6685478',
                                                                                                                  'out_octets': '157648239002',
                                                                                                                  'out_ucast_pkts': '127026517',
                                                                                                                  'poller_hash': 'NUE2QzY4cGVlcmludGZwb24gMTM3ODMy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODY5M0JFcGVlcmludGZwb24gMTQxOTg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '293617744584',
                                                                                                                  'in_ucast_pkts': '2135526991',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '9277672478420',
                                                                                                                  'out_ucast_pkts': '6358747766',
                                                                                                                  'poller_hash': 'ODcyNDM1cGVlcmludGZwb24gMTQzODAy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '171191420863',
                                                                                                                  'in_ucast_pkts': '399454065',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '1948138557532',
                                                                                                                  'out_ucast_pkts': '1505546976',
                                                                                                                  'poller_hash': 'QjBBMjc0cGVlcmludGZwb24gMTUxNTEw'},
                                                                                                          {       'in_nucast_pkts': '470',
                                                                                                                  'in_octets': '2501516450',
                                                                                                                  'in_ucast_pkts': '7313788',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186632',
                                                                                                                  'out_octets': '49014935556',
                                                                                                                  'out_ucast_pkts': '46627593',
                                                                                                                  'poller_hash': 'QjUxMTY5cGVlcmludGZwb24gMTU1MDIz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '498181554242',
                                                                                                                  'in_ucast_pkts': '5246290809',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '26620967230644',
                                                                                                                  'out_ucast_pkts': '18837897906',
                                                                                                                  'poller_hash': 'NDcxNzU1cGVlcmludGZwb24gMTY2MjEx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '137959005841',
                                                                                                                  'in_ucast_pkts': '5280337672',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '217',
                                                                                                                  'out_octets': '3115638108413',
                                                                                                                  'out_ucast_pkts': '2220388981',
                                                                                                                  'poller_hash': 'QzgzQjBFcGVlcmludGZwb24gMTY2Njc4'},
                                                                                                          {       'in_nucast_pkts': '57019',
                                                                                                                  'in_octets': '41774324518',
                                                                                                                  'in_ucast_pkts': '144266253',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3913938587',
                                                                                                                  'out_octets': '5623925090908',
                                                                                                                  'out_ucast_pkts': '324822623',
                                                                                                                  'poller_hash': 'QzgxREYzcGVlcmludGZwb24gMTg5ODc5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '80249061261',
                                                                                                                  'in_ucast_pkts': '498672635',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '2680553607091',
                                                                                                                  'out_ucast_pkts': '1883399228',
                                                                                                                  'poller_hash': 'QzgzMjQ2cGVlcmludGZwb24gMjA1MTIw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '211533567714',
                                                                                                                  'in_ucast_pkts': '616622013',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3379657791838',
                                                                                                                  'out_ucast_pkts': '2635314220',
                                                                                                                  'poller_hash': 'MkVENjdCcGVlcmludGZwb24gMjA1NTI2'},
                                                                                                          {       'in_nucast_pkts': '642518',
                                                                                                                  'in_octets': '213393183253',
                                                                                                                  'in_ucast_pkts': '816642142',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '89185721',
                                                                                                                  'out_octets': '5820353024167',
                                                                                                                  'out_ucast_pkts': '4633841659',
                                                                                                                  'poller_hash': 'QzgxNkU2cGVlcmludGZwb24gMjEzNjgx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDA1MHBlZXJpbnRmcG9uIDI0NDg1MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '927990395323',
                                                                                                                  'in_ucast_pkts': '4536997227',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '18475305747045',
                                                                                                                  'out_ucast_pkts': '13089081302',
                                                                                                                  'poller_hash': 'NTVGOUJDcGVlcmludGZwb24gMjQ0OTAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REMyMXBlZXJpbnRmcG9uIDI1NDA3Mg=='},
                                                                                                          {       'in_nucast_pkts': '14813',
                                                                                                                  'in_octets': '5535322238',
                                                                                                                  'in_ucast_pkts': '34607884',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1448004',
                                                                                                                  'out_octets': '167389906035',
                                                                                                                  'out_ucast_pkts': '115593807',
                                                                                                                  'poller_hash': 'MTI3NDdDNnBlZXJpbnRmcG9uIDI2MzE5MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '66518501041',
                                                                                                                  'in_ucast_pkts': '180567147',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '312194164772',
                                                                                                                  'out_ucast_pkts': '320928210',
                                                                                                                  'poller_hash': 'MTI3NDU4RnBlZXJpbnRmcG9uIDI3MzUxMw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '153222580083',
                                                                                         'out_octets': '485698315012844',
                                                                                         'out_ucast_pkts': '205163340095',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/2',
                                                                                                                 'id': '5/2',
                                                                                                                 'links': [       {       'connected_time': '15498431',
                                                                                                                                          'fiber_length': 7870,
                                                                                                                                          'poller_hash': 'NzFBNENDcGVlcnBvbnBvbiAyNTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '5470219',
                                                                                                                                          'fiber_length': 7313,
                                                                                                                                          'poller_hash': 'NzFBNENGcGVlcnBvbnBvbiAyNjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '47147',
                                                                                                                                          'fiber_length': 10084,
                                                                                                                                          'poller_hash': 'M0ZBOTBEcGVlcnBvbnBvbiAyNjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688382',
                                                                                                                                          'fiber_length': 7985,
                                                                                                                                          'poller_hash': 'ODcyNDMycGVlcnBvbnBvbiAyNjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '3113859',
                                                                                                                                          'fiber_length': 10020,
                                                                                                                                          'poller_hash': 'ODY5M0Q4cGVlcnBvbnBvbiAyNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '3126195',
                                                                                                                                          'fiber_length': 6390,
                                                                                                                                          'poller_hash': 'MTI5NUVDMXBlZXJwb25wb24gMjc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '15413616',
                                                                                                                                          'fiber_length': 7214,
                                                                                                                                          'poller_hash': 'MTI0ODRFcGVlcnBvbnBvbiAyNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '31199410',
                                                                                                                                          'fiber_length': 9966,
                                                                                                                                          'poller_hash': 'Q0JGRjVwZWVycG9ucG9uIDI3OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '15635754',
                                                                                                                                          'fiber_length': 6630,
                                                                                                                                          'poller_hash': 'MTMwMEQ2cGVlcnBvbnBvbiAyODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '64048647',
                                                                                                                                          'fiber_length': 9341,
                                                                                                                                          'poller_hash': 'MTc1RTcwcGVlcnBvbnBvbiAyODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '42328',
                                                                                                                                          'fiber_length': 8099,
                                                                                                                                          'poller_hash': 'RkQ4NzNDcGVlcnBvbnBvbiAyODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '41151',
                                                                                                                                          'fiber_length': 9861,
                                                                                                                                          'poller_hash': 'MzkwQzVFcGVlcnBvbnBvbiA0ODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '8925450',
                                                                                                                                          'fiber_length': 9323,
                                                                                                                                          'poller_hash': 'NDJCOUMzcGVlcnBvbnBvbiA1NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688383',
                                                                                                                                          'fiber_length': 7409,
                                                                                                                                          'poller_hash': 'NDJCRjFEcGVlcnBvbnBvbiA2NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '21816686',
                                                                                                                                          'fiber_length': 7080,
                                                                                                                                          'poller_hash': 'NDJDNzQ4cGVlcnBvbnBvbiA2OTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '3113920',
                                                                                                                                          'fiber_length': 11783,
                                                                                                                                          'poller_hash': 'NUE2QzY4cGVlcnBvbnBvbiAxMzc4MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '21816516',
                                                                                                                                          'fiber_length': 8718,
                                                                                                                                          'poller_hash': 'ODY5M0JFcGVlcnBvbnBvbiAxNDE5ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688383',
                                                                                                                                          'fiber_length': 8755,
                                                                                                                                          'poller_hash': 'ODcyNDM1cGVlcnBvbnBvbiAxNDM4MDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '21816699',
                                                                                                                                          'fiber_length': 5626,
                                                                                                                                          'poller_hash': 'QjBBMjc0cGVlcnBvbnBvbiAxNTE1MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '90240',
                                                                                                                                          'fiber_length': 6478,
                                                                                                                                          'poller_hash': 'QjUxMTY5cGVlcnBvbnBvbiAxNTUwMjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '995931',
                                                                                                                                          'fiber_length': 9432,
                                                                                                                                          'poller_hash': 'NDcxNzU1cGVlcnBvbnBvbiAxNjYyMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688378',
                                                                                                                                          'fiber_length': 6950,
                                                                                                                                          'poller_hash': 'QzgzQjBFcGVlcnBvbnBvbiAxNjY2Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '3113868',
                                                                                                                                          'fiber_length': 12133,
                                                                                                                                          'poller_hash': 'QzgxREYzcGVlcnBvbnBvbiAxODk4Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '30163283',
                                                                                                                                          'fiber_length': 5717,
                                                                                                                                          'poller_hash': 'QzgzMjQ2cGVlcnBvbnBvbiAyMDUxMjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '47694',
                                                                                                                                          'fiber_length': 10802,
                                                                                                                                          'poller_hash': 'MkVENjdCcGVlcnBvbnBvbiAyMDU1MjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '70485238',
                                                                                                                                          'fiber_length': 7294,
                                                                                                                                          'poller_hash': 'QzgxNkU2cGVlcnBvbnBvbiAyMTM2ODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '69402860',
                                                                                                                                          'fiber_length': 6156,
                                                                                                                                          'poller_hash': 'MTI3NDA1MHBlZXJwb25wb24gMjQ0ODUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '23019371',
                                                                                                                                          'fiber_length': 10891,
                                                                                                                                          'poller_hash': 'NTVGOUJDcGVlcnBvbnBvbiAyNDQ5MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '28343873',
                                                                                                                                          'fiber_length': 5754,
                                                                                                                                          'poller_hash': 'MTE3REMyMXBlZXJwb25wb24gMjU0MDcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688381',
                                                                                                                                          'fiber_length': 8478,
                                                                                                                                          'poller_hash': 'MTI3NDdDNnBlZXJwb25wb24gMjYzMTkx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '15219615',
                                                                                                                                          'fiber_length': 6948,
                                                                                                                                          'poller_hash': 'MTI3NDU4RnBlZXJwb25wb24gMjczNTEz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/2',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/3',
                                                                                         'in_discards': '55007',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '8417658',
                                                                                         'in_octets': '19948980943258',
                                                                                         'in_ucast_pkts': '77339737504',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1082939784236',
                                                                                                                  'in_ucast_pkts': '3128667823',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '15222932335505',
                                                                                                                  'out_ucast_pkts': '11835610199',
                                                                                                                  'poller_hash': 'MjQ4QUQ0cGVlcmludGZwb24gMzIx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2572407078465',
                                                                                                                  'in_ucast_pkts': '3771613008',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2819861003305',
                                                                                                                  'out_ucast_pkts': '4663743411',
                                                                                                                  'poller_hash': 'NzFBNTFBcGVlcmludGZwb24gMzI5'},
                                                                                                          {       'in_nucast_pkts': '181054',
                                                                                                                  'in_octets': '139839145579',
                                                                                                                  'in_ucast_pkts': '1091254656',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '16921922',
                                                                                                                  'out_octets': '3989707791574',
                                                                                                                  'out_ucast_pkts': '2925202328',
                                                                                                                  'poller_hash': 'NzZFRENEcGVlcmludGZwb24gMzM5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '161185898864',
                                                                                                                  'in_ucast_pkts': '1572099606',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4441347647900',
                                                                                                                  'out_ucast_pkts': '4744238257',
                                                                                                                  'poller_hash': 'NzFBNEY2cGVlcmludGZwb24gMzQw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDE4QjBwZWVyaW50ZnBvbiAzNDE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QkYyNDI5cGVlcmludGZwb24gMzQy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '626969060488',
                                                                                                                  'in_ucast_pkts': '4086073607',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7127650130313',
                                                                                                                  'out_ucast_pkts': '9825480261',
                                                                                                                  'poller_hash': 'ODQ1OTA5cGVlcmludGZwb24gMzQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '330205306008',
                                                                                                                  'in_ucast_pkts': '1890253469',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '5195624897762',
                                                                                                                  'out_ucast_pkts': '4853551418',
                                                                                                                  'poller_hash': 'NzI0QUQwcGVlcmludGZwb24gMzQ2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '120735094787',
                                                                                                                  'in_ucast_pkts': '1115828998',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4439568775719',
                                                                                                                  'out_ucast_pkts': '3065595611',
                                                                                                                  'poller_hash': 'QzgxRjcxcGVlcmludGZwb24gMzUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '842438257332',
                                                                                                                  'in_ucast_pkts': '4181290831',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '16405775303258',
                                                                                                                  'out_ucast_pkts': '12479383066',
                                                                                                                  'poller_hash': 'ODY5M0M1cGVlcmludGZwb24gMzU0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'N0UwRTlwZWVyaW50ZnBvbiAzNTU='},
                                                                                                          {       'in_nucast_pkts': '59647',
                                                                                                                  'in_octets': '35052492275',
                                                                                                                  'in_ucast_pkts': '329072288',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5128759',
                                                                                                                  'out_octets': '964042991148',
                                                                                                                  'out_ucast_pkts': '662333039',
                                                                                                                  'poller_hash': 'MzMwM0IxcGVlcmludGZwb24gMzU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '7030615822267',
                                                                                                                  'in_ucast_pkts': '10469450621',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '10740289541826',
                                                                                                                  'out_ucast_pkts': '339491145',
                                                                                                                  'poller_hash': 'NTU2MzgycGVlcmludGZwb24gMzU3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '84105052863',
                                                                                                                  'in_ucast_pkts': '5056633225',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '5025117925011',
                                                                                                                  'out_ucast_pkts': '15997719252',
                                                                                                                  'poller_hash': 'QTlDOEFBcGVlcmludGZwb24gMzU4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '145585917585',
                                                                                                                  'in_ucast_pkts': '1056496658',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1628359898019',
                                                                                                                  'out_ucast_pkts': '2138018853',
                                                                                                                  'poller_hash': 'NzIyMjA3cGVlcmludGZwb24gMzYw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTU2M0E5cGVlcmludGZwb24gMzYx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '833339544399',
                                                                                                                  'in_ucast_pkts': '5617391324',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '15443679623052',
                                                                                                                  'out_ucast_pkts': '14797750296',
                                                                                                                  'poller_hash': 'M0ZDODQ1cGVlcmludGZwb24gMzY0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '359390436122',
                                                                                                                  'in_ucast_pkts': '2236240146',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '5664278220074',
                                                                                                                  'out_ucast_pkts': '6209480659',
                                                                                                                  'poller_hash': 'ODQ1QkI0cGVlcmludGZwb24gMzY3'},
                                                                                                          {       'in_nucast_pkts': '186156',
                                                                                                                  'in_octets': '349532984192',
                                                                                                                  'in_ucast_pkts': '1509217629',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45076276',
                                                                                                                  'out_octets': '3783212143012',
                                                                                                                  'out_ucast_pkts': '2860432452',
                                                                                                                  'poller_hash': 'NDU5NjVEcGVlcmludGZwb24gMzY5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1100840934407',
                                                                                                                  'in_ucast_pkts': '10106123268',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '35064146098830',
                                                                                                                  'out_ucast_pkts': '28493369710',
                                                                                                                  'poller_hash': 'MkVEOUNEcGVlcmludGZwb24gMzcw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '440250072245',
                                                                                                                  'in_ucast_pkts': '3608902629',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '15226219448059',
                                                                                                                  'out_ucast_pkts': '12937185292',
                                                                                                                  'poller_hash': 'NzIzM0VBcGVlcmludGZwb24gMzcx'},
                                                                                                          {       'in_nucast_pkts': '54481',
                                                                                                                  'in_octets': '32115520283',
                                                                                                                  'in_ucast_pkts': '113947729',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6685861',
                                                                                                                  'out_octets': '705511764726',
                                                                                                                  'out_ucast_pkts': '531196435',
                                                                                                                  'poller_hash': 'QjBBMkE2cGVlcmludGZwb24gMTUwMDI0'},
                                                                                                          {       'in_nucast_pkts': '6825870',
                                                                                                                  'in_octets': '366764225064',
                                                                                                                  'in_ucast_pkts': '2967801017',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15595125988',
                                                                                                                  'out_octets': '37393486188257',
                                                                                                                  'out_ucast_pkts': '11183911375',
                                                                                                                  'poller_hash': 'QzgzQzIxcGVlcmludGZwb24gMTcyMDUy'},
                                                                                                          {       'in_nucast_pkts': '64908',
                                                                                                                  'in_octets': '43550689351',
                                                                                                                  'in_ucast_pkts': '105680005',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6356865',
                                                                                                                  'out_octets': '904841657689',
                                                                                                                  'out_ucast_pkts': '648227499',
                                                                                                                  'poller_hash': 'QzgzMTYzcGVlcmludGZwb24gMTcyNDQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '314867426441',
                                                                                                                  'in_ucast_pkts': '3609248501',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '22495735825037',
                                                                                                                  'out_ucast_pkts': '15011696532',
                                                                                                                  'poller_hash': 'QzgxRDU0cGVlcmludGZwb24gMjA2Mzgy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '110595665721',
                                                                                                                  'in_ucast_pkts': '643524565',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '3469789744542',
                                                                                                                  'out_ucast_pkts': '2415004597',
                                                                                                                  'poller_hash': 'MTI3NEE0NHBlZXJpbnRmcG9uIDIzNjIwNg=='},
                                                                                                          {       'in_nucast_pkts': '5603979',
                                                                                                                  'in_octets': '81116101640',
                                                                                                                  'in_ucast_pkts': '979565880',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '338850129',
                                                                                                                  'out_octets': '5645452422945',
                                                                                                                  'out_ucast_pkts': '3803059699',
                                                                                                                  'poller_hash': 'MTI3NDE5QXBlZXJpbnRmcG9uIDI0Mzc1MQ=='},
                                                                                                          {       'in_nucast_pkts': '64667',
                                                                                                                  'in_octets': '15932476656',
                                                                                                                  'in_ucast_pkts': '92133693',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6357809',
                                                                                                                  'out_octets': '677837916371',
                                                                                                                  'out_ucast_pkts': '502367444',
                                                                                                                  'poller_hash': 'MTE3REJCRXBlZXJpbnRmcG9uIDI0OTIzMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '91671758352',
                                                                                                                  'in_ucast_pkts': '813851319',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '4732356850273',
                                                                                                                  'out_ucast_pkts': '3256996060',
                                                                                                                  'poller_hash': 'MTE5NTE3M3BlZXJpbnRmcG9uIDI1MTUyOQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '115556592313',
                                                                                         'out_octets': '465832994158684',
                                                                                         'out_ucast_pkts': '223521673408',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/3',
                                                                                                                 'id': '5/3',
                                                                                                                 'links': [       {       'connected_time': '26736474',
                                                                                                                                          'fiber_length': 18562,
                                                                                                                                          'poller_hash': 'MjQ4QUQ0cGVlcnBvbnBvbiAzMjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '47457',
                                                                                                                                          'fiber_length': 18245,
                                                                                                                                          'poller_hash': 'NzFBNTFBcGVlcnBvbnBvbiAzMjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '9506531',
                                                                                                                                          'fiber_length': 18068,
                                                                                                                                          'poller_hash': 'NzZFRENEcGVlcnBvbnBvbiAzMzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '47651',
                                                                                                                                          'fiber_length': 17366,
                                                                                                                                          'poller_hash': 'NzFBNEY2cGVlcnBvbnBvbiAzNDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '24664576',
                                                                                                                                          'fiber_length': 17701,
                                                                                                                                          'poller_hash': 'NDE4QjBwZWVycG9ucG9uIDM0MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '47254',
                                                                                                                                          'fiber_length': 18232,
                                                                                                                                          'poller_hash': 'QkYyNDI5cGVlcnBvbnBvbiAzNDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '47731',
                                                                                                                                          'fiber_length': 17666,
                                                                                                                                          'poller_hash': 'ODQ1OTA5cGVlcnBvbnBvbiAzNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '26425410',
                                                                                                                                          'fiber_length': 18199,
                                                                                                                                          'poller_hash': 'NzI0QUQwcGVlcnBvbnBvbiAzNDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '3113835',
                                                                                                                                          'fiber_length': 18116,
                                                                                                                                          'poller_hash': 'QzgxRjcxcGVlcnBvbnBvbiAzNTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '585723',
                                                                                                                                          'fiber_length': 17990,
                                                                                                                                          'poller_hash': 'ODY5M0M1cGVlcnBvbnBvbiAzNTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '368751',
                                                                                                                                          'fiber_length': 18284,
                                                                                                                                          'poller_hash': 'N0UwRTlwZWVycG9ucG9uIDM1NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '3113830',
                                                                                                                                          'fiber_length': 17980,
                                                                                                                                          'poller_hash': 'MzMwM0IxcGVlcnBvbnBvbiAzNTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '29355974',
                                                                                                                                          'fiber_length': 18459,
                                                                                                                                          'poller_hash': 'NTU2MzgycGVlcnBvbnBvbiAzNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '30942276',
                                                                                                                                          'fiber_length': 18289,
                                                                                                                                          'poller_hash': 'QTlDOEFBcGVlcnBvbnBvbiAzNTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '47307',
                                                                                                                                          'fiber_length': 17608,
                                                                                                                                          'poller_hash': 'NzIyMjA3cGVlcnBvbnBvbiAzNjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '24301068',
                                                                                                                                          'fiber_length': 18269,
                                                                                                                                          'poller_hash': 'NTU2M0E5cGVlcnBvbnBvbiAzNjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '26617541',
                                                                                                                                          'fiber_length': 17448,
                                                                                                                                          'poller_hash': 'M0ZDODQ1cGVlcnBvbnBvbiAzNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '8200272',
                                                                                                                                          'fiber_length': 18194,
                                                                                                                                          'poller_hash': 'ODQ1QkI0cGVlcnBvbnBvbiAzNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '24669385',
                                                                                                                                          'fiber_length': 18274,
                                                                                                                                          'poller_hash': 'NDU5NjVEcGVlcnBvbnBvbiAzNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '22364992',
                                                                                                                                          'fiber_length': 18443,
                                                                                                                                          'poller_hash': 'MkVEOUNEcGVlcnBvbnBvbiAzNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '47663',
                                                                                                                                          'fiber_length': 18130,
                                                                                                                                          'poller_hash': 'NzIzM0VBcGVlcnBvbnBvbiAzNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '3113871',
                                                                                                                                          'fiber_length': 18440,
                                                                                                                                          'poller_hash': 'QjBBMkE2cGVlcnBvbnBvbiAxNTAwMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '21815711',
                                                                                                                                          'fiber_length': 18555,
                                                                                                                                          'poller_hash': 'QzgzQzIxcGVlcnBvbnBvbiAxNzIwNTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '2969268',
                                                                                                                                          'fiber_length': 17809,
                                                                                                                                          'poller_hash': 'QzgzMTYzcGVlcnBvbnBvbiAxNzI0NDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '70945612',
                                                                                                                                          'fiber_length': 18606,
                                                                                                                                          'poller_hash': 'QzgxRDU0cGVlcnBvbnBvbiAyMDYzODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '12635297',
                                                                                                                                          'fiber_length': 18502,
                                                                                                                                          'poller_hash': 'MTI3NEE0NHBlZXJwb25wb24gMjM2MjA2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '70093395',
                                                                                                                                          'fiber_length': 18525,
                                                                                                                                          'poller_hash': 'MTI3NDE5QXBlZXJwb25wb24gMjQzNzUx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '2969293',
                                                                                                                                          'fiber_length': 17885,
                                                                                                                                          'poller_hash': 'MTE3REJCRXBlZXJwb25wb24gMjQ5MjMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766},
                                                                                                                                  {       'connected_time': '29355887',
                                                                                                                                          'fiber_length': 17714,
                                                                                                                                          'poller_hash': 'MTE5NTE3M3BlZXJwb25wb24gMjUxNTI5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.766}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/3',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/4',
                                                                                         'in_discards': '1159',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5475129',
                                                                                         'in_octets': '18780772731985',
                                                                                         'in_ucast_pkts': '86532010542',
                                                                                         'links': [       {       'in_nucast_pkts': '1186',
                                                                                                                  'in_octets': '462326249847',
                                                                                                                  'in_ucast_pkts': '2693871334',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45027725',
                                                                                                                  'out_octets': '15715191799675',
                                                                                                                  'out_ucast_pkts': '10819452001',
                                                                                                                  'poller_hash': 'NTU2M0RGcGVlcmludGZwb24gMzc2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '234838016077',
                                                                                                                  'in_ucast_pkts': '2067772958',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '8603056441117',
                                                                                                                  'out_ucast_pkts': '7478821959',
                                                                                                                  'poller_hash': 'NzZCMDY0cGVlcmludGZwb24gMzc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1174374969783',
                                                                                                                  'in_ucast_pkts': '11784527404',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '31286379783879',
                                                                                                                  'out_ucast_pkts': '21884124056',
                                                                                                                  'poller_hash': 'NTU2MzkzcGVlcmludGZwb24gMzc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '133404008843',
                                                                                                                  'in_ucast_pkts': '1094710636',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6068813948816',
                                                                                                                  'out_ucast_pkts': '5836258020',
                                                                                                                  'poller_hash': 'MzI5Njc4cGVlcmludGZwb24gMzc5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '852586618',
                                                                                                                  'in_ucast_pkts': '3480891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '43432029999',
                                                                                                                  'out_ucast_pkts': '156388786',
                                                                                                                  'poller_hash': 'NEQxMTUycGVlcmludGZwb24gMzgw'},
                                                                                                          {       'in_nucast_pkts': '207599',
                                                                                                                  'in_octets': '488111459179',
                                                                                                                  'in_ucast_pkts': '1024832614',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45024923',
                                                                                                                  'out_octets': '4779245969763',
                                                                                                                  'out_ucast_pkts': '3509482053',
                                                                                                                  'poller_hash': 'N0UyMTBwZWVyaW50ZnBvbiAzODE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '119683100791',
                                                                                                                  'in_ucast_pkts': '1622783441',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4741999181743',
                                                                                                                  'out_ucast_pkts': '4834838478',
                                                                                                                  'poller_hash': 'MkZFOUIycGVlcmludGZwb24gMzgy'},
                                                                                                          {       'in_nucast_pkts': '82',
                                                                                                                  'in_octets': '402011816437',
                                                                                                                  'in_ucast_pkts': '1483299727',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '42437506',
                                                                                                                  'out_octets': '7890544400939',
                                                                                                                  'out_ucast_pkts': '5765802470',
                                                                                                                  'poller_hash': 'NTYxMTYycGVlcmludGZwb24gMzgz'},
                                                                                                          {       'in_nucast_pkts': '490057',
                                                                                                                  'in_octets': '96097125797',
                                                                                                                  'in_ucast_pkts': '230036022',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45021491',
                                                                                                                  'out_octets': '1720866987281',
                                                                                                                  'out_ucast_pkts': '1315728594',
                                                                                                                  'poller_hash': 'NzY5QzFDcGVlcmludGZwb24gMzg2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '265433967156',
                                                                                                                  'in_ucast_pkts': '1640561454',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4177465501144',
                                                                                                                  'out_ucast_pkts': '220662503',
                                                                                                                  'poller_hash': 'MzkwREM4cGVlcmludGZwb24gMzg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '526897125431',
                                                                                                                  'in_ucast_pkts': '2903329278',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6603814556288',
                                                                                                                  'out_ucast_pkts': '6856177080',
                                                                                                                  'poller_hash': 'NEQ0MjRFcGVlcmludGZwb24gMzg4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1258633370281',
                                                                                                                  'in_ucast_pkts': '143827289838',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '18874275111781',
                                                                                                                  'out_ucast_pkts': '13364166437',
                                                                                                                  'poller_hash': 'Q0JEMUQ2cGVlcmludGZwb24gMzkw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '93937902454',
                                                                                                                  'in_ucast_pkts': '625206157',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2270820101997',
                                                                                                                  'out_ucast_pkts': '1732276645',
                                                                                                                  'poller_hash': 'MzY2QzY3cGVlcmludGZwb24gMzkx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '286595945395',
                                                                                                                  'in_ucast_pkts': '2489959144',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '8650200131628',
                                                                                                                  'out_ucast_pkts': '6319898873',
                                                                                                                  'poller_hash': 'NTU2M0Q0cGVlcmludGZwb24gMzky'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '584835506384',
                                                                                                                  'in_ucast_pkts': '3216269940',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9343156908344',
                                                                                                                  'out_ucast_pkts': '8947268880',
                                                                                                                  'poller_hash': 'NEQ0MjRDcGVlcmludGZwb24gMzkz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '881529182283',
                                                                                                                  'in_ucast_pkts': '7346179017',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '33319642349566',
                                                                                                                  'out_ucast_pkts': '22232101578',
                                                                                                                  'poller_hash': 'NEQ0MjQ0cGVlcmludGZwb24gMzk2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '449830905952',
                                                                                                                  'in_ucast_pkts': '2299122214',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7771160290976',
                                                                                                                  'out_ucast_pkts': '7662805705',
                                                                                                                  'poller_hash': 'MzI5NjY3cGVlcmludGZwb24gMzk3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '485327153805',
                                                                                                                  'in_ucast_pkts': '4215084062',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '16108082844035',
                                                                                                                  'out_ucast_pkts': '17300870669',
                                                                                                                  'poller_hash': 'NEQ0MjNCcGVlcmludGZwb24gMzk4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '129397720911',
                                                                                                                  'in_ucast_pkts': '682419118',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1981870013459',
                                                                                                                  'out_ucast_pkts': '1731128645',
                                                                                                                  'poller_hash': 'NEQ0MUYycGVlcmludGZwb24gNDAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '5304648034973',
                                                                                                                  'in_ucast_pkts': '12061084943',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '27308157816730',
                                                                                                                  'out_ucast_pkts': '21981624298',
                                                                                                                  'poller_hash': 'NTYwMjk2cGVlcmludGZwb24gNDAx'},
                                                                                                          {       'in_nucast_pkts': '27',
                                                                                                                  'in_octets': '148685765838',
                                                                                                                  'in_ucast_pkts': '838426208',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6685784',
                                                                                                                  'out_octets': '2650478425474',
                                                                                                                  'out_ucast_pkts': '1888023483',
                                                                                                                  'poller_hash': 'REZGRTQycGVlcmludGZwb24gNDAz'},
                                                                                                          {       'in_nucast_pkts': '14000117',
                                                                                                                  'in_octets': '263173175927',
                                                                                                                  'in_ucast_pkts': '1464019983',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33120464598',
                                                                                                                  'out_octets': '50999486110972',
                                                                                                                  'out_ucast_pkts': '4890112756',
                                                                                                                  'poller_hash': 'NDI4MUU5cGVlcmludGZwb24gNTI3'},
                                                                                                          {       'in_nucast_pkts': '323997',
                                                                                                                  'in_octets': '5584177782',
                                                                                                                  'in_ucast_pkts': '74857765',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23687419',
                                                                                                                  'out_octets': '304912172831',
                                                                                                                  'out_ucast_pkts': '210412835',
                                                                                                                  'poller_hash': 'NDI4MUJGcGVlcmludGZwb24gNTM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '30230234256',
                                                                                                                  'in_ucast_pkts': '93385994',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '768967764029',
                                                                                                                  'out_ucast_pkts': '569772484',
                                                                                                                  'poller_hash': 'QTlDM0JCcGVlcmludGZwb24gMTQ2MjY2'},
                                                                                                          {       'in_nucast_pkts': '24',
                                                                                                                  'in_octets': '17364541394',
                                                                                                                  'in_ucast_pkts': '44401283',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6686022',
                                                                                                                  'out_octets': '533853327953',
                                                                                                                  'out_ucast_pkts': '378548363',
                                                                                                                  'poller_hash': 'QjUxMUI3cGVlcmludGZwb24gMTU0ODk5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '583799779563',
                                                                                                                  'in_ucast_pkts': '6890372824',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '20595726171093',
                                                                                                                  'out_ucast_pkts': '13733772363',
                                                                                                                  'poller_hash': 'QzgzQzFEcGVlcmludGZwb24gMTk3Nzk3'},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '41321523124',
                                                                                                                  'in_ucast_pkts': '215255015',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1494777',
                                                                                                                  'out_octets': '877683639817',
                                                                                                                  'out_ucast_pkts': '624034799',
                                                                                                                  'poller_hash': 'MTE5NTBGQXBlZXJpbnRmcG9uIDIxNjA3Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '30837273110',
                                                                                                                  'in_ucast_pkts': '180501386',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '963211212574',
                                                                                                                  'out_ucast_pkts': '673052137',
                                                                                                                  'poller_hash': 'MTE3REMwM3BlZXJpbnRmcG9uIDI1NDA1Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '50932034960',
                                                                                                                  'in_ucast_pkts': '205380017',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '473345050841',
                                                                                                                  'out_ucast_pkts': '369585268',
                                                                                                                  'poller_hash': 'MTI5NjA0MXBlZXJpbnRmcG9uIDI3NzQyNw=='},
                                                                                                          {       'in_nucast_pkts': '41975',
                                                                                                                  'in_octets': '2488974953',
                                                                                                                  'in_ucast_pkts': '3272670',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4669041',
                                                                                                                  'out_octets': '11392831063',
                                                                                                                  'out_ucast_pkts': '8859791',
                                                                                                                  'poller_hash': 'MTI5NjBEQnBlZXJpbnRmcG9uIDI4MzUxNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '191337033710',
                                                                                         'out_octets': '612711447289876',
                                                                                         'out_ucast_pkts': '250140423057',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/4',
                                                                                                                 'id': '5/4',
                                                                                                                 'links': [       {       'connected_time': '24668969',
                                                                                                                                          'fiber_length': 17810,
                                                                                                                                          'poller_hash': 'NTU2M0RGcGVlcnBvbnBvbiAzNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '9663884',
                                                                                                                                          'fiber_length': 17717,
                                                                                                                                          'poller_hash': 'NzZCMDY0cGVlcnBvbnBvbiAzNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47705',
                                                                                                                                          'fiber_length': 18447,
                                                                                                                                          'poller_hash': 'NTU2MzkzcGVlcnBvbnBvbiAzNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47588',
                                                                                                                                          'fiber_length': 18118,
                                                                                                                                          'poller_hash': 'MzI5Njc4cGVlcnBvbnBvbiAzNzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47752',
                                                                                                                                          'fiber_length': 18180,
                                                                                                                                          'poller_hash': 'NEQxMTUycGVlcnBvbnBvbiAzODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '24669065',
                                                                                                                                          'fiber_length': 18631,
                                                                                                                                          'poller_hash': 'N0UyMTBwZWVycG9ucG9uIDM4MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47863',
                                                                                                                                          'fiber_length': 18399,
                                                                                                                                          'poller_hash': 'MkZFOUIycGVlcnBvbnBvbiAzODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '23325118',
                                                                                                                                          'fiber_length': 17454,
                                                                                                                                          'poller_hash': 'NTYxMTYycGVlcnBvbnBvbiAzODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '24669156',
                                                                                                                                          'fiber_length': 18666,
                                                                                                                                          'poller_hash': 'NzY5QzFDcGVlcnBvbnBvbiAzODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47513',
                                                                                                                                          'fiber_length': 18545,
                                                                                                                                          'poller_hash': 'MzkwREM4cGVlcnBvbnBvbiAzODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '73292557',
                                                                                                                                          'fiber_length': 18280,
                                                                                                                                          'poller_hash': 'NEQ0MjRFcGVlcnBvbnBvbiAzODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '3113848',
                                                                                                                                          'fiber_length': 18096,
                                                                                                                                          'poller_hash': 'Q0JEMUQ2cGVlcnBvbnBvbiAzOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47909',
                                                                                                                                          'fiber_length': 18163,
                                                                                                                                          'poller_hash': 'MzY2QzY3cGVlcnBvbnBvbiAzOTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '10038949',
                                                                                                                                          'fiber_length': 17996,
                                                                                                                                          'poller_hash': 'NTU2M0Q0cGVlcnBvbnBvbiAzOTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '11345518',
                                                                                                                                          'fiber_length': 17538,
                                                                                                                                          'poller_hash': 'NEQ0MjRDcGVlcnBvbnBvbiAzOTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47644',
                                                                                                                                          'fiber_length': 17842,
                                                                                                                                          'poller_hash': 'NEQ0MjQ0cGVlcnBvbnBvbiAzOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47916',
                                                                                                                                          'poller_hash': 'MzI5NjY3cGVlcnBvbnBvbiAzOTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '47674',
                                                                                                                                          'fiber_length': 18214,
                                                                                                                                          'poller_hash': 'NEQ0MjNCcGVlcnBvbnBvbiAzOTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '29355792',
                                                                                                                                          'fiber_length': 17645,
                                                                                                                                          'poller_hash': 'NEQ0MUYycGVlcnBvbnBvbiA0MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '13127788',
                                                                                                                                          'fiber_length': 17621,
                                                                                                                                          'poller_hash': 'NTYwMjk2cGVlcnBvbnBvbiA0MDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '3113867',
                                                                                                                                          'fiber_length': 18366,
                                                                                                                                          'poller_hash': 'REZGRTQycGVlcnBvbnBvbiA0MDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '21815930',
                                                                                                                                          'fiber_length': 18546,
                                                                                                                                          'poller_hash': 'NDI4MUU5cGVlcnBvbnBvbiA1Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '1133960',
                                                                                                                                          'fiber_length': 18858,
                                                                                                                                          'poller_hash': 'NDI4MUJGcGVlcnBvbnBvbiA1MzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '3113872',
                                                                                                                                          'fiber_length': 17876,
                                                                                                                                          'poller_hash': 'QTlDM0JCcGVlcnBvbnBvbiAxNDYyNjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '3113963',
                                                                                                                                          'fiber_length': 18104,
                                                                                                                                          'poller_hash': 'QjUxMUI3cGVlcnBvbnBvbiAxNTQ4OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '21816163',
                                                                                                                                          'fiber_length': 18079,
                                                                                                                                          'poller_hash': 'QzgzQzFEcGVlcnBvbnBvbiAxOTc3OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '2969301',
                                                                                                                                          'fiber_length': 17588,
                                                                                                                                          'poller_hash': 'MTE5NTBGQXBlZXJwb25wb24gMjE2MDcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '1803250',
                                                                                                                                          'fiber_length': 17707,
                                                                                                                                          'poller_hash': 'MTE3REMwM3BlZXJwb25wb24gMjU0MDUy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '995949',
                                                                                                                                          'fiber_length': 18580,
                                                                                                                                          'poller_hash': 'MTI5NjA0MXBlZXJwb25wb24gMjc3NDI3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963},
                                                                                                                                  {       'connected_time': '2180897',
                                                                                                                                          'fiber_length': 18231,
                                                                                                                                          'poller_hash': 'MTI5NjBEQnBlZXJwb25wb24gMjgzNTE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.963}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/4',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/5',
                                                                                         'in_discards': '1169',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3473012',
                                                                                         'in_octets': '9810089205704',
                                                                                         'in_ucast_pkts': '45409030736',
                                                                                         'links': [       {       'in_nucast_pkts': '323172',
                                                                                                                  'in_octets': '5966862073',
                                                                                                                  'in_ucast_pkts': '25764544',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1411166',
                                                                                                                  'out_octets': '90635139092',
                                                                                                                  'out_ucast_pkts': '63651365',
                                                                                                                  'poller_hash': 'MzAwRTg4cGVlcmludGZwb24gMzYz'},
                                                                                                          {       'in_nucast_pkts': '385769',
                                                                                                                  'in_octets': '220515250412',
                                                                                                                  'in_ucast_pkts': '1574094457',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70637500',
                                                                                                                  'out_octets': '10337338268758',
                                                                                                                  'out_ucast_pkts': '7095751503',
                                                                                                                  'poller_hash': 'MUZENkRCcGVlcmludGZwb24gNDU3'},
                                                                                                          {       'in_nucast_pkts': '5547',
                                                                                                                  'in_octets': '16406546305',
                                                                                                                  'in_ucast_pkts': '120827565',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6671569',
                                                                                                                  'out_octets': '755537483562',
                                                                                                                  'out_ucast_pkts': '530705469',
                                                                                                                  'poller_hash': 'MTM4MEQ0cGVlcmludGZwb24gNDU4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '504224062963',
                                                                                                                  'in_ucast_pkts': '4216154172',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11265954596495',
                                                                                                                  'out_ucast_pkts': '5326064691',
                                                                                                                  'poller_hash': 'Q0MwQ0NwZWVyaW50ZnBvbiA0NTk='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '150462725727',
                                                                                                                  'in_ucast_pkts': '924222813',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3143938246612',
                                                                                                                  'out_ucast_pkts': '2287905258',
                                                                                                                  'poller_hash': 'Q0MwN0ZwZWVyaW50ZnBvbiA0NjA='},
                                                                                                          {       'in_nucast_pkts': '10566',
                                                                                                                  'in_octets': '85934469068',
                                                                                                                  'in_ucast_pkts': '414153171',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '143710625',
                                                                                                                  'out_octets': '1537112382041',
                                                                                                                  'out_ucast_pkts': '1144513557',
                                                                                                                  'poller_hash': 'MkIwNDA1cGVlcmludGZwb24gNDY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '157374702913',
                                                                                                                  'in_ucast_pkts': '707686655',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2599461724285',
                                                                                                                  'out_ucast_pkts': '2112960881',
                                                                                                                  'poller_hash': 'Q0MwQzJwZWVyaW50ZnBvbiA0ODU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzVGODQycGVlcmludGZwb24gNDk2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '877962591672',
                                                                                                                  'in_ucast_pkts': '2259933767',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '7737368515707',
                                                                                                                  'out_ucast_pkts': '6013743906',
                                                                                                                  'poller_hash': 'MzIxMEU0cGVlcmludGZwb24gNTEy'},
                                                                                                          {       'in_nucast_pkts': '1430991',
                                                                                                                  'in_octets': '3764256983929',
                                                                                                                  'in_ucast_pkts': '10741386854',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '224530418',
                                                                                                                  'out_octets': '54254153273571',
                                                                                                                  'out_ucast_pkts': '39229124958',
                                                                                                                  'poller_hash': 'MjZEOTkycGVlcmludGZwb24gNTUy'},
                                                                                                          {       'in_nucast_pkts': '10446',
                                                                                                                  'in_octets': '1654849605',
                                                                                                                  'in_ucast_pkts': '18965532',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1258919',
                                                                                                                  'out_octets': '127034887825',
                                                                                                                  'out_ucast_pkts': '84875518',
                                                                                                                  'poller_hash': 'MzIxODA4cGVlcmludGZwb24gNzE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '171131640662',
                                                                                                                  'in_ucast_pkts': '26476817380',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '3543876004186',
                                                                                                                  'out_ucast_pkts': '2620234533',
                                                                                                                  'poller_hash': 'QzgxRTc1cGVlcmludGZwb24gMjA1Mzg2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '237837834572',
                                                                                                                  'in_ucast_pkts': '1353685599',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '5425731575671',
                                                                                                                  'out_ucast_pkts': '4047028512',
                                                                                                                  'poller_hash': 'MTI3NTE1QXBlZXJpbnRmcG9uIDIxOTAzNQ=='},
                                                                                                          {       'in_nucast_pkts': '236933',
                                                                                                                  'in_octets': '2692104145',
                                                                                                                  'in_ucast_pkts': '7864267',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28772475',
                                                                                                                  'out_octets': '19760712839',
                                                                                                                  'out_ucast_pkts': '18660570',
                                                                                                                  'poller_hash': 'MTI3NDc4QXBlZXJpbnRmcG9uIDI3NDU0NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '53517534443',
                                                                                         'out_octets': '289505016380276',
                                                                                         'out_ucast_pkts': '154891485156',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/5',
                                                                                                                 'id': '5/5',
                                                                                                                 'links': [       {       'connected_time': '688395',
                                                                                                                                          'fiber_length': 1173,
                                                                                                                                          'poller_hash': 'MzAwRTg4cGVlcnBvbnBvbiAzNjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '9954635',
                                                                                                                                          'fiber_length': 1059,
                                                                                                                                          'poller_hash': 'MUZENkRCcGVlcnBvbnBvbiA0NTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '688356',
                                                                                                                                          'fiber_length': 1435,
                                                                                                                                          'poller_hash': 'MTM4MEQ0cGVlcnBvbnBvbiA0NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '8447159',
                                                                                                                                          'fiber_length': 1051,
                                                                                                                                          'poller_hash': 'Q0MwQ0NwZWVycG9ucG9uIDQ1OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '688384',
                                                                                                                                          'fiber_length': 1005,
                                                                                                                                          'poller_hash': 'Q0MwN0ZwZWVycG9ucG9uIDQ2MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '21815874',
                                                                                                                                          'fiber_length': 1254,
                                                                                                                                          'poller_hash': 'MkIwNDA1cGVlcnBvbnBvbiA0NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '47302',
                                                                                                                                          'fiber_length': 1213,
                                                                                                                                          'poller_hash': 'Q0MwQzJwZWVycG9ucG9uIDQ4NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '6668559',
                                                                                                                                          'fiber_length': 1168,
                                                                                                                                          'poller_hash': 'MzVGODQycGVlcnBvbnBvbiA0OTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '9776280',
                                                                                                                                          'fiber_length': 1483,
                                                                                                                                          'poller_hash': 'MzIxMEU0cGVlcnBvbnBvbiA1MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '11454477',
                                                                                                                                          'fiber_length': 1429,
                                                                                                                                          'poller_hash': 'MjZEOTkycGVlcnBvbnBvbiA1NTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '597547',
                                                                                                                                          'fiber_length': 1287,
                                                                                                                                          'poller_hash': 'MzIxODA4cGVlcnBvbnBvbiA3MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '688404',
                                                                                                                                          'fiber_length': 1272,
                                                                                                                                          'poller_hash': 'QzgxRTc1cGVlcnBvbnBvbiAyMDUzODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '6053404',
                                                                                                                                          'fiber_length': 1381,
                                                                                                                                          'poller_hash': 'MTI3NTE1QXBlZXJwb25wb24gMjE5MDM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592},
                                                                                                                                  {       'connected_time': '12804236',
                                                                                                                                          'fiber_length': 1332,
                                                                                                                                          'poller_hash': 'MTI3NDc4QXBlZXJwb25wb24gMjc0NTQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.592}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/5',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/6',
                                                                                         'in_discards': '4051',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1786226',
                                                                                         'in_octets': '9333188493671',
                                                                                         'in_ucast_pkts': '38804395451',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '529436055402',
                                                                                                                  'in_ucast_pkts': '117191576259',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '2891271805886',
                                                                                                                  'out_ucast_pkts': '2375790377',
                                                                                                                  'poller_hash': 'REZGRTBDcGVlcmludGZwb24gMTg1NzU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'REZGREQ3cGVlcmludGZwb24gMTg1NzY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '196762035158',
                                                                                                                  'in_ucast_pkts': '1878228492',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18403095043',
                                                                                                                  'out_octets': '30482924240108',
                                                                                                                  'out_ucast_pkts': '2402135',
                                                                                                                  'poller_hash': 'QzgzMEY1cGVlcmludGZwb24gMTg2NDk3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '788419896722',
                                                                                                                  'in_ucast_pkts': '61358028858',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '6438467397674',
                                                                                                                  'out_ucast_pkts': '4717625290',
                                                                                                                  'poller_hash': 'QzgzMEJBcGVlcmludGZwb24gMTg2NTI3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '322759068524',
                                                                                                                  'in_ucast_pkts': '40026424316',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '9757423755782',
                                                                                                                  'out_ucast_pkts': '9425505450',
                                                                                                                  'poller_hash': 'REZGRTJEcGVlcmludGZwb24gMTg2NTM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '94709421877',
                                                                                                                  'in_ucast_pkts': '307832904',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '1182748753006',
                                                                                                                  'out_ucast_pkts': '907873985',
                                                                                                                  'poller_hash': 'REZGREUwcGVlcmludGZwb24gMTg5NzY3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '924243346719',
                                                                                                                  'in_ucast_pkts': '117853600048',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62',
                                                                                                                  'out_octets': '5005860080866',
                                                                                                                  'out_ucast_pkts': '3895680490',
                                                                                                                  'poller_hash': 'QzgzQkM2cGVlcmludGZwb24gMTg5ODM5'},
                                                                                                          {       'in_nucast_pkts': '403514',
                                                                                                                  'in_octets': '156093661106',
                                                                                                                  'in_ucast_pkts': '1700312524',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '68247885',
                                                                                                                  'out_octets': '8924268409935',
                                                                                                                  'out_ucast_pkts': '5971284958',
                                                                                                                  'poller_hash': 'QzgzMENFcGVlcmludGZwb24gMTkzMzI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '34452348357',
                                                                                                                  'in_ucast_pkts': '133425433',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '807094366415',
                                                                                                                  'out_ucast_pkts': '688404605',
                                                                                                                  'poller_hash': 'QzgxRTJFcGVlcmludGZwb24gMTk4NTM5'},
                                                                                                          {       'in_nucast_pkts': '972300',
                                                                                                                  'in_octets': '808432611516',
                                                                                                                  'in_ucast_pkts': '2135048435',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '143674530',
                                                                                                                  'out_octets': '13988603012708',
                                                                                                                  'out_ucast_pkts': '10327520971',
                                                                                                                  'poller_hash': 'QzgxREYwcGVlcmludGZwb24gMjAwMzQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxOUJBcGVlcmludGZwb24gMjAwMzY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '198663368952',
                                                                                                                  'in_ucast_pkts': '1449620062',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '5069877918952',
                                                                                                                  'out_ucast_pkts': '4335687542',
                                                                                                                  'poller_hash': 'QzgxRUI4cGVlcmludGZwb24gMjAxMTE2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '191160664901',
                                                                                                                  'in_ucast_pkts': '14942095179',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '6720701526804',
                                                                                                                  'out_ucast_pkts': '4580253189',
                                                                                                                  'poller_hash': 'QzgxRTlGcGVlcmludGZwb24gMjAyODY4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '30345079441',
                                                                                                                  'in_ucast_pkts': '113984043',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '758894060512',
                                                                                                                  'out_ucast_pkts': '612714245',
                                                                                                                  'poller_hash': 'QzgxRTVGcGVlcmludGZwb24gMjA5MTU0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '459385553050',
                                                                                                                  'in_ucast_pkts': '15123969494',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '8323660077760',
                                                                                                                  'out_ucast_pkts': '5845094921',
                                                                                                                  'poller_hash': 'QzgxRTNFcGVlcmludGZwb24gMjA5NTU4'},
                                                                                                          {       'in_nucast_pkts': '538038',
                                                                                                                  'in_octets': '21412155682',
                                                                                                                  'in_ucast_pkts': '138812977',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '67889855',
                                                                                                                  'out_octets': '663764566871',
                                                                                                                  'out_ucast_pkts': '460461849',
                                                                                                                  'poller_hash': 'MTFCMjJCQnBlZXJpbnRmcG9uIDIxMjUwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '181851289884',
                                                                                                                  'in_ucast_pkts': '44201394582',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '6689982025236',
                                                                                                                  'out_ucast_pkts': '13201225051',
                                                                                                                  'poller_hash': 'MTE5NTEzRXBlZXJpbnRmcG9uIDIxMjYwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '691411647',
                                                                                                                  'in_ucast_pkts': '734229',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '155',
                                                                                                                  'out_octets': '12722487299',
                                                                                                                  'out_ucast_pkts': '54612531',
                                                                                                                  'poller_hash': 'MTE5NTE4RHBlZXJpbnRmcG9uIDIxMjc5MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '513613571027',
                                                                                                                  'in_ucast_pkts': '80195864982',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '186',
                                                                                                                  'out_octets': '13042241546499',
                                                                                                                  'out_ucast_pkts': '9883573946',
                                                                                                                  'poller_hash': 'MTE5NTE2RnBlZXJpbnRmcG9uIDIxMjkzNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '131967895340',
                                                                                                                  'in_ucast_pkts': '596119839',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2316722914974',
                                                                                                                  'out_ucast_pkts': '1663523147',
                                                                                                                  'poller_hash': 'MTI3NDlGQ3BlZXJpbnRmcG9uIDI0MTYyNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NjBFMHBlZXJpbnRmcG9uIDI4NDQ4OA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '5/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '92067491679',
                                                                                         'out_octets': '278336604090740',
                                                                                         'out_ucast_pkts': '112721812165',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/6',
                                                                                                                 'id': '5/6',
                                                                                                                 'links': [       {       'connected_time': '523134',
                                                                                                                                          'fiber_length': 11330,
                                                                                                                                          'poller_hash': 'REZGRTBDcGVlcnBvbnBvbiAxODU3NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '21816827',
                                                                                                                                          'fiber_length': 11229,
                                                                                                                                          'poller_hash': 'REZGREQ3cGVlcnBvbnBvbiAxODU3NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '1024524',
                                                                                                                                          'fiber_length': 11179,
                                                                                                                                          'poller_hash': 'QzgzMEY1cGVlcnBvbnBvbiAxODY0OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '4481109',
                                                                                                                                          'fiber_length': 11244,
                                                                                                                                          'poller_hash': 'QzgzMEJBcGVlcnBvbnBvbiAxODY1Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '30000762',
                                                                                                                                          'fiber_length': 11303,
                                                                                                                                          'poller_hash': 'REZGRTJEcGVlcnBvbnBvbiAxODY1Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '523138',
                                                                                                                                          'fiber_length': 11235,
                                                                                                                                          'poller_hash': 'REZGREUwcGVlcnBvbnBvbiAxODk3Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '688400',
                                                                                                                                          'fiber_length': 11286,
                                                                                                                                          'poller_hash': 'QzgzQkM2cGVlcnBvbnBvbiAxODk4Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '21815769',
                                                                                                                                          'fiber_length': 11293,
                                                                                                                                          'poller_hash': 'QzgzMENFcGVlcnBvbnBvbiAxOTMzMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '1981829',
                                                                                                                                          'fiber_length': 10817,
                                                                                                                                          'poller_hash': 'QzgxRTJFcGVlcnBvbnBvbiAxOTg1Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '21816495',
                                                                                                                                          'fiber_length': 10787,
                                                                                                                                          'poller_hash': 'QzgxREYwcGVlcnBvbnBvbiAyMDAzNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '24124661',
                                                                                                                                          'fiber_length': 11577,
                                                                                                                                          'poller_hash': 'QzgxOUJBcGVlcnBvbnBvbiAyMDAzNjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '72476750',
                                                                                                                                          'fiber_length': 10851,
                                                                                                                                          'poller_hash': 'QzgxRUI4cGVlcnBvbnBvbiAyMDExMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '6821488',
                                                                                                                                          'fiber_length': 11334,
                                                                                                                                          'poller_hash': 'QzgxRTlGcGVlcnBvbnBvbiAyMDI4Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '688409',
                                                                                                                                          'fiber_length': 11752,
                                                                                                                                          'poller_hash': 'QzgxRTVGcGVlcnBvbnBvbiAyMDkxNTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '688409',
                                                                                                                                          'fiber_length': 11325,
                                                                                                                                          'poller_hash': 'QzgxRTNFcGVlcnBvbnBvbiAyMDk1NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '30265335',
                                                                                                                                          'fiber_length': 11741,
                                                                                                                                          'poller_hash': 'MTFCMjJCQnBlZXJwb25wb24gMjEyNTAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '26632408',
                                                                                                                                          'fiber_length': 11334,
                                                                                                                                          'poller_hash': 'MTE5NTEzRXBlZXJwb25wb24gMjEyNjAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '21814690',
                                                                                                                                          'fiber_length': 11253,
                                                                                                                                          'poller_hash': 'MTE5NTE4RHBlZXJwb25wb24gMjEyNzkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '2137841',
                                                                                                                                          'fiber_length': 10917,
                                                                                                                                          'poller_hash': 'MTE5NTE2RnBlZXJwb25wb24gMjEyOTM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '523140',
                                                                                                                                          'fiber_length': 10910,
                                                                                                                                          'poller_hash': 'MTI3NDlGQ3BlZXJwb25wb24gMjQxNjI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147},
                                                                                                                                  {       'connected_time': '1292753',
                                                                                                                                          'fiber_length': 11350,
                                                                                                                                          'poller_hash': 'MTI5NjBFMHBlZXJwb25wb24gMjg0NDg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.147}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/6',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/7',
                                                                                         'in_discards': '60173',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '9994625',
                                                                                         'in_octets': '41821294880719',
                                                                                         'in_ucast_pkts': '111812096981',
                                                                                         'links': [       {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '9720169780',
                                                                                                                  'in_ucast_pkts': '58775370',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1102920',
                                                                                                                  'out_octets': '180519612770',
                                                                                                                  'out_ucast_pkts': '129223887',
                                                                                                                  'poller_hash': 'NjQwQjUxcGVlcmludGZwb24gODA5'},
                                                                                                          {       'in_nucast_pkts': '12251',
                                                                                                                  'in_octets': '6246676163',
                                                                                                                  'in_ucast_pkts': '41940401',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1446993',
                                                                                                                  'out_octets': '158629741225',
                                                                                                                  'out_ucast_pkts': '111969864',
                                                                                                                  'poller_hash': 'NDJDNTIzcGVlcmludGZwb24gODE3'},
                                                                                                          {       'in_nucast_pkts': '12078',
                                                                                                                  'in_octets': '16800899252',
                                                                                                                  'in_ucast_pkts': '83925712',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447308',
                                                                                                                  'out_octets': '277696738939',
                                                                                                                  'out_ucast_pkts': '206294021',
                                                                                                                  'poller_hash': 'NDJEMEI0cGVlcmludGZwb24gODMw'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '3629193117',
                                                                                                                  'in_ucast_pkts': '14652912',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1102962',
                                                                                                                  'out_octets': '55662378169',
                                                                                                                  'out_ucast_pkts': '44702510',
                                                                                                                  'poller_hash': 'NjQwQjNCcGVlcmludGZwb24gODU2'},
                                                                                                          {       'in_nucast_pkts': '11083684',
                                                                                                                  'in_octets': '191509581099',
                                                                                                                  'in_ucast_pkts': '1106928612',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '13548691361',
                                                                                                                  'out_octets': '24411996329342',
                                                                                                                  'out_ucast_pkts': '4375920048',
                                                                                                                  'poller_hash': 'NDJDNUI0cGVlcmludGZwb24gODU4'},
                                                                                                          {       'in_nucast_pkts': '14985',
                                                                                                                  'in_octets': '21025590410',
                                                                                                                  'in_ucast_pkts': '86679321',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447131',
                                                                                                                  'out_octets': '336479783739',
                                                                                                                  'out_ucast_pkts': '251133105',
                                                                                                                  'poller_hash': 'ODY5NDhDcGVlcmludGZwb24gOTAw'},
                                                                                                          {       'in_nucast_pkts': '22',
                                                                                                                  'in_octets': '1958904336',
                                                                                                                  'in_ucast_pkts': '12158949',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1102962',
                                                                                                                  'out_octets': '77065451899',
                                                                                                                  'out_ucast_pkts': '56667490',
                                                                                                                  'poller_hash': 'NjQwQUZCcGVlcmludGZwb24gMTM4MTgy'},
                                                                                                          {       'in_nucast_pkts': '264',
                                                                                                                  'in_octets': '11514896789',
                                                                                                                  'in_ucast_pkts': '54254777',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1102861',
                                                                                                                  'out_octets': '349605044589',
                                                                                                                  'out_ucast_pkts': '247297419',
                                                                                                                  'poller_hash': 'NDNCNTA4cGVlcmludGZwb24gMTM4MzM3'},
                                                                                                          {       'in_nucast_pkts': '12084',
                                                                                                                  'in_octets': '9369041341',
                                                                                                                  'in_ucast_pkts': '59598331',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447136',
                                                                                                                  'out_octets': '262513427262',
                                                                                                                  'out_ucast_pkts': '191684175',
                                                                                                                  'poller_hash': 'ODY5M0U0cGVlcmludGZwb24gMTM4NjM1'},
                                                                                                          {       'in_nucast_pkts': '11884',
                                                                                                                  'in_octets': '8221531110',
                                                                                                                  'in_ucast_pkts': '47223162',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447198',
                                                                                                                  'out_octets': '177419928398',
                                                                                                                  'out_ucast_pkts': '140769157',
                                                                                                                  'poller_hash': 'OTU2Rjc1cGVlcmludGZwb24gMTM5MjUx'},
                                                                                                          {       'in_nucast_pkts': '10511077',
                                                                                                                  'in_octets': '578268176164',
                                                                                                                  'in_ucast_pkts': '3447530806',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8836630826',
                                                                                                                  'out_octets': '22484296741538',
                                                                                                                  'out_ucast_pkts': '7555288461',
                                                                                                                  'poller_hash': 'NzZEMDUycGVlcmludGZwb24gMTQwMDYz'},
                                                                                                          {       'in_nucast_pkts': '14887',
                                                                                                                  'in_octets': '23218845494',
                                                                                                                  'in_ucast_pkts': '104964966',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447150',
                                                                                                                  'out_octets': '441006268941',
                                                                                                                  'out_ucast_pkts': '331748005',
                                                                                                                  'poller_hash': 'ODY5MjI2cGVlcmludGZwb24gMTQwNDU2'},
                                                                                                          {       'in_nucast_pkts': '886',
                                                                                                                  'in_octets': '53216365613',
                                                                                                                  'in_ucast_pkts': '59561008',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447171',
                                                                                                                  'out_octets': '9757896597',
                                                                                                                  'out_ucast_pkts': '42950110',
                                                                                                                  'poller_hash': 'ODQ0QkVGcGVlcmludGZwb24gMTQwNDc3'},
                                                                                                          {       'in_nucast_pkts': '325111',
                                                                                                                  'in_octets': '265101847396',
                                                                                                                  'in_ucast_pkts': '646397994',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18123280846',
                                                                                                                  'out_octets': '26761091910677',
                                                                                                                  'out_ucast_pkts': '1839007473',
                                                                                                                  'poller_hash': 'ODY5NDI3cGVlcmludGZwb24gMTQxMDEx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '203959142262',
                                                                                                                  'in_ucast_pkts': '699148958',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '2435153906963',
                                                                                                                  'out_ucast_pkts': '1928935990',
                                                                                                                  'poller_hash': 'ODY5M0ZBcGVlcmludGZwb24gMTQxMTA5'},
                                                                                                          {       'in_nucast_pkts': '15118',
                                                                                                                  'in_octets': '12472542766',
                                                                                                                  'in_ucast_pkts': '72544749',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447037',
                                                                                                                  'out_octets': '244080712324',
                                                                                                                  'out_ucast_pkts': '172225836',
                                                                                                                  'poller_hash': 'ODJBNjk2cGVlcmludGZwb24gMTQxNDYy'},
                                                                                                          {       'in_nucast_pkts': '14757',
                                                                                                                  'in_octets': '13571597016',
                                                                                                                  'in_ucast_pkts': '95021304',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447007',
                                                                                                                  'out_octets': '330071842171',
                                                                                                                  'out_ucast_pkts': '236334535',
                                                                                                                  'poller_hash': 'ODY5NDMxcGVlcmludGZwb24gMTQxNzk3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1048087646888',
                                                                                                                  'in_ucast_pkts': '2329056527',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '7695560605400',
                                                                                                                  'out_ucast_pkts': '10136780011',
                                                                                                                  'poller_hash': 'QTlDNDBFcGVlcmludGZwb24gMTQ1OTUw'},
                                                                                                          {       'in_nucast_pkts': '71624',
                                                                                                                  'in_octets': '860876007175',
                                                                                                                  'in_ucast_pkts': '936606374',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15671454',
                                                                                                                  'out_octets': '2180756487137',
                                                                                                                  'out_ucast_pkts': '1827592773',
                                                                                                                  'poller_hash': 'QjUxMUUwcGVlcmludGZwb24gMTU0OTA3'},
                                                                                                          {       'in_nucast_pkts': '14975',
                                                                                                                  'in_octets': '121924392410',
                                                                                                                  'in_ucast_pkts': '148212154',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447016',
                                                                                                                  'out_octets': '282541765475',
                                                                                                                  'out_ucast_pkts': '242073614',
                                                                                                                  'poller_hash': 'QzgyMEFDcGVlcmludGZwb24gMTU2NjE4'},
                                                                                                          {       'in_nucast_pkts': '12756',
                                                                                                                  'in_octets': '53263234947',
                                                                                                                  'in_ucast_pkts': '260152553',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447239',
                                                                                                                  'out_octets': '832023675489',
                                                                                                                  'out_ucast_pkts': '636435103',
                                                                                                                  'poller_hash': 'Q0JEMjI0cGVlcmludGZwb24gMTY2MzM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '235587257834',
                                                                                                                  'in_ucast_pkts': '1017499773',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '4627585027429',
                                                                                                                  'out_ucast_pkts': '3593548408',
                                                                                                                  'poller_hash': 'QzgzQjMzcGVlcmludGZwb24gMTkzNTgy'},
                                                                                                          {       'in_nucast_pkts': '14868',
                                                                                                                  'in_octets': '18591604746',
                                                                                                                  'in_ucast_pkts': '141504030',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447496',
                                                                                                                  'out_octets': '749823463560',
                                                                                                                  'out_ucast_pkts': '516978312',
                                                                                                                  'poller_hash': 'QzgxRTU2cGVlcmludGZwb24gMjA2ODQ5'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '245594598693',
                                                                                         'out_octets': '695128055613973',
                                                                                         'out_ucast_pkts': '274867487276',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/7',
                                                                                                                 'id': '5/7',
                                                                                                                 'links': [       {       'connected_time': '688404',
                                                                                                                                          'fiber_length': 1982,
                                                                                                                                          'poller_hash': 'NjQwQjUxcGVlcnBvbnBvbiA4MDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688407',
                                                                                                                                          'fiber_length': 2054,
                                                                                                                                          'poller_hash': 'NDJDNTIzcGVlcnBvbnBvbiA4MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688436',
                                                                                                                                          'fiber_length': 2433,
                                                                                                                                          'poller_hash': 'NDJEMEI0cGVlcnBvbnBvbiA4MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688407',
                                                                                                                                          'fiber_length': 2096,
                                                                                                                                          'poller_hash': 'NjQwQjNCcGVlcnBvbnBvbiA4NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '71459163',
                                                                                                                                          'fiber_length': 2217,
                                                                                                                                          'poller_hash': 'NDJDNUI0cGVlcnBvbnBvbiA4NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688410',
                                                                                                                                          'fiber_length': 1997,
                                                                                                                                          'poller_hash': 'ODY5NDhDcGVlcnBvbnBvbiA5MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688406',
                                                                                                                                          'fiber_length': 1862,
                                                                                                                                          'poller_hash': 'NjQwQUZCcGVlcnBvbnBvbiAxMzgxODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688395',
                                                                                                                                          'fiber_length': 2081,
                                                                                                                                          'poller_hash': 'NDNCNTA4cGVlcnBvbnBvbiAxMzgzMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688411',
                                                                                                                                          'fiber_length': 2139,
                                                                                                                                          'poller_hash': 'ODY5M0U0cGVlcnBvbnBvbiAxMzg2MzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688419',
                                                                                                                                          'fiber_length': 1945,
                                                                                                                                          'poller_hash': 'OTU2Rjc1cGVlcnBvbnBvbiAxMzkyNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '71306417',
                                                                                                                                          'fiber_length': 1945,
                                                                                                                                          'poller_hash': 'NzZEMDUycGVlcnBvbnBvbiAxNDAwNjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688409',
                                                                                                                                          'fiber_length': 2403,
                                                                                                                                          'poller_hash': 'ODY5MjI2cGVlcnBvbnBvbiAxNDA0NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688432',
                                                                                                                                          'fiber_length': 2092,
                                                                                                                                          'poller_hash': 'ODQ0QkVGcGVlcnBvbnBvbiAxNDA0Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '15498560',
                                                                                                                                          'fiber_length': 2094,
                                                                                                                                          'poller_hash': 'ODY5NDI3cGVlcnBvbnBvbiAxNDEwMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688401',
                                                                                                                                          'fiber_length': 2094,
                                                                                                                                          'poller_hash': 'ODY5M0ZBcGVlcnBvbnBvbiAxNDExMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688427',
                                                                                                                                          'fiber_length': 2381,
                                                                                                                                          'poller_hash': 'ODJBNjk2cGVlcnBvbnBvbiAxNDE0NjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688409',
                                                                                                                                          'fiber_length': 1966,
                                                                                                                                          'poller_hash': 'ODY5NDMxcGVlcnBvbnBvbiAxNDE3OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688411',
                                                                                                                                          'fiber_length': 2396,
                                                                                                                                          'poller_hash': 'QTlDNDBFcGVlcnBvbnBvbiAxNDU5NTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '6818853',
                                                                                                                                          'fiber_length': 2382,
                                                                                                                                          'poller_hash': 'QjUxMUUwcGVlcnBvbnBvbiAxNTQ5MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688402',
                                                                                                                                          'fiber_length': 2206,
                                                                                                                                          'poller_hash': 'QzgyMEFDcGVlcnBvbnBvbiAxNTY2MTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688412',
                                                                                                                                          'fiber_length': 1970,
                                                                                                                                          'poller_hash': 'Q0JEMjI0cGVlcnBvbnBvbiAxNjYzMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '5885243',
                                                                                                                                          'fiber_length': 2526,
                                                                                                                                          'poller_hash': 'QzgzQjMzcGVlcnBvbnBvbiAxOTM1ODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021},
                                                                                                                                  {       'connected_time': '688433',
                                                                                                                                          'fiber_length': 2427,
                                                                                                                                          'poller_hash': 'QzgxRTU2cGVlcnBvbnBvbiAyMDY4NDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.021}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/7',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/8',
                                                                                         'in_discards': '104742',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '13702539',
                                                                                         'in_octets': '26577992755783',
                                                                                         'in_ucast_pkts': '99224976276',
                                                                                         'links': [       {       'in_nucast_pkts': '12217',
                                                                                                                  'in_octets': '15136951266',
                                                                                                                  'in_ucast_pkts': '39640425',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1446854',
                                                                                                                  'out_octets': '170714713990',
                                                                                                                  'out_ucast_pkts': '132006254',
                                                                                                                  'poller_hash': 'MzIxNDNBcGVlcmludGZwb24gOA=='},
                                                                                                          {       'in_nucast_pkts': '537',
                                                                                                                  'in_octets': '35301706136',
                                                                                                                  'in_ucast_pkts': '113866372',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1102703',
                                                                                                                  'out_octets': '273273291585',
                                                                                                                  'out_ucast_pkts': '196758694',
                                                                                                                  'poller_hash': 'NDNDMUI4cGVlcmludGZwb24gMTY='},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '3426405944',
                                                                                                                  'in_ucast_pkts': '18833161',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1103082',
                                                                                                                  'out_octets': '70105152216',
                                                                                                                  'out_ucast_pkts': '50674190',
                                                                                                                  'poller_hash': 'NDNDMjRCcGVlcmludGZwb24gNzc='},
                                                                                                          {       'in_nucast_pkts': '14942',
                                                                                                                  'in_octets': '4652965064',
                                                                                                                  'in_ucast_pkts': '30343244',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447029',
                                                                                                                  'out_octets': '141417107829',
                                                                                                                  'out_ucast_pkts': '102103553',
                                                                                                                  'poller_hash': 'MzIyODFFcGVlcmludGZwb24gMTk1'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '4954682862',
                                                                                                                  'in_ucast_pkts': '13157805',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447074',
                                                                                                                  'out_octets': '51313001043',
                                                                                                                  'out_ucast_pkts': '41451707',
                                                                                                                  'poller_hash': 'MkIwNUJFcGVlcmludGZwb24gMzIy'},
                                                                                                          {       'in_nucast_pkts': '114911',
                                                                                                                  'in_octets': '6115642649',
                                                                                                                  'in_ucast_pkts': '22939358',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1100833',
                                                                                                                  'out_octets': '106528619350',
                                                                                                                  'out_ucast_pkts': '82520321',
                                                                                                                  'poller_hash': 'NEY5MUZGcGVlcmludGZwb24gNTE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1147385640305',
                                                                                                                  'in_ucast_pkts': '78083633948',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '22564136714527',
                                                                                                                  'out_ucast_pkts': '17162058102',
                                                                                                                  'poller_hash': 'NDI4MUQzcGVlcmludGZwb24gNTI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '290235109671',
                                                                                                                  'in_ucast_pkts': '1060773754',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62',
                                                                                                                  'out_octets': '5986896100211',
                                                                                                                  'out_ucast_pkts': '4348113754',
                                                                                                                  'poller_hash': 'MjA1REVEcGVlcmludGZwb24gNTQw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTEyNTBEcGVlcmludGZwb24gNTUx'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '633369016',
                                                                                                                  'in_ucast_pkts': '3690785',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447152',
                                                                                                                  'out_octets': '8534238957',
                                                                                                                  'out_ucast_pkts': '6917792',
                                                                                                                  'poller_hash': 'NDJEMkFFcGVlcmludGZwb24gNTU0'},
                                                                                                          {       'in_nucast_pkts': '160',
                                                                                                                  'in_octets': '637354174841',
                                                                                                                  'in_ucast_pkts': '3741039208',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '65719894',
                                                                                                                  'out_octets': '14404522781705',
                                                                                                                  'out_ucast_pkts': '10750397810',
                                                                                                                  'poller_hash': 'NTYxNzFDcGVlcmludGZwb24gNTY3'},
                                                                                                          {       'in_nucast_pkts': '15489',
                                                                                                                  'in_octets': '7544064331',
                                                                                                                  'in_ucast_pkts': '23645244',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447054',
                                                                                                                  'out_octets': '99996998869',
                                                                                                                  'out_ucast_pkts': '76853472',
                                                                                                                  'poller_hash': 'M0I2MURFcGVlcmludGZwb24gNjg5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzIxMERFcGVlcmludGZwb24gNzA0'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '18989522075',
                                                                                                                  'in_ucast_pkts': '43392524',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447083',
                                                                                                                  'out_octets': '266326628952',
                                                                                                                  'out_ucast_pkts': '201280082',
                                                                                                                  'poller_hash': 'MzIxOUJBcGVlcmludGZwb24gNzA3'},
                                                                                                          {       'in_nucast_pkts': '841138',
                                                                                                                  'in_octets': '1235421746844',
                                                                                                                  'in_ucast_pkts': '2573789681',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141487337',
                                                                                                                  'out_octets': '8069040338829',
                                                                                                                  'out_ucast_pkts': '6368832161',
                                                                                                                  'poller_hash': 'MzIyODI1cGVlcmludGZwb24gNzEx'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '7950428280',
                                                                                                                  'in_ucast_pkts': '66610030',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447218',
                                                                                                                  'out_octets': '208520518654',
                                                                                                                  'out_ucast_pkts': '142714498',
                                                                                                                  'poller_hash': 'NDJCOUQ3cGVlcmludGZwb24gNzIy'},
                                                                                                          {       'in_nucast_pkts': '11965',
                                                                                                                  'in_octets': '12566288977',
                                                                                                                  'in_ucast_pkts': '138653206',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447165',
                                                                                                                  'out_octets': '545874527277',
                                                                                                                  'out_ucast_pkts': '364554604',
                                                                                                                  'poller_hash': 'MzIxOEE1cGVlcmludGZwb24gNzMx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2026704984974',
                                                                                                                  'in_ucast_pkts': '5341146948',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '12094145591570',
                                                                                                                  'out_ucast_pkts': '13243070417',
                                                                                                                  'poller_hash': 'NTU2M0U5cGVlcmludGZwb24gNzM0'},
                                                                                                          {       'in_nucast_pkts': '33450',
                                                                                                                  'in_octets': '24485303282',
                                                                                                                  'in_ucast_pkts': '109350361',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1480856',
                                                                                                                  'out_octets': '421712553869',
                                                                                                                  'out_ucast_pkts': '320216834',
                                                                                                                  'poller_hash': 'MzAwRDlDcGVlcmludGZwb24gNzM5'},
                                                                                                          {       'in_nucast_pkts': '15',
                                                                                                                  'in_octets': '825980057',
                                                                                                                  'in_ucast_pkts': '6230982',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447195',
                                                                                                                  'out_octets': '20728232802',
                                                                                                                  'out_ucast_pkts': '15733400',
                                                                                                                  'poller_hash': 'MTI3NDBFQnBlZXJpbnRmcG9uIDc4OA=='},
                                                                                                          {       'in_nucast_pkts': '13592',
                                                                                                                  'in_octets': '9488352267',
                                                                                                                  'in_ucast_pkts': '63177291',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447090',
                                                                                                                  'out_octets': '224249597521',
                                                                                                                  'out_ucast_pkts': '166134257',
                                                                                                                  'poller_hash': 'NDJCRjFCcGVlcmludGZwb24gNzk5'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '2307601809',
                                                                                                                  'in_ucast_pkts': '16556334',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1102969',
                                                                                                                  'out_octets': '48315099442',
                                                                                                                  'out_ucast_pkts': '33947714',
                                                                                                                  'poller_hash': 'NjQwQjY3cGVlcmludGZwb24gODI4'},
                                                                                                          {       'in_nucast_pkts': '11851',
                                                                                                                  'in_octets': '528109164',
                                                                                                                  'in_ucast_pkts': '2670714',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447126',
                                                                                                                  'out_octets': '11206859508',
                                                                                                                  'out_ucast_pkts': '8724578',
                                                                                                                  'poller_hash': 'ODY5Mzc2cGVlcmludGZwb24gODQx'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '12438812648',
                                                                                                                  'in_ucast_pkts': '28322725',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447108',
                                                                                                                  'out_octets': '129795193071',
                                                                                                                  'out_ucast_pkts': '97250738',
                                                                                                                  'poller_hash': 'NDJCRjAzcGVlcmludGZwb24gODYz'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '2369014953',
                                                                                                                  'in_ucast_pkts': '10390361',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447215',
                                                                                                                  'out_octets': '63163760656',
                                                                                                                  'out_ucast_pkts': '46411837',
                                                                                                                  'poller_hash': 'ODY5M0ZEcGVlcmludGZwb24gMTM4Njk4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1277760028422',
                                                                                                                  'in_ucast_pkts': '4212119480',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '13750204723802',
                                                                                                                  'out_ucast_pkts': '10447209755',
                                                                                                                  'poller_hash': 'ODY5NDFBcGVlcmludGZwb24gMTQxMDMx'},
                                                                                                          {       'in_nucast_pkts': '14290',
                                                                                                                  'in_octets': '9772355441',
                                                                                                                  'in_ucast_pkts': '72986987',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447054',
                                                                                                                  'out_octets': '446250600385',
                                                                                                                  'out_ucast_pkts': '310587970',
                                                                                                                  'poller_hash': 'QzgzRTNBcGVlcmludGZwb24gMTU3Mjg0'},
                                                                                                          {       'in_nucast_pkts': '14832',
                                                                                                                  'in_octets': '20928440867',
                                                                                                                  'in_ucast_pkts': '78557893',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3028371824',
                                                                                                                  'out_octets': '4194992859360',
                                                                                                                  'out_ucast_pkts': '162792691',
                                                                                                                  'poller_hash': 'Q0JEMjIycGVlcmludGZwb24gMTY2NDA4'},
                                                                                                          {       'in_nucast_pkts': '12508',
                                                                                                                  'in_octets': '4071350087',
                                                                                                                  'in_ucast_pkts': '12222533',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447322',
                                                                                                                  'out_octets': '45757712199',
                                                                                                                  'out_ucast_pkts': '37642384',
                                                                                                                  'poller_hash': 'QzgxRTQwcGVlcmludGZwb24gMjEyMTk2'},
                                                                                                          {       'in_nucast_pkts': '13301',
                                                                                                                  'in_octets': '20412817777',
                                                                                                                  'in_ucast_pkts': '46510262',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1447207',
                                                                                                                  'out_octets': '219361557247',
                                                                                                                  'out_ucast_pkts': '163498481',
                                                                                                                  'poller_hash': 'QjUxMkI3cGVlcmludGZwb24gMjEyODY5'},
                                                                                                          {       'in_nucast_pkts': '14512',
                                                                                                                  'in_octets': '36232131643',
                                                                                                                  'in_ucast_pkts': '135294432',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1448053',
                                                                                                                  'out_octets': '559168172087',
                                                                                                                  'out_ucast_pkts': '409609394',
                                                                                                                  'poller_hash': 'MTI3MzQ3M3BlZXJpbnRmcG9uIDI0ODUyMw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '126449995489',
                                                                                         'out_octets': '560757680123900',
                                                                                         'out_ucast_pkts': '288986204690',
                                                                                         'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/8',
                                                                                                                 'id': '5/8',
                                                                                                                 'links': [       {       'connected_time': '688412',
                                                                                                                                          'fiber_length': 2222,
                                                                                                                                          'poller_hash': 'MzIxNDNBcGVlcnBvbnBvbiA4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688401',
                                                                                                                                          'fiber_length': 2113,
                                                                                                                                          'poller_hash': 'NDNDMUI4cGVlcnBvbnBvbiAxNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688396',
                                                                                                                                          'fiber_length': 2111,
                                                                                                                                          'poller_hash': 'NDNDMjRCcGVlcnBvbnBvbiA3Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688408',
                                                                                                                                          'fiber_length': 2107,
                                                                                                                                          'poller_hash': 'MzIyODFFcGVlcnBvbnBvbiAxOTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688406',
                                                                                                                                          'fiber_length': 2248,
                                                                                                                                          'poller_hash': 'MkIwNUJFcGVlcnBvbnBvbiAzMjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688176',
                                                                                                                                          'fiber_length': 2157,
                                                                                                                                          'poller_hash': 'NEY5MUZGcGVlcnBvbnBvbiA1MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688407',
                                                                                                                                          'fiber_length': 1955,
                                                                                                                                          'poller_hash': 'NDI4MUQzcGVlcnBvbnBvbiA1Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688413',
                                                                                                                                          'fiber_length': 2132,
                                                                                                                                          'poller_hash': 'MjA1REVEcGVlcnBvbnBvbiA1NDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '15499142',
                                                                                                                                          'fiber_length': 2089,
                                                                                                                                          'poller_hash': 'NTEyNTBEcGVlcnBvbnBvbiA1NTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688408',
                                                                                                                                          'fiber_length': 2148,
                                                                                                                                          'poller_hash': 'NDJEMkFFcGVlcnBvbnBvbiA1NTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '68237769',
                                                                                                                                          'fiber_length': 2120,
                                                                                                                                          'poller_hash': 'NTYxNzFDcGVlcnBvbnBvbiA1Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688411',
                                                                                                                                          'fiber_length': 2224,
                                                                                                                                          'poller_hash': 'M0I2MURFcGVlcnBvbnBvbiA2ODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688413',
                                                                                                                                          'fiber_length': 2188,
                                                                                                                                          'poller_hash': 'MzIxMERFcGVlcnBvbnBvbiA3MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688410',
                                                                                                                                          'fiber_length': 1983,
                                                                                                                                          'poller_hash': 'MzIxOUJBcGVlcnBvbnBvbiA3MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '20658378',
                                                                                                                                          'fiber_length': 1968,
                                                                                                                                          'poller_hash': 'MzIyODI1cGVlcnBvbnBvbiA3MTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688413',
                                                                                                                                          'fiber_length': 2227,
                                                                                                                                          'poller_hash': 'NDJCOUQ3cGVlcnBvbnBvbiA3MjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688420',
                                                                                                                                          'fiber_length': 1989,
                                                                                                                                          'poller_hash': 'MzIxOEE1cGVlcnBvbnBvbiA3MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '15769267',
                                                                                                                                          'fiber_length': 2094,
                                                                                                                                          'poller_hash': 'NTU2M0U5cGVlcnBvbnBvbiA3MzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688414',
                                                                                                                                          'fiber_length': 2138,
                                                                                                                                          'poller_hash': 'MzAwRDlDcGVlcnBvbnBvbiA3Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688411',
                                                                                                                                          'fiber_length': 2071,
                                                                                                                                          'poller_hash': 'MTI3NDBFQnBlZXJwb25wb24gNzg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688411',
                                                                                                                                          'fiber_length': 2144,
                                                                                                                                          'poller_hash': 'NDJCRjFCcGVlcnBvbnBvbiA3OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688406',
                                                                                                                                          'fiber_length': 1938,
                                                                                                                                          'poller_hash': 'NjQwQjY3cGVlcnBvbnBvbiA4Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688420',
                                                                                                                                          'fiber_length': 2119,
                                                                                                                                          'poller_hash': 'ODY5Mzc2cGVlcnBvbnBvbiA4NDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688408',
                                                                                                                                          'fiber_length': 2133,
                                                                                                                                          'poller_hash': 'NDJCRjAzcGVlcnBvbnBvbiA4NjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688418',
                                                                                                                                          'fiber_length': 2125,
                                                                                                                                          'poller_hash': 'ODY5M0ZEcGVlcnBvbnBvbiAxMzg2OTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688408',
                                                                                                                                          'fiber_length': 2243,
                                                                                                                                          'poller_hash': 'ODY5NDFBcGVlcnBvbnBvbiAxNDEwMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688415',
                                                                                                                                          'fiber_length': 2112,
                                                                                                                                          'poller_hash': 'QzgzRTNBcGVlcnBvbnBvbiAxNTcyODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688419',
                                                                                                                                          'fiber_length': 2006,
                                                                                                                                          'poller_hash': 'Q0JEMjIycGVlcnBvbnBvbiAxNjY0MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688409',
                                                                                                                                          'fiber_length': 1980,
                                                                                                                                          'poller_hash': 'QzgxRTQwcGVlcnBvbnBvbiAyMTIxOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688428',
                                                                                                                                          'fiber_length': 2230,
                                                                                                                                          'poller_hash': 'QjUxMkI3cGVlcnBvbnBvbiAyMTI4Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902},
                                                                                                                                  {       'connected_time': '688413',
                                                                                                                                          'fiber_length': 2096,
                                                                                                                                          'poller_hash': 'MTI3MzQ3M3BlZXJwb25wb24gMjQ4NTIz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.902}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/8',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNTEyNjAzNjYx',
                                                           'serial_number': '211512603661',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991832'},
                                                   {       'id': '6',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '6/1',
                                                                                         'in_discards': '42211',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '9904651',
                                                                                         'in_octets': '14991531256632',
                                                                                         'in_ucast_pkts': '73987214268',
                                                                                         'links': [       {       'in_nucast_pkts': '366',
                                                                                                                  'in_octets': '217782984688',
                                                                                                                  'in_ucast_pkts': '1758713255',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75606980',
                                                                                                                  'out_octets': '10578000550410',
                                                                                                                  'out_ucast_pkts': '7172572302',
                                                                                                                  'poller_hash': 'OEEwOTBwZWVyaW50ZnBvbiAxMjQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3040840369943',
                                                                                                                  'in_ucast_pkts': '11438225736',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '47154507820415',
                                                                                                                  'out_ucast_pkts': '35411164554',
                                                                                                                  'poller_hash': 'ODg5RkRwZWVyaW50ZnBvbiA1Nzk='},
                                                                                                          {       'in_nucast_pkts': '14863',
                                                                                                                  'in_octets': '7622399207',
                                                                                                                  'in_ucast_pkts': '38039469',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463814',
                                                                                                                  'out_octets': '161235625604',
                                                                                                                  'out_ucast_pkts': '119630717',
                                                                                                                  'poller_hash': 'QzgzQkE2cGVlcmludGZwb24gNTgw'},
                                                                                                          {       'in_nucast_pkts': '223',
                                                                                                                  'in_octets': '14957500527',
                                                                                                                  'in_ucast_pkts': '21818285',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119744',
                                                                                                                  'out_octets': '39695966139',
                                                                                                                  'out_ucast_pkts': '37386254',
                                                                                                                  'poller_hash': 'MTJGRkQ5cGVlcmludGZwb24gNTgy'},
                                                                                                          {       'in_nucast_pkts': '12005',
                                                                                                                  'in_octets': '25134249743',
                                                                                                                  'in_ucast_pkts': '47751907',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463884',
                                                                                                                  'out_octets': '170798791136',
                                                                                                                  'out_ucast_pkts': '132394510',
                                                                                                                  'poller_hash': 'QTlDODc5cGVlcmludGZwb24gNTgz'},
                                                                                                          {       'in_nucast_pkts': '36831671',
                                                                                                                  'in_octets': '1797081955845',
                                                                                                                  'in_ucast_pkts': '4649529023',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83353781129',
                                                                                                                  'out_octets': '133927933802718',
                                                                                                                  'out_ucast_pkts': '15265466363',
                                                                                                                  'poller_hash': 'ODg5RTJwZWVyaW50ZnBvbiA1ODQ='},
                                                                                                          {       'in_nucast_pkts': '740708',
                                                                                                                  'in_octets': '183827674100',
                                                                                                                  'in_ucast_pkts': '1661128349',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75613326',
                                                                                                                  'out_octets': '12871294173730',
                                                                                                                  'out_ucast_pkts': '8641882702',
                                                                                                                  'poller_hash': 'ODg5QkRwZWVyaW50ZnBvbiA1ODU='},
                                                                                                          {       'in_nucast_pkts': '3381',
                                                                                                                  'in_octets': '78851103318',
                                                                                                                  'in_ucast_pkts': '459222108',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27324140',
                                                                                                                  'out_octets': '3549923226663',
                                                                                                                  'out_ucast_pkts': '2509052173',
                                                                                                                  'poller_hash': 'OEEwNzZwZWVyaW50ZnBvbiA1ODY='},
                                                                                                          {       'in_nucast_pkts': '56922',
                                                                                                                  'in_octets': '61488454347',
                                                                                                                  'in_ucast_pkts': '145998276',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6776927',
                                                                                                                  'out_octets': '378159229707',
                                                                                                                  'out_ucast_pkts': '308446337',
                                                                                                                  'poller_hash': 'MUQwNjhEcGVlcmludGZwb24gNTg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1338463716881',
                                                                                                                  'in_ucast_pkts': '8009942926',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '40782014297710',
                                                                                                                  'out_ucast_pkts': '29566141784',
                                                                                                                  'poller_hash': 'ODg5NEZwZWVyaW50ZnBvbiA1ODg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '313452410460',
                                                                                                                  'in_ucast_pkts': '2590526640',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '17352177141514',
                                                                                                                  'out_ucast_pkts': '11801361485',
                                                                                                                  'poller_hash': 'OEJEQzhwZWVyaW50ZnBvbiA1ODk='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '336004994713',
                                                                                                                  'in_ucast_pkts': '1071341243',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4275298476848',
                                                                                                                  'out_ucast_pkts': '3250646993',
                                                                                                                  'poller_hash': 'OEEwNkRwZWVyaW50ZnBvbiA1OTA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '68154146783',
                                                                                                                  'in_ucast_pkts': '562116518',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2804632842739',
                                                                                                                  'out_ucast_pkts': '1991471643',
                                                                                                                  'poller_hash': 'ODhBMUVwZWVyaW50ZnBvbiA1OTE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '276219861792',
                                                                                                                  'in_ucast_pkts': '1518767293',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9308621815590',
                                                                                                                  'out_ucast_pkts': '6358465124',
                                                                                                                  'poller_hash': 'ODg5QTZwZWVyaW50ZnBvbiA1OTI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '150343919448',
                                                                                                                  'in_ucast_pkts': '1243560258',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '5783577860133',
                                                                                                                  'out_ucast_pkts': '4088959213',
                                                                                                                  'poller_hash': 'ODg5RjFwZWVyaW50ZnBvbiA1OTM='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '277428311559',
                                                                                                                  'in_ucast_pkts': '3076048934',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9293654946108',
                                                                                                                  'out_ucast_pkts': '2095521715',
                                                                                                                  'poller_hash': 'ODg5RUZwZWVyaW50ZnBvbiA1OTQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzRGRkNwZWVyaW50ZnBvbiA1OTU='},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '1770556986',
                                                                                                                  'in_ucast_pkts': '8306747',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119629',
                                                                                                                  'out_octets': '29801295936',
                                                                                                                  'out_ucast_pkts': '21010117',
                                                                                                                  'poller_hash': 'QzUwMDJwZWVyaW50ZnBvbiA1OTY='},
                                                                                                          {       'in_nucast_pkts': '787872',
                                                                                                                  'in_octets': '68770153914',
                                                                                                                  'in_ucast_pkts': '233287236',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75606043',
                                                                                                                  'out_octets': '1735607968031',
                                                                                                                  'out_ucast_pkts': '1239520494',
                                                                                                                  'poller_hash': 'QzUwMDRwZWVyaW50ZnBvbiA1OTc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '360032718082',
                                                                                                                  'in_ucast_pkts': '2889681745',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '16347137802612',
                                                                                                                  'out_ucast_pkts': '11113069166',
                                                                                                                  'poller_hash': 'QzU3MDNwZWVyaW50ZnBvbiA1OTg='},
                                                                                                          {       'in_nucast_pkts': '18326',
                                                                                                                  'in_octets': '35783339294',
                                                                                                                  'in_ucast_pkts': '245951703',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3188229',
                                                                                                                  'out_octets': '1195433354666',
                                                                                                                  'out_ucast_pkts': '829688936',
                                                                                                                  'poller_hash': 'QzU3NDVwZWVyaW50ZnBvbiA1OTk='},
                                                                                                          {       'in_nucast_pkts': '162',
                                                                                                                  'in_octets': '223422595204',
                                                                                                                  'in_ucast_pkts': '1771093812',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23532693',
                                                                                                                  'out_octets': '4997576277456',
                                                                                                                  'out_ucast_pkts': '3439787621',
                                                                                                                  'poller_hash': 'QzREREZwZWVyaW50ZnBvbiA2MDA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzRERUFwZWVyaW50ZnBvbiA2MDE='},
                                                                                                          {       'in_nucast_pkts': '14724',
                                                                                                                  'in_octets': '5298837087',
                                                                                                                  'in_ucast_pkts': '11923135',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119608',
                                                                                                                  'out_octets': '62626767483',
                                                                                                                  'out_ucast_pkts': '47861488',
                                                                                                                  'poller_hash': 'QzRERDFwZWVyaW50ZnBvbiA2MDI='},
                                                                                                          {       'in_nucast_pkts': '14402',
                                                                                                                  'in_octets': '18937474657',
                                                                                                                  'in_ucast_pkts': '45580790',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119711',
                                                                                                                  'out_octets': '487633314543',
                                                                                                                  'out_ucast_pkts': '353175908',
                                                                                                                  'poller_hash': 'QzRERkRwZWVyaW50ZnBvbiA2MDM='},
                                                                                                          {       'in_nucast_pkts': '19',
                                                                                                                  'in_octets': '8249711709',
                                                                                                                  'in_ucast_pkts': '69524452',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119387',
                                                                                                                  'out_octets': '214943491131',
                                                                                                                  'out_ucast_pkts': '148429919',
                                                                                                                  'poller_hash': 'Q0MwQzFwZWVyaW50ZnBvbiA2MDU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '341134059259',
                                                                                                                  'in_ucast_pkts': '2220442544',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '12229459173047',
                                                                                                                  'out_ucast_pkts': '8254388381',
                                                                                                                  'poller_hash': 'MTc1RTM5cGVlcmludGZwb24gNjA3'},
                                                                                                          {       'in_nucast_pkts': '15396',
                                                                                                                  'in_octets': '1293085795',
                                                                                                                  'in_ucast_pkts': '12394425',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463864',
                                                                                                                  'out_octets': '47741208514',
                                                                                                                  'out_ucast_pkts': '32599593',
                                                                                                                  'poller_hash': 'ODJBNTMycGVlcmludGZwb24gMTQwNzI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '256972643908',
                                                                                                                  'in_ucast_pkts': '1910934326',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '190',
                                                                                                                  'out_octets': '9258717304557',
                                                                                                                  'out_ucast_pkts': '6710554421',
                                                                                                                  'poller_hash': 'Q0JEMjMzcGVlcmludGZwb24gMTY4MjMy'},
                                                                                                          {       'in_nucast_pkts': '435544',
                                                                                                                  'in_octets': '113734770872',
                                                                                                                  'in_ucast_pkts': '729259407',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43147966',
                                                                                                                  'out_octets': '3755045982688',
                                                                                                                  'out_ucast_pkts': '2638891920',
                                                                                                                  'poller_hash': 'Q0MwQkZwZWVyaW50ZnBvbiAyNTc2NTg='},
                                                                                                          {       'in_nucast_pkts': '295',
                                                                                                                  'in_octets': '55377752882',
                                                                                                                  'in_ucast_pkts': '565689505',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1425',
                                                                                                                  'out_octets': '2305495252659',
                                                                                                                  'out_ucast_pkts': '1560648822',
                                                                                                                  'poller_hash': 'MTI3NDg2RnBlZXJpbnRmcG9uIDI3OTM2Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '196899051799',
                                                                                         'out_octets': '601893281075826',
                                                                                         'out_ucast_pkts': '238507708156',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/1',
                                                                                                                 'id': '6/1',
                                                                                                                 'links': [       {       'connected_time': '40931',
                                                                                                                                          'fiber_length': 830,
                                                                                                                                          'poller_hash': 'OEEwOTBwZWVycG9ucG9uIDEyNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '13161611',
                                                                                                                                          'fiber_length': 1170,
                                                                                                                                          'poller_hash': 'ODg5RkRwZWVycG9ucG9uIDU3OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688412',
                                                                                                                                          'fiber_length': 1270,
                                                                                                                                          'poller_hash': 'QzgzQkE2cGVlcnBvbnBvbiA1ODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688402',
                                                                                                                                          'fiber_length': 842,
                                                                                                                                          'poller_hash': 'MTJGRkQ5cGVlcnBvbnBvbiA1ODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688413',
                                                                                                                                          'fiber_length': 984,
                                                                                                                                          'poller_hash': 'QTlDODc5cGVlcnBvbnBvbiA1ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '40894',
                                                                                                                                          'fiber_length': 1004,
                                                                                                                                          'poller_hash': 'ODg5RTJwZWVycG9ucG9uIDU4NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '40906',
                                                                                                                                          'fiber_length': 902,
                                                                                                                                          'poller_hash': 'ODg5QkRwZWVycG9ucG9uIDU4NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '15498464',
                                                                                                                                          'fiber_length': 870,
                                                                                                                                          'poller_hash': 'OEEwNzZwZWVycG9ucG9uIDU4Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '3113515',
                                                                                                                                          'fiber_length': 1112,
                                                                                                                                          'poller_hash': 'MUQwNjhEcGVlcnBvbnBvbiA1ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '19446364',
                                                                                                                                          'fiber_length': 1073,
                                                                                                                                          'poller_hash': 'ODg5NEZwZWVycG9ucG9uIDU4OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688403',
                                                                                                                                          'fiber_length': 830,
                                                                                                                                          'poller_hash': 'OEJEQzhwZWVycG9ucG9uIDU4OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688408',
                                                                                                                                          'fiber_length': 809,
                                                                                                                                          'poller_hash': 'OEEwNkRwZWVycG9ucG9uIDU5MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '18858052',
                                                                                                                                          'fiber_length': 1097,
                                                                                                                                          'poller_hash': 'ODhBMUVwZWVycG9ucG9uIDU5MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688435',
                                                                                                                                          'fiber_length': 1133,
                                                                                                                                          'poller_hash': 'ODg5QTZwZWVycG9ucG9uIDU5Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '13219156',
                                                                                                                                          'fiber_length': 1099,
                                                                                                                                          'poller_hash': 'ODg5RjFwZWVycG9ucG9uIDU5Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '47398',
                                                                                                                                          'fiber_length': 1030,
                                                                                                                                          'poller_hash': 'ODg5RUZwZWVycG9ucG9uIDU5NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '41295',
                                                                                                                                          'fiber_length': 1138,
                                                                                                                                          'poller_hash': 'QzRGRkNwZWVycG9ucG9uIDU5NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688413',
                                                                                                                                          'fiber_length': 1154,
                                                                                                                                          'poller_hash': 'QzUwMDJwZWVycG9ucG9uIDU5Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '41171',
                                                                                                                                          'fiber_length': 1142,
                                                                                                                                          'poller_hash': 'QzUwMDRwZWVycG9ucG9uIDU5Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '13218205',
                                                                                                                                          'fiber_length': 1178,
                                                                                                                                          'poller_hash': 'QzU3MDNwZWVycG9ucG9uIDU5OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '68514702',
                                                                                                                                          'fiber_length': 1027,
                                                                                                                                          'poller_hash': 'QzU3NDVwZWVycG9ucG9uIDU5OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '13219155',
                                                                                                                                          'fiber_length': 1163,
                                                                                                                                          'poller_hash': 'QzREREZwZWVycG9ucG9uIDYwMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '47369',
                                                                                                                                          'fiber_length': 899,
                                                                                                                                          'poller_hash': 'QzRERUFwZWVycG9ucG9uIDYwMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688415',
                                                                                                                                          'fiber_length': 1139,
                                                                                                                                          'poller_hash': 'QzRERDFwZWVycG9ucG9uIDYwMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688414',
                                                                                                                                          'fiber_length': 1122,
                                                                                                                                          'poller_hash': 'QzRERkRwZWVycG9ucG9uIDYwMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688358',
                                                                                                                                          'fiber_length': 1197,
                                                                                                                                          'poller_hash': 'Q0MwQzFwZWVycG9ucG9uIDYwNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '47454',
                                                                                                                                          'fiber_length': 1163,
                                                                                                                                          'poller_hash': 'MTc1RTM5cGVlcnBvbnBvbiA2MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '688442',
                                                                                                                                          'fiber_length': 1301,
                                                                                                                                          'poller_hash': 'ODJBNTMycGVlcnBvbnBvbiAxNDA3MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '7865213',
                                                                                                                                          'fiber_length': 1088,
                                                                                                                                          'poller_hash': 'Q0JEMjMzcGVlcnBvbnBvbiAxNjgyMzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '25923133',
                                                                                                                                          'fiber_length': 1152,
                                                                                                                                          'poller_hash': 'Q0MwQkZwZWVycG9ucG9uIDI1NzY1OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707},
                                                                                                                                  {       'connected_time': '6400919',
                                                                                                                                          'fiber_length': 916,
                                                                                                                                          'poller_hash': 'MTI3NDg2RnBlZXJwb25wb24gMjc5MzY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.707}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/1',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/2',
                                                                                         'in_discards': '5615',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '9601929',
                                                                                         'in_octets': '13244502494423',
                                                                                         'in_ucast_pkts': '57808417212',
                                                                                         'links': [       {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '1308636043',
                                                                                                                  'in_ucast_pkts': '6473285',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1461871',
                                                                                                                  'out_octets': '28120988098',
                                                                                                                  'out_ucast_pkts': '21664604',
                                                                                                                  'poller_hash': 'NDJDNDBFcGVlcmludGZwb24gNTY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '434103703880',
                                                                                                                  'in_ucast_pkts': '1208678254',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4344061812092',
                                                                                                                  'out_ucast_pkts': '3388453620',
                                                                                                                  'poller_hash': 'ODg5QjdwZWVyaW50ZnBvbiA2MDg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '153191736628',
                                                                                                                  'in_ucast_pkts': '880880549',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '5327560952129',
                                                                                                                  'out_ucast_pkts': '3981514416',
                                                                                                                  'poller_hash': 'QzRERUNwZWVyaW50ZnBvbiA2MDk='},
                                                                                                          {       'in_nucast_pkts': '20',
                                                                                                                  'in_octets': '1515865747',
                                                                                                                  'in_ucast_pkts': '6421302',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119745',
                                                                                                                  'out_octets': '31567542553',
                                                                                                                  'out_ucast_pkts': '25880507',
                                                                                                                  'poller_hash': 'ODg5RTlwZWVyaW50ZnBvbiA2MTA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '795540980014',
                                                                                                                  'in_ucast_pkts': '4395297900',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '14811366790480',
                                                                                                                  'out_ucast_pkts': '12678147574',
                                                                                                                  'poller_hash': 'QzRGOUJwZWVyaW50ZnBvbiA2MTE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '81000695960',
                                                                                                                  'in_ucast_pkts': '257528207',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '626027854941',
                                                                                                                  'out_ucast_pkts': '584625284',
                                                                                                                  'poller_hash': 'QzRGRkFwZWVyaW50ZnBvbiA2MTM='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '404918539647',
                                                                                                                  'in_ucast_pkts': '4220523158',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '24200369063594',
                                                                                                                  'out_ucast_pkts': '16236970941',
                                                                                                                  'poller_hash': 'QzRGRjhwZWVyaW50ZnBvbiA2MTQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '523691200358',
                                                                                                                  'in_ucast_pkts': '2931181063',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '10567860334678',
                                                                                                                  'out_ucast_pkts': '7542637009',
                                                                                                                  'poller_hash': 'QzRGRUFwZWVyaW50ZnBvbiA2MTU='},
                                                                                                          {       'in_nucast_pkts': '4667170',
                                                                                                                  'in_octets': '50388400459',
                                                                                                                  'in_ucast_pkts': '185960096',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75614116',
                                                                                                                  'out_octets': '887086887145',
                                                                                                                  'out_ucast_pkts': '729332072',
                                                                                                                  'poller_hash': 'QzRGRkJwZWVyaW50ZnBvbiA2MTY='},
                                                                                                          {       'in_nucast_pkts': '758234',
                                                                                                                  'in_octets': '452041457508',
                                                                                                                  'in_ucast_pkts': '2079179850',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75585107',
                                                                                                                  'out_octets': '5568915967185',
                                                                                                                  'out_ucast_pkts': '4202586371',
                                                                                                                  'poller_hash': 'QzRERjZwZWVyaW50ZnBvbiA2MTc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '70047078405',
                                                                                                                  'in_ucast_pkts': '477442318',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2973947095223',
                                                                                                                  'out_ucast_pkts': '2063099395',
                                                                                                                  'poller_hash': 'QzU4NUZwZWVyaW50ZnBvbiA2MTg='},
                                                                                                          {       'in_nucast_pkts': '13398',
                                                                                                                  'in_octets': '6193282158',
                                                                                                                  'in_ucast_pkts': '28772459',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1108721',
                                                                                                                  'out_octets': '165511465437',
                                                                                                                  'out_ucast_pkts': '116313344',
                                                                                                                  'poller_hash': 'QzU3MTNwZWVyaW50ZnBvbiA2MTk='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '814121355929',
                                                                                                                  'in_ucast_pkts': '7865277151',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '43852631136757',
                                                                                                                  'out_ucast_pkts': '29337407185',
                                                                                                                  'poller_hash': 'Q0MwQzlwZWVyaW50ZnBvbiA2MjA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '222725364514',
                                                                                                                  'in_ucast_pkts': '486020552',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1876894309267',
                                                                                                                  'out_ucast_pkts': '1551821792',
                                                                                                                  'poller_hash': 'QzU3NzhwZWVyaW50ZnBvbiA2MjE='},
                                                                                                          {       'in_nucast_pkts': '85',
                                                                                                                  'in_octets': '259681113114',
                                                                                                                  'in_ucast_pkts': '2344649092',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23530378',
                                                                                                                  'out_octets': '8986014715752',
                                                                                                                  'out_ucast_pkts': '6054430676',
                                                                                                                  'poller_hash': 'QzRERkJwZWVyaW50ZnBvbiA2MjM='},
                                                                                                          {       'in_nucast_pkts': '106',
                                                                                                                  'in_octets': '146865990987',
                                                                                                                  'in_ucast_pkts': '303783776',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27326996',
                                                                                                                  'out_octets': '1556905173229',
                                                                                                                  'out_ucast_pkts': '1227837001',
                                                                                                                  'poller_hash': 'QzU3RThwZWVyaW50ZnBvbiA2MjQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '815885935683',
                                                                                                                  'in_ucast_pkts': '2448812465',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '12553686561675',
                                                                                                                  'out_ucast_pkts': '9284067750',
                                                                                                                  'poller_hash': 'QzU3RjFwZWVyaW50ZnBvbiA2MjU='},
                                                                                                          {       'in_nucast_pkts': '2538',
                                                                                                                  'in_octets': '489812510210',
                                                                                                                  'in_ucast_pkts': '3191781464',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75606187',
                                                                                                                  'out_octets': '7896808577736',
                                                                                                                  'out_ucast_pkts': '5485713657',
                                                                                                                  'poller_hash': 'Q0MwRDVwZWVyaW50ZnBvbiA2MjY='},
                                                                                                          {       'in_nucast_pkts': '636',
                                                                                                                  'in_octets': '5821300917',
                                                                                                                  'in_ucast_pkts': '21393255',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119735',
                                                                                                                  'out_octets': '120016045239',
                                                                                                                  'out_ucast_pkts': '87947413',
                                                                                                                  'poller_hash': 'NDNCNTRDcGVlcmludGZwb24gMTQxMTM5'},
                                                                                                          {       'in_nucast_pkts': '14208',
                                                                                                                  'in_octets': '5101089086',
                                                                                                                  'in_ucast_pkts': '9725901',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464127',
                                                                                                                  'out_octets': '20701379698',
                                                                                                                  'out_ucast_pkts': '18402248',
                                                                                                                  'poller_hash': 'QzgzQkQ4cGVlcmludGZwb24gMTk0Mzcw'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '71321167490',
                                                                                                                  'in_ucast_pkts': '113599102',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1381296',
                                                                                                                  'out_octets': '258337545616',
                                                                                                                  'out_ucast_pkts': '217772221',
                                                                                                                  'poller_hash': 'MTI3NTAwOHBlZXJpbnRmcG9uIDIyMTYwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '130936577180',
                                                                                                                  'in_ucast_pkts': '907415045',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '32',
                                                                                                                  'out_octets': '4873947763248',
                                                                                                                  'out_ucast_pkts': '3369877515',
                                                                                                                  'poller_hash': 'MTI3NDZCQnBlZXJpbnRmcG9uIDI2MjU3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDYwRXBlZXJpbnRmcG9uIDI2Mzg3Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '66275678793',
                                                                                         'out_octets': '320016985452599',
                                                                                         'out_ucast_pkts': '165939648592',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/2',
                                                                                                                 'id': '6/2',
                                                                                                                 'links': [       {       'connected_time': '688137',
                                                                                                                                          'fiber_length': 1041,
                                                                                                                                          'poller_hash': 'NDJDNDBFcGVlcnBvbnBvbiA1NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688407',
                                                                                                                                          'fiber_length': 891,
                                                                                                                                          'poller_hash': 'ODg5QjdwZWVycG9ucG9uIDYwOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '8385513',
                                                                                                                                          'fiber_length': 704,
                                                                                                                                          'poller_hash': 'QzRERUNwZWVycG9ucG9uIDYwOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688400',
                                                                                                                                          'fiber_length': 718,
                                                                                                                                          'poller_hash': 'ODg5RTlwZWVycG9ucG9uIDYxMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688408',
                                                                                                                                          'fiber_length': 911,
                                                                                                                                          'poller_hash': 'QzRGOUJwZWVycG9ucG9uIDYxMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '47941',
                                                                                                                                          'fiber_length': 935,
                                                                                                                                          'poller_hash': 'QzRGRkFwZWVycG9ucG9uIDYxMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688404',
                                                                                                                                          'fiber_length': 732,
                                                                                                                                          'poller_hash': 'QzRGRjhwZWVycG9ucG9uIDYxNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '21319272',
                                                                                                                                          'fiber_length': 1016,
                                                                                                                                          'poller_hash': 'QzRGRUFwZWVycG9ucG9uIDYxNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '40948',
                                                                                                                                          'fiber_length': 893,
                                                                                                                                          'poller_hash': 'QzRGRkJwZWVycG9ucG9uIDYxNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '41049',
                                                                                                                                          'fiber_length': 894,
                                                                                                                                          'poller_hash': 'QzRERjZwZWVycG9ucG9uIDYxNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '31279509',
                                                                                                                                          'fiber_length': 923,
                                                                                                                                          'poller_hash': 'QzU4NUZwZWVycG9ucG9uIDYxOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688407',
                                                                                                                                          'fiber_length': 871,
                                                                                                                                          'poller_hash': 'QzU3MTNwZWVycG9ucG9uIDYxOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '20997057',
                                                                                                                                          'fiber_length': 918,
                                                                                                                                          'poller_hash': 'Q0MwQzlwZWVycG9ucG9uIDYyMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '19105352',
                                                                                                                                          'fiber_length': 913,
                                                                                                                                          'poller_hash': 'QzU3NzhwZWVycG9ucG9uIDYyMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '13219225',
                                                                                                                                          'fiber_length': 878,
                                                                                                                                          'poller_hash': 'QzRERkJwZWVycG9ucG9uIDYyMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '15498391',
                                                                                                                                          'fiber_length': 947,
                                                                                                                                          'poller_hash': 'QzU3RThwZWVycG9ucG9uIDYyNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688409',
                                                                                                                                          'fiber_length': 893,
                                                                                                                                          'poller_hash': 'QzU3RjFwZWVycG9ucG9uIDYyNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '40403',
                                                                                                                                          'fiber_length': 947,
                                                                                                                                          'poller_hash': 'Q0MwRDVwZWVycG9ucG9uIDYyNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688413',
                                                                                                                                          'fiber_length': 880,
                                                                                                                                          'poller_hash': 'NDNCNTRDcGVlcnBvbnBvbiAxNDExMzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688421',
                                                                                                                                          'fiber_length': 1042,
                                                                                                                                          'poller_hash': 'QzgzQkQ4cGVlcnBvbnBvbiAxOTQzNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '648763',
                                                                                                                                          'fiber_length': 909,
                                                                                                                                          'poller_hash': 'MTI3NTAwOHBlZXJwb25wb24gMjIxNjAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '21175394',
                                                                                                                                          'fiber_length': 982,
                                                                                                                                          'poller_hash': 'MTI3NDZCQnBlZXJwb25wb24gMjYyNTc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '19441886',
                                                                                                                                          'fiber_length': 952,
                                                                                                                                          'poller_hash': 'MTI3NDYwRXBlZXJwb25wb24gMjYzODc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/2',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/3',
                                                                                         'in_discards': '18',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1058098',
                                                                                         'in_octets': '1680898360971',
                                                                                         'in_ucast_pkts': '11514905488',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '356583159235',
                                                                                                                  'in_ucast_pkts': '2555590677',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11326681380339',
                                                                                                                  'out_ucast_pkts': '7989817063',
                                                                                                                  'poller_hash': 'ODg5REVwZWVyaW50ZnBvbiA2MzA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'OEEyOENwZWVyaW50ZnBvbiA2MzE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '173832463007',
                                                                                                                  'in_ucast_pkts': '1508835090',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '10192484880360',
                                                                                                                  'out_ucast_pkts': '6945909093',
                                                                                                                  'poller_hash': 'MTI3NDMzNXBlZXJpbnRmcG9uIDYzMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '709831729901',
                                                                                                                  'in_ucast_pkts': '5212861463',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '20033470962067',
                                                                                                                  'out_ucast_pkts': '10036115478',
                                                                                                                  'poller_hash': 'ODg5QkFwZWVyaW50ZnBvbiA2MzM='},
                                                                                                          {       'in_nucast_pkts': '55',
                                                                                                                  'in_octets': '5478112281',
                                                                                                                  'in_ucast_pkts': '26998117',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119730',
                                                                                                                  'out_octets': '110457682467',
                                                                                                                  'out_ucast_pkts': '81434681',
                                                                                                                  'poller_hash': 'QzU3NENwZWVyaW50ZnBvbiA2MzQ='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '44148967109',
                                                                                         'out_octets': '114106571395601',
                                                                                         'out_ucast_pkts': '38432022148',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/3',
                                                                                                                 'id': '6/3',
                                                                                                                 'links': [       {       'connected_time': '13219209',
                                                                                                                                          'fiber_length': 1553,
                                                                                                                                          'poller_hash': 'ODg5REVwZWVycG9ucG9uIDYzMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '15224741',
                                                                                                                                          'fiber_length': 1752,
                                                                                                                                          'poller_hash': 'OEEyOENwZWVycG9ucG9uIDYzMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688422',
                                                                                                                                          'fiber_length': 1856,
                                                                                                                                          'poller_hash': 'MTI3NDMzNXBlZXJwb25wb24gNjMy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '29473757',
                                                                                                                                          'fiber_length': 1668,
                                                                                                                                          'poller_hash': 'ODg5QkFwZWVycG9ucG9uIDYzMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69},
                                                                                                                                  {       'connected_time': '688418',
                                                                                                                                          'fiber_length': 911,
                                                                                                                                          'poller_hash': 'QzU3NENwZWVycG9ucG9uIDYzNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.69}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/3',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/4',
                                                                                         'in_discards': '39278',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5927369',
                                                                                         'in_octets': '11411466434134',
                                                                                         'in_ucast_pkts': '52466519124',
                                                                                         'links': [       {       'in_nucast_pkts': '6183047',
                                                                                                                  'in_octets': '462704081998',
                                                                                                                  'in_ucast_pkts': '4153758055',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15649836734',
                                                                                                                  'out_octets': '41766975735081',
                                                                                                                  'out_ucast_pkts': '13942871592',
                                                                                                                  'poller_hash': 'ODg5MjVwZWVyaW50ZnBvbiA2MzU='},
                                                                                                          {       'in_nucast_pkts': '244754',
                                                                                                                  'in_octets': '533975708583',
                                                                                                                  'in_ucast_pkts': '2273440371',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23528274',
                                                                                                                  'out_octets': '5310094860518',
                                                                                                                  'out_ucast_pkts': '4409641477',
                                                                                                                  'poller_hash': 'ODg4RjJwZWVyaW50ZnBvbiA2MzY='},
                                                                                                          {       'in_nucast_pkts': '1185747',
                                                                                                                  'in_octets': '294693738986',
                                                                                                                  'in_ucast_pkts': '2280277097',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53387405',
                                                                                                                  'out_octets': '12654045516077',
                                                                                                                  'out_ucast_pkts': '8636031767',
                                                                                                                  'poller_hash': 'ODg4QTVwZWVyaW50ZnBvbiA2Mzc='},
                                                                                                          {       'in_nucast_pkts': '79',
                                                                                                                  'in_octets': '4874593975',
                                                                                                                  'in_ucast_pkts': '16720634',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119713',
                                                                                                                  'out_octets': '89519572121',
                                                                                                                  'out_ucast_pkts': '67839621',
                                                                                                                  'poller_hash': 'ODg4RjVwZWVyaW50ZnBvbiA2Mzg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '255563362594',
                                                                                                                  'in_ucast_pkts': '1437066921',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '190',
                                                                                                                  'out_octets': '7503618302919',
                                                                                                                  'out_ucast_pkts': '5216999021',
                                                                                                                  'poller_hash': 'MjA1REI5cGVlcmludGZwb24gNjM5'},
                                                                                                          {       'in_nucast_pkts': '299',
                                                                                                                  'in_octets': '1034540548473',
                                                                                                                  'in_ucast_pkts': '3628346778',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75653288',
                                                                                                                  'out_octets': '13783821698849',
                                                                                                                  'out_ucast_pkts': '10122595903',
                                                                                                                  'poller_hash': 'ODg5QTVwZWVyaW50ZnBvbiA2NDA='},
                                                                                                          {       'in_nucast_pkts': '298',
                                                                                                                  'in_octets': '20282351987',
                                                                                                                  'in_ucast_pkts': '68904238',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1512',
                                                                                                                  'out_octets': '522389069928',
                                                                                                                  'out_ucast_pkts': '451027063',
                                                                                                                  'poller_hash': 'MTE3REFEMnBlZXJpbnRmcG9uIDY0MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '478307205437',
                                                                                                                  'in_ucast_pkts': '4742986115',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '22814012295982',
                                                                                                                  'out_ucast_pkts': '15354707335',
                                                                                                                  'poller_hash': 'MzkwMUFBcGVlcmludGZwb24gNjQy'},
                                                                                                          {       'in_nucast_pkts': '2214',
                                                                                                                  'in_octets': '15006706457',
                                                                                                                  'in_ucast_pkts': '103709666',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119933',
                                                                                                                  'out_octets': '396478702294',
                                                                                                                  'out_ucast_pkts': '272198665',
                                                                                                                  'poller_hash': 'QzU3OENwZWVyaW50ZnBvbiA2NDM='},
                                                                                                          {       'in_nucast_pkts': '98560',
                                                                                                                  'in_octets': '39443458487',
                                                                                                                  'in_ucast_pkts': '158726890',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75672969',
                                                                                                                  'out_octets': '454384746327',
                                                                                                                  'out_ucast_pkts': '382257865',
                                                                                                                  'poller_hash': 'MTMwMDUzcGVlcmludGZwb24gNjQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '479987549883',
                                                                                                                  'in_ucast_pkts': '1804667832',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7765981897299',
                                                                                                                  'out_ucast_pkts': '5653769712',
                                                                                                                  'poller_hash': 'MTMwMDI2cGVlcmludGZwb24gNjQ2'},
                                                                                                          {       'in_nucast_pkts': '664844',
                                                                                                                  'in_octets': '528591584342',
                                                                                                                  'in_ucast_pkts': '2624798477',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66209265',
                                                                                                                  'out_octets': '12653108376390',
                                                                                                                  'out_ucast_pkts': '9035744537',
                                                                                                                  'poller_hash': 'MTJGRkQ0cGVlcmludGZwb24gNjQ3'},
                                                                                                          {       'in_nucast_pkts': '12208',
                                                                                                                  'in_octets': '7155963837',
                                                                                                                  'in_ucast_pkts': '54331719',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119711',
                                                                                                                  'out_octets': '299594896861',
                                                                                                                  'out_ucast_pkts': '213818761',
                                                                                                                  'poller_hash': 'MTMwMDhFcGVlcmludGZwb24gNjQ4'},
                                                                                                          {       'in_nucast_pkts': '321',
                                                                                                                  'in_octets': '157632738659',
                                                                                                                  'in_ucast_pkts': '691821891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75606886',
                                                                                                                  'out_octets': '3637548601275',
                                                                                                                  'out_ucast_pkts': '2581684114',
                                                                                                                  'poller_hash': 'MTJGRkYzcGVlcmludGZwb24gNjQ5'},
                                                                                                          {       'in_nucast_pkts': '903278',
                                                                                                                  'in_octets': '1779629366949',
                                                                                                                  'in_ucast_pkts': '5537064039',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '130014912',
                                                                                                                  'out_octets': '20869696689404',
                                                                                                                  'out_ucast_pkts': '15769619048',
                                                                                                                  'poller_hash': 'MUQwNjhCcGVlcmludGZwb24gNjUw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '124006365140',
                                                                                                                  'in_ucast_pkts': '1518939787',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '110',
                                                                                                                  'out_octets': '9393693034249',
                                                                                                                  'out_ucast_pkts': '6255264782',
                                                                                                                  'poller_hash': 'MUQwNjhGcGVlcmludGZwb24gNjUx'},
                                                                                                          {       'in_nucast_pkts': '12476',
                                                                                                                  'in_octets': '623141694',
                                                                                                                  'in_ucast_pkts': '2421956',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463860',
                                                                                                                  'out_octets': '13508612138',
                                                                                                                  'out_ucast_pkts': '10985466',
                                                                                                                  'poller_hash': 'ODJBNzUwcGVlcmludGZwb24gODc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgzMThEcGVlcmludGZwb24gMTkzMzgz'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '6838428539',
                                                                                                                  'in_ucast_pkts': '26027592',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464147',
                                                                                                                  'out_octets': '94367802746',
                                                                                                                  'out_ucast_pkts': '67814693',
                                                                                                                  'poller_hash': 'QzgxRjUycGVlcmludGZwb24gMTk3OTUw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NUU2NnBlZXJpbnRmcG9uIDI4MDMxNA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '89210447790',
                                                                                         'out_octets': '351323103096846',
                                                                                         'out_ucast_pkts': '164636843275',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/4',
                                                                                                                 'id': '6/4',
                                                                                                                 'links': [       {       'connected_time': '41107',
                                                                                                                                          'fiber_length': 1478,
                                                                                                                                          'poller_hash': 'ODg5MjVwZWVycG9ucG9uIDYzNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '13219243',
                                                                                                                                          'fiber_length': 1488,
                                                                                                                                          'poller_hash': 'ODg4RjJwZWVycG9ucG9uIDYzNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '28527661',
                                                                                                                                          'fiber_length': 1499,
                                                                                                                                          'poller_hash': 'ODg4QTVwZWVycG9ucG9uIDYzNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '688422',
                                                                                                                                          'fiber_length': 1541,
                                                                                                                                          'poller_hash': 'ODg4RjVwZWVycG9ucG9uIDYzOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '1378150',
                                                                                                                                          'fiber_length': 1728,
                                                                                                                                          'poller_hash': 'MjA1REI5cGVlcnBvbnBvbiA2Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '41227',
                                                                                                                                          'fiber_length': 1560,
                                                                                                                                          'poller_hash': 'ODg5QTVwZWVycG9ucG9uIDY0MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '688434',
                                                                                                                                          'fiber_length': 1565,
                                                                                                                                          'poller_hash': 'MTE3REFEMnBlZXJwb25wb24gNjQx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '66989',
                                                                                                                                          'fiber_length': 1452,
                                                                                                                                          'poller_hash': 'MzkwMUFBcGVlcnBvbnBvbiA2NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '688422',
                                                                                                                                          'fiber_length': 1524,
                                                                                                                                          'poller_hash': 'QzU3OENwZWVycG9ucG9uIDY0Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '41147',
                                                                                                                                          'fiber_length': 1436,
                                                                                                                                          'poller_hash': 'MTMwMDUzcGVlcnBvbnBvbiA2NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '15498581',
                                                                                                                                          'fiber_length': 1535,
                                                                                                                                          'poller_hash': 'MTMwMDI2cGVlcnBvbnBvbiA2NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '68194077',
                                                                                                                                          'fiber_length': 1506,
                                                                                                                                          'poller_hash': 'MTJGRkQ0cGVlcnBvbnBvbiA2NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '688431',
                                                                                                                                          'fiber_length': 1513,
                                                                                                                                          'poller_hash': 'MTMwMDhFcGVlcnBvbnBvbiA2NDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '41155',
                                                                                                                                          'fiber_length': 1499,
                                                                                                                                          'poller_hash': 'MTJGRkYzcGVlcnBvbnBvbiA2NDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '21817168',
                                                                                                                                          'fiber_length': 1785,
                                                                                                                                          'poller_hash': 'MUQwNjhCcGVlcnBvbnBvbiA2NTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '1052437',
                                                                                                                                          'fiber_length': 1856,
                                                                                                                                          'poller_hash': 'MUQwNjhGcGVlcnBvbnBvbiA2NTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '688459',
                                                                                                                                          'fiber_length': 1506,
                                                                                                                                          'poller_hash': 'ODJBNzUwcGVlcnBvbnBvbiA4Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '21815956',
                                                                                                                                          'fiber_length': 1779,
                                                                                                                                          'poller_hash': 'QzgzMThEcGVlcnBvbnBvbiAxOTMzODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '688433',
                                                                                                                                          'fiber_length': 1772,
                                                                                                                                          'poller_hash': 'QzgxRjUycGVlcnBvbnBvbiAxOTc5NTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84},
                                                                                                                                  {       'connected_time': '592922',
                                                                                                                                          'fiber_length': 1835,
                                                                                                                                          'poller_hash': 'MTI5NUU2NnBlZXJwb25wb24gMjgwMzE0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.84}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/4',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/5',
                                                                                         'in_discards': '2527',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '12876182',
                                                                                         'in_octets': '7855687992794',
                                                                                         'in_ucast_pkts': '45756994169',
                                                                                         'links': [       {       'in_nucast_pkts': '7438',
                                                                                                                  'in_octets': '6420382683',
                                                                                                                  'in_ucast_pkts': '24391573',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119775',
                                                                                                                  'out_octets': '96991097771',
                                                                                                                  'out_ucast_pkts': '76262059',
                                                                                                                  'poller_hash': 'NjQwQTc2cGVlcmludGZwb24gMzk0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '463848705454',
                                                                                                                  'in_ucast_pkts': '1847229686',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6295302986227',
                                                                                                                  'out_ucast_pkts': '4708800372',
                                                                                                                  'poller_hash': 'QzUwMzlwZWVyaW50ZnBvbiA2NTQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '254495743586',
                                                                                                                  'in_ucast_pkts': '44888386371',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '109',
                                                                                                                  'out_octets': '7944747759988',
                                                                                                                  'out_ucast_pkts': '5493642657',
                                                                                                                  'poller_hash': 'MjA1REMzcGVlcmludGZwb24gNjU1'},
                                                                                                          {       'in_nucast_pkts': '33230',
                                                                                                                  'in_octets': '9718321460',
                                                                                                                  'in_ucast_pkts': '59451736',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1153333',
                                                                                                                  'out_octets': '178022522861',
                                                                                                                  'out_ucast_pkts': '124840747',
                                                                                                                  'poller_hash': 'ODg5RTFwZWVyaW50ZnBvbiA2NTY='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '241122852245',
                                                                                                                  'in_ucast_pkts': '2366759624',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '13078415298323',
                                                                                                                  'out_ucast_pkts': '8864229251',
                                                                                                                  'poller_hash': 'ODcwNEZwZWVyaW50ZnBvbiA2NTc='},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '1409935644',
                                                                                                                  'in_ucast_pkts': '6802703',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119718',
                                                                                                                  'out_octets': '42421706427',
                                                                                                                  'out_ucast_pkts': '31350662',
                                                                                                                  'poller_hash': 'OEEyMzFwZWVyaW50ZnBvbiA2NTg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '148031049512',
                                                                                                                  'in_ucast_pkts': '945444811',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4589877883502',
                                                                                                                  'out_ucast_pkts': '3486795621',
                                                                                                                  'poller_hash': 'ODg4RkFwZWVyaW50ZnBvbiA2NTk='},
                                                                                                          {       'in_nucast_pkts': '806560',
                                                                                                                  'in_octets': '356485635191',
                                                                                                                  'in_ucast_pkts': '2848258929',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75587292',
                                                                                                                  'out_octets': '20724113740091',
                                                                                                                  'out_ucast_pkts': '14165794585',
                                                                                                                  'poller_hash': 'OEUxMDlwZWVyaW50ZnBvbiA2NjA='},
                                                                                                          {       'in_nucast_pkts': '75',
                                                                                                                  'in_octets': '13414369954',
                                                                                                                  'in_ucast_pkts': '74204486',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23529126',
                                                                                                                  'out_octets': '483338723217',
                                                                                                                  'out_ucast_pkts': '349102950',
                                                                                                                  'poller_hash': 'ODg1RTVwZWVyaW50ZnBvbiA2NjI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '11853116',
                                                                                                                  'in_ucast_pkts': '43940',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '18209624269',
                                                                                                                  'out_ucast_pkts': '80151115',
                                                                                                                  'poller_hash': 'ODg4OEZwZWVyaW50ZnBvbiA2NjM='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '52575376844',
                                                                                                                  'in_ucast_pkts': '111159679',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '128',
                                                                                                                  'out_octets': '408971399515',
                                                                                                                  'out_ucast_pkts': '380087245',
                                                                                                                  'poller_hash': 'MkIwMTgycGVlcmludGZwb24gNjY0'},
                                                                                                          {       'in_nucast_pkts': '449823',
                                                                                                                  'in_octets': '399463216665',
                                                                                                                  'in_ucast_pkts': '3980416604',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '47266953',
                                                                                                                  'out_octets': '27813794425205',
                                                                                                                  'out_ucast_pkts': '18700047395',
                                                                                                                  'poller_hash': 'ODg5RUFwZWVyaW50ZnBvbiA2NjU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'M0NGQTlwZWVyaW50ZnBvbiA2NjY='},
                                                                                                          {       'in_nucast_pkts': '393190',
                                                                                                                  'in_octets': '294333068606',
                                                                                                                  'in_ucast_pkts': '1225354534',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38333668',
                                                                                                                  'out_octets': '7415658221991',
                                                                                                                  'out_ucast_pkts': '5215326192',
                                                                                                                  'poller_hash': 'MTMwMDA2cGVlcmludGZwb24gNjY3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '101127006089',
                                                                                                                  'in_ucast_pkts': '1296996568',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7896770905820',
                                                                                                                  'out_ucast_pkts': '5255867099',
                                                                                                                  'poller_hash': 'MTMwMEM2cGVlcmludGZwb24gNjY4'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '5161899673',
                                                                                                                  'in_ucast_pkts': '20135056',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119708',
                                                                                                                  'out_octets': '94077149417',
                                                                                                                  'out_ucast_pkts': '69964685',
                                                                                                                  'poller_hash': 'MTMwMDlGcGVlcmludGZwb24gNjY5'},
                                                                                                          {       'in_nucast_pkts': '5',
                                                                                                                  'in_octets': '998324980',
                                                                                                                  'in_ucast_pkts': '5760836',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119700',
                                                                                                                  'out_octets': '32939887123',
                                                                                                                  'out_ucast_pkts': '22824330',
                                                                                                                  'poller_hash': 'MTMwMENCcGVlcmludGZwb24gNjcw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '314064913707',
                                                                                                                  'in_ucast_pkts': '82553151448',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '128',
                                                                                                                  'out_octets': '4576143080263',
                                                                                                                  'out_ucast_pkts': '3344980432',
                                                                                                                  'poller_hash': 'MzIxMkVGcGVlcmludGZwb24gNzM1'},
                                                                                                          {       'in_nucast_pkts': '11670',
                                                                                                                  'in_octets': '6258945068',
                                                                                                                  'in_ucast_pkts': '17644119',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463912',
                                                                                                                  'out_octets': '68791690329',
                                                                                                                  'out_ucast_pkts': '52105667',
                                                                                                                  'poller_hash': 'NDJEMzYxcGVlcmludGZwb24gNzgx'},
                                                                                                          {       'in_nucast_pkts': '947515',
                                                                                                                  'in_octets': '1157640598716',
                                                                                                                  'in_ucast_pkts': '4535728065',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '129897056',
                                                                                                                  'out_octets': '20724681785847',
                                                                                                                  'out_ucast_pkts': '15114355254',
                                                                                                                  'poller_hash': 'ODdGOEJEcGVlcmludGZwb24gMTQxNjc5'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '111046952804',
                                                                                         'out_octets': '363225220369292',
                                                                                         'out_ucast_pkts': '149532987423',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/5',
                                                                                                                 'id': '6/5',
                                                                                                                 'links': [       {       'connected_time': '688421',
                                                                                                                                          'fiber_length': 1321,
                                                                                                                                          'poller_hash': 'NjQwQTc2cGVlcnBvbnBvbiAzOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688420',
                                                                                                                                          'fiber_length': 1146,
                                                                                                                                          'poller_hash': 'QzUwMzlwZWVycG9ucG9uIDY1NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '24525703',
                                                                                                                                          'fiber_length': 1424,
                                                                                                                                          'poller_hash': 'MjA1REMzcGVlcnBvbnBvbiA2NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688422',
                                                                                                                                          'fiber_length': 1180,
                                                                                                                                          'poller_hash': 'ODg5RTFwZWVycG9ucG9uIDY1Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '31369146',
                                                                                                                                          'fiber_length': 1174,
                                                                                                                                          'poller_hash': 'ODcwNEZwZWVycG9ucG9uIDY1Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688417',
                                                                                                                                          'fiber_length': 1108,
                                                                                                                                          'poller_hash': 'OEEyMzFwZWVycG9ucG9uIDY1OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '47349',
                                                                                                                                          'fiber_length': 1088,
                                                                                                                                          'poller_hash': 'ODg4RkFwZWVycG9ucG9uIDY1OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '41195',
                                                                                                                                          'fiber_length': 1187,
                                                                                                                                          'poller_hash': 'OEUxMDlwZWVycG9ucG9uIDY2MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '13219277',
                                                                                                                                          'fiber_length': 1122,
                                                                                                                                          'poller_hash': 'ODg1RTVwZWVycG9ucG9uIDY2Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688422',
                                                                                                                                          'fiber_length': 1300,
                                                                                                                                          'poller_hash': 'ODg4OEZwZWVycG9ucG9uIDY2Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688434',
                                                                                                                                          'fiber_length': 1375,
                                                                                                                                          'poller_hash': 'MkIwMTgycGVlcnBvbnBvbiA2NjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '25137186',
                                                                                                                                          'fiber_length': 1342,
                                                                                                                                          'poller_hash': 'ODg5RUFwZWVycG9ucG9uIDY2NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '11453878',
                                                                                                                                          'fiber_length': 1229,
                                                                                                                                          'poller_hash': 'M0NGQTlwZWVycG9ucG9uIDY2Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '20919986',
                                                                                                                                          'fiber_length': 1559,
                                                                                                                                          'poller_hash': 'MTMwMDA2cGVlcnBvbnBvbiA2Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '31296076',
                                                                                                                                          'fiber_length': 1329,
                                                                                                                                          'poller_hash': 'MTMwMEM2cGVlcnBvbnBvbiA2Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688430',
                                                                                                                                          'fiber_length': 1401,
                                                                                                                                          'poller_hash': 'MTMwMDlGcGVlcnBvbnBvbiA2Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688431',
                                                                                                                                          'fiber_length': 1296,
                                                                                                                                          'poller_hash': 'MTMwMENCcGVlcnBvbnBvbiA2NzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688441',
                                                                                                                                          'fiber_length': 1469,
                                                                                                                                          'poller_hash': 'MzIxMkVGcGVlcnBvbnBvbiA3MzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '688436',
                                                                                                                                          'fiber_length': 1477,
                                                                                                                                          'poller_hash': 'NDJEMzYxcGVlcnBvbnBvbiA3ODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '21816367',
                                                                                                                                          'fiber_length': 1484,
                                                                                                                                          'poller_hash': 'ODdGOEJEcGVlcnBvbnBvbiAxNDE2Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/5',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/6',
                                                                                         'in_discards': '10344',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '8734725',
                                                                                         'in_octets': '49462119517773',
                                                                                         'in_ucast_pkts': '94018305019',
                                                                                         'links': [       {       'in_nucast_pkts': '21381',
                                                                                                                  'in_octets': '3864660766772',
                                                                                                                  'in_ucast_pkts': '6801131188',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27208229',
                                                                                                                  'out_octets': '13899593884188',
                                                                                                                  'out_ucast_pkts': '12554649569',
                                                                                                                  'poller_hash': 'MUZENkI3cGVlcmludGZwb24gNjcy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '11828950956',
                                                                                                                  'in_ucast_pkts': '136723052',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '604893281736',
                                                                                                                  'out_ucast_pkts': '448807159',
                                                                                                                  'poller_hash': 'MUQwNkNEcGVlcmludGZwb24gNjcz'},
                                                                                                          {       'in_nucast_pkts': '4498',
                                                                                                                  'in_octets': '205198210245',
                                                                                                                  'in_ucast_pkts': '992076266',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27327145',
                                                                                                                  'out_octets': '3855359549533',
                                                                                                                  'out_ucast_pkts': '2825199525',
                                                                                                                  'poller_hash': 'ODg5QzFwZWVyaW50ZnBvbiA2NzQ='},
                                                                                                          {       'in_nucast_pkts': '233',
                                                                                                                  'in_octets': '9605099303',
                                                                                                                  'in_ucast_pkts': '18425762',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '35074366',
                                                                                                                  'out_octets': '38064517457',
                                                                                                                  'out_ucast_pkts': '32039598',
                                                                                                                  'poller_hash': 'MzVFQjI0cGVlcmludGZwb24gNjc2'},
                                                                                                          {       'in_nucast_pkts': '14569',
                                                                                                                  'in_octets': '9380725467',
                                                                                                                  'in_ucast_pkts': '39437157',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119715',
                                                                                                                  'out_octets': '447196947757',
                                                                                                                  'out_ucast_pkts': '301503632',
                                                                                                                  'poller_hash': 'OEMzMDBwZWVyaW50ZnBvbiA2Nzc='},
                                                                                                          {       'in_nucast_pkts': '160894',
                                                                                                                  'in_octets': '17668306125',
                                                                                                                  'in_ucast_pkts': '205822117',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14151931',
                                                                                                                  'out_octets': '1271636992737',
                                                                                                                  'out_ucast_pkts': '845839190',
                                                                                                                  'poller_hash': 'QzU3M0VwZWVyaW50ZnBvbiA2Nzk='},
                                                                                                          {       'in_nucast_pkts': '41',
                                                                                                                  'in_octets': '5332439266',
                                                                                                                  'in_ucast_pkts': '7726935',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19517',
                                                                                                                  'out_octets': '7071338505',
                                                                                                                  'out_ucast_pkts': '7316591',
                                                                                                                  'poller_hash': 'QzU4NDlwZWVyaW50ZnBvbiA2ODA='},
                                                                                                          {       'in_nucast_pkts': '14780',
                                                                                                                  'in_octets': '15103265123',
                                                                                                                  'in_ucast_pkts': '26249140',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119705',
                                                                                                                  'out_octets': '65221456163',
                                                                                                                  'out_ucast_pkts': '60039058',
                                                                                                                  'poller_hash': 'Q0MwOUFwZWVyaW50ZnBvbiA2ODE='},
                                                                                                          {       'in_nucast_pkts': '32860',
                                                                                                                  'in_octets': '18711742461',
                                                                                                                  'in_ucast_pkts': '32834528',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120171',
                                                                                                                  'out_octets': '161031603161',
                                                                                                                  'out_ucast_pkts': '124175644',
                                                                                                                  'poller_hash': 'Q0MwNzVwZWVyaW50ZnBvbiA2ODI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTMwMDU4cGVlcmludGZwb24gNjgz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '470322831826',
                                                                                                                  'in_ucast_pkts': '5408363977',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '19230406529003',
                                                                                                                  'out_ucast_pkts': '12949695649',
                                                                                                                  'poller_hash': 'MTI0ODNGcGVlcmludGZwb24gNjg0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '69331969638',
                                                                                                                  'in_ucast_pkts': '235154476',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '790918907861',
                                                                                                                  'out_ucast_pkts': '726768307',
                                                                                                                  'poller_hash': 'ODg5RTZwZWVyaW50ZnBvbiA2ODU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '231374831298',
                                                                                                                  'in_ucast_pkts': '1738248115',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9457434507617',
                                                                                                                  'out_ucast_pkts': '6376420501',
                                                                                                                  'poller_hash': 'MTU1NDlGcGVlcmludGZwb24gNjg3'},
                                                                                                          {       'in_nucast_pkts': '94563',
                                                                                                                  'in_octets': '97704747022',
                                                                                                                  'in_ucast_pkts': '316212198',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83086810',
                                                                                                                  'out_octets': '1594483155932',
                                                                                                                  'out_ucast_pkts': '1295561950',
                                                                                                                  'poller_hash': 'MzVGOUIwcGVlcmludGZwb24gNzM2'},
                                                                                                          {       'in_nucast_pkts': '210',
                                                                                                                  'in_octets': '3988108755',
                                                                                                                  'in_ucast_pkts': '35555317',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119791',
                                                                                                                  'out_octets': '148546300735',
                                                                                                                  'out_ucast_pkts': '105194226',
                                                                                                                  'poller_hash': 'NjQwQUI2cGVlcmludGZwb24gODU5'},
                                                                                                          {       'in_nucast_pkts': '296',
                                                                                                                  'in_octets': '73920918660',
                                                                                                                  'in_ucast_pkts': '346618548',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2158',
                                                                                                                  'out_octets': '1166960350580',
                                                                                                                  'out_ucast_pkts': '899535896',
                                                                                                                  'poller_hash': 'Q0JFMDVCcGVlcmludGZwb24gMTcwNDY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '419510648922',
                                                                                                                  'in_ucast_pkts': '66681446254',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9019318618171',
                                                                                                                  'out_ucast_pkts': '6414552146',
                                                                                                                  'poller_hash': 'QzgxNUM4cGVlcmludGZwb24gMjAwNDUw'},
                                                                                                          {       'in_nucast_pkts': '11948',
                                                                                                                  'in_octets': '4273869246',
                                                                                                                  'in_ucast_pkts': '21737144',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1459633',
                                                                                                                  'out_octets': '77225665481',
                                                                                                                  'out_ucast_pkts': '63261385',
                                                                                                                  'poller_hash': 'MTI3NDlFQ3BlZXJpbnRmcG9uIDIzNDQwNw=='},
                                                                                                          {       'in_nucast_pkts': '18569',
                                                                                                                  'in_octets': '4136840327',
                                                                                                                  'in_ucast_pkts': '20909060',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '30466431',
                                                                                                                  'out_octets': '11569408346',
                                                                                                                  'out_ucast_pkts': '11226543',
                                                                                                                  'poller_hash': 'MTE5NEI4QnBlZXJpbnRmcG9uIDI3NDA1MQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '143527403135',
                                                                                         'out_octets': '424627299555121',
                                                                                         'out_ucast_pkts': '184020980752',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/6',
                                                                                                                 'id': '6/6',
                                                                                                                 'links': [       {       'connected_time': '15498304',
                                                                                                                                          'fiber_length': 1008,
                                                                                                                                          'poller_hash': 'MUZENkI3cGVlcnBvbnBvbiA2NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '8818315',
                                                                                                                                          'fiber_length': 1023,
                                                                                                                                          'poller_hash': 'MUQwNkNEcGVlcnBvbnBvbiA2NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '15498560',
                                                                                                                                          'fiber_length': 987,
                                                                                                                                          'poller_hash': 'ODg5QzFwZWVycG9ucG9uIDY3NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '15499171',
                                                                                                                                          'fiber_length': 1624,
                                                                                                                                          'poller_hash': 'MzVFQjI0cGVlcnBvbnBvbiA2NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '688432',
                                                                                                                                          'fiber_length': 916,
                                                                                                                                          'poller_hash': 'OEMzMDBwZWVycG9ucG9uIDY3Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '7712321',
                                                                                                                                          'fiber_length': 953,
                                                                                                                                          'poller_hash': 'QzU3M0VwZWVycG9ucG9uIDY3OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '12347',
                                                                                                                                          'fiber_length': 935,
                                                                                                                                          'poller_hash': 'QzU4NDlwZWVycG9ucG9uIDY4MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '688431',
                                                                                                                                          'fiber_length': 936,
                                                                                                                                          'poller_hash': 'Q0MwOUFwZWVycG9ucG9uIDY4MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '688418',
                                                                                                                                          'fiber_length': 998,
                                                                                                                                          'poller_hash': 'Q0MwNzVwZWVycG9ucG9uIDY4Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '15498593',
                                                                                                                                          'fiber_length': 1008,
                                                                                                                                          'poller_hash': 'MTMwMDU4cGVlcnBvbnBvbiA2ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '46993',
                                                                                                                                          'fiber_length': 1655,
                                                                                                                                          'poller_hash': 'MTI0ODNGcGVlcnBvbnBvbiA2ODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '47637',
                                                                                                                                          'fiber_length': 884,
                                                                                                                                          'poller_hash': 'ODg5RTZwZWVycG9ucG9uIDY4NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '65341989',
                                                                                                                                          'fiber_length': 1995,
                                                                                                                                          'poller_hash': 'MTU1NDlGcGVlcnBvbnBvbiA2ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '67790420',
                                                                                                                                          'fiber_length': 1047,
                                                                                                                                          'poller_hash': 'MzVGOUIwcGVlcnBvbnBvbiA3MzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '688442',
                                                                                                                                          'fiber_length': 1550,
                                                                                                                                          'poller_hash': 'NjQwQUI2cGVlcnBvbnBvbiA4NTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '6067696',
                                                                                                                                          'fiber_length': 1004,
                                                                                                                                          'poller_hash': 'Q0JFMDVCcGVlcnBvbnBvbiAxNzA0NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '687935',
                                                                                                                                          'fiber_length': 1640,
                                                                                                                                          'poller_hash': 'QzgxNUM4cGVlcnBvbnBvbiAyMDA0NTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '686551',
                                                                                                                                          'fiber_length': 1040,
                                                                                                                                          'poller_hash': 'MTI3NDlFQ3BlZXJwb25wb24gMjM0NDA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723},
                                                                                                                                  {       'connected_time': '13912989',
                                                                                                                                          'fiber_length': 1648,
                                                                                                                                          'poller_hash': 'MTE5NEI4QnBlZXJwb25wb24gMjc0MDUx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.723}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/6',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/7',
                                                                                         'in_discards': '196123',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1180624924',
                                                                                         'in_octets': '70594451101421',
                                                                                         'in_ucast_pkts': '177603083163',
                                                                                         'links': [       {       'in_nucast_pkts': '2879052',
                                                                                                                  'in_octets': '126270120057',
                                                                                                                  'in_ucast_pkts': '1609187359',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '198450959',
                                                                                                                  'out_octets': '8960602386217',
                                                                                                                  'out_ucast_pkts': '6326174385',
                                                                                                                  'poller_hash': 'ODhBMDBwZWVyaW50ZnBvbiA5MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '659953516948',
                                                                                                                  'in_ucast_pkts': '5243152427',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '15742330115948',
                                                                                                                  'out_ucast_pkts': '11238861341',
                                                                                                                  'poller_hash': 'ODhBMDVwZWVyaW50ZnBvbiA5Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '547796548079',
                                                                                                                  'in_ucast_pkts': '2203764737',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6655217345448',
                                                                                                                  'out_ucast_pkts': '5540578763',
                                                                                                                  'poller_hash': 'ODg5NzFwZWVyaW50ZnBvbiA5Mw=='},
                                                                                                          {       'in_nucast_pkts': '6393994',
                                                                                                                  'in_octets': '53184044507642',
                                                                                                                  'in_ucast_pkts': '107950796789',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '365350679',
                                                                                                                  'out_octets': '174768465404854',
                                                                                                                  'out_ucast_pkts': '148815009960',
                                                                                                                  'poller_hash': 'OUNBMERwZWVyaW50ZnBvbiA5NA=='},
                                                                                                          {       'in_nucast_pkts': '72',
                                                                                                                  'in_octets': '619063529632',
                                                                                                                  'in_ucast_pkts': '709474160',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27327588',
                                                                                                                  'out_octets': '1507775562569',
                                                                                                                  'out_ucast_pkts': '1313229537',
                                                                                                                  'poller_hash': 'ODg5RDFwZWVyaW50ZnBvbiA5NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '190803940818',
                                                                                                                  'in_ucast_pkts': '943963563',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3871253027287',
                                                                                                                  'out_ucast_pkts': '2939674429',
                                                                                                                  'poller_hash': 'ODhBMUZwZWVyaW50ZnBvbiA5Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '928196148639',
                                                                                                                  'in_ucast_pkts': '2003126254',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6666048656853',
                                                                                                                  'out_ucast_pkts': '5184696467',
                                                                                                                  'poller_hash': 'ODg5MjlwZWVyaW50ZnBvbiA5Nw=='},
                                                                                                          {       'in_nucast_pkts': '45698122',
                                                                                                                  'in_octets': '693591117403',
                                                                                                                  'in_ucast_pkts': '2418545542',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '383156522',
                                                                                                                  'out_octets': '8352076267073',
                                                                                                                  'out_ucast_pkts': '6516909192',
                                                                                                                  'poller_hash': 'QUM0NEVwZWVyaW50ZnBvbiA5OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '337034127590',
                                                                                                                  'in_ucast_pkts': '2115836914',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9816409719645',
                                                                                                                  'out_ucast_pkts': '7040273861',
                                                                                                                  'poller_hash': 'Q0MwOTVwZWVyaW50ZnBvbiA5OQ=='},
                                                                                                          {       'in_nucast_pkts': '329',
                                                                                                                  'in_octets': '9585366876',
                                                                                                                  'in_ucast_pkts': '25926697',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463801',
                                                                                                                  'out_octets': '108359475746',
                                                                                                                  'out_ucast_pkts': '81922903',
                                                                                                                  'poller_hash': 'Q0JEMjlEcGVlcmludGZwb24gMTAw'},
                                                                                                          {       'in_nucast_pkts': '14583',
                                                                                                                  'in_octets': '2894310630',
                                                                                                                  'in_ucast_pkts': '18263789',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119648',
                                                                                                                  'out_octets': '69361737664',
                                                                                                                  'out_ucast_pkts': '55198079',
                                                                                                                  'poller_hash': 'MTMwMEM1cGVlcmludGZwb24gMTAx'},
                                                                                                          {       'in_nucast_pkts': '14897',
                                                                                                                  'in_octets': '3135998675',
                                                                                                                  'in_ucast_pkts': '11577091',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119639',
                                                                                                                  'out_octets': '46148367339',
                                                                                                                  'out_ucast_pkts': '37031128',
                                                                                                                  'poller_hash': 'MTMwMDA1cGVlcmludGZwb24gMTAy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTMwMDlCcGVlcmludGZwb24gMTAz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '264265169654',
                                                                                                                  'in_ucast_pkts': '1434016587',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '8327046667144',
                                                                                                                  'out_ucast_pkts': '6004100854',
                                                                                                                  'poller_hash': 'MTJGRkZBcGVlcmludGZwb24gMTA0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '330442278135',
                                                                                                                  'in_ucast_pkts': '48089720924',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '128',
                                                                                                                  'out_octets': '2142803471622',
                                                                                                                  'out_ucast_pkts': '1819497319',
                                                                                                                  'poller_hash': 'MUQwNEVDcGVlcmludGZwb24gMTg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '245433310705',
                                                                                                                  'in_ucast_pkts': '1446998907',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '222',
                                                                                                                  'out_octets': '8094891974565',
                                                                                                                  'out_ucast_pkts': '5746842563',
                                                                                                                  'poller_hash': 'MkFEMEYxcGVlcmludGZwb24gMzAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '12205938781',
                                                                                                                  'in_ucast_pkts': '63058703',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '302598056283',
                                                                                                                  'out_ucast_pkts': '270496190',
                                                                                                                  'poller_hash': 'MkIwMTg2cGVlcmludGZwb24gNDM0'},
                                                                                                          {       'in_nucast_pkts': '4473',
                                                                                                                  'in_octets': '26727962413',
                                                                                                                  'in_ucast_pkts': '100371554',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465319',
                                                                                                                  'out_octets': '286926074231',
                                                                                                                  'out_ucast_pkts': '211824315',
                                                                                                                  'poller_hash': 'MzAwN0M1cGVlcmludGZwb24gNDQ5'},
                                                                                                          {       'in_nucast_pkts': '4432',
                                                                                                                  'in_octets': '144929379647',
                                                                                                                  'in_ucast_pkts': '589831033',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '35075152',
                                                                                                                  'out_octets': '3605241988876',
                                                                                                                  'out_ucast_pkts': '2699736265',
                                                                                                                  'poller_hash': 'MzVFOTY4cGVlcmludGZwb24gNTA4'},
                                                                                                          {       'in_nucast_pkts': '201',
                                                                                                                  'in_octets': '2601641211',
                                                                                                                  'in_ucast_pkts': '14541117',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119502',
                                                                                                                  'out_octets': '41748200566',
                                                                                                                  'out_ucast_pkts': '32645653',
                                                                                                                  'poller_hash': 'NjQwQjQ2cGVlcmludGZwb24gODY2'},
                                                                                                          {       'in_nucast_pkts': '401',
                                                                                                                  'in_octets': '6455753157',
                                                                                                                  'in_ucast_pkts': '68534506',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463929',
                                                                                                                  'out_octets': '248389225226',
                                                                                                                  'out_ucast_pkts': '169096743',
                                                                                                                  'poller_hash': 'NUE2MjFCcGVlcmludGZwb24gODY4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QTlDOTg2cGVlcmludGZwb24gODc5'},
                                                                                                          {       'in_nucast_pkts': '697167',
                                                                                                                  'in_octets': '641987478049',
                                                                                                                  'in_ucast_pkts': '1082782708',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83570468',
                                                                                                                  'out_octets': '3731133087841',
                                                                                                                  'out_ucast_pkts': '2955119879',
                                                                                                                  'poller_hash': 'NUE2QzlEcGVlcmludGZwb24gMTM3Njkz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '522220537689',
                                                                                                                  'in_ucast_pkts': '26395661214',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '1318373976056',
                                                                                                                  'out_ucast_pkts': '1223913040',
                                                                                                                  'poller_hash': 'QTlDNjc1cGVlcmludGZwb24gMTQ3MTY4'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '3368802544',
                                                                                                                  'in_ucast_pkts': '14583234',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465082',
                                                                                                                  'out_octets': '107574952194',
                                                                                                                  'out_ucast_pkts': '78602497',
                                                                                                                  'poller_hash': 'NDJDNDY3cGVlcmludGZwb24gMjU1ODI5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1010925905',
                                                                                                                  'in_ucast_pkts': '8208897',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '23663915731',
                                                                                                                  'out_ucast_pkts': '34940632',
                                                                                                                  'poller_hash': 'MTI5NzQyNnBlZXJpbnRmcG9uIDI2NDE1OQ=='},
                                                                                                          {       'in_nucast_pkts': '5506833',
                                                                                                                  'in_octets': '219034856077',
                                                                                                                  'in_ucast_pkts': '819600521',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '30673258',
                                                                                                                  'out_octets': '4213637549858',
                                                                                                                  'out_ucast_pkts': '2967684438',
                                                                                                                  'poller_hash': 'MTUyMDJDRnBlZXJpbnRmcG9uIDI2NDQ5NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '41856067008',
                                                                                         'out_octets': '454316781339363',
                                                                                         'out_ucast_pkts': '321998561855',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/7',
                                                                                                                 'id': '6/7',
                                                                                                                 'links': [       {       'connected_time': '29485979',
                                                                                                                                          'fiber_length': 977,
                                                                                                                                          'poller_hash': 'ODhBMDBwZWVycG9ucG9uIDkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '22050262',
                                                                                                                                          'fiber_length': 751,
                                                                                                                                          'poller_hash': 'ODhBMDVwZWVycG9ucG9uIDky',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688424',
                                                                                                                                          'fiber_length': 785,
                                                                                                                                          'poller_hash': 'ODg5NzFwZWVycG9ucG9uIDkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '5988529',
                                                                                                                                          'fiber_length': 1327,
                                                                                                                                          'poller_hash': 'OUNBMERwZWVycG9ucG9uIDk0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '15498546',
                                                                                                                                          'fiber_length': 1028,
                                                                                                                                          'poller_hash': 'ODg5RDFwZWVycG9ucG9uIDk1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '47526',
                                                                                                                                          'fiber_length': 815,
                                                                                                                                          'poller_hash': 'ODhBMUZwZWVycG9ucG9uIDk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '47388',
                                                                                                                                          'fiber_length': 953,
                                                                                                                                          'poller_hash': 'ODg5MjlwZWVycG9ucG9uIDk3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '72961103',
                                                                                                                                          'fiber_length': 897,
                                                                                                                                          'poller_hash': 'QUM0NEVwZWVycG9ucG9uIDk4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '1025287',
                                                                                                                                          'fiber_length': 825,
                                                                                                                                          'poller_hash': 'Q0MwOTVwZWVycG9ucG9uIDk5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688444',
                                                                                                                                          'fiber_length': 1184,
                                                                                                                                          'poller_hash': 'Q0JEMjlEcGVlcnBvbnBvbiAxMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688438',
                                                                                                                                          'fiber_length': 1161,
                                                                                                                                          'poller_hash': 'MTMwMEM1cGVlcnBvbnBvbiAxMDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688438',
                                                                                                                                          'fiber_length': 1075,
                                                                                                                                          'poller_hash': 'MTMwMDA1cGVlcnBvbnBvbiAxMDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '8385541',
                                                                                                                                          'fiber_length': 1065,
                                                                                                                                          'poller_hash': 'MTMwMDlCcGVlcnBvbnBvbiAxMDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '47551',
                                                                                                                                          'fiber_length': 1241,
                                                                                                                                          'poller_hash': 'MTJGRkZBcGVlcnBvbnBvbiAxMDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688453',
                                                                                                                                          'fiber_length': 1208,
                                                                                                                                          'poller_hash': 'MUQwNEVDcGVlcnBvbnBvbiAxODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688451',
                                                                                                                                          'fiber_length': 830,
                                                                                                                                          'poller_hash': 'MkFEMEYxcGVlcnBvbnBvbiAzMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '2144588',
                                                                                                                                          'fiber_length': 1186,
                                                                                                                                          'poller_hash': 'MkIwMTg2cGVlcnBvbnBvbiA0MzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688451',
                                                                                                                                          'fiber_length': 1170,
                                                                                                                                          'poller_hash': 'MzAwN0M1cGVlcnBvbnBvbiA0NDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '15499141',
                                                                                                                                          'fiber_length': 1213,
                                                                                                                                          'poller_hash': 'MzVFOTY4cGVlcnBvbnBvbiA1MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688447',
                                                                                                                                          'fiber_length': 1235,
                                                                                                                                          'poller_hash': 'NjQwQjQ2cGVlcnBvbnBvbiA4NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688466',
                                                                                                                                          'fiber_length': 1301,
                                                                                                                                          'poller_hash': 'NUE2MjFCcGVlcnBvbnBvbiA4Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688450',
                                                                                                                                          'fiber_length': 291,
                                                                                                                                          'poller_hash': 'QTlDOTg2cGVlcnBvbnBvbiA4Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '67790242',
                                                                                                                                          'fiber_length': 1210,
                                                                                                                                          'poller_hash': 'NUE2QzlEcGVlcnBvbnBvbiAxMzc2OTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688445',
                                                                                                                                          'fiber_length': 975,
                                                                                                                                          'poller_hash': 'QTlDNjc1cGVlcnBvbnBvbiAxNDcxNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688454',
                                                                                                                                          'fiber_length': 882,
                                                                                                                                          'poller_hash': 'NDJDNDY3cGVlcnBvbnBvbiAyNTU4Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688447',
                                                                                                                                          'fiber_length': 1013,
                                                                                                                                          'poller_hash': 'MTI5NzQyNnBlZXJwb25wb24gMjY0MTU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673},
                                                                                                                                  {       'connected_time': '688352',
                                                                                                                                          'fiber_length': 1017,
                                                                                                                                          'poller_hash': 'MTUyMDJDRnBlZXJwb25wb24gMjY0NDk0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.673}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/7',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/8',
                                                                                         'in_discards': '164983',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5258270',
                                                                                         'in_octets': '13332487121016',
                                                                                         'in_ucast_pkts': '56577246020',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MUQwNTAzcGVlcmludGZwb24gMjA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1229653996542',
                                                                                                                  'in_ucast_pkts': '19690654995',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '14066611180403',
                                                                                                                  'out_ucast_pkts': '10406457495',
                                                                                                                  'poller_hash': 'MUNDNEIzcGVlcmludGZwb24gMjI='},
                                                                                                          {       'in_nucast_pkts': '11998',
                                                                                                                  'in_octets': '1301596135',
                                                                                                                  'in_ucast_pkts': '7106502',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463574',
                                                                                                                  'out_octets': '77772694304',
                                                                                                                  'out_ucast_pkts': '54485199',
                                                                                                                  'poller_hash': 'MUQwMzQ3cGVlcmludGZwb24gMjY='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '542240122260',
                                                                                                                  'in_ucast_pkts': '135603381515',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '9085026882378',
                                                                                                                  'out_ucast_pkts': '6819809591',
                                                                                                                  'poller_hash': 'MUQwNTAwcGVlcmludGZwb24gNDc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '231449039856',
                                                                                                                  'in_ucast_pkts': '836425364',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '3190140125790',
                                                                                                                  'out_ucast_pkts': '2544346770',
                                                                                                                  'poller_hash': 'MUNBQjU2cGVlcmludGZwb24gNDk='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '50932380779',
                                                                                                                  'in_ucast_pkts': '471597244',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '126',
                                                                                                                  'out_octets': '1594737011383',
                                                                                                                  'out_ucast_pkts': '1165086850',
                                                                                                                  'poller_hash': 'MUQwNTA0cGVlcmludGZwb24gNjI='},
                                                                                                          {       'in_nucast_pkts': '536',
                                                                                                                  'in_octets': '6808607997',
                                                                                                                  'in_ucast_pkts': '21328169',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1438965',
                                                                                                                  'out_octets': '69572157072',
                                                                                                                  'out_ucast_pkts': '51768969',
                                                                                                                  'poller_hash': 'MjhGNjY0cGVlcmludGZwb24gNzg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '393866565684',
                                                                                                                  'in_ucast_pkts': '1932532367',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '222',
                                                                                                                  'out_octets': '5586345244993',
                                                                                                                  'out_ucast_pkts': '4098617665',
                                                                                                                  'poller_hash': 'Mjk4MjVGcGVlcmludGZwb24gMTMy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '560639487862',
                                                                                                                  'in_ucast_pkts': '50283447007',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '160',
                                                                                                                  'out_octets': '13204715801269',
                                                                                                                  'out_ucast_pkts': '9509533623',
                                                                                                                  'poller_hash': 'MUQwNjlCcGVlcmludGZwb24gMTQ5'},
                                                                                                          {       'in_nucast_pkts': '14701',
                                                                                                                  'in_octets': '5827260778',
                                                                                                                  'in_ucast_pkts': '24190997',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5000311342',
                                                                                                                  'out_octets': '6710131578104',
                                                                                                                  'out_ucast_pkts': '55388834',
                                                                                                                  'poller_hash': 'REZGQjU5cGVlcmludGZwb24gMjE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1293906236540',
                                                                                                                  'in_ucast_pkts': '3050171251',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '64',
                                                                                                                  'out_octets': '10621866392874',
                                                                                                                  'out_ucast_pkts': '8936934451',
                                                                                                                  'poller_hash': 'Mjk4NDM5cGVlcmludGZwb24gMjU0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MkIwMjk0cGVlcmludGZwb24gMjY3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '120799275468',
                                                                                                                  'in_ucast_pkts': '17680076177',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '126',
                                                                                                                  'out_octets': '2133605010264',
                                                                                                                  'out_ucast_pkts': '1839589338',
                                                                                                                  'poller_hash': 'MkIwNDUwcGVlcmludGZwb24gNDE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NEY5QjM3cGVlcmludGZwb24gNTMy'},
                                                                                                          {       'in_nucast_pkts': '265',
                                                                                                                  'in_octets': '273835599909',
                                                                                                                  'in_ucast_pkts': '1618089679',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83574339',
                                                                                                                  'out_octets': '6098986305922',
                                                                                                                  'out_ucast_pkts': '4311082233',
                                                                                                                  'poller_hash': 'MzVGOUIxcGVlcmludGZwb24gNjg4'},
                                                                                                          {       'in_nucast_pkts': '65454',
                                                                                                                  'in_octets': '1041713716',
                                                                                                                  'in_ucast_pkts': '5246522',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463946',
                                                                                                                  'out_octets': '28509932539',
                                                                                                                  'out_ucast_pkts': '22667750',
                                                                                                                  'poller_hash': 'NUE2NjJEcGVlcmludGZwb24gODA3'},
                                                                                                          {       'in_nucast_pkts': '859645',
                                                                                                                  'in_octets': '48473691804',
                                                                                                                  'in_ucast_pkts': '660389071',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '118286011',
                                                                                                                  'out_octets': '3678250237079',
                                                                                                                  'out_ucast_pkts': '2586029073',
                                                                                                                  'poller_hash': 'NDJDNzBBcGVlcmludGZwb24gODIw'},
                                                                                                          {       'in_nucast_pkts': '12060',
                                                                                                                  'in_octets': '16182047720',
                                                                                                                  'in_ucast_pkts': '174393578',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1497818',
                                                                                                                  'out_octets': '566883551510',
                                                                                                                  'out_ucast_pkts': '391140008',
                                                                                                                  'poller_hash': 'QzgxRDg1cGVlcmludGZwb24gMjExNjEz'},
                                                                                                          {       'in_nucast_pkts': '25437511',
                                                                                                                  'in_octets': '97021251620',
                                                                                                                  'in_ucast_pkts': '630938678',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22211568718',
                                                                                                                  'out_octets': '35025298874024',
                                                                                                                  'out_ucast_pkts': '4111661475',
                                                                                                                  'poller_hash': 'MkVEOUM0cGVlcmludGZwb24gMjE3MTQw'},
                                                                                                          {       'in_nucast_pkts': '12159',
                                                                                                                  'in_octets': '278595104',
                                                                                                                  'in_ucast_pkts': '1046233',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464490',
                                                                                                                  'out_octets': '10090293135',
                                                                                                                  'out_ucast_pkts': '7394441',
                                                                                                                  'poller_hash': 'MTI3NEI1QXBlZXJpbnRmcG9uIDIzMTYyMA=='},
                                                                                                          {       'in_nucast_pkts': '642083',
                                                                                                                  'in_octets': '180138581458',
                                                                                                                  'in_ucast_pkts': '1883213417',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82419359',
                                                                                                                  'out_octets': '16655854396546',
                                                                                                                  'out_ucast_pkts': '11159138337',
                                                                                                                  'poller_hash': 'MTI3NDMyOXBlZXJpbnRmcG9uIDIzNTQxOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '212958900526',
                                                                                                                  'in_ucast_pkts': '881871380',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '64',
                                                                                                                  'out_octets': '3272129119599',
                                                                                                                  'out_ucast_pkts': '2402602926',
                                                                                                                  'poller_hash': 'Mjk2RkVCcGVlcmludGZwb24gMjYyNjk4'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '6/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '150769264758',
                                                                                         'out_octets': '432582212777690',
                                                                                         'out_ucast_pkts': '168686474285',
                                                                                         'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/8',
                                                                                                                 'id': '6/8',
                                                                                                                 'links': [       {       'connected_time': '15499140',
                                                                                                                                          'fiber_length': 1192,
                                                                                                                                          'poller_hash': 'MUQwNTAzcGVlcnBvbnBvbiAyMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '72953878',
                                                                                                                                          'fiber_length': 1473,
                                                                                                                                          'poller_hash': 'MUNDNEIzcGVlcnBvbnBvbiAyMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688456',
                                                                                                                                          'fiber_length': 1199,
                                                                                                                                          'poller_hash': 'MUQwMzQ3cGVlcnBvbnBvbiAyNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '72476147',
                                                                                                                                          'fiber_length': 1177,
                                                                                                                                          'poller_hash': 'MUQwNTAwcGVlcnBvbnBvbiA0Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '29272053',
                                                                                                                                          'fiber_length': 1190,
                                                                                                                                          'poller_hash': 'MUNBQjU2cGVlcnBvbnBvbiA0OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '73621552',
                                                                                                                                          'fiber_length': 1458,
                                                                                                                                          'poller_hash': 'MUQwNTA0cGVlcnBvbnBvbiA2Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '677139',
                                                                                                                                          'fiber_length': 1312,
                                                                                                                                          'poller_hash': 'MjhGNjY0cGVlcnBvbnBvbiA3OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '19814845',
                                                                                                                                          'fiber_length': 1496,
                                                                                                                                          'poller_hash': 'Mjk4MjVGcGVlcnBvbnBvbiAxMzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688461',
                                                                                                                                          'fiber_length': 1398,
                                                                                                                                          'poller_hash': 'MUQwNjlCcGVlcnBvbnBvbiAxNDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688452',
                                                                                                                                          'fiber_length': 1533,
                                                                                                                                          'poller_hash': 'REZGQjU5cGVlcnBvbnBvbiAyMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '67850913',
                                                                                                                                          'fiber_length': 1214,
                                                                                                                                          'poller_hash': 'Mjk4NDM5cGVlcnBvbnBvbiAyNTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '21815361',
                                                                                                                                          'fiber_length': 1291,
                                                                                                                                          'poller_hash': 'MkIwMjk0cGVlcnBvbnBvbiAyNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '26779881',
                                                                                                                                          'fiber_length': 1306,
                                                                                                                                          'poller_hash': 'MkIwNDUwcGVlcnBvbnBvbiA0MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688446',
                                                                                                                                          'fiber_length': 1372,
                                                                                                                                          'poller_hash': 'NEY5QjM3cGVlcnBvbnBvbiA1MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '67790223',
                                                                                                                                          'fiber_length': 1154,
                                                                                                                                          'poller_hash': 'MzVGOUIxcGVlcnBvbnBvbiA2ODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688467',
                                                                                                                                          'fiber_length': 1487,
                                                                                                                                          'poller_hash': 'NUE2NjJEcGVlcnBvbnBvbiA4MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '15498764',
                                                                                                                                          'fiber_length': 1422,
                                                                                                                                          'poller_hash': 'NDJDNzBBcGVlcnBvbnBvbiA4MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688455',
                                                                                                                                          'fiber_length': 1337,
                                                                                                                                          'poller_hash': 'QzgxRDg1cGVlcnBvbnBvbiAyMTE2MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '41473',
                                                                                                                                          'fiber_length': 1109,
                                                                                                                                          'poller_hash': 'MkVEOUM0cGVlcnBvbnBvbiAyMTcxNDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688452',
                                                                                                                                          'fiber_length': 1363,
                                                                                                                                          'poller_hash': 'MTI3NEI1QXBlZXJwb25wb24gMjMxNjIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '67333726',
                                                                                                                                          'fiber_length': 1361,
                                                                                                                                          'poller_hash': 'MTI3NDMyOXBlZXJwb25wb24gMjM1NDE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '688455',
                                                                                                                                          'fiber_length': 1322,
                                                                                                                                          'poller_hash': 'Mjk2RkVCcGVlcnBvbnBvbiAyNjI2OTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/8',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNjEyNjA2Nzg5',
                                                           'serial_number': '211612606789',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42990963'},
                                                   {       'id': '7',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '7/1',
                                                                                         'in_discards': '1386',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7639915',
                                                                                         'in_octets': '4471513796790',
                                                                                         'in_ucast_pkts': '29770767165',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '349251449737',
                                                                                                                  'in_ucast_pkts': '4236197465',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '25419667488201',
                                                                                                                  'out_ucast_pkts': '16886663448',
                                                                                                                  'poller_hash': 'MTU1M0Q0cGVlcmludGZwb24gMzQ='},
                                                                                                          {       'in_nucast_pkts': '733528',
                                                                                                                  'in_octets': '72368341435',
                                                                                                                  'in_ucast_pkts': '941931829',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '101203936',
                                                                                                                  'out_octets': '5164641026899',
                                                                                                                  'out_ucast_pkts': '3646209021',
                                                                                                                  'poller_hash': 'MTU1MzIycGVlcmludGZwb24gMzU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '184801535871',
                                                                                                                  'in_ucast_pkts': '432776860',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3271921717978',
                                                                                                                  'out_ucast_pkts': '2491558784',
                                                                                                                  'poller_hash': 'MTc1RTZBcGVlcmludGZwb24gMzY='},
                                                                                                          {       'in_nucast_pkts': '256166',
                                                                                                                  'in_octets': '33853620148',
                                                                                                                  'in_ucast_pkts': '314071235',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23533891',
                                                                                                                  'out_octets': '2012667109402',
                                                                                                                  'out_ucast_pkts': '1347545303',
                                                                                                                  'poller_hash': 'MTU1NDM2cGVlcmludGZwb24gMzc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '322585475891',
                                                                                                                  'in_ucast_pkts': '3487590285',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '10290881112838',
                                                                                                                  'out_ucast_pkts': '2679847522',
                                                                                                                  'poller_hash': 'MTc1RTcycGVlcmludGZwb24gMzg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTc2MjM5cGVlcmludGZwb24gMzk='},
                                                                                                          {       'in_nucast_pkts': '260480',
                                                                                                                  'in_octets': '3936917756',
                                                                                                                  'in_ucast_pkts': '13624731',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23541529',
                                                                                                                  'out_octets': '45945615152',
                                                                                                                  'out_ucast_pkts': '38363358',
                                                                                                                  'poller_hash': 'Q0MwQzVwZWVyaW50ZnBvbiA0MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '113089914963',
                                                                                                                  'in_ucast_pkts': '1228128345',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6749315557507',
                                                                                                                  'out_ucast_pkts': '4638549040',
                                                                                                                  'poller_hash': 'MUM1Q0U5cGVlcmludGZwb24gNDE='},
                                                                                                          {       'in_nucast_pkts': '3388497',
                                                                                                                  'in_octets': '95266532404',
                                                                                                                  'in_ucast_pkts': '1186137348',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '163153270',
                                                                                                                  'out_octets': '6819039624998',
                                                                                                                  'out_ucast_pkts': '4602117467',
                                                                                                                  'poller_hash': 'MUM1Q0I3cGVlcmludGZwb24gNTU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '71245945058',
                                                                                                                  'in_ucast_pkts': '918109127',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '5803799180157',
                                                                                                                  'out_ucast_pkts': '3855394347',
                                                                                                                  'poller_hash': 'MkIwMjY0cGVlcmludGZwb24gMjcw'},
                                                                                                          {       'in_nucast_pkts': '5366134',
                                                                                                                  'in_octets': '174825717150',
                                                                                                                  'in_ucast_pkts': '1966286800',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '182095064',
                                                                                                                  'out_octets': '6945684513527',
                                                                                                                  'out_ucast_pkts': '4906281245',
                                                                                                                  'poller_hash': 'MjA1REM4cGVlcmludGZwb24gNDc0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '26345085674',
                                                                                                                  'in_ucast_pkts': '98843665',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '332285039955',
                                                                                                                  'out_ucast_pkts': '308682269',
                                                                                                                  'poller_hash': 'REZGREZBcGVlcmludGZwb24gMTcxODQ3'},
                                                                                                          {       'in_nucast_pkts': '21409',
                                                                                                                  'in_octets': '390089658784',
                                                                                                                  'in_ucast_pkts': '1036808211',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27344180',
                                                                                                                  'out_octets': '2096376198503',
                                                                                                                  'out_ucast_pkts': '2066235723',
                                                                                                                  'poller_hash': 'MjYwRTk2cGVlcmludGZwb24gMTk0MTA2'},
                                                                                                          {       'in_nucast_pkts': '293',
                                                                                                                  'in_octets': '1062250815533',
                                                                                                                  'in_ucast_pkts': '75373616',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1942',
                                                                                                                  'out_octets': '18196212933459',
                                                                                                                  'out_ucast_pkts': '14554307805',
                                                                                                                  'poller_hash': 'MTI3NDJFN3BlZXJpbnRmcG9uIDI0ODAzOA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '3753171298',
                                                                                         'out_octets': '137042381622806',
                                                                                         'out_ucast_pkts': '95780769729',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/1',
                                                                                                                 'id': '7/1',
                                                                                                                 'links': [       {       'connected_time': '13219168',
                                                                                                                                          'fiber_length': 1576,
                                                                                                                                          'poller_hash': 'MTU1M0Q0cGVlcnBvbnBvbiAzNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '13219219',
                                                                                                                                          'fiber_length': 1630,
                                                                                                                                          'poller_hash': 'MTU1MzIycGVlcnBvbnBvbiAzNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '3534799',
                                                                                                                                          'fiber_length': 1618,
                                                                                                                                          'poller_hash': 'MTc1RTZBcGVlcnBvbnBvbiAzNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '13219203',
                                                                                                                                          'fiber_length': 1592,
                                                                                                                                          'poller_hash': 'MTU1NDM2cGVlcnBvbnBvbiAzNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '64764785',
                                                                                                                                          'fiber_length': 1596,
                                                                                                                                          'poller_hash': 'MTc1RTcycGVlcnBvbnBvbiAzOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '15564960',
                                                                                                                                          'fiber_length': 1611,
                                                                                                                                          'poller_hash': 'MTc2MjM5cGVlcnBvbnBvbiAzOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '13219099',
                                                                                                                                          'fiber_length': 1587,
                                                                                                                                          'poller_hash': 'Q0MwQzVwZWVycG9ucG9uIDQw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '13219272',
                                                                                                                                          'fiber_length': 1583,
                                                                                                                                          'poller_hash': 'MUM1Q0U5cGVlcnBvbnBvbiA0MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '47626',
                                                                                                                                          'fiber_length': 1616,
                                                                                                                                          'poller_hash': 'MUM1Q0I3cGVlcnBvbnBvbiA1NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '31214041',
                                                                                                                                          'fiber_length': 1684,
                                                                                                                                          'poller_hash': 'MkIwMjY0cGVlcnBvbnBvbiAyNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '2150542',
                                                                                                                                          'fiber_length': 1681,
                                                                                                                                          'poller_hash': 'MjA1REM4cGVlcnBvbnBvbiA0NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '21815971',
                                                                                                                                          'fiber_length': 601,
                                                                                                                                          'poller_hash': 'REZGREZBcGVlcnBvbnBvbiAxNzE4NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '15498365',
                                                                                                                                          'fiber_length': 946,
                                                                                                                                          'poller_hash': 'MjYwRTk2cGVlcnBvbnBvbiAxOTQxMDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688457',
                                                                                                                                          'fiber_length': 1565,
                                                                                                                                          'poller_hash': 'MTI3NDJFN3BlZXJwb25wb24gMjQ4MDM4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/1',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/2',
                                                                                         'in_discards': '41123804',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1024336635',
                                                                                         'in_octets': '8547624246573',
                                                                                         'in_ucast_pkts': '43491213479',
                                                                                         'links': [       {       'in_nucast_pkts': '8730970',
                                                                                                                  'in_octets': '34979324322',
                                                                                                                  'in_ucast_pkts': '325343860',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8673421383',
                                                                                                                  'out_octets': '13237795399982',
                                                                                                                  'out_ucast_pkts': '1205947818',
                                                                                                                  'poller_hash': 'MjYzMjAwcGVlcmludGZwb24gNDQ0'},
                                                                                                          {       'in_nucast_pkts': '4041252',
                                                                                                                  'in_octets': '34614133066',
                                                                                                                  'in_ucast_pkts': '453754856',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3758956195',
                                                                                                                  'out_octets': '7042602668178',
                                                                                                                  'out_ucast_pkts': '1484417979',
                                                                                                                  'poller_hash': 'MjYzMjAxcGVlcmludGZwb24gNDQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTM4MTNDcGVlcmludGZwb24gNDUx'},
                                                                                                          {       'in_nucast_pkts': '17801657',
                                                                                                                  'in_octets': '2879405891833',
                                                                                                                  'in_ucast_pkts': '4400959361',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '104403874',
                                                                                                                  'out_octets': '6178245977018',
                                                                                                                  'out_ucast_pkts': '5534414511',
                                                                                                                  'poller_hash': 'Mzc4OEQwcGVlcmludGZwb24gNDUy'},
                                                                                                          {       'in_nucast_pkts': '6269634',
                                                                                                                  'in_octets': '20044117299',
                                                                                                                  'in_ucast_pkts': '239938814',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3780897893',
                                                                                                                  'out_octets': '5963264911777',
                                                                                                                  'out_ucast_pkts': '760301753',
                                                                                                                  'poller_hash': 'MjZEOTY2cGVlcmludGZwb24gNDY2'},
                                                                                                          {       'in_nucast_pkts': '2653517',
                                                                                                                  'in_octets': '54200169761',
                                                                                                                  'in_ucast_pkts': '612167640',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '576270951',
                                                                                                                  'out_octets': '2520394633059',
                                                                                                                  'out_ucast_pkts': '1281598688',
                                                                                                                  'poller_hash': 'Mzc4OEI2cGVlcmludGZwb24gNDY3'},
                                                                                                          {       'in_nucast_pkts': '7949034',
                                                                                                                  'in_octets': '34738079825',
                                                                                                                  'in_ucast_pkts': '412365981',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6036905516',
                                                                                                                  'out_octets': '9910428162398',
                                                                                                                  'out_ucast_pkts': '1442319280',
                                                                                                                  'poller_hash': 'MjZBQzZGcGVlcmludGZwb24gNDY5'},
                                                                                                          {       'in_nucast_pkts': '5085978',
                                                                                                                  'in_octets': '2594081482',
                                                                                                                  'in_ucast_pkts': '12064611',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4898057674',
                                                                                                                  'out_octets': '6493766818982',
                                                                                                                  'out_ucast_pkts': '23480694',
                                                                                                                  'poller_hash': 'MjYzMjIycGVlcmludGZwb24gNDcx'},
                                                                                                          {       'in_nucast_pkts': '18719946',
                                                                                                                  'in_octets': '4842233472515',
                                                                                                                  'in_ucast_pkts': '24001341635',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '89559502',
                                                                                                                  'out_octets': '76648560748600',
                                                                                                                  'out_ucast_pkts': '57335384233',
                                                                                                                  'poller_hash': 'MzVFOTYwcGVlcmludGZwb24gNDc1'},
                                                                                                          {       'in_nucast_pkts': '6518541',
                                                                                                                  'in_octets': '24927596922',
                                                                                                                  'in_ucast_pkts': '321333907',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7745618277',
                                                                                                                  'out_octets': '11693204295012',
                                                                                                                  'out_ucast_pkts': '1002363451',
                                                                                                                  'poller_hash': 'MjYyNUU3cGVlcmludGZwb24gNDc2'},
                                                                                                          {       'in_nucast_pkts': '1351049',
                                                                                                                  'in_octets': '23203833608',
                                                                                                                  'in_ucast_pkts': '277655853',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '190646173',
                                                                                                                  'out_octets': '1707134256106',
                                                                                                                  'out_ucast_pkts': '1101205368',
                                                                                                                  'poller_hash': 'NDU5NjhDcGVlcmludGZwb24gNDc3'},
                                                                                                          {       'in_nucast_pkts': '9522',
                                                                                                                  'in_octets': '19689510983',
                                                                                                                  'in_ucast_pkts': '83498930',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14496965',
                                                                                                                  'out_octets': '139028020970',
                                                                                                                  'out_ucast_pkts': '153797440',
                                                                                                                  'poller_hash': 'MTUzRDhwZWVyaW50ZnBvbiA0Nzg='},
                                                                                                          {       'in_nucast_pkts': '3007',
                                                                                                                  'in_octets': '134255210658',
                                                                                                                  'in_ucast_pkts': '417121203',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7173891',
                                                                                                                  'out_octets': '299164406553',
                                                                                                                  'out_ucast_pkts': '472619677',
                                                                                                                  'poller_hash': 'RDNFMzJwZWVyaW50ZnBvbiA0Nzk='},
                                                                                                          {       'in_nucast_pkts': '4631150',
                                                                                                                  'in_octets': '38882766114',
                                                                                                                  'in_ucast_pkts': '506979135',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3227605097',
                                                                                                                  'out_octets': '6546932644639',
                                                                                                                  'out_ucast_pkts': '1678970737',
                                                                                                                  'poller_hash': 'NDdFMzdwZWVyaW50ZnBvbiAyMzA5MjY='},
                                                                                                          {       'in_nucast_pkts': '6453843',
                                                                                                                  'in_octets': '6669581530',
                                                                                                                  'in_ucast_pkts': '59840162',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6753433685',
                                                                                                                  'out_octets': '9187899872543',
                                                                                                                  'out_ucast_pkts': '179759654',
                                                                                                                  'poller_hash': 'QkE4QzNwZWVyaW50ZnBvbiAyMzA5NDQ='},
                                                                                                          {       'in_nucast_pkts': '5144192',
                                                                                                                  'in_octets': '33948656829',
                                                                                                                  'in_ucast_pkts': '436889547',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6706929295',
                                                                                                                  'out_octets': '10859060650424',
                                                                                                                  'out_ucast_pkts': '1385275903',
                                                                                                                  'poller_hash': 'MTExQkMzcGVlcmludGZwb24gMjMwOTYy'},
                                                                                                          {       'in_nucast_pkts': '4869641',
                                                                                                                  'in_octets': '30639742617',
                                                                                                                  'in_ucast_pkts': '350471946',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '561550727',
                                                                                                                  'out_octets': '2292106567545',
                                                                                                                  'out_ucast_pkts': '1262233272',
                                                                                                                  'poller_hash': 'MTM0NDhGcGVlcmludGZwb24gMjMwOTgw'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '586493263497',
                                                                                         'out_octets': '930035314396536',
                                                                                         'out_ucast_pkts': '99567133781',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/2',
                                                                                                                 'id': '7/2',
                                                                                                                 'links': [       {       'connected_time': '4164919',
                                                                                                                                          'fiber_length': 2091,
                                                                                                                                          'poller_hash': 'MjYzMjAwcGVlcnBvbnBvbiA0NDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164937',
                                                                                                                                          'fiber_length': 2090,
                                                                                                                                          'poller_hash': 'MjYzMjAxcGVlcnBvbnBvbiA0NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164970',
                                                                                                                                          'fiber_length': 2085,
                                                                                                                                          'poller_hash': 'MTM4MTNDcGVlcnBvbnBvbiA0NTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4468821',
                                                                                                                                          'fiber_length': 1711,
                                                                                                                                          'poller_hash': 'Mzc4OEQwcGVlcnBvbnBvbiA0NTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164963',
                                                                                                                                          'fiber_length': 1937,
                                                                                                                                          'poller_hash': 'MjZEOTY2cGVlcnBvbnBvbiA0NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4165169',
                                                                                                                                          'fiber_length': 1561,
                                                                                                                                          'poller_hash': 'Mzc4OEI2cGVlcnBvbnBvbiA0Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164897',
                                                                                                                                          'fiber_length': 1893,
                                                                                                                                          'poller_hash': 'MjZBQzZGcGVlcnBvbnBvbiA0Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164929',
                                                                                                                                          'fiber_length': 1893,
                                                                                                                                          'poller_hash': 'MjYzMjIycGVlcnBvbnBvbiA0NzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '67790412',
                                                                                                                                          'fiber_length': 1593,
                                                                                                                                          'poller_hash': 'MzVFOTYwcGVlcnBvbnBvbiA0NzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4155296',
                                                                                                                                          'fiber_length': 1892,
                                                                                                                                          'poller_hash': 'MjYyNUU3cGVlcnBvbnBvbiA0NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4468870',
                                                                                                                                          'fiber_length': 1515,
                                                                                                                                          'poller_hash': 'NDU5NjhDcGVlcnBvbnBvbiA0Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164913',
                                                                                                                                          'fiber_length': 1456,
                                                                                                                                          'poller_hash': 'MTUzRDhwZWVycG9ucG9uIDQ3OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4165045',
                                                                                                                                          'fiber_length': 1620,
                                                                                                                                          'poller_hash': 'RDNFMzJwZWVycG9ucG9uIDQ3OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4165100',
                                                                                                                                          'fiber_length': 2004,
                                                                                                                                          'poller_hash': 'NDdFMzdwZWVycG9ucG9uIDIzMDkyNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164935',
                                                                                                                                          'fiber_length': 2004,
                                                                                                                                          'poller_hash': 'QkE4QzNwZWVycG9ucG9uIDIzMDk0NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164949',
                                                                                                                                          'fiber_length': 2004,
                                                                                                                                          'poller_hash': 'MTExQkMzcGVlcnBvbnBvbiAyMzA5NjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715},
                                                                                                                                  {       'connected_time': '4164955',
                                                                                                                                          'fiber_length': 2003,
                                                                                                                                          'poller_hash': 'MTM0NDhGcGVlcnBvbnBvbiAyMzA5ODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.715}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/2',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/3',
                                                                                         'in_discards': '5514',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2267526',
                                                                                         'in_octets': '3895516097638',
                                                                                         'in_ucast_pkts': '21073274070',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxRUQ5cGVlcmludGZwb24gNDY='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '163343893520',
                                                                                                                  'in_ucast_pkts': '662137203',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '55',
                                                                                                                  'out_octets': '4417468212601',
                                                                                                                  'out_ucast_pkts': '3311553706',
                                                                                                                  'poller_hash': 'MUNBQjFCcGVlcmludGZwb24gNTQ='},
                                                                                                          {       'in_nucast_pkts': '1188',
                                                                                                                  'in_octets': '101160340450',
                                                                                                                  'in_ucast_pkts': '1303931461',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '116864876',
                                                                                                                  'out_octets': '4919599312284',
                                                                                                                  'out_ucast_pkts': '3271028279',
                                                                                                                  'poller_hash': 'MkIwMUM2cGVlcmludGZwb24gNDY4'},
                                                                                                          {       'in_nucast_pkts': '283',
                                                                                                                  'in_octets': '166740007380',
                                                                                                                  'in_ucast_pkts': '1196146901',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1916',
                                                                                                                  'out_octets': '5787051856616',
                                                                                                                  'out_ucast_pkts': '4120755172',
                                                                                                                  'poller_hash': 'MjA1RDlDcGVlcmludGZwb24gNTQx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '143668847257',
                                                                                                                  'in_ucast_pkts': '1249415018',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '120',
                                                                                                                  'out_octets': '8506462874943',
                                                                                                                  'out_ucast_pkts': '5692440077',
                                                                                                                  'poller_hash': 'MjA1QjNDcGVlcmludGZwb24gNzAy'},
                                                                                                          {       'in_nucast_pkts': '1442',
                                                                                                                  'in_octets': '396708306360',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7713',
                                                                                                                  'out_octets': '5344357132276',
                                                                                                                  'out_ucast_pkts': '4023637363',
                                                                                                                  'poller_hash': 'OTU2RjRBcGVlcmludGZwb24gMTQ1NTE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '21541401522',
                                                                                                                  'in_ucast_pkts': '33859562',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149',
                                                                                                                  'out_octets': '92836170291',
                                                                                                                  'out_ucast_pkts': '128177109',
                                                                                                                  'poller_hash': 'QTlDNDFCcGVlcmludGZwb24gMTQ2MDMx'},
                                                                                                          {       'in_nucast_pkts': '38340',
                                                                                                                  'in_octets': '2834585558',
                                                                                                                  'in_ucast_pkts': '38891785',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15403089',
                                                                                                                  'out_octets': '222763755040',
                                                                                                                  'out_ucast_pkts': '154144071',
                                                                                                                  'poller_hash': 'MTI3NDlGRnBlZXJpbnRmcG9uIDE2Nzg0OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '363268840002',
                                                                                                                  'in_ucast_pkts': '1293017247',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '120',
                                                                                                                  'out_octets': '5404303454918',
                                                                                                                  'out_ucast_pkts': '4225601764',
                                                                                                                  'poller_hash': 'MTI3NTFEQXBlZXJpbnRmcG9uIDIyNDczOQ=='},
                                                                                                          {       'in_nucast_pkts': '14313',
                                                                                                                  'in_octets': '4759764367',
                                                                                                                  'in_ucast_pkts': '13761013',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464640',
                                                                                                                  'out_octets': '169908376600',
                                                                                                                  'out_ucast_pkts': '121188918',
                                                                                                                  'poller_hash': 'MTI3NDc5QXBlZXJpbnRmcG9uIDIyNDc1OQ=='},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '3518295021',
                                                                                                                  'in_ucast_pkts': '14725953',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464696',
                                                                                                                  'out_octets': '108269970399',
                                                                                                                  'out_ucast_pkts': '82149703',
                                                                                                                  'poller_hash': 'MTI3NTFDMXBlZXJpbnRmcG9uIDIyNDc2OQ=='},
                                                                                                          {       'in_nucast_pkts': '296',
                                                                                                                  'in_octets': '265919503409',
                                                                                                                  'in_ucast_pkts': '5525945238',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1246',
                                                                                                                  'out_octets': '4318340523388',
                                                                                                                  'out_ucast_pkts': '3176230830',
                                                                                                                  'poller_hash': 'MTA5NDhCNHBlZXJpbnRmcG9uIDI0OTYwMQ=='},
                                                                                                          {       'in_nucast_pkts': '14222',
                                                                                                                  'in_octets': '4260618485',
                                                                                                                  'in_ucast_pkts': '26984000',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465414',
                                                                                                                  'out_octets': '83792637439',
                                                                                                                  'out_ucast_pkts': '62961692',
                                                                                                                  'poller_hash': 'MTI3MEU0N3BlZXJpbnRmcG9uIDI3Njc2OQ=='},
                                                                                                          {       'in_nucast_pkts': '2384670',
                                                                                                                  'in_octets': '89413495185',
                                                                                                                  'in_ucast_pkts': '138526364',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '13558273',
                                                                                                                  'out_octets': '801683375994',
                                                                                                                  'out_ucast_pkts': '605054649',
                                                                                                                  'poller_hash': 'MTUyMDI5RHBlZXJpbnRmcG9uIDI3OTE1MA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '55413188754',
                                                                                         'out_octets': '179072880523660',
                                                                                         'out_ucast_pkts': '74337262394',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/3',
                                                                                                                 'id': '7/3',
                                                                                                                 'links': [       {       'connected_time': '6743564',
                                                                                                                                          'fiber_length': 1542,
                                                                                                                                          'poller_hash': 'QzgxRUQ5cGVlcnBvbnBvbiA0Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688464',
                                                                                                                                          'fiber_length': 1572,
                                                                                                                                          'poller_hash': 'MUNBQjFCcGVlcnBvbnBvbiA1NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '6743042',
                                                                                                                                          'fiber_length': 1710,
                                                                                                                                          'poller_hash': 'MkIwMUM2cGVlcnBvbnBvbiA0Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '583647',
                                                                                                                                          'fiber_length': 1682,
                                                                                                                                          'poller_hash': 'MjA1RDlDcGVlcnBvbnBvbiA1NDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688463',
                                                                                                                                          'fiber_length': 1667,
                                                                                                                                          'poller_hash': 'MjA1QjNDcGVlcnBvbnBvbiA3MDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688459',
                                                                                                                                          'fiber_length': 1707,
                                                                                                                                          'poller_hash': 'OTU2RjRBcGVlcnBvbnBvbiAxNDU1MTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688465',
                                                                                                                                          'fiber_length': 1539,
                                                                                                                                          'poller_hash': 'QTlDNDFCcGVlcnBvbnBvbiAxNDYwMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688458',
                                                                                                                                          'fiber_length': 1575,
                                                                                                                                          'poller_hash': 'MTI3NDlGRnBlZXJwb25wb24gMTY3ODQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688459',
                                                                                                                                          'fiber_length': 1573,
                                                                                                                                          'poller_hash': 'MTI3NTFEQXBlZXJwb25wb24gMjI0NzM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688463',
                                                                                                                                          'fiber_length': 1577,
                                                                                                                                          'poller_hash': 'MTI3NDc5QXBlZXJwb25wb24gMjI0NzU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688461',
                                                                                                                                          'fiber_length': 1572,
                                                                                                                                          'poller_hash': 'MTI3NTFDMXBlZXJwb25wb24gMjI0NzY5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '330749',
                                                                                                                                          'fiber_length': 1684,
                                                                                                                                          'poller_hash': 'MTA5NDhCNHBlZXJwb25wb24gMjQ5NjAx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688464',
                                                                                                                                          'fiber_length': 1546,
                                                                                                                                          'poller_hash': 'MTI3MEU0N3BlZXJwb25wb24gMjc2NzY5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619},
                                                                                                                                  {       'connected_time': '688363',
                                                                                                                                          'fiber_length': 1586,
                                                                                                                                          'poller_hash': 'MTUyMDI5RHBlZXJwb25wb24gMjc5MTUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.619}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/3',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/4',
                                                                                         'in_discards': '43067',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2150595',
                                                                                         'in_octets': '3991439618202',
                                                                                         'in_ucast_pkts': '23173930002',
                                                                                         'links': [       {       'in_nucast_pkts': '582',
                                                                                                                  'in_octets': '314983286771',
                                                                                                                  'in_ucast_pkts': '2592117679',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3733',
                                                                                                                  'out_octets': '18435312854967',
                                                                                                                  'out_ucast_pkts': '12693954656',
                                                                                                                  'poller_hash': 'MjA1REJDcGVlcmludGZwb24gMTE='},
                                                                                                          {       'in_nucast_pkts': '587',
                                                                                                                  'in_octets': '292896487529',
                                                                                                                  'in_ucast_pkts': '1196026428',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3170',
                                                                                                                  'out_octets': '5190992633329',
                                                                                                                  'out_ucast_pkts': '3885698101',
                                                                                                                  'poller_hash': 'MUQwNjg5cGVlcmludGZwb24gMTgz'},
                                                                                                          {       'in_nucast_pkts': '291',
                                                                                                                  'in_octets': '556934826765',
                                                                                                                  'in_ucast_pkts': '4982724954',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1409',
                                                                                                                  'out_octets': '15513369713051',
                                                                                                                  'out_ucast_pkts': '11838587760',
                                                                                                                  'poller_hash': 'MkRDRDE4cGVlcmludGZwb24gMjc2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTEyNUUzcGVlcmludGZwb24gNTUz'},
                                                                                                          {       'in_nucast_pkts': '827973',
                                                                                                                  'in_octets': '283465947780',
                                                                                                                  'in_ucast_pkts': '987428020',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93544822',
                                                                                                                  'out_octets': '4720769058327',
                                                                                                                  'out_ucast_pkts': '3634014972',
                                                                                                                  'poller_hash': 'MzVFOEU1cGVlcmludGZwb24gNzY1'},
                                                                                                          {       'in_nucast_pkts': '5312',
                                                                                                                  'in_octets': '2106259979',
                                                                                                                  'in_ucast_pkts': '13869220',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '509800',
                                                                                                                  'out_octets': '68617938690',
                                                                                                                  'out_ucast_pkts': '51517436',
                                                                                                                  'poller_hash': 'MjA1REIxcGVlcmludGZwb24gODQ3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '225524961380',
                                                                                                                  'in_ucast_pkts': '1974181731',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149',
                                                                                                                  'out_octets': '12885785410901',
                                                                                                                  'out_ucast_pkts': '8763417876',
                                                                                                                  'poller_hash': 'MTA5NDhGRHBlZXJpbnRmcG9uIDIwNzg0MA=='},
                                                                                                          {       'in_nucast_pkts': '12408',
                                                                                                                  'in_octets': '3815414753',
                                                                                                                  'in_ucast_pkts': '38487689',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464647',
                                                                                                                  'out_octets': '712557466527',
                                                                                                                  'out_ucast_pkts': '471995350',
                                                                                                                  'poller_hash': 'MTE5NTE4N3BlZXJpbnRmcG9uIDIxNjI0Nw=='},
                                                                                                          {       'in_nucast_pkts': '6470360',
                                                                                                                  'in_octets': '243476740482',
                                                                                                                  'in_ucast_pkts': '1023563348',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '35200942',
                                                                                                                  'out_octets': '10725007082850',
                                                                                                                  'out_ucast_pkts': '7367435187',
                                                                                                                  'poller_hash': 'MTUyMDJDN3BlZXJpbnRmcG9uIDI2MzIwMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTUyNDJENXBlZXJpbnRmcG9uIDI2NDA3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTUyMDJBNnBlZXJpbnRmcG9uIDI2NDg5OA=='},
                                                                                                          {       'in_nucast_pkts': '2720042',
                                                                                                                  'in_octets': '95235881612',
                                                                                                                  'in_ucast_pkts': '199530599',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15484518',
                                                                                                                  'out_octets': '1250276216436',
                                                                                                                  'out_ucast_pkts': '927523775',
                                                                                                                  'poller_hash': 'MTUyMDI4NnBlZXJpbnRmcG9uIDI3ODE4OA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '103581454874',
                                                                                         'out_octets': '267892921781775',
                                                                                         'out_ucast_pkts': '90618275671',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/4',
                                                                                                                 'id': '7/4',
                                                                                                                 'links': [       {       'connected_time': '688473',
                                                                                                                                          'fiber_length': 1708,
                                                                                                                                          'poller_hash': 'MjA1REJDcGVlcnBvbnBvbiAxMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '5882390',
                                                                                                                                          'fiber_length': 1677,
                                                                                                                                          'poller_hash': 'MUQwNjg5cGVlcnBvbnBvbiAxODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '688478',
                                                                                                                                          'fiber_length': 1670,
                                                                                                                                          'poller_hash': 'MkRDRDE4cGVlcnBvbnBvbiAyNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '15499058',
                                                                                                                                          'fiber_length': 1082,
                                                                                                                                          'poller_hash': 'NTEyNUUzcGVlcnBvbnBvbiA1NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -7.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '67790290',
                                                                                                                                          'fiber_length': 658,
                                                                                                                                          'poller_hash': 'MzVFOEU1cGVlcnBvbnBvbiA3NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '247592',
                                                                                                                                          'fiber_length': 1683,
                                                                                                                                          'poller_hash': 'MjA1REIxcGVlcnBvbnBvbiA4NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '688475',
                                                                                                                                          'fiber_length': 1680,
                                                                                                                                          'poller_hash': 'MTA5NDhGRHBlZXJwb25wb24gMjA3ODQw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '688469',
                                                                                                                                          'fiber_length': 1677,
                                                                                                                                          'poller_hash': 'MTE5NTE4N3BlZXJwb25wb24gMjE2MjQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '688366',
                                                                                                                                          'fiber_length': 1709,
                                                                                                                                          'poller_hash': 'MTUyMDJDN3BlZXJwb25wb24gMjYzMjAx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '688349',
                                                                                                                                          'fiber_length': 1704,
                                                                                                                                          'poller_hash': 'MTUyNDJENXBlZXJwb25wb24gMjY0MDc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '688351',
                                                                                                                                          'fiber_length': 603,
                                                                                                                                          'poller_hash': 'MTUyMDJBNnBlZXJwb25wb24gMjY0ODk4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515},
                                                                                                                                  {       'connected_time': '244052',
                                                                                                                                          'fiber_length': 1711,
                                                                                                                                          'poller_hash': 'MTUyMDI4NnBlZXJwb25wb24gMjc4MTg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.515}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/4',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/5',
                                                                                         'in_discards': '163914',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '4100537',
                                                                                         'in_octets': '7894593433563',
                                                                                         'in_ucast_pkts': '32483117308',
                                                                                         'links': [       {       'in_nucast_pkts': '885',
                                                                                                                  'in_octets': '166328916693',
                                                                                                                  'in_ucast_pkts': '535029877',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3662',
                                                                                                                  'out_octets': '1972200021367',
                                                                                                                  'out_ucast_pkts': '1514891048',
                                                                                                                  'poller_hash': 'MzIyNjAycGVlcmludGZwb24gMTU='},
                                                                                                          {       'in_nucast_pkts': '588',
                                                                                                                  'in_octets': '93161630072',
                                                                                                                  'in_ucast_pkts': '681148045',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3569',
                                                                                                                  'out_octets': '4149987183511',
                                                                                                                  'out_ucast_pkts': '6991802746',
                                                                                                                  'poller_hash': 'MUQwNjdCcGVlcmludGZwb24gMjQ='},
                                                                                                          {       'in_nucast_pkts': '294',
                                                                                                                  'in_octets': '227544243574',
                                                                                                                  'in_ucast_pkts': '2239265737',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1342',
                                                                                                                  'out_octets': '14387907269015',
                                                                                                                  'out_ucast_pkts': '9809126788',
                                                                                                                  'poller_hash': 'MUQwNjg3cGVlcmludGZwb24gMjc='},
                                                                                                          {       'in_nucast_pkts': '917110',
                                                                                                                  'in_octets': '391683198216',
                                                                                                                  'in_ucast_pkts': '1165101432',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145796909',
                                                                                                                  'out_octets': '9857876817968',
                                                                                                                  'out_ucast_pkts': '7007047530',
                                                                                                                  'poller_hash': 'MUQwNjkxcGVlcmludGZwb24gMjg='},
                                                                                                          {       'in_nucast_pkts': '751308',
                                                                                                                  'in_octets': '241509603118',
                                                                                                                  'in_ucast_pkts': '1921402154',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '102597507',
                                                                                                                  'out_octets': '10672929883136',
                                                                                                                  'out_ucast_pkts': '7367336187',
                                                                                                                  'poller_hash': 'MUQwNEYycGVlcmludGZwb24gMzE='},
                                                                                                          {       'in_nucast_pkts': '294',
                                                                                                                  'in_octets': '281450057321',
                                                                                                                  'in_ucast_pkts': '3035056868',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1492',
                                                                                                                  'out_octets': '12410677112663',
                                                                                                                  'out_ucast_pkts': '8443358054',
                                                                                                                  'poller_hash': 'MUQwNEY2cGVlcmludGZwb24gNDQ='},
                                                                                                          {       'in_nucast_pkts': '580811',
                                                                                                                  'in_octets': '387021719585',
                                                                                                                  'in_ucast_pkts': '1948038011',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '68592547',
                                                                                                                  'out_octets': '10752559959640',
                                                                                                                  'out_ucast_pkts': '7441000720',
                                                                                                                  'poller_hash': 'MjA1REZFcGVlcmludGZwb24gNDU='},
                                                                                                          {       'in_nucast_pkts': '34',
                                                                                                                  'in_octets': '32249782341',
                                                                                                                  'in_ucast_pkts': '168635261',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5553990',
                                                                                                                  'out_octets': '985441610547',
                                                                                                                  'out_ucast_pkts': '697157302',
                                                                                                                  'poller_hash': 'MjQ4QUNCcGVlcmludGZwb24gNTA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '87877111137',
                                                                                                                  'in_ucast_pkts': '335918136',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '175',
                                                                                                                  'out_octets': '1515979801657',
                                                                                                                  'out_ucast_pkts': '1269845918',
                                                                                                                  'poller_hash': 'Mjk4MjU0cGVlcmludGZwb24gODM='},
                                                                                                          {       'in_nucast_pkts': '294',
                                                                                                                  'in_octets': '118070127724',
                                                                                                                  'in_ucast_pkts': '696368668',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1892',
                                                                                                                  'out_octets': '2804822982271',
                                                                                                                  'out_ucast_pkts': '2185895902',
                                                                                                                  'poller_hash': 'MkIwMjZBcGVlcmludGZwb24gMjcz'},
                                                                                                          {       'in_nucast_pkts': '14605',
                                                                                                                  'in_octets': '12782358358',
                                                                                                                  'in_ucast_pkts': '33819403',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464198',
                                                                                                                  'out_octets': '86937142738',
                                                                                                                  'out_ucast_pkts': '68136931',
                                                                                                                  'poller_hash': 'NTEyNUUwcGVlcmludGZwb24gNTQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDk2RnBlZXJpbnRmcG9uIDIzNDU2NA=='},
                                                                                                          {       'in_nucast_pkts': '15168',
                                                                                                                  'in_octets': '14317919365',
                                                                                                                  'in_ucast_pkts': '26682363',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465327',
                                                                                                                  'out_octets': '133008787547',
                                                                                                                  'out_ucast_pkts': '109405672',
                                                                                                                  'poller_hash': 'MzAwMzlGcGVlcmludGZwb24gMjU1NDI2'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '72409436173',
                                                                                         'out_octets': '229197412858301',
                                                                                         'out_ucast_pkts': '97911143984',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/5',
                                                                                                                 'id': '7/5',
                                                                                                                 'links': [       {       'connected_time': '688464',
                                                                                                                                          'fiber_length': 1255,
                                                                                                                                          'poller_hash': 'MzIyNjAycGVlcnBvbnBvbiAxNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '31213046',
                                                                                                                                          'fiber_length': 1321,
                                                                                                                                          'poller_hash': 'MUQwNjdCcGVlcnBvbnBvbiAyNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '1118630',
                                                                                                                                          'fiber_length': 1201,
                                                                                                                                          'poller_hash': 'MUQwNjg3cGVlcnBvbnBvbiAyNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '21817073',
                                                                                                                                          'fiber_length': 1253,
                                                                                                                                          'poller_hash': 'MUQwNjkxcGVlcnBvbnBvbiAyOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '1142071',
                                                                                                                                          'fiber_length': 1168,
                                                                                                                                          'poller_hash': 'MUQwNEYycGVlcnBvbnBvbiAzMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '21817303',
                                                                                                                                          'fiber_length': 1182,
                                                                                                                                          'poller_hash': 'MUQwNEY2cGVlcnBvbnBvbiA0NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '31303880',
                                                                                                                                          'fiber_length': 1242,
                                                                                                                                          'poller_hash': 'MjA1REZFcGVlcnBvbnBvbiA0NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '3294100',
                                                                                                                                          'fiber_length': 1046,
                                                                                                                                          'poller_hash': 'MjQ4QUNCcGVlcnBvbnBvbiA1MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '21817288',
                                                                                                                                          'fiber_length': 1239,
                                                                                                                                          'poller_hash': 'Mjk4MjU0cGVlcnBvbnBvbiA4Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688462',
                                                                                                                                          'fiber_length': 1253,
                                                                                                                                          'poller_hash': 'MkIwMjZBcGVlcnBvbnBvbiAyNzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688480',
                                                                                                                                          'fiber_length': 1258,
                                                                                                                                          'poller_hash': 'NTEyNUUwcGVlcnBvbnBvbiA1NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '3539036',
                                                                                                                                          'fiber_length': 1244,
                                                                                                                                          'poller_hash': 'MTI3NDk2RnBlZXJwb25wb24gMjM0NTY0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688468',
                                                                                                                                          'fiber_length': 1686,
                                                                                                                                          'poller_hash': 'MzAwMzlGcGVlcnBvbnBvbiAyNTU0MjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/5',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/6',
                                                                                         'in_discards': '117448',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '6783902',
                                                                                         'in_octets': '7775620316623',
                                                                                         'in_ucast_pkts': '38075104224',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '174165131701',
                                                                                                                  'in_ucast_pkts': '781473559',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '173',
                                                                                                                  'out_octets': '2352370948535',
                                                                                                                  'out_ucast_pkts': '1811737325',
                                                                                                                  'poller_hash': 'MUQwNEYwcGVlcmludGZwb24gMjU='},
                                                                                                          {       'in_nucast_pkts': '3586163',
                                                                                                                  'in_octets': '237303090972',
                                                                                                                  'in_ucast_pkts': '3260542452',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '209299556',
                                                                                                                  'out_octets': '15041056609740',
                                                                                                                  'out_ucast_pkts': '10647145217',
                                                                                                                  'poller_hash': 'MUQwNEZDcGVlcmludGZwb24gMjk='},
                                                                                                          {       'in_nucast_pkts': '883',
                                                                                                                  'in_octets': '125548705447',
                                                                                                                  'in_ucast_pkts': '732352462',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4976',
                                                                                                                  'out_octets': '2707598870234',
                                                                                                                  'out_ucast_pkts': '1989665998',
                                                                                                                  'poller_hash': 'MUQwNEZEcGVlcmludGZwb24gNjM='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '25613269436',
                                                                                                                  'in_ucast_pkts': '98275361',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '84',
                                                                                                                  'out_octets': '675875791204',
                                                                                                                  'out_ucast_pkts': '580925814',
                                                                                                                  'poller_hash': 'MUQwMzQzcGVlcmludGZwb24gNjQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '93578649754',
                                                                                                                  'in_ucast_pkts': '1101373829',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '55',
                                                                                                                  'out_octets': '6794619215934',
                                                                                                                  'out_ucast_pkts': '4550961024',
                                                                                                                  'poller_hash': 'MUQwNTBBcGVlcmludGZwb24gNjk='},
                                                                                                          {       'in_nucast_pkts': '2680',
                                                                                                                  'in_octets': '7082820446',
                                                                                                                  'in_ucast_pkts': '13812961',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464275',
                                                                                                                  'out_octets': '39896573040',
                                                                                                                  'out_ucast_pkts': '32419540',
                                                                                                                  'poller_hash': 'MUQwNEY4cGVlcmludGZwb24gNzI='},
                                                                                                          {       'in_nucast_pkts': '290',
                                                                                                                  'in_octets': '1715503962572',
                                                                                                                  'in_ucast_pkts': '2620122891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1388',
                                                                                                                  'out_octets': '13413119535822',
                                                                                                                  'out_ucast_pkts': '11994185083',
                                                                                                                  'poller_hash': 'MzIxODg5cGVlcmludGZwb24gNzk='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '151598777130',
                                                                                                                  'in_ucast_pkts': '1053699351',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '126',
                                                                                                                  'out_octets': '4762940469499',
                                                                                                                  'out_ucast_pkts': '3359228697',
                                                                                                                  'poller_hash': 'Mjk4MjY4cGVlcmludGZwb24gMTM1'},
                                                                                                          {       'in_nucast_pkts': '2841659',
                                                                                                                  'in_octets': '140649660515',
                                                                                                                  'in_ucast_pkts': '416305957',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75869792',
                                                                                                                  'out_octets': '2785741158591',
                                                                                                                  'out_ucast_pkts': '2011783255',
                                                                                                                  'poller_hash': 'MzZDQzJGcGVlcmludGZwb24gMzA5'},
                                                                                                          {       'in_nucast_pkts': '30340',
                                                                                                                  'in_octets': '214048507',
                                                                                                                  'in_ucast_pkts': '778014',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464811',
                                                                                                                  'out_octets': '4770205631',
                                                                                                                  'out_ucast_pkts': '3943032',
                                                                                                                  'poller_hash': 'MzAwNjBBcGVlcmludGZwb24gMzEx'},
                                                                                                          {       'in_nucast_pkts': '495',
                                                                                                                  'in_octets': '323146291294',
                                                                                                                  'in_ucast_pkts': '1121148673',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '78317630',
                                                                                                                  'out_octets': '3167032182499',
                                                                                                                  'out_ucast_pkts': '2622528587',
                                                                                                                  'poller_hash': 'MjU5ODJBcGVlcmludGZwb24gMzUw'},
                                                                                                          {       'in_nucast_pkts': '45970536',
                                                                                                                  'in_octets': '1137356152878',
                                                                                                                  'in_ucast_pkts': '4168550630',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82673312',
                                                                                                                  'out_octets': '14279938229919',
                                                                                                                  'out_ucast_pkts': '10716155477',
                                                                                                                  'poller_hash': 'NUE2QkFFcGVlcmludGZwb24gNzUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJCRURGcGVlcmludGZwb24gODM3'},
                                                                                                          {       'in_nucast_pkts': '22789',
                                                                                                                  'in_octets': '924487610',
                                                                                                                  'in_ucast_pkts': '4007864',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120574',
                                                                                                                  'out_octets': '28293322989',
                                                                                                                  'out_ucast_pkts': '22914540',
                                                                                                                  'poller_hash': 'NjQwQTBBcGVlcmludGZwb24gODU3'},
                                                                                                          {       'in_nucast_pkts': '14766',
                                                                                                                  'in_octets': '19106671673',
                                                                                                                  'in_ucast_pkts': '124167407',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464410',
                                                                                                                  'out_octets': '444748064450',
                                                                                                                  'out_ucast_pkts': '329142721',
                                                                                                                  'poller_hash': 'QTlDM0U4cGVlcmludGZwb24gMTU1OTI1'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '8678286180',
                                                                                         'out_octets': '144047462968911',
                                                                                         'out_ucast_pkts': '98535488672',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/6',
                                                                                                                 'id': '7/6',
                                                                                                                 'links': [       {       'connected_time': '21815830',
                                                                                                                                          'fiber_length': 1024,
                                                                                                                                          'poller_hash': 'MUQwNEYwcGVlcnBvbnBvbiAyNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '68955576',
                                                                                                                                          'fiber_length': 1042,
                                                                                                                                          'poller_hash': 'MUQwNEZDcGVlcnBvbnBvbiAyOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688474',
                                                                                                                                          'fiber_length': 1042,
                                                                                                                                          'poller_hash': 'MUQwNEZEcGVlcnBvbnBvbiA2Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '2126558',
                                                                                                                                          'fiber_length': 1056,
                                                                                                                                          'poller_hash': 'MUQwMzQzcGVlcnBvbnBvbiA2NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '607321',
                                                                                                                                          'fiber_length': 1056,
                                                                                                                                          'poller_hash': 'MUQwNTBBcGVlcnBvbnBvbiA2OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688482',
                                                                                                                                          'fiber_length': 999,
                                                                                                                                          'poller_hash': 'MUQwNEY4cGVlcnBvbnBvbiA3Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '22280705',
                                                                                                                                          'fiber_length': 1045,
                                                                                                                                          'poller_hash': 'MzIxODg5cGVlcnBvbnBvbiA3OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688470',
                                                                                                                                          'fiber_length': 910,
                                                                                                                                          'poller_hash': 'Mjk4MjY4cGVlcnBvbnBvbiAxMzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '21433009',
                                                                                                                                          'fiber_length': 986,
                                                                                                                                          'poller_hash': 'MzZDQzJGcGVlcnBvbnBvbiAzMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688477',
                                                                                                                                          'fiber_length': 902,
                                                                                                                                          'poller_hash': 'MzAwNjBBcGVlcnBvbnBvbiAzMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '41653',
                                                                                                                                          'fiber_length': 992,
                                                                                                                                          'poller_hash': 'MjU5ODJBcGVlcnBvbnBvbiAzNTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '67790129',
                                                                                                                                          'fiber_length': 989,
                                                                                                                                          'poller_hash': 'NUE2QkFFcGVlcnBvbnBvbiA3NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '8468173',
                                                                                                                                          'fiber_length': 1032,
                                                                                                                                          'poller_hash': 'NDJCRURGcGVlcnBvbnBvbiA4Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688468',
                                                                                                                                          'fiber_length': 962,
                                                                                                                                          'poller_hash': 'NjQwQTBBcGVlcnBvbnBvbiA4NTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789},
                                                                                                                                  {       'connected_time': '688477',
                                                                                                                                          'fiber_length': 969,
                                                                                                                                          'poller_hash': 'QTlDM0U4cGVlcnBvbnBvbiAxNTU5MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -8.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.789}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/6',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/7',
                                                                                         'in_discards': '576313',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2225542',
                                                                                         'in_octets': '5494909866210',
                                                                                         'in_ucast_pkts': '24663802040',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '195643914613',
                                                                                                                  'in_ucast_pkts': '453135759',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '138',
                                                                                                                  'out_octets': '1862843802817',
                                                                                                                  'out_ucast_pkts': '1533788173',
                                                                                                                  'poller_hash': 'MUQwMzQ1cGVlcmludGZwb24gNDM='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '146305589772',
                                                                                                                  'in_ucast_pkts': '90832460840',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '115',
                                                                                                                  'out_octets': '2371519702257',
                                                                                                                  'out_ucast_pkts': '1847624428',
                                                                                                                  'poller_hash': 'MUQwNjk0cGVlcmludGZwb24gODQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '498832924835',
                                                                                                                  'in_ucast_pkts': '1867527296',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '4465740546109',
                                                                                                                  'out_ucast_pkts': '3709436189',
                                                                                                                  'poller_hash': 'Mjk4Mjg5cGVlcmludGZwb24gMTQz'},
                                                                                                          {       'in_nucast_pkts': '295',
                                                                                                                  'in_octets': '584458173370',
                                                                                                                  'in_ucast_pkts': '2722176725',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1778',
                                                                                                                  'out_octets': '11449128463770',
                                                                                                                  'out_ucast_pkts': '8437752760',
                                                                                                                  'poller_hash': 'MUQwNTEwcGVlcmludGZwb24gMjU1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '653638047376',
                                                                                                                  'in_ucast_pkts': '2384981450',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9157273985809',
                                                                                                                  'out_ucast_pkts': '7053190465',
                                                                                                                  'poller_hash': 'NDU5NjM4cGVlcmludGZwb24gNDcy'},
                                                                                                          {       'in_nucast_pkts': '298',
                                                                                                                  'in_octets': '226781213409',
                                                                                                                  'in_ucast_pkts': '446971699',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1454',
                                                                                                                  'out_octets': '638555745866',
                                                                                                                  'out_ucast_pkts': '682367607',
                                                                                                                  'poller_hash': 'QjA5RjM5cGVlcmludGZwb24gNDgw'},
                                                                                                          {       'in_nucast_pkts': '11902',
                                                                                                                  'in_octets': '7299988557',
                                                                                                                  'in_ucast_pkts': '45392140',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120052',
                                                                                                                  'out_octets': '177601982216',
                                                                                                                  'out_ucast_pkts': '123538236',
                                                                                                                  'poller_hash': 'NEY5QUU0cGVlcmludGZwb24gNTM5'},
                                                                                                          {       'in_nucast_pkts': '774325',
                                                                                                                  'in_octets': '620912527263',
                                                                                                                  'in_ucast_pkts': '4290628594',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82666182',
                                                                                                                  'out_octets': '17153343167003',
                                                                                                                  'out_ucast_pkts': '12494152970',
                                                                                                                  'poller_hash': 'MzVFOTZBcGVlcmludGZwb24gNTQy'},
                                                                                                          {       'in_nucast_pkts': '38076',
                                                                                                                  'in_octets': '2662733693',
                                                                                                                  'in_ucast_pkts': '7297399',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4383101',
                                                                                                                  'out_octets': '10215709746',
                                                                                                                  'out_ucast_pkts': '12038643',
                                                                                                                  'poller_hash': 'Mjk4MUZBcGVlcmludGZwb24gMTQyMDQ3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '105898319420',
                                                                                                                  'in_ucast_pkts': '286375361',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '175',
                                                                                                                  'out_octets': '815892944920',
                                                                                                                  'out_ucast_pkts': '699377166',
                                                                                                                  'poller_hash': 'ODcyNDQ1cGVlcmludGZwb24gMTQzMTY3'},
                                                                                                          {       'in_nucast_pkts': '292',
                                                                                                                  'in_octets': '200243195514',
                                                                                                                  'in_ucast_pkts': '724887725',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1315',
                                                                                                                  'out_octets': '2658846181298',
                                                                                                                  'out_ucast_pkts': '1921819834',
                                                                                                                  'poller_hash': 'MTE5NTE5MXBlZXJpbnRmcG9uIDIyNDYzOQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '7526934493',
                                                                                         'out_octets': '99227096003455',
                                                                                         'out_ucast_pkts': '66783374244',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/7',
                                                                                                                 'id': '7/7',
                                                                                                                 'links': [       {       'connected_time': '12019959',
                                                                                                                                          'fiber_length': 837,
                                                                                                                                          'poller_hash': 'MUQwMzQ1cGVlcnBvbnBvbiA0Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '21816848',
                                                                                                                                          'fiber_length': 1035,
                                                                                                                                          'poller_hash': 'MUQwNjk0cGVlcnBvbnBvbiA4NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '21533843',
                                                                                                                                          'fiber_length': 799,
                                                                                                                                          'poller_hash': 'Mjk4Mjg5cGVlcnBvbnBvbiAxNDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '688479',
                                                                                                                                          'fiber_length': 849,
                                                                                                                                          'poller_hash': 'MUQwNTEwcGVlcnBvbnBvbiAyNTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '71351983',
                                                                                                                                          'fiber_length': 1219,
                                                                                                                                          'poller_hash': 'NDU5NjM4cGVlcnBvbnBvbiA0NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '13238676',
                                                                                                                                          'fiber_length': 1461,
                                                                                                                                          'poller_hash': 'QjA5RjM5cGVlcnBvbnBvbiA0ODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '688463',
                                                                                                                                          'fiber_length': 748,
                                                                                                                                          'poller_hash': 'NEY5QUU0cGVlcnBvbnBvbiA1Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '67790372',
                                                                                                                                          'fiber_length': 1035,
                                                                                                                                          'poller_hash': 'MzVFOTZBcGVlcnBvbnBvbiA1NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '2024171',
                                                                                                                                          'fiber_length': 1047,
                                                                                                                                          'poller_hash': 'Mjk4MUZBcGVlcnBvbnBvbiAxNDIwNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '15498918',
                                                                                                                                          'fiber_length': 842,
                                                                                                                                          'poller_hash': 'ODcyNDQ1cGVlcnBvbnBvbiAxNDMxNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516},
                                                                                                                                  {       'connected_time': '688477',
                                                                                                                                          'fiber_length': 872,
                                                                                                                                          'poller_hash': 'MTE5NTE5MXBlZXJwb25wb24gMjI0NjM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.516}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/7',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/8',
                                                                                         'in_discards': '58811',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2694149',
                                                                                         'in_octets': '8210477761246',
                                                                                         'in_ucast_pkts': '28324045031',
                                                                                         'links': [       {       'in_nucast_pkts': '14215',
                                                                                                                  'in_octets': '15517818357',
                                                                                                                  'in_ucast_pkts': '56792414',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464322',
                                                                                                                  'out_octets': '241231029350',
                                                                                                                  'out_ucast_pkts': '196171080',
                                                                                                                  'poller_hash': 'MzIyODFEcGVlcmludGZwb24gNTIx'},
                                                                                                          {       'in_nucast_pkts': '583',
                                                                                                                  'in_octets': '66953477700',
                                                                                                                  'in_ucast_pkts': '387953859',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2555',
                                                                                                                  'out_octets': '1496602613174',
                                                                                                                  'out_ucast_pkts': '1119189959',
                                                                                                                  'poller_hash': 'MjA1REJCcGVlcmludGZwb24gNzc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NjQwQUY4cGVlcmludGZwb24gNzk2'},
                                                                                                          {       'in_nucast_pkts': '263',
                                                                                                                  'in_octets': '17757050606',
                                                                                                                  'in_ucast_pkts': '78207708',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82752801',
                                                                                                                  'out_octets': '326807614594',
                                                                                                                  'out_ucast_pkts': '264220210',
                                                                                                                  'poller_hash': 'NUE2NjI3cGVlcmludGZwb24gODU1'},
                                                                                                          {       'in_nucast_pkts': '112259',
                                                                                                                  'in_octets': '1443866521174',
                                                                                                                  'in_ucast_pkts': '1355923765',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82647393',
                                                                                                                  'out_octets': '1292647788174',
                                                                                                                  'out_ucast_pkts': '1432316429',
                                                                                                                  'poller_hash': 'NUE2NjhDcGVlcmludGZwb24gODYy'},
                                                                                                          {       'in_nucast_pkts': '498',
                                                                                                                  'in_octets': '2804335994',
                                                                                                                  'in_ucast_pkts': '6324231',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464380',
                                                                                                                  'out_octets': '25140647804',
                                                                                                                  'out_ucast_pkts': '20706447',
                                                                                                                  'poller_hash': 'NUE2NjU3cGVlcmludGZwb24gODcx'},
                                                                                                          {       'in_nucast_pkts': '45372',
                                                                                                                  'in_octets': '19158565519',
                                                                                                                  'in_ucast_pkts': '91256901',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5117448',
                                                                                                                  'out_octets': '547721214775',
                                                                                                                  'out_ucast_pkts': '397141776',
                                                                                                                  'poller_hash': 'NDJCRUMxcGVlcmludGZwb24gMTM3NzI0'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '5170669744',
                                                                                                                  'in_ucast_pkts': '18849813',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464479',
                                                                                                                  'out_octets': '99836235957',
                                                                                                                  'out_ucast_pkts': '79273371',
                                                                                                                  'poller_hash': 'ODY5NDA3cGVlcmludGZwb24gMTQyMzA2'},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '1435108300',
                                                                                                                  'in_ucast_pkts': '11252384',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1462378',
                                                                                                                  'out_octets': '73804079495',
                                                                                                                  'out_ucast_pkts': '51059746',
                                                                                                                  'poller_hash': 'ODY5M0E3cGVlcmludGZwb24gMTQyNDU2'},
                                                                                                          {       'in_nucast_pkts': '11811',
                                                                                                                  'in_octets': '1376929192',
                                                                                                                  'in_ucast_pkts': '5842297',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464407',
                                                                                                                  'out_octets': '47129504531',
                                                                                                                  'out_ucast_pkts': '37494054',
                                                                                                                  'poller_hash': 'ODcyNTQ0cGVlcmludGZwb24gMTQyNzM1'},
                                                                                                          {       'in_nucast_pkts': '290',
                                                                                                                  'in_octets': '95136350377',
                                                                                                                  'in_ucast_pkts': '812367490',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1480',
                                                                                                                  'out_octets': '2320238436424',
                                                                                                                  'out_ucast_pkts': '1672970699',
                                                                                                                  'poller_hash': 'QTlDN0Y5cGVlcmludGZwb24gMTQ5MTc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1363681588879',
                                                                                                                  'in_ucast_pkts': '3754966827',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2186478460565',
                                                                                                                  'out_ucast_pkts': '4439437311',
                                                                                                                  'poller_hash': 'NDRBQThGcGVlcmludGZwb24gMTYzMDY4'},
                                                                                                          {       'in_nucast_pkts': '294',
                                                                                                                  'in_octets': '95531496939',
                                                                                                                  'in_ucast_pkts': '552574613',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1220',
                                                                                                                  'out_octets': '2501524407610',
                                                                                                                  'out_ucast_pkts': '1797191512',
                                                                                                                  'poller_hash': 'MjA1RTA2cGVlcmludGZwb24gMjA3NTYz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '90269331956',
                                                                                                                  'in_ucast_pkts': '340180351',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '182',
                                                                                                                  'out_octets': '1536934147351',
                                                                                                                  'out_ucast_pkts': '1224198997',
                                                                                                                  'poller_hash': 'MTI3NDA4MXBlZXJpbnRmcG9uIDI0MTg5NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '82631856915',
                                                                                                                  'in_ucast_pkts': '397562586',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149',
                                                                                                                  'out_octets': '2766669720909',
                                                                                                                  'out_ucast_pkts': '2067954737',
                                                                                                                  'poller_hash': 'MTI3NDE4RHBlZXJpbnRmcG9uIDI0MjAzMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '37983413056',
                                                                                                                  'in_ucast_pkts': '4405846786',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '149',
                                                                                                                  'out_octets': '413367901268',
                                                                                                                  'out_ucast_pkts': '368021771',
                                                                                                                  'poller_hash': 'MTI3NDExQnBlZXJpbnRmcG9uIDI0MjcxOA=='},
                                                                                                          {       'in_nucast_pkts': '132680975',
                                                                                                                  'in_octets': '288795709436',
                                                                                                                  'in_ucast_pkts': '700078139',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88917767',
                                                                                                                  'out_octets': '1532352840541',
                                                                                                                  'out_ucast_pkts': '1392416080',
                                                                                                                  'poller_hash': 'MTI3NDA0QnBlZXJpbnRmcG9uIDI0MzI0OQ=='},
                                                                                                          {       'in_nucast_pkts': '294',
                                                                                                                  'in_octets': '172009394793',
                                                                                                                  'in_ucast_pkts': '1591577958',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2027',
                                                                                                                  'out_octets': '9420862223397',
                                                                                                                  'out_ucast_pkts': '6340101249',
                                                                                                                  'poller_hash': 'MTI3NDlGNXBlZXJpbnRmcG9uIDI0NTUyOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '411076133060',
                                                                                                                  'in_ucast_pkts': '22618985456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93',
                                                                                                                  'out_octets': '920242950788',
                                                                                                                  'out_ucast_pkts': '1501113988',
                                                                                                                  'poller_hash': 'MTI5NzQxRHBlZXJpbnRmcG9uIDI1NjU0Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '51675181528',
                                                                                                                  'in_ucast_pkts': '154660088',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '624319016570',
                                                                                                                  'out_ucast_pkts': '467533153',
                                                                                                                  'poller_hash': 'MTI3MEQ4RXBlZXJpbnRmcG9uIDI3NjY0OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3MEUwNXBlZXJpbnRmcG9uIDI3Njg0NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '165953686',
                                                                                         'out_octets': '91382509421201',
                                                                                         'out_ucast_pkts': '73430907304',
                                                                                         'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/8',
                                                                                                                 'id': '7/8',
                                                                                                                 'links': [       {       'connected_time': '688472',
                                                                                                                                          'fiber_length': 1707,
                                                                                                                                          'poller_hash': 'MzIyODFEcGVlcnBvbnBvbiA1MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '18640945',
                                                                                                                                          'fiber_length': 1696,
                                                                                                                                          'poller_hash': 'MjA1REJCcGVlcnBvbnBvbiA3Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688470',
                                                                                                                                          'fiber_length': 1278,
                                                                                                                                          'poller_hash': 'NjQwQUY4cGVlcnBvbnBvbiA3OTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '67790314',
                                                                                                                                          'fiber_length': 1251,
                                                                                                                                          'poller_hash': 'NUE2NjI3cGVlcnBvbnBvbiA4NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '67790307',
                                                                                                                                          'fiber_length': 593,
                                                                                                                                          'poller_hash': 'NUE2NjhDcGVlcnBvbnBvbiA4NjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688494',
                                                                                                                                          'fiber_length': 1303,
                                                                                                                                          'poller_hash': 'NUE2NjU3cGVlcnBvbnBvbiA4NzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '2356187',
                                                                                                                                          'fiber_length': 1290,
                                                                                                                                          'poller_hash': 'NDJCRUMxcGVlcnBvbnBvbiAxMzc3MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688478',
                                                                                                                                          'fiber_length': 1669,
                                                                                                                                          'poller_hash': 'ODY5NDA3cGVlcnBvbnBvbiAxNDIzMDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688228',
                                                                                                                                          'fiber_length': 1698,
                                                                                                                                          'poller_hash': 'ODY5M0E3cGVlcnBvbnBvbiAxNDI0NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688477',
                                                                                                                                          'fiber_length': 609,
                                                                                                                                          'poller_hash': 'ODcyNTQ0cGVlcnBvbnBvbiAxNDI3MzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688475',
                                                                                                                                          'fiber_length': 1363,
                                                                                                                                          'poller_hash': 'QTlDN0Y5cGVlcnBvbnBvbiAxNDkxNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '13124259',
                                                                                                                                          'fiber_length': 526,
                                                                                                                                          'poller_hash': 'NDRBQThGcGVlcnBvbnBvbiAxNjMwNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '129285',
                                                                                                                                          'fiber_length': 1683,
                                                                                                                                          'poller_hash': 'MjA1RTA2cGVlcnBvbnBvbiAyMDc1NjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '71386248',
                                                                                                                                          'fiber_length': 1905,
                                                                                                                                          'poller_hash': 'MTI3NDA4MXBlZXJwb25wb24gMjQxODk1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688476',
                                                                                                                                          'fiber_length': 2240,
                                                                                                                                          'poller_hash': 'MTI3NDE4RHBlZXJwb25wb24gMjQyMDMy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688471',
                                                                                                                                          'fiber_length': 2094,
                                                                                                                                          'poller_hash': 'MTI3NDExQnBlZXJwb25wb24gMjQyNzE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '70267782',
                                                                                                                                          'fiber_length': 2254,
                                                                                                                                          'poller_hash': 'MTI3NDA0QnBlZXJwb25wb24gMjQzMjQ5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '24528738',
                                                                                                                                          'fiber_length': 1656,
                                                                                                                                          'poller_hash': 'MTI3NDlGNXBlZXJwb25wb24gMjQ1NTI4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688526',
                                                                                                                                          'fiber_length': 1511,
                                                                                                                                          'poller_hash': 'MTI5NzQxRHBlZXJwb25wb24gMjU2NTQ2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '688474',
                                                                                                                                          'fiber_length': 616,
                                                                                                                                          'poller_hash': 'MTI3MEQ4RXBlZXJwb25wb24gMjc2NjQ5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728},
                                                                                                                                  {       'connected_time': '10305336',
                                                                                                                                          'fiber_length': 3095,
                                                                                                                                          'poller_hash': 'MTI3MEUwNXBlZXJwb25wb24gMjc2ODQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.728}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/8',
                                                                                                                 'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNTEyNjAzNzAx',
                                                           'serial_number': '211512603701',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991324'},
                                                   {       'id': '8',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '8/1',
                                                                                         'in_discards': '108105',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '49862734',
                                                                                         'in_octets': '55728231726017',
                                                                                         'in_ucast_pkts': '74178651375',
                                                                                         'links': [       {       'in_nucast_pkts': '12611',
                                                                                                                  'in_octets': '33243888094',
                                                                                                                  'in_ucast_pkts': '53467695',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1462042',
                                                                                                                  'out_octets': '228423413867',
                                                                                                                  'out_ucast_pkts': '172118685',
                                                                                                                  'poller_hash': 'QzgzQzAycGVlcmludGZwb24gNA=='},
                                                                                                          {       'in_nucast_pkts': '9664783',
                                                                                                                  'in_octets': '1259326056626',
                                                                                                                  'in_ucast_pkts': '7656200600',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18519500940',
                                                                                                                  'out_octets': '52865383785210',
                                                                                                                  'out_ucast_pkts': '20448640608',
                                                                                                                  'poller_hash': 'MzIxODU2cGVlcmludGZwb24gOQ=='},
                                                                                                          {       'in_nucast_pkts': '56266',
                                                                                                                  'in_octets': '8044049789',
                                                                                                                  'in_ucast_pkts': '34825580',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6626651',
                                                                                                                  'out_octets': '272900419322',
                                                                                                                  'out_ucast_pkts': '198706415',
                                                                                                                  'poller_hash': 'ODdGOTY1cGVlcmludGZwb24gMTk='},
                                                                                                          {       'in_nucast_pkts': '12',
                                                                                                                  'in_octets': '214801706814',
                                                                                                                  'in_ucast_pkts': '384323594',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18150621',
                                                                                                                  'out_octets': '1085395387072',
                                                                                                                  'out_ucast_pkts': '1002837189',
                                                                                                                  'poller_hash': 'MTM4MTNCcGVlcmludGZwb24gODE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1359891031505',
                                                                                                                  'in_ucast_pkts': '2547837759',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2846958216140',
                                                                                                                  'out_ucast_pkts': '3833086838',
                                                                                                                  'poller_hash': 'Q0MwNzRwZWVyaW50ZnBvbiA4Mg=='},
                                                                                                          {       'in_nucast_pkts': '300',
                                                                                                                  'in_octets': '552884068090',
                                                                                                                  'in_ucast_pkts': '4654033814',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53874366',
                                                                                                                  'out_octets': '14724436843209',
                                                                                                                  'out_ucast_pkts': '10196801527',
                                                                                                                  'poller_hash': 'MTYwRDE1cGVlcmludGZwb24gMTA2'},
                                                                                                          {       'in_nucast_pkts': '25',
                                                                                                                  'in_octets': '4701395179',
                                                                                                                  'in_ucast_pkts': '13585259',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11147468',
                                                                                                                  'out_octets': '60961203820',
                                                                                                                  'out_ucast_pkts': '49303785',
                                                                                                                  'poller_hash': 'MjM3QTI1cGVlcmludGZwb24gMTM4'},
                                                                                                          {       'in_nucast_pkts': '25123',
                                                                                                                  'in_octets': '380735961132',
                                                                                                                  'in_ucast_pkts': '912895999',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '32396207',
                                                                                                                  'out_octets': '2030780513366',
                                                                                                                  'out_ucast_pkts': '1779946935',
                                                                                                                  'poller_hash': 'QzU4NTJwZWVyaW50ZnBvbiAyNDg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2622044806959',
                                                                                                                  'in_ucast_pkts': '2377048695',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1670671312911',
                                                                                                                  'out_ucast_pkts': '2188061089',
                                                                                                                  'poller_hash': 'Mzc4ODJDcGVlcmludGZwb24gMzEy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '514635331522',
                                                                                                                  'in_ucast_pkts': '219588034636',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '1655412250790',
                                                                                                                  'out_ucast_pkts': '1452846242',
                                                                                                                  'poller_hash': 'Mjk4M0Y4cGVlcmludGZwb24gNDg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '198810881412',
                                                                                                                  'in_ucast_pkts': '8802932905',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '167',
                                                                                                                  'out_octets': '297116772035',
                                                                                                                  'out_ucast_pkts': '361993592',
                                                                                                                  'poller_hash': 'MkIwMUM4cGVlcmludGZwb24gNDk5'},
                                                                                                          {       'in_nucast_pkts': '59',
                                                                                                                  'in_octets': '1533697095',
                                                                                                                  'in_ucast_pkts': '5865607',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464395',
                                                                                                                  'out_octets': '40634573129',
                                                                                                                  'out_ucast_pkts': '32570721',
                                                                                                                  'poller_hash': 'MkJEMjY0cGVlcmludGZwb24gNTEz'},
                                                                                                          {       'in_nucast_pkts': '723',
                                                                                                                  'in_octets': '3609002311',
                                                                                                                  'in_ucast_pkts': '25786537',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464140',
                                                                                                                  'out_octets': '117601205758',
                                                                                                                  'out_ucast_pkts': '81141100',
                                                                                                                  'poller_hash': 'NTEyNUNBcGVlcmludGZwb24gNTQz'},
                                                                                                          {       'in_nucast_pkts': '520',
                                                                                                                  'in_octets': '370146386',
                                                                                                                  'in_ucast_pkts': '1927748',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464074',
                                                                                                                  'out_octets': '4018493629',
                                                                                                                  'out_ucast_pkts': '3494111',
                                                                                                                  'poller_hash': 'NTEyNTc2cGVlcmludGZwb24gNTY4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzVFOTcycGVlcmludGZwb24gNzQ4'},
                                                                                                          {       'in_nucast_pkts': '344908',
                                                                                                                  'in_octets': '14319180518',
                                                                                                                  'in_ucast_pkts': '78349702',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39761391',
                                                                                                                  'out_octets': '1169816886755',
                                                                                                                  'out_ucast_pkts': '779128894',
                                                                                                                  'poller_hash': 'MzVFOTRGcGVlcmludGZwb24gNzUy'},
                                                                                                          {       'in_nucast_pkts': '48028',
                                                                                                                  'in_octets': '64079122534',
                                                                                                                  'in_ucast_pkts': '205269869',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '80843200',
                                                                                                                  'out_octets': '316883391400',
                                                                                                                  'out_ucast_pkts': '289230546',
                                                                                                                  'poller_hash': 'Mzc4ODMycGVlcmludGZwb24gNzU2'},
                                                                                                          {       'in_nucast_pkts': '14009',
                                                                                                                  'in_octets': '1481818214',
                                                                                                                  'in_ucast_pkts': '10352205',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464251',
                                                                                                                  'out_octets': '32250818324',
                                                                                                                  'out_ucast_pkts': '24708180',
                                                                                                                  'poller_hash': 'QzgzOTJFcGVlcmludGZwb24gMTgwMjYy'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '8/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '16908341035',
                                                                                         'out_octets': '122727074459376',
                                                                                         'out_ucast_pkts': '93370971631',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/1',
                                                                                                                 'id': '8/1',
                                                                                                                 'links': [       {       'connected_time': '688481',
                                                                                                                                          'fiber_length': 1250,
                                                                                                                                          'poller_hash': 'QzgzQzAycGVlcnBvbnBvbiA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '21816923',
                                                                                                                                          'fiber_length': 1161,
                                                                                                                                          'poller_hash': 'MzIxODU2cGVlcnBvbnBvbiA5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '3113758',
                                                                                                                                          'fiber_length': 910,
                                                                                                                                          'poller_hash': 'ODdGOTY1cGVlcnBvbnBvbiAxOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '10111014',
                                                                                                                                          'fiber_length': 1649,
                                                                                                                                          'poller_hash': 'MTM4MTNCcGVlcnBvbnBvbiA4MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '47499',
                                                                                                                                          'fiber_length': 1228,
                                                                                                                                          'poller_hash': 'Q0MwNzRwZWVycG9ucG9uIDgy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '29986028',
                                                                                                                                          'fiber_length': 801,
                                                                                                                                          'poller_hash': 'MTYwRDE1cGVlcnBvbnBvbiAxMDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '6129406',
                                                                                                                                          'fiber_length': 902,
                                                                                                                                          'poller_hash': 'MjM3QTI1cGVlcnBvbnBvbiAxMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '18165231',
                                                                                                                                          'fiber_length': 1246,
                                                                                                                                          'poller_hash': 'QzU4NTJwZWVycG9ucG9uIDI0OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '47621',
                                                                                                                                          'fiber_length': 591,
                                                                                                                                          'poller_hash': 'Mzc4ODJDcGVlcnBvbnBvbiAzMTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '688483',
                                                                                                                                          'fiber_length': 671,
                                                                                                                                          'poller_hash': 'Mjk4M0Y4cGVlcnBvbnBvbiA0ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '13238750',
                                                                                                                                          'fiber_length': 1181,
                                                                                                                                          'poller_hash': 'MkIwMUM4cGVlcnBvbnBvbiA0OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '688481',
                                                                                                                                          'fiber_length': 1575,
                                                                                                                                          'poller_hash': 'MkJEMjY0cGVlcnBvbnBvbiA1MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '688492',
                                                                                                                                          'fiber_length': 1918,
                                                                                                                                          'poller_hash': 'NTEyNUNBcGVlcnBvbnBvbiA1NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '688500',
                                                                                                                                          'fiber_length': 1259,
                                                                                                                                          'poller_hash': 'NTEyNTc2cGVlcnBvbnBvbiA1Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '15499147',
                                                                                                                                          'fiber_length': 685,
                                                                                                                                          'poller_hash': 'MzVFOTcycGVlcnBvbnBvbiA3NDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '29575204',
                                                                                                                                          'fiber_length': 1188,
                                                                                                                                          'poller_hash': 'MzVFOTRGcGVlcnBvbnBvbiA3NTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '41056',
                                                                                                                                          'fiber_length': 1357,
                                                                                                                                          'poller_hash': 'Mzc4ODMycGVlcnBvbnBvbiA3NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725},
                                                                                                                                  {       'connected_time': '688489',
                                                                                                                                          'fiber_length': 682,
                                                                                                                                          'poller_hash': 'QzgzOTJFcGVlcnBvbnBvbiAxODAyNjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.725}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/1',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '8/2',
                                                                                         'in_discards': '145718',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '6920929',
                                                                                         'in_octets': '25144199881569',
                                                                                         'in_ucast_pkts': '87144906681',
                                                                                         'links': [       {       'in_nucast_pkts': '14562',
                                                                                                                  'in_octets': '20565259583',
                                                                                                                  'in_ucast_pkts': '31378274',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464003',
                                                                                                                  'out_octets': '120910774583',
                                                                                                                  'out_ucast_pkts': '99573494',
                                                                                                                  'poller_hash': 'MzIxNzdGcGVlcmludGZwb24gNTI='},
                                                                                                          {       'in_nucast_pkts': '859430',
                                                                                                                  'in_octets': '863337080838',
                                                                                                                  'in_ucast_pkts': '6858055097',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '223611074',
                                                                                                                  'out_octets': '36140263398297',
                                                                                                                  'out_ucast_pkts': '25007141617',
                                                                                                                  'poller_hash': 'MjYwRTk3cGVlcmludGZwb24gODg='},
                                                                                                          {       'in_nucast_pkts': '18301',
                                                                                                                  'in_octets': '272132073649',
                                                                                                                  'in_ucast_pkts': '424512067',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23369984',
                                                                                                                  'out_octets': '71580918886',
                                                                                                                  'out_ucast_pkts': '371286470',
                                                                                                                  'poller_hash': 'MjYwRTZBcGVlcmludGZwb24gODk='},
                                                                                                          {       'in_nucast_pkts': '4747236',
                                                                                                                  'in_octets': '4409333799',
                                                                                                                  'in_ucast_pkts': '27478767',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464037',
                                                                                                                  'out_octets': '94843238260',
                                                                                                                  'out_ucast_pkts': '66732287',
                                                                                                                  'poller_hash': 'MjhGNjUzcGVlcmludGZwb24gMTEw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '121003374091',
                                                                                                                  'in_ucast_pkts': '409554174',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '266',
                                                                                                                  'out_octets': '2069035265272',
                                                                                                                  'out_ucast_pkts': '1502300827',
                                                                                                                  'poller_hash': 'Mjk2RkM0cGVlcmludGZwb24gMTEx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '278235884731',
                                                                                                                  'in_ucast_pkts': '95575503828',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '4189516608357',
                                                                                                                  'out_ucast_pkts': '3323665271',
                                                                                                                  'poller_hash': 'MUQwNzE4cGVlcmludGZwb24gMTI4'},
                                                                                                          {       'in_nucast_pkts': '5369380',
                                                                                                                  'in_octets': '146484193517',
                                                                                                                  'in_ucast_pkts': '497037360',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9759374689',
                                                                                                                  'out_octets': '14992290919533',
                                                                                                                  'out_ucast_pkts': '1491665051',
                                                                                                                  'poller_hash': 'MjM3QUU1cGVlcmludGZwb24gMTQw'},
                                                                                                          {       'in_nucast_pkts': '19854994',
                                                                                                                  'in_octets': '322339857679',
                                                                                                                  'in_ucast_pkts': '2594330963',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27275911027',
                                                                                                                  'out_octets': '50997162115133',
                                                                                                                  'out_ucast_pkts': '10060836497',
                                                                                                                  'poller_hash': 'Mjk4NTc2cGVlcmludGZwb24gMTk3'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '6190098295',
                                                                                                                  'in_ucast_pkts': '46745118',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464114',
                                                                                                                  'out_octets': '221669410778',
                                                                                                                  'out_ucast_pkts': '156648617',
                                                                                                                  'poller_hash': 'MkJENjkzcGVlcmludGZwb24gMTk4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '245344089298',
                                                                                                                  'in_ucast_pkts': '1992094469',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '266',
                                                                                                                  'out_octets': '12033515352565',
                                                                                                                  'out_ucast_pkts': '8283567229',
                                                                                                                  'poller_hash': 'MUQwNjhDcGVlcmludGZwb24gMjAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '461477668147',
                                                                                                                  'in_ucast_pkts': '1370927236',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '3091078074513',
                                                                                                                  'out_ucast_pkts': '2608557638',
                                                                                                                  'poller_hash': 'Mjk4MzRDcGVlcmludGZwb24gMjAz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '361785983279',
                                                                                                                  'in_ucast_pkts': '49062472815',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '12093532033819',
                                                                                                                  'out_ucast_pkts': '8917871445',
                                                                                                                  'poller_hash': 'MkIwNUQ3cGVlcmludGZwb24gMjc0'},
                                                                                                          {       'in_nucast_pkts': '15558',
                                                                                                                  'in_octets': '28926689067',
                                                                                                                  'in_ucast_pkts': '34002267',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464037',
                                                                                                                  'out_octets': '69509774694',
                                                                                                                  'out_ucast_pkts': '59167842',
                                                                                                                  'poller_hash': 'MzAwNjIzcGVlcmludGZwb24gMjg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '271686873768',
                                                                                                                  'in_ucast_pkts': '108230933433',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '125',
                                                                                                                  'out_octets': '2011638086373',
                                                                                                                  'out_ucast_pkts': '1738497463',
                                                                                                                  'poller_hash': 'MjA1RTIwcGVlcmludGZwb24gMzA1'},
                                                                                                          {       'in_nucast_pkts': '10014',
                                                                                                                  'in_octets': '298040567717',
                                                                                                                  'in_ucast_pkts': '1568677074',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148485926',
                                                                                                                  'out_octets': '11058346003464',
                                                                                                                  'out_ucast_pkts': '8112337723',
                                                                                                                  'poller_hash': 'MkIwNDIxcGVlcmludGZwb24gNDM5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '351828381130',
                                                                                                                  'in_ucast_pkts': '1310709048',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '191',
                                                                                                                  'out_octets': '5538457341459',
                                                                                                                  'out_ucast_pkts': '4058871678',
                                                                                                                  'poller_hash': 'MjA1REUzcGVlcmludGZwb24gNDUw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '23178969172',
                                                                                                                  'in_ucast_pkts': '101667263',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '344350243556',
                                                                                                                  'out_ucast_pkts': '317022249',
                                                                                                                  'poller_hash': 'MkIwMTlDcGVlcmludGZwb24gNDYx'},
                                                                                                          {       'in_nucast_pkts': '10605',
                                                                                                                  'in_octets': '195658584179',
                                                                                                                  'in_ucast_pkts': '442992298',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27442082',
                                                                                                                  'out_octets': '751085708913',
                                                                                                                  'out_ucast_pkts': '702591268',
                                                                                                                  'poller_hash': 'MzVFQTg3cGVlcmludGZwb24gNDYy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '129256537503',
                                                                                                                  'in_ucast_pkts': '1188033406',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66',
                                                                                                                  'out_octets': '6808732116722',
                                                                                                                  'out_ucast_pkts': '4572221983',
                                                                                                                  'poller_hash': 'NDJDMEFGcGVlcmludGZwb24gNDYz'},
                                                                                                          {       'in_nucast_pkts': '11860',
                                                                                                                  'in_octets': '11323735954',
                                                                                                                  'in_ucast_pkts': '58066586',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119990',
                                                                                                                  'out_octets': '250244768355',
                                                                                                                  'out_ucast_pkts': '180866581',
                                                                                                                  'poller_hash': 'NEZBMjVEcGVlcmludGZwb24gNTIw'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '14386312082',
                                                                                                                  'in_ucast_pkts': '42866565',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464226',
                                                                                                                  'out_octets': '178678932006',
                                                                                                                  'out_ucast_pkts': '137710738',
                                                                                                                  'poller_hash': 'M0I2MjAwcGVlcmludGZwb24gNzA1'},
                                                                                                          {       'in_nucast_pkts': '15248',
                                                                                                                  'in_octets': '11612302394',
                                                                                                                  'in_ucast_pkts': '30553566',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464114',
                                                                                                                  'out_octets': '79214217446',
                                                                                                                  'out_ucast_pkts': '74370487',
                                                                                                                  'poller_hash': 'MzAwRjAwcGVlcmludGZwb24gNzA2'},
                                                                                                          {       'in_nucast_pkts': '14490',
                                                                                                                  'in_octets': '5979215056',
                                                                                                                  'in_ucast_pkts': '38867197',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463978',
                                                                                                                  'out_octets': '150299729619',
                                                                                                                  'out_ucast_pkts': '108829279',
                                                                                                                  'poller_hash': 'MjA1REJEcGVlcmludGZwb24gNzIx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '210530839960',
                                                                                                                  'in_ucast_pkts': '1235552554',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '266',
                                                                                                                  'out_octets': '5116382390451',
                                                                                                                  'out_ucast_pkts': '3706815966',
                                                                                                                  'poller_hash': 'MzIxOEU1cGVlcmludGZwb24gNzI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '40494736594',
                                                                                                                  'in_ucast_pkts': '211082762',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '1403644700367',
                                                                                                                  'out_ucast_pkts': '1031892037',
                                                                                                                  'poller_hash': 'MjA1RDlCcGVlcmludGZwb24gNzI1'},
                                                                                                          {       'in_nucast_pkts': '37630',
                                                                                                                  'in_octets': '752670577345',
                                                                                                                  'in_ucast_pkts': '3083592002',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '80818399',
                                                                                                                  'out_octets': '18310522966100',
                                                                                                                  'out_ucast_pkts': '13089566459',
                                                                                                                  'poller_hash': 'QkYyNTBFcGVlcmludGZwb24gMTc0Mzcw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '311552759272',
                                                                                                                  'in_ucast_pkts': '1688823049',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '9426640064411',
                                                                                                                  'out_ucast_pkts': '6515139743',
                                                                                                                  'poller_hash': 'QzgzQjdEcGVlcmludGZwb24gMTg2Njg3'},
                                                                                                          {       'in_nucast_pkts': '12162',
                                                                                                                  'in_octets': '9030959910',
                                                                                                                  'in_ucast_pkts': '13265086',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465077',
                                                                                                                  'out_octets': '57790012284',
                                                                                                                  'out_ucast_pkts': '47386820',
                                                                                                                  'poller_hash': 'MkIwNDcxcGVlcmludGZwb24gMjUxNjgz'},
                                                                                                          {       'in_nucast_pkts': '12039',
                                                                                                                  'in_octets': '7528115470',
                                                                                                                  'in_ucast_pkts': '38881411',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465208',
                                                                                                                  'out_octets': '346740958105',
                                                                                                                  'out_ucast_pkts': '241043003',
                                                                                                                  'poller_hash': 'MTE3REIwQnBlZXJpbnRmcG9uIDI1NDE5Mw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '8/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '110278742029',
                                                                                         'out_octets': '520396194918414',
                                                                                         'out_ucast_pkts': '270415089295',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/2',
                                                                                                                 'id': '8/2',
                                                                                                                 'links': [       {       'connected_time': '688478',
                                                                                                                                          'fiber_length': 1013,
                                                                                                                                          'poller_hash': 'MzIxNzdGcGVlcnBvbnBvbiA1Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '15469743',
                                                                                                                                          'fiber_length': 1368,
                                                                                                                                          'poller_hash': 'MjYwRTk3cGVlcnBvbnBvbiA4OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '13218993',
                                                                                                                                          'fiber_length': 1369,
                                                                                                                                          'poller_hash': 'MjYwRTZBcGVlcnBvbnBvbiA4OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688478',
                                                                                                                                          'fiber_length': 864,
                                                                                                                                          'poller_hash': 'MjhGNjUzcGVlcnBvbnBvbiAxMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '2138497',
                                                                                                                                          'fiber_length': 1090,
                                                                                                                                          'poller_hash': 'Mjk2RkM0cGVlcnBvbnBvbiAxMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '21817027',
                                                                                                                                          'fiber_length': 1003,
                                                                                                                                          'poller_hash': 'MUQwNzE4cGVlcnBvbnBvbiAxMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '15498457',
                                                                                                                                          'fiber_length': 992,
                                                                                                                                          'poller_hash': 'MjM3QUU1cGVlcnBvbnBvbiAxNDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '65075966',
                                                                                                                                          'fiber_length': 1018,
                                                                                                                                          'poller_hash': 'Mjk4NTc2cGVlcnBvbnBvbiAxOTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688479',
                                                                                                                                          'fiber_length': 1011,
                                                                                                                                          'poller_hash': 'MkJENjkzcGVlcnBvbnBvbiAxOTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '63688893',
                                                                                                                                          'fiber_length': 1083,
                                                                                                                                          'poller_hash': 'MUQwNjhDcGVlcnBvbnBvbiAyMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '10799232',
                                                                                                                                          'fiber_length': 1154,
                                                                                                                                          'poller_hash': 'Mjk4MzRDcGVlcnBvbnBvbiAyMDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '72951754',
                                                                                                                                          'fiber_length': 1098,
                                                                                                                                          'poller_hash': 'MkIwNUQ3cGVlcnBvbnBvbiAyNzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688489',
                                                                                                                                          'fiber_length': 899,
                                                                                                                                          'poller_hash': 'MzAwNjIzcGVlcnBvbnBvbiAyODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '5394487',
                                                                                                                                          'fiber_length': 1049,
                                                                                                                                          'poller_hash': 'MjA1RTIwcGVlcnBvbnBvbiAzMDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '21816307',
                                                                                                                                          'fiber_length': 992,
                                                                                                                                          'poller_hash': 'MkIwNDIxcGVlcnBvbnBvbiA0Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '6643424',
                                                                                                                                          'fiber_length': 878,
                                                                                                                                          'poller_hash': 'MjA1REUzcGVlcnBvbnBvbiA0NTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '15498618',
                                                                                                                                          'fiber_length': 1310,
                                                                                                                                          'poller_hash': 'MkIwMTlDcGVlcnBvbnBvbiA0NjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '12047422',
                                                                                                                                          'fiber_length': 1283,
                                                                                                                                          'poller_hash': 'MzVFQTg3cGVlcnBvbnBvbiA0NjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '11594465',
                                                                                                                                          'fiber_length': 1280,
                                                                                                                                          'poller_hash': 'NDJDMEFGcGVlcnBvbnBvbiA0NjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688481',
                                                                                                                                          'fiber_length': 1000,
                                                                                                                                          'poller_hash': 'NEZBMjVEcGVlcnBvbnBvbiA1MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688492',
                                                                                                                                          'fiber_length': 943,
                                                                                                                                          'poller_hash': 'M0I2MjAwcGVlcnBvbnBvbiA3MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688487',
                                                                                                                                          'fiber_length': 1705,
                                                                                                                                          'poller_hash': 'MzAwRjAwcGVlcnBvbnBvbiA3MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688491',
                                                                                                                                          'fiber_length': 1029,
                                                                                                                                          'poller_hash': 'MjA1REJEcGVlcnBvbnBvbiA3MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '15498720',
                                                                                                                                          'fiber_length': 1905,
                                                                                                                                          'poller_hash': 'MzIxOEU1cGVlcnBvbnBvbiA3MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688491',
                                                                                                                                          'fiber_length': 691,
                                                                                                                                          'poller_hash': 'MjA1RDlCcGVlcnBvbnBvbiA3MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '40565',
                                                                                                                                          'fiber_length': 984,
                                                                                                                                          'poller_hash': 'QkYyNTBFcGVlcnBvbnBvbiAxNzQzNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688486',
                                                                                                                                          'fiber_length': 732,
                                                                                                                                          'poller_hash': 'QzgzQjdEcGVlcnBvbnBvbiAxODY2ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688461',
                                                                                                                                          'fiber_length': 1121,
                                                                                                                                          'poller_hash': 'MkIwNDcxcGVlcnBvbnBvbiAyNTE2ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741},
                                                                                                                                  {       'connected_time': '688493',
                                                                                                                                          'fiber_length': 1009,
                                                                                                                                          'poller_hash': 'MTE3REIwQnBlZXJwb25wb24gMjU0MTkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.741}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/2',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '8/3',
                                                                                         'in_discards': '1222',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '915008',
                                                                                         'in_octets': '15773143581038',
                                                                                         'in_ucast_pkts': '26283143271',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '319346884849',
                                                                                                                  'in_ucast_pkts': '1230690026',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '4619396932272',
                                                                                                                  'out_ucast_pkts': '3426972138',
                                                                                                                  'poller_hash': 'MjA1RERDcGVlcmludGZwb24gNTQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '229627382686',
                                                                                                                  'in_ucast_pkts': '674914488',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '941148937318',
                                                                                                                  'out_ucast_pkts': '1143285096',
                                                                                                                  'poller_hash': 'MTMwMEM5cGVlcmludGZwb24gNTcz'},
                                                                                                          {       'in_nucast_pkts': '9034',
                                                                                                                  'in_octets': '199839168863',
                                                                                                                  'in_ucast_pkts': '403694400',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11805433',
                                                                                                                  'out_octets': '330441167530',
                                                                                                                  'out_ucast_pkts': '455761740',
                                                                                                                  'poller_hash': 'Mzc4OEJDcGVlcmludGZwb24gODAx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '127986034978',
                                                                                                                  'in_ucast_pkts': '718741852',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '2627448240657',
                                                                                                                  'out_ucast_pkts': '2028556457',
                                                                                                                  'poller_hash': 'MjA1REY0cGVlcmludGZwb24gODQ2'},
                                                                                                          {       'in_nucast_pkts': '19',
                                                                                                                  'in_octets': '26621368086',
                                                                                                                  'in_ucast_pkts': '99649732',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463863',
                                                                                                                  'out_octets': '320624456570',
                                                                                                                  'out_ucast_pkts': '252674428',
                                                                                                                  'poller_hash': 'ODcyNTBGcGVlcmludGZwb24gMTQzMDMz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '285308910333',
                                                                                                                  'in_ucast_pkts': '2057289254',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '8041361624400',
                                                                                                                  'out_ucast_pkts': '5639011386',
                                                                                                                  'poller_hash': 'OUMzMDIzcGVlcmludGZwb24gMTQ0MDEz'},
                                                                                                          {       'in_nucast_pkts': '13636',
                                                                                                                  'in_octets': '4790245659',
                                                                                                                  'in_ucast_pkts': '21722447',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464139',
                                                                                                                  'out_octets': '83355296149',
                                                                                                                  'out_ucast_pkts': '66293349',
                                                                                                                  'poller_hash': 'MUQwMzM3cGVlcmludGZwb24gMTYzNjUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '6264289449100',
                                                                                                                  'in_ucast_pkts': '481787634452',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '74',
                                                                                                                  'out_octets': '3852711187899',
                                                                                                                  'out_ucast_pkts': '4775909160',
                                                                                                                  'poller_hash': 'QjBBQUU4cGVlcmludGZwb24gMjE4ODQ1'},
                                                                                                          {       'in_nucast_pkts': '14124',
                                                                                                                  'in_octets': '15478487562',
                                                                                                                  'in_ucast_pkts': '63443404',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465328',
                                                                                                                  'out_octets': '208552746194',
                                                                                                                  'out_ucast_pkts': '152798573',
                                                                                                                  'poller_hash': 'MTI5NzQ4MnBlZXJpbnRmcG9uIDI1NjMzMw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '8/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '125919187',
                                                                                         'out_octets': '59051040788776',
                                                                                         'out_ucast_pkts': '47964386773',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/3',
                                                                                                                 'id': '8/3',
                                                                                                                 'links': [       {       'connected_time': '688487',
                                                                                                                                          'fiber_length': 549,
                                                                                                                                          'poller_hash': 'MjA1RERDcGVlcnBvbnBvbiA1NDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '47655',
                                                                                                                                          'fiber_length': 468,
                                                                                                                                          'poller_hash': 'MTMwMEM5cGVlcnBvbnBvbiA1NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '6500553',
                                                                                                                                          'fiber_length': 542,
                                                                                                                                          'poller_hash': 'Mzc4OEJDcGVlcnBvbnBvbiA4MDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '688489',
                                                                                                                                          'fiber_length': 559,
                                                                                                                                          'poller_hash': 'MjA1REY0cGVlcnBvbnBvbiA4NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '688489',
                                                                                                                                          'fiber_length': 554,
                                                                                                                                          'poller_hash': 'ODcyNTBGcGVlcnBvbnBvbiAxNDMwMzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '688501',
                                                                                                                                          'fiber_length': 589,
                                                                                                                                          'poller_hash': 'OUMzMDIzcGVlcnBvbnBvbiAxNDQwMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '688495',
                                                                                                                                          'fiber_length': 600,
                                                                                                                                          'poller_hash': 'MUQwMzM3cGVlcnBvbnBvbiAxNjM2NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '688496',
                                                                                                                                          'fiber_length': 559,
                                                                                                                                          'poller_hash': 'QjBBQUU4cGVlcnBvbnBvbiAyMTg4NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896},
                                                                                                                                  {       'connected_time': '688485',
                                                                                                                                          'fiber_length': 550,
                                                                                                                                          'poller_hash': 'MTI5NzQ4MnBlZXJwb25wb24gMjU2MzMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.896}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/3',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '8/4',
                                                                                         'in_discards': '482419',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7044122',
                                                                                         'in_octets': '26310955840286',
                                                                                         'in_ucast_pkts': '65883168481',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'Mjk4MzAxcGVlcmludGZwb24gMTgx'},
                                                                                                          {       'in_nucast_pkts': '22',
                                                                                                                  'in_octets': '4299983979730',
                                                                                                                  'in_ucast_pkts': '4215237363',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27173523',
                                                                                                                  'out_octets': '5509617720301',
                                                                                                                  'out_ucast_pkts': '4841490339',
                                                                                                                  'poller_hash': 'NzVDMURDcGVlcmludGZwb24gMjkx'},
                                                                                                          {       'in_nucast_pkts': '57580',
                                                                                                                  'in_octets': '39830397894',
                                                                                                                  'in_ucast_pkts': '207831380',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6603163',
                                                                                                                  'out_octets': '876659593410',
                                                                                                                  'out_ucast_pkts': '669271231',
                                                                                                                  'poller_hash': 'Mjk4MkM0cGVlcmludGZwb24gMzA2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '15734926330',
                                                                                                                  'in_ucast_pkts': '98002549',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '34935407655',
                                                                                                                  'out_ucast_pkts': '170937655',
                                                                                                                  'poller_hash': 'MzhGMjJBcGVlcmludGZwb24gMzEw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2596219143096',
                                                                                                                  'in_ucast_pkts': '3119489101',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3246989932726',
                                                                                                                  'out_ucast_pkts': '3605181533',
                                                                                                                  'poller_hash': 'MkExQUQzcGVlcmludGZwb24gMzUx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '364811589803',
                                                                                                                  'in_ucast_pkts': '3185093354',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '15796985294510',
                                                                                                                  'out_ucast_pkts': '10828685213',
                                                                                                                  'poller_hash': 'Mzk2MUE4cGVlcmludGZwb24gMzUy'},
                                                                                                          {       'in_nucast_pkts': '304',
                                                                                                                  'in_octets': '54577449616',
                                                                                                                  'in_ucast_pkts': '287694172',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '80838394',
                                                                                                                  'out_octets': '1091462197915',
                                                                                                                  'out_ucast_pkts': '863368820',
                                                                                                                  'poller_hash': 'MkM5NjU2cGVlcmludGZwb24gMzcz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '69037336659',
                                                                                                                  'in_ucast_pkts': '47468127409',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1484451426441',
                                                                                                                  'out_ucast_pkts': '1055553659',
                                                                                                                  'poller_hash': 'Mjk4MUUxcGVlcmludGZwb24gNDA0'},
                                                                                                          {       'in_nucast_pkts': '953213',
                                                                                                                  'in_octets': '1067234933448',
                                                                                                                  'in_ucast_pkts': '5590110946',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148489576',
                                                                                                                  'out_octets': '23308830238623',
                                                                                                                  'out_ucast_pkts': '16955430802',
                                                                                                                  'poller_hash': 'MkIwMThGcGVlcmludGZwb24gNDA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '535192464315',
                                                                                                                  'in_ucast_pkts': '1121974635',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '191',
                                                                                                                  'out_octets': '3051362288151',
                                                                                                                  'out_ucast_pkts': '2592887450',
                                                                                                                  'poller_hash': 'MjA1REY2cGVlcmludGZwb24gNDA2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '314429498712',
                                                                                                                  'in_ucast_pkts': '82684105815',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '99',
                                                                                                                  'out_octets': '5338069717696',
                                                                                                                  'out_ucast_pkts': '7905442774',
                                                                                                                  'poller_hash': 'MkIwNDUzcGVlcmludGZwb24gNDA3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96056912124',
                                                                                                                  'in_ucast_pkts': '884047599',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '5076147345211',
                                                                                                                  'out_ucast_pkts': '3431449616',
                                                                                                                  'poller_hash': 'MjA1RENCcGVlcmludGZwb24gNDA4'},
                                                                                                          {       'in_nucast_pkts': '670535',
                                                                                                                  'in_octets': '1911101621413',
                                                                                                                  'in_ucast_pkts': '6531995068',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '93456865',
                                                                                                                  'out_octets': '16693728419846',
                                                                                                                  'out_ucast_pkts': '12027919666',
                                                                                                                  'poller_hash': 'MkIwMThEcGVlcmludGZwb24gNDEx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '122646930660',
                                                                                                                  'in_ucast_pkts': '367279556',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1655367984396',
                                                                                                                  'out_ucast_pkts': '1394650866',
                                                                                                                  'poller_hash': 'MzhGRDdGcGVlcmludGZwb24gNDEz'},
                                                                                                          {       'in_nucast_pkts': '193577',
                                                                                                                  'in_octets': '1296768991',
                                                                                                                  'in_ucast_pkts': '4952783',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464171',
                                                                                                                  'out_octets': '30954048075',
                                                                                                                  'out_ucast_pkts': '23817324',
                                                                                                                  'poller_hash': 'MzAwRENCcGVlcmludGZwb24gNDE0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '49255058290',
                                                                                                                  'in_ucast_pkts': '303845739',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '4053133187724',
                                                                                                                  'out_ucast_pkts': '2764056035',
                                                                                                                  'poller_hash': 'MkIwMTk1cGVlcmludGZwb24gNDE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '230618543373',
                                                                                                                  'in_ucast_pkts': '31162658000',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '233',
                                                                                                                  'out_octets': '7522079505074',
                                                                                                                  'out_ucast_pkts': '5205587570',
                                                                                                                  'poller_hash': 'MkIwMTk0cGVlcmludGZwb24gNDE5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '59152700555',
                                                                                                                  'in_ucast_pkts': '101046137',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '200',
                                                                                                                  'out_octets': '182533714578',
                                                                                                                  'out_ucast_pkts': '187818957',
                                                                                                                  'poller_hash': 'MkIwMTlCcGVlcmludGZwb24gNDIx'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '6029158564',
                                                                                                                  'in_ucast_pkts': '35230928',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1275653',
                                                                                                                  'out_octets': '156149665458',
                                                                                                                  'out_ucast_pkts': '113992920',
                                                                                                                  'poller_hash': 'ODY5NDMycGVlcmludGZwb24gNDI2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '549598168934',
                                                                                                                  'in_ucast_pkts': '27059301396',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '200',
                                                                                                                  'out_octets': '3928631845757',
                                                                                                                  'out_ucast_pkts': '3167871949',
                                                                                                                  'poller_hash': 'MkIwNDNDcGVlcmludGZwb24gNDI3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '107771545222',
                                                                                                                  'in_ucast_pkts': '576014792',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '179',
                                                                                                                  'out_octets': '1875264942527',
                                                                                                                  'out_ucast_pkts': '1553528033',
                                                                                                                  'poller_hash': 'MkFGQjc0cGVlcmludGZwb24gNDMw'},
                                                                                                          {       'in_nucast_pkts': '9460285',
                                                                                                                  'in_octets': '29428829106',
                                                                                                                  'in_ucast_pkts': '343820560',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27029088242',
                                                                                                                  'out_octets': '37653936252096',
                                                                                                                  'out_ucast_pkts': '1222795098',
                                                                                                                  'poller_hash': 'MkIwMUI2cGVlcmludGZwb24gNDU0'},
                                                                                                          {       'in_nucast_pkts': '237',
                                                                                                                  'in_octets': '271826119047',
                                                                                                                  'in_ucast_pkts': '2540438645',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '92619877',
                                                                                                                  'out_octets': '12189714986270',
                                                                                                                  'out_ucast_pkts': '8273739637',
                                                                                                                  'poller_hash': 'MkIwMURFcGVlcmludGZwb24gNDU1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '146587691812',
                                                                                                                  'in_ucast_pkts': '737351817',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '174',
                                                                                                                  'out_octets': '5030749282825',
                                                                                                                  'out_ucast_pkts': '3462809234',
                                                                                                                  'poller_hash': 'ODcyNDFFcGVlcmludGZwb24gNTQ4'},
                                                                                                          {       'in_nucast_pkts': '12033',
                                                                                                                  'in_octets': '10687287610',
                                                                                                                  'in_ucast_pkts': '35022406',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464272',
                                                                                                                  'out_octets': '160399446628',
                                                                                                                  'out_ucast_pkts': '119028605',
                                                                                                                  'poller_hash': 'Q0JEMUU4cGVlcmludGZwb24gMTY0ODE1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '104414989899',
                                                                                                                  'in_ucast_pkts': '565865925',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '125',
                                                                                                                  'out_octets': '1573886685707',
                                                                                                                  'out_ucast_pkts': '1229730998',
                                                                                                                  'poller_hash': 'MjA1RTA1cGVlcmludGZwb24gMTg5ODI4'},
                                                                                                          {       'in_nucast_pkts': '38271',
                                                                                                                  'in_octets': '1330004258',
                                                                                                                  'in_ucast_pkts': '17444831',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15664490',
                                                                                                                  'out_octets': '104185937041',
                                                                                                                  'out_ucast_pkts': '69671748',
                                                                                                                  'poller_hash': 'MTI3NTA3NnBlZXJpbnRmcG9uIDIyMDM1OQ=='},
                                                                                                          {       'in_nucast_pkts': '14335',
                                                                                                                  'in_octets': '11664189642',
                                                                                                                  'in_ucast_pkts': '84568086',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1650852',
                                                                                                                  'out_octets': '447240729873',
                                                                                                                  'out_ucast_pkts': '307247602',
                                                                                                                  'poller_hash': 'MTI3NDlGMnBlZXJpbnRmcG9uIDIzMjU1OA=='},
                                                                                                          {       'in_nucast_pkts': '148',
                                                                                                                  'in_octets': '191902925754',
                                                                                                                  'in_ucast_pkts': '455470187',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '36317922',
                                                                                                                  'out_octets': '1381553424933',
                                                                                                                  'out_ucast_pkts': '1144765405',
                                                                                                                  'poller_hash': 'MTI3NDg2OXBlZXJpbnRmcG9uIDI3MjEyOQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '8/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '57341829589',
                                                                                         'out_octets': '322959281082031',
                                                                                         'out_ucast_pkts': '180511364833',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/4',
                                                                                                                 'id': '8/4',
                                                                                                                 'links': [       {       'connected_time': '21815991',
                                                                                                                                          'fiber_length': 2020,
                                                                                                                                          'poller_hash': 'Mjk4MzAxcGVlcnBvbnBvbiAxODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '15498467',
                                                                                                                                          'fiber_length': 1600,
                                                                                                                                          'poller_hash': 'NzVDMURDcGVlcnBvbnBvbiAyOTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '3113777',
                                                                                                                                          'fiber_length': 1202,
                                                                                                                                          'poller_hash': 'Mjk4MkM0cGVlcnBvbnBvbiAzMDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '48243',
                                                                                                                                          'fiber_length': 1277,
                                                                                                                                          'poller_hash': 'MzhGMjJBcGVlcnBvbnBvbiAzMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '47858',
                                                                                                                                          'fiber_length': 1923,
                                                                                                                                          'poller_hash': 'MkExQUQzcGVlcnBvbnBvbiAzNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '13216626',
                                                                                                                                          'fiber_length': 1224,
                                                                                                                                          'poller_hash': 'Mzk2MUE4cGVlcnBvbnBvbiAzNTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '40487',
                                                                                                                                          'fiber_length': 1322,
                                                                                                                                          'poller_hash': 'MkM5NjU2cGVlcnBvbnBvbiAzNzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '82401',
                                                                                                                                          'fiber_length': 1195,
                                                                                                                                          'poller_hash': 'Mjk4MUUxcGVlcnBvbnBvbiA0MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '21815964',
                                                                                                                                          'fiber_length': 1166,
                                                                                                                                          'poller_hash': 'MkIwMThGcGVlcnBvbnBvbiA0MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '21816989',
                                                                                                                                          'fiber_length': 1271,
                                                                                                                                          'poller_hash': 'MjA1REY2cGVlcnBvbnBvbiA0MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '688481',
                                                                                                                                          'fiber_length': 1927,
                                                                                                                                          'poller_hash': 'MkIwNDUzcGVlcnBvbnBvbiA0MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '65324014',
                                                                                                                                          'fiber_length': 1898,
                                                                                                                                          'poller_hash': 'MjA1RENCcGVlcnBvbnBvbiA0MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '71027301',
                                                                                                                                          'fiber_length': 1525,
                                                                                                                                          'poller_hash': 'MkIwMThEcGVlcnBvbnBvbiA0MTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '47987',
                                                                                                                                          'fiber_length': 1199,
                                                                                                                                          'poller_hash': 'MzhGRDdGcGVlcnBvbnBvbiA0MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '688497',
                                                                                                                                          'fiber_length': 1534,
                                                                                                                                          'poller_hash': 'MzAwRENCcGVlcnBvbnBvbiA0MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '13219345',
                                                                                                                                          'fiber_length': 1203,
                                                                                                                                          'poller_hash': 'MkIwMTk1cGVlcnBvbnBvbiA0MTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '21816088',
                                                                                                                                          'fiber_length': 1572,
                                                                                                                                          'poller_hash': 'MkIwMTk0cGVlcnBvbnBvbiA0MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '64764128',
                                                                                                                                          'fiber_length': 1782,
                                                                                                                                          'poller_hash': 'MkIwMTlCcGVlcnBvbnBvbiA0MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '597791',
                                                                                                                                          'fiber_length': 1872,
                                                                                                                                          'poller_hash': 'ODY5NDMycGVlcnBvbnBvbiA0MjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '4927279',
                                                                                                                                          'fiber_length': 1952,
                                                                                                                                          'poller_hash': 'MkIwNDNDcGVlcnBvbnBvbiA0Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '11510158',
                                                                                                                                          'fiber_length': 1337,
                                                                                                                                          'poller_hash': 'MkFGQjc0cGVlcnBvbnBvbiA0MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '21815553',
                                                                                                                                          'fiber_length': 1357,
                                                                                                                                          'poller_hash': 'MkIwMUI2cGVlcnBvbnBvbiA0NTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '70681266',
                                                                                                                                          'fiber_length': 1868,
                                                                                                                                          'poller_hash': 'MkIwMURFcGVlcnBvbnBvbiA0NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '688490',
                                                                                                                                          'fiber_length': 1351,
                                                                                                                                          'poller_hash': 'ODcyNDFFcGVlcnBvbnBvbiA1NDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '688500',
                                                                                                                                          'fiber_length': 1861,
                                                                                                                                          'poller_hash': 'Q0JEMUU4cGVlcnBvbnBvbiAxNjQ4MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '26374670',
                                                                                                                                          'fiber_length': 1728,
                                                                                                                                          'poller_hash': 'MjA1RTA1cGVlcnBvbnBvbiAxODk4Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '688499',
                                                                                                                                          'fiber_length': 2062,
                                                                                                                                          'poller_hash': 'MTI3NTA3NnBlZXJwb25wb24gMjIwMzU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '779210',
                                                                                                                                          'fiber_length': 1714,
                                                                                                                                          'poller_hash': 'MTI3NDlGMnBlZXJwb25wb24gMjMyNTU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618},
                                                                                                                                  {       'connected_time': '16160357',
                                                                                                                                          'fiber_length': 1421,
                                                                                                                                          'poller_hash': 'MTI3NDg2OXBlZXJwb25wb24gMjcyMTI5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.618}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/4',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '8/5',
                                                                                         'in_discards': '560494',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '6120859',
                                                                                         'in_octets': '114250327427704',
                                                                                         'in_ucast_pkts': '166300783435',
                                                                                         'links': [       {       'in_nucast_pkts': '35927039',
                                                                                                                  'in_octets': '444971703362',
                                                                                                                  'in_ucast_pkts': '1029626764',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10245610953',
                                                                                                                  'out_octets': '16346445312544',
                                                                                                                  'out_ucast_pkts': '2655212833',
                                                                                                                  'poller_hash': 'MUQwNjlEcGVlcmludGZwb24gMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '544101591545',
                                                                                                                  'in_ucast_pkts': '65887662926',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '191',
                                                                                                                  'out_octets': '4933902677574',
                                                                                                                  'out_ucast_pkts': '3745151059',
                                                                                                                  'poller_hash': 'Mjk2RkI4cGVlcmludGZwb24gMTEy'},
                                                                                                          {       'in_nucast_pkts': '11959',
                                                                                                                  'in_octets': '9525590952',
                                                                                                                  'in_ucast_pkts': '37192293',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464058',
                                                                                                                  'out_octets': '174240286849',
                                                                                                                  'out_ucast_pkts': '125727565',
                                                                                                                  'poller_hash': 'Mjk4MjQ4cGVlcmludGZwb24gMTEz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '138091907936',
                                                                                                                  'in_ucast_pkts': '838548078',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '212',
                                                                                                                  'out_octets': '3832809809129',
                                                                                                                  'out_ucast_pkts': '2762099443',
                                                                                                                  'poller_hash': 'MUQwNTBGcGVlcmludGZwb24gMTE0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '234231998907',
                                                                                                                  'in_ucast_pkts': '69628220214',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '167',
                                                                                                                  'out_octets': '4281811238213',
                                                                                                                  'out_ucast_pkts': '3200626853',
                                                                                                                  'poller_hash': 'Mjk2RkM5cGVlcmludGZwb24gMTE2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '679346643077',
                                                                                                                  'in_ucast_pkts': '1165693485',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '133',
                                                                                                                  'out_octets': '2591597955286',
                                                                                                                  'out_ucast_pkts': '1996957707',
                                                                                                                  'poller_hash': 'Mjk4MjUzcGVlcmludGZwb24gMTE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2050851863136',
                                                                                                                  'in_ucast_pkts': '93105551698',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '179',
                                                                                                                  'out_octets': '6646637458256',
                                                                                                                  'out_ucast_pkts': '6348015183',
                                                                                                                  'poller_hash': 'Mjk4MkI3cGVlcmludGZwb24gMTI1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '472871335693',
                                                                                                                  'in_ucast_pkts': '5402082284',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '191',
                                                                                                                  'out_octets': '16594035058712',
                                                                                                                  'out_ucast_pkts': '11163673444',
                                                                                                                  'poller_hash': 'MUQwNkQ3cGVlcmludGZwb24gMTI2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '145875838299',
                                                                                                                  'in_ucast_pkts': '1411159636',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '266',
                                                                                                                  'out_octets': '4761915001347',
                                                                                                                  'out_ucast_pkts': '3324317057',
                                                                                                                  'poller_hash': 'Mjk4MjhEcGVlcmludGZwb24gMTI5'},
                                                                                                          {       'in_nucast_pkts': '14957',
                                                                                                                  'in_octets': '2236458665',
                                                                                                                  'in_ucast_pkts': '5406427',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463881',
                                                                                                                  'out_octets': '26194622069',
                                                                                                                  'out_ucast_pkts': '20849784',
                                                                                                                  'poller_hash': 'MUQwNkExcGVlcmludGZwb24gMTMw'},
                                                                                                          {       'in_nucast_pkts': '15242',
                                                                                                                  'in_octets': '9331167664',
                                                                                                                  'in_ucast_pkts': '39139938',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463984',
                                                                                                                  'out_octets': '121426710105',
                                                                                                                  'out_ucast_pkts': '89219041',
                                                                                                                  'poller_hash': 'MzAwRERGcGVlcmludGZwb24gMTMx'},
                                                                                                          {       'in_nucast_pkts': '801935',
                                                                                                                  'in_octets': '381899528506',
                                                                                                                  'in_ucast_pkts': '2068877833',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76061852',
                                                                                                                  'out_octets': '9411526450184',
                                                                                                                  'out_ucast_pkts': '6736506404',
                                                                                                                  'poller_hash': 'MUQwNEYxcGVlcmludGZwb24gMTM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '644420474836',
                                                                                                                  'in_ucast_pkts': '3147027418',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '174',
                                                                                                                  'out_octets': '17346364060887',
                                                                                                                  'out_ucast_pkts': '12593649741',
                                                                                                                  'poller_hash': 'MUQwNkI5cGVlcmludGZwb24gMTQ4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '483240470794',
                                                                                                                  'in_ucast_pkts': '3429259977',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '92',
                                                                                                                  'out_octets': '12324813467874',
                                                                                                                  'out_ucast_pkts': '5157447139',
                                                                                                                  'poller_hash': 'MUQwNzE0cGVlcmludGZwb24gMTUx'},
                                                                                                          {       'in_nucast_pkts': '15211807',
                                                                                                                  'in_octets': '470760106692',
                                                                                                                  'in_ucast_pkts': '1685196461',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '32688566825',
                                                                                                                  'out_octets': '50598570120315',
                                                                                                                  'out_ucast_pkts': '5258405083',
                                                                                                                  'poller_hash': 'Mjk4MjFFcGVlcmludGZwb24gMTc0'},
                                                                                                          {       'in_nucast_pkts': '228',
                                                                                                                  'in_octets': '289317027879',
                                                                                                                  'in_ucast_pkts': '932933381',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '80889162',
                                                                                                                  'out_octets': '2344412752603',
                                                                                                                  'out_ucast_pkts': '1991617782',
                                                                                                                  'poller_hash': 'MTk5QTNBcGVlcmludGZwb24gMTc2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '566961056299',
                                                                                                                  'in_ucast_pkts': '1770494590',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7280234940563',
                                                                                                                  'out_ucast_pkts': '5492970134',
                                                                                                                  'poller_hash': 'MTc2MjRFcGVlcmludGZwb24gMTc4'},
                                                                                                          {       'in_nucast_pkts': '13063',
                                                                                                                  'in_octets': '7486637576',
                                                                                                                  'in_ucast_pkts': '26099618',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464032',
                                                                                                                  'out_octets': '82846547375',
                                                                                                                  'out_ucast_pkts': '64669109',
                                                                                                                  'poller_hash': 'Mjk4MjAwcGVlcmludGZwb24gMTg4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '889992295589',
                                                                                                                  'in_ucast_pkts': '98046241339',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '224',
                                                                                                                  'out_octets': '12862959859165',
                                                                                                                  'out_ucast_pkts': '9605761117',
                                                                                                                  'poller_hash': 'MUQwNjg4cGVlcmludGZwb24gMTk5'},
                                                                                                          {       'in_nucast_pkts': '21663',
                                                                                                                  'in_octets': '213685644321',
                                                                                                                  'in_ucast_pkts': '940607267',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '50208786',
                                                                                                                  'out_octets': '7797648554240',
                                                                                                                  'out_ucast_pkts': '5405007093',
                                                                                                                  'poller_hash': 'MjA1REQxcGVlcmludGZwb24gMjAx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3378910712772',
                                                                                                                  'in_ucast_pkts': '207147199751',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '266',
                                                                                                                  'out_octets': '17188440975244',
                                                                                                                  'out_ucast_pkts': '14904136612',
                                                                                                                  'poller_hash': 'MkIwMThBcGVlcmludGZwb24gMjA3'},
                                                                                                          {       'in_nucast_pkts': '23585',
                                                                                                                  'in_octets': '13403927991',
                                                                                                                  'in_ucast_pkts': '59089078',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464121',
                                                                                                                  'out_octets': '193318002063',
                                                                                                                  'out_ucast_pkts': '150545062',
                                                                                                                  'poller_hash': 'MkJENjhEcGVlcmludGZwb24gMjQ5'},
                                                                                                          {       'in_nucast_pkts': '89',
                                                                                                                  'in_octets': '717724481',
                                                                                                                  'in_ucast_pkts': '4559522',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11700022',
                                                                                                                  'out_octets': '13642307463',
                                                                                                                  'out_ucast_pkts': '10040572',
                                                                                                                  'poller_hash': 'MzVFQTQ0cGVlcmludGZwb24gNDQ4'},
                                                                                                          {       'in_nucast_pkts': '16707',
                                                                                                                  'in_octets': '7060355395',
                                                                                                                  'in_ucast_pkts': '59539725',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464202',
                                                                                                                  'out_octets': '250330173336',
                                                                                                                  'out_ucast_pkts': '177163459',
                                                                                                                  'poller_hash': 'MzAwREYxcGVlcmludGZwb24gNDg4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzVFOTM4cGVlcmludGZwb24gNDg5'},
                                                                                                          {       'in_nucast_pkts': '955',
                                                                                                                  'in_octets': '81049090267702',
                                                                                                                  'in_ucast_pkts': '68720615352',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '150423145',
                                                                                                                  'out_octets': '34036842973018',
                                                                                                                  'out_ucast_pkts': '56666971536',
                                                                                                                  'poller_hash': 'NjZENTNwZWVyaW50ZnBvbiAyMjQ0MzI='},
                                                                                                          {       'in_nucast_pkts': '546936',
                                                                                                                  'in_octets': '152696317549',
                                                                                                                  'in_ucast_pkts': '610162882',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '48258062',
                                                                                                                  'out_octets': '3550280734473',
                                                                                                                  'out_ucast_pkts': '2617468525',
                                                                                                                  'poller_hash': 'MUQwNjlGcGVlcmludGZwb24gMjYyNjA3'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '8/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '80904736708',
                                                                                         'out_octets': '470950089205943',
                                                                                         'out_ucast_pkts': '308660324444',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/5',
                                                                                                                 'id': '8/5',
                                                                                                                 'links': [       {       'connected_time': '41697',
                                                                                                                                          'fiber_length': 1341,
                                                                                                                                          'poller_hash': 'MUQwNjlEcGVlcnBvbnBvbiAz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '5080999',
                                                                                                                                          'fiber_length': 1450,
                                                                                                                                          'poller_hash': 'Mjk2RkI4cGVlcnBvbnBvbiAxMTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688494',
                                                                                                                                          'fiber_length': 1341,
                                                                                                                                          'poller_hash': 'Mjk4MjQ4cGVlcnBvbnBvbiAxMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '71552369',
                                                                                                                                          'fiber_length': 1303,
                                                                                                                                          'poller_hash': 'MUQwNTBGcGVlcnBvbnBvbiAxMTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '21944183',
                                                                                                                                          'fiber_length': 1292,
                                                                                                                                          'poller_hash': 'Mjk2RkM5cGVlcnBvbnBvbiAxMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '31213287',
                                                                                                                                          'fiber_length': 1364,
                                                                                                                                          'poller_hash': 'Mjk4MjUzcGVlcnBvbnBvbiAxMTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '11235575',
                                                                                                                                          'fiber_length': 1284,
                                                                                                                                          'poller_hash': 'Mjk4MkI3cGVlcnBvbnBvbiAxMjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '13219376',
                                                                                                                                          'fiber_length': 1355,
                                                                                                                                          'poller_hash': 'MUQwNkQ3cGVlcnBvbnBvbiAxMjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '674508',
                                                                                                                                          'fiber_length': 1254,
                                                                                                                                          'poller_hash': 'Mjk4MjhEcGVlcnBvbnBvbiAxMjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688508',
                                                                                                                                          'fiber_length': 1307,
                                                                                                                                          'poller_hash': 'MUQwNkExcGVlcnBvbnBvbiAxMzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688497',
                                                                                                                                          'fiber_length': 1312,
                                                                                                                                          'poller_hash': 'MzAwRERGcGVlcnBvbnBvbiAxMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '21816931',
                                                                                                                                          'fiber_length': 1371,
                                                                                                                                          'poller_hash': 'MUQwNEYxcGVlcnBvbnBvbiAxMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688500',
                                                                                                                                          'fiber_length': 1326,
                                                                                                                                          'poller_hash': 'MUQwNkI5cGVlcnBvbnBvbiAxNDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '4304167',
                                                                                                                                          'fiber_length': 1319,
                                                                                                                                          'poller_hash': 'MUQwNzE0cGVlcnBvbnBvbiAxNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '21816379',
                                                                                                                                          'fiber_length': 868,
                                                                                                                                          'poller_hash': 'Mjk4MjFFcGVlcnBvbnBvbiAxNzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '40631',
                                                                                                                                          'fiber_length': 1490,
                                                                                                                                          'poller_hash': 'MTk5QTNBcGVlcnBvbnBvbiAxNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '2846741',
                                                                                                                                          'fiber_length': 852,
                                                                                                                                          'poller_hash': 'MTc2MjRFcGVlcnBvbnBvbiAxNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688505',
                                                                                                                                          'fiber_length': 921,
                                                                                                                                          'poller_hash': 'Mjk4MjAwcGVlcnBvbnBvbiAxODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '17629509',
                                                                                                                                          'fiber_length': 871,
                                                                                                                                          'poller_hash': 'MUQwNjg4cGVlcnBvbnBvbiAxOTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '21763000',
                                                                                                                                          'fiber_length': 903,
                                                                                                                                          'poller_hash': 'MjA1REQxcGVlcnBvbnBvbiAyMDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688498',
                                                                                                                                          'fiber_length': 830,
                                                                                                                                          'poller_hash': 'MkIwMThBcGVlcnBvbnBvbiAyMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688500',
                                                                                                                                          'fiber_length': 887,
                                                                                                                                          'poller_hash': 'MkJENjhEcGVlcnBvbnBvbiAyNDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '5144138',
                                                                                                                                          'fiber_length': 1330,
                                                                                                                                          'poller_hash': 'MzVFQTQ0cGVlcnBvbnBvbiA0NDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '688506',
                                                                                                                                          'fiber_length': 1365,
                                                                                                                                          'poller_hash': 'MzAwREYxcGVlcnBvbnBvbiA0ODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '67790431',
                                                                                                                                          'fiber_length': 1377,
                                                                                                                                          'poller_hash': 'MzVFOTM4cGVlcnBvbnBvbiA0ODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '15493579',
                                                                                                                                          'fiber_length': 1688,
                                                                                                                                          'poller_hash': 'NjZENTNwZWVycG9ucG9uIDIyNDQzMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664},
                                                                                                                                  {       'connected_time': '20904113',
                                                                                                                                          'fiber_length': 1168,
                                                                                                                                          'poller_hash': 'MUQwNjlGcGVlcnBvbnBvbiAyNjI2MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.664}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/5',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '8/6',
                                                                                         'in_discards': '7381',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5014003',
                                                                                         'in_octets': '15657471012908',
                                                                                         'in_ucast_pkts': '41858904855',
                                                                                         'links': [       {       'in_nucast_pkts': '346586',
                                                                                                                  'in_octets': '153043905984',
                                                                                                                  'in_ucast_pkts': '1269657877',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43140836',
                                                                                                                  'out_octets': '5886684414645',
                                                                                                                  'out_ucast_pkts': '4081166040',
                                                                                                                  'poller_hash': 'MUQwNEZBcGVlcmludGZwb24gMTE1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '144761129526',
                                                                                                                  'in_ucast_pkts': '1157910890',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '113',
                                                                                                                  'out_octets': '6916726603455',
                                                                                                                  'out_ucast_pkts': '8830996015',
                                                                                                                  'poller_hash': 'Mjk2RkMxcGVlcmludGZwb24gMTE5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '274777998073',
                                                                                                                  'in_ucast_pkts': '10622603853',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '224',
                                                                                                                  'out_octets': '6381848279898',
                                                                                                                  'out_ucast_pkts': '4493281883',
                                                                                                                  'poller_hash': 'MUQwNTA4cGVlcmludGZwb24gMTIw'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '2461387758',
                                                                                                                  'in_ucast_pkts': '15551443',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463961',
                                                                                                                  'out_octets': '114353309032',
                                                                                                                  'out_ucast_pkts': '79949739',
                                                                                                                  'poller_hash': 'MUQwNTBDcGVlcmludGZwb24gMTIx'},
                                                                                                          {       'in_nucast_pkts': '150626',
                                                                                                                  'in_octets': '31782343245',
                                                                                                                  'in_ucast_pkts': '154490310',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17819593',
                                                                                                                  'out_octets': '670262893332',
                                                                                                                  'out_ucast_pkts': '494735476',
                                                                                                                  'poller_hash': 'MUQwNTEycGVlcmludGZwb24gMTIy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '371988178552',
                                                                                                                  'in_ucast_pkts': '1331301013',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '165',
                                                                                                                  'out_octets': '4561128588971',
                                                                                                                  'out_ucast_pkts': '3599274497',
                                                                                                                  'poller_hash': 'Mjk2RkM1cGVlcmludGZwb24gMTIz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '87600988904',
                                                                                                                  'in_ucast_pkts': '1050104658',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '117',
                                                                                                                  'out_octets': '6764244310376',
                                                                                                                  'out_ucast_pkts': '4538007308',
                                                                                                                  'poller_hash': 'MUQwNjlBcGVlcmludGZwb24gMTMz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '5568409984342',
                                                                                                                  'in_ucast_pkts': '369751969577',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '158',
                                                                                                                  'out_octets': '8468015401679',
                                                                                                                  'out_ucast_pkts': '7690817994',
                                                                                                                  'poller_hash': 'MUQwNEZCcGVlcmludGZwb24gMTM2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '783244328749',
                                                                                                                  'in_ucast_pkts': '80199830396',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '167',
                                                                                                                  'out_octets': '13104451312196',
                                                                                                                  'out_ucast_pkts': '9524772240',
                                                                                                                  'poller_hash': 'Mjk4MjI1cGVlcmludGZwb24gMTYx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTc1RUJBcGVlcmludGZwb24gMTc3'},
                                                                                                          {       'in_nucast_pkts': '133466',
                                                                                                                  'in_octets': '43220032043',
                                                                                                                  'in_ucast_pkts': '384479772',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14849776',
                                                                                                                  'out_octets': '2486109633593',
                                                                                                                  'out_ucast_pkts': '1671154750',
                                                                                                                  'poller_hash': 'Mjk4MjEycGVlcmludGZwb24gMTk2'},
                                                                                                          {       'in_nucast_pkts': '11892',
                                                                                                                  'in_octets': '4050873660',
                                                                                                                  'in_ucast_pkts': '20561219',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119988',
                                                                                                                  'out_octets': '99612986705',
                                                                                                                  'out_ucast_pkts': '72137236',
                                                                                                                  'poller_hash': 'NDNCMTU0cGVlcmludGZwb24gNjk3'},
                                                                                                          {       'in_nucast_pkts': '11867',
                                                                                                                  'in_octets': '1858120830',
                                                                                                                  'in_ucast_pkts': '9404594',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1456868',
                                                                                                                  'out_octets': '45162445772',
                                                                                                                  'out_ucast_pkts': '35565799',
                                                                                                                  'poller_hash': 'MzIyODRGcGVlcmludGZwb24gNzEy'},
                                                                                                          {       'in_nucast_pkts': '13034',
                                                                                                                  'in_octets': '7583101259',
                                                                                                                  'in_ucast_pkts': '14512786',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1463849',
                                                                                                                  'out_octets': '80565339589',
                                                                                                                  'out_ucast_pkts': '59948198',
                                                                                                                  'poller_hash': 'MzIxNzFCcGVlcmludGZwb24gNzIw'},
                                                                                                          {       'in_nucast_pkts': '15',
                                                                                                                  'in_octets': '819318223',
                                                                                                                  'in_ucast_pkts': '5803595',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464212',
                                                                                                                  'out_octets': '21929295035',
                                                                                                                  'out_ucast_pkts': '15994197',
                                                                                                                  'poller_hash': 'MzVFQUE1cGVlcmludGZwb24gNzQ0'},
                                                                                                          {       'in_nucast_pkts': '647650',
                                                                                                                  'in_octets': '139201975509',
                                                                                                                  'in_ucast_pkts': '730472217',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '92687941',
                                                                                                                  'out_octets': '3453325304789',
                                                                                                                  'out_ucast_pkts': '2598461225',
                                                                                                                  'poller_hash': 'ODcyNTAzcGVlcmludGZwb24gMTQyODMz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '213459506245',
                                                                                                                  'in_ucast_pkts': '21227061204',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '107',
                                                                                                                  'out_octets': '4383311779173',
                                                                                                                  'out_ucast_pkts': '8160923168',
                                                                                                                  'poller_hash': 'QzgzQjczcGVlcmludGZwb24gMTkzNTcy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '63004319958',
                                                                                                                  'in_ucast_pkts': '433948321',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33',
                                                                                                                  'out_octets': '1674836884611',
                                                                                                                  'out_ucast_pkts': '1216691207',
                                                                                                                  'poller_hash': 'MTFCRTNBMnBlZXJpbnRmcG9uIDI1Njc3Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '8/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '68227302372',
                                                                                         'out_octets': '247127927958243',
                                                                                         'out_ucast_pkts': '114384186321',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/6',
                                                                                                                 'id': '8/6',
                                                                                                                 'links': [       {       'connected_time': '18857359',
                                                                                                                                          'fiber_length': 1246,
                                                                                                                                          'poller_hash': 'MUQwNEZBcGVlcnBvbnBvbiAxMTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '30050458',
                                                                                                                                          'fiber_length': 1194,
                                                                                                                                          'poller_hash': 'Mjk2RkMxcGVlcnBvbnBvbiAxMTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '7581551',
                                                                                                                                          'fiber_length': 1194,
                                                                                                                                          'poller_hash': 'MUQwNTA4cGVlcnBvbnBvbiAxMjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688515',
                                                                                                                                          'fiber_length': 1195,
                                                                                                                                          'poller_hash': 'MUQwNTBDcGVlcnBvbnBvbiAxMjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '7695099',
                                                                                                                                          'fiber_length': 1251,
                                                                                                                                          'poller_hash': 'MUQwNTEycGVlcnBvbnBvbiAxMjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688502',
                                                                                                                                          'fiber_length': 1278,
                                                                                                                                          'poller_hash': 'Mjk2RkM1cGVlcnBvbnBvbiAxMjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '29450620',
                                                                                                                                          'fiber_length': 1234,
                                                                                                                                          'poller_hash': 'MUQwNjlBcGVlcnBvbnBvbiAxMzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688510',
                                                                                                                                          'fiber_length': 1128,
                                                                                                                                          'poller_hash': 'MUQwNEZCcGVlcnBvbnBvbiAxMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '16862227',
                                                                                                                                          'fiber_length': 1235,
                                                                                                                                          'poller_hash': 'Mjk4MjI1cGVlcnBvbnBvbiAxNjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688490',
                                                                                                                                          'fiber_length': 1006,
                                                                                                                                          'poller_hash': 'MTc1RUJBcGVlcnBvbnBvbiAxNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '6403842',
                                                                                                                                          'fiber_length': 1158,
                                                                                                                                          'poller_hash': 'Mjk4MjEycGVlcnBvbnBvbiAxOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688491',
                                                                                                                                          'fiber_length': 1112,
                                                                                                                                          'poller_hash': 'NDNCMTU0cGVlcnBvbnBvbiA2OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688500',
                                                                                                                                          'fiber_length': 1213,
                                                                                                                                          'poller_hash': 'MzIyODRGcGVlcnBvbnBvbiA3MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688505',
                                                                                                                                          'fiber_length': 1155,
                                                                                                                                          'poller_hash': 'MzIxNzFCcGVlcnBvbnBvbiA3MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '688524',
                                                                                                                                          'fiber_length': 1168,
                                                                                                                                          'poller_hash': 'MzVFQUE1cGVlcnBvbnBvbiA3NDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '21816400',
                                                                                                                                          'fiber_length': 1233,
                                                                                                                                          'poller_hash': 'ODcyNTAzcGVlcnBvbnBvbiAxNDI4MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '6413854',
                                                                                                                                          'fiber_length': 1184,
                                                                                                                                          'poller_hash': 'QzgzQjczcGVlcnBvbnBvbiAxOTM1NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405},
                                                                                                                                  {       'connected_time': '24546110',
                                                                                                                                          'fiber_length': 1190,
                                                                                                                                          'poller_hash': 'MTFCRTNBMnBlZXJwb25wb24gMjU2Nzc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.405}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/6',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '8/7',
                                                                                         'in_discards': '6805',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '909997',
                                                                                         'in_octets': '6773131220907',
                                                                                         'in_ucast_pkts': '31213040655',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '66992988241',
                                                                                                                  'in_ucast_pkts': '203905777',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '614359344547',
                                                                                                                  'out_ucast_pkts': '482492761',
                                                                                                                  'poller_hash': 'OUMzMDIycGVlcmludGZwb24gMTQ2MDgy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1114445892395',
                                                                                                                  'in_ucast_pkts': '2219974894',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '99',
                                                                                                                  'out_octets': '9734174418356',
                                                                                                                  'out_ucast_pkts': '6724146514',
                                                                                                                  'poller_hash': 'MjA1REYxcGVlcmludGZwb24gMTQ5MTQ3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '92132228907',
                                                                                                                  'in_ucast_pkts': '479617271',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1606686438815',
                                                                                                                  'out_ucast_pkts': '1172788603',
                                                                                                                  'poller_hash': 'MUQwNEY1cGVlcmludGZwb24gMTUwMTA3'},
                                                                                                          {       'in_nucast_pkts': '5282',
                                                                                                                  'in_octets': '967666172',
                                                                                                                  'in_ucast_pkts': '1320759',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464249',
                                                                                                                  'out_octets': '2261970364',
                                                                                                                  'out_ucast_pkts': '2164385',
                                                                                                                  'poller_hash': 'QjUxMkRDcGVlcmludGZwb24gMTU1MDI3'},
                                                                                                          {       'in_nucast_pkts': '19',
                                                                                                                  'in_octets': '12402796948',
                                                                                                                  'in_ucast_pkts': '17511027',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464356',
                                                                                                                  'out_octets': '61820249439',
                                                                                                                  'out_ucast_pkts': '50530195',
                                                                                                                  'poller_hash': 'QzgzODIxcGVlcmludGZwb24gMTc0NDg0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '130762744312',
                                                                                                                  'in_ucast_pkts': '237781294',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '312058573535',
                                                                                                                  'out_ucast_pkts': '330476155',
                                                                                                                  'poller_hash': 'QzgxRDcycGVlcmludGZwb24gMjA1MDkw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '133525152115',
                                                                                                                  'in_ucast_pkts': '1427259181',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12110898802',
                                                                                                                  'out_octets': '22867941440050',
                                                                                                                  'out_ucast_pkts': '4568828387',
                                                                                                                  'poller_hash': 'MTA5NDhEMXBlZXJpbnRmcG9uIDIxMzkxMw=='},
                                                                                                          {       'in_nucast_pkts': '12763',
                                                                                                                  'in_octets': '14392895080',
                                                                                                                  'in_ucast_pkts': '98152911',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464116',
                                                                                                                  'out_octets': '422381155003',
                                                                                                                  'out_ucast_pkts': '313396911',
                                                                                                                  'poller_hash': 'MTE5NTFCMHBlZXJpbnRmcG9uIDIxNjA4Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '95970349580',
                                                                                                                  'in_ucast_pkts': '131709041',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '91',
                                                                                                                  'out_octets': '128272121538',
                                                                                                                  'out_ucast_pkts': '152988469',
                                                                                                                  'poller_hash': 'MTI3NTExNHBlZXJpbnRmcG9uIDIyMTEyMA=='},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '4724637248',
                                                                                                                  'in_ucast_pkts': '34746781',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464577',
                                                                                                                  'out_octets': '201145667058',
                                                                                                                  'out_ucast_pkts': '143425766',
                                                                                                                  'poller_hash': 'REZGRTQ0cGVlcmludGZwb24gMjIxNDY1'},
                                                                                                          {       'in_nucast_pkts': '2374072',
                                                                                                                  'in_octets': '85366564152',
                                                                                                                  'in_ucast_pkts': '1128275564',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '159817338',
                                                                                                                  'out_octets': '6821200797821',
                                                                                                                  'out_ucast_pkts': '4610641053',
                                                                                                                  'poller_hash': 'MTI3NEFBN3BlZXJpbnRmcG9uIDIzMzczNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '155703074788',
                                                                                                                  'in_ucast_pkts': '1095009707',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '4236094221439',
                                                                                                                  'out_ucast_pkts': '2963544372',
                                                                                                                  'poller_hash': 'MTI3NDlBOHBlZXJpbnRmcG9uIDIzNDg1NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '592707568401',
                                                                                                                  'in_ucast_pkts': '2553208413',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33',
                                                                                                                  'out_octets': '6161468167992',
                                                                                                                  'out_ucast_pkts': '5201328353',
                                                                                                                  'poller_hash': 'MTI3NDE5MHBlZXJpbnRmcG9uIDI0MzA5Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '7648479679',
                                                                                                                  'in_ucast_pkts': '52996418',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '18355699777',
                                                                                                                  'out_ucast_pkts': '42298408',
                                                                                                                  'poller_hash': 'MTI3NDBCOHBlZXJpbnRmcG9uIDI0NjkxMA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '8/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '22362145924',
                                                                                         'out_octets': '149359235685662',
                                                                                         'out_ucast_pkts': '86910892590',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/7',
                                                                                                                 'id': '8/7',
                                                                                                                 'links': [       {       'connected_time': '688510',
                                                                                                                                          'fiber_length': 624,
                                                                                                                                          'poller_hash': 'OUMzMDIycGVlcnBvbnBvbiAxNDYwODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688502',
                                                                                                                                          'fiber_length': 1700,
                                                                                                                                          'poller_hash': 'MjA1REYxcGVlcnBvbnBvbiAxNDkxNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688512',
                                                                                                                                          'fiber_length': 615,
                                                                                                                                          'poller_hash': 'MUQwNEY1cGVlcnBvbnBvbiAxNTAxMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688527',
                                                                                                                                          'fiber_length': 599,
                                                                                                                                          'poller_hash': 'QjUxMkRDcGVlcnBvbnBvbiAxNTUwMjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688499',
                                                                                                                                          'fiber_length': 633,
                                                                                                                                          'poller_hash': 'QzgzODIxcGVlcnBvbnBvbiAxNzQ0ODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688502',
                                                                                                                                          'fiber_length': 1451,
                                                                                                                                          'poller_hash': 'QzgxRDcycGVlcnBvbnBvbiAyMDUwOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688504',
                                                                                                                                          'fiber_length': 607,
                                                                                                                                          'poller_hash': 'MTA5NDhEMXBlZXJwb25wb24gMjEzOTEz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688503',
                                                                                                                                          'fiber_length': 622,
                                                                                                                                          'poller_hash': 'MTE5NTFCMHBlZXJwb25wb24gMjE2MDgz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688503',
                                                                                                                                          'fiber_length': 1181,
                                                                                                                                          'poller_hash': 'MTI3NTExNHBlZXJwb25wb24gMjIxMTIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688506',
                                                                                                                                          'fiber_length': 621,
                                                                                                                                          'poller_hash': 'REZGRTQ0cGVlcnBvbnBvbiAyMjE0NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '64803881',
                                                                                                                                          'fiber_length': 1135,
                                                                                                                                          'poller_hash': 'MTI3NEFBN3BlZXJwb25wb24gMjMzNzM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688498',
                                                                                                                                          'fiber_length': 1364,
                                                                                                                                          'poller_hash': 'MTI3NDlBOHBlZXJwb25wb24gMjM0ODU0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '13208956',
                                                                                                                                          'fiber_length': 1552,
                                                                                                                                          'poller_hash': 'MTI3NDE5MHBlZXJwb25wb24gMjQzMDk3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514},
                                                                                                                                  {       'connected_time': '688505',
                                                                                                                                          'fiber_length': 1365,
                                                                                                                                          'poller_hash': 'MTI3NDBCOHBlZXJwb25wb24gMjQ2OTEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -9.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.514}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/7',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '8/8',
                                                                                         'in_discards': '6316',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '26675226',
                                                                                         'in_octets': '169679855366011',
                                                                                         'in_ucast_pkts': '171858318038',
                                                                                         'links': [       {       'in_nucast_pkts': '16192651',
                                                                                                                  'in_octets': '1556591803024',
                                                                                                                  'in_ucast_pkts': '2771493963',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '79374701',
                                                                                                                  'out_octets': '6582580067884',
                                                                                                                  'out_ucast_pkts': '6141309966',
                                                                                                                  'poller_hash': 'QkREOUZEcGVlcmludGZwb24gMTU2Nzc2'},
                                                                                                          {       'in_nucast_pkts': '2166',
                                                                                                                  'in_octets': '1181104688380',
                                                                                                                  'in_ucast_pkts': '5431688214',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '198796755',
                                                                                                                  'out_octets': '8512727407073',
                                                                                                                  'out_ucast_pkts': '7573079082',
                                                                                                                  'poller_hash': 'QzgxQkI1cGVlcmludGZwb24gMTU2Nzg4'},
                                                                                                          {       'in_nucast_pkts': '49642',
                                                                                                                  'in_octets': '194733125208',
                                                                                                                  'in_ucast_pkts': '627412543',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '79318979',
                                                                                                                  'out_octets': '2028732002600',
                                                                                                                  'out_ucast_pkts': '1679300819',
                                                                                                                  'poller_hash': 'QkREOUY0cGVlcmludGZwb24gMTU2ODA4'},
                                                                                                          {       'in_nucast_pkts': '3596754',
                                                                                                                  'in_octets': '1387632639885',
                                                                                                                  'in_ucast_pkts': '4663163956',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66865933',
                                                                                                                  'out_octets': '8715400518440',
                                                                                                                  'out_ucast_pkts': '7800065249',
                                                                                                                  'poller_hash': 'QzgxMUUycGVlcmludGZwb24gMTU2OTMx'},
                                                                                                          {       'in_nucast_pkts': '46069285',
                                                                                                                  'in_octets': '158305074114725',
                                                                                                                  'in_ucast_pkts': '135456167952',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '512920193',
                                                                                                                  'out_octets': '64512046185687',
                                                                                                                  'out_ucast_pkts': '119026071472',
                                                                                                                  'poller_hash': 'MkYzQjE3cGVlcmludGZwb24gMTU2OTgy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '280508898579',
                                                                                                                  'in_ucast_pkts': '1059681632',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '182',
                                                                                                                  'out_octets': '3045951279029',
                                                                                                                  'out_ucast_pkts': '2595037547',
                                                                                                                  'poller_hash': 'Q0JEMTg5cGVlcmludGZwb24gMTY3MjU3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '641341360086',
                                                                                                                  'in_ucast_pkts': '353428695334',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '3938940775490',
                                                                                                                  'out_ucast_pkts': '3018963808',
                                                                                                                  'poller_hash': 'REZGREIycGVlcmludGZwb24gMTcxOTY3'},
                                                                                                          {       'in_nucast_pkts': '725832',
                                                                                                                  'in_octets': '233587877947',
                                                                                                                  'in_ucast_pkts': '2366506149',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '99038472',
                                                                                                                  'out_octets': '14227848819899',
                                                                                                                  'out_ucast_pkts': '9535601907',
                                                                                                                  'poller_hash': 'Q0JEMUYzcGVlcmludGZwb24gMTc1MjI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '90763488689',
                                                                                                                  'in_ucast_pkts': '854186055',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '141',
                                                                                                                  'out_octets': '3482944353773',
                                                                                                                  'out_ucast_pkts': '2404240504',
                                                                                                                  'poller_hash': 'Q0JEMTk0cGVlcmludGZwb24gMTc1MjM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '71725232801',
                                                                                                                  'in_ucast_pkts': '187895963',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '207',
                                                                                                                  'out_octets': '567528691540',
                                                                                                                  'out_ucast_pkts': '511807214',
                                                                                                                  'poller_hash': 'Q0JEMTlDcGVlcmludGZwb24gMTc1MjQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '8241817367',
                                                                                                                  'in_ucast_pkts': '33839004',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '192',
                                                                                                                  'out_octets': '144424049476',
                                                                                                                  'out_ucast_pkts': '165000210',
                                                                                                                  'poller_hash': 'QzgzQjFFcGVlcmludGZwb24gMTc1Mzgx'},
                                                                                                          {       'in_nucast_pkts': '335319',
                                                                                                                  'in_octets': '366072960888',
                                                                                                                  'in_ucast_pkts': '3165777892',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '51229730',
                                                                                                                  'out_octets': '15520480445009',
                                                                                                                  'out_ucast_pkts': '11013737920',
                                                                                                                  'poller_hash': 'QzgzMUYzcGVlcmludGZwb24gMTc3MDA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '150973663423',
                                                                                                                  'in_ucast_pkts': '598898545',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75',
                                                                                                                  'out_octets': '4513205606103',
                                                                                                                  'out_ucast_pkts': '3404869554',
                                                                                                                  'poller_hash': 'QzgzQzE2cGVlcmludGZwb24gMTgwMTc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '529539636892',
                                                                                                                  'in_ucast_pkts': '74792150338',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '200',
                                                                                                                  'out_octets': '6956372559273',
                                                                                                                  'out_ucast_pkts': '5321474297',
                                                                                                                  'poller_hash': 'QzgzQkZBcGVlcmludGZwb24gMTgwNTMw'},
                                                                                                          {       'in_nucast_pkts': '613401',
                                                                                                                  'in_octets': '68467731802',
                                                                                                                  'in_ucast_pkts': '172767185',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '74641511',
                                                                                                                  'out_octets': '1072255058835',
                                                                                                                  'out_ucast_pkts': '827923932',
                                                                                                                  'poller_hash': 'QzgzQkQ1cGVlcmludGZwb24gMTgyMjQx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgzQkZGcGVlcmludGZwb24gMTgyMjYx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'REZGREI5cGVlcmludGZwb24gMTg2NjU3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '167868527313',
                                                                                                                  'in_ucast_pkts': '505218631',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '207',
                                                                                                                  'out_octets': '2747955116387',
                                                                                                                  'out_ucast_pkts': '2127899254',
                                                                                                                  'poller_hash': 'QzgzMTNGcGVlcmludGZwb24gMTg2NjY3'},
                                                                                                          {       'in_nucast_pkts': '238398',
                                                                                                                  'in_octets': '3424756202',
                                                                                                                  'in_ucast_pkts': '10235369',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '287614777',
                                                                                                                  'out_octets': '427601681766',
                                                                                                                  'out_ucast_pkts': '34395194',
                                                                                                                  'poller_hash': 'QzgzMTkzcGVlcmludGZwb24gMTg3NDY5'},
                                                                                                          {       'in_nucast_pkts': '14806',
                                                                                                                  'in_octets': '4082537689',
                                                                                                                  'in_ucast_pkts': '6992056',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464888',
                                                                                                                  'out_octets': '27032651252',
                                                                                                                  'out_ucast_pkts': '22884362',
                                                                                                                  'poller_hash': 'MTI3NDg0RXBlZXJpbnRmcG9uIDI2Mjc2NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3MEQ3MHBlZXJpbnRmcG9uIDI3NzM3Nw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '8/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '137119974849',
                                                                                         'out_octets': '368424228698178',
                                                                                         'out_ucast_pkts': '212106051292',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '8/8',
                                                                                                                 'id': '8/8',
                                                                                                                 'links': [       {       'connected_time': '73807588',
                                                                                                                                          'fiber_length': 11627,
                                                                                                                                          'poller_hash': 'QkREOUZEcGVlcnBvbnBvbiAxNTY3NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '73382281',
                                                                                                                                          'fiber_length': 11602,
                                                                                                                                          'poller_hash': 'QzgxQkI1cGVlcnBvbnBvbiAxNTY3ODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '73807576',
                                                                                                                                          'fiber_length': 11743,
                                                                                                                                          'poller_hash': 'QkREOUY0cGVlcnBvbnBvbiAxNTY4MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '2073770',
                                                                                                                                          'fiber_length': 11471,
                                                                                                                                          'poller_hash': 'QzgxMUUycGVlcnBvbnBvbiAxNTY5MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '73807968',
                                                                                                                                          'fiber_length': 11576,
                                                                                                                                          'poller_hash': 'MkYzQjE3cGVlcnBvbnBvbiAxNTY5ODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '10799647',
                                                                                                                                          'fiber_length': 11877,
                                                                                                                                          'poller_hash': 'Q0JEMTg5cGVlcnBvbnBvbiAxNjcyNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '688510',
                                                                                                                                          'fiber_length': 11687,
                                                                                                                                          'poller_hash': 'REZGREIycGVlcnBvbnBvbiAxNzE5Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '73290287',
                                                                                                                                          'fiber_length': 11721,
                                                                                                                                          'poller_hash': 'Q0JEMUYzcGVlcnBvbnBvbiAxNzUyMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '523248',
                                                                                                                                          'fiber_length': 11751,
                                                                                                                                          'poller_hash': 'Q0JEMTk0cGVlcnBvbnBvbiAxNzUyMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '21815841',
                                                                                                                                          'fiber_length': 11689,
                                                                                                                                          'poller_hash': 'Q0JEMTlDcGVlcnBvbnBvbiAxNzUyNDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '960570',
                                                                                                                                          'fiber_length': 11862,
                                                                                                                                          'poller_hash': 'QzgzQjFFcGVlcnBvbnBvbiAxNzUzODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '2078104',
                                                                                                                                          'fiber_length': 11482,
                                                                                                                                          'poller_hash': 'QzgzMUYzcGVlcnBvbnBvbiAxNzcwMDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '688516',
                                                                                                                                          'fiber_length': 11782,
                                                                                                                                          'poller_hash': 'QzgzQzE2cGVlcnBvbnBvbiAxODAxNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '21816278',
                                                                                                                                          'fiber_length': 11730,
                                                                                                                                          'poller_hash': 'QzgzQkZBcGVlcnBvbnBvbiAxODA1MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '21816763',
                                                                                                                                          'fiber_length': 11467,
                                                                                                                                          'poller_hash': 'QzgzQkQ1cGVlcnBvbnBvbiAxODIyNDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -29.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '21816000',
                                                                                                                                          'fiber_length': 11727,
                                                                                                                                          'poller_hash': 'QzgzQkZGcGVlcnBvbnBvbiAxODIyNjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '21816255',
                                                                                                                                          'fiber_length': 11837,
                                                                                                                                          'poller_hash': 'REZGREI5cGVlcnBvbnBvbiAxODY2NTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '1912560',
                                                                                                                                          'fiber_length': 11503,
                                                                                                                                          'poller_hash': 'QzgzMTNGcGVlcnBvbnBvbiAxODY2Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '523244',
                                                                                                                                          'fiber_length': 11584,
                                                                                                                                          'poller_hash': 'QzgzMTkzcGVlcnBvbnBvbiAxODc0Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '688514',
                                                                                                                                          'fiber_length': 11878,
                                                                                                                                          'poller_hash': 'MTI3NDg0RXBlZXJwb25wb24gMjYyNzY1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739},
                                                                                                                                  {       'connected_time': '9684744',
                                                                                                                                          'fiber_length': 11888,
                                                                                                                                          'poller_hash': 'MTI3MEQ3MHBlZXJwb25wb24gMjc3Mzc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.739}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '8/8',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNTA1NjAwMzA3',
                                                           'serial_number': '211505600307',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42990562'},
                                                   {       'id': '11',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '11/1',
                                                                                         'in_discards': '81775',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '51522558',
                                                                                         'in_octets': '24616470501506',
                                                                                         'in_ucast_pkts': '93156176491',
                                                                                         'links': [       {       'in_nucast_pkts': '23995752',
                                                                                                                  'in_octets': '370504917335',
                                                                                                                  'in_ucast_pkts': '1449703792',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '593628098',
                                                                                                                  'out_octets': '5258029738327',
                                                                                                                  'out_ucast_pkts': '3959080047',
                                                                                                                  'poller_hash': 'NDRBQUE4cGVlcmludGZwb24gNDg='},
                                                                                                          {       'in_nucast_pkts': '2746474',
                                                                                                                  'in_octets': '63104788927',
                                                                                                                  'in_ucast_pkts': '727686874',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7541501198',
                                                                                                                  'out_octets': '14189473223502',
                                                                                                                  'out_ucast_pkts': '2782649283',
                                                                                                                  'poller_hash': 'ODJBNkNCcGVlcmludGZwb24gMTA4'},
                                                                                                          {       'in_nucast_pkts': '2028268',
                                                                                                                  'in_octets': '2291105008541',
                                                                                                                  'in_ucast_pkts': '11532543456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '270929581',
                                                                                                                  'out_octets': '44524813207570',
                                                                                                                  'out_ucast_pkts': '31929987943',
                                                                                                                  'poller_hash': 'MjZDQzU0cGVlcmludGZwb24gMTUy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '475658995317',
                                                                                                                  'in_ucast_pkts': '2447596523',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '138',
                                                                                                                  'out_octets': '8400505288513',
                                                                                                                  'out_ucast_pkts': '6018250786',
                                                                                                                  'poller_hash': 'Mjk4NDQycGVlcmludGZwb24gMTkx'},
                                                                                                          {       'in_nucast_pkts': '10715',
                                                                                                                  'in_octets': '288649774850',
                                                                                                                  'in_ucast_pkts': '3035541448',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145122444',
                                                                                                                  'out_octets': '16871802604962',
                                                                                                                  'out_ucast_pkts': '12032903056',
                                                                                                                  'poller_hash': 'MUQwNjdGcGVlcmludGZwb24gMTkz'},
                                                                                                          {       'in_nucast_pkts': '1751703',
                                                                                                                  'in_octets': '1806937815605',
                                                                                                                  'in_ucast_pkts': '7731210128',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144919536',
                                                                                                                  'out_octets': '27427778164623',
                                                                                                                  'out_ucast_pkts': '20711249833',
                                                                                                                  'poller_hash': 'MUQwNkUxcGVlcmludGZwb24gMjA1'},
                                                                                                          {       'in_nucast_pkts': '20',
                                                                                                                  'in_octets': '44775377773',
                                                                                                                  'in_ucast_pkts': '73413123',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720619',
                                                                                                                  'out_octets': '458020761646',
                                                                                                                  'out_ucast_pkts': '339595150',
                                                                                                                  'poller_hash': 'MkJENDc4cGVlcmludGZwb24gMjQy'},
                                                                                                          {       'in_nucast_pkts': '2813',
                                                                                                                  'in_octets': '24554806497',
                                                                                                                  'in_ucast_pkts': '53623951',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720648',
                                                                                                                  'out_octets': '257053119877',
                                                                                                                  'out_ucast_pkts': '203804075',
                                                                                                                  'poller_hash': 'MkJENDc0cGVlcmludGZwb24gMjQz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '327674423559',
                                                                                                                  'in_ucast_pkts': '4503509965',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '10809212813676',
                                                                                                                  'out_ucast_pkts': '9961732038',
                                                                                                                  'poller_hash': 'N0UwRURwZWVyaW50ZnBvbiAyNTE='},
                                                                                                          {       'in_nucast_pkts': '188964',
                                                                                                                  'in_octets': '32542492272',
                                                                                                                  'in_ucast_pkts': '300208231',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '21678694',
                                                                                                                  'out_octets': '1711922642715',
                                                                                                                  'out_ucast_pkts': '1167534969',
                                                                                                                  'poller_hash': 'MUQwNjgzcGVlcmludGZwb24gMjU2'},
                                                                                                          {       'in_nucast_pkts': '2143773',
                                                                                                                  'in_octets': '1562238598180',
                                                                                                                  'in_ucast_pkts': '8015926679',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2018930714',
                                                                                                                  'out_octets': '32567408925774',
                                                                                                                  'out_ucast_pkts': '22378380304',
                                                                                                                  'poller_hash': 'ODY5M0EwcGVlcmludGZwb24gMjU3'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '62203302387',
                                                                                                                  'in_ucast_pkts': '226271381',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720682',
                                                                                                                  'out_octets': '629140298154',
                                                                                                                  'out_ucast_pkts': '494239529',
                                                                                                                  'poller_hash': 'MkIwMjY2cGVlcmludGZwb24gMjY5'},
                                                                                                          {       'in_nucast_pkts': '67099',
                                                                                                                  'in_octets': '20930811288',
                                                                                                                  'in_ucast_pkts': '54993181',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720550',
                                                                                                                  'out_octets': '144912625381',
                                                                                                                  'out_ucast_pkts': '118026055',
                                                                                                                  'poller_hash': 'MzAwNjRDcGVlcmludGZwb24gMjcy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzVGODRDcGVlcmludGZwb24gMjg4'},
                                                                                                          {       'in_nucast_pkts': '9516',
                                                                                                                  'in_octets': '52287579825',
                                                                                                                  'in_ucast_pkts': '237198510',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '13415176',
                                                                                                                  'out_octets': '1266897634338',
                                                                                                                  'out_ucast_pkts': '1001414512',
                                                                                                                  'poller_hash': 'MzAwN0FFcGVlcmludGZwb24gMjkw'},
                                                                                                          {       'in_nucast_pkts': '3451',
                                                                                                                  'in_octets': '591115546937',
                                                                                                                  'in_ucast_pkts': '2380118253',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33792523',
                                                                                                                  'out_octets': '11761725508562',
                                                                                                                  'out_ucast_pkts': '8202499538',
                                                                                                                  'poller_hash': 'MzVGOTUzcGVlcmludGZwb24gMzEz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2293733100764',
                                                                                                                  'in_ucast_pkts': '7045777794',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11273286154577',
                                                                                                                  'out_ucast_pkts': '13982576845',
                                                                                                                  'poller_hash': 'MzhGMjg1cGVlcmludGZwb24gMzIw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1243390822295',
                                                                                                                  'in_ucast_pkts': '153431345162',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82',
                                                                                                                  'out_octets': '9486556958500',
                                                                                                                  'out_ucast_pkts': '7359473781',
                                                                                                                  'poller_hash': 'MkIwMTUwcGVlcmludGZwb24gMzQz'},
                                                                                                          {       'in_nucast_pkts': '3848',
                                                                                                                  'in_octets': '367821721349',
                                                                                                                  'in_ucast_pkts': '2860495659',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '72837063',
                                                                                                                  'out_octets': '12483074364983',
                                                                                                                  'out_ucast_pkts': '8641015203',
                                                                                                                  'poller_hash': 'MkFEMEJDcGVlcmludGZwb24gNDkz'},
                                                                                                          {       'in_nucast_pkts': '2937',
                                                                                                                  'in_octets': '88957209788',
                                                                                                                  'in_ucast_pkts': '530711453',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83496319',
                                                                                                                  'out_octets': '4779695010978',
                                                                                                                  'out_ucast_pkts': '3226699286',
                                                                                                                  'poller_hash': 'MzVFOUNEcGVlcmludGZwb24gNTA0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '243082015734',
                                                                                                                  'in_ucast_pkts': '2071281673',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '112',
                                                                                                                  'out_octets': '11216537451373',
                                                                                                                  'out_ucast_pkts': '7648713625',
                                                                                                                  'poller_hash': 'MkIwMTlFcGVlcmludGZwb24gNTA1'},
                                                                                                          {       'in_nucast_pkts': '168',
                                                                                                                  'in_octets': '22140735047',
                                                                                                                  'in_ucast_pkts': '52366839',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1425841',
                                                                                                                  'out_octets': '130498193743',
                                                                                                                  'out_ucast_pkts': '114871893',
                                                                                                                  'poller_hash': 'MzAwRDk1cGVlcmludGZwb24gNTA5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '379243053914',
                                                                                                                  'in_ucast_pkts': '2271934954',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '9173905381904',
                                                                                                                  'out_ucast_pkts': '6349221367',
                                                                                                                  'poller_hash': 'MjA1REE0cGVlcmludGZwb24gNzMz'},
                                                                                                          {       'in_nucast_pkts': '11874735',
                                                                                                                  'in_octets': '432190365135',
                                                                                                                  'in_ucast_pkts': '2835842218',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '129779391',
                                                                                                                  'out_octets': '12172925921577',
                                                                                                                  'out_ucast_pkts': '8746107712',
                                                                                                                  'poller_hash': 'NDU5NkE1cGVlcmludGZwb24gMjA1NjIx'},
                                                                                                          {       'in_nucast_pkts': '53726',
                                                                                                                  'in_octets': '28129212154',
                                                                                                                  'in_ucast_pkts': '42174423',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6710277',
                                                                                                                  'out_octets': '109447245014',
                                                                                                                  'out_ucast_pkts': '88884870',
                                                                                                                  'poller_hash': 'MTE5NTFBN3BlZXJpbnRmcG9uIDIxNjcyMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '95844413562',
                                                                                                                  'in_ucast_pkts': '727488710',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '2181348364913',
                                                                                                                  'out_ucast_pkts': '1636470433',
                                                                                                                  'poller_hash': 'MTI3NTA3MnBlZXJpbnRmcG9uIDIyMTg1OA=='},
                                                                                                          {       'in_nucast_pkts': '117',
                                                                                                                  'in_octets': '6580856036',
                                                                                                                  'in_ucast_pkts': '14250980',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2196956',
                                                                                                                  'out_octets': '24046090708',
                                                                                                                  'out_ucast_pkts': '24685051',
                                                                                                                  'poller_hash': 'MTI5NzM0Q3BlZXJpbnRmcG9uIDI1NjM2Mw=='},
                                                                                                          {       'in_nucast_pkts': '22',
                                                                                                                  'in_octets': '65391626372',
                                                                                                                  'in_ucast_pkts': '71293568',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2607553',
                                                                                                                  'out_octets': '128935903957',
                                                                                                                  'out_ucast_pkts': '107209559',
                                                                                                                  'poller_hash': 'MzVFOTEzcGVlcmludGZwb24gMjU3NDE5'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '65236523077',
                                                                                         'out_octets': '434363105368699',
                                                                                         'out_ucast_pkts': '256948759737',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/1',
                                                                                                                 'id': '11/1',
                                                                                                                 'links': [       {       'connected_time': '24669263',
                                                                                                                                          'fiber_length': 18295,
                                                                                                                                          'poller_hash': 'NDRBQUE4cGVlcnBvbnBvbiA0OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '67790817',
                                                                                                                                          'fiber_length': 18242,
                                                                                                                                          'poller_hash': 'ODJBNkNCcGVlcnBvbnBvbiAxMDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '24664987',
                                                                                                                                          'fiber_length': 17944,
                                                                                                                                          'poller_hash': 'MjZDQzU0cGVlcnBvbnBvbiAxNTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '21816587',
                                                                                                                                          'fiber_length': 17672,
                                                                                                                                          'poller_hash': 'Mjk4NDQycGVlcnBvbnBvbiAxOTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '21817477',
                                                                                                                                          'fiber_length': 17554,
                                                                                                                                          'poller_hash': 'MUQwNjdGcGVlcnBvbnBvbiAxOTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '21817189',
                                                                                                                                          'fiber_length': 17696,
                                                                                                                                          'poller_hash': 'MUQwNkUxcGVlcnBvbnBvbiAyMDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '3113962',
                                                                                                                                          'fiber_length': 18869,
                                                                                                                                          'poller_hash': 'MkJENDc4cGVlcnBvbnBvbiAyNDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '3113990',
                                                                                                                                          'fiber_length': 17883,
                                                                                                                                          'poller_hash': 'MkJENDc0cGVlcnBvbnBvbiAyNDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '47920',
                                                                                                                                          'fiber_length': 18273,
                                                                                                                                          'poller_hash': 'N0UwRURwZWVycG9ucG9uIDI1MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '9960372',
                                                                                                                                          'fiber_length': 18806,
                                                                                                                                          'poller_hash': 'MUQwNjgzcGVlcnBvbnBvbiAyNTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '21816447',
                                                                                                                                          'fiber_length': 17723,
                                                                                                                                          'poller_hash': 'ODY5M0EwcGVlcnBvbnBvbiAyNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '3113953',
                                                                                                                                          'fiber_length': 17918,
                                                                                                                                          'poller_hash': 'MkIwMjY2cGVlcnBvbnBvbiAyNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '3113977',
                                                                                                                                          'fiber_length': 17982,
                                                                                                                                          'poller_hash': 'MzAwNjRDcGVlcnBvbnBvbiAyNzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '67790291',
                                                                                                                                          'fiber_length': 17982,
                                                                                                                                          'poller_hash': 'MzVGODRDcGVlcnBvbnBvbiAyODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '6158560',
                                                                                                                                          'fiber_length': 18151,
                                                                                                                                          'poller_hash': 'MzAwN0FFcGVlcnBvbnBvbiAyOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '15749555',
                                                                                                                                          'fiber_length': 18168,
                                                                                                                                          'poller_hash': 'MzVGOTUzcGVlcnBvbnBvbiAzMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '47622',
                                                                                                                                          'fiber_length': 17480,
                                                                                                                                          'poller_hash': 'MzhGMjg1cGVlcnBvbnBvbiAzMjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '21816299',
                                                                                                                                          'fiber_length': 20176,
                                                                                                                                          'poller_hash': 'MkIwMTUwcGVlcnBvbnBvbiAzNDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '64131621',
                                                                                                                                          'fiber_length': 18080,
                                                                                                                                          'poller_hash': 'MkFEMEJDcGVlcnBvbnBvbiA0OTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '67790585',
                                                                                                                                          'fiber_length': 18384,
                                                                                                                                          'poller_hash': 'MzVFOUNEcGVlcnBvbnBvbiA1MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '2327514',
                                                                                                                                          'fiber_length': 18560,
                                                                                                                                          'poller_hash': 'MkIwMTlFcGVlcnBvbnBvbiA1MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '670617',
                                                                                                                                          'fiber_length': 19562,
                                                                                                                                          'poller_hash': 'MzAwRDk1cGVlcnBvbnBvbiA1MDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '1646454',
                                                                                                                                          'fiber_length': 18040,
                                                                                                                                          'poller_hash': 'MjA1REE0cGVlcnBvbnBvbiA3MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '24669403',
                                                                                                                                          'fiber_length': 18101,
                                                                                                                                          'poller_hash': 'NDU5NkE1cGVlcnBvbnBvbiAyMDU2MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '3113968',
                                                                                                                                          'fiber_length': 18147,
                                                                                                                                          'poller_hash': 'MTE5NTFBN3BlZXJwb25wb24gMjE2NzIy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '3113962',
                                                                                                                                          'fiber_length': 17591,
                                                                                                                                          'poller_hash': 'MTI3NTA3MnBlZXJwb25wb24gMjIxODU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '1026015',
                                                                                                                                          'fiber_length': 17656,
                                                                                                                                          'poller_hash': 'MTI5NzM0Q3BlZXJwb25wb24gMjU2MzYz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688},
                                                                                                                                  {       'connected_time': '1213487',
                                                                                                                                          'fiber_length': 18824,
                                                                                                                                          'poller_hash': 'MzVFOTEzcGVlcnBvbnBvbiAyNTc0MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.688}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/1',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '11/2',
                                                                                         'in_discards': '16731',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2199123',
                                                                                         'in_octets': '21666066241607',
                                                                                         'in_ucast_pkts': '90873652540',
                                                                                         'links': [       {       'in_nucast_pkts': '12',
                                                                                                                  'in_octets': '12823604311',
                                                                                                                  'in_ucast_pkts': '64016436',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720760',
                                                                                                                  'out_octets': '412168046980',
                                                                                                                  'out_ucast_pkts': '292932788',
                                                                                                                  'poller_hash': 'NDJDNzAycGVlcmludGZwb24gMTgy'},
                                                                                                          {       'in_nucast_pkts': '50665',
                                                                                                                  'in_octets': '385176830089',
                                                                                                                  'in_ucast_pkts': '1680662745',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2254656',
                                                                                                                  'out_octets': '4280997515986',
                                                                                                                  'out_ucast_pkts': '3175673843',
                                                                                                                  'poller_hash': 'MkZFQTA1cGVlcmludGZwb24gMjA5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '605784893751',
                                                                                                                  'in_ucast_pkts': '2212673016',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4718465853771',
                                                                                                                  'out_ucast_pkts': '5438864360',
                                                                                                                  'poller_hash': 'NTVGOUM0cGVlcmludGZwb24gMjEw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1234407830421',
                                                                                                                  'in_ucast_pkts': '7061070794',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '25850242016602',
                                                                                                                  'out_ucast_pkts': '18890758956',
                                                                                                                  'poller_hash': 'MzI5NjM1cGVlcmludGZwb24gMjEx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '130301967802',
                                                                                                                  'in_ucast_pkts': '1496225785',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4439813772856',
                                                                                                                  'out_ucast_pkts': '4578328053',
                                                                                                                  'poller_hash': 'MkU5MUE4cGVlcmludGZwb24gMjEz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '776308712761',
                                                                                                                  'in_ucast_pkts': '2059048154',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4725623113103',
                                                                                                                  'out_ucast_pkts': '461531967',
                                                                                                                  'poller_hash': 'MzYyODExcGVlcmludGZwb24gMjE2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NEQ0MjNGcGVlcmludGZwb24gMjE3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '290303410117',
                                                                                                                  'in_ucast_pkts': '1174773463',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2797280335891',
                                                                                                                  'out_ucast_pkts': '3256154412',
                                                                                                                  'poller_hash': 'NzFEN0FFcGVlcmludGZwb24gMjE5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '300587837994',
                                                                                                                  'in_ucast_pkts': '2279709844',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7003510868921',
                                                                                                                  'out_ucast_pkts': '6603251605',
                                                                                                                  'poller_hash': 'QTlGMDhwZWVyaW50ZnBvbiAyMjA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1144637806514',
                                                                                                                  'in_ucast_pkts': '6469031392',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '29035185803645',
                                                                                                                  'out_ucast_pkts': '20283249639',
                                                                                                                  'poller_hash': 'MkVEOTJFcGVlcmludGZwb24gMjIx'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '421794526',
                                                                                                                  'in_ucast_pkts': '3031359',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '91919',
                                                                                                                  'out_octets': '22614230448',
                                                                                                                  'out_ucast_pkts': '15826640',
                                                                                                                  'poller_hash': 'QURCMDNwZWVyaW50ZnBvbiAyMjI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1790273861516',
                                                                                                                  'in_ucast_pkts': '19378838180',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '36685835266799',
                                                                                                                  'out_ucast_pkts': '29663067210',
                                                                                                                  'poller_hash': 'MTk5QTJFcGVlcmludGZwb24gMjI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '421123122322',
                                                                                                                  'in_ucast_pkts': '3503073918',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '18494402901878',
                                                                                                                  'out_ucast_pkts': '12788711225',
                                                                                                                  'poller_hash': 'Q0MwOENwZWVyaW50ZnBvbiAyMjc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'Q0MwQjhwZWVyaW50ZnBvbiAyMjg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '388957119287',
                                                                                                                  'in_ucast_pkts': '6146036459',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '23270866151625',
                                                                                                                  'out_ucast_pkts': '15684202556',
                                                                                                                  'poller_hash': 'Q0MwQjRwZWVyaW50ZnBvbiAyMjk='},
                                                                                                          {       'in_nucast_pkts': '475681',
                                                                                                                  'in_octets': '441748044594',
                                                                                                                  'in_ucast_pkts': '1692218021',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43301817',
                                                                                                                  'out_octets': '10099883567244',
                                                                                                                  'out_ucast_pkts': '7046742153',
                                                                                                                  'poller_hash': 'QzRGRTdwZWVyaW50ZnBvbiAyMzA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '509048167119',
                                                                                                                  'in_ucast_pkts': '2289279176',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7919304933877',
                                                                                                                  'out_ucast_pkts': '6465191809',
                                                                                                                  'poller_hash': 'MTJGRkU0cGVlcmludGZwb24gMjMx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '356443255584',
                                                                                                                  'in_ucast_pkts': '7079346551',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9852098274003',
                                                                                                                  'out_ucast_pkts': '15471411134',
                                                                                                                  'poller_hash': 'MTMwMEExcGVlcmludGZwb24gMjMy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '665671338867',
                                                                                                                  'in_ucast_pkts': '3187441351',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11512177217865',
                                                                                                                  'out_ucast_pkts': '10413784912',
                                                                                                                  'poller_hash': 'MTMwMEEycGVlcmludGZwb24gMjMz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '222814055389',
                                                                                                                  'in_ucast_pkts': '2940389082',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7109767816768',
                                                                                                                  'out_ucast_pkts': '10009866710',
                                                                                                                  'poller_hash': 'MTJGRkUxcGVlcmludGZwb24gMjM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTMwMDlDcGVlcmludGZwb24gMjM2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '157641945495',
                                                                                                                  'in_ucast_pkts': '651112078',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '819674358739',
                                                                                                                  'out_ucast_pkts': '1096654711',
                                                                                                                  'poller_hash': 'NzFEODk0cGVlcmludGZwb24gMjM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '232540305286',
                                                                                                                  'in_ucast_pkts': '1553385517',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7559114966770',
                                                                                                                  'out_ucast_pkts': '1245909078',
                                                                                                                  'poller_hash': 'MTMwMDkxcGVlcmludGZwb24gMjM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1132566196341',
                                                                                                                  'in_ucast_pkts': '3426746798',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '14378537708672',
                                                                                                                  'out_ucast_pkts': '11022256980',
                                                                                                                  'poller_hash': 'MjA1QkRDcGVlcmludGZwb24gMjk1'},
                                                                                                          {       'in_nucast_pkts': '53872',
                                                                                                                  'in_octets': '67204106042',
                                                                                                                  'in_ucast_pkts': '103342059',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720887',
                                                                                                                  'out_octets': '284928575721',
                                                                                                                  'out_ucast_pkts': '262411610',
                                                                                                                  'poller_hash': 'QzgzQkNGcGVlcmludGZwb24gMTc0OTk5'},
                                                                                                          {       'in_nucast_pkts': '2253964',
                                                                                                                  'in_octets': '795638034',
                                                                                                                  'in_ucast_pkts': '1942956',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1856852312',
                                                                                                                  'out_octets': '2496027215859',
                                                                                                                  'out_ucast_pkts': '2278158',
                                                                                                                  'poller_hash': 'QzgxQzUycGVlcmludGZwb24gMjA1MDgw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '331096658015',
                                                                                                                  'in_ucast_pkts': '30652761136',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '3404041477012',
                                                                                                                  'out_ucast_pkts': '6088578382',
                                                                                                                  'poller_hash': 'MTI3NDcwMnBlZXJpbnRmcG9uIDIzNjE4Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '61869152720',
                                                                                                                  'in_ucast_pkts': '158647580',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '21',
                                                                                                                  'out_octets': '518751538612',
                                                                                                                  'out_ucast_pkts': '424099703',
                                                                                                                  'poller_hash': 'MTE3REIyRXBlZXJpbnRmcG9uIDI1MTEzMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1328043959817',
                                                                                                                  'in_ucast_pkts': '10746913662',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '22343736429338',
                                                                                                                  'out_ucast_pkts': '20976038016',
                                                                                                                  'poller_hash': 'NEQ0MjM5cGVlcmludGZwb24gMjUzNDM2'},
                                                                                                          {       'in_nucast_pkts': '177080',
                                                                                                                  'in_octets': '137916330157',
                                                                                                                  'in_ucast_pkts': '1469507078',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18867298',
                                                                                                                  'out_octets': '5364328933484',
                                                                                                                  'out_ucast_pkts': '3645932598',
                                                                                                                  'poller_hash': 'MTI3MEQ1MXBlZXJpbnRmcG9uIDI3ODE0OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2064620939',
                                                                                                                  'in_ucast_pkts': '16073877',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '107068424889',
                                                                                                                  'out_ucast_pkts': '78665628',
                                                                                                                  'poller_hash': 'MTE5NEE5N3BlZXJpbnRmcG9uIDI4MjI3Mw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '248463967313',
                                                                                         'out_octets': '696188545306208',
                                                                                         'out_ucast_pkts': '262459664291',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/2',
                                                                                                                 'id': '11/2',
                                                                                                                 'links': [       {       'connected_time': '3114031',
                                                                                                                                          'fiber_length': 17415,
                                                                                                                                          'poller_hash': 'NDJDNzAycGVlcnBvbnBvbiAxODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '1363457',
                                                                                                                                          'fiber_length': 18174,
                                                                                                                                          'poller_hash': 'MkZFQTA1cGVlcnBvbnBvbiAyMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47455',
                                                                                                                                          'fiber_length': 17506,
                                                                                                                                          'poller_hash': 'NTVGOUM0cGVlcnBvbnBvbiAyMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '23325508',
                                                                                                                                          'fiber_length': 17700,
                                                                                                                                          'poller_hash': 'MzI5NjM1cGVlcnBvbnBvbiAyMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47804',
                                                                                                                                          'fiber_length': 17407,
                                                                                                                                          'poller_hash': 'MkU5MUE4cGVlcnBvbnBvbiAyMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '69225897',
                                                                                                                                          'fiber_length': 18008,
                                                                                                                                          'poller_hash': 'MzYyODExcGVlcnBvbnBvbiAyMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47971',
                                                                                                                                          'fiber_length': 17409,
                                                                                                                                          'poller_hash': 'NEQ0MjNGcGVlcnBvbnBvbiAyMTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '24653006',
                                                                                                                                          'fiber_length': 17759,
                                                                                                                                          'poller_hash': 'NzFEN0FFcGVlcnBvbnBvbiAyMTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47486',
                                                                                                                                          'fiber_length': 17522,
                                                                                                                                          'poller_hash': 'QTlGMDhwZWVycG9ucG9uIDIyMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47317',
                                                                                                                                          'fiber_length': 17988,
                                                                                                                                          'poller_hash': 'MkVEOTJFcGVlcnBvbnBvbiAyMjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '58586',
                                                                                                                                          'fiber_length': 17981,
                                                                                                                                          'poller_hash': 'QURCMDNwZWVycG9ucG9uIDIyMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47747',
                                                                                                                                          'fiber_length': 17342,
                                                                                                                                          'poller_hash': 'MTk5QTJFcGVlcnBvbnBvbiAyMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47289',
                                                                                                                                          'fiber_length': 17884,
                                                                                                                                          'poller_hash': 'Q0MwOENwZWVycG9ucG9uIDIyNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '69226064',
                                                                                                                                          'fiber_length': 18095,
                                                                                                                                          'poller_hash': 'Q0MwQjhwZWVycG9ucG9uIDIyOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47331',
                                                                                                                                          'fiber_length': 17879,
                                                                                                                                          'poller_hash': 'Q0MwQjRwZWVycG9ucG9uIDIyOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '24669034',
                                                                                                                                          'fiber_length': 17422,
                                                                                                                                          'poller_hash': 'QzRGRTdwZWVycG9ucG9uIDIzMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '996059',
                                                                                                                                          'fiber_length': 18054,
                                                                                                                                          'poller_hash': 'MTJGRkU0cGVlcnBvbnBvbiAyMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '48043',
                                                                                                                                          'fiber_length': 17745,
                                                                                                                                          'poller_hash': 'MTMwMEExcGVlcnBvbnBvbiAyMzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '3113961',
                                                                                                                                          'fiber_length': 17986,
                                                                                                                                          'poller_hash': 'MTMwMEEycGVlcnBvbnBvbiAyMzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '268687',
                                                                                                                                          'fiber_length': 17639,
                                                                                                                                          'poller_hash': 'MTJGRkUxcGVlcnBvbnBvbiAyMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '69226176',
                                                                                                                                          'fiber_length': 17519,
                                                                                                                                          'poller_hash': 'MTMwMDlDcGVlcnBvbnBvbiAyMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47934',
                                                                                                                                          'fiber_length': 18347,
                                                                                                                                          'poller_hash': 'NzFEODk0cGVlcnBvbnBvbiAyMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '67245829',
                                                                                                                                          'fiber_length': 18185,
                                                                                                                                          'poller_hash': 'MTMwMDkxcGVlcnBvbnBvbiAyMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '3113063',
                                                                                                                                          'fiber_length': 18000,
                                                                                                                                          'poller_hash': 'MjA1QkRDcGVlcnBvbnBvbiAyOTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '3113979',
                                                                                                                                          'fiber_length': 17525,
                                                                                                                                          'poller_hash': 'QzgzQkNGcGVlcnBvbnBvbiAxNzQ5OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '2969395',
                                                                                                                                          'fiber_length': 17761,
                                                                                                                                          'poller_hash': 'QzgxQzUycGVlcnBvbnBvbiAyMDUwODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -28.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '8579811',
                                                                                                                                          'fiber_length': 18215,
                                                                                                                                          'poller_hash': 'MTI3NDcwMnBlZXJwb25wb24gMjM2MTg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '3114005',
                                                                                                                                          'fiber_length': 18437,
                                                                                                                                          'poller_hash': 'MTE3REIyRXBlZXJwb25wb24gMjUxMTMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '47956',
                                                                                                                                          'fiber_length': 17672,
                                                                                                                                          'poller_hash': 'NEQ0MjM5cGVlcnBvbnBvbiAyNTM0MzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '8836008',
                                                                                                                                          'fiber_length': 18263,
                                                                                                                                          'poller_hash': 'MTI3MEQ1MXBlZXJwb25wb24gMjc4MTQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667},
                                                                                                                                  {       'connected_time': '2873093',
                                                                                                                                          'fiber_length': 17545,
                                                                                                                                          'poller_hash': 'MTE5NEE5N3BlZXJwb25wb24gMjgyMjcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.667}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/2',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '11/3',
                                                                                         'in_discards': '25504',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7858380',
                                                                                         'in_octets': '14380464507321',
                                                                                         'in_ucast_pkts': '69881592684',
                                                                                         'links': [       {       'in_nucast_pkts': '15895',
                                                                                                                  'in_octets': '539591788956',
                                                                                                                  'in_ucast_pkts': '951727088',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '55261211',
                                                                                                                  'out_octets': '1707445021963',
                                                                                                                  'out_ucast_pkts': '1425479312',
                                                                                                                  'poller_hash': 'ODY5M0UxcGVlcmludGZwb24gNw=='},
                                                                                                          {       'in_nucast_pkts': '4999',
                                                                                                                  'in_octets': '3201098135',
                                                                                                                  'in_ucast_pkts': '13470400',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '497761',
                                                                                                                  'out_octets': '56579642083',
                                                                                                                  'out_ucast_pkts': '44159572',
                                                                                                                  'poller_hash': 'MzIyN0YxcGVlcmludGZwb24gMTkw'},
                                                                                                          {       'in_nucast_pkts': '472392',
                                                                                                                  'in_octets': '86060050891',
                                                                                                                  'in_ucast_pkts': '606858471',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43307516',
                                                                                                                  'out_octets': '3491340125103',
                                                                                                                  'out_ucast_pkts': '2411544542',
                                                                                                                  'poller_hash': 'MkZFQzA4cGVlcmludGZwb24gMjQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '69886258408',
                                                                                                                  'in_ucast_pkts': '257444202',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '876515942884',
                                                                                                                  'out_ucast_pkts': '807027882',
                                                                                                                  'poller_hash': 'Mzk0MERFcGVlcmludGZwb24gMzE1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '302345920939',
                                                                                                                  'in_ucast_pkts': '310494896',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '296127375638',
                                                                                                                  'out_ucast_pkts': '424845624',
                                                                                                                  'poller_hash': 'MzI5NjdEcGVlcmludGZwb24gMzE2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MkZFQjYycGVlcmludGZwb24gMzE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '392723519514',
                                                                                                                  'in_ucast_pkts': '2912686552',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11078120275692',
                                                                                                                  'out_ucast_pkts': '7778721000',
                                                                                                                  'poller_hash': 'NTYwQzEwcGVlcmludGZwb24gMzE5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1033339023230',
                                                                                                                  'in_ucast_pkts': '10299356035',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '32784323960311',
                                                                                                                  'out_ucast_pkts': '22347479799',
                                                                                                                  'poller_hash': 'NTYwMjU4cGVlcmludGZwb24gMzIz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '442597800125',
                                                                                                                  'in_ucast_pkts': '1825637601',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6187844234126',
                                                                                                                  'out_ucast_pkts': '4555944444',
                                                                                                                  'poller_hash': 'MkZFQjVEcGVlcmludGZwb24gMzI0'},
                                                                                                          {       'in_nucast_pkts': '16536092',
                                                                                                                  'in_octets': '387461055298',
                                                                                                                  'in_ucast_pkts': '3174687259',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '20946732413',
                                                                                                                  'out_octets': '37550044658505',
                                                                                                                  'out_ucast_pkts': '6985138362',
                                                                                                                  'poller_hash': 'ODc0MTRwZWVyaW50ZnBvbiAzMjU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '498755433614',
                                                                                                                  'in_ucast_pkts': '3688804242',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '16121703744420',
                                                                                                                  'out_ucast_pkts': '7173665840',
                                                                                                                  'poller_hash': 'MkZFOUU1cGVlcmludGZwb24gMzMw'},
                                                                                                          {       'in_nucast_pkts': '49978',
                                                                                                                  'in_octets': '21622402510',
                                                                                                                  'in_ucast_pkts': '97539488',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4076844',
                                                                                                                  'out_octets': '462007352492',
                                                                                                                  'out_ucast_pkts': '343590504',
                                                                                                                  'poller_hash': 'MkZFQTBEcGVlcmludGZwb24gMzMx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '194649737388',
                                                                                                                  'in_ucast_pkts': '781157869',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2881960485554',
                                                                                                                  'out_ucast_pkts': '2436212660',
                                                                                                                  'poller_hash': 'MzkwOEY5cGVlcmludGZwb24gMzM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '932200058794',
                                                                                                                  'in_ucast_pkts': '5347896874',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '17436861497665',
                                                                                                                  'out_ucast_pkts': '12664301699',
                                                                                                                  'poller_hash': 'MkZFOUM5cGVlcmludGZwb24gMzM2'},
                                                                                                          {       'in_nucast_pkts': '3688',
                                                                                                                  'in_octets': '161852628717',
                                                                                                                  'in_ucast_pkts': '598537909',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27790844',
                                                                                                                  'out_octets': '1359732485616',
                                                                                                                  'out_ucast_pkts': '1108931423',
                                                                                                                  'poller_hash': 'MzVFOEZCcGVlcmludGZwb24gNzEw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '10862245217',
                                                                                                                  'in_ucast_pkts': '87699159',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '172',
                                                                                                                  'out_octets': '484311763963',
                                                                                                                  'out_ucast_pkts': '378802987',
                                                                                                                  'poller_hash': 'MzIxN0NDcGVlcmludGZwb24gNzI3'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '24126251371',
                                                                                                                  'in_ucast_pkts': '177352498',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720873',
                                                                                                                  'out_octets': '813670843902',
                                                                                                                  'out_ucast_pkts': '601043869',
                                                                                                                  'poller_hash': 'MzIxNzc5cGVlcmludGZwb24gNzI4'},
                                                                                                          {       'in_nucast_pkts': '54158',
                                                                                                                  'in_octets': '15176311322',
                                                                                                                  'in_ucast_pkts': '70038195',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6720776',
                                                                                                                  'out_octets': '712164327937',
                                                                                                                  'out_ucast_pkts': '512131825',
                                                                                                                  'poller_hash': 'MzIxOTRCcGVlcmludGZwb24gNzMw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '862122861',
                                                                                                                  'in_ucast_pkts': '5005525',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4576378343',
                                                                                                                  'out_octets': '6152953088334',
                                                                                                                  'out_ucast_pkts': '2383032',
                                                                                                                  'poller_hash': 'MzIxODI1cGVlcmludGZwb24gNzM4'},
                                                                                                          {       'in_nucast_pkts': '51424',
                                                                                                                  'in_octets': '33340582090',
                                                                                                                  'in_ucast_pkts': '136218338',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6407370',
                                                                                                                  'out_octets': '980454595075',
                                                                                                                  'out_ucast_pkts': '696829865',
                                                                                                                  'poller_hash': 'MTI3NEE2MHBlZXJpbnRmcG9uIDExNDc4MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '93560022661',
                                                                                                                  'in_ucast_pkts': '698785933',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '3559463101606',
                                                                                                                  'out_ucast_pkts': '2645017153',
                                                                                                                  'poller_hash': 'OUMyRjA4cGVlcmludGZwb24gMTQ5NzU4'},
                                                                                                          {       'in_nucast_pkts': '170611',
                                                                                                                  'in_octets': '34580550909',
                                                                                                                  'in_ucast_pkts': '41934402',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '20149158',
                                                                                                                  'out_octets': '44527335516',
                                                                                                                  'out_ucast_pkts': '47944443',
                                                                                                                  'poller_hash': 'QzgyMEQycGVlcmludGZwb24gMTYyMjk2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2167863362269',
                                                                                                                  'in_ucast_pkts': '442920531401',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '11450596700946',
                                                                                                                  'out_ucast_pkts': '8807519232',
                                                                                                                  'poller_hash': 'MTE5NTE0RHBlZXJpbnRmcG9uIDIxNzk2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '287247957729',
                                                                                                                  'in_ucast_pkts': '57731141044',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '8037748210112',
                                                                                                                  'out_ucast_pkts': '5583083504',
                                                                                                                  'poller_hash': 'MTI3NEE0Q3BlZXJpbnRmcG9uIDIzNTQ0OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '276252340980',
                                                                                                                  'in_ucast_pkts': '1884232501',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '86',
                                                                                                                  'out_octets': '10536060815107',
                                                                                                                  'out_ucast_pkts': '7289618910',
                                                                                                                  'poller_hash': 'MTI3NEE4QXBlZXJpbnRmcG9uIDIzNjEwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '241937932914',
                                                                                                                  'in_ucast_pkts': '1479219736',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '30',
                                                                                                                  'out_octets': '11199182335669',
                                                                                                                  'out_ucast_pkts': '7754297207',
                                                                                                                  'poller_hash': 'MTI3NDJBOHBlZXJpbnRmcG9uIDI0MTQ2NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '39882971944',
                                                                                                                  'in_ucast_pkts': '190391384',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '940788408198',
                                                                                                                  'out_ucast_pkts': '701824884',
                                                                                                                  'poller_hash': 'NEQxMTRDcGVlcmludGZwb24gMjcyMjk0'},
                                                                                                          {       'in_nucast_pkts': '16778',
                                                                                                                  'in_octets': '35938820116',
                                                                                                                  'in_ucast_pkts': '284852556',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2032364',
                                                                                                                  'out_octets': '1300019208686',
                                                                                                                  'out_ucast_pkts': '893205539',
                                                                                                                  'poller_hash': 'MTI5NjBEOHBlZXJpbnRmcG9uIDI4NDg0NQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '158392705697',
                                                                                         'out_octets': '483510740773129',
                                                                                         'out_ucast_pkts': '192382181026',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/3',
                                                                                                                 'id': '11/3',
                                                                                                                 'links': [       {       'connected_time': '11410495',
                                                                                                                                          'fiber_length': 18235,
                                                                                                                                          'poller_hash': 'ODY5M0UxcGVlcnBvbnBvbiA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '241694',
                                                                                                                                          'fiber_length': 18595,
                                                                                                                                          'poller_hash': 'MzIyN0YxcGVlcnBvbnBvbiAxOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '24669410',
                                                                                                                                          'fiber_length': 18437,
                                                                                                                                          'poller_hash': 'MkZFQzA4cGVlcnBvbnBvbiAyNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '47915',
                                                                                                                                          'fiber_length': 17535,
                                                                                                                                          'poller_hash': 'Mzk0MERFcGVlcnBvbnBvbiAzMTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '28950839',
                                                                                                                                          'fiber_length': 18041,
                                                                                                                                          'poller_hash': 'MzI5NjdEcGVlcnBvbnBvbiAzMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '3113969',
                                                                                                                                          'fiber_length': 18485,
                                                                                                                                          'poller_hash': 'MkZFQjYycGVlcnBvbnBvbiAzMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '23325143',
                                                                                                                                          'fiber_length': 17825,
                                                                                                                                          'poller_hash': 'NTYwQzEwcGVlcnBvbnBvbiAzMTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '66690374',
                                                                                                                                          'fiber_length': 17416,
                                                                                                                                          'poller_hash': 'NTYwMjU4cGVlcnBvbnBvbiAzMjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '48182',
                                                                                                                                          'fiber_length': 17816,
                                                                                                                                          'poller_hash': 'MkZFQjVEcGVlcnBvbnBvbiAzMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '24669203',
                                                                                                                                          'fiber_length': 17711,
                                                                                                                                          'poller_hash': 'ODc0MTRwZWVycG9ucG9uIDMyNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '13642953',
                                                                                                                                          'fiber_length': 17555,
                                                                                                                                          'poller_hash': 'MkZFOUU1cGVlcnBvbnBvbiAzMzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '2443388',
                                                                                                                                          'fiber_length': 18472,
                                                                                                                                          'poller_hash': 'MkZFQTBEcGVlcnBvbnBvbiAzMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '29417607',
                                                                                                                                          'fiber_length': 17812,
                                                                                                                                          'poller_hash': 'MzkwOEY5cGVlcnBvbnBvbiAzMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '48400',
                                                                                                                                          'fiber_length': 17819,
                                                                                                                                          'poller_hash': 'MkZFOUM5cGVlcnBvbnBvbiAzMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '12861944',
                                                                                                                                          'fiber_length': 17606,
                                                                                                                                          'poller_hash': 'MzVFOEZCcGVlcnBvbnBvbiA3MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -28.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '506878',
                                                                                                                                          'fiber_length': 18537,
                                                                                                                                          'poller_hash': 'MzIxN0NDcGVlcnBvbnBvbiA3Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '3113993',
                                                                                                                                          'fiber_length': 17923,
                                                                                                                                          'poller_hash': 'MzIxNzc5cGVlcnBvbnBvbiA3Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '3113982',
                                                                                                                                          'fiber_length': 18248,
                                                                                                                                          'poller_hash': 'MzIxOTRCcGVlcnBvbnBvbiA3MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '21816241',
                                                                                                                                          'fiber_length': 18744,
                                                                                                                                          'poller_hash': 'MzIxODI1cGVlcnBvbnBvbiA3Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '2969414',
                                                                                                                                          'fiber_length': 17768,
                                                                                                                                          'poller_hash': 'MTI3NEE2MHBlZXJwb25wb24gMTE0Nzgx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '3114039',
                                                                                                                                          'fiber_length': 17961,
                                                                                                                                          'poller_hash': 'OUMyRjA4cGVlcnBvbnBvbiAxNDk3NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '9239014',
                                                                                                                                          'fiber_length': 18282,
                                                                                                                                          'poller_hash': 'QzgyMEQycGVlcnBvbnBvbiAxNjIyOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '12942573',
                                                                                                                                          'fiber_length': 17535,
                                                                                                                                          'poller_hash': 'MTE5NTE0RHBlZXJwb25wb24gMjE3OTY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '3113952',
                                                                                                                                          'fiber_length': 17910,
                                                                                                                                          'poller_hash': 'MTI3NEE0Q3BlZXJwb25wb24gMjM1NDQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '1206725',
                                                                                                                                          'fiber_length': 18739,
                                                                                                                                          'poller_hash': 'MTI3NEE4QXBlZXJwb25wb24gMjM2MTAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '15575842',
                                                                                                                                          'fiber_length': 17826,
                                                                                                                                          'poller_hash': 'MTI3NDJBOHBlZXJwb25wb24gMjQxNDY1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '15813850',
                                                                                                                                          'fiber_length': 17629,
                                                                                                                                          'poller_hash': 'NEQxMTRDcGVlcnBvbnBvbiAyNzIyOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699},
                                                                                                                                  {       'connected_time': '951398',
                                                                                                                                          'fiber_length': 17577,
                                                                                                                                          'poller_hash': 'MTI5NjBEOHBlZXJwb25wb24gMjg0ODQ1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.699}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/3',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '11/4',
                                                                                         'in_discards': '5171720',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '13566326',
                                                                                         'in_octets': '124299157897915',
                                                                                         'in_ucast_pkts': '188692872267',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/4',
                                                                                         'oper_up': False,
                                                                                         'out_nucast_pkts': '70934214025',
                                                                                         'out_octets': '522460006485643',
                                                                                         'out_ucast_pkts': '345238888749',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS80',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/4',
                                                                                                                 'id': '11/4',
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/4',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS80',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '11/5',
                                                                                         'in_discards': '4',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '698820',
                                                                                         'in_octets': '978544795071',
                                                                                         'in_ucast_pkts': '5666290591',
                                                                                         'links': [       {       'in_nucast_pkts': '436103',
                                                                                                                  'in_octets': '475208273605',
                                                                                                                  'in_ucast_pkts': '1464439033',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43269328',
                                                                                                                  'out_octets': '7978303548788',
                                                                                                                  'out_ucast_pkts': '6403605840',
                                                                                                                  'poller_hash': 'MzI5NjU1cGVlcmludGZwb24gMzY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '138730074625',
                                                                                                                  'in_ucast_pkts': '1430507108',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '5340784566626',
                                                                                                                  'out_ucast_pkts': '3643271098',
                                                                                                                  'poller_hash': 'NDJDNkQycGVlcmludGZwb24gNDEw'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '1771962812',
                                                                                         'out_octets': '28319316986904',
                                                                                         'out_ucast_pkts': '19156591430',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS81',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/5',
                                                                                                                 'id': '11/5',
                                                                                                                 'links': [       {       'connected_time': '24669314',
                                                                                                                                          'fiber_length': 18019,
                                                                                                                                          'poller_hash': 'MzI5NjU1cGVlcnBvbnBvbiAzNjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.926},
                                                                                                                                  {       'connected_time': '2091613',
                                                                                                                                          'fiber_length': 18107,
                                                                                                                                          'poller_hash': 'NDJDNkQycGVlcnBvbnBvbiA0MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.926}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/5',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS81',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '11/6',
                                                                                         'in_discards': '517828',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '6380738',
                                                                                         'in_octets': '17269116079429',
                                                                                         'in_ucast_pkts': '52103512538',
                                                                                         'links': [       {       'in_nucast_pkts': '3321',
                                                                                                                  'in_octets': '89039843039',
                                                                                                                  'in_ucast_pkts': '190568151',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '21519927',
                                                                                                                  'out_octets': '283886901934',
                                                                                                                  'out_ucast_pkts': '284430128',
                                                                                                                  'poller_hash': 'MzVGOUZEcGVlcmludGZwb24gNTMz'},
                                                                                                          {       'in_nucast_pkts': '636',
                                                                                                                  'in_octets': '5660138300',
                                                                                                                  'in_ucast_pkts': '16923957',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120087',
                                                                                                                  'out_octets': '54070125933',
                                                                                                                  'out_ucast_pkts': '41863563',
                                                                                                                  'poller_hash': 'NEY5MkNGcGVlcmludGZwb24gNTM2'},
                                                                                                          {       'in_nucast_pkts': '110',
                                                                                                                  'in_octets': '1674092268000',
                                                                                                                  'in_ucast_pkts': '5114222919',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '65744501',
                                                                                                                  'out_octets': '16775109092921',
                                                                                                                  'out_ucast_pkts': '12962196692',
                                                                                                                  'poller_hash': 'MTFEMDE4cGVlcmludGZwb24gNTY0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '478393861318',
                                                                                                                  'in_ucast_pkts': '859247709',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '6998649661595',
                                                                                                                  'out_ucast_pkts': '11488143595',
                                                                                                                  'poller_hash': 'MzIxNzhFcGVlcmludGZwb24gNzA5'},
                                                                                                          {       'in_nucast_pkts': '18077709',
                                                                                                                  'in_octets': '738119705400',
                                                                                                                  'in_ucast_pkts': '3007608060',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '46297214534',
                                                                                                                  'out_octets': '80582474568245',
                                                                                                                  'out_ucast_pkts': '13072396157',
                                                                                                                  'poller_hash': 'MzIxNzIwcGVlcmludGZwb24gNzE5'},
                                                                                                          {       'in_nucast_pkts': '404335',
                                                                                                                  'in_octets': '1034836608690',
                                                                                                                  'in_ucast_pkts': '3055175465',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83521248',
                                                                                                                  'out_octets': '7638431108637',
                                                                                                                  'out_ucast_pkts': '6800845002',
                                                                                                                  'poller_hash': 'MzVGODQzcGVlcmludGZwb24gNzI5'},
                                                                                                          {       'in_nucast_pkts': '4793',
                                                                                                                  'in_octets': '309756736274',
                                                                                                                  'in_ucast_pkts': '568144457',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83116116',
                                                                                                                  'out_octets': '1173155527394',
                                                                                                                  'out_ucast_pkts': '985417435',
                                                                                                                  'poller_hash': 'MzVFOTVCcGVlcmludGZwb24gNzQw'},
                                                                                                          {       'in_nucast_pkts': '21704',
                                                                                                                  'in_octets': '84979106379',
                                                                                                                  'in_ucast_pkts': '314148003',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33278718',
                                                                                                                  'out_octets': '1897268710109',
                                                                                                                  'out_ucast_pkts': '1359604939',
                                                                                                                  'poller_hash': 'MzVFOTVDcGVlcmludGZwb24gNzU0'},
                                                                                                          {       'in_nucast_pkts': '1251353',
                                                                                                                  'in_octets': '50913880730',
                                                                                                                  'in_ucast_pkts': '327319308',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82761877',
                                                                                                                  'out_octets': '2810196622849',
                                                                                                                  'out_ucast_pkts': '1987899888',
                                                                                                                  'poller_hash': 'MzVFQTA4cGVlcmludGZwb24gNzU3'},
                                                                                                          {       'in_nucast_pkts': '2808',
                                                                                                                  'in_octets': '48536782183',
                                                                                                                  'in_ucast_pkts': '413500718',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33031870',
                                                                                                                  'out_octets': '1594603563706',
                                                                                                                  'out_ucast_pkts': '1156628508',
                                                                                                                  'poller_hash': 'MzVFQTI4cGVlcmludGZwb24gNzU5'},
                                                                                                          {       'in_nucast_pkts': '5488',
                                                                                                                  'in_octets': '12158851338',
                                                                                                                  'in_ucast_pkts': '38818641',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '519533',
                                                                                                                  'out_octets': '136838119728',
                                                                                                                  'out_ucast_pkts': '108478633',
                                                                                                                  'poller_hash': 'MzVFQTJDcGVlcmludGZwb24gNzYw'},
                                                                                                          {       'in_nucast_pkts': '271',
                                                                                                                  'in_octets': '497534777',
                                                                                                                  'in_ucast_pkts': '3279274',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120112',
                                                                                                                  'out_octets': '7955824782',
                                                                                                                  'out_ucast_pkts': '5861404',
                                                                                                                  'poller_hash': 'NDNCMkQ3cGVlcmludGZwb24gNzYx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '181671752739',
                                                                                                                  'in_ucast_pkts': '975137584',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4062300016877',
                                                                                                                  'out_ucast_pkts': '2939280562',
                                                                                                                  'poller_hash': 'MzIyODhDcGVlcmludGZwb24gNzYz'},
                                                                                                          {       'in_nucast_pkts': '10495',
                                                                                                                  'in_octets': '576305704697',
                                                                                                                  'in_ucast_pkts': '2578030805',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145137857',
                                                                                                                  'out_octets': '7893083064217',
                                                                                                                  'out_ucast_pkts': '6073784622',
                                                                                                                  'poller_hash': 'MzIyOENFcGVlcmludGZwb24gNzY0'},
                                                                                                          {       'in_nucast_pkts': '1550',
                                                                                                                  'in_octets': '22643983230',
                                                                                                                  'in_ucast_pkts': '55759384',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9748392',
                                                                                                                  'out_octets': '156561465015',
                                                                                                                  'out_ucast_pkts': '136481982',
                                                                                                                  'poller_hash': 'NUE2NjMycGVlcmludGZwb24gNzY4'},
                                                                                                          {       'in_nucast_pkts': '11966',
                                                                                                                  'in_octets': '2076431296',
                                                                                                                  'in_ucast_pkts': '3477341',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120090',
                                                                                                                  'out_octets': '8831810921',
                                                                                                                  'out_ucast_pkts': '7899344',
                                                                                                                  'poller_hash': 'NDNCNUI3cGVlcmludGZwb24gNzc5'},
                                                                                                          {       'in_nucast_pkts': '33514',
                                                                                                                  'in_octets': '20863308227',
                                                                                                                  'in_ucast_pkts': '67338159',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3582904',
                                                                                                                  'out_octets': '258593828678',
                                                                                                                  'out_ucast_pkts': '212996755',
                                                                                                                  'poller_hash': 'NUE2QTYwcGVlcmludGZwb24gMTM3ODI4'},
                                                                                                          {       'in_nucast_pkts': '58066',
                                                                                                                  'in_octets': '105070639727',
                                                                                                                  'in_ucast_pkts': '263362507',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9364183',
                                                                                                                  'out_octets': '547737904940',
                                                                                                                  'out_ucast_pkts': '674733829',
                                                                                                                  'poller_hash': 'ODJBNzQxcGVlcmludGZwb24gMTQxNjQx'},
                                                                                                          {       'in_nucast_pkts': '4721485',
                                                                                                                  'in_octets': '199582382316',
                                                                                                                  'in_ucast_pkts': '1333483139',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75542556',
                                                                                                                  'out_octets': '7235002350650',
                                                                                                                  'out_ucast_pkts': '5123399121',
                                                                                                                  'poller_hash': 'ODY5NDBFcGVlcmludGZwb24gMTQxODk3'},
                                                                                                          {       'in_nucast_pkts': '399327',
                                                                                                                  'in_octets': '204505333381',
                                                                                                                  'in_ucast_pkts': '1035549133',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '47333432',
                                                                                                                  'out_octets': '6594400870847',
                                                                                                                  'out_ucast_pkts': '4975054125',
                                                                                                                  'poller_hash': 'ODY5M0YxcGVlcmludGZwb24gMTQxOTA3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1994828769929',
                                                                                                                  'in_ucast_pkts': '5482127712',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '422',
                                                                                                                  'out_octets': '15838275637928',
                                                                                                                  'out_ucast_pkts': '12491823706',
                                                                                                                  'poller_hash': 'ODY5M0MwcGVlcmludGZwb24gMTQyMTM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '57200200206',
                                                                                                                  'in_ucast_pkts': '279638401',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '949201525038',
                                                                                                                  'out_ucast_pkts': '814805260',
                                                                                                                  'poller_hash': 'MTU4RkI2RXBlZXJpbnRmcG9uIDIxNTE2MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1533436140721',
                                                                                                                  'in_ucast_pkts': '36729982923',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15',
                                                                                                                  'out_octets': '6567041044091',
                                                                                                                  'out_ucast_pkts': '5489069679',
                                                                                                                  'poller_hash': 'MTI3NTBEN3BlZXJpbnRmcG9uIDIxNzk0Nw=='},
                                                                                                          {       'in_nucast_pkts': '12359',
                                                                                                                  'in_octets': '6817975690',
                                                                                                                  'in_ucast_pkts': '36563327',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1415867',
                                                                                                                  'out_octets': '146351024110',
                                                                                                                  'out_ucast_pkts': '111527071',
                                                                                                                  'poller_hash': 'MTI3NTAzNnBlZXJpbnRmcG9uIDIxNzk3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '170248150821',
                                                                                                                  'in_ucast_pkts': '479085542',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '1336350418314',
                                                                                                                  'out_ucast_pkts': '1114107720',
                                                                                                                  'poller_hash': 'MTI3NEE3OHBlZXJpbnRmcG9uIDIzMTUxMA=='},
                                                                                                          {       'in_nucast_pkts': '657174',
                                                                                                                  'in_octets': '248811250892',
                                                                                                                  'in_ucast_pkts': '561365950',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83664576',
                                                                                                                  'out_octets': '1571349737545',
                                                                                                                  'out_ucast_pkts': '1351286858',
                                                                                                                  'poller_hash': 'MTI3NDMzM3BlZXJpbnRmcG9uIDI0Njk5MA=='},
                                                                                                          {       'in_nucast_pkts': '14625',
                                                                                                                  'in_octets': '4362335716',
                                                                                                                  'in_ucast_pkts': '34159094',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465233',
                                                                                                                  'out_octets': '128420864762',
                                                                                                                  'out_ucast_pkts': '92747116',
                                                                                                                  'poller_hash': 'MTI3NDlCNHBlZXJpbnRmcG9uIDI0ODQxMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '65227167881',
                                                                                                                  'in_ucast_pkts': '294494954',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1030503100642',
                                                                                                                  'out_ucast_pkts': '757633119',
                                                                                                                  'poller_hash': 'MTI3NDczQXBlZXJpbnRmcG9uIDI2MzAxNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzQxMXBlZXJpbnRmcG9uIDI2NTAzMw=='},
                                                                                                          {       'in_nucast_pkts': '9435',
                                                                                                                  'in_octets': '1460324621',
                                                                                                                  'in_ucast_pkts': '2952225',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1149412',
                                                                                                                  'out_octets': '3551493216',
                                                                                                                  'out_ucast_pkts': '3662639',
                                                                                                                  'poller_hash': 'MTI5NUZCMnBlZXJpbnRmcG9uIDI4MDI0NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '67201366054',
                                                                                         'out_octets': '272037277474243',
                                                                                         'out_ucast_pkts': '144463060485',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/6',
                                                                                                                 'id': '11/6',
                                                                                                                 'links': [       {       'connected_time': '9885084',
                                                                                                                                          'fiber_length': 11053,
                                                                                                                                          'poller_hash': 'MzVGOUZEcGVlcnBvbnBvbiA1MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688525',
                                                                                                                                          'fiber_length': 12718,
                                                                                                                                          'poller_hash': 'NEY5MkNGcGVlcnBvbnBvbiA1MzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '67927947',
                                                                                                                                          'fiber_length': 10254,
                                                                                                                                          'poller_hash': 'MTFEMDE4cGVlcnBvbnBvbiA1NjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '23939405',
                                                                                                                                          'fiber_length': 12290,
                                                                                                                                          'poller_hash': 'MzIxNzhFcGVlcnBvbnBvbiA3MDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '21815946',
                                                                                                                                          'fiber_length': 11020,
                                                                                                                                          'poller_hash': 'MzIxNzIwcGVlcnBvbnBvbiA3MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '67790186',
                                                                                                                                          'fiber_length': 14054,
                                                                                                                                          'poller_hash': 'MzVGODQzcGVlcnBvbnBvbiA3Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '67790433',
                                                                                                                                          'fiber_length': 13813,
                                                                                                                                          'poller_hash': 'MzVFOTVCcGVlcnBvbnBvbiA3NDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '15499232',
                                                                                                                                          'fiber_length': 12485,
                                                                                                                                          'poller_hash': 'MzVFOTVDcGVlcnBvbnBvbiA3NTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '67790340',
                                                                                                                                          'fiber_length': 13308,
                                                                                                                                          'poller_hash': 'MzVFQTA4cGVlcnBvbnBvbiA3NTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '15448823',
                                                                                                                                          'fiber_length': 10435,
                                                                                                                                          'poller_hash': 'MzVFQTI4cGVlcnBvbnBvbiA3NTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '2609498',
                                                                                                                                          'fiber_length': 12716,
                                                                                                                                          'poller_hash': 'MzVFQTJDcGVlcnBvbnBvbiA3NjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688525',
                                                                                                                                          'fiber_length': 13539,
                                                                                                                                          'poller_hash': 'NDNCMkQ3cGVlcnBvbnBvbiA3NjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688532',
                                                                                                                                          'fiber_length': 10170,
                                                                                                                                          'poller_hash': 'MzIyODhDcGVlcnBvbnBvbiA3NjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '21816414',
                                                                                                                                          'fiber_length': 13404,
                                                                                                                                          'poller_hash': 'MzIyOENFcGVlcnBvbnBvbiA3NjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '4484892',
                                                                                                                                          'fiber_length': 10092,
                                                                                                                                          'poller_hash': 'NUE2NjMycGVlcnBvbnBvbiA3Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688527',
                                                                                                                                          'fiber_length': 10221,
                                                                                                                                          'poller_hash': 'NDNCNUI3cGVlcnBvbnBvbiA3Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '2351694',
                                                                                                                                          'fiber_length': 12566,
                                                                                                                                          'poller_hash': 'NUE2QTYwcGVlcnBvbnBvbiAxMzc4Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '4315678',
                                                                                                                                          'fiber_length': 10894,
                                                                                                                                          'poller_hash': 'ODJBNzQxcGVlcnBvbnBvbiAxNDE2NDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '21816203',
                                                                                                                                          'fiber_length': 10619,
                                                                                                                                          'poller_hash': 'ODY5NDBFcGVlcnBvbnBvbiAxNDE4OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '21321701',
                                                                                                                                          'fiber_length': 10542,
                                                                                                                                          'poller_hash': 'ODY5M0YxcGVlcnBvbnBvbiAxNDE5MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '9339790',
                                                                                                                                          'fiber_length': 10460,
                                                                                                                                          'poller_hash': 'ODY5M0MwcGVlcnBvbnBvbiAxNDIxMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '29301725',
                                                                                                                                          'fiber_length': 10515,
                                                                                                                                          'poller_hash': 'MTU4RkI2RXBlZXJwb25wb24gMjE1MTYx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688532',
                                                                                                                                          'fiber_length': 11528,
                                                                                                                                          'poller_hash': 'MTI3NTBEN3BlZXJwb25wb24gMjE3OTQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '665669',
                                                                                                                                          'fiber_length': 11543,
                                                                                                                                          'poller_hash': 'MTI3NTAzNnBlZXJwb25wb24gMjE3OTc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688535',
                                                                                                                                          'fiber_length': 10198,
                                                                                                                                          'poller_hash': 'MTI3NEE3OHBlZXJwb25wb24gMjMxNTEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '67911671',
                                                                                                                                          'fiber_length': 11636,
                                                                                                                                          'poller_hash': 'MTI3NDMzM3BlZXJwb25wb24gMjQ2OTkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688540',
                                                                                                                                          'fiber_length': 10160,
                                                                                                                                          'poller_hash': 'MTI3NDlCNHBlZXJwb25wb24gMjQ4NDEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '688465',
                                                                                                                                          'fiber_length': 10181,
                                                                                                                                          'poller_hash': 'MTI3NDczQXBlZXJwb25wb24gMjYzMDE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '17631129',
                                                                                                                                          'fiber_length': 11051,
                                                                                                                                          'poller_hash': 'MTI5NzQxMXBlZXJwb25wb24gMjY1MDMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756},
                                                                                                                                  {       'connected_time': '537662',
                                                                                                                                          'fiber_length': 10580,
                                                                                                                                          'poller_hash': 'MTI5NUZCMnBlZXJwb25wb24gMjgwMjQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.756}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/6',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '11/7',
                                                                                         'in_discards': '880',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '197470',
                                                                                         'in_octets': '1573458740198',
                                                                                         'in_ucast_pkts': '6091678938',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzIyODQ5cGVlcmludGZwb24gMjM5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '64261111878',
                                                                                                                  'in_ucast_pkts': '274085354',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '1575031711163',
                                                                                                                  'out_ucast_pkts': '1231201840',
                                                                                                                  'poller_hash': 'MzIyODQxcGVlcmludGZwb24gMjg1'},
                                                                                                          {       'in_nucast_pkts': '200061',
                                                                                                                  'in_octets': '36735276954',
                                                                                                                  'in_ucast_pkts': '250537158',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23313256',
                                                                                                                  'out_octets': '1310492361720',
                                                                                                                  'out_ucast_pkts': '940360750',
                                                                                                                  'poller_hash': 'MzIyNDIxcGVlcmludGZwb24gMjg2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '270993964124',
                                                                                                                  'in_ucast_pkts': '31434104065',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '86',
                                                                                                                  'out_octets': '5902432015925',
                                                                                                                  'out_ucast_pkts': '4628420006',
                                                                                                                  'poller_hash': 'MzIyNzNGcGVlcmludGZwb24gMzAx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '61211284511',
                                                                                                                  'in_ucast_pkts': '239435671',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '562769085520',
                                                                                                                  'out_ucast_pkts': '567000684',
                                                                                                                  'poller_hash': 'MzIyOUYycGVlcmludGZwb24gNDcw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJCRTlBcGVlcmludGZwb24gODEw'},
                                                                                                          {       'in_nucast_pkts': '65',
                                                                                                                  'in_octets': '175016373407',
                                                                                                                  'in_ucast_pkts': '451812546',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27288126',
                                                                                                                  'out_octets': '973375556760',
                                                                                                                  'out_ucast_pkts': '814317786',
                                                                                                                  'poller_hash': 'NUE2NUJFcGVlcmludGZwb24gODEx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'REZGQzMycGVlcmludGZwb24gMTgwNDkw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REI3NnBlZXJpbnRmcG9uIDI1NDI1OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDdFMXBlZXJpbnRmcG9uIDI2MjQ1Nw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '15147812968',
                                                                                         'out_octets': '45589733793651',
                                                                                         'out_ucast_pkts': '20076089342',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/7',
                                                                                                                 'id': '11/7',
                                                                                                                 'links': [       {       'connected_time': '21816609',
                                                                                                                                          'fiber_length': 5570,
                                                                                                                                          'poller_hash': 'MzIyODQ5cGVlcnBvbnBvbiAyMzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '21816409',
                                                                                                                                          'fiber_length': 6097,
                                                                                                                                          'poller_hash': 'MzIyODQxcGVlcnBvbnBvbiAyODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '10783334',
                                                                                                                                          'fiber_length': 5911,
                                                                                                                                          'poller_hash': 'MzIyNDIxcGVlcnBvbnBvbiAyODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '21816866',
                                                                                                                                          'fiber_length': 5933,
                                                                                                                                          'poller_hash': 'MzIyNzNGcGVlcnBvbnBvbiAzMDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '2060536',
                                                                                                                                          'fiber_length': 5975,
                                                                                                                                          'poller_hash': 'MzIyOUYycGVlcnBvbnBvbiA0NzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '6420495',
                                                                                                                                          'fiber_length': 6930,
                                                                                                                                          'poller_hash': 'NDJCRTlBcGVlcnBvbnBvbiA4MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '12619346',
                                                                                                                                          'fiber_length': 7171,
                                                                                                                                          'poller_hash': 'NUE2NUJFcGVlcnBvbnBvbiA4MTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '21816134',
                                                                                                                                          'fiber_length': 7260,
                                                                                                                                          'poller_hash': 'REZGQzMycGVlcnBvbnBvbiAxODA0OTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '23840230',
                                                                                                                                          'fiber_length': 6496,
                                                                                                                                          'poller_hash': 'MTE3REI3NnBlZXJwb25wb24gMjU0MjU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473},
                                                                                                                                  {       'connected_time': '21519489',
                                                                                                                                          'fiber_length': 6044,
                                                                                                                                          'poller_hash': 'MTI3NDdFMXBlZXJwb25wb24gMjYyNDU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.473}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/7',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '11/8',
                                                                                         'in_discards': '524329',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5075514',
                                                                                         'in_octets': '8888176980206',
                                                                                         'in_ucast_pkts': '32201961412',
                                                                                         'links': [       {       'in_nucast_pkts': '208',
                                                                                                                  'in_octets': '9934667130',
                                                                                                                  'in_ucast_pkts': '101392439',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120021',
                                                                                                                  'out_octets': '334279638785',
                                                                                                                  'out_ucast_pkts': '229953368',
                                                                                                                  'poller_hash': 'NEY5QjNDcGVlcmludGZwb24gMg=='},
                                                                                                          {       'in_nucast_pkts': '7032',
                                                                                                                  'in_octets': '269238786612',
                                                                                                                  'in_ucast_pkts': '910905115',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144999065',
                                                                                                                  'out_octets': '3217711317734',
                                                                                                                  'out_ucast_pkts': '2568367563',
                                                                                                                  'poller_hash': 'MzIxN0U2cGVlcmludGZwb24gNTY='},
                                                                                                          {       'in_nucast_pkts': '13924',
                                                                                                                  'in_octets': '1339517319',
                                                                                                                  'in_ucast_pkts': '3578697',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464273',
                                                                                                                  'out_octets': '20330314047',
                                                                                                                  'out_ucast_pkts': '16533886',
                                                                                                                  'poller_hash': 'MzIyQTQ3cGVlcmludGZwb24gNTE0'},
                                                                                                          {       'in_nucast_pkts': '635537',
                                                                                                                  'in_octets': '344261072226',
                                                                                                                  'in_ucast_pkts': '867688236',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83470597',
                                                                                                                  'out_octets': '3821345487448',
                                                                                                                  'out_ucast_pkts': '2940692213',
                                                                                                                  'poller_hash': 'MzVFOTREcGVlcmludGZwb24gODY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '147990754967',
                                                                                                                  'in_ucast_pkts': '134333307',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '111289611350',
                                                                                                                  'out_ucast_pkts': '176736797',
                                                                                                                  'poller_hash': 'ODY5M0MzcGVlcmludGZwb24gODc2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '53558053169',
                                                                                                                  'in_ucast_pkts': '4485374005',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '733332728608',
                                                                                                                  'out_ucast_pkts': '633870509',
                                                                                                                  'poller_hash': 'ODcyNEYzcGVlcmludGZwb24gOTA2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '110156565376',
                                                                                                                  'in_ucast_pkts': '504590052',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2486409596970',
                                                                                                                  'out_ucast_pkts': '1803377052',
                                                                                                                  'poller_hash': 'ODcyNTBFcGVlcmludGZwb24gMTM4MTcw'},
                                                                                                          {       'in_nucast_pkts': '13311',
                                                                                                                  'in_octets': '8880389846',
                                                                                                                  'in_ucast_pkts': '17490778',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464547',
                                                                                                                  'out_octets': '65461113173',
                                                                                                                  'out_ucast_pkts': '56886157',
                                                                                                                  'poller_hash': 'ODY5NDA5cGVlcmludGZwb24gMTQxODM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '37049525285',
                                                                                                                  'in_ucast_pkts': '106428875',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '86',
                                                                                                                  'out_octets': '289110741946',
                                                                                                                  'out_ucast_pkts': '291908107',
                                                                                                                  'poller_hash': 'ODY5NDkycGVlcmludGZwb24gMTQxODQ3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '333337637261',
                                                                                                                  'in_ucast_pkts': '2422220358',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '56',
                                                                                                                  'out_octets': '7654755595324',
                                                                                                                  'out_ucast_pkts': '5393438052',
                                                                                                                  'poller_hash': 'ODY5NDAycGVlcmludGZwb24gMTQyMDY4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '86128167253',
                                                                                                                  'in_ucast_pkts': '321919856',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '86',
                                                                                                                  'out_octets': '1494857790813',
                                                                                                                  'out_ucast_pkts': '1223295203',
                                                                                                                  'poller_hash': 'ODY5M0JDcGVlcmludGZwb24gMTQyMDg4'},
                                                                                                          {       'in_nucast_pkts': '247805',
                                                                                                                  'in_octets': '46872408502',
                                                                                                                  'in_ucast_pkts': '167420932',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28505071',
                                                                                                                  'out_octets': '1102800800037',
                                                                                                                  'out_ucast_pkts': '851701657',
                                                                                                                  'poller_hash': 'ODcyNTRCcGVlcmludGZwb24gMTQzODIy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '53613237579',
                                                                                                                  'in_ucast_pkts': '213228663',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '86',
                                                                                                                  'out_octets': '579115178391',
                                                                                                                  'out_ucast_pkts': '537030323',
                                                                                                                  'poller_hash': 'QTlDOEI0cGVlcmludGZwb24gMTQ3NDU2'},
                                                                                                          {       'in_nucast_pkts': '29',
                                                                                                                  'in_octets': '182641581597',
                                                                                                                  'in_ucast_pkts': '290508802',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9363913',
                                                                                                                  'out_octets': '672999506191',
                                                                                                                  'out_ucast_pkts': '586668026',
                                                                                                                  'poller_hash': 'QjUxMTc4cGVlcmludGZwb24gMTUwNTM5'},
                                                                                                          {       'in_nucast_pkts': '664073',
                                                                                                                  'in_octets': '206772621156',
                                                                                                                  'in_ucast_pkts': '729848416',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83519480',
                                                                                                                  'out_octets': '2296784527582',
                                                                                                                  'out_ucast_pkts': '1839693903',
                                                                                                                  'poller_hash': 'QjUxMUI5cGVlcmludGZwb24gMTUyMzAz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '63983738921',
                                                                                                                  'in_ucast_pkts': '276970456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '1131945570027',
                                                                                                                  'out_ucast_pkts': '934494007',
                                                                                                                  'poller_hash': 'QzgxMjI3cGVlcmludGZwb24gMTU2OTAx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '106818989799',
                                                                                                                  'in_ucast_pkts': '395029717',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6',
                                                                                                                  'out_octets': '1552726919848',
                                                                                                                  'out_ucast_pkts': '1192973000',
                                                                                                                  'poller_hash': 'QzgzOTU0cGVlcmludGZwb24gMTY0NzA1'},
                                                                                                          {       'in_nucast_pkts': '1485',
                                                                                                                  'in_octets': '39213265652',
                                                                                                                  'in_ucast_pkts': '90035054',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4729068',
                                                                                                                  'out_octets': '204556272886',
                                                                                                                  'out_ucast_pkts': '167804813',
                                                                                                                  'poller_hash': 'MTI5NzM4RXBlZXJpbnRmcG9uIDE2NTQ3OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '27193229648',
                                                                                                                  'in_ucast_pkts': '92789518',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '60',
                                                                                                                  'out_octets': '147982194706',
                                                                                                                  'out_ucast_pkts': '192306438',
                                                                                                                  'poller_hash': 'Q0JEMjdFcGVlcmludGZwb24gMTY4NTc0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEE3NnBlZXJpbnRmcG9uIDIzMzY3Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '202154734278',
                                                                                                                  'in_ucast_pkts': '39515599856',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '142',
                                                                                                                  'out_octets': '2123580409622',
                                                                                                                  'out_ucast_pkts': '1748743966',
                                                                                                                  'poller_hash': 'MTI3NEE1RXBlZXJpbnRmcG9uIDIzMzc1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEEyRXBlZXJpbnRmcG9uIDIzNDUyNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDFCRnBlZXJpbnRmcG9uIDI0MDMxMQ=='},
                                                                                                          {       'in_nucast_pkts': '11454',
                                                                                                                  'in_octets': '25922705432',
                                                                                                                  'in_ucast_pkts': '102608108',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465439',
                                                                                                                  'out_octets': '547607439196',
                                                                                                                  'out_ucast_pkts': '413589614',
                                                                                                                  'poller_hash': 'MTI3NDFDRHBlZXJpbnRmcG9uIDI0MjE3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '68884583216',
                                                                                                                  'in_ucast_pkts': '367071430',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2217911604078',
                                                                                                                  'out_ucast_pkts': '1553218206',
                                                                                                                  'poller_hash': 'MTI5NzREOXBlZXJpbnRmcG9uIDI1Njc4Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDZDQXBlZXJpbnRmcG9uIDI2MTQ1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDg2NXBlZXJpbnRmcG9uIDI3NDczNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '5593965093',
                                                                                                                  'in_ucast_pkts': '32504607',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '145918397322',
                                                                                                                  'out_ucast_pkts': '108749916',
                                                                                                                  'poller_hash': 'MTI3NDlDN3BlZXJpbnRmcG9uIDI4NDU1Mg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '11/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '168059310',
                                                                                         'out_octets': '112368248205227',
                                                                                         'out_ucast_pkts': '85674596479',
                                                                                         'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '11/8',
                                                                                                                 'id': '11/8',
                                                                                                                 'links': [       {       'connected_time': '688527',
                                                                                                                                          'fiber_length': 5646,
                                                                                                                                          'poller_hash': 'NEY5QjNDcGVlcnBvbnBvbiAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '21816603',
                                                                                                                                          'fiber_length': 6346,
                                                                                                                                          'poller_hash': 'MzIxN0U2cGVlcnBvbnBvbiA1Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '688547',
                                                                                                                                          'fiber_length': 6301,
                                                                                                                                          'poller_hash': 'MzIyQTQ3cGVlcnBvbnBvbiA1MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '67790417',
                                                                                                                                          'fiber_length': 5962,
                                                                                                                                          'poller_hash': 'MzVFOTREcGVlcnBvbnBvbiA4NjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '519881',
                                                                                                                                          'fiber_length': 5916,
                                                                                                                                          'poller_hash': 'ODY5M0MzcGVlcnBvbnBvbiA4NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '22383405',
                                                                                                                                          'fiber_length': 7935,
                                                                                                                                          'poller_hash': 'ODcyNEYzcGVlcnBvbnBvbiA5MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '688537',
                                                                                                                                          'fiber_length': 7113,
                                                                                                                                          'poller_hash': 'ODcyNTBFcGVlcnBvbnBvbiAxMzgxNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '688570',
                                                                                                                                          'fiber_length': 8130,
                                                                                                                                          'poller_hash': 'ODY5NDA5cGVlcnBvbnBvbiAxNDE4Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '21816220',
                                                                                                                                          'fiber_length': 8233,
                                                                                                                                          'poller_hash': 'ODY5NDkycGVlcnBvbnBvbiAxNDE4NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '30937693',
                                                                                                                                          'fiber_length': 8775,
                                                                                                                                          'poller_hash': 'ODY5NDAycGVlcnBvbnBvbiAxNDIwNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '28537793',
                                                                                                                                          'fiber_length': 8610,
                                                                                                                                          'poller_hash': 'ODY5M0JDcGVlcnBvbnBvbiAxNDIwODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '13219518',
                                                                                                                                          'fiber_length': 8085,
                                                                                                                                          'poller_hash': 'ODcyNTRCcGVlcnBvbnBvbiAxNDM4MjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '21816228',
                                                                                                                                          'fiber_length': 5679,
                                                                                                                                          'poller_hash': 'QTlDOEI0cGVlcnBvbnBvbiAxNDc0NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '4315564',
                                                                                                                                          'fiber_length': 7163,
                                                                                                                                          'poller_hash': 'QjUxMTc4cGVlcnBvbnBvbiAxNTA1Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '67790855',
                                                                                                                                          'fiber_length': 7486,
                                                                                                                                          'poller_hash': 'QjUxMUI5cGVlcnBvbnBvbiAxNTIzMDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '29301859',
                                                                                                                                          'fiber_length': 6489,
                                                                                                                                          'poller_hash': 'QzgxMjI3cGVlcnBvbnBvbiAxNTY5MDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '688542',
                                                                                                                                          'fiber_length': 7517,
                                                                                                                                          'poller_hash': 'QzgzOTU0cGVlcnBvbnBvbiAxNjQ3MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '2181353',
                                                                                                                                          'fiber_length': 7092,
                                                                                                                                          'poller_hash': 'MTI5NzM4RXBlZXJwb25wb24gMTY1NDc4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '68164046',
                                                                                                                                          'fiber_length': 8769,
                                                                                                                                          'poller_hash': 'Q0JEMjdFcGVlcnBvbnBvbiAxNjg1NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '4761198',
                                                                                                                                          'fiber_length': 9089,
                                                                                                                                          'poller_hash': 'MTI3NEE3NnBlZXJwb25wb24gMjMzNjcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '25491295',
                                                                                                                                          'fiber_length': 5732,
                                                                                                                                          'poller_hash': 'MTI3NEE1RXBlZXJwb25wb24gMjMzNzU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '3838593',
                                                                                                                                          'fiber_length': 7924,
                                                                                                                                          'poller_hash': 'MTI3NEEyRXBlZXJwb25wb24gMjM0NTI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '72668166',
                                                                                                                                          'fiber_length': 7231,
                                                                                                                                          'poller_hash': 'MTI3NDFCRnBlZXJwb25wb24gMjQwMzEx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '688574',
                                                                                                                                          'fiber_length': 6318,
                                                                                                                                          'poller_hash': 'MTI3NDFDRHBlZXJwb25wb24gMjQyMTc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '688539',
                                                                                                                                          'fiber_length': 6645,
                                                                                                                                          'poller_hash': 'MTI5NzREOXBlZXJwb25wb24gMjU2Nzg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '20148339',
                                                                                                                                          'fiber_length': 6689,
                                                                                                                                          'poller_hash': 'MTI3NDZDQXBlZXJwb25wb24gMjYxNDU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '8236946',
                                                                                                                                          'fiber_length': 5998,
                                                                                                                                          'poller_hash': 'MTI3NDg2NXBlZXJwb25wb24gMjc0NzM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73},
                                                                                                                                  {       'connected_time': '973050',
                                                                                                                                          'fiber_length': 6444,
                                                                                                                                          'poller_hash': 'MTI3NDlDN3BlZXJwb25wb24gMjg0NTUy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.73}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '11/8',
                                                                                                                 'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNDEwNjAxMzM0',
                                                           'serial_number': '211410601334',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991823'},
                                                   {       'id': '12',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '12/1',
                                                                                         'in_discards': '772275',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3411856',
                                                                                         'in_octets': '16596079062701',
                                                                                         'in_ucast_pkts': '61005533675',
                                                                                         'links': [       {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '7680599744',
                                                                                                                  'in_ucast_pkts': '46407311',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120471',
                                                                                                                  'out_octets': '150089357437',
                                                                                                                  'out_ucast_pkts': '115315392',
                                                                                                                  'poller_hash': 'NUE2QUVFcGVlcmludGZwb24gNzUx'},
                                                                                                          {       'in_nucast_pkts': '164',
                                                                                                                  'in_octets': '1391886037',
                                                                                                                  'in_ucast_pkts': '9252028',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '851191',
                                                                                                                  'out_octets': '30526009516',
                                                                                                                  'out_ucast_pkts': '22514538',
                                                                                                                  'poller_hash': 'NDNCMzhFcGVlcmludGZwb24gNzcz'},
                                                                                                          {       'in_nucast_pkts': '1333',
                                                                                                                  'in_octets': '5956758351',
                                                                                                                  'in_ucast_pkts': '55796253',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1476741',
                                                                                                                  'out_octets': '251705967841',
                                                                                                                  'out_ucast_pkts': '176675975',
                                                                                                                  'poller_hash': 'NUE2NjY4cGVlcmludGZwb24gNzc0'},
                                                                                                          {       'in_nucast_pkts': '858',
                                                                                                                  'in_octets': '84155512090',
                                                                                                                  'in_ucast_pkts': '335919033',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '99468244',
                                                                                                                  'out_octets': '1515296121291',
                                                                                                                  'out_ucast_pkts': '1158290962',
                                                                                                                  'poller_hash': 'NDJDRTRBcGVlcmludGZwb24gNzc1'},
                                                                                                          {       'in_nucast_pkts': '22594',
                                                                                                                  'in_octets': '186524256471',
                                                                                                                  'in_ucast_pkts': '602731047',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76676668',
                                                                                                                  'out_octets': '2043583326467',
                                                                                                                  'out_ucast_pkts': '1601911413',
                                                                                                                  'poller_hash': 'NUE2QjlBcGVlcmludGZwb24gNzc2'},
                                                                                                          {       'in_nucast_pkts': '20305381',
                                                                                                                  'in_octets': '95427504245',
                                                                                                                  'in_ucast_pkts': '1064081215',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76687687',
                                                                                                                  'out_octets': '3130961364770',
                                                                                                                  'out_ucast_pkts': '2123112404',
                                                                                                                  'poller_hash': 'NUE2QkNEcGVlcmludGZwb24gNzc4'},
                                                                                                          {       'in_nucast_pkts': '512',
                                                                                                                  'in_octets': '24333061128',
                                                                                                                  'in_ucast_pkts': '72894560',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1119864',
                                                                                                                  'out_octets': '234621537823',
                                                                                                                  'out_ucast_pkts': '184797375',
                                                                                                                  'poller_hash': 'NUE2QjhGcGVlcmludGZwb24gNzgw'},
                                                                                                          {       'in_nucast_pkts': '74',
                                                                                                                  'in_octets': '881413934',
                                                                                                                  'in_ucast_pkts': '2546810',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '109008',
                                                                                                                  'out_octets': '5639145061',
                                                                                                                  'out_ucast_pkts': '4379500',
                                                                                                                  'poller_hash': 'NjQwNzlBcGVlcmludGZwb24gNzgz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '189147273627',
                                                                                                                  'in_ucast_pkts': '26357816456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '90',
                                                                                                                  'out_octets': '2502180061617',
                                                                                                                  'out_ucast_pkts': '1949928457',
                                                                                                                  'poller_hash': 'NDJDRERDcGVlcmludGZwb24gNzkw'},
                                                                                                          {       'in_nucast_pkts': '2652',
                                                                                                                  'in_octets': '12329124854',
                                                                                                                  'in_ucast_pkts': '26442156',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '834603',
                                                                                                                  'out_octets': '81353254286',
                                                                                                                  'out_ucast_pkts': '67212033',
                                                                                                                  'poller_hash': 'NjQwQjJEcGVlcmludGZwb24gODA0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '649893956984',
                                                                                                                  'in_ucast_pkts': '17755138781',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '104',
                                                                                                                  'out_octets': '424171721774',
                                                                                                                  'out_ucast_pkts': '576997206',
                                                                                                                  'poller_hash': 'ODY5NDEwcGVlcmludGZwb24gODA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '110529187071',
                                                                                                                  'in_ucast_pkts': '4624212648',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '114',
                                                                                                                  'out_octets': '1204164832424',
                                                                                                                  'out_ucast_pkts': '1052101752',
                                                                                                                  'poller_hash': 'NDJDNzBCcGVlcmludGZwb24gODE0'},
                                                                                                          {       'in_nucast_pkts': '13',
                                                                                                                  'in_octets': '2875663724',
                                                                                                                  'in_ucast_pkts': '23291485',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120259',
                                                                                                                  'out_octets': '91518928839',
                                                                                                                  'out_ucast_pkts': '69675094',
                                                                                                                  'poller_hash': 'NDJDRDY2cGVlcmludGZwb24gODMy'},
                                                                                                          {       'in_nucast_pkts': '164',
                                                                                                                  'in_octets': '340964536',
                                                                                                                  'in_ucast_pkts': '1228300',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120490',
                                                                                                                  'out_octets': '3443800972',
                                                                                                                  'out_ucast_pkts': '2640091',
                                                                                                                  'poller_hash': 'NUE2NTkxcGVlcmludGZwb24gODM2'},
                                                                                                          {       'in_nucast_pkts': '21537',
                                                                                                                  'in_octets': '336096799412',
                                                                                                                  'in_ucast_pkts': '862311156',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22172600784',
                                                                                                                  'out_octets': '33754749343074',
                                                                                                                  'out_ucast_pkts': '3167656191',
                                                                                                                  'poller_hash': 'NDJCRjA4cGVlcmludGZwb24gMTM3NTIw'},
                                                                                                          {       'in_nucast_pkts': '1910',
                                                                                                                  'in_octets': '1247863904686',
                                                                                                                  'in_ucast_pkts': '6143498698',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76640852',
                                                                                                                  'out_octets': '21625634624962',
                                                                                                                  'out_ucast_pkts': '15250567380',
                                                                                                                  'poller_hash': 'NUE2QzhBcGVlcmludGZwb24gMTM4Njcy'},
                                                                                                          {       'in_nucast_pkts': '11811',
                                                                                                                  'in_octets': '6431672595',
                                                                                                                  'in_ucast_pkts': '19437957',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120417',
                                                                                                                  'out_octets': '20654366355',
                                                                                                                  'out_ucast_pkts': '24903413',
                                                                                                                  'poller_hash': 'ODQ0QkM3cGVlcmludGZwb24gMTQwMDQ5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '62335055385',
                                                                                                                  'in_ucast_pkts': '299330710',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1134645597345',
                                                                                                                  'out_ucast_pkts': '852185851',
                                                                                                                  'poller_hash': 'ODY5MjcxcGVlcmludGZwb24gMTQxMjAx'},
                                                                                                          {       'in_nucast_pkts': '4445503',
                                                                                                                  'in_octets': '600608554298',
                                                                                                                  'in_ucast_pkts': '2516191404',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '73763124',
                                                                                                                  'out_octets': '9077631950006',
                                                                                                                  'out_ucast_pkts': '6727140046',
                                                                                                                  'poller_hash': 'ODQ0QkVFcGVlcmludGZwb24gMTQxMjU4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MkZFQTFEcGVlcmludGZwb24gMTQzMTc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '190314426964',
                                                                                                                  'in_ucast_pkts': '22497905094',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '113',
                                                                                                                  'out_octets': '4493122957656',
                                                                                                                  'out_ucast_pkts': '3247798587',
                                                                                                                  'poller_hash': 'ODcyNTBBcGVlcmludGZwb24gMTQzNzYy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '51640387311',
                                                                                                                  'in_ucast_pkts': '192479159',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '137',
                                                                                                                  'out_octets': '662396714627',
                                                                                                                  'out_ucast_pkts': '566775004',
                                                                                                                  'poller_hash': 'QTlDOTRCcGVlcmludGZwb24gMTQ5MTA0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '312016351844',
                                                                                                                  'in_ucast_pkts': '1400819304',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '5818661704564',
                                                                                                                  'out_ucast_pkts': '4240295166',
                                                                                                                  'poller_hash': 'QjBBMjg4cGVlcmludGZwb24gMTUxNjEw'},
                                                                                                          {       'in_nucast_pkts': '910909',
                                                                                                                  'in_octets': '428639497060',
                                                                                                                  'in_ucast_pkts': '2858652966',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123018310',
                                                                                                                  'out_octets': '16645566496384',
                                                                                                                  'out_ucast_pkts': '11609563546',
                                                                                                                  'poller_hash': 'QzgzMEQ0cGVlcmludGZwb24gMTcyNDE0'},
                                                                                                          {       'in_nucast_pkts': '11499',
                                                                                                                  'in_octets': '10301949649',
                                                                                                                  'in_ucast_pkts': '67349484',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120757',
                                                                                                                  'out_octets': '220684922901',
                                                                                                                  'out_ucast_pkts': '176937005',
                                                                                                                  'poller_hash': 'MTE5NTE0NHBlZXJpbnRmcG9uIDIxMzY1MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '413125254761',
                                                                                                                  'in_ucast_pkts': '104454674996',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66',
                                                                                                                  'out_octets': '5217599652920',
                                                                                                                  'out_ucast_pkts': '4114985580',
                                                                                                                  'poller_hash': 'MTE5NTE0M3BlZXJpbnRmcG9uIDIxNTk1OQ=='},
                                                                                                          {       'in_nucast_pkts': '210472',
                                                                                                                  'in_octets': '276158214036',
                                                                                                                  'in_ucast_pkts': '727050033',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '24505549',
                                                                                                                  'out_octets': '2994270542936',
                                                                                                                  'out_ucast_pkts': '2341710240',
                                                                                                                  'poller_hash': 'MTI3NEE3RXBlZXJpbnRmcG9uIDIyNDY4OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEE3NXBlZXJpbnRmcG9uIDIzNjEyMA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '12/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '107370415453',
                                                                                         'out_octets': '360906811457768',
                                                                                         'out_ucast_pkts': '161784571431',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/1',
                                                                                                                 'id': '12/1',
                                                                                                                 'links': [       {       'connected_time': '538559',
                                                                                                                                          'fiber_length': 11153,
                                                                                                                                          'poller_hash': 'NUE2QUVFcGVlcnBvbnBvbiA3NTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538541',
                                                                                                                                          'fiber_length': 11191,
                                                                                                                                          'poller_hash': 'NDNCMzhFcGVlcnBvbnBvbiA3NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '709215',
                                                                                                                                          'fiber_length': 12027,
                                                                                                                                          'poller_hash': 'NUE2NjY4cGVlcnBvbnBvbiA3NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '4834258',
                                                                                                                                          'fiber_length': 12543,
                                                                                                                                          'poller_hash': 'NDJDRTRBcGVlcnBvbnBvbiA3NzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '67790120',
                                                                                                                                          'fiber_length': 11876,
                                                                                                                                          'poller_hash': 'NUE2QjlBcGVlcnBvbnBvbiA3NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '67790318',
                                                                                                                                          'fiber_length': 11201,
                                                                                                                                          'poller_hash': 'NUE2QkNEcGVlcnBvbnBvbiA3Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538558',
                                                                                                                                          'fiber_length': 13591,
                                                                                                                                          'poller_hash': 'NUE2QjhGcGVlcnBvbnBvbiA3ODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '69795',
                                                                                                                                          'fiber_length': 13217,
                                                                                                                                          'poller_hash': 'NjQwNzlBcGVlcnBvbnBvbiA3ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '21816141',
                                                                                                                                          'fiber_length': 13955,
                                                                                                                                          'poller_hash': 'NDJDRERDcGVlcnBvbnBvbiA3OTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538547',
                                                                                                                                          'fiber_length': 13247,
                                                                                                                                          'poller_hash': 'NjQwQjJEcGVlcnBvbnBvbiA4MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '21816439',
                                                                                                                                          'fiber_length': 13514,
                                                                                                                                          'poller_hash': 'ODY5NDEwcGVlcnBvbnBvbiA4MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '11412210',
                                                                                                                                          'fiber_length': 14700,
                                                                                                                                          'poller_hash': 'NDJDNzBCcGVlcnBvbnBvbiA4MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538547',
                                                                                                                                          'fiber_length': 13955,
                                                                                                                                          'poller_hash': 'NDJDRDY2cGVlcnBvbnBvbiA4MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538565',
                                                                                                                                          'fiber_length': 14735,
                                                                                                                                          'poller_hash': 'NUE2NTkxcGVlcnBvbnBvbiA4MzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '6572792',
                                                                                                                                          'fiber_length': 12379,
                                                                                                                                          'poller_hash': 'NDJCRjA4cGVlcnBvbnBvbiAxMzc1MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '67790303',
                                                                                                                                          'fiber_length': 11362,
                                                                                                                                          'poller_hash': 'NUE2QzhBcGVlcnBvbnBvbiAxMzg2NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538569',
                                                                                                                                          'fiber_length': 13301,
                                                                                                                                          'poller_hash': 'ODQ0QkM3cGVlcnBvbnBvbiAxNDAwNDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538549',
                                                                                                                                          'fiber_length': 13962,
                                                                                                                                          'poller_hash': 'ODY5MjcxcGVlcnBvbnBvbiAxNDEyMDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '67790661',
                                                                                                                                          'fiber_length': 14418,
                                                                                                                                          'poller_hash': 'ODQ0QkVFcGVlcnBvbnBvbiAxNDEyNTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538549',
                                                                                                                                          'fiber_length': 13000,
                                                                                                                                          'poller_hash': 'MkZFQTFEcGVlcnBvbnBvbiAxNDMxNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '9218555',
                                                                                                                                          'fiber_length': 14492,
                                                                                                                                          'poller_hash': 'ODcyNTBBcGVlcnBvbnBvbiAxNDM3NjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '21816297',
                                                                                                                                          'fiber_length': 13761,
                                                                                                                                          'poller_hash': 'QTlDOTRCcGVlcnBvbnBvbiAxNDkxMDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538555',
                                                                                                                                          'fiber_length': 18387,
                                                                                                                                          'poller_hash': 'QjBBMjg4cGVlcnBvbnBvbiAxNTE2MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '21816015',
                                                                                                                                          'fiber_length': 18774,
                                                                                                                                          'poller_hash': 'QzgzMEQ0cGVlcnBvbnBvbiAxNzI0MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '538562',
                                                                                                                                          'fiber_length': 14960,
                                                                                                                                          'poller_hash': 'MTE5NTE0NHBlZXJwb25wb24gMjEzNjUx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '2065619',
                                                                                                                                          'fiber_length': 13566,
                                                                                                                                          'poller_hash': 'MTE5NTE0M3BlZXJwb25wb24gMjE1OTU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '12865046',
                                                                                                                                          'fiber_length': 15029,
                                                                                                                                          'poller_hash': 'MTI3NEE3RXBlZXJwb25wb24gMjI0Njg5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668},
                                                                                                                                  {       'connected_time': '67050083',
                                                                                                                                          'fiber_length': 11631,
                                                                                                                                          'poller_hash': 'MTI3NEE3NXBlZXJwb25wb24gMjM2MTIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.668}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/1',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '12/2',
                                                                                         'in_discards': '73491',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '17259237',
                                                                                         'in_octets': '15084907031982',
                                                                                         'in_ucast_pkts': '59353390581',
                                                                                         'links': [       {       'in_nucast_pkts': '633602',
                                                                                                                  'in_octets': '286672742607',
                                                                                                                  'in_ucast_pkts': '1901269865',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71111847',
                                                                                                                  'out_octets': '8518861287690',
                                                                                                                  'out_ucast_pkts': '5911051380',
                                                                                                                  'poller_hash': 'NDJEMTU0cGVlcmludGZwb24gNTY5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTEyNUU0cGVlcmludGZwb24gNTcw'},
                                                                                                          {       'in_nucast_pkts': '78',
                                                                                                                  'in_octets': '1516029525',
                                                                                                                  'in_ucast_pkts': '6066078',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1660372',
                                                                                                                  'out_octets': '26355541883',
                                                                                                                  'out_ucast_pkts': '21119750',
                                                                                                                  'poller_hash': 'NTEyNTIzcGVlcmludGZwb24gNTcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTEyNUYwcGVlcmludGZwb24gNTcy'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '3677025873',
                                                                                                                  'in_ucast_pkts': '11693748',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120545',
                                                                                                                  'out_octets': '39455457522',
                                                                                                                  'out_ucast_pkts': '29691334',
                                                                                                                  'poller_hash': 'NDJDRTMycGVlcmludGZwb24gNTc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QjUwREM0cGVlcmludGZwb24gNjIy'},
                                                                                                          {       'in_nucast_pkts': '935033',
                                                                                                                  'in_octets': '284378211103',
                                                                                                                  'in_ucast_pkts': '870282862',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '123020404',
                                                                                                                  'out_octets': '4649039524098',
                                                                                                                  'out_ucast_pkts': '3507246475',
                                                                                                                  'poller_hash': 'NDJEMkFGcGVlcmludGZwb24gNjI5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTEyNTc1cGVlcmludGZwb24gNjk1'},
                                                                                                          {       'in_nucast_pkts': '12897',
                                                                                                                  'in_octets': '1022751382258',
                                                                                                                  'in_ucast_pkts': '1459357976',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122773309',
                                                                                                                  'out_octets': '2510490403929',
                                                                                                                  'out_ucast_pkts': '2415910591',
                                                                                                                  'poller_hash': 'NDJEMTk2cGVlcmludGZwb24gNjk2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJDMzhFcGVlcmludGZwb24gNjk4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '560583097788',
                                                                                                                  'in_ucast_pkts': '12830316745',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '22662435985101',
                                                                                                                  'out_ucast_pkts': '15751334406',
                                                                                                                  'poller_hash': 'NDJDMzhDcGVlcmludGZwb24gNjk5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '190496346784',
                                                                                                                  'in_ucast_pkts': '26947469307',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '137',
                                                                                                                  'out_octets': '3936401747777',
                                                                                                                  'out_ucast_pkts': '2845053537',
                                                                                                                  'poller_hash': 'NDJDMkQzcGVlcmludGZwb24gNzAw'},
                                                                                                          {       'in_nucast_pkts': '217',
                                                                                                                  'in_octets': '1594209131',
                                                                                                                  'in_ucast_pkts': '5685898',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120465',
                                                                                                                  'out_octets': '28201207249',
                                                                                                                  'out_ucast_pkts': '22146071',
                                                                                                                  'poller_hash': 'NUE2QkIycGVlcmludGZwb24gNzQx'},
                                                                                                          {       'in_nucast_pkts': '70371',
                                                                                                                  'in_octets': '47219009079',
                                                                                                                  'in_ucast_pkts': '60765529',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120473',
                                                                                                                  'out_octets': '128792266581',
                                                                                                                  'out_ucast_pkts': '100239284',
                                                                                                                  'poller_hash': 'NUE2QjgwcGVlcmludGZwb24gNzQy'},
                                                                                                          {       'in_nucast_pkts': '19',
                                                                                                                  'in_octets': '13311432235',
                                                                                                                  'in_ucast_pkts': '17483547',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '851239',
                                                                                                                  'out_octets': '61576493114',
                                                                                                                  'out_ucast_pkts': '46473240',
                                                                                                                  'poller_hash': 'NDNCM0I2cGVlcmludGZwb24gNzQ1'},
                                                                                                          {       'in_nucast_pkts': '45',
                                                                                                                  'in_octets': '2934199229',
                                                                                                                  'in_ucast_pkts': '13708560',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120474',
                                                                                                                  'out_octets': '51980907290',
                                                                                                                  'out_ucast_pkts': '39973066',
                                                                                                                  'poller_hash': 'NTEyNEU2cGVlcmludGZwb24gNzU1'},
                                                                                                          {       'in_nucast_pkts': '9663',
                                                                                                                  'in_octets': '19938757404',
                                                                                                                  'in_ucast_pkts': '36601494',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '920523',
                                                                                                                  'out_octets': '73470385616',
                                                                                                                  'out_ucast_pkts': '63457636',
                                                                                                                  'poller_hash': 'NUE2NUYycGVlcmludGZwb24gNzg0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '383778839488',
                                                                                                                  'in_ucast_pkts': '87432498400',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '113',
                                                                                                                  'out_octets': '10374964885076',
                                                                                                                  'out_ucast_pkts': '7246463853',
                                                                                                                  'poller_hash': 'NDJEMDkycGVlcmludGZwb24gNzkx'},
                                                                                                          {       'in_nucast_pkts': '33101068',
                                                                                                                  'in_octets': '83229599910',
                                                                                                                  'in_ucast_pkts': '770571419',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '74397042',
                                                                                                                  'out_octets': '2246149415031',
                                                                                                                  'out_ucast_pkts': '1568240044',
                                                                                                                  'poller_hash': 'NUE2NkM4cGVlcmludGZwb24gODE2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '232036',
                                                                                                                  'in_ucast_pkts': '793',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3410',
                                                                                                                  'out_octets': '1743887',
                                                                                                                  'out_ucast_pkts': '1518',
                                                                                                                  'poller_hash': 'MTI3NEE1MXBlZXJpbnRmcG9uIDg0NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '80483512674',
                                                                                                                  'in_ucast_pkts': '712907911',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '4307485090650',
                                                                                                                  'out_ucast_pkts': '2956760449',
                                                                                                                  'poller_hash': 'NDJEMUNFcGVlcmludGZwb24gODQ5'},
                                                                                                          {       'in_nucast_pkts': '6838',
                                                                                                                  'in_octets': '131951369874',
                                                                                                                  'in_ucast_pkts': '402438034',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76680156',
                                                                                                                  'out_octets': '1387949203861',
                                                                                                                  'out_ucast_pkts': '1090256969',
                                                                                                                  'poller_hash': 'NUE2NjE2cGVlcmludGZwb24gODUw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NUE2NjYxcGVlcmludGZwb24gODUx'},
                                                                                                          {       'in_nucast_pkts': '1681',
                                                                                                                  'in_octets': '7628622379',
                                                                                                                  'in_ucast_pkts': '45750066',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3685300',
                                                                                                                  'out_octets': '238096183433',
                                                                                                                  'out_ucast_pkts': '178897190',
                                                                                                                  'poller_hash': 'NUE2Qzg5cGVlcmludGZwb24gODcy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '373492844701',
                                                                                                                  'in_ucast_pkts': '26779468569',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '4816523333672',
                                                                                                                  'out_ucast_pkts': '3534687298',
                                                                                                                  'poller_hash': 'OTU2RTU4cGVlcmludGZwb24gOTA4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '258728136432',
                                                                                                                  'in_ucast_pkts': '978971078',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '3118605408463',
                                                                                                                  'out_ucast_pkts': '2277587403',
                                                                                                                  'poller_hash': 'ODY5M0NFcGVlcmludGZwb24gMTQxOTk3'},
                                                                                                          {       'in_nucast_pkts': '17',
                                                                                                                  'in_octets': '60226344667',
                                                                                                                  'in_ucast_pkts': '61716774',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1120669',
                                                                                                                  'out_octets': '117522699364',
                                                                                                                  'out_ucast_pkts': '99951171',
                                                                                                                  'poller_hash': 'ODcyNERFcGVlcmludGZwb24gMTQzMjM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '483385924539',
                                                                                                                  'in_ucast_pkts': '3766519016',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '81',
                                                                                                                  'out_octets': '12035486252307',
                                                                                                                  'out_ucast_pkts': '8682905663',
                                                                                                                  'poller_hash': 'QzgzQjY0cGVlcmludGZwb24gMTY2NjA4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '84374577947',
                                                                                                                  'in_ucast_pkts': '475568303',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '2309581282318',
                                                                                                                  'out_ucast_pkts': '1748389184',
                                                                                                                  'poller_hash': 'QzgxRTc5cGVlcmludGZwb24gMjA3MTc5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '391249884105',
                                                                                                                  'in_ucast_pkts': '1909279673',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '69',
                                                                                                                  'out_octets': '9686111166412',
                                                                                                                  'out_ucast_pkts': '6810406407',
                                                                                                                  'poller_hash': 'MTI3NTE5N3BlZXJpbnRmcG9uIDIyODcxMg=='},
                                                                                                          {       'in_nucast_pkts': '330',
                                                                                                                  'in_octets': '46027000764',
                                                                                                                  'in_ucast_pkts': '542274885',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28821846',
                                                                                                                  'out_octets': '2298904822333',
                                                                                                                  'out_ucast_pkts': '1537610439',
                                                                                                                  'poller_hash': 'MTI3NDBERHBlZXJpbnRmcG9uIDI0ODYzNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '38935389285',
                                                                                                                  'in_ucast_pkts': '110165657',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '421806032930',
                                                                                                                  'out_ucast_pkts': '347926572',
                                                                                                                  'poller_hash': 'MTI3NDdGRnBlZXJpbnRmcG9uIDI3NDU5Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '12/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '97558217707',
                                                                                         'out_octets': '379073954446271',
                                                                                         'out_ucast_pkts': '178548957703',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/2',
                                                                                                                 'id': '12/2',
                                                                                                                 'links': [       {       'connected_time': '65348135',
                                                                                                                                          'fiber_length': 17984,
                                                                                                                                          'poller_hash': 'NDJEMTU0cGVlcnBvbnBvbiA1Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538561',
                                                                                                                                          'fiber_length': 17257,
                                                                                                                                          'poller_hash': 'NTEyNUU0cGVlcnBvbnBvbiA1NzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '801523',
                                                                                                                                          'fiber_length': 18224,
                                                                                                                                          'poller_hash': 'NTEyNTIzcGVlcnBvbnBvbiA1NzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '67790067',
                                                                                                                                          'fiber_length': 18125,
                                                                                                                                          'poller_hash': 'NTEyNUYwcGVlcnBvbnBvbiA1NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538560',
                                                                                                                                          'fiber_length': 16752,
                                                                                                                                          'poller_hash': 'NDJDRTMycGVlcnBvbnBvbiA1Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '67790295',
                                                                                                                                          'fiber_length': 16627,
                                                                                                                                          'poller_hash': 'QjUwREM0cGVlcnBvbnBvbiA2MjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '21816866',
                                                                                                                                          'fiber_length': 15122,
                                                                                                                                          'poller_hash': 'NDJEMkFGcGVlcnBvbnBvbiA2Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538571',
                                                                                                                                          'fiber_length': 16710,
                                                                                                                                          'poller_hash': 'NTEyNTc1cGVlcnBvbnBvbiA2OTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '21816277',
                                                                                                                                          'fiber_length': 16662,
                                                                                                                                          'poller_hash': 'NDJEMTk2cGVlcnBvbnBvbiA2OTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '21816025',
                                                                                                                                          'fiber_length': 16220,
                                                                                                                                          'poller_hash': 'NDJDMzhFcGVlcnBvbnBvbiA2OTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '709198',
                                                                                                                                          'fiber_length': 16123,
                                                                                                                                          'poller_hash': 'NDJDMzhDcGVlcnBvbnBvbiA2OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '72954320',
                                                                                                                                          'fiber_length': 15500,
                                                                                                                                          'poller_hash': 'NDJDMkQzcGVlcnBvbnBvbiA3MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538568',
                                                                                                                                          'fiber_length': 16991,
                                                                                                                                          'poller_hash': 'NUE2QkIycGVlcnBvbnBvbiA3NDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538566',
                                                                                                                                          'fiber_length': 15829,
                                                                                                                                          'poller_hash': 'NUE2QjgwcGVlcnBvbnBvbiA3NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538551',
                                                                                                                                          'fiber_length': 13447,
                                                                                                                                          'poller_hash': 'NDNCM0I2cGVlcnBvbnBvbiA3NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538569',
                                                                                                                                          'fiber_length': 17126,
                                                                                                                                          'poller_hash': 'NTEyNEU2cGVlcnBvbnBvbiA3NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538571',
                                                                                                                                          'fiber_length': 17579,
                                                                                                                                          'poller_hash': 'NUE2NUYycGVlcnBvbnBvbiA3ODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '21816344',
                                                                                                                                          'fiber_length': 17293,
                                                                                                                                          'poller_hash': 'NDJEMDkycGVlcnBvbnBvbiA3OTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '67790146',
                                                                                                                                          'fiber_length': 17199,
                                                                                                                                          'poller_hash': 'NUE2NkM4cGVlcnBvbnBvbiA4MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538551',
                                                                                                                                          'fiber_length': 16166,
                                                                                                                                          'poller_hash': 'MTI3NEE1MXBlZXJwb25wb24gODQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '2788576',
                                                                                                                                          'fiber_length': 16450,
                                                                                                                                          'poller_hash': 'NDJEMUNFcGVlcnBvbnBvbiA4NDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '67790425',
                                                                                                                                          'fiber_length': 16947,
                                                                                                                                          'poller_hash': 'NUE2NjE2cGVlcnBvbnBvbiA4NTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '67790369',
                                                                                                                                          'fiber_length': 16617,
                                                                                                                                          'poller_hash': 'NUE2NjYxcGVlcnBvbnBvbiA4NTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '1722157',
                                                                                                                                          'fiber_length': 17525,
                                                                                                                                          'poller_hash': 'NUE2Qzg5cGVlcnBvbnBvbiA4NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538562',
                                                                                                                                          'fiber_length': 16316,
                                                                                                                                          'poller_hash': 'OTU2RTU4cGVlcnBvbnBvbiA5MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538561',
                                                                                                                                          'fiber_length': 14444,
                                                                                                                                          'poller_hash': 'ODY5M0NFcGVlcnBvbnBvbiAxNDE5OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538566',
                                                                                                                                          'fiber_length': 16376,
                                                                                                                                          'poller_hash': 'ODcyNERFcGVlcnBvbnBvbiAxNDMyMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538563',
                                                                                                                                          'fiber_length': 14175,
                                                                                                                                          'poller_hash': 'QzgzQjY0cGVlcnBvbnBvbiAxNjY2MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538555',
                                                                                                                                          'fiber_length': 16429,
                                                                                                                                          'poller_hash': 'QzgxRTc5cGVlcnBvbnBvbiAyMDcxNzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '538555',
                                                                                                                                          'fiber_length': 14462,
                                                                                                                                          'poller_hash': 'MTI3NTE5N3BlZXJwb25wb24gMjI4NzEy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '12530065',
                                                                                                                                          'fiber_length': 17357,
                                                                                                                                          'poller_hash': 'MTI3NDBERHBlZXJwb25wb24gMjQ4NjM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542},
                                                                                                                                  {       'connected_time': '12623764',
                                                                                                                                          'fiber_length': 18067,
                                                                                                                                          'poller_hash': 'MTI3NDdGRnBlZXJwb25wb24gMjc0NTk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.542}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/2',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '12/3',
                                                                                         'in_discards': '2764',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '885725',
                                                                                         'in_octets': '773280191910',
                                                                                         'in_ucast_pkts': '4913804413',
                                                                                         'links': [       {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '5479311255',
                                                                                                                  'in_ucast_pkts': '20582337',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434172',
                                                                                                                  'out_octets': '77701517796',
                                                                                                                  'out_ucast_pkts': '61106168',
                                                                                                                  'poller_hash': 'NUE2NUY1cGVlcmludGZwb24gNTc2'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '1029496047',
                                                                                                                  'in_ucast_pkts': '3061948',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434199',
                                                                                                                  'out_octets': '8248706687',
                                                                                                                  'out_ucast_pkts': '7340019',
                                                                                                                  'poller_hash': 'NUE2MUU2cGVlcmludGZwb24gODYw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxRDgwcGVlcmludGZwb24gMjA1NzQz'},
                                                                                                          {       'in_nucast_pkts': '7651166',
                                                                                                                  'in_octets': '118973115194',
                                                                                                                  'in_ucast_pkts': '1602211283',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15308074954',
                                                                                                                  'out_octets': '29146494482497',
                                                                                                                  'out_ucast_pkts': '6369319028',
                                                                                                                  'poller_hash': 'MTI3NEFGQ3BlZXJpbnRmcG9uIDIzNDExMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDA3MHBlZXJpbnRmcG9uIDI0NzEzNQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '12/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '12781519317',
                                                                                         'out_octets': '34694634971079',
                                                                                         'out_ucast_pkts': '13396788095',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8z',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/3',
                                                                                                                 'id': '12/3',
                                                                                                                 'links': [       {       'connected_time': '688556',
                                                                                                                                          'fiber_length': 1743,
                                                                                                                                          'poller_hash': 'NUE2NUY1cGVlcnBvbnBvbiA1NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.773},
                                                                                                                                  {       'connected_time': '688562',
                                                                                                                                          'fiber_length': 1728,
                                                                                                                                          'poller_hash': 'NUE2MUU2cGVlcnBvbnBvbiA4NjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.773},
                                                                                                                                  {       'connected_time': '9673962',
                                                                                                                                          'fiber_length': 1794,
                                                                                                                                          'poller_hash': 'QzgxRDgwcGVlcnBvbnBvbiAyMDU3NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.773},
                                                                                                                                  {       'connected_time': '4156016',
                                                                                                                                          'fiber_length': 1745,
                                                                                                                                          'poller_hash': 'MTI3NEFGQ3BlZXJwb25wb24gMjM0MTEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.773},
                                                                                                                                  {       'connected_time': '67752657',
                                                                                                                                          'fiber_length': 1610,
                                                                                                                                          'poller_hash': 'MTI3NDA3MHBlZXJwb25wb24gMjQ3MTM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.773}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/3',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8z',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '12/4',
                                                                                         'in_discards': '40203',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '6675719',
                                                                                         'in_octets': '6006010656047',
                                                                                         'in_ucast_pkts': '21493313867',
                                                                                         'links': [       {       'in_nucast_pkts': '520',
                                                                                                                  'in_octets': '185381546647',
                                                                                                                  'in_ucast_pkts': '607138461',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31561338',
                                                                                                                  'out_octets': '1772386638724',
                                                                                                                  'out_ucast_pkts': '1630841571',
                                                                                                                  'poller_hash': 'MTc2MjE1cGVlcmludGZwb24gODc='},
                                                                                                          {       'in_nucast_pkts': '1206703',
                                                                                                                  'in_octets': '878178065266',
                                                                                                                  'in_ucast_pkts': '2178269076',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '40279638',
                                                                                                                  'out_octets': '4388017432911',
                                                                                                                  'out_ucast_pkts': '3758449040',
                                                                                                                  'poller_hash': 'NTU1QzlGcGVlcmludGZwb24gNjUy'},
                                                                                                          {       'in_nucast_pkts': '654312',
                                                                                                                  'in_octets': '1381727826743',
                                                                                                                  'in_ucast_pkts': '4250401662',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76600459',
                                                                                                                  'out_octets': '14252832130077',
                                                                                                                  'out_ucast_pkts': '11081123046',
                                                                                                                  'poller_hash': 'NUE2MjIzcGVlcmludGZwb24gODEy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '76790005929',
                                                                                                                  'in_ucast_pkts': '581800318',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '94',
                                                                                                                  'out_octets': '3220116053399',
                                                                                                                  'out_ucast_pkts': '2229629808',
                                                                                                                  'poller_hash': 'NDJCRUEwcGVlcmludGZwb24gODIx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '205238582744',
                                                                                                                  'in_ucast_pkts': '1574683757',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '117',
                                                                                                                  'out_octets': '5626090873116',
                                                                                                                  'out_ucast_pkts': '4066998331',
                                                                                                                  'poller_hash': 'NDJDOEQzcGVlcmludGZwb24gODI3'},
                                                                                                          {       'in_nucast_pkts': '302479',
                                                                                                                  'in_octets': '56497584723',
                                                                                                                  'in_ucast_pkts': '96347000',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '34674642',
                                                                                                                  'out_octets': '160800427464',
                                                                                                                  'out_ucast_pkts': '145024322',
                                                                                                                  'poller_hash': 'NDJDODlDcGVlcmludGZwb24gODQz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '90452451260',
                                                                                                                  'in_ucast_pkts': '481681748',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '135',
                                                                                                                  'out_octets': '2712230052244',
                                                                                                                  'out_ucast_pkts': '1920562486',
                                                                                                                  'poller_hash': 'ODY5NDZEcGVlcmludGZwb24gMTQxODY3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '247750286306',
                                                                                                                  'in_ucast_pkts': '1704330457',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '7164466806485',
                                                                                                                  'out_ucast_pkts': '5141348929',
                                                                                                                  'poller_hash': 'ODY5M0VFcGVlcmludGZwb24gMTQxOTM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '81812068863',
                                                                                                                  'in_ucast_pkts': '241629595',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '113',
                                                                                                                  'out_octets': '913452489300',
                                                                                                                  'out_ucast_pkts': '772688256',
                                                                                                                  'poller_hash': 'ODY5M0ExcGVlcmludGZwb24gMTQyMDI3'},
                                                                                                          {       'in_nucast_pkts': '12406382',
                                                                                                                  'in_octets': '130608035937',
                                                                                                                  'in_ucast_pkts': '676977798',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '292316902',
                                                                                                                  'out_octets': '1013559505324',
                                                                                                                  'out_ucast_pkts': '1389514786',
                                                                                                                  'poller_hash': 'Q0MwRDRwZWVyaW50ZnBvbiAxNDU2NDA='},
                                                                                                          {       'in_nucast_pkts': '580',
                                                                                                                  'in_octets': '486019355646',
                                                                                                                  'in_ucast_pkts': '890973520',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45699978',
                                                                                                                  'out_octets': '1254044956152',
                                                                                                                  'out_ucast_pkts': '1296977127',
                                                                                                                  'poller_hash': 'MTU4RkE3MXBlZXJpbnRmcG9uIDI1NDU4MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REMxMXBlZXJpbnRmcG9uIDI1NDc0NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NjEyMXBlZXJpbnRmcG9uIDI4MzAxMA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '12/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '21543646555',
                                                                                         'out_octets': '102090817068502',
                                                                                         'out_ucast_pkts': '57192225456',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/4',
                                                                                                                 'id': '12/4',
                                                                                                                 'links': [       {       'connected_time': '18081316',
                                                                                                                                          'fiber_length': 2575,
                                                                                                                                          'poller_hash': 'MTc2MjE1cGVlcnBvbnBvbiA4Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '24135479',
                                                                                                                                          'fiber_length': 2162,
                                                                                                                                          'poller_hash': 'NTU1QzlGcGVlcnBvbnBvbiA2NTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '67790372',
                                                                                                                                          'fiber_length': 2877,
                                                                                                                                          'poller_hash': 'NUE2MjIzcGVlcnBvbnBvbiA4MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '688551',
                                                                                                                                          'fiber_length': 1936,
                                                                                                                                          'poller_hash': 'NDJCRUEwcGVlcnBvbnBvbiA4MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '688556',
                                                                                                                                          'fiber_length': 3399,
                                                                                                                                          'poller_hash': 'NDJDOEQzcGVlcnBvbnBvbiA4Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '15472874',
                                                                                                                                          'fiber_length': 2276,
                                                                                                                                          'poller_hash': 'NDJDODlDcGVlcnBvbnBvbiA4NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -12.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '21816868',
                                                                                                                                          'fiber_length': 5060,
                                                                                                                                          'poller_hash': 'ODY5NDZEcGVlcnBvbnBvbiAxNDE4Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '688550',
                                                                                                                                          'fiber_length': 4199,
                                                                                                                                          'poller_hash': 'ODY5M0VFcGVlcnBvbnBvbiAxNDE5Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '21816183',
                                                                                                                                          'fiber_length': 3645,
                                                                                                                                          'poller_hash': 'ODY5M0ExcGVlcnBvbnBvbiAxNDIwMjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '23328610',
                                                                                                                                          'fiber_length': 2300,
                                                                                                                                          'poller_hash': 'Q0MwRDRwZWVycG9ucG9uIDE0NTY0MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '27752524',
                                                                                                                                          'fiber_length': 2010,
                                                                                                                                          'poller_hash': 'MTU4RkE3MXBlZXJwb25wb24gMjU0NTgx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '27579615',
                                                                                                                                          'fiber_length': 1981,
                                                                                                                                          'poller_hash': 'MTE3REMxMXBlZXJwb25wb24gMjU0NzQ1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '2253186',
                                                                                                                                          'fiber_length': 4532,
                                                                                                                                          'poller_hash': 'MTI5NjEyMXBlZXJwb25wb24gMjgzMDEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/4',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '12/5',
                                                                                         'in_discards': '3978613',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '4220641',
                                                                                         'in_octets': '9214533624443',
                                                                                         'in_ucast_pkts': '24148475737',
                                                                                         'links': [       {       'in_nucast_pkts': '1466',
                                                                                                                  'in_octets': '302207125',
                                                                                                                  'in_ucast_pkts': '2080735',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1479154',
                                                                                                                  'out_octets': '14225030162',
                                                                                                                  'out_ucast_pkts': '10435176',
                                                                                                                  'poller_hash': 'NUE2NjA3cGVlcmludGZwb24gNzg2'},
                                                                                                          {       'in_nucast_pkts': '1795',
                                                                                                                  'in_octets': '3836296231',
                                                                                                                  'in_ucast_pkts': '6621711',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '942690',
                                                                                                                  'out_octets': '17097057510',
                                                                                                                  'out_ucast_pkts': '14022253',
                                                                                                                  'poller_hash': 'NTEyNTg4cGVlcmludGZwb24gNzg3'},
                                                                                                          {       'in_nucast_pkts': '1719646',
                                                                                                                  'in_octets': '191545204798',
                                                                                                                  'in_ucast_pkts': '347089528',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3738115',
                                                                                                                  'out_octets': '1150917768164',
                                                                                                                  'out_ucast_pkts': '909928206',
                                                                                                                  'poller_hash': 'NUE2QTEzcGVlcmludGZwb24gNzky'},
                                                                                                          {       'in_nucast_pkts': '11842123',
                                                                                                                  'in_octets': '111348306079',
                                                                                                                  'in_ucast_pkts': '394872470',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '69663174',
                                                                                                                  'out_octets': '790912485033',
                                                                                                                  'out_ucast_pkts': '747831047',
                                                                                                                  'poller_hash': 'NDJDOTE5cGVlcmludGZwb24gODAy'},
                                                                                                          {       'in_nucast_pkts': '30656',
                                                                                                                  'in_octets': '871055710790',
                                                                                                                  'in_ucast_pkts': '4406433883',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '84263560',
                                                                                                                  'out_octets': '15020446034700',
                                                                                                                  'out_ucast_pkts': '11193504916',
                                                                                                                  'poller_hash': 'NDJDOEVBcGVlcmludGZwb24gODY5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '74227161115',
                                                                                                                  'in_ucast_pkts': '590586616',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '3519338421137',
                                                                                                                  'out_ucast_pkts': '2478133928',
                                                                                                                  'poller_hash': 'ODcyNTA2cGVlcmludGZwb24gMTQyODEz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '378087964503',
                                                                                                                  'in_ucast_pkts': '9909425366',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '4304695168170',
                                                                                                                  'out_ucast_pkts': '3523191007',
                                                                                                                  'poller_hash': 'MTI3NEE1RHBlZXJpbnRmcG9uIDIzODU3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '893060494438',
                                                                                                                  'in_ucast_pkts': '91797434016',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '4101658962597',
                                                                                                                  'out_ucast_pkts': '3347386396',
                                                                                                                  'poller_hash': 'MTI3NDFEMHBlZXJpbnRmcG9uIDI0MTM3Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '12/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '159385058',
                                                                                         'out_octets': '80614120147748',
                                                                                         'out_ucast_pkts': '62119846400',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/5',
                                                                                                                 'id': '12/5',
                                                                                                                 'links': [       {       'connected_time': '709228',
                                                                                                                                          'fiber_length': 16887,
                                                                                                                                          'poller_hash': 'NUE2NjA3cGVlcnBvbnBvbiA3ODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934},
                                                                                                                                  {       'connected_time': '454487',
                                                                                                                                          'fiber_length': 19054,
                                                                                                                                          'poller_hash': 'NTEyNTg4cGVlcnBvbnBvbiA3ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934},
                                                                                                                                  {       'connected_time': '1745046',
                                                                                                                                          'fiber_length': 18150,
                                                                                                                                          'poller_hash': 'NUE2QTEzcGVlcnBvbnBvbiA3OTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934},
                                                                                                                                  {       'connected_time': '64718596',
                                                                                                                                          'fiber_length': 18015,
                                                                                                                                          'poller_hash': 'NDJDOTE5cGVlcnBvbnBvbiA4MDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934},
                                                                                                                                  {       'connected_time': '71474612',
                                                                                                                                          'fiber_length': 17122,
                                                                                                                                          'poller_hash': 'NDJDOEVBcGVlcnBvbnBvbiA4Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934},
                                                                                                                                  {       'connected_time': '538563',
                                                                                                                                          'fiber_length': 16833,
                                                                                                                                          'poller_hash': 'ODcyNTA2cGVlcnBvbnBvbiAxNDI4MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934},
                                                                                                                                  {       'connected_time': '3113974',
                                                                                                                                          'fiber_length': 14018,
                                                                                                                                          'poller_hash': 'MTI3NEE1RHBlZXJwb25wb24gMjM4NTc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934},
                                                                                                                                  {       'connected_time': '2607375',
                                                                                                                                          'fiber_length': 17111,
                                                                                                                                          'poller_hash': 'MTI3NDFEMHBlZXJwb25wb24gMjQxMzc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.934}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/5',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '12/6',
                                                                                         'in_discards': '64974',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7953319',
                                                                                         'in_octets': '9224968009185',
                                                                                         'in_ucast_pkts': '38776456235',
                                                                                         'links': [       {       'in_nucast_pkts': '11646',
                                                                                                                  'in_octets': '7657073813',
                                                                                                                  'in_ucast_pkts': '19461340',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1433870',
                                                                                                                  'out_octets': '133914192602',
                                                                                                                  'out_ucast_pkts': '94405904',
                                                                                                                  'poller_hash': 'ODcyNDMxcGVlcmludGZwb24gMTc='},
                                                                                                          {       'in_nucast_pkts': '14857',
                                                                                                                  'in_octets': '13758618998',
                                                                                                                  'in_ucast_pkts': '84155543',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434248',
                                                                                                                  'out_octets': '288668834521',
                                                                                                                  'out_ucast_pkts': '200047722',
                                                                                                                  'poller_hash': 'ODY5MzhBcGVlcmludGZwb24gOTA0'},
                                                                                                          {       'in_nucast_pkts': '14834',
                                                                                                                  'in_octets': '6121184055',
                                                                                                                  'in_ucast_pkts': '31150227',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434273',
                                                                                                                  'out_octets': '124010633563',
                                                                                                                  'out_ucast_pkts': '90677470',
                                                                                                                  'poller_hash': 'ODY5NDg0cGVlcmludGZwb24gMTQwMDkx'},
                                                                                                          {       'in_nucast_pkts': '15416',
                                                                                                                  'in_octets': '10008189306',
                                                                                                                  'in_ucast_pkts': '24092792',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434143',
                                                                                                                  'out_octets': '80831142299',
                                                                                                                  'out_ucast_pkts': '73208487',
                                                                                                                  'poller_hash': 'ODQ0QkY1cGVlcmludGZwb24gMTQwMzI2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '326513884245',
                                                                                                                  'in_ucast_pkts': '1225329103',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '160',
                                                                                                                  'out_octets': '5979743588738',
                                                                                                                  'out_ucast_pkts': '4369932925',
                                                                                                                  'poller_hash': 'ODcyNTdFcGVlcmludGZwb24gMTQ0NTI3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '297405270811',
                                                                                                                  'in_ucast_pkts': '60575896271',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '94',
                                                                                                                  'out_octets': '1394028613385',
                                                                                                                  'out_ucast_pkts': '1155969486',
                                                                                                                  'poller_hash': 'ODcyNTE0cGVlcmludGZwb24gMTQ0OTU4'},
                                                                                                          {       'in_nucast_pkts': '12367',
                                                                                                                  'in_octets': '37369166849',
                                                                                                                  'in_ucast_pkts': '106189432',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434299',
                                                                                                                  'out_octets': '436955397390',
                                                                                                                  'out_ucast_pkts': '352970569',
                                                                                                                  'poller_hash': 'ODcyNTU4cGVlcmludGZwb24gMTQ1Mjg5'},
                                                                                                          {       'in_nucast_pkts': '14301',
                                                                                                                  'in_octets': '2505695767',
                                                                                                                  'in_ucast_pkts': '14371066',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434250',
                                                                                                                  'out_octets': '173013499087',
                                                                                                                  'out_ucast_pkts': '117084458',
                                                                                                                  'poller_hash': 'ODcyNTA3cGVlcmludGZwb24gMTQ3NzI1'},
                                                                                                          {       'in_nucast_pkts': '14647',
                                                                                                                  'in_octets': '470649483',
                                                                                                                  'in_ucast_pkts': '1613951',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434448',
                                                                                                                  'out_octets': '3587132867',
                                                                                                                  'out_ucast_pkts': '3264338',
                                                                                                                  'poller_hash': 'MTE5NEU4Q3BlZXJpbnRmcG9uIDIxNzAyNg=='},
                                                                                                          {       'in_nucast_pkts': '12',
                                                                                                                  'in_octets': '1887066611',
                                                                                                                  'in_ucast_pkts': '3138506',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434781',
                                                                                                                  'out_octets': '5394064352',
                                                                                                                  'out_ucast_pkts': '4852765',
                                                                                                                  'poller_hash': 'MTI3NEE2RHBlZXJpbnRmcG9uIDIzMDg2Ng=='},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '12383247759',
                                                                                                                  'in_ucast_pkts': '58482166',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434872',
                                                                                                                  'out_octets': '243138196607',
                                                                                                                  'out_ucast_pkts': '176051768',
                                                                                                                  'poller_hash': 'MTI3NEIwRHBlZXJpbnRmcG9uIDIzMTM1OA=='},
                                                                                                          {       'in_nucast_pkts': '20',
                                                                                                                  'in_octets': '2668491302',
                                                                                                                  'in_ucast_pkts': '15320015',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434874',
                                                                                                                  'out_octets': '113568624979',
                                                                                                                  'out_ucast_pkts': '80917609',
                                                                                                                  'poller_hash': 'MTI3NEE5MXBlZXJpbnRmcG9uIDIzMTM2OA=='},
                                                                                                          {       'in_nucast_pkts': '20',
                                                                                                                  'in_octets': '12735422226',
                                                                                                                  'in_ucast_pkts': '37880261',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434887',
                                                                                                                  'out_octets': '320170306393',
                                                                                                                  'out_ucast_pkts': '227168018',
                                                                                                                  'poller_hash': 'MTI3NEE1QnBlZXJpbnRmcG9uIDIzMTUwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '320854869669',
                                                                                                                  'in_ucast_pkts': '48291231296',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '4119811317189',
                                                                                                                  'out_ucast_pkts': '3371372945',
                                                                                                                  'poller_hash': 'MTI3NEIwMnBlZXJpbnRmcG9uIDIzMzQ5Mw=='},
                                                                                                          {       'in_nucast_pkts': '14773',
                                                                                                                  'in_octets': '13117106314',
                                                                                                                  'in_ucast_pkts': '16911876',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1435040',
                                                                                                                  'out_octets': '47844823053',
                                                                                                                  'out_ucast_pkts': '40474799',
                                                                                                                  'poller_hash': 'MTI3NDBBMHBlZXJpbnRmcG9uIDI0MDU0OA=='},
                                                                                                          {       'in_nucast_pkts': '14711',
                                                                                                                  'in_octets': '3858174674',
                                                                                                                  'in_ucast_pkts': '9748848',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1435093',
                                                                                                                  'out_octets': '38802419120',
                                                                                                                  'out_ucast_pkts': '30918654',
                                                                                                                  'poller_hash': 'MTI3NDBCRHBlZXJpbnRmcG9uIDI0MjE5NQ=='},
                                                                                                          {       'in_nucast_pkts': '14833',
                                                                                                                  'in_octets': '2434273187',
                                                                                                                  'in_ucast_pkts': '11121837',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1435203',
                                                                                                                  'out_octets': '160586218110',
                                                                                                                  'out_ucast_pkts': '111309537',
                                                                                                                  'poller_hash': 'MTI5NzNFNXBlZXJpbnRmcG9uIDI1NDk0OQ=='},
                                                                                                          {       'in_nucast_pkts': '454670',
                                                                                                                  'in_octets': '51425991067',
                                                                                                                  'in_ucast_pkts': '526239563',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '56321021',
                                                                                                                  'out_octets': '2760545767479',
                                                                                                                  'out_ucast_pkts': '1898817472',
                                                                                                                  'poller_hash': 'MTI5NzQ4Q3BlZXJpbnRmcG9uIDI1NjE5Mg=='},
                                                                                                          {       'in_nucast_pkts': '12111',
                                                                                                                  'in_octets': '3336972692',
                                                                                                                  'in_ucast_pkts': '8808926',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1435250',
                                                                                                                  'out_octets': '81998244116',
                                                                                                                  'out_ucast_pkts': '59872512',
                                                                                                                  'poller_hash': 'MTI5NUVCOHBlZXJpbnRmcG9uIDI3OTMwNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTUyNDJCN3BlZXJpbnRmcG9uIDI4MjE2Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '12/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '1920377849',
                                                                                         'out_octets': '176217810489811',
                                                                                         'out_ucast_pkts': '127525159031',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/6',
                                                                                                                 'id': '12/6',
                                                                                                                 'links': [       {       'connected_time': '688551',
                                                                                                                                          'fiber_length': 1730,
                                                                                                                                          'poller_hash': 'ODcyNDMxcGVlcnBvbnBvbiAxNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688552',
                                                                                                                                          'fiber_length': 1745,
                                                                                                                                          'poller_hash': 'ODY5MzhBcGVlcnBvbnBvbiA5MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688560',
                                                                                                                                          'fiber_length': 1705,
                                                                                                                                          'poller_hash': 'ODY5NDg0cGVlcnBvbnBvbiAxNDAwOTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688574',
                                                                                                                                          'fiber_length': 1729,
                                                                                                                                          'poller_hash': 'ODQ0QkY1cGVlcnBvbnBvbiAxNDAzMjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688551',
                                                                                                                                          'fiber_length': 1716,
                                                                                                                                          'poller_hash': 'ODcyNTdFcGVlcnBvbnBvbiAxNDQ1Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688557',
                                                                                                                                          'fiber_length': 1743,
                                                                                                                                          'poller_hash': 'ODcyNTE0cGVlcnBvbnBvbiAxNDQ5NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688559',
                                                                                                                                          'fiber_length': 1718,
                                                                                                                                          'poller_hash': 'ODcyNTU4cGVlcnBvbnBvbiAxNDUyODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688562',
                                                                                                                                          'fiber_length': 1732,
                                                                                                                                          'poller_hash': 'ODcyNTA3cGVlcnBvbnBvbiAxNDc3MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688557',
                                                                                                                                          'fiber_length': 1788,
                                                                                                                                          'poller_hash': 'MTE5NEU4Q3BlZXJwb25wb24gMjE3MDI2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688559',
                                                                                                                                          'fiber_length': 1800,
                                                                                                                                          'poller_hash': 'MTI3NEE2RHBlZXJwb25wb24gMjMwODY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688557',
                                                                                                                                          'fiber_length': 1804,
                                                                                                                                          'poller_hash': 'MTI3NEIwRHBlZXJwb25wb24gMjMxMzU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688555',
                                                                                                                                          'fiber_length': 1817,
                                                                                                                                          'poller_hash': 'MTI3NEE5MXBlZXJwb25wb24gMjMxMzY4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688560',
                                                                                                                                          'fiber_length': 1789,
                                                                                                                                          'poller_hash': 'MTI3NEE1QnBlZXJwb25wb24gMjMxNTAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688551',
                                                                                                                                          'fiber_length': 1816,
                                                                                                                                          'poller_hash': 'MTI3NEIwMnBlZXJwb25wb24gMjMzNDkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688561',
                                                                                                                                          'fiber_length': 1812,
                                                                                                                                          'poller_hash': 'MTI3NDBBMHBlZXJwb25wb24gMjQwNTQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688557',
                                                                                                                                          'fiber_length': 1808,
                                                                                                                                          'poller_hash': 'MTI3NDBCRHBlZXJwb25wb24gMjQyMTk1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688553',
                                                                                                                                          'fiber_length': 1802,
                                                                                                                                          'poller_hash': 'MTI5NzNFNXBlZXJwb25wb24gMjU0OTQ5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '26102069',
                                                                                                                                          'fiber_length': 1634,
                                                                                                                                          'poller_hash': 'MTI5NzQ4Q3BlZXJwb25wb24gMjU2MTky',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688561',
                                                                                                                                          'fiber_length': 1658,
                                                                                                                                          'poller_hash': 'MTI5NUVCOHBlZXJwb25wb24gMjc5MzA2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407},
                                                                                                                                  {       'connected_time': '688454',
                                                                                                                                          'fiber_length': 1815,
                                                                                                                                          'poller_hash': 'MTUyNDJCN3BlZXJwb25wb24gMjgyMTY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.407}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/6',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '12/7',
                                                                                         'in_discards': '118068',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '963024',
                                                                                         'in_octets': '11792236448691',
                                                                                         'in_ucast_pkts': '23381682662',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1640869158223',
                                                                                                                  'in_ucast_pkts': '281456994457',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '137',
                                                                                                                  'out_octets': '2285283885184',
                                                                                                                  'out_ucast_pkts': '2698149548',
                                                                                                                  'poller_hash': 'ODY5M0Y0cGVlcmludGZwb24gODcz'},
                                                                                                          {       'in_nucast_pkts': '11793',
                                                                                                                  'in_octets': '8395939641',
                                                                                                                  'in_ucast_pkts': '43455970',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434207',
                                                                                                                  'out_octets': '162286012102',
                                                                                                                  'out_ucast_pkts': '113690705',
                                                                                                                  'poller_hash': 'QTlDMkJEcGVlcmludGZwb24gODc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1026664531021',
                                                                                                                  'in_ucast_pkts': '78750957416',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '90',
                                                                                                                  'out_octets': '2498798746062',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODY5M0Y5cGVlcmludGZwb24gMTM4Nzgx'},
                                                                                                          {       'in_nucast_pkts': '906',
                                                                                                                  'in_octets': '6933213506',
                                                                                                                  'in_ucast_pkts': '23234245',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1434228',
                                                                                                                  'out_octets': '119476391974',
                                                                                                                  'out_ucast_pkts': '83683939',
                                                                                                                  'poller_hash': 'NUE2OUM5cGVlcmludGZwb24gMTM5MTcy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1643322297651',
                                                                                                                  'in_ucast_pkts': '36393324448',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '4099892658813',
                                                                                                                  'out_ucast_pkts': '3552325179',
                                                                                                                  'poller_hash': 'ODY5NDc0cGVlcmludGZwb24gMTQxODc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '70062464910',
                                                                                                                  'in_ucast_pkts': '323705434',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '94',
                                                                                                                  'out_octets': '1731312572121',
                                                                                                                  'out_ucast_pkts': '1426996547',
                                                                                                                  'poller_hash': 'ODY5NDAwcGVlcmludGZwb24gMTQyMDA3'},
                                                                                                          {       'in_nucast_pkts': '608827',
                                                                                                                  'in_octets': '120156754603',
                                                                                                                  'in_ucast_pkts': '1106865886',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66282161',
                                                                                                                  'out_octets': '6270750813097',
                                                                                                                  'out_ucast_pkts': '4226265955',
                                                                                                                  'poller_hash': 'MTI3NDU5M3BlZXJpbnRmcG9uIDI0NjUxMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '65962708150',
                                                                                                                  'in_ucast_pkts': '146306891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '46',
                                                                                                                  'out_octets': '349886569189',
                                                                                                                  'out_ucast_pkts': '329253001',
                                                                                                                  'poller_hash': 'MTBBODY0MnBlZXJpbnRmcG9uIDI0OTI1Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '98464699699',
                                                                                                                  'in_ucast_pkts': '148192150',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '389345979327',
                                                                                                                  'out_ucast_pkts': '371860612',
                                                                                                                  'poller_hash': 'MTE3REJDRHBlZXJpbnRmcG9uIDI1NDkxNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '391029111833',
                                                                                                                  'in_ucast_pkts': '47946549428',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1603925272843',
                                                                                                                  'out_ucast_pkts': '1327827281',
                                                                                                                  'poller_hash': 'MTI5NzQ4QnBlZXJpbnRmcG9uIDI1NTgxOQ=='},
                                                                                                          {       'in_nucast_pkts': '241268',
                                                                                                                  'in_octets': '61264932901',
                                                                                                                  'in_ucast_pkts': '310690570',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18291499',
                                                                                                                  'out_octets': '832125635658',
                                                                                                                  'out_ucast_pkts': '662179248',
                                                                                                                  'poller_hash': 'MTI5NzNGRHBlZXJpbnRmcG9uIDI1NTg1OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '74166724903',
                                                                                                                  'in_ucast_pkts': '189380956',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43',
                                                                                                                  'out_octets': '585044190391',
                                                                                                                  'out_ucast_pkts': '517231570',
                                                                                                                  'poller_hash': 'MTI5NzRBOXBlZXJpbnRmcG9uIDI1NjYzNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '54385457970',
                                                                                                                  'in_ucast_pkts': '381506852',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '1943006668110',
                                                                                                                  'out_ucast_pkts': '1368748657',
                                                                                                                  'poller_hash': 'MTFCRTFEQ3BlZXJpbnRmcG9uIDI1ODAzNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '62284828309',
                                                                                                                  'in_ucast_pkts': '225313609',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '23',
                                                                                                                  'out_octets': '2067151778325',
                                                                                                                  'out_ucast_pkts': '1481997103',
                                                                                                                  'poller_hash': 'MTI3NDdCOHBlZXJpbnRmcG9uIDI2MjY0OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '4986301245',
                                                                                                                  'in_ucast_pkts': '31184593',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '297378356947',
                                                                                                                  'out_ucast_pkts': '243060673',
                                                                                                                  'poller_hash': 'MTI3NDczRHBlZXJpbnRmcG9uIDI2MzE0NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE5MkY5QnBlZXJpbnRmcG9uIDI2NjQzNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE5MkZGNXBlZXJpbnRmcG9uIDI2NjQ5NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE5MzA0MnBlZXJpbnRmcG9uIDI2NjUwNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '19072652694',
                                                                                                                  'in_ucast_pkts': '56542254',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '208230324202',
                                                                                                                  'out_ucast_pkts': '178419057',
                                                                                                                  'poller_hash': 'MTI3NDgwMHBlZXJpbnRmcG9uIDI3MjgzMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '97112081365',
                                                                                                                  'in_ucast_pkts': '21738596000',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1475807946359',
                                                                                                                  'out_ucast_pkts': '1044648714',
                                                                                                                  'poller_hash': 'MTI5NUVBM3BlZXJpbnRmcG9uIDI3OTI3Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3M0FERnBlZXJpbnRmcG9uIDI3OTQzNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxRTJEcGVlcmludGZwb24gMjc5NTI2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '826361575',
                                                                                                                  'in_ucast_pkts': '5004410',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '38533322150',
                                                                                                                  'out_ucast_pkts': '32037892',
                                                                                                                  'poller_hash': 'MTI5NjEyN3BlZXJpbnRmcG9uIDI4MDM5NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '777290892',
                                                                                                                  'in_ucast_pkts': '7642011',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '26998996509',
                                                                                                                  'out_ucast_pkts': '18543018',
                                                                                                                  'poller_hash': 'MTI5NjA0RXBlZXJpbnRmcG9uIDI4MjE5Mw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '12/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '74223635625',
                                                                                         'out_octets': '174556874501685',
                                                                                         'out_ucast_pkts': '57506200751',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/7',
                                                                                                                 'id': '12/7',
                                                                                                                 'links': [       {       'connected_time': '73959560',
                                                                                                                                          'fiber_length': 6307,
                                                                                                                                          'poller_hash': 'ODY5M0Y0cGVlcnBvbnBvbiA4NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688560',
                                                                                                                                          'fiber_length': 5939,
                                                                                                                                          'poller_hash': 'QTlDMkJEcGVlcnBvbnBvbiA4Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '21816411',
                                                                                                                                          'fiber_length': 6396,
                                                                                                                                          'poller_hash': 'ODY5M0Y5cGVlcnBvbnBvbiAxMzg3ODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688573',
                                                                                                                                          'fiber_length': 5768,
                                                                                                                                          'poller_hash': 'NUE2OUM5cGVlcnBvbnBvbiAxMzkxNzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688594',
                                                                                                                                          'fiber_length': 5409,
                                                                                                                                          'poller_hash': 'ODY5NDc0cGVlcnBvbnBvbiAxNDE4Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688554',
                                                                                                                                          'fiber_length': 5325,
                                                                                                                                          'poller_hash': 'ODY5NDAwcGVlcnBvbnBvbiAxNDIwMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '68532133',
                                                                                                                                          'fiber_length': 10424,
                                                                                                                                          'poller_hash': 'MTI3NDU5M3BlZXJwb25wb24gMjQ2NTEz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '65327368',
                                                                                                                                          'fiber_length': 9019,
                                                                                                                                          'poller_hash': 'MTBBODY0MnBlZXJwb25wb24gMjQ5MjUz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688564',
                                                                                                                                          'fiber_length': 5944,
                                                                                                                                          'poller_hash': 'MTE3REJDRHBlZXJwb25wb24gMjU0OTE0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688593',
                                                                                                                                          'fiber_length': 6310,
                                                                                                                                          'poller_hash': 'MTI5NzQ4QnBlZXJwb25wb24gMjU1ODE5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '26550122',
                                                                                                                                          'fiber_length': 6069,
                                                                                                                                          'poller_hash': 'MTI5NzNGRHBlZXJwb25wb24gMjU1ODU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '24804171',
                                                                                                                                          'fiber_length': 6952,
                                                                                                                                          'poller_hash': 'MTI5NzRBOXBlZXJwb25wb24gMjU2NjM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '538573',
                                                                                                                                          'fiber_length': 9759,
                                                                                                                                          'poller_hash': 'MTFCRTFEQ3BlZXJwb25wb24gMjU4MDM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '21086025',
                                                                                                                                          'fiber_length': 6983,
                                                                                                                                          'poller_hash': 'MTI3NDdCOHBlZXJwb25wb24gMjYyNjQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '20049931',
                                                                                                                                          'fiber_length': 6190,
                                                                                                                                          'poller_hash': 'MTI3NDczRHBlZXJwb25wb24gMjYzMTQ1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '16847561',
                                                                                                                                          'fiber_length': 7123,
                                                                                                                                          'poller_hash': 'MTE5MkY5QnBlZXJwb25wb24gMjY2NDM0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '16696858',
                                                                                                                                          'fiber_length': 6498,
                                                                                                                                          'poller_hash': 'MTE5MkZGNXBlZXJwb25wb24gMjY2NDk0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '16693433',
                                                                                                                                          'fiber_length': 6375,
                                                                                                                                          'poller_hash': 'MTE5MzA0MnBlZXJwb25wb24gMjY2NTA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '15470695',
                                                                                                                                          'fiber_length': 6205,
                                                                                                                                          'poller_hash': 'MTI3NDgwMHBlZXJwb25wb24gMjcyODMx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688571',
                                                                                                                                          'fiber_length': 7114,
                                                                                                                                          'poller_hash': 'MTI5NUVBM3BlZXJwb25wb24gMjc5Mjc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '5702777',
                                                                                                                                          'fiber_length': 5184,
                                                                                                                                          'poller_hash': 'MTI3M0FERnBlZXJwb25wb24gMjc5NDM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '5466100',
                                                                                                                                          'fiber_length': 5340,
                                                                                                                                          'poller_hash': 'QzgxRTJEcGVlcnBvbnBvbiAyNzk1MjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '688567',
                                                                                                                                          'fiber_length': 6296,
                                                                                                                                          'poller_hash': 'MTI5NjEyN3BlZXJwb25wb24gMjgwMzk0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585},
                                                                                                                                  {       'connected_time': '3135641',
                                                                                                                                          'fiber_length': 8784,
                                                                                                                                          'poller_hash': 'MTI5NjA0RXBlZXJwb25wb24gMjgyMTkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.585}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/7',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '12/8',
                                                                                         'in_discards': '9388',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '832521',
                                                                                         'in_octets': '5790886832',
                                                                                         'in_ucast_pkts': '24831286',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzVFQTJEcGVlcmludGZwb24gMTQ='},
                                                                                                          {       'in_nucast_pkts': '5052',
                                                                                                                  'in_octets': '28662172',
                                                                                                                  'in_ucast_pkts': '147346',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '382947',
                                                                                                                  'out_octets': '1212964345',
                                                                                                                  'out_ucast_pkts': '894324',
                                                                                                                  'poller_hash': 'REZGREZGcGVlcmludGZwb24gMTcz'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '12/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '305120155',
                                                                                         'out_octets': '204251663090',
                                                                                         'out_ucast_pkts': '80010846',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi84',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '12/8',
                                                                                                                 'id': '12/8',
                                                                                                                 'links': [       {       'connected_time': '688575',
                                                                                                                                          'fiber_length': 242,
                                                                                                                                          'poller_hash': 'MzVFQTJEcGVlcnBvbnBvbiAxNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.58},
                                                                                                                                  {       'connected_time': '688559',
                                                                                                                                          'fiber_length': 242,
                                                                                                                                          'poller_hash': 'REZGREZGcGVlcnBvbnBvbiAxNzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.58}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '12/8',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi84',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'NDcxOTA1MDEwMzY4',
                                                           'serial_number': '471905010368',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991247'},
                                                   {       'id': '13',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '13/1',
                                                                                         'in_discards': '1216',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '740114',
                                                                                         'in_octets': '2096582519852',
                                                                                         'in_ucast_pkts': '6320691735',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgxRDJBcGVlcmludGZwb24gMjAxMzg4'},
                                                                                                          {       'in_nucast_pkts': '11665',
                                                                                                                  'in_octets': '100634487610',
                                                                                                                  'in_ucast_pkts': '86046862',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1444329',
                                                                                                                  'out_octets': '90904980749',
                                                                                                                  'out_ucast_pkts': '120829431',
                                                                                                                  'poller_hash': 'QzgxRTU3cGVlcmludGZwb24gMjA1ODMz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '31218444575',
                                                                                                                  'in_ucast_pkts': '153576766',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '94',
                                                                                                                  'out_octets': '413332880300',
                                                                                                                  'out_ucast_pkts': '405048546',
                                                                                                                  'poller_hash': 'MTI3NDJBNXBlZXJpbnRmcG9uIDIyOTU4OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '63876646725',
                                                                                                                  'in_ucast_pkts': '364933106',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '762005095894',
                                                                                                                  'out_ucast_pkts': '677602634',
                                                                                                                  'poller_hash': 'MTI3NEEwRXBlZXJpbnRmcG9uIDIzMDYwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '39921228338',
                                                                                                                  'in_ucast_pkts': '103213548',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '280641149514',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEI0QXBlZXJpbnRmcG9uIDIzMTQxMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '77986738718',
                                                                                                                  'in_ucast_pkts': '662216586',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '2180813670493',
                                                                                                                  'out_ucast_pkts': '1598005181',
                                                                                                                  'poller_hash': 'MTI3NEFGRnBlZXJpbnRmcG9uIDIzMTUyMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '862816245',
                                                                                                                  'in_ucast_pkts': '2169219',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '16085624093',
                                                                                                                  'out_ucast_pkts': '60646113',
                                                                                                                  'poller_hash': 'MTI3NEExRHBlZXJpbnRmcG9uIDIzMTUzMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEIzMHBlZXJpbnRmcG9uIDIzMTU4MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEE4Q3BlZXJpbnRmcG9uIDIzMzgxOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEI0NHBlZXJpbnRmcG9uIDIzNDQzNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '212724707209',
                                                                                                                  'in_ucast_pkts': '5065624881',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37',
                                                                                                                  'out_octets': '3013811157740',
                                                                                                                  'out_ucast_pkts': '2296872542',
                                                                                                                  'poller_hash': 'MTI3NEEyOHBlZXJpbnRmcG9uIDIzNTAxOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REI0RnBlZXJpbnRmcG9uIDI0OTI3Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MkZFQUYzcGVlcmludGZwb24gMjUyODI1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '19732042760',
                                                                                                                  'in_ucast_pkts': '48502659',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '140785676232',
                                                                                                                  'out_ucast_pkts': '132638040',
                                                                                                                  'poller_hash': 'MTI3NDdDNHBlZXJpbnRmcG9uIDI1OTkxMg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '13/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '175816444',
                                                                                         'out_octets': '17805293959301',
                                                                                         'out_ucast_pkts': '14554471190',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/1',
                                                                                                                 'id': '13/1',
                                                                                                                 'links': [       {       'connected_time': '21815926',
                                                                                                                                          'fiber_length': 8151,
                                                                                                                                          'poller_hash': 'QzgxRDJBcGVlcnBvbnBvbiAyMDEzODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '688562',
                                                                                                                                          'fiber_length': 10406,
                                                                                                                                          'poller_hash': 'QzgxRTU3cGVlcnBvbnBvbiAyMDU4MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '2056952',
                                                                                                                                          'fiber_length': 9745,
                                                                                                                                          'poller_hash': 'MTI3NDJBNXBlZXJwb25wb24gMjI5NTg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '8980806',
                                                                                                                                          'fiber_length': 9878,
                                                                                                                                          'poller_hash': 'MTI3NEEwRXBlZXJwb25wb24gMjMwNjAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '8217978',
                                                                                                                                          'fiber_length': 9716,
                                                                                                                                          'poller_hash': 'MTI3NEI0QXBlZXJwb25wb24gMjMxNDEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '7595824',
                                                                                                                                          'fiber_length': 8900,
                                                                                                                                          'poller_hash': 'MTI3NEFGRnBlZXJwb25wb24gMjMxNTIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '7596460',
                                                                                                                                          'fiber_length': 8960,
                                                                                                                                          'poller_hash': 'MTI3NEExRHBlZXJwb25wb24gMjMxNTMw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '29747922',
                                                                                                                                          'fiber_length': 9418,
                                                                                                                                          'poller_hash': 'MTI3NEIzMHBlZXJwb25wb24gMjMxNTgw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '4576738',
                                                                                                                                          'fiber_length': 7327,
                                                                                                                                          'poller_hash': 'MTI3NEE4Q3BlZXJwb25wb24gMjMzODE5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '3634524',
                                                                                                                                          'fiber_length': 9207,
                                                                                                                                          'poller_hash': 'MTI3NEI0NHBlZXJwb25wb24gMjM0NDM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '7004732',
                                                                                                                                          'fiber_length': 8377,
                                                                                                                                          'poller_hash': 'MTI3NEEyOHBlZXJwb25wb24gMjM1MDE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '10913301',
                                                                                                                                          'fiber_length': 7339,
                                                                                                                                          'poller_hash': 'MTE3REI0RnBlZXJwb25wb24gMjQ5Mjc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '29465198',
                                                                                                                                          'fiber_length': 7459,
                                                                                                                                          'poller_hash': 'MkZFQUYzcGVlcnBvbnBvbiAyNTI4MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568},
                                                                                                                                  {       'connected_time': '16416929',
                                                                                                                                          'fiber_length': 8865,
                                                                                                                                          'poller_hash': 'MTI3NDdDNHBlZXJwb25wb24gMjU5OTEy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.568}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/1',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '13/2',
                                                                                         'in_discards': '717140',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1294916',
                                                                                         'in_octets': '7618454762116',
                                                                                         'in_ucast_pkts': '24544249808',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '78264383133',
                                                                                                                  'in_ucast_pkts': '297619804',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '1123892797549',
                                                                                                                  'out_ucast_pkts': '907867046',
                                                                                                                  'poller_hash': 'MTI3NEZGM3BlZXJpbnRmcG9uIDIyODUxNQ=='},
                                                                                                          {       'in_nucast_pkts': '14652',
                                                                                                                  'in_octets': '6486487994',
                                                                                                                  'in_ucast_pkts': '15207215',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1464646',
                                                                                                                  'out_octets': '104464494839',
                                                                                                                  'out_ucast_pkts': '79271787',
                                                                                                                  'poller_hash': 'MTI3NTAzQXBlZXJpbnRmcG9uIDIyODUzNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '94226339229',
                                                                                                                  'in_ucast_pkts': '304129863',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43',
                                                                                                                  'out_octets': '1219167609999',
                                                                                                                  'out_ucast_pkts': '1028394564',
                                                                                                                  'poller_hash': 'MTI3NDJEQnBlZXJpbnRmcG9uIDIyOTU0OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '250546673376',
                                                                                                                  'in_ucast_pkts': '207831004',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '44',
                                                                                                                  'out_octets': '78549376616',
                                                                                                                  'out_ucast_pkts': '207093604',
                                                                                                                  'poller_hash': 'MTI3NEIyNnBlZXJpbnRmcG9uIDIyOTU1OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '441745612553',
                                                                                                                  'in_ucast_pkts': '1479836709',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '4273621303522',
                                                                                                                  'out_ucast_pkts': '3534016833',
                                                                                                                  'poller_hash': 'MTI3NDJGOHBlZXJpbnRmcG9uIDIyOTU2OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '38516363480',
                                                                                                                  'in_ucast_pkts': '105530441',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '305220549728',
                                                                                                                  'out_ucast_pkts': '317511830',
                                                                                                                  'poller_hash': 'MTI3NEFGRXBlZXJpbnRmcG9uIDIyOTU3OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '793305122763',
                                                                                                                  'in_ucast_pkts': '250352909946',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43',
                                                                                                                  'out_octets': '4758019287442',
                                                                                                                  'out_ucast_pkts': '3835575889',
                                                                                                                  'poller_hash': 'MTI3NEE4OXBlZXJpbnRmcG9uIDIyOTY1MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '68042206030',
                                                                                                                  'in_ucast_pkts': '243903220',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '44',
                                                                                                                  'out_octets': '1048476343935',
                                                                                                                  'out_ucast_pkts': '827684392',
                                                                                                                  'poller_hash': 'MTI3NEFBMHBlZXJpbnRmcG9uIDIyOTcyMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '623637388428',
                                                                                                                  'in_ucast_pkts': '65626868322',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43',
                                                                                                                  'out_octets': '18683232478953',
                                                                                                                  'out_ucast_pkts': '13132013532',
                                                                                                                  'poller_hash': 'MTI3NEEzN3BlZXJpbnRmcG9uIDIyOTczMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '603821480599',
                                                                                                                  'in_ucast_pkts': '2316328056',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '167',
                                                                                                                  'out_octets': '8306422161531',
                                                                                                                  'out_ucast_pkts': '6169228909',
                                                                                                                  'poller_hash': 'MTI3NEIyRnBlZXJpbnRmcG9uIDIzMDA2NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTUzRTYzOXBlZXJpbnRmcG9uIDIzMDcwNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '461505791348',
                                                                                                                  'in_ucast_pkts': '527173990',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '760462284553',
                                                                                                                  'out_ucast_pkts': '765682507',
                                                                                                                  'poller_hash': 'MTI3NDk0NnBlZXJpbnRmcG9uIDIzMDkxNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '269837026390',
                                                                                                                  'in_ucast_pkts': '244639919',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43',
                                                                                                                  'out_octets': '211106747917',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEI2RnBlZXJpbnRmcG9uIDIzMTU2MA=='},
                                                                                                          {       'in_nucast_pkts': '2510028',
                                                                                                                  'in_octets': '1418585235575',
                                                                                                                  'in_ucast_pkts': '3938289468',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '121758713',
                                                                                                                  'out_octets': '9603518292591',
                                                                                                                  'out_ucast_pkts': '7713232754',
                                                                                                                  'poller_hash': 'MTI3NEFCOHBlZXJpbnRmcG9uIDIzMTU3MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '38434800184',
                                                                                                                  'in_ucast_pkts': '67015530',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '198963142794',
                                                                                                                  'out_ucast_pkts': '226430107',
                                                                                                                  'poller_hash': 'MTI3NEFEMHBlZXJpbnRmcG9uIDIzMjE4NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '20777396132',
                                                                                                                  'in_ucast_pkts': '181029269',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71',
                                                                                                                  'out_octets': '906809261719',
                                                                                                                  'out_ucast_pkts': '695411732',
                                                                                                                  'poller_hash': 'MTI3NEFEN3BlZXJpbnRmcG9uIDIzMjI3OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEE5NHBlZXJpbnRmcG9uIDIzMjMxOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEEwNXBlZXJpbnRmcG9uIDIzMjgxMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDlDQ3BlZXJpbnRmcG9uIDIzMzI1OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTBBODZCRHBlZXJpbnRmcG9uIDI0OTI5Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '13/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '175816437',
                                                                                         'out_octets': '78118905570427',
                                                                                         'out_ucast_pkts': '59864712642',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/2',
                                                                                                                 'id': '13/2',
                                                                                                                 'links': [       {       'connected_time': '10115362',
                                                                                                                                          'fiber_length': 7932,
                                                                                                                                          'poller_hash': 'MTI3NEZGM3BlZXJwb25wb24gMjI4NTE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '688565',
                                                                                                                                          'fiber_length': 8927,
                                                                                                                                          'poller_hash': 'MTI3NTAzQXBlZXJwb25wb24gMjI4NTM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '688560',
                                                                                                                                          'fiber_length': 7325,
                                                                                                                                          'poller_hash': 'MTI3NDJEQnBlZXJwb25wb24gMjI5NTQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '688554',
                                                                                                                                          'fiber_length': 10189,
                                                                                                                                          'poller_hash': 'MTI3NEIyNnBlZXJwb25wb24gMjI5NTU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '16867954',
                                                                                                                                          'fiber_length': 8561,
                                                                                                                                          'poller_hash': 'MTI3NDJGOHBlZXJwb25wb24gMjI5NTY4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '9686120',
                                                                                                                                          'fiber_length': 10041,
                                                                                                                                          'poller_hash': 'MTI3NEFGRXBlZXJwb25wb24gMjI5NTc4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '688565',
                                                                                                                                          'fiber_length': 8937,
                                                                                                                                          'poller_hash': 'MTI3NEE4OXBlZXJwb25wb24gMjI5NjUx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '688569',
                                                                                                                                          'fiber_length': 8622,
                                                                                                                                          'poller_hash': 'MTI3NEFBMHBlZXJwb25wb24gMjI5NzIx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '13114853',
                                                                                                                                          'fiber_length': 9773,
                                                                                                                                          'poller_hash': 'MTI3NEEzN3BlZXJwb25wb24gMjI5NzMx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '688569',
                                                                                                                                          'fiber_length': 10155,
                                                                                                                                          'poller_hash': 'MTI3NEIyRnBlZXJwb25wb24gMjMwMDY0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '1397632',
                                                                                                                                          'fiber_length': 9594,
                                                                                                                                          'poller_hash': 'MTUzRTYzOXBlZXJwb25wb24gMjMwNzA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '7279529',
                                                                                                                                          'fiber_length': 7847,
                                                                                                                                          'poller_hash': 'MTI3NDk0NnBlZXJwb25wb24gMjMwOTE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '5445301',
                                                                                                                                          'fiber_length': 8574,
                                                                                                                                          'poller_hash': 'MTI3NEI2RnBlZXJwb25wb24gMjMxNTYw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '7252767',
                                                                                                                                          'fiber_length': 7086,
                                                                                                                                          'poller_hash': 'MTI3NEFCOHBlZXJwb25wb24gMjMxNTcw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '23065539',
                                                                                                                                          'fiber_length': 10436,
                                                                                                                                          'poller_hash': 'MTI3NEFEMHBlZXJwb25wb24gMjMyMTg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '6659277',
                                                                                                                                          'fiber_length': 7722,
                                                                                                                                          'poller_hash': 'MTI3NEFEN3BlZXJwb25wb24gMjMyMjc5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '6561877',
                                                                                                                                          'fiber_length': 7716,
                                                                                                                                          'poller_hash': 'MTI3NEE5NHBlZXJwb25wb24gMjMyMzE5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '63977869',
                                                                                                                                          'fiber_length': 7024,
                                                                                                                                          'poller_hash': 'MTI3NEEwNXBlZXJwb25wb24gMjMyODEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '5369219',
                                                                                                                                          'fiber_length': 6898,
                                                                                                                                          'poller_hash': 'MTI3NDlDQ3BlZXJwb25wb24gMjMzMjU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284},
                                                                                                                                  {       'connected_time': '22578896',
                                                                                                                                          'fiber_length': 9070,
                                                                                                                                          'poller_hash': 'MTBBODZCRHBlZXJwb25wb24gMjQ5Mjk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.284}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/2',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '13/3',
                                                                                         'in_discards': '8',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '602',
                                                                                         'in_octets': '3492558657587',
                                                                                         'in_ucast_pkts': '9760703249',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '160186156779',
                                                                                                                  'in_ucast_pkts': '511860676',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '1975854924199',
                                                                                                                  'out_ucast_pkts': '1545797483',
                                                                                                                  'poller_hash': 'MTI3NDkxM3BlZXJpbnRmcG9uIDIzMzQ1Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '777561258578',
                                                                                                                  'in_ucast_pkts': '1865115614',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '72',
                                                                                                                  'out_octets': '6493085087255',
                                                                                                                  'out_ucast_pkts': '5438772427',
                                                                                                                  'poller_hash': 'MTI3NDk4N3BlZXJpbnRmcG9uIDIzNTcyMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '364242538236',
                                                                                                                  'in_ucast_pkts': '275615170351',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '89',
                                                                                                                  'out_octets': '3837196228998',
                                                                                                                  'out_ucast_pkts': '2932106333',
                                                                                                                  'poller_hash': 'MTI3NEExMnBlZXJpbnRmcG9uIDIzNjE1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '351119340',
                                                                                                                  'in_ucast_pkts': '2314962',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '14668988486',
                                                                                                                  'out_ucast_pkts': '60818091',
                                                                                                                  'poller_hash': 'MTI3NEE3QXBlZXJpbnRmcG9uIDIzNjE2Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3101767460',
                                                                                                                  'in_ucast_pkts': '13845176',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '51189734946',
                                                                                                                  'out_ucast_pkts': '89346996',
                                                                                                                  'poller_hash': 'MTI3NDg4QnBlZXJpbnRmcG9uIDIzNjIxNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '99659332619',
                                                                                                                  'in_ucast_pkts': '152334527',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '89',
                                                                                                                  'out_octets': '175903479987',
                                                                                                                  'out_ucast_pkts': '265648659',
                                                                                                                  'poller_hash': 'MTI3NEIxNHBlZXJpbnRmcG9uIDIzNzczMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '28909230837',
                                                                                                                  'in_ucast_pkts': '61555395',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '269997713211',
                                                                                                                  'out_ucast_pkts': '266305459',
                                                                                                                  'poller_hash': 'MTI3NEE1MHBlZXJpbnRmcG9uIDIzNzgyMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEEwN3BlZXJpbnRmcG9uIDIzODQ4MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '125292447451',
                                                                                                                  'in_ucast_pkts': '424741308',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '1187161297641',
                                                                                                                  'out_ucast_pkts': '1035964760',
                                                                                                                  'poller_hash': 'MTI3NEE5RnBlZXJpbnRmcG9uIDIzODg5NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96905352266',
                                                                                                                  'in_ucast_pkts': '308757125',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '1035835631025',
                                                                                                                  'out_ucast_pkts': '911606636',
                                                                                                                  'poller_hash': 'MTI3NDlGNHBlZXJpbnRmcG9uIDIzOTQxNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '25271240325',
                                                                                                                  'in_ucast_pkts': '110125705',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '448868144469',
                                                                                                                  'out_ucast_pkts': '407833550',
                                                                                                                  'poller_hash': 'MTI3NDA0NHBlZXJpbnRmcG9uIDI0MTc3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '55862141355',
                                                                                                                  'in_ucast_pkts': '127036306',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '266301608007',
                                                                                                                  'out_ucast_pkts': '264736115',
                                                                                                                  'poller_hash': 'MTI3NDFBRnBlZXJpbnRmcG9uIDI0NjQxNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3927579314',
                                                                                                                  'in_ucast_pkts': '32492765',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '53117209369',
                                                                                                                  'out_ucast_pkts': '74386473',
                                                                                                                  'poller_hash': 'MTI3NDMwRnBlZXJpbnRmcG9uIDI0ODQ4Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '220617927289',
                                                                                                                  'in_ucast_pkts': '1088862985',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4081011280258',
                                                                                                                  'out_ucast_pkts': '3034352720',
                                                                                                                  'poller_hash': 'MTI3NEE3OXBlZXJpbnRmcG9uIDI0OTA4Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzNGQnBlZXJpbnRmcG9uIDI1NjIxMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDJENHBlZXJpbnRmcG9uIDI2Mjk1NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE5MzAwQnBlZXJpbnRmcG9uIDI2NjQwNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDc4RnBlZXJpbnRmcG9uIDI3MjE3OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NjEyMnBlZXJpbnRmcG9uIDI4NDgzNQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '13/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '126948992',
                                                                                         'out_octets': '35283621688065',
                                                                                         'out_ucast_pkts': '28272717986',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/3',
                                                                                                                 'id': '13/3',
                                                                                                                 'links': [       {       'connected_time': '5265866',
                                                                                                                                          'fiber_length': 10341,
                                                                                                                                          'poller_hash': 'MTI3NDkxM3BlZXJwb25wb24gMjMzNDUz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '688568',
                                                                                                                                          'fiber_length': 9577,
                                                                                                                                          'poller_hash': 'MTI3NDk4N3BlZXJwb25wb24gMjM1NzIz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '688564',
                                                                                                                                          'fiber_length': 9235,
                                                                                                                                          'poller_hash': 'MTI3NEExMnBlZXJwb25wb24gMjM2MTU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '1047179',
                                                                                                                                          'fiber_length': 9053,
                                                                                                                                          'poller_hash': 'MTI3NEE3QXBlZXJwb25wb24gMjM2MTY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '967000',
                                                                                                                                          'fiber_length': 9831,
                                                                                                                                          'poller_hash': 'MTI3NDg4QnBlZXJwb25wb24gMjM2MjE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '4788781',
                                                                                                                                          'fiber_length': 9237,
                                                                                                                                          'poller_hash': 'MTI3NEIxNHBlZXJwb25wb24gMjM3NzMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '90689',
                                                                                                                                          'fiber_length': 9071,
                                                                                                                                          'poller_hash': 'MTI3NEE1MHBlZXJwb25wb24gMjM3ODIz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '73882299',
                                                                                                                                          'fiber_length': 8038,
                                                                                                                                          'poller_hash': 'MTI3NEEwN3BlZXJwb25wb24gMjM4NDgx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '73793265',
                                                                                                                                          'fiber_length': 8644,
                                                                                                                                          'poller_hash': 'MTI3NEE5RnBlZXJwb25wb24gMjM4ODk1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '25920265',
                                                                                                                                          'fiber_length': 9229,
                                                                                                                                          'poller_hash': 'MTI3NDlGNHBlZXJwb25wb24gMjM5NDE3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '2249869',
                                                                                                                                          'fiber_length': 8473,
                                                                                                                                          'poller_hash': 'MTI3NDA0NHBlZXJwb25wb24gMjQxNzc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '68797824',
                                                                                                                                          'fiber_length': 9844,
                                                                                                                                          'poller_hash': 'MTI3NDFBRnBlZXJwb25wb24gMjQ2NDE3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '66619055',
                                                                                                                                          'fiber_length': 4840,
                                                                                                                                          'poller_hash': 'MTI3NDMwRnBlZXJwb25wb24gMjQ4NDgz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '688571',
                                                                                                                                          'fiber_length': 5230,
                                                                                                                                          'poller_hash': 'MTI3NEE3OXBlZXJwb25wb24gMjQ5MDgz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '26096771',
                                                                                                                                          'fiber_length': 3582,
                                                                                                                                          'poller_hash': 'MTI5NzNGQnBlZXJwb25wb24gMjU2MjEy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '20488397',
                                                                                                                                          'fiber_length': 8843,
                                                                                                                                          'poller_hash': 'MTI3NDJENHBlZXJwb25wb24gMjYyOTU1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '16866847',
                                                                                                                                          'fiber_length': 5403,
                                                                                                                                          'poller_hash': 'MTE5MzAwQnBlZXJwb25wb24gMjY2NDA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '16071943',
                                                                                                                                          'fiber_length': 10197,
                                                                                                                                          'poller_hash': 'MTI3NDc4RnBlZXJwb25wb24gMjcyMTc5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279},
                                                                                                                                  {       'connected_time': '971153',
                                                                                                                                          'fiber_length': 5855,
                                                                                                                                          'poller_hash': 'MTI5NjEyMnBlZXJwb25wb24gMjg0ODM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.279}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/3',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '13/4',
                                                                                         'in_discards': '4',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '128770',
                                                                                         'in_octets': '6130950784580',
                                                                                         'in_ucast_pkts': '15366714958',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '272663983066',
                                                                                                                  'in_ucast_pkts': '74081456454',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '3803462566543',
                                                                                                                  'out_ucast_pkts': '2841283891',
                                                                                                                  'poller_hash': 'MTI3NEEyM3BlZXJpbnRmcG9uIDIzMzEwOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '157216073952',
                                                                                                                  'in_ucast_pkts': '280279556',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '49',
                                                                                                                  'out_octets': '721997483677',
                                                                                                                  'out_ucast_pkts': '671094586',
                                                                                                                  'poller_hash': 'MTI3NDlCM3BlZXJpbnRmcG9uIDIzMzUwMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '255084420198',
                                                                                                                  'in_ucast_pkts': '618996602',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '2367819241397',
                                                                                                                  'out_ucast_pkts': '2005838392',
                                                                                                                  'poller_hash': 'MTI3NEE2OHBlZXJpbnRmcG9uIDIzMzU1Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEIzOHBlZXJpbnRmcG9uIDIzMzg0OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '972968078297',
                                                                                                                  'in_ucast_pkts': '96219911399',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '99',
                                                                                                                  'out_octets': '5386609605013',
                                                                                                                  'out_ucast_pkts': '4437757882',
                                                                                                                  'poller_hash': 'MTI3NEE4MHBlZXJpbnRmcG9uIDIzNDQ0Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '49316725521',
                                                                                                                  'in_ucast_pkts': '4435800741',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '667081018999',
                                                                                                                  'out_ucast_pkts': '540103572',
                                                                                                                  'poller_hash': 'MTI3NEEwRnBlZXJpbnRmcG9uIDIzNTAyOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '389015036882',
                                                                                                                  'in_ucast_pkts': '103499865235',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '94',
                                                                                                                  'out_octets': '465229984277',
                                                                                                                  'out_ucast_pkts': '527740059',
                                                                                                                  'poller_hash': 'MTI3NEFENXBlZXJpbnRmcG9uIDIzNTEzMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '416522856857',
                                                                                                                  'in_ucast_pkts': '86778322643',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '104',
                                                                                                                  'out_octets': '2157454306471',
                                                                                                                  'out_ucast_pkts': '1964015540',
                                                                                                                  'poller_hash': 'MTI3NEFDOXBlZXJpbnRmcG9uIDIzNTQwOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '110952404187',
                                                                                                                  'in_ucast_pkts': '179499754',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '344493401493',
                                                                                                                  'out_ucast_pkts': '336343626',
                                                                                                                  'poller_hash': 'MTI3NEFBQXBlZXJpbnRmcG9uIDIzNTQyOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '115844164069',
                                                                                                                  'in_ucast_pkts': '872672990',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '127',
                                                                                                                  'out_octets': '4344781443973',
                                                                                                                  'out_ucast_pkts': '3142006588',
                                                                                                                  'poller_hash': 'MTI3NEE5NnBlZXJpbnRmcG9uIDIzODQ2MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '48570454802',
                                                                                                                  'in_ucast_pkts': '248966251',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '72',
                                                                                                                  'out_octets': '1146208883088',
                                                                                                                  'out_ucast_pkts': '860384380',
                                                                                                                  'poller_hash': 'MTI3NDlDM3BlZXJpbnRmcG9uIDIzOTM0NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '56336145810',
                                                                                                                  'in_ucast_pkts': '287160654',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '1551965307608',
                                                                                                                  'out_ucast_pkts': '1230438874',
                                                                                                                  'poller_hash': 'MTI3NDBCRnBlZXJpbnRmcG9uIDI0MDIyMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '70427988816',
                                                                                                                  'in_ucast_pkts': '232344884',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '885601656615',
                                                                                                                  'out_ucast_pkts': '756065912',
                                                                                                                  'poller_hash': 'MTI3NDdCOXBlZXJpbnRmcG9uIDI0MTg0NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '46205873351',
                                                                                                                  'in_ucast_pkts': '549923591',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '3393074682714',
                                                                                                                  'out_ucast_pkts': '2279598595',
                                                                                                                  'poller_hash': 'MTE3REFFRHBlZXJpbnRmcG9uIDI1Mjg4NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '391027327356',
                                                                                                                  'in_ucast_pkts': '707871206',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '2542947304042',
                                                                                                                  'out_ucast_pkts': '2033164860',
                                                                                                                  'poller_hash': 'MTFCRTE0MXBlZXJpbnRmcG9uIDI1Njg2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTFCRTMwMXBlZXJpbnRmcG9uIDI2NjM1MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '32342715359',
                                                                                                                  'in_ucast_pkts': '59747395',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '142409216355',
                                                                                                                  'out_ucast_pkts': '141396110',
                                                                                                                  'poller_hash': 'MTI3NDg1RHBlZXJpbnRmcG9uIDI3MzU3Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NjBFRnBlZXJpbnRmcG9uIDI4MjI1Mw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '13/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '175816445',
                                                                                         'out_octets': '54518535724748',
                                                                                         'out_ucast_pkts': '42785867532',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/4',
                                                                                                                 'id': '13/4',
                                                                                                                 'links': [       {       'connected_time': '5523610',
                                                                                                                                          'fiber_length': 10647,
                                                                                                                                          'poller_hash': 'MTI3NEEyM3BlZXJwb25wb24gMjMzMTA5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '1023927',
                                                                                                                                          'fiber_length': 9868,
                                                                                                                                          'poller_hash': 'MTI3NDlCM3BlZXJwb25wb24gMjMzNTAz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '4936633',
                                                                                                                                          'fiber_length': 10372,
                                                                                                                                          'poller_hash': 'MTI3NEE2OHBlZXJwb25wb24gMjMzNTUz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '22533668',
                                                                                                                                          'fiber_length': 10970,
                                                                                                                                          'poller_hash': 'MTI3NEIzOHBlZXJwb25wb24gMjMzODQ5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '619253',
                                                                                                                                          'fiber_length': 11211,
                                                                                                                                          'poller_hash': 'MTI3NEE4MHBlZXJwb25wb24gMjM0NDQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '2927154',
                                                                                                                                          'fiber_length': 5507,
                                                                                                                                          'poller_hash': 'MTI3NEEwRnBlZXJwb25wb24gMjM1MDI4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '65750201',
                                                                                                                                          'fiber_length': 7872,
                                                                                                                                          'poller_hash': 'MTI3NEFENXBlZXJwb25wb24gMjM1MTMy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '20538292',
                                                                                                                                          'fiber_length': 7564,
                                                                                                                                          'poller_hash': 'MTI3NEFDOXBlZXJwb25wb24gMjM1NDA4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '2345641',
                                                                                                                                          'fiber_length': 7555,
                                                                                                                                          'poller_hash': 'MTI3NEFBQXBlZXJwb25wb24gMjM1NDI4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '73895475',
                                                                                                                                          'fiber_length': 5681,
                                                                                                                                          'poller_hash': 'MTI3NEE5NnBlZXJwb25wb24gMjM4NDYx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '70168457',
                                                                                                                                          'fiber_length': 4793,
                                                                                                                                          'poller_hash': 'MTI3NDlDM3BlZXJwb25wb24gMjM5MzQ1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '688575',
                                                                                                                                          'fiber_length': 3679,
                                                                                                                                          'poller_hash': 'MTI3NDBCRnBlZXJwb25wb24gMjQwMjIx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '5172934',
                                                                                                                                          'fiber_length': 6153,
                                                                                                                                          'poller_hash': 'MTI3NDdCOXBlZXJwb25wb24gMjQxODQ1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '29057149',
                                                                                                                                          'fiber_length': 4793,
                                                                                                                                          'poller_hash': 'MTE3REFFRHBlZXJwb25wb24gMjUyODg1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '24026354',
                                                                                                                                          'fiber_length': 4999,
                                                                                                                                          'poller_hash': 'MTFCRTE0MXBlZXJwb25wb24gMjU2ODY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '688580',
                                                                                                                                          'fiber_length': 5221,
                                                                                                                                          'poller_hash': 'MTFCRTMwMXBlZXJwb25wb24gMjY2MzUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '15211345',
                                                                                                                                          'fiber_length': 4681,
                                                                                                                                          'poller_hash': 'MTI3NDg1RHBlZXJwb25wb24gMjczNTcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26},
                                                                                                                                  {       'connected_time': '2961268',
                                                                                                                                          'fiber_length': 5919,
                                                                                                                                          'poller_hash': 'MTI5NjBFRnBlZXJwb25wb24gMjgyMjUz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.26}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/4',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '13/5',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '260',
                                                                                         'in_octets': '2626623974102',
                                                                                         'in_ucast_pkts': '5732987461',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '77720088277',
                                                                                                                  'in_ucast_pkts': '13115490861',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '90',
                                                                                                                  'out_octets': '1010485845738',
                                                                                                                  'out_ucast_pkts': '838477811',
                                                                                                                  'poller_hash': 'MTI3NEFGQnBlZXJpbnRmcG9uIDIzMzg1OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '7808792866',
                                                                                                                  'in_ucast_pkts': '31285328',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '109500968392',
                                                                                                                  'out_ucast_pkts': '136243748',
                                                                                                                  'poller_hash': 'MTI3NEFFRHBlZXJpbnRmcG9uIDIzNjg2MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '129989411457',
                                                                                                                  'in_ucast_pkts': '43140135724',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '316160260398',
                                                                                                                  'out_ucast_pkts': '346450292',
                                                                                                                  'poller_hash': 'MTI3NEEwQnBlZXJpbnRmcG9uIDIzODM5MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEIwMHBlZXJpbnRmcG9uIDIzODYzOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '512221432654',
                                                                                                                  'in_ucast_pkts': '47895090303',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '942758876128',
                                                                                                                  'out_ucast_pkts': '959414086',
                                                                                                                  'poller_hash': 'MTI3NDBBNXBlZXJpbnRmcG9uIDIzOTIxMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '456374040948',
                                                                                                                  'in_ucast_pkts': '748945312',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '124',
                                                                                                                  'out_octets': '2294618666961',
                                                                                                                  'out_ucast_pkts': '1978601186',
                                                                                                                  'poller_hash': 'MTI3NDAyRHBlZXJpbnRmcG9uIDIzOTIyMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEIxRHBlZXJpbnRmcG9uIDIzOTY2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDlFMXBlZXJpbnRmcG9uIDIzOTc2MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '85300034263',
                                                                                                                  'in_ucast_pkts': '256652421',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '1089673798592',
                                                                                                                  'out_ucast_pkts': '893778226',
                                                                                                                  'poller_hash': 'MTI3NEFERnBlZXJpbnRmcG9uIDIzOTg0NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEFDRnBlZXJpbnRmcG9uIDI0MTczNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '191856152969',
                                                                                                                  'in_ucast_pkts': '26369288604',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '46',
                                                                                                                  'out_octets': '2465301971014',
                                                                                                                  'out_ucast_pkts': '1891448894',
                                                                                                                  'poller_hash': 'MTI3NDAyOXBlZXJpbnRmcG9uIDI0NDczMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96754856862',
                                                                                                                  'in_ucast_pkts': '328435888',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28',
                                                                                                                  'out_octets': '1646951485953',
                                                                                                                  'out_ucast_pkts': '1351642272',
                                                                                                                  'poller_hash': 'MTBBODZFN3BlZXJpbnRmcG9uIDI0OTQ3OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzM5RXBlZXJpbnRmcG9uIDI1Nzk1OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzM3NXBlZXJpbnRmcG9uIDI1OTUxMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTk5OUZGcGVlcmludGZwb24gMjYwMzg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTU5RjdGOXBlZXJpbnRmcG9uIDI2MjgwOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTU5RjdGRXBlZXJpbnRmcG9uIDI3MjQzMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '796218757',
                                                                                                                  'in_ucast_pkts': '3266865',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7398142548',
                                                                                                                  'out_ucast_pkts': '16778318',
                                                                                                                  'poller_hash': 'MTI3NDc5OHBlZXJpbnRmcG9uIDI3NDM3Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3MEQ5RHBlZXJpbnRmcG9uIDI3OTQwNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '13/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '125948361',
                                                                                         'out_octets': '16799079290594',
                                                                                         'out_ucast_pkts': '14473318532',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/5',
                                                                                                                 'id': '13/5',
                                                                                                                 'links': [       {       'connected_time': '688620',
                                                                                                                                          'fiber_length': 8378,
                                                                                                                                          'poller_hash': 'MTI3NEFGQnBlZXJwb25wb24gMjMzODU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '73710229',
                                                                                                                                          'fiber_length': 8056,
                                                                                                                                          'poller_hash': 'MTI3NEFFRHBlZXJwb25wb24gMjM2ODYx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '74404914',
                                                                                                                                          'fiber_length': 8499,
                                                                                                                                          'poller_hash': 'MTI3NEEwQnBlZXJwb25wb24gMjM4Mzkx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '73802908',
                                                                                                                                          'fiber_length': 3790,
                                                                                                                                          'poller_hash': 'MTI3NEIwMHBlZXJwb25wb24gMjM4NjM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '73621652',
                                                                                                                                          'fiber_length': 4728,
                                                                                                                                          'poller_hash': 'MTI3NDBBNXBlZXJwb25wb24gMjM5MjEx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '13672703',
                                                                                                                                          'fiber_length': 6819,
                                                                                                                                          'poller_hash': 'MTI3NDAyRHBlZXJwb25wb24gMjM5MjIx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '73271910',
                                                                                                                                          'fiber_length': 5305,
                                                                                                                                          'poller_hash': 'MTI3NEIxRHBlZXJwb25wb24gMjM5NjY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '68244312',
                                                                                                                                          'fiber_length': 9139,
                                                                                                                                          'poller_hash': 'MTI3NDlFMXBlZXJwb25wb24gMjM5NzYx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '688582',
                                                                                                                                          'fiber_length': 4054,
                                                                                                                                          'poller_hash': 'MTI3NEFERnBlZXJwb25wb24gMjM5ODQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '64351275',
                                                                                                                                          'fiber_length': 8901,
                                                                                                                                          'poller_hash': 'MTI3NEFDRnBlZXJwb25wb24gMjQxNzM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '688581',
                                                                                                                                          'fiber_length': 7233,
                                                                                                                                          'poller_hash': 'MTI3NDAyOXBlZXJwb25wb24gMjQ0NzMw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '64902557',
                                                                                                                                          'fiber_length': 6622,
                                                                                                                                          'poller_hash': 'MTBBODZFN3BlZXJwb25wb24gMjQ5NDc4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '23344366',
                                                                                                                                          'fiber_length': 6364,
                                                                                                                                          'poller_hash': 'MTI5NzM5RXBlZXJwb25wb24gMjU3OTU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '538989',
                                                                                                                                          'fiber_length': 6859,
                                                                                                                                          'poller_hash': 'MTI5NzM3NXBlZXJwb25wb24gMjU5NTEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '1761814',
                                                                                                                                          'fiber_length': 6967,
                                                                                                                                          'poller_hash': 'MTk5OUZGcGVlcnBvbnBvbiAyNjAzODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '13908384',
                                                                                                                                          'fiber_length': 6379,
                                                                                                                                          'poller_hash': 'MTU5RjdGOXBlZXJwb25wb24gMjYyODA4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '7677066',
                                                                                                                                          'fiber_length': 8563,
                                                                                                                                          'poller_hash': 'MTU5RjdGRXBlZXJwb25wb24gMjcyNDMw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '688644',
                                                                                                                                          'fiber_length': 8438,
                                                                                                                                          'poller_hash': 'MTI3NDc5OHBlZXJwb25wb24gMjc0Mzc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358},
                                                                                                                                  {       'connected_time': '5970382',
                                                                                                                                          'fiber_length': 8565,
                                                                                                                                          'poller_hash': 'MTI3MEQ5RHBlZXJwb25wb24gMjc5NDA2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.358}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/5',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '13/6',
                                                                                         'in_discards': '846',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '824431',
                                                                                         'in_octets': '51795641607232',
                                                                                         'in_ucast_pkts': '58275675274',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NTE4QnBlZXJpbnRmcG9uIDIyNDA2OA=='},
                                                                                                          {       'in_nucast_pkts': '185076',
                                                                                                                  'in_octets': '29274208424909',
                                                                                                                  'in_ucast_pkts': '26135505030',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '131447234',
                                                                                                                  'out_octets': '17085005155466',
                                                                                                                  'out_ucast_pkts': '17805780276',
                                                                                                                  'poller_hash': 'MTI3NEZBM3BlZXJpbnRmcG9uIDIyNDc5OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '214562418988',
                                                                                                                  'in_ucast_pkts': '13565412069',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '68',
                                                                                                                  'out_octets': '3597895512891',
                                                                                                                  'out_ucast_pkts': '2675361475',
                                                                                                                  'poller_hash': 'MTI3NTA0M3BlZXJpbnRmcG9uIDIyNjgzNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '296937424082',
                                                                                                                  'in_ucast_pkts': '86794845813',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '2761373519864',
                                                                                                                  'out_ucast_pkts': '2310631327',
                                                                                                                  'poller_hash': 'MTI3NTFEMHBlZXJpbnRmcG9uIDIyNjk3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '178302794693',
                                                                                                                  'in_ucast_pkts': '1124142024',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '4310603161820',
                                                                                                                  'out_ucast_pkts': '3196823674',
                                                                                                                  'poller_hash': 'MTI3NEZGQnBlZXJpbnRmcG9uIDIyNzA2NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NTFENnBlZXJpbnRmcG9uIDIyNzA3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '241655840388',
                                                                                                                  'in_ucast_pkts': '626015121',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '1671940173743',
                                                                                                                  'out_ucast_pkts': '1489212934',
                                                                                                                  'poller_hash': 'MTI3NTA1M3BlZXJpbnRmcG9uIDIyODMxOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDFEMnBlZXJpbnRmcG9uIDI0MTY4NA=='},
                                                                                                          {       'in_nucast_pkts': '1844',
                                                                                                                  'in_octets': '2087926991872',
                                                                                                                  'in_ucast_pkts': '4986827424',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '71494496',
                                                                                                                  'out_octets': '5564287166551',
                                                                                                                  'out_ucast_pkts': '6187781673',
                                                                                                                  'poller_hash': 'MTU4RkE4RnBlZXJpbnRmcG9uIDI0NTczNw=='},
                                                                                                          {       'in_nucast_pkts': '1669',
                                                                                                                  'in_octets': '365988922565',
                                                                                                                  'in_ucast_pkts': '361852727',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1097560',
                                                                                                                  'out_octets': '266929717615',
                                                                                                                  'out_ucast_pkts': '324896276',
                                                                                                                  'poller_hash': 'ODcyNTI0cGVlcmludGZwb24gMjQ2OTMw'},
                                                                                                          {       'in_nucast_pkts': '14614',
                                                                                                                  'in_octets': '1610199115',
                                                                                                                  'in_ucast_pkts': '11337651',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465307',
                                                                                                                  'out_octets': '62687067123',
                                                                                                                  'out_ucast_pkts': '49071074',
                                                                                                                  'poller_hash': 'MTI3NDBBMXBlZXJpbnRmcG9uIDI0Njk3MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzREQ3BlZXJpbnRmcG9uIDI1Njg4Nw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '13/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '23240919065',
                                                                                         'out_octets': '97047370484491',
                                                                                         'out_ucast_pkts': '65816590682',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/6',
                                                                                                                 'id': '13/6',
                                                                                                                 'links': [       {       'connected_time': '5966548',
                                                                                                                                          'fiber_length': 6412,
                                                                                                                                          'poller_hash': 'MTI3NTE4QnBlZXJwb25wb24gMjI0MDY4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '11227824',
                                                                                                                                          'fiber_length': 5439,
                                                                                                                                          'poller_hash': 'MTI3NEZBM3BlZXJwb25wb24gMjI0Nzk5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '688578',
                                                                                                                                          'fiber_length': 4900,
                                                                                                                                          'poller_hash': 'MTI3NTA0M3BlZXJwb25wb24gMjI2ODM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '10707545',
                                                                                                                                          'fiber_length': 6134,
                                                                                                                                          'poller_hash': 'MTI3NTFEMHBlZXJwb25wb24gMjI2OTc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '10361223',
                                                                                                                                          'fiber_length': 6572,
                                                                                                                                          'poller_hash': 'MTI3NEZGQnBlZXJwb25wb24gMjI3MDY1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '10356939',
                                                                                                                                          'fiber_length': 5952,
                                                                                                                                          'poller_hash': 'MTI3NTFENnBlZXJwb25wb24gMjI3MDc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '23690839',
                                                                                                                                          'fiber_length': 4954,
                                                                                                                                          'poller_hash': 'MTI3NTA1M3BlZXJwb25wb24gMjI4MzE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '66371019',
                                                                                                                                          'fiber_length': 5712,
                                                                                                                                          'poller_hash': 'MTI3NDFEMnBlZXJwb25wb24gMjQxNjg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '68954701',
                                                                                                                                          'fiber_length': 6393,
                                                                                                                                          'poller_hash': 'MTU4RkE4RnBlZXJwb25wb24gMjQ1NzM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '513271',
                                                                                                                                          'fiber_length': 6387,
                                                                                                                                          'poller_hash': 'ODcyNTI0cGVlcnBvbnBvbiAyNDY5MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '688587',
                                                                                                                                          'fiber_length': 5454,
                                                                                                                                          'poller_hash': 'MTI3NDBBMXBlZXJwb25wb24gMjQ2OTcw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288},
                                                                                                                                  {       'connected_time': '1671279',
                                                                                                                                          'fiber_length': 6301,
                                                                                                                                          'poller_hash': 'MTI5NzREQ3BlZXJwb25wb24gMjU2ODg3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.288}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/6',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '13/7',
                                                                                         'in_discards': '29461',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3207248',
                                                                                         'in_octets': '1820617322593',
                                                                                         'in_ucast_pkts': '8153409054',
                                                                                         'links': [       {       'in_nucast_pkts': '326788',
                                                                                                                  'in_octets': '1277086333',
                                                                                                                  'in_ucast_pkts': '3775239',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151573',
                                                                                                                  'out_octets': '12282175235',
                                                                                                                  'out_ucast_pkts': '10632653',
                                                                                                                  'poller_hash': 'MTI3NEEyNnBlZXJpbnRmcG9uIDI0ODYxNQ=='},
                                                                                                          {       'in_nucast_pkts': '5222',
                                                                                                                  'in_octets': '219966925',
                                                                                                                  'in_ucast_pkts': '655951',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '541459',
                                                                                                                  'out_octets': '7717400281',
                                                                                                                  'out_ucast_pkts': '5360167',
                                                                                                                  'poller_hash': 'MTBBODY0NHBlZXJpbnRmcG9uIDI0OTE4Mw=='},
                                                                                                          {       'in_nucast_pkts': '5647',
                                                                                                                  'in_octets': '1308860074',
                                                                                                                  'in_ucast_pkts': '4474062',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '541481',
                                                                                                                  'out_octets': '12171894786',
                                                                                                                  'out_ucast_pkts': '9552915',
                                                                                                                  'poller_hash': 'MTE3REMxOXBlZXJpbnRmcG9uIDI1Mjc0MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REI1OHBlZXJpbnRmcG9uIDI1MzQ4Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '123963984683',
                                                                                                                  'in_ucast_pkts': '417850523',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1298381240158',
                                                                                                                  'out_ucast_pkts': '995175132',
                                                                                                                  'poller_hash': 'MTE3REFGOHBlZXJpbnRmcG9uIDI1MzUwNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '77298465346',
                                                                                                                  'in_ucast_pkts': '424611174',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2322087573748',
                                                                                                                  'out_ucast_pkts': '1686896283',
                                                                                                                  'poller_hash': 'MTE3REJERnBlZXJpbnRmcG9uIDI1Mzk1Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '155991663999',
                                                                                                                  'in_ucast_pkts': '1055230506',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2983286586957',
                                                                                                                  'out_ucast_pkts': '2407653660',
                                                                                                                  'poller_hash': 'MTE3REJFMHBlZXJpbnRmcG9uIDI1NDAxMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REJDNHBlZXJpbnRmcG9uIDI1NDUxOQ=='},
                                                                                                          {       'in_nucast_pkts': '519876',
                                                                                                                  'in_octets': '156877185014',
                                                                                                                  'in_ucast_pkts': '496675981',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66419935',
                                                                                                                  'out_octets': '1922894862945',
                                                                                                                  'out_ucast_pkts': '1551779224',
                                                                                                                  'poller_hash': 'MTE3REI0Q3BlZXJpbnRmcG9uIDI1NDU1MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzM3OXBlZXJpbnRmcG9uIDI1NTg0OQ=='},
                                                                                                          {       'in_nucast_pkts': '460244',
                                                                                                                  'in_octets': '173074310049',
                                                                                                                  'in_ucast_pkts': '280432430',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '59910530',
                                                                                                                  'out_octets': '602662160050',
                                                                                                                  'out_ucast_pkts': '538760380',
                                                                                                                  'poller_hash': 'MTI5NzNERXBlZXJpbnRmcG9uIDI1NjUzNg=='},
                                                                                                          {       'in_nucast_pkts': '446236',
                                                                                                                  'in_octets': '42559913549',
                                                                                                                  'in_ucast_pkts': '147522985',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '59213475',
                                                                                                                  'out_octets': '387939606973',
                                                                                                                  'out_ucast_pkts': '350363838',
                                                                                                                  'poller_hash': 'MTI5NzM5NHBlZXJpbnRmcG9uIDI1NjU1Ng=='},
                                                                                                          {       'in_nucast_pkts': '9372',
                                                                                                                  'in_octets': '383538374',
                                                                                                                  'in_ucast_pkts': '1440878',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1151543',
                                                                                                                  'out_octets': '4494758017',
                                                                                                                  'out_ucast_pkts': '3777441',
                                                                                                                  'poller_hash': 'MTFCRTE1QXBlZXJpbnRmcG9uIDI1NzY5OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '31945536187',
                                                                                                                  'in_ucast_pkts': '88622810',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '652386297551',
                                                                                                                  'out_ucast_pkts': '481849515',
                                                                                                                  'poller_hash': 'MTI3NDg3NXBlZXJpbnRmcG9uIDI3NjQ5Mg=='},
                                                                                                          {       'in_nucast_pkts': '955',
                                                                                                                  'in_octets': '6497779201',
                                                                                                                  'in_ucast_pkts': '26449794',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '541560',
                                                                                                                  'out_octets': '107105677882',
                                                                                                                  'out_ucast_pkts': '76377925',
                                                                                                                  'poller_hash': 'MTI3M0FBNnBlZXJpbnRmcG9uIDI3ODQ3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NUU3NnBlZXJpbnRmcG9uIDI3OTA3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '21011760684',
                                                                                                                  'in_ucast_pkts': '39873971',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '146153223817',
                                                                                                                  'out_ucast_pkts': '116824924',
                                                                                                                  'poller_hash': 'MTI5NUU5QnBlZXJpbnRmcG9uIDI4MjkyMQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '13/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '433503461',
                                                                                         'out_octets': '31788991402273',
                                                                                         'out_ucast_pkts': '23865422608',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/7',
                                                                                                                 'id': '13/7',
                                                                                                                 'links': [       {       'connected_time': '538592',
                                                                                                                                          'fiber_length': 15042,
                                                                                                                                          'poller_hash': 'MTI3NEEyNnBlZXJwb25wb24gMjQ4NjE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '262468',
                                                                                                                                          'fiber_length': 18029,
                                                                                                                                          'poller_hash': 'MTBBODY0NHBlZXJwb25wb24gMjQ5MTgz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '262471',
                                                                                                                                          'fiber_length': 15491,
                                                                                                                                          'poller_hash': 'MTE3REMxOXBlZXJwb25wb24gMjUyNzQw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '28878991',
                                                                                                                                          'fiber_length': 17844,
                                                                                                                                          'poller_hash': 'MTE3REI1OHBlZXJwb25wb24gMjUzNDg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '538592',
                                                                                                                                          'fiber_length': 19395,
                                                                                                                                          'poller_hash': 'MTE3REFGOHBlZXJwb25wb24gMjUzNTA2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '538152',
                                                                                                                                          'fiber_length': 18280,
                                                                                                                                          'poller_hash': 'MTE3REJERnBlZXJwb25wb24gMjUzOTUy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '262469',
                                                                                                                                          'fiber_length': 15638,
                                                                                                                                          'poller_hash': 'MTE3REJFMHBlZXJwb25wb24gMjU0MDEy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '10653917',
                                                                                                                                          'fiber_length': 18060,
                                                                                                                                          'poller_hash': 'MTE3REJDNHBlZXJwb25wb24gMjU0NTE5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '27827087',
                                                                                                                                          'fiber_length': 15434,
                                                                                                                                          'poller_hash': 'MTE3REI0Q3BlZXJwb25wb24gMjU0NTUx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '10645451',
                                                                                                                                          'fiber_length': 18749,
                                                                                                                                          'poller_hash': 'MTI5NzM3OXBlZXJwb25wb24gMjU1ODQ5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '25148138',
                                                                                                                                          'fiber_length': 15671,
                                                                                                                                          'poller_hash': 'MTI5NzNERXBlZXJwb25wb24gMjU2NTM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '24902973',
                                                                                                                                          'fiber_length': 16744,
                                                                                                                                          'poller_hash': 'MTI5NzM5NHBlZXJwb25wb24gMjU2NTU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '538594',
                                                                                                                                          'fiber_length': 19325,
                                                                                                                                          'poller_hash': 'MTFCRTE1QXBlZXJwb25wb24gMjU3Njk4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '262468',
                                                                                                                                          'fiber_length': 13715,
                                                                                                                                          'poller_hash': 'MTI3NDg3NXBlZXJwb25wb24gMjc2NDky',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '262470',
                                                                                                                                          'fiber_length': 16790,
                                                                                                                                          'poller_hash': 'MTI3M0FBNnBlZXJwb25wb24gMjc4NDc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '7714292',
                                                                                                                                          'fiber_length': 17364,
                                                                                                                                          'poller_hash': 'MTI5NUU3NnBlZXJwb25wb24gMjc5MDc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666},
                                                                                                                                  {       'connected_time': '262468',
                                                                                                                                          'fiber_length': 16507,
                                                                                                                                          'poller_hash': 'MTI5NUU5QnBlZXJwb25wb24gMjgyOTIx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.666}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/7',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '13/8',
                                                                                         'in_discards': '7113',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5182557',
                                                                                         'in_octets': '5207738807207',
                                                                                         'in_ucast_pkts': '15643296962',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '189395509230',
                                                                                                                  'in_ucast_pkts': '847436561',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3874449075626',
                                                                                                                  'out_ucast_pkts': '2990669072',
                                                                                                                  'poller_hash': 'RUJCREIwcGVlcmludGZwb24gMTg2NTQ3'},
                                                                                                          {       'in_nucast_pkts': '857092',
                                                                                                                  'in_octets': '235401171611',
                                                                                                                  'in_ucast_pkts': '1660050427',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '150689642',
                                                                                                                  'out_octets': '9176969375762',
                                                                                                                  'out_ucast_pkts': '6315222642',
                                                                                                                  'poller_hash': 'REZGRTEycGVlcmludGZwb24gMTg2NTY3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgzMTJDcGVlcmludGZwb24gMTg2NTc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '376957066145',
                                                                                                                  'in_ucast_pkts': '52160572617',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '145',
                                                                                                                  'out_octets': '1154481537816',
                                                                                                                  'out_ucast_pkts': '1034587759',
                                                                                                                  'poller_hash': 'QzgzMTUwcGVlcmludGZwb24gMTg2NTg3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '435483833366',
                                                                                                                  'in_ucast_pkts': '1931366134',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '7189623658597',
                                                                                                                  'out_ucast_pkts': '1101550499',
                                                                                                                  'poller_hash': 'QzgzMEQ5cGVlcmludGZwb24gMTg2NTk3'},
                                                                                                          {       'in_nucast_pkts': '9308279',
                                                                                                                  'in_octets': '86671901662',
                                                                                                                  'in_ucast_pkts': '313113393',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11080557418',
                                                                                                                  'out_octets': '17170016731886',
                                                                                                                  'out_ucast_pkts': '1588731937',
                                                                                                                  'poller_hash': 'QzgxRTVDcGVlcmludGZwb24gMjA2OTY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '346604862775',
                                                                                                                  'in_ucast_pkts': '168676630770',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '49',
                                                                                                                  'out_octets': '5935049359061',
                                                                                                                  'out_ucast_pkts': '4347984767',
                                                                                                                  'poller_hash': 'MTE5NTE1Q3BlZXJpbnRmcG9uIDIxMjcxMA=='},
                                                                                                          {       'in_nucast_pkts': '2723386',
                                                                                                                  'in_octets': '342791430444',
                                                                                                                  'in_ucast_pkts': '399225449',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4362397818',
                                                                                                                  'out_octets': '6308977492132',
                                                                                                                  'out_ucast_pkts': '322244571',
                                                                                                                  'poller_hash': 'MTE5NTFBNXBlZXJpbnRmcG9uIDIxNTY0MQ=='},
                                                                                                          {       'in_nucast_pkts': '65149',
                                                                                                                  'in_octets': '6003450707',
                                                                                                                  'in_ucast_pkts': '53005179',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6782844',
                                                                                                                  'out_octets': '305934531387',
                                                                                                                  'out_ucast_pkts': '212107857',
                                                                                                                  'poller_hash': 'MTE5NEQ3RHBlZXJpbnRmcG9uIDIxNjIwNw=='},
                                                                                                          {       'in_nucast_pkts': '1546413',
                                                                                                                  'in_octets': '15253696280',
                                                                                                                  'in_ucast_pkts': '199415474',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77945071',
                                                                                                                  'out_octets': '1132681773845',
                                                                                                                  'out_ucast_pkts': '784197435',
                                                                                                                  'poller_hash': 'MTI3NDA1QnBlZXJpbnRmcG9uIDIxNzE1MA=='},
                                                                                                          {       'in_nucast_pkts': '1567792',
                                                                                                                  'in_octets': '9911340947',
                                                                                                                  'in_ucast_pkts': '126004591',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '79266059',
                                                                                                                  'out_octets': '678499820474',
                                                                                                                  'out_ucast_pkts': '460319462',
                                                                                                                  'poller_hash': 'MTE5NEU3MnBlZXJpbnRmcG9uIDIxNzE2MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '72940963966',
                                                                                                                  'in_ucast_pkts': '8800901096',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '122',
                                                                                                                  'out_octets': '286841383782',
                                                                                                                  'out_ucast_pkts': '365308984',
                                                                                                                  'poller_hash': 'MTI3NTA3OHBlZXJpbnRmcG9uIDIxODI1Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NTBDQ3BlZXJpbnRmcG9uIDIyNDY5OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3117635788',
                                                                                                                  'in_ucast_pkts': '27381705',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '23672633929',
                                                                                                                  'out_ucast_pkts': '87456778',
                                                                                                                  'poller_hash': 'MTI3NDkxN3BlZXJpbnRmcG9uIDIzNDU3NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '80146941947',
                                                                                                                  'in_ucast_pkts': '154730638',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '643041138373',
                                                                                                                  'out_ucast_pkts': '558030777',
                                                                                                                  'poller_hash': 'MTBBODcyQ3BlZXJpbnRmcG9uIDI1MDMzNw=='},
                                                                                                          {       'in_nucast_pkts': '45302',
                                                                                                                  'in_octets': '7885467821',
                                                                                                                  'in_ucast_pkts': '41625289',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5480457',
                                                                                                                  'out_octets': '294262623869',
                                                                                                                  'out_ucast_pkts': '202541461',
                                                                                                                  'poller_hash': 'MTI5NUVCQnBlZXJpbnRmcG9uIDI4MjM2Mw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '13/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '182784702541',
                                                                                         'out_octets': '303728079182418',
                                                                                         'out_ucast_pkts': '41584481361',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '13/8',
                                                                                                                 'id': '13/8',
                                                                                                                 'links': [       {       'connected_time': '15498553',
                                                                                                                                          'fiber_length': 9511,
                                                                                                                                          'poller_hash': 'RUJCREIwcGVlcnBvbnBvbiAxODY1NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '21816100',
                                                                                                                                          'fiber_length': 11880,
                                                                                                                                          'poller_hash': 'REZGRTEycGVlcnBvbnBvbiAxODY1Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '70353940',
                                                                                                                                          'fiber_length': 11246,
                                                                                                                                          'poller_hash': 'QzgzMTJDcGVlcnBvbnBvbiAxODY1Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '21816833',
                                                                                                                                          'fiber_length': 10640,
                                                                                                                                          'poller_hash': 'QzgzMTUwcGVlcnBvbnBvbiAxODY1ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '12508517',
                                                                                                                                          'fiber_length': 12067,
                                                                                                                                          'poller_hash': 'QzgzMEQ5cGVlcnBvbnBvbiAxODY1OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '10358766',
                                                                                                                                          'fiber_length': 14984,
                                                                                                                                          'poller_hash': 'QzgxRTVDcGVlcnBvbnBvbiAyMDY5NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '27989217',
                                                                                                                                          'fiber_length': 12265,
                                                                                                                                          'poller_hash': 'MTE5NTE1Q3BlZXJwb25wb24gMjEyNzEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '3114073',
                                                                                                                                          'fiber_length': 15649,
                                                                                                                                          'poller_hash': 'MTE5NTFBNXBlZXJwb25wb24gMjE1NjQx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '3114081',
                                                                                                                                          'fiber_length': 15371,
                                                                                                                                          'poller_hash': 'MTE5NEQ3RHBlZXJwb25wb24gMjE2MjA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '1911312',
                                                                                                                                          'fiber_length': 15190,
                                                                                                                                          'poller_hash': 'MTI3NDA1QnBlZXJwb25wb24gMjE3MTUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '1910110',
                                                                                                                                          'fiber_length': 15245,
                                                                                                                                          'poller_hash': 'MTE5NEU3MnBlZXJwb25wb24gMjE3MTYw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '15560080',
                                                                                                                                          'fiber_length': 8733,
                                                                                                                                          'poller_hash': 'MTI3NTA3OHBlZXJwb25wb24gMjE4MjU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '719770',
                                                                                                                                          'fiber_length': 10129,
                                                                                                                                          'poller_hash': 'MTI3NTBDQ3BlZXJwb25wb24gMjI0Njk5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '3114013',
                                                                                                                                          'fiber_length': 12577,
                                                                                                                                          'poller_hash': 'MTI3NDkxN3BlZXJwb25wb24gMjM0NTc0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '3114024',
                                                                                                                                          'fiber_length': 16382,
                                                                                                                                          'poller_hash': 'MTBBODcyQ3BlZXJwb25wb24gMjUwMzM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085},
                                                                                                                                  {       'connected_time': '2530205',
                                                                                                                                          'fiber_length': 13858,
                                                                                                                                          'poller_hash': 'MTI5NUVCQnBlZXJwb25wb24gMjgyMzYz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.085}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '13/8',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'NDcxOTA1MDA2NDE1',
                                                           'serial_number': '471905006415',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991830'},
                                                   {       'id': '14',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '14/1',
                                                                                         'in_discards': '2566',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1084595',
                                                                                         'in_octets': '3528161451478',
                                                                                         'in_ucast_pkts': '18950159770',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '464542793853',
                                                                                                                  'in_ucast_pkts': '1443501792',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '116',
                                                                                                                  'out_octets': '6941730523547',
                                                                                                                  'out_ucast_pkts': '5044044823',
                                                                                                                  'poller_hash': 'MTI3NEEyMnBlZXJpbnRmcG9uIDIzMzU4Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '38732215',
                                                                                                                  'in_ucast_pkts': '60853',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '160',
                                                                                                                  'out_octets': '12605121035',
                                                                                                                  'out_ucast_pkts': '54207661',
                                                                                                                  'poller_hash': 'MTI3NEE1QXBlZXJpbnRmcG9uIDIzMzcwMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDVEOXBlZXJpbnRmcG9uIDIzMzcxMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '10838262874',
                                                                                                                  'in_ucast_pkts': '25809481',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '160',
                                                                                                                  'out_octets': '88187343227',
                                                                                                                  'out_ucast_pkts': '124364653',
                                                                                                                  'poller_hash': 'MTI3NEE1MnBlZXJpbnRmcG9uIDIzNDA4MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '74440797396',
                                                                                                                  'in_ucast_pkts': '262101534',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '116',
                                                                                                                  'out_octets': '726394569519',
                                                                                                                  'out_ucast_pkts': '584819971',
                                                                                                                  'poller_hash': 'MTI3NEE1NXBlZXJpbnRmcG9uIDIzNDA5MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '174264850412',
                                                                                                                  'in_ucast_pkts': '913822117',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '116',
                                                                                                                  'out_octets': '4668963075941',
                                                                                                                  'out_ucast_pkts': '3494046631',
                                                                                                                  'poller_hash': 'MTI3NEE0QXBlZXJpbnRmcG9uIDIzNDM5Nw=='},
                                                                                                          {       'in_nucast_pkts': '3467660',
                                                                                                                  'in_octets': '236416050611',
                                                                                                                  'in_ucast_pkts': '3235405641',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '192261729',
                                                                                                                  'out_octets': '19191505275061',
                                                                                                                  'out_ucast_pkts': '12939750844',
                                                                                                                  'poller_hash': 'MTI3NEE2RnBlZXJpbnRmcG9uIDIzNDUzNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '480631280405',
                                                                                                                  'in_ucast_pkts': '93978306305',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '180',
                                                                                                                  'out_octets': '11132308048705',
                                                                                                                  'out_ucast_pkts': '7863239018',
                                                                                                                  'poller_hash': 'MTI3NEE0MnBlZXJpbnRmcG9uIDIzNDkzNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '149211246104',
                                                                                                                  'in_ucast_pkts': '1155771524',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '140',
                                                                                                                  'out_octets': '6924472655554',
                                                                                                                  'out_ucast_pkts': '4885641821',
                                                                                                                  'poller_hash': 'MTI3NDlBN3BlZXJpbnRmcG9uIDIzNjI2OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDFBMHBlZXJpbnRmcG9uIDIzOTIwMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '64422106433',
                                                                                                                  'in_ucast_pkts': '412556170',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '146',
                                                                                                                  'out_octets': '1895065204591',
                                                                                                                  'out_ucast_pkts': '1408615954',
                                                                                                                  'poller_hash': 'MTI3NDFCMHBlZXJpbnRmcG9uIDI0MDU4MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDMwQ3BlZXJpbnRmcG9uIDI0MTA0NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '398397587997',
                                                                                                                  'in_ucast_pkts': '40407305686',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '60',
                                                                                                                  'out_octets': '9438877424886',
                                                                                                                  'out_ucast_pkts': '7149563306',
                                                                                                                  'poller_hash': 'MTI3NDBDOXBlZXJpbnRmcG9uIDI0MTc1NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '167933331365',
                                                                                                                  'in_ucast_pkts': '786173351',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '80',
                                                                                                                  'out_octets': '3551815995783',
                                                                                                                  'out_ucast_pkts': '2577934097',
                                                                                                                  'poller_hash': 'MTI3NDMyM3BlZXJpbnRmcG9uIDI0MjkwNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDMzQXBlZXJpbnRmcG9uIDI0Njg3MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3770766415',
                                                                                                                  'in_ucast_pkts': '3803495',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7685956476',
                                                                                                                  'out_ucast_pkts': '23524100',
                                                                                                                  'poller_hash': 'MTI3NDAyQnBlZXJpbnRmcG9uIDI0Njg4MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REI0MXBlZXJpbnRmcG9uIDI0OTUyOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '247998334707',
                                                                                                                  'in_ucast_pkts': '349203323',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1487471332042',
                                                                                                                  'out_ucast_pkts': '1158896053',
                                                                                                                  'poller_hash': 'MTI5NzNDRnBlZXJpbnRmcG9uIDI2Mzk0MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '29169391521',
                                                                                                                  'in_ucast_pkts': '261558993',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '894772183788',
                                                                                                                  'out_ucast_pkts': '650148684',
                                                                                                                  'poller_hash': 'MTI5NzQ1RHBlZXJpbnRmcG9uIDI2NDM5MQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '14/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '395350142',
                                                                                         'out_octets': '86887711024536',
                                                                                         'out_ucast_pkts': '62430207360',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/1',
                                                                                                                 'id': '14/1',
                                                                                                                 'links': [       {       'connected_time': '688591',
                                                                                                                                          'fiber_length': 10641,
                                                                                                                                          'poller_hash': 'MTI3NEEyMnBlZXJwb25wb24gMjMzNTgz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '4674396',
                                                                                                                                          'fiber_length': 10556,
                                                                                                                                          'poller_hash': 'MTI3NEE1QXBlZXJwb25wb24gMjMzNzAz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '4673060',
                                                                                                                                          'fiber_length': 10444,
                                                                                                                                          'poller_hash': 'MTI3NDVEOXBlZXJwb25wb24gMjMzNzEz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '4229731',
                                                                                                                                          'fiber_length': 10640,
                                                                                                                                          'poller_hash': 'MTI3NEE1MnBlZXJwb25wb24gMjM0MDgw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '688592',
                                                                                                                                          'fiber_length': 10654,
                                                                                                                                          'poller_hash': 'MTI3NEE1NXBlZXJwb25wb24gMjM0MDkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '688588',
                                                                                                                                          'fiber_length': 10535,
                                                                                                                                          'poller_hash': 'MTI3NEE0QXBlZXJwb25wb24gMjM0Mzk3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '3624827',
                                                                                                                                          'fiber_length': 10571,
                                                                                                                                          'poller_hash': 'MTI3NEE2RnBlZXJwb25wb24gMjM0NTM0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '1880134',
                                                                                                                                          'fiber_length': 11199,
                                                                                                                                          'poller_hash': 'MTI3NEE0MnBlZXJwb25wb24gMjM0OTM0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '950833',
                                                                                                                                          'fiber_length': 10664,
                                                                                                                                          'poller_hash': 'MTI3NDlBN3BlZXJwb25wb24gMjM2MjY5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '73629940',
                                                                                                                                          'fiber_length': 10661,
                                                                                                                                          'poller_hash': 'MTI3NDFBMHBlZXJwb25wb24gMjM5MjAx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '72585550',
                                                                                                                                          'fiber_length': 12002,
                                                                                                                                          'poller_hash': 'MTI3NDFCMHBlZXJwb25wb24gMjQwNTgx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '72510083',
                                                                                                                                          'fiber_length': 11081,
                                                                                                                                          'poller_hash': 'MTI3NDMwQ3BlZXJwb25wb24gMjQxMDQ1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '688584',
                                                                                                                                          'fiber_length': 11475,
                                                                                                                                          'poller_hash': 'MTI3NDBDOXBlZXJwb25wb24gMjQxNzU1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '688591',
                                                                                                                                          'fiber_length': 10649,
                                                                                                                                          'poller_hash': 'MTI3NDMyM3BlZXJwb25wb24gMjQyOTA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '28172215',
                                                                                                                                          'fiber_length': 11514,
                                                                                                                                          'poller_hash': 'MTI3NDMzQXBlZXJwb25wb24gMjQ2ODcw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '68350753',
                                                                                                                                          'fiber_length': 11599,
                                                                                                                                          'poller_hash': 'MTI3NDAyQnBlZXJwb25wb24gMjQ2ODgw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '64910571',
                                                                                                                                          'fiber_length': 11202,
                                                                                                                                          'poller_hash': 'MTE3REI0MXBlZXJwb25wb24gMjQ5NTI4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '688587',
                                                                                                                                          'fiber_length': 10513,
                                                                                                                                          'poller_hash': 'MTI5NzNDRnBlZXJwb25wb24gMjYzOTQx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -30.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671},
                                                                                                                                  {       'connected_time': '674969',
                                                                                                                                          'fiber_length': 11689,
                                                                                                                                          'poller_hash': 'MTI5NzQ1RHBlZXJwb25wb24gMjY0Mzkx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.671}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/1',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '14/2',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '545',
                                                                                         'in_octets': '9293504537767',
                                                                                         'in_ucast_pkts': '23800972009',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDlGQXBlZXJpbnRmcG9uIDIzNDg5NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '6818502798',
                                                                                                                  'in_ucast_pkts': '36093911',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '160',
                                                                                                                  'out_octets': '89767618020',
                                                                                                                  'out_ucast_pkts': '119692513',
                                                                                                                  'poller_hash': 'MTI3NEIwRnBlZXJpbnRmcG9uIDIzNTEwMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDlGMXBlZXJpbnRmcG9uIDIzNTc1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '758793497519',
                                                                                                                  'in_ucast_pkts': '3117822550',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '12702597764202',
                                                                                                                  'out_ucast_pkts': '9081113400',
                                                                                                                  'poller_hash': 'MTI3NEFBNXBlZXJpbnRmcG9uIDIzNjExMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '445278542493',
                                                                                                                  'in_ucast_pkts': '79896118904',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '9003182060815',
                                                                                                                  'out_ucast_pkts': '6767623004',
                                                                                                                  'poller_hash': 'MTI3NEEzM3BlZXJpbnRmcG9uIDIzNjIzOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '153185200453',
                                                                                                                  'in_ucast_pkts': '177058364',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '140',
                                                                                                                  'out_octets': '31674399988',
                                                                                                                  'out_ucast_pkts': '110683356',
                                                                                                                  'poller_hash': 'MTI3NEIyMHBlZXJpbnRmcG9uIDIzODQxMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2683485248012',
                                                                                                                  'in_ucast_pkts': '41987091382',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '140',
                                                                                                                  'out_octets': '6842788804754',
                                                                                                                  'out_ucast_pkts': '6261696302',
                                                                                                                  'poller_hash': 'MTI3NEFDMnBlZXJpbnRmcG9uIDIzODYyOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '135247222800',
                                                                                                                  'in_ucast_pkts': '344841244',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '1152804177711',
                                                                                                                  'out_ucast_pkts': '964985480',
                                                                                                                  'poller_hash': 'MTI3NDBCNHBlZXJpbnRmcG9uIDI0MDE5MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '130573504830',
                                                                                                                  'in_ucast_pkts': '38949640492',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '116',
                                                                                                                  'out_octets': '987543452225',
                                                                                                                  'out_ucast_pkts': '830701433',
                                                                                                                  'poller_hash': 'MTI3NDMwNHBlZXJpbnRmcG9uIDI0MDIwMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '502385854099',
                                                                                                                  'in_ucast_pkts': '31867660126',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '110',
                                                                                                                  'out_octets': '10838859387546',
                                                                                                                  'out_ucast_pkts': '6163479683',
                                                                                                                  'poller_hash': 'MTI3NDIxNHBlZXJpbnRmcG9uIDI0MTcxNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '121829754625',
                                                                                                                  'in_ucast_pkts': '476481414',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '40',
                                                                                                                  'out_octets': '2350996311535',
                                                                                                                  'out_ucast_pkts': '1762672468',
                                                                                                                  'poller_hash': 'MTI3NDE5M3BlZXJpbnRmcG9uIDI0MjM2Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '129771244443',
                                                                                                                  'in_ucast_pkts': '323933596',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '110',
                                                                                                                  'out_octets': '351863410679',
                                                                                                                  'out_ucast_pkts': '491119131',
                                                                                                                  'poller_hash': 'MTI3NDA5N3BlZXJpbnRmcG9uIDI0Mjg4Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDJEMnBlZXJpbnRmcG9uIDI0Mjg5Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '208861945788',
                                                                                                                  'in_ucast_pkts': '425530294',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '130',
                                                                                                                  'out_octets': '1171408268363',
                                                                                                                  'out_ucast_pkts': '1055814527',
                                                                                                                  'poller_hash': 'MTI5NkZFMHBlZXJpbnRmcG9uIDI0MzcxMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '219567260068',
                                                                                                                  'in_ucast_pkts': '64831066406',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '904036045806',
                                                                                                                  'out_ucast_pkts': '787183517',
                                                                                                                  'poller_hash': 'MTI3NDZFMXBlZXJpbnRmcG9uIDI0OTA5Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '112430510647',
                                                                                                                  'in_ucast_pkts': '315316346',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1783964377144',
                                                                                                                  'out_ucast_pkts': '1299348899',
                                                                                                                  'poller_hash': 'MTE3REE2RHBlZXJpbnRmcG9uIDI1NDAwMg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '14/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '171679873',
                                                                                         'out_octets': '77283157742909',
                                                                                         'out_ucast_pkts': '60554732083',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/2',
                                                                                                                 'id': '14/2',
                                                                                                                 'links': [       {       'connected_time': '3036342',
                                                                                                                                          'fiber_length': 12104,
                                                                                                                                          'poller_hash': 'MTI3NDlGQXBlZXJwb25wb24gMjM0ODk0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '2776283',
                                                                                                                                          'fiber_length': 12190,
                                                                                                                                          'poller_hash': 'MTI3NEIwRnBlZXJwb25wb24gMjM1MTAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '77290',
                                                                                                                                          'fiber_length': 10768,
                                                                                                                                          'poller_hash': 'MTI3NDlGMXBlZXJwb25wb24gMjM1NzU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '688586',
                                                                                                                                          'fiber_length': 12069,
                                                                                                                                          'poller_hash': 'MTI3NEFBNXBlZXJwb25wb24gMjM2MTEw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '688595',
                                                                                                                                          'fiber_length': 12923,
                                                                                                                                          'poller_hash': 'MTI3NEEzM3BlZXJwb25wb24gMjM2MjM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '74394732',
                                                                                                                                          'fiber_length': 12100,
                                                                                                                                          'poller_hash': 'MTI3NEIyMHBlZXJwb25wb24gMjM4NDEx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '19073558',
                                                                                                                                          'fiber_length': 12721,
                                                                                                                                          'poller_hash': 'MTI3NEFDMnBlZXJwb25wb24gMjM4NjI5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '688587',
                                                                                                                                          'fiber_length': 11862,
                                                                                                                                          'poller_hash': 'MTI3NDBCNHBlZXJwb25wb24gMjQwMTkx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '688591',
                                                                                                                                          'fiber_length': 12362,
                                                                                                                                          'poller_hash': 'MTI3NDMwNHBlZXJwb25wb24gMjQwMjAx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '71976643',
                                                                                                                                          'fiber_length': 11323,
                                                                                                                                          'poller_hash': 'MTI3NDIxNHBlZXJwb25wb24gMjQxNzE0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '688587',
                                                                                                                                          'fiber_length': 11877,
                                                                                                                                          'poller_hash': 'MTI3NDE5M3BlZXJwb25wb24gMjQyMzYy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '70767365',
                                                                                                                                          'fiber_length': 11316,
                                                                                                                                          'poller_hash': 'MTI3NDA5N3BlZXJwb25wb24gMjQyODg3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '70765845',
                                                                                                                                          'fiber_length': 11193,
                                                                                                                                          'poller_hash': 'MTI3NDJEMnBlZXJwb25wb24gMjQyODk3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '70164143',
                                                                                                                                          'fiber_length': 11582,
                                                                                                                                          'poller_hash': 'MTI5NkZFMHBlZXJwb25wb24gMjQzNzEx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '688587',
                                                                                                                                          'fiber_length': 10517,
                                                                                                                                          'poller_hash': 'MTI3NDZFMXBlZXJwb25wb24gMjQ5MDkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444},
                                                                                                                                  {       'connected_time': '688599',
                                                                                                                                          'fiber_length': 12125,
                                                                                                                                          'poller_hash': 'MTE3REE2RHBlZXJwb25wb24gMjU0MDAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.444}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/2',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '14/3',
                                                                                         'in_discards': '11889',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2422834',
                                                                                         'in_octets': '5219617897514',
                                                                                         'in_ucast_pkts': '23518755121',
                                                                                         'links': [       {       'in_nucast_pkts': '12483',
                                                                                                                  'in_octets': '2761326444',
                                                                                                                  'in_ucast_pkts': '11123402',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465054',
                                                                                                                  'out_octets': '105724977184',
                                                                                                                  'out_ucast_pkts': '75653884',
                                                                                                                  'poller_hash': 'MTI3NDk5Q3BlZXJpbnRmcG9uIDIzNDg0NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '285629702355',
                                                                                                                  'in_ucast_pkts': '2333892997',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '140',
                                                                                                                  'out_octets': '6593979666414',
                                                                                                                  'out_ucast_pkts': '4695613786',
                                                                                                                  'poller_hash': 'MTI3NEEwMHBlZXJpbnRmcG9uIDIzNDg3NA=='},
                                                                                                          {       'in_nucast_pkts': '11759',
                                                                                                                  'in_octets': '6051245042',
                                                                                                                  'in_ucast_pkts': '10702104',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465092',
                                                                                                                  'out_octets': '48754060746',
                                                                                                                  'out_ucast_pkts': '39295509',
                                                                                                                  'poller_hash': 'MTI3NEEwRHBlZXJpbnRmcG9uIDIzNDkwNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '33799128668',
                                                                                                                  'in_ucast_pkts': '122028668',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '140',
                                                                                                                  'out_octets': '484263124752',
                                                                                                                  'out_ucast_pkts': '432976305',
                                                                                                                  'poller_hash': 'MTI3NDk5M3BlZXJpbnRmcG9uIDIzNDkyNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '277474153289',
                                                                                                                  'in_ucast_pkts': '586933743',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '140',
                                                                                                                  'out_octets': '1257180743871',
                                                                                                                  'out_ucast_pkts': '1215484373',
                                                                                                                  'poller_hash': 'MTI3NEEzQ3BlZXJpbnRmcG9uIDIzNTExMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDlDNnBlZXJpbnRmcG9uIDIzNTQ1OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '237016063401',
                                                                                                                  'in_ucast_pkts': '916568546',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '116',
                                                                                                                  'out_octets': '4768234449930',
                                                                                                                  'out_ucast_pkts': '3943486606',
                                                                                                                  'poller_hash': 'MTI3NEExOXBlZXJpbnRmcG9uIDIzNjI1OQ=='},
                                                                                                          {       'in_nucast_pkts': '11693',
                                                                                                                  'in_octets': '2671073402',
                                                                                                                  'in_ucast_pkts': '9946455',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465228',
                                                                                                                  'out_octets': '52464130038',
                                                                                                                  'out_ucast_pkts': '42537550',
                                                                                                                  'poller_hash': 'MTI3NDIxN3BlZXJpbnRmcG9uIDI0MTY0NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '50401923205',
                                                                                                                  'in_ucast_pkts': '260125908',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '80',
                                                                                                                  'out_octets': '1079527518651',
                                                                                                                  'out_ucast_pkts': '801548605',
                                                                                                                  'poller_hash': 'MTI3NDJFMHBlZXJpbnRmcG9uIDI0Mjg1Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '140356023830',
                                                                                                                  'in_ucast_pkts': '385230903',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '80',
                                                                                                                  'out_octets': '1434262779427',
                                                                                                                  'out_ucast_pkts': '1236713339',
                                                                                                                  'poller_hash': 'MTI5NzA2NnBlZXJpbnRmcG9uIDI0Mjg2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '226829363679',
                                                                                                                  'in_ucast_pkts': '6117388714',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '60',
                                                                                                                  'out_octets': '6154284587277',
                                                                                                                  'out_ucast_pkts': '4374873111',
                                                                                                                  'poller_hash': 'MTI5NkZDQnBlZXJpbnRmcG9uIDI0Mjg3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '249691296053',
                                                                                                                  'in_ucast_pkts': '1161133262',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '90',
                                                                                                                  'out_octets': '3137843610524',
                                                                                                                  'out_ucast_pkts': '2431553918',
                                                                                                                  'poller_hash': 'MTI3NDA3OXBlZXJpbnRmcG9uIDI0NTIwMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '346858204779',
                                                                                                                  'in_ucast_pkts': '53248524123',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '5313810094830',
                                                                                                                  'out_ucast_pkts': '3739705772',
                                                                                                                  'poller_hash': 'MTBBODZFRHBlZXJpbnRmcG9uIDI0OTQ5OA=='},
                                                                                                          {       'in_nucast_pkts': '19576',
                                                                                                                  'in_octets': '9976103705',
                                                                                                                  'in_ucast_pkts': '18933589',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2190700',
                                                                                                                  'out_octets': '71370889505',
                                                                                                                  'out_ucast_pkts': '58999481',
                                                                                                                  'poller_hash': 'MTBBODZBMXBlZXJpbnRmcG9uIDI1MDEwNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '40308285184',
                                                                                                                  'in_ucast_pkts': '144508462',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1053277925633',
                                                                                                                  'out_ucast_pkts': '839647390',
                                                                                                                  'poller_hash': 'MTE5MkZFOXBlZXJpbnRmcG9uIDI1ODIzMw=='},
                                                                                                          {       'in_nucast_pkts': '414940',
                                                                                                                  'in_octets': '108374894355',
                                                                                                                  'in_ucast_pkts': '722802903',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '50077548',
                                                                                                                  'out_octets': '2498861453360',
                                                                                                                  'out_ucast_pkts': '1971032579',
                                                                                                                  'poller_hash': 'MTI3NDhCMXBlZXJpbnRmcG9uIDI2MDM2NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDg0NHBlZXJpbnRmcG9uIDI2MjgxNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '21052005427',
                                                                                                                  'in_ucast_pkts': '107020169',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '710772122918',
                                                                                                                  'out_ucast_pkts': '551345528',
                                                                                                                  'poller_hash': 'MTI3NDc4RHBlZXJpbnRmcG9uIDI2Mzg2Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '14/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '171679876',
                                                                                         'out_octets': '85752913147781',
                                                                                         'out_ucast_pkts': '64284011150',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/3',
                                                                                                                 'id': '14/3',
                                                                                                                 'links': [       {       'connected_time': '688599',
                                                                                                                                          'fiber_length': 11442,
                                                                                                                                          'poller_hash': 'MTI3NDk5Q3BlZXJwb25wb24gMjM0ODQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '3363277',
                                                                                                                                          'fiber_length': 11490,
                                                                                                                                          'poller_hash': 'MTI3NEEwMHBlZXJwb25wb24gMjM0ODc0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688595',
                                                                                                                                          'fiber_length': 10619,
                                                                                                                                          'poller_hash': 'MTI3NEEwRHBlZXJwb25wb24gMjM0OTA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '2948034',
                                                                                                                                          'fiber_length': 10638,
                                                                                                                                          'poller_hash': 'MTI3NDk5M3BlZXJwb25wb24gMjM0OTI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '28280185',
                                                                                                                                          'fiber_length': 11468,
                                                                                                                                          'poller_hash': 'MTI3NEEzQ3BlZXJwb25wb24gMjM1MTEy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '2258054',
                                                                                                                                          'fiber_length': 11664,
                                                                                                                                          'poller_hash': 'MTI3NDlDNnBlZXJwb25wb24gMjM1NDU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688593',
                                                                                                                                          'fiber_length': 10655,
                                                                                                                                          'poller_hash': 'MTI3NEExOXBlZXJwb25wb24gMjM2MjU5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688598',
                                                                                                                                          'fiber_length': 10581,
                                                                                                                                          'poller_hash': 'MTI3NDIxN3BlZXJwb25wb24gMjQxNjQ0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688596',
                                                                                                                                          'fiber_length': 12393,
                                                                                                                                          'poller_hash': 'MTI3NDJFMHBlZXJwb25wb24gMjQyODU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688594',
                                                                                                                                          'fiber_length': 10438,
                                                                                                                                          'poller_hash': 'MTI5NzA2NnBlZXJwb25wb24gMjQyODY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688587',
                                                                                                                                          'fiber_length': 11975,
                                                                                                                                          'poller_hash': 'MTI5NkZDQnBlZXJwb25wb24gMjQyODc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '10385489',
                                                                                                                                          'fiber_length': 11566,
                                                                                                                                          'poller_hash': 'MTI3NDA3OXBlZXJwb25wb24gMjQ1MjAx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688595',
                                                                                                                                          'fiber_length': 10661,
                                                                                                                                          'poller_hash': 'MTBBODZFRHBlZXJwb25wb24gMjQ5NDk4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '1023956',
                                                                                                                                          'fiber_length': 12762,
                                                                                                                                          'poller_hash': 'MTBBODZBMXBlZXJwb25wb24gMjUwMTA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688594',
                                                                                                                                          'fiber_length': 11694,
                                                                                                                                          'poller_hash': 'MTE5MkZFOXBlZXJwb25wb24gMjU4MjMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '21777412',
                                                                                                                                          'fiber_length': 13616,
                                                                                                                                          'poller_hash': 'MTI3NDhCMXBlZXJwb25wb24gMjYwMzY0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '4750292',
                                                                                                                                          'fiber_length': 11534,
                                                                                                                                          'poller_hash': 'MTI3NDg0NHBlZXJwb25wb24gMjYyODE3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267},
                                                                                                                                  {       'connected_time': '688591',
                                                                                                                                          'fiber_length': 11520,
                                                                                                                                          'poller_hash': 'MTI3NDc4RHBlZXJwb25wb24gMjYzODY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.267}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/3',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '14/4',
                                                                                         'in_discards': '744',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '642783',
                                                                                         'in_octets': '621728834248',
                                                                                         'in_ucast_pkts': '3282798094',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3M0Y1NXBlZXJpbnRmcG9uIDI0MTgxNQ=='},
                                                                                                          {       'in_nucast_pkts': '676913',
                                                                                                                  'in_octets': '604907484750',
                                                                                                                  'in_ucast_pkts': '3258463965',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '83705938',
                                                                                                                  'out_octets': '11271252493136',
                                                                                                                  'out_ucast_pkts': '7963987273',
                                                                                                                  'poller_hash': 'MTI3NDMwOXBlZXJpbnRmcG9uIDI0NzEwNQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '14/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '105599789',
                                                                                         'out_octets': '11330136748957',
                                                                                         'out_ucast_pkts': '8002807539',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC80',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/4',
                                                                                                                 'id': '14/4',
                                                                                                                 'links': [       {       'connected_time': '23692706',
                                                                                                                                          'fiber_length': 14627,
                                                                                                                                          'poller_hash': 'MTI3M0Y1NXBlZXJwb25wb24gMjQxODE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.573},
                                                                                                                                  {       'connected_time': '67830839',
                                                                                                                                          'fiber_length': 14222,
                                                                                                                                          'poller_hash': 'MTI3NDMwOXBlZXJwb25wb24gMjQ3MTA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.573}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/4',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC80',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '14/5',
                                                                                         'in_discards': '1909',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '555675',
                                                                                         'in_octets': '2756142017109',
                                                                                         'in_ucast_pkts': '9633690783',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '195176625627',
                                                                                                                  'in_ucast_pkts': '1251500223',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '65',
                                                                                                                  'out_octets': '4238301800538',
                                                                                                                  'out_ucast_pkts': '3015790347',
                                                                                                                  'poller_hash': 'MTI3NDE5NnBlZXJpbnRmcG9uIDI0MzA3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDAzRnBlZXJpbnRmcG9uIDI0MzEwNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDJFRHBlZXJpbnRmcG9uIDI0MzcyMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '86063895666',
                                                                                                                  'in_ucast_pkts': '1086833623',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '25',
                                                                                                                  'out_octets': '6944086628284',
                                                                                                                  'out_ucast_pkts': '321015109',
                                                                                                                  'poller_hash': 'MTI3NEEzRnBlZXJpbnRmcG9uIDI0ODA4OA=='},
                                                                                                          {       'in_nucast_pkts': '587651',
                                                                                                                  'in_octets': '2102738826256',
                                                                                                                  'in_ucast_pkts': '5245381888',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '73043118',
                                                                                                                  'out_octets': '13385457878614',
                                                                                                                  'out_ucast_pkts': '10655604236',
                                                                                                                  'poller_hash': 'MTE3REFCN3BlZXJpbnRmcG9uIDI1MDI2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45457928',
                                                                                                                  'in_ucast_pkts': '70739',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '25',
                                                                                                                  'out_octets': '8083094952',
                                                                                                                  'out_ucast_pkts': '41643667',
                                                                                                                  'poller_hash': 'MTE3REI0NXBlZXJpbnRmcG9uIDI1MDI3Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzM5MHBlZXJpbnRmcG9uIDI1NzczOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '84890664013',
                                                                                                                  'in_ucast_pkts': '303963365',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '25',
                                                                                                                  'out_octets': '1470573590492',
                                                                                                                  'out_ucast_pkts': '1125143345',
                                                                                                                  'poller_hash': 'MTI5NzM2MHBlZXJpbnRmcG9uIDI1ODMxNQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '14/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '109361365',
                                                                                         'out_octets': '33526494819820',
                                                                                         'out_ucast_pkts': '24661078584',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/5',
                                                                                                                 'id': '14/5',
                                                                                                                 'links': [       {       'connected_time': '538597',
                                                                                                                                          'fiber_length': 18065,
                                                                                                                                          'poller_hash': 'MTI3NDE5NnBlZXJwb25wb24gMjQzMDc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729},
                                                                                                                                  {       'connected_time': '70334542',
                                                                                                                                          'fiber_length': 18628,
                                                                                                                                          'poller_hash': 'MTI3NDAzRnBlZXJwb25wb24gMjQzMTA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729},
                                                                                                                                  {       'connected_time': '70158734',
                                                                                                                                          'fiber_length': 17083,
                                                                                                                                          'poller_hash': 'MTI3NDJFRHBlZXJwb25wb24gMjQzNzIx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729},
                                                                                                                                  {       'connected_time': '67051846',
                                                                                                                                          'fiber_length': 17265,
                                                                                                                                          'poller_hash': 'MTI3NEEzRnBlZXJwb25wb24gMjQ4MDg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729},
                                                                                                                                  {       'connected_time': '64029754',
                                                                                                                                          'fiber_length': 19116,
                                                                                                                                          'poller_hash': 'MTE3REFCN3BlZXJwb25wb24gMjUwMjY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729},
                                                                                                                                  {       'connected_time': '64028908',
                                                                                                                                          'fiber_length': 19300,
                                                                                                                                          'poller_hash': 'MTE3REI0NXBlZXJwb25wb24gMjUwMjc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729},
                                                                                                                                  {       'connected_time': '23590596',
                                                                                                                                          'fiber_length': 18673,
                                                                                                                                          'poller_hash': 'MTI5NzM5MHBlZXJwb25wb24gMjU3NzM4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729},
                                                                                                                                  {       'connected_time': '538604',
                                                                                                                                          'fiber_length': 16497,
                                                                                                                                          'poller_hash': 'MTI5NzM2MHBlZXJwb25wb24gMjU4MzE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.729}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/5',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '14/6',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '161',
                                                                                         'in_octets': '1926723581521',
                                                                                         'in_ucast_pkts': '3507522627',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '618487218507',
                                                                                                                  'in_ucast_pkts': '61021579311',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '2571588353317',
                                                                                                                  'out_ucast_pkts': '2269650012',
                                                                                                                  'poller_hash': 'MTI3NDMxN3BlZXJpbnRmcG9uIDIzOTk0Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '198953239263',
                                                                                                                  'in_ucast_pkts': '4579553588',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '530802178326',
                                                                                                                  'out_ucast_pkts': '548698936',
                                                                                                                  'poller_hash': 'MTI3NDJFNHBlZXJpbnRmcG9uIDIzOTk1Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '46170742',
                                                                                                                  'in_ucast_pkts': '73938',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '146',
                                                                                                                  'out_octets': '12549329139',
                                                                                                                  'out_ucast_pkts': '53317212',
                                                                                                                  'poller_hash': 'MTI3NDFCQnBlZXJpbnRmcG9uIDI0MDI0MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDJGNnBlZXJpbnRmcG9uIDI0MDUzOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDI5QnBlZXJpbnRmcG9uIDI0MDU3MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDIyMHBlZXJpbnRmcG9uIDI0MTQ1NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '47466347698',
                                                                                                                  'in_ucast_pkts': '162539394',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '110',
                                                                                                                  'out_octets': '575876330730',
                                                                                                                  'out_ucast_pkts': '534309945',
                                                                                                                  'poller_hash': 'MTI3NEExMXBlZXJpbnRmcG9uIDI0MTU4NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '158620178358',
                                                                                                                  'in_ucast_pkts': '361389177',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '110',
                                                                                                                  'out_octets': '541859051869',
                                                                                                                  'out_ucast_pkts': '597152415',
                                                                                                                  'poller_hash': 'MTI3NEE5QXBlZXJpbnRmcG9uIDI0MTkxNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '127129670',
                                                                                                                  'in_ucast_pkts': '375689',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '25',
                                                                                                                  'out_octets': '10900815448',
                                                                                                                  'out_ucast_pkts': '48912134',
                                                                                                                  'poller_hash': 'MTI3NDFBMnBlZXJpbnRmcG9uIDI0NTg1Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEE0MHBlZXJpbnRmcG9uIDI0Nzc4NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '52902911989',
                                                                                                                  'in_ucast_pkts': '238322207',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1522235404086',
                                                                                                                  'out_ucast_pkts': '1157582200',
                                                                                                                  'poller_hash': 'MTI3NDc2MXBlZXJpbnRmcG9uIDI2MjkxNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDY5Q3BlZXJpbnRmcG9uIDI2NjA5OA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '14/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '116664167',
                                                                                         'out_octets': '8783066623214',
                                                                                         'out_ucast_pkts': '7894336825',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/6',
                                                                                                                 'id': '14/6',
                                                                                                                 'links': [       {       'connected_time': '688595',
                                                                                                                                          'fiber_length': 14764,
                                                                                                                                          'poller_hash': 'MTI3NDMxN3BlZXJwb25wb24gMjM5OTQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '1845110',
                                                                                                                                          'fiber_length': 13891,
                                                                                                                                          'poller_hash': 'MTI3NDJFNHBlZXJwb25wb24gMjM5OTU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '72680560',
                                                                                                                                          'fiber_length': 14932,
                                                                                                                                          'poller_hash': 'MTI3NDFCQnBlZXJwb25wb24gMjQwMjQx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '72602221',
                                                                                                                                          'fiber_length': 15489,
                                                                                                                                          'poller_hash': 'MTI3NDJGNnBlZXJwb25wb24gMjQwNTM4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '72597084',
                                                                                                                                          'fiber_length': 13062,
                                                                                                                                          'poller_hash': 'MTI3NDI5QnBlZXJwb25wb24gMjQwNTcx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '72411759',
                                                                                                                                          'fiber_length': 15536,
                                                                                                                                          'poller_hash': 'MTI3NDIyMHBlZXJwb25wb24gMjQxNDU1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '72160938',
                                                                                                                                          'fiber_length': 12941,
                                                                                                                                          'poller_hash': 'MTI3NEExMXBlZXJwb25wb24gMjQxNTg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '12029246',
                                                                                                                                          'fiber_length': 14448,
                                                                                                                                          'poller_hash': 'MTI3NEE5QXBlZXJwb25wb24gMjQxOTE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '68869275',
                                                                                                                                          'fiber_length': 13858,
                                                                                                                                          'poller_hash': 'MTI3NDFBMnBlZXJwb25wb24gMjQ1ODU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '23104090',
                                                                                                                                          'fiber_length': 16132,
                                                                                                                                          'poller_hash': 'MTI3NEE0MHBlZXJwb25wb24gMjQ3Nzg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '20565187',
                                                                                                                                          'fiber_length': 15505,
                                                                                                                                          'poller_hash': 'MTI3NDc2MXBlZXJwb25wb24gMjYyOTE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622},
                                                                                                                                  {       'connected_time': '17302188',
                                                                                                                                          'fiber_length': 15715,
                                                                                                                                          'poller_hash': 'MTI3NDY5Q3BlZXJwb25wb24gMjY2MDk4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.622}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/6',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '14/7',
                                                                                         'in_discards': '5965',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1654472',
                                                                                         'in_octets': '2097503341875',
                                                                                         'in_ucast_pkts': '12174197910',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '33003576664',
                                                                                                                  'in_ucast_pkts': '125593722',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '120',
                                                                                                                  'out_octets': '390041198654',
                                                                                                                  'out_ucast_pkts': '379114954',
                                                                                                                  'poller_hash': 'MTI3NDFCQ3BlZXJpbnRmcG9uIDIzOTAzOQ=='},
                                                                                                          {       'in_nucast_pkts': '11071',
                                                                                                                  'in_octets': '1883842790',
                                                                                                                  'in_ucast_pkts': '4351194',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1118803',
                                                                                                                  'out_octets': '23892596312',
                                                                                                                  'out_ucast_pkts': '19455282',
                                                                                                                  'poller_hash': 'MTI3NDlDMXBlZXJpbnRmcG9uIDIzOTU1MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDlDNXBlZXJpbnRmcG9uIDIzOTc1MQ=='},
                                                                                                          {       'in_nucast_pkts': '140549',
                                                                                                                  'in_octets': '53681278249',
                                                                                                                  'in_ucast_pkts': '136208253',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '16786562',
                                                                                                                  'out_octets': '361804620016',
                                                                                                                  'out_ucast_pkts': '320138669',
                                                                                                                  'poller_hash': 'MTI3NDJEOXBlZXJpbnRmcG9uIDIzOTkyNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '126087362924',
                                                                                                                  'in_ucast_pkts': '22039538596',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '81',
                                                                                                                  'out_octets': '2225236758916',
                                                                                                                  'out_ucast_pkts': '1761305511',
                                                                                                                  'poller_hash': 'MTI3NDJGNXBlZXJpbnRmcG9uIDI0MDY0OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96330232915',
                                                                                                                  'in_ucast_pkts': '291580994',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70',
                                                                                                                  'out_octets': '1439789343741',
                                                                                                                  'out_ucast_pkts': '1170818162',
                                                                                                                  'poller_hash': 'MTI3NEEwOHBlZXJpbnRmcG9uIDI0MTQ4OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '63448229131',
                                                                                                                  'in_ucast_pkts': '206747080',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70',
                                                                                                                  'out_octets': '629853116456',
                                                                                                                  'out_ucast_pkts': '587439378',
                                                                                                                  'poller_hash': 'MTI3NDA0Q3BlZXJpbnRmcG9uIDI0MTc0NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '88326949478',
                                                                                                                  'in_ucast_pkts': '506162564',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45',
                                                                                                                  'out_octets': '1994833514359',
                                                                                                                  'out_ucast_pkts': '1677927265',
                                                                                                                  'poller_hash': 'MTI3NDA2MXBlZXJpbnRmcG9uIDI0MTkwNQ=='},
                                                                                                          {       'in_nucast_pkts': '2822286',
                                                                                                                  'in_octets': '280912497924',
                                                                                                                  'in_ucast_pkts': '3866452063',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '174022001',
                                                                                                                  'out_octets': '23577958929019',
                                                                                                                  'out_ucast_pkts': '15880168294',
                                                                                                                  'poller_hash': 'MTI3NDBDOHBlZXJpbnRmcG9uIDI0MjEyNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDI5OHBlZXJpbnRmcG9uIDI0MjkzMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3M0Y2RnBlZXJpbnRmcG9uIDI0NDg5MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '170099689483',
                                                                                                                  'in_ucast_pkts': '371480380',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '50',
                                                                                                                  'out_octets': '780269856160',
                                                                                                                  'out_ucast_pkts': '730026288',
                                                                                                                  'poller_hash': 'MTI3NDA2N3BlZXJpbnRmcG9uIDI0ODUzMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '25791958147',
                                                                                                                  'in_ucast_pkts': '174277527',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '982855794984',
                                                                                                                  'out_ucast_pkts': '724702274',
                                                                                                                  'poller_hash': 'MTI3NDMzNnBlZXJpbnRmcG9uIDI0OTAzMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '115246062940',
                                                                                                                  'in_ucast_pkts': '30588253597',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '25',
                                                                                                                  'out_octets': '2190163003676',
                                                                                                                  'out_ucast_pkts': '1684268217',
                                                                                                                  'poller_hash': 'MTBBODZBQnBlZXJpbnRmcG9uIDI0OTEyMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTBBODY0MXBlZXJpbnRmcG9uIDI1MDMwNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDdENnBlZXJpbnRmcG9uIDI2MTM0Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDdBRnBlZXJpbnRmcG9uIDI2MjY3OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NkRFRHBlZXJpbnRmcG9uIDI2Mzk1MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE5MkZGNnBlZXJpbnRmcG9uIDI2NjQ4NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '14/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '361731953',
                                                                                         'out_octets': '59566811167142',
                                                                                         'out_ucast_pkts': '42864430920',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/7',
                                                                                                                 'id': '14/7',
                                                                                                                 'links': [       {       'connected_time': '73707903',
                                                                                                                                          'fiber_length': 11242,
                                                                                                                                          'poller_hash': 'MTI3NDFCQ3BlZXJwb25wb24gMjM5MDM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '523335',
                                                                                                                                          'fiber_length': 12154,
                                                                                                                                          'poller_hash': 'MTI3NDlDMXBlZXJwb25wb24gMjM5NTUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '73201399',
                                                                                                                                          'fiber_length': 12150,
                                                                                                                                          'poller_hash': 'MTI3NDlDNXBlZXJwb25wb24gMjM5NzUx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '7187005',
                                                                                                                                          'fiber_length': 12539,
                                                                                                                                          'poller_hash': 'MTI3NDJEOXBlZXJwb25wb24gMjM5OTI3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '3869134',
                                                                                                                                          'fiber_length': 11862,
                                                                                                                                          'poller_hash': 'MTI3NDJGNXBlZXJwb25wb24gMjQwNjQ4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '72168615',
                                                                                                                                          'fiber_length': 12289,
                                                                                                                                          'poller_hash': 'MTI3NEEwOHBlZXJwb25wb24gMjQxNDg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '71903261',
                                                                                                                                          'fiber_length': 11364,
                                                                                                                                          'poller_hash': 'MTI3NDA0Q3BlZXJwb25wb24gMjQxNzQ1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '24461472',
                                                                                                                                          'fiber_length': 11462,
                                                                                                                                          'poller_hash': 'MTI3NDA2MXBlZXJwb25wb24gMjQxOTA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '71289058',
                                                                                                                                          'fiber_length': 12061,
                                                                                                                                          'poller_hash': 'MTI3NDBDOHBlZXJwb25wb24gMjQyMTI1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '69143959',
                                                                                                                                          'fiber_length': 12108,
                                                                                                                                          'poller_hash': 'MTI3NDI5OHBlZXJwb25wb24gMjQyOTMw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '69388586',
                                                                                                                                          'fiber_length': 12061,
                                                                                                                                          'poller_hash': 'MTI3M0Y2RnBlZXJwb25wb24gMjQ0ODkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '66531890',
                                                                                                                                          'fiber_length': 11265,
                                                                                                                                          'poller_hash': 'MTI3NDA2N3BlZXJwb25wb24gMjQ4NTMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '523338',
                                                                                                                                          'fiber_length': 11219,
                                                                                                                                          'poller_hash': 'MTI3NDMzNnBlZXJwb25wb24gMjQ5MDMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '65768713',
                                                                                                                                          'fiber_length': 12029,
                                                                                                                                          'poller_hash': 'MTBBODZBQnBlZXJwb25wb24gMjQ5MTIz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '63961891',
                                                                                                                                          'fiber_length': 11878,
                                                                                                                                          'poller_hash': 'MTBBODY0MXBlZXJwb25wb24gMjUwMzA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '21621629',
                                                                                                                                          'fiber_length': 11465,
                                                                                                                                          'poller_hash': 'MTI3NDdENnBlZXJwb25wb24gMjYxMzQy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '20998295',
                                                                                                                                          'fiber_length': 11626,
                                                                                                                                          'poller_hash': 'MTI3NDdBRnBlZXJwb25wb24gMjYyNjc4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '18838847',
                                                                                                                                          'fiber_length': 12616,
                                                                                                                                          'poller_hash': 'MTI5NkRFRHBlZXJwb25wb24gMjYzOTUx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655},
                                                                                                                                  {       'connected_time': '16695529',
                                                                                                                                          'fiber_length': 11640,
                                                                                                                                          'poller_hash': 'MTE5MkZGNnBlZXJwb25wb24gMjY2NDg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.655}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/7',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '14/8',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3858070',
                                                                                         'in_octets': '963181260354',
                                                                                         'in_ucast_pkts': '6267931269',
                                                                                         'links': [       {       'in_nucast_pkts': '395278',
                                                                                                                  'in_octets': '298788193125',
                                                                                                                  'in_ucast_pkts': '1645732181',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70820682',
                                                                                                                  'out_octets': '7258814936712',
                                                                                                                  'out_ucast_pkts': '4943701088',
                                                                                                                  'poller_hash': 'MTM4MTM5cGVlcmludGZwb24gNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '61449529238',
                                                                                                                  'in_ucast_pkts': '541601342',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '146',
                                                                                                                  'out_octets': '3222822200898',
                                                                                                                  'out_ucast_pkts': '2199588349',
                                                                                                                  'poller_hash': 'MTI3NEIxMHBlZXJpbnRmcG9uIDIzOTg1NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '14/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '17759257719',
                                                                                         'out_octets': '53817039122024',
                                                                                         'out_ucast_pkts': '20764101591',
                                                                                         'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC84',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '14/8',
                                                                                                                 'id': '14/8',
                                                                                                                 'links': [       {       'connected_time': '15498547',
                                                                                                                                          'fiber_length': 1678,
                                                                                                                                          'poller_hash': 'MTM4MTM5cGVlcnBvbnBvbiA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -10.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.716},
                                                                                                                                  {       'connected_time': '73104368',
                                                                                                                                          'fiber_length': 1400,
                                                                                                                                          'poller_hash': 'MTI3NEIxMHBlZXJwb25wb24gMjM5ODU0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -11.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.716}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '14/8',
                                                                                                                 'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC84',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'NDcxOTA1MDA2Njgy',
                                                           'serial_number': '471905006682',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991388'},
                                                   {       'id': '15',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '15/1',
                                                                                         'in_discards': '1057',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '499586',
                                                                                         'in_octets': '1736195105863',
                                                                                         'in_ucast_pkts': '8992058911',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2683174784',
                                                                                                                  'in_ucast_pkts': '9508520',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '36166153281',
                                                                                                                  'out_ucast_pkts': '41298343',
                                                                                                                  'poller_hash': 'QzgxRTcxcGVlcmludGZwb24gMTkzOTIz'},
                                                                                                          {       'in_nucast_pkts': '526500',
                                                                                                                  'in_octets': '65423524573',
                                                                                                                  'in_ucast_pkts': '616125902',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '69208772',
                                                                                                                  'out_octets': '3342960444217',
                                                                                                                  'out_ucast_pkts': '2261332491',
                                                                                                                  'poller_hash': 'MTE3REJCMnBlZXJpbnRmcG9uIDI1MjgwNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45238237132',
                                                                                                                  'in_ucast_pkts': '8652368305',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '16',
                                                                                                                  'out_octets': '102849910346',
                                                                                                                  'out_ucast_pkts': '131482813',
                                                                                                                  'poller_hash': 'MTE3REIwRXBlZXJpbnRmcG9uIDI1MjgxNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '70459561442',
                                                                                                                  'in_ucast_pkts': '189401775',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10',
                                                                                                                  'out_octets': '530706138888',
                                                                                                                  'out_ucast_pkts': '475516923',
                                                                                                                  'poller_hash': 'MTE3REJCMHBlZXJpbnRmcG9uIDI1MzM1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '8817195169',
                                                                                                                  'in_ucast_pkts': '27041219',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '92518550762',
                                                                                                                  'out_ucast_pkts': '105955224',
                                                                                                                  'poller_hash': 'MTE3REMwMHBlZXJpbnRmcG9uIDI1MzQxNg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45196676748',
                                                                                                                  'in_ucast_pkts': '335476661',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '1649261899676',
                                                                                                                  'out_ucast_pkts': '1203897719',
                                                                                                                  'poller_hash': 'MTE3REIyMnBlZXJpbnRmcG9uIDI1MzQ2Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '6380159620',
                                                                                                                  'in_ucast_pkts': '23178655',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '16',
                                                                                                                  'out_octets': '88523206755',
                                                                                                                  'out_ucast_pkts': '97683470',
                                                                                                                  'poller_hash': 'MTE3REIzNHBlZXJpbnRmcG9uIDI1MzQ3Ng=='},
                                                                                                          {       'in_nucast_pkts': '5409',
                                                                                                                  'in_octets': '163972587008',
                                                                                                                  'in_ucast_pkts': '684524199',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '67684694',
                                                                                                                  'out_octets': '3417486951063',
                                                                                                                  'out_ucast_pkts': '2765092442',
                                                                                                                  'poller_hash': 'MTE3REI2RXBlZXJpbnRmcG9uIDI1Mzk5Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '176975052478',
                                                                                                                  'in_ucast_pkts': '47831526962',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '2606343779156',
                                                                                                                  'out_ucast_pkts': '2006067626',
                                                                                                                  'poller_hash': 'MTE3REEzQnBlZXJpbnRmcG9uIDI1NDE3Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '74074351421',
                                                                                                                  'in_ucast_pkts': '399539148',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1694735356644',
                                                                                                                  'out_ucast_pkts': '1259516995',
                                                                                                                  'poller_hash': 'MTE3REIzMHBlZXJpbnRmcG9uIDI1NDcyNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '140740344067',
                                                                                                                  'in_ucast_pkts': '61016728454',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '2615768850051',
                                                                                                                  'out_ucast_pkts': '2068800600',
                                                                                                                  'poller_hash': 'MTI5NzRENnBlZXJpbnRmcG9uIDI1NDc3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '70216702992',
                                                                                                                  'in_ucast_pkts': '427196979',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '2272157276591',
                                                                                                                  'out_ucast_pkts': '1693881743',
                                                                                                                  'poller_hash': 'MTI5NzRCMnBlZXJpbnRmcG9uIDI1NjI3Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '74308514060',
                                                                                                                  'in_ucast_pkts': '955442772',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3689331749908',
                                                                                                                  'out_ucast_pkts': '2457488621',
                                                                                                                  'poller_hash': 'MTI5NzBENHBlZXJpbnRmcG9uIDI1ODE4Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '16693248380',
                                                                                                                  'in_ucast_pkts': '44615614',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '181428968772',
                                                                                                                  'out_ucast_pkts': '178850269',
                                                                                                                  'poller_hash': 'MTI3NDg1Q3BlZXJpbnRmcG9uIDI2MTMxOQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '14633845057',
                                                                                                                  'in_ucast_pkts': '51834499',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '168840127629',
                                                                                                                  'out_ucast_pkts': '146870991',
                                                                                                                  'poller_hash': 'MTI5NzNFRHBlZXJpbnRmcG9uIDI2NTAwMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '18347781313',
                                                                                                                  'in_ucast_pkts': '65866143',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '282747110756',
                                                                                                                  'out_ucast_pkts': '249256851',
                                                                                                                  'poller_hash': 'MTI3NDg5NXBlZXJpbnRmcG9uIDI2NTAxMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEEyNXBlZXJpbnRmcG9uIDI3NDAzMQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '22667070328',
                                                                                                                  'in_ucast_pkts': '45626577',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '157375331237',
                                                                                                                  'out_ucast_pkts': '133725444',
                                                                                                                  'poller_hash': 'MTI5NUVBNnBlZXJpbnRmcG9uIDI4MDMyNA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '15/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '88927463',
                                                                                         'out_octets': '36940793478748',
                                                                                         'out_ucast_pkts': '27631299672',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '15/1',
                                                                                                                 'id': '15/1',
                                                                                                                 'links': [       {       'connected_time': '15037897',
                                                                                                                                          'fiber_length': 6516,
                                                                                                                                          'poller_hash': 'QzgxRTcxcGVlcnBvbnBvbiAxOTM5MjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '29484828',
                                                                                                                                          'fiber_length': 6914,
                                                                                                                                          'poller_hash': 'MTE3REJCMnBlZXJwb25wb24gMjUyODA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '29471886',
                                                                                                                                          'fiber_length': 6915,
                                                                                                                                          'poller_hash': 'MTE3REIwRXBlZXJwb25wb24gMjUyODE1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '523337',
                                                                                                                                          'fiber_length': 6929,
                                                                                                                                          'poller_hash': 'MTE3REJCMHBlZXJwb25wb24gMjUzMzU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '28968599',
                                                                                                                                          'fiber_length': 6977,
                                                                                                                                          'poller_hash': 'MTE3REMwMHBlZXJwb25wb24gMjUzNDE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '523339',
                                                                                                                                          'fiber_length': 6863,
                                                                                                                                          'poller_hash': 'MTE3REIyMnBlZXJwb25wb24gMjUzNDY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '28873352',
                                                                                                                                          'fiber_length': 6785,
                                                                                                                                          'poller_hash': 'MTE3REIzNHBlZXJwb25wb24gMjUzNDc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '14936535',
                                                                                                                                          'fiber_length': 6683,
                                                                                                                                          'poller_hash': 'MTE3REI2RXBlZXJwb25wb24gMjUzOTky',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '523339',
                                                                                                                                          'fiber_length': 6928,
                                                                                                                                          'poller_hash': 'MTE3REEzQnBlZXJwb25wb24gMjU0MTcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '26078305',
                                                                                                                                          'fiber_length': 6635,
                                                                                                                                          'poller_hash': 'MTE3REIzMHBlZXJwb25wb24gMjU0NzI1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '688595',
                                                                                                                                          'fiber_length': 6647,
                                                                                                                                          'poller_hash': 'MTI5NzRENnBlZXJwb25wb24gMjU0Nzc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '25944004',
                                                                                                                                          'fiber_length': 6626,
                                                                                                                                          'poller_hash': 'MTI5NzRCMnBlZXJwb25wb24gMjU2Mjcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '22986214',
                                                                                                                                          'fiber_length': 6878,
                                                                                                                                          'poller_hash': 'MTI5NzBENHBlZXJwb25wb24gMjU4MTgy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '688599',
                                                                                                                                          'fiber_length': 6586,
                                                                                                                                          'poller_hash': 'MTI3NDg1Q3BlZXJwb25wb24gMjYxMzE5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '17036726',
                                                                                                                                          'fiber_length': 6856,
                                                                                                                                          'poller_hash': 'MTI5NzNFRHBlZXJwb25wb24gMjY1MDAz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '17036075',
                                                                                                                                          'fiber_length': 6815,
                                                                                                                                          'poller_hash': 'MTI3NDg5NXBlZXJwb25wb24gMjY1MDEz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '13931908',
                                                                                                                                          'fiber_length': 6621,
                                                                                                                                          'poller_hash': 'MTI3NEEyNXBlZXJwb25wb24gMjc0MDMx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807},
                                                                                                                                  {       'connected_time': '4603596',
                                                                                                                                          'fiber_length': 6687,
                                                                                                                                          'poller_hash': 'MTI5NUVBNnBlZXJwb25wb24gMjgwMzI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.807}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '15/1',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '15/2',
                                                                                         'in_discards': '3853',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3457768',
                                                                                         'in_octets': '3606909852604',
                                                                                         'in_ucast_pkts': '17609550154',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '236363120846',
                                                                                                                  'in_ucast_pkts': '1398277995',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '8227355631983',
                                                                                                                  'out_ucast_pkts': '5781736918',
                                                                                                                  'poller_hash': 'MTE3REExMXBlZXJpbnRmcG9uIDI0OTYxMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '338938878696',
                                                                                                                  'in_ucast_pkts': '1884411055',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14',
                                                                                                                  'out_octets': '8441798087908',
                                                                                                                  'out_ucast_pkts': '5938021296',
                                                                                                                  'poller_hash': 'MTE3REI1MXBlZXJpbnRmcG9uIDI1MDEyNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTE3REFGRXBlZXJpbnRmcG9uIDI1MDI4Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '106530660415',
                                                                                                                  'in_ucast_pkts': '548980354',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '2944736963626',
                                                                                                                  'out_ucast_pkts': '2138729830',
                                                                                                                  'poller_hash': 'MTE3REMxQXBlZXJpbnRmcG9uIDI1MTA4NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '258439561573',
                                                                                                                  'in_ucast_pkts': '2721977347',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11948312902341',
                                                                                                                  'out_ucast_pkts': '8128390581',
                                                                                                                  'poller_hash': 'MTE3REFEMXBlZXJpbnRmcG9uIDI1MTc3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '85974671117',
                                                                                                                  'in_ucast_pkts': '922975445',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '5710108908743',
                                                                                                                  'out_ucast_pkts': '8072364026',
                                                                                                                  'poller_hash': 'MTE3REMwOXBlZXJpbnRmcG9uIDI1MjYyMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '160782063685',
                                                                                                                  'in_ucast_pkts': '30561320095',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12',
                                                                                                                  'out_octets': '2713951606825',
                                                                                                                  'out_ucast_pkts': '2047817019',
                                                                                                                  'poller_hash': 'MTE3REJGNnBlZXJpbnRmcG9uIDI1MjcwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '282495918827',
                                                                                                                  'in_ucast_pkts': '1001016803',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6',
                                                                                                                  'out_octets': '3421236967153',
                                                                                                                  'out_ucast_pkts': '2774293261',
                                                                                                                  'poller_hash': 'MTE3REFBOHBlZXJpbnRmcG9uIDI1MzM3Ng=='},
                                                                                                          {       'in_nucast_pkts': '2845273',
                                                                                                                  'in_octets': '104710526962',
                                                                                                                  'in_ucast_pkts': '1400990618',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '204884793',
                                                                                                                  'out_octets': '7954307086489',
                                                                                                                  'out_ucast_pkts': '5609899953',
                                                                                                                  'poller_hash': 'MTE3REMxRXBlZXJpbnRmcG9uIDI1MzM4Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '30570904128',
                                                                                                                  'in_ucast_pkts': '58338922',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '260838667082',
                                                                                                                  'out_ucast_pkts': '235752037',
                                                                                                                  'poller_hash': 'MTI5NzRFOXBlZXJpbnRmcG9uIDI1NjY1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTFCRTFFQXBlZXJpbnRmcG9uIDI1NzcwOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '59483385270',
                                                                                                                  'in_ucast_pkts': '227504100',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12',
                                                                                                                  'out_octets': '942520890227',
                                                                                                                  'out_ucast_pkts': '820662849',
                                                                                                                  'poller_hash': 'MTI5NzQ2OXBlZXJpbnRmcG9uIDI1NzcxOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '25273304171',
                                                                                                                  'in_ucast_pkts': '84631816',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '207883001786',
                                                                                                                  'out_ucast_pkts': '206278759',
                                                                                                                  'poller_hash': 'MTI5NzQyMXBlZXJpbnRmcG9uIDI1NzcyOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '15518169321',
                                                                                                                  'in_ucast_pkts': '62697067',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '126316107111',
                                                                                                                  'out_ucast_pkts': '146365603',
                                                                                                                  'poller_hash': 'MTI5NzM4QXBlZXJpbnRmcG9uIDI1OTg5Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '56177634478',
                                                                                                                  'in_ucast_pkts': '296104024',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1956862626195',
                                                                                                                  'out_ucast_pkts': '1444956721',
                                                                                                                  'poller_hash': 'MTI3NDZEMHBlZXJpbnRmcG9uIDI2MTIyNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '19281366609',
                                                                                                                  'in_ucast_pkts': '102256862',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8',
                                                                                                                  'out_octets': '400159724318',
                                                                                                                  'out_ucast_pkts': '329410060',
                                                                                                                  'poller_hash': 'MTI3NDcxMXBlZXJpbnRmcG9uIDI2MjYzOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI5NzQyQXBlZXJpbnRmcG9uIDI2NDQ3NA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDdGQnBlZXJpbnRmcG9uIDI2NDk5Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '40610898747',
                                                                                                                  'in_ucast_pkts': '61629158',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '16',
                                                                                                                  'out_octets': '178241700862',
                                                                                                                  'out_ucast_pkts': '163434892',
                                                                                                                  'poller_hash': 'MTI3NDY4MXBlZXJpbnRmcG9uIDI2NTcxNw=='},
                                                                                                          {       'in_nucast_pkts': '14822',
                                                                                                                  'in_octets': '3928597351',
                                                                                                                  'in_ucast_pkts': '15999438',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1465364',
                                                                                                                  'out_octets': '78778206726',
                                                                                                                  'out_ucast_pkts': '55199487',
                                                                                                                  'poller_hash': 'MTFCRTM2OXBlZXJpbnRmcG9uIDI2NzAwMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '143391161500',
                                                                                                                  'in_ucast_pkts': '178055765',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '462723864261',
                                                                                                                  'out_ucast_pkts': '362592193',
                                                                                                                  'poller_hash': 'MTI5NUVBQXBlZXJpbnRmcG9uIDI4MjEzNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05148',
                                                                                         'name': 'gpon '
                                                                                                 '15/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '2333890063',
                                                                                         'out_octets': '86876515552204',
                                                                                         'out_ucast_pkts': '60173381766',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '15/2',
                                                                                                                 'id': '15/2',
                                                                                                                 'links': [       {       'connected_time': '523341',
                                                                                                                                          'fiber_length': 6497,
                                                                                                                                          'poller_hash': 'MTE3REExMXBlZXJwb25wb24gMjQ5NjEy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '688608',
                                                                                                                                          'fiber_length': 6608,
                                                                                                                                          'poller_hash': 'MTE3REI1MXBlZXJwb25wb24gMjUwMTI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '15143026',
                                                                                                                                          'fiber_length': 6674,
                                                                                                                                          'poller_hash': 'MTE3REFGRXBlZXJwb25wb24gMjUwMjg3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '688603',
                                                                                                                                          'fiber_length': 6631,
                                                                                                                                          'poller_hash': 'MTE3REMxQXBlZXJwb25wb24gMjUxMDg0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '719128',
                                                                                                                                          'fiber_length': 6546,
                                                                                                                                          'poller_hash': 'MTE3REFEMXBlZXJwb25wb24gMjUxNzc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '26622824',
                                                                                                                                          'fiber_length': 6461,
                                                                                                                                          'poller_hash': 'MTE3REMwOXBlZXJwb25wb24gMjUyNjIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '688601',
                                                                                                                                          'fiber_length': 6730,
                                                                                                                                          'poller_hash': 'MTE3REJGNnBlZXJwb25wb24gMjUyNzAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '24794',
                                                                                                                                          'fiber_length': 6908,
                                                                                                                                          'poller_hash': 'MTE3REFBOHBlZXJwb25wb24gMjUzMzc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '29035616',
                                                                                                                                          'fiber_length': 6509,
                                                                                                                                          'poller_hash': 'MTE3REMxRXBlZXJwb25wb24gMjUzMzg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '24734684',
                                                                                                                                          'fiber_length': 6717,
                                                                                                                                          'poller_hash': 'MTI5NzRFOXBlZXJwb25wb24gMjU2NjU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '20585315',
                                                                                                                                          'fiber_length': 6742,
                                                                                                                                          'poller_hash': 'MTFCRTFFQXBlZXJwb25wb24gMjU3NzA4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '688592',
                                                                                                                                          'fiber_length': 6751,
                                                                                                                                          'poller_hash': 'MTI5NzQ2OXBlZXJwb25wb24gMjU3NzE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '23601025',
                                                                                                                                          'fiber_length': 6777,
                                                                                                                                          'poller_hash': 'MTI5NzQyMXBlZXJwb25wb24gMjU3NzI4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '22121595',
                                                                                                                                          'fiber_length': 6675,
                                                                                                                                          'poller_hash': 'MTI5NzM4QXBlZXJwb25wb24gMjU5ODky',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '21707981',
                                                                                                                                          'fiber_length': 6806,
                                                                                                                                          'poller_hash': 'MTI3NDZEMHBlZXJwb25wb24gMjYxMjI3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -16.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '688590',
                                                                                                                                          'fiber_length': 6555,
                                                                                                                                          'poller_hash': 'MTI3NDcxMXBlZXJwb25wb24gMjYyNjM4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '18149298',
                                                                                                                                          'fiber_length': 6818,
                                                                                                                                          'poller_hash': 'MTI5NzQyQXBlZXJwb25wb24gMjY0NDc0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '17646665',
                                                                                                                                          'fiber_length': 6656,
                                                                                                                                          'poller_hash': 'MTI3NDdGQnBlZXJwb25wb24gMjY0OTkz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '523342',
                                                                                                                                          'fiber_length': 6613,
                                                                                                                                          'poller_hash': 'MTI3NDY4MXBlZXJwb25wb24gMjY1NzE3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '688607',
                                                                                                                                          'fiber_length': 6621,
                                                                                                                                          'poller_hash': 'MTFCRTM2OXBlZXJwb25wb24gMjY3MDAz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601},
                                                                                                                                  {       'connected_time': '688604',
                                                                                                                                          'fiber_length': 6671,
                                                                                                                                          'poller_hash': 'MTI5NUVBQXBlZXJwb25wb24gMjgyMTM2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.601}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '15/2',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '15/3',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '0',
                                                                                         'in_octets': '1649239816',
                                                                                         'in_ucast_pkts': '5095998',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3MEQ2QnBlZXJpbnRmcG9uIDI3Njk5Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '592187652',
                                                                                                                  'in_ucast_pkts': '1597735',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4535845304',
                                                                                                                  'out_ucast_pkts': '10660471',
                                                                                                                  'poller_hash': 'MTI5NUU0OXBlZXJpbnRmcG9uIDI3OTI4Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '15/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '22087536',
                                                                                         'out_octets': '10864010674',
                                                                                         'out_ucast_pkts': '9458863',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8z',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '15/3',
                                                                                                                 'id': '15/3',
                                                                                                                 'links': [       {       'connected_time': '10132691',
                                                                                                                                          'fiber_length': 10196,
                                                                                                                                          'poller_hash': 'MTI3MEQ2QnBlZXJwb25wb24gMjc2OTk2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -13.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.295},
                                                                                                                                  {       'connected_time': '6490902',
                                                                                                                                          'fiber_length': 10561,
                                                                                                                                          'poller_hash': 'MTI5NUU0OXBlZXJwb25wb24gMjc5Mjg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.295}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '15/3',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8z',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': False,
                                                                                         'id': '15/4',
                                                                                         'name': 'gpon '
                                                                                                 '15/4',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS80'},
                                                                                 {       'admin_up': True,
                                                                                         'id': '15/5',
                                                                                         'name': 'gpon '
                                                                                                 '15/5',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS81'},
                                                                                 {       'admin_up': True,
                                                                                         'id': '15/6',
                                                                                         'name': 'gpon '
                                                                                                 '15/6',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS82'},
                                                                                 {       'admin_up': True,
                                                                                         'id': '15/7',
                                                                                         'name': 'gpon '
                                                                                                 '15/7',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS83'},
                                                                                 {       'admin_up': True,
                                                                                         'id': '15/8',
                                                                                         'in_discards': '432',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7775',
                                                                                         'in_octets': '43271909189',
                                                                                         'in_ucast_pkts': '138729880',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '15/8',
                                                                                         'oper_up': False,
                                                                                         'out_nucast_pkts': '95488999',
                                                                                         'out_octets': '729288581824',
                                                                                         'out_ucast_pkts': '452723213',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS84',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '15/8',
                                                                                                                 'id': '15/8',
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '15/8',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS84',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'NDcxOTA1MDEwMzk2',
                                                           'serial_number': '471905010396',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '35098472'},
                                                   {       'id': 'A',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': 'A/x1',
                                                                                         'in_discards': '112308',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '90946678',
                                                                                         'in_octets': '12363560361621981',
                                                                                         'in_ucast_pkts': '9184357454006',
                                                                                         'mac_address': '00:02:5d:db:26:9f',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'MCCL-SUBS '
                                                                                                 'A/x1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '158328468',
                                                                                         'out_octets': '1077096019951228',
                                                                                         'out_ucast_pkts': '3229638408761',
                                                                                         'poller_hash': 'MjExNDExNjAxMjY2TUNDTC1TVUJTIEEveDE=',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/x2',
                                                                                         'in_discards': '553495506',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2715766911',
                                                                                         'in_octets': '545494181657573',
                                                                                         'in_ucast_pkts': '573489466466',
                                                                                         'mac_address': '00:02:5d:db:26:a0',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01629',
                                                                                         'name': 'Data '
                                                                                                 'LAG '
                                                                                                 'A/x2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '250638254',
                                                                                         'out_octets': '453061188271301',
                                                                                         'out_ucast_pkts': '495992361484',
                                                                                         'poller_hash': 'MjExNDExNjAxMjY2RGF0YSBMQUcgQS94Mg==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/g1',
                                                                                         'in_discards': '88383584',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '428995177',
                                                                                         'in_octets': '2609115230652',
                                                                                         'in_ucast_pkts': '12254569793',
                                                                                         'mac_address': '00:02:5d:db:26:a3',
                                                                                         'manufacturer': 'FiberStore',
                                                                                         'model': '10GSFP-PC-30-2',
                                                                                         'name': 'Voice '
                                                                                                 'LAG '
                                                                                                 'A/g1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '10912932',
                                                                                         'out_octets': '2514008628096',
                                                                                         'out_ucast_pkts': '11303499881',
                                                                                         'poller_hash': 'MjExNDExNjAxMjY2Vm9pY2UgTEFHIEEvZzE=',
                                                                                         'speed': '1000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/g2',
                                                                                         'in_discards': '194040',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '328708904',
                                                                                         'in_octets': '195473735793318',
                                                                                         'in_ucast_pkts': '139011311490',
                                                                                         'mac_address': '00:02:5d:db:26:a4',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01660',
                                                                                         'name': 'ethernet '
                                                                                                 'A/g2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '75809388',
                                                                                         'out_octets': '2937369930699',
                                                                                         'out_ucast_pkts': '36938868128',
                                                                                         'poller_hash': 'MjExNDExNjAxMjY2ZXRoZXJuZXQgQS9nMg==',
                                                                                         'speed': '1000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/g3',
                                                                                         'in_discards': '68843',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '45959932',
                                                                                         'in_octets': '11790740102',
                                                                                         'in_ucast_pkts': '614526',
                                                                                         'mac_address': '00:02:5d:db:26:a5',
                                                                                         'manufacturer': 'FiberStore',
                                                                                         'model': 'SFP-1G85-5M-BR',
                                                                                         'name': 'ethernet '
                                                                                                 'A/g3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '4294035',
                                                                                         'out_octets': '536753710',
                                                                                         'out_ucast_pkts': '0',
                                                                                         'poller_hash': 'MjExNDExNjAxMjY2ZXRoZXJuZXQgQS9nMw==',
                                                                                         'speed': '1000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/g4',
                                                                                         'in_discards': '44389',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '938186539887',
                                                                                         'in_octets': '1281801809297371',
                                                                                         'in_ucast_pkts': '2892062677',
                                                                                         'mac_address': '00:02:5d:db:26:a6',
                                                                                         'manufacturer': '',
                                                                                         'model': 'SFP-SX',
                                                                                         'name': 'IPTV '
                                                                                                 'LAG '
                                                                                                 'A/g4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '35871866',
                                                                                         'out_octets': '353312658557',
                                                                                         'out_ucast_pkts': '2108227813',
                                                                                         'poller_hash': 'MjExNDExNjAxMjY2SVBUViBMQUcgQS9nNA==',
                                                                                         'speed': '1000000000'}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-02092',
                                                           'name': 'SCP-10GE',
                                                           'poller_hash': 'MjExNDExNjAxMjY2',
                                                           'serial_number': '211411601266',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42992083'},
                                                   {       'id': 'B',
                                                           'interfaces': [       {       'admin_up': False,
                                                                                         'id': 'B/x1',
                                                                                         'mac_address': '00:02:5d:db:28:5f',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'MCCL-SUBS '
                                                                                                 'B/x1',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'MjExNDExNjAxMzA5TUNDTC1TVUJTIEIveDE='},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/x2',
                                                                                         'in_discards': '1549139',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '44239519',
                                                                                         'in_octets': '6127400132',
                                                                                         'in_ucast_pkts': '0',
                                                                                         'mac_address': '00:02:5d:db:28:60',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01629',
                                                                                         'name': 'Data '
                                                                                                 'LAG '
                                                                                                 'B/x2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '42947805',
                                                                                         'out_octets': '5497319040',
                                                                                         'out_ucast_pkts': '0',
                                                                                         'poller_hash': 'MjExNDExNjAxMzA5RGF0YSBMQUcgQi94Mg==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/g1',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '42990902',
                                                                                         'in_octets': '5502835456',
                                                                                         'in_ucast_pkts': '0',
                                                                                         'mac_address': '00:02:5d:db:28:63',
                                                                                         'manufacturer': 'FiberStore',
                                                                                         'model': '10GSFP-PC-30-2',
                                                                                         'name': 'Voice '
                                                                                                 'LAG '
                                                                                                 'B/g1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '4294780',
                                                                                         'out_octets': '5999808172',
                                                                                         'out_ucast_pkts': '0',
                                                                                         'poller_hash': 'MjExNDExNjAxMzA5Vm9pY2UgTEFHIEIvZzE=',
                                                                                         'speed': '1000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/g2',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '201955',
                                                                                         'in_octets': '3078448894455',
                                                                                         'in_ucast_pkts': '21551247754',
                                                                                         'mac_address': '00:02:5d:db:28:64',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01660',
                                                                                         'name': 'ethernet '
                                                                                                 'B/g2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '54417812',
                                                                                         'out_octets': '3081968414830',
                                                                                         'out_ucast_pkts': '21480284128',
                                                                                         'poller_hash': 'MjExNDExNjAxMzA5ZXRoZXJuZXQgQi9nMg==',
                                                                                         'speed': '1000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/g3',
                                                                                         'in_discards': '687795',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '12214922',
                                                                                         'in_octets': '2443500304',
                                                                                         'in_ucast_pkts': '11866115',
                                                                                         'mac_address': '00:02:5d:db:28:65',
                                                                                         'manufacturer': 'FiberStore',
                                                                                         'model': 'SFP-1G85-5M',
                                                                                         'name': 'ethernet '
                                                                                                 'B/g3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '4294799',
                                                                                         'out_octets': '523964870',
                                                                                         'out_ucast_pkts': '0',
                                                                                         'poller_hash': 'MjExNDExNjAxMzA5ZXRoZXJuZXQgQi9nMw==',
                                                                                         'speed': '1000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/g4',
                                                                                         'in_discards': '49955',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1002483354991',
                                                                                         'in_octets': '1371128106385697',
                                                                                         'in_ucast_pkts': '2772839093',
                                                                                         'mac_address': '00:02:5d:db:28:66',
                                                                                         'manufacturer': '',
                                                                                         'model': 'SFP-SX',
                                                                                         'name': 'IPTV '
                                                                                                 'LAG '
                                                                                                 'B/g4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '9411874',
                                                                                         'out_octets': '320637872213',
                                                                                         'out_ucast_pkts': '1748066380',
                                                                                         'poller_hash': 'MjExNDExNjAxMzA5SVBUViBMQUcgQi9nNA==',
                                                                                         'speed': '1000000000'}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-02092',
                                                           'name': 'SCP-10GE',
                                                           'poller_hash': 'MjExNDExNjAxMzA5',
                                                           'serial_number': '211411601309',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '42991639'}],
                                'serial_number': '71412913092'},
                'peers': [       {       'counter_source': '',
                                         'interfaces': [       {       'id': '1/g1',
                                                                       'name': '1/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUNDNDY0cGVlcmludGYxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '1',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUNDNDY0cGVlcmludGZwb24gMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '1',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUNDNDY0cGVlcnBvbnBvbiAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'Milverton CO (844G)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "1" (location: '
                                                                  'GPON port "1/1"), '
                                                                  '"ONT went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-24T08:29:34.120Z',
                                         'poller_hash': 'MUNDNDY0',
                                         'serial_number': '1CC464',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '2/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:fc:af:f9:6d:00'],
                                                                       'id': '2/g1',
                                                                       'in_nucast_pkts': '208',
                                                                       'in_octets': '9934667130',
                                                                       'in_ucast_pkts': '101392439',
                                                                       'mac_address': 'cc:be:59:ab:61:0f',
                                                                       'mtu': 2000,
                                                                       'name': '2/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120021',
                                                                       'out_octets': '334279638785',
                                                                       'out_ucast_pkts': '229953368',
                                                                       'poller_hash': 'NEY5QjNDcGVlcmludGYyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '2',
                                                                       'in_nucast_pkts': '1120021',
                                                                       'in_octets': '334279638785',
                                                                       'in_ucast_pkts': '229953368',
                                                                       'links': [       {       'in_nucast_pkts': '1120021',
                                                                                                'in_octets': '334279638785',
                                                                                                'in_ucast_pkts': '229953368',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '208',
                                                                                                'out_octets': '9934667130',
                                                                                                'out_ucast_pkts': '101392439',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '208',
                                                                       'out_octets': '9934667130',
                                                                       'out_ucast_pkts': '101392439',
                                                                       'poller_hash': 'NEY5QjNDcGVlcmludGZwb24gMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '2',
                                                                                               'links': [       {       'connected_time': '688527',
                                                                                                                        'fiber_length': 5646,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.642,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5QjNDcGVlcnBvbnBvbiAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '6383 Road 124 Milverton (Brad Gascho '
                                                 '10400853)',
                                         'poller_hash': 'NEY5QjND',
                                         'serial_number': '4F9B3C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:61:0e',
                                         'uptime': '688527'},
                                 {       'counter_source': '3/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:62:99'],
                                                                       'id': '3/g1',
                                                                       'in_nucast_pkts': '35927039',
                                                                       'in_octets': '444971703362',
                                                                       'in_ucast_pkts': '1029626764',
                                                                       'mac_address': '00:06:31:e9:02:3e',
                                                                       'mtu': 2000,
                                                                       'name': '3/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10245610953',
                                                                       'out_octets': '16346445312544',
                                                                       'out_ucast_pkts': '2655212833',
                                                                       'poller_hash': 'MUQwNjlEcGVlcmludGYzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '3',
                                                                       'in_nucast_pkts': '10245610953',
                                                                       'in_octets': '16346445312544',
                                                                       'in_ucast_pkts': '2655212833',
                                                                       'links': [       {       'in_nucast_pkts': '10245610953',
                                                                                                'in_octets': '16346445312544',
                                                                                                'in_ucast_pkts': '2655212833',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '35927039',
                                                                                                'out_octets': '444971703362',
                                                                                                'out_ucast_pkts': '1029626764',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35927039',
                                                                       'out_octets': '444971703362',
                                                                       'out_ucast_pkts': '1029626764',
                                                                       'poller_hash': 'MUQwNjlEcGVlcmludGZwb24gMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '3',
                                                                                               'links': [       {       'connected_time': '41697',
                                                                                                                        'fiber_length': 1341,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjlEcGVlcnBvbnBvbiAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '6 WHALEY AV (PHILLIPS JODI)',
                                         'poller_hash': 'MUQwNjlE',
                                         'serial_number': '1D069D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:02:3b',
                                         'uptime': '41697'},
                                 {       'counter_source': '4/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3b:6d'],
                                                                       'id': '4/g1',
                                                                       'in_nucast_pkts': '12611',
                                                                       'in_octets': '33243888094',
                                                                       'in_ucast_pkts': '53467695',
                                                                       'mac_address': '84:d3:43:7e:34:ca',
                                                                       'mtu': 2000,
                                                                       'name': '4/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1462042',
                                                                       'out_octets': '228423413867',
                                                                       'out_ucast_pkts': '172118685',
                                                                       'poller_hash': 'QzgzQzAycGVlcmludGY0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '4',
                                                                       'in_nucast_pkts': '1462042',
                                                                       'in_octets': '228423413867',
                                                                       'in_ucast_pkts': '172118685',
                                                                       'links': [       {       'in_nucast_pkts': '1462042',
                                                                                                'in_octets': '228423413867',
                                                                                                'in_ucast_pkts': '172118685',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12611',
                                                                                                'out_octets': '33243888094',
                                                                                                'out_ucast_pkts': '53467695',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12611',
                                                                       'out_octets': '33243888094',
                                                                       'out_ucast_pkts': '53467695',
                                                                       'poller_hash': 'QzgzQzAycGVlcmludGZwb24gNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '4',
                                                                                               'links': [       {       'connected_time': '688481',
                                                                                                                        'fiber_length': 1250,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQzAycGVlcnBvbnBvbiA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '61 MAIN ST N (VATERS CARLIE)',
                                         'poller_hash': 'QzgzQzAy',
                                         'serial_number': 'C83C02',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:34:c7',
                                         'uptime': '688481'},
                                 {       'counter_source': '5/g1+5/g6+5/g7',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '08:62:66:9a:51:64'],
                                                                       'id': '5/g1',
                                                                       'in_nucast_pkts': '97337',
                                                                       'in_octets': '206478818630',
                                                                       'in_ucast_pkts': '930684151',
                                                                       'mac_address': '00:06:31:9c:c6:93',
                                                                       'mtu': 2000,
                                                                       'name': '5/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27298411',
                                                                       'out_octets': '3695635472289',
                                                                       'out_ucast_pkts': '2517691370',
                                                                       'poller_hash': 'MTM4MTM5cGVlcmludGY1L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:32:e5:17:f2:45'],
                                                                       'id': '5/g6',
                                                                       'in_nucast_pkts': '41777',
                                                                       'in_octets': '40447404472',
                                                                       'in_ucast_pkts': '397872661',
                                                                       'mac_address': '00:06:31:a1:12:72',
                                                                       'mtu': 2000,
                                                                       'name': '5/g6',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20520119',
                                                                       'out_octets': '2304778657028',
                                                                       'out_ucast_pkts': '1561655187',
                                                                       'poller_hash': 'MTM4MTM5cGVlcmludGY1L2c2',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:71:76'],
                                                                       'id': '5/g7',
                                                                       'in_nucast_pkts': '256164',
                                                                       'in_octets': '51861970023',
                                                                       'in_ucast_pkts': '317175369',
                                                                       'mac_address': '00:06:31:a1:12:73',
                                                                       'mtu': 2000,
                                                                       'name': '5/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23002152',
                                                                       'out_octets': '1258400807395',
                                                                       'out_ucast_pkts': '864354531',
                                                                       'poller_hash': 'MTM4MTM5cGVlcmludGY1L2c3',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '5',
                                                                       'in_nucast_pkts': '70820682',
                                                                       'in_octets': '7258814936712',
                                                                       'in_ucast_pkts': '4943701088',
                                                                       'links': [       {       'in_nucast_pkts': '70820682',
                                                                                                'in_octets': '7258814936712',
                                                                                                'in_ucast_pkts': '4943701088',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '395278',
                                                                                                'out_octets': '298788193125',
                                                                                                'out_ucast_pkts': '1645732181',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '395278',
                                                                       'out_octets': '298788193125',
                                                                       'out_ucast_pkts': '1645732181',
                                                                       'poller_hash': 'MTM4MTM5cGVlcmludGZwb24gNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '5',
                                                                                               'links': [       {       'connected_time': '15498547',
                                                                                                                        'fiber_length': 1678,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTM4MTM5cGVlcnBvbnBvbiA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '54 PUGH ST (WOODS RUTH A)',
                                         'poller_hash': 'MTM4MTM5',
                                         'serial_number': '138139',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:9c:c6:98',
                                         'uptime': '15498547'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '6/g1',
                                                                       'name': '6/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVFOUMzcGVlcmludGY2L2cx'},
                                                               {       'admin_up': False,
                                                                       'id': '6',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVFOUMzcGVlcmludGZwb24gNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '6',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOUMzcGVlcnBvbnBvbiA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'ont 6',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "6" (location: '
                                                                  'GPON port "7/1"), '
                                                                  '"ONT went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:38.730Z',
                                         'poller_hash': 'MzVFOUMz',
                                         'serial_number': '35E9C3',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '7/g1+7/g2+7/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:52:9a'],
                                                                       'id': '7/g1',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '51128679270',
                                                                       'in_ucast_pkts': '399167223',
                                                                       'mac_address': 'd0:76:8f:08:17:5e',
                                                                       'mtu': 2000,
                                                                       'name': '7/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24759027',
                                                                       'out_octets': '1484879449089',
                                                                       'out_ucast_pkts': '1040522996',
                                                                       'poller_hash': 'ODY5M0UxcGVlcmludGY3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:42:1a:7a:b7:88'],
                                                                       'id': '7/g2',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '483059600383',
                                                                       'in_ucast_pkts': '530044975',
                                                                       'mac_address': 'd0:76:8f:08:17:5f',
                                                                       'mtu': 2000,
                                                                       'name': '7/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5743211',
                                                                       'out_octets': '215839515810',
                                                                       'out_ucast_pkts': '358564635',
                                                                       'poller_hash': 'ODY5M0UxcGVlcmludGY3L2cy',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '38:c0:ea:44:be:46'],
                                                                       'id': '7/g3',
                                                                       'in_nucast_pkts': '15818',
                                                                       'in_octets': '5403509303',
                                                                       'in_ucast_pkts': '22514890',
                                                                       'mac_address': 'd0:76:8f:08:17:60',
                                                                       'mtu': 2000,
                                                                       'name': '7/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24758973',
                                                                       'out_octets': '6726057064',
                                                                       'out_ucast_pkts': '26391681',
                                                                       'poller_hash': 'ODY5M0UxcGVlcmludGY3L2cz',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '7',
                                                                       'in_nucast_pkts': '55261211',
                                                                       'in_octets': '1707445021963',
                                                                       'in_ucast_pkts': '1425479312',
                                                                       'links': [       {       'in_nucast_pkts': '55261211',
                                                                                                'in_octets': '1707445021963',
                                                                                                'in_ucast_pkts': '1425479312',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15895',
                                                                                                'out_octets': '539591788956',
                                                                                                'out_ucast_pkts': '951727088',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15895',
                                                                       'out_octets': '539591788956',
                                                                       'out_ucast_pkts': '951727088',
                                                                       'poller_hash': 'ODY5M0UxcGVlcmludGZwb24gNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '7',
                                                                                               'links': [       {       'connected_time': '11410495',
                                                                                                                        'fiber_length': 18235,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0UxcGVlcnBvbnBvbiA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '209 NELSON STREET (MUNICIPALITY OF '
                                                 'NORTH PERTH)',
                                         'poller_hash': 'ODY5M0Ux',
                                         'serial_number': '8693E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:5b',
                                         'uptime': '11410495'},
                                 {       'counter_source': '8/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e6:fc'],
                                                                       'id': '8/g1',
                                                                       'in_nucast_pkts': '12217',
                                                                       'in_octets': '15136951266',
                                                                       'in_ucast_pkts': '39640425',
                                                                       'mac_address': 'ec:4f:82:8d:5f:9b',
                                                                       'mtu': 2000,
                                                                       'name': '8/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1446854',
                                                                       'out_octets': '170714713990',
                                                                       'out_ucast_pkts': '132006254',
                                                                       'poller_hash': 'MzIxNDNBcGVlcmludGY4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '8',
                                                                       'in_nucast_pkts': '1446854',
                                                                       'in_octets': '170714713990',
                                                                       'in_ucast_pkts': '132006254',
                                                                       'links': [       {       'in_nucast_pkts': '1446854',
                                                                                                'in_octets': '170714713990',
                                                                                                'in_ucast_pkts': '132006254',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12217',
                                                                                                'out_octets': '15136951266',
                                                                                                'out_ucast_pkts': '39640425',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12217',
                                                                       'out_octets': '15136951266',
                                                                       'out_ucast_pkts': '39640425',
                                                                       'poller_hash': 'MzIxNDNBcGVlcmludGZwb24gOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '8',
                                                                                               'links': [       {       'connected_time': '688412',
                                                                                                                        'fiber_length': 2222,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxNDNBcGVlcnBvbnBvbiA4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '80 Gerber Drive Milverton (Randal '
                                                 'Leis 10434200)',
                                         'poller_hash': 'MzIxNDNB',
                                         'serial_number': '32143A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:5f:98',
                                         'uptime': '688412'},
                                 {       'counter_source': '9/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:94:a1'],
                                                                       'id': '9/g1',
                                                                       'in_nucast_pkts': '9664783',
                                                                       'in_octets': '1259326056626',
                                                                       'in_ucast_pkts': '7656200600',
                                                                       'mac_address': 'ec:4f:82:8d:80:7b',
                                                                       'mtu': 2000,
                                                                       'name': '9/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18519500940',
                                                                       'out_octets': '52865383785210',
                                                                       'out_ucast_pkts': '20448640608',
                                                                       'poller_hash': 'MzIxODU2cGVlcmludGY5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '9',
                                                                       'in_nucast_pkts': '18519500940',
                                                                       'in_octets': '52865383785210',
                                                                       'in_ucast_pkts': '20448640608',
                                                                       'links': [       {       'in_nucast_pkts': '18519500940',
                                                                                                'in_octets': '52865383785210',
                                                                                                'in_ucast_pkts': '20448640608',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9664783',
                                                                                                'out_octets': '1259326056626',
                                                                                                'out_ucast_pkts': '7656200600',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9664783',
                                                                       'out_octets': '1259326056626',
                                                                       'out_ucast_pkts': '7656200600',
                                                                       'poller_hash': 'MzIxODU2cGVlcmludGZwb24gOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '9',
                                                                                               'links': [       {       'connected_time': '21816923',
                                                                                                                        'fiber_length': 1161,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxODU2cGVlcnBvbnBvbiA5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '41 MILL ST W (GROPP KEN JR & INGRID)',
                                         'poller_hash': 'MzIxODU2',
                                         'serial_number': '321856',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:80:78',
                                         'uptime': '21816923'},
                                 {       'counter_source': '11/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '11/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '582',
                                                                       'in_octets': '314983286771',
                                                                       'in_ucast_pkts': '2592117679',
                                                                       'mac_address': 'ec:4f:82:0d:68:71',
                                                                       'name': '11/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '3733',
                                                                       'out_octets': '18435312854967',
                                                                       'out_ucast_pkts': '12693954656',
                                                                       'poller_hash': 'MjA1REJDcGVlcmludGYxMS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '11',
                                                                       'in_nucast_pkts': '3733',
                                                                       'in_octets': '18435312854967',
                                                                       'in_ucast_pkts': '12693954656',
                                                                       'links': [       {       'in_nucast_pkts': '3733',
                                                                                                'in_octets': '18435312854967',
                                                                                                'in_ucast_pkts': '12693954656',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '582',
                                                                                                'out_octets': '314983286771',
                                                                                                'out_ucast_pkts': '2592117679',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '582',
                                                                       'out_octets': '314983286771',
                                                                       'out_ucast_pkts': '2592117679',
                                                                       'poller_hash': 'MjA1REJDcGVlcmludGZwb24gMTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '11',
                                                                                               'links': [       {       'connected_time': '688473',
                                                                                                                        'fiber_length': 1708,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REJDcGVlcnBvbnBvbiAxMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 WILLIAM ST (PAULNITZ CATHY)',
                                         'poller_hash': 'MjA1REJD',
                                         'serial_number': '205DBC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:71',
                                         'uptime': '688473'},
                                 {       'counter_source': '12/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '12/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '47297301923',
                                                                       'in_ucast_pkts': '111946135',
                                                                       'mac_address': 'ec:4f:82:0d:6a:8d',
                                                                       'name': '12/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '111',
                                                                       'out_octets': '300968323037',
                                                                       'out_ucast_pkts': '310366662',
                                                                       'poller_hash': 'MjA1REY4cGVlcmludGYxMi9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '12',
                                                                       'in_nucast_pkts': '111',
                                                                       'in_octets': '300968323037',
                                                                       'in_ucast_pkts': '310366662',
                                                                       'links': [       {       'in_nucast_pkts': '111',
                                                                                                'in_octets': '300968323037',
                                                                                                'in_ucast_pkts': '310366662',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '47297301923',
                                                                                                'out_ucast_pkts': '111946135',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '47297301923',
                                                                       'out_ucast_pkts': '111946135',
                                                                       'poller_hash': 'MjA1REY4cGVlcmludGZwb24gMTI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '12',
                                                                                               'links': [       {       'connected_time': '688344',
                                                                                                                        'fiber_length': 1130,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REY4cGVlcnBvbnBvbiAxMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '42 Main St S Milverton (Cook '
                                                 'Millwrighting 10401478)',
                                         'poller_hash': 'MjA1REY4',
                                         'serial_number': '205DF8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:8d',
                                         'uptime': '688344'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '14/g1',
                                                                       'name': '14/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVFQTJEcGVlcmludGYxNC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '14',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzVFQTJEcGVlcmludGZwb24gMTQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '14',
                                                                                               'links': [       {       'connected_time': '688575',
                                                                                                                        'fiber_length': 242,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQTJEcGVlcnBvbnBvbiAxNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': 'INVENTORY ROOM (IPTV RESET)',
                                         'poller_hash': 'MzVFQTJE',
                                         'serial_number': '35EA2D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0b:a7',
                                         'uptime': '688575'},
                                 {       'counter_source': '15/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '15/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '885',
                                                                       'in_octets': '166328916693',
                                                                       'in_ucast_pkts': '535029877',
                                                                       'mac_address': 'ec:4f:82:8d:ed:d8',
                                                                       'name': '15/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '3662',
                                                                       'out_octets': '1972200021367',
                                                                       'out_ucast_pkts': '1514891048',
                                                                       'poller_hash': 'MzIyNjAycGVlcmludGYxNS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '15',
                                                                       'in_nucast_pkts': '3662',
                                                                       'in_octets': '1972200021367',
                                                                       'in_ucast_pkts': '1514891048',
                                                                       'links': [       {       'in_nucast_pkts': '3662',
                                                                                                'in_octets': '1972200021367',
                                                                                                'in_ucast_pkts': '1514891048',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '885',
                                                                                                'out_octets': '166328916693',
                                                                                                'out_ucast_pkts': '535029877',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '885',
                                                                       'out_octets': '166328916693',
                                                                       'out_ucast_pkts': '535029877',
                                                                       'poller_hash': 'MzIyNjAycGVlcmludGZwb24gMTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '15',
                                                                                               'links': [       {       'connected_time': '688464',
                                                                                                                        'fiber_length': 1255,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNjAycGVlcnBvbnBvbiAxNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '51 WILLIAM ST (ENNS JACOB)',
                                         'poller_hash': 'MzIyNjAy',
                                         'serial_number': '322602',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:ed:d8',
                                         'uptime': '688464'},
                                 {       'counter_source': '16/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '0c:80:63:34:53:53'],
                                                                       'id': '16/g1',
                                                                       'in_nucast_pkts': '537',
                                                                       'in_octets': '35301706136',
                                                                       'in_ucast_pkts': '113866372',
                                                                       'mac_address': 'cc:be:59:50:16:6e',
                                                                       'mtu': 2000,
                                                                       'name': '16/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102703',
                                                                       'out_octets': '273273291585',
                                                                       'out_ucast_pkts': '196758694',
                                                                       'poller_hash': 'NDNDMUI4cGVlcmludGYxNi9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '16',
                                                                       'in_nucast_pkts': '1102703',
                                                                       'in_octets': '273273291585',
                                                                       'in_ucast_pkts': '196758694',
                                                                       'links': [       {       'in_nucast_pkts': '1102703',
                                                                                                'in_octets': '273273291585',
                                                                                                'in_ucast_pkts': '196758694',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '537',
                                                                                                'out_octets': '35301706136',
                                                                                                'out_ucast_pkts': '113866372',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '537',
                                                                       'out_octets': '35301706136',
                                                                       'out_ucast_pkts': '113866372',
                                                                       'poller_hash': 'NDNDMUI4cGVlcmludGZwb24gMTY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '16',
                                                                                               'links': [       {       'connected_time': '688401',
                                                                                                                        'fiber_length': 2113,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.354,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.134}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNDMUI4cGVlcnBvbnBvbiAxNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '78 GERBER DR (LAURIAULT NICOLE)',
                                         'poller_hash': 'NDNDMUI4',
                                         'serial_number': '43C1B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:16:6d',
                                         'uptime': '688401'},
                                 {       'counter_source': '17/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:37:b0'],
                                                                       'id': '17/g1',
                                                                       'in_nucast_pkts': '11646',
                                                                       'in_octets': '7657073813',
                                                                       'in_ucast_pkts': '19461340',
                                                                       'mac_address': 'd0:76:8f:1b:9f:1e',
                                                                       'mtu': 2000,
                                                                       'name': '17/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1433870',
                                                                       'out_octets': '133914192602',
                                                                       'out_ucast_pkts': '94405904',
                                                                       'poller_hash': 'ODcyNDMxcGVlcmludGYxNy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '17',
                                                                       'in_nucast_pkts': '1433870',
                                                                       'in_octets': '133914192602',
                                                                       'in_ucast_pkts': '94405904',
                                                                       'links': [       {       'in_nucast_pkts': '1433870',
                                                                                                'in_octets': '133914192602',
                                                                                                'in_ucast_pkts': '94405904',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11646',
                                                                                                'out_octets': '7657073813',
                                                                                                'out_ucast_pkts': '19461340',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11646',
                                                                       'out_octets': '7657073813',
                                                                       'out_ucast_pkts': '19461340',
                                                                       'poller_hash': 'ODcyNDMxcGVlcmludGZwb24gMTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '17',
                                                                                               'links': [       {       'connected_time': '688551',
                                                                                                                        'fiber_length': 1730,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -25.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDMxcGVlcnBvbnBvbiAxNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 MURRAY CT (BOWMAN JOHN)',
                                         'poller_hash': 'ODcyNDMx',
                                         'serial_number': '872431',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:1b',
                                         'uptime': '688551'},
                                 {       'counter_source': '19/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:54:0d'],
                                                                       'id': '19/g1',
                                                                       'in_nucast_pkts': '56266',
                                                                       'in_octets': '8044049789',
                                                                       'in_ucast_pkts': '34825580',
                                                                       'mac_address': 'd0:76:8f:21:4c:36',
                                                                       'mtu': 2000,
                                                                       'name': '19/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6626651',
                                                                       'out_octets': '272900419322',
                                                                       'out_ucast_pkts': '198706415',
                                                                       'poller_hash': 'ODdGOTY1cGVlcmludGYxOS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '19',
                                                                       'in_nucast_pkts': '6626651',
                                                                       'in_octets': '272900419322',
                                                                       'in_ucast_pkts': '198706415',
                                                                       'links': [       {       'in_nucast_pkts': '6626651',
                                                                                                'in_octets': '272900419322',
                                                                                                'in_ucast_pkts': '198706415',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '56266',
                                                                                                'out_octets': '8044049789',
                                                                                                'out_ucast_pkts': '34825580',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '56266',
                                                                       'out_octets': '8044049789',
                                                                       'out_ucast_pkts': '34825580',
                                                                       'poller_hash': 'ODdGOTY1cGVlcmludGZwb24gMTk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '19',
                                                                                               'links': [       {       'connected_time': '3113758',
                                                                                                                        'fiber_length': 910,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODdGOTY1cGVlcnBvbnBvbiAxOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '18 Mill St. W. (MCCL Shop)',
                                         'poller_hash': 'ODdGOTY1',
                                         'serial_number': '87F965',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:21:4c:33',
                                         'uptime': '3113758'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '20/g1',
                                                                       'name': '20/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUQwNTAzcGVlcmludGYyMC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '20',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MUQwNTAzcGVlcmludGZwb24gMjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '20',
                                                                                               'links': [       {       'connected_time': '15499140',
                                                                                                                        'fiber_length': 1192,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTAzcGVlcnBvbnBvbiAyMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '8 ADELAIDE ST',
                                         'poller_hash': 'MUQwNTAz',
                                         'serial_number': '1D0503',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:d1',
                                         'uptime': '15499140'},
                                 {       'counter_source': '22/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '22/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1229653996542',
                                                                       'in_ucast_pkts': '19690654995',
                                                                       'mac_address': '00:06:31:e6:b1:01',
                                                                       'name': '22/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '14066611180403',
                                                                       'out_ucast_pkts': '10406457495',
                                                                       'poller_hash': 'MUNDNEIzcGVlcmludGYyMi9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '22',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '14066611180403',
                                                                       'in_ucast_pkts': '10406457495',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '14066611180403',
                                                                                                'in_ucast_pkts': '10406457495',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1229653996542',
                                                                                                'out_ucast_pkts': '19690654995',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1229653996542',
                                                                       'out_ucast_pkts': '19690654995',
                                                                       'poller_hash': 'MUNDNEIzcGVlcmludGZwb24gMjI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '22',
                                                                                               'links': [       {       'connected_time': '72953878',
                                                                                                                        'fiber_length': 1473,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUNDNEIzcGVlcnBvbnBvbiAyMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '6 Macbeth St (Dan Gerber 10400548)',
                                         'poller_hash': 'MUNDNEIz',
                                         'serial_number': '1CC4B3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e6:b1:01',
                                         'uptime': '72953878'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '23/g1',
                                                                       'name': '23/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUQwNEVFcGVlcmludGYyMy9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '23',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUQwNEVFcGVlcmludGZwb24gMjM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '23',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEVFcGVlcnBvbnBvbiAyMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '47 William St. Milverton (SABRI '
                                                 'AKHMAD)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "23" (location: '
                                                                  'GPON port "7/5"), '
                                                                  '"ONT Dying Gasp"',
                                         'offline_time': '2025-07-18T18:45:32.930Z',
                                         'poller_hash': 'MUQwNEVF',
                                         'serial_number': '1D04EE',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '24/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '24/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '588',
                                                                       'in_octets': '93161630072',
                                                                       'in_ucast_pkts': '681148045',
                                                                       'mac_address': '00:06:31:e9:01:09',
                                                                       'name': '24/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '3569',
                                                                       'out_octets': '4149987183511',
                                                                       'out_ucast_pkts': '6991802746',
                                                                       'poller_hash': 'MUQwNjdCcGVlcmludGYyNC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '24',
                                                                       'in_nucast_pkts': '3569',
                                                                       'in_octets': '4149987183511',
                                                                       'in_ucast_pkts': '6991802746',
                                                                       'links': [       {       'in_nucast_pkts': '3569',
                                                                                                'in_octets': '4149987183511',
                                                                                                'in_ucast_pkts': '6991802746',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '588',
                                                                                                'out_octets': '93161630072',
                                                                                                'out_ucast_pkts': '681148045',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '588',
                                                                       'out_octets': '93161630072',
                                                                       'out_ucast_pkts': '681148045',
                                                                       'poller_hash': 'MUQwNjdCcGVlcmludGZwb24gMjQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '24',
                                                                                               'links': [       {       'connected_time': '31213046',
                                                                                                                        'fiber_length': 1321,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjdCcGVlcnBvbnBvbiAyNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': 'William Nahrgang (12 George)',
                                         'poller_hash': 'MUQwNjdC',
                                         'serial_number': '1D067B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:09',
                                         'uptime': '31213046'},
                                 {       'counter_source': '25/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '25/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '174165131701',
                                                                       'in_ucast_pkts': '781473559',
                                                                       'mac_address': '00:06:31:e8:f3:26',
                                                                       'name': '25/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '173',
                                                                       'out_octets': '2352370948535',
                                                                       'out_ucast_pkts': '1811737325',
                                                                       'poller_hash': 'MUQwNEYwcGVlcmludGYyNS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '25',
                                                                       'in_nucast_pkts': '173',
                                                                       'in_octets': '2352370948535',
                                                                       'in_ucast_pkts': '1811737325',
                                                                       'links': [       {       'in_nucast_pkts': '173',
                                                                                                'in_octets': '2352370948535',
                                                                                                'in_ucast_pkts': '1811737325',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '174165131701',
                                                                                                'out_ucast_pkts': '781473559',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '174165131701',
                                                                       'out_ucast_pkts': '781473559',
                                                                       'poller_hash': 'MUQwNEYwcGVlcmludGZwb24gMjU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '25',
                                                                                               'links': [       {       'connected_time': '21815830',
                                                                                                                        'fiber_length': 1024,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEYwcGVlcnBvbnBvbiAyNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '27 William St. Milverton (Donald '
                                                 'Schneider)',
                                         'poller_hash': 'MUQwNEYw',
                                         'serial_number': '1D04F0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:26',
                                         'uptime': '21815830'},
                                 {       'counter_source': '26/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:57:30'],
                                                                       'id': '26/g1',
                                                                       'in_nucast_pkts': '11998',
                                                                       'in_octets': '1301596135',
                                                                       'in_ucast_pkts': '7106502',
                                                                       'mac_address': '00:06:31:e8:e4:38',
                                                                       'mtu': 2000,
                                                                       'name': '26/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463574',
                                                                       'out_octets': '77772694304',
                                                                       'out_ucast_pkts': '54485199',
                                                                       'poller_hash': 'MUQwMzQ3cGVlcmludGYyNi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '26',
                                                                       'in_nucast_pkts': '1463574',
                                                                       'in_octets': '77772694304',
                                                                       'in_ucast_pkts': '54485199',
                                                                       'links': [       {       'in_nucast_pkts': '1463574',
                                                                                                'in_octets': '77772694304',
                                                                                                'in_ucast_pkts': '54485199',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11998',
                                                                                                'out_octets': '1301596135',
                                                                                                'out_ucast_pkts': '7106502',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11998',
                                                                       'out_octets': '1301596135',
                                                                       'out_ucast_pkts': '7106502',
                                                                       'poller_hash': 'MUQwMzQ3cGVlcmludGZwb24gMjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '26',
                                                                                               'links': [       {       'connected_time': '688456',
                                                                                                                        'fiber_length': 1199,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwMzQ3cGVlcnBvbnBvbiAyNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '30 WILLIAM ST (HARRIS CASSIDY)',
                                         'poller_hash': 'MUQwMzQ3',
                                         'serial_number': '1D0347',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:e4:35',
                                         'uptime': '688456'},
                                 {       'counter_source': '27/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '27/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '294',
                                                                       'in_octets': '227544243574',
                                                                       'in_ucast_pkts': '2239265737',
                                                                       'mac_address': '00:06:31:e9:01:75',
                                                                       'name': '27/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1342',
                                                                       'out_octets': '14387907269015',
                                                                       'out_ucast_pkts': '9809126788',
                                                                       'poller_hash': 'MUQwNjg3cGVlcmludGYyNy9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '27',
                                                                       'in_nucast_pkts': '1342',
                                                                       'in_octets': '14387907269015',
                                                                       'in_ucast_pkts': '9809126788',
                                                                       'links': [       {       'in_nucast_pkts': '1342',
                                                                                                'in_octets': '14387907269015',
                                                                                                'in_ucast_pkts': '9809126788',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '294',
                                                                                                'out_octets': '227544243574',
                                                                                                'out_ucast_pkts': '2239265737',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '294',
                                                                       'out_octets': '227544243574',
                                                                       'out_ucast_pkts': '2239265737',
                                                                       'poller_hash': 'MUQwNjg3cGVlcmludGZwb24gMjc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '27',
                                                                                               'links': [       {       'connected_time': '1118630',
                                                                                                                        'fiber_length': 1201,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjg3cGVlcnBvbnBvbiAyNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '42 William St. Milverton (Robert '
                                                 'Sippel)',
                                         'poller_hash': 'MUQwNjg3',
                                         'serial_number': '1D0687',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:75',
                                         'uptime': '1118630'},
                                 {       'counter_source': '28/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d3:0f'],
                                                                       'id': '28/g1',
                                                                       'in_nucast_pkts': '917110',
                                                                       'in_octets': '391683198216',
                                                                       'in_ucast_pkts': '1165101432',
                                                                       'mac_address': '00:06:31:e9:01:d2',
                                                                       'mtu': 2000,
                                                                       'name': '28/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '145796909',
                                                                       'out_octets': '9857876817968',
                                                                       'out_ucast_pkts': '7007047530',
                                                                       'poller_hash': 'MUQwNjkxcGVlcmludGYyOC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '28',
                                                                       'in_nucast_pkts': '145796909',
                                                                       'in_octets': '9857876817968',
                                                                       'in_ucast_pkts': '7007047530',
                                                                       'links': [       {       'in_nucast_pkts': '145796909',
                                                                                                'in_octets': '9857876817968',
                                                                                                'in_ucast_pkts': '7007047530',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '917110',
                                                                                                'out_octets': '391683198216',
                                                                                                'out_ucast_pkts': '1165101432',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '917110',
                                                                       'out_octets': '391683198216',
                                                                       'out_ucast_pkts': '1165101432',
                                                                       'poller_hash': 'MUQwNjkxcGVlcmludGZwb24gMjg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '28',
                                                                                               'links': [       {       'connected_time': '21817073',
                                                                                                                        'fiber_length': 1253,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -7.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjkxcGVlcnBvbnBvbiAyOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '46 William St. Milverton (BRUBACHER '
                                                 'MARK)',
                                         'poller_hash': 'MUQwNjkx',
                                         'serial_number': '1D0691',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:cf',
                                         'uptime': '21817073'},
                                 {       'counter_source': '29/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:c7:1e'],
                                                                       'id': '29/g1',
                                                                       'in_nucast_pkts': '3586163',
                                                                       'in_octets': '237303090972',
                                                                       'in_ucast_pkts': '3260542452',
                                                                       'mac_address': '00:06:31:e8:f3:95',
                                                                       'mtu': 2000,
                                                                       'name': '29/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '209299556',
                                                                       'out_octets': '15041056609740',
                                                                       'out_ucast_pkts': '10647145217',
                                                                       'poller_hash': 'MUQwNEZDcGVlcmludGYyOS9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '29',
                                                                       'in_nucast_pkts': '209299556',
                                                                       'in_octets': '15041056609740',
                                                                       'in_ucast_pkts': '10647145217',
                                                                       'links': [       {       'in_nucast_pkts': '209299556',
                                                                                                'in_octets': '15041056609740',
                                                                                                'in_ucast_pkts': '10647145217',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3586163',
                                                                                                'out_octets': '237303090972',
                                                                                                'out_ucast_pkts': '3260542452',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3586163',
                                                                       'out_octets': '237303090972',
                                                                       'out_ucast_pkts': '3260542452',
                                                                       'poller_hash': 'MUQwNEZDcGVlcmludGZwb24gMjk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '29',
                                                                                               'links': [       {       'connected_time': '68955576',
                                                                                                                        'fiber_length': 1042,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEZDcGVlcnBvbnBvbiAyOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '3 John St. Milverton (Fred Jones '
                                                 '10401195)',
                                         'poller_hash': 'MUQwNEZD',
                                         'serial_number': '1D04FC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:92',
                                         'uptime': '68955576'},
                                 {       'counter_source': '31/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:e2:b8'],
                                                                       'id': '31/g1',
                                                                       'in_nucast_pkts': '751308',
                                                                       'in_octets': '241509603118',
                                                                       'in_ucast_pkts': '1921402154',
                                                                       'mac_address': '00:06:31:e8:f3:3b',
                                                                       'mtu': 2000,
                                                                       'name': '31/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '102597507',
                                                                       'out_octets': '10672929883136',
                                                                       'out_ucast_pkts': '7367336187',
                                                                       'poller_hash': 'MUQwNEYycGVlcmludGYzMS9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '31',
                                                                       'in_nucast_pkts': '102597507',
                                                                       'in_octets': '10672929883136',
                                                                       'in_ucast_pkts': '7367336187',
                                                                       'links': [       {       'in_nucast_pkts': '102597507',
                                                                                                'in_octets': '10672929883136',
                                                                                                'in_ucast_pkts': '7367336187',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '751308',
                                                                                                'out_octets': '241509603118',
                                                                                                'out_ucast_pkts': '1921402154',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '751308',
                                                                       'out_octets': '241509603118',
                                                                       'out_ucast_pkts': '1921402154',
                                                                       'poller_hash': 'MUQwNEYycGVlcmludGZwb24gMzE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '31',
                                                                                               'links': [       {       'connected_time': '1142071',
                                                                                                                        'fiber_length': 1168,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEYycGVlcnBvbnBvbiAzMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '49 WILLIAM ST (BOWMAN LAURA)',
                                         'poller_hash': 'MUQwNEYy',
                                         'serial_number': '1D04F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:38',
                                         'uptime': '1142071'},
                                 {       'counter_source': '34/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '34/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '349251449737',
                                                                       'in_ucast_pkts': '4236197465',
                                                                       'mac_address': '00:06:31:b6:8c:ec',
                                                                       'name': '34/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25419667488201',
                                                                       'out_ucast_pkts': '16886663448',
                                                                       'poller_hash': 'MTU1M0Q0cGVlcmludGYzNC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '34',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25419667488201',
                                                                       'in_ucast_pkts': '16886663448',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '25419667488201',
                                                                                                'in_ucast_pkts': '16886663448',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '349251449737',
                                                                                                'out_ucast_pkts': '4236197465',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '349251449737',
                                                                       'out_ucast_pkts': '4236197465',
                                                                       'poller_hash': 'MTU1M0Q0cGVlcmludGZwb24gMzQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '34',
                                                                                               'links': [       {       'connected_time': '13219168',
                                                                                                                        'fiber_length': 1576,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.206,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1M0Q0cGVlcnBvbnBvbiAzNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '60 William St. Suite 112 (DEZELL '
                                                 'BERNADETTE)',
                                         'poller_hash': 'MTU1M0Q0',
                                         'serial_number': '1553D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:3a:1a',
                                         'uptime': '13219168'},
                                 {       'counter_source': '35/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:d8:e4:4c'],
                                                                       'id': '35/g2',
                                                                       'in_nucast_pkts': '733528',
                                                                       'in_octets': '72368341435',
                                                                       'in_ucast_pkts': '941931829',
                                                                       'mac_address': '00:06:31:b6:a2:91',
                                                                       'mtu': 2000,
                                                                       'name': '35/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '101203936',
                                                                       'out_octets': '5164641026899',
                                                                       'out_ucast_pkts': '3646209021',
                                                                       'poller_hash': 'MTU1MzIycGVlcmludGYzNS9nMg==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '35',
                                                                       'in_nucast_pkts': '101203936',
                                                                       'in_octets': '5164641026899',
                                                                       'in_ucast_pkts': '3646209021',
                                                                       'links': [       {       'in_nucast_pkts': '101203936',
                                                                                                'in_octets': '5164641026899',
                                                                                                'in_ucast_pkts': '3646209021',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '733528',
                                                                                                'out_octets': '72368341435',
                                                                                                'out_ucast_pkts': '941931829',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '733528',
                                                                       'out_octets': '72368341435',
                                                                       'out_ucast_pkts': '941931829',
                                                                       'poller_hash': 'MTU1MzIycGVlcmludGZwb24gMzU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '35',
                                                                                               'links': [       {       'connected_time': '13219219',
                                                                                                                        'fiber_length': 1630,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.302,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1MzIycGVlcnBvbnBvbiAzNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '60 William St. Suite 108 (Betty '
                                                 'Brubacher 10401029)',
                                         'poller_hash': 'MTU1MzIy',
                                         'serial_number': '155322',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:39:68',
                                         'uptime': '13219219'},
                                 {       'counter_source': '36/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '36/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '184801535871',
                                                                       'in_ucast_pkts': '432776860',
                                                                       'mac_address': '00:06:31:dd:56:a7',
                                                                       'name': '36/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3271921717978',
                                                                       'out_ucast_pkts': '2491558784',
                                                                       'poller_hash': 'MTc1RTZBcGVlcmludGYzNi9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '36',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3271921717978',
                                                                       'in_ucast_pkts': '2491558784',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3271921717978',
                                                                                                'in_ucast_pkts': '2491558784',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '184801535871',
                                                                                                'out_ucast_pkts': '432776860',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '184801535871',
                                                                       'out_ucast_pkts': '432776860',
                                                                       'poller_hash': 'MTc1RTZBcGVlcmludGZwb24gMzY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '36',
                                                                                               'links': [       {       'connected_time': '3534799',
                                                                                                                        'fiber_length': 1618,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc1RTZBcGVlcnBvbnBvbiAzNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '60 William St. Suite 106 (Clara '
                                                 'Looser 10400400)',
                                         'poller_hash': 'MTc1RTZB',
                                         'serial_number': '175E6A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:56:a7',
                                         'uptime': '3534799'},
                                 {       'counter_source': '37/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6a:8b'],
                                                                       'id': '37/g1',
                                                                       'in_nucast_pkts': '256166',
                                                                       'in_octets': '33853620148',
                                                                       'in_ucast_pkts': '314071235',
                                                                       'mac_address': '00:06:31:b6:8a:48',
                                                                       'mtu': 2000,
                                                                       'name': '37/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23533891',
                                                                       'out_octets': '2012667109402',
                                                                       'out_ucast_pkts': '1347545303',
                                                                       'poller_hash': 'MTU1NDM2cGVlcmludGYzNy9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '37',
                                                                       'in_nucast_pkts': '23533891',
                                                                       'in_octets': '2012667109402',
                                                                       'in_ucast_pkts': '1347545303',
                                                                       'links': [       {       'in_nucast_pkts': '23533891',
                                                                                                'in_octets': '2012667109402',
                                                                                                'in_ucast_pkts': '1347545303',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '256166',
                                                                                                'out_octets': '33853620148',
                                                                                                'out_ucast_pkts': '314071235',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '256166',
                                                                       'out_octets': '33853620148',
                                                                       'out_ucast_pkts': '314071235',
                                                                       'poller_hash': 'MTU1NDM2cGVlcmludGZwb24gMzc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '37',
                                                                                               'links': [       {       'connected_time': '13219203',
                                                                                                                        'fiber_length': 1592,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.422,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1NDM2cGVlcnBvbnBvbiAzNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '60 William St. Suite 111 (ERB '
                                                 'CONNIE)',
                                         'poller_hash': 'MTU1NDM2',
                                         'serial_number': '155436',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:3a:7c',
                                         'uptime': '13219203'},
                                 {       'counter_source': '38/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '38/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '322585475891',
                                                                       'in_ucast_pkts': '3487590285',
                                                                       'mac_address': '00:06:31:dd:56:e7',
                                                                       'name': '38/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10290881112838',
                                                                       'out_ucast_pkts': '2679847522',
                                                                       'poller_hash': 'MTc1RTcycGVlcmludGYzOC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '38',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10290881112838',
                                                                       'in_ucast_pkts': '2679847522',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '10290881112838',
                                                                                                'in_ucast_pkts': '2679847522',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '322585475891',
                                                                                                'out_ucast_pkts': '3487590285',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '322585475891',
                                                                       'out_ucast_pkts': '3487590285',
                                                                       'poller_hash': 'MTc1RTcycGVlcmludGZwb24gMzg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '38',
                                                                                               'links': [       {       'connected_time': '64764785',
                                                                                                                        'fiber_length': 1596,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc1RTcycGVlcnBvbnBvbiAzOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '60 WILLIAM ST (ZEHR SALLY)',
                                         'poller_hash': 'MTc1RTcy',
                                         'serial_number': '175E72',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:56:e7',
                                         'uptime': '64764785'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '39/g1',
                                                                       'name': '39/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTc2MjM5cGVlcmludGYzOS9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '39',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTc2MjM5cGVlcmludGZwb24gMzk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '39',
                                                                                               'links': [       {       'connected_time': '15564960',
                                                                                                                        'fiber_length': 1611,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc2MjM5cGVlcnBvbnBvbiAzOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '60 William St. Suite 109',
                                         'poller_hash': 'MTc2MjM5',
                                         'serial_number': '176239',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:75:1f',
                                         'uptime': '15564960'},
                                 {       'counter_source': '40/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:71:6b'],
                                                                       'id': '40/g1',
                                                                       'in_nucast_pkts': '260480',
                                                                       'in_octets': '3936917756',
                                                                       'in_ucast_pkts': '13624731',
                                                                       'mac_address': '00:06:31:87:4f:e7',
                                                                       'mtu': 2000,
                                                                       'name': '40/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23541529',
                                                                       'out_octets': '45945615152',
                                                                       'out_ucast_pkts': '38363358',
                                                                       'poller_hash': 'Q0MwQzVwZWVyaW50ZjQwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '40',
                                                                       'in_nucast_pkts': '23541529',
                                                                       'in_octets': '45945615152',
                                                                       'in_ucast_pkts': '38363358',
                                                                       'links': [       {       'in_nucast_pkts': '23541529',
                                                                                                'in_octets': '45945615152',
                                                                                                'in_ucast_pkts': '38363358',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '260480',
                                                                                                'out_octets': '3936917756',
                                                                                                'out_ucast_pkts': '13624731',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '260480',
                                                                       'out_octets': '3936917756',
                                                                       'out_ucast_pkts': '13624731',
                                                                       'poller_hash': 'Q0MwQzVwZWVyaW50ZnBvbiA0MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '40',
                                                                                               'links': [       {       'connected_time': '13219099',
                                                                                                                        'fiber_length': 1587,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQzVwZWVycG9ucG9uIDQw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '60 WILLIAM ST (JANTZI AARON '
                                                 'J/MARTHA)',
                                         'poller_hash': 'Q0MwQzU=',
                                         'serial_number': 'CC0C5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:85',
                                         'uptime': '13219099'},
                                 {       'counter_source': '41/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '41/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '113089914963',
                                                                       'in_ucast_pkts': '1228128345',
                                                                       'mac_address': 'ec:4f:82:08:27:9d',
                                                                       'name': '41/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6749315557507',
                                                                       'out_ucast_pkts': '4638549040',
                                                                       'poller_hash': 'MUM1Q0U5cGVlcmludGY0MS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '41',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6749315557507',
                                                                       'in_ucast_pkts': '4638549040',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6749315557507',
                                                                                                'in_ucast_pkts': '4638549040',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '113089914963',
                                                                                                'out_ucast_pkts': '1228128345',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '113089914963',
                                                                       'out_ucast_pkts': '1228128345',
                                                                       'poller_hash': 'MUM1Q0U5cGVlcmludGZwb24gNDE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '41',
                                                                                               'links': [       {       'connected_time': '13219272',
                                                                                                                        'fiber_length': 1583,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.54,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUM1Q0U5cGVlcnBvbnBvbiA0MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '60 WILLIAM ST (SCHWEITZER DIANE)',
                                         'poller_hash': 'MUM1Q0U5',
                                         'serial_number': '1C5CE9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:08:27:9d',
                                         'uptime': '13219272'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '42/g1',
                                                                       'name': '42/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUQwNTBEcGVlcmludGY0Mi9nMQ=='},
                                                               {       'admin_up': False,
                                                                       'id': '42',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUQwNTBEcGVlcmludGZwb24gNDI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '42',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTBEcGVlcnBvbnBvbiA0Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '11 George St. Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "42" (location: '
                                                                  'GPON port "7/6"), '
                                                                  '"ONT went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:44.910Z',
                                         'poller_hash': 'MUQwNTBE',
                                         'serial_number': '1D050D',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '43/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '43/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '195643914613',
                                                                       'in_ucast_pkts': '453135759',
                                                                       'mac_address': '00:06:31:e8:e4:23',
                                                                       'name': '43/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '138',
                                                                       'out_octets': '1862843802817',
                                                                       'out_ucast_pkts': '1533788173',
                                                                       'poller_hash': 'MUQwMzQ1cGVlcmludGY0My9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '43',
                                                                       'in_nucast_pkts': '138',
                                                                       'in_octets': '1862843802817',
                                                                       'in_ucast_pkts': '1533788173',
                                                                       'links': [       {       'in_nucast_pkts': '138',
                                                                                                'in_octets': '1862843802817',
                                                                                                'in_ucast_pkts': '1533788173',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '195643914613',
                                                                                                'out_ucast_pkts': '453135759',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '195643914613',
                                                                       'out_ucast_pkts': '453135759',
                                                                       'poller_hash': 'MUQwMzQ1cGVlcmludGZwb24gNDM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '43',
                                                                                               'links': [       {       'connected_time': '12019959',
                                                                                                                        'fiber_length': 837,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwMzQ1cGVlcnBvbnBvbiA0Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '16 WILLIAM ST (CHAINEY KEVIN)',
                                         'poller_hash': 'MUQwMzQ1',
                                         'serial_number': '1D0345',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:e4:23',
                                         'uptime': '12019959'},
                                 {       'counter_source': '44/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '44/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '294',
                                                                       'in_octets': '281450057321',
                                                                       'in_ucast_pkts': '3035056868',
                                                                       'mac_address': '00:06:31:e8:f3:5c',
                                                                       'name': '44/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1492',
                                                                       'out_octets': '12410677112663',
                                                                       'out_ucast_pkts': '8443358054',
                                                                       'poller_hash': 'MUQwNEY2cGVlcmludGY0NC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '44',
                                                                       'in_nucast_pkts': '1492',
                                                                       'in_octets': '12410677112663',
                                                                       'in_ucast_pkts': '8443358054',
                                                                       'links': [       {       'in_nucast_pkts': '1492',
                                                                                                'in_octets': '12410677112663',
                                                                                                'in_ucast_pkts': '8443358054',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '294',
                                                                                                'out_octets': '281450057321',
                                                                                                'out_ucast_pkts': '3035056868',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '294',
                                                                       'out_octets': '281450057321',
                                                                       'out_ucast_pkts': '3035056868',
                                                                       'poller_hash': 'MUQwNEY2cGVlcmludGZwb24gNDQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '44',
                                                                                               'links': [       {       'connected_time': '21817303',
                                                                                                                        'fiber_length': 1182,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEY2cGVlcnBvbnBvbiA0NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '40 William St. Milverton (Merton '
                                                 'Kuepfer)',
                                         'poller_hash': 'MUQwNEY2',
                                         'serial_number': '1D04F6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:5c',
                                         'uptime': '21817303'},
                                 {       'counter_source': '45/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ea:76'],
                                                                       'id': '45/g1',
                                                                       'in_nucast_pkts': '580811',
                                                                       'in_octets': '387021719585',
                                                                       'in_ucast_pkts': '1948038011',
                                                                       'mac_address': 'ec:4f:82:0d:6a:c6',
                                                                       'mtu': 2000,
                                                                       'name': '45/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68592547',
                                                                       'out_octets': '10752559959640',
                                                                       'out_ucast_pkts': '7441000720',
                                                                       'poller_hash': 'MjA1REZFcGVlcmludGY0NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '45',
                                                                       'in_nucast_pkts': '68592547',
                                                                       'in_octets': '10752559959640',
                                                                       'in_ucast_pkts': '7441000720',
                                                                       'links': [       {       'in_nucast_pkts': '68592547',
                                                                                                'in_octets': '10752559959640',
                                                                                                'in_ucast_pkts': '7441000720',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '580811',
                                                                                                'out_octets': '387021719585',
                                                                                                'out_ucast_pkts': '1948038011',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '580811',
                                                                       'out_octets': '387021719585',
                                                                       'out_ucast_pkts': '1948038011',
                                                                       'poller_hash': 'MjA1REZFcGVlcmludGZwb24gNDU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '45',
                                                                                               'links': [       {       'connected_time': '31303880',
                                                                                                                        'fiber_length': 1242,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REZFcGVlcnBvbnBvbiA0NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '4 George St. Milverton (Tom Broda)',
                                         'poller_hash': 'MjA1REZF',
                                         'serial_number': '205DFE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:c3',
                                         'uptime': '31303880'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '46/g1',
                                                                       'name': '46/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRUQ5cGVlcmludGY0Ni9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '46',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxRUQ5cGVlcmludGZwb24gNDY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '46',
                                                                                               'links': [       {       'connected_time': '6743564',
                                                                                                                        'fiber_length': 1542,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUQ5cGVlcnBvbnBvbiA0Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 Sippel St. Milverton (Gerald Bain '
                                                 '10430059)',
                                         'poller_hash': 'QzgxRUQ5',
                                         'serial_number': 'C81ED9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4b:7f',
                                         'uptime': '6743564'},
                                 {       'counter_source': '47/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '47/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '542240122260',
                                                                       'in_ucast_pkts': '135603381515',
                                                                       'mac_address': '00:06:31:e8:f3:b6',
                                                                       'name': '47/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '9085026882378',
                                                                       'out_ucast_pkts': '6819809591',
                                                                       'poller_hash': 'MUQwNTAwcGVlcmludGY0Ny9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '47',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '9085026882378',
                                                                       'in_ucast_pkts': '6819809591',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '9085026882378',
                                                                                                'in_ucast_pkts': '6819809591',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '542240122260',
                                                                                                'out_ucast_pkts': '135603381515',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '542240122260',
                                                                       'out_ucast_pkts': '135603381515',
                                                                       'poller_hash': 'MUQwNTAwcGVlcmludGZwb24gNDc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '47',
                                                                                               'links': [       {       'connected_time': '72476147',
                                                                                                                        'fiber_length': 1177,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTAwcGVlcnBvbnBvbiA0Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '10 Adelaide St. Milverton (Jeremy '
                                                 'Sippel 10420159)',
                                         'poller_hash': 'MUQwNTAw',
                                         'serial_number': '1D0500',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:b6',
                                         'uptime': '72476147'},
                                 {       'counter_source': '48/g2+48/g3+48/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:e2:1e'],
                                                                       'id': '48/g2',
                                                                       'in_nucast_pkts': '2114878',
                                                                       'in_octets': '6750445500',
                                                                       'in_ucast_pkts': '39664936',
                                                                       'mac_address': 'cc:be:59:5a:ff:2f',
                                                                       'mtu': 2000,
                                                                       'name': '48/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '184602633',
                                                                       'out_octets': '225925775488',
                                                                       'out_ucast_pkts': '119995339',
                                                                       'poller_hash': 'NDRBQUE4cGVlcmludGY0OC9nMg==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:ad:4e:18:9f:1c'],
                                                                       'id': '48/g3',
                                                                       'in_nucast_pkts': '469404',
                                                                       'in_octets': '8234534627',
                                                                       'in_ucast_pkts': '74359111',
                                                                       'mac_address': 'cc:be:59:5a:ff:30',
                                                                       'mtu': 2000,
                                                                       'name': '48/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '570989',
                                                                       'out_octets': '34701672334',
                                                                       'out_ucast_pkts': '85370157',
                                                                       'poller_hash': 'NDRBQUE4cGVlcmludGY0OC9nMw==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:fb:6f'],
                                                                       'id': '48/g4',
                                                                       'in_nucast_pkts': '21411470',
                                                                       'in_octets': '355519937208',
                                                                       'in_ucast_pkts': '1335679745',
                                                                       'mac_address': 'cc:be:59:5a:ff:31',
                                                                       'mtu': 2000,
                                                                       'name': '48/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '408454476',
                                                                       'out_octets': '4997402290505',
                                                                       'out_ucast_pkts': '3753714551',
                                                                       'poller_hash': 'NDRBQUE4cGVlcmludGY0OC9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '48',
                                                                       'in_nucast_pkts': '593628098',
                                                                       'in_octets': '5258029738327',
                                                                       'in_ucast_pkts': '3959080047',
                                                                       'links': [       {       'in_nucast_pkts': '593628098',
                                                                                                'in_octets': '5258029738327',
                                                                                                'in_ucast_pkts': '3959080047',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23995752',
                                                                                                'out_octets': '370504917335',
                                                                                                'out_ucast_pkts': '1449703792',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23995752',
                                                                       'out_octets': '370504917335',
                                                                       'out_ucast_pkts': '1449703792',
                                                                       'poller_hash': 'NDRBQUE4cGVlcmludGZwb24gNDg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '48',
                                                                                               'links': [       {       'connected_time': '24669263',
                                                                                                                        'fiber_length': 18295,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.424,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDRBQUE4cGVlcnBvbnBvbiA0OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '200 Nelson St. Monkton (North Perth '
                                                 'Elma Logan Rec Complex)',
                                         'poller_hash': 'NDRBQUE4',
                                         'serial_number': '44AAA8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:5a:ff:2a',
                                         'uptime': '24669263'},
                                 {       'counter_source': '49/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '49/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '231449039856',
                                                                       'in_ucast_pkts': '836425364',
                                                                       'mac_address': '00:06:31:e5:cc:bc',
                                                                       'name': '49/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '3190140125790',
                                                                       'out_ucast_pkts': '2544346770',
                                                                       'poller_hash': 'MUNBQjU2cGVlcmludGY0OS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '49',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '3190140125790',
                                                                       'in_ucast_pkts': '2544346770',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '3190140125790',
                                                                                                'in_ucast_pkts': '2544346770',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '231449039856',
                                                                                                'out_ucast_pkts': '836425364',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '231449039856',
                                                                       'out_ucast_pkts': '836425364',
                                                                       'poller_hash': 'MUNBQjU2cGVlcmludGZwb24gNDk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '49',
                                                                                               'links': [       {       'connected_time': '29272053',
                                                                                                                        'fiber_length': 1190,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUNBQjU2cGVlcnBvbnBvbiA0OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '34 William St. Milverton (Melanie '
                                                 'Sippel)',
                                         'poller_hash': 'MUNBQjU2',
                                         'serial_number': '1CAB56',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e5:cc:bc',
                                         'uptime': '29272053'},
                                 {       'counter_source': '50/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:13:0f'],
                                                                       'id': '50/g2',
                                                                       'in_nucast_pkts': '34',
                                                                       'in_octets': '32249782341',
                                                                       'in_ucast_pkts': '168635261',
                                                                       'mac_address': 'ec:4f:82:32:25:dc',
                                                                       'mtu': 2000,
                                                                       'name': '50/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5553990',
                                                                       'out_octets': '985441610547',
                                                                       'out_ucast_pkts': '697157302',
                                                                       'poller_hash': 'MjQ4QUNCcGVlcmludGY1MC9nMg==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '50',
                                                                       'in_nucast_pkts': '5553990',
                                                                       'in_octets': '985441610547',
                                                                       'in_ucast_pkts': '697157302',
                                                                       'links': [       {       'in_nucast_pkts': '5553990',
                                                                                                'in_octets': '985441610547',
                                                                                                'in_ucast_pkts': '697157302',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '34',
                                                                                                'out_octets': '32249782341',
                                                                                                'out_ucast_pkts': '168635261',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '34',
                                                                       'out_octets': '32249782341',
                                                                       'out_ucast_pkts': '168635261',
                                                                       'poller_hash': 'MjQ4QUNCcGVlcmludGZwb24gNTA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '50',
                                                                                               'links': [       {       'connected_time': '3294100',
                                                                                                                        'fiber_length': 1046,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.116,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjQ4QUNCcGVlcnBvbnBvbiA1MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '45A & 45B William St. (Norma& Neil '
                                                 'Wilker)',
                                         'poller_hash': 'MjQ4QUNC',
                                         'serial_number': '248ACB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:32:25:d7',
                                         'uptime': '3294100'},
                                 {       'counter_source': '51/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e5:f4'],
                                                                       'id': '51/g1',
                                                                       'in_nucast_pkts': '11976',
                                                                       'in_octets': '6354362074',
                                                                       'in_ucast_pkts': '46961627',
                                                                       'mac_address': 'ec:4f:82:5b:72:26',
                                                                       'mtu': 2000,
                                                                       'name': '51/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464148',
                                                                       'out_octets': '155316913847',
                                                                       'out_ucast_pkts': '111019602',
                                                                       'poller_hash': 'MkIwMUI3cGVlcmludGY1MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '51',
                                                                       'in_nucast_pkts': '1464148',
                                                                       'in_octets': '155316913847',
                                                                       'in_ucast_pkts': '111019602',
                                                                       'links': [       {       'in_nucast_pkts': '1464148',
                                                                                                'in_octets': '155316913847',
                                                                                                'in_ucast_pkts': '111019602',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11976',
                                                                                                'out_octets': '6354362074',
                                                                                                'out_ucast_pkts': '46961627',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11976',
                                                                       'out_octets': '6354362074',
                                                                       'out_ucast_pkts': '46961627',
                                                                       'poller_hash': 'MkIwMUI3cGVlcmludGZwb24gNTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '51',
                                                                                               'links': [       {       'connected_time': '688339',
                                                                                                                        'fiber_length': 1214,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMUI3cGVlcnBvbnBvbiA1MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '37 Temperance St. Milverton (Chuck '
                                                 'Meyers 10401071)',
                                         'poller_hash': 'MkIwMUI3',
                                         'serial_number': '2B01B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:72:23',
                                         'uptime': '688339'},
                                 {       'counter_source': '52/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ec:d3'],
                                                                       'id': '52/g1',
                                                                       'in_nucast_pkts': '14562',
                                                                       'in_octets': '20565259583',
                                                                       'in_ucast_pkts': '31378274',
                                                                       'mac_address': 'ec:4f:82:8d:79:c3',
                                                                       'mtu': 2000,
                                                                       'name': '52/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464003',
                                                                       'out_octets': '120910774583',
                                                                       'out_ucast_pkts': '99573494',
                                                                       'poller_hash': 'MzIxNzdGcGVlcmludGY1Mi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '52',
                                                                       'in_nucast_pkts': '1464003',
                                                                       'in_octets': '120910774583',
                                                                       'in_ucast_pkts': '99573494',
                                                                       'links': [       {       'in_nucast_pkts': '1464003',
                                                                                                'in_octets': '120910774583',
                                                                                                'in_ucast_pkts': '99573494',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14562',
                                                                                                'out_octets': '20565259583',
                                                                                                'out_ucast_pkts': '31378274',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14562',
                                                                       'out_octets': '20565259583',
                                                                       'out_ucast_pkts': '31378274',
                                                                       'poller_hash': 'MzIxNzdGcGVlcmludGZwb24gNTI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '52',
                                                                                               'links': [       {       'connected_time': '688478',
                                                                                                                        'fiber_length': 1013,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxNzdGcGVlcnBvbnBvbiA1Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 MILL ST W (SHANNON KEELIE)',
                                         'poller_hash': 'MzIxNzdG',
                                         'serial_number': '32177F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:79:c0',
                                         'uptime': '688478'},
                                 {       'counter_source': '53/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ee:a2'],
                                                                       'id': '53/g1',
                                                                       'in_nucast_pkts': '927303',
                                                                       'in_octets': '914699400952',
                                                                       'in_ucast_pkts': '2464535569',
                                                                       'mac_address': 'f8:85:f9:17:15:4b',
                                                                       'mtu': 2000,
                                                                       'name': '53/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149411972',
                                                                       'out_octets': '12316123095237',
                                                                       'out_ucast_pkts': '9285011437',
                                                                       'poller_hash': 'REZGQjkycGVlcmludGY1My9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '53',
                                                                       'in_nucast_pkts': '149411972',
                                                                       'in_octets': '12316123095237',
                                                                       'in_ucast_pkts': '9285011437',
                                                                       'links': [       {       'in_nucast_pkts': '149411972',
                                                                                                'in_octets': '12316123095237',
                                                                                                'in_ucast_pkts': '9285011437',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '927303',
                                                                                                'out_octets': '914699400952',
                                                                                                'out_ucast_pkts': '2464535569',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '927303',
                                                                       'out_octets': '914699400952',
                                                                       'out_ucast_pkts': '2464535569',
                                                                       'poller_hash': 'REZGQjkycGVlcmludGZwb24gNTM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '53',
                                                                                               'links': [       {       'connected_time': '21815840',
                                                                                                                        'fiber_length': 1282,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGQjkycGVlcnBvbnBvbiA1Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 Wilfred St. Milverton (Steve '
                                                 'Court 10403171)',
                                         'poller_hash': 'REZGQjky',
                                         'serial_number': 'DFFB92',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:15:48',
                                         'uptime': '21815840'},
                                 {       'counter_source': '54/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '54/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '163343893520',
                                                                       'in_ucast_pkts': '662137203',
                                                                       'mac_address': '00:06:31:e5:ca:a9',
                                                                       'name': '54/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '55',
                                                                       'out_octets': '4417468212601',
                                                                       'out_ucast_pkts': '3311553706',
                                                                       'poller_hash': 'MUNBQjFCcGVlcmludGY1NC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '54',
                                                                       'in_nucast_pkts': '55',
                                                                       'in_octets': '4417468212601',
                                                                       'in_ucast_pkts': '3311553706',
                                                                       'links': [       {       'in_nucast_pkts': '55',
                                                                                                'in_octets': '4417468212601',
                                                                                                'in_ucast_pkts': '3311553706',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '163343893520',
                                                                                                'out_ucast_pkts': '662137203',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '163343893520',
                                                                       'out_ucast_pkts': '662137203',
                                                                       'poller_hash': 'MUNBQjFCcGVlcmludGZwb24gNTQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '54',
                                                                                               'links': [       {       'connected_time': '688464',
                                                                                                                        'fiber_length': 1572,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUNBQjFCcGVlcnBvbnBvbiA1NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '5 SIPPEL ST (CHAPMAN MICHELE)',
                                         'poller_hash': 'MUNBQjFC',
                                         'serial_number': '1CAB1B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e5:ca:a9',
                                         'uptime': '688464'},
                                 {       'counter_source': '55/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:fb:7e'],
                                                                       'id': '55/g2',
                                                                       'in_nucast_pkts': '3388497',
                                                                       'in_octets': '95266532404',
                                                                       'in_ucast_pkts': '1186137348',
                                                                       'mac_address': 'ec:4f:82:08:26:12',
                                                                       'mtu': 2000,
                                                                       'name': '55/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '163153270',
                                                                       'out_octets': '6819039624998',
                                                                       'out_ucast_pkts': '4602117467',
                                                                       'poller_hash': 'MUM1Q0I3cGVlcmludGY1NS9nMg==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '55',
                                                                       'in_nucast_pkts': '163153270',
                                                                       'in_octets': '6819039624998',
                                                                       'in_ucast_pkts': '4602117467',
                                                                       'links': [       {       'in_nucast_pkts': '163153270',
                                                                                                'in_octets': '6819039624998',
                                                                                                'in_ucast_pkts': '4602117467',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3388497',
                                                                                                'out_octets': '95266532404',
                                                                                                'out_ucast_pkts': '1186137348',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3388497',
                                                                       'out_octets': '95266532404',
                                                                       'out_ucast_pkts': '1186137348',
                                                                       'poller_hash': 'MUM1Q0I3cGVlcmludGZwb24gNTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '55',
                                                                                               'links': [       {       'connected_time': '47626',
                                                                                                                        'fiber_length': 1616,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.892,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUM1Q0I3cGVlcnBvbnBvbiA1NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '60 WILLIAM ST (LUCKHARDT MARION)',
                                         'poller_hash': 'MUM1Q0I3',
                                         'serial_number': '1C5CB7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:08:26:0d',
                                         'uptime': '47626'},
                                 {       'counter_source': '56/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a8:e9'],
                                                                       'id': '56/g1',
                                                                       'in_nucast_pkts': '7032',
                                                                       'in_octets': '269238786612',
                                                                       'in_ucast_pkts': '910905115',
                                                                       'mac_address': 'ec:4f:82:8d:7c:fb',
                                                                       'mtu': 2000,
                                                                       'name': '56/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '144999065',
                                                                       'out_octets': '3217711317734',
                                                                       'out_ucast_pkts': '2568367563',
                                                                       'poller_hash': 'MzIxN0U2cGVlcmludGY1Ni9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '56',
                                                                       'in_nucast_pkts': '144999065',
                                                                       'in_octets': '3217711317734',
                                                                       'in_ucast_pkts': '2568367563',
                                                                       'links': [       {       'in_nucast_pkts': '144999065',
                                                                                                'in_octets': '3217711317734',
                                                                                                'in_ucast_pkts': '2568367563',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7032',
                                                                                                'out_octets': '269238786612',
                                                                                                'out_ucast_pkts': '910905115',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7032',
                                                                       'out_octets': '269238786612',
                                                                       'out_ucast_pkts': '910905115',
                                                                       'poller_hash': 'MzIxN0U2cGVlcmludGZwb24gNTY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '56',
                                                                                               'links': [       {       'connected_time': '21816603',
                                                                                                                        'fiber_length': 6346,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxN0U2cGVlcnBvbnBvbiA1Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4345 LINE 61 (ROPP DAVID & BONNIE)',
                                         'poller_hash': 'MzIxN0U2',
                                         'serial_number': '3217E6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:7c:f8',
                                         'uptime': '21816603'},
                                 {       'counter_source': '58/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:c7:aa'],
                                                                       'id': '58/g3',
                                                                       'in_nucast_pkts': '2741903',
                                                                       'in_octets': '2416585084',
                                                                       'in_ucast_pkts': '18653020',
                                                                       'mac_address': 'ec:4f:82:08:1d:8b',
                                                                       'mtu': 2000,
                                                                       'name': '58/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2224499440',
                                                                       'out_octets': '2868922608599',
                                                                       'out_ucast_pkts': '73599187',
                                                                       'poller_hash': 'MUM1QkE2cGVlcmludGY1OC9nMw==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '58',
                                                                       'in_nucast_pkts': '2224499440',
                                                                       'in_octets': '2868922608599',
                                                                       'in_ucast_pkts': '73599187',
                                                                       'links': [       {       'in_nucast_pkts': '2224499440',
                                                                                                'in_octets': '2868922608599',
                                                                                                'in_ucast_pkts': '73599187',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2741903',
                                                                                                'out_octets': '2416585084',
                                                                                                'out_ucast_pkts': '18653020',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2741903',
                                                                       'out_octets': '2416585084',
                                                                       'out_ucast_pkts': '18653020',
                                                                       'poller_hash': 'MUM1QkE2cGVlcmludGZwb24gNTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '58',
                                                                                               'links': [       {       'connected_time': '17711637',
                                                                                                                        'fiber_length': 1245,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.716,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUM1QkE2cGVlcnBvbnBvbiA1OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '40 Temperance St. Milverton '
                                                 '(Mornington Communications)',
                                         'poller_hash': 'MUM1QkE2',
                                         'serial_number': '1C5BA6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:08:1d:85',
                                         'uptime': '17711637'},
                                 {       'counter_source': '59/g1+59/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c8:4f:86:fc:00:03'],
                                                                       'id': '59/g1',
                                                                       'in_nucast_pkts': '721',
                                                                       'in_octets': '101620273057549',
                                                                       'in_ucast_pkts': '83704730985',
                                                                       'mac_address': '14:21:03:04:fc:15',
                                                                       'mtu': 2000,
                                                                       'name': '59/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53768699',
                                                                       'out_octets': '11619498796210',
                                                                       'out_ucast_pkts': '52006892846',
                                                                       'poller_hash': 'MTYwNjQyOHBlZXJpbnRmNTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:be:b4:13:b6:c0'],
                                                                       'id': '59/g4',
                                                                       'in_nucast_pkts': '820',
                                                                       'in_octets': '3989802808817',
                                                                       'in_ucast_pkts': '3403949814',
                                                                       'mac_address': '14:21:03:04:fc:18',
                                                                       'mtu': 2000,
                                                                       'name': '59/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53766790',
                                                                       'out_octets': '412922640847',
                                                                       'out_ucast_pkts': '927797559',
                                                                       'poller_hash': 'MTYwNjQyOHBlZXJpbnRmNTkvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '59',
                                                                       'in_nucast_pkts': '107535489',
                                                                       'in_octets': '12032421437057',
                                                                       'in_ucast_pkts': '52934690405',
                                                                       'links': [       {       'in_nucast_pkts': '107535489',
                                                                                                'in_octets': '12032421437057',
                                                                                                'in_ucast_pkts': '52934690405',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1541',
                                                                                                'out_octets': '105610075866366',
                                                                                                'out_ucast_pkts': '87108680799',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1541',
                                                                       'out_octets': '105610075866366',
                                                                       'out_ucast_pkts': '87108680799',
                                                                       'poller_hash': 'MTYwNjQyOHBlZXJpbnRmcG9uIDU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '59',
                                                                                               'links': [       {       'connected_time': '28789464',
                                                                                                                        'fiber_length': 1245,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.006,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTYwNjQyOHBlZXJwb25wb24gNTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '40 Temperance St. Milverton (Perth '
                                                 'East Complex 10401202)',
                                         'poller_hash': 'MTYwNjQyOA==',
                                         'serial_number': '1606428',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '14:21:03:04:fc:11',
                                         'uptime': '28789464'},
                                 {       'counter_source': '60/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1b:96'],
                                                                       'id': '60/g1',
                                                                       'in_nucast_pkts': '463446',
                                                                       'in_octets': '158186346784',
                                                                       'in_ucast_pkts': '1063847162',
                                                                       'mac_address': '00:06:31:e8:f3:ef',
                                                                       'mtu': 2000,
                                                                       'name': '60/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58266095',
                                                                       'out_octets': '6525792929733',
                                                                       'out_ucast_pkts': '4552215267',
                                                                       'poller_hash': 'MUQwNTA2cGVlcmludGY2MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '60',
                                                                       'in_nucast_pkts': '58266095',
                                                                       'in_octets': '6525792929733',
                                                                       'in_ucast_pkts': '4552215267',
                                                                       'links': [       {       'in_nucast_pkts': '58266095',
                                                                                                'in_octets': '6525792929733',
                                                                                                'in_ucast_pkts': '4552215267',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '463446',
                                                                                                'out_octets': '158186346784',
                                                                                                'out_ucast_pkts': '1063847162',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '463446',
                                                                       'out_octets': '158186346784',
                                                                       'out_ucast_pkts': '1063847162',
                                                                       'poller_hash': 'MUQwNTA2cGVlcmludGZwb24gNjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '60',
                                                                                               'links': [       {       'connected_time': '24544077',
                                                                                                                        'fiber_length': 1251,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTA2cGVlcnBvbnBvbiA2MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '25A Temperance St. Milverton (Vince '
                                                 'Etchells)',
                                         'poller_hash': 'MUQwNTA2',
                                         'serial_number': '1D0506',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:ec',
                                         'uptime': '24544077'},
                                 {       'counter_source': '61/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '61/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '179741728032',
                                                                       'in_ucast_pkts': '1222960041',
                                                                       'mac_address': '00:06:31:e8:f3:65',
                                                                       'name': '61/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '4341598309688',
                                                                       'out_ucast_pkts': '3169524762',
                                                                       'poller_hash': 'MUQwNEY3cGVlcmludGY2MS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '61',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '4341598309688',
                                                                       'in_ucast_pkts': '3169524762',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '4341598309688',
                                                                                                'in_ucast_pkts': '3169524762',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '179741728032',
                                                                                                'out_ucast_pkts': '1222960041',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '179741728032',
                                                                       'out_ucast_pkts': '1222960041',
                                                                       'poller_hash': 'MUQwNEY3cGVlcmludGZwb24gNjE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '61',
                                                                                               'links': [       {       'connected_time': '21816522',
                                                                                                                        'fiber_length': 1250,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEY3cGVlcnBvbnBvbiA2MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '25 Temperance St. Milverton (Cathy '
                                                 'Wettlaufer)',
                                         'poller_hash': 'MUQwNEY3',
                                         'serial_number': '1D04F7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:65',
                                         'uptime': '21816522'},
                                 {       'counter_source': '62/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '62/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '50932380779',
                                                                       'in_ucast_pkts': '471597244',
                                                                       'mac_address': '00:06:31:e8:f3:da',
                                                                       'name': '62/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '126',
                                                                       'out_octets': '1594737011383',
                                                                       'out_ucast_pkts': '1165086850',
                                                                       'poller_hash': 'MUQwNTA0cGVlcmludGY2Mi9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '62',
                                                                       'in_nucast_pkts': '126',
                                                                       'in_octets': '1594737011383',
                                                                       'in_ucast_pkts': '1165086850',
                                                                       'links': [       {       'in_nucast_pkts': '126',
                                                                                                'in_octets': '1594737011383',
                                                                                                'in_ucast_pkts': '1165086850',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '50932380779',
                                                                                                'out_ucast_pkts': '471597244',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '50932380779',
                                                                       'out_ucast_pkts': '471597244',
                                                                       'poller_hash': 'MUQwNTA0cGVlcmludGZwb24gNjI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '62',
                                                                                               'links': [       {       'connected_time': '73621552',
                                                                                                                        'fiber_length': 1458,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTA0cGVlcnBvbnBvbiA2Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '17 MacBeth St. (David Schultz)',
                                         'poller_hash': 'MUQwNTA0',
                                         'serial_number': '1D0504',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:da',
                                         'uptime': '73621552'},
                                 {       'counter_source': '63/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '63/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '883',
                                                                       'in_octets': '125548705447',
                                                                       'in_ucast_pkts': '732352462',
                                                                       'mac_address': '00:06:31:e8:f3:9b',
                                                                       'name': '63/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '4976',
                                                                       'out_octets': '2707598870234',
                                                                       'out_ucast_pkts': '1989665998',
                                                                       'poller_hash': 'MUQwNEZEcGVlcmludGY2My9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '63',
                                                                       'in_nucast_pkts': '4976',
                                                                       'in_octets': '2707598870234',
                                                                       'in_ucast_pkts': '1989665998',
                                                                       'links': [       {       'in_nucast_pkts': '4976',
                                                                                                'in_octets': '2707598870234',
                                                                                                'in_ucast_pkts': '1989665998',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '883',
                                                                                                'out_octets': '125548705447',
                                                                                                'out_ucast_pkts': '732352462',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '883',
                                                                       'out_octets': '125548705447',
                                                                       'out_ucast_pkts': '732352462',
                                                                       'poller_hash': 'MUQwNEZEcGVlcmludGZwb24gNjM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '63',
                                                                                               'links': [       {       'connected_time': '688474',
                                                                                                                        'fiber_length': 1042,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEZEcGVlcnBvbnBvbiA2Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '10 JOHN ST (GOWING COLE)',
                                         'poller_hash': 'MUQwNEZE',
                                         'serial_number': '1D04FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:9b',
                                         'uptime': '688474'},
                                 {       'counter_source': '64/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '64/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25613269436',
                                                                       'in_ucast_pkts': '98275361',
                                                                       'mac_address': '00:06:31:e8:e4:11',
                                                                       'name': '64/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '84',
                                                                       'out_octets': '675875791204',
                                                                       'out_ucast_pkts': '580925814',
                                                                       'poller_hash': 'MUQwMzQzcGVlcmludGY2NC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '64',
                                                                       'in_nucast_pkts': '84',
                                                                       'in_octets': '675875791204',
                                                                       'in_ucast_pkts': '580925814',
                                                                       'links': [       {       'in_nucast_pkts': '84',
                                                                                                'in_octets': '675875791204',
                                                                                                'in_ucast_pkts': '580925814',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '25613269436',
                                                                                                'out_ucast_pkts': '98275361',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25613269436',
                                                                       'out_ucast_pkts': '98275361',
                                                                       'poller_hash': 'MUQwMzQzcGVlcmludGZwb24gNjQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '64',
                                                                                               'links': [       {       'connected_time': '2126558',
                                                                                                                        'fiber_length': 1056,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwMzQzcGVlcnBvbnBvbiA2NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '31 William St (Ken Lehmann 10400650)',
                                         'poller_hash': 'MUQwMzQz',
                                         'serial_number': '1D0343',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:e4:11',
                                         'uptime': '2126558'},
                                 {       'counter_source': '65/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ef:7b'],
                                                                       'id': '65/g1',
                                                                       'in_nucast_pkts': '917679',
                                                                       'in_octets': '364800707911',
                                                                       'in_ucast_pkts': '2052409027',
                                                                       'mac_address': '00:06:31:e8:e3:f0',
                                                                       'mtu': 2000,
                                                                       'name': '65/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149410495',
                                                                       'out_octets': '8664543806430',
                                                                       'out_ucast_pkts': '6432738827',
                                                                       'poller_hash': 'MUQwMzNGcGVlcmludGY2NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '65',
                                                                       'in_nucast_pkts': '149410495',
                                                                       'in_octets': '8664543806430',
                                                                       'in_ucast_pkts': '6432738827',
                                                                       'links': [       {       'in_nucast_pkts': '149410495',
                                                                                                'in_octets': '8664543806430',
                                                                                                'in_ucast_pkts': '6432738827',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '917679',
                                                                                                'out_octets': '364800707911',
                                                                                                'out_ucast_pkts': '2052409027',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '917679',
                                                                       'out_octets': '364800707911',
                                                                       'out_ucast_pkts': '2052409027',
                                                                       'poller_hash': 'MUQwMzNGcGVlcmludGZwb24gNjU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '65',
                                                                                               'links': [       {       'connected_time': '21816621',
                                                                                                                        'fiber_length': 1304,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwMzNGcGVlcnBvbnBvbiA2NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '9 WILFRED ST (PURVES ALI JOY)',
                                         'poller_hash': 'MUQwMzNG',
                                         'serial_number': '1D033F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:e3:ed',
                                         'uptime': '21816621'},
                                 {       'counter_source': '66/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '66/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '452346816936',
                                                                       'in_ucast_pkts': '74539523191',
                                                                       'mac_address': '00:06:31:e8:f3:e3',
                                                                       'name': '66/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '4407432749238',
                                                                       'out_ucast_pkts': '3288162902',
                                                                       'poller_hash': 'MUQwNTA1cGVlcmludGY2Ni9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '66',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '4407432749238',
                                                                       'in_ucast_pkts': '3288162902',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '4407432749238',
                                                                                                'in_ucast_pkts': '3288162902',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '452346816936',
                                                                                                'out_ucast_pkts': '74539523191',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '452346816936',
                                                                       'out_ucast_pkts': '74539523191',
                                                                       'poller_hash': 'MUQwNTA1cGVlcmludGZwb24gNjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '66',
                                                                                               'links': [       {       'connected_time': '21541485',
                                                                                                                        'fiber_length': 1303,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTA1cGVlcnBvbnBvbiA2Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '7 Wilfred St. Milverton (Gwen Moffat '
                                                 '10431370)',
                                         'poller_hash': 'MUQwNTA1',
                                         'serial_number': '1D0505',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:e3',
                                         'uptime': '21541485'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '67/g1',
                                                                       'name': '67/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NzU3QURwZWVyaW50ZjY3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '67',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NzU3QURwZWVyaW50ZnBvbiA2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '67',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzU3QURwZWVycG9ucG9uIDY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '10A Temperance St. Milverton (Julie '
                                                 'Gallinger)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "67" (location: '
                                                                  'GPON port "1/5"), '
                                                                  '"ONT went missing"',
                                         'offline_time': '2025-03-19T18:04:48.970Z',
                                         'poller_hash': 'NzU3QUQ=',
                                         'serial_number': '757AD',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '68/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '68/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '421453026879',
                                                                       'in_ucast_pkts': '1500968058',
                                                                       'mac_address': '00:06:31:52:57:d6',
                                                                       'name': '68/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9935722168654',
                                                                       'out_ucast_pkts': '7265312128',
                                                                       'poller_hash': 'ODg5RTdwZWVyaW50ZjY4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '68',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9935722168654',
                                                                       'in_ucast_pkts': '7265312128',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9935722168654',
                                                                                                'in_ucast_pkts': '7265312128',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '421453026879',
                                                                                                'out_ucast_pkts': '1500968058',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '421453026879',
                                                                       'out_ucast_pkts': '1500968058',
                                                                       'poller_hash': 'ODg5RTdwZWVyaW50ZnBvbiA2OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '68',
                                                                                               'links': [       {       'connected_time': '688336',
                                                                                                                        'fiber_length': 888,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RTdwZWVycG9ucG9uIDY4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '10B TEMPERANCE STREET (PATERSON JOE '
                                                 '& SHARON)',
                                         'poller_hash': 'ODg5RTc=',
                                         'serial_number': '889E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:cf',
                                         'uptime': '688336'},
                                 {       'counter_source': '69/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '69/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '93578649754',
                                                                       'in_ucast_pkts': '1101373829',
                                                                       'mac_address': '00:06:31:e8:f4:10',
                                                                       'name': '69/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '55',
                                                                       'out_octets': '6794619215934',
                                                                       'out_ucast_pkts': '4550961024',
                                                                       'poller_hash': 'MUQwNTBBcGVlcmludGY2OS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '69',
                                                                       'in_nucast_pkts': '55',
                                                                       'in_octets': '6794619215934',
                                                                       'in_ucast_pkts': '4550961024',
                                                                       'links': [       {       'in_nucast_pkts': '55',
                                                                                                'in_octets': '6794619215934',
                                                                                                'in_ucast_pkts': '4550961024',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '93578649754',
                                                                                                'out_ucast_pkts': '1101373829',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '93578649754',
                                                                       'out_ucast_pkts': '1101373829',
                                                                       'poller_hash': 'MUQwNTBBcGVlcmludGZwb24gNjk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '69',
                                                                                               'links': [       {       'connected_time': '607321',
                                                                                                                        'fiber_length': 1056,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTBBcGVlcnBvbnBvbiA2OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '12 John St. Milverton (Jim Abraham)',
                                         'poller_hash': 'MUQwNTBB',
                                         'serial_number': '1D050A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f4:10',
                                         'uptime': '607321'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '70/g1',
                                                                       'name': '70/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQjRBcGVlcmludGY3MC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '70',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQjRBcGVlcmludGZwb24gNzA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '70',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjRBcGVlcnBvbnBvbiA3MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '6 Main St N',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "70" (location: '
                                                                  'GPON port "7/8"), '
                                                                  '"ONT went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:41.870Z',
                                         'poller_hash': 'NjQwQjRB',
                                         'serial_number': '640B4A',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '71/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '71/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '73943621278',
                                                                       'in_ucast_pkts': '603235882',
                                                                       'mac_address': '00:06:31:e9:01:90',
                                                                       'name': '71/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '4145346202955',
                                                                       'out_ucast_pkts': '2841598088',
                                                                       'poller_hash': 'MUQwNjhBcGVlcmludGY3MS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '71',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '4145346202955',
                                                                       'in_ucast_pkts': '2841598088',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '4145346202955',
                                                                                                'in_ucast_pkts': '2841598088',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '73943621278',
                                                                                                'out_ucast_pkts': '603235882',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '73943621278',
                                                                       'out_ucast_pkts': '603235882',
                                                                       'poller_hash': 'MUQwNjhBcGVlcmludGZwb24gNzE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '71',
                                                                                               'links': [       {       'connected_time': '30482265',
                                                                                                                        'fiber_length': 1370,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjhBcGVlcnBvbnBvbiA3MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '53 Temperance St. Milverton (Heather '
                                                 'Arand 10401553)',
                                         'poller_hash': 'MUQwNjhB',
                                         'serial_number': '1D068A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:90',
                                         'uptime': '30482265'},
                                 {       'counter_source': '72/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:a4:b7:dd:29:f3'],
                                                                       'id': '72/g1',
                                                                       'in_nucast_pkts': '2680',
                                                                       'in_octets': '7082820446',
                                                                       'in_ucast_pkts': '13812961',
                                                                       'mac_address': '00:06:31:e8:f3:71',
                                                                       'mtu': 2000,
                                                                       'name': '72/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464275',
                                                                       'out_octets': '39896573040',
                                                                       'out_ucast_pkts': '32419540',
                                                                       'poller_hash': 'MUQwNEY4cGVlcmludGY3Mi9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '72',
                                                                       'in_nucast_pkts': '1464275',
                                                                       'in_octets': '39896573040',
                                                                       'in_ucast_pkts': '32419540',
                                                                       'links': [       {       'in_nucast_pkts': '1464275',
                                                                                                'in_octets': '39896573040',
                                                                                                'in_ucast_pkts': '32419540',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2680',
                                                                                                'out_octets': '7082820446',
                                                                                                'out_ucast_pkts': '13812961',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2680',
                                                                       'out_octets': '7082820446',
                                                                       'out_ucast_pkts': '13812961',
                                                                       'poller_hash': 'MUQwNEY4cGVlcmludGZwb24gNzI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '72',
                                                                                               'links': [       {       'connected_time': '688482',
                                                                                                                        'fiber_length': 999,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEY4cGVlcnBvbnBvbiA3Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '25 WILLIAM ST (ELROY WAGLER MASONRY)',
                                         'poller_hash': 'MUQwNEY4',
                                         'serial_number': '1D04F8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:6e',
                                         'uptime': '688482'},
                                 {       'counter_source': '73/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '73/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '542369992698',
                                                                       'in_ucast_pkts': '106235913943',
                                                                       'mac_address': '00:06:31:e5:cc:50',
                                                                       'name': '73/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '10041968279789',
                                                                       'out_ucast_pkts': '7367503586',
                                                                       'poller_hash': 'MUNBQjRBcGVlcmludGY3My9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '73',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '10041968279789',
                                                                       'in_ucast_pkts': '7367503586',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '10041968279789',
                                                                                                'in_ucast_pkts': '7367503586',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '542369992698',
                                                                                                'out_ucast_pkts': '106235913943',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '542369992698',
                                                                       'out_ucast_pkts': '106235913943',
                                                                       'poller_hash': 'MUNBQjRBcGVlcmludGZwb24gNzM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '73',
                                                                                               'links': [       {       'connected_time': '11231777',
                                                                                                                        'fiber_length': 988,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUNBQjRBcGVlcnBvbnBvbiA3Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '7 TEMPERANCE ST (MCKONE TYLER R)',
                                         'poller_hash': 'MUNBQjRB',
                                         'serial_number': '1CAB4A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e5:cc:50',
                                         'uptime': '11231777'},
                                 {       'counter_source': '74/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:ba:5e'],
                                                                       'id': '74/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '1475408762',
                                                                       'in_ucast_pkts': '13443559',
                                                                       'mac_address': 'ec:4f:82:59:8b:8f',
                                                                       'mtu': 2000,
                                                                       'name': '74/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464199',
                                                                       'out_octets': '62259095924',
                                                                       'out_ucast_pkts': '43307763',
                                                                       'poller_hash': 'MkFDNEI2cGVlcmludGY3NC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '74',
                                                                       'in_nucast_pkts': '1464199',
                                                                       'in_octets': '62259095924',
                                                                       'in_ucast_pkts': '43307763',
                                                                       'links': [       {       'in_nucast_pkts': '1464199',
                                                                                                'in_octets': '62259095924',
                                                                                                'in_ucast_pkts': '43307763',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '1475408762',
                                                                                                'out_ucast_pkts': '13443559',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '1475408762',
                                                                       'out_ucast_pkts': '13443559',
                                                                       'poller_hash': 'MkFDNEI2cGVlcmludGZwb24gNzQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '74',
                                                                                               'links': [       {       'connected_time': '688348',
                                                                                                                        'fiber_length': 962,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFDNEI2cGVlcnBvbnBvbiA3NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '8 Temperance St. Milverton (Mike '
                                                 'Crapper 10425901)',
                                         'poller_hash': 'MkFDNEI2',
                                         'serial_number': '2AC4B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:8b:8f',
                                         'uptime': '688348'},
                                 {       'counter_source': '77/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a0:ab:1b:88:c0:af'],
                                                                       'id': '77/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '3426405944',
                                                                       'in_ucast_pkts': '18833161',
                                                                       'mac_address': 'cc:be:59:50:17:94',
                                                                       'mtu': 2000,
                                                                       'name': '77/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1103082',
                                                                       'out_octets': '70105152216',
                                                                       'out_ucast_pkts': '50674190',
                                                                       'poller_hash': 'NDNDMjRCcGVlcmludGY3Ny9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '77',
                                                                       'in_nucast_pkts': '1103082',
                                                                       'in_octets': '70105152216',
                                                                       'in_ucast_pkts': '50674190',
                                                                       'links': [       {       'in_nucast_pkts': '1103082',
                                                                                                'in_octets': '70105152216',
                                                                                                'in_ucast_pkts': '50674190',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '3426405944',
                                                                                                'out_ucast_pkts': '18833161',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '3426405944',
                                                                       'out_ucast_pkts': '18833161',
                                                                       'poller_hash': 'NDNDMjRCcGVlcmludGZwb24gNzc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '77',
                                                                                               'links': [       {       'connected_time': '688396',
                                                                                                                        'fiber_length': 2111,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.408,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.226}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNDMjRCcGVlcnBvbnBvbiA3Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '74 Gerber Drive Milverton (Shannon '
                                                 'Stangen 10434431)',
                                         'poller_hash': 'NDNDMjRC',
                                         'serial_number': '43C24B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:17:93',
                                         'uptime': '688396'},
                                 {       'counter_source': '78/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '50:d4:f7:80:62:e9'],
                                                                       'id': '78/g1',
                                                                       'in_nucast_pkts': '536',
                                                                       'in_octets': '6808607997',
                                                                       'in_ucast_pkts': '21328169',
                                                                       'mac_address': 'ec:4f:82:3b:95:c7',
                                                                       'mtu': 2000,
                                                                       'name': '78/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1438965',
                                                                       'out_octets': '69572157072',
                                                                       'out_ucast_pkts': '51768969',
                                                                       'poller_hash': 'MjhGNjY0cGVlcmludGY3OC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '78',
                                                                       'in_nucast_pkts': '1438965',
                                                                       'in_octets': '69572157072',
                                                                       'in_ucast_pkts': '51768969',
                                                                       'links': [       {       'in_nucast_pkts': '1438965',
                                                                                                'in_octets': '69572157072',
                                                                                                'in_ucast_pkts': '51768969',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '536',
                                                                                                'out_octets': '6808607997',
                                                                                                'out_ucast_pkts': '21328169',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '536',
                                                                       'out_octets': '6808607997',
                                                                       'out_ucast_pkts': '21328169',
                                                                       'poller_hash': 'MjhGNjY0cGVlcmludGZwb24gNzg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '78',
                                                                                               'links': [       {       'connected_time': '677139',
                                                                                                                        'fiber_length': 1312,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjhGNjY0cGVlcnBvbnBvbiA3OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '8 Edward St. Milverton (Katrina '
                                                 'MacPherson)',
                                         'poller_hash': 'MjhGNjY0',
                                         'serial_number': '28F664',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:3b:95:c7',
                                         'uptime': '677139'},
                                 {       'counter_source': '79/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '79/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '290',
                                                                       'in_octets': '1715503962572',
                                                                       'in_ucast_pkts': '2620122891',
                                                                       'mac_address': 'ec:4f:82:8d:82:10',
                                                                       'name': '79/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1388',
                                                                       'out_octets': '13413119535822',
                                                                       'out_ucast_pkts': '11994185083',
                                                                       'poller_hash': 'MzIxODg5cGVlcmludGY3OS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '79',
                                                                       'in_nucast_pkts': '1388',
                                                                       'in_octets': '13413119535822',
                                                                       'in_ucast_pkts': '11994185083',
                                                                       'links': [       {       'in_nucast_pkts': '1388',
                                                                                                'in_octets': '13413119535822',
                                                                                                'in_ucast_pkts': '11994185083',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '290',
                                                                                                'out_octets': '1715503962572',
                                                                                                'out_ucast_pkts': '2620122891',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '290',
                                                                       'out_octets': '1715503962572',
                                                                       'out_ucast_pkts': '2620122891',
                                                                       'poller_hash': 'MzIxODg5cGVlcmludGZwb24gNzk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '79',
                                                                                               'links': [       {       'connected_time': '22280705',
                                                                                                                        'fiber_length': 1045,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxODg5cGVlcnBvbnBvbiA3OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29A William St. (Fred Manser)',
                                         'poller_hash': 'MzIxODg5',
                                         'serial_number': '321889',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:82:10',
                                         'uptime': '22280705'},
                                 {       'counter_source': '80/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '80/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '279622632223',
                                                                       'in_ucast_pkts': '1031731518',
                                                                       'mac_address': '00:06:31:e8:f3:c8',
                                                                       'name': '80/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '5441823747175',
                                                                       'out_ucast_pkts': '3904607707',
                                                                       'poller_hash': 'MUQwNTAycGVlcmludGY4MC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '80',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '5441823747175',
                                                                       'in_ucast_pkts': '3904607707',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '5441823747175',
                                                                                                'in_ucast_pkts': '3904607707',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '279622632223',
                                                                                                'out_ucast_pkts': '1031731518',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '279622632223',
                                                                       'out_ucast_pkts': '1031731518',
                                                                       'poller_hash': 'MUQwNTAycGVlcmludGZwb24gODA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '80',
                                                                                               'links': [       {       'connected_time': '688349',
                                                                                                                        'fiber_length': 841,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTAycGVlcnBvbnBvbiA4MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '11 Pugh St Milverton (Greg Storey)',
                                         'poller_hash': 'MUQwNTAy',
                                         'serial_number': '1D0502',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:c8',
                                         'uptime': '688349'},
                                 {       'counter_source': '81/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:a9:8a:75:88:59'],
                                                                       'id': '81/g1',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '214801706814',
                                                                       'in_ucast_pkts': '384323594',
                                                                       'mac_address': '00:06:31:9c:b9:d3',
                                                                       'mtu': 2000,
                                                                       'name': '81/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18150621',
                                                                       'out_octets': '1085395387072',
                                                                       'out_ucast_pkts': '1002837189',
                                                                       'poller_hash': 'MTM4MTNCcGVlcmludGY4MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '81',
                                                                       'in_nucast_pkts': '18150621',
                                                                       'in_octets': '1085395387072',
                                                                       'in_ucast_pkts': '1002837189',
                                                                       'links': [       {       'in_nucast_pkts': '18150621',
                                                                                                'in_octets': '1085395387072',
                                                                                                'in_ucast_pkts': '1002837189',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12',
                                                                                                'out_octets': '214801706814',
                                                                                                'out_ucast_pkts': '384323594',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '214801706814',
                                                                       'out_ucast_pkts': '384323594',
                                                                       'poller_hash': 'MTM4MTNCcGVlcmludGZwb24gODE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '81',
                                                                                               'links': [       {       'connected_time': '10111014',
                                                                                                                        'fiber_length': 1649,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTM4MTNCcGVlcnBvbnBvbiA4MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '4705 Line 64 Milverton (Wagler '
                                                 'Electric 10400140)',
                                         'poller_hash': 'MTM4MTNC',
                                         'serial_number': '13813B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:9c:b9:d8',
                                         'uptime': '10111014'},
                                 {       'counter_source': '82/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '82/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1359891031505',
                                                                       'in_ucast_pkts': '2547837759',
                                                                       'mac_address': '00:06:31:87:40:f3',
                                                                       'name': '82/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2846958216140',
                                                                       'out_ucast_pkts': '3833086838',
                                                                       'poller_hash': 'Q0MwNzRwZWVyaW50ZjgyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '82',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2846958216140',
                                                                       'in_ucast_pkts': '3833086838',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2846958216140',
                                                                                                'in_ucast_pkts': '3833086838',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1359891031505',
                                                                                                'out_ucast_pkts': '2547837759',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1359891031505',
                                                                       'out_ucast_pkts': '2547837759',
                                                                       'poller_hash': 'Q0MwNzRwZWVyaW50ZnBvbiA4Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '82',
                                                                                               'links': [       {       'connected_time': '47499',
                                                                                                                        'fiber_length': 1228,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwNzRwZWVycG9ucG9uIDgy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4713 Line 64 (Scott Steckly '
                                                 '10421864)',
                                         'poller_hash': 'Q0MwNzQ=',
                                         'serial_number': 'CC074',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:34',
                                         'uptime': '47499'},
                                 {       'counter_source': '83/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '83/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '87877111137',
                                                                       'in_ucast_pkts': '335918136',
                                                                       'mac_address': 'ec:4f:82:4f:7a:8f',
                                                                       'name': '83/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '175',
                                                                       'out_octets': '1515979801657',
                                                                       'out_ucast_pkts': '1269845918',
                                                                       'poller_hash': 'Mjk4MjU0cGVlcmludGY4My9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '83',
                                                                       'in_nucast_pkts': '175',
                                                                       'in_octets': '1515979801657',
                                                                       'in_ucast_pkts': '1269845918',
                                                                       'links': [       {       'in_nucast_pkts': '175',
                                                                                                'in_octets': '1515979801657',
                                                                                                'in_ucast_pkts': '1269845918',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '87877111137',
                                                                                                'out_ucast_pkts': '335918136',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '87877111137',
                                                                       'out_ucast_pkts': '335918136',
                                                                       'poller_hash': 'Mjk4MjU0cGVlcmludGZwb24gODM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '83',
                                                                                               'links': [       {       'connected_time': '21817288',
                                                                                                                        'fiber_length': 1239,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjU0cGVlcnBvbnBvbiA4Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '38 William St. Milverton (James '
                                                 'Brubacher)',
                                         'poller_hash': 'Mjk4MjU0',
                                         'serial_number': '298254',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:8f',
                                         'uptime': '21817288'},
                                 {       'counter_source': '84/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '84/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '146305589772',
                                                                       'in_ucast_pkts': '90832460840',
                                                                       'mac_address': '00:06:31:e9:01:ea',
                                                                       'name': '84/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '115',
                                                                       'out_octets': '2371519702257',
                                                                       'out_ucast_pkts': '1847624428',
                                                                       'poller_hash': 'MUQwNjk0cGVlcmludGY4NC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '84',
                                                                       'in_nucast_pkts': '115',
                                                                       'in_octets': '2371519702257',
                                                                       'in_ucast_pkts': '1847624428',
                                                                       'links': [       {       'in_nucast_pkts': '115',
                                                                                                'in_octets': '2371519702257',
                                                                                                'in_ucast_pkts': '1847624428',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '146305589772',
                                                                                                'out_ucast_pkts': '90832460840',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '146305589772',
                                                                       'out_ucast_pkts': '90832460840',
                                                                       'poller_hash': 'MUQwNjk0cGVlcmludGZwb24gODQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '84',
                                                                                               'links': [       {       'connected_time': '21816848',
                                                                                                                        'fiber_length': 1035,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjk0cGVlcnBvbnBvbiA4NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '9 John St. Milverton (Shannon Cairns '
                                                 '10432740)',
                                         'poller_hash': 'MUQwNjk0',
                                         'serial_number': '1D0694',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:ea',
                                         'uptime': '21816848'},
                                 {       'counter_source': '86/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '86/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3047270292',
                                                                       'in_ucast_pkts': '15927235',
                                                                       'mac_address': 'ec:4f:82:a2:b7:d0',
                                                                       'name': '86/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '100003059487',
                                                                       'out_ucast_pkts': '83334930',
                                                                       'poller_hash': 'MzQ5RTdEcGVlcmludGY4Ni9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '86',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '100003059487',
                                                                       'in_ucast_pkts': '83334930',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '100003059487',
                                                                                                'in_ucast_pkts': '83334930',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3047270292',
                                                                                                'out_ucast_pkts': '15927235',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3047270292',
                                                                       'out_ucast_pkts': '15927235',
                                                                       'poller_hash': 'MzQ5RTdEcGVlcmludGZwb24gODY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '86',
                                                                                               'links': [       {       'connected_time': '688357',
                                                                                                                        'fiber_length': 1024,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ5RTdEcGVlcnBvbnBvbiA4Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '11 WEST ST (STEVENS NCHOLAS)',
                                         'poller_hash': 'MzQ5RTdE',
                                         'serial_number': '349E7D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:a2:b7:d0',
                                         'uptime': '688357'},
                                 {       'counter_source': '87/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:91:62:00:96:13'],
                                                                       'id': '87/g1',
                                                                       'in_nucast_pkts': '520',
                                                                       'in_octets': '185381546647',
                                                                       'in_ucast_pkts': '607138461',
                                                                       'mac_address': '00:06:31:dd:74:03',
                                                                       'mtu': 2000,
                                                                       'name': '87/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31561338',
                                                                       'out_octets': '1772386638724',
                                                                       'out_ucast_pkts': '1630841571',
                                                                       'poller_hash': 'MTc2MjE1cGVlcmludGY4Ny9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '87',
                                                                       'in_nucast_pkts': '31561338',
                                                                       'in_octets': '1772386638724',
                                                                       'in_ucast_pkts': '1630841571',
                                                                       'links': [       {       'in_nucast_pkts': '31561338',
                                                                                                'in_octets': '1772386638724',
                                                                                                'in_ucast_pkts': '1630841571',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '520',
                                                                                                'out_octets': '185381546647',
                                                                                                'out_ucast_pkts': '607138461',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '520',
                                                                       'out_octets': '185381546647',
                                                                       'out_ucast_pkts': '607138461',
                                                                       'poller_hash': 'MTc2MjE1cGVlcmludGZwb24gODc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '87',
                                                                                               'links': [       {       'connected_time': '18081316',
                                                                                                                        'fiber_length': 2575,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc2MjE1cGVlcnBvbnBvbiA4Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4700 Line 61 (Perth East Shop '
                                                 '10401421)',
                                         'poller_hash': 'MTc2MjE1',
                                         'serial_number': '176215',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:73:ff',
                                         'uptime': '18081316'},
                                 {       'counter_source': '88/g3+88/g4+88/g5+88/g6+88/g7',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:ba:ee'],
                                                                       'id': '88/g3',
                                                                       'in_nucast_pkts': '109',
                                                                       'in_octets': '199533977495',
                                                                       'in_ucast_pkts': '1173172789',
                                                                       'mac_address': '00:06:31:f6:55:bd',
                                                                       'mtu': 2000,
                                                                       'name': '88/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27112297',
                                                                       'out_octets': '5283093283054',
                                                                       'out_ucast_pkts': '3631770359',
                                                                       'poller_hash': 'MjYwRTk3cGVlcmludGY4OC9nMw==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:0a:48'],
                                                                       'id': '88/g4',
                                                                       'in_nucast_pkts': '858116',
                                                                       'in_octets': '131596720436',
                                                                       'in_ucast_pkts': '1752058645',
                                                                       'mac_address': '00:06:31:f6:55:be',
                                                                       'mtu': 2000,
                                                                       'name': '88/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '115160971',
                                                                       'out_octets': '9680438243495',
                                                                       'out_ucast_pkts': '6856118383',
                                                                       'poller_hash': 'MjYwRTk3cGVlcmludGY4OC9nNA==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:47:41'],
                                                                       'id': '88/g5',
                                                                       'in_nucast_pkts': '106',
                                                                       'in_octets': '104825621134',
                                                                       'in_ucast_pkts': '798869964',
                                                                       'mac_address': 'ec:4f:82:27:32:a3',
                                                                       'mtu': 2000,
                                                                       'name': '88/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27112144',
                                                                       'out_octets': '4992427166998',
                                                                       'out_ucast_pkts': '3435241648',
                                                                       'poller_hash': 'MjYwRTk3cGVlcmludGY4OC9nNQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '58:d9:d5:1a:75:99'],
                                                                       'id': '88/g6',
                                                                       'in_nucast_pkts': '920',
                                                                       'in_octets': '198291752134',
                                                                       'in_ucast_pkts': '1136519167',
                                                                       'mac_address': 'ec:4f:82:27:32:a4',
                                                                       'mtu': 2000,
                                                                       'name': '88/g6',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27111703',
                                                                       'out_octets': '9702193192044',
                                                                       'out_ucast_pkts': '6608656683',
                                                                       'poller_hash': 'MjYwRTk3cGVlcmludGY4OC9nNg==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0a:44'],
                                                                       'id': '88/g7',
                                                                       'in_nucast_pkts': '179',
                                                                       'in_octets': '229089009639',
                                                                       'in_ucast_pkts': '1997434532',
                                                                       'mac_address': 'ec:4f:82:27:32:a5',
                                                                       'mtu': 2000,
                                                                       'name': '88/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27113959',
                                                                       'out_octets': '6482111512706',
                                                                       'out_ucast_pkts': '4475354544',
                                                                       'poller_hash': 'MjYwRTk3cGVlcmludGY4OC9nNw==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '88',
                                                                       'in_nucast_pkts': '223611074',
                                                                       'in_octets': '36140263398297',
                                                                       'in_ucast_pkts': '25007141617',
                                                                       'links': [       {       'in_nucast_pkts': '223611074',
                                                                                                'in_octets': '36140263398297',
                                                                                                'in_ucast_pkts': '25007141617',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '859430',
                                                                                                'out_octets': '863337080838',
                                                                                                'out_ucast_pkts': '6858055097',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '859430',
                                                                       'out_octets': '863337080838',
                                                                       'out_ucast_pkts': '6858055097',
                                                                       'poller_hash': 'MjYwRTk3cGVlcmludGZwb24gODg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '88',
                                                                                               'links': [       {       'connected_time': '15469743',
                                                                                                                        'fiber_length': 1368,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYwRTk3cGVlcnBvbnBvbiA4OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '14 MILL ST W',
                                         'poller_hash': 'MjYwRTk3',
                                         'serial_number': '260E97',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:f6:55:c0',
                                         'uptime': '15469743'},
                                 {       'counter_source': '89/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd4:76:a0:a3:53:9a'],
                                                                       'id': '89/g8',
                                                                       'in_nucast_pkts': '18301',
                                                                       'in_octets': '272132073649',
                                                                       'in_ucast_pkts': '424512067',
                                                                       'mac_address': 'ec:4f:82:25:3c:49',
                                                                       'mtu': 2000,
                                                                       'name': '89/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23369984',
                                                                       'out_octets': '71580918886',
                                                                       'out_ucast_pkts': '371286470',
                                                                       'poller_hash': 'MjYwRTZBcGVlcmludGY4OS9nOA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '89',
                                                                       'in_nucast_pkts': '23369984',
                                                                       'in_octets': '71580918886',
                                                                       'in_ucast_pkts': '371286470',
                                                                       'links': [       {       'in_nucast_pkts': '23369984',
                                                                                                'in_octets': '71580918886',
                                                                                                'in_ucast_pkts': '371286470',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18301',
                                                                                                'out_octets': '272132073649',
                                                                                                'out_ucast_pkts': '424512067',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18301',
                                                                       'out_octets': '272132073649',
                                                                       'out_ucast_pkts': '424512067',
                                                                       'poller_hash': 'MjYwRTZBcGVlcmludGZwb24gODk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '89',
                                                                                               'links': [       {       'connected_time': '13218993',
                                                                                                                        'fiber_length': 1369,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYwRTZBcGVlcnBvbnBvbiA4OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '14 MILL ST W (VERELLEN GREG R)',
                                         'poller_hash': 'MjYwRTZB',
                                         'serial_number': '260E6A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:f6:51:20',
                                         'uptime': '13218993'},
                                 {       'counter_source': '90/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:0b:d2'],
                                                                       'id': '90/g1',
                                                                       'in_nucast_pkts': '2879052',
                                                                       'in_octets': '126270120057',
                                                                       'in_ucast_pkts': '1609187359',
                                                                       'mac_address': '00:06:31:52:58:3a',
                                                                       'mtu': 2000,
                                                                       'name': '90/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '198450959',
                                                                       'out_octets': '8960602386217',
                                                                       'out_ucast_pkts': '6326174385',
                                                                       'poller_hash': 'ODhBMDBwZWVyaW50ZjkwL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '90',
                                                                       'in_nucast_pkts': '198450959',
                                                                       'in_octets': '8960602386217',
                                                                       'in_ucast_pkts': '6326174385',
                                                                       'links': [       {       'in_nucast_pkts': '198450959',
                                                                                                'in_octets': '8960602386217',
                                                                                                'in_ucast_pkts': '6326174385',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2879052',
                                                                                                'out_octets': '126270120057',
                                                                                                'out_ucast_pkts': '1609187359',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2879052',
                                                                       'out_octets': '126270120057',
                                                                       'out_ucast_pkts': '1609187359',
                                                                       'poller_hash': 'ODhBMDBwZWVyaW50ZnBvbiA5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '90',
                                                                                               'links': [       {       'connected_time': '29485979',
                                                                                                                        'fiber_length': 977,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODhBMDBwZWVycG9ucG9uIDkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '33 Mill st. E. Milverton (Frank '
                                                 'Huras)',
                                         'poller_hash': 'ODhBMDA=',
                                         'serial_number': '88A00',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:e8',
                                         'uptime': '29485979'},
                                 {       'counter_source': '92/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '92/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '659953516948',
                                                                       'in_ucast_pkts': '5243152427',
                                                                       'mac_address': '00:06:31:52:65:a6',
                                                                       'name': '92/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15742330115948',
                                                                       'out_ucast_pkts': '11238861341',
                                                                       'poller_hash': 'ODhBMDVwZWVyaW50ZjkyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '92',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15742330115948',
                                                                       'in_ucast_pkts': '11238861341',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '15742330115948',
                                                                                                'in_ucast_pkts': '11238861341',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '659953516948',
                                                                                                'out_ucast_pkts': '5243152427',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '659953516948',
                                                                       'out_ucast_pkts': '5243152427',
                                                                       'poller_hash': 'ODhBMDVwZWVyaW50ZnBvbiA5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '92',
                                                                                               'links': [       {       'connected_time': '22050262',
                                                                                                                        'fiber_length': 751,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODhBMDVwZWVycG9ucG9uIDky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '49 Mill St. E. Milverton (Bill Dale)',
                                         'poller_hash': 'ODhBMDU=',
                                         'serial_number': '88A05',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:ed',
                                         'uptime': '22050262'},
                                 {       'counter_source': '93/F1+93/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a1:c2:26'],
                                                                       'id': '93/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'name': '93/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODg5NzFwZWVyaW50ZjkzL0Yx'},
                                                               {       'admin_up': True,
                                                                       'id': '93/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '547796548079',
                                                                       'in_ucast_pkts': '2203764737',
                                                                       'mac_address': '00:06:31:52:67:06',
                                                                       'name': '93/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6655217345448',
                                                                       'out_ucast_pkts': '5540578763',
                                                                       'poller_hash': 'ODg5NzFwZWVyaW50ZjkzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '93',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6655217345448',
                                                                       'in_ucast_pkts': '5540578763',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6655217345448',
                                                                                                'in_ucast_pkts': '5540578763',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '547796548079',
                                                                                                'out_ucast_pkts': '2203764737',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '547796548079',
                                                                       'out_ucast_pkts': '2203764737',
                                                                       'poller_hash': 'ODg5NzFwZWVyaW50ZnBvbiA5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '93',
                                                                                               'links': [       {       'connected_time': '688424',
                                                                                                                        'fiber_length': 785,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5NzFwZWVycG9ucG9uIDkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '43 Mill St. E. Milverton (Micheal '
                                                 'Luckhardt 10423461)',
                                         'poller_hash': 'ODg5NzE=',
                                         'serial_number': '88971',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:59',
                                         'uptime': '688424'},
                                 {       'counter_source': '94/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c8:4f:86:a5:0f:14'],
                                                                       'id': '94/g3',
                                                                       'in_nucast_pkts': '6393994',
                                                                       'in_octets': '53184044507642',
                                                                       'in_ucast_pkts': '107950796789',
                                                                       'mac_address': '00:06:31:49:4d:e3',
                                                                       'mtu': 2000,
                                                                       'name': '94/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '365350679',
                                                                       'out_octets': '174768465404854',
                                                                       'out_ucast_pkts': '148815009960',
                                                                       'poller_hash': 'OUNBMERwZWVyaW50Zjk0L2cz',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '94',
                                                                       'in_nucast_pkts': '365350679',
                                                                       'in_octets': '174768465404854',
                                                                       'in_ucast_pkts': '148815009960',
                                                                       'links': [       {       'in_nucast_pkts': '365350679',
                                                                                                'in_octets': '174768465404854',
                                                                                                'in_ucast_pkts': '148815009960',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6393994',
                                                                                                'out_octets': '53184044507642',
                                                                                                'out_ucast_pkts': '107950796789',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6393994',
                                                                       'out_octets': '53184044507642',
                                                                       'out_ucast_pkts': '107950796789',
                                                                       'poller_hash': 'OUNBMERwZWVyaW50ZnBvbiA5NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '94',
                                                                                               'links': [       {       'connected_time': '5988529',
                                                                                                                        'fiber_length': 1327,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OUNBMERwZWVycG9ucG9uIDk0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '25 Mill St. E. Milverton (Perth East '
                                                 'Office 10411713)',
                                         'poller_hash': 'OUNBMEQ=',
                                         'serial_number': '9CA0D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:49:4d:e6',
                                         'uptime': '5988529'},
                                 {       'counter_source': '95/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:56:1f'],
                                                                       'id': '95/g1',
                                                                       'in_nucast_pkts': '72',
                                                                       'in_octets': '619063529632',
                                                                       'in_ucast_pkts': '709474160',
                                                                       'mac_address': '00:06:31:52:5a:72',
                                                                       'mtu': 2000,
                                                                       'name': '95/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27327588',
                                                                       'out_octets': '1507775562569',
                                                                       'out_ucast_pkts': '1313229537',
                                                                       'poller_hash': 'ODg5RDFwZWVyaW50Zjk1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '95',
                                                                       'in_nucast_pkts': '27327588',
                                                                       'in_octets': '1507775562569',
                                                                       'in_ucast_pkts': '1313229537',
                                                                       'links': [       {       'in_nucast_pkts': '27327588',
                                                                                                'in_octets': '1507775562569',
                                                                                                'in_ucast_pkts': '1313229537',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '72',
                                                                                                'out_octets': '619063529632',
                                                                                                'out_ucast_pkts': '709474160',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '72',
                                                                       'out_octets': '619063529632',
                                                                       'out_ucast_pkts': '709474160',
                                                                       'poller_hash': 'ODg5RDFwZWVyaW50ZnBvbiA5NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '95',
                                                                                               'links': [       {       'connected_time': '15498546',
                                                                                                                        'fiber_length': 1028,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RDFwZWVycG9ucG9uIDk1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '37 MILL ST E (HORTON LEXY)',
                                         'poller_hash': 'ODg5RDE=',
                                         'serial_number': '889D1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:b9',
                                         'uptime': '15498546'},
                                 {       'counter_source': '96/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '96/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '190803940818',
                                                                       'in_ucast_pkts': '943963563',
                                                                       'mac_address': '00:06:31:52:55:ee',
                                                                       'name': '96/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3871253027287',
                                                                       'out_ucast_pkts': '2939674429',
                                                                       'poller_hash': 'ODhBMUZwZWVyaW50Zjk2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '96',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3871253027287',
                                                                       'in_ucast_pkts': '2939674429',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3871253027287',
                                                                                                'in_ucast_pkts': '2939674429',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '190803940818',
                                                                                                'out_ucast_pkts': '943963563',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '190803940818',
                                                                       'out_ucast_pkts': '943963563',
                                                                       'poller_hash': 'ODhBMUZwZWVyaW50ZnBvbiA5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '96',
                                                                                               'links': [       {       'connected_time': '47526',
                                                                                                                        'fiber_length': 815,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODhBMUZwZWVycG9ucG9uIDk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '53 Mill St. E. Milverton (Clark '
                                                 'Moore 10401228)',
                                         'poller_hash': 'ODhBMUY=',
                                         'serial_number': '88A1F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:eb:07',
                                         'uptime': '47526'},
                                 {       'counter_source': '97/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '97/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '928196148639',
                                                                       'in_ucast_pkts': '2003126254',
                                                                       'mac_address': '00:06:31:52:65:ae',
                                                                       'name': '97/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6666048656853',
                                                                       'out_ucast_pkts': '5184696467',
                                                                       'poller_hash': 'ODg5MjlwZWVyaW50Zjk3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '97',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6666048656853',
                                                                       'in_ucast_pkts': '5184696467',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6666048656853',
                                                                                                'in_ucast_pkts': '5184696467',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '928196148639',
                                                                                                'out_ucast_pkts': '2003126254',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '928196148639',
                                                                       'out_ucast_pkts': '2003126254',
                                                                       'poller_hash': 'ODg5MjlwZWVyaW50ZnBvbiA5Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '97',
                                                                                               'links': [       {       'connected_time': '47388',
                                                                                                                        'fiber_length': 953,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5MjlwZWVycG9ucG9uIDk3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '31 Mill St. E. Milverton (Dave '
                                                 'Brand)',
                                         'poller_hash': 'ODg5Mjk=',
                                         'serial_number': '88929',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:11',
                                         'uptime': '47388'},
                                 {       'counter_source': '98/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:40:7f:96:4d:31',
                                                                                                        '1e:2d:5d:56:58:9f',
                                                                                                        '24:9a:d8:9b:4c:51',
                                                                                                        '26:fe:73:5a:c2:78',
                                                                                                        '28:d0:ea:d1:e1:10',
                                                                                                        '3e:9a:82:71:b8:5d',
                                                                                                        '44:db:d2:28:4a:7f',
                                                                                                        '44:db:d2:4a:59:ca',
                                                                                                        '4e:a7:31:1b:7e:2f',
                                                                                                        '52:00:6c:aa:87:59',
                                                                                                        '52:eb:16:84:bb:f0',
                                                                                                        '58:1c:f8:6d:29:9b',
                                                                                                        '5e:6f:dc:0f:75:4f',
                                                                                                        '60:22:32:45:a4:38',
                                                                                                        '64:00:6a:72:72:d6',
                                                                                                        '74:86:e2:07:24:0c',
                                                                                                        '74:86:e2:07:a1:9f',
                                                                                                        '78:45:58:43:c3:d9',
                                                                                                        'a8:3c:a5:20:a7:b5',
                                                                                                        'c2:88:9e:d8:3f:a9',
                                                                                                        'ca:51:32:b1:ca:30',
                                                                                                        'ca:d9:b4:e9:b5:b2',
                                                                                                        'cc:96:e5:20:76:47',
                                                                                                        'cc:96:e5:20:77:d9',
                                                                                                        'd0:8e:79:13:6d:c8',
                                                                                                        'd0:8e:79:13:6d:d7',
                                                                                                        'd4:f0:c9:0d:51:50'],
                                                                       'id': '98/g1',
                                                                       'in_nucast_pkts': '45698122',
                                                                       'in_octets': '693591117403',
                                                                       'in_ucast_pkts': '2418545542',
                                                                       'mac_address': '00:06:31:5e:d0:61',
                                                                       'mtu': 2000,
                                                                       'name': '98/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '383156522',
                                                                       'out_octets': '8352076267073',
                                                                       'out_ucast_pkts': '6516909192',
                                                                       'poller_hash': 'QUM0NEVwZWVyaW50Zjk4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '98',
                                                                       'in_nucast_pkts': '383156522',
                                                                       'in_octets': '8352076267073',
                                                                       'in_ucast_pkts': '6516909192',
                                                                       'links': [       {       'in_nucast_pkts': '383156522',
                                                                                                'in_octets': '8352076267073',
                                                                                                'in_ucast_pkts': '6516909192',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '45698122',
                                                                                                'out_octets': '693591117403',
                                                                                                'out_ucast_pkts': '2418545542',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45698122',
                                                                       'out_octets': '693591117403',
                                                                       'out_ucast_pkts': '2418545542',
                                                                       'poller_hash': 'QUM0NEVwZWVyaW50ZnBvbiA5OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '98',
                                                                                               'links': [       {       'connected_time': '72961103',
                                                                                                                        'fiber_length': 897,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QUM0NEVwZWVycG9ucG9uIDk4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '19 Mill St. E. Milverton (Perth East '
                                                 'Library)',
                                         'poller_hash': 'QUM0NEU=',
                                         'serial_number': 'AC44E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:57:d5:0e',
                                         'uptime': '72961103'},
                                 {       'counter_source': '99/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '99/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '337034127590',
                                                                       'in_ucast_pkts': '2115836914',
                                                                       'mac_address': '00:06:31:87:3e:db',
                                                                       'name': '99/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9816409719645',
                                                                       'out_ucast_pkts': '7040273861',
                                                                       'poller_hash': 'Q0MwOTVwZWVyaW50Zjk5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '99',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9816409719645',
                                                                       'in_ucast_pkts': '7040273861',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9816409719645',
                                                                                                'in_ucast_pkts': '7040273861',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '337034127590',
                                                                                                'out_ucast_pkts': '2115836914',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '337034127590',
                                                                       'out_ucast_pkts': '2115836914',
                                                                       'poller_hash': 'Q0MwOTVwZWVyaW50ZnBvbiA5OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '99',
                                                                                               'links': [       {       'connected_time': '1025287',
                                                                                                                        'fiber_length': 825,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwOTVwZWVycG9ucG9uIDk5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '41 Mill St. E. Milverton (Rob '
                                                 'Woodrow)',
                                         'poller_hash': 'Q0MwOTU=',
                                         'serial_number': 'CC095',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:55',
                                         'uptime': '1025287'},
                                 {       'counter_source': '100/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '54:af:97:39:64:4d'],
                                                                       'id': '100/g1',
                                                                       'in_nucast_pkts': '329',
                                                                       'in_octets': '9585366876',
                                                                       'in_ucast_pkts': '25926697',
                                                                       'mac_address': '84:d3:43:95:93:f5',
                                                                       'mtu': 2000,
                                                                       'name': '100/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463801',
                                                                       'out_octets': '108359475746',
                                                                       'out_ucast_pkts': '81922903',
                                                                       'poller_hash': 'Q0JEMjlEcGVlcmludGYxMDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '100',
                                                                       'in_nucast_pkts': '1463801',
                                                                       'in_octets': '108359475746',
                                                                       'in_ucast_pkts': '81922903',
                                                                       'links': [       {       'in_nucast_pkts': '1463801',
                                                                                                'in_octets': '108359475746',
                                                                                                'in_ucast_pkts': '81922903',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '329',
                                                                                                'out_octets': '9585366876',
                                                                                                'out_ucast_pkts': '25926697',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '329',
                                                                       'out_octets': '9585366876',
                                                                       'out_ucast_pkts': '25926697',
                                                                       'poller_hash': 'Q0JEMjlEcGVlcmludGZwb24gMTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '100',
                                                                                               'links': [       {       'connected_time': '688444',
                                                                                                                        'fiber_length': 1184,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjlEcGVlcnBvbnBvbiAxMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '34 Mill St. E. Milverton (Dale '
                                                 'Brubacher 10431281)',
                                         'poller_hash': 'Q0JEMjlE',
                                         'serial_number': 'CBD29D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:93:f2',
                                         'uptime': '688444'},
                                 {       'counter_source': '101/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:82:0a:f8'],
                                                                       'id': '101/g1',
                                                                       'in_nucast_pkts': '14583',
                                                                       'in_octets': '2894310630',
                                                                       'in_ucast_pkts': '18263789',
                                                                       'mac_address': '00:06:31:a6:67:5f',
                                                                       'mtu': 2000,
                                                                       'name': '101/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119648',
                                                                       'out_octets': '69361737664',
                                                                       'out_ucast_pkts': '55198079',
                                                                       'poller_hash': 'MTMwMEM1cGVlcmludGYxMDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '101',
                                                                       'in_nucast_pkts': '1119648',
                                                                       'in_octets': '69361737664',
                                                                       'in_ucast_pkts': '55198079',
                                                                       'links': [       {       'in_nucast_pkts': '1119648',
                                                                                                'in_octets': '69361737664',
                                                                                                'in_ucast_pkts': '55198079',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14583',
                                                                                                'out_octets': '2894310630',
                                                                                                'out_ucast_pkts': '18263789',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14583',
                                                                       'out_octets': '2894310630',
                                                                       'out_ucast_pkts': '18263789',
                                                                       'poller_hash': 'MTMwMEM1cGVlcmludGZwb24gMTAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '101',
                                                                                               'links': [       {       'connected_time': '688438',
                                                                                                                        'fiber_length': 1161,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMEM1cGVlcnBvbnBvbiAxMDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '40 MILL ST E (WARING ALICIA)',
                                         'poller_hash': 'MTMwMEM1',
                                         'serial_number': '1300C5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:85',
                                         'uptime': '688438'},
                                 {       'counter_source': '102/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:e3:ec'],
                                                                       'id': '102/g1',
                                                                       'in_nucast_pkts': '14897',
                                                                       'in_octets': '3135998675',
                                                                       'in_ucast_pkts': '11577091',
                                                                       'mac_address': '00:06:31:a6:6b:6f',
                                                                       'mtu': 2000,
                                                                       'name': '102/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119639',
                                                                       'out_octets': '46148367339',
                                                                       'out_ucast_pkts': '37031128',
                                                                       'poller_hash': 'MTMwMDA1cGVlcmludGYxMDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '102',
                                                                       'in_nucast_pkts': '1119639',
                                                                       'in_octets': '46148367339',
                                                                       'in_ucast_pkts': '37031128',
                                                                       'links': [       {       'in_nucast_pkts': '1119639',
                                                                                                'in_octets': '46148367339',
                                                                                                'in_ucast_pkts': '37031128',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14897',
                                                                                                'out_octets': '3135998675',
                                                                                                'out_ucast_pkts': '11577091',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14897',
                                                                       'out_octets': '3135998675',
                                                                       'out_ucast_pkts': '11577091',
                                                                       'poller_hash': 'MTMwMDA1cGVlcmludGZwb24gMTAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '102',
                                                                                               'links': [       {       'connected_time': '688438',
                                                                                                                        'fiber_length': 1075,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDA1cGVlcnBvbnBvbiAxMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '30 Mill St. E. Milverton (HABEL TOM)',
                                         'poller_hash': 'MTMwMDA1',
                                         'serial_number': '130005',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:c5',
                                         'uptime': '688438'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '103/g1',
                                                                       'name': '103/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTMwMDlCcGVlcmludGYxMDMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '103',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTMwMDlCcGVlcmludGZwb24gMTAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '103',
                                                                                               'links': [       {       'connected_time': '8385541',
                                                                                                                        'fiber_length': 1065,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDlCcGVlcnBvbnBvbiAxMDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '28 MILL ST E (JANTZI PAUL)',
                                         'poller_hash': 'MTMwMDlC',
                                         'serial_number': '13009B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:5b',
                                         'uptime': '8385541'},
                                 {       'counter_source': '104/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '104/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '264265169654',
                                                                       'in_ucast_pkts': '1434016587',
                                                                       'mac_address': '00:06:31:a6:69:83',
                                                                       'name': '104/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8327046667144',
                                                                       'out_ucast_pkts': '6004100854',
                                                                       'poller_hash': 'MTJGRkZBcGVlcmludGYxMDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '104',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8327046667144',
                                                                       'in_ucast_pkts': '6004100854',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '8327046667144',
                                                                                                'in_ucast_pkts': '6004100854',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '264265169654',
                                                                                                'out_ucast_pkts': '1434016587',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '264265169654',
                                                                       'out_ucast_pkts': '1434016587',
                                                                       'poller_hash': 'MTJGRkZBcGVlcmludGZwb24gMTA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '104',
                                                                                               'links': [       {       'connected_time': '47551',
                                                                                                                        'fiber_length': 1241,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTJGRkZBcGVlcnBvbnBvbiAxMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '18 Mill St. E. Milverton (Curvin '
                                                 'Martin)',
                                         'poller_hash': 'MTJGRkZB',
                                         'serial_number': '12FFFA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:ba',
                                         'uptime': '47551'},
                                 {       'counter_source': '105/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:ee'],
                                                                       'id': '105/g1',
                                                                       'in_nucast_pkts': '5714666',
                                                                       'in_octets': '1720670962447',
                                                                       'in_ucast_pkts': '8300401002',
                                                                       'mac_address': 'ec:4f:82:4f:78:ea',
                                                                       'mtu': 2000,
                                                                       'name': '105/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149404501',
                                                                       'out_octets': '20375364295764',
                                                                       'out_ucast_pkts': '16539027239',
                                                                       'poller_hash': 'Mjk4MjFGcGVlcmludGYxMDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '105',
                                                                       'in_nucast_pkts': '149404501',
                                                                       'in_octets': '20375364295764',
                                                                       'in_ucast_pkts': '16539027239',
                                                                       'links': [       {       'in_nucast_pkts': '149404501',
                                                                                                'in_octets': '20375364295764',
                                                                                                'in_ucast_pkts': '16539027239',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5714666',
                                                                                                'out_octets': '1720670962447',
                                                                                                'out_ucast_pkts': '8300401002',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5714666',
                                                                       'out_octets': '1720670962447',
                                                                       'out_ucast_pkts': '8300401002',
                                                                       'poller_hash': 'Mjk4MjFGcGVlcmludGZwb24gMTA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '105',
                                                                                               'links': [       {       'connected_time': '21815281',
                                                                                                                        'fiber_length': 1166,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjFGcGVlcnBvbnBvbiAxMDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 WEST ST (MURRAY JENNIFER)',
                                         'poller_hash': 'Mjk4MjFG',
                                         'serial_number': '29821F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:78:e7',
                                         'uptime': '21815281'},
                                 {       'counter_source': '106/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2c:75:23'],
                                                                       'id': '106/g1',
                                                                       'in_nucast_pkts': '300',
                                                                       'in_octets': '552884068090',
                                                                       'in_ucast_pkts': '4654033814',
                                                                       'mac_address': '00:06:31:b6:72:a8',
                                                                       'mtu': 2000,
                                                                       'name': '106/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53874366',
                                                                       'out_octets': '14724436843209',
                                                                       'out_ucast_pkts': '10196801527',
                                                                       'poller_hash': 'MTYwRDE1cGVlcmludGYxMDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '106',
                                                                       'in_nucast_pkts': '53874366',
                                                                       'in_octets': '14724436843209',
                                                                       'in_ucast_pkts': '10196801527',
                                                                       'links': [       {       'in_nucast_pkts': '53874366',
                                                                                                'in_octets': '14724436843209',
                                                                                                'in_ucast_pkts': '10196801527',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '300',
                                                                                                'out_octets': '552884068090',
                                                                                                'out_ucast_pkts': '4654033814',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '300',
                                                                       'out_octets': '552884068090',
                                                                       'out_ucast_pkts': '4654033814',
                                                                       'poller_hash': 'MTYwRDE1cGVlcmludGZwb24gMTA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '106',
                                                                                               'links': [       {       'connected_time': '29986028',
                                                                                                                        'fiber_length': 801,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.16,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTYwRDE1cGVlcnBvbnBvbiAxMDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '22 Mill St. W. Milverton (County Of '
                                                 'Perth EMS Base 10410443)',
                                         'poller_hash': 'MTYwRDE1',
                                         'serial_number': '160D15',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:b7:26:dc',
                                         'uptime': '29986028'},
                                 {       'counter_source': '108/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a4:92:12'],
                                                                       'id': '108/g1',
                                                                       'in_nucast_pkts': '2746474',
                                                                       'in_octets': '63104788927',
                                                                       'in_ucast_pkts': '727686874',
                                                                       'mac_address': '48:77:46:f8:79:df',
                                                                       'mtu': 2000,
                                                                       'name': '108/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7541501198',
                                                                       'out_octets': '14189473223502',
                                                                       'out_ucast_pkts': '2782649283',
                                                                       'poller_hash': 'ODJBNkNCcGVlcmludGYxMDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '108',
                                                                       'in_nucast_pkts': '7541501198',
                                                                       'in_octets': '14189473223502',
                                                                       'in_ucast_pkts': '2782649283',
                                                                       'links': [       {       'in_nucast_pkts': '7541501198',
                                                                                                'in_octets': '14189473223502',
                                                                                                'in_ucast_pkts': '2782649283',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2746474',
                                                                                                'out_octets': '63104788927',
                                                                                                'out_ucast_pkts': '727686874',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2746474',
                                                                       'out_octets': '63104788927',
                                                                       'out_ucast_pkts': '727686874',
                                                                       'poller_hash': 'ODJBNkNCcGVlcmludGZwb24gMTA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '108',
                                                                                               'links': [       {       'connected_time': '67790817',
                                                                                                                        'fiber_length': 18242,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNkNCcGVlcnBvbnBvbiAxMDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '23 WINSTANLEY ST (HEIBEIN SHERRY)',
                                         'poller_hash': 'ODJBNkNC',
                                         'serial_number': '82A6CB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:79:df',
                                         'uptime': '67790817'},
                                 {       'counter_source': '109/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '109/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '78129381355',
                                                                       'in_ucast_pkts': '599263128',
                                                                       'mac_address': 'ec:4f:82:4e:e6:27',
                                                                       'name': '109/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '2311647974021',
                                                                       'out_ucast_pkts': '1683981456',
                                                                       'poller_hash': 'Mjk2RkM3cGVlcmludGYxMDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '109',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '2311647974021',
                                                                       'in_ucast_pkts': '1683981456',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '2311647974021',
                                                                                                'in_ucast_pkts': '1683981456',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '78129381355',
                                                                                                'out_ucast_pkts': '599263128',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '78129381355',
                                                                       'out_ucast_pkts': '599263128',
                                                                       'poller_hash': 'Mjk2RkM3cGVlcmludGZwb24gMTA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '109',
                                                                                               'links': [       {       'connected_time': '15498467',
                                                                                                                        'fiber_length': 994,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkM3cGVlcnBvbnBvbiAxMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '18 TEMPERANCE ST (DALE BRIAN)',
                                         'poller_hash': 'Mjk2RkM3',
                                         'serial_number': '296FC7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4e:e6:27',
                                         'uptime': '15498467'},
                                 {       'counter_source': '110/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '40:5d:82:e1:9b:c9'],
                                                                       'id': '110/g1',
                                                                       'in_nucast_pkts': '4747236',
                                                                       'in_octets': '4409333799',
                                                                       'in_ucast_pkts': '27478767',
                                                                       'mac_address': 'ec:4f:82:3b:95:a5',
                                                                       'mtu': 2000,
                                                                       'name': '110/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464037',
                                                                       'out_octets': '94843238260',
                                                                       'out_ucast_pkts': '66732287',
                                                                       'poller_hash': 'MjhGNjUzcGVlcmludGYxMTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '110',
                                                                       'in_nucast_pkts': '1464037',
                                                                       'in_octets': '94843238260',
                                                                       'in_ucast_pkts': '66732287',
                                                                       'links': [       {       'in_nucast_pkts': '1464037',
                                                                                                'in_octets': '94843238260',
                                                                                                'in_ucast_pkts': '66732287',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4747236',
                                                                                                'out_octets': '4409333799',
                                                                                                'out_ucast_pkts': '27478767',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4747236',
                                                                       'out_octets': '4409333799',
                                                                       'out_ucast_pkts': '27478767',
                                                                       'poller_hash': 'MjhGNjUzcGVlcmludGZwb24gMTEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '110',
                                                                                               'links': [       {       'connected_time': '688478',
                                                                                                                        'fiber_length': 864,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjhGNjUzcGVlcnBvbnBvbiAxMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '27 Mill St. W. Milverton (Mary '
                                                 'Francis Lambert-Hill)',
                                         'poller_hash': 'MjhGNjUz',
                                         'serial_number': '28F653',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:3b:95:a5',
                                         'uptime': '688478'},
                                 {       'counter_source': '111/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '111/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '121003374091',
                                                                       'in_ucast_pkts': '409554174',
                                                                       'mac_address': 'ec:4f:82:4e:e6:0f',
                                                                       'name': '111/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '266',
                                                                       'out_octets': '2069035265272',
                                                                       'out_ucast_pkts': '1502300827',
                                                                       'poller_hash': 'Mjk2RkM0cGVlcmludGYxMTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '111',
                                                                       'in_nucast_pkts': '266',
                                                                       'in_octets': '2069035265272',
                                                                       'in_ucast_pkts': '1502300827',
                                                                       'links': [       {       'in_nucast_pkts': '266',
                                                                                                'in_octets': '2069035265272',
                                                                                                'in_ucast_pkts': '1502300827',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '121003374091',
                                                                                                'out_ucast_pkts': '409554174',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '121003374091',
                                                                       'out_ucast_pkts': '409554174',
                                                                       'poller_hash': 'Mjk2RkM0cGVlcmludGZwb24gMTEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '111',
                                                                                               'links': [       {       'connected_time': '2138497',
                                                                                                                        'fiber_length': 1090,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkM0cGVlcnBvbnBvbiAxMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '23 Mill St. W. Milverton (James '
                                                 'Wagler 10401047)',
                                         'poller_hash': 'Mjk2RkM0',
                                         'serial_number': '296FC4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4e:e6:0f',
                                         'uptime': '2138497'},
                                 {       'counter_source': '112/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '112/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '544101591545',
                                                                       'in_ucast_pkts': '65887662926',
                                                                       'mac_address': 'ec:4f:82:4e:e5:af',
                                                                       'name': '112/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '191',
                                                                       'out_octets': '4933902677574',
                                                                       'out_ucast_pkts': '3745151059',
                                                                       'poller_hash': 'Mjk2RkI4cGVlcmludGYxMTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '112',
                                                                       'in_nucast_pkts': '191',
                                                                       'in_octets': '4933902677574',
                                                                       'in_ucast_pkts': '3745151059',
                                                                       'links': [       {       'in_nucast_pkts': '191',
                                                                                                'in_octets': '4933902677574',
                                                                                                'in_ucast_pkts': '3745151059',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '544101591545',
                                                                                                'out_ucast_pkts': '65887662926',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '544101591545',
                                                                       'out_ucast_pkts': '65887662926',
                                                                       'poller_hash': 'Mjk2RkI4cGVlcmludGZwb24gMTEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '112',
                                                                                               'links': [       {       'connected_time': '5080999',
                                                                                                                        'fiber_length': 1450,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkI4cGVlcnBvbnBvbiAxMTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 Whaley Ave. Milverton (Dave '
                                                 'Paradis 10401473)',
                                         'poller_hash': 'Mjk2RkI4',
                                         'serial_number': '296FB8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4e:e5:af',
                                         'uptime': '5080999'},
                                 {       'counter_source': '113/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5b:5b'],
                                                                       'id': '113/g1',
                                                                       'in_nucast_pkts': '11959',
                                                                       'in_octets': '9525590952',
                                                                       'in_ucast_pkts': '37192293',
                                                                       'mac_address': 'ec:4f:82:4f:7a:32',
                                                                       'mtu': 2000,
                                                                       'name': '113/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464058',
                                                                       'out_octets': '174240286849',
                                                                       'out_ucast_pkts': '125727565',
                                                                       'poller_hash': 'Mjk4MjQ4cGVlcmludGYxMTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '113',
                                                                       'in_nucast_pkts': '1464058',
                                                                       'in_octets': '174240286849',
                                                                       'in_ucast_pkts': '125727565',
                                                                       'links': [       {       'in_nucast_pkts': '1464058',
                                                                                                'in_octets': '174240286849',
                                                                                                'in_ucast_pkts': '125727565',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11959',
                                                                                                'out_octets': '9525590952',
                                                                                                'out_ucast_pkts': '37192293',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11959',
                                                                       'out_octets': '9525590952',
                                                                       'out_ucast_pkts': '37192293',
                                                                       'poller_hash': 'Mjk4MjQ4cGVlcmludGZwb24gMTEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '113',
                                                                                               'links': [       {       'connected_time': '688494',
                                                                                                                        'fiber_length': 1341,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjQ4cGVlcnBvbnBvbiAxMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 Whaley Av. Milverton (MATHESON '
                                                 'TYLER)',
                                         'poller_hash': 'Mjk4MjQ4',
                                         'serial_number': '298248',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:2f',
                                         'uptime': '688494'},
                                 {       'counter_source': '114/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '114/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '138091907936',
                                                                       'in_ucast_pkts': '838548078',
                                                                       'mac_address': '00:06:31:e8:f4:3d',
                                                                       'name': '114/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '212',
                                                                       'out_octets': '3832809809129',
                                                                       'out_ucast_pkts': '2762099443',
                                                                       'poller_hash': 'MUQwNTBGcGVlcmludGYxMTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '114',
                                                                       'in_nucast_pkts': '212',
                                                                       'in_octets': '3832809809129',
                                                                       'in_ucast_pkts': '2762099443',
                                                                       'links': [       {       'in_nucast_pkts': '212',
                                                                                                'in_octets': '3832809809129',
                                                                                                'in_ucast_pkts': '2762099443',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '138091907936',
                                                                                                'out_ucast_pkts': '838548078',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '138091907936',
                                                                       'out_ucast_pkts': '838548078',
                                                                       'poller_hash': 'MUQwNTBGcGVlcmludGZwb24gMTE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '114',
                                                                                               'links': [       {       'connected_time': '71552369',
                                                                                                                        'fiber_length': 1303,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTBGcGVlcnBvbnBvbiAxMTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '19 Whaley Ave. Milverton (Monica&Joe '
                                                 'Phelan 10421673)',
                                         'poller_hash': 'MUQwNTBG',
                                         'serial_number': '1D050F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f4:3d',
                                         'uptime': '71552369'},
                                 {       'counter_source': '115/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ed:1e'],
                                                                       'id': '115/g1',
                                                                       'in_nucast_pkts': '346586',
                                                                       'in_octets': '153043905984',
                                                                       'in_ucast_pkts': '1269657877',
                                                                       'mac_address': '00:06:31:e8:f3:83',
                                                                       'mtu': 2000,
                                                                       'name': '115/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43140836',
                                                                       'out_octets': '5886684414645',
                                                                       'out_ucast_pkts': '4081166040',
                                                                       'poller_hash': 'MUQwNEZBcGVlcmludGYxMTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '115',
                                                                       'in_nucast_pkts': '43140836',
                                                                       'in_octets': '5886684414645',
                                                                       'in_ucast_pkts': '4081166040',
                                                                       'links': [       {       'in_nucast_pkts': '43140836',
                                                                                                'in_octets': '5886684414645',
                                                                                                'in_ucast_pkts': '4081166040',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '346586',
                                                                                                'out_octets': '153043905984',
                                                                                                'out_ucast_pkts': '1269657877',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '346586',
                                                                       'out_octets': '153043905984',
                                                                       'out_ucast_pkts': '1269657877',
                                                                       'poller_hash': 'MUQwNEZBcGVlcmludGZwb24gMTE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '115',
                                                                                               'links': [       {       'connected_time': '18857359',
                                                                                                                        'fiber_length': 1246,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEZBcGVlcnBvbnBvbiAxMTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '21 FULTON ST (BOWMAN RON & TRACEY)',
                                         'poller_hash': 'MUQwNEZB',
                                         'serial_number': '1D04FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:80',
                                         'uptime': '18857359'},
                                 {       'counter_source': '116/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '116/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '234231998907',
                                                                       'in_ucast_pkts': '69628220214',
                                                                       'mac_address': 'ec:4f:82:4e:e6:37',
                                                                       'name': '116/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '167',
                                                                       'out_octets': '4281811238213',
                                                                       'out_ucast_pkts': '3200626853',
                                                                       'poller_hash': 'Mjk2RkM5cGVlcmludGYxMTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '116',
                                                                       'in_nucast_pkts': '167',
                                                                       'in_octets': '4281811238213',
                                                                       'in_ucast_pkts': '3200626853',
                                                                       'links': [       {       'in_nucast_pkts': '167',
                                                                                                'in_octets': '4281811238213',
                                                                                                'in_ucast_pkts': '3200626853',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '234231998907',
                                                                                                'out_ucast_pkts': '69628220214',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '234231998907',
                                                                       'out_ucast_pkts': '69628220214',
                                                                       'poller_hash': 'Mjk2RkM5cGVlcmludGZwb24gMTE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '116',
                                                                                               'links': [       {       'connected_time': '21944183',
                                                                                                                        'fiber_length': 1292,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkM5cGVlcnBvbnBvbiAxMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '17 Whaley Ave. Milverton (Darrel '
                                                 'Roth)',
                                         'poller_hash': 'Mjk2RkM5',
                                         'serial_number': '296FC9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4e:e6:37',
                                         'uptime': '21944183'},
                                 {       'counter_source': '117/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '117/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '679346643077',
                                                                       'in_ucast_pkts': '1165693485',
                                                                       'mac_address': 'ec:4f:82:4f:7a:87',
                                                                       'name': '117/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '133',
                                                                       'out_octets': '2591597955286',
                                                                       'out_ucast_pkts': '1996957707',
                                                                       'poller_hash': 'Mjk4MjUzcGVlcmludGYxMTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '117',
                                                                       'in_nucast_pkts': '133',
                                                                       'in_octets': '2591597955286',
                                                                       'in_ucast_pkts': '1996957707',
                                                                       'links': [       {       'in_nucast_pkts': '133',
                                                                                                'in_octets': '2591597955286',
                                                                                                'in_ucast_pkts': '1996957707',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '679346643077',
                                                                                                'out_ucast_pkts': '1165693485',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '679346643077',
                                                                       'out_ucast_pkts': '1165693485',
                                                                       'poller_hash': 'Mjk4MjUzcGVlcmludGZwb24gMTE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '117',
                                                                                               'links': [       {       'connected_time': '31213287',
                                                                                                                        'fiber_length': 1364,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjUzcGVlcnBvbnBvbiAxMTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 Whaley Ave (FINNIE EVELYN)',
                                         'poller_hash': 'Mjk4MjUz',
                                         'serial_number': '298253',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:87',
                                         'uptime': '31213287'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '118/g1',
                                                                       'name': '118/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk2RkMwcGVlcmludGYxMTgvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '118',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk2RkMwcGVlcmludGZwb24gMTE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '118',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkMwcGVlcnBvbnBvbiAxMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '11 Whaley Ave. Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "118" '
                                                                  '(location: GPON '
                                                                  'port "8/5"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:10:38.340Z',
                                         'poller_hash': 'Mjk2RkMw',
                                         'serial_number': '296FC0',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '119/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '119/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '144761129526',
                                                                       'in_ucast_pkts': '1157910890',
                                                                       'mac_address': 'ec:4f:82:4e:e5:f7',
                                                                       'name': '119/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '113',
                                                                       'out_octets': '6916726603455',
                                                                       'out_ucast_pkts': '8830996015',
                                                                       'poller_hash': 'Mjk2RkMxcGVlcmludGYxMTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '119',
                                                                       'in_nucast_pkts': '113',
                                                                       'in_octets': '6916726603455',
                                                                       'in_ucast_pkts': '8830996015',
                                                                       'links': [       {       'in_nucast_pkts': '113',
                                                                                                'in_octets': '6916726603455',
                                                                                                'in_ucast_pkts': '8830996015',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '144761129526',
                                                                                                'out_ucast_pkts': '1157910890',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '144761129526',
                                                                       'out_ucast_pkts': '1157910890',
                                                                       'poller_hash': 'Mjk2RkMxcGVlcmludGZwb24gMTE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '119',
                                                                                               'links': [       {       'connected_time': '30050458',
                                                                                                                        'fiber_length': 1194,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkMxcGVlcnBvbnBvbiAxMTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 Fulton St. Milverton (Richard '
                                                 'Roulston 10401073)',
                                         'poller_hash': 'Mjk2RkMx',
                                         'serial_number': '296FC1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4e:e5:f7',
                                         'uptime': '30050458'},
                                 {       'counter_source': '120/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '120/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '274777998073',
                                                                       'in_ucast_pkts': '10622603853',
                                                                       'mac_address': '00:06:31:e8:f3:fe',
                                                                       'name': '120/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '224',
                                                                       'out_octets': '6381848279898',
                                                                       'out_ucast_pkts': '4493281883',
                                                                       'poller_hash': 'MUQwNTA4cGVlcmludGYxMjAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '120',
                                                                       'in_nucast_pkts': '224',
                                                                       'in_octets': '6381848279898',
                                                                       'in_ucast_pkts': '4493281883',
                                                                       'links': [       {       'in_nucast_pkts': '224',
                                                                                                'in_octets': '6381848279898',
                                                                                                'in_ucast_pkts': '4493281883',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '274777998073',
                                                                                                'out_ucast_pkts': '10622603853',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '274777998073',
                                                                       'out_ucast_pkts': '10622603853',
                                                                       'poller_hash': 'MUQwNTA4cGVlcmludGZwb24gMTIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '120',
                                                                                               'links': [       {       'connected_time': '7581551',
                                                                                                                        'fiber_length': 1194,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTA4cGVlcnBvbnBvbiAxMjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '14 Fulton St. Milverton (Nathan '
                                                 'Matheson 10426412)',
                                         'poller_hash': 'MUQwNTA4',
                                         'serial_number': '1D0508',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:fe',
                                         'uptime': '7581551'},
                                 {       'counter_source': '121/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:89'],
                                                                       'id': '121/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '2461387758',
                                                                       'in_ucast_pkts': '15551443',
                                                                       'mac_address': '00:06:31:e8:f4:25',
                                                                       'mtu': 2000,
                                                                       'name': '121/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463961',
                                                                       'out_octets': '114353309032',
                                                                       'out_ucast_pkts': '79949739',
                                                                       'poller_hash': 'MUQwNTBDcGVlcmludGYxMjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '121',
                                                                       'in_nucast_pkts': '1463961',
                                                                       'in_octets': '114353309032',
                                                                       'in_ucast_pkts': '79949739',
                                                                       'links': [       {       'in_nucast_pkts': '1463961',
                                                                                                'in_octets': '114353309032',
                                                                                                'in_ucast_pkts': '79949739',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '2461387758',
                                                                                                'out_ucast_pkts': '15551443',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '2461387758',
                                                                       'out_ucast_pkts': '15551443',
                                                                       'poller_hash': 'MUQwNTBDcGVlcmludGZwb24gMTIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '121',
                                                                                               'links': [       {       'connected_time': '688515',
                                                                                                                        'fiber_length': 1195,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTBDcGVlcnBvbnBvbiAxMjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '15 Fulton St. Milverton (Becky '
                                                 'Westman 10427702)',
                                         'poller_hash': 'MUQwNTBD',
                                         'serial_number': '1D050C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f4:22',
                                         'uptime': '688515'},
                                 {       'counter_source': '122/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:54'],
                                                                       'id': '122/g1',
                                                                       'in_nucast_pkts': '150626',
                                                                       'in_octets': '31782343245',
                                                                       'in_ucast_pkts': '154490310',
                                                                       'mac_address': '00:06:31:e8:f4:5b',
                                                                       'mtu': 2000,
                                                                       'name': '122/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17819593',
                                                                       'out_octets': '670262893332',
                                                                       'out_ucast_pkts': '494735476',
                                                                       'poller_hash': 'MUQwNTEycGVlcmludGYxMjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '122',
                                                                       'in_nucast_pkts': '17819593',
                                                                       'in_octets': '670262893332',
                                                                       'in_ucast_pkts': '494735476',
                                                                       'links': [       {       'in_nucast_pkts': '17819593',
                                                                                                'in_octets': '670262893332',
                                                                                                'in_ucast_pkts': '494735476',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '150626',
                                                                                                'out_octets': '31782343245',
                                                                                                'out_ucast_pkts': '154490310',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150626',
                                                                       'out_octets': '31782343245',
                                                                       'out_ucast_pkts': '154490310',
                                                                       'poller_hash': 'MUQwNTEycGVlcmludGZwb24gMTIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '122',
                                                                                               'links': [       {       'connected_time': '7695099',
                                                                                                                        'fiber_length': 1251,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTEycGVlcnBvbnBvbiAxMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '23 Fulton St. Milverton (Doug '
                                                 'Egerdee)',
                                         'poller_hash': 'MUQwNTEy',
                                         'serial_number': '1D0512',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f4:58',
                                         'uptime': '7695099'},
                                 {       'counter_source': '123/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '123/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '371988178552',
                                                                       'in_ucast_pkts': '1331301013',
                                                                       'mac_address': 'ec:4f:82:4e:e6:17',
                                                                       'name': '123/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '165',
                                                                       'out_octets': '4561128588971',
                                                                       'out_ucast_pkts': '3599274497',
                                                                       'poller_hash': 'Mjk2RkM1cGVlcmludGYxMjMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '123',
                                                                       'in_nucast_pkts': '165',
                                                                       'in_octets': '4561128588971',
                                                                       'in_ucast_pkts': '3599274497',
                                                                       'links': [       {       'in_nucast_pkts': '165',
                                                                                                'in_octets': '4561128588971',
                                                                                                'in_ucast_pkts': '3599274497',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '371988178552',
                                                                                                'out_ucast_pkts': '1331301013',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '371988178552',
                                                                       'out_ucast_pkts': '1331301013',
                                                                       'poller_hash': 'Mjk2RkM1cGVlcmludGZwb24gMTIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '123',
                                                                                               'links': [       {       'connected_time': '688502',
                                                                                                                        'fiber_length': 1278,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkM1cGVlcnBvbnBvbiAxMjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 Fulton St. Milverton (Jeff '
                                                 'Leblanc)',
                                         'poller_hash': 'Mjk2RkM1',
                                         'serial_number': '296FC5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4e:e6:17',
                                         'uptime': '688502'},
                                 {       'counter_source': '124/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2d:32:47'],
                                                                       'id': '124/g1',
                                                                       'in_nucast_pkts': '366',
                                                                       'in_octets': '217782984688',
                                                                       'in_ucast_pkts': '1758713255',
                                                                       'mac_address': '00:06:31:52:57:7a',
                                                                       'mtu': 2000,
                                                                       'name': '124/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75606980',
                                                                       'out_octets': '10578000550410',
                                                                       'out_ucast_pkts': '7172572302',
                                                                       'poller_hash': 'OEEwOTBwZWVyaW50ZjEyNC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '124',
                                                                       'in_nucast_pkts': '75606980',
                                                                       'in_octets': '10578000550410',
                                                                       'in_ucast_pkts': '7172572302',
                                                                       'links': [       {       'in_nucast_pkts': '75606980',
                                                                                                'in_octets': '10578000550410',
                                                                                                'in_ucast_pkts': '7172572302',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '366',
                                                                                                'out_octets': '217782984688',
                                                                                                'out_ucast_pkts': '1758713255',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '366',
                                                                       'out_octets': '217782984688',
                                                                       'out_ucast_pkts': '1758713255',
                                                                       'poller_hash': 'OEEwOTBwZWVyaW50ZnBvbiAxMjQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '124',
                                                                                               'links': [       {       'connected_time': '40931',
                                                                                                                        'fiber_length': 830,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEEwOTBwZWVycG9ucG9uIDEyNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '7 Gerber Dr. Milverton (Murray '
                                                 'Pfeifer)',
                                         'poller_hash': 'OEEwOTA=',
                                         'serial_number': '8A090',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:47:49:7c',
                                         'uptime': '40931'},
                                 {       'counter_source': '125/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '125/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2050851863136',
                                                                       'in_ucast_pkts': '93105551698',
                                                                       'mac_address': 'ec:4f:82:4f:7d:a7',
                                                                       'name': '125/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '179',
                                                                       'out_octets': '6646637458256',
                                                                       'out_ucast_pkts': '6348015183',
                                                                       'poller_hash': 'Mjk4MkI3cGVlcmludGYxMjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '125',
                                                                       'in_nucast_pkts': '179',
                                                                       'in_octets': '6646637458256',
                                                                       'in_ucast_pkts': '6348015183',
                                                                       'links': [       {       'in_nucast_pkts': '179',
                                                                                                'in_octets': '6646637458256',
                                                                                                'in_ucast_pkts': '6348015183',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2050851863136',
                                                                                                'out_ucast_pkts': '93105551698',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2050851863136',
                                                                       'out_ucast_pkts': '93105551698',
                                                                       'poller_hash': 'Mjk4MkI3cGVlcmludGZwb24gMTI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '125',
                                                                                               'links': [       {       'connected_time': '11235575',
                                                                                                                        'fiber_length': 1284,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MkI3cGVlcnBvbnBvbiAxMjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 Whaley Ave. Milverton (Adam Sage)',
                                         'poller_hash': 'Mjk4MkI3',
                                         'serial_number': '2982B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7d:a7',
                                         'uptime': '11235575'},
                                 {       'counter_source': '126/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '126/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '472871335693',
                                                                       'in_ucast_pkts': '5402082284',
                                                                       'mac_address': '00:06:31:e9:04:33',
                                                                       'name': '126/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '191',
                                                                       'out_octets': '16594035058712',
                                                                       'out_ucast_pkts': '11163673444',
                                                                       'poller_hash': 'MUQwNkQ3cGVlcmludGYxMjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '126',
                                                                       'in_nucast_pkts': '191',
                                                                       'in_octets': '16594035058712',
                                                                       'in_ucast_pkts': '11163673444',
                                                                       'links': [       {       'in_nucast_pkts': '191',
                                                                                                'in_octets': '16594035058712',
                                                                                                'in_ucast_pkts': '11163673444',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '472871335693',
                                                                                                'out_ucast_pkts': '5402082284',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '472871335693',
                                                                       'out_ucast_pkts': '5402082284',
                                                                       'poller_hash': 'MUQwNkQ3cGVlcmludGZwb24gMTI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '126',
                                                                                               'links': [       {       'connected_time': '13219376',
                                                                                                                        'fiber_length': 1355,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNkQ3cGVlcnBvbnBvbiAxMjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '8 WHALEY AV (STECKLY DOUG)',
                                         'poller_hash': 'MUQwNkQ3',
                                         'serial_number': '1D06D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:04:33',
                                         'uptime': '13219376'},
                                 {       'counter_source': '128/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '128/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '278235884731',
                                                                       'in_ucast_pkts': '95575503828',
                                                                       'mac_address': '00:06:31:e9:06:7c',
                                                                       'name': '128/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '4189516608357',
                                                                       'out_ucast_pkts': '3323665271',
                                                                       'poller_hash': 'MUQwNzE4cGVlcmludGYxMjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '128',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '4189516608357',
                                                                       'in_ucast_pkts': '3323665271',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '4189516608357',
                                                                                                'in_ucast_pkts': '3323665271',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '278235884731',
                                                                                                'out_ucast_pkts': '95575503828',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '278235884731',
                                                                       'out_ucast_pkts': '95575503828',
                                                                       'poller_hash': 'MUQwNzE4cGVlcmludGZwb24gMTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '128',
                                                                                               'links': [       {       'connected_time': '21817027',
                                                                                                                        'fiber_length': 1003,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNzE4cGVlcnBvbnBvbiAxMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '17 Mill St. W. (Peter Wall 10423210)',
                                         'poller_hash': 'MUQwNzE4',
                                         'serial_number': '1D0718',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:06:7c',
                                         'uptime': '21817027'},
                                 {       'counter_source': '129/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '129/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '145875838299',
                                                                       'in_ucast_pkts': '1411159636',
                                                                       'mac_address': 'ec:4f:82:4f:7c:57',
                                                                       'name': '129/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '266',
                                                                       'out_octets': '4761915001347',
                                                                       'out_ucast_pkts': '3324317057',
                                                                       'poller_hash': 'Mjk4MjhEcGVlcmludGYxMjkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '129',
                                                                       'in_nucast_pkts': '266',
                                                                       'in_octets': '4761915001347',
                                                                       'in_ucast_pkts': '3324317057',
                                                                       'links': [       {       'in_nucast_pkts': '266',
                                                                                                'in_octets': '4761915001347',
                                                                                                'in_ucast_pkts': '3324317057',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '145875838299',
                                                                                                'out_ucast_pkts': '1411159636',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '145875838299',
                                                                       'out_ucast_pkts': '1411159636',
                                                                       'poller_hash': 'Mjk4MjhEcGVlcmludGZwb24gMTI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '129',
                                                                                               'links': [       {       'connected_time': '674508',
                                                                                                                        'fiber_length': 1254,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjhEcGVlcnBvbnBvbiAxMjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 Whaley Ave. Milverton (Ron '
                                                 'Kipfer)',
                                         'poller_hash': 'Mjk4MjhE',
                                         'serial_number': '29828D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7c:57',
                                         'uptime': '674508'},
                                 {       'counter_source': '130/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:95:45'],
                                                                       'id': '130/g1',
                                                                       'in_nucast_pkts': '14957',
                                                                       'in_octets': '2236458665',
                                                                       'in_ucast_pkts': '5406427',
                                                                       'mac_address': '00:06:31:e9:02:62',
                                                                       'mtu': 2000,
                                                                       'name': '130/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463881',
                                                                       'out_octets': '26194622069',
                                                                       'out_ucast_pkts': '20849784',
                                                                       'poller_hash': 'MUQwNkExcGVlcmludGYxMzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '130',
                                                                       'in_nucast_pkts': '1463881',
                                                                       'in_octets': '26194622069',
                                                                       'in_ucast_pkts': '20849784',
                                                                       'links': [       {       'in_nucast_pkts': '1463881',
                                                                                                'in_octets': '26194622069',
                                                                                                'in_ucast_pkts': '20849784',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14957',
                                                                                                'out_octets': '2236458665',
                                                                                                'out_ucast_pkts': '5406427',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14957',
                                                                       'out_octets': '2236458665',
                                                                       'out_ucast_pkts': '5406427',
                                                                       'poller_hash': 'MUQwNkExcGVlcmludGZwb24gMTMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '130',
                                                                                               'links': [       {       'connected_time': '688508',
                                                                                                                        'fiber_length': 1307,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNkExcGVlcnBvbnBvbiAxMzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '14 Whaley Ave. Milverton (MCARTHUR '
                                                 'JIM)',
                                         'poller_hash': 'MUQwNkEx',
                                         'serial_number': '1D06A1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:02:5f',
                                         'uptime': '688508'},
                                 {       'counter_source': '131/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:08:d2:ed'],
                                                                       'id': '131/g1',
                                                                       'in_nucast_pkts': '15242',
                                                                       'in_octets': '9331167664',
                                                                       'in_ucast_pkts': '39139938',
                                                                       'mac_address': 'ec:4f:82:7d:c9:65',
                                                                       'mtu': 2000,
                                                                       'name': '131/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463984',
                                                                       'out_octets': '121426710105',
                                                                       'out_ucast_pkts': '89219041',
                                                                       'poller_hash': 'MzAwRERGcGVlcmludGYxMzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '131',
                                                                       'in_nucast_pkts': '1463984',
                                                                       'in_octets': '121426710105',
                                                                       'in_ucast_pkts': '89219041',
                                                                       'links': [       {       'in_nucast_pkts': '1463984',
                                                                                                'in_octets': '121426710105',
                                                                                                'in_ucast_pkts': '89219041',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15242',
                                                                                                'out_octets': '9331167664',
                                                                                                'out_ucast_pkts': '39139938',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15242',
                                                                       'out_octets': '9331167664',
                                                                       'out_ucast_pkts': '39139938',
                                                                       'poller_hash': 'MzAwRERGcGVlcmludGZwb24gMTMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '131',
                                                                                               'links': [       {       'connected_time': '688497',
                                                                                                                        'fiber_length': 1312,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRERGcGVlcnBvbnBvbiAxMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5 Whaley Ave. Milverton (DENOMME '
                                                 'ZACH 10438436)',
                                         'poller_hash': 'MzAwRERG',
                                         'serial_number': '300DDF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c9:65',
                                         'uptime': '688497'},
                                 {       'counter_source': '132/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '132/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '393866565684',
                                                                       'in_ucast_pkts': '1932532367',
                                                                       'mac_address': 'ec:4f:82:4f:7a:e7',
                                                                       'name': '132/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '222',
                                                                       'out_octets': '5586345244993',
                                                                       'out_ucast_pkts': '4098617665',
                                                                       'poller_hash': 'Mjk4MjVGcGVlcmludGYxMzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '132',
                                                                       'in_nucast_pkts': '222',
                                                                       'in_octets': '5586345244993',
                                                                       'in_ucast_pkts': '4098617665',
                                                                       'links': [       {       'in_nucast_pkts': '222',
                                                                                                'in_octets': '5586345244993',
                                                                                                'in_ucast_pkts': '4098617665',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '393866565684',
                                                                                                'out_ucast_pkts': '1932532367',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '393866565684',
                                                                       'out_ucast_pkts': '1932532367',
                                                                       'poller_hash': 'Mjk4MjVGcGVlcmludGZwb24gMTMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '132',
                                                                                               'links': [       {       'connected_time': '19814845',
                                                                                                                        'fiber_length': 1496,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjVGcGVlcnBvbnBvbiAxMzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '9 MacBeth St. Milverton (Don Gascho)',
                                         'poller_hash': 'Mjk4MjVG',
                                         'serial_number': '29825F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:e7',
                                         'uptime': '19814845'},
                                 {       'counter_source': '133/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '133/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '87600988904',
                                                                       'in_ucast_pkts': '1050104658',
                                                                       'mac_address': '00:06:31:e9:02:20',
                                                                       'name': '133/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '6764244310376',
                                                                       'out_ucast_pkts': '4538007308',
                                                                       'poller_hash': 'MUQwNjlBcGVlcmludGYxMzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '133',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '6764244310376',
                                                                       'in_ucast_pkts': '4538007308',
                                                                       'links': [       {       'in_nucast_pkts': '117',
                                                                                                'in_octets': '6764244310376',
                                                                                                'in_ucast_pkts': '4538007308',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '87600988904',
                                                                                                'out_ucast_pkts': '1050104658',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '87600988904',
                                                                       'out_ucast_pkts': '1050104658',
                                                                       'poller_hash': 'MUQwNjlBcGVlcmludGZwb24gMTMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '133',
                                                                                               'links': [       {       'connected_time': '29450620',
                                                                                                                        'fiber_length': 1234,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjlBcGVlcnBvbnBvbiAxMzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '18 Fulton St. Milverton (Joe '
                                                 'Schlueter)',
                                         'poller_hash': 'MUQwNjlB',
                                         'serial_number': '1D069A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:02:20',
                                         'uptime': '29450620'},
                                 {       'counter_source': '134/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:7f:72'],
                                                                       'id': '134/g1',
                                                                       'in_nucast_pkts': '801935',
                                                                       'in_octets': '381899528506',
                                                                       'in_ucast_pkts': '2068877833',
                                                                       'mac_address': '00:06:31:e8:f3:32',
                                                                       'mtu': 2000,
                                                                       'name': '134/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76061852',
                                                                       'out_octets': '9411526450184',
                                                                       'out_ucast_pkts': '6736506404',
                                                                       'poller_hash': 'MUQwNEYxcGVlcmludGYxMzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '134',
                                                                       'in_nucast_pkts': '76061852',
                                                                       'in_octets': '9411526450184',
                                                                       'in_ucast_pkts': '6736506404',
                                                                       'links': [       {       'in_nucast_pkts': '76061852',
                                                                                                'in_octets': '9411526450184',
                                                                                                'in_ucast_pkts': '6736506404',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '801935',
                                                                                                'out_octets': '381899528506',
                                                                                                'out_ucast_pkts': '2068877833',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '801935',
                                                                       'out_octets': '381899528506',
                                                                       'out_ucast_pkts': '2068877833',
                                                                       'poller_hash': 'MUQwNEYxcGVlcmludGZwb24gMTM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '134',
                                                                                               'links': [       {       'connected_time': '21816931',
                                                                                                                        'fiber_length': 1371,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEYxcGVlcnBvbnBvbiAxMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '4 WHALEY AVE. (Rod Steckly)',
                                         'poller_hash': 'MUQwNEYx',
                                         'serial_number': '1D04F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:2f',
                                         'uptime': '21816931'},
                                 {       'counter_source': '135/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '135/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '151598777130',
                                                                       'in_ucast_pkts': '1053699351',
                                                                       'mac_address': 'ec:4f:82:4f:7b:2f',
                                                                       'name': '135/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '126',
                                                                       'out_octets': '4762940469499',
                                                                       'out_ucast_pkts': '3359228697',
                                                                       'poller_hash': 'Mjk4MjY4cGVlcmludGYxMzUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '135',
                                                                       'in_nucast_pkts': '126',
                                                                       'in_octets': '4762940469499',
                                                                       'in_ucast_pkts': '3359228697',
                                                                       'links': [       {       'in_nucast_pkts': '126',
                                                                                                'in_octets': '4762940469499',
                                                                                                'in_ucast_pkts': '3359228697',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '151598777130',
                                                                                                'out_ucast_pkts': '1053699351',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '151598777130',
                                                                       'out_ucast_pkts': '1053699351',
                                                                       'poller_hash': 'Mjk4MjY4cGVlcmludGZwb24gMTM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '135',
                                                                                               'links': [       {       'connected_time': '688470',
                                                                                                                        'fiber_length': 910,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjY4cGVlcnBvbnBvbiAxMzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 William St. Milverton (ROPP ADAM)',
                                         'poller_hash': 'Mjk4MjY4',
                                         'serial_number': '298268',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7b:2f',
                                         'uptime': '688470'},
                                 {       'counter_source': '136/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '136/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5568409984342',
                                                                       'in_ucast_pkts': '369751969577',
                                                                       'mac_address': '00:06:31:e8:f3:89',
                                                                       'name': '136/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '8468015401679',
                                                                       'out_ucast_pkts': '7690817994',
                                                                       'poller_hash': 'MUQwNEZCcGVlcmludGYxMzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '136',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '8468015401679',
                                                                       'in_ucast_pkts': '7690817994',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '8468015401679',
                                                                                                'in_ucast_pkts': '7690817994',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '5568409984342',
                                                                                                'out_ucast_pkts': '369751969577',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5568409984342',
                                                                       'out_ucast_pkts': '369751969577',
                                                                       'poller_hash': 'MUQwNEZCcGVlcmludGZwb24gMTM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '136',
                                                                                               'links': [       {       'connected_time': '688510',
                                                                                                                        'fiber_length': 1128,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEZCcGVlcnBvbnBvbiAxMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '11 FULTON ST (DEVRIES MARTIN)',
                                         'poller_hash': 'MUQwNEZC',
                                         'serial_number': '1D04FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:89',
                                         'uptime': '688510'},
                                 {       'counter_source': '137/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:ed:f6'],
                                                                       'id': '137/g1',
                                                                       'in_nucast_pkts': '1769',
                                                                       'in_octets': '129283058005',
                                                                       'in_ucast_pkts': '287184325',
                                                                       'mac_address': 'ec:4f:82:67:71:d1',
                                                                       'mtu': 2000,
                                                                       'name': '137/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19475831',
                                                                       'out_octets': '1310866529577',
                                                                       'out_ucast_pkts': '976655744',
                                                                       'poller_hash': 'MjVEOTJFcGVlcmludGYxMzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137',
                                                                       'in_nucast_pkts': '19475831',
                                                                       'in_octets': '1310866529577',
                                                                       'in_ucast_pkts': '976655744',
                                                                       'links': [       {       'in_nucast_pkts': '19475831',
                                                                                                'in_octets': '1310866529577',
                                                                                                'in_ucast_pkts': '976655744',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1769',
                                                                                                'out_octets': '129283058005',
                                                                                                'out_ucast_pkts': '287184325',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1769',
                                                                       'out_octets': '129283058005',
                                                                       'out_ucast_pkts': '287184325',
                                                                       'poller_hash': 'MjVEOTJFcGVlcmludGZwb24gMTM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137',
                                                                                               'links': [       {       'connected_time': '10784823',
                                                                                                                        'fiber_length': 653,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.638,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjVEOTJFcGVlcnBvbnBvbiAxMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '35 Main St. S. Milverton (Randy '
                                                 'Martin 10434943)',
                                         'poller_hash': 'MjVEOTJF',
                                         'serial_number': '25D92E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:67:71:cd',
                                         'uptime': '10784823'},
                                 {       'counter_source': '138/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:94:ce'],
                                                                       'id': '138/g1',
                                                                       'in_nucast_pkts': '25',
                                                                       'in_octets': '4701395179',
                                                                       'in_ucast_pkts': '13585259',
                                                                       'mac_address': 'ec:4f:82:42:e7:e7',
                                                                       'mtu': 2000,
                                                                       'name': '138/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11147468',
                                                                       'out_octets': '60961203820',
                                                                       'out_ucast_pkts': '49303785',
                                                                       'poller_hash': 'MjM3QTI1cGVlcmludGYxMzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138',
                                                                       'in_nucast_pkts': '11147468',
                                                                       'in_octets': '60961203820',
                                                                       'in_ucast_pkts': '49303785',
                                                                       'links': [       {       'in_nucast_pkts': '11147468',
                                                                                                'in_octets': '60961203820',
                                                                                                'in_ucast_pkts': '49303785',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '25',
                                                                                                'out_octets': '4701395179',
                                                                                                'out_ucast_pkts': '13585259',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '25',
                                                                       'out_octets': '4701395179',
                                                                       'out_ucast_pkts': '13585259',
                                                                       'poller_hash': 'MjM3QTI1cGVlcmludGZwb24gMTM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138',
                                                                                               'links': [       {       'connected_time': '6129406',
                                                                                                                        'fiber_length': 902,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.846,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjM3QTI1cGVlcnBvbnBvbiAxMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '37 Mill St. W. Milverton (John '
                                                 'Gerber Construction 10401482)',
                                         'poller_hash': 'MjM3QTI1',
                                         'serial_number': '237A25',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:42:e7:e3',
                                         'uptime': '6129406'},
                                 {       'counter_source': '139/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '139/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '37307404101',
                                                                       'in_ucast_pkts': '475499357',
                                                                       'mac_address': 'ec:4f:82:4f:7b:cf',
                                                                       'name': '139/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '3040616867677',
                                                                       'out_ucast_pkts': '2050355436',
                                                                       'poller_hash': 'Mjk4MjdDcGVlcmludGYxMzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '139',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '3040616867677',
                                                                       'in_ucast_pkts': '2050355436',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '3040616867677',
                                                                                                'in_ucast_pkts': '2050355436',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '37307404101',
                                                                                                'out_ucast_pkts': '475499357',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '37307404101',
                                                                       'out_ucast_pkts': '475499357',
                                                                       'poller_hash': 'Mjk4MjdDcGVlcmludGZwb24gMTM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139',
                                                                                               'links': [       {       'connected_time': '24902851',
                                                                                                                        'fiber_length': 1361,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjdDcGVlcnBvbnBvbiAxMzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 Pierson St. Milverton (Mryna '
                                                 'Schmidt 10400672)',
                                         'poller_hash': 'Mjk4MjdD',
                                         'serial_number': '29827C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7b:cf',
                                         'uptime': '24902851'},
                                 {       'counter_source': '140/g2+140/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:12:e4:cd'],
                                                                       'id': '140/g2',
                                                                       'in_nucast_pkts': '212',
                                                                       'in_octets': '145163703459',
                                                                       'in_ucast_pkts': '492671297',
                                                                       'mac_address': 'ec:4f:82:42:ed:e8',
                                                                       'mtu': 2000,
                                                                       'name': '140/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27158690',
                                                                       'out_octets': '1971761512979',
                                                                       'out_ucast_pkts': '1487851247',
                                                                       'poller_hash': 'MjM3QUU1cGVlcmludGYxNDAvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2f:cc:9c'],
                                                                       'id': '140/g4',
                                                                       'in_nucast_pkts': '5369168',
                                                                       'in_octets': '1320490058',
                                                                       'in_ucast_pkts': '4366063',
                                                                       'mac_address': 'ec:4f:82:42:ed:ea',
                                                                       'mtu': 2000,
                                                                       'name': '140/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9732215999',
                                                                       'out_octets': '13020529406554',
                                                                       'out_ucast_pkts': '3813804',
                                                                       'poller_hash': 'MjM3QUU1cGVlcmludGYxNDAvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140',
                                                                       'in_nucast_pkts': '9759374689',
                                                                       'in_octets': '14992290919533',
                                                                       'in_ucast_pkts': '1491665051',
                                                                       'links': [       {       'in_nucast_pkts': '9759374689',
                                                                                                'in_octets': '14992290919533',
                                                                                                'in_ucast_pkts': '1491665051',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5369380',
                                                                                                'out_octets': '146484193517',
                                                                                                'out_ucast_pkts': '497037360',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5369380',
                                                                       'out_octets': '146484193517',
                                                                       'out_ucast_pkts': '497037360',
                                                                       'poller_hash': 'MjM3QUU1cGVlcmludGZwb24gMTQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140',
                                                                                               'links': [       {       'connected_time': '15498457',
                                                                                                                        'fiber_length': 992,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.912,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjM3QUU1cGVlcnBvbnBvbiAxNDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '14 Mill St. W. Milverton (Mill Wood '
                                                 'Manor)',
                                         'poller_hash': 'MjM3QUU1',
                                         'serial_number': '237AE5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:42:ed:e3',
                                         'uptime': '15498457'},
                                 {       'counter_source': '141/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e7:5f'],
                                                                       'id': '141/g1',
                                                                       'in_nucast_pkts': '19772',
                                                                       'in_octets': '86192202848',
                                                                       'in_ucast_pkts': '705114869',
                                                                       'mac_address': 'ec:4f:82:4f:7a:62',
                                                                       'mtu': 2000,
                                                                       'name': '141/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '72095703',
                                                                       'out_octets': '4467410172946',
                                                                       'out_ucast_pkts': '3026447457',
                                                                       'poller_hash': 'Mjk4MjRFcGVlcmludGYxNDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141',
                                                                       'in_nucast_pkts': '72095703',
                                                                       'in_octets': '4467410172946',
                                                                       'in_ucast_pkts': '3026447457',
                                                                       'links': [       {       'in_nucast_pkts': '72095703',
                                                                                                'in_octets': '4467410172946',
                                                                                                'in_ucast_pkts': '3026447457',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19772',
                                                                                                'out_octets': '86192202848',
                                                                                                'out_ucast_pkts': '705114869',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19772',
                                                                       'out_octets': '86192202848',
                                                                       'out_ucast_pkts': '705114869',
                                                                       'poller_hash': 'Mjk4MjRFcGVlcmludGZwb24gMTQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141',
                                                                                               'links': [       {       'connected_time': '31460818',
                                                                                                                        'fiber_length': 1185,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjRFcGVlcnBvbnBvbiAxNDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 West St. Milverton (Richard '
                                                 'Harrow)',
                                         'poller_hash': 'Mjk4MjRF',
                                         'serial_number': '29824E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:5f',
                                         'uptime': '31460818'},
                                 {       'counter_source': '143/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '498832924835',
                                                                       'in_ucast_pkts': '1867527296',
                                                                       'mac_address': 'ec:4f:82:4f:7c:37',
                                                                       'name': '143/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '4465740546109',
                                                                       'out_ucast_pkts': '3709436189',
                                                                       'poller_hash': 'Mjk4Mjg5cGVlcmludGYxNDMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '4465740546109',
                                                                       'in_ucast_pkts': '3709436189',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '4465740546109',
                                                                                                'in_ucast_pkts': '3709436189',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '498832924835',
                                                                                                'out_ucast_pkts': '1867527296',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '498832924835',
                                                                       'out_ucast_pkts': '1867527296',
                                                                       'poller_hash': 'Mjk4Mjg5cGVlcmludGZwb24gMTQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143',
                                                                                               'links': [       {       'connected_time': '21533843',
                                                                                                                        'fiber_length': 799,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4Mjg5cGVlcnBvbnBvbiAxNDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 William St. Milverton (Bill Moore '
                                                 '10404221)',
                                         'poller_hash': 'Mjk4Mjg5',
                                         'serial_number': '298289',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7c:37',
                                         'uptime': '21533843'},
                                 {       'counter_source': '144/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '144/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '21247329189',
                                                                       'in_ucast_pkts': '121713897',
                                                                       'mac_address': 'ec:4f:82:4f:77:e7',
                                                                       'name': '144/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '430559497447',
                                                                       'out_ucast_pkts': '329800327',
                                                                       'poller_hash': 'Mjk4MUZGcGVlcmludGYxNDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '430559497447',
                                                                       'in_ucast_pkts': '329800327',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '430559497447',
                                                                                                'in_ucast_pkts': '329800327',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '21247329189',
                                                                                                'out_ucast_pkts': '121713897',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '21247329189',
                                                                       'out_ucast_pkts': '121713897',
                                                                       'poller_hash': 'Mjk4MUZGcGVlcmludGZwb24gMTQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144',
                                                                                               'links': [       {       'connected_time': '1024852',
                                                                                                                        'fiber_length': 1128,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MUZGcGVlcnBvbnBvbiAxNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '18 West St. Milverton (Brian & Amy '
                                                 'Martin 10434104)',
                                         'poller_hash': 'Mjk4MUZG',
                                         'serial_number': '2981FF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:77:e7',
                                         'uptime': '1024852'},
                                 {       'counter_source': '145/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '34:0a:33:81:59:d5'],
                                                                       'id': '145/g1',
                                                                       'in_nucast_pkts': '41967',
                                                                       'in_octets': '9412763854',
                                                                       'in_ucast_pkts': '49563886',
                                                                       'mac_address': 'ec:4f:82:63:54:05',
                                                                       'mtu': 2000,
                                                                       'name': '145/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465360',
                                                                       'out_octets': '278502084128',
                                                                       'out_ucast_pkts': '208066315',
                                                                       'poller_hash': 'MkJEMjQxcGVlcmludGYxNDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145',
                                                                       'in_nucast_pkts': '1465360',
                                                                       'in_octets': '278502084128',
                                                                       'in_ucast_pkts': '208066315',
                                                                       'links': [       {       'in_nucast_pkts': '1465360',
                                                                                                'in_octets': '278502084128',
                                                                                                'in_ucast_pkts': '208066315',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '41967',
                                                                                                'out_octets': '9412763854',
                                                                                                'out_ucast_pkts': '49563886',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41967',
                                                                       'out_octets': '9412763854',
                                                                       'out_ucast_pkts': '49563886',
                                                                       'poller_hash': 'MkJEMjQxcGVlcmludGZwb24gMTQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145',
                                                                                               'links': [       {       'connected_time': '688345',
                                                                                                                        'fiber_length': 838,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJEMjQxcGVlcnBvbnBvbiAxNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '9 Pugh St. W. Milverton (Ashley '
                                                 'Mayberry)',
                                         'poller_hash': 'MkJEMjQx',
                                         'serial_number': '2BD241',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:54:05',
                                         'uptime': '688345'},
                                 {       'counter_source': '146/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15815488351',
                                                                       'in_ucast_pkts': '103363825',
                                                                       'mac_address': '00:06:31:e9:01:d8',
                                                                       'name': '146/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '451030515582',
                                                                       'out_ucast_pkts': '337854339',
                                                                       'poller_hash': 'MUQwNjkycGVlcmludGYxNDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '451030515582',
                                                                       'in_ucast_pkts': '337854339',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '451030515582',
                                                                                                'in_ucast_pkts': '337854339',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '15815488351',
                                                                                                'out_ucast_pkts': '103363825',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15815488351',
                                                                       'out_ucast_pkts': '103363825',
                                                                       'poller_hash': 'MUQwNjkycGVlcmludGZwb24gMTQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146',
                                                                                               'links': [       {       'connected_time': '16308318',
                                                                                                                        'fiber_length': 1000,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjkycGVlcnBvbnBvbiAxNDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '9 WEST ST (TRAINOR CASSIDY)',
                                         'poller_hash': 'MUQwNjky',
                                         'serial_number': '1D0692',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:d8',
                                         'uptime': '16308318'},
                                 {       'counter_source': '147/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '138930132649',
                                                                       'in_ucast_pkts': '321945455',
                                                                       'mac_address': 'ec:4f:82:4f:7a:ff',
                                                                       'name': '147/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '751411398454',
                                                                       'out_ucast_pkts': '677258796',
                                                                       'poller_hash': 'Mjk4MjYycGVlcmludGYxNDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '751411398454',
                                                                       'in_ucast_pkts': '677258796',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '751411398454',
                                                                                                'in_ucast_pkts': '677258796',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '138930132649',
                                                                                                'out_ucast_pkts': '321945455',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '138930132649',
                                                                       'out_ucast_pkts': '321945455',
                                                                       'poller_hash': 'Mjk4MjYycGVlcmludGZwb24gMTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147',
                                                                                               'links': [       {       'connected_time': '21816177',
                                                                                                                        'fiber_length': 1205,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjYycGVlcnBvbnBvbiAxNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '8 WEST ST (DAGG DEVIN)',
                                         'poller_hash': 'Mjk4MjYy',
                                         'serial_number': '298262',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:ff',
                                         'uptime': '21816177'},
                                 {       'counter_source': '148/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '148/G1',
                                                                       'in_errors': '1',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '644420474836',
                                                                       'in_ucast_pkts': '3147027418',
                                                                       'mac_address': '00:06:31:e9:03:25',
                                                                       'name': '148/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '174',
                                                                       'out_octets': '17346364060887',
                                                                       'out_ucast_pkts': '12593649741',
                                                                       'poller_hash': 'MUQwNkI5cGVlcmludGYxNDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '148',
                                                                       'in_nucast_pkts': '174',
                                                                       'in_octets': '17346364060887',
                                                                       'in_ucast_pkts': '12593649741',
                                                                       'links': [       {       'in_nucast_pkts': '174',
                                                                                                'in_octets': '17346364060887',
                                                                                                'in_ucast_pkts': '12593649741',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '644420474836',
                                                                                                'out_ucast_pkts': '3147027418',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '644420474836',
                                                                       'out_ucast_pkts': '3147027418',
                                                                       'poller_hash': 'MUQwNkI5cGVlcmludGZwb24gMTQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '148',
                                                                                               'links': [       {       'connected_time': '688500',
                                                                                                                        'fiber_length': 1326,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNkI5cGVlcnBvbnBvbiAxNDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '16 Whaley Ave. Milverton (Sherri '
                                                 'Gilbert)',
                                         'poller_hash': 'MUQwNkI5',
                                         'serial_number': '1D06B9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:03:25',
                                         'uptime': '688500'},
                                 {       'counter_source': '149/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '560639487862',
                                                                       'in_ucast_pkts': '50283447007',
                                                                       'mac_address': '00:06:31:e9:02:29',
                                                                       'name': '149/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '160',
                                                                       'out_octets': '13204715801269',
                                                                       'out_ucast_pkts': '9509533623',
                                                                       'poller_hash': 'MUQwNjlCcGVlcmludGYxNDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149',
                                                                       'in_nucast_pkts': '160',
                                                                       'in_octets': '13204715801269',
                                                                       'in_ucast_pkts': '9509533623',
                                                                       'links': [       {       'in_nucast_pkts': '160',
                                                                                                'in_octets': '13204715801269',
                                                                                                'in_ucast_pkts': '9509533623',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '560639487862',
                                                                                                'out_ucast_pkts': '50283447007',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '560639487862',
                                                                       'out_ucast_pkts': '50283447007',
                                                                       'poller_hash': 'MUQwNjlCcGVlcmludGZwb24gMTQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149',
                                                                                               'links': [       {       'connected_time': '688461',
                                                                                                                        'fiber_length': 1398,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjlCcGVlcnBvbnBvbiAxNDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '13 MacBeth St. Milverton (James '
                                                 'Adams-Kipfer)',
                                         'poller_hash': 'MUQwNjlC',
                                         'serial_number': '1D069B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:02:29',
                                         'uptime': '688461'},
                                 {       'counter_source': '150/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '150/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '63457684090',
                                                                       'in_ucast_pkts': '13092406543',
                                                                       'mac_address': 'ec:4f:82:4f:77:17',
                                                                       'name': '150/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '945931530684',
                                                                       'out_ucast_pkts': '711698087',
                                                                       'poller_hash': 'Mjk4MUU1cGVlcmludGYxNTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '150',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '945931530684',
                                                                       'in_ucast_pkts': '711698087',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '945931530684',
                                                                                                'in_ucast_pkts': '711698087',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '63457684090',
                                                                                                'out_ucast_pkts': '13092406543',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '63457684090',
                                                                       'out_ucast_pkts': '13092406543',
                                                                       'poller_hash': 'Mjk4MUU1cGVlcmludGZwb24gMTUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150',
                                                                                               'links': [       {       'connected_time': '6229370',
                                                                                                                        'fiber_length': 1269,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MUU1cGVlcnBvbnBvbiAxNTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29 West St. Milverton (Tim Ellis '
                                                 '10401774)',
                                         'poller_hash': 'Mjk4MUU1',
                                         'serial_number': '2981E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:77:17',
                                         'uptime': '6229370'},
                                 {       'counter_source': '151/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '151/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '483240470794',
                                                                       'in_ucast_pkts': '3429259977',
                                                                       'mac_address': '00:06:31:e9:06:58',
                                                                       'name': '151/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '92',
                                                                       'out_octets': '12324813467874',
                                                                       'out_ucast_pkts': '5157447139',
                                                                       'poller_hash': 'MUQwNzE0cGVlcmludGYxNTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '151',
                                                                       'in_nucast_pkts': '92',
                                                                       'in_octets': '12324813467874',
                                                                       'in_ucast_pkts': '5157447139',
                                                                       'links': [       {       'in_nucast_pkts': '92',
                                                                                                'in_octets': '12324813467874',
                                                                                                'in_ucast_pkts': '5157447139',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '483240470794',
                                                                                                'out_ucast_pkts': '3429259977',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '483240470794',
                                                                       'out_ucast_pkts': '3429259977',
                                                                       'poller_hash': 'MUQwNzE0cGVlcmludGZwb24gMTUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '151',
                                                                                               'links': [       {       'connected_time': '4304167',
                                                                                                                        'fiber_length': 1319,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNzE0cGVlcnBvbnBvbiAxNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '10 Whaley Ave. Milverton (Dave '
                                                 'Dickson)',
                                         'poller_hash': 'MUQwNzE0',
                                         'serial_number': '1D0714',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:06:58',
                                         'uptime': '4304167'},
                                 {       'counter_source': '152/g1+152/g4+152/g6',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a5:26'],
                                                                       'id': '152/g1',
                                                                       'in_nucast_pkts': '59',
                                                                       'in_octets': '583048539724',
                                                                       'in_ucast_pkts': '5590293060',
                                                                       'mac_address': 'ec:4f:82:d6:cb:14',
                                                                       'mtu': 2000,
                                                                       'name': '152/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43290695',
                                                                       'out_octets': '20385563611687',
                                                                       'out_ucast_pkts': '13967923864',
                                                                       'poller_hash': 'MjZDQzU0cGVlcmludGYxNTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '34:60:f9:0a:f2:d1'],
                                                                       'id': '152/g4',
                                                                       'in_nucast_pkts': '19194',
                                                                       'in_octets': '1573777406081',
                                                                       'in_ucast_pkts': '4142317490',
                                                                       'mac_address': 'ec:4f:82:d6:cb:17',
                                                                       'mtu': 2000,
                                                                       'name': '152/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43298001',
                                                                       'out_octets': '14115429974326',
                                                                       'out_ucast_pkts': '10880415083',
                                                                       'poller_hash': 'MjZDQzU0cGVlcmludGYxNTIvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:0e:10'],
                                                                       'id': '152/g6',
                                                                       'in_nucast_pkts': '2009015',
                                                                       'in_octets': '134279062736',
                                                                       'in_ucast_pkts': '1799932906',
                                                                       'mac_address': 'ec:4f:82:b0:30:83',
                                                                       'mtu': 2000,
                                                                       'name': '152/g6',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '184340885',
                                                                       'out_octets': '10023819621557',
                                                                       'out_ucast_pkts': '7081648996',
                                                                       'poller_hash': 'MjZDQzU0cGVlcmludGYxNTIvZzY=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '152',
                                                                       'in_nucast_pkts': '270929581',
                                                                       'in_octets': '44524813207570',
                                                                       'in_ucast_pkts': '31929987943',
                                                                       'links': [       {       'in_nucast_pkts': '270929581',
                                                                                                'in_octets': '44524813207570',
                                                                                                'in_ucast_pkts': '31929987943',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2028268',
                                                                                                'out_octets': '2291105008541',
                                                                                                'out_ucast_pkts': '11532543456',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2028268',
                                                                       'out_octets': '2291105008541',
                                                                       'out_ucast_pkts': '11532543456',
                                                                       'poller_hash': 'MjZDQzU0cGVlcmludGZwb24gMTUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '152',
                                                                                               'links': [       {       'connected_time': '24664987',
                                                                                                                        'fiber_length': 17944,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjZDQzU0cGVlcnBvbnBvbiAxNTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '214 Winstanley St. Monkton (8 Plex '
                                                 'Apt)',
                                         'poller_hash': 'MjZDQzU0',
                                         'serial_number': '26CC54',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:d6:cb:19',
                                         'uptime': '24664987'},
                                 {       'counter_source': '156/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d8:1b:e1:8f:89'],
                                                                       'id': '156/g1',
                                                                       'in_nucast_pkts': '305922',
                                                                       'in_octets': '9076408813',
                                                                       'in_ucast_pkts': '42794659',
                                                                       'mac_address': 'ec:4f:82:63:56:43',
                                                                       'mtu': 2000,
                                                                       'name': '156/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464242',
                                                                       'out_octets': '256162167272',
                                                                       'out_ucast_pkts': '196008874',
                                                                       'poller_hash': 'MkJEMzYwcGVlcmludGYxNTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156',
                                                                       'in_nucast_pkts': '1464242',
                                                                       'in_octets': '256162167272',
                                                                       'in_ucast_pkts': '196008874',
                                                                       'links': [       {       'in_nucast_pkts': '1464242',
                                                                                                'in_octets': '256162167272',
                                                                                                'in_ucast_pkts': '196008874',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '305922',
                                                                                                'out_octets': '9076408813',
                                                                                                'out_ucast_pkts': '42794659',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '305922',
                                                                       'out_octets': '9076408813',
                                                                       'out_ucast_pkts': '42794659',
                                                                       'poller_hash': 'MkJEMzYwcGVlcmludGZwb24gMTU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156',
                                                                                               'links': [       {       'connected_time': '688352',
                                                                                                                        'fiber_length': 1321,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJEMzYwcGVlcnBvbnBvbiAxNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '23 West St. Milverton (Tyson Londry '
                                                 '10431157)',
                                         'poller_hash': 'MkJEMzYw',
                                         'serial_number': '2BD360',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:56:43',
                                         'uptime': '688352'},
                                 {       'counter_source': '157/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '157/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '64733845993',
                                                                       'in_ucast_pkts': '393277218',
                                                                       'mac_address': '00:06:31:e8:e3:93',
                                                                       'name': '157/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1666065753599',
                                                                       'out_ucast_pkts': '1217642818',
                                                                       'poller_hash': 'MUQwMzM1cGVlcmludGYxNTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '157',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1666065753599',
                                                                       'in_ucast_pkts': '1217642818',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1666065753599',
                                                                                                'in_ucast_pkts': '1217642818',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '64733845993',
                                                                                                'out_ucast_pkts': '393277218',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '64733845993',
                                                                       'out_ucast_pkts': '393277218',
                                                                       'poller_hash': 'MUQwMzM1cGVlcmludGZwb24gMTU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157',
                                                                                               'links': [       {       'connected_time': '6231300',
                                                                                                                        'fiber_length': 1323,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwMzM1cGVlcnBvbnBvbiAxNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '12 Pierson St. Milverton (Kyle '
                                                 'Baillie)',
                                         'poller_hash': 'MUQwMzM1',
                                         'serial_number': '1D0335',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:e3:93',
                                         'uptime': '6231300'},
                                 {       'counter_source': '158/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d4:bc'],
                                                                       'id': '158/g1',
                                                                       'in_nucast_pkts': '11952',
                                                                       'in_octets': '4432356627',
                                                                       'in_ucast_pkts': '10982805',
                                                                       'mac_address': '00:06:31:e9:04:00',
                                                                       'mtu': 2000,
                                                                       'name': '158/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464011',
                                                                       'out_octets': '36995223884',
                                                                       'out_ucast_pkts': '27779625',
                                                                       'poller_hash': 'MUQwNkQxcGVlcmludGYxNTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '158',
                                                                       'in_nucast_pkts': '1464011',
                                                                       'in_octets': '36995223884',
                                                                       'in_ucast_pkts': '27779625',
                                                                       'links': [       {       'in_nucast_pkts': '1464011',
                                                                                                'in_octets': '36995223884',
                                                                                                'in_ucast_pkts': '27779625',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11952',
                                                                                                'out_octets': '4432356627',
                                                                                                'out_ucast_pkts': '10982805',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11952',
                                                                       'out_octets': '4432356627',
                                                                       'out_ucast_pkts': '10982805',
                                                                       'poller_hash': 'MUQwNkQxcGVlcmludGZwb24gMTU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '158',
                                                                                               'links': [       {       'connected_time': '688368',
                                                                                                                        'fiber_length': 962,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNkQxcGVlcnBvbnBvbiAxNTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '26 MAIN ST S (HOWE NICOLE)',
                                         'poller_hash': 'MUQwNkQx',
                                         'serial_number': '1D06D1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:03:fd',
                                         'uptime': '688368'},
                                 {       'counter_source': '159/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e9:b1'],
                                                                       'id': '159/g1',
                                                                       'in_nucast_pkts': '129023',
                                                                       'in_octets': '99400652464',
                                                                       'in_ucast_pkts': '384629769',
                                                                       'mac_address': 'd0:76:8f:aa:49:9f',
                                                                       'mtu': 2000,
                                                                       'name': '159/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14573150',
                                                                       'out_octets': '2144788445160',
                                                                       'out_ucast_pkts': '1511790438',
                                                                       'poller_hash': 'OTU2RTU5cGVlcmludGYxNTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '159',
                                                                       'in_nucast_pkts': '14573150',
                                                                       'in_octets': '2144788445160',
                                                                       'in_ucast_pkts': '1511790438',
                                                                       'links': [       {       'in_nucast_pkts': '14573150',
                                                                                                'in_octets': '2144788445160',
                                                                                                'in_ucast_pkts': '1511790438',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '129023',
                                                                                                'out_octets': '99400652464',
                                                                                                'out_ucast_pkts': '384629769',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '129023',
                                                                       'out_octets': '99400652464',
                                                                       'out_ucast_pkts': '384629769',
                                                                       'poller_hash': 'OTU2RTU5cGVlcmludGZwb24gMTU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '159',
                                                                                               'links': [       {       'connected_time': '6220910',
                                                                                                                        'fiber_length': 1234,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTU5cGVlcnBvbnBvbiAxNTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 West St. Milverton (Tom Broughton '
                                                 '10401292)',
                                         'poller_hash': 'OTU2RTU5',
                                         'serial_number': '956E59',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:49:9c',
                                         'uptime': '6220910'},
                                 {       'counter_source': '160/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:6a:f4'],
                                                                       'id': '160/g1',
                                                                       'in_nucast_pkts': '289819',
                                                                       'in_octets': '63887590398',
                                                                       'in_ucast_pkts': '734079880',
                                                                       'mac_address': '00:06:31:e8:f3:c2',
                                                                       'mtu': 2000,
                                                                       'name': '160/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36517218',
                                                                       'out_octets': '3781335881931',
                                                                       'out_ucast_pkts': '2525156997',
                                                                       'poller_hash': 'MUQwNTAxcGVlcmludGYxNjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '160',
                                                                       'in_nucast_pkts': '36517218',
                                                                       'in_octets': '3781335881931',
                                                                       'in_ucast_pkts': '2525156997',
                                                                       'links': [       {       'in_nucast_pkts': '36517218',
                                                                                                'in_octets': '3781335881931',
                                                                                                'in_ucast_pkts': '2525156997',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '289819',
                                                                                                'out_octets': '63887590398',
                                                                                                'out_ucast_pkts': '734079880',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '289819',
                                                                       'out_octets': '63887590398',
                                                                       'out_ucast_pkts': '734079880',
                                                                       'poller_hash': 'MUQwNTAxcGVlcmludGZwb24gMTYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '160',
                                                                                               'links': [       {       'connected_time': '16507268',
                                                                                                                        'fiber_length': 1172,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTAxcGVlcnBvbnBvbiAxNjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '1 Grosch St. Milverton (Joyce '
                                                 'Riddell 10432778)',
                                         'poller_hash': 'MUQwNTAx',
                                         'serial_number': '1D0501',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:bf',
                                         'uptime': '16507268'},
                                 {       'counter_source': '161/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '161/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '783244328749',
                                                                       'in_ucast_pkts': '80199830396',
                                                                       'mac_address': 'ec:4f:82:4f:79:17',
                                                                       'name': '161/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '167',
                                                                       'out_octets': '13104451312196',
                                                                       'out_ucast_pkts': '9524772240',
                                                                       'poller_hash': 'Mjk4MjI1cGVlcmludGYxNjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '161',
                                                                       'in_nucast_pkts': '167',
                                                                       'in_octets': '13104451312196',
                                                                       'in_ucast_pkts': '9524772240',
                                                                       'links': [       {       'in_nucast_pkts': '167',
                                                                                                'in_octets': '13104451312196',
                                                                                                'in_ucast_pkts': '9524772240',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '783244328749',
                                                                                                'out_ucast_pkts': '80199830396',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '783244328749',
                                                                       'out_ucast_pkts': '80199830396',
                                                                       'poller_hash': 'Mjk4MjI1cGVlcmludGZwb24gMTYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '161',
                                                                                               'links': [       {       'connected_time': '16862227',
                                                                                                                        'fiber_length': 1235,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjI1cGVlcnBvbnBvbiAxNjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 Fulton St. Milverton (Jill Ross '
                                                 '10430010)',
                                         'poller_hash': 'Mjk4MjI1',
                                         'serial_number': '298225',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:79:17',
                                         'uptime': '16862227'},
                                 {       'counter_source': '162/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '162/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '305534140760',
                                                                       'in_ucast_pkts': '456200196',
                                                                       'mac_address': '60:db:98:4e:6d:ec',
                                                                       'name': '162/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '790925644586',
                                                                       'out_ucast_pkts': '815214787',
                                                                       'poller_hash': 'QTlDMzdEcGVlcmludGYxNjIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '162',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '790925644586',
                                                                       'in_ucast_pkts': '815214787',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '790925644586',
                                                                                                'in_ucast_pkts': '815214787',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '305534140760',
                                                                                                'out_ucast_pkts': '456200196',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '305534140760',
                                                                       'out_ucast_pkts': '456200196',
                                                                       'poller_hash': 'QTlDMzdEcGVlcmludGZwb24gMTYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '162',
                                                                                               'links': [       {       'connected_time': '20037947',
                                                                                                                        'fiber_length': 1017,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMzdEcGVlcnBvbnBvbiAxNjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6A Main St. S. Milverton (ALLARD '
                                                 'PIERRE & PETA)',
                                         'poller_hash': 'QTlDMzdE',
                                         'serial_number': 'A9C37D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6d:ec',
                                         'uptime': '20037947'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '165/g1',
                                                                       'name': '165/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTYxMDI3cGVlcmludGYxNjUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '165',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTYxMDI3cGVlcmludGZwb24gMTY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '165',
                                                                                               'links': [       {       'connected_time': '24669181',
                                                                                                                        'fiber_length': 18088,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.264,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTYxMDI3cGVlcnBvbnBvbiAxNjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '11 MADDISON ST E',
                                         'poller_hash': 'MTYxMDI3',
                                         'serial_number': '161027',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:b7:29:ee',
                                         'uptime': '24669181'},
                                 {       'counter_source': '166/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '166/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '235229225999',
                                                                       'in_ucast_pkts': '2211419928',
                                                                       'mac_address': '00:06:31:b6:8b:b4',
                                                                       'name': '166/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11867083853474',
                                                                       'out_ucast_pkts': '8220990337',
                                                                       'poller_hash': 'MTU1M0M2cGVlcmludGYxNjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '166',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11867083853474',
                                                                       'in_ucast_pkts': '8220990337',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11867083853474',
                                                                                                'in_ucast_pkts': '8220990337',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '235229225999',
                                                                                                'out_ucast_pkts': '2211419928',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '235229225999',
                                                                       'out_ucast_pkts': '2211419928',
                                                                       'poller_hash': 'MTU1M0M2cGVlcmludGZwb24gMTY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166',
                                                                                               'links': [       {       'connected_time': '47673',
                                                                                                                        'fiber_length': 18000,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.528,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1M0M2cGVlcnBvbnBvbiAxNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '75 Maddison St. W. Monkton (Megan '
                                                 'Mclean)',
                                         'poller_hash': 'MTU1M0M2',
                                         'serial_number': '1553C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:3a:0c',
                                         'uptime': '47673'},
                                 {       'counter_source': '167/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:a3:1f'],
                                                                       'id': '167/g1',
                                                                       'in_nucast_pkts': '52021',
                                                                       'in_octets': '23344697215',
                                                                       'in_ucast_pkts': '138998116',
                                                                       'mac_address': '00:06:31:b5:a8:d3',
                                                                       'mtu': 2000,
                                                                       'name': '167/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5062359',
                                                                       'out_octets': '505698862636',
                                                                       'out_ucast_pkts': '361813972',
                                                                       'poller_hash': 'MTU1NDYwcGVlcmludGYxNjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '167',
                                                                       'in_nucast_pkts': '5062359',
                                                                       'in_octets': '505698862636',
                                                                       'in_ucast_pkts': '361813972',
                                                                       'links': [       {       'in_nucast_pkts': '5062359',
                                                                                                'in_octets': '505698862636',
                                                                                                'in_ucast_pkts': '361813972',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '52021',
                                                                                                'out_octets': '23344697215',
                                                                                                'out_ucast_pkts': '138998116',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '52021',
                                                                       'out_octets': '23344697215',
                                                                       'out_ucast_pkts': '138998116',
                                                                       'poller_hash': 'MTU1NDYwcGVlcmludGZwb24gMTY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167',
                                                                                               'links': [       {       'connected_time': '2969248',
                                                                                                                        'fiber_length': 17601,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.832,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1NDYwcGVlcnBvbnBvbiAxNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '164 WINSTANLEY ST (CARAPINHA KYLA)',
                                         'poller_hash': 'MTU1NDYw',
                                         'serial_number': '155460',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:3a:a6',
                                         'uptime': '2969248'},
                                 {       'counter_source': '168/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:51:dc'],
                                                                       'id': '168/g1',
                                                                       'in_nucast_pkts': '495801',
                                                                       'in_octets': '841834746815',
                                                                       'in_ucast_pkts': '1543522898',
                                                                       'mac_address': 'ec:4f:82:5b:79:2e',
                                                                       'mtu': 2000,
                                                                       'name': '168/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62167719',
                                                                       'out_octets': '7804345623434',
                                                                       'out_ucast_pkts': '6034653188',
                                                                       'poller_hash': 'MkIwMjk4cGVlcmludGYxNjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '168',
                                                                       'in_nucast_pkts': '62167719',
                                                                       'in_octets': '7804345623434',
                                                                       'in_ucast_pkts': '6034653188',
                                                                       'links': [       {       'in_nucast_pkts': '62167719',
                                                                                                'in_octets': '7804345623434',
                                                                                                'in_ucast_pkts': '6034653188',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '495801',
                                                                                                'out_octets': '841834746815',
                                                                                                'out_ucast_pkts': '1543522898',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '495801',
                                                                       'out_octets': '841834746815',
                                                                       'out_ucast_pkts': '1543522898',
                                                                       'poller_hash': 'MkIwMjk4cGVlcmludGZwb24gMTY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168',
                                                                                               'links': [       {       'connected_time': '25995562',
                                                                                                                        'fiber_length': 17964,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjk4cGVlcnBvbnBvbiAxNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '169 SMITH ST (DOLSON ANDREA)',
                                         'poller_hash': 'MkIwMjk4',
                                         'serial_number': '2B0298',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:79:2b',
                                         'uptime': '25995562'},
                                 {       'counter_source': '169/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:39:10'],
                                                                       'id': '169/g1',
                                                                       'in_nucast_pkts': '693220',
                                                                       'in_octets': '285064104106',
                                                                       'in_ucast_pkts': '2243418535',
                                                                       'mac_address': '00:06:31:b6:8c:18',
                                                                       'mtu': 2000,
                                                                       'name': '169/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '71945945',
                                                                       'out_octets': '7889520996466',
                                                                       'out_ucast_pkts': '5566846451',
                                                                       'poller_hash': 'MTU1M0JBcGVlcmludGYxNjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '169',
                                                                       'in_nucast_pkts': '71945945',
                                                                       'in_octets': '7889520996466',
                                                                       'in_ucast_pkts': '5566846451',
                                                                       'links': [       {       'in_nucast_pkts': '71945945',
                                                                                                'in_octets': '7889520996466',
                                                                                                'in_ucast_pkts': '5566846451',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '693220',
                                                                                                'out_octets': '285064104106',
                                                                                                'out_ucast_pkts': '2243418535',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '693220',
                                                                       'out_octets': '285064104106',
                                                                       'out_ucast_pkts': '2243418535',
                                                                       'poller_hash': 'MTU1M0JBcGVlcmludGZwb24gMTY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '169',
                                                                                               'links': [       {       'connected_time': '69226272',
                                                                                                                        'fiber_length': 18015,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.872,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1M0JBcGVlcnBvbnBvbiAxNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '93 Maddison Ave. Monkton (Mike '
                                                 'Duncan 10432186)',
                                         'poller_hash': 'MTU1M0JB',
                                         'serial_number': '1553BA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:3a:00',
                                         'uptime': '69226272'},
                                 {       'counter_source': '170/g1+170/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '08:96:ad:ec:1f:2d',
                                                                                                        '08:96:ad:ec:1f:40'],
                                                                       'id': '170/g1',
                                                                       'in_nucast_pkts': '4513486',
                                                                       'in_octets': '2055235892125',
                                                                       'in_ucast_pkts': '7450266353',
                                                                       'mac_address': 'cc:be:59:6f:ca:26',
                                                                       'mtu': 2000,
                                                                       'name': '170/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3903933',
                                                                       'out_octets': '42287265327221',
                                                                       'out_ucast_pkts': '29001397281',
                                                                       'poller_hash': 'NTVGNTM3cGVlcmludGYxNzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:50:56:a3:97:c5',
                                                                                                        'f8:6b:d9:26:63:e1'],
                                                                       'id': '170/g2',
                                                                       'in_nucast_pkts': '2378',
                                                                       'in_octets': '3335213729514',
                                                                       'in_ucast_pkts': '3370361858',
                                                                       'mac_address': 'cc:be:59:6f:ca:27',
                                                                       'mtu': 2000,
                                                                       'name': '170/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '168036731',
                                                                       'out_octets': '3801872651098',
                                                                       'out_ucast_pkts': '4411474544',
                                                                       'poller_hash': 'NTVGNTM3cGVlcmludGYxNzAvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '170',
                                                                       'in_nucast_pkts': '171940664',
                                                                       'in_octets': '46089137978319',
                                                                       'in_ucast_pkts': '33412871825',
                                                                       'links': [       {       'in_nucast_pkts': '171940664',
                                                                                                'in_octets': '46089137978319',
                                                                                                'in_ucast_pkts': '33412871825',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4515864',
                                                                                                'out_octets': '5390449621639',
                                                                                                'out_ucast_pkts': '10820628211',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4515864',
                                                                       'out_octets': '5390449621639',
                                                                       'out_ucast_pkts': '10820628211',
                                                                       'poller_hash': 'NTVGNTM3cGVlcmludGZwb24gMTcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '170',
                                                                                               'links': [       {       'connected_time': '18082129',
                                                                                                                        'fiber_length': 18437,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.814,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTVGNTM3cGVlcnBvbnBvbiAxNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '132 Maddison St. W. Monkton (WFS '
                                                 'Great Lakes 10424552)',
                                         'poller_hash': 'NTVGNTM3',
                                         'serial_number': '55F537',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6f:ca:22',
                                         'uptime': '18082129'},
                                 {       'counter_source': '171/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '171/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '121219855624',
                                                                       'in_ucast_pkts': '538310119',
                                                                       'mac_address': 'ec:4f:82:08:29:15',
                                                                       'name': '171/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2034673089853',
                                                                       'out_ucast_pkts': '1623148562',
                                                                       'poller_hash': 'MUM1RDE4cGVlcmludGYxNzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '171',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2034673089853',
                                                                       'in_ucast_pkts': '1623148562',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2034673089853',
                                                                                                'in_ucast_pkts': '1623148562',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '121219855624',
                                                                                                'out_ucast_pkts': '538310119',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '121219855624',
                                                                       'out_ucast_pkts': '538310119',
                                                                       'poller_hash': 'MUM1RDE4cGVlcmludGZwb24gMTcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '171',
                                                                                               'links': [       {       'connected_time': '19619401',
                                                                                                                        'fiber_length': 18461,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.548,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUM1RDE4cGVlcnBvbnBvbiAxNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '73 SCHADE ST (SCHMIDT RUTH)',
                                         'poller_hash': 'MUM1RDE4',
                                         'serial_number': '1C5D18',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:08:29:15',
                                         'uptime': '19619401'},
                                 {       'counter_source': '172/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b6:f5'],
                                                                       'id': '172/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '6069429757',
                                                                       'in_ucast_pkts': '21835788',
                                                                       'mac_address': 'ec:4f:82:63:5c:c1',
                                                                       'mtu': 2000,
                                                                       'name': '172/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464254',
                                                                       'out_octets': '95047624277',
                                                                       'out_ucast_pkts': '81917393',
                                                                       'poller_hash': 'MkJENjlGcGVlcmludGYxNzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '172',
                                                                       'in_nucast_pkts': '1464254',
                                                                       'in_octets': '95047624277',
                                                                       'in_ucast_pkts': '81917393',
                                                                       'links': [       {       'in_nucast_pkts': '1464254',
                                                                                                'in_octets': '95047624277',
                                                                                                'in_ucast_pkts': '81917393',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '6069429757',
                                                                                                'out_ucast_pkts': '21835788',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '6069429757',
                                                                       'out_ucast_pkts': '21835788',
                                                                       'poller_hash': 'MkJENjlGcGVlcmludGZwb24gMTcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172',
                                                                                               'links': [       {       'connected_time': '688353',
                                                                                                                        'fiber_length': 975,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENjlGcGVlcnBvbnBvbiAxNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '7 West St. Milverton (Ian Clark '
                                                 '10434646)',
                                         'poller_hash': 'MkJENjlG',
                                         'serial_number': '2BD69F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:5c:c1',
                                         'uptime': '688353'},
                                 {       'counter_source': '173/g1+173/g2+173/g3+173/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:eb:33'],
                                                                       'id': '173/g1',
                                                                       'in_nucast_pkts': '1219',
                                                                       'in_octets': '6173085',
                                                                       'in_ucast_pkts': '28924',
                                                                       'mac_address': 'f8:85:f9:17:28:b3',
                                                                       'mtu': 2000,
                                                                       'name': '173/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '97400',
                                                                       'out_octets': '302702674',
                                                                       'out_ucast_pkts': '214557',
                                                                       'poller_hash': 'REZGREZGcGVlcmludGYxNzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:e8:52'],
                                                                       'id': '173/g2',
                                                                       'in_nucast_pkts': '1189',
                                                                       'in_octets': '5619713',
                                                                       'in_ucast_pkts': '19299',
                                                                       'mac_address': 'f8:85:f9:17:28:b4',
                                                                       'mtu': 2000,
                                                                       'name': '173/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '96194',
                                                                       'out_octets': '302437969',
                                                                       'out_ucast_pkts': '212000',
                                                                       'poller_hash': 'REZGREZGcGVlcmludGYxNzMvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:ee:a3'],
                                                                       'id': '173/g3',
                                                                       'in_nucast_pkts': '1261',
                                                                       'in_octets': '10475445',
                                                                       'in_ucast_pkts': '63768',
                                                                       'mac_address': 'f8:85:f9:17:28:b5',
                                                                       'mtu': 2000,
                                                                       'name': '173/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '94674',
                                                                       'out_octets': '305322242',
                                                                       'out_ucast_pkts': '252860',
                                                                       'poller_hash': 'REZGREZGcGVlcmludGYxNzMvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f1:63'],
                                                                       'id': '173/g4',
                                                                       'in_nucast_pkts': '1383',
                                                                       'in_octets': '6393929',
                                                                       'in_ucast_pkts': '35355',
                                                                       'mac_address': 'f8:85:f9:17:28:b6',
                                                                       'mtu': 2000,
                                                                       'name': '173/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '94679',
                                                                       'out_octets': '302501460',
                                                                       'out_ucast_pkts': '214907',
                                                                       'poller_hash': 'REZGREZGcGVlcmludGYxNzMvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '173',
                                                                       'in_nucast_pkts': '382947',
                                                                       'in_octets': '1212964345',
                                                                       'in_ucast_pkts': '894324',
                                                                       'links': [       {       'in_nucast_pkts': '382947',
                                                                                                'in_octets': '1212964345',
                                                                                                'in_ucast_pkts': '894324',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5052',
                                                                                                'out_octets': '28662172',
                                                                                                'out_ucast_pkts': '147346',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5052',
                                                                       'out_octets': '28662172',
                                                                       'out_ucast_pkts': '147346',
                                                                       'poller_hash': 'REZGREZGcGVlcmludGZwb24gMTcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '173',
                                                                                               'links': [       {       'connected_time': '688559',
                                                                                                                        'fiber_length': 242,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREZGcGVlcnBvbnBvbiAxNzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'Inventory (Room)',
                                         'poller_hash': 'REZGREZG',
                                         'serial_number': 'DFFDFF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:28:b0',
                                         'uptime': '688559'},
                                 {       'counter_source': '174/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f2:c1'],
                                                                       'id': '174/g1',
                                                                       'in_nucast_pkts': '15211807',
                                                                       'in_octets': '470760106692',
                                                                       'in_ucast_pkts': '1685196461',
                                                                       'mac_address': 'ec:4f:82:4f:78:e2',
                                                                       'mtu': 2000,
                                                                       'name': '174/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32688566825',
                                                                       'out_octets': '50598570120315',
                                                                       'out_ucast_pkts': '5258405083',
                                                                       'poller_hash': 'Mjk4MjFFcGVlcmludGYxNzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174',
                                                                       'in_nucast_pkts': '32688566825',
                                                                       'in_octets': '50598570120315',
                                                                       'in_ucast_pkts': '5258405083',
                                                                       'links': [       {       'in_nucast_pkts': '32688566825',
                                                                                                'in_octets': '50598570120315',
                                                                                                'in_ucast_pkts': '5258405083',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15211807',
                                                                                                'out_octets': '470760106692',
                                                                                                'out_ucast_pkts': '1685196461',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15211807',
                                                                       'out_octets': '470760106692',
                                                                       'out_ucast_pkts': '1685196461',
                                                                       'poller_hash': 'Mjk4MjFFcGVlcmludGZwb24gMTc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174',
                                                                                               'links': [       {       'connected_time': '21816379',
                                                                                                                        'fiber_length': 868,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjFFcGVlcnBvbnBvbiAxNzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '24 PACIFIC AV (SMITH CHRISTINA)',
                                         'poller_hash': 'Mjk4MjFF',
                                         'serial_number': '29821E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:78:df',
                                         'uptime': '21816379'},
                                 {       'counter_source': '176/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '7c:5a:1c:82:a3:57'],
                                                                       'id': '176/g1',
                                                                       'in_nucast_pkts': '228',
                                                                       'in_octets': '289317027879',
                                                                       'in_ucast_pkts': '932933381',
                                                                       'mac_address': '00:06:31:e3:fd:0f',
                                                                       'mtu': 2000,
                                                                       'name': '176/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '80889162',
                                                                       'out_octets': '2344412752603',
                                                                       'out_ucast_pkts': '1991617782',
                                                                       'poller_hash': 'MTk5QTNBcGVlcmludGYxNzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '176',
                                                                       'in_nucast_pkts': '80889162',
                                                                       'in_octets': '2344412752603',
                                                                       'in_ucast_pkts': '1991617782',
                                                                       'links': [       {       'in_nucast_pkts': '80889162',
                                                                                                'in_octets': '2344412752603',
                                                                                                'in_ucast_pkts': '1991617782',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '228',
                                                                                                'out_octets': '289317027879',
                                                                                                'out_ucast_pkts': '932933381',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '228',
                                                                       'out_octets': '289317027879',
                                                                       'out_ucast_pkts': '932933381',
                                                                       'poller_hash': 'MTk5QTNBcGVlcmludGZwb24gMTc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '176',
                                                                                               'links': [       {       'connected_time': '40631',
                                                                                                                        'fiber_length': 1490,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.488,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTk5QTNBcGVlcnBvbnBvbiAxNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '15 Davis St. Milverton (Gary '
                                                 'Boshart)',
                                         'poller_hash': 'MTk5QTNB',
                                         'serial_number': '199A3A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:e3:fd:0b',
                                         'uptime': '40631'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '177/g1',
                                                                       'name': '177/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTc1RUJBcGVlcmludGYxNzcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '177',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTc1RUJBcGVlcmludGZwb24gMTc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '177',
                                                                                               'links': [       {       'connected_time': '688490',
                                                                                                                        'fiber_length': 1006,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc1RUJBcGVlcnBvbnBvbiAxNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '1 Pacific Ave. Milverton',
                                         'poller_hash': 'MTc1RUJB',
                                         'serial_number': '175EBA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:59:27',
                                         'uptime': '688490'},
                                 {       'counter_source': '178/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '178/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '566961056299',
                                                                       'in_ucast_pkts': '1770494590',
                                                                       'mac_address': '00:06:31:dd:75:c7',
                                                                       'name': '178/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7280234940563',
                                                                       'out_ucast_pkts': '5492970134',
                                                                       'poller_hash': 'MTc2MjRFcGVlcmludGYxNzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '178',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7280234940563',
                                                                       'in_ucast_pkts': '5492970134',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7280234940563',
                                                                                                'in_ucast_pkts': '5492970134',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '566961056299',
                                                                                                'out_ucast_pkts': '1770494590',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '566961056299',
                                                                       'out_ucast_pkts': '1770494590',
                                                                       'poller_hash': 'MTc2MjRFcGVlcmludGZwb24gMTc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '178',
                                                                                               'links': [       {       'connected_time': '2846741',
                                                                                                                        'fiber_length': 852,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc2MjRFcGVlcnBvbnBvbiAxNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '16 PACIFIC AV (SHOULDICE DARRYL & '
                                                 'CHARIS)',
                                         'poller_hash': 'MTc2MjRF',
                                         'serial_number': '17624E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:75:c7',
                                         'uptime': '2846741'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '179/g1',
                                                                       'name': '179/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk4MkM4cGVlcmludGYxNzkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '179',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk4MkM4cGVlcmludGZwb24gMTc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '179',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MkM4cGVlcnBvbnBvbiAxNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '55 William St. Apt#102 Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "179" '
                                                                  '(location: GPON '
                                                                  'port "7/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:38.310Z',
                                         'poller_hash': 'Mjk4MkM4',
                                         'serial_number': '2982C8',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '180/F1+180/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a2:fd:72',
                                                                                                        '00:03:e6:db:23:56'],
                                                                       'id': '180/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '574221260',
                                                                       'in_ucast_pkts': '3028887',
                                                                       'name': '180/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '114903503',
                                                                       'out_octets': '100875468162',
                                                                       'out_ucast_pkts': '1888002',
                                                                       'poller_hash': 'Mjk4MUU0cGVlcmludGYxODAvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '180/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '57495058435',
                                                                       'in_ucast_pkts': '261020943',
                                                                       'mac_address': 'ec:4f:82:4f:77:0f',
                                                                       'name': '180/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1592879246932',
                                                                       'out_ucast_pkts': '1130809874',
                                                                       'poller_hash': 'Mjk4MUU0cGVlcmludGYxODAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180',
                                                                       'in_nucast_pkts': '114903503',
                                                                       'in_octets': '1693754715094',
                                                                       'in_ucast_pkts': '1132697876',
                                                                       'links': [       {       'in_nucast_pkts': '114903503',
                                                                                                'in_octets': '1693754715094',
                                                                                                'in_ucast_pkts': '1132697876',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '58069279695',
                                                                                                'out_ucast_pkts': '264049830',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '58069279695',
                                                                       'out_ucast_pkts': '264049830',
                                                                       'poller_hash': 'Mjk4MUU0cGVlcmludGZwb24gMTgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180',
                                                                                               'links': [       {       'connected_time': '688358',
                                                                                                                        'fiber_length': 1304,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MUU0cGVlcnBvbnBvbiAxODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 West St. Milverton (Brennan Erb)',
                                         'poller_hash': 'Mjk4MUU0',
                                         'serial_number': '2981E4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:77:0f',
                                         'uptime': '688358'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '181/g1',
                                                                       'name': '181/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk4MzAxcGVlcmludGYxODEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '181',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'Mjk4MzAxcGVlcmludGZwb24gMTgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '181',
                                                                                               'links': [       {       'connected_time': '21815991',
                                                                                                                        'fiber_length': 2020,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MzAxcGVlcnBvbnBvbiAxODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '26 Main St. N. Milverton (Innovapost '
                                                 'Inc 10431404)',
                                         'poller_hash': 'Mjk4MzAx',
                                         'serial_number': '298301',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7f:f7',
                                         'uptime': '21815991'},
                                 {       'counter_source': '182/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2b:48:69'],
                                                                       'id': '182/g1',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '12823604311',
                                                                       'in_ucast_pkts': '64016436',
                                                                       'mac_address': 'cc:be:59:cb:ad:55',
                                                                       'mtu': 2000,
                                                                       'name': '182/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720760',
                                                                       'out_octets': '412168046980',
                                                                       'out_ucast_pkts': '292932788',
                                                                       'poller_hash': 'NDJDNzAycGVlcmludGYxODIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '182',
                                                                       'in_nucast_pkts': '6720760',
                                                                       'in_octets': '412168046980',
                                                                       'in_ucast_pkts': '292932788',
                                                                       'links': [       {       'in_nucast_pkts': '6720760',
                                                                                                'in_octets': '412168046980',
                                                                                                'in_ucast_pkts': '292932788',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12',
                                                                                                'out_octets': '12823604311',
                                                                                                'out_ucast_pkts': '64016436',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '12823604311',
                                                                       'out_ucast_pkts': '64016436',
                                                                       'poller_hash': 'NDJDNzAycGVlcmludGZwb24gMTgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182',
                                                                                               'links': [       {       'connected_time': '3114031',
                                                                                                                        'fiber_length': 17415,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNzAycGVlcnBvbnBvbiAxODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '191 Selby Av. Monkton (Roger Orth)',
                                         'poller_hash': 'NDJDNzAy',
                                         'serial_number': '42C702',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ad:52',
                                         'uptime': '3114031'},
                                 {       'counter_source': '183/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '183/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '587',
                                                                       'in_octets': '292896487529',
                                                                       'in_ucast_pkts': '1196026428',
                                                                       'mac_address': '00:06:31:e9:01:87',
                                                                       'name': '183/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '3170',
                                                                       'out_octets': '5190992633329',
                                                                       'out_ucast_pkts': '3885698101',
                                                                       'poller_hash': 'MUQwNjg5cGVlcmludGYxODMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '183',
                                                                       'in_nucast_pkts': '3170',
                                                                       'in_octets': '5190992633329',
                                                                       'in_ucast_pkts': '3885698101',
                                                                       'links': [       {       'in_nucast_pkts': '3170',
                                                                                                'in_octets': '5190992633329',
                                                                                                'in_ucast_pkts': '3885698101',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '587',
                                                                                                'out_octets': '292896487529',
                                                                                                'out_ucast_pkts': '1196026428',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '587',
                                                                       'out_octets': '292896487529',
                                                                       'out_ucast_pkts': '1196026428',
                                                                       'poller_hash': 'MUQwNjg5cGVlcmludGZwb24gMTgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '183',
                                                                                               'links': [       {       'connected_time': '5882390',
                                                                                                                        'fiber_length': 1677,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjg5cGVlcnBvbnBvbiAxODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '55 William St. Apt # 105 (Amanda '
                                                 'Bush)',
                                         'poller_hash': 'MUQwNjg5',
                                         'serial_number': '1D0689',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:87',
                                         'uptime': '5882390'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '184/g1',
                                                                       'name': '184/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCM0EycGVlcmludGYxODQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '184',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCM0EycGVlcmludGZwb24gMTg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0EycGVlcnBvbnBvbiAxODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '55 William St Apt 4 Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "184" '
                                                                  '(location: GPON '
                                                                  'port "7/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:38.360Z',
                                         'poller_hash': 'NDNCM0Ey',
                                         'serial_number': '43B3A2',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '187/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '187/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '330442278135',
                                                                       'in_ucast_pkts': '48089720924',
                                                                       'mac_address': '00:06:31:e8:f3:02',
                                                                       'name': '187/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '128',
                                                                       'out_octets': '2142803471622',
                                                                       'out_ucast_pkts': '1819497319',
                                                                       'poller_hash': 'MUQwNEVDcGVlcmludGYxODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '187',
                                                                       'in_nucast_pkts': '128',
                                                                       'in_octets': '2142803471622',
                                                                       'in_ucast_pkts': '1819497319',
                                                                       'links': [       {       'in_nucast_pkts': '128',
                                                                                                'in_octets': '2142803471622',
                                                                                                'in_ucast_pkts': '1819497319',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '330442278135',
                                                                                                'out_ucast_pkts': '48089720924',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '330442278135',
                                                                       'out_ucast_pkts': '48089720924',
                                                                       'poller_hash': 'MUQwNEVDcGVlcmludGZwb24gMTg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '187',
                                                                                               'links': [       {       'connected_time': '688453',
                                                                                                                        'fiber_length': 1208,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEVDcGVlcnBvbnBvbiAxODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '22 Mill St E Milverton (Becky '
                                                 'Steckly10422799)',
                                         'poller_hash': 'MUQwNEVD',
                                         'serial_number': '1D04EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:02',
                                         'uptime': '688453'},
                                 {       'counter_source': '188/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e1:33'],
                                                                       'id': '188/g1',
                                                                       'in_nucast_pkts': '13063',
                                                                       'in_octets': '7486637576',
                                                                       'in_ucast_pkts': '26099618',
                                                                       'mac_address': 'ec:4f:82:4f:77:f2',
                                                                       'mtu': 2000,
                                                                       'name': '188/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464032',
                                                                       'out_octets': '82846547375',
                                                                       'out_ucast_pkts': '64669109',
                                                                       'poller_hash': 'Mjk4MjAwcGVlcmludGYxODgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '188',
                                                                       'in_nucast_pkts': '1464032',
                                                                       'in_octets': '82846547375',
                                                                       'in_ucast_pkts': '64669109',
                                                                       'links': [       {       'in_nucast_pkts': '1464032',
                                                                                                'in_octets': '82846547375',
                                                                                                'in_ucast_pkts': '64669109',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13063',
                                                                                                'out_octets': '7486637576',
                                                                                                'out_ucast_pkts': '26099618',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13063',
                                                                       'out_octets': '7486637576',
                                                                       'out_ucast_pkts': '26099618',
                                                                       'poller_hash': 'Mjk4MjAwcGVlcmludGZwb24gMTg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '188',
                                                                                               'links': [       {       'connected_time': '688505',
                                                                                                                        'fiber_length': 921,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjAwcGVlcnBvbnBvbiAxODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '18 Pacific Ave. Milverton (Allissa '
                                                 'Wettlaufer)',
                                         'poller_hash': 'Mjk4MjAw',
                                         'serial_number': '298200',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:77:ef',
                                         'uptime': '688505'},
                                 {       'counter_source': '190/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ef:86'],
                                                                       'id': '190/g1',
                                                                       'in_nucast_pkts': '4999',
                                                                       'in_octets': '3201098135',
                                                                       'in_ucast_pkts': '13470400',
                                                                       'mac_address': 'ec:4f:82:8d:fd:53',
                                                                       'mtu': 2000,
                                                                       'name': '190/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '497761',
                                                                       'out_octets': '56579642083',
                                                                       'out_ucast_pkts': '44159572',
                                                                       'poller_hash': 'MzIyN0YxcGVlcmludGYxOTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '190',
                                                                       'in_nucast_pkts': '497761',
                                                                       'in_octets': '56579642083',
                                                                       'in_ucast_pkts': '44159572',
                                                                       'links': [       {       'in_nucast_pkts': '497761',
                                                                                                'in_octets': '56579642083',
                                                                                                'in_ucast_pkts': '44159572',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4999',
                                                                                                'out_octets': '3201098135',
                                                                                                'out_ucast_pkts': '13470400',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4999',
                                                                       'out_octets': '3201098135',
                                                                       'out_ucast_pkts': '13470400',
                                                                       'poller_hash': 'MzIyN0YxcGVlcmludGZwb24gMTkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '190',
                                                                                               'links': [       {       'connected_time': '241694',
                                                                                                                        'fiber_length': 18595,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyN0YxcGVlcnBvbnBvbiAxOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '135 Brock St (MORRISON W & V)',
                                         'poller_hash': 'MzIyN0Yx',
                                         'serial_number': '3227F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fd:50',
                                         'uptime': '241694'},
                                 {       'counter_source': '191/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '191/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '475658995317',
                                                                       'in_ucast_pkts': '2447596523',
                                                                       'mac_address': 'ec:4f:82:4f:89:ff',
                                                                       'name': '191/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '138',
                                                                       'out_octets': '8400505288513',
                                                                       'out_ucast_pkts': '6018250786',
                                                                       'poller_hash': 'Mjk4NDQycGVlcmludGYxOTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '191',
                                                                       'in_nucast_pkts': '138',
                                                                       'in_octets': '8400505288513',
                                                                       'in_ucast_pkts': '6018250786',
                                                                       'links': [       {       'in_nucast_pkts': '138',
                                                                                                'in_octets': '8400505288513',
                                                                                                'in_ucast_pkts': '6018250786',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '475658995317',
                                                                                                'out_ucast_pkts': '2447596523',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '475658995317',
                                                                       'out_ucast_pkts': '2447596523',
                                                                       'poller_hash': 'Mjk4NDQycGVlcmludGZwb24gMTkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '191',
                                                                                               'links': [       {       'connected_time': '21816587',
                                                                                                                        'fiber_length': 17672,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4NDQycGVlcnBvbnBvbiAxOTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '166 Winstanley St. Monkton (Danielle '
                                                 'MacDonald)',
                                         'poller_hash': 'Mjk4NDQy',
                                         'serial_number': '298442',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:89:ff',
                                         'uptime': '21816587'},
                                 {       'counter_source': '192/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b8:4b'],
                                                                       'id': '192/g1',
                                                                       'in_nucast_pkts': '8864',
                                                                       'in_octets': '228833052864',
                                                                       'in_ucast_pkts': '1297857919',
                                                                       'mac_address': 'ec:4f:82:8d:dc:23',
                                                                       'mtu': 2000,
                                                                       'name': '192/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45832228',
                                                                       'out_octets': '5291943696178',
                                                                       'out_ucast_pkts': '3820761452',
                                                                       'poller_hash': 'MzIyM0NCcGVlcmludGYxOTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '192',
                                                                       'in_nucast_pkts': '45832228',
                                                                       'in_octets': '5291943696178',
                                                                       'in_ucast_pkts': '3820761452',
                                                                       'links': [       {       'in_nucast_pkts': '45832228',
                                                                                                'in_octets': '5291943696178',
                                                                                                'in_ucast_pkts': '3820761452',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8864',
                                                                                                'out_octets': '228833052864',
                                                                                                'out_ucast_pkts': '1297857919',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8864',
                                                                       'out_octets': '228833052864',
                                                                       'out_ucast_pkts': '1297857919',
                                                                       'poller_hash': 'MzIyM0NCcGVlcmludGZwb24gMTky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '192',
                                                                                               'links': [       {       'connected_time': '19868105',
                                                                                                                        'fiber_length': 1319,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyM0NCcGVlcnBvbnBvbiAxOTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 PIERSON ST (KUEPFER JEFF & '
                                                 'SANDRA)',
                                         'poller_hash': 'MzIyM0NC',
                                         'serial_number': '3223CB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:dc:20',
                                         'uptime': '19868105'},
                                 {       'counter_source': '193/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:05:aa'],
                                                                       'id': '193/g1',
                                                                       'in_nucast_pkts': '10715',
                                                                       'in_octets': '288649774850',
                                                                       'in_ucast_pkts': '3035541448',
                                                                       'mac_address': '00:06:31:e9:01:30',
                                                                       'mtu': 2000,
                                                                       'name': '193/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '145122444',
                                                                       'out_octets': '16871802604962',
                                                                       'out_ucast_pkts': '12032903056',
                                                                       'poller_hash': 'MUQwNjdGcGVlcmludGYxOTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '193',
                                                                       'in_nucast_pkts': '145122444',
                                                                       'in_octets': '16871802604962',
                                                                       'in_ucast_pkts': '12032903056',
                                                                       'links': [       {       'in_nucast_pkts': '145122444',
                                                                                                'in_octets': '16871802604962',
                                                                                                'in_ucast_pkts': '12032903056',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10715',
                                                                                                'out_octets': '288649774850',
                                                                                                'out_ucast_pkts': '3035541448',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10715',
                                                                       'out_octets': '288649774850',
                                                                       'out_ucast_pkts': '3035541448',
                                                                       'poller_hash': 'MUQwNjdGcGVlcmludGZwb24gMTkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193',
                                                                                               'links': [       {       'connected_time': '21817477',
                                                                                                                        'fiber_length': 17554,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjdGcGVlcnBvbnBvbiAxOTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '201 Winstanley St. Monkton (Neil '
                                                 'Ward)',
                                         'poller_hash': 'MUQwNjdG',
                                         'serial_number': '1D067F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:2d',
                                         'uptime': '21817477'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '194/g1',
                                                                       'name': '194/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkJENUNBcGVlcmludGYxOTQvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '194',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkJENUNBcGVlcmludGZwb24gMTk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '194',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENUNBcGVlcnBvbnBvbiAxOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '171 Winstanley St. Monkton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "194" '
                                                                  '(location: GPON '
                                                                  'port "11/1"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-24T08:33:18.030Z',
                                         'poller_hash': 'MkJENUNB',
                                         'serial_number': '2BD5CA',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '195/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5f:c8'],
                                                                       'id': '195/g1',
                                                                       'in_nucast_pkts': '14942',
                                                                       'in_octets': '4652965064',
                                                                       'in_ucast_pkts': '30343244',
                                                                       'mac_address': 'ec:4f:82:8d:fe:bb',
                                                                       'mtu': 2000,
                                                                       'name': '195/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447029',
                                                                       'out_octets': '141417107829',
                                                                       'out_ucast_pkts': '102103553',
                                                                       'poller_hash': 'MzIyODFFcGVlcmludGYxOTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '195',
                                                                       'in_nucast_pkts': '1447029',
                                                                       'in_octets': '141417107829',
                                                                       'in_ucast_pkts': '102103553',
                                                                       'links': [       {       'in_nucast_pkts': '1447029',
                                                                                                'in_octets': '141417107829',
                                                                                                'in_ucast_pkts': '102103553',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14942',
                                                                                                'out_octets': '4652965064',
                                                                                                'out_ucast_pkts': '30343244',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14942',
                                                                       'out_octets': '4652965064',
                                                                       'out_ucast_pkts': '30343244',
                                                                       'poller_hash': 'MzIyODFFcGVlcmludGZwb24gMTk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '195',
                                                                                               'links': [       {       'connected_time': '688408',
                                                                                                                        'fiber_length': 2107,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODFFcGVlcnBvbnBvbiAxOTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '86 GERBER DR (NAFZIGER MIKE)',
                                         'poller_hash': 'MzIyODFF',
                                         'serial_number': '32281E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fe:b8',
                                         'uptime': '688408'},
                                 {       'counter_source': '196/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:58:9b'],
                                                                       'id': '196/g1',
                                                                       'in_nucast_pkts': '133466',
                                                                       'in_octets': '43220032043',
                                                                       'in_ucast_pkts': '384479772',
                                                                       'mac_address': 'ec:4f:82:4f:78:82',
                                                                       'mtu': 2000,
                                                                       'name': '196/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14849776',
                                                                       'out_octets': '2486109633593',
                                                                       'out_ucast_pkts': '1671154750',
                                                                       'poller_hash': 'Mjk4MjEycGVlcmludGYxOTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '196',
                                                                       'in_nucast_pkts': '14849776',
                                                                       'in_octets': '2486109633593',
                                                                       'in_ucast_pkts': '1671154750',
                                                                       'links': [       {       'in_nucast_pkts': '14849776',
                                                                                                'in_octets': '2486109633593',
                                                                                                'in_ucast_pkts': '1671154750',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '133466',
                                                                                                'out_octets': '43220032043',
                                                                                                'out_ucast_pkts': '384479772',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '133466',
                                                                       'out_octets': '43220032043',
                                                                       'out_ucast_pkts': '384479772',
                                                                       'poller_hash': 'Mjk4MjEycGVlcmludGZwb24gMTk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '196',
                                                                                               'links': [       {       'connected_time': '6403842',
                                                                                                                        'fiber_length': 1158,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjEycGVlcnBvbnBvbiAxOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6 PACIFIC AV (ROULSTON WAYNE)',
                                         'poller_hash': 'Mjk4MjEy',
                                         'serial_number': '298212',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:78:7f',
                                         'uptime': '6403842'},
                                 {       'counter_source': '197/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:f0:4f'],
                                                                       'id': '197/g1',
                                                                       'in_nucast_pkts': '19854994',
                                                                       'in_octets': '322339857679',
                                                                       'in_ucast_pkts': '2594330963',
                                                                       'mac_address': 'ec:4f:82:4f:93:a2',
                                                                       'mtu': 2000,
                                                                       'name': '197/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27275911027',
                                                                       'out_octets': '50997162115133',
                                                                       'out_ucast_pkts': '10060836497',
                                                                       'poller_hash': 'Mjk4NTc2cGVlcmludGYxOTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '197',
                                                                       'in_nucast_pkts': '27275911027',
                                                                       'in_octets': '50997162115133',
                                                                       'in_ucast_pkts': '10060836497',
                                                                       'links': [       {       'in_nucast_pkts': '27275911027',
                                                                                                'in_octets': '50997162115133',
                                                                                                'in_ucast_pkts': '10060836497',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19854994',
                                                                                                'out_octets': '322339857679',
                                                                                                'out_ucast_pkts': '2594330963',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19854994',
                                                                       'out_octets': '322339857679',
                                                                       'out_ucast_pkts': '2594330963',
                                                                       'poller_hash': 'Mjk4NTc2cGVlcmludGZwb24gMTk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '197',
                                                                                               'links': [       {       'connected_time': '65075966',
                                                                                                                        'fiber_length': 1018,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4NTc2cGVlcnBvbnBvbiAxOTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 PACIFIC AV (EISENHOFER MARJ)',
                                         'poller_hash': 'Mjk4NTc2',
                                         'serial_number': '298576',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:93:9f',
                                         'uptime': '65075966'},
                                 {       'counter_source': '198/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a5:da'],
                                                                       'id': '198/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '6190098295',
                                                                       'in_ucast_pkts': '46745118',
                                                                       'mac_address': 'ec:4f:82:63:5c:a9',
                                                                       'mtu': 2000,
                                                                       'name': '198/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464114',
                                                                       'out_octets': '221669410778',
                                                                       'out_ucast_pkts': '156648617',
                                                                       'poller_hash': 'MkJENjkzcGVlcmludGYxOTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '198',
                                                                       'in_nucast_pkts': '1464114',
                                                                       'in_octets': '221669410778',
                                                                       'in_ucast_pkts': '156648617',
                                                                       'links': [       {       'in_nucast_pkts': '1464114',
                                                                                                'in_octets': '221669410778',
                                                                                                'in_ucast_pkts': '156648617',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '6190098295',
                                                                                                'out_ucast_pkts': '46745118',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '6190098295',
                                                                       'out_ucast_pkts': '46745118',
                                                                       'poller_hash': 'MkJENjkzcGVlcmludGZwb24gMTk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198',
                                                                                               'links': [       {       'connected_time': '688479',
                                                                                                                        'fiber_length': 1011,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENjkzcGVlcnBvbnBvbiAxOTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '19 Pacific Av. Milverton (Wade '
                                                 'Goodwin 10426524)',
                                         'poller_hash': 'MkJENjkz',
                                         'serial_number': '2BD693',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:5c:a9',
                                         'uptime': '688479'},
                                 {       'counter_source': '199/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '199/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '889992295589',
                                                                       'in_ucast_pkts': '98046241339',
                                                                       'mac_address': '00:06:31:e9:01:7e',
                                                                       'name': '199/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '224',
                                                                       'out_octets': '12862959859165',
                                                                       'out_ucast_pkts': '9605761117',
                                                                       'poller_hash': 'MUQwNjg4cGVlcmludGYxOTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '199',
                                                                       'in_nucast_pkts': '224',
                                                                       'in_octets': '12862959859165',
                                                                       'in_ucast_pkts': '9605761117',
                                                                       'links': [       {       'in_nucast_pkts': '224',
                                                                                                'in_octets': '12862959859165',
                                                                                                'in_ucast_pkts': '9605761117',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '889992295589',
                                                                                                'out_ucast_pkts': '98046241339',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '889992295589',
                                                                       'out_ucast_pkts': '98046241339',
                                                                       'poller_hash': 'MUQwNjg4cGVlcmludGZwb24gMTk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '199',
                                                                                               'links': [       {       'connected_time': '17629509',
                                                                                                                        'fiber_length': 871,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjg4cGVlcnBvbnBvbiAxOTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '22 Pacific Ave. Milverton (Kevin '
                                                 'Rich)',
                                         'poller_hash': 'MUQwNjg4',
                                         'serial_number': '1D0688',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:7e',
                                         'uptime': '17629509'},
                                 {       'counter_source': '200/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '245344089298',
                                                                       'in_ucast_pkts': '1992094469',
                                                                       'mac_address': '00:06:31:e9:01:a2',
                                                                       'name': '200/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '266',
                                                                       'out_octets': '12033515352565',
                                                                       'out_ucast_pkts': '8283567229',
                                                                       'poller_hash': 'MUQwNjhDcGVlcmludGYyMDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200',
                                                                       'in_nucast_pkts': '266',
                                                                       'in_octets': '12033515352565',
                                                                       'in_ucast_pkts': '8283567229',
                                                                       'links': [       {       'in_nucast_pkts': '266',
                                                                                                'in_octets': '12033515352565',
                                                                                                'in_ucast_pkts': '8283567229',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '245344089298',
                                                                                                'out_ucast_pkts': '1992094469',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '245344089298',
                                                                       'out_ucast_pkts': '1992094469',
                                                                       'poller_hash': 'MUQwNjhDcGVlcmludGZwb24gMjAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200',
                                                                                               'links': [       {       'connected_time': '63688893',
                                                                                                                        'fiber_length': 1083,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjhDcGVlcnBvbnBvbiAyMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '22 Pacific Ave. Milverton (Robert '
                                                 'Paton)',
                                         'poller_hash': 'MUQwNjhD',
                                         'serial_number': '1D068C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:a2',
                                         'uptime': '63688893'},
                                 {       'counter_source': '201/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:3e:98:0e'],
                                                                       'id': '201/g1',
                                                                       'in_nucast_pkts': '21663',
                                                                       'in_octets': '213685644321',
                                                                       'in_ucast_pkts': '940607267',
                                                                       'mac_address': 'ec:4f:82:0d:69:31',
                                                                       'mtu': 2000,
                                                                       'name': '201/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '50208786',
                                                                       'out_octets': '7797648554240',
                                                                       'out_ucast_pkts': '5405007093',
                                                                       'poller_hash': 'MjA1REQxcGVlcmludGYyMDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '201',
                                                                       'in_nucast_pkts': '50208786',
                                                                       'in_octets': '7797648554240',
                                                                       'in_ucast_pkts': '5405007093',
                                                                       'links': [       {       'in_nucast_pkts': '50208786',
                                                                                                'in_octets': '7797648554240',
                                                                                                'in_ucast_pkts': '5405007093',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21663',
                                                                                                'out_octets': '213685644321',
                                                                                                'out_ucast_pkts': '940607267',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21663',
                                                                       'out_octets': '213685644321',
                                                                       'out_ucast_pkts': '940607267',
                                                                       'poller_hash': 'MjA1REQxcGVlcmludGZwb24gMjAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201',
                                                                                               'links': [       {       'connected_time': '21763000',
                                                                                                                        'fiber_length': 903,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REQxcGVlcnBvbnBvbiAyMDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '20 Pacific Ave. Milverton (Debbie '
                                                 'Timmermans10401639)',
                                         'poller_hash': 'MjA1REQx',
                                         'serial_number': '205DD1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:2e',
                                         'uptime': '21763000'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '202/g1',
                                                                       'name': '202/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk4MUU2cGVlcmludGYyMDIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '202',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk4MUU2cGVlcmludGZwb24gMjAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MUU2cGVlcnBvbnBvbiAyMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '3 PACIFIC AV',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "202" '
                                                                  '(location: GPON '
                                                                  'port "8/6"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:10:31.660Z',
                                         'poller_hash': 'Mjk4MUU2',
                                         'serial_number': '2981E6',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '203/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '203/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '461477668147',
                                                                       'in_ucast_pkts': '1370927236',
                                                                       'mac_address': 'ec:4f:82:4f:82:4f',
                                                                       'name': '203/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '3091078074513',
                                                                       'out_ucast_pkts': '2608557638',
                                                                       'poller_hash': 'Mjk4MzRDcGVlcmludGYyMDMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '203',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '3091078074513',
                                                                       'in_ucast_pkts': '2608557638',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '3091078074513',
                                                                                                'in_ucast_pkts': '2608557638',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '461477668147',
                                                                                                'out_ucast_pkts': '1370927236',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '461477668147',
                                                                       'out_ucast_pkts': '1370927236',
                                                                       'poller_hash': 'Mjk4MzRDcGVlcmludGZwb24gMjAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '203',
                                                                                               'links': [       {       'connected_time': '10799232',
                                                                                                                        'fiber_length': 1154,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MzRDcGVlcnBvbnBvbiAyMDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '33 Pacific Ave. Milverton (Jeffrey '
                                                 'Plager 10434465)',
                                         'poller_hash': 'Mjk4MzRD',
                                         'serial_number': '29834C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:82:4f',
                                         'uptime': '10799232'},
                                 {       'counter_source': '204/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d2:a1'],
                                                                       'id': '204/g1',
                                                                       'in_nucast_pkts': '14728',
                                                                       'in_octets': '6870349557',
                                                                       'in_ucast_pkts': '52466482',
                                                                       'mac_address': 'd0:76:8f:aa:54:47',
                                                                       'mtu': 2000,
                                                                       'name': '204/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464137',
                                                                       'out_octets': '186854302522',
                                                                       'out_ucast_pkts': '136376174',
                                                                       'poller_hash': 'OTU2RkFFcGVlcmludGYyMDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '204',
                                                                       'in_nucast_pkts': '1464137',
                                                                       'in_octets': '186854302522',
                                                                       'in_ucast_pkts': '136376174',
                                                                       'links': [       {       'in_nucast_pkts': '1464137',
                                                                                                'in_octets': '186854302522',
                                                                                                'in_ucast_pkts': '136376174',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14728',
                                                                                                'out_octets': '6870349557',
                                                                                                'out_ucast_pkts': '52466482',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14728',
                                                                       'out_octets': '6870349557',
                                                                       'out_ucast_pkts': '52466482',
                                                                       'poller_hash': 'OTU2RkFFcGVlcmludGZwb24gMjA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '204',
                                                                                               'links': [       {       'connected_time': '688343',
                                                                                                                        'fiber_length': 955,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RkFFcGVlcnBvbnBvbiAyMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 TEMPERANCE ST (RICHARDS MARTY)',
                                         'poller_hash': 'OTU2RkFF',
                                         'serial_number': '956FAE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:54:44',
                                         'uptime': '688343'},
                                 {       'counter_source': '205/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8a:e0'],
                                                                       'id': '205/g1',
                                                                       'in_nucast_pkts': '1751703',
                                                                       'in_octets': '1806937815605',
                                                                       'in_ucast_pkts': '7731210128',
                                                                       'mac_address': '00:06:31:e9:04:90',
                                                                       'mtu': 2000,
                                                                       'name': '205/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '144919536',
                                                                       'out_octets': '27427778164623',
                                                                       'out_ucast_pkts': '20711249833',
                                                                       'poller_hash': 'MUQwNkUxcGVlcmludGYyMDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205',
                                                                       'in_nucast_pkts': '144919536',
                                                                       'in_octets': '27427778164623',
                                                                       'in_ucast_pkts': '20711249833',
                                                                       'links': [       {       'in_nucast_pkts': '144919536',
                                                                                                'in_octets': '27427778164623',
                                                                                                'in_ucast_pkts': '20711249833',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1751703',
                                                                                                'out_octets': '1806937815605',
                                                                                                'out_ucast_pkts': '7731210128',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1751703',
                                                                       'out_octets': '1806937815605',
                                                                       'out_ucast_pkts': '7731210128',
                                                                       'poller_hash': 'MUQwNkUxcGVlcmludGZwb24gMjA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205',
                                                                                               'links': [       {       'connected_time': '21817189',
                                                                                                                        'fiber_length': 17696,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNkUxcGVlcnBvbnBvbiAyMDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '217 Winstanley St. Monkton (Mike '
                                                 'Chambers)',
                                         'poller_hash': 'MUQwNkUx',
                                         'serial_number': '1D06E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:04:8d',
                                         'uptime': '21817189'},
                                 {       'counter_source': '207/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '207/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3378910712772',
                                                                       'in_ucast_pkts': '207147199751',
                                                                       'mac_address': 'ec:4f:82:5b:70:bb',
                                                                       'name': '207/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '266',
                                                                       'out_octets': '17188440975244',
                                                                       'out_ucast_pkts': '14904136612',
                                                                       'poller_hash': 'MkIwMThBcGVlcmludGYyMDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '207',
                                                                       'in_nucast_pkts': '266',
                                                                       'in_octets': '17188440975244',
                                                                       'in_ucast_pkts': '14904136612',
                                                                       'links': [       {       'in_nucast_pkts': '266',
                                                                                                'in_octets': '17188440975244',
                                                                                                'in_ucast_pkts': '14904136612',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3378910712772',
                                                                                                'out_ucast_pkts': '207147199751',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3378910712772',
                                                                       'out_ucast_pkts': '207147199751',
                                                                       'poller_hash': 'MkIwMThBcGVlcmludGZwb24gMjA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207',
                                                                                               'links': [       {       'connected_time': '688498',
                                                                                                                        'fiber_length': 830,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMThBcGVlcnBvbnBvbiAyMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '28 Pacific Ave. Milverton (Jacob '
                                                 'Harman)',
                                         'poller_hash': 'MkIwMThB',
                                         'serial_number': '2B018A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:bb',
                                         'uptime': '688498'},
                                 {       'counter_source': '209/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c0:06:c3:1c:9b:46'],
                                                                       'id': '209/g1',
                                                                       'in_nucast_pkts': '50665',
                                                                       'in_octets': '385176830089',
                                                                       'in_ucast_pkts': '1680662745',
                                                                       'mac_address': 'ec:4f:82:b9:1a:dc',
                                                                       'mtu': 2000,
                                                                       'name': '209/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2254656',
                                                                       'out_octets': '4280997515986',
                                                                       'out_ucast_pkts': '3175673843',
                                                                       'poller_hash': 'MkZFQTA1cGVlcmludGYyMDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '209',
                                                                       'in_nucast_pkts': '2254656',
                                                                       'in_octets': '4280997515986',
                                                                       'in_ucast_pkts': '3175673843',
                                                                       'links': [       {       'in_nucast_pkts': '2254656',
                                                                                                'in_octets': '4280997515986',
                                                                                                'in_ucast_pkts': '3175673843',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '50665',
                                                                                                'out_octets': '385176830089',
                                                                                                'out_ucast_pkts': '1680662745',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '50665',
                                                                       'out_octets': '385176830089',
                                                                       'out_ucast_pkts': '1680662745',
                                                                       'poller_hash': 'MkZFQTA1cGVlcmludGZwb24gMjA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '209',
                                                                                               'links': [       {       'connected_time': '1363457',
                                                                                                                        'fiber_length': 18174,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.118,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQTA1cGVlcnBvbnBvbiAyMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '16 SCHADE ST (MAYER MICHAEL)',
                                         'poller_hash': 'MkZFQTA1',
                                         'serial_number': '2FEA05',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:1a:d8',
                                         'uptime': '1363457'},
                                 {       'counter_source': '210/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '210/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '605784893751',
                                                                       'in_ucast_pkts': '2212673016',
                                                                       'mac_address': 'cc:be:59:6f:ee:8a',
                                                                       'name': '210/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4718465853771',
                                                                       'out_ucast_pkts': '5438864360',
                                                                       'poller_hash': 'NTVGOUM0cGVlcmludGYyMTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '210',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4718465853771',
                                                                       'in_ucast_pkts': '5438864360',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4718465853771',
                                                                                                'in_ucast_pkts': '5438864360',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '605784893751',
                                                                                                'out_ucast_pkts': '2212673016',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '605784893751',
                                                                       'out_ucast_pkts': '2212673016',
                                                                       'poller_hash': 'NTVGOUM0cGVlcmludGZwb24gMjEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '210',
                                                                                               'links': [       {       'connected_time': '47455',
                                                                                                                        'fiber_length': 17506,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.262,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTVGOUM0cGVlcnBvbnBvbiAyMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '166 Fishleigh St. Monkton (Keith '
                                                 'Kuepfer 10403450)',
                                         'poller_hash': 'NTVGOUM0',
                                         'serial_number': '55F9C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6f:ee:8a',
                                         'uptime': '47455'},
                                 {       'counter_source': '211/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '211/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1234407830421',
                                                                       'in_ucast_pkts': '7061070794',
                                                                       'mac_address': 'ec:4f:82:92:fd:48',
                                                                       'name': '211/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25850242016602',
                                                                       'out_ucast_pkts': '18890758956',
                                                                       'poller_hash': 'MzI5NjM1cGVlcmludGYyMTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '211',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25850242016602',
                                                                       'in_ucast_pkts': '18890758956',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '25850242016602',
                                                                                                'in_ucast_pkts': '18890758956',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1234407830421',
                                                                                                'out_ucast_pkts': '7061070794',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1234407830421',
                                                                       'out_ucast_pkts': '7061070794',
                                                                       'poller_hash': 'MzI5NjM1cGVlcmludGZwb24gMjEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '211',
                                                                                               'links': [       {       'connected_time': '23325508',
                                                                                                                        'fiber_length': 17700,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.13,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzI5NjM1cGVlcnBvbnBvbiAyMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '150 Winstanley St. Monkton (Andre '
                                                 'Shtovba 10431113)',
                                         'poller_hash': 'MzI5NjM1',
                                         'serial_number': '329635',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:92:fd:48',
                                         'uptime': '23325508'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '212/g1',
                                                                       'name': '212/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTk5OUE3cGVlcmludGYyMTIvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '212',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTk5OUE3cGVlcmludGZwb24gMjEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTk5OUE3cGVlcnBvbnBvbiAyMTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '215 Nelson St. Monkton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "212" '
                                                                  '(location: GPON '
                                                                  'port "11/2"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-24T08:30:37.760Z',
                                         'poller_hash': 'MTk5OUE3',
                                         'serial_number': '1999A7',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '213/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '213/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '130301967802',
                                                                       'in_ucast_pkts': '1496225785',
                                                                       'mac_address': 'ec:4f:82:da:b0:48',
                                                                       'name': '213/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4439813772856',
                                                                       'out_ucast_pkts': '4578328053',
                                                                       'poller_hash': 'MkU5MUE4cGVlcmludGYyMTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '213',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4439813772856',
                                                                       'in_ucast_pkts': '4578328053',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4439813772856',
                                                                                                'in_ucast_pkts': '4578328053',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '130301967802',
                                                                                                'out_ucast_pkts': '1496225785',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '130301967802',
                                                                       'out_ucast_pkts': '1496225785',
                                                                       'poller_hash': 'MkU5MUE4cGVlcmludGZwb24gMjEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213',
                                                                                               'links': [       {       'connected_time': '47804',
                                                                                                                        'fiber_length': 17407,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkU5MUE4cGVlcnBvbnBvbiAyMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '187 Erskine St. Monkton (Robert '
                                                 'Doell 10432296)',
                                         'poller_hash': 'MkU5MUE4',
                                         'serial_number': '2E91A8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:da:b0:48',
                                         'uptime': '47804'},
                                 {       'counter_source': '216/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '216/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '776308712761',
                                                                       'in_ucast_pkts': '2059048154',
                                                                       'mac_address': 'ec:4f:82:bf:56:f0',
                                                                       'name': '216/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4725623113103',
                                                                       'out_ucast_pkts': '461531967',
                                                                       'poller_hash': 'MzYyODExcGVlcmludGYyMTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '216',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4725623113103',
                                                                       'in_ucast_pkts': '461531967',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4725623113103',
                                                                                                'in_ucast_pkts': '461531967',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '776308712761',
                                                                                                'out_ucast_pkts': '2059048154',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '776308712761',
                                                                       'out_ucast_pkts': '2059048154',
                                                                       'poller_hash': 'MzYyODExcGVlcmludGZwb24gMjE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216',
                                                                                               'links': [       {       'connected_time': '69225897',
                                                                                                                        'fiber_length': 18008,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.332,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzYyODExcGVlcnBvbnBvbiAyMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '53 MADDISON ST E (KUEPFER BRYCE)',
                                         'poller_hash': 'MzYyODEx',
                                         'serial_number': '362811',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:bf:56:f0',
                                         'uptime': '69225897'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '217/g1',
                                                                       'name': '217/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEQ0MjNGcGVlcmludGYyMTcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '217',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NEQ0MjNGcGVlcmludGZwb24gMjE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217',
                                                                                               'links': [       {       'connected_time': '47971',
                                                                                                                        'fiber_length': 17409,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.804,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjNGcGVlcnBvbnBvbiAyMTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '189 WINSTANLEY ST',
                                         'poller_hash': 'NEQ0MjNG',
                                         'serial_number': '4D423F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:66',
                                         'uptime': '47971'},
                                 {       'counter_source': '218/F1+218/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:17:6b:86'],
                                                                       'id': '218/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1186516945',
                                                                       'in_ucast_pkts': '6780160',
                                                                       'name': '218/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '4998847562',
                                                                       'out_octets': '6651422647776',
                                                                       'out_ucast_pkts': '3875161',
                                                                       'poller_hash': 'REZGQjU5cGVlcmludGYyMTgvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3c:e3'],
                                                                       'id': '218/g1',
                                                                       'in_nucast_pkts': '14701',
                                                                       'in_octets': '4640743833',
                                                                       'in_ucast_pkts': '17410837',
                                                                       'mac_address': 'f8:85:f9:17:13:83',
                                                                       'mtu': 2000,
                                                                       'name': '218/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463780',
                                                                       'out_octets': '58708930328',
                                                                       'out_ucast_pkts': '51513673',
                                                                       'poller_hash': 'REZGQjU5cGVlcmludGYyMTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '218',
                                                                       'in_nucast_pkts': '5000311342',
                                                                       'in_octets': '6710131578104',
                                                                       'in_ucast_pkts': '55388834',
                                                                       'links': [       {       'in_nucast_pkts': '5000311342',
                                                                                                'in_octets': '6710131578104',
                                                                                                'in_ucast_pkts': '55388834',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14701',
                                                                                                'out_octets': '5827260778',
                                                                                                'out_ucast_pkts': '24190997',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14701',
                                                                       'out_octets': '5827260778',
                                                                       'out_ucast_pkts': '24190997',
                                                                       'poller_hash': 'REZGQjU5cGVlcmludGZwb24gMjE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218',
                                                                                               'links': [       {       'connected_time': '688452',
                                                                                                                        'fiber_length': 1533,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGQjU5cGVlcnBvbnBvbiAyMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 MACBETH ST, Milverton (SCHMIDT '
                                                 'JOSH)',
                                         'poller_hash': 'REZGQjU5',
                                         'serial_number': 'DFFB59',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:13:80',
                                         'uptime': '688452'},
                                 {       'counter_source': '219/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '219/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '290303410117',
                                                                       'in_ucast_pkts': '1174773463',
                                                                       'mac_address': '48:77:46:a3:b3:fc',
                                                                       'name': '219/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2797280335891',
                                                                       'out_ucast_pkts': '3256154412',
                                                                       'poller_hash': 'NzFEN0FFcGVlcmludGYyMTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '219',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2797280335891',
                                                                       'in_ucast_pkts': '3256154412',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2797280335891',
                                                                                                'in_ucast_pkts': '3256154412',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '290303410117',
                                                                                                'out_ucast_pkts': '1174773463',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '290303410117',
                                                                       'out_ucast_pkts': '1174773463',
                                                                       'poller_hash': 'NzFEN0FFcGVlcmludGZwb24gMjE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '219',
                                                                                               'links': [       {       'connected_time': '24653006',
                                                                                                                        'fiber_length': 17759,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.672,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFEN0FFcGVlcnBvbnBvbiAyMTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '159 Market St. Monkton (Delford '
                                                 'Kuepfer)',
                                         'poller_hash': 'NzFEN0FF',
                                         'serial_number': '71D7AE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a3:b3:fc',
                                         'uptime': '24653006'},
                                 {       'counter_source': '220/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '220/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '300587837994',
                                                                       'in_ucast_pkts': '2279709844',
                                                                       'mac_address': '00:06:31:5e:ce:65',
                                                                       'name': '220/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7003510868921',
                                                                       'out_ucast_pkts': '6603251605',
                                                                       'poller_hash': 'QTlGMDhwZWVyaW50ZjIyMC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '220',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7003510868921',
                                                                       'in_ucast_pkts': '6603251605',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7003510868921',
                                                                                                'in_ucast_pkts': '6603251605',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '300587837994',
                                                                                                'out_ucast_pkts': '2279709844',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '300587837994',
                                                                       'out_ucast_pkts': '2279709844',
                                                                       'poller_hash': 'QTlGMDhwZWVyaW50ZnBvbiAyMjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '220',
                                                                                               'links': [       {       'connected_time': '47486',
                                                                                                                        'fiber_length': 17522,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlGMDhwZWVycG9ucG9uIDIyMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '167 Fishleigh St. Monkton (Mike '
                                                 'Danieli)',
                                         'poller_hash': 'QTlGMDg=',
                                         'serial_number': 'A9F08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:57:af:c8',
                                         'uptime': '47486'},
                                 {       'counter_source': '221/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '221/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1144637806514',
                                                                       'in_ucast_pkts': '6469031392',
                                                                       'mac_address': 'ec:4f:82:dc:ec:a4',
                                                                       'name': '221/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '29035185803645',
                                                                       'out_ucast_pkts': '20283249639',
                                                                       'poller_hash': 'MkVEOTJFcGVlcmludGYyMjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '221',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '29035185803645',
                                                                       'in_ucast_pkts': '20283249639',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '29035185803645',
                                                                                                'in_ucast_pkts': '20283249639',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1144637806514',
                                                                                                'out_ucast_pkts': '6469031392',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1144637806514',
                                                                       'out_ucast_pkts': '6469031392',
                                                                       'poller_hash': 'MkVEOTJFcGVlcmludGZwb24gMjIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221',
                                                                                               'links': [       {       'connected_time': '47317',
                                                                                                                        'fiber_length': 17988,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.672,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVEOTJFcGVlcnBvbnBvbiAyMjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '47 Maddison St. E. Monkton (Mane '
                                                 'Attraction 10431045)',
                                         'poller_hash': 'MkVEOTJF',
                                         'serial_number': '2ED92E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:dc:ec:a4',
                                         'uptime': '47317'},
                                 {       'counter_source': '222/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:48:10'],
                                                                       'id': '222/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '421794526',
                                                                       'in_ucast_pkts': '3031359',
                                                                       'mac_address': '00:06:31:65:58:72',
                                                                       'mtu': 2000,
                                                                       'name': '222/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '91919',
                                                                       'out_octets': '22614230448',
                                                                       'out_ucast_pkts': '15826640',
                                                                       'poller_hash': 'QURCMDNwZWVyaW50ZjIyMi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '222',
                                                                       'in_nucast_pkts': '91919',
                                                                       'in_octets': '22614230448',
                                                                       'in_ucast_pkts': '15826640',
                                                                       'links': [       {       'in_nucast_pkts': '91919',
                                                                                                'in_octets': '22614230448',
                                                                                                'in_ucast_pkts': '15826640',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '421794526',
                                                                                                'out_ucast_pkts': '3031359',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '421794526',
                                                                       'out_ucast_pkts': '3031359',
                                                                       'poller_hash': 'QURCMDNwZWVyaW50ZnBvbiAyMjI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '222',
                                                                                               'links': [       {       'connected_time': '58586',
                                                                                                                        'fiber_length': 17981,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QURCMDNwZWVycG9ucG9uIDIyMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '55 MADDISON ST W (REID MELISSA)',
                                         'poller_hash': 'QURCMDM=',
                                         'serial_number': 'ADB03',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:57:eb:c3',
                                         'uptime': '58586'},
                                 {       'counter_source': '224/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '224/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1790273861516',
                                                                       'in_ucast_pkts': '19378838180',
                                                                       'mac_address': '00:06:31:e3:fc:ab',
                                                                       'name': '224/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '36685835266799',
                                                                       'out_ucast_pkts': '29663067210',
                                                                       'poller_hash': 'MTk5QTJFcGVlcmludGYyMjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '224',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '36685835266799',
                                                                       'in_ucast_pkts': '29663067210',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '36685835266799',
                                                                                                'in_ucast_pkts': '29663067210',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1790273861516',
                                                                                                'out_ucast_pkts': '19378838180',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1790273861516',
                                                                       'out_ucast_pkts': '19378838180',
                                                                       'poller_hash': 'MTk5QTJFcGVlcmludGZwb24gMjI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224',
                                                                                               'links': [       {       'connected_time': '47747',
                                                                                                                        'fiber_length': 17342,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.842,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTk5QTJFcGVlcnBvbnBvbiAyMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '190 Selby Ave. Monkton (Mandy Smith)',
                                         'poller_hash': 'MTk5QTJF',
                                         'serial_number': '199A2E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:e3:fc:ab',
                                         'uptime': '47747'},
                                 {       'counter_source': '227/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '227/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '421123122322',
                                                                       'in_ucast_pkts': '3503073918',
                                                                       'mac_address': '00:06:31:87:45:43',
                                                                       'name': '227/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '18494402901878',
                                                                       'out_ucast_pkts': '12788711225',
                                                                       'poller_hash': 'Q0MwOENwZWVyaW50ZjIyNy9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '227',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '18494402901878',
                                                                       'in_ucast_pkts': '12788711225',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '18494402901878',
                                                                                                'in_ucast_pkts': '12788711225',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '421123122322',
                                                                                                'out_ucast_pkts': '3503073918',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '421123122322',
                                                                       'out_ucast_pkts': '3503073918',
                                                                       'poller_hash': 'Q0MwOENwZWVyaW50ZnBvbiAyMjc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '227',
                                                                                               'links': [       {       'connected_time': '47289',
                                                                                                                        'fiber_length': 17884,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwOENwZWVycG9ucG9uIDIyNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '45 Maddison St. W. Monkton (Darryl '
                                                 'Smith)',
                                         'poller_hash': 'Q0MwOEM=',
                                         'serial_number': 'CC08C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:4c',
                                         'uptime': '47289'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '228/g1',
                                                                       'name': '228/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Q0MwQjhwZWVyaW50ZjIyOC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '228',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'Q0MwQjhwZWVyaW50ZnBvbiAyMjg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '228',
                                                                                               'links': [       {       'connected_time': '69226064',
                                                                                                                        'fiber_length': 18095,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQjhwZWVycG9ucG9uIDIyOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '66 Maddison St. E. Monkton (Dan '
                                                 'Brown)',
                                         'poller_hash': 'Q0MwQjg=',
                                         'serial_number': 'CC0B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:78',
                                         'uptime': '69226064'},
                                 {       'counter_source': '229/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '388957119287',
                                                                       'in_ucast_pkts': '6146036459',
                                                                       'mac_address': '00:06:31:87:48:fb',
                                                                       'name': '229/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '23270866151625',
                                                                       'out_ucast_pkts': '15684202556',
                                                                       'poller_hash': 'Q0MwQjRwZWVyaW50ZjIyOS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '229',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '23270866151625',
                                                                       'in_ucast_pkts': '15684202556',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '23270866151625',
                                                                                                'in_ucast_pkts': '15684202556',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '388957119287',
                                                                                                'out_ucast_pkts': '6146036459',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '388957119287',
                                                                       'out_ucast_pkts': '6146036459',
                                                                       'poller_hash': 'Q0MwQjRwZWVyaW50ZnBvbiAyMjk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229',
                                                                                               'links': [       {       'connected_time': '47331',
                                                                                                                        'fiber_length': 17879,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQjRwZWVycG9ucG9uIDIyOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '30 Maddison St. W. Monkton (Nancy '
                                                 'Diehl)',
                                         'poller_hash': 'Q0MwQjQ=',
                                         'serial_number': 'CC0B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:74',
                                         'uptime': '47331'},
                                 {       'counter_source': '230/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e2:b4'],
                                                                       'id': '230/g1',
                                                                       'in_nucast_pkts': '475681',
                                                                       'in_octets': '441748044594',
                                                                       'in_ucast_pkts': '1692218021',
                                                                       'mac_address': '00:06:31:69:b4:41',
                                                                       'mtu': 2000,
                                                                       'name': '230/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43301817',
                                                                       'out_octets': '10099883567244',
                                                                       'out_ucast_pkts': '7046742153',
                                                                       'poller_hash': 'QzRGRTdwZWVyaW50ZjIzMC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230',
                                                                       'in_nucast_pkts': '43301817',
                                                                       'in_octets': '10099883567244',
                                                                       'in_ucast_pkts': '7046742153',
                                                                       'links': [       {       'in_nucast_pkts': '43301817',
                                                                                                'in_octets': '10099883567244',
                                                                                                'in_ucast_pkts': '7046742153',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '475681',
                                                                                                'out_octets': '441748044594',
                                                                                                'out_ucast_pkts': '1692218021',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '475681',
                                                                       'out_octets': '441748044594',
                                                                       'out_ucast_pkts': '1692218021',
                                                                       'poller_hash': 'QzRGRTdwZWVyaW50ZnBvbiAyMzA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230',
                                                                                               'links': [       {       'connected_time': '24669034',
                                                                                                                        'fiber_length': 17422,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGRTdwZWVycG9ucG9uIDIzMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '185 ERSKINE ST (IRELAND MORRIS)',
                                         'poller_hash': 'QzRGRTc=',
                                         'serial_number': 'C4FE7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:a7',
                                         'uptime': '24669034'},
                                 {       'counter_source': '231/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '231/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '509048167119',
                                                                       'in_ucast_pkts': '2289279176',
                                                                       'mac_address': '00:06:31:a6:6a:eb',
                                                                       'name': '231/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7919304933877',
                                                                       'out_ucast_pkts': '6465191809',
                                                                       'poller_hash': 'MTJGRkU0cGVlcmludGYyMzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '231',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7919304933877',
                                                                       'in_ucast_pkts': '6465191809',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7919304933877',
                                                                                                'in_ucast_pkts': '6465191809',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '509048167119',
                                                                                                'out_ucast_pkts': '2289279176',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '509048167119',
                                                                       'out_ucast_pkts': '2289279176',
                                                                       'poller_hash': 'MTJGRkU0cGVlcmludGZwb24gMjMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231',
                                                                                               'links': [       {       'connected_time': '996059',
                                                                                                                        'fiber_length': 18054,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTJGRkU0cGVlcnBvbnBvbiAyMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '170 Queen Ave. Monkton (Julia Cork)',
                                         'poller_hash': 'MTJGRkU0',
                                         'serial_number': '12FFE4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:a4',
                                         'uptime': '996059'},
                                 {       'counter_source': '232/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '232/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '356443255584',
                                                                       'in_ucast_pkts': '7079346551',
                                                                       'mac_address': '00:06:31:a6:65:e3',
                                                                       'name': '232/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9852098274003',
                                                                       'out_ucast_pkts': '15471411134',
                                                                       'poller_hash': 'MTMwMEExcGVlcmludGYyMzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '232',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9852098274003',
                                                                       'in_ucast_pkts': '15471411134',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9852098274003',
                                                                                                'in_ucast_pkts': '15471411134',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '356443255584',
                                                                                                'out_ucast_pkts': '7079346551',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '356443255584',
                                                                       'out_ucast_pkts': '7079346551',
                                                                       'poller_hash': 'MTMwMEExcGVlcmludGZwb24gMjMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232',
                                                                                               'links': [       {       'connected_time': '48043',
                                                                                                                        'fiber_length': 17745,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMEExcGVlcnBvbnBvbiAyMzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '203 king St. Monkton (Joel Leis)',
                                         'poller_hash': 'MTMwMEEx',
                                         'serial_number': '1300A1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:61',
                                         'uptime': '48043'},
                                 {       'counter_source': '233/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '665671338867',
                                                                       'in_ucast_pkts': '3187441351',
                                                                       'mac_address': '00:06:31:a6:73:2b',
                                                                       'name': '233/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11512177217865',
                                                                       'out_ucast_pkts': '10413784912',
                                                                       'poller_hash': 'MTMwMEEycGVlcmludGYyMzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '233',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11512177217865',
                                                                       'in_ucast_pkts': '10413784912',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11512177217865',
                                                                                                'in_ucast_pkts': '10413784912',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '665671338867',
                                                                                                'out_ucast_pkts': '3187441351',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '665671338867',
                                                                       'out_ucast_pkts': '3187441351',
                                                                       'poller_hash': 'MTMwMEEycGVlcmludGZwb24gMjMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233',
                                                                                               'links': [       {       'connected_time': '3113961',
                                                                                                                        'fiber_length': 17986,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMEEycGVlcnBvbnBvbiAyMzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '21 Maddison Av. E. Monkton (Kyle '
                                                 'Donkers10431767)',
                                         'poller_hash': 'MTMwMEEy',
                                         'serial_number': '1300A2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:62',
                                         'uptime': '3113961'},
                                 {       'counter_source': '234/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '222814055389',
                                                                       'in_ucast_pkts': '2940389082',
                                                                       'mac_address': '00:06:31:a6:6b:db',
                                                                       'name': '234/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7109767816768',
                                                                       'out_ucast_pkts': '10009866710',
                                                                       'poller_hash': 'MTJGRkUxcGVlcmludGYyMzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '234',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7109767816768',
                                                                       'in_ucast_pkts': '10009866710',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7109767816768',
                                                                                                'in_ucast_pkts': '10009866710',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '222814055389',
                                                                                                'out_ucast_pkts': '2940389082',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '222814055389',
                                                                       'out_ucast_pkts': '2940389082',
                                                                       'poller_hash': 'MTJGRkUxcGVlcmludGZwb24gMjM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234',
                                                                                               'links': [       {       'connected_time': '268687',
                                                                                                                        'fiber_length': 17639,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTJGRkUxcGVlcnBvbnBvbiAyMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '169 WINSTANLEY ST (KORNELSEN '
                                                 'DONOVEN)',
                                         'poller_hash': 'MTJGRkUx',
                                         'serial_number': '12FFE1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:a1',
                                         'uptime': '268687'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '236/g1',
                                                                       'name': '236/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTMwMDlDcGVlcmludGYyMzYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '236',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTMwMDlDcGVlcmludGZwb24gMjM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236',
                                                                                               'links': [       {       'connected_time': '69226176',
                                                                                                                        'fiber_length': 17519,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDlDcGVlcnBvbnBvbiAyMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '174 Winstanley St. Monkton',
                                         'poller_hash': 'MTMwMDlD',
                                         'serial_number': '13009C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:5c',
                                         'uptime': '69226176'},
                                 {       'counter_source': '237/F1+237/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:db:4c:3c',
                                                                                                        '00:03:e6:db:b7:34'],
                                                                       'id': '237/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'name': '237/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NzFEODk0cGVlcmludGYyMzcvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '237/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '157641945495',
                                                                       'in_ucast_pkts': '651112078',
                                                                       'mac_address': '48:77:46:a3:bb:2c',
                                                                       'name': '237/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '819674358739',
                                                                       'out_ucast_pkts': '1096654711',
                                                                       'poller_hash': 'NzFEODk0cGVlcmludGYyMzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '237',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '819674358739',
                                                                       'in_ucast_pkts': '1096654711',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '819674358739',
                                                                                                'in_ucast_pkts': '1096654711',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '157641945495',
                                                                                                'out_ucast_pkts': '651112078',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '157641945495',
                                                                       'out_ucast_pkts': '651112078',
                                                                       'poller_hash': 'NzFEODk0cGVlcmludGZwb24gMjM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '237',
                                                                                               'links': [       {       'connected_time': '47934',
                                                                                                                        'fiber_length': 18347,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.636,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFEODk0cGVlcnBvbnBvbiAyMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '142 Queen St. Line 57 Monkton '
                                                 '(Connie Bull)',
                                         'poller_hash': 'NzFEODk0',
                                         'serial_number': '71D894',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a3:bb:2c',
                                         'uptime': '47934'},
                                 {       'counter_source': '238/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '232540305286',
                                                                       'in_ucast_pkts': '1553385517',
                                                                       'mac_address': '00:06:31:a6:73:1b',
                                                                       'name': '238/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7559114966770',
                                                                       'out_ucast_pkts': '1245909078',
                                                                       'poller_hash': 'MTMwMDkxcGVlcmludGYyMzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '238',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7559114966770',
                                                                       'in_ucast_pkts': '1245909078',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7559114966770',
                                                                                                'in_ucast_pkts': '1245909078',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '232540305286',
                                                                                                'out_ucast_pkts': '1553385517',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '232540305286',
                                                                       'out_ucast_pkts': '1553385517',
                                                                       'poller_hash': 'MTMwMDkxcGVlcmludGZwb24gMjM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238',
                                                                                               'links': [       {       'connected_time': '67245829',
                                                                                                                        'fiber_length': 18185,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDkxcGVlcnBvbnBvbiAyMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '125 Maddison St. E. Monkton (Robert '
                                                 'McGuire)',
                                         'poller_hash': 'MTMwMDkx',
                                         'serial_number': '130091',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:51',
                                         'uptime': '67245829'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '239/g1',
                                                                       'name': '239/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyODQ5cGVlcmludGYyMzkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '239',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzIyODQ5cGVlcmludGZwb24gMjM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239',
                                                                                               'links': [       {       'connected_time': '21816609',
                                                                                                                        'fiber_length': 5570,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODQ5cGVlcnBvbnBvbiAyMzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4994 LINE 67 (MCDONALD RICHARD)',
                                         'poller_hash': 'MzIyODQ5',
                                         'serial_number': '322849',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:00:10',
                                         'uptime': '21816609'},
                                 {       'counter_source': '240/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:ba:c1'],
                                                                       'id': '240/g1',
                                                                       'in_nucast_pkts': '2527412',
                                                                       'in_octets': '1202356911933',
                                                                       'in_ucast_pkts': '6338618707',
                                                                       'mac_address': 'd0:76:8f:08:17:fe',
                                                                       'mtu': 2000,
                                                                       'name': '240/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '146310514',
                                                                       'out_octets': '27279713419672',
                                                                       'out_ucast_pkts': '19073082024',
                                                                       'poller_hash': 'ODY5M0Y1cGVlcmludGYyNDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '240',
                                                                       'in_nucast_pkts': '146310514',
                                                                       'in_octets': '27279713419672',
                                                                       'in_ucast_pkts': '19073082024',
                                                                       'links': [       {       'in_nucast_pkts': '146310514',
                                                                                                'in_octets': '27279713419672',
                                                                                                'in_ucast_pkts': '19073082024',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2527412',
                                                                                                'out_octets': '1202356911933',
                                                                                                'out_ucast_pkts': '6338618707',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2527412',
                                                                       'out_octets': '1202356911933',
                                                                       'out_ucast_pkts': '6338618707',
                                                                       'poller_hash': 'ODY5M0Y1cGVlcmludGZwb24gMjQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240',
                                                                                               'links': [       {       'connected_time': '21816210',
                                                                                                                        'fiber_length': 972,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0Y1cGVlcnBvbnBvbiAyNDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6 West St. Milverton (MCLAUGHLIN '
                                                 'CAMERON)',
                                         'poller_hash': 'ODY5M0Y1',
                                         'serial_number': '8693F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:fb',
                                         'uptime': '21816210'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '241/g1',
                                                                       'name': '241/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkJENjZGcGVlcmludGYyNDEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '241',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkJENjZGcGVlcmludGZwb24gMjQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENjZGcGVlcnBvbnBvbiAyNDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '12 Main St. S. Milverton (Execulink '
                                                 'Credit union)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "241" '
                                                                  '(location: GPON '
                                                                  'port "3/5"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2025-05-26T08:31:15.910Z',
                                         'poller_hash': 'MkJENjZG',
                                         'serial_number': '2BD66F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '242/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0f:ed'],
                                                                       'id': '242/g1',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '44775377773',
                                                                       'in_ucast_pkts': '73413123',
                                                                       'mac_address': 'ec:4f:82:63:58:73',
                                                                       'mtu': 2000,
                                                                       'name': '242/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720619',
                                                                       'out_octets': '458020761646',
                                                                       'out_ucast_pkts': '339595150',
                                                                       'poller_hash': 'MkJENDc4cGVlcmludGYyNDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '242',
                                                                       'in_nucast_pkts': '6720619',
                                                                       'in_octets': '458020761646',
                                                                       'in_ucast_pkts': '339595150',
                                                                       'links': [       {       'in_nucast_pkts': '6720619',
                                                                                                'in_octets': '458020761646',
                                                                                                'in_ucast_pkts': '339595150',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20',
                                                                                                'out_octets': '44775377773',
                                                                                                'out_ucast_pkts': '73413123',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20',
                                                                       'out_octets': '44775377773',
                                                                       'out_ucast_pkts': '73413123',
                                                                       'poller_hash': 'MkJENDc4cGVlcmludGZwb24gMjQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242',
                                                                                               'links': [       {       'connected_time': '3113962',
                                                                                                                        'fiber_length': 18869,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENDc4cGVlcnBvbnBvbiAyNDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '80 SCHADE ST (GOODMAN-SMITH ERIN)',
                                         'poller_hash': 'MkJENDc4',
                                         'serial_number': '2BD478',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:58:73',
                                         'uptime': '3113962'},
                                 {       'counter_source': '243/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:32:1b:49:bf:78'],
                                                                       'id': '243/g1',
                                                                       'in_nucast_pkts': '2813',
                                                                       'in_octets': '24554806497',
                                                                       'in_ucast_pkts': '53623951',
                                                                       'mac_address': 'ec:4f:82:63:58:6b',
                                                                       'mtu': 2000,
                                                                       'name': '243/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720648',
                                                                       'out_octets': '257053119877',
                                                                       'out_ucast_pkts': '203804075',
                                                                       'poller_hash': 'MkJENDc0cGVlcmludGYyNDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '243',
                                                                       'in_nucast_pkts': '6720648',
                                                                       'in_octets': '257053119877',
                                                                       'in_ucast_pkts': '203804075',
                                                                       'links': [       {       'in_nucast_pkts': '6720648',
                                                                                                'in_octets': '257053119877',
                                                                                                'in_ucast_pkts': '203804075',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2813',
                                                                                                'out_octets': '24554806497',
                                                                                                'out_ucast_pkts': '53623951',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2813',
                                                                       'out_octets': '24554806497',
                                                                       'out_ucast_pkts': '53623951',
                                                                       'poller_hash': 'MkJENDc0cGVlcmludGZwb24gMjQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243',
                                                                                               'links': [       {       'connected_time': '3113990',
                                                                                                                        'fiber_length': 17883,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENDc0cGVlcnBvbnBvbiAyNDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '29 MADDISON ST W (DOCKING MELANIE)',
                                         'poller_hash': 'MkJENDc0',
                                         'serial_number': '2BD474',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:58:6b',
                                         'uptime': '3113990'},
                                 {       'counter_source': '245/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f6:31'],
                                                                       'id': '245/g1',
                                                                       'in_nucast_pkts': '472392',
                                                                       'in_octets': '86060050891',
                                                                       'in_ucast_pkts': '606858471',
                                                                       'mac_address': 'ec:4f:82:b9:2a:f4',
                                                                       'mtu': 2000,
                                                                       'name': '245/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43307516',
                                                                       'out_octets': '3491340125103',
                                                                       'out_ucast_pkts': '2411544542',
                                                                       'poller_hash': 'MkZFQzA4cGVlcmludGYyNDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '245',
                                                                       'in_nucast_pkts': '43307516',
                                                                       'in_octets': '3491340125103',
                                                                       'in_ucast_pkts': '2411544542',
                                                                       'links': [       {       'in_nucast_pkts': '43307516',
                                                                                                'in_octets': '3491340125103',
                                                                                                'in_ucast_pkts': '2411544542',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '472392',
                                                                                                'out_octets': '86060050891',
                                                                                                'out_ucast_pkts': '606858471',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '472392',
                                                                       'out_octets': '86060050891',
                                                                       'out_ucast_pkts': '606858471',
                                                                       'poller_hash': 'MkZFQzA4cGVlcmludGZwb24gMjQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245',
                                                                                               'links': [       {       'connected_time': '24669410',
                                                                                                                        'fiber_length': 18437,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.738,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQzA4cGVlcnBvbnBvbiAyNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '103 Maddison St. W. Monkton '
                                                 '(Charlene Hamilton 10430268)',
                                         'poller_hash': 'MkZFQzA4',
                                         'serial_number': '2FEC08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:2a:f0',
                                         'uptime': '24669410'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '247/g1',
                                                                       'name': '247/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjY0NDE0cGVlcmludGYyNDcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '247',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjY0NDE0cGVlcmludGZwb24gMjQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjY0NDE0cGVlcnBvbnBvbiAyNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '54 Mill St. W. Milverton (Wesbell '
                                                 'Technologies)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "247" '
                                                                  '(location: GPON '
                                                                  'port "8/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-03-26T06:09:51.790Z',
                                         'poller_hash': 'MjY0NDE0',
                                         'serial_number': '264414',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '248/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:4c:a5:35:93:c2'],
                                                                       'id': '248/g2',
                                                                       'in_nucast_pkts': '25123',
                                                                       'in_octets': '380735961132',
                                                                       'in_ucast_pkts': '912895999',
                                                                       'mac_address': '00:06:31:69:f9:3e',
                                                                       'mtu': 2000,
                                                                       'name': '248/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32396207',
                                                                       'out_octets': '2030780513366',
                                                                       'out_ucast_pkts': '1779946935',
                                                                       'poller_hash': 'QzU4NTJwZWVyaW50ZjI0OC9nMg==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248',
                                                                       'in_nucast_pkts': '32396207',
                                                                       'in_octets': '2030780513366',
                                                                       'in_ucast_pkts': '1779946935',
                                                                       'links': [       {       'in_nucast_pkts': '32396207',
                                                                                                'in_octets': '2030780513366',
                                                                                                'in_ucast_pkts': '1779946935',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '25123',
                                                                                                'out_octets': '380735961132',
                                                                                                'out_ucast_pkts': '912895999',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '25123',
                                                                       'out_octets': '380735961132',
                                                                       'out_ucast_pkts': '912895999',
                                                                       'poller_hash': 'QzU4NTJwZWVyaW50ZnBvbiAyNDg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248',
                                                                                               'links': [       {       'connected_time': '18165231',
                                                                                                                        'fiber_length': 1246,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU4NTJwZWVycG9ucG9uIDI0OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '54 Mill St. W. Milverton (Wesbell '
                                                 'Technologies)',
                                         'poller_hash': 'QzU4NTI=',
                                         'serial_number': 'C5852',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:69:12',
                                         'uptime': '18165231'},
                                 {       'counter_source': '249/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '68:4a:76:d7:a4:20'],
                                                                       'id': '249/g1',
                                                                       'in_nucast_pkts': '23585',
                                                                       'in_octets': '13403927991',
                                                                       'in_ucast_pkts': '59089078',
                                                                       'mac_address': 'ec:4f:82:63:5c:9d',
                                                                       'mtu': 2000,
                                                                       'name': '249/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464121',
                                                                       'out_octets': '193318002063',
                                                                       'out_ucast_pkts': '150545062',
                                                                       'poller_hash': 'MkJENjhEcGVlcmludGYyNDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '249',
                                                                       'in_nucast_pkts': '1464121',
                                                                       'in_octets': '193318002063',
                                                                       'in_ucast_pkts': '150545062',
                                                                       'links': [       {       'in_nucast_pkts': '1464121',
                                                                                                'in_octets': '193318002063',
                                                                                                'in_ucast_pkts': '150545062',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23585',
                                                                                                'out_octets': '13403927991',
                                                                                                'out_ucast_pkts': '59089078',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23585',
                                                                       'out_octets': '13403927991',
                                                                       'out_ucast_pkts': '59089078',
                                                                       'poller_hash': 'MkJENjhEcGVlcmludGZwb24gMjQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249',
                                                                                               'links': [       {       'connected_time': '688500',
                                                                                                                        'fiber_length': 887,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENjhEcGVlcnBvbnBvbiAyNDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '32 Pacific Ave. Milverton (Rachel '
                                                 'Brennan)',
                                         'poller_hash': 'MkJENjhE',
                                         'serial_number': '2BD68D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:5c:9d',
                                         'uptime': '688500'},
                                 {       'counter_source': '251/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '327674423559',
                                                                       'in_ucast_pkts': '4503509965',
                                                                       'mac_address': '00:06:31:3f:d0:06',
                                                                       'name': '251/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10809212813676',
                                                                       'out_ucast_pkts': '9961732038',
                                                                       'poller_hash': 'N0UwRURwZWVyaW50ZjI1MS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '251',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10809212813676',
                                                                       'in_ucast_pkts': '9961732038',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '10809212813676',
                                                                                                'in_ucast_pkts': '9961732038',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '327674423559',
                                                                                                'out_ucast_pkts': '4503509965',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '327674423559',
                                                                       'out_ucast_pkts': '4503509965',
                                                                       'poller_hash': 'N0UwRURwZWVyaW50ZnBvbiAyNTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251',
                                                                                               'links': [       {       'connected_time': '47920',
                                                                                                                        'fiber_length': 18273,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'N0UwRURwZWVycG9ucG9uIDI1MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '51 SCHADE ST (JAKOBSEN SOREN & LEAH)',
                                         'poller_hash': 'N0UwRUQ=',
                                         'serial_number': '7E0ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:41:d3',
                                         'uptime': '47920'},
                                 {       'counter_source': '253/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2a:b0:41'],
                                                                       'id': '253/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '2880364493',
                                                                       'in_ucast_pkts': '15934851',
                                                                       'mac_address': 'ec:4f:82:63:54:61',
                                                                       'mtu': 2000,
                                                                       'name': '253/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464249',
                                                                       'out_octets': '78110485530',
                                                                       'out_ucast_pkts': '57736170',
                                                                       'poller_hash': 'MkJEMjZGcGVlcmludGYyNTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253',
                                                                       'in_nucast_pkts': '1464249',
                                                                       'in_octets': '78110485530',
                                                                       'in_ucast_pkts': '57736170',
                                                                       'links': [       {       'in_nucast_pkts': '1464249',
                                                                                                'in_octets': '78110485530',
                                                                                                'in_ucast_pkts': '57736170',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '2880364493',
                                                                                                'out_ucast_pkts': '15934851',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '2880364493',
                                                                       'out_ucast_pkts': '15934851',
                                                                       'poller_hash': 'MkJEMjZGcGVlcmludGZwb24gMjUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253',
                                                                                               'links': [       {       'connected_time': '688340',
                                                                                                                        'fiber_length': 1354,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJEMjZGcGVlcnBvbnBvbiAyNTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '49 Temperance St Milverton (D&J '
                                                 'Luyten 10433326)',
                                         'poller_hash': 'MkJEMjZG',
                                         'serial_number': '2BD26F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:54:61',
                                         'uptime': '688340'},
                                 {       'counter_source': '254/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1293906236540',
                                                                       'in_ucast_pkts': '3050171251',
                                                                       'mac_address': 'ec:4f:82:4f:89:b7',
                                                                       'name': '254/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '64',
                                                                       'out_octets': '10621866392874',
                                                                       'out_ucast_pkts': '8936934451',
                                                                       'poller_hash': 'Mjk4NDM5cGVlcmludGYyNTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '254',
                                                                       'in_nucast_pkts': '64',
                                                                       'in_octets': '10621866392874',
                                                                       'in_ucast_pkts': '8936934451',
                                                                       'links': [       {       'in_nucast_pkts': '64',
                                                                                                'in_octets': '10621866392874',
                                                                                                'in_ucast_pkts': '8936934451',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1293906236540',
                                                                                                'out_ucast_pkts': '3050171251',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1293906236540',
                                                                       'out_ucast_pkts': '3050171251',
                                                                       'poller_hash': 'Mjk4NDM5cGVlcmludGZwb24gMjU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254',
                                                                                               'links': [       {       'connected_time': '67850913',
                                                                                                                        'fiber_length': 1214,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4NDM5cGVlcnBvbnBvbiAyNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6 ADELAIDE ST (SANGUYO DAVID)',
                                         'poller_hash': 'Mjk4NDM5',
                                         'serial_number': '298439',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:89:b7',
                                         'uptime': '67850913'},
                                 {       'counter_source': '255/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '255/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '295',
                                                                       'in_octets': '584458173370',
                                                                       'in_ucast_pkts': '2722176725',
                                                                       'mac_address': '00:06:31:e8:f4:46',
                                                                       'name': '255/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1778',
                                                                       'out_octets': '11449128463770',
                                                                       'out_ucast_pkts': '8437752760',
                                                                       'poller_hash': 'MUQwNTEwcGVlcmludGYyNTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '255',
                                                                       'in_nucast_pkts': '1778',
                                                                       'in_octets': '11449128463770',
                                                                       'in_ucast_pkts': '8437752760',
                                                                       'links': [       {       'in_nucast_pkts': '1778',
                                                                                                'in_octets': '11449128463770',
                                                                                                'in_ucast_pkts': '8437752760',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '295',
                                                                                                'out_octets': '584458173370',
                                                                                                'out_ucast_pkts': '2722176725',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '295',
                                                                       'out_octets': '584458173370',
                                                                       'out_ucast_pkts': '2722176725',
                                                                       'poller_hash': 'MUQwNTEwcGVlcmludGZwb24gMjU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255',
                                                                                               'links': [       {       'connected_time': '688479',
                                                                                                                        'fiber_length': 849,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNTEwcGVlcnBvbnBvbiAyNTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '18 William St. Milverton (Landon '
                                                 'Sellers)',
                                         'poller_hash': 'MUQwNTEw',
                                         'serial_number': '1D0510',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f4:46',
                                         'uptime': '688479'},
                                 {       'counter_source': '256/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9d:af:21'],
                                                                       'id': '256/g1',
                                                                       'in_nucast_pkts': '188964',
                                                                       'in_octets': '32542492272',
                                                                       'in_ucast_pkts': '300208231',
                                                                       'mac_address': '00:06:31:e9:01:54',
                                                                       'mtu': 2000,
                                                                       'name': '256/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21678694',
                                                                       'out_octets': '1711922642715',
                                                                       'out_ucast_pkts': '1167534969',
                                                                       'poller_hash': 'MUQwNjgzcGVlcmludGYyNTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256',
                                                                       'in_nucast_pkts': '21678694',
                                                                       'in_octets': '1711922642715',
                                                                       'in_ucast_pkts': '1167534969',
                                                                       'links': [       {       'in_nucast_pkts': '21678694',
                                                                                                'in_octets': '1711922642715',
                                                                                                'in_ucast_pkts': '1167534969',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '188964',
                                                                                                'out_octets': '32542492272',
                                                                                                'out_ucast_pkts': '300208231',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '188964',
                                                                       'out_octets': '32542492272',
                                                                       'out_ucast_pkts': '300208231',
                                                                       'poller_hash': 'MUQwNjgzcGVlcmludGZwb24gMjU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256',
                                                                                               'links': [       {       'connected_time': '9960372',
                                                                                                                        'fiber_length': 18806,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjgzcGVlcnBvbnBvbiAyNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '174 Maddison St. E. Monkton (Roy '
                                                 'Beuerman)',
                                         'poller_hash': 'MUQwNjgz',
                                         'serial_number': '1D0683',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:51',
                                         'uptime': '9960372'},
                                 {       'counter_source': '257/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:63:38'],
                                                                       'id': '257/g1',
                                                                       'in_nucast_pkts': '2143773',
                                                                       'in_octets': '1562238598180',
                                                                       'in_ucast_pkts': '8015926679',
                                                                       'mac_address': 'd0:76:8f:08:15:56',
                                                                       'mtu': 2000,
                                                                       'name': '257/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2018930714',
                                                                       'out_octets': '32567408925774',
                                                                       'out_ucast_pkts': '22378380304',
                                                                       'poller_hash': 'ODY5M0EwcGVlcmludGYyNTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '257',
                                                                       'in_nucast_pkts': '2018930714',
                                                                       'in_octets': '32567408925774',
                                                                       'in_ucast_pkts': '22378380304',
                                                                       'links': [       {       'in_nucast_pkts': '2018930714',
                                                                                                'in_octets': '32567408925774',
                                                                                                'in_ucast_pkts': '22378380304',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2143773',
                                                                                                'out_octets': '1562238598180',
                                                                                                'out_ucast_pkts': '8015926679',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2143773',
                                                                       'out_octets': '1562238598180',
                                                                       'out_ucast_pkts': '8015926679',
                                                                       'poller_hash': 'ODY5M0EwcGVlcmludGZwb24gMjU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257',
                                                                                               'links': [       {       'connected_time': '21816447',
                                                                                                                        'fiber_length': 17723,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0EwcGVlcnBvbnBvbiAyNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '162 Winstanley St. Monkton (Jessie '
                                                 'Buchanan)',
                                         'poller_hash': 'ODY5M0Ew',
                                         'serial_number': '8693A0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:53',
                                         'uptime': '21816447'},
                                 {       'counter_source': '259/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f1:35'],
                                                                       'id': '259/g1',
                                                                       'in_nucast_pkts': '294734',
                                                                       'in_octets': '216883787476',
                                                                       'in_ucast_pkts': '594040066',
                                                                       'mac_address': '48:77:46:a2:1c:f0',
                                                                       'mtu': 2000,
                                                                       'name': '259/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26865254',
                                                                       'out_octets': '3170836455542',
                                                                       'out_ucast_pkts': '2328487646',
                                                                       'poller_hash': 'NzFBNENDcGVlcmludGYyNTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '259',
                                                                       'in_nucast_pkts': '26865254',
                                                                       'in_octets': '3170836455542',
                                                                       'in_ucast_pkts': '2328487646',
                                                                       'links': [       {       'in_nucast_pkts': '26865254',
                                                                                                'in_octets': '3170836455542',
                                                                                                'in_ucast_pkts': '2328487646',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '294734',
                                                                                                'out_octets': '216883787476',
                                                                                                'out_ucast_pkts': '594040066',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '294734',
                                                                       'out_octets': '216883787476',
                                                                       'out_ucast_pkts': '594040066',
                                                                       'poller_hash': 'NzFBNENDcGVlcmludGZwb24gMjU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '259',
                                                                                               'links': [       {       'connected_time': '15498431',
                                                                                                                        'fiber_length': 7870,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.53,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFBNENDcGVlcnBvbnBvbiAyNTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5223 Line 60 (Keith Reibeling '
                                                 '10400225)',
                                         'poller_hash': 'NzFBNEND',
                                         'serial_number': '71A4CC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a2:1c:ec',
                                         'uptime': '15498431'},
                                 {       'counter_source': '260/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '260/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '33105229337',
                                                                       'in_ucast_pkts': '288787776',
                                                                       'mac_address': '48:77:46:a2:1d:04',
                                                                       'name': '260/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1304370696857',
                                                                       'out_ucast_pkts': '974275016',
                                                                       'poller_hash': 'NzFBNENGcGVlcmludGYyNjAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '260',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1304370696857',
                                                                       'in_ucast_pkts': '974275016',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1304370696857',
                                                                                                'in_ucast_pkts': '974275016',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '33105229337',
                                                                                                'out_ucast_pkts': '288787776',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '33105229337',
                                                                       'out_ucast_pkts': '288787776',
                                                                       'poller_hash': 'NzFBNENGcGVlcmludGZwb24gMjYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '260',
                                                                                               'links': [       {       'connected_time': '5470219',
                                                                                                                        'fiber_length': 7313,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.112,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFBNENGcGVlcnBvbnBvbiAyNjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5161 Line 60 (Jeanette Young '
                                                 '10431137)',
                                         'poller_hash': 'NzFBNENG',
                                         'serial_number': '71A4CF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a2:1d:04',
                                         'uptime': '5470219'},
                                 {       'counter_source': '261/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '261/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '230812226566',
                                                                       'in_ucast_pkts': '2217385669',
                                                                       'mac_address': 'cc:be:59:2b:56:c7',
                                                                       'name': '261/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6608325717704',
                                                                       'out_ucast_pkts': '6569871492',
                                                                       'poller_hash': 'M0ZBOTBEcGVlcmludGYyNjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '261',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6608325717704',
                                                                       'in_ucast_pkts': '6569871492',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6608325717704',
                                                                                                'in_ucast_pkts': '6569871492',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '230812226566',
                                                                                                'out_ucast_pkts': '2217385669',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '230812226566',
                                                                       'out_ucast_pkts': '2217385669',
                                                                       'poller_hash': 'M0ZBOTBEcGVlcmludGZwb24gMjYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '261',
                                                                                               'links': [       {       'connected_time': '47147',
                                                                                                                        'fiber_length': 10084,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.388,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0ZBOTBEcGVlcnBvbnBvbiAyNjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5460 LINE 60 (HARNEST ALEXANDER)',
                                         'poller_hash': 'M0ZBOTBE',
                                         'serial_number': '3FA90D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:2b:56:c7',
                                         'uptime': '47147'},
                                 {       'counter_source': '262/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a8:81'],
                                                                       'id': '262/g1',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '23898763152',
                                                                       'in_ucast_pkts': '92316564',
                                                                       'mac_address': 'd0:76:8f:1b:9f:26',
                                                                       'mtu': 2000,
                                                                       'name': '262/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447114',
                                                                       'out_octets': '278476396992',
                                                                       'out_ucast_pkts': '213910594',
                                                                       'poller_hash': 'ODcyNDMycGVlcmludGYyNjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '262',
                                                                       'in_nucast_pkts': '1447114',
                                                                       'in_octets': '278476396992',
                                                                       'in_ucast_pkts': '213910594',
                                                                       'links': [       {       'in_nucast_pkts': '1447114',
                                                                                                'in_octets': '278476396992',
                                                                                                'in_ucast_pkts': '213910594',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23',
                                                                                                'out_octets': '23898763152',
                                                                                                'out_ucast_pkts': '92316564',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '23898763152',
                                                                       'out_ucast_pkts': '92316564',
                                                                       'poller_hash': 'ODcyNDMycGVlcmludGZwb24gMjYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262',
                                                                                               'links': [       {       'connected_time': '688382',
                                                                                                                        'fiber_length': 7985,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDMycGVlcnBvbnBvbiAyNjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5227 Line 60 (Mike Schultz 10421360)',
                                         'poller_hash': 'ODcyNDMy',
                                         'serial_number': '872432',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:23',
                                         'uptime': '688382'},
                                 {       'counter_source': '264/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ef:a7'],
                                                                       'id': '264/g1',
                                                                       'in_nucast_pkts': '66701',
                                                                       'in_octets': '25826350596',
                                                                       'in_ucast_pkts': '95915845',
                                                                       'mac_address': 'd0:76:8f:08:17:16',
                                                                       'mtu': 2000,
                                                                       'name': '264/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6672535',
                                                                       'out_octets': '352756150426',
                                                                       'out_ucast_pkts': '262049688',
                                                                       'poller_hash': 'ODY5M0Q4cGVlcmludGYyNjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '264',
                                                                       'in_nucast_pkts': '6672535',
                                                                       'in_octets': '352756150426',
                                                                       'in_ucast_pkts': '262049688',
                                                                       'links': [       {       'in_nucast_pkts': '6672535',
                                                                                                'in_octets': '352756150426',
                                                                                                'in_ucast_pkts': '262049688',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '66701',
                                                                                                'out_octets': '25826350596',
                                                                                                'out_ucast_pkts': '95915845',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66701',
                                                                       'out_octets': '25826350596',
                                                                       'out_ucast_pkts': '95915845',
                                                                       'poller_hash': 'ODY5M0Q4cGVlcmludGZwb24gMjY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264',
                                                                                               'links': [       {       'connected_time': '3113859',
                                                                                                                        'fiber_length': 10020,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0Q4cGVlcnBvbnBvbiAyNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5438 Line 60 (CARTER MIKE & BEV)',
                                         'poller_hash': 'ODY5M0Q4',
                                         'serial_number': '8693D8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:13',
                                         'uptime': '3113859'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266/g1',
                                                                       'name': '266/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkJENjk2cGVlcmludGYyNjYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '266',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkJENjk2cGVlcmludGZwb24gMjY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENjk2cGVlcnBvbnBvbiAyNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '8 Pacific Av. Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "266" '
                                                                  '(location: GPON '
                                                                  'port "8/6"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:10:31.610Z',
                                         'poller_hash': 'MkJENjk2',
                                         'serial_number': '2BD696',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '267/g1',
                                                                       'name': '267/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkIwMjk0cGVlcmludGYyNjcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '267',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MkIwMjk0cGVlcmludGZwb24gMjY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '267',
                                                                                               'links': [       {       'connected_time': '21815361',
                                                                                                                        'fiber_length': 1291,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjk0cGVlcnBvbnBvbiAyNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '9 Edward St. Milverton (Ezra '
                                                 'Streicher)',
                                         'poller_hash': 'MkIwMjk0',
                                         'serial_number': '2B0294',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:79:0b',
                                         'uptime': '21815361'},
                                 {       'counter_source': '268/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1c:3b'],
                                                                       'id': '268/g1',
                                                                       'in_nucast_pkts': '11930',
                                                                       'in_octets': '1177823599',
                                                                       'in_ucast_pkts': '3807716',
                                                                       'mac_address': 'ec:4f:82:5b:77:56',
                                                                       'mtu': 2000,
                                                                       'name': '268/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464152',
                                                                       'out_octets': '28392467417',
                                                                       'out_ucast_pkts': '22541139',
                                                                       'poller_hash': 'MkIwMjVEcGVlcmludGYyNjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '268',
                                                                       'in_nucast_pkts': '1464152',
                                                                       'in_octets': '28392467417',
                                                                       'in_ucast_pkts': '22541139',
                                                                       'links': [       {       'in_nucast_pkts': '1464152',
                                                                                                'in_octets': '28392467417',
                                                                                                'in_ucast_pkts': '22541139',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11930',
                                                                                                'out_octets': '1177823599',
                                                                                                'out_ucast_pkts': '3807716',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11930',
                                                                       'out_octets': '1177823599',
                                                                       'out_ucast_pkts': '3807716',
                                                                       'poller_hash': 'MkIwMjVEcGVlcmludGZwb24gMjY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '268',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 1342,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjVEcGVlcnBvbnBvbiAyNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6458 PERTH ROAD 131 (WAGLER  EZRA)',
                                         'poller_hash': 'MkIwMjVE',
                                         'serial_number': '2B025D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:77:53',
                                         'uptime': '688369'},
                                 {       'counter_source': '269/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2c:6b:78'],
                                                                       'id': '269/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '62203302387',
                                                                       'in_ucast_pkts': '226271381',
                                                                       'mac_address': 'ec:4f:82:5b:77:9e',
                                                                       'mtu': 2000,
                                                                       'name': '269/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720682',
                                                                       'out_octets': '629140298154',
                                                                       'out_ucast_pkts': '494239529',
                                                                       'poller_hash': 'MkIwMjY2cGVlcmludGYyNjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '269',
                                                                       'in_nucast_pkts': '6720682',
                                                                       'in_octets': '629140298154',
                                                                       'in_ucast_pkts': '494239529',
                                                                       'links': [       {       'in_nucast_pkts': '6720682',
                                                                                                'in_octets': '629140298154',
                                                                                                'in_ucast_pkts': '494239529',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '62203302387',
                                                                                                'out_ucast_pkts': '226271381',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '62203302387',
                                                                       'out_ucast_pkts': '226271381',
                                                                       'poller_hash': 'MkIwMjY2cGVlcmludGZwb24gMjY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '269',
                                                                                               'links': [       {       'connected_time': '3113953',
                                                                                                                        'fiber_length': 17918,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjY2cGVlcnBvbnBvbiAyNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 Maddison St. W. Monkton (Rachel '
                                                 'Chiarello)',
                                         'poller_hash': 'MkIwMjY2',
                                         'serial_number': '2B0266',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:77:9b',
                                         'uptime': '3113953'},
                                 {       'counter_source': '270/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '270/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '71245945058',
                                                                       'in_ucast_pkts': '918109127',
                                                                       'mac_address': 'ec:4f:82:5b:77:8b',
                                                                       'name': '270/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '5803799180157',
                                                                       'out_ucast_pkts': '3855394347',
                                                                       'poller_hash': 'MkIwMjY0cGVlcmludGYyNzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '270',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '5803799180157',
                                                                       'in_ucast_pkts': '3855394347',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '5803799180157',
                                                                                                'in_ucast_pkts': '3855394347',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '71245945058',
                                                                                                'out_ucast_pkts': '918109127',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '71245945058',
                                                                       'out_ucast_pkts': '918109127',
                                                                       'poller_hash': 'MkIwMjY0cGVlcmludGZwb24gMjcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '270',
                                                                                               'links': [       {       'connected_time': '31214041',
                                                                                                                        'fiber_length': 1684,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjY0cGVlcnBvbnBvbiAyNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '60 William St. Ste. 105 (Joyce '
                                                 'Westman 10433383)',
                                         'poller_hash': 'MkIwMjY0',
                                         'serial_number': '2B0264',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:77:8b',
                                         'uptime': '31214041'},
                                 {       'counter_source': '271/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:f8:2c:7f:5e:70'],
                                                                       'id': '271/g1',
                                                                       'in_nucast_pkts': '321295',
                                                                       'in_octets': '431830854033',
                                                                       'in_ucast_pkts': '1256073690',
                                                                       'mac_address': '00:06:31:f6:d5:08',
                                                                       'mtu': 2000,
                                                                       'name': '271/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11384417',
                                                                       'out_octets': '810425602509',
                                                                       'out_ucast_pkts': '1258994929',
                                                                       'poller_hash': 'MjYyNUU1cGVlcmludGYyNzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '271',
                                                                       'in_nucast_pkts': '11384417',
                                                                       'in_octets': '810425602509',
                                                                       'in_ucast_pkts': '1258994929',
                                                                       'links': [       {       'in_nucast_pkts': '11384417',
                                                                                                'in_octets': '810425602509',
                                                                                                'in_ucast_pkts': '1258994929',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '321295',
                                                                                                'out_octets': '431830854033',
                                                                                                'out_ucast_pkts': '1256073690',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '321295',
                                                                       'out_octets': '431830854033',
                                                                       'out_ucast_pkts': '1256073690',
                                                                       'poller_hash': 'MjYyNUU1cGVlcmludGZwb24gMjcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '271',
                                                                                               'links': [       {       'connected_time': '6175813',
                                                                                                                        'fiber_length': 1317,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYyNUU1cGVlcnBvbnBvbiAyNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '12 Main St. S. Milverton (Kindred '
                                                 'Credit Union)',
                                         'poller_hash': 'MjYyNUU1',
                                         'serial_number': '2625E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:f6:d5:0d',
                                         'uptime': '6175813'},
                                 {       'counter_source': '272/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e6:24'],
                                                                       'id': '272/g1',
                                                                       'in_nucast_pkts': '67099',
                                                                       'in_octets': '20930811288',
                                                                       'in_ucast_pkts': '54993181',
                                                                       'mac_address': 'ec:4f:82:7d:b9:c7',
                                                                       'mtu': 2000,
                                                                       'name': '272/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720550',
                                                                       'out_octets': '144912625381',
                                                                       'out_ucast_pkts': '118026055',
                                                                       'poller_hash': 'MzAwNjRDcGVlcmludGYyNzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272',
                                                                       'in_nucast_pkts': '6720550',
                                                                       'in_octets': '144912625381',
                                                                       'in_ucast_pkts': '118026055',
                                                                       'links': [       {       'in_nucast_pkts': '6720550',
                                                                                                'in_octets': '144912625381',
                                                                                                'in_ucast_pkts': '118026055',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '67099',
                                                                                                'out_octets': '20930811288',
                                                                                                'out_ucast_pkts': '54993181',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67099',
                                                                       'out_octets': '20930811288',
                                                                       'out_ucast_pkts': '54993181',
                                                                       'poller_hash': 'MzAwNjRDcGVlcmludGZwb24gMjcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272',
                                                                                               'links': [       {       'connected_time': '3113977',
                                                                                                                        'fiber_length': 17982,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwNjRDcGVlcnBvbnBvbiAyNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '157 Smith St. Monkton (Judith '
                                                 'Burwell)',
                                         'poller_hash': 'MzAwNjRD',
                                         'serial_number': '30064C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:b9:c7',
                                         'uptime': '3113977'},
                                 {       'counter_source': '273/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '273/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '294',
                                                                       'in_octets': '118070127724',
                                                                       'in_ucast_pkts': '696368668',
                                                                       'mac_address': 'ec:4f:82:5b:77:bb',
                                                                       'name': '273/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1892',
                                                                       'out_octets': '2804822982271',
                                                                       'out_ucast_pkts': '2185895902',
                                                                       'poller_hash': 'MkIwMjZBcGVlcmludGYyNzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '273',
                                                                       'in_nucast_pkts': '1892',
                                                                       'in_octets': '2804822982271',
                                                                       'in_ucast_pkts': '2185895902',
                                                                       'links': [       {       'in_nucast_pkts': '1892',
                                                                                                'in_octets': '2804822982271',
                                                                                                'in_ucast_pkts': '2185895902',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '294',
                                                                                                'out_octets': '118070127724',
                                                                                                'out_ucast_pkts': '696368668',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '294',
                                                                       'out_octets': '118070127724',
                                                                       'out_ucast_pkts': '696368668',
                                                                       'poller_hash': 'MkIwMjZBcGVlcmludGZwb24gMjcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '273',
                                                                                               'links': [       {       'connected_time': '688462',
                                                                                                                        'fiber_length': 1253,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjZBcGVlcnBvbnBvbiAyNzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '8 George St. Milverton (Stanley '
                                                 'Kuepfer 10400973)',
                                         'poller_hash': 'MkIwMjZB',
                                         'serial_number': '2B026A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:77:bb',
                                         'uptime': '688462'},
                                 {       'counter_source': '274/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '361785983279',
                                                                       'in_ucast_pkts': '49062472815',
                                                                       'mac_address': 'ec:4f:82:5b:93:23',
                                                                       'name': '274/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '12093532033819',
                                                                       'out_ucast_pkts': '8917871445',
                                                                       'poller_hash': 'MkIwNUQ3cGVlcmludGYyNzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '274',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '12093532033819',
                                                                       'in_ucast_pkts': '8917871445',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '12093532033819',
                                                                                                'in_ucast_pkts': '8917871445',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '361785983279',
                                                                                                'out_ucast_pkts': '49062472815',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '361785983279',
                                                                       'out_ucast_pkts': '49062472815',
                                                                       'poller_hash': 'MkIwNUQ3cGVlcmludGZwb24gMjc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274',
                                                                                               'links': [       {       'connected_time': '72951754',
                                                                                                                        'fiber_length': 1098,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNUQ3cGVlcnBvbnBvbiAyNzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29 PACIFIC AV (JUGDAVE VISHAAL)',
                                         'poller_hash': 'MkIwNUQ3',
                                         'serial_number': '2B05D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:93:23',
                                         'uptime': '72951754'},
                                 {       'counter_source': '276/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '276/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '291',
                                                                       'in_octets': '556934826765',
                                                                       'in_ucast_pkts': '4982724954',
                                                                       'mac_address': 'ec:4f:82:73:5f:5e',
                                                                       'name': '276/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1409',
                                                                       'out_octets': '15513369713051',
                                                                       'out_ucast_pkts': '11838587760',
                                                                       'poller_hash': 'MkRDRDE4cGVlcmludGYyNzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '276',
                                                                       'in_nucast_pkts': '1409',
                                                                       'in_octets': '15513369713051',
                                                                       'in_ucast_pkts': '11838587760',
                                                                       'links': [       {       'in_nucast_pkts': '1409',
                                                                                                'in_octets': '15513369713051',
                                                                                                'in_ucast_pkts': '11838587760',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '291',
                                                                                                'out_octets': '556934826765',
                                                                                                'out_ucast_pkts': '4982724954',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '291',
                                                                       'out_octets': '556934826765',
                                                                       'out_ucast_pkts': '4982724954',
                                                                       'poller_hash': 'MkRDRDE4cGVlcmludGZwb24gMjc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276',
                                                                                               'links': [       {       'connected_time': '688478',
                                                                                                                        'fiber_length': 1670,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkRDRDE4cGVlcnBvbnBvbiAyNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '55 WILLIAM ST (FELIX JAMES)',
                                         'poller_hash': 'MkRDRDE4',
                                         'serial_number': '2DCD18',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:73:5f:5e',
                                         'uptime': '688478'},
                                 {       'counter_source': '277/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'bc:fc:e7:22:8d:58'],
                                                                       'id': '277/g1',
                                                                       'in_nucast_pkts': '14613',
                                                                       'in_octets': '45525902578',
                                                                       'in_ucast_pkts': '315847210',
                                                                       'mac_address': '04:bc:9f:6e:59:f1',
                                                                       'mtu': 2000,
                                                                       'name': '277/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6705620',
                                                                       'out_octets': '1359056735249',
                                                                       'out_ucast_pkts': '1042813540',
                                                                       'poller_hash': 'MTI5NUVDMXBlZXJpbnRmMjc3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '277',
                                                                       'in_nucast_pkts': '6705620',
                                                                       'in_octets': '1359056735249',
                                                                       'in_ucast_pkts': '1042813540',
                                                                       'links': [       {       'in_nucast_pkts': '6705620',
                                                                                                'in_octets': '1359056735249',
                                                                                                'in_ucast_pkts': '1042813540',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14613',
                                                                                                'out_octets': '45525902578',
                                                                                                'out_ucast_pkts': '315847210',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14613',
                                                                       'out_octets': '45525902578',
                                                                       'out_ucast_pkts': '315847210',
                                                                       'poller_hash': 'MTI5NUVDMXBlZXJpbnRmcG9uIDI3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277',
                                                                                               'links': [       {       'connected_time': '3126195',
                                                                                                                        'fiber_length': 6390,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVDMXBlZXJwb25wb24gMjc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4926 Line 61 Milverton (Markus '
                                                 'Mueller)',
                                         'poller_hash': 'MTI5NUVDMQ==',
                                         'serial_number': '1295EC1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:59:ee',
                                         'uptime': '3126195'},
                                 {       'counter_source': '278/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '278/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2562795536683',
                                                                       'in_ucast_pkts': '9243630750',
                                                                       'mac_address': '00:06:31:8f:7e:e3',
                                                                       'name': '278/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '42158244888375',
                                                                       'out_ucast_pkts': '31181519067',
                                                                       'poller_hash': 'MTI0ODRFcGVlcmludGYyNzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '278',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '42158244888375',
                                                                       'in_ucast_pkts': '31181519067',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '42158244888375',
                                                                                                'in_ucast_pkts': '31181519067',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2562795536683',
                                                                                                'out_ucast_pkts': '9243630750',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2562795536683',
                                                                       'out_ucast_pkts': '9243630750',
                                                                       'poller_hash': 'MTI0ODRFcGVlcmludGZwb24gMjc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278',
                                                                                               'links': [       {       'connected_time': '15413616',
                                                                                                                        'fiber_length': 7214,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.612,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI0ODRFcGVlcnBvbnBvbiAyNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '4837 Line 61 Milverton (Grace '
                                                 'Mooney)',
                                         'poller_hash': 'MTI0ODRF',
                                         'serial_number': '12484E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:39:0e',
                                         'uptime': '15413616'},
                                 {       'counter_source': '279/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2d:48:f4'],
                                                                       'id': '279/g1',
                                                                       'in_nucast_pkts': '72',
                                                                       'in_octets': '66328677332',
                                                                       'in_ucast_pkts': '668128851',
                                                                       'mac_address': '00:06:31:87:4c:5f',
                                                                       'mtu': 2000,
                                                                       'name': '279/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53663600',
                                                                       'out_octets': '3909817787367',
                                                                       'out_ucast_pkts': '2620191485',
                                                                       'poller_hash': 'Q0JGRjVwZWVyaW50ZjI3OS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '279',
                                                                       'in_nucast_pkts': '53663600',
                                                                       'in_octets': '3909817787367',
                                                                       'in_ucast_pkts': '2620191485',
                                                                       'links': [       {       'in_nucast_pkts': '53663600',
                                                                                                'in_octets': '3909817787367',
                                                                                                'in_ucast_pkts': '2620191485',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '72',
                                                                                                'out_octets': '66328677332',
                                                                                                'out_ucast_pkts': '668128851',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '72',
                                                                       'out_octets': '66328677332',
                                                                       'out_ucast_pkts': '668128851',
                                                                       'poller_hash': 'Q0JGRjVwZWVyaW50ZnBvbiAyNzk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279',
                                                                                               'links': [       {       'connected_time': '31199410',
                                                                                                                        'fiber_length': 9966,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JGRjVwZWVycG9ucG9uIDI3OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '5433 Line 60 (Joanne Ohm)',
                                         'poller_hash': 'Q0JGRjU=',
                                         'serial_number': 'CBFF5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d0:b5',
                                         'uptime': '31199410'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '280/g1',
                                                                       'name': '280/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Q0JGRjRwZWVyaW50ZjI4MC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '280',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Q0JGRjRwZWVyaW50ZnBvbiAyODA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JGRjRwZWVycG9ucG9uIDI4MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '5250 Line 60 (Theo Kersten)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "280" '
                                                                  '(location: GPON '
                                                                  'port "5/2"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-11-30T14:00:17.930Z',
                                         'poller_hash': 'Q0JGRjQ=',
                                         'serial_number': 'CBFF4',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '281/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c0:56:27:b5:25:10'],
                                                                       'id': '281/g1',
                                                                       'in_nucast_pkts': '517194',
                                                                       'in_octets': '120563533209',
                                                                       'in_ucast_pkts': '456670317',
                                                                       'mac_address': '00:06:31:a6:67:9f',
                                                                       'mtu': 2000,
                                                                       'name': '281/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27081373',
                                                                       'out_octets': '2428181878168',
                                                                       'out_ucast_pkts': '1801769523',
                                                                       'poller_hash': 'MTMwMEQ2cGVlcmludGYyODEvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '281',
                                                                       'in_nucast_pkts': '27081373',
                                                                       'in_octets': '2428181878168',
                                                                       'in_ucast_pkts': '1801769523',
                                                                       'links': [       {       'in_nucast_pkts': '27081373',
                                                                                                'in_octets': '2428181878168',
                                                                                                'in_ucast_pkts': '1801769523',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '517194',
                                                                                                'out_octets': '120563533209',
                                                                                                'out_ucast_pkts': '456670317',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '517194',
                                                                       'out_octets': '120563533209',
                                                                       'out_ucast_pkts': '456670317',
                                                                       'poller_hash': 'MTMwMEQ2cGVlcmludGZwb24gMjgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '281',
                                                                                               'links': [       {       'connected_time': '15635754',
                                                                                                                        'fiber_length': 6630,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMEQ2cGVlcnBvbnBvbiAyODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4890 Line 61 Milverton (Eddie Jantzi '
                                                 '10425951)',
                                         'poller_hash': 'MTMwMEQ2',
                                         'serial_number': '1300D6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:96',
                                         'uptime': '15635754'},
                                 {       'counter_source': '282/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '567337392130',
                                                                       'in_ucast_pkts': '4382959852',
                                                                       'mac_address': '00:06:31:dd:56:d7',
                                                                       'name': '282/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '23129267556870',
                                                                       'out_ucast_pkts': '16115199838',
                                                                       'poller_hash': 'MTc1RTcwcGVlcmludGYyODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '282',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '23129267556870',
                                                                       'in_ucast_pkts': '16115199838',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '23129267556870',
                                                                                                'in_ucast_pkts': '16115199838',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '567337392130',
                                                                                                'out_ucast_pkts': '4382959852',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '567337392130',
                                                                       'out_ucast_pkts': '4382959852',
                                                                       'poller_hash': 'MTc1RTcwcGVlcmludGZwb24gMjgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282',
                                                                                               'links': [       {       'connected_time': '64048647',
                                                                                                                        'fiber_length': 9341,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc1RTcwcGVlcnBvbnBvbiAyODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '5340 Line 60 Milverton (C '
                                                 'VanDerStraeten 10434906)',
                                         'poller_hash': 'MTc1RTcw',
                                         'serial_number': '175E70',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:56:d7',
                                         'uptime': '64048647'},
                                 {       'counter_source': '283/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2b:48:d5'],
                                                                       'id': '283/g1',
                                                                       'in_nucast_pkts': '368',
                                                                       'in_octets': '131456322460',
                                                                       'in_ucast_pkts': '256298285',
                                                                       'mac_address': 'b8:94:70:1f:0a:05',
                                                                       'mtu': 2000,
                                                                       'name': '283/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '77874752',
                                                                       'out_octets': '671222814870',
                                                                       'out_ucast_pkts': '582706493',
                                                                       'poller_hash': 'RkQ4NzNDcGVlcmludGYyODMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '283',
                                                                       'in_nucast_pkts': '77874752',
                                                                       'in_octets': '671222814870',
                                                                       'in_ucast_pkts': '582706493',
                                                                       'links': [       {       'in_nucast_pkts': '77874752',
                                                                                                'in_octets': '671222814870',
                                                                                                'in_ucast_pkts': '582706493',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '368',
                                                                                                'out_octets': '131456322460',
                                                                                                'out_ucast_pkts': '256298285',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '368',
                                                                       'out_octets': '131456322460',
                                                                       'out_ucast_pkts': '256298285',
                                                                       'poller_hash': 'RkQ4NzNDcGVlcmludGZwb24gMjgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '283',
                                                                                               'links': [       {       'connected_time': '42328',
                                                                                                                        'fiber_length': 8099,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.342,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RkQ4NzNDcGVlcnBvbnBvbiAyODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '5208 Line 60 (Louis Byler)',
                                         'poller_hash': 'RkQ4NzND',
                                         'serial_number': 'FD873C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'b8:94:70:1f:0a:01',
                                         'uptime': '42328'},
                                 {       'counter_source': '284/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:5d:bd:9f'],
                                                                       'id': '284/g1',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '31223212088',
                                                                       'in_ucast_pkts': '159275601',
                                                                       'mac_address': '00:06:31:e9:01:f6',
                                                                       'mtu': 2000,
                                                                       'name': '284/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464145',
                                                                       'out_octets': '692089196995',
                                                                       'out_ucast_pkts': '488716889',
                                                                       'poller_hash': 'MUQwNjk1cGVlcmludGYyODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '284',
                                                                       'in_nucast_pkts': '1464145',
                                                                       'in_octets': '692089196995',
                                                                       'in_ucast_pkts': '488716889',
                                                                       'links': [       {       'in_nucast_pkts': '1464145',
                                                                                                'in_octets': '692089196995',
                                                                                                'in_ucast_pkts': '488716889',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12',
                                                                                                'out_octets': '31223212088',
                                                                                                'out_ucast_pkts': '159275601',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '31223212088',
                                                                       'out_ucast_pkts': '159275601',
                                                                       'poller_hash': 'MUQwNjk1cGVlcmludGZwb24gMjg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '284',
                                                                                               'links': [       {       'connected_time': '688359',
                                                                                                                        'fiber_length': 1301,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjk1cGVlcnBvbnBvbiAyODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '41 Temperance St. Milverton '
                                                 '(Charlene Lehman 10433199)',
                                         'poller_hash': 'MUQwNjk1',
                                         'serial_number': '1D0695',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:f3',
                                         'uptime': '688359'},
                                 {       'counter_source': '285/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '285/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '64261111878',
                                                                       'in_ucast_pkts': '274085354',
                                                                       'mac_address': 'ec:4f:82:8d:ff:d0',
                                                                       'name': '285/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '1575031711163',
                                                                       'out_ucast_pkts': '1231201840',
                                                                       'poller_hash': 'MzIyODQxcGVlcmludGYyODUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '285',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '1575031711163',
                                                                       'in_ucast_pkts': '1231201840',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '1575031711163',
                                                                                                'in_ucast_pkts': '1231201840',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '64261111878',
                                                                                                'out_ucast_pkts': '274085354',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '64261111878',
                                                                       'out_ucast_pkts': '274085354',
                                                                       'poller_hash': 'MzIyODQxcGVlcmludGZwb24gMjg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '285',
                                                                                               'links': [       {       'connected_time': '21816409',
                                                                                                                        'fiber_length': 6097,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODQxcGVlcnBvbnBvbiAyODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4952 Line 67 Milverton (Ivo Looser '
                                                 '10400216)',
                                         'poller_hash': 'MzIyODQx',
                                         'serial_number': '322841',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:ff:d0',
                                         'uptime': '21816409'},
                                 {       'counter_source': '286/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:ea:20'],
                                                                       'id': '286/g1',
                                                                       'in_nucast_pkts': '200061',
                                                                       'in_octets': '36735276954',
                                                                       'in_ucast_pkts': '250537158',
                                                                       'mac_address': 'ec:4f:82:8d:de:d3',
                                                                       'mtu': 2000,
                                                                       'name': '286/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23313256',
                                                                       'out_octets': '1310492361720',
                                                                       'out_ucast_pkts': '940360750',
                                                                       'poller_hash': 'MzIyNDIxcGVlcmludGYyODYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '286',
                                                                       'in_nucast_pkts': '23313256',
                                                                       'in_octets': '1310492361720',
                                                                       'in_ucast_pkts': '940360750',
                                                                       'links': [       {       'in_nucast_pkts': '23313256',
                                                                                                'in_octets': '1310492361720',
                                                                                                'in_ucast_pkts': '940360750',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '200061',
                                                                                                'out_octets': '36735276954',
                                                                                                'out_ucast_pkts': '250537158',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '200061',
                                                                       'out_octets': '36735276954',
                                                                       'out_ucast_pkts': '250537158',
                                                                       'poller_hash': 'MzIyNDIxcGVlcmludGZwb24gMjg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '286',
                                                                                               'links': [       {       'connected_time': '10783334',
                                                                                                                        'fiber_length': 5911,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDIxcGVlcnBvbnBvbiAyODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6717 Rd 140 Milverton (Beat Looser '
                                                 '10400601)',
                                         'poller_hash': 'MzIyNDIx',
                                         'serial_number': '322421',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:de:d0',
                                         'uptime': '10783334'},
                                 {       'counter_source': '287/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:bb:12'],
                                                                       'id': '287/g1',
                                                                       'in_nucast_pkts': '15558',
                                                                       'in_octets': '28926689067',
                                                                       'in_ucast_pkts': '34002267',
                                                                       'mac_address': 'ec:4f:82:7d:b9:75',
                                                                       'mtu': 2000,
                                                                       'name': '287/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464037',
                                                                       'out_octets': '69509774694',
                                                                       'out_ucast_pkts': '59167842',
                                                                       'poller_hash': 'MzAwNjIzcGVlcmludGYyODcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '287',
                                                                       'in_nucast_pkts': '1464037',
                                                                       'in_octets': '69509774694',
                                                                       'in_ucast_pkts': '59167842',
                                                                       'links': [       {       'in_nucast_pkts': '1464037',
                                                                                                'in_octets': '69509774694',
                                                                                                'in_ucast_pkts': '59167842',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15558',
                                                                                                'out_octets': '28926689067',
                                                                                                'out_ucast_pkts': '34002267',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15558',
                                                                       'out_octets': '28926689067',
                                                                       'out_ucast_pkts': '34002267',
                                                                       'poller_hash': 'MzAwNjIzcGVlcmludGZwb24gMjg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '287',
                                                                                               'links': [       {       'connected_time': '688489',
                                                                                                                        'fiber_length': 899,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwNjIzcGVlcnBvbnBvbiAyODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '36 PACIFIC AV (HORTON CHAD)',
                                         'poller_hash': 'MzAwNjIz',
                                         'serial_number': '300623',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:b9:75',
                                         'uptime': '688489'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '288/g1',
                                                                       'name': '288/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGODRDcGVlcmludGYyODgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '288',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzVGODRDcGVlcmludGZwb24gMjg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '288',
                                                                                               'links': [       {       'connected_time': '67790291',
                                                                                                                        'fiber_length': 17982,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGODRDcGVlcnBvbnBvbiAyODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '170 SMITH ST (JANZEN EDDIE)',
                                         'poller_hash': 'MzVGODRD',
                                         'serial_number': '35F84C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:36:28',
                                         'uptime': '67790291'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '289/g1',
                                                                       'name': '289/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjYyNUVBcGVlcmludGYyODkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '289',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjYyNUVBcGVlcmludGZwb24gMjg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '289',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYyNUVBcGVlcnBvbnBvbiAyODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '48 PUGH ST',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "289" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-24T08:29:34.070Z',
                                         'poller_hash': 'MjYyNUVB',
                                         'serial_number': '2625EA',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '290/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:32:e5:11:31:61'],
                                                                       'id': '290/g1',
                                                                       'in_nucast_pkts': '9516',
                                                                       'in_octets': '52287579825',
                                                                       'in_ucast_pkts': '237198510',
                                                                       'mac_address': 'ec:4f:82:7d:bc:8b',
                                                                       'mtu': 2000,
                                                                       'name': '290/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13415176',
                                                                       'out_octets': '1266897634338',
                                                                       'out_ucast_pkts': '1001414512',
                                                                       'poller_hash': 'MzAwN0FFcGVlcmludGYyOTAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '290',
                                                                       'in_nucast_pkts': '13415176',
                                                                       'in_octets': '1266897634338',
                                                                       'in_ucast_pkts': '1001414512',
                                                                       'links': [       {       'in_nucast_pkts': '13415176',
                                                                                                'in_octets': '1266897634338',
                                                                                                'in_ucast_pkts': '1001414512',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9516',
                                                                                                'out_octets': '52287579825',
                                                                                                'out_ucast_pkts': '237198510',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9516',
                                                                       'out_octets': '52287579825',
                                                                       'out_ucast_pkts': '237198510',
                                                                       'poller_hash': 'MzAwN0FFcGVlcmludGZwb24gMjkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '290',
                                                                                               'links': [       {       'connected_time': '6158560',
                                                                                                                        'fiber_length': 18151,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwN0FFcGVlcnBvbnBvbiAyOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '38 Maddison St. E. Monkton (Bjorn '
                                                 'Makinen)',
                                         'poller_hash': 'MzAwN0FF',
                                         'serial_number': '3007AE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:bc:8b',
                                         'uptime': '6158560'},
                                 {       'counter_source': '291/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:fd:74:0b:c4:d1'],
                                                                       'id': '291/g1',
                                                                       'in_nucast_pkts': '22',
                                                                       'in_octets': '4299983979730',
                                                                       'in_ucast_pkts': '4215237363',
                                                                       'mac_address': '48:77:46:a0:38:0c',
                                                                       'mtu': 2000,
                                                                       'name': '291/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27173523',
                                                                       'out_octets': '5509617720301',
                                                                       'out_ucast_pkts': '4841490339',
                                                                       'poller_hash': 'NzVDMURDcGVlcmludGYyOTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '291',
                                                                       'in_nucast_pkts': '27173523',
                                                                       'in_octets': '5509617720301',
                                                                       'in_ucast_pkts': '4841490339',
                                                                       'links': [       {       'in_nucast_pkts': '27173523',
                                                                                                'in_octets': '5509617720301',
                                                                                                'in_ucast_pkts': '4841490339',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22',
                                                                                                'out_octets': '4299983979730',
                                                                                                'out_ucast_pkts': '4215237363',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22',
                                                                       'out_octets': '4299983979730',
                                                                       'out_ucast_pkts': '4215237363',
                                                                       'poller_hash': 'NzVDMURDcGVlcmludGZwb24gMjkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '291',
                                                                                               'links': [       {       'connected_time': '15498467',
                                                                                                                        'fiber_length': 1600,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzVDMURDcGVlcnBvbnBvbiAyOTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '93 Main St. N. Milverton (Milverton '
                                                 'Vet Clinic 10400849)',
                                         'poller_hash': 'NzVDMURD',
                                         'serial_number': '75C1DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a0:38:11',
                                         'uptime': '15498467'},
                                 {       'counter_source': '292/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:0d:b9:57:ae:71'],
                                                                       'id': '292/g1',
                                                                       'in_nucast_pkts': '13017',
                                                                       'in_octets': '65333049689',
                                                                       'in_ucast_pkts': '316905604',
                                                                       'mac_address': 'ec:4f:82:4b:b7:57',
                                                                       'mtu': 2000,
                                                                       'name': '292/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27344767',
                                                                       'out_octets': '364581333822',
                                                                       'out_ucast_pkts': '439917403',
                                                                       'poller_hash': 'MjU5ODM3cGVlcmludGYyOTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '292',
                                                                       'in_nucast_pkts': '27344767',
                                                                       'in_octets': '364581333822',
                                                                       'in_ucast_pkts': '439917403',
                                                                       'links': [       {       'in_nucast_pkts': '27344767',
                                                                                                'in_octets': '364581333822',
                                                                                                'in_ucast_pkts': '439917403',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13017',
                                                                                                'out_octets': '65333049689',
                                                                                                'out_ucast_pkts': '316905604',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13017',
                                                                       'out_octets': '65333049689',
                                                                       'out_ucast_pkts': '316905604',
                                                                       'poller_hash': 'MjU5ODM3cGVlcmludGZwb24gMjky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '292',
                                                                                               'links': [       {       'connected_time': '15498562',
                                                                                                                        'fiber_length': 1677,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.516,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjU5ODM3cGVlcnBvbnBvbiAyOTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '6440 Perth Rd. 131 (Leis Feed '
                                                 '10422648)',
                                         'poller_hash': 'MjU5ODM3',
                                         'serial_number': '259837',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:4b:b7:53',
                                         'uptime': '15498562'},
                                 {       'counter_source': '293/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '293/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1301916765984',
                                                                       'in_ucast_pkts': '6395314093',
                                                                       'mac_address': 'ec:4f:82:5b:74:9b',
                                                                       'name': '293/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '4048851407208',
                                                                       'out_ucast_pkts': '3571713323',
                                                                       'poller_hash': 'MkIwMjA2cGVlcmludGYyOTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '293',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '4048851407208',
                                                                       'in_ucast_pkts': '3571713323',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '4048851407208',
                                                                                                'in_ucast_pkts': '3571713323',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1301916765984',
                                                                                                'out_ucast_pkts': '6395314093',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1301916765984',
                                                                       'out_ucast_pkts': '6395314093',
                                                                       'poller_hash': 'MkIwMjA2cGVlcmludGZwb24gMjkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '293',
                                                                                               'links': [       {       'connected_time': '10799098',
                                                                                                                        'fiber_length': 1761,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjA2cGVlcnBvbnBvbiAyOTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6432 Perth Rd 131 (Morris Gerber '
                                                 'Excavating 10401316)',
                                         'poller_hash': 'MkIwMjA2',
                                         'serial_number': '2B0206',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:74:9b',
                                         'uptime': '10799098'},
                                 {       'counter_source': '294/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '294/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96840905530',
                                                                       'in_ucast_pkts': '1143435300',
                                                                       'mac_address': 'ec:4f:82:5b:77:fb',
                                                                       'name': '294/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '4639166987305',
                                                                       'out_ucast_pkts': '3167564379',
                                                                       'poller_hash': 'MkIwMjcycGVlcmludGYyOTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '294',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '4639166987305',
                                                                       'in_ucast_pkts': '3167564379',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '4639166987305',
                                                                                                'in_ucast_pkts': '3167564379',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96840905530',
                                                                                                'out_ucast_pkts': '1143435300',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96840905530',
                                                                       'out_ucast_pkts': '1143435300',
                                                                       'poller_hash': 'MkIwMjcycGVlcmludGZwb24gMjk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '294',
                                                                                               'links': [       {       'connected_time': '688371',
                                                                                                                        'fiber_length': 1407,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMjcycGVlcnBvbnBvbiAyOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6468 Perth Rd. 131 (Jason Lantz '
                                                 '10430328)',
                                         'poller_hash': 'MkIwMjcy',
                                         'serial_number': '2B0272',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:77:fb',
                                         'uptime': '688371'},
                                 {       'counter_source': '295/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '295/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1132566196341',
                                                                       'in_ucast_pkts': '3426746798',
                                                                       'mac_address': 'ec:4f:82:0d:57:91',
                                                                       'name': '295/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '14378537708672',
                                                                       'out_ucast_pkts': '11022256980',
                                                                       'poller_hash': 'MjA1QkRDcGVlcmludGYyOTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '295',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '14378537708672',
                                                                       'in_ucast_pkts': '11022256980',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '14378537708672',
                                                                                                'in_ucast_pkts': '11022256980',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1132566196341',
                                                                                                'out_ucast_pkts': '3426746798',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1132566196341',
                                                                       'out_ucast_pkts': '3426746798',
                                                                       'poller_hash': 'MjA1QkRDcGVlcmludGZwb24gMjk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '295',
                                                                                               'links': [       {       'connected_time': '3113063',
                                                                                                                        'fiber_length': 18000,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1QkRDcGVlcnBvbnBvbiAyOTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '4 Winstanley St Monkton (Cherie '
                                                 'Smith 10430700)',
                                         'poller_hash': 'MjA1QkRD',
                                         'serial_number': '205BDC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:57:91',
                                         'uptime': '3113063'},
                                 {       'counter_source': '296/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f3:d2'],
                                                                       'id': '296/g1',
                                                                       'in_nucast_pkts': '523834',
                                                                       'in_octets': '223822807392',
                                                                       'in_ucast_pkts': '410406249',
                                                                       'mac_address': 'ec:4f:82:4f:87:5a',
                                                                       'mtu': 2000,
                                                                       'name': '296/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68938959',
                                                                       'out_octets': '1098796975886',
                                                                       'out_ucast_pkts': '958152557',
                                                                       'poller_hash': 'Mjk4M0VEcGVlcmludGYyOTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '296',
                                                                       'in_nucast_pkts': '68938959',
                                                                       'in_octets': '1098796975886',
                                                                       'in_ucast_pkts': '958152557',
                                                                       'links': [       {       'in_nucast_pkts': '68938959',
                                                                                                'in_octets': '1098796975886',
                                                                                                'in_ucast_pkts': '958152557',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '523834',
                                                                                                'out_octets': '223822807392',
                                                                                                'out_ucast_pkts': '410406249',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '523834',
                                                                       'out_octets': '223822807392',
                                                                       'out_ucast_pkts': '410406249',
                                                                       'poller_hash': 'Mjk4M0VEcGVlcmludGZwb24gMjk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '296',
                                                                                               'links': [       {       'connected_time': '29301849',
                                                                                                                        'fiber_length': 1271,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4M0VEcGVlcnBvbnBvbiAyOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6469 Perth Rd. 131 (Brian Nafziger '
                                                 '10427589)',
                                         'poller_hash': 'Mjk4M0VE',
                                         'serial_number': '2983ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:87:57',
                                         'uptime': '29301849'},
                                 {       'counter_source': '297/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '297/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '113482170920',
                                                                       'in_ucast_pkts': '1147660593',
                                                                       'mac_address': 'ec:4f:82:0d:6a:b1',
                                                                       'name': '297/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '6991124881989',
                                                                       'out_ucast_pkts': '4676447276',
                                                                       'poller_hash': 'MjA1REZDcGVlcmludGYyOTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '297',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '6991124881989',
                                                                       'in_ucast_pkts': '4676447276',
                                                                       'links': [       {       'in_nucast_pkts': '66',
                                                                                                'in_octets': '6991124881989',
                                                                                                'in_ucast_pkts': '4676447276',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '113482170920',
                                                                                                'out_ucast_pkts': '1147660593',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '113482170920',
                                                                       'out_ucast_pkts': '1147660593',
                                                                       'poller_hash': 'MjA1REZDcGVlcmludGZwb24gMjk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '297',
                                                                                               'links': [       {       'connected_time': '72824819',
                                                                                                                        'fiber_length': 1496,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REZDcGVlcnBvbnBvbiAyOTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '6480 Perth Rd. 131 (Fritz Ryter)',
                                         'poller_hash': 'MjA1REZD',
                                         'serial_number': '205DFC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:b1',
                                         'uptime': '72824819'},
                                 {       'counter_source': '298/g1+298/g3+298/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0b:ac'],
                                                                       'id': '298/g1',
                                                                       'in_nucast_pkts': '10395',
                                                                       'in_octets': '690692699145',
                                                                       'in_ucast_pkts': '1990680696',
                                                                       'mac_address': 'ec:4f:82:4f:8a:c2',
                                                                       'mtu': 2000,
                                                                       'name': '298/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150470168',
                                                                       'out_octets': '8647705084750',
                                                                       'out_ucast_pkts': '6721945526',
                                                                       'poller_hash': 'Mjk4NDVBcGVlcmludGYyOTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:06:98'],
                                                                       'id': '298/g3',
                                                                       'in_nucast_pkts': '9821050',
                                                                       'in_octets': '66790859887',
                                                                       'in_ucast_pkts': '838947426',
                                                                       'mac_address': 'ec:4f:82:4f:8a:c4',
                                                                       'mtu': 2000,
                                                                       'name': '298/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21603297309',
                                                                       'out_octets': '33156464586554',
                                                                       'out_ucast_pkts': '3330546706',
                                                                       'poller_hash': 'Mjk4NDVBcGVlcmludGYyOTgvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:05:66'],
                                                                       'id': '298/g4',
                                                                       'in_nucast_pkts': '10312641',
                                                                       'in_octets': '40734437005',
                                                                       'in_ucast_pkts': '469672907',
                                                                       'mac_address': 'ec:4f:82:4f:8a:c5',
                                                                       'mtu': 2000,
                                                                       'name': '298/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24042135602',
                                                                       'out_octets': '34569090433533',
                                                                       'out_ucast_pkts': '1990251079',
                                                                       'poller_hash': 'Mjk4NDVBcGVlcmludGYyOTgvZzQ=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '298',
                                                                       'in_nucast_pkts': '45795903079',
                                                                       'in_octets': '76373260104837',
                                                                       'in_ucast_pkts': '12042743311',
                                                                       'links': [       {       'in_nucast_pkts': '45795903079',
                                                                                                'in_octets': '76373260104837',
                                                                                                'in_ucast_pkts': '12042743311',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20144086',
                                                                                                'out_octets': '798217996037',
                                                                                                'out_ucast_pkts': '3299301029',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20144086',
                                                                       'out_octets': '798217996037',
                                                                       'out_ucast_pkts': '3299301029',
                                                                       'poller_hash': 'Mjk4NDVBcGVlcmludGZwb24gMjk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '298',
                                                                                               'links': [       {       'connected_time': '21816413',
                                                                                                                        'fiber_length': 1345,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4NDVBcGVlcnBvbnBvbiAyOTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6464 Perth Rd. 131 (Ross Gerber)',
                                         'poller_hash': 'Mjk4NDVB',
                                         'serial_number': '29845A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:8a:bf',
                                         'uptime': '21816413'},
                                 {       'counter_source': '299/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '299/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '340434104675',
                                                                       'in_ucast_pkts': '53739258918',
                                                                       'mac_address': 'ec:4f:82:4f:7a:67',
                                                                       'name': '299/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '147',
                                                                       'out_octets': '6922817845843',
                                                                       'out_ucast_pkts': '5128182132',
                                                                       'poller_hash': 'Mjk4MjRGcGVlcmludGYyOTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '299',
                                                                       'in_nucast_pkts': '147',
                                                                       'in_octets': '6922817845843',
                                                                       'in_ucast_pkts': '5128182132',
                                                                       'links': [       {       'in_nucast_pkts': '147',
                                                                                                'in_octets': '6922817845843',
                                                                                                'in_ucast_pkts': '5128182132',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '340434104675',
                                                                                                'out_ucast_pkts': '53739258918',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '340434104675',
                                                                       'out_ucast_pkts': '53739258918',
                                                                       'poller_hash': 'Mjk4MjRGcGVlcmludGZwb24gMjk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '299',
                                                                                               'links': [       {       'connected_time': '28453948',
                                                                                                                        'fiber_length': 1272,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -30.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjRGcGVlcnBvbnBvbiAyOTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6467 Perth Rd. 131 (Kate Arand)',
                                         'poller_hash': 'Mjk4MjRG',
                                         'serial_number': '29824F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:67',
                                         'uptime': '28453948'},
                                 {       'counter_source': '300/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '300/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '245433310705',
                                                                       'in_ucast_pkts': '1446998907',
                                                                       'mac_address': 'ec:4f:82:59:eb:f3',
                                                                       'name': '300/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '222',
                                                                       'out_octets': '8094891974565',
                                                                       'out_ucast_pkts': '5746842563',
                                                                       'poller_hash': 'MkFEMEYxcGVlcmludGYzMDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '300',
                                                                       'in_nucast_pkts': '222',
                                                                       'in_octets': '8094891974565',
                                                                       'in_ucast_pkts': '5746842563',
                                                                       'links': [       {       'in_nucast_pkts': '222',
                                                                                                'in_octets': '8094891974565',
                                                                                                'in_ucast_pkts': '5746842563',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '245433310705',
                                                                                                'out_ucast_pkts': '1446998907',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '245433310705',
                                                                       'out_ucast_pkts': '1446998907',
                                                                       'poller_hash': 'MkFEMEYxcGVlcmludGZwb24gMzAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '300',
                                                                                               'links': [       {       'connected_time': '688451',
                                                                                                                        'fiber_length': 830,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFEMEYxcGVlcnBvbnBvbiAzMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '47 MILL ST (COXON ROBERT)',
                                         'poller_hash': 'MkFEMEYx',
                                         'serial_number': '2AD0F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:eb:f3',
                                         'uptime': '688451'},
                                 {       'counter_source': '301/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '301/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '270993964124',
                                                                       'in_ucast_pkts': '31434104065',
                                                                       'mac_address': 'ec:4f:82:8d:f7:c0',
                                                                       'name': '301/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '86',
                                                                       'out_octets': '5902432015925',
                                                                       'out_ucast_pkts': '4628420006',
                                                                       'poller_hash': 'MzIyNzNGcGVlcmludGYzMDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '301',
                                                                       'in_nucast_pkts': '86',
                                                                       'in_octets': '5902432015925',
                                                                       'in_ucast_pkts': '4628420006',
                                                                       'links': [       {       'in_nucast_pkts': '86',
                                                                                                'in_octets': '5902432015925',
                                                                                                'in_ucast_pkts': '4628420006',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '270993964124',
                                                                                                'out_ucast_pkts': '31434104065',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '270993964124',
                                                                       'out_ucast_pkts': '31434104065',
                                                                       'poller_hash': 'MzIyNzNGcGVlcmludGZwb24gMzAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '301',
                                                                                               'links': [       {       'connected_time': '21816866',
                                                                                                                        'fiber_length': 5933,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNzNGcGVlcnBvbnBvbiAzMDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6680B Road 140 Milverton (Melanie '
                                                 'Elg 10427044)',
                                         'poller_hash': 'MzIyNzNG',
                                         'serial_number': '32273F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:f7:c0',
                                         'uptime': '21816866'},
                                 {       'counter_source': '302/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '302/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '211580395865',
                                                                       'in_ucast_pkts': '5087828380',
                                                                       'mac_address': 'ec:4f:82:0d:69:1c',
                                                                       'name': '302/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '3358233147162',
                                                                       'out_ucast_pkts': '2714752917',
                                                                       'poller_hash': 'MjA1RENGcGVlcmludGYzMDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '302',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '3358233147162',
                                                                       'in_ucast_pkts': '2714752917',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '3358233147162',
                                                                                                'in_ucast_pkts': '2714752917',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '211580395865',
                                                                                                'out_ucast_pkts': '5087828380',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '211580395865',
                                                                       'out_ucast_pkts': '5087828380',
                                                                       'poller_hash': 'MjA1RENGcGVlcmludGZwb24gMzAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '302',
                                                                                               'links': [       {       'connected_time': '2151357',
                                                                                                                        'fiber_length': 1351,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RENGcGVlcnBvbnBvbiAzMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '65 MAIN ST S (HAHN LEWIS)',
                                         'poller_hash': 'MjA1RENG',
                                         'serial_number': '205DCF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:1c',
                                         'uptime': '2151357'},
                                 {       'counter_source': '305/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '305/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '271686873768',
                                                                       'in_ucast_pkts': '108230933433',
                                                                       'mac_address': 'ec:4f:82:0d:6b:f5',
                                                                       'name': '305/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '125',
                                                                       'out_octets': '2011638086373',
                                                                       'out_ucast_pkts': '1738497463',
                                                                       'poller_hash': 'MjA1RTIwcGVlcmludGYzMDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '305',
                                                                       'in_nucast_pkts': '125',
                                                                       'in_octets': '2011638086373',
                                                                       'in_ucast_pkts': '1738497463',
                                                                       'links': [       {       'in_nucast_pkts': '125',
                                                                                                'in_octets': '2011638086373',
                                                                                                'in_ucast_pkts': '1738497463',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '271686873768',
                                                                                                'out_ucast_pkts': '108230933433',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '271686873768',
                                                                       'out_ucast_pkts': '108230933433',
                                                                       'poller_hash': 'MjA1RTIwcGVlcmludGZwb24gMzA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '305',
                                                                                               'links': [       {       'connected_time': '5394487',
                                                                                                                        'fiber_length': 1049,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RTIwcGVlcnBvbnBvbiAzMDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '13 Mill St.. W. Milverton (Trevor '
                                                 'Woods)',
                                         'poller_hash': 'MjA1RTIw',
                                         'serial_number': '205E20',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6b:f5',
                                         'uptime': '5394487'},
                                 {       'counter_source': '306/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f3:01'],
                                                                       'id': '306/g1',
                                                                       'in_nucast_pkts': '57580',
                                                                       'in_octets': '39830397894',
                                                                       'in_ucast_pkts': '207831380',
                                                                       'mac_address': 'ec:4f:82:4f:7e:12',
                                                                       'mtu': 2000,
                                                                       'name': '306/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6603163',
                                                                       'out_octets': '876659593410',
                                                                       'out_ucast_pkts': '669271231',
                                                                       'poller_hash': 'Mjk4MkM0cGVlcmludGYzMDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '306',
                                                                       'in_nucast_pkts': '6603163',
                                                                       'in_octets': '876659593410',
                                                                       'in_ucast_pkts': '669271231',
                                                                       'links': [       {       'in_nucast_pkts': '6603163',
                                                                                                'in_octets': '876659593410',
                                                                                                'in_ucast_pkts': '669271231',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '57580',
                                                                                                'out_octets': '39830397894',
                                                                                                'out_ucast_pkts': '207831380',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '57580',
                                                                       'out_octets': '39830397894',
                                                                       'out_ucast_pkts': '207831380',
                                                                       'poller_hash': 'Mjk4MkM0cGVlcmludGZwb24gMzA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '306',
                                                                                               'links': [       {       'connected_time': '3113777',
                                                                                                                        'fiber_length': 1202,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MkM0cGVlcnBvbnBvbiAzMDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '59 MAIN ST N (HOLMES  BRANDON)',
                                         'poller_hash': 'Mjk4MkM0',
                                         'serial_number': '2982C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7e:0f',
                                         'uptime': '3113777'},
                                 {       'counter_source': '307/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '307/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '144324544593',
                                                                       'in_ucast_pkts': '1092180976',
                                                                       'mac_address': 'ec:4f:82:4f:84:17',
                                                                       'name': '307/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '4278007633853',
                                                                       'out_ucast_pkts': '3043065068',
                                                                       'poller_hash': 'Mjk4Mzg1cGVlcmludGYzMDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '307',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '4278007633853',
                                                                       'in_ucast_pkts': '3043065068',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '4278007633853',
                                                                                                'in_ucast_pkts': '3043065068',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '144324544593',
                                                                                                'out_ucast_pkts': '1092180976',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '144324544593',
                                                                       'out_ucast_pkts': '1092180976',
                                                                       'poller_hash': 'Mjk4Mzg1cGVlcmludGZwb24gMzA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '307',
                                                                                               'links': [       {       'connected_time': '21816166',
                                                                                                                        'fiber_length': 1329,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4Mzg1cGVlcnBvbnBvbiAzMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6465 Perth Rd. 131 (Peggy '
                                                 'Birmingham)',
                                         'poller_hash': 'Mjk4Mzg1',
                                         'serial_number': '298385',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:84:17',
                                         'uptime': '21816166'},
                                 {       'counter_source': '308/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '308/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '382116362577',
                                                                       'in_ucast_pkts': '572266006',
                                                                       'mac_address': 'ec:4f:82:4f:79:7f',
                                                                       'name': '308/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7993589942051',
                                                                       'out_ucast_pkts': '5552480472',
                                                                       'poller_hash': 'Mjk4MjMycGVlcmludGYzMDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '308',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7993589942051',
                                                                       'in_ucast_pkts': '5552480472',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7993589942051',
                                                                                                'in_ucast_pkts': '5552480472',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '382116362577',
                                                                                                'out_ucast_pkts': '572266006',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '382116362577',
                                                                       'out_ucast_pkts': '572266006',
                                                                       'poller_hash': 'Mjk4MjMycGVlcmludGZwb24gMzA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '308',
                                                                                               'links': [       {       'connected_time': '14582749',
                                                                                                                        'fiber_length': 778,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjMycGVlcnBvbnBvbiAzMDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29 MAIN ST S (SCHWARTZENBERGER ALLAN '
                                                 '10438370)',
                                         'poller_hash': 'Mjk4MjMy',
                                         'serial_number': '298232',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:79:7f',
                                         'uptime': '14582749'},
                                 {       'counter_source': '309/g1+309/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:09:0d:f4:23:fe'],
                                                                       'id': '309/g1',
                                                                       'in_nucast_pkts': '2788115',
                                                                       'in_octets': '81297341914',
                                                                       'in_ucast_pkts': '212271859',
                                                                       'mac_address': 'ec:4f:82:c4:7a:60',
                                                                       'mtu': 2000,
                                                                       'name': '309/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '37930903',
                                                                       'out_octets': '1969352708704',
                                                                       'out_ucast_pkts': '1373230204',
                                                                       'poller_hash': 'MzZDQzJGcGVlcmludGYzMDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:9f:80:13:2f:f3'],
                                                                       'id': '309/g4',
                                                                       'in_nucast_pkts': '53544',
                                                                       'in_octets': '59352318601',
                                                                       'in_ucast_pkts': '204034098',
                                                                       'mac_address': 'ec:4f:82:c4:7a:63',
                                                                       'mtu': 2000,
                                                                       'name': '309/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '37938889',
                                                                       'out_octets': '816388449887',
                                                                       'out_ucast_pkts': '638553051',
                                                                       'poller_hash': 'MzZDQzJGcGVlcmludGYzMDkvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '309',
                                                                       'in_nucast_pkts': '75869792',
                                                                       'in_octets': '2785741158591',
                                                                       'in_ucast_pkts': '2011783255',
                                                                       'links': [       {       'in_nucast_pkts': '75869792',
                                                                                                'in_octets': '2785741158591',
                                                                                                'in_ucast_pkts': '2011783255',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2841659',
                                                                                                'out_octets': '140649660515',
                                                                                                'out_ucast_pkts': '416305957',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2841659',
                                                                       'out_octets': '140649660515',
                                                                       'out_ucast_pkts': '416305957',
                                                                       'poller_hash': 'MzZDQzJGcGVlcmludGZwb24gMzA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '309',
                                                                                               'links': [       {       'connected_time': '21433009',
                                                                                                                        'fiber_length': 986,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.584,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzZDQzJGcGVlcnBvbnBvbiAzMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': "34 MAIN ST N (KAREN'S KITCHEN)",
                                         'poller_hash': 'MzZDQzJG',
                                         'serial_number': '36CC2F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:c4:7a:5c',
                                         'uptime': '21433009'},
                                 {       'counter_source': '310/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '310/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15734926330',
                                                                       'in_ucast_pkts': '98002549',
                                                                       'mac_address': 'ec:4f:82:cb:a7:80',
                                                                       'name': '310/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '34935407655',
                                                                       'out_ucast_pkts': '170937655',
                                                                       'poller_hash': 'MzhGMjJBcGVlcmludGYzMTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '310',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '34935407655',
                                                                       'in_ucast_pkts': '170937655',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '34935407655',
                                                                                                'in_ucast_pkts': '170937655',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '15734926330',
                                                                                                'out_ucast_pkts': '98002549',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15734926330',
                                                                       'out_ucast_pkts': '98002549',
                                                                       'poller_hash': 'MzhGMjJBcGVlcmludGZwb24gMzEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '310',
                                                                                               'links': [       {       'connected_time': '48243',
                                                                                                                        'fiber_length': 1277,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzhGMjJBcGVlcnBvbnBvbiAzMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '15 Spencer St. Milverton (Township '
                                                 'Perth East)',
                                         'poller_hash': 'MzhGMjJB',
                                         'serial_number': '38F22A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cb:a7:80',
                                         'uptime': '48243'},
                                 {       'counter_source': '311/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b4:75:0e:54:f6:19'],
                                                                       'id': '311/g1',
                                                                       'in_nucast_pkts': '30340',
                                                                       'in_octets': '214048507',
                                                                       'in_ucast_pkts': '778014',
                                                                       'mac_address': 'ec:4f:82:7d:b9:43',
                                                                       'mtu': 2000,
                                                                       'name': '311/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464811',
                                                                       'out_octets': '4770205631',
                                                                       'out_ucast_pkts': '3943032',
                                                                       'poller_hash': 'MzAwNjBBcGVlcmludGYzMTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '311',
                                                                       'in_nucast_pkts': '1464811',
                                                                       'in_octets': '4770205631',
                                                                       'in_ucast_pkts': '3943032',
                                                                       'links': [       {       'in_nucast_pkts': '1464811',
                                                                                                'in_octets': '4770205631',
                                                                                                'in_ucast_pkts': '3943032',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '30340',
                                                                                                'out_octets': '214048507',
                                                                                                'out_ucast_pkts': '778014',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30340',
                                                                       'out_octets': '214048507',
                                                                       'out_ucast_pkts': '778014',
                                                                       'poller_hash': 'MzAwNjBBcGVlcmludGZwb24gMzEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '311',
                                                                                               'links': [       {       'connected_time': '688477',
                                                                                                                        'fiber_length': 902,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -6.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwNjBBcGVlcnBvbnBvbiAzMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '17 William St. Milverton (Barry '
                                                 'Brick)',
                                         'poller_hash': 'MzAwNjBB',
                                         'serial_number': '30060A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:b9:43',
                                         'uptime': '688477'},
                                 {       'counter_source': '312/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '312/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2622044806959',
                                                                       'in_ucast_pkts': '2377048695',
                                                                       'mac_address': 'ec:4f:82:be:a7:79',
                                                                       'name': '312/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1670671312911',
                                                                       'out_ucast_pkts': '2188061089',
                                                                       'poller_hash': 'Mzc4ODJDcGVlcmludGYzMTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '312',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1670671312911',
                                                                       'in_ucast_pkts': '2188061089',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1670671312911',
                                                                                                'in_ucast_pkts': '2188061089',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2622044806959',
                                                                                                'out_ucast_pkts': '2377048695',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2622044806959',
                                                                       'out_ucast_pkts': '2377048695',
                                                                       'poller_hash': 'Mzc4ODJDcGVlcmludGZwb24gMzEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '312',
                                                                                               'links': [       {       'connected_time': '47621',
                                                                                                                        'fiber_length': 591,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.698,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzc4ODJDcGVlcnBvbnBvbiAzMTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '9 Main St. N. Milverton (Milv. '
                                                 'Accounting&Oak Fin)',
                                         'poller_hash': 'Mzc4ODJD',
                                         'serial_number': '37882C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:be:a7:79',
                                         'uptime': '47621'},
                                 {       'counter_source': '313/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:7c:3f:0b:45:d8'],
                                                                       'id': '313/g1',
                                                                       'in_nucast_pkts': '3451',
                                                                       'in_octets': '591115546937',
                                                                       'in_ucast_pkts': '2380118253',
                                                                       'mac_address': 'ec:4f:82:7e:39:3d',
                                                                       'mtu': 2000,
                                                                       'name': '313/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33792523',
                                                                       'out_octets': '11761725508562',
                                                                       'out_ucast_pkts': '8202499538',
                                                                       'poller_hash': 'MzVGOTUzcGVlcmludGYzMTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '313',
                                                                       'in_nucast_pkts': '33792523',
                                                                       'in_octets': '11761725508562',
                                                                       'in_ucast_pkts': '8202499538',
                                                                       'links': [       {       'in_nucast_pkts': '33792523',
                                                                                                'in_octets': '11761725508562',
                                                                                                'in_ucast_pkts': '8202499538',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3451',
                                                                                                'out_octets': '591115546937',
                                                                                                'out_ucast_pkts': '2380118253',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3451',
                                                                       'out_octets': '591115546937',
                                                                       'out_ucast_pkts': '2380118253',
                                                                       'poller_hash': 'MzVGOTUzcGVlcmludGZwb24gMzEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '313',
                                                                                               'links': [       {       'connected_time': '15749555',
                                                                                                                        'fiber_length': 18168,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOTUzcGVlcnBvbnBvbiAzMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '82 MADDISON ST E (COCHRANE  DEREK)',
                                         'poller_hash': 'MzVGOTUz',
                                         'serial_number': '35F953',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:39:3d',
                                         'uptime': '15749555'},
                                 {       'counter_source': '315/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '315/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '69886258408',
                                                                       'in_ucast_pkts': '257444202',
                                                                       'mac_address': 'cc:be:59:29:04:8f',
                                                                       'name': '315/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '876515942884',
                                                                       'out_ucast_pkts': '807027882',
                                                                       'poller_hash': 'Mzk0MERFcGVlcmludGYzMTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '315',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '876515942884',
                                                                       'in_ucast_pkts': '807027882',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '876515942884',
                                                                                                'in_ucast_pkts': '807027882',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '69886258408',
                                                                                                'out_ucast_pkts': '257444202',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '69886258408',
                                                                       'out_ucast_pkts': '257444202',
                                                                       'poller_hash': 'Mzk0MERFcGVlcmludGZwb24gMzE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '315',
                                                                                               'links': [       {       'connected_time': '47915',
                                                                                                                        'fiber_length': 17535,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzk0MERFcGVlcnBvbnBvbiAzMTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '159 Fishleigh St. Monkton (Henry '
                                                 'Vander Hyden)',
                                         'poller_hash': 'Mzk0MERF',
                                         'serial_number': '3940DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:29:04:8f',
                                         'uptime': '47915'},
                                 {       'counter_source': '316/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '316/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '302345920939',
                                                                       'in_ucast_pkts': '310494896',
                                                                       'mac_address': 'ec:4f:82:92:ff:88',
                                                                       'name': '316/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '296127375638',
                                                                       'out_ucast_pkts': '424845624',
                                                                       'poller_hash': 'MzI5NjdEcGVlcmludGYzMTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '316',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '296127375638',
                                                                       'in_ucast_pkts': '424845624',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '296127375638',
                                                                                                'in_ucast_pkts': '424845624',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '302345920939',
                                                                                                'out_ucast_pkts': '310494896',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '302345920939',
                                                                       'out_ucast_pkts': '310494896',
                                                                       'poller_hash': 'MzI5NjdEcGVlcmludGZwb24gMzE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '316',
                                                                                               'links': [       {       'connected_time': '28950839',
                                                                                                                        'fiber_length': 18041,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.13,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzI5NjdEcGVlcnBvbnBvbiAzMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '34 MADDISON ST E MONKTON (FELLOWSHIP '
                                                 'HAVEN MENNO)',
                                         'poller_hash': 'MzI5NjdE',
                                         'serial_number': '32967D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:92:ff:88',
                                         'uptime': '28950839'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '318/g1',
                                                                       'name': '318/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkZFQjYycGVlcmludGYzMTgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '318',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MkZFQjYycGVlcmludGZwb24gMzE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '318',
                                                                                               'links': [       {       'connected_time': '3113969',
                                                                                                                        'fiber_length': 18485,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.966,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQjYycGVlcnBvbnBvbiAzMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '138 Maddison St. E. Monkton',
                                         'poller_hash': 'MkZFQjYy',
                                         'serial_number': '2FEB62',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:25:c0',
                                         'uptime': '3113969'},
                                 {       'counter_source': '319/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '319/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '392723519514',
                                                                       'in_ucast_pkts': '2912686552',
                                                                       'mac_address': 'cc:be:59:70:80:ee',
                                                                       'name': '319/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11078120275692',
                                                                       'out_ucast_pkts': '7778721000',
                                                                       'poller_hash': 'NTYwQzEwcGVlcmludGYzMTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '319',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11078120275692',
                                                                       'in_ucast_pkts': '7778721000',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11078120275692',
                                                                                                'in_ucast_pkts': '7778721000',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '392723519514',
                                                                                                'out_ucast_pkts': '2912686552',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '392723519514',
                                                                       'out_ucast_pkts': '2912686552',
                                                                       'poller_hash': 'NTYwQzEwcGVlcmludGZwb24gMzE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '319',
                                                                                               'links': [       {       'connected_time': '23325143',
                                                                                                                        'fiber_length': 17825,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYwQzEwcGVlcnBvbnBvbiAzMTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '164 Smith St. Monkton (Bob Mclean)',
                                         'poller_hash': 'NTYwQzEw',
                                         'serial_number': '560C10',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:70:80:ee',
                                         'uptime': '23325143'},
                                 {       'counter_source': '320/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '320/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2293733100764',
                                                                       'in_ucast_pkts': '7045777794',
                                                                       'mac_address': 'ec:4f:82:cb:aa:58',
                                                                       'name': '320/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11273286154577',
                                                                       'out_ucast_pkts': '13982576845',
                                                                       'poller_hash': 'MzhGMjg1cGVlcmludGYzMjAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '320',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11273286154577',
                                                                       'in_ucast_pkts': '13982576845',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11273286154577',
                                                                                                'in_ucast_pkts': '13982576845',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2293733100764',
                                                                                                'out_ucast_pkts': '7045777794',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2293733100764',
                                                                       'out_ucast_pkts': '7045777794',
                                                                       'poller_hash': 'MzhGMjg1cGVlcmludGZwb24gMzIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '320',
                                                                                               'links': [       {       'connected_time': '47622',
                                                                                                                        'fiber_length': 17480,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzhGMjg1cGVlcnBvbnBvbiAzMjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '205 Winstanley St. Monkton (TJ Choi)',
                                         'poller_hash': 'MzhGMjg1',
                                         'serial_number': '38F285',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cb:aa:58',
                                         'uptime': '47622'},
                                 {       'counter_source': '321/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '321/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1082939784236',
                                                                       'in_ucast_pkts': '3128667823',
                                                                       'mac_address': 'ec:4f:82:32:26:1f',
                                                                       'name': '321/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15222932335505',
                                                                       'out_ucast_pkts': '11835610199',
                                                                       'poller_hash': 'MjQ4QUQ0cGVlcmludGYzMjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '321',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15222932335505',
                                                                       'in_ucast_pkts': '11835610199',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '15222932335505',
                                                                                                'in_ucast_pkts': '11835610199',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1082939784236',
                                                                                                'out_ucast_pkts': '3128667823',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1082939784236',
                                                                       'out_ucast_pkts': '3128667823',
                                                                       'poller_hash': 'MjQ4QUQ0cGVlcmludGZwb24gMzIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '321',
                                                                                               'links': [       {       'connected_time': '26736474',
                                                                                                                        'fiber_length': 18562,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.854,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjQ4QUQ0cGVlcnBvbnBvbiAzMjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': 'Derick Stolz 10430219 (52 schade st)',
                                         'poller_hash': 'MjQ4QUQ0',
                                         'serial_number': '248AD4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:32:26:1f',
                                         'uptime': '26736474'},
                                 {       'counter_source': '322/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:e3:85'],
                                                                       'id': '322/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '4954682862',
                                                                       'in_ucast_pkts': '13157805',
                                                                       'mac_address': 'ec:4f:82:5b:92:5e',
                                                                       'mtu': 2000,
                                                                       'name': '322/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447074',
                                                                       'out_octets': '51313001043',
                                                                       'out_ucast_pkts': '41451707',
                                                                       'poller_hash': 'MkIwNUJFcGVlcmludGYzMjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '322',
                                                                       'in_nucast_pkts': '1447074',
                                                                       'in_octets': '51313001043',
                                                                       'in_ucast_pkts': '41451707',
                                                                       'links': [       {       'in_nucast_pkts': '1447074',
                                                                                                'in_octets': '51313001043',
                                                                                                'in_ucast_pkts': '41451707',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '4954682862',
                                                                                                'out_ucast_pkts': '13157805',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '4954682862',
                                                                       'out_ucast_pkts': '13157805',
                                                                       'poller_hash': 'MkIwNUJFcGVlcmludGZwb24gMzIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '322',
                                                                                               'links': [       {       'connected_time': '688406',
                                                                                                                        'fiber_length': 2248,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNUJFcGVlcnBvbnBvbiAzMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '68 GERBER DR',
                                         'poller_hash': 'MkIwNUJF',
                                         'serial_number': '2B05BE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:92:5b',
                                         'uptime': '688406'},
                                 {       'counter_source': '323/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '323/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1033339023230',
                                                                       'in_ucast_pkts': '10299356035',
                                                                       'mac_address': 'cc:be:59:70:33:2e',
                                                                       'name': '323/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '32784323960311',
                                                                       'out_ucast_pkts': '22347479799',
                                                                       'poller_hash': 'NTYwMjU4cGVlcmludGYzMjMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '323',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '32784323960311',
                                                                       'in_ucast_pkts': '22347479799',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '32784323960311',
                                                                                                'in_ucast_pkts': '22347479799',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1033339023230',
                                                                                                'out_ucast_pkts': '10299356035',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1033339023230',
                                                                       'out_ucast_pkts': '10299356035',
                                                                       'poller_hash': 'NTYwMjU4cGVlcmludGZwb24gMzIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '323',
                                                                                               'links': [       {       'connected_time': '66690374',
                                                                                                                        'fiber_length': 17416,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.308,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYwMjU4cGVlcnBvbnBvbiAzMjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '185 Winstanley St. Monkton (Becky '
                                                 'Beuerman)',
                                         'poller_hash': 'NTYwMjU4',
                                         'serial_number': '560258',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:70:33:2e',
                                         'uptime': '66690374'},
                                 {       'counter_source': '324/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '324/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '442597800125',
                                                                       'in_ucast_pkts': '1825637601',
                                                                       'mac_address': 'ec:4f:82:b9:25:98',
                                                                       'name': '324/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6187844234126',
                                                                       'out_ucast_pkts': '4555944444',
                                                                       'poller_hash': 'MkZFQjVEcGVlcmludGYzMjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '324',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6187844234126',
                                                                       'in_ucast_pkts': '4555944444',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6187844234126',
                                                                                                'in_ucast_pkts': '4555944444',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '442597800125',
                                                                                                'out_ucast_pkts': '1825637601',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '442597800125',
                                                                       'out_ucast_pkts': '1825637601',
                                                                       'poller_hash': 'MkZFQjVEcGVlcmludGZwb24gMzI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '324',
                                                                                               'links': [       {       'connected_time': '48182',
                                                                                                                        'fiber_length': 17816,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.302,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQjVEcGVlcnBvbnBvbiAzMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '25 Maddison St. W. Monkton (Dean '
                                                 'Smith)',
                                         'poller_hash': 'MkZFQjVE',
                                         'serial_number': '2FEB5D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:25:98',
                                         'uptime': '48182'},
                                 {       'counter_source': '325/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:99:4d:4b'],
                                                                       'id': '325/g1',
                                                                       'in_nucast_pkts': '16536092',
                                                                       'in_octets': '387461055298',
                                                                       'in_ucast_pkts': '3174687259',
                                                                       'mac_address': '00:06:31:4c:15:6d',
                                                                       'mtu': 2000,
                                                                       'name': '325/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20946732413',
                                                                       'out_octets': '37550044658505',
                                                                       'out_ucast_pkts': '6985138362',
                                                                       'poller_hash': 'ODc0MTRwZWVyaW50ZjMyNS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '325',
                                                                       'in_nucast_pkts': '20946732413',
                                                                       'in_octets': '37550044658505',
                                                                       'in_ucast_pkts': '6985138362',
                                                                       'links': [       {       'in_nucast_pkts': '20946732413',
                                                                                                'in_octets': '37550044658505',
                                                                                                'in_ucast_pkts': '6985138362',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16536092',
                                                                                                'out_octets': '387461055298',
                                                                                                'out_ucast_pkts': '3174687259',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16536092',
                                                                       'out_octets': '387461055298',
                                                                       'out_ucast_pkts': '3174687259',
                                                                       'poller_hash': 'ODc0MTRwZWVyaW50ZnBvbiAzMjU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '325',
                                                                                               'links': [       {       'connected_time': '24669203',
                                                                                                                        'fiber_length': 17711,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODc0MTRwZWVycG9ucG9uIDMyNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '148 Winstanley St. Monkton (Stan & '
                                                 'Glenda Stolz 10430578)',
                                         'poller_hash': 'ODc0MTQ=',
                                         'serial_number': '87414',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:d4:fc',
                                         'uptime': '24669203'},
                                 {       'counter_source': '329/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '329/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2572407078465',
                                                                       'in_ucast_pkts': '3771613008',
                                                                       'mac_address': '48:77:46:a2:1f:5c',
                                                                       'name': '329/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2819861003305',
                                                                       'out_ucast_pkts': '4663743411',
                                                                       'poller_hash': 'NzFBNTFBcGVlcmludGYzMjkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '329',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2819861003305',
                                                                       'in_ucast_pkts': '4663743411',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2819861003305',
                                                                                                'in_ucast_pkts': '4663743411',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2572407078465',
                                                                                                'out_ucast_pkts': '3771613008',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2572407078465',
                                                                       'out_ucast_pkts': '3771613008',
                                                                       'poller_hash': 'NzFBNTFBcGVlcmludGZwb24gMzI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '329',
                                                                                               'links': [       {       'connected_time': '47457',
                                                                                                                        'fiber_length': 18245,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.446,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFBNTFBcGVlcnBvbnBvbiAzMjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '124 Maddison St. W. Monkton (Walsh '
                                                 'Truck 10430198)',
                                         'poller_hash': 'NzFBNTFB',
                                         'serial_number': '71A51A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a2:1f:5c',
                                         'uptime': '47457'},
                                 {       'counter_source': '330/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '330/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '498755433614',
                                                                       'in_ucast_pkts': '3688804242',
                                                                       'mac_address': 'ec:4f:82:b9:19:d8',
                                                                       'name': '330/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '16121703744420',
                                                                       'out_ucast_pkts': '7173665840',
                                                                       'poller_hash': 'MkZFOUU1cGVlcmludGYzMzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '330',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '16121703744420',
                                                                       'in_ucast_pkts': '7173665840',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '16121703744420',
                                                                                                'in_ucast_pkts': '7173665840',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '498755433614',
                                                                                                'out_ucast_pkts': '3688804242',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '498755433614',
                                                                       'out_ucast_pkts': '3688804242',
                                                                       'poller_hash': 'MkZFOUU1cGVlcmludGZwb24gMzMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '330',
                                                                                               'links': [       {       'connected_time': '13642953',
                                                                                                                        'fiber_length': 17555,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.338,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFOUU1cGVlcnBvbnBvbiAzMzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '175 Winstanley St. Monkton (Vince '
                                                 'Priestrap)',
                                         'poller_hash': 'MkZFOUU1',
                                         'serial_number': '2FE9E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:19:d8',
                                         'uptime': '13642953'},
                                 {       'counter_source': '331/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b4:4f'],
                                                                       'id': '331/g1',
                                                                       'in_nucast_pkts': '49978',
                                                                       'in_octets': '21622402510',
                                                                       'in_ucast_pkts': '97539488',
                                                                       'mac_address': 'ec:4f:82:b9:1b:1c',
                                                                       'mtu': 2000,
                                                                       'name': '331/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4076844',
                                                                       'out_octets': '462007352492',
                                                                       'out_ucast_pkts': '343590504',
                                                                       'poller_hash': 'MkZFQTBEcGVlcmludGYzMzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '331',
                                                                       'in_nucast_pkts': '4076844',
                                                                       'in_octets': '462007352492',
                                                                       'in_ucast_pkts': '343590504',
                                                                       'links': [       {       'in_nucast_pkts': '4076844',
                                                                                                'in_octets': '462007352492',
                                                                                                'in_ucast_pkts': '343590504',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '49978',
                                                                                                'out_octets': '21622402510',
                                                                                                'out_ucast_pkts': '97539488',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49978',
                                                                       'out_octets': '21622402510',
                                                                       'out_ucast_pkts': '97539488',
                                                                       'poller_hash': 'MkZFQTBEcGVlcmludGZwb24gMzMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '331',
                                                                                               'links': [       {       'connected_time': '2443388',
                                                                                                                        'fiber_length': 18472,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.542,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQTBEcGVlcnBvbnBvbiAzMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '163 BROCK STREET MONKTON (GREGORY '
                                                 'DAWN)',
                                         'poller_hash': 'MkZFQTBE',
                                         'serial_number': '2FEA0D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:1b:18',
                                         'uptime': '2443388'},
                                 {       'counter_source': '334/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '334/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '194649737388',
                                                                       'in_ucast_pkts': '781157869',
                                                                       'mac_address': 'ec:4f:82:cc:5e:0c',
                                                                       'name': '334/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2881960485554',
                                                                       'out_ucast_pkts': '2436212660',
                                                                       'poller_hash': 'MzkwOEY5cGVlcmludGYzMzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '334',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2881960485554',
                                                                       'in_ucast_pkts': '2436212660',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2881960485554',
                                                                                                'in_ucast_pkts': '2436212660',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '194649737388',
                                                                                                'out_ucast_pkts': '781157869',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '194649737388',
                                                                       'out_ucast_pkts': '781157869',
                                                                       'poller_hash': 'MzkwOEY5cGVlcmludGZwb24gMzM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '334',
                                                                                               'links': [       {       'connected_time': '29417607',
                                                                                                                        'fiber_length': 17812,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.206,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzkwOEY5cGVlcnBvbnBvbiAzMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '33 Maddison St. W. (Paul Ward)',
                                         'poller_hash': 'MzkwOEY5',
                                         'serial_number': '3908F9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cc:5e:0c',
                                         'uptime': '29417607'},
                                 {       'counter_source': '336/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '336/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '932200058794',
                                                                       'in_ucast_pkts': '5347896874',
                                                                       'mac_address': 'ec:4f:82:b9:18:f8',
                                                                       'name': '336/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '17436861497665',
                                                                       'out_ucast_pkts': '12664301699',
                                                                       'poller_hash': 'MkZFOUM5cGVlcmludGYzMzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '336',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '17436861497665',
                                                                       'in_ucast_pkts': '12664301699',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '17436861497665',
                                                                                                'in_ucast_pkts': '12664301699',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '932200058794',
                                                                                                'out_ucast_pkts': '5347896874',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '932200058794',
                                                                       'out_ucast_pkts': '5347896874',
                                                                       'poller_hash': 'MkZFOUM5cGVlcmludGZwb24gMzM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '336',
                                                                                               'links': [       {       'connected_time': '48400',
                                                                                                                        'fiber_length': 17819,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.172,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFOUM5cGVlcnBvbnBvbiAzMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '37 Maddison St. W. Monkton (Amanda '
                                                 'Pfeifer)',
                                         'poller_hash': 'MkZFOUM5',
                                         'serial_number': '2FE9C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:18:f8',
                                         'uptime': '48400'},
                                 {       'counter_source': '339/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:27'],
                                                                       'id': '339/g1',
                                                                       'in_nucast_pkts': '181054',
                                                                       'in_octets': '139839145579',
                                                                       'in_ucast_pkts': '1091254656',
                                                                       'mac_address': '48:77:46:ce:41:0a',
                                                                       'mtu': 2000,
                                                                       'name': '339/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16921922',
                                                                       'out_octets': '3989707791574',
                                                                       'out_ucast_pkts': '2925202328',
                                                                       'poller_hash': 'NzZFRENEcGVlcmludGYzMzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '339',
                                                                       'in_nucast_pkts': '16921922',
                                                                       'in_octets': '3989707791574',
                                                                       'in_ucast_pkts': '2925202328',
                                                                       'links': [       {       'in_nucast_pkts': '16921922',
                                                                                                'in_octets': '3989707791574',
                                                                                                'in_ucast_pkts': '2925202328',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '181054',
                                                                                                'out_octets': '139839145579',
                                                                                                'out_ucast_pkts': '1091254656',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '181054',
                                                                       'out_octets': '139839145579',
                                                                       'out_ucast_pkts': '1091254656',
                                                                       'poller_hash': 'NzZFRENEcGVlcmludGZwb24gMzM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '339',
                                                                                               'links': [       {       'connected_time': '9506531',
                                                                                                                        'fiber_length': 18068,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzZFRENEcGVlcnBvbnBvbiAzMzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '189 QUEEN AVENUE MONKTON (MORRISON '
                                                 'COLTON)',
                                         'poller_hash': 'NzZFRENE',
                                         'serial_number': '76EDCD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:ce:41:06',
                                         'uptime': '9506531'},
                                 {       'counter_source': '340/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '340/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '161185898864',
                                                                       'in_ucast_pkts': '1572099606',
                                                                       'mac_address': '48:77:46:a2:1e:3c',
                                                                       'name': '340/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4441347647900',
                                                                       'out_ucast_pkts': '4744238257',
                                                                       'poller_hash': 'NzFBNEY2cGVlcmludGYzNDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '340',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4441347647900',
                                                                       'in_ucast_pkts': '4744238257',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4441347647900',
                                                                                                'in_ucast_pkts': '4744238257',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '161185898864',
                                                                                                'out_ucast_pkts': '1572099606',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '161185898864',
                                                                       'out_ucast_pkts': '1572099606',
                                                                       'poller_hash': 'NzFBNEY2cGVlcmludGZwb24gMzQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '340',
                                                                                               'links': [       {       'connected_time': '47651',
                                                                                                                        'fiber_length': 17366,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.58,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFBNEY2cGVlcnBvbnBvbiAzNDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '186 Selby Ave. Monkton (YUNDT JUD)',
                                         'poller_hash': 'NzFBNEY2',
                                         'serial_number': '71A4F6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a2:1e:3c',
                                         'uptime': '47651'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '341/g1',
                                                                       'name': '341/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDE4QjBwZWVyaW50ZjM0MS9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '341',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDE4QjBwZWVyaW50ZnBvbiAzNDE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '341',
                                                                                               'links': [       {       'connected_time': '24664576',
                                                                                                                        'fiber_length': 17701,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDE4QjBwZWVycG9ucG9uIDM0MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712GX',
                                         'name': '146 Winstanley St. Monkton (Robert '
                                                 'Straeten)',
                                         'poller_hash': 'NDE4QjA=',
                                         'serial_number': '418B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:20:02:f8',
                                         'uptime': '24664576'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '342/g1',
                                                                       'name': '342/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QkYyNDI5cGVlcmludGYzNDIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '342',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QkYyNDI5cGVlcmludGZwb24gMzQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '342',
                                                                                               'links': [       {       'connected_time': '47254',
                                                                                                                        'fiber_length': 18232,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QkYyNDI5cGVlcnBvbnBvbiAzNDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '43 SCHADE ST (DUNCAN KELLY)',
                                         'poller_hash': 'QkYyNDI5',
                                         'serial_number': 'BF2429',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'f8:85:f9:6f:1b:d5',
                                         'uptime': '47254'},
                                 {       'counter_source': '343/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '343/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1243390822295',
                                                                       'in_ucast_pkts': '153431345162',
                                                                       'mac_address': 'ec:4f:82:5b:6e:eb',
                                                                       'name': '343/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '82',
                                                                       'out_octets': '9486556958500',
                                                                       'out_ucast_pkts': '7359473781',
                                                                       'poller_hash': 'MkIwMTUwcGVlcmludGYzNDMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '343',
                                                                       'in_nucast_pkts': '82',
                                                                       'in_octets': '9486556958500',
                                                                       'in_ucast_pkts': '7359473781',
                                                                       'links': [       {       'in_nucast_pkts': '82',
                                                                                                'in_octets': '9486556958500',
                                                                                                'in_ucast_pkts': '7359473781',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1243390822295',
                                                                                                'out_ucast_pkts': '153431345162',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1243390822295',
                                                                       'out_ucast_pkts': '153431345162',
                                                                       'poller_hash': 'MkIwMTUwcGVlcmludGZwb24gMzQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '343',
                                                                                               'links': [       {       'connected_time': '21816299',
                                                                                                                        'fiber_length': 20176,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTUwcGVlcnBvbnBvbiAzNDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5734 LINE 57 (VAN NULAND MARIA '
                                                 '(MELANIE))',
                                         'poller_hash': 'MkIwMTUw',
                                         'serial_number': '2B0150',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:6e:eb',
                                         'uptime': '21816299'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '344/g1',
                                                                       'name': '344/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwRTcycGVlcmludGYzNDQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '344',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwRTcycGVlcmludGZwb24gMzQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '344',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRTcycGVlcnBvbnBvbiAzNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'test',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "344" '
                                                                  '(location: GPON '
                                                                  'port "13/8"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:53.960Z',
                                         'poller_hash': 'MzAwRTcy',
                                         'serial_number': '300E72',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '345/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '345/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '626969060488',
                                                                       'in_ucast_pkts': '4086073607',
                                                                       'mac_address': '48:77:46:ce:a9:b6',
                                                                       'name': '345/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7127650130313',
                                                                       'out_ucast_pkts': '9825480261',
                                                                       'poller_hash': 'ODQ1OTA5cGVlcmludGYzNDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '345',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7127650130313',
                                                                       'in_ucast_pkts': '9825480261',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7127650130313',
                                                                                                'in_ucast_pkts': '9825480261',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '626969060488',
                                                                                                'out_ucast_pkts': '4086073607',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '626969060488',
                                                                       'out_ucast_pkts': '4086073607',
                                                                       'poller_hash': 'ODQ1OTA5cGVlcmludGZwb24gMzQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '345',
                                                                                               'links': [       {       'connected_time': '47731',
                                                                                                                        'fiber_length': 17666,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.074,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ1OTA5cGVlcnBvbnBvbiAzNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '211 SELBY AV (HENRY NATHAN)',
                                         'poller_hash': 'ODQ1OTA5',
                                         'serial_number': '845909',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:ce:a9:b6',
                                         'uptime': '47731'},
                                 {       'counter_source': '346/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '346/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '330205306008',
                                                                       'in_ucast_pkts': '1890253469',
                                                                       'mac_address': '48:77:46:3b:c5:aa',
                                                                       'name': '346/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5195624897762',
                                                                       'out_ucast_pkts': '4853551418',
                                                                       'poller_hash': 'NzI0QUQwcGVlcmludGYzNDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '346',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5195624897762',
                                                                       'in_ucast_pkts': '4853551418',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '5195624897762',
                                                                                                'in_ucast_pkts': '4853551418',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '330205306008',
                                                                                                'out_ucast_pkts': '1890253469',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '330205306008',
                                                                       'out_ucast_pkts': '1890253469',
                                                                       'poller_hash': 'NzI0QUQwcGVlcmludGZwb24gMzQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '346',
                                                                                               'links': [       {       'connected_time': '26425410',
                                                                                                                        'fiber_length': 18199,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzI0QUQwcGVlcnBvbnBvbiAzNDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6208 Rd. 164 Monkton (Ginny White)',
                                         'poller_hash': 'NzI0QUQw',
                                         'serial_number': '724AD0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:3b:c5:aa',
                                         'uptime': '26425410'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '348/g1',
                                                                       'name': '348/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1REM1cGVlcmludGYzNDgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '348',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1REM1cGVlcmludGZwb24gMzQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '348',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REM1cGVlcnBvbnBvbiAzNDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '26 MAIN ST N',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "348" '
                                                                  '(location: GPON '
                                                                  'port "8/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:09:57.320Z',
                                         'poller_hash': 'MjA1REM1',
                                         'serial_number': '205DC5',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '350/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2d:32:6b'],
                                                                       'id': '350/g1',
                                                                       'in_nucast_pkts': '495',
                                                                       'in_octets': '323146291294',
                                                                       'in_ucast_pkts': '1121148673',
                                                                       'mac_address': 'ec:4f:82:4b:b6:ef',
                                                                       'mtu': 2000,
                                                                       'name': '350/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '78317630',
                                                                       'out_octets': '3167032182499',
                                                                       'out_ucast_pkts': '2622528587',
                                                                       'poller_hash': 'MjU5ODJBcGVlcmludGYzNTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '350',
                                                                       'in_nucast_pkts': '78317630',
                                                                       'in_octets': '3167032182499',
                                                                       'in_ucast_pkts': '2622528587',
                                                                       'links': [       {       'in_nucast_pkts': '78317630',
                                                                                                'in_octets': '3167032182499',
                                                                                                'in_ucast_pkts': '2622528587',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '495',
                                                                                                'out_octets': '323146291294',
                                                                                                'out_ucast_pkts': '1121148673',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '495',
                                                                       'out_octets': '323146291294',
                                                                       'out_ucast_pkts': '1121148673',
                                                                       'poller_hash': 'MjU5ODJBcGVlcmludGZwb24gMzUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '350',
                                                                                               'links': [       {       'connected_time': '41653',
                                                                                                                        'fiber_length': 992,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.37,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjU5ODJBcGVlcnBvbnBvbiAzNTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '36 Main St. N. Milverton (Milverton '
                                                 'Foodtown)',
                                         'poller_hash': 'MjU5ODJB',
                                         'serial_number': '25982A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:4b:b6:eb',
                                         'uptime': '41653'},
                                 {       'counter_source': '351/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '351/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2596219143096',
                                                                       'in_ucast_pkts': '3119489101',
                                                                       'mac_address': 'ec:4f:82:54:e7:d7',
                                                                       'name': '351/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3246989932726',
                                                                       'out_ucast_pkts': '3605181533',
                                                                       'poller_hash': 'MkExQUQzcGVlcmludGYzNTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '351',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3246989932726',
                                                                       'in_ucast_pkts': '3605181533',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3246989932726',
                                                                                                'in_ucast_pkts': '3605181533',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2596219143096',
                                                                                                'out_ucast_pkts': '3119489101',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2596219143096',
                                                                       'out_ucast_pkts': '3119489101',
                                                                       'poller_hash': 'MkExQUQzcGVlcmludGZwb24gMzUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '351',
                                                                                               'links': [       {       'connected_time': '47858',
                                                                                                                        'fiber_length': 1923,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.856,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkExQUQzcGVlcnBvbnBvbiAzNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '28 Main St. N. Milverton (Milverton '
                                                 'Family Dental)',
                                         'poller_hash': 'MkExQUQz',
                                         'serial_number': '2A1AD3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:54:e7:d7',
                                         'uptime': '47858'},
                                 {       'counter_source': '352/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '352/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '364811589803',
                                                                       'in_ucast_pkts': '3185093354',
                                                                       'mac_address': 'ec:4f:82:e0:0a:d8',
                                                                       'name': '352/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15796985294510',
                                                                       'out_ucast_pkts': '10828685213',
                                                                       'poller_hash': 'Mzk2MUE4cGVlcmludGYzNTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '352',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15796985294510',
                                                                       'in_ucast_pkts': '10828685213',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '15796985294510',
                                                                                                'in_ucast_pkts': '10828685213',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '364811589803',
                                                                                                'out_ucast_pkts': '3185093354',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '364811589803',
                                                                       'out_ucast_pkts': '3185093354',
                                                                       'poller_hash': 'Mzk2MUE4cGVlcmludGZwb24gMzUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '352',
                                                                                               'links': [       {       'connected_time': '13216626',
                                                                                                                        'fiber_length': 1224,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.57,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzk2MUE4cGVlcnBvbnBvbiAzNTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '7 SPENCER ST (MARK JUTZI FUNERAL '
                                                 'HOME)',
                                         'poller_hash': 'Mzk2MUE4',
                                         'serial_number': '3961A8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:e0:0a:d8',
                                         'uptime': '13216626'},
                                 {       'counter_source': '353/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '353/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '120735094787',
                                                                       'in_ucast_pkts': '1115828998',
                                                                       'mac_address': '84:d3:43:7d:50:3f',
                                                                       'name': '353/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4439568775719',
                                                                       'out_ucast_pkts': '3065595611',
                                                                       'poller_hash': 'QzgxRjcxcGVlcmludGYzNTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '353',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4439568775719',
                                                                       'in_ucast_pkts': '3065595611',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4439568775719',
                                                                                                'in_ucast_pkts': '3065595611',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '120735094787',
                                                                                                'out_ucast_pkts': '1115828998',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '120735094787',
                                                                       'out_ucast_pkts': '1115828998',
                                                                       'poller_hash': 'QzgxRjcxcGVlcmludGZwb24gMzUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '353',
                                                                                               'links': [       {       'connected_time': '3113835',
                                                                                                                        'fiber_length': 18116,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRjcxcGVlcnBvbnBvbiAzNTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '73 Maddison St. E. Monkton (Nancy '
                                                 'Mcgoldrick)',
                                         'poller_hash': 'QzgxRjcx',
                                         'serial_number': 'C81F71',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:50:3f',
                                         'uptime': '3113835'},
                                 {       'counter_source': '354/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '354/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '842438257332',
                                                                       'in_ucast_pkts': '4181290831',
                                                                       'mac_address': 'd0:76:8f:08:16:7b',
                                                                       'name': '354/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '16405775303258',
                                                                       'out_ucast_pkts': '12479383066',
                                                                       'poller_hash': 'ODY5M0M1cGVlcmludGYzNTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '354',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '16405775303258',
                                                                       'in_ucast_pkts': '12479383066',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '16405775303258',
                                                                                                'in_ucast_pkts': '12479383066',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '842438257332',
                                                                                                'out_ucast_pkts': '4181290831',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '842438257332',
                                                                       'out_ucast_pkts': '4181290831',
                                                                       'poller_hash': 'ODY5M0M1cGVlcmludGZwb24gMzU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '354',
                                                                                               'links': [       {       'connected_time': '585723',
                                                                                                                        'fiber_length': 17990,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0M1cGVlcnBvbnBvbiAzNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'THOMAS DAHM (1-3 MADDISON ST E '
                                                 'MONKTON)',
                                         'poller_hash': 'ODY5M0M1',
                                         'serial_number': '8693C5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:7b',
                                         'uptime': '585723'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '355/g1',
                                                                       'name': '355/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'N0UwRTlwZWVyaW50ZjM1NS9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '355',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'N0UwRTlwZWVyaW50ZnBvbiAzNTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '355',
                                                                                               'links': [       {       'connected_time': '368751',
                                                                                                                        'fiber_length': 18284,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'N0UwRTlwZWVycG9ucG9uIDM1NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '47 Schade St. Monkton (Fay Baker)',
                                         'poller_hash': 'N0UwRTk=',
                                         'serial_number': '7E0E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:41:cf',
                                         'uptime': '368751'},
                                 {       'counter_source': '356/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:92:d2'],
                                                                       'id': '356/g1',
                                                                       'in_nucast_pkts': '59647',
                                                                       'in_octets': '35052492275',
                                                                       'in_ucast_pkts': '329072288',
                                                                       'mac_address': 'ec:4f:82:96:69:2c',
                                                                       'mtu': 2000,
                                                                       'name': '356/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5128759',
                                                                       'out_octets': '964042991148',
                                                                       'out_ucast_pkts': '662333039',
                                                                       'poller_hash': 'MzMwM0IxcGVlcmludGYzNTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '356',
                                                                       'in_nucast_pkts': '5128759',
                                                                       'in_octets': '964042991148',
                                                                       'in_ucast_pkts': '662333039',
                                                                       'links': [       {       'in_nucast_pkts': '5128759',
                                                                                                'in_octets': '964042991148',
                                                                                                'in_ucast_pkts': '662333039',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '59647',
                                                                                                'out_octets': '35052492275',
                                                                                                'out_ucast_pkts': '329072288',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '59647',
                                                                       'out_octets': '35052492275',
                                                                       'out_ucast_pkts': '329072288',
                                                                       'poller_hash': 'MzMwM0IxcGVlcmludGZwb24gMzU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '356',
                                                                                               'links': [       {       'connected_time': '3113830',
                                                                                                                        'fiber_length': 17980,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.262,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzMwM0IxcGVlcnBvbnBvbiAzNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '31 MADDISON ST E (DOHERTY KERRI)',
                                         'poller_hash': 'MzMwM0Ix',
                                         'serial_number': '3303B1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:96:69:28',
                                         'uptime': '3113830'},
                                 {       'counter_source': '357/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '357/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7030615822267',
                                                                       'in_ucast_pkts': '10469450621',
                                                                       'mac_address': 'cc:be:59:eb:38:66',
                                                                       'name': '357/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10740289541826',
                                                                       'out_ucast_pkts': '339491145',
                                                                       'poller_hash': 'NTU2MzgycGVlcmludGYzNTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '357',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10740289541826',
                                                                       'in_ucast_pkts': '339491145',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '10740289541826',
                                                                                                'in_ucast_pkts': '339491145',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '7030615822267',
                                                                                                'out_ucast_pkts': '10469450621',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7030615822267',
                                                                       'out_ucast_pkts': '10469450621',
                                                                       'poller_hash': 'NTU2MzgycGVlcmludGZwb24gMzU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '357',
                                                                                               'links': [       {       'connected_time': '29355974',
                                                                                                                        'fiber_length': 18459,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.652,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU2MzgycGVlcnBvbnBvbiAzNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '69 Schade St. Monkton (Jay Yundt)',
                                         'poller_hash': 'NTU2Mzgy',
                                         'serial_number': '556382',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:38:66',
                                         'uptime': '29355974'},
                                 {       'counter_source': '358/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '358/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '84105052863',
                                                                       'in_ucast_pkts': '5056633225',
                                                                       'mac_address': '60:db:98:4e:97:54',
                                                                       'name': '358/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '5025117925011',
                                                                       'out_ucast_pkts': '15997719252',
                                                                       'poller_hash': 'QTlDOEFBcGVlcmludGYzNTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '358',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '5025117925011',
                                                                       'in_ucast_pkts': '15997719252',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '5025117925011',
                                                                                                'in_ucast_pkts': '15997719252',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '84105052863',
                                                                                                'out_ucast_pkts': '5056633225',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '84105052863',
                                                                       'out_ucast_pkts': '5056633225',
                                                                       'poller_hash': 'QTlDOEFBcGVlcmludGZwb24gMzU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '358',
                                                                                               'links': [       {       'connected_time': '30942276',
                                                                                                                        'fiber_length': 18289,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOEFBcGVlcnBvbnBvbiAzNTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '101 MADISON ST E MONKTON (GREEN '
                                                 'MARIA)',
                                         'poller_hash': 'QTlDOEFB',
                                         'serial_number': 'A9C8AA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:97:54',
                                         'uptime': '30942276'},
                                 {       'counter_source': '360/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '360/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '145585917585',
                                                                       'in_ucast_pkts': '1056496658',
                                                                       'mac_address': '48:77:46:3a:7f:62',
                                                                       'name': '360/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1628359898019',
                                                                       'out_ucast_pkts': '2138018853',
                                                                       'poller_hash': 'NzIyMjA3cGVlcmludGYzNjAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '360',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1628359898019',
                                                                       'in_ucast_pkts': '2138018853',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1628359898019',
                                                                                                'in_ucast_pkts': '2138018853',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '145585917585',
                                                                                                'out_ucast_pkts': '1056496658',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '145585917585',
                                                                       'out_ucast_pkts': '1056496658',
                                                                       'poller_hash': 'NzIyMjA3cGVlcmludGZwb24gMzYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '360',
                                                                                               'links': [       {       'connected_time': '47307',
                                                                                                                        'fiber_length': 17608,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzIyMjA3cGVlcnBvbnBvbiAzNjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '168 Winstanley St. Monkton (Don '
                                                 'Buchanan)',
                                         'poller_hash': 'NzIyMjA3',
                                         'serial_number': '722207',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:3a:7f:62',
                                         'uptime': '47307'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '361/g1',
                                                                       'name': '361/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTU2M0E5cGVlcmludGYzNjEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '361',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTU2M0E5cGVlcmludGZwb24gMzYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '361',
                                                                                               'links': [       {       'connected_time': '24301068',
                                                                                                                        'fiber_length': 18269,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.624,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU2M0E5cGVlcnBvbnBvbiAzNjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '130 Maddison St. E. Monkton (Don '
                                                 'Hymers)',
                                         'poller_hash': 'NTU2M0E5',
                                         'serial_number': '5563A9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:39:9e',
                                         'uptime': '24301068'},
                                 {       'counter_source': '363/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '1c:7e:e5:37:f4:9d'],
                                                                       'id': '363/g1',
                                                                       'in_nucast_pkts': '323172',
                                                                       'in_octets': '5966862073',
                                                                       'in_ucast_pkts': '25764544',
                                                                       'mac_address': 'ec:4f:82:7d:ca:b7',
                                                                       'mtu': 2000,
                                                                       'name': '363/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1411166',
                                                                       'out_octets': '90635139092',
                                                                       'out_ucast_pkts': '63651365',
                                                                       'poller_hash': 'MzAwRTg4cGVlcmludGYzNjMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '363',
                                                                       'in_nucast_pkts': '1411166',
                                                                       'in_octets': '90635139092',
                                                                       'in_ucast_pkts': '63651365',
                                                                       'links': [       {       'in_nucast_pkts': '1411166',
                                                                                                'in_octets': '90635139092',
                                                                                                'in_ucast_pkts': '63651365',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '323172',
                                                                                                'out_octets': '5966862073',
                                                                                                'out_ucast_pkts': '25764544',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '323172',
                                                                       'out_octets': '5966862073',
                                                                       'out_ucast_pkts': '25764544',
                                                                       'poller_hash': 'MzAwRTg4cGVlcmludGZwb24gMzYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '363',
                                                                                               'links': [       {       'connected_time': '688395',
                                                                                                                        'fiber_length': 1173,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRTg4cGVlcnBvbnBvbiAzNjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '7 Llana Heights Main Fl (Robyn Yost)',
                                         'poller_hash': 'MzAwRTg4',
                                         'serial_number': '300E88',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:ca:b7',
                                         'uptime': '688395'},
                                 {       'counter_source': '364/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '364/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '833339544399',
                                                                       'in_ucast_pkts': '5617391324',
                                                                       'mac_address': 'cc:be:59:52:91:4f',
                                                                       'name': '364/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15443679623052',
                                                                       'out_ucast_pkts': '14797750296',
                                                                       'poller_hash': 'M0ZDODQ1cGVlcmludGYzNjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '364',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15443679623052',
                                                                       'in_ucast_pkts': '14797750296',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '15443679623052',
                                                                                                'in_ucast_pkts': '14797750296',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '833339544399',
                                                                                                'out_ucast_pkts': '5617391324',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '833339544399',
                                                                       'out_ucast_pkts': '5617391324',
                                                                       'poller_hash': 'M0ZDODQ1cGVlcmludGZwb24gMzY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '364',
                                                                                               'links': [       {       'connected_time': '26617541',
                                                                                                                        'fiber_length': 17448,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0ZDODQ1cGVlcnBvbnBvbiAzNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '202 WINSTANLEY ST (KIRKHAM DOUG & '
                                                 'KELLI)',
                                         'poller_hash': 'M0ZDODQ1',
                                         'serial_number': '3FC845',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:52:91:4f',
                                         'uptime': '26617541'},
                                 {       'counter_source': '366/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cc:d8'],
                                                                       'id': '366/g1',
                                                                       'in_nucast_pkts': '436103',
                                                                       'in_octets': '475208273605',
                                                                       'in_ucast_pkts': '1464439033',
                                                                       'mac_address': 'ec:4f:82:92:fe:4c',
                                                                       'mtu': 2000,
                                                                       'name': '366/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43269328',
                                                                       'out_octets': '7978303548788',
                                                                       'out_ucast_pkts': '6403605840',
                                                                       'poller_hash': 'MzI5NjU1cGVlcmludGYzNjYvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '366',
                                                                       'in_nucast_pkts': '43269328',
                                                                       'in_octets': '7978303548788',
                                                                       'in_ucast_pkts': '6403605840',
                                                                       'links': [       {       'in_nucast_pkts': '43269328',
                                                                                                'in_octets': '7978303548788',
                                                                                                'in_ucast_pkts': '6403605840',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '436103',
                                                                                                'out_octets': '475208273605',
                                                                                                'out_ucast_pkts': '1464439033',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '436103',
                                                                       'out_octets': '475208273605',
                                                                       'out_ucast_pkts': '1464439033',
                                                                       'poller_hash': 'MzI5NjU1cGVlcmludGZwb24gMzY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '366',
                                                                                               'links': [       {       'connected_time': '24669314',
                                                                                                                        'fiber_length': 18019,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -7.542,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzI5NjU1cGVlcnBvbnBvbiAzNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '48 MADDISON ST E (VILLEMAIRE '
                                                 'JACQUELINE)',
                                         'poller_hash': 'MzI5NjU1',
                                         'serial_number': '329655',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:92:fe:48',
                                         'uptime': '24669314'},
                                 {       'counter_source': '367/F1+367/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:17:68:76',
                                                                                                        '5c:0f:fb:17:72:46'],
                                                                       'id': '367/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2630692297',
                                                                       'in_ucast_pkts': '8596911',
                                                                       'name': '367/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODQ1QkI0cGVlcmludGYzNjcvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '367/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '356759743825',
                                                                       'in_ucast_pkts': '2227643235',
                                                                       'mac_address': '48:77:46:ce:bf:0e',
                                                                       'name': '367/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5664278220074',
                                                                       'out_ucast_pkts': '6209480659',
                                                                       'poller_hash': 'ODQ1QkI0cGVlcmludGYzNjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '367',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5664278220074',
                                                                       'in_ucast_pkts': '6209480659',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '5664278220074',
                                                                                                'in_ucast_pkts': '6209480659',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '359390436122',
                                                                                                'out_ucast_pkts': '2236240146',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '359390436122',
                                                                       'out_ucast_pkts': '2236240146',
                                                                       'poller_hash': 'ODQ1QkI0cGVlcmludGZwb24gMzY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '367',
                                                                                               'links': [       {       'connected_time': '8200272',
                                                                                                                        'fiber_length': 18194,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.402,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ1QkI0cGVlcnBvbnBvbiAzNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '193 QUEEN AVE (EARL JEFF & SUE)',
                                         'poller_hash': 'ODQ1QkI0',
                                         'serial_number': '845BB4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:ce:bf:0e',
                                         'uptime': '8200272'},
                                 {       'counter_source': '369/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:1e:e3:12:34:56'],
                                                                       'id': '369/g1',
                                                                       'in_nucast_pkts': '186156',
                                                                       'in_octets': '349532984192',
                                                                       'in_ucast_pkts': '1509217629',
                                                                       'mac_address': 'cc:be:59:61:19:7e',
                                                                       'mtu': 2000,
                                                                       'name': '369/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45076276',
                                                                       'out_octets': '3783212143012',
                                                                       'out_ucast_pkts': '2860432452',
                                                                       'poller_hash': 'NDU5NjVEcGVlcmludGYzNjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '369',
                                                                       'in_nucast_pkts': '45076276',
                                                                       'in_octets': '3783212143012',
                                                                       'in_ucast_pkts': '2860432452',
                                                                       'links': [       {       'in_nucast_pkts': '45076276',
                                                                                                'in_octets': '3783212143012',
                                                                                                'in_ucast_pkts': '2860432452',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '186156',
                                                                                                'out_octets': '349532984192',
                                                                                                'out_ucast_pkts': '1509217629',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '186156',
                                                                       'out_octets': '349532984192',
                                                                       'out_ucast_pkts': '1509217629',
                                                                       'poller_hash': 'NDU5NjVEcGVlcmludGZwb24gMzY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '369',
                                                                                               'links': [       {       'connected_time': '24669385',
                                                                                                                        'fiber_length': 18274,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.73,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDU5NjVEcGVlcnBvbnBvbiAzNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '111 Maddison St. W. Monkton (Steve '
                                                 'Smith Auto Sales)',
                                         'poller_hash': 'NDU5NjVE',
                                         'serial_number': '45965D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:61:19:7a',
                                         'uptime': '24669385'},
                                 {       'counter_source': '370/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '370/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1100840934407',
                                                                       'in_ucast_pkts': '10106123268',
                                                                       'mac_address': 'ec:4f:82:dc:f1:9c',
                                                                       'name': '370/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '35064146098830',
                                                                       'out_ucast_pkts': '28493369710',
                                                                       'poller_hash': 'MkVEOUNEcGVlcmludGYzNzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '370',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '35064146098830',
                                                                       'in_ucast_pkts': '28493369710',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '35064146098830',
                                                                                                'in_ucast_pkts': '28493369710',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1100840934407',
                                                                                                'out_ucast_pkts': '10106123268',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1100840934407',
                                                                       'out_ucast_pkts': '10106123268',
                                                                       'poller_hash': 'MkVEOUNEcGVlcmludGZwb24gMzcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '370',
                                                                                               'links': [       {       'connected_time': '22364992',
                                                                                                                        'fiber_length': 18443,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.518,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVEOUNEcGVlcnBvbnBvbiAzNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '147 Brock St. Monkton (Kevin '
                                                 'Coulson)',
                                         'poller_hash': 'MkVEOUNE',
                                         'serial_number': '2ED9CD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:dc:f1:9c',
                                         'uptime': '22364992'},
                                 {       'counter_source': '371/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '371/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '440250072245',
                                                                       'in_ucast_pkts': '3608902629',
                                                                       'mac_address': '48:77:46:3b:0e:7a',
                                                                       'name': '371/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15226219448059',
                                                                       'out_ucast_pkts': '12937185292',
                                                                       'poller_hash': 'NzIzM0VBcGVlcmludGYzNzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '371',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15226219448059',
                                                                       'in_ucast_pkts': '12937185292',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '15226219448059',
                                                                                                'in_ucast_pkts': '12937185292',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '440250072245',
                                                                                                'out_ucast_pkts': '3608902629',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '440250072245',
                                                                       'out_ucast_pkts': '3608902629',
                                                                       'poller_hash': 'NzIzM0VBcGVlcmludGZwb24gMzcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '371',
                                                                                               'links': [       {       'connected_time': '47663',
                                                                                                                        'fiber_length': 18130,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzIzM0VBcGVlcnBvbnBvbiAzNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '20 Schade St. Monkton (Ruby Rennick)',
                                         'poller_hash': 'NzIzM0VB',
                                         'serial_number': '7233EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:3b:0e:7a',
                                         'uptime': '47663'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '372/g1',
                                                                       'name': '372/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODAwQTlwZWVyaW50ZjM3Mi9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '372',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODAwQTlwZWVyaW50ZnBvbiAzNzI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '372',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODAwQTlwZWVycG9ucG9uIDM3Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '221 Winstanley St. Monkton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "372" '
                                                                  '(location: GPON '
                                                                  'port "5/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-24T08:36:33.620Z',
                                         'poller_hash': 'ODAwQTk=',
                                         'serial_number': '800A9',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '373/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2d:7a:35'],
                                                                       'id': '373/g1',
                                                                       'in_nucast_pkts': '304',
                                                                       'in_octets': '54577449616',
                                                                       'in_ucast_pkts': '287694172',
                                                                       'mac_address': 'ec:4f:82:ca:e7:5c',
                                                                       'mtu': 2000,
                                                                       'name': '373/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '80838394',
                                                                       'out_octets': '1091462197915',
                                                                       'out_ucast_pkts': '863368820',
                                                                       'poller_hash': 'MkM5NjU2cGVlcmludGYzNzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '373',
                                                                       'in_nucast_pkts': '80838394',
                                                                       'in_octets': '1091462197915',
                                                                       'in_ucast_pkts': '863368820',
                                                                       'links': [       {       'in_nucast_pkts': '80838394',
                                                                                                'in_octets': '1091462197915',
                                                                                                'in_ucast_pkts': '863368820',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '304',
                                                                                                'out_octets': '54577449616',
                                                                                                'out_ucast_pkts': '287694172',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '304',
                                                                       'out_octets': '54577449616',
                                                                       'out_ucast_pkts': '287694172',
                                                                       'poller_hash': 'MkM5NjU2cGVlcmludGZwb24gMzcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '373',
                                                                                               'links': [       {       'connected_time': '40487',
                                                                                                                        'fiber_length': 1322,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkM5NjU2cGVlcnBvbnBvbiAzNzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6611 Perth Rd. 131 Milverton (The '
                                                 'Garage&R&J Wagler)',
                                         'poller_hash': 'MkM5NjU2',
                                         'serial_number': '2C9656',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:ca:e7:58',
                                         'uptime': '40487'},
                                 {       'counter_source': '376/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b7:19'],
                                                                       'id': '376/g1',
                                                                       'in_nucast_pkts': '1186',
                                                                       'in_octets': '462326249847',
                                                                       'in_ucast_pkts': '2693871334',
                                                                       'mac_address': 'cc:be:59:eb:3b:52',
                                                                       'mtu': 2000,
                                                                       'name': '376/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45027725',
                                                                       'out_octets': '15715191799675',
                                                                       'out_ucast_pkts': '10819452001',
                                                                       'poller_hash': 'NTU2M0RGcGVlcmludGYzNzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '376',
                                                                       'in_nucast_pkts': '45027725',
                                                                       'in_octets': '15715191799675',
                                                                       'in_ucast_pkts': '10819452001',
                                                                       'links': [       {       'in_nucast_pkts': '45027725',
                                                                                                'in_octets': '15715191799675',
                                                                                                'in_ucast_pkts': '10819452001',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1186',
                                                                                                'out_octets': '462326249847',
                                                                                                'out_ucast_pkts': '2693871334',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1186',
                                                                       'out_octets': '462326249847',
                                                                       'out_ucast_pkts': '2693871334',
                                                                       'poller_hash': 'NTU2M0RGcGVlcmludGZwb24gMzc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '376',
                                                                                               'links': [       {       'connected_time': '24668969',
                                                                                                                        'fiber_length': 17810,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.098,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU2M0RGcGVlcnBvbnBvbiAzNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '197 King Ave, Monkton (Dana Verwey)',
                                         'poller_hash': 'NTU2M0RG',
                                         'serial_number': '5563DF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:3b:4e',
                                         'uptime': '24668969'},
                                 {       'counter_source': '377/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '377/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '234838016077',
                                                                       'in_ucast_pkts': '2067772958',
                                                                       'mac_address': '48:77:46:8c:2e:14',
                                                                       'name': '377/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8603056441117',
                                                                       'out_ucast_pkts': '7478821959',
                                                                       'poller_hash': 'NzZCMDY0cGVlcmludGYzNzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '377',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8603056441117',
                                                                       'in_ucast_pkts': '7478821959',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '8603056441117',
                                                                                                'in_ucast_pkts': '7478821959',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '234838016077',
                                                                                                'out_ucast_pkts': '2067772958',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '234838016077',
                                                                       'out_ucast_pkts': '2067772958',
                                                                       'poller_hash': 'NzZCMDY0cGVlcmludGZwb24gMzc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '377',
                                                                                               'links': [       {       'connected_time': '9663884',
                                                                                                                        'fiber_length': 17717,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzZCMDY0cGVlcnBvbnBvbiAzNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '205 King Ave, Monkton (Mervin '
                                                 'Subject)',
                                         'poller_hash': 'NzZCMDY0',
                                         'serial_number': '76B064',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:8c:2e:14',
                                         'uptime': '9663884'},
                                 {       'counter_source': '378/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '378/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1174374969783',
                                                                       'in_ucast_pkts': '11784527404',
                                                                       'mac_address': 'cc:be:59:eb:38:ee',
                                                                       'name': '378/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '31286379783879',
                                                                       'out_ucast_pkts': '21884124056',
                                                                       'poller_hash': 'NTU2MzkzcGVlcmludGYzNzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '378',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '31286379783879',
                                                                       'in_ucast_pkts': '21884124056',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '31286379783879',
                                                                                                'in_ucast_pkts': '21884124056',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1174374969783',
                                                                                                'out_ucast_pkts': '11784527404',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1174374969783',
                                                                       'out_ucast_pkts': '11784527404',
                                                                       'poller_hash': 'NTU2MzkzcGVlcmludGZwb24gMzc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '378',
                                                                                               'links': [       {       'connected_time': '47705',
                                                                                                                        'fiber_length': 18447,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.99,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU2MzkzcGVlcnBvbnBvbiAzNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '72 Schade St, Monkton (Geoff & Jodie '
                                                 'Tummers 10430143)',
                                         'poller_hash': 'NTU2Mzkz',
                                         'serial_number': '556393',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:38:ee',
                                         'uptime': '47705'},
                                 {       'counter_source': '379/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '379/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '133404008843',
                                                                       'in_ucast_pkts': '1094710636',
                                                                       'mac_address': 'ec:4f:82:92:ff:60',
                                                                       'name': '379/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6068813948816',
                                                                       'out_ucast_pkts': '5836258020',
                                                                       'poller_hash': 'MzI5Njc4cGVlcmludGYzNzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '379',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6068813948816',
                                                                       'in_ucast_pkts': '5836258020',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6068813948816',
                                                                                                'in_ucast_pkts': '5836258020',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '133404008843',
                                                                                                'out_ucast_pkts': '1094710636',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '133404008843',
                                                                       'out_ucast_pkts': '1094710636',
                                                                       'poller_hash': 'MzI5Njc4cGVlcmludGZwb24gMzc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '379',
                                                                                               'links': [       {       'connected_time': '47588',
                                                                                                                        'fiber_length': 18118,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.79,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzI5Njc4cGVlcnBvbnBvbiAzNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '92 MADDISON ST E (HOWES BEN '
                                                 '10430142)',
                                         'poller_hash': 'MzI5Njc4',
                                         'serial_number': '329678',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:92:ff:60',
                                         'uptime': '47588'},
                                 {       'counter_source': '380/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '380/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '852586618',
                                                                       'in_ucast_pkts': '3480891',
                                                                       'mac_address': 'cc:be:59:b7:3c:ff',
                                                                       'name': '380/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '43432029999',
                                                                       'out_ucast_pkts': '156388786',
                                                                       'poller_hash': 'NEQxMTUycGVlcmludGYzODAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '380',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '43432029999',
                                                                       'in_ucast_pkts': '156388786',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '43432029999',
                                                                                                'in_ucast_pkts': '156388786',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '852586618',
                                                                                                'out_ucast_pkts': '3480891',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '852586618',
                                                                       'out_ucast_pkts': '3480891',
                                                                       'poller_hash': 'NEQxMTUycGVlcmludGZwb24gMzgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '380',
                                                                                               'links': [       {       'connected_time': '47752',
                                                                                                                        'fiber_length': 18180,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.09,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQxMTUycGVlcnBvbnBvbiAzODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '206 Queen Ave, Monkton',
                                         'poller_hash': 'NEQxMTUy',
                                         'serial_number': '4D1152',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:b7:3c:ff',
                                         'uptime': '47752'},
                                 {       'counter_source': '381/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f6:0e'],
                                                                       'id': '381/g1',
                                                                       'in_nucast_pkts': '207599',
                                                                       'in_octets': '488111459179',
                                                                       'in_ucast_pkts': '1024832614',
                                                                       'mac_address': '00:06:31:3f:ff:24',
                                                                       'mtu': 2000,
                                                                       'name': '381/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45024923',
                                                                       'out_octets': '4779245969763',
                                                                       'out_ucast_pkts': '3509482053',
                                                                       'poller_hash': 'N0UyMTBwZWVyaW50ZjM4MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '381',
                                                                       'in_nucast_pkts': '45024923',
                                                                       'in_octets': '4779245969763',
                                                                       'in_ucast_pkts': '3509482053',
                                                                       'links': [       {       'in_nucast_pkts': '45024923',
                                                                                                'in_octets': '4779245969763',
                                                                                                'in_ucast_pkts': '3509482053',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '207599',
                                                                                                'out_octets': '488111459179',
                                                                                                'out_ucast_pkts': '1024832614',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '207599',
                                                                       'out_octets': '488111459179',
                                                                       'out_ucast_pkts': '1024832614',
                                                                       'poller_hash': 'N0UyMTBwZWVyaW50ZnBvbiAzODE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '381',
                                                                                               'links': [       {       'connected_time': '24669065',
                                                                                                                        'fiber_length': 18631,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'N0UyMTBwZWVycG9ucG9uIDM4MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '125 BROCK STREET MONKTON '
                                                 '(QUACKENBUSH PARKER 10436791)',
                                         'poller_hash': 'N0UyMTA=',
                                         'serial_number': '7E210',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:42:f6',
                                         'uptime': '24669065'},
                                 {       'counter_source': '382/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '382/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '119683100791',
                                                                       'in_ucast_pkts': '1622783441',
                                                                       'mac_address': 'ec:4f:82:b9:18:40',
                                                                       'name': '382/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4741999181743',
                                                                       'out_ucast_pkts': '4834838478',
                                                                       'poller_hash': 'MkZFOUIycGVlcmludGYzODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '382',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4741999181743',
                                                                       'in_ucast_pkts': '4834838478',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4741999181743',
                                                                                                'in_ucast_pkts': '4834838478',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '119683100791',
                                                                                                'out_ucast_pkts': '1622783441',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '119683100791',
                                                                       'out_ucast_pkts': '1622783441',
                                                                       'poller_hash': 'MkZFOUIycGVlcmludGZwb24gMzgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '382',
                                                                                               'links': [       {       'connected_time': '47863',
                                                                                                                        'fiber_length': 18399,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.126,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFOUIycGVlcnBvbnBvbiAzODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '149 BROCK STREET MONKTON (BROUGHTON '
                                                 'JOANNE)',
                                         'poller_hash': 'MkZFOUIy',
                                         'serial_number': '2FE9B2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:18:40',
                                         'uptime': '47863'},
                                 {       'counter_source': '383/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:93:c9'],
                                                                       'id': '383/g1',
                                                                       'in_nucast_pkts': '82',
                                                                       'in_octets': '402011816437',
                                                                       'in_ucast_pkts': '1483299727',
                                                                       'mac_address': 'cc:be:59:70:ab:86',
                                                                       'mtu': 2000,
                                                                       'name': '383/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '42437506',
                                                                       'out_octets': '7890544400939',
                                                                       'out_ucast_pkts': '5765802470',
                                                                       'poller_hash': 'NTYxMTYycGVlcmludGYzODMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '383',
                                                                       'in_nucast_pkts': '42437506',
                                                                       'in_octets': '7890544400939',
                                                                       'in_ucast_pkts': '5765802470',
                                                                       'links': [       {       'in_nucast_pkts': '42437506',
                                                                                                'in_octets': '7890544400939',
                                                                                                'in_ucast_pkts': '5765802470',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '82',
                                                                                                'out_octets': '402011816437',
                                                                                                'out_ucast_pkts': '1483299727',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82',
                                                                       'out_octets': '402011816437',
                                                                       'out_ucast_pkts': '1483299727',
                                                                       'poller_hash': 'NTYxMTYycGVlcmludGZwb24gMzgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '383',
                                                                                               'links': [       {       'connected_time': '23325118',
                                                                                                                        'fiber_length': 17454,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYxMTYycGVlcnBvbnBvbiAzODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '206 Winstanely St., Monkton (Nicole '
                                                 'Willfong 10434242)',
                                         'poller_hash': 'NTYxMTYy',
                                         'serial_number': '561162',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:70:ab:82',
                                         'uptime': '23325118'},
                                 {       'counter_source': '386/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:81:fd:0c'],
                                                                       'id': '386/g1',
                                                                       'in_nucast_pkts': '490057',
                                                                       'in_octets': '96097125797',
                                                                       'in_ucast_pkts': '230036022',
                                                                       'mac_address': '48:77:46:8b:8b:d8',
                                                                       'mtu': 2000,
                                                                       'name': '386/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45021491',
                                                                       'out_octets': '1720866987281',
                                                                       'out_ucast_pkts': '1315728594',
                                                                       'poller_hash': 'NzY5QzFDcGVlcmludGYzODYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '386',
                                                                       'in_nucast_pkts': '45021491',
                                                                       'in_octets': '1720866987281',
                                                                       'in_ucast_pkts': '1315728594',
                                                                       'links': [       {       'in_nucast_pkts': '45021491',
                                                                                                'in_octets': '1720866987281',
                                                                                                'in_ucast_pkts': '1315728594',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '490057',
                                                                                                'out_octets': '96097125797',
                                                                                                'out_ucast_pkts': '230036022',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '490057',
                                                                       'out_octets': '96097125797',
                                                                       'out_ucast_pkts': '230036022',
                                                                       'poller_hash': 'NzY5QzFDcGVlcmludGZwb24gMzg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '386',
                                                                                               'links': [       {       'connected_time': '24669156',
                                                                                                                        'fiber_length': 18666,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzY5QzFDcGVlcnBvbnBvbiAzODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '89 Schade St., Monkton (Gord &Helen '
                                                 'Young)',
                                         'poller_hash': 'NzY5QzFD',
                                         'serial_number': '769C1C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:8b:8b:d4',
                                         'uptime': '24669156'},
                                 {       'counter_source': '387/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '387/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '265433967156',
                                                                       'in_ucast_pkts': '1640561454',
                                                                       'mac_address': 'ec:4f:82:cc:84:84',
                                                                       'name': '387/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4177465501144',
                                                                       'out_ucast_pkts': '220662503',
                                                                       'poller_hash': 'MzkwREM4cGVlcmludGYzODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '387',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4177465501144',
                                                                       'in_ucast_pkts': '220662503',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4177465501144',
                                                                                                'in_ucast_pkts': '220662503',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '265433967156',
                                                                                                'out_ucast_pkts': '1640561454',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '265433967156',
                                                                       'out_ucast_pkts': '1640561454',
                                                                       'poller_hash': 'MzkwREM4cGVlcmludGZwb24gMzg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '387',
                                                                                               'links': [       {       'connected_time': '47513',
                                                                                                                        'fiber_length': 18545,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.384,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzkwREM4cGVlcnBvbnBvbiAzODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '61 SCHADE ST (BROUGHTON KELLY)',
                                         'poller_hash': 'MzkwREM4',
                                         'serial_number': '390DC8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cc:84:84',
                                         'uptime': '47513'},
                                 {       'counter_source': '388/F1+388/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a4:d3:80'],
                                                                       'id': '388/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3710788851',
                                                                       'in_ucast_pkts': '38456949',
                                                                       'name': '388/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NEQ0MjRFcGVlcmludGYzODgvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '388/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '523186336580',
                                                                       'in_ucast_pkts': '2864872329',
                                                                       'mac_address': 'cc:be:59:ef:f7:de',
                                                                       'name': '388/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6603814556288',
                                                                       'out_ucast_pkts': '6856177080',
                                                                       'poller_hash': 'NEQ0MjRFcGVlcmludGYzODgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '388',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6603814556288',
                                                                       'in_ucast_pkts': '6856177080',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6603814556288',
                                                                                                'in_ucast_pkts': '6856177080',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '526897125431',
                                                                                                'out_ucast_pkts': '2903329278',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '526897125431',
                                                                       'out_ucast_pkts': '2903329278',
                                                                       'poller_hash': 'NEQ0MjRFcGVlcmludGZwb24gMzg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '388',
                                                                                               'links': [       {       'connected_time': '73292557',
                                                                                                                        'fiber_length': 18280,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.172,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjRFcGVlcnBvbnBvbiAzODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '22 Winstanley St., Monkton (Natasha '
                                                 'Baxter 10433874)',
                                         'poller_hash': 'NEQ0MjRF',
                                         'serial_number': '4D424E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:de',
                                         'uptime': '73292557'},
                                 {       'counter_source': '390/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '390/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1258633370281',
                                                                       'in_ucast_pkts': '143827289838',
                                                                       'mac_address': '84:d3:43:95:8d:ba',
                                                                       'name': '390/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '18874275111781',
                                                                       'out_ucast_pkts': '13364166437',
                                                                       'poller_hash': 'Q0JEMUQ2cGVlcmludGYzOTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '390',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '18874275111781',
                                                                       'in_ucast_pkts': '13364166437',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '18874275111781',
                                                                                                'in_ucast_pkts': '13364166437',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1258633370281',
                                                                                                'out_ucast_pkts': '143827289838',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1258633370281',
                                                                       'out_ucast_pkts': '143827289838',
                                                                       'poller_hash': 'Q0JEMUQ2cGVlcmludGZwb24gMzkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '390',
                                                                                               'links': [       {       'connected_time': '3113848',
                                                                                                                        'fiber_length': 18096,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUQ2cGVlcnBvbnBvbiAzOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '22 Maddison St., East, Monkton (Greg '
                                                 '& Kathy Leung)',
                                         'poller_hash': 'Q0JEMUQ2',
                                         'serial_number': 'CBD1D6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8d:ba',
                                         'uptime': '3113848'},
                                 {       'counter_source': '391/F1+391/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a8:0e:56',
                                                                                                        '00:03:e6:a8:0e:58'],
                                                                       'id': '391/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '632942303',
                                                                       'in_ucast_pkts': '4333530',
                                                                       'name': '391/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzY2QzY3cGVlcmludGYzOTEvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '391/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '93304960151',
                                                                       'in_ucast_pkts': '620872627',
                                                                       'mac_address': 'ec:4f:82:c1:79:a0',
                                                                       'name': '391/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2270820101997',
                                                                       'out_ucast_pkts': '1732276645',
                                                                       'poller_hash': 'MzY2QzY3cGVlcmludGYzOTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '391',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2270820101997',
                                                                       'in_ucast_pkts': '1732276645',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2270820101997',
                                                                                                'in_ucast_pkts': '1732276645',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '93937902454',
                                                                                                'out_ucast_pkts': '625206157',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '93937902454',
                                                                       'out_ucast_pkts': '625206157',
                                                                       'poller_hash': 'MzY2QzY3cGVlcmludGZwb24gMzkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '391',
                                                                                               'links': [       {       'connected_time': '47909',
                                                                                                                        'fiber_length': 18163,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzY2QzY3cGVlcnBvbnBvbiAzOTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '23 Schade St., Monkton (Sherry '
                                                 'Heibein)',
                                         'poller_hash': 'MzY2QzY3',
                                         'serial_number': '366C67',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:c1:79:a0',
                                         'uptime': '47909'},
                                 {       'counter_source': '392/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '392/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '286595945395',
                                                                       'in_ucast_pkts': '2489959144',
                                                                       'mac_address': 'cc:be:59:eb:3a:f6',
                                                                       'name': '392/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8650200131628',
                                                                       'out_ucast_pkts': '6319898873',
                                                                       'poller_hash': 'NTU2M0Q0cGVlcmludGYzOTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '392',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8650200131628',
                                                                       'in_ucast_pkts': '6319898873',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '8650200131628',
                                                                                                'in_ucast_pkts': '6319898873',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '286595945395',
                                                                                                'out_ucast_pkts': '2489959144',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '286595945395',
                                                                       'out_ucast_pkts': '2489959144',
                                                                       'poller_hash': 'NTU2M0Q0cGVlcmludGZwb24gMzky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '392',
                                                                                               'links': [       {       'connected_time': '10038949',
                                                                                                                        'fiber_length': 17996,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.592,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU2M0Q0cGVlcnBvbnBvbiAzOTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '59 Maddison St. West, Monkton (Kevin '
                                                 'Cook 10430147)',
                                         'poller_hash': 'NTU2M0Q0',
                                         'serial_number': '5563D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:3a:f6',
                                         'uptime': '10038949'},
                                 {       'counter_source': '393/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '393/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '584835506384',
                                                                       'in_ucast_pkts': '3216269940',
                                                                       'mac_address': 'cc:be:59:ef:f7:ce',
                                                                       'name': '393/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9343156908344',
                                                                       'out_ucast_pkts': '8947268880',
                                                                       'poller_hash': 'NEQ0MjRDcGVlcmludGYzOTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '393',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9343156908344',
                                                                       'in_ucast_pkts': '8947268880',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9343156908344',
                                                                                                'in_ucast_pkts': '8947268880',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '584835506384',
                                                                                                'out_ucast_pkts': '3216269940',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '584835506384',
                                                                       'out_ucast_pkts': '3216269940',
                                                                       'poller_hash': 'NEQ0MjRDcGVlcmludGZwb24gMzkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '393',
                                                                                               'links': [       {       'connected_time': '11345518',
                                                                                                                        'fiber_length': 17538,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.072,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjRDcGVlcnBvbnBvbiAzOTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '170 Fishleigh St., Monkton (Ben '
                                                 'Fizell)',
                                         'poller_hash': 'NEQ0MjRD',
                                         'serial_number': '4D424C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:ce',
                                         'uptime': '11345518'},
                                 {       'counter_source': '394/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c4:41:1e:77:6f:d0'],
                                                                       'id': '394/g1',
                                                                       'in_nucast_pkts': '7438',
                                                                       'in_octets': '6420382683',
                                                                       'in_ucast_pkts': '24391573',
                                                                       'mac_address': '44:65:7f:33:1f:f9',
                                                                       'mtu': 2000,
                                                                       'name': '394/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119775',
                                                                       'out_octets': '96991097771',
                                                                       'out_ucast_pkts': '76262059',
                                                                       'poller_hash': 'NjQwQTc2cGVlcmludGYzOTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '394',
                                                                       'in_nucast_pkts': '1119775',
                                                                       'in_octets': '96991097771',
                                                                       'in_ucast_pkts': '76262059',
                                                                       'links': [       {       'in_nucast_pkts': '1119775',
                                                                                                'in_octets': '96991097771',
                                                                                                'in_ucast_pkts': '76262059',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7438',
                                                                                                'out_octets': '6420382683',
                                                                                                'out_ucast_pkts': '24391573',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7438',
                                                                       'out_octets': '6420382683',
                                                                       'out_ucast_pkts': '24391573',
                                                                       'poller_hash': 'NjQwQTc2cGVlcmludGZwb24gMzk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '394',
                                                                                               'links': [       {       'connected_time': '688421',
                                                                                                                        'fiber_length': 1321,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.908,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.174}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQTc2cGVlcnBvbnBvbiAzOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '91 Mill St E Milverton (Carey '
                                                 'Soehner 10431225)',
                                         'poller_hash': 'NjQwQTc2',
                                         'serial_number': '640A76',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:1f:f8',
                                         'uptime': '688421'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '395/g1',
                                                                       'name': '395/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2REE2cGVlcmludGYzOTUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '395',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2REE2cGVlcmludGZwb24gMzk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '395',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2REE2cGVlcnBvbnBvbiAzOTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '213 Selby Ave, Monkton (Craig Burek '
                                                 '10435498)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "395" '
                                                                  '(location: GPON '
                                                                  'port "5/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-24T08:27:53.990Z',
                                         'poller_hash': 'OTU2REE2',
                                         'serial_number': '956DA6',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '396/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '396/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '881529182283',
                                                                       'in_ucast_pkts': '7346179017',
                                                                       'mac_address': 'cc:be:59:ef:f7:8e',
                                                                       'name': '396/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '33319642349566',
                                                                       'out_ucast_pkts': '22232101578',
                                                                       'poller_hash': 'NEQ0MjQ0cGVlcmludGYzOTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '396',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '33319642349566',
                                                                       'in_ucast_pkts': '22232101578',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '33319642349566',
                                                                                                'in_ucast_pkts': '22232101578',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '881529182283',
                                                                                                'out_ucast_pkts': '7346179017',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '881529182283',
                                                                       'out_ucast_pkts': '7346179017',
                                                                       'poller_hash': 'NEQ0MjQ0cGVlcmludGZwb24gMzk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '396',
                                                                                               'links': [       {       'connected_time': '47644',
                                                                                                                        'fiber_length': 17842,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.144,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjQ0cGVlcnBvbnBvbiAzOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '22 Maddison St. West, Monkton '
                                                 '(George & Pat Worth)',
                                         'poller_hash': 'NEQ0MjQ0',
                                         'serial_number': '4D4244',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:8e',
                                         'uptime': '47644'},
                                 {       'counter_source': '397/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '397/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '449830905952',
                                                                       'in_ucast_pkts': '2299122214',
                                                                       'mac_address': 'ec:4f:82:92:fe:d8',
                                                                       'name': '397/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7771160290976',
                                                                       'out_ucast_pkts': '7662805705',
                                                                       'poller_hash': 'MzI5NjY3cGVlcmludGYzOTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '397',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7771160290976',
                                                                       'in_ucast_pkts': '7662805705',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7771160290976',
                                                                                                'in_ucast_pkts': '7662805705',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '449830905952',
                                                                                                'out_ucast_pkts': '2299122214',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '449830905952',
                                                                       'out_ucast_pkts': '2299122214',
                                                                       'poller_hash': 'MzI5NjY3cGVlcmludGZwb24gMzk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '397',
                                                                                               'links': [       {       'connected_time': '47916',
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.342,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzI5NjY3cGVlcnBvbnBvbiAzOTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '211 Yonge St., Monkton (Terry & '
                                                 'Leslie Brown)',
                                         'poller_hash': 'MzI5NjY3',
                                         'serial_number': '329667',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:92:fe:d8',
                                         'uptime': '47916'},
                                 {       'counter_source': '398/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '398/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '485327153805',
                                                                       'in_ucast_pkts': '4215084062',
                                                                       'mac_address': 'cc:be:59:ef:f7:46',
                                                                       'name': '398/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '16108082844035',
                                                                       'out_ucast_pkts': '17300870669',
                                                                       'poller_hash': 'NEQ0MjNCcGVlcmludGYzOTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '398',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '16108082844035',
                                                                       'in_ucast_pkts': '17300870669',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '16108082844035',
                                                                                                'in_ucast_pkts': '17300870669',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '485327153805',
                                                                                                'out_ucast_pkts': '4215084062',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '485327153805',
                                                                       'out_ucast_pkts': '4215084062',
                                                                       'poller_hash': 'NEQ0MjNCcGVlcmludGZwb24gMzk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '398',
                                                                                               'links': [       {       'connected_time': '47674',
                                                                                                                        'fiber_length': 18214,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.578,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjNCcGVlcnBvbnBvbiAzOTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '100 MADDISON ST E (BOYD RON & KATHY)',
                                         'poller_hash': 'NEQ0MjNC',
                                         'serial_number': '4D423B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:46',
                                         'uptime': '47674'},
                                 {       'counter_source': '400/F1+400/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:eb:7d:1e'],
                                                                       'id': '400/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2593699835',
                                                                       'in_ucast_pkts': '16081417',
                                                                       'name': '400/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NEQ0MUYycGVlcmludGY0MDAvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '400/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '126804021076',
                                                                       'in_ucast_pkts': '666337701',
                                                                       'mac_address': 'cc:be:59:ef:f4:fe',
                                                                       'name': '400/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1981870013459',
                                                                       'out_ucast_pkts': '1731128645',
                                                                       'poller_hash': 'NEQ0MUYycGVlcmludGY0MDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '400',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1981870013459',
                                                                       'in_ucast_pkts': '1731128645',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1981870013459',
                                                                                                'in_ucast_pkts': '1731128645',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '129397720911',
                                                                                                'out_ucast_pkts': '682419118',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '129397720911',
                                                                       'out_ucast_pkts': '682419118',
                                                                       'poller_hash': 'NEQ0MUYycGVlcmludGZwb24gNDAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '400',
                                                                                               'links': [       {       'connected_time': '29355792',
                                                                                                                        'fiber_length': 17645,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.242,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MUYycGVlcnBvbnBvbiA0MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '218 Selby Ave, Monkton (Steve & '
                                                 'Krystle Pettapiece)',
                                         'poller_hash': 'NEQ0MUYy',
                                         'serial_number': '4D41F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f4:fe',
                                         'uptime': '29355792'},
                                 {       'counter_source': '401/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '401/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5304648034973',
                                                                       'in_ucast_pkts': '12061084943',
                                                                       'mac_address': 'cc:be:59:70:35:1e',
                                                                       'name': '401/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '27308157816730',
                                                                       'out_ucast_pkts': '21981624298',
                                                                       'poller_hash': 'NTYwMjk2cGVlcmludGY0MDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '401',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '27308157816730',
                                                                       'in_ucast_pkts': '21981624298',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '27308157816730',
                                                                                                'in_ucast_pkts': '21981624298',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '5304648034973',
                                                                                                'out_ucast_pkts': '12061084943',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5304648034973',
                                                                       'out_ucast_pkts': '12061084943',
                                                                       'poller_hash': 'NTYwMjk2cGVlcmludGZwb24gNDAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '401',
                                                                                               'links': [       {       'connected_time': '13127788',
                                                                                                                        'fiber_length': 17621,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYwMjk2cGVlcnBvbnBvbiA0MDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '210 Winstanley Ave, Monkton (Donna '
                                                 'McConnell)',
                                         'poller_hash': 'NTYwMjk2',
                                         'serial_number': '560296',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:70:35:1e',
                                         'uptime': '13127788'},
                                 {       'counter_source': '403/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:2b'],
                                                                       'id': '403/g1',
                                                                       'in_nucast_pkts': '27',
                                                                       'in_octets': '148685765838',
                                                                       'in_ucast_pkts': '838426208',
                                                                       'mac_address': 'f8:85:f9:17:2a:cb',
                                                                       'mtu': 2000,
                                                                       'name': '403/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6685784',
                                                                       'out_octets': '2650478425474',
                                                                       'out_ucast_pkts': '1888023483',
                                                                       'poller_hash': 'REZGRTQycGVlcmludGY0MDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '403',
                                                                       'in_nucast_pkts': '6685784',
                                                                       'in_octets': '2650478425474',
                                                                       'in_ucast_pkts': '1888023483',
                                                                       'links': [       {       'in_nucast_pkts': '6685784',
                                                                                                'in_octets': '2650478425474',
                                                                                                'in_ucast_pkts': '1888023483',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27',
                                                                                                'out_octets': '148685765838',
                                                                                                'out_ucast_pkts': '838426208',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27',
                                                                       'out_octets': '148685765838',
                                                                       'out_ucast_pkts': '838426208',
                                                                       'poller_hash': 'REZGRTQycGVlcmludGZwb24gNDAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '403',
                                                                                               'links': [       {       'connected_time': '3113867',
                                                                                                                        'fiber_length': 18366,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTQycGVlcnBvbnBvbiA0MDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '124 Maddison St. East, Monkton '
                                                 '(Vanessa Schiestel 10435517)',
                                         'poller_hash': 'REZGRTQy',
                                         'serial_number': 'DFFE42',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2a:c8',
                                         'uptime': '3113867'},
                                 {       'counter_source': '404/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '404/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '69037336659',
                                                                       'in_ucast_pkts': '47468127409',
                                                                       'mac_address': 'ec:4f:82:4f:76:f7',
                                                                       'name': '404/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1484451426441',
                                                                       'out_ucast_pkts': '1055553659',
                                                                       'poller_hash': 'Mjk4MUUxcGVlcmludGY0MDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '404',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1484451426441',
                                                                       'in_ucast_pkts': '1055553659',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1484451426441',
                                                                                                'in_ucast_pkts': '1055553659',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '69037336659',
                                                                                                'out_ucast_pkts': '47468127409',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '69037336659',
                                                                       'out_ucast_pkts': '47468127409',
                                                                       'poller_hash': 'Mjk4MUUxcGVlcmludGZwb24gNDA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '404',
                                                                                               'links': [       {       'connected_time': '82401',
                                                                                                                        'fiber_length': 1195,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MUUxcGVlcnBvbnBvbiA0MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '57 MAIN ST N (DZODAN DEJAN)',
                                         'poller_hash': 'Mjk4MUUx',
                                         'serial_number': '2981E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:76:f7',
                                         'uptime': '82401'},
                                 {       'counter_source': '405/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d4:9b'],
                                                                       'id': '405/g1',
                                                                       'in_nucast_pkts': '953213',
                                                                       'in_octets': '1067234933448',
                                                                       'in_ucast_pkts': '5590110946',
                                                                       'mac_address': 'ec:4f:82:5b:70:e6',
                                                                       'mtu': 2000,
                                                                       'name': '405/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '148489576',
                                                                       'out_octets': '23308830238623',
                                                                       'out_ucast_pkts': '16955430802',
                                                                       'poller_hash': 'MkIwMThGcGVlcmludGY0MDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '405',
                                                                       'in_nucast_pkts': '148489576',
                                                                       'in_octets': '23308830238623',
                                                                       'in_ucast_pkts': '16955430802',
                                                                       'links': [       {       'in_nucast_pkts': '148489576',
                                                                                                'in_octets': '23308830238623',
                                                                                                'in_ucast_pkts': '16955430802',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '953213',
                                                                                                'out_octets': '1067234933448',
                                                                                                'out_ucast_pkts': '5590110946',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '953213',
                                                                       'out_octets': '1067234933448',
                                                                       'out_ucast_pkts': '5590110946',
                                                                       'poller_hash': 'MkIwMThGcGVlcmludGZwb24gNDA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '405',
                                                                                               'links': [       {       'connected_time': '21815964',
                                                                                                                        'fiber_length': 1166,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMThGcGVlcnBvbnBvbiA0MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 MAIN ST N (MACDONALD ALANNA)',
                                         'poller_hash': 'MkIwMThG',
                                         'serial_number': '2B018F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:e3',
                                         'uptime': '21815964'},
                                 {       'counter_source': '406/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '406/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '535192464315',
                                                                       'in_ucast_pkts': '1121974635',
                                                                       'mac_address': 'ec:4f:82:0d:6a:7b',
                                                                       'name': '406/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '191',
                                                                       'out_octets': '3051362288151',
                                                                       'out_ucast_pkts': '2592887450',
                                                                       'poller_hash': 'MjA1REY2cGVlcmludGY0MDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '406',
                                                                       'in_nucast_pkts': '191',
                                                                       'in_octets': '3051362288151',
                                                                       'in_ucast_pkts': '2592887450',
                                                                       'links': [       {       'in_nucast_pkts': '191',
                                                                                                'in_octets': '3051362288151',
                                                                                                'in_ucast_pkts': '2592887450',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '535192464315',
                                                                                                'out_ucast_pkts': '1121974635',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '535192464315',
                                                                       'out_ucast_pkts': '1121974635',
                                                                       'poller_hash': 'MjA1REY2cGVlcmludGZwb24gNDA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '406',
                                                                                               'links': [       {       'connected_time': '21816989',
                                                                                                                        'fiber_length': 1271,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REY2cGVlcnBvbnBvbiA0MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '45 Main St. N. Milverton (Greg '
                                                 'Guenther)',
                                         'poller_hash': 'MjA1REY2',
                                         'serial_number': '205DF6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:7b',
                                         'uptime': '21816989'},
                                 {       'counter_source': '407/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '407/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '314429498712',
                                                                       'in_ucast_pkts': '82684105815',
                                                                       'mac_address': 'ec:4f:82:5b:87:03',
                                                                       'name': '407/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '99',
                                                                       'out_octets': '5338069717696',
                                                                       'out_ucast_pkts': '7905442774',
                                                                       'poller_hash': 'MkIwNDUzcGVlcmludGY0MDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '407',
                                                                       'in_nucast_pkts': '99',
                                                                       'in_octets': '5338069717696',
                                                                       'in_ucast_pkts': '7905442774',
                                                                       'links': [       {       'in_nucast_pkts': '99',
                                                                                                'in_octets': '5338069717696',
                                                                                                'in_ucast_pkts': '7905442774',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '314429498712',
                                                                                                'out_ucast_pkts': '82684105815',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '314429498712',
                                                                       'out_ucast_pkts': '82684105815',
                                                                       'poller_hash': 'MkIwNDUzcGVlcmludGZwb24gNDA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '407',
                                                                                               'links': [       {       'connected_time': '688481',
                                                                                                                        'fiber_length': 1927,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDUzcGVlcnBvbnBvbiA0MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '40 Main St. N. Milverton (Eric '
                                                 'Sider)',
                                         'poller_hash': 'MkIwNDUz',
                                         'serial_number': '2B0453',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:87:03',
                                         'uptime': '688481'},
                                 {       'counter_source': '408/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '408/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96056912124',
                                                                       'in_ucast_pkts': '884047599',
                                                                       'mac_address': 'ec:4f:82:0d:68:f8',
                                                                       'name': '408/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '5076147345211',
                                                                       'out_ucast_pkts': '3431449616',
                                                                       'poller_hash': 'MjA1RENCcGVlcmludGY0MDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '408',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '5076147345211',
                                                                       'in_ucast_pkts': '3431449616',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '5076147345211',
                                                                                                'in_ucast_pkts': '3431449616',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96056912124',
                                                                                                'out_ucast_pkts': '884047599',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96056912124',
                                                                       'out_ucast_pkts': '884047599',
                                                                       'poller_hash': 'MjA1RENCcGVlcmludGZwb24gNDA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '408',
                                                                                               'links': [       {       'connected_time': '65324014',
                                                                                                                        'fiber_length': 1898,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RENCcGVlcnBvbnBvbiA0MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '44 Main St. N. Milverton (David '
                                                 'Weber)',
                                         'poller_hash': 'MjA1RENC',
                                         'serial_number': '205DCB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:f8',
                                         'uptime': '65324014'},
                                 {       'counter_source': '409/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:0a:d2'],
                                                                       'id': '409/g1',
                                                                       'in_nucast_pkts': '2861729',
                                                                       'in_octets': '191671062003',
                                                                       'in_ucast_pkts': '2613210116',
                                                                       'mac_address': 'ec:4f:82:5b:70:1e',
                                                                       'mtu': 2000,
                                                                       'name': '409/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '199404850',
                                                                       'out_octets': '14602589163063',
                                                                       'out_ucast_pkts': '10340161682',
                                                                       'poller_hash': 'MkIwMTc2cGVlcmludGY0MDkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '409',
                                                                       'in_nucast_pkts': '199404850',
                                                                       'in_octets': '14602589163063',
                                                                       'in_ucast_pkts': '10340161682',
                                                                       'links': [       {       'in_nucast_pkts': '199404850',
                                                                                                'in_octets': '14602589163063',
                                                                                                'in_ucast_pkts': '10340161682',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2861729',
                                                                                                'out_octets': '191671062003',
                                                                                                'out_ucast_pkts': '2613210116',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2861729',
                                                                       'out_octets': '191671062003',
                                                                       'out_ucast_pkts': '2613210116',
                                                                       'poller_hash': 'MkIwMTc2cGVlcmludGZwb24gNDA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '409',
                                                                                               'links': [       {       'connected_time': '30076165',
                                                                                                                        'fiber_length': 1147,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTc2cGVlcnBvbnBvbiA0MDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6 Grosch St. Milverton (Jim '
                                                 'Partridge)',
                                         'poller_hash': 'MkIwMTc2',
                                         'serial_number': '2B0176',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:1b',
                                         'uptime': '30076165'},
                                 {       'counter_source': '410/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '410/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '138730074625',
                                                                       'in_ucast_pkts': '1430507108',
                                                                       'mac_address': 'cc:be:59:cb:ab:d2',
                                                                       'name': '410/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '5340784566626',
                                                                       'out_ucast_pkts': '3643271098',
                                                                       'poller_hash': 'NDJDNkQycGVlcmludGY0MTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '410',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '5340784566626',
                                                                       'in_ucast_pkts': '3643271098',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '5340784566626',
                                                                                                'in_ucast_pkts': '3643271098',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '138730074625',
                                                                                                'out_ucast_pkts': '1430507108',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '138730074625',
                                                                       'out_ucast_pkts': '1430507108',
                                                                       'poller_hash': 'NDJDNkQycGVlcmludGZwb24gNDEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '410',
                                                                                               'links': [       {       'connected_time': '2091613',
                                                                                                                        'fiber_length': 18107,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkQycGVlcnBvbnBvbiA0MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '42 MADDISON ST E (MARTIN GERALD)',
                                         'poller_hash': 'NDJDNkQy',
                                         'serial_number': '42C6D2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ab:d2',
                                         'uptime': '2091613'},
                                 {       'counter_source': '411/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9d:d9:8a'],
                                                                       'id': '411/g1',
                                                                       'in_nucast_pkts': '670535',
                                                                       'in_octets': '1911101621413',
                                                                       'in_ucast_pkts': '6531995068',
                                                                       'mac_address': 'ec:4f:82:5b:70:d6',
                                                                       'mtu': 2000,
                                                                       'name': '411/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '93456865',
                                                                       'out_octets': '16693728419846',
                                                                       'out_ucast_pkts': '12027919666',
                                                                       'poller_hash': 'MkIwMThEcGVlcmludGY0MTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '411',
                                                                       'in_nucast_pkts': '93456865',
                                                                       'in_octets': '16693728419846',
                                                                       'in_ucast_pkts': '12027919666',
                                                                       'links': [       {       'in_nucast_pkts': '93456865',
                                                                                                'in_octets': '16693728419846',
                                                                                                'in_ucast_pkts': '12027919666',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '670535',
                                                                                                'out_octets': '1911101621413',
                                                                                                'out_ucast_pkts': '6531995068',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '670535',
                                                                       'out_octets': '1911101621413',
                                                                       'out_ucast_pkts': '6531995068',
                                                                       'poller_hash': 'MkIwMThEcGVlcmludGZwb24gNDEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '411',
                                                                                               'links': [       {       'connected_time': '71027301',
                                                                                                                        'fiber_length': 1525,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMThEcGVlcnBvbnBvbiA0MTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '74 MAIN ST N (MCCORMICK PHIL & '
                                                 'CAROL)',
                                         'poller_hash': 'MkIwMThE',
                                         'serial_number': '2B018D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:d3',
                                         'uptime': '71027301'},
                                 {       'counter_source': '412/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '412/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '32291728406',
                                                                       'in_ucast_pkts': '4414529026',
                                                                       'mac_address': '60:db:98:4e:66:f4',
                                                                       'name': '412/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1254672908408',
                                                                       'out_ucast_pkts': '898636499',
                                                                       'poller_hash': 'QTlDMjlFcGVlcmludGY0MTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '412',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1254672908408',
                                                                       'in_ucast_pkts': '898636499',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1254672908408',
                                                                                                'in_ucast_pkts': '898636499',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '32291728406',
                                                                                                'out_ucast_pkts': '4414529026',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '32291728406',
                                                                       'out_ucast_pkts': '4414529026',
                                                                       'poller_hash': 'QTlDMjlFcGVlcmludGZwb24gNDEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '412',
                                                                                               'links': [       {       'connected_time': '688352',
                                                                                                                        'fiber_length': 1008,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMjlFcGVlcnBvbnBvbiA0MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 MAIN ST S (GERBER CAITLYN)',
                                         'poller_hash': 'QTlDMjlF',
                                         'serial_number': 'A9C29E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:66:f4',
                                         'uptime': '688352'},
                                 {       'counter_source': '413/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '413/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '122646930660',
                                                                       'in_ucast_pkts': '367279556',
                                                                       'mac_address': 'ec:4f:82:cc:02:34',
                                                                       'name': '413/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1655367984396',
                                                                       'out_ucast_pkts': '1394650866',
                                                                       'poller_hash': 'MzhGRDdGcGVlcmludGY0MTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '413',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1655367984396',
                                                                       'in_ucast_pkts': '1394650866',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1655367984396',
                                                                                                'in_ucast_pkts': '1394650866',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '122646930660',
                                                                                                'out_ucast_pkts': '367279556',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '122646930660',
                                                                       'out_ucast_pkts': '367279556',
                                                                       'poller_hash': 'MzhGRDdGcGVlcmludGZwb24gNDEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '413',
                                                                                               'links': [       {       'connected_time': '47987',
                                                                                                                        'fiber_length': 1199,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.338,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzhGRDdGcGVlcnBvbnBvbiA0MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '91 MAIN ST N (KUEPFER TRAVIS)',
                                         'poller_hash': 'MzhGRDdG',
                                         'serial_number': '38FD7F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cc:02:34',
                                         'uptime': '47987'},
                                 {       'counter_source': '414/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '30:68:93:36:15:5d'],
                                                                       'id': '414/g1',
                                                                       'in_nucast_pkts': '193577',
                                                                       'in_octets': '1296768991',
                                                                       'in_ucast_pkts': '4952783',
                                                                       'mac_address': 'ec:4f:82:7d:c9:3d',
                                                                       'mtu': 2000,
                                                                       'name': '414/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464171',
                                                                       'out_octets': '30954048075',
                                                                       'out_ucast_pkts': '23817324',
                                                                       'poller_hash': 'MzAwRENCcGVlcmludGY0MTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '414',
                                                                       'in_nucast_pkts': '1464171',
                                                                       'in_octets': '30954048075',
                                                                       'in_ucast_pkts': '23817324',
                                                                       'links': [       {       'in_nucast_pkts': '1464171',
                                                                                                'in_octets': '30954048075',
                                                                                                'in_ucast_pkts': '23817324',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '193577',
                                                                                                'out_octets': '1296768991',
                                                                                                'out_ucast_pkts': '4952783',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '193577',
                                                                       'out_octets': '1296768991',
                                                                       'out_ucast_pkts': '4952783',
                                                                       'poller_hash': 'MzAwRENCcGVlcmludGZwb24gNDE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '414',
                                                                                               'links': [       {       'connected_time': '688497',
                                                                                                                        'fiber_length': 1534,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRENCcGVlcnBvbnBvbiA0MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '76 MAIN ST N (NAFZIGER COLE)',
                                         'poller_hash': 'MzAwRENC',
                                         'serial_number': '300DCB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c9:3d',
                                         'uptime': '688497'},
                                 {       'counter_source': '417/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '417/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '120799275468',
                                                                       'in_ucast_pkts': '17680076177',
                                                                       'mac_address': 'ec:4f:82:5b:86:eb',
                                                                       'name': '417/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '126',
                                                                       'out_octets': '2133605010264',
                                                                       'out_ucast_pkts': '1839589338',
                                                                       'poller_hash': 'MkIwNDUwcGVlcmludGY0MTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '417',
                                                                       'in_nucast_pkts': '126',
                                                                       'in_octets': '2133605010264',
                                                                       'in_ucast_pkts': '1839589338',
                                                                       'links': [       {       'in_nucast_pkts': '126',
                                                                                                'in_octets': '2133605010264',
                                                                                                'in_ucast_pkts': '1839589338',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '120799275468',
                                                                                                'out_ucast_pkts': '17680076177',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '120799275468',
                                                                       'out_ucast_pkts': '17680076177',
                                                                       'poller_hash': 'MkIwNDUwcGVlcmludGZwb24gNDE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '417',
                                                                                               'links': [       {       'connected_time': '26779881',
                                                                                                                        'fiber_length': 1306,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDUwcGVlcnBvbnBvbiA0MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 EDWARD ST (HERD CATHY)',
                                         'poller_hash': 'MkIwNDUw',
                                         'serial_number': '2B0450',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:86:eb',
                                         'uptime': '26779881'},
                                 {       'counter_source': '418/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '418/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '49255058290',
                                                                       'in_ucast_pkts': '303845739',
                                                                       'mac_address': 'ec:4f:82:5b:71:13',
                                                                       'name': '418/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '4053133187724',
                                                                       'out_ucast_pkts': '2764056035',
                                                                       'poller_hash': 'MkIwMTk1cGVlcmludGY0MTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '418',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '4053133187724',
                                                                       'in_ucast_pkts': '2764056035',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '4053133187724',
                                                                                                'in_ucast_pkts': '2764056035',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '49255058290',
                                                                                                'out_ucast_pkts': '303845739',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '49255058290',
                                                                       'out_ucast_pkts': '303845739',
                                                                       'poller_hash': 'MkIwMTk1cGVlcmludGZwb24gNDE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '418',
                                                                                               'links': [       {       'connected_time': '13219345',
                                                                                                                        'fiber_length': 1203,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTk1cGVlcnBvbnBvbiA0MTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '67 Main St. N. (Michele Duskocy '
                                                 '10400255)',
                                         'poller_hash': 'MkIwMTk1',
                                         'serial_number': '2B0195',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:71:13',
                                         'uptime': '13219345'},
                                 {       'counter_source': '419/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '419/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '230618543373',
                                                                       'in_ucast_pkts': '31162658000',
                                                                       'mac_address': 'ec:4f:82:5b:71:0b',
                                                                       'name': '419/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '233',
                                                                       'out_octets': '7522079505074',
                                                                       'out_ucast_pkts': '5205587570',
                                                                       'poller_hash': 'MkIwMTk0cGVlcmludGY0MTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '419',
                                                                       'in_nucast_pkts': '233',
                                                                       'in_octets': '7522079505074',
                                                                       'in_ucast_pkts': '5205587570',
                                                                       'links': [       {       'in_nucast_pkts': '233',
                                                                                                'in_octets': '7522079505074',
                                                                                                'in_ucast_pkts': '5205587570',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '230618543373',
                                                                                                'out_ucast_pkts': '31162658000',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '230618543373',
                                                                       'out_ucast_pkts': '31162658000',
                                                                       'poller_hash': 'MkIwMTk0cGVlcmludGZwb24gNDE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '419',
                                                                                               'links': [       {       'connected_time': '21816088',
                                                                                                                        'fiber_length': 1572,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTk0cGVlcnBvbnBvbiA0MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '68 MAIN ST N (YOST RACHEL)',
                                         'poller_hash': 'MkIwMTk0',
                                         'serial_number': '2B0194',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:71:0b',
                                         'uptime': '21816088'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '420/g1',
                                                                       'name': '420/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1RTFFcGVlcmludGY0MjAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '420',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1RTFFcGVlcmludGZwb24gNDIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '420',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RTFFcGVlcnBvbnBvbiA0MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '55 WILLIAM ST (HUBERTS KIRK)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "420" '
                                                                  '(location: GPON '
                                                                  'port "7/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-02-15T17:33:29.800Z',
                                         'poller_hash': 'MjA1RTFF',
                                         'serial_number': '205E1E',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '421/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '421/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '59152700555',
                                                                       'in_ucast_pkts': '101046137',
                                                                       'mac_address': 'ec:4f:82:5b:71:43',
                                                                       'name': '421/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '200',
                                                                       'out_octets': '182533714578',
                                                                       'out_ucast_pkts': '187818957',
                                                                       'poller_hash': 'MkIwMTlCcGVlcmludGY0MjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '421',
                                                                       'in_nucast_pkts': '200',
                                                                       'in_octets': '182533714578',
                                                                       'in_ucast_pkts': '187818957',
                                                                       'links': [       {       'in_nucast_pkts': '200',
                                                                                                'in_octets': '182533714578',
                                                                                                'in_ucast_pkts': '187818957',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '59152700555',
                                                                                                'out_ucast_pkts': '101046137',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '59152700555',
                                                                       'out_ucast_pkts': '101046137',
                                                                       'poller_hash': 'MkIwMTlCcGVlcmludGZwb24gNDIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '421',
                                                                                               'links': [       {       'connected_time': '64764128',
                                                                                                                        'fiber_length': 1782,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTlCcGVlcnBvbnBvbiA0MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "66 Main St. N. Milverton (Burn's "
                                                 'Presbyterian Church)',
                                         'poller_hash': 'MkIwMTlC',
                                         'serial_number': '2B019B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:71:43',
                                         'uptime': '64764128'},
                                 {       'counter_source': '422/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '422/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '56113053889',
                                                                       'in_ucast_pkts': '12966358882',
                                                                       'mac_address': 'ec:4f:82:5b:7f:ab',
                                                                       'name': '422/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '214',
                                                                       'out_octets': '346587674878',
                                                                       'out_ucast_pkts': '377951399',
                                                                       'poller_hash': 'MkIwMzY4cGVlcmludGY0MjIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '422',
                                                                       'in_nucast_pkts': '214',
                                                                       'in_octets': '346587674878',
                                                                       'in_ucast_pkts': '377951399',
                                                                       'links': [       {       'in_nucast_pkts': '214',
                                                                                                'in_octets': '346587674878',
                                                                                                'in_ucast_pkts': '377951399',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '56113053889',
                                                                                                'out_ucast_pkts': '12966358882',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '56113053889',
                                                                       'out_ucast_pkts': '12966358882',
                                                                       'poller_hash': 'MkIwMzY4cGVlcmludGZwb24gNDIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '422',
                                                                                               'links': [       {       'connected_time': '688338',
                                                                                                                        'fiber_length': 1205,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMzY4cGVlcnBvbnBvbiA0MjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '33 Temperance St. Milverton '
                                                 '(Sarah&Mary Martin 10402460)',
                                         'poller_hash': 'MkIwMzY4',
                                         'serial_number': '2B0368',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:7f:ab',
                                         'uptime': '688338'},
                                 {       'counter_source': '423/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:37:86:43:b0:d8'],
                                                                       'id': '423/g1',
                                                                       'in_nucast_pkts': '11670',
                                                                       'in_octets': '7912780094',
                                                                       'in_ucast_pkts': '41275628',
                                                                       'mac_address': '44:65:7f:7b:c5:5d',
                                                                       'mtu': 2000,
                                                                       'name': '423/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464172',
                                                                       'out_octets': '153163083184',
                                                                       'out_ucast_pkts': '112900608',
                                                                       'poller_hash': 'NUE2NkJBcGVlcmludGY0MjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '423',
                                                                       'in_nucast_pkts': '1464172',
                                                                       'in_octets': '153163083184',
                                                                       'in_ucast_pkts': '112900608',
                                                                       'links': [       {       'in_nucast_pkts': '1464172',
                                                                                                'in_octets': '153163083184',
                                                                                                'in_ucast_pkts': '112900608',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11670',
                                                                                                'out_octets': '7912780094',
                                                                                                'out_ucast_pkts': '41275628',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11670',
                                                                       'out_octets': '7912780094',
                                                                       'out_ucast_pkts': '41275628',
                                                                       'poller_hash': 'NUE2NkJBcGVlcmludGZwb24gNDIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '423',
                                                                                               'links': [       {       'connected_time': '688392',
                                                                                                                        'fiber_length': 1243,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NkJBcGVlcnBvbnBvbiA0MjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '6475 Perh Rd. 131 Milverton (Megan '
                                                 'Ray)',
                                         'poller_hash': 'NUE2NkJB',
                                         'serial_number': '5A66BA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c5:5d',
                                         'uptime': '688392'},
                                 {       'counter_source': '424/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '424/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '151366931829',
                                                                       'in_ucast_pkts': '26613710912',
                                                                       'mac_address': 'ec:4f:82:5b:87:0b',
                                                                       'name': '424/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '4269273357183',
                                                                       'out_ucast_pkts': '3111817644',
                                                                       'poller_hash': 'MkIwNDU0cGVlcmludGY0MjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '424',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '4269273357183',
                                                                       'in_ucast_pkts': '3111817644',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '4269273357183',
                                                                                                'in_ucast_pkts': '3111817644',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '151366931829',
                                                                                                'out_ucast_pkts': '26613710912',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '151366931829',
                                                                       'out_ucast_pkts': '26613710912',
                                                                       'poller_hash': 'MkIwNDU0cGVlcmludGZwb24gNDI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '424',
                                                                                               'links': [       {       'connected_time': '21815758',
                                                                                                                        'fiber_length': 848,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDU0cGVlcnBvbnBvbiA0MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 Pugh St. W. Milverton (Don '
                                                 'Hollinger)',
                                         'poller_hash': 'MkIwNDU0',
                                         'serial_number': '2B0454',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:87:0b',
                                         'uptime': '21815758'},
                                 {       'counter_source': '425/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '425/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '129576721655',
                                                                       'in_ucast_pkts': '545239413',
                                                                       'mac_address': 'ec:4f:82:5b:86:63',
                                                                       'name': '425/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '24',
                                                                       'out_octets': '2981215854333',
                                                                       'out_ucast_pkts': '2131144503',
                                                                       'poller_hash': 'MkIwNDNGcGVlcmludGY0MjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '425',
                                                                       'in_nucast_pkts': '24',
                                                                       'in_octets': '2981215854333',
                                                                       'in_ucast_pkts': '2131144503',
                                                                       'links': [       {       'in_nucast_pkts': '24',
                                                                                                'in_octets': '2981215854333',
                                                                                                'in_ucast_pkts': '2131144503',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '129576721655',
                                                                                                'out_ucast_pkts': '545239413',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '129576721655',
                                                                       'out_ucast_pkts': '545239413',
                                                                       'poller_hash': 'MkIwNDNGcGVlcmludGZwb24gNDI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '425',
                                                                                               'links': [       {       'connected_time': '688343',
                                                                                                                        'fiber_length': 1157,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDNGcGVlcnBvbnBvbiA0MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '8 Grosch St. Milverton (Angela '
                                                 'Schneider)',
                                         'poller_hash': 'MkIwNDNG',
                                         'serial_number': '2B043F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:86:63',
                                         'uptime': '688343'},
                                 {       'counter_source': '426/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a3:be'],
                                                                       'id': '426/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '6029158564',
                                                                       'in_ucast_pkts': '35230928',
                                                                       'mac_address': 'd0:76:8f:08:19:e6',
                                                                       'mtu': 2000,
                                                                       'name': '426/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1275653',
                                                                       'out_octets': '156149665458',
                                                                       'out_ucast_pkts': '113992920',
                                                                       'poller_hash': 'ODY5NDMycGVlcmludGY0MjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '426',
                                                                       'in_nucast_pkts': '1275653',
                                                                       'in_octets': '156149665458',
                                                                       'in_ucast_pkts': '113992920',
                                                                       'links': [       {       'in_nucast_pkts': '1275653',
                                                                                                'in_octets': '156149665458',
                                                                                                'in_ucast_pkts': '113992920',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '6029158564',
                                                                                                'out_ucast_pkts': '35230928',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '6029158564',
                                                                       'out_ucast_pkts': '35230928',
                                                                       'poller_hash': 'ODY5NDMycGVlcmludGZwb24gNDI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '426',
                                                                                               'links': [       {       'connected_time': '597791',
                                                                                                                        'fiber_length': 1872,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDMycGVlcnBvbnBvbiA0MjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '50 MAIN ST N (CAMERON JEFF)',
                                         'poller_hash': 'ODY5NDMy',
                                         'serial_number': '869432',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:e3',
                                         'uptime': '597791'},
                                 {       'counter_source': '427/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '427/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '549598168934',
                                                                       'in_ucast_pkts': '27059301396',
                                                                       'mac_address': 'ec:4f:82:5b:86:4b',
                                                                       'name': '427/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '200',
                                                                       'out_octets': '3928631845757',
                                                                       'out_ucast_pkts': '3167871949',
                                                                       'poller_hash': 'MkIwNDNDcGVlcmludGY0MjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '427',
                                                                       'in_nucast_pkts': '200',
                                                                       'in_octets': '3928631845757',
                                                                       'in_ucast_pkts': '3167871949',
                                                                       'links': [       {       'in_nucast_pkts': '200',
                                                                                                'in_octets': '3928631845757',
                                                                                                'in_ucast_pkts': '3167871949',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '549598168934',
                                                                                                'out_ucast_pkts': '27059301396',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '549598168934',
                                                                       'out_ucast_pkts': '27059301396',
                                                                       'poller_hash': 'MkIwNDNDcGVlcmludGZwb24gNDI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '427',
                                                                                               'links': [       {       'connected_time': '4927279',
                                                                                                                        'fiber_length': 1952,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDNDcGVlcnBvbnBvbiA0Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '38 Main St. N. Milverton (Chris '
                                                 'Donnelly)',
                                         'poller_hash': 'MkIwNDND',
                                         'serial_number': '2B043C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:86:4b',
                                         'uptime': '4927279'},
                                 {       'counter_source': '428/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a0:e9'],
                                                                       'id': '428/g1',
                                                                       'in_nucast_pkts': '75',
                                                                       'in_octets': '153963567826',
                                                                       'in_ucast_pkts': '409357579',
                                                                       'mac_address': 'ec:4f:82:5b:87:7e',
                                                                       'mtu': 2000,
                                                                       'name': '428/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15193058',
                                                                       'out_octets': '1844170042210',
                                                                       'out_ucast_pkts': '1287954961',
                                                                       'poller_hash': 'MkIwNDYycGVlcmludGY0MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '428',
                                                                       'in_nucast_pkts': '15193058',
                                                                       'in_octets': '1844170042210',
                                                                       'in_ucast_pkts': '1287954961',
                                                                       'links': [       {       'in_nucast_pkts': '15193058',
                                                                                                'in_octets': '1844170042210',
                                                                                                'in_ucast_pkts': '1287954961',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '75',
                                                                                                'out_octets': '153963567826',
                                                                                                'out_ucast_pkts': '409357579',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75',
                                                                       'out_octets': '153963567826',
                                                                       'out_ucast_pkts': '409357579',
                                                                       'poller_hash': 'MkIwNDYycGVlcmludGZwb24gNDI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '428',
                                                                                               'links': [       {       'connected_time': '6487380',
                                                                                                                        'fiber_length': 851,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDYycGVlcnBvbnBvbiA0Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 Pugh St. W. Milverton (Ken Zehr '
                                                 '10429938)',
                                         'poller_hash': 'MkIwNDYy',
                                         'serial_number': '2B0462',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:87:7b',
                                         'uptime': '6487380'},
                                 {       'counter_source': '429/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a6:41'],
                                                                       'id': '429/g1',
                                                                       'in_nucast_pkts': '9028',
                                                                       'in_octets': '1278341985153',
                                                                       'in_ucast_pkts': '5311681210',
                                                                       'mac_address': 'ec:4f:82:5b:86:ce',
                                                                       'mtu': 2000,
                                                                       'name': '429/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149457820',
                                                                       'out_octets': '17748755111674',
                                                                       'out_ucast_pkts': '13097536764',
                                                                       'poller_hash': 'MkIwNDRDcGVlcmludGY0MjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '429',
                                                                       'in_nucast_pkts': '149457820',
                                                                       'in_octets': '17748755111674',
                                                                       'in_ucast_pkts': '13097536764',
                                                                       'links': [       {       'in_nucast_pkts': '149457820',
                                                                                                'in_octets': '17748755111674',
                                                                                                'in_ucast_pkts': '13097536764',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9028',
                                                                                                'out_octets': '1278341985153',
                                                                                                'out_ucast_pkts': '5311681210',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9028',
                                                                       'out_octets': '1278341985153',
                                                                       'out_ucast_pkts': '5311681210',
                                                                       'poller_hash': 'MkIwNDRDcGVlcmludGZwb24gNDI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '429',
                                                                                               'links': [       {       'connected_time': '21815431',
                                                                                                                        'fiber_length': 1163,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDRDcGVlcnBvbnBvbiA0Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '31 Temperance St. Milverton (Peter '
                                                 'Ryter 10425942)',
                                         'poller_hash': 'MkIwNDRD',
                                         'serial_number': '2B044C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:86:cb',
                                         'uptime': '21815431'},
                                 {       'counter_source': '430/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '430/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '107771545222',
                                                                       'in_ucast_pkts': '576014792',
                                                                       'mac_address': 'ec:4f:82:5b:40:0b',
                                                                       'name': '430/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '179',
                                                                       'out_octets': '1875264942527',
                                                                       'out_ucast_pkts': '1553528033',
                                                                       'poller_hash': 'MkFGQjc0cGVlcmludGY0MzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '430',
                                                                       'in_nucast_pkts': '179',
                                                                       'in_octets': '1875264942527',
                                                                       'in_ucast_pkts': '1553528033',
                                                                       'links': [       {       'in_nucast_pkts': '179',
                                                                                                'in_octets': '1875264942527',
                                                                                                'in_ucast_pkts': '1553528033',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '107771545222',
                                                                                                'out_ucast_pkts': '576014792',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '107771545222',
                                                                       'out_ucast_pkts': '576014792',
                                                                       'poller_hash': 'MkFGQjc0cGVlcmludGZwb24gNDMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '430',
                                                                                               'links': [       {       'connected_time': '11510158',
                                                                                                                        'fiber_length': 1337,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFGQjc0cGVlcnBvbnBvbiA0MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 Spencer St. Milverton (Jason '
                                                 'Yantzi)',
                                         'poller_hash': 'MkFGQjc0',
                                         'serial_number': '2AFB74',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:40:0b',
                                         'uptime': '11510158'},
                                 {       'counter_source': '431/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '431/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3207379064',
                                                                       'in_ucast_pkts': '18501115',
                                                                       'mac_address': 'ec:4f:82:5b:86:8b',
                                                                       'name': '431/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '70509937817',
                                                                       'out_ucast_pkts': '64210620',
                                                                       'poller_hash': 'MkIwNDQ0cGVlcmludGY0MzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '431',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '70509937817',
                                                                       'in_ucast_pkts': '64210620',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '70509937817',
                                                                                                'in_ucast_pkts': '64210620',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3207379064',
                                                                                                'out_ucast_pkts': '18501115',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3207379064',
                                                                       'out_ucast_pkts': '18501115',
                                                                       'poller_hash': 'MkIwNDQ0cGVlcmludGZwb24gNDMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '431',
                                                                                               'links': [       {       'connected_time': '21815373',
                                                                                                                        'fiber_length': 1137,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDQ0cGVlcnBvbnBvbiA0MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '42 MAIN ST S (2495008 ONTARIO LTD.)',
                                         'poller_hash': 'MkIwNDQ0',
                                         'serial_number': '2B0444',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:86:8b',
                                         'uptime': '21815373'},
                                 {       'counter_source': '432/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2d:34:75'],
                                                                       'id': '432/g1',
                                                                       'in_nucast_pkts': '231',
                                                                       'in_octets': '62444615293',
                                                                       'in_ucast_pkts': '520375644',
                                                                       'mac_address': 'ec:4f:82:7e:09:64',
                                                                       'mtu': 2000,
                                                                       'name': '432/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '69117245',
                                                                       'out_octets': '2865613830770',
                                                                       'out_ucast_pkts': '1951612121',
                                                                       'poller_hash': 'MzVFOTZDcGVlcmludGY0MzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '432',
                                                                       'in_nucast_pkts': '69117245',
                                                                       'in_octets': '2865613830770',
                                                                       'in_ucast_pkts': '1951612121',
                                                                       'links': [       {       'in_nucast_pkts': '69117245',
                                                                                                'in_octets': '2865613830770',
                                                                                                'in_ucast_pkts': '1951612121',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '231',
                                                                                                'out_octets': '62444615293',
                                                                                                'out_ucast_pkts': '520375644',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '231',
                                                                       'out_octets': '62444615293',
                                                                       'out_ucast_pkts': '520375644',
                                                                       'poller_hash': 'MzVFOTZDcGVlcmludGZwb24gNDMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '432',
                                                                                               'links': [       {       'connected_time': '29381420',
                                                                                                                        'fiber_length': 1030,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTZDcGVlcnBvbnBvbiA0MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '24 temperance St. Milverton (Royal '
                                                 'Can. Legion)',
                                         'poller_hash': 'MzVFOTZD',
                                         'serial_number': '35E96C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:64',
                                         'uptime': '29381420'},
                                 {       'counter_source': '434/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '434/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12205938781',
                                                                       'in_ucast_pkts': '63058703',
                                                                       'mac_address': 'ec:4f:82:5b:70:9b',
                                                                       'name': '434/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '302598056283',
                                                                       'out_ucast_pkts': '270496190',
                                                                       'poller_hash': 'MkIwMTg2cGVlcmludGY0MzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '434',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '302598056283',
                                                                       'in_ucast_pkts': '270496190',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '302598056283',
                                                                                                'in_ucast_pkts': '270496190',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '12205938781',
                                                                                                'out_ucast_pkts': '63058703',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12205938781',
                                                                       'out_ucast_pkts': '63058703',
                                                                       'poller_hash': 'MkIwMTg2cGVlcmludGZwb24gNDM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '434',
                                                                                               'links': [       {       'connected_time': '2144588',
                                                                                                                        'fiber_length': 1186,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTg2cGVlcnBvbnBvbiA0MzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '32 Mill St. E. Milverton (Brendan '
                                                 'Barrett)',
                                         'poller_hash': 'MkIwMTg2',
                                         'serial_number': '2B0186',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:9b',
                                         'uptime': '2144588'},
                                 {       'counter_source': '439/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b7:fa'],
                                                                       'id': '439/g1',
                                                                       'in_nucast_pkts': '10014',
                                                                       'in_octets': '298040567717',
                                                                       'in_ucast_pkts': '1568677074',
                                                                       'mac_address': 'ec:4f:82:5b:85:76',
                                                                       'mtu': 2000,
                                                                       'name': '439/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '148485926',
                                                                       'out_octets': '11058346003464',
                                                                       'out_ucast_pkts': '8112337723',
                                                                       'poller_hash': 'MkIwNDIxcGVlcmludGY0MzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '439',
                                                                       'in_nucast_pkts': '148485926',
                                                                       'in_octets': '11058346003464',
                                                                       'in_ucast_pkts': '8112337723',
                                                                       'links': [       {       'in_nucast_pkts': '148485926',
                                                                                                'in_octets': '11058346003464',
                                                                                                'in_ucast_pkts': '8112337723',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10014',
                                                                                                'out_octets': '298040567717',
                                                                                                'out_ucast_pkts': '1568677074',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10014',
                                                                       'out_octets': '298040567717',
                                                                       'out_ucast_pkts': '1568677074',
                                                                       'poller_hash': 'MkIwNDIxcGVlcmludGZwb24gNDM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '439',
                                                                                               'links': [       {       'connected_time': '21816307',
                                                                                                                        'fiber_length': 992,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDIxcGVlcnBvbnBvbiA0Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 Mill St. W. Milverton (Llori '
                                                 'Nicholls)',
                                         'poller_hash': 'MkIwNDIx',
                                         'serial_number': '2B0421',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:85:73',
                                         'uptime': '21816307'},
                                 {       'counter_source': '444/g1+444/g2+444/g3+444/g4+444/g6+444/g7+444/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:db:46:da'],
                                                                       'id': '444/g1',
                                                                       'in_nucast_pkts': '1382901',
                                                                       'in_octets': '337494150',
                                                                       'in_ucast_pkts': '1123284',
                                                                       'mac_address': '00:06:31:ca:5e:33',
                                                                       'mtu': 2000,
                                                                       'name': '444/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1608889463',
                                                                       'out_octets': '2144839540587',
                                                                       'out_ucast_pkts': '847674',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGY0NDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:6e:9a'],
                                                                       'id': '444/g2',
                                                                       'in_nucast_pkts': '1392591',
                                                                       'in_octets': '346645672',
                                                                       'in_ucast_pkts': '1195711',
                                                                       'mac_address': '00:06:31:ca:5e:34',
                                                                       'mtu': 2000,
                                                                       'name': '444/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2393401902',
                                                                       'out_octets': '3213127267838',
                                                                       'out_ucast_pkts': '906337',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGY0NDQvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5b:d8'],
                                                                       'id': '444/g3',
                                                                       'in_nucast_pkts': '1324353',
                                                                       'in_octets': '4270752573',
                                                                       'in_ucast_pkts': '45032391',
                                                                       'mac_address': '00:06:31:ca:5e:35',
                                                                       'mtu': 2000,
                                                                       'name': '444/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55121813',
                                                                       'out_octets': '220331638609',
                                                                       'out_ucast_pkts': '142654575',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGY0NDQvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:48:5f'],
                                                                       'id': '444/g4',
                                                                       'in_nucast_pkts': '808210',
                                                                       'in_octets': '25967109232',
                                                                       'in_ucast_pkts': '246256475',
                                                                       'mac_address': '00:06:31:ca:5e:36',
                                                                       'mtu': 2000,
                                                                       'name': '444/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1983922271',
                                                                       'out_octets': '4108281035282',
                                                                       'out_ucast_pkts': '970258638',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGY0NDQvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:30:52'],
                                                                       'id': '444/g6',
                                                                       'in_nucast_pkts': '1371776',
                                                                       'in_octets': '321163991',
                                                                       'in_ucast_pkts': '991614',
                                                                       'mac_address': 'ec:4f:82:27:85:41',
                                                                       'mtu': 2000,
                                                                       'name': '444/g6',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2300731126',
                                                                       'out_octets': '3088737155025',
                                                                       'out_ucast_pkts': '729071',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGY0NDQvZzY=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:74:50'],
                                                                       'id': '444/g7',
                                                                       'in_nucast_pkts': '1206042',
                                                                       'in_octets': '1426543399',
                                                                       'in_ucast_pkts': '11929674',
                                                                       'mac_address': 'ec:4f:82:27:85:42',
                                                                       'mtu': 2000,
                                                                       'name': '444/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '105438602',
                                                                       'out_octets': '135388916734',
                                                                       'out_ucast_pkts': '36196354',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGY0NDQvZzc=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:42:65:20'],
                                                                       'id': '444/g8',
                                                                       'in_nucast_pkts': '1245097',
                                                                       'in_octets': '2309615305',
                                                                       'in_ucast_pkts': '18814711',
                                                                       'mac_address': 'ec:4f:82:27:85:43',
                                                                       'mtu': 2000,
                                                                       'name': '444/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '225916206',
                                                                       'out_octets': '327089845907',
                                                                       'out_ucast_pkts': '54355169',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGY0NDQvZzg=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '444',
                                                                       'in_nucast_pkts': '8673421383',
                                                                       'in_octets': '13237795399982',
                                                                       'in_ucast_pkts': '1205947818',
                                                                       'links': [       {       'in_nucast_pkts': '8673421383',
                                                                                                'in_octets': '13237795399982',
                                                                                                'in_ucast_pkts': '1205947818',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8730970',
                                                                                                'out_octets': '34979324322',
                                                                                                'out_ucast_pkts': '325343860',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8730970',
                                                                       'out_octets': '34979324322',
                                                                       'out_ucast_pkts': '325343860',
                                                                       'poller_hash': 'MjYzMjAwcGVlcmludGZwb24gNDQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '444',
                                                                                               'links': [       {       'connected_time': '4164919',
                                                                                                                        'fiber_length': 2091,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYzMjAwcGVlcnBvbnBvbiA0NDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST (BRUCE WILLIAM)',
                                         'poller_hash': 'MjYzMjAw',
                                         'serial_number': '263200',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ca:5e:38',
                                         'uptime': '4164919'},
                                 {       'counter_source': '445/g2+445/g3+445/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:15:4f:0a'],
                                                                       'id': '445/g2',
                                                                       'in_nucast_pkts': '1430502',
                                                                       'in_octets': '336054737',
                                                                       'in_ucast_pkts': '1029228',
                                                                       'mac_address': '00:06:31:ca:5e:04',
                                                                       'mtu': 2000,
                                                                       'name': '445/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2688068185',
                                                                       'out_octets': '3614604977834',
                                                                       'out_ucast_pkts': '774661',
                                                                       'poller_hash': 'MjYzMjAxcGVlcmludGY0NDUvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5b:e0'],
                                                                       'id': '445/g3',
                                                                       'in_nucast_pkts': '1287283',
                                                                       'in_octets': '14550174879',
                                                                       'in_ucast_pkts': '191592843',
                                                                       'mac_address': '00:06:31:ca:5e:05',
                                                                       'mtu': 2000,
                                                                       'name': '445/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1015774456',
                                                                       'out_octets': '2208535675818',
                                                                       'out_ucast_pkts': '630232084',
                                                                       'poller_hash': 'MjYzMjAxcGVlcmludGY0NDUvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:58:28'],
                                                                       'id': '445/g4',
                                                                       'in_nucast_pkts': '1323467',
                                                                       'in_octets': '19727903450',
                                                                       'in_ucast_pkts': '261132785',
                                                                       'mac_address': '00:06:31:ca:5e:06',
                                                                       'mtu': 2000,
                                                                       'name': '445/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55113554',
                                                                       'out_octets': '1219462014526',
                                                                       'out_ucast_pkts': '853411234',
                                                                       'poller_hash': 'MjYzMjAxcGVlcmludGY0NDUvZzQ=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '445',
                                                                       'in_nucast_pkts': '3758956195',
                                                                       'in_octets': '7042602668178',
                                                                       'in_ucast_pkts': '1484417979',
                                                                       'links': [       {       'in_nucast_pkts': '3758956195',
                                                                                                'in_octets': '7042602668178',
                                                                                                'in_ucast_pkts': '1484417979',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4041252',
                                                                                                'out_octets': '34614133066',
                                                                                                'out_ucast_pkts': '453754856',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4041252',
                                                                       'out_octets': '34614133066',
                                                                       'out_ucast_pkts': '453754856',
                                                                       'poller_hash': 'MjYzMjAxcGVlcmludGZwb24gNDQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '445',
                                                                                               'links': [       {       'connected_time': '4164937',
                                                                                                                        'fiber_length': 2090,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYzMjAxcGVlcnBvbnBvbiA0NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST',
                                         'poller_hash': 'MjYzMjAx',
                                         'serial_number': '263201',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ca:5e:08',
                                         'uptime': '4164937'},
                                 {       'counter_source': '446/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '446/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '99755286517',
                                                                       'in_ucast_pkts': '488590493',
                                                                       'mac_address': 'ec:4f:82:0d:6a:f0',
                                                                       'name': '446/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '3041233838786',
                                                                       'out_ucast_pkts': '10023435811',
                                                                       'poller_hash': 'MjA1RTAzcGVlcmludGY0NDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '446',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '3041233838786',
                                                                       'in_ucast_pkts': '10023435811',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '3041233838786',
                                                                                                'in_ucast_pkts': '10023435811',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '99755286517',
                                                                                                'out_ucast_pkts': '488590493',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '99755286517',
                                                                       'out_ucast_pkts': '488590493',
                                                                       'poller_hash': 'MjA1RTAzcGVlcmludGZwb24gNDQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '446',
                                                                                               'links': [       {       'connected_time': '9448326',
                                                                                                                        'fiber_length': 1377,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RTAzcGVlcnBvbnBvbiA0NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '57 Temperance St. Milverton (Jim '
                                                 'Schneuker 10401435)',
                                         'poller_hash': 'MjA1RTAz',
                                         'serial_number': '205E03',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:f0',
                                         'uptime': '9448326'},
                                 {       'counter_source': '447/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '447/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '752047738474',
                                                                       'in_ucast_pkts': '260199755',
                                                                       'mac_address': 'ec:4f:82:0d:6a:33',
                                                                       'name': '447/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '150',
                                                                       'out_octets': '1970273090760',
                                                                       'out_ucast_pkts': '1930719927',
                                                                       'poller_hash': 'MjA1REVFcGVlcmludGY0NDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '447',
                                                                       'in_nucast_pkts': '150',
                                                                       'in_octets': '1970273090760',
                                                                       'in_ucast_pkts': '1930719927',
                                                                       'links': [       {       'in_nucast_pkts': '150',
                                                                                                'in_octets': '1970273090760',
                                                                                                'in_ucast_pkts': '1930719927',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '752047738474',
                                                                                                'out_ucast_pkts': '260199755',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '752047738474',
                                                                       'out_ucast_pkts': '260199755',
                                                                       'poller_hash': 'MjA1REVFcGVlcmludGZwb24gNDQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '447',
                                                                                               'links': [       {       'connected_time': '21816546',
                                                                                                                        'fiber_length': 1560,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REVFcGVlcnBvbnBvbiA0NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '76 MAIN ST S (UHC OF WESTERN ONTARIO '
                                                 'MCO 839)',
                                         'poller_hash': 'MjA1REVF',
                                         'serial_number': '205DEE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:33',
                                         'uptime': '21816546'},
                                 {       'counter_source': '448/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:03:8e'],
                                                                       'id': '448/g1',
                                                                       'in_nucast_pkts': '89',
                                                                       'in_octets': '717724481',
                                                                       'in_ucast_pkts': '4559522',
                                                                       'mac_address': 'ec:4f:82:7e:0b:ec',
                                                                       'mtu': 2000,
                                                                       'name': '448/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11700022',
                                                                       'out_octets': '13642307463',
                                                                       'out_ucast_pkts': '10040572',
                                                                       'poller_hash': 'MzVFQTQ0cGVlcmludGY0NDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '448',
                                                                       'in_nucast_pkts': '11700022',
                                                                       'in_octets': '13642307463',
                                                                       'in_ucast_pkts': '10040572',
                                                                       'links': [       {       'in_nucast_pkts': '11700022',
                                                                                                'in_octets': '13642307463',
                                                                                                'in_ucast_pkts': '10040572',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '89',
                                                                                                'out_octets': '717724481',
                                                                                                'out_ucast_pkts': '4559522',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '89',
                                                                       'out_octets': '717724481',
                                                                       'out_ucast_pkts': '4559522',
                                                                       'poller_hash': 'MzVFQTQ0cGVlcmludGZwb24gNDQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '448',
                                                                                               'links': [       {       'connected_time': '5144138',
                                                                                                                        'fiber_length': 1330,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQTQ0cGVlcnBvbnBvbiA0NDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '21 Whaley Av. Milverton (Murray '
                                                 'Schneider)',
                                         'poller_hash': 'MzVFQTQ0',
                                         'serial_number': '35EA44',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0b:ec',
                                         'uptime': '5144138'},
                                 {       'counter_source': '449/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c4:e9:0a:56:56:57'],
                                                                       'id': '449/g1',
                                                                       'in_nucast_pkts': '4473',
                                                                       'in_octets': '26727962413',
                                                                       'in_ucast_pkts': '100371554',
                                                                       'mac_address': 'ec:4f:82:7d:bc:b9',
                                                                       'mtu': 2000,
                                                                       'name': '449/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465319',
                                                                       'out_octets': '286926074231',
                                                                       'out_ucast_pkts': '211824315',
                                                                       'poller_hash': 'MzAwN0M1cGVlcmludGY0NDkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '449',
                                                                       'in_nucast_pkts': '1465319',
                                                                       'in_octets': '286926074231',
                                                                       'in_ucast_pkts': '211824315',
                                                                       'links': [       {       'in_nucast_pkts': '1465319',
                                                                                                'in_octets': '286926074231',
                                                                                                'in_ucast_pkts': '211824315',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4473',
                                                                                                'out_octets': '26727962413',
                                                                                                'out_ucast_pkts': '100371554',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4473',
                                                                       'out_octets': '26727962413',
                                                                       'out_ucast_pkts': '100371554',
                                                                       'poller_hash': 'MzAwN0M1cGVlcmludGZwb24gNDQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '449',
                                                                                               'links': [       {       'connected_time': '688451',
                                                                                                                        'fiber_length': 1170,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwN0M1cGVlcnBvbnBvbiA0NDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '26 Mill St. E. Milverton (Tony '
                                                 'Douglas)',
                                         'poller_hash': 'MzAwN0M1',
                                         'serial_number': '3007C5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:bc:b9',
                                         'uptime': '688451'},
                                 {       'counter_source': '450/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '450/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '351828381130',
                                                                       'in_ucast_pkts': '1310709048',
                                                                       'mac_address': 'ec:4f:82:0d:69:d0',
                                                                       'name': '450/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '191',
                                                                       'out_octets': '5538457341459',
                                                                       'out_ucast_pkts': '4058871678',
                                                                       'poller_hash': 'MjA1REUzcGVlcmludGY0NTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '450',
                                                                       'in_nucast_pkts': '191',
                                                                       'in_octets': '5538457341459',
                                                                       'in_ucast_pkts': '4058871678',
                                                                       'links': [       {       'in_nucast_pkts': '191',
                                                                                                'in_octets': '5538457341459',
                                                                                                'in_ucast_pkts': '4058871678',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '351828381130',
                                                                                                'out_ucast_pkts': '1310709048',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '351828381130',
                                                                       'out_ucast_pkts': '1310709048',
                                                                       'poller_hash': 'MjA1REUzcGVlcmludGZwb24gNDUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '450',
                                                                                               'links': [       {       'connected_time': '6643424',
                                                                                                                        'fiber_length': 878,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REUzcGVlcnBvbnBvbiA0NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '34 Pacific Av Milverton (Jeremy '
                                                 'Matheson)',
                                         'poller_hash': 'MjA1REUz',
                                         'serial_number': '205DE3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:d0',
                                         'uptime': '6643424'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '451/g1',
                                                                       'name': '451/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTM4MTNDcGVlcmludGY0NTEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '451',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTM4MTNDcGVlcmludGZwb24gNDUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '451',
                                                                                               'links': [       {       'connected_time': '4164970',
                                                                                                                        'fiber_length': 2085,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTM4MTNDcGVlcnBvbnBvbiA0NTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 William St. Milverton (Knollcrest '
                                                 'Data Rm. 3)',
                                         'poller_hash': 'MTM4MTND',
                                         'serial_number': '13813C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:9c:c2:08',
                                         'uptime': '4164970'},
                                 {       'counter_source': '452/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:15:65:d4:bb:53',
                                                                                                        '00:21:b7:81:63:4b',
                                                                                                        '00:23:24:73:26:76',
                                                                                                        '00:23:24:73:26:f4',
                                                                                                        '00:d8:61:1d:63:0c',
                                                                                                        '00:d8:61:44:bf:14',
                                                                                                        '00:d8:61:45:96:e0',
                                                                                                        '00:d8:61:b5:97:bf',
                                                                                                        '1c:ee:c9:20:75:c7',
                                                                                                        '1c:ee:c9:2a:5b:83',
                                                                                                        '24:9a:d8:2b:8b:69',
                                                                                                        '24:9a:d8:6d:ec:18',
                                                                                                        '30:9c:23:76:22:06',
                                                                                                        '30:9c:23:76:22:35',
                                                                                                        '3c:e3:6b:44:d2:e7',
                                                                                                        '3c:e3:6b:e3:96:18',
                                                                                                        '44:37:e6:7f:79:61',
                                                                                                        '44:db:d2:22:8b:e2',
                                                                                                        '44:db:d2:44:b6:da',
                                                                                                        '44:db:d2:6b:ae:fc',
                                                                                                        '44:db:d2:6b:b0:9d',
                                                                                                        '4c:cc:6a:1c:7c:a8',
                                                                                                        '6c:4b:90:ed:6c:04',
                                                                                                        '6c:4b:90:ed:6c:e9',
                                                                                                        '80:5e:0c:44:4c:85',
                                                                                                        '80:5e:0c:cb:31:02',
                                                                                                        '80:5e:0c:cd:ac:4b',
                                                                                                        '80:5e:0c:f4:ab:9c',
                                                                                                        '80:5e:c0:14:73:d4',
                                                                                                        '80:5e:c0:1c:fe:22',
                                                                                                        '80:5e:c0:1c:fe:c2',
                                                                                                        '80:5e:c0:1d:00:03',
                                                                                                        '80:5e:c0:1d:01:9e',
                                                                                                        '80:5e:c0:1d:03:36',
                                                                                                        '80:5e:c0:2f:63:44',
                                                                                                        '80:5e:c0:2f:64:18',
                                                                                                        '80:5e:c0:2f:69:51',
                                                                                                        'cc:d5:39:1d:4d:82',
                                                                                                        'cc:d5:39:1d:4d:c0',
                                                                                                        'cc:d5:39:1d:4d:c1',
                                                                                                        'd8:cb:8a:92:e1:bc',
                                                                                                        'f4:93:9f:f0:69:cd',
                                                                                                        'fc:4d:d4:d4:0a:a8',
                                                                                                        'fc:4d:d4:f4:32:bb'],
                                                                       'id': '452/g2',
                                                                       'in_nucast_pkts': '17801657',
                                                                       'in_octets': '2879405891833',
                                                                       'in_ucast_pkts': '4400959361',
                                                                       'mac_address': 'ec:4f:82:be:ac:9e',
                                                                       'mtu': 2000,
                                                                       'name': '452/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '104403874',
                                                                       'out_octets': '6178245977018',
                                                                       'out_ucast_pkts': '5534414511',
                                                                       'poller_hash': 'Mzc4OEQwcGVlcmludGY0NTIvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '452',
                                                                       'in_nucast_pkts': '104403874',
                                                                       'in_octets': '6178245977018',
                                                                       'in_ucast_pkts': '5534414511',
                                                                       'links': [       {       'in_nucast_pkts': '104403874',
                                                                                                'in_octets': '6178245977018',
                                                                                                'in_ucast_pkts': '5534414511',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '17801657',
                                                                                                'out_octets': '2879405891833',
                                                                                                'out_ucast_pkts': '4400959361',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17801657',
                                                                       'out_octets': '2879405891833',
                                                                       'out_ucast_pkts': '4400959361',
                                                                       'poller_hash': 'Mzc4OEQwcGVlcmludGZwb24gNDUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '452',
                                                                                               'links': [       {       'connected_time': '4468821',
                                                                                                                        'fiber_length': 1711,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.02,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzc4OEQwcGVlcnBvbnBvbiA0NTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '50 William St.  Milverton '
                                                 '(Knollcrest Data Rm 3)',
                                         'poller_hash': 'Mzc4OEQw',
                                         'serial_number': '3788D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:be:ac:99',
                                         'uptime': '4468821'},
                                 {       'counter_source': '453/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '453/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5704110957',
                                                                       'in_ucast_pkts': '13473315',
                                                                       'mac_address': 'ec:4f:82:5b:84:3b',
                                                                       'name': '453/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '30733253407',
                                                                       'out_ucast_pkts': '39683778',
                                                                       'poller_hash': 'MkIwM0ZBcGVlcmludGY0NTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '453',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '30733253407',
                                                                       'in_ucast_pkts': '39683778',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '30733253407',
                                                                                                'in_ucast_pkts': '39683778',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '5704110957',
                                                                                                'out_ucast_pkts': '13473315',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5704110957',
                                                                       'out_ucast_pkts': '13473315',
                                                                       'poller_hash': 'MkIwM0ZBcGVlcmludGZwb24gNDUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '453',
                                                                                               'links': [       {       'connected_time': '72961741',
                                                                                                                        'fiber_length': 813,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwM0ZBcGVlcnBvbnBvbiA0NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '39 Main St. S. Milverton (United '
                                                 'Church 10401217)',
                                         'poller_hash': 'MkIwM0ZB',
                                         'serial_number': '2B03FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:84:3b',
                                         'uptime': '72961741'},
                                 {       'counter_source': '454/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:0b:3a'],
                                                                       'id': '454/g4',
                                                                       'in_nucast_pkts': '9460285',
                                                                       'in_octets': '29428829106',
                                                                       'in_ucast_pkts': '343820560',
                                                                       'mac_address': 'ec:4f:82:5b:72:21',
                                                                       'mtu': 2000,
                                                                       'name': '454/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27029088242',
                                                                       'out_octets': '37653936252096',
                                                                       'out_ucast_pkts': '1222795098',
                                                                       'poller_hash': 'MkIwMUI2cGVlcmludGY0NTQvZzQ=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '454',
                                                                       'in_nucast_pkts': '27029088242',
                                                                       'in_octets': '37653936252096',
                                                                       'in_ucast_pkts': '1222795098',
                                                                       'links': [       {       'in_nucast_pkts': '27029088242',
                                                                                                'in_octets': '37653936252096',
                                                                                                'in_ucast_pkts': '1222795098',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9460285',
                                                                                                'out_octets': '29428829106',
                                                                                                'out_ucast_pkts': '343820560',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9460285',
                                                                       'out_octets': '29428829106',
                                                                       'out_ucast_pkts': '343820560',
                                                                       'poller_hash': 'MkIwMUI2cGVlcmludGZwb24gNDU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '454',
                                                                                               'links': [       {       'connected_time': '21815553',
                                                                                                                        'fiber_length': 1357,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMUI2cGVlcnBvbnBvbiA0NTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 Spencer St. Milverton (Neil '
                                                 'Schlueter)',
                                         'poller_hash': 'MkIwMUI2',
                                         'serial_number': '2B01B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:72:1b',
                                         'uptime': '21815553'},
                                 {       'counter_source': '455/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:04:42'],
                                                                       'id': '455/g1',
                                                                       'in_nucast_pkts': '237',
                                                                       'in_octets': '271826119047',
                                                                       'in_ucast_pkts': '2540438645',
                                                                       'mac_address': 'ec:4f:82:5b:73:5e',
                                                                       'mtu': 2000,
                                                                       'name': '455/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '92619877',
                                                                       'out_octets': '12189714986270',
                                                                       'out_ucast_pkts': '8273739637',
                                                                       'poller_hash': 'MkIwMURFcGVlcmludGY0NTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '455',
                                                                       'in_nucast_pkts': '92619877',
                                                                       'in_octets': '12189714986270',
                                                                       'in_ucast_pkts': '8273739637',
                                                                       'links': [       {       'in_nucast_pkts': '92619877',
                                                                                                'in_octets': '12189714986270',
                                                                                                'in_ucast_pkts': '8273739637',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '237',
                                                                                                'out_octets': '271826119047',
                                                                                                'out_ucast_pkts': '2540438645',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '237',
                                                                       'out_octets': '271826119047',
                                                                       'out_ucast_pkts': '2540438645',
                                                                       'poller_hash': 'MkIwMURFcGVlcmludGZwb24gNDU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '455',
                                                                                               'links': [       {       'connected_time': '70681266',
                                                                                                                        'fiber_length': 1868,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMURFcGVlcnBvbnBvbiA0NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '46 Main St. N. Milverton (John '
                                                 'Kilmer)',
                                         'poller_hash': 'MkIwMURF',
                                         'serial_number': '2B01DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:73:5b',
                                         'uptime': '70681266'},
                                 {       'counter_source': '457/g1+457/g2+457/g3+457/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:71:55'],
                                                                       'id': '457/g1',
                                                                       'in_nucast_pkts': '204086',
                                                                       'in_octets': '61330785644',
                                                                       'in_ucast_pkts': '516927538',
                                                                       'mac_address': 'ec:4f:82:34:64:73',
                                                                       'mtu': 2000,
                                                                       'name': '457/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17649240',
                                                                       'out_octets': '3197957951832',
                                                                       'out_ucast_pkts': '2149786594',
                                                                       'poller_hash': 'MUZENkRCcGVlcmludGY0NTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:04:54'],
                                                                       'id': '457/g2',
                                                                       'in_nucast_pkts': '38',
                                                                       'in_octets': '62536682734',
                                                                       'in_ucast_pkts': '649389975',
                                                                       'mac_address': 'ec:4f:82:34:64:74',
                                                                       'mtu': 2000,
                                                                       'name': '457/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17649120',
                                                                       'out_octets': '3392898001392',
                                                                       'out_ucast_pkts': '2296115284',
                                                                       'poller_hash': 'MUZENkRCcGVlcmludGY0NTcvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:eb:89'],
                                                                       'id': '457/g3',
                                                                       'in_nucast_pkts': '130',
                                                                       'in_octets': '54782077773',
                                                                       'in_ucast_pkts': '163687958',
                                                                       'mac_address': 'ec:4f:82:34:64:75',
                                                                       'mtu': 2000,
                                                                       'name': '457/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17669774',
                                                                       'out_octets': '936513889104',
                                                                       'out_ucast_pkts': '713791033',
                                                                       'poller_hash': 'MUZENkRCcGVlcmludGY0NTcvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:94:c1'],
                                                                       'id': '457/g4',
                                                                       'in_nucast_pkts': '181515',
                                                                       'in_octets': '41865704261',
                                                                       'in_ucast_pkts': '244088986',
                                                                       'mac_address': 'ec:4f:82:34:64:76',
                                                                       'mtu': 2000,
                                                                       'name': '457/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17669366',
                                                                       'out_octets': '2809968426430',
                                                                       'out_ucast_pkts': '1936058592',
                                                                       'poller_hash': 'MUZENkRCcGVlcmludGY0NTcvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '457',
                                                                       'in_nucast_pkts': '70637500',
                                                                       'in_octets': '10337338268758',
                                                                       'in_ucast_pkts': '7095751503',
                                                                       'links': [       {       'in_nucast_pkts': '70637500',
                                                                                                'in_octets': '10337338268758',
                                                                                                'in_ucast_pkts': '7095751503',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '385769',
                                                                                                'out_octets': '220515250412',
                                                                                                'out_ucast_pkts': '1574094457',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '385769',
                                                                       'out_octets': '220515250412',
                                                                       'out_ucast_pkts': '1574094457',
                                                                       'poller_hash': 'MUZENkRCcGVlcmludGZwb24gNDU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '457',
                                                                                               'links': [       {       'connected_time': '9954635',
                                                                                                                        'fiber_length': 1059,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.188,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUZENkRCcGVlcnBvbnBvbiA0NTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': "2 Llana Heights Milverton (Apt #'s "
                                                 '11-12-14-15)',
                                         'poller_hash': 'MUZENkRC',
                                         'serial_number': '1FD6DB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:34:64:6f',
                                         'uptime': '9954635'},
                                 {       'counter_source': '458/g1+458/g2+458/g3+458/g5+458/g6+458/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:04:d2'],
                                                                       'id': '458/g1',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '1530211746',
                                                                       'in_ucast_pkts': '5192973',
                                                                       'mac_address': '00:06:31:9c:bb:b3',
                                                                       'mtu': 2000,
                                                                       'name': '458/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102918',
                                                                       'out_octets': '46535157510',
                                                                       'out_ucast_pkts': '32958389',
                                                                       'poller_hash': 'MTM4MEQ0cGVlcmludGY0NTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:04:9c'],
                                                                       'id': '458/g2',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '41114088',
                                                                       'in_ucast_pkts': '108366',
                                                                       'mac_address': '00:06:31:9c:bb:b4',
                                                                       'mtu': 2000,
                                                                       'name': '458/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102919',
                                                                       'out_octets': '150483755',
                                                                       'out_ucast_pkts': '190527',
                                                                       'poller_hash': 'MTM4MEQ0cGVlcmludGY0NTgvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '30:68:93:49:c2:5d'],
                                                                       'id': '458/g3',
                                                                       'in_nucast_pkts': '402',
                                                                       'in_octets': '2546745882',
                                                                       'in_ucast_pkts': '31270664',
                                                                       'mac_address': '00:06:31:9c:bb:b5',
                                                                       'mtu': 2000,
                                                                       'name': '458/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102492',
                                                                       'out_octets': '199336464848',
                                                                       'out_ucast_pkts': '132266828',
                                                                       'poller_hash': 'MTM4MEQ0cGVlcmludGY0NTgvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:f0:3f'],
                                                                       'id': '458/g5',
                                                                       'in_nucast_pkts': '6',
                                                                       'in_octets': '6543317727',
                                                                       'in_ucast_pkts': '43875248',
                                                                       'mac_address': '00:06:31:a1:10:95',
                                                                       'mtu': 2000,
                                                                       'name': '458/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1157390',
                                                                       'out_octets': '268035953017',
                                                                       'out_ucast_pkts': '183470006',
                                                                       'poller_hash': 'MTM4MEQ0cGVlcmludGY0NTgvZzU=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a4:2a:95:3e:22:80'],
                                                                       'id': '458/g6',
                                                                       'in_nucast_pkts': '5111',
                                                                       'in_octets': '2506492568',
                                                                       'in_ucast_pkts': '11278717',
                                                                       'mac_address': '00:06:31:a1:10:96',
                                                                       'mtu': 2000,
                                                                       'name': '458/g6',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102937',
                                                                       'out_octets': '83810685411',
                                                                       'out_ucast_pkts': '64575278',
                                                                       'poller_hash': 'MTM4MEQ0cGVlcmludGY0NTgvZzY=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b6:e3'],
                                                                       'id': '458/g8',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '3238664294',
                                                                       'in_ucast_pkts': '29101597',
                                                                       'mac_address': '00:06:31:a1:10:98',
                                                                       'mtu': 2000,
                                                                       'name': '458/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102913',
                                                                       'out_octets': '157668739021',
                                                                       'out_ucast_pkts': '117244441',
                                                                       'poller_hash': 'MTM4MEQ0cGVlcmludGY0NTgvZzg=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '458',
                                                                       'in_nucast_pkts': '6671569',
                                                                       'in_octets': '755537483562',
                                                                       'in_ucast_pkts': '530705469',
                                                                       'links': [       {       'in_nucast_pkts': '6671569',
                                                                                                'in_octets': '755537483562',
                                                                                                'in_ucast_pkts': '530705469',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5547',
                                                                                                'out_octets': '16406546305',
                                                                                                'out_ucast_pkts': '120827565',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5547',
                                                                       'out_octets': '16406546305',
                                                                       'out_ucast_pkts': '120827565',
                                                                       'poller_hash': 'MTM4MEQ0cGVlcmludGZwb24gNDU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '458',
                                                                                               'links': [       {       'connected_time': '688356',
                                                                                                                        'fiber_length': 1435,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTM4MEQ0cGVlcnBvbnBvbiA0NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': "2 Llana Heights Milverton (Apt #'s "
                                                 '21 thr 34)',
                                         'poller_hash': 'MTM4MEQ0',
                                         'serial_number': '1380D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:9c:bb:b8',
                                         'uptime': '688356'},
                                 {       'counter_source': '459/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '459/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '504224062963',
                                                                       'in_ucast_pkts': '4216154172',
                                                                       'mac_address': '00:06:31:87:4f:db',
                                                                       'name': '459/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11265954596495',
                                                                       'out_ucast_pkts': '5326064691',
                                                                       'poller_hash': 'Q0MwQ0NwZWVyaW50ZjQ1OS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '459',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11265954596495',
                                                                       'in_ucast_pkts': '5326064691',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11265954596495',
                                                                                                'in_ucast_pkts': '5326064691',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '504224062963',
                                                                                                'out_ucast_pkts': '4216154172',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '504224062963',
                                                                       'out_ucast_pkts': '4216154172',
                                                                       'poller_hash': 'Q0MwQ0NwZWVyaW50ZnBvbiA0NTk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '459',
                                                                                               'links': [       {       'connected_time': '8447159',
                                                                                                                        'fiber_length': 1051,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQ0NwZWVycG9ucG9uIDQ1OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '5 LLANA HEIGHTS DR (PRESCOTT BRENDA)',
                                         'poller_hash': 'Q0MwQ0M=',
                                         'serial_number': 'CC0CC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:8c',
                                         'uptime': '8447159'},
                                 {       'counter_source': '460/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '460/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '150462725727',
                                                                       'in_ucast_pkts': '924222813',
                                                                       'mac_address': '00:06:31:87:41:43',
                                                                       'name': '460/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3143938246612',
                                                                       'out_ucast_pkts': '2287905258',
                                                                       'poller_hash': 'Q0MwN0ZwZWVyaW50ZjQ2MC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '460',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3143938246612',
                                                                       'in_ucast_pkts': '2287905258',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3143938246612',
                                                                                                'in_ucast_pkts': '2287905258',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '150462725727',
                                                                                                'out_ucast_pkts': '924222813',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '150462725727',
                                                                       'out_ucast_pkts': '924222813',
                                                                       'poller_hash': 'Q0MwN0ZwZWVyaW50ZnBvbiA0NjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '460',
                                                                                               'links': [       {       'connected_time': '688384',
                                                                                                                        'fiber_length': 1005,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwN0ZwZWVycG9ucG9uIDQ2MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '3 Llana Heights Milverton (Deewar '
                                                 'Randy)',
                                         'poller_hash': 'Q0MwN0Y=',
                                         'serial_number': 'CC07F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:3f',
                                         'uptime': '688384'},
                                 {       'counter_source': '461/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '461/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '23178969172',
                                                                       'in_ucast_pkts': '101667263',
                                                                       'mac_address': 'ec:4f:82:5b:71:4b',
                                                                       'name': '461/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '344350243556',
                                                                       'out_ucast_pkts': '317022249',
                                                                       'poller_hash': 'MkIwMTlDcGVlcmludGY0NjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '461',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '344350243556',
                                                                       'in_ucast_pkts': '317022249',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '344350243556',
                                                                                                'in_ucast_pkts': '317022249',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '23178969172',
                                                                                                'out_ucast_pkts': '101667263',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '23178969172',
                                                                       'out_ucast_pkts': '101667263',
                                                                       'poller_hash': 'MkIwMTlDcGVlcmludGZwb24gNDYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '461',
                                                                                               'links': [       {       'connected_time': '15498618',
                                                                                                                        'fiber_length': 1310,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTlDcGVlcnBvbnBvbiA0NjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '81 Main St. N. Milverton (Carrie '
                                                 'Wagler)',
                                         'poller_hash': 'MkIwMTlD',
                                         'serial_number': '2B019C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:71:4b',
                                         'uptime': '15498618'},
                                 {       'counter_source': '462/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6f:13:67:f3:3b'],
                                                                       'id': '462/g1',
                                                                       'in_nucast_pkts': '10605',
                                                                       'in_octets': '195658584179',
                                                                       'in_ucast_pkts': '442992298',
                                                                       'mac_address': 'ec:4f:82:7e:0c:b5',
                                                                       'mtu': 2000,
                                                                       'name': '462/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27442082',
                                                                       'out_octets': '751085708913',
                                                                       'out_ucast_pkts': '702591268',
                                                                       'poller_hash': 'MzVFQTg3cGVlcmludGY0NjIvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '462',
                                                                       'in_nucast_pkts': '27442082',
                                                                       'in_octets': '751085708913',
                                                                       'in_ucast_pkts': '702591268',
                                                                       'links': [       {       'in_nucast_pkts': '27442082',
                                                                                                'in_octets': '751085708913',
                                                                                                'in_ucast_pkts': '702591268',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10605',
                                                                                                'out_octets': '195658584179',
                                                                                                'out_ucast_pkts': '442992298',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10605',
                                                                       'out_octets': '195658584179',
                                                                       'out_ucast_pkts': '442992298',
                                                                       'poller_hash': 'MzVFQTg3cGVlcmludGZwb24gNDYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '462',
                                                                                               'links': [       {       'connected_time': '12047422',
                                                                                                                        'fiber_length': 1283,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQTg3cGVlcnBvbnBvbiA0NjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '39 Main St. N. Milverton (Guenthers '
                                                 'Bakery)',
                                         'poller_hash': 'MzVFQTg3',
                                         'serial_number': '35EA87',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0c:b5',
                                         'uptime': '12047422'},
                                 {       'counter_source': '463/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '463/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '129256537503',
                                                                       'in_ucast_pkts': '1188033406',
                                                                       'mac_address': 'cc:be:59:cb:7a:ba',
                                                                       'name': '463/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '6808732116722',
                                                                       'out_ucast_pkts': '4572221983',
                                                                       'poller_hash': 'NDJDMEFGcGVlcmludGY0NjMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '463',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '6808732116722',
                                                                       'in_ucast_pkts': '4572221983',
                                                                       'links': [       {       'in_nucast_pkts': '66',
                                                                                                'in_octets': '6808732116722',
                                                                                                'in_ucast_pkts': '4572221983',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '129256537503',
                                                                                                'out_ucast_pkts': '1188033406',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '129256537503',
                                                                       'out_ucast_pkts': '1188033406',
                                                                       'poller_hash': 'NDJDMEFGcGVlcmludGZwb24gNDYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '463',
                                                                                               'links': [       {       'connected_time': '11594465',
                                                                                                                        'fiber_length': 1280,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMEFGcGVlcnBvbnBvbiA0NjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '37 Main St. N. Milverton (Doris '
                                                 'Coultes)',
                                         'poller_hash': 'NDJDMEFG',
                                         'serial_number': '42C0AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:7a:ba',
                                         'uptime': '11594465'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '464/g1',
                                                                       'name': '464/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGOUE1cGVlcmludGY0NjQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '464',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGOUE1cGVlcmludGZwb24gNDY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '464',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUE1cGVlcnBvbnBvbiA0NjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '41 Main St. N. Milverton (Ginnie '
                                                 'Perry)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "464" '
                                                                  '(location: GPON '
                                                                  'port "7/7"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2024-08-08T18:08:42.660Z',
                                         'poller_hash': 'MzVGOUE1',
                                         'serial_number': '35F9A5',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '465/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a5:06:0d'],
                                                                       'id': '465/g1',
                                                                       'in_nucast_pkts': '10566',
                                                                       'in_octets': '85934469068',
                                                                       'in_ucast_pkts': '414153171',
                                                                       'mac_address': 'ec:4f:82:5b:84:96',
                                                                       'mtu': 2000,
                                                                       'name': '465/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '143710625',
                                                                       'out_octets': '1537112382041',
                                                                       'out_ucast_pkts': '1144513557',
                                                                       'poller_hash': 'MkIwNDA1cGVlcmludGY0NjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '465',
                                                                       'in_nucast_pkts': '143710625',
                                                                       'in_octets': '1537112382041',
                                                                       'in_ucast_pkts': '1144513557',
                                                                       'links': [       {       'in_nucast_pkts': '143710625',
                                                                                                'in_octets': '1537112382041',
                                                                                                'in_ucast_pkts': '1144513557',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10566',
                                                                                                'out_octets': '85934469068',
                                                                                                'out_ucast_pkts': '414153171',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10566',
                                                                       'out_octets': '85934469068',
                                                                       'out_ucast_pkts': '414153171',
                                                                       'poller_hash': 'MkIwNDA1cGVlcmludGZwb24gNDY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '465',
                                                                                               'links': [       {       'connected_time': '21815874',
                                                                                                                        'fiber_length': 1254,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDA1cGVlcnBvbnBvbiA0NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 Llana Heights Milverton (Dirk '
                                                 'Willemsen 10401603)',
                                         'poller_hash': 'MkIwNDA1',
                                         'serial_number': '2B0405',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:84:93',
                                         'uptime': '21815874'},
                                 {       'counter_source': '466/g2+466/g5+466/g6+466/g7+466/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:70:b2'],
                                                                       'id': '466/g2',
                                                                       'in_nucast_pkts': '1349306',
                                                                       'in_octets': '328360024',
                                                                       'in_ucast_pkts': '1142324',
                                                                       'mac_address': 'ec:4f:82:28:e5:3e',
                                                                       'mtu': 2000,
                                                                       'name': '466/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1449284395',
                                                                       'out_octets': '1927659818222',
                                                                       'out_ucast_pkts': '876255',
                                                                       'poller_hash': 'MjZEOTY2cGVlcmludGY0NjYvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:58:de'],
                                                                       'id': '466/g5',
                                                                       'in_nucast_pkts': '1242299',
                                                                       'in_octets': '10201528004',
                                                                       'in_ucast_pkts': '131539871',
                                                                       'mac_address': 'ec:4f:82:b0:3e:fe',
                                                                       'mtu': 2000,
                                                                       'name': '466/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '509428549',
                                                                       'out_octets': '1232518466822',
                                                                       'out_ucast_pkts': '424174413',
                                                                       'poller_hash': 'MjZEOTY2cGVlcmludGY0NjYvZzU=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:43:97:58'],
                                                                       'id': '466/g6',
                                                                       'in_nucast_pkts': '1152010',
                                                                       'in_octets': '7866560021',
                                                                       'in_ucast_pkts': '101893839',
                                                                       'mac_address': 'ec:4f:82:b0:3e:ff',
                                                                       'mtu': 2000,
                                                                       'name': '466/g6',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1419823643',
                                                                       'out_octets': '2332750321630',
                                                                       'out_ucast_pkts': '325997874',
                                                                       'poller_hash': 'MjZEOTY2cGVlcmludGY0NjYvZzY=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:7d:9a'],
                                                                       'id': '466/g7',
                                                                       'in_nucast_pkts': '1330150',
                                                                       'in_octets': '1340254784',
                                                                       'in_ucast_pkts': '4362749',
                                                                       'mac_address': 'ec:4f:82:b0:3f:00',
                                                                       'mtu': 2000,
                                                                       'name': '466/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55133439',
                                                                       'out_octets': '32033689385',
                                                                       'out_ucast_pkts': '8476295',
                                                                       'poller_hash': 'MjZEOTY2cGVlcmludGY0NjYvZzc=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:6f:ae'],
                                                                       'id': '466/g8',
                                                                       'in_nucast_pkts': '1195869',
                                                                       'in_octets': '307414466',
                                                                       'in_ucast_pkts': '1000031',
                                                                       'mac_address': 'ec:4f:82:b0:3f:01',
                                                                       'mtu': 2000,
                                                                       'name': '466/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '347227867',
                                                                       'out_octets': '438302615718',
                                                                       'out_ucast_pkts': '776916',
                                                                       'poller_hash': 'MjZEOTY2cGVlcmludGY0NjYvZzg=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '466',
                                                                       'in_nucast_pkts': '3780897893',
                                                                       'in_octets': '5963264911777',
                                                                       'in_ucast_pkts': '760301753',
                                                                       'links': [       {       'in_nucast_pkts': '3780897893',
                                                                                                'in_octets': '5963264911777',
                                                                                                'in_ucast_pkts': '760301753',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6269634',
                                                                                                'out_octets': '20044117299',
                                                                                                'out_ucast_pkts': '239938814',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6269634',
                                                                       'out_octets': '20044117299',
                                                                       'out_ucast_pkts': '239938814',
                                                                       'poller_hash': 'MjZEOTY2cGVlcmludGZwb24gNDY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '466',
                                                                                               'links': [       {       'connected_time': '4164963',
                                                                                                                        'fiber_length': 1937,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjZEOTY2cGVlcnBvbnBvbiA0NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 William St. Milverton (Data Rm 2 '
                                                 'Knollcrest)',
                                         'poller_hash': 'MjZEOTY2',
                                         'serial_number': '26D966',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:28:e5:42',
                                         'uptime': '4164963'},
                                 {       'counter_source': '467/g1+467/g2+467/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:58:38'],
                                                                       'id': '467/g1',
                                                                       'in_nucast_pkts': '1318831',
                                                                       'in_octets': '1494347237',
                                                                       'in_ucast_pkts': '9861061',
                                                                       'mac_address': 'ec:4f:82:be:ab:cd',
                                                                       'mtu': 2000,
                                                                       'name': '467/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '513899797',
                                                                       'out_octets': '684876981875',
                                                                       'out_ucast_pkts': '33311460',
                                                                       'poller_hash': 'Mzc4OEI2cGVlcmludGY0NjcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a1:8c:bb'],
                                                                       'id': '467/g2',
                                                                       'in_nucast_pkts': '11272',
                                                                       'in_octets': '49787706006',
                                                                       'in_ucast_pkts': '584460633',
                                                                       'mac_address': 'ec:4f:82:be:ab:ce',
                                                                       'mtu': 2000,
                                                                       'name': '467/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7247599',
                                                                       'out_octets': '1671953786850',
                                                                       'out_ucast_pkts': '1133810351',
                                                                       'poller_hash': 'Mzc4OEI2cGVlcmludGY0NjcvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:60:16'],
                                                                       'id': '467/g3',
                                                                       'in_nucast_pkts': '1323414',
                                                                       'in_octets': '2918116518',
                                                                       'in_ucast_pkts': '17845946',
                                                                       'mac_address': 'ec:4f:82:be:ab:cf',
                                                                       'mtu': 2000,
                                                                       'name': '467/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55123555',
                                                                       'out_octets': '163563864334',
                                                                       'out_ucast_pkts': '114476877',
                                                                       'poller_hash': 'Mzc4OEI2cGVlcmludGY0NjcvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '467',
                                                                       'in_nucast_pkts': '576270951',
                                                                       'in_octets': '2520394633059',
                                                                       'in_ucast_pkts': '1281598688',
                                                                       'links': [       {       'in_nucast_pkts': '576270951',
                                                                                                'in_octets': '2520394633059',
                                                                                                'in_ucast_pkts': '1281598688',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2653517',
                                                                                                'out_octets': '54200169761',
                                                                                                'out_ucast_pkts': '612167640',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2653517',
                                                                       'out_octets': '54200169761',
                                                                       'out_ucast_pkts': '612167640',
                                                                       'poller_hash': 'Mzc4OEI2cGVlcmludGZwb24gNDY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '467',
                                                                                               'links': [       {       'connected_time': '4165169',
                                                                                                                        'fiber_length': 1561,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.198,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzc4OEI2cGVlcnBvbnBvbiA0Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '50 WILLIAM ST',
                                         'poller_hash': 'Mzc4OEI2',
                                         'serial_number': '3788B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:be:ab:c9',
                                         'uptime': '4165169'},
                                 {       'counter_source': '468/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b7:df'],
                                                                       'id': '468/g1',
                                                                       'in_nucast_pkts': '1188',
                                                                       'in_octets': '101160340450',
                                                                       'in_ucast_pkts': '1303931461',
                                                                       'mac_address': 'ec:4f:82:5b:72:9e',
                                                                       'mtu': 2000,
                                                                       'name': '468/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '116864876',
                                                                       'out_octets': '4919599312284',
                                                                       'out_ucast_pkts': '3271028279',
                                                                       'poller_hash': 'MkIwMUM2cGVlcmludGY0NjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '468',
                                                                       'in_nucast_pkts': '116864876',
                                                                       'in_octets': '4919599312284',
                                                                       'in_ucast_pkts': '3271028279',
                                                                       'links': [       {       'in_nucast_pkts': '116864876',
                                                                                                'in_octets': '4919599312284',
                                                                                                'in_ucast_pkts': '3271028279',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1188',
                                                                                                'out_octets': '101160340450',
                                                                                                'out_ucast_pkts': '1303931461',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1188',
                                                                       'out_octets': '101160340450',
                                                                       'out_ucast_pkts': '1303931461',
                                                                       'poller_hash': 'MkIwMUM2cGVlcmludGZwb24gNDY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '468',
                                                                                               'links': [       {       'connected_time': '6743042',
                                                                                                                        'fiber_length': 1710,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMUM2cGVlcnBvbnBvbiA0Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 William St.Apt # 101 (Catharine '
                                                 'Birch)',
                                         'poller_hash': 'MkIwMUM2',
                                         'serial_number': '2B01C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:72:9b',
                                         'uptime': '6743042'},
                                 {       'counter_source': '469/g1+469/g2+469/g4+469/g5+469/g7+469/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5b:1c'],
                                                                       'id': '469/g1',
                                                                       'in_nucast_pkts': '1348992',
                                                                       'in_octets': '5692562691',
                                                                       'in_ucast_pkts': '69356826',
                                                                       'mac_address': 'ec:4f:82:28:82:55',
                                                                       'mtu': 2000,
                                                                       'name': '469/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1373648123',
                                                                       'out_octets': '2125675264019',
                                                                       'out_ucast_pkts': '222384836',
                                                                       'poller_hash': 'MjZBQzZGcGVlcmludGY0NjkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:7d:82'],
                                                                       'id': '469/g2',
                                                                       'in_nucast_pkts': '1257961',
                                                                       'in_octets': '718876321',
                                                                       'in_ucast_pkts': '2105446',
                                                                       'mac_address': 'ec:4f:82:28:82:56',
                                                                       'mtu': 2000,
                                                                       'name': '469/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1384114513',
                                                                       'out_octets': '1830590139819',
                                                                       'out_ucast_pkts': '4305343',
                                                                       'poller_hash': 'MjZBQzZGcGVlcmludGY0NjkvZzI=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:31:50'],
                                                                       'id': '469/g4',
                                                                       'in_nucast_pkts': '1374560',
                                                                       'in_octets': '344938840',
                                                                       'in_ucast_pkts': '1242759',
                                                                       'mac_address': 'ec:4f:82:28:82:58',
                                                                       'mtu': 2000,
                                                                       'name': '469/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1963166627',
                                                                       'out_octets': '2627728289048',
                                                                       'out_ucast_pkts': '954606',
                                                                       'poller_hash': 'MjZBQzZGcGVlcmludGY0NjkvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5b:0c'],
                                                                       'id': '469/g5',
                                                                       'in_nucast_pkts': '1235153',
                                                                       'in_octets': '5407664643',
                                                                       'in_ucast_pkts': '42430594',
                                                                       'mac_address': 'ec:4f:82:64:27:80',
                                                                       'mtu': 2000,
                                                                       'name': '469/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '50380811',
                                                                       'out_octets': '335124792683',
                                                                       'out_ucast_pkts': '240586036',
                                                                       'poller_hash': 'MjZBQzZGcGVlcmludGY0NjkvZzU=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:80:1c'],
                                                                       'id': '469/g7',
                                                                       'in_nucast_pkts': '1332241',
                                                                       'in_octets': '22252690186',
                                                                       'in_ucast_pkts': '296294598',
                                                                       'mac_address': 'ec:4f:82:64:27:82',
                                                                       'mtu': 2000,
                                                                       'name': '469/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55118217',
                                                                       'out_octets': '1388549427823',
                                                                       'out_ucast_pkts': '973391732',
                                                                       'poller_hash': 'MjZBQzZGcGVlcmludGY0NjkvZzc=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:17:75:1a'],
                                                                       'id': '469/g8',
                                                                       'in_nucast_pkts': '1400127',
                                                                       'in_octets': '321347144',
                                                                       'in_ucast_pkts': '935758',
                                                                       'mac_address': 'ec:4f:82:64:27:83',
                                                                       'mtu': 2000,
                                                                       'name': '469/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1210477225',
                                                                       'out_octets': '1602760249006',
                                                                       'out_ucast_pkts': '696727',
                                                                       'poller_hash': 'MjZBQzZGcGVlcmludGY0NjkvZzg=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '469',
                                                                       'in_nucast_pkts': '6036905516',
                                                                       'in_octets': '9910428162398',
                                                                       'in_ucast_pkts': '1442319280',
                                                                       'links': [       {       'in_nucast_pkts': '6036905516',
                                                                                                'in_octets': '9910428162398',
                                                                                                'in_ucast_pkts': '1442319280',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7949034',
                                                                                                'out_octets': '34738079825',
                                                                                                'out_ucast_pkts': '412365981',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7949034',
                                                                       'out_octets': '34738079825',
                                                                       'out_ucast_pkts': '412365981',
                                                                       'poller_hash': 'MjZBQzZGcGVlcmludGZwb24gNDY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '469',
                                                                                               'links': [       {       'connected_time': '4164897',
                                                                                                                        'fiber_length': 1893,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjZBQzZGcGVlcnBvbnBvbiA0Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST',
                                         'poller_hash': 'MjZBQzZG',
                                         'serial_number': '26AC6F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:28:82:5a',
                                         'uptime': '4164897'},
                                 {       'counter_source': '470/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '470/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '61211284511',
                                                                       'in_ucast_pkts': '239435671',
                                                                       'mac_address': 'ec:4f:82:8e:0d:58',
                                                                       'name': '470/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '562769085520',
                                                                       'out_ucast_pkts': '567000684',
                                                                       'poller_hash': 'MzIyOUYycGVlcmludGY0NzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '470',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '562769085520',
                                                                       'in_ucast_pkts': '567000684',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '562769085520',
                                                                                                'in_ucast_pkts': '567000684',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '61211284511',
                                                                                                'out_ucast_pkts': '239435671',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '61211284511',
                                                                       'out_ucast_pkts': '239435671',
                                                                       'poller_hash': 'MzIyOUYycGVlcmludGZwb24gNDcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '470',
                                                                                               'links': [       {       'connected_time': '2060536',
                                                                                                                        'fiber_length': 5975,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOUYycGVlcnBvbnBvbiA0NzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6680A Road 140 Milverton (Bruce Elg '
                                                 '10401527)',
                                         'poller_hash': 'MzIyOUYy',
                                         'serial_number': '3229F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:0d:58',
                                         'uptime': '2060536'},
                                 {       'counter_source': '471/g2+471/g3+471/g4+471/g5+471/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a4:da:68'],
                                                                       'id': '471/g2',
                                                                       'in_nucast_pkts': '981218',
                                                                       'in_octets': '304473201',
                                                                       'in_ucast_pkts': '1216130',
                                                                       'mac_address': '00:06:31:ca:5e:24',
                                                                       'mtu': 2000,
                                                                       'name': '471/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '602536903',
                                                                       'out_octets': '796518667325',
                                                                       'out_ucast_pkts': '974295',
                                                                       'poller_hash': 'MjYzMjIycGVlcmludGY0NzEvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:7d:a2'],
                                                                       'id': '471/g3',
                                                                       'in_nucast_pkts': '712245',
                                                                       'in_octets': '876104897',
                                                                       'in_ucast_pkts': '4131991',
                                                                       'mac_address': '00:06:31:ca:5e:25',
                                                                       'mtu': 2000,
                                                                       'name': '471/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39546302',
                                                                       'out_octets': '31123566162',
                                                                       'out_ucast_pkts': '11984824',
                                                                       'poller_hash': 'MjYzMjIycGVlcmludGY0NzEvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a4:db:58'],
                                                                       'id': '471/g4',
                                                                       'in_nucast_pkts': '1439076',
                                                                       'in_octets': '372798003',
                                                                       'in_ucast_pkts': '1346464',
                                                                       'mac_address': '00:06:31:ca:5e:26',
                                                                       'mtu': 2000,
                                                                       'name': '471/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3141107353',
                                                                       'out_octets': '4231846162689',
                                                                       'out_ucast_pkts': '1070384',
                                                                       'poller_hash': 'MjYzMjIycGVlcmludGY0NzEvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:73:9e'],
                                                                       'id': '471/g5',
                                                                       'in_nucast_pkts': '696748',
                                                                       'in_octets': '60504543',
                                                                       'in_ucast_pkts': '194424',
                                                                       'mac_address': 'ec:4f:82:27:85:91',
                                                                       'mtu': 2000,
                                                                       'name': '471/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41910545',
                                                                       'out_octets': '19094211843',
                                                                       'out_ucast_pkts': '78192',
                                                                       'poller_hash': 'MjYzMjIycGVlcmludGY0NzEvZzU=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5a:b2'],
                                                                       'id': '471/g8',
                                                                       'in_nucast_pkts': '1256691',
                                                                       'in_octets': '980200838',
                                                                       'in_ucast_pkts': '5175602',
                                                                       'mac_address': 'ec:4f:82:27:85:94',
                                                                       'mtu': 2000,
                                                                       'name': '471/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1072956571',
                                                                       'out_octets': '1415184210963',
                                                                       'out_ucast_pkts': '9372999',
                                                                       'poller_hash': 'MjYzMjIycGVlcmludGY0NzEvZzg=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '471',
                                                                       'in_nucast_pkts': '4898057674',
                                                                       'in_octets': '6493766818982',
                                                                       'in_ucast_pkts': '23480694',
                                                                       'links': [       {       'in_nucast_pkts': '4898057674',
                                                                                                'in_octets': '6493766818982',
                                                                                                'in_ucast_pkts': '23480694',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5085978',
                                                                                                'out_octets': '2594081482',
                                                                                                'out_ucast_pkts': '12064611',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5085978',
                                                                       'out_octets': '2594081482',
                                                                       'out_ucast_pkts': '12064611',
                                                                       'poller_hash': 'MjYzMjIycGVlcmludGZwb24gNDcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '471',
                                                                                               'links': [       {       'connected_time': '4164929',
                                                                                                                        'fiber_length': 1893,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYzMjIycGVlcnBvbnBvbiA0NzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST (ALBRECHT MISS '
                                                 'CATHERINE)',
                                         'poller_hash': 'MjYzMjIy',
                                         'serial_number': '263222',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ca:5e:28',
                                         'uptime': '4164929'},
                                 {       'counter_source': '472/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '472/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '653638047376',
                                                                       'in_ucast_pkts': '2384981450',
                                                                       'mac_address': 'cc:be:59:61:18:52',
                                                                       'name': '472/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9157273985809',
                                                                       'out_ucast_pkts': '7053190465',
                                                                       'poller_hash': 'NDU5NjM4cGVlcmludGY0NzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '472',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9157273985809',
                                                                       'in_ucast_pkts': '7053190465',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9157273985809',
                                                                                                'in_ucast_pkts': '7053190465',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '653638047376',
                                                                                                'out_ucast_pkts': '2384981450',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '653638047376',
                                                                       'out_ucast_pkts': '2384981450',
                                                                       'poller_hash': 'NDU5NjM4cGVlcmludGZwb24gNDcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '472',
                                                                                               'links': [       {       'connected_time': '71351983',
                                                                                                                        'fiber_length': 1219,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.414,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDU5NjM4cGVlcnBvbnBvbiA0NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '27 Main St. N. Milverton (Koertland '
                                                 'Garage)',
                                         'poller_hash': 'NDU5NjM4',
                                         'serial_number': '459638',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:61:18:52',
                                         'uptime': '71351983'},
                                 {       'counter_source': '473/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '473/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '76837579929',
                                                                       'in_ucast_pkts': '557865808',
                                                                       'mac_address': 'ec:4f:82:0d:69:a3',
                                                                       'name': '473/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '50',
                                                                       'out_octets': '2004026278759',
                                                                       'out_ucast_pkts': '1562216345',
                                                                       'poller_hash': 'MjA1RERFcGVlcmludGY0NzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '473',
                                                                       'in_nucast_pkts': '50',
                                                                       'in_octets': '2004026278759',
                                                                       'in_ucast_pkts': '1562216345',
                                                                       'links': [       {       'in_nucast_pkts': '50',
                                                                                                'in_octets': '2004026278759',
                                                                                                'in_ucast_pkts': '1562216345',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '76837579929',
                                                                                                'out_ucast_pkts': '557865808',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '76837579929',
                                                                       'out_ucast_pkts': '557865808',
                                                                       'poller_hash': 'MjA1RERFcGVlcmludGZwb24gNDcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '473',
                                                                                               'links': [       {       'connected_time': '688346',
                                                                                                                        'fiber_length': 843,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RERFcGVlcnBvbnBvbiA0NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '12 PUGH ST W (PRISECARU TIMOTEI)',
                                         'poller_hash': 'MjA1RERF',
                                         'serial_number': '205DDE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:a3',
                                         'uptime': '688346'},
                                 {       'counter_source': '474/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:80:66'],
                                                                       'id': '474/g1',
                                                                       'in_nucast_pkts': '5366134',
                                                                       'in_octets': '174825717150',
                                                                       'in_ucast_pkts': '1966286800',
                                                                       'mac_address': 'ec:4f:82:0d:68:e0',
                                                                       'mtu': 2000,
                                                                       'name': '474/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '182095064',
                                                                       'out_octets': '6945684513527',
                                                                       'out_ucast_pkts': '4906281245',
                                                                       'poller_hash': 'MjA1REM4cGVlcmludGY0NzQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '474',
                                                                       'in_nucast_pkts': '182095064',
                                                                       'in_octets': '6945684513527',
                                                                       'in_ucast_pkts': '4906281245',
                                                                       'links': [       {       'in_nucast_pkts': '182095064',
                                                                                                'in_octets': '6945684513527',
                                                                                                'in_ucast_pkts': '4906281245',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5366134',
                                                                                                'out_octets': '174825717150',
                                                                                                'out_ucast_pkts': '1966286800',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5366134',
                                                                       'out_octets': '174825717150',
                                                                       'out_ucast_pkts': '1966286800',
                                                                       'poller_hash': 'MjA1REM4cGVlcmludGZwb24gNDc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '474',
                                                                                               'links': [       {       'connected_time': '2150542',
                                                                                                                        'fiber_length': 1681,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -7.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REM4cGVlcnBvbnBvbiA0NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '60 WILLIAM ST STE 104 (COSSEY JEAN)',
                                         'poller_hash': 'MjA1REM4',
                                         'serial_number': '205DC8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:dd',
                                         'uptime': '2150542'},
                                 {       'counter_source': '475/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:fa:30:a0:18:10'],
                                                                       'id': '475/g1',
                                                                       'in_nucast_pkts': '18719946',
                                                                       'in_octets': '4842233472515',
                                                                       'in_ucast_pkts': '24001341635',
                                                                       'mac_address': 'ec:4f:82:7e:09:40',
                                                                       'mtu': 2000,
                                                                       'name': '475/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '89559502',
                                                                       'out_octets': '76648560748600',
                                                                       'out_ucast_pkts': '57335384233',
                                                                       'poller_hash': 'MzVFOTYwcGVlcmludGY0NzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '475',
                                                                       'in_nucast_pkts': '89559502',
                                                                       'in_octets': '76648560748600',
                                                                       'in_ucast_pkts': '57335384233',
                                                                       'links': [       {       'in_nucast_pkts': '89559502',
                                                                                                'in_octets': '76648560748600',
                                                                                                'in_ucast_pkts': '57335384233',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18719946',
                                                                                                'out_octets': '4842233472515',
                                                                                                'out_ucast_pkts': '24001341635',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18719946',
                                                                       'out_octets': '4842233472515',
                                                                       'out_ucast_pkts': '24001341635',
                                                                       'poller_hash': 'MzVFOTYwcGVlcmludGZwb24gNDc1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '475',
                                                                                               'links': [       {       'connected_time': '67790412',
                                                                                                                        'fiber_length': 1593,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTYwcGVlcnBvbnBvbiA0NzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '50 William St. Milverton (Knollcrest '
                                                 'Main Internet&Elevator phone)',
                                         'poller_hash': 'MzVFOTYw',
                                         'serial_number': '35E960',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:40',
                                         'uptime': '67790412'},
                                 {       'counter_source': '476/g1+476/g3+476/g5+476/g6+476/g7',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a4:db:66'],
                                                                       'id': '476/g1',
                                                                       'in_nucast_pkts': '1357606',
                                                                       'in_octets': '928988450',
                                                                       'in_ucast_pkts': '6631608',
                                                                       'mac_address': '00:06:31:f6:d7:78',
                                                                       'mtu': 2000,
                                                                       'name': '476/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1145650157',
                                                                       'out_octets': '1529085916656',
                                                                       'out_ucast_pkts': '19577086',
                                                                       'poller_hash': 'MjYyNUU3cGVlcmludGY0NzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:32:0c'],
                                                                       'id': '476/g3',
                                                                       'in_nucast_pkts': '1232450',
                                                                       'in_octets': '133676836',
                                                                       'in_ucast_pkts': '356821',
                                                                       'mac_address': '00:06:31:f6:d7:7a',
                                                                       'mtu': 2000,
                                                                       'name': '476/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2197724610',
                                                                       'out_octets': '2951583144134',
                                                                       'out_ucast_pkts': '209399',
                                                                       'poller_hash': 'MjYyNUU3cGVlcmludGY0NzYvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:73:18'],
                                                                       'id': '476/g5',
                                                                       'in_nucast_pkts': '1383338',
                                                                       'in_octets': '342777229',
                                                                       'in_ucast_pkts': '1091930',
                                                                       'mac_address': 'ec:4f:82:25:6b:f3',
                                                                       'mtu': 2000,
                                                                       'name': '476/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2078279247',
                                                                       'out_octets': '2783025740622',
                                                                       'out_ucast_pkts': '805650',
                                                                       'poller_hash': 'MjYyNUU3cGVlcmludGY0NzYvZzU=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:43:96:d2'],
                                                                       'id': '476/g6',
                                                                       'in_nucast_pkts': '1315820',
                                                                       'in_octets': '20213375703',
                                                                       'in_ucast_pkts': '276498630',
                                                                       'mac_address': 'ec:4f:82:25:6b:f4',
                                                                       'mtu': 2000,
                                                                       'name': '476/g6',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1565167884',
                                                                       'out_octets': '3296221107611',
                                                                       'out_ucast_pkts': '868052810',
                                                                       'poller_hash': 'MjYyNUU3cGVlcmludGY0NzYvZzY=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:58:16'],
                                                                       'id': '476/g7',
                                                                       'in_nucast_pkts': '1229327',
                                                                       'in_octets': '3308778704',
                                                                       'in_ucast_pkts': '36754918',
                                                                       'mac_address': 'ec:4f:82:25:6b:f5',
                                                                       'mtu': 2000,
                                                                       'name': '476/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '758796379',
                                                                       'out_octets': '1133288385989',
                                                                       'out_ucast_pkts': '113718506',
                                                                       'poller_hash': 'MjYyNUU3cGVlcmludGY0NzYvZzc=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '476',
                                                                       'in_nucast_pkts': '7745618277',
                                                                       'in_octets': '11693204295012',
                                                                       'in_ucast_pkts': '1002363451',
                                                                       'links': [       {       'in_nucast_pkts': '7745618277',
                                                                                                'in_octets': '11693204295012',
                                                                                                'in_ucast_pkts': '1002363451',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6518541',
                                                                                                'out_octets': '24927596922',
                                                                                                'out_ucast_pkts': '321333907',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6518541',
                                                                       'out_octets': '24927596922',
                                                                       'out_ucast_pkts': '321333907',
                                                                       'poller_hash': 'MjYyNUU3cGVlcmludGZwb24gNDc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '476',
                                                                                               'links': [       {       'connected_time': '4155296',
                                                                                                                        'fiber_length': 1892,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYyNUU3cGVlcnBvbnBvbiA0NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST',
                                         'poller_hash': 'MjYyNUU3',
                                         'serial_number': '2625E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:f6:d7:7d',
                                         'uptime': '4155296'},
                                 {       'counter_source': '477/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:60:02'],
                                                                       'id': '477/g4',
                                                                       'in_nucast_pkts': '1351049',
                                                                       'in_octets': '23203833608',
                                                                       'in_ucast_pkts': '277655853',
                                                                       'mac_address': 'cc:be:59:61:1a:f9',
                                                                       'mtu': 2000,
                                                                       'name': '477/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '190646173',
                                                                       'out_octets': '1707134256106',
                                                                       'out_ucast_pkts': '1101205368',
                                                                       'poller_hash': 'NDU5NjhDcGVlcmludGY0NzcvZzQ=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '477',
                                                                       'in_nucast_pkts': '190646173',
                                                                       'in_octets': '1707134256106',
                                                                       'in_ucast_pkts': '1101205368',
                                                                       'links': [       {       'in_nucast_pkts': '190646173',
                                                                                                'in_octets': '1707134256106',
                                                                                                'in_ucast_pkts': '1101205368',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1351049',
                                                                                                'out_octets': '23203833608',
                                                                                                'out_ucast_pkts': '277655853',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1351049',
                                                                       'out_octets': '23203833608',
                                                                       'out_ucast_pkts': '277655853',
                                                                       'poller_hash': 'NDU5NjhDcGVlcmludGZwb24gNDc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '477',
                                                                                               'links': [       {       'connected_time': '4468870',
                                                                                                                        'fiber_length': 1515,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.808,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDU5NjhDcGVlcnBvbnBvbiA0Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '50 William St. Milverton (Knollcrest '
                                                 'Data Rm. 1)',
                                         'poller_hash': 'NDU5NjhD',
                                         'serial_number': '45968C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:61:1a:f2',
                                         'uptime': '4468870'},
                                 {       'counter_source': '478/f1+478/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d2:0d'],
                                                                       'id': '478/f1',
                                                                       'in_nucast_pkts': '6',
                                                                       'in_octets': '5603147358',
                                                                       'in_ucast_pkts': '24699962',
                                                                       'mac_address': '00:06:31:06:d1:e3',
                                                                       'mtu': 1608,
                                                                       'name': '478/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7248029',
                                                                       'out_octets': '30644929682',
                                                                       'out_ucast_pkts': '39801679',
                                                                       'poller_hash': 'MTUzRDhwZWVyaW50ZjQ3OC9mMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:42:a1:2c:71:d0'],
                                                                       'id': '478/g1',
                                                                       'in_nucast_pkts': '9516',
                                                                       'in_octets': '14086363625',
                                                                       'in_ucast_pkts': '58798968',
                                                                       'mac_address': '00:06:31:08:e6:f4',
                                                                       'mtu': 1608,
                                                                       'name': '478/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7248936',
                                                                       'out_octets': '108383091288',
                                                                       'out_ucast_pkts': '113995761',
                                                                       'poller_hash': 'MTUzRDhwZWVyaW50ZjQ3OC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '478',
                                                                       'in_nucast_pkts': '14496965',
                                                                       'in_octets': '139028020970',
                                                                       'in_ucast_pkts': '153797440',
                                                                       'links': [       {       'in_nucast_pkts': '14496965',
                                                                                                'in_octets': '139028020970',
                                                                                                'in_ucast_pkts': '153797440',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9522',
                                                                                                'out_octets': '19689510983',
                                                                                                'out_ucast_pkts': '83498930',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9522',
                                                                       'out_octets': '19689510983',
                                                                       'out_ucast_pkts': '83498930',
                                                                       'poller_hash': 'MTUzRDhwZWVyaW50ZnBvbiA0Nzg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '478',
                                                                                               'links': [       {       'connected_time': '4164913',
                                                                                                                        'fiber_length': 1456,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUzRDhwZWVycG9ucG9uIDQ3OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '714G',
                                         'name': '50 William St. Milverton '
                                                 '(Countryside Midwifery Services '
                                                 '10400769)',
                                         'poller_hash': 'MTUzRDg=',
                                         'serial_number': '153D8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:e6:f4',
                                         'uptime': '4164913'},
                                 {       'counter_source': '479/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c8:4f:86:c6:42:10'],
                                                                       'id': '479/g1',
                                                                       'in_nucast_pkts': '3007',
                                                                       'in_octets': '134255210658',
                                                                       'in_ucast_pkts': '417121203',
                                                                       'mac_address': '00:06:31:68:34:fe',
                                                                       'mtu': 2000,
                                                                       'name': '479/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7173891',
                                                                       'out_octets': '299164406553',
                                                                       'out_ucast_pkts': '472619677',
                                                                       'poller_hash': 'RDNFMzJwZWVyaW50ZjQ3OS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '479',
                                                                       'in_nucast_pkts': '7173891',
                                                                       'in_octets': '299164406553',
                                                                       'in_ucast_pkts': '472619677',
                                                                       'links': [       {       'in_nucast_pkts': '7173891',
                                                                                                'in_octets': '299164406553',
                                                                                                'in_ucast_pkts': '472619677',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3007',
                                                                                                'out_octets': '134255210658',
                                                                                                'out_ucast_pkts': '417121203',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3007',
                                                                       'out_octets': '134255210658',
                                                                       'out_ucast_pkts': '417121203',
                                                                       'poller_hash': 'RDNFMzJwZWVyaW50ZnBvbiA0Nzk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '479',
                                                                                               'links': [       {       'connected_time': '4165045',
                                                                                                                        'fiber_length': 1620,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RDNFMzJwZWVycG9ucG9uIDQ3OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '50 William St. Milverton (Dr. Hodes '
                                                 '& Anderson 10401327)',
                                         'poller_hash': 'RDNFMzI=',
                                         'serial_number': 'D3E32',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:6a:4e:f2',
                                         'uptime': '4165045'},
                                 {       'counter_source': '480/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '480/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '298',
                                                                       'in_octets': '226781213409',
                                                                       'in_ucast_pkts': '446971699',
                                                                       'mac_address': '60:db:98:73:81:3c',
                                                                       'name': '480/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1454',
                                                                       'out_octets': '638555745866',
                                                                       'out_ucast_pkts': '682367607',
                                                                       'poller_hash': 'QjA5RjM5cGVlcmludGY0ODAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '480',
                                                                       'in_nucast_pkts': '1454',
                                                                       'in_octets': '638555745866',
                                                                       'in_ucast_pkts': '682367607',
                                                                       'links': [       {       'in_nucast_pkts': '1454',
                                                                                                'in_octets': '638555745866',
                                                                                                'in_ucast_pkts': '682367607',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '298',
                                                                                                'out_octets': '226781213409',
                                                                                                'out_ucast_pkts': '446971699',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '298',
                                                                       'out_octets': '226781213409',
                                                                       'out_ucast_pkts': '446971699',
                                                                       'poller_hash': 'QjA5RjM5cGVlcmludGZwb24gNDgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '480',
                                                                                               'links': [       {       'connected_time': '13238676',
                                                                                                                        'fiber_length': 1461,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjA5RjM5cGVlcnBvbnBvbiA0ODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6619 PERTH ROAD 131 (MILVERTON '
                                                 'CHRISTIAN FELLOWSHIP)',
                                         'poller_hash': 'QjA5RjM5',
                                         'serial_number': 'B09F39',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:81:3c',
                                         'uptime': '13238676'},
                                 {       'counter_source': '481/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:82:08:01'],
                                                                       'id': '481/g2',
                                                                       'in_nucast_pkts': '797853',
                                                                       'in_octets': '1227044624553',
                                                                       'in_ucast_pkts': '5147749106',
                                                                       'mac_address': 'ec:4f:82:cc:79:39',
                                                                       'mtu': 2000,
                                                                       'name': '481/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '77791438',
                                                                       'out_octets': '18002206058634',
                                                                       'out_ucast_pkts': '13502299906',
                                                                       'poller_hash': 'MzkwQzVFcGVlcmludGY0ODEvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '481',
                                                                       'in_nucast_pkts': '77791438',
                                                                       'in_octets': '18002206058634',
                                                                       'in_ucast_pkts': '13502299906',
                                                                       'links': [       {       'in_nucast_pkts': '77791438',
                                                                                                'in_octets': '18002206058634',
                                                                                                'in_ucast_pkts': '13502299906',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '797853',
                                                                                                'out_octets': '1227044624553',
                                                                                                'out_ucast_pkts': '5147749106',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '797853',
                                                                       'out_octets': '1227044624553',
                                                                       'out_ucast_pkts': '5147749106',
                                                                       'poller_hash': 'MzkwQzVFcGVlcmludGZwb24gNDgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '481',
                                                                                               'links': [       {       'connected_time': '41151',
                                                                                                                        'fiber_length': 9861,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.482,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzkwQzVFcGVlcnBvbnBvbiA0ODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5384 Line 60 (Don&Barb Boyne)',
                                         'poller_hash': 'MzkwQzVF',
                                         'serial_number': '390C5E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cc:79:34',
                                         'uptime': '41151'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '482/g1',
                                                                       'name': '482/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEVFMDhwZWVyaW50ZjQ4Mi9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '482',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEVFMDhwZWVyaW50ZnBvbiA0ODI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '482',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEVFMDhwZWVycG9ucG9uIDQ4Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '8 Pugh St. Milverton (MCCL Cell '
                                                 'Tower)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "482" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-24T08:29:34.020Z',
                                         'poller_hash': 'NEVFMDg=',
                                         'serial_number': '4EE08',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '485/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '485/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '157374702913',
                                                                       'in_ucast_pkts': '707686655',
                                                                       'mac_address': '00:06:31:87:48:e3',
                                                                       'name': '485/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2599461724285',
                                                                       'out_ucast_pkts': '2112960881',
                                                                       'poller_hash': 'Q0MwQzJwZWVyaW50ZjQ4NS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '485',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2599461724285',
                                                                       'in_ucast_pkts': '2112960881',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2599461724285',
                                                                                                'in_ucast_pkts': '2112960881',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '157374702913',
                                                                                                'out_ucast_pkts': '707686655',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '157374702913',
                                                                       'out_ucast_pkts': '707686655',
                                                                       'poller_hash': 'Q0MwQzJwZWVyaW50ZnBvbiA0ODU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '485',
                                                                                               'links': [       {       'connected_time': '47302',
                                                                                                                        'fiber_length': 1213,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQzJwZWVycG9ucG9uIDQ4NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '22 Arena Dr. Milverton (Hairstyling '
                                                 'Headquarters 10401026)',
                                         'poller_hash': 'Q0MwQzI=',
                                         'serial_number': 'CC0C2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:82',
                                         'uptime': '47302'},
                                 {       'counter_source': '487/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '487/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '514635331522',
                                                                       'in_ucast_pkts': '219588034636',
                                                                       'mac_address': 'ec:4f:82:4f:87:af',
                                                                       'name': '487/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '1655412250790',
                                                                       'out_ucast_pkts': '1452846242',
                                                                       'poller_hash': 'Mjk4M0Y4cGVlcmludGY0ODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '487',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '1655412250790',
                                                                       'in_ucast_pkts': '1452846242',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '1655412250790',
                                                                                                'in_ucast_pkts': '1452846242',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '514635331522',
                                                                                                'out_ucast_pkts': '219588034636',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '514635331522',
                                                                       'out_ucast_pkts': '219588034636',
                                                                       'poller_hash': 'Mjk4M0Y4cGVlcmludGZwb24gNDg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '487',
                                                                                               'links': [       {       'connected_time': '688483',
                                                                                                                        'fiber_length': 671,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4M0Y4cGVlcnBvbnBvbiA0ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6 MILL ST E APT 2 (STREAMLINE '
                                                 'ENGINEERING INC)',
                                         'poller_hash': 'Mjk4M0Y4',
                                         'serial_number': '2983F8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:87:af',
                                         'uptime': '688483'},
                                 {       'counter_source': '488/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:2f:d0:36:09:f1'],
                                                                       'id': '488/g1',
                                                                       'in_nucast_pkts': '16707',
                                                                       'in_octets': '7060355395',
                                                                       'in_ucast_pkts': '59539725',
                                                                       'mac_address': 'ec:4f:82:7d:c9:89',
                                                                       'mtu': 2000,
                                                                       'name': '488/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464202',
                                                                       'out_octets': '250330173336',
                                                                       'out_ucast_pkts': '177163459',
                                                                       'poller_hash': 'MzAwREYxcGVlcmludGY0ODgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '488',
                                                                       'in_nucast_pkts': '1464202',
                                                                       'in_octets': '250330173336',
                                                                       'in_ucast_pkts': '177163459',
                                                                       'links': [       {       'in_nucast_pkts': '1464202',
                                                                                                'in_octets': '250330173336',
                                                                                                'in_ucast_pkts': '177163459',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16707',
                                                                                                'out_octets': '7060355395',
                                                                                                'out_ucast_pkts': '59539725',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16707',
                                                                       'out_octets': '7060355395',
                                                                       'out_ucast_pkts': '59539725',
                                                                       'poller_hash': 'MzAwREYxcGVlcmludGZwb24gNDg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '488',
                                                                                               'links': [       {       'connected_time': '688506',
                                                                                                                        'fiber_length': 1365,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwREYxcGVlcnBvbnBvbiA0ODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '1 Whaley Ave. Milverton (Nick Beggs)',
                                         'poller_hash': 'MzAwREYx',
                                         'serial_number': '300DF1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c9:89',
                                         'uptime': '688506'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '489/g1',
                                                                       'name': '489/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVFOTM4cGVlcmludGY0ODkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '489',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzVFOTM4cGVlcmludGZwb24gNDg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '489',
                                                                                               'links': [       {       'connected_time': '67790431',
                                                                                                                        'fiber_length': 1377,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTM4cGVlcnBvbnBvbiA0ODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '22 Whaley Av. Milverton (Roy Wolfe)',
                                         'poller_hash': 'MzVFOTM4',
                                         'serial_number': '35E938',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:08:c8',
                                         'uptime': '67790431'},
                                 {       'counter_source': '490/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '490/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '224166089019',
                                                                       'in_ucast_pkts': '1804314409',
                                                                       'mac_address': '00:06:31:b6:8b:e4',
                                                                       'name': '490/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4833385123037',
                                                                       'out_ucast_pkts': '1072283128',
                                                                       'poller_hash': 'MTU1M0JGcGVlcmludGY0OTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '490',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4833385123037',
                                                                       'in_ucast_pkts': '1072283128',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4833385123037',
                                                                                                'in_ucast_pkts': '1072283128',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '224166089019',
                                                                                                'out_ucast_pkts': '1804314409',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '224166089019',
                                                                       'out_ucast_pkts': '1804314409',
                                                                       'poller_hash': 'MTU1M0JGcGVlcmludGZwb24gNDkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '490',
                                                                                               'links': [       {       'connected_time': '47253',
                                                                                                                        'fiber_length': 17584,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.174,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1M0JGcGVlcnBvbnBvbiA0OTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '195 Winstanley St Monkton (Ron McKee '
                                                 '10433696)',
                                         'poller_hash': 'MTU1M0JG',
                                         'serial_number': '1553BF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:3a:05',
                                         'uptime': '47253'},
                                 {       'counter_source': '491/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '491/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '471054396504',
                                                                       'in_ucast_pkts': '4207853903',
                                                                       'mac_address': 'cc:be:59:cb:42:f2',
                                                                       'name': '491/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '14346508048285',
                                                                       'out_ucast_pkts': '10087272070',
                                                                       'poller_hash': 'NDJCOUI2cGVlcmludGY0OTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '491',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '14346508048285',
                                                                       'in_ucast_pkts': '10087272070',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '14346508048285',
                                                                                                'in_ucast_pkts': '10087272070',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '471054396504',
                                                                                                'out_ucast_pkts': '4207853903',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '471054396504',
                                                                       'out_ucast_pkts': '4207853903',
                                                                       'poller_hash': 'NDJCOUI2cGVlcmludGZwb24gNDkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '491',
                                                                                               'links': [       {       'connected_time': '2146718',
                                                                                                                        'fiber_length': 1328,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUI2cGVlcnBvbnBvbiA0OTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '45 Temperance St (Lindsay Hammer '
                                                 '10423568)',
                                         'poller_hash': 'NDJCOUI2',
                                         'serial_number': '42B9B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:42:f2',
                                         'uptime': '2146718'},
                                 {       'counter_source': '493/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:e3:19'],
                                                                       'id': '493/g1',
                                                                       'in_nucast_pkts': '3848',
                                                                       'in_octets': '367821721349',
                                                                       'in_ucast_pkts': '2860495659',
                                                                       'mac_address': 'ec:4f:82:59:ea:4e',
                                                                       'mtu': 2000,
                                                                       'name': '493/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '72837063',
                                                                       'out_octets': '12483074364983',
                                                                       'out_ucast_pkts': '8641015203',
                                                                       'poller_hash': 'MkFEMEJDcGVlcmludGY0OTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '493',
                                                                       'in_nucast_pkts': '72837063',
                                                                       'in_octets': '12483074364983',
                                                                       'in_ucast_pkts': '8641015203',
                                                                       'links': [       {       'in_nucast_pkts': '72837063',
                                                                                                'in_octets': '12483074364983',
                                                                                                'in_ucast_pkts': '8641015203',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3848',
                                                                                                'out_octets': '367821721349',
                                                                                                'out_ucast_pkts': '2860495659',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3848',
                                                                       'out_octets': '367821721349',
                                                                       'out_ucast_pkts': '2860495659',
                                                                       'poller_hash': 'MkFEMEJDcGVlcmludGZwb24gNDkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '493',
                                                                                               'links': [       {       'connected_time': '64131621',
                                                                                                                        'fiber_length': 18080,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFEMEJDcGVlcnBvbnBvbiA0OTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '49 Maddison St W (A & A McKenzie)',
                                         'poller_hash': 'MkFEMEJD',
                                         'serial_number': '2AD0BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:ea:4b',
                                         'uptime': '64131621'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '495/g1',
                                                                       'name': '495/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGODUxcGVlcmludGY0OTUvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '495',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGODUxcGVlcmludGZwb24gNDk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '495',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGODUxcGVlcnBvbnBvbiA0OTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '165 Smith St. Monkton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "495" '
                                                                  '(location: GPON '
                                                                  'port "11/1"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-24T08:33:17.110Z',
                                         'poller_hash': 'MzVGODUx',
                                         'serial_number': '35F851',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '496/g1',
                                                                       'name': '496/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGODQycGVlcmludGY0OTYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '496',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzVGODQycGVlcmludGZwb24gNDk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '496',
                                                                                               'links': [       {       'connected_time': '6668559',
                                                                                                                        'fiber_length': 1168,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGODQycGVlcnBvbnBvbiA0OTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '7 llana Heights Upper Apt (Dereth '
                                                 'Kipfer)',
                                         'poller_hash': 'MzVGODQy',
                                         'serial_number': '35F842',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:36:0a',
                                         'uptime': '6668559'},
                                 {       'counter_source': '497/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b6:26'],
                                                                       'id': '497/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '601970829',
                                                                       'in_ucast_pkts': '1300130',
                                                                       'mac_address': 'cc:be:59:ab:88:cf',
                                                                       'mtu': 2000,
                                                                       'name': '497/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120044',
                                                                       'out_octets': '4476123378',
                                                                       'out_ucast_pkts': '3646325',
                                                                       'poller_hash': 'NEZBMjQ0cGVlcmludGY0OTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '497',
                                                                       'in_nucast_pkts': '1120044',
                                                                       'in_octets': '4476123378',
                                                                       'in_ucast_pkts': '3646325',
                                                                       'links': [       {       'in_nucast_pkts': '1120044',
                                                                                                'in_octets': '4476123378',
                                                                                                'in_ucast_pkts': '3646325',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '601970829',
                                                                                                'out_ucast_pkts': '1300130',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '601970829',
                                                                       'out_ucast_pkts': '1300130',
                                                                       'poller_hash': 'NEZBMjQ0cGVlcmludGZwb24gNDk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '497',
                                                                                               'links': [       {       'connected_time': '688345',
                                                                                                                        'fiber_length': 862,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.074,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.336}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMjQ0cGVlcnBvbnBvbiA0OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '3 West St Milverton (Joe Kuepfer '
                                                 '10426185)',
                                         'poller_hash': 'NEZBMjQ0',
                                         'serial_number': '4FA244',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:88:ce',
                                         'uptime': '688345'},
                                 {       'counter_source': '499/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '499/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '198810881412',
                                                                       'in_ucast_pkts': '8802932905',
                                                                       'mac_address': 'ec:4f:82:5b:72:ab',
                                                                       'name': '499/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '167',
                                                                       'out_octets': '297116772035',
                                                                       'out_ucast_pkts': '361993592',
                                                                       'poller_hash': 'MkIwMUM4cGVlcmludGY0OTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '499',
                                                                       'in_nucast_pkts': '167',
                                                                       'in_octets': '297116772035',
                                                                       'in_ucast_pkts': '361993592',
                                                                       'links': [       {       'in_nucast_pkts': '167',
                                                                                                'in_octets': '297116772035',
                                                                                                'in_ucast_pkts': '361993592',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '198810881412',
                                                                                                'out_ucast_pkts': '8802932905',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '198810881412',
                                                                       'out_ucast_pkts': '8802932905',
                                                                       'poller_hash': 'MkIwMUM4cGVlcmludGZwb24gNDk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '499',
                                                                                               'links': [       {       'connected_time': '13238750',
                                                                                                                        'fiber_length': 1181,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMUM4cGVlcnBvbnBvbiA0OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '53 MAIN ST N (HESSON CHRISTIAN '
                                                 'FELLOWSHIP)',
                                         'poller_hash': 'MkIwMUM4',
                                         'serial_number': '2B01C8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:72:ab',
                                         'uptime': '13238750'},
                                 {       'counter_source': '500/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:38:8c'],
                                                                       'id': '500/g1',
                                                                       'in_nucast_pkts': '14466',
                                                                       'in_octets': '2249402765',
                                                                       'in_ucast_pkts': '22153661',
                                                                       'mac_address': 'ec:4f:82:5b:70:76',
                                                                       'mtu': 2000,
                                                                       'name': '500/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464215',
                                                                       'out_octets': '185160171225',
                                                                       'out_ucast_pkts': '124934828',
                                                                       'poller_hash': 'MkIwMTgxcGVlcmludGY1MDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '500',
                                                                       'in_nucast_pkts': '1464215',
                                                                       'in_octets': '185160171225',
                                                                       'in_ucast_pkts': '124934828',
                                                                       'links': [       {       'in_nucast_pkts': '1464215',
                                                                                                'in_octets': '185160171225',
                                                                                                'in_ucast_pkts': '124934828',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14466',
                                                                                                'out_octets': '2249402765',
                                                                                                'out_ucast_pkts': '22153661',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14466',
                                                                       'out_octets': '2249402765',
                                                                       'out_ucast_pkts': '22153661',
                                                                       'poller_hash': 'MkIwMTgxcGVlcmludGZwb24gNTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '500',
                                                                                               'links': [       {       'connected_time': '688380',
                                                                                                                        'fiber_length': 1426,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTgxcGVlcnBvbnBvbiA1MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6472 Perth Road 131 (Brett Kipfer)',
                                         'poller_hash': 'MkIwMTgx',
                                         'serial_number': '2B0181',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:73',
                                         'uptime': '688380'},
                                 {       'counter_source': '503/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '503/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '111407538674',
                                                                       'in_ucast_pkts': '419282533',
                                                                       'mac_address': 'ec:4f:82:a2:b8:18',
                                                                       'name': '503/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2109549162142',
                                                                       'out_ucast_pkts': '1642253180',
                                                                       'poller_hash': 'MzQ5RTg1cGVlcmludGY1MDMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '503',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2109549162142',
                                                                       'in_ucast_pkts': '1642253180',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2109549162142',
                                                                                                'in_ucast_pkts': '1642253180',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '111407538674',
                                                                                                'out_ucast_pkts': '419282533',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '111407538674',
                                                                       'out_ucast_pkts': '419282533',
                                                                       'poller_hash': 'MzQ5RTg1cGVlcmludGZwb24gNTAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '503',
                                                                                               'links': [       {       'connected_time': '21536360',
                                                                                                                        'fiber_length': 835,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ5RTg1cGVlcnBvbnBvbiA1MDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '27 Main St S Milverton (Richard '
                                                 'Smallbone 10433849)',
                                         'poller_hash': 'MzQ5RTg1',
                                         'serial_number': '349E85',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:a2:b8:18',
                                         'uptime': '21536360'},
                                 {       'counter_source': '504/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:db:db'],
                                                                       'id': '504/g1',
                                                                       'in_nucast_pkts': '2937',
                                                                       'in_octets': '88957209788',
                                                                       'in_ucast_pkts': '530711453',
                                                                       'mac_address': 'ec:4f:82:7e:0a:87',
                                                                       'mtu': 2000,
                                                                       'name': '504/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83496319',
                                                                       'out_octets': '4779695010978',
                                                                       'out_ucast_pkts': '3226699286',
                                                                       'poller_hash': 'MzVFOUNEcGVlcmludGY1MDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '504',
                                                                       'in_nucast_pkts': '83496319',
                                                                       'in_octets': '4779695010978',
                                                                       'in_ucast_pkts': '3226699286',
                                                                       'links': [       {       'in_nucast_pkts': '83496319',
                                                                                                'in_octets': '4779695010978',
                                                                                                'in_ucast_pkts': '3226699286',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2937',
                                                                                                'out_octets': '88957209788',
                                                                                                'out_ucast_pkts': '530711453',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2937',
                                                                       'out_octets': '88957209788',
                                                                       'out_ucast_pkts': '530711453',
                                                                       'poller_hash': 'MzVFOUNEcGVlcmludGZwb24gNTA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '504',
                                                                                               'links': [       {       'connected_time': '67790585',
                                                                                                                        'fiber_length': 18384,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOUNEcGVlcnBvbnBvbiA1MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '100 Maddison S W Monkton (Gillian '
                                                 'Ward)',
                                         'poller_hash': 'MzVFOUNE',
                                         'serial_number': '35E9CD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0a:87',
                                         'uptime': '67790585'},
                                 {       'counter_source': '505/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '505/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '243082015734',
                                                                       'in_ucast_pkts': '2071281673',
                                                                       'mac_address': 'ec:4f:82:5b:71:5b',
                                                                       'name': '505/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '112',
                                                                       'out_octets': '11216537451373',
                                                                       'out_ucast_pkts': '7648713625',
                                                                       'poller_hash': 'MkIwMTlFcGVlcmludGY1MDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '505',
                                                                       'in_nucast_pkts': '112',
                                                                       'in_octets': '11216537451373',
                                                                       'in_ucast_pkts': '7648713625',
                                                                       'links': [       {       'in_nucast_pkts': '112',
                                                                                                'in_octets': '11216537451373',
                                                                                                'in_ucast_pkts': '7648713625',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '243082015734',
                                                                                                'out_ucast_pkts': '2071281673',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '243082015734',
                                                                       'out_ucast_pkts': '2071281673',
                                                                       'poller_hash': 'MkIwMTlFcGVlcmludGZwb24gNTA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '505',
                                                                                               'links': [       {       'connected_time': '2327514',
                                                                                                                        'fiber_length': 18560,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTlFcGVlcnBvbnBvbiA1MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '146 Brock St Monkton (Kelly '
                                                 'Battams-Trussler 10430830)',
                                         'poller_hash': 'MkIwMTlF',
                                         'serial_number': '2B019E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:71:5b',
                                         'uptime': '2327514'},
                                 {       'counter_source': '506/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:f5:a2:09:16:62'],
                                                                       'id': '506/g1',
                                                                       'in_nucast_pkts': '29196',
                                                                       'in_octets': '2695353400',
                                                                       'in_ucast_pkts': '12193942',
                                                                       'mac_address': 'ec:4f:82:7d:ca:91',
                                                                       'mtu': 2000,
                                                                       'name': '506/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464824',
                                                                       'out_octets': '70923633368',
                                                                       'out_ucast_pkts': '51505076',
                                                                       'poller_hash': 'MzAwRTc1cGVlcmludGY1MDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '506',
                                                                       'in_nucast_pkts': '1464824',
                                                                       'in_octets': '70923633368',
                                                                       'in_ucast_pkts': '51505076',
                                                                       'links': [       {       'in_nucast_pkts': '1464824',
                                                                                                'in_octets': '70923633368',
                                                                                                'in_ucast_pkts': '51505076',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '29196',
                                                                                                'out_octets': '2695353400',
                                                                                                'out_ucast_pkts': '12193942',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29196',
                                                                       'out_octets': '2695353400',
                                                                       'out_ucast_pkts': '12193942',
                                                                       'poller_hash': 'MzAwRTc1cGVlcmludGZwb24gNTA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '506',
                                                                                               'links': [       {       'connected_time': '688359',
                                                                                                                        'fiber_length': 822,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRTc1cGVlcnBvbnBvbiA1MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '25 MAIN ST S (HALL MICHAEL)',
                                         'poller_hash': 'MzAwRTc1',
                                         'serial_number': '300E75',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:ca:91',
                                         'uptime': '688359'},
                                 {       'counter_source': '508/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '90:98:38:54:67:72'],
                                                                       'id': '508/g1',
                                                                       'in_nucast_pkts': '4432',
                                                                       'in_octets': '144929379647',
                                                                       'in_ucast_pkts': '589831033',
                                                                       'mac_address': 'ec:4f:82:7e:09:58',
                                                                       'mtu': 2000,
                                                                       'name': '508/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35075152',
                                                                       'out_octets': '3605241988876',
                                                                       'out_ucast_pkts': '2699736265',
                                                                       'poller_hash': 'MzVFOTY4cGVlcmludGY1MDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '508',
                                                                       'in_nucast_pkts': '35075152',
                                                                       'in_octets': '3605241988876',
                                                                       'in_ucast_pkts': '2699736265',
                                                                       'links': [       {       'in_nucast_pkts': '35075152',
                                                                                                'in_octets': '3605241988876',
                                                                                                'in_ucast_pkts': '2699736265',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4432',
                                                                                                'out_octets': '144929379647',
                                                                                                'out_ucast_pkts': '589831033',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4432',
                                                                       'out_octets': '144929379647',
                                                                       'out_ucast_pkts': '589831033',
                                                                       'poller_hash': 'MzVFOTY4cGVlcmludGZwb24gNTA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '508',
                                                                                               'links': [       {       'connected_time': '15499141',
                                                                                                                        'fiber_length': 1213,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTY4cGVlcnBvbnBvbiA1MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '36 MILL ST E (NAFZIGER BRETT '
                                                 '10438396)',
                                         'poller_hash': 'MzVFOTY4',
                                         'serial_number': '35E968',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:58',
                                         'uptime': '15499141'},
                                 {       'counter_source': '509/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:63:4c:6d:fe:81'],
                                                                       'id': '509/g1',
                                                                       'in_nucast_pkts': '168',
                                                                       'in_octets': '22140735047',
                                                                       'in_ucast_pkts': '52366839',
                                                                       'mac_address': 'ec:4f:82:7d:c8:d1',
                                                                       'mtu': 2000,
                                                                       'name': '509/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1425841',
                                                                       'out_octets': '130498193743',
                                                                       'out_ucast_pkts': '114871893',
                                                                       'poller_hash': 'MzAwRDk1cGVlcmludGY1MDkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '509',
                                                                       'in_nucast_pkts': '1425841',
                                                                       'in_octets': '130498193743',
                                                                       'in_ucast_pkts': '114871893',
                                                                       'links': [       {       'in_nucast_pkts': '1425841',
                                                                                                'in_octets': '130498193743',
                                                                                                'in_ucast_pkts': '114871893',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '168',
                                                                                                'out_octets': '22140735047',
                                                                                                'out_ucast_pkts': '52366839',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '168',
                                                                       'out_octets': '22140735047',
                                                                       'out_ucast_pkts': '52366839',
                                                                       'poller_hash': 'MzAwRDk1cGVlcmludGZwb24gNTA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '509',
                                                                                               'links': [       {       'connected_time': '670617',
                                                                                                                        'fiber_length': 19562,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRDk1cGVlcnBvbnBvbiA1MDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5792 LINE 57 (VAN NULAND MARIA '
                                                 '(MELANIE))',
                                         'poller_hash': 'MzAwRDk1',
                                         'serial_number': '300D95',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c8:d1',
                                         'uptime': '670617'},
                                 {       'counter_source': '512/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '512/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '877962591672',
                                                                       'in_ucast_pkts': '2259933767',
                                                                       'mac_address': 'ec:4f:82:8d:44:e8',
                                                                       'name': '512/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '7737368515707',
                                                                       'out_ucast_pkts': '6013743906',
                                                                       'poller_hash': 'MzIxMEU0cGVlcmludGY1MTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '512',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '7737368515707',
                                                                       'in_ucast_pkts': '6013743906',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '7737368515707',
                                                                                                'in_ucast_pkts': '6013743906',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '877962591672',
                                                                                                'out_ucast_pkts': '2259933767',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '877962591672',
                                                                       'out_ucast_pkts': '2259933767',
                                                                       'poller_hash': 'MzIxMEU0cGVlcmludGZwb24gNTEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '512',
                                                                                               'links': [       {       'connected_time': '9776280',
                                                                                                                        'fiber_length': 1483,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMEU0cGVlcnBvbnBvbiA1MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 Arena Dr. Milverton (Denise '
                                                 'Smith)',
                                         'poller_hash': 'MzIxMEU0',
                                         'serial_number': '3210E4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:44:e8',
                                         'uptime': '9776280'},
                                 {       'counter_source': '513/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6f:13:40:a4:f3'],
                                                                       'id': '513/g1',
                                                                       'in_nucast_pkts': '59',
                                                                       'in_octets': '1533697095',
                                                                       'in_ucast_pkts': '5865607',
                                                                       'mac_address': 'ec:4f:82:63:54:4b',
                                                                       'mtu': 2000,
                                                                       'name': '513/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464395',
                                                                       'out_octets': '40634573129',
                                                                       'out_ucast_pkts': '32570721',
                                                                       'poller_hash': 'MkJEMjY0cGVlcmludGY1MTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '513',
                                                                       'in_nucast_pkts': '1464395',
                                                                       'in_octets': '40634573129',
                                                                       'in_ucast_pkts': '32570721',
                                                                       'links': [       {       'in_nucast_pkts': '1464395',
                                                                                                'in_octets': '40634573129',
                                                                                                'in_ucast_pkts': '32570721',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '59',
                                                                                                'out_octets': '1533697095',
                                                                                                'out_ucast_pkts': '5865607',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '59',
                                                                       'out_octets': '1533697095',
                                                                       'out_ucast_pkts': '5865607',
                                                                       'poller_hash': 'MkJEMjY0cGVlcmludGZwb24gNTEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '513',
                                                                                               'links': [       {       'connected_time': '688481',
                                                                                                                        'fiber_length': 1575,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJEMjY0cGVlcnBvbnBvbiA1MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '82 Main St N (William & Krista '
                                                 'Kuepfer)',
                                         'poller_hash': 'MkJEMjY0',
                                         'serial_number': '2BD264',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:54:4b',
                                         'uptime': '688481'},
                                 {       'counter_source': '514/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e7:bb'],
                                                                       'id': '514/g1',
                                                                       'in_nucast_pkts': '13924',
                                                                       'in_octets': '1339517319',
                                                                       'in_ucast_pkts': '3578697',
                                                                       'mac_address': 'ec:4f:82:8e:10:03',
                                                                       'mtu': 2000,
                                                                       'name': '514/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464273',
                                                                       'out_octets': '20330314047',
                                                                       'out_ucast_pkts': '16533886',
                                                                       'poller_hash': 'MzIyQTQ3cGVlcmludGY1MTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '514',
                                                                       'in_nucast_pkts': '1464273',
                                                                       'in_octets': '20330314047',
                                                                       'in_ucast_pkts': '16533886',
                                                                       'links': [       {       'in_nucast_pkts': '1464273',
                                                                                                'in_octets': '20330314047',
                                                                                                'in_ucast_pkts': '16533886',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13924',
                                                                                                'out_octets': '1339517319',
                                                                                                'out_ucast_pkts': '3578697',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13924',
                                                                       'out_octets': '1339517319',
                                                                       'out_ucast_pkts': '3578697',
                                                                       'poller_hash': 'MzIyQTQ3cGVlcmludGZwb24gNTE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '514',
                                                                                               'links': [       {       'connected_time': '688547',
                                                                                                                        'fiber_length': 6301,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyQTQ3cGVlcnBvbnBvbiA1MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4344 LINE 61 (JANTZI JARRELL)',
                                         'poller_hash': 'MzIyQTQ3',
                                         'serial_number': '322A47',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:10:00',
                                         'uptime': '688547'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '516/g1',
                                                                       'name': '516/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGOUUwcGVlcmludGY1MTYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '516',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzVGOUUwcGVlcmludGZwb24gNTE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '516',
                                                                                               'links': [       {       'connected_time': '67790133',
                                                                                                                        'fiber_length': 1193,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUUwcGVlcnBvbnBvbiA1MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '27 Temperance St. Milverton (Chris '
                                                 'Potwarka 10401128)',
                                         'poller_hash': 'MzVGOUUw',
                                         'serial_number': '35F9E0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3a:e4',
                                         'uptime': '67790133'},
                                 {       'counter_source': '517/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:31:92:dc:5c:f9'],
                                                                       'id': '517/g1',
                                                                       'in_nucast_pkts': '114911',
                                                                       'in_octets': '6115642649',
                                                                       'in_ucast_pkts': '22939358',
                                                                       'mac_address': 'cc:be:59:ab:34:0f',
                                                                       'mtu': 2000,
                                                                       'name': '517/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1100833',
                                                                       'out_octets': '106528619350',
                                                                       'out_ucast_pkts': '82520321',
                                                                       'poller_hash': 'NEY5MUZGcGVlcmludGY1MTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '517',
                                                                       'in_nucast_pkts': '1100833',
                                                                       'in_octets': '106528619350',
                                                                       'in_ucast_pkts': '82520321',
                                                                       'links': [       {       'in_nucast_pkts': '1100833',
                                                                                                'in_octets': '106528619350',
                                                                                                'in_ucast_pkts': '82520321',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '114911',
                                                                                                'out_octets': '6115642649',
                                                                                                'out_ucast_pkts': '22939358',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '114911',
                                                                       'out_octets': '6115642649',
                                                                       'out_ucast_pkts': '22939358',
                                                                       'poller_hash': 'NEY5MUZGcGVlcmludGZwb24gNTE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '517',
                                                                                               'links': [       {       'connected_time': '688176',
                                                                                                                        'fiber_length': 2157,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.334,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.304}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5MUZGcGVlcnBvbnBvbiA1MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '61 Reagan St (Philip & Rosemary '
                                                 'Lebold,10411732)',
                                         'poller_hash': 'NEY5MUZG',
                                         'serial_number': '4F91FF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:34:0e',
                                         'uptime': '688176'},
                                 {       'counter_source': '520/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:94:11'],
                                                                       'id': '520/g1',
                                                                       'in_nucast_pkts': '11860',
                                                                       'in_octets': '11323735954',
                                                                       'in_ucast_pkts': '58066586',
                                                                       'mac_address': 'cc:be:59:ab:89:01',
                                                                       'mtu': 2000,
                                                                       'name': '520/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119990',
                                                                       'out_octets': '250244768355',
                                                                       'out_ucast_pkts': '180866581',
                                                                       'poller_hash': 'NEZBMjVEcGVlcmludGY1MjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '520',
                                                                       'in_nucast_pkts': '1119990',
                                                                       'in_octets': '250244768355',
                                                                       'in_ucast_pkts': '180866581',
                                                                       'links': [       {       'in_nucast_pkts': '1119990',
                                                                                                'in_octets': '250244768355',
                                                                                                'in_ucast_pkts': '180866581',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11860',
                                                                                                'out_octets': '11323735954',
                                                                                                'out_ucast_pkts': '58066586',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11860',
                                                                       'out_octets': '11323735954',
                                                                       'out_ucast_pkts': '58066586',
                                                                       'poller_hash': 'NEZBMjVEcGVlcmludGZwb24gNTIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '520',
                                                                                               'links': [       {       'connected_time': '688481',
                                                                                                                        'fiber_length': 1000,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.812,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.182}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMjVEcGVlcnBvbnBvbiA1MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '25 MILL ST W (ENNS ANTHONY)',
                                         'poller_hash': 'NEZBMjVE',
                                         'serial_number': '4FA25D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:89:00',
                                         'uptime': '688481'},
                                 {       'counter_source': '521/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:56:f9'],
                                                                       'id': '521/g1',
                                                                       'in_nucast_pkts': '14215',
                                                                       'in_octets': '15517818357',
                                                                       'in_ucast_pkts': '56792414',
                                                                       'mac_address': 'ec:4f:82:8d:fe:b3',
                                                                       'mtu': 2000,
                                                                       'name': '521/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464322',
                                                                       'out_octets': '241231029350',
                                                                       'out_ucast_pkts': '196171080',
                                                                       'poller_hash': 'MzIyODFEcGVlcmludGY1MjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '521',
                                                                       'in_nucast_pkts': '1464322',
                                                                       'in_octets': '241231029350',
                                                                       'in_ucast_pkts': '196171080',
                                                                       'links': [       {       'in_nucast_pkts': '1464322',
                                                                                                'in_octets': '241231029350',
                                                                                                'in_ucast_pkts': '196171080',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14215',
                                                                                                'out_octets': '15517818357',
                                                                                                'out_ucast_pkts': '56792414',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14215',
                                                                       'out_octets': '15517818357',
                                                                       'out_ucast_pkts': '56792414',
                                                                       'poller_hash': 'MzIyODFEcGVlcmludGZwb24gNTIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '521',
                                                                                               'links': [       {       'connected_time': '688472',
                                                                                                                        'fiber_length': 1707,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODFEcGVlcnBvbnBvbiA1MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '58 MAIN ST N (SYDNEY ROCKLAND)',
                                         'poller_hash': 'MzIyODFE',
                                         'serial_number': '32281D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fe:b0',
                                         'uptime': '688472'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '522/g1',
                                                                       'name': '522/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEY4QzlCcGVlcmludGY1MjIvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '522',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEY4QzlCcGVlcmludGZwb24gNTIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '522',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY4QzlCcGVlcnBvbnBvbiA1MjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '82 Gerber Drive Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "522" '
                                                                  '(location: GPON '
                                                                  'port "5/8"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:08:44.270Z',
                                         'poller_hash': 'NEY4QzlC',
                                         'serial_number': '4F8C9B',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '524/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '524/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '44766556446',
                                                                       'in_ucast_pkts': '587943962',
                                                                       'mac_address': 'ec:4f:82:0d:67:24',
                                                                       'name': '524/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '52',
                                                                       'out_octets': '3742026782464',
                                                                       'out_ucast_pkts': '2492691468',
                                                                       'poller_hash': 'MjA1RDk3cGVlcmludGY1MjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '524',
                                                                       'in_nucast_pkts': '52',
                                                                       'in_octets': '3742026782464',
                                                                       'in_ucast_pkts': '2492691468',
                                                                       'links': [       {       'in_nucast_pkts': '52',
                                                                                                'in_octets': '3742026782464',
                                                                                                'in_ucast_pkts': '2492691468',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '44766556446',
                                                                                                'out_ucast_pkts': '587943962',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '44766556446',
                                                                       'out_ucast_pkts': '587943962',
                                                                       'poller_hash': 'MjA1RDk3cGVlcmludGZwb24gNTI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '524',
                                                                                               'links': [       {       'connected_time': '21319494',
                                                                                                                        'fiber_length': 1116,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RDk3cGVlcnBvbnBvbiA1MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 MAIN ST S (MCCUTCHEN WINNIE)',
                                         'poller_hash': 'MjA1RDk3',
                                         'serial_number': '205D97',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:67:24',
                                         'uptime': '21319494'},
                                 {       'counter_source': '527/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:70:10'],
                                                                       'id': '527/g1',
                                                                       'in_nucast_pkts': '14000117',
                                                                       'in_octets': '263173175927',
                                                                       'in_ucast_pkts': '1464019983',
                                                                       'mac_address': 'cc:be:59:ca:df:55',
                                                                       'mtu': 2000,
                                                                       'name': '527/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33120464598',
                                                                       'out_octets': '50999486110972',
                                                                       'out_ucast_pkts': '4890112756',
                                                                       'poller_hash': 'NDI4MUU5cGVlcmludGY1MjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '527',
                                                                       'in_nucast_pkts': '33120464598',
                                                                       'in_octets': '50999486110972',
                                                                       'in_ucast_pkts': '4890112756',
                                                                       'links': [       {       'in_nucast_pkts': '33120464598',
                                                                                                'in_octets': '50999486110972',
                                                                                                'in_ucast_pkts': '4890112756',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14000117',
                                                                                                'out_octets': '263173175927',
                                                                                                'out_ucast_pkts': '1464019983',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14000117',
                                                                       'out_octets': '263173175927',
                                                                       'out_ucast_pkts': '1464019983',
                                                                       'poller_hash': 'NDI4MUU5cGVlcmludGZwb24gNTI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '527',
                                                                                               'links': [       {       'connected_time': '21815930',
                                                                                                                        'fiber_length': 18546,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUU5cGVlcnBvbnBvbiA1Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '161 Brock St Monkton (Wilma Acheson '
                                                 '10434491)',
                                         'poller_hash': 'NDI4MUU5',
                                         'serial_number': '4281E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:df:52',
                                         'uptime': '21815930'},
                                 {       'counter_source': '528/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '528/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1147385640305',
                                                                       'in_ucast_pkts': '78083633948',
                                                                       'mac_address': 'cc:be:59:ca:de:a2',
                                                                       'name': '528/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '22564136714527',
                                                                       'out_ucast_pkts': '17162058102',
                                                                       'poller_hash': 'NDI4MUQzcGVlcmludGY1MjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '528',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '22564136714527',
                                                                       'in_ucast_pkts': '17162058102',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '22564136714527',
                                                                                                'in_ucast_pkts': '17162058102',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1147385640305',
                                                                                                'out_ucast_pkts': '78083633948',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1147385640305',
                                                                       'out_ucast_pkts': '78083633948',
                                                                       'poller_hash': 'NDI4MUQzcGVlcmludGZwb24gNTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '528',
                                                                                               'links': [       {       'connected_time': '688407',
                                                                                                                        'fiber_length': 1955,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUQzcGVlcnBvbnBvbiA1Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '94 GERBER DR (BUDDELL WILLIAM)',
                                         'poller_hash': 'NDI4MUQz',
                                         'serial_number': '4281D3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:a2',
                                         'uptime': '688407'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '532/g1',
                                                                       'name': '532/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEY5QjM3cGVlcmludGY1MzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '532',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NEY5QjM3cGVlcmludGZwb24gNTMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '532',
                                                                                               'links': [       {       'connected_time': '688446',
                                                                                                                        'fiber_length': 1372,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.59,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.034}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5QjM3cGVlcnBvbnBvbiA1MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5 MacBeth St Milverton (Damion '
                                                 'Bijman 10434605)',
                                         'poller_hash': 'NEY5QjM3',
                                         'serial_number': '4F9B37',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:60:f8',
                                         'uptime': '688446'},
                                 {       'counter_source': '533/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:16:f9:42:b0:f6'],
                                                                       'id': '533/g1',
                                                                       'in_nucast_pkts': '3321',
                                                                       'in_octets': '89039843039',
                                                                       'in_ucast_pkts': '190568151',
                                                                       'mac_address': 'ec:4f:82:7e:3b:3b',
                                                                       'mtu': 2000,
                                                                       'name': '533/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21519927',
                                                                       'out_octets': '283886901934',
                                                                       'out_ucast_pkts': '284430128',
                                                                       'poller_hash': 'MzVGOUZEcGVlcmludGY1MzMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '533',
                                                                       'in_nucast_pkts': '21519927',
                                                                       'in_octets': '283886901934',
                                                                       'in_ucast_pkts': '284430128',
                                                                       'links': [       {       'in_nucast_pkts': '21519927',
                                                                                                'in_octets': '283886901934',
                                                                                                'in_ucast_pkts': '284430128',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3321',
                                                                                                'out_octets': '89039843039',
                                                                                                'out_ucast_pkts': '190568151',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3321',
                                                                       'out_octets': '89039843039',
                                                                       'out_ucast_pkts': '190568151',
                                                                       'poller_hash': 'MzVGOUZEcGVlcmludGZwb24gNTMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '533',
                                                                                               'links': [       {       'connected_time': '9885084',
                                                                                                                        'fiber_length': 11053,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUZEcGVlcnBvbnBvbiA1MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5843 Perth Road 119 (Larry P Janzti '
                                                 '10400600)',
                                         'poller_hash': 'MzVGOUZE',
                                         'serial_number': '35F9FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3b:3b',
                                         'uptime': '9885084'},
                                 {       'counter_source': '534/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:43:97:c6'],
                                                                       'id': '534/g1',
                                                                       'in_nucast_pkts': '323997',
                                                                       'in_octets': '5584177782',
                                                                       'in_ucast_pkts': '74857765',
                                                                       'mac_address': 'cc:be:59:ca:de:05',
                                                                       'mtu': 2000,
                                                                       'name': '534/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23687419',
                                                                       'out_octets': '304912172831',
                                                                       'out_ucast_pkts': '210412835',
                                                                       'poller_hash': 'NDI4MUJGcGVlcmludGY1MzQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '534',
                                                                       'in_nucast_pkts': '23687419',
                                                                       'in_octets': '304912172831',
                                                                       'in_ucast_pkts': '210412835',
                                                                       'links': [       {       'in_nucast_pkts': '23687419',
                                                                                                'in_octets': '304912172831',
                                                                                                'in_ucast_pkts': '210412835',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '323997',
                                                                                                'out_octets': '5584177782',
                                                                                                'out_ucast_pkts': '74857765',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '323997',
                                                                       'out_octets': '5584177782',
                                                                       'out_ucast_pkts': '74857765',
                                                                       'poller_hash': 'NDI4MUJGcGVlcmludGZwb24gNTM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '534',
                                                                                               'links': [       {       'connected_time': '1133960',
                                                                                                                        'fiber_length': 18858,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUJGcGVlcnBvbnBvbiA1MzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '76 Schade St Monkton (Murray '
                                                 'Crawford 10434552)',
                                         'poller_hash': 'NDI4MUJG',
                                         'serial_number': '4281BF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:02',
                                         'uptime': '1133960'},
                                 {       'counter_source': '535/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '535/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1068491653712',
                                                                       'in_ucast_pkts': '2387367482',
                                                                       'mac_address': 'cc:be:59:ca:de:3a',
                                                                       'name': '535/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '22448255535000',
                                                                       'out_ucast_pkts': '16014211482',
                                                                       'poller_hash': 'NDI4MUM2cGVlcmludGY1MzUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '535',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '22448255535000',
                                                                       'in_ucast_pkts': '16014211482',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '22448255535000',
                                                                                                'in_ucast_pkts': '16014211482',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1068491653712',
                                                                                                'out_ucast_pkts': '2387367482',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1068491653712',
                                                                       'out_ucast_pkts': '2387367482',
                                                                       'poller_hash': 'NDI4MUM2cGVlcmludGZwb24gNTM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '535',
                                                                                               'links': [       {       'connected_time': '1131830',
                                                                                                                        'fiber_length': 18593,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUM2cGVlcnBvbnBvbiA1MzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '139 Brock St Monkton (Angela Pestell '
                                                 '10412111)',
                                         'poller_hash': 'NDI4MUM2',
                                         'serial_number': '4281C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:3a',
                                         'uptime': '1131830'},
                                 {       'counter_source': '536/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:32:1b:49:95:0f'],
                                                                       'id': '536/g1',
                                                                       'in_nucast_pkts': '636',
                                                                       'in_octets': '5660138300',
                                                                       'in_ucast_pkts': '16923957',
                                                                       'mac_address': 'cc:be:59:ab:37:dd',
                                                                       'mtu': 2000,
                                                                       'name': '536/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120087',
                                                                       'out_octets': '54070125933',
                                                                       'out_ucast_pkts': '41863563',
                                                                       'poller_hash': 'NEY5MkNGcGVlcmludGY1MzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '536',
                                                                       'in_nucast_pkts': '1120087',
                                                                       'in_octets': '54070125933',
                                                                       'in_ucast_pkts': '41863563',
                                                                       'links': [       {       'in_nucast_pkts': '1120087',
                                                                                                'in_octets': '54070125933',
                                                                                                'in_ucast_pkts': '41863563',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '636',
                                                                                                'out_octets': '5660138300',
                                                                                                'out_ucast_pkts': '16923957',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '636',
                                                                       'out_octets': '5660138300',
                                                                       'out_ucast_pkts': '16923957',
                                                                       'poller_hash': 'NEY5MkNGcGVlcmludGZwb24gNTM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '536',
                                                                                               'links': [       {       'connected_time': '688525',
                                                                                                                        'fiber_length': 12718,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.864,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.138}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5MkNGcGVlcnBvbnBvbiA1MzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5713 Perth Rd 119 Brunner (Harry '
                                                 'Walmsley 10402219)',
                                         'poller_hash': 'NEY5MkNG',
                                         'serial_number': '4F92CF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:37:dc',
                                         'uptime': '688525'},
                                 {       'counter_source': '537/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '537/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '670633589940',
                                                                       'in_ucast_pkts': '149252515727',
                                                                       'mac_address': 'cc:be:59:ca:f2:6a',
                                                                       'name': '537/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '13311437601478',
                                                                       'out_ucast_pkts': '10445343543',
                                                                       'poller_hash': 'NDI4NDRDcGVlcmludGY1MzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '537',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '13311437601478',
                                                                       'in_ucast_pkts': '10445343543',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '13311437601478',
                                                                                                'in_ucast_pkts': '10445343543',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '670633589940',
                                                                                                'out_ucast_pkts': '149252515727',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '670633589940',
                                                                       'out_ucast_pkts': '149252515727',
                                                                       'poller_hash': 'NDI4NDRDcGVlcmludGZwb24gNTM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '537',
                                                                                               'links': [       {       'connected_time': '13124654',
                                                                                                                        'fiber_length': 18297,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4NDRDcGVlcnBvbnBvbiA1Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '126 Brock St (Kevin Mclean, '
                                                 '10422928)',
                                         'poller_hash': 'NDI4NDRD',
                                         'serial_number': '42844C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:f2:6a',
                                         'uptime': '13124654'},
                                 {       'counter_source': '539/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:ec:33'],
                                                                       'id': '539/g1',
                                                                       'in_nucast_pkts': '11902',
                                                                       'in_octets': '7299988557',
                                                                       'in_ucast_pkts': '45392140',
                                                                       'mac_address': 'cc:be:59:ab:5f:99',
                                                                       'mtu': 2000,
                                                                       'name': '539/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120052',
                                                                       'out_octets': '177601982216',
                                                                       'out_ucast_pkts': '123538236',
                                                                       'poller_hash': 'NEY5QUU0cGVlcmludGY1MzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '539',
                                                                       'in_nucast_pkts': '1120052',
                                                                       'in_octets': '177601982216',
                                                                       'in_ucast_pkts': '123538236',
                                                                       'links': [       {       'in_nucast_pkts': '1120052',
                                                                                                'in_octets': '177601982216',
                                                                                                'in_ucast_pkts': '123538236',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11902',
                                                                                                'out_octets': '7299988557',
                                                                                                'out_ucast_pkts': '45392140',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11902',
                                                                       'out_octets': '7299988557',
                                                                       'out_ucast_pkts': '45392140',
                                                                       'poller_hash': 'NEY5QUU0cGVlcmludGZwb24gNTM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '539',
                                                                                               'links': [       {       'connected_time': '688463',
                                                                                                                        'fiber_length': 748,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.17,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.246}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5QUU0cGVlcnBvbnBvbiA1Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '8A WILLIAM ST (HOLST SAMUEL)',
                                         'poller_hash': 'NEY5QUU0',
                                         'serial_number': '4F9AE4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:5f:98',
                                         'uptime': '688463'},
                                 {       'counter_source': '540/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '540/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '290235109671',
                                                                       'in_ucast_pkts': '1060773754',
                                                                       'mac_address': 'ec:4f:82:0d:6a:2a',
                                                                       'name': '540/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '62',
                                                                       'out_octets': '5986896100211',
                                                                       'out_ucast_pkts': '4348113754',
                                                                       'poller_hash': 'MjA1REVEcGVlcmludGY1NDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '540',
                                                                       'in_nucast_pkts': '62',
                                                                       'in_octets': '5986896100211',
                                                                       'in_ucast_pkts': '4348113754',
                                                                       'links': [       {       'in_nucast_pkts': '62',
                                                                                                'in_octets': '5986896100211',
                                                                                                'in_ucast_pkts': '4348113754',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '290235109671',
                                                                                                'out_ucast_pkts': '1060773754',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '290235109671',
                                                                       'out_ucast_pkts': '1060773754',
                                                                       'poller_hash': 'MjA1REVEcGVlcmludGZwb24gNTQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '540',
                                                                                               'links': [       {       'connected_time': '688413',
                                                                                                                        'fiber_length': 2132,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REVEcGVlcnBvbnBvbiA1NDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '82 Reagan St. Milverton (Craig & '
                                                 'Julie Purvis 10435983)',
                                         'poller_hash': 'MjA1REVE',
                                         'serial_number': '205DED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:2a',
                                         'uptime': '688413'},
                                 {       'counter_source': '541/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '541/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '283',
                                                                       'in_octets': '166740007380',
                                                                       'in_ucast_pkts': '1196146901',
                                                                       'mac_address': 'ec:4f:82:0d:67:51',
                                                                       'name': '541/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1916',
                                                                       'out_octets': '5787051856616',
                                                                       'out_ucast_pkts': '4120755172',
                                                                       'poller_hash': 'MjA1RDlDcGVlcmludGY1NDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '541',
                                                                       'in_nucast_pkts': '1916',
                                                                       'in_octets': '5787051856616',
                                                                       'in_ucast_pkts': '4120755172',
                                                                       'links': [       {       'in_nucast_pkts': '1916',
                                                                                                'in_octets': '5787051856616',
                                                                                                'in_ucast_pkts': '4120755172',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '283',
                                                                                                'out_octets': '166740007380',
                                                                                                'out_ucast_pkts': '1196146901',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '283',
                                                                       'out_octets': '166740007380',
                                                                       'out_ucast_pkts': '1196146901',
                                                                       'poller_hash': 'MjA1RDlDcGVlcmludGZwb24gNTQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '541',
                                                                                               'links': [       {       'connected_time': '583647',
                                                                                                                        'fiber_length': 1682,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -24.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RDlDcGVlcnBvbnBvbiA1NDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 William St Milverton (Jonathan '
                                                 'Siderius 10434882)',
                                         'poller_hash': 'MjA1RDlD',
                                         'serial_number': '205D9C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:67:51',
                                         'uptime': '583647'},
                                 {       'counter_source': '542/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '98:25:4a:de:d1:91'],
                                                                       'id': '542/g1',
                                                                       'in_nucast_pkts': '774325',
                                                                       'in_octets': '620912527263',
                                                                       'in_ucast_pkts': '4290628594',
                                                                       'mac_address': 'ec:4f:82:7e:09:5e',
                                                                       'mtu': 2000,
                                                                       'name': '542/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82666182',
                                                                       'out_octets': '17153343167003',
                                                                       'out_ucast_pkts': '12494152970',
                                                                       'poller_hash': 'MzVFOTZBcGVlcmludGY1NDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '542',
                                                                       'in_nucast_pkts': '82666182',
                                                                       'in_octets': '17153343167003',
                                                                       'in_ucast_pkts': '12494152970',
                                                                       'links': [       {       'in_nucast_pkts': '82666182',
                                                                                                'in_octets': '17153343167003',
                                                                                                'in_ucast_pkts': '12494152970',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '774325',
                                                                                                'out_octets': '620912527263',
                                                                                                'out_ucast_pkts': '4290628594',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '774325',
                                                                       'out_octets': '620912527263',
                                                                       'out_ucast_pkts': '4290628594',
                                                                       'poller_hash': 'MzVFOTZBcGVlcmludGZwb24gNTQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '542',
                                                                                               'links': [       {       'connected_time': '67790372',
                                                                                                                        'fiber_length': 1035,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTZBcGVlcnBvbnBvbiA1NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5 John St (Becky Hastings & Trevor '
                                                 'Lines)',
                                         'poller_hash': 'MzVFOTZB',
                                         'serial_number': '35E96A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:5e',
                                         'uptime': '67790372'},
                                 {       'counter_source': '543/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a0:ab:1b:7b:68:2d'],
                                                                       'id': '543/g1',
                                                                       'in_nucast_pkts': '723',
                                                                       'in_octets': '3609002311',
                                                                       'in_ucast_pkts': '25786537',
                                                                       'mac_address': 'cc:be:59:b9:c5:c6',
                                                                       'mtu': 2000,
                                                                       'name': '543/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464140',
                                                                       'out_octets': '117601205758',
                                                                       'out_ucast_pkts': '81141100',
                                                                       'poller_hash': 'NTEyNUNBcGVlcmludGY1NDMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '543',
                                                                       'in_nucast_pkts': '1464140',
                                                                       'in_octets': '117601205758',
                                                                       'in_ucast_pkts': '81141100',
                                                                       'links': [       {       'in_nucast_pkts': '1464140',
                                                                                                'in_octets': '117601205758',
                                                                                                'in_ucast_pkts': '81141100',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '723',
                                                                                                'out_octets': '3609002311',
                                                                                                'out_ucast_pkts': '25786537',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '723',
                                                                       'out_octets': '3609002311',
                                                                       'out_ucast_pkts': '25786537',
                                                                       'poller_hash': 'NTEyNUNBcGVlcmludGZwb24gNTQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '543',
                                                                                               'links': [       {       'connected_time': '688492',
                                                                                                                        'fiber_length': 1918,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUNBcGVlcnBvbnBvbiA1NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '52 Main St N Milverton (Diane '
                                                 'Schweitzer 10401082)',
                                         'poller_hash': 'NTEyNUNB',
                                         'serial_number': '5125CA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c5:c6',
                                         'uptime': '688492'},
                                 {       'counter_source': '544/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '544/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '319346884849',
                                                                       'in_ucast_pkts': '1230690026',
                                                                       'mac_address': 'ec:4f:82:0d:69:91',
                                                                       'name': '544/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '4619396932272',
                                                                       'out_ucast_pkts': '3426972138',
                                                                       'poller_hash': 'MjA1RERDcGVlcmludGY1NDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '544',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '4619396932272',
                                                                       'in_ucast_pkts': '3426972138',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '4619396932272',
                                                                                                'in_ucast_pkts': '3426972138',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '319346884849',
                                                                                                'out_ucast_pkts': '1230690026',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '319346884849',
                                                                       'out_ucast_pkts': '1230690026',
                                                                       'poller_hash': 'MjA1RERDcGVlcmludGZwb24gNTQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '544',
                                                                                               'links': [       {       'connected_time': '688487',
                                                                                                                        'fiber_length': 549,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RERDcGVlcnBvbnBvbiA1NDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '4 Main St S Apt 2 Milverton (DUNBAR '
                                                 'SARAH)',
                                         'poller_hash': 'MjA1RERD',
                                         'serial_number': '205DDC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:91',
                                         'uptime': '688487'},
                                 {       'counter_source': '545/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1b:3e'],
                                                                       'id': '545/g1',
                                                                       'in_nucast_pkts': '14605',
                                                                       'in_octets': '12782358358',
                                                                       'in_ucast_pkts': '33819403',
                                                                       'mac_address': 'cc:be:59:b9:c6:08',
                                                                       'mtu': 2000,
                                                                       'name': '545/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464198',
                                                                       'out_octets': '86937142738',
                                                                       'out_ucast_pkts': '68136931',
                                                                       'poller_hash': 'NTEyNUUwcGVlcmludGY1NDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '545',
                                                                       'in_nucast_pkts': '1464198',
                                                                       'in_octets': '86937142738',
                                                                       'in_ucast_pkts': '68136931',
                                                                       'links': [       {       'in_nucast_pkts': '1464198',
                                                                                                'in_octets': '86937142738',
                                                                                                'in_ucast_pkts': '68136931',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14605',
                                                                                                'out_octets': '12782358358',
                                                                                                'out_ucast_pkts': '33819403',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14605',
                                                                       'out_octets': '12782358358',
                                                                       'out_ucast_pkts': '33819403',
                                                                       'poller_hash': 'NTEyNUUwcGVlcmludGZwb24gNTQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '545',
                                                                                               'links': [       {       'connected_time': '688480',
                                                                                                                        'fiber_length': 1258,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUUwcGVlcnBvbnBvbiA1NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5 GEORGE ST (KRAFT STEPHANIE)',
                                         'poller_hash': 'NTEyNUUw',
                                         'serial_number': '5125E0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c6:08',
                                         'uptime': '688480'},
                                 {       'counter_source': '546/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '546/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '145959188740',
                                                                       'in_ucast_pkts': '522187134',
                                                                       'mac_address': 'cc:be:59:cb:43:5a',
                                                                       'name': '546/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '1758110118038',
                                                                       'out_ucast_pkts': '1439168946',
                                                                       'poller_hash': 'NDJCOUMzcGVlcmludGY1NDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '546',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '1758110118038',
                                                                       'in_ucast_pkts': '1439168946',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '1758110118038',
                                                                                                'in_ucast_pkts': '1439168946',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '145959188740',
                                                                                                'out_ucast_pkts': '522187134',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '145959188740',
                                                                       'out_ucast_pkts': '522187134',
                                                                       'poller_hash': 'NDJCOUMzcGVlcmludGZwb24gNTQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '546',
                                                                                               'links': [       {       'connected_time': '8925450',
                                                                                                                        'fiber_length': 9323,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUMzcGVlcnBvbnBvbiA1NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6322 Rd. 148 (Kurt Gsell)',
                                         'poller_hash': 'NDJCOUMz',
                                         'serial_number': '42B9C3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:5a',
                                         'uptime': '8925450'},
                                 {       'counter_source': '548/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '548/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '146587691812',
                                                                       'in_ucast_pkts': '737351817',
                                                                       'mac_address': 'd0:76:8f:1b:9e:83',
                                                                       'name': '548/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '174',
                                                                       'out_octets': '5030749282825',
                                                                       'out_ucast_pkts': '3462809234',
                                                                       'poller_hash': 'ODcyNDFFcGVlcmludGY1NDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '548',
                                                                       'in_nucast_pkts': '174',
                                                                       'in_octets': '5030749282825',
                                                                       'in_ucast_pkts': '3462809234',
                                                                       'links': [       {       'in_nucast_pkts': '174',
                                                                                                'in_octets': '5030749282825',
                                                                                                'in_ucast_pkts': '3462809234',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '146587691812',
                                                                                                'out_ucast_pkts': '737351817',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '146587691812',
                                                                       'out_ucast_pkts': '737351817',
                                                                       'poller_hash': 'ODcyNDFFcGVlcmludGZwb24gNTQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '548',
                                                                                               'links': [       {       'connected_time': '688490',
                                                                                                                        'fiber_length': 1351,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDFFcGVlcnBvbnBvbiA1NDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 Main St. N. Milverton (Brenda '
                                                 'Robertson)',
                                         'poller_hash': 'ODcyNDFF',
                                         'serial_number': '87241E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9e:83',
                                         'uptime': '688490'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '549/g1',
                                                                       'name': '549/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNEUxcGVlcmludGY1NDkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '549',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNEUxcGVlcmludGZwb24gNTQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '549',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNEUxcGVlcnBvbnBvbiA1NDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '17 MAIN ST N',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "549" '
                                                                  '(location: GPON '
                                                                  'port "8/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-06-28T16:49:14.360Z',
                                         'poller_hash': 'NTEyNEUx',
                                         'serial_number': '5124E1',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '550/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:79:d0:b7'],
                                                                       'id': '550/g1',
                                                                       'in_nucast_pkts': '459',
                                                                       'in_octets': '62749926320',
                                                                       'in_ucast_pkts': '466444616',
                                                                       'mac_address': 'cc:be:59:cb:44:ad',
                                                                       'mtu': 2000,
                                                                       'name': '550/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35084250',
                                                                       'out_octets': '2332622550828',
                                                                       'out_ucast_pkts': '1617325733',
                                                                       'poller_hash': 'NDJCOUVEcGVlcmludGY1NTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '550',
                                                                       'in_nucast_pkts': '35084250',
                                                                       'in_octets': '2332622550828',
                                                                       'in_ucast_pkts': '1617325733',
                                                                       'links': [       {       'in_nucast_pkts': '35084250',
                                                                                                'in_octets': '2332622550828',
                                                                                                'in_ucast_pkts': '1617325733',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '459',
                                                                                                'out_octets': '62749926320',
                                                                                                'out_ucast_pkts': '466444616',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '459',
                                                                       'out_octets': '62749926320',
                                                                       'out_ucast_pkts': '466444616',
                                                                       'poller_hash': 'NDJCOUVEcGVlcmludGZwb24gNTUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '550',
                                                                                               'links': [       {       'connected_time': '15498714',
                                                                                                                        'fiber_length': 1133,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUVEcGVlcnBvbnBvbiA1NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 Main St S Milverton (Milverton '
                                                 'Mini Mart 10420723)',
                                         'poller_hash': 'NDJCOUVE',
                                         'serial_number': '42B9ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:aa',
                                         'uptime': '15498714'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '551/g1',
                                                                       'name': '551/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNTBEcGVlcmludGY1NTEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '551',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTEyNTBEcGVlcmludGZwb24gNTUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '551',
                                                                                               'links': [       {       'connected_time': '15499142',
                                                                                                                        'fiber_length': 2089,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTBEcGVlcnBvbnBvbiA1NTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': 'Rueben Kuepfer 10431068 (85 Reagan '
                                                 'St)',
                                         'poller_hash': 'NTEyNTBE',
                                         'serial_number': '51250D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:8f',
                                         'uptime': '15499142'},
                                 {       'counter_source': '552/g4+552/g5+552/g7',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b4:b0:24:8b:b9:11'],
                                                                       'id': '552/g4',
                                                                       'in_nucast_pkts': '167748',
                                                                       'in_octets': '1409205309965',
                                                                       'in_ucast_pkts': '6354693195',
                                                                       'mac_address': 'ec:4f:82:28:e4:b0',
                                                                       'mtu': 2000,
                                                                       'name': '552/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '77887394',
                                                                       'out_octets': '42785781939607',
                                                                       'out_ucast_pkts': '30138363984',
                                                                       'poller_hash': 'MjZEOTkycGVlcmludGY1NTIvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '88:d7:f6:b8:30:24'],
                                                                       'id': '552/g5',
                                                                       'in_nucast_pkts': '294',
                                                                       'in_octets': '276816214231',
                                                                       'in_ucast_pkts': '971421241',
                                                                       'mac_address': 'ec:4f:82:b0:3f:de',
                                                                       'mtu': 2000,
                                                                       'name': '552/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '77852297',
                                                                       'out_octets': '4649745337596',
                                                                       'out_ucast_pkts': '3405641564',
                                                                       'poller_hash': 'MjZEOTkycGVlcmludGY1NTIvZzU=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f6:1b'],
                                                                       'id': '552/g7',
                                                                       'in_nucast_pkts': '1262949',
                                                                       'in_octets': '2078235459733',
                                                                       'in_ucast_pkts': '3415272418',
                                                                       'mac_address': 'ec:4f:82:b0:3f:e0',
                                                                       'mtu': 2000,
                                                                       'name': '552/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68790727',
                                                                       'out_octets': '6818625996368',
                                                                       'out_ucast_pkts': '5685119410',
                                                                       'poller_hash': 'MjZEOTkycGVlcmludGY1NTIvZzc=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '552',
                                                                       'in_nucast_pkts': '224530418',
                                                                       'in_octets': '54254153273571',
                                                                       'in_ucast_pkts': '39229124958',
                                                                       'links': [       {       'in_nucast_pkts': '224530418',
                                                                                                'in_octets': '54254153273571',
                                                                                                'in_ucast_pkts': '39229124958',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1430991',
                                                                                                'out_octets': '3764256983929',
                                                                                                'out_ucast_pkts': '10741386854',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1430991',
                                                                       'out_octets': '3764256983929',
                                                                       'out_ucast_pkts': '10741386854',
                                                                       'poller_hash': 'MjZEOTkycGVlcmludGZwb24gNTUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '552',
                                                                                               'links': [       {       'connected_time': '11454477',
                                                                                                                        'fiber_length': 1429,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjZEOTkycGVlcnBvbnBvbiA1NTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '1 LLANA HEIGHTS DR (KNIPE JACQUELYN)',
                                         'poller_hash': 'MjZEOTky',
                                         'serial_number': '26D992',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:28:e4:b2',
                                         'uptime': '11454477'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '553/g1',
                                                                       'name': '553/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNUUzcGVlcmludGY1NTMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '553',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTEyNUUzcGVlcmludGZwb24gNTUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '553',
                                                                                               'links': [       {       'connected_time': '15499058',
                                                                                                                        'fiber_length': 1082,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUUzcGVlcnBvbnBvbiA1NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '32 MAIN ST N',
                                         'poller_hash': 'NTEyNUUz',
                                         'serial_number': '5125E3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c6:11',
                                         'uptime': '15499058'},
                                 {       'counter_source': '554/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:9c:9d'],
                                                                       'id': '554/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '633369016',
                                                                       'in_ucast_pkts': '3690785',
                                                                       'mac_address': 'cc:be:59:cc:0a:b5',
                                                                       'mtu': 2000,
                                                                       'name': '554/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447152',
                                                                       'out_octets': '8534238957',
                                                                       'out_ucast_pkts': '6917792',
                                                                       'poller_hash': 'NDJEMkFFcGVlcmludGY1NTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '554',
                                                                       'in_nucast_pkts': '1447152',
                                                                       'in_octets': '8534238957',
                                                                       'in_ucast_pkts': '6917792',
                                                                       'links': [       {       'in_nucast_pkts': '1447152',
                                                                                                'in_octets': '8534238957',
                                                                                                'in_ucast_pkts': '6917792',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '633369016',
                                                                                                'out_ucast_pkts': '3690785',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '633369016',
                                                                       'out_ucast_pkts': '3690785',
                                                                       'poller_hash': 'NDJEMkFFcGVlcmludGZwb24gNTU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '554',
                                                                                               'links': [       {       'connected_time': '688408',
                                                                                                                        'fiber_length': 2148,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMkFFcGVlcnBvbnBvbiA1NTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '79 GERBER DR (DOHERTY LARRY & DONNA)',
                                         'poller_hash': 'NDJEMkFF',
                                         'serial_number': '42D2AE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:0a:b2',
                                         'uptime': '688408'},
                                 {       'counter_source': '555/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '555/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '43543160581',
                                                                       'in_ucast_pkts': '4590173464',
                                                                       'mac_address': 'cc:be:59:cb:c6:c2',
                                                                       'name': '555/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '25',
                                                                       'out_octets': '1147333703828',
                                                                       'out_ucast_pkts': '888658367',
                                                                       'poller_hash': 'NDJDQTMwcGVlcmludGY1NTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '555',
                                                                       'in_nucast_pkts': '25',
                                                                       'in_octets': '1147333703828',
                                                                       'in_ucast_pkts': '888658367',
                                                                       'links': [       {       'in_nucast_pkts': '25',
                                                                                                'in_octets': '1147333703828',
                                                                                                'in_ucast_pkts': '888658367',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '43543160581',
                                                                                                'out_ucast_pkts': '4590173464',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '43543160581',
                                                                       'out_ucast_pkts': '4590173464',
                                                                       'poller_hash': 'NDJDQTMwcGVlcmludGZwb24gNTU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '555',
                                                                                               'links': [       {       'connected_time': '688333',
                                                                                                                        'fiber_length': 1340,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDQTMwcGVlcnBvbnBvbiA1NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 MAIN ST N APT 2 MILVERTON '
                                                 '(FRIESEN HENRY)',
                                         'poller_hash': 'NDJDQTMw',
                                         'serial_number': '42CA30',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:c6:c2',
                                         'uptime': '688333'},
                                 {       'counter_source': '564/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:a9:8a:22:67:6e'],
                                                                       'id': '564/g1',
                                                                       'in_nucast_pkts': '110',
                                                                       'in_octets': '1674092268000',
                                                                       'in_ucast_pkts': '5114222919',
                                                                       'mac_address': '00:06:31:a2:c5:70',
                                                                       'mtu': 2000,
                                                                       'name': '564/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '65744501',
                                                                       'out_octets': '16775109092921',
                                                                       'out_ucast_pkts': '12962196692',
                                                                       'poller_hash': 'MTFEMDE4cGVlcmludGY1NjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '564',
                                                                       'in_nucast_pkts': '65744501',
                                                                       'in_octets': '16775109092921',
                                                                       'in_ucast_pkts': '12962196692',
                                                                       'links': [       {       'in_nucast_pkts': '65744501',
                                                                                                'in_octets': '16775109092921',
                                                                                                'in_ucast_pkts': '12962196692',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '110',
                                                                                                'out_octets': '1674092268000',
                                                                                                'out_ucast_pkts': '5114222919',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '110',
                                                                       'out_octets': '1674092268000',
                                                                       'out_ucast_pkts': '5114222919',
                                                                       'poller_hash': 'MTFEMDE4cGVlcmludGZwb24gNTY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '564',
                                                                                               'links': [       {       'connected_time': '67927947',
                                                                                                                        'fiber_length': 10254,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.798,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFEMDE4cGVlcnBvbnBvbiA1NjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': 'Peaceful Meadows Retirement (Office '
                                                 'ONT)',
                                         'poller_hash': 'MTFEMDE4',
                                         'serial_number': '11D018',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:8b:c0:d8',
                                         'uptime': '67927947'},
                                 {       'counter_source': '565/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:8e:57'],
                                                                       'id': '565/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '1308636043',
                                                                       'in_ucast_pkts': '6473285',
                                                                       'mac_address': 'cc:be:59:cb:95:b5',
                                                                       'mtu': 2000,
                                                                       'name': '565/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1461871',
                                                                       'out_octets': '28120988098',
                                                                       'out_ucast_pkts': '21664604',
                                                                       'poller_hash': 'NDJDNDBFcGVlcmludGY1NjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '565',
                                                                       'in_nucast_pkts': '1461871',
                                                                       'in_octets': '28120988098',
                                                                       'in_ucast_pkts': '21664604',
                                                                       'links': [       {       'in_nucast_pkts': '1461871',
                                                                                                'in_octets': '28120988098',
                                                                                                'in_ucast_pkts': '21664604',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '1308636043',
                                                                                                'out_ucast_pkts': '6473285',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '1308636043',
                                                                       'out_ucast_pkts': '6473285',
                                                                       'poller_hash': 'NDJDNDBFcGVlcmludGZwb24gNTY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '565',
                                                                                               'links': [       {       'connected_time': '688137',
                                                                                                                        'fiber_length': 1041,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNDBFcGVlcnBvbnBvbiA1NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '36 Gerber Ct Milverton (Norman Bell '
                                                 '10435005)',
                                         'poller_hash': 'NDJDNDBF',
                                         'serial_number': '42C40E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:95:b2',
                                         'uptime': '688137'},
                                 {       'counter_source': '567/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:79:de:01'],
                                                                       'id': '567/g1',
                                                                       'in_nucast_pkts': '160',
                                                                       'in_octets': '637354174841',
                                                                       'in_ucast_pkts': '3741039208',
                                                                       'mac_address': 'cc:be:59:70:d9:56',
                                                                       'mtu': 2000,
                                                                       'name': '567/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '65719894',
                                                                       'out_octets': '14404522781705',
                                                                       'out_ucast_pkts': '10750397810',
                                                                       'poller_hash': 'NTYxNzFDcGVlcmludGY1NjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '567',
                                                                       'in_nucast_pkts': '65719894',
                                                                       'in_octets': '14404522781705',
                                                                       'in_ucast_pkts': '10750397810',
                                                                       'links': [       {       'in_nucast_pkts': '65719894',
                                                                                                'in_octets': '14404522781705',
                                                                                                'in_ucast_pkts': '10750397810',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160',
                                                                                                'out_octets': '637354174841',
                                                                                                'out_ucast_pkts': '3741039208',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160',
                                                                       'out_octets': '637354174841',
                                                                       'out_ucast_pkts': '3741039208',
                                                                       'poller_hash': 'NTYxNzFDcGVlcmludGZwb24gNTY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '567',
                                                                                               'links': [       {       'connected_time': '68237769',
                                                                                                                        'fiber_length': 2120,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.908,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYxNzFDcGVlcnBvbnBvbiA1Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '63 Reagan St Milverton (Megan Hines '
                                                 '10435292)',
                                         'poller_hash': 'NTYxNzFD',
                                         'serial_number': '56171C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:70:d9:52',
                                         'uptime': '68237769'},
                                 {       'counter_source': '568/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:8d:28:55:cd:f1'],
                                                                       'id': '568/g1',
                                                                       'in_nucast_pkts': '520',
                                                                       'in_octets': '370146386',
                                                                       'in_ucast_pkts': '1927748',
                                                                       'mac_address': 'cc:be:59:b9:c4:ca',
                                                                       'mtu': 2000,
                                                                       'name': '568/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464074',
                                                                       'out_octets': '4018493629',
                                                                       'out_ucast_pkts': '3494111',
                                                                       'poller_hash': 'NTEyNTc2cGVlcmludGY1NjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '568',
                                                                       'in_nucast_pkts': '1464074',
                                                                       'in_octets': '4018493629',
                                                                       'in_ucast_pkts': '3494111',
                                                                       'links': [       {       'in_nucast_pkts': '1464074',
                                                                                                'in_octets': '4018493629',
                                                                                                'in_ucast_pkts': '3494111',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '520',
                                                                                                'out_octets': '370146386',
                                                                                                'out_ucast_pkts': '1927748',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '520',
                                                                       'out_octets': '370146386',
                                                                       'out_ucast_pkts': '1927748',
                                                                       'poller_hash': 'NTEyNTc2cGVlcmludGZwb24gNTY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '568',
                                                                                               'links': [       {       'connected_time': '688500',
                                                                                                                        'fiber_length': 1259,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTc2cGVlcnBvbnBvbiA1Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '89 Main St N Milverton (Doug & Anna '
                                                 'Hodgins 10403089)',
                                         'poller_hash': 'NTEyNTc2',
                                         'serial_number': '512576',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c4:ca',
                                         'uptime': '688500'},
                                 {       'counter_source': '569/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e7:e3'],
                                                                       'id': '569/g1',
                                                                       'in_nucast_pkts': '633602',
                                                                       'in_octets': '286672742607',
                                                                       'in_ucast_pkts': '1901269865',
                                                                       'mac_address': 'cc:be:59:cb:ff:e5',
                                                                       'mtu': 2000,
                                                                       'name': '569/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '71111847',
                                                                       'out_octets': '8518861287690',
                                                                       'out_ucast_pkts': '5911051380',
                                                                       'poller_hash': 'NDJEMTU0cGVlcmludGY1NjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '569',
                                                                       'in_nucast_pkts': '71111847',
                                                                       'in_octets': '8518861287690',
                                                                       'in_ucast_pkts': '5911051380',
                                                                       'links': [       {       'in_nucast_pkts': '71111847',
                                                                                                'in_octets': '8518861287690',
                                                                                                'in_ucast_pkts': '5911051380',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '633602',
                                                                                                'out_octets': '286672742607',
                                                                                                'out_ucast_pkts': '1901269865',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '633602',
                                                                       'out_octets': '286672742607',
                                                                       'out_ucast_pkts': '1901269865',
                                                                       'poller_hash': 'NDJEMTU0cGVlcmludGZwb24gNTY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '569',
                                                                                               'links': [       {       'connected_time': '65348135',
                                                                                                                        'fiber_length': 17984,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMTU0cGVlcnBvbnBvbiA1Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4535 Line 49 (Helmut&Liz Mergl '
                                                 '10400543)',
                                         'poller_hash': 'NDJEMTU0',
                                         'serial_number': '42D154',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ff:e2',
                                         'uptime': '65348135'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '570/g1',
                                                                       'name': '570/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNUU0cGVlcmludGY1NzAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '570',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTEyNUU0cGVlcmludGZwb24gNTcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '570',
                                                                                               'links': [       {       'connected_time': '538561',
                                                                                                                        'fiber_length': 17257,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUU0cGVlcnBvbnBvbiA1NzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': "5183 ROAD 130 (ST PAUL'S LUTHERAN "
                                                 'CHURCH)',
                                         'poller_hash': 'NTEyNUU0',
                                         'serial_number': '5125E4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c6:14',
                                         'uptime': '538561'},
                                 {       'counter_source': '571/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:97'],
                                                                       'id': '571/g1',
                                                                       'in_nucast_pkts': '78',
                                                                       'in_octets': '1516029525',
                                                                       'in_ucast_pkts': '6066078',
                                                                       'mac_address': 'cc:be:59:b9:c3:d1',
                                                                       'mtu': 2000,
                                                                       'name': '571/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1660372',
                                                                       'out_octets': '26355541883',
                                                                       'out_ucast_pkts': '21119750',
                                                                       'poller_hash': 'NTEyNTIzcGVlcmludGY1NzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '571',
                                                                       'in_nucast_pkts': '1660372',
                                                                       'in_octets': '26355541883',
                                                                       'in_ucast_pkts': '21119750',
                                                                       'links': [       {       'in_nucast_pkts': '1660372',
                                                                                                'in_octets': '26355541883',
                                                                                                'in_ucast_pkts': '21119750',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '78',
                                                                                                'out_octets': '1516029525',
                                                                                                'out_ucast_pkts': '6066078',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '78',
                                                                       'out_octets': '1516029525',
                                                                       'out_ucast_pkts': '6066078',
                                                                       'poller_hash': 'NTEyNTIzcGVlcmludGZwb24gNTcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '571',
                                                                                               'links': [       {       'connected_time': '801523',
                                                                                                                        'fiber_length': 18224,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTIzcGVlcnBvbnBvbiA1NzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4538 Line 49 (Samuel Ebersol '
                                                 '10430136)',
                                         'poller_hash': 'NTEyNTIz',
                                         'serial_number': '512523',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:d1',
                                         'uptime': '801523'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '572/g1',
                                                                       'name': '572/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNUYwcGVlcmludGY1NzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '572',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTEyNUYwcGVlcmludGZwb24gNTcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '572',
                                                                                               'links': [       {       'connected_time': '67790067',
                                                                                                                        'fiber_length': 18125,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUYwcGVlcnBvbnBvbiA1NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4514 Line 49 (Isaak Janzen 10429928)',
                                         'poller_hash': 'NTEyNUYw',
                                         'serial_number': '5125F0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c6:38',
                                         'uptime': '67790067'},
                                 {       'counter_source': '573/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '573/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '229627382686',
                                                                       'in_ucast_pkts': '674914488',
                                                                       'mac_address': '00:06:31:a6:67:3b',
                                                                       'name': '573/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '941148937318',
                                                                       'out_ucast_pkts': '1143285096',
                                                                       'poller_hash': 'MTMwMEM5cGVlcmludGY1NzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '573',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '941148937318',
                                                                       'in_ucast_pkts': '1143285096',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '941148937318',
                                                                                                'in_ucast_pkts': '1143285096',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '229627382686',
                                                                                                'out_ucast_pkts': '674914488',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '229627382686',
                                                                       'out_ucast_pkts': '674914488',
                                                                       'poller_hash': 'MTMwMEM5cGVlcmludGZwb24gNTcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '573',
                                                                                               'links': [       {       'connected_time': '47655',
                                                                                                                        'fiber_length': 468,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMEM5cGVlcnBvbnBvbiA1NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4 Main St. S. Unit 1 Milverton '
                                                 '(Huron view Insurance 10411009)',
                                         'poller_hash': 'MTMwMEM5',
                                         'serial_number': '1300C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:89',
                                         'uptime': '47655'},
                                 {       'counter_source': '574/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1c:eb'],
                                                                       'id': '574/g1',
                                                                       'in_nucast_pkts': '54011',
                                                                       'in_octets': '23194716933',
                                                                       'in_ucast_pkts': '109200478',
                                                                       'mac_address': 'd0:76:8f:08:19:76',
                                                                       'mtu': 2000,
                                                                       'name': '574/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6782021',
                                                                       'out_octets': '398968497561',
                                                                       'out_ucast_pkts': '314532842',
                                                                       'poller_hash': 'ODY5NDI0cGVlcmludGY1NzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '574',
                                                                       'in_nucast_pkts': '6782021',
                                                                       'in_octets': '398968497561',
                                                                       'in_ucast_pkts': '314532842',
                                                                       'links': [       {       'in_nucast_pkts': '6782021',
                                                                                                'in_octets': '398968497561',
                                                                                                'in_ucast_pkts': '314532842',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '54011',
                                                                                                'out_octets': '23194716933',
                                                                                                'out_ucast_pkts': '109200478',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '54011',
                                                                       'out_octets': '23194716933',
                                                                       'out_ucast_pkts': '109200478',
                                                                       'poller_hash': 'ODY5NDI0cGVlcmludGZwb24gNTc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '574',
                                                                                               'links': [       {       'connected_time': '3113880',
                                                                                                                        'fiber_length': 17483,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDI0cGVlcnBvbnBvbiA1NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '184 SELBY AV (BOYD DON)',
                                         'poller_hash': 'ODY5NDI0',
                                         'serial_number': '869424',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:73',
                                         'uptime': '3113880'},
                                 {       'counter_source': '576/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d2:28'],
                                                                       'id': '576/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '5479311255',
                                                                       'in_ucast_pkts': '20582337',
                                                                       'mac_address': '44:65:7f:7b:c3:0e',
                                                                       'mtu': 2000,
                                                                       'name': '576/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434172',
                                                                       'out_octets': '77701517796',
                                                                       'out_ucast_pkts': '61106168',
                                                                       'poller_hash': 'NUE2NUY1cGVlcmludGY1NzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '576',
                                                                       'in_nucast_pkts': '1434172',
                                                                       'in_octets': '77701517796',
                                                                       'in_ucast_pkts': '61106168',
                                                                       'links': [       {       'in_nucast_pkts': '1434172',
                                                                                                'in_octets': '77701517796',
                                                                                                'in_ucast_pkts': '61106168',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '5479311255',
                                                                                                'out_ucast_pkts': '20582337',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '5479311255',
                                                                       'out_ucast_pkts': '20582337',
                                                                       'poller_hash': 'NUE2NUY1cGVlcmludGZwb24gNTc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '576',
                                                                                               'links': [       {       'connected_time': '688556',
                                                                                                                        'fiber_length': 1743,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NUY1cGVlcnBvbnBvbiA1NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1 Murray Ct (Dawn Griffin 10435906)',
                                         'poller_hash': 'NUE2NUY1',
                                         'serial_number': '5A65F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:0e',
                                         'uptime': '688556'},
                                 {       'counter_source': '578/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:58'],
                                                                       'id': '578/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '3677025873',
                                                                       'in_ucast_pkts': '11693748',
                                                                       'mac_address': 'cc:be:59:cb:e6:d5',
                                                                       'mtu': 2000,
                                                                       'name': '578/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120545',
                                                                       'out_octets': '39455457522',
                                                                       'out_ucast_pkts': '29691334',
                                                                       'poller_hash': 'NDJDRTMycGVlcmludGY1NzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '578',
                                                                       'in_nucast_pkts': '1120545',
                                                                       'in_octets': '39455457522',
                                                                       'in_ucast_pkts': '29691334',
                                                                       'links': [       {       'in_nucast_pkts': '1120545',
                                                                                                'in_octets': '39455457522',
                                                                                                'in_ucast_pkts': '29691334',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '3677025873',
                                                                                                'out_ucast_pkts': '11693748',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '3677025873',
                                                                       'out_ucast_pkts': '11693748',
                                                                       'poller_hash': 'NDJDRTMycGVlcmludGZwb24gNTc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '578',
                                                                                               'links': [       {       'connected_time': '538560',
                                                                                                                        'fiber_length': 16752,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTMycGVlcnBvbnBvbiA1Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4661 LINE 49 (ROES JASON & ANGELA)',
                                         'poller_hash': 'NDJDRTMy',
                                         'serial_number': '42CE32',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e6:d2',
                                         'uptime': '538560'},
                                 {       'counter_source': '579/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '579/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3040840369943',
                                                                       'in_ucast_pkts': '11438225736',
                                                                       'mac_address': '00:06:31:52:64:1e',
                                                                       'name': '579/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '47154507820415',
                                                                       'out_ucast_pkts': '35411164554',
                                                                       'poller_hash': 'ODg5RkRwZWVyaW50ZjU3OS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '579',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '47154507820415',
                                                                       'in_ucast_pkts': '35411164554',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '47154507820415',
                                                                                                'in_ucast_pkts': '35411164554',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3040840369943',
                                                                                                'out_ucast_pkts': '11438225736',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3040840369943',
                                                                       'out_ucast_pkts': '11438225736',
                                                                       'poller_hash': 'ODg5RkRwZWVyaW50ZnBvbiA1Nzk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '579',
                                                                                               'links': [       {       'connected_time': '13161611',
                                                                                                                        'fiber_length': 1170,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RkRwZWVycG9ucG9uIDU3OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '12 JO-LENE CT (WOODTHORPE STEPHEN)',
                                         'poller_hash': 'ODg5RkQ=',
                                         'serial_number': '889FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:e5',
                                         'uptime': '13161611'},
                                 {       'counter_source': '580/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b8:cb'],
                                                                       'id': '580/g1',
                                                                       'in_nucast_pkts': '14863',
                                                                       'in_octets': '7622399207',
                                                                       'in_ucast_pkts': '38039469',
                                                                       'mac_address': '84:d3:43:7e:31:ea',
                                                                       'mtu': 2000,
                                                                       'name': '580/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463814',
                                                                       'out_octets': '161235625604',
                                                                       'out_ucast_pkts': '119630717',
                                                                       'poller_hash': 'QzgzQkE2cGVlcmludGY1ODAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '580',
                                                                       'in_nucast_pkts': '1463814',
                                                                       'in_octets': '161235625604',
                                                                       'in_ucast_pkts': '119630717',
                                                                       'links': [       {       'in_nucast_pkts': '1463814',
                                                                                                'in_octets': '161235625604',
                                                                                                'in_ucast_pkts': '119630717',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14863',
                                                                                                'out_octets': '7622399207',
                                                                                                'out_ucast_pkts': '38039469',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14863',
                                                                       'out_octets': '7622399207',
                                                                       'out_ucast_pkts': '38039469',
                                                                       'poller_hash': 'QzgzQkE2cGVlcmludGZwb24gNTgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '580',
                                                                                               'links': [       {       'connected_time': '688412',
                                                                                                                        'fiber_length': 1270,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkE2cGVlcnBvbnBvbiA1ODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 Jo-Lene Ct. Milverton (Doug '
                                                 'Biesinger 10402290)',
                                         'poller_hash': 'QzgzQkE2',
                                         'serial_number': 'C83BA6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:e7',
                                         'uptime': '688412'},
                                 {       'counter_source': '582/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '6c:b0:ce:e2:f6:4a'],
                                                                       'id': '582/g1',
                                                                       'in_nucast_pkts': '223',
                                                                       'in_octets': '14957500527',
                                                                       'in_ucast_pkts': '21818285',
                                                                       'mac_address': '00:06:31:a6:6a:5f',
                                                                       'mtu': 2000,
                                                                       'name': '582/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119744',
                                                                       'out_octets': '39695966139',
                                                                       'out_ucast_pkts': '37386254',
                                                                       'poller_hash': 'MTJGRkQ5cGVlcmludGY1ODIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '582',
                                                                       'in_nucast_pkts': '1119744',
                                                                       'in_octets': '39695966139',
                                                                       'in_ucast_pkts': '37386254',
                                                                       'links': [       {       'in_nucast_pkts': '1119744',
                                                                                                'in_octets': '39695966139',
                                                                                                'in_ucast_pkts': '37386254',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '223',
                                                                                                'out_octets': '14957500527',
                                                                                                'out_ucast_pkts': '21818285',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '223',
                                                                       'out_octets': '14957500527',
                                                                       'out_ucast_pkts': '21818285',
                                                                       'poller_hash': 'MTJGRkQ5cGVlcmludGZwb24gNTgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '582',
                                                                                               'links': [       {       'connected_time': '688402',
                                                                                                                        'fiber_length': 842,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTJGRkQ5cGVlcnBvbnBvbiA1ODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '12 GERBER DR (HUBER ERNA & KEVIN)',
                                         'poller_hash': 'MTJGRkQ5',
                                         'serial_number': '12FFD9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:99',
                                         'uptime': '688402'},
                                 {       'counter_source': '583/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d6:74'],
                                                                       'id': '583/g1',
                                                                       'in_nucast_pkts': '12005',
                                                                       'in_octets': '25134249743',
                                                                       'in_ucast_pkts': '47751907',
                                                                       'mac_address': '60:db:98:4e:95:cf',
                                                                       'mtu': 2000,
                                                                       'name': '583/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463884',
                                                                       'out_octets': '170798791136',
                                                                       'out_ucast_pkts': '132394510',
                                                                       'poller_hash': 'QTlDODc5cGVlcmludGY1ODMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '583',
                                                                       'in_nucast_pkts': '1463884',
                                                                       'in_octets': '170798791136',
                                                                       'in_ucast_pkts': '132394510',
                                                                       'links': [       {       'in_nucast_pkts': '1463884',
                                                                                                'in_octets': '170798791136',
                                                                                                'in_ucast_pkts': '132394510',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12005',
                                                                                                'out_octets': '25134249743',
                                                                                                'out_ucast_pkts': '47751907',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12005',
                                                                       'out_octets': '25134249743',
                                                                       'out_ucast_pkts': '47751907',
                                                                       'poller_hash': 'QTlDODc5cGVlcmludGZwb24gNTgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '583',
                                                                                               'links': [       {       'connected_time': '688413',
                                                                                                                        'fiber_length': 984,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDODc5cGVlcnBvbnBvbiA1ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 Gerber Dr. Milverton (LUDWIG '
                                                 'RYAN)',
                                         'poller_hash': 'QTlDODc5',
                                         'serial_number': 'A9C879',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:95:cc',
                                         'uptime': '688413'},
                                 {       'counter_source': '584/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5c:16'],
                                                                       'id': '584/g1',
                                                                       'in_nucast_pkts': '36831671',
                                                                       'in_octets': '1797081955845',
                                                                       'in_ucast_pkts': '4649529023',
                                                                       'mac_address': '00:06:31:52:65:62',
                                                                       'mtu': 2000,
                                                                       'name': '584/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83353781129',
                                                                       'out_octets': '133927933802718',
                                                                       'out_ucast_pkts': '15265466363',
                                                                       'poller_hash': 'ODg5RTJwZWVyaW50ZjU4NC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '584',
                                                                       'in_nucast_pkts': '83353781129',
                                                                       'in_octets': '133927933802718',
                                                                       'in_ucast_pkts': '15265466363',
                                                                       'links': [       {       'in_nucast_pkts': '83353781129',
                                                                                                'in_octets': '133927933802718',
                                                                                                'in_ucast_pkts': '15265466363',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '36831671',
                                                                                                'out_octets': '1797081955845',
                                                                                                'out_ucast_pkts': '4649529023',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36831671',
                                                                       'out_octets': '1797081955845',
                                                                       'out_ucast_pkts': '4649529023',
                                                                       'poller_hash': 'ODg5RTJwZWVyaW50ZnBvbiA1ODQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '584',
                                                                                               'links': [       {       'connected_time': '40894',
                                                                                                                        'fiber_length': 1004,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RTJwZWVycG9ucG9uIDU4NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '36 GERBER DR (PARKER JIM & SHEILA)',
                                         'poller_hash': 'ODg5RTI=',
                                         'serial_number': '889E2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:ca',
                                         'uptime': '40894'},
                                 {       'counter_source': '585/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:82:08:0c'],
                                                                       'id': '585/g1',
                                                                       'in_nucast_pkts': '740708',
                                                                       'in_octets': '183827674100',
                                                                       'in_ucast_pkts': '1661128349',
                                                                       'mac_address': '00:06:31:52:5a:0a',
                                                                       'mtu': 2000,
                                                                       'name': '585/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75613326',
                                                                       'out_octets': '12871294173730',
                                                                       'out_ucast_pkts': '8641882702',
                                                                       'poller_hash': 'ODg5QkRwZWVyaW50ZjU4NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '585',
                                                                       'in_nucast_pkts': '75613326',
                                                                       'in_octets': '12871294173730',
                                                                       'in_ucast_pkts': '8641882702',
                                                                       'links': [       {       'in_nucast_pkts': '75613326',
                                                                                                'in_octets': '12871294173730',
                                                                                                'in_ucast_pkts': '8641882702',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '740708',
                                                                                                'out_octets': '183827674100',
                                                                                                'out_ucast_pkts': '1661128349',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '740708',
                                                                       'out_octets': '183827674100',
                                                                       'out_ucast_pkts': '1661128349',
                                                                       'poller_hash': 'ODg5QkRwZWVyaW50ZnBvbiA1ODU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '585',
                                                                                               'links': [       {       'connected_time': '40906',
                                                                                                                        'fiber_length': 902,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5QkRwZWVycG9ucG9uIDU4NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '24 Gerber Dr. Milverton (Mark '
                                                 'Matheson)',
                                         'poller_hash': 'ODg5QkQ=',
                                         'serial_number': '889BD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:a5',
                                         'uptime': '40906'},
                                 {       'counter_source': '586/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:79:59:c8:1f:de'],
                                                                       'id': '586/g1',
                                                                       'in_nucast_pkts': '3381',
                                                                       'in_octets': '78851103318',
                                                                       'in_ucast_pkts': '459222108',
                                                                       'mac_address': '00:06:31:52:71:82',
                                                                       'mtu': 2000,
                                                                       'name': '586/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27324140',
                                                                       'out_octets': '3549923226663',
                                                                       'out_ucast_pkts': '2509052173',
                                                                       'poller_hash': 'OEEwNzZwZWVyaW50ZjU4Ni9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '586',
                                                                       'in_nucast_pkts': '27324140',
                                                                       'in_octets': '3549923226663',
                                                                       'in_ucast_pkts': '2509052173',
                                                                       'links': [       {       'in_nucast_pkts': '27324140',
                                                                                                'in_octets': '3549923226663',
                                                                                                'in_ucast_pkts': '2509052173',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3381',
                                                                                                'out_octets': '78851103318',
                                                                                                'out_ucast_pkts': '459222108',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3381',
                                                                       'out_octets': '78851103318',
                                                                       'out_ucast_pkts': '459222108',
                                                                       'poller_hash': 'OEEwNzZwZWVyaW50ZnBvbiA1ODY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '586',
                                                                                               'links': [       {       'connected_time': '15498464',
                                                                                                                        'fiber_length': 870,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEEwNzZwZWVycG9ucG9uIDU4Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '16 Gerber Dr. Milverton (Peter '
                                                 'Flanagan)',
                                         'poller_hash': 'OEEwNzY=',
                                         'serial_number': '8A076',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:47:49:62',
                                         'uptime': '15498464'},
                                 {       'counter_source': '587/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e5:5e'],
                                                                       'id': '587/g1',
                                                                       'in_nucast_pkts': '56922',
                                                                       'in_octets': '61488454347',
                                                                       'in_ucast_pkts': '145998276',
                                                                       'mac_address': '00:06:31:e9:01:ae',
                                                                       'mtu': 2000,
                                                                       'name': '587/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6776927',
                                                                       'out_octets': '378159229707',
                                                                       'out_ucast_pkts': '308446337',
                                                                       'poller_hash': 'MUQwNjhEcGVlcmludGY1ODcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '587',
                                                                       'in_nucast_pkts': '6776927',
                                                                       'in_octets': '378159229707',
                                                                       'in_ucast_pkts': '308446337',
                                                                       'links': [       {       'in_nucast_pkts': '6776927',
                                                                                                'in_octets': '378159229707',
                                                                                                'in_ucast_pkts': '308446337',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '56922',
                                                                                                'out_octets': '61488454347',
                                                                                                'out_ucast_pkts': '145998276',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '56922',
                                                                       'out_octets': '61488454347',
                                                                       'out_ucast_pkts': '145998276',
                                                                       'poller_hash': 'MUQwNjhEcGVlcmludGZwb24gNTg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '587',
                                                                                               'links': [       {       'connected_time': '3113515',
                                                                                                                        'fiber_length': 1112,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjhEcGVlcnBvbnBvbiA1ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '44 Gerber Dr. Milverton (Ryan '
                                                 'Harloff)',
                                         'poller_hash': 'MUQwNjhE',
                                         'serial_number': '1D068D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:ab',
                                         'uptime': '3113515'},
                                 {       'counter_source': '588/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '588/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1338463716881',
                                                                       'in_ucast_pkts': '8009942926',
                                                                       'mac_address': '00:06:31:52:5a:16',
                                                                       'name': '588/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '40782014297710',
                                                                       'out_ucast_pkts': '29566141784',
                                                                       'poller_hash': 'ODg5NEZwZWVyaW50ZjU4OC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '588',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '40782014297710',
                                                                       'in_ucast_pkts': '29566141784',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '40782014297710',
                                                                                                'in_ucast_pkts': '29566141784',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1338463716881',
                                                                                                'out_ucast_pkts': '8009942926',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1338463716881',
                                                                       'out_ucast_pkts': '8009942926',
                                                                       'poller_hash': 'ODg5NEZwZWVyaW50ZnBvbiA1ODg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '588',
                                                                                               'links': [       {       'connected_time': '19446364',
                                                                                                                        'fiber_length': 1073,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5NEZwZWVycG9ucG9uIDU4OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '48 GERBER DR (BAILLIE TERRY)',
                                         'poller_hash': 'ODg5NEY=',
                                         'serial_number': '8894F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:37',
                                         'uptime': '19446364'},
                                 {       'counter_source': '589/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '589/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '313452410460',
                                                                       'in_ucast_pkts': '2590526640',
                                                                       'mac_address': '00:06:31:51:61:10',
                                                                       'name': '589/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '17352177141514',
                                                                       'out_ucast_pkts': '11801361485',
                                                                       'poller_hash': 'OEJEQzhwZWVyaW50ZjU4OS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '589',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '17352177141514',
                                                                       'in_ucast_pkts': '11801361485',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '17352177141514',
                                                                                                'in_ucast_pkts': '11801361485',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '313452410460',
                                                                                                'out_ucast_pkts': '2590526640',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '313452410460',
                                                                       'out_ucast_pkts': '2590526640',
                                                                       'poller_hash': 'OEJEQzhwZWVyaW50ZnBvbiA1ODk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '589',
                                                                                               'links': [       {       'connected_time': '688403',
                                                                                                                        'fiber_length': 830,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEJEQzhwZWVycG9ucG9uIDU4OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '3 GERBER DR (SCHMIDT CALVIN)',
                                         'poller_hash': 'OEJEQzg=',
                                         'serial_number': '8BDC8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:47:66:b4',
                                         'uptime': '688403'},
                                 {       'counter_source': '590/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '590/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '336004994713',
                                                                       'in_ucast_pkts': '1071341243',
                                                                       'mac_address': '00:06:31:52:6f:96',
                                                                       'name': '590/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4275298476848',
                                                                       'out_ucast_pkts': '3250646993',
                                                                       'poller_hash': 'OEEwNkRwZWVyaW50ZjU5MC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '590',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4275298476848',
                                                                       'in_ucast_pkts': '3250646993',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4275298476848',
                                                                                                'in_ucast_pkts': '3250646993',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '336004994713',
                                                                                                'out_ucast_pkts': '1071341243',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '336004994713',
                                                                       'out_ucast_pkts': '1071341243',
                                                                       'poller_hash': 'OEEwNkRwZWVyaW50ZnBvbiA1OTA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '590',
                                                                                               'links': [       {       'connected_time': '688408',
                                                                                                                        'fiber_length': 809,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEEwNkRwZWVycG9ucG9uIDU5MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4 Gerber Dr. Milverton (Sean '
                                                 'McCurdy)',
                                         'poller_hash': 'OEEwNkQ=',
                                         'serial_number': '8A06D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:47:49:59',
                                         'uptime': '688408'},
                                 {       'counter_source': '591/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '591/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '68154146783',
                                                                       'in_ucast_pkts': '562116518',
                                                                       'mac_address': '00:06:31:52:65:e6',
                                                                       'name': '591/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2804632842739',
                                                                       'out_ucast_pkts': '1991471643',
                                                                       'poller_hash': 'ODhBMUVwZWVyaW50ZjU5MS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '591',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2804632842739',
                                                                       'in_ucast_pkts': '1991471643',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2804632842739',
                                                                                                'in_ucast_pkts': '1991471643',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '68154146783',
                                                                                                'out_ucast_pkts': '562116518',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '68154146783',
                                                                       'out_ucast_pkts': '562116518',
                                                                       'poller_hash': 'ODhBMUVwZWVyaW50ZnBvbiA1OTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '591',
                                                                                               'links': [       {       'connected_time': '18858052',
                                                                                                                        'fiber_length': 1097,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODhBMUVwZWVycG9ucG9uIDU5MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '8 Colbalt St. S. Milverton (Elroy '
                                                 'Horst)',
                                         'poller_hash': 'ODhBMUU=',
                                         'serial_number': '88A1E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:eb:06',
                                         'uptime': '18858052'},
                                 {       'counter_source': '592/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '592/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '276219861792',
                                                                       'in_ucast_pkts': '1518767293',
                                                                       'mac_address': '00:06:31:52:62:86',
                                                                       'name': '592/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9308621815590',
                                                                       'out_ucast_pkts': '6358465124',
                                                                       'poller_hash': 'ODg5QTZwZWVyaW50ZjU5Mi9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '592',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9308621815590',
                                                                       'in_ucast_pkts': '6358465124',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9308621815590',
                                                                                                'in_ucast_pkts': '6358465124',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '276219861792',
                                                                                                'out_ucast_pkts': '1518767293',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '276219861792',
                                                                       'out_ucast_pkts': '1518767293',
                                                                       'poller_hash': 'ODg5QTZwZWVyaW50ZnBvbiA1OTI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '592',
                                                                                               'links': [       {       'connected_time': '688435',
                                                                                                                        'fiber_length': 1133,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5QTZwZWVycG9ucG9uIDU5Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '47 Gerber Dr. Milverton (Ron Hanna)',
                                         'poller_hash': 'ODg5QTY=',
                                         'serial_number': '889A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:8e',
                                         'uptime': '688435'},
                                 {       'counter_source': '593/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '593/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '150343919448',
                                                                       'in_ucast_pkts': '1243560258',
                                                                       'mac_address': '00:06:31:52:58:1e',
                                                                       'name': '593/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5783577860133',
                                                                       'out_ucast_pkts': '4088959213',
                                                                       'poller_hash': 'ODg5RjFwZWVyaW50ZjU5My9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '593',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5783577860133',
                                                                       'in_ucast_pkts': '4088959213',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '5783577860133',
                                                                                                'in_ucast_pkts': '4088959213',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '150343919448',
                                                                                                'out_ucast_pkts': '1243560258',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '150343919448',
                                                                       'out_ucast_pkts': '1243560258',
                                                                       'poller_hash': 'ODg5RjFwZWVyaW50ZnBvbiA1OTM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '593',
                                                                                               'links': [       {       'connected_time': '13219156',
                                                                                                                        'fiber_length': 1099,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RjFwZWVycG9ucG9uIDU5Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '52 Gerber Dr. Milverton (Urius '
                                                 'Bowman)',
                                         'poller_hash': 'ODg5RjE=',
                                         'serial_number': '889F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:d9',
                                         'uptime': '13219156'},
                                 {       'counter_source': '594/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '594/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '277428311559',
                                                                       'in_ucast_pkts': '3076048934',
                                                                       'mac_address': '00:06:31:52:58:4e',
                                                                       'name': '594/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9293654946108',
                                                                       'out_ucast_pkts': '2095521715',
                                                                       'poller_hash': 'ODg5RUZwZWVyaW50ZjU5NC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '594',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9293654946108',
                                                                       'in_ucast_pkts': '2095521715',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9293654946108',
                                                                                                'in_ucast_pkts': '2095521715',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '277428311559',
                                                                                                'out_ucast_pkts': '3076048934',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '277428311559',
                                                                       'out_ucast_pkts': '3076048934',
                                                                       'poller_hash': 'ODg5RUZwZWVyaW50ZnBvbiA1OTQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '594',
                                                                                               'links': [       {       'connected_time': '47398',
                                                                                                                        'fiber_length': 1030,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RUZwZWVycG9ucG9uIDU5NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '28 Gerber Dr. Milverton (Norman '
                                                 'Steckly)',
                                         'poller_hash': 'ODg5RUY=',
                                         'serial_number': '889EF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:d7',
                                         'uptime': '47398'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '595/g1',
                                                                       'name': '595/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzRGRkNwZWVyaW50ZjU5NS9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '595',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzRGRkNwZWVyaW50ZnBvbiA1OTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '595',
                                                                                               'links': [       {       'connected_time': '41295',
                                                                                                                        'fiber_length': 1138,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGRkNwZWVycG9ucG9uIDU5NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '39 Gerber Dr. Milverton (Allan '
                                                 'Steckley)',
                                         'poller_hash': 'QzRGRkM=',
                                         'serial_number': 'C4FFC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:bc',
                                         'uptime': '41295'},
                                 {       'counter_source': '596/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d5:13'],
                                                                       'id': '596/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '1770556986',
                                                                       'in_ucast_pkts': '8306747',
                                                                       'mac_address': '00:06:31:69:b4:b9',
                                                                       'mtu': 2000,
                                                                       'name': '596/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119629',
                                                                       'out_octets': '29801295936',
                                                                       'out_ucast_pkts': '21010117',
                                                                       'poller_hash': 'QzUwMDJwZWVyaW50ZjU5Ni9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '596',
                                                                       'in_nucast_pkts': '1119629',
                                                                       'in_octets': '29801295936',
                                                                       'in_ucast_pkts': '21010117',
                                                                       'links': [       {       'in_nucast_pkts': '1119629',
                                                                                                'in_octets': '29801295936',
                                                                                                'in_ucast_pkts': '21010117',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '1770556986',
                                                                                                'out_ucast_pkts': '8306747',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '1770556986',
                                                                       'out_ucast_pkts': '8306747',
                                                                       'poller_hash': 'QzUwMDJwZWVyaW50ZnBvbiA1OTY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '596',
                                                                                               'links': [       {       'connected_time': '688413',
                                                                                                                        'fiber_length': 1154,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzUwMDJwZWVycG9ucG9uIDU5Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '31 Gerber Dr. Milverton (Nadine '
                                                 'Hanna)',
                                         'poller_hash': 'QzUwMDI=',
                                         'serial_number': 'C5002',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:c2',
                                         'uptime': '688413'},
                                 {       'counter_source': '597/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:12'],
                                                                       'id': '597/g1',
                                                                       'in_nucast_pkts': '787872',
                                                                       'in_octets': '68770153914',
                                                                       'in_ucast_pkts': '233287236',
                                                                       'mac_address': '00:06:31:69:b4:a9',
                                                                       'mtu': 2000,
                                                                       'name': '597/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75606043',
                                                                       'out_octets': '1735607968031',
                                                                       'out_ucast_pkts': '1239520494',
                                                                       'poller_hash': 'QzUwMDRwZWVyaW50ZjU5Ny9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '597',
                                                                       'in_nucast_pkts': '75606043',
                                                                       'in_octets': '1735607968031',
                                                                       'in_ucast_pkts': '1239520494',
                                                                       'links': [       {       'in_nucast_pkts': '75606043',
                                                                                                'in_octets': '1735607968031',
                                                                                                'in_ucast_pkts': '1239520494',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '787872',
                                                                                                'out_octets': '68770153914',
                                                                                                'out_ucast_pkts': '233287236',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '787872',
                                                                       'out_octets': '68770153914',
                                                                       'out_ucast_pkts': '233287236',
                                                                       'poller_hash': 'QzUwMDRwZWVyaW50ZnBvbiA1OTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '597',
                                                                                               'links': [       {       'connected_time': '41171',
                                                                                                                        'fiber_length': 1142,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzUwMDRwZWVycG9ucG9uIDU5Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '23 Gerber Dr. Milverton (Joe '
                                                 'Streicher)',
                                         'poller_hash': 'QzUwMDQ=',
                                         'serial_number': 'C5004',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:c4',
                                         'uptime': '41171'},
                                 {       'counter_source': '598/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '598/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '360032718082',
                                                                       'in_ucast_pkts': '2889681745',
                                                                       'mac_address': '00:06:31:69:ea:41',
                                                                       'name': '598/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '16347137802612',
                                                                       'out_ucast_pkts': '11113069166',
                                                                       'poller_hash': 'QzU3MDNwZWVyaW50ZjU5OC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '598',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '16347137802612',
                                                                       'in_ucast_pkts': '11113069166',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '16347137802612',
                                                                                                'in_ucast_pkts': '11113069166',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '360032718082',
                                                                                                'out_ucast_pkts': '2889681745',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '360032718082',
                                                                       'out_ucast_pkts': '2889681745',
                                                                       'poller_hash': 'QzU3MDNwZWVyaW50ZnBvbiA1OTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '598',
                                                                                               'links': [       {       'connected_time': '13218205',
                                                                                                                        'fiber_length': 1178,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3MDNwZWVycG9ucG9uIDU5OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '19 Gerber Dr. Milverton (Rob McEwen)',
                                         'poller_hash': 'QzU3MDM=',
                                         'serial_number': 'C5703',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:67:c3',
                                         'uptime': '13218205'},
                                 {       'counter_source': '599/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4f:b7:4d'],
                                                                       'id': '599/g1',
                                                                       'in_nucast_pkts': '18326',
                                                                       'in_octets': '35783339294',
                                                                       'in_ucast_pkts': '245951703',
                                                                       'mac_address': '00:06:31:69:f0:55',
                                                                       'mtu': 2000,
                                                                       'name': '599/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3188229',
                                                                       'out_octets': '1195433354666',
                                                                       'out_ucast_pkts': '829688936',
                                                                       'poller_hash': 'QzU3NDVwZWVyaW50ZjU5OS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '599',
                                                                       'in_nucast_pkts': '3188229',
                                                                       'in_octets': '1195433354666',
                                                                       'in_ucast_pkts': '829688936',
                                                                       'links': [       {       'in_nucast_pkts': '3188229',
                                                                                                'in_octets': '1195433354666',
                                                                                                'in_ucast_pkts': '829688936',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18326',
                                                                                                'out_octets': '35783339294',
                                                                                                'out_ucast_pkts': '245951703',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18326',
                                                                       'out_octets': '35783339294',
                                                                       'out_ucast_pkts': '245951703',
                                                                       'poller_hash': 'QzU3NDVwZWVyaW50ZnBvbiA1OTk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '599',
                                                                                               'links': [       {       'connected_time': '68514702',
                                                                                                                        'fiber_length': 1027,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3NDVwZWVycG9ucG9uIDU5OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '40 GERBER DR (AMSTEIN DEBORAH)',
                                         'poller_hash': 'QzU3NDU=',
                                         'serial_number': 'C5745',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:68:05',
                                         'uptime': '68514702'},
                                 {       'counter_source': '600/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2c:70:52'],
                                                                       'id': '600/g1',
                                                                       'in_nucast_pkts': '162',
                                                                       'in_octets': '223422595204',
                                                                       'in_ucast_pkts': '1771093812',
                                                                       'mac_address': '00:06:31:6d:9b:bf',
                                                                       'mtu': 2000,
                                                                       'name': '600/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23532693',
                                                                       'out_octets': '4997576277456',
                                                                       'out_ucast_pkts': '3439787621',
                                                                       'poller_hash': 'QzREREZwZWVyaW50ZjYwMC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '600',
                                                                       'in_nucast_pkts': '23532693',
                                                                       'in_octets': '4997576277456',
                                                                       'in_ucast_pkts': '3439787621',
                                                                       'links': [       {       'in_nucast_pkts': '23532693',
                                                                                                'in_octets': '4997576277456',
                                                                                                'in_ucast_pkts': '3439787621',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '162',
                                                                                                'out_octets': '223422595204',
                                                                                                'out_ucast_pkts': '1771093812',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '162',
                                                                       'out_octets': '223422595204',
                                                                       'out_ucast_pkts': '1771093812',
                                                                       'poller_hash': 'QzREREZwZWVyaW50ZnBvbiA2MDA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '600',
                                                                                               'links': [       {       'connected_time': '13219155',
                                                                                                                        'fiber_length': 1163,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzREREZwZWVycG9ucG9uIDYwMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '15 Gerber Dr. Milverton (Jeff '
                                                 'Keller)',
                                         'poller_hash': 'QzREREY=',
                                         'serial_number': 'C4DDF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:5e:9f',
                                         'uptime': '13219155'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '601/g1',
                                                                       'name': '601/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzRERUFwZWVyaW50ZjYwMS9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '601',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzRERUFwZWVyaW50ZnBvbiA2MDE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '601',
                                                                                               'links': [       {       'connected_time': '47369',
                                                                                                                        'fiber_length': 899,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRERUFwZWVycG9ucG9uIDYwMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '19 St. Marys St. S. Milverton '
                                                 '(KUEPFER MARTHA)',
                                         'poller_hash': 'QzRERUE=',
                                         'serial_number': 'C4DEA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:5e:aa',
                                         'uptime': '47369'},
                                 {       'counter_source': '602/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e8:eb'],
                                                                       'id': '602/g1',
                                                                       'in_nucast_pkts': '14724',
                                                                       'in_octets': '5298837087',
                                                                       'in_ucast_pkts': '11923135',
                                                                       'mac_address': '00:06:31:69:a1:55',
                                                                       'mtu': 2000,
                                                                       'name': '602/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119608',
                                                                       'out_octets': '62626767483',
                                                                       'out_ucast_pkts': '47861488',
                                                                       'poller_hash': 'QzRERDFwZWVyaW50ZjYwMi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '602',
                                                                       'in_nucast_pkts': '1119608',
                                                                       'in_octets': '62626767483',
                                                                       'in_ucast_pkts': '47861488',
                                                                       'links': [       {       'in_nucast_pkts': '1119608',
                                                                                                'in_octets': '62626767483',
                                                                                                'in_ucast_pkts': '47861488',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14724',
                                                                                                'out_octets': '5298837087',
                                                                                                'out_ucast_pkts': '11923135',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14724',
                                                                       'out_octets': '5298837087',
                                                                       'out_ucast_pkts': '11923135',
                                                                       'poller_hash': 'QzRERDFwZWVyaW50ZnBvbiA2MDI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '602',
                                                                                               'links': [       {       'connected_time': '688415',
                                                                                                                        'fiber_length': 1139,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRERDFwZWVycG9ucG9uIDYwMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4 Jo-Lenne Ct. Milverton (Jay '
                                                 'Schultz)',
                                         'poller_hash': 'QzRERDE=',
                                         'serial_number': 'C4DD1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:5e:91',
                                         'uptime': '688415'},
                                 {       'counter_source': '603/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7e:12'],
                                                                       'id': '603/g1',
                                                                       'in_nucast_pkts': '14402',
                                                                       'in_octets': '18937474657',
                                                                       'in_ucast_pkts': '45580790',
                                                                       'mac_address': '00:06:31:69:a1:ed',
                                                                       'mtu': 2000,
                                                                       'name': '603/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119711',
                                                                       'out_octets': '487633314543',
                                                                       'out_ucast_pkts': '353175908',
                                                                       'poller_hash': 'QzRERkRwZWVyaW50ZjYwMy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '603',
                                                                       'in_nucast_pkts': '1119711',
                                                                       'in_octets': '487633314543',
                                                                       'in_ucast_pkts': '353175908',
                                                                       'links': [       {       'in_nucast_pkts': '1119711',
                                                                                                'in_octets': '487633314543',
                                                                                                'in_ucast_pkts': '353175908',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14402',
                                                                                                'out_octets': '18937474657',
                                                                                                'out_ucast_pkts': '45580790',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14402',
                                                                       'out_octets': '18937474657',
                                                                       'out_ucast_pkts': '45580790',
                                                                       'poller_hash': 'QzRERkRwZWVyaW50ZnBvbiA2MDM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '603',
                                                                                               'links': [       {       'connected_time': '688414',
                                                                                                                        'fiber_length': 1122,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRERkRwZWVycG9ucG9uIDYwMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '8 JO-LENE CT (FARIES JACINTA)',
                                         'poller_hash': 'QzRERkQ=',
                                         'serial_number': 'C4DFD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:5e:bd',
                                         'uptime': '688414'},
                                 {       'counter_source': '605/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a4:f0'],
                                                                       'id': '605/g1',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '8249711709',
                                                                       'in_ucast_pkts': '69524452',
                                                                       'mac_address': '00:06:31:87:4f:1f',
                                                                       'mtu': 2000,
                                                                       'name': '605/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119387',
                                                                       'out_octets': '214943491131',
                                                                       'out_ucast_pkts': '148429919',
                                                                       'poller_hash': 'Q0MwQzFwZWVyaW50ZjYwNS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '605',
                                                                       'in_nucast_pkts': '1119387',
                                                                       'in_octets': '214943491131',
                                                                       'in_ucast_pkts': '148429919',
                                                                       'links': [       {       'in_nucast_pkts': '1119387',
                                                                                                'in_octets': '214943491131',
                                                                                                'in_ucast_pkts': '148429919',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19',
                                                                                                'out_octets': '8249711709',
                                                                                                'out_ucast_pkts': '69524452',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19',
                                                                       'out_octets': '8249711709',
                                                                       'out_ucast_pkts': '69524452',
                                                                       'poller_hash': 'Q0MwQzFwZWVyaW50ZnBvbiA2MDU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '605',
                                                                                               'links': [       {       'connected_time': '688358',
                                                                                                                        'fiber_length': 1197,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQzFwZWVycG9ucG9uIDYwNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '11 JO-LENE CT (BARKER PATRICIA)',
                                         'poller_hash': 'Q0MwQzE=',
                                         'serial_number': 'CC0C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:81',
                                         'uptime': '688358'},
                                 {       'counter_source': '607/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '607/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '341134059259',
                                                                       'in_ucast_pkts': '2220442544',
                                                                       'mac_address': '00:06:31:dd:55:1f',
                                                                       'name': '607/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12229459173047',
                                                                       'out_ucast_pkts': '8254388381',
                                                                       'poller_hash': 'MTc1RTM5cGVlcmludGY2MDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '607',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12229459173047',
                                                                       'in_ucast_pkts': '8254388381',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '12229459173047',
                                                                                                'in_ucast_pkts': '8254388381',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '341134059259',
                                                                                                'out_ucast_pkts': '2220442544',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '341134059259',
                                                                       'out_ucast_pkts': '2220442544',
                                                                       'poller_hash': 'MTc1RTM5cGVlcmludGZwb24gNjA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '607',
                                                                                               'links': [       {       'connected_time': '47454',
                                                                                                                        'fiber_length': 1163,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTc1RTM5cGVlcnBvbnBvbiA2MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '3 Jo-Lene Ct. Milverton (Bill&June '
                                                 'Coulter)',
                                         'poller_hash': 'MTc1RTM5',
                                         'serial_number': '175E39',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:dd:55:1f',
                                         'uptime': '47454'},
                                 {       'counter_source': '608/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '608/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '434103703880',
                                                                       'in_ucast_pkts': '1208678254',
                                                                       'mac_address': '00:06:31:52:5a:86',
                                                                       'name': '608/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4344061812092',
                                                                       'out_ucast_pkts': '3388453620',
                                                                       'poller_hash': 'ODg5QjdwZWVyaW50ZjYwOC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '608',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4344061812092',
                                                                       'in_ucast_pkts': '3388453620',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4344061812092',
                                                                                                'in_ucast_pkts': '3388453620',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '434103703880',
                                                                                                'out_ucast_pkts': '1208678254',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '434103703880',
                                                                       'out_ucast_pkts': '1208678254',
                                                                       'poller_hash': 'ODg5QjdwZWVyaW50ZnBvbiA2MDg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '608',
                                                                                               'links': [       {       'connected_time': '688407',
                                                                                                                        'fiber_length': 891,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5QjdwZWVycG9ucG9uIDYwOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '8 Gerber Ct. Milverton (Maynard '
                                                 'Streicher)',
                                         'poller_hash': 'ODg5Qjc=',
                                         'serial_number': '889B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:9f',
                                         'uptime': '688407'},
                                 {       'counter_source': '609/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '609/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '153191736628',
                                                                       'in_ucast_pkts': '880880549',
                                                                       'mac_address': '00:06:31:6d:9a:53',
                                                                       'name': '609/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5327560952129',
                                                                       'out_ucast_pkts': '3981514416',
                                                                       'poller_hash': 'QzRERUNwZWVyaW50ZjYwOS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '609',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5327560952129',
                                                                       'in_ucast_pkts': '3981514416',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '5327560952129',
                                                                                                'in_ucast_pkts': '3981514416',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '153191736628',
                                                                                                'out_ucast_pkts': '880880549',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '153191736628',
                                                                       'out_ucast_pkts': '880880549',
                                                                       'poller_hash': 'QzRERUNwZWVyaW50ZnBvbiA2MDk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '609',
                                                                                               'links': [       {       'connected_time': '8385513',
                                                                                                                        'fiber_length': 704,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRERUNwZWVycG9ucG9uIDYwOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '5 St. Marys St. S. Milverton (Wayne '
                                                 'Poole 10421583)',
                                         'poller_hash': 'QzRERUM=',
                                         'serial_number': 'C4DEC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:5e:ac',
                                         'uptime': '8385513'},
                                 {       'counter_source': '610/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:8b:cd:60:08:0c'],
                                                                       'id': '610/g1',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '1515865747',
                                                                       'in_ucast_pkts': '6421302',
                                                                       'mac_address': '00:06:31:52:64:42',
                                                                       'mtu': 2000,
                                                                       'name': '610/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119745',
                                                                       'out_octets': '31567542553',
                                                                       'out_ucast_pkts': '25880507',
                                                                       'poller_hash': 'ODg5RTlwZWVyaW50ZjYxMC9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '610',
                                                                       'in_nucast_pkts': '1119745',
                                                                       'in_octets': '31567542553',
                                                                       'in_ucast_pkts': '25880507',
                                                                       'links': [       {       'in_nucast_pkts': '1119745',
                                                                                                'in_octets': '31567542553',
                                                                                                'in_ucast_pkts': '25880507',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20',
                                                                                                'out_octets': '1515865747',
                                                                                                'out_ucast_pkts': '6421302',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20',
                                                                       'out_octets': '1515865747',
                                                                       'out_ucast_pkts': '6421302',
                                                                       'poller_hash': 'ODg5RTlwZWVyaW50ZnBvbiA2MTA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '610',
                                                                                               'links': [       {       'connected_time': '688400',
                                                                                                                        'fiber_length': 718,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RTlwZWVycG9ucG9uIDYxMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '2 St. Marys St. S. Milverton (Darcy '
                                                 'Bauman 10433545)',
                                         'poller_hash': 'ODg5RTk=',
                                         'serial_number': '889E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:d1',
                                         'uptime': '688400'},
                                 {       'counter_source': '611/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '611/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '795540980014',
                                                                       'in_ucast_pkts': '4395297900',
                                                                       'mac_address': '00:06:31:69:b2:4d',
                                                                       'name': '611/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '14811366790480',
                                                                       'out_ucast_pkts': '12678147574',
                                                                       'poller_hash': 'QzRGOUJwZWVyaW50ZjYxMS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '611',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '14811366790480',
                                                                       'in_ucast_pkts': '12678147574',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '14811366790480',
                                                                                                'in_ucast_pkts': '12678147574',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '795540980014',
                                                                                                'out_ucast_pkts': '4395297900',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '795540980014',
                                                                       'out_ucast_pkts': '4395297900',
                                                                       'poller_hash': 'QzRGOUJwZWVyaW50ZnBvbiA2MTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '611',
                                                                                               'links': [       {       'connected_time': '688408',
                                                                                                                        'fiber_length': 911,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGOUJwZWVycG9ucG9uIDYxMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '17 ST MARYS ST S (Kristin Brooks '
                                                 '10438137)',
                                         'poller_hash': 'QzRGOUI=',
                                         'serial_number': 'C4F9B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:5b',
                                         'uptime': '688408'},
                                 {       'counter_source': '613/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '613/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '81000695960',
                                                                       'in_ucast_pkts': '257528207',
                                                                       'mac_address': '00:06:31:69:b3:cd',
                                                                       'name': '613/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '626027854941',
                                                                       'out_ucast_pkts': '584625284',
                                                                       'poller_hash': 'QzRGRkFwZWVyaW50ZjYxMy9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '613',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '626027854941',
                                                                       'in_ucast_pkts': '584625284',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '626027854941',
                                                                                                'in_ucast_pkts': '584625284',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '81000695960',
                                                                                                'out_ucast_pkts': '257528207',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '81000695960',
                                                                       'out_ucast_pkts': '257528207',
                                                                       'poller_hash': 'QzRGRkFwZWVyaW50ZnBvbiA2MTM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '613',
                                                                                               'links': [       {       'connected_time': '47941',
                                                                                                                        'fiber_length': 935,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGRkFwZWVycG9ucG9uIDYxMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '15 ST MARYS ST S (MOHR RON)',
                                         'poller_hash': 'QzRGRkE=',
                                         'serial_number': 'C4FFA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:ba',
                                         'uptime': '47941'},
                                 {       'counter_source': '614/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '614/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '404918539647',
                                                                       'in_ucast_pkts': '4220523158',
                                                                       'mac_address': '00:06:31:69:b3:d5',
                                                                       'name': '614/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '24200369063594',
                                                                       'out_ucast_pkts': '16236970941',
                                                                       'poller_hash': 'QzRGRjhwZWVyaW50ZjYxNC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '614',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '24200369063594',
                                                                       'in_ucast_pkts': '16236970941',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '24200369063594',
                                                                                                'in_ucast_pkts': '16236970941',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '404918539647',
                                                                                                'out_ucast_pkts': '4220523158',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '404918539647',
                                                                       'out_ucast_pkts': '4220523158',
                                                                       'poller_hash': 'QzRGRjhwZWVyaW50ZnBvbiA2MTQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '614',
                                                                                               'links': [       {       'connected_time': '688404',
                                                                                                                        'fiber_length': 732,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGRjhwZWVycG9ucG9uIDYxNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4 St. Marys St. S. Milverton (Rick '
                                                 'Baier)',
                                         'poller_hash': 'QzRGRjg=',
                                         'serial_number': 'C4FF8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:b8',
                                         'uptime': '688404'},
                                 {       'counter_source': '615/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '615/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '523691200358',
                                                                       'in_ucast_pkts': '2931181063',
                                                                       'mac_address': '00:06:31:69:b2:f5',
                                                                       'name': '615/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10567860334678',
                                                                       'out_ucast_pkts': '7542637009',
                                                                       'poller_hash': 'QzRGRUFwZWVyaW50ZjYxNS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '615',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10567860334678',
                                                                       'in_ucast_pkts': '7542637009',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '10567860334678',
                                                                                                'in_ucast_pkts': '7542637009',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '523691200358',
                                                                                                'out_ucast_pkts': '2931181063',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '523691200358',
                                                                       'out_ucast_pkts': '2931181063',
                                                                       'poller_hash': 'QzRGRUFwZWVyaW50ZnBvbiA2MTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '615',
                                                                                               'links': [       {       'connected_time': '21319272',
                                                                                                                        'fiber_length': 1016,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGRUFwZWVycG9ucG9uIDYxNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '8 ST MARYS ST S (WADDELL DALE)',
                                         'poller_hash': 'QzRGRUE=',
                                         'serial_number': 'C4FEA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:aa',
                                         'uptime': '21319272'},
                                 {       'counter_source': '616/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a0:ab:1b:48:c4:02'],
                                                                       'id': '616/g1',
                                                                       'in_nucast_pkts': '4667170',
                                                                       'in_octets': '50388400459',
                                                                       'in_ucast_pkts': '185960096',
                                                                       'mac_address': '00:06:31:69:b4:79',
                                                                       'mtu': 2000,
                                                                       'name': '616/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75614116',
                                                                       'out_octets': '887086887145',
                                                                       'out_ucast_pkts': '729332072',
                                                                       'poller_hash': 'QzRGRkJwZWVyaW50ZjYxNi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '616',
                                                                       'in_nucast_pkts': '75614116',
                                                                       'in_octets': '887086887145',
                                                                       'in_ucast_pkts': '729332072',
                                                                       'links': [       {       'in_nucast_pkts': '75614116',
                                                                                                'in_octets': '887086887145',
                                                                                                'in_ucast_pkts': '729332072',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4667170',
                                                                                                'out_octets': '50388400459',
                                                                                                'out_ucast_pkts': '185960096',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4667170',
                                                                       'out_octets': '50388400459',
                                                                       'out_ucast_pkts': '185960096',
                                                                       'poller_hash': 'QzRGRkJwZWVyaW50ZnBvbiA2MTY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '616',
                                                                                               'links': [       {       'connected_time': '40948',
                                                                                                                        'fiber_length': 893,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGRkJwZWVycG9ucG9uIDYxNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '12 Gerber Ct. Milverton (William '
                                                 'Westman)',
                                         'poller_hash': 'QzRGRkI=',
                                         'serial_number': 'C4FFB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:bb',
                                         'uptime': '40948'},
                                 {       'counter_source': '617/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f2:0f'],
                                                                       'id': '617/g1',
                                                                       'in_nucast_pkts': '758234',
                                                                       'in_octets': '452041457508',
                                                                       'in_ucast_pkts': '2079179850',
                                                                       'mac_address': '00:06:31:69:a1:fd',
                                                                       'mtu': 2000,
                                                                       'name': '617/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75585107',
                                                                       'out_octets': '5568915967185',
                                                                       'out_ucast_pkts': '4202586371',
                                                                       'poller_hash': 'QzRERjZwZWVyaW50ZjYxNy9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '617',
                                                                       'in_nucast_pkts': '75585107',
                                                                       'in_octets': '5568915967185',
                                                                       'in_ucast_pkts': '4202586371',
                                                                       'links': [       {       'in_nucast_pkts': '75585107',
                                                                                                'in_octets': '5568915967185',
                                                                                                'in_ucast_pkts': '4202586371',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '758234',
                                                                                                'out_octets': '452041457508',
                                                                                                'out_ucast_pkts': '2079179850',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '758234',
                                                                       'out_octets': '452041457508',
                                                                       'out_ucast_pkts': '2079179850',
                                                                       'poller_hash': 'QzRERjZwZWVyaW50ZnBvbiA2MTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '617',
                                                                                               'links': [       {       'connected_time': '41049',
                                                                                                                        'fiber_length': 894,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRERjZwZWVycG9ucG9uIDYxNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '24 Gerber Ct. Milverton (John '
                                                 'Rasmussen)',
                                         'poller_hash': 'QzRERjY=',
                                         'serial_number': 'C4DF6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:5e:b6',
                                         'uptime': '41049'},
                                 {       'counter_source': '618/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '618/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '70047078405',
                                                                       'in_ucast_pkts': '477442318',
                                                                       'mac_address': '00:06:31:69:f6:61',
                                                                       'name': '618/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2973947095223',
                                                                       'out_ucast_pkts': '2063099395',
                                                                       'poller_hash': 'QzU4NUZwZWVyaW50ZjYxOC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '618',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2973947095223',
                                                                       'in_ucast_pkts': '2063099395',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2973947095223',
                                                                                                'in_ucast_pkts': '2063099395',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '70047078405',
                                                                                                'out_ucast_pkts': '477442318',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '70047078405',
                                                                       'out_ucast_pkts': '477442318',
                                                                       'poller_hash': 'QzU4NUZwZWVyaW50ZnBvbiA2MTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '618',
                                                                                               'links': [       {       'connected_time': '31279509',
                                                                                                                        'fiber_length': 923,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU4NUZwZWVycG9ucG9uIDYxOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '28 Gerber Ct. Milverton (George '
                                                 'Tanner)',
                                         'poller_hash': 'QzU4NUY=',
                                         'serial_number': 'C585F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:69:1f',
                                         'uptime': '31279509'},
                                 {       'counter_source': '619/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:32'],
                                                                       'id': '619/g1',
                                                                       'in_nucast_pkts': '13398',
                                                                       'in_octets': '6193282158',
                                                                       'in_ucast_pkts': '28772459',
                                                                       'mac_address': '00:06:31:69:ef:85',
                                                                       'mtu': 2000,
                                                                       'name': '619/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1108721',
                                                                       'out_octets': '165511465437',
                                                                       'out_ucast_pkts': '116313344',
                                                                       'poller_hash': 'QzU3MTNwZWVyaW50ZjYxOS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '619',
                                                                       'in_nucast_pkts': '1108721',
                                                                       'in_octets': '165511465437',
                                                                       'in_ucast_pkts': '116313344',
                                                                       'links': [       {       'in_nucast_pkts': '1108721',
                                                                                                'in_octets': '165511465437',
                                                                                                'in_ucast_pkts': '116313344',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13398',
                                                                                                'out_octets': '6193282158',
                                                                                                'out_ucast_pkts': '28772459',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13398',
                                                                       'out_octets': '6193282158',
                                                                       'out_ucast_pkts': '28772459',
                                                                       'poller_hash': 'QzU3MTNwZWVyaW50ZnBvbiA2MTk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '619',
                                                                                               'links': [       {       'connected_time': '688407',
                                                                                                                        'fiber_length': 871,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3MTNwZWVycG9ucG9uIDYxOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '16 GERBER CT (STOKER BRANDON)',
                                         'poller_hash': 'QzU3MTM=',
                                         'serial_number': 'C5713',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:67:d3',
                                         'uptime': '688407'},
                                 {       'counter_source': '620/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '620/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '814121355929',
                                                                       'in_ucast_pkts': '7865277151',
                                                                       'mac_address': '00:06:31:87:4e:73',
                                                                       'name': '620/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '43852631136757',
                                                                       'out_ucast_pkts': '29337407185',
                                                                       'poller_hash': 'Q0MwQzlwZWVyaW50ZjYyMC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '620',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '43852631136757',
                                                                       'in_ucast_pkts': '29337407185',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '43852631136757',
                                                                                                'in_ucast_pkts': '29337407185',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '814121355929',
                                                                                                'out_ucast_pkts': '7865277151',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '814121355929',
                                                                       'out_ucast_pkts': '7865277151',
                                                                       'poller_hash': 'Q0MwQzlwZWVyaW50ZnBvbiA2MjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '620',
                                                                                               'links': [       {       'connected_time': '20997057',
                                                                                                                        'fiber_length': 918,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQzlwZWVycG9ucG9uIDYyMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '10 St. Marys St. S. Milverton (Ron '
                                                 'Schweitzer)',
                                         'poller_hash': 'Q0MwQzk=',
                                         'serial_number': 'CC0C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:89',
                                         'uptime': '20997057'},
                                 {       'counter_source': '621/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '621/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '222725364514',
                                                                       'in_ucast_pkts': '486020552',
                                                                       'mac_address': '00:06:31:69:ee:19',
                                                                       'name': '621/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1876894309267',
                                                                       'out_ucast_pkts': '1551821792',
                                                                       'poller_hash': 'QzU3NzhwZWVyaW50ZjYyMS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '621',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1876894309267',
                                                                       'in_ucast_pkts': '1551821792',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1876894309267',
                                                                                                'in_ucast_pkts': '1551821792',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '222725364514',
                                                                                                'out_ucast_pkts': '486020552',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '222725364514',
                                                                       'out_ucast_pkts': '486020552',
                                                                       'poller_hash': 'QzU3NzhwZWVyaW50ZnBvbiA2MjE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '621',
                                                                                               'links': [       {       'connected_time': '19105352',
                                                                                                                        'fiber_length': 913,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3NzhwZWVycG9ucG9uIDYyMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '3 Gerber Ct. Milverton (Durell '
                                                 'Brubacher)',
                                         'poller_hash': 'QzU3Nzg=',
                                         'serial_number': 'C5778',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:68:38',
                                         'uptime': '19105352'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '622/g1',
                                                                       'name': '622/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QjUwREM0cGVlcmludGY2MjIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '622',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QjUwREM0cGVlcmludGZwb24gNjIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '622',
                                                                                               'links': [       {       'connected_time': '67790295',
                                                                                                                        'fiber_length': 16627,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUwREM0cGVlcnBvbnBvbiA2MjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4699 Line 49 (Darryl Terpstra '
                                                 '10430074)',
                                         'poller_hash': 'QjUwREM0',
                                         'serial_number': 'B50DC4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8e:f5:00',
                                         'uptime': '67790295'},
                                 {       'counter_source': '623/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:94:74'],
                                                                       'id': '623/g1',
                                                                       'in_nucast_pkts': '85',
                                                                       'in_octets': '259681113114',
                                                                       'in_ucast_pkts': '2344649092',
                                                                       'mac_address': '00:06:31:69:a2:1d',
                                                                       'mtu': 2000,
                                                                       'name': '623/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23530378',
                                                                       'out_octets': '8986014715752',
                                                                       'out_ucast_pkts': '6054430676',
                                                                       'poller_hash': 'QzRERkJwZWVyaW50ZjYyMy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '623',
                                                                       'in_nucast_pkts': '23530378',
                                                                       'in_octets': '8986014715752',
                                                                       'in_ucast_pkts': '6054430676',
                                                                       'links': [       {       'in_nucast_pkts': '23530378',
                                                                                                'in_octets': '8986014715752',
                                                                                                'in_ucast_pkts': '6054430676',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '85',
                                                                                                'out_octets': '259681113114',
                                                                                                'out_ucast_pkts': '2344649092',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '85',
                                                                       'out_octets': '259681113114',
                                                                       'out_ucast_pkts': '2344649092',
                                                                       'poller_hash': 'QzRERkJwZWVyaW50ZnBvbiA2MjM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '623',
                                                                                               'links': [       {       'connected_time': '13219225',
                                                                                                                        'fiber_length': 878,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRERkJwZWVycG9ucG9uIDYyMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '16 ST MARYS ST S (SCHMIDT DAVID & '
                                                 'KAREN)',
                                         'poller_hash': 'QzRERkI=',
                                         'serial_number': 'C4DFB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:5e:bb',
                                         'uptime': '13219225'},
                                 {       'counter_source': '624/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '40:b0:76:33:aa:00'],
                                                                       'id': '624/g1',
                                                                       'in_nucast_pkts': '106',
                                                                       'in_octets': '146865990987',
                                                                       'in_ucast_pkts': '303783776',
                                                                       'mac_address': '00:06:31:69:f4:c5',
                                                                       'mtu': 2000,
                                                                       'name': '624/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27326996',
                                                                       'out_octets': '1556905173229',
                                                                       'out_ucast_pkts': '1227837001',
                                                                       'poller_hash': 'QzU3RThwZWVyaW50ZjYyNC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '624',
                                                                       'in_nucast_pkts': '27326996',
                                                                       'in_octets': '1556905173229',
                                                                       'in_ucast_pkts': '1227837001',
                                                                       'links': [       {       'in_nucast_pkts': '27326996',
                                                                                                'in_octets': '1556905173229',
                                                                                                'in_ucast_pkts': '1227837001',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '106',
                                                                                                'out_octets': '146865990987',
                                                                                                'out_ucast_pkts': '303783776',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '106',
                                                                       'out_octets': '146865990987',
                                                                       'out_ucast_pkts': '303783776',
                                                                       'poller_hash': 'QzU3RThwZWVyaW50ZnBvbiA2MjQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '624',
                                                                                               'links': [       {       'connected_time': '15498391',
                                                                                                                        'fiber_length': 947,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3RThwZWVycG9ucG9uIDYyNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '13 St. Marys St. S. Milverton '
                                                 '(Heather Westman)',
                                         'poller_hash': 'QzU3RTg=',
                                         'serial_number': 'C57E8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:68:a8',
                                         'uptime': '15498391'},
                                 {       'counter_source': '625/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '625/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '815885935683',
                                                                       'in_ucast_pkts': '2448812465',
                                                                       'mac_address': '00:06:31:69:f5:09',
                                                                       'name': '625/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12553686561675',
                                                                       'out_ucast_pkts': '9284067750',
                                                                       'poller_hash': 'QzU3RjFwZWVyaW50ZjYyNS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '625',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12553686561675',
                                                                       'in_ucast_pkts': '9284067750',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '12553686561675',
                                                                                                'in_ucast_pkts': '9284067750',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '815885935683',
                                                                                                'out_ucast_pkts': '2448812465',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '815885935683',
                                                                       'out_ucast_pkts': '2448812465',
                                                                       'poller_hash': 'QzU3RjFwZWVyaW50ZnBvbiA2MjU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '625',
                                                                                               'links': [       {       'connected_time': '688409',
                                                                                                                        'fiber_length': 893,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3RjFwZWVycG9ucG9uIDYyNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '20 GERBER CT. (SCHWARTZENTRUBER '
                                                 'GRANT)',
                                         'poller_hash': 'QzU3RjE=',
                                         'serial_number': 'C57F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:68:b1',
                                         'uptime': '688409'},
                                 {       'counter_source': '626/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '10:c3:7b:a3:51:38'],
                                                                       'id': '626/g4',
                                                                       'in_nucast_pkts': '2538',
                                                                       'in_octets': '489812510210',
                                                                       'in_ucast_pkts': '3191781464',
                                                                       'mac_address': '00:06:31:87:4f:2a',
                                                                       'mtu': 2000,
                                                                       'name': '626/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75606187',
                                                                       'out_octets': '7896808577736',
                                                                       'out_ucast_pkts': '5485713657',
                                                                       'poller_hash': 'Q0MwRDVwZWVyaW50ZjYyNi9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '626',
                                                                       'in_nucast_pkts': '75606187',
                                                                       'in_octets': '7896808577736',
                                                                       'in_ucast_pkts': '5485713657',
                                                                       'links': [       {       'in_nucast_pkts': '75606187',
                                                                                                'in_octets': '7896808577736',
                                                                                                'in_ucast_pkts': '5485713657',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2538',
                                                                                                'out_octets': '489812510210',
                                                                                                'out_ucast_pkts': '3191781464',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2538',
                                                                       'out_octets': '489812510210',
                                                                       'out_ucast_pkts': '3191781464',
                                                                       'poller_hash': 'Q0MwRDVwZWVyaW50ZnBvbiA2MjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '626',
                                                                                               'links': [       {       'connected_time': '40403',
                                                                                                                        'fiber_length': 947,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwRDVwZWVycG9ucG9uIDYyNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '32 Gerber Ct. Milverton (Ken Leduc)',
                                         'poller_hash': 'Q0MwRDU=',
                                         'serial_number': 'CC0D5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:95',
                                         'uptime': '40403'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '627/g1',
                                                                       'name': '627/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwREZEcGVlcmludGY2MjcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '627',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwREZEcGVlcmludGZwb24gNjI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '627',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwREZEcGVlcnBvbnBvbiA2Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '9 St. Marys St. N. Milverton (Alex '
                                                 'Campbell 10435536)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "627" '
                                                                  '(location: GPON '
                                                                  'port "6/2"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-08-04T18:40:28.240Z',
                                         'poller_hash': 'MzAwREZE',
                                         'serial_number': '300DFD',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '629/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:73'],
                                                                       'id': '629/g1',
                                                                       'in_nucast_pkts': '935033',
                                                                       'in_octets': '284378211103',
                                                                       'in_ucast_pkts': '870282862',
                                                                       'mac_address': 'cc:be:59:cc:0a:bd',
                                                                       'mtu': 2000,
                                                                       'name': '629/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '123020404',
                                                                       'out_octets': '4649039524098',
                                                                       'out_ucast_pkts': '3507246475',
                                                                       'poller_hash': 'NDJEMkFGcGVlcmludGY2MjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '629',
                                                                       'in_nucast_pkts': '123020404',
                                                                       'in_octets': '4649039524098',
                                                                       'in_ucast_pkts': '3507246475',
                                                                       'links': [       {       'in_nucast_pkts': '123020404',
                                                                                                'in_octets': '4649039524098',
                                                                                                'in_ucast_pkts': '3507246475',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '935033',
                                                                                                'out_octets': '284378211103',
                                                                                                'out_ucast_pkts': '870282862',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '935033',
                                                                       'out_octets': '284378211103',
                                                                       'out_ucast_pkts': '870282862',
                                                                       'poller_hash': 'NDJEMkFGcGVlcmludGZwb24gNjI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '629',
                                                                                               'links': [       {       'connected_time': '21816866',
                                                                                                                        'fiber_length': 15122,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMkFGcGVlcnBvbnBvbiA2Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5216 Line 49 (MCDONNELL MAGGIE)',
                                         'poller_hash': 'NDJEMkFG',
                                         'serial_number': '42D2AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:0a:ba',
                                         'uptime': '21816866'},
                                 {       'counter_source': '630/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '630/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '356583159235',
                                                                       'in_ucast_pkts': '2555590677',
                                                                       'mac_address': '00:06:31:52:65:1e',
                                                                       'name': '630/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11326681380339',
                                                                       'out_ucast_pkts': '7989817063',
                                                                       'poller_hash': 'ODg5REVwZWVyaW50ZjYzMC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '630',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11326681380339',
                                                                       'in_ucast_pkts': '7989817063',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11326681380339',
                                                                                                'in_ucast_pkts': '7989817063',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '356583159235',
                                                                                                'out_ucast_pkts': '2555590677',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '356583159235',
                                                                       'out_ucast_pkts': '2555590677',
                                                                       'poller_hash': 'ODg5REVwZWVyaW50ZnBvbiA2MzA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '630',
                                                                                               'links': [       {       'connected_time': '13219209',
                                                                                                                        'fiber_length': 1553,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5REVwZWVycG9ucG9uIDYzMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '109 Mill St. E. Milverton (Murray '
                                                 'Stewart)',
                                         'poller_hash': 'ODg5REU=',
                                         'serial_number': '889DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:c6',
                                         'uptime': '13219209'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '631/g1',
                                                                       'name': '631/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OEEyOENwZWVyaW50ZjYzMS9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '631',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'OEEyOENwZWVyaW50ZnBvbiA2MzE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '631',
                                                                                               'links': [       {       'connected_time': '15224741',
                                                                                                                        'fiber_length': 1752,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEEyOENwZWVycG9ucG9uIDYzMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '109B Mill St. E. Milverton (Paul '
                                                 'Jantzi 10403500)',
                                         'poller_hash': 'OEEyOEM=',
                                         'serial_number': '8A28C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:47:4b:78',
                                         'uptime': '15224741'},
                                 {       'counter_source': '632/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '632/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '173832463007',
                                                                       'in_ucast_pkts': '1508835090',
                                                                       'mac_address': '04:bc:9f:5b:e5:a9',
                                                                       'name': '632/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10192484880360',
                                                                       'out_ucast_pkts': '6945909093',
                                                                       'poller_hash': 'MTI3NDMzNXBlZXJpbnRmNjMyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '632',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10192484880360',
                                                                       'in_ucast_pkts': '6945909093',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '10192484880360',
                                                                                                'in_ucast_pkts': '6945909093',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '173832463007',
                                                                                                'out_ucast_pkts': '1508835090',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '173832463007',
                                                                       'out_ucast_pkts': '1508835090',
                                                                       'poller_hash': 'MTI3NDMzNXBlZXJpbnRmcG9uIDYzMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '632',
                                                                                               'links': [       {       'connected_time': '688422',
                                                                                                                        'fiber_length': 1856,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMzNXBlZXJwb25wb24gNjMy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '113 MILL ST E (KIPFER REX)',
                                         'poller_hash': 'MTI3NDMzNQ==',
                                         'serial_number': '1274335',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:a9',
                                         'uptime': '688422'},
                                 {       'counter_source': '633/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '633/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '709831729901',
                                                                       'in_ucast_pkts': '5212861463',
                                                                       'mac_address': '00:06:31:52:66:fe',
                                                                       'name': '633/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '20033470962067',
                                                                       'out_ucast_pkts': '10036115478',
                                                                       'poller_hash': 'ODg5QkFwZWVyaW50ZjYzMy9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '633',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '20033470962067',
                                                                       'in_ucast_pkts': '10036115478',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '20033470962067',
                                                                                                'in_ucast_pkts': '10036115478',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '709831729901',
                                                                                                'out_ucast_pkts': '5212861463',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '709831729901',
                                                                       'out_ucast_pkts': '5212861463',
                                                                       'poller_hash': 'ODg5QkFwZWVyaW50ZnBvbiA2MzM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '633',
                                                                                               'links': [       {       'connected_time': '29473757',
                                                                                                                        'fiber_length': 1668,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5QkFwZWVycG9ucG9uIDYzMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4495 LINE 64 (KUEPFER KEN & '
                                                 'CHRISTENA)',
                                         'poller_hash': 'ODg5QkE=',
                                         'serial_number': '889BA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:a2',
                                         'uptime': '29473757'},
                                 {       'counter_source': '634/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ac:9e:17:b2:fb:f8'],
                                                                       'id': '634/g1',
                                                                       'in_nucast_pkts': '55',
                                                                       'in_octets': '5478112281',
                                                                       'in_ucast_pkts': '26998117',
                                                                       'mac_address': '00:06:31:69:ef:f5',
                                                                       'mtu': 2000,
                                                                       'name': '634/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119730',
                                                                       'out_octets': '110457682467',
                                                                       'out_ucast_pkts': '81434681',
                                                                       'poller_hash': 'QzU3NENwZWVyaW50ZjYzNC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '634',
                                                                       'in_nucast_pkts': '1119730',
                                                                       'in_octets': '110457682467',
                                                                       'in_ucast_pkts': '81434681',
                                                                       'links': [       {       'in_nucast_pkts': '1119730',
                                                                                                'in_octets': '110457682467',
                                                                                                'in_ucast_pkts': '81434681',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '55',
                                                                                                'out_octets': '5478112281',
                                                                                                'out_ucast_pkts': '26998117',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55',
                                                                       'out_octets': '5478112281',
                                                                       'out_ucast_pkts': '26998117',
                                                                       'poller_hash': 'QzU3NENwZWVyaW50ZnBvbiA2MzQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '634',
                                                                                               'links': [       {       'connected_time': '688418',
                                                                                                                        'fiber_length': 911,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3NENwZWVycG9ucG9uIDYzNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4 GERBER CT (KIPFER COLBY)',
                                         'poller_hash': 'QzU3NEM=',
                                         'serial_number': 'C574C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:68:0c',
                                         'uptime': '688418'},
                                 {       'counter_source': '635/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1d:bc'],
                                                                       'id': '635/g1',
                                                                       'in_nucast_pkts': '6183047',
                                                                       'in_octets': '462704081998',
                                                                       'in_ucast_pkts': '4153758055',
                                                                       'mac_address': '00:06:31:52:5e:b2',
                                                                       'mtu': 2000,
                                                                       'name': '635/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15649836734',
                                                                       'out_octets': '41766975735081',
                                                                       'out_ucast_pkts': '13942871592',
                                                                       'poller_hash': 'ODg5MjVwZWVyaW50ZjYzNS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '635',
                                                                       'in_nucast_pkts': '15649836734',
                                                                       'in_octets': '41766975735081',
                                                                       'in_ucast_pkts': '13942871592',
                                                                       'links': [       {       'in_nucast_pkts': '15649836734',
                                                                                                'in_octets': '41766975735081',
                                                                                                'in_ucast_pkts': '13942871592',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6183047',
                                                                                                'out_octets': '462704081998',
                                                                                                'out_ucast_pkts': '4153758055',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6183047',
                                                                       'out_octets': '462704081998',
                                                                       'out_ucast_pkts': '4153758055',
                                                                       'poller_hash': 'ODg5MjVwZWVyaW50ZnBvbiA2MzU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '635',
                                                                                               'links': [       {       'connected_time': '41107',
                                                                                                                        'fiber_length': 1478,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5MjVwZWVycG9ucG9uIDYzNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '105 Mill St. E. Milverton (Ted '
                                                 'Newman)',
                                         'poller_hash': 'ODg5MjU=',
                                         'serial_number': '88925',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:0d',
                                         'uptime': '41107'},
                                 {       'counter_source': '636/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ef:18'],
                                                                       'id': '636/g4',
                                                                       'in_nucast_pkts': '244754',
                                                                       'in_octets': '533975708583',
                                                                       'in_ucast_pkts': '2273440371',
                                                                       'mac_address': '00:06:31:52:5a:ad',
                                                                       'mtu': 2000,
                                                                       'name': '636/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23528274',
                                                                       'out_octets': '5310094860518',
                                                                       'out_ucast_pkts': '4409641477',
                                                                       'poller_hash': 'ODg4RjJwZWVyaW50ZjYzNi9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '636',
                                                                       'in_nucast_pkts': '23528274',
                                                                       'in_octets': '5310094860518',
                                                                       'in_ucast_pkts': '4409641477',
                                                                       'links': [       {       'in_nucast_pkts': '23528274',
                                                                                                'in_octets': '5310094860518',
                                                                                                'in_ucast_pkts': '4409641477',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '244754',
                                                                                                'out_octets': '533975708583',
                                                                                                'out_ucast_pkts': '2273440371',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '244754',
                                                                       'out_octets': '533975708583',
                                                                       'out_ucast_pkts': '2273440371',
                                                                       'poller_hash': 'ODg4RjJwZWVyaW50ZnBvbiA2MzY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '636',
                                                                                               'links': [       {       'connected_time': '13219243',
                                                                                                                        'fiber_length': 1488,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg4RjJwZWVycG9ucG9uIDYzNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '107 Mill St. E. Milverton (Alvin '
                                                 'Streicher)',
                                         'poller_hash': 'ODg4RjI=',
                                         'serial_number': '888F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:e9:da',
                                         'uptime': '13219243'},
                                 {       'counter_source': '637/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:05'],
                                                                       'id': '637/g1',
                                                                       'in_nucast_pkts': '1185747',
                                                                       'in_octets': '294693738986',
                                                                       'in_ucast_pkts': '2280277097',
                                                                       'mac_address': '00:06:31:52:40:ea',
                                                                       'mtu': 2000,
                                                                       'name': '637/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53387405',
                                                                       'out_octets': '12654045516077',
                                                                       'out_ucast_pkts': '8636031767',
                                                                       'poller_hash': 'ODg4QTVwZWVyaW50ZjYzNy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '637',
                                                                       'in_nucast_pkts': '53387405',
                                                                       'in_octets': '12654045516077',
                                                                       'in_ucast_pkts': '8636031767',
                                                                       'links': [       {       'in_nucast_pkts': '53387405',
                                                                                                'in_octets': '12654045516077',
                                                                                                'in_ucast_pkts': '8636031767',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1185747',
                                                                                                'out_octets': '294693738986',
                                                                                                'out_ucast_pkts': '2280277097',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1185747',
                                                                       'out_octets': '294693738986',
                                                                       'out_ucast_pkts': '2280277097',
                                                                       'poller_hash': 'ODg4QTVwZWVyaW50ZnBvbiA2Mzc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '637',
                                                                                               'links': [       {       'connected_time': '28527661',
                                                                                                                        'fiber_length': 1499,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg4QTVwZWVycG9ucG9uIDYzNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '103 Mill St. E. Milverton (Jim '
                                                 'Harloff)',
                                         'poller_hash': 'ODg4QTU=',
                                         'serial_number': '888A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:e9:8d',
                                         'uptime': '28527661'},
                                 {       'counter_source': '638/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'bc:22:28:df:fb:c7'],
                                                                       'id': '638/g1',
                                                                       'in_nucast_pkts': '79',
                                                                       'in_octets': '4874593975',
                                                                       'in_ucast_pkts': '16720634',
                                                                       'mac_address': '00:06:31:52:66:c2',
                                                                       'mtu': 2000,
                                                                       'name': '638/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119713',
                                                                       'out_octets': '89519572121',
                                                                       'out_ucast_pkts': '67839621',
                                                                       'poller_hash': 'ODg4RjVwZWVyaW50ZjYzOC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '638',
                                                                       'in_nucast_pkts': '1119713',
                                                                       'in_octets': '89519572121',
                                                                       'in_ucast_pkts': '67839621',
                                                                       'links': [       {       'in_nucast_pkts': '1119713',
                                                                                                'in_octets': '89519572121',
                                                                                                'in_ucast_pkts': '67839621',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '79',
                                                                                                'out_octets': '4874593975',
                                                                                                'out_ucast_pkts': '16720634',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '79',
                                                                       'out_octets': '4874593975',
                                                                       'out_ucast_pkts': '16720634',
                                                                       'poller_hash': 'ODg4RjVwZWVyaW50ZnBvbiA2Mzg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '638',
                                                                                               'links': [       {       'connected_time': '688422',
                                                                                                                        'fiber_length': 1541,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg4RjVwZWVycG9ucG9uIDYzOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '99 Mill St. E. Milverton (C. '
                                                 'Winegarden-Bynsdrop 10430437)',
                                         'poller_hash': 'ODg4RjU=',
                                         'serial_number': '888F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:e9:dd',
                                         'uptime': '688422'},
                                 {       'counter_source': '639/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '639/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '255563362594',
                                                                       'in_ucast_pkts': '1437066921',
                                                                       'mac_address': 'ec:4f:82:0d:68:56',
                                                                       'name': '639/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '190',
                                                                       'out_octets': '7503618302919',
                                                                       'out_ucast_pkts': '5216999021',
                                                                       'poller_hash': 'MjA1REI5cGVlcmludGY2MzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '639',
                                                                       'in_nucast_pkts': '190',
                                                                       'in_octets': '7503618302919',
                                                                       'in_ucast_pkts': '5216999021',
                                                                       'links': [       {       'in_nucast_pkts': '190',
                                                                                                'in_octets': '7503618302919',
                                                                                                'in_ucast_pkts': '5216999021',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '255563362594',
                                                                                                'out_ucast_pkts': '1437066921',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '255563362594',
                                                                       'out_ucast_pkts': '1437066921',
                                                                       'poller_hash': 'MjA1REI5cGVlcmludGZwb24gNjM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '639',
                                                                                               'links': [       {       'connected_time': '1378150',
                                                                                                                        'fiber_length': 1728,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REI5cGVlcnBvbnBvbiA2Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '104 Mill St. E. Milverton (Doug Fyfe '
                                                 '10402350)',
                                         'poller_hash': 'MjA1REI5',
                                         'serial_number': '205DB9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:56',
                                         'uptime': '1378150'},
                                 {       'counter_source': '640/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:94:1a'],
                                                                       'id': '640/g1',
                                                                       'in_nucast_pkts': '299',
                                                                       'in_octets': '1034540548473',
                                                                       'in_ucast_pkts': '3628346778',
                                                                       'mac_address': '00:06:31:52:5c:5a',
                                                                       'mtu': 2000,
                                                                       'name': '640/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75653288',
                                                                       'out_octets': '13783821698849',
                                                                       'out_ucast_pkts': '10122595903',
                                                                       'poller_hash': 'ODg5QTVwZWVyaW50ZjY0MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '640',
                                                                       'in_nucast_pkts': '75653288',
                                                                       'in_octets': '13783821698849',
                                                                       'in_ucast_pkts': '10122595903',
                                                                       'links': [       {       'in_nucast_pkts': '75653288',
                                                                                                'in_octets': '13783821698849',
                                                                                                'in_ucast_pkts': '10122595903',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '299',
                                                                                                'out_octets': '1034540548473',
                                                                                                'out_ucast_pkts': '3628346778',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '299',
                                                                       'out_octets': '1034540548473',
                                                                       'out_ucast_pkts': '3628346778',
                                                                       'poller_hash': 'ODg5QTVwZWVyaW50ZnBvbiA2NDA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '640',
                                                                                               'links': [       {       'connected_time': '41227',
                                                                                                                        'fiber_length': 1560,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5QTVwZWVycG9ucG9uIDY0MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '97 Mill St. E. Milverton (Scott '
                                                 'Steckly 10402469)',
                                         'poller_hash': 'ODg5QTU=',
                                         'serial_number': '889A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:8d',
                                         'uptime': '41227'},
                                 {       'counter_source': '641/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '641/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '298',
                                                                       'in_octets': '20282351987',
                                                                       'in_ucast_pkts': '68904238',
                                                                       'mac_address': 'b8:94:70:a1:46:35',
                                                                       'name': '641/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1512',
                                                                       'out_octets': '522389069928',
                                                                       'out_ucast_pkts': '451027063',
                                                                       'poller_hash': 'MTE3REFEMnBlZXJpbnRmNjQxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '641',
                                                                       'in_nucast_pkts': '1512',
                                                                       'in_octets': '522389069928',
                                                                       'in_ucast_pkts': '451027063',
                                                                       'links': [       {       'in_nucast_pkts': '1512',
                                                                                                'in_octets': '522389069928',
                                                                                                'in_ucast_pkts': '451027063',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '298',
                                                                                                'out_octets': '20282351987',
                                                                                                'out_ucast_pkts': '68904238',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '298',
                                                                       'out_octets': '20282351987',
                                                                       'out_ucast_pkts': '68904238',
                                                                       'poller_hash': 'MTE3REFEMnBlZXJpbnRmcG9uIDY0MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '641',
                                                                                               'links': [       {       'connected_time': '688434',
                                                                                                                        'fiber_length': 1565,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFEMnBlZXJwb25wb24gNjQx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '94 Mill St. E. Milverton',
                                         'poller_hash': 'MTE3REFEMg==',
                                         'serial_number': '117DAD2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:46:35',
                                         'uptime': '688434'},
                                 {       'counter_source': '642/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '642/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '478307205437',
                                                                       'in_ucast_pkts': '4742986115',
                                                                       'mac_address': 'ec:4f:82:cc:23:8c',
                                                                       'name': '642/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '22814012295982',
                                                                       'out_ucast_pkts': '15354707335',
                                                                       'poller_hash': 'MzkwMUFBcGVlcmludGY2NDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '642',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '22814012295982',
                                                                       'in_ucast_pkts': '15354707335',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '22814012295982',
                                                                                                'in_ucast_pkts': '15354707335',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '478307205437',
                                                                                                'out_ucast_pkts': '4742986115',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '478307205437',
                                                                       'out_ucast_pkts': '4742986115',
                                                                       'poller_hash': 'MzkwMUFBcGVlcmludGZwb24gNjQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '642',
                                                                                               'links': [       {       'connected_time': '66989',
                                                                                                                        'fiber_length': 1452,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.304,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzkwMUFBcGVlcnBvbnBvbiA2NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '86 Mill St. E. Milverton (John '
                                                 'Spencer)',
                                         'poller_hash': 'MzkwMUFB',
                                         'serial_number': '3901AA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cc:23:8c',
                                         'uptime': '66989'},
                                 {       'counter_source': '643/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '50:91:e3:71:ac:20'],
                                                                       'id': '643/g1',
                                                                       'in_nucast_pkts': '2214',
                                                                       'in_octets': '15006706457',
                                                                       'in_ucast_pkts': '103709666',
                                                                       'mac_address': '00:06:31:69:f0:c5',
                                                                       'mtu': 2000,
                                                                       'name': '643/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119933',
                                                                       'out_octets': '396478702294',
                                                                       'out_ucast_pkts': '272198665',
                                                                       'poller_hash': 'QzU3OENwZWVyaW50ZjY0My9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '643',
                                                                       'in_nucast_pkts': '1119933',
                                                                       'in_octets': '396478702294',
                                                                       'in_ucast_pkts': '272198665',
                                                                       'links': [       {       'in_nucast_pkts': '1119933',
                                                                                                'in_octets': '396478702294',
                                                                                                'in_ucast_pkts': '272198665',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2214',
                                                                                                'out_octets': '15006706457',
                                                                                                'out_ucast_pkts': '103709666',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2214',
                                                                       'out_octets': '15006706457',
                                                                       'out_ucast_pkts': '103709666',
                                                                       'poller_hash': 'QzU3OENwZWVyaW50ZnBvbiA2NDM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '643',
                                                                                               'links': [       {       'connected_time': '688422',
                                                                                                                        'fiber_length': 1524,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3OENwZWVycG9ucG9uIDY0Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '101 MILL ST E (BAIER ADAM)',
                                         'poller_hash': 'QzU3OEM=',
                                         'serial_number': 'C578C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:68:4c',
                                         'uptime': '688422'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '644/g1',
                                                                       'name': '644/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'RDNFNEVwZWVyaW50ZjY0NC9nMQ=='},
                                                               {       'admin_up': False,
                                                                       'id': '644',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'RDNFNEVwZWVyaW50ZnBvbiA2NDQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '644',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RDNFNEVwZWVycG9ucG9uIDY0NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '111 Mill St. E. Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "644" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:12:02.420Z',
                                         'poller_hash': 'RDNFNEU=',
                                         'serial_number': 'D3E4E',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '645/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:9f:80:0b:99:69'],
                                                                       'id': '645/g1',
                                                                       'in_nucast_pkts': '98560',
                                                                       'in_octets': '39443458487',
                                                                       'in_ucast_pkts': '158726890',
                                                                       'mac_address': '00:06:31:a6:6f:1f',
                                                                       'mtu': 2000,
                                                                       'name': '645/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75672969',
                                                                       'out_octets': '454384746327',
                                                                       'out_ucast_pkts': '382257865',
                                                                       'poller_hash': 'MTMwMDUzcGVlcmludGY2NDUvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '645',
                                                                       'in_nucast_pkts': '75672969',
                                                                       'in_octets': '454384746327',
                                                                       'in_ucast_pkts': '382257865',
                                                                       'links': [       {       'in_nucast_pkts': '75672969',
                                                                                                'in_octets': '454384746327',
                                                                                                'in_ucast_pkts': '382257865',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '98560',
                                                                                                'out_octets': '39443458487',
                                                                                                'out_ucast_pkts': '158726890',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '98560',
                                                                       'out_octets': '39443458487',
                                                                       'out_ucast_pkts': '158726890',
                                                                       'poller_hash': 'MTMwMDUzcGVlcmludGZwb24gNjQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '645',
                                                                                               'links': [       {       'connected_time': '41147',
                                                                                                                        'fiber_length': 1436,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDUzcGVlcnBvbnBvbiA2NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '90 Mill St. E. Milverton (Debbie '
                                                 'Tanner)',
                                         'poller_hash': 'MTMwMDUz',
                                         'serial_number': '130053',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:13',
                                         'uptime': '41147'},
                                 {       'counter_source': '646/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '646/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '479987549883',
                                                                       'in_ucast_pkts': '1804667832',
                                                                       'mac_address': '00:06:31:a6:6d:eb',
                                                                       'name': '646/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7765981897299',
                                                                       'out_ucast_pkts': '5653769712',
                                                                       'poller_hash': 'MTMwMDI2cGVlcmludGY2NDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '646',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7765981897299',
                                                                       'in_ucast_pkts': '5653769712',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7765981897299',
                                                                                                'in_ucast_pkts': '5653769712',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '479987549883',
                                                                                                'out_ucast_pkts': '1804667832',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '479987549883',
                                                                       'out_ucast_pkts': '1804667832',
                                                                       'poller_hash': 'MTMwMDI2cGVlcmludGZwb24gNjQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '646',
                                                                                               'links': [       {       'connected_time': '15498581',
                                                                                                                        'fiber_length': 1535,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDI2cGVlcnBvbnBvbiA2NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '98 Mill St. E. Milverton (Krystina '
                                                 'Stachelski 10401363)',
                                         'poller_hash': 'MTMwMDI2',
                                         'serial_number': '130026',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:e6',
                                         'uptime': '15498581'},
                                 {       'counter_source': '647/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1b:8b'],
                                                                       'id': '647/g1',
                                                                       'in_nucast_pkts': '664844',
                                                                       'in_octets': '528591584342',
                                                                       'in_ucast_pkts': '2624798477',
                                                                       'mac_address': '00:06:31:a6:6a:a7',
                                                                       'mtu': 2000,
                                                                       'name': '647/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66209265',
                                                                       'out_octets': '12653108376390',
                                                                       'out_ucast_pkts': '9035744537',
                                                                       'poller_hash': 'MTJGRkQ0cGVlcmludGY2NDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '647',
                                                                       'in_nucast_pkts': '66209265',
                                                                       'in_octets': '12653108376390',
                                                                       'in_ucast_pkts': '9035744537',
                                                                       'links': [       {       'in_nucast_pkts': '66209265',
                                                                                                'in_octets': '12653108376390',
                                                                                                'in_ucast_pkts': '9035744537',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '664844',
                                                                                                'out_octets': '528591584342',
                                                                                                'out_ucast_pkts': '2624798477',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '664844',
                                                                       'out_octets': '528591584342',
                                                                       'out_ucast_pkts': '2624798477',
                                                                       'poller_hash': 'MTJGRkQ0cGVlcmludGZwb24gNjQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '647',
                                                                                               'links': [       {       'connected_time': '68194077',
                                                                                                                        'fiber_length': 1506,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTJGRkQ0cGVlcnBvbnBvbiA2NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '84 Mill St. E. Milverton (Sandi '
                                                 'Jacobs 10403134)',
                                         'poller_hash': 'MTJGRkQ0',
                                         'serial_number': '12FFD4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:94',
                                         'uptime': '68194077'},
                                 {       'counter_source': '648/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7d:d0'],
                                                                       'id': '648/g1',
                                                                       'in_nucast_pkts': '12208',
                                                                       'in_octets': '7155963837',
                                                                       'in_ucast_pkts': '54331719',
                                                                       'mac_address': '00:06:31:a6:72:af',
                                                                       'mtu': 2000,
                                                                       'name': '648/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119711',
                                                                       'out_octets': '299594896861',
                                                                       'out_ucast_pkts': '213818761',
                                                                       'poller_hash': 'MTMwMDhFcGVlcmludGY2NDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '648',
                                                                       'in_nucast_pkts': '1119711',
                                                                       'in_octets': '299594896861',
                                                                       'in_ucast_pkts': '213818761',
                                                                       'links': [       {       'in_nucast_pkts': '1119711',
                                                                                                'in_octets': '299594896861',
                                                                                                'in_ucast_pkts': '213818761',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12208',
                                                                                                'out_octets': '7155963837',
                                                                                                'out_ucast_pkts': '54331719',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12208',
                                                                       'out_octets': '7155963837',
                                                                       'out_ucast_pkts': '54331719',
                                                                       'poller_hash': 'MTMwMDhFcGVlcmludGZwb24gNjQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '648',
                                                                                               'links': [       {       'connected_time': '688431',
                                                                                                                        'fiber_length': 1513,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDhFcGVlcnBvbnBvbiA2NDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '96 MILL ST E (MORWOOD NATHAN)',
                                         'poller_hash': 'MTMwMDhF',
                                         'serial_number': '13008E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:4e',
                                         'uptime': '688431'},
                                 {       'counter_source': '649/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0f:93'],
                                                                       'id': '649/g1',
                                                                       'in_nucast_pkts': '321',
                                                                       'in_octets': '157632738659',
                                                                       'in_ucast_pkts': '691821891',
                                                                       'mac_address': '00:06:31:a6:65:77',
                                                                       'mtu': 2000,
                                                                       'name': '649/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75606886',
                                                                       'out_octets': '3637548601275',
                                                                       'out_ucast_pkts': '2581684114',
                                                                       'poller_hash': 'MTJGRkYzcGVlcmludGY2NDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '649',
                                                                       'in_nucast_pkts': '75606886',
                                                                       'in_octets': '3637548601275',
                                                                       'in_ucast_pkts': '2581684114',
                                                                       'links': [       {       'in_nucast_pkts': '75606886',
                                                                                                'in_octets': '3637548601275',
                                                                                                'in_ucast_pkts': '2581684114',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '321',
                                                                                                'out_octets': '157632738659',
                                                                                                'out_ucast_pkts': '691821891',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '321',
                                                                       'out_octets': '157632738659',
                                                                       'out_ucast_pkts': '691821891',
                                                                       'poller_hash': 'MTJGRkYzcGVlcmludGZwb24gNjQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '649',
                                                                                               'links': [       {       'connected_time': '41155',
                                                                                                                        'fiber_length': 1499,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTJGRkYzcGVlcnBvbnBvbiA2NDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '92 MILL ST E (GRAUL SCHERI)',
                                         'poller_hash': 'MTJGRkYz',
                                         'serial_number': '12FFF3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:b3',
                                         'uptime': '41155'},
                                 {       'counter_source': '650/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ed:bb'],
                                                                       'id': '650/g1',
                                                                       'in_nucast_pkts': '903278',
                                                                       'in_octets': '1779629366949',
                                                                       'in_ucast_pkts': '5537064039',
                                                                       'mac_address': '00:06:31:e9:01:9c',
                                                                       'mtu': 2000,
                                                                       'name': '650/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '130014912',
                                                                       'out_octets': '20869696689404',
                                                                       'out_ucast_pkts': '15769619048',
                                                                       'poller_hash': 'MUQwNjhCcGVlcmludGY2NTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '650',
                                                                       'in_nucast_pkts': '130014912',
                                                                       'in_octets': '20869696689404',
                                                                       'in_ucast_pkts': '15769619048',
                                                                       'links': [       {       'in_nucast_pkts': '130014912',
                                                                                                'in_octets': '20869696689404',
                                                                                                'in_ucast_pkts': '15769619048',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '903278',
                                                                                                'out_octets': '1779629366949',
                                                                                                'out_ucast_pkts': '5537064039',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '903278',
                                                                       'out_octets': '1779629366949',
                                                                       'out_ucast_pkts': '5537064039',
                                                                       'poller_hash': 'MUQwNjhCcGVlcmludGZwb24gNjUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '650',
                                                                                               'links': [       {       'connected_time': '21817168',
                                                                                                                        'fiber_length': 1785,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjhCcGVlcnBvbnBvbiA2NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '112 Mill St. E. Milverton (Earl '
                                                 'Leis)',
                                         'poller_hash': 'MUQwNjhC',
                                         'serial_number': '1D068B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:99',
                                         'uptime': '21817168'},
                                 {       'counter_source': '651/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '651/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '124006365140',
                                                                       'in_ucast_pkts': '1518939787',
                                                                       'mac_address': '00:06:31:e9:01:bd',
                                                                       'name': '651/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '110',
                                                                       'out_octets': '9393693034249',
                                                                       'out_ucast_pkts': '6255264782',
                                                                       'poller_hash': 'MUQwNjhGcGVlcmludGY2NTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '651',
                                                                       'in_nucast_pkts': '110',
                                                                       'in_octets': '9393693034249',
                                                                       'in_ucast_pkts': '6255264782',
                                                                       'links': [       {       'in_nucast_pkts': '110',
                                                                                                'in_octets': '9393693034249',
                                                                                                'in_ucast_pkts': '6255264782',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '124006365140',
                                                                                                'out_ucast_pkts': '1518939787',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '124006365140',
                                                                       'out_ucast_pkts': '1518939787',
                                                                       'poller_hash': 'MUQwNjhGcGVlcmludGZwb24gNjUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '651',
                                                                                               'links': [       {       'connected_time': '1052437',
                                                                                                                        'fiber_length': 1856,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjhGcGVlcnBvbnBvbiA2NTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '108 Mill St. E. Milverton (Larry '
                                                 'Kropf)',
                                         'poller_hash': 'MUQwNjhG',
                                         'serial_number': '1D068F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:bd',
                                         'uptime': '1052437'},
                                 {       'counter_source': '652/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:9a:18:18:7f:b0'],
                                                                       'id': '652/g1',
                                                                       'in_nucast_pkts': '1206703',
                                                                       'in_octets': '878178065266',
                                                                       'in_ucast_pkts': '2178269076',
                                                                       'mac_address': 'cc:be:59:eb:01:52',
                                                                       'mtu': 2000,
                                                                       'name': '652/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '40279638',
                                                                       'out_octets': '4388017432911',
                                                                       'out_ucast_pkts': '3758449040',
                                                                       'poller_hash': 'NTU1QzlGcGVlcmludGY2NTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '652',
                                                                       'in_nucast_pkts': '40279638',
                                                                       'in_octets': '4388017432911',
                                                                       'in_ucast_pkts': '3758449040',
                                                                       'links': [       {       'in_nucast_pkts': '40279638',
                                                                                                'in_octets': '4388017432911',
                                                                                                'in_ucast_pkts': '3758449040',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1206703',
                                                                                                'out_octets': '878178065266',
                                                                                                'out_ucast_pkts': '2178269076',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1206703',
                                                                       'out_octets': '878178065266',
                                                                       'out_ucast_pkts': '2178269076',
                                                                       'poller_hash': 'NTU1QzlGcGVlcmludGZwb24gNjUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '652',
                                                                                               'links': [       {       'connected_time': '24135479',
                                                                                                                        'fiber_length': 2162,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.636,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU1QzlGcGVlcnBvbnBvbiA2NTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '6380 Perth Rd 131 (MASONAL STONE)',
                                         'poller_hash': 'NTU1QzlG',
                                         'serial_number': '555C9F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:01:4e',
                                         'uptime': '24135479'},
                                 {       'counter_source': '653/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:f0:12'],
                                                                       'id': '653/g1',
                                                                       'in_nucast_pkts': '18',
                                                                       'in_octets': '6477136642',
                                                                       'in_ucast_pkts': '31125299',
                                                                       'mac_address': 'cc:be:59:cb:6e:2d',
                                                                       'mtu': 2000,
                                                                       'name': '653/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447193',
                                                                       'out_octets': '238612823926',
                                                                       'out_ucast_pkts': '172371251',
                                                                       'poller_hash': 'NDJCRjFEcGVlcmludGY2NTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '653',
                                                                       'in_nucast_pkts': '1447193',
                                                                       'in_octets': '238612823926',
                                                                       'in_ucast_pkts': '172371251',
                                                                       'links': [       {       'in_nucast_pkts': '1447193',
                                                                                                'in_octets': '238612823926',
                                                                                                'in_ucast_pkts': '172371251',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18',
                                                                                                'out_octets': '6477136642',
                                                                                                'out_ucast_pkts': '31125299',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18',
                                                                       'out_octets': '6477136642',
                                                                       'out_ucast_pkts': '31125299',
                                                                       'poller_hash': 'NDJCRjFEcGVlcmludGZwb24gNjUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '653',
                                                                                               'links': [       {       'connected_time': '688383',
                                                                                                                        'fiber_length': 7409,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjFEcGVlcnBvbnBvbiA2NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6238 Road 140 Milverton (Rick Diehl '
                                                 '10404130)',
                                         'poller_hash': 'NDJCRjFE',
                                         'serial_number': '42BF1D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6e:2a',
                                         'uptime': '688383'},
                                 {       'counter_source': '654/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '654/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '463848705454',
                                                                       'in_ucast_pkts': '1847229686',
                                                                       'mac_address': '00:06:31:69:b5:65',
                                                                       'name': '654/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6295302986227',
                                                                       'out_ucast_pkts': '4708800372',
                                                                       'poller_hash': 'QzUwMzlwZWVyaW50ZjY1NC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '654',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6295302986227',
                                                                       'in_ucast_pkts': '4708800372',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6295302986227',
                                                                                                'in_ucast_pkts': '4708800372',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '463848705454',
                                                                                                'out_ucast_pkts': '1847229686',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '463848705454',
                                                                       'out_ucast_pkts': '1847229686',
                                                                       'poller_hash': 'QzUwMzlwZWVyaW50ZnBvbiA2NTQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '654',
                                                                                               'links': [       {       'connected_time': '688420',
                                                                                                                        'fiber_length': 1146,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzUwMzlwZWVycG9ucG9uIDY1NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '81 Mill St. E. Milverton (Brian '
                                                 'Hall)',
                                         'poller_hash': 'QzUwMzk=',
                                         'serial_number': 'C5039',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:60:f9',
                                         'uptime': '688420'},
                                 {       'counter_source': '655/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '655/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '254495743586',
                                                                       'in_ucast_pkts': '44888386371',
                                                                       'mac_address': 'ec:4f:82:0d:68:b0',
                                                                       'name': '655/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '109',
                                                                       'out_octets': '7944747759988',
                                                                       'out_ucast_pkts': '5493642657',
                                                                       'poller_hash': 'MjA1REMzcGVlcmludGY2NTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '655',
                                                                       'in_nucast_pkts': '109',
                                                                       'in_octets': '7944747759988',
                                                                       'in_ucast_pkts': '5493642657',
                                                                       'links': [       {       'in_nucast_pkts': '109',
                                                                                                'in_octets': '7944747759988',
                                                                                                'in_ucast_pkts': '5493642657',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '254495743586',
                                                                                                'out_ucast_pkts': '44888386371',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '254495743586',
                                                                       'out_ucast_pkts': '44888386371',
                                                                       'poller_hash': 'MjA1REMzcGVlcmludGZwb24gNjU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '655',
                                                                                               'links': [       {       'connected_time': '24525703',
                                                                                                                        'fiber_length': 1424,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REMzcGVlcnBvbnBvbiA2NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '78 MILL ST E (CARTER ROGER & '
                                                 'MICHELLE)',
                                         'poller_hash': 'MjA1REMz',
                                         'serial_number': '205DC3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:b0',
                                         'uptime': '24525703'},
                                 {       'counter_source': '656/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2b:47:fd'],
                                                                       'id': '656/g1',
                                                                       'in_nucast_pkts': '33230',
                                                                       'in_octets': '9718321460',
                                                                       'in_ucast_pkts': '59451736',
                                                                       'mac_address': '00:06:31:52:61:6a',
                                                                       'mtu': 2000,
                                                                       'name': '656/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1153333',
                                                                       'out_octets': '178022522861',
                                                                       'out_ucast_pkts': '124840747',
                                                                       'poller_hash': 'ODg5RTFwZWVyaW50ZjY1Ni9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '656',
                                                                       'in_nucast_pkts': '1153333',
                                                                       'in_octets': '178022522861',
                                                                       'in_ucast_pkts': '124840747',
                                                                       'links': [       {       'in_nucast_pkts': '1153333',
                                                                                                'in_octets': '178022522861',
                                                                                                'in_ucast_pkts': '124840747',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '33230',
                                                                                                'out_octets': '9718321460',
                                                                                                'out_ucast_pkts': '59451736',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33230',
                                                                       'out_octets': '9718321460',
                                                                       'out_ucast_pkts': '59451736',
                                                                       'poller_hash': 'ODg5RTFwZWVyaW50ZnBvbiA2NTY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '656',
                                                                                               'links': [       {       'connected_time': '688422',
                                                                                                                        'fiber_length': 1180,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RTFwZWVycG9ucG9uIDY1Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '83 MILL ST E (PAJOT MIKE)',
                                         'poller_hash': 'ODg5RTE=',
                                         'serial_number': '889E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:c9',
                                         'uptime': '688422'},
                                 {       'counter_source': '657/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '657/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '241122852245',
                                                                       'in_ucast_pkts': '2366759624',
                                                                       'mac_address': '00:06:31:52:41:9e',
                                                                       'name': '657/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '13078415298323',
                                                                       'out_ucast_pkts': '8864229251',
                                                                       'poller_hash': 'ODcwNEZwZWVyaW50ZjY1Ny9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '657',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '13078415298323',
                                                                       'in_ucast_pkts': '8864229251',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '13078415298323',
                                                                                                'in_ucast_pkts': '8864229251',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '241122852245',
                                                                                                'out_ucast_pkts': '2366759624',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '241122852245',
                                                                       'out_ucast_pkts': '2366759624',
                                                                       'poller_hash': 'ODcwNEZwZWVyaW50ZnBvbiA2NTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '657',
                                                                                               'links': [       {       'connected_time': '31369146',
                                                                                                                        'fiber_length': 1174,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcwNEZwZWVycG9ucG9uIDY1Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '71 Mill St. E. Milverton (Valerie '
                                                 'Yost 10430631)',
                                         'poller_hash': 'ODcwNEY=',
                                         'serial_number': '8704F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:d1:37',
                                         'uptime': '31369146'},
                                 {       'counter_source': '658/F1+658/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:db:2f:90',
                                                                                                        '00:03:e6:db:30:de'],
                                                                       'id': '658/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'name': '658/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'OEEyMzFwZWVyaW50ZjY1OC9GMQ=='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:95:9d'],
                                                                       'id': '658/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '1409935644',
                                                                       'in_ucast_pkts': '6802703',
                                                                       'mac_address': '00:06:31:52:7a:a2',
                                                                       'mtu': 2000,
                                                                       'name': '658/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119718',
                                                                       'out_octets': '42421706427',
                                                                       'out_ucast_pkts': '31350662',
                                                                       'poller_hash': 'OEEyMzFwZWVyaW50ZjY1OC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '658',
                                                                       'in_nucast_pkts': '1119718',
                                                                       'in_octets': '42421706427',
                                                                       'in_ucast_pkts': '31350662',
                                                                       'links': [       {       'in_nucast_pkts': '1119718',
                                                                                                'in_octets': '42421706427',
                                                                                                'in_ucast_pkts': '31350662',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '1409935644',
                                                                                                'out_ucast_pkts': '6802703',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '1409935644',
                                                                       'out_ucast_pkts': '6802703',
                                                                       'poller_hash': 'OEEyMzFwZWVyaW50ZnBvbiA2NTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '658',
                                                                                               'links': [       {       'connected_time': '688417',
                                                                                                                        'fiber_length': 1108,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEEyMzFwZWVycG9ucG9uIDY1OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '77 Mill St. E. Milverton (Josh '
                                                 'Ludwig, 10431732)',
                                         'poller_hash': 'OEEyMzE=',
                                         'serial_number': '8A231',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:47:4b:1d',
                                         'uptime': '688417'},
                                 {       'counter_source': '659/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '659/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '148031049512',
                                                                       'in_ucast_pkts': '945444811',
                                                                       'mac_address': '00:06:31:52:50:1e',
                                                                       'name': '659/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4589877883502',
                                                                       'out_ucast_pkts': '3486795621',
                                                                       'poller_hash': 'ODg4RkFwZWVyaW50ZjY1OS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '659',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4589877883502',
                                                                       'in_ucast_pkts': '3486795621',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4589877883502',
                                                                                                'in_ucast_pkts': '3486795621',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '148031049512',
                                                                                                'out_ucast_pkts': '945444811',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '148031049512',
                                                                       'out_ucast_pkts': '945444811',
                                                                       'poller_hash': 'ODg4RkFwZWVyaW50ZnBvbiA2NTk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '659',
                                                                                               'links': [       {       'connected_time': '47349',
                                                                                                                        'fiber_length': 1088,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg4RkFwZWVycG9ucG9uIDY1OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '75 Mill St. E. Milverton (Louise '
                                                 'Campbell)',
                                         'poller_hash': 'ODg4RkE=',
                                         'serial_number': '888FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:e9:e2',
                                         'uptime': '47349'},
                                 {       'counter_source': '660/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:2e:73'],
                                                                       'id': '660/g1',
                                                                       'in_nucast_pkts': '806560',
                                                                       'in_octets': '356485635191',
                                                                       'in_ucast_pkts': '2848258929',
                                                                       'mac_address': '00:06:31:52:62:e2',
                                                                       'mtu': 2000,
                                                                       'name': '660/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75587292',
                                                                       'out_octets': '20724113740091',
                                                                       'out_ucast_pkts': '14165794585',
                                                                       'poller_hash': 'OEUxMDlwZWVyaW50ZjY2MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '660',
                                                                       'in_nucast_pkts': '75587292',
                                                                       'in_octets': '20724113740091',
                                                                       'in_ucast_pkts': '14165794585',
                                                                       'links': [       {       'in_nucast_pkts': '75587292',
                                                                                                'in_octets': '20724113740091',
                                                                                                'in_ucast_pkts': '14165794585',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '806560',
                                                                                                'out_octets': '356485635191',
                                                                                                'out_ucast_pkts': '2848258929',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '806560',
                                                                       'out_octets': '356485635191',
                                                                       'out_ucast_pkts': '2848258929',
                                                                       'poller_hash': 'OEUxMDlwZWVyaW50ZnBvbiA2NjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '660',
                                                                                               'links': [       {       'connected_time': '41195',
                                                                                                                        'fiber_length': 1187,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEUxMDlwZWVycG9ucG9uIDY2MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '69 Mill St. E. Milverton (Lois '
                                                 'McEwin 10401294)',
                                         'poller_hash': 'OEUxMDk=',
                                         'serial_number': '8E109',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ef:6f',
                                         'uptime': '41195'},
                                 {       'counter_source': '662/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b0:6e:bf:95:30:18'],
                                                                       'id': '662/g1',
                                                                       'in_nucast_pkts': '75',
                                                                       'in_octets': '13414369954',
                                                                       'in_ucast_pkts': '74204486',
                                                                       'mac_address': '00:06:31:52:41:32',
                                                                       'mtu': 2000,
                                                                       'name': '662/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23529126',
                                                                       'out_octets': '483338723217',
                                                                       'out_ucast_pkts': '349102950',
                                                                       'poller_hash': 'ODg1RTVwZWVyaW50ZjY2Mi9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '662',
                                                                       'in_nucast_pkts': '23529126',
                                                                       'in_octets': '483338723217',
                                                                       'in_ucast_pkts': '349102950',
                                                                       'links': [       {       'in_nucast_pkts': '23529126',
                                                                                                'in_octets': '483338723217',
                                                                                                'in_ucast_pkts': '349102950',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '75',
                                                                                                'out_octets': '13414369954',
                                                                                                'out_ucast_pkts': '74204486',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75',
                                                                       'out_octets': '13414369954',
                                                                       'out_ucast_pkts': '74204486',
                                                                       'poller_hash': 'ODg1RTVwZWVyaW50ZnBvbiA2NjI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '662',
                                                                                               'links': [       {       'connected_time': '13219277',
                                                                                                                        'fiber_length': 1122,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg1RTVwZWVycG9ucG9uIDY2Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '73 Mill St. E. Milverton (Theresa '
                                                 'Hopf)',
                                         'poller_hash': 'ODg1RTU=',
                                         'serial_number': '885E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:e6:cd',
                                         'uptime': '13219277'},
                                 {       'counter_source': '663/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '663/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11853116',
                                                                       'in_ucast_pkts': '43940',
                                                                       'mac_address': '00:06:31:52:3f:e6',
                                                                       'name': '663/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '18209624269',
                                                                       'out_ucast_pkts': '80151115',
                                                                       'poller_hash': 'ODg4OEZwZWVyaW50ZjY2My9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '663',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '18209624269',
                                                                       'in_ucast_pkts': '80151115',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '18209624269',
                                                                                                'in_ucast_pkts': '80151115',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '11853116',
                                                                                                'out_ucast_pkts': '43940',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11853116',
                                                                       'out_ucast_pkts': '43940',
                                                                       'poller_hash': 'ODg4OEZwZWVyaW50ZnBvbiA2NjM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '663',
                                                                                               'links': [       {       'connected_time': '688422',
                                                                                                                        'fiber_length': 1300,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg4OEZwZWVycG9ucG9uIDY2Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '93 Mill St. E. Milverton (Dan '
                                                 'Crowley)',
                                         'poller_hash': 'ODg4OEY=',
                                         'serial_number': '8888F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:e9:77',
                                         'uptime': '688422'},
                                 {       'counter_source': '664/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '664/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '52575376844',
                                                                       'in_ucast_pkts': '111159679',
                                                                       'mac_address': 'ec:4f:82:5b:70:7b',
                                                                       'name': '664/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '128',
                                                                       'out_octets': '408971399515',
                                                                       'out_ucast_pkts': '380087245',
                                                                       'poller_hash': 'MkIwMTgycGVlcmludGY2NjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '664',
                                                                       'in_nucast_pkts': '128',
                                                                       'in_octets': '408971399515',
                                                                       'in_ucast_pkts': '380087245',
                                                                       'links': [       {       'in_nucast_pkts': '128',
                                                                                                'in_octets': '408971399515',
                                                                                                'in_ucast_pkts': '380087245',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '52575376844',
                                                                                                'out_ucast_pkts': '111159679',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '52575376844',
                                                                       'out_ucast_pkts': '111159679',
                                                                       'poller_hash': 'MkIwMTgycGVlcmludGZwb24gNjY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '664',
                                                                                               'links': [       {       'connected_time': '688434',
                                                                                                                        'fiber_length': 1375,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTgycGVlcnBvbnBvbiA2NjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '78 Mill St. E. Milverton (MARTIN '
                                                 'CHADRICK)',
                                         'poller_hash': 'MkIwMTgy',
                                         'serial_number': '2B0182',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:7b',
                                         'uptime': '688434'},
                                 {       'counter_source': '665/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cc:54'],
                                                                       'id': '665/g1',
                                                                       'in_nucast_pkts': '449823',
                                                                       'in_octets': '399463216665',
                                                                       'in_ucast_pkts': '3980416604',
                                                                       'mac_address': '00:06:31:52:58:92',
                                                                       'mtu': 2000,
                                                                       'name': '665/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '47266953',
                                                                       'out_octets': '27813794425205',
                                                                       'out_ucast_pkts': '18700047395',
                                                                       'poller_hash': 'ODg5RUFwZWVyaW50ZjY2NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '665',
                                                                       'in_nucast_pkts': '47266953',
                                                                       'in_octets': '27813794425205',
                                                                       'in_ucast_pkts': '18700047395',
                                                                       'links': [       {       'in_nucast_pkts': '47266953',
                                                                                                'in_octets': '27813794425205',
                                                                                                'in_ucast_pkts': '18700047395',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '449823',
                                                                                                'out_octets': '399463216665',
                                                                                                'out_ucast_pkts': '3980416604',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '449823',
                                                                       'out_octets': '399463216665',
                                                                       'out_ucast_pkts': '3980416604',
                                                                       'poller_hash': 'ODg5RUFwZWVyaW50ZnBvbiA2NjU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '665',
                                                                                               'links': [       {       'connected_time': '25137186',
                                                                                                                        'fiber_length': 1342,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RUFwZWVycG9ucG9uIDY2NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '87 Mill St. E. Milverton (David '
                                                 'Kuehn)',
                                         'poller_hash': 'ODg5RUE=',
                                         'serial_number': '889EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:d2',
                                         'uptime': '25137186'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '666/g1',
                                                                       'name': '666/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'M0NGQTlwZWVyaW50ZjY2Ni9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '666',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'M0NGQTlwZWVyaW50ZnBvbiA2NjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '666',
                                                                                               'links': [       {       'connected_time': '11453878',
                                                                                                                        'fiber_length': 1229,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0NGQTlwZWVycG9ucG9uIDY2Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712GX',
                                         'name': '95 Mill St. E. Milverton (Pumphouse '
                                                 'Perth East)',
                                         'poller_hash': 'M0NGQTk=',
                                         'serial_number': '3CFA9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:18:c3:a4',
                                         'uptime': '11453878'},
                                 {       'counter_source': '667/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7c:bd'],
                                                                       'id': '667/g1',
                                                                       'in_nucast_pkts': '393190',
                                                                       'in_octets': '294333068606',
                                                                       'in_ucast_pkts': '1225354534',
                                                                       'mac_address': '00:06:31:a6:6b:0f',
                                                                       'mtu': 2000,
                                                                       'name': '667/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38333668',
                                                                       'out_octets': '7415658221991',
                                                                       'out_ucast_pkts': '5215326192',
                                                                       'poller_hash': 'MTMwMDA2cGVlcmludGY2NjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '667',
                                                                       'in_nucast_pkts': '38333668',
                                                                       'in_octets': '7415658221991',
                                                                       'in_ucast_pkts': '5215326192',
                                                                       'links': [       {       'in_nucast_pkts': '38333668',
                                                                                                'in_octets': '7415658221991',
                                                                                                'in_ucast_pkts': '5215326192',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '393190',
                                                                                                'out_octets': '294333068606',
                                                                                                'out_ucast_pkts': '1225354534',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '393190',
                                                                       'out_octets': '294333068606',
                                                                       'out_ucast_pkts': '1225354534',
                                                                       'poller_hash': 'MTMwMDA2cGVlcmludGZwb24gNjY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '667',
                                                                                               'links': [       {       'connected_time': '20919986',
                                                                                                                        'fiber_length': 1559,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDA2cGVlcnBvbnBvbiA2Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '86A MILL ST E (CAMPBELL MIKE & '
                                                 'TAMMY)',
                                         'poller_hash': 'MTMwMDA2',
                                         'serial_number': '130006',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f0:c6',
                                         'uptime': '20919986'},
                                 {       'counter_source': '668/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '668/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '101127006089',
                                                                       'in_ucast_pkts': '1296996568',
                                                                       'mac_address': '00:06:31:a6:67:23',
                                                                       'name': '668/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7896770905820',
                                                                       'out_ucast_pkts': '5255867099',
                                                                       'poller_hash': 'MTMwMEM2cGVlcmludGY2NjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '668',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7896770905820',
                                                                       'in_ucast_pkts': '5255867099',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7896770905820',
                                                                                                'in_ucast_pkts': '5255867099',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '101127006089',
                                                                                                'out_ucast_pkts': '1296996568',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '101127006089',
                                                                       'out_ucast_pkts': '1296996568',
                                                                       'poller_hash': 'MTMwMEM2cGVlcmludGZwb24gNjY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '668',
                                                                                               'links': [       {       'connected_time': '31296076',
                                                                                                                        'fiber_length': 1329,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMEM2cGVlcnBvbnBvbiA2Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '76 Mill St. E. Milverton (Helena '
                                                 'Mulder)',
                                         'poller_hash': 'MTMwMEM2',
                                         'serial_number': '1300C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:86',
                                         'uptime': '31296076'},
                                 {       'counter_source': '669/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d4:f8'],
                                                                       'id': '669/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '5161899673',
                                                                       'in_ucast_pkts': '20135056',
                                                                       'mac_address': '00:06:31:a6:65:f7',
                                                                       'mtu': 2000,
                                                                       'name': '669/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119708',
                                                                       'out_octets': '94077149417',
                                                                       'out_ucast_pkts': '69964685',
                                                                       'poller_hash': 'MTMwMDlGcGVlcmludGY2NjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '669',
                                                                       'in_nucast_pkts': '1119708',
                                                                       'in_octets': '94077149417',
                                                                       'in_ucast_pkts': '69964685',
                                                                       'links': [       {       'in_nucast_pkts': '1119708',
                                                                                                'in_octets': '94077149417',
                                                                                                'in_ucast_pkts': '69964685',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '5161899673',
                                                                                                'out_ucast_pkts': '20135056',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '5161899673',
                                                                       'out_ucast_pkts': '20135056',
                                                                       'poller_hash': 'MTMwMDlGcGVlcmludGZwb24gNjY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '669',
                                                                                               'links': [       {       'connected_time': '688430',
                                                                                                                        'fiber_length': 1401,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDlGcGVlcnBvbnBvbiA2Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '82 Mill St. E. Milverton (Craig '
                                                 'Kipfer)',
                                         'poller_hash': 'MTMwMDlG',
                                         'serial_number': '13009F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:5f',
                                         'uptime': '688430'},
                                 {       'counter_source': '670/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:f1:56'],
                                                                       'id': '670/g1',
                                                                       'in_nucast_pkts': '5',
                                                                       'in_octets': '998324980',
                                                                       'in_ucast_pkts': '5760836',
                                                                       'mac_address': '00:06:31:a6:67:2f',
                                                                       'mtu': 2000,
                                                                       'name': '670/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119700',
                                                                       'out_octets': '32939887123',
                                                                       'out_ucast_pkts': '22824330',
                                                                       'poller_hash': 'MTMwMENCcGVlcmludGY2NzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '670',
                                                                       'in_nucast_pkts': '1119700',
                                                                       'in_octets': '32939887123',
                                                                       'in_ucast_pkts': '22824330',
                                                                       'links': [       {       'in_nucast_pkts': '1119700',
                                                                                                'in_octets': '32939887123',
                                                                                                'in_ucast_pkts': '22824330',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5',
                                                                                                'out_octets': '998324980',
                                                                                                'out_ucast_pkts': '5760836',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5',
                                                                       'out_octets': '998324980',
                                                                       'out_ucast_pkts': '5760836',
                                                                       'poller_hash': 'MTMwMENCcGVlcmludGZwb24gNjcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '670',
                                                                                               'links': [       {       'connected_time': '688431',
                                                                                                                        'fiber_length': 1296,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMENCcGVlcnBvbnBvbiA2NzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '76 MILL ST E (RICHARDS ALEXIS)',
                                         'poller_hash': 'MTMwMENC',
                                         'serial_number': '1300CB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:8b',
                                         'uptime': '688431'},
                                 {       'counter_source': '672/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:4c:a5:6d:0c:b0'],
                                                                       'id': '672/g4',
                                                                       'in_nucast_pkts': '21381',
                                                                       'in_octets': '3864660766772',
                                                                       'in_ucast_pkts': '6801131188',
                                                                       'mac_address': 'ec:4f:82:34:63:56',
                                                                       'mtu': 2000,
                                                                       'name': '672/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27208229',
                                                                       'out_octets': '13899593884188',
                                                                       'out_ucast_pkts': '12554649569',
                                                                       'poller_hash': 'MUZENkI3cGVlcmludGY2NzIvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '672',
                                                                       'in_nucast_pkts': '27208229',
                                                                       'in_octets': '13899593884188',
                                                                       'in_ucast_pkts': '12554649569',
                                                                       'links': [       {       'in_nucast_pkts': '27208229',
                                                                                                'in_octets': '13899593884188',
                                                                                                'in_ucast_pkts': '12554649569',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21381',
                                                                                                'out_octets': '3864660766772',
                                                                                                'out_ucast_pkts': '6801131188',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21381',
                                                                       'out_octets': '3864660766772',
                                                                       'out_ucast_pkts': '6801131188',
                                                                       'poller_hash': 'MUZENkI3cGVlcmludGZwb24gNjcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '672',
                                                                                               'links': [       {       'connected_time': '15498304',
                                                                                                                        'fiber_length': 1008,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.542,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUZENkI3cGVlcnBvbnBvbiA2NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '68 Mill St. E. Milverton (Milverton '
                                                 'Public School)',
                                         'poller_hash': 'MUZENkI3',
                                         'serial_number': '1FD6B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:34:63:4f',
                                         'uptime': '15498304'},
                                 {       'counter_source': '673/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '673/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11828950956',
                                                                       'in_ucast_pkts': '136723052',
                                                                       'mac_address': '00:06:31:e9:03:d9',
                                                                       'name': '673/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '158',
                                                                       'out_octets': '604893281736',
                                                                       'out_ucast_pkts': '448807159',
                                                                       'poller_hash': 'MUQwNkNEcGVlcmludGY2NzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '673',
                                                                       'in_nucast_pkts': '158',
                                                                       'in_octets': '604893281736',
                                                                       'in_ucast_pkts': '448807159',
                                                                       'links': [       {       'in_nucast_pkts': '158',
                                                                                                'in_octets': '604893281736',
                                                                                                'in_ucast_pkts': '448807159',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '11828950956',
                                                                                                'out_ucast_pkts': '136723052',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11828950956',
                                                                       'out_ucast_pkts': '136723052',
                                                                       'poller_hash': 'MUQwNkNEcGVlcmludGZwb24gNjcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '673',
                                                                                               'links': [       {       'connected_time': '8818315',
                                                                                                                        'fiber_length': 1023,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNkNEcGVlcnBvbnBvbiA2NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '60 Mill St. E. Milverton (William '
                                                 'Jantzi 10432843)',
                                         'poller_hash': 'MUQwNkNE',
                                         'serial_number': '1D06CD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:03:d9',
                                         'uptime': '8818315'},
                                 {       'counter_source': '674/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c8:9e:43:d6:7a:0a'],
                                                                       'id': '674/g1',
                                                                       'in_nucast_pkts': '4498',
                                                                       'in_octets': '205198210245',
                                                                       'in_ucast_pkts': '992076266',
                                                                       'mac_address': '00:06:31:52:63:ea',
                                                                       'mtu': 2000,
                                                                       'name': '674/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27327145',
                                                                       'out_octets': '3855359549533',
                                                                       'out_ucast_pkts': '2825199525',
                                                                       'poller_hash': 'ODg5QzFwZWVyaW50ZjY3NC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '674',
                                                                       'in_nucast_pkts': '27327145',
                                                                       'in_octets': '3855359549533',
                                                                       'in_ucast_pkts': '2825199525',
                                                                       'links': [       {       'in_nucast_pkts': '27327145',
                                                                                                'in_octets': '3855359549533',
                                                                                                'in_ucast_pkts': '2825199525',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4498',
                                                                                                'out_octets': '205198210245',
                                                                                                'out_ucast_pkts': '992076266',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4498',
                                                                       'out_octets': '205198210245',
                                                                       'out_ucast_pkts': '992076266',
                                                                       'poller_hash': 'ODg5QzFwZWVyaW50ZnBvbiA2NzQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '674',
                                                                                               'links': [       {       'connected_time': '15498560',
                                                                                                                        'fiber_length': 987,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5QzFwZWVycG9ucG9uIDY3NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '55 Mill St. E. Milverton (Kelly '
                                                 'Hargrave 10434003)',
                                         'poller_hash': 'ODg5QzE=',
                                         'serial_number': '889C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:a9',
                                         'uptime': '15498560'},
                                 {       'counter_source': '676/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c8:d3:a3:59:59:e3'],
                                                                       'id': '676/g1',
                                                                       'in_nucast_pkts': '233',
                                                                       'in_octets': '9605099303',
                                                                       'in_ucast_pkts': '18425762',
                                                                       'mac_address': 'ec:4f:82:7e:0e:8c',
                                                                       'mtu': 2000,
                                                                       'name': '676/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35074366',
                                                                       'out_octets': '38064517457',
                                                                       'out_ucast_pkts': '32039598',
                                                                       'poller_hash': 'MzVFQjI0cGVlcmludGY2NzYvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '676',
                                                                       'in_nucast_pkts': '35074366',
                                                                       'in_octets': '38064517457',
                                                                       'in_ucast_pkts': '32039598',
                                                                       'links': [       {       'in_nucast_pkts': '35074366',
                                                                                                'in_octets': '38064517457',
                                                                                                'in_ucast_pkts': '32039598',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '233',
                                                                                                'out_octets': '9605099303',
                                                                                                'out_ucast_pkts': '18425762',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '233',
                                                                       'out_octets': '9605099303',
                                                                       'out_ucast_pkts': '18425762',
                                                                       'poller_hash': 'MzVFQjI0cGVlcmludGZwb24gNjc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '676',
                                                                                               'links': [       {       'connected_time': '15499171',
                                                                                                                        'fiber_length': 1624,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQjI0cGVlcnBvbnBvbiA2NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '7 Colbalt St. N. Milverton (Judy '
                                                 'Hamilton 10422751)',
                                         'poller_hash': 'MzVFQjI0',
                                         'serial_number': '35EB24',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0e:8c',
                                         'uptime': '15499171'},
                                 {       'counter_source': '677/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:2c:63'],
                                                                       'id': '677/g1',
                                                                       'in_nucast_pkts': '14569',
                                                                       'in_octets': '9380725467',
                                                                       'in_ucast_pkts': '39437157',
                                                                       'mac_address': '00:06:31:49:14:c6',
                                                                       'mtu': 2000,
                                                                       'name': '677/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119715',
                                                                       'out_octets': '447196947757',
                                                                       'out_ucast_pkts': '301503632',
                                                                       'poller_hash': 'OEMzMDBwZWVyaW50ZjY3Ny9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '677',
                                                                       'in_nucast_pkts': '1119715',
                                                                       'in_octets': '447196947757',
                                                                       'in_ucast_pkts': '301503632',
                                                                       'links': [       {       'in_nucast_pkts': '1119715',
                                                                                                'in_octets': '447196947757',
                                                                                                'in_ucast_pkts': '301503632',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14569',
                                                                                                'out_octets': '9380725467',
                                                                                                'out_ucast_pkts': '39437157',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14569',
                                                                       'out_octets': '9380725467',
                                                                       'out_ucast_pkts': '39437157',
                                                                       'poller_hash': 'OEMzMDBwZWVyaW50ZnBvbiA2Nzc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '677',
                                                                                               'links': [       {       'connected_time': '688432',
                                                                                                                        'fiber_length': 916,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OEMzMDBwZWVycG9ucG9uIDY3Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '61 Mill St. E. Milverton (Jeff '
                                                 'Scott)',
                                         'poller_hash': 'OEMzMDA=',
                                         'serial_number': '8C300',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:47:6b:ec',
                                         'uptime': '688432'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '678/g1',
                                                                       'name': '678/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODg5RUNwZWVyaW50ZjY3OC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '678',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODg5RUNwZWVyaW50ZnBvbiA2Nzg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '678',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RUNwZWVycG9ucG9uIDY3OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '65 Mill St. E. Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "678" '
                                                                  '(location: GPON '
                                                                  'port "6/6"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:12:01.470Z',
                                         'poller_hash': 'ODg5RUM=',
                                         'serial_number': '889EC',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '679/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e2:9e'],
                                                                       'id': '679/g1',
                                                                       'in_nucast_pkts': '160894',
                                                                       'in_octets': '17668306125',
                                                                       'in_ucast_pkts': '205822117',
                                                                       'mac_address': '00:06:31:69:ee:e5',
                                                                       'mtu': 2000,
                                                                       'name': '679/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14151931',
                                                                       'out_octets': '1271636992737',
                                                                       'out_ucast_pkts': '845839190',
                                                                       'poller_hash': 'QzU3M0VwZWVyaW50ZjY3OS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '679',
                                                                       'in_nucast_pkts': '14151931',
                                                                       'in_octets': '1271636992737',
                                                                       'in_ucast_pkts': '845839190',
                                                                       'links': [       {       'in_nucast_pkts': '14151931',
                                                                                                'in_octets': '1271636992737',
                                                                                                'in_ucast_pkts': '845839190',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160894',
                                                                                                'out_octets': '17668306125',
                                                                                                'out_ucast_pkts': '205822117',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160894',
                                                                       'out_octets': '17668306125',
                                                                       'out_ucast_pkts': '205822117',
                                                                       'poller_hash': 'QzU3M0VwZWVyaW50ZnBvbiA2Nzk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '679',
                                                                                               'links': [       {       'connected_time': '7712321',
                                                                                                                        'fiber_length': 953,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3M0VwZWVycG9ucG9uIDY3OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '63 MILL ST E (GRAY JACKIE)',
                                         'poller_hash': 'QzU3M0U=',
                                         'serial_number': 'C573E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:67:fe',
                                         'uptime': '7712321'},
                                 {       'counter_source': '680/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ac:8b:a9:11:38:c1'],
                                                                       'id': '680/g1',
                                                                       'in_nucast_pkts': '41',
                                                                       'in_octets': '5332439266',
                                                                       'in_ucast_pkts': '7726935',
                                                                       'mac_address': '00:06:31:69:f8:ad',
                                                                       'mtu': 2000,
                                                                       'name': '680/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19517',
                                                                       'out_octets': '7071338505',
                                                                       'out_ucast_pkts': '7316591',
                                                                       'poller_hash': 'QzU4NDlwZWVyaW50ZjY4MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '680',
                                                                       'in_nucast_pkts': '19517',
                                                                       'in_octets': '7071338505',
                                                                       'in_ucast_pkts': '7316591',
                                                                       'links': [       {       'in_nucast_pkts': '19517',
                                                                                                'in_octets': '7071338505',
                                                                                                'in_ucast_pkts': '7316591',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '41',
                                                                                                'out_octets': '5332439266',
                                                                                                'out_ucast_pkts': '7726935',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41',
                                                                       'out_octets': '5332439266',
                                                                       'out_ucast_pkts': '7726935',
                                                                       'poller_hash': 'QzU4NDlwZWVyaW50ZnBvbiA2ODA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '680',
                                                                                               'links': [       {       'connected_time': '12347',
                                                                                                                        'fiber_length': 935,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU4NDlwZWVycG9ucG9uIDY4MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '52 MILL ST E (RICHARD MITCHELL)',
                                         'poller_hash': 'QzU4NDk=',
                                         'serial_number': 'C5849',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:69:09',
                                         'uptime': '12347'},
                                 {       'counter_source': '681/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:92:a6'],
                                                                       'id': '681/g1',
                                                                       'in_nucast_pkts': '14780',
                                                                       'in_octets': '15103265123',
                                                                       'in_ucast_pkts': '26249140',
                                                                       'mac_address': '00:06:31:87:48:8f',
                                                                       'mtu': 2000,
                                                                       'name': '681/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119705',
                                                                       'out_octets': '65221456163',
                                                                       'out_ucast_pkts': '60039058',
                                                                       'poller_hash': 'Q0MwOUFwZWVyaW50ZjY4MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '681',
                                                                       'in_nucast_pkts': '1119705',
                                                                       'in_octets': '65221456163',
                                                                       'in_ucast_pkts': '60039058',
                                                                       'links': [       {       'in_nucast_pkts': '1119705',
                                                                                                'in_octets': '65221456163',
                                                                                                'in_ucast_pkts': '60039058',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14780',
                                                                                                'out_octets': '15103265123',
                                                                                                'out_ucast_pkts': '26249140',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14780',
                                                                       'out_octets': '15103265123',
                                                                       'out_ucast_pkts': '26249140',
                                                                       'poller_hash': 'Q0MwOUFwZWVyaW50ZnBvbiA2ODE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '681',
                                                                                               'links': [       {       'connected_time': '688431',
                                                                                                                        'fiber_length': 936,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwOUFwZWVycG9ucG9uIDY4MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '62 MILL ST E (RICHARDSON CLAIRE)',
                                         'poller_hash': 'Q0MwOUE=',
                                         'serial_number': 'CC09A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:5a',
                                         'uptime': '688431'},
                                 {       'counter_source': '682/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:9f:80:75:ac:4f'],
                                                                       'id': '682/g1',
                                                                       'in_nucast_pkts': '32860',
                                                                       'in_octets': '18711742461',
                                                                       'in_ucast_pkts': '32834528',
                                                                       'mac_address': '00:06:31:87:40:ff',
                                                                       'mtu': 2000,
                                                                       'name': '682/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120171',
                                                                       'out_octets': '161031603161',
                                                                       'out_ucast_pkts': '124175644',
                                                                       'poller_hash': 'Q0MwNzVwZWVyaW50ZjY4Mi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '682',
                                                                       'in_nucast_pkts': '1120171',
                                                                       'in_octets': '161031603161',
                                                                       'in_ucast_pkts': '124175644',
                                                                       'links': [       {       'in_nucast_pkts': '1120171',
                                                                                                'in_octets': '161031603161',
                                                                                                'in_ucast_pkts': '124175644',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '32860',
                                                                                                'out_octets': '18711742461',
                                                                                                'out_ucast_pkts': '32834528',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32860',
                                                                       'out_octets': '18711742461',
                                                                       'out_ucast_pkts': '32834528',
                                                                       'poller_hash': 'Q0MwNzVwZWVyaW50ZnBvbiA2ODI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '682',
                                                                                               'links': [       {       'connected_time': '688418',
                                                                                                                        'fiber_length': 998,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwNzVwZWVycG9ucG9uIDY4Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '48 Mill St. E. Milverton (Michelle '
                                                 'Corbett)',
                                         'poller_hash': 'Q0MwNzU=',
                                         'serial_number': 'CC075',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:35',
                                         'uptime': '688418'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '683/g1',
                                                                       'name': '683/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTMwMDU4cGVlcmludGY2ODMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '683',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTMwMDU4cGVlcmludGZwb24gNjgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '683',
                                                                                               'links': [       {       'connected_time': '15498593',
                                                                                                                        'fiber_length': 1008,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwMDU4cGVlcnBvbnBvbiA2ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '68 Mill St. E. Milverton (Nursery '
                                                 'School)',
                                         'poller_hash': 'MTMwMDU4',
                                         'serial_number': '130058',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:f1:18',
                                         'uptime': '15498593'},
                                 {       'counter_source': '684/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '684/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '470322831826',
                                                                       'in_ucast_pkts': '5408363977',
                                                                       'mac_address': '00:06:31:8f:7f:83',
                                                                       'name': '684/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '19230406529003',
                                                                       'out_ucast_pkts': '12949695649',
                                                                       'poller_hash': 'MTI0ODNGcGVlcmludGY2ODQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '684',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '19230406529003',
                                                                       'in_ucast_pkts': '12949695649',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '19230406529003',
                                                                                                'in_ucast_pkts': '12949695649',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '470322831826',
                                                                                                'out_ucast_pkts': '5408363977',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '470322831826',
                                                                       'out_ucast_pkts': '5408363977',
                                                                       'poller_hash': 'MTI0ODNGcGVlcmludGZwb24gNjg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '684',
                                                                                               'links': [       {       'connected_time': '46993',
                                                                                                                        'fiber_length': 1655,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.92,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI0ODNGcGVlcnBvbnBvbiA2ODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '3 Cobalt St. N. Milverton (Rick '
                                                 'Banks)',
                                         'poller_hash': 'MTI0ODNG',
                                         'serial_number': '12483F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:90:38:ff',
                                         'uptime': '46993'},
                                 {       'counter_source': '685/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '685/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '69331969638',
                                                                       'in_ucast_pkts': '235154476',
                                                                       'mac_address': '00:06:31:52:64:16',
                                                                       'name': '685/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '790918907861',
                                                                       'out_ucast_pkts': '726768307',
                                                                       'poller_hash': 'ODg5RTZwZWVyaW50ZjY4NS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '685',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '790918907861',
                                                                       'in_ucast_pkts': '726768307',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '790918907861',
                                                                                                'in_ucast_pkts': '726768307',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '69331969638',
                                                                                                'out_ucast_pkts': '235154476',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '69331969638',
                                                                       'out_ucast_pkts': '235154476',
                                                                       'poller_hash': 'ODg5RTZwZWVyaW50ZnBvbiA2ODU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '685',
                                                                                               'links': [       {       'connected_time': '47637',
                                                                                                                        'fiber_length': 884,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODg5RTZwZWVycG9ucG9uIDY4NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '56 Mill St. E. Milverton (Mark '
                                                 'Gingerich 10432800)',
                                         'poller_hash': 'ODg5RTY=',
                                         'serial_number': '889E6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:ea:ce',
                                         'uptime': '47637'},
                                 {       'counter_source': '687/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '687/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '231374831298',
                                                                       'in_ucast_pkts': '1738248115',
                                                                       'mac_address': '00:06:31:b6:7e:4c',
                                                                       'name': '687/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9457434507617',
                                                                       'out_ucast_pkts': '6376420501',
                                                                       'poller_hash': 'MTU1NDlGcGVlcmludGY2ODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '687',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9457434507617',
                                                                       'in_ucast_pkts': '6376420501',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9457434507617',
                                                                                                'in_ucast_pkts': '6376420501',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '231374831298',
                                                                                                'out_ucast_pkts': '1738248115',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '231374831298',
                                                                       'out_ucast_pkts': '1738248115',
                                                                       'poller_hash': 'MTU1NDlGcGVlcmludGZwb24gNjg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '687',
                                                                                               'links': [       {       'connected_time': '65341989',
                                                                                                                        'fiber_length': 1995,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.348,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU1NDlGcGVlcnBvbnBvbiA2ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '40 Cobalt St. N. Milverton (Lorne '
                                                 'Reis)',
                                         'poller_hash': 'MTU1NDlG',
                                         'serial_number': '15549F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:ad:3a:e5',
                                         'uptime': '65341989'},
                                 {       'counter_source': '688/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a9:11'],
                                                                       'id': '688/g1',
                                                                       'in_nucast_pkts': '265',
                                                                       'in_octets': '273835599909',
                                                                       'in_ucast_pkts': '1618089679',
                                                                       'mac_address': 'ec:4f:82:7e:3a:57',
                                                                       'mtu': 2000,
                                                                       'name': '688/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83574339',
                                                                       'out_octets': '6098986305922',
                                                                       'out_ucast_pkts': '4311082233',
                                                                       'poller_hash': 'MzVGOUIxcGVlcmludGY2ODgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '688',
                                                                       'in_nucast_pkts': '83574339',
                                                                       'in_octets': '6098986305922',
                                                                       'in_ucast_pkts': '4311082233',
                                                                       'links': [       {       'in_nucast_pkts': '83574339',
                                                                                                'in_octets': '6098986305922',
                                                                                                'in_ucast_pkts': '4311082233',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '265',
                                                                                                'out_octets': '273835599909',
                                                                                                'out_ucast_pkts': '1618089679',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '265',
                                                                       'out_octets': '273835599909',
                                                                       'out_ucast_pkts': '1618089679',
                                                                       'poller_hash': 'MzVGOUIxcGVlcmludGZwb24gNjg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '688',
                                                                                               'links': [       {       'connected_time': '67790223',
                                                                                                                        'fiber_length': 1154,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUIxcGVlcnBvbnBvbiA2ODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '36 William St (Evan Hargrave)',
                                         'poller_hash': 'MzVGOUIx',
                                         'serial_number': '35F9B1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3a:57',
                                         'uptime': '67790223'},
                                 {       'counter_source': '689/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:4b:52'],
                                                                       'id': '689/g1',
                                                                       'in_nucast_pkts': '15489',
                                                                       'in_octets': '7544064331',
                                                                       'in_ucast_pkts': '23645244',
                                                                       'mac_address': 'ec:4f:82:ae:bc:ff',
                                                                       'mtu': 2000,
                                                                       'name': '689/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447054',
                                                                       'out_octets': '99996998869',
                                                                       'out_ucast_pkts': '76853472',
                                                                       'poller_hash': 'M0I2MURFcGVlcmludGY2ODkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '689',
                                                                       'in_nucast_pkts': '1447054',
                                                                       'in_octets': '99996998869',
                                                                       'in_ucast_pkts': '76853472',
                                                                       'links': [       {       'in_nucast_pkts': '1447054',
                                                                                                'in_octets': '99996998869',
                                                                                                'in_ucast_pkts': '76853472',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15489',
                                                                                                'out_octets': '7544064331',
                                                                                                'out_ucast_pkts': '23645244',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15489',
                                                                       'out_octets': '7544064331',
                                                                       'out_ucast_pkts': '23645244',
                                                                       'poller_hash': 'M0I2MURFcGVlcmludGZwb24gNjg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '689',
                                                                                               'links': [       {       'connected_time': '688411',
                                                                                                                        'fiber_length': 2224,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0I2MURFcGVlcnBvbnBvbiA2ODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '72 GERBER DR (MCCLUSKIE CODY)',
                                         'poller_hash': 'M0I2MURF',
                                         'serial_number': '3B61DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:ae:bc:ff',
                                         'uptime': '688411'},
                                 {       'counter_source': '690/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '690/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96633036714',
                                                                       'in_ucast_pkts': '5329082604',
                                                                       'mac_address': 'ec:4f:82:a2:b8:2a',
                                                                       'name': '690/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '144',
                                                                       'out_octets': '5208321421769',
                                                                       'out_ucast_pkts': '7807743112',
                                                                       'poller_hash': 'MzQ5RTg3cGVlcmludGY2OTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '690',
                                                                       'in_nucast_pkts': '144',
                                                                       'in_octets': '5208321421769',
                                                                       'in_ucast_pkts': '7807743112',
                                                                       'links': [       {       'in_nucast_pkts': '144',
                                                                                                'in_octets': '5208321421769',
                                                                                                'in_ucast_pkts': '7807743112',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96633036714',
                                                                                                'out_ucast_pkts': '5329082604',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96633036714',
                                                                       'out_ucast_pkts': '5329082604',
                                                                       'poller_hash': 'MzQ5RTg3cGVlcmludGZwb24gNjkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '690',
                                                                                               'links': [       {       'connected_time': '20839243',
                                                                                                                        'fiber_length': 1046,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ5RTg3cGVlcnBvbnBvbiA2OTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 MAIN ST S STE 10 (GROPP BARRY)',
                                         'poller_hash': 'MzQ5RTg3',
                                         'serial_number': '349E87',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:a2:b8:2a',
                                         'uptime': '20839243'},
                                 {       'counter_source': '693/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '693/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '141418925013',
                                                                       'in_ucast_pkts': '914908673',
                                                                       'mac_address': 'cc:be:59:cb:af:82',
                                                                       'name': '693/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '4064085761154',
                                                                       'out_ucast_pkts': '2986767287',
                                                                       'poller_hash': 'NDJDNzQ4cGVlcmludGY2OTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '693',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '4064085761154',
                                                                       'in_ucast_pkts': '2986767287',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '4064085761154',
                                                                                                'in_ucast_pkts': '2986767287',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '141418925013',
                                                                                                'out_ucast_pkts': '914908673',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '141418925013',
                                                                       'out_ucast_pkts': '914908673',
                                                                       'poller_hash': 'NDJDNzQ4cGVlcmludGZwb24gNjkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '693',
                                                                                               'links': [       {       'connected_time': '21816686',
                                                                                                                        'fiber_length': 7080,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNzQ4cGVlcnBvbnBvbiA2OTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4987 Line 59 (Prima Valley Farms Ltd '
                                                 '10400278)',
                                         'poller_hash': 'NDJDNzQ4',
                                         'serial_number': '42C748',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:af:82',
                                         'uptime': '21816686'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '694/g1',
                                                                       'name': '694/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1QjM4cGVlcmludGY2OTQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '694',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MjA1QjM4cGVlcmludGZwb24gNjk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '694',
                                                                                               'links': [       {       'connected_time': '30595765',
                                                                                                                        'fiber_length': 1115,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1QjM4cGVlcnBvbnBvbiA2OTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 MAIN ST S (COOK CONSTANCE)',
                                         'poller_hash': 'MjA1QjM4',
                                         'serial_number': '205B38',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:51:cd',
                                         'uptime': '30595765'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '695/g1',
                                                                       'name': '695/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNTc1cGVlcmludGY2OTUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '695',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTEyNTc1cGVlcmludGZwb24gNjk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '695',
                                                                                               'links': [       {       'connected_time': '538571',
                                                                                                                        'fiber_length': 16710,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTc1cGVlcnBvbnBvbiA2OTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4650 Line 49 (Linda Debus10403081)',
                                         'poller_hash': 'NTEyNTc1',
                                         'serial_number': '512575',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c4:c7',
                                         'uptime': '538571'},
                                 {       'counter_source': '696/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0f:42'],
                                                                       'id': '696/g1',
                                                                       'in_nucast_pkts': '12897',
                                                                       'in_octets': '1022751382258',
                                                                       'in_ucast_pkts': '1459357976',
                                                                       'mac_address': 'cc:be:59:cc:01:f5',
                                                                       'mtu': 2000,
                                                                       'name': '696/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '122773309',
                                                                       'out_octets': '2510490403929',
                                                                       'out_ucast_pkts': '2415910591',
                                                                       'poller_hash': 'NDJEMTk2cGVlcmludGY2OTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '696',
                                                                       'in_nucast_pkts': '122773309',
                                                                       'in_octets': '2510490403929',
                                                                       'in_ucast_pkts': '2415910591',
                                                                       'links': [       {       'in_nucast_pkts': '122773309',
                                                                                                'in_octets': '2510490403929',
                                                                                                'in_ucast_pkts': '2415910591',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12897',
                                                                                                'out_octets': '1022751382258',
                                                                                                'out_ucast_pkts': '1459357976',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12897',
                                                                       'out_octets': '1022751382258',
                                                                       'out_ucast_pkts': '1459357976',
                                                                       'poller_hash': 'NDJEMTk2cGVlcmludGZwb24gNjk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '696',
                                                                                               'links': [       {       'connected_time': '21816277',
                                                                                                                        'fiber_length': 16662,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMTk2cGVlcnBvbnBvbiA2OTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4650 Line 49-3 (J & J Millwrighting '
                                                 '10403280)',
                                         'poller_hash': 'NDJEMTk2',
                                         'serial_number': '42D196',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:01:f2',
                                         'uptime': '21816277'},
                                 {       'counter_source': '697/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:c0'],
                                                                       'id': '697/g1',
                                                                       'in_nucast_pkts': '11892',
                                                                       'in_octets': '4050873660',
                                                                       'in_ucast_pkts': '20561219',
                                                                       'mac_address': 'cc:be:59:4f:fb:16',
                                                                       'mtu': 2000,
                                                                       'name': '697/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119988',
                                                                       'out_octets': '99612986705',
                                                                       'out_ucast_pkts': '72137236',
                                                                       'poller_hash': 'NDNCMTU0cGVlcmludGY2OTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '697',
                                                                       'in_nucast_pkts': '1119988',
                                                                       'in_octets': '99612986705',
                                                                       'in_ucast_pkts': '72137236',
                                                                       'links': [       {       'in_nucast_pkts': '1119988',
                                                                                                'in_octets': '99612986705',
                                                                                                'in_ucast_pkts': '72137236',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11892',
                                                                                                'out_octets': '4050873660',
                                                                                                'out_ucast_pkts': '20561219',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11892',
                                                                       'out_octets': '4050873660',
                                                                       'out_ucast_pkts': '20561219',
                                                                       'poller_hash': 'NDNCMTU0cGVlcmludGZwb24gNjk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '697',
                                                                                               'links': [       {       'connected_time': '688491',
                                                                                                                        'fiber_length': 1112,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.636,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.29}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMTU0cGVlcnBvbnBvbiA2OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '14 Pacific Ave (Mitchell '
                                                 'Medeiros10434078)',
                                         'poller_hash': 'NDNCMTU0',
                                         'serial_number': '43B154',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:fb:15',
                                         'uptime': '688491'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '698/g1',
                                                                       'name': '698/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDMzhFcGVlcmludGY2OTgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '698',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJDMzhFcGVlcmludGZwb24gNjk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '698',
                                                                                               'links': [       {       'connected_time': '21816025',
                                                                                                                        'fiber_length': 16220,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMzhFcGVlcnBvbnBvbiA2OTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5340A Line 49 (Malcolm McDonald '
                                                 '10400307)',
                                         'poller_hash': 'NDJDMzhF',
                                         'serial_number': '42C38E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:91:b2',
                                         'uptime': '21816025'},
                                 {       'counter_source': '699/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '699/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '560583097788',
                                                                       'in_ucast_pkts': '12830316745',
                                                                       'mac_address': 'cc:be:59:cb:91:a2',
                                                                       'name': '699/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '22662435985101',
                                                                       'out_ucast_pkts': '15751334406',
                                                                       'poller_hash': 'NDJDMzhDcGVlcmludGY2OTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '699',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '22662435985101',
                                                                       'in_ucast_pkts': '15751334406',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '22662435985101',
                                                                                                'in_ucast_pkts': '15751334406',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '560583097788',
                                                                                                'out_ucast_pkts': '12830316745',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '560583097788',
                                                                       'out_ucast_pkts': '12830316745',
                                                                       'poller_hash': 'NDJDMzhDcGVlcmludGZwb24gNjk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '699',
                                                                                               'links': [       {       'connected_time': '709198',
                                                                                                                        'fiber_length': 16123,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMzhDcGVlcnBvbnBvbiA2OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5340B Line 49 Gadshill (Stu McDonald '
                                                 '10422004)',
                                         'poller_hash': 'NDJDMzhD',
                                         'serial_number': '42C38C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:91:a2',
                                         'uptime': '709198'},
                                 {       'counter_source': '700/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '700/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '190496346784',
                                                                       'in_ucast_pkts': '26947469307',
                                                                       'mac_address': 'cc:be:59:cb:8b:da',
                                                                       'name': '700/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '137',
                                                                       'out_octets': '3936401747777',
                                                                       'out_ucast_pkts': '2845053537',
                                                                       'poller_hash': 'NDJDMkQzcGVlcmludGY3MDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '700',
                                                                       'in_nucast_pkts': '137',
                                                                       'in_octets': '3936401747777',
                                                                       'in_ucast_pkts': '2845053537',
                                                                       'links': [       {       'in_nucast_pkts': '137',
                                                                                                'in_octets': '3936401747777',
                                                                                                'in_ucast_pkts': '2845053537',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '190496346784',
                                                                                                'out_ucast_pkts': '26947469307',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '190496346784',
                                                                       'out_ucast_pkts': '26947469307',
                                                                       'poller_hash': 'NDJDMkQzcGVlcmludGZwb24gNzAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '700',
                                                                                               'links': [       {       'connected_time': '72954320',
                                                                                                                        'fiber_length': 15500,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMkQzcGVlcnBvbnBvbiA3MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5256 Line 49 Gadshill (M & K DeJong '
                                                 '10427640)',
                                         'poller_hash': 'NDJDMkQz',
                                         'serial_number': '42C2D3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:8b:da',
                                         'uptime': '72954320'},
                                 {       'counter_source': '702/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '702/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '143668847257',
                                                                       'in_ucast_pkts': '1249415018',
                                                                       'mac_address': 'ec:4f:82:0d:51:f1',
                                                                       'name': '702/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '120',
                                                                       'out_octets': '8506462874943',
                                                                       'out_ucast_pkts': '5692440077',
                                                                       'poller_hash': 'MjA1QjNDcGVlcmludGY3MDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '702',
                                                                       'in_nucast_pkts': '120',
                                                                       'in_octets': '8506462874943',
                                                                       'in_ucast_pkts': '5692440077',
                                                                       'links': [       {       'in_nucast_pkts': '120',
                                                                                                'in_octets': '8506462874943',
                                                                                                'in_ucast_pkts': '5692440077',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '143668847257',
                                                                                                'out_ucast_pkts': '1249415018',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '143668847257',
                                                                       'out_ucast_pkts': '1249415018',
                                                                       'poller_hash': 'MjA1QjNDcGVlcmludGZwb24gNzAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '702',
                                                                                               'links': [       {       'connected_time': '688463',
                                                                                                                        'fiber_length': 1667,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1QjNDcGVlcnBvbnBvbiA3MDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 William St Apt 3',
                                         'poller_hash': 'MjA1QjND',
                                         'serial_number': '205B3C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:51:f1',
                                         'uptime': '688463'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '704/g1',
                                                                       'name': '704/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIxMERFcGVlcmludGY3MDQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '704',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzIxMERFcGVlcmludGZwb24gNzA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '704',
                                                                                               'links': [       {       'connected_time': '688413',
                                                                                                                        'fiber_length': 2188,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMERFcGVlcnBvbnBvbiA3MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '76 GERBER DR',
                                         'poller_hash': 'MzIxMERF',
                                         'serial_number': '3210DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:44:b8',
                                         'uptime': '688413'},
                                 {       'counter_source': '705/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a6:2b'],
                                                                       'id': '705/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '14386312082',
                                                                       'in_ucast_pkts': '42866565',
                                                                       'mac_address': 'ec:4f:82:ae:bd:43',
                                                                       'mtu': 2000,
                                                                       'name': '705/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464226',
                                                                       'out_octets': '178678932006',
                                                                       'out_ucast_pkts': '137710738',
                                                                       'poller_hash': 'M0I2MjAwcGVlcmludGY3MDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '705',
                                                                       'in_nucast_pkts': '1464226',
                                                                       'in_octets': '178678932006',
                                                                       'in_ucast_pkts': '137710738',
                                                                       'links': [       {       'in_nucast_pkts': '1464226',
                                                                                                'in_octets': '178678932006',
                                                                                                'in_ucast_pkts': '137710738',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '14386312082',
                                                                                                'out_ucast_pkts': '42866565',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '14386312082',
                                                                       'out_ucast_pkts': '42866565',
                                                                       'poller_hash': 'M0I2MjAwcGVlcmludGZwb24gNzA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '705',
                                                                                               'links': [       {       'connected_time': '688492',
                                                                                                                        'fiber_length': 943,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0I2MjAwcGVlcnBvbnBvbiA3MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '15 Pacific Ave (Steven & Debbie '
                                                 'Emerson, 10404480)',
                                         'poller_hash': 'M0I2MjAw',
                                         'serial_number': '3B6200',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:ae:bd:43',
                                         'uptime': '688492'},
                                 {       'counter_source': '706/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ed:97'],
                                                                       'id': '706/g1',
                                                                       'in_nucast_pkts': '15248',
                                                                       'in_octets': '11612302394',
                                                                       'in_ucast_pkts': '30553566',
                                                                       'mac_address': 'ec:4f:82:7d:cb:a7',
                                                                       'mtu': 2000,
                                                                       'name': '706/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464114',
                                                                       'out_octets': '79214217446',
                                                                       'out_ucast_pkts': '74370487',
                                                                       'poller_hash': 'MzAwRjAwcGVlcmludGY3MDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '706',
                                                                       'in_nucast_pkts': '1464114',
                                                                       'in_octets': '79214217446',
                                                                       'in_ucast_pkts': '74370487',
                                                                       'links': [       {       'in_nucast_pkts': '1464114',
                                                                                                'in_octets': '79214217446',
                                                                                                'in_ucast_pkts': '74370487',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15248',
                                                                                                'out_octets': '11612302394',
                                                                                                'out_ucast_pkts': '30553566',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15248',
                                                                       'out_octets': '11612302394',
                                                                       'out_ucast_pkts': '30553566',
                                                                       'poller_hash': 'MzAwRjAwcGVlcmludGZwb24gNzA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '706',
                                                                                               'links': [       {       'connected_time': '688487',
                                                                                                                        'fiber_length': 1705,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRjAwcGVlcnBvbnBvbiA3MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '60 MAIN ST N (NEWMAN ASHLYNNE)',
                                         'poller_hash': 'MzAwRjAw',
                                         'serial_number': '300F00',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:cb:a7',
                                         'uptime': '688487'},
                                 {       'counter_source': '707/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a8:50'],
                                                                       'id': '707/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '18989522075',
                                                                       'in_ucast_pkts': '43392524',
                                                                       'mac_address': 'ec:4f:82:8d:8b:9b',
                                                                       'mtu': 2000,
                                                                       'name': '707/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447083',
                                                                       'out_octets': '266326628952',
                                                                       'out_ucast_pkts': '201280082',
                                                                       'poller_hash': 'MzIxOUJBcGVlcmludGY3MDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '707',
                                                                       'in_nucast_pkts': '1447083',
                                                                       'in_octets': '266326628952',
                                                                       'in_ucast_pkts': '201280082',
                                                                       'links': [       {       'in_nucast_pkts': '1447083',
                                                                                                'in_octets': '266326628952',
                                                                                                'in_ucast_pkts': '201280082',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '18989522075',
                                                                                                'out_ucast_pkts': '43392524',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '18989522075',
                                                                       'out_ucast_pkts': '43392524',
                                                                       'poller_hash': 'MzIxOUJBcGVlcmludGZwb24gNzA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '707',
                                                                                               'links': [       {       'connected_time': '688410',
                                                                                                                        'fiber_length': 1983,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxOUJBcGVlcnBvbnBvbiA3MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '90 GERBER DR (LIVINGSTONE MATTHEW)',
                                         'poller_hash': 'MzIxOUJB',
                                         'serial_number': '3219BA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:8b:98',
                                         'uptime': '688410'},
                                 {       'counter_source': '708/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '708/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '35958964854',
                                                                       'in_ucast_pkts': '136828569',
                                                                       'mac_address': 'ec:4f:82:8d:dd:e0',
                                                                       'name': '708/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '573100907147',
                                                                       'out_ucast_pkts': '419335398',
                                                                       'poller_hash': 'MzIyNDAzcGVlcmludGY3MDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '708',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '573100907147',
                                                                       'in_ucast_pkts': '419335398',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '573100907147',
                                                                                                'in_ucast_pkts': '419335398',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '35958964854',
                                                                                                'out_ucast_pkts': '136828569',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '35958964854',
                                                                       'out_ucast_pkts': '136828569',
                                                                       'poller_hash': 'MzIyNDAzcGVlcmludGZwb24gNzA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '708',
                                                                                               'links': [       {       'connected_time': '29747800',
                                                                                                                        'fiber_length': 766,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDAzcGVlcnBvbnBvbiA3MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '37 Main St. S. Milverton (Vicki '
                                                 'Baier)',
                                         'poller_hash': 'MzIyNDAz',
                                         'serial_number': '322403',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:dd:e0',
                                         'uptime': '29747800'},
                                 {       'counter_source': '709/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '709/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '478393861318',
                                                                       'in_ucast_pkts': '859247709',
                                                                       'mac_address': 'ec:4f:82:8d:7a:38',
                                                                       'name': '709/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '6998649661595',
                                                                       'out_ucast_pkts': '11488143595',
                                                                       'poller_hash': 'MzIxNzhFcGVlcmludGY3MDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '709',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '6998649661595',
                                                                       'in_ucast_pkts': '11488143595',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '6998649661595',
                                                                                                'in_ucast_pkts': '11488143595',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '478393861318',
                                                                                                'out_ucast_pkts': '859247709',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '478393861318',
                                                                       'out_ucast_pkts': '859247709',
                                                                       'poller_hash': 'MzIxNzhFcGVlcmludGZwb24gNzA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '709',
                                                                                               'links': [       {       'connected_time': '23939405',
                                                                                                                        'fiber_length': 12290,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxNzhFcGVlcnBvbnBvbiA3MDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5752 Perth Rd. 119 (Micheal Weber '
                                                 '10400697)',
                                         'poller_hash': 'MzIxNzhF',
                                         'serial_number': '32178E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:7a:38',
                                         'uptime': '23939405'},
                                 {       'counter_source': '710/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a4:2b:8c:7c:8c:8d'],
                                                                       'id': '710/g1',
                                                                       'in_nucast_pkts': '3688',
                                                                       'in_octets': '161852628717',
                                                                       'in_ucast_pkts': '598537909',
                                                                       'mac_address': 'ec:4f:82:7e:08:11',
                                                                       'mtu': 2000,
                                                                       'name': '710/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27790844',
                                                                       'out_octets': '1359732485616',
                                                                       'out_ucast_pkts': '1108931423',
                                                                       'poller_hash': 'MzVFOEZCcGVlcmludGY3MTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '710',
                                                                       'in_nucast_pkts': '27790844',
                                                                       'in_octets': '1359732485616',
                                                                       'in_ucast_pkts': '1108931423',
                                                                       'links': [       {       'in_nucast_pkts': '27790844',
                                                                                                'in_octets': '1359732485616',
                                                                                                'in_ucast_pkts': '1108931423',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3688',
                                                                                                'out_octets': '161852628717',
                                                                                                'out_ucast_pkts': '598537909',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3688',
                                                                       'out_octets': '161852628717',
                                                                       'out_ucast_pkts': '598537909',
                                                                       'poller_hash': 'MzVFOEZCcGVlcmludGZwb24gNzEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '710',
                                                                                               'links': [       {       'connected_time': '12861944',
                                                                                                                        'fiber_length': 17606,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -24.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOEZCcGVlcnBvbnBvbiA3MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '170 Winstanley St Monkton (STECKLY '
                                                 'JEREMY)',
                                         'poller_hash': 'MzVFOEZC',
                                         'serial_number': '35E8FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:08:11',
                                         'uptime': '12861944'},
                                 {       'counter_source': '711/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:8a:7c'],
                                                                       'id': '711/g1',
                                                                       'in_nucast_pkts': '841138',
                                                                       'in_octets': '1235421746844',
                                                                       'in_ucast_pkts': '2573789681',
                                                                       'mac_address': 'ec:4f:82:8d:fe:f3',
                                                                       'mtu': 2000,
                                                                       'name': '711/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '141487337',
                                                                       'out_octets': '8069040338829',
                                                                       'out_ucast_pkts': '6368832161',
                                                                       'poller_hash': 'MzIyODI1cGVlcmludGY3MTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '711',
                                                                       'in_nucast_pkts': '141487337',
                                                                       'in_octets': '8069040338829',
                                                                       'in_ucast_pkts': '6368832161',
                                                                       'links': [       {       'in_nucast_pkts': '141487337',
                                                                                                'in_octets': '8069040338829',
                                                                                                'in_ucast_pkts': '6368832161',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '841138',
                                                                                                'out_octets': '1235421746844',
                                                                                                'out_ucast_pkts': '2573789681',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '841138',
                                                                       'out_octets': '1235421746844',
                                                                       'out_ucast_pkts': '2573789681',
                                                                       'poller_hash': 'MzIyODI1cGVlcmludGZwb24gNzEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '711',
                                                                                               'links': [       {       'connected_time': '20658378',
                                                                                                                        'fiber_length': 1968,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODI1cGVlcnBvbnBvbiA3MTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '93 GERBER DR (WITT BROOKLYNN)',
                                         'poller_hash': 'MzIyODI1',
                                         'serial_number': '322825',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fe:f0',
                                         'uptime': '20658378'},
                                 {       'counter_source': '712/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ed:00'],
                                                                       'id': '712/g1',
                                                                       'in_nucast_pkts': '11867',
                                                                       'in_octets': '1858120830',
                                                                       'in_ucast_pkts': '9404594',
                                                                       'mac_address': 'ec:4f:82:8e:00:43',
                                                                       'mtu': 2000,
                                                                       'name': '712/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1456868',
                                                                       'out_octets': '45162445772',
                                                                       'out_ucast_pkts': '35565799',
                                                                       'poller_hash': 'MzIyODRGcGVlcmludGY3MTIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '712',
                                                                       'in_nucast_pkts': '1456868',
                                                                       'in_octets': '45162445772',
                                                                       'in_ucast_pkts': '35565799',
                                                                       'links': [       {       'in_nucast_pkts': '1456868',
                                                                                                'in_octets': '45162445772',
                                                                                                'in_ucast_pkts': '35565799',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11867',
                                                                                                'out_octets': '1858120830',
                                                                                                'out_ucast_pkts': '9404594',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11867',
                                                                       'out_octets': '1858120830',
                                                                       'out_ucast_pkts': '9404594',
                                                                       'poller_hash': 'MzIyODRGcGVlcmludGZwb24gNzEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '712',
                                                                                               'links': [       {       'connected_time': '688500',
                                                                                                                        'fiber_length': 1213,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODRGcGVlcnBvbnBvbiA3MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 Pacific Ave Milverton (Matthew '
                                                 'Bancroft 10436288)',
                                         'poller_hash': 'MzIyODRG',
                                         'serial_number': '32284F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:00:40',
                                         'uptime': '688500'},
                                 {       'counter_source': '715/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '715/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '102496007989',
                                                                       'in_ucast_pkts': '22318210147',
                                                                       'mac_address': 'ec:4f:82:a2:b6:44',
                                                                       'name': '715/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '98',
                                                                       'out_octets': '4909918565667',
                                                                       'out_ucast_pkts': '3354070159',
                                                                       'poller_hash': 'MzQ5RTUxcGVlcmludGY3MTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '715',
                                                                       'in_nucast_pkts': '98',
                                                                       'in_octets': '4909918565667',
                                                                       'in_ucast_pkts': '3354070159',
                                                                       'links': [       {       'in_nucast_pkts': '98',
                                                                                                'in_octets': '4909918565667',
                                                                                                'in_ucast_pkts': '3354070159',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '102496007989',
                                                                                                'out_ucast_pkts': '22318210147',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '102496007989',
                                                                       'out_ucast_pkts': '22318210147',
                                                                       'poller_hash': 'MzQ5RTUxcGVlcmludGZwb24gNzE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '715',
                                                                                               'links': [       {       'connected_time': '68802632',
                                                                                                                        'fiber_length': 1119,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ5RTUxcGVlcnBvbnBvbiA3MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 Main St S Ste 17 Milverton '
                                                 '(Katherine Priestap 10434089)',
                                         'poller_hash': 'MzQ5RTUx',
                                         'serial_number': '349E51',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:a2:b6:44',
                                         'uptime': '68802632'},
                                 {       'counter_source': '717/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:7a:a8'],
                                                                       'id': '717/g1',
                                                                       'in_nucast_pkts': '10446',
                                                                       'in_octets': '1654849605',
                                                                       'in_ucast_pkts': '18965532',
                                                                       'mac_address': 'ec:4f:82:8d:7e:0b',
                                                                       'mtu': 2000,
                                                                       'name': '717/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1258919',
                                                                       'out_octets': '127034887825',
                                                                       'out_ucast_pkts': '84875518',
                                                                       'poller_hash': 'MzIxODA4cGVlcmludGY3MTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '717',
                                                                       'in_nucast_pkts': '1258919',
                                                                       'in_octets': '127034887825',
                                                                       'in_ucast_pkts': '84875518',
                                                                       'links': [       {       'in_nucast_pkts': '1258919',
                                                                                                'in_octets': '127034887825',
                                                                                                'in_ucast_pkts': '84875518',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10446',
                                                                                                'out_octets': '1654849605',
                                                                                                'out_ucast_pkts': '18965532',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10446',
                                                                       'out_octets': '1654849605',
                                                                       'out_ucast_pkts': '18965532',
                                                                       'poller_hash': 'MzIxODA4cGVlcmludGZwb24gNzE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '717',
                                                                                               'links': [       {       'connected_time': '597547',
                                                                                                                        'fiber_length': 1287,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxODA4cGVlcnBvbnBvbiA3MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '24 Arena Dr Milverton (Shirley Weber '
                                                 '10401054)',
                                         'poller_hash': 'MzIxODA4',
                                         'serial_number': '321808',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:7e:08',
                                         'uptime': '597547'},
                                 {       'counter_source': '719/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b8:4e'],
                                                                       'id': '719/g1',
                                                                       'in_nucast_pkts': '18077709',
                                                                       'in_octets': '738119705400',
                                                                       'in_ucast_pkts': '3007608060',
                                                                       'mac_address': 'ec:4f:82:8d:76:cb',
                                                                       'mtu': 2000,
                                                                       'name': '719/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '46297214534',
                                                                       'out_octets': '80582474568245',
                                                                       'out_ucast_pkts': '13072396157',
                                                                       'poller_hash': 'MzIxNzIwcGVlcmludGY3MTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '719',
                                                                       'in_nucast_pkts': '46297214534',
                                                                       'in_octets': '80582474568245',
                                                                       'in_ucast_pkts': '13072396157',
                                                                       'links': [       {       'in_nucast_pkts': '46297214534',
                                                                                                'in_octets': '80582474568245',
                                                                                                'in_ucast_pkts': '13072396157',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18077709',
                                                                                                'out_octets': '738119705400',
                                                                                                'out_ucast_pkts': '3007608060',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18077709',
                                                                       'out_octets': '738119705400',
                                                                       'out_ucast_pkts': '3007608060',
                                                                       'poller_hash': 'MzIxNzIwcGVlcmludGZwb24gNzE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '719',
                                                                                               'links': [       {       'connected_time': '21815946',
                                                                                                                        'fiber_length': 11020,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxNzIwcGVlcnBvbnBvbiA3MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5842 Perth Road 119 Brunner (Don & '
                                                 'Eva Anderson 10430386)',
                                         'poller_hash': 'MzIxNzIw',
                                         'serial_number': '321720',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:76:c8',
                                         'uptime': '21815946'},
                                 {       'counter_source': '720/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1d:e8'],
                                                                       'id': '720/g1',
                                                                       'in_nucast_pkts': '13034',
                                                                       'in_octets': '7583101259',
                                                                       'in_ucast_pkts': '14512786',
                                                                       'mac_address': 'ec:4f:82:8d:76:a3',
                                                                       'mtu': 2000,
                                                                       'name': '720/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463849',
                                                                       'out_octets': '80565339589',
                                                                       'out_ucast_pkts': '59948198',
                                                                       'poller_hash': 'MzIxNzFCcGVlcmludGY3MjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '720',
                                                                       'in_nucast_pkts': '1463849',
                                                                       'in_octets': '80565339589',
                                                                       'in_ucast_pkts': '59948198',
                                                                       'links': [       {       'in_nucast_pkts': '1463849',
                                                                                                'in_octets': '80565339589',
                                                                                                'in_ucast_pkts': '59948198',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13034',
                                                                                                'out_octets': '7583101259',
                                                                                                'out_ucast_pkts': '14512786',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13034',
                                                                       'out_octets': '7583101259',
                                                                       'out_ucast_pkts': '14512786',
                                                                       'poller_hash': 'MzIxNzFCcGVlcmludGZwb24gNzIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '720',
                                                                                               'links': [       {       'connected_time': '688505',
                                                                                                                        'fiber_length': 1155,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxNzFCcGVlcnBvbnBvbiA3MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 PACIFIC AV (MILOVANCEV CHLOE AND '
                                                 'NATHAN)',
                                         'poller_hash': 'MzIxNzFC',
                                         'serial_number': '32171B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:76:a0',
                                         'uptime': '688505'},
                                 {       'counter_source': '721/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:64:8c'],
                                                                       'id': '721/g1',
                                                                       'in_nucast_pkts': '14490',
                                                                       'in_octets': '5979215056',
                                                                       'in_ucast_pkts': '38867197',
                                                                       'mac_address': 'ec:4f:82:0d:68:7d',
                                                                       'mtu': 2000,
                                                                       'name': '721/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463978',
                                                                       'out_octets': '150299729619',
                                                                       'out_ucast_pkts': '108829279',
                                                                       'poller_hash': 'MjA1REJEcGVlcmludGY3MjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '721',
                                                                       'in_nucast_pkts': '1463978',
                                                                       'in_octets': '150299729619',
                                                                       'in_ucast_pkts': '108829279',
                                                                       'links': [       {       'in_nucast_pkts': '1463978',
                                                                                                'in_octets': '150299729619',
                                                                                                'in_ucast_pkts': '108829279',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14490',
                                                                                                'out_octets': '5979215056',
                                                                                                'out_ucast_pkts': '38867197',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14490',
                                                                       'out_octets': '5979215056',
                                                                       'out_ucast_pkts': '38867197',
                                                                       'poller_hash': 'MjA1REJEcGVlcmludGZwb24gNzIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '721',
                                                                                               'links': [       {       'connected_time': '688491',
                                                                                                                        'fiber_length': 1029,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REJEcGVlcnBvbnBvbiA3MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '23 PACIFIC AV (WILLIAMS TY)',
                                         'poller_hash': 'MjA1REJE',
                                         'serial_number': '205DBD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:7a',
                                         'uptime': '688491'},
                                 {       'counter_source': '722/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:8f:0b'],
                                                                       'id': '722/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '7950428280',
                                                                       'in_ucast_pkts': '66610030',
                                                                       'mac_address': 'cc:be:59:cb:43:fd',
                                                                       'mtu': 2000,
                                                                       'name': '722/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447218',
                                                                       'out_octets': '208520518654',
                                                                       'out_ucast_pkts': '142714498',
                                                                       'poller_hash': 'NDJCOUQ3cGVlcmludGY3MjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '722',
                                                                       'in_nucast_pkts': '1447218',
                                                                       'in_octets': '208520518654',
                                                                       'in_ucast_pkts': '142714498',
                                                                       'links': [       {       'in_nucast_pkts': '1447218',
                                                                                                'in_octets': '208520518654',
                                                                                                'in_ucast_pkts': '142714498',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '7950428280',
                                                                                                'out_ucast_pkts': '66610030',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '7950428280',
                                                                       'out_ucast_pkts': '66610030',
                                                                       'poller_hash': 'NDJCOUQ3cGVlcmludGZwb24gNzIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '722',
                                                                                               'links': [       {       'connected_time': '688413',
                                                                                                                        'fiber_length': 2227,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUQ3cGVlcnBvbnBvbiA3MjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '75 Gerber Drive Milverton (Cezary '
                                                 'Cybulski 10434099)',
                                         'poller_hash': 'NDJCOUQ3',
                                         'serial_number': '42B9D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:fa',
                                         'uptime': '688413'},
                                 {       'counter_source': '724/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '724/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '210530839960',
                                                                       'in_ucast_pkts': '1235552554',
                                                                       'mac_address': 'ec:4f:82:8d:84:f0',
                                                                       'name': '724/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '266',
                                                                       'out_octets': '5116382390451',
                                                                       'out_ucast_pkts': '3706815966',
                                                                       'poller_hash': 'MzIxOEU1cGVlcmludGY3MjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '724',
                                                                       'in_nucast_pkts': '266',
                                                                       'in_octets': '5116382390451',
                                                                       'in_ucast_pkts': '3706815966',
                                                                       'links': [       {       'in_nucast_pkts': '266',
                                                                                                'in_octets': '5116382390451',
                                                                                                'in_ucast_pkts': '3706815966',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '210530839960',
                                                                                                'out_ucast_pkts': '1235552554',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '210530839960',
                                                                       'out_ucast_pkts': '1235552554',
                                                                       'poller_hash': 'MzIxOEU1cGVlcmludGZwb24gNzI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '724',
                                                                                               'links': [       {       'connected_time': '15498720',
                                                                                                                        'fiber_length': 1905,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxOEU1cGVlcnBvbnBvbiA3MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '54 MAIN ST N (GRANT ERIKA)',
                                         'poller_hash': 'MzIxOEU1',
                                         'serial_number': '3218E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:84:f0',
                                         'uptime': '15498720'},
                                 {       'counter_source': '725/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '725/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '40494736594',
                                                                       'in_ucast_pkts': '211082762',
                                                                       'mac_address': 'ec:4f:82:0d:67:48',
                                                                       'name': '725/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '1403644700367',
                                                                       'out_ucast_pkts': '1031892037',
                                                                       'poller_hash': 'MjA1RDlCcGVlcmludGY3MjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '725',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '1403644700367',
                                                                       'in_ucast_pkts': '1031892037',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '1403644700367',
                                                                                                'in_ucast_pkts': '1031892037',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '40494736594',
                                                                                                'out_ucast_pkts': '211082762',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '40494736594',
                                                                       'out_ucast_pkts': '211082762',
                                                                       'poller_hash': 'MjA1RDlCcGVlcmludGZwb24gNzI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '725',
                                                                                               'links': [       {       'connected_time': '688491',
                                                                                                                        'fiber_length': 691,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RDlCcGVlcnBvbnBvbiA3MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '4 Mill St E Milverton Apt 2 (Maxine '
                                                 'Kuepfer)',
                                         'poller_hash': 'MjA1RDlC',
                                         'serial_number': '205D9B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:67:48',
                                         'uptime': '688491'},
                                 {       'counter_source': '727/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '727/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10862245217',
                                                                       'in_ucast_pkts': '87699159',
                                                                       'mac_address': 'ec:4f:82:8d:7c:28',
                                                                       'name': '727/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '172',
                                                                       'out_octets': '484311763963',
                                                                       'out_ucast_pkts': '378802987',
                                                                       'poller_hash': 'MzIxN0NDcGVlcmludGY3MjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '727',
                                                                       'in_nucast_pkts': '172',
                                                                       'in_octets': '484311763963',
                                                                       'in_ucast_pkts': '378802987',
                                                                       'links': [       {       'in_nucast_pkts': '172',
                                                                                                'in_octets': '484311763963',
                                                                                                'in_ucast_pkts': '378802987',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '10862245217',
                                                                                                'out_ucast_pkts': '87699159',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10862245217',
                                                                       'out_ucast_pkts': '87699159',
                                                                       'poller_hash': 'MzIxN0NDcGVlcmludGZwb24gNzI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '727',
                                                                                               'links': [       {       'connected_time': '506878',
                                                                                                                        'fiber_length': 18537,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxN0NDcGVlcnBvbnBvbiA3Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '68 Schade St Monkton (Ross Rennick '
                                                 '10434031)',
                                         'poller_hash': 'MzIxN0ND',
                                         'serial_number': '3217CC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:7c:28',
                                         'uptime': '506878'},
                                 {       'counter_source': '728/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2f:2f:eb'],
                                                                       'id': '728/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '24126251371',
                                                                       'in_ucast_pkts': '177352498',
                                                                       'mac_address': 'ec:4f:82:8d:79:93',
                                                                       'mtu': 2000,
                                                                       'name': '728/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720873',
                                                                       'out_octets': '813670843902',
                                                                       'out_ucast_pkts': '601043869',
                                                                       'poller_hash': 'MzIxNzc5cGVlcmludGY3MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '728',
                                                                       'in_nucast_pkts': '6720873',
                                                                       'in_octets': '813670843902',
                                                                       'in_ucast_pkts': '601043869',
                                                                       'links': [       {       'in_nucast_pkts': '6720873',
                                                                                                'in_octets': '813670843902',
                                                                                                'in_ucast_pkts': '601043869',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '24126251371',
                                                                                                'out_ucast_pkts': '177352498',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '24126251371',
                                                                       'out_ucast_pkts': '177352498',
                                                                       'poller_hash': 'MzIxNzc5cGVlcmludGZwb24gNzI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '728',
                                                                                               'links': [       {       'connected_time': '3113993',
                                                                                                                        'fiber_length': 17923,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxNzc5cGVlcnBvbnBvbiA3Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '160 Smith St Monkton (Carl Steicher '
                                                 '10434033)',
                                         'poller_hash': 'MzIxNzc5',
                                         'serial_number': '321779',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:79:90',
                                         'uptime': '3113993'},
                                 {       'counter_source': '729/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:32:1b:52:b3:c6'],
                                                                       'id': '729/g1',
                                                                       'in_nucast_pkts': '404335',
                                                                       'in_octets': '1034836608690',
                                                                       'in_ucast_pkts': '3055175465',
                                                                       'mac_address': 'ec:4f:82:7e:36:0d',
                                                                       'mtu': 2000,
                                                                       'name': '729/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83521248',
                                                                       'out_octets': '7638431108637',
                                                                       'out_ucast_pkts': '6800845002',
                                                                       'poller_hash': 'MzVGODQzcGVlcmludGY3MjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '729',
                                                                       'in_nucast_pkts': '83521248',
                                                                       'in_octets': '7638431108637',
                                                                       'in_ucast_pkts': '6800845002',
                                                                       'links': [       {       'in_nucast_pkts': '83521248',
                                                                                                'in_octets': '7638431108637',
                                                                                                'in_ucast_pkts': '6800845002',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '404335',
                                                                                                'out_octets': '1034836608690',
                                                                                                'out_ucast_pkts': '3055175465',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '404335',
                                                                       'out_octets': '1034836608690',
                                                                       'out_ucast_pkts': '3055175465',
                                                                       'poller_hash': 'MzVGODQzcGVlcmludGZwb24gNzI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '729',
                                                                                               'links': [       {       'connected_time': '67790186',
                                                                                                                        'fiber_length': 14054,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGODQzcGVlcnBvbnBvbiA3Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5620 Perth Rd 119 Brunner (James & '
                                                 'Donna Jantzi 10423020)',
                                         'poller_hash': 'MzVGODQz',
                                         'serial_number': '35F843',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:36:0d',
                                         'uptime': '67790186'},
                                 {       'counter_source': '730/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:69:89'],
                                                                       'id': '730/g1',
                                                                       'in_nucast_pkts': '54158',
                                                                       'in_octets': '15176311322',
                                                                       'in_ucast_pkts': '70038195',
                                                                       'mac_address': 'ec:4f:82:8d:88:23',
                                                                       'mtu': 2000,
                                                                       'name': '730/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720776',
                                                                       'out_octets': '712164327937',
                                                                       'out_ucast_pkts': '512131825',
                                                                       'poller_hash': 'MzIxOTRCcGVlcmludGY3MzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '730',
                                                                       'in_nucast_pkts': '6720776',
                                                                       'in_octets': '712164327937',
                                                                       'in_ucast_pkts': '512131825',
                                                                       'links': [       {       'in_nucast_pkts': '6720776',
                                                                                                'in_octets': '712164327937',
                                                                                                'in_ucast_pkts': '512131825',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '54158',
                                                                                                'out_octets': '15176311322',
                                                                                                'out_ucast_pkts': '70038195',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '54158',
                                                                       'out_octets': '15176311322',
                                                                       'out_ucast_pkts': '70038195',
                                                                       'poller_hash': 'MzIxOTRCcGVlcmludGZwb24gNzMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '730',
                                                                                               'links': [       {       'connected_time': '3113982',
                                                                                                                        'fiber_length': 18248,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxOTRCcGVlcnBvbnBvbiA3MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 Schade St Monkton (STAHLKE '
                                                 'KRISTEN)',
                                         'poller_hash': 'MzIxOTRC',
                                         'serial_number': '32194B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:88:20',
                                         'uptime': '3113982'},
                                 {       'counter_source': '731/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:26'],
                                                                       'id': '731/g1',
                                                                       'in_nucast_pkts': '11965',
                                                                       'in_octets': '12566288977',
                                                                       'in_ucast_pkts': '138653206',
                                                                       'mac_address': 'ec:4f:82:8d:82:f3',
                                                                       'mtu': 2000,
                                                                       'name': '731/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447165',
                                                                       'out_octets': '545874527277',
                                                                       'out_ucast_pkts': '364554604',
                                                                       'poller_hash': 'MzIxOEE1cGVlcmludGY3MzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '731',
                                                                       'in_nucast_pkts': '1447165',
                                                                       'in_octets': '545874527277',
                                                                       'in_ucast_pkts': '364554604',
                                                                       'links': [       {       'in_nucast_pkts': '1447165',
                                                                                                'in_octets': '545874527277',
                                                                                                'in_ucast_pkts': '364554604',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11965',
                                                                                                'out_octets': '12566288977',
                                                                                                'out_ucast_pkts': '138653206',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11965',
                                                                       'out_octets': '12566288977',
                                                                       'out_ucast_pkts': '138653206',
                                                                       'poller_hash': 'MzIxOEE1cGVlcmludGZwb24gNzMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '731',
                                                                                               'links': [       {       'connected_time': '688420',
                                                                                                                        'fiber_length': 1989,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxOEE1cGVlcnBvbnBvbiA3MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '91 GERBER DR (DE KLEER SARAH)',
                                         'poller_hash': 'MzIxOEE1',
                                         'serial_number': '3218A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:82:f0',
                                         'uptime': '688420'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '732/g1',
                                                                       'name': '732/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwNzlEcGVlcmludGY3MzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '732',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwNzlEcGVlcmludGZwb24gNzMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '732',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwNzlEcGVlcnBvbnBvbiA3MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '55 William St apt 207 Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "732" '
                                                                  '(location: GPON '
                                                                  'port "7/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:39.850Z',
                                         'poller_hash': 'MzAwNzlE',
                                         'serial_number': '30079D',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '733/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '733/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '379243053914',
                                                                       'in_ucast_pkts': '2271934954',
                                                                       'mac_address': 'ec:4f:82:0d:67:99',
                                                                       'name': '733/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '9173905381904',
                                                                       'out_ucast_pkts': '6349221367',
                                                                       'poller_hash': 'MjA1REE0cGVlcmludGY3MzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '733',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '9173905381904',
                                                                       'in_ucast_pkts': '6349221367',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '9173905381904',
                                                                                                'in_ucast_pkts': '6349221367',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '379243053914',
                                                                                                'out_ucast_pkts': '2271934954',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '379243053914',
                                                                       'out_ucast_pkts': '2271934954',
                                                                       'poller_hash': 'MjA1REE0cGVlcmludGZwb24gNzMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '733',
                                                                                               'links': [       {       'connected_time': '1646454',
                                                                                                                        'fiber_length': 18040,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REE0cGVlcnBvbnBvbiA3MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '43 Maddison St E Monkton (Christine '
                                                 'Rose 10432948)',
                                         'poller_hash': 'MjA1REE0',
                                         'serial_number': '205DA4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:67:99',
                                         'uptime': '1646454'},
                                 {       'counter_source': '734/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '734/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2026704984974',
                                                                       'in_ucast_pkts': '5341146948',
                                                                       'mac_address': 'cc:be:59:eb:3b:9e',
                                                                       'name': '734/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12094145591570',
                                                                       'out_ucast_pkts': '13243070417',
                                                                       'poller_hash': 'NTU2M0U5cGVlcmludGY3MzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '734',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12094145591570',
                                                                       'in_ucast_pkts': '13243070417',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '12094145591570',
                                                                                                'in_ucast_pkts': '13243070417',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2026704984974',
                                                                                                'out_ucast_pkts': '5341146948',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2026704984974',
                                                                       'out_ucast_pkts': '5341146948',
                                                                       'poller_hash': 'NTU2M0U5cGVlcmludGZwb24gNzM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '734',
                                                                                               'links': [       {       'connected_time': '15769267',
                                                                                                                        'fiber_length': 2094,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.438,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU2M0U5cGVlcnBvbnBvbiA3MzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '71 REAGAN ST (CAO QUI PHUC)',
                                         'poller_hash': 'NTU2M0U5',
                                         'serial_number': '5563E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:3b:9e',
                                         'uptime': '15769267'},
                                 {       'counter_source': '735/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '735/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '314064913707',
                                                                       'in_ucast_pkts': '82553151448',
                                                                       'mac_address': 'ec:4f:82:8d:55:40',
                                                                       'name': '735/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '128',
                                                                       'out_octets': '4576143080263',
                                                                       'out_ucast_pkts': '3344980432',
                                                                       'poller_hash': 'MzIxMkVGcGVlcmludGY3MzUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '735',
                                                                       'in_nucast_pkts': '128',
                                                                       'in_octets': '4576143080263',
                                                                       'in_ucast_pkts': '3344980432',
                                                                       'links': [       {       'in_nucast_pkts': '128',
                                                                                                'in_octets': '4576143080263',
                                                                                                'in_ucast_pkts': '3344980432',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '314064913707',
                                                                                                'out_ucast_pkts': '82553151448',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '314064913707',
                                                                       'out_ucast_pkts': '82553151448',
                                                                       'poller_hash': 'MzIxMkVGcGVlcmludGZwb24gNzM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '735',
                                                                                               'links': [       {       'connected_time': '688441',
                                                                                                                        'fiber_length': 1469,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMkVGcGVlcnBvbnBvbiA3MzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '80 Mill St. E. Milverton (Lynda '
                                                 'Kuepfer)',
                                         'poller_hash': 'MzIxMkVG',
                                         'serial_number': '3212EF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:55:40',
                                         'uptime': '688441'},
                                 {       'counter_source': '736/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a4:2b:b0:ee:cc:2b'],
                                                                       'id': '736/g1',
                                                                       'in_nucast_pkts': '94563',
                                                                       'in_octets': '97704747022',
                                                                       'in_ucast_pkts': '316212198',
                                                                       'mac_address': 'ec:4f:82:7e:3a:54',
                                                                       'mtu': 2000,
                                                                       'name': '736/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83086810',
                                                                       'out_octets': '1594483155932',
                                                                       'out_ucast_pkts': '1295561950',
                                                                       'poller_hash': 'MzVGOUIwcGVlcmludGY3MzYvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '736',
                                                                       'in_nucast_pkts': '83086810',
                                                                       'in_octets': '1594483155932',
                                                                       'in_ucast_pkts': '1295561950',
                                                                       'links': [       {       'in_nucast_pkts': '83086810',
                                                                                                'in_octets': '1594483155932',
                                                                                                'in_ucast_pkts': '1295561950',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '94563',
                                                                                                'out_octets': '97704747022',
                                                                                                'out_ucast_pkts': '316212198',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '94563',
                                                                       'out_octets': '97704747022',
                                                                       'out_ucast_pkts': '316212198',
                                                                       'poller_hash': 'MzVGOUIwcGVlcmludGZwb24gNzM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '736',
                                                                                               'links': [       {       'connected_time': '67790420',
                                                                                                                        'fiber_length': 1047,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUIwcGVlcnBvbnBvbiA3MzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '50 Mill St E Milverton (Barb '
                                                 'Chalmers 10400152)',
                                         'poller_hash': 'MzVGOUIw',
                                         'serial_number': '35F9B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3a:54',
                                         'uptime': '67790420'},
                                 {       'counter_source': '738/F1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:70:de',
                                                                                                        '00:03:e6:a5:73:90'],
                                                                       'id': '738/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '862122861',
                                                                       'in_ucast_pkts': '5005525',
                                                                       'name': '738/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '4576378343',
                                                                       'out_octets': '6152953088334',
                                                                       'out_ucast_pkts': '2383032',
                                                                       'poller_hash': 'MzIxODI1cGVlcmludGY3MzgvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '738',
                                                                       'in_nucast_pkts': '4576378343',
                                                                       'in_octets': '6152953088334',
                                                                       'in_ucast_pkts': '2383032',
                                                                       'links': [       {       'in_nucast_pkts': '4576378343',
                                                                                                'in_octets': '6152953088334',
                                                                                                'in_ucast_pkts': '2383032',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '862122861',
                                                                                                'out_ucast_pkts': '5005525',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '862122861',
                                                                       'out_ucast_pkts': '5005525',
                                                                       'poller_hash': 'MzIxODI1cGVlcmludGZwb24gNzM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '738',
                                                                                               'links': [       {       'connected_time': '21816241',
                                                                                                                        'fiber_length': 18744,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxODI1cGVlcnBvbnBvbiA3Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '93 Schade St Monkton (Steve Smith '
                                                 '10434116)',
                                         'poller_hash': 'MzIxODI1',
                                         'serial_number': '321825',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:7e:f0',
                                         'uptime': '21816241'},
                                 {       'counter_source': '739/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:c5:2f'],
                                                                       'id': '739/g1',
                                                                       'in_nucast_pkts': '33450',
                                                                       'in_octets': '24485303282',
                                                                       'in_ucast_pkts': '109350361',
                                                                       'mac_address': 'ec:4f:82:7d:c8:df',
                                                                       'mtu': 2000,
                                                                       'name': '739/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1480856',
                                                                       'out_octets': '421712553869',
                                                                       'out_ucast_pkts': '320216834',
                                                                       'poller_hash': 'MzAwRDlDcGVlcmludGY3MzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '739',
                                                                       'in_nucast_pkts': '1480856',
                                                                       'in_octets': '421712553869',
                                                                       'in_ucast_pkts': '320216834',
                                                                       'links': [       {       'in_nucast_pkts': '1480856',
                                                                                                'in_octets': '421712553869',
                                                                                                'in_ucast_pkts': '320216834',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '33450',
                                                                                                'out_octets': '24485303282',
                                                                                                'out_ucast_pkts': '109350361',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33450',
                                                                       'out_octets': '24485303282',
                                                                       'out_ucast_pkts': '109350361',
                                                                       'poller_hash': 'MzAwRDlDcGVlcmludGZwb24gNzM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '739',
                                                                                               'links': [       {       'connected_time': '688414',
                                                                                                                        'fiber_length': 2138,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRDlDcGVlcnBvbnBvbiA3Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '87 Gerber Dr (CUNHA CARLOS)',
                                         'poller_hash': 'MzAwRDlD',
                                         'serial_number': '300D9C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c8:df',
                                         'uptime': '688414'},
                                 {       'counter_source': '740/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:31:bf:e5:79:00'],
                                                                       'id': '740/g1',
                                                                       'in_nucast_pkts': '4793',
                                                                       'in_octets': '309756736274',
                                                                       'in_ucast_pkts': '568144457',
                                                                       'mac_address': 'ec:4f:82:7e:09:31',
                                                                       'mtu': 2000,
                                                                       'name': '740/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83116116',
                                                                       'out_octets': '1173155527394',
                                                                       'out_ucast_pkts': '985417435',
                                                                       'poller_hash': 'MzVFOTVCcGVlcmludGY3NDAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '740',
                                                                       'in_nucast_pkts': '83116116',
                                                                       'in_octets': '1173155527394',
                                                                       'in_ucast_pkts': '985417435',
                                                                       'links': [       {       'in_nucast_pkts': '83116116',
                                                                                                'in_octets': '1173155527394',
                                                                                                'in_ucast_pkts': '985417435',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4793',
                                                                                                'out_octets': '309756736274',
                                                                                                'out_ucast_pkts': '568144457',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4793',
                                                                       'out_octets': '309756736274',
                                                                       'out_ucast_pkts': '568144457',
                                                                       'poller_hash': 'MzVFOTVCcGVlcmludGZwb24gNzQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '740',
                                                                                               'links': [       {       'connected_time': '67790433',
                                                                                                                        'fiber_length': 13813,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTVCcGVlcnBvbnBvbiA3NDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5845 Perth Rd 119 (Matthew & Sheila '
                                                 'Ropp 10420100)',
                                         'poller_hash': 'MzVFOTVC',
                                         'serial_number': '35E95B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:31',
                                         'uptime': '67790433'},
                                 {       'counter_source': '741/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '10:27:f5:63:45:ad'],
                                                                       'id': '741/g1',
                                                                       'in_nucast_pkts': '217',
                                                                       'in_octets': '1594209131',
                                                                       'in_ucast_pkts': '5685898',
                                                                       'mac_address': '44:65:7f:7b:d4:45',
                                                                       'mtu': 2000,
                                                                       'name': '741/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120465',
                                                                       'out_octets': '28201207249',
                                                                       'out_ucast_pkts': '22146071',
                                                                       'poller_hash': 'NUE2QkIycGVlcmludGY3NDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '741',
                                                                       'in_nucast_pkts': '1120465',
                                                                       'in_octets': '28201207249',
                                                                       'in_ucast_pkts': '22146071',
                                                                       'links': [       {       'in_nucast_pkts': '1120465',
                                                                                                'in_octets': '28201207249',
                                                                                                'in_ucast_pkts': '22146071',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '217',
                                                                                                'out_octets': '1594209131',
                                                                                                'out_ucast_pkts': '5685898',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '217',
                                                                       'out_octets': '1594209131',
                                                                       'out_ucast_pkts': '5685898',
                                                                       'poller_hash': 'NUE2QkIycGVlcmludGZwb24gNzQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '741',
                                                                                               'links': [       {       'connected_time': '538568',
                                                                                                                        'fiber_length': 16991,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QkIycGVlcnBvbnBvbiA3NDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5173 Road 150 Gadshill (NATYWARY '
                                                 'TYLER)',
                                         'poller_hash': 'NUE2QkIy',
                                         'serial_number': '5A6BB2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d4:45',
                                         'uptime': '538568'},
                                 {       'counter_source': '742/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c0:a0:bb:e7:a9:40'],
                                                                       'id': '742/g1',
                                                                       'in_nucast_pkts': '70371',
                                                                       'in_octets': '47219009079',
                                                                       'in_ucast_pkts': '60765529',
                                                                       'mac_address': '44:65:7f:7b:d3:af',
                                                                       'mtu': 2000,
                                                                       'name': '742/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120473',
                                                                       'out_octets': '128792266581',
                                                                       'out_ucast_pkts': '100239284',
                                                                       'poller_hash': 'NUE2QjgwcGVlcmludGY3NDIvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '742',
                                                                       'in_nucast_pkts': '1120473',
                                                                       'in_octets': '128792266581',
                                                                       'in_ucast_pkts': '100239284',
                                                                       'links': [       {       'in_nucast_pkts': '1120473',
                                                                                                'in_octets': '128792266581',
                                                                                                'in_ucast_pkts': '100239284',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '70371',
                                                                                                'out_octets': '47219009079',
                                                                                                'out_ucast_pkts': '60765529',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '70371',
                                                                       'out_octets': '47219009079',
                                                                       'out_ucast_pkts': '60765529',
                                                                       'poller_hash': 'NUE2QjgwcGVlcmludGZwb24gNzQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '742',
                                                                                               'links': [       {       'connected_time': '538566',
                                                                                                                        'fiber_length': 15829,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjgwcGVlcnBvbnBvbiA3NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5280 LINE 49 (YOUNGBLUT BILL & '
                                                 'SARAH)',
                                         'poller_hash': 'NUE2Qjgw',
                                         'serial_number': '5A6B80',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:af',
                                         'uptime': '538566'},
                                 {       'counter_source': '744/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'bc:ee:7b:ec:b5:38'],
                                                                       'id': '744/g1',
                                                                       'in_nucast_pkts': '15',
                                                                       'in_octets': '819318223',
                                                                       'in_ucast_pkts': '5803595',
                                                                       'mac_address': 'ec:4f:82:7e:0d:0f',
                                                                       'mtu': 2000,
                                                                       'name': '744/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464212',
                                                                       'out_octets': '21929295035',
                                                                       'out_ucast_pkts': '15994197',
                                                                       'poller_hash': 'MzVFQUE1cGVlcmludGY3NDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '744',
                                                                       'in_nucast_pkts': '1464212',
                                                                       'in_octets': '21929295035',
                                                                       'in_ucast_pkts': '15994197',
                                                                       'links': [       {       'in_nucast_pkts': '1464212',
                                                                                                'in_octets': '21929295035',
                                                                                                'in_ucast_pkts': '15994197',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15',
                                                                                                'out_octets': '819318223',
                                                                                                'out_ucast_pkts': '5803595',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15',
                                                                       'out_octets': '819318223',
                                                                       'out_ucast_pkts': '5803595',
                                                                       'poller_hash': 'MzVFQUE1cGVlcmludGZwb24gNzQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '744',
                                                                                               'links': [       {       'connected_time': '688524',
                                                                                                                        'fiber_length': 1168,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQUE1cGVlcnBvbnBvbiA3NDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '17 Fulton St Milverton (Edward Wiebe '
                                                 '10426867)',
                                         'poller_hash': 'MzVFQUE1',
                                         'serial_number': '35EAA5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0d:0f',
                                         'uptime': '688524'},
                                 {       'counter_source': '745/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '38:d5:47:87:96:d0'],
                                                                       'id': '745/g1',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '13311432235',
                                                                       'in_ucast_pkts': '17483547',
                                                                       'mac_address': 'cc:be:59:4f:ff:da',
                                                                       'mtu': 2000,
                                                                       'name': '745/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '851239',
                                                                       'out_octets': '61576493114',
                                                                       'out_ucast_pkts': '46473240',
                                                                       'poller_hash': 'NDNCM0I2cGVlcmludGY3NDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '745',
                                                                       'in_nucast_pkts': '851239',
                                                                       'in_octets': '61576493114',
                                                                       'in_ucast_pkts': '46473240',
                                                                       'links': [       {       'in_nucast_pkts': '851239',
                                                                                                'in_octets': '61576493114',
                                                                                                'in_ucast_pkts': '46473240',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19',
                                                                                                'out_octets': '13311432235',
                                                                                                'out_ucast_pkts': '17483547',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19',
                                                                       'out_octets': '13311432235',
                                                                       'out_ucast_pkts': '17483547',
                                                                       'poller_hash': 'NDNCM0I2cGVlcmludGZwb24gNzQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '745',
                                                                                               'links': [       {       'connected_time': '538551',
                                                                                                                        'fiber_length': 13447,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.742,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.22}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0I2cGVlcnBvbnBvbiA3NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '4934 LINE 49 (LEYBOURNE KEVIN)',
                                         'poller_hash': 'NDNCM0I2',
                                         'serial_number': '43B3B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:ff:d9',
                                         'uptime': '538551'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '748/g1',
                                                                       'name': '748/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVFOTcycGVlcmludGY3NDgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '748',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzVFOTcycGVlcmludGZwb24gNzQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '748',
                                                                                               'links': [       {       'connected_time': '15499147',
                                                                                                                        'fiber_length': 685,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTcycGVlcnBvbnBvbiA3NDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1 Main St. N. Milverton (Mane Appeal '
                                                 '(Phone))',
                                         'poller_hash': 'MzVFOTcy',
                                         'serial_number': '35E972',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:76',
                                         'uptime': '15499147'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '750/g1',
                                                                       'name': '750/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwN0FGcGVlcmludGY3NTAvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '750',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwN0FGcGVlcmludGZwb24gNzUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '750',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwN0FGcGVlcnBvbnBvbiA3NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '5907 Perth Road 119 Brunner',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "750" '
                                                                  '(location: GPON '
                                                                  'port "11/6"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:03:10.770Z',
                                         'poller_hash': 'MzAwN0FG',
                                         'serial_number': '3007AF',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '751/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:37:15'],
                                                                       'id': '751/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '7680599744',
                                                                       'in_ucast_pkts': '46407311',
                                                                       'mac_address': '44:65:7f:7b:d1:f9',
                                                                       'mtu': 2000,
                                                                       'name': '751/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120471',
                                                                       'out_octets': '150089357437',
                                                                       'out_ucast_pkts': '115315392',
                                                                       'poller_hash': 'NUE2QUVFcGVlcmludGY3NTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '751',
                                                                       'in_nucast_pkts': '1120471',
                                                                       'in_octets': '150089357437',
                                                                       'in_ucast_pkts': '115315392',
                                                                       'links': [       {       'in_nucast_pkts': '1120471',
                                                                                                'in_octets': '150089357437',
                                                                                                'in_ucast_pkts': '115315392',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '7680599744',
                                                                                                'out_ucast_pkts': '46407311',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '7680599744',
                                                                       'out_ucast_pkts': '46407311',
                                                                       'poller_hash': 'NUE2QUVFcGVlcmludGZwb24gNzUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '751',
                                                                                               'links': [       {       'connected_time': '538559',
                                                                                                                        'fiber_length': 11153,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QUVFcGVlcnBvbnBvbiA3NTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5358 Road 140 Brunner (Martin '
                                                 'Brodhagen 10401375)',
                                         'poller_hash': 'NUE2QUVF',
                                         'serial_number': '5A6AEE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d1:f9',
                                         'uptime': '538559'},
                                 {       'counter_source': '752/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:70:3c'],
                                                                       'id': '752/g1',
                                                                       'in_nucast_pkts': '344908',
                                                                       'in_octets': '14319180518',
                                                                       'in_ucast_pkts': '78349702',
                                                                       'mac_address': 'ec:4f:82:7e:09:0d',
                                                                       'mtu': 2000,
                                                                       'name': '752/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39761391',
                                                                       'out_octets': '1169816886755',
                                                                       'out_ucast_pkts': '779128894',
                                                                       'poller_hash': 'MzVFOTRGcGVlcmludGY3NTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '752',
                                                                       'in_nucast_pkts': '39761391',
                                                                       'in_octets': '1169816886755',
                                                                       'in_ucast_pkts': '779128894',
                                                                       'links': [       {       'in_nucast_pkts': '39761391',
                                                                                                'in_octets': '1169816886755',
                                                                                                'in_ucast_pkts': '779128894',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '344908',
                                                                                                'out_octets': '14319180518',
                                                                                                'out_ucast_pkts': '78349702',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '344908',
                                                                       'out_octets': '14319180518',
                                                                       'out_ucast_pkts': '78349702',
                                                                       'poller_hash': 'MzVFOTRGcGVlcmludGZwb24gNzUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '752',
                                                                                               'links': [       {       'connected_time': '29575204',
                                                                                                                        'fiber_length': 1188,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTRGcGVlcnBvbnBvbiA3NTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '53 MILL ST W (PAULI ROGER)',
                                         'poller_hash': 'MzVFOTRG',
                                         'serial_number': '35E94F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:0d',
                                         'uptime': '29575204'},
                                 {       'counter_source': '753/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c0:c9:e3:09:df:e5'],
                                                                       'id': '753/g1',
                                                                       'in_nucast_pkts': '45970536',
                                                                       'in_octets': '1137356152878',
                                                                       'in_ucast_pkts': '4168550630',
                                                                       'mac_address': '44:65:7f:7b:d4:39',
                                                                       'mtu': 2000,
                                                                       'name': '753/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82673312',
                                                                       'out_octets': '14279938229919',
                                                                       'out_ucast_pkts': '10716155477',
                                                                       'poller_hash': 'NUE2QkFFcGVlcmludGY3NTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '753',
                                                                       'in_nucast_pkts': '82673312',
                                                                       'in_octets': '14279938229919',
                                                                       'in_ucast_pkts': '10716155477',
                                                                       'links': [       {       'in_nucast_pkts': '82673312',
                                                                                                'in_octets': '14279938229919',
                                                                                                'in_ucast_pkts': '10716155477',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '45970536',
                                                                                                'out_octets': '1137356152878',
                                                                                                'out_ucast_pkts': '4168550630',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45970536',
                                                                       'out_octets': '1137356152878',
                                                                       'out_ucast_pkts': '4168550630',
                                                                       'poller_hash': 'NUE2QkFFcGVlcmludGZwb24gNzUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '753',
                                                                                               'links': [       {       'connected_time': '67790129',
                                                                                                                        'fiber_length': 989,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QkFFcGVlcnBvbnBvbiA3NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '23 William St Milverton (C & H '
                                                 'Lichti 10424574)',
                                         'poller_hash': 'NUE2QkFF',
                                         'serial_number': '5A6BAE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d4:39',
                                         'uptime': '67790129'},
                                 {       'counter_source': '754/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:47:32:21:9a:59'],
                                                                       'id': '754/g1',
                                                                       'in_nucast_pkts': '21704',
                                                                       'in_octets': '84979106379',
                                                                       'in_ucast_pkts': '314148003',
                                                                       'mac_address': 'ec:4f:82:7e:09:34',
                                                                       'mtu': 2000,
                                                                       'name': '754/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33278718',
                                                                       'out_octets': '1897268710109',
                                                                       'out_ucast_pkts': '1359604939',
                                                                       'poller_hash': 'MzVFOTVDcGVlcmludGY3NTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '754',
                                                                       'in_nucast_pkts': '33278718',
                                                                       'in_octets': '1897268710109',
                                                                       'in_ucast_pkts': '1359604939',
                                                                       'links': [       {       'in_nucast_pkts': '33278718',
                                                                                                'in_octets': '1897268710109',
                                                                                                'in_ucast_pkts': '1359604939',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21704',
                                                                                                'out_octets': '84979106379',
                                                                                                'out_ucast_pkts': '314148003',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21704',
                                                                       'out_octets': '84979106379',
                                                                       'out_ucast_pkts': '314148003',
                                                                       'poller_hash': 'MzVFOTVDcGVlcmludGZwb24gNzU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '754',
                                                                                               'links': [       {       'connected_time': '15499232',
                                                                                                                        'fiber_length': 12485,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTVDcGVlcnBvbnBvbiA3NTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': 'Arnold Yantzi 10400583 (5734 Perth '
                                                 'Rd 119)',
                                         'poller_hash': 'MzVFOTVD',
                                         'serial_number': '35E95C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:34',
                                         'uptime': '15499232'},
                                 {       'counter_source': '755/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '68:ff:7b:33:69:73'],
                                                                       'id': '755/g1',
                                                                       'in_nucast_pkts': '45',
                                                                       'in_octets': '2934199229',
                                                                       'in_ucast_pkts': '13708560',
                                                                       'mac_address': 'cc:be:59:b9:c3:1a',
                                                                       'mtu': 2000,
                                                                       'name': '755/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120474',
                                                                       'out_octets': '51980907290',
                                                                       'out_ucast_pkts': '39973066',
                                                                       'poller_hash': 'NTEyNEU2cGVlcmludGY3NTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '755',
                                                                       'in_nucast_pkts': '1120474',
                                                                       'in_octets': '51980907290',
                                                                       'in_ucast_pkts': '39973066',
                                                                       'links': [       {       'in_nucast_pkts': '1120474',
                                                                                                'in_octets': '51980907290',
                                                                                                'in_ucast_pkts': '39973066',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '45',
                                                                                                'out_octets': '2934199229',
                                                                                                'out_ucast_pkts': '13708560',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45',
                                                                       'out_octets': '2934199229',
                                                                       'out_ucast_pkts': '13708560',
                                                                       'poller_hash': 'NTEyNEU2cGVlcmludGZwb24gNzU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '755',
                                                                                               'links': [       {       'connected_time': '538569',
                                                                                                                        'fiber_length': 17126,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNEU2cGVlcnBvbnBvbiA3NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4621 Line 49 Brunner (Andrew Debus '
                                                 '10435392)',
                                         'poller_hash': 'NTEyNEU2',
                                         'serial_number': '5124E6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:1a',
                                         'uptime': '538569'},
                                 {       'counter_source': '756/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '2c:b0:5d:26:9f:9d'],
                                                                       'id': '756/g1',
                                                                       'in_nucast_pkts': '48028',
                                                                       'in_octets': '64079122534',
                                                                       'in_ucast_pkts': '205269869',
                                                                       'mac_address': 'ec:4f:82:be:a7:ad',
                                                                       'mtu': 2000,
                                                                       'name': '756/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '80843200',
                                                                       'out_octets': '316883391400',
                                                                       'out_ucast_pkts': '289230546',
                                                                       'poller_hash': 'Mzc4ODMycGVlcmludGY3NTYvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '756',
                                                                       'in_nucast_pkts': '80843200',
                                                                       'in_octets': '316883391400',
                                                                       'in_ucast_pkts': '289230546',
                                                                       'links': [       {       'in_nucast_pkts': '80843200',
                                                                                                'in_octets': '316883391400',
                                                                                                'in_ucast_pkts': '289230546',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '48028',
                                                                                                'out_octets': '64079122534',
                                                                                                'out_ucast_pkts': '205269869',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '48028',
                                                                       'out_octets': '64079122534',
                                                                       'out_ucast_pkts': '205269869',
                                                                       'poller_hash': 'Mzc4ODMycGVlcmludGZwb24gNzU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '756',
                                                                                               'links': [       {       'connected_time': '41056',
                                                                                                                        'fiber_length': 1357,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.49,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzc4ODMycGVlcnBvbnBvbiA3NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '6 Spencer St Milverton (Parks '
                                                 'Livestock 10420629)',
                                         'poller_hash': 'Mzc4ODMy',
                                         'serial_number': '378832',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:be:a7:a9',
                                         'uptime': '41056'},
                                 {       'counter_source': '757/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:9f:80:73:c0:9b'],
                                                                       'id': '757/g1',
                                                                       'in_nucast_pkts': '1251353',
                                                                       'in_octets': '50913880730',
                                                                       'in_ucast_pkts': '327319308',
                                                                       'mac_address': 'ec:4f:82:7e:0b:38',
                                                                       'mtu': 2000,
                                                                       'name': '757/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82761877',
                                                                       'out_octets': '2810196622849',
                                                                       'out_ucast_pkts': '1987899888',
                                                                       'poller_hash': 'MzVFQTA4cGVlcmludGY3NTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '757',
                                                                       'in_nucast_pkts': '82761877',
                                                                       'in_octets': '2810196622849',
                                                                       'in_ucast_pkts': '1987899888',
                                                                       'links': [       {       'in_nucast_pkts': '82761877',
                                                                                                'in_octets': '2810196622849',
                                                                                                'in_ucast_pkts': '1987899888',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1251353',
                                                                                                'out_octets': '50913880730',
                                                                                                'out_ucast_pkts': '327319308',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1251353',
                                                                       'out_octets': '50913880730',
                                                                       'out_ucast_pkts': '327319308',
                                                                       'poller_hash': 'MzVFQTA4cGVlcmludGZwb24gNzU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '757',
                                                                                               'links': [       {       'connected_time': '67790340',
                                                                                                                        'fiber_length': 13308,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQTA4cGVlcnBvbnBvbiA3NTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5673 Perth Road 119 (Donald & Nelda '
                                                 'Erb)',
                                         'poller_hash': 'MzVFQTA4',
                                         'serial_number': '35EA08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0b:38',
                                         'uptime': '67790340'},
                                 {       'counter_source': '759/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:dd:a9:92:18:70'],
                                                                       'id': '759/g1',
                                                                       'in_nucast_pkts': '2808',
                                                                       'in_octets': '48536782183',
                                                                       'in_ucast_pkts': '413500718',
                                                                       'mac_address': 'ec:4f:82:7e:0b:98',
                                                                       'mtu': 2000,
                                                                       'name': '759/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33031870',
                                                                       'out_octets': '1594603563706',
                                                                       'out_ucast_pkts': '1156628508',
                                                                       'poller_hash': 'MzVFQTI4cGVlcmludGY3NTkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '759',
                                                                       'in_nucast_pkts': '33031870',
                                                                       'in_octets': '1594603563706',
                                                                       'in_ucast_pkts': '1156628508',
                                                                       'links': [       {       'in_nucast_pkts': '33031870',
                                                                                                'in_octets': '1594603563706',
                                                                                                'in_ucast_pkts': '1156628508',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2808',
                                                                                                'out_octets': '48536782183',
                                                                                                'out_ucast_pkts': '413500718',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2808',
                                                                       'out_octets': '48536782183',
                                                                       'out_ucast_pkts': '413500718',
                                                                       'poller_hash': 'MzVFQTI4cGVlcmludGZwb24gNzU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '759',
                                                                                               'links': [       {       'connected_time': '15448823',
                                                                                                                        'fiber_length': 10435,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQTI4cGVlcnBvbnBvbiA3NTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5894 Perth Road 119 (Robt & Esther '
                                                 'Bitzer 10400237)',
                                         'poller_hash': 'MzVFQTI4',
                                         'serial_number': '35EA28',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0b:98',
                                         'uptime': '15448823'},
                                 {       'counter_source': '760/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f2:ef'],
                                                                       'id': '760/g1',
                                                                       'in_nucast_pkts': '5488',
                                                                       'in_octets': '12158851338',
                                                                       'in_ucast_pkts': '38818641',
                                                                       'mac_address': 'ec:4f:82:7e:0b:a4',
                                                                       'mtu': 2000,
                                                                       'name': '760/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '519533',
                                                                       'out_octets': '136838119728',
                                                                       'out_ucast_pkts': '108478633',
                                                                       'poller_hash': 'MzVFQTJDcGVlcmludGY3NjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '760',
                                                                       'in_nucast_pkts': '519533',
                                                                       'in_octets': '136838119728',
                                                                       'in_ucast_pkts': '108478633',
                                                                       'links': [       {       'in_nucast_pkts': '519533',
                                                                                                'in_octets': '136838119728',
                                                                                                'in_ucast_pkts': '108478633',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5488',
                                                                                                'out_octets': '12158851338',
                                                                                                'out_ucast_pkts': '38818641',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5488',
                                                                       'out_octets': '12158851338',
                                                                       'out_ucast_pkts': '38818641',
                                                                       'poller_hash': 'MzVFQTJDcGVlcmludGZwb24gNzYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '760',
                                                                                               'links': [       {       'connected_time': '2609498',
                                                                                                                        'fiber_length': 12716,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFQTJDcGVlcnBvbnBvbiA3NjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5720 PERTH ROAD 119 (BROADHURST '
                                                 'EVAN)',
                                         'poller_hash': 'MzVFQTJD',
                                         'serial_number': '35EA2C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0b:a4',
                                         'uptime': '2609498'},
                                 {       'counter_source': '761/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:50:e6:c3:5a:05'],
                                                                       'id': '761/g1',
                                                                       'in_nucast_pkts': '271',
                                                                       'in_octets': '497534777',
                                                                       'in_ucast_pkts': '3279274',
                                                                       'mac_address': 'cc:be:59:4f:fe:1c',
                                                                       'mtu': 2000,
                                                                       'name': '761/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120112',
                                                                       'out_octets': '7955824782',
                                                                       'out_ucast_pkts': '5861404',
                                                                       'poller_hash': 'NDNCMkQ3cGVlcmludGY3NjEvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '761',
                                                                       'in_nucast_pkts': '1120112',
                                                                       'in_octets': '7955824782',
                                                                       'in_ucast_pkts': '5861404',
                                                                       'links': [       {       'in_nucast_pkts': '1120112',
                                                                                                'in_octets': '7955824782',
                                                                                                'in_ucast_pkts': '5861404',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '271',
                                                                                                'out_octets': '497534777',
                                                                                                'out_ucast_pkts': '3279274',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '271',
                                                                       'out_octets': '497534777',
                                                                       'out_ucast_pkts': '3279274',
                                                                       'poller_hash': 'NDNCMkQ3cGVlcmludGZwb24gNzYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '761',
                                                                                               'links': [       {       'connected_time': '688525',
                                                                                                                        'fiber_length': 13539,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.37,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.392}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMkQ3cGVlcnBvbnBvbiA3NjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5668 Perth Rd 119 Brunner (Simon '
                                                 'Jantzi 10400499)',
                                         'poller_hash': 'NDNCMkQ3',
                                         'serial_number': '43B2D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:fe:1b',
                                         'uptime': '688525'},
                                 {       'counter_source': '763/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '763/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '181671752739',
                                                                       'in_ucast_pkts': '975137584',
                                                                       'mac_address': 'ec:4f:82:8e:02:28',
                                                                       'name': '763/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4062300016877',
                                                                       'out_ucast_pkts': '2939280562',
                                                                       'poller_hash': 'MzIyODhDcGVlcmludGY3NjMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '763',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4062300016877',
                                                                       'in_ucast_pkts': '2939280562',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4062300016877',
                                                                                                'in_ucast_pkts': '2939280562',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '181671752739',
                                                                                                'out_ucast_pkts': '975137584',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '181671752739',
                                                                       'out_ucast_pkts': '975137584',
                                                                       'poller_hash': 'MzIyODhDcGVlcmludGZwb24gNzYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '763',
                                                                                               'links': [       {       'connected_time': '688532',
                                                                                                                        'fiber_length': 10170,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODhDcGVlcnBvbnBvbiA3NjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5903 Perth Rd 119 Brunner (Marie '
                                                 'Lindner 10431377)',
                                         'poller_hash': 'MzIyODhD',
                                         'serial_number': '32288C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:02:28',
                                         'uptime': '688532'},
                                 {       'counter_source': '764/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:95:55'],
                                                                       'id': '764/g1',
                                                                       'in_nucast_pkts': '10495',
                                                                       'in_octets': '576305704697',
                                                                       'in_ucast_pkts': '2578030805',
                                                                       'mac_address': 'ec:4f:82:8e:04:3b',
                                                                       'mtu': 2000,
                                                                       'name': '764/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '145137857',
                                                                       'out_octets': '7893083064217',
                                                                       'out_ucast_pkts': '6073784622',
                                                                       'poller_hash': 'MzIyOENFcGVlcmludGY3NjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '764',
                                                                       'in_nucast_pkts': '145137857',
                                                                       'in_octets': '7893083064217',
                                                                       'in_ucast_pkts': '6073784622',
                                                                       'links': [       {       'in_nucast_pkts': '145137857',
                                                                                                'in_octets': '7893083064217',
                                                                                                'in_ucast_pkts': '6073784622',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10495',
                                                                                                'out_octets': '576305704697',
                                                                                                'out_ucast_pkts': '2578030805',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10495',
                                                                       'out_octets': '576305704697',
                                                                       'out_ucast_pkts': '2578030805',
                                                                       'poller_hash': 'MzIyOENFcGVlcmludGZwb24gNzY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '764',
                                                                                               'links': [       {       'connected_time': '21816414',
                                                                                                                        'fiber_length': 13404,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOENFcGVlcnBvbnBvbiA3NjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5678 Perth Rd 119 Brunner (Orval '
                                                 'Zehr 10400708)',
                                         'poller_hash': 'MzIyOENF',
                                         'serial_number': '3228CE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:04:38',
                                         'uptime': '21816414'},
                                 {       'counter_source': '765/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:fa:10'],
                                                                       'id': '765/g1',
                                                                       'in_nucast_pkts': '827973',
                                                                       'in_octets': '283465947780',
                                                                       'in_ucast_pkts': '987428020',
                                                                       'mac_address': 'ec:4f:82:7e:07:cf',
                                                                       'mtu': 2000,
                                                                       'name': '765/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '93544822',
                                                                       'out_octets': '4720769058327',
                                                                       'out_ucast_pkts': '3634014972',
                                                                       'poller_hash': 'MzVFOEU1cGVlcmludGY3NjUvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '765',
                                                                       'in_nucast_pkts': '93544822',
                                                                       'in_octets': '4720769058327',
                                                                       'in_ucast_pkts': '3634014972',
                                                                       'links': [       {       'in_nucast_pkts': '93544822',
                                                                                                'in_octets': '4720769058327',
                                                                                                'in_ucast_pkts': '3634014972',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '827973',
                                                                                                'out_octets': '283465947780',
                                                                                                'out_ucast_pkts': '987428020',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '827973',
                                                                       'out_octets': '283465947780',
                                                                       'out_ucast_pkts': '987428020',
                                                                       'poller_hash': 'MzVFOEU1cGVlcmludGZwb24gNzY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '765',
                                                                                               'links': [       {       'connected_time': '67790290',
                                                                                                                        'fiber_length': 658,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOEU1cGVlcnBvbnBvbiA3NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5 Main St N Milverton (V & M Lemieux '
                                                 '10400112)',
                                         'poller_hash': 'MzVFOEU1',
                                         'serial_number': '35E8E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:07:cf',
                                         'uptime': '67790290'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '766/g1',
                                                                       'name': '766/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1REMycGVlcmludGY3NjYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '766',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1REMycGVlcmludGZwb24gNzY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '766',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REMycGVlcnBvbnBvbiA3NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '7 MAIN ST N',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "766" '
                                                                  '(location: GPON '
                                                                  'port "8/1"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-02-04T20:48:49.300Z',
                                         'poller_hash': 'MjA1REMy',
                                         'serial_number': '205DC2',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '767/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:b3:70:12:51:39'],
                                                                       'id': '767/g1',
                                                                       'in_nucast_pkts': '42',
                                                                       'in_octets': '16115801888',
                                                                       'in_ucast_pkts': '60127040',
                                                                       'mac_address': 'cc:be:59:4f:ff:68',
                                                                       'mtu': 2000,
                                                                       'name': '767/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120144',
                                                                       'out_octets': '313170434269',
                                                                       'out_ucast_pkts': '244097670',
                                                                       'poller_hash': 'NDNCMzdEcGVlcmludGY3NjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '767',
                                                                       'in_nucast_pkts': '1120144',
                                                                       'in_octets': '313170434269',
                                                                       'in_ucast_pkts': '244097670',
                                                                       'links': [       {       'in_nucast_pkts': '1120144',
                                                                                                'in_octets': '313170434269',
                                                                                                'in_ucast_pkts': '244097670',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '42',
                                                                                                'out_octets': '16115801888',
                                                                                                'out_ucast_pkts': '60127040',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '42',
                                                                       'out_octets': '16115801888',
                                                                       'out_ucast_pkts': '60127040',
                                                                       'poller_hash': 'NDNCMzdEcGVlcmludGZwb24gNzY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '767',
                                                                                               'links': [       {       'connected_time': '688334',
                                                                                                                        'fiber_length': 847,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.908,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.026}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMzdEcGVlcnBvbnBvbiA3Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '19 TEMPERANCE ST (BOWEY JOHN)',
                                         'poller_hash': 'NDNCMzdE',
                                         'serial_number': '43B37D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:ff:67',
                                         'uptime': '688334'},
                                 {       'counter_source': '768/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:07:b6:74:04:e1'],
                                                                       'id': '768/g1',
                                                                       'in_nucast_pkts': '1550',
                                                                       'in_octets': '22643983230',
                                                                       'in_ucast_pkts': '55759384',
                                                                       'mac_address': '44:65:7f:7b:c3:c5',
                                                                       'mtu': 2000,
                                                                       'name': '768/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9748392',
                                                                       'out_octets': '156561465015',
                                                                       'out_ucast_pkts': '136481982',
                                                                       'poller_hash': 'NUE2NjMycGVlcmludGY3NjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '768',
                                                                       'in_nucast_pkts': '9748392',
                                                                       'in_octets': '156561465015',
                                                                       'in_ucast_pkts': '136481982',
                                                                       'links': [       {       'in_nucast_pkts': '9748392',
                                                                                                'in_octets': '156561465015',
                                                                                                'in_ucast_pkts': '136481982',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1550',
                                                                                                'out_octets': '22643983230',
                                                                                                'out_ucast_pkts': '55759384',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1550',
                                                                       'out_octets': '22643983230',
                                                                       'out_ucast_pkts': '55759384',
                                                                       'poller_hash': 'NUE2NjMycGVlcmludGZwb24gNzY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '768',
                                                                                               'links': [       {       'connected_time': '4484892',
                                                                                                                        'fiber_length': 10092,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjMycGVlcnBvbnBvbiA3Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5909 PERTH ROAD 119 (FRIESEN KEVIN)',
                                         'poller_hash': 'NUE2NjMy',
                                         'serial_number': '5A6632',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:c5',
                                         'uptime': '4484892'},
                                 {       'counter_source': '771/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '28:3b:82:66:18:9e'],
                                                                       'id': '771/g1',
                                                                       'in_nucast_pkts': '743',
                                                                       'in_octets': '15978955369',
                                                                       'in_ucast_pkts': '76340151',
                                                                       'mac_address': 'cc:be:59:50:00:58',
                                                                       'mtu': 2000,
                                                                       'name': '771/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1121595',
                                                                       'out_octets': '453956222853',
                                                                       'out_ucast_pkts': '346817938',
                                                                       'poller_hash': 'NDNCM0Y0cGVlcmludGY3NzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '771',
                                                                       'in_nucast_pkts': '1121595',
                                                                       'in_octets': '453956222853',
                                                                       'in_ucast_pkts': '346817938',
                                                                       'links': [       {       'in_nucast_pkts': '1121595',
                                                                                                'in_octets': '453956222853',
                                                                                                'in_ucast_pkts': '346817938',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '743',
                                                                                                'out_octets': '15978955369',
                                                                                                'out_ucast_pkts': '76340151',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '743',
                                                                       'out_octets': '15978955369',
                                                                       'out_ucast_pkts': '76340151',
                                                                       'poller_hash': 'NDNCM0Y0cGVlcmludGZwb24gNzcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '771',
                                                                                               'links': [       {       'connected_time': '688344',
                                                                                                                        'fiber_length': 1220,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.234,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.32}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0Y0cGVlcnBvbnBvbiA3NzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '15 West St Milverton (Maggie Enns '
                                                 '10430560)',
                                         'poller_hash': 'NDNCM0Y0',
                                         'serial_number': '43B3F4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:00:57',
                                         'uptime': '688344'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '772/g1',
                                                                       'name': '772/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2RUYxcGVlcmludGY3NzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '772',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2RUYxcGVlcmludGZwb24gNzcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '772',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RUYxcGVlcnBvbnBvbiA3NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '19 MAIN ST N Apt 1',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "772" '
                                                                  '(location: GPON '
                                                                  'port "7/8"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:41.920Z',
                                         'poller_hash': 'OTU2RUYx',
                                         'serial_number': '956EF1',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '773/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '6c:cd:d6:eb:84:7d'],
                                                                       'id': '773/g1',
                                                                       'in_nucast_pkts': '164',
                                                                       'in_octets': '1391886037',
                                                                       'in_ucast_pkts': '9252028',
                                                                       'mac_address': 'cc:be:59:4f:ff:8a',
                                                                       'mtu': 2000,
                                                                       'name': '773/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '851191',
                                                                       'out_octets': '30526009516',
                                                                       'out_ucast_pkts': '22514538',
                                                                       'poller_hash': 'NDNCMzhFcGVlcmludGY3NzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '773',
                                                                       'in_nucast_pkts': '851191',
                                                                       'in_octets': '30526009516',
                                                                       'in_ucast_pkts': '22514538',
                                                                       'links': [       {       'in_nucast_pkts': '851191',
                                                                                                'in_octets': '30526009516',
                                                                                                'in_ucast_pkts': '22514538',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '164',
                                                                                                'out_octets': '1391886037',
                                                                                                'out_ucast_pkts': '9252028',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '164',
                                                                       'out_octets': '1391886037',
                                                                       'out_ucast_pkts': '9252028',
                                                                       'poller_hash': 'NDNCMzhFcGVlcmludGZwb24gNzcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '773',
                                                                                               'links': [       {       'connected_time': '538541',
                                                                                                                        'fiber_length': 11191,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.652,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.258}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMzhFcGVlcnBvbnBvbiA3NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '4978 Line 52 Brunner (Andrew '
                                                 'Brodhagen 10433259)',
                                         'poller_hash': 'NDNCMzhF',
                                         'serial_number': '43B38E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:ff:89',
                                         'uptime': '538541'},
                                 {       'counter_source': '774/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '2c:fd:a1:c2:80:48'],
                                                                       'id': '774/g1',
                                                                       'in_nucast_pkts': '1333',
                                                                       'in_octets': '5956758351',
                                                                       'in_ucast_pkts': '55796253',
                                                                       'mac_address': '44:65:7f:7b:c4:67',
                                                                       'mtu': 2000,
                                                                       'name': '774/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1476741',
                                                                       'out_octets': '251705967841',
                                                                       'out_ucast_pkts': '176675975',
                                                                       'poller_hash': 'NUE2NjY4cGVlcmludGY3NzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '774',
                                                                       'in_nucast_pkts': '1476741',
                                                                       'in_octets': '251705967841',
                                                                       'in_ucast_pkts': '176675975',
                                                                       'links': [       {       'in_nucast_pkts': '1476741',
                                                                                                'in_octets': '251705967841',
                                                                                                'in_ucast_pkts': '176675975',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1333',
                                                                                                'out_octets': '5956758351',
                                                                                                'out_ucast_pkts': '55796253',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1333',
                                                                       'out_octets': '5956758351',
                                                                       'out_ucast_pkts': '55796253',
                                                                       'poller_hash': 'NUE2NjY4cGVlcmludGZwb24gNzc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '774',
                                                                                               'links': [       {       'connected_time': '709215',
                                                                                                                        'fiber_length': 12027,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjY4cGVlcnBvbnBvbiA3NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4911 Line 52 Brunner (M & L '
                                                 'Brodhagen 10400659)',
                                         'poller_hash': 'NUE2NjY4',
                                         'serial_number': '5A6668',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:67',
                                         'uptime': '709215'},
                                 {       'counter_source': '775/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:45'],
                                                                       'id': '775/g1',
                                                                       'in_nucast_pkts': '858',
                                                                       'in_octets': '84155512090',
                                                                       'in_ucast_pkts': '335919033',
                                                                       'mac_address': 'cc:be:59:cb:e7:95',
                                                                       'mtu': 2000,
                                                                       'name': '775/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '99468244',
                                                                       'out_octets': '1515296121291',
                                                                       'out_ucast_pkts': '1158290962',
                                                                       'poller_hash': 'NDJDRTRBcGVlcmludGY3NzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '775',
                                                                       'in_nucast_pkts': '99468244',
                                                                       'in_octets': '1515296121291',
                                                                       'in_ucast_pkts': '1158290962',
                                                                       'links': [       {       'in_nucast_pkts': '99468244',
                                                                                                'in_octets': '1515296121291',
                                                                                                'in_ucast_pkts': '1158290962',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '858',
                                                                                                'out_octets': '84155512090',
                                                                                                'out_ucast_pkts': '335919033',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '858',
                                                                       'out_octets': '84155512090',
                                                                       'out_ucast_pkts': '335919033',
                                                                       'poller_hash': 'NDJDRTRBcGVlcmludGZwb24gNzc1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '775',
                                                                                               'links': [       {       'connected_time': '4834258',
                                                                                                                        'fiber_length': 12543,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTRBcGVlcnBvbnBvbiA3NzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4860 Line 52 Brunner (Ronald '
                                                 'Raycraft 10400550)',
                                         'poller_hash': 'NDJDRTRB',
                                         'serial_number': '42CE4A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e7:92',
                                         'uptime': '4834258'},
                                 {       'counter_source': '776/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:2f:d0:d5:3c:61'],
                                                                       'id': '776/g1',
                                                                       'in_nucast_pkts': '22594',
                                                                       'in_octets': '186524256471',
                                                                       'in_ucast_pkts': '602731047',
                                                                       'mac_address': '44:65:7f:7b:d3:fd',
                                                                       'mtu': 2000,
                                                                       'name': '776/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76676668',
                                                                       'out_octets': '2043583326467',
                                                                       'out_ucast_pkts': '1601911413',
                                                                       'poller_hash': 'NUE2QjlBcGVlcmludGY3NzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '776',
                                                                       'in_nucast_pkts': '76676668',
                                                                       'in_octets': '2043583326467',
                                                                       'in_ucast_pkts': '1601911413',
                                                                       'links': [       {       'in_nucast_pkts': '76676668',
                                                                                                'in_octets': '2043583326467',
                                                                                                'in_ucast_pkts': '1601911413',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22594',
                                                                                                'out_octets': '186524256471',
                                                                                                'out_ucast_pkts': '602731047',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22594',
                                                                       'out_octets': '186524256471',
                                                                       'out_ucast_pkts': '602731047',
                                                                       'poller_hash': 'NUE2QjlBcGVlcmludGZwb24gNzc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '776',
                                                                                               'links': [       {       'connected_time': '67790120',
                                                                                                                        'fiber_length': 11876,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjlBcGVlcnBvbnBvbiA3NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5088 Line 52 Brunner (Jim Van Herk '
                                                 '10400809)',
                                         'poller_hash': 'NUE2QjlB',
                                         'serial_number': '5A6B9A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:fd',
                                         'uptime': '67790120'},
                                 {       'counter_source': '777/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '777/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '583',
                                                                       'in_octets': '66953477700',
                                                                       'in_ucast_pkts': '387953859',
                                                                       'mac_address': 'ec:4f:82:0d:68:68',
                                                                       'name': '777/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '2555',
                                                                       'out_octets': '1496602613174',
                                                                       'out_ucast_pkts': '1119189959',
                                                                       'poller_hash': 'MjA1REJCcGVlcmludGY3NzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '777',
                                                                       'in_nucast_pkts': '2555',
                                                                       'in_octets': '1496602613174',
                                                                       'in_ucast_pkts': '1119189959',
                                                                       'links': [       {       'in_nucast_pkts': '2555',
                                                                                                'in_octets': '1496602613174',
                                                                                                'in_ucast_pkts': '1119189959',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '583',
                                                                                                'out_octets': '66953477700',
                                                                                                'out_ucast_pkts': '387953859',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '583',
                                                                       'out_octets': '66953477700',
                                                                       'out_ucast_pkts': '387953859',
                                                                       'poller_hash': 'MjA1REJCcGVlcmludGZwb24gNzc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '777',
                                                                                               'links': [       {       'connected_time': '18640945',
                                                                                                                        'fiber_length': 1696,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REJCcGVlcnBvbnBvbiA3Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 William St Apt 203 Milverton '
                                                 '(Diane Playford 10435440)',
                                         'poller_hash': 'MjA1REJC',
                                         'serial_number': '205DBB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:68',
                                         'uptime': '18640945'},
                                 {       'counter_source': '778/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:84:6a:57:6c:f5'],
                                                                       'id': '778/g1',
                                                                       'in_nucast_pkts': '20305381',
                                                                       'in_octets': '95427504245',
                                                                       'in_ucast_pkts': '1064081215',
                                                                       'mac_address': '44:65:7f:7b:d4:96',
                                                                       'mtu': 2000,
                                                                       'name': '778/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76687687',
                                                                       'out_octets': '3130961364770',
                                                                       'out_ucast_pkts': '2123112404',
                                                                       'poller_hash': 'NUE2QkNEcGVlcmludGY3NzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '778',
                                                                       'in_nucast_pkts': '76687687',
                                                                       'in_octets': '3130961364770',
                                                                       'in_ucast_pkts': '2123112404',
                                                                       'links': [       {       'in_nucast_pkts': '76687687',
                                                                                                'in_octets': '3130961364770',
                                                                                                'in_ucast_pkts': '2123112404',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20305381',
                                                                                                'out_octets': '95427504245',
                                                                                                'out_ucast_pkts': '1064081215',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20305381',
                                                                       'out_octets': '95427504245',
                                                                       'out_ucast_pkts': '1064081215',
                                                                       'poller_hash': 'NUE2QkNEcGVlcmludGZwb24gNzc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '778',
                                                                                               'links': [       {       'connected_time': '67790318',
                                                                                                                        'fiber_length': 11201,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QkNEcGVlcnBvbnBvbiA3Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5020 Line 52 Brunner (B & J Natywary '
                                                 '10400723)',
                                         'poller_hash': 'NUE2QkNE',
                                         'serial_number': '5A6BCD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d4:96',
                                         'uptime': '67790318'},
                                 {       'counter_source': '779/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:66:df'],
                                                                       'id': '779/g1',
                                                                       'in_nucast_pkts': '11966',
                                                                       'in_octets': '2076431296',
                                                                       'in_ucast_pkts': '3477341',
                                                                       'mac_address': 'cc:be:59:50:03:e0',
                                                                       'mtu': 2000,
                                                                       'name': '779/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120090',
                                                                       'out_octets': '8831810921',
                                                                       'out_ucast_pkts': '7899344',
                                                                       'poller_hash': 'NDNCNUI3cGVlcmludGY3NzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '779',
                                                                       'in_nucast_pkts': '1120090',
                                                                       'in_octets': '8831810921',
                                                                       'in_ucast_pkts': '7899344',
                                                                       'links': [       {       'in_nucast_pkts': '1120090',
                                                                                                'in_octets': '8831810921',
                                                                                                'in_ucast_pkts': '7899344',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11966',
                                                                                                'out_octets': '2076431296',
                                                                                                'out_ucast_pkts': '3477341',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11966',
                                                                       'out_octets': '2076431296',
                                                                       'out_ucast_pkts': '3477341',
                                                                       'poller_hash': 'NDNCNUI3cGVlcmludGZwb24gNzc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '779',
                                                                                               'links': [       {       'connected_time': '688527',
                                                                                                                        'fiber_length': 10221,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.52,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.958}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUI3cGVlcnBvbnBvbiA3Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '6121 Ana St Brunner (Stephan '
                                                 'Gingerich 10435309)',
                                         'poller_hash': 'NDNCNUI3',
                                         'serial_number': '43B5B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:03:df',
                                         'uptime': '688527'},
                                 {       'counter_source': '780/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:52:a1:96:ba:ed'],
                                                                       'id': '780/g1',
                                                                       'in_nucast_pkts': '512',
                                                                       'in_octets': '24333061128',
                                                                       'in_ucast_pkts': '72894560',
                                                                       'mac_address': '44:65:7f:7b:d3:dc',
                                                                       'mtu': 2000,
                                                                       'name': '780/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119864',
                                                                       'out_octets': '234621537823',
                                                                       'out_ucast_pkts': '184797375',
                                                                       'poller_hash': 'NUE2QjhGcGVlcmludGY3ODAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '780',
                                                                       'in_nucast_pkts': '1119864',
                                                                       'in_octets': '234621537823',
                                                                       'in_ucast_pkts': '184797375',
                                                                       'links': [       {       'in_nucast_pkts': '1119864',
                                                                                                'in_octets': '234621537823',
                                                                                                'in_ucast_pkts': '184797375',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '512',
                                                                                                'out_octets': '24333061128',
                                                                                                'out_ucast_pkts': '72894560',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '512',
                                                                       'out_octets': '24333061128',
                                                                       'out_ucast_pkts': '72894560',
                                                                       'poller_hash': 'NUE2QjhGcGVlcmludGZwb24gNzgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '780',
                                                                                               'links': [       {       'connected_time': '538558',
                                                                                                                        'fiber_length': 13591,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjhGcGVlcnBvbnBvbiA3ODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5133 Rd 140 (Paul and Carolyn '
                                                 'Visser)',
                                         'poller_hash': 'NUE2QjhG',
                                         'serial_number': '5A6B8F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:dc',
                                         'uptime': '538558'},
                                 {       'counter_source': '781/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f0:c9'],
                                                                       'id': '781/g1',
                                                                       'in_nucast_pkts': '11670',
                                                                       'in_octets': '6258945068',
                                                                       'in_ucast_pkts': '17644119',
                                                                       'mac_address': 'cc:be:59:cc:10:4d',
                                                                       'mtu': 2000,
                                                                       'name': '781/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463912',
                                                                       'out_octets': '68791690329',
                                                                       'out_ucast_pkts': '52105667',
                                                                       'poller_hash': 'NDJEMzYxcGVlcmludGY3ODEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '781',
                                                                       'in_nucast_pkts': '1463912',
                                                                       'in_octets': '68791690329',
                                                                       'in_ucast_pkts': '52105667',
                                                                       'links': [       {       'in_nucast_pkts': '1463912',
                                                                                                'in_octets': '68791690329',
                                                                                                'in_ucast_pkts': '52105667',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11670',
                                                                                                'out_octets': '6258945068',
                                                                                                'out_ucast_pkts': '17644119',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11670',
                                                                       'out_octets': '6258945068',
                                                                       'out_ucast_pkts': '17644119',
                                                                       'poller_hash': 'NDJEMzYxcGVlcmludGZwb24gNzgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '781',
                                                                                               'links': [       {       'connected_time': '688436',
                                                                                                                        'fiber_length': 1477,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMzYxcGVlcnBvbnBvbiA3ODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '85B Mill St E Milverton (Tricia '
                                                 'Nahrgang 10435488)',
                                         'poller_hash': 'NDJEMzYx',
                                         'serial_number': '42D361',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:10:4a',
                                         'uptime': '688436'},
                                 {       'counter_source': '782/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0b:6d'],
                                                                       'id': '782/g1',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '80418833084',
                                                                       'in_ucast_pkts': '187845635',
                                                                       'mac_address': 'ec:4f:82:7e:08:c2',
                                                                       'mtu': 2000,
                                                                       'name': '782/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6782058',
                                                                       'out_octets': '588321884380',
                                                                       'out_ucast_pkts': '464457459',
                                                                       'poller_hash': 'MzVFOTM2cGVlcmludGY3ODIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '782',
                                                                       'in_nucast_pkts': '6782058',
                                                                       'in_octets': '588321884380',
                                                                       'in_ucast_pkts': '464457459',
                                                                       'links': [       {       'in_nucast_pkts': '6782058',
                                                                                                'in_octets': '588321884380',
                                                                                                'in_ucast_pkts': '464457459',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '31',
                                                                                                'out_octets': '80418833084',
                                                                                                'out_ucast_pkts': '187845635',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '80418833084',
                                                                       'out_ucast_pkts': '187845635',
                                                                       'poller_hash': 'MzVFOTM2cGVlcmludGZwb24gNzgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '782',
                                                                                               'links': [       {       'connected_time': '3113915',
                                                                                                                        'fiber_length': 18800,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTM2cGVlcnBvbnBvbiA3ODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '85 Schade St Monkton (Jesse Hamilton '
                                                 '10434996)',
                                         'poller_hash': 'MzVFOTM2',
                                         'serial_number': '35E936',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:08:c2',
                                         'uptime': '3113915'},
                                 {       'counter_source': '783/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '8c:90:2d:72:7d:85'],
                                                                       'id': '783/g1',
                                                                       'in_nucast_pkts': '74',
                                                                       'in_octets': '881413934',
                                                                       'in_ucast_pkts': '2546810',
                                                                       'mac_address': '44:65:7f:33:1a:3f',
                                                                       'mtu': 2000,
                                                                       'name': '783/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '109008',
                                                                       'out_octets': '5639145061',
                                                                       'out_ucast_pkts': '4379500',
                                                                       'poller_hash': 'NjQwNzlBcGVlcmludGY3ODMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '783',
                                                                       'in_nucast_pkts': '109008',
                                                                       'in_octets': '5639145061',
                                                                       'in_ucast_pkts': '4379500',
                                                                       'links': [       {       'in_nucast_pkts': '109008',
                                                                                                'in_octets': '5639145061',
                                                                                                'in_ucast_pkts': '4379500',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '74',
                                                                                                'out_octets': '881413934',
                                                                                                'out_ucast_pkts': '2546810',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '74',
                                                                       'out_octets': '881413934',
                                                                       'out_ucast_pkts': '2546810',
                                                                       'poller_hash': 'NjQwNzlBcGVlcmludGZwb24gNzgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '783',
                                                                                               'links': [       {       'connected_time': '69795',
                                                                                                                        'fiber_length': 13217,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.718,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.352}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwNzlBcGVlcnBvbnBvbiA3ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5154 Road 140 (Joseph White '
                                                 '10435270)',
                                         'poller_hash': 'NjQwNzlB',
                                         'serial_number': '64079A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:1a:3e',
                                         'uptime': '69795'},
                                 {       'counter_source': '784/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ee:1d'],
                                                                       'id': '784/g1',
                                                                       'in_nucast_pkts': '9663',
                                                                       'in_octets': '19938757404',
                                                                       'in_ucast_pkts': '36601494',
                                                                       'mac_address': '44:65:7f:7b:c3:05',
                                                                       'mtu': 2000,
                                                                       'name': '784/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '920523',
                                                                       'out_octets': '73470385616',
                                                                       'out_ucast_pkts': '63457636',
                                                                       'poller_hash': 'NUE2NUYycGVlcmludGY3ODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '784',
                                                                       'in_nucast_pkts': '920523',
                                                                       'in_octets': '73470385616',
                                                                       'in_ucast_pkts': '63457636',
                                                                       'links': [       {       'in_nucast_pkts': '920523',
                                                                                                'in_octets': '73470385616',
                                                                                                'in_ucast_pkts': '63457636',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9663',
                                                                                                'out_octets': '19938757404',
                                                                                                'out_ucast_pkts': '36601494',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9663',
                                                                       'out_octets': '19938757404',
                                                                       'out_ucast_pkts': '36601494',
                                                                       'poller_hash': 'NUE2NUYycGVlcmludGZwb24gNzg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '784',
                                                                                               'links': [       {       'connected_time': '538571',
                                                                                                                        'fiber_length': 17579,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NUYycGVlcnBvbnBvbiA3ODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4910 Line 48 Gadshill (Dan '
                                                 'Groenstege 10421236)',
                                         'poller_hash': 'NUE2NUYy',
                                         'serial_number': '5A65F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:05',
                                         'uptime': '538571'},
                                 {       'counter_source': '786/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:54:2e:a4:13:a9'],
                                                                       'id': '786/g1',
                                                                       'in_nucast_pkts': '1466',
                                                                       'in_octets': '302207125',
                                                                       'in_ucast_pkts': '2080735',
                                                                       'mac_address': '44:65:7f:7b:c3:44',
                                                                       'mtu': 2000,
                                                                       'name': '786/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1479154',
                                                                       'out_octets': '14225030162',
                                                                       'out_ucast_pkts': '10435176',
                                                                       'poller_hash': 'NUE2NjA3cGVlcmludGY3ODYvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '786',
                                                                       'in_nucast_pkts': '1479154',
                                                                       'in_octets': '14225030162',
                                                                       'in_ucast_pkts': '10435176',
                                                                       'links': [       {       'in_nucast_pkts': '1479154',
                                                                                                'in_octets': '14225030162',
                                                                                                'in_ucast_pkts': '10435176',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1466',
                                                                                                'out_octets': '302207125',
                                                                                                'out_ucast_pkts': '2080735',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1466',
                                                                       'out_octets': '302207125',
                                                                       'out_ucast_pkts': '2080735',
                                                                       'poller_hash': 'NUE2NjA3cGVlcmludGZwb24gNzg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '786',
                                                                                               'links': [       {       'connected_time': '709228',
                                                                                                                        'fiber_length': 16887,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjA3cGVlcnBvbnBvbiA3ODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5368 Perth Rd 135B Brunner (Glen '
                                                 'Diehl 10400698)',
                                         'poller_hash': 'NUE2NjA3',
                                         'serial_number': '5A6607',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:44',
                                         'uptime': '709228'},
                                 {       'counter_source': '787/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c4:e9:84:87:af:6d'],
                                                                       'id': '787/g1',
                                                                       'in_nucast_pkts': '1795',
                                                                       'in_octets': '3836296231',
                                                                       'in_ucast_pkts': '6621711',
                                                                       'mac_address': 'cc:be:59:b9:c5:00',
                                                                       'mtu': 2000,
                                                                       'name': '787/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '942690',
                                                                       'out_octets': '17097057510',
                                                                       'out_ucast_pkts': '14022253',
                                                                       'poller_hash': 'NTEyNTg4cGVlcmludGY3ODcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '787',
                                                                       'in_nucast_pkts': '942690',
                                                                       'in_octets': '17097057510',
                                                                       'in_ucast_pkts': '14022253',
                                                                       'links': [       {       'in_nucast_pkts': '942690',
                                                                                                'in_octets': '17097057510',
                                                                                                'in_ucast_pkts': '14022253',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1795',
                                                                                                'out_octets': '3836296231',
                                                                                                'out_ucast_pkts': '6621711',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1795',
                                                                       'out_octets': '3836296231',
                                                                       'out_ucast_pkts': '6621711',
                                                                       'poller_hash': 'NTEyNTg4cGVlcmludGZwb24gNzg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '787',
                                                                                               'links': [       {       'connected_time': '454487',
                                                                                                                        'fiber_length': 19054,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTg4cGVlcnBvbnBvbiA3ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4660 Line 52 Brunner (Rick Reis '
                                                 '10400560)',
                                         'poller_hash': 'NTEyNTg4',
                                         'serial_number': '512588',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c5:00',
                                         'uptime': '454487'},
                                 {       'counter_source': '788/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:72:7d:cd'],
                                                                       'id': '788/g1',
                                                                       'in_nucast_pkts': '15',
                                                                       'in_octets': '825980057',
                                                                       'in_ucast_pkts': '6230982',
                                                                       'mac_address': '04:bc:9f:5b:d3:5c',
                                                                       'mtu': 2000,
                                                                       'name': '788/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447195',
                                                                       'out_octets': '20728232802',
                                                                       'out_ucast_pkts': '15733400',
                                                                       'poller_hash': 'MTI3NDBFQnBlZXJpbnRmNzg4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '788',
                                                                       'in_nucast_pkts': '1447195',
                                                                       'in_octets': '20728232802',
                                                                       'in_ucast_pkts': '15733400',
                                                                       'links': [       {       'in_nucast_pkts': '1447195',
                                                                                                'in_octets': '20728232802',
                                                                                                'in_ucast_pkts': '15733400',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15',
                                                                                                'out_octets': '825980057',
                                                                                                'out_ucast_pkts': '6230982',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15',
                                                                       'out_octets': '825980057',
                                                                       'out_ucast_pkts': '6230982',
                                                                       'poller_hash': 'MTI3NDBFQnBlZXJpbnRmcG9uIDc4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '788',
                                                                                               'links': [       {       'connected_time': '688411',
                                                                                                                        'fiber_length': 2071,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBFQnBlZXJwb25wb24gNzg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '81 Reagan St Milverton (K Saysangkhy '
                                                 '104355220)',
                                         'poller_hash': 'MTI3NDBFQg==',
                                         'serial_number': '12740EB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d3:59',
                                         'uptime': '688411'},
                                 {       'counter_source': '790/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '790/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '189147273627',
                                                                       'in_ucast_pkts': '26357816456',
                                                                       'mac_address': 'cc:be:59:cb:e4:22',
                                                                       'name': '790/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '90',
                                                                       'out_octets': '2502180061617',
                                                                       'out_ucast_pkts': '1949928457',
                                                                       'poller_hash': 'NDJDRERDcGVlcmludGY3OTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '790',
                                                                       'in_nucast_pkts': '90',
                                                                       'in_octets': '2502180061617',
                                                                       'in_ucast_pkts': '1949928457',
                                                                       'links': [       {       'in_nucast_pkts': '90',
                                                                                                'in_octets': '2502180061617',
                                                                                                'in_ucast_pkts': '1949928457',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '189147273627',
                                                                                                'out_ucast_pkts': '26357816456',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '189147273627',
                                                                       'out_ucast_pkts': '26357816456',
                                                                       'poller_hash': 'NDJDRERDcGVlcmludGZwb24gNzkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '790',
                                                                                               'links': [       {       'connected_time': '21816141',
                                                                                                                        'fiber_length': 13955,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRERDcGVlcnBvbnBvbiA3OTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5099 Line 49 Gadshill (Andy '
                                                 'Groenestege 10400881)',
                                         'poller_hash': 'NDJDRERD',
                                         'serial_number': '42CDDC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e4:22',
                                         'uptime': '21816141'},
                                 {       'counter_source': '791/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '791/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '383778839488',
                                                                       'in_ucast_pkts': '87432498400',
                                                                       'mac_address': 'cc:be:59:cb:f9:d2',
                                                                       'name': '791/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '113',
                                                                       'out_octets': '10374964885076',
                                                                       'out_ucast_pkts': '7246463853',
                                                                       'poller_hash': 'NDJEMDkycGVlcmludGY3OTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '791',
                                                                       'in_nucast_pkts': '113',
                                                                       'in_octets': '10374964885076',
                                                                       'in_ucast_pkts': '7246463853',
                                                                       'links': [       {       'in_nucast_pkts': '113',
                                                                                                'in_octets': '10374964885076',
                                                                                                'in_ucast_pkts': '7246463853',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '383778839488',
                                                                                                'out_ucast_pkts': '87432498400',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '383778839488',
                                                                       'out_ucast_pkts': '87432498400',
                                                                       'poller_hash': 'NDJEMDkycGVlcmludGZwb24gNzkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '791',
                                                                                               'links': [       {       'connected_time': '21816344',
                                                                                                                        'fiber_length': 17293,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMDkycGVlcnBvbnBvbiA3OTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4606 Line 49 Gadshill (J & A Debus '
                                                 '10400821)',
                                         'poller_hash': 'NDJEMDky',
                                         'serial_number': '42D092',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:f9:d2',
                                         'uptime': '21816344'},
                                 {       'counter_source': '792/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:d9:f5:8c:e9:e0'],
                                                                       'id': '792/g1',
                                                                       'in_nucast_pkts': '1719646',
                                                                       'in_octets': '191545204798',
                                                                       'in_ucast_pkts': '347089528',
                                                                       'mac_address': '44:65:7f:7b:cf:68',
                                                                       'mtu': 2000,
                                                                       'name': '792/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3738115',
                                                                       'out_octets': '1150917768164',
                                                                       'out_ucast_pkts': '909928206',
                                                                       'poller_hash': 'NUE2QTEzcGVlcmludGY3OTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '792',
                                                                       'in_nucast_pkts': '3738115',
                                                                       'in_octets': '1150917768164',
                                                                       'in_ucast_pkts': '909928206',
                                                                       'links': [       {       'in_nucast_pkts': '3738115',
                                                                                                'in_octets': '1150917768164',
                                                                                                'in_ucast_pkts': '909928206',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1719646',
                                                                                                'out_octets': '191545204798',
                                                                                                'out_ucast_pkts': '347089528',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1719646',
                                                                       'out_octets': '191545204798',
                                                                       'out_ucast_pkts': '347089528',
                                                                       'poller_hash': 'NUE2QTEzcGVlcmludGZwb24gNzky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '792',
                                                                                               'links': [       {       'connected_time': '1745046',
                                                                                                                        'fiber_length': 18150,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QTEzcGVlcnBvbnBvbiA3OTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4723 LINE 52 (ZONDERVAN TONY & '
                                                 'CYNTHIA)',
                                         'poller_hash': 'NUE2QTEz',
                                         'serial_number': '5A6A13',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:cf:68',
                                         'uptime': '1745046'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '796/g1',
                                                                       'name': '796/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQUY4cGVlcmludGY3OTYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '796',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NjQwQUY4cGVlcmludGZwb24gNzk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '796',
                                                                                               'links': [       {       'connected_time': '688470',
                                                                                                                        'fiber_length': 1278,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -24.35,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.35}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUY4cGVlcnBvbnBvbiA3OTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '11 Main St N Ste 2 Milverton',
                                         'poller_hash': 'NjQwQUY4',
                                         'serial_number': '640AF8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:00',
                                         'uptime': '688470'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '798/g1',
                                                                       'name': '798/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2QjkzcGVlcmludGY3OTgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '798',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2QjkzcGVlcmludGZwb24gNzk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '798',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjkzcGVlcnBvbnBvbiA3OTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '6314 ROAD 124 (POOLE VICTOR)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "798" '
                                                                  '(location: GPON '
                                                                  'port "11/8"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-07-23T17:28:44.990Z',
                                         'poller_hash': 'NUE2Qjkz',
                                         'serial_number': '5A6B93',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '799/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:ea:83'],
                                                                       'id': '799/g1',
                                                                       'in_nucast_pkts': '13592',
                                                                       'in_octets': '9488352267',
                                                                       'in_ucast_pkts': '63177291',
                                                                       'mac_address': 'cc:be:59:cb:6e:1d',
                                                                       'mtu': 2000,
                                                                       'name': '799/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447090',
                                                                       'out_octets': '224249597521',
                                                                       'out_ucast_pkts': '166134257',
                                                                       'poller_hash': 'NDJCRjFCcGVlcmludGY3OTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '799',
                                                                       'in_nucast_pkts': '1447090',
                                                                       'in_octets': '224249597521',
                                                                       'in_ucast_pkts': '166134257',
                                                                       'links': [       {       'in_nucast_pkts': '1447090',
                                                                                                'in_octets': '224249597521',
                                                                                                'in_ucast_pkts': '166134257',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13592',
                                                                                                'out_octets': '9488352267',
                                                                                                'out_ucast_pkts': '63177291',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13592',
                                                                       'out_octets': '9488352267',
                                                                       'out_ucast_pkts': '63177291',
                                                                       'poller_hash': 'NDJCRjFCcGVlcmludGZwb24gNzk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '799',
                                                                                               'links': [       {       'connected_time': '688411',
                                                                                                                        'fiber_length': 2144,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjFCcGVlcnBvbnBvbiA3OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '77A REAGAN ST (LARA MARCELLO)',
                                         'poller_hash': 'NDJCRjFC',
                                         'serial_number': '42BF1B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6e:1a',
                                         'uptime': '688411'},
                                 {       'counter_source': '801/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e0:23:ff:2c:38:ee'],
                                                                       'id': '801/g1',
                                                                       'in_nucast_pkts': '9034',
                                                                       'in_octets': '199839168863',
                                                                       'in_ucast_pkts': '403694400',
                                                                       'mac_address': 'ec:4f:82:be:ab:fd',
                                                                       'mtu': 2000,
                                                                       'name': '801/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11805433',
                                                                       'out_octets': '330441167530',
                                                                       'out_ucast_pkts': '455761740',
                                                                       'poller_hash': 'Mzc4OEJDcGVlcmludGY4MDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '801',
                                                                       'in_nucast_pkts': '11805433',
                                                                       'in_octets': '330441167530',
                                                                       'in_ucast_pkts': '455761740',
                                                                       'links': [       {       'in_nucast_pkts': '11805433',
                                                                                                'in_octets': '330441167530',
                                                                                                'in_ucast_pkts': '455761740',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9034',
                                                                                                'out_octets': '199839168863',
                                                                                                'out_ucast_pkts': '403694400',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9034',
                                                                       'out_octets': '199839168863',
                                                                       'out_ucast_pkts': '403694400',
                                                                       'poller_hash': 'Mzc4OEJDcGVlcmludGZwb24gODAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '801',
                                                                                               'links': [       {       'connected_time': '6500553',
                                                                                                                        'fiber_length': 542,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.462,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzc4OEJDcGVlcnBvbnBvbiA4MDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '20 Main St N Milverton (Ellis HHBC '
                                                 '10433408)',
                                         'poller_hash': 'Mzc4OEJD',
                                         'serial_number': '3788BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:be:ab:f9',
                                         'uptime': '6500553'},
                                 {       'counter_source': '802/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:09:0f:09:01:00'],
                                                                       'id': '802/g4',
                                                                       'in_nucast_pkts': '11842123',
                                                                       'in_octets': '111348306079',
                                                                       'in_ucast_pkts': '394872470',
                                                                       'mac_address': 'cc:be:59:cb:be:10',
                                                                       'mtu': 2000,
                                                                       'name': '802/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '69663174',
                                                                       'out_octets': '790912485033',
                                                                       'out_ucast_pkts': '747831047',
                                                                       'poller_hash': 'NDJDOTE5cGVlcmludGY4MDIvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '802',
                                                                       'in_nucast_pkts': '69663174',
                                                                       'in_octets': '790912485033',
                                                                       'in_ucast_pkts': '747831047',
                                                                       'links': [       {       'in_nucast_pkts': '69663174',
                                                                                                'in_octets': '790912485033',
                                                                                                'in_ucast_pkts': '747831047',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11842123',
                                                                                                'out_octets': '111348306079',
                                                                                                'out_ucast_pkts': '394872470',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11842123',
                                                                       'out_octets': '111348306079',
                                                                       'out_ucast_pkts': '394872470',
                                                                       'poller_hash': 'NDJDOTE5cGVlcmludGZwb24gODAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '802',
                                                                                               'links': [       {       'connected_time': '64718596',
                                                                                                                        'fiber_length': 18015,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOTE5cGVlcnBvbnBvbiA4MDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4743 Line 52 Brunner (Milverton '
                                                 'Layer Farms)',
                                         'poller_hash': 'NDJDOTE5',
                                         'serial_number': '42C919',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:be:0a',
                                         'uptime': '64718596'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '803/g1',
                                                                       'name': '803/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEZBMjQ5cGVlcmludGY4MDMvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '803',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEZBMjQ5cGVlcmludGZwb24gODAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '803',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMjQ5cGVlcnBvbnBvbiA4MDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '8 MacBeth St Milverton (Abe '
                                                 'Hildebrandt 10435622)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "803" '
                                                                  '(location: GPON '
                                                                  'port "6/8"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:12:15.170Z',
                                         'poller_hash': 'NEZBMjQ5',
                                         'serial_number': '4FA249',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '804/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c0:56:27:68:f2:88'],
                                                                       'id': '804/g1',
                                                                       'in_nucast_pkts': '2652',
                                                                       'in_octets': '12329124854',
                                                                       'in_ucast_pkts': '26442156',
                                                                       'mac_address': '44:65:7f:33:21:6b',
                                                                       'mtu': 2000,
                                                                       'name': '804/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '834603',
                                                                       'out_octets': '81353254286',
                                                                       'out_ucast_pkts': '67212033',
                                                                       'poller_hash': 'NjQwQjJEcGVlcmludGY4MDQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '804',
                                                                       'in_nucast_pkts': '834603',
                                                                       'in_octets': '81353254286',
                                                                       'in_ucast_pkts': '67212033',
                                                                       'links': [       {       'in_nucast_pkts': '834603',
                                                                                                'in_octets': '81353254286',
                                                                                                'in_ucast_pkts': '67212033',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2652',
                                                                                                'out_octets': '12329124854',
                                                                                                'out_ucast_pkts': '26442156',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2652',
                                                                       'out_octets': '12329124854',
                                                                       'out_ucast_pkts': '26442156',
                                                                       'poller_hash': 'NjQwQjJEcGVlcmludGZwb24gODA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '804',
                                                                                               'links': [       {       'connected_time': '538547',
                                                                                                                        'fiber_length': 13247,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.724,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.492}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjJEcGVlcnBvbnBvbiA4MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5375 Rd 144 Brunner (Lisa Paxton '
                                                 '10432593)',
                                         'poller_hash': 'NjQwQjJE',
                                         'serial_number': '640B2D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:6a',
                                         'uptime': '538547'},
                                 {       'counter_source': '805/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '805/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '649893956984',
                                                                       'in_ucast_pkts': '17755138781',
                                                                       'mac_address': 'd0:76:8f:08:18:d3',
                                                                       'name': '805/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '104',
                                                                       'out_octets': '424171721774',
                                                                       'out_ucast_pkts': '576997206',
                                                                       'poller_hash': 'ODY5NDEwcGVlcmludGY4MDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '805',
                                                                       'in_nucast_pkts': '104',
                                                                       'in_octets': '424171721774',
                                                                       'in_ucast_pkts': '576997206',
                                                                       'links': [       {       'in_nucast_pkts': '104',
                                                                                                'in_octets': '424171721774',
                                                                                                'in_ucast_pkts': '576997206',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '649893956984',
                                                                                                'out_ucast_pkts': '17755138781',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '649893956984',
                                                                       'out_ucast_pkts': '17755138781',
                                                                       'poller_hash': 'ODY5NDEwcGVlcmludGZwb24gODA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '805',
                                                                                               'links': [       {       'connected_time': '21816439',
                                                                                                                        'fiber_length': 13514,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDEwcGVlcnBvbnBvbiA4MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5375 Road 144 Brunner (W & J Paxton '
                                                 '10400532)',
                                         'poller_hash': 'ODY5NDEw',
                                         'serial_number': '869410',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:d3',
                                         'uptime': '21816439'},
                                 {       'counter_source': '806/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:ad:24:60:e2:2b'],
                                                                       'id': '806/g1',
                                                                       'in_nucast_pkts': '53',
                                                                       'in_octets': '4986174164',
                                                                       'in_ucast_pkts': '43187695',
                                                                       'mac_address': '44:65:7f:7b:d4:21',
                                                                       'mtu': 2000,
                                                                       'name': '806/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1462308',
                                                                       'out_octets': '186686989602',
                                                                       'out_ucast_pkts': '129976292',
                                                                       'poller_hash': 'NUE2QkE2cGVlcmludGY4MDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '806',
                                                                       'in_nucast_pkts': '1462308',
                                                                       'in_octets': '186686989602',
                                                                       'in_ucast_pkts': '129976292',
                                                                       'links': [       {       'in_nucast_pkts': '1462308',
                                                                                                'in_octets': '186686989602',
                                                                                                'in_ucast_pkts': '129976292',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '53',
                                                                                                'out_octets': '4986174164',
                                                                                                'out_ucast_pkts': '43187695',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53',
                                                                       'out_octets': '4986174164',
                                                                       'out_ucast_pkts': '43187695',
                                                                       'poller_hash': 'NUE2QkE2cGVlcmludGZwb24gODA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '806',
                                                                                               'links': [       {       'connected_time': '688366',
                                                                                                                        'fiber_length': 961,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QkE2cGVlcnBvbnBvbiA4MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5 West St. Milverton (Michael '
                                                 'Gordon)',
                                         'poller_hash': 'NUE2QkE2',
                                         'serial_number': '5A6BA6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d4:21',
                                         'uptime': '688366'},
                                 {       'counter_source': '807/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '1c:61:b4:5c:5c:94'],
                                                                       'id': '807/g1',
                                                                       'in_nucast_pkts': '65454',
                                                                       'in_octets': '1041713716',
                                                                       'in_ucast_pkts': '5246522',
                                                                       'mac_address': '44:65:7f:7b:c3:b6',
                                                                       'mtu': 2000,
                                                                       'name': '807/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463946',
                                                                       'out_octets': '28509932539',
                                                                       'out_ucast_pkts': '22667750',
                                                                       'poller_hash': 'NUE2NjJEcGVlcmludGY4MDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '807',
                                                                       'in_nucast_pkts': '1463946',
                                                                       'in_octets': '28509932539',
                                                                       'in_ucast_pkts': '22667750',
                                                                       'links': [       {       'in_nucast_pkts': '1463946',
                                                                                                'in_octets': '28509932539',
                                                                                                'in_ucast_pkts': '22667750',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '65454',
                                                                                                'out_octets': '1041713716',
                                                                                                'out_ucast_pkts': '5246522',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '65454',
                                                                       'out_octets': '1041713716',
                                                                       'out_ucast_pkts': '5246522',
                                                                       'poller_hash': 'NUE2NjJEcGVlcmludGZwb24gODA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '807',
                                                                                               'links': [       {       'connected_time': '688467',
                                                                                                                        'fiber_length': 1487,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjJEcGVlcnBvbnBvbiA4MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '40 Tye St Milverton (B & P Huber '
                                                 '10403559)',
                                         'poller_hash': 'NUE2NjJE',
                                         'serial_number': '5A662D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:b6',
                                         'uptime': '688467'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '808/g1',
                                                                       'name': '808/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQUREcGVlcmludGY4MDgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '808',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQUREcGVlcmludGZwb24gODA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '808',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUREcGVlcnBvbnBvbiA4MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '13 Main St N Apt 5 Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "808" '
                                                                  '(location: GPON '
                                                                  'port "7/8"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:41.960Z',
                                         'poller_hash': 'NjQwQURE',
                                         'serial_number': '640ADD',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '809/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:96'],
                                                                       'id': '809/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '9720169780',
                                                                       'in_ucast_pkts': '58775370',
                                                                       'mac_address': '44:65:7f:33:21:b3',
                                                                       'mtu': 2000,
                                                                       'name': '809/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102920',
                                                                       'out_octets': '180519612770',
                                                                       'out_ucast_pkts': '129223887',
                                                                       'poller_hash': 'NjQwQjUxcGVlcmludGY4MDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '809',
                                                                       'in_nucast_pkts': '1102920',
                                                                       'in_octets': '180519612770',
                                                                       'in_ucast_pkts': '129223887',
                                                                       'links': [       {       'in_nucast_pkts': '1102920',
                                                                                                'in_octets': '180519612770',
                                                                                                'in_ucast_pkts': '129223887',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '9720169780',
                                                                                                'out_ucast_pkts': '58775370',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '9720169780',
                                                                       'out_ucast_pkts': '58775370',
                                                                       'poller_hash': 'NjQwQjUxcGVlcmludGZwb24gODA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '809',
                                                                                               'links': [       {       'connected_time': '688404',
                                                                                                                        'fiber_length': 1982,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.652,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.666}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjUxcGVlcnBvbnBvbiA4MDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '87 Reagan St. Milverton (Emily Mills '
                                                 '10435602)',
                                         'poller_hash': 'NjQwQjUx',
                                         'serial_number': '640B51',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:b2',
                                         'uptime': '688404'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '810/g1',
                                                                       'name': '810/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCRTlBcGVlcmludGY4MTAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '810',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJCRTlBcGVlcmludGZwb24gODEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '810',
                                                                                               'links': [       {       'connected_time': '6420495',
                                                                                                                        'fiber_length': 6930,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRTlBcGVlcnBvbnBvbiA4MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4895 Line 67 House (Ron and Deborah '
                                                 'Riddell)',
                                         'poller_hash': 'NDJCRTlB',
                                         'serial_number': '42BE9A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6a:12',
                                         'uptime': '6420495'},
                                 {       'counter_source': '811/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:22:32:43:8e:43'],
                                                                       'id': '811/g1',
                                                                       'in_nucast_pkts': '65',
                                                                       'in_octets': '175016373407',
                                                                       'in_ucast_pkts': '451812546',
                                                                       'mac_address': '44:65:7f:7b:c2:69',
                                                                       'mtu': 2000,
                                                                       'name': '811/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27288126',
                                                                       'out_octets': '973375556760',
                                                                       'out_ucast_pkts': '814317786',
                                                                       'poller_hash': 'NUE2NUJFcGVlcmludGY4MTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '811',
                                                                       'in_nucast_pkts': '27288126',
                                                                       'in_octets': '973375556760',
                                                                       'in_ucast_pkts': '814317786',
                                                                       'links': [       {       'in_nucast_pkts': '27288126',
                                                                                                'in_octets': '973375556760',
                                                                                                'in_ucast_pkts': '814317786',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '65',
                                                                                                'out_octets': '175016373407',
                                                                                                'out_ucast_pkts': '451812546',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '65',
                                                                       'out_octets': '175016373407',
                                                                       'out_ucast_pkts': '451812546',
                                                                       'poller_hash': 'NUE2NUJFcGVlcmludGZwb24gODEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '811',
                                                                                               'links': [       {       'connected_time': '12619346',
                                                                                                                        'fiber_length': 7171,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NUJFcGVlcnBvbnBvbiA4MTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4895 Line 67 Barn (Ron & Deb Riddell '
                                                 '10426917)',
                                         'poller_hash': 'NUE2NUJF',
                                         'serial_number': '5A65BE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c2:69',
                                         'uptime': '12619346'},
                                 {       'counter_source': '812/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e0:6d'],
                                                                       'id': '812/g1',
                                                                       'in_nucast_pkts': '654312',
                                                                       'in_octets': '1381727826743',
                                                                       'in_ucast_pkts': '4250401662',
                                                                       'mac_address': '44:65:7f:7b:b7:98',
                                                                       'mtu': 2000,
                                                                       'name': '812/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76600459',
                                                                       'out_octets': '14252832130077',
                                                                       'out_ucast_pkts': '11081123046',
                                                                       'poller_hash': 'NUE2MjIzcGVlcmludGY4MTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '812',
                                                                       'in_nucast_pkts': '76600459',
                                                                       'in_octets': '14252832130077',
                                                                       'in_ucast_pkts': '11081123046',
                                                                       'links': [       {       'in_nucast_pkts': '76600459',
                                                                                                'in_octets': '14252832130077',
                                                                                                'in_ucast_pkts': '11081123046',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '654312',
                                                                                                'out_octets': '1381727826743',
                                                                                                'out_ucast_pkts': '4250401662',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '654312',
                                                                       'out_octets': '1381727826743',
                                                                       'out_ucast_pkts': '4250401662',
                                                                       'poller_hash': 'NUE2MjIzcGVlcmludGZwb24gODEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '812',
                                                                                               'links': [       {       'connected_time': '67790372',
                                                                                                                        'fiber_length': 2877,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2MjIzcGVlcnBvbnBvbiA4MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4725 LINE 61 (MUELLER STEFAN)',
                                         'poller_hash': 'NUE2MjIz',
                                         'serial_number': '5A6223',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:b7:98',
                                         'uptime': '67790372'},
                                 {       'counter_source': '813/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '813/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '58790081170',
                                                                       'in_ucast_pkts': '463629882',
                                                                       'mac_address': 'cc:be:59:cb:6a:7a',
                                                                       'name': '813/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '2527133650555',
                                                                       'out_ucast_pkts': '1762970241',
                                                                       'poller_hash': 'NDJCRUE3cGVlcmludGY4MTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '813',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '2527133650555',
                                                                       'in_ucast_pkts': '1762970241',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '2527133650555',
                                                                                                'in_ucast_pkts': '1762970241',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '58790081170',
                                                                                                'out_ucast_pkts': '463629882',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '58790081170',
                                                                       'out_ucast_pkts': '463629882',
                                                                       'poller_hash': 'NDJCRUE3cGVlcmludGZwb24gODEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '813',
                                                                                               'links': [       {       'connected_time': '73034707',
                                                                                                                        'fiber_length': 8792,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRUE3cGVlcnBvbnBvbiA4MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5088 Line 58 Milverton (Franz & Susy '
                                                 'Arnold 10400118)',
                                         'poller_hash': 'NDJCRUE3',
                                         'serial_number': '42BEA7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6a:7a',
                                         'uptime': '73034707'},
                                 {       'counter_source': '814/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '814/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '110529187071',
                                                                       'in_ucast_pkts': '4624212648',
                                                                       'mac_address': 'cc:be:59:cb:ad:9a',
                                                                       'name': '814/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '114',
                                                                       'out_octets': '1204164832424',
                                                                       'out_ucast_pkts': '1052101752',
                                                                       'poller_hash': 'NDJDNzBCcGVlcmludGY4MTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '814',
                                                                       'in_nucast_pkts': '114',
                                                                       'in_octets': '1204164832424',
                                                                       'in_ucast_pkts': '1052101752',
                                                                       'links': [       {       'in_nucast_pkts': '114',
                                                                                                'in_octets': '1204164832424',
                                                                                                'in_ucast_pkts': '1052101752',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '110529187071',
                                                                                                'out_ucast_pkts': '4624212648',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '110529187071',
                                                                       'out_ucast_pkts': '4624212648',
                                                                       'poller_hash': 'NDJDNzBCcGVlcmludGZwb24gODE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '814',
                                                                                               'links': [       {       'connected_time': '11412210',
                                                                                                                        'fiber_length': 14700,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNzBCcGVlcnBvbnBvbiA4MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5137 Line 49 Gadshill (MJD FARMS '
                                                 'LTD)',
                                         'poller_hash': 'NDJDNzBC',
                                         'serial_number': '42C70B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ad:9a',
                                         'uptime': '11412210'},
                                 {       'counter_source': '815/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d5:a3'],
                                                                       'id': '815/g1',
                                                                       'in_nucast_pkts': '66618778',
                                                                       'in_octets': '381633014468',
                                                                       'in_ucast_pkts': '2717653620',
                                                                       'mac_address': 'cc:be:59:cb:6a:55',
                                                                       'mtu': 2000,
                                                                       'name': '815/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63485708336',
                                                                       'out_octets': '93626962161218',
                                                                       'out_ucast_pkts': '5748029905',
                                                                       'poller_hash': 'NDJCRUEycGVlcmludGY4MTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '815',
                                                                       'in_nucast_pkts': '63485708336',
                                                                       'in_octets': '93626962161218',
                                                                       'in_ucast_pkts': '5748029905',
                                                                       'links': [       {       'in_nucast_pkts': '63485708336',
                                                                                                'in_octets': '93626962161218',
                                                                                                'in_ucast_pkts': '5748029905',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '66618778',
                                                                                                'out_octets': '381633014468',
                                                                                                'out_ucast_pkts': '2717653620',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66618778',
                                                                       'out_octets': '381633014468',
                                                                       'out_ucast_pkts': '2717653620',
                                                                       'poller_hash': 'NDJCRUEycGVlcmludGZwb24gODE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '815',
                                                                                               'links': [       {       'connected_time': '21815978',
                                                                                                                        'fiber_length': 9078,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRUEycGVlcnBvbnBvbiA4MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5126 Line 58 Milverton (J & B '
                                                 'Crowley 10404213)',
                                         'poller_hash': 'NDJCRUEy',
                                         'serial_number': '42BEA2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6a:52',
                                         'uptime': '21815978'},
                                 {       'counter_source': '816/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '08:36:c9:ed:65:1d'],
                                                                       'id': '816/g1',
                                                                       'in_nucast_pkts': '33101068',
                                                                       'in_octets': '83229599910',
                                                                       'in_ucast_pkts': '770571419',
                                                                       'mac_address': '44:65:7f:7b:c5:87',
                                                                       'mtu': 2000,
                                                                       'name': '816/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '74397042',
                                                                       'out_octets': '2246149415031',
                                                                       'out_ucast_pkts': '1568240044',
                                                                       'poller_hash': 'NUE2NkM4cGVlcmludGY4MTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '816',
                                                                       'in_nucast_pkts': '74397042',
                                                                       'in_octets': '2246149415031',
                                                                       'in_ucast_pkts': '1568240044',
                                                                       'links': [       {       'in_nucast_pkts': '74397042',
                                                                                                'in_octets': '2246149415031',
                                                                                                'in_ucast_pkts': '1568240044',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '33101068',
                                                                                                'out_octets': '83229599910',
                                                                                                'out_ucast_pkts': '770571419',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33101068',
                                                                       'out_octets': '83229599910',
                                                                       'out_ucast_pkts': '770571419',
                                                                       'poller_hash': 'NUE2NkM4cGVlcmludGZwb24gODE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '816',
                                                                                               'links': [       {       'connected_time': '67790146',
                                                                                                                        'fiber_length': 17199,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NkM4cGVlcnBvbnBvbiA4MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4595 Line 49 (PEHLKE MARJORIE)',
                                         'poller_hash': 'NUE2NkM4',
                                         'serial_number': '5A66C8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c5:87',
                                         'uptime': '67790146'},
                                 {       'counter_source': '817/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e3:64'],
                                                                       'id': '817/g1',
                                                                       'in_nucast_pkts': '12251',
                                                                       'in_octets': '6246676163',
                                                                       'in_ucast_pkts': '41940401',
                                                                       'mac_address': 'cc:be:59:cb:9e:5d',
                                                                       'mtu': 2000,
                                                                       'name': '817/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1446993',
                                                                       'out_octets': '158629741225',
                                                                       'out_ucast_pkts': '111969864',
                                                                       'poller_hash': 'NDJDNTIzcGVlcmludGY4MTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '817',
                                                                       'in_nucast_pkts': '1446993',
                                                                       'in_octets': '158629741225',
                                                                       'in_ucast_pkts': '111969864',
                                                                       'links': [       {       'in_nucast_pkts': '1446993',
                                                                                                'in_octets': '158629741225',
                                                                                                'in_ucast_pkts': '111969864',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12251',
                                                                                                'out_octets': '6246676163',
                                                                                                'out_ucast_pkts': '41940401',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12251',
                                                                       'out_octets': '6246676163',
                                                                       'out_ucast_pkts': '41940401',
                                                                       'poller_hash': 'NDJDNTIzcGVlcmludGZwb24gODE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '817',
                                                                                               'links': [       {       'connected_time': '688407',
                                                                                                                        'fiber_length': 2054,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNTIzcGVlcnBvbnBvbiA4MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '84 Reagan St Milverton (Martin '
                                                 'Andrews 10435676)',
                                         'poller_hash': 'NDJDNTIz',
                                         'serial_number': '42C523',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:9e:5a',
                                         'uptime': '688407'},
                                 {       'counter_source': '820/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:05:8a'],
                                                                       'id': '820/g1',
                                                                       'in_nucast_pkts': '859645',
                                                                       'in_octets': '48473691804',
                                                                       'in_ucast_pkts': '660389071',
                                                                       'mac_address': 'cc:be:59:cb:ad:95',
                                                                       'mtu': 2000,
                                                                       'name': '820/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '118286011',
                                                                       'out_octets': '3678250237079',
                                                                       'out_ucast_pkts': '2586029073',
                                                                       'poller_hash': 'NDJDNzBBcGVlcmludGY4MjAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '820',
                                                                       'in_nucast_pkts': '118286011',
                                                                       'in_octets': '3678250237079',
                                                                       'in_ucast_pkts': '2586029073',
                                                                       'links': [       {       'in_nucast_pkts': '118286011',
                                                                                                'in_octets': '3678250237079',
                                                                                                'in_ucast_pkts': '2586029073',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '859645',
                                                                                                'out_octets': '48473691804',
                                                                                                'out_ucast_pkts': '660389071',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '859645',
                                                                       'out_octets': '48473691804',
                                                                       'out_ucast_pkts': '660389071',
                                                                       'poller_hash': 'NDJDNzBBcGVlcmludGZwb24gODIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '820',
                                                                                               'links': [       {       'connected_time': '15498764',
                                                                                                                        'fiber_length': 1422,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNzBBcGVlcnBvbnBvbiA4MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 MacBeth St (David Sloat 10401199)',
                                         'poller_hash': 'NDJDNzBB',
                                         'serial_number': '42C70A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ad:92',
                                         'uptime': '15498764'},
                                 {       'counter_source': '821/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '821/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '76790005929',
                                                                       'in_ucast_pkts': '581800318',
                                                                       'mac_address': 'cc:be:59:cb:6a:42',
                                                                       'name': '821/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '94',
                                                                       'out_octets': '3220116053399',
                                                                       'out_ucast_pkts': '2229629808',
                                                                       'poller_hash': 'NDJCRUEwcGVlcmludGY4MjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '821',
                                                                       'in_nucast_pkts': '94',
                                                                       'in_octets': '3220116053399',
                                                                       'in_ucast_pkts': '2229629808',
                                                                       'links': [       {       'in_nucast_pkts': '94',
                                                                                                'in_octets': '3220116053399',
                                                                                                'in_ucast_pkts': '2229629808',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '76790005929',
                                                                                                'out_ucast_pkts': '581800318',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '76790005929',
                                                                       'out_ucast_pkts': '581800318',
                                                                       'poller_hash': 'NDJCRUEwcGVlcmludGZwb24gODIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '821',
                                                                                               'links': [       {       'connected_time': '688551',
                                                                                                                        'fiber_length': 1936,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRUEwcGVlcnBvbnBvbiA4MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4633 Line 61 Milverton (James '
                                                 'Kuepfer 10400294)',
                                         'poller_hash': 'NDJCRUEw',
                                         'serial_number': '42BEA0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6a:42',
                                         'uptime': '688551'},
                                 {       'counter_source': '823/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:d0:53'],
                                                                       'id': '823/g1',
                                                                       'in_nucast_pkts': '15',
                                                                       'in_octets': '6561547602',
                                                                       'in_ucast_pkts': '6941177',
                                                                       'mac_address': 'ec:4f:82:7d:cb:97',
                                                                       'mtu': 2000,
                                                                       'name': '823/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464380',
                                                                       'out_octets': '3102915082',
                                                                       'out_ucast_pkts': '4531554',
                                                                       'poller_hash': 'MzAwRUY4cGVlcmludGY4MjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '823',
                                                                       'in_nucast_pkts': '1464380',
                                                                       'in_octets': '3102915082',
                                                                       'in_ucast_pkts': '4531554',
                                                                       'links': [       {       'in_nucast_pkts': '1464380',
                                                                                                'in_octets': '3102915082',
                                                                                                'in_ucast_pkts': '4531554',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15',
                                                                                                'out_octets': '6561547602',
                                                                                                'out_ucast_pkts': '6941177',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15',
                                                                       'out_octets': '6561547602',
                                                                       'out_ucast_pkts': '6941177',
                                                                       'poller_hash': 'MzAwRUY4cGVlcmludGZwb24gODIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '823',
                                                                                               'links': [       {       'connected_time': '688358',
                                                                                                                        'fiber_length': 889,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRUY4cGVlcnBvbnBvbiA4MjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '44 Main St S Milverton (JL Hallahan '
                                                 'Ltd 10435438)',
                                         'poller_hash': 'MzAwRUY4',
                                         'serial_number': '300EF8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:cb:97',
                                         'uptime': '688358'},
                                 {       'counter_source': '824/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:63:7d:7a:80:cb'],
                                                                       'id': '824/g1',
                                                                       'in_nucast_pkts': '2754',
                                                                       'in_octets': '13109530874',
                                                                       'in_ucast_pkts': '25826247',
                                                                       'mac_address': '44:65:7f:7b:c4:58',
                                                                       'mtu': 2000,
                                                                       'name': '824/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6443563',
                                                                       'out_octets': '81088845913',
                                                                       'out_ucast_pkts': '67198824',
                                                                       'poller_hash': 'NUE2NjYzcGVlcmludGY4MjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '824',
                                                                       'in_nucast_pkts': '6443563',
                                                                       'in_octets': '81088845913',
                                                                       'in_ucast_pkts': '67198824',
                                                                       'links': [       {       'in_nucast_pkts': '6443563',
                                                                                                'in_octets': '81088845913',
                                                                                                'in_ucast_pkts': '67198824',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2754',
                                                                                                'out_octets': '13109530874',
                                                                                                'out_ucast_pkts': '25826247',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2754',
                                                                       'out_octets': '13109530874',
                                                                       'out_ucast_pkts': '25826247',
                                                                       'poller_hash': 'NUE2NjYzcGVlcmludGZwb24gODI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '824',
                                                                                               'links': [       {       'connected_time': '2969325',
                                                                                                                        'fiber_length': 18200,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjYzcGVlcnBvbnBvbiA4MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '198 Queen St Monkton (Peter Enns '
                                                 '10431166)',
                                         'poller_hash': 'NUE2NjYz',
                                         'serial_number': '5A6663',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:58',
                                         'uptime': '2969325'},
                                 {       'counter_source': '825/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0f:39'],
                                                                       'id': '825/g1',
                                                                       'in_nucast_pkts': '11204',
                                                                       'in_octets': '206490459383',
                                                                       'in_ucast_pkts': '1443954117',
                                                                       'mac_address': 'cc:be:59:cb:77:85',
                                                                       'mtu': 2000,
                                                                       'name': '825/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150650055',
                                                                       'out_octets': '8121686598315',
                                                                       'out_ucast_pkts': '5625215652',
                                                                       'poller_hash': 'NDJDMDQ4cGVlcmludGY4MjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '825',
                                                                       'in_nucast_pkts': '150650055',
                                                                       'in_octets': '8121686598315',
                                                                       'in_ucast_pkts': '5625215652',
                                                                       'links': [       {       'in_nucast_pkts': '150650055',
                                                                                                'in_octets': '8121686598315',
                                                                                                'in_ucast_pkts': '5625215652',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11204',
                                                                                                'out_octets': '206490459383',
                                                                                                'out_ucast_pkts': '1443954117',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11204',
                                                                       'out_octets': '206490459383',
                                                                       'out_ucast_pkts': '1443954117',
                                                                       'poller_hash': 'NDJDMDQ4cGVlcmludGZwb24gODI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '825',
                                                                                               'links': [       {       'connected_time': '21816174',
                                                                                                                        'fiber_length': 17534,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMDQ4cGVlcnBvbnBvbiA4MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '194 SELBY AV (SMITH LORI)',
                                         'poller_hash': 'NDJDMDQ4',
                                         'serial_number': '42C048',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:77:82',
                                         'uptime': '21816174'},
                                 {       'counter_source': '827/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '827/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '205238582744',
                                                                       'in_ucast_pkts': '1574683757',
                                                                       'mac_address': 'cc:be:59:cb:bb:da',
                                                                       'name': '827/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '5626090873116',
                                                                       'out_ucast_pkts': '4066998331',
                                                                       'poller_hash': 'NDJDOEQzcGVlcmludGY4MjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '827',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '5626090873116',
                                                                       'in_ucast_pkts': '4066998331',
                                                                       'links': [       {       'in_nucast_pkts': '117',
                                                                                                'in_octets': '5626090873116',
                                                                                                'in_ucast_pkts': '4066998331',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '205238582744',
                                                                                                'out_ucast_pkts': '1574683757',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '205238582744',
                                                                       'out_ucast_pkts': '1574683757',
                                                                       'poller_hash': 'NDJDOEQzcGVlcmludGZwb24gODI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '827',
                                                                                               'links': [       {       'connected_time': '688556',
                                                                                                                        'fiber_length': 3399,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOEQzcGVlcnBvbnBvbiA4Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4775 LINE 61 (BAKER TYLER & JULIE)',
                                         'poller_hash': 'NDJDOEQz',
                                         'serial_number': '42C8D3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bb:da',
                                         'uptime': '688556'},
                                 {       'counter_source': '828/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:48:ef'],
                                                                       'id': '828/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '2307601809',
                                                                       'in_ucast_pkts': '16556334',
                                                                       'mac_address': '44:65:7f:33:21:df',
                                                                       'mtu': 2000,
                                                                       'name': '828/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102969',
                                                                       'out_octets': '48315099442',
                                                                       'out_ucast_pkts': '33947714',
                                                                       'poller_hash': 'NjQwQjY3cGVlcmludGY4MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '828',
                                                                       'in_nucast_pkts': '1102969',
                                                                       'in_octets': '48315099442',
                                                                       'in_ucast_pkts': '33947714',
                                                                       'links': [       {       'in_nucast_pkts': '1102969',
                                                                                                'in_octets': '48315099442',
                                                                                                'in_ucast_pkts': '33947714',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '2307601809',
                                                                                                'out_ucast_pkts': '16556334',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '2307601809',
                                                                       'out_ucast_pkts': '16556334',
                                                                       'poller_hash': 'NjQwQjY3cGVlcmludGZwb24gODI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '828',
                                                                                               'links': [       {       'connected_time': '688406',
                                                                                                                        'fiber_length': 1938,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.394,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.21}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjY3cGVlcnBvbnBvbiA4Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '2 Coulter St Milverton (Lynn Giles '
                                                 '10434418)',
                                         'poller_hash': 'NjQwQjY3',
                                         'serial_number': '640B67',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:de',
                                         'uptime': '688406'},
                                 {       'counter_source': '829/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '829/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '499475773766',
                                                                       'in_ucast_pkts': '4263014785',
                                                                       'mac_address': 'cc:be:59:cb:9b:8a',
                                                                       'name': '829/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '140',
                                                                       'out_octets': '13974742213691',
                                                                       'out_ucast_pkts': '9725663706',
                                                                       'poller_hash': 'NDJDNEM5cGVlcmludGY4MjkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '829',
                                                                       'in_nucast_pkts': '140',
                                                                       'in_octets': '13974742213691',
                                                                       'in_ucast_pkts': '9725663706',
                                                                       'links': [       {       'in_nucast_pkts': '140',
                                                                                                'in_octets': '13974742213691',
                                                                                                'in_ucast_pkts': '9725663706',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '499475773766',
                                                                                                'out_ucast_pkts': '4263014785',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '499475773766',
                                                                       'out_ucast_pkts': '4263014785',
                                                                       'poller_hash': 'NDJDNEM5cGVlcmludGZwb24gODI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '829',
                                                                                               'links': [       {       'connected_time': '11726986',
                                                                                                                        'fiber_length': 18483,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNEM5cGVlcnBvbnBvbiA4Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '140 Brock St Monkton (R & B Buchanan '
                                                 '10430719)',
                                         'poller_hash': 'NDJDNEM5',
                                         'serial_number': '42C4C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:9b:8a',
                                         'uptime': '11726986'},
                                 {       'counter_source': '830/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:54:ab:1b'],
                                                                       'id': '830/g2',
                                                                       'in_nucast_pkts': '12078',
                                                                       'in_octets': '16800899252',
                                                                       'in_ucast_pkts': '83925712',
                                                                       'mac_address': 'cc:be:59:cb:fa:e6',
                                                                       'mtu': 2000,
                                                                       'name': '830/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447308',
                                                                       'out_octets': '277696738939',
                                                                       'out_ucast_pkts': '206294021',
                                                                       'poller_hash': 'NDJEMEI0cGVlcmludGY4MzAvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '830',
                                                                       'in_nucast_pkts': '1447308',
                                                                       'in_octets': '277696738939',
                                                                       'in_ucast_pkts': '206294021',
                                                                       'links': [       {       'in_nucast_pkts': '1447308',
                                                                                                'in_octets': '277696738939',
                                                                                                'in_ucast_pkts': '206294021',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12078',
                                                                                                'out_octets': '16800899252',
                                                                                                'out_ucast_pkts': '83925712',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12078',
                                                                       'out_octets': '16800899252',
                                                                       'out_ucast_pkts': '83925712',
                                                                       'poller_hash': 'NDJEMEI0cGVlcmludGZwb24gODMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '830',
                                                                                               'links': [       {       'connected_time': '688436',
                                                                                                                        'fiber_length': 2433,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMEI0cGVlcnBvbnBvbiA4MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '9 Colbalt St. S. Milverton (Jamie '
                                                 'Scott)',
                                         'poller_hash': 'NDJEMEI0',
                                         'serial_number': '42D0B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:fa:e2',
                                         'uptime': '688436'},
                                 {       'counter_source': '832/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:45:7f'],
                                                                       'id': '832/g1',
                                                                       'in_nucast_pkts': '13',
                                                                       'in_octets': '2875663724',
                                                                       'in_ucast_pkts': '23291485',
                                                                       'mac_address': 'cc:be:59:cb:e0:75',
                                                                       'mtu': 2000,
                                                                       'name': '832/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120259',
                                                                       'out_octets': '91518928839',
                                                                       'out_ucast_pkts': '69675094',
                                                                       'poller_hash': 'NDJDRDY2cGVlcmludGY4MzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '832',
                                                                       'in_nucast_pkts': '1120259',
                                                                       'in_octets': '91518928839',
                                                                       'in_ucast_pkts': '69675094',
                                                                       'links': [       {       'in_nucast_pkts': '1120259',
                                                                                                'in_octets': '91518928839',
                                                                                                'in_ucast_pkts': '69675094',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13',
                                                                                                'out_octets': '2875663724',
                                                                                                'out_ucast_pkts': '23291485',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13',
                                                                       'out_octets': '2875663724',
                                                                       'out_ucast_pkts': '23291485',
                                                                       'poller_hash': 'NDJDRDY2cGVlcmludGZwb24gODMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '832',
                                                                                               'links': [       {       'connected_time': '538547',
                                                                                                                        'fiber_length': 13955,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRDY2cGVlcnBvbnBvbiA4MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5316 Road 144 Brunner (W & J Paxton '
                                                 '10435748)',
                                         'poller_hash': 'NDJDRDY2',
                                         'serial_number': '42CD66',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e0:72',
                                         'uptime': '538547'},
                                 {       'counter_source': '836/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:d2:94:ad:58:1e'],
                                                                       'id': '836/g1',
                                                                       'in_nucast_pkts': '164',
                                                                       'in_octets': '340964536',
                                                                       'in_ucast_pkts': '1228300',
                                                                       'mac_address': '44:65:7f:7b:c1:e2',
                                                                       'mtu': 2000,
                                                                       'name': '836/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120490',
                                                                       'out_octets': '3443800972',
                                                                       'out_ucast_pkts': '2640091',
                                                                       'poller_hash': 'NUE2NTkxcGVlcmludGY4MzYvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '836',
                                                                       'in_nucast_pkts': '1120490',
                                                                       'in_octets': '3443800972',
                                                                       'in_ucast_pkts': '2640091',
                                                                       'links': [       {       'in_nucast_pkts': '1120490',
                                                                                                'in_octets': '3443800972',
                                                                                                'in_ucast_pkts': '2640091',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '164',
                                                                                                'out_octets': '340964536',
                                                                                                'out_ucast_pkts': '1228300',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '164',
                                                                       'out_octets': '340964536',
                                                                       'out_ucast_pkts': '1228300',
                                                                       'poller_hash': 'NUE2NTkxcGVlcmludGZwb24gODM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '836',
                                                                                               'links': [       {       'connected_time': '538565',
                                                                                                                        'fiber_length': 14735,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NTkxcGVlcnBvbnBvbiA4MzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5177 Line 49 Gadshill (John Mueller '
                                                 '10400461)',
                                         'poller_hash': 'NUE2NTkx',
                                         'serial_number': '5A6591',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c1:e2',
                                         'uptime': '538565'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '837/g1',
                                                                       'name': '837/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCRURGcGVlcmludGY4MzcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '837',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJCRURGcGVlcmludGZwb24gODM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '837',
                                                                                               'links': [       {       'connected_time': '8468173',
                                                                                                                        'fiber_length': 1032,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRURGcGVlcnBvbnBvbiA4Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '8 John St Milverton (Craig Lines '
                                                 '10422212)',
                                         'poller_hash': 'NDJCRURG',
                                         'serial_number': '42BEDF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6c:3a',
                                         'uptime': '8468173'},
                                 {       'counter_source': '838/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:0b:82:ed:dc:35',
                                                                                                        '00:18:39:73:29:7c',
                                                                                                        '1c:61:b4:03:cd:d5'],
                                                                       'id': '838/g1',
                                                                       'in_nucast_pkts': '86584',
                                                                       'in_octets': '87193062249',
                                                                       'in_ucast_pkts': '532479831',
                                                                       'mac_address': '44:65:7f:33:21:7b',
                                                                       'mtu': 2000,
                                                                       'name': '838/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5224046',
                                                                       'out_octets': '1677918461929',
                                                                       'out_ucast_pkts': '1238787631',
                                                                       'poller_hash': 'NjQwQjM1cGVlcmludGY4MzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '838',
                                                                       'in_nucast_pkts': '5224046',
                                                                       'in_octets': '1677918461929',
                                                                       'in_ucast_pkts': '1238787631',
                                                                       'links': [       {       'in_nucast_pkts': '5224046',
                                                                                                'in_octets': '1677918461929',
                                                                                                'in_ucast_pkts': '1238787631',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '86584',
                                                                                                'out_octets': '87193062249',
                                                                                                'out_ucast_pkts': '532479831',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '86584',
                                                                       'out_octets': '87193062249',
                                                                       'out_ucast_pkts': '532479831',
                                                                       'poller_hash': 'NjQwQjM1cGVlcmludGZwb24gODM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '838',
                                                                                               'links': [       {       'connected_time': '3113778',
                                                                                                                        'fiber_length': 18444,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.362,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.186}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjM1cGVlcnBvbnBvbiA4Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '150 Brock St Monkton (Ryan Buchanan '
                                                 '10430196)',
                                         'poller_hash': 'NjQwQjM1',
                                         'serial_number': '640B35',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:7a',
                                         'uptime': '3113778'},
                                 {       'counter_source': '841/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:58:c7'],
                                                                       'id': '841/g1',
                                                                       'in_nucast_pkts': '11851',
                                                                       'in_octets': '528109164',
                                                                       'in_ucast_pkts': '2670714',
                                                                       'mac_address': 'd0:76:8f:08:14:06',
                                                                       'mtu': 2000,
                                                                       'name': '841/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447126',
                                                                       'out_octets': '11206859508',
                                                                       'out_ucast_pkts': '8724578',
                                                                       'poller_hash': 'ODY5Mzc2cGVlcmludGY4NDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '841',
                                                                       'in_nucast_pkts': '1447126',
                                                                       'in_octets': '11206859508',
                                                                       'in_ucast_pkts': '8724578',
                                                                       'links': [       {       'in_nucast_pkts': '1447126',
                                                                                                'in_octets': '11206859508',
                                                                                                'in_ucast_pkts': '8724578',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11851',
                                                                                                'out_octets': '528109164',
                                                                                                'out_ucast_pkts': '2670714',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11851',
                                                                       'out_octets': '528109164',
                                                                       'out_ucast_pkts': '2670714',
                                                                       'poller_hash': 'ODY5Mzc2cGVlcmludGZwb24gODQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '841',
                                                                                               'links': [       {       'connected_time': '688420',
                                                                                                                        'fiber_length': 2119,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5Mzc2cGVlcnBvbnBvbiA4NDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '85 GERBER DR (SNYDER ASHLEY)',
                                         'poller_hash': 'ODY5Mzc2',
                                         'serial_number': '869376',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:14:03',
                                         'uptime': '688420'},
                                 {       'counter_source': '843/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:cb:ac'],
                                                                       'id': '843/g1',
                                                                       'in_nucast_pkts': '302479',
                                                                       'in_octets': '56497584723',
                                                                       'in_ucast_pkts': '96347000',
                                                                       'mac_address': 'cc:be:59:cb:ba:25',
                                                                       'mtu': 2000,
                                                                       'name': '843/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '34674642',
                                                                       'out_octets': '160800427464',
                                                                       'out_ucast_pkts': '145024322',
                                                                       'poller_hash': 'NDJDODlDcGVlcmludGY4NDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '843',
                                                                       'in_nucast_pkts': '34674642',
                                                                       'in_octets': '160800427464',
                                                                       'in_ucast_pkts': '145024322',
                                                                       'links': [       {       'in_nucast_pkts': '34674642',
                                                                                                'in_octets': '160800427464',
                                                                                                'in_ucast_pkts': '145024322',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '302479',
                                                                                                'out_octets': '56497584723',
                                                                                                'out_ucast_pkts': '96347000',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '302479',
                                                                       'out_octets': '56497584723',
                                                                       'out_ucast_pkts': '96347000',
                                                                       'poller_hash': 'NDJDODlDcGVlcmludGZwb24gODQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '843',
                                                                                               'links': [       {       'connected_time': '15472874',
                                                                                                                        'fiber_length': 2276,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDODlDcGVlcnBvbnBvbiA4NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4660 LINE 61 (MILVERTON ROD & GUN '
                                                 'CLUB INC)',
                                         'poller_hash': 'NDJDODlD',
                                         'serial_number': '42C89C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ba:22',
                                         'uptime': '15472874'},
                                 {       'counter_source': '844/F1+844/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:bb:34',
                                                                                                        '00:03:e6:db:30:70'],
                                                                       'id': '844/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45420',
                                                                       'in_ucast_pkts': '266',
                                                                       'name': '844/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '3410',
                                                                       'out_octets': '1522978',
                                                                       'out_ucast_pkts': '164',
                                                                       'poller_hash': 'MTI3NEE1MXBlZXJpbnRmODQ0L0Yx'},
                                                               {       'admin_up': True,
                                                                       'id': '844/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '186616',
                                                                       'in_ucast_pkts': '527',
                                                                       'mac_address': '04:bc:9f:5c:1e:89',
                                                                       'name': '844/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '220909',
                                                                       'out_ucast_pkts': '1354',
                                                                       'poller_hash': 'MTI3NEE1MXBlZXJpbnRmODQ0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '844',
                                                                       'in_nucast_pkts': '3410',
                                                                       'in_octets': '1743887',
                                                                       'in_ucast_pkts': '1518',
                                                                       'links': [       {       'in_nucast_pkts': '3410',
                                                                                                'in_octets': '1743887',
                                                                                                'in_ucast_pkts': '1518',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '232036',
                                                                                                'out_ucast_pkts': '793',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '232036',
                                                                       'out_ucast_pkts': '793',
                                                                       'poller_hash': 'MTI3NEE1MXBlZXJpbnRmcG9uIDg0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '844',
                                                                                               'links': [       {       'connected_time': '538551',
                                                                                                                        'fiber_length': 16166,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1MXBlZXJwb25wb24gODQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5333 Line 49 Gadshill (Gerald Pol '
                                                 '10401245)',
                                         'poller_hash': 'MTI3NEE1MQ==',
                                         'serial_number': '1274A51',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:89',
                                         'uptime': '538551'},
                                 {       'counter_source': '845/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d3:75'],
                                                                       'id': '845/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '4649414274',
                                                                       'in_ucast_pkts': '17097234',
                                                                       'mac_address': 'cc:be:59:cb:a1:ed',
                                                                       'mtu': 2000,
                                                                       'name': '845/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3187206',
                                                                       'out_octets': '91802601251',
                                                                       'out_ucast_pkts': '65272219',
                                                                       'poller_hash': 'NDJDNTk1cGVlcmludGY4NDUvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '845',
                                                                       'in_nucast_pkts': '3187206',
                                                                       'in_octets': '91802601251',
                                                                       'in_ucast_pkts': '65272219',
                                                                       'links': [       {       'in_nucast_pkts': '3187206',
                                                                                                'in_octets': '91802601251',
                                                                                                'in_ucast_pkts': '65272219',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '4649414274',
                                                                                                'out_ucast_pkts': '17097234',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '4649414274',
                                                                       'out_ucast_pkts': '17097234',
                                                                       'poller_hash': 'NDJDNTk1cGVlcmludGZwb24gODQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '845',
                                                                                               'links': [       {       'connected_time': '1478640',
                                                                                                                        'fiber_length': 18505,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNTk1cGVlcnBvbnBvbiA4NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '136 Brock St Monkton (Dale '
                                                 'MacCannell 10435836)',
                                         'poller_hash': 'NDJDNTk1',
                                         'serial_number': '42C595',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:a1:ea',
                                         'uptime': '1478640'},
                                 {       'counter_source': '846/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '846/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '127986034978',
                                                                       'in_ucast_pkts': '718741852',
                                                                       'mac_address': 'ec:4f:82:0d:6a:69',
                                                                       'name': '846/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '2627448240657',
                                                                       'out_ucast_pkts': '2028556457',
                                                                       'poller_hash': 'MjA1REY0cGVlcmludGY4NDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '846',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '2627448240657',
                                                                       'in_ucast_pkts': '2028556457',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '2627448240657',
                                                                                                'in_ucast_pkts': '2028556457',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '127986034978',
                                                                                                'out_ucast_pkts': '718741852',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '127986034978',
                                                                       'out_ucast_pkts': '718741852',
                                                                       'poller_hash': 'MjA1REY0cGVlcmludGZwb24gODQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '846',
                                                                                               'links': [       {       'connected_time': '688489',
                                                                                                                        'fiber_length': 559,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REY0cGVlcnBvbnBvbiA4NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '4 MAIN ST S (CASKANETTE BENJAMEN)',
                                         'poller_hash': 'MjA1REY0',
                                         'serial_number': '205DF4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:69',
                                         'uptime': '688489'},
                                 {       'counter_source': '847/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b9:1f'],
                                                                       'id': '847/g1',
                                                                       'in_nucast_pkts': '5312',
                                                                       'in_octets': '2106259979',
                                                                       'in_ucast_pkts': '13869220',
                                                                       'mac_address': 'ec:4f:82:0d:68:11',
                                                                       'mtu': 2000,
                                                                       'name': '847/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '509800',
                                                                       'out_octets': '68617938690',
                                                                       'out_ucast_pkts': '51517436',
                                                                       'poller_hash': 'MjA1REIxcGVlcmludGY4NDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '847',
                                                                       'in_nucast_pkts': '509800',
                                                                       'in_octets': '68617938690',
                                                                       'in_ucast_pkts': '51517436',
                                                                       'links': [       {       'in_nucast_pkts': '509800',
                                                                                                'in_octets': '68617938690',
                                                                                                'in_ucast_pkts': '51517436',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5312',
                                                                                                'out_octets': '2106259979',
                                                                                                'out_ucast_pkts': '13869220',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5312',
                                                                       'out_octets': '2106259979',
                                                                       'out_ucast_pkts': '13869220',
                                                                       'poller_hash': 'MjA1REIxcGVlcmludGZwb24gODQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '847',
                                                                                               'links': [       {       'connected_time': '247592',
                                                                                                                        'fiber_length': 1683,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -7.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REIxcGVlcnBvbnBvbiA4NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 WILLIAM ST (ANDRE JOLY)',
                                         'poller_hash': 'MjA1REIx',
                                         'serial_number': '205DB1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:0e',
                                         'uptime': '247592'},
                                 {       'counter_source': '849/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '849/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '80483512674',
                                                                       'in_ucast_pkts': '712907911',
                                                                       'mac_address': 'cc:be:59:cc:03:b2',
                                                                       'name': '849/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '4307485090650',
                                                                       'out_ucast_pkts': '2956760449',
                                                                       'poller_hash': 'NDJEMUNFcGVlcmludGY4NDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '849',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '4307485090650',
                                                                       'in_ucast_pkts': '2956760449',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '4307485090650',
                                                                                                'in_ucast_pkts': '2956760449',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '80483512674',
                                                                                                'out_ucast_pkts': '712907911',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '80483512674',
                                                                       'out_ucast_pkts': '712907911',
                                                                       'poller_hash': 'NDJEMUNFcGVlcmludGZwb24gODQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '849',
                                                                                               'links': [       {       'connected_time': '2788576',
                                                                                                                        'fiber_length': 16450,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMUNFcGVlcnBvbnBvbiA4NDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5113 Perth Road 135 Rostock (B & D '
                                                 'Debus 10402399)',
                                         'poller_hash': 'NDJEMUNF',
                                         'serial_number': '42D1CE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:03:b2',
                                         'uptime': '2788576'},
                                 {       'counter_source': '850/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:dd:a9:91:e4:90'],
                                                                       'id': '850/g1',
                                                                       'in_nucast_pkts': '6838',
                                                                       'in_octets': '131951369874',
                                                                       'in_ucast_pkts': '402438034',
                                                                       'mac_address': '44:65:7f:7b:c3:71',
                                                                       'mtu': 2000,
                                                                       'name': '850/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76680156',
                                                                       'out_octets': '1387949203861',
                                                                       'out_ucast_pkts': '1090256969',
                                                                       'poller_hash': 'NUE2NjE2cGVlcmludGY4NTAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '850',
                                                                       'in_nucast_pkts': '76680156',
                                                                       'in_octets': '1387949203861',
                                                                       'in_ucast_pkts': '1090256969',
                                                                       'links': [       {       'in_nucast_pkts': '76680156',
                                                                                                'in_octets': '1387949203861',
                                                                                                'in_ucast_pkts': '1090256969',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6838',
                                                                                                'out_octets': '131951369874',
                                                                                                'out_ucast_pkts': '402438034',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6838',
                                                                       'out_octets': '131951369874',
                                                                       'out_ucast_pkts': '402438034',
                                                                       'poller_hash': 'NUE2NjE2cGVlcmludGZwb24gODUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '850',
                                                                                               'links': [       {       'connected_time': '67790425',
                                                                                                                        'fiber_length': 16947,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjE2cGVlcnBvbnBvbiA4NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4860 Line 48 Gadshill (Bert '
                                                 'Groenestege 10400636)',
                                         'poller_hash': 'NUE2NjE2',
                                         'serial_number': '5A6616',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:71',
                                         'uptime': '67790425'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '851/g1',
                                                                       'name': '851/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2NjYxcGVlcmludGY4NTEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '851',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NUE2NjYxcGVlcmludGZwb24gODUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '851',
                                                                                               'links': [       {       'connected_time': '67790369',
                                                                                                                        'fiber_length': 16617,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjYxcGVlcnBvbnBvbiA4NTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4672 Line 49 Brunner (Stan Debus '
                                                 '10400507)',
                                         'poller_hash': 'NUE2NjYx',
                                         'serial_number': '5A6661',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:52',
                                         'uptime': '67790369'},
                                 {       'counter_source': '853/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:62:1f:51:e9:39'],
                                                                       'id': '853/g1',
                                                                       'in_nucast_pkts': '1370',
                                                                       'in_octets': '845195844',
                                                                       'in_ucast_pkts': '3428462',
                                                                       'mac_address': '44:65:7f:7b:c4:5e',
                                                                       'mtu': 2000,
                                                                       'name': '853/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447177',
                                                                       'out_octets': '12405562249',
                                                                       'out_ucast_pkts': '9708271',
                                                                       'poller_hash': 'NUE2NjY1cGVlcmludGY4NTMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '853',
                                                                       'in_nucast_pkts': '1447177',
                                                                       'in_octets': '12405562249',
                                                                       'in_ucast_pkts': '9708271',
                                                                       'links': [       {       'in_nucast_pkts': '1447177',
                                                                                                'in_octets': '12405562249',
                                                                                                'in_ucast_pkts': '9708271',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1370',
                                                                                                'out_octets': '845195844',
                                                                                                'out_ucast_pkts': '3428462',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1370',
                                                                       'out_octets': '845195844',
                                                                       'out_ucast_pkts': '3428462',
                                                                       'poller_hash': 'NUE2NjY1cGVlcmludGZwb24gODUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '853',
                                                                                               'links': [       {       'connected_time': '688391',
                                                                                                                        'fiber_length': 8704,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjY1cGVlcnBvbnBvbiA4NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5087 Line 58 Milverton (Dale Zehr '
                                                 '10401364)',
                                         'poller_hash': 'NUE2NjY1',
                                         'serial_number': '5A6665',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:5e',
                                         'uptime': '688391'},
                                 {       'counter_source': '855/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d4:d4'],
                                                                       'id': '855/g1',
                                                                       'in_nucast_pkts': '263',
                                                                       'in_octets': '17757050606',
                                                                       'in_ucast_pkts': '78207708',
                                                                       'mac_address': '44:65:7f:7b:c3:a4',
                                                                       'mtu': 2000,
                                                                       'name': '855/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82752801',
                                                                       'out_octets': '326807614594',
                                                                       'out_ucast_pkts': '264220210',
                                                                       'poller_hash': 'NUE2NjI3cGVlcmludGY4NTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '855',
                                                                       'in_nucast_pkts': '82752801',
                                                                       'in_octets': '326807614594',
                                                                       'in_ucast_pkts': '264220210',
                                                                       'links': [       {       'in_nucast_pkts': '82752801',
                                                                                                'in_octets': '326807614594',
                                                                                                'in_ucast_pkts': '264220210',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '263',
                                                                                                'out_octets': '17757050606',
                                                                                                'out_ucast_pkts': '78207708',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '263',
                                                                       'out_octets': '17757050606',
                                                                       'out_ucast_pkts': '78207708',
                                                                       'poller_hash': 'NUE2NjI3cGVlcmludGZwb24gODU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '855',
                                                                                               'links': [       {       'connected_time': '67790314',
                                                                                                                        'fiber_length': 1251,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjI3cGVlcnBvbnBvbiA4NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '87 Min St N Milverton (George '
                                                 'Raycraft 10400878)',
                                         'poller_hash': 'NUE2NjI3',
                                         'serial_number': '5A6627',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:a4',
                                         'uptime': '67790314'},
                                 {       'counter_source': '856/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:7b:8c:14:90:8f'],
                                                                       'id': '856/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '3629193117',
                                                                       'in_ucast_pkts': '14652912',
                                                                       'mac_address': '44:65:7f:33:21:87',
                                                                       'mtu': 2000,
                                                                       'name': '856/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102962',
                                                                       'out_octets': '55662378169',
                                                                       'out_ucast_pkts': '44702510',
                                                                       'poller_hash': 'NjQwQjNCcGVlcmludGY4NTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '856',
                                                                       'in_nucast_pkts': '1102962',
                                                                       'in_octets': '55662378169',
                                                                       'in_ucast_pkts': '44702510',
                                                                       'links': [       {       'in_nucast_pkts': '1102962',
                                                                                                'in_octets': '55662378169',
                                                                                                'in_ucast_pkts': '44702510',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '3629193117',
                                                                                                'out_ucast_pkts': '14652912',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '3629193117',
                                                                       'out_ucast_pkts': '14652912',
                                                                       'poller_hash': 'NjQwQjNCcGVlcmludGZwb24gODU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '856',
                                                                                               'links': [       {       'connected_time': '688407',
                                                                                                                        'fiber_length': 2096,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.58,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.088}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjNCcGVlcnBvbnBvbiA4NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '69 Reagan St Milverton (Richard '
                                                 'Battrick, 10436036)',
                                         'poller_hash': 'NjQwQjNC',
                                         'serial_number': '640B3B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:86',
                                         'uptime': '688407'},
                                 {       'counter_source': '857/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:f5:a2:00:96:a6'],
                                                                       'id': '857/g1',
                                                                       'in_nucast_pkts': '22789',
                                                                       'in_octets': '924487610',
                                                                       'in_ucast_pkts': '4007864',
                                                                       'mac_address': '44:65:7f:33:1f:21',
                                                                       'mtu': 2000,
                                                                       'name': '857/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120574',
                                                                       'out_octets': '28293322989',
                                                                       'out_ucast_pkts': '22914540',
                                                                       'poller_hash': 'NjQwQTBBcGVlcmludGY4NTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '857',
                                                                       'in_nucast_pkts': '1120574',
                                                                       'in_octets': '28293322989',
                                                                       'in_ucast_pkts': '22914540',
                                                                       'links': [       {       'in_nucast_pkts': '1120574',
                                                                                                'in_octets': '28293322989',
                                                                                                'in_ucast_pkts': '22914540',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22789',
                                                                                                'out_octets': '924487610',
                                                                                                'out_ucast_pkts': '4007864',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22789',
                                                                       'out_octets': '924487610',
                                                                       'out_ucast_pkts': '4007864',
                                                                       'poller_hash': 'NjQwQTBBcGVlcmludGZwb24gODU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '857',
                                                                                               'links': [       {       'connected_time': '688468',
                                                                                                                        'fiber_length': 962,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.518,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.484}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQTBBcGVlcnBvbnBvbiA4NTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '29 William St (Addition) (Lorilee '
                                                 'Streicher 10401618)',
                                         'poller_hash': 'NjQwQTBB',
                                         'serial_number': '640A0A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:1f:20',
                                         'uptime': '688468'},
                                 {       'counter_source': '858/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d5:0a'],
                                                                       'id': '858/g3',
                                                                       'in_nucast_pkts': '11083684',
                                                                       'in_octets': '191509581099',
                                                                       'in_ucast_pkts': '1106928612',
                                                                       'mac_address': 'cc:be:59:cb:a2:e7',
                                                                       'mtu': 2000,
                                                                       'name': '858/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13548691361',
                                                                       'out_octets': '24411996329342',
                                                                       'out_ucast_pkts': '4375920048',
                                                                       'poller_hash': 'NDJDNUI0cGVlcmludGY4NTgvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '858',
                                                                       'in_nucast_pkts': '13548691361',
                                                                       'in_octets': '24411996329342',
                                                                       'in_ucast_pkts': '4375920048',
                                                                       'links': [       {       'in_nucast_pkts': '13548691361',
                                                                                                'in_octets': '24411996329342',
                                                                                                'in_ucast_pkts': '4375920048',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11083684',
                                                                                                'out_octets': '191509581099',
                                                                                                'out_ucast_pkts': '1106928612',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11083684',
                                                                       'out_octets': '191509581099',
                                                                       'out_ucast_pkts': '1106928612',
                                                                       'poller_hash': 'NDJDNUI0cGVlcmludGZwb24gODU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '858',
                                                                                               'links': [       {       'connected_time': '71459163',
                                                                                                                        'fiber_length': 2217,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNUI0cGVlcnBvbnBvbiA4NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '67 Reagan St. Milverton (Joy '
                                                 'Sheridan 10435977)',
                                         'poller_hash': 'NDJDNUI0',
                                         'serial_number': '42C5B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:a2:e2',
                                         'uptime': '71459163'},
                                 {       'counter_source': '859/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:59:c0:c1:50:28'],
                                                                       'id': '859/g1',
                                                                       'in_nucast_pkts': '210',
                                                                       'in_octets': '3988108755',
                                                                       'in_ucast_pkts': '35555317',
                                                                       'mac_address': '44:65:7f:33:20:79',
                                                                       'mtu': 2000,
                                                                       'name': '859/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119791',
                                                                       'out_octets': '148546300735',
                                                                       'out_ucast_pkts': '105194226',
                                                                       'poller_hash': 'NjQwQUI2cGVlcmludGY4NTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '859',
                                                                       'in_nucast_pkts': '1119791',
                                                                       'in_octets': '148546300735',
                                                                       'in_ucast_pkts': '105194226',
                                                                       'links': [       {       'in_nucast_pkts': '1119791',
                                                                                                'in_octets': '148546300735',
                                                                                                'in_ucast_pkts': '105194226',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '210',
                                                                                                'out_octets': '3988108755',
                                                                                                'out_ucast_pkts': '35555317',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '210',
                                                                       'out_octets': '3988108755',
                                                                       'out_ucast_pkts': '35555317',
                                                                       'poller_hash': 'NjQwQUI2cGVlcmludGZwb24gODU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '859',
                                                                                               'links': [       {       'connected_time': '688442',
                                                                                                                        'fiber_length': 1550,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.548,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.2}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUI2cGVlcnBvbnBvbiA4NTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5 CObalt St (Matthew Yost 10436053)',
                                         'poller_hash': 'NjQwQUI2',
                                         'serial_number': '640AB6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:20:78',
                                         'uptime': '688442'},
                                 {       'counter_source': '860/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:b5:29:55'],
                                                                       'id': '860/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '1029496047',
                                                                       'in_ucast_pkts': '3061948',
                                                                       'mac_address': '44:65:7f:7b:b6:e1',
                                                                       'mtu': 2000,
                                                                       'name': '860/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434199',
                                                                       'out_octets': '8248706687',
                                                                       'out_ucast_pkts': '7340019',
                                                                       'poller_hash': 'NUE2MUU2cGVlcmludGY4NjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '860',
                                                                       'in_nucast_pkts': '1434199',
                                                                       'in_octets': '8248706687',
                                                                       'in_ucast_pkts': '7340019',
                                                                       'links': [       {       'in_nucast_pkts': '1434199',
                                                                                                'in_octets': '8248706687',
                                                                                                'in_ucast_pkts': '7340019',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '1029496047',
                                                                                                'out_ucast_pkts': '3061948',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '1029496047',
                                                                       'out_ucast_pkts': '3061948',
                                                                       'poller_hash': 'NUE2MUU2cGVlcmludGZwb24gODYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '860',
                                                                                               'links': [       {       'connected_time': '688562',
                                                                                                                        'fiber_length': 1728,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2MUU2cGVlcnBvbnBvbiA4NjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5 Murray Ct Milverton (Lorne '
                                                 'Steckley 10402394)',
                                         'poller_hash': 'NUE2MUU2',
                                         'serial_number': '5A61E6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:b6:e1',
                                         'uptime': '688562'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '861/g1',
                                                                       'name': '861/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzQ5RThBcGVlcmludGY4NjEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '861',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzQ5RThBcGVlcmludGZwb24gODYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '861',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ5RThBcGVlcnBvbnBvbiA4NjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '111 Mill St E (Adam Harloff)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "861" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-03-05T18:23:12.620Z',
                                         'poller_hash': 'MzQ5RThB',
                                         'serial_number': '349E8A',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '862/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '98:da:c4:63:4b:2a'],
                                                                       'id': '862/g1',
                                                                       'in_nucast_pkts': '112259',
                                                                       'in_octets': '1443866521174',
                                                                       'in_ucast_pkts': '1355923765',
                                                                       'mac_address': '44:65:7f:7b:c4:d3',
                                                                       'mtu': 2000,
                                                                       'name': '862/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82647393',
                                                                       'out_octets': '1292647788174',
                                                                       'out_ucast_pkts': '1432316429',
                                                                       'poller_hash': 'NUE2NjhDcGVlcmludGY4NjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '862',
                                                                       'in_nucast_pkts': '82647393',
                                                                       'in_octets': '1292647788174',
                                                                       'in_ucast_pkts': '1432316429',
                                                                       'links': [       {       'in_nucast_pkts': '82647393',
                                                                                                'in_octets': '1292647788174',
                                                                                                'in_ucast_pkts': '1432316429',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '112259',
                                                                                                'out_octets': '1443866521174',
                                                                                                'out_ucast_pkts': '1355923765',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '112259',
                                                                       'out_octets': '1443866521174',
                                                                       'out_ucast_pkts': '1355923765',
                                                                       'poller_hash': 'NUE2NjhDcGVlcmludGZwb24gODYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '862',
                                                                                               'links': [       {       'connected_time': '67790307',
                                                                                                                        'fiber_length': 593,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjhDcGVlcnBvbnBvbiA4NjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '12 Main St N Milverton (Wicked Snips '
                                                 '10402369)',
                                         'poller_hash': 'NUE2NjhD',
                                         'serial_number': '5A668C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:d3',
                                         'uptime': '67790307'},
                                 {       'counter_source': '863/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a5:e3'],
                                                                       'id': '863/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '12438812648',
                                                                       'in_ucast_pkts': '28322725',
                                                                       'mac_address': 'cc:be:59:cb:6d:5d',
                                                                       'mtu': 2000,
                                                                       'name': '863/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447108',
                                                                       'out_octets': '129795193071',
                                                                       'out_ucast_pkts': '97250738',
                                                                       'poller_hash': 'NDJCRjAzcGVlcmludGY4NjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '863',
                                                                       'in_nucast_pkts': '1447108',
                                                                       'in_octets': '129795193071',
                                                                       'in_ucast_pkts': '97250738',
                                                                       'links': [       {       'in_nucast_pkts': '1447108',
                                                                                                'in_octets': '129795193071',
                                                                                                'in_ucast_pkts': '97250738',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '12438812648',
                                                                                                'out_ucast_pkts': '28322725',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '12438812648',
                                                                       'out_ucast_pkts': '28322725',
                                                                       'poller_hash': 'NDJCRjAzcGVlcmludGZwb24gODYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '863',
                                                                                               'links': [       {       'connected_time': '688408',
                                                                                                                        'fiber_length': 2133,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjAzcGVlcnBvbnBvbiA4NjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '88 GERBER DRIVE (DEWAR JAMIE & '
                                                 'ANGIE)',
                                         'poller_hash': 'NDJCRjAz',
                                         'serial_number': '42BF03',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6d:5a',
                                         'uptime': '688408'},
                                 {       'counter_source': '865/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:eb:ab'],
                                                                       'id': '865/g1',
                                                                       'in_nucast_pkts': '635537',
                                                                       'in_octets': '344261072226',
                                                                       'in_ucast_pkts': '867688236',
                                                                       'mac_address': 'ec:4f:82:7e:09:07',
                                                                       'mtu': 2000,
                                                                       'name': '865/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83470597',
                                                                       'out_octets': '3821345487448',
                                                                       'out_ucast_pkts': '2940692213',
                                                                       'poller_hash': 'MzVFOTREcGVlcmludGY4NjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '865',
                                                                       'in_nucast_pkts': '83470597',
                                                                       'in_octets': '3821345487448',
                                                                       'in_ucast_pkts': '2940692213',
                                                                       'links': [       {       'in_nucast_pkts': '83470597',
                                                                                                'in_octets': '3821345487448',
                                                                                                'in_ucast_pkts': '2940692213',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '635537',
                                                                                                'out_octets': '344261072226',
                                                                                                'out_ucast_pkts': '867688236',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '635537',
                                                                       'out_octets': '344261072226',
                                                                       'out_ucast_pkts': '867688236',
                                                                       'poller_hash': 'MzVFOTREcGVlcmludGZwb24gODY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '865',
                                                                                               'links': [       {       'connected_time': '67790417',
                                                                                                                        'fiber_length': 5962,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTREcGVlcnBvbnBvbiA4NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4316 Line 61 Milverton (E & P Hill '
                                                 '10405285)',
                                         'poller_hash': 'MzVFOTRE',
                                         'serial_number': '35E94D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:07',
                                         'uptime': '67790417'},
                                 {       'counter_source': '866/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a0:04:60:ca:69:e9'],
                                                                       'id': '866/g1',
                                                                       'in_nucast_pkts': '201',
                                                                       'in_octets': '2601641211',
                                                                       'in_ucast_pkts': '14541117',
                                                                       'mac_address': '44:65:7f:33:21:9d',
                                                                       'mtu': 2000,
                                                                       'name': '866/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119502',
                                                                       'out_octets': '41748200566',
                                                                       'out_ucast_pkts': '32645653',
                                                                       'poller_hash': 'NjQwQjQ2cGVlcmludGY4NjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '866',
                                                                       'in_nucast_pkts': '1119502',
                                                                       'in_octets': '41748200566',
                                                                       'in_ucast_pkts': '32645653',
                                                                       'links': [       {       'in_nucast_pkts': '1119502',
                                                                                                'in_octets': '41748200566',
                                                                                                'in_ucast_pkts': '32645653',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '201',
                                                                                                'out_octets': '2601641211',
                                                                                                'out_ucast_pkts': '14541117',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '201',
                                                                       'out_octets': '2601641211',
                                                                       'out_ucast_pkts': '14541117',
                                                                       'poller_hash': 'NjQwQjQ2cGVlcmludGZwb24gODY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '866',
                                                                                               'links': [       {       'connected_time': '688447',
                                                                                                                        'fiber_length': 1235,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.328,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.376}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjQ2cGVlcnBvbnBvbiA4NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '46 Mill St E (Kyle Woodrow 10430299)',
                                         'poller_hash': 'NjQwQjQ2',
                                         'serial_number': '640B46',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:9c',
                                         'uptime': '688447'},
                                 {       'counter_source': '868/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:42:a1:5b:fc:71'],
                                                                       'id': '868/g1',
                                                                       'in_nucast_pkts': '401',
                                                                       'in_octets': '6455753157',
                                                                       'in_ucast_pkts': '68534506',
                                                                       'mac_address': '44:65:7f:7b:b7:80',
                                                                       'mtu': 2000,
                                                                       'name': '868/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463929',
                                                                       'out_octets': '248389225226',
                                                                       'out_ucast_pkts': '169096743',
                                                                       'poller_hash': 'NUE2MjFCcGVlcmludGY4NjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '868',
                                                                       'in_nucast_pkts': '1463929',
                                                                       'in_octets': '248389225226',
                                                                       'in_ucast_pkts': '169096743',
                                                                       'links': [       {       'in_nucast_pkts': '1463929',
                                                                                                'in_octets': '248389225226',
                                                                                                'in_ucast_pkts': '169096743',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '401',
                                                                                                'out_octets': '6455753157',
                                                                                                'out_ucast_pkts': '68534506',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '401',
                                                                       'out_octets': '6455753157',
                                                                       'out_ucast_pkts': '68534506',
                                                                       'poller_hash': 'NUE2MjFCcGVlcmludGZwb24gODY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '868',
                                                                                               'links': [       {       'connected_time': '688466',
                                                                                                                        'fiber_length': 1301,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2MjFCcGVlcnBvbnBvbiA4Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '44 Mill St E Milverton (Douglas Fink '
                                                 '10425982)',
                                         'poller_hash': 'NUE2MjFC',
                                         'serial_number': '5A621B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:b7:80',
                                         'uptime': '688466'},
                                 {       'counter_source': '869/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:42:a1:29:f2:75'],
                                                                       'id': '869/g1',
                                                                       'in_nucast_pkts': '30656',
                                                                       'in_octets': '871055710790',
                                                                       'in_ucast_pkts': '4406433883',
                                                                       'mac_address': 'cc:be:59:cb:bc:95',
                                                                       'mtu': 2000,
                                                                       'name': '869/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '84263560',
                                                                       'out_octets': '15020446034700',
                                                                       'out_ucast_pkts': '11193504916',
                                                                       'poller_hash': 'NDJDOEVBcGVlcmludGY4NjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '869',
                                                                       'in_nucast_pkts': '84263560',
                                                                       'in_octets': '15020446034700',
                                                                       'in_ucast_pkts': '11193504916',
                                                                       'links': [       {       'in_nucast_pkts': '84263560',
                                                                                                'in_octets': '15020446034700',
                                                                                                'in_ucast_pkts': '11193504916',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '30656',
                                                                                                'out_octets': '871055710790',
                                                                                                'out_ucast_pkts': '4406433883',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30656',
                                                                       'out_octets': '871055710790',
                                                                       'out_ucast_pkts': '4406433883',
                                                                       'poller_hash': 'NDJDOEVBcGVlcmludGZwb24gODY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '869',
                                                                                               'links': [       {       'connected_time': '71474612',
                                                                                                                        'fiber_length': 17122,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOEVBcGVlcnBvbnBvbiA4Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5383 Perth Road 135 (Ryan & Mandy '
                                                 'Kuntze10421167)',
                                         'poller_hash': 'NDJDOEVB',
                                         'serial_number': '42C8EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bc:92',
                                         'uptime': '71474612'},
                                 {       'counter_source': '870/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '870/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '151263615691',
                                                                       'in_ucast_pkts': '387275792',
                                                                       'mac_address': 'd0:76:8f:08:14:7b',
                                                                       'name': '870/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '168',
                                                                       'out_octets': '1506341880046',
                                                                       'out_ucast_pkts': '1241717014',
                                                                       'poller_hash': 'ODY5Mzg1cGVlcmludGY4NzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '870',
                                                                       'in_nucast_pkts': '168',
                                                                       'in_octets': '1506341880046',
                                                                       'in_ucast_pkts': '1241717014',
                                                                       'links': [       {       'in_nucast_pkts': '168',
                                                                                                'in_octets': '1506341880046',
                                                                                                'in_ucast_pkts': '1241717014',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '151263615691',
                                                                                                'out_ucast_pkts': '387275792',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '151263615691',
                                                                       'out_ucast_pkts': '387275792',
                                                                       'poller_hash': 'ODY5Mzg1cGVlcmludGZwb24gODcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '870',
                                                                                               'links': [       {       'connected_time': '3113822',
                                                                                                                        'fiber_length': 17582,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5Mzg1cGVlcnBvbnBvbiA4NzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '208 Selby Av (Stephanie Priestap '
                                                 '10436138)',
                                         'poller_hash': 'ODY5Mzg1',
                                         'serial_number': '869385',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:14:7b',
                                         'uptime': '3113822'},
                                 {       'counter_source': '871/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b0:4e:26:7b:bf:a3'],
                                                                       'id': '871/g1',
                                                                       'in_nucast_pkts': '498',
                                                                       'in_octets': '2804335994',
                                                                       'in_ucast_pkts': '6324231',
                                                                       'mac_address': '44:65:7f:7b:c4:34',
                                                                       'mtu': 2000,
                                                                       'name': '871/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464380',
                                                                       'out_octets': '25140647804',
                                                                       'out_ucast_pkts': '20706447',
                                                                       'poller_hash': 'NUE2NjU3cGVlcmludGY4NzEvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '871',
                                                                       'in_nucast_pkts': '1464380',
                                                                       'in_octets': '25140647804',
                                                                       'in_ucast_pkts': '20706447',
                                                                       'links': [       {       'in_nucast_pkts': '1464380',
                                                                                                'in_octets': '25140647804',
                                                                                                'in_ucast_pkts': '20706447',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '498',
                                                                                                'out_octets': '2804335994',
                                                                                                'out_ucast_pkts': '6324231',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '498',
                                                                       'out_octets': '2804335994',
                                                                       'out_ucast_pkts': '6324231',
                                                                       'poller_hash': 'NUE2NjU3cGVlcmludGZwb24gODcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '871',
                                                                                               'links': [       {       'connected_time': '688494',
                                                                                                                        'fiber_length': 1303,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjU3cGVlcnBvbnBvbiA4NzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '6 Fulton St Milverton (P & M '
                                                 'Bueckert 10421013)',
                                         'poller_hash': 'NUE2NjU3',
                                         'serial_number': '5A6657',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:34',
                                         'uptime': '688494'},
                                 {       'counter_source': '872/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:32:1b:63:f0:6a'],
                                                                       'id': '872/g1',
                                                                       'in_nucast_pkts': '1681',
                                                                       'in_octets': '7628622379',
                                                                       'in_ucast_pkts': '45750066',
                                                                       'mac_address': '44:65:7f:7b:d6:ca',
                                                                       'mtu': 2000,
                                                                       'name': '872/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3685300',
                                                                       'out_octets': '238096183433',
                                                                       'out_ucast_pkts': '178897190',
                                                                       'poller_hash': 'NUE2Qzg5cGVlcmludGY4NzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '872',
                                                                       'in_nucast_pkts': '3685300',
                                                                       'in_octets': '238096183433',
                                                                       'in_ucast_pkts': '178897190',
                                                                       'links': [       {       'in_nucast_pkts': '3685300',
                                                                                                'in_octets': '238096183433',
                                                                                                'in_ucast_pkts': '178897190',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1681',
                                                                                                'out_octets': '7628622379',
                                                                                                'out_ucast_pkts': '45750066',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1681',
                                                                       'out_octets': '7628622379',
                                                                       'out_ucast_pkts': '45750066',
                                                                       'poller_hash': 'NUE2Qzg5cGVlcmludGZwb24gODcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '872',
                                                                                               'links': [       {       'connected_time': '1722157',
                                                                                                                        'fiber_length': 17525,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2Qzg5cGVlcnBvbnBvbiA4NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4575 Line 49 (Roosendale Theo)',
                                         'poller_hash': 'NUE2Qzg5',
                                         'serial_number': '5A6C89',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d6:ca',
                                         'uptime': '1722157'},
                                 {       'counter_source': '873/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '873/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1640869158223',
                                                                       'in_ucast_pkts': '281456994457',
                                                                       'mac_address': 'd0:76:8f:08:17:f3',
                                                                       'name': '873/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '137',
                                                                       'out_octets': '2285283885184',
                                                                       'out_ucast_pkts': '2698149548',
                                                                       'poller_hash': 'ODY5M0Y0cGVlcmludGY4NzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '873',
                                                                       'in_nucast_pkts': '137',
                                                                       'in_octets': '2285283885184',
                                                                       'in_ucast_pkts': '2698149548',
                                                                       'links': [       {       'in_nucast_pkts': '137',
                                                                                                'in_octets': '2285283885184',
                                                                                                'in_ucast_pkts': '2698149548',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1640869158223',
                                                                                                'out_ucast_pkts': '281456994457',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1640869158223',
                                                                       'out_ucast_pkts': '281456994457',
                                                                       'poller_hash': 'ODY5M0Y0cGVlcmludGZwb24gODcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '873',
                                                                                               'links': [       {       'connected_time': '73959560',
                                                                                                                        'fiber_length': 6307,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0Y0cGVlcnBvbnBvbiA4NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4731 PERTH LINE 55 (SLITS PEDRO & '
                                                 'JOLANDA)',
                                         'poller_hash': 'ODY5M0Y0',
                                         'serial_number': '8693F4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:f3',
                                         'uptime': '73959560'},
                                 {       'counter_source': '874/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:59:c0:bf:74:f6'],
                                                                       'id': '874/g1',
                                                                       'in_nucast_pkts': '835',
                                                                       'in_octets': '22464188834',
                                                                       'in_ucast_pkts': '203897243',
                                                                       'mac_address': 'cc:be:59:50:1b:56',
                                                                       'mtu': 2000,
                                                                       'name': '874/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4959035',
                                                                       'out_octets': '1361173533313',
                                                                       'out_ucast_pkts': '984511694',
                                                                       'poller_hash': 'NDNDNDJDcGVlcmludGY4NzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '874',
                                                                       'in_nucast_pkts': '4959035',
                                                                       'in_octets': '1361173533313',
                                                                       'in_ucast_pkts': '984511694',
                                                                       'links': [       {       'in_nucast_pkts': '4959035',
                                                                                                'in_octets': '1361173533313',
                                                                                                'in_ucast_pkts': '984511694',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '835',
                                                                                                'out_octets': '22464188834',
                                                                                                'out_ucast_pkts': '203897243',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '835',
                                                                       'out_octets': '22464188834',
                                                                       'out_ucast_pkts': '203897243',
                                                                       'poller_hash': 'NDNDNDJDcGVlcmludGZwb24gODc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '874',
                                                                                               'links': [       {       'connected_time': '2969198',
                                                                                                                        'fiber_length': 17691,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.444,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.052}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNDNDJDcGVlcnBvbnBvbiA4NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '153 Winstanley St. Monkton (Ben '
                                                 'Champion)',
                                         'poller_hash': 'NDNDNDJD',
                                         'serial_number': '43C42C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:1b:55',
                                         'uptime': '2969198'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '875/g1',
                                                                       'name': '875/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2QjZGcGVlcmludGY4NzUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '875',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2QjZGcGVlcmludGZwb24gODc1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '875',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjZGcGVlcnBvbnBvbiA4NzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '67 MAIN ST (BOOSE ERIC)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "875" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-24T08:29:33.970Z',
                                         'poller_hash': 'NUE2QjZG',
                                         'serial_number': '5A6B6F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '876/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '876/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '147990754967',
                                                                       'in_ucast_pkts': '134333307',
                                                                       'mac_address': 'd0:76:8f:08:16:6b',
                                                                       'name': '876/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '111289611350',
                                                                       'out_ucast_pkts': '176736797',
                                                                       'poller_hash': 'ODY5M0MzcGVlcmludGY4NzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '876',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '111289611350',
                                                                       'in_ucast_pkts': '176736797',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '111289611350',
                                                                                                'in_ucast_pkts': '176736797',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '147990754967',
                                                                                                'out_ucast_pkts': '134333307',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '147990754967',
                                                                       'out_ucast_pkts': '134333307',
                                                                       'poller_hash': 'ODY5M0MzcGVlcmludGZwb24gODc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '876',
                                                                                               'links': [       {       'connected_time': '519881',
                                                                                                                        'fiber_length': 5916,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0MzcGVlcnBvbnBvbiA4NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'Beside the Still Water',
                                         'poller_hash': 'ODY5M0Mz',
                                         'serial_number': '8693C3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:6b',
                                         'uptime': '519881'},
                                 {       'counter_source': '877/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f3:7a'],
                                                                       'id': '877/g1',
                                                                       'in_nucast_pkts': '11793',
                                                                       'in_octets': '8395939641',
                                                                       'in_ucast_pkts': '43455970',
                                                                       'mac_address': '60:db:98:4e:67:ef',
                                                                       'mtu': 2000,
                                                                       'name': '877/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434207',
                                                                       'out_octets': '162286012102',
                                                                       'out_ucast_pkts': '113690705',
                                                                       'poller_hash': 'QTlDMkJEcGVlcmludGY4NzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '877',
                                                                       'in_nucast_pkts': '1434207',
                                                                       'in_octets': '162286012102',
                                                                       'in_ucast_pkts': '113690705',
                                                                       'links': [       {       'in_nucast_pkts': '1434207',
                                                                                                'in_octets': '162286012102',
                                                                                                'in_ucast_pkts': '113690705',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11793',
                                                                                                'out_octets': '8395939641',
                                                                                                'out_ucast_pkts': '43455970',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11793',
                                                                       'out_octets': '8395939641',
                                                                       'out_ucast_pkts': '43455970',
                                                                       'poller_hash': 'QTlDMkJEcGVlcmludGZwb24gODc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '877',
                                                                                               'links': [       {       'connected_time': '688560',
                                                                                                                        'fiber_length': 5939,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkJEcGVlcnBvbnBvbiA4Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4676 Line 55 (MICHAEL GIBSON)',
                                         'poller_hash': 'QTlDMkJE',
                                         'serial_number': 'A9C2BD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:67:ec',
                                         'uptime': '688560'},
                                 {       'counter_source': '878/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f3:22'],
                                                                       'id': '878/g1',
                                                                       'in_nucast_pkts': '12476',
                                                                       'in_octets': '623141694',
                                                                       'in_ucast_pkts': '2421956',
                                                                       'mac_address': '48:77:46:f8:7b:6e',
                                                                       'mtu': 2000,
                                                                       'name': '878/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463860',
                                                                       'out_octets': '13508612138',
                                                                       'out_ucast_pkts': '10985466',
                                                                       'poller_hash': 'ODJBNzUwcGVlcmludGY4NzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '878',
                                                                       'in_nucast_pkts': '1463860',
                                                                       'in_octets': '13508612138',
                                                                       'in_ucast_pkts': '10985466',
                                                                       'links': [       {       'in_nucast_pkts': '1463860',
                                                                                                'in_octets': '13508612138',
                                                                                                'in_ucast_pkts': '10985466',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12476',
                                                                                                'out_octets': '623141694',
                                                                                                'out_ucast_pkts': '2421956',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12476',
                                                                       'out_octets': '623141694',
                                                                       'out_ucast_pkts': '2421956',
                                                                       'poller_hash': 'ODJBNzUwcGVlcmludGZwb24gODc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '878',
                                                                                               'links': [       {       'connected_time': '688459',
                                                                                                                        'fiber_length': 1506,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNzUwcGVlcnBvbnBvbiA4Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '88 MILL ST E (GASCHO BRIAN & LIZ)',
                                         'poller_hash': 'ODJBNzUw',
                                         'serial_number': '82A750',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:7b:6e',
                                         'uptime': '688459'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '879/g1',
                                                                       'name': '879/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QTlDOTg2cGVlcmludGY4NzkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '879',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QTlDOTg2cGVlcmludGZwb24gODc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '879',
                                                                                               'links': [       {       'connected_time': '688450',
                                                                                                                        'fiber_length': 291,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOTg2cGVlcnBvbnBvbiA4Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 Mill St. E. Milverton (Tech '
                                                 'Office Basement)',
                                         'poller_hash': 'QTlDOTg2',
                                         'serial_number': 'A9C986',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:9e:34',
                                         'uptime': '688450'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '886/g1',
                                                                       'name': '886/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzU3RkFwZWVyaW50Zjg4Ni9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '886',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzU3RkFwZWVyaW50ZnBvbiA4ODY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '886',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU3RkFwZWVycG9ucG9uIDg4Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '7 Pugh St. Milverton (Shawn Gilbert)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "886" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-24T08:29:33.920Z',
                                         'poller_hash': 'QzU3RkE=',
                                         'serial_number': 'C57FA',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '893/g1',
                                                                       'name': '893/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzRGRTlwZWVyaW50Zjg5My9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '893',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzRGRTlwZWVyaW50ZnBvbiA4OTM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '893',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzRGRTlwZWVycG9ucG9uIDg5Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '14 Pugh St. Milverton (COOK GERALD)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "893" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-24T08:29:33.870Z',
                                         'poller_hash': 'QzRGRTk=',
                                         'serial_number': 'C4FE9',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '897/g1',
                                                                       'name': '897/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5M0NDcGVlcmludGY4OTcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '897',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5M0NDcGVlcmludGZwb24gODk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '897',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0NDcGVlcnBvbnBvbiA4OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '7 Cruikshank St. Milverton',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "897" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-24T08:29:33.820Z',
                                         'poller_hash': 'ODY5M0ND',
                                         'serial_number': '8693CC',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '900/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:37:e7'],
                                                                       'id': '900/g1',
                                                                       'in_nucast_pkts': '14985',
                                                                       'in_octets': '21025590410',
                                                                       'in_ucast_pkts': '86679321',
                                                                       'mac_address': 'd0:76:8f:08:1c:b6',
                                                                       'mtu': 2000,
                                                                       'name': '900/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447131',
                                                                       'out_octets': '336479783739',
                                                                       'out_ucast_pkts': '251133105',
                                                                       'poller_hash': 'ODY5NDhDcGVlcmludGY5MDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '900',
                                                                       'in_nucast_pkts': '1447131',
                                                                       'in_octets': '336479783739',
                                                                       'in_ucast_pkts': '251133105',
                                                                       'links': [       {       'in_nucast_pkts': '1447131',
                                                                                                'in_octets': '336479783739',
                                                                                                'in_ucast_pkts': '251133105',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14985',
                                                                                                'out_octets': '21025590410',
                                                                                                'out_ucast_pkts': '86679321',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14985',
                                                                       'out_octets': '21025590410',
                                                                       'out_ucast_pkts': '86679321',
                                                                       'poller_hash': 'ODY5NDhDcGVlcmludGZwb24gOTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '900',
                                                                                               'links': [       {       'connected_time': '688410',
                                                                                                                        'fiber_length': 1997,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDhDcGVlcnBvbnBvbiA5MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '93 Reagan St Milverton (VANT VOORT M '
                                                 '& C)',
                                         'poller_hash': 'ODY5NDhD',
                                         'serial_number': '86948C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1c:b3',
                                         'uptime': '688410'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '902/g1',
                                                                       'name': '902/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzQ4MEREcGVlcmludGY5MDIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '902',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzQ4MEREcGVlcmludGZwb24gOTAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '902',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ4MEREcGVlcnBvbnBvbiA5MDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '55 WILLIAM ST APT 1',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "902" '
                                                                  '(location: GPON '
                                                                  'port "7/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:38.410Z',
                                         'poller_hash': 'MzQ4MERE',
                                         'serial_number': '3480DD',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '904/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5a:69'],
                                                                       'id': '904/g1',
                                                                       'in_nucast_pkts': '14857',
                                                                       'in_octets': '13758618998',
                                                                       'in_ucast_pkts': '84155543',
                                                                       'mac_address': 'd0:76:8f:08:14:a6',
                                                                       'mtu': 2000,
                                                                       'name': '904/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434248',
                                                                       'out_octets': '288668834521',
                                                                       'out_ucast_pkts': '200047722',
                                                                       'poller_hash': 'ODY5MzhBcGVlcmludGY5MDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '904',
                                                                       'in_nucast_pkts': '1434248',
                                                                       'in_octets': '288668834521',
                                                                       'in_ucast_pkts': '200047722',
                                                                       'links': [       {       'in_nucast_pkts': '1434248',
                                                                                                'in_octets': '288668834521',
                                                                                                'in_ucast_pkts': '200047722',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14857',
                                                                                                'out_octets': '13758618998',
                                                                                                'out_ucast_pkts': '84155543',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14857',
                                                                       'out_octets': '13758618998',
                                                                       'out_ucast_pkts': '84155543',
                                                                       'poller_hash': 'ODY5MzhBcGVlcmludGZwb24gOTA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '904',
                                                                                               'links': [       {       'connected_time': '688552',
                                                                                                                        'fiber_length': 1745,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzhBcGVlcnBvbnBvbiA5MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 MURRAY CT (ARMSTRONG SAMANTHA)',
                                         'poller_hash': 'ODY5MzhB',
                                         'serial_number': '86938A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:14:a3',
                                         'uptime': '688552'},
                                 {       'counter_source': '905/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f4:92:bf:8d:a7:f8'],
                                                                       'id': '905/g1',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '470604128300',
                                                                       'in_ucast_pkts': '486686968',
                                                                       'mac_address': 'd0:76:8f:1b:a1:ae',
                                                                       'mtu': 2000,
                                                                       'name': '905/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464543',
                                                                       'out_octets': '421370797828',
                                                                       'out_ucast_pkts': '434516754',
                                                                       'poller_hash': 'ODcyNDgzcGVlcmludGY5MDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '905',
                                                                       'in_nucast_pkts': '1464543',
                                                                       'in_octets': '421370797828',
                                                                       'in_ucast_pkts': '434516754',
                                                                       'links': [       {       'in_nucast_pkts': '1464543',
                                                                                                'in_octets': '421370797828',
                                                                                                'in_ucast_pkts': '434516754',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19',
                                                                                                'out_octets': '470604128300',
                                                                                                'out_ucast_pkts': '486686968',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19',
                                                                       'out_octets': '470604128300',
                                                                       'out_ucast_pkts': '486686968',
                                                                       'poller_hash': 'ODcyNDgzcGVlcmludGZwb24gOTA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '905',
                                                                                               'links': [       {       'connected_time': '688368',
                                                                                                                        'fiber_length': 2355,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDgzcGVlcnBvbnBvbiA5MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'PERCON NEW SHOP 10434163 (4679 LINE '
                                                 '61)',
                                         'poller_hash': 'ODcyNDgz',
                                         'serial_number': '872483',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a1:ab',
                                         'uptime': '688368'},
                                 {       'counter_source': '906/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '906/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '53558053169',
                                                                       'in_ucast_pkts': '4485374005',
                                                                       'mac_address': 'd0:76:8f:1b:a5:2b',
                                                                       'name': '906/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '733332728608',
                                                                       'out_ucast_pkts': '633870509',
                                                                       'poller_hash': 'ODcyNEYzcGVlcmludGY5MDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '906',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '733332728608',
                                                                       'in_ucast_pkts': '633870509',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '733332728608',
                                                                                                'in_ucast_pkts': '633870509',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '53558053169',
                                                                                                'out_ucast_pkts': '4485374005',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '53558053169',
                                                                       'out_ucast_pkts': '4485374005',
                                                                       'poller_hash': 'ODcyNEYzcGVlcmludGZwb24gOTA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '906',
                                                                                               'links': [       {       'connected_time': '22383405',
                                                                                                                        'fiber_length': 7935,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNEYzcGVlcnBvbnBvbiA5MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6168 RD 124 MILVERTON (SCHULTZ '
                                                 'LONNIE)',
                                         'poller_hash': 'ODcyNEYz',
                                         'serial_number': '8724F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:2b',
                                         'uptime': '22383405'},
                                 {       'counter_source': '908/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '908/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '373492844701',
                                                                       'in_ucast_pkts': '26779468569',
                                                                       'mac_address': 'd0:76:8f:aa:49:94',
                                                                       'name': '908/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '4816523333672',
                                                                       'out_ucast_pkts': '3534687298',
                                                                       'poller_hash': 'OTU2RTU4cGVlcmludGY5MDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '908',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '4816523333672',
                                                                       'in_ucast_pkts': '3534687298',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '4816523333672',
                                                                                                'in_ucast_pkts': '3534687298',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '373492844701',
                                                                                                'out_ucast_pkts': '26779468569',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '373492844701',
                                                                       'out_ucast_pkts': '26779468569',
                                                                       'poller_hash': 'OTU2RTU4cGVlcmludGZwb24gOTA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '908',
                                                                                               'links': [       {       'connected_time': '538562',
                                                                                                                        'fiber_length': 16316,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTU4cGVlcnBvbnBvbiA5MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5095 PERTH RD 135 ROSTOCK (DEBUS '
                                                 'JOSH)',
                                         'poller_hash': 'OTU2RTU4',
                                         'serial_number': '956E58',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:49:94',
                                         'uptime': '538562'},
                                 {       'counter_source': '910/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'dc:62:79:c7:74:71'],
                                                                       'id': '910/g1',
                                                                       'in_nucast_pkts': '20569',
                                                                       'in_octets': '197253958927',
                                                                       'in_ucast_pkts': '606034054',
                                                                       'mac_address': '60:db:98:8f:04:45',
                                                                       'mtu': 2000,
                                                                       'name': '910/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32346044',
                                                                       'out_octets': '1830384241804',
                                                                       'out_ucast_pkts': '1545447936',
                                                                       'poller_hash': 'QjUxMkRCcGVlcmludGY5MTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '910',
                                                                       'in_nucast_pkts': '32346044',
                                                                       'in_octets': '1830384241804',
                                                                       'in_ucast_pkts': '1545447936',
                                                                       'links': [       {       'in_nucast_pkts': '32346044',
                                                                                                'in_octets': '1830384241804',
                                                                                                'in_ucast_pkts': '1545447936',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20569',
                                                                                                'out_octets': '197253958927',
                                                                                                'out_ucast_pkts': '606034054',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20569',
                                                                       'out_octets': '197253958927',
                                                                       'out_ucast_pkts': '606034054',
                                                                       'poller_hash': 'QjUxMkRCcGVlcmludGZwb24gOTEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '910',
                                                                                               'links': [       {       'connected_time': '14251264',
                                                                                                                        'fiber_length': 10790,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMkRCcGVlcnBvbnBvbiA5MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5956 PERTH RD 119-5 (ELROY WAGLER '
                                                 'MASONRY)',
                                         'poller_hash': 'QjUxMkRC',
                                         'serial_number': 'B512DB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:04:45',
                                         'uptime': '14251264'},
                                 {       'counter_source': '114781/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7c:02'],
                                                                       'id': '114781/g1',
                                                                       'in_nucast_pkts': '51424',
                                                                       'in_octets': '33340582090',
                                                                       'in_ucast_pkts': '136218338',
                                                                       'mac_address': '04:bc:9f:5c:1f:04',
                                                                       'mtu': 2000,
                                                                       'name': '114781/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6407370',
                                                                       'out_octets': '980454595075',
                                                                       'out_ucast_pkts': '696829865',
                                                                       'poller_hash': 'MTI3NEE2MHBlZXJpbnRmMTE0NzgxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '114781',
                                                                       'in_nucast_pkts': '6407370',
                                                                       'in_octets': '980454595075',
                                                                       'in_ucast_pkts': '696829865',
                                                                       'links': [       {       'in_nucast_pkts': '6407370',
                                                                                                'in_octets': '980454595075',
                                                                                                'in_ucast_pkts': '696829865',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '51424',
                                                                                                'out_octets': '33340582090',
                                                                                                'out_ucast_pkts': '136218338',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '51424',
                                                                       'out_octets': '33340582090',
                                                                       'out_ucast_pkts': '136218338',
                                                                       'poller_hash': 'MTI3NEE2MHBlZXJpbnRmcG9uIDExNDc4MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '114781',
                                                                                               'links': [       {       'connected_time': '2969414',
                                                                                                                        'fiber_length': 17768,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE2MHBlZXJwb25wb24gMTE0Nzgx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '170 MARKET ST (REITZEL ETHAN)',
                                         'poller_hash': 'MTI3NEE2MA==',
                                         'serial_number': '1274A60',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:01',
                                         'uptime': '2969414'},
                                 {       'counter_source': '137520/F1+137520/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:2e:38'],
                                                                       'id': '137520/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2303100302',
                                                                       'in_ucast_pkts': '13950349',
                                                                       'name': '137520/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '22069755669',
                                                                       'out_octets': '29525545679520',
                                                                       'out_ucast_pkts': '9369307',
                                                                       'poller_hash': 'NDJCRjA4cGVlcmludGYxMzc1MjAvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:56:81:c5:c5:ba'],
                                                                       'id': '137520/g1',
                                                                       'in_nucast_pkts': '21537',
                                                                       'in_octets': '333793699110',
                                                                       'in_ucast_pkts': '848360807',
                                                                       'mac_address': 'cc:be:59:cb:6d:85',
                                                                       'mtu': 2000,
                                                                       'name': '137520/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '102845115',
                                                                       'out_octets': '4229203663554',
                                                                       'out_ucast_pkts': '3158286884',
                                                                       'poller_hash': 'NDJCRjA4cGVlcmludGYxMzc1MjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137520',
                                                                       'in_nucast_pkts': '22172600784',
                                                                       'in_octets': '33754749343074',
                                                                       'in_ucast_pkts': '3167656191',
                                                                       'links': [       {       'in_nucast_pkts': '22172600784',
                                                                                                'in_octets': '33754749343074',
                                                                                                'in_ucast_pkts': '3167656191',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21537',
                                                                                                'out_octets': '336096799412',
                                                                                                'out_ucast_pkts': '862311156',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21537',
                                                                       'out_octets': '336096799412',
                                                                       'out_ucast_pkts': '862311156',
                                                                       'poller_hash': 'NDJCRjA4cGVlcmludGZwb24gMTM3NTIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137520',
                                                                                               'links': [       {       'connected_time': '6572792',
                                                                                                                        'fiber_length': 12379,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjA4cGVlcnBvbnBvbiAxMzc1MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4900 LINE 52 (DEGROOT JIM & WENDY)',
                                         'poller_hash': 'NDJCRjA4',
                                         'serial_number': '42BF08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6d:82',
                                         'uptime': '6572792'},
                                 {       'counter_source': '137693/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f4:f0'],
                                                                       'id': '137693/g1',
                                                                       'in_nucast_pkts': '697167',
                                                                       'in_octets': '641987478049',
                                                                       'in_ucast_pkts': '1082782708',
                                                                       'mac_address': '44:65:7f:7b:d7:06',
                                                                       'mtu': 2000,
                                                                       'name': '137693/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83570468',
                                                                       'out_octets': '3731133087841',
                                                                       'out_ucast_pkts': '2955119879',
                                                                       'poller_hash': 'NUE2QzlEcGVlcmludGYxMzc2OTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137693',
                                                                       'in_nucast_pkts': '83570468',
                                                                       'in_octets': '3731133087841',
                                                                       'in_ucast_pkts': '2955119879',
                                                                       'links': [       {       'in_nucast_pkts': '83570468',
                                                                                                'in_octets': '3731133087841',
                                                                                                'in_ucast_pkts': '2955119879',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '697167',
                                                                                                'out_octets': '641987478049',
                                                                                                'out_ucast_pkts': '1082782708',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '697167',
                                                                       'out_octets': '641987478049',
                                                                       'out_ucast_pkts': '1082782708',
                                                                       'poller_hash': 'NUE2QzlEcGVlcmludGZwb24gMTM3Njkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137693',
                                                                                               'links': [       {       'connected_time': '67790242',
                                                                                                                        'fiber_length': 1210,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzlEcGVlcnBvbnBvbiAxMzc2OTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '24 MILL ST E (WEBER JUSTIN)',
                                         'poller_hash': 'NUE2QzlE',
                                         'serial_number': '5A6C9D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d7:06',
                                         'uptime': '67790242'},
                                 {       'counter_source': '137697/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:50:9a'],
                                                                       'id': '137697/g1',
                                                                       'in_nucast_pkts': '14069',
                                                                       'in_octets': '35272410729',
                                                                       'in_ucast_pkts': '149430233',
                                                                       'mac_address': '44:65:7f:7b:d5:da',
                                                                       'mtu': 2000,
                                                                       'name': '137697/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '85888364',
                                                                       'out_octets': '611268189356',
                                                                       'out_ucast_pkts': '503434866',
                                                                       'poller_hash': 'NUE2QzM5cGVlcmludGYxMzc2OTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137697',
                                                                       'in_nucast_pkts': '85888364',
                                                                       'in_octets': '611268189356',
                                                                       'in_ucast_pkts': '503434866',
                                                                       'links': [       {       'in_nucast_pkts': '85888364',
                                                                                                'in_octets': '611268189356',
                                                                                                'in_ucast_pkts': '503434866',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14069',
                                                                                                'out_octets': '35272410729',
                                                                                                'out_ucast_pkts': '149430233',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14069',
                                                                       'out_octets': '35272410729',
                                                                       'out_ucast_pkts': '149430233',
                                                                       'poller_hash': 'NUE2QzM5cGVlcmludGZwb24gMTM3Njk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137697',
                                                                                               'links': [       {       'connected_time': '67790150',
                                                                                                                        'fiber_length': 1188,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzM5cGVlcnBvbnBvbiAxMzc2OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '22 WEST ST (STECKLY STEVE)',
                                         'poller_hash': 'NUE2QzM5',
                                         'serial_number': '5A6C39',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d5:da',
                                         'uptime': '67790150'},
                                 {       'counter_source': '137724/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:81:fa:a4'],
                                                                       'id': '137724/g1',
                                                                       'in_nucast_pkts': '45372',
                                                                       'in_octets': '19158565519',
                                                                       'in_ucast_pkts': '91256901',
                                                                       'mac_address': 'cc:be:59:cb:6b:4d',
                                                                       'mtu': 2000,
                                                                       'name': '137724/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5117448',
                                                                       'out_octets': '547721214775',
                                                                       'out_ucast_pkts': '397141776',
                                                                       'poller_hash': 'NDJCRUMxcGVlcmludGYxMzc3MjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137724',
                                                                       'in_nucast_pkts': '5117448',
                                                                       'in_octets': '547721214775',
                                                                       'in_ucast_pkts': '397141776',
                                                                       'links': [       {       'in_nucast_pkts': '5117448',
                                                                                                'in_octets': '547721214775',
                                                                                                'in_ucast_pkts': '397141776',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '45372',
                                                                                                'out_octets': '19158565519',
                                                                                                'out_ucast_pkts': '91256901',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45372',
                                                                       'out_octets': '19158565519',
                                                                       'out_ucast_pkts': '91256901',
                                                                       'poller_hash': 'NDJCRUMxcGVlcmludGZwb24gMTM3NzI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137724',
                                                                                               'links': [       {       'connected_time': '2356187',
                                                                                                                        'fiber_length': 1290,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRUMxcGVlcnBvbnBvbiAxMzc3MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '47 MAIN ST N (WOLF JONATHAN)',
                                         'poller_hash': 'NDJCRUMx',
                                         'serial_number': '42BEC1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6b:4a',
                                         'uptime': '2356187'},
                                 {       'counter_source': '137813/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:f5:a2:8d:fb:8e'],
                                                                       'id': '137813/g1',
                                                                       'in_nucast_pkts': '1464023',
                                                                       'in_octets': '146829894050',
                                                                       'in_ucast_pkts': '467800737',
                                                                       'mac_address': '44:65:7f:7b:d5:17',
                                                                       'mtu': 2000,
                                                                       'name': '137813/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '85951116',
                                                                       'out_octets': '1235358423734',
                                                                       'out_ucast_pkts': '899020712',
                                                                       'poller_hash': 'NUE2QkY4cGVlcmludGYxMzc4MTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137813',
                                                                       'in_nucast_pkts': '85951116',
                                                                       'in_octets': '1235358423734',
                                                                       'in_ucast_pkts': '899020712',
                                                                       'links': [       {       'in_nucast_pkts': '85951116',
                                                                                                'in_octets': '1235358423734',
                                                                                                'in_ucast_pkts': '899020712',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1464023',
                                                                                                'out_octets': '146829894050',
                                                                                                'out_ucast_pkts': '467800737',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464023',
                                                                       'out_octets': '146829894050',
                                                                       'out_ucast_pkts': '467800737',
                                                                       'poller_hash': 'NUE2QkY4cGVlcmludGZwb24gMTM3ODEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137813',
                                                                                               'links': [       {       'connected_time': '67789945',
                                                                                                                        'fiber_length': 1118,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QkY4cGVlcnBvbnBvbiAxMzc4MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '14 WEST ST (SALZMAN BRUCE)',
                                         'poller_hash': 'NUE2QkY4',
                                         'serial_number': '5A6BF8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d5:17',
                                         'uptime': '67789945'},
                                 {       'counter_source': '137828/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3a:e9'],
                                                                       'id': '137828/g1',
                                                                       'in_nucast_pkts': '33514',
                                                                       'in_octets': '20863308227',
                                                                       'in_ucast_pkts': '67338159',
                                                                       'mac_address': '44:65:7f:7b:d0:4f',
                                                                       'mtu': 2000,
                                                                       'name': '137828/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3582904',
                                                                       'out_octets': '258593828678',
                                                                       'out_ucast_pkts': '212996755',
                                                                       'poller_hash': 'NUE2QTYwcGVlcmludGYxMzc4MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137828',
                                                                       'in_nucast_pkts': '3582904',
                                                                       'in_octets': '258593828678',
                                                                       'in_ucast_pkts': '212996755',
                                                                       'links': [       {       'in_nucast_pkts': '3582904',
                                                                                                'in_octets': '258593828678',
                                                                                                'in_ucast_pkts': '212996755',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '33514',
                                                                                                'out_octets': '20863308227',
                                                                                                'out_ucast_pkts': '67338159',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33514',
                                                                       'out_octets': '20863308227',
                                                                       'out_ucast_pkts': '67338159',
                                                                       'poller_hash': 'NUE2QTYwcGVlcmludGZwb24gMTM3ODI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137828',
                                                                                               'links': [       {       'connected_time': '2351694',
                                                                                                                        'fiber_length': 12566,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QTYwcGVlcnBvbnBvbiAxMzc4Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5735 PERTH ROAD 119 (NAFZIGER JAMES)',
                                         'poller_hash': 'NUE2QTYw',
                                         'serial_number': '5A6A60',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d0:4f',
                                         'uptime': '2351694'},
                                 {       'counter_source': '137832/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '98:da:c4:30:8b:40'],
                                                                       'id': '137832/g1',
                                                                       'in_nucast_pkts': '4760',
                                                                       'in_octets': '13090094586',
                                                                       'in_ucast_pkts': '39974737',
                                                                       'mac_address': '44:65:7f:7b:d6:67',
                                                                       'mtu': 2000,
                                                                       'name': '137832/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6685478',
                                                                       'out_octets': '157648239002',
                                                                       'out_ucast_pkts': '127026517',
                                                                       'poller_hash': 'NUE2QzY4cGVlcmludGYxMzc4MzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137832',
                                                                       'in_nucast_pkts': '6685478',
                                                                       'in_octets': '157648239002',
                                                                       'in_ucast_pkts': '127026517',
                                                                       'links': [       {       'in_nucast_pkts': '6685478',
                                                                                                'in_octets': '157648239002',
                                                                                                'in_ucast_pkts': '127026517',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4760',
                                                                                                'out_octets': '13090094586',
                                                                                                'out_ucast_pkts': '39974737',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4760',
                                                                       'out_octets': '13090094586',
                                                                       'out_ucast_pkts': '39974737',
                                                                       'poller_hash': 'NUE2QzY4cGVlcmludGZwb24gMTM3ODMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137832',
                                                                                               'links': [       {       'connected_time': '3113920',
                                                                                                                        'fiber_length': 11783,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzY4cGVlcnBvbnBvbiAxMzc4MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '6515 ROAD 154 (ERB MARK & RACHEL)',
                                         'poller_hash': 'NUE2QzY4',
                                         'serial_number': '5A6C68',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d6:67',
                                         'uptime': '3113920'},
                                 {       'counter_source': '138170/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '138170/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '110156565376',
                                                                       'in_ucast_pkts': '504590052',
                                                                       'mac_address': 'd0:76:8f:1b:a6:03',
                                                                       'name': '138170/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2486409596970',
                                                                       'out_ucast_pkts': '1803377052',
                                                                       'poller_hash': 'ODcyNTBFcGVlcmludGYxMzgxNzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138170',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2486409596970',
                                                                       'in_ucast_pkts': '1803377052',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2486409596970',
                                                                                                'in_ucast_pkts': '1803377052',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '110156565376',
                                                                                                'out_ucast_pkts': '504590052',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '110156565376',
                                                                       'out_ucast_pkts': '504590052',
                                                                       'poller_hash': 'ODcyNTBFcGVlcmludGZwb24gMTM4MTcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138170',
                                                                                               'links': [       {       'connected_time': '688537',
                                                                                                                        'fiber_length': 7113,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTBFcGVlcnBvbnBvbiAxMzgxNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6283 ROAD 124 (BAUMAN GREG)',
                                         'poller_hash': 'ODcyNTBF',
                                         'serial_number': '87250E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a6:03',
                                         'uptime': '688537'},
                                 {       'counter_source': '138182/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:35:e3'],
                                                                       'id': '138182/g1',
                                                                       'in_nucast_pkts': '22',
                                                                       'in_octets': '1958904336',
                                                                       'in_ucast_pkts': '12158949',
                                                                       'mac_address': '44:65:7f:33:21:07',
                                                                       'mtu': 2000,
                                                                       'name': '138182/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102962',
                                                                       'out_octets': '77065451899',
                                                                       'out_ucast_pkts': '56667490',
                                                                       'poller_hash': 'NjQwQUZCcGVlcmludGYxMzgxODIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138182',
                                                                       'in_nucast_pkts': '1102962',
                                                                       'in_octets': '77065451899',
                                                                       'in_ucast_pkts': '56667490',
                                                                       'links': [       {       'in_nucast_pkts': '1102962',
                                                                                                'in_octets': '77065451899',
                                                                                                'in_ucast_pkts': '56667490',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22',
                                                                                                'out_octets': '1958904336',
                                                                                                'out_ucast_pkts': '12158949',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22',
                                                                       'out_octets': '1958904336',
                                                                       'out_ucast_pkts': '12158949',
                                                                       'poller_hash': 'NjQwQUZCcGVlcmludGZwb24gMTM4MTgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138182',
                                                                                               'links': [       {       'connected_time': '688406',
                                                                                                                        'fiber_length': 1862,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.626,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.232}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUZCcGVlcnBvbnBvbiAxMzgxODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '94 REAGAN ST (DEWAR MELANIE)',
                                         'poller_hash': 'NjQwQUZC',
                                         'serial_number': '640AFB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:06',
                                         'uptime': '688406'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '138228/g1',
                                                                       'name': '138228/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDNEE5cGVlcmludGYxMzgyMjgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138228',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDNEE5cGVlcmludGZwb24gMTM4MjI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138228',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNEE5cGVlcnBvbnBvbiAxMzgyMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '199 WINSTANLEY ST (WHETSTONE JOSH)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "138228" '
                                                                  '(location: GPON '
                                                                  'port "4/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-24T08:32:13.440Z',
                                         'poller_hash': 'NDJDNEE5',
                                         'serial_number': '42C4A9',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '138337/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '20:23:51:49:b5:a9'],
                                                                       'id': '138337/g1',
                                                                       'in_nucast_pkts': '264',
                                                                       'in_octets': '11514896789',
                                                                       'in_ucast_pkts': '54254777',
                                                                       'mac_address': 'cc:be:59:50:02:82',
                                                                       'mtu': 2000,
                                                                       'name': '138337/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1102861',
                                                                       'out_octets': '349605044589',
                                                                       'out_ucast_pkts': '247297419',
                                                                       'poller_hash': 'NDNCNTA4cGVlcmludGYxMzgzMzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138337',
                                                                       'in_nucast_pkts': '1102861',
                                                                       'in_octets': '349605044589',
                                                                       'in_ucast_pkts': '247297419',
                                                                       'links': [       {       'in_nucast_pkts': '1102861',
                                                                                                'in_octets': '349605044589',
                                                                                                'in_ucast_pkts': '247297419',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '264',
                                                                                                'out_octets': '11514896789',
                                                                                                'out_ucast_pkts': '54254777',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '264',
                                                                       'out_octets': '11514896789',
                                                                       'out_ucast_pkts': '54254777',
                                                                       'poller_hash': 'NDNCNTA4cGVlcmludGZwb24gMTM4MzM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138337',
                                                                                               'links': [       {       'connected_time': '688395',
                                                                                                                        'fiber_length': 2081,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.562,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.436}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNTA4cGVlcnBvbnBvbiAxMzgzMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '73 REAGAN ST (MILLER PAUL)',
                                         'poller_hash': 'NDNCNTA4',
                                         'serial_number': '43B508',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:02:81',
                                         'uptime': '688395'},
                                 {       'counter_source': '138635/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f4:da'],
                                                                       'id': '138635/g1',
                                                                       'in_nucast_pkts': '12084',
                                                                       'in_octets': '9369041341',
                                                                       'in_ucast_pkts': '59598331',
                                                                       'mac_address': 'd0:76:8f:08:17:76',
                                                                       'mtu': 2000,
                                                                       'name': '138635/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447136',
                                                                       'out_octets': '262513427262',
                                                                       'out_ucast_pkts': '191684175',
                                                                       'poller_hash': 'ODY5M0U0cGVlcmludGYxMzg2MzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138635',
                                                                       'in_nucast_pkts': '1447136',
                                                                       'in_octets': '262513427262',
                                                                       'in_ucast_pkts': '191684175',
                                                                       'links': [       {       'in_nucast_pkts': '1447136',
                                                                                                'in_octets': '262513427262',
                                                                                                'in_ucast_pkts': '191684175',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12084',
                                                                                                'out_octets': '9369041341',
                                                                                                'out_ucast_pkts': '59598331',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12084',
                                                                       'out_octets': '9369041341',
                                                                       'out_ucast_pkts': '59598331',
                                                                       'poller_hash': 'ODY5M0U0cGVlcmludGZwb24gMTM4NjM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138635',
                                                                                               'links': [       {       'connected_time': '688411',
                                                                                                                        'fiber_length': 2139,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0U0cGVlcnBvbnBvbiAxMzg2MzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '75 REAGAN ST (MILLER PATRICK)',
                                         'poller_hash': 'ODY5M0U0',
                                         'serial_number': '8693E4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:73',
                                         'uptime': '688411'},
                                 {       'counter_source': '138672/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:63:7d:4f:07:30'],
                                                                       'id': '138672/g1',
                                                                       'in_nucast_pkts': '1910',
                                                                       'in_octets': '1247863904686',
                                                                       'in_ucast_pkts': '6143498698',
                                                                       'mac_address': '44:65:7f:7b:d6:cd',
                                                                       'mtu': 2000,
                                                                       'name': '138672/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76640852',
                                                                       'out_octets': '21625634624962',
                                                                       'out_ucast_pkts': '15250567380',
                                                                       'poller_hash': 'NUE2QzhBcGVlcmludGYxMzg2NzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138672',
                                                                       'in_nucast_pkts': '76640852',
                                                                       'in_octets': '21625634624962',
                                                                       'in_ucast_pkts': '15250567380',
                                                                       'links': [       {       'in_nucast_pkts': '76640852',
                                                                                                'in_octets': '21625634624962',
                                                                                                'in_ucast_pkts': '15250567380',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1910',
                                                                                                'out_octets': '1247863904686',
                                                                                                'out_ucast_pkts': '6143498698',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1910',
                                                                       'out_octets': '1247863904686',
                                                                       'out_ucast_pkts': '6143498698',
                                                                       'poller_hash': 'NUE2QzhBcGVlcmludGZwb24gMTM4Njcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138672',
                                                                                               'links': [       {       'connected_time': '67790303',
                                                                                                                        'fiber_length': 11362,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzhBcGVlcnBvbnBvbiAxMzg2NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5357 ROAD 140 (ROOBROECK JEREMY)',
                                         'poller_hash': 'NUE2QzhB',
                                         'serial_number': '5A6C8A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d6:cd',
                                         'uptime': '67790303'},
                                 {       'counter_source': '138698/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:ed:c9'],
                                                                       'id': '138698/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '2369014953',
                                                                       'in_ucast_pkts': '10390361',
                                                                       'mac_address': 'd0:76:8f:08:18:3e',
                                                                       'mtu': 2000,
                                                                       'name': '138698/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447215',
                                                                       'out_octets': '63163760656',
                                                                       'out_ucast_pkts': '46411837',
                                                                       'poller_hash': 'ODY5M0ZEcGVlcmludGYxMzg2OTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138698',
                                                                       'in_nucast_pkts': '1447215',
                                                                       'in_octets': '63163760656',
                                                                       'in_ucast_pkts': '46411837',
                                                                       'links': [       {       'in_nucast_pkts': '1447215',
                                                                                                'in_octets': '63163760656',
                                                                                                'in_ucast_pkts': '46411837',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '2369014953',
                                                                                                'out_ucast_pkts': '10390361',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '2369014953',
                                                                       'out_ucast_pkts': '10390361',
                                                                       'poller_hash': 'ODY5M0ZEcGVlcmludGZwb24gMTM4Njk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138698',
                                                                                               'links': [       {       'connected_time': '688418',
                                                                                                                        'fiber_length': 2125,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0ZEcGVlcnBvbnBvbiAxMzg2OTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '81 GERBER DR (MARSHALL GLEN)',
                                         'poller_hash': 'ODY5M0ZE',
                                         'serial_number': '8693FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:3b',
                                         'uptime': '688418'},
                                 {       'counter_source': '138781/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '138781/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1026664531021',
                                                                       'in_ucast_pkts': '78750957416',
                                                                       'mac_address': 'd0:76:8f:08:18:1b',
                                                                       'name': '138781/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '90',
                                                                       'out_octets': '2498798746062',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODY5M0Y5cGVlcmludGYxMzg3ODEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138781',
                                                                       'in_nucast_pkts': '90',
                                                                       'in_octets': '2498798746062',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '90',
                                                                                                'in_octets': '2498798746062',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1026664531021',
                                                                                                'out_ucast_pkts': '78750957416',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1026664531021',
                                                                       'out_ucast_pkts': '78750957416',
                                                                       'poller_hash': 'ODY5M0Y5cGVlcmludGZwb24gMTM4Nzgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138781',
                                                                                               'links': [       {       'connected_time': '21816411',
                                                                                                                        'fiber_length': 6396,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0Y5cGVlcnBvbnBvbiAxMzg3ODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4728 PERTH LINE 55 (KUNTZE HOWARD)',
                                         'poller_hash': 'ODY5M0Y5',
                                         'serial_number': '8693F9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:1b',
                                         'uptime': '21816411'},
                                 {       'counter_source': '139172/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '34:97:f6:a7:1a:ec'],
                                                                       'id': '139172/g1',
                                                                       'in_nucast_pkts': '906',
                                                                       'in_octets': '6933213506',
                                                                       'in_ucast_pkts': '23234245',
                                                                       'mac_address': '44:65:7f:7b:ce:8a',
                                                                       'mtu': 2000,
                                                                       'name': '139172/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434228',
                                                                       'out_octets': '119476391974',
                                                                       'out_ucast_pkts': '83683939',
                                                                       'poller_hash': 'NUE2OUM5cGVlcmludGYxMzkxNzIvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139172',
                                                                       'in_nucast_pkts': '1434228',
                                                                       'in_octets': '119476391974',
                                                                       'in_ucast_pkts': '83683939',
                                                                       'links': [       {       'in_nucast_pkts': '1434228',
                                                                                                'in_octets': '119476391974',
                                                                                                'in_ucast_pkts': '83683939',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '906',
                                                                                                'out_octets': '6933213506',
                                                                                                'out_ucast_pkts': '23234245',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '906',
                                                                       'out_octets': '6933213506',
                                                                       'out_ucast_pkts': '23234245',
                                                                       'poller_hash': 'NUE2OUM5cGVlcmludGZwb24gMTM5MTcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139172',
                                                                                               'links': [       {       'connected_time': '688573',
                                                                                                                        'fiber_length': 5768,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2OUM5cGVlcnBvbnBvbiAxMzkxNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4677 PERTH LINE 55 (KUNTZE MURRAY)',
                                         'poller_hash': 'NUE2OUM5',
                                         'serial_number': '5A69C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:ce:8a',
                                         'uptime': '688573'},
                                 {       'counter_source': '139194/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '7a:23:3f:12:f3:92'],
                                                                       'id': '139194/g1',
                                                                       'in_nucast_pkts': '628',
                                                                       'in_octets': '43009546343',
                                                                       'in_ucast_pkts': '69419158',
                                                                       'mac_address': '44:65:7f:7b:d3:58',
                                                                       'mtu': 2000,
                                                                       'name': '139194/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463508',
                                                                       'out_octets': '169246511663',
                                                                       'out_ucast_pkts': '137830063',
                                                                       'poller_hash': 'NUE2QjYzcGVlcmludGYxMzkxOTQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139194',
                                                                       'in_nucast_pkts': '1463508',
                                                                       'in_octets': '169246511663',
                                                                       'in_ucast_pkts': '137830063',
                                                                       'links': [       {       'in_nucast_pkts': '1463508',
                                                                                                'in_octets': '169246511663',
                                                                                                'in_ucast_pkts': '137830063',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '628',
                                                                                                'out_octets': '43009546343',
                                                                                                'out_ucast_pkts': '69419158',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '628',
                                                                       'out_octets': '43009546343',
                                                                       'out_ucast_pkts': '69419158',
                                                                       'poller_hash': 'NUE2QjYzcGVlcmludGZwb24gMTM5MTk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139194',
                                                                                               'links': [       {       'connected_time': '688367',
                                                                                                                        'fiber_length': 931,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjYzcGVlcnBvbnBvbiAxMzkxOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4 WEST ST (OJIGBO RUTE)',
                                         'poller_hash': 'NUE2QjYz',
                                         'serial_number': '5A6B63',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:58',
                                         'uptime': '688367'},
                                 {       'counter_source': '139251/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d5:e5'],
                                                                       'id': '139251/g1',
                                                                       'in_nucast_pkts': '11884',
                                                                       'in_octets': '8221531110',
                                                                       'in_ucast_pkts': '47223162',
                                                                       'mac_address': 'd0:76:8f:aa:52:7f',
                                                                       'mtu': 2000,
                                                                       'name': '139251/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447198',
                                                                       'out_octets': '177419928398',
                                                                       'out_ucast_pkts': '140769157',
                                                                       'poller_hash': 'OTU2Rjc1cGVlcmludGYxMzkyNTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139251',
                                                                       'in_nucast_pkts': '1447198',
                                                                       'in_octets': '177419928398',
                                                                       'in_ucast_pkts': '140769157',
                                                                       'links': [       {       'in_nucast_pkts': '1447198',
                                                                                                'in_octets': '177419928398',
                                                                                                'in_ucast_pkts': '140769157',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11884',
                                                                                                'out_octets': '8221531110',
                                                                                                'out_ucast_pkts': '47223162',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11884',
                                                                       'out_octets': '8221531110',
                                                                       'out_ucast_pkts': '47223162',
                                                                       'poller_hash': 'OTU2Rjc1cGVlcmludGZwb24gMTM5MjUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139251',
                                                                                               'links': [       {       'connected_time': '688419',
                                                                                                                        'fiber_length': 1945,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2Rjc1cGVlcnBvbnBvbiAxMzkyNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '92 REAGAN ST (KROEKER JOHN)',
                                         'poller_hash': 'OTU2Rjc1',
                                         'serial_number': '956F75',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:52:7c',
                                         'uptime': '688419'},
                                 {       'counter_source': '140049/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f0:c5'],
                                                                       'id': '140049/g1',
                                                                       'in_nucast_pkts': '11811',
                                                                       'in_octets': '6431672595',
                                                                       'in_ucast_pkts': '19437957',
                                                                       'mac_address': 'd0:76:8f:06:e5:a4',
                                                                       'mtu': 2000,
                                                                       'name': '140049/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120417',
                                                                       'out_octets': '20654366355',
                                                                       'out_ucast_pkts': '24903413',
                                                                       'poller_hash': 'ODQ0QkM3cGVlcmludGYxNDAwNDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140049',
                                                                       'in_nucast_pkts': '1120417',
                                                                       'in_octets': '20654366355',
                                                                       'in_ucast_pkts': '24903413',
                                                                       'links': [       {       'in_nucast_pkts': '1120417',
                                                                                                'in_octets': '20654366355',
                                                                                                'in_ucast_pkts': '24903413',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11811',
                                                                                                'out_octets': '6431672595',
                                                                                                'out_ucast_pkts': '19437957',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11811',
                                                                       'out_octets': '6431672595',
                                                                       'out_ucast_pkts': '19437957',
                                                                       'poller_hash': 'ODQ0QkM3cGVlcmludGZwb24gMTQwMDQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140049',
                                                                                               'links': [       {       'connected_time': '538569',
                                                                                                                        'fiber_length': 13301,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkM3cGVlcnBvbnBvbiAxNDAwNDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5163 ROAD 140 (WHITE JOSEPH)',
                                         'poller_hash': 'ODQ0QkM3',
                                         'serial_number': '844BC7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:a4',
                                         'uptime': '538569'},
                                 {       'counter_source': '140063/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:1d'],
                                                                       'id': '140063/g1',
                                                                       'in_nucast_pkts': '10511077',
                                                                       'in_octets': '578268176164',
                                                                       'in_ucast_pkts': '3447530806',
                                                                       'mac_address': '48:77:46:cd:55:32',
                                                                       'mtu': 2000,
                                                                       'name': '140063/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8836630826',
                                                                       'out_octets': '22484296741538',
                                                                       'out_ucast_pkts': '7555288461',
                                                                       'poller_hash': 'NzZEMDUycGVlcmludGYxNDAwNjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140063',
                                                                       'in_nucast_pkts': '8836630826',
                                                                       'in_octets': '22484296741538',
                                                                       'in_ucast_pkts': '7555288461',
                                                                       'links': [       {       'in_nucast_pkts': '8836630826',
                                                                                                'in_octets': '22484296741538',
                                                                                                'in_ucast_pkts': '7555288461',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10511077',
                                                                                                'out_octets': '578268176164',
                                                                                                'out_ucast_pkts': '3447530806',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10511077',
                                                                       'out_octets': '578268176164',
                                                                       'out_ucast_pkts': '3447530806',
                                                                       'poller_hash': 'NzZEMDUycGVlcmludGZwb24gMTQwMDYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140063',
                                                                                               'links': [       {       'connected_time': '71306417',
                                                                                                                        'fiber_length': 1945,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.892,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzZEMDUycGVlcnBvbnBvbiAxNDAwNjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '86 REAGAN ST (YANTZI MARCUS)',
                                         'poller_hash': 'NzZEMDUy',
                                         'serial_number': '76D052',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:cd:55:2e',
                                         'uptime': '71306417'},
                                 {       'counter_source': '140091/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:e1'],
                                                                       'id': '140091/g1',
                                                                       'in_nucast_pkts': '14834',
                                                                       'in_octets': '6121184055',
                                                                       'in_ucast_pkts': '31150227',
                                                                       'mac_address': 'd0:76:8f:08:1c:76',
                                                                       'mtu': 2000,
                                                                       'name': '140091/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434273',
                                                                       'out_octets': '124010633563',
                                                                       'out_ucast_pkts': '90677470',
                                                                       'poller_hash': 'ODY5NDg0cGVlcmludGYxNDAwOTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140091',
                                                                       'in_nucast_pkts': '1434273',
                                                                       'in_octets': '124010633563',
                                                                       'in_ucast_pkts': '90677470',
                                                                       'links': [       {       'in_nucast_pkts': '1434273',
                                                                                                'in_octets': '124010633563',
                                                                                                'in_ucast_pkts': '90677470',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14834',
                                                                                                'out_octets': '6121184055',
                                                                                                'out_ucast_pkts': '31150227',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14834',
                                                                       'out_octets': '6121184055',
                                                                       'out_ucast_pkts': '31150227',
                                                                       'poller_hash': 'ODY5NDg0cGVlcmludGZwb24gMTQwMDkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140091',
                                                                                               'links': [       {       'connected_time': '688560',
                                                                                                                        'fiber_length': 1705,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDg0cGVlcnBvbnBvbiAxNDAwOTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '9 MURRAY CT (GREER SANDRA)',
                                         'poller_hash': 'ODY5NDg0',
                                         'serial_number': '869484',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1c:73',
                                         'uptime': '688560'},
                                 {       'counter_source': '140326/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ef:bd'],
                                                                       'id': '140326/g1',
                                                                       'in_nucast_pkts': '15416',
                                                                       'in_octets': '10008189306',
                                                                       'in_ucast_pkts': '24092792',
                                                                       'mac_address': 'd0:76:8f:06:e6:2e',
                                                                       'mtu': 2000,
                                                                       'name': '140326/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434143',
                                                                       'out_octets': '80831142299',
                                                                       'out_ucast_pkts': '73208487',
                                                                       'poller_hash': 'ODQ0QkY1cGVlcmludGYxNDAzMjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140326',
                                                                       'in_nucast_pkts': '1434143',
                                                                       'in_octets': '80831142299',
                                                                       'in_ucast_pkts': '73208487',
                                                                       'links': [       {       'in_nucast_pkts': '1434143',
                                                                                                'in_octets': '80831142299',
                                                                                                'in_ucast_pkts': '73208487',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15416',
                                                                                                'out_octets': '10008189306',
                                                                                                'out_ucast_pkts': '24092792',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15416',
                                                                       'out_octets': '10008189306',
                                                                       'out_ucast_pkts': '24092792',
                                                                       'poller_hash': 'ODQ0QkY1cGVlcmludGZwb24gMTQwMzI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140326',
                                                                                               'links': [       {       'connected_time': '688574',
                                                                                                                        'fiber_length': 1729,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkY1cGVlcnBvbnBvbiAxNDAzMjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '7 MURRAY CT (JOSE NAJMA)',
                                         'poller_hash': 'ODQ0QkY1',
                                         'serial_number': '844BF5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:2e',
                                         'uptime': '688574'},
                                 {       'counter_source': '140456/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:9f'],
                                                                       'id': '140456/g1',
                                                                       'in_nucast_pkts': '14887',
                                                                       'in_octets': '23218845494',
                                                                       'in_ucast_pkts': '104964966',
                                                                       'mac_address': 'd0:76:8f:08:09:86',
                                                                       'mtu': 2000,
                                                                       'name': '140456/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447150',
                                                                       'out_octets': '441006268941',
                                                                       'out_ucast_pkts': '331748005',
                                                                       'poller_hash': 'ODY5MjI2cGVlcmludGYxNDA0NTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140456',
                                                                       'in_nucast_pkts': '1447150',
                                                                       'in_octets': '441006268941',
                                                                       'in_ucast_pkts': '331748005',
                                                                       'links': [       {       'in_nucast_pkts': '1447150',
                                                                                                'in_octets': '441006268941',
                                                                                                'in_ucast_pkts': '331748005',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14887',
                                                                                                'out_octets': '23218845494',
                                                                                                'out_ucast_pkts': '104964966',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14887',
                                                                       'out_octets': '23218845494',
                                                                       'out_ucast_pkts': '104964966',
                                                                       'poller_hash': 'ODY5MjI2cGVlcmludGZwb24gMTQwNDU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140456',
                                                                                               'links': [       {       'connected_time': '688409',
                                                                                                                        'fiber_length': 2403,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MjI2cGVlcnBvbnBvbiAxNDA0NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '65 GERBER DR (NORRIS DEANNE)',
                                         'poller_hash': 'ODY5MjI2',
                                         'serial_number': '869226',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:09:83',
                                         'uptime': '688409'},
                                 {       'counter_source': '140477/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:f4:11:60:9e:a9'],
                                                                       'id': '140477/g1',
                                                                       'in_nucast_pkts': '886',
                                                                       'in_octets': '53216365613',
                                                                       'in_ucast_pkts': '59561008',
                                                                       'mac_address': 'd0:76:8f:06:e6:1c',
                                                                       'mtu': 2000,
                                                                       'name': '140477/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447171',
                                                                       'out_octets': '9757896597',
                                                                       'out_ucast_pkts': '42950110',
                                                                       'poller_hash': 'ODQ0QkVGcGVlcmludGYxNDA0NzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140477',
                                                                       'in_nucast_pkts': '1447171',
                                                                       'in_octets': '9757896597',
                                                                       'in_ucast_pkts': '42950110',
                                                                       'links': [       {       'in_nucast_pkts': '1447171',
                                                                                                'in_octets': '9757896597',
                                                                                                'in_ucast_pkts': '42950110',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '886',
                                                                                                'out_octets': '53216365613',
                                                                                                'out_ucast_pkts': '59561008',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '886',
                                                                       'out_octets': '53216365613',
                                                                       'out_ucast_pkts': '59561008',
                                                                       'poller_hash': 'ODQ0QkVGcGVlcmludGZwb24gMTQwNDc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140477',
                                                                                               'links': [       {       'connected_time': '688432',
                                                                                                                        'fiber_length': 2092,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkVGcGVlcnBvbnBvbiAxNDA0Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '83 REAGAN ST (RUAN ZHIHENG)',
                                         'poller_hash': 'ODQ0QkVG',
                                         'serial_number': '844BEF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:1c',
                                         'uptime': '688432'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '140527/g1',
                                                                       'name': '140527/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODQ0QkQ2cGVlcmludGYxNDA1MjcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140527',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODQ0QkQ2cGVlcmludGZwb24gMTQwNTI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140527',
                                                                                               'links': [       {       'connected_time': '67790238',
                                                                                                                        'fiber_length': 1397,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkQ2cGVlcnBvbnBvbiAxNDA1Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '6466 PERTH ROAD 131 (WALL FRANK)',
                                         'poller_hash': 'ODQ0QkQ2',
                                         'serial_number': '844BD6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:d1',
                                         'uptime': '67790238'},
                                 {       'counter_source': '140669/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '140669/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '52528117846',
                                                                       'in_ucast_pkts': '226983349',
                                                                       'mac_address': 'd0:76:8f:08:19:13',
                                                                       'name': '140669/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1041748484237',
                                                                       'out_ucast_pkts': '758268984',
                                                                       'poller_hash': 'ODY5NDE4cGVlcmludGYxNDA2NjkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140669',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1041748484237',
                                                                       'in_ucast_pkts': '758268984',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1041748484237',
                                                                                                'in_ucast_pkts': '758268984',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '52528117846',
                                                                                                'out_ucast_pkts': '226983349',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '52528117846',
                                                                       'out_ucast_pkts': '226983349',
                                                                       'poller_hash': 'ODY5NDE4cGVlcmludGZwb24gMTQwNjY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140669',
                                                                                               'links': [       {       'connected_time': '6050414',
                                                                                                                        'fiber_length': 1088,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDE4cGVlcnBvbnBvbiAxNDA2Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '36 MAIN ST S (JOHNSON JAMIE & HEIDI)',
                                         'poller_hash': 'ODY5NDE4',
                                         'serial_number': '869418',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:13',
                                         'uptime': '6050414'},
                                 {       'counter_source': '140724/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:8a'],
                                                                       'id': '140724/g1',
                                                                       'in_nucast_pkts': '15396',
                                                                       'in_octets': '1293085795',
                                                                       'in_ucast_pkts': '12394425',
                                                                       'mac_address': '48:77:46:f8:75:14',
                                                                       'mtu': 2000,
                                                                       'name': '140724/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463864',
                                                                       'out_octets': '47741208514',
                                                                       'out_ucast_pkts': '32599593',
                                                                       'poller_hash': 'ODJBNTMycGVlcmludGYxNDA3MjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140724',
                                                                       'in_nucast_pkts': '1463864',
                                                                       'in_octets': '47741208514',
                                                                       'in_ucast_pkts': '32599593',
                                                                       'links': [       {       'in_nucast_pkts': '1463864',
                                                                                                'in_octets': '47741208514',
                                                                                                'in_ucast_pkts': '32599593',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15396',
                                                                                                'out_octets': '1293085795',
                                                                                                'out_ucast_pkts': '12394425',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15396',
                                                                       'out_octets': '1293085795',
                                                                       'out_ucast_pkts': '12394425',
                                                                       'poller_hash': 'ODJBNTMycGVlcmludGZwb24gMTQwNzI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140724',
                                                                                               'links': [       {       'connected_time': '688442',
                                                                                                                        'fiber_length': 1301,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNTMycGVlcnBvbnBvbiAxNDA3MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '15 JO-LENE CT (HALE RICK)',
                                         'poller_hash': 'ODJBNTMy',
                                         'serial_number': '82A532',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:75:14',
                                         'uptime': '688442'},
                                 {       'counter_source': '141011/F1+141011/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:46:4a',
                                                                                                        '00:03:e6:eb:4b:14'],
                                                                       'id': '141011/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4909178855',
                                                                       'in_ucast_pkts': '26596501',
                                                                       'name': '141011/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '18088633714',
                                                                       'out_octets': '24422826438606',
                                                                       'out_ucast_pkts': '7724894',
                                                                       'poller_hash': 'ODY5NDI3cGVlcmludGYxNDEwMTEvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:10'],
                                                                       'id': '141011/g1',
                                                                       'in_nucast_pkts': '325111',
                                                                       'in_octets': '260192668541',
                                                                       'in_ucast_pkts': '619801493',
                                                                       'mac_address': 'd0:76:8f:08:19:8e',
                                                                       'mtu': 2000,
                                                                       'name': '141011/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '34647132',
                                                                       'out_octets': '2338265472071',
                                                                       'out_ucast_pkts': '1831282579',
                                                                       'poller_hash': 'ODY5NDI3cGVlcmludGYxNDEwMTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141011',
                                                                       'in_nucast_pkts': '18123280846',
                                                                       'in_octets': '26761091910677',
                                                                       'in_ucast_pkts': '1839007473',
                                                                       'links': [       {       'in_nucast_pkts': '18123280846',
                                                                                                'in_octets': '26761091910677',
                                                                                                'in_ucast_pkts': '1839007473',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '325111',
                                                                                                'out_octets': '265101847396',
                                                                                                'out_ucast_pkts': '646397994',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '325111',
                                                                       'out_octets': '265101847396',
                                                                       'out_ucast_pkts': '646397994',
                                                                       'poller_hash': 'ODY5NDI3cGVlcmludGZwb24gMTQxMDEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141011',
                                                                                               'links': [       {       'connected_time': '15498560',
                                                                                                                        'fiber_length': 2094,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDI3cGVlcnBvbnBvbiAxNDEwMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '79 REAGAN ST (ALVAREZ JOE)',
                                         'poller_hash': 'ODY5NDI3',
                                         'serial_number': '869427',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:8b',
                                         'uptime': '15498560'},
                                 {       'counter_source': '141031/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141031/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1277760028422',
                                                                       'in_ucast_pkts': '4212119480',
                                                                       'mac_address': 'd0:76:8f:08:19:23',
                                                                       'name': '141031/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '13750204723802',
                                                                       'out_ucast_pkts': '10447209755',
                                                                       'poller_hash': 'ODY5NDFBcGVlcmludGYxNDEwMzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141031',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '13750204723802',
                                                                       'in_ucast_pkts': '10447209755',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '13750204723802',
                                                                                                'in_ucast_pkts': '10447209755',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1277760028422',
                                                                                                'out_ucast_pkts': '4212119480',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1277760028422',
                                                                       'out_ucast_pkts': '4212119480',
                                                                       'poller_hash': 'ODY5NDFBcGVlcmludGZwb24gMTQxMDMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141031',
                                                                                               'links': [       {       'connected_time': '688408',
                                                                                                                        'fiber_length': 2243,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDFBcGVlcnBvbnBvbiAxNDEwMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '70 GERBER DR (DARASCH SAMANTHA)',
                                         'poller_hash': 'ODY5NDFB',
                                         'serial_number': '86941A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:23',
                                         'uptime': '688408'},
                                 {       'counter_source': '141041/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141041/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '429341258370',
                                                                       'in_ucast_pkts': '26928436449',
                                                                       'mac_address': 'd0:76:8f:08:18:93',
                                                                       'name': '141041/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '119',
                                                                       'out_octets': '3330760003758',
                                                                       'out_ucast_pkts': '2619571145',
                                                                       'poller_hash': 'ODY5NDA4cGVlcmludGYxNDEwNDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141041',
                                                                       'in_nucast_pkts': '119',
                                                                       'in_octets': '3330760003758',
                                                                       'in_ucast_pkts': '2619571145',
                                                                       'links': [       {       'in_nucast_pkts': '119',
                                                                                                'in_octets': '3330760003758',
                                                                                                'in_ucast_pkts': '2619571145',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '429341258370',
                                                                                                'out_ucast_pkts': '26928436449',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '429341258370',
                                                                       'out_ucast_pkts': '26928436449',
                                                                       'poller_hash': 'ODY5NDA4cGVlcmludGZwb24gMTQxMDQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141041',
                                                                                               'links': [       {       'connected_time': '72955401',
                                                                                                                        'fiber_length': 17977,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDA4cGVlcnBvbnBvbiAxNDEwNDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '154 SMITH ST (GINGERICH BRAEDEN)',
                                         'poller_hash': 'ODY5NDA4',
                                         'serial_number': '869408',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:93',
                                         'uptime': '72955401'},
                                 {       'counter_source': '141109/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141109/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '203959142262',
                                                                       'in_ucast_pkts': '699148958',
                                                                       'mac_address': 'd0:76:8f:08:18:23',
                                                                       'name': '141109/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '2435153906963',
                                                                       'out_ucast_pkts': '1928935990',
                                                                       'poller_hash': 'ODY5M0ZBcGVlcmludGYxNDExMDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141109',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '2435153906963',
                                                                       'in_ucast_pkts': '1928935990',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '2435153906963',
                                                                                                'in_ucast_pkts': '1928935990',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '203959142262',
                                                                                                'out_ucast_pkts': '699148958',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '203959142262',
                                                                       'out_ucast_pkts': '699148958',
                                                                       'poller_hash': 'ODY5M0ZBcGVlcmludGZwb24gMTQxMTA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141109',
                                                                                               'links': [       {       'connected_time': '688401',
                                                                                                                        'fiber_length': 2094,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0ZBcGVlcnBvbnBvbiAxNDExMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '89 REAGAN ST MILVERTON (VINK JASON)',
                                         'poller_hash': 'ODY5M0ZB',
                                         'serial_number': '8693FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:23',
                                         'uptime': '688401'},
                                 {       'counter_source': '141139/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:62:b8:2f:ef:61'],
                                                                       'id': '141139/g1',
                                                                       'in_nucast_pkts': '636',
                                                                       'in_octets': '5821300917',
                                                                       'in_ucast_pkts': '21393255',
                                                                       'mac_address': 'cc:be:59:50:03:0a',
                                                                       'mtu': 2000,
                                                                       'name': '141139/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1119735',
                                                                       'out_octets': '120016045239',
                                                                       'out_ucast_pkts': '87947413',
                                                                       'poller_hash': 'NDNCNTRDcGVlcmludGYxNDExMzkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141139',
                                                                       'in_nucast_pkts': '1119735',
                                                                       'in_octets': '120016045239',
                                                                       'in_ucast_pkts': '87947413',
                                                                       'links': [       {       'in_nucast_pkts': '1119735',
                                                                                                'in_octets': '120016045239',
                                                                                                'in_ucast_pkts': '87947413',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '636',
                                                                                                'out_octets': '5821300917',
                                                                                                'out_ucast_pkts': '21393255',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '636',
                                                                       'out_octets': '5821300917',
                                                                       'out_ucast_pkts': '21393255',
                                                                       'poller_hash': 'NDNCNTRDcGVlcmludGZwb24gMTQxMTM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141139',
                                                                                               'links': [       {       'connected_time': '688413',
                                                                                                                        'fiber_length': 880,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.38,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.326}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNTRDcGVlcnBvbnBvbiAxNDExMzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '20 ST MARYS ST S (HIEBERT JUDITH)',
                                         'poller_hash': 'NDNCNTRD',
                                         'serial_number': '43B54C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:03:09',
                                         'uptime': '688413'},
                                 {       'counter_source': '141201/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141201/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '62335055385',
                                                                       'in_ucast_pkts': '299330710',
                                                                       'mac_address': 'd0:76:8f:08:0b:db',
                                                                       'name': '141201/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1134645597345',
                                                                       'out_ucast_pkts': '852185851',
                                                                       'poller_hash': 'ODY5MjcxcGVlcmludGYxNDEyMDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141201',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1134645597345',
                                                                       'in_ucast_pkts': '852185851',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1134645597345',
                                                                                                'in_ucast_pkts': '852185851',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '62335055385',
                                                                                                'out_ucast_pkts': '299330710',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '62335055385',
                                                                       'out_ucast_pkts': '299330710',
                                                                       'poller_hash': 'ODY5MjcxcGVlcmludGZwb24gMTQxMjAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141201',
                                                                                               'links': [       {       'connected_time': '538549',
                                                                                                                        'fiber_length': 13962,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MjcxcGVlcnBvbnBvbiAxNDEyMDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5317 ROAD 144 (GRAUL RODNEY)',
                                         'poller_hash': 'ODY5Mjcx',
                                         'serial_number': '869271',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:0b:db',
                                         'uptime': '538549'},
                                 {       'counter_source': '141258/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:f5:a2:a3:14:e0'],
                                                                       'id': '141258/g1',
                                                                       'in_nucast_pkts': '4445503',
                                                                       'in_octets': '600608554298',
                                                                       'in_ucast_pkts': '2516191404',
                                                                       'mac_address': 'd0:76:8f:06:e6:19',
                                                                       'mtu': 2000,
                                                                       'name': '141258/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '73763124',
                                                                       'out_octets': '9077631950006',
                                                                       'out_ucast_pkts': '6727140046',
                                                                       'poller_hash': 'ODQ0QkVFcGVlcmludGYxNDEyNTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141258',
                                                                       'in_nucast_pkts': '73763124',
                                                                       'in_octets': '9077631950006',
                                                                       'in_ucast_pkts': '6727140046',
                                                                       'links': [       {       'in_nucast_pkts': '73763124',
                                                                                                'in_octets': '9077631950006',
                                                                                                'in_ucast_pkts': '6727140046',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4445503',
                                                                                                'out_octets': '600608554298',
                                                                                                'out_ucast_pkts': '2516191404',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4445503',
                                                                       'out_octets': '600608554298',
                                                                       'out_ucast_pkts': '2516191404',
                                                                       'poller_hash': 'ODQ0QkVFcGVlcmludGZwb24gMTQxMjU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141258',
                                                                                               'links': [       {       'connected_time': '67790661',
                                                                                                                        'fiber_length': 14418,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkVFcGVlcnBvbnBvbiAxNDEyNTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5136 LINE 49 (BAUER ELONA)',
                                         'poller_hash': 'ODQ0QkVF',
                                         'serial_number': '844BEE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:19',
                                         'uptime': '67790661'},
                                 {       'counter_source': '141462/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f3:b1'],
                                                                       'id': '141462/g1',
                                                                       'in_nucast_pkts': '15118',
                                                                       'in_octets': '12472542766',
                                                                       'in_ucast_pkts': '72544749',
                                                                       'mac_address': '48:77:46:f8:79:40',
                                                                       'mtu': 2000,
                                                                       'name': '141462/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447037',
                                                                       'out_octets': '244080712324',
                                                                       'out_ucast_pkts': '172225836',
                                                                       'poller_hash': 'ODJBNjk2cGVlcmludGYxNDE0NjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141462',
                                                                       'in_nucast_pkts': '1447037',
                                                                       'in_octets': '244080712324',
                                                                       'in_ucast_pkts': '172225836',
                                                                       'links': [       {       'in_nucast_pkts': '1447037',
                                                                                                'in_octets': '244080712324',
                                                                                                'in_ucast_pkts': '172225836',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15118',
                                                                                                'out_octets': '12472542766',
                                                                                                'out_ucast_pkts': '72544749',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15118',
                                                                       'out_octets': '12472542766',
                                                                       'out_ucast_pkts': '72544749',
                                                                       'poller_hash': 'ODJBNjk2cGVlcmludGZwb24gMTQxNDYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141462',
                                                                                               'links': [       {       'connected_time': '688427',
                                                                                                                        'fiber_length': 2381,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNjk2cGVlcnBvbnBvbiAxNDE0NjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '66 GERBER DR (SCHEERER JEFF & KAREN)',
                                         'poller_hash': 'ODJBNjk2',
                                         'serial_number': '82A696',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:79:40',
                                         'uptime': '688427'},
                                 {       'counter_source': '141641/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f4:e2:c6:f2:3a:e9'],
                                                                       'id': '141641/g1',
                                                                       'in_nucast_pkts': '58066',
                                                                       'in_octets': '105070639727',
                                                                       'in_ucast_pkts': '263362507',
                                                                       'mac_address': '48:77:46:f8:7b:41',
                                                                       'mtu': 2000,
                                                                       'name': '141641/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9364183',
                                                                       'out_octets': '547737904940',
                                                                       'out_ucast_pkts': '674733829',
                                                                       'poller_hash': 'ODJBNzQxcGVlcmludGYxNDE2NDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141641',
                                                                       'in_nucast_pkts': '9364183',
                                                                       'in_octets': '547737904940',
                                                                       'in_ucast_pkts': '674733829',
                                                                       'links': [       {       'in_nucast_pkts': '9364183',
                                                                                                'in_octets': '547737904940',
                                                                                                'in_ucast_pkts': '674733829',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '58066',
                                                                                                'out_octets': '105070639727',
                                                                                                'out_ucast_pkts': '263362507',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58066',
                                                                       'out_octets': '105070639727',
                                                                       'out_ucast_pkts': '263362507',
                                                                       'poller_hash': 'ODJBNzQxcGVlcmludGZwb24gMTQxNjQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141641',
                                                                                               'links': [       {       'connected_time': '4315678',
                                                                                                                        'fiber_length': 10894,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNzQxcGVlcnBvbnBvbiAxNDE2NDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5960 PERTH RD 119 (715276 ONTARIO '
                                                 'INC)',
                                         'poller_hash': 'ODJBNzQx',
                                         'serial_number': '82A741',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:7b:41',
                                         'uptime': '4315678'},
                                 {       'counter_source': '141679/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:63:b1'],
                                                                       'id': '141679/g1',
                                                                       'in_nucast_pkts': '947515',
                                                                       'in_octets': '1157640598716',
                                                                       'in_ucast_pkts': '4535728065',
                                                                       'mac_address': 'd0:76:8f:21:46:f6',
                                                                       'mtu': 2000,
                                                                       'name': '141679/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '129897056',
                                                                       'out_octets': '20724681785847',
                                                                       'out_ucast_pkts': '15114355254',
                                                                       'poller_hash': 'ODdGOEJEcGVlcmludGYxNDE2NzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141679',
                                                                       'in_nucast_pkts': '129897056',
                                                                       'in_octets': '20724681785847',
                                                                       'in_ucast_pkts': '15114355254',
                                                                       'links': [       {       'in_nucast_pkts': '129897056',
                                                                                                'in_octets': '20724681785847',
                                                                                                'in_ucast_pkts': '15114355254',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '947515',
                                                                                                'out_octets': '1157640598716',
                                                                                                'out_ucast_pkts': '4535728065',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '947515',
                                                                       'out_octets': '1157640598716',
                                                                       'out_ucast_pkts': '4535728065',
                                                                       'poller_hash': 'ODdGOEJEcGVlcmludGZwb24gMTQxNjc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141679',
                                                                                               'links': [       {       'connected_time': '21816367',
                                                                                                                        'fiber_length': 1484,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODdGOEJEcGVlcnBvbnBvbiAxNDE2Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '85A MILL ST E (KNUDSEN KAREN)',
                                         'poller_hash': 'ODdGOEJE',
                                         'serial_number': '87F8BD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:21:46:f3',
                                         'uptime': '21816367'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '141774/g1',
                                                                       'name': '141774/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQUZGcGVlcmludGYxNDE3NzQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141774',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NjQwQUZGcGVlcmludGZwb24gMTQxNzc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141774',
                                                                                               'links': [       {       'connected_time': '688346',
                                                                                                                        'fiber_length': 901,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.412,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.162}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUZGcGVlcnBvbnBvbiAxNDE3NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '55 MAIN ST S',
                                         'poller_hash': 'NjQwQUZG',
                                         'serial_number': '640AFF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:0e',
                                         'uptime': '688346'},
                                 {       'counter_source': '141797/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:65:8a'],
                                                                       'id': '141797/g1',
                                                                       'in_nucast_pkts': '14757',
                                                                       'in_octets': '13571597016',
                                                                       'in_ucast_pkts': '95021304',
                                                                       'mac_address': 'd0:76:8f:08:19:de',
                                                                       'mtu': 2000,
                                                                       'name': '141797/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447007',
                                                                       'out_octets': '330071842171',
                                                                       'out_ucast_pkts': '236334535',
                                                                       'poller_hash': 'ODY5NDMxcGVlcmludGYxNDE3OTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141797',
                                                                       'in_nucast_pkts': '1447007',
                                                                       'in_octets': '330071842171',
                                                                       'in_ucast_pkts': '236334535',
                                                                       'links': [       {       'in_nucast_pkts': '1447007',
                                                                                                'in_octets': '330071842171',
                                                                                                'in_ucast_pkts': '236334535',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14757',
                                                                                                'out_octets': '13571597016',
                                                                                                'out_ucast_pkts': '95021304',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14757',
                                                                       'out_octets': '13571597016',
                                                                       'out_ucast_pkts': '95021304',
                                                                       'poller_hash': 'ODY5NDMxcGVlcmludGZwb24gMTQxNzk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141797',
                                                                                               'links': [       {       'connected_time': '688409',
                                                                                                                        'fiber_length': 1966,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDMxcGVlcnBvbnBvbiAxNDE3OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '90 REAGAN ST (GRAHAM DAVE)',
                                         'poller_hash': 'ODY5NDMx',
                                         'serial_number': '869431',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:db',
                                         'uptime': '688409'},
                                 {       'counter_source': '141837/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:36:c9'],
                                                                       'id': '141837/g1',
                                                                       'in_nucast_pkts': '13311',
                                                                       'in_octets': '8880389846',
                                                                       'in_ucast_pkts': '17490778',
                                                                       'mac_address': 'd0:76:8f:08:18:9e',
                                                                       'mtu': 2000,
                                                                       'name': '141837/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464547',
                                                                       'out_octets': '65461113173',
                                                                       'out_ucast_pkts': '56886157',
                                                                       'poller_hash': 'ODY5NDA5cGVlcmludGYxNDE4MzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141837',
                                                                       'in_nucast_pkts': '1464547',
                                                                       'in_octets': '65461113173',
                                                                       'in_ucast_pkts': '56886157',
                                                                       'links': [       {       'in_nucast_pkts': '1464547',
                                                                                                'in_octets': '65461113173',
                                                                                                'in_ucast_pkts': '56886157',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13311',
                                                                                                'out_octets': '8880389846',
                                                                                                'out_ucast_pkts': '17490778',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13311',
                                                                       'out_octets': '8880389846',
                                                                       'out_ucast_pkts': '17490778',
                                                                       'poller_hash': 'ODY5NDA5cGVlcmludGZwb24gMTQxODM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141837',
                                                                                               'links': [       {       'connected_time': '688570',
                                                                                                                        'fiber_length': 8130,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDA5cGVlcnBvbnBvbiAxNDE4Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4509 LINE 61 (KUEPFER ALVIN)',
                                         'poller_hash': 'ODY5NDA5',
                                         'serial_number': '869409',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:9b',
                                         'uptime': '688570'},
                                 {       'counter_source': '141847/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141847/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '37049525285',
                                                                       'in_ucast_pkts': '106428875',
                                                                       'mac_address': 'd0:76:8f:08:1c:e3',
                                                                       'name': '141847/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '86',
                                                                       'out_octets': '289110741946',
                                                                       'out_ucast_pkts': '291908107',
                                                                       'poller_hash': 'ODY5NDkycGVlcmludGYxNDE4NDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141847',
                                                                       'in_nucast_pkts': '86',
                                                                       'in_octets': '289110741946',
                                                                       'in_ucast_pkts': '291908107',
                                                                       'links': [       {       'in_nucast_pkts': '86',
                                                                                                'in_octets': '289110741946',
                                                                                                'in_ucast_pkts': '291908107',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '37049525285',
                                                                                                'out_ucast_pkts': '106428875',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '37049525285',
                                                                       'out_ucast_pkts': '106428875',
                                                                       'poller_hash': 'ODY5NDkycGVlcmludGZwb24gMTQxODQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141847',
                                                                                               'links': [       {       'connected_time': '21816220',
                                                                                                                        'fiber_length': 8233,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDkycGVlcnBvbnBvbiAxNDE4NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4509 LINE 61 (LEGACY STONE INC)',
                                         'poller_hash': 'ODY5NDky',
                                         'serial_number': '869492',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1c:e3',
                                         'uptime': '21816220'},
                                 {       'counter_source': '141867/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141867/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '90452451260',
                                                                       'in_ucast_pkts': '481681748',
                                                                       'mac_address': 'd0:76:8f:08:1b:bb',
                                                                       'name': '141867/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '135',
                                                                       'out_octets': '2712230052244',
                                                                       'out_ucast_pkts': '1920562486',
                                                                       'poller_hash': 'ODY5NDZEcGVlcmludGYxNDE4NjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141867',
                                                                       'in_nucast_pkts': '135',
                                                                       'in_octets': '2712230052244',
                                                                       'in_ucast_pkts': '1920562486',
                                                                       'links': [       {       'in_nucast_pkts': '135',
                                                                                                'in_octets': '2712230052244',
                                                                                                'in_ucast_pkts': '1920562486',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '90452451260',
                                                                                                'out_ucast_pkts': '481681748',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '90452451260',
                                                                       'out_ucast_pkts': '481681748',
                                                                       'poller_hash': 'ODY5NDZEcGVlcmludGZwb24gMTQxODY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141867',
                                                                                               'links': [       {       'connected_time': '21816868',
                                                                                                                        'fiber_length': 5060,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDZEcGVlcnBvbnBvbiAxNDE4Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6137 PERTH ROAD 131 (HADDOW BRIAN & '
                                                 'REID)',
                                         'poller_hash': 'ODY5NDZE',
                                         'serial_number': '86946D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1b:bb',
                                         'uptime': '21816868'},
                                 {       'counter_source': '141877/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141877/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1643322297651',
                                                                       'in_ucast_pkts': '36393324448',
                                                                       'mac_address': 'd0:76:8f:08:1b:f3',
                                                                       'name': '141877/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '4099892658813',
                                                                       'out_ucast_pkts': '3552325179',
                                                                       'poller_hash': 'ODY5NDc0cGVlcmludGYxNDE4NzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141877',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '4099892658813',
                                                                       'in_ucast_pkts': '3552325179',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '4099892658813',
                                                                                                'in_ucast_pkts': '3552325179',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1643322297651',
                                                                                                'out_ucast_pkts': '36393324448',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1643322297651',
                                                                       'out_ucast_pkts': '36393324448',
                                                                       'poller_hash': 'ODY5NDc0cGVlcmludGZwb24gMTQxODc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141877',
                                                                                               'links': [       {       'connected_time': '688594',
                                                                                                                        'fiber_length': 5409,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDc0cGVlcnBvbnBvbiAxNDE4Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4643 PERTH LINE 55 (REIS COURTNEY)',
                                         'poller_hash': 'ODY5NDc0',
                                         'serial_number': '869474',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1b:f3',
                                         'uptime': '688594'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '141887/g1',
                                                                       'name': '141887/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5NDBGcGVlcmludGYxNDE4ODcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141887',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5NDBGcGVlcmludGZwb24gMTQxODg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141887',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDBGcGVlcnBvbnBvbiAxNDE4ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '12 WILLIAM ST',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "141887" '
                                                                  '(location: GPON '
                                                                  'port "7/7"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:34.870Z',
                                         'poller_hash': 'ODY5NDBG',
                                         'serial_number': '86940F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '141897/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:72:8e'],
                                                                       'id': '141897/g1',
                                                                       'in_nucast_pkts': '4721485',
                                                                       'in_octets': '199582382316',
                                                                       'in_ucast_pkts': '1333483139',
                                                                       'mac_address': 'd0:76:8f:08:18:c6',
                                                                       'mtu': 2000,
                                                                       'name': '141897/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75542556',
                                                                       'out_octets': '7235002350650',
                                                                       'out_ucast_pkts': '5123399121',
                                                                       'poller_hash': 'ODY5NDBFcGVlcmludGYxNDE4OTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141897',
                                                                       'in_nucast_pkts': '75542556',
                                                                       'in_octets': '7235002350650',
                                                                       'in_ucast_pkts': '5123399121',
                                                                       'links': [       {       'in_nucast_pkts': '75542556',
                                                                                                'in_octets': '7235002350650',
                                                                                                'in_ucast_pkts': '5123399121',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4721485',
                                                                                                'out_octets': '199582382316',
                                                                                                'out_ucast_pkts': '1333483139',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4721485',
                                                                       'out_octets': '199582382316',
                                                                       'out_ucast_pkts': '1333483139',
                                                                       'poller_hash': 'ODY5NDBFcGVlcmludGZwb24gMTQxODk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141897',
                                                                                               'links': [       {       'connected_time': '21816203',
                                                                                                                        'fiber_length': 10619,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDBFcGVlcnBvbnBvbiAxNDE4OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5926 SOMERSET DR (HARRIS  STEVE)',
                                         'poller_hash': 'ODY5NDBF',
                                         'serial_number': '86940E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:c3',
                                         'uptime': '21816203'},
                                 {       'counter_source': '141907/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:65:27'],
                                                                       'id': '141907/g1',
                                                                       'in_nucast_pkts': '399327',
                                                                       'in_octets': '204505333381',
                                                                       'in_ucast_pkts': '1035549133',
                                                                       'mac_address': 'd0:76:8f:08:17:de',
                                                                       'mtu': 2000,
                                                                       'name': '141907/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '47333432',
                                                                       'out_octets': '6594400870847',
                                                                       'out_ucast_pkts': '4975054125',
                                                                       'poller_hash': 'ODY5M0YxcGVlcmludGYxNDE5MDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141907',
                                                                       'in_nucast_pkts': '47333432',
                                                                       'in_octets': '6594400870847',
                                                                       'in_ucast_pkts': '4975054125',
                                                                       'links': [       {       'in_nucast_pkts': '47333432',
                                                                                                'in_octets': '6594400870847',
                                                                                                'in_ucast_pkts': '4975054125',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '399327',
                                                                                                'out_octets': '204505333381',
                                                                                                'out_ucast_pkts': '1035549133',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '399327',
                                                                       'out_octets': '204505333381',
                                                                       'out_ucast_pkts': '1035549133',
                                                                       'poller_hash': 'ODY5M0YxcGVlcmludGZwb24gMTQxOTA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141907',
                                                                                               'links': [       {       'connected_time': '21321701',
                                                                                                                        'fiber_length': 10542,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0YxcGVlcnBvbnBvbiAxNDE5MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5927 SOMERSET DR (HALLAHAN JAMIE & '
                                                 'LAURA)',
                                         'poller_hash': 'ODY5M0Yx',
                                         'serial_number': '8693F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:db',
                                         'uptime': '21321701'},
                                 {       'counter_source': '141917/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141917/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '501769043660',
                                                                       'in_ucast_pkts': '2019428890',
                                                                       'mac_address': 'd0:76:8f:08:18:fb',
                                                                       'name': '141917/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '8901787679893',
                                                                       'out_ucast_pkts': '6462736051',
                                                                       'poller_hash': 'ODY5NDE1cGVlcmludGYxNDE5MTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141917',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '8901787679893',
                                                                       'in_ucast_pkts': '6462736051',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '8901787679893',
                                                                                                'in_ucast_pkts': '6462736051',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '501769043660',
                                                                                                'out_ucast_pkts': '2019428890',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '501769043660',
                                                                       'out_ucast_pkts': '2019428890',
                                                                       'poller_hash': 'ODY5NDE1cGVlcmludGZwb24gMTQxOTE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141917',
                                                                                               'links': [       {       'connected_time': '688374',
                                                                                                                        'fiber_length': 10296,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDE1cGVlcnBvbnBvbiAxNDE5MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5241 LINE 58 (BANNON NIKKI)',
                                         'poller_hash': 'ODY5NDE1',
                                         'serial_number': '869415',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:fb',
                                         'uptime': '688374'},
                                 {       'counter_source': '141927/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141927/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1338142167095',
                                                                       'in_ucast_pkts': '3253097975',
                                                                       'mac_address': 'd0:76:8f:08:18:a3',
                                                                       'name': '141927/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '7856821298234',
                                                                       'out_ucast_pkts': '6339414335',
                                                                       'poller_hash': 'ODY5NDBBcGVlcmludGYxNDE5MjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141927',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '7856821298234',
                                                                       'in_ucast_pkts': '6339414335',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '7856821298234',
                                                                                                'in_ucast_pkts': '6339414335',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1338142167095',
                                                                                                'out_ucast_pkts': '3253097975',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1338142167095',
                                                                       'out_ucast_pkts': '3253097975',
                                                                       'poller_hash': 'ODY5NDBBcGVlcmludGZwb24gMTQxOTI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141927',
                                                                                               'links': [       {       'connected_time': '13121918',
                                                                                                                        'fiber_length': 11030,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDBBcGVlcnBvbnBvbiAxNDE5Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4821 PERTH LINE 55 (LOOSER CHRIS)',
                                         'poller_hash': 'ODY5NDBB',
                                         'serial_number': '86940A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:a3',
                                         'uptime': '13121918'},
                                 {       'counter_source': '141937/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141937/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '247750286306',
                                                                       'in_ucast_pkts': '1704330457',
                                                                       'mac_address': 'd0:76:8f:08:17:c3',
                                                                       'name': '141937/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '7164466806485',
                                                                       'out_ucast_pkts': '5141348929',
                                                                       'poller_hash': 'ODY5M0VFcGVlcmludGYxNDE5MzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141937',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '7164466806485',
                                                                       'in_ucast_pkts': '5141348929',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '7164466806485',
                                                                                                'in_ucast_pkts': '5141348929',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '247750286306',
                                                                                                'out_ucast_pkts': '1704330457',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '247750286306',
                                                                       'out_ucast_pkts': '1704330457',
                                                                       'poller_hash': 'ODY5M0VFcGVlcmludGZwb24gMTQxOTM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141937',
                                                                                               'links': [       {       'connected_time': '688550',
                                                                                                                        'fiber_length': 4199,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0VFcGVlcnBvbnBvbiAxNDE5Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6212 PERTH ROAD 131 (MCCORMACK '
                                                 'KELLY)',
                                         'poller_hash': 'ODY5M0VF',
                                         'serial_number': '8693EE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:c3',
                                         'uptime': '688550'},
                                 {       'counter_source': '141947/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141947/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '97241683566',
                                                                       'in_ucast_pkts': '331057998',
                                                                       'mac_address': 'd0:76:8f:08:15:cb',
                                                                       'name': '141947/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '1613851457507',
                                                                       'out_ucast_pkts': '1243777444',
                                                                       'poller_hash': 'ODY5M0FGcGVlcmludGYxNDE5NDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141947',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '1613851457507',
                                                                       'in_ucast_pkts': '1243777444',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '1613851457507',
                                                                                                'in_ucast_pkts': '1243777444',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '97241683566',
                                                                                                'out_ucast_pkts': '331057998',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '97241683566',
                                                                       'out_ucast_pkts': '331057998',
                                                                       'poller_hash': 'ODY5M0FGcGVlcmludGZwb24gMTQxOTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141947',
                                                                                               'links': [       {       'connected_time': '2082693',
                                                                                                                        'fiber_length': 11201,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0FGcGVlcnBvbnBvbiAxNDE5NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6146 ROAD 136 (KUEPFER GERALD & '
                                                 'DARLENE)',
                                         'poller_hash': 'ODY5M0FG',
                                         'serial_number': '8693AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:cb',
                                         'uptime': '2082693'},
                                 {       'counter_source': '141957/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141957/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8524470669',
                                                                       'in_ucast_pkts': '85189087',
                                                                       'mac_address': 'd0:76:8f:08:17:43',
                                                                       'name': '141957/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '263731270759',
                                                                       'out_ucast_pkts': '233840961',
                                                                       'poller_hash': 'ODY5M0RFcGVlcmludGYxNDE5NTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141957',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '263731270759',
                                                                       'in_ucast_pkts': '233840961',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '263731270759',
                                                                                                'in_ucast_pkts': '233840961',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '8524470669',
                                                                                                'out_ucast_pkts': '85189087',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8524470669',
                                                                       'out_ucast_pkts': '85189087',
                                                                       'poller_hash': 'ODY5M0RFcGVlcmludGZwb24gMTQxOTU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141957',
                                                                                               'links': [       {       'connected_time': '21817170',
                                                                                                                        'fiber_length': 9364,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0RFcGVlcnBvbnBvbiAxNDE5NTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4977 PERTH LINE 55 (MOGK BROTHERS)',
                                         'poller_hash': 'ODY5M0RF',
                                         'serial_number': '8693DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:43',
                                         'uptime': '21817170'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '141987/g1',
                                                                       'name': '141987/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5M0JFcGVlcmludGYxNDE5ODcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141987',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODY5M0JFcGVlcmludGZwb24gMTQxOTg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141987',
                                                                                               'links': [       {       'connected_time': '21816516',
                                                                                                                        'fiber_length': 8718,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0JFcGVlcnBvbnBvbiAxNDE5ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4994 PERTH LINE 55 (ZEHR SUSAN)',
                                         'poller_hash': 'ODY5M0JF',
                                         'serial_number': '8693BE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:43',
                                         'uptime': '21816516'},
                                 {       'counter_source': '141997/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141997/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '258728136432',
                                                                       'in_ucast_pkts': '978971078',
                                                                       'mac_address': 'd0:76:8f:08:16:c3',
                                                                       'name': '141997/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '3118605408463',
                                                                       'out_ucast_pkts': '2277587403',
                                                                       'poller_hash': 'ODY5M0NFcGVlcmludGYxNDE5OTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141997',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '3118605408463',
                                                                       'in_ucast_pkts': '2277587403',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '3118605408463',
                                                                                                'in_ucast_pkts': '2277587403',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '258728136432',
                                                                                                'out_ucast_pkts': '978971078',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '258728136432',
                                                                       'out_ucast_pkts': '978971078',
                                                                       'poller_hash': 'ODY5M0NFcGVlcmludGZwb24gMTQxOTk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141997',
                                                                                               'links': [       {       'connected_time': '538561',
                                                                                                                        'fiber_length': 14444,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0NFcGVlcnBvbnBvbiAxNDE5OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4868 LINE 49 (DEGROOT TY)',
                                         'poller_hash': 'ODY5M0NF',
                                         'serial_number': '8693CE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:c3',
                                         'uptime': '538561'},
                                 {       'counter_source': '142007/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142007/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '70062464910',
                                                                       'in_ucast_pkts': '323705434',
                                                                       'mac_address': 'd0:76:8f:08:18:53',
                                                                       'name': '142007/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '94',
                                                                       'out_octets': '1731312572121',
                                                                       'out_ucast_pkts': '1426996547',
                                                                       'poller_hash': 'ODY5NDAwcGVlcmludGYxNDIwMDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142007',
                                                                       'in_nucast_pkts': '94',
                                                                       'in_octets': '1731312572121',
                                                                       'in_ucast_pkts': '1426996547',
                                                                       'links': [       {       'in_nucast_pkts': '94',
                                                                                                'in_octets': '1731312572121',
                                                                                                'in_ucast_pkts': '1426996547',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '70062464910',
                                                                                                'out_ucast_pkts': '323705434',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '70062464910',
                                                                       'out_ucast_pkts': '323705434',
                                                                       'poller_hash': 'ODY5NDAwcGVlcmludGZwb24gMTQyMDA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142007',
                                                                                               'links': [       {       'connected_time': '688554',
                                                                                                                        'fiber_length': 5325,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDAwcGVlcnBvbnBvbiAxNDIwMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4631 PERTH LINE 55 (ROPP TITUS)',
                                         'poller_hash': 'ODY5NDAw',
                                         'serial_number': '869400',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:53',
                                         'uptime': '688554'},
                                 {       'counter_source': '142027/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142027/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '81812068863',
                                                                       'in_ucast_pkts': '241629595',
                                                                       'mac_address': 'd0:76:8f:08:15:5b',
                                                                       'name': '142027/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '113',
                                                                       'out_octets': '913452489300',
                                                                       'out_ucast_pkts': '772688256',
                                                                       'poller_hash': 'ODY5M0ExcGVlcmludGYxNDIwMjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142027',
                                                                       'in_nucast_pkts': '113',
                                                                       'in_octets': '913452489300',
                                                                       'in_ucast_pkts': '772688256',
                                                                       'links': [       {       'in_nucast_pkts': '113',
                                                                                                'in_octets': '913452489300',
                                                                                                'in_ucast_pkts': '772688256',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '81812068863',
                                                                                                'out_ucast_pkts': '241629595',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '81812068863',
                                                                       'out_ucast_pkts': '241629595',
                                                                       'poller_hash': 'ODY5M0ExcGVlcmludGZwb24gMTQyMDI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142027',
                                                                                               'links': [       {       'connected_time': '21816183',
                                                                                                                        'fiber_length': 3645,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0ExcGVlcnBvbnBvbiAxNDIwMjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6250 PERTH ROAD 131 (RAYCRAFT KEN & '
                                                 'DARLENE)',
                                         'poller_hash': 'ODY5M0Ex',
                                         'serial_number': '8693A1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:5b',
                                         'uptime': '21816183'},
                                 {       'counter_source': '142047/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:8e:a7'],
                                                                       'id': '142047/g1',
                                                                       'in_nucast_pkts': '38076',
                                                                       'in_octets': '2662733693',
                                                                       'in_ucast_pkts': '7297399',
                                                                       'mac_address': 'ec:4f:82:4f:77:c2',
                                                                       'mtu': 2000,
                                                                       'name': '142047/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4383101',
                                                                       'out_octets': '10215709746',
                                                                       'out_ucast_pkts': '12038643',
                                                                       'poller_hash': 'Mjk4MUZBcGVlcmludGYxNDIwNDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142047',
                                                                       'in_nucast_pkts': '4383101',
                                                                       'in_octets': '10215709746',
                                                                       'in_ucast_pkts': '12038643',
                                                                       'links': [       {       'in_nucast_pkts': '4383101',
                                                                                                'in_octets': '10215709746',
                                                                                                'in_ucast_pkts': '12038643',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '38076',
                                                                                                'out_octets': '2662733693',
                                                                                                'out_ucast_pkts': '7297399',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38076',
                                                                       'out_octets': '2662733693',
                                                                       'out_ucast_pkts': '7297399',
                                                                       'poller_hash': 'Mjk4MUZBcGVlcmludGZwb24gMTQyMDQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142047',
                                                                                               'links': [       {       'connected_time': '2024171',
                                                                                                                        'fiber_length': 1047,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MUZBcGVlcnBvbnBvbiAxNDIwNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 JOHN ST (HURAS JENNIFER)',
                                         'poller_hash': 'Mjk4MUZB',
                                         'serial_number': '2981FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:77:bf',
                                         'uptime': '2024171'},
                                 {       'counter_source': '142068/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142068/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '333337637261',
                                                                       'in_ucast_pkts': '2422220358',
                                                                       'mac_address': 'd0:76:8f:08:18:63',
                                                                       'name': '142068/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '56',
                                                                       'out_octets': '7654755595324',
                                                                       'out_ucast_pkts': '5393438052',
                                                                       'poller_hash': 'ODY5NDAycGVlcmludGYxNDIwNjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142068',
                                                                       'in_nucast_pkts': '56',
                                                                       'in_octets': '7654755595324',
                                                                       'in_ucast_pkts': '5393438052',
                                                                       'links': [       {       'in_nucast_pkts': '56',
                                                                                                'in_octets': '7654755595324',
                                                                                                'in_ucast_pkts': '5393438052',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '333337637261',
                                                                                                'out_ucast_pkts': '2422220358',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '333337637261',
                                                                       'out_ucast_pkts': '2422220358',
                                                                       'poller_hash': 'ODY5NDAycGVlcmludGZwb24gMTQyMDY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142068',
                                                                                               'links': [       {       'connected_time': '30937693',
                                                                                                                        'fiber_length': 8775,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDAycGVlcnBvbnBvbiAxNDIwNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4561 LINE 61 (LUPTON MACKENZIE '
                                                 '(MAC))',
                                         'poller_hash': 'ODY5NDAy',
                                         'serial_number': '869402',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:63',
                                         'uptime': '30937693'},
                                 {       'counter_source': '142078/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142078/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '139093459877',
                                                                       'in_ucast_pkts': '5164304904',
                                                                       'mac_address': 'd0:76:8f:08:16:e3',
                                                                       'name': '142078/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '4451293077739',
                                                                       'out_ucast_pkts': '3384813338',
                                                                       'poller_hash': 'ODY5M0QycGVlcmludGYxNDIwNzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142078',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '4451293077739',
                                                                       'in_ucast_pkts': '3384813338',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '4451293077739',
                                                                                                'in_ucast_pkts': '3384813338',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '139093459877',
                                                                                                'out_ucast_pkts': '5164304904',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '139093459877',
                                                                       'out_ucast_pkts': '5164304904',
                                                                       'poller_hash': 'ODY5M0QycGVlcmludGZwb24gMTQyMDc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142078',
                                                                                               'links': [       {       'connected_time': '21816880',
                                                                                                                        'fiber_length': 10386,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0QycGVlcnBvbnBvbiAxNDIwNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4861 PERTH LINE 55 (ZOOK EPHRAIM)',
                                         'poller_hash': 'ODY5M0Qy',
                                         'serial_number': '8693D2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:e3',
                                         'uptime': '21816880'},
                                 {       'counter_source': '142088/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142088/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '86128167253',
                                                                       'in_ucast_pkts': '321919856',
                                                                       'mac_address': 'd0:76:8f:08:16:33',
                                                                       'name': '142088/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '86',
                                                                       'out_octets': '1494857790813',
                                                                       'out_ucast_pkts': '1223295203',
                                                                       'poller_hash': 'ODY5M0JDcGVlcmludGYxNDIwODgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142088',
                                                                       'in_nucast_pkts': '86',
                                                                       'in_octets': '1494857790813',
                                                                       'in_ucast_pkts': '1223295203',
                                                                       'links': [       {       'in_nucast_pkts': '86',
                                                                                                'in_octets': '1494857790813',
                                                                                                'in_ucast_pkts': '1223295203',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '86128167253',
                                                                                                'out_ucast_pkts': '321919856',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '86128167253',
                                                                       'out_ucast_pkts': '321919856',
                                                                       'poller_hash': 'ODY5M0JDcGVlcmludGZwb24gMTQyMDg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142088',
                                                                                               'links': [       {       'connected_time': '28537793',
                                                                                                                        'fiber_length': 8610,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0JDcGVlcnBvbnBvbiAxNDIwODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4540 LINE 61 (BUCHAN SCOTT)',
                                         'poller_hash': 'ODY5M0JD',
                                         'serial_number': '8693BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:33',
                                         'uptime': '28537793'},
                                 {       'counter_source': '142138/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142138/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1994828769929',
                                                                       'in_ucast_pkts': '5482127712',
                                                                       'mac_address': 'd0:76:8f:08:16:53',
                                                                       'name': '142138/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '422',
                                                                       'out_octets': '15838275637928',
                                                                       'out_ucast_pkts': '12491823706',
                                                                       'poller_hash': 'ODY5M0MwcGVlcmludGYxNDIxMzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142138',
                                                                       'in_nucast_pkts': '422',
                                                                       'in_octets': '15838275637928',
                                                                       'in_ucast_pkts': '12491823706',
                                                                       'links': [       {       'in_nucast_pkts': '422',
                                                                                                'in_octets': '15838275637928',
                                                                                                'in_ucast_pkts': '12491823706',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1994828769929',
                                                                                                'out_ucast_pkts': '5482127712',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1994828769929',
                                                                       'out_ucast_pkts': '5482127712',
                                                                       'poller_hash': 'ODY5M0MwcGVlcmludGZwb24gMTQyMTM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142138',
                                                                                               'links': [       {       'connected_time': '9339790',
                                                                                                                        'fiber_length': 10460,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0MwcGVlcnBvbnBvbiAxNDIxMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5592 ROAD 126 (KOBUS DAN & BECCA)',
                                         'poller_hash': 'ODY5M0Mw',
                                         'serial_number': '8693C0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:53',
                                         'uptime': '9339790'},
                                 {       'counter_source': '142244/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142244/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '19639727250',
                                                                       'in_ucast_pkts': '138626751',
                                                                       'mac_address': 'd0:76:8f:08:15:fb',
                                                                       'name': '142244/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '797490008257',
                                                                       'out_ucast_pkts': '576831432',
                                                                       'poller_hash': 'ODY5M0I1cGVlcmludGYxNDIyNDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142244',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '797490008257',
                                                                       'in_ucast_pkts': '576831432',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '797490008257',
                                                                                                'in_ucast_pkts': '576831432',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '19639727250',
                                                                                                'out_ucast_pkts': '138626751',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '19639727250',
                                                                       'out_ucast_pkts': '138626751',
                                                                       'poller_hash': 'ODY5M0I1cGVlcmludGZwb24gMTQyMjQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142244',
                                                                                               'links': [       {       'connected_time': '18864638',
                                                                                                                        'fiber_length': 10604,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0I1cGVlcnBvbnBvbiAxNDIyNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5940 PERTH ROAD 119 (FRIJTERS '
                                                 'MARIETTE)',
                                         'poller_hash': 'ODY5M0I1',
                                         'serial_number': '8693B5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:fb',
                                         'uptime': '18864638'},
                                 {       'counter_source': '142254/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142254/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '58183189697',
                                                                       'in_ucast_pkts': '259479199',
                                                                       'mac_address': 'd0:76:8f:08:16:8b',
                                                                       'name': '142254/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1088240746343',
                                                                       'out_ucast_pkts': '845403095',
                                                                       'poller_hash': 'ODY5M0M3cGVlcmludGYxNDIyNTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142254',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1088240746343',
                                                                       'in_ucast_pkts': '845403095',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1088240746343',
                                                                                                'in_ucast_pkts': '845403095',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '58183189697',
                                                                                                'out_ucast_pkts': '259479199',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '58183189697',
                                                                       'out_ucast_pkts': '259479199',
                                                                       'poller_hash': 'ODY5M0M3cGVlcmludGZwb24gMTQyMjU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142254',
                                                                                               'links': [       {       'connected_time': '688346',
                                                                                                                        'fiber_length': 10585,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0M3cGVlcnBvbnBvbiAxNDIyNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5940 PERTH ROAD 119 (HUBER RACHEL)',
                                         'poller_hash': 'ODY5M0M3',
                                         'serial_number': '8693C7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:8b',
                                         'uptime': '688346'},
                                 {       'counter_source': '142264/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e9:26'],
                                                                       'id': '142264/g1',
                                                                       'in_nucast_pkts': '752945',
                                                                       'in_octets': '119660496513',
                                                                       'in_ucast_pkts': '1083768890',
                                                                       'mac_address': 'd0:76:8f:08:17:6e',
                                                                       'mtu': 2000,
                                                                       'name': '142264/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '95001851',
                                                                       'out_octets': '6167832674671',
                                                                       'out_ucast_pkts': '4226557460',
                                                                       'poller_hash': 'ODY5M0UzcGVlcmludGYxNDIyNjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142264',
                                                                       'in_nucast_pkts': '95001851',
                                                                       'in_octets': '6167832674671',
                                                                       'in_ucast_pkts': '4226557460',
                                                                       'links': [       {       'in_nucast_pkts': '95001851',
                                                                                                'in_octets': '6167832674671',
                                                                                                'in_ucast_pkts': '4226557460',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '752945',
                                                                                                'out_octets': '119660496513',
                                                                                                'out_ucast_pkts': '1083768890',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '752945',
                                                                       'out_octets': '119660496513',
                                                                       'out_ucast_pkts': '1083768890',
                                                                       'poller_hash': 'ODY5M0UzcGVlcmludGZwb24gMTQyMjY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142264',
                                                                                               'links': [       {       'connected_time': '71368108',
                                                                                                                        'fiber_length': 10574,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0UzcGVlcnBvbnBvbiAxNDIyNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5940 PERTH ROAD 119 (BANNON JOANNE)',
                                         'poller_hash': 'ODY5M0Uz',
                                         'serial_number': '8693E3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:6b',
                                         'uptime': '71368108'},
                                 {       'counter_source': '142286/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142286/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '85128767875',
                                                                       'in_ucast_pkts': '343233681',
                                                                       'mac_address': 'd0:76:8f:08:17:53',
                                                                       'name': '142286/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1538863338824',
                                                                       'out_ucast_pkts': '1167879766',
                                                                       'poller_hash': 'ODY5M0UwcGVlcmludGYxNDIyODYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142286',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1538863338824',
                                                                       'in_ucast_pkts': '1167879766',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1538863338824',
                                                                                                'in_ucast_pkts': '1167879766',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '85128767875',
                                                                                                'out_ucast_pkts': '343233681',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '85128767875',
                                                                       'out_ucast_pkts': '343233681',
                                                                       'poller_hash': 'ODY5M0UwcGVlcmludGZwb24gMTQyMjg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142286',
                                                                                               'links': [       {       'connected_time': '688349',
                                                                                                                        'fiber_length': 10461,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0UwcGVlcnBvbnBvbiAxNDIyODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6127 ANA ST (NAFZIGER KEN)',
                                         'poller_hash': 'ODY5M0Uw',
                                         'serial_number': '8693E0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:53',
                                         'uptime': '688349'},
                                 {       'counter_source': '142296/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142296/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '55847101573',
                                                                       'in_ucast_pkts': '231915260',
                                                                       'mac_address': 'd0:76:8f:08:16:d3',
                                                                       'name': '142296/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1065560722477',
                                                                       'out_ucast_pkts': '811391259',
                                                                       'poller_hash': 'ODY5M0QwcGVlcmludGYxNDIyOTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142296',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1065560722477',
                                                                       'in_ucast_pkts': '811391259',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1065560722477',
                                                                                                'in_ucast_pkts': '811391259',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '55847101573',
                                                                                                'out_ucast_pkts': '231915260',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '55847101573',
                                                                       'out_ucast_pkts': '231915260',
                                                                       'poller_hash': 'ODY5M0QwcGVlcmludGZwb24gMTQyMjk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142296',
                                                                                               'links': [       {       'connected_time': '6552385',
                                                                                                                        'fiber_length': 10442,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0QwcGVlcnBvbnBvbiAxNDIyOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6112 ANA ST (CRESSMAN STEVE & JUDY)',
                                         'poller_hash': 'ODY5M0Qw',
                                         'serial_number': '8693D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:d3',
                                         'uptime': '6552385'},
                                 {       'counter_source': '142306/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:95:5e'],
                                                                       'id': '142306/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '5170669744',
                                                                       'in_ucast_pkts': '18849813',
                                                                       'mac_address': 'd0:76:8f:08:18:8e',
                                                                       'mtu': 2000,
                                                                       'name': '142306/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464479',
                                                                       'out_octets': '99836235957',
                                                                       'out_ucast_pkts': '79273371',
                                                                       'poller_hash': 'ODY5NDA3cGVlcmludGYxNDIzMDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142306',
                                                                       'in_nucast_pkts': '1464479',
                                                                       'in_octets': '99836235957',
                                                                       'in_ucast_pkts': '79273371',
                                                                       'links': [       {       'in_nucast_pkts': '1464479',
                                                                                                'in_octets': '99836235957',
                                                                                                'in_ucast_pkts': '79273371',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '5170669744',
                                                                                                'out_ucast_pkts': '18849813',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '5170669744',
                                                                       'out_ucast_pkts': '18849813',
                                                                       'poller_hash': 'ODY5NDA3cGVlcmludGZwb24gMTQyMzA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142306',
                                                                                               'links': [       {       'connected_time': '688478',
                                                                                                                        'fiber_length': 1669,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDA3cGVlcnBvbnBvbiAxNDIzMDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 WILLIAM ST (GRAVES ANDREA)',
                                         'poller_hash': 'ODY5NDA3',
                                         'serial_number': '869407',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:8b',
                                         'uptime': '688478'},
                                 {       'counter_source': '142316/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142316/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7035734623',
                                                                       'in_ucast_pkts': '41494881',
                                                                       'mac_address': 'd0:76:8f:08:18:73',
                                                                       'name': '142316/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '224869884836',
                                                                       'out_ucast_pkts': '169078705',
                                                                       'poller_hash': 'ODY5NDA0cGVlcmludGYxNDIzMTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142316',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '224869884836',
                                                                       'in_ucast_pkts': '169078705',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '224869884836',
                                                                                                'in_ucast_pkts': '169078705',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '7035734623',
                                                                                                'out_ucast_pkts': '41494881',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7035734623',
                                                                       'out_ucast_pkts': '41494881',
                                                                       'poller_hash': 'ODY5NDA0cGVlcmludGZwb24gMTQyMzE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142316',
                                                                                               'links': [       {       'connected_time': '688380',
                                                                                                                        'fiber_length': 10495,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDA0cGVlcnBvbnBvbiAxNDIzMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5916 PERTH ROAD 119 (ROTH DERRICK)',
                                         'poller_hash': 'ODY5NDA0',
                                         'serial_number': '869404',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:73',
                                         'uptime': '688380'},
                                 {       'counter_source': '142336/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142336/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '41142730647',
                                                                       'in_ucast_pkts': '161134276',
                                                                       'mac_address': 'd0:76:8f:08:17:23',
                                                                       'name': '142336/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '583938080550',
                                                                       'out_ucast_pkts': '435761469',
                                                                       'poller_hash': 'ODY5M0RBcGVlcmludGYxNDIzMzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142336',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '583938080550',
                                                                       'in_ucast_pkts': '435761469',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '583938080550',
                                                                                                'in_ucast_pkts': '435761469',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '41142730647',
                                                                                                'out_ucast_pkts': '161134276',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '41142730647',
                                                                       'out_ucast_pkts': '161134276',
                                                                       'poller_hash': 'ODY5M0RBcGVlcmludGZwb24gMTQyMzM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142336',
                                                                                               'links': [       {       'connected_time': '688353',
                                                                                                                        'fiber_length': 10377,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0RBcGVlcnBvbnBvbiAxNDIzMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5922 PERTH ROAD 119 (NAFZIGER TOM & '
                                                 'ALYCIA)',
                                         'poller_hash': 'ODY5M0RB',
                                         'serial_number': '8693DA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:23',
                                         'uptime': '688353'},
                                 {       'counter_source': '142366/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142366/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '111949369665',
                                                                       'in_ucast_pkts': '47381067021',
                                                                       'mac_address': 'd0:76:8f:08:1b:eb',
                                                                       'name': '142366/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '264410448598',
                                                                       'out_ucast_pkts': '255515672',
                                                                       'poller_hash': 'ODY5NDczcGVlcmludGYxNDIzNjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142366',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '264410448598',
                                                                       'in_ucast_pkts': '255515672',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '264410448598',
                                                                                                'in_ucast_pkts': '255515672',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '111949369665',
                                                                                                'out_ucast_pkts': '47381067021',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '111949369665',
                                                                       'out_ucast_pkts': '47381067021',
                                                                       'poller_hash': 'ODY5NDczcGVlcmludGZwb24gMTQyMzY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142366',
                                                                                               'links': [       {       'connected_time': '13582346',
                                                                                                                        'fiber_length': 10555,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDczcGVlcnBvbnBvbiAxNDIzNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5929 SOMERSET DR (GERBER TOM & '
                                                 'SHARON)',
                                         'poller_hash': 'ODY5NDcz',
                                         'serial_number': '869473',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1b:eb',
                                         'uptime': '13582346'},
                                 {       'counter_source': '142396/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142396/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '98126934260',
                                                                       'in_ucast_pkts': '277737030',
                                                                       'mac_address': 'd0:76:8f:08:18:2b',
                                                                       'name': '142396/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1176580628118',
                                                                       'out_ucast_pkts': '854856246',
                                                                       'poller_hash': 'ODY5M0ZCcGVlcmludGYxNDIzOTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142396',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1176580628118',
                                                                       'in_ucast_pkts': '854856246',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1176580628118',
                                                                                                'in_ucast_pkts': '854856246',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '98126934260',
                                                                                                'out_ucast_pkts': '277737030',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '98126934260',
                                                                       'out_ucast_pkts': '277737030',
                                                                       'poller_hash': 'ODY5M0ZCcGVlcmludGZwb24gMTQyMzk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142396',
                                                                                               'links': [       {       'connected_time': '688346',
                                                                                                                        'fiber_length': 10471,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0ZCcGVlcnBvbnBvbiAxNDIzOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6108 ANA ST (DUNN ROBIN & WENDY)',
                                         'poller_hash': 'ODY5M0ZC',
                                         'serial_number': '8693FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:2b',
                                         'uptime': '688346'},
                                 {       'counter_source': '142406/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142406/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10015870208',
                                                                       'in_ucast_pkts': '4312390751',
                                                                       'mac_address': 'd0:76:8f:08:14:73',
                                                                       'name': '142406/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '92979438877',
                                                                       'out_ucast_pkts': '87314502',
                                                                       'poller_hash': 'ODY5Mzg0cGVlcmludGYxNDI0MDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142406',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '92979438877',
                                                                       'in_ucast_pkts': '87314502',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '92979438877',
                                                                                                'in_ucast_pkts': '87314502',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '10015870208',
                                                                                                'out_ucast_pkts': '4312390751',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10015870208',
                                                                       'out_ucast_pkts': '4312390751',
                                                                       'poller_hash': 'ODY5Mzg0cGVlcmludGZwb24gMTQyNDA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142406',
                                                                                               'links': [       {       'connected_time': '18660841',
                                                                                                                        'fiber_length': 10582,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5Mzg0cGVlcnBvbnBvbiAxNDI0MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5922 SOMERSET DR (WAGLER CHRISTOPHER '
                                                 '& ROSALIND)',
                                         'poller_hash': 'ODY5Mzg0',
                                         'serial_number': '869384',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:14:73',
                                         'uptime': '18660841'},
                                 {       'counter_source': '142416/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142416/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7362389747',
                                                                       'in_ucast_pkts': '36232489',
                                                                       'mac_address': 'd0:76:8f:08:15:b3',
                                                                       'name': '142416/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '223759019455',
                                                                       'out_ucast_pkts': '175045468',
                                                                       'poller_hash': 'ODY5M0FDcGVlcmludGYxNDI0MTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142416',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '223759019455',
                                                                       'in_ucast_pkts': '175045468',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '223759019455',
                                                                                                'in_ucast_pkts': '175045468',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '7362389747',
                                                                                                'out_ucast_pkts': '36232489',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7362389747',
                                                                       'out_ucast_pkts': '36232489',
                                                                       'poller_hash': 'ODY5M0FDcGVlcmludGZwb24gMTQyNDE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142416',
                                                                                               'links': [       {       'connected_time': '21815921',
                                                                                                                        'fiber_length': 10536,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0FDcGVlcnBvbnBvbiAxNDI0MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10400848 (GERBER STAN & ERMA)',
                                         'poller_hash': 'ODY5M0FD',
                                         'serial_number': '8693AC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:b3',
                                         'uptime': '21815921'},
                                 {       'counter_source': '142456/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2f:2d:12'],
                                                                       'id': '142456/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '1435108300',
                                                                       'in_ucast_pkts': '11252384',
                                                                       'mac_address': 'd0:76:8f:08:15:8e',
                                                                       'mtu': 2000,
                                                                       'name': '142456/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1462378',
                                                                       'out_octets': '73804079495',
                                                                       'out_ucast_pkts': '51059746',
                                                                       'poller_hash': 'ODY5M0E3cGVlcmludGYxNDI0NTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142456',
                                                                       'in_nucast_pkts': '1462378',
                                                                       'in_octets': '73804079495',
                                                                       'in_ucast_pkts': '51059746',
                                                                       'links': [       {       'in_nucast_pkts': '1462378',
                                                                                                'in_octets': '73804079495',
                                                                                                'in_ucast_pkts': '51059746',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '1435108300',
                                                                                                'out_ucast_pkts': '11252384',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '1435108300',
                                                                       'out_ucast_pkts': '11252384',
                                                                       'poller_hash': 'ODY5M0E3cGVlcmludGZwb24gMTQyNDU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142456',
                                                                                               'links': [       {       'connected_time': '688228',
                                                                                                                        'fiber_length': 1698,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0E3cGVlcnBvbnBvbiAxNDI0NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 WILLIAM ST (CONLAN MANDY)',
                                         'poller_hash': 'ODY5M0E3',
                                         'serial_number': '8693A7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:8b',
                                         'uptime': '688228'},
                                 {       'counter_source': '142466/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142466/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '35198371045',
                                                                       'in_ucast_pkts': '82263663',
                                                                       'mac_address': 'd0:76:8f:08:15:e3',
                                                                       'name': '142466/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '358224715890',
                                                                       'out_ucast_pkts': '299747892',
                                                                       'poller_hash': 'ODY5M0IycGVlcmludGYxNDI0NjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142466',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '358224715890',
                                                                       'in_ucast_pkts': '299747892',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '358224715890',
                                                                                                'in_ucast_pkts': '299747892',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '35198371045',
                                                                                                'out_ucast_pkts': '82263663',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '35198371045',
                                                                       'out_ucast_pkts': '82263663',
                                                                       'poller_hash': 'ODY5M0IycGVlcmludGZwb24gMTQyNDY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142466',
                                                                                               'links': [       {       'connected_time': '688353',
                                                                                                                        'fiber_length': 10593,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0IycGVlcnBvbnBvbiAxNDI0NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5931 SOMERSET DR (LICHTY HOWARD)',
                                         'poller_hash': 'ODY5M0Iy',
                                         'serial_number': '8693B2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:e3',
                                         'uptime': '688353'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '142476/g1',
                                                                       'name': '142476/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5NDJGcGVlcmludGYxNDI0NzYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142476',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODY5NDJGcGVlcmludGZwb24gMTQyNDc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142476',
                                                                                               'links': [       {       'connected_time': '688354',
                                                                                                                        'fiber_length': 10373,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDJGcGVlcnBvbnBvbiAxNDI0NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6109 ANA ST (WALL GEORGE (ESTATE '
                                                 'OF))',
                                         'poller_hash': 'ODY5NDJG',
                                         'serial_number': '86942F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:cb',
                                         'uptime': '688354'},
                                 {       'counter_source': '142496/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142496/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '101535944997',
                                                                       'in_ucast_pkts': '9168315573',
                                                                       'mac_address': '04:bc:9f:5c:18:f9',
                                                                       'name': '142496/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3609851534314',
                                                                       'out_ucast_pkts': '2545108758',
                                                                       'poller_hash': 'MTI3NDk5RnBlZXJpbnRmMTQyNDk2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '142496',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3609851534314',
                                                                       'in_ucast_pkts': '2545108758',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3609851534314',
                                                                                                'in_ucast_pkts': '2545108758',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '101535944997',
                                                                                                'out_ucast_pkts': '9168315573',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '101535944997',
                                                                       'out_ucast_pkts': '9168315573',
                                                                       'poller_hash': 'MTI3NDk5RnBlZXJpbnRmcG9uIDE0MjQ5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142496',
                                                                                               'links': [       {       'connected_time': '4060597',
                                                                                                                        'fiber_length': 10552,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk5RnBlZXJwb25wb24gMTQyNDk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6143 ANA ST (NAFZIGER CHRIS)',
                                         'poller_hash': 'MTI3NDk5Rg==',
                                         'serial_number': '127499F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:18:f9',
                                         'uptime': '4060597'},
                                 {       'counter_source': '142506/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1c:67'],
                                                                       'id': '142506/g1',
                                                                       'in_nucast_pkts': '14763',
                                                                       'in_octets': '6407478930',
                                                                       'in_ucast_pkts': '24295667',
                                                                       'mac_address': 'd0:76:8f:08:15:7e',
                                                                       'mtu': 2000,
                                                                       'name': '142506/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464403',
                                                                       'out_octets': '91063223357',
                                                                       'out_ucast_pkts': '69457078',
                                                                       'poller_hash': 'ODY5M0E1cGVlcmludGYxNDI1MDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142506',
                                                                       'in_nucast_pkts': '1464403',
                                                                       'in_octets': '91063223357',
                                                                       'in_ucast_pkts': '69457078',
                                                                       'links': [       {       'in_nucast_pkts': '1464403',
                                                                                                'in_octets': '91063223357',
                                                                                                'in_ucast_pkts': '69457078',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14763',
                                                                                                'out_octets': '6407478930',
                                                                                                'out_ucast_pkts': '24295667',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14763',
                                                                       'out_octets': '6407478930',
                                                                       'out_ucast_pkts': '24295667',
                                                                       'poller_hash': 'ODY5M0E1cGVlcmludGZwb24gMTQyNTA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142506',
                                                                                               'links': [       {       'connected_time': '688356',
                                                                                                                        'fiber_length': 10370,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0E1cGVlcnBvbnBvbiAxNDI1MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5597 DUKE ST (HIGGS TERRY)',
                                         'poller_hash': 'ODY5M0E1',
                                         'serial_number': '8693A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:7b',
                                         'uptime': '688356'},
                                 {       'counter_source': '142556/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142556/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '31620576089',
                                                                       'in_ucast_pkts': '166688457',
                                                                       'mac_address': 'd0:76:8f:08:17:1b',
                                                                       'name': '142556/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '814883807358',
                                                                       'out_ucast_pkts': '600510609',
                                                                       'poller_hash': 'ODY5M0Q5cGVlcmludGYxNDI1NTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142556',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '814883807358',
                                                                       'in_ucast_pkts': '600510609',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '814883807358',
                                                                                                'in_ucast_pkts': '600510609',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '31620576089',
                                                                                                'out_ucast_pkts': '166688457',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '31620576089',
                                                                       'out_ucast_pkts': '166688457',
                                                                       'poller_hash': 'ODY5M0Q5cGVlcmludGZwb24gMTQyNTU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142556',
                                                                                               'links': [       {       'connected_time': '688349',
                                                                                                                        'fiber_length': 10566,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0Q5cGVlcnBvbnBvbiAxNDI1NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5934 PERTH ROAD 119 (YANTZI RHODA)',
                                         'poller_hash': 'ODY5M0Q5',
                                         'serial_number': '8693D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:1b',
                                         'uptime': '688349'},
                                 {       'counter_source': '142605/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142605/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12139013374',
                                                                       'in_ucast_pkts': '131607262',
                                                                       'mac_address': 'd0:76:8f:08:16:1b',
                                                                       'name': '142605/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '387120254706',
                                                                       'out_ucast_pkts': '269441326',
                                                                       'poller_hash': 'ODY5M0I5cGVlcmludGYxNDI2MDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142605',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '387120254706',
                                                                       'in_ucast_pkts': '269441326',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '387120254706',
                                                                                                'in_ucast_pkts': '269441326',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '12139013374',
                                                                                                'out_ucast_pkts': '131607262',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12139013374',
                                                                       'out_ucast_pkts': '131607262',
                                                                       'poller_hash': 'ODY5M0I5cGVlcmludGZwb24gMTQyNjA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142605',
                                                                                               'links': [       {       'connected_time': '71555921',
                                                                                                                        'fiber_length': 10498,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0I5cGVlcnBvbnBvbiAxNDI2MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5923 SOMERSET DR (NAFZIGER WAYNE)',
                                         'poller_hash': 'ODY5M0I5',
                                         'serial_number': '8693B9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:1b',
                                         'uptime': '71555921'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '142725/g1',
                                                                       'name': '142725/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODcyNTUzcGVlcmludGYxNDI3MjUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142725',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODcyNTUzcGVlcmludGZwb24gMTQyNzI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142725',
                                                                                               'links': [       {       'connected_time': '19808706',
                                                                                                                        'fiber_length': 10716,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTUzcGVlcnBvbnBvbiAxNDI3MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6202 ROAD 148 (JANTZI SHAWN)',
                                         'poller_hash': 'ODcyNTUz',
                                         'serial_number': '872553',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a8:2b',
                                         'uptime': '19808706'},
                                 {       'counter_source': '142735/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7d:6d'],
                                                                       'id': '142735/g1',
                                                                       'in_nucast_pkts': '11811',
                                                                       'in_octets': '1376929192',
                                                                       'in_ucast_pkts': '5842297',
                                                                       'mac_address': 'd0:76:8f:1b:a7:b6',
                                                                       'mtu': 2000,
                                                                       'name': '142735/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464407',
                                                                       'out_octets': '47129504531',
                                                                       'out_ucast_pkts': '37494054',
                                                                       'poller_hash': 'ODcyNTQ0cGVlcmludGYxNDI3MzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142735',
                                                                       'in_nucast_pkts': '1464407',
                                                                       'in_octets': '47129504531',
                                                                       'in_ucast_pkts': '37494054',
                                                                       'links': [       {       'in_nucast_pkts': '1464407',
                                                                                                'in_octets': '47129504531',
                                                                                                'in_ucast_pkts': '37494054',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11811',
                                                                                                'out_octets': '1376929192',
                                                                                                'out_ucast_pkts': '5842297',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11811',
                                                                       'out_octets': '1376929192',
                                                                       'out_ucast_pkts': '5842297',
                                                                       'poller_hash': 'ODcyNTQ0cGVlcmludGZwb24gMTQyNzM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142735',
                                                                                               'links': [       {       'connected_time': '688477',
                                                                                                                        'fiber_length': 609,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTQ0cGVlcnBvbnBvbiAxNDI3MzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 MAIN ST N APT 2 (STREICHER '
                                                 'DUSTIN)',
                                         'poller_hash': 'ODcyNTQ0',
                                         'serial_number': '872544',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a7:b3',
                                         'uptime': '688477'},
                                 {       'counter_source': '142803/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:0c:d1:09'],
                                                                       'id': '142803/g1',
                                                                       'in_nucast_pkts': '952122',
                                                                       'in_octets': '515850706480',
                                                                       'in_ucast_pkts': '3792982580',
                                                                       'mac_address': 'd0:76:8f:1b:a1:b6',
                                                                       'mtu': 2000,
                                                                       'name': '142803/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149539546',
                                                                       'out_octets': '14737857952500',
                                                                       'out_ucast_pkts': '11090321461',
                                                                       'poller_hash': 'ODcyNDg0cGVlcmludGYxNDI4MDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142803',
                                                                       'in_nucast_pkts': '149539546',
                                                                       'in_octets': '14737857952500',
                                                                       'in_ucast_pkts': '11090321461',
                                                                       'links': [       {       'in_nucast_pkts': '149539546',
                                                                                                'in_octets': '14737857952500',
                                                                                                'in_ucast_pkts': '11090321461',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '952122',
                                                                                                'out_octets': '515850706480',
                                                                                                'out_ucast_pkts': '3792982580',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '952122',
                                                                       'out_octets': '515850706480',
                                                                       'out_ucast_pkts': '3792982580',
                                                                       'poller_hash': 'ODcyNDg0cGVlcmludGZwb24gMTQyODAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142803',
                                                                                               'links': [       {       'connected_time': '21816583',
                                                                                                                        'fiber_length': 10332,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDg0cGVlcnBvbnBvbiAxNDI4MDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6113 ANA ST (ERB DARYL)',
                                         'poller_hash': 'ODcyNDg0',
                                         'serial_number': '872484',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a1:b3',
                                         'uptime': '21816583'},
                                 {       'counter_source': '142813/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142813/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '74227161115',
                                                                       'in_ucast_pkts': '590586616',
                                                                       'mac_address': 'd0:76:8f:1b:a5:c3',
                                                                       'name': '142813/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '3519338421137',
                                                                       'out_ucast_pkts': '2478133928',
                                                                       'poller_hash': 'ODcyNTA2cGVlcmludGYxNDI4MTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142813',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '3519338421137',
                                                                       'in_ucast_pkts': '2478133928',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '3519338421137',
                                                                                                'in_ucast_pkts': '2478133928',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '74227161115',
                                                                                                'out_ucast_pkts': '590586616',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '74227161115',
                                                                       'out_ucast_pkts': '590586616',
                                                                       'poller_hash': 'ODcyNTA2cGVlcmludGZwb24gMTQyODEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142813',
                                                                                               'links': [       {       'connected_time': '538563',
                                                                                                                        'fiber_length': 16833,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTA2cGVlcnBvbnBvbiAxNDI4MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5368 PERTH ROAD 135 (DIEHL BRAD)',
                                         'poller_hash': 'ODcyNTA2',
                                         'serial_number': '872506',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:c3',
                                         'uptime': '538563'},
                                 {       'counter_source': '142833/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f0:17'],
                                                                       'id': '142833/g1',
                                                                       'in_nucast_pkts': '647650',
                                                                       'in_octets': '139201975509',
                                                                       'in_ucast_pkts': '730472217',
                                                                       'mac_address': 'd0:76:8f:1b:a5:ae',
                                                                       'mtu': 2000,
                                                                       'name': '142833/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '92687941',
                                                                       'out_octets': '3453325304789',
                                                                       'out_ucast_pkts': '2598461225',
                                                                       'poller_hash': 'ODcyNTAzcGVlcmludGYxNDI4MzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142833',
                                                                       'in_nucast_pkts': '92687941',
                                                                       'in_octets': '3453325304789',
                                                                       'in_ucast_pkts': '2598461225',
                                                                       'links': [       {       'in_nucast_pkts': '92687941',
                                                                                                'in_octets': '3453325304789',
                                                                                                'in_ucast_pkts': '2598461225',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '647650',
                                                                                                'out_octets': '139201975509',
                                                                                                'out_ucast_pkts': '730472217',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '647650',
                                                                       'out_octets': '139201975509',
                                                                       'out_ucast_pkts': '730472217',
                                                                       'poller_hash': 'ODcyNTAzcGVlcmludGZwb24gMTQyODMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142833',
                                                                                               'links': [       {       'connected_time': '21816400',
                                                                                                                        'fiber_length': 1233,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTAzcGVlcnBvbnBvbiAxNDI4MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 FULTON ST (KOCH JEREMY)',
                                         'poller_hash': 'ODcyNTAz',
                                         'serial_number': '872503',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:ab',
                                         'uptime': '21816400'},
                                 {       'counter_source': '143033/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2d:2f:c8'],
                                                                       'id': '143033/g1',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '26621368086',
                                                                       'in_ucast_pkts': '99649732',
                                                                       'mac_address': 'd0:76:8f:1b:a6:0e',
                                                                       'mtu': 2000,
                                                                       'name': '143033/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463863',
                                                                       'out_octets': '320624456570',
                                                                       'out_ucast_pkts': '252674428',
                                                                       'poller_hash': 'ODcyNTBGcGVlcmludGYxNDMwMzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143033',
                                                                       'in_nucast_pkts': '1463863',
                                                                       'in_octets': '320624456570',
                                                                       'in_ucast_pkts': '252674428',
                                                                       'links': [       {       'in_nucast_pkts': '1463863',
                                                                                                'in_octets': '320624456570',
                                                                                                'in_ucast_pkts': '252674428',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19',
                                                                                                'out_octets': '26621368086',
                                                                                                'out_ucast_pkts': '99649732',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19',
                                                                       'out_octets': '26621368086',
                                                                       'out_ucast_pkts': '99649732',
                                                                       'poller_hash': 'ODcyNTBGcGVlcmludGZwb24gMTQzMDMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143033',
                                                                                               'links': [       {       'connected_time': '688489',
                                                                                                                        'fiber_length': 554,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTBGcGVlcnBvbnBvbiAxNDMwMzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 MAIN ST S (BURGER AELAURA)',
                                         'poller_hash': 'ODcyNTBG',
                                         'serial_number': '87250F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a6:0b',
                                         'uptime': '688489'},
                                 {       'counter_source': '143147/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143147/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '811310843311',
                                                                       'in_ucast_pkts': '10258512801',
                                                                       'mac_address': 'd0:76:8f:1b:a4:e3',
                                                                       'name': '143147/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8150663018680',
                                                                       'out_ucast_pkts': '6113749652',
                                                                       'poller_hash': 'ODcyNEVBcGVlcmludGYxNDMxNDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143147',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8150663018680',
                                                                       'in_ucast_pkts': '6113749652',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '8150663018680',
                                                                                                'in_ucast_pkts': '6113749652',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '811310843311',
                                                                                                'out_ucast_pkts': '10258512801',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '811310843311',
                                                                       'out_ucast_pkts': '10258512801',
                                                                       'poller_hash': 'ODcyNEVBcGVlcmludGZwb24gMTQzMTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143147',
                                                                                               'links': [       {       'connected_time': '688334',
                                                                                                                        'fiber_length': 1122,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNEVBcGVlcnBvbnBvbiAxNDMxNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 MILL ST E (FRESHNUP INC.)',
                                         'poller_hash': 'ODcyNEVB',
                                         'serial_number': '8724EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a4:e3',
                                         'uptime': '688334'},
                                 {       'counter_source': '143167/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143167/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '105898319420',
                                                                       'in_ucast_pkts': '286375361',
                                                                       'mac_address': 'd0:76:8f:1b:9f:bb',
                                                                       'name': '143167/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '175',
                                                                       'out_octets': '815892944920',
                                                                       'out_ucast_pkts': '699377166',
                                                                       'poller_hash': 'ODcyNDQ1cGVlcmludGYxNDMxNjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143167',
                                                                       'in_nucast_pkts': '175',
                                                                       'in_octets': '815892944920',
                                                                       'in_ucast_pkts': '699377166',
                                                                       'links': [       {       'in_nucast_pkts': '175',
                                                                                                'in_octets': '815892944920',
                                                                                                'in_ucast_pkts': '699377166',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '105898319420',
                                                                                                'out_ucast_pkts': '286375361',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '105898319420',
                                                                       'out_ucast_pkts': '286375361',
                                                                       'poller_hash': 'ODcyNDQ1cGVlcmludGZwb24gMTQzMTY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143167',
                                                                                               'links': [       {       'connected_time': '15498918',
                                                                                                                        'fiber_length': 842,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDQ1cGVlcnBvbnBvbiAxNDMxNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 JOHN ST (ALBRECHT MIKE)',
                                         'poller_hash': 'ODcyNDQ1',
                                         'serial_number': '872445',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:bb',
                                         'uptime': '15498918'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '143177/g1',
                                                                       'name': '143177/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkZFQTFEcGVlcmludGYxNDMxNzcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143177',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MkZFQTFEcGVlcmludGZwb24gMTQzMTc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143177',
                                                                                               'links': [       {       'connected_time': '538549',
                                                                                                                        'fiber_length': 13000,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.686,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQTFEcGVlcnBvbnBvbiAxNDMxNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5020 LINE 49 (ROCDALE FARMS)',
                                         'poller_hash': 'MkZFQTFE',
                                         'serial_number': '2FEA1D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:1b:98',
                                         'uptime': '538549'},
                                 {       'counter_source': '143207/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143207/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '66589636596',
                                                                       'in_ucast_pkts': '164878656',
                                                                       'mac_address': 'd0:76:8f:1b:9e:93',
                                                                       'name': '143207/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '376110368002',
                                                                       'out_ucast_pkts': '340422356',
                                                                       'poller_hash': 'ODcyNDIwcGVlcmludGYxNDMyMDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143207',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '376110368002',
                                                                       'in_ucast_pkts': '340422356',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '376110368002',
                                                                                                'in_ucast_pkts': '340422356',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '66589636596',
                                                                                                'out_ucast_pkts': '164878656',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '66589636596',
                                                                       'out_ucast_pkts': '164878656',
                                                                       'poller_hash': 'ODcyNDIwcGVlcmludGZwb24gMTQzMjA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143207',
                                                                                               'links': [       {       'connected_time': '21816353',
                                                                                                                        'fiber_length': 2416,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDIwcGVlcnBvbnBvbiAxNDMyMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "4679 LINE 61 (GJ'S HARVEST CENTRE "
                                                 'INC.)',
                                         'poller_hash': 'ODcyNDIw',
                                         'serial_number': '872420',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9e:93',
                                         'uptime': '21816353'},
                                 {       'counter_source': '143237/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d6:7b'],
                                                                       'id': '143237/g1',
                                                                       'in_nucast_pkts': '17',
                                                                       'in_octets': '60226344667',
                                                                       'in_ucast_pkts': '61716774',
                                                                       'mac_address': 'd0:76:8f:1b:a4:86',
                                                                       'mtu': 2000,
                                                                       'name': '143237/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120669',
                                                                       'out_octets': '117522699364',
                                                                       'out_ucast_pkts': '99951171',
                                                                       'poller_hash': 'ODcyNERFcGVlcmludGYxNDMyMzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143237',
                                                                       'in_nucast_pkts': '1120669',
                                                                       'in_octets': '117522699364',
                                                                       'in_ucast_pkts': '99951171',
                                                                       'links': [       {       'in_nucast_pkts': '1120669',
                                                                                                'in_octets': '117522699364',
                                                                                                'in_ucast_pkts': '99951171',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '17',
                                                                                                'out_octets': '60226344667',
                                                                                                'out_ucast_pkts': '61716774',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17',
                                                                       'out_octets': '60226344667',
                                                                       'out_ucast_pkts': '61716774',
                                                                       'poller_hash': 'ODcyNERFcGVlcmludGZwb24gMTQzMjM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143237',
                                                                                               'links': [       {       'connected_time': '538566',
                                                                                                                        'fiber_length': 16376,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNERFcGVlcnBvbnBvbiAxNDMyMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5376 LINE 49 (RIDLEY RYAN)',
                                         'poller_hash': 'ODcyNERF',
                                         'serial_number': '8724DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a4:83',
                                         'uptime': '538566'},
                                 {       'counter_source': '143762/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143762/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '190314426964',
                                                                       'in_ucast_pkts': '22497905094',
                                                                       'mac_address': 'd0:76:8f:1b:a5:e3',
                                                                       'name': '143762/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '113',
                                                                       'out_octets': '4493122957656',
                                                                       'out_ucast_pkts': '3247798587',
                                                                       'poller_hash': 'ODcyNTBBcGVlcmludGYxNDM3NjIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143762',
                                                                       'in_nucast_pkts': '113',
                                                                       'in_octets': '4493122957656',
                                                                       'in_ucast_pkts': '3247798587',
                                                                       'links': [       {       'in_nucast_pkts': '113',
                                                                                                'in_octets': '4493122957656',
                                                                                                'in_ucast_pkts': '3247798587',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '190314426964',
                                                                                                'out_ucast_pkts': '22497905094',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '190314426964',
                                                                       'out_ucast_pkts': '22497905094',
                                                                       'poller_hash': 'ODcyNTBBcGVlcmludGZwb24gMTQzNzYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143762',
                                                                                               'links': [       {       'connected_time': '9218555',
                                                                                                                        'fiber_length': 14492,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTBBcGVlcnBvbnBvbiAxNDM3NjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5267 ROAD 144 (GRAUL STEVE)',
                                         'poller_hash': 'ODcyNTBB',
                                         'serial_number': '87250A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:e3',
                                         'uptime': '9218555'},
                                 {       'counter_source': '143802/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143802/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '293617744584',
                                                                       'in_ucast_pkts': '2135526991',
                                                                       'mac_address': 'd0:76:8f:1b:9f:3b',
                                                                       'name': '143802/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '9277672478420',
                                                                       'out_ucast_pkts': '6358747766',
                                                                       'poller_hash': 'ODcyNDM1cGVlcmludGYxNDM4MDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143802',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '9277672478420',
                                                                       'in_ucast_pkts': '6358747766',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '9277672478420',
                                                                                                'in_ucast_pkts': '6358747766',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '293617744584',
                                                                                                'out_ucast_pkts': '2135526991',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '293617744584',
                                                                       'out_ucast_pkts': '2135526991',
                                                                       'poller_hash': 'ODcyNDM1cGVlcmludGZwb24gMTQzODAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143802',
                                                                                               'links': [       {       'connected_time': '688383',
                                                                                                                        'fiber_length': 8755,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDM1cGVlcnBvbnBvbiAxNDM4MDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6104 RD 140 (DYCK PETER)',
                                         'poller_hash': 'ODcyNDM1',
                                         'serial_number': '872435',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:3b',
                                         'uptime': '688383'},
                                 {       'counter_source': '143822/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:54:de'],
                                                                       'id': '143822/g1',
                                                                       'in_nucast_pkts': '247805',
                                                                       'in_octets': '46872408502',
                                                                       'in_ucast_pkts': '167420932',
                                                                       'mac_address': 'd0:76:8f:1b:a7:ee',
                                                                       'mtu': 2000,
                                                                       'name': '143822/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '28505071',
                                                                       'out_octets': '1102800800037',
                                                                       'out_ucast_pkts': '851701657',
                                                                       'poller_hash': 'ODcyNTRCcGVlcmludGYxNDM4MjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143822',
                                                                       'in_nucast_pkts': '28505071',
                                                                       'in_octets': '1102800800037',
                                                                       'in_ucast_pkts': '851701657',
                                                                       'links': [       {       'in_nucast_pkts': '28505071',
                                                                                                'in_octets': '1102800800037',
                                                                                                'in_ucast_pkts': '851701657',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '247805',
                                                                                                'out_octets': '46872408502',
                                                                                                'out_ucast_pkts': '167420932',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '247805',
                                                                       'out_octets': '46872408502',
                                                                       'out_ucast_pkts': '167420932',
                                                                       'poller_hash': 'ODcyNTRCcGVlcmludGZwb24gMTQzODIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143822',
                                                                                               'links': [       {       'connected_time': '13219518',
                                                                                                                        'fiber_length': 8085,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTRCcGVlcnBvbnBvbiAxNDM4MjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6161 ROAD 124 (JANTZI BEN)',
                                         'poller_hash': 'ODcyNTRC',
                                         'serial_number': '87254B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a7:eb',
                                         'uptime': '13219518'},
                                 {       'counter_source': '144013/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '144013/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '285308910333',
                                                                       'in_ucast_pkts': '2057289254',
                                                                       'mac_address': 'd0:76:8f:e7:26:9e',
                                                                       'name': '144013/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '8041361624400',
                                                                       'out_ucast_pkts': '5639011386',
                                                                       'poller_hash': 'OUMzMDIzcGVlcmludGYxNDQwMTMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144013',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '8041361624400',
                                                                       'in_ucast_pkts': '5639011386',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '8041361624400',
                                                                                                'in_ucast_pkts': '5639011386',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '285308910333',
                                                                                                'out_ucast_pkts': '2057289254',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '285308910333',
                                                                       'out_ucast_pkts': '2057289254',
                                                                       'poller_hash': 'OUMzMDIzcGVlcmludGZwb24gMTQ0MDEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144013',
                                                                                               'links': [       {       'connected_time': '688501',
                                                                                                                        'fiber_length': 589,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OUMzMDIzcGVlcnBvbnBvbiAxNDQwMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '16 MAIN ST N (MILLIGAN JACKIE)',
                                         'poller_hash': 'OUMzMDIz',
                                         'serial_number': '9C3023',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:e7:26:9e',
                                         'uptime': '688501'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '144486/g1',
                                                                       'name': '144486/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OUMyRUZCcGVlcmludGYxNDQ0ODYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144486',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OUMyRUZCcGVlcmludGZwb24gMTQ0NDg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144486',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OUMyRUZCcGVlcnBvbnBvbiAxNDQ0ODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '13 MAIN ST N (FRIESEN HENRY)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "144486" '
                                                                  '(location: GPON '
                                                                  'port "8/7"), "ONT '
                                                                  'Dying Gasp", '
                                                                  '(suppressed)',
                                         'offline_time': '2025-03-15T18:32:26.380Z',
                                         'poller_hash': 'OUMyRUZC',
                                         'serial_number': '9C2EFB',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '144527/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '144527/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '326513884245',
                                                                       'in_ucast_pkts': '1225329103',
                                                                       'mac_address': 'd0:76:8f:1b:a9:83',
                                                                       'name': '144527/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '160',
                                                                       'out_octets': '5979743588738',
                                                                       'out_ucast_pkts': '4369932925',
                                                                       'poller_hash': 'ODcyNTdFcGVlcmludGYxNDQ1MjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144527',
                                                                       'in_nucast_pkts': '160',
                                                                       'in_octets': '5979743588738',
                                                                       'in_ucast_pkts': '4369932925',
                                                                       'links': [       {       'in_nucast_pkts': '160',
                                                                                                'in_octets': '5979743588738',
                                                                                                'in_ucast_pkts': '4369932925',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '326513884245',
                                                                                                'out_ucast_pkts': '1225329103',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '326513884245',
                                                                       'out_ucast_pkts': '1225329103',
                                                                       'poller_hash': 'ODcyNTdFcGVlcmludGZwb24gMTQ0NTI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144527',
                                                                                               'links': [       {       'connected_time': '688551',
                                                                                                                        'fiber_length': 1716,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTdFcGVlcnBvbnBvbiAxNDQ1Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 MURRAY CT (ALBRECHT NIKKETA)',
                                         'poller_hash': 'ODcyNTdF',
                                         'serial_number': '87257E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a9:83',
                                         'uptime': '688551'},
                                 {       'counter_source': '144958/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '144958/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '297405270811',
                                                                       'in_ucast_pkts': '60575896271',
                                                                       'mac_address': 'd0:76:8f:1b:a6:33',
                                                                       'name': '144958/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '94',
                                                                       'out_octets': '1394028613385',
                                                                       'out_ucast_pkts': '1155969486',
                                                                       'poller_hash': 'ODcyNTE0cGVlcmludGYxNDQ5NTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144958',
                                                                       'in_nucast_pkts': '94',
                                                                       'in_octets': '1394028613385',
                                                                       'in_ucast_pkts': '1155969486',
                                                                       'links': [       {       'in_nucast_pkts': '94',
                                                                                                'in_octets': '1394028613385',
                                                                                                'in_ucast_pkts': '1155969486',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '297405270811',
                                                                                                'out_ucast_pkts': '60575896271',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '297405270811',
                                                                       'out_ucast_pkts': '60575896271',
                                                                       'poller_hash': 'ODcyNTE0cGVlcmludGZwb24gMTQ0OTU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144958',
                                                                                               'links': [       {       'connected_time': '688557',
                                                                                                                        'fiber_length': 1743,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTE0cGVlcnBvbnBvbiAxNDQ5NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 MURRAY CT MILVERTON (RIDDELL '
                                                 'VANESSA)',
                                         'poller_hash': 'ODcyNTE0',
                                         'serial_number': '872514',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a6:33',
                                         'uptime': '688557'},
                                 {       'counter_source': '145289/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e4:c4'],
                                                                       'id': '145289/g1',
                                                                       'in_nucast_pkts': '12367',
                                                                       'in_octets': '37369166849',
                                                                       'in_ucast_pkts': '106189432',
                                                                       'mac_address': 'd0:76:8f:1b:a8:56',
                                                                       'mtu': 2000,
                                                                       'name': '145289/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434299',
                                                                       'out_octets': '436955397390',
                                                                       'out_ucast_pkts': '352970569',
                                                                       'poller_hash': 'ODcyNTU4cGVlcmludGYxNDUyODkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145289',
                                                                       'in_nucast_pkts': '1434299',
                                                                       'in_octets': '436955397390',
                                                                       'in_ucast_pkts': '352970569',
                                                                       'links': [       {       'in_nucast_pkts': '1434299',
                                                                                                'in_octets': '436955397390',
                                                                                                'in_ucast_pkts': '352970569',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12367',
                                                                                                'out_octets': '37369166849',
                                                                                                'out_ucast_pkts': '106189432',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12367',
                                                                       'out_octets': '37369166849',
                                                                       'out_ucast_pkts': '106189432',
                                                                       'poller_hash': 'ODcyNTU4cGVlcmludGZwb24gMTQ1Mjg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145289',
                                                                                               'links': [       {       'connected_time': '688559',
                                                                                                                        'fiber_length': 1718,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTU4cGVlcnBvbnBvbiAxNDUyODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 MURRAY CT (DOST MATTHEW)',
                                         'poller_hash': 'ODcyNTU4',
                                         'serial_number': '872558',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a8:53',
                                         'uptime': '688559'},
                                 {       'counter_source': '145508/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145508/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '22117500414',
                                                                       'in_ucast_pkts': '163054916',
                                                                       'mac_address': 'd0:76:8f:aa:52:0c',
                                                                       'name': '145508/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '522818871620',
                                                                       'out_ucast_pkts': '380055566',
                                                                       'poller_hash': 'OTU2RjY3cGVlcmludGYxNDU1MDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145508',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '522818871620',
                                                                       'in_ucast_pkts': '380055566',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '522818871620',
                                                                                                'in_ucast_pkts': '380055566',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '22117500414',
                                                                                                'out_ucast_pkts': '163054916',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '22117500414',
                                                                       'out_ucast_pkts': '163054916',
                                                                       'poller_hash': 'OTU2RjY3cGVlcmludGZwb24gMTQ1NTA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145508',
                                                                                               'links': [       {       'connected_time': '688356',
                                                                                                                        'fiber_length': 10397,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjY3cGVlcnBvbnBvbiAxNDU1MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "5920 PERTH ROAD 119 (JAY'S GARDEN "
                                                 'GAZEBOS)',
                                         'poller_hash': 'OTU2RjY3',
                                         'serial_number': '956F67',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:52:0c',
                                         'uptime': '688356'},
                                 {       'counter_source': '145518/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145518/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '1442',
                                                                       'in_octets': '396708306360',
                                                                       'in_ucast_pkts': '0',
                                                                       'mac_address': 'd0:76:8f:aa:51:24',
                                                                       'name': '145518/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '7713',
                                                                       'out_octets': '5344357132276',
                                                                       'out_ucast_pkts': '4023637363',
                                                                       'poller_hash': 'OTU2RjRBcGVlcmludGYxNDU1MTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145518',
                                                                       'in_nucast_pkts': '7713',
                                                                       'in_octets': '5344357132276',
                                                                       'in_ucast_pkts': '4023637363',
                                                                       'links': [       {       'in_nucast_pkts': '7713',
                                                                                                'in_octets': '5344357132276',
                                                                                                'in_ucast_pkts': '4023637363',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1442',
                                                                                                'out_octets': '396708306360',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1442',
                                                                       'out_octets': '396708306360',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'OTU2RjRBcGVlcmludGZwb24gMTQ1NTE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145518',
                                                                                               'links': [       {       'connected_time': '688459',
                                                                                                                        'fiber_length': 1707,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjRBcGVlcnBvbnBvbiAxNDU1MTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 WILLIAM ST (WICE DAVID)',
                                         'poller_hash': 'OTU2RjRB',
                                         'serial_number': '956F4A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:51:24',
                                         'uptime': '688459'},
                                 {       'counter_source': '145640/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:4f:09:c6:0a',
                                                                                                        '00:80:a3:d7:b0:51',
                                                                                                        '24:0f:9b:14:2e:5d',
                                                                                                        'd0:67:e5:ab:f7:82',
                                                                                                        'f0:9f:c2:d0:dc:59'],
                                                                       'id': '145640/g1',
                                                                       'in_nucast_pkts': '12406382',
                                                                       'in_octets': '130608035937',
                                                                       'in_ucast_pkts': '676977798',
                                                                       'mac_address': '00:06:31:87:4d:97',
                                                                       'mtu': 2000,
                                                                       'name': '145640/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '292316902',
                                                                       'out_octets': '1013559505324',
                                                                       'out_ucast_pkts': '1389514786',
                                                                       'poller_hash': 'Q0MwRDRwZWVyaW50ZjE0NTY0MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145640',
                                                                       'in_nucast_pkts': '292316902',
                                                                       'in_octets': '1013559505324',
                                                                       'in_ucast_pkts': '1389514786',
                                                                       'links': [       {       'in_nucast_pkts': '292316902',
                                                                                                'in_octets': '1013559505324',
                                                                                                'in_ucast_pkts': '1389514786',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12406382',
                                                                                                'out_octets': '130608035937',
                                                                                                'out_ucast_pkts': '676977798',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12406382',
                                                                       'out_octets': '130608035937',
                                                                       'out_ucast_pkts': '676977798',
                                                                       'poller_hash': 'Q0MwRDRwZWVyaW50ZnBvbiAxNDU2NDA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145640',
                                                                                               'links': [       {       'connected_time': '23328610',
                                                                                                                        'fiber_length': 2300,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwRDRwZWVycG9ucG9uIDE0NTY0MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '6372 PERTH ROAD 131 (COUNTY OF '
                                                 'PERTH)',
                                         'poller_hash': 'Q0MwRDQ=',
                                         'serial_number': 'CC0D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:94',
                                         'uptime': '23328610'},
                                 {       'counter_source': '145950/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145950/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1048087646888',
                                                                       'in_ucast_pkts': '2329056527',
                                                                       'mac_address': '60:db:98:4e:72:74',
                                                                       'name': '145950/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '7695560605400',
                                                                       'out_ucast_pkts': '10136780011',
                                                                       'poller_hash': 'QTlDNDBFcGVlcmludGYxNDU5NTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145950',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '7695560605400',
                                                                       'in_ucast_pkts': '10136780011',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '7695560605400',
                                                                                                'in_ucast_pkts': '10136780011',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1048087646888',
                                                                                                'out_ucast_pkts': '2329056527',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1048087646888',
                                                                       'out_ucast_pkts': '2329056527',
                                                                       'poller_hash': 'QTlDNDBFcGVlcmludGZwb24gMTQ1OTUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145950',
                                                                                               'links': [       {       'connected_time': '688411',
                                                                                                                        'fiber_length': 2396,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNDBFcGVlcnBvbnBvbiAxNDU5NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '63 GERBER DR (FALCONER DAVE)',
                                         'poller_hash': 'QTlDNDBF',
                                         'serial_number': 'A9C40E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:72:74',
                                         'uptime': '688411'},
                                 {       'counter_source': '146031/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146031/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '21541401522',
                                                                       'in_ucast_pkts': '33859562',
                                                                       'mac_address': '60:db:98:4e:72:dc',
                                                                       'name': '146031/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '149',
                                                                       'out_octets': '92836170291',
                                                                       'out_ucast_pkts': '128177109',
                                                                       'poller_hash': 'QTlDNDFCcGVlcmludGYxNDYwMzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146031',
                                                                       'in_nucast_pkts': '149',
                                                                       'in_octets': '92836170291',
                                                                       'in_ucast_pkts': '128177109',
                                                                       'links': [       {       'in_nucast_pkts': '149',
                                                                                                'in_octets': '92836170291',
                                                                                                'in_ucast_pkts': '128177109',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '21541401522',
                                                                                                'out_ucast_pkts': '33859562',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '21541401522',
                                                                       'out_ucast_pkts': '33859562',
                                                                       'poller_hash': 'QTlDNDFCcGVlcmludGZwb24gMTQ2MDMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146031',
                                                                                               'links': [       {       'connected_time': '688465',
                                                                                                                        'fiber_length': 1539,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNDFCcGVlcnBvbnBvbiAxNDYwMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '59 WILLIAM ST (MARTIN RENITA)',
                                         'poller_hash': 'QTlDNDFC',
                                         'serial_number': 'A9C41B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:72:dc',
                                         'uptime': '688465'},
                                 {       'counter_source': '146082/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146082/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '66992988241',
                                                                       'in_ucast_pkts': '203905777',
                                                                       'mac_address': 'd0:76:8f:e7:26:95',
                                                                       'name': '146082/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '614359344547',
                                                                       'out_ucast_pkts': '482492761',
                                                                       'poller_hash': 'OUMzMDIycGVlcmludGYxNDYwODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146082',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '614359344547',
                                                                       'in_ucast_pkts': '482492761',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '614359344547',
                                                                                                'in_ucast_pkts': '482492761',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '66992988241',
                                                                                                'out_ucast_pkts': '203905777',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '66992988241',
                                                                       'out_ucast_pkts': '203905777',
                                                                       'poller_hash': 'OUMzMDIycGVlcmludGZwb24gMTQ2MDgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146082',
                                                                                               'links': [       {       'connected_time': '688510',
                                                                                                                        'fiber_length': 624,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OUMzMDIycGVlcnBvbnBvbiAxNDYwODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '10 MAIN ST N (RAHMAN HABIBUR)',
                                         'poller_hash': 'OUMzMDIy',
                                         'serial_number': '9C3022',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:e7:26:95',
                                         'uptime': '688510'},
                                 {       'counter_source': '146266/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146266/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '30230234256',
                                                                       'in_ucast_pkts': '93385994',
                                                                       'mac_address': '60:db:98:4e:6f:dc',
                                                                       'name': '146266/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '768967764029',
                                                                       'out_ucast_pkts': '569772484',
                                                                       'poller_hash': 'QTlDM0JCcGVlcmludGYxNDYyNjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146266',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '768967764029',
                                                                       'in_ucast_pkts': '569772484',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '768967764029',
                                                                                                'in_ucast_pkts': '569772484',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '30230234256',
                                                                                                'out_ucast_pkts': '93385994',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '30230234256',
                                                                       'out_ucast_pkts': '93385994',
                                                                       'poller_hash': 'QTlDM0JCcGVlcmludGZwb24gMTQ2MjY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146266',
                                                                                               'links': [       {       'connected_time': '3113872',
                                                                                                                        'fiber_length': 17876,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0JCcGVlcnBvbnBvbiAxNDYyNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '58 MADDISON ST E (REDECOP JOHN)',
                                         'poller_hash': 'QTlDM0JC',
                                         'serial_number': 'A9C3BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6f:dc',
                                         'uptime': '3113872'},
                                 {       'counter_source': '146815/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146815/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '124029822675',
                                                                       'in_ucast_pkts': '571355331',
                                                                       'mac_address': '60:db:98:4e:8d:7c',
                                                                       'name': '146815/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1',
                                                                       'out_octets': '4859318172377',
                                                                       'out_ucast_pkts': '3488857195',
                                                                       'poller_hash': 'QTlDNzZGcGVlcmludGYxNDY4MTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146815',
                                                                       'in_nucast_pkts': '1',
                                                                       'in_octets': '4859318172377',
                                                                       'in_ucast_pkts': '3488857195',
                                                                       'links': [       {       'in_nucast_pkts': '1',
                                                                                                'in_octets': '4859318172377',
                                                                                                'in_ucast_pkts': '3488857195',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '124029822675',
                                                                                                'out_ucast_pkts': '571355331',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '124029822675',
                                                                       'out_ucast_pkts': '571355331',
                                                                       'poller_hash': 'QTlDNzZGcGVlcmludGZwb24gMTQ2ODE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146815',
                                                                                               'links': [       {       'connected_time': '2969243',
                                                                                                                        'fiber_length': 17827,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNzZGcGVlcnBvbnBvbiAxNDY4MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '163 MARKET ST (BOESE PETER)',
                                         'poller_hash': 'QTlDNzZG',
                                         'serial_number': 'A9C76F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:8d:7c',
                                         'uptime': '2969243'},
                                 {       'counter_source': '146937/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:54:ff'],
                                                                       'id': '146937/g1',
                                                                       'in_nucast_pkts': '67732',
                                                                       'in_octets': '6826278755',
                                                                       'in_ucast_pkts': '39386224',
                                                                       'mac_address': '60:db:98:4e:99:37',
                                                                       'mtu': 2000,
                                                                       'name': '146937/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6782122',
                                                                       'out_octets': '122484660432',
                                                                       'out_ucast_pkts': '94322323',
                                                                       'poller_hash': 'QTlDOEU2cGVlcmludGYxNDY5MzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '146937',
                                                                       'in_nucast_pkts': '6782122',
                                                                       'in_octets': '122484660432',
                                                                       'in_ucast_pkts': '94322323',
                                                                       'links': [       {       'in_nucast_pkts': '6782122',
                                                                                                'in_octets': '122484660432',
                                                                                                'in_ucast_pkts': '94322323',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '67732',
                                                                                                'out_octets': '6826278755',
                                                                                                'out_ucast_pkts': '39386224',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67732',
                                                                       'out_octets': '6826278755',
                                                                       'out_ucast_pkts': '39386224',
                                                                       'poller_hash': 'QTlDOEU2cGVlcmludGZwb24gMTQ2OTM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146937',
                                                                                               'links': [       {       'connected_time': '3113850',
                                                                                                                        'fiber_length': 17505,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOEU2cGVlcnBvbnBvbiAxNDY5Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '195 SELBY AV (SMITH EDWARD)',
                                         'poller_hash': 'QTlDOEU2',
                                         'serial_number': 'A9C8E6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:99:34',
                                         'uptime': '3113850'},
                                 {       'counter_source': '147047/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147047/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '507166049595',
                                                                       'in_ucast_pkts': '36619831759',
                                                                       'mac_address': '60:db:98:4e:70:94',
                                                                       'name': '147047/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '47',
                                                                       'out_octets': '9333350595378',
                                                                       'out_ucast_pkts': '6810867559',
                                                                       'poller_hash': 'QTlDM0QycGVlcmludGYxNDcwNDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147047',
                                                                       'in_nucast_pkts': '47',
                                                                       'in_octets': '9333350595378',
                                                                       'in_ucast_pkts': '6810867559',
                                                                       'links': [       {       'in_nucast_pkts': '47',
                                                                                                'in_octets': '9333350595378',
                                                                                                'in_ucast_pkts': '6810867559',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '507166049595',
                                                                                                'out_ucast_pkts': '36619831759',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '507166049595',
                                                                       'out_ucast_pkts': '36619831759',
                                                                       'poller_hash': 'QTlDM0QycGVlcmludGZwb24gMTQ3MDQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147047',
                                                                                               'links': [       {       'connected_time': '2015390',
                                                                                                                        'fiber_length': 17958,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0QycGVlcnBvbnBvbiAxNDcwNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '128 WINSTANLEY ST (SWINKELS IAN)',
                                         'poller_hash': 'QTlDM0Qy',
                                         'serial_number': 'A9C3D2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:70:94',
                                         'uptime': '2015390'},
                                 {       'counter_source': '147057/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d5:cf'],
                                                                       'id': '147057/g1',
                                                                       'in_nucast_pkts': '68322',
                                                                       'in_octets': '41314882320',
                                                                       'in_ucast_pkts': '105183159',
                                                                       'mac_address': '60:db:98:4e:70:37',
                                                                       'mtu': 2000,
                                                                       'name': '147057/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6782051',
                                                                       'out_octets': '975689852294',
                                                                       'out_ucast_pkts': '695763092',
                                                                       'poller_hash': 'QTlDM0M2cGVlcmludGYxNDcwNTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '147057',
                                                                       'in_nucast_pkts': '6782051',
                                                                       'in_octets': '975689852294',
                                                                       'in_ucast_pkts': '695763092',
                                                                       'links': [       {       'in_nucast_pkts': '6782051',
                                                                                                'in_octets': '975689852294',
                                                                                                'in_ucast_pkts': '695763092',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '68322',
                                                                                                'out_octets': '41314882320',
                                                                                                'out_ucast_pkts': '105183159',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68322',
                                                                       'out_octets': '41314882320',
                                                                       'out_ucast_pkts': '105183159',
                                                                       'poller_hash': 'QTlDM0M2cGVlcmludGZwb24gMTQ3MDU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147057',
                                                                                               'links': [       {       'connected_time': '3113826',
                                                                                                                        'fiber_length': 18318,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0M2cGVlcnBvbnBvbiAxNDcwNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '39 SCHADE ST (CHANSE JORDAN)',
                                         'poller_hash': 'QTlDM0M2',
                                         'serial_number': 'A9C3C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:70:34',
                                         'uptime': '3113826'},
                                 {       'counter_source': '147087/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147087/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '430726391380',
                                                                       'in_ucast_pkts': '82335425455',
                                                                       'mac_address': '60:db:98:4e:a0:04',
                                                                       'name': '147087/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '147',
                                                                       'out_octets': '2753426688244',
                                                                       'out_ucast_pkts': '2242940688',
                                                                       'poller_hash': 'QTlDOUMwcGVlcmludGYxNDcwODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147087',
                                                                       'in_nucast_pkts': '147',
                                                                       'in_octets': '2753426688244',
                                                                       'in_ucast_pkts': '2242940688',
                                                                       'links': [       {       'in_nucast_pkts': '147',
                                                                                                'in_octets': '2753426688244',
                                                                                                'in_ucast_pkts': '2242940688',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '430726391380',
                                                                                                'out_ucast_pkts': '82335425455',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '430726391380',
                                                                       'out_ucast_pkts': '82335425455',
                                                                       'poller_hash': 'QTlDOUMwcGVlcmludGZwb24gMTQ3MDg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147087',
                                                                                               'links': [       {       'connected_time': '72955601',
                                                                                                                        'fiber_length': 18167,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOUMwcGVlcnBvbnBvbiAxNDcwODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '78 MADDISON ST E (HIEBERT ABRAHAM)',
                                         'poller_hash': 'QTlDOUMw',
                                         'serial_number': 'A9C9C0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:a0:04',
                                         'uptime': '72955601'},
                                 {       'counter_source': '147168/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147168/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '522220537689',
                                                                       'in_ucast_pkts': '26395661214',
                                                                       'mac_address': '60:db:98:4e:85:ac',
                                                                       'name': '147168/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '1318373976056',
                                                                       'out_ucast_pkts': '1223913040',
                                                                       'poller_hash': 'QTlDNjc1cGVlcmludGYxNDcxNjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147168',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '1318373976056',
                                                                       'in_ucast_pkts': '1223913040',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '1318373976056',
                                                                                                'in_ucast_pkts': '1223913040',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '522220537689',
                                                                                                'out_ucast_pkts': '26395661214',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '522220537689',
                                                                       'out_ucast_pkts': '26395661214',
                                                                       'poller_hash': 'QTlDNjc1cGVlcmludGZwb24gMTQ3MTY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147168',
                                                                                               'links': [       {       'connected_time': '688445',
                                                                                                                        'fiber_length': 975,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNjc1cGVlcnBvbnBvbiAxNDcxNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29 MILL ST E (HASENPFLUG ED & '
                                                 'LOUISE)',
                                         'poller_hash': 'QTlDNjc1',
                                         'serial_number': 'A9C675',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:85:ac',
                                         'uptime': '688445'},
                                 {       'counter_source': '147366/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147366/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45291303846',
                                                                       'in_ucast_pkts': '295219888',
                                                                       'mac_address': '60:db:98:4e:8d:dc',
                                                                       'name': '147366/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '880087551014',
                                                                       'out_ucast_pkts': '628923100',
                                                                       'poller_hash': 'QTlDNzdCcGVlcmludGYxNDczNjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147366',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '880087551014',
                                                                       'in_ucast_pkts': '628923100',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '880087551014',
                                                                                                'in_ucast_pkts': '628923100',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45291303846',
                                                                                                'out_ucast_pkts': '295219888',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45291303846',
                                                                       'out_ucast_pkts': '295219888',
                                                                       'poller_hash': 'QTlDNzdCcGVlcmludGZwb24gMTQ3MzY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147366',
                                                                                               'links': [       {       'connected_time': '9084563',
                                                                                                                        'fiber_length': 10866,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNzdCcGVlcnBvbnBvbiAxNDczNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5948 PERTH ROAD 119 (ENGLER DON)',
                                         'poller_hash': 'QTlDNzdC',
                                         'serial_number': 'A9C77B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:8d:dc',
                                         'uptime': '9084563'},
                                 {       'counter_source': '147456/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147456/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '53613237579',
                                                                       'in_ucast_pkts': '213228663',
                                                                       'mac_address': '60:db:98:4e:97:a4',
                                                                       'name': '147456/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '86',
                                                                       'out_octets': '579115178391',
                                                                       'out_ucast_pkts': '537030323',
                                                                       'poller_hash': 'QTlDOEI0cGVlcmludGYxNDc0NTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147456',
                                                                       'in_nucast_pkts': '86',
                                                                       'in_octets': '579115178391',
                                                                       'in_ucast_pkts': '537030323',
                                                                       'links': [       {       'in_nucast_pkts': '86',
                                                                                                'in_octets': '579115178391',
                                                                                                'in_ucast_pkts': '537030323',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '53613237579',
                                                                                                'out_ucast_pkts': '213228663',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '53613237579',
                                                                       'out_ucast_pkts': '213228663',
                                                                       'poller_hash': 'QTlDOEI0cGVlcmludGZwb24gMTQ3NDU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147456',
                                                                                               'links': [       {       'connected_time': '21816228',
                                                                                                                        'fiber_length': 5679,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOEI0cGVlcnBvbnBvbiAxNDc0NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4249 LINE 61 (JANTZI NELSON)',
                                         'poller_hash': 'QTlDOEI0',
                                         'serial_number': 'A9C8B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:97:a4',
                                         'uptime': '21816228'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '147627/g1',
                                                                       'name': '147627/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QTlDMkI4cGVlcmludGYxNDc2MjcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147627',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QTlDMkI4cGVlcmludGZwb24gMTQ3NjI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147627',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkI4cGVlcnBvbnBvbiAxNDc2Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '13 MAIN ST N',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "147627" '
                                                                  '(location: GPON '
                                                                  'port "8/7"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:10:24.750Z',
                                         'poller_hash': 'QTlDMkI4',
                                         'serial_number': 'A9C2B8',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '147725/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:ac:6e'],
                                                                       'id': '147725/g1',
                                                                       'in_nucast_pkts': '14301',
                                                                       'in_octets': '2505695767',
                                                                       'in_ucast_pkts': '14371066',
                                                                       'mac_address': 'd0:76:8f:1b:a5:ce',
                                                                       'mtu': 2000,
                                                                       'name': '147725/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434250',
                                                                       'out_octets': '173013499087',
                                                                       'out_ucast_pkts': '117084458',
                                                                       'poller_hash': 'ODcyNTA3cGVlcmludGYxNDc3MjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '147725',
                                                                       'in_nucast_pkts': '1434250',
                                                                       'in_octets': '173013499087',
                                                                       'in_ucast_pkts': '117084458',
                                                                       'links': [       {       'in_nucast_pkts': '1434250',
                                                                                                'in_octets': '173013499087',
                                                                                                'in_ucast_pkts': '117084458',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14301',
                                                                                                'out_octets': '2505695767',
                                                                                                'out_ucast_pkts': '14371066',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14301',
                                                                       'out_octets': '2505695767',
                                                                       'out_ucast_pkts': '14371066',
                                                                       'poller_hash': 'ODcyNTA3cGVlcmludGZwb24gMTQ3NzI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147725',
                                                                                               'links': [       {       'connected_time': '688562',
                                                                                                                        'fiber_length': 1732,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTA3cGVlcnBvbnBvbiAxNDc3MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '17 MURRAY CT (AUSTIN MICHELLE)',
                                         'poller_hash': 'ODcyNTA3',
                                         'serial_number': '872507',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:cb',
                                         'uptime': '688562'},
                                 {       'counter_source': '147854/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d4:dd'],
                                                                       'id': '147854/g1',
                                                                       'in_nucast_pkts': '14882',
                                                                       'in_octets': '6881749883',
                                                                       'in_ucast_pkts': '44945106',
                                                                       'mac_address': '60:db:98:4e:89:d7',
                                                                       'mtu': 2000,
                                                                       'name': '147854/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464417',
                                                                       'out_octets': '197250513106',
                                                                       'out_ucast_pkts': '146960407',
                                                                       'poller_hash': 'QTlDNkZBcGVlcmludGYxNDc4NTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '147854',
                                                                       'in_nucast_pkts': '1464417',
                                                                       'in_octets': '197250513106',
                                                                       'in_ucast_pkts': '146960407',
                                                                       'links': [       {       'in_nucast_pkts': '1464417',
                                                                                                'in_octets': '197250513106',
                                                                                                'in_ucast_pkts': '146960407',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14882',
                                                                                                'out_octets': '6881749883',
                                                                                                'out_ucast_pkts': '44945106',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14882',
                                                                       'out_octets': '6881749883',
                                                                       'out_ucast_pkts': '44945106',
                                                                       'poller_hash': 'QTlDNkZBcGVlcmludGZwb24gMTQ3ODU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147854',
                                                                                               'links': [       {       'connected_time': '688351',
                                                                                                                        'fiber_length': 1183,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNkZBcGVlcnBvbnBvbiAxNDc4NTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29 TEMPERENCE ST, MILVERTON (Melissa '
                                                 '& Chris Stahlke)',
                                         'poller_hash': 'QTlDNkZB',
                                         'serial_number': 'A9C6FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:89:d4',
                                         'uptime': '688351'},
                                 {       'counter_source': '149104/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149104/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '51640387311',
                                                                       'in_ucast_pkts': '192479159',
                                                                       'mac_address': '60:db:98:4e:9c:5c',
                                                                       'name': '149104/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '137',
                                                                       'out_octets': '662396714627',
                                                                       'out_ucast_pkts': '566775004',
                                                                       'poller_hash': 'QTlDOTRCcGVlcmludGYxNDkxMDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149104',
                                                                       'in_nucast_pkts': '137',
                                                                       'in_octets': '662396714627',
                                                                       'in_ucast_pkts': '566775004',
                                                                       'links': [       {       'in_nucast_pkts': '137',
                                                                                                'in_octets': '662396714627',
                                                                                                'in_ucast_pkts': '566775004',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '51640387311',
                                                                                                'out_ucast_pkts': '192479159',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '51640387311',
                                                                       'out_ucast_pkts': '192479159',
                                                                       'poller_hash': 'QTlDOTRCcGVlcmludGZwb24gMTQ5MTA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149104',
                                                                                               'links': [       {       'connected_time': '21816297',
                                                                                                                        'fiber_length': 13761,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOTRCcGVlcnBvbnBvbiAxNDkxMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5331 ROAD 144 (DAVIDSON FRASER & '
                                                 'STARLA)',
                                         'poller_hash': 'QTlDOTRC',
                                         'serial_number': 'A9C94B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:9c:5c',
                                         'uptime': '21816297'},
                                 {       'counter_source': '149127/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b7:ce'],
                                                                       'id': '149127/g1',
                                                                       'in_nucast_pkts': '56291',
                                                                       'in_octets': '185719339003',
                                                                       'in_ucast_pkts': '161646148',
                                                                       'mac_address': '60:db:98:4e:99:2f',
                                                                       'mtu': 2000,
                                                                       'name': '149127/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6781892',
                                                                       'out_octets': '196033823123',
                                                                       'out_ucast_pkts': '211262605',
                                                                       'poller_hash': 'QTlDOEU1cGVlcmludGYxNDkxMjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '149127',
                                                                       'in_nucast_pkts': '6781892',
                                                                       'in_octets': '196033823123',
                                                                       'in_ucast_pkts': '211262605',
                                                                       'links': [       {       'in_nucast_pkts': '6781892',
                                                                                                'in_octets': '196033823123',
                                                                                                'in_ucast_pkts': '211262605',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '56291',
                                                                                                'out_octets': '185719339003',
                                                                                                'out_ucast_pkts': '161646148',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '56291',
                                                                       'out_octets': '185719339003',
                                                                       'out_ucast_pkts': '161646148',
                                                                       'poller_hash': 'QTlDOEU1cGVlcmludGZwb24gMTQ5MTI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149127',
                                                                                               'links': [       {       'connected_time': '3113845',
                                                                                                                        'fiber_length': 18838,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOEU1cGVlcnBvbnBvbiAxNDkxMjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '77 SCHADE ST (VAN DYK HENRY)',
                                         'poller_hash': 'QTlDOEU1',
                                         'serial_number': 'A9C8E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:99:2c',
                                         'uptime': '3113845'},
                                 {       'counter_source': '149147/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149147/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1114445892395',
                                                                       'in_ucast_pkts': '2219974894',
                                                                       'mac_address': 'ec:4f:82:0d:6a:4e',
                                                                       'name': '149147/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '99',
                                                                       'out_octets': '9734174418356',
                                                                       'out_ucast_pkts': '6724146514',
                                                                       'poller_hash': 'MjA1REYxcGVlcmludGYxNDkxNDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149147',
                                                                       'in_nucast_pkts': '99',
                                                                       'in_octets': '9734174418356',
                                                                       'in_ucast_pkts': '6724146514',
                                                                       'links': [       {       'in_nucast_pkts': '99',
                                                                                                'in_octets': '9734174418356',
                                                                                                'in_ucast_pkts': '6724146514',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1114445892395',
                                                                                                'out_ucast_pkts': '2219974894',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1114445892395',
                                                                       'out_ucast_pkts': '2219974894',
                                                                       'poller_hash': 'MjA1REYxcGVlcmludGZwb24gMTQ5MTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149147',
                                                                                               'links': [       {       'connected_time': '688502',
                                                                                                                        'fiber_length': 1700,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REYxcGVlcnBvbnBvbiAxNDkxNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 WILLIAM STREET APT 201 (HARVEY '
                                                 'JANE)',
                                         'poller_hash': 'MjA1REYx',
                                         'serial_number': '205DF1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6a:4e',
                                         'uptime': '688502'},
                                 {       'counter_source': '149178/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149178/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '290',
                                                                       'in_octets': '95136350377',
                                                                       'in_ucast_pkts': '812367490',
                                                                       'mac_address': '60:db:98:4e:91:cc',
                                                                       'name': '149178/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1480',
                                                                       'out_octets': '2320238436424',
                                                                       'out_ucast_pkts': '1672970699',
                                                                       'poller_hash': 'QTlDN0Y5cGVlcmludGYxNDkxNzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149178',
                                                                       'in_nucast_pkts': '1480',
                                                                       'in_octets': '2320238436424',
                                                                       'in_ucast_pkts': '1672970699',
                                                                       'links': [       {       'in_nucast_pkts': '1480',
                                                                                                'in_octets': '2320238436424',
                                                                                                'in_ucast_pkts': '1672970699',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '290',
                                                                                                'out_octets': '95136350377',
                                                                                                'out_ucast_pkts': '812367490',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '290',
                                                                       'out_octets': '95136350377',
                                                                       'out_ucast_pkts': '812367490',
                                                                       'poller_hash': 'QTlDN0Y5cGVlcmludGZwb24gMTQ5MTc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149178',
                                                                                               'links': [       {       'connected_time': '688475',
                                                                                                                        'fiber_length': 1363,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDN0Y5cGVlcnBvbnBvbiAxNDkxNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 MAIN ST N (KLASSEN CAMERON)',
                                         'poller_hash': 'QTlDN0Y5',
                                         'serial_number': 'A9C7F9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:91:cc',
                                         'uptime': '688475'},
                                 {       'counter_source': '149655/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:fa:10:92'],
                                                                       'id': '149655/g1',
                                                                       'in_nucast_pkts': '119253',
                                                                       'in_octets': '1012519937',
                                                                       'in_ucast_pkts': '11282223',
                                                                       'mac_address': '60:db:98:73:91:e7',
                                                                       'mtu': 2000,
                                                                       'name': '149655/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49519180',
                                                                       'out_octets': '82644586229',
                                                                       'out_ucast_pkts': '44686831',
                                                                       'poller_hash': 'QjBBMTRFcGVlcmludGYxNDk2NTUvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '149655',
                                                                       'in_nucast_pkts': '49519180',
                                                                       'in_octets': '82644586229',
                                                                       'in_ucast_pkts': '44686831',
                                                                       'links': [       {       'in_nucast_pkts': '49519180',
                                                                                                'in_octets': '82644586229',
                                                                                                'in_ucast_pkts': '44686831',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '119253',
                                                                                                'out_octets': '1012519937',
                                                                                                'out_ucast_pkts': '11282223',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '119253',
                                                                       'out_octets': '1012519937',
                                                                       'out_ucast_pkts': '11282223',
                                                                       'poller_hash': 'QjBBMTRFcGVlcmludGZwb24gMTQ5NjU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149655',
                                                                                               'links': [       {       'connected_time': '2138232',
                                                                                                                        'fiber_length': 10885,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMTRFcGVlcnBvbnBvbiAxNDk2NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5944 PERTH ROAD 119 (PRONG DOUGLAS & '
                                                 'LISA)',
                                         'poller_hash': 'QjBBMTRF',
                                         'serial_number': 'B0A14E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:91:e4',
                                         'uptime': '2138232'},
                                 {       'counter_source': '149758/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149758/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '93560022661',
                                                                       'in_ucast_pkts': '698785933',
                                                                       'mac_address': 'd0:76:8f:e7:1c:ab',
                                                                       'name': '149758/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '3559463101606',
                                                                       'out_ucast_pkts': '2645017153',
                                                                       'poller_hash': 'OUMyRjA4cGVlcmludGYxNDk3NTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149758',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '3559463101606',
                                                                       'in_ucast_pkts': '2645017153',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '3559463101606',
                                                                                                'in_ucast_pkts': '2645017153',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '93560022661',
                                                                                                'out_ucast_pkts': '698785933',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '93560022661',
                                                                       'out_ucast_pkts': '698785933',
                                                                       'poller_hash': 'OUMyRjA4cGVlcmludGZwb24gMTQ5NzU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149758',
                                                                                               'links': [       {       'connected_time': '3114039',
                                                                                                                        'fiber_length': 17961,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OUMyRjA4cGVlcnBvbnBvbiAxNDk3NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '163 SMITH ST (FRIESEN CORNELIUS)',
                                         'poller_hash': 'OUMyRjA4',
                                         'serial_number': '9C2F08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:e7:1c:ab',
                                         'uptime': '3114039'},
                                 {       'counter_source': '150024/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:b5:8e'],
                                                                       'id': '150024/g1',
                                                                       'in_nucast_pkts': '54481',
                                                                       'in_octets': '32115520283',
                                                                       'in_ucast_pkts': '113947729',
                                                                       'mac_address': '60:db:98:73:9c:a7',
                                                                       'mtu': 2000,
                                                                       'name': '150024/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6685861',
                                                                       'out_octets': '705511764726',
                                                                       'out_ucast_pkts': '531196435',
                                                                       'poller_hash': 'QjBBMkE2cGVlcmludGYxNTAwMjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '150024',
                                                                       'in_nucast_pkts': '6685861',
                                                                       'in_octets': '705511764726',
                                                                       'in_ucast_pkts': '531196435',
                                                                       'links': [       {       'in_nucast_pkts': '6685861',
                                                                                                'in_octets': '705511764726',
                                                                                                'in_ucast_pkts': '531196435',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '54481',
                                                                                                'out_octets': '32115520283',
                                                                                                'out_ucast_pkts': '113947729',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '54481',
                                                                       'out_octets': '32115520283',
                                                                       'out_ucast_pkts': '113947729',
                                                                       'poller_hash': 'QjBBMkE2cGVlcmludGZwb24gMTUwMDI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150024',
                                                                                               'links': [       {       'connected_time': '3113871',
                                                                                                                        'fiber_length': 18440,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMkE2cGVlcnBvbnBvbiAxNTAwMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '167 MADDISON ST E (MACHAN TORIE)',
                                         'poller_hash': 'QjBBMkE2',
                                         'serial_number': 'B0A2A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:9c:a4',
                                         'uptime': '3113871'},
                                 {       'counter_source': '150107/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '150107/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '92132228907',
                                                                       'in_ucast_pkts': '479617271',
                                                                       'mac_address': '00:06:31:e8:f3:53',
                                                                       'name': '150107/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1606686438815',
                                                                       'out_ucast_pkts': '1172788603',
                                                                       'poller_hash': 'MUQwNEY1cGVlcmludGYxNTAxMDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '150107',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1606686438815',
                                                                       'in_ucast_pkts': '1172788603',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1606686438815',
                                                                                                'in_ucast_pkts': '1172788603',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '92132228907',
                                                                                                'out_ucast_pkts': '479617271',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '92132228907',
                                                                       'out_ucast_pkts': '479617271',
                                                                       'poller_hash': 'MUQwNEY1cGVlcmludGZwb24gMTUwMTA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150107',
                                                                                               'links': [       {       'connected_time': '688512',
                                                                                                                        'fiber_length': 615,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEY1cGVlcnBvbnBvbiAxNTAxMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '16 MAIN ST N (SCHMITT TONY)',
                                         'poller_hash': 'MUQwNEY1',
                                         'serial_number': '1D04F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:53',
                                         'uptime': '688512'},
                                 {       'counter_source': '150118/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:bb:2f'],
                                                                       'id': '150118/g1',
                                                                       'in_nucast_pkts': '12466',
                                                                       'in_octets': '24536744342',
                                                                       'in_ucast_pkts': '88184358',
                                                                       'mac_address': 'd0:76:8f:06:e5:89',
                                                                       'mtu': 2000,
                                                                       'name': '150118/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464287',
                                                                       'out_octets': '461410088118',
                                                                       'out_ucast_pkts': '336262861',
                                                                       'poller_hash': 'ODQ0QkJFcGVlcmludGYxNTAxMTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '150118',
                                                                       'in_nucast_pkts': '1464287',
                                                                       'in_octets': '461410088118',
                                                                       'in_ucast_pkts': '336262861',
                                                                       'links': [       {       'in_nucast_pkts': '1464287',
                                                                                                'in_octets': '461410088118',
                                                                                                'in_ucast_pkts': '336262861',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12466',
                                                                                                'out_octets': '24536744342',
                                                                                                'out_ucast_pkts': '88184358',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12466',
                                                                       'out_octets': '24536744342',
                                                                       'out_ucast_pkts': '88184358',
                                                                       'poller_hash': 'ODQ0QkJFcGVlcmludGZwb24gMTUwMTE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150118',
                                                                                               'links': [       {       'connected_time': '688372',
                                                                                                                        'fiber_length': 1211,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkJFcGVlcnBvbnBvbiAxNTAxMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '20 WEST ST (SCHNEIDER EVAN)',
                                         'poller_hash': 'ODQ0QkJF',
                                         'serial_number': '844BBE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:89',
                                         'uptime': '688372'},
                                 {       'counter_source': '150539/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b0:4e:26:e9:95:54'],
                                                                       'id': '150539/g1',
                                                                       'in_nucast_pkts': '29',
                                                                       'in_octets': '182641581597',
                                                                       'in_ucast_pkts': '290508802',
                                                                       'mac_address': '60:db:98:8f:00:1c',
                                                                       'mtu': 2000,
                                                                       'name': '150539/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9363913',
                                                                       'out_octets': '672999506191',
                                                                       'out_ucast_pkts': '586668026',
                                                                       'poller_hash': 'QjUxMTc4cGVlcmludGYxNTA1MzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '150539',
                                                                       'in_nucast_pkts': '9363913',
                                                                       'in_octets': '672999506191',
                                                                       'in_ucast_pkts': '586668026',
                                                                       'links': [       {       'in_nucast_pkts': '9363913',
                                                                                                'in_octets': '672999506191',
                                                                                                'in_ucast_pkts': '586668026',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '29',
                                                                                                'out_octets': '182641581597',
                                                                                                'out_ucast_pkts': '290508802',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29',
                                                                       'out_octets': '182641581597',
                                                                       'out_ucast_pkts': '290508802',
                                                                       'poller_hash': 'QjUxMTc4cGVlcmludGZwb24gMTUwNTM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150539',
                                                                                               'links': [       {       'connected_time': '4315564',
                                                                                                                        'fiber_length': 7163,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMTc4cGVlcnBvbnBvbiAxNTA1Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '6248 ROAD 124 (CLAASSEN ELISABETH '
                                                 '(ISABELLE))',
                                         'poller_hash': 'QjUxMTc4',
                                         'serial_number': 'B51178',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:00:1c',
                                         'uptime': '4315564'},
                                 {       'counter_source': '151227/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '10:27:f5:32:b8:bb'],
                                                                       'id': '151227/g1',
                                                                       'in_nucast_pkts': '31819',
                                                                       'in_octets': '54864245648',
                                                                       'in_ucast_pkts': '95136356',
                                                                       'mac_address': '60:db:98:8e:ff:f8',
                                                                       'mtu': 2000,
                                                                       'name': '151227/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '85703255',
                                                                       'out_octets': '182649242925',
                                                                       'out_ucast_pkts': '162762136',
                                                                       'poller_hash': 'QjUxMTZDcGVlcmludGYxNTEyMjcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '151227',
                                                                       'in_nucast_pkts': '85703255',
                                                                       'in_octets': '182649242925',
                                                                       'in_ucast_pkts': '162762136',
                                                                       'links': [       {       'in_nucast_pkts': '85703255',
                                                                                                'in_octets': '182649242925',
                                                                                                'in_ucast_pkts': '162762136',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '31819',
                                                                                                'out_octets': '54864245648',
                                                                                                'out_ucast_pkts': '95136356',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31819',
                                                                       'out_octets': '54864245648',
                                                                       'out_ucast_pkts': '95136356',
                                                                       'poller_hash': 'QjUxMTZDcGVlcmludGZwb24gMTUxMjI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '151227',
                                                                                               'links': [       {       'connected_time': '67790643',
                                                                                                                        'fiber_length': 10181,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMTZDcGVlcnBvbnBvbiAxNTEyMjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5917 PERTH ROAD 119 (GERBER KATHRYN)',
                                         'poller_hash': 'QjUxMTZD',
                                         'serial_number': 'B5116C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8e:ff:f8',
                                         'uptime': '67790643'},
                                 {       'counter_source': '151510/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '151510/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '171191420863',
                                                                       'in_ucast_pkts': '399454065',
                                                                       'mac_address': '60:db:98:73:9b:14',
                                                                       'name': '151510/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '1948138557532',
                                                                       'out_ucast_pkts': '1505546976',
                                                                       'poller_hash': 'QjBBMjc0cGVlcmludGYxNTE1MTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '151510',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '1948138557532',
                                                                       'in_ucast_pkts': '1505546976',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '1948138557532',
                                                                                                'in_ucast_pkts': '1505546976',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '171191420863',
                                                                                                'out_ucast_pkts': '399454065',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '171191420863',
                                                                       'out_ucast_pkts': '399454065',
                                                                       'poller_hash': 'QjBBMjc0cGVlcmludGZwb24gMTUxNTEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '151510',
                                                                                               'links': [       {       'connected_time': '21816699',
                                                                                                                        'fiber_length': 5626,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMjc0cGVlcnBvbnBvbiAxNTE1MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6414 ROAD 140 (ALIGN PLUMBING)',
                                         'poller_hash': 'QjBBMjc0',
                                         'serial_number': 'B0A274',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:9b:14',
                                         'uptime': '21816699'},
                                 {       'counter_source': '151610/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '151610/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '312016351844',
                                                                       'in_ucast_pkts': '1400819304',
                                                                       'mac_address': '60:db:98:73:9b:b4',
                                                                       'name': '151610/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '5818661704564',
                                                                       'out_ucast_pkts': '4240295166',
                                                                       'poller_hash': 'QjBBMjg4cGVlcmludGYxNTE2MTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '151610',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '5818661704564',
                                                                       'in_ucast_pkts': '4240295166',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '5818661704564',
                                                                                                'in_ucast_pkts': '4240295166',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '312016351844',
                                                                                                'out_ucast_pkts': '1400819304',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '312016351844',
                                                                       'out_ucast_pkts': '1400819304',
                                                                       'poller_hash': 'QjBBMjg4cGVlcmludGZwb24gMTUxNjEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '151610',
                                                                                               'links': [       {       'connected_time': '538555',
                                                                                                                        'fiber_length': 18387,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMjg4cGVlcnBvbnBvbiAxNTE2MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5307 ROAD 130 (BALL ANDREW)',
                                         'poller_hash': 'QjBBMjg4',
                                         'serial_number': 'B0A288',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:9b:b4',
                                         'uptime': '538555'},
                                 {       'counter_source': '151628/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b7:e0'],
                                                                       'id': '151628/g1',
                                                                       'in_nucast_pkts': '132634',
                                                                       'in_octets': '10852915982',
                                                                       'in_ucast_pkts': '21194325',
                                                                       'mac_address': '60:db:98:8e:ff:98',
                                                                       'mtu': 2000,
                                                                       'name': '151628/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14580570',
                                                                       'out_octets': '127108458661',
                                                                       'out_ucast_pkts': '96700608',
                                                                       'poller_hash': 'QjUxMTRDcGVlcmludGYxNTE2MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '151628',
                                                                       'in_nucast_pkts': '14580570',
                                                                       'in_octets': '127108458661',
                                                                       'in_ucast_pkts': '96700608',
                                                                       'links': [       {       'in_nucast_pkts': '14580570',
                                                                                                'in_octets': '127108458661',
                                                                                                'in_ucast_pkts': '96700608',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '132634',
                                                                                                'out_octets': '10852915982',
                                                                                                'out_ucast_pkts': '21194325',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '132634',
                                                                       'out_octets': '10852915982',
                                                                       'out_ucast_pkts': '21194325',
                                                                       'poller_hash': 'QjUxMTRDcGVlcmludGZwb24gMTUxNjI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '151628',
                                                                                               'links': [       {       'connected_time': '6230566',
                                                                                                                        'fiber_length': 10606,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMTRDcGVlcnBvbnBvbiAxNTE2Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5593 DUKE ST (LEIS WARREN)',
                                         'poller_hash': 'QjUxMTRD',
                                         'serial_number': 'B5114C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8e:ff:98',
                                         'uptime': '6230566'},
                                 {       'counter_source': '152303/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b8:38'],
                                                                       'id': '152303/g1',
                                                                       'in_nucast_pkts': '664073',
                                                                       'in_octets': '206772621156',
                                                                       'in_ucast_pkts': '729848416',
                                                                       'mac_address': '60:db:98:8f:00:df',
                                                                       'mtu': 2000,
                                                                       'name': '152303/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83519480',
                                                                       'out_octets': '2296784527582',
                                                                       'out_ucast_pkts': '1839693903',
                                                                       'poller_hash': 'QjUxMUI5cGVlcmludGYxNTIzMDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '152303',
                                                                       'in_nucast_pkts': '83519480',
                                                                       'in_octets': '2296784527582',
                                                                       'in_ucast_pkts': '1839693903',
                                                                       'links': [       {       'in_nucast_pkts': '83519480',
                                                                                                'in_octets': '2296784527582',
                                                                                                'in_ucast_pkts': '1839693903',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '664073',
                                                                                                'out_octets': '206772621156',
                                                                                                'out_ucast_pkts': '729848416',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '664073',
                                                                       'out_octets': '206772621156',
                                                                       'out_ucast_pkts': '729848416',
                                                                       'poller_hash': 'QjUxMUI5cGVlcmludGZwb24gMTUyMzAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '152303',
                                                                                               'links': [       {       'connected_time': '67790855',
                                                                                                                        'fiber_length': 7486,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMUI5cGVlcnBvbnBvbiAxNTIzMDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '6214 ROAD 124 (KUNTZE JOHN)',
                                         'poller_hash': 'QjUxMUI5',
                                         'serial_number': 'B511B9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:00:df',
                                         'uptime': '67790855'},
                                 {       'counter_source': '152377/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '152377/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '61840548718',
                                                                       'in_ucast_pkts': '300762560',
                                                                       'mac_address': '60:db:98:73:98:04',
                                                                       'name': '152377/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1369935464804',
                                                                       'out_ucast_pkts': '951651726',
                                                                       'poller_hash': 'QjBBMjEycGVlcmludGYxNTIzNzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '152377',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1369935464804',
                                                                       'in_ucast_pkts': '951651726',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1369935464804',
                                                                                                'in_ucast_pkts': '951651726',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '61840548718',
                                                                                                'out_ucast_pkts': '300762560',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '61840548718',
                                                                       'out_ucast_pkts': '300762560',
                                                                       'poller_hash': 'QjBBMjEycGVlcmludGZwb24gMTUyMzc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '152377',
                                                                                               'links': [       {       'connected_time': '21816063',
                                                                                                                        'fiber_length': 10354,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMjEycGVlcnBvbnBvbiAxNTIzNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6114 ANA ST (KRAEMER RICHARD & '
                                                 'FRANCES)',
                                         'poller_hash': 'QjBBMjEy',
                                         'serial_number': 'B0A212',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:98:04',
                                         'uptime': '21816063'},
                                 {       'counter_source': '154899/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:4b:fe:1f:59:d8'],
                                                                       'id': '154899/g1',
                                                                       'in_nucast_pkts': '24',
                                                                       'in_octets': '17364541394',
                                                                       'in_ucast_pkts': '44401283',
                                                                       'mac_address': '60:db:98:8f:00:d9',
                                                                       'mtu': 2000,
                                                                       'name': '154899/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6686022',
                                                                       'out_octets': '533853327953',
                                                                       'out_ucast_pkts': '378548363',
                                                                       'poller_hash': 'QjUxMUI3cGVlcmludGYxNTQ4OTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '154899',
                                                                       'in_nucast_pkts': '6686022',
                                                                       'in_octets': '533853327953',
                                                                       'in_ucast_pkts': '378548363',
                                                                       'links': [       {       'in_nucast_pkts': '6686022',
                                                                                                'in_octets': '533853327953',
                                                                                                'in_ucast_pkts': '378548363',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '24',
                                                                                                'out_octets': '17364541394',
                                                                                                'out_ucast_pkts': '44401283',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24',
                                                                       'out_octets': '17364541394',
                                                                       'out_ucast_pkts': '44401283',
                                                                       'poller_hash': 'QjUxMUI3cGVlcmludGZwb24gMTU0ODk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '154899',
                                                                                               'links': [       {       'connected_time': '3113963',
                                                                                                                        'fiber_length': 18104,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMUI3cGVlcnBvbnBvbiAxNTQ4OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '56 MADDISON ST W (BOYD RYAN)',
                                         'poller_hash': 'QjUxMUI3',
                                         'serial_number': 'B511B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:00:d9',
                                         'uptime': '3113963'},
                                 {       'counter_source': '154907/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e0:63:da:83:f9:f0'],
                                                                       'id': '154907/g1',
                                                                       'in_nucast_pkts': '71624',
                                                                       'in_octets': '860876007175',
                                                                       'in_ucast_pkts': '936606374',
                                                                       'mac_address': '60:db:98:8f:01:54',
                                                                       'mtu': 2000,
                                                                       'name': '154907/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15671454',
                                                                       'out_octets': '2180756487137',
                                                                       'out_ucast_pkts': '1827592773',
                                                                       'poller_hash': 'QjUxMUUwcGVlcmludGYxNTQ5MDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '154907',
                                                                       'in_nucast_pkts': '15671454',
                                                                       'in_octets': '2180756487137',
                                                                       'in_ucast_pkts': '1827592773',
                                                                       'links': [       {       'in_nucast_pkts': '15671454',
                                                                                                'in_octets': '2180756487137',
                                                                                                'in_ucast_pkts': '1827592773',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '71624',
                                                                                                'out_octets': '860876007175',
                                                                                                'out_ucast_pkts': '936606374',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '71624',
                                                                       'out_octets': '860876007175',
                                                                       'out_ucast_pkts': '936606374',
                                                                       'poller_hash': 'QjUxMUUwcGVlcmludGZwb24gMTU0OTA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '154907',
                                                                                               'links': [       {       'connected_time': '6818853',
                                                                                                                        'fiber_length': 2382,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMUUwcGVlcnBvbnBvbiAxNTQ5MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '64 GERBER DR (LOESCH JOSEPH)',
                                         'poller_hash': 'QjUxMUUw',
                                         'serial_number': 'B511E0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:01:54',
                                         'uptime': '6818853'},
                                 {       'counter_source': '155023/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:98:e8:5f:d5:94'],
                                                                       'id': '155023/g1',
                                                                       'in_nucast_pkts': '470',
                                                                       'in_octets': '2501516450',
                                                                       'in_ucast_pkts': '7313788',
                                                                       'mac_address': '60:db:98:8e:ff:ef',
                                                                       'mtu': 2000,
                                                                       'name': '155023/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '186632',
                                                                       'out_octets': '49014935556',
                                                                       'out_ucast_pkts': '46627593',
                                                                       'poller_hash': 'QjUxMTY5cGVlcmludGYxNTUwMjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '155023',
                                                                       'in_nucast_pkts': '186632',
                                                                       'in_octets': '49014935556',
                                                                       'in_ucast_pkts': '46627593',
                                                                       'links': [       {       'in_nucast_pkts': '186632',
                                                                                                'in_octets': '49014935556',
                                                                                                'in_ucast_pkts': '46627593',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '470',
                                                                                                'out_octets': '2501516450',
                                                                                                'out_ucast_pkts': '7313788',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '470',
                                                                       'out_octets': '2501516450',
                                                                       'out_ucast_pkts': '7313788',
                                                                       'poller_hash': 'QjUxMTY5cGVlcmludGZwb24gMTU1MDIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '155023',
                                                                                               'links': [       {       'connected_time': '90240',
                                                                                                                        'fiber_length': 6478,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMTY5cGVlcnBvbnBvbiAxNTUwMjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '6313 ROAD 140 (FALCONER JOHN)',
                                         'poller_hash': 'QjUxMTY5',
                                         'serial_number': 'B51169',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8e:ff:ef',
                                         'uptime': '90240'},
                                 {       'counter_source': '155027/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:21:91:f8:32:4d'],
                                                                       'id': '155027/g1',
                                                                       'in_nucast_pkts': '5282',
                                                                       'in_octets': '967666172',
                                                                       'in_ucast_pkts': '1320759',
                                                                       'mac_address': '60:db:98:8f:04:48',
                                                                       'mtu': 2000,
                                                                       'name': '155027/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464249',
                                                                       'out_octets': '2261970364',
                                                                       'out_ucast_pkts': '2164385',
                                                                       'poller_hash': 'QjUxMkRDcGVlcmludGYxNTUwMjcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '155027',
                                                                       'in_nucast_pkts': '1464249',
                                                                       'in_octets': '2261970364',
                                                                       'in_ucast_pkts': '2164385',
                                                                       'links': [       {       'in_nucast_pkts': '1464249',
                                                                                                'in_octets': '2261970364',
                                                                                                'in_ucast_pkts': '2164385',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5282',
                                                                                                'out_octets': '967666172',
                                                                                                'out_ucast_pkts': '1320759',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5282',
                                                                       'out_octets': '967666172',
                                                                       'out_ucast_pkts': '1320759',
                                                                       'poller_hash': 'QjUxMkRDcGVlcmludGZwb24gMTU1MDI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '155027',
                                                                                               'links': [       {       'connected_time': '688527',
                                                                                                                        'fiber_length': 599,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMkRDcGVlcnBvbnBvbiAxNTUwMjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '14 MAIN ST N (NORTHERN DREAM INK)',
                                         'poller_hash': 'QjUxMkRD',
                                         'serial_number': 'B512DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:04:48',
                                         'uptime': '688527'},
                                 {       'counter_source': '155925/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:bb:a8'],
                                                                       'id': '155925/g1',
                                                                       'in_nucast_pkts': '14766',
                                                                       'in_octets': '19106671673',
                                                                       'in_ucast_pkts': '124167407',
                                                                       'mac_address': '60:db:98:4e:71:47',
                                                                       'mtu': 2000,
                                                                       'name': '155925/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464410',
                                                                       'out_octets': '444748064450',
                                                                       'out_ucast_pkts': '329142721',
                                                                       'poller_hash': 'QTlDM0U4cGVlcmludGYxNTU5MjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '155925',
                                                                       'in_nucast_pkts': '1464410',
                                                                       'in_octets': '444748064450',
                                                                       'in_ucast_pkts': '329142721',
                                                                       'links': [       {       'in_nucast_pkts': '1464410',
                                                                                                'in_octets': '444748064450',
                                                                                                'in_ucast_pkts': '329142721',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14766',
                                                                                                'out_octets': '19106671673',
                                                                                                'out_ucast_pkts': '124167407',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14766',
                                                                       'out_octets': '19106671673',
                                                                       'out_ucast_pkts': '124167407',
                                                                       'poller_hash': 'QTlDM0U4cGVlcmludGZwb24gMTU1OTI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '155925',
                                                                                               'links': [       {       'connected_time': '688477',
                                                                                                                        'fiber_length': 969,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0U4cGVlcnBvbnBvbiAxNTU5MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 WILLIAM ST (WAGNER ASHLEY)',
                                         'poller_hash': 'QTlDM0U4',
                                         'serial_number': 'A9C3E8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:71:44',
                                         'uptime': '688477'},
                                 {       'counter_source': '156618/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b5:71'],
                                                                       'id': '156618/g1',
                                                                       'in_nucast_pkts': '14975',
                                                                       'in_octets': '121924392410',
                                                                       'in_ucast_pkts': '148212154',
                                                                       'mac_address': '84:d3:43:7d:5a:1a',
                                                                       'mtu': 2000,
                                                                       'name': '156618/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447016',
                                                                       'out_octets': '282541765475',
                                                                       'out_ucast_pkts': '242073614',
                                                                       'poller_hash': 'QzgyMEFDcGVlcmludGYxNTY2MTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156618',
                                                                       'in_nucast_pkts': '1447016',
                                                                       'in_octets': '282541765475',
                                                                       'in_ucast_pkts': '242073614',
                                                                       'links': [       {       'in_nucast_pkts': '1447016',
                                                                                                'in_octets': '282541765475',
                                                                                                'in_ucast_pkts': '242073614',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14975',
                                                                                                'out_octets': '121924392410',
                                                                                                'out_ucast_pkts': '148212154',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14975',
                                                                       'out_octets': '121924392410',
                                                                       'out_ucast_pkts': '148212154',
                                                                       'poller_hash': 'QzgyMEFDcGVlcmludGZwb24gMTU2NjE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156618',
                                                                                               'links': [       {       'connected_time': '688402',
                                                                                                                        'fiber_length': 2206,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgyMEFDcGVlcnBvbnBvbiAxNTY2MTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '65 REAGAN ST (JORDAO JOSH)',
                                         'poller_hash': 'QzgyMEFD',
                                         'serial_number': 'C820AC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:5a:17',
                                         'uptime': '688402'},
                                 {       'counter_source': '156776/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f4:92:bf:84:2a:53'],
                                                                       'id': '156776/g1',
                                                                       'in_nucast_pkts': '16192651',
                                                                       'in_octets': '1556591803024',
                                                                       'in_ucast_pkts': '2771493963',
                                                                       'mac_address': '84:d3:43:34:02:a5',
                                                                       'mtu': 2000,
                                                                       'name': '156776/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '79374701',
                                                                       'out_octets': '6582580067884',
                                                                       'out_ucast_pkts': '6141309966',
                                                                       'poller_hash': 'QkREOUZEcGVlcmludGYxNTY3NzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156776',
                                                                       'in_nucast_pkts': '79374701',
                                                                       'in_octets': '6582580067884',
                                                                       'in_ucast_pkts': '6141309966',
                                                                       'links': [       {       'in_nucast_pkts': '79374701',
                                                                                                'in_octets': '6582580067884',
                                                                                                'in_ucast_pkts': '6141309966',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16192651',
                                                                                                'out_octets': '1556591803024',
                                                                                                'out_ucast_pkts': '2771493963',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16192651',
                                                                       'out_octets': '1556591803024',
                                                                       'out_ucast_pkts': '2771493963',
                                                                       'poller_hash': 'QkREOUZEcGVlcmludGZwb24gMTU2Nzc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156776',
                                                                                               'links': [       {       'connected_time': '73807588',
                                                                                                                        'fiber_length': 11627,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.578,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QkREOUZEcGVlcnBvbnBvbiAxNTY3NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '4072 PERTH LINE 72 (MILESTONE '
                                                 'EQUIPMENT LTD)',
                                         'poller_hash': 'QkREOUZE',
                                         'serial_number': 'BDD9FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '84:d3:43:34:02:a1',
                                         'uptime': '73807588'},
                                 {       'counter_source': '156788/g1+156788/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:eb:40:35:aa:f0'],
                                                                       'id': '156788/g1',
                                                                       'in_nucast_pkts': '1467',
                                                                       'in_octets': '1155672942544',
                                                                       'in_ucast_pkts': '5272617497',
                                                                       'mac_address': '84:d3:43:7d:32:62',
                                                                       'mtu': 2000,
                                                                       'name': '156788/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '99397388',
                                                                       'out_octets': '8466805037720',
                                                                       'out_ucast_pkts': '7418478287',
                                                                       'poller_hash': 'QzgxQkI1cGVlcmludGYxNTY3ODgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:d4:c4:d6:79:5c'],
                                                                       'id': '156788/g2',
                                                                       'in_nucast_pkts': '699',
                                                                       'in_octets': '25431745836',
                                                                       'in_ucast_pkts': '159070717',
                                                                       'mac_address': '84:d3:43:7d:32:63',
                                                                       'mtu': 2000,
                                                                       'name': '156788/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '99399367',
                                                                       'out_octets': '45922369353',
                                                                       'out_ucast_pkts': '154600795',
                                                                       'poller_hash': 'QzgxQkI1cGVlcmludGYxNTY3ODgvZzI=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156788',
                                                                       'in_nucast_pkts': '198796755',
                                                                       'in_octets': '8512727407073',
                                                                       'in_ucast_pkts': '7573079082',
                                                                       'links': [       {       'in_nucast_pkts': '198796755',
                                                                                                'in_octets': '8512727407073',
                                                                                                'in_ucast_pkts': '7573079082',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2166',
                                                                                                'out_octets': '1181104688380',
                                                                                                'out_ucast_pkts': '5431688214',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2166',
                                                                       'out_octets': '1181104688380',
                                                                       'out_ucast_pkts': '5431688214',
                                                                       'poller_hash': 'QzgxQkI1cGVlcmludGZwb24gMTU2Nzg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156788',
                                                                                               'links': [       {       'connected_time': '73382281',
                                                                                                                        'fiber_length': 11602,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxQkI1cGVlcnBvbnBvbiAxNTY3ODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4044 PERTH LINE 72 (MILLBANK FAMILY '
                                                 'FURNITURE)',
                                         'poller_hash': 'QzgxQkI1',
                                         'serial_number': 'C81BB5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:32:5f',
                                         'uptime': '73382281'},
                                 {       'counter_source': '156808/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:32:b1:e9:de:bc'],
                                                                       'id': '156808/g1',
                                                                       'in_nucast_pkts': '49642',
                                                                       'in_octets': '194733125208',
                                                                       'in_ucast_pkts': '627412543',
                                                                       'mac_address': '84:d3:43:34:02:5d',
                                                                       'mtu': 2000,
                                                                       'name': '156808/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '79318979',
                                                                       'out_octets': '2028732002600',
                                                                       'out_ucast_pkts': '1679300819',
                                                                       'poller_hash': 'QkREOUY0cGVlcmludGYxNTY4MDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156808',
                                                                       'in_nucast_pkts': '79318979',
                                                                       'in_octets': '2028732002600',
                                                                       'in_ucast_pkts': '1679300819',
                                                                       'links': [       {       'in_nucast_pkts': '79318979',
                                                                                                'in_octets': '2028732002600',
                                                                                                'in_ucast_pkts': '1679300819',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '49642',
                                                                                                'out_octets': '194733125208',
                                                                                                'out_ucast_pkts': '627412543',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49642',
                                                                       'out_octets': '194733125208',
                                                                       'out_ucast_pkts': '627412543',
                                                                       'poller_hash': 'QkREOUY0cGVlcmludGZwb24gMTU2ODA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156808',
                                                                                               'links': [       {       'connected_time': '73807576',
                                                                                                                        'fiber_length': 11743,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.16,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QkREOUY0cGVlcnBvbnBvbiAxNTY4MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '4082 PERTH LINE 72 (GOSPEL ECHOES)',
                                         'poller_hash': 'QkREOUY0',
                                         'serial_number': 'BDD9F4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '84:d3:43:34:02:59',
                                         'uptime': '73807576'},
                                 {       'counter_source': '156871/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '156871/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '89852096030',
                                                                       'in_ucast_pkts': '121120392',
                                                                       'mac_address': '84:d3:43:7d:4a:1f',
                                                                       'name': '156871/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '316429571559',
                                                                       'out_ucast_pkts': '305085255',
                                                                       'poller_hash': 'QzgxRUFEcGVlcmludGYxNTY4NzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '156871',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '316429571559',
                                                                       'in_ucast_pkts': '305085255',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '316429571559',
                                                                                                'in_ucast_pkts': '305085255',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '89852096030',
                                                                                                'out_ucast_pkts': '121120392',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '89852096030',
                                                                       'out_ucast_pkts': '121120392',
                                                                       'poller_hash': 'QzgxRUFEcGVlcmludGZwb24gMTU2ODcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156871',
                                                                                               'links': [       {       'connected_time': '24119266',
                                                                                                                        'fiber_length': 622,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUFEcGVlcnBvbnBvbiAxNTY4NzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 MAIN ST N (1262174 ONTARIO INC)',
                                         'poller_hash': 'QzgxRUFE',
                                         'serial_number': 'C81EAD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:1f',
                                         'uptime': '24119266'},
                                 {       'counter_source': '156901/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '156901/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '63983738921',
                                                                       'in_ucast_pkts': '276970456',
                                                                       'mac_address': '84:d3:43:7c:e5:ef',
                                                                       'name': '156901/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '1131945570027',
                                                                       'out_ucast_pkts': '934494007',
                                                                       'poller_hash': 'QzgxMjI3cGVlcmludGYxNTY5MDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '156901',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '1131945570027',
                                                                       'in_ucast_pkts': '934494007',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '1131945570027',
                                                                                                'in_ucast_pkts': '934494007',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '63983738921',
                                                                                                'out_ucast_pkts': '276970456',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '63983738921',
                                                                       'out_ucast_pkts': '276970456',
                                                                       'poller_hash': 'QzgxMjI3cGVlcmludGZwb24gMTU2OTAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156901',
                                                                                               'links': [       {       'connected_time': '29301859',
                                                                                                                        'fiber_length': 6489,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxMjI3cGVlcnBvbnBvbiAxNTY5MDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6313 ROAD 124 (STREICHER MURRAY & '
                                                 'JOANNE)',
                                         'poller_hash': 'QzgxMjI3',
                                         'serial_number': 'C81227',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7c:e5:ef',
                                         'uptime': '29301859'},
                                 {       'counter_source': '156931/g1+156931/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e0:55:3d:89:d9:c5'],
                                                                       'id': '156931/g1',
                                                                       'in_nucast_pkts': '3588238',
                                                                       'in_octets': '1368773497754',
                                                                       'in_ucast_pkts': '4578020848',
                                                                       'mac_address': '84:d3:43:7c:e3:ca',
                                                                       'mtu': 2000,
                                                                       'name': '156931/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43616850',
                                                                       'out_octets': '8694045373783',
                                                                       'out_ucast_pkts': '7715831396',
                                                                       'poller_hash': 'QzgxMUUycGVlcmludGYxNTY5MzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:0d:b9:55:21:19'],
                                                                       'id': '156931/g2',
                                                                       'in_nucast_pkts': '8516',
                                                                       'in_octets': '18859142131',
                                                                       'in_ucast_pkts': '85143108',
                                                                       'mac_address': '84:d3:43:7c:e3:cb',
                                                                       'mtu': 2000,
                                                                       'name': '156931/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23249083',
                                                                       'out_octets': '21355144657',
                                                                       'out_ucast_pkts': '84233853',
                                                                       'poller_hash': 'QzgxMUUycGVlcmludGYxNTY5MzEvZzI=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156931',
                                                                       'in_nucast_pkts': '66865933',
                                                                       'in_octets': '8715400518440',
                                                                       'in_ucast_pkts': '7800065249',
                                                                       'links': [       {       'in_nucast_pkts': '66865933',
                                                                                                'in_octets': '8715400518440',
                                                                                                'in_ucast_pkts': '7800065249',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3596754',
                                                                                                'out_octets': '1387632639885',
                                                                                                'out_ucast_pkts': '4663163956',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3596754',
                                                                       'out_octets': '1387632639885',
                                                                       'out_ucast_pkts': '4663163956',
                                                                       'poller_hash': 'QzgxMUUycGVlcmludGZwb24gMTU2OTMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156931',
                                                                                               'links': [       {       'connected_time': '2073770',
                                                                                                                        'fiber_length': 11471,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxMUUycGVlcnBvbnBvbiAxNTY5MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4060 PERTH LINE 72 (MILLBANK HOME '
                                                 'BAKERY & COUNTRY)',
                                         'poller_hash': 'QzgxMUUy',
                                         'serial_number': 'C811E2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7c:e3:c7',
                                         'uptime': '2073770'},
                                 {       'counter_source': '156951/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d3:09'],
                                                                       'id': '156951/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '19082349991',
                                                                       'in_ucast_pkts': '60054411',
                                                                       'mac_address': '84:d3:43:7c:e4:aa',
                                                                       'mtu': 2000,
                                                                       'name': '156951/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464401',
                                                                       'out_octets': '301368232894',
                                                                       'out_ucast_pkts': '210551882',
                                                                       'poller_hash': 'QzgxMUZFcGVlcmludGYxNTY5NTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156951',
                                                                       'in_nucast_pkts': '1464401',
                                                                       'in_octets': '301368232894',
                                                                       'in_ucast_pkts': '210551882',
                                                                       'links': [       {       'in_nucast_pkts': '1464401',
                                                                                                'in_octets': '301368232894',
                                                                                                'in_ucast_pkts': '210551882',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '19082349991',
                                                                                                'out_ucast_pkts': '60054411',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '19082349991',
                                                                       'out_ucast_pkts': '60054411',
                                                                       'poller_hash': 'QzgxMUZFcGVlcmludGZwb24gMTU2OTUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156951',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 2777,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxMUZFcGVlcnBvbnBvbiAxNTY5NTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '65 PUGH ST (REKTOR SHANE)',
                                         'poller_hash': 'QzgxMUZF',
                                         'serial_number': 'C811FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7c:e4:a7',
                                         'uptime': '688370'},
                                 {       'counter_source': '156982/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c8:4f:86:fc:00:02'],
                                                                       'id': '156982/g1',
                                                                       'in_nucast_pkts': '46069285',
                                                                       'in_octets': '158305074114725',
                                                                       'in_ucast_pkts': '135456167952',
                                                                       'mac_address': 'ec:4f:82:7a:25:5b',
                                                                       'mtu': 2000,
                                                                       'name': '156982/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '512920193',
                                                                       'out_octets': '64512046185687',
                                                                       'out_ucast_pkts': '119026071472',
                                                                       'poller_hash': 'MkYzQjE3cGVlcmludGYxNTY5ODIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156982',
                                                                       'in_nucast_pkts': '512920193',
                                                                       'in_octets': '64512046185687',
                                                                       'in_ucast_pkts': '119026071472',
                                                                       'links': [       {       'in_nucast_pkts': '512920193',
                                                                                                'in_octets': '64512046185687',
                                                                                                'in_ucast_pkts': '119026071472',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '46069285',
                                                                                                'out_octets': '158305074114725',
                                                                                                'out_ucast_pkts': '135456167952',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '46069285',
                                                                       'out_octets': '158305074114725',
                                                                       'out_ucast_pkts': '135456167952',
                                                                       'poller_hash': 'MkYzQjE3cGVlcmludGZwb24gMTU2OTgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156982',
                                                                                               'links': [       {       'connected_time': '73807968',
                                                                                                                        'fiber_length': 11576,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.354,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkYzQjE3cGVlcnBvbnBvbiAxNTY5ODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '7003 Raymond St (MIllbank Hardware)',
                                         'poller_hash': 'MkYzQjE3',
                                         'serial_number': '2F3B17',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:7a:25:57',
                                         'uptime': '73807968'},
                                 {       'counter_source': '157284/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:bc:9e'],
                                                                       'id': '157284/g1',
                                                                       'in_nucast_pkts': '14290',
                                                                       'in_octets': '9772355441',
                                                                       'in_ucast_pkts': '72986987',
                                                                       'mac_address': '84:d3:43:7e:46:8a',
                                                                       'mtu': 2000,
                                                                       'name': '157284/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447054',
                                                                       'out_octets': '446250600385',
                                                                       'out_ucast_pkts': '310587970',
                                                                       'poller_hash': 'QzgzRTNBcGVlcmludGYxNTcyODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157284',
                                                                       'in_nucast_pkts': '1447054',
                                                                       'in_octets': '446250600385',
                                                                       'in_ucast_pkts': '310587970',
                                                                       'links': [       {       'in_nucast_pkts': '1447054',
                                                                                                'in_octets': '446250600385',
                                                                                                'in_ucast_pkts': '310587970',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14290',
                                                                                                'out_octets': '9772355441',
                                                                                                'out_ucast_pkts': '72986987',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14290',
                                                                       'out_octets': '9772355441',
                                                                       'out_ucast_pkts': '72986987',
                                                                       'poller_hash': 'QzgzRTNBcGVlcmludGZwb24gMTU3Mjg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157284',
                                                                                               'links': [       {       'connected_time': '688415',
                                                                                                                        'fiber_length': 2112,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzRTNBcGVlcnBvbnBvbiAxNTcyODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '83 GERBER DR (JAMES COLLEEN)',
                                         'poller_hash': 'QzgzRTNB',
                                         'serial_number': 'C83E3A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:46:87',
                                         'uptime': '688415'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '157444/g1',
                                                                       'name': '157444/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRDhFcGVlcmludGYxNTc0NDQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '157444',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxRDhFcGVlcmludGZwb24gMTU3NDQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157444',
                                                                                               'links': [       {       'connected_time': '6377864',
                                                                                                                        'fiber_length': 10142,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDhFcGVlcnBvbnBvbiAxNTc0NDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5905 PERTH ROAD 119 (JANTZI MARY J)',
                                         'poller_hash': 'QzgxRDhF',
                                         'serial_number': 'C81D8E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:41:27',
                                         'uptime': '6377864'},
                                 {       'counter_source': '157743/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b7:a2'],
                                                                       'id': '157743/g1',
                                                                       'in_nucast_pkts': '14880',
                                                                       'in_octets': '4248064986',
                                                                       'in_ucast_pkts': '9326994',
                                                                       'mac_address': '84:d3:43:7d:49:aa',
                                                                       'mtu': 2000,
                                                                       'name': '157743/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464287',
                                                                       'out_octets': '78166586253',
                                                                       'out_ucast_pkts': '56106253',
                                                                       'poller_hash': 'QzgxRTlFcGVlcmludGYxNTc3NDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157743',
                                                                       'in_nucast_pkts': '1464287',
                                                                       'in_octets': '78166586253',
                                                                       'in_ucast_pkts': '56106253',
                                                                       'links': [       {       'in_nucast_pkts': '1464287',
                                                                                                'in_octets': '78166586253',
                                                                                                'in_ucast_pkts': '56106253',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14880',
                                                                                                'out_octets': '4248064986',
                                                                                                'out_ucast_pkts': '9326994',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14880',
                                                                       'out_octets': '4248064986',
                                                                       'out_ucast_pkts': '9326994',
                                                                       'poller_hash': 'QzgxRTlFcGVlcmludGZwb24gMTU3NzQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157743',
                                                                                               'links': [       {       'connected_time': '688365',
                                                                                                                        'fiber_length': 2845,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTlFcGVlcnBvbnBvbiAxNTc3NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '147 PUGH ST (BLOUSE ASHLEE)',
                                         'poller_hash': 'QzgxRTlF',
                                         'serial_number': 'C81E9E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:49:a7',
                                         'uptime': '688365'},
                                 {       'counter_source': '158000/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '158000/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '165706073113',
                                                                       'in_ucast_pkts': '668258293',
                                                                       'mac_address': '84:d3:43:7c:e4:e7',
                                                                       'name': '158000/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '3953326007530',
                                                                       'out_ucast_pkts': '3112066813',
                                                                       'poller_hash': 'QzgxMjA2cGVlcmludGYxNTgwMDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '158000',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '3953326007530',
                                                                       'in_ucast_pkts': '3112066813',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '3953326007530',
                                                                                                'in_ucast_pkts': '3112066813',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '165706073113',
                                                                                                'out_ucast_pkts': '668258293',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '165706073113',
                                                                       'out_ucast_pkts': '668258293',
                                                                       'poller_hash': 'QzgxMjA2cGVlcmludGZwb24gMTU4MDAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '158000',
                                                                                               'links': [       {       'connected_time': '688363',
                                                                                                                        'fiber_length': 2865,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxMjA2cGVlcnBvbnBvbiAxNTgwMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '145 PUGH ST (HAMMER PAM)',
                                         'poller_hash': 'QzgxMjA2',
                                         'serial_number': 'C81206',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7c:e4:e7',
                                         'uptime': '688363'},
                                 {       'counter_source': '162296/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b5:5b'],
                                                                       'id': '162296/g1',
                                                                       'in_nucast_pkts': '170611',
                                                                       'in_octets': '34580550909',
                                                                       'in_ucast_pkts': '41934402',
                                                                       'mac_address': '84:d3:43:7d:5b:4a',
                                                                       'mtu': 2000,
                                                                       'name': '162296/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20149158',
                                                                       'out_octets': '44527335516',
                                                                       'out_ucast_pkts': '47944443',
                                                                       'poller_hash': 'QzgyMEQycGVlcmludGYxNjIyOTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '162296',
                                                                       'in_nucast_pkts': '20149158',
                                                                       'in_octets': '44527335516',
                                                                       'in_ucast_pkts': '47944443',
                                                                       'links': [       {       'in_nucast_pkts': '20149158',
                                                                                                'in_octets': '44527335516',
                                                                                                'in_ucast_pkts': '47944443',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '170611',
                                                                                                'out_octets': '34580550909',
                                                                                                'out_ucast_pkts': '41934402',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '170611',
                                                                       'out_octets': '34580550909',
                                                                       'out_ucast_pkts': '41934402',
                                                                       'poller_hash': 'QzgyMEQycGVlcmludGZwb24gMTYyMjk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '162296',
                                                                                               'links': [       {       'connected_time': '9239014',
                                                                                                                        'fiber_length': 18282,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgyMEQycGVlcnBvbnBvbiAxNjIyOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '160 QUEEN AVENUE MONKTON (GERBER '
                                                 'MICHAEL)',
                                         'poller_hash': 'QzgyMEQy',
                                         'serial_number': 'C820D2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:5b:47',
                                         'uptime': '9239014'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '162473/g1',
                                                                       'name': '162473/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzQjc0cGVlcmludGYxNjI0NzMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '162473',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgzQjc0cGVlcmludGZwb24gMTYyNDcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '162473',
                                                                                               'links': [       {       'connected_time': '21816543',
                                                                                                                        'fiber_length': 10575,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjc0cGVlcnBvbnBvbiAxNjI0NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5920 SOMERSET DR (WIEBE GERHARD)',
                                         'poller_hash': 'QzgzQjc0',
                                         'serial_number': 'C83B74',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:30:57',
                                         'uptime': '21816543'},
                                 {       'counter_source': '163068/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '163068/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1363681588879',
                                                                       'in_ucast_pkts': '3754966827',
                                                                       'mac_address': 'cc:be:59:5a:fe:62',
                                                                       'name': '163068/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2186478460565',
                                                                       'out_ucast_pkts': '4439437311',
                                                                       'poller_hash': 'NDRBQThGcGVlcmludGYxNjMwNjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '163068',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2186478460565',
                                                                       'in_ucast_pkts': '4439437311',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2186478460565',
                                                                                                'in_ucast_pkts': '4439437311',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1363681588879',
                                                                                                'out_ucast_pkts': '3754966827',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1363681588879',
                                                                       'out_ucast_pkts': '3754966827',
                                                                       'poller_hash': 'NDRBQThGcGVlcmludGZwb24gMTYzMDY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '163068',
                                                                                               'links': [       {       'connected_time': '13124259',
                                                                                                                        'fiber_length': 526,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.452,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDRBQThGcGVlcnBvbnBvbiAxNjMwNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '8 MAIN ST N MILVERTON (VILLAGE '
                                                 'PHARMACY)',
                                         'poller_hash': 'NDRBQThG',
                                         'serial_number': '44AA8F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:5a:fe:62',
                                         'uptime': '13124259'},
                                 {       'counter_source': '163309/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '163309/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '67348657846',
                                                                       'in_ucast_pkts': '274571166',
                                                                       'mac_address': '84:d3:43:7e:01:cf',
                                                                       'name': '163309/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1631398275052',
                                                                       'out_ucast_pkts': '1167797243',
                                                                       'poller_hash': 'QzgzNUEzcGVlcmludGYxNjMzMDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '163309',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1631398275052',
                                                                       'in_ucast_pkts': '1167797243',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1631398275052',
                                                                                                'in_ucast_pkts': '1167797243',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '67348657846',
                                                                                                'out_ucast_pkts': '274571166',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '67348657846',
                                                                       'out_ucast_pkts': '274571166',
                                                                       'poller_hash': 'QzgzNUEzcGVlcmludGZwb24gMTYzMzA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '163309',
                                                                                               'links': [       {       'connected_time': '688387',
                                                                                                                        'fiber_length': 10436,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzNUEzcGVlcnBvbnBvbiAxNjMzMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5590 ROAD 126 (ZACH TREVOR)',
                                         'poller_hash': 'QzgzNUEz',
                                         'serial_number': 'C835A3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:01:cf',
                                         'uptime': '688387'},
                                 {       'counter_source': '163653/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7d:83'],
                                                                       'id': '163653/g1',
                                                                       'in_nucast_pkts': '13636',
                                                                       'in_octets': '4790245659',
                                                                       'in_ucast_pkts': '21722447',
                                                                       'mac_address': '00:06:31:e8:e3:a8',
                                                                       'mtu': 2000,
                                                                       'name': '163653/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464139',
                                                                       'out_octets': '83355296149',
                                                                       'out_ucast_pkts': '66293349',
                                                                       'poller_hash': 'MUQwMzM3cGVlcmludGYxNjM2NTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '163653',
                                                                       'in_nucast_pkts': '1464139',
                                                                       'in_octets': '83355296149',
                                                                       'in_ucast_pkts': '66293349',
                                                                       'links': [       {       'in_nucast_pkts': '1464139',
                                                                                                'in_octets': '83355296149',
                                                                                                'in_ucast_pkts': '66293349',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13636',
                                                                                                'out_octets': '4790245659',
                                                                                                'out_ucast_pkts': '21722447',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13636',
                                                                       'out_octets': '4790245659',
                                                                       'out_ucast_pkts': '21722447',
                                                                       'poller_hash': 'MUQwMzM3cGVlcmludGZwb24gMTYzNjUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '163653',
                                                                                               'links': [       {       'connected_time': '688495',
                                                                                                                        'fiber_length': 600,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwMzM3cGVlcnBvbnBvbiAxNjM2NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '16 MAIN ST N (PATTERSON REBEKKAH)',
                                         'poller_hash': 'MUQwMzM3',
                                         'serial_number': '1D0337',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:e3:a5',
                                         'uptime': '688495'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '164156/g1',
                                                                       'name': '164156/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QjBBQUYxcGVlcmludGYxNjQxNTYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '164156',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QjBBQUYxcGVlcmludGZwb24gMTY0MTU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164156',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBQUYxcGVlcnBvbnBvbiAxNjQxNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '55 WILLIAM ST 207-2',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "164156" '
                                                                  '(location: GPON '
                                                                  'port "7/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-26T06:02:39.900Z',
                                         'poller_hash': 'QjBBQUYx',
                                         'serial_number': 'B0AAF1',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '164705/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '164705/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '106818989799',
                                                                       'in_ucast_pkts': '395029717',
                                                                       'mac_address': '84:d3:43:7e:1f:57',
                                                                       'name': '164705/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '6',
                                                                       'out_octets': '1552726919848',
                                                                       'out_ucast_pkts': '1192973000',
                                                                       'poller_hash': 'QzgzOTU0cGVlcmludGYxNjQ3MDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '164705',
                                                                       'in_nucast_pkts': '6',
                                                                       'in_octets': '1552726919848',
                                                                       'in_ucast_pkts': '1192973000',
                                                                       'links': [       {       'in_nucast_pkts': '6',
                                                                                                'in_octets': '1552726919848',
                                                                                                'in_ucast_pkts': '1192973000',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '106818989799',
                                                                                                'out_ucast_pkts': '395029717',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '106818989799',
                                                                       'out_ucast_pkts': '395029717',
                                                                       'poller_hash': 'QzgzOTU0cGVlcmludGZwb24gMTY0NzA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164705',
                                                                                               'links': [       {       'connected_time': '688542',
                                                                                                                        'fiber_length': 7517,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzOTU0cGVlcnBvbnBvbiAxNjQ3MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6213 Rd 124 (VAN NES MICHELLE)',
                                         'poller_hash': 'QzgzOTU0',
                                         'serial_number': 'C83954',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:1f:57',
                                         'uptime': '688542'},
                                 {       'counter_source': '164805/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '164805/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '39249919677',
                                                                       'in_ucast_pkts': '72036623',
                                                                       'mac_address': '84:d3:43:95:8f:9a',
                                                                       'name': '164805/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '164944559804',
                                                                       'out_ucast_pkts': '152153486',
                                                                       'poller_hash': 'Q0JEMjEycGVlcmludGYxNjQ4MDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '164805',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '164944559804',
                                                                       'in_ucast_pkts': '152153486',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '164944559804',
                                                                                                'in_ucast_pkts': '152153486',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '39249919677',
                                                                                                'out_ucast_pkts': '72036623',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '39249919677',
                                                                       'out_ucast_pkts': '72036623',
                                                                       'poller_hash': 'Q0JEMjEycGVlcmludGZwb24gMTY0ODA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164805',
                                                                                               'links': [       {       'connected_time': '21815666',
                                                                                                                        'fiber_length': 10161,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjEycGVlcnBvbnBvbiAxNjQ4MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5908 PERTH ROAD 119 (JANTZI MYRON)',
                                         'poller_hash': 'Q0JEMjEy',
                                         'serial_number': 'CBD212',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8f:9a',
                                         'uptime': '21815666'},
                                 {       'counter_source': '164815/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:fa:9f'],
                                                                       'id': '164815/g1',
                                                                       'in_nucast_pkts': '12033',
                                                                       'in_octets': '10687287610',
                                                                       'in_ucast_pkts': '35022406',
                                                                       'mac_address': '84:d3:43:95:8e:4d',
                                                                       'mtu': 2000,
                                                                       'name': '164815/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464272',
                                                                       'out_octets': '160399446628',
                                                                       'out_ucast_pkts': '119028605',
                                                                       'poller_hash': 'Q0JEMUU4cGVlcmludGYxNjQ4MTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '164815',
                                                                       'in_nucast_pkts': '1464272',
                                                                       'in_octets': '160399446628',
                                                                       'in_ucast_pkts': '119028605',
                                                                       'links': [       {       'in_nucast_pkts': '1464272',
                                                                                                'in_octets': '160399446628',
                                                                                                'in_ucast_pkts': '119028605',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12033',
                                                                                                'out_octets': '10687287610',
                                                                                                'out_ucast_pkts': '35022406',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12033',
                                                                       'out_octets': '10687287610',
                                                                       'out_ucast_pkts': '35022406',
                                                                       'poller_hash': 'Q0JEMUU4cGVlcmludGZwb24gMTY0ODE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164815',
                                                                                               'links': [       {       'connected_time': '688500',
                                                                                                                        'fiber_length': 1861,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUU4cGVlcnBvbnBvbiAxNjQ4MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '48 MAIN ST N (SCHLOSSER KYLE)',
                                         'poller_hash': 'Q0JEMUU4',
                                         'serial_number': 'CBD1E8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8e:4a',
                                         'uptime': '688500'},
                                 {       'counter_source': '165100/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '165100/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '366565236459',
                                                                       'in_ucast_pkts': '32096242129',
                                                                       'mac_address': '84:d3:43:95:8a:6a',
                                                                       'name': '165100/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '16071240654554',
                                                                       'out_ucast_pkts': '11187686402',
                                                                       'poller_hash': 'Q0JEMTZDcGVlcmludGYxNjUxMDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '165100',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '16071240654554',
                                                                       'in_ucast_pkts': '11187686402',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '16071240654554',
                                                                                                'in_ucast_pkts': '11187686402',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '366565236459',
                                                                                                'out_ucast_pkts': '32096242129',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '366565236459',
                                                                       'out_ucast_pkts': '32096242129',
                                                                       'poller_hash': 'Q0JEMTZDcGVlcmludGZwb24gMTY1MTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '165100',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 2840,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTZDcGVlcnBvbnBvbiAxNjUxMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '149 PUGH ST (WRIGHT BRITTANY)',
                                         'poller_hash': 'Q0JEMTZD',
                                         'serial_number': 'CBD16C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8a:6a',
                                         'uptime': '688369'},
                                 {       'counter_source': '165478/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:6e:84:00:aa:1d'],
                                                                       'id': '165478/g1',
                                                                       'in_nucast_pkts': '1485',
                                                                       'in_octets': '39213265652',
                                                                       'in_ucast_pkts': '90035054',
                                                                       'mac_address': '04:bc:9f:6f:00:59',
                                                                       'mtu': 2000,
                                                                       'name': '165478/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4729068',
                                                                       'out_octets': '204556272886',
                                                                       'out_ucast_pkts': '167804813',
                                                                       'poller_hash': 'MTI5NzM4RXBlZXJpbnRmMTY1NDc4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '165478',
                                                                       'in_nucast_pkts': '4729068',
                                                                       'in_octets': '204556272886',
                                                                       'in_ucast_pkts': '167804813',
                                                                       'links': [       {       'in_nucast_pkts': '4729068',
                                                                                                'in_octets': '204556272886',
                                                                                                'in_ucast_pkts': '167804813',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1485',
                                                                                                'out_octets': '39213265652',
                                                                                                'out_ucast_pkts': '90035054',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1485',
                                                                       'out_octets': '39213265652',
                                                                       'out_ucast_pkts': '90035054',
                                                                       'poller_hash': 'MTI5NzM4RXBlZXJpbnRmcG9uIDE2NTQ3OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '165478',
                                                                                               'links': [       {       'connected_time': '2181353',
                                                                                                                        'fiber_length': 7092,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM4RXBlZXJwb25wb24gMTY1NDc4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4399 LINE 61 (HAVERKAMP RALPH)',
                                         'poller_hash': 'MTI5NzM4RQ==',
                                         'serial_number': '129738E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:00:56',
                                         'uptime': '2181353'},
                                 {       'counter_source': '166211/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '166211/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '498181554242',
                                                                       'in_ucast_pkts': '5246290809',
                                                                       'mac_address': 'cc:be:59:6e:86:1e',
                                                                       'name': '166211/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '26620967230644',
                                                                       'out_ucast_pkts': '18837897906',
                                                                       'poller_hash': 'NDcxNzU1cGVlcmludGYxNjYyMTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '166211',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '26620967230644',
                                                                       'in_ucast_pkts': '18837897906',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '26620967230644',
                                                                                                'in_ucast_pkts': '18837897906',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '498181554242',
                                                                                                'out_ucast_pkts': '5246290809',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '498181554242',
                                                                       'out_ucast_pkts': '5246290809',
                                                                       'poller_hash': 'NDcxNzU1cGVlcmludGZwb24gMTY2MjEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166211',
                                                                                               'links': [       {       'connected_time': '995931',
                                                                                                                        'fiber_length': 9432,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDcxNzU1cGVlcnBvbnBvbiAxNjYyMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5387 LINE 60 (COXON TYE M)',
                                         'poller_hash': 'NDcxNzU1',
                                         'serial_number': '471755',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6e:86:1e',
                                         'uptime': '995931'},
                                 {       'counter_source': '166338/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d6:69'],
                                                                       'id': '166338/g1',
                                                                       'in_nucast_pkts': '12756',
                                                                       'in_octets': '53263234947',
                                                                       'in_ucast_pkts': '260152553',
                                                                       'mac_address': '84:d3:43:95:90:2d',
                                                                       'mtu': 2000,
                                                                       'name': '166338/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447239',
                                                                       'out_octets': '832023675489',
                                                                       'out_ucast_pkts': '636435103',
                                                                       'poller_hash': 'Q0JEMjI0cGVlcmludGYxNjYzMzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '166338',
                                                                       'in_nucast_pkts': '1447239',
                                                                       'in_octets': '832023675489',
                                                                       'in_ucast_pkts': '636435103',
                                                                       'links': [       {       'in_nucast_pkts': '1447239',
                                                                                                'in_octets': '832023675489',
                                                                                                'in_ucast_pkts': '636435103',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12756',
                                                                                                'out_octets': '53263234947',
                                                                                                'out_ucast_pkts': '260152553',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12756',
                                                                       'out_octets': '53263234947',
                                                                       'out_ucast_pkts': '260152553',
                                                                       'poller_hash': 'Q0JEMjI0cGVlcmludGZwb24gMTY2MzM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166338',
                                                                                               'links': [       {       'connected_time': '688412',
                                                                                                                        'fiber_length': 1970,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjI0cGVlcnBvbnBvbiAxNjYzMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '91 REAGAN ST (CASTRO FATIMA  D)',
                                         'poller_hash': 'Q0JEMjI0',
                                         'serial_number': 'CBD224',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:90:2a',
                                         'uptime': '688412'},
                                 {       'counter_source': '166408/F1+166408/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:14:3c:4a',
                                                                                                        '5c:0f:fb:17:66:c8'],
                                                                       'id': '166408/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9804074871',
                                                                       'in_ucast_pkts': '24145031',
                                                                       'name': '166408/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '3026924630',
                                                                       'out_octets': '3984739016610',
                                                                       'out_ucast_pkts': '3714521',
                                                                       'poller_hash': 'Q0JEMjIycGVlcmludGYxNjY0MDgvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f7:00'],
                                                                       'id': '166408/g1',
                                                                       'in_nucast_pkts': '14832',
                                                                       'in_octets': '11124365996',
                                                                       'in_ucast_pkts': '54412862',
                                                                       'mac_address': '84:d3:43:95:90:1d',
                                                                       'mtu': 2000,
                                                                       'name': '166408/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447194',
                                                                       'out_octets': '210253842750',
                                                                       'out_ucast_pkts': '159078170',
                                                                       'poller_hash': 'Q0JEMjIycGVlcmludGYxNjY0MDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '166408',
                                                                       'in_nucast_pkts': '3028371824',
                                                                       'in_octets': '4194992859360',
                                                                       'in_ucast_pkts': '162792691',
                                                                       'links': [       {       'in_nucast_pkts': '3028371824',
                                                                                                'in_octets': '4194992859360',
                                                                                                'in_ucast_pkts': '162792691',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14832',
                                                                                                'out_octets': '20928440867',
                                                                                                'out_ucast_pkts': '78557893',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14832',
                                                                       'out_octets': '20928440867',
                                                                       'out_ucast_pkts': '78557893',
                                                                       'poller_hash': 'Q0JEMjIycGVlcmludGZwb24gMTY2NDA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166408',
                                                                                               'links': [       {       'connected_time': '688419',
                                                                                                                        'fiber_length': 2006,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjIycGVlcnBvbnBvbiAxNjY0MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '89 GERBER DR (ALVES JORGE)',
                                         'poller_hash': 'Q0JEMjIy',
                                         'serial_number': 'CBD222',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:90:1a',
                                         'uptime': '688419'},
                                 {       'counter_source': '166608/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '166608/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '483385924539',
                                                                       'in_ucast_pkts': '3766519016',
                                                                       'mac_address': '84:d3:43:7e:2f:d7',
                                                                       'name': '166608/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '81',
                                                                       'out_octets': '12035486252307',
                                                                       'out_ucast_pkts': '8682905663',
                                                                       'poller_hash': 'QzgzQjY0cGVlcmludGYxNjY2MDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '166608',
                                                                       'in_nucast_pkts': '81',
                                                                       'in_octets': '12035486252307',
                                                                       'in_ucast_pkts': '8682905663',
                                                                       'links': [       {       'in_nucast_pkts': '81',
                                                                                                'in_octets': '12035486252307',
                                                                                                'in_ucast_pkts': '8682905663',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '483385924539',
                                                                                                'out_ucast_pkts': '3766519016',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '483385924539',
                                                                       'out_ucast_pkts': '3766519016',
                                                                       'poller_hash': 'QzgzQjY0cGVlcmludGZwb24gMTY2NjA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166608',
                                                                                               'links': [       {       'connected_time': '538563',
                                                                                                                        'fiber_length': 14175,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjY0cGVlcnBvbnBvbiAxNjY2MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4939 LINE 49 (BENNEWIES TREVOR)',
                                         'poller_hash': 'QzgzQjY0',
                                         'serial_number': 'C83B64',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:2f:d7',
                                         'uptime': '538563'},
                                 {       'counter_source': '166678/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '166678/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '137959005841',
                                                                       'in_ucast_pkts': '5280337672',
                                                                       'mac_address': '84:d3:43:7e:2d:27',
                                                                       'name': '166678/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '217',
                                                                       'out_octets': '3115638108413',
                                                                       'out_ucast_pkts': '2220388981',
                                                                       'poller_hash': 'QzgzQjBFcGVlcmludGYxNjY2NzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '166678',
                                                                       'in_nucast_pkts': '217',
                                                                       'in_octets': '3115638108413',
                                                                       'in_ucast_pkts': '2220388981',
                                                                       'links': [       {       'in_nucast_pkts': '217',
                                                                                                'in_octets': '3115638108413',
                                                                                                'in_ucast_pkts': '2220388981',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '137959005841',
                                                                                                'out_ucast_pkts': '5280337672',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '137959005841',
                                                                       'out_ucast_pkts': '5280337672',
                                                                       'poller_hash': 'QzgzQjBFcGVlcmludGZwb24gMTY2Njc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166678',
                                                                                               'links': [       {       'connected_time': '688378',
                                                                                                                        'fiber_length': 6950,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjBFcGVlcnBvbnBvbiAxNjY2Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6280 ROAD 140',
                                         'poller_hash': 'QzgzQjBF',
                                         'serial_number': 'C83B0E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:2d:27',
                                         'uptime': '688378'},
                                 {       'counter_source': '167177/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '167177/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '246013802141',
                                                                       'in_ucast_pkts': '770623819',
                                                                       'mac_address': '84:d3:43:95:8a:72',
                                                                       'name': '167177/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '47',
                                                                       'out_octets': '3002657171821',
                                                                       'out_ucast_pkts': '2454583627',
                                                                       'poller_hash': 'Q0JEMTZEcGVlcmludGYxNjcxNzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '167177',
                                                                       'in_nucast_pkts': '47',
                                                                       'in_octets': '3002657171821',
                                                                       'in_ucast_pkts': '2454583627',
                                                                       'links': [       {       'in_nucast_pkts': '47',
                                                                                                'in_octets': '3002657171821',
                                                                                                'in_ucast_pkts': '2454583627',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '246013802141',
                                                                                                'out_ucast_pkts': '770623819',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '246013802141',
                                                                       'out_ucast_pkts': '770623819',
                                                                       'poller_hash': 'Q0JEMTZEcGVlcmludGZwb24gMTY3MTc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167177',
                                                                                               'links': [       {       'connected_time': '688364',
                                                                                                                        'fiber_length': 2823,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTZEcGVlcnBvbnBvbiAxNjcxNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '151 PUGH ST (PICKERING RICHARD)',
                                         'poller_hash': 'Q0JEMTZE',
                                         'serial_number': 'CBD16D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8a:72',
                                         'uptime': '688364'},
                                 {       'counter_source': '167257/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '167257/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '280508898579',
                                                                       'in_ucast_pkts': '1059681632',
                                                                       'mac_address': '84:d3:43:95:8b:52',
                                                                       'name': '167257/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '182',
                                                                       'out_octets': '3045951279029',
                                                                       'out_ucast_pkts': '2595037547',
                                                                       'poller_hash': 'Q0JEMTg5cGVlcmludGYxNjcyNTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '167257',
                                                                       'in_nucast_pkts': '182',
                                                                       'in_octets': '3045951279029',
                                                                       'in_ucast_pkts': '2595037547',
                                                                       'links': [       {       'in_nucast_pkts': '182',
                                                                                                'in_octets': '3045951279029',
                                                                                                'in_ucast_pkts': '2595037547',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '280508898579',
                                                                                                'out_ucast_pkts': '1059681632',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '280508898579',
                                                                       'out_ucast_pkts': '1059681632',
                                                                       'poller_hash': 'Q0JEMTg5cGVlcmludGZwb24gMTY3MjU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167257',
                                                                                               'links': [       {       'connected_time': '10799647',
                                                                                                                        'fiber_length': 11877,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTg5cGVlcnBvbnBvbiAxNjcyNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6975 FREEBORN ST (CHAPMAN GORDON R)',
                                         'poller_hash': 'Q0JEMTg5',
                                         'serial_number': 'CBD189',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8b:52',
                                         'uptime': '10799647'},
                                 {       'counter_source': '167848/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:bc:38'],
                                                                       'id': '167848/g1',
                                                                       'in_nucast_pkts': '38340',
                                                                       'in_octets': '2834585558',
                                                                       'in_ucast_pkts': '38891785',
                                                                       'mac_address': '04:bc:9f:5c:1b:fc',
                                                                       'mtu': 2000,
                                                                       'name': '167848/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15403089',
                                                                       'out_octets': '222763755040',
                                                                       'out_ucast_pkts': '154144071',
                                                                       'poller_hash': 'MTI3NDlGRnBlZXJpbnRmMTY3ODQ4L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '167848',
                                                                       'in_nucast_pkts': '15403089',
                                                                       'in_octets': '222763755040',
                                                                       'in_ucast_pkts': '154144071',
                                                                       'links': [       {       'in_nucast_pkts': '15403089',
                                                                                                'in_octets': '222763755040',
                                                                                                'in_ucast_pkts': '154144071',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '38340',
                                                                                                'out_octets': '2834585558',
                                                                                                'out_ucast_pkts': '38891785',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38340',
                                                                       'out_octets': '2834585558',
                                                                       'out_ucast_pkts': '38891785',
                                                                       'poller_hash': 'MTI3NDlGRnBlZXJpbnRmcG9uIDE2Nzg0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167848',
                                                                                               'links': [       {       'connected_time': '688458',
                                                                                                                        'fiber_length': 1575,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlGRnBlZXJwb25wb24gMTY3ODQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '57 WILLIAM ST (PLAGER JEFFREY)',
                                         'poller_hash': 'MTI3NDlGRg==',
                                         'serial_number': '12749FF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:f9',
                                         'uptime': '688458'},
                                 {       'counter_source': '168232/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '168232/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '256972643908',
                                                                       'in_ucast_pkts': '1910934326',
                                                                       'mac_address': '84:d3:43:95:90:a2',
                                                                       'name': '168232/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '190',
                                                                       'out_octets': '9258717304557',
                                                                       'out_ucast_pkts': '6710554421',
                                                                       'poller_hash': 'Q0JEMjMzcGVlcmludGYxNjgyMzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '168232',
                                                                       'in_nucast_pkts': '190',
                                                                       'in_octets': '9258717304557',
                                                                       'in_ucast_pkts': '6710554421',
                                                                       'links': [       {       'in_nucast_pkts': '190',
                                                                                                'in_octets': '9258717304557',
                                                                                                'in_ucast_pkts': '6710554421',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '256972643908',
                                                                                                'out_ucast_pkts': '1910934326',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '256972643908',
                                                                       'out_ucast_pkts': '1910934326',
                                                                       'poller_hash': 'Q0JEMjMzcGVlcmludGZwb24gMTY4MjMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168232',
                                                                                               'links': [       {       'connected_time': '7865213',
                                                                                                                        'fiber_length': 1088,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjMzcGVlcnBvbnBvbiAxNjgyMzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '32 GERBER DRIVE (NAFZIGER BARB)',
                                         'poller_hash': 'Q0JEMjMz',
                                         'serial_number': 'CBD233',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:90:a2',
                                         'uptime': '7865213'},
                                 {       'counter_source': '168574/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '168574/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '27193229648',
                                                                       'in_ucast_pkts': '92789518',
                                                                       'mac_address': '84:d3:43:95:92:fa',
                                                                       'name': '168574/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '60',
                                                                       'out_octets': '147982194706',
                                                                       'out_ucast_pkts': '192306438',
                                                                       'poller_hash': 'Q0JEMjdFcGVlcmludGYxNjg1NzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '168574',
                                                                       'in_nucast_pkts': '60',
                                                                       'in_octets': '147982194706',
                                                                       'in_ucast_pkts': '192306438',
                                                                       'links': [       {       'in_nucast_pkts': '60',
                                                                                                'in_octets': '147982194706',
                                                                                                'in_ucast_pkts': '192306438',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '27193229648',
                                                                                                'out_ucast_pkts': '92789518',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '27193229648',
                                                                       'out_ucast_pkts': '92789518',
                                                                       'poller_hash': 'Q0JEMjdFcGVlcmludGZwb24gMTY4NTc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168574',
                                                                                               'links': [       {       'connected_time': '68164046',
                                                                                                                        'fiber_length': 8769,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjdFcGVlcnBvbnBvbiAxNjg1NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4541 LINE 61 (NAFZIGER STEVE)',
                                         'poller_hash': 'Q0JEMjdF',
                                         'serial_number': 'CBD27E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:92:fa',
                                         'uptime': '68164046'},
                                 {       'counter_source': '168941/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d6:5e'],
                                                                       'id': '168941/g1',
                                                                       'in_nucast_pkts': '1003',
                                                                       'in_octets': '224107518321',
                                                                       'in_ucast_pkts': '2146511638',
                                                                       'mac_address': '84:d3:43:95:8b:f5',
                                                                       'mtu': 2000,
                                                                       'name': '168941/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '34480913',
                                                                       'out_octets': '14172092545986',
                                                                       'out_ucast_pkts': '9472786524',
                                                                       'poller_hash': 'Q0JEMTlEcGVlcmludGYxNjg5NDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '168941',
                                                                       'in_nucast_pkts': '34480913',
                                                                       'in_octets': '14172092545986',
                                                                       'in_ucast_pkts': '9472786524',
                                                                       'links': [       {       'in_nucast_pkts': '34480913',
                                                                                                'in_octets': '14172092545986',
                                                                                                'in_ucast_pkts': '9472786524',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1003',
                                                                                                'out_octets': '224107518321',
                                                                                                'out_ucast_pkts': '2146511638',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1003',
                                                                       'out_octets': '224107518321',
                                                                       'out_ucast_pkts': '2146511638',
                                                                       'poller_hash': 'Q0JEMTlEcGVlcmludGZwb24gMTY4OTQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168941',
                                                                                               'links': [       {       'connected_time': '15218523',
                                                                                                                        'fiber_length': 2785,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTlEcGVlcnBvbnBvbiAxNjg5NDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '77 PUGH ST (LUDWIG VICTOR)',
                                         'poller_hash': 'Q0JEMTlE',
                                         'serial_number': 'CBD19D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8b:f2',
                                         'uptime': '15218523'},
                                 {       'counter_source': '168970/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:75'],
                                                                       'id': '168970/g1',
                                                                       'in_nucast_pkts': '13426',
                                                                       'in_octets': '9677610564',
                                                                       'in_ucast_pkts': '35643509',
                                                                       'mac_address': '84:d3:43:96:00:65',
                                                                       'mtu': 2000,
                                                                       'name': '168970/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464282',
                                                                       'out_octets': '126398124933',
                                                                       'out_ucast_pkts': '96983254',
                                                                       'poller_hash': 'Q0JFMDJCcGVlcmludGYxNjg5NzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '168970',
                                                                       'in_nucast_pkts': '1464282',
                                                                       'in_octets': '126398124933',
                                                                       'in_ucast_pkts': '96983254',
                                                                       'links': [       {       'in_nucast_pkts': '1464282',
                                                                                                'in_octets': '126398124933',
                                                                                                'in_ucast_pkts': '96983254',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13426',
                                                                                                'out_octets': '9677610564',
                                                                                                'out_ucast_pkts': '35643509',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13426',
                                                                       'out_octets': '9677610564',
                                                                       'out_ucast_pkts': '35643509',
                                                                       'poller_hash': 'Q0JFMDJCcGVlcmludGZwb24gMTY4OTcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168970',
                                                                                               'links': [       {       'connected_time': '688375',
                                                                                                                        'fiber_length': 2828,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JFMDJCcGVlcnBvbnBvbiAxNjg5NzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '153 PUGH ST (ZURBRIGG CHANTELLE)',
                                         'poller_hash': 'Q0JFMDJC',
                                         'serial_number': 'CBE02B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:96:00:62',
                                         'uptime': '688375'},
                                 {       'counter_source': '170465/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '170465/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '296',
                                                                       'in_octets': '73920918660',
                                                                       'in_ucast_pkts': '346618548',
                                                                       'mac_address': '84:d3:43:96:01:e2',
                                                                       'name': '170465/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '2158',
                                                                       'out_octets': '1166960350580',
                                                                       'out_ucast_pkts': '899535896',
                                                                       'poller_hash': 'Q0JFMDVCcGVlcmludGYxNzA0NjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '170465',
                                                                       'in_nucast_pkts': '2158',
                                                                       'in_octets': '1166960350580',
                                                                       'in_ucast_pkts': '899535896',
                                                                       'links': [       {       'in_nucast_pkts': '2158',
                                                                                                'in_octets': '1166960350580',
                                                                                                'in_ucast_pkts': '899535896',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '296',
                                                                                                'out_octets': '73920918660',
                                                                                                'out_ucast_pkts': '346618548',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '296',
                                                                       'out_octets': '73920918660',
                                                                       'out_ucast_pkts': '346618548',
                                                                       'poller_hash': 'Q0JFMDVCcGVlcmludGZwb24gMTcwNDY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '170465',
                                                                                               'links': [       {       'connected_time': '6067696',
                                                                                                                        'fiber_length': 1004,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JFMDVCcGVlcnBvbnBvbiAxNzA0NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '54 MILL ST E (STOREY GRANT)',
                                         'poller_hash': 'Q0JFMDVC',
                                         'serial_number': 'CBE05B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:96:01:e2',
                                         'uptime': '6067696'},
                                 {       'counter_source': '170742/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:ca:a7'],
                                                                       'id': '170742/g1',
                                                                       'in_nucast_pkts': '6574',
                                                                       'in_octets': '567362730',
                                                                       'in_ucast_pkts': '2455557',
                                                                       'mac_address': '84:d3:43:7e:32:7a',
                                                                       'mtu': 2000,
                                                                       'name': '170742/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '772208',
                                                                       'out_octets': '23442932646',
                                                                       'out_ucast_pkts': '18067540',
                                                                       'poller_hash': 'QzgzQkI4cGVlcmludGYxNzA3NDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '170742',
                                                                       'in_nucast_pkts': '772208',
                                                                       'in_octets': '23442932646',
                                                                       'in_ucast_pkts': '18067540',
                                                                       'links': [       {       'in_nucast_pkts': '772208',
                                                                                                'in_octets': '23442932646',
                                                                                                'in_ucast_pkts': '18067540',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6574',
                                                                                                'out_octets': '567362730',
                                                                                                'out_ucast_pkts': '2455557',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6574',
                                                                       'out_octets': '567362730',
                                                                       'out_ucast_pkts': '2455557',
                                                                       'poller_hash': 'QzgzQkI4cGVlcmludGZwb24gMTcwNzQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '170742',
                                                                                               'links': [       {       'connected_time': '373561',
                                                                                                                        'fiber_length': 2808,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkI4cGVlcnBvbnBvbiAxNzA3NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '81 PUGH ST (HEAD JOSH)',
                                         'poller_hash': 'QzgzQkI4',
                                         'serial_number': 'C83BB8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:77',
                                         'uptime': '373561'},
                                 {       'counter_source': '170949/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:fd:74:1f:86:f6'],
                                                                       'id': '170949/g1',
                                                                       'in_nucast_pkts': '408',
                                                                       'in_octets': '181887662914',
                                                                       'in_ucast_pkts': '1166043982',
                                                                       'mac_address': 'f8:85:f9:17:29:3b',
                                                                       'mtu': 2000,
                                                                       'name': '170949/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '44367200',
                                                                       'out_octets': '4279798656659',
                                                                       'out_ucast_pkts': '3126969233',
                                                                       'poller_hash': 'REZGRTEwcGVlcmludGYxNzA5NDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '170949',
                                                                       'in_nucast_pkts': '44367200',
                                                                       'in_octets': '4279798656659',
                                                                       'in_ucast_pkts': '3126969233',
                                                                       'links': [       {       'in_nucast_pkts': '44367200',
                                                                                                'in_octets': '4279798656659',
                                                                                                'in_ucast_pkts': '3126969233',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '408',
                                                                                                'out_octets': '181887662914',
                                                                                                'out_ucast_pkts': '1166043982',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '408',
                                                                       'out_octets': '181887662914',
                                                                       'out_ucast_pkts': '1166043982',
                                                                       'poller_hash': 'REZGRTEwcGVlcmludGZwb24gMTcwOTQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '170949',
                                                                                               'links': [       {       'connected_time': '19523274',
                                                                                                                        'fiber_length': 10488,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTEwcGVlcnBvbnBvbiAxNzA5NDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5208 LINE 58 (GINGRICH MIKE)',
                                         'poller_hash': 'REZGRTEw',
                                         'serial_number': 'DFFE10',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:29:38',
                                         'uptime': '19523274'},
                                 {       'counter_source': '171847/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '171847/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '26345085674',
                                                                       'in_ucast_pkts': '98843665',
                                                                       'mac_address': 'f8:85:f9:17:28:88',
                                                                       'name': '171847/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '332285039955',
                                                                       'out_ucast_pkts': '308682269',
                                                                       'poller_hash': 'REZGREZBcGVlcmludGYxNzE4NDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '171847',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '332285039955',
                                                                       'in_ucast_pkts': '308682269',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '332285039955',
                                                                                                'in_ucast_pkts': '308682269',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '26345085674',
                                                                                                'out_ucast_pkts': '98843665',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '26345085674',
                                                                       'out_ucast_pkts': '98843665',
                                                                       'poller_hash': 'REZGREZBcGVlcmludGZwb24gMTcxODQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '171847',
                                                                                               'links': [       {       'connected_time': '21815971',
                                                                                                                        'fiber_length': 601,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREZBcGVlcnBvbnBvbiAxNzE4NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 MAIN ST N (PHYSIOTHERAPY ALLIANCE)',
                                         'poller_hash': 'REZGREZB',
                                         'serial_number': 'DFFDFA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:28:88',
                                         'uptime': '21815971'},
                                 {       'counter_source': '171928/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '171928/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '129837250126',
                                                                       'in_ucast_pkts': '294748857',
                                                                       'mac_address': '84:d3:43:7e:32:ff',
                                                                       'name': '171928/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '678522684122',
                                                                       'out_ucast_pkts': '635337906',
                                                                       'poller_hash': 'QzgzQkM5cGVlcmludGYxNzE5MjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '171928',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '678522684122',
                                                                       'in_ucast_pkts': '635337906',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '678522684122',
                                                                                                'in_ucast_pkts': '635337906',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '129837250126',
                                                                                                'out_ucast_pkts': '294748857',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '129837250126',
                                                                       'out_ucast_pkts': '294748857',
                                                                       'poller_hash': 'QzgzQkM5cGVlcmludGZwb24gMTcxOTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '171928',
                                                                                               'links': [       {       'connected_time': '15498459',
                                                                                                                        'fiber_length': 12050,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkM5cGVlcnBvbnBvbiAxNzE5Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6976 MILLBANK MAIN ST (WYE CROSS '
                                                 'PROPERTIES)',
                                         'poller_hash': 'QzgzQkM5',
                                         'serial_number': 'C83BC9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:ff',
                                         'uptime': '15498459'},
                                 {       'counter_source': '171938/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '171938/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '27406028917',
                                                                       'in_ucast_pkts': '114944376',
                                                                       'mac_address': '84:d3:43:7e:32:2f',
                                                                       'name': '171938/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '152',
                                                                       'out_octets': '226121164658',
                                                                       'out_ucast_pkts': '252549702',
                                                                       'poller_hash': 'QzgzQkFGcGVlcmludGYxNzE5MzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '171938',
                                                                       'in_nucast_pkts': '152',
                                                                       'in_octets': '226121164658',
                                                                       'in_ucast_pkts': '252549702',
                                                                       'links': [       {       'in_nucast_pkts': '152',
                                                                                                'in_octets': '226121164658',
                                                                                                'in_ucast_pkts': '252549702',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '27406028917',
                                                                                                'out_ucast_pkts': '114944376',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '27406028917',
                                                                       'out_ucast_pkts': '114944376',
                                                                       'poller_hash': 'QzgzQkFGcGVlcmludGZwb24gMTcxOTM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '171938',
                                                                                               'links': [       {       'connected_time': '3969452',
                                                                                                                        'fiber_length': 11832,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkFGcGVlcnBvbnBvbiAxNzE5Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4082 WATERLOO ST (WEBERS FOAM '
                                                 'INSULATION)',
                                         'poller_hash': 'QzgzQkFG',
                                         'serial_number': 'C83BAF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:2f',
                                         'uptime': '3969452'},
                                 {       'counter_source': '171967/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '171967/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '641341360086',
                                                                       'in_ucast_pkts': '353428695334',
                                                                       'mac_address': 'f8:85:f9:17:26:48',
                                                                       'name': '171967/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '3938940775490',
                                                                       'out_ucast_pkts': '3018963808',
                                                                       'poller_hash': 'REZGREIycGVlcmludGYxNzE5NjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '171967',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '3938940775490',
                                                                       'in_ucast_pkts': '3018963808',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '3938940775490',
                                                                                                'in_ucast_pkts': '3018963808',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '641341360086',
                                                                                                'out_ucast_pkts': '353428695334',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '641341360086',
                                                                       'out_ucast_pkts': '353428695334',
                                                                       'poller_hash': 'REZGREIycGVlcmludGZwb24gMTcxOTY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '171967',
                                                                                               'links': [       {       'connected_time': '688510',
                                                                                                                        'fiber_length': 11687,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREIycGVlcnBvbnBvbiAxNzE5Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6992 MILLBANK MAIN ST (BULMAR WYATT)',
                                         'poller_hash': 'REZGREIy',
                                         'serial_number': 'DFFDB2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:26:48',
                                         'uptime': '688510'},
                                 {       'counter_source': '172052/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:82:01:8f'],
                                                                       'id': '172052/g1',
                                                                       'in_nucast_pkts': '6825870',
                                                                       'in_octets': '366764225064',
                                                                       'in_ucast_pkts': '2967801017',
                                                                       'mac_address': '84:d3:43:7e:35:c2',
                                                                       'mtu': 2000,
                                                                       'name': '172052/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15595125988',
                                                                       'out_octets': '37393486188257',
                                                                       'out_ucast_pkts': '11183911375',
                                                                       'poller_hash': 'QzgzQzIxcGVlcmludGYxNzIwNTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '172052',
                                                                       'in_nucast_pkts': '15595125988',
                                                                       'in_octets': '37393486188257',
                                                                       'in_ucast_pkts': '11183911375',
                                                                       'links': [       {       'in_nucast_pkts': '15595125988',
                                                                                                'in_octets': '37393486188257',
                                                                                                'in_ucast_pkts': '11183911375',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6825870',
                                                                                                'out_octets': '366764225064',
                                                                                                'out_ucast_pkts': '2967801017',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6825870',
                                                                       'out_octets': '366764225064',
                                                                       'out_ucast_pkts': '2967801017',
                                                                       'poller_hash': 'QzgzQzIxcGVlcmludGZwb24gMTcyMDUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172052',
                                                                                               'links': [       {       'connected_time': '21815711',
                                                                                                                        'fiber_length': 18555,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQzIxcGVlcnBvbnBvbiAxNzIwNTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '64 SCHADE ST (YUNDT D DAWN)',
                                         'poller_hash': 'QzgzQzIx',
                                         'serial_number': 'C83C21',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:35:bf',
                                         'uptime': '21815711'},
                                 {       'counter_source': '172239/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '172239/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '392537914599',
                                                                       'in_ucast_pkts': '108641936525',
                                                                       'mac_address': '84:d3:43:7d:e0:57',
                                                                       'name': '172239/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '74',
                                                                       'out_octets': '4948946299963',
                                                                       'out_ucast_pkts': '3725195165',
                                                                       'poller_hash': 'QzgzMTc0cGVlcmludGYxNzIyMzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '172239',
                                                                       'in_nucast_pkts': '74',
                                                                       'in_octets': '4948946299963',
                                                                       'in_ucast_pkts': '3725195165',
                                                                       'links': [       {       'in_nucast_pkts': '74',
                                                                                                'in_octets': '4948946299963',
                                                                                                'in_ucast_pkts': '3725195165',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '392537914599',
                                                                                                'out_ucast_pkts': '108641936525',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '392537914599',
                                                                       'out_ucast_pkts': '108641936525',
                                                                       'poller_hash': 'QzgzMTc0cGVlcmludGZwb24gMTcyMjM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172239',
                                                                                               'links': [       {       'connected_time': '688348',
                                                                                                                        'fiber_length': 11718,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTc0cGVlcnBvbnBvbiAxNzIyMzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6994 MILLBANK MAIN STREET (LICHTI '
                                                 'MATTHEW & SARAH)',
                                         'poller_hash': 'QzgzMTc0',
                                         'serial_number': 'C83174',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e0:57',
                                         'uptime': '688348'},
                                 {       'counter_source': '172269/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '172269/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '354652191154',
                                                                       'in_ucast_pkts': '1328745384',
                                                                       'mac_address': '84:d3:43:7d:df:07',
                                                                       'name': '172269/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '5078768878050',
                                                                       'out_ucast_pkts': '3825240824',
                                                                       'poller_hash': 'QzgzMTRBcGVlcmludGYxNzIyNjkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '172269',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '5078768878050',
                                                                       'in_ucast_pkts': '3825240824',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '5078768878050',
                                                                                                'in_ucast_pkts': '3825240824',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '354652191154',
                                                                                                'out_ucast_pkts': '1328745384',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '354652191154',
                                                                       'out_ucast_pkts': '1328745384',
                                                                       'poller_hash': 'QzgzMTRBcGVlcmludGZwb24gMTcyMjY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172269',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 2773,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTRBcGVlcnBvbnBvbiAxNzIyNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '73 PUGH ST (SZYMIKOWSKI DARIUSZ)',
                                         'poller_hash': 'QzgzMTRB',
                                         'serial_number': 'C8314A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:df:07',
                                         'uptime': '688370'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '172369/g1',
                                                                       'name': '172369/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzQkU2cGVlcmludGYxNzIzNjkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '172369',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgzQkU2cGVlcmludGZwb24gMTcyMzY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172369',
                                                                                               'links': [       {       'connected_time': '21816677',
                                                                                                                        'fiber_length': 1485,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkU2cGVlcnBvbnBvbiAxNzIzNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6476 PERTH ROAD 131 (DYCK ABRAM W)',
                                         'poller_hash': 'QzgzQkU2',
                                         'serial_number': 'C83BE6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:33:e7',
                                         'uptime': '21816677'},
                                 {       'counter_source': '172414/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:64:04'],
                                                                       'id': '172414/g1',
                                                                       'in_nucast_pkts': '910909',
                                                                       'in_octets': '428639497060',
                                                                       'in_ucast_pkts': '2858652966',
                                                                       'mac_address': '84:d3:43:7d:db:5a',
                                                                       'mtu': 2000,
                                                                       'name': '172414/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '123018310',
                                                                       'out_octets': '16645566496384',
                                                                       'out_ucast_pkts': '11609563546',
                                                                       'poller_hash': 'QzgzMEQ0cGVlcmludGYxNzI0MTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '172414',
                                                                       'in_nucast_pkts': '123018310',
                                                                       'in_octets': '16645566496384',
                                                                       'in_ucast_pkts': '11609563546',
                                                                       'links': [       {       'in_nucast_pkts': '123018310',
                                                                                                'in_octets': '16645566496384',
                                                                                                'in_ucast_pkts': '11609563546',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '910909',
                                                                                                'out_octets': '428639497060',
                                                                                                'out_ucast_pkts': '2858652966',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '910909',
                                                                       'out_octets': '428639497060',
                                                                       'out_ucast_pkts': '2858652966',
                                                                       'poller_hash': 'QzgzMEQ0cGVlcmludGZwb24gMTcyNDE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172414',
                                                                                               'links': [       {       'connected_time': '21816015',
                                                                                                                        'fiber_length': 18774,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMEQ0cGVlcnBvbnBvbiAxNzI0MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5339 ROAD 130 (STADELMANN MARK)',
                                         'poller_hash': 'QzgzMEQ0',
                                         'serial_number': 'C830D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:db:57',
                                         'uptime': '21816015'},
                                 {       'counter_source': '172444/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e9:43'],
                                                                       'id': '172444/g1',
                                                                       'in_nucast_pkts': '64908',
                                                                       'in_octets': '43550689351',
                                                                       'in_ucast_pkts': '105680005',
                                                                       'mac_address': '84:d3:43:7d:df:d2',
                                                                       'mtu': 2000,
                                                                       'name': '172444/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6356865',
                                                                       'out_octets': '904841657689',
                                                                       'out_ucast_pkts': '648227499',
                                                                       'poller_hash': 'QzgzMTYzcGVlcmludGYxNzI0NDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '172444',
                                                                       'in_nucast_pkts': '6356865',
                                                                       'in_octets': '904841657689',
                                                                       'in_ucast_pkts': '648227499',
                                                                       'links': [       {       'in_nucast_pkts': '6356865',
                                                                                                'in_octets': '904841657689',
                                                                                                'in_ucast_pkts': '648227499',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '64908',
                                                                                                'out_octets': '43550689351',
                                                                                                'out_ucast_pkts': '105680005',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '64908',
                                                                       'out_octets': '43550689351',
                                                                       'out_ucast_pkts': '105680005',
                                                                       'poller_hash': 'QzgzMTYzcGVlcmludGZwb24gMTcyNDQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172444',
                                                                                               'links': [       {       'connected_time': '2969268',
                                                                                                                        'fiber_length': 17809,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTYzcGVlcnBvbnBvbiAxNzI0NDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '165 MARKET ST (SUBJECT HELENA & '
                                                 'FRED)',
                                         'poller_hash': 'QzgzMTYz',
                                         'serial_number': 'C83163',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:df:cf',
                                         'uptime': '2969268'},
                                 {       'counter_source': '172521/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '172521/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '28257824664',
                                                                       'in_ucast_pkts': '136492831',
                                                                       'mac_address': '84:d3:43:7d:de:e7',
                                                                       'name': '172521/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '582785009363',
                                                                       'out_ucast_pkts': '447523324',
                                                                       'poller_hash': 'QzgzMTQ2cGVlcmludGYxNzI1MjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '172521',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '582785009363',
                                                                       'in_ucast_pkts': '447523324',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '582785009363',
                                                                                                'in_ucast_pkts': '447523324',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '28257824664',
                                                                                                'out_ucast_pkts': '136492831',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '28257824664',
                                                                       'out_ucast_pkts': '136492831',
                                                                       'poller_hash': 'QzgzMTQ2cGVlcmludGZwb24gMTcyNTIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172521',
                                                                                               'links': [       {       'connected_time': '596470',
                                                                                                                        'fiber_length': 10555,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTQ2cGVlcnBvbnBvbiAxNzI1MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5921 SOMERSET DR (HART BEN)',
                                         'poller_hash': 'QzgzMTQ2',
                                         'serial_number': 'C83146',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:de:e7',
                                         'uptime': '596470'},
                                 {       'counter_source': '172566/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e8:25'],
                                                                       'id': '172566/g1',
                                                                       'in_nucast_pkts': '1837',
                                                                       'in_octets': '7063643270',
                                                                       'in_ucast_pkts': '37803317',
                                                                       'mac_address': '84:d3:43:7d:e4:0a',
                                                                       'mtu': 2000,
                                                                       'name': '172566/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '380390',
                                                                       'out_octets': '360137241448',
                                                                       'out_ucast_pkts': '250424212',
                                                                       'poller_hash': 'QzgzMUVBcGVlcmludGYxNzI1NjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '172566',
                                                                       'in_nucast_pkts': '380390',
                                                                       'in_octets': '360137241448',
                                                                       'in_ucast_pkts': '250424212',
                                                                       'links': [       {       'in_nucast_pkts': '380390',
                                                                                                'in_octets': '360137241448',
                                                                                                'in_ucast_pkts': '250424212',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1837',
                                                                                                'out_octets': '7063643270',
                                                                                                'out_ucast_pkts': '37803317',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1837',
                                                                       'out_octets': '7063643270',
                                                                       'out_ucast_pkts': '37803317',
                                                                       'poller_hash': 'QzgzMUVBcGVlcmludGZwb24gMTcyNTY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172566',
                                                                                               'links': [       {       'connected_time': '688357',
                                                                                                                        'fiber_length': 1279,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMUVBcGVlcnBvbnBvbiAxNzI1NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '31 WEST ST (TOPHAM-LINTHOME TYLER)',
                                         'poller_hash': 'QzgzMUVB',
                                         'serial_number': 'C831EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e4:07',
                                         'uptime': '688357'},
                                 {       'counter_source': '172599/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '172599/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '767296',
                                                                       'in_octets': '118391865171',
                                                                       'in_ucast_pkts': '872760014',
                                                                       'mac_address': 'd0:76:8f:07:18:dd',
                                                                       'name': '172599/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '4815944432',
                                                                       'out_octets': '10601189274562',
                                                                       'out_ucast_pkts': '2865522368',
                                                                       'poller_hash': 'ODY3OEEycGVlcmludGYxNzI1OTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '172599',
                                                                       'in_nucast_pkts': '4815944432',
                                                                       'in_octets': '10601189274562',
                                                                       'in_ucast_pkts': '2865522368',
                                                                       'links': [       {       'in_nucast_pkts': '4815944432',
                                                                                                'in_octets': '10601189274562',
                                                                                                'in_ucast_pkts': '2865522368',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '767296',
                                                                                                'out_octets': '118391865171',
                                                                                                'out_ucast_pkts': '872760014',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '767296',
                                                                       'out_octets': '118391865171',
                                                                       'out_ucast_pkts': '872760014',
                                                                       'poller_hash': 'ODY3OEEycGVlcmludGZwb24gMTcyNTk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172599',
                                                                                               'links': [       {       'connected_time': '688356',
                                                                                                                        'fiber_length': 1115,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY3OEEycGVlcnBvbnBvbiAxNzI1OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 MAIN ST S (KUEPFER RACHEL)',
                                         'poller_hash': 'ODY3OEEy',
                                         'serial_number': '8678A2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:07:18:dd',
                                         'uptime': '688356'},
                                 {       'counter_source': '172610/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ec:2f'],
                                                                       'id': '172610/g1',
                                                                       'in_nucast_pkts': '981641',
                                                                       'in_octets': '2258603719014',
                                                                       'in_ucast_pkts': '5963475083',
                                                                       'mac_address': 'f8:85:f9:17:22:7b',
                                                                       'mtu': 2000,
                                                                       'name': '172610/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149352519',
                                                                       'out_octets': '13585090105034',
                                                                       'out_ucast_pkts': '11883156656',
                                                                       'poller_hash': 'REZGRDM4cGVlcmludGYxNzI2MTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '172610',
                                                                       'in_nucast_pkts': '149352519',
                                                                       'in_octets': '13585090105034',
                                                                       'in_ucast_pkts': '11883156656',
                                                                       'links': [       {       'in_nucast_pkts': '149352519',
                                                                                                'in_octets': '13585090105034',
                                                                                                'in_ucast_pkts': '11883156656',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '981641',
                                                                                                'out_octets': '2258603719014',
                                                                                                'out_ucast_pkts': '5963475083',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '981641',
                                                                       'out_octets': '2258603719014',
                                                                       'out_ucast_pkts': '5963475083',
                                                                       'poller_hash': 'REZGRDM4cGVlcmludGZwb24gMTcyNjEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172610',
                                                                                               'links': [       {       'connected_time': '21816175',
                                                                                                                        'fiber_length': 11627,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRDM4cGVlcnBvbnBvbiAxNzI2MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4066 WATERLOO ST (BARRY ALEXANDER)',
                                         'poller_hash': 'REZGRDM4',
                                         'serial_number': 'DFFD38',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:22:78',
                                         'uptime': '21816175'},
                                 {       'counter_source': '174370/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:eb:b6:58:a7:d5'],
                                                                       'id': '174370/g1',
                                                                       'in_nucast_pkts': '37630',
                                                                       'in_octets': '752670577345',
                                                                       'in_ucast_pkts': '3083592002',
                                                                       'mac_address': 'f8:85:f9:6f:22:f1',
                                                                       'mtu': 2000,
                                                                       'name': '174370/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '80818399',
                                                                       'out_octets': '18310522966100',
                                                                       'out_ucast_pkts': '13089566459',
                                                                       'poller_hash': 'QkYyNTBFcGVlcmludGYxNzQzNzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174370',
                                                                       'in_nucast_pkts': '80818399',
                                                                       'in_octets': '18310522966100',
                                                                       'in_ucast_pkts': '13089566459',
                                                                       'links': [       {       'in_nucast_pkts': '80818399',
                                                                                                'in_octets': '18310522966100',
                                                                                                'in_ucast_pkts': '13089566459',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '37630',
                                                                                                'out_octets': '752670577345',
                                                                                                'out_ucast_pkts': '3083592002',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '37630',
                                                                       'out_octets': '752670577345',
                                                                       'out_ucast_pkts': '3083592002',
                                                                       'poller_hash': 'QkYyNTBFcGVlcmludGZwb24gMTc0Mzcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174370',
                                                                                               'links': [       {       'connected_time': '40565',
                                                                                                                        'fiber_length': 984,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QkYyNTBFcGVlcnBvbnBvbiAxNzQzNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '25 PACIFIC AV (DOBES SEBASTIAN)',
                                         'poller_hash': 'QkYyNTBF',
                                         'serial_number': 'BF250E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'f8:85:f9:6f:22:ed',
                                         'uptime': '40565'},
                                 {       'counter_source': '174484/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:bf'],
                                                                       'id': '174484/g1',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '12402796948',
                                                                       'in_ucast_pkts': '17511027',
                                                                       'mac_address': '84:d3:43:7e:15:c2',
                                                                       'mtu': 2000,
                                                                       'name': '174484/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464356',
                                                                       'out_octets': '61820249439',
                                                                       'out_ucast_pkts': '50530195',
                                                                       'poller_hash': 'QzgzODIxcGVlcmludGYxNzQ0ODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174484',
                                                                       'in_nucast_pkts': '1464356',
                                                                       'in_octets': '61820249439',
                                                                       'in_ucast_pkts': '50530195',
                                                                       'links': [       {       'in_nucast_pkts': '1464356',
                                                                                                'in_octets': '61820249439',
                                                                                                'in_ucast_pkts': '50530195',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19',
                                                                                                'out_octets': '12402796948',
                                                                                                'out_ucast_pkts': '17511027',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19',
                                                                       'out_octets': '12402796948',
                                                                       'out_ucast_pkts': '17511027',
                                                                       'poller_hash': 'QzgzODIxcGVlcmludGZwb24gMTc0NDg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174484',
                                                                                               'links': [       {       'connected_time': '688499',
                                                                                                                        'fiber_length': 633,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzODIxcGVlcnBvbnBvbiAxNzQ0ODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 MAIN ST N (KAUR LOVEPREET)',
                                         'poller_hash': 'QzgzODIx',
                                         'serial_number': 'C83821',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:15:bf',
                                         'uptime': '688499'},
                                 {       'counter_source': '174999/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7d:62'],
                                                                       'id': '174999/g1',
                                                                       'in_nucast_pkts': '53872',
                                                                       'in_octets': '67204106042',
                                                                       'in_ucast_pkts': '103342059',
                                                                       'mac_address': '84:d3:43:7e:33:32',
                                                                       'mtu': 2000,
                                                                       'name': '174999/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6720887',
                                                                       'out_octets': '284928575721',
                                                                       'out_ucast_pkts': '262411610',
                                                                       'poller_hash': 'QzgzQkNGcGVlcmludGYxNzQ5OTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174999',
                                                                       'in_nucast_pkts': '6720887',
                                                                       'in_octets': '284928575721',
                                                                       'in_ucast_pkts': '262411610',
                                                                       'links': [       {       'in_nucast_pkts': '6720887',
                                                                                                'in_octets': '284928575721',
                                                                                                'in_ucast_pkts': '262411610',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '53872',
                                                                                                'out_octets': '67204106042',
                                                                                                'out_ucast_pkts': '103342059',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53872',
                                                                       'out_octets': '67204106042',
                                                                       'out_ucast_pkts': '103342059',
                                                                       'poller_hash': 'QzgzQkNGcGVlcmludGZwb24gMTc0OTk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174999',
                                                                                               'links': [       {       'connected_time': '3113979',
                                                                                                                        'fiber_length': 17525,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkNGcGVlcnBvbnBvbiAxNzQ5OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '188 ERSKINE ST (RAMIREZ CARLOS)',
                                         'poller_hash': 'QzgzQkNG',
                                         'serial_number': 'C83BCF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:33:2f',
                                         'uptime': '3113979'},
                                 {       'counter_source': '175146/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '175146/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '88180927156',
                                                                       'in_ucast_pkts': '486466288',
                                                                       'mac_address': '60:db:98:72:f6:af',
                                                                       'name': '175146/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '1678403168165',
                                                                       'out_ucast_pkts': '1259393553',
                                                                       'poller_hash': 'QjA4RUY2cGVlcmludGYxNzUxNDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '175146',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '1678403168165',
                                                                       'in_ucast_pkts': '1259393553',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '1678403168165',
                                                                                                'in_ucast_pkts': '1259393553',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '88180927156',
                                                                                                'out_ucast_pkts': '486466288',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '88180927156',
                                                                       'out_ucast_pkts': '486466288',
                                                                       'poller_hash': 'QjA4RUY2cGVlcmludGZwb24gMTc1MTQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175146',
                                                                                               'links': [       {       'connected_time': '687625',
                                                                                                                        'fiber_length': 11694,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjA4RUY2cGVlcnBvbnBvbiAxNzUxNDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6988 MILLBANK MAIN ST (WEBER KACIE)',
                                         'poller_hash': 'QjA4RUY2',
                                         'serial_number': 'B08EF6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:72:f6:af',
                                         'uptime': '687625'},
                                 {       'counter_source': '175214/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '175214/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '516305446523',
                                                                       'in_ucast_pkts': '79724217018',
                                                                       'mac_address': '84:d3:43:7e:34:1f',
                                                                       'name': '175214/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '86',
                                                                       'out_octets': '8098216499401',
                                                                       'out_ucast_pkts': '5704130253',
                                                                       'poller_hash': 'QzgzQkVEcGVlcmludGYxNzUyMTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '175214',
                                                                       'in_nucast_pkts': '86',
                                                                       'in_octets': '8098216499401',
                                                                       'in_ucast_pkts': '5704130253',
                                                                       'links': [       {       'in_nucast_pkts': '86',
                                                                                                'in_octets': '8098216499401',
                                                                                                'in_ucast_pkts': '5704130253',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '516305446523',
                                                                                                'out_ucast_pkts': '79724217018',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '516305446523',
                                                                       'out_ucast_pkts': '79724217018',
                                                                       'poller_hash': 'QzgzQkVEcGVlcmludGZwb24gMTc1MjE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175214',
                                                                                               'links': [       {       'connected_time': '688362',
                                                                                                                        'fiber_length': 2801,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkVEcGVlcnBvbnBvbiAxNzUyMTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '85 PUGH ST (CARTER ASHLEY)',
                                         'poller_hash': 'QzgzQkVE',
                                         'serial_number': 'C83BED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:34:1f',
                                         'uptime': '688362'},
                                 {       'counter_source': '175224/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cf:e5'],
                                                                       'id': '175224/g1',
                                                                       'in_nucast_pkts': '725832',
                                                                       'in_octets': '233587877947',
                                                                       'in_ucast_pkts': '2366506149',
                                                                       'mac_address': '84:d3:43:95:8e:a5',
                                                                       'mtu': 2000,
                                                                       'name': '175224/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '99038472',
                                                                       'out_octets': '14227848819899',
                                                                       'out_ucast_pkts': '9535601907',
                                                                       'poller_hash': 'Q0JEMUYzcGVlcmludGYxNzUyMjQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '175224',
                                                                       'in_nucast_pkts': '99038472',
                                                                       'in_octets': '14227848819899',
                                                                       'in_ucast_pkts': '9535601907',
                                                                       'links': [       {       'in_nucast_pkts': '99038472',
                                                                                                'in_octets': '14227848819899',
                                                                                                'in_ucast_pkts': '9535601907',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '725832',
                                                                                                'out_octets': '233587877947',
                                                                                                'out_ucast_pkts': '2366506149',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '725832',
                                                                       'out_octets': '233587877947',
                                                                       'out_ucast_pkts': '2366506149',
                                                                       'poller_hash': 'Q0JEMUYzcGVlcmludGZwb24gMTc1MjI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175224',
                                                                                               'links': [       {       'connected_time': '73290287',
                                                                                                                        'fiber_length': 11721,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUYzcGVlcnBvbnBvbiAxNzUyMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6996 FREEBORN ST (BIGAM RONALD)',
                                         'poller_hash': 'Q0JEMUYz',
                                         'serial_number': 'CBD1F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8e:a2',
                                         'uptime': '73290287'},
                                 {       'counter_source': '175234/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '175234/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '90763488689',
                                                                       'in_ucast_pkts': '854186055',
                                                                       'mac_address': '84:d3:43:95:8b:aa',
                                                                       'name': '175234/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '3482944353773',
                                                                       'out_ucast_pkts': '2404240504',
                                                                       'poller_hash': 'Q0JEMTk0cGVlcmludGYxNzUyMzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '175234',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '3482944353773',
                                                                       'in_ucast_pkts': '2404240504',
                                                                       'links': [       {       'in_nucast_pkts': '141',
                                                                                                'in_octets': '3482944353773',
                                                                                                'in_ucast_pkts': '2404240504',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '90763488689',
                                                                                                'out_ucast_pkts': '854186055',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '90763488689',
                                                                       'out_ucast_pkts': '854186055',
                                                                       'poller_hash': 'Q0JEMTk0cGVlcmludGZwb24gMTc1MjM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175234',
                                                                                               'links': [       {       'connected_time': '523248',
                                                                                                                        'fiber_length': 11751,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTk0cGVlcnBvbnBvbiAxNzUyMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6984 FREEBORN ST (CUNNINGHAM ELAINE)',
                                         'poller_hash': 'Q0JEMTk0',
                                         'serial_number': 'CBD194',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8b:aa',
                                         'uptime': '523248'},
                                 {       'counter_source': '175244/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '175244/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '71725232801',
                                                                       'in_ucast_pkts': '187895963',
                                                                       'mac_address': '84:d3:43:95:8b:ea',
                                                                       'name': '175244/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '207',
                                                                       'out_octets': '567528691540',
                                                                       'out_ucast_pkts': '511807214',
                                                                       'poller_hash': 'Q0JEMTlDcGVlcmludGYxNzUyNDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '175244',
                                                                       'in_nucast_pkts': '207',
                                                                       'in_octets': '567528691540',
                                                                       'in_ucast_pkts': '511807214',
                                                                       'links': [       {       'in_nucast_pkts': '207',
                                                                                                'in_octets': '567528691540',
                                                                                                'in_ucast_pkts': '511807214',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '71725232801',
                                                                                                'out_ucast_pkts': '187895963',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '71725232801',
                                                                       'out_ucast_pkts': '187895963',
                                                                       'poller_hash': 'Q0JEMTlDcGVlcmludGZwb24gMTc1MjQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175244',
                                                                                               'links': [       {       'connected_time': '21815841',
                                                                                                                        'fiber_length': 11689,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTlDcGVlcnBvbnBvbiAxNzUyNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6994 FREEBORN ST (ERB LLOYD & '
                                                 'LAURENE)',
                                         'poller_hash': 'Q0JEMTlD',
                                         'serial_number': 'CBD19C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8b:ea',
                                         'uptime': '21815841'},
                                 {       'counter_source': '175361/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:39:b5'],
                                                                       'id': '175361/g1',
                                                                       'in_nucast_pkts': '452292',
                                                                       'in_octets': '183323476460',
                                                                       'in_ucast_pkts': '464803530',
                                                                       'mac_address': '84:d3:43:95:8c:ed',
                                                                       'mtu': 2000,
                                                                       'name': '175361/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '44806989',
                                                                       'out_octets': '1220034012146',
                                                                       'out_ucast_pkts': '1068628776',
                                                                       'poller_hash': 'Q0JEMUJDcGVlcmludGYxNzUzNjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '175361',
                                                                       'in_nucast_pkts': '44806989',
                                                                       'in_octets': '1220034012146',
                                                                       'in_ucast_pkts': '1068628776',
                                                                       'links': [       {       'in_nucast_pkts': '44806989',
                                                                                                'in_octets': '1220034012146',
                                                                                                'in_ucast_pkts': '1068628776',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '452292',
                                                                                                'out_octets': '183323476460',
                                                                                                'out_ucast_pkts': '464803530',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '452292',
                                                                       'out_octets': '183323476460',
                                                                       'out_ucast_pkts': '464803530',
                                                                       'poller_hash': 'Q0JEMUJDcGVlcmludGZwb24gMTc1MzYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175361',
                                                                                               'links': [       {       'connected_time': '69480770',
                                                                                                                        'fiber_length': 11800,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUJDcGVlcnBvbnBvbiAxNzUzNjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6980 MILLBANK MAIN ST (ELIZABETH '
                                                 'WATSON FLORAL)',
                                         'poller_hash': 'Q0JEMUJD',
                                         'serial_number': 'CBD1BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8c:ea',
                                         'uptime': '69480770'},
                                 {       'counter_source': '175381/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '175381/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8241817367',
                                                                       'in_ucast_pkts': '33839004',
                                                                       'mac_address': '84:d3:43:7e:2d:a7',
                                                                       'name': '175381/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '192',
                                                                       'out_octets': '144424049476',
                                                                       'out_ucast_pkts': '165000210',
                                                                       'poller_hash': 'QzgzQjFFcGVlcmludGYxNzUzODEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '175381',
                                                                       'in_nucast_pkts': '192',
                                                                       'in_octets': '144424049476',
                                                                       'in_ucast_pkts': '165000210',
                                                                       'links': [       {       'in_nucast_pkts': '192',
                                                                                                'in_octets': '144424049476',
                                                                                                'in_ucast_pkts': '165000210',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '8241817367',
                                                                                                'out_ucast_pkts': '33839004',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8241817367',
                                                                       'out_ucast_pkts': '33839004',
                                                                       'poller_hash': 'QzgzQjFFcGVlcmludGZwb24gMTc1Mzgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175381',
                                                                                               'links': [       {       'connected_time': '960570',
                                                                                                                        'fiber_length': 11862,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjFFcGVlcnBvbnBvbiAxNzUzODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6985 FREEBORN ST (GERBER BETTY)',
                                         'poller_hash': 'QzgzQjFF',
                                         'serial_number': 'C83B1E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:2d:a7',
                                         'uptime': '960570'},
                                 {       'counter_source': '175856/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3c:f9'],
                                                                       'id': '175856/g1',
                                                                       'in_nucast_pkts': '11966',
                                                                       'in_octets': '8257788770',
                                                                       'in_ucast_pkts': '47271769',
                                                                       'mac_address': '84:d3:43:7d:d8:aa',
                                                                       'mtu': 2000,
                                                                       'name': '175856/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464468',
                                                                       'out_octets': '297421827931',
                                                                       'out_ucast_pkts': '215697682',
                                                                       'poller_hash': 'QzgzMDdFcGVlcmludGYxNzU4NTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '175856',
                                                                       'in_nucast_pkts': '1464468',
                                                                       'in_octets': '297421827931',
                                                                       'in_ucast_pkts': '215697682',
                                                                       'links': [       {       'in_nucast_pkts': '1464468',
                                                                                                'in_octets': '297421827931',
                                                                                                'in_ucast_pkts': '215697682',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11966',
                                                                                                'out_octets': '8257788770',
                                                                                                'out_ucast_pkts': '47271769',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11966',
                                                                       'out_octets': '8257788770',
                                                                       'out_ucast_pkts': '47271769',
                                                                       'poller_hash': 'QzgzMDdFcGVlcmludGZwb24gMTc1ODU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175856',
                                                                                               'links': [       {       'connected_time': '688389',
                                                                                                                        'fiber_length': 2721,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMDdFcGVlcnBvbnBvbiAxNzU4NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '125 PUGH ST (KIRCH STEVE)',
                                         'poller_hash': 'QzgzMDdF',
                                         'serial_number': 'C8307E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:d8:a7',
                                         'uptime': '688389'},
                                 {       'counter_source': '177005/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:52:96'],
                                                                       'id': '177005/g1',
                                                                       'in_nucast_pkts': '335319',
                                                                       'in_octets': '366072960888',
                                                                       'in_ucast_pkts': '3165777892',
                                                                       'mac_address': '84:d3:43:7d:e4:52',
                                                                       'mtu': 2000,
                                                                       'name': '177005/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '51229730',
                                                                       'out_octets': '15520480445009',
                                                                       'out_ucast_pkts': '11013737920',
                                                                       'poller_hash': 'QzgzMUYzcGVlcmludGYxNzcwMDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '177005',
                                                                       'in_nucast_pkts': '51229730',
                                                                       'in_octets': '15520480445009',
                                                                       'in_ucast_pkts': '11013737920',
                                                                       'links': [       {       'in_nucast_pkts': '51229730',
                                                                                                'in_octets': '15520480445009',
                                                                                                'in_ucast_pkts': '11013737920',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '335319',
                                                                                                'out_octets': '366072960888',
                                                                                                'out_ucast_pkts': '3165777892',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '335319',
                                                                       'out_octets': '366072960888',
                                                                       'out_ucast_pkts': '3165777892',
                                                                       'poller_hash': 'QzgzMUYzcGVlcmludGZwb24gMTc3MDA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '177005',
                                                                                               'links': [       {       'connected_time': '2078104',
                                                                                                                        'fiber_length': 11482,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMUYzcGVlcnBvbnBvbiAxNzcwMDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6995 MILLBANK MAIN ST (KROPF BRUCE)',
                                         'poller_hash': 'QzgzMUYz',
                                         'serial_number': 'C831F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e4:4f',
                                         'uptime': '2078104'},
                                 {       'counter_source': '177025/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '34:0a:33:89:92:79'],
                                                                       'id': '177025/g1',
                                                                       'in_nucast_pkts': '4129',
                                                                       'in_octets': '57441885',
                                                                       'in_ucast_pkts': '176802',
                                                                       'mac_address': '84:d3:43:7e:05:3a',
                                                                       'mtu': 2000,
                                                                       'name': '177025/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465893',
                                                                       'out_octets': '353484219',
                                                                       'out_ucast_pkts': '346087',
                                                                       'poller_hash': 'QzgzNjEwcGVlcmludGYxNzcwMjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '177025',
                                                                       'in_nucast_pkts': '1465893',
                                                                       'in_octets': '353484219',
                                                                       'in_ucast_pkts': '346087',
                                                                       'links': [       {       'in_nucast_pkts': '1465893',
                                                                                                'in_octets': '353484219',
                                                                                                'in_ucast_pkts': '346087',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4129',
                                                                                                'out_octets': '57441885',
                                                                                                'out_ucast_pkts': '176802',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4129',
                                                                       'out_octets': '57441885',
                                                                       'out_ucast_pkts': '176802',
                                                                       'poller_hash': 'QzgzNjEwcGVlcmludGZwb24gMTc3MDI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '177025',
                                                                                               'links': [       {       'connected_time': '688348',
                                                                                                                        'fiber_length': 11846,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzNjEwcGVlcnBvbnBvbiAxNzcwMjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4082B WATERLOO ST (ORANGE BARN)',
                                         'poller_hash': 'QzgzNjEw',
                                         'serial_number': 'C83610',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:05:37',
                                         'uptime': '688348'},
                                 {       'counter_source': '177266/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:37:a5'],
                                                                       'id': '177266/g1',
                                                                       'in_nucast_pkts': '66835',
                                                                       'in_octets': '61977279961',
                                                                       'in_ucast_pkts': '321198641',
                                                                       'mac_address': 'ec:4f:82:4f:7a:5a',
                                                                       'mtu': 2000,
                                                                       'name': '177266/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6781920',
                                                                       'out_octets': '1536095483149',
                                                                       'out_ucast_pkts': '1168103753',
                                                                       'poller_hash': 'Mjk4MjREcGVlcmludGYxNzcyNjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '177266',
                                                                       'in_nucast_pkts': '6781920',
                                                                       'in_octets': '1536095483149',
                                                                       'in_ucast_pkts': '1168103753',
                                                                       'links': [       {       'in_nucast_pkts': '6781920',
                                                                                                'in_octets': '1536095483149',
                                                                                                'in_ucast_pkts': '1168103753',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '66835',
                                                                                                'out_octets': '61977279961',
                                                                                                'out_ucast_pkts': '321198641',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66835',
                                                                       'out_octets': '61977279961',
                                                                       'out_ucast_pkts': '321198641',
                                                                       'poller_hash': 'Mjk4MjREcGVlcmludGZwb24gMTc3MjY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '177266',
                                                                                               'links': [       {       'connected_time': '3113836',
                                                                                                                        'fiber_length': 17559,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjREcGVlcnBvbnBvbiAxNzcyNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '162 FISHLEIGH ST (MATTHEWS GRANT)',
                                         'poller_hash': 'Mjk4MjRE',
                                         'serial_number': '29824D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4f:7a:57',
                                         'uptime': '3113836'},
                                 {       'counter_source': '180167/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '180167/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '82759220394',
                                                                       'in_ucast_pkts': '401896319',
                                                                       'mac_address': '84:d3:43:7e:34:5f',
                                                                       'name': '180167/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '168',
                                                                       'out_octets': '2541203125015',
                                                                       'out_ucast_pkts': '2062066654',
                                                                       'poller_hash': 'QzgzQkY1cGVlcmludGYxODAxNjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180167',
                                                                       'in_nucast_pkts': '168',
                                                                       'in_octets': '2541203125015',
                                                                       'in_ucast_pkts': '2062066654',
                                                                       'links': [       {       'in_nucast_pkts': '168',
                                                                                                'in_octets': '2541203125015',
                                                                                                'in_ucast_pkts': '2062066654',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '82759220394',
                                                                                                'out_ucast_pkts': '401896319',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '82759220394',
                                                                       'out_ucast_pkts': '401896319',
                                                                       'poller_hash': 'QzgzQkY1cGVlcmludGZwb24gMTgwMTY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180167',
                                                                                               'links': [       {       'connected_time': '3113802',
                                                                                                                        'fiber_length': 17705,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkY1cGVlcnBvbnBvbiAxODAxNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '220 SELBY AV (GREENWALD ROBERT)',
                                         'poller_hash': 'QzgzQkY1',
                                         'serial_number': 'C83BF5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:34:5f',
                                         'uptime': '3113802'},
                                 {       'counter_source': '180177/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '180177/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '150973663423',
                                                                       'in_ucast_pkts': '598898545',
                                                                       'mac_address': '84:d3:43:7e:35:67',
                                                                       'name': '180177/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '75',
                                                                       'out_octets': '4513205606103',
                                                                       'out_ucast_pkts': '3404869554',
                                                                       'poller_hash': 'QzgzQzE2cGVlcmludGYxODAxNzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180177',
                                                                       'in_nucast_pkts': '75',
                                                                       'in_octets': '4513205606103',
                                                                       'in_ucast_pkts': '3404869554',
                                                                       'links': [       {       'in_nucast_pkts': '75',
                                                                                                'in_octets': '4513205606103',
                                                                                                'in_ucast_pkts': '3404869554',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '150973663423',
                                                                                                'out_ucast_pkts': '598898545',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '150973663423',
                                                                       'out_ucast_pkts': '598898545',
                                                                       'poller_hash': 'QzgzQzE2cGVlcmludGZwb24gMTgwMTc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180177',
                                                                                               'links': [       {       'connected_time': '688516',
                                                                                                                        'fiber_length': 11782,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQzE2cGVlcnBvbnBvbiAxODAxNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4038 PERTH LINE 72 (KUEPFER JOE)',
                                         'poller_hash': 'QzgzQzE2',
                                         'serial_number': 'C83C16',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:35:67',
                                         'uptime': '688516'},
                                 {       'counter_source': '180262/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ee:d8'],
                                                                       'id': '180262/g1',
                                                                       'in_nucast_pkts': '14009',
                                                                       'in_octets': '1481818214',
                                                                       'in_ucast_pkts': '10352205',
                                                                       'mac_address': '84:d3:43:7e:1e:2a',
                                                                       'mtu': 2000,
                                                                       'name': '180262/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464251',
                                                                       'out_octets': '32250818324',
                                                                       'out_ucast_pkts': '24708180',
                                                                       'poller_hash': 'QzgzOTJFcGVlcmludGYxODAyNjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '180262',
                                                                       'in_nucast_pkts': '1464251',
                                                                       'in_octets': '32250818324',
                                                                       'in_ucast_pkts': '24708180',
                                                                       'links': [       {       'in_nucast_pkts': '1464251',
                                                                                                'in_octets': '32250818324',
                                                                                                'in_ucast_pkts': '24708180',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14009',
                                                                                                'out_octets': '1481818214',
                                                                                                'out_ucast_pkts': '10352205',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14009',
                                                                       'out_octets': '1481818214',
                                                                       'out_ucast_pkts': '10352205',
                                                                       'poller_hash': 'QzgzOTJFcGVlcmludGZwb24gMTgwMjYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180262',
                                                                                               'links': [       {       'connected_time': '688489',
                                                                                                                        'fiber_length': 682,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzOTJFcGVlcnBvbnBvbiAxODAyNjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 MILL ST E APT 1 (MANE APPEAL)',
                                         'poller_hash': 'QzgzOTJF',
                                         'serial_number': 'C8392E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:1e:27',
                                         'uptime': '688489'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '180272/g1',
                                                                       'name': '180272/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzQkM1cGVlcmludGYxODAyNzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180272',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzQkM1cGVlcmludGZwb24gMTgwMjcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180272',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkM1cGVlcnBvbnBvbiAxODAyNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '162 FISHLEIGH ST',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "180272" '
                                                                  '(location: GPON '
                                                                  'port "11/3"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-06-16T18:42:37.610Z',
                                         'poller_hash': 'QzgzQkM1',
                                         'serial_number': 'C83BC5',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '180282/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '180282/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '64636598501',
                                                                       'in_ucast_pkts': '248477589',
                                                                       'mac_address': '84:d3:43:7e:09:3f',
                                                                       'name': '180282/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '536529126999',
                                                                       'out_ucast_pkts': '551210987',
                                                                       'poller_hash': 'QzgzNjkxcGVlcmludGYxODAyODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180282',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '536529126999',
                                                                       'in_ucast_pkts': '551210987',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '536529126999',
                                                                                                'in_ucast_pkts': '551210987',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '64636598501',
                                                                                                'out_ucast_pkts': '248477589',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '64636598501',
                                                                       'out_ucast_pkts': '248477589',
                                                                       'poller_hash': 'QzgzNjkxcGVlcmludGZwb24gMTgwMjgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180282',
                                                                                               'links': [       {       'connected_time': '27666955',
                                                                                                                        'fiber_length': 11784,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzNjkxcGVlcnBvbnBvbiAxODAyODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6985 PARK ST (ORANGE BARN)',
                                         'poller_hash': 'QzgzNjkx',
                                         'serial_number': 'C83691',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:09:3f',
                                         'uptime': '27666955'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '180490/g1',
                                                                       'name': '180490/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'REZGQzMycGVlcmludGYxODA0OTAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180490',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'REZGQzMycGVlcmludGZwb24gMTgwNDkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180490',
                                                                                               'links': [       {       'connected_time': '21816134',
                                                                                                                        'fiber_length': 7260,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGQzMycGVlcnBvbnBvbiAxODA0OTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4851 LINE 67 (KUEPFER BRIAN)',
                                         'poller_hash': 'REZGQzMy',
                                         'serial_number': 'DFFC32',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:1a:48',
                                         'uptime': '21816134'},
                                 {       'counter_source': '180530/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '180530/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '529539636892',
                                                                       'in_ucast_pkts': '74792150338',
                                                                       'mac_address': '84:d3:43:7e:34:87',
                                                                       'name': '180530/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '200',
                                                                       'out_octets': '6956372559273',
                                                                       'out_ucast_pkts': '5321474297',
                                                                       'poller_hash': 'QzgzQkZBcGVlcmludGYxODA1MzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180530',
                                                                       'in_nucast_pkts': '200',
                                                                       'in_octets': '6956372559273',
                                                                       'in_ucast_pkts': '5321474297',
                                                                       'links': [       {       'in_nucast_pkts': '200',
                                                                                                'in_octets': '6956372559273',
                                                                                                'in_ucast_pkts': '5321474297',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '529539636892',
                                                                                                'out_ucast_pkts': '74792150338',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '529539636892',
                                                                       'out_ucast_pkts': '74792150338',
                                                                       'poller_hash': 'QzgzQkZBcGVlcmludGZwb24gMTgwNTMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180530',
                                                                                               'links': [       {       'connected_time': '21816278',
                                                                                                                        'fiber_length': 11730,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkZBcGVlcnBvbnBvbiAxODA1MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6989 FREEBORN ST (JANTZI JOHN & '
                                                 'MALLORY)',
                                         'poller_hash': 'QzgzQkZB',
                                         'serial_number': 'C83BFA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:34:87',
                                         'uptime': '21816278'},
                                 {       'counter_source': '180764/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '180764/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1063976737495',
                                                                       'in_ucast_pkts': '2223679454',
                                                                       'mac_address': 'f8:85:f9:17:27:f8',
                                                                       'name': '180764/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '70',
                                                                       'out_octets': '7054263226137',
                                                                       'out_ucast_pkts': '5189420177',
                                                                       'poller_hash': 'REZGREU4cGVlcmludGYxODA3NjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180764',
                                                                       'in_nucast_pkts': '70',
                                                                       'in_octets': '7054263226137',
                                                                       'in_ucast_pkts': '5189420177',
                                                                       'links': [       {       'in_nucast_pkts': '70',
                                                                                                'in_octets': '7054263226137',
                                                                                                'in_ucast_pkts': '5189420177',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1063976737495',
                                                                                                'out_ucast_pkts': '2223679454',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1063976737495',
                                                                       'out_ucast_pkts': '2223679454',
                                                                       'poller_hash': 'REZGREU4cGVlcmludGZwb24gMTgwNzY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180764',
                                                                                               'links': [       {       'connected_time': '688364',
                                                                                                                        'fiber_length': 2578,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREU4cGVlcnBvbnBvbiAxODA3NjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '97 PUGH ST (SKOMRA MARGARET)',
                                         'poller_hash': 'REZGREU4',
                                         'serial_number': 'DFFDE8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:27:f8',
                                         'uptime': '688364'},
                                 {       'counter_source': '181001/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '181001/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '66181739690',
                                                                       'in_ucast_pkts': '17302872467',
                                                                       'mac_address': '84:d3:43:7e:30:17',
                                                                       'name': '181001/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '322163000643',
                                                                       'out_ucast_pkts': '321808779',
                                                                       'poller_hash': 'QzgzQjZDcGVlcmludGYxODEwMDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '181001',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '322163000643',
                                                                       'in_ucast_pkts': '321808779',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '322163000643',
                                                                                                'in_ucast_pkts': '321808779',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '66181739690',
                                                                                                'out_ucast_pkts': '17302872467',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '66181739690',
                                                                       'out_ucast_pkts': '17302872467',
                                                                       'poller_hash': 'QzgzQjZDcGVlcmludGZwb24gMTgxMDAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '181001',
                                                                                               'links': [       {       'connected_time': '21816725',
                                                                                                                        'fiber_length': 11426,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjZDcGVlcnBvbnBvbiAxODEwMDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6990 CHURCH ST (ZEHR DYLAN)',
                                         'poller_hash': 'QzgzQjZD',
                                         'serial_number': 'C83B6C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:30:17',
                                         'uptime': '21816725'},
                                 {       'counter_source': '182241/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:e9:8f'],
                                                                       'id': '182241/g1',
                                                                       'in_nucast_pkts': '613401',
                                                                       'in_octets': '68467731802',
                                                                       'in_ucast_pkts': '172767185',
                                                                       'mac_address': '84:d3:43:7e:33:62',
                                                                       'mtu': 2000,
                                                                       'name': '182241/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '74641511',
                                                                       'out_octets': '1072255058835',
                                                                       'out_ucast_pkts': '827923932',
                                                                       'poller_hash': 'QzgzQkQ1cGVlcmludGYxODIyNDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '182241',
                                                                       'in_nucast_pkts': '74641511',
                                                                       'in_octets': '1072255058835',
                                                                       'in_ucast_pkts': '827923932',
                                                                       'links': [       {       'in_nucast_pkts': '74641511',
                                                                                                'in_octets': '1072255058835',
                                                                                                'in_ucast_pkts': '827923932',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '613401',
                                                                                                'out_octets': '68467731802',
                                                                                                'out_ucast_pkts': '172767185',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '613401',
                                                                       'out_octets': '68467731802',
                                                                       'out_ucast_pkts': '172767185',
                                                                       'poller_hash': 'QzgzQkQ1cGVlcmludGZwb24gMTgyMjQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182241',
                                                                                               'links': [       {       'connected_time': '21816763',
                                                                                                                        'fiber_length': 11467,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -24.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkQ1cGVlcnBvbnBvbiAxODIyNDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6999 MILLBANK MAIN ST (JANTZI EZRA & '
                                                 'ELIZABETH)',
                                         'poller_hash': 'QzgzQkQ1',
                                         'serial_number': 'C83BD5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:33:5f',
                                         'uptime': '21816763'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '182261/g1',
                                                                       'name': '182261/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzQkZGcGVlcmludGYxODIyNjEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '182261',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgzQkZGcGVlcmludGZwb24gMTgyMjYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182261',
                                                                                               'links': [       {       'connected_time': '21816000',
                                                                                                                        'fiber_length': 11727,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkZGcGVlcnBvbnBvbiAxODIyNjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6990 FREEBORN ST (ROBERTS RICHARD & '
                                                 'JUDY)',
                                         'poller_hash': 'QzgzQkZG',
                                         'serial_number': 'C83BFF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:34:af',
                                         'uptime': '21816000'},
                                 {       'counter_source': '182271/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:ef:95'],
                                                                       'id': '182271/g1',
                                                                       'in_nucast_pkts': '199457',
                                                                       'in_octets': '17872795505',
                                                                       'in_ucast_pkts': '46241494',
                                                                       'mac_address': '84:d3:43:7e:0a:da',
                                                                       'mtu': 2000,
                                                                       'name': '182271/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23990096',
                                                                       'out_octets': '165540549854',
                                                                       'out_ucast_pkts': '144515071',
                                                                       'poller_hash': 'QzgzNkM0cGVlcmludGYxODIyNzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '182271',
                                                                       'in_nucast_pkts': '23990096',
                                                                       'in_octets': '165540549854',
                                                                       'in_ucast_pkts': '144515071',
                                                                       'links': [       {       'in_nucast_pkts': '23990096',
                                                                                                'in_octets': '165540549854',
                                                                                                'in_ucast_pkts': '144515071',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '199457',
                                                                                                'out_octets': '17872795505',
                                                                                                'out_ucast_pkts': '46241494',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '199457',
                                                                       'out_octets': '17872795505',
                                                                       'out_ucast_pkts': '46241494',
                                                                       'poller_hash': 'QzgzNkM0cGVlcmludGZwb24gMTgyMjcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182271',
                                                                                               'links': [       {       'connected_time': '21816857',
                                                                                                                        'fiber_length': 11627,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzNkM0cGVlcnBvbnBvbiAxODIyNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6991 MILLBANK MAIN ST (WIEBE EVA)',
                                         'poller_hash': 'QzgzNkM0',
                                         'serial_number': 'C836C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:0a:d7',
                                         'uptime': '21816857'},
                                 {       'counter_source': '182292/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ec:2e'],
                                                                       'id': '182292/g1',
                                                                       'in_nucast_pkts': '444104',
                                                                       'in_octets': '142814072481',
                                                                       'in_ucast_pkts': '499266109',
                                                                       'mac_address': 'f8:85:f9:17:15:13',
                                                                       'mtu': 2000,
                                                                       'name': '182292/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '60233653',
                                                                       'out_octets': '3499755642020',
                                                                       'out_ucast_pkts': '2733245579',
                                                                       'poller_hash': 'REZGQjhCcGVlcmludGYxODIyOTIvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '182292',
                                                                       'in_nucast_pkts': '60233653',
                                                                       'in_octets': '3499755642020',
                                                                       'in_ucast_pkts': '2733245579',
                                                                       'links': [       {       'in_nucast_pkts': '60233653',
                                                                                                'in_octets': '3499755642020',
                                                                                                'in_ucast_pkts': '2733245579',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '444104',
                                                                                                'out_octets': '142814072481',
                                                                                                'out_ucast_pkts': '499266109',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '444104',
                                                                       'out_octets': '142814072481',
                                                                       'out_ucast_pkts': '499266109',
                                                                       'poller_hash': 'REZGQjhCcGVlcmludGZwb24gMTgyMjky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182292',
                                                                                               'links': [       {       'connected_time': '25305799',
                                                                                                                        'fiber_length': 12067,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGQjhCcGVlcnBvbnBvbiAxODIyOTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4063 NORFOLK ST (SADIWNYK CATHERINE)',
                                         'poller_hash': 'REZGQjhC',
                                         'serial_number': 'DFFB8B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:15:10',
                                         'uptime': '25305799'},
                                 {       'counter_source': '182302/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '182302/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '28316392064',
                                                                       'in_ucast_pkts': '123791302',
                                                                       'mac_address': '84:d3:43:7e:34:37',
                                                                       'name': '182302/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '850946041316',
                                                                       'out_ucast_pkts': '722498072',
                                                                       'poller_hash': 'QzgzQkYwcGVlcmludGYxODIzMDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '182302',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '850946041316',
                                                                       'in_ucast_pkts': '722498072',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '850946041316',
                                                                                                'in_ucast_pkts': '722498072',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '28316392064',
                                                                                                'out_ucast_pkts': '123791302',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '28316392064',
                                                                       'out_ucast_pkts': '123791302',
                                                                       'poller_hash': 'QzgzQkYwcGVlcmludGZwb24gMTgyMzAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182302',
                                                                                               'links': [       {       'connected_time': '21816094',
                                                                                                                        'fiber_length': 11750,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkYwcGVlcnBvbnBvbiAxODIzMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6996 MILLBANK MAIN ST (GERBER FLOYD)',
                                         'poller_hash': 'QzgzQkYw',
                                         'serial_number': 'C83BF0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:34:37',
                                         'uptime': '21816094'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '182335/g1',
                                                                       'name': '182335/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzMEU4cGVlcmludGYxODIzMzUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '182335',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzMEU4cGVlcmludGZwb24gMTgyMzM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182335',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMEU4cGVlcnBvbnBvbiAxODIzMzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '6990 RUTHERFORD ST (BURNS BLAKE)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "182335" '
                                                                  '(location: GPON '
                                                                  'port "1/7"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-04-01T00:17:09.260Z',
                                         'poller_hash': 'QzgzMEU4',
                                         'serial_number': 'C830E8',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '182345/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:bc:30'],
                                                                       'id': '182345/g1',
                                                                       'in_nucast_pkts': '66951',
                                                                       'in_octets': '19630593708',
                                                                       'in_ucast_pkts': '112356840',
                                                                       'mac_address': 'f8:85:f9:17:28:c3',
                                                                       'mtu': 2000,
                                                                       'name': '182345/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6781939',
                                                                       'out_octets': '523511833952',
                                                                       'out_ucast_pkts': '378514507',
                                                                       'poller_hash': 'REZGRTAxcGVlcmludGYxODIzNDUvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '182345',
                                                                       'in_nucast_pkts': '6781939',
                                                                       'in_octets': '523511833952',
                                                                       'in_ucast_pkts': '378514507',
                                                                       'links': [       {       'in_nucast_pkts': '6781939',
                                                                                                'in_octets': '523511833952',
                                                                                                'in_ucast_pkts': '378514507',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '66951',
                                                                                                'out_octets': '19630593708',
                                                                                                'out_ucast_pkts': '112356840',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66951',
                                                                       'out_octets': '19630593708',
                                                                       'out_ucast_pkts': '112356840',
                                                                       'poller_hash': 'REZGRTAxcGVlcmludGZwb24gMTgyMzQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182345',
                                                                                               'links': [       {       'connected_time': '3113810',
                                                                                                                        'fiber_length': 18301,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTAxcGVlcnBvbnBvbiAxODIzNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 SCHADE ST (BAST SANDI)',
                                         'poller_hash': 'REZGRTAx',
                                         'serial_number': 'DFFE01',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:28:c0',
                                         'uptime': '3113810'},
                                 {       'counter_source': '183656/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '183656/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '222688948643',
                                                                       'in_ucast_pkts': '26549638688',
                                                                       'mac_address': '84:d3:43:7d:dc:7f',
                                                                       'name': '183656/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '74',
                                                                       'out_octets': '2692933599529',
                                                                       'out_ucast_pkts': '2120661577',
                                                                       'poller_hash': 'QzgzMEY5cGVlcmludGYxODM2NTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '183656',
                                                                       'in_nucast_pkts': '74',
                                                                       'in_octets': '2692933599529',
                                                                       'in_ucast_pkts': '2120661577',
                                                                       'links': [       {       'in_nucast_pkts': '74',
                                                                                                'in_octets': '2692933599529',
                                                                                                'in_ucast_pkts': '2120661577',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '222688948643',
                                                                                                'out_ucast_pkts': '26549638688',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '222688948643',
                                                                       'out_ucast_pkts': '26549638688',
                                                                       'poller_hash': 'QzgzMEY5cGVlcmludGZwb24gMTgzNjU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '183656',
                                                                                               'links': [       {       'connected_time': '688348',
                                                                                                                        'fiber_length': 11445,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMEY5cGVlcnBvbnBvbiAxODM2NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6989 CHURCH ST (ZWART JANELLE)',
                                         'poller_hash': 'QzgzMEY5',
                                         'serial_number': 'C830F9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:dc:7f',
                                         'uptime': '688348'},
                                 {       'counter_source': '183666/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '183666/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '505822873053',
                                                                       'in_ucast_pkts': '14951399657',
                                                                       'mac_address': 'f8:85:f9:17:25:c8',
                                                                       'name': '183666/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '7370525443020',
                                                                       'out_ucast_pkts': '5640560103',
                                                                       'poller_hash': 'REZGREEycGVlcmludGYxODM2NjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '183666',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '7370525443020',
                                                                       'in_ucast_pkts': '5640560103',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '7370525443020',
                                                                                                'in_ucast_pkts': '5640560103',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '505822873053',
                                                                                                'out_ucast_pkts': '14951399657',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '505822873053',
                                                                       'out_ucast_pkts': '14951399657',
                                                                       'poller_hash': 'REZGREEycGVlcmludGZwb24gMTgzNjY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '183666',
                                                                                               'links': [       {       'connected_time': '74480379',
                                                                                                                        'fiber_length': 11424,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREEycGVlcnBvbnBvbiAxODM2NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6991 CHURCH ST (TOMBLIN PAUL)',
                                         'poller_hash': 'REZGREEy',
                                         'serial_number': 'DFFDA2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:25:c8',
                                         'uptime': '74480379'},
                                 {       'counter_source': '183676/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '183676/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '67126724320',
                                                                       'in_ucast_pkts': '336828729',
                                                                       'mac_address': 'f8:85:f9:17:2a:c0',
                                                                       'name': '183676/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1382342806805',
                                                                       'out_ucast_pkts': '1067039743',
                                                                       'poller_hash': 'REZGRTQxcGVlcmludGYxODM2NzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '183676',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1382342806805',
                                                                       'in_ucast_pkts': '1067039743',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1382342806805',
                                                                                                'in_ucast_pkts': '1067039743',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '67126724320',
                                                                                                'out_ucast_pkts': '336828729',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '67126724320',
                                                                       'out_ucast_pkts': '336828729',
                                                                       'poller_hash': 'REZGRTQxcGVlcmludGZwb24gMTgzNjc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '183676',
                                                                                               'links': [       {       'connected_time': '21815991',
                                                                                                                        'fiber_length': 11481,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTQxcGVlcnBvbnBvbiAxODM2NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4059 PRINCESS ST (BUECKERT HENRY & '
                                                 'ANNA)',
                                         'poller_hash': 'REZGRTQx',
                                         'serial_number': 'DFFE41',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2a:c0',
                                         'uptime': '21815991'},
                                 {       'counter_source': '184715/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '184715/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '486259273444',
                                                                       'in_ucast_pkts': '130161343401',
                                                                       'mac_address': 'f8:85:f9:17:12:b8',
                                                                       'name': '184715/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '3271780468616',
                                                                       'out_ucast_pkts': '2673734730',
                                                                       'poller_hash': 'REZGQjQwcGVlcmludGYxODQ3MTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '184715',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '3271780468616',
                                                                       'in_ucast_pkts': '2673734730',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '3271780468616',
                                                                                                'in_ucast_pkts': '2673734730',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '486259273444',
                                                                                                'out_ucast_pkts': '130161343401',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '486259273444',
                                                                       'out_ucast_pkts': '130161343401',
                                                                       'poller_hash': 'REZGQjQwcGVlcmludGZwb24gMTg0NzE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184715',
                                                                                               'links': [       {       'connected_time': '523096',
                                                                                                                        'fiber_length': 11102,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGQjQwcGVlcnBvbnBvbiAxODQ3MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4084 PRINCESS ST (REINDL MELISSA)',
                                         'poller_hash': 'REZGQjQw',
                                         'serial_number': 'DFFB40',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:12:b8',
                                         'uptime': '523096'},
                                 {       'counter_source': '184745/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '184745/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '503125359973',
                                                                       'in_ucast_pkts': '31683836879',
                                                                       'mac_address': '84:d3:43:7e:30:f7',
                                                                       'name': '184745/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '4458013344774',
                                                                       'out_ucast_pkts': '3488774014',
                                                                       'poller_hash': 'QzgzQjg4cGVlcmludGYxODQ3NDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '184745',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '4458013344774',
                                                                       'in_ucast_pkts': '3488774014',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '4458013344774',
                                                                                                'in_ucast_pkts': '3488774014',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '503125359973',
                                                                                                'out_ucast_pkts': '31683836879',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '503125359973',
                                                                       'out_ucast_pkts': '31683836879',
                                                                       'poller_hash': 'QzgzQjg4cGVlcmludGZwb24gMTg0NzQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184745',
                                                                                               'links': [       {       'connected_time': '21816369',
                                                                                                                        'fiber_length': 11986,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjg4cGVlcnBvbnBvbiAxODQ3NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4051 NORFOLK ST MILLBANK (HORST BILL '
                                                 'ACUTIONEER)',
                                         'poller_hash': 'QzgzQjg4',
                                         'serial_number': 'C83B88',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:30:f7',
                                         'uptime': '21816369'},
                                 {       'counter_source': '184755/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '184755/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '92675074820',
                                                                       'in_ucast_pkts': '229543532',
                                                                       'mac_address': '84:d3:43:7e:32:97',
                                                                       'name': '184755/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '147',
                                                                       'out_octets': '677396310975',
                                                                       'out_ucast_pkts': '602545077',
                                                                       'poller_hash': 'QzgzQkJDcGVlcmludGYxODQ3NTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '184755',
                                                                       'in_nucast_pkts': '147',
                                                                       'in_octets': '677396310975',
                                                                       'in_ucast_pkts': '602545077',
                                                                       'links': [       {       'in_nucast_pkts': '147',
                                                                                                'in_octets': '677396310975',
                                                                                                'in_ucast_pkts': '602545077',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '92675074820',
                                                                                                'out_ucast_pkts': '229543532',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '92675074820',
                                                                       'out_ucast_pkts': '229543532',
                                                                       'poller_hash': 'QzgzQkJDcGVlcmludGZwb24gMTg0NzU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184755',
                                                                                               'links': [       {       'connected_time': '21816264',
                                                                                                                        'fiber_length': 11595,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkJDcGVlcnBvbnBvbiAxODQ3NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4049 PRINCESS ST (BIEHN RUTH ANNE)',
                                         'poller_hash': 'QzgzQkJD',
                                         'serial_number': 'C83BBC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:97',
                                         'uptime': '21816264'},
                                 {       'counter_source': '184765/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8c:77'],
                                                                       'id': '184765/g1',
                                                                       'in_nucast_pkts': '262925',
                                                                       'in_octets': '90496395418',
                                                                       'in_ucast_pkts': '1163163512',
                                                                       'mac_address': '84:d3:43:7e:31:32',
                                                                       'mtu': 2000,
                                                                       'name': '184765/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30390529',
                                                                       'out_octets': '3504477759681',
                                                                       'out_ucast_pkts': '2355516950',
                                                                       'poller_hash': 'QzgzQjhGcGVlcmludGYxODQ3NjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '184765',
                                                                       'in_nucast_pkts': '30390529',
                                                                       'in_octets': '3504477759681',
                                                                       'in_ucast_pkts': '2355516950',
                                                                       'links': [       {       'in_nucast_pkts': '30390529',
                                                                                                'in_octets': '3504477759681',
                                                                                                'in_ucast_pkts': '2355516950',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '262925',
                                                                                                'out_octets': '90496395418',
                                                                                                'out_ucast_pkts': '1163163512',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '262925',
                                                                       'out_octets': '90496395418',
                                                                       'out_ucast_pkts': '1163163512',
                                                                       'poller_hash': 'QzgzQjhGcGVlcmludGZwb24gMTg0NzY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184765',
                                                                                               'links': [       {       'connected_time': '15498586',
                                                                                                                        'fiber_length': 12020,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjhGcGVlcnBvbnBvbiAxODQ3NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6935 NICKLIN ST MILLBANK (LICHTY '
                                                 'DEL)',
                                         'poller_hash': 'QzgzQjhG',
                                         'serial_number': 'C83B8F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:2f',
                                         'uptime': '15498586'},
                                 {       'counter_source': '184775/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '184775/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '179998021223',
                                                                       'in_ucast_pkts': '1110679394',
                                                                       'mac_address': '84:d3:43:7e:0c:f7',
                                                                       'name': '184775/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '7554978269371',
                                                                       'out_ucast_pkts': '5646401367',
                                                                       'poller_hash': 'QzgzNzA4cGVlcmludGYxODQ3NzUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '184775',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '7554978269371',
                                                                       'in_ucast_pkts': '5646401367',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '7554978269371',
                                                                                                'in_ucast_pkts': '5646401367',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '179998021223',
                                                                                                'out_ucast_pkts': '1110679394',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '179998021223',
                                                                       'out_ucast_pkts': '1110679394',
                                                                       'poller_hash': 'QzgzNzA4cGVlcmludGZwb24gMTg0Nzc1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184775',
                                                                                               'links': [       {       'connected_time': '688362',
                                                                                                                        'fiber_length': 11858,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzNzA4cGVlcnBvbnBvbiAxODQ3NzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4050 NORFOLK ST (WAGLER KRISTA)',
                                         'poller_hash': 'QzgzNzA4',
                                         'serial_number': 'C83708',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:0c:f7',
                                         'uptime': '688362'},
                                 {       'counter_source': '184785/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '184785/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '123510130289',
                                                                       'in_ucast_pkts': '197162462',
                                                                       'mac_address': '84:d3:43:7e:32:57',
                                                                       'name': '184785/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '343160165537',
                                                                       'out_ucast_pkts': '361522999',
                                                                       'poller_hash': 'QzgzQkI0cGVlcmludGYxODQ3ODUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '184785',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '343160165537',
                                                                       'in_ucast_pkts': '361522999',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '343160165537',
                                                                                                'in_ucast_pkts': '361522999',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '123510130289',
                                                                                                'out_ucast_pkts': '197162462',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '123510130289',
                                                                       'out_ucast_pkts': '197162462',
                                                                       'poller_hash': 'QzgzQkI0cGVlcmludGZwb24gMTg0Nzg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184785',
                                                                                               'links': [       {       'connected_time': '21815790',
                                                                                                                        'fiber_length': 11112,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkI0cGVlcnBvbnBvbiAxODQ3ODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4086 PRINCESS ST MILLBANK (ZEHR '
                                                 'DAVID)',
                                         'poller_hash': 'QzgzQkI0',
                                         'serial_number': 'C83BB4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:57',
                                         'uptime': '21815790'},
                                 {       'counter_source': '185171/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:14:d7'],
                                                                       'id': '185171/g1',
                                                                       'in_nucast_pkts': '513285',
                                                                       'in_octets': '104516592021',
                                                                       'in_ucast_pkts': '930580447',
                                                                       'mac_address': '84:d3:43:7e:31:12',
                                                                       'mtu': 2000,
                                                                       'name': '185171/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67509329',
                                                                       'out_octets': '5607462190608',
                                                                       'out_ucast_pkts': '3829093714',
                                                                       'poller_hash': 'QzgzQjhCcGVlcmludGYxODUxNzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '185171',
                                                                       'in_nucast_pkts': '67509329',
                                                                       'in_octets': '5607462190608',
                                                                       'in_ucast_pkts': '3829093714',
                                                                       'links': [       {       'in_nucast_pkts': '67509329',
                                                                                                'in_octets': '5607462190608',
                                                                                                'in_ucast_pkts': '3829093714',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '513285',
                                                                                                'out_octets': '104516592021',
                                                                                                'out_ucast_pkts': '930580447',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '513285',
                                                                       'out_octets': '104516592021',
                                                                       'out_ucast_pkts': '930580447',
                                                                       'poller_hash': 'QzgzQjhCcGVlcmludGZwb24gMTg1MTcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185171',
                                                                                               'links': [       {       'connected_time': '28453801',
                                                                                                                        'fiber_length': 12118,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjhCcGVlcnBvbnBvbiAxODUxNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6947 NICKLIN ST (EDWARDS ERNEST & '
                                                 'DONNA)',
                                         'poller_hash': 'QzgzQjhC',
                                         'serial_number': 'C83B8B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:0f',
                                         'uptime': '28453801'},
                                 {       'counter_source': '185181/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185181/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8531467426',
                                                                       'in_ucast_pkts': '35041838',
                                                                       'mac_address': '84:d3:43:7d:d8:f7',
                                                                       'name': '185181/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '146',
                                                                       'out_octets': '118875295991',
                                                                       'out_ucast_pkts': '137745264',
                                                                       'poller_hash': 'QzgzMDg4cGVlcmludGYxODUxODEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185181',
                                                                       'in_nucast_pkts': '146',
                                                                       'in_octets': '118875295991',
                                                                       'in_ucast_pkts': '137745264',
                                                                       'links': [       {       'in_nucast_pkts': '146',
                                                                                                'in_octets': '118875295991',
                                                                                                'in_ucast_pkts': '137745264',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '8531467426',
                                                                                                'out_ucast_pkts': '35041838',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8531467426',
                                                                       'out_ucast_pkts': '35041838',
                                                                       'poller_hash': 'QzgzMDg4cGVlcmludGZwb24gMTg1MTgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185181',
                                                                                               'links': [       {       'connected_time': '2041884',
                                                                                                                        'fiber_length': 11982,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMDg4cGVlcnBvbnBvbiAxODUxODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4053 NORFOLK ST (WAGLER RICHARD)',
                                         'poller_hash': 'QzgzMDg4',
                                         'serial_number': 'C83088',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:d8:f7',
                                         'uptime': '2041884'},
                                 {       'counter_source': '185236/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185236/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11207505808',
                                                                       'in_ucast_pkts': '26958749',
                                                                       'mac_address': 'f8:85:f9:17:2a:b0',
                                                                       'name': '185236/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '49571635091',
                                                                       'out_ucast_pkts': '58244569',
                                                                       'poller_hash': 'REZGRTNGcGVlcmludGYxODUyMzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185236',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '49571635091',
                                                                       'in_ucast_pkts': '58244569',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '49571635091',
                                                                                                'in_ucast_pkts': '58244569',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '11207505808',
                                                                                                'out_ucast_pkts': '26958749',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11207505808',
                                                                       'out_ucast_pkts': '26958749',
                                                                       'poller_hash': 'REZGRTNGcGVlcmludGZwb24gMTg1MjM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185236',
                                                                                               'links': [       {       'connected_time': '14746251',
                                                                                                                        'fiber_length': 11066,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTNGcGVlcnBvbnBvbiAxODUyMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4091 PRINCESS ST (REIMER DARREN & '
                                                 'LISA)',
                                         'poller_hash': 'REZGRTNG',
                                         'serial_number': 'DFFE3F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2a:b0',
                                         'uptime': '14746251'},
                                 {       'counter_source': '185256/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185256/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '113765594256',
                                                                       'in_ucast_pkts': '560177583',
                                                                       'mac_address': 'f8:85:f9:17:1c:38',
                                                                       'name': '185256/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '2316304197789',
                                                                       'out_ucast_pkts': '1908363399',
                                                                       'poller_hash': 'REZGQzcwcGVlcmludGYxODUyNTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185256',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '2316304197789',
                                                                       'in_ucast_pkts': '1908363399',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '2316304197789',
                                                                                                'in_ucast_pkts': '1908363399',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '113765594256',
                                                                                                'out_ucast_pkts': '560177583',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '113765594256',
                                                                       'out_ucast_pkts': '560177583',
                                                                       'poller_hash': 'REZGQzcwcGVlcmludGZwb24gMTg1MjU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185256',
                                                                                               'links': [       {       'connected_time': '69274060',
                                                                                                                        'fiber_length': 11600,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGQzcwcGVlcnBvbnBvbiAxODUyNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4056 WATERLOO ST (GIESBRECHT ABRAM & '
                                                 'KATHARINA)',
                                         'poller_hash': 'REZGQzcw',
                                         'serial_number': 'DFFC70',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:1c:38',
                                         'uptime': '69274060'},
                                 {       'counter_source': '185388/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185388/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '114423833352',
                                                                       'in_ucast_pkts': '447369076',
                                                                       'mac_address': 'f8:85:f9:17:25:98',
                                                                       'name': '185388/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2575618660266',
                                                                       'out_ucast_pkts': '1794520043',
                                                                       'poller_hash': 'REZGRDlDcGVlcmludGYxODUzODgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185388',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2575618660266',
                                                                       'in_ucast_pkts': '1794520043',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2575618660266',
                                                                                                'in_ucast_pkts': '1794520043',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '114423833352',
                                                                                                'out_ucast_pkts': '447369076',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '114423833352',
                                                                       'out_ucast_pkts': '447369076',
                                                                       'poller_hash': 'REZGRDlDcGVlcmludGZwb24gMTg1Mzg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185388',
                                                                                               'links': [       {       'connected_time': '68525743',
                                                                                                                        'fiber_length': 11150,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRDlDcGVlcnBvbnBvbiAxODUzODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4094 PRINCESS ST (DISTEFAN RIA)',
                                         'poller_hash': 'REZGRDlD',
                                         'serial_number': 'DFFD9C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:25:98',
                                         'uptime': '68525743'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '185408/g1',
                                                                       'name': '185408/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'REZGQzAycGVlcmludGYxODU0MDgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185408',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'REZGQzAycGVlcmludGZwb24gMTg1NDA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185408',
                                                                                               'links': [       {       'connected_time': '21815953',
                                                                                                                        'fiber_length': 12003,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGQzAycGVlcnBvbnBvbiAxODU0MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4058 NORFOLK ST (KUEPFER RAYMOND K)',
                                         'poller_hash': 'REZGQzAy',
                                         'serial_number': 'DFFC02',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:18:c8',
                                         'uptime': '21815953'},
                                 {       'counter_source': '185418/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185418/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '159373492039',
                                                                       'in_ucast_pkts': '490358623',
                                                                       'mac_address': 'f8:85:f9:17:2b:08',
                                                                       'name': '185418/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '2362233602308',
                                                                       'out_ucast_pkts': '1831573515',
                                                                       'poller_hash': 'REZGRTRBcGVlcmludGYxODU0MTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185418',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '2362233602308',
                                                                       'in_ucast_pkts': '1831573515',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '2362233602308',
                                                                                                'in_ucast_pkts': '1831573515',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '159373492039',
                                                                                                'out_ucast_pkts': '490358623',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '159373492039',
                                                                       'out_ucast_pkts': '490358623',
                                                                       'poller_hash': 'REZGRTRBcGVlcmludGZwb24gMTg1NDE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185418',
                                                                                               'links': [       {       'connected_time': '523102',
                                                                                                                        'fiber_length': 11587,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTRBcGVlcnBvbnBvbiAxODU0MTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4053 PRINCESS ST (VAN BAKEL BRYAN)',
                                         'poller_hash': 'REZGRTRB',
                                         'serial_number': 'DFFE4A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2b:08',
                                         'uptime': '523102'},
                                 {       'counter_source': '185428/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185428/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '24568458634',
                                                                       'in_ucast_pkts': '4394564576',
                                                                       'mac_address': 'f8:85:f9:17:2a:40',
                                                                       'name': '185428/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '120',
                                                                       'out_octets': '689099899600',
                                                                       'out_ucast_pkts': '559536211',
                                                                       'poller_hash': 'REZGRTMxcGVlcmludGYxODU0MjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185428',
                                                                       'in_nucast_pkts': '120',
                                                                       'in_octets': '689099899600',
                                                                       'in_ucast_pkts': '559536211',
                                                                       'links': [       {       'in_nucast_pkts': '120',
                                                                                                'in_octets': '689099899600',
                                                                                                'in_ucast_pkts': '559536211',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '24568458634',
                                                                                                'out_ucast_pkts': '4394564576',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '24568458634',
                                                                       'out_ucast_pkts': '4394564576',
                                                                       'poller_hash': 'REZGRTMxcGVlcmludGZwb24gMTg1NDI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185428',
                                                                                               'links': [       {       'connected_time': '22882706',
                                                                                                                        'fiber_length': 12024,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTMxcGVlcnBvbnBvbiAxODU0Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4052 NORFOLK ST (KUEPFER NELSON)',
                                         'poller_hash': 'REZGRTMx',
                                         'serial_number': 'DFFE31',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2a:40',
                                         'uptime': '22882706'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '185438/g1',
                                                                       'name': '185438/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxREQ1cGVlcmludGYxODU0MzgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185438',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxREQ1cGVlcmludGZwb24gMTg1NDM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185438',
                                                                                               'links': [       {       'connected_time': '6029064',
                                                                                                                        'fiber_length': 12028,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREQ1cGVlcnBvbnBvbiAxODU0Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4056 NORFOLK ST (ZACHARIAS JAKE)',
                                         'poller_hash': 'QzgxREQ1',
                                         'serial_number': 'C81DD5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:43:5f',
                                         'uptime': '6029064'},
                                 {       'counter_source': '185497/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185497/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1534004368891',
                                                                       'in_ucast_pkts': '44851822037',
                                                                       'mac_address': '84:d3:43:7d:38:4f',
                                                                       'name': '185497/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '1721015254856',
                                                                       'out_ucast_pkts': '1628818068',
                                                                       'poller_hash': 'QzgxQzczcGVlcmludGYxODU0OTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185497',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '1721015254856',
                                                                       'in_ucast_pkts': '1628818068',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '1721015254856',
                                                                                                'in_ucast_pkts': '1628818068',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1534004368891',
                                                                                                'out_ucast_pkts': '44851822037',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1534004368891',
                                                                       'out_ucast_pkts': '44851822037',
                                                                       'poller_hash': 'QzgxQzczcGVlcmludGZwb24gMTg1NDk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185497',
                                                                                               'links': [       {       'connected_time': '21816941',
                                                                                                                        'fiber_length': 11543,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxQzczcGVlcnBvbnBvbiAxODU0OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6974 CHURCH ST (MILLBANK CHEESE AND '
                                                 'BUTTER)',
                                         'poller_hash': 'QzgxQzcz',
                                         'serial_number': 'C81C73',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:38:4f',
                                         'uptime': '21816941'},
                                 {       'counter_source': '185701/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185701/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '169112998667',
                                                                       'in_ucast_pkts': '168574049',
                                                                       'mac_address': '84:d3:43:7d:23:9f',
                                                                       'name': '185701/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '60741735729',
                                                                       'out_ucast_pkts': '127997431',
                                                                       'poller_hash': 'QzgxOUREcGVlcmludGYxODU3MDEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185701',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '60741735729',
                                                                       'in_ucast_pkts': '127997431',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '60741735729',
                                                                                                'in_ucast_pkts': '127997431',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '169112998667',
                                                                                                'out_ucast_pkts': '168574049',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '169112998667',
                                                                       'out_ucast_pkts': '168574049',
                                                                       'poller_hash': 'QzgxOUREcGVlcmludGZwb24gMTg1NzAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185701',
                                                                                               'links': [       {       'connected_time': '6774594',
                                                                                                                        'fiber_length': 11600,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxOUREcGVlcnBvbnBvbiAxODU3MDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6974 CHURCH ST (OXFORD MILKWAY - '
                                                 'MILLBANK DEPO)',
                                         'poller_hash': 'QzgxOURE',
                                         'serial_number': 'C819DD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:23:9f',
                                         'uptime': '6774594'},
                                 {       'counter_source': '185711/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185711/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '458965789610',
                                                                       'in_ucast_pkts': '26998100584',
                                                                       'mac_address': 'f8:85:f9:17:25:38',
                                                                       'name': '185711/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '3136034658086',
                                                                       'out_ucast_pkts': '2824624561',
                                                                       'poller_hash': 'REZGRDkwcGVlcmludGYxODU3MTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185711',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '3136034658086',
                                                                       'in_ucast_pkts': '2824624561',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '3136034658086',
                                                                                                'in_ucast_pkts': '2824624561',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '458965789610',
                                                                                                'out_ucast_pkts': '26998100584',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '458965789610',
                                                                       'out_ucast_pkts': '26998100584',
                                                                       'poller_hash': 'REZGRDkwcGVlcmludGZwb24gMTg1NzEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185711',
                                                                                               'links': [       {       'connected_time': '9398572',
                                                                                                                        'fiber_length': 11196,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRDkwcGVlcnBvbnBvbiAxODU3MTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4079 PRINCESS ST (MAGDANZ VERONIKA)',
                                         'poller_hash': 'REZGRDkw',
                                         'serial_number': 'DFFD90',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:25:38',
                                         'uptime': '9398572'},
                                 {       'counter_source': '185721/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185721/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '211003831310',
                                                                       'in_ucast_pkts': '2351554557',
                                                                       'mac_address': '84:d3:43:7d:3f:c7',
                                                                       'name': '185721/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '14147598260495',
                                                                       'out_ucast_pkts': '9460227816',
                                                                       'poller_hash': 'QzgxRDYycGVlcmludGYxODU3MjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185721',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '14147598260495',
                                                                       'in_ucast_pkts': '9460227816',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '14147598260495',
                                                                                                'in_ucast_pkts': '9460227816',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '211003831310',
                                                                                                'out_ucast_pkts': '2351554557',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '211003831310',
                                                                       'out_ucast_pkts': '2351554557',
                                                                       'poller_hash': 'QzgxRDYycGVlcmludGZwb24gMTg1NzIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185721',
                                                                                               'links': [       {       'connected_time': '69480697',
                                                                                                                        'fiber_length': 12027,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDYycGVlcnBvbnBvbiAxODU3MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6945 NICKLIN ST (ROES IVAN & RETA)',
                                         'poller_hash': 'QzgxRDYy',
                                         'serial_number': 'C81D62',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:3f:c7',
                                         'uptime': '69480697'},
                                 {       'counter_source': '185756/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '185756/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '529436055402',
                                                                       'in_ucast_pkts': '117191576259',
                                                                       'mac_address': 'f8:85:f9:17:29:18',
                                                                       'name': '185756/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '2891271805886',
                                                                       'out_ucast_pkts': '2375790377',
                                                                       'poller_hash': 'REZGRTBDcGVlcmludGYxODU3NTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185756',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '2891271805886',
                                                                       'in_ucast_pkts': '2375790377',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '2891271805886',
                                                                                                'in_ucast_pkts': '2375790377',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '529436055402',
                                                                                                'out_ucast_pkts': '117191576259',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '529436055402',
                                                                       'out_ucast_pkts': '117191576259',
                                                                       'poller_hash': 'REZGRTBDcGVlcmludGZwb24gMTg1NzU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185756',
                                                                                               'links': [       {       'connected_time': '523134',
                                                                                                                        'fiber_length': 11330,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTBDcGVlcnBvbnBvbiAxODU3NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6967 PERTH ST (BRENNEMAN JEREMY & '
                                                 'BEVERLEY)',
                                         'poller_hash': 'REZGRTBD',
                                         'serial_number': 'DFFE0C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:29:18',
                                         'uptime': '523134'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '185766/g1',
                                                                       'name': '185766/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'REZGREQ3cGVlcmludGYxODU3NjYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '185766',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'REZGREQ3cGVlcmludGZwb24gMTg1NzY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185766',
                                                                                               'links': [       {       'connected_time': '21816827',
                                                                                                                        'fiber_length': 11229,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREQ3cGVlcnBvbnBvbiAxODU3NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4090 HILLCREST ST (WAGLER A M)',
                                         'poller_hash': 'REZGREQ3',
                                         'serial_number': 'DFFDD7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:27:70',
                                         'uptime': '21816827'},
                                 {       'counter_source': '186335/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186335/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '14520245020',
                                                                       'in_ucast_pkts': '43236942',
                                                                       'mac_address': '84:d3:43:7d:dc:f7',
                                                                       'name': '186335/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '193097894521',
                                                                       'out_ucast_pkts': '156950931',
                                                                       'poller_hash': 'QzgzMTA4cGVlcmludGYxODYzMzUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186335',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '193097894521',
                                                                       'in_ucast_pkts': '156950931',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '193097894521',
                                                                                                'in_ucast_pkts': '156950931',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '14520245020',
                                                                                                'out_ucast_pkts': '43236942',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '14520245020',
                                                                       'out_ucast_pkts': '43236942',
                                                                       'poller_hash': 'QzgzMTA4cGVlcmludGZwb24gMTg2MzM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186335',
                                                                                               'links': [       {       'connected_time': '21815895',
                                                                                                                        'fiber_length': 11417,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTA4cGVlcnBvbnBvbiAxODYzMzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6939 ELGIN ST (GERBER RAY)',
                                         'poller_hash': 'QzgzMTA4',
                                         'serial_number': 'C83108',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:dc:f7',
                                         'uptime': '21815895'},
                                 {       'counter_source': '186477/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186477/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1586174610',
                                                                       'in_ucast_pkts': '6765755',
                                                                       'mac_address': '84:d3:43:7d:df:ff',
                                                                       'name': '186477/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '147',
                                                                       'out_octets': '23210325710',
                                                                       'out_ucast_pkts': '59501367',
                                                                       'poller_hash': 'QzgzMTY5cGVlcmludGYxODY0NzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186477',
                                                                       'in_nucast_pkts': '147',
                                                                       'in_octets': '23210325710',
                                                                       'in_ucast_pkts': '59501367',
                                                                       'links': [       {       'in_nucast_pkts': '147',
                                                                                                'in_octets': '23210325710',
                                                                                                'in_ucast_pkts': '59501367',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1586174610',
                                                                                                'out_ucast_pkts': '6765755',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1586174610',
                                                                       'out_ucast_pkts': '6765755',
                                                                       'poller_hash': 'QzgzMTY5cGVlcmludGZwb24gMTg2NDc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186477',
                                                                                               'links': [       {       'connected_time': '2139849',
                                                                                                                        'fiber_length': 12256,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTY5cGVlcnBvbnBvbiAxODY0Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6927 NICKLIN ST MILLBANK (BAUMAN '
                                                 'LAVERNE)',
                                         'poller_hash': 'QzgzMTY5',
                                         'serial_number': 'C83169',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:df:ff',
                                         'uptime': '2139849'},
                                 {       'counter_source': '186487/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186487/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '668448129871',
                                                                       'in_ucast_pkts': '61752969470',
                                                                       'mac_address': '84:d3:43:7d:dd:57',
                                                                       'name': '186487/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '70',
                                                                       'out_octets': '4358435713162',
                                                                       'out_ucast_pkts': '3666681058',
                                                                       'poller_hash': 'QzgzMTE0cGVlcmludGYxODY0ODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186487',
                                                                       'in_nucast_pkts': '70',
                                                                       'in_octets': '4358435713162',
                                                                       'in_ucast_pkts': '3666681058',
                                                                       'links': [       {       'in_nucast_pkts': '70',
                                                                                                'in_octets': '4358435713162',
                                                                                                'in_ucast_pkts': '3666681058',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '668448129871',
                                                                                                'out_ucast_pkts': '61752969470',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '668448129871',
                                                                       'out_ucast_pkts': '61752969470',
                                                                       'poller_hash': 'QzgzMTE0cGVlcmludGZwb24gMTg2NDg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186487',
                                                                                               'links': [       {       'connected_time': '688368',
                                                                                                                        'fiber_length': 12284,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTE0cGVlcnBvbnBvbiAxODY0ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6972 NICKLIN ST MILLBANK (WAGLER '
                                                 'ARLENE)',
                                         'poller_hash': 'QzgzMTE0',
                                         'serial_number': 'C83114',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:dd:57',
                                         'uptime': '688368'},
                                 {       'counter_source': '186497/F1+186497/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:30:e8'],
                                                                       'id': '186497/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2614873160',
                                                                       'in_ucast_pkts': '15523658',
                                                                       'name': '186497/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '18403094888',
                                                                       'out_octets': '24647508135332',
                                                                       'out_ucast_pkts': '2402135',
                                                                       'poller_hash': 'QzgzMEY1cGVlcmludGYxODY0OTcvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '186497/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '194147161998',
                                                                       'in_ucast_pkts': '1862704834',
                                                                       'mac_address': '84:d3:43:7d:dc:5f',
                                                                       'name': '186497/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '5835416104776',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgzMEY1cGVlcmludGYxODY0OTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186497',
                                                                       'in_nucast_pkts': '18403095043',
                                                                       'in_octets': '30482924240108',
                                                                       'in_ucast_pkts': '2402135',
                                                                       'links': [       {       'in_nucast_pkts': '18403095043',
                                                                                                'in_octets': '30482924240108',
                                                                                                'in_ucast_pkts': '2402135',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '196762035158',
                                                                                                'out_ucast_pkts': '1878228492',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '196762035158',
                                                                       'out_ucast_pkts': '1878228492',
                                                                       'poller_hash': 'QzgzMEY1cGVlcmludGZwb24gMTg2NDk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186497',
                                                                                               'links': [       {       'connected_time': '1024524',
                                                                                                                        'fiber_length': 11179,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMEY1cGVlcnBvbnBvbiAxODY0OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6963 PERTH ST MILLBANK (BRICK MARIE)',
                                         'poller_hash': 'QzgzMEY1',
                                         'serial_number': 'C830F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:dc:5f',
                                         'uptime': '1024524'},
                                 {       'counter_source': '186527/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186527/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '788419896722',
                                                                       'in_ucast_pkts': '61358028858',
                                                                       'mac_address': '84:d3:43:7d:da:87',
                                                                       'name': '186527/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '6438467397674',
                                                                       'out_ucast_pkts': '4717625290',
                                                                       'poller_hash': 'QzgzMEJBcGVlcmludGYxODY1MjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186527',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '6438467397674',
                                                                       'in_ucast_pkts': '4717625290',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '6438467397674',
                                                                                                'in_ucast_pkts': '4717625290',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '788419896722',
                                                                                                'out_ucast_pkts': '61358028858',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '788419896722',
                                                                       'out_ucast_pkts': '61358028858',
                                                                       'poller_hash': 'QzgzMEJBcGVlcmludGZwb24gMTg2NTI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186527',
                                                                                               'links': [       {       'connected_time': '4481109',
                                                                                                                        'fiber_length': 11244,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMEJBcGVlcnBvbnBvbiAxODY1Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4083 HILLCREST ST (ERB BARBARA)',
                                         'poller_hash': 'QzgzMEJB',
                                         'serial_number': 'C830BA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:da:87',
                                         'uptime': '4481109'},
                                 {       'counter_source': '186537/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186537/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '322759068524',
                                                                       'in_ucast_pkts': '40026424316',
                                                                       'mac_address': 'f8:85:f9:17:2a:20',
                                                                       'name': '186537/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '9757423755782',
                                                                       'out_ucast_pkts': '9425505450',
                                                                       'poller_hash': 'REZGRTJEcGVlcmludGYxODY1MzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186537',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '9757423755782',
                                                                       'in_ucast_pkts': '9425505450',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '9757423755782',
                                                                                                'in_ucast_pkts': '9425505450',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '322759068524',
                                                                                                'out_ucast_pkts': '40026424316',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '322759068524',
                                                                       'out_ucast_pkts': '40026424316',
                                                                       'poller_hash': 'REZGRTJEcGVlcmludGZwb24gMTg2NTM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186537',
                                                                                               'links': [       {       'connected_time': '30000762',
                                                                                                                        'fiber_length': 11303,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTJEcGVlcnBvbnBvbiAxODY1Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6969 PERTH ST (WAGLER PAUL & '
                                                 'DARLENE)',
                                         'poller_hash': 'REZGRTJE',
                                         'serial_number': 'DFFE2D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2a:20',
                                         'uptime': '30000762'},
                                 {       'counter_source': '186547/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186547/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '189395509230',
                                                                       'in_ucast_pkts': '847436561',
                                                                       'mac_address': 'b8:94:70:17:db:31',
                                                                       'name': '186547/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3874449075626',
                                                                       'out_ucast_pkts': '2990669072',
                                                                       'poller_hash': 'RUJCREIwcGVlcmludGYxODY1NDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186547',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3874449075626',
                                                                       'in_ucast_pkts': '2990669072',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3874449075626',
                                                                                                'in_ucast_pkts': '2990669072',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '189395509230',
                                                                                                'out_ucast_pkts': '847436561',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '189395509230',
                                                                       'out_ucast_pkts': '847436561',
                                                                       'poller_hash': 'RUJCREIwcGVlcmludGZwb24gMTg2NTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186547',
                                                                                               'links': [       {       'connected_time': '15498553',
                                                                                                                        'fiber_length': 9511,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.386,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RUJCREIwcGVlcnBvbnBvbiAxODY1NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5101 PERTH LINE 55 (VAN DER WAL '
                                                 'JOHN)',
                                         'poller_hash': 'RUJCREIw',
                                         'serial_number': 'EBBDB0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'b8:94:70:17:db:31',
                                         'uptime': '15498553'},
                                 {       'counter_source': '186567/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b5:41'],
                                                                       'id': '186567/g1',
                                                                       'in_nucast_pkts': '857092',
                                                                       'in_octets': '235401171611',
                                                                       'in_ucast_pkts': '1660050427',
                                                                       'mac_address': 'f8:85:f9:17:29:4b',
                                                                       'mtu': 2000,
                                                                       'name': '186567/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150689642',
                                                                       'out_octets': '9176969375762',
                                                                       'out_ucast_pkts': '6315222642',
                                                                       'poller_hash': 'REZGRTEycGVlcmludGYxODY1NjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '186567',
                                                                       'in_nucast_pkts': '150689642',
                                                                       'in_octets': '9176969375762',
                                                                       'in_ucast_pkts': '6315222642',
                                                                       'links': [       {       'in_nucast_pkts': '150689642',
                                                                                                'in_octets': '9176969375762',
                                                                                                'in_ucast_pkts': '6315222642',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '857092',
                                                                                                'out_octets': '235401171611',
                                                                                                'out_ucast_pkts': '1660050427',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '857092',
                                                                       'out_octets': '235401171611',
                                                                       'out_ucast_pkts': '1660050427',
                                                                       'poller_hash': 'REZGRTEycGVlcmludGZwb24gMTg2NTY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186567',
                                                                                               'links': [       {       'connected_time': '21816100',
                                                                                                                        'fiber_length': 11880,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTEycGVlcnBvbnBvbiAxODY1Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5518 ROAD 144 (PAXTON WAYNE & JUDY)',
                                         'poller_hash': 'REZGRTEy',
                                         'serial_number': 'DFFE12',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:29:48',
                                         'uptime': '21816100'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '186577/g1',
                                                                       'name': '186577/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzMTJDcGVlcmludGYxODY1NzcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186577',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgzMTJDcGVlcmludGZwb24gMTg2NTc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186577',
                                                                                               'links': [       {       'connected_time': '70353940',
                                                                                                                        'fiber_length': 11246,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTJDcGVlcnBvbnBvbiAxODY1Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5578 ROAD 144 (GINGRICH MAYNARD)',
                                         'poller_hash': 'QzgzMTJD',
                                         'serial_number': 'C8312C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:de:17',
                                         'uptime': '70353940'},
                                 {       'counter_source': '186587/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186587/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '376957066145',
                                                                       'in_ucast_pkts': '52160572617',
                                                                       'mac_address': '84:d3:43:7d:df:37',
                                                                       'name': '186587/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '1154481537816',
                                                                       'out_ucast_pkts': '1034587759',
                                                                       'poller_hash': 'QzgzMTUwcGVlcmludGYxODY1ODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186587',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '1154481537816',
                                                                       'in_ucast_pkts': '1034587759',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '1154481537816',
                                                                                                'in_ucast_pkts': '1034587759',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '376957066145',
                                                                                                'out_ucast_pkts': '52160572617',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '376957066145',
                                                                       'out_ucast_pkts': '52160572617',
                                                                       'poller_hash': 'QzgzMTUwcGVlcmludGZwb24gMTg2NTg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186587',
                                                                                               'links': [       {       'connected_time': '21816833',
                                                                                                                        'fiber_length': 10640,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTUwcGVlcnBvbnBvbiAxODY1ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5174 PERTH LINE 55 (BYLER JOHN)',
                                         'poller_hash': 'QzgzMTUw',
                                         'serial_number': 'C83150',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:df:37',
                                         'uptime': '21816833'},
                                 {       'counter_source': '186597/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186597/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '435483833366',
                                                                       'in_ucast_pkts': '1931366134',
                                                                       'mac_address': '84:d3:43:7d:db:7f',
                                                                       'name': '186597/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '7189623658597',
                                                                       'out_ucast_pkts': '1101550499',
                                                                       'poller_hash': 'QzgzMEQ5cGVlcmludGYxODY1OTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186597',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '7189623658597',
                                                                       'in_ucast_pkts': '1101550499',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '7189623658597',
                                                                                                'in_ucast_pkts': '1101550499',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '435483833366',
                                                                                                'out_ucast_pkts': '1931366134',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '435483833366',
                                                                       'out_ucast_pkts': '1931366134',
                                                                       'poller_hash': 'QzgzMEQ5cGVlcmludGZwb24gMTg2NTk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186597',
                                                                                               'links': [       {       'connected_time': '12508517',
                                                                                                                        'fiber_length': 12067,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMEQ5cGVlcnBvbnBvbiAxODY1OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5323 PERTH LINE 55 (MUELLER GEORGE)',
                                         'poller_hash': 'QzgzMEQ5',
                                         'serial_number': 'C830D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:db:7f',
                                         'uptime': '12508517'},
                                 {       'counter_source': '186607/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186607/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '371241092760',
                                                                       'in_ucast_pkts': '100735757227',
                                                                       'mac_address': 'f8:85:f9:17:29:f0',
                                                                       'name': '186607/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '74',
                                                                       'out_octets': '14243865456975',
                                                                       'out_ucast_pkts': '9773719799',
                                                                       'poller_hash': 'REZGRTI3cGVlcmludGYxODY2MDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186607',
                                                                       'in_nucast_pkts': '74',
                                                                       'in_octets': '14243865456975',
                                                                       'in_ucast_pkts': '9773719799',
                                                                       'links': [       {       'in_nucast_pkts': '74',
                                                                                                'in_octets': '14243865456975',
                                                                                                'in_ucast_pkts': '9773719799',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '371241092760',
                                                                                                'out_ucast_pkts': '100735757227',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '371241092760',
                                                                       'out_ucast_pkts': '100735757227',
                                                                       'poller_hash': 'REZGRTI3cGVlcmludGZwb24gMTg2NjA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186607',
                                                                                               'links': [       {       'connected_time': '688351',
                                                                                                                        'fiber_length': 11602,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTI3cGVlcnBvbnBvbiAxODY2MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6973 CHURCH ST (HANDORF BEN)',
                                         'poller_hash': 'REZGRTI3',
                                         'serial_number': 'DFFE27',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:29:f0',
                                         'uptime': '688351'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '186657/g1',
                                                                       'name': '186657/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'REZGREI5cGVlcmludGYxODY2NTcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186657',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'REZGREI5cGVlcmludGZwb24gMTg2NjU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186657',
                                                                                               'links': [       {       'connected_time': '21816255',
                                                                                                                        'fiber_length': 11837,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREI5cGVlcnBvbnBvbiAxODY2NTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4088 PERTH LINE 72 (MORNINGSTAR '
                                                 'CHRISTIAN SCHOOL)',
                                         'poller_hash': 'REZGREI5',
                                         'serial_number': 'DFFDB9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:26:80',
                                         'uptime': '21816255'},
                                 {       'counter_source': '186667/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186667/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '167868527313',
                                                                       'in_ucast_pkts': '505218631',
                                                                       'mac_address': '84:d3:43:7d:de:af',
                                                                       'name': '186667/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '207',
                                                                       'out_octets': '2747955116387',
                                                                       'out_ucast_pkts': '2127899254',
                                                                       'poller_hash': 'QzgzMTNGcGVlcmludGYxODY2NjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186667',
                                                                       'in_nucast_pkts': '207',
                                                                       'in_octets': '2747955116387',
                                                                       'in_ucast_pkts': '2127899254',
                                                                       'links': [       {       'in_nucast_pkts': '207',
                                                                                                'in_octets': '2747955116387',
                                                                                                'in_ucast_pkts': '2127899254',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '167868527313',
                                                                                                'out_ucast_pkts': '505218631',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '167868527313',
                                                                       'out_ucast_pkts': '505218631',
                                                                       'poller_hash': 'QzgzMTNGcGVlcmludGZwb24gMTg2NjY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186667',
                                                                                               'links': [       {       'connected_time': '1912560',
                                                                                                                        'fiber_length': 11503,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTNGcGVlcnBvbnBvbiAxODY2Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6998 MILLBANK MAIN ST (JANZEN '
                                                 'GERHARD)',
                                         'poller_hash': 'QzgzMTNG',
                                         'serial_number': 'C8313F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:de:af',
                                         'uptime': '1912560'},
                                 {       'counter_source': '186687/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186687/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '311552759272',
                                                                       'in_ucast_pkts': '1688823049',
                                                                       'mac_address': '84:d3:43:7e:30:9f',
                                                                       'name': '186687/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '9426640064411',
                                                                       'out_ucast_pkts': '6515139743',
                                                                       'poller_hash': 'QzgzQjdEcGVlcmludGYxODY2ODcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186687',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '9426640064411',
                                                                       'in_ucast_pkts': '6515139743',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '9426640064411',
                                                                                                'in_ucast_pkts': '6515139743',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '311552759272',
                                                                                                'out_ucast_pkts': '1688823049',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '311552759272',
                                                                       'out_ucast_pkts': '1688823049',
                                                                       'poller_hash': 'QzgzQjdEcGVlcmludGZwb24gMTg2Njg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186687',
                                                                                               'links': [       {       'connected_time': '688486',
                                                                                                                        'fiber_length': 732,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjdEcGVlcnBvbnBvbiAxODY2ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '38 PACIFIC AVE MILVERTON (VAN '
                                                 'ROOSMALEN KRISTIE)',
                                         'poller_hash': 'QzgzQjdE',
                                         'serial_number': 'C83B7D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:30:9f',
                                         'uptime': '688486'},
                                 {       'counter_source': '187469/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2f:cc:0f',
                                                                                                        '5c:0f:fb:2f:ce:bb',
                                                                                                        '84:d3:43:6d:ea:40'],
                                                                       'id': '187469/g1',
                                                                       'in_nucast_pkts': '238398',
                                                                       'in_octets': '3424756202',
                                                                       'in_ucast_pkts': '10235369',
                                                                       'mac_address': '84:d3:43:7d:e1:52',
                                                                       'mtu': 2000,
                                                                       'name': '187469/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '287614777',
                                                                       'out_octets': '427601681766',
                                                                       'out_ucast_pkts': '34395194',
                                                                       'poller_hash': 'QzgzMTkzcGVlcmludGYxODc0NjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '187469',
                                                                       'in_nucast_pkts': '287614777',
                                                                       'in_octets': '427601681766',
                                                                       'in_ucast_pkts': '34395194',
                                                                       'links': [       {       'in_nucast_pkts': '287614777',
                                                                                                'in_octets': '427601681766',
                                                                                                'in_ucast_pkts': '34395194',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '238398',
                                                                                                'out_octets': '3424756202',
                                                                                                'out_ucast_pkts': '10235369',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '238398',
                                                                       'out_octets': '3424756202',
                                                                       'out_ucast_pkts': '10235369',
                                                                       'poller_hash': 'QzgzMTkzcGVlcmludGZwb24gMTg3NDY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '187469',
                                                                                               'links': [       {       'connected_time': '523244',
                                                                                                                        'fiber_length': 11584,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTkzcGVlcnBvbnBvbiAxODc0Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6998 FREEBORN ST (RICHARDS BILL)',
                                         'poller_hash': 'QzgzMTkz',
                                         'serial_number': 'C83193',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e1:4f',
                                         'uptime': '523244'},
                                 {       'counter_source': '189767/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '189767/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '94709421877',
                                                                       'in_ucast_pkts': '307832904',
                                                                       'mac_address': 'f8:85:f9:17:27:b8',
                                                                       'name': '189767/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '1182748753006',
                                                                       'out_ucast_pkts': '907873985',
                                                                       'poller_hash': 'REZGREUwcGVlcmludGYxODk3NjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '189767',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '1182748753006',
                                                                       'in_ucast_pkts': '907873985',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '1182748753006',
                                                                                                'in_ucast_pkts': '907873985',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '94709421877',
                                                                                                'out_ucast_pkts': '307832904',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '94709421877',
                                                                       'out_ucast_pkts': '307832904',
                                                                       'poller_hash': 'REZGREUwcGVlcmludGZwb24gMTg5NzY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '189767',
                                                                                               'links': [       {       'connected_time': '523138',
                                                                                                                        'fiber_length': 11235,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGREUwcGVlcnBvbnBvbiAxODk3Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4094 HILLCREST ST MILLBANK '
                                                 '(LINGELBACH BRITTANY)',
                                         'poller_hash': 'REZGREUw',
                                         'serial_number': 'DFFDE0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:27:b8',
                                         'uptime': '523138'},
                                 {       'counter_source': '189807/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3a:0d'],
                                                                       'id': '189807/g1',
                                                                       'in_nucast_pkts': '6509970',
                                                                       'in_octets': '334390036668',
                                                                       'in_ucast_pkts': '2007918184',
                                                                       'mac_address': '84:d3:43:7e:32:fa',
                                                                       'mtu': 2000,
                                                                       'name': '189807/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8551117092',
                                                                       'out_octets': '18670267170392',
                                                                       'out_ucast_pkts': '5226717142',
                                                                       'poller_hash': 'QzgzQkM4cGVlcmludGYxODk4MDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '189807',
                                                                       'in_nucast_pkts': '8551117092',
                                                                       'in_octets': '18670267170392',
                                                                       'in_ucast_pkts': '5226717142',
                                                                       'links': [       {       'in_nucast_pkts': '8551117092',
                                                                                                'in_octets': '18670267170392',
                                                                                                'in_ucast_pkts': '5226717142',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6509970',
                                                                                                'out_octets': '334390036668',
                                                                                                'out_ucast_pkts': '2007918184',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6509970',
                                                                       'out_octets': '334390036668',
                                                                       'out_ucast_pkts': '2007918184',
                                                                       'poller_hash': 'QzgzQkM4cGVlcmludGZwb24gMTg5ODA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '189807',
                                                                                               'links': [       {       'connected_time': '69798267',
                                                                                                                        'fiber_length': 12028,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkM4cGVlcnBvbnBvbiAxODk4MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4059 NORFOLK ST (POOLE BRIAN)',
                                         'poller_hash': 'QzgzQkM4',
                                         'serial_number': 'C83BC8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:f7',
                                         'uptime': '69798267'},
                                 {       'counter_source': '189828/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '189828/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '104414989899',
                                                                       'in_ucast_pkts': '565865925',
                                                                       'mac_address': 'ec:4f:82:0d:6b:02',
                                                                       'name': '189828/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '125',
                                                                       'out_octets': '1573886685707',
                                                                       'out_ucast_pkts': '1229730998',
                                                                       'poller_hash': 'MjA1RTA1cGVlcmludGYxODk4MjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '189828',
                                                                       'in_nucast_pkts': '125',
                                                                       'in_octets': '1573886685707',
                                                                       'in_ucast_pkts': '1229730998',
                                                                       'links': [       {       'in_nucast_pkts': '125',
                                                                                                'in_octets': '1573886685707',
                                                                                                'in_ucast_pkts': '1229730998',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '104414989899',
                                                                                                'out_ucast_pkts': '565865925',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '104414989899',
                                                                       'out_ucast_pkts': '565865925',
                                                                       'poller_hash': 'MjA1RTA1cGVlcmludGZwb24gMTg5ODI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '189828',
                                                                                               'links': [       {       'connected_time': '26374670',
                                                                                                                        'fiber_length': 1728,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RTA1cGVlcnBvbnBvbiAxODk4Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '62 MAIN STREET N (FOBEL RICHARD)',
                                         'poller_hash': 'MjA1RTA1',
                                         'serial_number': '205E05',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6b:02',
                                         'uptime': '26374670'},
                                 {       'counter_source': '189839/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '189839/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '924243346719',
                                                                       'in_ucast_pkts': '117853600048',
                                                                       'mac_address': '84:d3:43:7e:32:e7',
                                                                       'name': '189839/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '62',
                                                                       'out_octets': '5005860080866',
                                                                       'out_ucast_pkts': '3895680490',
                                                                       'poller_hash': 'QzgzQkM2cGVlcmludGYxODk4MzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '189839',
                                                                       'in_nucast_pkts': '62',
                                                                       'in_octets': '5005860080866',
                                                                       'in_ucast_pkts': '3895680490',
                                                                       'links': [       {       'in_nucast_pkts': '62',
                                                                                                'in_octets': '5005860080866',
                                                                                                'in_ucast_pkts': '3895680490',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '924243346719',
                                                                                                'out_ucast_pkts': '117853600048',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '924243346719',
                                                                       'out_ucast_pkts': '117853600048',
                                                                       'poller_hash': 'QzgzQkM2cGVlcmludGZwb24gMTg5ODM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '189839',
                                                                                               'links': [       {       'connected_time': '688400',
                                                                                                                        'fiber_length': 11286,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkM2cGVlcnBvbnBvbiAxODk4Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6933 PERTH ROAD 121 (STREICHER ADAM)',
                                         'poller_hash': 'QzgzQkM2',
                                         'serial_number': 'C83BC6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:e7',
                                         'uptime': '688400'},
                                 {       'counter_source': '189879/F1+189879/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:17:6e:84'],
                                                                       'id': '189879/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2969215883',
                                                                       'in_ucast_pkts': '18702524',
                                                                       'name': '189879/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '3907252614',
                                                                       'out_octets': '5196259845165',
                                                                       'out_ucast_pkts': '702260',
                                                                       'poller_hash': 'QzgxREYzcGVlcmludGYxODk4NzkvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3c:49'],
                                                                       'id': '189879/g1',
                                                                       'in_nucast_pkts': '57019',
                                                                       'in_octets': '38805108635',
                                                                       'in_ucast_pkts': '125563729',
                                                                       'mac_address': '84:d3:43:7d:44:52',
                                                                       'mtu': 2000,
                                                                       'name': '189879/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6685973',
                                                                       'out_octets': '427665245743',
                                                                       'out_ucast_pkts': '324120363',
                                                                       'poller_hash': 'QzgxREYzcGVlcmludGYxODk4NzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '189879',
                                                                       'in_nucast_pkts': '3913938587',
                                                                       'in_octets': '5623925090908',
                                                                       'in_ucast_pkts': '324822623',
                                                                       'links': [       {       'in_nucast_pkts': '3913938587',
                                                                                                'in_octets': '5623925090908',
                                                                                                'in_ucast_pkts': '324822623',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '57019',
                                                                                                'out_octets': '41774324518',
                                                                                                'out_ucast_pkts': '144266253',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '57019',
                                                                       'out_octets': '41774324518',
                                                                       'out_ucast_pkts': '144266253',
                                                                       'poller_hash': 'QzgxREYzcGVlcmludGZwb24gMTg5ODc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '189879',
                                                                                               'links': [       {       'connected_time': '3113868',
                                                                                                                        'fiber_length': 12133,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREYzcGVlcnBvbnBvbiAxODk4Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6540 ROAD 154 (SATCHELL SCOTT)',
                                         'poller_hash': 'QzgxREYz',
                                         'serial_number': 'C81DF3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:4f',
                                         'uptime': '3113868'},
                                 {       'counter_source': '192865/F1+192865/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a8:0d:a6',
                                                                                                        '5c:0f:fb:17:64:16',
                                                                                                        '5c:0f:fb:17:72:10'],
                                                                       'id': '192865/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4433810217',
                                                                       'in_ucast_pkts': '28538823',
                                                                       'name': '192865/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '40638813085',
                                                                       'out_octets': '54991792954870',
                                                                       'out_ucast_pkts': '4676487',
                                                                       'poller_hash': 'QzgxREE1cGVlcmludGYxOTI4NjUvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '192865/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '403160099966',
                                                                       'in_ucast_pkts': '78847152686',
                                                                       'mac_address': '84:d3:43:7d:41:df',
                                                                       'name': '192865/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '172',
                                                                       'out_octets': '5009822072768',
                                                                       'out_ucast_pkts': '4033790255',
                                                                       'poller_hash': 'QzgxREE1cGVlcmludGYxOTI4NjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '192865',
                                                                       'in_nucast_pkts': '40638813257',
                                                                       'in_octets': '60001615027638',
                                                                       'in_ucast_pkts': '4038466742',
                                                                       'links': [       {       'in_nucast_pkts': '40638813257',
                                                                                                'in_octets': '60001615027638',
                                                                                                'in_ucast_pkts': '4038466742',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '407593910183',
                                                                                                'out_ucast_pkts': '78875691509',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '407593910183',
                                                                       'out_ucast_pkts': '78875691509',
                                                                       'poller_hash': 'QzgxREE1cGVlcmludGZwb24gMTkyODY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '192865',
                                                                                               'links': [       {       'connected_time': '1026215',
                                                                                                                        'fiber_length': 11451,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREE1cGVlcnBvbnBvbiAxOTI4NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6983 CHURCH ST (KUEPFER WILLIAM R)',
                                         'poller_hash': 'QzgxREE1',
                                         'serial_number': 'C81DA5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:41:df',
                                         'uptime': '1026215'},
                                 {       'counter_source': '193318/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '193318/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '152109406897',
                                                                       'in_ucast_pkts': '1263685400',
                                                                       'mac_address': '60:db:98:4e:9a:7c',
                                                                       'name': '193318/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '6164660759485',
                                                                       'out_ucast_pkts': '4448146166',
                                                                       'poller_hash': 'QTlDOTBGcGVlcmludGYxOTMzMTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '193318',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '6164660759485',
                                                                       'in_ucast_pkts': '4448146166',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '6164660759485',
                                                                                                'in_ucast_pkts': '4448146166',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '152109406897',
                                                                                                'out_ucast_pkts': '1263685400',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '152109406897',
                                                                       'out_ucast_pkts': '1263685400',
                                                                       'poller_hash': 'QTlDOTBGcGVlcmludGZwb24gMTkzMzE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193318',
                                                                                               'links': [       {       'connected_time': '21817023',
                                                                                                                        'fiber_length': 9662,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOTBGcGVlcnBvbnBvbiAxOTMzMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4924 PERTH LINE 55 (WESTMAN STEPHEN '
                                                 '& LINDA)',
                                         'poller_hash': 'QTlDOTBG',
                                         'serial_number': 'A9C90F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:9a:7c',
                                         'uptime': '21817023'},
                                 {       'counter_source': '193328/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8a:72'],
                                                                       'id': '193328/g1',
                                                                       'in_nucast_pkts': '403514',
                                                                       'in_octets': '156093661106',
                                                                       'in_ucast_pkts': '1700312524',
                                                                       'mac_address': '84:d3:43:7d:db:2a',
                                                                       'mtu': 2000,
                                                                       'name': '193328/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68247885',
                                                                       'out_octets': '8924268409935',
                                                                       'out_ucast_pkts': '5971284958',
                                                                       'poller_hash': 'QzgzMENFcGVlcmludGYxOTMzMjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '193328',
                                                                       'in_nucast_pkts': '68247885',
                                                                       'in_octets': '8924268409935',
                                                                       'in_ucast_pkts': '5971284958',
                                                                       'links': [       {       'in_nucast_pkts': '68247885',
                                                                                                'in_octets': '8924268409935',
                                                                                                'in_ucast_pkts': '5971284958',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '403514',
                                                                                                'out_octets': '156093661106',
                                                                                                'out_ucast_pkts': '1700312524',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '403514',
                                                                       'out_octets': '156093661106',
                                                                       'out_ucast_pkts': '1700312524',
                                                                       'poller_hash': 'QzgzMENFcGVlcmludGZwb24gMTkzMzI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193328',
                                                                                               'links': [       {       'connected_time': '21815769',
                                                                                                                        'fiber_length': 11293,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMENFcGVlcnBvbnBvbiAxOTMzMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4084 WYNDHAM ST (ROPP DAVID & JANET)',
                                         'poller_hash': 'QzgzMENF',
                                         'serial_number': 'C830CE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:db:27',
                                         'uptime': '21815769'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '193383/g1',
                                                                       'name': '193383/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzMThEcGVlcmludGYxOTMzODMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '193383',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgzMThEcGVlcmludGZwb24gMTkzMzgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193383',
                                                                                               'links': [       {       'connected_time': '21815956',
                                                                                                                        'fiber_length': 1779,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMThEcGVlcnBvbnBvbiAxOTMzODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '110 MILL ST E (WAGLER ALVIN)',
                                         'poller_hash': 'QzgzMThE',
                                         'serial_number': 'C8318D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e1:1f',
                                         'uptime': '21815956'},
                                 {       'counter_source': '193572/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '193572/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '213459506245',
                                                                       'in_ucast_pkts': '21227061204',
                                                                       'mac_address': '84:d3:43:7e:30:4f',
                                                                       'name': '193572/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '107',
                                                                       'out_octets': '4383311779173',
                                                                       'out_ucast_pkts': '8160923168',
                                                                       'poller_hash': 'QzgzQjczcGVlcmludGYxOTM1NzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '193572',
                                                                       'in_nucast_pkts': '107',
                                                                       'in_octets': '4383311779173',
                                                                       'in_ucast_pkts': '8160923168',
                                                                       'links': [       {       'in_nucast_pkts': '107',
                                                                                                'in_octets': '4383311779173',
                                                                                                'in_ucast_pkts': '8160923168',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '213459506245',
                                                                                                'out_ucast_pkts': '21227061204',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '213459506245',
                                                                       'out_ucast_pkts': '21227061204',
                                                                       'poller_hash': 'QzgzQjczcGVlcmludGZwb24gMTkzNTcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193572',
                                                                                               'links': [       {       'connected_time': '6413854',
                                                                                                                        'fiber_length': 1184,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjczcGVlcnBvbnBvbiAxOTM1NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '9 PACIFIC AV (SCHAEFER ROBERT)',
                                         'poller_hash': 'QzgzQjcz',
                                         'serial_number': 'C83B73',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:30:4f',
                                         'uptime': '6413854'},
                                 {       'counter_source': '193582/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '193582/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '235587257834',
                                                                       'in_ucast_pkts': '1017499773',
                                                                       'mac_address': '84:d3:43:7e:2e:4f',
                                                                       'name': '193582/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '4627585027429',
                                                                       'out_ucast_pkts': '3593548408',
                                                                       'poller_hash': 'QzgzQjMzcGVlcmludGYxOTM1ODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '193582',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '4627585027429',
                                                                       'in_ucast_pkts': '3593548408',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '4627585027429',
                                                                                                'in_ucast_pkts': '3593548408',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '235587257834',
                                                                                                'out_ucast_pkts': '1017499773',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '235587257834',
                                                                       'out_ucast_pkts': '1017499773',
                                                                       'poller_hash': 'QzgzQjMzcGVlcmludGZwb24gMTkzNTgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193582',
                                                                                               'links': [       {       'connected_time': '5885243',
                                                                                                                        'fiber_length': 2526,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjMzcGVlcnBvbnBvbiAxOTM1ODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 COBALT STREET S (LEHMANN TRAVIS)',
                                         'poller_hash': 'QzgzQjMz',
                                         'serial_number': 'C83B33',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:2e:4f',
                                         'uptime': '5885243'},
                                 {       'counter_source': '193923/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '193923/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2683174784',
                                                                       'in_ucast_pkts': '9508520',
                                                                       'mac_address': '84:d3:43:7d:48:3f',
                                                                       'name': '193923/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '36166153281',
                                                                       'out_ucast_pkts': '41298343',
                                                                       'poller_hash': 'QzgxRTcxcGVlcmludGYxOTM5MjMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '193923',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '36166153281',
                                                                       'in_ucast_pkts': '41298343',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '36166153281',
                                                                                                'in_ucast_pkts': '41298343',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2683174784',
                                                                                                'out_ucast_pkts': '9508520',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2683174784',
                                                                       'out_ucast_pkts': '9508520',
                                                                       'poller_hash': 'QzgxRTcxcGVlcmludGZwb24gMTkzOTIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193923',
                                                                                               'links': [       {       'connected_time': '15037897',
                                                                                                                        'fiber_length': 6516,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTcxcGVlcnBvbnBvbiAxOTM5MjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4461 PERTH LINE 72 (TAKACH DON)',
                                         'poller_hash': 'QzgxRTcx',
                                         'serial_number': 'C81E71',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:48:3f',
                                         'uptime': '15037897'},
                                 {       'counter_source': '194106/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:4c:a5:f1:71:7a'],
                                                                       'id': '194106/g1',
                                                                       'in_nucast_pkts': '21409',
                                                                       'in_octets': '390089658784',
                                                                       'in_ucast_pkts': '1036808211',
                                                                       'mac_address': '00:06:31:f6:55:ab',
                                                                       'mtu': 2000,
                                                                       'name': '194106/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27344180',
                                                                       'out_octets': '2096376198503',
                                                                       'out_ucast_pkts': '2066235723',
                                                                       'poller_hash': 'MjYwRTk2cGVlcmludGYxOTQxMDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '194106',
                                                                       'in_nucast_pkts': '27344180',
                                                                       'in_octets': '2096376198503',
                                                                       'in_ucast_pkts': '2066235723',
                                                                       'links': [       {       'in_nucast_pkts': '27344180',
                                                                                                'in_octets': '2096376198503',
                                                                                                'in_ucast_pkts': '2066235723',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21409',
                                                                                                'out_octets': '390089658784',
                                                                                                'out_ucast_pkts': '1036808211',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21409',
                                                                       'out_octets': '390089658784',
                                                                       'out_ucast_pkts': '1036808211',
                                                                       'poller_hash': 'MjYwRTk2cGVlcmludGZwb24gMTk0MTA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '194106',
                                                                                               'links': [       {       'connected_time': '15498365',
                                                                                                                        'fiber_length': 946,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjYwRTk2cGVlcnBvbnBvbiAxOTQxMDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '1 Main St. S. Milverton (Zehr '
                                                 'Insurance)',
                                         'poller_hash': 'MjYwRTk2',
                                         'serial_number': '260E96',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:f6:55:b0',
                                         'uptime': '15498365'},
                                 {       'counter_source': '194370/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ed:0b'],
                                                                       'id': '194370/g1',
                                                                       'in_nucast_pkts': '14208',
                                                                       'in_octets': '5101089086',
                                                                       'in_ucast_pkts': '9725901',
                                                                       'mac_address': '84:d3:43:7e:33:7a',
                                                                       'mtu': 2000,
                                                                       'name': '194370/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464127',
                                                                       'out_octets': '20701379698',
                                                                       'out_ucast_pkts': '18402248',
                                                                       'poller_hash': 'QzgzQkQ4cGVlcmludGYxOTQzNzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '194370',
                                                                       'in_nucast_pkts': '1464127',
                                                                       'in_octets': '20701379698',
                                                                       'in_ucast_pkts': '18402248',
                                                                       'links': [       {       'in_nucast_pkts': '1464127',
                                                                                                'in_octets': '20701379698',
                                                                                                'in_ucast_pkts': '18402248',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14208',
                                                                                                'out_octets': '5101089086',
                                                                                                'out_ucast_pkts': '9725901',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14208',
                                                                       'out_octets': '5101089086',
                                                                       'out_ucast_pkts': '9725901',
                                                                       'poller_hash': 'QzgzQkQ4cGVlcmludGZwb24gMTk0Mzcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '194370',
                                                                                               'links': [       {       'connected_time': '688421',
                                                                                                                        'fiber_length': 1042,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkQ4cGVlcnBvbnBvbiAxOTQzNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 ST MARYS ST S (FRIESEN PETER)',
                                         'poller_hash': 'QzgzQkQ4',
                                         'serial_number': 'C83BD8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:33:77',
                                         'uptime': '688421'},
                                 {       'counter_source': '196085/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1e:8d'],
                                                                       'id': '196085/g1',
                                                                       'in_nucast_pkts': '11973',
                                                                       'in_octets': '4430497054',
                                                                       'in_ucast_pkts': '10505042',
                                                                       'mac_address': '84:d3:43:7d:22:7a',
                                                                       'mtu': 2000,
                                                                       'name': '196085/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464319',
                                                                       'out_octets': '142284415759',
                                                                       'out_ucast_pkts': '106930872',
                                                                       'poller_hash': 'QzgxOUI4cGVlcmludGYxOTYwODUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '196085',
                                                                       'in_nucast_pkts': '1464319',
                                                                       'in_octets': '142284415759',
                                                                       'in_ucast_pkts': '106930872',
                                                                       'links': [       {       'in_nucast_pkts': '1464319',
                                                                                                'in_octets': '142284415759',
                                                                                                'in_ucast_pkts': '106930872',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11973',
                                                                                                'out_octets': '4430497054',
                                                                                                'out_ucast_pkts': '10505042',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11973',
                                                                       'out_octets': '4430497054',
                                                                       'out_ucast_pkts': '10505042',
                                                                       'poller_hash': 'QzgxOUI4cGVlcmludGZwb24gMTk2MDg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '196085',
                                                                                               'links': [       {       'connected_time': '688361',
                                                                                                                        'fiber_length': 3088,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxOUI4cGVlcnBvbnBvbiAxOTYwODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '132 PUGH ST (GRAHAM JENNIFER)',
                                         'poller_hash': 'QzgxOUI4',
                                         'serial_number': 'C819B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:22:77',
                                         'uptime': '688361'},
                                 {       'counter_source': '196244/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '196244/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2581066627733',
                                                                       'in_ucast_pkts': '5676346867',
                                                                       'mac_address': '84:d3:43:7d:dd:67',
                                                                       'name': '196244/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '17420649061494',
                                                                       'out_ucast_pkts': '12915323556',
                                                                       'poller_hash': 'QzgzMTE2cGVlcmludGYxOTYyNDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '196244',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '17420649061494',
                                                                       'in_ucast_pkts': '12915323556',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '17420649061494',
                                                                                                'in_ucast_pkts': '12915323556',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2581066627733',
                                                                                                'out_ucast_pkts': '5676346867',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2581066627733',
                                                                       'out_ucast_pkts': '5676346867',
                                                                       'poller_hash': 'QzgzMTE2cGVlcmludGZwb24gMTk2MjQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '196244',
                                                                                               'links': [       {       'connected_time': '688360',
                                                                                                                        'fiber_length': 2839,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTE2cGVlcnBvbnBvbiAxOTYyNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '155 PUGH ST (ZIERL MATTHEW)',
                                         'poller_hash': 'QzgzMTE2',
                                         'serial_number': 'C83116',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:dd:67',
                                         'uptime': '688360'},
                                 {       'counter_source': '196905/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '196905/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '86417134426',
                                                                       'in_ucast_pkts': '21921868883',
                                                                       'mac_address': '84:d3:43:7d:41:07',
                                                                       'name': '196905/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '2016692764010',
                                                                       'out_ucast_pkts': '1496282246',
                                                                       'poller_hash': 'QzgxRDhBcGVlcmludGYxOTY5MDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '196905',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '2016692764010',
                                                                       'in_ucast_pkts': '1496282246',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '2016692764010',
                                                                                                'in_ucast_pkts': '1496282246',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '86417134426',
                                                                                                'out_ucast_pkts': '21921868883',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '86417134426',
                                                                       'out_ucast_pkts': '21921868883',
                                                                       'poller_hash': 'QzgxRDhBcGVlcmludGZwb24gMTk2OTA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '196905',
                                                                                               'links': [       {       'connected_time': '688348',
                                                                                                                        'fiber_length': 11613,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDhBcGVlcnBvbnBvbiAxOTY5MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6987 MILLBANK MAIN ST (FRIESEN JAKE)',
                                         'poller_hash': 'QzgxRDhB',
                                         'serial_number': 'C81D8A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:41:07',
                                         'uptime': '688348'},
                                 {       'counter_source': '197797/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '197797/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '583799779563',
                                                                       'in_ucast_pkts': '6890372824',
                                                                       'mac_address': '84:d3:43:7e:35:9f',
                                                                       'name': '197797/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '20595726171093',
                                                                       'out_ucast_pkts': '13733772363',
                                                                       'poller_hash': 'QzgzQzFEcGVlcmludGYxOTc3OTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '197797',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '20595726171093',
                                                                       'in_ucast_pkts': '13733772363',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '20595726171093',
                                                                                                'in_ucast_pkts': '13733772363',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '583799779563',
                                                                                                'out_ucast_pkts': '6890372824',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '583799779563',
                                                                       'out_ucast_pkts': '6890372824',
                                                                       'poller_hash': 'QzgzQzFEcGVlcmludGZwb24gMTk3Nzk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '197797',
                                                                                               'links': [       {       'connected_time': '21816163',
                                                                                                                        'fiber_length': 18079,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQzFEcGVlcnBvbnBvbiAxOTc3OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '77 MADDISON ST W (MCEWEN COLLEEN)',
                                         'poller_hash': 'QzgzQzFE',
                                         'serial_number': 'C83C1D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:35:9f',
                                         'uptime': '21816163'},
                                 {       'counter_source': '197950/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b5:cc'],
                                                                       'id': '197950/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '6838428539',
                                                                       'in_ucast_pkts': '26027592',
                                                                       'mac_address': '84:d3:43:7d:4f:4a',
                                                                       'mtu': 2000,
                                                                       'name': '197950/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464147',
                                                                       'out_octets': '94367802746',
                                                                       'out_ucast_pkts': '67814693',
                                                                       'poller_hash': 'QzgxRjUycGVlcmludGYxOTc5NTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '197950',
                                                                       'in_nucast_pkts': '1464147',
                                                                       'in_octets': '94367802746',
                                                                       'in_ucast_pkts': '67814693',
                                                                       'links': [       {       'in_nucast_pkts': '1464147',
                                                                                                'in_octets': '94367802746',
                                                                                                'in_ucast_pkts': '67814693',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '6838428539',
                                                                                                'out_ucast_pkts': '26027592',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '6838428539',
                                                                       'out_ucast_pkts': '26027592',
                                                                       'poller_hash': 'QzgxRjUycGVlcmludGZwb24gMTk3OTUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '197950',
                                                                                               'links': [       {       'connected_time': '688433',
                                                                                                                        'fiber_length': 1772,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRjUycGVlcnBvbnBvbiAxOTc5NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '102 MILL ST E MILVERTON (JANTZI '
                                                 'STEPHEN)',
                                         'poller_hash': 'QzgxRjUy',
                                         'serial_number': 'C81F52',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4f:47',
                                         'uptime': '688433'},
                                 {       'counter_source': '198234/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '198234/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '154521852554',
                                                                       'in_ucast_pkts': '223167394',
                                                                       'mac_address': '84:d3:43:7e:00:e7',
                                                                       'name': '198234/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1182343710051',
                                                                       'out_ucast_pkts': '894853727',
                                                                       'poller_hash': 'QzgzNTg2cGVlcmludGYxOTgyMzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '198234',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1182343710051',
                                                                       'in_ucast_pkts': '894853727',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1182343710051',
                                                                                                'in_ucast_pkts': '894853727',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '154521852554',
                                                                                                'out_ucast_pkts': '223167394',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '154521852554',
                                                                       'out_ucast_pkts': '223167394',
                                                                       'poller_hash': 'QzgzNTg2cGVlcmludGZwb24gMTk4MjM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198234',
                                                                                               'links': [       {       'connected_time': '523090',
                                                                                                                        'fiber_length': 11461,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzNTg2cGVlcnBvbnBvbiAxOTgyMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4060 PRINCESS ST (SCHMITT GEOFF)',
                                         'poller_hash': 'QzgzNTg2',
                                         'serial_number': 'C83586',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:00:e7',
                                         'uptime': '523090'},
                                 {       'counter_source': '198244/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '198244/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '332008495016',
                                                                       'in_ucast_pkts': '628618286',
                                                                       'mac_address': '84:d3:43:7d:e5:8f',
                                                                       'name': '198244/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '1798814245244',
                                                                       'out_ucast_pkts': '1639611895',
                                                                       'poller_hash': 'QzgzMjFCcGVlcmludGYxOTgyNDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '198244',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '1798814245244',
                                                                       'in_ucast_pkts': '1639611895',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '1798814245244',
                                                                                                'in_ucast_pkts': '1639611895',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '332008495016',
                                                                                                'out_ucast_pkts': '628618286',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '332008495016',
                                                                       'out_ucast_pkts': '628618286',
                                                                       'poller_hash': 'QzgzMjFCcGVlcmludGZwb24gMTk4MjQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198244',
                                                                                               'links': [       {       'connected_time': '14513940',
                                                                                                                        'fiber_length': 11426,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMjFCcGVlcnBvbnBvbiAxOTgyNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4087 QUEEN ST (MARTIN MERLIN)',
                                         'poller_hash': 'QzgzMjFC',
                                         'serial_number': 'C8321B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e5:8f',
                                         'uptime': '14513940'},
                                 {       'counter_source': '198254/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '198254/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '252664499640',
                                                                       'in_ucast_pkts': '69637711361',
                                                                       'mac_address': '84:d3:43:7d:4a:ff',
                                                                       'name': '198254/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '147',
                                                                       'out_octets': '4297810388964',
                                                                       'out_ucast_pkts': '3258849809',
                                                                       'poller_hash': 'QzgxRUM5cGVlcmludGYxOTgyNTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '198254',
                                                                       'in_nucast_pkts': '147',
                                                                       'in_octets': '4297810388964',
                                                                       'in_ucast_pkts': '3258849809',
                                                                       'links': [       {       'in_nucast_pkts': '147',
                                                                                                'in_octets': '4297810388964',
                                                                                                'in_ucast_pkts': '3258849809',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '252664499640',
                                                                                                'out_ucast_pkts': '69637711361',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '252664499640',
                                                                       'out_ucast_pkts': '69637711361',
                                                                       'poller_hash': 'QzgxRUM5cGVlcmludGZwb24gMTk4MjU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198254',
                                                                                               'links': [       {       'connected_time': '21816523',
                                                                                                                        'fiber_length': 11407,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUM5cGVlcnBvbnBvbiAxOTgyNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6948 ELGIN ST (PETERS ISAAK & '
                                                 'SUSANA)',
                                         'poller_hash': 'QzgxRUM5',
                                         'serial_number': 'C81EC9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:ff',
                                         'uptime': '21816523'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '198264/g1',
                                                                       'name': '198264/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRTIzcGVlcmludGYxOTgyNjQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '198264',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxRTIzcGVlcmludGZwb24gMTk4MjY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198264',
                                                                                               'links': [       {       'connected_time': '21816305',
                                                                                                                        'fiber_length': 11387,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTIzcGVlcnBvbnBvbiAxOTgyNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4091 QUEEN ST (WAGLER EARL)',
                                         'poller_hash': 'QzgxRTIz',
                                         'serial_number': 'C81E23',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:45:cf',
                                         'uptime': '21816305'},
                                 {       'counter_source': '198384/g1+198384/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:73:5a'],
                                                                       'id': '198384/g1',
                                                                       'in_nucast_pkts': '603962',
                                                                       'in_octets': '72195780646',
                                                                       'in_ucast_pkts': '775313567',
                                                                       'mac_address': '84:d3:43:7d:47:c2',
                                                                       'mtu': 2000,
                                                                       'name': '198384/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68318699',
                                                                       'out_octets': '4845967949016',
                                                                       'out_ucast_pkts': '3289947539',
                                                                       'poller_hash': 'QzgxRTYxcGVlcmludGYxOTgzODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:0e:26'],
                                                                       'id': '198384/g3',
                                                                       'in_nucast_pkts': '2975740',
                                                                       'in_octets': '18008880235',
                                                                       'in_ucast_pkts': '189563890',
                                                                       'mac_address': '84:d3:43:7d:47:c4',
                                                                       'mtu': 2000,
                                                                       'name': '198384/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '208879318',
                                                                       'out_octets': '1076870471238',
                                                                       'out_ucast_pkts': '720307292',
                                                                       'poller_hash': 'QzgxRTYxcGVlcmludGYxOTgzODQvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '198384',
                                                                       'in_nucast_pkts': '277198017',
                                                                       'in_octets': '5922838420254',
                                                                       'in_ucast_pkts': '4010254831',
                                                                       'links': [       {       'in_nucast_pkts': '277198017',
                                                                                                'in_octets': '5922838420254',
                                                                                                'in_ucast_pkts': '4010254831',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3579702',
                                                                                                'out_octets': '90204660881',
                                                                                                'out_ucast_pkts': '964877457',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3579702',
                                                                       'out_octets': '90204660881',
                                                                       'out_ucast_pkts': '964877457',
                                                                       'poller_hash': 'QzgxRTYxcGVlcmludGZwb24gMTk4Mzg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198384',
                                                                                               'links': [       {       'connected_time': '29739904',
                                                                                                                        'fiber_length': 11720,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTYxcGVlcnBvbnBvbiAxOTgzODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6927 HORIZON TL (KROPF AUDREY)',
                                         'poller_hash': 'QzgxRTYx',
                                         'serial_number': 'C81E61',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:bf',
                                         'uptime': '29739904'},
                                 {       'counter_source': '198539/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '198539/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '34452348357',
                                                                       'in_ucast_pkts': '133425433',
                                                                       'mac_address': '84:d3:43:7d:46:27',
                                                                       'name': '198539/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '807094366415',
                                                                       'out_ucast_pkts': '688404605',
                                                                       'poller_hash': 'QzgxRTJFcGVlcmludGYxOTg1MzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '198539',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '807094366415',
                                                                       'in_ucast_pkts': '688404605',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '807094366415',
                                                                                                'in_ucast_pkts': '688404605',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '34452348357',
                                                                                                'out_ucast_pkts': '133425433',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '34452348357',
                                                                       'out_ucast_pkts': '133425433',
                                                                       'poller_hash': 'QzgxRTJFcGVlcmludGZwb24gMTk4NTM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198539',
                                                                                               'links': [       {       'connected_time': '1981829',
                                                                                                                        'fiber_length': 10817,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTJFcGVlcnBvbnBvbiAxOTg1Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6950 PERTH ROAD 121 (KUEPFER IVAN & '
                                                 'ANNAMAE)',
                                         'poller_hash': 'QzgxRTJF',
                                         'serial_number': 'C81E2E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:46:27',
                                         'uptime': '1981829'},
                                 {       'counter_source': '198579/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '198579/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12193300315',
                                                                       'in_ucast_pkts': '34311446',
                                                                       'mac_address': '84:d3:43:7d:43:57',
                                                                       'name': '198579/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '85416465331',
                                                                       'out_ucast_pkts': '83360488',
                                                                       'poller_hash': 'QzgxREQ0cGVlcmludGYxOTg1NzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '198579',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '85416465331',
                                                                       'in_ucast_pkts': '83360488',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '85416465331',
                                                                                                'in_ucast_pkts': '83360488',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '12193300315',
                                                                                                'out_ucast_pkts': '34311446',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12193300315',
                                                                       'out_ucast_pkts': '34311446',
                                                                       'poller_hash': 'QzgxREQ0cGVlcmludGZwb24gMTk4NTc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198579',
                                                                                               'links': [       {       'connected_time': '21816077',
                                                                                                                        'fiber_length': 11905,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREQ0cGVlcnBvbnBvbiAxOTg1Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4102 WYNDHAM ST (SOULLIERE LINDA)',
                                         'poller_hash': 'QzgxREQ0',
                                         'serial_number': 'C81DD4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:43:57',
                                         'uptime': '21816077'},
                                 {       'counter_source': '200176/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200176/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '93759621423',
                                                                       'in_ucast_pkts': '8998288164',
                                                                       'mac_address': '84:d3:43:7d:4a:cf',
                                                                       'name': '200176/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '1700146081646',
                                                                       'out_ucast_pkts': '1289461895',
                                                                       'poller_hash': 'QzgxRUMzcGVlcmludGYyMDAxNzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200176',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '1700146081646',
                                                                       'in_ucast_pkts': '1289461895',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '1700146081646',
                                                                                                'in_ucast_pkts': '1289461895',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '93759621423',
                                                                                                'out_ucast_pkts': '8998288164',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '93759621423',
                                                                       'out_ucast_pkts': '8998288164',
                                                                       'poller_hash': 'QzgxRUMzcGVlcmludGZwb24gMjAwMTc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200176',
                                                                                               'links': [       {       'connected_time': '688377',
                                                                                                                        'fiber_length': 11424,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUMzcGVlcnBvbnBvbiAyMDAxNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6927 PERTH ROAD 121 (SCHEERER MARK & '
                                                 'SANDY)',
                                         'poller_hash': 'QzgxRUMz',
                                         'serial_number': 'C81EC3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:cf',
                                         'uptime': '688377'},
                                 {       'counter_source': '200186/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200186/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '873312804834',
                                                                       'in_ucast_pkts': '61385302694',
                                                                       'mac_address': '84:d3:43:7d:4d:6f',
                                                                       'name': '200186/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '3057680426785',
                                                                       'out_ucast_pkts': '2622263255',
                                                                       'poller_hash': 'QzgxRjE3cGVlcmludGYyMDAxODYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200186',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '3057680426785',
                                                                       'in_ucast_pkts': '2622263255',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '3057680426785',
                                                                                                'in_ucast_pkts': '2622263255',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '873312804834',
                                                                                                'out_ucast_pkts': '61385302694',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '873312804834',
                                                                       'out_ucast_pkts': '61385302694',
                                                                       'poller_hash': 'QzgxRjE3cGVlcmludGZwb24gMjAwMTg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200186',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 11378,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRjE3cGVlcnBvbnBvbiAyMDAxODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4095 QUEEN ST (HOBDEN ROBERT & '
                                                 'DIANE)',
                                         'poller_hash': 'QzgxRjE3',
                                         'serial_number': 'C81F17',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4d:6f',
                                         'uptime': '688370'},
                                 {       'counter_source': '200196/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:39:aa'],
                                                                       'id': '200196/g1',
                                                                       'in_nucast_pkts': '14897',
                                                                       'in_octets': '7756183021',
                                                                       'in_ucast_pkts': '25714983',
                                                                       'mac_address': '84:d3:43:7d:48:ca',
                                                                       'mtu': 2000,
                                                                       'name': '200196/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464381',
                                                                       'out_octets': '198411442477',
                                                                       'out_ucast_pkts': '148388707',
                                                                       'poller_hash': 'QzgxRTgycGVlcmludGYyMDAxOTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '200196',
                                                                       'in_nucast_pkts': '1464381',
                                                                       'in_octets': '198411442477',
                                                                       'in_ucast_pkts': '148388707',
                                                                       'links': [       {       'in_nucast_pkts': '1464381',
                                                                                                'in_octets': '198411442477',
                                                                                                'in_ucast_pkts': '148388707',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14897',
                                                                                                'out_octets': '7756183021',
                                                                                                'out_ucast_pkts': '25714983',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14897',
                                                                       'out_octets': '7756183021',
                                                                       'out_ucast_pkts': '25714983',
                                                                       'poller_hash': 'QzgxRTgycGVlcmludGZwb24gMjAwMTk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200196',
                                                                                               'links': [       {       'connected_time': '688379',
                                                                                                                        'fiber_length': 3117,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTgycGVlcnBvbnBvbiAyMDAxOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '136 PUGH ST (ROBARTS CHARLOTTE)',
                                         'poller_hash': 'QzgxRTgy',
                                         'serial_number': 'C81E82',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:48:c7',
                                         'uptime': '688379'},
                                 {       'counter_source': '200345/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1d:85'],
                                                                       'id': '200345/g1',
                                                                       'in_nucast_pkts': '972300',
                                                                       'in_octets': '808432611516',
                                                                       'in_ucast_pkts': '2135048435',
                                                                       'mac_address': '84:d3:43:7d:44:3a',
                                                                       'mtu': 2000,
                                                                       'name': '200345/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '143674530',
                                                                       'out_octets': '13988603012708',
                                                                       'out_ucast_pkts': '10327520971',
                                                                       'poller_hash': 'QzgxREYwcGVlcmludGYyMDAzNDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '200345',
                                                                       'in_nucast_pkts': '143674530',
                                                                       'in_octets': '13988603012708',
                                                                       'in_ucast_pkts': '10327520971',
                                                                       'links': [       {       'in_nucast_pkts': '143674530',
                                                                                                'in_octets': '13988603012708',
                                                                                                'in_ucast_pkts': '10327520971',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '972300',
                                                                                                'out_octets': '808432611516',
                                                                                                'out_ucast_pkts': '2135048435',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '972300',
                                                                       'out_octets': '808432611516',
                                                                       'out_ucast_pkts': '2135048435',
                                                                       'poller_hash': 'QzgxREYwcGVlcmludGZwb24gMjAwMzQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200345',
                                                                                               'links': [       {       'connected_time': '21816495',
                                                                                                                        'fiber_length': 10787,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREYwcGVlcnBvbnBvbiAyMDAzNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6960 PERTH ROAD 121 (KUEPFER JAMES & '
                                                 'SARAH)',
                                         'poller_hash': 'QzgxREYw',
                                         'serial_number': 'C81DF0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:37',
                                         'uptime': '21816495'},
                                 {       'counter_source': '200355/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200355/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '242963451651',
                                                                       'in_ucast_pkts': '86734570812',
                                                                       'mac_address': '84:d3:43:7d:42:07',
                                                                       'name': '200355/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2845741619342',
                                                                       'out_ucast_pkts': '2066748858',
                                                                       'poller_hash': 'QzgxREFBcGVlcmludGYyMDAzNTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200355',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2845741619342',
                                                                       'in_ucast_pkts': '2066748858',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2845741619342',
                                                                                                'in_ucast_pkts': '2066748858',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '242963451651',
                                                                                                'out_ucast_pkts': '86734570812',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '242963451651',
                                                                       'out_ucast_pkts': '86734570812',
                                                                       'poller_hash': 'QzgxREFBcGVlcmludGZwb24gMjAwMzU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200355',
                                                                                               'links': [       {       'connected_time': '21815971',
                                                                                                                        'fiber_length': 11896,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREFBcGVlcnBvbnBvbiAyMDAzNTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4104 WYNDHAM ST (TAM KEITH & JACQUI)',
                                         'poller_hash': 'QzgxREFB',
                                         'serial_number': 'C81DAA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:42:07',
                                         'uptime': '21815971'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '200365/g1',
                                                                       'name': '200365/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxOUJBcGVlcmludGYyMDAzNjUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200365',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxOUJBcGVlcmludGZwb24gMjAwMzY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200365',
                                                                                               'links': [       {       'connected_time': '24124661',
                                                                                                                        'fiber_length': 11577,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxOUJBcGVlcnBvbnBvbiAyMDAzNjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6914 PERTH ROAD 121 (BARTSCH JOHAN)',
                                         'poller_hash': 'QzgxOUJB',
                                         'serial_number': 'C819BA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:22:87',
                                         'uptime': '24124661'},
                                 {       'counter_source': '200385/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200385/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '100566926582',
                                                                       'in_ucast_pkts': '1287234984',
                                                                       'mac_address': '84:d3:43:7d:49:67',
                                                                       'name': '200385/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '144',
                                                                       'out_octets': '5346872929929',
                                                                       'out_ucast_pkts': '3571137124',
                                                                       'poller_hash': 'QzgxRTk2cGVlcmludGYyMDAzODUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200385',
                                                                       'in_nucast_pkts': '144',
                                                                       'in_octets': '5346872929929',
                                                                       'in_ucast_pkts': '3571137124',
                                                                       'links': [       {       'in_nucast_pkts': '144',
                                                                                                'in_octets': '5346872929929',
                                                                                                'in_ucast_pkts': '3571137124',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '100566926582',
                                                                                                'out_ucast_pkts': '1287234984',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '100566926582',
                                                                       'out_ucast_pkts': '1287234984',
                                                                       'poller_hash': 'QzgxRTk2cGVlcmludGZwb24gMjAwMzg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200385',
                                                                                               'links': [       {       'connected_time': '21815653',
                                                                                                                        'fiber_length': 11637,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTk2cGVlcnBvbnBvbiAyMDAzODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4071 WATERLOO ST (STREICHER ELLA & '
                                                 'ERVIN)',
                                         'poller_hash': 'QzgxRTk2',
                                         'serial_number': 'C81E96',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:49:67',
                                         'uptime': '21815653'},
                                 {       'counter_source': '200430/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200430/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '13780343502',
                                                                       'in_ucast_pkts': '45384089',
                                                                       'mac_address': '84:d3:43:7d:42:bf',
                                                                       'name': '200430/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '633474524415',
                                                                       'out_ucast_pkts': '443512410',
                                                                       'poller_hash': 'QzgxREMxcGVlcmludGYyMDA0MzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200430',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '633474524415',
                                                                       'in_ucast_pkts': '443512410',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '633474524415',
                                                                                                'in_ucast_pkts': '443512410',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '13780343502',
                                                                                                'out_ucast_pkts': '45384089',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '13780343502',
                                                                       'out_ucast_pkts': '45384089',
                                                                       'poller_hash': 'QzgxREMxcGVlcmludGZwb24gMjAwNDMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200430',
                                                                                               'links': [       {       'connected_time': '688360',
                                                                                                                        'fiber_length': 11708,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREMxcGVlcnBvbnBvbiAyMDA0MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6928 HORIZON TL (GREEN DARRYL)',
                                         'poller_hash': 'QzgxREMx',
                                         'serial_number': 'C81DC1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:42:bf',
                                         'uptime': '688360'},
                                 {       'counter_source': '200440/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200440/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7956582242',
                                                                       'in_ucast_pkts': '34454319',
                                                                       'mac_address': '84:d3:43:7d:4c:c7',
                                                                       'name': '200440/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '120',
                                                                       'out_octets': '95739241939',
                                                                       'out_ucast_pkts': '123491827',
                                                                       'poller_hash': 'QzgxRjAycGVlcmludGYyMDA0NDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200440',
                                                                       'in_nucast_pkts': '120',
                                                                       'in_octets': '95739241939',
                                                                       'in_ucast_pkts': '123491827',
                                                                       'links': [       {       'in_nucast_pkts': '120',
                                                                                                'in_octets': '95739241939',
                                                                                                'in_ucast_pkts': '123491827',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '7956582242',
                                                                                                'out_ucast_pkts': '34454319',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7956582242',
                                                                       'out_ucast_pkts': '34454319',
                                                                       'poller_hash': 'QzgxRjAycGVlcmludGZwb24gMjAwNDQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200440',
                                                                                               'links': [       {       'connected_time': '1130993',
                                                                                                                        'fiber_length': 11683,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRjAycGVlcnBvbnBvbiAyMDA0NDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4105 CONNIE ST (GERBER DARLENE)',
                                         'poller_hash': 'QzgxRjAy',
                                         'serial_number': 'C81F02',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4c:c7',
                                         'uptime': '1130993'},
                                 {       'counter_source': '200450/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200450/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '419510648922',
                                                                       'in_ucast_pkts': '66681446254',
                                                                       'mac_address': '84:d3:43:7d:02:f7',
                                                                       'name': '200450/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9019318618171',
                                                                       'out_ucast_pkts': '6414552146',
                                                                       'poller_hash': 'QzgxNUM4cGVlcmludGYyMDA0NTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200450',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9019318618171',
                                                                       'in_ucast_pkts': '6414552146',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9019318618171',
                                                                                                'in_ucast_pkts': '6414552146',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '419510648922',
                                                                                                'out_ucast_pkts': '66681446254',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '419510648922',
                                                                       'out_ucast_pkts': '66681446254',
                                                                       'poller_hash': 'QzgxNUM4cGVlcmludGZwb24gMjAwNDUw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200450',
                                                                                               'links': [       {       'connected_time': '687935',
                                                                                                                        'fiber_length': 1640,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxNUM4cGVlcnBvbnBvbiAyMDA0NTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 COLBALT ST (MCKONE STEPHANIE)',
                                         'poller_hash': 'QzgxNUM4',
                                         'serial_number': 'C815C8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:02:f7',
                                         'uptime': '687935'},
                                 {       'counter_source': '200856/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200856/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '383854024099',
                                                                       'in_ucast_pkts': '2248804416',
                                                                       'mac_address': '84:d3:43:7d:47:a7',
                                                                       'name': '200856/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '9917490085363',
                                                                       'out_ucast_pkts': '7039437168',
                                                                       'poller_hash': 'QzgxRTVFcGVlcmludGYyMDA4NTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200856',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '9917490085363',
                                                                       'in_ucast_pkts': '7039437168',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '9917490085363',
                                                                                                'in_ucast_pkts': '7039437168',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '383854024099',
                                                                                                'out_ucast_pkts': '2248804416',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '383854024099',
                                                                       'out_ucast_pkts': '2248804416',
                                                                       'poller_hash': 'QzgxRTVFcGVlcmludGZwb24gMjAwODU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200856',
                                                                                               'links': [       {       'connected_time': '688379',
                                                                                                                        'fiber_length': 11923,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTVFcGVlcnBvbnBvbiAyMDA4NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6946 NICKLIN ST (JANSEN SHERI)',
                                         'poller_hash': 'QzgxRTVF',
                                         'serial_number': 'C81E5E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:a7',
                                         'uptime': '688379'},
                                 {       'counter_source': '200866/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '200866/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '254313639796',
                                                                       'in_ucast_pkts': '4534362706',
                                                                       'mac_address': '84:d3:43:7d:47:9f',
                                                                       'name': '200866/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '352422748404',
                                                                       'out_ucast_pkts': '333696555',
                                                                       'poller_hash': 'QzgxRTVEcGVlcmludGYyMDA4NjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '200866',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '352422748404',
                                                                       'in_ucast_pkts': '333696555',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '352422748404',
                                                                                                'in_ucast_pkts': '333696555',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '254313639796',
                                                                                                'out_ucast_pkts': '4534362706',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '254313639796',
                                                                       'out_ucast_pkts': '4534362706',
                                                                       'poller_hash': 'QzgxRTVEcGVlcmludGZwb24gMjAwODY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200866',
                                                                                               'links': [       {       'connected_time': '1913956',
                                                                                                                        'fiber_length': 11818,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTVEcGVlcnBvbnBvbiAyMDA4NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6936 HORIZON TL (KEEPSAKES BY '
                                                 'KLASSEN)',
                                         'poller_hash': 'QzgxRTVE',
                                         'serial_number': 'C81E5D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:9f',
                                         'uptime': '1913956'},
                                 {       'counter_source': '201116/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '201116/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '198663368952',
                                                                       'in_ucast_pkts': '1449620062',
                                                                       'mac_address': '84:d3:43:7d:4a:77',
                                                                       'name': '201116/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '5069877918952',
                                                                       'out_ucast_pkts': '4335687542',
                                                                       'poller_hash': 'QzgxRUI4cGVlcmludGYyMDExMTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '201116',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '5069877918952',
                                                                       'in_ucast_pkts': '4335687542',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '5069877918952',
                                                                                                'in_ucast_pkts': '4335687542',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '198663368952',
                                                                                                'out_ucast_pkts': '1449620062',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '198663368952',
                                                                       'out_ucast_pkts': '1449620062',
                                                                       'poller_hash': 'QzgxRUI4cGVlcmludGZwb24gMjAxMTE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201116',
                                                                                               'links': [       {       'connected_time': '72476750',
                                                                                                                        'fiber_length': 10851,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUI4cGVlcnBvbnBvbiAyMDExMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6958 PERTH ROAD 121 (WOELK JOHN)',
                                         'poller_hash': 'QzgxRUI4',
                                         'serial_number': 'C81EB8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:77',
                                         'uptime': '72476750'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '201285/g1',
                                                                       'name': '201285/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxNjA5cGVlcmludGYyMDEyODUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '201285',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxNjA5cGVlcmludGZwb24gMjAxMjg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201285',
                                                                                               'links': [       {       'connected_time': '19095402',
                                                                                                                        'fiber_length': 11526,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxNjA5cGVlcnBvbnBvbiAyMDEyODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4097 QUEEN ST (ALBRECHT NATHAN)',
                                         'poller_hash': 'QzgxNjA5',
                                         'serial_number': 'C81609',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:04:ff',
                                         'uptime': '19095402'},
                                 {       'counter_source': '201317/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '201317/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '85208868407',
                                                                       'in_ucast_pkts': '554079434',
                                                                       'mac_address': '84:d3:43:7d:47:b7',
                                                                       'name': '201317/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '4193121677740',
                                                                       'out_ucast_pkts': '3066135511',
                                                                       'poller_hash': 'QzgxRTYwcGVlcmludGYyMDEzMTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '201317',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '4193121677740',
                                                                       'in_ucast_pkts': '3066135511',
                                                                       'links': [       {       'in_nucast_pkts': '117',
                                                                                                'in_octets': '4193121677740',
                                                                                                'in_ucast_pkts': '3066135511',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '85208868407',
                                                                                                'out_ucast_pkts': '554079434',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '85208868407',
                                                                       'out_ucast_pkts': '554079434',
                                                                       'poller_hash': 'QzgxRTYwcGVlcmludGZwb24gMjAxMzE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201317',
                                                                                               'links': [       {       'connected_time': '10654124',
                                                                                                                        'fiber_length': 11453,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTYwcGVlcnBvbnBvbiAyMDEzMTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4081 QUEEN ST (JANTZI MURRAY & '
                                                 'TRACILEE)',
                                         'poller_hash': 'QzgxRTYw',
                                         'serial_number': 'C81E60',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:b7',
                                         'uptime': '10654124'},
                                 {       'counter_source': '201356/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '201356/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '186214273425',
                                                                       'in_ucast_pkts': '665263535',
                                                                       'mac_address': '04:bc:9f:36:61:5c',
                                                                       'name': '201356/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1839243489862',
                                                                       'out_ucast_pkts': '1647351579',
                                                                       'poller_hash': 'MTAwQUFGOXBlZXJpbnRmMjAxMzU2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '201356',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1839243489862',
                                                                       'in_ucast_pkts': '1647351579',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1839243489862',
                                                                                                'in_ucast_pkts': '1647351579',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '186214273425',
                                                                                                'out_ucast_pkts': '665263535',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '186214273425',
                                                                       'out_ucast_pkts': '665263535',
                                                                       'poller_hash': 'MTAwQUFGOXBlZXJpbnRmcG9uIDIwMTM1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201356',
                                                                                               'links': [       {       'connected_time': '47827',
                                                                                                                        'fiber_length': 11660,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.436,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTAwQUFGOXBlZXJwb25wb24gMjAxMzU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': "6979 MILLBANK MAIN ST (ZEHR'S "
                                                 'COUNTRY MARKET)',
                                         'poller_hash': 'MTAwQUFGOQ==',
                                         'serial_number': '100AAF9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '04:bc:9f:36:61:5c',
                                         'uptime': '47827'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '201388/g1',
                                                                       'name': '201388/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRDJBcGVlcmludGYyMDEzODgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '201388',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxRDJBcGVlcmludGZwb24gMjAxMzg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201388',
                                                                                               'links': [       {       'connected_time': '21815926',
                                                                                                                        'fiber_length': 8151,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDJBcGVlcnBvbnBvbiAyMDEzODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6939 ROAD 140 (LANNIN DON)',
                                         'poller_hash': 'QzgxRDJB',
                                         'serial_number': 'C81D2A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:3e:07',
                                         'uptime': '21815926'},
                                 {       'counter_source': '201654/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '201654/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '577135536638',
                                                                       'in_ucast_pkts': '1383781709',
                                                                       'mac_address': '84:d3:43:7d:4a:17',
                                                                       'name': '201654/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '74',
                                                                       'out_octets': '5211627569264',
                                                                       'out_ucast_pkts': '3813577860',
                                                                       'poller_hash': 'QzgxRUFDcGVlcmludGYyMDE2NTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '201654',
                                                                       'in_nucast_pkts': '74',
                                                                       'in_octets': '5211627569264',
                                                                       'in_ucast_pkts': '3813577860',
                                                                       'links': [       {       'in_nucast_pkts': '74',
                                                                                                'in_octets': '5211627569264',
                                                                                                'in_ucast_pkts': '3813577860',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '577135536638',
                                                                                                'out_ucast_pkts': '1383781709',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '577135536638',
                                                                       'out_ucast_pkts': '1383781709',
                                                                       'poller_hash': 'QzgxRUFDcGVlcmludGZwb24gMjAxNjU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201654',
                                                                                               'links': [       {       'connected_time': '523084',
                                                                                                                        'fiber_length': 11803,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUFDcGVlcnBvbnBvbiAyMDE2NTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4092 WATERLOO ST (ELGIE CHAD)',
                                         'poller_hash': 'QzgxRUFD',
                                         'serial_number': 'C81EAC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:17',
                                         'uptime': '523084'},
                                 {       'counter_source': '202595/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '202595/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '246737583297',
                                                                       'in_ucast_pkts': '129189300924',
                                                                       'mac_address': '84:d3:43:7d:20:07',
                                                                       'name': '202595/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '99',
                                                                       'out_octets': '995607284861',
                                                                       'out_ucast_pkts': '872432682',
                                                                       'poller_hash': 'QzgxOTZBcGVlcmludGYyMDI1OTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '202595',
                                                                       'in_nucast_pkts': '99',
                                                                       'in_octets': '995607284861',
                                                                       'in_ucast_pkts': '872432682',
                                                                       'links': [       {       'in_nucast_pkts': '99',
                                                                                                'in_octets': '995607284861',
                                                                                                'in_ucast_pkts': '872432682',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '246737583297',
                                                                                                'out_ucast_pkts': '129189300924',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '246737583297',
                                                                       'out_ucast_pkts': '129189300924',
                                                                       'poller_hash': 'QzgxOTZBcGVlcmludGZwb24gMjAyNTk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202595',
                                                                                               'links': [       {       'connected_time': '688353',
                                                                                                                        'fiber_length': 11573,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxOTZBcGVlcnBvbnBvbiAyMDI1OTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6975 CHURCH ST (STREICHER FANNIE)',
                                         'poller_hash': 'QzgxOTZB',
                                         'serial_number': 'C8196A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:20:07',
                                         'uptime': '688353'},
                                 {       'counter_source': '202674/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:cf:b8:4b'],
                                                                       'id': '202674/g1',
                                                                       'in_nucast_pkts': '378844',
                                                                       'in_octets': '141974769016',
                                                                       'in_ucast_pkts': '841270511',
                                                                       'mac_address': '84:d3:43:7d:45:42',
                                                                       'mtu': 2000,
                                                                       'name': '202674/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '46232990',
                                                                       'out_octets': '2225715288393',
                                                                       'out_ucast_pkts': '1698640027',
                                                                       'poller_hash': 'QzgxRTExcGVlcmludGYyMDI2NzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '202674',
                                                                       'in_nucast_pkts': '46232990',
                                                                       'in_octets': '2225715288393',
                                                                       'in_ucast_pkts': '1698640027',
                                                                       'links': [       {       'in_nucast_pkts': '46232990',
                                                                                                'in_octets': '2225715288393',
                                                                                                'in_ucast_pkts': '1698640027',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '378844',
                                                                                                'out_octets': '141974769016',
                                                                                                'out_ucast_pkts': '841270511',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '378844',
                                                                       'out_octets': '141974769016',
                                                                       'out_ucast_pkts': '841270511',
                                                                       'poller_hash': 'QzgxRTExcGVlcmludGZwb24gMjAyNjc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202674',
                                                                                               'links': [       {       'connected_time': '21815627',
                                                                                                                        'fiber_length': 11546,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTExcGVlcnBvbnBvbiAyMDI2NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6984 CHURCH ST (BRUBACHER DARREN & '
                                                 'DARLENE)',
                                         'poller_hash': 'QzgxRTEx',
                                         'serial_number': 'C81E11',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:45:3f',
                                         'uptime': '21815627'},
                                 {       'counter_source': '202684/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e9:d2'],
                                                                       'id': '202684/g1',
                                                                       'in_nucast_pkts': '12710',
                                                                       'in_octets': '9726280181',
                                                                       'in_ucast_pkts': '53205514',
                                                                       'mac_address': '84:d3:43:7d:48:9a',
                                                                       'mtu': 2000,
                                                                       'name': '202684/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464375',
                                                                       'out_octets': '202018200868',
                                                                       'out_ucast_pkts': '143471998',
                                                                       'poller_hash': 'QzgxRTdDcGVlcmludGYyMDI2ODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '202684',
                                                                       'in_nucast_pkts': '1464375',
                                                                       'in_octets': '202018200868',
                                                                       'in_ucast_pkts': '143471998',
                                                                       'links': [       {       'in_nucast_pkts': '1464375',
                                                                                                'in_octets': '202018200868',
                                                                                                'in_ucast_pkts': '143471998',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12710',
                                                                                                'out_octets': '9726280181',
                                                                                                'out_ucast_pkts': '53205514',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12710',
                                                                       'out_octets': '9726280181',
                                                                       'out_ucast_pkts': '53205514',
                                                                       'poller_hash': 'QzgxRTdDcGVlcmludGZwb24gMjAyNjg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202684',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 2930,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTdDcGVlcnBvbnBvbiAyMDI2ODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '140 PUGH ST (WEST JAMES)',
                                         'poller_hash': 'QzgxRTdD',
                                         'serial_number': 'C81E7C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:48:97',
                                         'uptime': '688370'},
                                 {       'counter_source': '202818/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '202818/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '441195473245',
                                                                       'in_ucast_pkts': '5591397608',
                                                                       'mac_address': '84:d3:43:7d:4a:37',
                                                                       'name': '202818/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '7970225170103',
                                                                       'out_ucast_pkts': '6104260530',
                                                                       'poller_hash': 'QzgxRUIwcGVlcmludGYyMDI4MTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '202818',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '7970225170103',
                                                                       'in_ucast_pkts': '6104260530',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '7970225170103',
                                                                                                'in_ucast_pkts': '6104260530',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '441195473245',
                                                                                                'out_ucast_pkts': '5591397608',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '441195473245',
                                                                       'out_ucast_pkts': '5591397608',
                                                                       'poller_hash': 'QzgxRUIwcGVlcmludGZwb24gMjAyODE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202818',
                                                                                               'links': [       {       'connected_time': '73426872',
                                                                                                                        'fiber_length': 9385,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUIwcGVlcnBvbnBvbiAyMDI4MTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6237 ROAD 136 (DIEHL ERIC & SUE)',
                                         'poller_hash': 'QzgxRUIw',
                                         'serial_number': 'C81EB0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:37',
                                         'uptime': '73426872'},
                                 {       'counter_source': '202828/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '202828/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '370390449562',
                                                                       'in_ucast_pkts': '635659650',
                                                                       'mac_address': '84:d3:43:7d:4d:87',
                                                                       'name': '202828/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1709963262316',
                                                                       'out_ucast_pkts': '1305378022',
                                                                       'poller_hash': 'QzgxRjFBcGVlcmludGYyMDI4MjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '202828',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1709963262316',
                                                                       'in_ucast_pkts': '1305378022',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1709963262316',
                                                                                                'in_ucast_pkts': '1305378022',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '370390449562',
                                                                                                'out_ucast_pkts': '635659650',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '370390449562',
                                                                       'out_ucast_pkts': '635659650',
                                                                       'poller_hash': 'QzgxRjFBcGVlcmludGZwb24gMjAyODI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202828',
                                                                                               'links': [       {       'connected_time': '688354',
                                                                                                                        'fiber_length': 11518,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRjFBcGVlcnBvbnBvbiAyMDI4Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4076 QUEEN ST (WATKINS JOANNA)',
                                         'poller_hash': 'QzgxRjFB',
                                         'serial_number': 'C81F1A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4d:87',
                                         'uptime': '688354'},
                                 {       'counter_source': '202868/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '202868/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '191160664901',
                                                                       'in_ucast_pkts': '14942095179',
                                                                       'mac_address': '84:d3:43:7d:49:af',
                                                                       'name': '202868/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '6720701526804',
                                                                       'out_ucast_pkts': '4580253189',
                                                                       'poller_hash': 'QzgxRTlGcGVlcmludGYyMDI4NjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '202868',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '6720701526804',
                                                                       'in_ucast_pkts': '4580253189',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '6720701526804',
                                                                                                'in_ucast_pkts': '4580253189',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '191160664901',
                                                                                                'out_ucast_pkts': '14942095179',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '191160664901',
                                                                       'out_ucast_pkts': '14942095179',
                                                                       'poller_hash': 'QzgxRTlGcGVlcmludGZwb24gMjAyODY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202868',
                                                                                               'links': [       {       'connected_time': '6821488',
                                                                                                                        'fiber_length': 11334,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTlGcGVlcnBvbnBvbiAyMDI4Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4074 WYNDHAM ST (CRUMMER TOM)',
                                         'poller_hash': 'QzgxRTlG',
                                         'serial_number': 'C81E9F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:49:af',
                                         'uptime': '6821488'},
                                 {       'counter_source': '202878/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '202878/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '20407853315',
                                                                       'in_ucast_pkts': '113961296',
                                                                       'mac_address': '84:d3:43:7d:4a:6f',
                                                                       'name': '202878/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '315884441967',
                                                                       'out_ucast_pkts': '253315771',
                                                                       'poller_hash': 'QzgxRUI3cGVlcmludGYyMDI4NzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '202878',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '315884441967',
                                                                       'in_ucast_pkts': '253315771',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '315884441967',
                                                                                                'in_ucast_pkts': '253315771',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '20407853315',
                                                                                                'out_ucast_pkts': '113961296',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '20407853315',
                                                                       'out_ucast_pkts': '113961296',
                                                                       'poller_hash': 'QzgxRUI3cGVlcmludGZwb24gMjAyODc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202878',
                                                                                               'links': [       {       'connected_time': '21815859',
                                                                                                                        'fiber_length': 11039,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUI3cGVlcnBvbnBvbiAyMDI4Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4088 PRINCESS ST (LICHTI NANCY)',
                                         'poller_hash': 'QzgxRUI3',
                                         'serial_number': 'C81EB7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:6f',
                                         'uptime': '21815859'},
                                 {       'counter_source': '203813/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8c:4b'],
                                                                       'id': '203813/g1',
                                                                       'in_nucast_pkts': '337236',
                                                                       'in_octets': '3789793684',
                                                                       'in_ucast_pkts': '12179639',
                                                                       'mac_address': '84:d3:43:7d:49:22',
                                                                       'mtu': 2000,
                                                                       'name': '203813/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447542',
                                                                       'out_octets': '52025578820',
                                                                       'out_ucast_pkts': '39170411',
                                                                       'poller_hash': 'QzgxRThEcGVlcmludGYyMDM4MTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '203813',
                                                                       'in_nucast_pkts': '1447542',
                                                                       'in_octets': '52025578820',
                                                                       'in_ucast_pkts': '39170411',
                                                                       'links': [       {       'in_nucast_pkts': '1447542',
                                                                                                'in_octets': '52025578820',
                                                                                                'in_ucast_pkts': '39170411',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '337236',
                                                                                                'out_octets': '3789793684',
                                                                                                'out_ucast_pkts': '12179639',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '337236',
                                                                       'out_octets': '3789793684',
                                                                       'out_ucast_pkts': '12179639',
                                                                       'poller_hash': 'QzgxRThEcGVlcmludGZwb24gMjAzODEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '203813',
                                                                                               'links': [       {       'connected_time': '688414',
                                                                                                                        'fiber_length': 8409,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRThEcGVlcnBvbnBvbiAyMDM4MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4866 LINE 59 (MILLER WILLIAM A)',
                                         'poller_hash': 'QzgxRThE',
                                         'serial_number': 'C81E8D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:49:1f',
                                         'uptime': '688414'},
                                 {       'counter_source': '204223/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '204223/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '258293253090',
                                                                       'in_ucast_pkts': '1362058557',
                                                                       'mac_address': '84:d3:43:7d:45:37',
                                                                       'name': '204223/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '62',
                                                                       'out_octets': '3806740244785',
                                                                       'out_ucast_pkts': '3200725239',
                                                                       'poller_hash': 'QzgxRTEwcGVlcmludGYyMDQyMjMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '204223',
                                                                       'in_nucast_pkts': '62',
                                                                       'in_octets': '3806740244785',
                                                                       'in_ucast_pkts': '3200725239',
                                                                       'links': [       {       'in_nucast_pkts': '62',
                                                                                                'in_octets': '3806740244785',
                                                                                                'in_ucast_pkts': '3200725239',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '258293253090',
                                                                                                'out_ucast_pkts': '1362058557',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '258293253090',
                                                                       'out_ucast_pkts': '1362058557',
                                                                       'poller_hash': 'QzgxRTEwcGVlcmludGZwb24gMjA0MjIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '204223',
                                                                                               'links': [       {       'connected_time': '21816474',
                                                                                                                        'fiber_length': 8168,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTEwcGVlcnBvbnBvbiAyMDQyMjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4885 LINE 59 (WAGNER KYLE & CRYSTAL)',
                                         'poller_hash': 'QzgxRTEw',
                                         'serial_number': 'C81E10',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:45:37',
                                         'uptime': '21816474'},
                                 {       'counter_source': '204925/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '204925/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '132812031037',
                                                                       'in_ucast_pkts': '43141185377',
                                                                       'mac_address': '84:d3:43:7d:48:2f',
                                                                       'name': '204925/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '454316814274',
                                                                       'out_ucast_pkts': '430736318',
                                                                       'poller_hash': 'QzgxRTZGcGVlcmludGYyMDQ5MjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '204925',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '454316814274',
                                                                       'in_ucast_pkts': '430736318',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '454316814274',
                                                                                                'in_ucast_pkts': '430736318',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '132812031037',
                                                                                                'out_ucast_pkts': '43141185377',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '132812031037',
                                                                       'out_ucast_pkts': '43141185377',
                                                                       'poller_hash': 'QzgxRTZGcGVlcmludGZwb24gMjA0OTI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '204925',
                                                                                               'links': [       {       'connected_time': '21816227',
                                                                                                                        'fiber_length': 1127,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTZGcGVlcnBvbnBvbiAyMDQ5MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '30 MAIN ST S (OAK TREE FINANCIAL)',
                                         'poller_hash': 'QzgxRTZG',
                                         'serial_number': 'C81E6F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:48:2f',
                                         'uptime': '21816227'},
                                 {       'counter_source': '204970/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '204970/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '22928764110',
                                                                       'in_ucast_pkts': '36826673',
                                                                       'mac_address': '84:d3:43:7d:46:67',
                                                                       'name': '204970/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '67876652101',
                                                                       'out_ucast_pkts': '70173301',
                                                                       'poller_hash': 'QzgxRTM2cGVlcmludGYyMDQ5NzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '204970',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '67876652101',
                                                                       'in_ucast_pkts': '70173301',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '67876652101',
                                                                                                'in_ucast_pkts': '70173301',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '22928764110',
                                                                                                'out_ucast_pkts': '36826673',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '22928764110',
                                                                       'out_ucast_pkts': '36826673',
                                                                       'poller_hash': 'QzgxRTM2cGVlcmludGZwb24gMjA0OTcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '204970',
                                                                                               'links': [       {       'connected_time': '21815545',
                                                                                                                        'fiber_length': 11872,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTM2cGVlcnBvbnBvbiAyMDQ5NzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4106 WYNDHAM ST MILLBANK (WAGLER '
                                                 'WAYNE & DIANNE)',
                                         'poller_hash': 'QzgxRTM2',
                                         'serial_number': 'C81E36',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:46:67',
                                         'uptime': '21815545'},
                                 {       'counter_source': '205070/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '205070/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '272194299947',
                                                                       'in_ucast_pkts': '897610253',
                                                                       'mac_address': '84:d3:43:7d:47:37',
                                                                       'name': '205070/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3533233105464',
                                                                       'out_ucast_pkts': '2745436998',
                                                                       'poller_hash': 'QzgxRTUwcGVlcmludGYyMDUwNzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205070',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3533233105464',
                                                                       'in_ucast_pkts': '2745436998',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3533233105464',
                                                                                                'in_ucast_pkts': '2745436998',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '272194299947',
                                                                                                'out_ucast_pkts': '897610253',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '272194299947',
                                                                       'out_ucast_pkts': '897610253',
                                                                       'poller_hash': 'QzgxRTUwcGVlcmludGZwb24gMjA1MDcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205070',
                                                                                               'links': [       {       'connected_time': '10014597',
                                                                                                                        'fiber_length': 11100,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTUwcGVlcnBvbnBvbiAyMDUwNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4092 PRINCESS ST (DELONG JOHN & '
                                                 'JOCELYN)',
                                         'poller_hash': 'QzgxRTUw',
                                         'serial_number': 'C81E50',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:37',
                                         'uptime': '10014597'},
                                 {       'counter_source': '205080/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:eb:71:ce',
                                                                                                        'd0:76:8f:1f:64:14'],
                                                                       'id': '205080/g1',
                                                                       'in_nucast_pkts': '2253964',
                                                                       'in_octets': '795638034',
                                                                       'in_ucast_pkts': '1942956',
                                                                       'mac_address': '84:d3:43:7d:37:4a',
                                                                       'mtu': 2000,
                                                                       'name': '205080/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1856852312',
                                                                       'out_octets': '2496027215859',
                                                                       'out_ucast_pkts': '2278158',
                                                                       'poller_hash': 'QzgxQzUycGVlcmludGYyMDUwODAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205080',
                                                                       'in_nucast_pkts': '1856852312',
                                                                       'in_octets': '2496027215859',
                                                                       'in_ucast_pkts': '2278158',
                                                                       'links': [       {       'in_nucast_pkts': '1856852312',
                                                                                                'in_octets': '2496027215859',
                                                                                                'in_ucast_pkts': '2278158',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2253964',
                                                                                                'out_octets': '795638034',
                                                                                                'out_ucast_pkts': '1942956',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2253964',
                                                                       'out_octets': '795638034',
                                                                       'out_ucast_pkts': '1942956',
                                                                       'poller_hash': 'QzgxQzUycGVlcmludGZwb24gMjA1MDgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205080',
                                                                                               'links': [       {       'connected_time': '2969395',
                                                                                                                        'fiber_length': 17761,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -24.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxQzUycGVlcnBvbnBvbiAyMDUwODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '169 MARKET ST (HOWARD MARDLING '
                                                 '10431857)',
                                         'poller_hash': 'QzgxQzUy',
                                         'serial_number': 'C81C52',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:37:47',
                                         'uptime': '2969395'},
                                 {       'counter_source': '205090/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '205090/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '130762744312',
                                                                       'in_ucast_pkts': '237781294',
                                                                       'mac_address': '84:d3:43:7d:40:47',
                                                                       'name': '205090/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '312058573535',
                                                                       'out_ucast_pkts': '330476155',
                                                                       'poller_hash': 'QzgxRDcycGVlcmludGYyMDUwOTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205090',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '312058573535',
                                                                       'in_ucast_pkts': '330476155',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '312058573535',
                                                                                                'in_ucast_pkts': '330476155',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '130762744312',
                                                                                                'out_ucast_pkts': '237781294',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '130762744312',
                                                                       'out_ucast_pkts': '237781294',
                                                                       'poller_hash': 'QzgxRDcycGVlcmludGZwb24gMjA1MDkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205090',
                                                                                               'links': [       {       'connected_time': '688502',
                                                                                                                        'fiber_length': 1451,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDcycGVlcnBvbnBvbiAyMDUwOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6611D PERTH ROAD 131 (ACRE75)',
                                         'poller_hash': 'QzgxRDcy',
                                         'serial_number': 'C81D72',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:40:47',
                                         'uptime': '688502'},
                                 {       'counter_source': '205120/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '205120/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '80249061261',
                                                                       'in_ucast_pkts': '498672635',
                                                                       'mac_address': '84:d3:43:7d:e6:e7',
                                                                       'name': '205120/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '2680553607091',
                                                                       'out_ucast_pkts': '1883399228',
                                                                       'poller_hash': 'QzgzMjQ2cGVlcmludGYyMDUxMjAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205120',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '2680553607091',
                                                                       'in_ucast_pkts': '1883399228',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '2680553607091',
                                                                                                'in_ucast_pkts': '1883399228',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '80249061261',
                                                                                                'out_ucast_pkts': '498672635',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '80249061261',
                                                                       'out_ucast_pkts': '498672635',
                                                                       'poller_hash': 'QzgzMjQ2cGVlcmludGZwb24gMjA1MTIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205120',
                                                                                               'links': [       {       'connected_time': '30163283',
                                                                                                                        'fiber_length': 5717,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMjQ2cGVlcnBvbnBvbiAyMDUxMjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6420 ROAD 140 (STREICHER SAM & '
                                                 'ESTHER)',
                                         'poller_hash': 'QzgzMjQ2',
                                         'serial_number': 'C83246',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e6:e7',
                                         'uptime': '30163283'},
                                 {       'counter_source': '205376/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:ac:eb'],
                                                                       'id': '205376/g1',
                                                                       'in_nucast_pkts': '12768',
                                                                       'in_octets': '14287321011',
                                                                       'in_ucast_pkts': '74465891',
                                                                       'mac_address': '84:d3:43:7d:44:1a',
                                                                       'mtu': 2000,
                                                                       'name': '205376/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464414',
                                                                       'out_octets': '508182822283',
                                                                       'out_ucast_pkts': '361073827',
                                                                       'poller_hash': 'QzgxREVDcGVlcmludGYyMDUzNzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205376',
                                                                       'in_nucast_pkts': '1464414',
                                                                       'in_octets': '508182822283',
                                                                       'in_ucast_pkts': '361073827',
                                                                       'links': [       {       'in_nucast_pkts': '1464414',
                                                                                                'in_octets': '508182822283',
                                                                                                'in_ucast_pkts': '361073827',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12768',
                                                                                                'out_octets': '14287321011',
                                                                                                'out_ucast_pkts': '74465891',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12768',
                                                                       'out_octets': '14287321011',
                                                                       'out_ucast_pkts': '74465891',
                                                                       'poller_hash': 'QzgxREVDcGVlcmludGZwb24gMjA1Mzc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205376',
                                                                                               'links': [       {       'connected_time': '688367',
                                                                                                                        'fiber_length': 2563,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREVDcGVlcnBvbnBvbiAyMDUzNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "99 PUGH ST (D'HULSTER EMIEL)",
                                         'poller_hash': 'QzgxREVD',
                                         'serial_number': 'C81DEC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:17',
                                         'uptime': '688367'},
                                 {       'counter_source': '205386/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '205386/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '171131640662',
                                                                       'in_ucast_pkts': '26476817380',
                                                                       'mac_address': '84:d3:43:7d:48:5f',
                                                                       'name': '205386/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '3543876004186',
                                                                       'out_ucast_pkts': '2620234533',
                                                                       'poller_hash': 'QzgxRTc1cGVlcmludGYyMDUzODYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205386',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '3543876004186',
                                                                       'in_ucast_pkts': '2620234533',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '3543876004186',
                                                                                                'in_ucast_pkts': '2620234533',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '171131640662',
                                                                                                'out_ucast_pkts': '26476817380',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '171131640662',
                                                                       'out_ucast_pkts': '26476817380',
                                                                       'poller_hash': 'QzgxRTc1cGVlcmludGZwb24gMjA1Mzg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205386',
                                                                                               'links': [       {       'connected_time': '688404',
                                                                                                                        'fiber_length': 1272,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTc1cGVlcnBvbnBvbiAyMDUzODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '8 LLANA HEIGHTS DR (STECKLY JAMES)',
                                         'poller_hash': 'QzgxRTc1',
                                         'serial_number': 'C81E75',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:48:5f',
                                         'uptime': '688404'},
                                 {       'counter_source': '205406/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:e9:a5'],
                                                                       'id': '205406/g1',
                                                                       'in_nucast_pkts': '218524',
                                                                       'in_octets': '187964415910',
                                                                       'in_ucast_pkts': '1072731480',
                                                                       'mac_address': 'ec:4f:82:b9:22:84',
                                                                       'mtu': 2000,
                                                                       'name': '205406/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19805635',
                                                                       'out_octets': '4215902741878',
                                                                       'out_ucast_pkts': '2985392652',
                                                                       'poller_hash': 'MkZFQUZBcGVlcmludGYyMDU0MDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205406',
                                                                       'in_nucast_pkts': '19805635',
                                                                       'in_octets': '4215902741878',
                                                                       'in_ucast_pkts': '2985392652',
                                                                       'links': [       {       'in_nucast_pkts': '19805635',
                                                                                                'in_octets': '4215902741878',
                                                                                                'in_ucast_pkts': '2985392652',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '218524',
                                                                                                'out_octets': '187964415910',
                                                                                                'out_ucast_pkts': '1072731480',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '218524',
                                                                       'out_octets': '187964415910',
                                                                       'out_ucast_pkts': '1072731480',
                                                                       'poller_hash': 'MkZFQUZBcGVlcmludGZwb24gMjA1NDA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205406',
                                                                                               'links': [       {       'connected_time': '10996353',
                                                                                                                        'fiber_length': 1019,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.69,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQUZBcGVlcnBvbnBvbiAyMDU0MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '3 GROSCH ST MILVERTON (NAHRGANG '
                                                 'GEORGE)',
                                         'poller_hash': 'MkZFQUZB',
                                         'serial_number': '2FEAFA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:22:80',
                                         'uptime': '10996353'},
                                 {       'counter_source': '205526/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '205526/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '211533567714',
                                                                       'in_ucast_pkts': '616622013',
                                                                       'mac_address': 'ec:4f:82:dc:d7:0c',
                                                                       'name': '205526/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3379657791838',
                                                                       'out_ucast_pkts': '2635314220',
                                                                       'poller_hash': 'MkVENjdCcGVlcmludGYyMDU1MjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205526',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3379657791838',
                                                                       'in_ucast_pkts': '2635314220',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3379657791838',
                                                                                                'in_ucast_pkts': '2635314220',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '211533567714',
                                                                                                'out_ucast_pkts': '616622013',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '211533567714',
                                                                       'out_ucast_pkts': '616622013',
                                                                       'poller_hash': 'MkVENjdCcGVlcmludGZwb24gMjA1NTI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205526',
                                                                                               'links': [       {       'connected_time': '47694',
                                                                                                                        'fiber_length': 10802,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.518,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVENjdCcGVlcnBvbnBvbiAyMDU1MjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6374 ROAD 154 (VAN NULAND MARIA)',
                                         'poller_hash': 'MkVENjdC',
                                         'serial_number': '2ED67B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:dc:d7:0c',
                                         'uptime': '47694'},
                                 {       'counter_source': '205536/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:c0:a4'],
                                                                       'id': '205536/g1',
                                                                       'in_nucast_pkts': '12176',
                                                                       'in_octets': '13874742607',
                                                                       'in_ucast_pkts': '69030630',
                                                                       'mac_address': '84:d3:43:7d:40:9a',
                                                                       'mtu': 2000,
                                                                       'name': '205536/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464456',
                                                                       'out_octets': '254688242062',
                                                                       'out_ucast_pkts': '195684526',
                                                                       'poller_hash': 'QzgxRDdDcGVlcmludGYyMDU1MzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205536',
                                                                       'in_nucast_pkts': '1464456',
                                                                       'in_octets': '254688242062',
                                                                       'in_ucast_pkts': '195684526',
                                                                       'links': [       {       'in_nucast_pkts': '1464456',
                                                                                                'in_octets': '254688242062',
                                                                                                'in_ucast_pkts': '195684526',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12176',
                                                                                                'out_octets': '13874742607',
                                                                                                'out_ucast_pkts': '69030630',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12176',
                                                                       'out_octets': '13874742607',
                                                                       'out_ucast_pkts': '69030630',
                                                                       'poller_hash': 'QzgxRDdDcGVlcmludGZwb24gMjA1NTM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205536',
                                                                                               'links': [       {       'connected_time': '688376',
                                                                                                                        'fiber_length': 2597,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDdDcGVlcnBvbnBvbiAyMDU1MzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '93 PUGH ST (SAMUEL MARK & GILLIAN)',
                                         'poller_hash': 'QzgxRDdD',
                                         'serial_number': 'C81D7C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:40:97',
                                         'uptime': '688376'},
                                 {       'counter_source': '205621/g1+205621/g2+205621/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:5f:67:a4:50:5d'],
                                                                       'id': '205621/g1',
                                                                       'in_nucast_pkts': '11858155',
                                                                       'in_octets': '138182533611',
                                                                       'in_ucast_pkts': '792833747',
                                                                       'mac_address': 'cc:be:59:61:1b:be',
                                                                       'mtu': 2000,
                                                                       'name': '205621/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43259403',
                                                                       'out_octets': '4059775222953',
                                                                       'out_ucast_pkts': '2917749078',
                                                                       'poller_hash': 'NDU5NkE1cGVlcmludGYyMDU2MjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:83:e7:7c:39:98'],
                                                                       'id': '205621/g2',
                                                                       'in_nucast_pkts': '8670',
                                                                       'in_octets': '239914240538',
                                                                       'in_ucast_pkts': '1922170195',
                                                                       'mac_address': 'cc:be:59:61:1b:bf',
                                                                       'mtu': 2000,
                                                                       'name': '205621/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43254884',
                                                                       'out_octets': '7387395705841',
                                                                       'out_ucast_pkts': '5243421513',
                                                                       'poller_hash': 'NDU5NkE1cGVlcmludGYyMDU2MjEvZzI=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:2f:d0:93:60:3b'],
                                                                       'id': '205621/g4',
                                                                       'in_nucast_pkts': '7910',
                                                                       'in_octets': '54093590986',
                                                                       'in_ucast_pkts': '120838276',
                                                                       'mac_address': 'cc:be:59:61:1b:c1',
                                                                       'mtu': 2000,
                                                                       'name': '205621/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43265104',
                                                                       'out_octets': '725754992783',
                                                                       'out_ucast_pkts': '584937121',
                                                                       'poller_hash': 'NDU5NkE1cGVlcmludGYyMDU2MjEvZzQ=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205621',
                                                                       'in_nucast_pkts': '129779391',
                                                                       'in_octets': '12172925921577',
                                                                       'in_ucast_pkts': '8746107712',
                                                                       'links': [       {       'in_nucast_pkts': '129779391',
                                                                                                'in_octets': '12172925921577',
                                                                                                'in_ucast_pkts': '8746107712',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11874735',
                                                                                                'out_octets': '432190365135',
                                                                                                'out_ucast_pkts': '2835842218',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11874735',
                                                                       'out_octets': '432190365135',
                                                                       'out_ucast_pkts': '2835842218',
                                                                       'poller_hash': 'NDU5NkE1cGVlcmludGZwb24gMjA1NjIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205621',
                                                                                               'links': [       {       'connected_time': '24669403',
                                                                                                                        'fiber_length': 18101,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.538,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDU5NkE1cGVlcnBvbnBvbiAyMDU2MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '9 & 11 Maddison Street E '
                                                 '(Multi-customer)',
                                         'poller_hash': 'NDU5NkE1',
                                         'serial_number': '4596A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:61:1b:ba',
                                         'uptime': '24669403'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '205743/g1',
                                                                       'name': '205743/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRDgwcGVlcmludGYyMDU3NDMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205743',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxRDgwcGVlcmludGZwb24gMjA1NzQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205743',
                                                                                               'links': [       {       'connected_time': '9673962',
                                                                                                                        'fiber_length': 1794,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDgwcGVlcnBvbnBvbiAyMDU3NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (SOUTH SIDE ESTATES '
                                                 'INC)',
                                         'poller_hash': 'QzgxRDgw',
                                         'serial_number': 'C81D80',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:40:b7',
                                         'uptime': '9673962'},
                                 {       'counter_source': '205833/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9d:ad:ab'],
                                                                       'id': '205833/g1',
                                                                       'in_nucast_pkts': '11665',
                                                                       'in_octets': '100634487610',
                                                                       'in_ucast_pkts': '86046862',
                                                                       'mac_address': '84:d3:43:7d:47:72',
                                                                       'mtu': 2000,
                                                                       'name': '205833/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1444329',
                                                                       'out_octets': '90904980749',
                                                                       'out_ucast_pkts': '120829431',
                                                                       'poller_hash': 'QzgxRTU3cGVlcmludGYyMDU4MzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205833',
                                                                       'in_nucast_pkts': '1444329',
                                                                       'in_octets': '90904980749',
                                                                       'in_ucast_pkts': '120829431',
                                                                       'links': [       {       'in_nucast_pkts': '1444329',
                                                                                                'in_octets': '90904980749',
                                                                                                'in_ucast_pkts': '120829431',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11665',
                                                                                                'out_octets': '100634487610',
                                                                                                'out_ucast_pkts': '86046862',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11665',
                                                                       'out_octets': '100634487610',
                                                                       'out_ucast_pkts': '86046862',
                                                                       'poller_hash': 'QzgxRTU3cGVlcmludGZwb24gMjA1ODMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205833',
                                                                                               'links': [       {       'connected_time': '688562',
                                                                                                                        'fiber_length': 10406,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTU3cGVlcnBvbnBvbiAyMDU4MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7201 ROAD 140 (SCOTT KRISTINE)',
                                         'poller_hash': 'QzgxRTU3',
                                         'serial_number': 'C81E57',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:6f',
                                         'uptime': '688562'},
                                 {       'counter_source': '206060/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:ae:ae'],
                                                                       'id': '206060/g1',
                                                                       'in_nucast_pkts': '7501',
                                                                       'in_octets': '10094452733',
                                                                       'in_ucast_pkts': '46248377',
                                                                       'mac_address': '84:d3:43:7d:43:42',
                                                                       'mtu': 2000,
                                                                       'name': '206060/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '701966',
                                                                       'out_octets': '293623966425',
                                                                       'out_ucast_pkts': '200498549',
                                                                       'poller_hash': 'QzgxREQxcGVlcmludGYyMDYwNjAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '206060',
                                                                       'in_nucast_pkts': '701966',
                                                                       'in_octets': '293623966425',
                                                                       'in_ucast_pkts': '200498549',
                                                                       'links': [       {       'in_nucast_pkts': '701966',
                                                                                                'in_octets': '293623966425',
                                                                                                'in_ucast_pkts': '200498549',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7501',
                                                                                                'out_octets': '10094452733',
                                                                                                'out_ucast_pkts': '46248377',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7501',
                                                                       'out_octets': '10094452733',
                                                                       'out_ucast_pkts': '46248377',
                                                                       'poller_hash': 'QzgxREQxcGVlcmludGZwb24gMjA2MDYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '206060',
                                                                                               'links': [       {       'connected_time': '340396',
                                                                                                                        'fiber_length': 2882,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREQxcGVlcnBvbnBvbiAyMDYwNjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '152 PUGH ST (BURKE AMY)',
                                         'poller_hash': 'QzgxREQx',
                                         'serial_number': 'C81DD1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:43:3f',
                                         'uptime': '340396'},
                                 {       'counter_source': '206382/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '206382/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '314867426441',
                                                                       'in_ucast_pkts': '3609248501',
                                                                       'mac_address': '84:d3:43:7d:3f:57',
                                                                       'name': '206382/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '22495735825037',
                                                                       'out_ucast_pkts': '15011696532',
                                                                       'poller_hash': 'QzgxRDU0cGVlcmludGYyMDYzODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '206382',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '22495735825037',
                                                                       'in_ucast_pkts': '15011696532',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '22495735825037',
                                                                                                'in_ucast_pkts': '15011696532',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '314867426441',
                                                                                                'out_ucast_pkts': '3609248501',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '314867426441',
                                                                       'out_ucast_pkts': '3609248501',
                                                                       'poller_hash': 'QzgxRDU0cGVlcmludGZwb24gMjA2Mzgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '206382',
                                                                                               'links': [       {       'connected_time': '70945612',
                                                                                                                        'fiber_length': 18606,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDU0cGVlcnBvbnBvbiAyMDYzODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '60 SCHADE ST (ARMSTRONG GERALD & '
                                                 'BARB)',
                                         'poller_hash': 'QzgxRDU0',
                                         'serial_number': 'C81D54',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:3f:57',
                                         'uptime': '70945612'},
                                 {       'counter_source': '206849/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b7:1a'],
                                                                       'id': '206849/g1',
                                                                       'in_nucast_pkts': '14868',
                                                                       'in_octets': '18591604746',
                                                                       'in_ucast_pkts': '141504030',
                                                                       'mac_address': '84:d3:43:7d:47:6a',
                                                                       'mtu': 2000,
                                                                       'name': '206849/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447496',
                                                                       'out_octets': '749823463560',
                                                                       'out_ucast_pkts': '516978312',
                                                                       'poller_hash': 'QzgxRTU2cGVlcmludGYyMDY4NDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '206849',
                                                                       'in_nucast_pkts': '1447496',
                                                                       'in_octets': '749823463560',
                                                                       'in_ucast_pkts': '516978312',
                                                                       'links': [       {       'in_nucast_pkts': '1447496',
                                                                                                'in_octets': '749823463560',
                                                                                                'in_ucast_pkts': '516978312',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14868',
                                                                                                'out_octets': '18591604746',
                                                                                                'out_ucast_pkts': '141504030',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14868',
                                                                       'out_octets': '18591604746',
                                                                       'out_ucast_pkts': '141504030',
                                                                       'poller_hash': 'QzgxRTU2cGVlcmludGZwb24gMjA2ODQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '206849',
                                                                                               'links': [       {       'connected_time': '688433',
                                                                                                                        'fiber_length': 2427,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTU2cGVlcnBvbnBvbiAyMDY4NDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 COBALT STREET S (JUTZI TREVOR)',
                                         'poller_hash': 'QzgxRTU2',
                                         'serial_number': 'C81E56',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:67',
                                         'uptime': '688433'},
                                 {       'counter_source': '206966/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:40:fb'],
                                                                       'id': '206966/g1',
                                                                       'in_nucast_pkts': '9308279',
                                                                       'in_octets': '86671901662',
                                                                       'in_ucast_pkts': '313113393',
                                                                       'mac_address': '84:d3:43:7d:47:9a',
                                                                       'mtu': 2000,
                                                                       'name': '206966/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11080557418',
                                                                       'out_octets': '17170016731886',
                                                                       'out_ucast_pkts': '1588731937',
                                                                       'poller_hash': 'QzgxRTVDcGVlcmludGYyMDY5NjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '206966',
                                                                       'in_nucast_pkts': '11080557418',
                                                                       'in_octets': '17170016731886',
                                                                       'in_ucast_pkts': '1588731937',
                                                                       'links': [       {       'in_nucast_pkts': '11080557418',
                                                                                                'in_octets': '17170016731886',
                                                                                                'in_ucast_pkts': '1588731937',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9308279',
                                                                                                'out_octets': '86671901662',
                                                                                                'out_ucast_pkts': '313113393',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9308279',
                                                                       'out_octets': '86671901662',
                                                                       'out_ucast_pkts': '313113393',
                                                                       'poller_hash': 'QzgxRTVDcGVlcmludGZwb24gMjA2OTY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '206966',
                                                                                               'links': [       {       'connected_time': '10358766',
                                                                                                                        'fiber_length': 14984,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTVDcGVlcnBvbnBvbiAyMDY5NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5610 PERTH LINE 55 (DAVIDSON KELLY)',
                                         'poller_hash': 'QzgxRTVD',
                                         'serial_number': 'C81E5C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:97',
                                         'uptime': '10358766'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '207159/g1',
                                                                       'name': '207159/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRDg2cGVlcmludGYyMDcxNTkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '207159',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRDg2cGVlcmludGZwb24gMjA3MTU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207159',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDg2cGVlcnBvbnBvbiAyMDcxNTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '5 MADDISON ST E (SPEARS BARBARA)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "207159" '
                                                                  '(location: GPON '
                                                                  'port "11/3"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2024-11-30T17:22:27.820Z',
                                         'poller_hash': 'QzgxRDg2',
                                         'serial_number': 'C81D86',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '207179/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '207179/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '84374577947',
                                                                       'in_ucast_pkts': '475568303',
                                                                       'mac_address': '84:d3:43:7d:48:7f',
                                                                       'name': '207179/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '2309581282318',
                                                                       'out_ucast_pkts': '1748389184',
                                                                       'poller_hash': 'QzgxRTc5cGVlcmludGYyMDcxNzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '207179',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '2309581282318',
                                                                       'in_ucast_pkts': '1748389184',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '2309581282318',
                                                                                                'in_ucast_pkts': '1748389184',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '84374577947',
                                                                                                'out_ucast_pkts': '475568303',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '84374577947',
                                                                       'out_ucast_pkts': '475568303',
                                                                       'poller_hash': 'QzgxRTc5cGVlcmludGZwb24gMjA3MTc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207179',
                                                                                               'links': [       {       'connected_time': '538555',
                                                                                                                        'fiber_length': 16429,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTc5cGVlcnBvbnBvbiAyMDcxNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4806 LINE 48 GADSHILL (LAURA '
                                                 'DEWEERD)',
                                         'poller_hash': 'QzgxRTc5',
                                         'serial_number': 'C81E79',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:48:7f',
                                         'uptime': '538555'},
                                 {       'counter_source': '207563/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '207563/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '294',
                                                                       'in_octets': '95531496939',
                                                                       'in_ucast_pkts': '552574613',
                                                                       'mac_address': 'ec:4f:82:0d:6b:0b',
                                                                       'name': '207563/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1220',
                                                                       'out_octets': '2501524407610',
                                                                       'out_ucast_pkts': '1797191512',
                                                                       'poller_hash': 'MjA1RTA2cGVlcmludGYyMDc1NjMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '207563',
                                                                       'in_nucast_pkts': '1220',
                                                                       'in_octets': '2501524407610',
                                                                       'in_ucast_pkts': '1797191512',
                                                                       'links': [       {       'in_nucast_pkts': '1220',
                                                                                                'in_octets': '2501524407610',
                                                                                                'in_ucast_pkts': '1797191512',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '294',
                                                                                                'out_octets': '95531496939',
                                                                                                'out_ucast_pkts': '552574613',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '294',
                                                                       'out_octets': '95531496939',
                                                                       'out_ucast_pkts': '552574613',
                                                                       'poller_hash': 'MjA1RTA2cGVlcmludGZwb24gMjA3NTYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207563',
                                                                                               'links': [       {       'connected_time': '129285',
                                                                                                                        'fiber_length': 1683,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RTA2cGVlcnBvbnBvbiAyMDc1NjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 WILLIAM ST (BARONETTE ROGER)',
                                         'poller_hash': 'MjA1RTA2',
                                         'serial_number': '205E06',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:6b:0b',
                                         'uptime': '129285'},
                                 {       'counter_source': '207840/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '207840/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '225524961380',
                                                                       'in_ucast_pkts': '1974181731',
                                                                       'mac_address': 'f8:85:f9:a8:25:59',
                                                                       'name': '207840/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '149',
                                                                       'out_octets': '12885785410901',
                                                                       'out_ucast_pkts': '8763417876',
                                                                       'poller_hash': 'MTA5NDhGRHBlZXJpbnRmMjA3ODQwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '207840',
                                                                       'in_nucast_pkts': '149',
                                                                       'in_octets': '12885785410901',
                                                                       'in_ucast_pkts': '8763417876',
                                                                       'links': [       {       'in_nucast_pkts': '149',
                                                                                                'in_octets': '12885785410901',
                                                                                                'in_ucast_pkts': '8763417876',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '225524961380',
                                                                                                'out_ucast_pkts': '1974181731',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '225524961380',
                                                                       'out_ucast_pkts': '1974181731',
                                                                       'poller_hash': 'MTA5NDhGRHBlZXJpbnRmcG9uIDIwNzg0MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207840',
                                                                                               'links': [       {       'connected_time': '688475',
                                                                                                                        'fiber_length': 1680,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTA5NDhGRHBlZXJwb25wb24gMjA3ODQw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 WILLIAM STREET APT 207 (FRASER '
                                                 'KATHY)',
                                         'poller_hash': 'MTA5NDhGRA==',
                                         'serial_number': '10948FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:a8:25:59',
                                         'uptime': '688475'},
                                 {       'counter_source': '208658/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '208658/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '417406299765',
                                                                       'in_ucast_pkts': '951380939',
                                                                       'mac_address': '84:d3:43:7d:42:df',
                                                                       'name': '208658/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '4157617099043',
                                                                       'out_ucast_pkts': '3004911992',
                                                                       'poller_hash': 'QzgxREM1cGVlcmludGYyMDg2NTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '208658',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '4157617099043',
                                                                       'in_ucast_pkts': '3004911992',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '4157617099043',
                                                                                                'in_ucast_pkts': '3004911992',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '417406299765',
                                                                                                'out_ucast_pkts': '951380939',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '417406299765',
                                                                       'out_ucast_pkts': '951380939',
                                                                       'poller_hash': 'QzgxREM1cGVlcmludGZwb24gMjA4NjU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208658',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 12065,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREM1cGVlcnBvbnBvbiAyMDg2NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4069 NORFOLK ST (HORST PAULINE)',
                                         'poller_hash': 'QzgxREM1',
                                         'serial_number': 'C81DC5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:42:df',
                                         'uptime': '688369'},
                                 {       'counter_source': '208668/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '208668/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '36883163884',
                                                                       'in_ucast_pkts': '155391641',
                                                                       'mac_address': '84:d3:43:7d:39:3f',
                                                                       'name': '208668/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '584212465730',
                                                                       'out_ucast_pkts': '503315994',
                                                                       'poller_hash': 'QzgxQzkxcGVlcmludGYyMDg2NjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '208668',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '584212465730',
                                                                       'in_ucast_pkts': '503315994',
                                                                       'links': [       {       'in_nucast_pkts': '100',
                                                                                                'in_octets': '584212465730',
                                                                                                'in_ucast_pkts': '503315994',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '36883163884',
                                                                                                'out_ucast_pkts': '155391641',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '36883163884',
                                                                       'out_ucast_pkts': '155391641',
                                                                       'poller_hash': 'QzgxQzkxcGVlcmludGZwb24gMjA4NjY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208668',
                                                                                               'links': [       {       'connected_time': '6576700',
                                                                                                                        'fiber_length': 11779,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxQzkxcGVlcnBvbnBvbiAyMDg2Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (WILLIS PAULINE)',
                                         'poller_hash': 'QzgxQzkx',
                                         'serial_number': 'C81C91',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:39:3f',
                                         'uptime': '6576700'},
                                 {       'counter_source': '208943/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '208943/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '118608755568',
                                                                       'in_ucast_pkts': '26089801693',
                                                                       'mac_address': 'b8:94:70:c7:07:6f',
                                                                       'name': '208943/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '147',
                                                                       'out_octets': '1177131489725',
                                                                       'out_ucast_pkts': '990390808',
                                                                       'poller_hash': 'MTFCMjVGOHBlZXJpbnRmMjA4OTQzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '208943',
                                                                       'in_nucast_pkts': '147',
                                                                       'in_octets': '1177131489725',
                                                                       'in_ucast_pkts': '990390808',
                                                                       'links': [       {       'in_nucast_pkts': '147',
                                                                                                'in_octets': '1177131489725',
                                                                                                'in_ucast_pkts': '990390808',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '118608755568',
                                                                                                'out_ucast_pkts': '26089801693',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '118608755568',
                                                                       'out_ucast_pkts': '26089801693',
                                                                       'poller_hash': 'MTFCMjVGOHBlZXJpbnRmcG9uIDIwODk0Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208943',
                                                                                               'links': [       {       'connected_time': '72951357',
                                                                                                                        'fiber_length': 11763,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCMjVGOHBlZXJwb25wb24gMjA4OTQz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (VAN ALLEN BARBIE)',
                                         'poller_hash': 'MTFCMjVGOA==',
                                         'serial_number': '11B25F8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:c7:07:6f',
                                         'uptime': '72951357'},
                                 {       'counter_source': '209026/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '209026/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '183258536936',
                                                                       'in_ucast_pkts': '945791834',
                                                                       'mac_address': 'b8:94:70:c7:08:3f',
                                                                       'name': '209026/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '3929788040697',
                                                                       'out_ucast_pkts': '2842901744',
                                                                       'poller_hash': 'MTFCMjYxMnBlZXJpbnRmMjA5MDI2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '209026',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '3929788040697',
                                                                       'in_ucast_pkts': '2842901744',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '3929788040697',
                                                                                                'in_ucast_pkts': '2842901744',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '183258536936',
                                                                                                'out_ucast_pkts': '945791834',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '183258536936',
                                                                       'out_ucast_pkts': '945791834',
                                                                       'poller_hash': 'MTFCMjYxMnBlZXJpbnRmcG9uIDIwOTAyNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '209026',
                                                                                               'links': [       {       'connected_time': '688377',
                                                                                                                        'fiber_length': 11816,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCMjYxMnBlZXJwb25wb24gMjA5MDI2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (VAN ALLEN TAUNYA)',
                                         'poller_hash': 'MTFCMjYxMg==',
                                         'serial_number': '11B2612',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:c7:08:3f',
                                         'uptime': '688377'},
                                 {       'counter_source': '209134/g1+209134/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:72:c8'],
                                                                       'id': '209134/g1',
                                                                       'in_nucast_pkts': '5538620',
                                                                       'in_octets': '76599466374',
                                                                       'in_ucast_pkts': '1001102508',
                                                                       'mac_address': '84:d3:43:7d:47:d2',
                                                                       'mtu': 2000,
                                                                       'name': '209134/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '176775157',
                                                                       'out_octets': '4257016632304',
                                                                       'out_ucast_pkts': '2989224780',
                                                                       'poller_hash': 'QzgxRTYzcGVlcmludGYyMDkxMzQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:61:34'],
                                                                       'id': '209134/g2',
                                                                       'in_nucast_pkts': '36241477',
                                                                       'in_octets': '311522802184',
                                                                       'in_ucast_pkts': '271207070',
                                                                       'mac_address': '84:d3:43:7d:47:d3',
                                                                       'mtu': 2000,
                                                                       'name': '209134/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22115204166',
                                                                       'out_octets': '29177599505427',
                                                                       'out_ucast_pkts': '112101870',
                                                                       'poller_hash': 'QzgxRTYzcGVlcmludGYyMDkxMzQvZzI=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '209134',
                                                                       'in_nucast_pkts': '22291979323',
                                                                       'in_octets': '33434616137731',
                                                                       'in_ucast_pkts': '3101326650',
                                                                       'links': [       {       'in_nucast_pkts': '22291979323',
                                                                                                'in_octets': '33434616137731',
                                                                                                'in_ucast_pkts': '3101326650',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '41780097',
                                                                                                'out_octets': '388122268558',
                                                                                                'out_ucast_pkts': '1272309578',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41780097',
                                                                       'out_octets': '388122268558',
                                                                       'out_ucast_pkts': '1272309578',
                                                                       'poller_hash': 'QzgxRTYzcGVlcmludGZwb24gMjA5MTM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '209134',
                                                                                               'links': [       {       'connected_time': '21816835',
                                                                                                                        'fiber_length': 1162,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTYzcGVlcnBvbnBvbiAyMDkxMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 GROSCH ST (WETTLAUFER DWIGHT)',
                                         'poller_hash': 'QzgxRTYz',
                                         'serial_number': 'C81E63',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:cf',
                                         'uptime': '21816835'},
                                 {       'counter_source': '209154/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '209154/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '30345079441',
                                                                       'in_ucast_pkts': '113984043',
                                                                       'mac_address': '84:d3:43:7d:47:af',
                                                                       'name': '209154/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '758894060512',
                                                                       'out_ucast_pkts': '612714245',
                                                                       'poller_hash': 'QzgxRTVGcGVlcmludGYyMDkxNTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '209154',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '758894060512',
                                                                       'in_ucast_pkts': '612714245',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '758894060512',
                                                                                                'in_ucast_pkts': '612714245',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '30345079441',
                                                                                                'out_ucast_pkts': '113984043',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '30345079441',
                                                                       'out_ucast_pkts': '113984043',
                                                                       'poller_hash': 'QzgxRTVGcGVlcmludGZwb24gMjA5MTU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '209154',
                                                                                               'links': [       {       'connected_time': '688409',
                                                                                                                        'fiber_length': 11752,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTVGcGVlcnBvbnBvbiAyMDkxNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (BAKER MARNIE)',
                                         'poller_hash': 'QzgxRTVG',
                                         'serial_number': 'C81E5F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:af',
                                         'uptime': '688409'},
                                 {       'counter_source': '209174/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:ca:e9'],
                                                                       'id': '209174/g1',
                                                                       'in_nucast_pkts': '14435',
                                                                       'in_octets': '5332988487',
                                                                       'in_ucast_pkts': '53935805',
                                                                       'mac_address': '84:d3:43:7d:43:8a',
                                                                       'mtu': 2000,
                                                                       'name': '209174/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464442',
                                                                       'out_octets': '164111846735',
                                                                       'out_ucast_pkts': '113068850',
                                                                       'poller_hash': 'QzgxRERBcGVlcmludGYyMDkxNzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '209174',
                                                                       'in_nucast_pkts': '1464442',
                                                                       'in_octets': '164111846735',
                                                                       'in_ucast_pkts': '113068850',
                                                                       'links': [       {       'in_nucast_pkts': '1464442',
                                                                                                'in_octets': '164111846735',
                                                                                                'in_ucast_pkts': '113068850',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14435',
                                                                                                'out_octets': '5332988487',
                                                                                                'out_ucast_pkts': '53935805',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14435',
                                                                       'out_octets': '5332988487',
                                                                       'out_ucast_pkts': '53935805',
                                                                       'poller_hash': 'QzgxRERBcGVlcmludGZwb24gMjA5MTc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '209174',
                                                                                               'links': [       {       'connected_time': '688367',
                                                                                                                        'fiber_length': 2899,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRERBcGVlcnBvbnBvbiAyMDkxNzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '148 PUGH ST (CARREON ANNE)',
                                         'poller_hash': 'QzgxRERB',
                                         'serial_number': 'C81DDA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:43:87',
                                         'uptime': '688367'},
                                 {       'counter_source': '209558/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '209558/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '459385553050',
                                                                       'in_ucast_pkts': '15123969494',
                                                                       'mac_address': '84:d3:43:7d:46:a7',
                                                                       'name': '209558/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '8323660077760',
                                                                       'out_ucast_pkts': '5845094921',
                                                                       'poller_hash': 'QzgxRTNFcGVlcmludGYyMDk1NTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '209558',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '8323660077760',
                                                                       'in_ucast_pkts': '5845094921',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '8323660077760',
                                                                                                'in_ucast_pkts': '5845094921',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '459385553050',
                                                                                                'out_ucast_pkts': '15123969494',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '459385553050',
                                                                       'out_ucast_pkts': '15123969494',
                                                                       'poller_hash': 'QzgxRTNFcGVlcmludGZwb24gMjA5NTU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '209558',
                                                                                               'links': [       {       'connected_time': '688409',
                                                                                                                        'fiber_length': 11325,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTNFcGVlcnBvbnBvbiAyMDk1NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6932 PERTH ST (DERKSEN ANDREAS)',
                                         'poller_hash': 'QzgxRTNF',
                                         'serial_number': 'C81E3E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:46:a7',
                                         'uptime': '688409'},
                                 {       'counter_source': '210085/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:84:99:cd'],
                                                                       'id': '210085/g1',
                                                                       'in_nucast_pkts': '916665',
                                                                       'in_octets': '4572262513',
                                                                       'in_ucast_pkts': '14000535',
                                                                       'mac_address': 'b8:94:70:c7:06:da',
                                                                       'mtu': 2000,
                                                                       'name': '210085/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149217273',
                                                                       'out_octets': '50650310535',
                                                                       'out_ucast_pkts': '27712829',
                                                                       'poller_hash': 'MTFCMjVFNXBlZXJpbnRmMjEwMDg1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '210085',
                                                                       'in_nucast_pkts': '149217273',
                                                                       'in_octets': '50650310535',
                                                                       'in_ucast_pkts': '27712829',
                                                                       'links': [       {       'in_nucast_pkts': '149217273',
                                                                                                'in_octets': '50650310535',
                                                                                                'in_ucast_pkts': '27712829',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '916665',
                                                                                                'out_octets': '4572262513',
                                                                                                'out_ucast_pkts': '14000535',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '916665',
                                                                       'out_octets': '4572262513',
                                                                       'out_ucast_pkts': '14000535',
                                                                       'poller_hash': 'MTFCMjVFNXBlZXJpbnRmcG9uIDIxMDA4NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '210085',
                                                                                               'links': [       {       'connected_time': '21816086',
                                                                                                                        'fiber_length': 10633,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCMjVFNXBlZXJwb25wb24gMjEwMDg1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5592 DUKE ST (COUNTRY COUSINS)',
                                         'poller_hash': 'MTFCMjVFNQ==',
                                         'serial_number': '11B25E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:c7:06:d7',
                                         'uptime': '21816086'},
                                 {       'counter_source': '211613/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:84:92:69'],
                                                                       'id': '211613/g1',
                                                                       'in_nucast_pkts': '12060',
                                                                       'in_octets': '16182047720',
                                                                       'in_ucast_pkts': '174393578',
                                                                       'mac_address': '84:d3:43:7d:40:e2',
                                                                       'mtu': 2000,
                                                                       'name': '211613/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1497818',
                                                                       'out_octets': '566883551510',
                                                                       'out_ucast_pkts': '391140008',
                                                                       'poller_hash': 'QzgxRDg1cGVlcmludGYyMTE2MTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '211613',
                                                                       'in_nucast_pkts': '1497818',
                                                                       'in_octets': '566883551510',
                                                                       'in_ucast_pkts': '391140008',
                                                                       'links': [       {       'in_nucast_pkts': '1497818',
                                                                                                'in_octets': '566883551510',
                                                                                                'in_ucast_pkts': '391140008',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12060',
                                                                                                'out_octets': '16182047720',
                                                                                                'out_ucast_pkts': '174393578',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12060',
                                                                       'out_octets': '16182047720',
                                                                       'out_ucast_pkts': '174393578',
                                                                       'poller_hash': 'QzgxRDg1cGVlcmludGZwb24gMjExNjEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '211613',
                                                                                               'links': [       {       'connected_time': '688455',
                                                                                                                        'fiber_length': 1337,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDg1cGVlcnBvbnBvbiAyMTE2MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 EDWARD ST MILVERTON (METZGER '
                                                 'TREVOR & BRANDY)',
                                         'poller_hash': 'QzgxRDg1',
                                         'serial_number': 'C81D85',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:40:df',
                                         'uptime': '688455'},
                                 {       'counter_source': '211623/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:ce:dd'],
                                                                       'id': '211623/g1',
                                                                       'in_nucast_pkts': '13656',
                                                                       'in_octets': '12328811283',
                                                                       'in_ucast_pkts': '38714282',
                                                                       'mac_address': '84:d3:43:7d:39:32',
                                                                       'mtu': 2000,
                                                                       'name': '211623/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464342',
                                                                       'out_octets': '214429032528',
                                                                       'out_ucast_pkts': '156795547',
                                                                       'poller_hash': 'QzgxQzhGcGVlcmludGYyMTE2MjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '211623',
                                                                       'in_nucast_pkts': '1464342',
                                                                       'in_octets': '214429032528',
                                                                       'in_ucast_pkts': '156795547',
                                                                       'links': [       {       'in_nucast_pkts': '1464342',
                                                                                                'in_octets': '214429032528',
                                                                                                'in_ucast_pkts': '156795547',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13656',
                                                                                                'out_octets': '12328811283',
                                                                                                'out_ucast_pkts': '38714282',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13656',
                                                                       'out_octets': '12328811283',
                                                                       'out_ucast_pkts': '38714282',
                                                                       'poller_hash': 'QzgxQzhGcGVlcmludGZwb24gMjExNjIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '211623',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 3109,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxQzhGcGVlcnBvbnBvbiAyMTE2MjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '100 PUGH ST (BOIVIN JASON)',
                                         'poller_hash': 'QzgxQzhG',
                                         'serial_number': 'C81C8F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:39:2f',
                                         'uptime': '688370'},
                                 {       'counter_source': '211775/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cd:b4'],
                                                                       'id': '211775/g1',
                                                                       'in_nucast_pkts': '14783',
                                                                       'in_octets': '182813237532',
                                                                       'in_ucast_pkts': '173591980',
                                                                       'mac_address': '84:d3:43:7d:40:f2',
                                                                       'mtu': 2000,
                                                                       'name': '211775/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1463272',
                                                                       'out_octets': '178903361866',
                                                                       'out_ucast_pkts': '193076187',
                                                                       'poller_hash': 'QzgxRDg3cGVlcmludGYyMTE3NzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '211775',
                                                                       'in_nucast_pkts': '1463272',
                                                                       'in_octets': '178903361866',
                                                                       'in_ucast_pkts': '193076187',
                                                                       'links': [       {       'in_nucast_pkts': '1463272',
                                                                                                'in_octets': '178903361866',
                                                                                                'in_ucast_pkts': '193076187',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14783',
                                                                                                'out_octets': '182813237532',
                                                                                                'out_ucast_pkts': '173591980',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14783',
                                                                       'out_octets': '182813237532',
                                                                       'out_ucast_pkts': '173591980',
                                                                       'poller_hash': 'QzgxRDg3cGVlcmludGZwb24gMjExNzc1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '211775',
                                                                                               'links': [       {       'connected_time': '688364',
                                                                                                                        'fiber_length': 3098,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDg3cGVlcnBvbnBvbiAyMTE3NzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '104 PUGH ST (MOHAMED SOMSACK)',
                                         'poller_hash': 'QzgxRDg3',
                                         'serial_number': 'C81D87',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:40:ef',
                                         'uptime': '688364'},
                                 {       'counter_source': '211785/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '211785/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '32453739266',
                                                                       'in_ucast_pkts': '74319388',
                                                                       'mac_address': '84:d3:43:7d:d8:37',
                                                                       'name': '211785/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '312007839850',
                                                                       'out_ucast_pkts': '301622566',
                                                                       'poller_hash': 'QzgzMDcwcGVlcmludGYyMTE3ODUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '211785',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '312007839850',
                                                                       'in_ucast_pkts': '301622566',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '312007839850',
                                                                                                'in_ucast_pkts': '301622566',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '32453739266',
                                                                                                'out_ucast_pkts': '74319388',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '32453739266',
                                                                       'out_ucast_pkts': '74319388',
                                                                       'poller_hash': 'QzgzMDcwcGVlcmludGZwb24gMjExNzg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '211785',
                                                                                               'links': [       {       'connected_time': '21816130',
                                                                                                                        'fiber_length': 8705,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMDcwcGVlcnBvbnBvbiAyMTE3ODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4838 LINE 59 (WAGLER LEN)',
                                         'poller_hash': 'QzgzMDcw',
                                         'serial_number': 'C83070',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:d8:37',
                                         'uptime': '21816130'},
                                 {       'counter_source': '211805/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cb:d0'],
                                                                       'id': '211805/g1',
                                                                       'in_nucast_pkts': '13859',
                                                                       'in_octets': '13202304341',
                                                                       'in_ucast_pkts': '35863505',
                                                                       'mac_address': 'b8:94:70:ba:e2:a8',
                                                                       'mtu': 2000,
                                                                       'name': '211805/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464368',
                                                                       'out_octets': '106459896686',
                                                                       'out_ucast_pkts': '83688118',
                                                                       'poller_hash': 'MTE5NTE2OXBlZXJpbnRmMjExODA1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '211805',
                                                                       'in_nucast_pkts': '1464368',
                                                                       'in_octets': '106459896686',
                                                                       'in_ucast_pkts': '83688118',
                                                                       'links': [       {       'in_nucast_pkts': '1464368',
                                                                                                'in_octets': '106459896686',
                                                                                                'in_ucast_pkts': '83688118',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13859',
                                                                                                'out_octets': '13202304341',
                                                                                                'out_ucast_pkts': '35863505',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13859',
                                                                       'out_octets': '13202304341',
                                                                       'out_ucast_pkts': '35863505',
                                                                       'poller_hash': 'MTE5NTE2OXBlZXJpbnRmcG9uIDIxMTgwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '211805',
                                                                                               'links': [       {       'connected_time': '688371',
                                                                                                                        'fiber_length': 2582,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE2OXBlZXJwb25wb24gMjExODA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '95 PUGH ST (ROOPRAM ANTHONY)',
                                         'poller_hash': 'MTE5NTE2OQ==',
                                         'serial_number': '1195169',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e2:a5',
                                         'uptime': '688371'},
                                 {       'counter_source': '212196/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e1:ac'],
                                                                       'id': '212196/g1',
                                                                       'in_nucast_pkts': '12508',
                                                                       'in_octets': '4071350087',
                                                                       'in_ucast_pkts': '12222533',
                                                                       'mac_address': '84:d3:43:7d:46:ba',
                                                                       'mtu': 2000,
                                                                       'name': '212196/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447322',
                                                                       'out_octets': '45757712199',
                                                                       'out_ucast_pkts': '37642384',
                                                                       'poller_hash': 'QzgxRTQwcGVlcmludGYyMTIxOTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '212196',
                                                                       'in_nucast_pkts': '1447322',
                                                                       'in_octets': '45757712199',
                                                                       'in_ucast_pkts': '37642384',
                                                                       'links': [       {       'in_nucast_pkts': '1447322',
                                                                                                'in_octets': '45757712199',
                                                                                                'in_ucast_pkts': '37642384',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12508',
                                                                                                'out_octets': '4071350087',
                                                                                                'out_ucast_pkts': '12222533',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12508',
                                                                       'out_octets': '4071350087',
                                                                       'out_ucast_pkts': '12222533',
                                                                       'poller_hash': 'QzgxRTQwcGVlcmludGZwb24gMjEyMTk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212196',
                                                                                               'links': [       {       'connected_time': '688409',
                                                                                                                        'fiber_length': 1980,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTQwcGVlcnBvbnBvbiAyMTIxOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '92 GERBER DR (MEYER JACKIE)',
                                         'poller_hash': 'QzgxRTQw',
                                         'serial_number': 'C81E40',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:46:b7',
                                         'uptime': '688409'},
                                 {       'counter_source': '212206/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212206/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12756133463',
                                                                       'in_ucast_pkts': '47585746',
                                                                       'mac_address': 'b8:94:70:c6:ff:67',
                                                                       'name': '212206/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '173997181358',
                                                                       'out_ucast_pkts': '153966442',
                                                                       'poller_hash': 'MTFCMjRGN3BlZXJpbnRmMjEyMjA2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212206',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '173997181358',
                                                                       'in_ucast_pkts': '153966442',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '173997181358',
                                                                                                'in_ucast_pkts': '153966442',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '12756133463',
                                                                                                'out_ucast_pkts': '47585746',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12756133463',
                                                                       'out_ucast_pkts': '47585746',
                                                                       'poller_hash': 'MTFCMjRGN3BlZXJpbnRmcG9uIDIxMjIwNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212206',
                                                                                               'links': [       {       'connected_time': '511297',
                                                                                                                        'fiber_length': 11246,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCMjRGN3BlZXJwb25wb24gMjEyMjA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4075 PRINCESS ST (MISENER SCOTT)',
                                         'poller_hash': 'MTFCMjRGNw==',
                                         'serial_number': '11B24F7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:c6:ff:67',
                                         'uptime': '511297'},
                                 {       'counter_source': '212416/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:8d:3c'],
                                                                       'id': '212416/g1',
                                                                       'in_nucast_pkts': '12117',
                                                                       'in_octets': '13949976607',
                                                                       'in_ucast_pkts': '49923038',
                                                                       'mac_address': 'b8:94:70:ba:e1:58',
                                                                       'mtu': 2000,
                                                                       'name': '212416/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464236',
                                                                       'out_octets': '157425846690',
                                                                       'out_ucast_pkts': '129190380',
                                                                       'poller_hash': 'MTE5NTEzRnBlZXJpbnRmMjEyNDE2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '212416',
                                                                       'in_nucast_pkts': '1464236',
                                                                       'in_octets': '157425846690',
                                                                       'in_ucast_pkts': '129190380',
                                                                       'links': [       {       'in_nucast_pkts': '1464236',
                                                                                                'in_octets': '157425846690',
                                                                                                'in_ucast_pkts': '129190380',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12117',
                                                                                                'out_octets': '13949976607',
                                                                                                'out_ucast_pkts': '49923038',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12117',
                                                                       'out_octets': '13949976607',
                                                                       'out_ucast_pkts': '49923038',
                                                                       'poller_hash': 'MTE5NTEzRnBlZXJpbnRmcG9uIDIxMjQxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212416',
                                                                                               'links': [       {       'connected_time': '688354',
                                                                                                                        'fiber_length': 11624,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTEzRnBlZXJwb25wb24gMjEyNDE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6985 RUTHERFORD ST (NIEWINSKI DAVE)',
                                         'poller_hash': 'MTE5NTEzRg==',
                                         'serial_number': '119513F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e1:55',
                                         'uptime': '688354'},
                                 {       'counter_source': '212457/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212457/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '64411257135',
                                                                       'in_ucast_pkts': '272338851',
                                                                       'mac_address': 'b8:94:70:c6:eb:7f',
                                                                       'name': '212457/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1284120723732',
                                                                       'out_ucast_pkts': '898851951',
                                                                       'poller_hash': 'MTFCMjI3QXBlZXJpbnRmMjEyNDU3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212457',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1284120723732',
                                                                       'in_ucast_pkts': '898851951',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1284120723732',
                                                                                                'in_ucast_pkts': '898851951',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '64411257135',
                                                                                                'out_ucast_pkts': '272338851',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '64411257135',
                                                                       'out_ucast_pkts': '272338851',
                                                                       'poller_hash': 'MTFCMjI3QXBlZXJpbnRmcG9uIDIxMjQ1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212457',
                                                                                               'links': [       {       'connected_time': '16596261',
                                                                                                                        'fiber_length': 11480,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCMjI3QXBlZXJwb25wb24gMjEyNDU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4072 QUEEN ST (THOMAS JIM)',
                                         'poller_hash': 'MTFCMjI3QQ==',
                                         'serial_number': '11B227A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:c6:eb:7f',
                                         'uptime': '16596261'},
                                 {       'counter_source': '212500/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:93:fb'],
                                                                       'id': '212500/g1',
                                                                       'in_nucast_pkts': '538038',
                                                                       'in_octets': '21412155682',
                                                                       'in_ucast_pkts': '138812977',
                                                                       'mac_address': 'b8:94:70:c6:ed:8a',
                                                                       'mtu': 2000,
                                                                       'name': '212500/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67889855',
                                                                       'out_octets': '663764566871',
                                                                       'out_ucast_pkts': '460461849',
                                                                       'poller_hash': 'MTFCMjJCQnBlZXJpbnRmMjEyNTAwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '212500',
                                                                       'in_nucast_pkts': '67889855',
                                                                       'in_octets': '663764566871',
                                                                       'in_ucast_pkts': '460461849',
                                                                       'links': [       {       'in_nucast_pkts': '67889855',
                                                                                                'in_octets': '663764566871',
                                                                                                'in_ucast_pkts': '460461849',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '538038',
                                                                                                'out_octets': '21412155682',
                                                                                                'out_ucast_pkts': '138812977',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '538038',
                                                                       'out_octets': '21412155682',
                                                                       'out_ucast_pkts': '138812977',
                                                                       'poller_hash': 'MTFCMjJCQnBlZXJpbnRmcG9uIDIxMjUwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212500',
                                                                                               'links': [       {       'connected_time': '30265335',
                                                                                                                        'fiber_length': 11741,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCMjJCQnBlZXJwb25wb24gMjEyNTAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (ROES SHARON)',
                                         'poller_hash': 'MTFCMjJCQg==',
                                         'serial_number': '11B22BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:c6:ed:87',
                                         'uptime': '30265335'},
                                 {       'counter_source': '212520/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212520/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '191991797699',
                                                                       'in_ucast_pkts': '964950917',
                                                                       'mac_address': 'b8:94:70:ba:e2:ed',
                                                                       'name': '212520/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '120',
                                                                       'out_octets': '4168810036943',
                                                                       'out_ucast_pkts': '3153060804',
                                                                       'poller_hash': 'MTE5NTE3MnBlZXJpbnRmMjEyNTIwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212520',
                                                                       'in_nucast_pkts': '120',
                                                                       'in_octets': '4168810036943',
                                                                       'in_ucast_pkts': '3153060804',
                                                                       'links': [       {       'in_nucast_pkts': '120',
                                                                                                'in_octets': '4168810036943',
                                                                                                'in_ucast_pkts': '3153060804',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '191991797699',
                                                                                                'out_ucast_pkts': '964950917',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '191991797699',
                                                                       'out_ucast_pkts': '964950917',
                                                                       'poller_hash': 'MTE5NTE3MnBlZXJpbnRmcG9uIDIxMjUyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212520',
                                                                                               'links': [       {       'connected_time': '22282804',
                                                                                                                        'fiber_length': 11756,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE3MnBlZXJwb25wb24gMjEyNTIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (FRIESEN HELENA & '
                                                 'JACOB)',
                                         'poller_hash': 'MTE5NTE3Mg==',
                                         'serial_number': '1195172',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e2:ed',
                                         'uptime': '22282804'},
                                 {       'counter_source': '212600/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212600/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '181851289884',
                                                                       'in_ucast_pkts': '44201394582',
                                                                       'mac_address': 'b8:94:70:ba:e1:4d',
                                                                       'name': '212600/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '6689982025236',
                                                                       'out_ucast_pkts': '13201225051',
                                                                       'poller_hash': 'MTE5NTEzRXBlZXJpbnRmMjEyNjAwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212600',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '6689982025236',
                                                                       'in_ucast_pkts': '13201225051',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '6689982025236',
                                                                                                'in_ucast_pkts': '13201225051',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '181851289884',
                                                                                                'out_ucast_pkts': '44201394582',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '181851289884',
                                                                       'out_ucast_pkts': '44201394582',
                                                                       'poller_hash': 'MTE5NTEzRXBlZXJpbnRmcG9uIDIxMjYwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212600',
                                                                                               'links': [       {       'connected_time': '26632408',
                                                                                                                        'fiber_length': 11334,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTEzRXBlZXJwb25wb24gMjEyNjAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4095 WYNDHAM ST (SCHEERER JOHN)',
                                         'poller_hash': 'MTE5NTEzRQ==',
                                         'serial_number': '119513E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e1:4d',
                                         'uptime': '26632408'},
                                 {       'counter_source': '212710/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212710/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '346604862775',
                                                                       'in_ucast_pkts': '168676630770',
                                                                       'mac_address': 'b8:94:70:ba:e2:3d',
                                                                       'name': '212710/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '49',
                                                                       'out_octets': '5935049359061',
                                                                       'out_ucast_pkts': '4347984767',
                                                                       'poller_hash': 'MTE5NTE1Q3BlZXJpbnRmMjEyNzEwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212710',
                                                                       'in_nucast_pkts': '49',
                                                                       'in_octets': '5935049359061',
                                                                       'in_ucast_pkts': '4347984767',
                                                                       'links': [       {       'in_nucast_pkts': '49',
                                                                                                'in_octets': '5935049359061',
                                                                                                'in_ucast_pkts': '4347984767',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '346604862775',
                                                                                                'out_ucast_pkts': '168676630770',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '346604862775',
                                                                       'out_ucast_pkts': '168676630770',
                                                                       'poller_hash': 'MTE5NTE1Q3BlZXJpbnRmcG9uIDIxMjcxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212710',
                                                                                               'links': [       {       'connected_time': '27989217',
                                                                                                                        'fiber_length': 12265,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE1Q3BlZXJwb25wb24gMjEyNzEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5479 ROAD 144 (DEGROOT CHRIS & '
                                                 'ANGELA)',
                                         'poller_hash': 'MTE5NTE1Qw==',
                                         'serial_number': '119515C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e2:3d',
                                         'uptime': '27989217'},
                                 {       'counter_source': '212720/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212720/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '114189082198',
                                                                       'in_ucast_pkts': '364936847',
                                                                       'mac_address': 'b8:94:70:ba:cb:95',
                                                                       'name': '212720/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '1834558033898',
                                                                       'out_ucast_pkts': '1488736010',
                                                                       'poller_hash': 'MTE5NEU4N3BlZXJpbnRmMjEyNzIwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212720',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '1834558033898',
                                                                       'in_ucast_pkts': '1488736010',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '1834558033898',
                                                                                                'in_ucast_pkts': '1488736010',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '114189082198',
                                                                                                'out_ucast_pkts': '364936847',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '114189082198',
                                                                       'out_ucast_pkts': '364936847',
                                                                       'poller_hash': 'MTE5NEU4N3BlZXJpbnRmcG9uIDIxMjcyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212720',
                                                                                               'links': [       {       'connected_time': '21792971',
                                                                                                                        'fiber_length': 11745,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NEU4N3BlZXJwb25wb24gMjEyNzIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6986 MILLBANK MAIN ST (KLASSEN JAKE '
                                                 '& ANNA)',
                                         'poller_hash': 'MTE5NEU4Nw==',
                                         'serial_number': '1194E87',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:cb:95',
                                         'uptime': '21792971'},
                                 {       'counter_source': '212790/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212790/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '691411647',
                                                                       'in_ucast_pkts': '734229',
                                                                       'mac_address': 'b8:94:70:ba:e3:c5',
                                                                       'name': '212790/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '12722487299',
                                                                       'out_ucast_pkts': '54612531',
                                                                       'poller_hash': 'MTE5NTE4RHBlZXJpbnRmMjEyNzkwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212790',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '12722487299',
                                                                       'in_ucast_pkts': '54612531',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '12722487299',
                                                                                                'in_ucast_pkts': '54612531',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '691411647',
                                                                                                'out_ucast_pkts': '734229',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '691411647',
                                                                       'out_ucast_pkts': '734229',
                                                                       'poller_hash': 'MTE5NTE4RHBlZXJpbnRmcG9uIDIxMjc5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212790',
                                                                                               'links': [       {       'connected_time': '21814690',
                                                                                                                        'fiber_length': 11253,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE4RHBlZXJwb25wb24gMjEyNzkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4095 HILLCREST ST (FRIESEN HEINRICH '
                                                 '& EVA)',
                                         'poller_hash': 'MTE5NTE4RA==',
                                         'serial_number': '119518D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e3:c5',
                                         'uptime': '21814690'},
                                 {       'counter_source': '212869/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:94:a0'],
                                                                       'id': '212869/g1',
                                                                       'in_nucast_pkts': '13301',
                                                                       'in_octets': '20412817777',
                                                                       'in_ucast_pkts': '46510262',
                                                                       'mac_address': '60:db:98:8f:03:d9',
                                                                       'mtu': 2000,
                                                                       'name': '212869/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1447207',
                                                                       'out_octets': '219361557247',
                                                                       'out_ucast_pkts': '163498481',
                                                                       'poller_hash': 'QjUxMkI3cGVlcmludGYyMTI4NjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '212869',
                                                                       'in_nucast_pkts': '1447207',
                                                                       'in_octets': '219361557247',
                                                                       'in_ucast_pkts': '163498481',
                                                                       'links': [       {       'in_nucast_pkts': '1447207',
                                                                                                'in_octets': '219361557247',
                                                                                                'in_ucast_pkts': '163498481',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13301',
                                                                                                'out_octets': '20412817777',
                                                                                                'out_ucast_pkts': '46510262',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13301',
                                                                       'out_octets': '20412817777',
                                                                       'out_ucast_pkts': '46510262',
                                                                       'poller_hash': 'QjUxMkI3cGVlcmludGZwb24gMjEyODY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212869',
                                                                                               'links': [       {       'connected_time': '688428',
                                                                                                                        'fiber_length': 2230,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMkI3cGVlcnBvbnBvbiAyMTI4Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '77 GERBER DR (SZCZEPANEK KRZYSZTOF '
                                                 '(KRIS))',
                                         'poller_hash': 'QjUxMkI3',
                                         'serial_number': 'B512B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:03:d9',
                                         'uptime': '688428'},
                                 {       'counter_source': '212936/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212936/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '513613571027',
                                                                       'in_ucast_pkts': '80195864982',
                                                                       'mac_address': 'b8:94:70:ba:e2:d5',
                                                                       'name': '212936/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '186',
                                                                       'out_octets': '13042241546499',
                                                                       'out_ucast_pkts': '9883573946',
                                                                       'poller_hash': 'MTE5NTE2RnBlZXJpbnRmMjEyOTM2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212936',
                                                                       'in_nucast_pkts': '186',
                                                                       'in_octets': '13042241546499',
                                                                       'in_ucast_pkts': '9883573946',
                                                                       'links': [       {       'in_nucast_pkts': '186',
                                                                                                'in_octets': '13042241546499',
                                                                                                'in_ucast_pkts': '9883573946',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '513613571027',
                                                                                                'out_ucast_pkts': '80195864982',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '513613571027',
                                                                       'out_ucast_pkts': '80195864982',
                                                                       'poller_hash': 'MTE5NTE2RnBlZXJpbnRmcG9uIDIxMjkzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212936',
                                                                                               'links': [       {       'connected_time': '2137841',
                                                                                                                        'fiber_length': 10917,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE2RnBlZXJwb25wb24gMjEyOTM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6944 PERTH ROAD 121 (BARTSCH JOHN & '
                                                 'SUSANA)',
                                         'poller_hash': 'MTE5NTE2Rg==',
                                         'serial_number': '119516F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e2:d5',
                                         'uptime': '2137841'},
                                 {       'counter_source': '213517/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '213517/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '201574888377',
                                                                       'in_ucast_pkts': '56876949483',
                                                                       'mac_address': '84:d3:43:95:8a:0a',
                                                                       'name': '213517/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '5857435651419',
                                                                       'out_ucast_pkts': '4131900180',
                                                                       'poller_hash': 'Q0JEMTYwcGVlcmludGYyMTM1MTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '213517',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '5857435651419',
                                                                       'in_ucast_pkts': '4131900180',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '5857435651419',
                                                                                                'in_ucast_pkts': '4131900180',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '201574888377',
                                                                                                'out_ucast_pkts': '56876949483',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '201574888377',
                                                                       'out_ucast_pkts': '56876949483',
                                                                       'poller_hash': 'Q0JEMTYwcGVlcmludGZwb24gMjEzNTE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213517',
                                                                                               'links': [       {       'connected_time': '28863',
                                                                                                                        'fiber_length': 12028,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTYwcGVlcnBvbnBvbiAyMTM1MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6964 MILLBANK MAIN ST (DORSCHT '
                                                 'RANDY)',
                                         'poller_hash': 'Q0JEMTYw',
                                         'serial_number': 'CBD160',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8a:0a',
                                         'uptime': '28863'},
                                 {       'counter_source': '213651/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:8f:f1'],
                                                                       'id': '213651/g1',
                                                                       'in_nucast_pkts': '11499',
                                                                       'in_octets': '10301949649',
                                                                       'in_ucast_pkts': '67349484',
                                                                       'mac_address': 'b8:94:70:ba:e1:80',
                                                                       'mtu': 2000,
                                                                       'name': '213651/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1120757',
                                                                       'out_octets': '220684922901',
                                                                       'out_ucast_pkts': '176937005',
                                                                       'poller_hash': 'MTE5NTE0NHBlZXJpbnRmMjEzNjUxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '213651',
                                                                       'in_nucast_pkts': '1120757',
                                                                       'in_octets': '220684922901',
                                                                       'in_ucast_pkts': '176937005',
                                                                       'links': [       {       'in_nucast_pkts': '1120757',
                                                                                                'in_octets': '220684922901',
                                                                                                'in_ucast_pkts': '176937005',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11499',
                                                                                                'out_octets': '10301949649',
                                                                                                'out_ucast_pkts': '67349484',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11499',
                                                                       'out_octets': '10301949649',
                                                                       'out_ucast_pkts': '67349484',
                                                                       'poller_hash': 'MTE5NTE0NHBlZXJpbnRmcG9uIDIxMzY1MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213651',
                                                                                               'links': [       {       'connected_time': '538562',
                                                                                                                        'fiber_length': 14960,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE0NHBlZXJwb25wb24gMjEzNjUx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5201 LINE 49 (NOON JOSHUA)',
                                         'poller_hash': 'MTE5NTE0NA==',
                                         'serial_number': '1195144',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e1:7d',
                                         'uptime': '538562'},
                                 {       'counter_source': '213681/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:a7:bf'],
                                                                       'id': '213681/g1',
                                                                       'in_nucast_pkts': '642518',
                                                                       'in_octets': '213393183253',
                                                                       'in_ucast_pkts': '816642142',
                                                                       'mac_address': '84:d3:43:7d:0b:ea',
                                                                       'mtu': 2000,
                                                                       'name': '213681/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '89185721',
                                                                       'out_octets': '5820353024167',
                                                                       'out_ucast_pkts': '4633841659',
                                                                       'poller_hash': 'QzgxNkU2cGVlcmludGYyMTM2ODEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '213681',
                                                                       'in_nucast_pkts': '89185721',
                                                                       'in_octets': '5820353024167',
                                                                       'in_ucast_pkts': '4633841659',
                                                                       'links': [       {       'in_nucast_pkts': '89185721',
                                                                                                'in_octets': '5820353024167',
                                                                                                'in_ucast_pkts': '4633841659',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '642518',
                                                                                                'out_octets': '213393183253',
                                                                                                'out_ucast_pkts': '816642142',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '642518',
                                                                       'out_octets': '213393183253',
                                                                       'out_ucast_pkts': '816642142',
                                                                       'poller_hash': 'QzgxNkU2cGVlcmludGZwb24gMjEzNjgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213681',
                                                                                               'links': [       {       'connected_time': '70485238',
                                                                                                                        'fiber_length': 7294,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxNkU2cGVlcnBvbnBvbiAyMTM2ODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5125 LINE 60 (EBERSOL ISRAEL & '
                                                 'CHRISTENA)',
                                         'poller_hash': 'QzgxNkU2',
                                         'serial_number': 'C816E6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:0b:e7',
                                         'uptime': '70485238'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '213780/g1',
                                                                       'name': '213780/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxREY0cGVlcmludGYyMTM3ODAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '213780',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxREY0cGVlcmludGZwb24gMjEzNzgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213780',
                                                                                               'links': [       {       'connected_time': '20994506',
                                                                                                                        'fiber_length': 11969,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREY0cGVlcnBvbnBvbiAyMTM3ODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4062 NORFOLK ST (GUENTHER DAVID)',
                                         'poller_hash': 'QzgxREY0',
                                         'serial_number': 'C81DF4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:57',
                                         'uptime': '20994506'},
                                 {       'counter_source': '213790/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:df:f4'],
                                                                       'id': '213790/g1',
                                                                       'in_nucast_pkts': '14799',
                                                                       'in_octets': '9565012653',
                                                                       'in_ucast_pkts': '76138734',
                                                                       'mac_address': 'b8:94:70:ba:e3:50',
                                                                       'mtu': 2000,
                                                                       'name': '213790/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464742',
                                                                       'out_octets': '289507400617',
                                                                       'out_ucast_pkts': '206715935',
                                                                       'poller_hash': 'MTE5NTE3RXBlZXJpbnRmMjEzNzkwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '213790',
                                                                       'in_nucast_pkts': '1464742',
                                                                       'in_octets': '289507400617',
                                                                       'in_ucast_pkts': '206715935',
                                                                       'links': [       {       'in_nucast_pkts': '1464742',
                                                                                                'in_octets': '289507400617',
                                                                                                'in_ucast_pkts': '206715935',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14799',
                                                                                                'out_octets': '9565012653',
                                                                                                'out_ucast_pkts': '76138734',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14799',
                                                                       'out_octets': '9565012653',
                                                                       'out_ucast_pkts': '76138734',
                                                                       'poller_hash': 'MTE5NTE3RXBlZXJpbnRmcG9uIDIxMzc5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213790',
                                                                                               'links': [       {       'connected_time': '688393',
                                                                                                                        'fiber_length': 11818,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE3RXBlZXJwb25wb24gMjEzNzkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6935 HORIZON TL (MCARTHUR DOROTHY & '
                                                 'DOUG)',
                                         'poller_hash': 'MTE5NTE3RQ==',
                                         'serial_number': '119517E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e3:4d',
                                         'uptime': '688393'},
                                 {       'counter_source': '213913/F1+213913/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2f:c6:e4'],
                                                                       'id': '213913/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '17178265318',
                                                                       'in_ucast_pkts': '37755074',
                                                                       'name': '213913/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '12110898620',
                                                                       'out_octets': '16062056666504',
                                                                       'out_ucast_pkts': '4095374',
                                                                       'poller_hash': 'MTA5NDhEMXBlZXJpbnRmMjEzOTEzL0Yx'},
                                                               {       'admin_up': True,
                                                                       'id': '213913/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '116346886797',
                                                                       'in_ucast_pkts': '1389504107',
                                                                       'mac_address': 'f8:85:f9:a8:23:cd',
                                                                       'name': '213913/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '182',
                                                                       'out_octets': '6805884773546',
                                                                       'out_ucast_pkts': '4564733013',
                                                                       'poller_hash': 'MTA5NDhEMXBlZXJpbnRmMjEzOTEzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '213913',
                                                                       'in_nucast_pkts': '12110898802',
                                                                       'in_octets': '22867941440050',
                                                                       'in_ucast_pkts': '4568828387',
                                                                       'links': [       {       'in_nucast_pkts': '12110898802',
                                                                                                'in_octets': '22867941440050',
                                                                                                'in_ucast_pkts': '4568828387',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '133525152115',
                                                                                                'out_ucast_pkts': '1427259181',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '133525152115',
                                                                       'out_ucast_pkts': '1427259181',
                                                                       'poller_hash': 'MTA5NDhEMXBlZXJpbnRmcG9uIDIxMzkxMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213913',
                                                                                               'links': [       {       'connected_time': '688504',
                                                                                                                        'fiber_length': 607,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTA5NDhEMXBlZXJwb25wb24gMjEzOTEz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '10 MAIN ST N (JACOBS CHAD)',
                                         'poller_hash': 'MTA5NDhEMQ==',
                                         'serial_number': '10948D1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:a8:23:cd',
                                         'uptime': '688504'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '213934/g1',
                                                                       'name': '213934/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE5NTEwNXBlZXJpbnRmMjEzOTM0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '213934',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE5NTEwNXBlZXJpbnRmcG9uIDIxMzkzNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213934',
                                                                                               'links': [       {       'connected_time': '20916304',
                                                                                                                        'fiber_length': 11791,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTEwNXBlZXJwb25wb24gMjEzOTM0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (ZEHR ALMA)',
                                         'poller_hash': 'MTE5NTEwNQ==',
                                         'serial_number': '1195105',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:df:85',
                                         'uptime': '20916304'},
                                 {       'counter_source': '215021/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '215021/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '488343378505',
                                                                       'in_ucast_pkts': '69583600077',
                                                                       'mac_address': 'b8:94:70:ba:e1:8d',
                                                                       'name': '215021/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '3065258784324',
                                                                       'out_ucast_pkts': '2395717967',
                                                                       'poller_hash': 'MTE5NTE0NnBlZXJpbnRmMjE1MDIxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '215021',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '3065258784324',
                                                                       'in_ucast_pkts': '2395717967',
                                                                       'links': [       {       'in_nucast_pkts': '148',
                                                                                                'in_octets': '3065258784324',
                                                                                                'in_ucast_pkts': '2395717967',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '488343378505',
                                                                                                'out_ucast_pkts': '69583600077',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '488343378505',
                                                                       'out_ucast_pkts': '69583600077',
                                                                       'poller_hash': 'MTE5NTE0NnBlZXJpbnRmcG9uIDIxNTAyMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215021',
                                                                                               'links': [       {       'connected_time': '12765140',
                                                                                                                        'fiber_length': 11644,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE0NnBlZXJwb25wb24gMjE1MDIx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6983 MILLBANK MAIN ST (KROEKER '
                                                 'ISAAK)',
                                         'poller_hash': 'MTE5NTE0Ng==',
                                         'serial_number': '1195146',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e1:8d',
                                         'uptime': '12765140'},
                                 {       'counter_source': '215031/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e4:f0'],
                                                                       'id': '215031/g1',
                                                                       'in_nucast_pkts': '925310',
                                                                       'in_octets': '1145224937597',
                                                                       'in_ucast_pkts': '4822153098',
                                                                       'mac_address': 'b8:94:70:ba:e4:28',
                                                                       'mtu': 2000,
                                                                       'name': '215031/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '137784697',
                                                                       'out_octets': '15441594055366',
                                                                       'out_ucast_pkts': '11455853966',
                                                                       'poller_hash': 'MTE5NTE5OXBlZXJpbnRmMjE1MDMxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '215031',
                                                                       'in_nucast_pkts': '137784697',
                                                                       'in_octets': '15441594055366',
                                                                       'in_ucast_pkts': '11455853966',
                                                                       'links': [       {       'in_nucast_pkts': '137784697',
                                                                                                'in_octets': '15441594055366',
                                                                                                'in_ucast_pkts': '11455853966',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '925310',
                                                                                                'out_octets': '1145224937597',
                                                                                                'out_ucast_pkts': '4822153098',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '925310',
                                                                       'out_octets': '1145224937597',
                                                                       'out_ucast_pkts': '4822153098',
                                                                       'poller_hash': 'MTE5NTE5OXBlZXJpbnRmcG9uIDIxNTAzMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215031',
                                                                                               'links': [       {       'connected_time': '19970731',
                                                                                                                        'fiber_length': 11916,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE5OXBlZXJwb25wb24gMjE1MDMx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6963 MILLBANK MAIN STREET (GOMES '
                                                 'JOHN & NELIA)',
                                         'poller_hash': 'MTE5NTE5OQ==',
                                         'serial_number': '1195199',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e4:25',
                                         'uptime': '19970731'},
                                 {       'counter_source': '215161/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '215161/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '57200200206',
                                                                       'in_ucast_pkts': '279638401',
                                                                       'mac_address': '1c:8b:76:00:37:c1',
                                                                       'name': '215161/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '949201525038',
                                                                       'out_ucast_pkts': '814805260',
                                                                       'poller_hash': 'MTU4RkI2RXBlZXJpbnRmMjE1MTYxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '215161',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '949201525038',
                                                                       'in_ucast_pkts': '814805260',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '949201525038',
                                                                                                'in_ucast_pkts': '814805260',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '57200200206',
                                                                                                'out_ucast_pkts': '279638401',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '57200200206',
                                                                       'out_ucast_pkts': '279638401',
                                                                       'poller_hash': 'MTU4RkI2RXBlZXJpbnRmcG9uIDIxNTE2MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215161',
                                                                                               'links': [       {       'connected_time': '29301725',
                                                                                                                        'fiber_length': 10515,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.494,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU4RkI2RXBlZXJwb25wb24gMjE1MTYx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '5592 DUKE STREET UNIT C (SEW-WEAR '
                                                 'MFG CO LTD)',
                                         'poller_hash': 'MTU4RkI2RQ==',
                                         'serial_number': '158FB6E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '1c:8b:76:00:37:c1',
                                         'uptime': '29301725'},
                                 {       'counter_source': '215641/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:90:d8'],
                                                                       'id': '215641/g1',
                                                                       'in_nucast_pkts': '2723386',
                                                                       'in_octets': '342791430444',
                                                                       'in_ucast_pkts': '399225449',
                                                                       'mac_address': 'b8:94:70:ba:e4:88',
                                                                       'mtu': 2000,
                                                                       'name': '215641/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4362397818',
                                                                       'out_octets': '6308977492132',
                                                                       'out_ucast_pkts': '322244571',
                                                                       'poller_hash': 'MTE5NTFBNXBlZXJpbnRmMjE1NjQxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '215641',
                                                                       'in_nucast_pkts': '4362397818',
                                                                       'in_octets': '6308977492132',
                                                                       'in_ucast_pkts': '322244571',
                                                                       'links': [       {       'in_nucast_pkts': '4362397818',
                                                                                                'in_octets': '6308977492132',
                                                                                                'in_ucast_pkts': '322244571',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2723386',
                                                                                                'out_octets': '342791430444',
                                                                                                'out_ucast_pkts': '399225449',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2723386',
                                                                       'out_octets': '342791430444',
                                                                       'out_ucast_pkts': '399225449',
                                                                       'poller_hash': 'MTE5NTFBNXBlZXJpbnRmcG9uIDIxNTY0MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215641',
                                                                                               'links': [       {       'connected_time': '3114073',
                                                                                                                        'fiber_length': 15649,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTFBNXBlZXJwb25wb24gMjE1NjQx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5657 PERTH LINE 55 (HAHN BRUCE)',
                                         'poller_hash': 'MTE5NTFBNQ==',
                                         'serial_number': '11951A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e4:85',
                                         'uptime': '3114073'},
                                 {       'counter_source': '215664/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '215664/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '422035260559',
                                                                       'in_ucast_pkts': '32142966629',
                                                                       'mac_address': 'b8:94:70:ba:e2:65',
                                                                       'name': '215664/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '123',
                                                                       'out_octets': '4512030014842',
                                                                       'out_ucast_pkts': '4308826404',
                                                                       'poller_hash': 'MTE5NTE2MXBlZXJpbnRmMjE1NjY0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '215664',
                                                                       'in_nucast_pkts': '123',
                                                                       'in_octets': '4512030014842',
                                                                       'in_ucast_pkts': '4308826404',
                                                                       'links': [       {       'in_nucast_pkts': '123',
                                                                                                'in_octets': '4512030014842',
                                                                                                'in_ucast_pkts': '4308826404',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '422035260559',
                                                                                                'out_ucast_pkts': '32142966629',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '422035260559',
                                                                       'out_ucast_pkts': '32142966629',
                                                                       'poller_hash': 'MTE5NTE2MXBlZXJpbnRmcG9uIDIxNTY2NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215664',
                                                                                               'links': [       {       'connected_time': '19801590',
                                                                                                                        'fiber_length': 11966,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE2MXBlZXJwb25wb24gMjE1NjY0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6972 MILLBANK MAIN STREET (MARTIN '
                                                 'ROGER)',
                                         'poller_hash': 'MTE5NTE2MQ==',
                                         'serial_number': '1195161',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e2:65',
                                         'uptime': '19801590'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '215773/g1',
                                                                       'name': '215773/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTFCMjVENHBlZXJpbnRmMjE1NzczL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '215773',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTFCMjVENHBlZXJpbnRmcG9uIDIxNTc3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215773',
                                                                                               'links': [       {       'connected_time': '19363903',
                                                                                                                        'fiber_length': 8450,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCMjVENHBlZXJwb25wb24gMjE1Nzcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4863 LINE 59 (DAVIDSON TIM)',
                                         'poller_hash': 'MTFCMjVENA==',
                                         'serial_number': '11B25D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:c7:06:4f',
                                         'uptime': '19363903'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '215835/g1',
                                                                       'name': '215835/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE5NDVEMXBlZXJpbnRmMjE1ODM1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '215835',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE5NDVEMXBlZXJpbnRmcG9uIDIxNTgzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215835',
                                                                                               'links': [       {       'connected_time': '19197442',
                                                                                                                        'fiber_length': 8831,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NDVEMXBlZXJwb25wb24gMjE1ODM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4822 LINE 59 (GRAVELRIDGE WI HALL)',
                                         'poller_hash': 'MTE5NDVEMQ==',
                                         'serial_number': '11945D1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:85:e5',
                                         'uptime': '19197442'},
                                 {       'counter_source': '215959/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '215959/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '413125254761',
                                                                       'in_ucast_pkts': '104454674996',
                                                                       'mac_address': 'b8:94:70:ba:e1:75',
                                                                       'name': '215959/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '5217599652920',
                                                                       'out_ucast_pkts': '4114985580',
                                                                       'poller_hash': 'MTE5NTE0M3BlZXJpbnRmMjE1OTU5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '215959',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '5217599652920',
                                                                       'in_ucast_pkts': '4114985580',
                                                                       'links': [       {       'in_nucast_pkts': '66',
                                                                                                'in_octets': '5217599652920',
                                                                                                'in_ucast_pkts': '4114985580',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '413125254761',
                                                                                                'out_ucast_pkts': '104454674996',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '413125254761',
                                                                       'out_ucast_pkts': '104454674996',
                                                                       'poller_hash': 'MTE5NTE0M3BlZXJpbnRmcG9uIDIxNTk1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215959',
                                                                                               'links': [       {       'connected_time': '2065619',
                                                                                                                        'fiber_length': 13566,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE0M3BlZXJwb25wb24gMjE1OTU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5262 LINE 52 (BVH EXCAVATION & '
                                                 'CONTRACTING)',
                                         'poller_hash': 'MTE5NTE0Mw==',
                                         'serial_number': '1195143',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e1:75',
                                         'uptime': '2065619'},
                                 {       'counter_source': '216073/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b8:d2'],
                                                                       'id': '216073/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '41321523124',
                                                                       'in_ucast_pkts': '215255015',
                                                                       'mac_address': 'b8:94:70:ba:df:30',
                                                                       'mtu': 2000,
                                                                       'name': '216073/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1494777',
                                                                       'out_octets': '877683639817',
                                                                       'out_ucast_pkts': '624034799',
                                                                       'poller_hash': 'MTE5NTBGQXBlZXJpbnRmMjE2MDczL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216073',
                                                                       'in_nucast_pkts': '1494777',
                                                                       'in_octets': '877683639817',
                                                                       'in_ucast_pkts': '624034799',
                                                                       'links': [       {       'in_nucast_pkts': '1494777',
                                                                                                'in_octets': '877683639817',
                                                                                                'in_ucast_pkts': '624034799',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '41321523124',
                                                                                                'out_ucast_pkts': '215255015',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '41321523124',
                                                                       'out_ucast_pkts': '215255015',
                                                                       'poller_hash': 'MTE5NTBGQXBlZXJpbnRmcG9uIDIxNjA3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216073',
                                                                                               'links': [       {       'connected_time': '2969301',
                                                                                                                        'fiber_length': 17588,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTBGQXBlZXJwb25wb24gMjE2MDcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '194 WINSTANLEY ST (THOM KENNEDY)',
                                         'poller_hash': 'MTE5NTBGQQ==',
                                         'serial_number': '11950FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:df:2d',
                                         'uptime': '2969301'},
                                 {       'counter_source': '216083/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f0:5b'],
                                                                       'id': '216083/g1',
                                                                       'in_nucast_pkts': '12763',
                                                                       'in_octets': '14392895080',
                                                                       'in_ucast_pkts': '98152911',
                                                                       'mac_address': 'b8:94:70:ba:e4:e0',
                                                                       'mtu': 2000,
                                                                       'name': '216083/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464116',
                                                                       'out_octets': '422381155003',
                                                                       'out_ucast_pkts': '313396911',
                                                                       'poller_hash': 'MTE5NTFCMHBlZXJpbnRmMjE2MDgzL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216083',
                                                                       'in_nucast_pkts': '1464116',
                                                                       'in_octets': '422381155003',
                                                                       'in_ucast_pkts': '313396911',
                                                                       'links': [       {       'in_nucast_pkts': '1464116',
                                                                                                'in_octets': '422381155003',
                                                                                                'in_ucast_pkts': '313396911',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12763',
                                                                                                'out_octets': '14392895080',
                                                                                                'out_ucast_pkts': '98152911',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12763',
                                                                       'out_octets': '14392895080',
                                                                       'out_ucast_pkts': '98152911',
                                                                       'poller_hash': 'MTE5NTFCMHBlZXJpbnRmcG9uIDIxNjA4Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216083',
                                                                                               'links': [       {       'connected_time': '688503',
                                                                                                                        'fiber_length': 622,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTFCMHBlZXJwb25wb24gMjE2MDgz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 MAIN ST N (WITMER TIM)',
                                         'poller_hash': 'MTE5NTFCMA==',
                                         'serial_number': '11951B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e4:dd',
                                         'uptime': '688503'},
                                 {       'counter_source': '216207/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:8e:86'],
                                                                       'id': '216207/g1',
                                                                       'in_nucast_pkts': '65149',
                                                                       'in_octets': '6003450707',
                                                                       'in_ucast_pkts': '53005179',
                                                                       'mac_address': 'b8:94:70:ba:c3:48',
                                                                       'mtu': 2000,
                                                                       'name': '216207/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6782844',
                                                                       'out_octets': '305934531387',
                                                                       'out_ucast_pkts': '212107857',
                                                                       'poller_hash': 'MTE5NEQ3RHBlZXJpbnRmMjE2MjA3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216207',
                                                                       'in_nucast_pkts': '6782844',
                                                                       'in_octets': '305934531387',
                                                                       'in_ucast_pkts': '212107857',
                                                                       'links': [       {       'in_nucast_pkts': '6782844',
                                                                                                'in_octets': '305934531387',
                                                                                                'in_ucast_pkts': '212107857',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '65149',
                                                                                                'out_octets': '6003450707',
                                                                                                'out_ucast_pkts': '53005179',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '65149',
                                                                       'out_octets': '6003450707',
                                                                       'out_ucast_pkts': '53005179',
                                                                       'poller_hash': 'MTE5NEQ3RHBlZXJpbnRmcG9uIDIxNjIwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216207',
                                                                                               'links': [       {       'connected_time': '3114081',
                                                                                                                        'fiber_length': 15371,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NEQ3RHBlZXJwb25wb24gMjE2MjA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5640 PERTH LINE 55 (RAYCRAFT RANDY)',
                                         'poller_hash': 'MTE5NEQ3RA==',
                                         'serial_number': '1194D7D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:c3:45',
                                         'uptime': '3114081'},
                                 {       'counter_source': '216247/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:8d:31'],
                                                                       'id': '216247/g1',
                                                                       'in_nucast_pkts': '12408',
                                                                       'in_octets': '3815414753',
                                                                       'in_ucast_pkts': '38487689',
                                                                       'mac_address': 'b8:94:70:ba:e3:98',
                                                                       'mtu': 2000,
                                                                       'name': '216247/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464647',
                                                                       'out_octets': '712557466527',
                                                                       'out_ucast_pkts': '471995350',
                                                                       'poller_hash': 'MTE5NTE4N3BlZXJpbnRmMjE2MjQ3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216247',
                                                                       'in_nucast_pkts': '1464647',
                                                                       'in_octets': '712557466527',
                                                                       'in_ucast_pkts': '471995350',
                                                                       'links': [       {       'in_nucast_pkts': '1464647',
                                                                                                'in_octets': '712557466527',
                                                                                                'in_ucast_pkts': '471995350',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12408',
                                                                                                'out_octets': '3815414753',
                                                                                                'out_ucast_pkts': '38487689',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12408',
                                                                       'out_octets': '3815414753',
                                                                       'out_ucast_pkts': '38487689',
                                                                       'poller_hash': 'MTE5NTE4N3BlZXJpbnRmcG9uIDIxNjI0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216247',
                                                                                               'links': [       {       'connected_time': '688469',
                                                                                                                        'fiber_length': 1677,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE4N3BlZXJwb25wb24gMjE2MjQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 WILLIAM ST (MURAWSKY PAUL)',
                                         'poller_hash': 'MTE5NTE4Nw==',
                                         'serial_number': '1195187',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e3:95',
                                         'uptime': '688469'},
                                 {       'counter_source': '216722/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:eb:e2'],
                                                                       'id': '216722/g1',
                                                                       'in_nucast_pkts': '53726',
                                                                       'in_octets': '28129212154',
                                                                       'in_ucast_pkts': '42174423',
                                                                       'mac_address': 'b8:94:70:ba:e4:98',
                                                                       'mtu': 2000,
                                                                       'name': '216722/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6710277',
                                                                       'out_octets': '109447245014',
                                                                       'out_ucast_pkts': '88884870',
                                                                       'poller_hash': 'MTE5NTFBN3BlZXJpbnRmMjE2NzIyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216722',
                                                                       'in_nucast_pkts': '6710277',
                                                                       'in_octets': '109447245014',
                                                                       'in_ucast_pkts': '88884870',
                                                                       'links': [       {       'in_nucast_pkts': '6710277',
                                                                                                'in_octets': '109447245014',
                                                                                                'in_ucast_pkts': '88884870',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '53726',
                                                                                                'out_octets': '28129212154',
                                                                                                'out_ucast_pkts': '42174423',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53726',
                                                                       'out_octets': '28129212154',
                                                                       'out_ucast_pkts': '42174423',
                                                                       'poller_hash': 'MTE5NTFBN3BlZXJpbnRmcG9uIDIxNjcyMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216722',
                                                                                               'links': [       {       'connected_time': '3113968',
                                                                                                                        'fiber_length': 18147,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTFBN3BlZXJwb25wb24gMjE2NzIy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 MADDISON ST E (THIESSEN GERHARD)',
                                         'poller_hash': 'MTE5NTFBNw==',
                                         'serial_number': '11951A7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e4:95',
                                         'uptime': '3113968'},
                                 {       'counter_source': '216732/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:ad:08'],
                                                                       'id': '216732/g1',
                                                                       'in_nucast_pkts': '14795',
                                                                       'in_octets': '6666357485',
                                                                       'in_ucast_pkts': '28796125',
                                                                       'mac_address': 'b8:94:70:ba:e5:10',
                                                                       'mtu': 2000,
                                                                       'name': '216732/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464398',
                                                                       'out_octets': '106255835577',
                                                                       'out_ucast_pkts': '80422526',
                                                                       'poller_hash': 'MTE5NTFCNnBlZXJpbnRmMjE2NzMyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216732',
                                                                       'in_nucast_pkts': '1464398',
                                                                       'in_octets': '106255835577',
                                                                       'in_ucast_pkts': '80422526',
                                                                       'links': [       {       'in_nucast_pkts': '1464398',
                                                                                                'in_octets': '106255835577',
                                                                                                'in_ucast_pkts': '80422526',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14795',
                                                                                                'out_octets': '6666357485',
                                                                                                'out_ucast_pkts': '28796125',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14795',
                                                                       'out_octets': '6666357485',
                                                                       'out_ucast_pkts': '28796125',
                                                                       'poller_hash': 'MTE5NTFCNnBlZXJpbnRmcG9uIDIxNjczMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216732',
                                                                                               'links': [       {       'connected_time': '688378',
                                                                                                                        'fiber_length': 2720,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTFCNnBlZXJwb25wb24gMjE2NzMy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '129 PUGH ST (ANTHONY DAVE)',
                                         'poller_hash': 'MTE5NTFCNg==',
                                         'serial_number': '11951B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e5:0d',
                                         'uptime': '688378'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '217005/g1',
                                                                       'name': '217005/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTA5NDkwNnBlZXJpbnRmMjE3MDA1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '217005',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTA5NDkwNnBlZXJpbnRmcG9uIDIxNzAwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217005',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTA5NDkwNnBlZXJwb25wb24gMjE3MDA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '2 MAIN ST N (FOSTER ANNE)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "217005" '
                                                                  '(location: GPON '
                                                                  'port "8/7"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-02-01T02:01:53.400Z',
                                         'poller_hash': 'MTA5NDkwNg==',
                                         'serial_number': '1094906',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '217016/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '217016/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '228518053151',
                                                                       'in_ucast_pkts': '1171483603',
                                                                       'mac_address': 'b8:94:70:ba:df:e5',
                                                                       'name': '217016/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '47',
                                                                       'out_octets': '3525560860788',
                                                                       'out_ucast_pkts': '2662007248',
                                                                       'poller_hash': 'MTE5NTExMXBlZXJpbnRmMjE3MDE2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '217016',
                                                                       'in_nucast_pkts': '47',
                                                                       'in_octets': '3525560860788',
                                                                       'in_ucast_pkts': '2662007248',
                                                                       'links': [       {       'in_nucast_pkts': '47',
                                                                                                'in_octets': '3525560860788',
                                                                                                'in_ucast_pkts': '2662007248',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '228518053151',
                                                                                                'out_ucast_pkts': '1171483603',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '228518053151',
                                                                       'out_ucast_pkts': '1171483603',
                                                                       'poller_hash': 'MTE5NTExMXBlZXJpbnRmcG9uIDIxNzAxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217016',
                                                                                               'links': [       {       'connected_time': '688364',
                                                                                                                        'fiber_length': 3351,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTExMXBlZXJwb25wb24gMjE3MDE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '78 PUGH ST (ENSINGER JASON)',
                                         'poller_hash': 'MTE5NTExMQ==',
                                         'serial_number': '1195111',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:df:e5',
                                         'uptime': '688364'},
                                 {       'counter_source': '217026/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:dd:4a'],
                                                                       'id': '217026/g1',
                                                                       'in_nucast_pkts': '14647',
                                                                       'in_octets': '470649483',
                                                                       'in_ucast_pkts': '1613951',
                                                                       'mac_address': 'b8:94:70:ba:cb:c0',
                                                                       'mtu': 2000,
                                                                       'name': '217026/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434448',
                                                                       'out_octets': '3587132867',
                                                                       'out_ucast_pkts': '3264338',
                                                                       'poller_hash': 'MTE5NEU4Q3BlZXJpbnRmMjE3MDI2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '217026',
                                                                       'in_nucast_pkts': '1434448',
                                                                       'in_octets': '3587132867',
                                                                       'in_ucast_pkts': '3264338',
                                                                       'links': [       {       'in_nucast_pkts': '1434448',
                                                                                                'in_octets': '3587132867',
                                                                                                'in_ucast_pkts': '3264338',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14647',
                                                                                                'out_octets': '470649483',
                                                                                                'out_ucast_pkts': '1613951',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14647',
                                                                       'out_octets': '470649483',
                                                                       'out_ucast_pkts': '1613951',
                                                                       'poller_hash': 'MTE5NEU4Q3BlZXJpbnRmcG9uIDIxNzAyNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217026',
                                                                                               'links': [       {       'connected_time': '688557',
                                                                                                                        'fiber_length': 1788,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NEU4Q3BlZXJwb25wb24gMjE3MDI2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (ROTH CAROLYN)',
                                         'poller_hash': 'MTE5NEU4Qw==',
                                         'serial_number': '1194E8C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:cb:bd',
                                         'uptime': '688557'},
                                 {       'counter_source': '217140/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e6:66'],
                                                                       'id': '217140/g1',
                                                                       'in_nucast_pkts': '25437511',
                                                                       'in_octets': '97021251620',
                                                                       'in_ucast_pkts': '630938678',
                                                                       'mac_address': 'ec:4f:82:dc:f1:58',
                                                                       'mtu': 2000,
                                                                       'name': '217140/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22211568718',
                                                                       'out_octets': '35025298874024',
                                                                       'out_ucast_pkts': '4111661475',
                                                                       'poller_hash': 'MkVEOUM0cGVlcmludGYyMTcxNDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '217140',
                                                                       'in_nucast_pkts': '22211568718',
                                                                       'in_octets': '35025298874024',
                                                                       'in_ucast_pkts': '4111661475',
                                                                       'links': [       {       'in_nucast_pkts': '22211568718',
                                                                                                'in_octets': '35025298874024',
                                                                                                'in_ucast_pkts': '4111661475',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '25437511',
                                                                                                'out_octets': '97021251620',
                                                                                                'out_ucast_pkts': '630938678',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '25437511',
                                                                       'out_octets': '97021251620',
                                                                       'out_ucast_pkts': '630938678',
                                                                       'poller_hash': 'MkVEOUM0cGVlcmludGZwb24gMjE3MTQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217140',
                                                                                               'links': [       {       'connected_time': '41473',
                                                                                                                        'fiber_length': 1109,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.61,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVEOUM0cGVlcnBvbnBvbiAyMTcxNDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '32 WILLIAM ST (LEHMANN DON & SHARON)',
                                         'poller_hash': 'MkVEOUM0',
                                         'serial_number': '2ED9C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:dc:f1:54',
                                         'uptime': '41473'},
                                 {       'counter_source': '217150/g2+217150/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:43:96:ba'],
                                                                       'id': '217150/g2',
                                                                       'in_nucast_pkts': '787455',
                                                                       'in_octets': '520939807',
                                                                       'in_ucast_pkts': '1017981',
                                                                       'mac_address': '04:bc:9f:5b:ce:dd',
                                                                       'mtu': 2000,
                                                                       'name': '217150/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38962695',
                                                                       'out_octets': '18737248242',
                                                                       'out_ucast_pkts': '2303618',
                                                                       'poller_hash': 'MTI3NDA1QnBlZXJpbnRmMjE3MTUwL2cy',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:3e:b8'],
                                                                       'id': '217150/g3',
                                                                       'in_nucast_pkts': '758958',
                                                                       'in_octets': '14732756473',
                                                                       'in_ucast_pkts': '198397493',
                                                                       'mac_address': '04:bc:9f:5b:ce:de',
                                                                       'mtu': 2000,
                                                                       'name': '217150/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38982376',
                                                                       'out_octets': '1113944525603',
                                                                       'out_ucast_pkts': '781893817',
                                                                       'poller_hash': 'MTI3NDA1QnBlZXJpbnRmMjE3MTUwL2cz',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '217150',
                                                                       'in_nucast_pkts': '77945071',
                                                                       'in_octets': '1132681773845',
                                                                       'in_ucast_pkts': '784197435',
                                                                       'links': [       {       'in_nucast_pkts': '77945071',
                                                                                                'in_octets': '1132681773845',
                                                                                                'in_ucast_pkts': '784197435',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1546413',
                                                                                                'out_octets': '15253696280',
                                                                                                'out_ucast_pkts': '199415474',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1546413',
                                                                       'out_octets': '15253696280',
                                                                       'out_ucast_pkts': '199415474',
                                                                       'poller_hash': 'MTI3NDA1QnBlZXJpbnRmcG9uIDIxNzE1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217150',
                                                                                               'links': [       {       'connected_time': '1911312',
                                                                                                                        'fiber_length': 15190,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA1QnBlZXJwb25wb24gMjE3MTUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5619B PERTH LINE 55 (MCEWEN R '
                                                 'DOUGLAS)',
                                         'poller_hash': 'MTI3NDA1Qg==',
                                         'serial_number': '127405B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:d9',
                                         'uptime': '1911312'},
                                 {       'counter_source': '217160/g1+217160/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:3f:5e'],
                                                                       'id': '217160/g1',
                                                                       'in_nucast_pkts': '777404',
                                                                       'in_octets': '9188419319',
                                                                       'in_ucast_pkts': '122059143',
                                                                       'mac_address': 'b8:94:70:ba:ca:f0',
                                                                       'mtu': 2000,
                                                                       'name': '217160/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38111219',
                                                                       'out_octets': '641717948413',
                                                                       'out_ucast_pkts': '445899661',
                                                                       'poller_hash': 'MTE5NEU3MnBlZXJpbnRmMjE3MTYwL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:43:96:e0'],
                                                                       'id': '217160/g3',
                                                                       'in_nucast_pkts': '790388',
                                                                       'in_octets': '722921628',
                                                                       'in_ucast_pkts': '3945448',
                                                                       'mac_address': 'b8:94:70:ba:ca:f2',
                                                                       'mtu': 2000,
                                                                       'name': '217160/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41154840',
                                                                       'out_octets': '36781872061',
                                                                       'out_ucast_pkts': '14419801',
                                                                       'poller_hash': 'MTE5NEU3MnBlZXJpbnRmMjE3MTYwL2cz',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '217160',
                                                                       'in_nucast_pkts': '79266059',
                                                                       'in_octets': '678499820474',
                                                                       'in_ucast_pkts': '460319462',
                                                                       'links': [       {       'in_nucast_pkts': '79266059',
                                                                                                'in_octets': '678499820474',
                                                                                                'in_ucast_pkts': '460319462',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1567792',
                                                                                                'out_octets': '9911340947',
                                                                                                'out_ucast_pkts': '126004591',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1567792',
                                                                       'out_octets': '9911340947',
                                                                       'out_ucast_pkts': '126004591',
                                                                       'poller_hash': 'MTE5NEU3MnBlZXJpbnRmcG9uIDIxNzE2MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217160',
                                                                                               'links': [       {       'connected_time': '1910110',
                                                                                                                        'fiber_length': 15245,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NEU3MnBlZXJwb25wb24gMjE3MTYw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5619A PERTH LINE 55 (MCEWEN BRIAN)',
                                         'poller_hash': 'MTE5NEU3Mg==',
                                         'serial_number': '1194E72',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:ca:ed',
                                         'uptime': '1910110'},
                                 {       'counter_source': '217947/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '217947/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1533436140721',
                                                                       'in_ucast_pkts': '36729982923',
                                                                       'mac_address': '04:bc:9f:5c:52:b9',
                                                                       'name': '217947/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '15',
                                                                       'out_octets': '6567041044091',
                                                                       'out_ucast_pkts': '5489069679',
                                                                       'poller_hash': 'MTI3NTBEN3BlZXJpbnRmMjE3OTQ3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '217947',
                                                                       'in_nucast_pkts': '15',
                                                                       'in_octets': '6567041044091',
                                                                       'in_ucast_pkts': '5489069679',
                                                                       'links': [       {       'in_nucast_pkts': '15',
                                                                                                'in_octets': '6567041044091',
                                                                                                'in_ucast_pkts': '5489069679',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1533436140721',
                                                                                                'out_ucast_pkts': '36729982923',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1533436140721',
                                                                       'out_ucast_pkts': '36729982923',
                                                                       'poller_hash': 'MTI3NTBEN3BlZXJpbnRmcG9uIDIxNzk0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217947',
                                                                                               'links': [       {       'connected_time': '688532',
                                                                                                                        'fiber_length': 11528,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTBEN3BlZXJwb25wb24gMjE3OTQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5808 PERTH ROAD 119 (FORLER '
                                                 'RACHELLE)',
                                         'poller_hash': 'MTI3NTBENw==',
                                         'serial_number': '12750D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:52:b9',
                                         'uptime': '688532'},
                                 {       'counter_source': '217967/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '217967/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2167863362269',
                                                                       'in_ucast_pkts': '442920531401',
                                                                       'mac_address': 'b8:94:70:ba:e1:c5',
                                                                       'name': '217967/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '11450596700946',
                                                                       'out_ucast_pkts': '8807519232',
                                                                       'poller_hash': 'MTE5NTE0RHBlZXJpbnRmMjE3OTY3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '217967',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '11450596700946',
                                                                       'in_ucast_pkts': '8807519232',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '11450596700946',
                                                                                                'in_ucast_pkts': '8807519232',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2167863362269',
                                                                                                'out_ucast_pkts': '442920531401',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2167863362269',
                                                                       'out_ucast_pkts': '442920531401',
                                                                       'poller_hash': 'MTE5NTE0RHBlZXJpbnRmcG9uIDIxNzk2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217967',
                                                                                               'links': [       {       'connected_time': '12942573',
                                                                                                                        'fiber_length': 17535,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE0RHBlZXJwb25wb24gMjE3OTY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '207 SELBY AV (REDEEMER LUTHERAN '
                                                 'CHURCH)',
                                         'poller_hash': 'MTE5NTE0RA==',
                                         'serial_number': '119514D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e1:c5',
                                         'uptime': '12942573'},
                                 {       'counter_source': '217977/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e0:e6'],
                                                                       'id': '217977/g1',
                                                                       'in_nucast_pkts': '12359',
                                                                       'in_octets': '6817975690',
                                                                       'in_ucast_pkts': '36563327',
                                                                       'mac_address': '04:bc:9f:5c:4d:b4',
                                                                       'mtu': 2000,
                                                                       'name': '217977/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1415867',
                                                                       'out_octets': '146351024110',
                                                                       'out_ucast_pkts': '111527071',
                                                                       'poller_hash': 'MTI3NTAzNnBlZXJpbnRmMjE3OTc3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '217977',
                                                                       'in_nucast_pkts': '1415867',
                                                                       'in_octets': '146351024110',
                                                                       'in_ucast_pkts': '111527071',
                                                                       'links': [       {       'in_nucast_pkts': '1415867',
                                                                                                'in_octets': '146351024110',
                                                                                                'in_ucast_pkts': '111527071',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12359',
                                                                                                'out_octets': '6817975690',
                                                                                                'out_ucast_pkts': '36563327',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12359',
                                                                       'out_octets': '6817975690',
                                                                       'out_ucast_pkts': '36563327',
                                                                       'poller_hash': 'MTI3NTAzNnBlZXJpbnRmcG9uIDIxNzk3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217977',
                                                                                               'links': [       {       'connected_time': '665669',
                                                                                                                        'fiber_length': 11543,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTAzNnBlZXJwb25wb24gMjE3OTc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5804 PERTH ROAD 119 (CNOSSEN SIMON & '
                                                 'ANDREA)',
                                         'poller_hash': 'MTI3NTAzNg==',
                                         'serial_number': '1275036',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4d:b1',
                                         'uptime': '665669'},
                                 {       'counter_source': '218072/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e6:0e'],
                                                                       'id': '218072/g1',
                                                                       'in_nucast_pkts': '105949',
                                                                       'in_octets': '20920149465',
                                                                       'in_ucast_pkts': '265895097',
                                                                       'mac_address': '04:bc:9f:5c:52:6c',
                                                                       'mtu': 2000,
                                                                       'name': '218072/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14237494',
                                                                       'out_octets': '1425464949951',
                                                                       'out_ucast_pkts': '947457510',
                                                                       'poller_hash': 'MTI3NTBDRHBlZXJpbnRmMjE4MDcyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '218072',
                                                                       'in_nucast_pkts': '14237494',
                                                                       'in_octets': '1425464949951',
                                                                       'in_ucast_pkts': '947457510',
                                                                       'links': [       {       'in_nucast_pkts': '14237494',
                                                                                                'in_octets': '1425464949951',
                                                                                                'in_ucast_pkts': '947457510',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '105949',
                                                                                                'out_octets': '20920149465',
                                                                                                'out_ucast_pkts': '265895097',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '105949',
                                                                       'out_octets': '20920149465',
                                                                       'out_ucast_pkts': '265895097',
                                                                       'poller_hash': 'MTI3NTBDRHBlZXJpbnRmcG9uIDIxODA3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218072',
                                                                                               'links': [       {       'connected_time': '6075955',
                                                                                                                        'fiber_length': 779,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTBDRHBlZXJwb25wb24gMjE4MDcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '31 MAIN ST S (HOBBS KEN & GUDRUN)',
                                         'poller_hash': 'MTI3NTBDRA==',
                                         'serial_number': '12750CD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:52:69',
                                         'uptime': '6075955'},
                                 {       'counter_source': '218227/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e8:1e'],
                                                                       'id': '218227/g1',
                                                                       'in_nucast_pkts': '11963',
                                                                       'in_octets': '1585080934',
                                                                       'in_ucast_pkts': '6250030',
                                                                       'mac_address': '04:bc:9f:5c:50:14',
                                                                       'mtu': 2000,
                                                                       'name': '218227/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464439',
                                                                       'out_octets': '34913218510',
                                                                       'out_ucast_pkts': '25571942',
                                                                       'poller_hash': 'MTI3NTA4MnBlZXJpbnRmMjE4MjI3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '218227',
                                                                       'in_nucast_pkts': '1464439',
                                                                       'in_octets': '34913218510',
                                                                       'in_ucast_pkts': '25571942',
                                                                       'links': [       {       'in_nucast_pkts': '1464439',
                                                                                                'in_octets': '34913218510',
                                                                                                'in_ucast_pkts': '25571942',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11963',
                                                                                                'out_octets': '1585080934',
                                                                                                'out_ucast_pkts': '6250030',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11963',
                                                                       'out_octets': '1585080934',
                                                                       'out_ucast_pkts': '6250030',
                                                                       'poller_hash': 'MTI3NTA4MnBlZXJpbnRmcG9uIDIxODIyNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218227',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 3044,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTA4MnBlZXJwb25wb24gMjE4MjI3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '120 PUGH ST (SCHMIDT KELSEY)',
                                         'poller_hash': 'MTI3NTA4Mg==',
                                         'serial_number': '1275082',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:50:11',
                                         'uptime': '688370'},
                                 {       'counter_source': '218257/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '218257/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '72940963966',
                                                                       'in_ucast_pkts': '8800901096',
                                                                       'mac_address': '04:bc:9f:5c:4f:c1',
                                                                       'name': '218257/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '286841383782',
                                                                       'out_ucast_pkts': '365308984',
                                                                       'poller_hash': 'MTI3NTA3OHBlZXJpbnRmMjE4MjU3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '218257',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '286841383782',
                                                                       'in_ucast_pkts': '365308984',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '286841383782',
                                                                                                'in_ucast_pkts': '365308984',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '72940963966',
                                                                                                'out_ucast_pkts': '8800901096',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '72940963966',
                                                                       'out_ucast_pkts': '8800901096',
                                                                       'poller_hash': 'MTI3NTA3OHBlZXJpbnRmcG9uIDIxODI1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218257',
                                                                                               'links': [       {       'connected_time': '15560080',
                                                                                                                        'fiber_length': 8733,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTA3OHBlZXJwb25wb24gMjE4MjU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5005 PERTH LINE 55 (ALBRECHT LESLIE)',
                                         'poller_hash': 'MTI3NTA3OA==',
                                         'serial_number': '1275078',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4f:c1',
                                         'uptime': '15560080'},
                                 {       'counter_source': '218466/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e4:e5'],
                                                                       'id': '218466/g1',
                                                                       'in_nucast_pkts': '12066',
                                                                       'in_octets': '6957140903',
                                                                       'in_ucast_pkts': '24910329',
                                                                       'mac_address': '04:bc:9f:5c:58:7c',
                                                                       'mtu': 2000,
                                                                       'name': '218466/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464322',
                                                                       'out_octets': '244623327780',
                                                                       'out_ucast_pkts': '171797600',
                                                                       'poller_hash': 'MTI3NTE4RnBlZXJpbnRmMjE4NDY2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '218466',
                                                                       'in_nucast_pkts': '1464322',
                                                                       'in_octets': '244623327780',
                                                                       'in_ucast_pkts': '171797600',
                                                                       'links': [       {       'in_nucast_pkts': '1464322',
                                                                                                'in_octets': '244623327780',
                                                                                                'in_ucast_pkts': '171797600',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12066',
                                                                                                'out_octets': '6957140903',
                                                                                                'out_ucast_pkts': '24910329',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12066',
                                                                       'out_octets': '6957140903',
                                                                       'out_ucast_pkts': '24910329',
                                                                       'poller_hash': 'MTI3NTE4RnBlZXJpbnRmcG9uIDIxODQ2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218466',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 2543,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTE4RnBlZXJwb25wb24gMjE4NDY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '103 PUGH ST (SABOL SAMANTHA)',
                                         'poller_hash': 'MTI3NTE4Rg==',
                                         'serial_number': '127518F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:58:79',
                                         'uptime': '688369'},
                                 {       'counter_source': '218845/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '218845/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6264289449100',
                                                                       'in_ucast_pkts': '481787634452',
                                                                       'mac_address': '60:db:98:73:e2:5c',
                                                                       'name': '218845/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '74',
                                                                       'out_octets': '3852711187899',
                                                                       'out_ucast_pkts': '4775909160',
                                                                       'poller_hash': 'QjBBQUU4cGVlcmludGYyMTg4NDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '218845',
                                                                       'in_nucast_pkts': '74',
                                                                       'in_octets': '3852711187899',
                                                                       'in_ucast_pkts': '4775909160',
                                                                       'links': [       {       'in_nucast_pkts': '74',
                                                                                                'in_octets': '3852711187899',
                                                                                                'in_ucast_pkts': '4775909160',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '6264289449100',
                                                                                                'out_ucast_pkts': '481787634452',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6264289449100',
                                                                       'out_ucast_pkts': '481787634452',
                                                                       'poller_hash': 'QjBBQUU4cGVlcmludGZwb24gMjE4ODQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218845',
                                                                                               'links': [       {       'connected_time': '688496',
                                                                                                                        'fiber_length': 559,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBQUU4cGVlcnBvbnBvbiAyMTg4NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '4 MAIN ST S (MACEWAN ADAM)',
                                         'poller_hash': 'QjBBQUU4',
                                         'serial_number': 'B0AAE8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:e2:5c',
                                         'uptime': '688496'},
                                 {       'counter_source': '219035/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '219035/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '237837834572',
                                                                       'in_ucast_pkts': '1353685599',
                                                                       'mac_address': '04:bc:9f:5c:56:d1',
                                                                       'name': '219035/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '5425731575671',
                                                                       'out_ucast_pkts': '4047028512',
                                                                       'poller_hash': 'MTI3NTE1QXBlZXJpbnRmMjE5MDM1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '219035',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '5425731575671',
                                                                       'in_ucast_pkts': '4047028512',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '5425731575671',
                                                                                                'in_ucast_pkts': '4047028512',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '237837834572',
                                                                                                'out_ucast_pkts': '1353685599',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '237837834572',
                                                                       'out_ucast_pkts': '1353685599',
                                                                       'poller_hash': 'MTI3NTE1QXBlZXJpbnRmcG9uIDIxOTAzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '219035',
                                                                                               'links': [       {       'connected_time': '6053404',
                                                                                                                        'fiber_length': 1381,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTE1QXBlZXJwb25wb24gMjE5MDM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '27 ARENA DR (PUGH TIM & ROSE)',
                                         'poller_hash': 'MTI3NTE1QQ==',
                                         'serial_number': '127515A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:56:d1',
                                         'uptime': '6053404'},
                                 {       'counter_source': '220359/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:d7:82'],
                                                                       'id': '220359/g1',
                                                                       'in_nucast_pkts': '38271',
                                                                       'in_octets': '1330004258',
                                                                       'in_ucast_pkts': '17444831',
                                                                       'mac_address': '04:bc:9f:5c:4f:b4',
                                                                       'mtu': 2000,
                                                                       'name': '220359/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15664490',
                                                                       'out_octets': '104185937041',
                                                                       'out_ucast_pkts': '69671748',
                                                                       'poller_hash': 'MTI3NTA3NnBlZXJpbnRmMjIwMzU5L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '220359',
                                                                       'in_nucast_pkts': '15664490',
                                                                       'in_octets': '104185937041',
                                                                       'in_ucast_pkts': '69671748',
                                                                       'links': [       {       'in_nucast_pkts': '15664490',
                                                                                                'in_octets': '104185937041',
                                                                                                'in_ucast_pkts': '69671748',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '38271',
                                                                                                'out_octets': '1330004258',
                                                                                                'out_ucast_pkts': '17444831',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38271',
                                                                       'out_octets': '1330004258',
                                                                       'out_ucast_pkts': '17444831',
                                                                       'poller_hash': 'MTI3NTA3NnBlZXJpbnRmcG9uIDIyMDM1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '220359',
                                                                                               'links': [       {       'connected_time': '688499',
                                                                                                                        'fiber_length': 2062,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTA3NnBlZXJwb25wb24gMjIwMzU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '26 MAIN ST N (DULIES ALMUTH)',
                                         'poller_hash': 'MTI3NTA3Ng==',
                                         'serial_number': '1275076',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4f:b1',
                                         'uptime': '688499'},
                                 {       'counter_source': '221120/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '221120/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '95970349580',
                                                                       'in_ucast_pkts': '131709041',
                                                                       'mac_address': '04:bc:9f:5c:54:a1',
                                                                       'name': '221120/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '91',
                                                                       'out_octets': '128272121538',
                                                                       'out_ucast_pkts': '152988469',
                                                                       'poller_hash': 'MTI3NTExNHBlZXJpbnRmMjIxMTIwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '221120',
                                                                       'in_nucast_pkts': '91',
                                                                       'in_octets': '128272121538',
                                                                       'in_ucast_pkts': '152988469',
                                                                       'links': [       {       'in_nucast_pkts': '91',
                                                                                                'in_octets': '128272121538',
                                                                                                'in_ucast_pkts': '152988469',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '95970349580',
                                                                                                'out_ucast_pkts': '131709041',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '95970349580',
                                                                       'out_ucast_pkts': '131709041',
                                                                       'poller_hash': 'MTI3NTExNHBlZXJpbnRmcG9uIDIyMTEyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221120',
                                                                                               'links': [       {       'connected_time': '688503',
                                                                                                                        'fiber_length': 1181,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTExNHBlZXJwb25wb24gMjIxMTIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '69 MAIN ST N (PRAYER HOUSE OF '
                                                 'MILVERTON)',
                                         'poller_hash': 'MTI3NTExNA==',
                                         'serial_number': '1275114',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:54:a1',
                                         'uptime': '688503'},
                                 {       'counter_source': '221465/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2c:71:b1'],
                                                                       'id': '221465/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '4724637248',
                                                                       'in_ucast_pkts': '34746781',
                                                                       'mac_address': 'f8:85:f9:17:2a:db',
                                                                       'mtu': 2000,
                                                                       'name': '221465/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464577',
                                                                       'out_octets': '201145667058',
                                                                       'out_ucast_pkts': '143425766',
                                                                       'poller_hash': 'REZGRTQ0cGVlcmludGYyMjE0NjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '221465',
                                                                       'in_nucast_pkts': '1464577',
                                                                       'in_octets': '201145667058',
                                                                       'in_ucast_pkts': '143425766',
                                                                       'links': [       {       'in_nucast_pkts': '1464577',
                                                                                                'in_octets': '201145667058',
                                                                                                'in_ucast_pkts': '143425766',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '4724637248',
                                                                                                'out_ucast_pkts': '34746781',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '4724637248',
                                                                       'out_ucast_pkts': '34746781',
                                                                       'poller_hash': 'REZGRTQ0cGVlcmludGZwb24gMjIxNDY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221465',
                                                                                               'links': [       {       'connected_time': '688506',
                                                                                                                        'fiber_length': 621,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTQ0cGVlcnBvbnBvbiAyMjE0NjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 MAIN ST N (SUTHERLAND RAVEN)',
                                         'poller_hash': 'REZGRTQ0',
                                         'serial_number': 'DFFE44',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2a:d8',
                                         'uptime': '688506'},
                                 {       'counter_source': '221600/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b4:fb:e4:8e:43:79'],
                                                                       'id': '221600/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '71321167490',
                                                                       'in_ucast_pkts': '113599102',
                                                                       'mac_address': '04:bc:9f:5c:4c:44',
                                                                       'mtu': 2000,
                                                                       'name': '221600/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1381296',
                                                                       'out_octets': '258337545616',
                                                                       'out_ucast_pkts': '217772221',
                                                                       'poller_hash': 'MTI3NTAwOHBlZXJpbnRmMjIxNjAwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '221600',
                                                                       'in_nucast_pkts': '1381296',
                                                                       'in_octets': '258337545616',
                                                                       'in_ucast_pkts': '217772221',
                                                                       'links': [       {       'in_nucast_pkts': '1381296',
                                                                                                'in_octets': '258337545616',
                                                                                                'in_ucast_pkts': '217772221',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '71321167490',
                                                                                                'out_ucast_pkts': '113599102',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '71321167490',
                                                                       'out_ucast_pkts': '113599102',
                                                                       'poller_hash': 'MTI3NTAwOHBlZXJpbnRmcG9uIDIyMTYwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221600',
                                                                                               'links': [       {       'connected_time': '648763',
                                                                                                                        'fiber_length': 909,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTAwOHBlZXJwb25wb24gMjIxNjAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 ST MARYS ST N (DRAGOI CRISTIAN & '
                                                 'ADRIANA)',
                                         'poller_hash': 'MTI3NTAwOA==',
                                         'serial_number': '1275008',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4c:41',
                                         'uptime': '648763'},
                                 {       'counter_source': '221858/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '221858/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '95844413562',
                                                                       'in_ucast_pkts': '727488710',
                                                                       'mac_address': '04:bc:9f:5c:4f:91',
                                                                       'name': '221858/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '2181348364913',
                                                                       'out_ucast_pkts': '1636470433',
                                                                       'poller_hash': 'MTI3NTA3MnBlZXJpbnRmMjIxODU4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '221858',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '2181348364913',
                                                                       'in_ucast_pkts': '1636470433',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '2181348364913',
                                                                                                'in_ucast_pkts': '1636470433',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '95844413562',
                                                                                                'out_ucast_pkts': '727488710',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '95844413562',
                                                                       'out_ucast_pkts': '727488710',
                                                                       'poller_hash': 'MTI3NTA3MnBlZXJpbnRmcG9uIDIyMTg1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221858',
                                                                                               'links': [       {       'connected_time': '3113962',
                                                                                                                        'fiber_length': 17591,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTA3MnBlZXJwb25wb24gMjIxODU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '160 FISHLEIGH ST MONKTON (MCGIRR '
                                                 'KENNETH  10433784)',
                                         'poller_hash': 'MTI3NTA3Mg==',
                                         'serial_number': '1275072',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4f:91',
                                         'uptime': '3113962'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '224068/g1',
                                                                       'name': '224068/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NTE4QnBlZXJpbnRmMjI0MDY4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '224068',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NTE4QnBlZXJpbnRmcG9uIDIyNDA2OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224068',
                                                                                               'links': [       {       'connected_time': '5966548',
                                                                                                                        'fiber_length': 6412,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTE4QnBlZXJwb25wb24gMjI0MDY4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4482 PERTH LINE 72 (HOERLE MARY)',
                                         'poller_hash': 'MTI3NTE4Qg==',
                                         'serial_number': '127518B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:58:59',
                                         'uptime': '5966548'},
                                 {       'counter_source': '224432/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:09:0f:09:01:02'],
                                                                       'id': '224432/g1',
                                                                       'in_nucast_pkts': '955',
                                                                       'in_octets': '81049090267702',
                                                                       'in_ucast_pkts': '68720615352',
                                                                       'mac_address': '00:06:31:2e:ad:4e',
                                                                       'mtu': 2000,
                                                                       'name': '224432/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150423145',
                                                                       'out_octets': '34036842973018',
                                                                       'out_ucast_pkts': '56666971536',
                                                                       'poller_hash': 'NjZENTNwZWVyaW50ZjIyNDQzMi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '224432',
                                                                       'in_nucast_pkts': '150423145',
                                                                       'in_octets': '34036842973018',
                                                                       'in_ucast_pkts': '56666971536',
                                                                       'links': [       {       'in_nucast_pkts': '150423145',
                                                                                                'in_octets': '34036842973018',
                                                                                                'in_ucast_pkts': '56666971536',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '955',
                                                                                                'out_octets': '81049090267702',
                                                                                                'out_ucast_pkts': '68720615352',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '955',
                                                                       'out_octets': '81049090267702',
                                                                       'out_ucast_pkts': '68720615352',
                                                                       'poller_hash': 'NjZENTNwZWVyaW50ZnBvbiAyMjQ0MzI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224432',
                                                                                               'links': [       {       'connected_time': '15493579',
                                                                                                                        'fiber_length': 1688,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjZENTNwZWVycG9ucG9uIDIyNDQzMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '25 WHALEY AVE MILVERTON (BOSHART '
                                                 'INDUSTRIES)',
                                         'poller_hash': 'NjZENTM=',
                                         'serial_number': '66D53',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:2e:ad:53',
                                         'uptime': '15493579'},
                                 {       'counter_source': '224639/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '224639/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '292',
                                                                       'in_octets': '200243195514',
                                                                       'in_ucast_pkts': '724887725',
                                                                       'mac_address': 'b8:94:70:ba:e3:e5',
                                                                       'name': '224639/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1315',
                                                                       'out_octets': '2658846181298',
                                                                       'out_ucast_pkts': '1921819834',
                                                                       'poller_hash': 'MTE5NTE5MXBlZXJpbnRmMjI0NjM5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '224639',
                                                                       'in_nucast_pkts': '1315',
                                                                       'in_octets': '2658846181298',
                                                                       'in_ucast_pkts': '1921819834',
                                                                       'links': [       {       'in_nucast_pkts': '1315',
                                                                                                'in_octets': '2658846181298',
                                                                                                'in_ucast_pkts': '1921819834',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '292',
                                                                                                'out_octets': '200243195514',
                                                                                                'out_ucast_pkts': '724887725',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '292',
                                                                       'out_octets': '200243195514',
                                                                       'out_ucast_pkts': '724887725',
                                                                       'poller_hash': 'MTE5NTE5MXBlZXJpbnRmcG9uIDIyNDYzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224639',
                                                                                               'links': [       {       'connected_time': '688477',
                                                                                                                        'fiber_length': 872,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE5MXBlZXJwb25wb24gMjI0NjM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "22 WILLIAM ST (O'MALLEY CAITLIN)",
                                         'poller_hash': 'MTE5NTE5MQ==',
                                         'serial_number': '1195191',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e3:e5',
                                         'uptime': '688477'},
                                 {       'counter_source': '224689/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:d5:6d'],
                                                                       'id': '224689/g1',
                                                                       'in_nucast_pkts': '210472',
                                                                       'in_octets': '276158214036',
                                                                       'in_ucast_pkts': '727050033',
                                                                       'mac_address': '04:bc:9f:5c:1f:f4',
                                                                       'mtu': 2000,
                                                                       'name': '224689/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24505549',
                                                                       'out_octets': '2994270542936',
                                                                       'out_ucast_pkts': '2341710240',
                                                                       'poller_hash': 'MTI3NEE3RXBlZXJpbnRmMjI0Njg5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '224689',
                                                                       'in_nucast_pkts': '24505549',
                                                                       'in_octets': '2994270542936',
                                                                       'in_ucast_pkts': '2341710240',
                                                                       'links': [       {       'in_nucast_pkts': '24505549',
                                                                                                'in_octets': '2994270542936',
                                                                                                'in_ucast_pkts': '2341710240',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '210472',
                                                                                                'out_octets': '276158214036',
                                                                                                'out_ucast_pkts': '727050033',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '210472',
                                                                       'out_octets': '276158214036',
                                                                       'out_ucast_pkts': '727050033',
                                                                       'poller_hash': 'MTI3NEE3RXBlZXJpbnRmcG9uIDIyNDY4OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224689',
                                                                                               'links': [       {       'connected_time': '12865046',
                                                                                                                        'fiber_length': 15029,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE3RXBlZXJwb25wb24gMjI0Njg5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5198 LINE 49 (GRAUL NICOLE)',
                                         'poller_hash': 'MTI3NEE3RQ==',
                                         'serial_number': '1274A7E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:f1',
                                         'uptime': '12865046'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '224699/g1',
                                                                       'name': '224699/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NTBDQ3BlZXJpbnRmMjI0Njk5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '224699',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NTBDQ3BlZXJpbnRmcG9uIDIyNDY5OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224699',
                                                                                               'links': [       {       'connected_time': '719770',
                                                                                                                        'fiber_length': 10129,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTBDQ3BlZXJwb25wb24gMjI0Njk5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5134 PERTH LINE 55 (JANTZI MARTIN & '
                                                 'PAULINE)',
                                         'poller_hash': 'MTI3NTBDQw==',
                                         'serial_number': '12750CC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:52:61',
                                         'uptime': '719770'},
                                 {       'counter_source': '224739/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '224739/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '363268840002',
                                                                       'in_ucast_pkts': '1293017247',
                                                                       'mac_address': '04:bc:9f:5c:5a:d1',
                                                                       'name': '224739/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '120',
                                                                       'out_octets': '5404303454918',
                                                                       'out_ucast_pkts': '4225601764',
                                                                       'poller_hash': 'MTI3NTFEQXBlZXJpbnRmMjI0NzM5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '224739',
                                                                       'in_nucast_pkts': '120',
                                                                       'in_octets': '5404303454918',
                                                                       'in_ucast_pkts': '4225601764',
                                                                       'links': [       {       'in_nucast_pkts': '120',
                                                                                                'in_octets': '5404303454918',
                                                                                                'in_ucast_pkts': '4225601764',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '363268840002',
                                                                                                'out_ucast_pkts': '1293017247',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '363268840002',
                                                                       'out_ucast_pkts': '1293017247',
                                                                       'poller_hash': 'MTI3NTFEQXBlZXJpbnRmcG9uIDIyNDczOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224739',
                                                                                               'links': [       {       'connected_time': '688459',
                                                                                                                        'fiber_length': 1573,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTFEQXBlZXJwb25wb24gMjI0NzM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '57 WILLIAM ST APT 202 (HOWATT '
                                                 'MITCHELL)',
                                         'poller_hash': 'MTI3NTFEQQ==',
                                         'serial_number': '12751DA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:5a:d1',
                                         'uptime': '688459'},
                                 {       'counter_source': '224759/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:89:07'],
                                                                       'id': '224759/g1',
                                                                       'in_nucast_pkts': '14313',
                                                                       'in_octets': '4759764367',
                                                                       'in_ucast_pkts': '13761013',
                                                                       'mac_address': '04:bc:9f:5c:08:d4',
                                                                       'mtu': 2000,
                                                                       'name': '224759/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464640',
                                                                       'out_octets': '169908376600',
                                                                       'out_ucast_pkts': '121188918',
                                                                       'poller_hash': 'MTI3NDc5QXBlZXJpbnRmMjI0NzU5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '224759',
                                                                       'in_nucast_pkts': '1464640',
                                                                       'in_octets': '169908376600',
                                                                       'in_ucast_pkts': '121188918',
                                                                       'links': [       {       'in_nucast_pkts': '1464640',
                                                                                                'in_octets': '169908376600',
                                                                                                'in_ucast_pkts': '121188918',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14313',
                                                                                                'out_octets': '4759764367',
                                                                                                'out_ucast_pkts': '13761013',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14313',
                                                                       'out_octets': '4759764367',
                                                                       'out_ucast_pkts': '13761013',
                                                                       'poller_hash': 'MTI3NDc5QXBlZXJpbnRmcG9uIDIyNDc1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224759',
                                                                                               'links': [       {       'connected_time': '688463',
                                                                                                                        'fiber_length': 1577,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc5QXBlZXJwb25wb24gMjI0NzU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '57 WILLIAM ST (EVANS HEATHER)',
                                         'poller_hash': 'MTI3NDc5QQ==',
                                         'serial_number': '127479A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:08:d1',
                                         'uptime': '688463'},
                                 {       'counter_source': '224769/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:94:87'],
                                                                       'id': '224769/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '3518295021',
                                                                       'in_ucast_pkts': '14725953',
                                                                       'mac_address': '04:bc:9f:5c:5a:0c',
                                                                       'mtu': 2000,
                                                                       'name': '224769/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464696',
                                                                       'out_octets': '108269970399',
                                                                       'out_ucast_pkts': '82149703',
                                                                       'poller_hash': 'MTI3NTFDMXBlZXJpbnRmMjI0NzY5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '224769',
                                                                       'in_nucast_pkts': '1464696',
                                                                       'in_octets': '108269970399',
                                                                       'in_ucast_pkts': '82149703',
                                                                       'links': [       {       'in_nucast_pkts': '1464696',
                                                                                                'in_octets': '108269970399',
                                                                                                'in_ucast_pkts': '82149703',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '3518295021',
                                                                                                'out_ucast_pkts': '14725953',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '3518295021',
                                                                       'out_ucast_pkts': '14725953',
                                                                       'poller_hash': 'MTI3NTFDMXBlZXJpbnRmcG9uIDIyNDc2OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224769',
                                                                                               'links': [       {       'connected_time': '688461',
                                                                                                                        'fiber_length': 1572,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTFDMXBlZXJwb25wb24gMjI0NzY5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '57 WILLIAM ST (OWUSU SELINA)',
                                         'poller_hash': 'MTI3NTFDMQ==',
                                         'serial_number': '12751C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:5a:09',
                                         'uptime': '688461'},
                                 {       'counter_source': '224799/G1+224799/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '224799/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '22888927615',
                                                                       'in_ucast_pkts': '139272156',
                                                                       'mac_address': '04:bc:9f:5c:49:19',
                                                                       'name': '224799/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '78961089115',
                                                                       'out_ucast_pkts': '147036053',
                                                                       'poller_hash': 'MTI3NEZBM3BlZXJpbnRmMjI0Nzk5L0cx'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:90:7f:19:e4:a2'],
                                                                       'id': '224799/g1',
                                                                       'in_nucast_pkts': '185076',
                                                                       'in_octets': '29251319497294',
                                                                       'in_ucast_pkts': '25996232874',
                                                                       'mac_address': '04:bc:9f:5c:49:1c',
                                                                       'mtu': 2000,
                                                                       'name': '224799/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '131447227',
                                                                       'out_octets': '17006044066351',
                                                                       'out_ucast_pkts': '17658744223',
                                                                       'poller_hash': 'MTI3NEZBM3BlZXJpbnRmMjI0Nzk5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '224799',
                                                                       'in_nucast_pkts': '131447234',
                                                                       'in_octets': '17085005155466',
                                                                       'in_ucast_pkts': '17805780276',
                                                                       'links': [       {       'in_nucast_pkts': '131447234',
                                                                                                'in_octets': '17085005155466',
                                                                                                'in_ucast_pkts': '17805780276',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '185076',
                                                                                                'out_octets': '29274208424909',
                                                                                                'out_ucast_pkts': '26135505030',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '185076',
                                                                       'out_octets': '29274208424909',
                                                                       'out_ucast_pkts': '26135505030',
                                                                       'poller_hash': 'MTI3NEZBM3BlZXJpbnRmcG9uIDIyNDc5OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224799',
                                                                                               'links': [       {       'connected_time': '11227824',
                                                                                                                        'fiber_length': 5439,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEZBM3BlZXJwb25wb24gMjI0Nzk5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4570 PERTH LINE 72 (BATEMAN '
                                                 'EQUIPMENT CORPORATION)',
                                         'poller_hash': 'MTI3NEZBMw==',
                                         'serial_number': '1274FA3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:49:19',
                                         'uptime': '11227824'},
                                 {       'counter_source': '224872/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '224872/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '112332165902',
                                                                       'in_ucast_pkts': '889776891',
                                                                       'mac_address': '04:bc:9f:5c:5a:e9',
                                                                       'name': '224872/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '72',
                                                                       'out_octets': '5184818483767',
                                                                       'out_ucast_pkts': '3602206490',
                                                                       'poller_hash': 'MTI3NTFERHBlZXJpbnRmMjI0ODcyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '224872',
                                                                       'in_nucast_pkts': '72',
                                                                       'in_octets': '5184818483767',
                                                                       'in_ucast_pkts': '3602206490',
                                                                       'links': [       {       'in_nucast_pkts': '72',
                                                                                                'in_octets': '5184818483767',
                                                                                                'in_ucast_pkts': '3602206490',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '112332165902',
                                                                                                'out_ucast_pkts': '889776891',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '112332165902',
                                                                       'out_ucast_pkts': '889776891',
                                                                       'poller_hash': 'MTI3NTFERHBlZXJpbnRmcG9uIDIyNDg3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224872',
                                                                                               'links': [       {       'connected_time': '68521034',
                                                                                                                        'fiber_length': 11981,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTFERHBlZXJwb25wb24gMjI0ODcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6968 MILLBANK MAIN ST (FRANKLIN '
                                                 'GARY)',
                                         'poller_hash': 'MTI3NTFERA==',
                                         'serial_number': '12751DD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:5a:e9',
                                         'uptime': '68521034'},
                                 {       'counter_source': '226837/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '226837/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '214562418988',
                                                                       'in_ucast_pkts': '13565412069',
                                                                       'mac_address': '04:bc:9f:5c:4e:19',
                                                                       'name': '226837/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '68',
                                                                       'out_octets': '3597895512891',
                                                                       'out_ucast_pkts': '2675361475',
                                                                       'poller_hash': 'MTI3NTA0M3BlZXJpbnRmMjI2ODM3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '226837',
                                                                       'in_nucast_pkts': '68',
                                                                       'in_octets': '3597895512891',
                                                                       'in_ucast_pkts': '2675361475',
                                                                       'links': [       {       'in_nucast_pkts': '68',
                                                                                                'in_octets': '3597895512891',
                                                                                                'in_ucast_pkts': '2675361475',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '214562418988',
                                                                                                'out_ucast_pkts': '13565412069',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '214562418988',
                                                                       'out_ucast_pkts': '13565412069',
                                                                       'poller_hash': 'MTI3NTA0M3BlZXJpbnRmcG9uIDIyNjgzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '226837',
                                                                                               'links': [       {       'connected_time': '688578',
                                                                                                                        'fiber_length': 4900,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTA0M3BlZXJwb25wb24gMjI2ODM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4619 PERTH LINE 72 (MCTAVISH '
                                                 'COLLEEN)',
                                         'poller_hash': 'MTI3NTA0Mw==',
                                         'serial_number': '1275043',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4e:19',
                                         'uptime': '688578'},
                                 {       'counter_source': '226975/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '226975/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '296937424082',
                                                                       'in_ucast_pkts': '86794845813',
                                                                       'mac_address': '04:bc:9f:5c:5a:81',
                                                                       'name': '226975/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '2761373519864',
                                                                       'out_ucast_pkts': '2310631327',
                                                                       'poller_hash': 'MTI3NTFEMHBlZXJpbnRmMjI2OTc1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '226975',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '2761373519864',
                                                                       'in_ucast_pkts': '2310631327',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '2761373519864',
                                                                                                'in_ucast_pkts': '2310631327',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '296937424082',
                                                                                                'out_ucast_pkts': '86794845813',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '296937424082',
                                                                       'out_ucast_pkts': '86794845813',
                                                                       'poller_hash': 'MTI3NTFEMHBlZXJpbnRmcG9uIDIyNjk3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '226975',
                                                                                               'links': [       {       'connected_time': '10707545',
                                                                                                                        'fiber_length': 6134,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTFEMHBlZXJwb25wb24gMjI2OTc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4516 PERTH LINE 72 (REIST STEVEN)',
                                         'poller_hash': 'MTI3NTFEMA==',
                                         'serial_number': '12751D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:5a:81',
                                         'uptime': '10707545'},
                                 {       'counter_source': '227065/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '227065/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '178302794693',
                                                                       'in_ucast_pkts': '1124142024',
                                                                       'mac_address': '04:bc:9f:5c:4b:d9',
                                                                       'name': '227065/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '4310603161820',
                                                                       'out_ucast_pkts': '3196823674',
                                                                       'poller_hash': 'MTI3NEZGQnBlZXJpbnRmMjI3MDY1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '227065',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '4310603161820',
                                                                       'in_ucast_pkts': '3196823674',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '4310603161820',
                                                                                                'in_ucast_pkts': '3196823674',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '178302794693',
                                                                                                'out_ucast_pkts': '1124142024',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '178302794693',
                                                                       'out_ucast_pkts': '1124142024',
                                                                       'poller_hash': 'MTI3NEZGQnBlZXJpbnRmcG9uIDIyNzA2NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '227065',
                                                                                               'links': [       {       'connected_time': '10361223',
                                                                                                                        'fiber_length': 6572,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEZGQnBlZXJwb25wb24gMjI3MDY1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4492 PERTH LINE 72 (ALBRECHT AMMON)',
                                         'poller_hash': 'MTI3NEZGQg==',
                                         'serial_number': '1274FFB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4b:d9',
                                         'uptime': '10361223'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '227075/g1',
                                                                       'name': '227075/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NTFENnBlZXJpbnRmMjI3MDc1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '227075',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NTFENnBlZXJpbnRmcG9uIDIyNzA3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '227075',
                                                                                               'links': [       {       'connected_time': '10356939',
                                                                                                                        'fiber_length': 5952,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTFENnBlZXJwb25wb24gMjI3MDc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4528 PERTH LINE 72 (THIESSEN JACOB)',
                                         'poller_hash': 'MTI3NTFENg==',
                                         'serial_number': '12751D6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:5a:b1',
                                         'uptime': '10356939'},
                                 {       'counter_source': '228318/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '228318/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '241655840388',
                                                                       'in_ucast_pkts': '626015121',
                                                                       'mac_address': '04:bc:9f:5c:4e:99',
                                                                       'name': '228318/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '1671940173743',
                                                                       'out_ucast_pkts': '1489212934',
                                                                       'poller_hash': 'MTI3NTA1M3BlZXJpbnRmMjI4MzE4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '228318',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '1671940173743',
                                                                       'in_ucast_pkts': '1489212934',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '1671940173743',
                                                                                                'in_ucast_pkts': '1489212934',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '241655840388',
                                                                                                'out_ucast_pkts': '626015121',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '241655840388',
                                                                       'out_ucast_pkts': '626015121',
                                                                       'poller_hash': 'MTI3NTA1M3BlZXJpbnRmcG9uIDIyODMxOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '228318',
                                                                                               'links': [       {       'connected_time': '23690839',
                                                                                                                        'fiber_length': 4954,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTA1M3BlZXJwb25wb24gMjI4MzE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4620 PERTH LINE 72 (MCKERTCHER '
                                                 'JOYCE)',
                                         'poller_hash': 'MTI3NTA1Mw==',
                                         'serial_number': '1275053',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4e:99',
                                         'uptime': '23690839'},
                                 {       'counter_source': '228515/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '228515/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '78264383133',
                                                                       'in_ucast_pkts': '297619804',
                                                                       'mac_address': '04:bc:9f:5c:4b:99',
                                                                       'name': '228515/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '1123892797549',
                                                                       'out_ucast_pkts': '907867046',
                                                                       'poller_hash': 'MTI3NEZGM3BlZXJpbnRmMjI4NTE1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '228515',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '1123892797549',
                                                                       'in_ucast_pkts': '907867046',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '1123892797549',
                                                                                                'in_ucast_pkts': '907867046',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '78264383133',
                                                                                                'out_ucast_pkts': '297619804',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '78264383133',
                                                                       'out_ucast_pkts': '297619804',
                                                                       'poller_hash': 'MTI3NEZGM3BlZXJpbnRmcG9uIDIyODUxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '228515',
                                                                                               'links': [       {       'connected_time': '10115362',
                                                                                                                        'fiber_length': 7932,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEZGM3BlZXJwb25wb24gMjI4NTE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4317 PERTH LINE 72 (KUEPFER SANFORD)',
                                         'poller_hash': 'MTI3NEZGMw==',
                                         'serial_number': '1274FF3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4b:99',
                                         'uptime': '10115362'},
                                 {       'counter_source': '228535/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ea:4a'],
                                                                       'id': '228535/g1',
                                                                       'in_nucast_pkts': '14652',
                                                                       'in_octets': '6486487994',
                                                                       'in_ucast_pkts': '15207215',
                                                                       'mac_address': '04:bc:9f:5c:4d:d4',
                                                                       'mtu': 2000,
                                                                       'name': '228535/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464646',
                                                                       'out_octets': '104464494839',
                                                                       'out_ucast_pkts': '79271787',
                                                                       'poller_hash': 'MTI3NTAzQXBlZXJpbnRmMjI4NTM1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '228535',
                                                                       'in_nucast_pkts': '1464646',
                                                                       'in_octets': '104464494839',
                                                                       'in_ucast_pkts': '79271787',
                                                                       'links': [       {       'in_nucast_pkts': '1464646',
                                                                                                'in_octets': '104464494839',
                                                                                                'in_ucast_pkts': '79271787',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14652',
                                                                                                'out_octets': '6486487994',
                                                                                                'out_ucast_pkts': '15207215',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14652',
                                                                       'out_octets': '6486487994',
                                                                       'out_ucast_pkts': '15207215',
                                                                       'poller_hash': 'MTI3NTAzQXBlZXJpbnRmcG9uIDIyODUzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '228535',
                                                                                               'links': [       {       'connected_time': '688565',
                                                                                                                        'fiber_length': 8927,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTAzQXBlZXJwb25wb24gMjI4NTM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4244 PERTH LINE 72 (ALBRECHT ALEX)',
                                         'poller_hash': 'MTI3NTAzQQ==',
                                         'serial_number': '127503A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4d:d1',
                                         'uptime': '688565'},
                                 {       'counter_source': '228712/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '228712/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '391249884105',
                                                                       'in_ucast_pkts': '1909279673',
                                                                       'mac_address': '04:bc:9f:5c:58:b9',
                                                                       'name': '228712/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '69',
                                                                       'out_octets': '9686111166412',
                                                                       'out_ucast_pkts': '6810406407',
                                                                       'poller_hash': 'MTI3NTE5N3BlZXJpbnRmMjI4NzEyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '228712',
                                                                       'in_nucast_pkts': '69',
                                                                       'in_octets': '9686111166412',
                                                                       'in_ucast_pkts': '6810406407',
                                                                       'links': [       {       'in_nucast_pkts': '69',
                                                                                                'in_octets': '9686111166412',
                                                                                                'in_ucast_pkts': '6810406407',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '391249884105',
                                                                                                'out_ucast_pkts': '1909279673',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '391249884105',
                                                                       'out_ucast_pkts': '1909279673',
                                                                       'poller_hash': 'MTI3NTE5N3BlZXJpbnRmcG9uIDIyODcxMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '228712',
                                                                                               'links': [       {       'connected_time': '538555',
                                                                                                                        'fiber_length': 14462,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTE5N3BlZXJwb25wb24gMjI4NzEy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4870 LINE 49 (TAMAS ALEXANDRA)',
                                         'poller_hash': 'MTI3NTE5Nw==',
                                         'serial_number': '1275197',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:58:b9',
                                         'uptime': '538555'},
                                 {       'counter_source': '229548/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229548/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '94226339229',
                                                                       'in_ucast_pkts': '304129863',
                                                                       'mac_address': '04:bc:9f:5b:e2:d9',
                                                                       'name': '229548/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '43',
                                                                       'out_octets': '1219167609999',
                                                                       'out_ucast_pkts': '1028394564',
                                                                       'poller_hash': 'MTI3NDJEQnBlZXJpbnRmMjI5NTQ4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229548',
                                                                       'in_nucast_pkts': '43',
                                                                       'in_octets': '1219167609999',
                                                                       'in_ucast_pkts': '1028394564',
                                                                       'links': [       {       'in_nucast_pkts': '43',
                                                                                                'in_octets': '1219167609999',
                                                                                                'in_ucast_pkts': '1028394564',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '94226339229',
                                                                                                'out_ucast_pkts': '304129863',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '94226339229',
                                                                       'out_ucast_pkts': '304129863',
                                                                       'poller_hash': 'MTI3NDJEQnBlZXJpbnRmcG9uIDIyOTU0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229548',
                                                                                               'links': [       {       'connected_time': '688560',
                                                                                                                        'fiber_length': 7325,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJEQnBlZXJwb25wb24gMjI5NTQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4375 PERTH LINE 72 (FRIESEN ARNO)',
                                         'poller_hash': 'MTI3NDJEQg==',
                                         'serial_number': '12742DB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e2:d9',
                                         'uptime': '688560'},
                                 {       'counter_source': '229558/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229558/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '250546673376',
                                                                       'in_ucast_pkts': '207831004',
                                                                       'mac_address': '04:bc:9f:5c:25:31',
                                                                       'name': '229558/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '44',
                                                                       'out_octets': '78549376616',
                                                                       'out_ucast_pkts': '207093604',
                                                                       'poller_hash': 'MTI3NEIyNnBlZXJpbnRmMjI5NTU4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229558',
                                                                       'in_nucast_pkts': '44',
                                                                       'in_octets': '78549376616',
                                                                       'in_ucast_pkts': '207093604',
                                                                       'links': [       {       'in_nucast_pkts': '44',
                                                                                                'in_octets': '78549376616',
                                                                                                'in_ucast_pkts': '207093604',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '250546673376',
                                                                                                'out_ucast_pkts': '207831004',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '250546673376',
                                                                       'out_ucast_pkts': '207831004',
                                                                       'poller_hash': 'MTI3NEIyNnBlZXJpbnRmcG9uIDIyOTU1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229558',
                                                                                               'links': [       {       'connected_time': '688554',
                                                                                                                        'fiber_length': 10189,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIyNnBlZXJwb25wb24gMjI5NTU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4138 PERTH LINE 72 (BEREAN CHRISTIAN '
                                                 'FELLOWSHIP)',
                                         'poller_hash': 'MTI3NEIyNg==',
                                         'serial_number': '1274B26',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:25:31',
                                         'uptime': '688554'},
                                 {       'counter_source': '229568/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229568/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '441745612553',
                                                                       'in_ucast_pkts': '1479836709',
                                                                       'mac_address': '04:bc:9f:5b:e3:c1',
                                                                       'name': '229568/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '4273621303522',
                                                                       'out_ucast_pkts': '3534016833',
                                                                       'poller_hash': 'MTI3NDJGOHBlZXJpbnRmMjI5NTY4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229568',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '4273621303522',
                                                                       'in_ucast_pkts': '3534016833',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '4273621303522',
                                                                                                'in_ucast_pkts': '3534016833',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '441745612553',
                                                                                                'out_ucast_pkts': '1479836709',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '441745612553',
                                                                       'out_ucast_pkts': '1479836709',
                                                                       'poller_hash': 'MTI3NDJGOHBlZXJpbnRmcG9uIDIyOTU2OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229568',
                                                                                               'links': [       {       'connected_time': '16867954',
                                                                                                                        'fiber_length': 8561,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJGOHBlZXJwb25wb24gMjI5NTY4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7007 ROAD 124 (LEIS ROBERT)',
                                         'poller_hash': 'MTI3NDJGOA==',
                                         'serial_number': '12742F8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:c1',
                                         'uptime': '16867954'},
                                 {       'counter_source': '229578/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229578/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '38516363480',
                                                                       'in_ucast_pkts': '105530441',
                                                                       'mac_address': '04:bc:9f:5c:23:f1',
                                                                       'name': '229578/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '305220549728',
                                                                       'out_ucast_pkts': '317511830',
                                                                       'poller_hash': 'MTI3NEFGRXBlZXJpbnRmMjI5NTc4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229578',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '305220549728',
                                                                       'in_ucast_pkts': '317511830',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '305220549728',
                                                                                                'in_ucast_pkts': '317511830',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '38516363480',
                                                                                                'out_ucast_pkts': '105530441',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '38516363480',
                                                                       'out_ucast_pkts': '105530441',
                                                                       'poller_hash': 'MTI3NEFGRXBlZXJpbnRmcG9uIDIyOTU3OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229578',
                                                                                               'links': [       {       'connected_time': '9686120',
                                                                                                                        'fiber_length': 10041,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFGRXBlZXJwb25wb24gMjI5NTc4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4147 PERTH LINE 72 (GERBER SHARON)',
                                         'poller_hash': 'MTI3NEFGRQ==',
                                         'serial_number': '1274AFE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:23:f1',
                                         'uptime': '9686120'},
                                 {       'counter_source': '229588/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229588/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '31218444575',
                                                                       'in_ucast_pkts': '153576766',
                                                                       'mac_address': '04:bc:9f:5b:e1:29',
                                                                       'name': '229588/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '94',
                                                                       'out_octets': '413332880300',
                                                                       'out_ucast_pkts': '405048546',
                                                                       'poller_hash': 'MTI3NDJBNXBlZXJpbnRmMjI5NTg4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229588',
                                                                       'in_nucast_pkts': '94',
                                                                       'in_octets': '413332880300',
                                                                       'in_ucast_pkts': '405048546',
                                                                       'links': [       {       'in_nucast_pkts': '94',
                                                                                                'in_octets': '413332880300',
                                                                                                'in_ucast_pkts': '405048546',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '31218444575',
                                                                                                'out_ucast_pkts': '153576766',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '31218444575',
                                                                       'out_ucast_pkts': '153576766',
                                                                       'poller_hash': 'MTI3NDJBNXBlZXJpbnRmcG9uIDIyOTU4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229588',
                                                                                               'links': [       {       'connected_time': '2056952',
                                                                                                                        'fiber_length': 9745,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJBNXBlZXJwb25wb24gMjI5NTg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4896 PERTH LINE 72 (CARTER MARY)',
                                         'poller_hash': 'MTI3NDJBNQ==',
                                         'serial_number': '12742A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e1:29',
                                         'uptime': '2056952'},
                                 {       'counter_source': '229651/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229651/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '793305122763',
                                                                       'in_ucast_pkts': '250352909946',
                                                                       'mac_address': '04:bc:9f:5c:20:49',
                                                                       'name': '229651/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '43',
                                                                       'out_octets': '4758019287442',
                                                                       'out_ucast_pkts': '3835575889',
                                                                       'poller_hash': 'MTI3NEE4OXBlZXJpbnRmMjI5NjUxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229651',
                                                                       'in_nucast_pkts': '43',
                                                                       'in_octets': '4758019287442',
                                                                       'in_ucast_pkts': '3835575889',
                                                                       'links': [       {       'in_nucast_pkts': '43',
                                                                                                'in_octets': '4758019287442',
                                                                                                'in_ucast_pkts': '3835575889',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '793305122763',
                                                                                                'out_ucast_pkts': '250352909946',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '793305122763',
                                                                       'out_ucast_pkts': '250352909946',
                                                                       'poller_hash': 'MTI3NEE4OXBlZXJpbnRmcG9uIDIyOTY1MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229651',
                                                                                               'links': [       {       'connected_time': '688565',
                                                                                                                        'fiber_length': 8937,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE4OXBlZXJwb25wb24gMjI5NjUx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6946 ROAD 124 (JANTZI JONATHAN)',
                                         'poller_hash': 'MTI3NEE4OQ==',
                                         'serial_number': '1274A89',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:49',
                                         'uptime': '688565'},
                                 {       'counter_source': '229721/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229721/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '68042206030',
                                                                       'in_ucast_pkts': '243903220',
                                                                       'mac_address': '04:bc:9f:5c:21:01',
                                                                       'name': '229721/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '44',
                                                                       'out_octets': '1048476343935',
                                                                       'out_ucast_pkts': '827684392',
                                                                       'poller_hash': 'MTI3NEFBMHBlZXJpbnRmMjI5NzIxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229721',
                                                                       'in_nucast_pkts': '44',
                                                                       'in_octets': '1048476343935',
                                                                       'in_ucast_pkts': '827684392',
                                                                       'links': [       {       'in_nucast_pkts': '44',
                                                                                                'in_octets': '1048476343935',
                                                                                                'in_ucast_pkts': '827684392',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '68042206030',
                                                                                                'out_ucast_pkts': '243903220',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '68042206030',
                                                                       'out_ucast_pkts': '243903220',
                                                                       'poller_hash': 'MTI3NEFBMHBlZXJpbnRmcG9uIDIyOTcyMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229721',
                                                                                               'links': [       {       'connected_time': '688569',
                                                                                                                        'fiber_length': 8622,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFBMHBlZXJwb25wb24gMjI5NzIx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6997 ROAD 124 (KUEPFER DANNIE)',
                                         'poller_hash': 'MTI3NEFBMA==',
                                         'serial_number': '1274AA0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:21:01',
                                         'uptime': '688569'},
                                 {       'counter_source': '229731/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229731/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '623637388428',
                                                                       'in_ucast_pkts': '65626868322',
                                                                       'mac_address': '04:bc:9f:5c:1d:b9',
                                                                       'name': '229731/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '43',
                                                                       'out_octets': '18683232478953',
                                                                       'out_ucast_pkts': '13132013532',
                                                                       'poller_hash': 'MTI3NEEzN3BlZXJpbnRmMjI5NzMxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229731',
                                                                       'in_nucast_pkts': '43',
                                                                       'in_octets': '18683232478953',
                                                                       'in_ucast_pkts': '13132013532',
                                                                       'links': [       {       'in_nucast_pkts': '43',
                                                                                                'in_octets': '18683232478953',
                                                                                                'in_ucast_pkts': '13132013532',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '623637388428',
                                                                                                'out_ucast_pkts': '65626868322',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '623637388428',
                                                                       'out_ucast_pkts': '65626868322',
                                                                       'poller_hash': 'MTI3NEEzN3BlZXJpbnRmcG9uIDIyOTczMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229731',
                                                                                               'links': [       {       'connected_time': '13114853',
                                                                                                                        'fiber_length': 9773,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEzN3BlZXJwb25wb24gMjI5NzMx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4172 PERTH LINE 72 (FENTON EUGENE & '
                                                 'LESLEY)',
                                         'poller_hash': 'MTI3NEEzNw==',
                                         'serial_number': '1274A37',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:b9',
                                         'uptime': '13114853'},
                                 {       'counter_source': '230019/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:ec:54'],
                                                                       'id': '230019/g1',
                                                                       'in_nucast_pkts': '11996',
                                                                       'in_octets': '10996664098',
                                                                       'in_ucast_pkts': '23695548',
                                                                       'mac_address': '04:bc:9f:5c:23:ac',
                                                                       'mtu': 2000,
                                                                       'name': '230019/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464635',
                                                                       'out_octets': '125443874257',
                                                                       'out_ucast_pkts': '94061558',
                                                                       'poller_hash': 'MTI3NEFGNXBlZXJpbnRmMjMwMDE5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230019',
                                                                       'in_nucast_pkts': '1464635',
                                                                       'in_octets': '125443874257',
                                                                       'in_ucast_pkts': '94061558',
                                                                       'links': [       {       'in_nucast_pkts': '1464635',
                                                                                                'in_octets': '125443874257',
                                                                                                'in_ucast_pkts': '94061558',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11996',
                                                                                                'out_octets': '10996664098',
                                                                                                'out_ucast_pkts': '23695548',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11996',
                                                                       'out_octets': '10996664098',
                                                                       'out_ucast_pkts': '23695548',
                                                                       'poller_hash': 'MTI3NEFGNXBlZXJpbnRmcG9uIDIzMDAxOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230019',
                                                                                               'links': [       {       'connected_time': '688367',
                                                                                                                        'fiber_length': 2797,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFGNXBlZXJwb25wb24gMjMwMDE5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '57 PUGH ST (HEALEY RYAN)',
                                         'poller_hash': 'MTI3NEFGNQ==',
                                         'serial_number': '1274AF5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:23:a9',
                                         'uptime': '688367'},
                                 {       'counter_source': '230064/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '230064/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '603821480599',
                                                                       'in_ucast_pkts': '2316328056',
                                                                       'mac_address': '04:bc:9f:5c:25:79',
                                                                       'name': '230064/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '167',
                                                                       'out_octets': '8306422161531',
                                                                       'out_ucast_pkts': '6169228909',
                                                                       'poller_hash': 'MTI3NEIyRnBlZXJpbnRmMjMwMDY0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '230064',
                                                                       'in_nucast_pkts': '167',
                                                                       'in_octets': '8306422161531',
                                                                       'in_ucast_pkts': '6169228909',
                                                                       'links': [       {       'in_nucast_pkts': '167',
                                                                                                'in_octets': '8306422161531',
                                                                                                'in_ucast_pkts': '6169228909',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '603821480599',
                                                                                                'out_ucast_pkts': '2316328056',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '603821480599',
                                                                       'out_ucast_pkts': '2316328056',
                                                                       'poller_hash': 'MTI3NEIyRnBlZXJpbnRmcG9uIDIzMDA2NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230064',
                                                                                               'links': [       {       'connected_time': '688569',
                                                                                                                        'fiber_length': 10155,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIyRnBlZXJwb25wb24gMjMwMDY0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4137 PERTH LINE 72 (KUEPFER JOHN K)',
                                         'poller_hash': 'MTI3NEIyRg==',
                                         'serial_number': '1274B2F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:25:79',
                                         'uptime': '688569'},
                                 {       'counter_source': '230600/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '230600/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '63876646725',
                                                                       'in_ucast_pkts': '364933106',
                                                                       'mac_address': '04:bc:9f:5c:1c:71',
                                                                       'name': '230600/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '762005095894',
                                                                       'out_ucast_pkts': '677602634',
                                                                       'poller_hash': 'MTI3NEEwRXBlZXJpbnRmMjMwNjAwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '230600',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '762005095894',
                                                                       'in_ucast_pkts': '677602634',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '762005095894',
                                                                                                'in_ucast_pkts': '677602634',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '63876646725',
                                                                                                'out_ucast_pkts': '364933106',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '63876646725',
                                                                       'out_ucast_pkts': '364933106',
                                                                       'poller_hash': 'MTI3NEEwRXBlZXJpbnRmcG9uIDIzMDYwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230600',
                                                                                               'links': [       {       'connected_time': '8980806',
                                                                                                                        'fiber_length': 9878,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwRXBlZXJwb25wb24gMjMwNjAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4898 PERTH LINE 72 (CARTER CHERYL)',
                                         'poller_hash': 'MTI3NEEwRQ==',
                                         'serial_number': '1274A0E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:71',
                                         'uptime': '8980806'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '230704/g1',
                                                                       'name': '230704/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTUzRTYzOXBlZXJpbnRmMjMwNzA0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '230704',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUzRTYzOXBlZXJpbnRmcG9uIDIzMDcwNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230704',
                                                                                               'links': [       {       'connected_time': '1397632',
                                                                                                                        'fiber_length': 9594,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUzRTYzOXBlZXJwb25wb24gMjMwNzA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6888 ROAD 124 (KIPFER LORETTA)',
                                         'poller_hash': 'MTUzRTYzOQ==',
                                         'serial_number': '153E639',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '04:bc:9f:e8:9d:e0',
                                         'uptime': '1397632'},
                                 {       'counter_source': '230774/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '230774/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '23441823793',
                                                                       'in_ucast_pkts': '120617855',
                                                                       'mac_address': '04:bc:9f:5b:d5:29',
                                                                       'name': '230774/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '568993365585',
                                                                       'out_ucast_pkts': '415045839',
                                                                       'poller_hash': 'MTI3NDEyNXBlZXJpbnRmMjMwNzc0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '230774',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '568993365585',
                                                                       'in_ucast_pkts': '415045839',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '568993365585',
                                                                                                'in_ucast_pkts': '415045839',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '23441823793',
                                                                                                'out_ucast_pkts': '120617855',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '23441823793',
                                                                       'out_ucast_pkts': '120617855',
                                                                       'poller_hash': 'MTI3NDEyNXBlZXJpbnRmcG9uIDIzMDc3NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230774',
                                                                                               'links': [       {       'connected_time': '688352',
                                                                                                                        'fiber_length': 1332,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDEyNXBlZXJwb25wb24gMjMwNzc0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '17 WEST ST (ALLEN JIM)',
                                         'poller_hash': 'MTI3NDEyNQ==',
                                         'serial_number': '1274125',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d5:29',
                                         'uptime': '688352'},
                                 {       'counter_source': '230866/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:ad'],
                                                                       'id': '230866/g1',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '1887066611',
                                                                       'in_ucast_pkts': '3138506',
                                                                       'mac_address': '04:bc:9f:5c:1f:6c',
                                                                       'mtu': 2000,
                                                                       'name': '230866/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434781',
                                                                       'out_octets': '5394064352',
                                                                       'out_ucast_pkts': '4852765',
                                                                       'poller_hash': 'MTI3NEE2RHBlZXJpbnRmMjMwODY2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230866',
                                                                       'in_nucast_pkts': '1434781',
                                                                       'in_octets': '5394064352',
                                                                       'in_ucast_pkts': '4852765',
                                                                       'links': [       {       'in_nucast_pkts': '1434781',
                                                                                                'in_octets': '5394064352',
                                                                                                'in_ucast_pkts': '4852765',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12',
                                                                                                'out_octets': '1887066611',
                                                                                                'out_ucast_pkts': '3138506',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '1887066611',
                                                                       'out_ucast_pkts': '3138506',
                                                                       'poller_hash': 'MTI3NEE2RHBlZXJpbnRmcG9uIDIzMDg2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230866',
                                                                                               'links': [       {       'connected_time': '688559',
                                                                                                                        'fiber_length': 1800,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE2RHBlZXJwb25wb24gMjMwODY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (BAUMAN ROSEANNE)',
                                         'poller_hash': 'MTI3NEE2RA==',
                                         'serial_number': '1274A6D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:69',
                                         'uptime': '688559'},
                                 {       'counter_source': '230916/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '230916/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '461505791348',
                                                                       'in_ucast_pkts': '527173990',
                                                                       'mac_address': '04:bc:9f:5c:16:31',
                                                                       'name': '230916/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '760462284553',
                                                                       'out_ucast_pkts': '765682507',
                                                                       'poller_hash': 'MTI3NDk0NnBlZXJpbnRmMjMwOTE2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '230916',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '760462284553',
                                                                       'in_ucast_pkts': '765682507',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '760462284553',
                                                                                                'in_ucast_pkts': '765682507',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '461505791348',
                                                                                                'out_ucast_pkts': '527173990',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '461505791348',
                                                                       'out_ucast_pkts': '527173990',
                                                                       'poller_hash': 'MTI3NDk0NnBlZXJpbnRmcG9uIDIzMDkxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230916',
                                                                                               'links': [       {       'connected_time': '7279529',
                                                                                                                        'fiber_length': 7847,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk0NnBlZXJwb25wb24gMjMwOTE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4348 PERTH LINE 72 (BAUMAN JOSEPH M)',
                                         'poller_hash': 'MTI3NDk0Ng==',
                                         'serial_number': '1274946',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:16:31',
                                         'uptime': '7279529'},
                                 {       'counter_source': '230926/g1+230926/g3+230926/g5+230926/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:52:86'],
                                                                       'id': '230926/g1',
                                                                       'in_nucast_pkts': '1361865',
                                                                       'in_octets': '14602097900',
                                                                       'in_ucast_pkts': '193109121',
                                                                       'mac_address': '00:06:31:26:12:b8',
                                                                       'mtu': 2000,
                                                                       'name': '230926/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1668795340',
                                                                       'out_octets': '3110314519364',
                                                                       'out_ucast_pkts': '638544227',
                                                                       'poller_hash': 'NDdFMzdwZWVyaW50ZjIzMDkyNi9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:80:08'],
                                                                       'id': '230926/g3',
                                                                       'in_nucast_pkts': '975450',
                                                                       'in_octets': '10092661037',
                                                                       'in_ucast_pkts': '128949508',
                                                                       'mac_address': '00:06:31:26:12:ba',
                                                                       'mtu': 2000,
                                                                       'name': '230926/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '48086555',
                                                                       'out_octets': '602538038425',
                                                                       'out_ucast_pkts': '416112465',
                                                                       'poller_hash': 'NDdFMzdwZWVyaW50ZjIzMDkyNi9nMw==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:7f:8a'],
                                                                       'id': '230926/g5',
                                                                       'in_nucast_pkts': '1062968',
                                                                       'in_octets': '2469222290',
                                                                       'in_ucast_pkts': '31246991',
                                                                       'mac_address': '00:06:31:22:cc:e5',
                                                                       'mtu': 2000,
                                                                       'name': '230926/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1135143309',
                                                                       'out_octets': '1671875390284',
                                                                       'out_ucast_pkts': '120599300',
                                                                       'poller_hash': 'NDdFMzdwZWVyaW50ZjIzMDkyNi9nNQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:58:aa'],
                                                                       'id': '230926/g8',
                                                                       'in_nucast_pkts': '1230867',
                                                                       'in_octets': '11718784887',
                                                                       'in_ucast_pkts': '153673515',
                                                                       'mac_address': '00:06:31:22:cc:e8',
                                                                       'mtu': 2000,
                                                                       'name': '230926/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '375579893',
                                                                       'out_octets': '1162204696566',
                                                                       'out_ucast_pkts': '503714745',
                                                                       'poller_hash': 'NDdFMzdwZWVyaW50ZjIzMDkyNi9nOA==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230926',
                                                                       'in_nucast_pkts': '3227605097',
                                                                       'in_octets': '6546932644639',
                                                                       'in_ucast_pkts': '1678970737',
                                                                       'links': [       {       'in_nucast_pkts': '3227605097',
                                                                                                'in_octets': '6546932644639',
                                                                                                'in_ucast_pkts': '1678970737',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4631150',
                                                                                                'out_octets': '38882766114',
                                                                                                'out_ucast_pkts': '506979135',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4631150',
                                                                       'out_octets': '38882766114',
                                                                       'out_ucast_pkts': '506979135',
                                                                       'poller_hash': 'NDdFMzdwZWVyaW50ZnBvbiAyMzA5MjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230926',
                                                                                               'links': [       {       'connected_time': '4165100',
                                                                                                                        'fiber_length': 2004,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDdFMzdwZWVycG9ucG9uIDIzMDkyNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST (WOODS RUTH A)',
                                         'poller_hash': 'NDdFMzc=',
                                         'serial_number': '47E37',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:26:12:bd',
                                         'uptime': '4165100'},
                                 {       'counter_source': '230944/g1+230944/g4+230944/g5+230944/g7+230944/g8',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:52:38'],
                                                                       'id': '230944/g1',
                                                                       'in_nucast_pkts': '1286351',
                                                                       'in_octets': '2623435061',
                                                                       'in_ucast_pkts': '26782131',
                                                                       'mac_address': '00:06:31:4f:9f:5b',
                                                                       'mtu': 2000,
                                                                       'name': '230944/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '123205844',
                                                                       'out_octets': '224889197560',
                                                                       'out_ucast_pkts': '83792858',
                                                                       'poller_hash': 'QkE4QzNwZWVyaW50ZjIzMDk0NC9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:17:75:30'],
                                                                       'id': '230944/g4',
                                                                       'in_nucast_pkts': '1432094',
                                                                       'in_octets': '372042317',
                                                                       'in_ucast_pkts': '1172021',
                                                                       'mac_address': '00:06:31:4f:9f:5e',
                                                                       'mtu': 2000,
                                                                       'name': '230944/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2504106785',
                                                                       'out_octets': '3363967193793',
                                                                       'out_ucast_pkts': '837311',
                                                                       'poller_hash': 'QkE4QzNwZWVyaW50ZjIzMDk0NC9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:6f:84'],
                                                                       'id': '230944/g5',
                                                                       'in_nucast_pkts': '1298079',
                                                                       'in_octets': '325640085',
                                                                       'in_ucast_pkts': '1167810',
                                                                       'mac_address': '00:06:31:6a:1b:79',
                                                                       'mtu': 2000,
                                                                       'name': '230944/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2169298356',
                                                                       'out_octets': '2905103049319',
                                                                       'out_ucast_pkts': '894266',
                                                                       'poller_hash': 'QkE4QzNwZWVyaW50ZjIzMDk0NC9nNQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:43:96:ae'],
                                                                       'id': '230944/g7',
                                                                       'in_nucast_pkts': '1315309',
                                                                       'in_octets': '2495044703',
                                                                       'in_ucast_pkts': '27669898',
                                                                       'mac_address': '00:06:31:6a:1b:7b',
                                                                       'mtu': 2000,
                                                                       'name': '230944/g7',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1206416252',
                                                                       'out_octets': '1711421227507',
                                                                       'out_ucast_pkts': '87050615',
                                                                       'poller_hash': 'QkE4QzNwZWVyaW50ZjIzMDk0NC9nNw==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5a:a0'],
                                                                       'id': '230944/g8',
                                                                       'in_nucast_pkts': '1122010',
                                                                       'in_octets': '853419364',
                                                                       'in_ucast_pkts': '3048302',
                                                                       'mac_address': '00:06:31:6a:1b:7c',
                                                                       'mtu': 2000,
                                                                       'name': '230944/g8',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '750406448',
                                                                       'out_octets': '982519204364',
                                                                       'out_ucast_pkts': '7184604',
                                                                       'poller_hash': 'QkE4QzNwZWVyaW50ZjIzMDk0NC9nOA==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230944',
                                                                       'in_nucast_pkts': '6753433685',
                                                                       'in_octets': '9187899872543',
                                                                       'in_ucast_pkts': '179759654',
                                                                       'links': [       {       'in_nucast_pkts': '6753433685',
                                                                                                'in_octets': '9187899872543',
                                                                                                'in_ucast_pkts': '179759654',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6453843',
                                                                                                'out_octets': '6669581530',
                                                                                                'out_ucast_pkts': '59840162',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6453843',
                                                                       'out_octets': '6669581530',
                                                                       'out_ucast_pkts': '59840162',
                                                                       'poller_hash': 'QkE4QzNwZWVyaW50ZnBvbiAyMzA5NDQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230944',
                                                                                               'links': [       {       'connected_time': '4164935',
                                                                                                                        'fiber_length': 2004,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QkE4QzNwZWVycG9ucG9uIDIzMDk0NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST',
                                         'poller_hash': 'QkE4QzM=',
                                         'serial_number': 'BA8C3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:4f:9f:60',
                                         'uptime': '4164935'},
                                 {       'counter_source': '230962/g1+230962/g2+230962/g3+230962/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5a:32'],
                                                                       'id': '230962/g1',
                                                                       'in_nucast_pkts': '1325599',
                                                                       'in_octets': '28023934118',
                                                                       'in_ucast_pkts': '377327007',
                                                                       'mac_address': '00:06:31:8a:da:03',
                                                                       'mtu': 2000,
                                                                       'name': '230962/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1039796796',
                                                                       'out_octets': '3039514221008',
                                                                       'out_ucast_pkts': '1198570760',
                                                                       'poller_hash': 'MTExQkMzcGVlcmludGYyMzA5NjIvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:5a:e4'],
                                                                       'id': '230962/g2',
                                                                       'in_nucast_pkts': '1281763',
                                                                       'in_octets': '1099769980',
                                                                       'in_ucast_pkts': '4762710',
                                                                       'mac_address': '00:06:31:8a:da:04',
                                                                       'mtu': 2000,
                                                                       'name': '230962/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63438501',
                                                                       'out_octets': '44053397850',
                                                                       'out_ucast_pkts': '13083070',
                                                                       'poller_hash': 'MTExQkMzcGVlcmludGYyMzA5NjIvZzI=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:57:d0'],
                                                                       'id': '230962/g3',
                                                                       'in_nucast_pkts': '1284009',
                                                                       'in_octets': '4720588020',
                                                                       'in_ucast_pkts': '54591749',
                                                                       'mac_address': '00:06:31:8a:da:05',
                                                                       'mtu': 2000,
                                                                       'name': '230962/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1059092596',
                                                                       'out_octets': '1625410881482',
                                                                       'out_ucast_pkts': '173536495',
                                                                       'poller_hash': 'MTExQkMzcGVlcmludGYyMzA5NjIvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:74:46'],
                                                                       'id': '230962/g4',
                                                                       'in_nucast_pkts': '1252821',
                                                                       'in_octets': '104364711',
                                                                       'in_ucast_pkts': '208081',
                                                                       'mac_address': '00:06:31:8a:da:06',
                                                                       'mtu': 2000,
                                                                       'name': '230962/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4544601402',
                                                                       'out_octets': '6150082150084',
                                                                       'out_ucast_pkts': '85578',
                                                                       'poller_hash': 'MTExQkMzcGVlcmludGYyMzA5NjIvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230962',
                                                                       'in_nucast_pkts': '6706929295',
                                                                       'in_octets': '10859060650424',
                                                                       'in_ucast_pkts': '1385275903',
                                                                       'links': [       {       'in_nucast_pkts': '6706929295',
                                                                                                'in_octets': '10859060650424',
                                                                                                'in_ucast_pkts': '1385275903',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5144192',
                                                                                                'out_octets': '33948656829',
                                                                                                'out_ucast_pkts': '436889547',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5144192',
                                                                       'out_octets': '33948656829',
                                                                       'out_ucast_pkts': '436889547',
                                                                       'poller_hash': 'MTExQkMzcGVlcmludGZwb24gMjMwOTYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230962',
                                                                                               'links': [       {       'connected_time': '4164949',
                                                                                                                        'fiber_length': 2004,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTExQkMzcGVlcnBvbnBvbiAyMzA5NjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST',
                                         'poller_hash': 'MTExQkMz',
                                         'serial_number': '111BC3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:8a:da:08',
                                         'uptime': '4164949'},
                                 {       'counter_source': '230980/g1+230980/g3+230980/g4+230980/g5',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:80:2c'],
                                                                       'id': '230980/g1',
                                                                       'in_nucast_pkts': '1131563',
                                                                       'in_octets': '2078994269',
                                                                       'in_ucast_pkts': '18580554',
                                                                       'mac_address': '00:06:31:8a:fa:a3',
                                                                       'mtu': 2000,
                                                                       'name': '230980/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '293479593',
                                                                       'out_octets': '426275196682',
                                                                       'out_ucast_pkts': '57236205',
                                                                       'poller_hash': 'MTM0NDhGcGVlcmludGYyMzA5ODAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:7c:5e'],
                                                                       'id': '230980/g3',
                                                                       'in_nucast_pkts': '1332161',
                                                                       'in_octets': '1626560500',
                                                                       'in_ucast_pkts': '8065194',
                                                                       'mac_address': '00:06:31:8a:fa:a5',
                                                                       'mtu': 2000,
                                                                       'name': '230980/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55136076',
                                                                       'out_octets': '51545101923',
                                                                       'out_ucast_pkts': '22641012',
                                                                       'poller_hash': 'MTM0NDhGcGVlcmludGYyMzA5ODAvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:3f:3e'],
                                                                       'id': '230980/g4',
                                                                       'in_nucast_pkts': '1081568',
                                                                       'in_octets': '3525525984',
                                                                       'in_ucast_pkts': '37823935',
                                                                       'mac_address': '00:06:31:8a:fa:a6',
                                                                       'mtu': 2000,
                                                                       'name': '230980/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '157836792',
                                                                       'out_octets': '329182785441',
                                                                       'out_ucast_pkts': '115590303',
                                                                       'poller_hash': 'MTM0NDhGcGVlcmludGYyMzA5ODAvZzQ=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:52:08'],
                                                                       'id': '230980/g5',
                                                                       'in_nucast_pkts': '1324349',
                                                                       'in_octets': '23408661864',
                                                                       'in_ucast_pkts': '286002263',
                                                                       'mac_address': '00:06:31:96:3b:28',
                                                                       'mtu': 2000,
                                                                       'name': '230980/g5',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55098266',
                                                                       'out_octets': '1485103483499',
                                                                       'out_ucast_pkts': '1066765752',
                                                                       'poller_hash': 'MTM0NDhGcGVlcmludGYyMzA5ODAvZzU=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230980',
                                                                       'in_nucast_pkts': '561550727',
                                                                       'in_octets': '2292106567545',
                                                                       'in_ucast_pkts': '1262233272',
                                                                       'links': [       {       'in_nucast_pkts': '561550727',
                                                                                                'in_octets': '2292106567545',
                                                                                                'in_ucast_pkts': '1262233272',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4869641',
                                                                                                'out_octets': '30639742617',
                                                                                                'out_ucast_pkts': '350471946',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4869641',
                                                                       'out_octets': '30639742617',
                                                                       'out_ucast_pkts': '350471946',
                                                                       'poller_hash': 'MTM0NDhGcGVlcmludGZwb24gMjMwOTgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230980',
                                                                                               'links': [       {       'connected_time': '4164955',
                                                                                                                        'fiber_length': 2003,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTM0NDhGcGVlcnBvbnBvbiAyMzA5ODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '50 WILLIAM ST (POOLE MADALENA)',
                                         'poller_hash': 'MTM0NDhG',
                                         'serial_number': '13448F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:8a:fa:a8',
                                         'uptime': '4164955'},
                                 {       'counter_source': '231018/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e8:76'],
                                                                       'id': '231018/g1',
                                                                       'in_nucast_pkts': '903366',
                                                                       'in_octets': '847867046049',
                                                                       'in_ucast_pkts': '4626636566',
                                                                       'mac_address': '04:bc:9f:5b:d0:84',
                                                                       'mtu': 2000,
                                                                       'name': '231018/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '125470665',
                                                                       'out_octets': '16677848189884',
                                                                       'out_ucast_pkts': '11814447761',
                                                                       'poller_hash': 'MTI3NDA5MHBlZXJpbnRmMjMxMDE4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231018',
                                                                       'in_nucast_pkts': '125470665',
                                                                       'in_octets': '16677848189884',
                                                                       'in_ucast_pkts': '11814447761',
                                                                       'links': [       {       'in_nucast_pkts': '125470665',
                                                                                                'in_octets': '16677848189884',
                                                                                                'in_ucast_pkts': '11814447761',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '903366',
                                                                                                'out_octets': '847867046049',
                                                                                                'out_ucast_pkts': '4626636566',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '903366',
                                                                       'out_octets': '847867046049',
                                                                       'out_ucast_pkts': '4626636566',
                                                                       'poller_hash': 'MTI3NDA5MHBlZXJpbnRmcG9uIDIzMTAxOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231018',
                                                                                               'links': [       {       'connected_time': '8458080',
                                                                                                                        'fiber_length': 3328,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA5MHBlZXJwb25wb24gMjMxMDE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '62 PUGH ST (LAUFER PETER)',
                                         'poller_hash': 'MTI3NDA5MA==',
                                         'serial_number': '1274090',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d0:81',
                                         'uptime': '8458080'},
                                 {       'counter_source': '231358/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:c3'],
                                                                       'id': '231358/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '12383247759',
                                                                       'in_ucast_pkts': '58482166',
                                                                       'mac_address': '04:bc:9f:5c:24:6c',
                                                                       'mtu': 2000,
                                                                       'name': '231358/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434872',
                                                                       'out_octets': '243138196607',
                                                                       'out_ucast_pkts': '176051768',
                                                                       'poller_hash': 'MTI3NEIwRHBlZXJpbnRmMjMxMzU4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231358',
                                                                       'in_nucast_pkts': '1434872',
                                                                       'in_octets': '243138196607',
                                                                       'in_ucast_pkts': '176051768',
                                                                       'links': [       {       'in_nucast_pkts': '1434872',
                                                                                                'in_octets': '243138196607',
                                                                                                'in_ucast_pkts': '176051768',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '12383247759',
                                                                                                'out_ucast_pkts': '58482166',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '12383247759',
                                                                       'out_ucast_pkts': '58482166',
                                                                       'poller_hash': 'MTI3NEIwRHBlZXJpbnRmcG9uIDIzMTM1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231358',
                                                                                               'links': [       {       'connected_time': '688557',
                                                                                                                        'fiber_length': 1804,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIwRHBlZXJwb25wb24gMjMxMzU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (JONES LOGAN)',
                                         'poller_hash': 'MTI3NEIwRA==',
                                         'serial_number': '1274B0D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:69',
                                         'uptime': '688557'},
                                 {       'counter_source': '231368/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:1d'],
                                                                       'id': '231368/g1',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '2668491302',
                                                                       'in_ucast_pkts': '15320015',
                                                                       'mac_address': '04:bc:9f:5c:20:8c',
                                                                       'mtu': 2000,
                                                                       'name': '231368/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434874',
                                                                       'out_octets': '113568624979',
                                                                       'out_ucast_pkts': '80917609',
                                                                       'poller_hash': 'MTI3NEE5MXBlZXJpbnRmMjMxMzY4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231368',
                                                                       'in_nucast_pkts': '1434874',
                                                                       'in_octets': '113568624979',
                                                                       'in_ucast_pkts': '80917609',
                                                                       'links': [       {       'in_nucast_pkts': '1434874',
                                                                                                'in_octets': '113568624979',
                                                                                                'in_ucast_pkts': '80917609',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20',
                                                                                                'out_octets': '2668491302',
                                                                                                'out_ucast_pkts': '15320015',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20',
                                                                       'out_octets': '2668491302',
                                                                       'out_ucast_pkts': '15320015',
                                                                       'poller_hash': 'MTI3NEE5MXBlZXJpbnRmcG9uIDIzMTM2OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231368',
                                                                                               'links': [       {       'connected_time': '688555',
                                                                                                                        'fiber_length': 1817,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE5MXBlZXJwb25wb24gMjMxMzY4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (PRIALA GHEORGHE)',
                                         'poller_hash': 'MTI3NEE5MQ==',
                                         'serial_number': '1274A91',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:89',
                                         'uptime': '688555'},
                                 {       'counter_source': '231410/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '231410/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '39921228338',
                                                                       'in_ucast_pkts': '103213548',
                                                                       'mac_address': '04:bc:9f:5c:26:51',
                                                                       'name': '231410/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '280641149514',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEI0QXBlZXJpbnRmMjMxNDEwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '231410',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '280641149514',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '280641149514',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '39921228338',
                                                                                                'out_ucast_pkts': '103213548',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '39921228338',
                                                                       'out_ucast_pkts': '103213548',
                                                                       'poller_hash': 'MTI3NEI0QXBlZXJpbnRmcG9uIDIzMTQxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231410',
                                                                                               'links': [       {       'connected_time': '8217978',
                                                                                                                        'fiber_length': 9716,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEI0QXBlZXJwb25wb24gMjMxNDEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4899 PERTH LINE 72 (KUEPFER ROY)',
                                         'poller_hash': 'MTI3NEI0QQ==',
                                         'serial_number': '1274B4A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:26:51',
                                         'uptime': '8217978'},
                                 {       'counter_source': '231500/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:60'],
                                                                       'id': '231500/g1',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '12735422226',
                                                                       'in_ucast_pkts': '37880261',
                                                                       'mac_address': '04:bc:9f:5c:1e:dc',
                                                                       'mtu': 2000,
                                                                       'name': '231500/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1434887',
                                                                       'out_octets': '320170306393',
                                                                       'out_ucast_pkts': '227168018',
                                                                       'poller_hash': 'MTI3NEE1QnBlZXJpbnRmMjMxNTAwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231500',
                                                                       'in_nucast_pkts': '1434887',
                                                                       'in_octets': '320170306393',
                                                                       'in_ucast_pkts': '227168018',
                                                                       'links': [       {       'in_nucast_pkts': '1434887',
                                                                                                'in_octets': '320170306393',
                                                                                                'in_ucast_pkts': '227168018',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20',
                                                                                                'out_octets': '12735422226',
                                                                                                'out_ucast_pkts': '37880261',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20',
                                                                       'out_octets': '12735422226',
                                                                       'out_ucast_pkts': '37880261',
                                                                       'poller_hash': 'MTI3NEE1QnBlZXJpbnRmcG9uIDIzMTUwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231500',
                                                                                               'links': [       {       'connected_time': '688560',
                                                                                                                        'fiber_length': 1789,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1QnBlZXJwb25wb24gMjMxNTAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (HOSSFELD KRYSTAL)',
                                         'poller_hash': 'MTI3NEE1Qg==',
                                         'serial_number': '1274A5B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:d9',
                                         'uptime': '688560'},
                                 {       'counter_source': '231510/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '231510/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '170248150821',
                                                                       'in_ucast_pkts': '479085542',
                                                                       'mac_address': '04:bc:9f:5c:1f:c1',
                                                                       'name': '231510/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '1336350418314',
                                                                       'out_ucast_pkts': '1114107720',
                                                                       'poller_hash': 'MTI3NEE3OHBlZXJpbnRmMjMxNTEwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '231510',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '1336350418314',
                                                                       'in_ucast_pkts': '1114107720',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '1336350418314',
                                                                                                'in_ucast_pkts': '1114107720',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '170248150821',
                                                                                                'out_ucast_pkts': '479085542',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '170248150821',
                                                                       'out_ucast_pkts': '479085542',
                                                                       'poller_hash': 'MTI3NEE3OHBlZXJpbnRmcG9uIDIzMTUxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231510',
                                                                                               'links': [       {       'connected_time': '688535',
                                                                                                                        'fiber_length': 10198,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE3OHBlZXJwb25wb24gMjMxNTEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5899 PERTH RD 119 (BRAUN AARON & '
                                                 'MARGARETHA)',
                                         'poller_hash': 'MTI3NEE3OA==',
                                         'serial_number': '1274A78',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:c1',
                                         'uptime': '688535'},
                                 {       'counter_source': '231520/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '231520/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '77986738718',
                                                                       'in_ucast_pkts': '662216586',
                                                                       'mac_address': '04:bc:9f:5c:23:f9',
                                                                       'name': '231520/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '2180813670493',
                                                                       'out_ucast_pkts': '1598005181',
                                                                       'poller_hash': 'MTI3NEFGRnBlZXJpbnRmMjMxNTIwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '231520',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '2180813670493',
                                                                       'in_ucast_pkts': '1598005181',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '2180813670493',
                                                                                                'in_ucast_pkts': '1598005181',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '77986738718',
                                                                                                'out_ucast_pkts': '662216586',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '77986738718',
                                                                       'out_ucast_pkts': '662216586',
                                                                       'poller_hash': 'MTI3NEFGRnBlZXJpbnRmcG9uIDIzMTUyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231520',
                                                                                               'links': [       {       'connected_time': '7595824',
                                                                                                                        'fiber_length': 8900,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFGRnBlZXJwb25wb24gMjMxNTIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4976 PERTH LINE 72 (CARTER KEN & '
                                                 'ANITA)',
                                         'poller_hash': 'MTI3NEFGRg==',
                                         'serial_number': '1274AFF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:23:f9',
                                         'uptime': '7595824'},
                                 {       'counter_source': '231530/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '231530/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '862816245',
                                                                       'in_ucast_pkts': '2169219',
                                                                       'mac_address': '04:bc:9f:5c:1c:e9',
                                                                       'name': '231530/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '16085624093',
                                                                       'out_ucast_pkts': '60646113',
                                                                       'poller_hash': 'MTI3NEExRHBlZXJpbnRmMjMxNTMwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '231530',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '16085624093',
                                                                       'in_ucast_pkts': '60646113',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '16085624093',
                                                                                                'in_ucast_pkts': '60646113',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '862816245',
                                                                                                'out_ucast_pkts': '2169219',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '862816245',
                                                                       'out_ucast_pkts': '2169219',
                                                                       'poller_hash': 'MTI3NEExRHBlZXJpbnRmcG9uIDIzMTUzMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231530',
                                                                                               'links': [       {       'connected_time': '7596460',
                                                                                                                        'fiber_length': 8960,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEExRHBlZXJwb25wb24gMjMxNTMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4976 PERTH LINE 72 (CARTER '
                                                 'CATHERINE)',
                                         'poller_hash': 'MTI3NEExRA==',
                                         'serial_number': '1274A1D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:e9',
                                         'uptime': '7596460'},
                                 {       'counter_source': '231560/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '231560/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '269837026390',
                                                                       'in_ucast_pkts': '244639919',
                                                                       'mac_address': '04:bc:9f:5c:27:79',
                                                                       'name': '231560/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '43',
                                                                       'out_octets': '211106747917',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEI2RnBlZXJpbnRmMjMxNTYwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '231560',
                                                                       'in_nucast_pkts': '43',
                                                                       'in_octets': '211106747917',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '43',
                                                                                                'in_octets': '211106747917',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '269837026390',
                                                                                                'out_ucast_pkts': '244639919',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '269837026390',
                                                                       'out_ucast_pkts': '244639919',
                                                                       'poller_hash': 'MTI3NEI2RnBlZXJpbnRmcG9uIDIzMTU2MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231560',
                                                                                               'links': [       {       'connected_time': '5445301',
                                                                                                                        'fiber_length': 8574,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEI2RnBlZXJwb25wb24gMjMxNTYw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4276 PERTH LINE 72 (BAUMAN AMSEY S)',
                                         'poller_hash': 'MTI3NEI2Rg==',
                                         'serial_number': '1274B6F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:27:79',
                                         'uptime': '5445301'},
                                 {       'counter_source': '231570/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd4:ca:6d:32:23:73'],
                                                                       'id': '231570/g1',
                                                                       'in_nucast_pkts': '2510028',
                                                                       'in_octets': '1418585235575',
                                                                       'in_ucast_pkts': '3938289468',
                                                                       'mac_address': '04:bc:9f:5c:21:c4',
                                                                       'mtu': 2000,
                                                                       'name': '231570/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '121758713',
                                                                       'out_octets': '9603518292591',
                                                                       'out_ucast_pkts': '7713232754',
                                                                       'poller_hash': 'MTI3NEFCOHBlZXJpbnRmMjMxNTcwL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231570',
                                                                       'in_nucast_pkts': '121758713',
                                                                       'in_octets': '9603518292591',
                                                                       'in_ucast_pkts': '7713232754',
                                                                       'links': [       {       'in_nucast_pkts': '121758713',
                                                                                                'in_octets': '9603518292591',
                                                                                                'in_ucast_pkts': '7713232754',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2510028',
                                                                                                'out_octets': '1418585235575',
                                                                                                'out_ucast_pkts': '3938289468',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2510028',
                                                                       'out_octets': '1418585235575',
                                                                       'out_ucast_pkts': '3938289468',
                                                                       'poller_hash': 'MTI3NEFCOHBlZXJpbnRmcG9uIDIzMTU3MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231570',
                                                                                               'links': [       {       'connected_time': '7252767',
                                                                                                                        'fiber_length': 7086,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFCOHBlZXJwb25wb24gMjMxNTcw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4392 PERTH LINE 72 (NEWTON FURNITURE '
                                                 'INC)',
                                         'poller_hash': 'MTI3NEFCOA==',
                                         'serial_number': '1274AB8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:21:c1',
                                         'uptime': '7252767'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '231580/g1',
                                                                       'name': '231580/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEIzMHBlZXJpbnRmMjMxNTgwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '231580',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEIzMHBlZXJpbnRmcG9uIDIzMTU4MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231580',
                                                                                               'links': [       {       'connected_time': '29747922',
                                                                                                                        'fiber_length': 9418,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIzMHBlZXJwb25wb24gMjMxNTgw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4939 PERTH LINE 72 (KUEPFER ALVIN K '
                                                 '& ANNA MAE)',
                                         'poller_hash': 'MTI3NEIzMA==',
                                         'serial_number': '1274B30',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:25:81',
                                         'uptime': '29747922'},
                                 {       'counter_source': '231620/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:df:c8'],
                                                                       'id': '231620/g1',
                                                                       'in_nucast_pkts': '12159',
                                                                       'in_octets': '278595104',
                                                                       'in_ucast_pkts': '1046233',
                                                                       'mac_address': '04:bc:9f:5c:26:d4',
                                                                       'mtu': 2000,
                                                                       'name': '231620/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464490',
                                                                       'out_octets': '10090293135',
                                                                       'out_ucast_pkts': '7394441',
                                                                       'poller_hash': 'MTI3NEI1QXBlZXJpbnRmMjMxNjIwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231620',
                                                                       'in_nucast_pkts': '1464490',
                                                                       'in_octets': '10090293135',
                                                                       'in_ucast_pkts': '7394441',
                                                                       'links': [       {       'in_nucast_pkts': '1464490',
                                                                                                'in_octets': '10090293135',
                                                                                                'in_ucast_pkts': '7394441',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12159',
                                                                                                'out_octets': '278595104',
                                                                                                'out_ucast_pkts': '1046233',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12159',
                                                                       'out_octets': '278595104',
                                                                       'out_ucast_pkts': '1046233',
                                                                       'poller_hash': 'MTI3NEI1QXBlZXJpbnRmcG9uIDIzMTYyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231620',
                                                                                               'links': [       {       'connected_time': '688452',
                                                                                                                        'fiber_length': 1363,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEI1QXBlZXJwb25wb24gMjMxNjIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 EDWARD ST (CHALMERS JENNIFER)',
                                         'poller_hash': 'MTI3NEI1QQ==',
                                         'serial_number': '1274B5A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:26:d1',
                                         'uptime': '688452'},
                                 {       'counter_source': '232114/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e7:8f'],
                                                                       'id': '232114/g1',
                                                                       'in_nucast_pkts': '12922',
                                                                       'in_octets': '3508305069',
                                                                       'in_ucast_pkts': '9929327',
                                                                       'mac_address': '04:bc:9f:5c:25:dc',
                                                                       'mtu': 2000,
                                                                       'name': '232114/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464498',
                                                                       'out_octets': '19793819238',
                                                                       'out_ucast_pkts': '18225532',
                                                                       'poller_hash': 'MTI3NEIzQnBlZXJpbnRmMjMyMTE0L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '232114',
                                                                       'in_nucast_pkts': '1464498',
                                                                       'in_octets': '19793819238',
                                                                       'in_ucast_pkts': '18225532',
                                                                       'links': [       {       'in_nucast_pkts': '1464498',
                                                                                                'in_octets': '19793819238',
                                                                                                'in_ucast_pkts': '18225532',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12922',
                                                                                                'out_octets': '3508305069',
                                                                                                'out_ucast_pkts': '9929327',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12922',
                                                                       'out_octets': '3508305069',
                                                                       'out_ucast_pkts': '9929327',
                                                                       'poller_hash': 'MTI3NEIzQnBlZXJpbnRmcG9uIDIzMjExNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232114',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 11269,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIzQnBlZXJwb25wb24gMjMyMTE0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4092 QUEEN ST (BAUMAN DARRELL)',
                                         'poller_hash': 'MTI3NEIzQg==',
                                         'serial_number': '1274B3B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:25:d9',
                                         'uptime': '688369'},
                                 {       'counter_source': '232184/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '232184/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '38434800184',
                                                                       'in_ucast_pkts': '67015530',
                                                                       'mac_address': '04:bc:9f:5c:22:81',
                                                                       'name': '232184/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '198963142794',
                                                                       'out_ucast_pkts': '226430107',
                                                                       'poller_hash': 'MTI3NEFEMHBlZXJpbnRmMjMyMTg0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '232184',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '198963142794',
                                                                       'in_ucast_pkts': '226430107',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '198963142794',
                                                                                                'in_ucast_pkts': '226430107',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '38434800184',
                                                                                                'out_ucast_pkts': '67015530',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '38434800184',
                                                                       'out_ucast_pkts': '67015530',
                                                                       'poller_hash': 'MTI3NEFEMHBlZXJpbnRmcG9uIDIzMjE4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232184',
                                                                                               'links': [       {       'connected_time': '23065539',
                                                                                                                        'fiber_length': 10436,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFEMHBlZXJwb25wb24gMjMyMTg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4119 PERTH LINE 72 (LIVESTOCK '
                                                 'NUTRITIONAL SERVICES)',
                                         'poller_hash': 'MTI3NEFEMA==',
                                         'serial_number': '1274AD0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:81',
                                         'uptime': '23065539'},
                                 {       'counter_source': '232279/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '232279/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '20777396132',
                                                                       'in_ucast_pkts': '181029269',
                                                                       'mac_address': '04:bc:9f:5c:22:b9',
                                                                       'name': '232279/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '906809261719',
                                                                       'out_ucast_pkts': '695411732',
                                                                       'poller_hash': 'MTI3NEFEN3BlZXJpbnRmMjMyMjc5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '232279',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '906809261719',
                                                                       'in_ucast_pkts': '695411732',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '906809261719',
                                                                                                'in_ucast_pkts': '695411732',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '20777396132',
                                                                                                'out_ucast_pkts': '181029269',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '20777396132',
                                                                       'out_ucast_pkts': '181029269',
                                                                       'poller_hash': 'MTI3NEFEN3BlZXJpbnRmcG9uIDIzMjI3OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232279',
                                                                                               'links': [       {       'connected_time': '6659277',
                                                                                                                        'fiber_length': 7722,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFEN3BlZXJwb25wb24gMjMyMjc5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4341 PERTH LINE 72 (JANTZI LEVI K)',
                                         'poller_hash': 'MTI3NEFENw==',
                                         'serial_number': '1274AD7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:b9',
                                         'uptime': '6659277'},
                                 {       'counter_source': '232309/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e9:1b'],
                                                                       'id': '232309/g1',
                                                                       'in_nucast_pkts': '14588',
                                                                       'in_octets': '21117233674',
                                                                       'in_ucast_pkts': '68319815',
                                                                       'mac_address': '04:bc:9f:5c:18:bc',
                                                                       'mtu': 2000,
                                                                       'name': '232309/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464731',
                                                                       'out_octets': '160563655881',
                                                                       'out_ucast_pkts': '133673872',
                                                                       'poller_hash': 'MTI3NDk5N3BlZXJpbnRmMjMyMzA5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '232309',
                                                                       'in_nucast_pkts': '1464731',
                                                                       'in_octets': '160563655881',
                                                                       'in_ucast_pkts': '133673872',
                                                                       'links': [       {       'in_nucast_pkts': '1464731',
                                                                                                'in_octets': '160563655881',
                                                                                                'in_ucast_pkts': '133673872',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14588',
                                                                                                'out_octets': '21117233674',
                                                                                                'out_ucast_pkts': '68319815',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14588',
                                                                       'out_octets': '21117233674',
                                                                       'out_ucast_pkts': '68319815',
                                                                       'poller_hash': 'MTI3NDk5N3BlZXJpbnRmcG9uIDIzMjMwOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232309',
                                                                                               'links': [       {       'connected_time': '688364',
                                                                                                                        'fiber_length': 3081,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk5N3BlZXJwb25wb24gMjMyMzA5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '108 PUGH ST (SANGWA MARIE)',
                                         'poller_hash': 'MTI3NDk5Nw==',
                                         'serial_number': '1274997',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:18:b9',
                                         'uptime': '688364'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '232319/g1',
                                                                       'name': '232319/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEE5NHBlZXJpbnRmMjMyMzE5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '232319',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEE5NHBlZXJpbnRmcG9uIDIzMjMxOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232319',
                                                                                               'links': [       {       'connected_time': '6561877',
                                                                                                                        'fiber_length': 7716,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE5NHBlZXJwb25wb24gMjMyMzE5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4341 PERTH LINE 72 (JANTZI JIM)',
                                         'poller_hash': 'MTI3NEE5NA==',
                                         'serial_number': '1274A94',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:a1',
                                         'uptime': '6561877'},
                                 {       'counter_source': '232558/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e6:03'],
                                                                       'id': '232558/g1',
                                                                       'in_nucast_pkts': '14335',
                                                                       'in_octets': '11664189642',
                                                                       'in_ucast_pkts': '84568086',
                                                                       'mac_address': '04:bc:9f:5c:1b:94',
                                                                       'mtu': 2000,
                                                                       'name': '232558/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1650852',
                                                                       'out_octets': '447240729873',
                                                                       'out_ucast_pkts': '307247602',
                                                                       'poller_hash': 'MTI3NDlGMnBlZXJpbnRmMjMyNTU4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '232558',
                                                                       'in_nucast_pkts': '1650852',
                                                                       'in_octets': '447240729873',
                                                                       'in_ucast_pkts': '307247602',
                                                                       'links': [       {       'in_nucast_pkts': '1650852',
                                                                                                'in_octets': '447240729873',
                                                                                                'in_ucast_pkts': '307247602',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14335',
                                                                                                'out_octets': '11664189642',
                                                                                                'out_ucast_pkts': '84568086',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14335',
                                                                       'out_octets': '11664189642',
                                                                       'out_ucast_pkts': '84568086',
                                                                       'poller_hash': 'MTI3NDlGMnBlZXJpbnRmcG9uIDIzMjU1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232558',
                                                                                               'links': [       {       'connected_time': '779210',
                                                                                                                        'fiber_length': 1714,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlGMnBlZXJwb25wb24gMjMyNTU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'ont 232558',
                                         'poller_hash': 'MTI3NDlGMg==',
                                         'serial_number': '12749F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:91',
                                         'uptime': '779210'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '232810/g1',
                                                                       'name': '232810/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEEwNXBlZXJpbnRmMjMyODEwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '232810',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEEwNXBlZXJpbnRmcG9uIDIzMjgxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232810',
                                                                                               'links': [       {       'connected_time': '63977869',
                                                                                                                        'fiber_length': 7024,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwNXBlZXJwb25wb24gMjMyODEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4404 PERTH LINE 72 (BAUMAN JESSE)',
                                         'poller_hash': 'MTI3NEEwNQ==',
                                         'serial_number': '1274A05',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:29',
                                         'uptime': '63977869'},
                                 {       'counter_source': '233109/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233109/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '272663983066',
                                                                       'in_ucast_pkts': '74081456454',
                                                                       'mac_address': '04:bc:9f:5c:1d:19',
                                                                       'name': '233109/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '3803462566543',
                                                                       'out_ucast_pkts': '2841283891',
                                                                       'poller_hash': 'MTI3NEEyM3BlZXJpbnRmMjMzMTA5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233109',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '3803462566543',
                                                                       'in_ucast_pkts': '2841283891',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '3803462566543',
                                                                                                'in_ucast_pkts': '2841283891',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '272663983066',
                                                                                                'out_ucast_pkts': '74081456454',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '272663983066',
                                                                       'out_ucast_pkts': '74081456454',
                                                                       'poller_hash': 'MTI3NEEyM3BlZXJpbnRmcG9uIDIzMzEwOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233109',
                                                                                               'links': [       {       'connected_time': '5523610',
                                                                                                                        'fiber_length': 10647,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEyM3BlZXJwb25wb24gMjMzMTA5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6574 ROAD 116 (MCCOY CHRISTINE)',
                                         'poller_hash': 'MTI3NEEyMw==',
                                         'serial_number': '1274A23',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:19',
                                         'uptime': '5523610'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '233258/g1',
                                                                       'name': '233258/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDlDQ3BlZXJpbnRmMjMzMjU4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233258',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDlDQ3BlZXJpbnRmcG9uIDIzMzI1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233258',
                                                                                               'links': [       {       'connected_time': '5369219',
                                                                                                                        'fiber_length': 6898,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlDQ3BlZXJwb25wb24gMjMzMjU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4413 PERTH LINE 72 (BAUMAN TOBIAS)',
                                         'poller_hash': 'MTI3NDlDQw==',
                                         'serial_number': '12749CC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1a:61',
                                         'uptime': '5369219'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '233289/g1',
                                                                       'name': '233289/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEFCNXBlZXJpbnRmMjMzMjg5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233289',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEFCNXBlZXJpbnRmcG9uIDIzMzI4OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233289',
                                                                                               'links': [       {       'connected_time': '5354267',
                                                                                                                        'fiber_length': 9878,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFCNXBlZXJwb25wb24gMjMzMjg5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4892 PERTH LINE 55 (WAGLER MRS '
                                                 'KATIE)',
                                         'poller_hash': 'MTI3NEFCNQ==',
                                         'serial_number': '1274AB5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:21:a9',
                                         'uptime': '5354267'},
                                 {       'counter_source': '233453/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233453/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '160186156779',
                                                                       'in_ucast_pkts': '511860676',
                                                                       'mac_address': '04:bc:9f:5c:14:99',
                                                                       'name': '233453/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '1975854924199',
                                                                       'out_ucast_pkts': '1545797483',
                                                                       'poller_hash': 'MTI3NDkxM3BlZXJpbnRmMjMzNDUzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233453',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '1975854924199',
                                                                       'in_ucast_pkts': '1545797483',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '1975854924199',
                                                                                                'in_ucast_pkts': '1545797483',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '160186156779',
                                                                                                'out_ucast_pkts': '511860676',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '160186156779',
                                                                       'out_ucast_pkts': '511860676',
                                                                       'poller_hash': 'MTI3NDkxM3BlZXJpbnRmcG9uIDIzMzQ1Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233453',
                                                                                               'links': [       {       'connected_time': '5265866',
                                                                                                                        'fiber_length': 10341,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDkxM3BlZXJwb25wb24gMjMzNDUz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6662 ROAD 116 (SCHMIDT TIM)',
                                         'poller_hash': 'MTI3NDkxMw==',
                                         'serial_number': '1274913',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:14:99',
                                         'uptime': '5265866'},
                                 {       'counter_source': '233493/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233493/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '320854869669',
                                                                       'in_ucast_pkts': '48291231296',
                                                                       'mac_address': '04:bc:9f:5c:24:11',
                                                                       'name': '233493/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '4119811317189',
                                                                       'out_ucast_pkts': '3371372945',
                                                                       'poller_hash': 'MTI3NEIwMnBlZXJpbnRmMjMzNDkzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233493',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '4119811317189',
                                                                       'in_ucast_pkts': '3371372945',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '4119811317189',
                                                                                                'in_ucast_pkts': '3371372945',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '320854869669',
                                                                                                'out_ucast_pkts': '48291231296',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '320854869669',
                                                                       'out_ucast_pkts': '48291231296',
                                                                       'poller_hash': 'MTI3NEIwMnBlZXJpbnRmcG9uIDIzMzQ5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233493',
                                                                                               'links': [       {       'connected_time': '688551',
                                                                                                                        'fiber_length': 1816,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -26.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIwMnBlZXJwb25wb24gMjMzNDkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (KAUR  SIMRAN)',
                                         'poller_hash': 'MTI3NEIwMg==',
                                         'serial_number': '1274B02',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:11',
                                         'uptime': '688551'},
                                 {       'counter_source': '233503/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233503/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '157216073952',
                                                                       'in_ucast_pkts': '280279556',
                                                                       'mac_address': '04:bc:9f:5c:19:99',
                                                                       'name': '233503/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '49',
                                                                       'out_octets': '721997483677',
                                                                       'out_ucast_pkts': '671094586',
                                                                       'poller_hash': 'MTI3NDlCM3BlZXJpbnRmMjMzNTAzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233503',
                                                                       'in_nucast_pkts': '49',
                                                                       'in_octets': '721997483677',
                                                                       'in_ucast_pkts': '671094586',
                                                                       'links': [       {       'in_nucast_pkts': '49',
                                                                                                'in_octets': '721997483677',
                                                                                                'in_ucast_pkts': '671094586',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '157216073952',
                                                                                                'out_ucast_pkts': '280279556',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '157216073952',
                                                                       'out_ucast_pkts': '280279556',
                                                                       'poller_hash': 'MTI3NDlCM3BlZXJpbnRmcG9uIDIzMzUwMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233503',
                                                                                               'links': [       {       'connected_time': '1023927',
                                                                                                                        'fiber_length': 9868,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlCM3BlZXJwb25wb24gMjMzNTAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6755 ROAD 116 (JANTZI GERALD W)',
                                         'poller_hash': 'MTI3NDlCMw==',
                                         'serial_number': '12749B3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:19:99',
                                         'uptime': '1023927'},
                                 {       'counter_source': '233553/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233553/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '255084420198',
                                                                       'in_ucast_pkts': '618996602',
                                                                       'mac_address': '04:bc:9f:5c:1f:41',
                                                                       'name': '233553/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '2367819241397',
                                                                       'out_ucast_pkts': '2005838392',
                                                                       'poller_hash': 'MTI3NEE2OHBlZXJpbnRmMjMzNTUzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233553',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '2367819241397',
                                                                       'in_ucast_pkts': '2005838392',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '2367819241397',
                                                                                                'in_ucast_pkts': '2005838392',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '255084420198',
                                                                                                'out_ucast_pkts': '618996602',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '255084420198',
                                                                       'out_ucast_pkts': '618996602',
                                                                       'poller_hash': 'MTI3NEE2OHBlZXJpbnRmcG9uIDIzMzU1Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233553',
                                                                                               'links': [       {       'connected_time': '4936633',
                                                                                                                        'fiber_length': 10372,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE2OHBlZXJwb25wb24gMjMzNTUz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6621 ROAD 116 (ZEHR TERRY & SUE)',
                                         'poller_hash': 'MTI3NEE2OA==',
                                         'serial_number': '1274A68',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:41',
                                         'uptime': '4936633'},
                                 {       'counter_source': '233583/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233583/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '464542793853',
                                                                       'in_ucast_pkts': '1443501792',
                                                                       'mac_address': '04:bc:9f:5c:1d:11',
                                                                       'name': '233583/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '116',
                                                                       'out_octets': '6941730523547',
                                                                       'out_ucast_pkts': '5044044823',
                                                                       'poller_hash': 'MTI3NEEyMnBlZXJpbnRmMjMzNTgzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233583',
                                                                       'in_nucast_pkts': '116',
                                                                       'in_octets': '6941730523547',
                                                                       'in_ucast_pkts': '5044044823',
                                                                       'links': [       {       'in_nucast_pkts': '116',
                                                                                                'in_octets': '6941730523547',
                                                                                                'in_ucast_pkts': '5044044823',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '464542793853',
                                                                                                'out_ucast_pkts': '1443501792',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '464542793853',
                                                                       'out_ucast_pkts': '1443501792',
                                                                       'poller_hash': 'MTI3NEEyMnBlZXJpbnRmcG9uIDIzMzU4Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233583',
                                                                                               'links': [       {       'connected_time': '688591',
                                                                                                                        'fiber_length': 10641,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEyMnBlZXJwb25wb24gMjMzNTgz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6396 PERTH ROAD 121 (DEGROOT STEVEN)',
                                         'poller_hash': 'MTI3NEEyMg==',
                                         'serial_number': '1274A22',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:11',
                                         'uptime': '688591'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '233673/g1',
                                                                       'name': '233673/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEE3NnBlZXJpbnRmMjMzNjczL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233673',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEE3NnBlZXJpbnRmcG9uIDIzMzY3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233673',
                                                                                               'links': [       {       'connected_time': '4761198',
                                                                                                                        'fiber_length': 9089,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE3NnBlZXJwb25wb24gMjMzNjcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4577 LINE 61 (VAN NES JAMES)',
                                         'poller_hash': 'MTI3NEE3Ng==',
                                         'serial_number': '1274A76',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:b1',
                                         'uptime': '4761198'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '233693/g1',
                                                                       'name': '233693/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEE3MHBlZXJpbnRmMjMzNjkzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233693',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEE3MHBlZXJpbnRmcG9uIDIzMzY5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233693',
                                                                                               'links': [       {       'connected_time': '4678524',
                                                                                                                        'fiber_length': 1006,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE3MHBlZXJwb25wb24gMjMzNjkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 TEMPERANCE ST (LEIS STANLEY)',
                                         'poller_hash': 'MTI3NEE3MA==',
                                         'serial_number': '1274A70',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:81',
                                         'uptime': '4678524'},
                                 {       'counter_source': '233703/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233703/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '38732215',
                                                                       'in_ucast_pkts': '60853',
                                                                       'mac_address': '04:bc:9f:5c:1e:d1',
                                                                       'name': '233703/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '160',
                                                                       'out_octets': '12605121035',
                                                                       'out_ucast_pkts': '54207661',
                                                                       'poller_hash': 'MTI3NEE1QXBlZXJpbnRmMjMzNzAzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233703',
                                                                       'in_nucast_pkts': '160',
                                                                       'in_octets': '12605121035',
                                                                       'in_ucast_pkts': '54207661',
                                                                       'links': [       {       'in_nucast_pkts': '160',
                                                                                                'in_octets': '12605121035',
                                                                                                'in_ucast_pkts': '54207661',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '38732215',
                                                                                                'out_ucast_pkts': '60853',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '38732215',
                                                                       'out_ucast_pkts': '60853',
                                                                       'poller_hash': 'MTI3NEE1QXBlZXJpbnRmcG9uIDIzMzcwMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233703',
                                                                                               'links': [       {       'connected_time': '4674396',
                                                                                                                        'fiber_length': 10556,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1QXBlZXJwb25wb24gMjMzNzAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6401 PERTH ROAD 121 (MILES WILLIAM)',
                                         'poller_hash': 'MTI3NEE1QQ==',
                                         'serial_number': '1274A5A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:d1',
                                         'uptime': '4674396'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '233713/g1',
                                                                       'name': '233713/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDVEOXBlZXJpbnRmMjMzNzEzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233713',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDVEOXBlZXJpbnRmcG9uIDIzMzcxMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233713',
                                                                                               'links': [       {       'connected_time': '4673060',
                                                                                                                        'fiber_length': 10444,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDVEOXBlZXJwb25wb24gMjMzNzEz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4107 LINE 61 (THIESSEN JACOB)',
                                         'poller_hash': 'MTI3NDVEOQ==',
                                         'serial_number': '12745D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:fa:c9',
                                         'uptime': '4673060'},
                                 {       'counter_source': '233736/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:d4:bc'],
                                                                       'id': '233736/g2',
                                                                       'in_nucast_pkts': '2374072',
                                                                       'in_octets': '85366564152',
                                                                       'in_ucast_pkts': '1128275564',
                                                                       'mac_address': '04:bc:9f:5c:21:3d',
                                                                       'mtu': 2000,
                                                                       'name': '233736/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '159817338',
                                                                       'out_octets': '6821200797821',
                                                                       'out_ucast_pkts': '4610641053',
                                                                       'poller_hash': 'MTI3NEFBN3BlZXJpbnRmMjMzNzM2L2cy',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '233736',
                                                                       'in_nucast_pkts': '159817338',
                                                                       'in_octets': '6821200797821',
                                                                       'in_ucast_pkts': '4610641053',
                                                                       'links': [       {       'in_nucast_pkts': '159817338',
                                                                                                'in_octets': '6821200797821',
                                                                                                'in_ucast_pkts': '4610641053',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2374072',
                                                                                                'out_octets': '85366564152',
                                                                                                'out_ucast_pkts': '1128275564',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2374072',
                                                                       'out_octets': '85366564152',
                                                                       'out_ucast_pkts': '1128275564',
                                                                       'poller_hash': 'MTI3NEFBN3BlZXJpbnRmcG9uIDIzMzczNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233736',
                                                                                               'links': [       {       'connected_time': '64803881',
                                                                                                                        'fiber_length': 1135,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFBN3BlZXJwb25wb24gMjMzNzM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '77 MAIN ST N (RASMUSSEN FLORENCE)',
                                         'poller_hash': 'MTI3NEFBNw==',
                                         'serial_number': '1274AA7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:21:39',
                                         'uptime': '64803881'},
                                 {       'counter_source': '233756/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233756/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '202154734278',
                                                                       'in_ucast_pkts': '39515599856',
                                                                       'mac_address': '04:bc:9f:5c:1e:f1',
                                                                       'name': '233756/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '2123580409622',
                                                                       'out_ucast_pkts': '1748743966',
                                                                       'poller_hash': 'MTI3NEE1RXBlZXJpbnRmMjMzNzU2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233756',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '2123580409622',
                                                                       'in_ucast_pkts': '1748743966',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '2123580409622',
                                                                                                'in_ucast_pkts': '1748743966',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '202154734278',
                                                                                                'out_ucast_pkts': '39515599856',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '202154734278',
                                                                       'out_ucast_pkts': '39515599856',
                                                                       'poller_hash': 'MTI3NEE1RXBlZXJpbnRmcG9uIDIzMzc1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233756',
                                                                                               'links': [       {       'connected_time': '25491295',
                                                                                                                        'fiber_length': 5732,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1RXBlZXJwb25wb24gMjMzNzU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6433 ROAD 124 (ANDERSON RYAN)',
                                         'poller_hash': 'MTI3NEE1RQ==',
                                         'serial_number': '1274A5E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:f1',
                                         'uptime': '25491295'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '233819/g1',
                                                                       'name': '233819/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEE4Q3BlZXJpbnRmMjMzODE5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233819',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEE4Q3BlZXJpbnRmcG9uIDIzMzgxOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233819',
                                                                                               'links': [       {       'connected_time': '4576738',
                                                                                                                        'fiber_length': 7327,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE4Q3BlZXJwb25wb24gMjMzODE5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6573 ROAD 140 (EBERSOL ELI)',
                                         'poller_hash': 'MTI3NEE4Qw==',
                                         'serial_number': '1274A8C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:61',
                                         'uptime': '4576738'},
                                 {       'counter_source': '233839/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:09:ab'],
                                                                       'id': '233839/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '30435597033',
                                                                       'in_ucast_pkts': '130995778',
                                                                       'mac_address': '04:bc:9f:5c:23:4c',
                                                                       'mtu': 2000,
                                                                       'name': '233839/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465097',
                                                                       'out_octets': '394459439029',
                                                                       'out_ucast_pkts': '321474202',
                                                                       'poller_hash': 'MTI3NEFFOXBlZXJpbnRmMjMzODM5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '233839',
                                                                       'in_nucast_pkts': '1465097',
                                                                       'in_octets': '394459439029',
                                                                       'in_ucast_pkts': '321474202',
                                                                       'links': [       {       'in_nucast_pkts': '1465097',
                                                                                                'in_octets': '394459439029',
                                                                                                'in_ucast_pkts': '321474202',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '30435597033',
                                                                                                'out_ucast_pkts': '130995778',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '30435597033',
                                                                       'out_ucast_pkts': '130995778',
                                                                       'poller_hash': 'MTI3NEFFOXBlZXJpbnRmcG9uIDIzMzgzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233839',
                                                                                               'links': [       {       'connected_time': '688355',
                                                                                                                        'fiber_length': 1139,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFFOXBlZXJwb25wb24gMjMzODM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 MAIN ST S UNIT 102 (GOETZ LORI)',
                                         'poller_hash': 'MTI3NEFFOQ==',
                                         'serial_number': '1274AE9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:23:49',
                                         'uptime': '688355'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '233849/g1',
                                                                       'name': '233849/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEIzOHBlZXJpbnRmMjMzODQ5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233849',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEIzOHBlZXJpbnRmcG9uIDIzMzg0OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233849',
                                                                                               'links': [       {       'connected_time': '22533668',
                                                                                                                        'fiber_length': 10970,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIzOHBlZXJwb25wb24gMjMzODQ5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6563 ROAD 116 (JANTZI SHARON)',
                                         'poller_hash': 'MTI3NEIzOA==',
                                         'serial_number': '1274B38',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:25:c1',
                                         'uptime': '22533668'},
                                 {       'counter_source': '233859/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233859/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '77720088277',
                                                                       'in_ucast_pkts': '13115490861',
                                                                       'mac_address': '04:bc:9f:5c:23:d9',
                                                                       'name': '233859/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '90',
                                                                       'out_octets': '1010485845738',
                                                                       'out_ucast_pkts': '838477811',
                                                                       'poller_hash': 'MTI3NEFGQnBlZXJpbnRmMjMzODU5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233859',
                                                                       'in_nucast_pkts': '90',
                                                                       'in_octets': '1010485845738',
                                                                       'in_ucast_pkts': '838477811',
                                                                       'links': [       {       'in_nucast_pkts': '90',
                                                                                                'in_octets': '1010485845738',
                                                                                                'in_ucast_pkts': '838477811',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '77720088277',
                                                                                                'out_ucast_pkts': '13115490861',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '77720088277',
                                                                       'out_ucast_pkts': '13115490861',
                                                                       'poller_hash': 'MTI3NEFGQnBlZXJpbnRmcG9uIDIzMzg1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233859',
                                                                                               'links': [       {       'connected_time': '688620',
                                                                                                                        'fiber_length': 8378,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFGQnBlZXJwb25wb24gMjMzODU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6752 PERTH ROAD 121 (ZEHR CHERYL)',
                                         'poller_hash': 'MTI3NEFGQg==',
                                         'serial_number': '1274AFB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:23:d9',
                                         'uptime': '688620'},
                                 {       'counter_source': '234080/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234080/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '10838262874',
                                                                       'in_ucast_pkts': '25809481',
                                                                       'mac_address': '04:bc:9f:5c:1e:91',
                                                                       'name': '234080/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '160',
                                                                       'out_octets': '88187343227',
                                                                       'out_ucast_pkts': '124364653',
                                                                       'poller_hash': 'MTI3NEE1MnBlZXJpbnRmMjM0MDgwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234080',
                                                                       'in_nucast_pkts': '160',
                                                                       'in_octets': '88187343227',
                                                                       'in_ucast_pkts': '124364653',
                                                                       'links': [       {       'in_nucast_pkts': '160',
                                                                                                'in_octets': '88187343227',
                                                                                                'in_ucast_pkts': '124364653',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '10838262874',
                                                                                                'out_ucast_pkts': '25809481',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '10838262874',
                                                                       'out_ucast_pkts': '25809481',
                                                                       'poller_hash': 'MTI3NEE1MnBlZXJpbnRmcG9uIDIzNDA4MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234080',
                                                                                               'links': [       {       'connected_time': '4229731',
                                                                                                                        'fiber_length': 10640,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1MnBlZXJwb25wb24gMjM0MDgw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6395 PERTH ROAD 121 (JANTZI '
                                                 'EXCAVATING & BULLDOZING)',
                                         'poller_hash': 'MTI3NEE1Mg==',
                                         'serial_number': '1274A52',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:91',
                                         'uptime': '4229731'},
                                 {       'counter_source': '234090/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234090/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '74440797396',
                                                                       'in_ucast_pkts': '262101534',
                                                                       'mac_address': '04:bc:9f:5c:1e:a9',
                                                                       'name': '234090/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '116',
                                                                       'out_octets': '726394569519',
                                                                       'out_ucast_pkts': '584819971',
                                                                       'poller_hash': 'MTI3NEE1NXBlZXJpbnRmMjM0MDkwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234090',
                                                                       'in_nucast_pkts': '116',
                                                                       'in_octets': '726394569519',
                                                                       'in_ucast_pkts': '584819971',
                                                                       'links': [       {       'in_nucast_pkts': '116',
                                                                                                'in_octets': '726394569519',
                                                                                                'in_ucast_pkts': '584819971',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '74440797396',
                                                                                                'out_ucast_pkts': '262101534',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '74440797396',
                                                                       'out_ucast_pkts': '262101534',
                                                                       'poller_hash': 'MTI3NEE1NXBlZXJpbnRmcG9uIDIzNDA5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234090',
                                                                                               'links': [       {       'connected_time': '688592',
                                                                                                                        'fiber_length': 10654,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1NXBlZXJwb25wb24gMjM0MDkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6400 PERTH ROAD 121 (KUEPFER VICTOR)',
                                         'poller_hash': 'MTI3NEE1NQ==',
                                         'serial_number': '1274A55',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:a9',
                                         'uptime': '688592'},
                                 {       'counter_source': '234110/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:34:ac:02'],
                                                                       'id': '234110/g4',
                                                                       'in_nucast_pkts': '7651166',
                                                                       'in_octets': '118973115194',
                                                                       'in_ucast_pkts': '1602211283',
                                                                       'mac_address': '04:bc:9f:5c:23:e7',
                                                                       'mtu': 2000,
                                                                       'name': '234110/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15308074954',
                                                                       'out_octets': '29146494482497',
                                                                       'out_ucast_pkts': '6369319028',
                                                                       'poller_hash': 'MTI3NEFGQ3BlZXJpbnRmMjM0MTEwL2c0',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '234110',
                                                                       'in_nucast_pkts': '15308074954',
                                                                       'in_octets': '29146494482497',
                                                                       'in_ucast_pkts': '6369319028',
                                                                       'links': [       {       'in_nucast_pkts': '15308074954',
                                                                                                'in_octets': '29146494482497',
                                                                                                'in_ucast_pkts': '6369319028',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7651166',
                                                                                                'out_octets': '118973115194',
                                                                                                'out_ucast_pkts': '1602211283',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7651166',
                                                                       'out_octets': '118973115194',
                                                                       'out_ucast_pkts': '1602211283',
                                                                       'poller_hash': 'MTI3NEFGQ3BlZXJpbnRmcG9uIDIzNDExMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234110',
                                                                                               'links': [       {       'connected_time': '4156016',
                                                                                                                        'fiber_length': 1745,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFGQ3BlZXJwb25wb24gMjM0MTEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 MURRAY CT MILVERTON (WILLIAM HORST '
                                                 '10400332)',
                                         'poller_hash': 'MTI3NEFGQw==',
                                         'serial_number': '1274AFC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:23:e1',
                                         'uptime': '4156016'},
                                 {       'counter_source': '234347/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234347/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '48905771558',
                                                                       'in_ucast_pkts': '142565640',
                                                                       'mac_address': '04:bc:9f:5c:1c:f9',
                                                                       'name': '234347/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '704846907500',
                                                                       'out_ucast_pkts': '600982353',
                                                                       'poller_hash': 'MTI3NEExRnBlZXJpbnRmMjM0MzQ3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234347',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '704846907500',
                                                                       'in_ucast_pkts': '600982353',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '704846907500',
                                                                                                'in_ucast_pkts': '600982353',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '48905771558',
                                                                                                'out_ucast_pkts': '142565640',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '48905771558',
                                                                       'out_ucast_pkts': '142565640',
                                                                       'poller_hash': 'MTI3NEExRnBlZXJpbnRmcG9uIDIzNDM0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234347',
                                                                                               'links': [       {       'connected_time': '4055332',
                                                                                                                        'fiber_length': 11790,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEExRnBlZXJwb25wb24gMjM0MzQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6906 ELGIN ST (HERRFORT MARLENE)',
                                         'poller_hash': 'MTI3NEExRg==',
                                         'serial_number': '1274A1F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:f9',
                                         'uptime': '4055332'},
                                 {       'counter_source': '234367/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234367/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '38496656814',
                                                                       'in_ucast_pkts': '128804503',
                                                                       'mac_address': '04:bc:9f:5c:21:69',
                                                                       'name': '234367/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '592686974138',
                                                                       'out_ucast_pkts': '482546907',
                                                                       'poller_hash': 'MTI3NEFBRHBlZXJpbnRmMjM0MzY3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234367',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '592686974138',
                                                                       'in_ucast_pkts': '482546907',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '592686974138',
                                                                                                'in_ucast_pkts': '482546907',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '38496656814',
                                                                                                'out_ucast_pkts': '128804503',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '38496656814',
                                                                       'out_ucast_pkts': '128804503',
                                                                       'poller_hash': 'MTI3NEFBRHBlZXJpbnRmcG9uIDIzNDM2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234367',
                                                                                               'links': [       {       'connected_time': '3971929',
                                                                                                                        'fiber_length': 1011,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFBRHBlZXJwb25wb24gMjM0MzY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 MILL ST W (BRAUN DAVID & '
                                                 'MARGARETHA)',
                                         'poller_hash': 'MTI3NEFBRA==',
                                         'serial_number': '1274AAD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:21:69',
                                         'uptime': '3971929'},
                                 {       'counter_source': '234397/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234397/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '174264850412',
                                                                       'in_ucast_pkts': '913822117',
                                                                       'mac_address': '04:bc:9f:5c:1e:51',
                                                                       'name': '234397/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '116',
                                                                       'out_octets': '4668963075941',
                                                                       'out_ucast_pkts': '3494046631',
                                                                       'poller_hash': 'MTI3NEE0QXBlZXJpbnRmMjM0Mzk3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234397',
                                                                       'in_nucast_pkts': '116',
                                                                       'in_octets': '4668963075941',
                                                                       'in_ucast_pkts': '3494046631',
                                                                       'links': [       {       'in_nucast_pkts': '116',
                                                                                                'in_octets': '4668963075941',
                                                                                                'in_ucast_pkts': '3494046631',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '174264850412',
                                                                                                'out_ucast_pkts': '913822117',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '174264850412',
                                                                       'out_ucast_pkts': '913822117',
                                                                       'poller_hash': 'MTI3NEE0QXBlZXJpbnRmcG9uIDIzNDM5Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234397',
                                                                                               'links': [       {       'connected_time': '688588',
                                                                                                                        'fiber_length': 10535,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE0QXBlZXJwb25wb24gMjM0Mzk3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4112 LINE 61 (JANTZI NICK)',
                                         'poller_hash': 'MTI3NEE0QQ==',
                                         'serial_number': '1274A4A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:51',
                                         'uptime': '688588'},
                                 {       'counter_source': '234407/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e6:be'],
                                                                       'id': '234407/g1',
                                                                       'in_nucast_pkts': '11948',
                                                                       'in_octets': '4273869246',
                                                                       'in_ucast_pkts': '21737144',
                                                                       'mac_address': '04:bc:9f:5c:1b:64',
                                                                       'mtu': 2000,
                                                                       'name': '234407/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1459633',
                                                                       'out_octets': '77225665481',
                                                                       'out_ucast_pkts': '63261385',
                                                                       'poller_hash': 'MTI3NDlFQ3BlZXJpbnRmMjM0NDA3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '234407',
                                                                       'in_nucast_pkts': '1459633',
                                                                       'in_octets': '77225665481',
                                                                       'in_ucast_pkts': '63261385',
                                                                       'links': [       {       'in_nucast_pkts': '1459633',
                                                                                                'in_octets': '77225665481',
                                                                                                'in_ucast_pkts': '63261385',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11948',
                                                                                                'out_octets': '4273869246',
                                                                                                'out_ucast_pkts': '21737144',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11948',
                                                                       'out_octets': '4273869246',
                                                                       'out_ucast_pkts': '21737144',
                                                                       'poller_hash': 'MTI3NDlFQ3BlZXJpbnRmcG9uIDIzNDQwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234407',
                                                                                               'links': [       {       'connected_time': '686551',
                                                                                                                        'fiber_length': 1040,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlFQ3BlZXJwb25wb24gMjM0NDA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '57 MILL ST E MILVERTON (HUNTER BILL)',
                                         'poller_hash': 'MTI3NDlFQw==',
                                         'serial_number': '12749EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:61',
                                         'uptime': '686551'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '234427/g1',
                                                                       'name': '234427/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEFCMHBlZXJpbnRmMjM0NDI3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234427',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEFCMHBlZXJpbnRmcG9uIDIzNDQyNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234427',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFCMHBlZXJwb25wb24gMjM0NDI3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4740 LINE 67 (5588269 ONTARIO INC)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "234427" '
                                                                  '(location: GPON '
                                                                  'port "13/1"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2024-06-12T05:57:26.730Z',
                                         'poller_hash': 'MTI3NEFCMA==',
                                         'serial_number': '1274AB0',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '234437/g1',
                                                                       'name': '234437/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEI0NHBlZXJpbnRmMjM0NDM3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234437',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEI0NHBlZXJpbnRmcG9uIDIzNDQzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234437',
                                                                                               'links': [       {       'connected_time': '3634524',
                                                                                                                        'fiber_length': 9207,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEI0NHBlZXJwb25wb24gMjM0NDM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4712 LINE 67 (ALBRECHT ELDON & '
                                                 'CHRISTENA)',
                                         'poller_hash': 'MTI3NEI0NA==',
                                         'serial_number': '1274B44',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:26:21',
                                         'uptime': '3634524'},
                                 {       'counter_source': '234447/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234447/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '972968078297',
                                                                       'in_ucast_pkts': '96219911399',
                                                                       'mac_address': '04:bc:9f:5c:20:01',
                                                                       'name': '234447/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '99',
                                                                       'out_octets': '5386609605013',
                                                                       'out_ucast_pkts': '4437757882',
                                                                       'poller_hash': 'MTI3NEE4MHBlZXJpbnRmMjM0NDQ3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234447',
                                                                       'in_nucast_pkts': '99',
                                                                       'in_octets': '5386609605013',
                                                                       'in_ucast_pkts': '4437757882',
                                                                       'links': [       {       'in_nucast_pkts': '99',
                                                                                                'in_octets': '5386609605013',
                                                                                                'in_ucast_pkts': '4437757882',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '972968078297',
                                                                                                'out_ucast_pkts': '96219911399',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '972968078297',
                                                                       'out_ucast_pkts': '96219911399',
                                                                       'poller_hash': 'MTI3NEE4MHBlZXJpbnRmcG9uIDIzNDQ0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234447',
                                                                                               'links': [       {       'connected_time': '619253',
                                                                                                                        'fiber_length': 11211,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE4MHBlZXJwb25wb24gMjM0NDQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "6878 ROAD 116 (JOHN'S TRANSIT)",
                                         'poller_hash': 'MTI3NEE4MA==',
                                         'serial_number': '1274A80',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:01',
                                         'uptime': '619253'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '234524/g1',
                                                                       'name': '234524/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEEyRXBlZXJpbnRmMjM0NTI0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234524',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEEyRXBlZXJpbnRmcG9uIDIzNDUyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234524',
                                                                                               'links': [       {       'connected_time': '3838593',
                                                                                                                        'fiber_length': 7924,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEyRXBlZXJwb25wb24gMjM0NTI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4493 LINE 61 (ALBRECHT KEN K)',
                                         'poller_hash': 'MTI3NEEyRQ==',
                                         'serial_number': '1274A2E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:71',
                                         'uptime': '3838593'},
                                 {       'counter_source': '234534/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:f9:fb:fc'],
                                                                       'id': '234534/g1',
                                                                       'in_nucast_pkts': '3467660',
                                                                       'in_octets': '236416050611',
                                                                       'in_ucast_pkts': '3235405641',
                                                                       'mac_address': '04:bc:9f:5c:1f:7c',
                                                                       'mtu': 2000,
                                                                       'name': '234534/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '192261729',
                                                                       'out_octets': '19191505275061',
                                                                       'out_ucast_pkts': '12939750844',
                                                                       'poller_hash': 'MTI3NEE2RnBlZXJpbnRmMjM0NTM0L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '234534',
                                                                       'in_nucast_pkts': '192261729',
                                                                       'in_octets': '19191505275061',
                                                                       'in_ucast_pkts': '12939750844',
                                                                       'links': [       {       'in_nucast_pkts': '192261729',
                                                                                                'in_octets': '19191505275061',
                                                                                                'in_ucast_pkts': '12939750844',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3467660',
                                                                                                'out_octets': '236416050611',
                                                                                                'out_ucast_pkts': '3235405641',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3467660',
                                                                       'out_octets': '236416050611',
                                                                       'out_ucast_pkts': '3235405641',
                                                                       'poller_hash': 'MTI3NEE2RnBlZXJpbnRmcG9uIDIzNDUzNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234534',
                                                                                               'links': [       {       'connected_time': '3624827',
                                                                                                                        'fiber_length': 10571,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE2RnBlZXJwb25wb24gMjM0NTM0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6405 PERTH ROAD 121 (LEIS ALBERT)',
                                         'poller_hash': 'MTI3NEE2Rg==',
                                         'serial_number': '1274A6F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:79',
                                         'uptime': '3624827'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '234564/g1',
                                                                       'name': '234564/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDk2RnBlZXJpbnRmMjM0NTY0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234564',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDk2RnBlZXJpbnRmcG9uIDIzNDU2NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234564',
                                                                                               'links': [       {       'connected_time': '3539036',
                                                                                                                        'fiber_length': 1244,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk2RnBlZXJwb25wb24gMjM0NTY0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 GEORGE ST (WAGLER CATHERINE)',
                                         'poller_hash': 'MTI3NDk2Rg==',
                                         'serial_number': '127496F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:17:79',
                                         'uptime': '3539036'},
                                 {       'counter_source': '234574/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234574/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3117635788',
                                                                       'in_ucast_pkts': '27381705',
                                                                       'mac_address': '04:bc:9f:5c:14:b9',
                                                                       'name': '234574/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '23672633929',
                                                                       'out_ucast_pkts': '87456778',
                                                                       'poller_hash': 'MTI3NDkxN3BlZXJpbnRmMjM0NTc0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234574',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '23672633929',
                                                                       'in_ucast_pkts': '87456778',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '23672633929',
                                                                                                'in_ucast_pkts': '87456778',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3117635788',
                                                                                                'out_ucast_pkts': '27381705',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3117635788',
                                                                       'out_ucast_pkts': '27381705',
                                                                       'poller_hash': 'MTI3NDkxN3BlZXJpbnRmcG9uIDIzNDU3NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234574',
                                                                                               'links': [       {       'connected_time': '3114013',
                                                                                                                        'fiber_length': 12577,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDkxN3BlZXJwb25wb24gMjM0NTc0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5384 PERTH LINE 55 (PARTRIDGE RYAN)',
                                         'poller_hash': 'MTI3NDkxNw==',
                                         'serial_number': '1274917',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:14:b9',
                                         'uptime': '3114013'},
                                 {       'counter_source': '234844/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e2:7d'],
                                                                       'id': '234844/g1',
                                                                       'in_nucast_pkts': '12483',
                                                                       'in_octets': '2761326444',
                                                                       'in_ucast_pkts': '11123402',
                                                                       'mac_address': '04:bc:9f:5c:18:e4',
                                                                       'mtu': 2000,
                                                                       'name': '234844/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465054',
                                                                       'out_octets': '105724977184',
                                                                       'out_ucast_pkts': '75653884',
                                                                       'poller_hash': 'MTI3NDk5Q3BlZXJpbnRmMjM0ODQ0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '234844',
                                                                       'in_nucast_pkts': '1465054',
                                                                       'in_octets': '105724977184',
                                                                       'in_ucast_pkts': '75653884',
                                                                       'links': [       {       'in_nucast_pkts': '1465054',
                                                                                                'in_octets': '105724977184',
                                                                                                'in_ucast_pkts': '75653884',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12483',
                                                                                                'out_octets': '2761326444',
                                                                                                'out_ucast_pkts': '11123402',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12483',
                                                                       'out_octets': '2761326444',
                                                                       'out_ucast_pkts': '11123402',
                                                                       'poller_hash': 'MTI3NDk5Q3BlZXJpbnRmcG9uIDIzNDg0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234844',
                                                                                               'links': [       {       'connected_time': '688599',
                                                                                                                        'fiber_length': 11442,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk5Q3BlZXJwb25wb24gMjM0ODQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4011 LINE 61 (BANKS BRETT)',
                                         'poller_hash': 'MTI3NDk5Qw==',
                                         'serial_number': '127499C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:18:e1',
                                         'uptime': '688599'},
                                 {       'counter_source': '234854/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234854/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '155703074788',
                                                                       'in_ucast_pkts': '1095009707',
                                                                       'mac_address': '04:bc:9f:5c:19:41',
                                                                       'name': '234854/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '4236094221439',
                                                                       'out_ucast_pkts': '2963544372',
                                                                       'poller_hash': 'MTI3NDlBOHBlZXJpbnRmMjM0ODU0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234854',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '4236094221439',
                                                                       'in_ucast_pkts': '2963544372',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '4236094221439',
                                                                                                'in_ucast_pkts': '2963544372',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '155703074788',
                                                                                                'out_ucast_pkts': '1095009707',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '155703074788',
                                                                       'out_ucast_pkts': '1095009707',
                                                                       'poller_hash': 'MTI3NDlBOHBlZXJpbnRmcG9uIDIzNDg1NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234854',
                                                                                               'links': [       {       'connected_time': '688498',
                                                                                                                        'fiber_length': 1364,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlBOHBlZXJwb25wb24gMjM0ODU0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 MAIN ST N (DURAN PEREZ SERGIO '
                                                 'TOMAS)',
                                         'poller_hash': 'MTI3NDlBOA==',
                                         'serial_number': '12749A8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:19:41',
                                         'uptime': '688498'},
                                 {       'counter_source': '234874/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234874/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '285629702355',
                                                                       'in_ucast_pkts': '2333892997',
                                                                       'mac_address': '04:bc:9f:5c:1c:01',
                                                                       'name': '234874/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '140',
                                                                       'out_octets': '6593979666414',
                                                                       'out_ucast_pkts': '4695613786',
                                                                       'poller_hash': 'MTI3NEEwMHBlZXJpbnRmMjM0ODc0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234874',
                                                                       'in_nucast_pkts': '140',
                                                                       'in_octets': '6593979666414',
                                                                       'in_ucast_pkts': '4695613786',
                                                                       'links': [       {       'in_nucast_pkts': '140',
                                                                                                'in_octets': '6593979666414',
                                                                                                'in_ucast_pkts': '4695613786',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '285629702355',
                                                                                                'out_ucast_pkts': '2333892997',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '285629702355',
                                                                       'out_ucast_pkts': '2333892997',
                                                                       'poller_hash': 'MTI3NEEwMHBlZXJpbnRmcG9uIDIzNDg3NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234874',
                                                                                               'links': [       {       'connected_time': '3363277',
                                                                                                                        'fiber_length': 11490,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwMHBlZXJwb25wb24gMjM0ODc0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4007 LINE 61 (BINKLE MURRAY)',
                                         'poller_hash': 'MTI3NEEwMA==',
                                         'serial_number': '1274A00',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:01',
                                         'uptime': '3363277'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '234894/g1',
                                                                       'name': '234894/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDlGQXBlZXJpbnRmMjM0ODk0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234894',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDlGQXBlZXJpbnRmcG9uIDIzNDg5NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234894',
                                                                                               'links': [       {       'connected_time': '3036342',
                                                                                                                        'fiber_length': 12104,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlGQXBlZXJwb25wb24gMjM0ODk0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4228 LINE 61 (GASCHO DAN)',
                                         'poller_hash': 'MTI3NDlGQQ==',
                                         'serial_number': '12749FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:d1',
                                         'uptime': '3036342'},
                                 {       'counter_source': '234904/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:eb:31'],
                                                                       'id': '234904/g1',
                                                                       'in_nucast_pkts': '11759',
                                                                       'in_octets': '6051245042',
                                                                       'in_ucast_pkts': '10702104',
                                                                       'mac_address': '04:bc:9f:5c:1c:6c',
                                                                       'mtu': 2000,
                                                                       'name': '234904/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465092',
                                                                       'out_octets': '48754060746',
                                                                       'out_ucast_pkts': '39295509',
                                                                       'poller_hash': 'MTI3NEEwRHBlZXJpbnRmMjM0OTA0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '234904',
                                                                       'in_nucast_pkts': '1465092',
                                                                       'in_octets': '48754060746',
                                                                       'in_ucast_pkts': '39295509',
                                                                       'links': [       {       'in_nucast_pkts': '1465092',
                                                                                                'in_octets': '48754060746',
                                                                                                'in_ucast_pkts': '39295509',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11759',
                                                                                                'out_octets': '6051245042',
                                                                                                'out_ucast_pkts': '10702104',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11759',
                                                                       'out_octets': '6051245042',
                                                                       'out_ucast_pkts': '10702104',
                                                                       'poller_hash': 'MTI3NEEwRHBlZXJpbnRmcG9uIDIzNDkwNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234904',
                                                                                               'links': [       {       'connected_time': '688595',
                                                                                                                        'fiber_length': 10619,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwRHBlZXJwb25wb24gMjM0OTA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4091 LINE 61 (KUEPFER ROGER)',
                                         'poller_hash': 'MTI3NEEwRA==',
                                         'serial_number': '1274A0D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:69',
                                         'uptime': '688595'},
                                 {       'counter_source': '234914/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:71:d9'],
                                                                       'id': '234914/g1',
                                                                       'in_nucast_pkts': '123838',
                                                                       'in_octets': '30398753323',
                                                                       'in_ucast_pkts': '374665557',
                                                                       'mac_address': '04:bc:9f:5c:4b:cc',
                                                                       'mtu': 2000,
                                                                       'name': '234914/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13466096',
                                                                       'out_octets': '1956672456503',
                                                                       'out_ucast_pkts': '1309326951',
                                                                       'poller_hash': 'MTI3NEZGOXBlZXJpbnRmMjM0OTE0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '234914',
                                                                       'in_nucast_pkts': '13466096',
                                                                       'in_octets': '1956672456503',
                                                                       'in_ucast_pkts': '1309326951',
                                                                       'links': [       {       'in_nucast_pkts': '13466096',
                                                                                                'in_octets': '1956672456503',
                                                                                                'in_ucast_pkts': '1309326951',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '123838',
                                                                                                'out_octets': '30398753323',
                                                                                                'out_ucast_pkts': '374665557',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '123838',
                                                                       'out_octets': '30398753323',
                                                                       'out_ucast_pkts': '374665557',
                                                                       'poller_hash': 'MTI3NEZGOXBlZXJpbnRmcG9uIDIzNDkxNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234914',
                                                                                               'links': [       {       'connected_time': '5868607',
                                                                                                                        'fiber_length': 7931,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEZGOXBlZXJwb25wb24gMjM0OTE0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4908 LINE 59 (DIEHL FRAZER FARMING '
                                                 'LTD)',
                                         'poller_hash': 'MTI3NEZGOQ==',
                                         'serial_number': '1274FF9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4b:c9',
                                         'uptime': '5868607'},
                                 {       'counter_source': '234924/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234924/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '33799128668',
                                                                       'in_ucast_pkts': '122028668',
                                                                       'mac_address': '04:bc:9f:5c:18:99',
                                                                       'name': '234924/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '140',
                                                                       'out_octets': '484263124752',
                                                                       'out_ucast_pkts': '432976305',
                                                                       'poller_hash': 'MTI3NDk5M3BlZXJpbnRmMjM0OTI0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234924',
                                                                       'in_nucast_pkts': '140',
                                                                       'in_octets': '484263124752',
                                                                       'in_ucast_pkts': '432976305',
                                                                       'links': [       {       'in_nucast_pkts': '140',
                                                                                                'in_octets': '484263124752',
                                                                                                'in_ucast_pkts': '432976305',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '33799128668',
                                                                                                'out_ucast_pkts': '122028668',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '33799128668',
                                                                       'out_ucast_pkts': '122028668',
                                                                       'poller_hash': 'MTI3NDk5M3BlZXJpbnRmcG9uIDIzNDkyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234924',
                                                                                               'links': [       {       'connected_time': '2948034',
                                                                                                                        'fiber_length': 10638,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk5M3BlZXJwb25wb24gMjM0OTI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4087 LINE 61 (GOOD DWIGHT)',
                                         'poller_hash': 'MTI3NDk5Mw==',
                                         'serial_number': '1274993',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:18:99',
                                         'uptime': '2948034'},
                                 {       'counter_source': '234934/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '234934/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '480631280405',
                                                                       'in_ucast_pkts': '93978306305',
                                                                       'mac_address': '04:bc:9f:5c:1e:11',
                                                                       'name': '234934/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '180',
                                                                       'out_octets': '11132308048705',
                                                                       'out_ucast_pkts': '7863239018',
                                                                       'poller_hash': 'MTI3NEE0MnBlZXJpbnRmMjM0OTM0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '234934',
                                                                       'in_nucast_pkts': '180',
                                                                       'in_octets': '11132308048705',
                                                                       'in_ucast_pkts': '7863239018',
                                                                       'links': [       {       'in_nucast_pkts': '180',
                                                                                                'in_octets': '11132308048705',
                                                                                                'in_ucast_pkts': '7863239018',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '480631280405',
                                                                                                'out_ucast_pkts': '93978306305',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '480631280405',
                                                                       'out_ucast_pkts': '93978306305',
                                                                       'poller_hash': 'MTI3NEE0MnBlZXJpbnRmcG9uIDIzNDkzNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '234934',
                                                                                               'links': [       {       'connected_time': '1880134',
                                                                                                                        'fiber_length': 11199,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE0MnBlZXJwb25wb24gMjM0OTM0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4164 LINE 61 (SIEGNER PAUL)',
                                         'poller_hash': 'MTI3NEE0Mg==',
                                         'serial_number': '1274A42',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:11',
                                         'uptime': '1880134'},
                                 {       'counter_source': '235018/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235018/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '212724707209',
                                                                       'in_ucast_pkts': '5065624881',
                                                                       'mac_address': '04:bc:9f:5c:1d:41',
                                                                       'name': '235018/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '37',
                                                                       'out_octets': '3013811157740',
                                                                       'out_ucast_pkts': '2296872542',
                                                                       'poller_hash': 'MTI3NEEyOHBlZXJpbnRmMjM1MDE4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235018',
                                                                       'in_nucast_pkts': '37',
                                                                       'in_octets': '3013811157740',
                                                                       'in_ucast_pkts': '2296872542',
                                                                       'links': [       {       'in_nucast_pkts': '37',
                                                                                                'in_octets': '3013811157740',
                                                                                                'in_ucast_pkts': '2296872542',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '212724707209',
                                                                                                'out_ucast_pkts': '5065624881',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '212724707209',
                                                                       'out_ucast_pkts': '5065624881',
                                                                       'poller_hash': 'MTI3NEEyOHBlZXJpbnRmcG9uIDIzNTAxOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235018',
                                                                                               'links': [       {       'connected_time': '7004732',
                                                                                                                        'fiber_length': 8377,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEyOHBlZXJwb25wb24gMjM1MDE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4788 LINE 67 (EGLI ROGER)',
                                         'poller_hash': 'MTI3NEEyOA==',
                                         'serial_number': '1274A28',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:41',
                                         'uptime': '7004732'},
                                 {       'counter_source': '235028/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235028/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '49316725521',
                                                                       'in_ucast_pkts': '4435800741',
                                                                       'mac_address': '04:bc:9f:5c:1c:79',
                                                                       'name': '235028/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '667081018999',
                                                                       'out_ucast_pkts': '540103572',
                                                                       'poller_hash': 'MTI3NEEwRnBlZXJpbnRmMjM1MDI4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235028',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '667081018999',
                                                                       'in_ucast_pkts': '540103572',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '667081018999',
                                                                                                'in_ucast_pkts': '540103572',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '49316725521',
                                                                                                'out_ucast_pkts': '4435800741',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '49316725521',
                                                                       'out_ucast_pkts': '4435800741',
                                                                       'poller_hash': 'MTI3NEEwRnBlZXJpbnRmcG9uIDIzNTAyOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235028',
                                                                                               'links': [       {       'connected_time': '2927154',
                                                                                                                        'fiber_length': 5507,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwRnBlZXJwb25wb24gMjM1MDI4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4410 LINE 64 (EGLI URS)',
                                         'poller_hash': 'MTI3NEEwRg==',
                                         'serial_number': '1274A0F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:79',
                                         'uptime': '2927154'},
                                 {       'counter_source': '235102/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235102/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6818502798',
                                                                       'in_ucast_pkts': '36093911',
                                                                       'mac_address': '04:bc:9f:5c:24:79',
                                                                       'name': '235102/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '160',
                                                                       'out_octets': '89767618020',
                                                                       'out_ucast_pkts': '119692513',
                                                                       'poller_hash': 'MTI3NEIwRnBlZXJpbnRmMjM1MTAyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235102',
                                                                       'in_nucast_pkts': '160',
                                                                       'in_octets': '89767618020',
                                                                       'in_ucast_pkts': '119692513',
                                                                       'links': [       {       'in_nucast_pkts': '160',
                                                                                                'in_octets': '89767618020',
                                                                                                'in_ucast_pkts': '119692513',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '6818502798',
                                                                                                'out_ucast_pkts': '36093911',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6818502798',
                                                                       'out_ucast_pkts': '36093911',
                                                                       'poller_hash': 'MTI3NEIwRnBlZXJpbnRmcG9uIDIzNTEwMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235102',
                                                                                               'links': [       {       'connected_time': '2776283',
                                                                                                                        'fiber_length': 12190,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIwRnBlZXJwb25wb24gMjM1MTAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4232 LINE 61 (ERB WAYNE)',
                                         'poller_hash': 'MTI3NEIwRg==',
                                         'serial_number': '1274B0F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:79',
                                         'uptime': '2776283'},
                                 {       'counter_source': '235112/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235112/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '277474153289',
                                                                       'in_ucast_pkts': '586933743',
                                                                       'mac_address': '04:bc:9f:5c:1d:e1',
                                                                       'name': '235112/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '140',
                                                                       'out_octets': '1257180743871',
                                                                       'out_ucast_pkts': '1215484373',
                                                                       'poller_hash': 'MTI3NEEzQ3BlZXJpbnRmMjM1MTEyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235112',
                                                                       'in_nucast_pkts': '140',
                                                                       'in_octets': '1257180743871',
                                                                       'in_ucast_pkts': '1215484373',
                                                                       'links': [       {       'in_nucast_pkts': '140',
                                                                                                'in_octets': '1257180743871',
                                                                                                'in_ucast_pkts': '1215484373',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '277474153289',
                                                                                                'out_ucast_pkts': '586933743',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '277474153289',
                                                                       'out_ucast_pkts': '586933743',
                                                                       'poller_hash': 'MTI3NEEzQ3BlZXJpbnRmcG9uIDIzNTExMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235112',
                                                                                               'links': [       {       'connected_time': '28280185',
                                                                                                                        'fiber_length': 11468,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEzQ3BlZXJwb25wb24gMjM1MTEy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4013 LINE 61 (KUEPFER ROSS D)',
                                         'poller_hash': 'MTI3NEEzQw==',
                                         'serial_number': '1274A3C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:e1',
                                         'uptime': '28280185'},
                                 {       'counter_source': '235122/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d5:a3'],
                                                                       'id': '235122/g1',
                                                                       'in_nucast_pkts': '14957',
                                                                       'in_octets': '3567755315',
                                                                       'in_ucast_pkts': '20607622',
                                                                       'mac_address': '04:bc:9f:5c:19:2c',
                                                                       'mtu': 2000,
                                                                       'name': '235122/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464888',
                                                                       'out_octets': '83165836258',
                                                                       'out_ucast_pkts': '62055901',
                                                                       'poller_hash': 'MTI3NDlBNXBlZXJpbnRmMjM1MTIyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '235122',
                                                                       'in_nucast_pkts': '1464888',
                                                                       'in_octets': '83165836258',
                                                                       'in_ucast_pkts': '62055901',
                                                                       'links': [       {       'in_nucast_pkts': '1464888',
                                                                                                'in_octets': '83165836258',
                                                                                                'in_ucast_pkts': '62055901',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14957',
                                                                                                'out_octets': '3567755315',
                                                                                                'out_ucast_pkts': '20607622',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14957',
                                                                       'out_octets': '3567755315',
                                                                       'out_ucast_pkts': '20607622',
                                                                       'poller_hash': 'MTI3NDlBNXBlZXJpbnRmcG9uIDIzNTEyMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235122',
                                                                                               'links': [       {       'connected_time': '688358',
                                                                                                                        'fiber_length': 2555,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlBNXBlZXJwb25wb24gMjM1MTIy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '101 PUGH ST MILVERTON (HENHOEFFER '
                                                 'BONNIE)',
                                         'poller_hash': 'MTI3NDlBNQ==',
                                         'serial_number': '12749A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:19:29',
                                         'uptime': '688358'},
                                 {       'counter_source': '235132/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235132/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '389015036882',
                                                                       'in_ucast_pkts': '103499865235',
                                                                       'mac_address': '04:bc:9f:5c:22:a9',
                                                                       'name': '235132/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '94',
                                                                       'out_octets': '465229984277',
                                                                       'out_ucast_pkts': '527740059',
                                                                       'poller_hash': 'MTI3NEFENXBlZXJpbnRmMjM1MTMyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235132',
                                                                       'in_nucast_pkts': '94',
                                                                       'in_octets': '465229984277',
                                                                       'in_ucast_pkts': '527740059',
                                                                       'links': [       {       'in_nucast_pkts': '94',
                                                                                                'in_octets': '465229984277',
                                                                                                'in_ucast_pkts': '527740059',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '389015036882',
                                                                                                'out_ucast_pkts': '103499865235',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '389015036882',
                                                                       'out_ucast_pkts': '103499865235',
                                                                       'poller_hash': 'MTI3NEFENXBlZXJpbnRmcG9uIDIzNTEzMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235132',
                                                                                               'links': [       {       'connected_time': '65750201',
                                                                                                                        'fiber_length': 7872,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFENXBlZXJwb25wb24gMjM1MTMy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6551 ROAD 124 (MARTIN AARON F)',
                                         'poller_hash': 'MTI3NEFENQ==',
                                         'serial_number': '1274AD5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:a9',
                                         'uptime': '65750201'},
                                 {       'counter_source': '235408/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235408/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '416522856857',
                                                                       'in_ucast_pkts': '86778322643',
                                                                       'mac_address': '04:bc:9f:5c:22:49',
                                                                       'name': '235408/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '104',
                                                                       'out_octets': '2157454306471',
                                                                       'out_ucast_pkts': '1964015540',
                                                                       'poller_hash': 'MTI3NEFDOXBlZXJpbnRmMjM1NDA4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235408',
                                                                       'in_nucast_pkts': '104',
                                                                       'in_octets': '2157454306471',
                                                                       'in_ucast_pkts': '1964015540',
                                                                       'links': [       {       'in_nucast_pkts': '104',
                                                                                                'in_octets': '2157454306471',
                                                                                                'in_ucast_pkts': '1964015540',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '416522856857',
                                                                                                'out_ucast_pkts': '86778322643',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '416522856857',
                                                                       'out_ucast_pkts': '86778322643',
                                                                       'poller_hash': 'MTI3NEFDOXBlZXJpbnRmcG9uIDIzNTQwOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235408',
                                                                                               'links': [       {       'connected_time': '20538292',
                                                                                                                        'fiber_length': 7564,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFDOXBlZXJwb25wb24gMjM1NDA4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6502 ROAD 124 (JANTZI RONALD)',
                                         'poller_hash': 'MTI3NEFDOQ==',
                                         'serial_number': '1274AC9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:49',
                                         'uptime': '20538292'},
                                 {       'counter_source': '235418/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:53:52'],
                                                                       'id': '235418/g1',
                                                                       'in_nucast_pkts': '642083',
                                                                       'in_octets': '180138581458',
                                                                       'in_ucast_pkts': '1883213417',
                                                                       'mac_address': '04:bc:9f:5b:e5:4c',
                                                                       'mtu': 2000,
                                                                       'name': '235418/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82419359',
                                                                       'out_octets': '16655854396546',
                                                                       'out_ucast_pkts': '11159138337',
                                                                       'poller_hash': 'MTI3NDMyOXBlZXJpbnRmMjM1NDE4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '235418',
                                                                       'in_nucast_pkts': '82419359',
                                                                       'in_octets': '16655854396546',
                                                                       'in_ucast_pkts': '11159138337',
                                                                       'links': [       {       'in_nucast_pkts': '82419359',
                                                                                                'in_octets': '16655854396546',
                                                                                                'in_ucast_pkts': '11159138337',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '642083',
                                                                                                'out_octets': '180138581458',
                                                                                                'out_ucast_pkts': '1883213417',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '642083',
                                                                       'out_octets': '180138581458',
                                                                       'out_ucast_pkts': '1883213417',
                                                                       'poller_hash': 'MTI3NDMyOXBlZXJpbnRmcG9uIDIzNTQxOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235418',
                                                                                               'links': [       {       'connected_time': '67333726',
                                                                                                                        'fiber_length': 1361,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMyOXBlZXJwb25wb24gMjM1NDE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 EDWARD ST (HOLST NELSON)',
                                         'poller_hash': 'MTI3NDMyOQ==',
                                         'serial_number': '1274329',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:49',
                                         'uptime': '67333726'},
                                 {       'counter_source': '235428/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235428/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '110952404187',
                                                                       'in_ucast_pkts': '179499754',
                                                                       'mac_address': '04:bc:9f:5c:21:51',
                                                                       'name': '235428/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '344493401493',
                                                                       'out_ucast_pkts': '336343626',
                                                                       'poller_hash': 'MTI3NEFBQXBlZXJpbnRmMjM1NDI4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235428',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '344493401493',
                                                                       'in_ucast_pkts': '336343626',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '344493401493',
                                                                                                'in_ucast_pkts': '336343626',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '110952404187',
                                                                                                'out_ucast_pkts': '179499754',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '110952404187',
                                                                       'out_ucast_pkts': '179499754',
                                                                       'poller_hash': 'MTI3NEFBQXBlZXJpbnRmcG9uIDIzNTQyOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235428',
                                                                                               'links': [       {       'connected_time': '2345641',
                                                                                                                        'fiber_length': 7555,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFBQXBlZXJwb25wb24gMjM1NDI4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6497 ROAD 124 (WAGLER JEREMY)',
                                         'poller_hash': 'MTI3NEFBQQ==',
                                         'serial_number': '1274AAA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:21:51',
                                         'uptime': '2345641'},
                                 {       'counter_source': '235448/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235448/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '287247957729',
                                                                       'in_ucast_pkts': '57731141044',
                                                                       'mac_address': '04:bc:9f:5c:1e:61',
                                                                       'name': '235448/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '8037748210112',
                                                                       'out_ucast_pkts': '5583083504',
                                                                       'poller_hash': 'MTI3NEE0Q3BlZXJpbnRmMjM1NDQ4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235448',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '8037748210112',
                                                                       'in_ucast_pkts': '5583083504',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '8037748210112',
                                                                                                'in_ucast_pkts': '5583083504',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '287247957729',
                                                                                                'out_ucast_pkts': '57731141044',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '287247957729',
                                                                       'out_ucast_pkts': '57731141044',
                                                                       'poller_hash': 'MTI3NEE0Q3BlZXJpbnRmcG9uIDIzNTQ0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235448',
                                                                                               'links': [       {       'connected_time': '3113952',
                                                                                                                        'fiber_length': 17910,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE0Q3BlZXJwb25wb24gMjM1NDQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '162 SMITH ST (SMITH BRENDA)',
                                         'poller_hash': 'MTI3NEE0Qw==',
                                         'serial_number': '1274A4C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:61',
                                         'uptime': '3113952'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '235458/g1',
                                                                       'name': '235458/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDlDNnBlZXJpbnRmMjM1NDU4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235458',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDlDNnBlZXJpbnRmcG9uIDIzNTQ1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235458',
                                                                                               'links': [       {       'connected_time': '2258054',
                                                                                                                        'fiber_length': 11664,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlDNnBlZXJwb25wb24gMjM1NDU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4001 LINE 61 (GINGERICH CLARE)',
                                         'poller_hash': 'MTI3NDlDNg==',
                                         'serial_number': '12749C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1a:31',
                                         'uptime': '2258054'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '235516/g1',
                                                                       'name': '235516/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDk1RXBlZXJpbnRmMjM1NTE2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235516',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDk1RXBlZXJpbnRmcG9uIDIzNTUxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235516',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk1RXBlZXJwb25wb24gMjM1NTE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4094A LINE 61 (STECKLY SHEILA)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "235516" '
                                                                  '(location: GPON '
                                                                  'port "14/3"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-03-26T16:44:08.790Z',
                                         'poller_hash': 'MTI3NDk1RQ==',
                                         'serial_number': '127495E',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '235723/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235723/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '777561258578',
                                                                       'in_ucast_pkts': '1865115614',
                                                                       'mac_address': '04:bc:9f:5c:18:39',
                                                                       'name': '235723/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '72',
                                                                       'out_octets': '6493085087255',
                                                                       'out_ucast_pkts': '5438772427',
                                                                       'poller_hash': 'MTI3NDk4N3BlZXJpbnRmMjM1NzIzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235723',
                                                                       'in_nucast_pkts': '72',
                                                                       'in_octets': '6493085087255',
                                                                       'in_ucast_pkts': '5438772427',
                                                                       'links': [       {       'in_nucast_pkts': '72',
                                                                                                'in_octets': '6493085087255',
                                                                                                'in_ucast_pkts': '5438772427',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '777561258578',
                                                                                                'out_ucast_pkts': '1865115614',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '777561258578',
                                                                       'out_ucast_pkts': '1865115614',
                                                                       'poller_hash': 'MTI3NDk4N3BlZXJpbnRmcG9uIDIzNTcyMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235723',
                                                                                               'links': [       {       'connected_time': '688568',
                                                                                                                        'fiber_length': 9577,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk4N3BlZXJwb25wb24gMjM1NzIz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3984 LINE 67 (ZEHR TERRY & SUE)',
                                         'poller_hash': 'MTI3NDk4Nw==',
                                         'serial_number': '1274987',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:18:39',
                                         'uptime': '688568'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '235756/g1',
                                                                       'name': '235756/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDlGMXBlZXJpbnRmMjM1NzU2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235756',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDlGMXBlZXJpbnRmcG9uIDIzNTc1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235756',
                                                                                               'links': [       {       'connected_time': '77290',
                                                                                                                        'fiber_length': 10768,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlGMXBlZXJwb25wb24gMjM1NzU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6376 PERTH ROAD 121 (WAGLER CHARLES)',
                                         'poller_hash': 'MTI3NDlGMQ==',
                                         'serial_number': '12749F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:89',
                                         'uptime': '77290'},
                                 {       'counter_source': '236100/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236100/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '276252340980',
                                                                       'in_ucast_pkts': '1884232501',
                                                                       'mac_address': '04:bc:9f:5c:20:51',
                                                                       'name': '236100/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '86',
                                                                       'out_octets': '10536060815107',
                                                                       'out_ucast_pkts': '7289618910',
                                                                       'poller_hash': 'MTI3NEE4QXBlZXJpbnRmMjM2MTAwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236100',
                                                                       'in_nucast_pkts': '86',
                                                                       'in_octets': '10536060815107',
                                                                       'in_ucast_pkts': '7289618910',
                                                                       'links': [       {       'in_nucast_pkts': '86',
                                                                                                'in_octets': '10536060815107',
                                                                                                'in_ucast_pkts': '7289618910',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '276252340980',
                                                                                                'out_ucast_pkts': '1884232501',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '276252340980',
                                                                       'out_ucast_pkts': '1884232501',
                                                                       'poller_hash': 'MTI3NEE4QXBlZXJpbnRmcG9uIDIzNjEwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236100',
                                                                                               'links': [       {       'connected_time': '1206725',
                                                                                                                        'fiber_length': 18739,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE4QXBlZXJwb25wb24gMjM2MTAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5583 ROAD 164 MONKTON (FRANCIS '
                                                 'CHRISTINE)',
                                         'poller_hash': 'MTI3NEE4QQ==',
                                         'serial_number': '1274A8A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:51',
                                         'uptime': '1206725'},
                                 {       'counter_source': '236110/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236110/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '758793497519',
                                                                       'in_ucast_pkts': '3117822550',
                                                                       'mac_address': '04:bc:9f:5c:21:29',
                                                                       'name': '236110/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '12702597764202',
                                                                       'out_ucast_pkts': '9081113400',
                                                                       'poller_hash': 'MTI3NEFBNXBlZXJpbnRmMjM2MTEwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236110',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '12702597764202',
                                                                       'in_ucast_pkts': '9081113400',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '12702597764202',
                                                                                                'in_ucast_pkts': '9081113400',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '758793497519',
                                                                                                'out_ucast_pkts': '3117822550',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '758793497519',
                                                                       'out_ucast_pkts': '3117822550',
                                                                       'poller_hash': 'MTI3NEFBNXBlZXJpbnRmcG9uIDIzNjExMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236110',
                                                                                               'links': [       {       'connected_time': '688586',
                                                                                                                        'fiber_length': 12069,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFBNXBlZXJwb25wb24gMjM2MTEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6261 PERTH ROAD 121 (ROES SCOTT)',
                                         'poller_hash': 'MTI3NEFBNQ==',
                                         'serial_number': '1274AA5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:21:29',
                                         'uptime': '688586'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '236120/g1',
                                                                       'name': '236120/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEE3NXBlZXJpbnRmMjM2MTIwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236120',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEE3NXBlZXJpbnRmcG9uIDIzNjEyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236120',
                                                                                               'links': [       {       'connected_time': '67050083',
                                                                                                                        'fiber_length': 11631,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE3NXBlZXJwb25wb24gMjM2MTIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5325 ROAD 140 (BRODHAGEN ELVEY)',
                                         'poller_hash': 'MTI3NEE3NQ==',
                                         'serial_number': '1274A75',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:a9',
                                         'uptime': '67050083'},
                                 {       'counter_source': '236156/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236156/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '364242538236',
                                                                       'in_ucast_pkts': '275615170351',
                                                                       'mac_address': '04:bc:9f:5c:1c:91',
                                                                       'name': '236156/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '89',
                                                                       'out_octets': '3837196228998',
                                                                       'out_ucast_pkts': '2932106333',
                                                                       'poller_hash': 'MTI3NEExMnBlZXJpbnRmMjM2MTU2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236156',
                                                                       'in_nucast_pkts': '89',
                                                                       'in_octets': '3837196228998',
                                                                       'in_ucast_pkts': '2932106333',
                                                                       'links': [       {       'in_nucast_pkts': '89',
                                                                                                'in_octets': '3837196228998',
                                                                                                'in_ucast_pkts': '2932106333',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '364242538236',
                                                                                                'out_ucast_pkts': '275615170351',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '364242538236',
                                                                       'out_ucast_pkts': '275615170351',
                                                                       'poller_hash': 'MTI3NEExMnBlZXJpbnRmcG9uIDIzNjE1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236156',
                                                                                               'links': [       {       'connected_time': '688564',
                                                                                                                        'fiber_length': 9235,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEExMnBlZXJwb25wb24gMjM2MTU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6535 PERTH ROAD 121 (BRINSTON BRET)',
                                         'poller_hash': 'MTI3NEExMg==',
                                         'serial_number': '1274A12',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:91',
                                         'uptime': '688564'},
                                 {       'counter_source': '236166/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236166/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '351119340',
                                                                       'in_ucast_pkts': '2314962',
                                                                       'mac_address': '04:bc:9f:5c:1f:d1',
                                                                       'name': '236166/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '14668988486',
                                                                       'out_ucast_pkts': '60818091',
                                                                       'poller_hash': 'MTI3NEE3QXBlZXJpbnRmMjM2MTY2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236166',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '14668988486',
                                                                       'in_ucast_pkts': '60818091',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '14668988486',
                                                                                                'in_ucast_pkts': '60818091',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '351119340',
                                                                                                'out_ucast_pkts': '2314962',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '351119340',
                                                                       'out_ucast_pkts': '2314962',
                                                                       'poller_hash': 'MTI3NEE3QXBlZXJpbnRmcG9uIDIzNjE2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236166',
                                                                                               'links': [       {       'connected_time': '1047179',
                                                                                                                        'fiber_length': 9053,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE3QXBlZXJwb25wb24gMjM2MTY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6535 PERTH ROAD 121 (ROES NORMAN)',
                                         'poller_hash': 'MTI3NEE3QQ==',
                                         'serial_number': '1274A7A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:d1',
                                         'uptime': '1047179'},
                                 {       'counter_source': '236186/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236186/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '331096658015',
                                                                       'in_ucast_pkts': '30652761136',
                                                                       'mac_address': '04:bc:9f:5c:04:11',
                                                                       'name': '236186/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '3404041477012',
                                                                       'out_ucast_pkts': '6088578382',
                                                                       'poller_hash': 'MTI3NDcwMnBlZXJpbnRmMjM2MTg2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236186',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '3404041477012',
                                                                       'in_ucast_pkts': '6088578382',
                                                                       'links': [       {       'in_nucast_pkts': '142',
                                                                                                'in_octets': '3404041477012',
                                                                                                'in_ucast_pkts': '6088578382',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '331096658015',
                                                                                                'out_ucast_pkts': '30652761136',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '331096658015',
                                                                       'out_ucast_pkts': '30652761136',
                                                                       'poller_hash': 'MTI3NDcwMnBlZXJpbnRmcG9uIDIzNjE4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236186',
                                                                                               'links': [       {       'connected_time': '8579811',
                                                                                                                        'fiber_length': 18215,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDcwMnBlZXJwb25wb24gMjM2MTg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '124 WINSTANLEY ST (COOK TAMMY)',
                                         'poller_hash': 'MTI3NDcwMg==',
                                         'serial_number': '1274702',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:04:11',
                                         'uptime': '8579811'},
                                 {       'counter_source': '236206/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236206/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '110595665721',
                                                                       'in_ucast_pkts': '643524565',
                                                                       'mac_address': '04:bc:9f:5c:1e:21',
                                                                       'name': '236206/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '155',
                                                                       'out_octets': '3469789744542',
                                                                       'out_ucast_pkts': '2415004597',
                                                                       'poller_hash': 'MTI3NEE0NHBlZXJpbnRmMjM2MjA2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236206',
                                                                       'in_nucast_pkts': '155',
                                                                       'in_octets': '3469789744542',
                                                                       'in_ucast_pkts': '2415004597',
                                                                       'links': [       {       'in_nucast_pkts': '155',
                                                                                                'in_octets': '3469789744542',
                                                                                                'in_ucast_pkts': '2415004597',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '110595665721',
                                                                                                'out_ucast_pkts': '643524565',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '110595665721',
                                                                       'out_ucast_pkts': '643524565',
                                                                       'poller_hash': 'MTI3NEE0NHBlZXJpbnRmcG9uIDIzNjIwNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236206',
                                                                                               'links': [       {       'connected_time': '12635297',
                                                                                                                        'fiber_length': 18502,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE0NHBlZXJwb25wb24gMjM2MjA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '139 MADDISON ST E (WEIR TERESA & '
                                                 'ROBIN)',
                                         'poller_hash': 'MTI3NEE0NA==',
                                         'serial_number': '1274A44',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:21',
                                         'uptime': '12635297'},
                                 {       'counter_source': '236216/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236216/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3101767460',
                                                                       'in_ucast_pkts': '13845176',
                                                                       'mac_address': '04:bc:9f:5c:10:59',
                                                                       'name': '236216/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '51189734946',
                                                                       'out_ucast_pkts': '89346996',
                                                                       'poller_hash': 'MTI3NDg4QnBlZXJpbnRmMjM2MjE2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236216',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '51189734946',
                                                                       'in_ucast_pkts': '89346996',
                                                                       'links': [       {       'in_nucast_pkts': '145',
                                                                                                'in_octets': '51189734946',
                                                                                                'in_ucast_pkts': '89346996',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3101767460',
                                                                                                'out_ucast_pkts': '13845176',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3101767460',
                                                                       'out_ucast_pkts': '13845176',
                                                                       'poller_hash': 'MTI3NDg4QnBlZXJpbnRmcG9uIDIzNjIxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236216',
                                                                                               'links': [       {       'connected_time': '967000',
                                                                                                                        'fiber_length': 9831,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg4QnBlZXJwb25wb24gMjM2MjE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6476 PERTH ROAD 121 (ROPP WAYNE)',
                                         'poller_hash': 'MTI3NDg4Qg==',
                                         'serial_number': '127488B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:10:59',
                                         'uptime': '967000'},
                                 {       'counter_source': '236239/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236239/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '445278542493',
                                                                       'in_ucast_pkts': '79896118904',
                                                                       'mac_address': '04:bc:9f:5c:1d:99',
                                                                       'name': '236239/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '9003182060815',
                                                                       'out_ucast_pkts': '6767623004',
                                                                       'poller_hash': 'MTI3NEEzM3BlZXJpbnRmMjM2MjM5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236239',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '9003182060815',
                                                                       'in_ucast_pkts': '6767623004',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '9003182060815',
                                                                                                'in_ucast_pkts': '6767623004',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '445278542493',
                                                                                                'out_ucast_pkts': '79896118904',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '445278542493',
                                                                       'out_ucast_pkts': '79896118904',
                                                                       'poller_hash': 'MTI3NEEzM3BlZXJpbnRmcG9uIDIzNjIzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236239',
                                                                                               'links': [       {       'connected_time': '688595',
                                                                                                                        'fiber_length': 12923,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEzM3BlZXJwb25wb24gMjM2MjM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6176 PERTH ROAD 121 (ALBRECHT '
                                                 'SANDRA)',
                                         'poller_hash': 'MTI3NEEzMw==',
                                         'serial_number': '1274A33',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:99',
                                         'uptime': '688595'},
                                 {       'counter_source': '236259/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236259/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '237016063401',
                                                                       'in_ucast_pkts': '916568546',
                                                                       'mac_address': '04:bc:9f:5c:1c:c9',
                                                                       'name': '236259/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '116',
                                                                       'out_octets': '4768234449930',
                                                                       'out_ucast_pkts': '3943486606',
                                                                       'poller_hash': 'MTI3NEExOXBlZXJpbnRmMjM2MjU5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236259',
                                                                       'in_nucast_pkts': '116',
                                                                       'in_octets': '4768234449930',
                                                                       'in_ucast_pkts': '3943486606',
                                                                       'links': [       {       'in_nucast_pkts': '116',
                                                                                                'in_octets': '4768234449930',
                                                                                                'in_ucast_pkts': '3943486606',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '237016063401',
                                                                                                'out_ucast_pkts': '916568546',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '237016063401',
                                                                       'out_ucast_pkts': '916568546',
                                                                       'poller_hash': 'MTI3NEExOXBlZXJpbnRmcG9uIDIzNjI1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236259',
                                                                                               'links': [       {       'connected_time': '688593',
                                                                                                                        'fiber_length': 10655,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEExOXBlZXJwb25wb24gMjM2MjU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4094B LINE 61 (NEUFELD HELENA)',
                                         'poller_hash': 'MTI3NEExOQ==',
                                         'serial_number': '1274A19',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:c9',
                                         'uptime': '688593'},
                                 {       'counter_source': '236269/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236269/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '149211246104',
                                                                       'in_ucast_pkts': '1155771524',
                                                                       'mac_address': '04:bc:9f:5c:19:39',
                                                                       'name': '236269/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '140',
                                                                       'out_octets': '6924472655554',
                                                                       'out_ucast_pkts': '4885641821',
                                                                       'poller_hash': 'MTI3NDlBN3BlZXJpbnRmMjM2MjY5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236269',
                                                                       'in_nucast_pkts': '140',
                                                                       'in_octets': '6924472655554',
                                                                       'in_ucast_pkts': '4885641821',
                                                                       'links': [       {       'in_nucast_pkts': '140',
                                                                                                'in_octets': '6924472655554',
                                                                                                'in_ucast_pkts': '4885641821',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '149211246104',
                                                                                                'out_ucast_pkts': '1155771524',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '149211246104',
                                                                       'out_ucast_pkts': '1155771524',
                                                                       'poller_hash': 'MTI3NDlBN3BlZXJpbnRmcG9uIDIzNjI2OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236269',
                                                                                               'links': [       {       'connected_time': '950833',
                                                                                                                        'fiber_length': 10664,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlBN3BlZXJwb25wb24gMjM2MjY5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6404 PERTH ROAD 121 (ROTH IRENE I)',
                                         'poller_hash': 'MTI3NDlBNw==',
                                         'serial_number': '12749A7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:19:39',
                                         'uptime': '950833'},
                                 {       'counter_source': '236861/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '236861/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7808792866',
                                                                       'in_ucast_pkts': '31285328',
                                                                       'mac_address': '04:bc:9f:5c:23:69',
                                                                       'name': '236861/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '109500968392',
                                                                       'out_ucast_pkts': '136243748',
                                                                       'poller_hash': 'MTI3NEFFRHBlZXJpbnRmMjM2ODYxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '236861',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '109500968392',
                                                                       'in_ucast_pkts': '136243748',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '109500968392',
                                                                                                'in_ucast_pkts': '136243748',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '7808792866',
                                                                                                'out_ucast_pkts': '31285328',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7808792866',
                                                                       'out_ucast_pkts': '31285328',
                                                                       'poller_hash': 'MTI3NEFFRHBlZXJpbnRmcG9uIDIzNjg2MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236861',
                                                                                               'links': [       {       'connected_time': '73710229',
                                                                                                                        'fiber_length': 8056,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFFRHBlZXJwb25wb24gMjM2ODYx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6720 PERTH ROAD 121 (KUEPFER KEN)',
                                         'poller_hash': 'MTI3NEFFRA==',
                                         'serial_number': '1274AED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:23:69',
                                         'uptime': '73710229'},
                                 {       'counter_source': '237733/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '237733/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '99659332619',
                                                                       'in_ucast_pkts': '152334527',
                                                                       'mac_address': '04:bc:9f:5c:24:a1',
                                                                       'name': '237733/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '89',
                                                                       'out_octets': '175903479987',
                                                                       'out_ucast_pkts': '265648659',
                                                                       'poller_hash': 'MTI3NEIxNHBlZXJpbnRmMjM3NzMzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '237733',
                                                                       'in_nucast_pkts': '89',
                                                                       'in_octets': '175903479987',
                                                                       'in_ucast_pkts': '265648659',
                                                                       'links': [       {       'in_nucast_pkts': '89',
                                                                                                'in_octets': '175903479987',
                                                                                                'in_ucast_pkts': '265648659',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '99659332619',
                                                                                                'out_ucast_pkts': '152334527',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '99659332619',
                                                                       'out_ucast_pkts': '152334527',
                                                                       'poller_hash': 'MTI3NEIxNHBlZXJpbnRmcG9uIDIzNzczMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '237733',
                                                                                               'links': [       {       'connected_time': '4788781',
                                                                                                                        'fiber_length': 9237,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIxNHBlZXJwb25wb24gMjM3NzMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6551 PERTH ROAD 121 (MARTIN MERVIN M '
                                                 'J)',
                                         'poller_hash': 'MTI3NEIxNA==',
                                         'serial_number': '1274B14',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:a1',
                                         'uptime': '4788781'},
                                 {       'counter_source': '237823/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '237823/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '28909230837',
                                                                       'in_ucast_pkts': '61555395',
                                                                       'mac_address': '04:bc:9f:5c:1e:81',
                                                                       'name': '237823/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '269997713211',
                                                                       'out_ucast_pkts': '266305459',
                                                                       'poller_hash': 'MTI3NEE1MHBlZXJpbnRmMjM3ODIzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '237823',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '269997713211',
                                                                       'in_ucast_pkts': '266305459',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '269997713211',
                                                                                                'in_ucast_pkts': '266305459',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '28909230837',
                                                                                                'out_ucast_pkts': '61555395',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '28909230837',
                                                                       'out_ucast_pkts': '61555395',
                                                                       'poller_hash': 'MTI3NEE1MHBlZXJpbnRmcG9uIDIzNzgyMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '237823',
                                                                                               'links': [       {       'connected_time': '90689',
                                                                                                                        'fiber_length': 9071,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1MHBlZXJwb25wb24gMjM3ODIz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6566 PERTH ROAD 121 (ZEHR BRYAN & '
                                                 'HEIDI)',
                                         'poller_hash': 'MTI3NEE1MA==',
                                         'serial_number': '1274A50',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:81',
                                         'uptime': '90689'},
                                 {       'counter_source': '238391/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238391/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '129989411457',
                                                                       'in_ucast_pkts': '43140135724',
                                                                       'mac_address': '04:bc:9f:5c:1c:59',
                                                                       'name': '238391/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '316160260398',
                                                                       'out_ucast_pkts': '346450292',
                                                                       'poller_hash': 'MTI3NEEwQnBlZXJpbnRmMjM4MzkxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238391',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '316160260398',
                                                                       'in_ucast_pkts': '346450292',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '316160260398',
                                                                                                'in_ucast_pkts': '346450292',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '129989411457',
                                                                                                'out_ucast_pkts': '43140135724',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '129989411457',
                                                                       'out_ucast_pkts': '43140135724',
                                                                       'poller_hash': 'MTI3NEEwQnBlZXJpbnRmcG9uIDIzODM5MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238391',
                                                                                               'links': [       {       'connected_time': '74404914',
                                                                                                                        'fiber_length': 8499,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwQnBlZXJwb25wb24gMjM4Mzkx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6765 PERTH ROAD 121 (ADAMS STEVE)',
                                         'poller_hash': 'MTI3NEEwQg==',
                                         'serial_number': '1274A0B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:59',
                                         'uptime': '74404914'},
                                 {       'counter_source': '238411/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238411/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '153185200453',
                                                                       'in_ucast_pkts': '177058364',
                                                                       'mac_address': '04:bc:9f:5c:25:01',
                                                                       'name': '238411/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '140',
                                                                       'out_octets': '31674399988',
                                                                       'out_ucast_pkts': '110683356',
                                                                       'poller_hash': 'MTI3NEIyMHBlZXJpbnRmMjM4NDExL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238411',
                                                                       'in_nucast_pkts': '140',
                                                                       'in_octets': '31674399988',
                                                                       'in_ucast_pkts': '110683356',
                                                                       'links': [       {       'in_nucast_pkts': '140',
                                                                                                'in_octets': '31674399988',
                                                                                                'in_ucast_pkts': '110683356',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '153185200453',
                                                                                                'out_ucast_pkts': '177058364',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '153185200453',
                                                                       'out_ucast_pkts': '177058364',
                                                                       'poller_hash': 'MTI3NEIyMHBlZXJpbnRmcG9uIDIzODQxMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238411',
                                                                                               'links': [       {       'connected_time': '74394732',
                                                                                                                        'fiber_length': 12100,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIyMHBlZXJwb25wb24gMjM4NDEx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4230 LINE 61 (SCHWARTZENTRUBER '
                                                 'JOYCE)',
                                         'poller_hash': 'MTI3NEIyMA==',
                                         'serial_number': '1274B20',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:25:01',
                                         'uptime': '74394732'},
                                 {       'counter_source': '238461/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238461/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '115844164069',
                                                                       'in_ucast_pkts': '872672990',
                                                                       'mac_address': '04:bc:9f:5c:20:b1',
                                                                       'name': '238461/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '127',
                                                                       'out_octets': '4344781443973',
                                                                       'out_ucast_pkts': '3142006588',
                                                                       'poller_hash': 'MTI3NEE5NnBlZXJpbnRmMjM4NDYxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238461',
                                                                       'in_nucast_pkts': '127',
                                                                       'in_octets': '4344781443973',
                                                                       'in_ucast_pkts': '3142006588',
                                                                       'links': [       {       'in_nucast_pkts': '127',
                                                                                                'in_octets': '4344781443973',
                                                                                                'in_ucast_pkts': '3142006588',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '115844164069',
                                                                                                'out_ucast_pkts': '872672990',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '115844164069',
                                                                       'out_ucast_pkts': '872672990',
                                                                       'poller_hash': 'MTI3NEE5NnBlZXJpbnRmcG9uIDIzODQ2MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238461',
                                                                                               'links': [       {       'connected_time': '73895475',
                                                                                                                        'fiber_length': 5681,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE5NnBlZXJwb25wb24gMjM4NDYx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6886 ROAD 129 (COOK WILLIAM & JEAN)',
                                         'poller_hash': 'MTI3NEE5Ng==',
                                         'serial_number': '1274A96',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:b1',
                                         'uptime': '73895475'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '238481/g1',
                                                                       'name': '238481/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEEwN3BlZXJpbnRmMjM4NDgxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238481',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEEwN3BlZXJpbnRmcG9uIDIzODQ4MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238481',
                                                                                               'links': [       {       'connected_time': '73882299',
                                                                                                                        'fiber_length': 8038,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwN3BlZXJwb25wb24gMjM4NDgx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6653 PERTH ROAD 121 (KUEPFER MANUEL '
                                                 'A & RUTH ANNE)',
                                         'poller_hash': 'MTI3NEEwNw==',
                                         'serial_number': '1274A07',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:39',
                                         'uptime': '73882299'},
                                 {       'counter_source': '238575/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238575/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '378087964503',
                                                                       'in_ucast_pkts': '9909425366',
                                                                       'mac_address': '04:bc:9f:5c:1e:e9',
                                                                       'name': '238575/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '4304695168170',
                                                                       'out_ucast_pkts': '3523191007',
                                                                       'poller_hash': 'MTI3NEE1RHBlZXJpbnRmMjM4NTc1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238575',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '4304695168170',
                                                                       'in_ucast_pkts': '3523191007',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '4304695168170',
                                                                                                'in_ucast_pkts': '3523191007',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '378087964503',
                                                                                                'out_ucast_pkts': '9909425366',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '378087964503',
                                                                       'out_ucast_pkts': '9909425366',
                                                                       'poller_hash': 'MTI3NEE1RHBlZXJpbnRmcG9uIDIzODU3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238575',
                                                                                               'links': [       {       'connected_time': '3113974',
                                                                                                                        'fiber_length': 14018,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE1RHBlZXJwb25wb24gMjM4NTc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5311 LINE 52 (ERNST ERIKA)',
                                         'poller_hash': 'MTI3NEE1RA==',
                                         'serial_number': '1274A5D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:e9',
                                         'uptime': '3113974'},
                                 {       'counter_source': '238629/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238629/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2683485248012',
                                                                       'in_ucast_pkts': '41987091382',
                                                                       'mac_address': '04:bc:9f:5c:22:11',
                                                                       'name': '238629/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '140',
                                                                       'out_octets': '6842788804754',
                                                                       'out_ucast_pkts': '6261696302',
                                                                       'poller_hash': 'MTI3NEFDMnBlZXJpbnRmMjM4NjI5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238629',
                                                                       'in_nucast_pkts': '140',
                                                                       'in_octets': '6842788804754',
                                                                       'in_ucast_pkts': '6261696302',
                                                                       'links': [       {       'in_nucast_pkts': '140',
                                                                                                'in_octets': '6842788804754',
                                                                                                'in_ucast_pkts': '6261696302',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2683485248012',
                                                                                                'out_ucast_pkts': '41987091382',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2683485248012',
                                                                       'out_ucast_pkts': '41987091382',
                                                                       'poller_hash': 'MTI3NEFDMnBlZXJpbnRmcG9uIDIzODYyOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238629',
                                                                                               'links': [       {       'connected_time': '19073558',
                                                                                                                        'fiber_length': 12721,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFDMnBlZXJwb25wb24gMjM4NjI5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6193 PERTH ROAD 121 (HYATT PAUL & '
                                                 'KIMBERLY)',
                                         'poller_hash': 'MTI3NEFDMg==',
                                         'serial_number': '1274AC2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:11',
                                         'uptime': '19073558'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '238639/g1',
                                                                       'name': '238639/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEIwMHBlZXJpbnRmMjM4NjM5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238639',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEIwMHBlZXJpbnRmcG9uIDIzODYzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238639',
                                                                                               'links': [       {       'connected_time': '73802908',
                                                                                                                        'fiber_length': 3790,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIwMHBlZXJwb25wb24gMjM4NjM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4435 LINE 67 (KUEPFER ALMEDA & '
                                                 'CLARE)',
                                         'poller_hash': 'MTI3NEIwMA==',
                                         'serial_number': '1274B00',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:01',
                                         'uptime': '73802908'},
                                 {       'counter_source': '238895/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238895/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '125292447451',
                                                                       'in_ucast_pkts': '424741308',
                                                                       'mac_address': '04:bc:9f:5c:20:f9',
                                                                       'name': '238895/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '1187161297641',
                                                                       'out_ucast_pkts': '1035964760',
                                                                       'poller_hash': 'MTI3NEE5RnBlZXJpbnRmMjM4ODk1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '238895',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '1187161297641',
                                                                       'in_ucast_pkts': '1035964760',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '1187161297641',
                                                                                                'in_ucast_pkts': '1035964760',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '125292447451',
                                                                                                'out_ucast_pkts': '424741308',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '125292447451',
                                                                       'out_ucast_pkts': '424741308',
                                                                       'poller_hash': 'MTI3NEE5RnBlZXJpbnRmcG9uIDIzODg5NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238895',
                                                                                               'links': [       {       'connected_time': '73793265',
                                                                                                                        'fiber_length': 8644,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE5RnBlZXJwb25wb24gMjM4ODk1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6602 PERTH ROAD 121 (HOOVER SAMUEL '
                                                 'B)',
                                         'poller_hash': 'MTI3NEE5Rg==',
                                         'serial_number': '1274A9F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:f9',
                                         'uptime': '73793265'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '239029/g1',
                                                                       'name': '239029/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDA2RHBlZXJpbnRmMjM5MDI5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239029',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDA2RHBlZXJpbnRmcG9uIDIzOTAyOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239029',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA2RHBlZXJwb25wb24gMjM5MDI5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '70 MAIN ST S (1000669244ONTARIO)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "239029" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-24T08:29:33.710Z',
                                         'poller_hash': 'MTI3NDA2RA==',
                                         'serial_number': '127406D',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '239039/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239039/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '33003576664',
                                                                       'in_ucast_pkts': '125593722',
                                                                       'mac_address': '04:bc:9f:5b:d9:e1',
                                                                       'name': '239039/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '120',
                                                                       'out_octets': '390041198654',
                                                                       'out_ucast_pkts': '379114954',
                                                                       'poller_hash': 'MTI3NDFCQ3BlZXJpbnRmMjM5MDM5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239039',
                                                                       'in_nucast_pkts': '120',
                                                                       'in_octets': '390041198654',
                                                                       'in_ucast_pkts': '379114954',
                                                                       'links': [       {       'in_nucast_pkts': '120',
                                                                                                'in_octets': '390041198654',
                                                                                                'in_ucast_pkts': '379114954',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '33003576664',
                                                                                                'out_ucast_pkts': '125593722',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '33003576664',
                                                                       'out_ucast_pkts': '125593722',
                                                                       'poller_hash': 'MTI3NDFCQ3BlZXJpbnRmcG9uIDIzOTAzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239039',
                                                                                               'links': [       {       'connected_time': '73707903',
                                                                                                                        'fiber_length': 11242,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFCQ3BlZXJwb25wb24gMjM5MDM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7033 PERTH ROAD 121 (SCHELLENBERGER '
                                                 'FRIEDA)',
                                         'poller_hash': 'MTI3NDFCQw==',
                                         'serial_number': '12741BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:e1',
                                         'uptime': '73707903'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '239201/g1',
                                                                       'name': '239201/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDFBMHBlZXJpbnRmMjM5MjAxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239201',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDFBMHBlZXJpbnRmcG9uIDIzOTIwMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239201',
                                                                                               'links': [       {       'connected_time': '73629940',
                                                                                                                        'fiber_length': 10661,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFBMHBlZXJwb25wb24gMjM5MjAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6390 PERTH ROAD 121 (BUECKERT HENRY '
                                                 '& TINA)',
                                         'poller_hash': 'MTI3NDFBMA==',
                                         'serial_number': '12741A0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:01',
                                         'uptime': '73629940'},
                                 {       'counter_source': '239211/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239211/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '512221432654',
                                                                       'in_ucast_pkts': '47895090303',
                                                                       'mac_address': '04:bc:9f:5b:d1:29',
                                                                       'name': '239211/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '122',
                                                                       'out_octets': '942758876128',
                                                                       'out_ucast_pkts': '959414086',
                                                                       'poller_hash': 'MTI3NDBBNXBlZXJpbnRmMjM5MjExL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239211',
                                                                       'in_nucast_pkts': '122',
                                                                       'in_octets': '942758876128',
                                                                       'in_ucast_pkts': '959414086',
                                                                       'links': [       {       'in_nucast_pkts': '122',
                                                                                                'in_octets': '942758876128',
                                                                                                'in_ucast_pkts': '959414086',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '512221432654',
                                                                                                'out_ucast_pkts': '47895090303',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '512221432654',
                                                                       'out_ucast_pkts': '47895090303',
                                                                       'poller_hash': 'MTI3NDBBNXBlZXJpbnRmcG9uIDIzOTIxMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239211',
                                                                                               'links': [       {       'connected_time': '73621652',
                                                                                                                        'fiber_length': 4728,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBBNXBlZXJwb25wb24gMjM5MjEx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4360 LINE 67 (MARTIN AARON F)',
                                         'poller_hash': 'MTI3NDBBNQ==',
                                         'serial_number': '12740A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d1:29',
                                         'uptime': '73621652'},
                                 {       'counter_source': '239221/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239221/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '456374040948',
                                                                       'in_ucast_pkts': '748945312',
                                                                       'mac_address': '04:bc:9f:5b:cd:69',
                                                                       'name': '239221/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '124',
                                                                       'out_octets': '2294618666961',
                                                                       'out_ucast_pkts': '1978601186',
                                                                       'poller_hash': 'MTI3NDAyRHBlZXJpbnRmMjM5MjIxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239221',
                                                                       'in_nucast_pkts': '124',
                                                                       'in_octets': '2294618666961',
                                                                       'in_ucast_pkts': '1978601186',
                                                                       'links': [       {       'in_nucast_pkts': '124',
                                                                                                'in_octets': '2294618666961',
                                                                                                'in_ucast_pkts': '1978601186',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '456374040948',
                                                                                                'out_ucast_pkts': '748945312',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '456374040948',
                                                                       'out_ucast_pkts': '748945312',
                                                                       'poller_hash': 'MTI3NDAyRHBlZXJpbnRmcG9uIDIzOTIyMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239221',
                                                                                               'links': [       {       'connected_time': '13672703',
                                                                                                                        'fiber_length': 6819,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDAyRHBlZXJwb25wb24gMjM5MjIx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4155 LINE 67 (ERB KENNETH & LAURIE)',
                                         'poller_hash': 'MTI3NDAyRA==',
                                         'serial_number': '127402D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cd:69',
                                         'uptime': '13672703'},
                                 {       'counter_source': '239345/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239345/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '48570454802',
                                                                       'in_ucast_pkts': '248966251',
                                                                       'mac_address': '04:bc:9f:5c:1a:19',
                                                                       'name': '239345/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '72',
                                                                       'out_octets': '1146208883088',
                                                                       'out_ucast_pkts': '860384380',
                                                                       'poller_hash': 'MTI3NDlDM3BlZXJpbnRmMjM5MzQ1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239345',
                                                                       'in_nucast_pkts': '72',
                                                                       'in_octets': '1146208883088',
                                                                       'in_ucast_pkts': '860384380',
                                                                       'links': [       {       'in_nucast_pkts': '72',
                                                                                                'in_octets': '1146208883088',
                                                                                                'in_ucast_pkts': '860384380',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '48570454802',
                                                                                                'out_ucast_pkts': '248966251',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '48570454802',
                                                                       'out_ucast_pkts': '248966251',
                                                                       'poller_hash': 'MTI3NDlDM3BlZXJpbnRmcG9uIDIzOTM0NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239345',
                                                                                               'links': [       {       'connected_time': '70168457',
                                                                                                                        'fiber_length': 4793,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlDM3BlZXJwb25wb24gMjM5MzQ1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6799 ROAD 129 (JANTZI PERRY & '
                                                 'KRISTA)',
                                         'poller_hash': 'MTI3NDlDMw==',
                                         'serial_number': '12749C3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1a:19',
                                         'uptime': '70168457'},
                                 {       'counter_source': '239417/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239417/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96905352266',
                                                                       'in_ucast_pkts': '308757125',
                                                                       'mac_address': '04:bc:9f:5c:1b:a1',
                                                                       'name': '239417/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '1035835631025',
                                                                       'out_ucast_pkts': '911606636',
                                                                       'poller_hash': 'MTI3NDlGNHBlZXJpbnRmMjM5NDE3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239417',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '1035835631025',
                                                                       'in_ucast_pkts': '911606636',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '1035835631025',
                                                                                                'in_ucast_pkts': '911606636',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96905352266',
                                                                                                'out_ucast_pkts': '308757125',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96905352266',
                                                                       'out_ucast_pkts': '308757125',
                                                                       'poller_hash': 'MTI3NDlGNHBlZXJpbnRmcG9uIDIzOTQxNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239417',
                                                                                               'links': [       {       'connected_time': '25920265',
                                                                                                                        'fiber_length': 9229,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlGNHBlZXJwb25wb24gMjM5NDE3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3948 LINE 67 (SCHMIDT JAMES)',
                                         'poller_hash': 'MTI3NDlGNA==',
                                         'serial_number': '12749F4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:a1',
                                         'uptime': '25920265'},
                                 {       'counter_source': '239550/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:69:6d'],
                                                                       'id': '239550/g1',
                                                                       'in_nucast_pkts': '11071',
                                                                       'in_octets': '1883842790',
                                                                       'in_ucast_pkts': '4351194',
                                                                       'mac_address': '04:bc:9f:5c:1a:0c',
                                                                       'mtu': 2000,
                                                                       'name': '239550/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1118803',
                                                                       'out_octets': '23892596312',
                                                                       'out_ucast_pkts': '19455282',
                                                                       'poller_hash': 'MTI3NDlDMXBlZXJpbnRmMjM5NTUwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '239550',
                                                                       'in_nucast_pkts': '1118803',
                                                                       'in_octets': '23892596312',
                                                                       'in_ucast_pkts': '19455282',
                                                                       'links': [       {       'in_nucast_pkts': '1118803',
                                                                                                'in_octets': '23892596312',
                                                                                                'in_ucast_pkts': '19455282',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11071',
                                                                                                'out_octets': '1883842790',
                                                                                                'out_ucast_pkts': '4351194',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11071',
                                                                       'out_octets': '1883842790',
                                                                       'out_ucast_pkts': '4351194',
                                                                       'poller_hash': 'MTI3NDlDMXBlZXJpbnRmcG9uIDIzOTU1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239550',
                                                                                               'links': [       {       'connected_time': '523335',
                                                                                                                        'fiber_length': 12154,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlDMXBlZXJwb25wb24gMjM5NTUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7116 PERTH ROAD 121 (KUEPFER GARY & '
                                                 'MARY LOU)',
                                         'poller_hash': 'MTI3NDlDMQ==',
                                         'serial_number': '12749C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1a:09',
                                         'uptime': '523335'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '239667/g1',
                                                                       'name': '239667/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEIxRHBlZXJpbnRmMjM5NjY3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239667',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEIxRHBlZXJpbnRmcG9uIDIzOTY2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239667',
                                                                                               'links': [       {       'connected_time': '73271910',
                                                                                                                        'fiber_length': 5305,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIxRHBlZXJwb25wb24gMjM5NjY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4287 LINE 67 (KUEPFER CALVIN)',
                                         'poller_hash': 'MTI3NEIxRA==',
                                         'serial_number': '1274B1D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:e9',
                                         'uptime': '73271910'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '239751/g1',
                                                                       'name': '239751/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDlDNXBlZXJpbnRmMjM5NzUxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239751',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDlDNXBlZXJpbnRmcG9uIDIzOTc1MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239751',
                                                                                               'links': [       {       'connected_time': '73201399',
                                                                                                                        'fiber_length': 12150,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlDNXBlZXJwb25wb24gMjM5NzUx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7112 PERTH ROAD 121',
                                         'poller_hash': 'MTI3NDlDNQ==',
                                         'serial_number': '12749C5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1a:29',
                                         'uptime': '73201399'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '239761/g1',
                                                                       'name': '239761/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDlFMXBlZXJpbnRmMjM5NzYxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239761',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDlFMXBlZXJpbnRmcG9uIDIzOTc2MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239761',
                                                                                               'links': [       {       'connected_time': '68244312',
                                                                                                                        'fiber_length': 9139,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlFMXBlZXJwb25wb24gMjM5NzYx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6825 PERTH ROAD 121 (KUEPFER DAVID & '
                                                 'ANNA MAE)',
                                         'poller_hash': 'MTI3NDlFMQ==',
                                         'serial_number': '12749E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:09',
                                         'uptime': '68244312'},
                                 {       'counter_source': '239844/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239844/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '85300034263',
                                                                       'in_ucast_pkts': '256652421',
                                                                       'mac_address': '04:bc:9f:5c:22:f9',
                                                                       'name': '239844/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '1089673798592',
                                                                       'out_ucast_pkts': '893778226',
                                                                       'poller_hash': 'MTI3NEFERnBlZXJpbnRmMjM5ODQ0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239844',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '1089673798592',
                                                                       'in_ucast_pkts': '893778226',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '1089673798592',
                                                                                                'in_ucast_pkts': '893778226',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '85300034263',
                                                                                                'out_ucast_pkts': '256652421',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '85300034263',
                                                                       'out_ucast_pkts': '256652421',
                                                                       'poller_hash': 'MTI3NEFERnBlZXJpbnRmcG9uIDIzOTg0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239844',
                                                                                               'links': [       {       'connected_time': '688582',
                                                                                                                        'fiber_length': 4054,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFERnBlZXJwb25wb24gMjM5ODQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4434 LINE 67 (BANCROFT PETER)',
                                         'poller_hash': 'MTI3NEFERg==',
                                         'serial_number': '1274ADF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:f9',
                                         'uptime': '688582'},
                                 {       'counter_source': '239854/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239854/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '61449529238',
                                                                       'in_ucast_pkts': '541601342',
                                                                       'mac_address': '04:bc:9f:5c:24:81',
                                                                       'name': '239854/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '146',
                                                                       'out_octets': '3222822200898',
                                                                       'out_ucast_pkts': '2199588349',
                                                                       'poller_hash': 'MTI3NEIxMHBlZXJpbnRmMjM5ODU0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239854',
                                                                       'in_nucast_pkts': '146',
                                                                       'in_octets': '3222822200898',
                                                                       'in_ucast_pkts': '2199588349',
                                                                       'links': [       {       'in_nucast_pkts': '146',
                                                                                                'in_octets': '3222822200898',
                                                                                                'in_ucast_pkts': '2199588349',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '61449529238',
                                                                                                'out_ucast_pkts': '541601342',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '61449529238',
                                                                       'out_ucast_pkts': '541601342',
                                                                       'poller_hash': 'MTI3NEIxMHBlZXJpbnRmcG9uIDIzOTg1NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239854',
                                                                                               'links': [       {       'connected_time': '73104368',
                                                                                                                        'fiber_length': 1400,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIxMHBlZXJwb25wb24gMjM5ODU0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '54 PUGH ST UNIT 4 (GERBER BUD)',
                                         'poller_hash': 'MTI3NEIxMA==',
                                         'serial_number': '1274B10',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:81',
                                         'uptime': '73104368'},
                                 {       'counter_source': '239927/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:ab:f5'],
                                                                       'id': '239927/g2',
                                                                       'in_nucast_pkts': '140549',
                                                                       'in_octets': '53681278249',
                                                                       'in_ucast_pkts': '136208253',
                                                                       'mac_address': '04:bc:9f:5b:e2:cd',
                                                                       'mtu': 2000,
                                                                       'name': '239927/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16786562',
                                                                       'out_octets': '361804620016',
                                                                       'out_ucast_pkts': '320138669',
                                                                       'poller_hash': 'MTI3NDJEOXBlZXJpbnRmMjM5OTI3L2cy',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '239927',
                                                                       'in_nucast_pkts': '16786562',
                                                                       'in_octets': '361804620016',
                                                                       'in_ucast_pkts': '320138669',
                                                                       'links': [       {       'in_nucast_pkts': '16786562',
                                                                                                'in_octets': '361804620016',
                                                                                                'in_ucast_pkts': '320138669',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '140549',
                                                                                                'out_octets': '53681278249',
                                                                                                'out_ucast_pkts': '136208253',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '140549',
                                                                       'out_octets': '53681278249',
                                                                       'out_ucast_pkts': '136208253',
                                                                       'poller_hash': 'MTI3NDJEOXBlZXJpbnRmcG9uIDIzOTkyNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239927',
                                                                                               'links': [       {       'connected_time': '7187005',
                                                                                                                        'fiber_length': 12539,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJEOXBlZXJwb25wb24gMjM5OTI3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4111 LINE 73 (MORALES HUGO)',
                                         'poller_hash': 'MTI3NDJEOQ==',
                                         'serial_number': '12742D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e2:c9',
                                         'uptime': '7187005'},
                                 {       'counter_source': '239947/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239947/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '618487218507',
                                                                       'in_ucast_pkts': '61021579311',
                                                                       'mac_address': '04:bc:9f:5b:e4:b9',
                                                                       'name': '239947/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '2571588353317',
                                                                       'out_ucast_pkts': '2269650012',
                                                                       'poller_hash': 'MTI3NDMxN3BlZXJpbnRmMjM5OTQ3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239947',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '2571588353317',
                                                                       'in_ucast_pkts': '2269650012',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '2571588353317',
                                                                                                'in_ucast_pkts': '2269650012',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '618487218507',
                                                                                                'out_ucast_pkts': '61021579311',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '618487218507',
                                                                       'out_ucast_pkts': '61021579311',
                                                                       'poller_hash': 'MTI3NDMxN3BlZXJpbnRmcG9uIDIzOTk0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239947',
                                                                                               'links': [       {       'connected_time': '688595',
                                                                                                                        'fiber_length': 14764,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMxN3BlZXJwb25wb24gMjM5OTQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7374 PERTH ROAD 121 (BAUMAN DEAN)',
                                         'poller_hash': 'MTI3NDMxNw==',
                                         'serial_number': '1274317',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e4:b9',
                                         'uptime': '688595'},
                                 {       'counter_source': '239957/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '239957/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '198953239263',
                                                                       'in_ucast_pkts': '4579553588',
                                                                       'mac_address': '04:bc:9f:5b:e3:21',
                                                                       'name': '239957/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '530802178326',
                                                                       'out_ucast_pkts': '548698936',
                                                                       'poller_hash': 'MTI3NDJFNHBlZXJpbnRmMjM5OTU3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '239957',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '530802178326',
                                                                       'in_ucast_pkts': '548698936',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '530802178326',
                                                                                                'in_ucast_pkts': '548698936',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '198953239263',
                                                                                                'out_ucast_pkts': '4579553588',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '198953239263',
                                                                       'out_ucast_pkts': '4579553588',
                                                                       'poller_hash': 'MTI3NDJFNHBlZXJpbnRmcG9uIDIzOTk1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239957',
                                                                                               'links': [       {       'connected_time': '1845110',
                                                                                                                        'fiber_length': 13891,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJFNHBlZXJwb25wb24gMjM5OTU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7272 PERTH ROAD 121 (MARTIN '
                                                 'BENJAMIN)',
                                         'poller_hash': 'MTI3NDJFNA==',
                                         'serial_number': '12742E4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:21',
                                         'uptime': '1845110'},
                                 {       'counter_source': '240191/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '240191/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '135247222800',
                                                                       'in_ucast_pkts': '344841244',
                                                                       'mac_address': '04:bc:9f:5b:d1:a1',
                                                                       'name': '240191/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '1152804177711',
                                                                       'out_ucast_pkts': '964985480',
                                                                       'poller_hash': 'MTI3NDBCNHBlZXJpbnRmMjQwMTkxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240191',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '1152804177711',
                                                                       'in_ucast_pkts': '964985480',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '1152804177711',
                                                                                                'in_ucast_pkts': '964985480',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '135247222800',
                                                                                                'out_ucast_pkts': '344841244',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '135247222800',
                                                                       'out_ucast_pkts': '344841244',
                                                                       'poller_hash': 'MTI3NDBCNHBlZXJpbnRmcG9uIDI0MDE5MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240191',
                                                                                               'links': [       {       'connected_time': '688587',
                                                                                                                        'fiber_length': 11862,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBCNHBlZXJwb25wb24gMjQwMTkx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4224 LINE 61 (CARTER ALVIN)',
                                         'poller_hash': 'MTI3NDBCNA==',
                                         'serial_number': '12740B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d1:a1',
                                         'uptime': '688587'},
                                 {       'counter_source': '240201/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '240201/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '130573504830',
                                                                       'in_ucast_pkts': '38949640492',
                                                                       'mac_address': '04:bc:9f:5b:e4:21',
                                                                       'name': '240201/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '116',
                                                                       'out_octets': '987543452225',
                                                                       'out_ucast_pkts': '830701433',
                                                                       'poller_hash': 'MTI3NDMwNHBlZXJpbnRmMjQwMjAxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240201',
                                                                       'in_nucast_pkts': '116',
                                                                       'in_octets': '987543452225',
                                                                       'in_ucast_pkts': '830701433',
                                                                       'links': [       {       'in_nucast_pkts': '116',
                                                                                                'in_octets': '987543452225',
                                                                                                'in_ucast_pkts': '830701433',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '130573504830',
                                                                                                'out_ucast_pkts': '38949640492',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '130573504830',
                                                                       'out_ucast_pkts': '38949640492',
                                                                       'poller_hash': 'MTI3NDMwNHBlZXJpbnRmcG9uIDI0MDIwMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240201',
                                                                                               'links': [       {       'connected_time': '688591',
                                                                                                                        'fiber_length': 12362,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMwNHBlZXJwb25wb24gMjQwMjAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4229 LINE 61 (ZEHR SAWYER)',
                                         'poller_hash': 'MTI3NDMwNA==',
                                         'serial_number': '1274304',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e4:21',
                                         'uptime': '688591'},
                                 {       'counter_source': '240221/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '240221/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '56336145810',
                                                                       'in_ucast_pkts': '287160654',
                                                                       'mac_address': '04:bc:9f:5b:d1:f9',
                                                                       'name': '240221/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '1551965307608',
                                                                       'out_ucast_pkts': '1230438874',
                                                                       'poller_hash': 'MTI3NDBCRnBlZXJpbnRmMjQwMjIxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240221',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '1551965307608',
                                                                       'in_ucast_pkts': '1230438874',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '1551965307608',
                                                                                                'in_ucast_pkts': '1230438874',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '56336145810',
                                                                                                'out_ucast_pkts': '287160654',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '56336145810',
                                                                       'out_ucast_pkts': '287160654',
                                                                       'poller_hash': 'MTI3NDBCRnBlZXJpbnRmcG9uIDI0MDIyMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240221',
                                                                                               'links': [       {       'connected_time': '688575',
                                                                                                                        'fiber_length': 3679,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBCRnBlZXJwb25wb24gMjQwMjIx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6672 ROAD 129 (KUEPFER PAUL)',
                                         'poller_hash': 'MTI3NDBCRg==',
                                         'serial_number': '12740BF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d1:f9',
                                         'uptime': '688575'},
                                 {       'counter_source': '240241/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '240241/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '46170742',
                                                                       'in_ucast_pkts': '73938',
                                                                       'mac_address': '04:bc:9f:5b:d9:d9',
                                                                       'name': '240241/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '146',
                                                                       'out_octets': '12549329139',
                                                                       'out_ucast_pkts': '53317212',
                                                                       'poller_hash': 'MTI3NDFCQnBlZXJpbnRmMjQwMjQxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240241',
                                                                       'in_nucast_pkts': '146',
                                                                       'in_octets': '12549329139',
                                                                       'in_ucast_pkts': '53317212',
                                                                       'links': [       {       'in_nucast_pkts': '146',
                                                                                                'in_octets': '12549329139',
                                                                                                'in_ucast_pkts': '53317212',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '46170742',
                                                                                                'out_ucast_pkts': '73938',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '46170742',
                                                                       'out_ucast_pkts': '73938',
                                                                       'poller_hash': 'MTI3NDFCQnBlZXJpbnRmcG9uIDI0MDI0MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240241',
                                                                                               'links': [       {       'connected_time': '72680560',
                                                                                                                        'fiber_length': 14932,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFCQnBlZXJwb25wb24gMjQwMjQx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7401 PERTH ROAD 121 (SAUDER MERVIN)',
                                         'poller_hash': 'MTI3NDFCQg==',
                                         'serial_number': '12741BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:d9',
                                         'uptime': '72680560'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '240311/g1',
                                                                       'name': '240311/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDFCRnBlZXJpbnRmMjQwMzExL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240311',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDFCRnBlZXJpbnRmcG9uIDI0MDMxMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240311',
                                                                                               'links': [       {       'connected_time': '72668166',
                                                                                                                        'fiber_length': 7231,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFCRnBlZXJwb25wb24gMjQwMzEx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4410 LINE 61 (KUEPFER VERNON & '
                                                 'DARLENE)',
                                         'poller_hash': 'MTI3NDFCRg==',
                                         'serial_number': '12741BF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:f9',
                                         'uptime': '72668166'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '240538/g1',
                                                                       'name': '240538/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDJGNnBlZXJpbnRmMjQwNTM4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240538',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDJGNnBlZXJpbnRmcG9uIDI0MDUzOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240538',
                                                                                               'links': [       {       'connected_time': '72602221',
                                                                                                                        'fiber_length': 15489,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJGNnBlZXJwb25wb24gMjQwNTM4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7452 PERTH ROAD 121 (FERNBANK '
                                                 'PAROCHIAL SCHOOL)',
                                         'poller_hash': 'MTI3NDJGNg==',
                                         'serial_number': '12742F6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:b1',
                                         'uptime': '72602221'},
                                 {       'counter_source': '240548/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1c:d5'],
                                                                       'id': '240548/g1',
                                                                       'in_nucast_pkts': '14773',
                                                                       'in_octets': '13117106314',
                                                                       'in_ucast_pkts': '16911876',
                                                                       'mac_address': '04:bc:9f:5b:d1:04',
                                                                       'mtu': 2000,
                                                                       'name': '240548/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1435040',
                                                                       'out_octets': '47844823053',
                                                                       'out_ucast_pkts': '40474799',
                                                                       'poller_hash': 'MTI3NDBBMHBlZXJpbnRmMjQwNTQ4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '240548',
                                                                       'in_nucast_pkts': '1435040',
                                                                       'in_octets': '47844823053',
                                                                       'in_ucast_pkts': '40474799',
                                                                       'links': [       {       'in_nucast_pkts': '1435040',
                                                                                                'in_octets': '47844823053',
                                                                                                'in_ucast_pkts': '40474799',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14773',
                                                                                                'out_octets': '13117106314',
                                                                                                'out_ucast_pkts': '16911876',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14773',
                                                                       'out_octets': '13117106314',
                                                                       'out_ucast_pkts': '16911876',
                                                                       'poller_hash': 'MTI3NDBBMHBlZXJpbnRmcG9uIDI0MDU0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240548',
                                                                                               'links': [       {       'connected_time': '688561',
                                                                                                                        'fiber_length': 1812,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBBMHBlZXJwb25wb24gMjQwNTQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (IONCE ANCA)',
                                         'poller_hash': 'MTI3NDBBMA==',
                                         'serial_number': '12740A0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d1:01',
                                         'uptime': '688561'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '240571/g1',
                                                                       'name': '240571/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDI5QnBlZXJpbnRmMjQwNTcxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240571',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDI5QnBlZXJpbnRmcG9uIDI0MDU3MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240571',
                                                                                               'links': [       {       'connected_time': '72597084',
                                                                                                                        'fiber_length': 13062,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDI5QnBlZXJwb25wb24gMjQwNTcx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7212 PERTH ROAD 121 (BURKHART JAMES '
                                                 'M)',
                                         'poller_hash': 'MTI3NDI5Qg==',
                                         'serial_number': '127429B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e0:d9',
                                         'uptime': '72597084'},
                                 {       'counter_source': '240581/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '240581/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '64422106433',
                                                                       'in_ucast_pkts': '412556170',
                                                                       'mac_address': '04:bc:9f:5b:d9:81',
                                                                       'name': '240581/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '146',
                                                                       'out_octets': '1895065204591',
                                                                       'out_ucast_pkts': '1408615954',
                                                                       'poller_hash': 'MTI3NDFCMHBlZXJpbnRmMjQwNTgxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240581',
                                                                       'in_nucast_pkts': '146',
                                                                       'in_octets': '1895065204591',
                                                                       'in_ucast_pkts': '1408615954',
                                                                       'links': [       {       'in_nucast_pkts': '146',
                                                                                                'in_octets': '1895065204591',
                                                                                                'in_ucast_pkts': '1408615954',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '64422106433',
                                                                                                'out_ucast_pkts': '412556170',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '64422106433',
                                                                       'out_ucast_pkts': '412556170',
                                                                       'poller_hash': 'MTI3NDFCMHBlZXJpbnRmcG9uIDI0MDU4MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240581',
                                                                                               'links': [       {       'connected_time': '72585550',
                                                                                                                        'fiber_length': 12002,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFCMHBlZXJwb25wb24gMjQwNTgx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4187 LINE 61 (ZEHR TIM & LILLIAN)',
                                         'poller_hash': 'MTI3NDFCMA==',
                                         'serial_number': '12741B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:81',
                                         'uptime': '72585550'},
                                 {       'counter_source': '240648/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '240648/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '126087362924',
                                                                       'in_ucast_pkts': '22039538596',
                                                                       'mac_address': '04:bc:9f:5b:e3:a9',
                                                                       'name': '240648/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '81',
                                                                       'out_octets': '2225236758916',
                                                                       'out_ucast_pkts': '1761305511',
                                                                       'poller_hash': 'MTI3NDJGNXBlZXJpbnRmMjQwNjQ4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '240648',
                                                                       'in_nucast_pkts': '81',
                                                                       'in_octets': '2225236758916',
                                                                       'in_ucast_pkts': '1761305511',
                                                                       'links': [       {       'in_nucast_pkts': '81',
                                                                                                'in_octets': '2225236758916',
                                                                                                'in_ucast_pkts': '1761305511',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '126087362924',
                                                                                                'out_ucast_pkts': '22039538596',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '126087362924',
                                                                       'out_ucast_pkts': '22039538596',
                                                                       'poller_hash': 'MTI3NDJGNXBlZXJpbnRmcG9uIDI0MDY0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '240648',
                                                                                               'links': [       {       'connected_time': '3869134',
                                                                                                                        'fiber_length': 11862,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJGNXBlZXJwb25wb24gMjQwNjQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7082 PERTH ROAD 121 (MORALES MAYDA)',
                                         'poller_hash': 'MTI3NDJGNQ==',
                                         'serial_number': '12742F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:a9',
                                         'uptime': '3869134'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '241045/g1',
                                                                       'name': '241045/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDMwQ3BlZXJpbnRmMjQxMDQ1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241045',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDMwQ3BlZXJpbnRmcG9uIDI0MTA0NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241045',
                                                                                               'links': [       {       'connected_time': '72510083',
                                                                                                                        'fiber_length': 11081,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMwQ3BlZXJwb25wb24gMjQxMDQ1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4144 LINE 61 (KUEPFER FLOYD)',
                                         'poller_hash': 'MTI3NDMwQw==',
                                         'serial_number': '127430C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e4:61',
                                         'uptime': '72510083'},
                                 {       'counter_source': '241376/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241376/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '893060494438',
                                                                       'in_ucast_pkts': '91797434016',
                                                                       'mac_address': '04:bc:9f:5b:da:81',
                                                                       'name': '241376/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '71',
                                                                       'out_octets': '4101658962597',
                                                                       'out_ucast_pkts': '3347386396',
                                                                       'poller_hash': 'MTI3NDFEMHBlZXJpbnRmMjQxMzc2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241376',
                                                                       'in_nucast_pkts': '71',
                                                                       'in_octets': '4101658962597',
                                                                       'in_ucast_pkts': '3347386396',
                                                                       'links': [       {       'in_nucast_pkts': '71',
                                                                                                'in_octets': '4101658962597',
                                                                                                'in_ucast_pkts': '3347386396',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '893060494438',
                                                                                                'out_ucast_pkts': '91797434016',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '893060494438',
                                                                       'out_ucast_pkts': '91797434016',
                                                                       'poller_hash': 'MTI3NDFEMHBlZXJpbnRmcG9uIDI0MTM3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241376',
                                                                                               'links': [       {       'connected_time': '2607375',
                                                                                                                        'fiber_length': 17111,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFEMHBlZXJwb25wb24gMjQxMzc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5346 PERTH ROAD 135 (SANDY PEAK '
                                                 'DAIRY LTD)',
                                         'poller_hash': 'MTI3NDFEMA==',
                                         'serial_number': '12741D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:da:81',
                                         'uptime': '2607375'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '241455/g1',
                                                                       'name': '241455/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDIyMHBlZXJpbnRmMjQxNDU1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241455',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDIyMHBlZXJpbnRmcG9uIDI0MTQ1NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241455',
                                                                                               'links': [       {       'connected_time': '72411759',
                                                                                                                        'fiber_length': 15536,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDIyMHBlZXJwb25wb24gMjQxNDU1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7446 PERTH ROAD 121 (BURKHART WAYNE)',
                                         'poller_hash': 'MTI3NDIyMA==',
                                         'serial_number': '1274220',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:dd:01',
                                         'uptime': '72411759'},
                                 {       'counter_source': '241465/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241465/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '241937932914',
                                                                       'in_ucast_pkts': '1479219736',
                                                                       'mac_address': '04:bc:9f:5b:e1:41',
                                                                       'name': '241465/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '30',
                                                                       'out_octets': '11199182335669',
                                                                       'out_ucast_pkts': '7754297207',
                                                                       'poller_hash': 'MTI3NDJBOHBlZXJpbnRmMjQxNDY1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241465',
                                                                       'in_nucast_pkts': '30',
                                                                       'in_octets': '11199182335669',
                                                                       'in_ucast_pkts': '7754297207',
                                                                       'links': [       {       'in_nucast_pkts': '30',
                                                                                                'in_octets': '11199182335669',
                                                                                                'in_ucast_pkts': '7754297207',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '241937932914',
                                                                                                'out_ucast_pkts': '1479219736',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '241937932914',
                                                                       'out_ucast_pkts': '1479219736',
                                                                       'poller_hash': 'MTI3NDJBOHBlZXJpbnRmcG9uIDI0MTQ2NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241465',
                                                                                               'links': [       {       'connected_time': '15575842',
                                                                                                                        'fiber_length': 17826,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJBOHBlZXJwb25wb24gMjQxNDY1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '155 WINSTANLEY ST MONKTON (SIMPSON '
                                                 'DENNIS)',
                                         'poller_hash': 'MTI3NDJBOA==',
                                         'serial_number': '12742A8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e1:41',
                                         'uptime': '15575842'},
                                 {       'counter_source': '241488/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241488/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96330232915',
                                                                       'in_ucast_pkts': '291580994',
                                                                       'mac_address': '04:bc:9f:5c:1c:41',
                                                                       'name': '241488/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '70',
                                                                       'out_octets': '1439789343741',
                                                                       'out_ucast_pkts': '1170818162',
                                                                       'poller_hash': 'MTI3NEEwOHBlZXJpbnRmMjQxNDg4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241488',
                                                                       'in_nucast_pkts': '70',
                                                                       'in_octets': '1439789343741',
                                                                       'in_ucast_pkts': '1170818162',
                                                                       'links': [       {       'in_nucast_pkts': '70',
                                                                                                'in_octets': '1439789343741',
                                                                                                'in_ucast_pkts': '1170818162',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96330232915',
                                                                                                'out_ucast_pkts': '291580994',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96330232915',
                                                                       'out_ucast_pkts': '291580994',
                                                                       'poller_hash': 'MTI3NEEwOHBlZXJpbnRmcG9uIDI0MTQ4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241488',
                                                                                               'links': [       {       'connected_time': '72168615',
                                                                                                                        'fiber_length': 12289,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwOHBlZXJwb25wb24gMjQxNDg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7120 PERTH ROAD 121 (ZEHR LAVERNE)',
                                         'poller_hash': 'MTI3NEEwOA==',
                                         'serial_number': '1274A08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:41',
                                         'uptime': '72168615'},
                                 {       'counter_source': '241584/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241584/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '47466347698',
                                                                       'in_ucast_pkts': '162539394',
                                                                       'mac_address': '04:bc:9f:5c:1c:89',
                                                                       'name': '241584/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '110',
                                                                       'out_octets': '575876330730',
                                                                       'out_ucast_pkts': '534309945',
                                                                       'poller_hash': 'MTI3NEExMXBlZXJpbnRmMjQxNTg0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241584',
                                                                       'in_nucast_pkts': '110',
                                                                       'in_octets': '575876330730',
                                                                       'in_ucast_pkts': '534309945',
                                                                       'links': [       {       'in_nucast_pkts': '110',
                                                                                                'in_octets': '575876330730',
                                                                                                'in_ucast_pkts': '534309945',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '47466347698',
                                                                                                'out_ucast_pkts': '162539394',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '47466347698',
                                                                       'out_ucast_pkts': '162539394',
                                                                       'poller_hash': 'MTI3NEExMXBlZXJpbnRmcG9uIDI0MTU4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241584',
                                                                                               'links': [       {       'connected_time': '72160938',
                                                                                                                        'fiber_length': 12941,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEExMXBlZXJwb25wb24gMjQxNTg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4133 LINE 73 (MARTIN MELVIN & '
                                                 'SHARON)',
                                         'poller_hash': 'MTI3NEExMQ==',
                                         'serial_number': '1274A11',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:89',
                                         'uptime': '72160938'},
                                 {       'counter_source': '241624/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241624/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '131967895340',
                                                                       'in_ucast_pkts': '596119839',
                                                                       'mac_address': '04:bc:9f:5c:1b:e1',
                                                                       'name': '241624/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2316722914974',
                                                                       'out_ucast_pkts': '1663523147',
                                                                       'poller_hash': 'MTI3NDlGQ3BlZXJpbnRmMjQxNjI0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241624',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2316722914974',
                                                                       'in_ucast_pkts': '1663523147',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2316722914974',
                                                                                                'in_ucast_pkts': '1663523147',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '131967895340',
                                                                                                'out_ucast_pkts': '596119839',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '131967895340',
                                                                       'out_ucast_pkts': '596119839',
                                                                       'poller_hash': 'MTI3NDlGQ3BlZXJpbnRmcG9uIDI0MTYyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241624',
                                                                                               'links': [       {       'connected_time': '523140',
                                                                                                                        'fiber_length': 10910,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlGQ3BlZXJwb25wb24gMjQxNjI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6946B PERTH ROAD 121 (THIBIDEAU '
                                                 'KALEB)',
                                         'poller_hash': 'MTI3NDlGQw==',
                                         'serial_number': '12749FC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:e1',
                                         'uptime': '523140'},
                                 {       'counter_source': '241644/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1f:f8'],
                                                                       'id': '241644/g1',
                                                                       'in_nucast_pkts': '11693',
                                                                       'in_octets': '2671073402',
                                                                       'in_ucast_pkts': '9946455',
                                                                       'mac_address': '04:bc:9f:5b:dc:bc',
                                                                       'mtu': 2000,
                                                                       'name': '241644/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465228',
                                                                       'out_octets': '52464130038',
                                                                       'out_ucast_pkts': '42537550',
                                                                       'poller_hash': 'MTI3NDIxN3BlZXJpbnRmMjQxNjQ0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '241644',
                                                                       'in_nucast_pkts': '1465228',
                                                                       'in_octets': '52464130038',
                                                                       'in_ucast_pkts': '42537550',
                                                                       'links': [       {       'in_nucast_pkts': '1465228',
                                                                                                'in_octets': '52464130038',
                                                                                                'in_ucast_pkts': '42537550',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11693',
                                                                                                'out_octets': '2671073402',
                                                                                                'out_ucast_pkts': '9946455',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11693',
                                                                       'out_octets': '2671073402',
                                                                       'out_ucast_pkts': '9946455',
                                                                       'poller_hash': 'MTI3NDIxN3BlZXJpbnRmcG9uIDI0MTY0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241644',
                                                                                               'links': [       {       'connected_time': '688598',
                                                                                                                        'fiber_length': 10581,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDIxN3BlZXJwb25wb24gMjQxNjQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4093 LINE 61 (YODER LYNFORD)',
                                         'poller_hash': 'MTI3NDIxNw==',
                                         'serial_number': '1274217',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:dc:b9',
                                         'uptime': '688598'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '241684/g1',
                                                                       'name': '241684/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDFEMnBlZXJpbnRmMjQxNjg0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241684',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDFEMnBlZXJpbnRmcG9uIDI0MTY4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241684',
                                                                                               'links': [       {       'connected_time': '66371019',
                                                                                                                        'fiber_length': 5712,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFEMnBlZXJwb25wb24gMjQxNjg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4560 PERTH LINE 72 (KUEPFER LESTER)',
                                         'poller_hash': 'MTI3NDFEMg==',
                                         'serial_number': '12741D2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:da:91',
                                         'uptime': '66371019'},
                                 {       'counter_source': '241714/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241714/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '502385854099',
                                                                       'in_ucast_pkts': '31867660126',
                                                                       'mac_address': '04:bc:9f:5b:dc:a1',
                                                                       'name': '241714/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '110',
                                                                       'out_octets': '10838859387546',
                                                                       'out_ucast_pkts': '6163479683',
                                                                       'poller_hash': 'MTI3NDIxNHBlZXJpbnRmMjQxNzE0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241714',
                                                                       'in_nucast_pkts': '110',
                                                                       'in_octets': '10838859387546',
                                                                       'in_ucast_pkts': '6163479683',
                                                                       'links': [       {       'in_nucast_pkts': '110',
                                                                                                'in_octets': '10838859387546',
                                                                                                'in_ucast_pkts': '6163479683',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '502385854099',
                                                                                                'out_ucast_pkts': '31867660126',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '502385854099',
                                                                       'out_ucast_pkts': '31867660126',
                                                                       'poller_hash': 'MTI3NDIxNHBlZXJpbnRmcG9uIDI0MTcxNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241714',
                                                                                               'links': [       {       'connected_time': '71976643',
                                                                                                                        'fiber_length': 11323,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDIxNHBlZXJwb25wb24gMjQxNzE0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6345 PERTH ROAD 121 (VAN DEN TEMPEL '
                                                 'LAUREN)',
                                         'poller_hash': 'MTI3NDIxNA==',
                                         'serial_number': '1274214',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:dc:a1',
                                         'uptime': '71976643'},
                                 {       'counter_source': '241724/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241724/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '64049020388',
                                                                       'in_ucast_pkts': '4636480439',
                                                                       'mac_address': 'ec:4f:82:0d:68:95',
                                                                       'name': '241724/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1364415834006',
                                                                       'out_ucast_pkts': '1023611975',
                                                                       'poller_hash': 'MjA1REMwcGVlcmludGYyNDE3MjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '241724',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1364415834006',
                                                                       'in_ucast_pkts': '1023611975',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1364415834006',
                                                                                                'in_ucast_pkts': '1023611975',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '64049020388',
                                                                                                'out_ucast_pkts': '4636480439',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '64049020388',
                                                                       'out_ucast_pkts': '4636480439',
                                                                       'poller_hash': 'MjA1REMwcGVlcmludGZwb24gMjQxNzI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241724',
                                                                                               'links': [       {       'connected_time': '688357',
                                                                                                                        'fiber_length': 1131,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REMwcGVlcnBvbnBvbiAyNDE3MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '15 MAIN ST APT. 202 (GOETZ MATTHEW)',
                                         'poller_hash': 'MjA1REMw',
                                         'serial_number': '205DC0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:95',
                                         'uptime': '688357'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '241735/g1',
                                                                       'name': '241735/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEFDRnBlZXJpbnRmMjQxNzM1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241735',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEFDRnBlZXJpbnRmcG9uIDI0MTczNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241735',
                                                                                               'links': [       {       'connected_time': '64351275',
                                                                                                                        'fiber_length': 8901,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFDRnBlZXJwb25wb24gMjQxNzM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6800 PERTH ROAD 121 (KUEPFER WAYNE '
                                                 'S)',
                                         'poller_hash': 'MTI3NEFDRg==',
                                         'serial_number': '1274ACF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:79',
                                         'uptime': '64351275'},
                                 {       'counter_source': '241745/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241745/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '63448229131',
                                                                       'in_ucast_pkts': '206747080',
                                                                       'mac_address': '04:bc:9f:5b:ce:61',
                                                                       'name': '241745/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '70',
                                                                       'out_octets': '629853116456',
                                                                       'out_ucast_pkts': '587439378',
                                                                       'poller_hash': 'MTI3NDA0Q3BlZXJpbnRmMjQxNzQ1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241745',
                                                                       'in_nucast_pkts': '70',
                                                                       'in_octets': '629853116456',
                                                                       'in_ucast_pkts': '587439378',
                                                                       'links': [       {       'in_nucast_pkts': '70',
                                                                                                'in_octets': '629853116456',
                                                                                                'in_ucast_pkts': '587439378',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '63448229131',
                                                                                                'out_ucast_pkts': '206747080',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '63448229131',
                                                                       'out_ucast_pkts': '206747080',
                                                                       'poller_hash': 'MTI3NDA0Q3BlZXJpbnRmcG9uIDI0MTc0NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241745',
                                                                                               'links': [       {       'connected_time': '71903261',
                                                                                                                        'fiber_length': 11364,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA0Q3BlZXJwb25wb24gMjQxNzQ1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7034 PERTH ROAD 121 (BAUMAN MURRAY)',
                                         'poller_hash': 'MTI3NDA0Qw==',
                                         'serial_number': '127404C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:61',
                                         'uptime': '71903261'},
                                 {       'counter_source': '241755/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241755/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '398397587997',
                                                                       'in_ucast_pkts': '40407305686',
                                                                       'mac_address': '04:bc:9f:5b:d2:49',
                                                                       'name': '241755/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '60',
                                                                       'out_octets': '9438877424886',
                                                                       'out_ucast_pkts': '7149563306',
                                                                       'poller_hash': 'MTI3NDBDOXBlZXJpbnRmMjQxNzU1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241755',
                                                                       'in_nucast_pkts': '60',
                                                                       'in_octets': '9438877424886',
                                                                       'in_ucast_pkts': '7149563306',
                                                                       'links': [       {       'in_nucast_pkts': '60',
                                                                                                'in_octets': '9438877424886',
                                                                                                'in_ucast_pkts': '7149563306',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '398397587997',
                                                                                                'out_ucast_pkts': '40407305686',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '398397587997',
                                                                       'out_ucast_pkts': '40407305686',
                                                                       'poller_hash': 'MTI3NDBDOXBlZXJpbnRmcG9uIDI0MTc1NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241755',
                                                                                               'links': [       {       'connected_time': '688584',
                                                                                                                        'fiber_length': 11475,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBDOXBlZXJwb25wb24gMjQxNzU1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4192 LINE 61 (RAYNER WILLIAM)',
                                         'poller_hash': 'MTI3NDBDOQ==',
                                         'serial_number': '12740C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d2:49',
                                         'uptime': '688584'},
                                 {       'counter_source': '241775/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241775/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25271240325',
                                                                       'in_ucast_pkts': '110125705',
                                                                       'mac_address': '04:bc:9f:5b:ce:21',
                                                                       'name': '241775/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '448868144469',
                                                                       'out_ucast_pkts': '407833550',
                                                                       'poller_hash': 'MTI3NDA0NHBlZXJpbnRmMjQxNzc1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241775',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '448868144469',
                                                                       'in_ucast_pkts': '407833550',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '448868144469',
                                                                                                'in_ucast_pkts': '407833550',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '25271240325',
                                                                                                'out_ucast_pkts': '110125705',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25271240325',
                                                                       'out_ucast_pkts': '110125705',
                                                                       'poller_hash': 'MTI3NDA0NHBlZXJpbnRmcG9uIDI0MTc3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241775',
                                                                                               'links': [       {       'connected_time': '2249869',
                                                                                                                        'fiber_length': 8473,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA0NHBlZXJwb25wb24gMjQxNzc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4019 LINE 67 (KUEPFER JUDY)',
                                         'poller_hash': 'MTI3NDA0NA==',
                                         'serial_number': '1274044',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:21',
                                         'uptime': '2249869'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '241815/g1',
                                                                       'name': '241815/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3M0Y1NXBlZXJpbnRmMjQxODE1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241815',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3M0Y1NXBlZXJpbnRmcG9uIDI0MTgxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241815',
                                                                                               'links': [       {       'connected_time': '23692706',
                                                                                                                        'fiber_length': 14627,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0Y1NXBlZXJwb25wb24gMjQxODE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6238 ROAD 116 (NAFZIGER POULTRY INC)',
                                         'poller_hash': 'MTI3M0Y1NQ==',
                                         'serial_number': '1273F55',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:c6:a9',
                                         'uptime': '23692706'},
                                 {       'counter_source': '241845/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241845/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '70427988816',
                                                                       'in_ucast_pkts': '232344884',
                                                                       'mac_address': '04:bc:9f:5c:09:c9',
                                                                       'name': '241845/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '885601656615',
                                                                       'out_ucast_pkts': '756065912',
                                                                       'poller_hash': 'MTI3NDdCOXBlZXJpbnRmMjQxODQ1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241845',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '885601656615',
                                                                       'in_ucast_pkts': '756065912',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '885601656615',
                                                                                                'in_ucast_pkts': '756065912',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '70427988816',
                                                                                                'out_ucast_pkts': '232344884',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '70427988816',
                                                                       'out_ucast_pkts': '232344884',
                                                                       'poller_hash': 'MTI3NDdCOXBlZXJpbnRmcG9uIDI0MTg0NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241845',
                                                                                               'links': [       {       'connected_time': '5172934',
                                                                                                                        'fiber_length': 6153,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdCOXBlZXJwb25wb24gMjQxODQ1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4346 LINE 64 (JANTZI JASON & '
                                                 'ROCHELLE)',
                                         'poller_hash': 'MTI3NDdCOQ==',
                                         'serial_number': '12747B9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:09:c9',
                                         'uptime': '5172934'},
                                 {       'counter_source': '241855/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241855/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2048199528',
                                                                       'in_ucast_pkts': '14379354',
                                                                       'mac_address': '04:bc:9f:5b:cd:b9',
                                                                       'name': '241855/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '44483617115',
                                                                       'out_ucast_pkts': '32052460',
                                                                       'poller_hash': 'MTI3NDAzN3BlZXJpbnRmMjQxODU1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241855',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '44483617115',
                                                                       'in_ucast_pkts': '32052460',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '44483617115',
                                                                                                'in_ucast_pkts': '32052460',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2048199528',
                                                                                                'out_ucast_pkts': '14379354',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2048199528',
                                                                       'out_ucast_pkts': '14379354',
                                                                       'poller_hash': 'MTI3NDAzN3BlZXJpbnRmcG9uIDI0MTg1NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241855',
                                                                                               'links': [       {       'connected_time': '449345',
                                                                                                                        'fiber_length': 1121,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDAzN3BlZXJwb25wb24gMjQxODU1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 MAIN ST S APT 203 (Psi)',
                                         'poller_hash': 'MTI3NDAzNw==',
                                         'serial_number': '1274037',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cd:b9',
                                         'uptime': '449345'},
                                 {       'counter_source': '241875/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:73'],
                                                                       'id': '241875/g1',
                                                                       'in_nucast_pkts': '133',
                                                                       'in_octets': '45309852481',
                                                                       'in_ucast_pkts': '77839351',
                                                                       'mac_address': '04:bc:9f:5b:e6:0c',
                                                                       'mtu': 2000,
                                                                       'name': '241875/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1439353',
                                                                       'out_octets': '293127104363',
                                                                       'out_ucast_pkts': '213013907',
                                                                       'poller_hash': 'MTI3NDM0MXBlZXJpbnRmMjQxODc1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '241875',
                                                                       'in_nucast_pkts': '1439353',
                                                                       'in_octets': '293127104363',
                                                                       'in_ucast_pkts': '213013907',
                                                                       'links': [       {       'in_nucast_pkts': '1439353',
                                                                                                'in_octets': '293127104363',
                                                                                                'in_ucast_pkts': '213013907',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '133',
                                                                                                'out_octets': '45309852481',
                                                                                                'out_ucast_pkts': '77839351',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '133',
                                                                       'out_octets': '45309852481',
                                                                       'out_ucast_pkts': '77839351',
                                                                       'poller_hash': 'MTI3NDM0MXBlZXJpbnRmcG9uIDI0MTg3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241875',
                                                                                               'links': [       {       'connected_time': '676588',
                                                                                                                        'fiber_length': 1146,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDM0MXBlZXJwb25wb24gMjQxODc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 MAIN ST S (GERBER CASSIDY)',
                                         'poller_hash': 'MTI3NDM0MQ==',
                                         'serial_number': '1274341',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e6:09',
                                         'uptime': '676588'},
                                 {       'counter_source': '241895/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241895/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '90269331956',
                                                                       'in_ucast_pkts': '340180351',
                                                                       'mac_address': '04:bc:9f:5b:d0:09',
                                                                       'name': '241895/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '182',
                                                                       'out_octets': '1536934147351',
                                                                       'out_ucast_pkts': '1224198997',
                                                                       'poller_hash': 'MTI3NDA4MXBlZXJpbnRmMjQxODk1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241895',
                                                                       'in_nucast_pkts': '182',
                                                                       'in_octets': '1536934147351',
                                                                       'in_ucast_pkts': '1224198997',
                                                                       'links': [       {       'in_nucast_pkts': '182',
                                                                                                'in_octets': '1536934147351',
                                                                                                'in_ucast_pkts': '1224198997',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '90269331956',
                                                                                                'out_ucast_pkts': '340180351',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '90269331956',
                                                                       'out_ucast_pkts': '340180351',
                                                                       'poller_hash': 'MTI3NDA4MXBlZXJpbnRmcG9uIDI0MTg5NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241895',
                                                                                               'links': [       {       'connected_time': '71386248',
                                                                                                                        'fiber_length': 1905,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA4MXBlZXJwb25wb24gMjQxODk1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4710 LINE 64 (MUELLER PAUL)',
                                         'poller_hash': 'MTI3NDA4MQ==',
                                         'serial_number': '1274081',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d0:09',
                                         'uptime': '71386248'},
                                 {       'counter_source': '241905/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241905/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '88326949478',
                                                                       'in_ucast_pkts': '506162564',
                                                                       'mac_address': '04:bc:9f:5b:cf:09',
                                                                       'name': '241905/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '45',
                                                                       'out_octets': '1994833514359',
                                                                       'out_ucast_pkts': '1677927265',
                                                                       'poller_hash': 'MTI3NDA2MXBlZXJpbnRmMjQxOTA1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241905',
                                                                       'in_nucast_pkts': '45',
                                                                       'in_octets': '1994833514359',
                                                                       'in_ucast_pkts': '1677927265',
                                                                       'links': [       {       'in_nucast_pkts': '45',
                                                                                                'in_octets': '1994833514359',
                                                                                                'in_ucast_pkts': '1677927265',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '88326949478',
                                                                                                'out_ucast_pkts': '506162564',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '88326949478',
                                                                       'out_ucast_pkts': '506162564',
                                                                       'poller_hash': 'MTI3NDA2MXBlZXJpbnRmcG9uIDI0MTkwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241905',
                                                                                               'links': [       {       'connected_time': '24461472',
                                                                                                                        'fiber_length': 11462,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA2MXBlZXJwb25wb24gMjQxOTA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7020 PERTH ROAD 121 (BARBU THOMAS & '
                                                 'ROSE)',
                                         'poller_hash': 'MTI3NDA2MQ==',
                                         'serial_number': '1274061',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cf:09',
                                         'uptime': '24461472'},
                                 {       'counter_source': '241915/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241915/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '158620178358',
                                                                       'in_ucast_pkts': '361389177',
                                                                       'mac_address': '04:bc:9f:5c:20:d1',
                                                                       'name': '241915/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '110',
                                                                       'out_octets': '541859051869',
                                                                       'out_ucast_pkts': '597152415',
                                                                       'poller_hash': 'MTI3NEE5QXBlZXJpbnRmMjQxOTE1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241915',
                                                                       'in_nucast_pkts': '110',
                                                                       'in_octets': '541859051869',
                                                                       'in_ucast_pkts': '597152415',
                                                                       'links': [       {       'in_nucast_pkts': '110',
                                                                                                'in_octets': '541859051869',
                                                                                                'in_ucast_pkts': '597152415',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '158620178358',
                                                                                                'out_ucast_pkts': '361389177',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '158620178358',
                                                                       'out_ucast_pkts': '361389177',
                                                                       'poller_hash': 'MTI3NEE5QXBlZXJpbnRmcG9uIDI0MTkxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241915',
                                                                                               'links': [       {       'connected_time': '12029246',
                                                                                                                        'fiber_length': 14448,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE5QXBlZXJwb25wb24gMjQxOTE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4059 LINE 76 (MILLBANK MFG INC)',
                                         'poller_hash': 'MTI3NEE5QQ==',
                                         'serial_number': '1274A9A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:d1',
                                         'uptime': '12029246'},
                                 {       'counter_source': '242032/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242032/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '82631856915',
                                                                       'in_ucast_pkts': '397562586',
                                                                       'mac_address': '04:bc:9f:5b:d8:69',
                                                                       'name': '242032/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '149',
                                                                       'out_octets': '2766669720909',
                                                                       'out_ucast_pkts': '2067954737',
                                                                       'poller_hash': 'MTI3NDE4RHBlZXJpbnRmMjQyMDMyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242032',
                                                                       'in_nucast_pkts': '149',
                                                                       'in_octets': '2766669720909',
                                                                       'in_ucast_pkts': '2067954737',
                                                                       'links': [       {       'in_nucast_pkts': '149',
                                                                                                'in_octets': '2766669720909',
                                                                                                'in_ucast_pkts': '2067954737',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '82631856915',
                                                                                                'out_ucast_pkts': '397562586',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '82631856915',
                                                                       'out_ucast_pkts': '397562586',
                                                                       'poller_hash': 'MTI3NDE4RHBlZXJpbnRmcG9uIDI0MjAzMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242032',
                                                                                               'links': [       {       'connected_time': '688476',
                                                                                                                        'fiber_length': 2240,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE4RHBlZXJwb25wb24gMjQyMDMy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4784 LINE 64 (FABER JOHN)',
                                         'poller_hash': 'MTI3NDE4RA==',
                                         'serial_number': '127418D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:69',
                                         'uptime': '688476'},
                                 {       'counter_source': '242125/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:fa:10:50'],
                                                                       'id': '242125/g1',
                                                                       'in_nucast_pkts': '2822286',
                                                                       'in_octets': '280912497924',
                                                                       'in_ucast_pkts': '3866452063',
                                                                       'mac_address': '04:bc:9f:5b:d2:44',
                                                                       'mtu': 2000,
                                                                       'name': '242125/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '174022001',
                                                                       'out_octets': '23577958929019',
                                                                       'out_ucast_pkts': '15880168294',
                                                                       'poller_hash': 'MTI3NDBDOHBlZXJpbnRmMjQyMTI1L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '242125',
                                                                       'in_nucast_pkts': '174022001',
                                                                       'in_octets': '23577958929019',
                                                                       'in_ucast_pkts': '15880168294',
                                                                       'links': [       {       'in_nucast_pkts': '174022001',
                                                                                                'in_octets': '23577958929019',
                                                                                                'in_ucast_pkts': '15880168294',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2822286',
                                                                                                'out_octets': '280912497924',
                                                                                                'out_ucast_pkts': '3866452063',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2822286',
                                                                       'out_octets': '280912497924',
                                                                       'out_ucast_pkts': '3866452063',
                                                                       'poller_hash': 'MTI3NDBDOHBlZXJpbnRmcG9uIDI0MjEyNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242125',
                                                                                               'links': [       {       'connected_time': '71289058',
                                                                                                                        'fiber_length': 12061,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBDOHBlZXJwb25wb24gMjQyMTI1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7110 PERTH ROAD 121 (ZAWADZKI MAREK)',
                                         'poller_hash': 'MTI3NDBDOA==',
                                         'serial_number': '12740C8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d2:41',
                                         'uptime': '71289058'},
                                 {       'counter_source': '242175/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '6c:63:f8:39:1e:d7'],
                                                                       'id': '242175/g1',
                                                                       'in_nucast_pkts': '11454',
                                                                       'in_octets': '25922705432',
                                                                       'in_ucast_pkts': '102608108',
                                                                       'mac_address': '04:bc:9f:5b:da:6c',
                                                                       'mtu': 2000,
                                                                       'name': '242175/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465439',
                                                                       'out_octets': '547607439196',
                                                                       'out_ucast_pkts': '413589614',
                                                                       'poller_hash': 'MTI3NDFDRHBlZXJpbnRmMjQyMTc1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '242175',
                                                                       'in_nucast_pkts': '1465439',
                                                                       'in_octets': '547607439196',
                                                                       'in_ucast_pkts': '413589614',
                                                                       'links': [       {       'in_nucast_pkts': '1465439',
                                                                                                'in_octets': '547607439196',
                                                                                                'in_ucast_pkts': '413589614',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11454',
                                                                                                'out_octets': '25922705432',
                                                                                                'out_ucast_pkts': '102608108',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11454',
                                                                       'out_octets': '25922705432',
                                                                       'out_ucast_pkts': '102608108',
                                                                       'poller_hash': 'MTI3NDFDRHBlZXJpbnRmcG9uIDI0MjE3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242175',
                                                                                               'links': [       {       'connected_time': '688574',
                                                                                                                        'fiber_length': 6318,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFDRHBlZXJwb25wb24gMjQyMTc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4199 LINE 61 MILVERTON (BOSHART '
                                                 'INDUSTRIES)',
                                         'poller_hash': 'MTI3NDFDRA==',
                                         'serial_number': '12741CD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:da:69',
                                         'uptime': '688574'},
                                 {       'counter_source': '242185/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:54:a4:67'],
                                                                       'id': '242185/g1',
                                                                       'in_nucast_pkts': '12019',
                                                                       'in_octets': '4046134495',
                                                                       'in_ucast_pkts': '17186622',
                                                                       'mac_address': '04:bc:9f:6e:e2:89',
                                                                       'mtu': 2000,
                                                                       'name': '242185/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465086',
                                                                       'out_octets': '65723722795',
                                                                       'out_ucast_pkts': '52617894',
                                                                       'poller_hash': 'MTI5NkZENHBlZXJpbnRmMjQyMTg1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '242185',
                                                                       'in_nucast_pkts': '1465086',
                                                                       'in_octets': '65723722795',
                                                                       'in_ucast_pkts': '52617894',
                                                                       'links': [       {       'in_nucast_pkts': '1465086',
                                                                                                'in_octets': '65723722795',
                                                                                                'in_ucast_pkts': '52617894',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12019',
                                                                                                'out_octets': '4046134495',
                                                                                                'out_ucast_pkts': '17186622',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12019',
                                                                       'out_octets': '4046134495',
                                                                       'out_ucast_pkts': '17186622',
                                                                       'poller_hash': 'MTI5NkZENHBlZXJpbnRmcG9uIDI0MjE4NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242185',
                                                                                               'links': [       {       'connected_time': '688364',
                                                                                                                        'fiber_length': 3350,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NkZENHBlZXJwb25wb24gMjQyMTg1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '58 PUGH ST (PATHAMMAVONG LEE)',
                                         'poller_hash': 'MTI5NkZENA==',
                                         'serial_number': '1296FD4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:e2:86',
                                         'uptime': '688364'},
                                 {       'counter_source': '242195/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:39:c0'],
                                                                       'id': '242195/g1',
                                                                       'in_nucast_pkts': '14711',
                                                                       'in_octets': '3858174674',
                                                                       'in_ucast_pkts': '9748848',
                                                                       'mac_address': '04:bc:9f:5b:d1:ec',
                                                                       'mtu': 2000,
                                                                       'name': '242195/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1435093',
                                                                       'out_octets': '38802419120',
                                                                       'out_ucast_pkts': '30918654',
                                                                       'poller_hash': 'MTI3NDBCRHBlZXJpbnRmMjQyMTk1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '242195',
                                                                       'in_nucast_pkts': '1435093',
                                                                       'in_octets': '38802419120',
                                                                       'in_ucast_pkts': '30918654',
                                                                       'links': [       {       'in_nucast_pkts': '1435093',
                                                                                                'in_octets': '38802419120',
                                                                                                'in_ucast_pkts': '30918654',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14711',
                                                                                                'out_octets': '3858174674',
                                                                                                'out_ucast_pkts': '9748848',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14711',
                                                                       'out_octets': '3858174674',
                                                                       'out_ucast_pkts': '9748848',
                                                                       'poller_hash': 'MTI3NDBCRHBlZXJpbnRmcG9uIDI0MjE5NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242195',
                                                                                               'links': [       {       'connected_time': '688557',
                                                                                                                        'fiber_length': 1808,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBCRHBlZXJwb25wb24gMjQyMTk1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (POLANIC ADRIANA)',
                                         'poller_hash': 'MTI3NDBCRA==',
                                         'serial_number': '12740BD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d1:e9',
                                         'uptime': '688557'},
                                 {       'counter_source': '242362/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242362/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '121829754625',
                                                                       'in_ucast_pkts': '476481414',
                                                                       'mac_address': '04:bc:9f:5b:d8:99',
                                                                       'name': '242362/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '40',
                                                                       'out_octets': '2350996311535',
                                                                       'out_ucast_pkts': '1762672468',
                                                                       'poller_hash': 'MTI3NDE5M3BlZXJpbnRmMjQyMzYyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242362',
                                                                       'in_nucast_pkts': '40',
                                                                       'in_octets': '2350996311535',
                                                                       'in_ucast_pkts': '1762672468',
                                                                       'links': [       {       'in_nucast_pkts': '40',
                                                                                                'in_octets': '2350996311535',
                                                                                                'in_ucast_pkts': '1762672468',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '121829754625',
                                                                                                'out_ucast_pkts': '476481414',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '121829754625',
                                                                       'out_ucast_pkts': '476481414',
                                                                       'poller_hash': 'MTI3NDE5M3BlZXJpbnRmcG9uIDI0MjM2Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242362',
                                                                                               'links': [       {       'connected_time': '688587',
                                                                                                                        'fiber_length': 11877,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5M3BlZXJwb25wb24gMjQyMzYy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4223 LINE 61 (DIETRICH JESSE & '
                                                 'JENNA)',
                                         'poller_hash': 'MTI3NDE5Mw==',
                                         'serial_number': '1274193',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:99',
                                         'uptime': '688587'},
                                 {       'counter_source': '242718/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242718/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '37983413056',
                                                                       'in_ucast_pkts': '4405846786',
                                                                       'mac_address': '04:bc:9f:5b:d4:d9',
                                                                       'name': '242718/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '149',
                                                                       'out_octets': '413367901268',
                                                                       'out_ucast_pkts': '368021771',
                                                                       'poller_hash': 'MTI3NDExQnBlZXJpbnRmMjQyNzE4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242718',
                                                                       'in_nucast_pkts': '149',
                                                                       'in_octets': '413367901268',
                                                                       'in_ucast_pkts': '368021771',
                                                                       'links': [       {       'in_nucast_pkts': '149',
                                                                                                'in_octets': '413367901268',
                                                                                                'in_ucast_pkts': '368021771',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '37983413056',
                                                                                                'out_ucast_pkts': '4405846786',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '37983413056',
                                                                       'out_ucast_pkts': '4405846786',
                                                                       'poller_hash': 'MTI3NDExQnBlZXJpbnRmcG9uIDI0MjcxOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242718',
                                                                                               'links': [       {       'connected_time': '688471',
                                                                                                                        'fiber_length': 2094,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDExQnBlZXJwb25wb24gMjQyNzE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4773 LINE 64 (BYLER JAMES)',
                                         'poller_hash': 'MTI3NDExQg==',
                                         'serial_number': '127411B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d4:d9',
                                         'uptime': '688471'},
                                 {       'counter_source': '242857/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242857/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '50401923205',
                                                                       'in_ucast_pkts': '260125908',
                                                                       'mac_address': '04:bc:9f:5b:e3:01',
                                                                       'name': '242857/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '80',
                                                                       'out_octets': '1079527518651',
                                                                       'out_ucast_pkts': '801548605',
                                                                       'poller_hash': 'MTI3NDJFMHBlZXJpbnRmMjQyODU3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242857',
                                                                       'in_nucast_pkts': '80',
                                                                       'in_octets': '1079527518651',
                                                                       'in_ucast_pkts': '801548605',
                                                                       'links': [       {       'in_nucast_pkts': '80',
                                                                                                'in_octets': '1079527518651',
                                                                                                'in_ucast_pkts': '801548605',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '50401923205',
                                                                                                'out_ucast_pkts': '260125908',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '50401923205',
                                                                       'out_ucast_pkts': '260125908',
                                                                       'poller_hash': 'MTI3NDJFMHBlZXJpbnRmcG9uIDI0Mjg1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242857',
                                                                                               'links': [       {       'connected_time': '688596',
                                                                                                                        'fiber_length': 12393,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJFMHBlZXJwb25wb24gMjQyODU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3994 LINE 61 (VAN NES JAMES)',
                                         'poller_hash': 'MTI3NDJFMA==',
                                         'serial_number': '12742E0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:01',
                                         'uptime': '688596'},
                                 {       'counter_source': '242867/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242867/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '140356023830',
                                                                       'in_ucast_pkts': '385230903',
                                                                       'mac_address': '04:bc:9f:6e:e7:16',
                                                                       'name': '242867/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '80',
                                                                       'out_octets': '1434262779427',
                                                                       'out_ucast_pkts': '1236713339',
                                                                       'poller_hash': 'MTI5NzA2NnBlZXJpbnRmMjQyODY3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242867',
                                                                       'in_nucast_pkts': '80',
                                                                       'in_octets': '1434262779427',
                                                                       'in_ucast_pkts': '1236713339',
                                                                       'links': [       {       'in_nucast_pkts': '80',
                                                                                                'in_octets': '1434262779427',
                                                                                                'in_ucast_pkts': '1236713339',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '140356023830',
                                                                                                'out_ucast_pkts': '385230903',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '140356023830',
                                                                       'out_ucast_pkts': '385230903',
                                                                       'poller_hash': 'MTI5NzA2NnBlZXJpbnRmcG9uIDI0Mjg2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242867',
                                                                                               'links': [       {       'connected_time': '688594',
                                                                                                                        'fiber_length': 10438,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzA2NnBlZXJwb25wb24gMjQyODY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4104 LINE 61 (BAUMAN FRED)',
                                         'poller_hash': 'MTI5NzA2Ng==',
                                         'serial_number': '1297066',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:e7:16',
                                         'uptime': '688594'},
                                 {       'counter_source': '242877/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242877/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '226829363679',
                                                                       'in_ucast_pkts': '6117388714',
                                                                       'mac_address': '04:bc:9f:6e:e2:3e',
                                                                       'name': '242877/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '60',
                                                                       'out_octets': '6154284587277',
                                                                       'out_ucast_pkts': '4374873111',
                                                                       'poller_hash': 'MTI5NkZDQnBlZXJpbnRmMjQyODc3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242877',
                                                                       'in_nucast_pkts': '60',
                                                                       'in_octets': '6154284587277',
                                                                       'in_ucast_pkts': '4374873111',
                                                                       'links': [       {       'in_nucast_pkts': '60',
                                                                                                'in_octets': '6154284587277',
                                                                                                'in_ucast_pkts': '4374873111',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '226829363679',
                                                                                                'out_ucast_pkts': '6117388714',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '226829363679',
                                                                       'out_ucast_pkts': '6117388714',
                                                                       'poller_hash': 'MTI5NkZDQnBlZXJpbnRmcG9uIDI0Mjg3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242877',
                                                                                               'links': [       {       'connected_time': '688587',
                                                                                                                        'fiber_length': 11975,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NkZDQnBlZXJwb25wb24gMjQyODc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4009 LINE 61 (CHALMERS KEVIN)',
                                         'poller_hash': 'MTI5NkZDQg==',
                                         'serial_number': '1296FCB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:e2:3e',
                                         'uptime': '688587'},
                                 {       'counter_source': '242887/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242887/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '129771244443',
                                                                       'in_ucast_pkts': '323933596',
                                                                       'mac_address': '04:bc:9f:5b:d0:b9',
                                                                       'name': '242887/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '110',
                                                                       'out_octets': '351863410679',
                                                                       'out_ucast_pkts': '491119131',
                                                                       'poller_hash': 'MTI3NDA5N3BlZXJpbnRmMjQyODg3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242887',
                                                                       'in_nucast_pkts': '110',
                                                                       'in_octets': '351863410679',
                                                                       'in_ucast_pkts': '491119131',
                                                                       'links': [       {       'in_nucast_pkts': '110',
                                                                                                'in_octets': '351863410679',
                                                                                                'in_ucast_pkts': '491119131',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '129771244443',
                                                                                                'out_ucast_pkts': '323933596',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '129771244443',
                                                                       'out_ucast_pkts': '323933596',
                                                                       'poller_hash': 'MTI3NDA5N3BlZXJpbnRmcG9uIDI0Mjg4Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242887',
                                                                                               'links': [       {       'connected_time': '70767365',
                                                                                                                        'fiber_length': 11316,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA5N3BlZXJwb25wb24gMjQyODg3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6346 PERTH ROAD 121 (ZEHR GERALD & '
                                                 'ALICE BARN)',
                                         'poller_hash': 'MTI3NDA5Nw==',
                                         'serial_number': '1274097',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d0:b9',
                                         'uptime': '70767365'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '242897/g1',
                                                                       'name': '242897/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDJEMnBlZXJpbnRmMjQyODk3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242897',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDJEMnBlZXJpbnRmcG9uIDI0Mjg5Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242897',
                                                                                               'links': [       {       'connected_time': '70765845',
                                                                                                                        'fiber_length': 11193,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJEMnBlZXJwb25wb24gMjQyODk3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6346 PERTH ROAD 121 (ZEHR GERALD & '
                                                 'ALICE HOUSE)',
                                         'poller_hash': 'MTI3NDJEMg==',
                                         'serial_number': '12742D2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e2:91',
                                         'uptime': '70765845'},
                                 {       'counter_source': '242907/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242907/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '167933331365',
                                                                       'in_ucast_pkts': '786173351',
                                                                       'mac_address': '04:bc:9f:5b:e5:19',
                                                                       'name': '242907/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '80',
                                                                       'out_octets': '3551815995783',
                                                                       'out_ucast_pkts': '2577934097',
                                                                       'poller_hash': 'MTI3NDMyM3BlZXJpbnRmMjQyOTA3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242907',
                                                                       'in_nucast_pkts': '80',
                                                                       'in_octets': '3551815995783',
                                                                       'in_ucast_pkts': '2577934097',
                                                                       'links': [       {       'in_nucast_pkts': '80',
                                                                                                'in_octets': '3551815995783',
                                                                                                'in_ucast_pkts': '2577934097',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '167933331365',
                                                                                                'out_ucast_pkts': '786173351',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '167933331365',
                                                                       'out_ucast_pkts': '786173351',
                                                                       'poller_hash': 'MTI3NDMyM3BlZXJpbnRmcG9uIDI0MjkwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242907',
                                                                                               'links': [       {       'connected_time': '688591',
                                                                                                                        'fiber_length': 10649,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMyM3BlZXJwb25wb24gMjQyOTA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6394 PERTH ROAD 121 (YANTZI AMY)',
                                         'poller_hash': 'MTI3NDMyMw==',
                                         'serial_number': '1274323',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:19',
                                         'uptime': '688591'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '242930/g1',
                                                                       'name': '242930/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDI5OHBlZXJpbnRmMjQyOTMwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242930',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDI5OHBlZXJpbnRmcG9uIDI0MjkzMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242930',
                                                                                               'links': [       {       'connected_time': '69143959',
                                                                                                                        'fiber_length': 12108,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDI5OHBlZXJwb25wb24gMjQyOTMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7092 PERTH ROAD 121 (WEBER SARAH & '
                                                 'EDNA)',
                                         'poller_hash': 'MTI3NDI5OA==',
                                         'serial_number': '1274298',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e0:c1',
                                         'uptime': '69143959'},
                                 {       'counter_source': '243077/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '243077/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '195176625627',
                                                                       'in_ucast_pkts': '1251500223',
                                                                       'mac_address': '04:bc:9f:5b:d8:b1',
                                                                       'name': '243077/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '65',
                                                                       'out_octets': '4238301800538',
                                                                       'out_ucast_pkts': '3015790347',
                                                                       'poller_hash': 'MTI3NDE5NnBlZXJpbnRmMjQzMDc3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '243077',
                                                                       'in_nucast_pkts': '65',
                                                                       'in_octets': '4238301800538',
                                                                       'in_ucast_pkts': '3015790347',
                                                                       'links': [       {       'in_nucast_pkts': '65',
                                                                                                'in_octets': '4238301800538',
                                                                                                'in_ucast_pkts': '3015790347',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '195176625627',
                                                                                                'out_ucast_pkts': '1251500223',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '195176625627',
                                                                       'out_ucast_pkts': '1251500223',
                                                                       'poller_hash': 'MTI3NDE5NnBlZXJpbnRmcG9uIDI0MzA3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243077',
                                                                                               'links': [       {       'connected_time': '538597',
                                                                                                                        'fiber_length': 18065,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5NnBlZXJwb25wb24gMjQzMDc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4165 LINE 50 (JANTZI ROSS)',
                                         'poller_hash': 'MTI3NDE5Ng==',
                                         'serial_number': '1274196',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:b1',
                                         'uptime': '538597'},
                                 {       'counter_source': '243097/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '243097/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '592707568401',
                                                                       'in_ucast_pkts': '2553208413',
                                                                       'mac_address': '04:bc:9f:5b:d8:81',
                                                                       'name': '243097/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '33',
                                                                       'out_octets': '6161468167992',
                                                                       'out_ucast_pkts': '5201328353',
                                                                       'poller_hash': 'MTI3NDE5MHBlZXJpbnRmMjQzMDk3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '243097',
                                                                       'in_nucast_pkts': '33',
                                                                       'in_octets': '6161468167992',
                                                                       'in_ucast_pkts': '5201328353',
                                                                       'links': [       {       'in_nucast_pkts': '33',
                                                                                                'in_octets': '6161468167992',
                                                                                                'in_ucast_pkts': '5201328353',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '592707568401',
                                                                                                'out_ucast_pkts': '2553208413',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '592707568401',
                                                                       'out_ucast_pkts': '2553208413',
                                                                       'poller_hash': 'MTI3NDE5MHBlZXJpbnRmcG9uIDI0MzA5Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243097',
                                                                                               'links': [       {       'connected_time': '13208956',
                                                                                                                        'fiber_length': 1552,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5MHBlZXJwb25wb24gMjQzMDk3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '78 MAIN ST N (KUEPFER JOHN & EILEEN)',
                                         'poller_hash': 'MTI3NDE5MA==',
                                         'serial_number': '1274190',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:81',
                                         'uptime': '13208956'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '243107/g1',
                                                                       'name': '243107/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDAzRnBlZXJpbnRmMjQzMTA3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '243107',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDAzRnBlZXJpbnRmcG9uIDI0MzEwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243107',
                                                                                               'links': [       {       'connected_time': '70334542',
                                                                                                                        'fiber_length': 18628,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDAzRnBlZXJwb25wb24gMjQzMTA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4227 LINE 50 (BRUNK MELVIN)',
                                         'poller_hash': 'MTI3NDAzRg==',
                                         'serial_number': '127403F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cd:f9',
                                         'uptime': '70334542'},
                                 {       'counter_source': '243239/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e0:15'],
                                                                       'id': '243239/g1',
                                                                       'in_nucast_pkts': '12253',
                                                                       'in_octets': '14746860714',
                                                                       'in_ucast_pkts': '87296061',
                                                                       'mac_address': '04:bc:9f:5b:d8:ec',
                                                                       'mtu': 2000,
                                                                       'name': '243239/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465166',
                                                                       'out_octets': '542406624599',
                                                                       'out_ucast_pkts': '399673695',
                                                                       'poller_hash': 'MTI3NDE5RHBlZXJpbnRmMjQzMjM5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '243239',
                                                                       'in_nucast_pkts': '1465166',
                                                                       'in_octets': '542406624599',
                                                                       'in_ucast_pkts': '399673695',
                                                                       'links': [       {       'in_nucast_pkts': '1465166',
                                                                                                'in_octets': '542406624599',
                                                                                                'in_ucast_pkts': '399673695',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12253',
                                                                                                'out_octets': '14746860714',
                                                                                                'out_ucast_pkts': '87296061',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12253',
                                                                       'out_octets': '14746860714',
                                                                       'out_ucast_pkts': '87296061',
                                                                       'poller_hash': 'MTI3NDE5RHBlZXJpbnRmcG9uIDI0MzIzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243239',
                                                                                               'links': [       {       'connected_time': '688363',
                                                                                                                        'fiber_length': 2549,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5RHBlZXJwb25wb24gMjQzMjM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '105 PUGH ST (HEIDENREICH MIKE & '
                                                 'JEANETTE)',
                                         'poller_hash': 'MTI3NDE5RA==',
                                         'serial_number': '127419D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:e9',
                                         'uptime': '688363'},
                                 {       'counter_source': '243249/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e1:28'],
                                                                       'id': '243249/g1',
                                                                       'in_nucast_pkts': '132680975',
                                                                       'in_octets': '288795709436',
                                                                       'in_ucast_pkts': '700078139',
                                                                       'mac_address': '04:bc:9f:5b:ce:5c',
                                                                       'mtu': 2000,
                                                                       'name': '243249/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '88917767',
                                                                       'out_octets': '1532352840541',
                                                                       'out_ucast_pkts': '1392416080',
                                                                       'poller_hash': 'MTI3NDA0QnBlZXJpbnRmMjQzMjQ5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '243249',
                                                                       'in_nucast_pkts': '88917767',
                                                                       'in_octets': '1532352840541',
                                                                       'in_ucast_pkts': '1392416080',
                                                                       'links': [       {       'in_nucast_pkts': '88917767',
                                                                                                'in_octets': '1532352840541',
                                                                                                'in_ucast_pkts': '1392416080',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '132680975',
                                                                                                'out_octets': '288795709436',
                                                                                                'out_ucast_pkts': '700078139',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '132680975',
                                                                       'out_octets': '288795709436',
                                                                       'out_ucast_pkts': '700078139',
                                                                       'poller_hash': 'MTI3NDA0QnBlZXJpbnRmcG9uIDI0MzI0OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243249',
                                                                                               'links': [       {       'connected_time': '70267782',
                                                                                                                        'fiber_length': 2254,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA0QnBlZXJwb25wb24gMjQzMjQ5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4742 LINE 64 (BIEHN LOWELL)',
                                         'poller_hash': 'MTI3NDA0Qg==',
                                         'serial_number': '127404B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:59',
                                         'uptime': '70267782'},
                                 {       'counter_source': '243711/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '243711/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '208861945788',
                                                                       'in_ucast_pkts': '425530294',
                                                                       'mac_address': '04:bc:9f:6e:e2:e6',
                                                                       'name': '243711/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '130',
                                                                       'out_octets': '1171408268363',
                                                                       'out_ucast_pkts': '1055814527',
                                                                       'poller_hash': 'MTI5NkZFMHBlZXJpbnRmMjQzNzExL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '243711',
                                                                       'in_nucast_pkts': '130',
                                                                       'in_octets': '1171408268363',
                                                                       'in_ucast_pkts': '1055814527',
                                                                       'links': [       {       'in_nucast_pkts': '130',
                                                                                                'in_octets': '1171408268363',
                                                                                                'in_ucast_pkts': '1055814527',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '208861945788',
                                                                                                'out_ucast_pkts': '425530294',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '208861945788',
                                                                       'out_ucast_pkts': '425530294',
                                                                       'poller_hash': 'MTI5NkZFMHBlZXJpbnRmcG9uIDI0MzcxMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243711',
                                                                                               'links': [       {       'connected_time': '70164143',
                                                                                                                        'fiber_length': 11582,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NkZFMHBlZXJwb25wb24gMjQzNzEx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4198 LINE 61 (CONRAD JOHN)',
                                         'poller_hash': 'MTI5NkZFMA==',
                                         'serial_number': '1296FE0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:e2:e6',
                                         'uptime': '70164143'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '243721/g1',
                                                                       'name': '243721/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDJFRHBlZXJpbnRmMjQzNzIxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '243721',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDJFRHBlZXJpbnRmcG9uIDI0MzcyMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243721',
                                                                                               'links': [       {       'connected_time': '70158734',
                                                                                                                        'fiber_length': 17083,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJFRHBlZXJwb25wb24gMjQzNzIx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4084 LINE 50 (MANTLE BRUCE)',
                                         'poller_hash': 'MTI3NDJFRA==',
                                         'serial_number': '12742ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:69',
                                         'uptime': '70158734'},
                                 {       'counter_source': '243751/g1+243751/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:c5:00'],
                                                                       'id': '243751/g1',
                                                                       'in_nucast_pkts': '2800707',
                                                                       'in_octets': '16679957962',
                                                                       'in_ucast_pkts': '153841009',
                                                                       'mac_address': '04:bc:9f:5b:d8:d4',
                                                                       'mtu': 2000,
                                                                       'name': '243751/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '169429324',
                                                                       'out_octets': '803301173724',
                                                                       'out_ucast_pkts': '536538675',
                                                                       'poller_hash': 'MTI3NDE5QXBlZXJpbnRmMjQzNzUxL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:c7:32'],
                                                                       'id': '243751/g2',
                                                                       'in_nucast_pkts': '2803272',
                                                                       'in_octets': '64436143678',
                                                                       'in_ucast_pkts': '825724871',
                                                                       'mac_address': '04:bc:9f:5b:d8:d5',
                                                                       'mtu': 2000,
                                                                       'name': '243751/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '169420805',
                                                                       'out_octets': '4842151249221',
                                                                       'out_ucast_pkts': '3266521024',
                                                                       'poller_hash': 'MTI3NDE5QXBlZXJpbnRmMjQzNzUxL2cy',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '243751',
                                                                       'in_nucast_pkts': '338850129',
                                                                       'in_octets': '5645452422945',
                                                                       'in_ucast_pkts': '3803059699',
                                                                       'links': [       {       'in_nucast_pkts': '338850129',
                                                                                                'in_octets': '5645452422945',
                                                                                                'in_ucast_pkts': '3803059699',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5603979',
                                                                                                'out_octets': '81116101640',
                                                                                                'out_ucast_pkts': '979565880',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5603979',
                                                                       'out_octets': '81116101640',
                                                                       'out_ucast_pkts': '979565880',
                                                                       'poller_hash': 'MTI3NDE5QXBlZXJpbnRmcG9uIDI0Mzc1MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243751',
                                                                                               'links': [       {       'connected_time': '70093395',
                                                                                                                        'fiber_length': 18525,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5QXBlZXJwb25wb24gMjQzNzUx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '148 MADDISON ST E MONKTON (WICK '
                                                 'GARY)',
                                         'poller_hash': 'MTI3NDE5QQ==',
                                         'serial_number': '127419A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:d1',
                                         'uptime': '70093395'},
                                 {       'counter_source': '244730/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '244730/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '191856152969',
                                                                       'in_ucast_pkts': '26369288604',
                                                                       'mac_address': '04:bc:9f:5b:cd:49',
                                                                       'name': '244730/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '46',
                                                                       'out_octets': '2465301971014',
                                                                       'out_ucast_pkts': '1891448894',
                                                                       'poller_hash': 'MTI3NDAyOXBlZXJpbnRmMjQ0NzMwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '244730',
                                                                       'in_nucast_pkts': '46',
                                                                       'in_octets': '2465301971014',
                                                                       'in_ucast_pkts': '1891448894',
                                                                       'links': [       {       'in_nucast_pkts': '46',
                                                                                                'in_octets': '2465301971014',
                                                                                                'in_ucast_pkts': '1891448894',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '191856152969',
                                                                                                'out_ucast_pkts': '26369288604',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '191856152969',
                                                                       'out_ucast_pkts': '26369288604',
                                                                       'poller_hash': 'MTI3NDAyOXBlZXJpbnRmcG9uIDI0NDczMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '244730',
                                                                                               'links': [       {       'connected_time': '688581',
                                                                                                                        'fiber_length': 7233,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDAyOXBlZXJwb25wb24gMjQ0NzMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4133 LINE 67 (KUEPFER TREVOR)',
                                         'poller_hash': 'MTI3NDAyOQ==',
                                         'serial_number': '1274029',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cd:49',
                                         'uptime': '688581'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '244810/g1',
                                                                       'name': '244810/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDE5RnBlZXJpbnRmMjQ0ODEwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '244810',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDE5RnBlZXJpbnRmcG9uIDI0NDgxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '244810',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5RnBlZXJwb25wb24gMjQ0ODEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '6700 ROAD 129 (JANTZI DOUG)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "244810" '
                                                                  '(location: GPON '
                                                                  'port "13/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-08-05T16:14:46.360Z',
                                         'poller_hash': 'MTI3NDE5Rg==',
                                         'serial_number': '127419F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '244850/g1',
                                                                       'name': '244850/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDA1MHBlZXJpbnRmMjQ0ODUwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '244850',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDA1MHBlZXJpbnRmcG9uIDI0NDg1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '244850',
                                                                                               'links': [       {       'connected_time': '69402860',
                                                                                                                        'fiber_length': 6156,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA1MHBlZXJwb25wb24gMjQ0ODUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6455 ROAD 140 (STREICHER MERLIN)',
                                         'poller_hash': 'MTI3NDA1MA==',
                                         'serial_number': '1274050',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:81',
                                         'uptime': '69402860'},
                                 {       'counter_source': '244870/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '244870/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6912990883',
                                                                       'in_ucast_pkts': '23463037',
                                                                       'mac_address': '04:bc:9f:5b:d8:79',
                                                                       'name': '244870/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '123408950684',
                                                                       'out_ucast_pkts': '107319026',
                                                                       'poller_hash': 'MTI3NDE4RnBlZXJpbnRmMjQ0ODcwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '244870',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '123408950684',
                                                                       'in_ucast_pkts': '107319026',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '123408950684',
                                                                                                'in_ucast_pkts': '107319026',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '6912990883',
                                                                                                'out_ucast_pkts': '23463037',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6912990883',
                                                                       'out_ucast_pkts': '23463037',
                                                                       'poller_hash': 'MTI3NDE4RnBlZXJpbnRmcG9uIDI0NDg3MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '244870',
                                                                                               'links': [       {       'connected_time': '14067777',
                                                                                                                        'fiber_length': 11906,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE4RnBlZXJwb25wb24gMjQ0ODcw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6940 HORIZON TRAIL (KUEPFER CALVIN '
                                                 'R)',
                                         'poller_hash': 'MTI3NDE4Rg==',
                                         'serial_number': '127418F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:79',
                                         'uptime': '14067777'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '244890/g1',
                                                                       'name': '244890/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3M0Y2RnBlZXJpbnRmMjQ0ODkwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '244890',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3M0Y2RnBlZXJpbnRmcG9uIDI0NDg5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '244890',
                                                                                               'links': [       {       'connected_time': '69388586',
                                                                                                                        'fiber_length': 12061,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0Y2RnBlZXJwb25wb24gMjQ0ODkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7098 PERTH ROAD 121 (MARTIN JAMES)',
                                         'poller_hash': 'MTI3M0Y2Rg==',
                                         'serial_number': '1273F6F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:c7:79',
                                         'uptime': '69388586'},
                                 {       'counter_source': '244900/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '244900/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '927990395323',
                                                                       'in_ucast_pkts': '4536997227',
                                                                       'mac_address': 'cc:be:59:6f:ee:4a',
                                                                       'name': '244900/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '18475305747045',
                                                                       'out_ucast_pkts': '13089081302',
                                                                       'poller_hash': 'NTVGOUJDcGVlcmludGYyNDQ5MDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '244900',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '18475305747045',
                                                                       'in_ucast_pkts': '13089081302',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '18475305747045',
                                                                                                'in_ucast_pkts': '13089081302',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '927990395323',
                                                                                                'out_ucast_pkts': '4536997227',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '927990395323',
                                                                       'out_ucast_pkts': '4536997227',
                                                                       'poller_hash': 'NTVGOUJDcGVlcmludGZwb24gMjQ0OTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '244900',
                                                                                               'links': [       {       'connected_time': '23019371',
                                                                                                                        'fiber_length': 10891,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTVGOUJDcGVlcnBvbnBvbiAyNDQ5MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6432 ROAD 154 (OHM PERRY & JENNIFER)',
                                         'poller_hash': 'NTVGOUJD',
                                         'serial_number': '55F9BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6f:ee:4a',
                                         'uptime': '23019371'},
                                 {       'counter_source': '245201/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '245201/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '249691296053',
                                                                       'in_ucast_pkts': '1161133262',
                                                                       'mac_address': '04:bc:9f:5b:cf:c9',
                                                                       'name': '245201/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '90',
                                                                       'out_octets': '3137843610524',
                                                                       'out_ucast_pkts': '2431553918',
                                                                       'poller_hash': 'MTI3NDA3OXBlZXJpbnRmMjQ1MjAxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '245201',
                                                                       'in_nucast_pkts': '90',
                                                                       'in_octets': '3137843610524',
                                                                       'in_ucast_pkts': '2431553918',
                                                                       'links': [       {       'in_nucast_pkts': '90',
                                                                                                'in_octets': '3137843610524',
                                                                                                'in_ucast_pkts': '2431553918',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '249691296053',
                                                                                                'out_ucast_pkts': '1161133262',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '249691296053',
                                                                       'out_ucast_pkts': '1161133262',
                                                                       'poller_hash': 'MTI3NDA3OXBlZXJpbnRmcG9uIDI0NTIwMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245201',
                                                                                               'links': [       {       'connected_time': '10385489',
                                                                                                                        'fiber_length': 11566,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA3OXBlZXJwb25wb24gMjQ1MjAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4067 LINE 61 (VAN DEN TEMPEL TED & '
                                                 'LIZ)',
                                         'poller_hash': 'MTI3NDA3OQ==',
                                         'serial_number': '1274079',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cf:c9',
                                         'uptime': '10385489'},
                                 {       'counter_source': '245528/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '245528/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '294',
                                                                       'in_octets': '172009394793',
                                                                       'in_ucast_pkts': '1591577958',
                                                                       'mac_address': '04:bc:9f:5c:1b:a9',
                                                                       'name': '245528/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '2027',
                                                                       'out_octets': '9420862223397',
                                                                       'out_ucast_pkts': '6340101249',
                                                                       'poller_hash': 'MTI3NDlGNXBlZXJpbnRmMjQ1NTI4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '245528',
                                                                       'in_nucast_pkts': '2027',
                                                                       'in_octets': '9420862223397',
                                                                       'in_ucast_pkts': '6340101249',
                                                                       'links': [       {       'in_nucast_pkts': '2027',
                                                                                                'in_octets': '9420862223397',
                                                                                                'in_ucast_pkts': '6340101249',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '294',
                                                                                                'out_octets': '172009394793',
                                                                                                'out_ucast_pkts': '1591577958',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '294',
                                                                       'out_octets': '172009394793',
                                                                       'out_ucast_pkts': '1591577958',
                                                                       'poller_hash': 'MTI3NDlGNXBlZXJpbnRmcG9uIDI0NTUyOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245528',
                                                                                               'links': [       {       'connected_time': '24528738',
                                                                                                                        'fiber_length': 1656,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlGNXBlZXJwb25wb24gMjQ1NTI4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4741 LINE 64 (SCHADE MARY)',
                                         'poller_hash': 'MTI3NDlGNQ==',
                                         'serial_number': '12749F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1b:a9',
                                         'uptime': '24528738'},
                                 {       'counter_source': '245572/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '74:ac:b9:1f:74:ed'],
                                                                       'id': '245572/g1',
                                                                       'in_nucast_pkts': '156945',
                                                                       'in_octets': '218890006536',
                                                                       'in_ucast_pkts': '488588224',
                                                                       'mac_address': '04:bc:9f:5c:1c:0c',
                                                                       'mtu': 2000,
                                                                       'name': '245572/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465059',
                                                                       'out_octets': '454904259724',
                                                                       'out_ucast_pkts': '457122834',
                                                                       'poller_hash': 'MTI3NEEwMXBlZXJpbnRmMjQ1NTcyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '245572',
                                                                       'in_nucast_pkts': '1465059',
                                                                       'in_octets': '454904259724',
                                                                       'in_ucast_pkts': '457122834',
                                                                       'links': [       {       'in_nucast_pkts': '1465059',
                                                                                                'in_octets': '454904259724',
                                                                                                'in_ucast_pkts': '457122834',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '156945',
                                                                                                'out_octets': '218890006536',
                                                                                                'out_ucast_pkts': '488588224',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '156945',
                                                                       'out_octets': '218890006536',
                                                                       'out_ucast_pkts': '488588224',
                                                                       'poller_hash': 'MTI3NEEwMXBlZXJpbnRmcG9uIDI0NTU3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245572',
                                                                                               'links': [       {       'connected_time': '688358',
                                                                                                                        'fiber_length': 1123,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEwMXBlZXJwb25wb24gMjQ1NTcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 MAIN ST S (ELLIS CHARLES)',
                                         'poller_hash': 'MTI3NEEwMQ==',
                                         'serial_number': '1274A01',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:09',
                                         'uptime': '688358'},
                                 {       'counter_source': '245737/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b4:fb:e4:d3:ff:15'],
                                                                       'id': '245737/g1',
                                                                       'in_nucast_pkts': '1844',
                                                                       'in_octets': '2087926991872',
                                                                       'in_ucast_pkts': '4986827424',
                                                                       'mac_address': '1c:8b:76:00:30:cd',
                                                                       'mtu': 2000,
                                                                       'name': '245737/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '71494496',
                                                                       'out_octets': '5564287166551',
                                                                       'out_ucast_pkts': '6187781673',
                                                                       'poller_hash': 'MTU4RkE4RnBlZXJpbnRmMjQ1NzM3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '245737',
                                                                       'in_nucast_pkts': '71494496',
                                                                       'in_octets': '5564287166551',
                                                                       'in_ucast_pkts': '6187781673',
                                                                       'links': [       {       'in_nucast_pkts': '71494496',
                                                                                                'in_octets': '5564287166551',
                                                                                                'in_ucast_pkts': '6187781673',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1844',
                                                                                                'out_octets': '2087926991872',
                                                                                                'out_ucast_pkts': '4986827424',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1844',
                                                                       'out_octets': '2087926991872',
                                                                       'out_ucast_pkts': '4986827424',
                                                                       'poller_hash': 'MTU4RkE4RnBlZXJpbnRmcG9uIDI0NTczNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245737',
                                                                                               'links': [       {       'connected_time': '68954701',
                                                                                                                        'fiber_length': 6393,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.866,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU4RkE4RnBlZXJwb25wb24gMjQ1NzM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '4478 PERTH LINE 72 (NEW STAR '
                                                 'RENTALS&EQUIPMENT LTD)',
                                         'poller_hash': 'MTU4RkE4Rg==',
                                         'serial_number': '158FA8F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '1c:8b:76:00:30:c9',
                                         'uptime': '68954701'},
                                 {       'counter_source': '245857/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '245857/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '127129670',
                                                                       'in_ucast_pkts': '375689',
                                                                       'mac_address': '04:bc:9f:5b:d9:11',
                                                                       'name': '245857/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '25',
                                                                       'out_octets': '10900815448',
                                                                       'out_ucast_pkts': '48912134',
                                                                       'poller_hash': 'MTI3NDFBMnBlZXJpbnRmMjQ1ODU3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '245857',
                                                                       'in_nucast_pkts': '25',
                                                                       'in_octets': '10900815448',
                                                                       'in_ucast_pkts': '48912134',
                                                                       'links': [       {       'in_nucast_pkts': '25',
                                                                                                'in_octets': '10900815448',
                                                                                                'in_ucast_pkts': '48912134',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '127129670',
                                                                                                'out_ucast_pkts': '375689',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '127129670',
                                                                       'out_ucast_pkts': '375689',
                                                                       'poller_hash': 'MTI3NDFBMnBlZXJpbnRmcG9uIDI0NTg1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245857',
                                                                                               'links': [       {       'connected_time': '68869275',
                                                                                                                        'fiber_length': 13858,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFBMnBlZXJwb25wb24gMjQ1ODU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7285 PERTH ROAD 121 (ALBRECHT LLOYD)',
                                         'poller_hash': 'MTI3NDFBMg==',
                                         'serial_number': '12741A2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:11',
                                         'uptime': '68869275'},
                                 {       'counter_source': '246417/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '246417/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '55862141355',
                                                                       'in_ucast_pkts': '127036306',
                                                                       'mac_address': '04:bc:9f:5b:d9:79',
                                                                       'name': '246417/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '266301608007',
                                                                       'out_ucast_pkts': '264736115',
                                                                       'poller_hash': 'MTI3NDFBRnBlZXJpbnRmMjQ2NDE3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '246417',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '266301608007',
                                                                       'in_ucast_pkts': '264736115',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '266301608007',
                                                                                                'in_ucast_pkts': '264736115',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '55862141355',
                                                                                                'out_ucast_pkts': '127036306',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '55862141355',
                                                                       'out_ucast_pkts': '127036306',
                                                                       'poller_hash': 'MTI3NDFBRnBlZXJpbnRmcG9uIDI0NjQxNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246417',
                                                                                               'links': [       {       'connected_time': '68797824',
                                                                                                                        'fiber_length': 9844,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFBRnBlZXJwb25wb24gMjQ2NDE3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6475 PERTH ROAD 121 (POOLE MENNONITE '
                                                 'CHURCH)',
                                         'poller_hash': 'MTI3NDFBRg==',
                                         'serial_number': '12741AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:79',
                                         'uptime': '68797824'},
                                 {       'counter_source': '246513/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ed:2b'],
                                                                       'id': '246513/g1',
                                                                       'in_nucast_pkts': '608827',
                                                                       'in_octets': '120156754603',
                                                                       'in_ucast_pkts': '1106865886',
                                                                       'mac_address': '04:bc:9f:5b:f8:9c',
                                                                       'mtu': 2000,
                                                                       'name': '246513/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66282161',
                                                                       'out_octets': '6270750813097',
                                                                       'out_ucast_pkts': '4226265955',
                                                                       'poller_hash': 'MTI3NDU5M3BlZXJpbnRmMjQ2NTEzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '246513',
                                                                       'in_nucast_pkts': '66282161',
                                                                       'in_octets': '6270750813097',
                                                                       'in_ucast_pkts': '4226265955',
                                                                       'links': [       {       'in_nucast_pkts': '66282161',
                                                                                                'in_octets': '6270750813097',
                                                                                                'in_ucast_pkts': '4226265955',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '608827',
                                                                                                'out_octets': '120156754603',
                                                                                                'out_ucast_pkts': '1106865886',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '608827',
                                                                       'out_octets': '120156754603',
                                                                       'out_ucast_pkts': '1106865886',
                                                                       'poller_hash': 'MTI3NDU5M3BlZXJpbnRmcG9uIDI0NjUxMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246513',
                                                                                               'links': [       {       'connected_time': '68532133',
                                                                                                                        'fiber_length': 10424,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDU5M3BlZXJwb25wb24gMjQ2NTEz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5454 ROAD 126 (NEFF ALLEN & IRMA)',
                                         'poller_hash': 'MTI3NDU5Mw==',
                                         'serial_number': '1274593',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:f8:99',
                                         'uptime': '68532133'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '246561/g1',
                                                                       'name': '246561/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDA3QXBlZXJpbnRmMjQ2NTYxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '246561',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDA3QXBlZXJpbnRmcG9uIDI0NjU2MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246561',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA3QXBlZXJwb25wb24gMjQ2NTYx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '55 WILLIAM ST (OWUSU SELINA)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "246561" '
                                                                  '(location: GPON '
                                                                  'port "7/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-06-28T21:03:22.690Z',
                                         'poller_hash': 'MTI3NDA3QQ==',
                                         'serial_number': '127407A',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '246571/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:70:6e'],
                                                                       'id': '246571/g1',
                                                                       'in_nucast_pkts': '670571',
                                                                       'in_octets': '1099345285121',
                                                                       'in_ucast_pkts': '3136493142',
                                                                       'mac_address': '04:bc:9f:5b:d3:6c',
                                                                       'mtu': 2000,
                                                                       'name': '246571/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '89055775',
                                                                       'out_octets': '12246326613504',
                                                                       'out_ucast_pkts': '9561568069',
                                                                       'poller_hash': 'MTI3NDBFRHBlZXJpbnRmMjQ2NTcxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '246571',
                                                                       'in_nucast_pkts': '89055775',
                                                                       'in_octets': '12246326613504',
                                                                       'in_ucast_pkts': '9561568069',
                                                                       'links': [       {       'in_nucast_pkts': '89055775',
                                                                                                'in_octets': '12246326613504',
                                                                                                'in_ucast_pkts': '9561568069',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '670571',
                                                                                                'out_octets': '1099345285121',
                                                                                                'out_ucast_pkts': '3136493142',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '670571',
                                                                       'out_octets': '1099345285121',
                                                                       'out_ucast_pkts': '3136493142',
                                                                       'poller_hash': 'MTI3NDBFRHBlZXJpbnRmcG9uIDI0NjU3MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246571',
                                                                                               'links': [       {       'connected_time': '68517793',
                                                                                                                        'fiber_length': 20771,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBFRHBlZXJwb25wb24gMjQ2NTcx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2288 LINE 45 (MCMILLAN ROBERT)',
                                         'poller_hash': 'MTI3NDBFRA==',
                                         'serial_number': '12740ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d3:69',
                                         'uptime': '68517793'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '246870/g1',
                                                                       'name': '246870/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDMzQXBlZXJpbnRmMjQ2ODcwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '246870',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDMzQXBlZXJpbnRmcG9uIDI0Njg3MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246870',
                                                                                               'links': [       {       'connected_time': '28172215',
                                                                                                                        'fiber_length': 11514,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMzQXBlZXJwb25wb24gMjQ2ODcw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4184 LINE 61 (FAIRHAVEN CH DAY '
                                                 'SCHOOL)',
                                         'poller_hash': 'MTI3NDMzQQ==',
                                         'serial_number': '127433A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:d1',
                                         'uptime': '28172215'},
                                 {       'counter_source': '246880/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '246880/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3770766415',
                                                                       'in_ucast_pkts': '3803495',
                                                                       'mac_address': '04:bc:9f:5b:cd:59',
                                                                       'name': '246880/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7685956476',
                                                                       'out_ucast_pkts': '23524100',
                                                                       'poller_hash': 'MTI3NDAyQnBlZXJpbnRmMjQ2ODgwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '246880',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7685956476',
                                                                       'in_ucast_pkts': '23524100',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7685956476',
                                                                                                'in_ucast_pkts': '23524100',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3770766415',
                                                                                                'out_ucast_pkts': '3803495',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3770766415',
                                                                       'out_ucast_pkts': '3803495',
                                                                       'poller_hash': 'MTI3NDAyQnBlZXJpbnRmcG9uIDI0Njg4MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246880',
                                                                                               'links': [       {       'connected_time': '68350753',
                                                                                                                        'fiber_length': 11599,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDAyQnBlZXJwb25wb24gMjQ2ODgw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4186 LINE 61 (FAIRHAVEN AMISH '
                                                 'MENNONITE CH)',
                                         'poller_hash': 'MTI3NDAyQg==',
                                         'serial_number': '127402B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cd:59',
                                         'uptime': '68350753'},
                                 {       'counter_source': '246900/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '7c:5a:1c:8b:fa:33'],
                                                                       'id': '246900/g1',
                                                                       'in_nucast_pkts': '6',
                                                                       'in_octets': '13055530373067',
                                                                       'in_ucast_pkts': '13609361751',
                                                                       'mac_address': '04:bc:9f:5b:cf:24',
                                                                       'mtu': 2000,
                                                                       'name': '246900/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35094860',
                                                                       'out_octets': '5799129896248',
                                                                       'out_ucast_pkts': '8737277792',
                                                                       'poller_hash': 'MTI3NDA2NHBlZXJpbnRmMjQ2OTAwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '246900',
                                                                       'in_nucast_pkts': '35094860',
                                                                       'in_octets': '5799129896248',
                                                                       'in_ucast_pkts': '8737277792',
                                                                       'links': [       {       'in_nucast_pkts': '35094860',
                                                                                                'in_octets': '5799129896248',
                                                                                                'in_ucast_pkts': '8737277792',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6',
                                                                                                'out_octets': '13055530373067',
                                                                                                'out_ucast_pkts': '13609361751',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6',
                                                                       'out_octets': '13055530373067',
                                                                       'out_ucast_pkts': '13609361751',
                                                                       'poller_hash': 'MTI3NDA2NHBlZXJpbnRmcG9uIDI0NjkwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246900',
                                                                                               'links': [       {       'connected_time': '15498714',
                                                                                                                        'fiber_length': 11849,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA2NHBlZXJwb25wb24gMjQ2OTAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4078 WATERLOO ST (PRECISION '
                                                 'WOODCRAFT)',
                                         'poller_hash': 'MTI3NDA2NA==',
                                         'serial_number': '1274064',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cf:21',
                                         'uptime': '15498714'},
                                 {       'counter_source': '246910/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '246910/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7648479679',
                                                                       'in_ucast_pkts': '52996418',
                                                                       'mac_address': '04:bc:9f:5b:d1:c1',
                                                                       'name': '246910/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '18355699777',
                                                                       'out_ucast_pkts': '42298408',
                                                                       'poller_hash': 'MTI3NDBCOHBlZXJpbnRmMjQ2OTEwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '246910',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '18355699777',
                                                                       'in_ucast_pkts': '42298408',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '18355699777',
                                                                                                'in_ucast_pkts': '42298408',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '7648479679',
                                                                                                'out_ucast_pkts': '52996418',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7648479679',
                                                                       'out_ucast_pkts': '52996418',
                                                                       'poller_hash': 'MTI3NDBCOHBlZXJpbnRmcG9uIDI0NjkxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246910',
                                                                                               'links': [       {       'connected_time': '688505',
                                                                                                                        'fiber_length': 1365,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBCOHBlZXJwb25wb24gMjQ2OTEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 MAIN ST N APT 6 MILVERTON',
                                         'poller_hash': 'MTI3NDBCOA==',
                                         'serial_number': '12740B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d1:c1',
                                         'uptime': '688505'},
                                 {       'counter_source': '246930/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '34:56:fe:2c:cc:ab'],
                                                                       'id': '246930/g1',
                                                                       'in_nucast_pkts': '1669',
                                                                       'in_octets': '365988922565',
                                                                       'in_ucast_pkts': '361852727',
                                                                       'mac_address': 'd0:76:8f:1b:a6:b6',
                                                                       'mtu': 2000,
                                                                       'name': '246930/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1097560',
                                                                       'out_octets': '266929717615',
                                                                       'out_ucast_pkts': '324896276',
                                                                       'poller_hash': 'ODcyNTI0cGVlcmludGYyNDY5MzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '246930',
                                                                       'in_nucast_pkts': '1097560',
                                                                       'in_octets': '266929717615',
                                                                       'in_ucast_pkts': '324896276',
                                                                       'links': [       {       'in_nucast_pkts': '1097560',
                                                                                                'in_octets': '266929717615',
                                                                                                'in_ucast_pkts': '324896276',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1669',
                                                                                                'out_octets': '365988922565',
                                                                                                'out_ucast_pkts': '361852727',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1669',
                                                                       'out_octets': '365988922565',
                                                                       'out_ucast_pkts': '361852727',
                                                                       'poller_hash': 'ODcyNTI0cGVlcmludGZwb24gMjQ2OTMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246930',
                                                                                               'links': [       {       'connected_time': '513271',
                                                                                                                        'fiber_length': 6387,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTI0cGVlcnBvbnBvbiAyNDY5MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4481 PERTH LINE 72 NEWTON (MITCHELL '
                                                 'MILL SYSTEMS)',
                                         'poller_hash': 'ODcyNTI0',
                                         'serial_number': '872524',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a6:b3',
                                         'uptime': '513271'},
                                 {       'counter_source': '246970/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:68:bd'],
                                                                       'id': '246970/g1',
                                                                       'in_nucast_pkts': '14614',
                                                                       'in_octets': '1610199115',
                                                                       'in_ucast_pkts': '11337651',
                                                                       'mac_address': '04:bc:9f:5b:d1:0c',
                                                                       'mtu': 2000,
                                                                       'name': '246970/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465307',
                                                                       'out_octets': '62687067123',
                                                                       'out_ucast_pkts': '49071074',
                                                                       'poller_hash': 'MTI3NDBBMXBlZXJpbnRmMjQ2OTcwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '246970',
                                                                       'in_nucast_pkts': '1465307',
                                                                       'in_octets': '62687067123',
                                                                       'in_ucast_pkts': '49071074',
                                                                       'links': [       {       'in_nucast_pkts': '1465307',
                                                                                                'in_octets': '62687067123',
                                                                                                'in_ucast_pkts': '49071074',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14614',
                                                                                                'out_octets': '1610199115',
                                                                                                'out_ucast_pkts': '11337651',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14614',
                                                                       'out_octets': '1610199115',
                                                                       'out_ucast_pkts': '11337651',
                                                                       'poller_hash': 'MTI3NDBBMXBlZXJpbnRmcG9uIDI0Njk3MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246970',
                                                                                               'links': [       {       'connected_time': '688587',
                                                                                                                        'fiber_length': 5454,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBBMXBlZXJwb25wb24gMjQ2OTcw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4575 PERTH LINE 72 (WYATT JUDY)',
                                         'poller_hash': 'MTI3NDBBMQ==',
                                         'serial_number': '12740A1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d1:09',
                                         'uptime': '688587'},
                                 {       'counter_source': '246990/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6f:66'],
                                                                       'id': '246990/g1',
                                                                       'in_nucast_pkts': '657174',
                                                                       'in_octets': '248811250892',
                                                                       'in_ucast_pkts': '561365950',
                                                                       'mac_address': '04:bc:9f:5b:e5:9c',
                                                                       'mtu': 2000,
                                                                       'name': '246990/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83664576',
                                                                       'out_octets': '1571349737545',
                                                                       'out_ucast_pkts': '1351286858',
                                                                       'poller_hash': 'MTI3NDMzM3BlZXJpbnRmMjQ2OTkwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '246990',
                                                                       'in_nucast_pkts': '83664576',
                                                                       'in_octets': '1571349737545',
                                                                       'in_ucast_pkts': '1351286858',
                                                                       'links': [       {       'in_nucast_pkts': '83664576',
                                                                                                'in_octets': '1571349737545',
                                                                                                'in_ucast_pkts': '1351286858',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '657174',
                                                                                                'out_octets': '248811250892',
                                                                                                'out_ucast_pkts': '561365950',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '657174',
                                                                       'out_octets': '248811250892',
                                                                       'out_ucast_pkts': '561365950',
                                                                       'poller_hash': 'MTI3NDMzM3BlZXJpbnRmcG9uIDI0Njk5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246990',
                                                                                               'links': [       {       'connected_time': '67911671',
                                                                                                                        'fiber_length': 11636,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMzM3BlZXJwb25wb24gMjQ2OTkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5964 PERTH ROAD 119 (GERBER JOSEPH)',
                                         'poller_hash': 'MTI3NDMzMw==',
                                         'serial_number': '1274333',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:99',
                                         'uptime': '67911671'},
                                 {       'counter_source': '247105/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6a:07'],
                                                                       'id': '247105/g1',
                                                                       'in_nucast_pkts': '676913',
                                                                       'in_octets': '604907484750',
                                                                       'in_ucast_pkts': '3258463965',
                                                                       'mac_address': '04:bc:9f:5b:e4:4c',
                                                                       'mtu': 2000,
                                                                       'name': '247105/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '83705938',
                                                                       'out_octets': '11271252493136',
                                                                       'out_ucast_pkts': '7963987273',
                                                                       'poller_hash': 'MTI3NDMwOXBlZXJpbnRmMjQ3MTA1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '247105',
                                                                       'in_nucast_pkts': '83705938',
                                                                       'in_octets': '11271252493136',
                                                                       'in_ucast_pkts': '7963987273',
                                                                       'links': [       {       'in_nucast_pkts': '83705938',
                                                                                                'in_octets': '11271252493136',
                                                                                                'in_ucast_pkts': '7963987273',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '676913',
                                                                                                'out_octets': '604907484750',
                                                                                                'out_ucast_pkts': '3258463965',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '676913',
                                                                       'out_octets': '604907484750',
                                                                       'out_ucast_pkts': '3258463965',
                                                                       'poller_hash': 'MTI3NDMwOXBlZXJpbnRmcG9uIDI0NzEwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247105',
                                                                                               'links': [       {       'connected_time': '67830839',
                                                                                                                        'fiber_length': 14222,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMwOXBlZXJwb25wb24gMjQ3MTA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6194 ROAD 116 (WAGLER RUTH)',
                                         'poller_hash': 'MTI3NDMwOQ==',
                                         'serial_number': '1274309',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e4:49',
                                         'uptime': '67830839'},
                                 {       'counter_source': '247115/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '247115/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '81699210715',
                                                                       'in_ucast_pkts': '597133113',
                                                                       'mac_address': '04:bc:9f:5b:ce:f9',
                                                                       'name': '247115/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1',
                                                                       'out_octets': '2894855343354',
                                                                       'out_ucast_pkts': '2031580405',
                                                                       'poller_hash': 'MTI3NDA1RnBlZXJpbnRmMjQ3MTE1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '247115',
                                                                       'in_nucast_pkts': '1',
                                                                       'in_octets': '2894855343354',
                                                                       'in_ucast_pkts': '2031580405',
                                                                       'links': [       {       'in_nucast_pkts': '1',
                                                                                                'in_octets': '2894855343354',
                                                                                                'in_ucast_pkts': '2031580405',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '81699210715',
                                                                                                'out_ucast_pkts': '597133113',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '81699210715',
                                                                       'out_ucast_pkts': '597133113',
                                                                       'poller_hash': 'MTI3NDA1RnBlZXJpbnRmcG9uIDI0NzExNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247115',
                                                                                               'links': [       {       'connected_time': '688361',
                                                                                                                        'fiber_length': 2752,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA1RnBlZXJwb25wb24gMjQ3MTE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '69 PUGH ST MILVERTON (SHANTZ '
                                                 'TIFFANY)',
                                         'poller_hash': 'MTI3NDA1Rg==',
                                         'serial_number': '127405F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:f9',
                                         'uptime': '688361'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '247135/g1',
                                                                       'name': '247135/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDA3MHBlZXJpbnRmMjQ3MTM1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '247135',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDA3MHBlZXJpbnRmcG9uIDI0NzEzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247135',
                                                                                               'links': [       {       'connected_time': '67752657',
                                                                                                                        'fiber_length': 1610,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA3MHBlZXJwb25wb24gMjQ3MTM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '33 MURRAY CT (SOUTHSIDE ESTATES INC)',
                                         'poller_hash': 'MTI3NDA3MA==',
                                         'serial_number': '1274070',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cf:81',
                                         'uptime': '67752657'},
                                 {       'counter_source': '247754/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:7a:24'],
                                                                       'id': '247754/g1',
                                                                       'in_nucast_pkts': '310230',
                                                                       'in_octets': '278465944530',
                                                                       'in_ucast_pkts': '1237054984',
                                                                       'mac_address': '04:bc:9f:5b:d8:f4',
                                                                       'mtu': 2000,
                                                                       'name': '247754/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35084752',
                                                                       'out_octets': '8017314721521',
                                                                       'out_ucast_pkts': '5795179460',
                                                                       'poller_hash': 'MTI3NDE5RXBlZXJpbnRmMjQ3NzU0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '247754',
                                                                       'in_nucast_pkts': '35084752',
                                                                       'in_octets': '8017314721521',
                                                                       'in_ucast_pkts': '5795179460',
                                                                       'links': [       {       'in_nucast_pkts': '35084752',
                                                                                                'in_octets': '8017314721521',
                                                                                                'in_ucast_pkts': '5795179460',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '310230',
                                                                                                'out_octets': '278465944530',
                                                                                                'out_ucast_pkts': '1237054984',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '310230',
                                                                       'out_octets': '278465944530',
                                                                       'out_ucast_pkts': '1237054984',
                                                                       'poller_hash': 'MTI3NDE5RXBlZXJpbnRmcG9uIDI0Nzc1NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247754',
                                                                                               'links': [       {       'connected_time': '15498571',
                                                                                                                        'fiber_length': 2592,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5RXBlZXJwb25wb24gMjQ3NzU0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '113 PUGH ST (MAYNE STEPH)',
                                         'poller_hash': 'MTI3NDE5RQ==',
                                         'serial_number': '127419E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:f1',
                                         'uptime': '15498571'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '247784/g1',
                                                                       'name': '247784/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEE0MHBlZXJpbnRmMjQ3Nzg0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '247784',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEE0MHBlZXJpbnRmcG9uIDI0Nzc4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247784',
                                                                                               'links': [       {       'connected_time': '23104090',
                                                                                                                        'fiber_length': 16132,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE0MHBlZXJwb25wb24gMjQ3Nzg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7455 PERTH ROAD 121 (KUEPFER MURRAY '
                                                 '& ELIZABETH)',
                                         'poller_hash': 'MTI3NEE0MA==',
                                         'serial_number': '1274A40',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1e:01',
                                         'uptime': '23104090'},
                                 {       'counter_source': '248038/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '248038/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '293',
                                                                       'in_octets': '1062250815533',
                                                                       'in_ucast_pkts': '75373616',
                                                                       'mac_address': '04:bc:9f:5b:e3:39',
                                                                       'name': '248038/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1942',
                                                                       'out_octets': '18196212933459',
                                                                       'out_ucast_pkts': '14554307805',
                                                                       'poller_hash': 'MTI3NDJFN3BlZXJpbnRmMjQ4MDM4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '248038',
                                                                       'in_nucast_pkts': '1942',
                                                                       'in_octets': '18196212933459',
                                                                       'in_ucast_pkts': '14554307805',
                                                                       'links': [       {       'in_nucast_pkts': '1942',
                                                                                                'in_octets': '18196212933459',
                                                                                                'in_ucast_pkts': '14554307805',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '293',
                                                                                                'out_octets': '1062250815533',
                                                                                                'out_ucast_pkts': '75373616',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '293',
                                                                       'out_octets': '1062250815533',
                                                                       'out_ucast_pkts': '75373616',
                                                                       'poller_hash': 'MTI3NDJFN3BlZXJpbnRmcG9uIDI0ODAzOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248038',
                                                                                               'links': [       {       'connected_time': '688457',
                                                                                                                        'fiber_length': 1565,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -8.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJFN3BlZXJwb25wb24gMjQ4MDM4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '80 MAIN ST N (GOFTON LAURA)',
                                         'poller_hash': 'MTI3NDJFNw==',
                                         'serial_number': '12742E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:39',
                                         'uptime': '688457'},
                                 {       'counter_source': '248088/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '248088/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '86063895666',
                                                                       'in_ucast_pkts': '1086833623',
                                                                       'mac_address': '04:bc:9f:5c:1d:f9',
                                                                       'name': '248088/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '25',
                                                                       'out_octets': '6944086628284',
                                                                       'out_ucast_pkts': '321015109',
                                                                       'poller_hash': 'MTI3NEEzRnBlZXJpbnRmMjQ4MDg4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '248088',
                                                                       'in_nucast_pkts': '25',
                                                                       'in_octets': '6944086628284',
                                                                       'in_ucast_pkts': '321015109',
                                                                       'links': [       {       'in_nucast_pkts': '25',
                                                                                                'in_octets': '6944086628284',
                                                                                                'in_ucast_pkts': '321015109',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '86063895666',
                                                                                                'out_ucast_pkts': '1086833623',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '86063895666',
                                                                       'out_ucast_pkts': '1086833623',
                                                                       'poller_hash': 'MTI3NEEzRnBlZXJpbnRmcG9uIDI0ODA4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248088',
                                                                                               'links': [       {       'connected_time': '67051846',
                                                                                                                        'fiber_length': 17265,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEzRnBlZXJwb25wb24gMjQ4MDg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4097 LINE 50 (NEEB RICHARD)',
                                         'poller_hash': 'MTI3NEEzRg==',
                                         'serial_number': '1274A3F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:f9',
                                         'uptime': '67051846'},
                                 {       'counter_source': '248148/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '248148/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '142550186599',
                                                                       'in_ucast_pkts': '973956642',
                                                                       'mac_address': '04:bc:9f:5b:e5:01',
                                                                       'name': '248148/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '29',
                                                                       'out_octets': '4200341330739',
                                                                       'out_ucast_pkts': '3035075336',
                                                                       'poller_hash': 'MTI3NDMyMHBlZXJpbnRmMjQ4MTQ4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '248148',
                                                                       'in_nucast_pkts': '29',
                                                                       'in_octets': '4200341330739',
                                                                       'in_ucast_pkts': '3035075336',
                                                                       'links': [       {       'in_nucast_pkts': '29',
                                                                                                'in_octets': '4200341330739',
                                                                                                'in_ucast_pkts': '3035075336',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '142550186599',
                                                                                                'out_ucast_pkts': '973956642',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '142550186599',
                                                                       'out_ucast_pkts': '973956642',
                                                                       'poller_hash': 'MTI3NDMyMHBlZXJpbnRmcG9uIDI0ODE0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248148',
                                                                                               'links': [       {       'connected_time': '21472',
                                                                                                                        'fiber_length': 21244,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMyMHBlZXJwb25wb24gMjQ4MTQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4234 LINE 46 (CATANIA JACKIE)',
                                         'poller_hash': 'MTI3NDMyMA==',
                                         'serial_number': '1274320',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:01',
                                         'uptime': '21472'},
                                 {       'counter_source': '248410/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e7:42'],
                                                                       'id': '248410/g1',
                                                                       'in_nucast_pkts': '14625',
                                                                       'in_octets': '4362335716',
                                                                       'in_ucast_pkts': '34159094',
                                                                       'mac_address': '04:bc:9f:5c:19:a4',
                                                                       'mtu': 2000,
                                                                       'name': '248410/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465233',
                                                                       'out_octets': '128420864762',
                                                                       'out_ucast_pkts': '92747116',
                                                                       'poller_hash': 'MTI3NDlCNHBlZXJpbnRmMjQ4NDEwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248410',
                                                                       'in_nucast_pkts': '1465233',
                                                                       'in_octets': '128420864762',
                                                                       'in_ucast_pkts': '92747116',
                                                                       'links': [       {       'in_nucast_pkts': '1465233',
                                                                                                'in_octets': '128420864762',
                                                                                                'in_ucast_pkts': '92747116',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14625',
                                                                                                'out_octets': '4362335716',
                                                                                                'out_ucast_pkts': '34159094',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14625',
                                                                       'out_octets': '4362335716',
                                                                       'out_ucast_pkts': '34159094',
                                                                       'poller_hash': 'MTI3NDlCNHBlZXJpbnRmcG9uIDI0ODQxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248410',
                                                                                               'links': [       {       'connected_time': '688540',
                                                                                                                        'fiber_length': 10160,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlCNHBlZXJwb25wb24gMjQ4NDEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5913 PERTH ROAD 119 (THOMPSON STACY)',
                                         'poller_hash': 'MTI3NDlCNA==',
                                         'serial_number': '12749B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:19:a1',
                                         'uptime': '688540'},
                                 {       'counter_source': '248483/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '248483/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3927579314',
                                                                       'in_ucast_pkts': '32492765',
                                                                       'mac_address': '04:bc:9f:5b:e4:79',
                                                                       'name': '248483/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '53117209369',
                                                                       'out_ucast_pkts': '74386473',
                                                                       'poller_hash': 'MTI3NDMwRnBlZXJpbnRmMjQ4NDgzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '248483',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '53117209369',
                                                                       'in_ucast_pkts': '74386473',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '53117209369',
                                                                                                'in_ucast_pkts': '74386473',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3927579314',
                                                                                                'out_ucast_pkts': '32492765',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3927579314',
                                                                       'out_ucast_pkts': '32492765',
                                                                       'poller_hash': 'MTI3NDMwRnBlZXJpbnRmcG9uIDI0ODQ4Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248483',
                                                                                               'links': [       {       'connected_time': '66619055',
                                                                                                                        'fiber_length': 4840,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMwRnBlZXJwb25wb24gMjQ4NDgz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6799 ROAD 129 (JANTZI LEVI K)',
                                         'poller_hash': 'MTI3NDMwRg==',
                                         'serial_number': '127430F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e4:79',
                                         'uptime': '66619055'},
                                 {       'counter_source': '248523/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1d:22'],
                                                                       'id': '248523/g1',
                                                                       'in_nucast_pkts': '14512',
                                                                       'in_octets': '36232131643',
                                                                       'in_ucast_pkts': '135294432',
                                                                       'mac_address': '04:bc:9f:5b:6f:9c',
                                                                       'mtu': 2000,
                                                                       'name': '248523/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1448053',
                                                                       'out_octets': '559168172087',
                                                                       'out_ucast_pkts': '409609394',
                                                                       'poller_hash': 'MTI3MzQ3M3BlZXJpbnRmMjQ4NTIzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248523',
                                                                       'in_nucast_pkts': '1448053',
                                                                       'in_octets': '559168172087',
                                                                       'in_ucast_pkts': '409609394',
                                                                       'links': [       {       'in_nucast_pkts': '1448053',
                                                                                                'in_octets': '559168172087',
                                                                                                'in_ucast_pkts': '409609394',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14512',
                                                                                                'out_octets': '36232131643',
                                                                                                'out_ucast_pkts': '135294432',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14512',
                                                                       'out_octets': '36232131643',
                                                                       'out_ucast_pkts': '135294432',
                                                                       'poller_hash': 'MTI3MzQ3M3BlZXJpbnRmcG9uIDI0ODUyMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248523',
                                                                                               'links': [       {       'connected_time': '688413',
                                                                                                                        'fiber_length': 2096,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MzQ3M3BlZXJwb25wb24gMjQ4NTIz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '82 GERBER DR (ANANTHAN SUBATHA)',
                                         'poller_hash': 'MTI3MzQ3Mw==',
                                         'serial_number': '1273473',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:6f:99',
                                         'uptime': '688413'},
                                 {       'counter_source': '248533/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '248533/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '170099689483',
                                                                       'in_ucast_pkts': '371480380',
                                                                       'mac_address': '04:bc:9f:5b:cf:39',
                                                                       'name': '248533/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '50',
                                                                       'out_octets': '780269856160',
                                                                       'out_ucast_pkts': '730026288',
                                                                       'poller_hash': 'MTI3NDA2N3BlZXJpbnRmMjQ4NTMzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '248533',
                                                                       'in_nucast_pkts': '50',
                                                                       'in_octets': '780269856160',
                                                                       'in_ucast_pkts': '730026288',
                                                                       'links': [       {       'in_nucast_pkts': '50',
                                                                                                'in_octets': '780269856160',
                                                                                                'in_ucast_pkts': '730026288',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '170099689483',
                                                                                                'out_ucast_pkts': '371480380',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '170099689483',
                                                                       'out_ucast_pkts': '371480380',
                                                                       'poller_hash': 'MTI3NDA2N3BlZXJpbnRmcG9uIDI0ODUzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248533',
                                                                                               'links': [       {       'connected_time': '66531890',
                                                                                                                        'fiber_length': 11265,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA2N3BlZXJwb25wb24gMjQ4NTMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7035 PERTH ROAD 121 (SCHULTZ DELMAR '
                                                 '& JANET)',
                                         'poller_hash': 'MTI3NDA2Nw==',
                                         'serial_number': '1274067',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cf:39',
                                         'uptime': '66531890'},
                                 {       'counter_source': '248615/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:32:e5:b2:c9:cd'],
                                                                       'id': '248615/g1',
                                                                       'in_nucast_pkts': '326788',
                                                                       'in_octets': '1277086333',
                                                                       'in_ucast_pkts': '3775239',
                                                                       'mac_address': '04:bc:9f:5c:1d:34',
                                                                       'mtu': 2000,
                                                                       'name': '248615/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151573',
                                                                       'out_octets': '12282175235',
                                                                       'out_ucast_pkts': '10632653',
                                                                       'poller_hash': 'MTI3NEEyNnBlZXJpbnRmMjQ4NjE1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248615',
                                                                       'in_nucast_pkts': '1151573',
                                                                       'in_octets': '12282175235',
                                                                       'in_ucast_pkts': '10632653',
                                                                       'links': [       {       'in_nucast_pkts': '1151573',
                                                                                                'in_octets': '12282175235',
                                                                                                'in_ucast_pkts': '10632653',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '326788',
                                                                                                'out_octets': '1277086333',
                                                                                                'out_ucast_pkts': '3775239',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '326788',
                                                                       'out_octets': '1277086333',
                                                                       'out_ucast_pkts': '3775239',
                                                                       'poller_hash': 'MTI3NEEyNnBlZXJpbnRmcG9uIDI0ODYxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248615',
                                                                                               'links': [       {       'connected_time': '538592',
                                                                                                                        'fiber_length': 15042,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEyNnBlZXJwb25wb24gMjQ4NjE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2541 LINE 54 (ERB HAROLD)',
                                         'poller_hash': 'MTI3NEEyNg==',
                                         'serial_number': '1274A26',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:31',
                                         'uptime': '538592'},
                                 {       'counter_source': '248635/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d2:67'],
                                                                       'id': '248635/g1',
                                                                       'in_nucast_pkts': '330',
                                                                       'in_octets': '46027000764',
                                                                       'in_ucast_pkts': '542274885',
                                                                       'mac_address': '04:bc:9f:5b:d2:ec',
                                                                       'mtu': 2000,
                                                                       'name': '248635/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '28821846',
                                                                       'out_octets': '2298904822333',
                                                                       'out_ucast_pkts': '1537610439',
                                                                       'poller_hash': 'MTI3NDBERHBlZXJpbnRmMjQ4NjM1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248635',
                                                                       'in_nucast_pkts': '28821846',
                                                                       'in_octets': '2298904822333',
                                                                       'in_ucast_pkts': '1537610439',
                                                                       'links': [       {       'in_nucast_pkts': '28821846',
                                                                                                'in_octets': '2298904822333',
                                                                                                'in_ucast_pkts': '1537610439',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '330',
                                                                                                'out_octets': '46027000764',
                                                                                                'out_ucast_pkts': '542274885',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '330',
                                                                       'out_octets': '46027000764',
                                                                       'out_ucast_pkts': '542274885',
                                                                       'poller_hash': 'MTI3NDBERHBlZXJpbnRmcG9uIDI0ODYzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248635',
                                                                                               'links': [       {       'connected_time': '12530065',
                                                                                                                        'fiber_length': 17357,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBERHBlZXJwb25wb24gMjQ4NjM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5225 ROAD 130 (STEPHENSON BOB)',
                                         'poller_hash': 'MTI3NDBERA==',
                                         'serial_number': '12740DD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d2:e9',
                                         'uptime': '12530065'},
                                 {       'counter_source': '249033/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249033/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25791958147',
                                                                       'in_ucast_pkts': '174277527',
                                                                       'mac_address': '04:bc:9f:5b:e5:b1',
                                                                       'name': '249033/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '982855794984',
                                                                       'out_ucast_pkts': '724702274',
                                                                       'poller_hash': 'MTI3NDMzNnBlZXJpbnRmMjQ5MDMzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249033',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '982855794984',
                                                                       'in_ucast_pkts': '724702274',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '982855794984',
                                                                                                'in_ucast_pkts': '724702274',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '25791958147',
                                                                                                'out_ucast_pkts': '174277527',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25791958147',
                                                                       'out_ucast_pkts': '174277527',
                                                                       'poller_hash': 'MTI3NDMzNnBlZXJpbnRmcG9uIDI0OTAzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249033',
                                                                                               'links': [       {       'connected_time': '523338',
                                                                                                                        'fiber_length': 11219,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMzNnBlZXJwb25wb24gMjQ5MDMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7027 PERTH ROAD 121 (TERRY-ROSS '
                                                 'ROBIN)',
                                         'poller_hash': 'MTI3NDMzNg==',
                                         'serial_number': '1274336',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:b1',
                                         'uptime': '523338'},
                                 {       'counter_source': '249083/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249083/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '220617927289',
                                                                       'in_ucast_pkts': '1088862985',
                                                                       'mac_address': '04:bc:9f:5c:1f:c9',
                                                                       'name': '249083/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4081011280258',
                                                                       'out_ucast_pkts': '3034352720',
                                                                       'poller_hash': 'MTI3NEE3OXBlZXJpbnRmMjQ5MDgzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249083',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4081011280258',
                                                                       'in_ucast_pkts': '3034352720',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4081011280258',
                                                                                                'in_ucast_pkts': '3034352720',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '220617927289',
                                                                                                'out_ucast_pkts': '1088862985',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '220617927289',
                                                                       'out_ucast_pkts': '1088862985',
                                                                       'poller_hash': 'MTI3NEE3OXBlZXJpbnRmcG9uIDI0OTA4Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249083',
                                                                                               'links': [       {       'connected_time': '688571',
                                                                                                                        'fiber_length': 5230,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE3OXBlZXJwb25wb24gMjQ5MDgz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4464 LINE 64 (MCNALLY SHANE)',
                                         'poller_hash': 'MTI3NEE3OQ==',
                                         'serial_number': '1274A79',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1f:c9',
                                         'uptime': '688571'},
                                 {       'counter_source': '249093/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249093/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '219567260068',
                                                                       'in_ucast_pkts': '64831066406',
                                                                       'mac_address': '04:bc:9f:5c:03:09',
                                                                       'name': '249093/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '904036045806',
                                                                       'out_ucast_pkts': '787183517',
                                                                       'poller_hash': 'MTI3NDZFMXBlZXJpbnRmMjQ5MDkzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249093',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '904036045806',
                                                                       'in_ucast_pkts': '787183517',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '904036045806',
                                                                                                'in_ucast_pkts': '787183517',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '219567260068',
                                                                                                'out_ucast_pkts': '64831066406',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '219567260068',
                                                                       'out_ucast_pkts': '64831066406',
                                                                       'poller_hash': 'MTI3NDZFMXBlZXJpbnRmcG9uIDI0OTA5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249093',
                                                                                               'links': [       {       'connected_time': '688587',
                                                                                                                        'fiber_length': 10517,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDZFMXBlZXJwb25wb24gMjQ5MDkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4113 LINE 61 (JANTZI GABRIEL)',
                                         'poller_hash': 'MTI3NDZFMQ==',
                                         'serial_number': '12746E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:03:09',
                                         'uptime': '688587'},
                                 {       'counter_source': '249123/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249123/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '115246062940',
                                                                       'in_ucast_pkts': '30588253597',
                                                                       'mac_address': 'f8:85:f9:b4:ed:46',
                                                                       'name': '249123/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '25',
                                                                       'out_octets': '2190163003676',
                                                                       'out_ucast_pkts': '1684268217',
                                                                       'poller_hash': 'MTBBODZBQnBlZXJpbnRmMjQ5MTIzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249123',
                                                                       'in_nucast_pkts': '25',
                                                                       'in_octets': '2190163003676',
                                                                       'in_ucast_pkts': '1684268217',
                                                                       'links': [       {       'in_nucast_pkts': '25',
                                                                                                'in_octets': '2190163003676',
                                                                                                'in_ucast_pkts': '1684268217',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '115246062940',
                                                                                                'out_ucast_pkts': '30588253597',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '115246062940',
                                                                       'out_ucast_pkts': '30588253597',
                                                                       'poller_hash': 'MTBBODZBQnBlZXJpbnRmcG9uIDI0OTEyMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249123',
                                                                                               'links': [       {       'connected_time': '65768713',
                                                                                                                        'fiber_length': 12029,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODZBQnBlZXJwb25wb24gMjQ5MTIz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7106 PERTH ROAD 121 (SCHULTZ DALE)',
                                         'poller_hash': 'MTBBODZBQg==',
                                         'serial_number': '10A86AB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:ed:46',
                                         'uptime': '65768713'},
                                 {       'counter_source': '249183/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a0:88:95'],
                                                                       'id': '249183/g1',
                                                                       'in_nucast_pkts': '5222',
                                                                       'in_octets': '219966925',
                                                                       'in_ucast_pkts': '655951',
                                                                       'mac_address': 'f8:85:f9:b4:ea:11',
                                                                       'mtu': 2000,
                                                                       'name': '249183/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '541459',
                                                                       'out_octets': '7717400281',
                                                                       'out_ucast_pkts': '5360167',
                                                                       'poller_hash': 'MTBBODY0NHBlZXJpbnRmMjQ5MTgzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '249183',
                                                                       'in_nucast_pkts': '541459',
                                                                       'in_octets': '7717400281',
                                                                       'in_ucast_pkts': '5360167',
                                                                       'links': [       {       'in_nucast_pkts': '541459',
                                                                                                'in_octets': '7717400281',
                                                                                                'in_ucast_pkts': '5360167',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5222',
                                                                                                'out_octets': '219966925',
                                                                                                'out_ucast_pkts': '655951',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5222',
                                                                       'out_octets': '219966925',
                                                                       'out_ucast_pkts': '655951',
                                                                       'poller_hash': 'MTBBODY0NHBlZXJpbnRmcG9uIDI0OTE4Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249183',
                                                                                               'links': [       {       'connected_time': '262468',
                                                                                                                        'fiber_length': 18029,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODY0NHBlZXJwb25wb24gMjQ5MTgz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2288 LINE 47 (BALFOUR WENDY F)',
                                         'poller_hash': 'MTBBODY0NA==',
                                         'serial_number': '10A8644',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:ea:0e',
                                         'uptime': '262468'},
                                 {       'counter_source': '249203/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:e9:c6'],
                                                                       'id': '249203/g1',
                                                                       'in_nucast_pkts': '9760',
                                                                       'in_octets': '4458023950',
                                                                       'in_ucast_pkts': '20010700',
                                                                       'mac_address': 'b8:94:70:a1:45:68',
                                                                       'mtu': 2000,
                                                                       'name': '249203/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151530',
                                                                       'out_octets': '95094135465',
                                                                       'out_ucast_pkts': '68339439',
                                                                       'poller_hash': 'MTE3REFCOHBlZXJpbnRmMjQ5MjAzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '249203',
                                                                       'in_nucast_pkts': '1151530',
                                                                       'in_octets': '95094135465',
                                                                       'in_ucast_pkts': '68339439',
                                                                       'links': [       {       'in_nucast_pkts': '1151530',
                                                                                                'in_octets': '95094135465',
                                                                                                'in_ucast_pkts': '68339439',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9760',
                                                                                                'out_octets': '4458023950',
                                                                                                'out_ucast_pkts': '20010700',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9760',
                                                                       'out_octets': '4458023950',
                                                                       'out_ucast_pkts': '20010700',
                                                                       'poller_hash': 'MTE3REFCOHBlZXJpbnRmcG9uIDI0OTIwMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249203',
                                                                                               'links': [       {       'connected_time': '538374',
                                                                                                                        'fiber_length': 22126,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFCOHBlZXJwb25wb24gMjQ5MjAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4290 LINE 46 (BRUBACHER PAUL & '
                                                 'CATHY)',
                                         'poller_hash': 'MTE3REFCOA==',
                                         'serial_number': '117DAB8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:45:65',
                                         'uptime': '538374'},
                                 {       'counter_source': '249223/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a0:8a:58'],
                                                                       'id': '249223/g1',
                                                                       'in_nucast_pkts': '11590',
                                                                       'in_octets': '4799404396',
                                                                       'in_ucast_pkts': '29984667',
                                                                       'mac_address': 'f8:85:f9:b4:ed:59',
                                                                       'mtu': 2000,
                                                                       'name': '249223/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151348',
                                                                       'out_octets': '193278073945',
                                                                       'out_ucast_pkts': '133354499',
                                                                       'poller_hash': 'MTBBODZBRHBlZXJpbnRmMjQ5MjIzL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '249223',
                                                                       'in_nucast_pkts': '1151348',
                                                                       'in_octets': '193278073945',
                                                                       'in_ucast_pkts': '133354499',
                                                                       'links': [       {       'in_nucast_pkts': '1151348',
                                                                                                'in_octets': '193278073945',
                                                                                                'in_ucast_pkts': '133354499',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11590',
                                                                                                'out_octets': '4799404396',
                                                                                                'out_ucast_pkts': '29984667',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11590',
                                                                       'out_octets': '4799404396',
                                                                       'out_ucast_pkts': '29984667',
                                                                       'poller_hash': 'MTBBODZBRHBlZXJpbnRmcG9uIDI0OTIyMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249223',
                                                                                               'links': [       {       'connected_time': '538379',
                                                                                                                        'fiber_length': 20363,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODZBRHBlZXJwb25wb24gMjQ5MjIz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4149 LINE 46 (HARWOOD PAUL D)',
                                         'poller_hash': 'MTBBODZBRA==',
                                         'serial_number': '10A86AD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:ed:56',
                                         'uptime': '538379'},
                                 {       'counter_source': '249233/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f0:dd'],
                                                                       'id': '249233/g1',
                                                                       'in_nucast_pkts': '64667',
                                                                       'in_octets': '15932476656',
                                                                       'in_ucast_pkts': '92133693',
                                                                       'mac_address': 'b8:94:70:a1:4d:98',
                                                                       'mtu': 2000,
                                                                       'name': '249233/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6357809',
                                                                       'out_octets': '677837916371',
                                                                       'out_ucast_pkts': '502367444',
                                                                       'poller_hash': 'MTE3REJCRXBlZXJpbnRmMjQ5MjMzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '249233',
                                                                       'in_nucast_pkts': '6357809',
                                                                       'in_octets': '677837916371',
                                                                       'in_ucast_pkts': '502367444',
                                                                       'links': [       {       'in_nucast_pkts': '6357809',
                                                                                                'in_octets': '677837916371',
                                                                                                'in_ucast_pkts': '502367444',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '64667',
                                                                                                'out_octets': '15932476656',
                                                                                                'out_ucast_pkts': '92133693',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '64667',
                                                                       'out_octets': '15932476656',
                                                                       'out_ucast_pkts': '92133693',
                                                                       'poller_hash': 'MTE3REJCRXBlZXJpbnRmcG9uIDI0OTIzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249233',
                                                                                               'links': [       {       'connected_time': '2969293',
                                                                                                                        'fiber_length': 17885,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJCRXBlZXJwb25wb24gMjQ5MjMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '140 WINSTANLEY ST (BROWN ADAM)',
                                         'poller_hash': 'MTE3REJCRQ==',
                                         'serial_number': '117DBBE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4d:95',
                                         'uptime': '2969293'},
                                 {       'counter_source': '249253/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249253/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '65962708150',
                                                                       'in_ucast_pkts': '146306891',
                                                                       'mac_address': 'f8:85:f9:b4:e9:fe',
                                                                       'name': '249253/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '46',
                                                                       'out_octets': '349886569189',
                                                                       'out_ucast_pkts': '329253001',
                                                                       'poller_hash': 'MTBBODY0MnBlZXJpbnRmMjQ5MjUzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249253',
                                                                       'in_nucast_pkts': '46',
                                                                       'in_octets': '349886569189',
                                                                       'in_ucast_pkts': '329253001',
                                                                       'links': [       {       'in_nucast_pkts': '46',
                                                                                                'in_octets': '349886569189',
                                                                                                'in_ucast_pkts': '329253001',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '65962708150',
                                                                                                'out_ucast_pkts': '146306891',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '65962708150',
                                                                       'out_ucast_pkts': '146306891',
                                                                       'poller_hash': 'MTBBODY0MnBlZXJpbnRmcG9uIDI0OTI1Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249253',
                                                                                               'links': [       {       'connected_time': '65327368',
                                                                                                                        'fiber_length': 9019,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODY0MnBlZXJwb25wb24gMjQ5MjUz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4460 LINE 52 (ROPP RICHARD)',
                                         'poller_hash': 'MTBBODY0Mg==',
                                         'serial_number': '10A8642',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:e9:fe',
                                         'uptime': '65327368'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '249276/g1',
                                                                       'name': '249276/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REI0RnBlZXJpbnRmMjQ5Mjc2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249276',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REI0RnBlZXJpbnRmcG9uIDI0OTI3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249276',
                                                                                               'links': [       {       'connected_time': '10913301',
                                                                                                                        'fiber_length': 7339,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI0RnBlZXJwb25wb24gMjQ5Mjc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6571 ROAD 140 (EBERSOL AMANDA)',
                                         'poller_hash': 'MTE3REI0Rg==',
                                         'serial_number': '117DB4F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:1d',
                                         'uptime': '10913301'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '249296/g1',
                                                                       'name': '249296/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTBBODZCRHBlZXJpbnRmMjQ5Mjk2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249296',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTBBODZCRHBlZXJpbnRmcG9uIDI0OTI5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249296',
                                                                                               'links': [       {       'connected_time': '22578896',
                                                                                                                        'fiber_length': 9070,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODZCRHBlZXJwb25wb24gMjQ5Mjk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6943 ROAD 124 (ALBRECHT MURRAY & '
                                                 'LORETTA)',
                                         'poller_hash': 'MTBBODZCRA==',
                                         'serial_number': '10A86BD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:ed:d6',
                                         'uptime': '22578896'},
                                 {       'counter_source': '249478/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249478/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96754856862',
                                                                       'in_ucast_pkts': '328435888',
                                                                       'mac_address': 'f8:85:f9:b4:ef:26',
                                                                       'name': '249478/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '1646951485953',
                                                                       'out_ucast_pkts': '1351642272',
                                                                       'poller_hash': 'MTBBODZFN3BlZXJpbnRmMjQ5NDc4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249478',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '1646951485953',
                                                                       'in_ucast_pkts': '1351642272',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '1646951485953',
                                                                                                'in_ucast_pkts': '1351642272',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96754856862',
                                                                                                'out_ucast_pkts': '328435888',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96754856862',
                                                                       'out_ucast_pkts': '328435888',
                                                                       'poller_hash': 'MTBBODZFN3BlZXJpbnRmcG9uIDI0OTQ3OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249478',
                                                                                               'links': [       {       'connected_time': '64902557',
                                                                                                                        'fiber_length': 6622,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODZFN3BlZXJwb25wb24gMjQ5NDc4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4213 LINE 67 (MARTIN PAUL M)',
                                         'poller_hash': 'MTBBODZFNw==',
                                         'serial_number': '10A86E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:ef:26',
                                         'uptime': '64902557'},
                                 {       'counter_source': '249498/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249498/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '346858204779',
                                                                       'in_ucast_pkts': '53248524123',
                                                                       'mac_address': 'f8:85:f9:b4:ef:56',
                                                                       'name': '249498/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5313810094830',
                                                                       'out_ucast_pkts': '3739705772',
                                                                       'poller_hash': 'MTBBODZFRHBlZXJpbnRmMjQ5NDk4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249498',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5313810094830',
                                                                       'in_ucast_pkts': '3739705772',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '5313810094830',
                                                                                                'in_ucast_pkts': '3739705772',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '346858204779',
                                                                                                'out_ucast_pkts': '53248524123',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '346858204779',
                                                                       'out_ucast_pkts': '53248524123',
                                                                       'poller_hash': 'MTBBODZFRHBlZXJpbnRmcG9uIDI0OTQ5OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249498',
                                                                                               'links': [       {       'connected_time': '688595',
                                                                                                                        'fiber_length': 10661,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODZFRHBlZXJwb25wb24gMjQ5NDk4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6409 PERTH ROAD 121 (SCHERING LUCAS)',
                                         'poller_hash': 'MTBBODZFRA==',
                                         'serial_number': '10A86ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:ef:56',
                                         'uptime': '688595'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '249528/g1',
                                                                       'name': '249528/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REI0MXBlZXJpbnRmMjQ5NTI4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249528',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REI0MXBlZXJpbnRmcG9uIDI0OTUyOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249528',
                                                                                               'links': [       {       'connected_time': '64910571',
                                                                                                                        'fiber_length': 11202,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI0MXBlZXJwb25wb24gMjQ5NTI4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4174 LINE 61 (KUEPFER SARAH)',
                                         'poller_hash': 'MTE3REI0MQ==',
                                         'serial_number': '117DB41',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:49:ad',
                                         'uptime': '64910571'},
                                 {       'counter_source': '249601/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249601/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '296',
                                                                       'in_octets': '265919503409',
                                                                       'in_ucast_pkts': '5525945238',
                                                                       'mac_address': 'f8:85:f9:a8:22:c8',
                                                                       'name': '249601/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1246',
                                                                       'out_octets': '4318340523388',
                                                                       'out_ucast_pkts': '3176230830',
                                                                       'poller_hash': 'MTA5NDhCNHBlZXJpbnRmMjQ5NjAxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249601',
                                                                       'in_nucast_pkts': '1246',
                                                                       'in_octets': '4318340523388',
                                                                       'in_ucast_pkts': '3176230830',
                                                                       'links': [       {       'in_nucast_pkts': '1246',
                                                                                                'in_octets': '4318340523388',
                                                                                                'in_ucast_pkts': '3176230830',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '296',
                                                                                                'out_octets': '265919503409',
                                                                                                'out_ucast_pkts': '5525945238',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '296',
                                                                       'out_octets': '265919503409',
                                                                       'out_ucast_pkts': '5525945238',
                                                                       'poller_hash': 'MTA5NDhCNHBlZXJpbnRmcG9uIDI0OTYwMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249601',
                                                                                               'links': [       {       'connected_time': '330749',
                                                                                                                        'fiber_length': 1684,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTA5NDhCNHBlZXJwb25wb24gMjQ5NjAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '55 WILLIAM ST Unit 1 (Rahman '
                                                 'Habibur)',
                                         'poller_hash': 'MTA5NDhCNA==',
                                         'serial_number': '10948B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:a8:22:c8',
                                         'uptime': '330749'},
                                 {       'counter_source': '249612/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249612/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '236363120846',
                                                                       'in_ucast_pkts': '1398277995',
                                                                       'mac_address': 'b8:94:70:a1:40:2d',
                                                                       'name': '249612/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '8227355631983',
                                                                       'out_ucast_pkts': '5781736918',
                                                                       'poller_hash': 'MTE3REExMXBlZXJpbnRmMjQ5NjEyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '249612',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '8227355631983',
                                                                       'in_ucast_pkts': '5781736918',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '8227355631983',
                                                                                                'in_ucast_pkts': '5781736918',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '236363120846',
                                                                                                'out_ucast_pkts': '1398277995',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '236363120846',
                                                                       'out_ucast_pkts': '1398277995',
                                                                       'poller_hash': 'MTE3REExMXBlZXJpbnRmcG9uIDI0OTYxMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249612',
                                                                                               'links': [       {       'connected_time': '523341',
                                                                                                                        'fiber_length': 6497,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REExMXBlZXJwb25wb24gMjQ5NjEy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4454 PERTH LINE 72 (WILLIAMS JUSTIN)',
                                         'poller_hash': 'MTE3REExMQ==',
                                         'serial_number': '117DA11',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:40:2d',
                                         'uptime': '523341'},
                                 {       'counter_source': '249622/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:c7:19'],
                                                                       'id': '249622/g1',
                                                                       'in_nucast_pkts': '15022',
                                                                       'in_octets': '214672509125',
                                                                       'in_ucast_pkts': '310021559',
                                                                       'mac_address': 'f8:85:f9:b4:e9:99',
                                                                       'mtu': 2000,
                                                                       'name': '249622/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465249',
                                                                       'out_octets': '571727594377',
                                                                       'out_ucast_pkts': '453592433',
                                                                       'poller_hash': 'MTBBODYzNXBlZXJpbnRmMjQ5NjIyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '249622',
                                                                       'in_nucast_pkts': '1465249',
                                                                       'in_octets': '571727594377',
                                                                       'in_ucast_pkts': '453592433',
                                                                       'links': [       {       'in_nucast_pkts': '1465249',
                                                                                                'in_octets': '571727594377',
                                                                                                'in_ucast_pkts': '453592433',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15022',
                                                                                                'out_octets': '214672509125',
                                                                                                'out_ucast_pkts': '310021559',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15022',
                                                                       'out_octets': '214672509125',
                                                                       'out_ucast_pkts': '310021559',
                                                                       'poller_hash': 'MTBBODYzNXBlZXJpbnRmcG9uIDI0OTYyMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249622',
                                                                                               'links': [       {       'connected_time': '688372',
                                                                                                                        'fiber_length': 2949,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODYzNXBlZXJwb25wb24gMjQ5NjIy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '51 COULTER ST (MARCHESE ANTHONY)',
                                         'poller_hash': 'MTBBODYzNQ==',
                                         'serial_number': '10A8635',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:e9:96',
                                         'uptime': '688372'},
                                 {       'counter_source': '250034/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '250034/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '119202698810',
                                                                       'in_ucast_pkts': '917440662',
                                                                       'mac_address': 'b8:94:70:a1:49:6d',
                                                                       'name': '250034/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2987252149522',
                                                                       'out_ucast_pkts': '2159398981',
                                                                       'poller_hash': 'MTE3REIzOXBlZXJpbnRmMjUwMDM0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250034',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2987252149522',
                                                                       'in_ucast_pkts': '2159398981',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2987252149522',
                                                                                                'in_ucast_pkts': '2159398981',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '119202698810',
                                                                                                'out_ucast_pkts': '917440662',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '119202698810',
                                                                       'out_ucast_pkts': '917440662',
                                                                       'poller_hash': 'MTE3REIzOXBlZXJpbnRmcG9uIDI1MDAzNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250034',
                                                                                               'links': [       {       'connected_time': '523101',
                                                                                                                        'fiber_length': 11498,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIzOXBlZXJwb25wb24gMjUwMDM0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6951 ELGIN ST (POOLE SAMANTHA)',
                                         'poller_hash': 'MTE3REIzOQ==',
                                         'serial_number': '117DB39',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:49:6d',
                                         'uptime': '523101'},
                                 {       'counter_source': '250044/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:72:e1'],
                                                                       'id': '250044/g1',
                                                                       'in_nucast_pkts': '593380',
                                                                       'in_octets': '254211694243',
                                                                       'in_ucast_pkts': '2110349588',
                                                                       'mac_address': 'b8:94:70:a1:46:e8',
                                                                       'mtu': 2000,
                                                                       'name': '250044/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '77346422',
                                                                       'out_octets': '12963656033255',
                                                                       'out_ucast_pkts': '8774499208',
                                                                       'poller_hash': 'MTE3REFFOHBlZXJpbnRmMjUwMDQ0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '250044',
                                                                       'in_nucast_pkts': '77346422',
                                                                       'in_octets': '12963656033255',
                                                                       'in_ucast_pkts': '8774499208',
                                                                       'links': [       {       'in_nucast_pkts': '77346422',
                                                                                                'in_octets': '12963656033255',
                                                                                                'in_ucast_pkts': '8774499208',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '593380',
                                                                                                'out_octets': '254211694243',
                                                                                                'out_ucast_pkts': '2110349588',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '593380',
                                                                       'out_octets': '254211694243',
                                                                       'out_ucast_pkts': '2110349588',
                                                                       'poller_hash': 'MTE3REFFOHBlZXJpbnRmcG9uIDI1MDA0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250044',
                                                                                               'links': [       {       'connected_time': '64566130',
                                                                                                                        'fiber_length': 20621,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFFOHBlZXJwb25wb24gMjUwMDQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2542 LINE 45 (ROSE JOHN & LOIS)',
                                         'poller_hash': 'MTE3REFFOA==',
                                         'serial_number': '117DAE8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:46:e5',
                                         'uptime': '64566130'},
                                 {       'counter_source': '250054/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '250054/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '141415249998',
                                                                       'in_ucast_pkts': '34710156580',
                                                                       'mac_address': 'b8:94:70:a1:3e:95',
                                                                       'name': '250054/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '2271014758427',
                                                                       'out_ucast_pkts': '1719259161',
                                                                       'poller_hash': 'MTE3RDlERXBlZXJpbnRmMjUwMDU0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250054',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '2271014758427',
                                                                       'in_ucast_pkts': '1719259161',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '2271014758427',
                                                                                                'in_ucast_pkts': '1719259161',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '141415249998',
                                                                                                'out_ucast_pkts': '34710156580',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '141415249998',
                                                                       'out_ucast_pkts': '34710156580',
                                                                       'poller_hash': 'MTE3RDlERXBlZXJpbnRmcG9uIDI1MDA1NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250054',
                                                                                               'links': [       {       'connected_time': '262226',
                                                                                                                        'fiber_length': 20346,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3RDlERXBlZXJwb25wb24gMjUwMDU0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2520 LINE 45 (ACKERSVILLER BRADLEY)',
                                         'poller_hash': 'MTE3RDlERQ==',
                                         'serial_number': '117D9DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:3e:95',
                                         'uptime': '262226'},
                                 {       'counter_source': '250064/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:eb:52'],
                                                                       'id': '250064/g1',
                                                                       'in_nucast_pkts': '12111',
                                                                       'in_octets': '6524050749',
                                                                       'in_ucast_pkts': '28516334',
                                                                       'mac_address': 'b8:94:70:a1:4a:a8',
                                                                       'mtu': 2000,
                                                                       'name': '250064/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465060',
                                                                       'out_octets': '89590013163',
                                                                       'out_ucast_pkts': '69032541',
                                                                       'poller_hash': 'MTE3REI2MHBlZXJpbnRmMjUwMDY0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '250064',
                                                                       'in_nucast_pkts': '1465060',
                                                                       'in_octets': '89590013163',
                                                                       'in_ucast_pkts': '69032541',
                                                                       'links': [       {       'in_nucast_pkts': '1465060',
                                                                                                'in_octets': '89590013163',
                                                                                                'in_ucast_pkts': '69032541',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12111',
                                                                                                'out_octets': '6524050749',
                                                                                                'out_ucast_pkts': '28516334',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12111',
                                                                       'out_octets': '6524050749',
                                                                       'out_ucast_pkts': '28516334',
                                                                       'poller_hash': 'MTE3REI2MHBlZXJpbnRmcG9uIDI1MDA2NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250064',
                                                                                               'links': [       {       'connected_time': '688374',
                                                                                                                        'fiber_length': 2784,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI2MHBlZXJwb25wb24gMjUwMDY0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '61 PUGH ST (WOODTHORPE PETER)',
                                         'poller_hash': 'MTE3REI2MA==',
                                         'serial_number': '117DB60',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:a5',
                                         'uptime': '688374'},
                                 {       'counter_source': '250094/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '250094/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '51862057841',
                                                                       'in_ucast_pkts': '299859186',
                                                                       'mac_address': 'f8:85:f9:b4:f0:d6',
                                                                       'name': '250094/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '1609948040095',
                                                                       'out_ucast_pkts': '1191846291',
                                                                       'poller_hash': 'MTBBODcxRHBlZXJpbnRmMjUwMDk0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250094',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '1609948040095',
                                                                       'in_ucast_pkts': '1191846291',
                                                                       'links': [       {       'in_nucast_pkts': '4',
                                                                                                'in_octets': '1609948040095',
                                                                                                'in_ucast_pkts': '1191846291',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '51862057841',
                                                                                                'out_ucast_pkts': '299859186',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '51862057841',
                                                                       'out_ucast_pkts': '299859186',
                                                                       'poller_hash': 'MTBBODcxRHBlZXJpbnRmcG9uIDI1MDA5NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250094',
                                                                                               'links': [       {       'connected_time': '538347',
                                                                                                                        'fiber_length': 22266,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODcxRHBlZXJwb25wb24gMjUwMDk0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2700 LINE 45 (BALFOUR WESLEY)',
                                         'poller_hash': 'MTBBODcxRA==',
                                         'serial_number': '10A871D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:f0:d6',
                                         'uptime': '538347'},
                                 {       'counter_source': '250104/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f0:f3'],
                                                                       'id': '250104/g1',
                                                                       'in_nucast_pkts': '19576',
                                                                       'in_octets': '9976103705',
                                                                       'in_ucast_pkts': '18933589',
                                                                       'mac_address': 'f8:85:f9:b4:ec:f9',
                                                                       'mtu': 2000,
                                                                       'name': '250104/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2190700',
                                                                       'out_octets': '71370889505',
                                                                       'out_ucast_pkts': '58999481',
                                                                       'poller_hash': 'MTBBODZBMXBlZXJpbnRmMjUwMTA0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '250104',
                                                                       'in_nucast_pkts': '2190700',
                                                                       'in_octets': '71370889505',
                                                                       'in_ucast_pkts': '58999481',
                                                                       'links': [       {       'in_nucast_pkts': '2190700',
                                                                                                'in_octets': '71370889505',
                                                                                                'in_ucast_pkts': '58999481',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19576',
                                                                                                'out_octets': '9976103705',
                                                                                                'out_ucast_pkts': '18933589',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19576',
                                                                       'out_octets': '9976103705',
                                                                       'out_ucast_pkts': '18933589',
                                                                       'poller_hash': 'MTBBODZBMXBlZXJpbnRmcG9uIDI1MDEwNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250104',
                                                                                               'links': [       {       'connected_time': '1023956',
                                                                                                                        'fiber_length': 12762,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODZBMXBlZXJwb25wb24gMjUwMTA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6405 ROAD 116 (METZGER CHARLENE)',
                                         'poller_hash': 'MTBBODZBMQ==',
                                         'serial_number': '10A86A1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:ec:f6',
                                         'uptime': '1023956'},
                                 {       'counter_source': '250124/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '250124/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '338938878696',
                                                                       'in_ucast_pkts': '1884411055',
                                                                       'mac_address': 'b8:94:70:a1:4a:2d',
                                                                       'name': '250124/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '8441798087908',
                                                                       'out_ucast_pkts': '5938021296',
                                                                       'poller_hash': 'MTE3REI1MXBlZXJpbnRmMjUwMTI0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250124',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '8441798087908',
                                                                       'in_ucast_pkts': '5938021296',
                                                                       'links': [       {       'in_nucast_pkts': '14',
                                                                                                'in_octets': '8441798087908',
                                                                                                'in_ucast_pkts': '5938021296',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '338938878696',
                                                                                                'out_ucast_pkts': '1884411055',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '338938878696',
                                                                       'out_ucast_pkts': '1884411055',
                                                                       'poller_hash': 'MTE3REI1MXBlZXJpbnRmcG9uIDI1MDEyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250124',
                                                                                               'links': [       {       'connected_time': '688608',
                                                                                                                        'fiber_length': 6608,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI1MXBlZXJwb25wb24gMjUwMTI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4442 PERTH LINE 72 (HIEBERT JACOB)',
                                         'poller_hash': 'MTE3REI1MQ==',
                                         'serial_number': '117DB51',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:2d',
                                         'uptime': '688608'},
                                 {       'counter_source': '250144/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '250144/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '220598880140',
                                                                       'in_ucast_pkts': '951749113',
                                                                       'mac_address': 'b8:94:70:a1:49:e5',
                                                                       'name': '250144/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '57',
                                                                       'out_octets': '10298633928438',
                                                                       'out_ucast_pkts': '2914650736',
                                                                       'poller_hash': 'MTE3REI0OHBlZXJpbnRmMjUwMTQ0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250144',
                                                                       'in_nucast_pkts': '57',
                                                                       'in_octets': '10298633928438',
                                                                       'in_ucast_pkts': '2914650736',
                                                                       'links': [       {       'in_nucast_pkts': '57',
                                                                                                'in_octets': '10298633928438',
                                                                                                'in_ucast_pkts': '2914650736',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '220598880140',
                                                                                                'out_ucast_pkts': '951749113',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '220598880140',
                                                                       'out_ucast_pkts': '951749113',
                                                                       'poller_hash': 'MTE3REI0OHBlZXJpbnRmcG9uIDI1MDE0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250144',
                                                                                               'links': [       {       'connected_time': '64132371',
                                                                                                                        'fiber_length': 20505,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI0OHBlZXJwb25wb24gMjUwMTQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4168 LINE 46 (LOUWAGIE DENNIS)',
                                         'poller_hash': 'MTE3REI0OA==',
                                         'serial_number': '117DB48',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:49:e5',
                                         'uptime': '64132371'},
                                 {       'counter_source': '250247/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:e9:6e'],
                                                                       'id': '250247/g1',
                                                                       'in_nucast_pkts': '11485',
                                                                       'in_octets': '7134496001',
                                                                       'in_ucast_pkts': '12722342',
                                                                       'mac_address': 'b8:94:70:a1:45:d8',
                                                                       'mtu': 2000,
                                                                       'name': '250247/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151478',
                                                                       'out_octets': '53813366910',
                                                                       'out_ucast_pkts': '41525180',
                                                                       'poller_hash': 'MTE3REFDNnBlZXJpbnRmMjUwMjQ3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '250247',
                                                                       'in_nucast_pkts': '1151478',
                                                                       'in_octets': '53813366910',
                                                                       'in_ucast_pkts': '41525180',
                                                                       'links': [       {       'in_nucast_pkts': '1151478',
                                                                                                'in_octets': '53813366910',
                                                                                                'in_ucast_pkts': '41525180',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11485',
                                                                                                'out_octets': '7134496001',
                                                                                                'out_ucast_pkts': '12722342',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11485',
                                                                       'out_octets': '7134496001',
                                                                       'out_ucast_pkts': '12722342',
                                                                       'poller_hash': 'MTE3REFDNnBlZXJpbnRmcG9uIDI1MDI0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250247',
                                                                                               'links': [       {       'connected_time': '538379',
                                                                                                                        'fiber_length': 20723,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFDNnBlZXJwb25wb24gMjUwMjQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4179 LINE 46 (PLAYERDISE FARMS LTD)',
                                         'poller_hash': 'MTE3REFDNg==',
                                         'serial_number': '117DAC6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:45:d5',
                                         'uptime': '538379'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '250257/g1',
                                                                       'name': '250257/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDBDNnBlZXJpbnRmMjUwMjU3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250257',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDBDNnBlZXJpbnRmcG9uIDI1MDI1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250257',
                                                                                               'links': [       {       'connected_time': '64044936',
                                                                                                                        'fiber_length': 22154,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDBDNnBlZXJwb25wb24gMjUwMjU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4913 ROAD 122 (PLAYERDISE FARMS LTD)',
                                         'poller_hash': 'MTI3NDBDNg==',
                                         'serial_number': '12740C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d2:31',
                                         'uptime': '64044936'},
                                 {       'counter_source': '250267/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:74:af'],
                                                                       'id': '250267/g1',
                                                                       'in_nucast_pkts': '587651',
                                                                       'in_octets': '2102738826256',
                                                                       'in_ucast_pkts': '5245381888',
                                                                       'mac_address': 'b8:94:70:a1:45:60',
                                                                       'mtu': 2000,
                                                                       'name': '250267/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '73043118',
                                                                       'out_octets': '13385457878614',
                                                                       'out_ucast_pkts': '10655604236',
                                                                       'poller_hash': 'MTE3REFCN3BlZXJpbnRmMjUwMjY3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '250267',
                                                                       'in_nucast_pkts': '73043118',
                                                                       'in_octets': '13385457878614',
                                                                       'in_ucast_pkts': '10655604236',
                                                                       'links': [       {       'in_nucast_pkts': '73043118',
                                                                                                'in_octets': '13385457878614',
                                                                                                'in_ucast_pkts': '10655604236',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '587651',
                                                                                                'out_octets': '2102738826256',
                                                                                                'out_ucast_pkts': '5245381888',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '587651',
                                                                       'out_octets': '2102738826256',
                                                                       'out_ucast_pkts': '5245381888',
                                                                       'poller_hash': 'MTE3REFCN3BlZXJpbnRmcG9uIDI1MDI2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250267',
                                                                                               'links': [       {       'connected_time': '64029754',
                                                                                                                        'fiber_length': 19116,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFCN3BlZXJwb25wb24gMjUwMjY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5135 ROAD 122 (REIGER JOE)',
                                         'poller_hash': 'MTE3REFCNw==',
                                         'serial_number': '117DAB7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:45:5d',
                                         'uptime': '64029754'},
                                 {       'counter_source': '250277/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '250277/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45457928',
                                                                       'in_ucast_pkts': '70739',
                                                                       'mac_address': 'b8:94:70:a1:49:cd',
                                                                       'name': '250277/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '25',
                                                                       'out_octets': '8083094952',
                                                                       'out_ucast_pkts': '41643667',
                                                                       'poller_hash': 'MTE3REI0NXBlZXJpbnRmMjUwMjc3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250277',
                                                                       'in_nucast_pkts': '25',
                                                                       'in_octets': '8083094952',
                                                                       'in_ucast_pkts': '41643667',
                                                                       'links': [       {       'in_nucast_pkts': '25',
                                                                                                'in_octets': '8083094952',
                                                                                                'in_ucast_pkts': '41643667',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45457928',
                                                                                                'out_ucast_pkts': '70739',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45457928',
                                                                       'out_ucast_pkts': '70739',
                                                                       'poller_hash': 'MTE3REI0NXBlZXJpbnRmcG9uIDI1MDI3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250277',
                                                                                               'links': [       {       'connected_time': '64028908',
                                                                                                                        'fiber_length': 19300,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI0NXBlZXJwb25wb24gMjUwMjc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5122 ROAD 122 (VAN STRAATEN DAVE)',
                                         'poller_hash': 'MTE3REI0NQ==',
                                         'serial_number': '117DB45',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:49:cd',
                                         'uptime': '64028908'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '250287/g1',
                                                                       'name': '250287/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REFGRXBlZXJpbnRmMjUwMjg3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250287',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REFGRXBlZXJpbnRmcG9uIDI1MDI4Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250287',
                                                                                               'links': [       {       'connected_time': '15143026',
                                                                                                                        'fiber_length': 6674,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFGRXBlZXJwb25wb24gMjUwMjg3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4434 PERTH LINE 72 (SCHMIDT '
                                                 'KENDRICK)',
                                         'poller_hash': 'MTE3REFGRQ==',
                                         'serial_number': '117DAFE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:47:95',
                                         'uptime': '15143026'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '250307/g1',
                                                                       'name': '250307/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTBBODY0MXBlZXJpbnRmMjUwMzA3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250307',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTBBODY0MXBlZXJpbnRmcG9uIDI1MDMwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250307',
                                                                                               'links': [       {       'connected_time': '63961891',
                                                                                                                        'fiber_length': 11878,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODY0MXBlZXJwb25wb24gMjUwMzA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7088 PERTH ROAD 121 (SAUDER MAYNARD)',
                                         'poller_hash': 'MTBBODY0MQ==',
                                         'serial_number': '10A8641',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:e9:f6',
                                         'uptime': '63961891'},
                                 {       'counter_source': '250337/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '250337/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '80146941947',
                                                                       'in_ucast_pkts': '154730638',
                                                                       'mac_address': 'f8:85:f9:b4:f1:4e',
                                                                       'name': '250337/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '643041138373',
                                                                       'out_ucast_pkts': '558030777',
                                                                       'poller_hash': 'MTBBODcyQ3BlZXJpbnRmMjUwMzM3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '250337',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '643041138373',
                                                                       'in_ucast_pkts': '558030777',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '643041138373',
                                                                                                'in_ucast_pkts': '558030777',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '80146941947',
                                                                                                'out_ucast_pkts': '154730638',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '80146941947',
                                                                       'out_ucast_pkts': '154730638',
                                                                       'poller_hash': 'MTBBODcyQ3BlZXJpbnRmcG9uIDI1MDMzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250337',
                                                                                               'links': [       {       'connected_time': '3114024',
                                                                                                                        'fiber_length': 16382,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTBBODcyQ3BlZXJwb25wb24gMjUwMzM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5719 PERTH LINE 55 (CHARMAINE VALDEZ '
                                                 'JOYCE)',
                                         'poller_hash': 'MTBBODcyQw==',
                                         'serial_number': '10A872C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:b4:f1:4e',
                                         'uptime': '3114024'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '251024/g1',
                                                                       'name': '251024/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REFDMXBlZXJpbnRmMjUxMDI0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251024',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REFDMXBlZXJpbnRmcG9uIDI1MTAyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251024',
                                                                                               'links': [       {       'connected_time': '63443605',
                                                                                                                        'fiber_length': 11556,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFDMXBlZXJwb25wb24gMjUxMDI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4081 PERTH LINE 72 (BARTSCH DAVID)',
                                         'poller_hash': 'MTE3REFDMQ==',
                                         'serial_number': '117DAC1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:45:ad',
                                         'uptime': '63443605'},
                                 {       'counter_source': '251084/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251084/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '106530660415',
                                                                       'in_ucast_pkts': '548980354',
                                                                       'mac_address': 'b8:94:70:a1:50:75',
                                                                       'name': '251084/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '2944736963626',
                                                                       'out_ucast_pkts': '2138729830',
                                                                       'poller_hash': 'MTE3REMxQXBlZXJpbnRmMjUxMDg0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251084',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '2944736963626',
                                                                       'in_ucast_pkts': '2138729830',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '2944736963626',
                                                                                                'in_ucast_pkts': '2138729830',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '106530660415',
                                                                                                'out_ucast_pkts': '548980354',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '106530660415',
                                                                       'out_ucast_pkts': '548980354',
                                                                       'poller_hash': 'MTE3REMxQXBlZXJpbnRmcG9uIDI1MTA4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251084',
                                                                                               'links': [       {       'connected_time': '688603',
                                                                                                                        'fiber_length': 6631,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMxQXBlZXJwb25wb24gMjUxMDg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4438 PERTH LINE 72 (POOLE TYLER)',
                                         'poller_hash': 'MTE3REMxQQ==',
                                         'serial_number': '117DC1A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:75',
                                         'uptime': '688603'},
                                 {       'counter_source': '251133/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251133/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '61869152720',
                                                                       'in_ucast_pkts': '158647580',
                                                                       'mac_address': 'b8:94:70:a1:49:15',
                                                                       'name': '251133/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '21',
                                                                       'out_octets': '518751538612',
                                                                       'out_ucast_pkts': '424099703',
                                                                       'poller_hash': 'MTE3REIyRXBlZXJpbnRmMjUxMTMzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251133',
                                                                       'in_nucast_pkts': '21',
                                                                       'in_octets': '518751538612',
                                                                       'in_ucast_pkts': '424099703',
                                                                       'links': [       {       'in_nucast_pkts': '21',
                                                                                                'in_octets': '518751538612',
                                                                                                'in_ucast_pkts': '424099703',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '61869152720',
                                                                                                'out_ucast_pkts': '158647580',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '61869152720',
                                                                       'out_ucast_pkts': '158647580',
                                                                       'poller_hash': 'MTE3REIyRXBlZXJpbnRmcG9uIDI1MTEzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251133',
                                                                                               'links': [       {       'connected_time': '3114005',
                                                                                                                        'fiber_length': 18437,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIyRXBlZXJwb25wb24gMjUxMTMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 SCHADE ST MONKTON (WALL GEORGE)',
                                         'poller_hash': 'MTE3REIyRQ==',
                                         'serial_number': '117DB2E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:49:15',
                                         'uptime': '3114005'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '251153/g1',
                                                                       'name': '251153/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REI3MXBlZXJpbnRmMjUxMTUzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251153',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REI3MXBlZXJpbnRmcG9uIDI1MTE1Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251153',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI3MXBlZXJwb25wb24gMjUxMTUz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '14 CRUICKSHANK ST (BRAUN JACOB)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "251153" '
                                                                  '(location: GPON '
                                                                  'port "3/6"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-06-14T21:35:57.390Z',
                                         'poller_hash': 'MTE3REI3MQ==',
                                         'serial_number': '117DB71',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '251163/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:c7:9d'],
                                                                       'id': '251163/g1',
                                                                       'in_nucast_pkts': '12425',
                                                                       'in_octets': '10688741581',
                                                                       'in_ucast_pkts': '25277908',
                                                                       'mac_address': 'b8:94:70:a1:4c:38',
                                                                       'mtu': 2000,
                                                                       'name': '251163/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465153',
                                                                       'out_octets': '134931551825',
                                                                       'out_ucast_pkts': '99815627',
                                                                       'poller_hash': 'MTE3REI5MnBlZXJpbnRmMjUxMTYzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '251163',
                                                                       'in_nucast_pkts': '1465153',
                                                                       'in_octets': '134931551825',
                                                                       'in_ucast_pkts': '99815627',
                                                                       'links': [       {       'in_nucast_pkts': '1465153',
                                                                                                'in_octets': '134931551825',
                                                                                                'in_ucast_pkts': '99815627',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12425',
                                                                                                'out_octets': '10688741581',
                                                                                                'out_ucast_pkts': '25277908',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12425',
                                                                       'out_octets': '10688741581',
                                                                       'out_ucast_pkts': '25277908',
                                                                       'poller_hash': 'MTE3REI5MnBlZXJpbnRmcG9uIDI1MTE2Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251163',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 3318,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI5MnBlZXJwb25wb24gMjUxMTYz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '70 PUGH ST (ZINKEN JONATHAN)',
                                         'poller_hash': 'MTE3REI5Mg==',
                                         'serial_number': '117DB92',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4c:35',
                                         'uptime': '688370'},
                                 {       'counter_source': '251529/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251529/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '91671758352',
                                                                       'in_ucast_pkts': '813851319',
                                                                       'mac_address': 'b8:94:70:ba:e2:f5',
                                                                       'name': '251529/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '4732356850273',
                                                                       'out_ucast_pkts': '3256996060',
                                                                       'poller_hash': 'MTE5NTE3M3BlZXJpbnRmMjUxNTI5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251529',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '4732356850273',
                                                                       'in_ucast_pkts': '3256996060',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '4732356850273',
                                                                                                'in_ucast_pkts': '3256996060',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '91671758352',
                                                                                                'out_ucast_pkts': '813851319',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '91671758352',
                                                                       'out_ucast_pkts': '813851319',
                                                                       'poller_hash': 'MTE5NTE3M3BlZXJpbnRmcG9uIDI1MTUyOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251529',
                                                                                               'links': [       {       'connected_time': '29355887',
                                                                                                                        'fiber_length': 17714,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE3M3BlZXJwb25wb24gMjUxNTI5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '212 WINSTANLEY ST (HULLEY LAWRENCE)',
                                         'poller_hash': 'MTE5NTE3Mw==',
                                         'serial_number': '1195173',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e2:f5',
                                         'uptime': '29355887'},
                                 {       'counter_source': '251539/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:c6:d7'],
                                                                       'id': '251539/g1',
                                                                       'in_nucast_pkts': '12625',
                                                                       'in_octets': '53159130509',
                                                                       'in_ucast_pkts': '182844077',
                                                                       'mac_address': 'b8:94:70:a1:4e:80',
                                                                       'mtu': 2000,
                                                                       'name': '251539/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465170',
                                                                       'out_octets': '738229245613',
                                                                       'out_ucast_pkts': '541294983',
                                                                       'poller_hash': 'MTE3REJEQnBlZXJpbnRmMjUxNTM5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '251539',
                                                                       'in_nucast_pkts': '1465170',
                                                                       'in_octets': '738229245613',
                                                                       'in_ucast_pkts': '541294983',
                                                                       'links': [       {       'in_nucast_pkts': '1465170',
                                                                                                'in_octets': '738229245613',
                                                                                                'in_ucast_pkts': '541294983',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12625',
                                                                                                'out_octets': '53159130509',
                                                                                                'out_ucast_pkts': '182844077',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12625',
                                                                       'out_octets': '53159130509',
                                                                       'out_ucast_pkts': '182844077',
                                                                       'poller_hash': 'MTE3REJEQnBlZXJpbnRmcG9uIDI1MTUzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251539',
                                                                                               'links': [       {       'connected_time': '688376',
                                                                                                                        'fiber_length': 2857,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJEQnBlZXJwb25wb24gMjUxNTM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '143 PUGH ST (COURCHESNE BRIAN)',
                                         'poller_hash': 'MTE3REJEQg==',
                                         'serial_number': '117DBDB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4e:7d',
                                         'uptime': '688376'},
                                 {       'counter_source': '251569/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251569/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '13429344474',
                                                                       'in_ucast_pkts': '59313160',
                                                                       'mac_address': '04:bc:9f:5b:ce:a1',
                                                                       'name': '251569/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '582538260235',
                                                                       'out_ucast_pkts': '474729193',
                                                                       'poller_hash': 'MTI3NDA1NHBlZXJpbnRmMjUxNTY5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251569',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '582538260235',
                                                                       'in_ucast_pkts': '474729193',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '582538260235',
                                                                                                'in_ucast_pkts': '474729193',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '13429344474',
                                                                                                'out_ucast_pkts': '59313160',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '13429344474',
                                                                       'out_ucast_pkts': '59313160',
                                                                       'poller_hash': 'MTI3NDA1NHBlZXJpbnRmcG9uIDI1MTU2OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251569',
                                                                                               'links': [       {       'connected_time': '688325',
                                                                                                                        'fiber_length': 1370,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA1NHBlZXJwb25wb24gMjUxNTY5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 MAIN ST N (DYCK BENJAMIN)',
                                         'poller_hash': 'MTI3NDA1NA==',
                                         'serial_number': '1274054',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:a1',
                                         'uptime': '688325'},
                                 {       'counter_source': '251619/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251619/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '143690653886',
                                                                       'in_ucast_pkts': '1061493372',
                                                                       'mac_address': 'b8:94:70:a1:43:3d',
                                                                       'name': '251619/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1',
                                                                       'out_octets': '3481876808386',
                                                                       'out_ucast_pkts': '2410879825',
                                                                       'poller_hash': 'MTE3REE3M3BlZXJpbnRmMjUxNjE5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251619',
                                                                       'in_nucast_pkts': '1',
                                                                       'in_octets': '3481876808386',
                                                                       'in_ucast_pkts': '2410879825',
                                                                       'links': [       {       'in_nucast_pkts': '1',
                                                                                                'in_octets': '3481876808386',
                                                                                                'in_ucast_pkts': '2410879825',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '143690653886',
                                                                                                'out_ucast_pkts': '1061493372',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '143690653886',
                                                                       'out_ucast_pkts': '1061493372',
                                                                       'poller_hash': 'MTE3REE3M3BlZXJpbnRmcG9uIDI1MTYxOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251619',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 2573,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REE3M3BlZXJwb25wb24gMjUxNjE5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '111 PUGH ST (BECKER JANE)',
                                         'poller_hash': 'MTE3REE3Mw==',
                                         'serial_number': '117DA73',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:43:3d',
                                         'uptime': '688369'},
                                 {       'counter_source': '251683/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f0:2d'],
                                                                       'id': '251683/g1',
                                                                       'in_nucast_pkts': '12162',
                                                                       'in_octets': '9030959910',
                                                                       'in_ucast_pkts': '13265086',
                                                                       'mac_address': 'ec:4f:82:5b:87:f6',
                                                                       'mtu': 2000,
                                                                       'name': '251683/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465077',
                                                                       'out_octets': '57790012284',
                                                                       'out_ucast_pkts': '47386820',
                                                                       'poller_hash': 'MkIwNDcxcGVlcmludGYyNTE2ODMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '251683',
                                                                       'in_nucast_pkts': '1465077',
                                                                       'in_octets': '57790012284',
                                                                       'in_ucast_pkts': '47386820',
                                                                       'links': [       {       'in_nucast_pkts': '1465077',
                                                                                                'in_octets': '57790012284',
                                                                                                'in_ucast_pkts': '47386820',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12162',
                                                                                                'out_octets': '9030959910',
                                                                                                'out_ucast_pkts': '13265086',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12162',
                                                                       'out_octets': '9030959910',
                                                                       'out_ucast_pkts': '13265086',
                                                                       'poller_hash': 'MkIwNDcxcGVlcmludGZwb24gMjUxNjgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251683',
                                                                                               'links': [       {       'connected_time': '688461',
                                                                                                                        'fiber_length': 1121,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDcxcGVlcnBvbnBvbiAyNTE2ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '31 PACIFIC AV (MCKAY ROBERT)',
                                         'poller_hash': 'MkIwNDcx',
                                         'serial_number': '2B0471',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:87:f3',
                                         'uptime': '688461'},
                                 {       'counter_source': '251775/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251775/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '258439561573',
                                                                       'in_ucast_pkts': '2721977347',
                                                                       'mac_address': 'b8:94:70:a1:46:2d',
                                                                       'name': '251775/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11948312902341',
                                                                       'out_ucast_pkts': '8128390581',
                                                                       'poller_hash': 'MTE3REFEMXBlZXJpbnRmMjUxNzc1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251775',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11948312902341',
                                                                       'in_ucast_pkts': '8128390581',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11948312902341',
                                                                                                'in_ucast_pkts': '8128390581',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '258439561573',
                                                                                                'out_ucast_pkts': '2721977347',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '258439561573',
                                                                       'out_ucast_pkts': '2721977347',
                                                                       'poller_hash': 'MTE3REFEMXBlZXJpbnRmcG9uIDI1MTc3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251775',
                                                                                               'links': [       {       'connected_time': '719128',
                                                                                                                        'fiber_length': 6546,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFEMXBlZXJwb25wb24gMjUxNzc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4458 PERTH LINE 72 (WINSTON MIKE & '
                                                 'SHARON)',
                                         'poller_hash': 'MTE3REFEMQ==',
                                         'serial_number': '117DAD1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:46:2d',
                                         'uptime': '719128'},
                                 {       'counter_source': '251805/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '251805/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '286214671837',
                                                                       'in_ucast_pkts': '813104367',
                                                                       'mac_address': 'b8:94:70:a1:50:dd',
                                                                       'name': '251805/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '3280063497982',
                                                                       'out_ucast_pkts': '2595178367',
                                                                       'poller_hash': 'MTE3REMyN3BlZXJpbnRmMjUxODA1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '251805',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '3280063497982',
                                                                       'in_ucast_pkts': '2595178367',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '3280063497982',
                                                                                                'in_ucast_pkts': '2595178367',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '286214671837',
                                                                                                'out_ucast_pkts': '813104367',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '286214671837',
                                                                       'out_ucast_pkts': '813104367',
                                                                       'poller_hash': 'MTE3REMyN3BlZXJpbnRmcG9uIDI1MTgwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251805',
                                                                                               'links': [       {       'connected_time': '13204769',
                                                                                                                        'fiber_length': 20919,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMyN3BlZXJwb25wb24gMjUxODA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2580 LINE 45 (KINGSBURY STEVE)',
                                         'poller_hash': 'MTE3REMyNw==',
                                         'serial_number': '117DC27',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:dd',
                                         'uptime': '13204769'},
                                 {       'counter_source': '251815/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:c6:a0'],
                                                                       'id': '251815/g1',
                                                                       'in_nucast_pkts': '556573',
                                                                       'in_octets': '1522043202725',
                                                                       'in_ucast_pkts': '4382952624',
                                                                       'mac_address': 'b8:94:70:a1:4b:78',
                                                                       'mtu': 2000,
                                                                       'name': '251815/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '70727695',
                                                                       'out_octets': '16993377415206',
                                                                       'out_ucast_pkts': '12347163019',
                                                                       'poller_hash': 'MTE3REI3QXBlZXJpbnRmMjUxODE1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '251815',
                                                                       'in_nucast_pkts': '70727695',
                                                                       'in_octets': '16993377415206',
                                                                       'in_ucast_pkts': '12347163019',
                                                                       'links': [       {       'in_nucast_pkts': '70727695',
                                                                                                'in_octets': '16993377415206',
                                                                                                'in_ucast_pkts': '12347163019',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '556573',
                                                                                                'out_octets': '1522043202725',
                                                                                                'out_ucast_pkts': '4382952624',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '556573',
                                                                       'out_octets': '1522043202725',
                                                                       'out_ucast_pkts': '4382952624',
                                                                       'poller_hash': 'MTE3REI3QXBlZXJpbnRmcG9uIDI1MTgxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251815',
                                                                                               'links': [       {       'connected_time': '30330383',
                                                                                                                        'fiber_length': 20155,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI3QXBlZXJwb25wb24gMjUxODE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4119 LINE 46 (QUAST CHERYL)',
                                         'poller_hash': 'MTE3REI3QQ==',
                                         'serial_number': '117DB7A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4b:75',
                                         'uptime': '30330383'},
                                 {       'counter_source': '252600/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ea:8c'],
                                                                       'id': '252600/g1',
                                                                       'in_nucast_pkts': '14966',
                                                                       'in_octets': '9485851040',
                                                                       'in_ucast_pkts': '22626189',
                                                                       'mac_address': 'b8:94:70:a1:50:08',
                                                                       'mtu': 2000,
                                                                       'name': '252600/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465159',
                                                                       'out_octets': '105645393221',
                                                                       'out_ucast_pkts': '75527400',
                                                                       'poller_hash': 'MTE3REMwQ3BlZXJpbnRmMjUyNjAwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '252600',
                                                                       'in_nucast_pkts': '1465159',
                                                                       'in_octets': '105645393221',
                                                                       'in_ucast_pkts': '75527400',
                                                                       'links': [       {       'in_nucast_pkts': '1465159',
                                                                                                'in_octets': '105645393221',
                                                                                                'in_ucast_pkts': '75527400',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14966',
                                                                                                'out_octets': '9485851040',
                                                                                                'out_ucast_pkts': '22626189',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14966',
                                                                       'out_octets': '9485851040',
                                                                       'out_ucast_pkts': '22626189',
                                                                       'poller_hash': 'MTE3REMwQ3BlZXJpbnRmcG9uIDI1MjYwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252600',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 2756,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMwQ3BlZXJwb25wb24gMjUyNjAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '121 PUGH ST (STONE STEPHEN)',
                                         'poller_hash': 'MTE3REMwQw==',
                                         'serial_number': '117DC0C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:05',
                                         'uptime': '688370'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '252610/g1',
                                                                       'name': '252610/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3RDk3RHBlZXJpbnRmMjUyNjEwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '252610',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3RDk3RHBlZXJpbnRmcG9uIDI1MjYxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252610',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3RDk3RHBlZXJwb25wb24gMjUyNjEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '25 MURRAY CT (SMITH JESSICA)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "252610" '
                                                                  '(location: GPON '
                                                                  'port "12/6"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2025-07-24T17:41:56.480Z',
                                         'poller_hash': 'MTE3RDk3RA==',
                                         'serial_number': '117D97D',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '252620/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '252620/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '85974671117',
                                                                       'in_ucast_pkts': '922975445',
                                                                       'mac_address': 'b8:94:70:a1:4f:ed',
                                                                       'name': '252620/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5710108908743',
                                                                       'out_ucast_pkts': '8072364026',
                                                                       'poller_hash': 'MTE3REMwOXBlZXJpbnRmMjUyNjIwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '252620',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5710108908743',
                                                                       'in_ucast_pkts': '8072364026',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '5710108908743',
                                                                                                'in_ucast_pkts': '8072364026',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '85974671117',
                                                                                                'out_ucast_pkts': '922975445',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '85974671117',
                                                                       'out_ucast_pkts': '922975445',
                                                                       'poller_hash': 'MTE3REMwOXBlZXJpbnRmcG9uIDI1MjYyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252620',
                                                                                               'links': [       {       'connected_time': '26622824',
                                                                                                                        'fiber_length': 6461,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMwOXBlZXJwb25wb24gMjUyNjIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7006 ROAD 129 (WAGLER BRIAN)',
                                         'poller_hash': 'MTE3REMwOQ==',
                                         'serial_number': '117DC09',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4f:ed',
                                         'uptime': '26622824'},
                                 {       'counter_source': '252700/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '252700/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '160782063685',
                                                                       'in_ucast_pkts': '30561320095',
                                                                       'mac_address': 'b8:94:70:a1:4f:55',
                                                                       'name': '252700/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '2713951606825',
                                                                       'out_ucast_pkts': '2047817019',
                                                                       'poller_hash': 'MTE3REJGNnBlZXJpbnRmMjUyNzAwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '252700',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '2713951606825',
                                                                       'in_ucast_pkts': '2047817019',
                                                                       'links': [       {       'in_nucast_pkts': '12',
                                                                                                'in_octets': '2713951606825',
                                                                                                'in_ucast_pkts': '2047817019',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '160782063685',
                                                                                                'out_ucast_pkts': '30561320095',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '160782063685',
                                                                       'out_ucast_pkts': '30561320095',
                                                                       'poller_hash': 'MTE3REJGNnBlZXJpbnRmcG9uIDI1MjcwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252700',
                                                                                               'links': [       {       'connected_time': '688601',
                                                                                                                        'fiber_length': 6730,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJGNnBlZXJwb25wb24gMjUyNzAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4433 PERTH LINE 72 (HAFEMANN '
                                                 'CHRISTIAN)',
                                         'poller_hash': 'MTE3REJGNg==',
                                         'serial_number': '117DBF6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4f:55',
                                         'uptime': '688601'},
                                 {       'counter_source': '252740/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ea:ce'],
                                                                       'id': '252740/g1',
                                                                       'in_nucast_pkts': '5647',
                                                                       'in_octets': '1308860074',
                                                                       'in_ucast_pkts': '4474062',
                                                                       'mac_address': 'b8:94:70:a1:50:70',
                                                                       'mtu': 2000,
                                                                       'name': '252740/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '541481',
                                                                       'out_octets': '12171894786',
                                                                       'out_ucast_pkts': '9552915',
                                                                       'poller_hash': 'MTE3REMxOXBlZXJpbnRmMjUyNzQwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '252740',
                                                                       'in_nucast_pkts': '541481',
                                                                       'in_octets': '12171894786',
                                                                       'in_ucast_pkts': '9552915',
                                                                       'links': [       {       'in_nucast_pkts': '541481',
                                                                                                'in_octets': '12171894786',
                                                                                                'in_ucast_pkts': '9552915',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5647',
                                                                                                'out_octets': '1308860074',
                                                                                                'out_ucast_pkts': '4474062',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5647',
                                                                       'out_octets': '1308860074',
                                                                       'out_ucast_pkts': '4474062',
                                                                       'poller_hash': 'MTE3REMxOXBlZXJpbnRmcG9uIDI1Mjc0MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252740',
                                                                                               'links': [       {       'connected_time': '262471',
                                                                                                                        'fiber_length': 15491,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMxOXBlZXJwb25wb24gMjUyNzQw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5264 ROAD 108 (ROTH TIM)',
                                         'poller_hash': 'MTE3REMxOQ==',
                                         'serial_number': '117DC19',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:6d',
                                         'uptime': '262471'},
                                 {       'counter_source': '252750/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a1:e5:81'],
                                                                       'id': '252750/g1',
                                                                       'in_nucast_pkts': '12010',
                                                                       'in_octets': '14075602334',
                                                                       'in_ucast_pkts': '33480663',
                                                                       'mac_address': 'b8:94:70:a1:4f:30',
                                                                       'mtu': 2000,
                                                                       'name': '252750/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465217',
                                                                       'out_octets': '154977066180',
                                                                       'out_ucast_pkts': '119402844',
                                                                       'poller_hash': 'MTE3REJGMXBlZXJpbnRmMjUyNzUwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '252750',
                                                                       'in_nucast_pkts': '1465217',
                                                                       'in_octets': '154977066180',
                                                                       'in_ucast_pkts': '119402844',
                                                                       'links': [       {       'in_nucast_pkts': '1465217',
                                                                                                'in_octets': '154977066180',
                                                                                                'in_ucast_pkts': '119402844',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12010',
                                                                                                'out_octets': '14075602334',
                                                                                                'out_ucast_pkts': '33480663',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12010',
                                                                       'out_octets': '14075602334',
                                                                       'out_ucast_pkts': '33480663',
                                                                       'poller_hash': 'MTE3REJGMXBlZXJpbnRmcG9uIDI1Mjc1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252750',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 3309,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJGMXBlZXJwb25wb24gMjUyNzUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '66 PUGH ST (PEARCE DOUG)',
                                         'poller_hash': 'MTE3REJGMQ==',
                                         'serial_number': '117DBF1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4f:2d',
                                         'uptime': '688369'},
                                 {       'counter_source': '252805/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ea:60'],
                                                                       'id': '252805/g1',
                                                                       'in_nucast_pkts': '526500',
                                                                       'in_octets': '65423524573',
                                                                       'in_ucast_pkts': '616125902',
                                                                       'mac_address': 'b8:94:70:a1:4d:38',
                                                                       'mtu': 2000,
                                                                       'name': '252805/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '69208772',
                                                                       'out_octets': '3342960444217',
                                                                       'out_ucast_pkts': '2261332491',
                                                                       'poller_hash': 'MTE3REJCMnBlZXJpbnRmMjUyODA1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '252805',
                                                                       'in_nucast_pkts': '69208772',
                                                                       'in_octets': '3342960444217',
                                                                       'in_ucast_pkts': '2261332491',
                                                                       'links': [       {       'in_nucast_pkts': '69208772',
                                                                                                'in_octets': '3342960444217',
                                                                                                'in_ucast_pkts': '2261332491',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '526500',
                                                                                                'out_octets': '65423524573',
                                                                                                'out_ucast_pkts': '616125902',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '526500',
                                                                       'out_octets': '65423524573',
                                                                       'out_ucast_pkts': '616125902',
                                                                       'poller_hash': 'MTE3REJCMnBlZXJpbnRmcG9uIDI1MjgwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252805',
                                                                                               'links': [       {       'connected_time': '29484828',
                                                                                                                        'fiber_length': 6914,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJCMnBlZXJwb25wb24gMjUyODA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4463 COMMUNITY CENTRE DR (CARTER '
                                                 'RUTH ANN)',
                                         'poller_hash': 'MTE3REJCMg==',
                                         'serial_number': '117DBB2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4d:35',
                                         'uptime': '29484828'},
                                 {       'counter_source': '252815/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '252815/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45238237132',
                                                                       'in_ucast_pkts': '8652368305',
                                                                       'mac_address': 'b8:94:70:a1:48:15',
                                                                       'name': '252815/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '102849910346',
                                                                       'out_ucast_pkts': '131482813',
                                                                       'poller_hash': 'MTE3REIwRXBlZXJpbnRmMjUyODE1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '252815',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '102849910346',
                                                                       'in_ucast_pkts': '131482813',
                                                                       'links': [       {       'in_nucast_pkts': '16',
                                                                                                'in_octets': '102849910346',
                                                                                                'in_ucast_pkts': '131482813',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45238237132',
                                                                                                'out_ucast_pkts': '8652368305',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45238237132',
                                                                       'out_ucast_pkts': '8652368305',
                                                                       'poller_hash': 'MTE3REIwRXBlZXJpbnRmcG9uIDI1MjgxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252815',
                                                                                               'links': [       {       'connected_time': '29471886',
                                                                                                                        'fiber_length': 6915,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIwRXBlZXJwb25wb24gMjUyODE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4452 COMMUNITY CENTRE DR (GERBER '
                                                 'RONALD)',
                                         'poller_hash': 'MTE3REIwRQ==',
                                         'serial_number': '117DB0E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:48:15',
                                         'uptime': '29471886'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '252825/g1',
                                                                       'name': '252825/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkZFQUYzcGVlcmludGYyNTI4MjUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '252825',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MkZFQUYzcGVlcmludGZwb24gMjUyODI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252825',
                                                                                               'links': [       {       'connected_time': '29465198',
                                                                                                                        'fiber_length': 7459,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.716,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQUYzcGVlcnBvbnBvbiAyNTI4MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6879A ROAD 140 (KUEPFER MARVIN & '
                                                 'SARAH)',
                                         'poller_hash': 'MkZFQUYz',
                                         'serial_number': '2FEAF3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:22:48',
                                         'uptime': '29465198'},
                                 {       'counter_source': '252855/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ed:83'],
                                                                       'id': '252855/g1',
                                                                       'in_nucast_pkts': '11789',
                                                                       'in_octets': '5915718053',
                                                                       'in_ucast_pkts': '35114400',
                                                                       'mac_address': 'b8:94:70:a1:4c:68',
                                                                       'mtu': 2000,
                                                                       'name': '252855/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465199',
                                                                       'out_octets': '115650603427',
                                                                       'out_ucast_pkts': '86134650',
                                                                       'poller_hash': 'MTE3REI5OHBlZXJpbnRmMjUyODU1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '252855',
                                                                       'in_nucast_pkts': '1465199',
                                                                       'in_octets': '115650603427',
                                                                       'in_ucast_pkts': '86134650',
                                                                       'links': [       {       'in_nucast_pkts': '1465199',
                                                                                                'in_octets': '115650603427',
                                                                                                'in_ucast_pkts': '86134650',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11789',
                                                                                                'out_octets': '5915718053',
                                                                                                'out_ucast_pkts': '35114400',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11789',
                                                                       'out_octets': '5915718053',
                                                                       'out_ucast_pkts': '35114400',
                                                                       'poller_hash': 'MTE3REI5OHBlZXJpbnRmcG9uIDI1Mjg1NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252855',
                                                                                               'links': [       {       'connected_time': '688375',
                                                                                                                        'fiber_length': 3184,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI5OHBlZXJwb25wb24gMjUyODU1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '46 AVERY PL (BENDER BOB & JANINE)',
                                         'poller_hash': 'MTE3REI5OA==',
                                         'serial_number': '117DB98',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4c:65',
                                         'uptime': '688375'},
                                 {       'counter_source': '252885/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '252885/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '46205873351',
                                                                       'in_ucast_pkts': '549923591',
                                                                       'mac_address': 'b8:94:70:a1:47:0d',
                                                                       'name': '252885/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '3393074682714',
                                                                       'out_ucast_pkts': '2279598595',
                                                                       'poller_hash': 'MTE3REFFRHBlZXJpbnRmMjUyODg1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '252885',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '3393074682714',
                                                                       'in_ucast_pkts': '2279598595',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '3393074682714',
                                                                                                'in_ucast_pkts': '2279598595',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '46205873351',
                                                                                                'out_ucast_pkts': '549923591',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '46205873351',
                                                                       'out_ucast_pkts': '549923591',
                                                                       'poller_hash': 'MTE3REFFRHBlZXJpbnRmcG9uIDI1Mjg4NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252885',
                                                                                               'links': [       {       'connected_time': '29057149',
                                                                                                                        'fiber_length': 4793,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFFRHBlZXJwb25wb24gMjUyODg1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6569 ROAD 129 (DIETRICH DIANE)',
                                         'poller_hash': 'MTE3REFFRA==',
                                         'serial_number': '117DAED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:47:0d',
                                         'uptime': '29057149'},
                                 {       'counter_source': '253356/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253356/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '70459561442',
                                                                       'in_ucast_pkts': '189401775',
                                                                       'mac_address': 'b8:94:70:a1:4d:25',
                                                                       'name': '253356/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '530706138888',
                                                                       'out_ucast_pkts': '475516923',
                                                                       'poller_hash': 'MTE3REJCMHBlZXJpbnRmMjUzMzU2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253356',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '530706138888',
                                                                       'in_ucast_pkts': '475516923',
                                                                       'links': [       {       'in_nucast_pkts': '10',
                                                                                                'in_octets': '530706138888',
                                                                                                'in_ucast_pkts': '475516923',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '70459561442',
                                                                                                'out_ucast_pkts': '189401775',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '70459561442',
                                                                       'out_ucast_pkts': '189401775',
                                                                       'poller_hash': 'MTE3REJCMHBlZXJpbnRmcG9uIDI1MzM1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253356',
                                                                                               'links': [       {       'connected_time': '523337',
                                                                                                                        'fiber_length': 6929,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJCMHBlZXJwb25wb24gMjUzMzU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4464 COMMUNITY CENTRE DR (GINGRICH '
                                                 'NATHAN)',
                                         'poller_hash': 'MTE3REJCMA==',
                                         'serial_number': '117DBB0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4d:25',
                                         'uptime': '523337'},
                                 {       'counter_source': '253376/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253376/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '282495918827',
                                                                       'in_ucast_pkts': '1001016803',
                                                                       'mac_address': 'b8:94:70:a1:44:e5',
                                                                       'name': '253376/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '6',
                                                                       'out_octets': '3421236967153',
                                                                       'out_ucast_pkts': '2774293261',
                                                                       'poller_hash': 'MTE3REFBOHBlZXJpbnRmMjUzMzc2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253376',
                                                                       'in_nucast_pkts': '6',
                                                                       'in_octets': '3421236967153',
                                                                       'in_ucast_pkts': '2774293261',
                                                                       'links': [       {       'in_nucast_pkts': '6',
                                                                                                'in_octets': '3421236967153',
                                                                                                'in_ucast_pkts': '2774293261',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '282495918827',
                                                                                                'out_ucast_pkts': '1001016803',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '282495918827',
                                                                       'out_ucast_pkts': '1001016803',
                                                                       'poller_hash': 'MTE3REFBOHBlZXJpbnRmcG9uIDI1MzM3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253376',
                                                                                               'links': [       {       'connected_time': '24794',
                                                                                                                        'fiber_length': 6908,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFBOHBlZXJwb25wb24gMjUzMzc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4458 COMMUNITY CENTRE DR (HAWLEY '
                                                 'ALAN)',
                                         'poller_hash': 'MTE3REFBOA==',
                                         'serial_number': '117DAA8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:44:e5',
                                         'uptime': '24794'},
                                 {       'counter_source': '253386/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:33:31:e6'],
                                                                       'id': '253386/g2',
                                                                       'in_nucast_pkts': '2845273',
                                                                       'in_octets': '104710526962',
                                                                       'in_ucast_pkts': '1400990618',
                                                                       'mac_address': 'b8:94:70:a1:50:99',
                                                                       'mtu': 2000,
                                                                       'name': '253386/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '204884793',
                                                                       'out_octets': '7954307086489',
                                                                       'out_ucast_pkts': '5609899953',
                                                                       'poller_hash': 'MTE3REMxRXBlZXJpbnRmMjUzMzg2L2cy',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253386',
                                                                       'in_nucast_pkts': '204884793',
                                                                       'in_octets': '7954307086489',
                                                                       'in_ucast_pkts': '5609899953',
                                                                       'links': [       {       'in_nucast_pkts': '204884793',
                                                                                                'in_octets': '7954307086489',
                                                                                                'in_ucast_pkts': '5609899953',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2845273',
                                                                                                'out_octets': '104710526962',
                                                                                                'out_ucast_pkts': '1400990618',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2845273',
                                                                       'out_octets': '104710526962',
                                                                       'out_ucast_pkts': '1400990618',
                                                                       'poller_hash': 'MTE3REMxRXBlZXJpbnRmcG9uIDI1MzM4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253386',
                                                                                               'links': [       {       'connected_time': '29035616',
                                                                                                                        'fiber_length': 6509,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMxRXBlZXJwb25wb24gMjUzMzg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4446 PERTH LINE 72 (WAGLER SHIRLEY)',
                                         'poller_hash': 'MTE3REMxRQ==',
                                         'serial_number': '117DC1E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:95',
                                         'uptime': '29035616'},
                                 {       'counter_source': '253396/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253396/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '199656288604',
                                                                       'in_ucast_pkts': '1511768662',
                                                                       'mac_address': 'b8:94:70:a1:40:8d',
                                                                       'name': '253396/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '7224961585690',
                                                                       'out_ucast_pkts': '5136990995',
                                                                       'poller_hash': 'MTE3REExRHBlZXJpbnRmMjUzMzk2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253396',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '7224961585690',
                                                                       'in_ucast_pkts': '5136990995',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '7224961585690',
                                                                                                'in_ucast_pkts': '5136990995',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '199656288604',
                                                                                                'out_ucast_pkts': '1511768662',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '199656288604',
                                                                       'out_ucast_pkts': '1511768662',
                                                                       'poller_hash': 'MTE3REExRHBlZXJpbnRmcG9uIDI1MzM5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253396',
                                                                                               'links': [       {       'connected_time': '538378',
                                                                                                                        'fiber_length': 24757,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REExRHBlZXJwb25wb24gMjUzMzk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4538 LINE 46 (ROSE JOHN P)',
                                         'poller_hash': 'MTE3REExRA==',
                                         'serial_number': '117DA1D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:40:8d',
                                         'uptime': '538378'},
                                 {       'counter_source': '253416/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253416/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8817195169',
                                                                       'in_ucast_pkts': '27041219',
                                                                       'mac_address': 'b8:94:70:a1:4f:a5',
                                                                       'name': '253416/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '92518550762',
                                                                       'out_ucast_pkts': '105955224',
                                                                       'poller_hash': 'MTE3REMwMHBlZXJpbnRmMjUzNDE2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253416',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '92518550762',
                                                                       'in_ucast_pkts': '105955224',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '92518550762',
                                                                                                'in_ucast_pkts': '105955224',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '8817195169',
                                                                                                'out_ucast_pkts': '27041219',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8817195169',
                                                                       'out_ucast_pkts': '27041219',
                                                                       'poller_hash': 'MTE3REMwMHBlZXJpbnRmcG9uIDI1MzQxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253416',
                                                                                               'links': [       {       'connected_time': '28968599',
                                                                                                                        'fiber_length': 6977,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMwMHBlZXJwb25wb24gMjUzNDE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4466 COMMUNITY CENTRE DR (JANTZI '
                                                 'EMMANUEL)',
                                         'poller_hash': 'MTE3REMwMA==',
                                         'serial_number': '117DC00',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4f:a5',
                                         'uptime': '28968599'},
                                 {       'counter_source': '253436/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253436/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1328043959817',
                                                                       'in_ucast_pkts': '10746913662',
                                                                       'mac_address': 'cc:be:59:ef:f7:36',
                                                                       'name': '253436/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '22343736429338',
                                                                       'out_ucast_pkts': '20976038016',
                                                                       'poller_hash': 'NEQ0MjM5cGVlcmludGYyNTM0MzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '253436',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '22343736429338',
                                                                       'in_ucast_pkts': '20976038016',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '22343736429338',
                                                                                                'in_ucast_pkts': '20976038016',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1328043959817',
                                                                                                'out_ucast_pkts': '10746913662',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1328043959817',
                                                                       'out_ucast_pkts': '10746913662',
                                                                       'poller_hash': 'NEQ0MjM5cGVlcmludGZwb24gMjUzNDM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253436',
                                                                                               'links': [       {       'connected_time': '47956',
                                                                                                                        'fiber_length': 17672,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.79,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjM5cGVlcnBvbnBvbiAyNTM0MzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '214 SELBY AVENUE (WARD KELLY-ANN)',
                                         'poller_hash': 'NEQ0MjM5',
                                         'serial_number': '4D4239',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:36',
                                         'uptime': '47956'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '253446/g1',
                                                                       'name': '253446/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REFGQ3BlZXJpbnRmMjUzNDQ2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253446',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REFGQ3BlZXJpbnRmcG9uIDI1MzQ0Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253446',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFGQ3BlZXJwb25wb24gMjUzNDQ2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '6889 ROAD 124 (JANTZI REUBEN K)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "253446" '
                                                                  '(location: GPON '
                                                                  'port "13/2"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-05-12T23:10:00.920Z',
                                         'poller_hash': 'MTE3REFGQw==',
                                         'serial_number': '117DAFC',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '253466/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253466/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45196676748',
                                                                       'in_ucast_pkts': '335476661',
                                                                       'mac_address': 'b8:94:70:a1:48:b5',
                                                                       'name': '253466/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '1649261899676',
                                                                       'out_ucast_pkts': '1203897719',
                                                                       'poller_hash': 'MTE3REIyMnBlZXJpbnRmMjUzNDY2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253466',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '1649261899676',
                                                                       'in_ucast_pkts': '1203897719',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '1649261899676',
                                                                                                'in_ucast_pkts': '1203897719',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45196676748',
                                                                                                'out_ucast_pkts': '335476661',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45196676748',
                                                                       'out_ucast_pkts': '335476661',
                                                                       'poller_hash': 'MTE3REIyMnBlZXJpbnRmcG9uIDI1MzQ2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253466',
                                                                                               'links': [       {       'connected_time': '523339',
                                                                                                                        'fiber_length': 6863,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIyMnBlZXJwb25wb24gMjUzNDY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4461 COMMUNITY CENTRE DR (WAGLER '
                                                 'EDGAR)',
                                         'poller_hash': 'MTE3REIyMg==',
                                         'serial_number': '117DB22',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:48:b5',
                                         'uptime': '523339'},
                                 {       'counter_source': '253476/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253476/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6380159620',
                                                                       'in_ucast_pkts': '23178655',
                                                                       'mac_address': 'b8:94:70:a1:49:45',
                                                                       'name': '253476/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '88523206755',
                                                                       'out_ucast_pkts': '97683470',
                                                                       'poller_hash': 'MTE3REIzNHBlZXJpbnRmMjUzNDc2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253476',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '88523206755',
                                                                       'in_ucast_pkts': '97683470',
                                                                       'links': [       {       'in_nucast_pkts': '16',
                                                                                                'in_octets': '88523206755',
                                                                                                'in_ucast_pkts': '97683470',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '6380159620',
                                                                                                'out_ucast_pkts': '23178655',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6380159620',
                                                                       'out_ucast_pkts': '23178655',
                                                                       'poller_hash': 'MTE3REIzNHBlZXJpbnRmcG9uIDI1MzQ3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253476',
                                                                                               'links': [       {       'connected_time': '28873352',
                                                                                                                        'fiber_length': 6785,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIzNHBlZXJwb25wb24gMjUzNDc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4455 COMMUNITY CENTRE DR (ROLFE '
                                                 'ROSANNE)',
                                         'poller_hash': 'MTE3REIzNA==',
                                         'serial_number': '117DB34',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:49:45',
                                         'uptime': '28873352'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '253486/g1',
                                                                       'name': '253486/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REI1OHBlZXJpbnRmMjUzNDg2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253486',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REI1OHBlZXJpbnRmcG9uIDI1MzQ4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253486',
                                                                                               'links': [       {       'connected_time': '28878991',
                                                                                                                        'fiber_length': 17844,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI1OHBlZXJwb25wb24gMjUzNDg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2460 LINE 47 (GERBER COLLIN)',
                                         'poller_hash': 'MTE3REI1OA==',
                                         'serial_number': '117DB58',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:65',
                                         'uptime': '28878991'},
                                 {       'counter_source': '253496/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f4:6e'],
                                                                       'id': '253496/g1',
                                                                       'in_nucast_pkts': '4543',
                                                                       'in_octets': '587983643',
                                                                       'in_ucast_pkts': '1870851',
                                                                       'mac_address': 'b8:94:70:a1:50:f8',
                                                                       'mtu': 2000,
                                                                       'name': '253496/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '541495',
                                                                       'out_octets': '10391722720',
                                                                       'out_ucast_pkts': '7816061',
                                                                       'poller_hash': 'MTE3REMyQXBlZXJpbnRmMjUzNDk2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253496',
                                                                       'in_nucast_pkts': '541495',
                                                                       'in_octets': '10391722720',
                                                                       'in_ucast_pkts': '7816061',
                                                                       'links': [       {       'in_nucast_pkts': '541495',
                                                                                                'in_octets': '10391722720',
                                                                                                'in_ucast_pkts': '7816061',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4543',
                                                                                                'out_octets': '587983643',
                                                                                                'out_ucast_pkts': '1870851',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4543',
                                                                       'out_octets': '587983643',
                                                                       'out_ucast_pkts': '1870851',
                                                                       'poller_hash': 'MTE3REMyQXBlZXJpbnRmcG9uIDI1MzQ5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253496',
                                                                                               'links': [       {       'connected_time': '262228',
                                                                                                                        'fiber_length': 22072,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMyQXBlZXJwb25wb24gMjUzNDk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4847 ROAD 109 (HORNE ARTHUR)',
                                         'poller_hash': 'MTE3REMyQQ==',
                                         'serial_number': '117DC2A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:f5',
                                         'uptime': '262228'},
                                 {       'counter_source': '253506/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253506/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '123963984683',
                                                                       'in_ucast_pkts': '417850523',
                                                                       'mac_address': 'b8:94:70:a1:47:65',
                                                                       'name': '253506/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1298381240158',
                                                                       'out_ucast_pkts': '995175132',
                                                                       'poller_hash': 'MTE3REFGOHBlZXJpbnRmMjUzNTA2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253506',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1298381240158',
                                                                       'in_ucast_pkts': '995175132',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1298381240158',
                                                                                                'in_ucast_pkts': '995175132',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '123963984683',
                                                                                                'out_ucast_pkts': '417850523',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '123963984683',
                                                                       'out_ucast_pkts': '417850523',
                                                                       'poller_hash': 'MTE3REFGOHBlZXJpbnRmcG9uIDI1MzUwNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253506',
                                                                                               'links': [       {       'connected_time': '538592',
                                                                                                                        'fiber_length': 19395,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFGOHBlZXJwb25wb24gMjUzNTA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2624 LINE 47 (PATCH DALE)',
                                         'poller_hash': 'MTE3REFGOA==',
                                         'serial_number': '117DAF8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:47:65',
                                         'uptime': '538592'},
                                 {       'counter_source': '253527/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:70:e7'],
                                                                       'id': '253527/g1',
                                                                       'in_nucast_pkts': '256768',
                                                                       'in_octets': '209580692829',
                                                                       'in_ucast_pkts': '257107761',
                                                                       'mac_address': 'b8:94:70:a1:4f:48',
                                                                       'mtu': 2000,
                                                                       'name': '253527/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31045783',
                                                                       'out_octets': '415548364182',
                                                                       'out_ucast_pkts': '446691296',
                                                                       'poller_hash': 'MTE3REJGNHBlZXJpbnRmMjUzNTI3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253527',
                                                                       'in_nucast_pkts': '31045783',
                                                                       'in_octets': '415548364182',
                                                                       'in_ucast_pkts': '446691296',
                                                                       'links': [       {       'in_nucast_pkts': '31045783',
                                                                                                'in_octets': '415548364182',
                                                                                                'in_ucast_pkts': '446691296',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '256768',
                                                                                                'out_octets': '209580692829',
                                                                                                'out_ucast_pkts': '257107761',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '256768',
                                                                       'out_octets': '209580692829',
                                                                       'out_ucast_pkts': '257107761',
                                                                       'poller_hash': 'MTE3REJGNHBlZXJpbnRmcG9uIDI1MzUyNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253527',
                                                                                               'links': [       {       'connected_time': '13642123',
                                                                                                                        'fiber_length': 1020,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJGNHBlZXJwb25wb24gMjUzNTI3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6B MAIN ST S (NEW ORLEANS PIZZA)',
                                         'poller_hash': 'MTE3REJGNA==',
                                         'serial_number': '117DBF4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4f:45',
                                         'uptime': '13642123'},
                                 {       'counter_source': '253537/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253537/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12955731561',
                                                                       'in_ucast_pkts': '61565163',
                                                                       'mac_address': 'b8:94:70:a1:4d:f5',
                                                                       'name': '253537/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '396613973719',
                                                                       'out_ucast_pkts': '313195627',
                                                                       'poller_hash': 'MTE3REJDQXBlZXJpbnRmMjUzNTM3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253537',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '396613973719',
                                                                       'in_ucast_pkts': '313195627',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '396613973719',
                                                                                                'in_ucast_pkts': '313195627',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '12955731561',
                                                                                                'out_ucast_pkts': '61565163',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12955731561',
                                                                       'out_ucast_pkts': '61565163',
                                                                       'poller_hash': 'MTE3REJDQXBlZXJpbnRmcG9uIDI1MzUzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253537',
                                                                                               'links': [       {       'connected_time': '28787924',
                                                                                                                        'fiber_length': 21660,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJDQXBlZXJwb25wb24gMjUzNTM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2620 LINE 45 (SATCHELL DOUGLAS & '
                                                 'BONNIE)',
                                         'poller_hash': 'MTE3REJDQQ==',
                                         'serial_number': '117DBCA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4d:f5',
                                         'uptime': '28787924'},
                                 {       'counter_source': '253952/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '253952/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '77298465346',
                                                                       'in_ucast_pkts': '424611174',
                                                                       'mac_address': 'b8:94:70:a1:4e:9d',
                                                                       'name': '253952/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2322087573748',
                                                                       'out_ucast_pkts': '1686896283',
                                                                       'poller_hash': 'MTE3REJERnBlZXJpbnRmMjUzOTUyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '253952',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2322087573748',
                                                                       'in_ucast_pkts': '1686896283',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2322087573748',
                                                                                                'in_ucast_pkts': '1686896283',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '77298465346',
                                                                                                'out_ucast_pkts': '424611174',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '77298465346',
                                                                       'out_ucast_pkts': '424611174',
                                                                       'poller_hash': 'MTE3REJERnBlZXJpbnRmcG9uIDI1Mzk1Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253952',
                                                                                               'links': [       {       'connected_time': '538152',
                                                                                                                        'fiber_length': 18280,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJERnBlZXJwb25wb24gMjUzOTUy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2499 LINE 47 (ROTH JOHN & INGRID)',
                                         'poller_hash': 'MTE3REJERg==',
                                         'serial_number': '117DBDF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4e:9d',
                                         'uptime': '538152'},
                                 {       'counter_source': '253992/g1+253992/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd4:6e:0e:9f:e4:9b'],
                                                                       'id': '253992/g1',
                                                                       'in_nucast_pkts': '4529',
                                                                       'in_octets': '26013089316',
                                                                       'in_ucast_pkts': '123976902',
                                                                       'mac_address': 'b8:94:70:a1:4b:18',
                                                                       'mtu': 2000,
                                                                       'name': '253992/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33839768',
                                                                       'out_octets': '551150827226',
                                                                       'out_ucast_pkts': '447337235',
                                                                       'poller_hash': 'MTE3REI2RXBlZXJpbnRmMjUzOTkyL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:ee:0c:28:e6:4f'],
                                                                       'id': '253992/g2',
                                                                       'in_nucast_pkts': '880',
                                                                       'in_octets': '137959497692',
                                                                       'in_ucast_pkts': '560547297',
                                                                       'mac_address': 'b8:94:70:a1:4b:19',
                                                                       'mtu': 2000,
                                                                       'name': '253992/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33844926',
                                                                       'out_octets': '2866336123837',
                                                                       'out_ucast_pkts': '2317755207',
                                                                       'poller_hash': 'MTE3REI2RXBlZXJpbnRmMjUzOTkyL2cy',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253992',
                                                                       'in_nucast_pkts': '67684694',
                                                                       'in_octets': '3417486951063',
                                                                       'in_ucast_pkts': '2765092442',
                                                                       'links': [       {       'in_nucast_pkts': '67684694',
                                                                                                'in_octets': '3417486951063',
                                                                                                'in_ucast_pkts': '2765092442',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5409',
                                                                                                'out_octets': '163972587008',
                                                                                                'out_ucast_pkts': '684524199',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5409',
                                                                       'out_octets': '163972587008',
                                                                       'out_ucast_pkts': '684524199',
                                                                       'poller_hash': 'MTE3REI2RXBlZXJpbnRmcG9uIDI1Mzk5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253992',
                                                                                               'links': [       {       'connected_time': '14936535',
                                                                                                                        'fiber_length': 6683,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI2RXBlZXJwb25wb24gMjUzOTky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4467 PERTH LINE 72 (BUECKERT PETER & '
                                                 'TRUDY)',
                                         'poller_hash': 'MTE3REI2RQ==',
                                         'serial_number': '117DB6E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4b:15',
                                         'uptime': '14936535'},
                                 {       'counter_source': '254002/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254002/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '112430510647',
                                                                       'in_ucast_pkts': '315316346',
                                                                       'mac_address': 'b8:94:70:a1:43:0d',
                                                                       'name': '254002/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1783964377144',
                                                                       'out_ucast_pkts': '1299348899',
                                                                       'poller_hash': 'MTE3REE2RHBlZXJpbnRmMjU0MDAyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254002',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1783964377144',
                                                                       'in_ucast_pkts': '1299348899',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1783964377144',
                                                                                                'in_ucast_pkts': '1299348899',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '112430510647',
                                                                                                'out_ucast_pkts': '315316346',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '112430510647',
                                                                       'out_ucast_pkts': '315316346',
                                                                       'poller_hash': 'MTE3REE2RHBlZXJpbnRmcG9uIDI1NDAwMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254002',
                                                                                               'links': [       {       'connected_time': '688599',
                                                                                                                        'fiber_length': 12125,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REE2RHBlZXJwb25wb24gMjU0MDAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6267 PERTH ROAD 121 (ROES TYRELL)',
                                         'poller_hash': 'MTE3REE2RA==',
                                         'serial_number': '117DA6D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:43:0d',
                                         'uptime': '688599'},
                                 {       'counter_source': '254012/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254012/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '155991663999',
                                                                       'in_ucast_pkts': '1055230506',
                                                                       'mac_address': 'b8:94:70:a1:4e:a5',
                                                                       'name': '254012/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2983286586957',
                                                                       'out_ucast_pkts': '2407653660',
                                                                       'poller_hash': 'MTE3REJFMHBlZXJpbnRmMjU0MDEyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254012',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2983286586957',
                                                                       'in_ucast_pkts': '2407653660',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2983286586957',
                                                                                                'in_ucast_pkts': '2407653660',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '155991663999',
                                                                                                'out_ucast_pkts': '1055230506',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '155991663999',
                                                                       'out_ucast_pkts': '1055230506',
                                                                       'poller_hash': 'MTE3REJFMHBlZXJpbnRmcG9uIDI1NDAxMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254012',
                                                                                               'links': [       {       'connected_time': '262469',
                                                                                                                        'fiber_length': 15638,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJFMHBlZXJwb25wb24gMjU0MDEy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5261 ROAD 108 (NEEBVIEW ACRES LTD.)',
                                         'poller_hash': 'MTE3REJFMA==',
                                         'serial_number': '117DBE0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4e:a5',
                                         'uptime': '262469'},
                                 {       'counter_source': '254022/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254022/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '268386125620',
                                                                       'in_ucast_pkts': '34888193610',
                                                                       'mac_address': 'b8:94:70:a1:46:45',
                                                                       'name': '254022/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '1944950496886',
                                                                       'out_ucast_pkts': '1588880632',
                                                                       'poller_hash': 'MTE3REFENHBlZXJpbnRmMjU0MDIyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254022',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '1944950496886',
                                                                       'in_ucast_pkts': '1588880632',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '1944950496886',
                                                                                                'in_ucast_pkts': '1588880632',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '268386125620',
                                                                                                'out_ucast_pkts': '34888193610',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '268386125620',
                                                                       'out_ucast_pkts': '34888193610',
                                                                       'poller_hash': 'MTE3REFENHBlZXJpbnRmcG9uIDI1NDAyMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254022',
                                                                                               'links': [       {       'connected_time': '538376',
                                                                                                                        'fiber_length': 20673,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFENHBlZXJwb25wb24gMjU0MDIy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4188 LINE 46 (VAN MEURS MARTEN)',
                                         'poller_hash': 'MTE3REFENA==',
                                         'serial_number': '117DAD4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:46:45',
                                         'uptime': '538376'},
                                 {       'counter_source': '254052/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254052/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '30837273110',
                                                                       'in_ucast_pkts': '180501386',
                                                                       'mac_address': 'b8:94:70:a1:4f:bd',
                                                                       'name': '254052/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '963211212574',
                                                                       'out_ucast_pkts': '673052137',
                                                                       'poller_hash': 'MTE3REMwM3BlZXJpbnRmMjU0MDUyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254052',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '963211212574',
                                                                       'in_ucast_pkts': '673052137',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '963211212574',
                                                                                                'in_ucast_pkts': '673052137',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '30837273110',
                                                                                                'out_ucast_pkts': '180501386',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '30837273110',
                                                                       'out_ucast_pkts': '180501386',
                                                                       'poller_hash': 'MTE3REMwM3BlZXJpbnRmcG9uIDI1NDA1Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254052',
                                                                                               'links': [       {       'connected_time': '1803250',
                                                                                                                        'fiber_length': 17707,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMwM3BlZXJwb25wb24gMjU0MDUy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '220 WINSTANLEY ST MONKTON (HEWITT '
                                                 'ROBERT)',
                                         'poller_hash': 'MTE3REMwMw==',
                                         'serial_number': '117DC03',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4f:bd',
                                         'uptime': '1803250'},
                                 {       'counter_source': '254062/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:f5:fa'],
                                                                       'id': '254062/g1',
                                                                       'in_nucast_pkts': '14524',
                                                                       'in_octets': '5118339443',
                                                                       'in_ucast_pkts': '19347928',
                                                                       'mac_address': 'b8:94:70:a1:48:40',
                                                                       'mtu': 2000,
                                                                       'name': '254062/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465204',
                                                                       'out_octets': '98243891135',
                                                                       'out_ucast_pkts': '76570423',
                                                                       'poller_hash': 'MTE3REIxM3BlZXJpbnRmMjU0MDYyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254062',
                                                                       'in_nucast_pkts': '1465204',
                                                                       'in_octets': '98243891135',
                                                                       'in_ucast_pkts': '76570423',
                                                                       'links': [       {       'in_nucast_pkts': '1465204',
                                                                                                'in_octets': '98243891135',
                                                                                                'in_ucast_pkts': '76570423',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14524',
                                                                                                'out_octets': '5118339443',
                                                                                                'out_ucast_pkts': '19347928',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14524',
                                                                       'out_octets': '5118339443',
                                                                       'out_ucast_pkts': '19347928',
                                                                       'poller_hash': 'MTE3REIxM3BlZXJpbnRmcG9uIDI1NDA2Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254062',
                                                                                               'links': [       {       'connected_time': '688368',
                                                                                                                        'fiber_length': 2610,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIxM3BlZXJwb25wb24gMjU0MDYy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '115 PUGH ST (BASHNIAK MYKOLA)',
                                         'poller_hash': 'MTE3REIxMw==',
                                         'serial_number': '117DB13',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:48:3d',
                                         'uptime': '688368'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '254072/g1',
                                                                       'name': '254072/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REMyMXBlZXJpbnRmMjU0MDcyL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254072',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REMyMXBlZXJpbnRmcG9uIDI1NDA3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254072',
                                                                                               'links': [       {       'connected_time': '28343873',
                                                                                                                        'fiber_length': 5754,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMyMXBlZXJwb25wb24gMjU0MDcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4987 LINE 61 (STREICHER RON)',
                                         'poller_hash': 'MTE3REMyMQ==',
                                         'serial_number': '117DC21',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:ad',
                                         'uptime': '28343873'},
                                 {       'counter_source': '254173/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254173/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '176975052478',
                                                                       'in_ucast_pkts': '47831526962',
                                                                       'mac_address': 'b8:94:70:a1:41:7d',
                                                                       'name': '254173/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '2606343779156',
                                                                       'out_ucast_pkts': '2006067626',
                                                                       'poller_hash': 'MTE3REEzQnBlZXJpbnRmMjU0MTczL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254173',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '2606343779156',
                                                                       'in_ucast_pkts': '2006067626',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '2606343779156',
                                                                                                'in_ucast_pkts': '2006067626',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '176975052478',
                                                                                                'out_ucast_pkts': '47831526962',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '176975052478',
                                                                       'out_ucast_pkts': '47831526962',
                                                                       'poller_hash': 'MTE3REEzQnBlZXJpbnRmcG9uIDI1NDE3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254173',
                                                                                               'links': [       {       'connected_time': '523339',
                                                                                                                        'fiber_length': 6928,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REEzQnBlZXJwb25wb24gMjU0MTcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4460 COMMUNITY CENTRE DR (ALBRECHT '
                                                 'GREG)',
                                         'poller_hash': 'MTE3REEzQg==',
                                         'serial_number': '117DA3B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:41:7d',
                                         'uptime': '523339'},
                                 {       'counter_source': '254193/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:75:49'],
                                                                       'id': '254193/g1',
                                                                       'in_nucast_pkts': '12039',
                                                                       'in_octets': '7528115470',
                                                                       'in_ucast_pkts': '38881411',
                                                                       'mac_address': 'b8:94:70:a1:48:00',
                                                                       'mtu': 2000,
                                                                       'name': '254193/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465208',
                                                                       'out_octets': '346740958105',
                                                                       'out_ucast_pkts': '241043003',
                                                                       'poller_hash': 'MTE3REIwQnBlZXJpbnRmMjU0MTkzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254193',
                                                                       'in_nucast_pkts': '1465208',
                                                                       'in_octets': '346740958105',
                                                                       'in_ucast_pkts': '241043003',
                                                                       'links': [       {       'in_nucast_pkts': '1465208',
                                                                                                'in_octets': '346740958105',
                                                                                                'in_ucast_pkts': '241043003',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12039',
                                                                                                'out_octets': '7528115470',
                                                                                                'out_ucast_pkts': '38881411',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12039',
                                                                       'out_octets': '7528115470',
                                                                       'out_ucast_pkts': '38881411',
                                                                       'poller_hash': 'MTE3REIwQnBlZXJpbnRmcG9uIDI1NDE5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254193',
                                                                                               'links': [       {       'connected_time': '688493',
                                                                                                                        'fiber_length': 1009,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIwQnBlZXJwb25wb24gMjU0MTkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 MILL ST W (ANSTETT ADAM)',
                                         'poller_hash': 'MTE3REIwQg==',
                                         'serial_number': '117DB0B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:47:fd',
                                         'uptime': '688493'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '254258/g1',
                                                                       'name': '254258/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REI3NnBlZXJpbnRmMjU0MjU4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254258',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REI3NnBlZXJpbnRmcG9uIDI1NDI1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254258',
                                                                                               'links': [       {       'connected_time': '23840230',
                                                                                                                        'fiber_length': 6496,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI3NnBlZXJwb25wb24gMjU0MjU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4917 LINE 67 (EBERSOL ALVIN)',
                                         'poller_hash': 'MTE3REI3Ng==',
                                         'serial_number': '117DB76',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4b:55',
                                         'uptime': '23840230'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '254519/g1',
                                                                       'name': '254519/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REJDNHBlZXJpbnRmMjU0NTE5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254519',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REJDNHBlZXJpbnRmcG9uIDI1NDUxOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254519',
                                                                                               'links': [       {       'connected_time': '10653917',
                                                                                                                        'fiber_length': 18060,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJDNHBlZXJwb25wb24gMjU0NTE5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2435 LINE 47 (GERBER JAMES)',
                                         'poller_hash': 'MTE3REJDNA==',
                                         'serial_number': '117DBC4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4d:c5',
                                         'uptime': '10653917'},
                                 {       'counter_source': '254541/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:ec:ca'],
                                                                       'id': '254541/g1',
                                                                       'in_nucast_pkts': '5899',
                                                                       'in_octets': '30736083744',
                                                                       'in_ucast_pkts': '39212158',
                                                                       'mac_address': 'b8:94:70:a1:3c:40',
                                                                       'mtu': 2000,
                                                                       'name': '254541/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '647464',
                                                                       'out_octets': '121256067482',
                                                                       'out_ucast_pkts': '93443142',
                                                                       'poller_hash': 'MTE3RDk5M3BlZXJpbnRmMjU0NTQxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254541',
                                                                       'in_nucast_pkts': '647464',
                                                                       'in_octets': '121256067482',
                                                                       'in_ucast_pkts': '93443142',
                                                                       'links': [       {       'in_nucast_pkts': '647464',
                                                                                                'in_octets': '121256067482',
                                                                                                'in_ucast_pkts': '93443142',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5899',
                                                                                                'out_octets': '30736083744',
                                                                                                'out_ucast_pkts': '39212158',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5899',
                                                                       'out_octets': '30736083744',
                                                                       'out_ucast_pkts': '39212158',
                                                                       'poller_hash': 'MTE3RDk5M3BlZXJpbnRmcG9uIDI1NDU0MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254541',
                                                                                               'links': [       {       'connected_time': '688375',
                                                                                                                        'fiber_length': 3236,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3RDk5M3BlZXJwb25wb24gMjU0NTQx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29 AVERY PL (MCDOWELL PATTI)',
                                         'poller_hash': 'MTE3RDk5Mw==',
                                         'serial_number': '117D993',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:3c:3d',
                                         'uptime': '688375'},
                                 {       'counter_source': '254551/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6f:be'],
                                                                       'id': '254551/g1',
                                                                       'in_nucast_pkts': '519876',
                                                                       'in_octets': '156877185014',
                                                                       'in_ucast_pkts': '496675981',
                                                                       'mac_address': 'b8:94:70:a1:4a:08',
                                                                       'mtu': 2000,
                                                                       'name': '254551/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66419935',
                                                                       'out_octets': '1922894862945',
                                                                       'out_ucast_pkts': '1551779224',
                                                                       'poller_hash': 'MTE3REI0Q3BlZXJpbnRmMjU0NTUxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254551',
                                                                       'in_nucast_pkts': '66419935',
                                                                       'in_octets': '1922894862945',
                                                                       'in_ucast_pkts': '1551779224',
                                                                       'links': [       {       'in_nucast_pkts': '66419935',
                                                                                                'in_octets': '1922894862945',
                                                                                                'in_ucast_pkts': '1551779224',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '519876',
                                                                                                'out_octets': '156877185014',
                                                                                                'out_ucast_pkts': '496675981',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '519876',
                                                                       'out_octets': '156877185014',
                                                                       'out_ucast_pkts': '496675981',
                                                                       'poller_hash': 'MTE3REI0Q3BlZXJpbnRmcG9uIDI1NDU1MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254551',
                                                                                               'links': [       {       'connected_time': '27827087',
                                                                                                                        'fiber_length': 15434,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI0Q3BlZXJwb25wb24gMjU0NTUx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2483 LINE 51 (JANTZI ROBERT)',
                                         'poller_hash': 'MTE3REI0Qw==',
                                         'serial_number': '117DB4C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:05',
                                         'uptime': '27827087'},
                                 {       'counter_source': '254571/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6c:b1'],
                                                                       'id': '254571/g1',
                                                                       'in_nucast_pkts': '5370',
                                                                       'in_octets': '3753327020',
                                                                       'in_ucast_pkts': '7535005',
                                                                       'mac_address': 'b8:94:70:a1:46:c0',
                                                                       'mtu': 2000,
                                                                       'name': '254571/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '541504',
                                                                       'out_octets': '31709746042',
                                                                       'out_ucast_pkts': '25903738',
                                                                       'poller_hash': 'MTE3REFFM3BlZXJpbnRmMjU0NTcxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254571',
                                                                       'in_nucast_pkts': '541504',
                                                                       'in_octets': '31709746042',
                                                                       'in_ucast_pkts': '25903738',
                                                                       'links': [       {       'in_nucast_pkts': '541504',
                                                                                                'in_octets': '31709746042',
                                                                                                'in_ucast_pkts': '25903738',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5370',
                                                                                                'out_octets': '3753327020',
                                                                                                'out_ucast_pkts': '7535005',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5370',
                                                                       'out_octets': '3753327020',
                                                                       'out_ucast_pkts': '7535005',
                                                                       'poller_hash': 'MTE3REFFM3BlZXJpbnRmcG9uIDI1NDU3MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254571',
                                                                                               'links': [       {       'connected_time': '262229',
                                                                                                                        'fiber_length': 20133,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFFM3BlZXJwb25wb24gMjU0NTcx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2493 LINE 45 (HYDE AMBER)',
                                         'poller_hash': 'MTE3REFFMw==',
                                         'serial_number': '117DAE3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:46:bd',
                                         'uptime': '262229'},
                                 {       'counter_source': '254581/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:30:44:3e:1a:55'],
                                                                       'id': '254581/g1',
                                                                       'in_nucast_pkts': '580',
                                                                       'in_octets': '486019355646',
                                                                       'in_ucast_pkts': '890973520',
                                                                       'mac_address': '1c:8b:76:00:2f:dd',
                                                                       'mtu': 2000,
                                                                       'name': '254581/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45699978',
                                                                       'out_octets': '1254044956152',
                                                                       'out_ucast_pkts': '1296977127',
                                                                       'poller_hash': 'MTU4RkE3MXBlZXJpbnRmMjU0NTgxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254581',
                                                                       'in_nucast_pkts': '45699978',
                                                                       'in_octets': '1254044956152',
                                                                       'in_ucast_pkts': '1296977127',
                                                                       'links': [       {       'in_nucast_pkts': '45699978',
                                                                                                'in_octets': '1254044956152',
                                                                                                'in_ucast_pkts': '1296977127',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '580',
                                                                                                'out_octets': '486019355646',
                                                                                                'out_ucast_pkts': '890973520',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '580',
                                                                       'out_octets': '486019355646',
                                                                       'out_ucast_pkts': '890973520',
                                                                       'poller_hash': 'MTU4RkE3MXBlZXJpbnRmcG9uIDI1NDU4MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254581',
                                                                                               'links': [       {       'connected_time': '27752524',
                                                                                                                        'fiber_length': 2010,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.692,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU4RkE3MXBlZXJwb25wb24gMjU0NTgx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '6433 PERTH ROAD 131 (GROWMARK,INC. '
                                                 'D/B/A CENTRAL)',
                                         'poller_hash': 'MTU4RkE3MQ==',
                                         'serial_number': '158FA71',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '1c:8b:76:00:2f:d9',
                                         'uptime': '27752524'},
                                 {       'counter_source': '254593/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6d:14'],
                                                                       'id': '254593/g1',
                                                                       'in_nucast_pkts': '6020',
                                                                       'in_octets': '2110822379',
                                                                       'in_ucast_pkts': '8734020',
                                                                       'mac_address': 'b8:94:70:a1:46:88',
                                                                       'mtu': 2000,
                                                                       'name': '254593/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '690126',
                                                                       'out_octets': '63154248071',
                                                                       'out_ucast_pkts': '43433248',
                                                                       'poller_hash': 'MTE3REFEQ3BlZXJpbnRmMjU0NTkzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254593',
                                                                       'in_nucast_pkts': '690126',
                                                                       'in_octets': '63154248071',
                                                                       'in_ucast_pkts': '43433248',
                                                                       'links': [       {       'in_nucast_pkts': '690126',
                                                                                                'in_octets': '63154248071',
                                                                                                'in_ucast_pkts': '43433248',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6020',
                                                                                                'out_octets': '2110822379',
                                                                                                'out_ucast_pkts': '8734020',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6020',
                                                                       'out_octets': '2110822379',
                                                                       'out_ucast_pkts': '8734020',
                                                                       'poller_hash': 'MTE3REFEQ3BlZXJpbnRmcG9uIDI1NDU5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254593',
                                                                                               'links': [       {       'connected_time': '334508',
                                                                                                                        'fiber_length': 21416,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFEQ3BlZXJwb25wb24gMjU0NTkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4247 LINE 46 (THOMAS CHRISTIAN)',
                                         'poller_hash': 'MTE3REFEQw==',
                                         'serial_number': '117DADC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:46:85',
                                         'uptime': '334508'},
                                 {       'counter_source': '254725/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254725/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '74074351421',
                                                                       'in_ucast_pkts': '399539148',
                                                                       'mac_address': 'b8:94:70:a1:49:25',
                                                                       'name': '254725/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1694735356644',
                                                                       'out_ucast_pkts': '1259516995',
                                                                       'poller_hash': 'MTE3REIzMHBlZXJpbnRmMjU0NzI1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254725',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1694735356644',
                                                                       'in_ucast_pkts': '1259516995',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1694735356644',
                                                                                                'in_ucast_pkts': '1259516995',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '74074351421',
                                                                                                'out_ucast_pkts': '399539148',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '74074351421',
                                                                       'out_ucast_pkts': '399539148',
                                                                       'poller_hash': 'MTE3REIzMHBlZXJpbnRmcG9uIDI1NDcyNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254725',
                                                                                               'links': [       {       'connected_time': '26078305',
                                                                                                                        'fiber_length': 6635,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REIzMHBlZXJwb25wb24gMjU0NzI1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4463 PERTH LINE 72 (WAGLER LANELLE)',
                                         'poller_hash': 'MTE3REIzMA==',
                                         'serial_number': '117DB30',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:49:25',
                                         'uptime': '26078305'},
                                 {       'counter_source': '254735/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254735/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '56351041477',
                                                                       'in_ucast_pkts': '243847697',
                                                                       'mac_address': 'b8:94:70:a1:4e:dd',
                                                                       'name': '254735/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '1037449028476',
                                                                       'out_ucast_pkts': '815697476',
                                                                       'poller_hash': 'MTE3REJFN3BlZXJpbnRmMjU0NzM1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254735',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '1037449028476',
                                                                       'in_ucast_pkts': '815697476',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '1037449028476',
                                                                                                'in_ucast_pkts': '815697476',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '56351041477',
                                                                                                'out_ucast_pkts': '243847697',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '56351041477',
                                                                       'out_ucast_pkts': '243847697',
                                                                       'poller_hash': 'MTE3REJFN3BlZXJpbnRmcG9uIDI1NDczNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254735',
                                                                                               'links': [       {       'connected_time': '538376',
                                                                                                                        'fiber_length': 24561,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJFN3BlZXJwb25wb24gMjU0NzM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4540 LINE 46 (RITSMA JOAN)',
                                         'poller_hash': 'MTE3REJFNw==',
                                         'serial_number': '117DBE7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4e:dd',
                                         'uptime': '538376'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '254745/g1',
                                                                       'name': '254745/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE3REMxMXBlZXJpbnRmMjU0NzQ1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254745',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE3REMxMXBlZXJpbnRmcG9uIDI1NDc0NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254745',
                                                                                               'links': [       {       'connected_time': '27579615',
                                                                                                                        'fiber_length': 1981,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMxMXBlZXJwb25wb24gMjU0NzQ1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4619 LINE 61 (WAGLER JONATHAN)',
                                         'poller_hash': 'MTE3REMxMQ==',
                                         'serial_number': '117DC11',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:2d',
                                         'uptime': '27579615'},
                                 {       'counter_source': '254775/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254775/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '140740344067',
                                                                       'in_ucast_pkts': '61016728454',
                                                                       'mac_address': '04:bc:9f:6f:0a:96',
                                                                       'name': '254775/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '2615768850051',
                                                                       'out_ucast_pkts': '2068800600',
                                                                       'poller_hash': 'MTI5NzRENnBlZXJpbnRmMjU0Nzc1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254775',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '2615768850051',
                                                                       'in_ucast_pkts': '2068800600',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '2615768850051',
                                                                                                'in_ucast_pkts': '2068800600',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '140740344067',
                                                                                                'out_ucast_pkts': '61016728454',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '140740344067',
                                                                       'out_ucast_pkts': '61016728454',
                                                                       'poller_hash': 'MTI5NzRENnBlZXJpbnRmcG9uIDI1NDc3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254775',
                                                                                               'links': [       {       'connected_time': '688595',
                                                                                                                        'fiber_length': 6647,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRENnBlZXJwb25wb24gMjU0Nzc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7007 ROAD 129 (KLASSEN ANNA)',
                                         'poller_hash': 'MTI5NzRENg==',
                                         'serial_number': '12974D6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:0a:96',
                                         'uptime': '688595'},
                                 {       'counter_source': '254914/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254914/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '98464699699',
                                                                       'in_ucast_pkts': '148192150',
                                                                       'mac_address': 'b8:94:70:a1:4e:0d',
                                                                       'name': '254914/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '389345979327',
                                                                       'out_ucast_pkts': '371860612',
                                                                       'poller_hash': 'MTE3REJDRHBlZXJpbnRmMjU0OTE0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254914',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '389345979327',
                                                                       'in_ucast_pkts': '371860612',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '389345979327',
                                                                                                'in_ucast_pkts': '371860612',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '98464699699',
                                                                                                'out_ucast_pkts': '148192150',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '98464699699',
                                                                       'out_ucast_pkts': '148192150',
                                                                       'poller_hash': 'MTE3REJDRHBlZXJpbnRmcG9uIDI1NDkxNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254914',
                                                                                               'links': [       {       'connected_time': '688564',
                                                                                                                        'fiber_length': 5944,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJDRHBlZXJwb25wb24gMjU0OTE0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6033 PERTH ROAD 119 (GASCHO KEVIN & '
                                                 'NADINE)',
                                         'poller_hash': 'MTE3REJDRA==',
                                         'serial_number': '117DBCD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4e:0d',
                                         'uptime': '688564'},
                                 {       'counter_source': '254949/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:73:d3'],
                                                                       'id': '254949/g1',
                                                                       'in_nucast_pkts': '14833',
                                                                       'in_octets': '2434273187',
                                                                       'in_ucast_pkts': '11121837',
                                                                       'mac_address': '04:bc:9f:6f:03:11',
                                                                       'mtu': 2000,
                                                                       'name': '254949/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1435203',
                                                                       'out_octets': '160586218110',
                                                                       'out_ucast_pkts': '111309537',
                                                                       'poller_hash': 'MTI5NzNFNXBlZXJpbnRmMjU0OTQ5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254949',
                                                                       'in_nucast_pkts': '1435203',
                                                                       'in_octets': '160586218110',
                                                                       'in_ucast_pkts': '111309537',
                                                                       'links': [       {       'in_nucast_pkts': '1435203',
                                                                                                'in_octets': '160586218110',
                                                                                                'in_ucast_pkts': '111309537',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14833',
                                                                                                'out_octets': '2434273187',
                                                                                                'out_ucast_pkts': '11121837',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14833',
                                                                       'out_octets': '2434273187',
                                                                       'out_ucast_pkts': '11121837',
                                                                       'poller_hash': 'MTI5NzNFNXBlZXJpbnRmcG9uIDI1NDk0OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254949',
                                                                                               'links': [       {       'connected_time': '688553',
                                                                                                                        'fiber_length': 1802,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNFNXBlZXJwb25wb24gMjU0OTQ5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 MURRAY CT (WEBER JOSH)',
                                         'poller_hash': 'MTI5NzNFNQ==',
                                         'serial_number': '12973E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:03:0e',
                                         'uptime': '688553'},
                                 {       'counter_source': '255426/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a1:e6:7e'],
                                                                       'id': '255426/g1',
                                                                       'in_nucast_pkts': '15168',
                                                                       'in_octets': '14317919365',
                                                                       'in_ucast_pkts': '26682363',
                                                                       'mac_address': 'ec:4f:82:7d:b4:6d',
                                                                       'mtu': 2000,
                                                                       'name': '255426/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465327',
                                                                       'out_octets': '133008787547',
                                                                       'out_ucast_pkts': '109405672',
                                                                       'poller_hash': 'MzAwMzlGcGVlcmludGYyNTU0MjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '255426',
                                                                       'in_nucast_pkts': '1465327',
                                                                       'in_octets': '133008787547',
                                                                       'in_ucast_pkts': '109405672',
                                                                       'links': [       {       'in_nucast_pkts': '1465327',
                                                                                                'in_octets': '133008787547',
                                                                                                'in_ucast_pkts': '109405672',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15168',
                                                                                                'out_octets': '14317919365',
                                                                                                'out_ucast_pkts': '26682363',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15168',
                                                                       'out_octets': '14317919365',
                                                                       'out_ucast_pkts': '26682363',
                                                                       'poller_hash': 'MzAwMzlGcGVlcmludGZwb24gMjU1NDI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255426',
                                                                                               'links': [       {       'connected_time': '688468',
                                                                                                                        'fiber_length': 1686,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwMzlGcGVlcnBvbnBvbiAyNTU0MjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '55 WILLIAM ST (BRUCK SAMANTHA)',
                                         'poller_hash': 'MzAwMzlG',
                                         'serial_number': '30039F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:b4:6d',
                                         'uptime': '688468'},
                                 {       'counter_source': '255702/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a1:e5:c3'],
                                                                       'id': '255702/g1',
                                                                       'in_nucast_pkts': '502469',
                                                                       'in_octets': '181157846893',
                                                                       'in_ucast_pkts': '1056279218',
                                                                       'mac_address': '04:bc:9f:6f:00:31',
                                                                       'mtu': 2000,
                                                                       'name': '255702/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '64693803',
                                                                       'out_octets': '4018052321968',
                                                                       'out_ucast_pkts': '2952573173',
                                                                       'poller_hash': 'MTI5NzM4OXBlZXJpbnRmMjU1NzAyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '255702',
                                                                       'in_nucast_pkts': '64693803',
                                                                       'in_octets': '4018052321968',
                                                                       'in_ucast_pkts': '2952573173',
                                                                       'links': [       {       'in_nucast_pkts': '64693803',
                                                                                                'in_octets': '4018052321968',
                                                                                                'in_ucast_pkts': '2952573173',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '502469',
                                                                                                'out_octets': '181157846893',
                                                                                                'out_ucast_pkts': '1056279218',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '502469',
                                                                       'out_octets': '181157846893',
                                                                       'out_ucast_pkts': '1056279218',
                                                                       'poller_hash': 'MTI5NzM4OXBlZXJpbnRmcG9uIDI1NTcwMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255702',
                                                                                               'links': [       {       'connected_time': '26960047',
                                                                                                                        'fiber_length': 3100,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM4OXBlZXJwb25wb24gMjU1NzAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '72 AVERY PL (MCINTOSH MARGO)',
                                         'poller_hash': 'MTI5NzM4OQ==',
                                         'serial_number': '1297389',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:00:2e',
                                         'uptime': '26960047'},
                                 {       'counter_source': '255819/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '255819/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '391029111833',
                                                                       'in_ucast_pkts': '47946549428',
                                                                       'mac_address': '04:bc:9f:6f:08:3e',
                                                                       'name': '255819/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1603925272843',
                                                                       'out_ucast_pkts': '1327827281',
                                                                       'poller_hash': 'MTI5NzQ4QnBlZXJpbnRmMjU1ODE5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '255819',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1603925272843',
                                                                       'in_ucast_pkts': '1327827281',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1603925272843',
                                                                                                'in_ucast_pkts': '1327827281',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '391029111833',
                                                                                                'out_ucast_pkts': '47946549428',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '391029111833',
                                                                       'out_ucast_pkts': '47946549428',
                                                                       'poller_hash': 'MTI5NzQ4QnBlZXJpbnRmcG9uIDI1NTgxOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255819',
                                                                                               'links': [       {       'connected_time': '688593',
                                                                                                                        'fiber_length': 6310,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ4QnBlZXJwb25wb24gMjU1ODE5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6034 PERTH ROAD 119 (SLITS PEDRO & '
                                                 'JOLANDA)',
                                         'poller_hash': 'MTI5NzQ4Qg==',
                                         'serial_number': '129748B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:08:3e',
                                         'uptime': '688593'},
                                 {       'counter_source': '255829/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d5:40'],
                                                                       'id': '255829/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '3368802544',
                                                                       'in_ucast_pkts': '14583234',
                                                                       'mac_address': 'cc:be:59:cb:98:7d',
                                                                       'mtu': 2000,
                                                                       'name': '255829/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465082',
                                                                       'out_octets': '107574952194',
                                                                       'out_ucast_pkts': '78602497',
                                                                       'poller_hash': 'NDJDNDY3cGVlcmludGYyNTU4MjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '255829',
                                                                       'in_nucast_pkts': '1465082',
                                                                       'in_octets': '107574952194',
                                                                       'in_ucast_pkts': '78602497',
                                                                       'links': [       {       'in_nucast_pkts': '1465082',
                                                                                                'in_octets': '107574952194',
                                                                                                'in_ucast_pkts': '78602497',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '3368802544',
                                                                                                'out_ucast_pkts': '14583234',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '3368802544',
                                                                       'out_ucast_pkts': '14583234',
                                                                       'poller_hash': 'NDJDNDY3cGVlcmludGZwb24gMjU1ODI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255829',
                                                                                               'links': [       {       'connected_time': '688454',
                                                                                                                        'fiber_length': 882,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNDY3cGVlcnBvbnBvbiAyNTU4Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '51 MILL ST E (HIPPE AMANDA)',
                                         'poller_hash': 'NDJDNDY3',
                                         'serial_number': '42C467',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:98:7a',
                                         'uptime': '688454'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '255849/g1',
                                                                       'name': '255849/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzM3OXBlZXJpbnRmMjU1ODQ5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '255849',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzM3OXBlZXJpbnRmcG9uIDI1NTg0OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255849',
                                                                                               'links': [       {       'connected_time': '10645451',
                                                                                                                        'fiber_length': 18749,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM3OXBlZXJwb25wb24gMjU1ODQ5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2568 LINE 47 (SCHNEIDER LARRY)',
                                         'poller_hash': 'MTI5NzM3OQ==',
                                         'serial_number': '1297379',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:ff:ae',
                                         'uptime': '10645451'},
                                 {       'counter_source': '255859/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:73:23'],
                                                                       'id': '255859/g1',
                                                                       'in_nucast_pkts': '241268',
                                                                       'in_octets': '61264932901',
                                                                       'in_ucast_pkts': '310690570',
                                                                       'mac_address': '04:bc:9f:6f:03:d1',
                                                                       'mtu': 2000,
                                                                       'name': '255859/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18291499',
                                                                       'out_octets': '832125635658',
                                                                       'out_ucast_pkts': '662179248',
                                                                       'poller_hash': 'MTI5NzNGRHBlZXJpbnRmMjU1ODU5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '255859',
                                                                       'in_nucast_pkts': '18291499',
                                                                       'in_octets': '832125635658',
                                                                       'in_ucast_pkts': '662179248',
                                                                       'links': [       {       'in_nucast_pkts': '18291499',
                                                                                                'in_octets': '832125635658',
                                                                                                'in_ucast_pkts': '662179248',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '241268',
                                                                                                'out_octets': '61264932901',
                                                                                                'out_ucast_pkts': '310690570',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '241268',
                                                                       'out_octets': '61264932901',
                                                                       'out_ucast_pkts': '310690570',
                                                                       'poller_hash': 'MTI5NzNGRHBlZXJpbnRmcG9uIDI1NTg1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255859',
                                                                                               'links': [       {       'connected_time': '26550122',
                                                                                                                        'fiber_length': 6069,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNGRHBlZXJwb25wb24gMjU1ODU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6025 PERTH ROAD 119 (SCHEERER KEN & '
                                                 'CONNIE)',
                                         'poller_hash': 'MTI5NzNGRA==',
                                         'serial_number': '12973FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:03:ce',
                                         'uptime': '26550122'},
                                 {       'counter_source': '255888/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:c7:3a'],
                                                                       'id': '255888/g1',
                                                                       'in_nucast_pkts': '12577',
                                                                       'in_octets': '15721190108',
                                                                       'in_ucast_pkts': '18517931',
                                                                       'mac_address': '04:bc:9f:6f:02:a1',
                                                                       'mtu': 2000,
                                                                       'name': '255888/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465330',
                                                                       'out_octets': '43993550596',
                                                                       'out_ucast_pkts': '37731321',
                                                                       'poller_hash': 'MTI5NzNEN3BlZXJpbnRmMjU1ODg4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '255888',
                                                                       'in_nucast_pkts': '1465330',
                                                                       'in_octets': '43993550596',
                                                                       'in_ucast_pkts': '37731321',
                                                                       'links': [       {       'in_nucast_pkts': '1465330',
                                                                                                'in_octets': '43993550596',
                                                                                                'in_ucast_pkts': '37731321',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12577',
                                                                                                'out_octets': '15721190108',
                                                                                                'out_ucast_pkts': '18517931',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12577',
                                                                       'out_octets': '15721190108',
                                                                       'out_ucast_pkts': '18517931',
                                                                       'poller_hash': 'MTI5NzNEN3BlZXJpbnRmcG9uIDI1NTg4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255888',
                                                                                               'links': [       {       'connected_time': '688357',
                                                                                                                        'fiber_length': 1141,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNEN3BlZXJwb25wb24gMjU1ODg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 MAIN ST S (KETA LISA)',
                                         'poller_hash': 'MTI5NzNENw==',
                                         'serial_number': '12973D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:02:9e',
                                         'uptime': '688357'},
                                 {       'counter_source': '256192/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:73:5f'],
                                                                       'id': '256192/g1',
                                                                       'in_nucast_pkts': '454670',
                                                                       'in_octets': '51425991067',
                                                                       'in_ucast_pkts': '526239563',
                                                                       'mac_address': '04:bc:9f:6f:08:49',
                                                                       'mtu': 2000,
                                                                       'name': '256192/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '56321021',
                                                                       'out_octets': '2760545767479',
                                                                       'out_ucast_pkts': '1898817472',
                                                                       'poller_hash': 'MTI5NzQ4Q3BlZXJpbnRmMjU2MTkyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256192',
                                                                       'in_nucast_pkts': '56321021',
                                                                       'in_octets': '2760545767479',
                                                                       'in_ucast_pkts': '1898817472',
                                                                       'links': [       {       'in_nucast_pkts': '56321021',
                                                                                                'in_octets': '2760545767479',
                                                                                                'in_ucast_pkts': '1898817472',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '454670',
                                                                                                'out_octets': '51425991067',
                                                                                                'out_ucast_pkts': '526239563',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '454670',
                                                                       'out_octets': '51425991067',
                                                                       'out_ucast_pkts': '526239563',
                                                                       'poller_hash': 'MTI5NzQ4Q3BlZXJpbnRmcG9uIDI1NjE5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256192',
                                                                                               'links': [       {       'connected_time': '26102069',
                                                                                                                        'fiber_length': 1634,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ4Q3BlZXJwb25wb24gMjU2MTky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '33 MURRAY CT (BRAUN HELENA)',
                                         'poller_hash': 'MTI5NzQ4Qw==',
                                         'serial_number': '129748C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:08:46',
                                         'uptime': '26102069'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '256212/g1',
                                                                       'name': '256212/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzNGQnBlZXJpbnRmMjU2MjEyL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256212',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzNGQnBlZXJpbnRmcG9uIDI1NjIxMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256212',
                                                                                               'links': [       {       'connected_time': '26096771',
                                                                                                                        'fiber_length': 3582,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNGQnBlZXJwb25wb24gMjU2MjEy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "6684 ROAD 129 (JANTZI'S SHOE BOX)",
                                         'poller_hash': 'MTI5NzNGQg==',
                                         'serial_number': '12973FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:03:be',
                                         'uptime': '26096771'},
                                 {       'counter_source': '256272/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256272/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '70216702992',
                                                                       'in_ucast_pkts': '427196979',
                                                                       'mac_address': '04:bc:9f:6f:09:76',
                                                                       'name': '256272/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '2272157276591',
                                                                       'out_ucast_pkts': '1693881743',
                                                                       'poller_hash': 'MTI5NzRCMnBlZXJpbnRmMjU2MjcyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256272',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '2272157276591',
                                                                       'in_ucast_pkts': '1693881743',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '2272157276591',
                                                                                                'in_ucast_pkts': '1693881743',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '70216702992',
                                                                                                'out_ucast_pkts': '427196979',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '70216702992',
                                                                       'out_ucast_pkts': '427196979',
                                                                       'poller_hash': 'MTI5NzRCMnBlZXJpbnRmcG9uIDI1NjI3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256272',
                                                                                               'links': [       {       'connected_time': '25944004',
                                                                                                                        'fiber_length': 6626,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRCMnBlZXJwb25wb24gMjU2Mjcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7009 ROAD 129 (MARTIN HOWARD)',
                                         'poller_hash': 'MTI5NzRCMg==',
                                         'serial_number': '12974B2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:09:76',
                                         'uptime': '25944004'},
                                 {       'counter_source': '256292/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256292/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '85807266371',
                                                                       'in_ucast_pkts': '372378688',
                                                                       'mac_address': '04:bc:9f:6f:03:7e',
                                                                       'name': '256292/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '2274862873802',
                                                                       'out_ucast_pkts': '1687085022',
                                                                       'poller_hash': 'MTI5NzNGM3BlZXJpbnRmMjU2MjkyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256292',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '2274862873802',
                                                                       'in_ucast_pkts': '1687085022',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '2274862873802',
                                                                                                'in_ucast_pkts': '1687085022',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '85807266371',
                                                                                                'out_ucast_pkts': '372378688',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '85807266371',
                                                                       'out_ucast_pkts': '372378688',
                                                                       'poller_hash': 'MTI5NzNGM3BlZXJpbnRmcG9uIDI1NjI5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256292',
                                                                                               'links': [       {       'connected_time': '538368',
                                                                                                                        'fiber_length': 21078,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNGM3BlZXJwb25wb24gMjU2Mjky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4219 LINE 46 (BRUULSEMA HAROLD)',
                                         'poller_hash': 'MTI5NzNGMw==',
                                         'serial_number': '12973F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:03:7e',
                                         'uptime': '538368'},
                                 {       'counter_source': '256333/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a4:35:f3'],
                                                                       'id': '256333/g1',
                                                                       'in_nucast_pkts': '14124',
                                                                       'in_octets': '15478487562',
                                                                       'in_ucast_pkts': '63443404',
                                                                       'mac_address': '04:bc:9f:6f:07:f9',
                                                                       'mtu': 2000,
                                                                       'name': '256333/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465328',
                                                                       'out_octets': '208552746194',
                                                                       'out_ucast_pkts': '152798573',
                                                                       'poller_hash': 'MTI5NzQ4MnBlZXJpbnRmMjU2MzMzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256333',
                                                                       'in_nucast_pkts': '1465328',
                                                                       'in_octets': '208552746194',
                                                                       'in_ucast_pkts': '152798573',
                                                                       'links': [       {       'in_nucast_pkts': '1465328',
                                                                                                'in_octets': '208552746194',
                                                                                                'in_ucast_pkts': '152798573',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14124',
                                                                                                'out_octets': '15478487562',
                                                                                                'out_ucast_pkts': '63443404',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14124',
                                                                       'out_octets': '15478487562',
                                                                       'out_ucast_pkts': '63443404',
                                                                       'poller_hash': 'MTI5NzQ4MnBlZXJpbnRmcG9uIDI1NjMzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256333',
                                                                                               'links': [       {       'connected_time': '688485',
                                                                                                                        'fiber_length': 550,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ4MnBlZXJwb25wb24gMjU2MzMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 MAIN ST S (GARDNER ROBYN)',
                                         'poller_hash': 'MTI5NzQ4Mg==',
                                         'serial_number': '1297482',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:07:f6',
                                         'uptime': '688485'},
                                 {       'counter_source': '256363/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:b3:70:20:13:d1'],
                                                                       'id': '256363/g1',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '6580856036',
                                                                       'in_ucast_pkts': '14250980',
                                                                       'mac_address': '04:bc:9f:6e:fe:49',
                                                                       'mtu': 2000,
                                                                       'name': '256363/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2196956',
                                                                       'out_octets': '24046090708',
                                                                       'out_ucast_pkts': '24685051',
                                                                       'poller_hash': 'MTI5NzM0Q3BlZXJpbnRmMjU2MzYzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256363',
                                                                       'in_nucast_pkts': '2196956',
                                                                       'in_octets': '24046090708',
                                                                       'in_ucast_pkts': '24685051',
                                                                       'links': [       {       'in_nucast_pkts': '2196956',
                                                                                                'in_octets': '24046090708',
                                                                                                'in_ucast_pkts': '24685051',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '117',
                                                                                                'out_octets': '6580856036',
                                                                                                'out_ucast_pkts': '14250980',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '6580856036',
                                                                       'out_ucast_pkts': '14250980',
                                                                       'poller_hash': 'MTI5NzM0Q3BlZXJpbnRmcG9uIDI1NjM2Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256363',
                                                                                               'links': [       {       'connected_time': '1026015',
                                                                                                                        'fiber_length': 17656,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM0Q3BlZXJwb25wb24gMjU2MzYz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '218 WINSTANLEY ST (PEARL '
                                                 'INSPIRATIONS)',
                                         'poller_hash': 'MTI5NzM0Qw==',
                                         'serial_number': '129734C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:fe:46',
                                         'uptime': '1026015'},
                                 {       'counter_source': '256442/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:74:88'],
                                                                       'id': '256442/g1',
                                                                       'in_nucast_pkts': '9601',
                                                                       'in_octets': '3179935927',
                                                                       'in_ucast_pkts': '16421375',
                                                                       'mac_address': '04:bc:9f:6f:09:49',
                                                                       'mtu': 2000,
                                                                       'name': '256442/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151535',
                                                                       'out_octets': '112012841839',
                                                                       'out_ucast_pkts': '79426269',
                                                                       'poller_hash': 'MTI5NzRBQ3BlZXJpbnRmMjU2NDQyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256442',
                                                                       'in_nucast_pkts': '1151535',
                                                                       'in_octets': '112012841839',
                                                                       'in_ucast_pkts': '79426269',
                                                                       'links': [       {       'in_nucast_pkts': '1151535',
                                                                                                'in_octets': '112012841839',
                                                                                                'in_ucast_pkts': '79426269',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9601',
                                                                                                'out_octets': '3179935927',
                                                                                                'out_ucast_pkts': '16421375',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9601',
                                                                       'out_octets': '3179935927',
                                                                       'out_ucast_pkts': '16421375',
                                                                       'poller_hash': 'MTI5NzRBQ3BlZXJpbnRmcG9uIDI1NjQ0Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256442',
                                                                                               'links': [       {       'connected_time': '538350',
                                                                                                                        'fiber_length': 22718,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRBQ3BlZXJwb25wb24gMjU2NDQy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2733 LINE 45 (ROSE STEVEN & KERRY)',
                                         'poller_hash': 'MTI5NzRBQw==',
                                         'serial_number': '12974AC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:09:46',
                                         'uptime': '538350'},
                                 {       'counter_source': '256462/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:75:9b'],
                                                                       'id': '256462/g1',
                                                                       'in_nucast_pkts': '234760',
                                                                       'in_octets': '291979630890',
                                                                       'in_ucast_pkts': '634793017',
                                                                       'mac_address': '04:bc:9f:6f:06:b9',
                                                                       'mtu': 2000,
                                                                       'name': '256462/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27881900',
                                                                       'out_octets': '3249671507104',
                                                                       'out_ucast_pkts': '2501565286',
                                                                       'poller_hash': 'MTI5NzQ1QXBlZXJpbnRmMjU2NDYyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256462',
                                                                       'in_nucast_pkts': '27881900',
                                                                       'in_octets': '3249671507104',
                                                                       'in_ucast_pkts': '2501565286',
                                                                       'links': [       {       'in_nucast_pkts': '27881900',
                                                                                                'in_octets': '3249671507104',
                                                                                                'in_ucast_pkts': '2501565286',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '234760',
                                                                                                'out_octets': '291979630890',
                                                                                                'out_ucast_pkts': '634793017',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '234760',
                                                                       'out_octets': '291979630890',
                                                                       'out_ucast_pkts': '634793017',
                                                                       'poller_hash': 'MTI5NzQ1QXBlZXJpbnRmcG9uIDI1NjQ2Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256462',
                                                                                               'links': [       {       'connected_time': '12377084',
                                                                                                                        'fiber_length': 10048,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ1QXBlZXJwb25wb24gMjU2NDYy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5205 LINE 58 (YOUNG JACOB)',
                                         'poller_hash': 'MTI5NzQ1QQ==',
                                         'serial_number': '129745A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:06:b6',
                                         'uptime': '12377084'},
                                 {       'counter_source': '256536/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:70:c0'],
                                                                       'id': '256536/g1',
                                                                       'in_nucast_pkts': '460244',
                                                                       'in_octets': '173074310049',
                                                                       'in_ucast_pkts': '280432430',
                                                                       'mac_address': '04:bc:9f:6f:02:d9',
                                                                       'mtu': 2000,
                                                                       'name': '256536/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '59910530',
                                                                       'out_octets': '602662160050',
                                                                       'out_ucast_pkts': '538760380',
                                                                       'poller_hash': 'MTI5NzNERXBlZXJpbnRmMjU2NTM2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256536',
                                                                       'in_nucast_pkts': '59910530',
                                                                       'in_octets': '602662160050',
                                                                       'in_ucast_pkts': '538760380',
                                                                       'links': [       {       'in_nucast_pkts': '59910530',
                                                                                                'in_octets': '602662160050',
                                                                                                'in_ucast_pkts': '538760380',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '460244',
                                                                                                'out_octets': '173074310049',
                                                                                                'out_ucast_pkts': '280432430',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '460244',
                                                                       'out_octets': '173074310049',
                                                                       'out_ucast_pkts': '280432430',
                                                                       'poller_hash': 'MTI5NzNERXBlZXJpbnRmcG9uIDI1NjUzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256536',
                                                                                               'links': [       {       'connected_time': '25148138',
                                                                                                                        'fiber_length': 15671,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNERXBlZXJwb25wb24gMjU2NTM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2339 LINE 51 (WAGLER CHRISTINE)',
                                         'poller_hash': 'MTI5NzNERQ==',
                                         'serial_number': '12973DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:02:d6',
                                         'uptime': '25148138'},
                                 {       'counter_source': '256546/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256546/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '411076133060',
                                                                       'in_ucast_pkts': '22618985456',
                                                                       'mac_address': '04:bc:9f:6f:04:ce',
                                                                       'name': '256546/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '93',
                                                                       'out_octets': '920242950788',
                                                                       'out_ucast_pkts': '1501113988',
                                                                       'poller_hash': 'MTI5NzQxRHBlZXJpbnRmMjU2NTQ2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256546',
                                                                       'in_nucast_pkts': '93',
                                                                       'in_octets': '920242950788',
                                                                       'in_ucast_pkts': '1501113988',
                                                                       'links': [       {       'in_nucast_pkts': '93',
                                                                                                'in_octets': '920242950788',
                                                                                                'in_ucast_pkts': '1501113988',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '411076133060',
                                                                                                'out_ucast_pkts': '22618985456',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '411076133060',
                                                                       'out_ucast_pkts': '22618985456',
                                                                       'poller_hash': 'MTI5NzQxRHBlZXJpbnRmcG9uIDI1NjU0Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256546',
                                                                                               'links': [       {       'connected_time': '688526',
                                                                                                                        'fiber_length': 1511,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQxRHBlZXJwb25wb24gMjU2NTQ2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4721 LINE 64 (ARRO CRUSHING LTD)',
                                         'poller_hash': 'MTI5NzQxRA==',
                                         'serial_number': '129741D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:04:ce',
                                         'uptime': '688526'},
                                 {       'counter_source': '256556/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:70:89'],
                                                                       'id': '256556/g1',
                                                                       'in_nucast_pkts': '446236',
                                                                       'in_octets': '42559913549',
                                                                       'in_ucast_pkts': '147522985',
                                                                       'mac_address': '04:bc:9f:6f:00:89',
                                                                       'mtu': 2000,
                                                                       'name': '256556/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '59213475',
                                                                       'out_octets': '387939606973',
                                                                       'out_ucast_pkts': '350363838',
                                                                       'poller_hash': 'MTI5NzM5NHBlZXJpbnRmMjU2NTU2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256556',
                                                                       'in_nucast_pkts': '59213475',
                                                                       'in_octets': '387939606973',
                                                                       'in_ucast_pkts': '350363838',
                                                                       'links': [       {       'in_nucast_pkts': '59213475',
                                                                                                'in_octets': '387939606973',
                                                                                                'in_ucast_pkts': '350363838',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '446236',
                                                                                                'out_octets': '42559913549',
                                                                                                'out_ucast_pkts': '147522985',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '446236',
                                                                       'out_octets': '42559913549',
                                                                       'out_ucast_pkts': '147522985',
                                                                       'poller_hash': 'MTI5NzM5NHBlZXJpbnRmcG9uIDI1NjU1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256556',
                                                                                               'links': [       {       'connected_time': '24902973',
                                                                                                                        'fiber_length': 16744,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -24.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM5NHBlZXJwb25wb24gMjU2NTU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2606 LINE 51 (CIPP-FREIBURGER '
                                                 'DEBBIE)',
                                         'poller_hash': 'MTI5NzM5NA==',
                                         'serial_number': '1297394',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:00:86',
                                         'uptime': '24902973'},
                                 {       'counter_source': '256596/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256596/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '21496092414',
                                                                       'in_ucast_pkts': '41520178',
                                                                       'mac_address': '04:bc:9f:6f:01:8e',
                                                                       'name': '256596/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '234110108756',
                                                                       'out_ucast_pkts': '182312687',
                                                                       'poller_hash': 'MTI5NzNCNXBlZXJpbnRmMjU2NTk2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256596',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '234110108756',
                                                                       'in_ucast_pkts': '182312687',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '234110108756',
                                                                                                'in_ucast_pkts': '182312687',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '21496092414',
                                                                                                'out_ucast_pkts': '41520178',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '21496092414',
                                                                       'out_ucast_pkts': '41520178',
                                                                       'poller_hash': 'MTI5NzNCNXBlZXJpbnRmcG9uIDI1NjU5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256596',
                                                                                               'links': [       {       'connected_time': '688365',
                                                                                                                        'fiber_length': 990,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNCNXBlZXJwb25wb24gMjU2NTk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 CRUICKSHANK APT 1 (REIS KEVIN)',
                                         'poller_hash': 'MTI5NzNCNQ==',
                                         'serial_number': '12973B5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:01:8e',
                                         'uptime': '688365'},
                                 {       'counter_source': '256636/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256636/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '74166724903',
                                                                       'in_ucast_pkts': '189380956',
                                                                       'mac_address': '04:bc:9f:6f:09:2e',
                                                                       'name': '256636/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '43',
                                                                       'out_octets': '585044190391',
                                                                       'out_ucast_pkts': '517231570',
                                                                       'poller_hash': 'MTI5NzRBOXBlZXJpbnRmMjU2NjM2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256636',
                                                                       'in_nucast_pkts': '43',
                                                                       'in_octets': '585044190391',
                                                                       'in_ucast_pkts': '517231570',
                                                                       'links': [       {       'in_nucast_pkts': '43',
                                                                                                'in_octets': '585044190391',
                                                                                                'in_ucast_pkts': '517231570',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '74166724903',
                                                                                                'out_ucast_pkts': '189380956',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '74166724903',
                                                                       'out_ucast_pkts': '189380956',
                                                                       'poller_hash': 'MTI5NzRBOXBlZXJpbnRmcG9uIDI1NjYzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256636',
                                                                                               'links': [       {       'connected_time': '24804171',
                                                                                                                        'fiber_length': 6952,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRBOXBlZXJwb25wb24gMjU2NjM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6000 PERTH ROAD 119 (ROPP STEPHEN)',
                                         'poller_hash': 'MTI5NzRBOQ==',
                                         'serial_number': '12974A9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:09:2e',
                                         'uptime': '24804171'},
                                 {       'counter_source': '256656/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256656/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '30570904128',
                                                                       'in_ucast_pkts': '58338922',
                                                                       'mac_address': '04:bc:9f:6f:0b:2e',
                                                                       'name': '256656/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '260838667082',
                                                                       'out_ucast_pkts': '235752037',
                                                                       'poller_hash': 'MTI5NzRFOXBlZXJpbnRmMjU2NjU2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256656',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '260838667082',
                                                                       'in_ucast_pkts': '235752037',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '260838667082',
                                                                                                'in_ucast_pkts': '235752037',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '30570904128',
                                                                                                'out_ucast_pkts': '58338922',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '30570904128',
                                                                       'out_ucast_pkts': '58338922',
                                                                       'poller_hash': 'MTI5NzRFOXBlZXJpbnRmcG9uIDI1NjY1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256656',
                                                                                               'links': [       {       'connected_time': '24734684',
                                                                                                                        'fiber_length': 6717,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRFOXBlZXJwb25wb24gMjU2NjU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4441 PERTH LINE 72 (KLASSEN '
                                                 'KATHARIN)',
                                         'poller_hash': 'MTI5NzRFOQ==',
                                         'serial_number': '12974E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:0b:2e',
                                         'uptime': '24734684'},
                                 {       'counter_source': '256666/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256666/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '49632428656',
                                                                       'in_ucast_pkts': '62216055',
                                                                       'mac_address': '04:bc:9f:6f:06:9e',
                                                                       'name': '256666/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '100000718199',
                                                                       'out_ucast_pkts': '126863025',
                                                                       'poller_hash': 'MTI5NzQ1N3BlZXJpbnRmMjU2NjY2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256666',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '100000718199',
                                                                       'in_ucast_pkts': '126863025',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '100000718199',
                                                                                                'in_ucast_pkts': '126863025',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '49632428656',
                                                                                                'out_ucast_pkts': '62216055',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '49632428656',
                                                                       'out_ucast_pkts': '62216055',
                                                                       'poller_hash': 'MTI5NzQ1N3BlZXJpbnRmcG9uIDI1NjY2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256666',
                                                                                               'links': [       {       'connected_time': '262227',
                                                                                                                        'fiber_length': 21562,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ1N3BlZXJwb25wb24gMjU2NjY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4878 ROAD 109 (REFRESHING WIND '
                                                 'MINISTRIES)',
                                         'poller_hash': 'MTI5NzQ1Nw==',
                                         'serial_number': '1297457',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:06:9e',
                                         'uptime': '262227'},
                                 {       'counter_source': '256696/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c4:ad:34:b1:22:bf'],
                                                                       'id': '256696/g1',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '29356322786',
                                                                       'in_ucast_pkts': '29115592',
                                                                       'mac_address': '04:bc:9f:6e:fe:71',
                                                                       'mtu': 2000,
                                                                       'name': '256696/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465446',
                                                                       'out_octets': '24720525771',
                                                                       'out_ucast_pkts': '29283774',
                                                                       'poller_hash': 'MTI5NzM1MXBlZXJpbnRmMjU2Njk2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256696',
                                                                       'in_nucast_pkts': '1465446',
                                                                       'in_octets': '24720525771',
                                                                       'in_ucast_pkts': '29283774',
                                                                       'links': [       {       'in_nucast_pkts': '1465446',
                                                                                                'in_octets': '24720525771',
                                                                                                'in_ucast_pkts': '29283774',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '26',
                                                                                                'out_octets': '29356322786',
                                                                                                'out_ucast_pkts': '29115592',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '29356322786',
                                                                       'out_ucast_pkts': '29115592',
                                                                       'poller_hash': 'MTI5NzM1MXBlZXJpbnRmcG9uIDI1NjY5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256696',
                                                                                               'links': [       {       'connected_time': '688351',
                                                                                                                        'fiber_length': 2787,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM1MXBlZXJwb25wb24gMjU2Njk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4676 HAWTHORNE CT (BADEN WOODTURNING '
                                                 'INC.)',
                                         'poller_hash': 'MTI5NzM1MQ==',
                                         'serial_number': '1297351',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:fe:6e',
                                         'uptime': '688351'},
                                 {       'counter_source': '256726/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f3:47'],
                                                                       'id': '256726/g1',
                                                                       'in_nucast_pkts': '12023',
                                                                       'in_octets': '1890947956',
                                                                       'in_ucast_pkts': '12918320',
                                                                       'mac_address': '04:bc:9f:6f:09:61',
                                                                       'mtu': 2000,
                                                                       'name': '256726/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465244',
                                                                       'out_octets': '66652945507',
                                                                       'out_ucast_pkts': '46663131',
                                                                       'poller_hash': 'MTI5NzRBRnBlZXJpbnRmMjU2NzI2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256726',
                                                                       'in_nucast_pkts': '1465244',
                                                                       'in_octets': '66652945507',
                                                                       'in_ucast_pkts': '46663131',
                                                                       'links': [       {       'in_nucast_pkts': '1465244',
                                                                                                'in_octets': '66652945507',
                                                                                                'in_ucast_pkts': '46663131',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12023',
                                                                                                'out_octets': '1890947956',
                                                                                                'out_ucast_pkts': '12918320',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12023',
                                                                       'out_octets': '1890947956',
                                                                       'out_ucast_pkts': '12918320',
                                                                       'poller_hash': 'MTI5NzRBRnBlZXJpbnRmcG9uIDI1NjcyNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256726',
                                                                                               'links': [       {       'connected_time': '688369',
                                                                                                                        'fiber_length': 2561,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRBRnBlZXJwb25wb24gMjU2NzI2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '109 PUGH ST (DAVIDSON DIANN)',
                                         'poller_hash': 'MTI5NzRBRg==',
                                         'serial_number': '12974AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:09:5e',
                                         'uptime': '688369'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '256736/g1',
                                                                       'name': '256736/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzQzRHBlZXJpbnRmMjU2NzM2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256736',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzQzRHBlZXJpbnRmcG9uIDI1NjczNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256736',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQzRHBlZXJwb25wb24gMjU2NzM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4447 PERTH LINE 72 (PURCELL DOUG)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "256736" '
                                                                  '(location: GPON '
                                                                  'port "15/2"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-03-13T14:41:14.980Z',
                                         'poller_hash': 'MTI5NzQzRA==',
                                         'serial_number': '129743D',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '256776/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256776/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '63004319958',
                                                                       'in_ucast_pkts': '433948321',
                                                                       'mac_address': 'b8:94:70:ce:6a:84',
                                                                       'name': '256776/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '33',
                                                                       'out_octets': '1674836884611',
                                                                       'out_ucast_pkts': '1216691207',
                                                                       'poller_hash': 'MTFCRTNBMnBlZXJpbnRmMjU2Nzc2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256776',
                                                                       'in_nucast_pkts': '33',
                                                                       'in_octets': '1674836884611',
                                                                       'in_ucast_pkts': '1216691207',
                                                                       'links': [       {       'in_nucast_pkts': '33',
                                                                                                'in_octets': '1674836884611',
                                                                                                'in_ucast_pkts': '1216691207',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '63004319958',
                                                                                                'out_ucast_pkts': '433948321',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '63004319958',
                                                                       'out_ucast_pkts': '433948321',
                                                                       'poller_hash': 'MTFCRTNBMnBlZXJpbnRmcG9uIDI1Njc3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256776',
                                                                                               'links': [       {       'connected_time': '24546110',
                                                                                                                        'fiber_length': 1190,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTNBMnBlZXJwb25wb24gMjU2Nzc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 FULTON ST (TOTH DON)',
                                         'poller_hash': 'MTFCRTNBMg==',
                                         'serial_number': '11BE3A2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:6a:84',
                                         'uptime': '24546110'},
                                 {       'counter_source': '256786/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256786/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '68884583216',
                                                                       'in_ucast_pkts': '367071430',
                                                                       'mac_address': '04:bc:9f:6f:0a:ae',
                                                                       'name': '256786/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2217911604078',
                                                                       'out_ucast_pkts': '1553218206',
                                                                       'poller_hash': 'MTI5NzREOXBlZXJpbnRmMjU2Nzg2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256786',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2217911604078',
                                                                       'in_ucast_pkts': '1553218206',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2217911604078',
                                                                                                'in_ucast_pkts': '1553218206',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '68884583216',
                                                                                                'out_ucast_pkts': '367071430',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '68884583216',
                                                                       'out_ucast_pkts': '367071430',
                                                                       'poller_hash': 'MTI5NzREOXBlZXJpbnRmcG9uIDI1Njc4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256786',
                                                                                               'links': [       {       'connected_time': '688539',
                                                                                                                        'fiber_length': 6645,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzREOXBlZXJwb25wb24gMjU2Nzg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4373 LINE 61 (VIANA KYLA)',
                                         'poller_hash': 'MTI5NzREOQ==',
                                         'serial_number': '12974D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:0a:ae',
                                         'uptime': '688539'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '256836/g1',
                                                                       'name': '256836/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzRFQ3BlZXJpbnRmMjU2ODM2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256836',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzRFQ3BlZXJpbnRmcG9uIDI1NjgzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256836',
                                                                                               'links': [       {       'connected_time': '24110006',
                                                                                                                        'fiber_length': 11854,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRFQ3BlZXJwb25wb24gMjU2ODM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4094 WATERLOO ST (HERRFORT RALPH)',
                                         'poller_hash': 'MTI5NzRFQw==',
                                         'serial_number': '12974EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:0b:46',
                                         'uptime': '24110006'},
                                 {       'counter_source': '256867/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256867/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '391027327356',
                                                                       'in_ucast_pkts': '707871206',
                                                                       'mac_address': 'b8:94:70:ce:57:7c',
                                                                       'name': '256867/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '2542947304042',
                                                                       'out_ucast_pkts': '2033164860',
                                                                       'poller_hash': 'MTFCRTE0MXBlZXJpbnRmMjU2ODY3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256867',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '2542947304042',
                                                                       'in_ucast_pkts': '2033164860',
                                                                       'links': [       {       'in_nucast_pkts': '28',
                                                                                                'in_octets': '2542947304042',
                                                                                                'in_ucast_pkts': '2033164860',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '391027327356',
                                                                                                'out_ucast_pkts': '707871206',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '391027327356',
                                                                       'out_ucast_pkts': '707871206',
                                                                       'poller_hash': 'MTFCRTE0MXBlZXJpbnRmcG9uIDI1Njg2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256867',
                                                                                               'links': [       {       'connected_time': '24026354',
                                                                                                                        'fiber_length': 4999,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTE0MXBlZXJwb25wb24gMjU2ODY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6837 ROAD 129 (JANTZI RAYMOND)',
                                         'poller_hash': 'MTFCRTE0MQ==',
                                         'serial_number': '11BE141',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:57:7c',
                                         'uptime': '24026354'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '256887/g1',
                                                                       'name': '256887/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzREQ3BlZXJpbnRmMjU2ODg3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256887',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzREQ3BlZXJpbnRmcG9uIDI1Njg4Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256887',
                                                                                               'links': [       {       'connected_time': '1671279',
                                                                                                                        'fiber_length': 6301,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzREQ3BlZXJwb25wb24gMjU2ODg3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4555 PERTH LINE 72 (ALBRECHT NORMAN)',
                                         'poller_hash': 'MTI5NzREQw==',
                                         'serial_number': '12974DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:0a:c6',
                                         'uptime': '1671279'},
                                 {       'counter_source': '257419/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b7:85'],
                                                                       'id': '257419/g1',
                                                                       'in_nucast_pkts': '22',
                                                                       'in_octets': '65391626372',
                                                                       'in_ucast_pkts': '71293568',
                                                                       'mac_address': 'ec:4f:82:7e:08:59',
                                                                       'mtu': 2000,
                                                                       'name': '257419/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2607553',
                                                                       'out_octets': '128935903957',
                                                                       'out_ucast_pkts': '107209559',
                                                                       'poller_hash': 'MzVFOTEzcGVlcmludGYyNTc0MTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '257419',
                                                                       'in_nucast_pkts': '2607553',
                                                                       'in_octets': '128935903957',
                                                                       'in_ucast_pkts': '107209559',
                                                                       'links': [       {       'in_nucast_pkts': '2607553',
                                                                                                'in_octets': '128935903957',
                                                                                                'in_ucast_pkts': '107209559',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22',
                                                                                                'out_octets': '65391626372',
                                                                                                'out_ucast_pkts': '71293568',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22',
                                                                       'out_octets': '65391626372',
                                                                       'out_ucast_pkts': '71293568',
                                                                       'poller_hash': 'MzVFOTEzcGVlcmludGZwb24gMjU3NDE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257419',
                                                                                               'links': [       {       'connected_time': '1213487',
                                                                                                                        'fiber_length': 18824,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTEzcGVlcnBvbnBvbiAyNTc0MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5818 LINE 55 (ROPP KRISTI)',
                                         'poller_hash': 'MzVFOTEz',
                                         'serial_number': '35E913',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:08:59',
                                         'uptime': '1213487'},
                                 {       'counter_source': '257658/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:6f:d9'],
                                                                       'id': '257658/g1',
                                                                       'in_nucast_pkts': '435544',
                                                                       'in_octets': '113734770872',
                                                                       'in_ucast_pkts': '729259407',
                                                                       'mac_address': '00:06:31:87:4c:3f',
                                                                       'mtu': 2000,
                                                                       'name': '257658/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43147966',
                                                                       'out_octets': '3755045982688',
                                                                       'out_ucast_pkts': '2638891920',
                                                                       'poller_hash': 'Q0MwQkZwZWVyaW50ZjI1NzY1OC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '257658',
                                                                       'in_nucast_pkts': '43147966',
                                                                       'in_octets': '3755045982688',
                                                                       'in_ucast_pkts': '2638891920',
                                                                       'links': [       {       'in_nucast_pkts': '43147966',
                                                                                                'in_octets': '3755045982688',
                                                                                                'in_ucast_pkts': '2638891920',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '435544',
                                                                                                'out_octets': '113734770872',
                                                                                                'out_ucast_pkts': '729259407',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '435544',
                                                                       'out_octets': '113734770872',
                                                                       'out_ucast_pkts': '729259407',
                                                                       'poller_hash': 'Q0MwQkZwZWVyaW50ZnBvbiAyNTc2NTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257658',
                                                                                               'links': [       {       'connected_time': '25923133',
                                                                                                                        'fiber_length': 1152,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQkZwZWVycG9ucG9uIDI1NzY1OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '7 JO-LENE CT (ALEXANDER ADRIAN)',
                                         'poller_hash': 'Q0MwQkY=',
                                         'serial_number': 'CC0BF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:d1:7f',
                                         'uptime': '25923133'},
                                 {       'counter_source': '257678/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257678/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12508235323',
                                                                       'in_ucast_pkts': '71406581',
                                                                       'mac_address': 'b8:94:70:ce:4d:14',
                                                                       'name': '257678/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '710239396238',
                                                                       'out_ucast_pkts': '507261354',
                                                                       'poller_hash': 'MTFCREZGNHBlZXJpbnRmMjU3Njc4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257678',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '710239396238',
                                                                       'in_ucast_pkts': '507261354',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '710239396238',
                                                                                                'in_ucast_pkts': '507261354',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '12508235323',
                                                                                                'out_ucast_pkts': '71406581',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12508235323',
                                                                       'out_ucast_pkts': '71406581',
                                                                       'poller_hash': 'MTFCREZGNHBlZXJpbnRmcG9uIDI1NzY3OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257678',
                                                                                               'links': [       {       'connected_time': '688358',
                                                                                                                        'fiber_length': 1143,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCREZGNHBlZXJwb25wb24gMjU3Njc4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 MAIN ST S (HASTINGS MANDY)',
                                         'poller_hash': 'MTFCREZGNA==',
                                         'serial_number': '11BDFF4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:4d:14',
                                         'uptime': '688358'},
                                 {       'counter_source': '257698/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:75:59'],
                                                                       'id': '257698/g1',
                                                                       'in_nucast_pkts': '9372',
                                                                       'in_octets': '383538374',
                                                                       'in_ucast_pkts': '1440878',
                                                                       'mac_address': 'b8:94:70:ce:58:47',
                                                                       'mtu': 2000,
                                                                       'name': '257698/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151543',
                                                                       'out_octets': '4494758017',
                                                                       'out_ucast_pkts': '3777441',
                                                                       'poller_hash': 'MTFCRTE1QXBlZXJpbnRmMjU3Njk4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '257698',
                                                                       'in_nucast_pkts': '1151543',
                                                                       'in_octets': '4494758017',
                                                                       'in_ucast_pkts': '3777441',
                                                                       'links': [       {       'in_nucast_pkts': '1151543',
                                                                                                'in_octets': '4494758017',
                                                                                                'in_ucast_pkts': '3777441',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9372',
                                                                                                'out_octets': '383538374',
                                                                                                'out_ucast_pkts': '1440878',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9372',
                                                                       'out_octets': '383538374',
                                                                       'out_ucast_pkts': '1440878',
                                                                       'poller_hash': 'MTFCRTE1QXBlZXJpbnRmcG9uIDI1NzY5OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257698',
                                                                                               'links': [       {       'connected_time': '538594',
                                                                                                                        'fiber_length': 19325,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTE1QXBlZXJwb25wb24gMjU3Njk4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2621 LINE 47 (WILSON JENNIFER)',
                                         'poller_hash': 'MTFCRTE1QQ==',
                                         'serial_number': '11BE15A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:58:44',
                                         'uptime': '538594'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '257708/g1',
                                                                       'name': '257708/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTFCRTFFQXBlZXJpbnRmMjU3NzA4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257708',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTFCRTFFQXBlZXJpbnRmcG9uIDI1NzcwOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257708',
                                                                                               'links': [       {       'connected_time': '20585315',
                                                                                                                        'fiber_length': 6742,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTFFQXBlZXJwb25wb24gMjU3NzA4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4430 PERTH LINE 72 NEWTON (A&A '
                                                 'CONTRACTING)',
                                         'poller_hash': 'MTFCRTFFQQ==',
                                         'serial_number': '11BE1EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:5c:c4',
                                         'uptime': '20585315'},
                                 {       'counter_source': '257718/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257718/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '59483385270',
                                                                       'in_ucast_pkts': '227504100',
                                                                       'mac_address': '04:bc:9f:6f:07:2e',
                                                                       'name': '257718/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '942520890227',
                                                                       'out_ucast_pkts': '820662849',
                                                                       'poller_hash': 'MTI5NzQ2OXBlZXJpbnRmMjU3NzE4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257718',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '942520890227',
                                                                       'in_ucast_pkts': '820662849',
                                                                       'links': [       {       'in_nucast_pkts': '12',
                                                                                                'in_octets': '942520890227',
                                                                                                'in_ucast_pkts': '820662849',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '59483385270',
                                                                                                'out_ucast_pkts': '227504100',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '59483385270',
                                                                       'out_ucast_pkts': '227504100',
                                                                       'poller_hash': 'MTI5NzQ2OXBlZXJpbnRmcG9uIDI1NzcxOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257718',
                                                                                               'links': [       {       'connected_time': '688592',
                                                                                                                        'fiber_length': 6751,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ2OXBlZXJwb25wb24gMjU3NzE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4445 PERTH LINE 72 (A & A '
                                                 'CONTRACTING)',
                                         'poller_hash': 'MTI5NzQ2OQ==',
                                         'serial_number': '1297469',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:07:2e',
                                         'uptime': '688592'},
                                 {       'counter_source': '257728/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257728/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25273304171',
                                                                       'in_ucast_pkts': '84631816',
                                                                       'mac_address': '04:bc:9f:6f:04:ee',
                                                                       'name': '257728/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '207883001786',
                                                                       'out_ucast_pkts': '206278759',
                                                                       'poller_hash': 'MTI5NzQyMXBlZXJpbnRmMjU3NzI4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257728',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '207883001786',
                                                                       'in_ucast_pkts': '206278759',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '207883001786',
                                                                                                'in_ucast_pkts': '206278759',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '25273304171',
                                                                                                'out_ucast_pkts': '84631816',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25273304171',
                                                                       'out_ucast_pkts': '84631816',
                                                                       'poller_hash': 'MTI5NzQyMXBlZXJpbnRmcG9uIDI1NzcyOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257728',
                                                                                               'links': [       {       'connected_time': '23601025',
                                                                                                                        'fiber_length': 6777,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQyMXBlZXJwb25wb24gMjU3NzI4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4443 PERTH LINE 72 (ALBRECHT DIANE)',
                                         'poller_hash': 'MTI5NzQyMQ==',
                                         'serial_number': '1297421',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:04:ee',
                                         'uptime': '23601025'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '257738/g1',
                                                                       'name': '257738/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzM5MHBlZXJpbnRmMjU3NzM4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257738',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzM5MHBlZXJpbnRmcG9uIDI1NzczOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257738',
                                                                                               'links': [       {       'connected_time': '23590596',
                                                                                                                        'fiber_length': 18673,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM5MHBlZXJwb25wb24gMjU3NzM4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4223 LINE 50 (VAN LEEUWEN WILLIAM)',
                                         'poller_hash': 'MTI5NzM5MA==',
                                         'serial_number': '1297390',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:00:66',
                                         'uptime': '23590596'},
                                 {       'counter_source': '257768/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257768/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '253829124417',
                                                                       'in_ucast_pkts': '593917159',
                                                                       'mac_address': '04:bc:9f:6f:04:be',
                                                                       'name': '257768/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '2421897280533',
                                                                       'out_ucast_pkts': '1920057416',
                                                                       'poller_hash': 'MTI5NzQxQnBlZXJpbnRmMjU3NzY4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257768',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '2421897280533',
                                                                       'in_ucast_pkts': '1920057416',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '2421897280533',
                                                                                                'in_ucast_pkts': '1920057416',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '253829124417',
                                                                                                'out_ucast_pkts': '593917159',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '253829124417',
                                                                       'out_ucast_pkts': '593917159',
                                                                       'poller_hash': 'MTI5NzQxQnBlZXJpbnRmcG9uIDI1Nzc2OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257768',
                                                                                               'links': [       {       'connected_time': '538373',
                                                                                                                        'fiber_length': 24373,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQxQnBlZXJwb25wb24gMjU3NzY4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4505 LINE 46 (ROTH GARY & KIM)',
                                         'poller_hash': 'MTI5NzQxQg==',
                                         'serial_number': '129741B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:04:be',
                                         'uptime': '538373'},
                                 {       'counter_source': '257929/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257929/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '156947775074',
                                                                       'in_ucast_pkts': '452910833',
                                                                       'mac_address': '04:bc:9f:6e:f8:1e',
                                                                       'name': '257929/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '2947640596519',
                                                                       'out_ucast_pkts': '2164368684',
                                                                       'poller_hash': 'MTI5NzI4N3BlZXJpbnRmMjU3OTI5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257929',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '2947640596519',
                                                                       'in_ucast_pkts': '2164368684',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '2947640596519',
                                                                                                'in_ucast_pkts': '2164368684',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '156947775074',
                                                                                                'out_ucast_pkts': '452910833',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '156947775074',
                                                                       'out_ucast_pkts': '452910833',
                                                                       'poller_hash': 'MTI5NzI4N3BlZXJpbnRmcG9uIDI1NzkyOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257929',
                                                                                               'links': [       {       'connected_time': '538371',
                                                                                                                        'fiber_length': 24681,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzI4N3BlZXJwb25wb24gMjU3OTI5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4535 LINE 46 (VANSTRAATEN AMY)',
                                         'poller_hash': 'MTI5NzI4Nw==',
                                         'serial_number': '1297287',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:f8:1e',
                                         'uptime': '538371'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '257959/g1',
                                                                       'name': '257959/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzM5RXBlZXJpbnRmMjU3OTU5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257959',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzM5RXBlZXJpbnRmcG9uIDI1Nzk1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257959',
                                                                                               'links': [       {       'connected_time': '23344366',
                                                                                                                        'fiber_length': 6364,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM5RXBlZXJwb25wb24gMjU3OTU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6802 ROAD 124 (KUEPFER NELSON & '
                                                 'DORIS)',
                                         'poller_hash': 'MTI5NzM5RQ==',
                                         'serial_number': '129739E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:00:d6',
                                         'uptime': '23344366'},
                                 {       'counter_source': '258036/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '258036/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '54385457970',
                                                                       'in_ucast_pkts': '381506852',
                                                                       'mac_address': 'b8:94:70:ce:5c:54',
                                                                       'name': '258036/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '1943006668110',
                                                                       'out_ucast_pkts': '1368748657',
                                                                       'poller_hash': 'MTFCRTFEQ3BlZXJpbnRmMjU4MDM2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '258036',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '1943006668110',
                                                                       'in_ucast_pkts': '1368748657',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '1943006668110',
                                                                                                'in_ucast_pkts': '1368748657',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '54385457970',
                                                                                                'out_ucast_pkts': '381506852',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '54385457970',
                                                                       'out_ucast_pkts': '381506852',
                                                                       'poller_hash': 'MTFCRTFEQ3BlZXJpbnRmcG9uIDI1ODAzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '258036',
                                                                                               'links': [       {       'connected_time': '538573',
                                                                                                                        'fiber_length': 9759,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTFEQ3BlZXJwb25wb24gMjU4MDM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4398 LINE 52 (WEBER DAN & TAMMY)',
                                         'poller_hash': 'MTFCRTFEQw==',
                                         'serial_number': '11BE1DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:5c:54',
                                         'uptime': '538573'},
                                 {       'counter_source': '258182/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '258182/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '74308514060',
                                                                       'in_ucast_pkts': '955442772',
                                                                       'mac_address': '04:bc:9f:6e:ea:86',
                                                                       'name': '258182/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3689331749908',
                                                                       'out_ucast_pkts': '2457488621',
                                                                       'poller_hash': 'MTI5NzBENHBlZXJpbnRmMjU4MTgyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '258182',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3689331749908',
                                                                       'in_ucast_pkts': '2457488621',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3689331749908',
                                                                                                'in_ucast_pkts': '2457488621',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '74308514060',
                                                                                                'out_ucast_pkts': '955442772',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '74308514060',
                                                                       'out_ucast_pkts': '955442772',
                                                                       'poller_hash': 'MTI5NzBENHBlZXJpbnRmcG9uIDI1ODE4Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '258182',
                                                                                               'links': [       {       'connected_time': '22986214',
                                                                                                                        'fiber_length': 6878,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzBENHBlZXJwb25wb24gMjU4MTgy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4457 COMMUNITY CENTRE DRIVE (SHANTZ '
                                                 'MARIANNE)',
                                         'poller_hash': 'MTI5NzBENA==',
                                         'serial_number': '12970D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:ea:86',
                                         'uptime': '22986214'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '258192/g1',
                                                                       'name': '258192/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTFCRTM5NXBlZXJpbnRmMjU4MTkyL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '258192',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTFCRTM5NXBlZXJpbnRmcG9uIDI1ODE5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '258192',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTM5NXBlZXJwb25wb24gMjU4MTky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4230 LINE 67 (KUEPFER ROSEANA)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "258192" '
                                                                  '(location: GPON '
                                                                  'port "13/5"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2024-11-27T19:41:38.430Z',
                                         'poller_hash': 'MTFCRTM5NQ==',
                                         'serial_number': '11BE395',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '258233/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '258233/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '40308285184',
                                                                       'in_ucast_pkts': '144508462',
                                                                       'mac_address': 'b8:94:70:b9:d6:a5',
                                                                       'name': '258233/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1053277925633',
                                                                       'out_ucast_pkts': '839647390',
                                                                       'poller_hash': 'MTE5MkZFOXBlZXJpbnRmMjU4MjMzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '258233',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1053277925633',
                                                                       'in_ucast_pkts': '839647390',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1053277925633',
                                                                                                'in_ucast_pkts': '839647390',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '40308285184',
                                                                                                'out_ucast_pkts': '144508462',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '40308285184',
                                                                       'out_ucast_pkts': '144508462',
                                                                       'poller_hash': 'MTE5MkZFOXBlZXJpbnRmcG9uIDI1ODIzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '258233',
                                                                                               'links': [       {       'connected_time': '688594',
                                                                                                                        'fiber_length': 11694,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MkZFOXBlZXJwb25wb24gMjU4MjMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3996A LINE 61 (BARTSCH DYCK PETER)',
                                         'poller_hash': 'MTE5MkZFOQ==',
                                         'serial_number': '1192FE9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:d6:a5',
                                         'uptime': '688594'},
                                 {       'counter_source': '258263/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:6e:e7'],
                                                                       'id': '258263/g1',
                                                                       'in_nucast_pkts': '9630',
                                                                       'in_octets': '833120305',
                                                                       'in_ucast_pkts': '3024575',
                                                                       'mac_address': '04:bc:9f:6e:f7:29',
                                                                       'mtu': 2000,
                                                                       'name': '258263/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151535',
                                                                       'out_octets': '5911150371',
                                                                       'out_ucast_pkts': '5637661',
                                                                       'poller_hash': 'MTI5NzI2OHBlZXJpbnRmMjU4MjYzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '258263',
                                                                       'in_nucast_pkts': '1151535',
                                                                       'in_octets': '5911150371',
                                                                       'in_ucast_pkts': '5637661',
                                                                       'links': [       {       'in_nucast_pkts': '1151535',
                                                                                                'in_octets': '5911150371',
                                                                                                'in_ucast_pkts': '5637661',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9630',
                                                                                                'out_octets': '833120305',
                                                                                                'out_ucast_pkts': '3024575',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9630',
                                                                       'out_octets': '833120305',
                                                                       'out_ucast_pkts': '3024575',
                                                                       'poller_hash': 'MTI5NzI2OHBlZXJpbnRmcG9uIDI1ODI2Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '258263',
                                                                                               'links': [       {       'connected_time': '538377',
                                                                                                                        'fiber_length': 24193,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzI2OHBlZXJwb25wb24gMjU4MjYz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4499 LINE 46 (VANSTRAATEN CATHY)',
                                         'poller_hash': 'MTI5NzI2OA==',
                                         'serial_number': '1297268',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:f7:26',
                                         'uptime': '538377'},
                                 {       'counter_source': '258315/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '258315/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '84890664013',
                                                                       'in_ucast_pkts': '303963365',
                                                                       'mac_address': '04:bc:9f:6e:fe:e6',
                                                                       'name': '258315/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '25',
                                                                       'out_octets': '1470573590492',
                                                                       'out_ucast_pkts': '1125143345',
                                                                       'poller_hash': 'MTI5NzM2MHBlZXJpbnRmMjU4MzE1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '258315',
                                                                       'in_nucast_pkts': '25',
                                                                       'in_octets': '1470573590492',
                                                                       'in_ucast_pkts': '1125143345',
                                                                       'links': [       {       'in_nucast_pkts': '25',
                                                                                                'in_octets': '1470573590492',
                                                                                                'in_ucast_pkts': '1125143345',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '84890664013',
                                                                                                'out_ucast_pkts': '303963365',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '84890664013',
                                                                       'out_ucast_pkts': '303963365',
                                                                       'poller_hash': 'MTI5NzM2MHBlZXJpbnRmcG9uIDI1ODMxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '258315',
                                                                                               'links': [       {       'connected_time': '538604',
                                                                                                                        'fiber_length': 16497,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM2MHBlZXJwb25wb24gMjU4MzE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4028 LINE 50 (YODER ARNOLD)',
                                         'poller_hash': 'MTI5NzM2MA==',
                                         'serial_number': '1297360',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:fe:e6',
                                         'uptime': '538604'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '259510/g1',
                                                                       'name': '259510/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzM3NXBlZXJpbnRmMjU5NTEwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '259510',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzM3NXBlZXJpbnRmcG9uIDI1OTUxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '259510',
                                                                                               'links': [       {       'connected_time': '538989',
                                                                                                                        'fiber_length': 6859,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM3NXBlZXJwb25wb24gMjU5NTEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "4158 LINE 67 (ALBRECHT'S CHAIN SAW "
                                                 'SERVICE)',
                                         'poller_hash': 'MTI5NzM3NQ==',
                                         'serial_number': '1297375',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:ff:8e',
                                         'uptime': '538989'},
                                 {       'counter_source': '259892/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '259892/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15518169321',
                                                                       'in_ucast_pkts': '62697067',
                                                                       'mac_address': '04:bc:9f:6f:00:36',
                                                                       'name': '259892/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '126316107111',
                                                                       'out_ucast_pkts': '146365603',
                                                                       'poller_hash': 'MTI5NzM4QXBlZXJpbnRmMjU5ODkyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '259892',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '126316107111',
                                                                       'in_ucast_pkts': '146365603',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '126316107111',
                                                                                                'in_ucast_pkts': '146365603',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '15518169321',
                                                                                                'out_ucast_pkts': '62697067',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15518169321',
                                                                       'out_ucast_pkts': '62697067',
                                                                       'poller_hash': 'MTI5NzM4QXBlZXJpbnRmcG9uIDI1OTg5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '259892',
                                                                                               'links': [       {       'connected_time': '22121595',
                                                                                                                        'fiber_length': 6675,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzM4QXBlZXJwb25wb24gMjU5ODky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "4464 PERTH LINE 72 (JACOBS' FAMILY "
                                                 'SERVICE)',
                                         'poller_hash': 'MTI5NzM4QQ==',
                                         'serial_number': '129738A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:00:36',
                                         'uptime': '22121595'},
                                 {       'counter_source': '259912/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '259912/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '19732042760',
                                                                       'in_ucast_pkts': '48502659',
                                                                       'mac_address': '04:bc:9f:5c:0a:21',
                                                                       'name': '259912/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '140785676232',
                                                                       'out_ucast_pkts': '132638040',
                                                                       'poller_hash': 'MTI3NDdDNHBlZXJpbnRmMjU5OTEyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '259912',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '140785676232',
                                                                       'in_ucast_pkts': '132638040',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '140785676232',
                                                                                                'in_ucast_pkts': '132638040',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '19732042760',
                                                                                                'out_ucast_pkts': '48502659',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '19732042760',
                                                                       'out_ucast_pkts': '48502659',
                                                                       'poller_hash': 'MTI3NDdDNHBlZXJpbnRmcG9uIDI1OTkxMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '259912',
                                                                                               'links': [       {       'connected_time': '16416929',
                                                                                                                        'fiber_length': 8865,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdDNHBlZXJwb25wb24gMjU5OTEy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4785 LINE 67 (WAGLER WELDING)',
                                         'poller_hash': 'MTI3NDdDNA==',
                                         'serial_number': '12747C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0a:21',
                                         'uptime': '16416929'},
                                 {       'counter_source': '260364/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:70:7e'],
                                                                       'id': '260364/g1',
                                                                       'in_nucast_pkts': '414940',
                                                                       'in_octets': '108374894355',
                                                                       'in_ucast_pkts': '722802903',
                                                                       'mac_address': '04:bc:9f:5c:11:8c',
                                                                       'mtu': 2000,
                                                                       'name': '260364/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '50077548',
                                                                       'out_octets': '2498861453360',
                                                                       'out_ucast_pkts': '1971032579',
                                                                       'poller_hash': 'MTI3NDhCMXBlZXJpbnRmMjYwMzY0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '260364',
                                                                       'in_nucast_pkts': '50077548',
                                                                       'in_octets': '2498861453360',
                                                                       'in_ucast_pkts': '1971032579',
                                                                       'links': [       {       'in_nucast_pkts': '50077548',
                                                                                                'in_octets': '2498861453360',
                                                                                                'in_ucast_pkts': '1971032579',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '414940',
                                                                                                'out_octets': '108374894355',
                                                                                                'out_ucast_pkts': '722802903',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '414940',
                                                                       'out_octets': '108374894355',
                                                                       'out_ucast_pkts': '722802903',
                                                                       'poller_hash': 'MTI3NDhCMXBlZXJpbnRmcG9uIDI2MDM2NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '260364',
                                                                                               'links': [       {       'connected_time': '21777412',
                                                                                                                        'fiber_length': 13616,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDhCMXBlZXJwb25wb24gMjYwMzY0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6475 ROAD 116 (SCHULTZ WILLIAM)',
                                         'poller_hash': 'MTI3NDhCMQ==',
                                         'serial_number': '12748B1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:11:89',
                                         'uptime': '21777412'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '260387/g1',
                                                                       'name': '260387/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTk5OUZGcGVlcmludGYyNjAzODcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '260387',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTk5OUZGcGVlcmludGZwb24gMjYwMzg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '260387',
                                                                                               'links': [       {       'connected_time': '1761814',
                                                                                                                        'fiber_length': 6967,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.468,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTk5OUZGcGVlcnBvbnBvbiAyNjAzODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4186 LINE 67 (JANTZI MAHLON K)',
                                         'poller_hash': 'MTk5OUZG',
                                         'serial_number': '1999FF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:e3:fb:33',
                                         'uptime': '1761814'},
                                 {       'counter_source': '261227/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '261227/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '56177634478',
                                                                       'in_ucast_pkts': '296104024',
                                                                       'mac_address': '04:bc:9f:5c:02:81',
                                                                       'name': '261227/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1956862626195',
                                                                       'out_ucast_pkts': '1444956721',
                                                                       'poller_hash': 'MTI3NDZEMHBlZXJpbnRmMjYxMjI3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '261227',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1956862626195',
                                                                       'in_ucast_pkts': '1444956721',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1956862626195',
                                                                                                'in_ucast_pkts': '1444956721',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '56177634478',
                                                                                                'out_ucast_pkts': '296104024',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '56177634478',
                                                                       'out_ucast_pkts': '296104024',
                                                                       'poller_hash': 'MTI3NDZEMHBlZXJpbnRmcG9uIDI2MTIyNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '261227',
                                                                                               'links': [       {       'connected_time': '21707981',
                                                                                                                        'fiber_length': 6806,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDZEMHBlZXJwb25wb24gMjYxMjI3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4429 PERTH LINE 72 (TOMIUK TIM)',
                                         'poller_hash': 'MTI3NDZEMA==',
                                         'serial_number': '12746D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:02:81',
                                         'uptime': '21707981'},
                                 {       'counter_source': '261319/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '261319/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '16693248380',
                                                                       'in_ucast_pkts': '44615614',
                                                                       'mac_address': '04:bc:9f:5c:0e:e1',
                                                                       'name': '261319/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '181428968772',
                                                                       'out_ucast_pkts': '178850269',
                                                                       'poller_hash': 'MTI3NDg1Q3BlZXJpbnRmMjYxMzE5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '261319',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '181428968772',
                                                                       'in_ucast_pkts': '178850269',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '181428968772',
                                                                                                'in_ucast_pkts': '178850269',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '16693248380',
                                                                                                'out_ucast_pkts': '44615614',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '16693248380',
                                                                       'out_ucast_pkts': '44615614',
                                                                       'poller_hash': 'MTI3NDg1Q3BlZXJpbnRmcG9uIDI2MTMxOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '261319',
                                                                                               'links': [       {       'connected_time': '688599',
                                                                                                                        'fiber_length': 6586,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg1Q3BlZXJwb25wb24gMjYxMzE5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7008 ROAD 129 (NEWTON IRON INC)',
                                         'poller_hash': 'MTI3NDg1Qw==',
                                         'serial_number': '127485C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0e:e1',
                                         'uptime': '688599'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '261342/g1',
                                                                       'name': '261342/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDdENnBlZXJpbnRmMjYxMzQyL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '261342',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDdENnBlZXJpbnRmcG9uIDI2MTM0Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '261342',
                                                                                               'links': [       {       'connected_time': '21621629',
                                                                                                                        'fiber_length': 11465,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdENnBlZXJwb25wb24gMjYxMzQy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7026 PERTH ROAD 121 (SCHMIDT '
                                                 'KATHRYN)',
                                         'poller_hash': 'MTI3NDdENg==',
                                         'serial_number': '12747D6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0a:b1',
                                         'uptime': '21621629'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '261456/g1',
                                                                       'name': '261456/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDZDQXBlZXJpbnRmMjYxNDU2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '261456',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDZDQXBlZXJpbnRmcG9uIDI2MTQ1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '261456',
                                                                                               'links': [       {       'connected_time': '20148339',
                                                                                                                        'fiber_length': 6689,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -26.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDZDQXBlZXJwb25wb24gMjYxNDU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4372 LINE 61 (ALBRECHT JOHN)',
                                         'poller_hash': 'MTI3NDZDQQ==',
                                         'serial_number': '12746CA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:02:51',
                                         'uptime': '20148339'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '262457/g1',
                                                                       'name': '262457/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDdFMXBlZXJpbnRmMjYyNDU3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262457',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDdFMXBlZXJpbnRmcG9uIDI2MjQ1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262457',
                                                                                               'links': [       {       'connected_time': '21519489',
                                                                                                                        'fiber_length': 6044,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdFMXBlZXJwb25wb24gMjYyNDU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4951 LINE 67 (KUEPFER MANUEL & '
                                                 'IRENE)',
                                         'poller_hash': 'MTI3NDdFMQ==',
                                         'serial_number': '12747E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0b:09',
                                         'uptime': '21519489'},
                                 {       'counter_source': '262577/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262577/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '130936577180',
                                                                       'in_ucast_pkts': '907415045',
                                                                       'mac_address': '04:bc:9f:5c:01:d9',
                                                                       'name': '262577/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '32',
                                                                       'out_octets': '4873947763248',
                                                                       'out_ucast_pkts': '3369877515',
                                                                       'poller_hash': 'MTI3NDZCQnBlZXJpbnRmMjYyNTc3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262577',
                                                                       'in_nucast_pkts': '32',
                                                                       'in_octets': '4873947763248',
                                                                       'in_ucast_pkts': '3369877515',
                                                                       'links': [       {       'in_nucast_pkts': '32',
                                                                                                'in_octets': '4873947763248',
                                                                                                'in_ucast_pkts': '3369877515',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '130936577180',
                                                                                                'out_ucast_pkts': '907415045',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '130936577180',
                                                                       'out_ucast_pkts': '907415045',
                                                                       'poller_hash': 'MTI3NDZCQnBlZXJpbnRmcG9uIDI2MjU3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262577',
                                                                                               'links': [       {       'connected_time': '21175394',
                                                                                                                        'fiber_length': 982,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDZCQnBlZXJwb25wb24gMjYyNTc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 ST MARYS ST S (KERSTEN THEO)',
                                         'poller_hash': 'MTI3NDZCQg==',
                                         'serial_number': '12746BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:01:d9',
                                         'uptime': '21175394'},
                                 {       'counter_source': '262607/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:88:e6'],
                                                                       'id': '262607/g1',
                                                                       'in_nucast_pkts': '546936',
                                                                       'in_octets': '152696317549',
                                                                       'in_ucast_pkts': '610162882',
                                                                       'mac_address': '00:06:31:e9:02:50',
                                                                       'mtu': 2000,
                                                                       'name': '262607/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '48258062',
                                                                       'out_octets': '3550280734473',
                                                                       'out_ucast_pkts': '2617468525',
                                                                       'poller_hash': 'MUQwNjlGcGVlcmludGYyNjI2MDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '262607',
                                                                       'in_nucast_pkts': '48258062',
                                                                       'in_octets': '3550280734473',
                                                                       'in_ucast_pkts': '2617468525',
                                                                       'links': [       {       'in_nucast_pkts': '48258062',
                                                                                                'in_octets': '3550280734473',
                                                                                                'in_ucast_pkts': '2617468525',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '546936',
                                                                                                'out_octets': '152696317549',
                                                                                                'out_ucast_pkts': '610162882',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '546936',
                                                                       'out_octets': '152696317549',
                                                                       'out_ucast_pkts': '610162882',
                                                                       'poller_hash': 'MUQwNjlGcGVlcmludGZwb24gMjYyNjA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262607',
                                                                                               'links': [       {       'connected_time': '20904113',
                                                                                                                        'fiber_length': 1168,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjlGcGVlcnBvbnBvbiAyNjI2MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '11 PACIFIC AV (WEBER SARAH)',
                                         'poller_hash': 'MUQwNjlG',
                                         'serial_number': '1D069F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:02:4d',
                                         'uptime': '20904113'},
                                 {       'counter_source': '262638/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262638/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '19281366609',
                                                                       'in_ucast_pkts': '102256862',
                                                                       'mac_address': '04:bc:9f:5c:04:89',
                                                                       'name': '262638/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '400159724318',
                                                                       'out_ucast_pkts': '329410060',
                                                                       'poller_hash': 'MTI3NDcxMXBlZXJpbnRmMjYyNjM4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262638',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '400159724318',
                                                                       'in_ucast_pkts': '329410060',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '400159724318',
                                                                                                'in_ucast_pkts': '329410060',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '19281366609',
                                                                                                'out_ucast_pkts': '102256862',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '19281366609',
                                                                       'out_ucast_pkts': '102256862',
                                                                       'poller_hash': 'MTI3NDcxMXBlZXJpbnRmcG9uIDI2MjYzOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262638',
                                                                                               'links': [       {       'connected_time': '688590',
                                                                                                                        'fiber_length': 6555,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDcxMXBlZXJwb25wb24gMjYyNjM4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4444 PERTH LINE 72 (MARTIN CHRIS)',
                                         'poller_hash': 'MTI3NDcxMQ==',
                                         'serial_number': '1274711',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:04:89',
                                         'uptime': '688590'},
                                 {       'counter_source': '262648/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262648/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '62284828309',
                                                                       'in_ucast_pkts': '225313609',
                                                                       'mac_address': '04:bc:9f:5c:09:c1',
                                                                       'name': '262648/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '2067151778325',
                                                                       'out_ucast_pkts': '1481997103',
                                                                       'poller_hash': 'MTI3NDdCOHBlZXJpbnRmMjYyNjQ4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262648',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '2067151778325',
                                                                       'in_ucast_pkts': '1481997103',
                                                                       'links': [       {       'in_nucast_pkts': '23',
                                                                                                'in_octets': '2067151778325',
                                                                                                'in_ucast_pkts': '1481997103',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '62284828309',
                                                                                                'out_ucast_pkts': '225313609',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '62284828309',
                                                                       'out_ucast_pkts': '225313609',
                                                                       'poller_hash': 'MTI3NDdCOHBlZXJpbnRmcG9uIDI2MjY0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262648',
                                                                                               'links': [       {       'connected_time': '21086025',
                                                                                                                        'fiber_length': 6983,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdCOHBlZXJwb25wb24gMjYyNjQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5977 PERTH ROAD 119 (JANTZI BRIAN & '
                                                 'BEV)',
                                         'poller_hash': 'MTI3NDdCOA==',
                                         'serial_number': '12747B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:09:c1',
                                         'uptime': '21086025'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '262678/g1',
                                                                       'name': '262678/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDdBRnBlZXJpbnRmMjYyNjc4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262678',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDdBRnBlZXJpbnRmcG9uIDI2MjY3OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262678',
                                                                                               'links': [       {       'connected_time': '20998295',
                                                                                                                        'fiber_length': 11626,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdBRnBlZXJwb25wb24gMjYyNjc4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7014 PERTH ROAD 121 (KUEPFER WAYNE)',
                                         'poller_hash': 'MTI3NDdBRg==',
                                         'serial_number': '12747AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:09:79',
                                         'uptime': '20998295'},
                                 {       'counter_source': '262698/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262698/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '212958900526',
                                                                       'in_ucast_pkts': '881871380',
                                                                       'mac_address': 'ec:4f:82:4e:e7:47',
                                                                       'name': '262698/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '64',
                                                                       'out_octets': '3272129119599',
                                                                       'out_ucast_pkts': '2402602926',
                                                                       'poller_hash': 'Mjk2RkVCcGVlcmludGYyNjI2OTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '262698',
                                                                       'in_nucast_pkts': '64',
                                                                       'in_octets': '3272129119599',
                                                                       'in_ucast_pkts': '2402602926',
                                                                       'links': [       {       'in_nucast_pkts': '64',
                                                                                                'in_octets': '3272129119599',
                                                                                                'in_ucast_pkts': '2402602926',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '212958900526',
                                                                                                'out_ucast_pkts': '881871380',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '212958900526',
                                                                       'out_ucast_pkts': '881871380',
                                                                       'poller_hash': 'Mjk2RkVCcGVlcmludGZwb24gMjYyNjk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262698',
                                                                                               'links': [       {       'connected_time': '688455',
                                                                                                                        'fiber_length': 1322,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk2RkVCcGVlcnBvbnBvbiAyNjI2OTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 EDWARD ST (WATSON BRENNAN)',
                                         'poller_hash': 'Mjk2RkVC',
                                         'serial_number': '296FEB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:4e:e7:47',
                                         'uptime': '688455'},
                                 {       'counter_source': '262765/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8c:a3'],
                                                                       'id': '262765/g1',
                                                                       'in_nucast_pkts': '14806',
                                                                       'in_octets': '4082537689',
                                                                       'in_ucast_pkts': '6992056',
                                                                       'mac_address': '04:bc:9f:5c:0e:74',
                                                                       'mtu': 2000,
                                                                       'name': '262765/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1464888',
                                                                       'out_octets': '27032651252',
                                                                       'out_ucast_pkts': '22884362',
                                                                       'poller_hash': 'MTI3NDg0RXBlZXJpbnRmMjYyNzY1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '262765',
                                                                       'in_nucast_pkts': '1464888',
                                                                       'in_octets': '27032651252',
                                                                       'in_ucast_pkts': '22884362',
                                                                       'links': [       {       'in_nucast_pkts': '1464888',
                                                                                                'in_octets': '27032651252',
                                                                                                'in_ucast_pkts': '22884362',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14806',
                                                                                                'out_octets': '4082537689',
                                                                                                'out_ucast_pkts': '6992056',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14806',
                                                                       'out_octets': '4082537689',
                                                                       'out_ucast_pkts': '6992056',
                                                                       'poller_hash': 'MTI3NDg0RXBlZXJpbnRmcG9uIDI2Mjc2NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262765',
                                                                                               'links': [       {       'connected_time': '688514',
                                                                                                                        'fiber_length': 11878,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg0RXBlZXJwb25wb24gMjYyNzY1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4042 WATERLOO ST (PENNER JOHAN)',
                                         'poller_hash': 'MTI3NDg0RQ==',
                                         'serial_number': '127484E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0e:71',
                                         'uptime': '688514'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '262808/g1',
                                                                       'name': '262808/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTU5RjdGOXBlZXJpbnRmMjYyODA4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262808',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTU5RjdGOXBlZXJpbnRmcG9uIDI2MjgwOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262808',
                                                                                               'links': [       {       'connected_time': '13908384',
                                                                                                                        'fiber_length': 6379,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.256,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU5RjdGOXBlZXJwb25wb24gMjYyODA4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6751 ROAD 124 (KUEPFER ARNOLD)',
                                         'poller_hash': 'MTU5RjdGOQ==',
                                         'serial_number': '159F7F9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '1c:8b:76:66:6b:51',
                                         'uptime': '13908384'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '262817/g1',
                                                                       'name': '262817/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDg0NHBlZXJpbnRmMjYyODE3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262817',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDg0NHBlZXJpbnRmcG9uIDI2MjgxNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262817',
                                                                                               'links': [       {       'connected_time': '4750292',
                                                                                                                        'fiber_length': 11534,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg0NHBlZXJwb25wb24gMjYyODE3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4003 LINE 61 (JANTZI ALLAN)',
                                         'poller_hash': 'MTI3NDg0NA==',
                                         'serial_number': '1274844',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0e:21',
                                         'uptime': '4750292'},
                                 {       'counter_source': '262915/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262915/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '52902911989',
                                                                       'in_ucast_pkts': '238322207',
                                                                       'mac_address': '04:bc:9f:5c:07:09',
                                                                       'name': '262915/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1522235404086',
                                                                       'out_ucast_pkts': '1157582200',
                                                                       'poller_hash': 'MTI3NDc2MXBlZXJpbnRmMjYyOTE1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262915',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1522235404086',
                                                                       'in_ucast_pkts': '1157582200',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1522235404086',
                                                                                                'in_ucast_pkts': '1157582200',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '52902911989',
                                                                                                'out_ucast_pkts': '238322207',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '52902911989',
                                                                       'out_ucast_pkts': '238322207',
                                                                       'poller_hash': 'MTI3NDc2MXBlZXJpbnRmcG9uIDI2MjkxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262915',
                                                                                               'links': [       {       'connected_time': '20565187',
                                                                                                                        'fiber_length': 15505,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc2MXBlZXJwb25wb24gMjYyOTE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7464 PERTH ROAD 121 (ALBRECHT '
                                                 'EDWARD)',
                                         'poller_hash': 'MTI3NDc2MQ==',
                                         'serial_number': '1274761',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:07:09',
                                         'uptime': '20565187'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '262955/g1',
                                                                       'name': '262955/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDJENHBlZXJpbnRmMjYyOTU1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262955',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDJENHBlZXJpbnRmcG9uIDI2Mjk1NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262955',
                                                                                               'links': [       {       'connected_time': '20488397',
                                                                                                                        'fiber_length': 8843,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJENHBlZXJwb25wb24gMjYyOTU1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4010 LINE 67 (KUEPFER PAUL A & MARY)',
                                         'poller_hash': 'MTI3NDJENA==',
                                         'serial_number': '12742D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e2:a1',
                                         'uptime': '20488397'},
                                 {       'counter_source': '263015/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '263015/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '65227167881',
                                                                       'in_ucast_pkts': '294494954',
                                                                       'mac_address': '04:bc:9f:5c:05:d1',
                                                                       'name': '263015/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1030503100642',
                                                                       'out_ucast_pkts': '757633119',
                                                                       'poller_hash': 'MTI3NDczQXBlZXJpbnRmMjYzMDE1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263015',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1030503100642',
                                                                       'in_ucast_pkts': '757633119',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1030503100642',
                                                                                                'in_ucast_pkts': '757633119',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '65227167881',
                                                                                                'out_ucast_pkts': '294494954',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '65227167881',
                                                                       'out_ucast_pkts': '294494954',
                                                                       'poller_hash': 'MTI3NDczQXBlZXJpbnRmcG9uIDI2MzAxNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263015',
                                                                                               'links': [       {       'connected_time': '688465',
                                                                                                                        'fiber_length': 10181,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDczQXBlZXJwb25wb24gMjYzMDE1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5595 ROAD 126 (ROTH BRENT)',
                                         'poller_hash': 'MTI3NDczQQ==',
                                         'serial_number': '127473A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:05:d1',
                                         'uptime': '688465'},
                                 {       'counter_source': '263145/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '263145/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4986301245',
                                                                       'in_ucast_pkts': '31184593',
                                                                       'mac_address': '04:bc:9f:5c:05:e9',
                                                                       'name': '263145/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '297378356947',
                                                                       'out_ucast_pkts': '243060673',
                                                                       'poller_hash': 'MTI3NDczRHBlZXJpbnRmMjYzMTQ1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263145',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '297378356947',
                                                                       'in_ucast_pkts': '243060673',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '297378356947',
                                                                                                'in_ucast_pkts': '243060673',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '4986301245',
                                                                                                'out_ucast_pkts': '31184593',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4986301245',
                                                                       'out_ucast_pkts': '31184593',
                                                                       'poller_hash': 'MTI3NDczRHBlZXJpbnRmcG9uIDI2MzE0NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263145',
                                                                                               'links': [       {       'connected_time': '20049931',
                                                                                                                        'fiber_length': 6190,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDczRHBlZXJwb25wb24gMjYzMTQ1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6019 PERTH ROAD 119 (GERLACH CONNIE)',
                                         'poller_hash': 'MTI3NDczRA==',
                                         'serial_number': '127473D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:05:e9',
                                         'uptime': '20049931'},
                                 {       'counter_source': '263191/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8a:d5'],
                                                                       'id': '263191/g1',
                                                                       'in_nucast_pkts': '14813',
                                                                       'in_octets': '5535322238',
                                                                       'in_ucast_pkts': '34607884',
                                                                       'mac_address': '04:bc:9f:5c:0a:34',
                                                                       'mtu': 2000,
                                                                       'name': '263191/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1448004',
                                                                       'out_octets': '167389906035',
                                                                       'out_ucast_pkts': '115593807',
                                                                       'poller_hash': 'MTI3NDdDNnBlZXJpbnRmMjYzMTkxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '263191',
                                                                       'in_nucast_pkts': '1448004',
                                                                       'in_octets': '167389906035',
                                                                       'in_ucast_pkts': '115593807',
                                                                       'links': [       {       'in_nucast_pkts': '1448004',
                                                                                                'in_octets': '167389906035',
                                                                                                'in_ucast_pkts': '115593807',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14813',
                                                                                                'out_octets': '5535322238',
                                                                                                'out_ucast_pkts': '34607884',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14813',
                                                                       'out_octets': '5535322238',
                                                                       'out_ucast_pkts': '34607884',
                                                                       'poller_hash': 'MTI3NDdDNnBlZXJpbnRmcG9uIDI2MzE5MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263191',
                                                                                               'links': [       {       'connected_time': '688381',
                                                                                                                        'fiber_length': 8478,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdDNnBlZXJwb25wb24gMjYzMTkx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5250 LINE 60 (CARTER JORDAN)',
                                         'poller_hash': 'MTI3NDdDNg==',
                                         'serial_number': '12747C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0a:31',
                                         'uptime': '688381'},
                                 {       'counter_source': '263201/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '263201/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '6470360',
                                                                       'in_octets': '243476740482',
                                                                       'in_ucast_pkts': '1023563348',
                                                                       'mac_address': '04:bc:9f:24:09:26',
                                                                       'name': '263201/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '35200942',
                                                                       'out_octets': '10725007082850',
                                                                       'out_ucast_pkts': '7367435187',
                                                                       'poller_hash': 'MTUyMDJDN3BlZXJpbnRmMjYzMjAxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263201',
                                                                       'in_nucast_pkts': '35200942',
                                                                       'in_octets': '10725007082850',
                                                                       'in_ucast_pkts': '7367435187',
                                                                       'links': [       {       'in_nucast_pkts': '35200942',
                                                                                                'in_octets': '10725007082850',
                                                                                                'in_ucast_pkts': '7367435187',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6470360',
                                                                                                'out_octets': '243476740482',
                                                                                                'out_ucast_pkts': '1023563348',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6470360',
                                                                       'out_octets': '243476740482',
                                                                       'out_ucast_pkts': '1023563348',
                                                                       'poller_hash': 'MTUyMDJDN3BlZXJpbnRmcG9uIDI2MzIwMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263201',
                                                                                               'links': [       {       'connected_time': '688366',
                                                                                                                        'fiber_length': 1709,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.416,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.888}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUyMDJDN3BlZXJwb25wb24gMjYzMjAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': '55 WILLIAM ST (ASSELSTINE MAGDELINA)',
                                         'poller_hash': 'MTUyMDJDNw==',
                                         'serial_number': '15202C7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.2.0.2.2',
                                         'system_mac_address': '04:bc:9f:24:09:26',
                                         'uptime': '688366'},
                                 {       'counter_source': '263866/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '263866/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '21052005427',
                                                                       'in_ucast_pkts': '107020169',
                                                                       'mac_address': '04:bc:9f:5c:08:69',
                                                                       'name': '263866/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '710772122918',
                                                                       'out_ucast_pkts': '551345528',
                                                                       'poller_hash': 'MTI3NDc4RHBlZXJpbnRmMjYzODY2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263866',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '710772122918',
                                                                       'in_ucast_pkts': '551345528',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '710772122918',
                                                                                                'in_ucast_pkts': '551345528',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '21052005427',
                                                                                                'out_ucast_pkts': '107020169',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '21052005427',
                                                                       'out_ucast_pkts': '107020169',
                                                                       'poller_hash': 'MTI3NDc4RHBlZXJpbnRmcG9uIDI2Mzg2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263866',
                                                                                               'links': [       {       'connected_time': '688591',
                                                                                                                        'fiber_length': 11520,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc4RHBlZXJwb25wb24gMjYzODY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4006 LINE 61 (STREICHER LAVERNE)',
                                         'poller_hash': 'MTI3NDc4RA==',
                                         'serial_number': '127478D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:08:69',
                                         'uptime': '688591'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '263876/g1',
                                                                       'name': '263876/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDYwRXBlZXJpbnRmMjYzODc2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263876',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDYwRXBlZXJpbnRmcG9uIDI2Mzg3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263876',
                                                                                               'links': [       {       'connected_time': '19441886',
                                                                                                                        'fiber_length': 952,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDYwRXBlZXJwb25wb24gMjYzODc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 ST MARYS ST N (GROPP KEN (ESTATE '
                                                 'OF))',
                                         'poller_hash': 'MTI3NDYwRQ==',
                                         'serial_number': '127460E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:fc:71',
                                         'uptime': '19441886'},
                                 {       'counter_source': '263896/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '263896/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '23941028851',
                                                                       'in_ucast_pkts': '229513627',
                                                                       'mac_address': '04:bc:9f:6f:02:ee',
                                                                       'name': '263896/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1195536599962',
                                                                       'out_ucast_pkts': '823461452',
                                                                       'poller_hash': 'MTI5NzNFMXBlZXJpbnRmMjYzODk2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263896',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1195536599962',
                                                                       'in_ucast_pkts': '823461452',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1195536599962',
                                                                                                'in_ucast_pkts': '823461452',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '23941028851',
                                                                                                'out_ucast_pkts': '229513627',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '23941028851',
                                                                       'out_ucast_pkts': '229513627',
                                                                       'poller_hash': 'MTI5NzNFMXBlZXJpbnRmcG9uIDI2Mzg5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263896',
                                                                                               'links': [       {       'connected_time': '688333',
                                                                                                                        'fiber_length': 1336,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNFMXBlZXJwb25wb24gMjYzODk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '85 MAIN ST N (THEDORFF SANDY)',
                                         'poller_hash': 'MTI5NzNFMQ==',
                                         'serial_number': '12973E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:02:ee',
                                         'uptime': '688333'},
                                 {       'counter_source': '263941/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '263941/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '247998334707',
                                                                       'in_ucast_pkts': '349203323',
                                                                       'mac_address': '04:bc:9f:6f:02:5e',
                                                                       'name': '263941/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1487471332042',
                                                                       'out_ucast_pkts': '1158896053',
                                                                       'poller_hash': 'MTI5NzNDRnBlZXJpbnRmMjYzOTQxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263941',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1487471332042',
                                                                       'in_ucast_pkts': '1158896053',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1487471332042',
                                                                                                'in_ucast_pkts': '1158896053',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '247998334707',
                                                                                                'out_ucast_pkts': '349203323',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '247998334707',
                                                                       'out_ucast_pkts': '349203323',
                                                                       'poller_hash': 'MTI5NzNDRnBlZXJpbnRmcG9uIDI2Mzk0MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263941',
                                                                                               'links': [       {       'connected_time': '688587',
                                                                                                                        'fiber_length': 10513,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -25.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNDRnBlZXJwb25wb24gMjYzOTQx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4109 LINE 61 (HILDEBRANDT ROSANA)',
                                         'poller_hash': 'MTI5NzNDRg==',
                                         'serial_number': '12973CF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:02:5e',
                                         'uptime': '688587'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '263951/g1',
                                                                       'name': '263951/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NkRFRHBlZXJpbnRmMjYzOTUxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '263951',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NkRFRHBlZXJpbnRmcG9uIDI2Mzk1MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263951',
                                                                                               'links': [       {       'connected_time': '18838847',
                                                                                                                        'fiber_length': 12616,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NkRFRHBlZXJwb25wb24gMjYzOTUx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7118 PERTH ROAD 121 (BAUMAN ABNER)',
                                         'poller_hash': 'MTI5NkRFRA==',
                                         'serial_number': '1296DED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:d3:4e',
                                         'uptime': '18838847'},
                                 {       'counter_source': '264077/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '264077/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'mac_address': '04:bc:9f:26:49:4a',
                                                                       'name': '264077/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUyNDJENXBlZXJpbnRmMjY0MDc3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264077',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUyNDJENXBlZXJpbnRmcG9uIDI2NDA3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264077',
                                                                                               'links': [       {       'connected_time': '688349',
                                                                                                                        'fiber_length': 1704,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.506,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.71}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUyNDJENXBlZXJwb25wb24gMjY0MDc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': '55 WILLIAM ST (ALBRECHT MATTHEW)',
                                         'poller_hash': 'MTUyNDJENQ==',
                                         'serial_number': '15242D5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.4.0.0.57',
                                         'system_mac_address': '04:bc:9f:26:49:4a',
                                         'uptime': '688349'},
                                 {       'counter_source': '264159/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '264159/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1010925905',
                                                                       'in_ucast_pkts': '8208897',
                                                                       'mac_address': '04:bc:9f:6f:05:16',
                                                                       'name': '264159/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '23663915731',
                                                                       'out_ucast_pkts': '34940632',
                                                                       'poller_hash': 'MTI5NzQyNnBlZXJpbnRmMjY0MTU5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264159',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '23663915731',
                                                                       'in_ucast_pkts': '34940632',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '23663915731',
                                                                                                'in_ucast_pkts': '34940632',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1010925905',
                                                                                                'out_ucast_pkts': '8208897',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1010925905',
                                                                       'out_ucast_pkts': '8208897',
                                                                       'poller_hash': 'MTI5NzQyNnBlZXJpbnRmcG9uIDI2NDE1OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264159',
                                                                                               'links': [       {       'connected_time': '688447',
                                                                                                                        'fiber_length': 1013,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQyNnBlZXJwb25wb24gMjY0MTU5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "13 MILL ST E (KIPFER'S GARAGE LTD)",
                                         'poller_hash': 'MTI5NzQyNg==',
                                         'serial_number': '1297426',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:05:16',
                                         'uptime': '688447'},
                                 {       'counter_source': '264391/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '264391/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '29169391521',
                                                                       'in_ucast_pkts': '261558993',
                                                                       'mac_address': '04:bc:9f:6f:06:ce',
                                                                       'name': '264391/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '894772183788',
                                                                       'out_ucast_pkts': '650148684',
                                                                       'poller_hash': 'MTI5NzQ1RHBlZXJpbnRmMjY0MzkxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264391',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '894772183788',
                                                                       'in_ucast_pkts': '650148684',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '894772183788',
                                                                                                'in_ucast_pkts': '650148684',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '29169391521',
                                                                                                'out_ucast_pkts': '261558993',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '29169391521',
                                                                       'out_ucast_pkts': '261558993',
                                                                       'poller_hash': 'MTI5NzQ1RHBlZXJpbnRmcG9uIDI2NDM5MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264391',
                                                                                               'links': [       {       'connected_time': '674969',
                                                                                                                        'fiber_length': 11689,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ1RHBlZXJwb25wb24gMjY0Mzkx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4188 LINE 61 (BULMER D)',
                                         'poller_hash': 'MTI5NzQ1RA==',
                                         'serial_number': '129745D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:06:ce',
                                         'uptime': '674969'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '264474/g1',
                                                                       'name': '264474/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzQyQXBlZXJpbnRmMjY0NDc0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264474',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzQyQXBlZXJpbnRmcG9uIDI2NDQ3NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264474',
                                                                                               'links': [       {       'connected_time': '18149298',
                                                                                                                        'fiber_length': 6818,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQyQXBlZXJwb25wb24gMjY0NDc0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4424 PERTH LINE 72 (STECKLY ELSIE)',
                                         'poller_hash': 'MTI5NzQyQQ==',
                                         'serial_number': '129742A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:05:36',
                                         'uptime': '18149298'},
                                 {       'counter_source': '264494/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '264494/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '5506833',
                                                                       'in_octets': '219034856077',
                                                                       'in_ucast_pkts': '819600521',
                                                                       'mac_address': '04:bc:9f:24:09:6e',
                                                                       'name': '264494/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '30673258',
                                                                       'out_octets': '4213637549858',
                                                                       'out_ucast_pkts': '2967684438',
                                                                       'poller_hash': 'MTUyMDJDRnBlZXJpbnRmMjY0NDk0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264494',
                                                                       'in_nucast_pkts': '30673258',
                                                                       'in_octets': '4213637549858',
                                                                       'in_ucast_pkts': '2967684438',
                                                                       'links': [       {       'in_nucast_pkts': '30673258',
                                                                                                'in_octets': '4213637549858',
                                                                                                'in_ucast_pkts': '2967684438',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5506833',
                                                                                                'out_octets': '219034856077',
                                                                                                'out_ucast_pkts': '819600521',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5506833',
                                                                       'out_octets': '219034856077',
                                                                       'out_ucast_pkts': '819600521',
                                                                       'poller_hash': 'MTUyMDJDRnBlZXJpbnRmcG9uIDI2NDQ5NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264494',
                                                                                               'links': [       {       'connected_time': '688352',
                                                                                                                        'fiber_length': 1017,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.838,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.86}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUyMDJDRnBlZXJwb25wb24gMjY0NDk0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': '13 MILL ST E (SOCKETT KAREN)',
                                         'poller_hash': 'MTUyMDJDRg==',
                                         'serial_number': '15202CF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.2.0.2.2',
                                         'system_mac_address': '04:bc:9f:24:09:6e',
                                         'uptime': '688352'},
                                 {       'counter_source': '264898/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '264898/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'mac_address': '04:bc:9f:24:07:fd',
                                                                       'name': '264898/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUyMDJBNnBlZXJpbnRmMjY0ODk4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264898',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUyMDJBNnBlZXJpbnRmcG9uIDI2NDg5OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264898',
                                                                                               'links': [       {       'connected_time': '688351',
                                                                                                                        'fiber_length': 603,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.522,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.876}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUyMDJBNnBlZXJwb25wb24gMjY0ODk4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': "16 MAIN ST N (O'HEARN DANIEL)",
                                         'poller_hash': 'MTUyMDJBNg==',
                                         'serial_number': '15202A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.2.0.2.2',
                                         'system_mac_address': '04:bc:9f:24:07:fd',
                                         'uptime': '688351'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '264993/g1',
                                                                       'name': '264993/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDdGQnBlZXJpbnRmMjY0OTkzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264993',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDdGQnBlZXJpbnRmcG9uIDI2NDk5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264993',
                                                                                               'links': [       {       'connected_time': '17646665',
                                                                                                                        'fiber_length': 6656,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdGQnBlZXJwb25wb24gMjY0OTkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4462 PERTH LINE 72',
                                         'poller_hash': 'MTI3NDdGQg==',
                                         'serial_number': '12747FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0b:d9',
                                         'uptime': '17646665'},
                                 {       'counter_source': '265003/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '265003/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '14633845057',
                                                                       'in_ucast_pkts': '51834499',
                                                                       'mac_address': '04:bc:9f:6f:03:4e',
                                                                       'name': '265003/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '168840127629',
                                                                       'out_ucast_pkts': '146870991',
                                                                       'poller_hash': 'MTI5NzNFRHBlZXJpbnRmMjY1MDAzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '265003',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '168840127629',
                                                                       'in_ucast_pkts': '146870991',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '168840127629',
                                                                                                'in_ucast_pkts': '146870991',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '14633845057',
                                                                                                'out_ucast_pkts': '51834499',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '14633845057',
                                                                       'out_ucast_pkts': '51834499',
                                                                       'poller_hash': 'MTI5NzNFRHBlZXJpbnRmcG9uIDI2NTAwMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265003',
                                                                                               'links': [       {       'connected_time': '17036726',
                                                                                                                        'fiber_length': 6856,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNFRHBlZXJwb25wb24gMjY1MDAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7022 ROAD 129 (NEWTON RUSTICS)',
                                         'poller_hash': 'MTI5NzNFRA==',
                                         'serial_number': '12973ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:03:4e',
                                         'uptime': '17036726'},
                                 {       'counter_source': '265013/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '265013/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '18347781313',
                                                                       'in_ucast_pkts': '65866143',
                                                                       'mac_address': '04:bc:9f:5c:10:a9',
                                                                       'name': '265013/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '282747110756',
                                                                       'out_ucast_pkts': '249256851',
                                                                       'poller_hash': 'MTI3NDg5NXBlZXJpbnRmMjY1MDEzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '265013',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '282747110756',
                                                                       'in_ucast_pkts': '249256851',
                                                                       'links': [       {       'in_nucast_pkts': '8',
                                                                                                'in_octets': '282747110756',
                                                                                                'in_ucast_pkts': '249256851',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '18347781313',
                                                                                                'out_ucast_pkts': '65866143',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '18347781313',
                                                                       'out_ucast_pkts': '65866143',
                                                                       'poller_hash': 'MTI3NDg5NXBlZXJpbnRmcG9uIDI2NTAxMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265013',
                                                                                               'links': [       {       'connected_time': '17036075',
                                                                                                                        'fiber_length': 6815,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg5NXBlZXJwb25wb24gMjY1MDEz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "7022 ROAD 129 (THIESSEN BROTHER'S "
                                                 'CUSTOM WOOD)',
                                         'poller_hash': 'MTI3NDg5NQ==',
                                         'serial_number': '1274895',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:10:a9',
                                         'uptime': '17036075'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '265033/g1',
                                                                       'name': '265033/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzQxMXBlZXJpbnRmMjY1MDMzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '265033',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzQxMXBlZXJpbnRmcG9uIDI2NTAzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265033',
                                                                                               'links': [       {       'connected_time': '17631129',
                                                                                                                        'fiber_length': 11051,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQxMXBlZXJwb25wb24gMjY1MDMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5542 ROAD 126 (GASCHO KEITH & '
                                                 'ANNETTE)',
                                         'poller_hash': 'MTI5NzQxMQ==',
                                         'serial_number': '1297411',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:04:6e',
                                         'uptime': '17631129'},
                                 {       'counter_source': '265647/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8a:30'],
                                                                       'id': '265647/g1',
                                                                       'in_nucast_pkts': '4585',
                                                                       'in_octets': '2566691564',
                                                                       'in_ucast_pkts': '11276635',
                                                                       'mac_address': '04:bc:9f:5c:0f:a4',
                                                                       'mtu': 2000,
                                                                       'name': '265647/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '541542',
                                                                       'out_octets': '34862678993',
                                                                       'out_ucast_pkts': '26565068',
                                                                       'poller_hash': 'MTI3NDg3NHBlZXJpbnRmMjY1NjQ3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '265647',
                                                                       'in_nucast_pkts': '541542',
                                                                       'in_octets': '34862678993',
                                                                       'in_ucast_pkts': '26565068',
                                                                       'links': [       {       'in_nucast_pkts': '541542',
                                                                                                'in_octets': '34862678993',
                                                                                                'in_ucast_pkts': '26565068',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4585',
                                                                                                'out_octets': '2566691564',
                                                                                                'out_ucast_pkts': '11276635',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4585',
                                                                       'out_octets': '2566691564',
                                                                       'out_ucast_pkts': '11276635',
                                                                       'poller_hash': 'MTI3NDg3NHBlZXJpbnRmcG9uIDI2NTY0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265647',
                                                                                               'links': [       {       'connected_time': '262226',
                                                                                                                        'fiber_length': 20065,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg3NHBlZXJwb25wb24gMjY1NjQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4820 ROAD 108 (SCHULER KEVIN)',
                                         'poller_hash': 'MTI3NDg3NA==',
                                         'serial_number': '1274874',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0f:a1',
                                         'uptime': '262226'},
                                 {       'counter_source': '265697/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cd:3b'],
                                                                       'id': '265697/g1',
                                                                       'in_nucast_pkts': '308643',
                                                                       'in_octets': '35578357865',
                                                                       'in_ucast_pkts': '113373992',
                                                                       'mac_address': '04:bc:9f:5c:05:14',
                                                                       'mtu': 2000,
                                                                       'name': '265697/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39970992',
                                                                       'out_octets': '524069379735',
                                                                       'out_ucast_pkts': '441328923',
                                                                       'poller_hash': 'MTI3NDcyMnBlZXJpbnRmMjY1Njk3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '265697',
                                                                       'in_nucast_pkts': '39970992',
                                                                       'in_octets': '524069379735',
                                                                       'in_ucast_pkts': '441328923',
                                                                       'links': [       {       'in_nucast_pkts': '39970992',
                                                                                                'in_octets': '524069379735',
                                                                                                'in_ucast_pkts': '441328923',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '308643',
                                                                                                'out_octets': '35578357865',
                                                                                                'out_ucast_pkts': '113373992',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '308643',
                                                                       'out_octets': '35578357865',
                                                                       'out_ucast_pkts': '113373992',
                                                                       'poller_hash': 'MTI3NDcyMnBlZXJpbnRmcG9uIDI2NTY5Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265697',
                                                                                               'links': [       {       'connected_time': '17471855',
                                                                                                                        'fiber_length': 19400,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDcyMnBlZXJwb25wb24gMjY1Njk3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4056 LINE 46 (ROES MELVIN)',
                                         'poller_hash': 'MTI3NDcyMg==',
                                         'serial_number': '1274722',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:05:11',
                                         'uptime': '17471855'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '265707/g1',
                                                                       'name': '265707/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDVGQ3BlZXJpbnRmMjY1NzA3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '265707',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDVGQ3BlZXJpbnRmcG9uIDI2NTcwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265707',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDVGQ3BlZXJwb25wb24gMjY1NzA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4460B PERTH LINE 72',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "265707" '
                                                                  '(location: GPON '
                                                                  'port "15/2"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-06-26T09:43:38.880Z',
                                         'poller_hash': 'MTI3NDVGQw==',
                                         'serial_number': '12745FC',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '265717/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '265717/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '40610898747',
                                                                       'in_ucast_pkts': '61629158',
                                                                       'mac_address': '04:bc:9f:5c:00:09',
                                                                       'name': '265717/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '178241700862',
                                                                       'out_ucast_pkts': '163434892',
                                                                       'poller_hash': 'MTI3NDY4MXBlZXJpbnRmMjY1NzE3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '265717',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '178241700862',
                                                                       'in_ucast_pkts': '163434892',
                                                                       'links': [       {       'in_nucast_pkts': '16',
                                                                                                'in_octets': '178241700862',
                                                                                                'in_ucast_pkts': '163434892',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '40610898747',
                                                                                                'out_ucast_pkts': '61629158',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '40610898747',
                                                                       'out_ucast_pkts': '61629158',
                                                                       'poller_hash': 'MTI3NDY4MXBlZXJpbnRmcG9uIDI2NTcxNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265717',
                                                                                               'links': [       {       'connected_time': '523342',
                                                                                                                        'fiber_length': 6613,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDY4MXBlZXJwb25wb24gMjY1NzE3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4460B PERTH LINE 72 (PETERS WILHELM '
                                                 '& ELIZABETH)',
                                         'poller_hash': 'MTI3NDY4MQ==',
                                         'serial_number': '1274681',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:00:09',
                                         'uptime': '523342'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266098/g1',
                                                                       'name': '266098/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDY5Q3BlZXJpbnRmMjY2MDk4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266098',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDY5Q3BlZXJpbnRmcG9uIDI2NjA5OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266098',
                                                                                               'links': [       {       'connected_time': '17302188',
                                                                                                                        'fiber_length': 15715,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDY5Q3BlZXJwb25wb24gMjY2MDk4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7477 PERTH ROAD 121 (SAUDER ORVIE)',
                                         'poller_hash': 'MTI3NDY5Qw==',
                                         'serial_number': '127469C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:00:e1',
                                         'uptime': '17302188'},
                                 {       'counter_source': '266107/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7e:33'],
                                                                       'id': '266107/g1',
                                                                       'in_nucast_pkts': '9557',
                                                                       'in_octets': '11427437360',
                                                                       'in_ucast_pkts': '55988585',
                                                                       'mac_address': '04:bc:9f:5c:02:6c',
                                                                       'mtu': 2000,
                                                                       'name': '266107/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1151496',
                                                                       'out_octets': '296804632211',
                                                                       'out_ucast_pkts': '212321408',
                                                                       'poller_hash': 'MTI3NDZDRHBlZXJpbnRmMjY2MTA3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '266107',
                                                                       'in_nucast_pkts': '1151496',
                                                                       'in_octets': '296804632211',
                                                                       'in_ucast_pkts': '212321408',
                                                                       'links': [       {       'in_nucast_pkts': '1151496',
                                                                                                'in_octets': '296804632211',
                                                                                                'in_ucast_pkts': '212321408',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9557',
                                                                                                'out_octets': '11427437360',
                                                                                                'out_ucast_pkts': '55988585',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9557',
                                                                       'out_octets': '11427437360',
                                                                       'out_ucast_pkts': '55988585',
                                                                       'poller_hash': 'MTI3NDZDRHBlZXJpbnRmcG9uIDI2NjEwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266107',
                                                                                               'links': [       {       'connected_time': '538377',
                                                                                                                        'fiber_length': 25365,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDZDRHBlZXJwb25wb24gMjY2MTA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4578 LINE 46 (PETERSON MONTGOMERY)',
                                         'poller_hash': 'MTI3NDZDRA==',
                                         'serial_number': '12746CD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:02:69',
                                         'uptime': '538377'},
                                 {       'counter_source': '266310/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '266310/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '95544726483',
                                                                       'in_ucast_pkts': '179516480',
                                                                       'mac_address': '04:bc:9f:5b:f1:91',
                                                                       'name': '266310/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '976682089103',
                                                                       'out_ucast_pkts': '726155109',
                                                                       'poller_hash': 'MTI3NDRCMnBlZXJpbnRmMjY2MzEwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266310',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '976682089103',
                                                                       'in_ucast_pkts': '726155109',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '976682089103',
                                                                                                'in_ucast_pkts': '726155109',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '95544726483',
                                                                                                'out_ucast_pkts': '179516480',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '95544726483',
                                                                       'out_ucast_pkts': '179516480',
                                                                       'poller_hash': 'MTI3NDRCMnBlZXJpbnRmcG9uIDI2NjMxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266310',
                                                                                               'links': [       {       'connected_time': '262226',
                                                                                                                        'fiber_length': 19850,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDRCMnBlZXJwb25wb24gMjY2MzEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2339 LINE 45 (HART DAWN)',
                                         'poller_hash': 'MTI3NDRCMg==',
                                         'serial_number': '12744B2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:f1:91',
                                         'uptime': '262226'},
                                 {       'counter_source': '266330/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '266330/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15961784857',
                                                                       'in_ucast_pkts': '28934145',
                                                                       'mac_address': '04:bc:9f:6f:03:66',
                                                                       'name': '266330/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '74653534510',
                                                                       'out_ucast_pkts': '75092596',
                                                                       'poller_hash': 'MTI5NzNGMHBlZXJpbnRmMjY2MzMwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266330',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '74653534510',
                                                                       'in_ucast_pkts': '75092596',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '74653534510',
                                                                                                'in_ucast_pkts': '75092596',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '15961784857',
                                                                                                'out_ucast_pkts': '28934145',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15961784857',
                                                                       'out_ucast_pkts': '28934145',
                                                                       'poller_hash': 'MTI5NzNGMHBlZXJpbnRmcG9uIDI2NjMzMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266330',
                                                                                               'links': [       {       'connected_time': '523099',
                                                                                                                        'fiber_length': 11041,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNGMHBlZXJwb25wb24gMjY2MzMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4085 PRINCESS ST (BARTEL ALLISON)',
                                         'poller_hash': 'MTI5NzNGMA==',
                                         'serial_number': '12973F0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:03:66',
                                         'uptime': '523099'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266350/g1',
                                                                       'name': '266350/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTFCRTMwMXBlZXJpbnRmMjY2MzUwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266350',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTFCRTMwMXBlZXJpbnRmcG9uIDI2NjM1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266350',
                                                                                               'links': [       {       'connected_time': '688580',
                                                                                                                        'fiber_length': 5221,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTMwMXBlZXJwb25wb24gMjY2MzUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6841 ROAD 129 (REIMER ABRAM)',
                                         'poller_hash': 'MTFCRTMwMQ==',
                                         'serial_number': '11BE301',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:65:7c',
                                         'uptime': '688580'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266404/g1',
                                                                       'name': '266404/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE5MzAwQnBlZXJpbnRmMjY2NDA0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266404',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE5MzAwQnBlZXJpbnRmcG9uIDI2NjQwNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266404',
                                                                                               'links': [       {       'connected_time': '16866847',
                                                                                                                        'fiber_length': 5403,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MzAwQnBlZXJwb25wb24gMjY2NDA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6529 ROAD 129 (KUEPFER EDWARD)',
                                         'poller_hash': 'MTE5MzAwQg==',
                                         'serial_number': '119300B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:d7:b5',
                                         'uptime': '16866847'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266434/g1',
                                                                       'name': '266434/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE5MkY5QnBlZXJpbnRmMjY2NDM0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266434',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE5MkY5QnBlZXJpbnRmcG9uIDI2NjQzNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266434',
                                                                                               'links': [       {       'connected_time': '16847561',
                                                                                                                        'fiber_length': 7123,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MkY5QnBlZXJwb25wb24gMjY2NDM0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5961 PERTH ROAD 119 (GASCHO ANDRE)',
                                         'poller_hash': 'MTE5MkY5Qg==',
                                         'serial_number': '1192F9B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:d4:35',
                                         'uptime': '16847561'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266484/g1',
                                                                       'name': '266484/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE5MkZGNnBlZXJpbnRmMjY2NDg0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266484',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NjgyZ3BvbiAxNC83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE5MkZGNnBlZXJpbnRmcG9uIDI2NjQ4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266484',
                                                                                               'links': [       {       'connected_time': '16695529',
                                                                                                                        'fiber_length': 11640,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NjgycG9uZ3BvbiAxNC83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MkZGNnBlZXJwb25wb24gMjY2NDg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7002 PERTH ROAD 121 (KUEPFER JEREMY)',
                                         'poller_hash': 'MTE5MkZGNg==',
                                         'serial_number': '1192FF6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:d7:0d',
                                         'uptime': '16695529'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266494/g1',
                                                                       'name': '266494/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE5MkZGNXBlZXJpbnRmMjY2NDk0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266494',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE5MkZGNXBlZXJpbnRmcG9uIDI2NjQ5NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266494',
                                                                                               'links': [       {       'connected_time': '16696858',
                                                                                                                        'fiber_length': 6498,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MkZGNXBlZXJwb25wb24gMjY2NDk0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6005 PERTH ROAD 119 (GASCHO BRIAN)',
                                         'poller_hash': 'MTE5MkZGNQ==',
                                         'serial_number': '1192FF5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:d7:05',
                                         'uptime': '16696858'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '266504/g1',
                                                                       'name': '266504/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTE5MzA0MnBlZXJpbnRmMjY2NTA0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '266504',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTE5MzA0MnBlZXJpbnRmcG9uIDI2NjUwNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266504',
                                                                                               'links': [       {       'connected_time': '16693433',
                                                                                                                        'fiber_length': 6375,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MzA0MnBlZXJwb25wb24gMjY2NTA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6013 PERTH ROAD 119 (GEERTSMA PETER)',
                                         'poller_hash': 'MTE5MzA0Mg==',
                                         'serial_number': '1193042',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:d9:6d',
                                         'uptime': '16693433'},
                                 {       'counter_source': '267003/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8a:51'],
                                                                       'id': '267003/g1',
                                                                       'in_nucast_pkts': '14822',
                                                                       'in_octets': '3928597351',
                                                                       'in_ucast_pkts': '15999438',
                                                                       'mac_address': 'b8:94:70:ce:68:bf',
                                                                       'mtu': 2000,
                                                                       'name': '267003/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465364',
                                                                       'out_octets': '78778206726',
                                                                       'out_ucast_pkts': '55199487',
                                                                       'poller_hash': 'MTFCRTM2OXBlZXJpbnRmMjY3MDAzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '267003',
                                                                       'in_nucast_pkts': '1465364',
                                                                       'in_octets': '78778206726',
                                                                       'in_ucast_pkts': '55199487',
                                                                       'links': [       {       'in_nucast_pkts': '1465364',
                                                                                                'in_octets': '78778206726',
                                                                                                'in_ucast_pkts': '55199487',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14822',
                                                                                                'out_octets': '3928597351',
                                                                                                'out_ucast_pkts': '15999438',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14822',
                                                                       'out_octets': '3928597351',
                                                                       'out_ucast_pkts': '15999438',
                                                                       'poller_hash': 'MTFCRTM2OXBlZXJpbnRmcG9uIDI2NzAwMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '267003',
                                                                                               'links': [       {       'connected_time': '688607',
                                                                                                                        'fiber_length': 6621,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTM2OXBlZXJwb25wb24gMjY3MDAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4436 PERTH LINE 72 (KUEPFER RAYMOND)',
                                         'poller_hash': 'MTFCRTM2OQ==',
                                         'serial_number': '11BE369',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:68:bc',
                                         'uptime': '688607'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '272039/g1',
                                                                       'name': '272039/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NzNCMHBlZXJpbnRmMjcyMDM5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '272039',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NzNCMHBlZXJpbnRmcG9uIDI3MjAzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272039',
                                                                                               'links': [       {       'connected_time': '16333063',
                                                                                                                        'fiber_length': 19984,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzNCMHBlZXJwb25wb24gMjcyMDM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2460 LINE 45 (GERBER TIMOTHY)',
                                         'poller_hash': 'MTI5NzNCMA==',
                                         'serial_number': '12973B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:01:66',
                                         'uptime': '16333063'},
                                 {       'counter_source': '272129/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:ed:66'],
                                                                       'id': '272129/g1',
                                                                       'in_nucast_pkts': '148',
                                                                       'in_octets': '191902925754',
                                                                       'in_ucast_pkts': '455470187',
                                                                       'mac_address': '04:bc:9f:5c:0f:4c',
                                                                       'mtu': 2000,
                                                                       'name': '272129/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36317922',
                                                                       'out_octets': '1381553424933',
                                                                       'out_ucast_pkts': '1144765405',
                                                                       'poller_hash': 'MTI3NDg2OXBlZXJpbnRmMjcyMTI5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272129',
                                                                       'in_nucast_pkts': '36317922',
                                                                       'in_octets': '1381553424933',
                                                                       'in_ucast_pkts': '1144765405',
                                                                       'links': [       {       'in_nucast_pkts': '36317922',
                                                                                                'in_octets': '1381553424933',
                                                                                                'in_ucast_pkts': '1144765405',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '148',
                                                                                                'out_octets': '191902925754',
                                                                                                'out_ucast_pkts': '455470187',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '148',
                                                                       'out_octets': '191902925754',
                                                                       'out_ucast_pkts': '455470187',
                                                                       'poller_hash': 'MTI3NDg2OXBlZXJpbnRmcG9uIDI3MjEyOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272129',
                                                                                               'links': [       {       'connected_time': '16160357',
                                                                                                                        'fiber_length': 1421,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg2OXBlZXJwb25wb24gMjcyMTI5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 SPENCER ST (AW MILLWRIGHTS)',
                                         'poller_hash': 'MTI3NDg2OQ==',
                                         'serial_number': '1274869',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0f:49',
                                         'uptime': '16160357'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '272179/g1',
                                                                       'name': '272179/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDc4RnBlZXJpbnRmMjcyMTc5L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '272179',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDc4RnBlZXJpbnRmcG9uIDI3MjE3OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272179',
                                                                                               'links': [       {       'connected_time': '16071943',
                                                                                                                        'fiber_length': 10197,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc4RnBlZXJwb25wb24gMjcyMTc5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6477A PERTH ROAD 121 (KUEPFER '
                                                 'CLAYTON J AND SHARON)',
                                         'poller_hash': 'MTI3NDc4Rg==',
                                         'serial_number': '127478F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:08:79',
                                         'uptime': '16071943'},
                                 {       'counter_source': '272229/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:58:85'],
                                                                       'id': '272229/g1',
                                                                       'in_nucast_pkts': '14500',
                                                                       'in_octets': '9592759743',
                                                                       'in_ucast_pkts': '75891941',
                                                                       'mac_address': '04:bc:9f:5c:09:0c',
                                                                       'mtu': 2000,
                                                                       'name': '272229/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465411',
                                                                       'out_octets': '392652009078',
                                                                       'out_ucast_pkts': '290165301',
                                                                       'poller_hash': 'MTI3NDdBMXBlZXJpbnRmMjcyMjI5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272229',
                                                                       'in_nucast_pkts': '1465411',
                                                                       'in_octets': '392652009078',
                                                                       'in_ucast_pkts': '290165301',
                                                                       'links': [       {       'in_nucast_pkts': '1465411',
                                                                                                'in_octets': '392652009078',
                                                                                                'in_ucast_pkts': '290165301',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14500',
                                                                                                'out_octets': '9592759743',
                                                                                                'out_ucast_pkts': '75891941',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14500',
                                                                       'out_octets': '9592759743',
                                                                       'out_ucast_pkts': '75891941',
                                                                       'poller_hash': 'MTI3NDdBMXBlZXJpbnRmcG9uIDI3MjIyOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272229',
                                                                                               'links': [       {       'connected_time': '688362',
                                                                                                                        'fiber_length': 1106,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdBMXBlZXJwb25wb24gMjcyMjI5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '34 MAIN ST S MILVERTON (FOSTER ANNE)',
                                         'poller_hash': 'MTI3NDdBMQ==',
                                         'serial_number': '12747A1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:09:09',
                                         'uptime': '688362'},
                                 {       'counter_source': '272272/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:57:b5:0e'],
                                                                       'id': '272272/g1',
                                                                       'in_nucast_pkts': '5196',
                                                                       'in_octets': '4835430372',
                                                                       'in_ucast_pkts': '13887485',
                                                                       'mac_address': '04:bc:9f:5c:0a:0c',
                                                                       'mtu': 2000,
                                                                       'name': '272272/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '541539',
                                                                       'out_octets': '95566663649',
                                                                       'out_ucast_pkts': '71523342',
                                                                       'poller_hash': 'MTI3NDdDMXBlZXJpbnRmMjcyMjcyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272272',
                                                                       'in_nucast_pkts': '541539',
                                                                       'in_octets': '95566663649',
                                                                       'in_ucast_pkts': '71523342',
                                                                       'links': [       {       'in_nucast_pkts': '541539',
                                                                                                'in_octets': '95566663649',
                                                                                                'in_ucast_pkts': '71523342',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5196',
                                                                                                'out_octets': '4835430372',
                                                                                                'out_ucast_pkts': '13887485',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5196',
                                                                       'out_octets': '4835430372',
                                                                       'out_ucast_pkts': '13887485',
                                                                       'poller_hash': 'MTI3NDdDMXBlZXJpbnRmcG9uIDI3MjI3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272272',
                                                                                               'links': [       {       'connected_time': '262228',
                                                                                                                        'fiber_length': 22256,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdDMXBlZXJwb25wb24gMjcyMjcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5001 ROAD 109 (ROSE JOANNE)',
                                         'poller_hash': 'MTI3NDdDMQ==',
                                         'serial_number': '12747C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0a:09',
                                         'uptime': '262228'},
                                 {       'counter_source': '272294/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '272294/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '39882971944',
                                                                       'in_ucast_pkts': '190391384',
                                                                       'mac_address': 'cc:be:59:b7:3c:cf',
                                                                       'name': '272294/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '940788408198',
                                                                       'out_ucast_pkts': '701824884',
                                                                       'poller_hash': 'NEQxMTRDcGVlcmludGYyNzIyOTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '272294',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '940788408198',
                                                                       'in_ucast_pkts': '701824884',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '940788408198',
                                                                                                'in_ucast_pkts': '701824884',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '39882971944',
                                                                                                'out_ucast_pkts': '190391384',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '39882971944',
                                                                       'out_ucast_pkts': '190391384',
                                                                       'poller_hash': 'NEQxMTRDcGVlcmludGZwb24gMjcyMjk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272294',
                                                                                               'links': [       {       'connected_time': '15813850',
                                                                                                                        'fiber_length': 17629,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.516,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQxMTRDcGVlcnBvbnBvbiAyNzIyOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '165 WINSTANLEY ST (KLOOSTRA ALEX)',
                                         'poller_hash': 'NEQxMTRD',
                                         'serial_number': '4D114C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:b7:3c:cf',
                                         'uptime': '15813850'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '272430/g1',
                                                                       'name': '272430/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTU5RjdGRXBlZXJpbnRmMjcyNDMwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '272430',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTU5RjdGRXBlZXJpbnRmcG9uIDI3MjQzMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272430',
                                                                                               'links': [       {       'connected_time': '7677066',
                                                                                                                        'fiber_length': 8563,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.466,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU5RjdGRXBlZXJwb25wb24gMjcyNDMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '6780 PERTH ROAD 121 (ALBRECHT ALVIN '
                                                 'K)',
                                         'poller_hash': 'MTU5RjdGRQ==',
                                         'serial_number': '159F7FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '1c:8b:76:66:6b:79',
                                         'uptime': '7677066'},
                                 {       'counter_source': '272557/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '272557/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '69931149231',
                                                                       'in_ucast_pkts': '353607186',
                                                                       'mac_address': '04:bc:9f:5c:0f:89',
                                                                       'name': '272557/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1332045539730',
                                                                       'out_ucast_pkts': '978914800',
                                                                       'poller_hash': 'MTI3NDg3MXBlZXJpbnRmMjcyNTU3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '272557',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1332045539730',
                                                                       'in_ucast_pkts': '978914800',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1332045539730',
                                                                                                'in_ucast_pkts': '978914800',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '69931149231',
                                                                                                'out_ucast_pkts': '353607186',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '69931149231',
                                                                       'out_ucast_pkts': '353607186',
                                                                       'poller_hash': 'MTI3NDg3MXBlZXJpbnRmcG9uIDI3MjU1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272557',
                                                                                               'links': [       {       'connected_time': '262228',
                                                                                                                        'fiber_length': 21896,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg3MXBlZXJwb25wb24gMjcyNTU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4852 ROAD 109 (COOK CHRIS)',
                                         'poller_hash': 'MTI3NDg3MQ==',
                                         'serial_number': '1274871',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0f:89',
                                         'uptime': '262228'},
                                 {       'counter_source': '272577/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '272577/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '28775729668',
                                                                       'in_ucast_pkts': '257200647',
                                                                       'mac_address': '04:bc:9f:5c:11:09',
                                                                       'name': '272577/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1987845766194',
                                                                       'out_ucast_pkts': '1354257246',
                                                                       'poller_hash': 'MTI3NDhBMXBlZXJpbnRmMjcyNTc3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '272577',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1987845766194',
                                                                       'in_ucast_pkts': '1354257246',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1987845766194',
                                                                                                'in_ucast_pkts': '1354257246',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '28775729668',
                                                                                                'out_ucast_pkts': '257200647',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '28775729668',
                                                                       'out_ucast_pkts': '257200647',
                                                                       'poller_hash': 'MTI3NDhBMXBlZXJpbnRmcG9uIDI3MjU3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272577',
                                                                                               'links': [       {       'connected_time': '15479602',
                                                                                                                        'fiber_length': 18809,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDhBMXBlZXJwb25wb24gMjcyNTc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4943 ROAD 108 (COOK FRANCES)',
                                         'poller_hash': 'MTI3NDhBMQ==',
                                         'serial_number': '12748A1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:11:09',
                                         'uptime': '15479602'},
                                 {       'counter_source': '272831/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '272831/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '19072652694',
                                                                       'in_ucast_pkts': '56542254',
                                                                       'mac_address': '04:bc:9f:5c:0c:01',
                                                                       'name': '272831/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '208230324202',
                                                                       'out_ucast_pkts': '178419057',
                                                                       'poller_hash': 'MTI3NDgwMHBlZXJpbnRmMjcyODMxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '272831',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '208230324202',
                                                                       'in_ucast_pkts': '178419057',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '208230324202',
                                                                                                'in_ucast_pkts': '178419057',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '19072652694',
                                                                                                'out_ucast_pkts': '56542254',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '19072652694',
                                                                       'out_ucast_pkts': '56542254',
                                                                       'poller_hash': 'MTI3NDgwMHBlZXJpbnRmcG9uIDI3MjgzMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272831',
                                                                                               'links': [       {       'connected_time': '15470695',
                                                                                                                        'fiber_length': 6205,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDgwMHBlZXJwb25wb24gMjcyODMx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6023 PERTH ROAD 119 (ALBRECHT '
                                                 'ARNOLD)',
                                         'poller_hash': 'MTI3NDgwMA==',
                                         'serial_number': '1274800',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0c:01',
                                         'uptime': '15470695'},
                                 {       'counter_source': '273513/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '273513/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '66518501041',
                                                                       'in_ucast_pkts': '180567147',
                                                                       'mac_address': '04:bc:9f:5b:f8:79',
                                                                       'name': '273513/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '312194164772',
                                                                       'out_ucast_pkts': '320928210',
                                                                       'poller_hash': 'MTI3NDU4RnBlZXJpbnRmMjczNTEzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '273513',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '312194164772',
                                                                       'in_ucast_pkts': '320928210',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '312194164772',
                                                                                                'in_ucast_pkts': '320928210',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '66518501041',
                                                                                                'out_ucast_pkts': '180567147',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '66518501041',
                                                                       'out_ucast_pkts': '180567147',
                                                                       'poller_hash': 'MTI3NDU4RnBlZXJpbnRmcG9uIDI3MzUxMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '273513',
                                                                                               'links': [       {       'connected_time': '15219615',
                                                                                                                        'fiber_length': 6948,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDU4RnBlZXJwb25wb24gMjczNTEz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5083 LINE 60 (GINGRICH TONY)',
                                         'poller_hash': 'MTI3NDU4Rg==',
                                         'serial_number': '127458F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:f8:79',
                                         'uptime': '15219615'},
                                 {       'counter_source': '273572/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '273572/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '32342715359',
                                                                       'in_ucast_pkts': '59747395',
                                                                       'mac_address': '04:bc:9f:5c:0e:e9',
                                                                       'name': '273572/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '142409216355',
                                                                       'out_ucast_pkts': '141396110',
                                                                       'poller_hash': 'MTI3NDg1RHBlZXJpbnRmMjczNTcyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '273572',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '142409216355',
                                                                       'in_ucast_pkts': '141396110',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '142409216355',
                                                                                                'in_ucast_pkts': '141396110',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '32342715359',
                                                                                                'out_ucast_pkts': '59747395',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '32342715359',
                                                                       'out_ucast_pkts': '59747395',
                                                                       'poller_hash': 'MTI3NDg1RHBlZXJpbnRmcG9uIDI3MzU3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '273572',
                                                                                               'links': [       {       'connected_time': '15211345',
                                                                                                                        'fiber_length': 4681,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg1RHBlZXJwb25wb24gMjczNTcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6800 ROAD 129 (KUEPFER CARL A)',
                                         'poller_hash': 'MTI3NDg1RA==',
                                         'serial_number': '127485D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0e:e9',
                                         'uptime': '15211345'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '273881/g1',
                                                                       'name': '273881/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDgxRXBlZXJpbnRmMjczODgxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '273881',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDgxRXBlZXJpbnRmcG9uIDI3Mzg4MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '273881',
                                                                                               'links': [       {       'connected_time': '4587986',
                                                                                                                        'fiber_length': 1503,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDgxRXBlZXJwb25wb24gMjczODgx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '17 PIERSON ST (MORNINGTON '
                                                 'COMMUNICATIONS)',
                                         'poller_hash': 'MTI3NDgxRQ==',
                                         'serial_number': '127481E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0c:f1',
                                         'uptime': '4587986'},
                                 {       'counter_source': '273901/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '273901/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '98559074013',
                                                                       'in_ucast_pkts': '587784435',
                                                                       'mac_address': '04:bc:9f:5c:10:99',
                                                                       'name': '273901/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2754079757779',
                                                                       'out_ucast_pkts': '1923240785',
                                                                       'poller_hash': 'MTI3NDg5M3BlZXJpbnRmMjczOTAxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '273901',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2754079757779',
                                                                       'in_ucast_pkts': '1923240785',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2754079757779',
                                                                                                'in_ucast_pkts': '1923240785',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '98559074013',
                                                                                                'out_ucast_pkts': '587784435',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '98559074013',
                                                                       'out_ucast_pkts': '587784435',
                                                                       'poller_hash': 'MTI3NDg5M3BlZXJpbnRmcG9uIDI3MzkwMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '273901',
                                                                                               'links': [       {       'connected_time': '688363',
                                                                                                                        'fiber_length': 2734,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg5M3BlZXJwb25wb24gMjczOTAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '141 PUGH ST (HULLEY CRYSTAL)',
                                         'poller_hash': 'MTI3NDg5Mw==',
                                         'serial_number': '1274893',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:10:99',
                                         'uptime': '688363'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '274031/g1',
                                                                       'name': '274031/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEEyNXBlZXJpbnRmMjc0MDMxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274031',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEEyNXBlZXJpbnRmcG9uIDI3NDAzMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274031',
                                                                                               'links': [       {       'connected_time': '13931908',
                                                                                                                        'fiber_length': 6621,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEEyNXBlZXJwb25wb24gMjc0MDMx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7019 ROAD 129 (STREICHER CARL)',
                                         'poller_hash': 'MTI3NEEyNQ==',
                                         'serial_number': '1274A25',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1d:29',
                                         'uptime': '13931908'},
                                 {       'counter_source': '274051/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:39:8f:7e:b0:30'],
                                                                       'id': '274051/g1',
                                                                       'in_nucast_pkts': '18569',
                                                                       'in_octets': '4136840327',
                                                                       'in_ucast_pkts': '20909060',
                                                                       'mac_address': 'b8:94:70:ba:b3:b8',
                                                                       'mtu': 2000,
                                                                       'name': '274051/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30466431',
                                                                       'out_octets': '11569408346',
                                                                       'out_ucast_pkts': '11226543',
                                                                       'poller_hash': 'MTE5NEI4QnBlZXJpbnRmMjc0MDUxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274051',
                                                                       'in_nucast_pkts': '30466431',
                                                                       'in_octets': '11569408346',
                                                                       'in_ucast_pkts': '11226543',
                                                                       'links': [       {       'in_nucast_pkts': '30466431',
                                                                                                'in_octets': '11569408346',
                                                                                                'in_ucast_pkts': '11226543',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18569',
                                                                                                'out_octets': '4136840327',
                                                                                                'out_ucast_pkts': '20909060',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18569',
                                                                       'out_octets': '4136840327',
                                                                       'out_ucast_pkts': '20909060',
                                                                       'poller_hash': 'MTE5NEI4QnBlZXJpbnRmcG9uIDI3NDA1MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274051',
                                                                                               'links': [       {       'connected_time': '13912989',
                                                                                                                        'fiber_length': 1648,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NEI4QnBlZXJwb25wb24gMjc0MDUx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 COBALT ST N (AVON MAITLAND '
                                                 'DISTRICT SCHOOL)',
                                         'poller_hash': 'MTE5NEI4Qg==',
                                         'serial_number': '1194B8B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:b3:b5',
                                         'uptime': '13912989'},
                                 {       'counter_source': '274376/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274376/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '796218757',
                                                                       'in_ucast_pkts': '3266865',
                                                                       'mac_address': '04:bc:9f:5c:08:c1',
                                                                       'name': '274376/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7398142548',
                                                                       'out_ucast_pkts': '16778318',
                                                                       'poller_hash': 'MTI3NDc5OHBlZXJpbnRmMjc0Mzc2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274376',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7398142548',
                                                                       'in_ucast_pkts': '16778318',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7398142548',
                                                                                                'in_ucast_pkts': '16778318',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '796218757',
                                                                                                'out_ucast_pkts': '3266865',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '796218757',
                                                                       'out_ucast_pkts': '3266865',
                                                                       'poller_hash': 'MTI3NDc5OHBlZXJpbnRmcG9uIDI3NDM3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274376',
                                                                                               'links': [       {       'connected_time': '688644',
                                                                                                                        'fiber_length': 8438,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc5OHBlZXJwb25wb24gMjc0Mzc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6752 PERTH ROAD 121 (NITHVALLEY '
                                                 'FARMS)',
                                         'poller_hash': 'MTI3NDc5OA==',
                                         'serial_number': '1274798',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:08:c1',
                                         'uptime': '688644'},
                                 {       'counter_source': '274386/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274386/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '15510286359',
                                                                       'in_ucast_pkts': '35786330',
                                                                       'mac_address': '04:bc:9f:5c:0d:49',
                                                                       'name': '274386/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '120409960686',
                                                                       'out_ucast_pkts': '116169530',
                                                                       'poller_hash': 'MTI3NDgyOXBlZXJpbnRmMjc0Mzg2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274386',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '120409960686',
                                                                       'in_ucast_pkts': '116169530',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '120409960686',
                                                                                                'in_ucast_pkts': '116169530',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '15510286359',
                                                                                                'out_ucast_pkts': '35786330',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '15510286359',
                                                                       'out_ucast_pkts': '35786330',
                                                                       'poller_hash': 'MTI3NDgyOXBlZXJpbnRmcG9uIDI3NDM4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274386',
                                                                                               'links': [       {       'connected_time': '13405471',
                                                                                                                        'fiber_length': 11794,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDgyOXBlZXJwb25wb24gMjc0Mzg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6908 PERTH ROAD 121 (DJ KUEPFER & '
                                                 'FAMILY INC)',
                                         'poller_hash': 'MTI3NDgyOQ==',
                                         'serial_number': '1274829',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0d:49',
                                         'uptime': '13405471'},
                                 {       'counter_source': '274396/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:d2:79:8e'],
                                                                       'id': '274396/g1',
                                                                       'in_nucast_pkts': '11979',
                                                                       'in_octets': '7334671967',
                                                                       'in_ucast_pkts': '19306036',
                                                                       'mac_address': '04:bc:9f:5c:0f:14',
                                                                       'mtu': 2000,
                                                                       'name': '274396/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465241',
                                                                       'out_octets': '85041597432',
                                                                       'out_ucast_pkts': '70415645',
                                                                       'poller_hash': 'MTI3NDg2MnBlZXJpbnRmMjc0Mzk2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274396',
                                                                       'in_nucast_pkts': '1465241',
                                                                       'in_octets': '85041597432',
                                                                       'in_ucast_pkts': '70415645',
                                                                       'links': [       {       'in_nucast_pkts': '1465241',
                                                                                                'in_octets': '85041597432',
                                                                                                'in_ucast_pkts': '70415645',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11979',
                                                                                                'out_octets': '7334671967',
                                                                                                'out_ucast_pkts': '19306036',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11979',
                                                                       'out_octets': '7334671967',
                                                                       'out_ucast_pkts': '19306036',
                                                                       'poller_hash': 'MTI3NDg2MnBlZXJpbnRmcG9uIDI3NDM5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274396',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 3168,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg2MnBlZXJwb25wb24gMjc0Mzk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '38 AVERY PL (ALENCAR ISABEL)',
                                         'poller_hash': 'MTI3NDg2Mg==',
                                         'serial_number': '1274862',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0f:11',
                                         'uptime': '688370'},
                                 {       'counter_source': '274416/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ec:de'],
                                                                       'id': '274416/g1',
                                                                       'in_nucast_pkts': '15547',
                                                                       'in_octets': '7174213108',
                                                                       'in_ucast_pkts': '50551816',
                                                                       'mac_address': '04:bc:9f:5b:eb:24',
                                                                       'mtu': 2000,
                                                                       'name': '274416/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1499048',
                                                                       'out_octets': '175273632415',
                                                                       'out_ucast_pkts': '121398423',
                                                                       'poller_hash': 'MTI3NDNFNHBlZXJpbnRmMjc0NDE2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274416',
                                                                       'in_nucast_pkts': '1499048',
                                                                       'in_octets': '175273632415',
                                                                       'in_ucast_pkts': '121398423',
                                                                       'links': [       {       'in_nucast_pkts': '1499048',
                                                                                                'in_octets': '175273632415',
                                                                                                'in_ucast_pkts': '121398423',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15547',
                                                                                                'out_octets': '7174213108',
                                                                                                'out_ucast_pkts': '50551816',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15547',
                                                                       'out_octets': '7174213108',
                                                                       'out_ucast_pkts': '50551816',
                                                                       'poller_hash': 'MTI3NDNFNHBlZXJpbnRmcG9uIDI3NDQxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274416',
                                                                                               'links': [       {       'connected_time': '688341',
                                                                                                                        'fiber_length': 1348,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDNFNHBlZXJwb25wb24gMjc0NDE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 PIERSON ST (HENRY SHAWN)',
                                         'poller_hash': 'MTI3NDNFNA==',
                                         'serial_number': '12743E4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:eb:21',
                                         'uptime': '688341'},
                                 {       'counter_source': '274494/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:3e:b2:36'],
                                                                       'id': '274494/g1',
                                                                       'in_nucast_pkts': '60',
                                                                       'in_octets': '13410363296',
                                                                       'in_ucast_pkts': '30653618',
                                                                       'mac_address': '04:bc:9f:5c:0c:0c',
                                                                       'mtu': 2000,
                                                                       'name': '274494/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1123734',
                                                                       'out_octets': '99462387213',
                                                                       'out_ucast_pkts': '78667256',
                                                                       'poller_hash': 'MTI3NDgwMXBlZXJpbnRmMjc0NDk0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274494',
                                                                       'in_nucast_pkts': '1123734',
                                                                       'in_octets': '99462387213',
                                                                       'in_ucast_pkts': '78667256',
                                                                       'links': [       {       'in_nucast_pkts': '1123734',
                                                                                                'in_octets': '99462387213',
                                                                                                'in_ucast_pkts': '78667256',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '60',
                                                                                                'out_octets': '13410363296',
                                                                                                'out_ucast_pkts': '30653618',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '60',
                                                                       'out_octets': '13410363296',
                                                                       'out_ucast_pkts': '30653618',
                                                                       'poller_hash': 'MTI3NDgwMXBlZXJpbnRmcG9uIDI3NDQ5NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274494',
                                                                                               'links': [       {       'connected_time': '688348',
                                                                                                                        'fiber_length': 997,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDgwMXBlZXJwb25wb24gMjc0NDk0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 TEMPERANCE ST (BULMER STUART)',
                                         'poller_hash': 'MTI3NDgwMQ==',
                                         'serial_number': '1274801',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0c:09',
                                         'uptime': '688348'},
                                 {       'counter_source': '274544/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8c:82'],
                                                                       'id': '274544/g1',
                                                                       'in_nucast_pkts': '236933',
                                                                       'in_octets': '2692104145',
                                                                       'in_ucast_pkts': '7864267',
                                                                       'mac_address': '04:bc:9f:5c:08:54',
                                                                       'mtu': 2000,
                                                                       'name': '274544/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '28772475',
                                                                       'out_octets': '19760712839',
                                                                       'out_ucast_pkts': '18660570',
                                                                       'poller_hash': 'MTI3NDc4QXBlZXJpbnRmMjc0NTQ0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274544',
                                                                       'in_nucast_pkts': '28772475',
                                                                       'in_octets': '19760712839',
                                                                       'in_ucast_pkts': '18660570',
                                                                       'links': [       {       'in_nucast_pkts': '28772475',
                                                                                                'in_octets': '19760712839',
                                                                                                'in_ucast_pkts': '18660570',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '236933',
                                                                                                'out_octets': '2692104145',
                                                                                                'out_ucast_pkts': '7864267',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '236933',
                                                                       'out_octets': '2692104145',
                                                                       'out_ucast_pkts': '7864267',
                                                                       'poller_hash': 'MTI3NDc4QXBlZXJpbnRmcG9uIDI3NDU0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274544',
                                                                                               'links': [       {       'connected_time': '12804236',
                                                                                                                        'fiber_length': 1332,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc4QXBlZXJwb25wb24gMjc0NTQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 ARENA DR (NICHOLSON JOHN)',
                                         'poller_hash': 'MTI3NDc4QQ==',
                                         'serial_number': '127478A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:08:51',
                                         'uptime': '12804236'},
                                 {       'counter_source': '274596/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274596/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '38935389285',
                                                                       'in_ucast_pkts': '110165657',
                                                                       'mac_address': '04:bc:9f:5c:0b:f9',
                                                                       'name': '274596/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '421806032930',
                                                                       'out_ucast_pkts': '347926572',
                                                                       'poller_hash': 'MTI3NDdGRnBlZXJpbnRmMjc0NTk2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274596',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '421806032930',
                                                                       'in_ucast_pkts': '347926572',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '421806032930',
                                                                                                'in_ucast_pkts': '347926572',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '38935389285',
                                                                                                'out_ucast_pkts': '110165657',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '38935389285',
                                                                       'out_ucast_pkts': '110165657',
                                                                       'poller_hash': 'MTI3NDdGRnBlZXJpbnRmcG9uIDI3NDU5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274596',
                                                                                               'links': [       {       'connected_time': '12623764',
                                                                                                                        'fiber_length': 18067,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdGRnBlZXJwb25wb24gMjc0NTk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4513 LINE 49 (WAGLER MURRAY)',
                                         'poller_hash': 'MTI3NDdGRg==',
                                         'serial_number': '12747FF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0b:f9',
                                         'uptime': '12623764'},
                                 {       'counter_source': '274646/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274646/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '57193564731',
                                                                       'in_ucast_pkts': '227235849',
                                                                       'mac_address': '04:bc:9f:5c:09:31',
                                                                       'name': '274646/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '922816005341',
                                                                       'out_ucast_pkts': '705505821',
                                                                       'poller_hash': 'MTI3NDdBNnBlZXJpbnRmMjc0NjQ2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274646',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '922816005341',
                                                                       'in_ucast_pkts': '705505821',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '922816005341',
                                                                                                'in_ucast_pkts': '705505821',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '57193564731',
                                                                                                'out_ucast_pkts': '227235849',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '57193564731',
                                                                       'out_ucast_pkts': '227235849',
                                                                       'poller_hash': 'MTI3NDdBNnBlZXJpbnRmcG9uIDI3NDY0Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274646',
                                                                                               'links': [       {       'connected_time': '12443958',
                                                                                                                        'fiber_length': 988,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdBNnBlZXJwb25wb24gMjc0NjQ2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 CRUICKSHANK ST (TIGANI JAMES)',
                                         'poller_hash': 'MTI3NDdBNg==',
                                         'serial_number': '12747A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:09:31',
                                         'uptime': '12443958'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '274737/g1',
                                                                       'name': '274737/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDg2NXBlZXJpbnRmMjc0NzM3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274737',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDg2NXBlZXJpbnRmcG9uIDI3NDczNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274737',
                                                                                               'links': [       {       'connected_time': '8236946',
                                                                                                                        'fiber_length': 5998,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg2NXBlZXJwb25wb24gMjc0NzM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4300 LINE 61 (KUEPFER MERVIN)',
                                         'poller_hash': 'MTI3NDg2NQ==',
                                         'serial_number': '1274865',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0f:29',
                                         'uptime': '8236946'},
                                 {       'counter_source': '274747/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:39:8f:ed:41:5e'],
                                                                       'id': '274747/g1',
                                                                       'in_nucast_pkts': '958',
                                                                       'in_octets': '24619132911',
                                                                       'in_ucast_pkts': '76842851',
                                                                       'mac_address': '04:bc:9f:5c:06:24',
                                                                       'mtu': 2000,
                                                                       'name': '274747/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465249',
                                                                       'out_octets': '250852823793',
                                                                       'out_ucast_pkts': '199799491',
                                                                       'poller_hash': 'MTI3NDc0NHBlZXJpbnRmMjc0NzQ3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274747',
                                                                       'in_nucast_pkts': '1465249',
                                                                       'in_octets': '250852823793',
                                                                       'in_ucast_pkts': '199799491',
                                                                       'links': [       {       'in_nucast_pkts': '1465249',
                                                                                                'in_octets': '250852823793',
                                                                                                'in_ucast_pkts': '199799491',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '958',
                                                                                                'out_octets': '24619132911',
                                                                                                'out_ucast_pkts': '76842851',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '958',
                                                                       'out_octets': '24619132911',
                                                                       'out_ucast_pkts': '76842851',
                                                                       'poller_hash': 'MTI3NDc0NHBlZXJpbnRmcG9uIDI3NDc0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274747',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 3182,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc0NHBlZXJwb25wb24gMjc0NzQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '42 AVERY PL (VAN NIEROP DAVE)',
                                         'poller_hash': 'MTI3NDc0NA==',
                                         'serial_number': '1274744',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:06:21',
                                         'uptime': '688370'},
                                 {       'counter_source': '276482/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '276482/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '50829907764',
                                                                       'in_ucast_pkts': '374318656',
                                                                       'mac_address': 'b8:94:70:b9:dd:ad',
                                                                       'name': '276482/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1795952028331',
                                                                       'out_ucast_pkts': '1251365948',
                                                                       'poller_hash': 'MTE5MzBDQXBlZXJpbnRmMjc2NDgyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '276482',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1795952028331',
                                                                       'in_ucast_pkts': '1251365948',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1795952028331',
                                                                                                'in_ucast_pkts': '1251365948',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '50829907764',
                                                                                                'out_ucast_pkts': '374318656',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '50829907764',
                                                                       'out_ucast_pkts': '374318656',
                                                                       'poller_hash': 'MTE5MzBDQXBlZXJpbnRmcG9uIDI3NjQ4Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276482',
                                                                                               'links': [       {       'connected_time': '688354',
                                                                                                                        'fiber_length': 10132,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MzBDQXBlZXJwb25wb24gMjc2NDgy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5911 PERTH ROAD 119 (YANTZI AMY)',
                                         'poller_hash': 'MTE5MzBDQQ==',
                                         'serial_number': '11930CA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:dd:ad',
                                         'uptime': '688354'},
                                 {       'counter_source': '276492/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '276492/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '31945536187',
                                                                       'in_ucast_pkts': '88622810',
                                                                       'mac_address': '04:bc:9f:5c:0f:a9',
                                                                       'name': '276492/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '652386297551',
                                                                       'out_ucast_pkts': '481849515',
                                                                       'poller_hash': 'MTI3NDg3NXBlZXJpbnRmMjc2NDkyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '276492',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '652386297551',
                                                                       'in_ucast_pkts': '481849515',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '652386297551',
                                                                                                'in_ucast_pkts': '481849515',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '31945536187',
                                                                                                'out_ucast_pkts': '88622810',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '31945536187',
                                                                       'out_ucast_pkts': '88622810',
                                                                       'poller_hash': 'MTI3NDg3NXBlZXJpbnRmcG9uIDI3NjQ5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276492',
                                                                                               'links': [       {       'connected_time': '262468',
                                                                                                                        'fiber_length': 13715,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg3NXBlZXJwb25wb24gMjc2NDky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5445 ROAD 108 (LUCKHARDT SHELLEY)',
                                         'poller_hash': 'MTI3NDg3NQ==',
                                         'serial_number': '1274875',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0f:a9',
                                         'uptime': '262468'},
                                 {       'counter_source': '276649/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '276649/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '51675181528',
                                                                       'in_ucast_pkts': '154660088',
                                                                       'mac_address': '04:bc:9f:5a:38:71',
                                                                       'name': '276649/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '624319016570',
                                                                       'out_ucast_pkts': '467533153',
                                                                       'poller_hash': 'MTI3MEQ4RXBlZXJpbnRmMjc2NjQ5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '276649',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '624319016570',
                                                                       'in_ucast_pkts': '467533153',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '624319016570',
                                                                                                'in_ucast_pkts': '467533153',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '51675181528',
                                                                                                'out_ucast_pkts': '154660088',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '51675181528',
                                                                       'out_ucast_pkts': '154660088',
                                                                       'poller_hash': 'MTI3MEQ4RXBlZXJpbnRmcG9uIDI3NjY0OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276649',
                                                                                               'links': [       {       'connected_time': '688474',
                                                                                                                        'fiber_length': 616,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ4RXBlZXJwb25wb24gMjc2NjQ5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 MAIN ST N (COOPER JAMES)',
                                         'poller_hash': 'MTI3MEQ4RQ==',
                                         'serial_number': '1270D8E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:38:71',
                                         'uptime': '688474'},
                                 {       'counter_source': '276769/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f1:4d'],
                                                                       'id': '276769/g1',
                                                                       'in_nucast_pkts': '14222',
                                                                       'in_octets': '4260618485',
                                                                       'in_ucast_pkts': '26984000',
                                                                       'mac_address': '04:bc:9f:5a:3e:3c',
                                                                       'mtu': 2000,
                                                                       'name': '276769/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465414',
                                                                       'out_octets': '83792637439',
                                                                       'out_ucast_pkts': '62961692',
                                                                       'poller_hash': 'MTI3MEU0N3BlZXJpbnRmMjc2NzY5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '276769',
                                                                       'in_nucast_pkts': '1465414',
                                                                       'in_octets': '83792637439',
                                                                       'in_ucast_pkts': '62961692',
                                                                       'links': [       {       'in_nucast_pkts': '1465414',
                                                                                                'in_octets': '83792637439',
                                                                                                'in_ucast_pkts': '62961692',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14222',
                                                                                                'out_octets': '4260618485',
                                                                                                'out_ucast_pkts': '26984000',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14222',
                                                                       'out_octets': '4260618485',
                                                                       'out_ucast_pkts': '26984000',
                                                                       'poller_hash': 'MTI3MEU0N3BlZXJpbnRmcG9uIDI3Njc2OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276769',
                                                                                               'links': [       {       'connected_time': '688464',
                                                                                                                        'fiber_length': 1546,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEU0N3BlZXJwb25wb24gMjc2NzY5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '61 WILLIAM ST (ROTHWELL SARAH)',
                                         'poller_hash': 'MTI3MEU0Nw==',
                                         'serial_number': '1270E47',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:3e:39',
                                         'uptime': '688464'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '276844/g1',
                                                                       'name': '276844/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3MEUwNXBlZXJpbnRmMjc2ODQ0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '276844',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3MEUwNXBlZXJpbnRmcG9uIDI3Njg0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276844',
                                                                                               'links': [       {       'connected_time': '10305336',
                                                                                                                        'fiber_length': 3095,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEUwNXBlZXJwb25wb24gMjc2ODQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6484 ROAD 136 (COXON MRS LLOYD)',
                                         'poller_hash': 'MTI3MEUwNQ==',
                                         'serial_number': '1270E05',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:3c:29',
                                         'uptime': '10305336'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '276996/g1',
                                                                       'name': '276996/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3MEQ2QnBlZXJpbnRmMjc2OTk2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '276996',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3MEQ2QnBlZXJpbnRmcG9uIDI3Njk5Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276996',
                                                                                               'links': [       {       'connected_time': '10132691',
                                                                                                                        'fiber_length': 10196,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ2QnBlZXJwb25wb24gMjc2OTk2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5897 PERTH ROAD 119 (WAGLER SARAH)',
                                         'poller_hash': 'MTI3MEQ2Qg==',
                                         'serial_number': '1270D6B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:37:59',
                                         'uptime': '10132691'},
                                 {       'counter_source': '277036/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '277036/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '33245879400',
                                                                       'in_ucast_pkts': '192362671',
                                                                       'mac_address': '04:bc:9f:5a:37:71',
                                                                       'name': '277036/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '843574328311',
                                                                       'out_ucast_pkts': '648273917',
                                                                       'poller_hash': 'MTI3MEQ2RXBlZXJpbnRmMjc3MDM2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '277036',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '843574328311',
                                                                       'in_ucast_pkts': '648273917',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '843574328311',
                                                                                                'in_ucast_pkts': '648273917',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '33245879400',
                                                                                                'out_ucast_pkts': '192362671',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '33245879400',
                                                                       'out_ucast_pkts': '192362671',
                                                                       'poller_hash': 'MTI3MEQ2RXBlZXJpbnRmcG9uIDI3NzAzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277036',
                                                                                               'links': [       {       'connected_time': '688343',
                                                                                                                        'fiber_length': 975,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -12.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ2RXBlZXJwb25wb24gMjc3MDM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 TEMPERANCE ST (DEBUS DAVID)',
                                         'poller_hash': 'MTI3MEQ2RQ==',
                                         'serial_number': '1270D6E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:37:71',
                                         'uptime': '688343'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '277367/g1',
                                                                       'name': '277367/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3MzlEOXBlZXJpbnRmMjc3MzY3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '277367',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3MzlEOXBlZXJpbnRmcG9uIDI3NzM2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277367',
                                                                                               'links': [       {       'connected_time': '9698186',
                                                                                                                        'fiber_length': 1639,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MzlEOXBlZXJwb25wb24gMjc3MzY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 WILLIAM ST (METERGY SOLUTIONS '
                                                 'INC)',
                                         'poller_hash': 'MTI3MzlEOQ==',
                                         'serial_number': '12739D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:9a:c9',
                                         'uptime': '9698186'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '277377/g1',
                                                                       'name': '277377/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3MEQ3MHBlZXJpbnRmMjc3Mzc3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '277377',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzA3Z3BvbiA4Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3MEQ3MHBlZXJpbnRmcG9uIDI3NzM3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277377',
                                                                                               'links': [       {       'connected_time': '9684744',
                                                                                                                        'fiber_length': 11888,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzA3cG9uZ3BvbiA4Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ3MHBlZXJwb25wb24gMjc3Mzc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4022 PERTH LINE 72 (KIPFER SAM S)',
                                         'poller_hash': 'MTI3MEQ3MA==',
                                         'serial_number': '1270D70',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:37:81',
                                         'uptime': '9684744'},
                                 {       'counter_source': '277427/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '277427/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '50932034960',
                                                                       'in_ucast_pkts': '205380017',
                                                                       'mac_address': '04:bc:9f:6e:65:ee',
                                                                       'name': '277427/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '473345050841',
                                                                       'out_ucast_pkts': '369585268',
                                                                       'poller_hash': 'MTI5NjA0MXBlZXJpbnRmMjc3NDI3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '277427',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '473345050841',
                                                                       'in_ucast_pkts': '369585268',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '473345050841',
                                                                                                'in_ucast_pkts': '369585268',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '50932034960',
                                                                                                'out_ucast_pkts': '205380017',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '50932034960',
                                                                       'out_ucast_pkts': '205380017',
                                                                       'poller_hash': 'MTI5NjA0MXBlZXJpbnRmcG9uIDI3NzQyNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277427',
                                                                                               'links': [       {       'connected_time': '995949',
                                                                                                                        'fiber_length': 18580,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjA0MXBlZXJwb25wb24gMjc3NDI3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5930 LINE 57 (PELKMANS ELS)',
                                         'poller_hash': 'MTI5NjA0MQ==',
                                         'serial_number': '1296041',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:65:ee',
                                         'uptime': '995949'},
                                 {       'counter_source': '277437/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:e6:bb'],
                                                                       'id': '277437/g1',
                                                                       'in_nucast_pkts': '14388',
                                                                       'in_octets': '14434489786',
                                                                       'in_ucast_pkts': '91303309',
                                                                       'mac_address': '04:bc:9f:6e:58:a1',
                                                                       'mtu': 2000,
                                                                       'name': '277437/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1465060',
                                                                       'out_octets': '829966742160',
                                                                       'out_ucast_pkts': '565926629',
                                                                       'poller_hash': 'MTI5NUU5N3BlZXJpbnRmMjc3NDM3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '277437',
                                                                       'in_nucast_pkts': '1465060',
                                                                       'in_octets': '829966742160',
                                                                       'in_ucast_pkts': '565926629',
                                                                       'links': [       {       'in_nucast_pkts': '1465060',
                                                                                                'in_octets': '829966742160',
                                                                                                'in_ucast_pkts': '565926629',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14388',
                                                                                                'out_octets': '14434489786',
                                                                                                'out_ucast_pkts': '91303309',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14388',
                                                                       'out_octets': '14434489786',
                                                                       'out_ucast_pkts': '91303309',
                                                                       'poller_hash': 'MTI5NUU5N3BlZXJpbnRmcG9uIDI3NzQzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277437',
                                                                                               'links': [       {       'connected_time': '688366',
                                                                                                                        'fiber_length': 3062,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU5N3BlZXJwb25wb24gMjc3NDM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '112 PUGH ST (WITEN STACEY & '
                                                 'LAURENCE)',
                                         'poller_hash': 'MTI5NUU5Nw==',
                                         'serial_number': '1295E97',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:58:9e',
                                         'uptime': '688366'},
                                 {       'counter_source': '278148/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f0:2f'],
                                                                       'id': '278148/g1',
                                                                       'in_nucast_pkts': '177080',
                                                                       'in_octets': '137916330157',
                                                                       'in_ucast_pkts': '1469507078',
                                                                       'mac_address': '04:bc:9f:5a:36:8c',
                                                                       'mtu': 2000,
                                                                       'name': '278148/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18867298',
                                                                       'out_octets': '5364328933484',
                                                                       'out_ucast_pkts': '3645932598',
                                                                       'poller_hash': 'MTI3MEQ1MXBlZXJpbnRmMjc4MTQ4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '278148',
                                                                       'in_nucast_pkts': '18867298',
                                                                       'in_octets': '5364328933484',
                                                                       'in_ucast_pkts': '3645932598',
                                                                       'links': [       {       'in_nucast_pkts': '18867298',
                                                                                                'in_octets': '5364328933484',
                                                                                                'in_ucast_pkts': '3645932598',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '177080',
                                                                                                'out_octets': '137916330157',
                                                                                                'out_ucast_pkts': '1469507078',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '177080',
                                                                       'out_octets': '137916330157',
                                                                       'out_ucast_pkts': '1469507078',
                                                                       'poller_hash': 'MTI3MEQ1MXBlZXJpbnRmcG9uIDI3ODE0OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278148',
                                                                                               'links': [       {       'connected_time': '8836008',
                                                                                                                        'fiber_length': 18263,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ1MXBlZXJwb25wb24gMjc4MTQ4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 SCHADE ST (JARMUTH KEN & NORA)',
                                         'poller_hash': 'MTI3MEQ1MQ==',
                                         'serial_number': '1270D51',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:36:89',
                                         'uptime': '8836008'},
                                 {       'counter_source': '278188/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '278188/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '2720042',
                                                                       'in_octets': '95235881612',
                                                                       'in_ucast_pkts': '199530599',
                                                                       'mac_address': '04:bc:9f:24:06:dd',
                                                                       'name': '278188/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '15484518',
                                                                       'out_octets': '1250276216436',
                                                                       'out_ucast_pkts': '927523775',
                                                                       'poller_hash': 'MTUyMDI4NnBlZXJpbnRmMjc4MTg4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '278188',
                                                                       'in_nucast_pkts': '15484518',
                                                                       'in_octets': '1250276216436',
                                                                       'in_ucast_pkts': '927523775',
                                                                       'links': [       {       'in_nucast_pkts': '15484518',
                                                                                                'in_octets': '1250276216436',
                                                                                                'in_ucast_pkts': '927523775',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2720042',
                                                                                                'out_octets': '95235881612',
                                                                                                'out_ucast_pkts': '199530599',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2720042',
                                                                       'out_octets': '95235881612',
                                                                       'out_ucast_pkts': '199530599',
                                                                       'poller_hash': 'MTUyMDI4NnBlZXJpbnRmcG9uIDI3ODE4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278188',
                                                                                               'links': [       {       'connected_time': '244052',
                                                                                                                        'fiber_length': 1711,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.194,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.802}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUyMDI4NnBlZXJwb25wb24gMjc4MTg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': '55 WILLIAM ST AP 104 (FOX LORI)',
                                         'poller_hash': 'MTUyMDI4Ng==',
                                         'serial_number': '1520286',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.4.0.0.57',
                                         'system_mac_address': '04:bc:9f:24:06:dd',
                                         'uptime': '244052'},
                                 {       'counter_source': '278195/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '278195/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '14758921129',
                                                                       'in_ucast_pkts': '29363363',
                                                                       'mac_address': '04:bc:9f:5a:38:91',
                                                                       'name': '278195/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '51767495671',
                                                                       'out_ucast_pkts': '54994029',
                                                                       'poller_hash': 'MTI3MEQ5MnBlZXJpbnRmMjc4MTk1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '278195',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '51767495671',
                                                                       'in_ucast_pkts': '54994029',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '51767495671',
                                                                                                'in_ucast_pkts': '54994029',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '14758921129',
                                                                                                'out_ucast_pkts': '29363363',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '14758921129',
                                                                       'out_ucast_pkts': '29363363',
                                                                       'poller_hash': 'MTI3MEQ5MnBlZXJpbnRmcG9uIDI3ODE5NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278195',
                                                                                               'links': [       {       'connected_time': '8387845',
                                                                                                                        'fiber_length': 11732,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ5MnBlZXJwb25wb24gMjc4MTk1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6905 PERTH ROAD 121 (WAGLER HAROLD)',
                                         'poller_hash': 'MTI3MEQ5Mg==',
                                         'serial_number': '1270D92',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:38:91',
                                         'uptime': '8387845'},
                                 {       'counter_source': '278477/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '8c:90:2d:44:30:a8'],
                                                                       'id': '278477/g1',
                                                                       'in_nucast_pkts': '955',
                                                                       'in_octets': '6497779201',
                                                                       'in_ucast_pkts': '26449794',
                                                                       'mac_address': '04:bc:9f:5b:a1:34',
                                                                       'mtu': 2000,
                                                                       'name': '278477/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '541560',
                                                                       'out_octets': '107105677882',
                                                                       'out_ucast_pkts': '76377925',
                                                                       'poller_hash': 'MTI3M0FBNnBlZXJpbnRmMjc4NDc3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '278477',
                                                                       'in_nucast_pkts': '541560',
                                                                       'in_octets': '107105677882',
                                                                       'in_ucast_pkts': '76377925',
                                                                       'links': [       {       'in_nucast_pkts': '541560',
                                                                                                'in_octets': '107105677882',
                                                                                                'in_ucast_pkts': '76377925',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '955',
                                                                                                'out_octets': '6497779201',
                                                                                                'out_ucast_pkts': '26449794',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '955',
                                                                       'out_octets': '6497779201',
                                                                       'out_ucast_pkts': '26449794',
                                                                       'poller_hash': 'MTI3M0FBNnBlZXJpbnRmcG9uIDI3ODQ3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278477',
                                                                                               'links': [       {       'connected_time': '262470',
                                                                                                                        'fiber_length': 16790,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0FBNnBlZXJwb25wb24gMjc4NDc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2390 LINE 47 (SAUDER CANDICE)',
                                         'poller_hash': 'MTI3M0FBNg==',
                                         'serial_number': '1273AA6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:a1:31',
                                         'uptime': '262470'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '279077/g1',
                                                                       'name': '279077/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NUU3NnBlZXJpbnRmMjc5MDc3L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279077',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NUU3NnBlZXJpbnRmcG9uIDI3OTA3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279077',
                                                                                               'links': [       {       'connected_time': '7714292',
                                                                                                                        'fiber_length': 17364,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU3NnBlZXJwb25wb24gMjc5MDc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2332 LINE 47 (GERBER PAUL)',
                                         'poller_hash': 'MTI5NUU3Ng==',
                                         'serial_number': '1295E76',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:57:96',
                                         'uptime': '7714292'},
                                 {       'counter_source': '279150/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279150/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '2384670',
                                                                       'in_octets': '89413495185',
                                                                       'in_ucast_pkts': '138526364',
                                                                       'mac_address': '04:bc:9f:24:07:ac',
                                                                       'name': '279150/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '13558273',
                                                                       'out_octets': '801683375994',
                                                                       'out_ucast_pkts': '605054649',
                                                                       'poller_hash': 'MTUyMDI5RHBlZXJpbnRmMjc5MTUwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279150',
                                                                       'in_nucast_pkts': '13558273',
                                                                       'in_octets': '801683375994',
                                                                       'in_ucast_pkts': '605054649',
                                                                       'links': [       {       'in_nucast_pkts': '13558273',
                                                                                                'in_octets': '801683375994',
                                                                                                'in_ucast_pkts': '605054649',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2384670',
                                                                                                'out_octets': '89413495185',
                                                                                                'out_ucast_pkts': '138526364',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzAxZ3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2384670',
                                                                       'out_octets': '89413495185',
                                                                       'out_ucast_pkts': '138526364',
                                                                       'poller_hash': 'MTUyMDI5RHBlZXJpbnRmcG9uIDI3OTE1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279150',
                                                                                               'links': [       {       'connected_time': '688363',
                                                                                                                        'fiber_length': 1586,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzAxcG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -10.888,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.256}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUyMDI5RHBlZXJwb25wb24gMjc5MTUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': '57 WILLIAM ST APT 201 (BOBAN '
                                                 'AISWARYA)',
                                         'poller_hash': 'MTUyMDI5RA==',
                                         'serial_number': '152029D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '25.1.0.0.33',
                                         'system_mac_address': '04:bc:9f:24:07:ac',
                                         'uptime': '688363'},
                                 {       'counter_source': '279157/g3+279157/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:03:ea'],
                                                                       'id': '279157/g3',
                                                                       'in_nucast_pkts': '408406',
                                                                       'in_octets': '34643528348',
                                                                       'in_ucast_pkts': '462851915',
                                                                       'mac_address': '04:bc:9f:5a:3b:36',
                                                                       'mtu': 2000,
                                                                       'name': '279157/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '85930690',
                                                                       'out_octets': '2701367590060',
                                                                       'out_ucast_pkts': '1900656829',
                                                                       'poller_hash': 'MTI3MERFNnBlZXJpbnRmMjc5MTU3L2cz',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:13:e2'],
                                                                       'id': '279157/g4',
                                                                       'in_nucast_pkts': '341356',
                                                                       'in_octets': '1580108798',
                                                                       'in_ucast_pkts': '8935411',
                                                                       'mac_address': '04:bc:9f:5a:3b:37',
                                                                       'mtu': 2000,
                                                                       'name': '279157/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '77562141',
                                                                       'out_octets': '63809154097',
                                                                       'out_ucast_pkts': '25973321',
                                                                       'poller_hash': 'MTI3MERFNnBlZXJpbnRmMjc5MTU3L2c0',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '279157',
                                                                       'in_nucast_pkts': '163492831',
                                                                       'in_octets': '2765176744157',
                                                                       'in_ucast_pkts': '1926630150',
                                                                       'links': [       {       'in_nucast_pkts': '163492831',
                                                                                                'in_octets': '2765176744157',
                                                                                                'in_ucast_pkts': '1926630150',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '749762',
                                                                                                'out_octets': '36223637146',
                                                                                                'out_ucast_pkts': '471787326',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '749762',
                                                                       'out_octets': '36223637146',
                                                                       'out_ucast_pkts': '471787326',
                                                                       'poller_hash': 'MTI3MERFNnBlZXJpbnRmcG9uIDI3OTE1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279157',
                                                                                               'links': [       {       'connected_time': '7351722',
                                                                                                                        'fiber_length': 1076,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -9.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MERFNnBlZXJwb25wb24gMjc5MTU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 GROSCH ST (DOUPE DOLORES)',
                                         'poller_hash': 'MTI3MERFNg==',
                                         'serial_number': '1270DE6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:3b:31',
                                         'uptime': '7351722'},
                                 {       'counter_source': '279276/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279276/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '97112081365',
                                                                       'in_ucast_pkts': '21738596000',
                                                                       'mac_address': '04:bc:9f:6e:58:fe',
                                                                       'name': '279276/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1475807946359',
                                                                       'out_ucast_pkts': '1044648714',
                                                                       'poller_hash': 'MTI5NUVBM3BlZXJpbnRmMjc5Mjc2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279276',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1475807946359',
                                                                       'in_ucast_pkts': '1044648714',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1475807946359',
                                                                                                'in_ucast_pkts': '1044648714',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '97112081365',
                                                                                                'out_ucast_pkts': '21738596000',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '97112081365',
                                                                       'out_ucast_pkts': '21738596000',
                                                                       'poller_hash': 'MTI5NUVBM3BlZXJpbnRmcG9uIDI3OTI3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279276',
                                                                                               'links': [       {       'connected_time': '688571',
                                                                                                                        'fiber_length': 7114,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVBM3BlZXJwb25wb24gMjc5Mjc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5979 PERTH ROAD 119 (MARTIN DARRYK)',
                                         'poller_hash': 'MTI5NUVBMw==',
                                         'serial_number': '1295EA3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:58:fe',
                                         'uptime': '688571'},
                                 {       'counter_source': '279286/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279286/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '592187652',
                                                                       'in_ucast_pkts': '1597735',
                                                                       'mac_address': '04:bc:9f:6e:56:2e',
                                                                       'name': '279286/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4535845304',
                                                                       'out_ucast_pkts': '10660471',
                                                                       'poller_hash': 'MTI5NUU0OXBlZXJpbnRmMjc5Mjg2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279286',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4535845304',
                                                                       'in_ucast_pkts': '10660471',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4535845304',
                                                                                                'in_ucast_pkts': '10660471',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '592187652',
                                                                                                'out_ucast_pkts': '1597735',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '592187652',
                                                                       'out_ucast_pkts': '1597735',
                                                                       'poller_hash': 'MTI5NUU0OXBlZXJpbnRmcG9uIDI3OTI4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279286',
                                                                                               'links': [       {       'connected_time': '6490902',
                                                                                                                        'fiber_length': 10561,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -11.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU0OXBlZXJwb25wb24gMjc5Mjg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5883 PERTH ROAD 119 (JACOBS HERB K)',
                                         'poller_hash': 'MTI5NUU0OQ==',
                                         'serial_number': '1295E49',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:56:2e',
                                         'uptime': '6490902'},
                                 {       'counter_source': '279306/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:ec:a9'],
                                                                       'id': '279306/g1',
                                                                       'in_nucast_pkts': '12111',
                                                                       'in_octets': '3336972692',
                                                                       'in_ucast_pkts': '8808926',
                                                                       'mac_address': '04:bc:9f:6e:59:a9',
                                                                       'mtu': 2000,
                                                                       'name': '279306/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1435250',
                                                                       'out_octets': '81998244116',
                                                                       'out_ucast_pkts': '59872512',
                                                                       'poller_hash': 'MTI5NUVCOHBlZXJpbnRmMjc5MzA2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '279306',
                                                                       'in_nucast_pkts': '1435250',
                                                                       'in_octets': '81998244116',
                                                                       'in_ucast_pkts': '59872512',
                                                                       'links': [       {       'in_nucast_pkts': '1435250',
                                                                                                'in_octets': '81998244116',
                                                                                                'in_ucast_pkts': '59872512',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12111',
                                                                                                'out_octets': '3336972692',
                                                                                                'out_ucast_pkts': '8808926',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12111',
                                                                       'out_octets': '3336972692',
                                                                       'out_ucast_pkts': '8808926',
                                                                       'poller_hash': 'MTI5NUVCOHBlZXJpbnRmcG9uIDI3OTMwNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279306',
                                                                                               'links': [       {       'connected_time': '688561',
                                                                                                                        'fiber_length': 1658,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVCOHBlZXJwb25wb24gMjc5MzA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '33 MURRAY CT (SHAW DANIELLE)',
                                         'poller_hash': 'MTI5NUVCOA==',
                                         'serial_number': '1295EB8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:59:a6',
                                         'uptime': '688561'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '279336/g1',
                                                                       'name': '279336/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3MzExN3BlZXJpbnRmMjc5MzM2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279336',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3MzExN3BlZXJpbnRmcG9uIDI3OTMzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279336',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MzExN3BlZXJwb25wb24gMjc5MzM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '6600 PERTH ROAD 121 (SAUVE JOE)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "279336" '
                                                                  '(location: GPON '
                                                                  'port "13/3"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-08-05T13:18:54.160Z',
                                         'poller_hash': 'MTI3MzExNw==',
                                         'serial_number': '1273117',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '279366/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279366/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '295',
                                                                       'in_octets': '55377752882',
                                                                       'in_ucast_pkts': '565689505',
                                                                       'mac_address': '04:bc:9f:5c:0f:79',
                                                                       'name': '279366/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1425',
                                                                       'out_octets': '2305495252659',
                                                                       'out_ucast_pkts': '1560648822',
                                                                       'poller_hash': 'MTI3NDg2RnBlZXJpbnRmMjc5MzY2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279366',
                                                                       'in_nucast_pkts': '1425',
                                                                       'in_octets': '2305495252659',
                                                                       'in_ucast_pkts': '1560648822',
                                                                       'links': [       {       'in_nucast_pkts': '1425',
                                                                                                'in_octets': '2305495252659',
                                                                                                'in_ucast_pkts': '1560648822',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '295',
                                                                                                'out_octets': '55377752882',
                                                                                                'out_ucast_pkts': '565689505',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '295',
                                                                       'out_octets': '55377752882',
                                                                       'out_ucast_pkts': '565689505',
                                                                       'poller_hash': 'MTI3NDg2RnBlZXJpbnRmcG9uIDI3OTM2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279366',
                                                                                               'links': [       {       'connected_time': '6400919',
                                                                                                                        'fiber_length': 916,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg2RnBlZXJwb25wb24gMjc5MzY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '8 GERBER DR (KNIPE DOROTHY)',
                                         'poller_hash': 'MTI3NDg2Rg==',
                                         'serial_number': '127486F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0f:79',
                                         'uptime': '6400919'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '279406/g1',
                                                                       'name': '279406/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3MEQ5RHBlZXJpbnRmMjc5NDA2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279406',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy81'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3MEQ5RHBlZXJpbnRmcG9uIDI3OTQwNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279406',
                                                                                               'links': [       {       'connected_time': '5970382',
                                                                                                                        'fiber_length': 8565,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy81',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ5RHBlZXJwb25wb24gMjc5NDA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6771 PERTH ROAD 121 (KUEPFER SAM A)',
                                         'poller_hash': 'MTI3MEQ5RA==',
                                         'serial_number': '1270D9D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:38:e9',
                                         'uptime': '5970382'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '279436/g1',
                                                                       'name': '279436/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3M0FERnBlZXJpbnRmMjc5NDM2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279436',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3M0FERnBlZXJpbnRmcG9uIDI3OTQzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279436',
                                                                                               'links': [       {       'connected_time': '5702777',
                                                                                                                        'fiber_length': 5184,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0FERnBlZXJwb25wb24gMjc5NDM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4622 PERTH LINE 55 (EBERSOL RONALD)',
                                         'poller_hash': 'MTI3M0FERg==',
                                         'serial_number': '1273ADF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:a2:f9',
                                         'uptime': '5702777'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '279526/g1',
                                                                       'name': '279526/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgxRTJEcGVlcmludGYyNzk1MjYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '279526',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgxRTJEcGVlcmludGZwb24gMjc5NTI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279526',
                                                                                               'links': [       {       'connected_time': '5466100',
                                                                                                                        'fiber_length': 5340,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTJEcGVlcnBvbnBvbiAyNzk1MjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4626 PERTH LINE 55 (JANTZI MARVIN A)',
                                         'poller_hash': 'QzgxRTJE',
                                         'serial_number': 'C81E2D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:46:1f',
                                         'uptime': '5466100'},
                                 {       'counter_source': '279984/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279984/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5940634675',
                                                                       'in_ucast_pkts': '16764845',
                                                                       'mac_address': '04:bc:9f:5b:93:99',
                                                                       'name': '279984/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '44381762322',
                                                                       'out_ucast_pkts': '40081754',
                                                                       'poller_hash': 'MTI3MzhGM3BlZXJpbnRmMjc5OTg0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279984',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '44381762322',
                                                                       'in_ucast_pkts': '40081754',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '44381762322',
                                                                                                'in_ucast_pkts': '40081754',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '5940634675',
                                                                                                'out_ucast_pkts': '16764845',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5940634675',
                                                                       'out_ucast_pkts': '16764845',
                                                                       'poller_hash': 'MTI3MzhGM3BlZXJpbnRmcG9uIDI3OTk4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279984',
                                                                                               'links': [       {       'connected_time': '688343',
                                                                                                                        'fiber_length': 1599,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MzhGM3BlZXJwb25wb24gMjc5OTg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '86 MAIN ST N (FRIESEN SHAWN)',
                                         'poller_hash': 'MTI3MzhGMw==',
                                         'serial_number': '12738F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:93:99',
                                         'uptime': '688343'},
                                 {       'counter_source': '280224/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '280224/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '85367984980',
                                                                       'in_ucast_pkts': '567973699',
                                                                       'mac_address': '04:bc:9f:6e:66:06',
                                                                       'name': '280224/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1494757532031',
                                                                       'out_ucast_pkts': '1062950934',
                                                                       'poller_hash': 'MTI5NjA0NHBlZXJpbnRmMjgwMjI0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '280224',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1494757532031',
                                                                       'in_ucast_pkts': '1062950934',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1494757532031',
                                                                                                'in_ucast_pkts': '1062950934',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '85367984980',
                                                                                                'out_ucast_pkts': '567973699',
                                                                                                'poller_hash': 'MjExNTEyNjAzNzE3Z3BvbiAzLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '85367984980',
                                                                       'out_ucast_pkts': '567973699',
                                                                       'poller_hash': 'MTI5NjA0NHBlZXJpbnRmcG9uIDI4MDIyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280224',
                                                                                               'links': [       {       'connected_time': '688355',
                                                                                                                        'fiber_length': 11792,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNzE3cG9uZ3BvbiAzLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjA0NHBlZXJwb25wb24gMjgwMjI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6932 HORIZON TL (NICKEL AMANDA)',
                                         'poller_hash': 'MTI5NjA0NA==',
                                         'serial_number': '1296044',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:66:06',
                                         'uptime': '688355'},
                                 {       'counter_source': '280244/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4f:b1:d8'],
                                                                       'id': '280244/g1',
                                                                       'in_nucast_pkts': '9435',
                                                                       'in_octets': '1460324621',
                                                                       'in_ucast_pkts': '2952225',
                                                                       'mac_address': '04:bc:9f:6e:61:79',
                                                                       'mtu': 2000,
                                                                       'name': '280244/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1149412',
                                                                       'out_octets': '3551493216',
                                                                       'out_ucast_pkts': '3662639',
                                                                       'poller_hash': 'MTI5NUZCMnBlZXJpbnRmMjgwMjQ0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '280244',
                                                                       'in_nucast_pkts': '1149412',
                                                                       'in_octets': '3551493216',
                                                                       'in_ucast_pkts': '3662639',
                                                                       'links': [       {       'in_nucast_pkts': '1149412',
                                                                                                'in_octets': '3551493216',
                                                                                                'in_ucast_pkts': '3662639',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9435',
                                                                                                'out_octets': '1460324621',
                                                                                                'out_ucast_pkts': '2952225',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9435',
                                                                       'out_octets': '1460324621',
                                                                       'out_ucast_pkts': '2952225',
                                                                       'poller_hash': 'MTI5NUZCMnBlZXJpbnRmcG9uIDI4MDI0NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280244',
                                                                                               'links': [       {       'connected_time': '537662',
                                                                                                                        'fiber_length': 10580,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUZCMnBlZXJwb25wb24gMjgwMjQ0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5915 SOMERSET DR (TOMBS GREG)',
                                         'poller_hash': 'MTI5NUZCMg==',
                                         'serial_number': '1295FB2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:61:76',
                                         'uptime': '537662'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '280314/g1',
                                                                       'name': '280314/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NUU2NnBlZXJpbnRmMjgwMzE0L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '280314',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2Nzg5Z3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NUU2NnBlZXJpbnRmcG9uIDI4MDMxNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280314',
                                                                                               'links': [       {       'connected_time': '592922',
                                                                                                                        'fiber_length': 1835,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2Nzg5cG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU2NnBlZXJwb25wb24gMjgwMzE0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4494 LINE 64 (ZEHR JACOB)',
                                         'poller_hash': 'MTI5NUU2Ng==',
                                         'serial_number': '1295E66',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:57:16',
                                         'uptime': '592922'},
                                 {       'counter_source': '280324/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '280324/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '22667070328',
                                                                       'in_ucast_pkts': '45626577',
                                                                       'mac_address': '04:bc:9f:6e:59:16',
                                                                       'name': '280324/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '157375331237',
                                                                       'out_ucast_pkts': '133725444',
                                                                       'poller_hash': 'MTI5NUVBNnBlZXJpbnRmMjgwMzI0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '280324',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '157375331237',
                                                                       'in_ucast_pkts': '133725444',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '157375331237',
                                                                                                'in_ucast_pkts': '133725444',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '22667070328',
                                                                                                'out_ucast_pkts': '45626577',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8x'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '22667070328',
                                                                       'out_ucast_pkts': '45626577',
                                                                       'poller_hash': 'MTI5NUVBNnBlZXJpbnRmcG9uIDI4MDMyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280324',
                                                                                               'links': [       {       'connected_time': '4603596',
                                                                                                                        'fiber_length': 6687,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8x',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVBNnBlZXJwb25wb24gMjgwMzI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4464 CRESCENT AV (ALBRECHT PAUL & '
                                                 'MARILYN)',
                                         'poller_hash': 'MTI5NUVBNg==',
                                         'serial_number': '1295EA6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:59:16',
                                         'uptime': '4603596'},
                                 {       'counter_source': '280394/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '280394/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '826361575',
                                                                       'in_ucast_pkts': '5004410',
                                                                       'mac_address': '04:bc:9f:6e:6d:1e',
                                                                       'name': '280394/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '38533322150',
                                                                       'out_ucast_pkts': '32037892',
                                                                       'poller_hash': 'MTI5NjEyN3BlZXJpbnRmMjgwMzk0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '280394',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '38533322150',
                                                                       'in_ucast_pkts': '32037892',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '38533322150',
                                                                                                'in_ucast_pkts': '32037892',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '826361575',
                                                                                                'out_ucast_pkts': '5004410',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '826361575',
                                                                       'out_ucast_pkts': '5004410',
                                                                       'poller_hash': 'MTI5NjEyN3BlZXJpbnRmcG9uIDI4MDM5NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280394',
                                                                                               'links': [       {       'connected_time': '688567',
                                                                                                                        'fiber_length': 6296,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjEyN3BlZXJwb25wb24gMjgwMzk0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6015 PERTH ROAD 119 (KUEPFER JUDITH)',
                                         'poller_hash': 'MTI5NjEyNw==',
                                         'serial_number': '1296127',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6d:1e',
                                         'uptime': '688567'},
                                 {       'counter_source': '280404/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '280404/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '92409522',
                                                                       'in_ucast_pkts': '200582',
                                                                       'mac_address': '04:bc:9f:6e:65:56',
                                                                       'name': '280404/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1433884571',
                                                                       'out_ucast_pkts': '5623640',
                                                                       'poller_hash': 'MTI5NjAyRXBlZXJpbnRmMjgwNDA0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '280404',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1433884571',
                                                                       'in_ucast_pkts': '5623640',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1433884571',
                                                                                                'in_ucast_pkts': '5623640',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '92409522',
                                                                                                'out_ucast_pkts': '200582',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '92409522',
                                                                       'out_ucast_pkts': '200582',
                                                                       'poller_hash': 'MTI5NjAyRXBlZXJpbnRmcG9uIDI4MDQwNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280404',
                                                                                               'links': [       {       'connected_time': '4257088',
                                                                                                                        'fiber_length': 11703,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjAyRXBlZXJwb25wb24gMjgwNDA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4072 WATERLOO ST (MILLBANK '
                                                 'ENTERPRISES INC)',
                                         'poller_hash': 'MTI5NjAyRQ==',
                                         'serial_number': '129602E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:65:56',
                                         'uptime': '4257088'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '281630/g1',
                                                                       'name': '281630/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NjAxMnBlZXJpbnRmMjgxNjMwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '281630',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NjAxMnBlZXJpbnRmcG9uIDI4MTYzMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '281630',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjAxMnBlZXJwb25wb24gMjgxNjMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4460C PERTH LINE 72 (WALL AARON)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "281630" '
                                                                  '(location: GPON '
                                                                  'port "15/2"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-08-04T23:27:55.300Z',
                                         'poller_hash': 'MTI5NjAxMg==',
                                         'serial_number': '1296012',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '282037/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282037/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '111558773650',
                                                                       'in_ucast_pkts': '163495298',
                                                                       'mac_address': '04:bc:9f:6e:6d:be',
                                                                       'name': '282037/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '634962669695',
                                                                       'out_ucast_pkts': '480373355',
                                                                       'poller_hash': 'MTI5NjEzQnBlZXJpbnRmMjgyMDM3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282037',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '634962669695',
                                                                       'in_ucast_pkts': '480373355',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '634962669695',
                                                                                                'in_ucast_pkts': '480373355',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '111558773650',
                                                                                                'out_ucast_pkts': '163495298',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '111558773650',
                                                                       'out_ucast_pkts': '163495298',
                                                                       'poller_hash': 'MTI5NjEzQnBlZXJpbnRmcG9uIDI4MjAzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282037',
                                                                                               'links': [       {       'connected_time': '688370',
                                                                                                                        'fiber_length': 3334,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjEzQnBlZXJwb25wb24gMjgyMDM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '74 PUGH ST (SCOTCHMER WAYNE)',
                                         'poller_hash': 'MTI5NjEzQg==',
                                         'serial_number': '129613B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6d:be',
                                         'uptime': '688370'},
                                 {       'counter_source': '282047/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282047/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1751979952',
                                                                       'in_ucast_pkts': '7627007',
                                                                       'mac_address': '04:bc:9f:6e:6c:66',
                                                                       'name': '282047/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '19026514314',
                                                                       'out_ucast_pkts': '19149689',
                                                                       'poller_hash': 'MTI5NjExMHBlZXJpbnRmMjgyMDQ3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282047',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '19026514314',
                                                                       'in_ucast_pkts': '19149689',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '19026514314',
                                                                                                'in_ucast_pkts': '19149689',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1751979952',
                                                                                                'out_ucast_pkts': '7627007',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1751979952',
                                                                       'out_ucast_pkts': '7627007',
                                                                       'poller_hash': 'MTI5NjExMHBlZXJpbnRmcG9uIDI4MjA0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282047',
                                                                                               'links': [       {       'connected_time': '3564005',
                                                                                                                        'fiber_length': 12013,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjExMHBlZXJwb25wb24gMjgyMDQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6896 PERTH ROAD 121 (T & D '
                                                 'RESTORATIONS INC)',
                                         'poller_hash': 'MTI5NjExMA==',
                                         'serial_number': '1296110',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6c:66',
                                         'uptime': '3564005'},
                                 {       'counter_source': '282070/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282070/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25826325862',
                                                                       'in_ucast_pkts': '97153165',
                                                                       'mac_address': '04:bc:9f:6e:55:9e',
                                                                       'name': '282070/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '297251118819',
                                                                       'out_ucast_pkts': '247648327',
                                                                       'poller_hash': 'MTI5NUUzN3BlZXJpbnRmMjgyMDcwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282070',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '297251118819',
                                                                       'in_ucast_pkts': '247648327',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '297251118819',
                                                                                                'in_ucast_pkts': '247648327',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '25826325862',
                                                                                                'out_ucast_pkts': '97153165',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25826325862',
                                                                       'out_ucast_pkts': '97153165',
                                                                       'poller_hash': 'MTI5NUUzN3BlZXJpbnRmcG9uIDI4MjA3MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282070',
                                                                                               'links': [       {       'connected_time': '688362',
                                                                                                                        'fiber_length': 3096,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUUzN3BlZXJwb25wb24gMjgyMDcw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '54 AVERY PL (KUHL WADE)',
                                         'poller_hash': 'MTI5NUUzNw==',
                                         'serial_number': '1295E37',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:55:9e',
                                         'uptime': '688362'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '282080/g1',
                                                                       'name': '282080/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NjAwRXBlZXJpbnRmMjgyMDgwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282080',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NjAwRXBlZXJpbnRmcG9uIDI4MjA4MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282080',
                                                                                               'links': [       {       'connected_time': '688374',
                                                                                                                        'fiber_length': 3074,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjAwRXBlZXJwb25wb24gMjgyMDgw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '128 PUGH ST',
                                         'poller_hash': 'MTI5NjAwRQ==',
                                         'serial_number': '129600E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:64:56',
                                         'uptime': '688374'},
                                 {       'counter_source': '282136/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282136/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '143391161500',
                                                                       'in_ucast_pkts': '178055765',
                                                                       'mac_address': '04:bc:9f:6e:59:36',
                                                                       'name': '282136/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '462723864261',
                                                                       'out_ucast_pkts': '362592193',
                                                                       'poller_hash': 'MTI5NUVBQXBlZXJpbnRmMjgyMTM2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282136',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '462723864261',
                                                                       'in_ucast_pkts': '362592193',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '462723864261',
                                                                                                'in_ucast_pkts': '362592193',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '143391161500',
                                                                                                'out_ucast_pkts': '178055765',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzk2Z3BvbiAxNS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '143391161500',
                                                                       'out_ucast_pkts': '178055765',
                                                                       'poller_hash': 'MTI5NUVBQXBlZXJpbnRmcG9uIDI4MjEzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282136',
                                                                                               'links': [       {       'connected_time': '688604',
                                                                                                                        'fiber_length': 6671,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzk2cG9uZ3BvbiAxNS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVBQXBlZXJwb25wb24gMjgyMTM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4432 PERTH LINE 72 (FABER SUSAN)',
                                         'poller_hash': 'MTI5NUVBQQ==',
                                         'serial_number': '1295EAA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:59:36',
                                         'uptime': '688604'},
                                 {       'counter_source': '282166/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282166/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'mac_address': '04:bc:9f:26:48:3c',
                                                                       'name': '282166/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUyNDJCN3BlZXJpbnRmMjgyMTY2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282166',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi82'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUyNDJCN3BlZXJpbnRmcG9uIDI4MjE2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282166',
                                                                                               'links': [       {       'connected_time': '688454',
                                                                                                                        'fiber_length': 1815,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi82',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.186,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.676}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUyNDJCN3BlZXJwb25wb24gMjgyMTY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': '25 MURRAY CT (COLE DANIELLE)',
                                         'poller_hash': 'MTUyNDJCNw==',
                                         'serial_number': '15242B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.4.0.0.57',
                                         'system_mac_address': '04:bc:9f:26:48:3c',
                                         'uptime': '688454'},
                                 {       'counter_source': '282193/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282193/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '777290892',
                                                                       'in_ucast_pkts': '7642011',
                                                                       'mac_address': '04:bc:9f:6e:66:56',
                                                                       'name': '282193/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '26998996509',
                                                                       'out_ucast_pkts': '18543018',
                                                                       'poller_hash': 'MTI5NjA0RXBlZXJpbnRmMjgyMTkzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282193',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '26998996509',
                                                                       'in_ucast_pkts': '18543018',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '26998996509',
                                                                                                'in_ucast_pkts': '18543018',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '777290892',
                                                                                                'out_ucast_pkts': '7642011',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '777290892',
                                                                       'out_ucast_pkts': '7642011',
                                                                       'poller_hash': 'MTI5NjA0RXBlZXJpbnRmcG9uIDI4MjE5Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282193',
                                                                                               'links': [       {       'connected_time': '3135641',
                                                                                                                        'fiber_length': 8784,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjA0RXBlZXJwb25wb24gMjgyMTkz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4486 LINE 52 (DEBUS RON)',
                                         'poller_hash': 'MTI5NjA0RQ==',
                                         'serial_number': '129604E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:66:56',
                                         'uptime': '3135641'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '282253/g1',
                                                                       'name': '282253/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NjBFRnBlZXJpbnRmMjgyMjUzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282253',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NjBFRnBlZXJpbnRmcG9uIDI4MjI1Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282253',
                                                                                               'links': [       {       'connected_time': '2961268',
                                                                                                                        'fiber_length': 5919,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjBFRnBlZXJwb25wb24gMjgyMjUz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4374 LINE 64 (ALBRECHT DAVID Z)',
                                         'poller_hash': 'MTI5NjBFRg==',
                                         'serial_number': '12960EF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6b:5e',
                                         'uptime': '2961268'},
                                 {       'counter_source': '282273/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282273/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2064620939',
                                                                       'in_ucast_pkts': '16073877',
                                                                       'mac_address': 'b8:94:70:ba:ac:15',
                                                                       'name': '282273/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '107068424889',
                                                                       'out_ucast_pkts': '78665628',
                                                                       'poller_hash': 'MTE5NEE5N3BlZXJpbnRmMjgyMjczL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282273',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '107068424889',
                                                                       'in_ucast_pkts': '78665628',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '107068424889',
                                                                                                'in_ucast_pkts': '78665628',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2064620939',
                                                                                                'out_ucast_pkts': '16073877',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8y'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2064620939',
                                                                       'out_ucast_pkts': '16073877',
                                                                       'poller_hash': 'MTE5NEE5N3BlZXJpbnRmcG9uIDI4MjI3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282273',
                                                                                               'links': [       {       'connected_time': '2873093',
                                                                                                                        'fiber_length': 17545,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8y',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NEE5N3BlZXJwb25wb24gMjgyMjcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '202 SELBY AV (TOM GOODYEAR)',
                                         'poller_hash': 'MTE5NEE5Nw==',
                                         'serial_number': '1194A97',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:ac:15',
                                         'uptime': '2873093'},
                                 {       'counter_source': '282303/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282303/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2104009264',
                                                                       'in_ucast_pkts': '6686474',
                                                                       'mac_address': '04:bc:9f:6e:66:b6',
                                                                       'name': '282303/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '38985987894',
                                                                       'out_ucast_pkts': '31969022',
                                                                       'poller_hash': 'MTI5NjA1QXBlZXJpbnRmMjgyMzAzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282303',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '38985987894',
                                                                       'in_ucast_pkts': '31969022',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '38985987894',
                                                                                                'in_ucast_pkts': '31969022',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2104009264',
                                                                                                'out_ucast_pkts': '6686474',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2104009264',
                                                                       'out_ucast_pkts': '6686474',
                                                                       'poller_hash': 'MTI5NjA1QXBlZXJpbnRmcG9uIDI4MjMwMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282303',
                                                                                               'links': [       {       'connected_time': '688352',
                                                                                                                        'fiber_length': 11508,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjA1QXBlZXJwb25wb24gMjgyMzAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6982 CHURCH ST (HORST LAVERNE)',
                                         'poller_hash': 'MTI5NjA1QQ==',
                                         'serial_number': '129605A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:66:b6',
                                         'uptime': '688352'},
                                 {       'counter_source': '282363/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4f:bc:28'],
                                                                       'id': '282363/g1',
                                                                       'in_nucast_pkts': '45302',
                                                                       'in_octets': '7885467821',
                                                                       'in_ucast_pkts': '41625289',
                                                                       'mac_address': '04:bc:9f:6e:59:c1',
                                                                       'mtu': 2000,
                                                                       'name': '282363/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5480457',
                                                                       'out_octets': '294262623869',
                                                                       'out_ucast_pkts': '202541461',
                                                                       'poller_hash': 'MTI5NUVCQnBlZXJpbnRmMjgyMzYzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '282363',
                                                                       'in_nucast_pkts': '5480457',
                                                                       'in_octets': '294262623869',
                                                                       'in_ucast_pkts': '202541461',
                                                                       'links': [       {       'in_nucast_pkts': '5480457',
                                                                                                'in_octets': '294262623869',
                                                                                                'in_ucast_pkts': '202541461',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '45302',
                                                                                                'out_octets': '7885467821',
                                                                                                'out_ucast_pkts': '41625289',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45302',
                                                                       'out_octets': '7885467821',
                                                                       'out_ucast_pkts': '41625289',
                                                                       'poller_hash': 'MTI5NUVCQnBlZXJpbnRmcG9uIDI4MjM2Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282363',
                                                                                               'links': [       {       'connected_time': '2530205',
                                                                                                                        'fiber_length': 13858,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVCQnBlZXJwb25wb24gMjgyMzYz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5499 PERTH LINE 55 (MEULLER LINDA)',
                                         'poller_hash': 'MTI5NUVCQg==',
                                         'serial_number': '1295EBB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:59:be',
                                         'uptime': '2530205'},
                                 {       'counter_source': '282921/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282921/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '21011760684',
                                                                       'in_ucast_pkts': '39873971',
                                                                       'mac_address': '04:bc:9f:6e:58:be',
                                                                       'name': '282921/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '146153223817',
                                                                       'out_ucast_pkts': '116824924',
                                                                       'poller_hash': 'MTI5NUU5QnBlZXJpbnRmMjgyOTIxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282921',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '146153223817',
                                                                       'in_ucast_pkts': '116824924',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '146153223817',
                                                                                                'in_ucast_pkts': '116824924',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '21011760684',
                                                                                                'out_ucast_pkts': '39873971',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy83'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '21011760684',
                                                                       'out_ucast_pkts': '39873971',
                                                                       'poller_hash': 'MTI5NUU5QnBlZXJpbnRmcG9uIDI4MjkyMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282921',
                                                                                               'links': [       {       'connected_time': '262468',
                                                                                                                        'fiber_length': 16507,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy83',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU5QnBlZXJwb25wb24gMjgyOTIx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2263 LINE 51 (LINDNER JANE)',
                                         'poller_hash': 'MTI5NUU5Qg==',
                                         'serial_number': '1295E9B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:58:be',
                                         'uptime': '262468'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '283010/g1',
                                                                       'name': '283010/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NjEyMXBlZXJpbnRmMjgzMDEwL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '283010',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzY4Z3BvbiAxMi80'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NjEyMXBlZXJpbnRmcG9uIDI4MzAxMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '283010',
                                                                                               'links': [       {       'connected_time': '2253186',
                                                                                                                        'fiber_length': 4532,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzY4cG9uZ3BvbiAxMi80',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjEyMXBlZXJwb25wb24gMjgzMDEw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6190 PERTH ROAD 131 (DAVIDSON JEREMY '
                                                 '& MALINDA)',
                                         'poller_hash': 'MTI5NjEyMQ==',
                                         'serial_number': '1296121',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6c:ee',
                                         'uptime': '2253186'},
                                 {       'counter_source': '283516/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:01:b3:c7'],
                                                                       'id': '283516/g2',
                                                                       'in_nucast_pkts': '41975',
                                                                       'in_octets': '2488974953',
                                                                       'in_ucast_pkts': '3272670',
                                                                       'mac_address': '04:bc:9f:6e:6a:c2',
                                                                       'mtu': 2000,
                                                                       'name': '283516/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4669041',
                                                                       'out_octets': '11392831063',
                                                                       'out_ucast_pkts': '8859791',
                                                                       'poller_hash': 'MTI5NjBEQnBlZXJpbnRmMjgzNTE2L2cy',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '283516',
                                                                       'in_nucast_pkts': '4669041',
                                                                       'in_octets': '11392831063',
                                                                       'in_ucast_pkts': '8859791',
                                                                       'links': [       {       'in_nucast_pkts': '4669041',
                                                                                                'in_octets': '11392831063',
                                                                                                'in_ucast_pkts': '8859791',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '41975',
                                                                                                'out_octets': '2488974953',
                                                                                                'out_ucast_pkts': '3272670',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41975',
                                                                       'out_octets': '2488974953',
                                                                       'out_ucast_pkts': '3272670',
                                                                       'poller_hash': 'MTI5NjBEQnBlZXJpbnRmcG9uIDI4MzUxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '283516',
                                                                                               'links': [       {       'connected_time': '2180897',
                                                                                                                        'fiber_length': 18231,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjBEQnBlZXJwb25wb24gMjgzNTE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 SCHADE ST MONKTON (HUBER DERRICK)',
                                         'poller_hash': 'MTI5NjBEQg==',
                                         'serial_number': '12960DB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6a:be',
                                         'uptime': '2180897'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '284488/g1',
                                                                       'name': '284488/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NjBFMHBlZXJpbnRmMjg0NDg4L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '284488',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjYxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NjBFMHBlZXJpbnRmcG9uIDI4NDQ4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '284488',
                                                                                               'links': [       {       'connected_time': '1292753',
                                                                                                                        'fiber_length': 11350,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjYxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjBFMHBlZXJwb25wb24gMjg0NDg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6931 PERTH ROAD 121 (ED ZEHR)',
                                         'poller_hash': 'MTI5NjBFMA==',
                                         'serial_number': '12960E0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6a:e6',
                                         'uptime': '1292753'},
                                 {       'counter_source': '284552/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '284552/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5593965093',
                                                                       'in_ucast_pkts': '32504607',
                                                                       'mac_address': '04:bc:9f:5c:1a:39',
                                                                       'name': '284552/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '145918397322',
                                                                       'out_ucast_pkts': '108749916',
                                                                       'poller_hash': 'MTI3NDlDN3BlZXJpbnRmMjg0NTUyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '284552',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '145918397322',
                                                                       'in_ucast_pkts': '108749916',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '145918397322',
                                                                                                'in_ucast_pkts': '108749916',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '5593965093',
                                                                                                'out_ucast_pkts': '32504607',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS84'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5593965093',
                                                                       'out_ucast_pkts': '32504607',
                                                                       'poller_hash': 'MTI3NDlDN3BlZXJpbnRmcG9uIDI4NDU1Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '284552',
                                                                                               'links': [       {       'connected_time': '973050',
                                                                                                                        'fiber_length': 6444,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS84',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDlDN3BlZXJwb25wb24gMjg0NTUy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'VICTOR POOLE (6314 RD 124)',
                                         'poller_hash': 'MTI3NDlDNw==',
                                         'serial_number': '12749C7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1a:39',
                                         'uptime': '973050'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '284835/g1',
                                                                       'name': '284835/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI5NjEyMnBlZXJpbnRmMjg0ODM1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '284835',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDA2NDE1Z3BvbiAxMy8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI5NjEyMnBlZXJpbnRmcG9uIDI4NDgzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '284835',
                                                                                               'links': [       {       'connected_time': '971153',
                                                                                                                        'fiber_length': 5855,
                                                                                                                        'poller_hash': 'NDcxOTA1MDA2NDE1cG9uZ3BvbiAxMy8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjEyMnBlZXJwb25wb24gMjg0ODM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6478 ROAD 129 (KUEPFER ELDON R)',
                                         'poller_hash': 'MTI5NjEyMg==',
                                         'serial_number': '1296122',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6c:f6',
                                         'uptime': '971153'},
                                 {       'counter_source': '284845/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:38:34'],
                                                                       'id': '284845/g1',
                                                                       'in_nucast_pkts': '16778',
                                                                       'in_octets': '35938820116',
                                                                       'in_ucast_pkts': '284852556',
                                                                       'mac_address': '04:bc:9f:6e:6a:a9',
                                                                       'mtu': 2000,
                                                                       'name': '284845/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2032364',
                                                                       'out_octets': '1300019208686',
                                                                       'out_ucast_pkts': '893205539',
                                                                       'poller_hash': 'MTI5NjBEOHBlZXJpbnRmMjg0ODQ1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '284845',
                                                                       'in_nucast_pkts': '2032364',
                                                                       'in_octets': '1300019208686',
                                                                       'in_ucast_pkts': '893205539',
                                                                       'links': [       {       'in_nucast_pkts': '2032364',
                                                                                                'in_octets': '1300019208686',
                                                                                                'in_ucast_pkts': '893205539',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16778',
                                                                                                'out_octets': '35938820116',
                                                                                                'out_ucast_pkts': '284852556',
                                                                                                'poller_hash': 'MjExNDEwNjAxMzM0Z3BvbiAxMS8z'}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16778',
                                                                       'out_octets': '35938820116',
                                                                       'out_ucast_pkts': '284852556',
                                                                       'poller_hash': 'MTI5NjBEOHBlZXJpbnRmcG9uIDI4NDg0NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '284845',
                                                                                               'links': [       {       'connected_time': '951398',
                                                                                                                        'fiber_length': 17577,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMzM0cG9uZ3BvbiAxMS8z',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjBEOHBlZXJwb25wb24gMjg0ODQ1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '208 WINSTANLEY ST (KAREN AND GORD '
                                                 'BENDER)',
                                         'poller_hash': 'MTI5NjBEOA==',
                                         'serial_number': '12960D8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6a:a6',
                                         'uptime': '951398'},
                                 {       'counter_source': '286284/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:01:b9:e1'],
                                                                       'id': '286284/g1',
                                                                       'in_nucast_pkts': '7349',
                                                                       'in_octets': '14918303722',
                                                                       'in_ucast_pkts': '32145046',
                                                                       'mac_address': '04:bc:9f:6e:58:91',
                                                                       'mtu': 2000,
                                                                       'name': '286284/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '723590',
                                                                       'out_octets': '89931862425',
                                                                       'out_ucast_pkts': '74636195',
                                                                       'poller_hash': 'MTI5NUU5NXBlZXJpbnRmMjg2Mjg0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '286284',
                                                                       'in_nucast_pkts': '723590',
                                                                       'in_octets': '89931862425',
                                                                       'in_ucast_pkts': '74636195',
                                                                       'links': [       {       'in_nucast_pkts': '723590',
                                                                                                'in_octets': '89931862425',
                                                                                                'in_ucast_pkts': '74636195',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7349',
                                                                                                'out_octets': '14918303722',
                                                                                                'out_ucast_pkts': '32145046',
                                                                                                'poller_hash': 'MjExNTEyNjAzNjgxZ3BvbiA0LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7349',
                                                                       'out_octets': '14918303722',
                                                                       'out_ucast_pkts': '32145046',
                                                                       'poller_hash': 'MTI5NUU5NXBlZXJpbnRmcG9uIDI4NjI4NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '286284',
                                                                                               'links': [       {       'connected_time': '351587',
                                                                                                                        'fiber_length': 3064,
                                                                                                                        'poller_hash': 'MjExNTEyNjAzNjgxcG9uZ3BvbiA0LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU5NXBlZXJwb25wb24gMjg2Mjg0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '58 AVERY PL (KARDAS BRANT)',
                                         'poller_hash': 'MTI5NUU5NQ==',
                                         'serial_number': '1295E95',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:58:8e',
                                         'uptime': '351587'},
                                 {       'counter_source': '286336/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '286336/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5421787',
                                                                       'in_ucast_pkts': '21112',
                                                                       'mac_address': '04:bc:9f:6e:64:f6',
                                                                       'name': '286336/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '221033891',
                                                                       'out_ucast_pkts': '175166',
                                                                       'poller_hash': 'MTI5NjAyMnBlZXJpbnRmMjg2MzM2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '286336',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '221033891',
                                                                       'in_ucast_pkts': '175166',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '221033891',
                                                                                                'in_ucast_pkts': '175166',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '5421787',
                                                                                                'out_ucast_pkts': '21112',
                                                                                                'poller_hash': 'MjExNDEwNjAxMjg5Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '5421787',
                                                                       'out_ucast_pkts': '21112',
                                                                       'poller_hash': 'MTI5NjAyMnBlZXJpbnRmcG9uIDI4NjMzNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '286336',
                                                                                               'links': [       {       'connected_time': '8527',
                                                                                                                        'fiber_length': 21602,
                                                                                                                        'poller_hash': 'MjExNDEwNjAxMjg5cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjAyMnBlZXJwb25wb24gMjg2MzM2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4957 ROAD 109 (ROTH CHRIS)',
                                         'poller_hash': 'MTI5NjAyMg==',
                                         'serial_number': '1296022',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:64:f6',
                                         'uptime': '8527'}]
        }
        id_field = 'mac_address' if not obj.stations else ('mac_address' if hasattr(obj.stations[0], 'mac_address') and getattr(obj.stations[0], 'mac_address') else ('imsi' if hasattr(obj.stations[0], 'imsi') else 'serial'))
        cur_info = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('info', 'ap', obj.model) + [('vendor', 'str'), ('model', 'str'), ('sw_version', 'str')]}
        if 'rf_freq' in cur_info and cur_info.get('rf_freq') and (cur_info['rf_freq'] <= 0 or int.bit_length(cur_info['rf_freq']) >= 32):
            cur_info.pop('rf_freq')
        cur_stats = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'ap', obj.model)}
        cur_sta_stats = {getattr(sta, id_field): {x: getattr(sta, x, None) for (x, data_type) in sorted(self.apdd.get_metric_fields('stats', 'cpe', obj.model)) + [(id_field, 'str'), ('ip_address', 'str'), ('name', 'str'), ('dev_macs', 'str')]} for sta in sorted(obj.stations)}
        cur_ne_msg = getattr(obj, 'network_element_update_msg', None)
        cur_ne_msg = cur_ne_msg.pb if cur_ne_msg else None
        self.assertEqual(cur_info, exp_info)
        self.assertEqual(cur_stats, exp_stats)
        self.assertEqual(cur_sta_stats, exp_sta_stats)
        self.assertTrue(self.compare_pb_dict(cur_ne_msg, exp_ne_msg))

    def test_calix_exa_E7_20_3_4_51_25_02(self):
        obj = self.run_data("calix.exa.E7-20.3.4.51.25.02")
        exp_info = {
                'model': 'E7-20',
                'sta_count': 795,
                'sw_version': '3.4.51.25',
                'vendor': 'Calix'
        }
        exp_stats = {
         'serial_number': '71412913076', 'system_mac_address': None
        }
        exp_sta_stats = {
                None: {       'dev_macs': [],
                              'ip_address': None,
                              'mac_address': None,
                              'name': '4903 Perth Rd. 119',
                              'serial_number': 'F100',
                              'system_mac_address': None},
                '00:06:31:08:49:ae': {       'dev_macs': ['d8:ec:5e:f6:14:31'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:49:ae',
                                             'name': '4881 Tiffany Ct. Gadshill (Jim '
                                                     'Brubacher)',
                                             'serial_number': 'F0F5',
                                             'system_mac_address': None},
                '00:06:31:08:49:e7': {       'dev_macs': ['cc:be:59:c3:47:fc'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:49:e7',
                                             'name': '4917 Linhaven Dr. (Shawn '
                                                     'Gilhula)',
                                             'serial_number': 'F0FF',
                                             'system_mac_address': None},
                '00:06:31:08:49:ed': {       'dev_macs': ['5c:0f:fb:2a:0a:08'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:49:ed',
                                             'name': '4910 Perth Road 119 (John '
                                                     'Dunsmore)',
                                             'serial_number': 'F0FE',
                                             'system_mac_address': None},
                '00:06:31:08:4a:05': {       'dev_macs': ['d0:76:8f:1f:55:6d'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:4a:05',
                                             'name': '4900 Perth Road 119 (Jason&Trish '
                                                     'Bowles)',
                                             'serial_number': 'F0F6',
                                             'system_mac_address': None},
                '00:06:31:08:4a:80': {       'dev_macs': ['9c:a2:f4:5f:23:41'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:4a:80',
                                             'name': '4002 PERTH ROAD 56 (WILLIAMSON '
                                                     'DOUGLAS & BETTY)',
                                             'serial_number': 'F328',
                                             'system_mac_address': None},
                '00:06:31:08:4c:42': {       'dev_macs': ['0c:80:63:f6:b9:68'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:4c:42',
                                             'name': '4902 Perth Road 119 (Steve '
                                                     'Winkel)',
                                             'serial_number': 'F436',
                                             'system_mac_address': None},
                '00:06:31:08:4c:90': {       'dev_macs': ['3c:84:6a:75:21:53'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:4c:90',
                                             'name': '4916 Perth Road 119 (Mark '
                                                     'Walker)',
                                             'serial_number': 'F431',
                                             'system_mac_address': None},
                '00:06:31:08:4d:71': {       'dev_macs': ['ec:4f:82:12:f2:d4'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:4d:71',
                                             'name': '4888 Linehaven Dr. (J. Bolhuis '
                                                     '10423884)',
                                             'serial_number': 'F43B',
                                             'system_mac_address': None},
                '00:06:31:08:da:eb': {       'dev_macs': ['94:2a:6f:12:58:bc'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:08:da:eb',
                                             'name': '2835 Line 45 (TMP International '
                                                     '10425548)',
                                             'serial_number': '153CB',
                                             'system_mac_address': None},
                '00:06:31:11:ee:d7': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:11:ee:d7',
                                             'name': '4017 Perth Line 56-1 (Kenneth '
                                                     'Jantzi 10432211)',
                                             'serial_number': '130AD',
                                             'system_mac_address': None},
                '00:06:31:11:ef:0a': {       'dev_macs': ['e8:9c:25:5b:16:d8'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:11:ef:0a',
                                             'name': '5451 Perth Rd. 119 B (Marvin '
                                                     'Good)',
                                             'serial_number': '130A2',
                                             'system_mac_address': None},
                '00:06:31:11:ef:dc': {       'dev_macs': ['a8:63:7d:4e:fe:c0'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:11:ef:dc',
                                             'name': '4921 Perth Road 119 (Stephen '
                                                     'Weadick)',
                                             'serial_number': '1303E',
                                             'system_mac_address': None},
                '00:06:31:11:f0:12': {       'dev_macs': ['74:da:88:25:b9:4b'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:11:f0:12',
                                             'name': '4906 Linhaven Dr. (Darly '
                                                     'Iszczuk)',
                                             'serial_number': '13048',
                                             'system_mac_address': None},
                '00:06:31:15:a9:55': {       'dev_macs': ['0c:9d:92:ce:56:f0'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:15:a9:55',
                                             'name': '4927 Perth Road 119 (Ivan '
                                                     'Messerschmidt)',
                                             'serial_number': '3D0FB',
                                             'system_mac_address': None},
                '00:06:31:18:c1:2e': {       'dev_macs': ['84:16:f9:4f:27:84'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:18:c1:2e',
                                             'name': '4919 Linhaven Dr. Gadshill (Ken '
                                                     'Huether)',
                                             'serial_number': '3CFAE',
                                             'system_mac_address': None},
                '00:06:31:1f:08:e5': {       'dev_macs': ['f0:76:1c:7a:84:d8'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:1f:08:e5',
                                             'name': '2860 Line 45 (Robert St. Pierre)',
                                             'serial_number': '41548',
                                             'system_mac_address': None},
                '00:06:31:20:01:68': {       'dev_macs': ['60:db:98:55:b7:88'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:20:01:68',
                                             'name': '4872 Perth Road 119 (Jeremy&Dana '
                                                     'Satchell)',
                                             'serial_number': '415A7',
                                             'system_mac_address': None},
                '00:06:31:3c:cc:ef': {       'dev_macs': ['60:db:98:09:1e:fb'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:3c:cc:ef',
                                             'name': '2844 Line 45 (Gordon Hill)',
                                             'serial_number': '86C07',
                                             'system_mac_address': None},
                '00:06:31:5f:68:cf': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:5f:68:cf',
                                             'name': '4006 Andy St. Gadshill (Jen '
                                                     'Rose)',
                                             'serial_number': 'C580F',
                                             'system_mac_address': None},
                '00:06:31:b7:5c:46': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:b7:5c:46',
                                             'name': '4892 LINHAVEN DR (READINGS JOE)',
                                             'serial_number': '16427F',
                                             'system_mac_address': None},
                '00:06:31:e3:fa:e3': {       'dev_macs': ['14:21:03:d1:8b:22'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:e3:fa:e3',
                                             'name': '4959 PERTH ROAD 119 (FERNADES '
                                                     'ROMUALD)',
                                             'serial_number': '1999F5',
                                             'system_mac_address': None},
                '00:06:31:e3:fd:13': {       'dev_macs': ['44:65:7f:41:94:90'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:e3:fd:13',
                                             'name': '4984 Perth Rd. 119 Gadshill '
                                                     '(Daniel Kuepfer 10435144)',
                                             'serial_number': '199A3B',
                                             'system_mac_address': None},
                '00:06:31:e8:f3:ad': {       'dev_macs': ['60:db:98:82:01:dc'],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:e8:f3:ad',
                                             'name': '4883 TIFFANY CT (WILLIG MIKE)',
                                             'serial_number': '1D04FF',
                                             'system_mac_address': None},
                '00:06:31:e9:01:3f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '00:06:31:e9:01:3f',
                                             'name': '1 PONDVIEW UNIT 102 (SHANTZ ROSS '
                                                     '10436359)',
                                             'serial_number': '1D0681',
                                             'system_mac_address': None},
                '04:bc:9f:22:fa:c3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:22:fa:c3',
                                             'name': 'u4g-test',
                                             'serial_number': '151E4BC',
                                             'system_mac_address': None},
                '04:bc:9f:36:53:34': {       'dev_macs': ['9c:05:d6:ab:d9:f7'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:36:53:34',
                                             'name': '4011 Perth Line 56 (Elgin Roth '
                                                     'Transport)',
                                             'serial_number': '100A934',
                                             'system_mac_address': None},
                '04:bc:9f:5a:35:91': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5a:35:91',
                                             'name': '4 BERDUX PL (GRILLI JOHN)',
                                             'serial_number': '1270D32',
                                             'system_mac_address': None},
                '04:bc:9f:5a:36:59': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5a:36:59',
                                             'name': '14 FIRELLA PL (LICHTY TODD)',
                                             'serial_number': '1270D4B',
                                             'system_mac_address': None},
                '04:bc:9f:5a:37:69': {       'dev_macs': ['cc:be:59:32:95:16'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5a:37:69',
                                             'name': '1 POND VIEW DR APT 301 (STEINMAN '
                                                     'ANNA MAE)',
                                             'serial_number': '1270D6D',
                                             'system_mac_address': None},
                '04:bc:9f:5a:3e:91': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5a:3e:91',
                                             'name': '2859 LINE 45 (FAULHAFER ROBERT & '
                                                     'DOROTHEA)',
                                             'serial_number': '1270E52',
                                             'system_mac_address': None},
                '04:bc:9f:5b:9b:d9': {       'dev_macs': ['e0:46:ee:1e:4a:4f'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:9b:d9',
                                             'name': '1818 MOSER-YOUNG RD (DOYLE '
                                                     'MATTHEW & ALYSSA)',
                                             'serial_number': '12739FB',
                                             'system_mac_address': None},
                '04:bc:9f:5b:9d:89': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:9d:89',
                                             'name': '34 GALENA ST (SEHER STEFAN)',
                                             'serial_number': '1273A31',
                                             'system_mac_address': None},
                '04:bc:9f:5b:a6:11': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:a6:11',
                                             'name': '4021 PERTH LINE 56 (KIPP DEBBIE '
                                                     '& NATHAN)',
                                             'serial_number': '1273B42',
                                             'system_mac_address': None},
                '04:bc:9f:5b:a6:31': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:a6:31',
                                             'name': '4980 PERTH ROAD 119 (MCMURRAY '
                                                     'STEVE & WENDY)',
                                             'serial_number': '1273B46',
                                             'system_mac_address': None},
                '04:bc:9f:5b:c6:c1': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:c6:c1',
                                             'name': '162 LAWRENCE ST WELLESLEY '
                                                     '(STENDEL TROY)',
                                             'serial_number': '1273F58',
                                             'system_mac_address': None},
                '04:bc:9f:5b:cd:c1': {       'dev_macs': ['14:21:03:86:d0:a8'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:cd:c1',
                                             'name': '19 CRAB APPLE CT (DIETRICH '
                                                     'LINDSEY)',
                                             'serial_number': '1274038',
                                             'system_mac_address': None},
                '04:bc:9f:5b:ce:41': {       'dev_macs': ['d0:76:8f:1f:6b:2b'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:ce:41',
                                             'name': '175 FERRIS DRIVE (WILSON GEOFF)',
                                             'serial_number': '1274048',
                                             'system_mac_address': None},
                '04:bc:9f:5b:ce:b9': {       'dev_macs': ['14:21:03:8b:73:86'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:ce:b9',
                                             'name': '50 SCHWEITZER CR (MOORE DARRYL '
                                                     'W)',
                                             'serial_number': '1274057',
                                             'system_mac_address': None},
                '04:bc:9f:5b:ce:f1': {       'dev_macs': ['b8:94:70:f3:e9:e1'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:ce:f1',
                                             'name': '60 PARKVIEW DR (CHATFIELD '
                                                     'RICHARD)',
                                             'serial_number': '127405E',
                                             'system_mac_address': None},
                '04:bc:9f:5b:cf:79': {       'dev_macs': ['14:21:03:8b:78:56'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:cf:79',
                                             'name': '64 POND VIEW DR (DELUTIS GREG)',
                                             'serial_number': '127406F',
                                             'system_mac_address': None},
                '04:bc:9f:5b:d4:41': {       'dev_macs': ['14:21:03:8b:68:c8'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:d4:41',
                                             'name': '4576 WEIMAR LINE (WILZ HERMANN & '
                                                     'LISA)',
                                             'serial_number': '1274108',
                                             'system_mac_address': None},
                '04:bc:9f:5b:d8:71': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:d8:71',
                                             'name': '3637 NAFZIGER RD (HUNT KEVIN)',
                                             'serial_number': '127418E',
                                             'system_mac_address': None},
                '04:bc:9f:5b:d8:a9': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:d8:a9',
                                             'name': '3771 NAFZIGER RD (FALK RON)',
                                             'serial_number': '1274195',
                                             'system_mac_address': None},
                '04:bc:9f:5b:d9:e9': {       'dev_macs': ['7c:10:c9:02:8d:70'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:d9:e9',
                                             'name': '4931 PERTH ROAD 119 (STECKLEY '
                                                     'TERRY)',
                                             'serial_number': '12741BD',
                                             'system_mac_address': None},
                '04:bc:9f:5b:e1:81': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:e1:81',
                                             'name': '4957 PERTH ROAD 119 (KOCHER DON '
                                                     '& MARY)',
                                             'serial_number': '12742B0',
                                             'system_mac_address': None},
                '04:bc:9f:5b:e1:c1': {       'dev_macs': ['14:21:03:8b:79:11'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:e1:c1',
                                             'name': '146 LAWRENCE ST (HEAD MIKE)',
                                             'serial_number': '12742B8',
                                             'system_mac_address': None},
                '04:bc:9f:5b:e2:81': {       'dev_macs': ['14:21:03:8b:71:13'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:e2:81',
                                             'name': '19 WATER ST (FUTHER DAVID)',
                                             'serial_number': '12742D0',
                                             'system_mac_address': None},
                '04:bc:9f:5b:e3:79': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:e3:79',
                                             'name': '116 REINER CR (ROSE ANNA)',
                                             'serial_number': '12742EF',
                                             'system_mac_address': None},
                '04:bc:9f:5b:e4:b1': {       'dev_macs': ['14:21:03:8b:6a:28'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:e4:b1',
                                             'name': '83 GERBER MEADOWS DR (MARTIN '
                                                     'ANDREW J)',
                                             'serial_number': '1274316',
                                             'system_mac_address': None},
                '04:bc:9f:5b:e5:29': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:e5:29',
                                             'name': '1096 HENRY ST (GERBER EUNICE)',
                                             'serial_number': '1274325',
                                             'system_mac_address': None},
                '04:bc:9f:5b:f8:d1': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:f8:d1',
                                             'name': '1416 HUTCHISON RD (ROHR '
                                                     'ELECTRIC)',
                                             'serial_number': '127459A',
                                             'system_mac_address': None},
                '04:bc:9f:5b:fc:19': {       'dev_macs': ['14:21:03:d1:8a:f6'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:fc:19',
                                             'name': "1189 QUEEN'S BUSH RD (WAGNER "
                                                     'CARRIE)',
                                             'serial_number': '1274603',
                                             'system_mac_address': None},
                '04:bc:9f:5b:fd:79': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:fd:79',
                                             'name': '4920 PERTH ROAD 119 (BRANDON '
                                                     'ROB)',
                                             'serial_number': '127462F',
                                             'system_mac_address': None},
                '04:bc:9f:5b:ff:41': {       'dev_macs': ['14:21:03:d1:89:ac'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5b:ff:41',
                                             'name': "1221 QUEEN'S BUSH RD (DICASA "
                                                     'MICHAEL)',
                                             'serial_number': '1274668',
                                             'system_mac_address': None},
                '04:bc:9f:5c:00:89': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:00:89',
                                             'name': '40 ZINKANN CR (LANG TODD)',
                                             'serial_number': '1274691',
                                             'system_mac_address': None},
                '04:bc:9f:5c:00:b9': {       'dev_macs': ['5c:0f:fb:33:c1:92'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:00:b9',
                                             'name': '1060 MOLESWORTH ST (STROH TOM)',
                                             'serial_number': '1274697',
                                             'system_mac_address': None},
                '04:bc:9f:5c:01:31': {       'dev_macs': ['14:21:03:d1:7f:1a'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:01:31',
                                             'name': '107 GERBER MEADOWS DR (JANTZI '
                                                     'CHARLOTTE)',
                                             'serial_number': '12746A6',
                                             'system_mac_address': None},
                '04:bc:9f:5c:05:d9': {       'dev_macs': ['00:0d:b9:5d:9a:05'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:05:d9',
                                             'name': "1214 QUEEN'S BUSH RD (LEIS FEED "
                                                     '& SUPPLY LTD)',
                                             'serial_number': '127473B',
                                             'system_mac_address': None},
                '04:bc:9f:5c:05:f1': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:05:f1',
                                             'name': '146 DAVID ST (COMMUNITY FOOD '
                                                     'CUPBOARD)',
                                             'serial_number': '127473E',
                                             'system_mac_address': None},
                '04:bc:9f:5c:06:49': {       'dev_macs': ['ec:4f:82:2d:7d:dd'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:06:49',
                                             'name': '16 LAWRENCE ST (KOCH KIM)',
                                             'serial_number': '1274749',
                                             'system_mac_address': None},
                '04:bc:9f:5c:09:99': {       'dev_macs': ['14:21:03:d1:8c:1f'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:09:99',
                                             'name': '42 GALENA ST WELLESLEY (ROMANOVA '
                                                     'ELENA)',
                                             'serial_number': '12747B3',
                                             'system_mac_address': None},
                '04:bc:9f:5c:0a:a9': {       'dev_macs': ['14:21:03:d1:7c:86'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:0a:a9',
                                             'name': '29 WELWOOD AV (SCHLUETER BETH)',
                                             'serial_number': '12747D5',
                                             'system_mac_address': None},
                '04:bc:9f:5c:0a:d1': {       'dev_macs': ['d0:76:8f:02:f1:d8'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:0a:d1',
                                             'name': '84 POND VIEW DR (BOLLER PATRICK)',
                                             'serial_number': '12747DA',
                                             'system_mac_address': None},
                '04:bc:9f:5c:0a:e1': {       'dev_macs': ['14:21:03:d1:7e:ee'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:0a:e1',
                                             'name': '1416 HUTCHISON RD (HURREN '
                                                     'RICHARD)',
                                             'serial_number': '12747DC',
                                             'system_mac_address': None},
                '04:bc:9f:5c:0c:81': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:0c:81',
                                             'name': '1416 HUTCHISON RD (CENTRE LINE '
                                                     'CONSTRUCTION LTD)',
                                             'serial_number': '1274810',
                                             'system_mac_address': None},
                '04:bc:9f:5c:0c:b1': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:0c:b1',
                                             'name': '2060 PERTH LINE (JANTZI GABRIEL)',
                                             'serial_number': '1274816',
                                             'system_mac_address': None},
                '04:bc:9f:5c:0d:f9': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:0d:f9',
                                             'name': '1129 HENRY ST (INSPIRING MINDS '
                                                     'EARLY LEARNING)',
                                             'serial_number': '127483F',
                                             'system_mac_address': None},
                '04:bc:9f:5c:10:c9': {       'dev_macs': ['14:21:03:a9:75:22'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:10:c9',
                                             'name': '4903 PERTH ROAD 119 (ALLIN JODI)',
                                             'serial_number': '1274899',
                                             'system_mac_address': None},
                '04:bc:9f:5c:11:69': {       'dev_macs': ['84:d3:43:6d:ea:da'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:11:69',
                                             'name': '1086 HENRY ST (KOCH PAUL R)',
                                             'serial_number': '12748AD',
                                             'system_mac_address': None},
                '04:bc:9f:5c:11:a1': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:11:a1',
                                             'name': '5590 PERTH LINE 56A (KIPFER '
                                                     'JESSICA)',
                                             'serial_number': '12748B4',
                                             'system_mac_address': None},
                '04:bc:9f:5c:17:59': {       'dev_macs': ['14:21:03:d1:8c:6c'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:17:59',
                                             'name': '5 BERDUX PL (SCHWARTZ TREVOR)',
                                             'serial_number': '127496B',
                                             'system_mac_address': None},
                '04:bc:9f:5c:1c:e1': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:1c:e1',
                                             'name': '4913 LINHAVEN DR (PLAYER BILL & '
                                                     'CINDIE)',
                                             'serial_number': '1274A1C',
                                             'system_mac_address': None},
                '04:bc:9f:5c:20:09': {       'dev_macs': ['b8:94:70:f3:e3:c7'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:20:09',
                                             'name': '5565 KING ST (WILSON DAVE)',
                                             'serial_number': '1274A81',
                                             'system_mac_address': None},
                '04:bc:9f:5c:20:21': {       'dev_macs': ['b8:94:70:f3:ca:2b'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:20:21',
                                             'name': '4896 PERTH ROAD 119 (MOHR DAVID '
                                                     '& LYNDA)',
                                             'serial_number': '1274A84',
                                             'system_mac_address': None},
                '04:bc:9f:5c:22:39': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:22:39',
                                             'name': "1401 QUEEN'S BUSH RD (WOOLWICH "
                                                     'COMMUNITY HEALTH CENT)',
                                             'serial_number': '1274AC7',
                                             'system_mac_address': None},
                '04:bc:9f:5c:22:d1': {       'dev_macs': ['b8:94:70:9e:cd:d5'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:22:d1',
                                             'name': '61 LAWRENCE STREET (DOPKO MATT)',
                                             'serial_number': '1274ADA',
                                             'system_mac_address': None},
                '04:bc:9f:5c:24:21': {       'dev_macs': ['cc:be:59:c5:5d:18'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:24:21',
                                             'name': '37 ZINKANN CR WELLESLEY (DYSON '
                                                     'CATHERINE)',
                                             'serial_number': '1274B04',
                                             'system_mac_address': None},
                '04:bc:9f:5c:24:61': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:24:61',
                                             'name': '4876 PERTH ROAD 119 (ROESNER '
                                                     'NANCY)',
                                             'serial_number': '1274B0C',
                                             'system_mac_address': None},
                '04:bc:9f:5c:25:49': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:25:49',
                                             'name': '2848 LINE 45 GADSHILL (WALKER '
                                                     'LARRY)',
                                             'serial_number': '1274B29',
                                             'system_mac_address': None},
                '04:bc:9f:5c:27:89': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:27:89',
                                             'name': '4917 PERTH ROAD 119 (BROWN JOHN '
                                                     'T)',
                                             'serial_number': '1274B71',
                                             'system_mac_address': None},
                '04:bc:9f:5c:4a:79': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:4a:79',
                                             'name': '2088 PERTH LINE (LANTZ LORNE & '
                                                     'IRENE)',
                                             'serial_number': '1274FCF',
                                             'system_mac_address': None},
                '04:bc:9f:5c:4c:39': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:4c:39',
                                             'name': '2086 PERTH LINE 56',
                                             'serial_number': '1275007',
                                             'system_mac_address': None},
                '04:bc:9f:5c:50:09': {       'dev_macs': ['b8:94:70:f3:e0:83'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:50:09',
                                             'name': '1 SHADYWOOD CT (FRASER SHEILA)',
                                             'serial_number': '1275081',
                                             'system_mac_address': None},
                '04:bc:9f:5c:51:c9': {       'dev_macs': ['e4:6c:d1:4f:ac:9a'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:51:c9',
                                             'name': '26 FERRIS DR (NAFZIGER JAMES)',
                                             'serial_number': '12750B9',
                                             'system_mac_address': None},
                '04:bc:9f:5c:52:21': {       'dev_macs': ['b8:94:70:f3:e0:99'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:52:21',
                                             'name': '3710 NAFZIGER RD (LENG MILDRED)',
                                             'serial_number': '12750C4',
                                             'system_mac_address': None},
                '04:bc:9f:5c:53:39': {       'dev_macs': ['00:01:21:06:75:ef'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:53:39',
                                             'name': "1401 QUEEN'S BUSH ROAD (TOWNSHIP "
                                                     'OF WELLESLEY)',
                                             'serial_number': '12750E7',
                                             'system_mac_address': None},
                '04:bc:9f:5c:56:41': {       'dev_macs': ['b8:94:70:f3:ec:96'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:56:41',
                                             'name': '4918 LINHAVEN DR (SCHWARTZ '
                                                     'TREVOR)',
                                             'serial_number': '1275148',
                                             'system_mac_address': None},
                '04:bc:9f:5c:58:91': {       'dev_macs': ['d0:76:8f:1f:64:1f'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:58:91',
                                             'name': "1139 QUEEN'S BUSH RD (HARDY "
                                                     'FRANK)',
                                             'serial_number': '1275192',
                                             'system_mac_address': None},
                '04:bc:9f:5c:5a:21': {       'dev_macs': ['b8:94:70:f3:e5:b6'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:5c:5a:21',
                                             'name': '6 EDGEWOOD CT (GUENTHER JOANNE)',
                                             'serial_number': '12751C4',
                                             'system_mac_address': None},
                '04:bc:9f:6e:48:06': {       'dev_macs': ['e4:6c:d1:4e:e9:4f'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:48:06',
                                             'name': 'ont 282243',
                                             'serial_number': '1295C84',
                                             'system_mac_address': None},
                '04:bc:9f:6e:56:5e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:56:5e',
                                             'name': '4940 LINHAVEN DR GADSHILL (RIEHL '
                                                     'TIM & CASEY)',
                                             'serial_number': '1295E4F',
                                             'system_mac_address': None},
                '04:bc:9f:6e:57:ee': {       'dev_macs': ['e4:6c:d1:4f:b7:79'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:57:ee',
                                             'name': '138 LAWRENCE ST (MORAN JOHN)',
                                             'serial_number': '1295E81',
                                             'system_mac_address': None},
                '04:bc:9f:6e:58:46': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:58:46',
                                             'name': "1220 QUEEN'S BUSH RD (WELLESLEY "
                                                     'SERVICE CENTRE LTD)',
                                             'serial_number': '1295E8C',
                                             'system_mac_address': None},
                '04:bc:9f:6e:58:ce': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:58:ce',
                                             'name': '4908 PERTH ROAD 119 (GARY '
                                                     'PATTERSON)',
                                             'serial_number': '1295E9D',
                                             'system_mac_address': None},
                '04:bc:9f:6e:5a:06': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:5a:06',
                                             'name': '4030 PERTH LINE 56 (GERBER '
                                                     'PHILIP & ESTHER)',
                                             'serial_number': '1295EC4',
                                             'system_mac_address': None},
                '04:bc:9f:6e:5a:7e': {       'dev_macs': ['04:bc:9f:46:8f:99'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:5a:7e',
                                             'name': '1870 PERTH LINE (BRENNAN ANNE)',
                                             'serial_number': '1295ED3',
                                             'system_mac_address': None},
                '04:bc:9f:6e:62:ae': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:62:ae',
                                             'name': '4908 LINHAVEN DR GADSHILL '
                                                     '(PINKNEY BRANDON)',
                                             'serial_number': '1295FD9',
                                             'system_mac_address': None},
                '04:bc:9f:6e:62:d6': {       'dev_macs': ['60:db:98:08:ff:7c'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:62:d6',
                                             'name': '8 POND VIEW DR (SCHELTER-BROWN '
                                                     'BONNE)',
                                             'serial_number': '1295FDE',
                                             'system_mac_address': None},
                '04:bc:9f:6e:63:56': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:63:56',
                                             'name': '2076 ALBERT ST (DETZLER A & '
                                                     'SMITH C)',
                                             'serial_number': '1295FEE',
                                             'system_mac_address': None},
                '04:bc:9f:6e:63:8e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:63:8e',
                                             'name': '4898 LINHAVEN DR (SMITS JAMES)',
                                             'serial_number': '1295FF5',
                                             'system_mac_address': None},
                '04:bc:9f:6e:6c:6e': {       'dev_macs': ['14:21:03:a9:74:a9'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:6c:6e',
                                             'name': '3634 NAFZIGER RD (SCHULTZ EMILY)',
                                             'serial_number': '1296111',
                                             'system_mac_address': None},
                '04:bc:9f:6e:db:8e': {       'dev_macs': ['00:0c:29:e3:f7:44'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6e:db:8e',
                                             'name': '14 JACK KOEHLER LN (MACMILLAN '
                                                     'BRAD)',
                                             'serial_number': '1296EF5',
                                             'system_mac_address': None},
                '04:bc:9f:6f:06:fe': {       'dev_macs': ['14:21:03:a9:72:d0'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6f:06:fe',
                                             'name': '16 WELWOOD AV (HARNACK ROBERT)',
                                             'serial_number': '1297463',
                                             'system_mac_address': None},
                '04:bc:9f:6f:07:4e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6f:07:4e',
                                             'name': '7 DAVID ST (JUKES CALEB)',
                                             'serial_number': '129746D',
                                             'system_mac_address': None},
                '04:bc:9f:6f:08:ce': {       'dev_macs': ['14:21:03:a9:6d:c9'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6f:08:ce',
                                             'name': '36 WELWOOD AV (GRUBER JEFF)',
                                             'serial_number': '129749D',
                                             'system_mac_address': None},
                '04:bc:9f:6f:08:fe': {       'dev_macs': ['14:21:03:a9:71:86'],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6f:08:fe',
                                             'name': '37 LAWRENCE ST (VAN DER MAAS '
                                                     'PETER)',
                                             'serial_number': '12974A3',
                                             'system_mac_address': None},
                '04:bc:9f:6f:0a:66': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6f:0a:66',
                                             'name': '1416 HUTCHISON RD (DOERING '
                                                     'ELECTRIC INC)',
                                             'serial_number': '12974D0',
                                             'system_mac_address': None},
                '04:bc:9f:6f:0b:3e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '04:bc:9f:6f:0b:3e',
                                             'name': '2018 PERTH LINE (SNELL EMILY)',
                                             'serial_number': '12974EB',
                                             'system_mac_address': None},
                '10:74:c5:fe:c0:54': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '10:74:c5:fe:c0:54',
                                             'name': '4200 (Test)',
                                             'serial_number': '1C888C6',
                                             'system_mac_address': None},
                '1c:8b:76:00:31:c9': {       'dev_macs': ['d4:6a:91:b7:c6:72'],
                                             'ip_address': None,
                                             'mac_address': '1c:8b:76:00:31:c9',
                                             'name': '3685 NAFZIGER RD (SCHMIDTSVILLE '
                                                     'RESTURANT)',
                                             'serial_number': '158FAAF',
                                             'system_mac_address': None},
                '1c:8b:76:00:37:71': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '1c:8b:76:00:37:71',
                                             'name': '3546 NAFZIGER RD (WELLESLEY '
                                                     'WOODWORKS)',
                                             'serial_number': '158FB64',
                                             'system_mac_address': None},
                '44:65:7f:33:0c:c8': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:0c:c8',
                                             'name': '44 VILLAGE RD',
                                             'serial_number': '6400EC',
                                             'system_mac_address': None},
                '44:65:7f:33:0c:d6': {       'dev_macs': ['48:77:46:9a:4e:87'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:0c:d6',
                                             'name': '1087 HENRY ST (Hugh Cameron)',
                                             'serial_number': '6400F3',
                                             'system_mac_address': None},
                '44:65:7f:33:0d:0e': {       'dev_macs': ['44:65:7f:38:03:55'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:0d:0e',
                                             'name': '107 Village Rd Wellesley (SHEWAN '
                                                     'KELLY)',
                                             'serial_number': '64010E',
                                             'system_mac_address': None},
                '44:65:7f:33:0d:d0': {       'dev_macs': ['cc:32:e5:54:05:0a'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:0d:d0',
                                             'name': "1144 QUEEN'S BUSH RD (JACOBS "
                                                     'CORY)',
                                             'serial_number': '64016E',
                                             'system_mac_address': None},
                '44:65:7f:33:20:86': {       'dev_macs': ['9c:3d:cf:4c:02:9c'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:20:86',
                                             'name': '71 VILLAGE RD (LEBOLD STEVE)',
                                             'serial_number': '640ABD',
                                             'system_mac_address': None},
                '44:65:7f:33:20:96': {       'dev_macs': ['6c:cd:d6:f8:ee:7a'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:20:96',
                                             'name': "1015 Queen's Bush Wellesley "
                                                     '(Scott Kittell 10435598)',
                                             'serial_number': '640AC5',
                                             'system_mac_address': None},
                '44:65:7f:33:20:98': {       'dev_macs': ['44:65:7f:88:d3:ea'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:20:98',
                                             'name': '17 Parkview Dr Wellesley '
                                                     '(Jo-Anne Kropf 10435560)',
                                             'serial_number': '640AC6',
                                             'system_mac_address': None},
                '44:65:7f:33:20:c4': {       'dev_macs': ['48:77:46:9a:52:80'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:20:c4',
                                             'name': "1122 QUEEN'S BUSH RD (DIETZ JON "
                                                     'W)',
                                             'serial_number': '640ADC',
                                             'system_mac_address': None},
                '44:65:7f:33:20:fc': {       'dev_macs': ['bc:0f:9a:70:bc:6b'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:20:fc',
                                             'name': '1023 Maple Leaf Lane Wellesley '
                                                     '(MARTIN LUCILLE)',
                                             'serial_number': '640AF6',
                                             'system_mac_address': None},
                '44:65:7f:33:21:0c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:0c',
                                             'name': '87 FERRIS DR (MACDONALD CALUM)',
                                             'serial_number': '640AFE',
                                             'system_mac_address': None},
                '44:65:7f:33:21:24': {       'dev_macs': ['48:77:46:9a:50:fd'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:24',
                                             'name': '4 SHADYWOOD CT (BAUMAN CHARLES)',
                                             'serial_number': '640B0A',
                                             'system_mac_address': None},
                '44:65:7f:33:21:32': {       'dev_macs': ['cc:f4:11:45:6e:c1'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:32',
                                             'name': '5447 KING ST (WURDELL PAUL)',
                                             'serial_number': '640B11',
                                             'system_mac_address': None},
                '44:65:7f:33:21:34': {       'dev_macs': ['48:77:46:9a:62:25'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:34',
                                             'name': '15 NITHBANK RD (ELLIOTT '
                                                     'ADRIENNE)',
                                             'serial_number': '640B12',
                                             'system_mac_address': None},
                '44:65:7f:33:21:44': {       'dev_macs': ['b0:95:75:b3:56:8e'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:44',
                                             'name': '24 WATER ST (PETERS BRYAN)',
                                             'serial_number': '640B1A',
                                             'system_mac_address': None},
                '44:65:7f:33:21:48': {       'dev_macs': ['d8:07:b6:a8:fa:99'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:48',
                                             'name': '28 REINER CR (SNYDER SCOTT)',
                                             'serial_number': '640B1C',
                                             'system_mac_address': None},
                '44:65:7f:33:21:56': {       'dev_macs': ['58:cb:52:de:3d:2e'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:56',
                                             'name': '35 Ferris Dr Wellesley (Natalie '
                                                     'Parco 10436039)',
                                             'serial_number': '640B23',
                                             'system_mac_address': None},
                '44:65:7f:33:21:58': {       'dev_macs': ['44:65:7f:88:47:bf'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:58',
                                             'name': '3850 Nafziger Rd Wellesley '
                                                     '(Irene Cook 10434806)',
                                             'serial_number': '640B24',
                                             'system_mac_address': None},
                '44:65:7f:33:21:5e': {       'dev_macs': ['f8:27:2e:06:3d:f6'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:5e',
                                             'name': '1126 HENRY ST (CALDWELL MICHAEL)',
                                             'serial_number': '640B27',
                                             'system_mac_address': None},
                '44:65:7f:33:21:92': {       'dev_macs': ['44:65:7f:89:0e:8e'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:92',
                                             'name': "1316 Queen's Bush Rd Wellesley "
                                                     '(Josh Zehr 10434844)',
                                             'serial_number': '640B41',
                                             'system_mac_address': None},
                '44:65:7f:33:21:ac': {       'dev_macs': ['44:65:7f:89:0c:cc'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:ac',
                                             'name': '1099 Henry St Wellesley (Jeff '
                                                     'Hill 10435307)',
                                             'serial_number': '640B4E',
                                             'system_mac_address': None},
                '44:65:7f:33:21:b4': {       'dev_macs': ['d0:76:8f:02:f0:ba'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:b4',
                                             'name': '3701 NAFZIGER RD (ROPP COURTNEY)',
                                             'serial_number': '640B52',
                                             'system_mac_address': None},
                '44:65:7f:33:21:e2': {       'dev_macs': ['cc:be:59:7e:b7:3d'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:e2',
                                             'name': '32 Zinkann Cr Wellesley (Jeff '
                                                     'Kneller 10435335)',
                                             'serial_number': '640B69',
                                             'system_mac_address': None},
                '44:65:7f:33:21:f8': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:21:f8',
                                             'name': '13 VILLAGE RD',
                                             'serial_number': '640B74',
                                             'system_mac_address': None},
                '44:65:7f:33:22:06': {       'dev_macs': ['44:65:7f:72:7e:0c'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:22:06',
                                             'name': '11 Reiner Cr Wellesley (Andrew '
                                                     'Scott 10434900)',
                                             'serial_number': '640B7A',
                                             'system_mac_address': None},
                '44:65:7f:33:22:0c': {       'dev_macs': ['b4:75:0e:4c:93:28'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:22:0c',
                                             'name': '105 Brown St Wellesley (Beverly '
                                                     'Meyer 10435259)',
                                             'serial_number': '640B7D',
                                             'system_mac_address': None},
                '44:65:7f:33:22:1a': {       'dev_macs': ['30:fd:38:f4:5f:dc'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:33:22:1a',
                                             'name': '3648 Nafziger Rd Unit B '
                                                     'Wellesley (Shawn Beggs 10435918)',
                                             'serial_number': '640B84',
                                             'system_mac_address': None},
                '44:65:7f:7b:b9:c9': {       'dev_macs': ['d0:76:8f:02:f0:af'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:b9:c9',
                                             'name': "1020 QUEEN'S BUSH RD "
                                                     '(WESTMORLAND BRENDAN)',
                                             'serial_number': '5A62DE',
                                             'system_mac_address': None},
                '44:65:7f:7b:c2:ea': {       'dev_macs': ['44:65:7f:88:d3:cf'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:c2:ea',
                                             'name': "1308 Queen's Bush Rd Wellesley "
                                                     '(Carrie Harman 10435242)',
                                             'serial_number': '5A65E9',
                                             'system_mac_address': None},
                '44:65:7f:7b:c2:f6': {       'dev_macs': ['d0:60:8c:03:51:4c'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:c2:f6',
                                             'name': '63 Village Rd Wellesley (Xinbin '
                                                     'Zhou 10434842)',
                                             'serial_number': '5A65ED',
                                             'system_mac_address': None},
                '44:65:7f:7b:c3:aa': {       'dev_macs': ['44:65:7f:38:09:6a'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:c3:aa',
                                             'name': '8 Edgewood Cr Wellesley (Roger '
                                                     'Zehr 10434453)',
                                             'serial_number': '5A6629',
                                             'system_mac_address': None},
                '44:65:7f:7b:c4:4c': {       'dev_macs': ['14:21:03:8b:73:a7'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:c4:4c',
                                             'name': '18 Parkview Dr Wellesley '
                                                     '(Matthew Jones 10435652)',
                                             'serial_number': '5A665F',
                                             'system_mac_address': None},
                '44:65:7f:7b:c4:7c': {       'dev_macs': ['44:65:7f:89:0d:da'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:c4:7c',
                                             'name': '1420 Hutchinson Rd Wellesley (K '
                                                     '& K Liquidation 10435446)',
                                             'serial_number': '5A666F',
                                             'system_mac_address': None},
                '44:65:7f:7b:c4:a3': {       'dev_macs': ['00:62:ec:3b:6c:57'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:c4:a3',
                                             'name': '2076 WENDELL CT (BOGART SAM)',
                                             'serial_number': '5A667C',
                                             'system_mac_address': None},
                '44:65:7f:7b:ce:63': {       'dev_macs': ['bc:a5:11:00:fd:5a'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:ce:63',
                                             'name': '1107 HENRY ST (SOULLIERE CINDY)',
                                             'serial_number': '5A69BC',
                                             'system_mac_address': None},
                '44:65:7f:7b:d0:22': {       'dev_macs': ['d4:be:d9:85:3f:5d'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d0:22',
                                             'name': '125 DAVID ST (LUDINGTON GORDON)',
                                             'serial_number': '5A6A51',
                                             'system_mac_address': None},
                '44:65:7f:7b:d3:76': {       'dev_macs': ['60:38:e0:a8:30:8d'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d3:76',
                                             'name': '1834 MOSER-YOUNG RD (MAINPRIZE '
                                                     'PAUL)',
                                             'serial_number': '5A6B6D',
                                             'system_mac_address': None},
                '44:65:7f:7b:d3:c7': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d3:c7',
                                             'name': '1810 MOSER-YOUNG RD (SPURR '
                                                     'BRENT)',
                                             'serial_number': '5A6B88',
                                             'system_mac_address': None},
                '44:65:7f:7b:d3:cd': {       'dev_macs': ['24:05:88:22:85:78'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d3:cd',
                                             'name': '102 SCHWEITZER CR (OHLHAUSEN '
                                                     'JEFF)',
                                             'serial_number': '5A6B8A',
                                             'system_mac_address': None},
                '44:65:7f:7b:d3:d6': {       'dev_macs': ['a4:2b:8c:fd:6d:a1'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d3:d6',
                                             'name': '3620 NAFZIGER RD (FIRST ST PAULS '
                                                     'LUTHERAN CHURCH)',
                                             'serial_number': '5A6B8D',
                                             'system_mac_address': None},
                '44:65:7f:7b:d3:e2': {       'dev_macs': [       '5c:e9:31:8c:d8:64',
                                                                 '74:fe:ce:31:9f:89'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d3:e2',
                                             'name': '4939 Perth Rd 119 Gadshill (J & '
                                                     'M Patterson 10423710)',
                                             'serial_number': '5A6B91',
                                             'system_mac_address': None},
                '44:65:7f:7b:d3:eb': {       'dev_macs': ['24:2d:6c:40:58:01'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d3:eb',
                                             'name': '119 FERRIS DR (CHAKRABURTTY '
                                                     'SAIBAL)',
                                             'serial_number': '5A6B94',
                                             'system_mac_address': None},
                '44:65:7f:7b:d4:30': {       'dev_macs': ['d8:07:b6:c5:34:07'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d4:30',
                                             'name': '3536 NAFZIGER RD (DANIEL BISCH '
                                                     'GENERAL CONTRACTO)',
                                             'serial_number': '5A6BAB',
                                             'system_mac_address': None},
                '44:65:7f:7b:d6:b8': {       'dev_macs': ['18:fd:74:7a:b8:52'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d6:b8',
                                             'name': "1245 QUEEN'S BUSH RD (GUENTHER "
                                                     'TECH INC)',
                                             'serial_number': '5A6C83',
                                             'system_mac_address': None},
                '44:65:7f:7b:d6:e2': {       'dev_macs': ['24:05:88:0f:65:ff'],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d6:e2',
                                             'name': '16 VILLAGE RD (EGGERT MIKE)',
                                             'serial_number': '5A6C91',
                                             'system_mac_address': None},
                '44:65:7f:7b:d6:e8': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '44:65:7f:7b:d6:e8',
                                             'name': '16 SCHWEITZER CR (AZIZ SHAHZAD)',
                                             'serial_number': '5A6C93',
                                             'system_mac_address': None},
                '48:77:46:0c:04:5e': {       'dev_macs': ['b8:94:70:f3:e6:50'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:0c:04:5e',
                                             'name': '3691 NAFZIGER RD (G/A REPAIR '
                                                     'MACHINERY INC)',
                                             'serial_number': '6DD524',
                                             'system_mac_address': None},
                '48:77:46:3a:7f:ea': {       'dev_macs': [       'b8:94:70:bc:ba:3d',
                                                                 '60:db:98:09:1f:b6',
                                                                 'd0:76:8f:91:d4:d2'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:3a:7f:ea',
                                             'name': '1020 MAPLE LEAF ST (BRODRECHT '
                                                     'ANDREA)',
                                             'serial_number': '722218',
                                             'system_mac_address': None},
                '48:77:46:3b:8f:12': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:3b:8f:12',
                                             'name': '4937 PERTH ROAD 119 (BANNON AMY)',
                                             'serial_number': '7243FD',
                                             'system_mac_address': None},
                '48:77:46:3b:c8:3a': {       'dev_macs': ['b8:94:70:f3:e4:a3'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:3b:c8:3a',
                                             'name': '4885 Tiffany Ct. Gadshill (Paul '
                                                     'Duden)',
                                             'serial_number': '724B22',
                                             'system_mac_address': None},
                '48:77:46:47:06:46': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:47:06:46',
                                             'name': '1 POND VIEW DR (PENTELBURY '
                                                     'LAURENE)',
                                             'serial_number': '73ABE1',
                                             'system_mac_address': None},
                '48:77:46:47:07:f6': {       'dev_macs': ['00:03:e6:dd:30:b2'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:47:07:f6',
                                             'name': '2184 Perth Line Wellesley (Alex '
                                                     'Albrecht 10436172)',
                                             'serial_number': '73AC11',
                                             'system_mac_address': None},
                '48:77:46:8c:2c:dc': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:8c:2c:dc',
                                             'name': '4009 Line 52 (Erik Vanaarsen)',
                                             'serial_number': '76B03D',
                                             'system_mac_address': None},
                '48:77:46:8c:6d:5c': {       'dev_macs': ['14:59:c0:ac:a6:ee'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:8c:6d:5c',
                                             'name': '2632 NOTRE DAME DR (BROWN WADE)',
                                             'serial_number': '76B84D',
                                             'system_mac_address': None},
                '48:77:46:a2:1b:dc': {       'dev_macs': ['14:21:03:a9:71:39'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:a2:1b:dc',
                                             'name': '83 GALENA ST (RUBY MYRON)',
                                             'serial_number': '71A4AA',
                                             'system_mac_address': None},
                '48:77:46:a2:1c:ac': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:a2:1c:ac',
                                             'name': '4011 Frederick St. Gadshill '
                                                     '(Sheila Loerzel)',
                                             'serial_number': '71A4C4',
                                             'system_mac_address': None},
                '48:77:46:a3:bb:54': {       'dev_macs': ['44:65:7f:89:0b:b5'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:a3:bb:54',
                                             'name': '135 GERBER MEADOWS DRIVE '
                                                     '(LOBRUTTO JOANNE)',
                                             'serial_number': '71D899',
                                             'system_mac_address': None},
                '48:77:46:cd:54:e6': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:cd:54:e6',
                                             'name': '4949 Perth Road 119 (Elizabeth '
                                                     'Berg 10431695)',
                                             'serial_number': '76D049',
                                             'system_mac_address': None},
                '48:77:46:ce:a6:ce': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:ce:a6:ce',
                                             'name': '5525 Perth Rd. 119 (Heather '
                                                     'Stever)',
                                             'serial_number': '8458AC',
                                             'system_mac_address': None},
                '48:77:46:f8:71:a2': {       'dev_macs': ['d0:76:8f:02:f1:49'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:f8:71:a2',
                                             'name': '1022 MOLESWORTH ST (ROTH '
                                                     'STEPHENIE)',
                                             'serial_number': '82A40C',
                                             'system_mac_address': None},
                '48:77:46:f8:74:3c': {       'dev_macs': ['d0:76:8f:02:f2:eb'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:f8:74:3c',
                                             'name': "1185 QUEEN'S BUSH RD (SAUDER "
                                                     'JONATHAN)',
                                             'serial_number': '82A4EA',
                                             'system_mac_address': None},
                '48:77:46:f8:74:87': {       'dev_macs': ['ec:ad:e0:4a:41:f2'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:f8:74:87',
                                             'name': '1125 MOLESWORTH ST (WHITSITT '
                                                     'WARREN)',
                                             'serial_number': '82A503',
                                             'system_mac_address': None},
                '48:77:46:f8:76:af': {       'dev_macs': ['3c:7c:3f:d1:b2:50'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:f8:76:af',
                                             'name': '134 GERBER MEADOWS DR (CAUCEAN '
                                                     'FLORIN)',
                                             'serial_number': '82A5BB',
                                             'system_mac_address': None},
                '48:77:46:f8:7b:11': {       'dev_macs': ['10:27:f5:b3:77:79'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:f8:7b:11',
                                             'name': '91 POND VIEW DR (BEVERS PAULINE)',
                                             'serial_number': '82A731',
                                             'system_mac_address': None},
                '48:77:46:f8:7b:4a': {       'dev_macs': ['f0:a7:31:21:45:55'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:f8:7b:4a',
                                             'name': '3697 NAFZIGER RD (DAMMEIER RYAN)',
                                             'serial_number': '82A744',
                                             'system_mac_address': None},
                '48:77:46:f8:7b:65': {       'dev_macs': ['60:db:98:09:39:cb'],
                                             'ip_address': None,
                                             'mac_address': '48:77:46:f8:7b:65',
                                             'name': '1831 MOSER-YOUNG RD (ESBOUGH '
                                                     'MALLORY)',
                                             'serial_number': '82A74D',
                                             'system_mac_address': None},
                '60:db:98:4e:67:2c': {       'dev_macs': ['14:21:03:92:ed:57'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:67:2c',
                                             'name': '49 DAVID ST (BUONO TERESA)',
                                             'serial_number': 'A9C2A5',
                                             'system_mac_address': None},
                '60:db:98:4e:67:7c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:67:7c',
                                             'name': '5596 PERTH LINE 56A (KUEPFER '
                                                     'BARBARA)',
                                             'serial_number': 'A9C2AF',
                                             'system_mac_address': None},
                '60:db:98:4e:67:a4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:67:a4',
                                             'name': '195 FERRIS DRIVE WELLESLEY (KUNG '
                                                     'KIMBERLEY)',
                                             'serial_number': 'A9C2B4',
                                             'system_mac_address': None},
                '60:db:98:4e:67:dc': {       'dev_macs': ['b8:94:70:bc:b8:d2'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:67:dc',
                                             'name': '4010 Fredrick St.',
                                             'serial_number': 'A9C2BB',
                                             'system_mac_address': None},
                '60:db:98:4e:69:3c': {       'dev_macs': ['40:f2:e9:dc:d0:64'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:69:3c',
                                             'name': '68 POND VIEW DR (MARTIN BRETT)',
                                             'serial_number': 'A9C2E7',
                                             'system_mac_address': None},
                '60:db:98:4e:69:ac': {       'dev_macs': [       '00:03:e6:db:2e:b7',
                                                                 '00:03:e6:db:48:37',
                                                                 '00:03:e6:db:4b:65',
                                                                 '00:03:e6:db:51:db',
                                                                 '60:db:98:55:4e:9a'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:69:ac',
                                             'name': '2 PARKVIEW DR (YOST BRIAN)',
                                             'serial_number': 'A9C2F5',
                                             'system_mac_address': None},
                '60:db:98:4e:6d:54': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:6d:54',
                                             'name': "1201 QUEEN'S BUSH RD (COOKS "
                                                     'PHARMACY)',
                                             'serial_number': 'A9C36A',
                                             'system_mac_address': None},
                '60:db:98:4e:6d:e4': {       'dev_macs': ['d0:76:8f:1f:64:cf'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:6d:e4',
                                             'name': "1155 QUEEN'S BUSH RD (BERTRAM "
                                                     'ALLISTER)',
                                             'serial_number': 'A9C37C',
                                             'system_mac_address': None},
                '60:db:98:4e:6e:dc': {       'dev_macs': ['14:21:03:a9:75:6f'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:6e:dc',
                                             'name': '12 SHADYWOOD CT WELLESLEY (BROWN '
                                                     'SHANE)',
                                             'serial_number': 'A9C39B',
                                             'system_mac_address': None},
                '60:db:98:4e:6e:ec': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:6e:ec',
                                             'name': '1317 GREENWOOD HILL RD (JANTZI '
                                                     'RANDELL 10436707)',
                                             'serial_number': 'A9C39D',
                                             'system_mac_address': None},
                '60:db:98:4e:6f:14': {       'dev_macs': ['d0:76:8f:91:d6:7f'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:6f:14',
                                             'name': '26 POND VIEW DR (ALBRECHT ORIE)',
                                             'serial_number': 'A9C3A2',
                                             'system_mac_address': None},
                '60:db:98:4e:6f:5c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:6f:5c',
                                             'name': '1907 PERTH LINE 56 (MANNHARDT '
                                                     'JOSHUA)',
                                             'serial_number': 'A9C3AB',
                                             'system_mac_address': None},
                '60:db:98:4e:70:2c': {       'dev_macs': ['d0:76:8f:91:d5:98'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:70:2c',
                                             'name': '7 SHADYWOOD CT (LEIS GARY)',
                                             'serial_number': 'A9C3C5',
                                             'system_mac_address': None},
                '60:db:98:4e:71:0c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:71:0c',
                                             'name': '21 JACK KOEHLER LN (WENDE JASON)',
                                             'serial_number': 'A9C3E1',
                                             'system_mac_address': None},
                '60:db:98:4e:71:a4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:71:a4',
                                             'name': "1044 QUEEN'S BUSH RD (PANKIW "
                                                     'ANDRE)',
                                             'serial_number': 'A9C3F4',
                                             'system_mac_address': None},
                '60:db:98:4e:71:b4': {       'dev_macs': ['d0:76:8f:02:f1:ac'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:71:b4',
                                             'name': '82 LAWRENCE ST (MACKINNON '
                                                     'MATTHEW)',
                                             'serial_number': 'A9C3F6',
                                             'system_mac_address': None},
                '60:db:98:4e:72:9c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:72:9c',
                                             'name': '1561 MOSER-YOUNG RD (OPENSPACE '
                                                     'SOLUTIONS INC)',
                                             'serial_number': 'A9C413',
                                             'system_mac_address': None},
                '60:db:98:4e:72:cc': {       'dev_macs': [       '00:03:e6:db:49:4d',
                                                                 '00:03:e6:db:4f:8d',
                                                                 'd0:76:8f:93:c8:1b'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:72:cc',
                                             'name': '55 ZOEGER CT (MORRISON LISA)',
                                             'serial_number': 'A9C419',
                                             'system_mac_address': None},
                '60:db:98:4e:91:c4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:91:c4',
                                             'name': '2858 Line 45 (Mary VanStaveren)',
                                             'serial_number': 'A9C7F8',
                                             'system_mac_address': None},
                '60:db:98:4e:95:54': {       'dev_macs': ['98:25:4a:ea:4d:dd'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:95:54',
                                             'name': '31 Nithbank Rd Wellesley (Chris '
                                                     'Tillery 10434314)',
                                             'serial_number': 'A9C86A',
                                             'system_mac_address': None},
                '60:db:98:4e:97:e4': {       'dev_macs': ['d0:76:8f:02:f5:48'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:97:e4',
                                             'name': '35 NITHBANK RD (ERB CAROLYN)',
                                             'serial_number': 'A9C8BC',
                                             'system_mac_address': None},
                '60:db:98:4e:99:f4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:4e:99:f4',
                                             'name': '1214 HUTCHISON RD (JANTZI '
                                                     'CONRAD)',
                                             'serial_number': 'A9C8FE',
                                             'system_mac_address': None},
                '60:db:98:72:ec:27': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:72:ec:27',
                                             'name': '114 Nithbank Rd (Bryce Bramhill '
                                                     '10434766)',
                                             'serial_number': 'B08DA5',
                                             'system_mac_address': None},
                '60:db:98:72:f5:c7': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:72:f5:c7',
                                             'name': '2847 Line 45 (Sean Bell)',
                                             'serial_number': 'B08ED9',
                                             'system_mac_address': None},
                '60:db:98:73:9b:cc': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:73:9b:cc',
                                             'name': '5039 PERTH ROAD 119 (RIEHL '
                                                     'ROBIN)',
                                             'serial_number': 'B0A28B',
                                             'system_mac_address': None},
                '60:db:98:73:9c:44': {       'dev_macs': ['60:db:98:82:0a:8a'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:73:9c:44',
                                             'name': '1825 MOSER-YOUNG RD (DOERING '
                                                     'CHARLES)',
                                             'serial_number': 'B0A29A',
                                             'system_mac_address': None},
                '60:db:98:73:9c:6c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:73:9c:6c',
                                             'name': '4879 TIFFANY CT (ROULSTON ROD & '
                                                     'GLORIA)',
                                             'serial_number': 'B0A29F',
                                             'system_mac_address': None},
                '60:db:98:73:9f:44': {       'dev_macs': ['60:db:98:09:1a:e6'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:73:9f:44',
                                             'name': '4424 WEIMAR LINE (COBER VALERIE)',
                                             'serial_number': 'B0A2FA',
                                             'system_mac_address': None},
                '60:db:98:73:e2:e3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:73:e2:e3',
                                             'name': '1020 MOLESWORTH ST (BENDER '
                                                     'WYATT)',
                                             'serial_number': 'B0AAF7',
                                             'system_mac_address': None},
                '60:db:98:8e:fd:e5': {       'dev_macs': ['60:db:98:09:38:ad'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:8e:fd:e5',
                                             'name': '1010 DOERING ST (FRIES  ROBERT)',
                                             'serial_number': 'B510BB',
                                             'system_mac_address': None},
                '60:db:98:8e:ff:b0': {       'dev_macs': ['60:db:98:81:f3:2a'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:8e:ff:b0',
                                             'name': '4894 Linhaven Dr Gadshill (Lucas '
                                                     'Commerford 10435700)',
                                             'serial_number': 'B51154',
                                             'system_mac_address': None},
                '60:db:98:8f:00:3d': {       'dev_macs': ['e0:b2:60:5f:70:f0'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:8f:00:3d',
                                             'name': '4525 WEIMAR LINE (WEIMAR '
                                                     'MANUFACTURING INC)',
                                             'serial_number': 'B51183',
                                             'system_mac_address': None},
                '60:db:98:8f:00:6a': {       'dev_macs': ['d0:76:8f:02:ee:aa'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:8f:00:6a',
                                             'name': '1821 MOSER-YOUNG RD (KROETSCH '
                                                     'PAUL)',
                                             'serial_number': 'B51192',
                                             'system_mac_address': None},
                '60:db:98:8f:01:4e': {       'dev_macs': ['10:27:f5:a8:23:15'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:8f:01:4e',
                                             'name': '2832 NOTRE DAME DR (LAM EDMUND)',
                                             'serial_number': 'B511DE',
                                             'system_mac_address': None},
                '60:db:98:8f:01:75': {       'dev_macs': ['3c:5c:f1:63:9e:c1'],
                                             'ip_address': None,
                                             'mac_address': '60:db:98:8f:01:75',
                                             'name': '62 GALENA ST (HARFORD MATTHEW)',
                                             'serial_number': 'B511EB',
                                             'system_mac_address': None},
                '84:d3:43:7c:98:87': {       'dev_macs': ['84:d3:43:0f:b7:fa'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7c:98:87',
                                             'name': '7 PARKVIEW DR (FRANKLIN PAUL)',
                                             'serial_number': 'C8087A',
                                             'system_mac_address': None},
                '84:d3:43:7c:ff:cf': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7c:ff:cf',
                                             'name': '167 FERRIS DR (MITCHELL DAVID)',
                                             'serial_number': 'C81563',
                                             'system_mac_address': None},
                '84:d3:43:7d:0e:77': {       'dev_macs': ['84:d3:43:0f:bb:f9'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:0e:77',
                                             'name': "1147 QUEEN'S BUSH RD (POOLE "
                                                     'MATTHEW)',
                                             'serial_number': 'C81738',
                                             'system_mac_address': None},
                '84:d3:43:7d:0f:ff': {       'dev_macs': ['84:d3:43:0f:62:20'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:0f:ff',
                                             'name': "1140 QUEEN'S BUSH RD (Chad "
                                                     'KIESWETTER 10438397)',
                                             'serial_number': 'C81769',
                                             'system_mac_address': None},
                '84:d3:43:7d:14:a7': {       'dev_macs': ['b8:94:70:9e:cb:8e'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:14:a7',
                                             'name': '1 POND VIEW DR (LICHTY NORMA)',
                                             'serial_number': 'C817FE',
                                             'system_mac_address': None},
                '84:d3:43:7d:31:bf': {       'dev_macs': ['b8:94:70:9e:b7:bf'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:31:bf',
                                             'name': '68 WELWOOD AV (SCHMIDT DARLENE)',
                                             'serial_number': 'C81BA1',
                                             'system_mac_address': None},
                '84:d3:43:7d:38:77': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:38:77',
                                             'name': '2073 ALBERT ST (ERICKSON KAREN)',
                                             'serial_number': 'C81C78',
                                             'system_mac_address': None},
                '84:d3:43:7d:3f:37': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:3f:37',
                                             'name': '4900 Linhaven Dr. Gadshill '
                                                     '(Ronald Hundt 10423790)',
                                             'serial_number': 'C81D50',
                                             'system_mac_address': None},
                '84:d3:43:7d:3f:e7': {       'dev_macs': ['84:d3:43:0f:bb:28'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:3f:e7',
                                             'name': '74 SCHWEITZER CR (ENNS JAMIE)',
                                             'serial_number': 'C81D66',
                                             'system_mac_address': None},
                '84:d3:43:7d:40:37': {       'dev_macs': ['84:d3:43:0f:b6:a5'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:40:37',
                                             'name': '206 FERRIS DR (KUBE AMY & CHRIS)',
                                             'serial_number': 'C81D70',
                                             'system_mac_address': None},
                '84:d3:43:7d:40:f7': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:40:f7',
                                             'name': '49 PARKVIEW DR (KIENITZ JASVEEN)',
                                             'serial_number': 'C81D88',
                                             'system_mac_address': None},
                '84:d3:43:7d:41:b7': {       'dev_macs': ['b8:94:70:9d:ae:3a'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:41:b7',
                                             'name': '2827 LINE 45 (VANZANDWIJK GARY)',
                                             'serial_number': 'C81DA0',
                                             'system_mac_address': None},
                '84:d3:43:7d:41:cf': {       'dev_macs': ['b8:94:70:bc:b2:60'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:41:cf',
                                             'name': '112 REINER CR (DA SILVA HUGO)',
                                             'serial_number': 'C81DA3',
                                             'system_mac_address': None},
                '84:d3:43:7d:42:cf': {       'dev_macs': ['84:d3:43:0f:b9:b2'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:42:cf',
                                             'name': "1220 QUEEN'S BUSH RD (KUEPFER "
                                                     'CRAIG)',
                                             'serial_number': 'C81DC3',
                                             'system_mac_address': None},
                '84:d3:43:7d:43:af': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:43:af',
                                             'name': '4026 WEIMAR LINE (BROWN '
                                                     'ALEXANDER)',
                                             'serial_number': 'C81DDF',
                                             'system_mac_address': None},
                '84:d3:43:7d:43:d7': {       'dev_macs': ['b8:94:70:9e:cb:f1'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:43:d7',
                                             'name': '2830 LINE 45 (SZUSZ RYAN)',
                                             'serial_number': 'C81DE4',
                                             'system_mac_address': None},
                '84:d3:43:7d:44:2f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:44:2f',
                                             'name': '5 SHADYWOOD CT (HARKINS RYAN)',
                                             'serial_number': 'C81DEF',
                                             'system_mac_address': None},
                '84:d3:43:7d:44:b7': {       'dev_macs': ['b8:94:70:bc:ba:7f'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:44:b7',
                                             'name': '26 PARKVIEW DR (MYERS BOB)',
                                             'serial_number': 'C81E00',
                                             'system_mac_address': None},
                '84:d3:43:7d:44:ef': {       'dev_macs': ['20:7c:14:a0:b8:fe'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:44:ef',
                                             'name': '58 Nithbank Rd Wellesley (CHOW '
                                                     'ANDY)',
                                             'serial_number': 'C81E07',
                                             'system_mac_address': None},
                '84:d3:43:7d:44:f7': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:44:f7',
                                             'name': '3706 NAFZIGER RD (JENNIFER '
                                                     'PITSCH RMT)',
                                             'serial_number': 'C81E08',
                                             'system_mac_address': None},
                '84:d3:43:7d:45:1f': {       'dev_macs': ['60:db:98:09:1e:98'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:45:1f',
                                             'name': "1134 QUEEN'S BUSH RD (SCHOTT "
                                                     'JEFF)',
                                             'serial_number': 'C81E0D',
                                             'system_mac_address': None},
                '84:d3:43:7d:45:7f': {       'dev_macs': ['b8:94:70:9d:ae:9d'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:45:7f',
                                             'name': '3835 NAFZIGER RD (JANZEN PETER)',
                                             'serial_number': 'C81E19',
                                             'system_mac_address': None},
                '84:d3:43:7d:47:17': {       'dev_macs': ['84:d3:43:6d:e9:bc'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:47:17',
                                             'name': '2 LAWRENCE STREET (BARKER '
                                                     'ANGELA)',
                                             'serial_number': 'C81E4C',
                                             'system_mac_address': None},
                '84:d3:43:7d:47:7f': {       'dev_macs': ['84:d3:43:0f:63:a1'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:47:7f',
                                             'name': '92 BROWN ST (GERBER CONNOR)',
                                             'serial_number': 'C81E59',
                                             'system_mac_address': None},
                '84:d3:43:7d:4a:0f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:4a:0f',
                                             'name': '162 LAWRENCE ST (CAMPBELL PAYNE '
                                                     'MORGAN)',
                                             'serial_number': 'C81EAB',
                                             'system_mac_address': None},
                '84:d3:43:7d:4b:07': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:4b:07',
                                             'name': '3685 NAFZIGER RD (KUEPFER SHANE)',
                                             'serial_number': 'C81ECA',
                                             'system_mac_address': None},
                '84:d3:43:7d:4b:ef': {       'dev_macs': ['60:22:32:96:32:86'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:4b:ef',
                                             'name': '83 Parkview Dr Wellesley (Dave '
                                                     'Wintrip 10434356)',
                                             'serial_number': 'C81EE7',
                                             'system_mac_address': None},
                '84:d3:43:7d:4c:0f': {       'dev_macs': ['60:db:98:08:ff:d4'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:4c:0f',
                                             'name': '116 BROWN STREET (LICHTY SARAH)',
                                             'serial_number': 'C81EEB',
                                             'system_mac_address': None},
                '84:d3:43:7d:4d:47': {       'dev_macs': ['b8:94:70:f3:e4:61'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:4d:47',
                                             'name': '126 Ferris Dr Wellesley (Tara '
                                                     'Proper 10434940)',
                                             'serial_number': 'C81F12',
                                             'system_mac_address': None},
                '84:d3:43:7d:4d:57': {       'dev_macs': ['14:21:03:a9:7f:5c'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:4d:57',
                                             'name': '2343 PERTH LINE 56 (R&G DEEN '
                                                     'FARMS LTD)',
                                             'serial_number': 'C81F14',
                                             'system_mac_address': None},
                '84:d3:43:7d:5a:af': {       'dev_macs': ['60:db:98:09:1c:f6'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:5a:af',
                                             'name': '2853 LINE 45 (LECOMTE OLIVIA)',
                                             'serial_number': 'C820BF',
                                             'system_mac_address': None},
                '84:d3:43:7d:db:bf': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:db:bf',
                                             'name': '118 Schweitzer Cr Wellesley '
                                                     '(Esteban Giron 10434568)',
                                             'serial_number': 'C830E1',
                                             'system_mac_address': None},
                '84:d3:43:7d:de:3f': {       'dev_macs': ['60:db:98:09:1e:6c'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:de:3f',
                                             'name': "1294 QUEEN'S BUSH ROAD (WINKLER "
                                                     'MESLISSA)',
                                             'serial_number': 'C83131',
                                             'system_mac_address': None},
                '84:d3:43:7d:e2:ff': {       'dev_macs': ['d0:76:8f:de:de:92'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7d:e2:ff',
                                             'name': '63 NITHBANK RD (DAY RYAN)',
                                             'serial_number': 'C831C9',
                                             'system_mac_address': None},
                '84:d3:43:7e:07:57': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:07:57',
                                             'name': '145 SCHWEITZER CR (MCCLEMENTS '
                                                     'TIM)',
                                             'serial_number': 'C83654',
                                             'system_mac_address': None},
                '84:d3:43:7e:15:b7': {       'dev_macs': ['84:d3:43:6d:ea:fb'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:15:b7',
                                             'name': '14 VILLAGE RD (SCHMIDT RICHARD)',
                                             'serial_number': 'C83820',
                                             'system_mac_address': None},
                '84:d3:43:7e:1c:d7': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:1c:d7',
                                             'name': '196 FERRIS DR (FIFE PAUL)',
                                             'serial_number': 'C83904',
                                             'system_mac_address': None},
                '84:d3:43:7e:1e:97': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:1e:97',
                                             'name': '60 REINER CR MILVERTON (WELLMAN '
                                                     'MICHELLE)',
                                             'serial_number': 'C8393C',
                                             'system_mac_address': None},
                '84:d3:43:7e:1e:b7': {       'dev_macs': ['84:d3:43:0f:63:6a'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:1e:b7',
                                             'name': '4147 WEIMAR LINE (KREUTZKAMP '
                                                     'JOHN)',
                                             'serial_number': 'C83940',
                                             'system_mac_address': None},
                '84:d3:43:7e:2d:6f': {       'dev_macs': ['18:fd:74:1d:05:ce'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:2d:6f',
                                             'name': '3580 NAFZIGER RD WELLESLEY '
                                                     '(WELLESLEY VET CLINIC)',
                                             'serial_number': 'C83B17',
                                             'system_mac_address': None},
                '84:d3:43:7e:2f:df': {       'dev_macs': ['60:db:98:09:3b:0a'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:2f:df',
                                             'name': '157 Lawrence St Wellesley (Lisa '
                                                     'Snider 10435028)',
                                             'serial_number': 'C83B65',
                                             'system_mac_address': None},
                '84:d3:43:7e:2f:ff': {       'dev_macs': ['84:d3:43:0f:b8:f7'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:2f:ff',
                                             'name': '37 PARKVIEW DR (KORCZYNSKI '
                                                     'JENNA)',
                                             'serial_number': 'C83B69',
                                             'system_mac_address': None},
                '84:d3:43:7e:30:2f': {       'dev_macs': ['d0:76:8f:02:f7:6e'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:30:2f',
                                             'name': 'ont 164167',
                                             'serial_number': 'C83B6F',
                                             'system_mac_address': None},
                '84:d3:43:7e:31:47': {       'dev_macs': ['b8:94:70:f3:e7:d1'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:31:47',
                                             'name': '5421 KING ST (NICKEL KIM)',
                                             'serial_number': 'C83B92',
                                             'system_mac_address': None},
                '84:d3:43:7e:31:5f': {       'dev_macs': ['84:d3:43:6d:e9:4e'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:31:5f',
                                             'name': '5 FIRELLA PL (ROBINSON MARTIN)',
                                             'serial_number': 'C83B95',
                                             'system_mac_address': None},
                '84:d3:43:7e:31:8f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:31:8f',
                                             'name': '13 FIRELLA PL (ARENDOQUE LORD)',
                                             'serial_number': 'C83B9B',
                                             'system_mac_address': None},
                '84:d3:43:7e:31:b7': {       'dev_macs': ['14:21:03:8b:70:84'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:31:b7',
                                             'name': '2 POND VIEW DR (FOWLER ANNE)',
                                             'serial_number': 'C83BA0',
                                             'system_mac_address': None},
                '84:d3:43:7e:31:df': {       'dev_macs': ['cc:be:59:7e:d4:8f'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:31:df',
                                             'name': '32 POND VIEW DR (ERB FRANKLIN & '
                                                     'JEAN)',
                                             'serial_number': 'C83BA5',
                                             'system_mac_address': None},
                '84:d3:43:7e:31:ef': {       'dev_macs': ['b8:94:70:bc:b2:b8'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:31:ef',
                                             'name': '1006 Doering St (HURREN RICHARD '
                                                     '& KELLY)',
                                             'serial_number': 'C83BA7',
                                             'system_mac_address': None},
                '84:d3:43:7e:32:a7': {       'dev_macs': ['84:d3:43:6d:fa:5d'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:32:a7',
                                             'name': '11 POND VIEW DR (LICHTY SAM S)',
                                             'serial_number': 'C83BBE',
                                             'system_mac_address': None},
                '84:d3:43:7e:32:bf': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:32:bf',
                                             'name': '3832 NAFZIGER RD (ADAM KATRINA)',
                                             'serial_number': 'C83BC1',
                                             'system_mac_address': None},
                '84:d3:43:7e:32:d7': {       'dev_macs': ['f0:2f:74:10:95:c0'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:32:d7',
                                             'name': '1287 HUTCHISON RD (LEIS FEED & '
                                                     'SUPPLY LTD)',
                                             'serial_number': 'C83BC4',
                                             'system_mac_address': None},
                '84:d3:43:7e:33:4f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:33:4f',
                                             'name': '2854 LINE 45 (JANZEN AARON & '
                                                     'MARIA)',
                                             'serial_number': 'C83BD3',
                                             'system_mac_address': None},
                '84:d3:43:7e:34:e7': {       'dev_macs': ['60:db:98:09:1d:2d'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:34:e7',
                                             'name': '20 WELWOOD AV (SEIP TERRY)',
                                             'serial_number': 'C83C06',
                                             'system_mac_address': None},
                '84:d3:43:7e:35:8f': {       'dev_macs': ['60:db:98:09:38:d9'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:35:8f',
                                             'name': '4005 FREDERICK ST (SKINNER '
                                                     'SCOTT)',
                                             'serial_number': 'C83C1B',
                                             'system_mac_address': None},
                '84:d3:43:7e:35:b7': {       'dev_macs': ['44:65:7f:88:45:a3'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:35:b7',
                                             'name': '3709 NAFZIGER ROAD (HANER ARDEN)',
                                             'serial_number': 'C83C20',
                                             'system_mac_address': None},
                '84:d3:43:7e:44:6f': {       'dev_macs': ['84:d3:43:6d:ed:21'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:44:6f',
                                             'name': '15 ZINKANN CR (MASSEL SHELLY)',
                                             'serial_number': 'C83DF7',
                                             'system_mac_address': None},
                '84:d3:43:7e:45:37': {       'dev_macs': ['84:d3:43:0f:b6:fd'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:45:37',
                                             'name': '117 LAWRENCE ST (BUTLER STACI)',
                                             'serial_number': 'C83E10',
                                             'system_mac_address': None},
                '84:d3:43:7e:45:67': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:7e:45:67',
                                             'name': '4639 WEIMAR LINE (HOOVER EDWIN)',
                                             'serial_number': 'C83E16',
                                             'system_mac_address': None},
                '84:d3:43:95:8a:32': {       'dev_macs': ['84:d3:43:6d:e9:f3'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8a:32',
                                             'name': '5324A Road 106 (LICHTI JASON)',
                                             'serial_number': 'CBD165',
                                             'system_mac_address': None},
                '84:d3:43:95:8b:82': {       'dev_macs': ['60:db:98:09:1d:7a'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8b:82',
                                             'name': '102 GERBER MEADOWS DR (LADD '
                                                     'ANITA)',
                                             'serial_number': 'CBD18F',
                                             'system_mac_address': None},
                '84:d3:43:95:8b:8a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8b:8a',
                                             'name': '5470 KING ST (BAILLIE BRAD)',
                                             'serial_number': 'CBD190',
                                             'system_mac_address': None},
                '84:d3:43:95:8c:7a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8c:7a',
                                             'name': '4225 WEIMAR LINE (MARTIN MARY)',
                                             'serial_number': 'CBD1AE',
                                             'system_mac_address': None},
                '84:d3:43:95:8c:ca': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8c:ca',
                                             'name': '1010 MAPLE LEAF ST (GERVAIS '
                                                     'STEVE)',
                                             'serial_number': 'CBD1B8',
                                             'system_mac_address': None},
                '84:d3:43:95:8c:da': {       'dev_macs': ['84:d3:43:0f:b9:7b'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8c:da',
                                             'name': '87 REINER CR (PATTERSON '
                                                     'JENNIFER)',
                                             'serial_number': 'CBD1BA',
                                             'system_mac_address': None},
                '84:d3:43:95:8d:62': {       'dev_macs': ['60:db:98:55:b4:70'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8d:62',
                                             'name': '126 LAWRENCE ST (VOTEARY JIM)',
                                             'serial_number': 'CBD1CB',
                                             'system_mac_address': None},
                '84:d3:43:95:8d:b2': {       'dev_macs': ['5c:0f:fb:2a:c7:9e'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8d:b2',
                                             'name': '1 POND VIEW DR (GAJDOS ANITA)',
                                             'serial_number': 'CBD1D5',
                                             'system_mac_address': None},
                '84:d3:43:95:8d:f2': {       'dev_macs': ['60:db:98:09:1f:e2'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8d:f2',
                                             'name': '3524B NAFZIGER RD (BARTLETT '
                                                     'BRETT)',
                                             'serial_number': 'CBD1DD',
                                             'system_mac_address': None},
                '84:d3:43:95:8e:da': {       'dev_macs': ['84:d3:43:6d:f9:c3'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8e:da',
                                             'name': "1031 QUEEN'S BUSH RD (PITT "
                                                     'BRENDA)',
                                             'serial_number': 'CBD1FA',
                                             'system_mac_address': None},
                '84:d3:43:95:8f:3a': {       'dev_macs': ['60:db:98:55:ba:27'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8f:3a',
                                             'name': '5329 Road 106 (Scott Rupert)',
                                             'serial_number': 'CBD206',
                                             'system_mac_address': None},
                '84:d3:43:95:8f:72': {       'dev_macs': ['60:db:98:09:13:56'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:8f:72',
                                             'name': '23 GERBER MEADOWS DR (SHANTZ '
                                                     'BRENDA)',
                                             'serial_number': 'CBD20D',
                                             'system_mac_address': None},
                '84:d3:43:95:90:ea': {       'dev_macs': ['84:d3:43:0f:b7:60'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:95:90:ea',
                                             'name': '64 REINER CR (GOREL SHANTELLE)',
                                             'serial_number': 'CBD23C',
                                             'system_mac_address': None},
                '84:d3:43:96:00:82': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:96:00:82',
                                             'name': '4332 WEIMAR LINE (ROSS JEFF)',
                                             'serial_number': 'CBE02F',
                                             'system_mac_address': None},
                '84:d3:43:96:02:92': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:96:02:92',
                                             'name': "1202 QUEEN'S BUSH RD (BORSATO "
                                                     'LARRY)',
                                             'serial_number': 'CBE071',
                                             'system_mac_address': None},
                '84:d3:43:96:04:02': {       'dev_macs': ['60:db:98:09:1d:a6'],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:96:04:02',
                                             'name': '1096 MOLESWORTH ST (MCGOVERN '
                                                     'MATT)',
                                             'serial_number': 'CBE09F',
                                             'system_mac_address': None},
                '84:d3:43:b3:79:6d': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:b3:79:6d',
                                             'name': '2840 LINE 45 (TAYLOR BRIAN)',
                                             'serial_number': 'DCF440',
                                             'system_mac_address': None},
                '84:d3:43:b3:7d:75': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': '84:d3:43:b3:7d:75',
                                             'name': '90 PARKVIEW DR WELLESLEY '
                                                     '(TEMPLETON JOANN)',
                                             'serial_number': 'DCF4C1',
                                             'system_mac_address': None},
                'b8:94:70:1f:09:f1': {       'dev_macs': ['14:21:03:92:2b:fb'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:1f:09:f1',
                                             'name': '3937 WEIMAR LINE (BORONKA PETER)',
                                             'serial_number': 'FD873A',
                                             'system_mac_address': None},
                'b8:94:70:a1:46:5d': {       'dev_macs': ['5c:0f:fb:3d:51:70'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:46:5d',
                                             'name': '19 PARKVIEW DR WELLESLEY '
                                                     '(SCHUMACHER RICHARD)',
                                             'serial_number': '117DAD7',
                                             'system_mac_address': None},
                'b8:94:70:a1:4a:9d': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:4a:9d',
                                             'name': '5277 PERTH ROAD 119 (SUN RISE '
                                                     'SEAMLESS EAVESTROUGHI)',
                                             'serial_number': '117DB5F',
                                             'system_mac_address': None},
                'b8:94:70:a1:4a:ed': {       'dev_macs': ['d0:76:8f:1f:63:c7'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:4a:ed',
                                             'name': '1010 DOERING STREET (GUENTHER '
                                                     'ROY)',
                                             'serial_number': '117DB69',
                                             'system_mac_address': None},
                'b8:94:70:a1:4a:fd': {       'dev_macs': ['14:21:03:92:eb:73'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:4a:fd',
                                             'name': '102 LAWRENCE ST (WELCHES '
                                                     'MARGARET T)',
                                             'serial_number': '117DB6B',
                                             'system_mac_address': None},
                'b8:94:70:a1:4d:05': {       'dev_macs': ['14:21:03:a9:74:ca'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:4d:05',
                                             'name': "1328 QUEEN'S BUSH RD (DE ARTOLA "
                                                     'THOMAS)',
                                             'serial_number': '117DBAC',
                                             'system_mac_address': None},
                'b8:94:70:a1:4f:3d': {       'dev_macs': ['14:21:03:92:c8:fd'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:4f:3d',
                                             'name': '21 PARKVIEW DR (LUKAS MATTHEW)',
                                             'serial_number': '117DBF3',
                                             'system_mac_address': None},
                'b8:94:70:a1:50:9d': {       'dev_macs': ['14:21:03:a9:77:8a'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:50:9d',
                                             'name': '1 POND VIEW DRIVE UNIT 201 (BAST '
                                                     'DARLENE)',
                                             'serial_number': '117DC1F',
                                             'system_mac_address': None},
                'b8:94:70:a1:51:0d': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:51:0d',
                                             'name': '138 GERBER MEADOWS DR (SMITH '
                                                     'GREGORY)',
                                             'serial_number': '117DC2D',
                                             'system_mac_address': None},
                'b8:94:70:a1:51:35': {       'dev_macs': ['44:65:7f:79:dc:fc'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:a1:51:35',
                                             'name': '26 JACK KOEHLER LINE (NG ALBERT)',
                                             'serial_number': '117DC32',
                                             'system_mac_address': None},
                'b8:94:70:b9:d7:f5': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:b9:d7:f5',
                                             'name': '1416 HUTCHISON RD (RON C ESBAUGH '
                                                     'CONST LTD)',
                                             'serial_number': '1193013',
                                             'system_mac_address': None},
                'b8:94:70:b9:df:7d': {       'dev_macs': ['60:db:98:09:3c:d8'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:b9:df:7d',
                                             'name': '48 REINER CR (ALBRECHT ALLISON)',
                                             'serial_number': '1193104',
                                             'system_mac_address': None},
                'b8:94:70:ba:34:35': {       'dev_macs': ['b8:94:70:f3:e0:57'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:34:35',
                                             'name': '2660 LINE 47 GADSHILL (JAIME '
                                                     'NEEB)',
                                             'serial_number': '1193B9B',
                                             'system_mac_address': None},
                'b8:94:70:ba:e0:6d': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:e0:6d',
                                             'name': '39 REINER CR (SUI DANIEL)',
                                             'serial_number': '1195122',
                                             'system_mac_address': None},
                'b8:94:70:ba:e0:a5': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:e0:a5',
                                             'name': '5517 BROWN ST (ELLISON WHITE '
                                                     'MIKE & JULIE)',
                                             'serial_number': '1195129',
                                             'system_mac_address': None},
                'b8:94:70:ba:e1:65': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:e1:65',
                                             'name': "1193 QUEEN'S BUSH RD (1001090591 "
                                                     'ONTARIO INC)',
                                             'serial_number': '1195141',
                                             'system_mac_address': None},
                'b8:94:70:ba:e2:1d': {       'dev_macs': ['04:bc:9f:46:91:30'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:e2:1d',
                                             'name': '111 VILLAGE RD (BELL NANCY)',
                                             'serial_number': '1195158',
                                             'system_mac_address': None},
                'b8:94:70:ba:e3:15': {       'dev_macs': ['04:bc:9f:46:ac:e7'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:e3:15',
                                             'name': '5446 ROAD 106 (FORAN GERALD)',
                                             'serial_number': '1195177',
                                             'system_mac_address': None},
                'b8:94:70:ba:e4:ed': {       'dev_macs': ['b8:94:70:f3:e1:5f'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:e4:ed',
                                             'name': '16 DAVID ST (BRENNAN KARALEE)',
                                             'serial_number': '11951B2',
                                             'system_mac_address': None},
                'b8:94:70:ba:e5:15': {       'dev_macs': ['d0:76:8f:91:d2:f9'],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ba:e5:15',
                                             'name': '12 LAWRENCE ST (GIVLIN FRANCIS)',
                                             'serial_number': '11951B7',
                                             'system_mac_address': None},
                'b8:94:70:ce:5c:e4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'b8:94:70:ce:5c:e4',
                                             'name': '2312 PERTH LINE (SCHULTZ '
                                                     'NICHOLAS)',
                                             'serial_number': '11BE1EE',
                                             'system_mac_address': None},
                'cc:be:59:0c:b2:b4': {       'dev_macs': [       'd0:76:8f:1f:64:6c',
                                                                 'd0:76:8f:02:f2:ca',
                                                                 '60:db:98:09:3a:44'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:0c:b2:b4',
                                             'name': 'Pond View MDU',
                                             'serial_number': '3ADD0E',
                                             'system_mac_address': None},
                'cc:be:59:47:32:ff': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:47:32:ff',
                                             'name': '4896 Linhaven Dr. Gadshill '
                                                     '(Trevor Wraith)',
                                             'serial_number': '433E67',
                                             'system_mac_address': None},
                'cc:be:59:4f:f7:6d': {       'dev_macs': ['44:65:7f:41:cb:29'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:4f:f7:6d',
                                             'name': '1177 Henry St Wellesley (Murray '
                                                     'Bremner 10434791)',
                                             'serial_number': '43AF83',
                                             'system_mac_address': None},
                'cc:be:59:4f:fe:9f': {       'dev_macs': ['44:65:7f:38:04:51'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:4f:fe:9f',
                                             'name': "1098 Queen's Bush Rd Wellesley "
                                                     '(Dan Koebel 10434542)',
                                             'serial_number': '43B319',
                                             'system_mac_address': None},
                'cc:be:59:4f:ff:61': {       'dev_macs': ['d0:76:8f:02:f1:b7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:4f:ff:61',
                                             'name': '34 Parkview Dr Wellesley (Eric '
                                                     'Jeffray10434740)',
                                             'serial_number': '43B37A',
                                             'system_mac_address': None},
                'cc:be:59:4f:ff:b9': {       'dev_macs': ['d0:76:8f:1f:5a:f8'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:4f:ff:b9',
                                             'name': '87 PARKVIEW DR (DAVIDSON IVAN)',
                                             'serial_number': '43B3A6',
                                             'system_mac_address': None},
                'cc:be:59:4f:ff:dd': {       'dev_macs': ['74:fe:ce:31:9c:d1'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:4f:ff:dd',
                                             'name': '1260 GREENWOOD HILL RD '
                                                     '(SLOTHOUBER STEPHEN)',
                                             'serial_number': '43B3B8',
                                             'system_mac_address': None},
                'cc:be:59:50:00:0f': {       'dev_macs': ['38:2c:4a:c7:10:75'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:00:0f',
                                             'name': '69 Parkview Dr (Jon Lambert '
                                                     '10434823)',
                                             'serial_number': '43B3D1',
                                             'system_mac_address': None},
                'cc:be:59:50:00:25': {       'dev_macs': ['cc:be:59:31:4e:fa'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:00:25',
                                             'name': '1385 Greenwood Hill Rd Wellesley '
                                                     '(Nelson Hoover 10435286)',
                                             'serial_number': '43B3DC',
                                             'system_mac_address': None},
                'cc:be:59:50:00:67': {       'dev_macs': ['78:54:2e:f0:6c:03'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:00:67',
                                             'name': '1031 Doering St Wellesley '
                                                     '(Lawrence Beach 10434628)',
                                             'serial_number': '43B3FD',
                                             'system_mac_address': None},
                'cc:be:59:50:00:a3': {       'dev_macs': ['00:90:7f:b9:86:b4'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:00:a3',
                                             'name': '3706 Nafziger Rd Wellesley (Adds '
                                                     '& Ends 10435557)',
                                             'serial_number': '43B41A',
                                             'system_mac_address': None},
                'cc:be:59:50:00:dd': {       'dev_macs': ['44:65:7f:38:00:58'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:00:dd',
                                             'name': "1205 Queen's Bush Rd Wellesley "
                                                     '(Kristen Lobrutto 10434675)',
                                             'serial_number': '43B437',
                                             'system_mac_address': None},
                'cc:be:59:50:00:f9': {       'dev_macs': ['44:65:7f:89:09:48'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:00:f9',
                                             'name': '16 Zinkann Cr Wellesley '
                                                     '(Brigitte Belcourt 10434818)',
                                             'serial_number': '43B445',
                                             'system_mac_address': None},
                'cc:be:59:50:02:a3': {       'dev_macs': ['44:65:7f:38:06:40'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:02:a3',
                                             'name': '5 Jack Koehler LN (Dawn Gellert, '
                                                     '10434571)',
                                             'serial_number': '43B519',
                                             'system_mac_address': None},
                'cc:be:59:50:02:d9': {       'dev_macs': ['38:86:f7:fd:46:8d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:02:d9',
                                             'name': '67 NITHBANK RD (PRSA STEVEN)',
                                             'serial_number': '43B534',
                                             'system_mac_address': None},
                'cc:be:59:50:03:85': {       'dev_macs': ['b0:a7:b9:3c:d7:92'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:03:85',
                                             'name': '1119 Molesworth St Wellesley '
                                                     '(Dean Savory 10434599)',
                                             'serial_number': '43B58A',
                                             'system_mac_address': None},
                'cc:be:59:50:03:db': {       'dev_macs': ['28:bd:89:d2:ca:2e'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:03:db',
                                             'name': '71 Parkview Rd Wellesley (Mike '
                                                     'Masschaele 10434324)',
                                             'serial_number': '43B5B5',
                                             'system_mac_address': None},
                'cc:be:59:50:03:dd': {       'dev_macs': ['e8:48:b8:c0:bd:95'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:03:dd',
                                             'name': '11 CAMPBELL PL (MOSER MICHAEL)',
                                             'serial_number': '43B5B6',
                                             'system_mac_address': None},
                'cc:be:59:50:03:e7': {       'dev_macs': ['44:65:7f:38:34:cc'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:03:e7',
                                             'name': '61 Parkview Dr (Scott Kelterborn '
                                                     '10434621)',
                                             'serial_number': '43B5BB',
                                             'system_mac_address': None},
                'cc:be:59:50:04:09': {       'dev_macs': ['60:db:98:09:1e:2a'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:04:09',
                                             'name': 'Harris (Temp)',
                                             'serial_number': '43B5CC',
                                             'system_mac_address': None},
                'cc:be:59:50:04:1f': {       'dev_macs': ['30:23:03:6c:1a:ae'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:04:1f',
                                             'name': '135 Ferris Dr Wellesley (Steven '
                                                     'Curliss 10434479)',
                                             'serial_number': '43B5D7',
                                             'system_mac_address': None},
                'cc:be:59:50:04:2b': {       'dev_macs': ['44:65:7f:37:fc:d4'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:04:2b',
                                             'name': '53 PARKVIEW DR (HOCHBAN ERIC)',
                                             'serial_number': '43B5DD',
                                             'system_mac_address': None},
                'cc:be:59:50:04:3f': {       'dev_macs': ['60:38:e0:0f:b6:99'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:04:3f',
                                             'name': '3603 Nafziger Rd Wellesley '
                                                     '(Richard Nafziger 10434874)',
                                             'serial_number': '43B5E7',
                                             'system_mac_address': None},
                'cc:be:59:50:04:6d': {       'dev_macs': ['60:a4:b7:a4:fc:75'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:04:6d',
                                             'name': '88 Ferris Dr (Jordan Wisniewski, '
                                                     '10434389)',
                                             'serial_number': '43B5FE',
                                             'system_mac_address': None},
                'cc:be:59:50:17:4f': {       'dev_macs': ['14:21:03:8b:74:36'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:17:4f',
                                             'name': '5106 Perth Rd. 119 (Fallon '
                                                     'Martin)',
                                             'serial_number': '43C229',
                                             'system_mac_address': None},
                'cc:be:59:50:1b:87': {       'dev_macs': ['b0:39:56:6b:cf:43'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:50:1b:87',
                                             'name': '5525 King St. Nithburg (Merlene '
                                                     'Austen)',
                                             'serial_number': '43C445',
                                             'system_mac_address': None},
                'cc:be:59:53:6f:6f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:53:6f:6f',
                                             'name': '4005 Andy St. Gadshill (Michael '
                                                     'Bean)',
                                             'serial_number': '3FE409',
                                             'system_mac_address': None},
                'cc:be:59:5a:ff:9a': {       'dev_macs': ['f4:1e:57:18:67:ed'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:5a:ff:9a',
                                             'name': '1426 Hutchison Rd Wellesley '
                                                     '(Jantzi Plumbing Inc 10430891)',
                                             'serial_number': '44AAB6',
                                             'system_mac_address': None},
                'cc:be:59:6a:df:7e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:6a:df:7e',
                                             'name': '4947 Perth Road 119 (Margaret '
                                                     'Rieck)',
                                             'serial_number': '46CE4E',
                                             'system_mac_address': None},
                'cc:be:59:6d:a9:86': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:6d:a9:86',
                                             'name': '4891 Linhaven Dr. (Gerry '
                                                     'Schmidt)',
                                             'serial_number': '400D68',
                                             'system_mac_address': None},
                'cc:be:59:6d:c9:2e': {       'dev_macs': ['ec:4f:82:13:26:16'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:6d:c9:2e',
                                             'name': '4887 Tiffany Crt. (Jonathan '
                                                     'Spiers 10423980)',
                                             'serial_number': '401152',
                                             'system_mac_address': None},
                'cc:be:59:6f:c9:4a': {       'dev_macs': ['cc:be:59:33:63:aa'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:6f:c9:4a',
                                             'name': "4955 Perth Rd. 119 (O'DRISCOLL "
                                                     'MICHELLE)',
                                             'serial_number': '55F51C',
                                             'system_mac_address': None},
                'cc:be:59:70:c9:1a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:70:c9:1a',
                                             'name': '2867 LINE 45 (WEBER PAYTON)',
                                             'serial_number': '561515',
                                             'system_mac_address': None},
                'cc:be:59:70:ea:fe': {       'dev_macs': ['ec:4f:82:a8:83:7a'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:70:ea:fe',
                                             'name': '4897 Perth Rd 119 Gadshill (Neil '
                                                     'McCann 104323786)',
                                             'serial_number': '561951',
                                             'system_mac_address': None},
                'cc:be:59:ab:20:e0': {       'dev_macs': ['74:83:c2:70:8d:f1'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:20:e0',
                                             'name': '41 Zoeger Ct Wellesley (David '
                                                     'Green 10434462)',
                                             'serial_number': '4F8C99',
                                             'system_mac_address': None},
                'cc:be:59:ab:23:62': {       'dev_macs': ['cc:be:59:c3:46:ca'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:23:62',
                                             'name': '1032 Molesworth St Wellesley '
                                                     '(Terry Steckley 10434287)',
                                             'serial_number': '4F8DCD',
                                             'system_mac_address': None},
                'cc:be:59:ab:23:6c': {       'dev_macs': ['cc:be:59:7e:d6:87'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:23:6c',
                                             'name': '3889 Nafziger Rd (Kent '
                                                     'Mccullough, 10434313)',
                                             'serial_number': '4F8DD2',
                                             'system_mac_address': None},
                'cc:be:59:ab:25:a0': {       'dev_macs': ['cc:be:59:c3:47:75'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:25:a0',
                                             'name': '29 Lawrence St Wellesley (Fred '
                                                     'Lindner 10434643)',
                                             'serial_number': '4F8ED3',
                                             'system_mac_address': None},
                'cc:be:59:ab:33:2e': {       'dev_macs': ['cc:be:59:c3:c8:0f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:33:2e',
                                             'name': '142 Ferris Drive Wellesley '
                                                     '(William Murray 10434537)',
                                             'serial_number': '4F91CE',
                                             'system_mac_address': None},
                'cc:be:59:ab:33:a0': {       'dev_macs': ['14:21:03:92:ef:25'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:33:a0',
                                             'name': '40 Pond View Dr (Sheri Zupko, '
                                                     '10434413)',
                                             'serial_number': '4F91E9',
                                             'system_mac_address': None},
                'cc:be:59:ab:33:c6': {       'dev_macs': ['00:18:61:34:f3:c7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:33:c6',
                                             'name': '40 Reiner Cr Wellesley (Len '
                                                     'Glass 10434419)',
                                             'serial_number': '4F91F0',
                                             'system_mac_address': None},
                'cc:be:59:ab:60:58': {       'dev_macs': ['84:d3:43:0f:bb:80'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:60:58',
                                             'name': '98 Lawrence St Wellesley (Jason '
                                                     'Neeb 10434715)',
                                             'serial_number': '4F9B0E',
                                             'system_mac_address': None},
                'cc:be:59:ab:77:86': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:77:86',
                                             'name': '92 Reiner Cr Wellesley',
                                             'serial_number': '4F9E9E',
                                             'system_mac_address': None},
                'cc:be:59:ab:85:a6': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:85:a6',
                                             'name': '163 Ferris Dr Wellesley',
                                             'serial_number': '4FA177',
                                             'system_mac_address': None},
                'cc:be:59:ab:86:4a': {       'dev_macs': ['44:65:7f:37:fd:25'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:86:4a',
                                             'name': '1327 Greenwood Hill Rd Wellesley '
                                                     '(Kevin Bauman 10434641)',
                                             'serial_number': '4FA19C',
                                             'system_mac_address': None},
                'cc:be:59:ab:86:a6': {       'dev_macs': ['cc:be:59:c3:48:d4'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:86:a6',
                                             'name': '106 Schweitzer Cr Wellesley (Bob '
                                                     'Tracey 10434512)',
                                             'serial_number': '4FA1C1',
                                             'system_mac_address': None},
                'cc:be:59:ab:88:d4': {       'dev_macs': ['38:d5:47:40:2b:f4'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:88:d4',
                                             'name': '109 Brown St Wellesley (John '
                                                     'Thaler 10434553)',
                                             'serial_number': '4FA247',
                                             'system_mac_address': None},
                'cc:be:59:ab:88:fe': {       'dev_macs': ['04:42:1a:10:43:f8'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:88:fe',
                                             'name': '118 David St Wellesley (Daryl '
                                                     'Kropf 10434350)',
                                             'serial_number': '4FA25C',
                                             'system_mac_address': None},
                'cc:be:59:ab:89:02': {       'dev_macs': ['cc:be:59:c3:c6:3b'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ab:89:02',
                                             'name': '72 Reiner Cr Wellesley (Chris '
                                                     'Dietrich 10404238)',
                                             'serial_number': '4FA25E',
                                             'system_mac_address': None},
                'cc:be:59:b1:12:4b': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b1:12:4b',
                                             'name': '2845 Line 45 (John Bax 10427651)',
                                             'serial_number': '4E6434',
                                             'system_mac_address': None},
                'cc:be:59:b7:3b:37': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b7:3b:37',
                                             'name': '4914 Linhaven Dr. Gadshill '
                                                     '(Richard Schneider)',
                                             'serial_number': '4D1119',
                                             'system_mac_address': None},
                'cc:be:59:b7:3b:3f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b7:3b:3f',
                                             'name': '4893 LINHAVEN DR (SCHADE JOHN & '
                                                     'BRENDA)',
                                             'serial_number': '4D111A',
                                             'system_mac_address': None},
                'cc:be:59:b9:bf:c9': {       'dev_macs': ['44:65:7f:38:34:d5'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:bf:c9',
                                             'name': '14 Nithbank Rd Wellesley (Marian '
                                                     'Weber 10434753)',
                                             'serial_number': '5123CB',
                                             'system_mac_address': None},
                'cc:be:59:b9:c2:fc': {       'dev_macs': ['cc:be:59:ff:c2:68'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c2:fc',
                                             'name': '12 Reiner Cr Wellesley (Kim '
                                                     'Heinmiller 10434608)',
                                             'serial_number': '5124DC',
                                             'system_mac_address': None},
                'cc:be:59:b9:c3:4a': {       'dev_macs': ['a8:42:a1:f5:ce:e9'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c3:4a',
                                             'name': '90 SCHWEITZER CR (NAGY DARYL)',
                                             'serial_number': '5124F6',
                                             'system_mac_address': None},
                'cc:be:59:b9:c3:56': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c3:56',
                                             'name': '1019 Molesworth St Wellesley',
                                             'serial_number': '5124FA',
                                             'system_mac_address': None},
                'cc:be:59:b9:c3:62': {       'dev_macs': [       '00:03:e6:a2:74:22',
                                                                 '00:03:e6:da:37:d8',
                                                                 'cc:be:59:c3:c6:4d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c3:62',
                                             'name': '2869 Line 45 Gadshill (Neil '
                                                     'Jenkins 10434799)',
                                             'serial_number': '5124FE',
                                             'system_mac_address': None},
                'cc:be:59:b9:c3:83': {       'dev_macs': ['44:65:7f:41:8e:8d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c3:83',
                                             'name': '13 ParkviewDr Wellesley (Mary '
                                                     'Lichty 10434481)',
                                             'serial_number': '512509',
                                             'system_mac_address': None},
                'cc:be:59:b9:c3:a7': {       'dev_macs': ['84:d3:43:6d:ea:4b'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c3:a7',
                                             'name': '103 DAVID ST (STRYKER BRITTANY)',
                                             'serial_number': '512515',
                                             'system_mac_address': None},
                'cc:be:59:b9:c3:b0': {       'dev_macs': ['60:db:98:82:07:f6'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c3:b0',
                                             'name': '3894 Nafziger Rd Welleseley '
                                                     '(Pete Weber 10434296)',
                                             'serial_number': '512518',
                                             'system_mac_address': None},
                'cc:be:59:b9:c3:e9': {       'dev_macs': ['44:65:7f:37:fd:c7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c3:e9',
                                             'name': '45 PARKVIEW DR (GILCHRIST '
                                                     'SHELDON)',
                                             'serial_number': '51252B',
                                             'system_mac_address': None},
                'cc:be:59:b9:c4:10': {       'dev_macs': ['44:65:7f:38:0f:ac'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c4:10',
                                             'name': '1159 Henry St Wellelsey (Kirk '
                                                     'Bergey 10434750)',
                                             'serial_number': '512538',
                                             'system_mac_address': None},
                'cc:be:59:b9:c4:31': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c4:31',
                                             'name': '4008 Perth Line 56 (Robert '
                                                     'Wagler)',
                                             'serial_number': '512543',
                                             'system_mac_address': None},
                'cc:be:59:b9:c4:52': {       'dev_macs': ['cc:be:59:c3:c9:ad'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c4:52',
                                             'name': '28 David St Wellesley (Ray & '
                                                     'Carol Rankin 10434730)',
                                             'serial_number': '51254E',
                                             'system_mac_address': None},
                'cc:be:59:b9:c4:a9': {       'dev_macs': ['44:65:7f:38:34:a8'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c4:a9',
                                             'name': '31 Parkview Dr Wellesley (Wayne '
                                                     'Jantzi 10434738)',
                                             'serial_number': '51256B',
                                             'system_mac_address': None},
                'cc:be:59:b9:c5:27': {       'dev_macs': ['cc:be:59:c3:48:29'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c5:27',
                                             'name': '1052 Henry St Wellesley (Judy & '
                                                     'Murray Bender 10411529)',
                                             'serial_number': '512595',
                                             'system_mac_address': None},
                'cc:be:59:b9:c5:b1': {       'dev_macs': ['cc:be:59:c3:c9:bf'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c5:b1',
                                             'name': "1072 Queen's Bush Rd Wellesley "
                                                     '(Neil & Linda Lackey10434339)',
                                             'serial_number': '5125C3',
                                             'system_mac_address': None},
                'cc:be:59:b9:c6:23': {       'dev_macs': ['44:65:7f:41:92:08'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c6:23',
                                             'name': '3 CAMPBELL PL (PFOHL JEFF)',
                                             'serial_number': '5125E9',
                                             'system_mac_address': None},
                'cc:be:59:b9:c6:50': {       'dev_macs': ['b0:6a:41:d0:c9:7c'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c6:50',
                                             'name': '1256 Greenwoodhill Dr. (Anson '
                                                     'Field)',
                                             'serial_number': '5125F8',
                                             'system_mac_address': None},
                'cc:be:59:b9:c6:9b': {       'dev_macs': ['f0:9f:c2:6d:70:56'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c6:9b',
                                             'name': '8 Parkview Dr Wellesley (D & D '
                                                     'Underwood 10434295)',
                                             'serial_number': '512611',
                                             'system_mac_address': None},
                'cc:be:59:b9:c9:fe': {       'dev_macs': ['cc:be:59:c5:60:15'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:b9:c9:fe',
                                             'name': "1014 Queen's Bush Rd Wellesley "
                                                     '(Jeff Weeks 10434549)',
                                             'serial_number': '512732',
                                             'system_mac_address': None},
                'cc:be:59:bb:0e:34': {       'dev_macs': ['44:65:7f:38:35:4a'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:bb:0e:34',
                                             'name': '5507 King St. Nithburg (Kasey '
                                                     'Hale)',
                                             'serial_number': '5A47F3',
                                             'system_mac_address': None},
                'cc:be:59:ca:db:ba': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:db:ba',
                                             'name': '39 ZINKANN CR (HOFFMAN KARI)',
                                             'serial_number': '428176',
                                             'system_mac_address': None},
                'cc:be:59:ca:dc:82': {       'dev_macs': ['cc:be:59:c3:c7:c7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dc:82',
                                             'name': '1132 Molesworth St (Laura '
                                                     'Lucchetti 10434476)',
                                             'serial_number': '42818F',
                                             'system_mac_address': None},
                'cc:be:59:ca:dc:a2': {       'dev_macs': ['b8:94:70:bc:bb:3a'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dc:a2',
                                             'name': '1091 Henry St Wellesley (Ross '
                                                     'Leis 10434674)',
                                             'serial_number': '428193',
                                             'system_mac_address': None},
                'cc:be:59:ca:dc:c2': {       'dev_macs': ['cc:be:59:c3:48:4d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dc:c2',
                                             'name': "1285 Queen's Bush Rd Wellesley "
                                                     '(Brian & Felicity Finn 10434499)',
                                             'serial_number': '428197',
                                             'system_mac_address': None},
                'cc:be:59:ca:dc:ca': {       'dev_macs': ['cc:be:59:c3:52:eb'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dc:ca',
                                             'name': '65 Lawrence St Wellesley (Lee '
                                                     'and Cory Becker 10434327)',
                                             'serial_number': '428198',
                                             'system_mac_address': None},
                'cc:be:59:ca:dd:52': {       'dev_macs': ['cc:be:59:c3:48:dd'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dd:52',
                                             'name': '1079 MOLESWORTH ST (MUNRO JERED)',
                                             'serial_number': '4281A9',
                                             'system_mac_address': None},
                'cc:be:59:ca:dd:6a': {       'dev_macs': ['cc:be:59:c3:46:82'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dd:6a',
                                             'name': '127 Ferris Dr (Mike Butler '
                                                     '10434478)',
                                             'serial_number': '4281AC',
                                             'system_mac_address': None},
                'cc:be:59:ca:dd:72': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dd:72',
                                             'name': '1128 Henry St Wellesley (Peter '
                                                     'Brand 10434670)',
                                             'serial_number': '4281AD',
                                             'system_mac_address': None},
                'cc:be:59:ca:dd:ba': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dd:ba',
                                             'name': '130 FERRIS DR (MCCUTCHEON '
                                                     'DARRYL)',
                                             'serial_number': '4281B6',
                                             'system_mac_address': None},
                'cc:be:59:ca:dd:c2': {       'dev_macs': ['cc:be:59:c3:52:be'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dd:c2',
                                             'name': '48 Lawrence St Wellesley (Brent '
                                                     '& Sarah Kipfer 10434373)',
                                             'serial_number': '4281B7',
                                             'system_mac_address': None},
                'cc:be:59:ca:dd:d2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:dd:d2',
                                             'name': "1313 Queen's Bush Rd Wellesley "
                                                     '(Sandra Erb 10434424)',
                                             'serial_number': '4281B9',
                                             'system_mac_address': None},
                'cc:be:59:ca:de:12': {       'dev_macs': ['cc:be:59:c3:47:99'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:de:12',
                                             'name': '183 FERRIS DR (GILBERT FIONA & '
                                                     'MARK)',
                                             'serial_number': '4281C1',
                                             'system_mac_address': None},
                'cc:be:59:ca:de:42': {       'dev_macs': ['cc:be:59:c3:48:c2'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:de:42',
                                             'name': '147 FERRIS DR (VONA STEFANO)',
                                             'serial_number': '4281C7',
                                             'system_mac_address': None},
                'cc:be:59:ca:de:ca': {       'dev_macs': ['dc:2c:6e:c7:55:a9'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:de:ca',
                                             'name': '174 Ferris Drive Wellesley '
                                                     '(Jeremy Szabo 10434290)',
                                             'serial_number': '4281D8',
                                             'system_mac_address': None},
                'cc:be:59:ca:de:d2': {       'dev_macs': ['cc:be:59:c3:46:94'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:de:d2',
                                             'name': '51 Zinkann Cr Wellesley (PAUL '
                                                     'KEVIN)',
                                             'serial_number': '4281D9',
                                             'system_mac_address': None},
                'cc:be:59:ca:de:fa': {       'dev_macs': ['cc:be:59:c3:53:84'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:de:fa',
                                             'name': '55 Reiner Cr Wellesley (Brad & '
                                                     'Kim Gerber 10434538)',
                                             'serial_number': '4281DE',
                                             'system_mac_address': None},
                'cc:be:59:ca:df:6a': {       'dev_macs': ['cc:be:59:7e:b6:2f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:df:6a',
                                             'name': '49 Zoeger Cr Wellesley (Matthew '
                                                     'Harney 10434400)',
                                             'serial_number': '4281EC',
                                             'system_mac_address': None},
                'cc:be:59:ca:df:e2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:df:e2',
                                             'name': "1019 QUEEN'S BUSH RD (LEIS "
                                                     'DOLORES)',
                                             'serial_number': '4281FB',
                                             'system_mac_address': None},
                'cc:be:59:ca:df:ea': {       'dev_macs': ['cc:be:59:c3:c9:5c'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:df:ea',
                                             'name': '1120 Molesworth St Wellesley '
                                                     '(Brett Moffett 10434484)',
                                             'serial_number': '4281FC',
                                             'system_mac_address': None},
                'cc:be:59:ca:df:f2': {       'dev_macs': ['cc:be:59:7e:b5:c3'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:df:f2',
                                             'name': '121 Lawrence St Wellesley '
                                                     '(Gregory Parks 10434291)',
                                             'serial_number': '4281FD',
                                             'system_mac_address': None},
                'cc:be:59:ca:e2:e2': {       'dev_macs': ['cc:be:59:c3:c9:0b'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:e2:e2',
                                             'name': '1065 Molesworth St Wellesley '
                                                     '(Michelle Martin 10434298)',
                                             'serial_number': '42825B',
                                             'system_mac_address': None},
                'cc:be:59:ca:e4:2a': {       'dev_macs': ['cc:be:59:7e:d3:ae'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:e4:2a',
                                             'name': '49 Lawrence St (Mcdonnell)',
                                             'serial_number': '428284',
                                             'system_mac_address': None},
                'cc:be:59:ca:e5:0a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ca:e5:0a',
                                             'name': '104 FERRIS DR (LOUTER SANDY)',
                                             'serial_number': '4282A0',
                                             'system_mac_address': None},
                'cc:be:59:cb:42:42': {       'dev_macs': ['44:65:7f:37:fe:72'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:42:42',
                                             'name': '1021 Maple Leaf St Wellesley '
                                                     '(Ronald Bisch 10434692)',
                                             'serial_number': '42B9A0',
                                             'system_mac_address': None},
                'cc:be:59:cb:42:b2': {       'dev_macs': ['f0:9f:c2:15:09:9f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:42:b2',
                                             'name': '30 JACK KOEHLER LN (DEWHURST '
                                                     'ADRIAN)',
                                             'serial_number': '42B9AE',
                                             'system_mac_address': None},
                'cc:be:59:cb:42:ea': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:42:ea',
                                             'name': '130 Brown St Wellesley',
                                             'serial_number': '42B9B5',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:22': {       'dev_macs': ['d0:76:8f:02:ee:1b'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:22',
                                             'name': '92 FERRIS DR (BOGOSLOWSKI PAUL)',
                                             'serial_number': '42B9BC',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:42': {       'dev_macs': ['44:65:7f:38:35:92'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:42',
                                             'name': "1005 Queen's Bush Rd (Jeremy "
                                                     'Stemmler 10434548)',
                                             'serial_number': '42B9C0',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:4a': {       'dev_macs': [       '00:03:e6:a2:64:fb',
                                                                 '00:03:e6:aa:4e:fd',
                                                                 'cc:be:59:c3:47:5a'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:4a',
                                             'name': '86 NITHBANK RD (KUEPFER GLENN)',
                                             'serial_number': '42B9C1',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:62': {       'dev_macs': ['44:65:7f:38:35:ad'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:62',
                                             'name': '75 Village Rd Wellesley (K & M '
                                                     'Gerber 10400460)',
                                             'serial_number': '42B9C4',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:7a': {       'dev_macs': ['44:65:7f:37:fe:d5'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:7a',
                                             'name': '110 NITHBANK RD (SZIMANSKI JODI)',
                                             'serial_number': '42B9C7',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:82': {       'dev_macs': ['44:65:7f:88:25:6f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:82',
                                             'name': '3 Ferris Dr Wellesley (Kristin '
                                                     'Good 10435391)',
                                             'serial_number': '42B9C8',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:8a': {       'dev_macs': ['14:21:03:a9:72:ba'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:8a',
                                             'name': '146 Gerber Meadows Dr Wellesley '
                                                     '(K&G Kingsbury 10434610)',
                                             'serial_number': '42B9C9',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:92': {       'dev_macs': [       '00:03:e6:da:37:cc',
                                                                 '44:65:7f:37:fc:dd'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:92',
                                             'name': '18 FIRELLA PL (ANYAN GARY & '
                                                     'JESSIE)',
                                             'serial_number': '42B9CA',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:b2': {       'dev_macs': ['ac:15:a2:c8:aa:75'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:b2',
                                             'name': '1205 GREENWOOD HILL RD '
                                                     '(JOHNSTONE TANYA)',
                                             'serial_number': '42B9CE',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:c2': {       'dev_macs': ['44:65:7f:41:8f:02'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:c2',
                                             'name': '22 Lawrence St Wellesley '
                                                     '(Elizabeth Eades-Kirby 10434655)',
                                             'serial_number': '42B9D0',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:ca': {       'dev_macs': ['60:db:98:09:3a:7b'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:ca',
                                             'name': '21 Firella Pl Wellesley (Philip '
                                                     'Lepage 10434718)',
                                             'serial_number': '42B9D1',
                                             'system_mac_address': None},
                'cc:be:59:cb:43:ea': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:43:ea',
                                             'name': '22 JACK KOEHLER LN (MCKAY '
                                                     'MICHAEL)',
                                             'serial_number': '42B9D5',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:1a': {       'dev_macs': ['d0:76:8f:02:ed:ad'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:1a',
                                             'name': '36 VILLAGE RD (MILLER AMANDA)',
                                             'serial_number': '42B9DB',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:22': {       'dev_macs': ['44:65:7f:38:35:5c'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:22',
                                             'name': "1127 Queen's Bush Rd (Keith Erb "
                                                     '10434601)',
                                             'serial_number': '42B9DC',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:32': {       'dev_macs': ['cc:be:59:c3:c9:a4'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:32',
                                             'name': '167 Gerber Meadows Dr Wellesley '
                                                     '(Andrea Gebbie 10434736)',
                                             'serial_number': '42B9DE',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:3a': {       'dev_macs': ['cc:be:59:c3:52:e2'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:3a',
                                             'name': '9 Village Rd Wellesley (Mark '
                                                     'Norris 10434645)',
                                             'serial_number': '42B9DF',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:6a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:6a',
                                             'name': '126 SCHWEITZER CR (LAROCCA '
                                                     'CURTIS)',
                                             'serial_number': '42B9E5',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:7a': {       'dev_macs': ['14:21:03:a9:7a:ef'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:7a',
                                             'name': '115 GERBER MEADOWS DR (JACOB '
                                                     'EDWARD (ED))',
                                             'serial_number': '42B9E7',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:8a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:8a',
                                             'name': "1025 Queen's Bush Rd Wellesley "
                                                     '(Jacob Reed 10436059)',
                                             'serial_number': '42B9E9',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:92': {       'dev_macs': ['44:65:7f:38:08:4a'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:92',
                                             'name': '38 PARKVIEW DR (TABONE TESHIA)',
                                             'serial_number': '42B9EA',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:a2': {       'dev_macs': ['44:65:7f:38:34:f0'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:a2',
                                             'name': '15 Parkview Dr Wellesley (David '
                                                     'Roth 10434761)',
                                             'serial_number': '42B9EC',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:b2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:b2',
                                             'name': '36 Zinkann Cr (Andrea Wiebe '
                                                     '10436178)',
                                             'serial_number': '42B9EE',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:ca': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:ca',
                                             'name': '82 GALENA ST',
                                             'serial_number': '42B9F1',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:da': {       'dev_macs': ['14:21:03:d1:8c:40'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:da',
                                             'name': '41 Parkview Drive Wellesley '
                                                     '(Jennifer & Jeff Sommer '
                                                     '10434517)',
                                             'serial_number': '42B9F3',
                                             'system_mac_address': None},
                'cc:be:59:cb:44:fa': {       'dev_macs': ['60:db:98:59:a4:87'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:44:fa',
                                             'name': "1145 Queen's Bush Rd (Charles & "
                                                     'Kathleen Stemmler, 10434555)',
                                             'serial_number': '42B9F7',
                                             'system_mac_address': None},
                'cc:be:59:cb:45:02': {       'dev_macs': ['44:65:7f:37:fc:8c'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:45:02',
                                             'name': '39 PARKVIEW DR (SIEWERT JASON)',
                                             'serial_number': '42B9F8',
                                             'system_mac_address': None},
                'cc:be:59:cb:68:fa': {       'dev_macs': ['44:65:7f:88:d4:4d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:68:fa',
                                             'name': '64 Zinkann Cr Wellesley (E & M '
                                                     'Krueger 10435337)',
                                             'serial_number': '42BE77',
                                             'system_mac_address': None},
                'cc:be:59:cb:6a:22': {       'dev_macs': ['44:65:7f:88:2d:34'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6a:22',
                                             'name': '12 Campbell Pl Wellesley (David '
                                                     'Redecop 10435159)',
                                             'serial_number': '42BE9C',
                                             'system_mac_address': None},
                'cc:be:59:cb:6a:3a': {       'dev_macs': ['60:db:98:55:b9:fb'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6a:3a',
                                             'name': '67 Reiner Ct Wellesley (Jeff '
                                                     'Lawrence 10434859)',
                                             'serial_number': '42BE9F',
                                             'system_mac_address': None},
                'cc:be:59:cb:6b:42': {       'dev_macs': ['d0:76:8f:02:ee:ec'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6b:42',
                                             'name': '2849 Line 45 (Jason Rose)',
                                             'serial_number': '42BEC0',
                                             'system_mac_address': None},
                'cc:be:59:cb:6b:ea': {       'dev_macs': ['ec:4f:82:a4:90:a1'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6b:ea',
                                             'name': '4014 Andy St GadsHill (Megan '
                                                     'Lunn 10435814)',
                                             'serial_number': '42BED5',
                                             'system_mac_address': None},
                'cc:be:59:cb:6c:4a': {       'dev_macs': ['48:77:46:9a:54:66'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6c:4a',
                                             'name': '40 WELWOOD AV (PENNY TOM)',
                                             'serial_number': '42BEE1',
                                             'system_mac_address': None},
                'cc:be:59:cb:6d:ca': {       'dev_macs': ['44:65:7f:89:0d:2f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6d:ca',
                                             'name': '24 Pondview Dr (Eva Irwin, '
                                                     '10434918)',
                                             'serial_number': '42BF11',
                                             'system_mac_address': None},
                'cc:be:59:cb:6d:ea': {       'dev_macs': ['44:65:7f:89:0e:d6'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6d:ea',
                                             'name': '1969 PERTH LINE 56 (SCHIEDEL '
                                                     'ALLISON)',
                                             'serial_number': '42BF15',
                                             'system_mac_address': None},
                'cc:be:59:cb:6d:fa': {       'dev_macs': ['44:65:7f:89:0c:4e'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6d:fa',
                                             'name': '182 Ferris Dr (Diane Wakem '
                                                     '10435238)',
                                             'serial_number': '42BF17',
                                             'system_mac_address': None},
                'cc:be:59:cb:6e:32': {       'dev_macs': [       '00:03:e6:9a:4a:a3',
                                                                 '00:03:e6:eb:46:da',
                                                                 'd0:76:8f:1f:57:d5'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6e:32',
                                             'name': '1005 CATHERINE ST (GARTLAND '
                                                     'CASSADI)',
                                             'serial_number': '42BF1E',
                                             'system_mac_address': None},
                'cc:be:59:cb:6e:52': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6e:52',
                                             'name': '56 ZINKANN CR (PARKINSON HOLLY)',
                                             'serial_number': '42BF22',
                                             'system_mac_address': None},
                'cc:be:59:cb:6e:da': {       'dev_macs': [       '00:03:e6:da:b1:10',
                                                                 '00:03:e6:da:ba:b5',
                                                                 '44:65:7f:72:7d:22'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:6e:da',
                                             'name': '43 Pond View Dr Wellesley (Peter '
                                                     'Erb 10434856)',
                                             'serial_number': '42BF33',
                                             'system_mac_address': None},
                'cc:be:59:cb:71:1a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:71:1a',
                                             'name': '47 Reiner Cr Wellesley',
                                             'serial_number': '42BF7B',
                                             'system_mac_address': None},
                'cc:be:59:cb:71:3a': {       'dev_macs': ['44:65:7f:88:47:ad'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:71:3a',
                                             'name': '1116 Henry St Wellesley (Allen '
                                                     'Bowslaugh 10434962)',
                                             'serial_number': '42BF7F',
                                             'system_mac_address': None},
                'cc:be:59:cb:79:ba': {       'dev_macs': ['44:65:7f:88:46:72'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:79:ba',
                                             'name': '3758 Nafziger Rd (Rhonda Lee '
                                                     'Fetterly10434984)',
                                             'serial_number': '42C08F',
                                             'system_mac_address': None},
                'cc:be:59:cb:83:ea': {       'dev_macs': ['44:65:7f:89:0d:fe'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:83:ea',
                                             'name': '171 FERRIS DR (MERLAU NELMA J)',
                                             'serial_number': '42C1D5',
                                             'system_mac_address': None},
                'cc:be:59:cb:84:ca': {       'dev_macs': ['44:65:7f:79:d7:02'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:84:ca',
                                             'name': '23 Galena St Wellesley (Andrew '
                                                     'Moore 10435707)',
                                             'serial_number': '42C1F1',
                                             'system_mac_address': None},
                'cc:be:59:cb:86:02': {       'dev_macs': ['44:65:7f:88:d4:5f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:86:02',
                                             'name': '110 Gerber Meadows Dr Wellesley '
                                                     '(Alan Bowman 10435225)',
                                             'serial_number': '42C218',
                                             'system_mac_address': None},
                'cc:be:59:cb:86:3a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:86:3a',
                                             'name': '91 Reiner Ct Wellesley (Justin '
                                                     'Martin 10435594)',
                                             'serial_number': '42C21F',
                                             'system_mac_address': None},
                'cc:be:59:cb:94:a2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:94:a2',
                                             'name': '51 Reiner Cr (Ted Richards, '
                                                     '10410746)',
                                             'serial_number': '42C3EC',
                                             'system_mac_address': None},
                'cc:be:59:cb:97:f2': {       'dev_macs': ['44:65:7f:89:0c:f0'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:97:f2',
                                             'name': '17 ZINKANN CR (NAFZIGER HEIDI)',
                                             'serial_number': '42C456',
                                             'system_mac_address': None},
                'cc:be:59:cb:97:fa': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:97:fa',
                                             'name': '42 Schweitzel Cr Wellesley',
                                             'serial_number': '42C457',
                                             'system_mac_address': None},
                'cc:be:59:cb:98:0a': {       'dev_macs': ['44:65:7f:72:7d:3d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:98:0a',
                                             'name': '14 Campbell Pl (Julie Whitney, '
                                                     '10435162)',
                                             'serial_number': '42C459',
                                             'system_mac_address': None},
                'cc:be:59:cb:99:fa': {       'dev_macs': ['ec:4f:82:ac:df:05'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:99:fa',
                                             'name': '122 SCHWEITZER CR (ANTONIOU '
                                                     'LOUIE)',
                                             'serial_number': '42C497',
                                             'system_mac_address': None},
                'cc:be:59:cb:9a:7a': {       'dev_macs': ['cc:be:59:c3:c4:f7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:9a:7a',
                                             'name': '70 Parkview Dr Wellesley (Kevin '
                                                     'Zehr 10434638)',
                                             'serial_number': '42C4A7',
                                             'system_mac_address': None},
                'cc:be:59:cb:9b:02': {       'dev_macs': ['58:d5:6e:e7:49:ce'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:9b:02',
                                             'name': '3766 Nafziger Rd (Marvin Martin, '
                                                     '10435662)',
                                             'serial_number': '42C4B8',
                                             'system_mac_address': None},
                'cc:be:59:cb:9e:42': {       'dev_macs': ['44:65:7f:88:45:b5'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:9e:42',
                                             'name': '12 Berdux Pl (Michael '
                                                     'Hayter,10434949)',
                                             'serial_number': '42C520',
                                             'system_mac_address': None},
                'cc:be:59:cb:a9:fa': {       'dev_macs': ['00:03:e6:aa:4d:57'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:a9:fa',
                                             'name': '4886 Linhaven Dr. Gadshill '
                                                     '(Brian Deleemans 10435582)',
                                             'serial_number': '42C697',
                                             'system_mac_address': None},
                'cc:be:59:cb:aa:32': {       'dev_macs': ['44:65:7f:88:47:65'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:aa:32',
                                             'name': '1018 Doering St Wellesley (Jim '
                                                     'Capling 10435327)',
                                             'serial_number': '42C69E',
                                             'system_mac_address': None},
                'cc:be:59:cb:aa:8a': {       'dev_macs': ['e4:38:83:18:98:11'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:aa:8a',
                                             'name': '139 VILLAGE RD (TURNER PAUL)',
                                             'serial_number': '42C6A9',
                                             'system_mac_address': None},
                'cc:be:59:cb:aa:92': {       'dev_macs': [       '00:03:e6:da:b5:75',
                                                                 '00:03:e6:da:bf:bf',
                                                                 '44:65:7f:88:d3:99'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:aa:92',
                                             'name': '3767 Nafziger Rd Wellesley '
                                                     '(Derek Switzer 10435073)',
                                             'serial_number': '42C6AA',
                                             'system_mac_address': None},
                'cc:be:59:cb:ab:1a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ab:1a',
                                             'name': '1890 Perth Line Wellesley (Wayne '
                                                     'Jantzi 10434423)',
                                             'serial_number': '42C6BB',
                                             'system_mac_address': None},
                'cc:be:59:cb:ac:3a': {       'dev_macs': ['84:d3:43:0f:b5:ea'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ac:3a',
                                             'name': '120 Village Rd Wellesley (Keith '
                                                     'Kuepfer 10435601)',
                                             'serial_number': '42C6DF',
                                             'system_mac_address': None},
                'cc:be:59:cb:ac:7a': {       'dev_macs': ['d0:76:8f:1f:63:e8'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ac:7a',
                                             'name': '5527 KING ST (KAUNTZ ROBERT & '
                                                     'PATRICIA)',
                                             'serial_number': '42C6E7',
                                             'system_mac_address': None},
                'cc:be:59:cb:ac:a2': {       'dev_macs': ['44:65:7f:88:47:2f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ac:a2',
                                             'name': '10 Evelyn Dr Wellesley (Nancy '
                                                     'Cook 10435554)',
                                             'serial_number': '42C6EC',
                                             'system_mac_address': None},
                'cc:be:59:cb:ac:e2': {       'dev_macs': ['44:65:7f:89:0f:c9'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ac:e2',
                                             'name': '47 ParkviewDr Wellesley (Sherri '
                                                     'Mula 10435398)',
                                             'serial_number': '42C6F4',
                                             'system_mac_address': None},
                'cc:be:59:cb:ad:a2': {       'dev_macs': ['c8:4f:86:ac:ef:2f'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ad:a2',
                                             'name': '1400 Greenwood Hill Dr Wellesley '
                                                     '(Bio-Ag Consultants 10434635)',
                                             'serial_number': '42C70C',
                                             'system_mac_address': None},
                'cc:be:59:cb:ad:ca': {       'dev_macs': ['44:65:7f:88:d4:b0'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ad:ca',
                                             'name': '3524A Nafziger Rd Wellesley '
                                                     '(Jodi Kipp 10435639)',
                                             'serial_number': '42C711',
                                             'system_mac_address': None},
                'cc:be:59:cb:b6:32': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:b6:32',
                                             'name': '4922 Linhaven Dr. Gadshill '
                                                     '(Protemp Refridgeration '
                                                     '10436516)',
                                             'serial_number': '42C81E',
                                             'system_mac_address': None},
                'cc:be:59:cb:b7:8a': {       'dev_macs': ['48:77:46:9a:4b:e4'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:b7:8a',
                                             'name': '12 PARKVIEW DR (NAFZIGER '
                                                     'JUANITA)',
                                             'serial_number': '42C849',
                                             'system_mac_address': None},
                'cc:be:59:cb:b8:12': {       'dev_macs': ['44:65:7f:41:94:1b'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:b8:12',
                                             'name': 'CLARE WAGLER (30 PONDVIEW DR '
                                                     'WELL)',
                                             'serial_number': '42C85A',
                                             'system_mac_address': None},
                'cc:be:59:cb:ba:12': {       'dev_macs': ['44:65:7f:37:fe:60'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ba:12',
                                             'name': '139 GERBER MEADOWS DR (DIETRICH '
                                                     'VALERIE A)',
                                             'serial_number': '42C89A',
                                             'system_mac_address': None},
                'cc:be:59:cb:ba:b2': {       'dev_macs': ['cc:be:59:2f:31:80'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ba:b2',
                                             'name': '5506 King St. Nithburg (Jennifer '
                                                     'Birmingham)',
                                             'serial_number': '42C8AE',
                                             'system_mac_address': None},
                'cc:be:59:cb:bb:4a': {       'dev_macs': ['b8:94:70:f3:e0:a4'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bb:4a',
                                             'name': '5 Edgewood Ct Wellelsey (Connie '
                                                     'Finlay 10434980)',
                                             'serial_number': '42C8C1',
                                             'system_mac_address': None},
                'cc:be:59:cb:bc:2a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bc:2a',
                                             'name': '73 WELWOOD AV (MCLEOD ANGUS & '
                                                     'HARTMANN GERT)',
                                             'serial_number': '42C8DD',
                                             'system_mac_address': None},
                'cc:be:59:cb:bc:32': {       'dev_macs': ['74:b6:b6:00:36:a0'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bc:32',
                                             'name': '3600 Nafziger Rd Wellesley '
                                                     '(Daniel Predhomme 10435026)',
                                             'serial_number': '42C8DE',
                                             'system_mac_address': None},
                'cc:be:59:cb:bc:6a': {       'dev_macs': [       '00:03:e6:eb:41:08',
                                                                 'b8:94:70:f3:df:65'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bc:6a',
                                             'name': '81 WELWOOD AV (DIENESCH '
                                                     'MICHELLE)',
                                             'serial_number': '42C8E5',
                                             'system_mac_address': None},
                'cc:be:59:cb:bd:0a': {       'dev_macs': ['44:65:7f:89:0f:81'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bd:0a',
                                             'name': '1380 Greenwood Hill Rd Wellesley '
                                                     '(Michael Lorentz 10435027)',
                                             'serial_number': '42C8F9',
                                             'system_mac_address': None},
                'cc:be:59:cb:bd:52': {       'dev_macs': ['b8:94:70:f3:e2:eb'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bd:52',
                                             'name': '54 PARKVIEW DR (FROKLAGE HARRY)',
                                             'serial_number': '42C902',
                                             'system_mac_address': None},
                'cc:be:59:cb:bd:62': {       'dev_macs': ['58:d9:d5:f9:2f:61'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bd:62',
                                             'name': "1116 QUEEN'S BUSH RD (COOK "
                                                     'DANIEL)',
                                             'serial_number': '42C904',
                                             'system_mac_address': None},
                'cc:be:59:cb:bd:ba': {       'dev_macs': ['44:65:7f:72:7d:a9'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bd:ba',
                                             'name': '14 Park ViewDr Wellesley (Paul '
                                                     'Abell 10434965)',
                                             'serial_number': '42C90F',
                                             'system_mac_address': None},
                'cc:be:59:cb:bd:ea': {       'dev_macs': ['44:65:7f:89:0d:e3'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bd:ea',
                                             'name': '1251 Greenwood Hill Rd Wellesley '
                                                     '(Steve & Ruth Hoffele 10434887)',
                                             'serial_number': '42C915',
                                             'system_mac_address': None},
                'cc:be:59:cb:bf:22': {       'dev_macs': ['44:65:7f:88:d3:c6'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:bf:22',
                                             'name': '77 VILLAGE RD (KANE THOMAS(TOM))',
                                             'serial_number': '42C93C',
                                             'system_mac_address': None},
                'cc:be:59:cb:c3:e2': {       'dev_macs': ['44:65:7f:88:d1:e0'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:c3:e2',
                                             'name': '28 Zinkann Cr Wellesley (Scott '
                                                     'Hache)',
                                             'serial_number': '42C9D4',
                                             'system_mac_address': None},
                'cc:be:59:cb:c5:8a': {       'dev_macs': ['44:65:7f:88:d3:e1'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:c5:8a',
                                             'name': '21 Village Rd Wellesley (Linda '
                                                     'Hergott 10435445)',
                                             'serial_number': '42CA09',
                                             'system_mac_address': None},
                'cc:be:59:cb:c8:72': {       'dev_macs': ['44:65:7f:88:d5:b5'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:c8:72',
                                             'name': '1039 MOLESWORTH ST (LOGAN JULIE)',
                                             'serial_number': '42CA66',
                                             'system_mac_address': None},
                'cc:be:59:cb:e1:1a': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e1:1a',
                                             'name': '20 FIRELLA PL (ROUBLE MARK)',
                                             'serial_number': '42CD7B',
                                             'system_mac_address': None},
                'cc:be:59:cb:e1:d2': {       'dev_macs': ['04:bc:9f:46:91:ca'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e1:d2',
                                             'name': '4016 ANDY ST (MARTIN JEFF)',
                                             'serial_number': '42CD92',
                                             'system_mac_address': None},
                'cc:be:59:cb:e2:da': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e2:da',
                                             'name': '1129 MOLESWORTH ST (PRICE '
                                                     'MATTHEW)',
                                             'serial_number': '42CDB3',
                                             'system_mac_address': None},
                'cc:be:59:cb:e4:da': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e4:da',
                                             'name': '162 Gerber Meadows Dr (Neil '
                                                     'MacDougall, 10435414)',
                                             'serial_number': '42CDF3',
                                             'system_mac_address': None},
                'cc:be:59:cb:e6:72': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e6:72',
                                             'name': '1368 Greenwood Hill (Jerry '
                                                     'Mailloux, 10435021)',
                                             'serial_number': '42CE26',
                                             'system_mac_address': None},
                'cc:be:59:cb:e6:ea': {       'dev_macs': ['44:65:7f:89:0e:fa'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e6:ea',
                                             'name': '1025 Doering St Wellesley '
                                                     '(Gretchen Robertson 10435569)',
                                             'serial_number': '42CE35',
                                             'system_mac_address': None},
                'cc:be:59:cb:e6:f2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e6:f2',
                                             'name': '2188 Perth Line Wellesley (P & B '
                                                     'Albrecht 10427546)',
                                             'serial_number': '42CE36',
                                             'system_mac_address': None},
                'cc:be:59:cb:e7:52': {       'dev_macs': [       '00:03:e6:a5:6e:93',
                                                                 '00:03:e6:a5:6e:a1',
                                                                 '00:03:e6:a5:6f:34',
                                                                 'cc:be:59:7e:b5:ba'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e7:52',
                                             'name': '1080 Henry St (Brock Gerber '
                                                     '10434382)',
                                             'serial_number': '42CE42',
                                             'system_mac_address': None},
                'cc:be:59:cb:e7:c2': {       'dev_macs': ['7c:5e:98:d5:73:21'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e7:c2',
                                             'name': '130 SCHWEITZER CR (HILL '
                                                     'JONATHAN)',
                                             'serial_number': '42CE50',
                                             'system_mac_address': None},
                'cc:be:59:cb:e8:12': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e8:12',
                                             'name': '3604 NAFZIGER RD (SCHNARR '
                                                     'MEGHAN)',
                                             'serial_number': '42CE5A',
                                             'system_mac_address': None},
                'cc:be:59:cb:e8:2a': {       'dev_macs': ['44:65:7f:89:0c:f9'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e8:2a',
                                             'name': '17 FIRELLA PL (AUSTIN KARLIE)',
                                             'serial_number': '42CE5D',
                                             'system_mac_address': None},
                'cc:be:59:cb:e9:d2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e9:d2',
                                             'name': '50 ZINKANN CR (BILODEAU MAURICE '
                                                     '& MONICA)',
                                             'serial_number': '42CE92',
                                             'system_mac_address': None},
                'cc:be:59:cb:e9:da': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e9:da',
                                             'name': '32 PARKVIEW DR (GOLDRICK TINA)',
                                             'serial_number': '42CE93',
                                             'system_mac_address': None},
                'cc:be:59:cb:e9:f2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:e9:f2',
                                             'name': '3625 NAFZIGER RD (KLASSEN FRAN)',
                                             'serial_number': '42CE96',
                                             'system_mac_address': None},
                'cc:be:59:cb:f8:b2': {       'dev_macs': [       '00:03:e6:da:b0:e6',
                                                                 '44:65:7f:89:0d:41'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:f8:b2',
                                             'name': '1169 Henry St Wellesley (Derek '
                                                     'Brick 10435296)',
                                             'serial_number': '42D06E',
                                             'system_mac_address': None},
                'cc:be:59:cb:f9:2a': {       'dev_macs': ['44:65:7f:88:d5:25'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:f9:2a',
                                             'name': "1212 QUEEN'S BUSH RD (MARTIN "
                                                     'MARJORIE)',
                                             'serial_number': '42D07D',
                                             'system_mac_address': None},
                'cc:be:59:cb:f9:52': {       'dev_macs': ['84:d3:43:0f:b9:65'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:f9:52',
                                             'name': '102 NITHBANK RD (HENNIGE JULIA)',
                                             'serial_number': '42D082',
                                             'system_mac_address': None},
                'cc:be:59:cb:f9:c2': {       'dev_macs': ['44:65:7f:72:7d:df'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:f9:c2',
                                             'name': '78 Pond View Dr Wellesley (Paul '
                                                     'Perkins 10434981)',
                                             'serial_number': '42D090',
                                             'system_mac_address': None},
                'cc:be:59:cb:fb:72': {       'dev_macs': [       '00:03:e6:da:34:50',
                                                                 '00:03:e6:eb:3c:ce',
                                                                 '44:65:7f:41:8d:c7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:fb:72',
                                             'name': '4012 Andy St Gadshill (April '
                                                     'Rose 10435328)',
                                             'serial_number': '42D0C6',
                                             'system_mac_address': None},
                'cc:be:59:cb:fc:fa': {       'dev_macs': ['44:65:7f:89:0e:bb'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:fc:fa',
                                             'name': '97 Lawrence St Wellesley (Michel '
                                                     'Houle 10434807)',
                                             'serial_number': '42D0F7',
                                             'system_mac_address': None},
                'cc:be:59:cb:ff:52': {       'dev_macs': ['44:65:7f:88:42:9d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cb:ff:52',
                                             'name': '73 David St Wellesley (Stephen '
                                                     'Bader 10434843)',
                                             'serial_number': '42D142',
                                             'system_mac_address': None},
                'cc:be:59:cc:04:6a': {       'dev_macs': ['44:65:7f:88:d3:6c'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cc:04:6a',
                                             'name': '3648 Nafziger Rd Wellesley '
                                                     '(Linda Vincent 10435908)',
                                             'serial_number': '42D1E5',
                                             'system_mac_address': None},
                'cc:be:59:cc:05:0a': {       'dev_macs': ['60:db:98:09:3c:80'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cc:05:0a',
                                             'name': '2480 GERBER RD (WAGLER PHILIP)',
                                             'serial_number': '42D1F9',
                                             'system_mac_address': None},
                'cc:be:59:cc:05:22': {       'dev_macs': ['d0:76:8f:02:f3:43'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cc:05:22',
                                             'name': '42 ZINKANN CR (FITZPATRICK '
                                                     'SAMANTHA)',
                                             'serial_number': '42D1FC',
                                             'system_mac_address': None},
                'cc:be:59:cc:06:ba': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cc:06:ba',
                                             'name': '2014 PERTH LINE (WAGLER DAVID)',
                                             'serial_number': '42D22F',
                                             'system_mac_address': None},
                'cc:be:59:cc:0a:72': {       'dev_macs': [       '00:03:e6:da:b4:1a',
                                                                 '00:03:e6:da:bf:48',
                                                                 '00:03:e6:da:c0:aa',
                                                                 'b8:94:70:84:96:73'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cc:0a:72',
                                             'name': '75 GERBER MEADOWS DR (GERBER '
                                                     'NICOLE)',
                                             'serial_number': '42D2A6',
                                             'system_mac_address': None},
                'cc:be:59:cc:0b:82': {       'dev_macs': ['44:65:7f:89:0e:19'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cc:0b:82',
                                             'name': '2 Berdux Pl Wellesley (Bruce '
                                                     'Rohr 10434970)',
                                             'serial_number': '42D2C8',
                                             'system_mac_address': None},
                'cc:be:59:cc:11:ca': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:cc:11:ca',
                                             'name': '56 VILLAGE RD (CROZIER SCOT)',
                                             'serial_number': '42D391',
                                             'system_mac_address': None},
                'cc:be:59:eb:02:66': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:eb:02:66',
                                             'name': '3737 Nafziger Rd Wellesley '
                                                     '(Wellesley Auto Care 10434379)',
                                             'serial_number': '555CC2',
                                             'system_mac_address': None},
                'cc:be:59:ef:bf:ce': {       'dev_macs': ['04:d9:f5:5a:23:08'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ef:bf:ce',
                                             'name': "1235 Queen'sBush Rd (Erb "
                                                     'Financial Services 10434422)',
                                             'serial_number': '4D3B4C',
                                             'system_mac_address': None},
                'cc:be:59:ef:f7:26': {       'dev_macs': ['44:65:7f:88:c9:55'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ef:f7:26',
                                             'name': '3648 Nafziger Rd Unit A '
                                                     'Wellesley (Eleanore Watters '
                                                     '10435900)',
                                             'serial_number': '4D4237',
                                             'system_mac_address': None},
                'cc:be:59:ef:f7:6e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ef:f7:6e',
                                             'name': '2868 Line 45 (Simon Kuepfer)',
                                             'serial_number': '4D4240',
                                             'system_mac_address': None},
                'cc:be:59:ef:f8:86': {       'dev_macs': [       '5c:0f:fb:3d:80:18',
                                                                 '5c:0f:fb:3d:80:34'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:ef:f8:86',
                                             'name': '2870 LINE 45 (HARNOCK PATRICIA)',
                                             'serial_number': '4D4263',
                                             'system_mac_address': None},
                'cc:be:59:f8:44:f7': {       'dev_macs': ['d0:76:8f:91:d2:54'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:44:f7',
                                             'name': '28 Welwood Av Wellesley (David '
                                                     'Honderich 10434636)',
                                             'serial_number': '5741A0',
                                             'system_mac_address': None},
                'cc:be:59:f8:4a:47': {       'dev_macs': ['84:d3:43:6d:eb:d7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:4a:47',
                                             'name': '59 Parkview Dr Wellesley (Bryan '
                                                     'Sansom 10436194)',
                                             'serial_number': '574246',
                                             'system_mac_address': None},
                'cc:be:59:f8:57:17': {       'dev_macs': ['44:65:7f:37:fd:2e'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:57:17',
                                             'name': "1032 Queen's Bush Rd (Steve "
                                                     'Tippin, 10434550)',
                                             'serial_number': '5743DD',
                                             'system_mac_address': None},
                'cc:be:59:f8:5d:9f': {       'dev_macs': ['44:65:7f:38:36:3d'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:5d:9f',
                                             'name': '64 VILLAGE RD (BAIRD MICHAEL '
                                                     '(MIKE))',
                                             'serial_number': '5744A9',
                                             'system_mac_address': None},
                'cc:be:59:f8:5f:3f': {       'dev_macs': ['cc:be:59:c3:c9:b6'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:5f:3f',
                                             'name': '54 SCHWEITZER CR (WILLIAMS '
                                                     'JUSTIN & ASHLEY)',
                                             'serial_number': '5744DC',
                                             'system_mac_address': None},
                'cc:be:59:f8:63:ef': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:63:ef',
                                             'name': '7 Edgewood Ct Wellesley (D &S '
                                                     'Hergott 10429943)',
                                             'serial_number': '574571',
                                             'system_mac_address': None},
                'cc:be:59:f8:64:57': {       'dev_macs': ['bc:ee:7b:f4:85:f0'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:64:57',
                                             'name': "1082 Queen's Bush Rd (Chris "
                                                     'Dietz 10434525)',
                                             'serial_number': '57457E',
                                             'system_mac_address': None},
                'cc:be:59:f8:64:7f': {       'dev_macs': ['28:87:ba:54:92:61'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:64:7f',
                                             'name': '5498 King St. Nithburg '
                                                     '(Linda&Ben Veiledal 10433920)',
                                             'serial_number': '574583',
                                             'system_mac_address': None},
                'cc:be:59:f8:64:ff': {       'dev_macs': ['44:65:7f:37:fc:a7'],
                                             'ip_address': None,
                                             'mac_address': 'cc:be:59:f8:64:ff',
                                             'name': '19 REINER CR (BENDER TIFFANY)',
                                             'serial_number': '574592',
                                             'system_mac_address': None},
                'd0:76:8f:06:e5:38': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e5:38',
                                             'name': '3952 WEIMAR LINE',
                                             'serial_number': '844BA3',
                                             'system_mac_address': None},
                'd0:76:8f:06:e5:56': {       'dev_macs': ['04:d9:f5:a3:af:d0'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e5:56',
                                             'name': '4575 WEIMAR LINE (SHERK AMOS)',
                                             'serial_number': '844BAD',
                                             'system_mac_address': None},
                'd0:76:8f:06:e5:a1': {       'dev_macs': ['b8:94:70:f3:e2:5c'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e5:a1',
                                             'name': '4161 WEIMAR LINE (KAYS ROSALIND)',
                                             'serial_number': '844BC6',
                                             'system_mac_address': None},
                'd0:76:8f:06:e5:bf': {       'dev_macs': ['a8:63:7d:7e:06:20'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e5:bf',
                                             'name': '1765 MOSER-YOUNG RD (DEVRIES '
                                                     'CLARENCE & KATIE)',
                                             'serial_number': '844BD0',
                                             'system_mac_address': None},
                'd0:76:8f:06:e5:fb': {       'dev_macs': ['3c:84:6a:0e:82:1b'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e5:fb',
                                             'name': '78 SCHWEITZER CR (WEBBER DAVID)',
                                             'serial_number': '844BE4',
                                             'system_mac_address': None},
                'd0:76:8f:06:e6:25': {       'dev_macs': ['d0:76:8f:02:ea:53'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e6:25',
                                             'name': '202 FERRIS DR (FOSTER KEDRIC)',
                                             'serial_number': '844BF2',
                                             'system_mac_address': None},
                'd0:76:8f:06:e6:3a': {       'dev_macs': ['88:de:7c:2b:cf:10'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e6:3a',
                                             'name': '4025 WEIMAR LINE (NENADOV DAVID)',
                                             'serial_number': '844BF9',
                                             'system_mac_address': None},
                'd0:76:8f:06:e6:61': {       'dev_macs': ['d0:76:8f:02:f3:dd'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e6:61',
                                             'name': '82 SCHWEITZER CR (BOESE JEFF)',
                                             'serial_number': '844C06',
                                             'system_mac_address': None},
                'd0:76:8f:06:e6:67': {       'dev_macs': ['28:bd:89:ed:93:51'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e6:67',
                                             'name': '4187 WEIMAR LINE (JANZEN TERRY)',
                                             'serial_number': '844C08',
                                             'system_mac_address': None},
                'd0:76:8f:06:e6:73': {       'dev_macs': ['3c:84:6a:75:20:33'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e6:73',
                                             'name': '1732 MOSER-YOUNG RD (MARTIN '
                                                     'MELANIE)',
                                             'serial_number': '844C0C',
                                             'system_mac_address': None},
                'd0:76:8f:06:e6:7f': {       'dev_macs': ['0c:ea:14:5a:b9:a4'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:06:e6:7f',
                                             'name': '4124 WEIMAR LINE (BUSCH BRAD)',
                                             'serial_number': '844C10',
                                             'system_mac_address': None},
                'd0:76:8f:07:18:0e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:07:18:0e',
                                             'name': '106 LAWRENCE ST (MOL RHONDDA)',
                                             'serial_number': '86788B',
                                             'system_mac_address': None},
                'd0:76:8f:07:19:2e': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:07:19:2e',
                                             'name': "1235 QUEEN'S BUSH RD (HORST "
                                                     'LOIS)',
                                             'serial_number': '8678AB',
                                             'system_mac_address': None},
                'd0:76:8f:08:0f:1b': {       'dev_macs': ['f0:09:0d:25:0d:4f'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:0f:1b',
                                             'name': "1317 QUEEN'S BUSH RD (MACARTHUR "
                                                     'JAMES)',
                                             'serial_number': '8692D9',
                                             'system_mac_address': None},
                'd0:76:8f:08:12:db': {       'dev_macs': ['d0:76:8f:02:f2:3b'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:12:db',
                                             'name': "1152 QUEEN'S BUSH RD (QUAST "
                                                     'LORRAINE)',
                                             'serial_number': '869351',
                                             'system_mac_address': None},
                'd0:76:8f:08:13:5b': {       'dev_macs': ['d0:76:8f:02:f5:cc'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:13:5b',
                                             'name': '4019 WEIMAR LINE (ZIEGLER CAROL)',
                                             'serial_number': '869361',
                                             'system_mac_address': None},
                'd0:76:8f:08:13:73': {       'dev_macs': ['00:03:e6:a6:7e:ce'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:13:73',
                                             'name': '4010 Andy St Gadshill (Katie '
                                                     'Waters 10434612)',
                                             'serial_number': '869364',
                                             'system_mac_address': None},
                'd0:76:8f:08:13:93': {       'dev_macs': [       '00:03:e6:dd:31:a8',
                                                                 'd0:76:8f:02:f1:6a'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:13:93',
                                             'name': '4211 WEIMAR LINE (MARSHALL '
                                                     'DAVID)',
                                             'serial_number': '869368',
                                             'system_mac_address': None},
                'd0:76:8f:08:13:ab': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:13:ab',
                                             'name': '1133 MOLESWORTH ST (BEDELL '
                                                     'MEGAN)',
                                             'serial_number': '86936B',
                                             'system_mac_address': None},
                'd0:76:8f:08:13:e3': {       'dev_macs': ['d0:76:8f:02:f7:9a'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:13:e3',
                                             'name': '4140 WEIMAR LINE (FINK DOUGLAS)',
                                             'serial_number': '869372',
                                             'system_mac_address': None},
                'd0:76:8f:08:14:db': {       'dev_macs': ['d4:5d:64:1b:57:00'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:14:db',
                                             'name': '1010 Hutchison Rd Wellesley '
                                                     '(HBEND ENTERPRISES)',
                                             'serial_number': '869391',
                                             'system_mac_address': None},
                'd0:76:8f:08:14:e3': {       'dev_macs': ['d0:76:8f:02:ef:b2'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:14:e3',
                                             'name': '4107 WEIMAR LINE (BRUBACHER '
                                                     'MENNO)',
                                             'serial_number': '869392',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:03': {       'dev_macs': [       '00:03:e6:db:4d:e0',
                                                                 '00:03:e6:db:50:94',
                                                                 '00:03:e6:ec:14:10',
                                                                 '68:d7:9a:66:cb:50'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:03',
                                             'name': '1690 CARMEL-KOCH RD (STRAUS DAN)',
                                             'serial_number': '869396',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:0b': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:0b',
                                             'name': '3875 NAFZIGER RD (MORDEN WANDA)',
                                             'serial_number': '869397',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:23': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:23',
                                             'name': '60 Village Rd Wellesley (James '
                                                     'Doyle 10434763)',
                                             'serial_number': '86939A',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:3b': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:3b',
                                             'name': '1826 MOSER-YOUNG RD (ENGLISH '
                                                     'SCOTT)',
                                             'serial_number': '86939D',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:43': {       'dev_macs': [       '00:03:e6:dd:2e:e4',
                                                                 '00:03:e6:dd:31:60',
                                                                 '00:03:e6:eb:49:1c'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:43',
                                             'name': '2400 Notre Dame Dr. (Russ& Linda '
                                                     'Straus)',
                                             'serial_number': '86939E',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:93': {       'dev_macs': ['d0:76:8f:02:f5:74'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:93',
                                             'name': '1064 DOERING ST (BAST KIM)',
                                             'serial_number': '8693A8',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:c3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:c3',
                                             'name': '2069 WENDELL CT (EAST SEAN)',
                                             'serial_number': '8693AE',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:d3': {       'dev_macs': [       '5c:0f:fb:17:74:cc',
                                                                 '84:d3:43:6d:e8:7d'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:d3',
                                             'name': '4175 WEIMAR LINE (MARSHALL '
                                                     'DAVID)',
                                             'serial_number': '8693B0',
                                             'system_mac_address': None},
                'd0:76:8f:08:15:f3': {       'dev_macs': [       '5c:0f:fb:2a:14:2a',
                                                                 'd0:76:8f:02:f4:09'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:15:f3',
                                             'name': '3706 NAFZIGER UNIT 6 (MCCL '
                                                     'WELLESLEY STORE)',
                                             'serial_number': '8693B4',
                                             'system_mac_address': None},
                'd0:76:8f:08:16:63': {       'dev_macs': ['d0:76:8f:02:f4:82'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:16:63',
                                             'name': '1112 MOLESWORTH ST (MILLER '
                                                     'CHERIE)',
                                             'serial_number': '8693C2',
                                             'system_mac_address': None},
                'd0:76:8f:08:16:9b': {       'dev_macs': ['d0:76:8f:02:e4:c8'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:16:9b',
                                             'name': '4090 WEIMAR LINE (ENGLER PAULA)',
                                             'serial_number': '8693C9',
                                             'system_mac_address': None},
                'd0:76:8f:08:17:33': {       'dev_macs': ['d0:76:8f:02:eb:ea'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:17:33',
                                             'name': '3998 WEIMAR LINE (YOUNG DENNIS '
                                                     '10436227)',
                                             'serial_number': '8693DC',
                                             'system_mac_address': None},
                'd0:76:8f:08:17:63': {       'dev_macs': ['d0:76:8f:1f:62:0f'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:17:63',
                                             'name': '2090 NITHVIEW CT (FREUND DENNIS)',
                                             'serial_number': '8693E2',
                                             'system_mac_address': None},
                'd0:76:8f:08:17:8b': {       'dev_macs': ['44:65:7f:88:cb:32'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:17:8b',
                                             'name': '6 POND VIEW DR (JANTZI SANDRA)',
                                             'serial_number': '8693E7',
                                             'system_mac_address': None},
                'd0:76:8f:08:17:9b': {       'dev_macs': ['14:21:03:a9:6d:7c'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:17:9b',
                                             'name': '1 POND VIEW DR (HORST MAHLON)',
                                             'serial_number': '8693E9',
                                             'system_mac_address': None},
                'd0:76:8f:08:17:bb': {       'dev_macs': ['d0:76:8f:02:f1:c2'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:17:bb',
                                             'name': '4018 WEIMAR LINE (JACOBS LISA '
                                                     '10436609)',
                                             'serial_number': '8693ED',
                                             'system_mac_address': None},
                'd0:76:8f:08:18:bb': {       'dev_macs': ['cc:be:59:2d:a4:96'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:18:bb',
                                             'name': '32 VILLAGE RD (BIGNEY KRISTOFER)',
                                             'serial_number': '86940D',
                                             'system_mac_address': None},
                'd0:76:8f:08:19:9b': {       'dev_macs': ['ec:4f:82:ac:f1:05'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:19:9b',
                                             'name': '17 LAWRENCE ST (BISCH DORIS & '
                                                     'MURRAY)',
                                             'serial_number': '869429',
                                             'system_mac_address': None},
                'd0:76:8f:08:19:ab': {       'dev_macs': ['d0:76:8f:02:f5:53'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:19:ab',
                                             'name': '1 Schweitzer Cr Wellesley (Robin '
                                                     'Palermo 10435541)',
                                             'serial_number': '86942B',
                                             'system_mac_address': None},
                'd0:76:8f:08:19:b3': {       'dev_macs': ['d0:76:8f:02:ed:4a'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:19:b3',
                                             'name': '1836 CARMEL-KOCH RD (STRAUS '
                                                     'THOMAS & STACY 10436566)',
                                             'serial_number': '86942C',
                                             'system_mac_address': None},
                'd0:76:8f:08:1c:13': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:1c:13',
                                             'name': '4080 WEIMAR LINE (THOMPSON DALE)',
                                             'serial_number': '869478',
                                             'system_mac_address': None},
                'd0:76:8f:08:1c:33': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:1c:33',
                                             'name': '67 CRAB APPLE CT (NYMAN ROBERT)',
                                             'serial_number': '86947C',
                                             'system_mac_address': None},
                'd0:76:8f:08:1c:83': {       'dev_macs': ['04:d9:f5:74:47:10'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:08:1c:83',
                                             'name': '38 GALENA ST (GRANT JACOB)',
                                             'serial_number': '869486',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9e:53': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9e:53',
                                             'name': '151 FERRIS DR (WETTLAUFER '
                                                     'MATTHEW)',
                                             'serial_number': '872418',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9e:73': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9e:73',
                                             'name': '1116 MOLESWORTH ST (PIASECKI '
                                                     'LUKE)',
                                             'serial_number': '87241C',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9e:d3': {       'dev_macs': ['d0:76:8f:1f:5b:3a'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9e:d3',
                                             'name': '1095 HENRY ST (JANZEN KARL)',
                                             'serial_number': '872428',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9f:03': {       'dev_macs': ['d0:76:8f:1f:5a:a0'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9f:03',
                                             'name': '4054 WEIMAR LINE (DANIS-HARKNESS '
                                                     'SUZANNE)',
                                             'serial_number': '87242E',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9f:6b': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9f:6b',
                                             'name': '47 CRAB APPLE CT (BILLINGS '
                                                     'KANDICE)',
                                             'serial_number': '87243B',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9f:9b': {       'dev_macs': ['84:d3:43:a5:63:cb'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9f:9b',
                                             'name': '2082 NITHVIEW CT (HACKBART DAN)',
                                             'serial_number': '872441',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9f:ab': {       'dev_macs': ['d0:76:8f:1f:52:4a'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9f:ab',
                                             'name': '143 VILLAGE RD (EVANS GARY)',
                                             'serial_number': '872443',
                                             'system_mac_address': None},
                'd0:76:8f:1b:9f:c3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:9f:c3',
                                             'name': '21 Zoeger Ct Wellesley',
                                             'serial_number': '872446',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a0:9b': {       'dev_macs': ['48:77:46:f8:b1:c0'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a0:9b',
                                             'name': '63 GERBER MEADOWS DR (CLARK '
                                                     'CATRINA)',
                                             'serial_number': '872461',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a5:03': {       'dev_macs': ['18:c2:41:82:a6:01'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a5:03',
                                             'name': "1245 QUEEN'S BUSH RD",
                                             'serial_number': '8724EE',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a5:33': {       'dev_macs': ['60:db:98:09:3c:1d'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a5:33',
                                             'name': '53 DAVID ST (STEINMANN HAROLD)',
                                             'serial_number': '8724F4',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a5:83': {       'dev_macs': ['2c:3f:0b:ac:78:a5'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a5:83',
                                             'name': '2196 GERBER RD (WELLESLEY HOME '
                                                     'CENTRE)',
                                             'serial_number': '8724FE',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a5:b3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a5:b3',
                                             'name': '43 CRAB APPLE CT',
                                             'serial_number': '872504',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a6:8b': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a6:8b',
                                             'name': '2086 NITHVIEW CT (BERARIU '
                                                     'SIMONA)',
                                             'serial_number': '87251F',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a6:a3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a6:a3',
                                             'name': '105 LAWRENCE ST (WALKER CHRIS)',
                                             'serial_number': '872522',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a7:f3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a7:f3',
                                             'name': '79 GERBER MEADOWS DR',
                                             'serial_number': '87254C',
                                             'system_mac_address': None},
                'd0:76:8f:1b:a8:a3': {       'dev_macs': [       '00:03:e6:dd:31:2e',
                                                                 '00:03:e6:eb:3c:d0'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:1b:a8:a3',
                                             'name': '107 DAVID ST (SCHROTH ANDRE)',
                                             'serial_number': '872562',
                                             'system_mac_address': None},
                'd0:76:8f:aa:47:c4': {       'dev_macs': ['d0:76:8f:91:d6:27'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:47:c4',
                                             'name': '44 ZINKANN CR (QUINT MELISSA)',
                                             'serial_number': '956E1E',
                                             'system_mac_address': None},
                'd0:76:8f:aa:48:4c': {       'dev_macs': ['d0:76:8f:91:d5:fb'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:48:4c',
                                             'name': '126 Brown St (Will Bender)',
                                             'serial_number': '956E2F',
                                             'system_mac_address': None},
                'd0:76:8f:aa:48:5c': {       'dev_macs': ['d0:76:8f:91:d3:30'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:48:5c',
                                             'name': '41 LAWRENCE ST (PETRIE CHERYL)',
                                             'serial_number': '956E31',
                                             'system_mac_address': None},
                'd0:76:8f:aa:48:64': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:48:64',
                                             'name': '187 FERRIS DR (TRUSSELL JASON)',
                                             'serial_number': '956E32',
                                             'system_mac_address': None},
                'd0:76:8f:aa:48:ec': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:48:ec',
                                             'name': '3705 NAFZIGER RD (HORST CARL)',
                                             'serial_number': '956E43',
                                             'system_mac_address': None},
                'd0:76:8f:aa:49:04': {       'dev_macs': ['d0:76:8f:91:d0:0d'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:49:04',
                                             'name': '69 LAWRENCE ST (ANDERSON PILAR)',
                                             'serial_number': '956E46',
                                             'system_mac_address': None},
                'd0:76:8f:aa:49:14': {       'dev_macs': ['d0:76:8f:8d:0d:cc'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:49:14',
                                             'name': '89 PARKVIEW DR (HARCOURT CHRIS)',
                                             'serial_number': '956E48',
                                             'system_mac_address': None},
                'd0:76:8f:aa:49:1c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:49:1c',
                                             'name': '3 LAWRENCE ST (POOLE ELECTRIC '
                                                     'LTD.)',
                                             'serial_number': '956E49',
                                             'system_mac_address': None},
                'd0:76:8f:aa:49:74': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:49:74',
                                             'name': '35 ZOEGER CT (BUERKLE HEATHER)',
                                             'serial_number': '956E54',
                                             'system_mac_address': None},
                'd0:76:8f:aa:49:cc': {       'dev_macs': ['00:01:21:2a:eb:dd'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:49:cc',
                                             'name': "1379 QUEEN'S BUSH RD (TOWNSHIP "
                                                     'OF WELLESLEY)',
                                             'serial_number': '956E5F',
                                             'system_mac_address': None},
                'd0:76:8f:aa:4b:d4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:4b:d4',
                                             'name': '2080 Nithview Ct Nithberg '
                                                     '(Robert Witzel 10436127)',
                                             'serial_number': '956EA0',
                                             'system_mac_address': None},
                'd0:76:8f:aa:4e:c4': {       'dev_macs': ['d0:76:8f:8d:0f:42'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:4e:c4',
                                             'name': '95 Parkview Dr (Bonnie Roth)',
                                             'serial_number': '956EFE',
                                             'system_mac_address': None},
                'd0:76:8f:aa:4e:e4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:4e:e4',
                                             'name': '3706 NAFZIGER RD (KATES '
                                                     'KINETICS)',
                                             'serial_number': '956F02',
                                             'system_mac_address': None},
                'd0:76:8f:aa:4f:2c': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:4f:2c',
                                             'name': '73 LAWRENCE ST (SINDALL ASHLEY)',
                                             'serial_number': '956F0B',
                                             'system_mac_address': None},
                'd0:76:8f:aa:4f:74': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:4f:74',
                                             'name': '131 GERBER MEADOWS DR (ROBERTSON '
                                                     'AMY)',
                                             'serial_number': '956F14',
                                             'system_mac_address': None},
                'd0:76:8f:aa:50:4c': {       'dev_macs': ['d0:76:8f:91:ca:da'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:50:4c',
                                             'name': '3890 NAFZIGER RD (GOOD JEFF)',
                                             'serial_number': '956F2F',
                                             'system_mac_address': None},
                'd0:76:8f:aa:51:04': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:51:04',
                                             'name': '3700 NAFZIGER RD (RESOLUTE '
                                                     'WELLNESS INC)',
                                             'serial_number': '956F46',
                                             'system_mac_address': None},
                'd0:76:8f:aa:51:94': {       'dev_macs': ['d0:76:8f:02:f5:11'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:51:94',
                                             'name': '13 DAVID ST (BLACKBURN LISA)',
                                             'serial_number': '956F58',
                                             'system_mac_address': None},
                'd0:76:8f:aa:51:a4': {       'dev_macs': ['d0:76:8f:91:d6:1c'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:51:a4',
                                             'name': '86 SCHWEITZER CR (LIVINGSTONE '
                                                     'CHAD)',
                                             'serial_number': '956F5A',
                                             'system_mac_address': None},
                'd0:76:8f:aa:52:44': {       'dev_macs': ['d0:76:8f:02:f4:b9'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:52:44',
                                             'name': '3974 WEIMAR LINE (VOGEL BARB & '
                                                     'KLAUS)',
                                             'serial_number': '956F6E',
                                             'system_mac_address': None},
                'd0:76:8f:aa:52:64': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:52:64',
                                             'name': '15 VILLAGE RD (HUBER PAUL)',
                                             'serial_number': '956F72',
                                             'system_mac_address': None},
                'd0:76:8f:aa:52:d4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:52:d4',
                                             'name': '121 DAVID ST (COBER BRUCE)',
                                             'serial_number': '956F80',
                                             'system_mac_address': None},
                'd0:76:8f:aa:52:e4': {       'dev_macs': ['d0:76:8f:91:d4:85'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:52:e4',
                                             'name': '2839 LINE 45 (MCDOWELL CAROLINE)',
                                             'serial_number': '956F82',
                                             'system_mac_address': None},
                'd0:76:8f:aa:52:f4': {       'dev_macs': ['7c:5a:1c:9b:15:0e'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:52:f4',
                                             'name': '1315  Hutchinson Rd Nithburg '
                                                     '(Leis Pet Distribution 10434255)',
                                             'serial_number': '956F84',
                                             'system_mac_address': None},
                'd0:76:8f:aa:53:2c': {       'dev_macs': ['b8:94:70:9e:d0:06'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:53:2c',
                                             'name': '2 SHADYWOOD CT (ENS GERHARD)',
                                             'serial_number': '956F8B',
                                             'system_mac_address': None},
                'd0:76:8f:aa:53:34': {       'dev_macs': ['d0:76:8f:91:d3:88'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:53:34',
                                             'name': '1038 MOLESWORTH ST (SHAW TRICIA)',
                                             'serial_number': '956F8C',
                                             'system_mac_address': None},
                'd0:76:8f:aa:53:54': {       'dev_macs': ['d0:76:8f:91:d4:22'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:53:54',
                                             'name': '66 ZINKANN CR (GOLDEN SARAH)',
                                             'serial_number': '956F90',
                                             'system_mac_address': None},
                'd0:76:8f:aa:53:8c': {       'dev_macs': ['d0:76:8f:91:d4:7a'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:53:8c',
                                             'name': '158 FERRIS DR (TOMPKINS TANYA)',
                                             'serial_number': '956F97',
                                             'system_mac_address': None},
                'd0:76:8f:aa:53:f4': {       'dev_macs': ['d0:76:8f:91:d5:1f'],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:aa:53:f4',
                                             'name': '138 FERRIS DR (LANTEIGNE CONNIE)',
                                             'serial_number': '956FA4',
                                             'system_mac_address': None},
                'd0:76:8f:e7:24:43': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'd0:76:8f:e7:24:43',
                                             'name': '81 POND VIEW DR (FUTHER KELSEY)',
                                             'serial_number': '9C2FE0',
                                             'system_mac_address': None},
                'ec:4f:82:08:1d:6d': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:08:1d:6d',
                                             'name': '5432 PERTH ROAD 119 (MARTIN '
                                                     'HARLAN & MICHELLE)',
                                             'serial_number': '1C5BA3',
                                             'system_mac_address': None},
                'ec:4f:82:0d:68:5f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:0d:68:5f',
                                             'name': '5484 KING ST (HOLMES MICHAEL)',
                                             'serial_number': '205DBA',
                                             'system_mac_address': None},
                'ec:4f:82:0d:69:5b': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:0d:69:5b',
                                             'name': '2073 WENDELL CT (DARCY MARTIN '
                                                     '10432882)',
                                             'serial_number': '205DD6',
                                             'system_mac_address': None},
                'ec:4f:82:0d:69:7f': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:0d:69:7f',
                                             'name': "1235 QUEEN'S BUSH RD (JOHNSON "
                                                     'VAL)',
                                             'serial_number': '205DDA',
                                             'system_mac_address': None},
                'ec:4f:82:0d:69:e2': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:0d:69:e2',
                                             'name': '5556 KING ST (KURUCZ ROSE)',
                                             'serial_number': '205DE5',
                                             'system_mac_address': None},
                'ec:4f:82:28:82:2a': {       'dev_macs': ['94:83:c4:60:11:10'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:28:82:2a',
                                             'name': '2076 Perth line 56 (DORTMAN BROS '
                                                     'BARN EQUIP)',
                                             'serial_number': '26AC6C',
                                             'system_mac_address': None},
                'ec:4f:82:32:25:67': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:32:25:67',
                                             'name': '4910 Linhaven Dr. Gadshill '
                                                     '(Sheila Hausleitner)',
                                             'serial_number': '248ABD',
                                             'system_mac_address': None},
                'ec:4f:82:44:18:57': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:44:18:57',
                                             'name': '5321 PERTH ROAD 119 (CEDAR CREST '
                                                     'ACRES INC)',
                                             'serial_number': '23A00A',
                                             'system_mac_address': None},
                'ec:4f:82:59:ea:b3': {       'dev_macs': ['60:db:98:09:3c:3e'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:59:ea:b3',
                                             'name': '5465 King St. Nithburg (James '
                                                     'Green)',
                                             'serial_number': '2AD0C9',
                                             'system_mac_address': None},
                'ec:4f:82:59:eb:23': {       'dev_macs': [       '00:03:e6:a1:bf:84',
                                                                 '00:03:e6:a4:d9:c4',
                                                                 'ec:4f:82:a9:a9:1a'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:59:eb:23',
                                             'name': '5479 King St. Nithburg (Shannon '
                                                     'Gingerich)',
                                             'serial_number': '2AD0D7',
                                             'system_mac_address': None},
                'ec:4f:82:59:ec:53': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:59:ec:53',
                                             'name': '5514 KING ST (COLLINS GARY)',
                                             'serial_number': '2AD0FD',
                                             'system_mac_address': None},
                'ec:4f:82:59:ec:bb': {       'dev_macs': ['04:bc:9f:46:90:ac'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:59:ec:bb',
                                             'name': '5485 KING ST (MARTIN KEITH)',
                                             'serial_number': '2AD10A',
                                             'system_mac_address': None},
                'ec:4f:82:59:ed:73': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:59:ed:73',
                                             'name': '5535 KING ST (WILSON CAROL D)',
                                             'serial_number': '2AD121',
                                             'system_mac_address': None},
                'ec:4f:82:5b:70:a3': {       'dev_macs': ['ec:4f:82:ac:f0:1b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:5b:70:a3',
                                             'name': '5554 King St. Nithburg (Raymond '
                                                     'Robinson 10433731)',
                                             'serial_number': '2B0187',
                                             'system_mac_address': None},
                'ec:4f:82:5b:70:cb': {       'dev_macs': ['ec:4f:82:ac:f0:51'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:5b:70:cb',
                                             'name': '5577 King St. Nithburg (John '
                                                     'Vleeming 10433729)',
                                             'serial_number': '2B018C',
                                             'system_mac_address': None},
                'ec:4f:82:5b:73:3b': {       'dev_macs': ['ec:4f:82:a8:87:6a'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:5b:73:3b',
                                             'name': '2698 Line 47 (Sharon Schnarr)',
                                             'serial_number': '2B01DA',
                                             'system_mac_address': None},
                'ec:4f:82:5b:80:bb': {       'dev_macs': ['ec:4f:82:ac:f0:fc'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:5b:80:bb',
                                             'name': '2068 Wendell Ct. Nithburg (Lisa '
                                                     'Koch 10433828)',
                                             'serial_number': '2B038A',
                                             'system_mac_address': None},
                'ec:4f:82:5b:88:23': {       'dev_macs': ['60:db:98:09:1f:ab'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:5b:88:23',
                                             'name': '5587 King St. Nithburg (DREHER '
                                                     'KEVIN 10433735)',
                                             'serial_number': '2B0477',
                                             'system_mac_address': None},
                'ec:4f:82:63:5a:e7': {       'dev_macs': ['48:f8:b3:f5:9c:29'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:63:5a:e7',
                                             'name': '5597 PERTH LINE 56A (CHOWN '
                                                     'CHRISTOPHER)',
                                             'serial_number': '2BD5B2',
                                             'system_mac_address': None},
                'ec:4f:82:7d:bb:7f': {       'dev_macs': ['24:f5:a2:7f:2b:32'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:bb:7f',
                                             'name': '2084 NithviewCt Nithview (Rola '
                                                     'Scott 10434213)',
                                             'serial_number': '300728',
                                             'system_mac_address': None},
                'ec:4f:82:7d:bc:a7': {       'dev_macs': ['b8:94:70:f3:e4:6c'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:bc:a7',
                                             'name': '2091 Nithview Ct. Nithburg (Jeff '
                                                     'Foster)',
                                             'serial_number': '3007BC',
                                             'system_mac_address': None},
                'ec:4f:82:7d:c8:a9': {       'dev_macs': ['d0:76:8f:91:ce:81'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:c8:a9',
                                             'name': '2067 Wendell Ct. Nithburg '
                                                     '(SUTHERLAND HEATHER)',
                                             'serial_number': '300D81',
                                             'system_mac_address': None},
                'ec:4f:82:7d:c8:c9': {       'dev_macs': ['44:65:7f:37:fe:06'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:c8:c9',
                                             'name': '5576 KING ST (WILLIAMS DAVE)',
                                             'serial_number': '300D91',
                                             'system_mac_address': None},
                'ec:4f:82:7d:c9:4d': {       'dev_macs': ['d4:01:c3:0a:d7:1b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:c9:4d',
                                             'name': '5444 KING ST (KUEPFER PETER)',
                                             'serial_number': '300DD3',
                                             'system_mac_address': None},
                'ec:4f:82:7d:ca:b3': {       'dev_macs': ['78:98:e8:33:66:2b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:ca:b3',
                                             'name': '5407 King St. Nithburg (Chantal '
                                                     'Valliere 10434016)',
                                             'serial_number': '300E86',
                                             'system_mac_address': None},
                'ec:4f:82:7d:ca:c1': {       'dev_macs': ['cc:be:59:c5:5f:6a'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:ca:c1',
                                             'name': '5500 King St. Nithburg (Sheila '
                                                     'Gustafson)',
                                             'serial_number': '300E8D',
                                             'system_mac_address': None},
                'ec:4f:82:7d:ca:e7': {       'dev_macs': ['24:f5:a2:ef:29:2a'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:ca:e7',
                                             'name': '2089 Nithview Ct. Nithburg '
                                                     '(HAWTHORNTHWAITE ROGER)',
                                             'serial_number': '300EA0',
                                             'system_mac_address': None},
                'ec:4f:82:7d:cb:4d': {       'dev_macs': ['ec:4f:82:a9:a6:92'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:cb:4d',
                                             'name': '5455 King St. Nithburg (Steven '
                                                     'Careless)',
                                             'serial_number': '300ED3',
                                             'system_mac_address': None},
                'ec:4f:82:7d:cb:61': {       'dev_macs': ['ec:4f:82:ac:ef:ca'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:cb:61',
                                             'name': '5573 King St. Nithburg (Sean & '
                                                     'Tara Haselden)',
                                             'serial_number': '300EDD',
                                             'system_mac_address': None},
                'ec:4f:82:7d:cb:a3': {       'dev_macs': ['74:da:da:97:2c:18'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7d:cb:a3',
                                             'name': '2074 Wendell Ct. Nithburg '
                                                     '(Wilmer&Marjorie Gerber)',
                                             'serial_number': '300EFE',
                                             'system_mac_address': None},
                'ec:4f:82:7e:08:a4': {       'dev_macs': ['90:ec:77:8e:7e:59'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:08:a4',
                                             'name': '1430 Hutchison Rd Nithburg (T & '
                                                     'T Power Group 10431156)',
                                             'serial_number': '35E92C',
                                             'system_mac_address': None},
                'ec:4f:82:7e:08:aa': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:08:aa',
                                             'name': '1000 MAPLE LEAF ST (TOWNSHIP OF '
                                                     'WELLESLEY)',
                                             'serial_number': '35E92E',
                                             'system_mac_address': None},
                'ec:4f:82:7e:08:ce': {       'dev_macs': ['70:8b:cd:ae:8b:48'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:08:ce',
                                             'name': '2249 PERTH LINE 56 (SCHULTZ '
                                                     'KEITH & FAYE)',
                                             'serial_number': '35E93A',
                                             'system_mac_address': None},
                'ec:4f:82:7e:09:6a': {       'dev_macs': ['38:94:ed:bf:72:bf'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:09:6a',
                                             'name': '64 FERRIS DR (FANE PAMELA)',
                                             'serial_number': '35E96E',
                                             'system_mac_address': None},
                'ec:4f:82:7e:0a:69': {       'dev_macs': ['cc:be:59:32:94:86'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:0a:69',
                                             'name': '1 Pond View Unit 303 (Mary Lou '
                                                     'Shantz 10434317)',
                                             'serial_number': '35E9C3',
                                             'system_mac_address': None},
                'ec:4f:82:7e:0a:75': {       'dev_macs': ['04:d4:c4:c3:1b:a8'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:0a:75',
                                             'name': '120 David St Wellesley (Larry '
                                                     'Roes, 10411398)',
                                             'serial_number': '35E9C7',
                                             'system_mac_address': None},
                'ec:4f:82:7e:3a:36': {       'dev_macs': ['48:f8:b3:64:c8:d4'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:3a:36',
                                             'name': '5466 Brown St. Nithburg (Nancy '
                                                     'Firth)',
                                             'serial_number': '35F9A6',
                                             'system_mac_address': None},
                'ec:4f:82:7e:3a:60': {       'dev_macs': ['cc:f4:11:7e:5c:77'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:3a:60',
                                             'name': '107 SCHWEITZER CR (MAGUIRE MARC)',
                                             'serial_number': '35F9B4',
                                             'system_mac_address': None},
                'ec:4f:82:7e:3a:c9': {       'dev_macs': ['ec:4f:82:a9:a8:e4'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:3a:c9',
                                             'name': '5499 BROWN ST (ROES RICK)',
                                             'serial_number': '35F9D7',
                                             'system_mac_address': None},
                'ec:4f:82:7e:3b:0b': {       'dev_macs': ['cc:be:59:2d:a5:53'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:7e:3b:0b',
                                             'name': '2291 Perth Line 56 (Ruthann Erb '
                                                     '10433944)',
                                             'serial_number': '35F9ED',
                                             'system_mac_address': None},
                'ec:4f:82:8d:3e:88': {       'dev_macs': ['cc:be:59:c5:5f:e8'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:3e:88',
                                             'name': '1100 MOLESWORTH STREET (FRITZ '
                                                     'MICHAEL)',
                                             'serial_number': '321018',
                                             'system_mac_address': None},
                'ec:4f:82:8d:43:d8': {       'dev_macs': ['e4:6c:d1:4e:f1:b0'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:43:d8',
                                             'name': '2087 Nithview Ct. Nithburg '
                                                     '(Jeannine Campbell 10434015)',
                                             'serial_number': '3210C2',
                                             'system_mac_address': None},
                'ec:4f:82:8d:43:e8': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:43:e8',
                                             'name': '2050 Perth Line 56 Nithburg (Ken '
                                                     'Snider)',
                                             'serial_number': '3210C4',
                                             'system_mac_address': None},
                'ec:4f:82:8d:44:88': {       'dev_macs': ['00:03:e6:a8:0e:9e'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:44:88',
                                             'name': '2268 PERTH LINE (JANTZI AARON & '
                                                     'MELISSA)',
                                             'serial_number': '3210D8',
                                             'system_mac_address': None},
                'ec:4f:82:8d:44:98': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:44:98',
                                             'name': '5470 Brown St. Nithburg (John '
                                                     'Tabbert 10412107)',
                                             'serial_number': '3210DA',
                                             'system_mac_address': None},
                'ec:4f:82:8d:45:a0': {       'dev_macs': ['ec:4f:82:a9:a7:a9'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:45:a0',
                                             'name': '5469 BROWN ST (MCHUGH ROBERT)',
                                             'serial_number': '3210FB',
                                             'system_mac_address': None},
                'ec:4f:82:8d:46:08': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:46:08',
                                             'name': '5569 KING ST (TAYLOR KATHLEEN)',
                                             'serial_number': '321108',
                                             'system_mac_address': None},
                'ec:4f:82:8d:47:88': {       'dev_macs': ['00:03:e6:a4:d9:84'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:47:88',
                                             'name': '5490 King St. Nithburg (Jody '
                                                     'Broda)',
                                             'serial_number': '321138',
                                             'system_mac_address': None},
                'ec:4f:82:8d:48:08': {       'dev_macs': ['ec:4f:82:ac:e3:2b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:48:08',
                                             'name': '5489 Brown St. Nithburg (Janice '
                                                     'Castell)',
                                             'serial_number': '321148',
                                             'system_mac_address': None},
                'ec:4f:82:8d:48:28': {       'dev_macs': ['48:77:46:9a:55:ce'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:48:28',
                                             'name': '2039B Perth Line 56 Nithburg '
                                                     '(Delmer & Mary Erb)',
                                             'serial_number': '32114C',
                                             'system_mac_address': None},
                'ec:4f:82:8d:48:c0': {       'dev_macs': ['60:db:98:08:f8:b2'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:48:c0',
                                             'name': '5471 BROWN ST (ABBOTT TYLER)',
                                             'serial_number': '32115F',
                                             'system_mac_address': None},
                'ec:4f:82:8d:49:28': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:49:28',
                                             'name': '4912 PERTH ROAD 119 (DOWNEY BILL '
                                                     '& KAREN)',
                                             'serial_number': '32116C',
                                             'system_mac_address': None},
                'ec:4f:82:8d:49:40': {       'dev_macs': ['b8:94:70:f3:e7:fd'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:49:40',
                                             'name': '5428 King St. Nithburg (Raymond '
                                                     'Lemieux)',
                                             'serial_number': '32116F',
                                             'system_mac_address': None},
                'ec:4f:82:8d:49:58': {       'dev_macs': ['48:77:46:f8:b1:89'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:49:58',
                                             'name': '5476 King St. Nithburg (Ed '
                                                     'Jernigan)',
                                             'serial_number': '321172',
                                             'system_mac_address': None},
                'ec:4f:82:8d:4a:28': {       'dev_macs': ['44:65:7f:38:10:ba'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:4a:28',
                                             'name': "1099 Queen's Bush Rd Wellesley "
                                                     '(Stephen Bender 10434690)',
                                             'serial_number': '32118C',
                                             'system_mac_address': None},
                'ec:4f:82:8d:4a:48': {       'dev_macs': ['e4:6c:d1:4f:b1:c2'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:4a:48',
                                             'name': '5478 BROWN ST (LINGELBACH SARAH)',
                                             'serial_number': '321190',
                                             'system_mac_address': None},
                'ec:4f:82:8d:4d:a0': {       'dev_macs': ['cc:be:59:2d:9d:e8'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:4d:a0',
                                             'name': '5483 BROWN ST (SHUSHACK PAUL)',
                                             'serial_number': '3211FB',
                                             'system_mac_address': None},
                'ec:4f:82:8d:4e:f8': {       'dev_macs': [       '00:03:e6:a1:c1:ba',
                                                                 '5c:0f:fb:2f:ce:d4',
                                                                 'd0:76:8f:1f:5a:5e'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:4e:f8',
                                             'name': '5521 Brown St. Nithburg (John '
                                                     'MacDonald)',
                                             'serial_number': '321226',
                                             'system_mac_address': None},
                'ec:4f:82:8d:50:68': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:50:68',
                                             'name': '4920 LINHAVEN DR (FAULHAFER DON)',
                                             'serial_number': '321254',
                                             'system_mac_address': None},
                'ec:4f:82:8d:51:88': {       'dev_macs': ['b8:94:70:f3:d6:e3'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:51:88',
                                             'name': '2070 Wendell ct. Nithburg (D & R '
                                                     'Lantz)',
                                             'serial_number': '321278',
                                             'system_mac_address': None},
                'ec:4f:82:8d:52:38': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:52:38',
                                             'name': '5586 PERTH LINE 56A (DAN ZEHR '
                                                     '10400068)',
                                             'serial_number': '32128E',
                                             'system_mac_address': None},
                'ec:4f:82:8d:55:d0': {       'dev_macs': ['ec:4f:82:a9:a8:ed'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:55:d0',
                                             'name': '2074 Albert St. Nithburg (Jim '
                                                     'Brubacher)',
                                             'serial_number': '321301',
                                             'system_mac_address': None},
                'ec:4f:82:8d:55:e0': {       'dev_macs': ['ec:4f:82:ad:07:46'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:55:e0',
                                             'name': '2066 WENDELL CT (PRINCE JODY & '
                                                     'MICHELLE)',
                                             'serial_number': '321303',
                                             'system_mac_address': None},
                'ec:4f:82:8d:75:f8': {       'dev_macs': ['00:03:e6:a5:70:70'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:75:f8',
                                             'name': '5460 King St. Nithburg (Robert '
                                                     'Litke)',
                                             'serial_number': '321706',
                                             'system_mac_address': None},
                'ec:4f:82:8d:db:b8': {       'dev_macs': ['cc:be:59:7e:b7:a0'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:db:b8',
                                             'name': "1358 Queen's Bush Rd Wellesley "
                                                     '(Rick Dietrich 10434312)',
                                             'serial_number': '3223BE',
                                             'system_mac_address': None},
                'ec:4f:82:8d:db:d0': {       'dev_macs': ['cc:be:59:7e:ba:dc'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:db:d0',
                                             'name': "1303 Queen's Bush Rd Wellesley "
                                                     '(Karyn Shantz 10434342)',
                                             'serial_number': '3223C1',
                                             'system_mac_address': None},
                'ec:4f:82:8d:de:88': {       'dev_macs': ['14:21:03:8b:6a:1d'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:de:88',
                                             'name': '3840 Nafziger Rd-2 Wellesley '
                                                     '(Barry & Denise Bender 10434362)',
                                             'serial_number': '322418',
                                             'system_mac_address': None},
                'ec:4f:82:8d:de:b0': {       'dev_macs': ['cc:be:59:7e:b6:da'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:de:b0',
                                             'name': '15 ZOEGER CT (MARTIN RAY)',
                                             'serial_number': '32241D',
                                             'system_mac_address': None},
                'ec:4f:82:8d:de:b8': {       'dev_macs': ['84:7b:eb:ec:5e:b6'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:de:b8',
                                             'name': '157 David St (Wellesley Men '
                                                     'Chrch)',
                                             'serial_number': '32241E',
                                             'system_mac_address': None},
                'ec:4f:82:8d:de:e8': {       'dev_macs': ['cc:be:59:7e:b7:a9'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:de:e8',
                                             'name': '3609 Nafziger Rd Wellesley '
                                                     '(Christie Howes 10434371)',
                                             'serial_number': '322424',
                                             'system_mac_address': None},
                'ec:4f:82:8d:df:40': {       'dev_macs': ['cc:be:59:7e:b6:d1'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:df:40',
                                             'name': '1155 Henry St. Wellesley (Ron & '
                                                     'Sheila Cressman 10434337)',
                                             'serial_number': '32242F',
                                             'system_mac_address': None},
                'ec:4f:82:8d:ed:c8': {       'dev_macs': ['cc:be:59:7e:b5:9f'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:ed:c8',
                                             'name': "1300 Queen's Bush Rd Wellesley "
                                                     '(Bill & Bonnie Love 10434529)',
                                             'serial_number': '322600',
                                             'system_mac_address': None},
                'ec:4f:82:8d:ee:30': {       'dev_macs': ['18:e8:29:44:b5:13'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:ee:30',
                                             'name': '25 Zoeger Ct Wellesley (David '
                                                     'Henderson 10434421)',
                                             'serial_number': '32260D',
                                             'system_mac_address': None},
                'ec:4f:82:8d:ee:88': {       'dev_macs': [       '00:03:e6:da:a7:15',
                                                                 '5c:0f:fb:17:6e:95',
                                                                 'cc:be:59:2f:2f:b5'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:ee:88',
                                             'name': '113 Lawrence St Wellesley (Aimee '
                                                     'Higgins 10434513)',
                                             'serial_number': '322618',
                                             'system_mac_address': None},
                'ec:4f:82:8d:f3:30': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:f3:30',
                                             'name': '31 David St Wellesley (G & S '
                                                     'Poole 10434325)',
                                             'serial_number': '3226AD',
                                             'system_mac_address': None},
                'ec:4f:82:8d:f3:d8': {       'dev_macs': ['cc:be:59:7e:d2:a0'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:f3:d8',
                                             'name': '10 Berdux Pl Wellesley (Calvin '
                                                     'Schmidt 10434396)',
                                             'serial_number': '3226C2',
                                             'system_mac_address': None},
                'ec:4f:82:8d:f4:08': {       'dev_macs': ['14:21:03:8b:6c:59'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:f4:08',
                                             'name': '1147 HENRY ST (KRALL MELISSA)',
                                             'serial_number': '3226C8',
                                             'system_mac_address': None},
                'ec:4f:82:8d:f5:58': {       'dev_macs': ['b8:94:70:f3:e2:46'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:f5:58',
                                             'name': '13 POND VIEW DR (PETERS DIANE & '
                                                     'GARRY)',
                                             'serial_number': '3226F2',
                                             'system_mac_address': None},
                'ec:4f:82:8d:f7:10': {       'dev_macs': ['cc:be:59:7e:b9:8f'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:f7:10',
                                             'name': "1291 Queen's Bush Rd Wellesley "
                                                     '(Dan Gingrich 10434293)',
                                             'serial_number': '322729',
                                             'system_mac_address': None},
                'ec:4f:82:8d:f7:a0': {       'dev_macs': ['cc:be:59:7e:b5:4e'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:f7:a0',
                                             'name': "1352 Queen's Bush Rd Wellesley "
                                                     '(Jeff Moyer 10434297)',
                                             'serial_number': '32273B',
                                             'system_mac_address': None},
                'ec:4f:82:8d:fa:80': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:fa:80',
                                             'name': '61 ZINKANN CR (RYERSE PATRICK)',
                                             'serial_number': '322797',
                                             'system_mac_address': None},
                'ec:4f:82:8d:fd:58': {       'dev_macs': ['d0:76:8f:02:ee:47'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:fd:58',
                                             'name': '1124 HENRY ST (BYER CAMERON & '
                                                     'JAYNE)',
                                             'serial_number': '3227F2',
                                             'system_mac_address': None},
                'ec:4f:82:8d:fe:88': {       'dev_macs': ['cc:be:59:7e:ba:16'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:fe:88',
                                             'name': "1287 QUEEN'S BUSH RD (GLAUSER "
                                                     'TED)',
                                             'serial_number': '322818',
                                             'system_mac_address': None},
                'ec:4f:82:8d:fe:d8': {       'dev_macs': ['cc:be:59:7e:ba:8b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:fe:d8',
                                             'name': '3631 Nafziger Rd Wellesley '
                                                     '(James & Karen Gerber 10432639)',
                                             'serial_number': '322822',
                                             'system_mac_address': None},
                'ec:4f:82:8d:ff:a0': {       'dev_macs': ['14:21:03:a9:70:1b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8d:ff:a0',
                                             'name': "1346 Queen's Bush Rd Wellesley "
                                                     '(Sherry Cormier 10434322)',
                                             'serial_number': '32283B',
                                             'system_mac_address': None},
                'ec:4f:82:8e:00:00': {       'dev_macs': ['14:21:03:8b:72:f7'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:00:00',
                                             'name': '1300 Greenwood Hill Dr Wellesley '
                                                     '(A & B Detzler 10434514)',
                                             'serial_number': '322847',
                                             'system_mac_address': None},
                'ec:4f:82:8e:00:e8': {       'dev_macs': [       '5c:0f:fb:33:c1:8d',
                                                                 '5c:0f:fb:33:c1:9a',
                                                                 'cc:be:59:7e:b6:80'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:00:e8',
                                             'name': '1035 Doering St Wellesley (Paul '
                                                     'Cook 10434450)',
                                             'serial_number': '322864',
                                             'system_mac_address': None},
                'ec:4f:82:8e:01:50': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:01:50',
                                             'name': '45 POND VIEW DR (DOTZERT DORIS)',
                                             'serial_number': '322871',
                                             'system_mac_address': None},
                'ec:4f:82:8e:01:78': {       'dev_macs': ['cc:be:59:c5:5e:f5'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:01:78',
                                             'name': '1026 Molesworth St Wellesley '
                                                     '(Ron & Brenda Poole 10411884)',
                                             'serial_number': '322876',
                                             'system_mac_address': None},
                'ec:4f:82:8e:01:98': {       'dev_macs': ['cc:be:59:7e:b8:a5'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:01:98',
                                             'name': '115 David St Wellesley (Arnold & '
                                                     'Arnetta Erb 10402642)',
                                             'serial_number': '32287A',
                                             'system_mac_address': None},
                'ec:4f:82:8e:03:d0': {       'dev_macs': ['cc:be:59:7e:b8:ff'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:03:d0',
                                             'name': '21 LAWRENCE ST (ROBERTS ANN)',
                                             'serial_number': '3228C1',
                                             'system_mac_address': None},
                'ec:4f:82:8e:05:e0': {       'dev_macs': ['44:65:7f:37:fc:b9'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:05:e0',
                                             'name': "1097 QUEEN'S BUSH RD (SEBBEN "
                                                     'TED)',
                                             'serial_number': '322903',
                                             'system_mac_address': None},
                'ec:4f:82:8e:06:00': {       'dev_macs': ['44:65:7f:37:fe:7b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:06:00',
                                             'name': '101 Brown St Wellesley (Mary '
                                                     'Jane Falk 10434554)',
                                             'serial_number': '322907',
                                             'system_mac_address': None},
                'ec:4f:82:8e:06:40': {       'dev_macs': ['b8:94:70:f3:e2:67'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:06:40',
                                             'name': '2838 LINE 45 (ADAIR MEGAN)',
                                             'serial_number': '32290F',
                                             'system_mac_address': None},
                'ec:4f:82:8e:06:58': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:06:58',
                                             'name': '3640 NAFZIGER RD (MCCREA LIZ & '
                                                     'JEFF)',
                                             'serial_number': '322912',
                                             'system_mac_address': None},
                'ec:4f:82:8e:08:20': {       'dev_macs': ['14:21:03:d1:7d:2b'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:08:20',
                                             'name': '81 David St Wellesley (Rob Greer '
                                                     '10434764)',
                                             'serial_number': '32294B',
                                             'system_mac_address': None},
                'ec:4f:82:8e:08:28': {       'dev_macs': ['44:65:7f:38:35:53'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:08:28',
                                             'name': "1010 QUEEN'S BUSH RD (SPROULE "
                                                     'BOB)',
                                             'serial_number': '32294C',
                                             'system_mac_address': None},
                'ec:4f:82:8e:0b:40': {       'dev_macs': ['b8:94:70:bc:be:7e'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:8e:0b:40',
                                             'name': '18 POND VIEW DR (MULLEN THOMAS)',
                                             'serial_number': '3229AF',
                                             'system_mac_address': None},
                'ec:4f:82:92:fd:d0': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:92:fd:d0',
                                             'name': '2864 LINE 45 (ROTH HAROLD)',
                                             'serial_number': '329646',
                                             'system_mac_address': None},
                'ec:4f:82:92:ff:80': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:92:ff:80',
                                             'name': '2837 Line 45 (Jamie Leach '
                                                     '10423704)',
                                             'serial_number': '32967C',
                                             'system_mac_address': None},
                'ec:4f:82:a1:ae:b3': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:a1:ae:b3',
                                             'name': '2005A Perth Line 56 Wellesley '
                                                     '(Emily Lindner 10436103)',
                                             'serial_number': '348108',
                                             'system_mac_address': None},
                'ec:4f:82:ae:ae:ff': {       'dev_macs': ['bc:22:28:85:b9:b2'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:ae:ae:ff',
                                             'name': '5568 KING ST (MEADOWS BRETT)',
                                             'serial_number': '3B5ADE',
                                             'system_mac_address': None},
                'ec:4f:82:ae:bb:93': {       'dev_macs': ['68:d7:9a:5c:bd:4f'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:ae:bb:93',
                                             'name': '5555 King St. Nithburg (Phillip '
                                                     'Garrett)',
                                             'serial_number': '3B6128',
                                             'system_mac_address': None},
                'ec:4f:82:ae:bd:0d': {       'dev_macs': ['ec:4f:82:a9:a4:fd'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:ae:bd:0d',
                                             'name': '2039A Perth Line 56 Nithburg '
                                                     '(Shawn Conn)',
                                             'serial_number': '3B61E5',
                                             'system_mac_address': None},
                'ec:4f:82:b9:1a:58': {       'dev_macs': ['70:a7:41:a6:7e:b1'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:b9:1a:58',
                                             'name': '2828 Line 45 (Paul&Linda Bogle '
                                                     '10433906)',
                                             'serial_number': '2FE9F5',
                                             'system_mac_address': None},
                'ec:4f:82:b9:1a:f8': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:b9:1a:f8',
                                             'name': '4025 Perth Line 56 (Mark '
                                                     'Nafziger)',
                                             'serial_number': '2FEA09',
                                             'system_mac_address': None},
                'ec:4f:82:b9:22:70': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:b9:22:70',
                                             'name': '4924 Linhaven Dr. Gadshill (Adam '
                                                     '& Jenny Riehl 10431415)',
                                             'serial_number': '2FEAF8',
                                             'system_mac_address': None},
                'ec:4f:82:be:a8:09': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:be:a8:09',
                                             'name': '1172 Henry St Wellesley '
                                                     '(Futher-Franklin Funeral '
                                                     '10435038)',
                                             'serial_number': '37883E',
                                             'system_mac_address': None},
                'ec:4f:82:cc:10:fc': {       'dev_macs': ['04:bc:9f:46:ac:bb'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:cc:10:fc',
                                             'name': '2846 LINE 45 (WALSH VICTORIA)',
                                             'serial_number': '38FF58',
                                             'system_mac_address': None},
                'ec:4f:82:da:16:34': {       'dev_macs': ['60:db:98:55:b5:57'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:da:16:34',
                                             'name': '2829 Line 45 Gadshill (Randy '
                                                     'Swartzentruber 10424337)',
                                             'serial_number': '2E7E69',
                                             'system_mac_address': None},
                'ec:4f:82:db:ef:1c': {       'dev_macs': ['cc:be:59:7e:b6:5c'],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:db:ef:1c',
                                             'name': '2863 Line 45 (Greg Ganselves)',
                                             'serial_number': '2EB97E',
                                             'system_mac_address': None},
                'ec:4f:82:dc:ed:ec': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:dc:ed:ec',
                                             'name': '2857 Line 45 (Hayward Martin)',
                                             'serial_number': '2ED957',
                                             'system_mac_address': None},
                'ec:4f:82:dc:f1:a4': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'ec:4f:82:dc:f1:a4',
                                             'name': '5559 Perth Rd. 119 (Shane Stever '
                                                     '10420470)',
                                             'serial_number': '2ED9CE',
                                             'system_mac_address': None},
                'f8:85:f9:17:21:18': {       'dev_macs': ['60:db:98:09:1a:a4'],
                                             'ip_address': None,
                                             'mac_address': 'f8:85:f9:17:21:18',
                                             'name': '70 NITHBANK RD (NOLAN DENNIS)',
                                             'serial_number': 'DFFD0C',
                                             'system_mac_address': None},
                'f8:85:f9:17:28:c8': {       'dev_macs': ['14:21:03:8b:73:91'],
                                             'ip_address': None,
                                             'mac_address': 'f8:85:f9:17:28:c8',
                                             'name': '142 LAWRENCE ST (REED GEOFF)',
                                             'serial_number': 'DFFE02',
                                             'system_mac_address': None},
                'f8:85:f9:17:29:00': {       'dev_macs': ['60:db:98:59:6e:00'],
                                             'ip_address': None,
                                             'mac_address': 'f8:85:f9:17:29:00',
                                             'name': '54 Zinkann Cr Wellesley '
                                                     '(Stephanie Gohl 10434787)',
                                             'serial_number': 'DFFE09',
                                             'system_mac_address': None},
                'f8:85:f9:17:29:28': {       'dev_macs': ['84:d3:43:6d:ef:f7'],
                                             'ip_address': None,
                                             'mac_address': 'f8:85:f9:17:29:28',
                                             'name': '40 VILLAGE RD (ERB WENDY)',
                                             'serial_number': 'DFFE0E',
                                             'system_mac_address': None},
                'f8:85:f9:17:29:d0': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'f8:85:f9:17:29:d0',
                                             'name': '3956 WEIMAR LINE (ARNOTT TAMARA)',
                                             'serial_number': 'DFFE23',
                                             'system_mac_address': None},
                'f8:85:f9:17:2a:f0': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'f8:85:f9:17:2a:f0',
                                             'name': '3610 NAFZIGER RD (BILLSON '
                                                     'JOANNA)',
                                             'serial_number': 'DFFE47',
                                             'system_mac_address': None},
                'f8:85:f9:6f:20:b5': {       'dev_macs': [],
                                             'ip_address': None,
                                             'mac_address': 'f8:85:f9:6f:20:b5',
                                             'name': '4838 Perth Road 119 (Jean '
                                                     'Langille)',
                                             'serial_number': 'BF24C6',
                                             'system_mac_address': None}
        }
        exp_ne_msg = {
                'data': {       'interfaces': [       {       'admin_up': True,
                                                              'id': 'front',
                                                              'mac_address': 'cc:be:59:47:ab:98',
                                                              'name': 'Craft-FE: front',
                                                              'oper_up': True,
                                                              'poller_hash': 'NzE0MTI5MTMwNzZmcm9udA=='},
                                                      {       'admin_up': False,
                                                              'id': 'rear-A',
                                                              'name': 'Craft-FE: '
                                                                      'rear-A',
                                                              'oper_up': False,
                                                              'poller_hash': 'NzE0MTI5MTMwNzZyZWFyLUE='},
                                                      {       'admin_up': False,
                                                              'id': 'rear-B',
                                                              'name': 'Craft-FE: '
                                                                      'rear-B',
                                                              'oper_up': False,
                                                              'poller_hash': 'NzE0MTI5MTMwNzZyZWFyLUI='}],
                                'manufacturer': 'Calix',
                                'model': 'E7-20',
                                'modules': [       {       'id': '1',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '1/1',
                                                                                         'in_discards': '172748',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '4437786',
                                                                                         'in_octets': '44590283729828',
                                                                                         'in_ucast_pkts': '95995365925',
                                                                                         'links': [       {       'in_nucast_pkts': '156787',
                                                                                                                  'in_octets': '8183751538',
                                                                                                                  'in_ucast_pkts': '19730132',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9693795',
                                                                                                                  'out_octets': '54596088812',
                                                                                                                  'out_ucast_pkts': '47878708',
                                                                                                                  'poller_hash': 'MzIyNkYycGVlcmludGZwb24gMTcy'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '3760461523',
                                                                                                                  'in_ucast_pkts': '18361801',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1339071',
                                                                                                                  'out_octets': '70549520952',
                                                                                                                  'out_ucast_pkts': '55798354',
                                                                                                                  'poller_hash': 'NEQ0MjM3cGVlcmludGZwb24gNTI3'},
                                                                                                          {       'in_nucast_pkts': '2384',
                                                                                                                  'in_octets': '145409401582',
                                                                                                                  'in_ucast_pkts': '542846153',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5601482',
                                                                                                                  'out_octets': '2202292228204',
                                                                                                                  'out_ucast_pkts': '1601830828',
                                                                                                                  'poller_hash': 'NjQwQjg0cGVlcmludGZwb24gNTI4'},
                                                                                                          {       'in_nucast_pkts': '6',
                                                                                                                  'in_octets': '75996546078',
                                                                                                                  'in_ucast_pkts': '404755094',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9870781',
                                                                                                                  'out_octets': '1267066622970',
                                                                                                                  'out_ucast_pkts': '898055207',
                                                                                                                  'poller_hash': 'NDJEMUU1cGVlcmludGZwb24gNTI5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '89321075993',
                                                                                                                  'in_ucast_pkts': '331453889',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '57',
                                                                                                                  'out_octets': '1207739454853',
                                                                                                                  'out_ucast_pkts': '1027257800',
                                                                                                                  'poller_hash': 'NzNBQkUxcGVlcmludGZwb24gMTM3NTQ3'},
                                                                                                          {       'in_nucast_pkts': '36',
                                                                                                                  'in_octets': '288978149469',
                                                                                                                  'in_ucast_pkts': '540920784',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5611055',
                                                                                                                  'out_octets': '1282082640440',
                                                                                                                  'out_ucast_pkts': '1073003968',
                                                                                                                  'poller_hash': 'NjQwQjFBcGVlcmludGZwb24gMTM3NjY2'},
                                                                                                          {       'in_nucast_pkts': '826836',
                                                                                                                  'in_octets': '7638134032',
                                                                                                                  'in_ucast_pkts': '32792287',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39418315',
                                                                                                                  'out_octets': '83131146754',
                                                                                                                  'out_ucast_pkts': '75114098',
                                                                                                                  'poller_hash': 'NUE2QTUxcGVlcmludGZwb24gMTM4MTEx'},
                                                                                                          {       'in_nucast_pkts': '469',
                                                                                                                  'in_octets': '4680914934648',
                                                                                                                  'in_ucast_pkts': '8871026434',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14071240',
                                                                                                                  'out_octets': '8058128001026',
                                                                                                                  'out_ucast_pkts': '8896216037',
                                                                                                                  'poller_hash': 'NUE2QzgzcGVlcmludGZwb24gMTM4ODEy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1861246674171',
                                                                                                                  'in_ucast_pkts': '2745302225',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '75',
                                                                                                                  'out_octets': '3810103788860',
                                                                                                                  'out_ucast_pkts': '3897870947',
                                                                                                                  'poller_hash': 'ODY5Mzk3cGVlcmludGZwb24gMTQwMDM1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '16525172990',
                                                                                                                  'in_ucast_pkts': '65487729',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '90122780074',
                                                                                                                  'out_ucast_pkts': '135168013',
                                                                                                                  'poller_hash': 'ODY3OEFCcGVlcmludGZwb24gMTQwNTE2'},
                                                                                                          {       'in_nucast_pkts': '29621',
                                                                                                                  'in_octets': '44442529034',
                                                                                                                  'in_ucast_pkts': '153916249',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9918791',
                                                                                                                  'out_octets': '100857311262',
                                                                                                                  'out_ucast_pkts': '130528996',
                                                                                                                  'poller_hash': 'ODcyNEVFcGVlcmludGZwb24gMTQzNzMy'},
                                                                                                          {       'in_nucast_pkts': '7947516',
                                                                                                                  'in_octets': '347613346840',
                                                                                                                  'in_ucast_pkts': '1194881380',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10903553127',
                                                                                                                  'out_octets': '18621243132098',
                                                                                                                  'out_ucast_pkts': '3106591191',
                                                                                                                  'poller_hash': 'OTU2RjJGcGVlcmludGZwb24gMTQ1MTE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '957741445244',
                                                                                                                  'in_ucast_pkts': '116913296200',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '1046097147561',
                                                                                                                  'out_ucast_pkts': '1313849486',
                                                                                                                  'poller_hash': 'OTU2RTQ5cGVlcmludGZwb24gMTQ1Nzcy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '476070975611',
                                                                                                                  'in_ucast_pkts': '69745778124',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '169',
                                                                                                                  'out_octets': '817716268781',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QTlDMzZBcGVlcmludGZwb24gMTQ2NDUx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '187111430646',
                                                                                                                  'in_ucast_pkts': '1038889776',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '20',
                                                                                                                  'out_octets': '3909870071172',
                                                                                                                  'out_ucast_pkts': '2828664552',
                                                                                                                  'poller_hash': 'QzgzQkMxcGVlcmludGZwb24gMTYyNDgz'},
                                                                                                          {       'in_nucast_pkts': '948815',
                                                                                                                  'in_octets': '49910692484',
                                                                                                                  'in_ucast_pkts': '260000959',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61546449',
                                                                                                                  'out_octets': '1567682492530',
                                                                                                                  'out_ucast_pkts': '1286149536',
                                                                                                                  'poller_hash': 'QzgzQkJFcGVlcmludGZwb24gMTY0MjU2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3160681212',
                                                                                                                  'in_ucast_pkts': '10878071',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '39187841749',
                                                                                                                  'out_ucast_pkts': '35116132',
                                                                                                                  'poller_hash': 'MTI3NDE5NXBlZXJpbnRmcG9uIDE2Nzk1OA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '580888822682',
                                                                                                                  'in_ucast_pkts': '23049604361',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '16602389041182',
                                                                                                                  'out_ucast_pkts': '12057190323',
                                                                                                                  'poller_hash': 'Q0JFMDcxcGVlcmludGZwb24gMTY4Mzg5'},
                                                                                                          {       'in_nucast_pkts': '63',
                                                                                                                  'in_octets': '2412648999213',
                                                                                                                  'in_ucast_pkts': '2157430160',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878567',
                                                                                                                  'out_octets': '1986554904429',
                                                                                                                  'out_ucast_pkts': '2127243194',
                                                                                                                  'poller_hash': 'QzgzQjE3cGVlcmludGZwb24gMTg0NzM1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '623082883234',
                                                                                                                  'in_ucast_pkts': '25204925642',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '16304780654239',
                                                                                                                  'out_ucast_pkts': '12160570696',
                                                                                                                  'poller_hash': 'REZGRTQ3cGVlcmludGZwb24gMTg2NTA3'},
                                                                                                          {       'in_nucast_pkts': '54874',
                                                                                                                  'in_octets': '58492491491',
                                                                                                                  'in_ucast_pkts': '148063402',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3212341',
                                                                                                                  'out_octets': '467658485197',
                                                                                                                  'out_ucast_pkts': '353470255',
                                                                                                                  'poller_hash': 'QzgxREMzcGVlcmludGZwb24gMTk2MjM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '58472040479',
                                                                                                                  'in_ucast_pkts': '327474330',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1087824903634',
                                                                                                                  'out_ucast_pkts': '812411350',
                                                                                                                  'poller_hash': 'MTE5NTE0MXBlZXJpbnRmcG9uIDIxMjkxNg=='},
                                                                                                          {       'in_nucast_pkts': '832139',
                                                                                                                  'in_octets': '560997896276',
                                                                                                                  'in_ucast_pkts': '6180244033',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43750319',
                                                                                                                  'out_octets': '14964221716612',
                                                                                                                  'out_ucast_pkts': '11110998925',
                                                                                                                  'poller_hash': 'MTE5NTFCN3BlZXJpbnRmcG9uIDIxNjE4Nw=='},
                                                                                                          {       'in_nucast_pkts': '147771',
                                                                                                                  'in_octets': '50596051326',
                                                                                                                  'in_ucast_pkts': '335151754',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878607',
                                                                                                                  'out_octets': '2538778560317',
                                                                                                                  'out_ucast_pkts': '1754066759',
                                                                                                                  'poller_hash': 'MTE5NTFCMnBlZXJpbnRmcG9uIDIxNzY0Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '87465860317',
                                                                                                                  'in_ucast_pkts': '71390913',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '20',
                                                                                                                  'out_octets': '36935808451',
                                                                                                                  'out_ucast_pkts': '82425144',
                                                                                                                  'poller_hash': 'MTI3NDczRXBlZXJpbnRmcG9uIDIzMDcxNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '454964383673',
                                                                                                                  'in_ucast_pkts': '182468851289',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '48',
                                                                                                                  'out_octets': '6050323459431',
                                                                                                                  'out_ucast_pkts': '4786241060',
                                                                                                                  'poller_hash': 'MTI3NDE4RXBlZXJpbnRmcG9uIDI0MzY4MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '290551080274',
                                                                                                                  'in_ucast_pkts': '304899716',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '528399700125',
                                                                                                                  'out_ucast_pkts': '488640813',
                                                                                                                  'poller_hash': 'MTU4RkI2NHBlZXJpbnRmcG9uIDI0NjQ2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '114311097246',
                                                                                                                  'in_ucast_pkts': '47532727073',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18',
                                                                                                                  'out_octets': '890434808114',
                                                                                                                  'out_ucast_pkts': '770707866',
                                                                                                                  'poller_hash': 'MTI3NDgzRnBlZXJpbnRmcG9uIDI2MjY4OA=='},
                                                                                                          {       'in_nucast_pkts': '3876',
                                                                                                                  'in_octets': '158377159985',
                                                                                                                  'in_ucast_pkts': '396630105',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4829510',
                                                                                                                  'out_octets': '434188533211',
                                                                                                                  'out_ucast_pkts': '508224534',
                                                                                                                  'poller_hash': 'MTI3NDczQnBlZXJpbnRmcG9uIDI3NDA3MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '169726796686',
                                                                                                                  'in_ucast_pkts': '184720273',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '53653861395',
                                                                                                                  'out_ucast_pkts': '168717362',
                                                                                                                  'poller_hash': 'MTI5NUU4Q3BlZXJpbnRmcG9uIDI4MDQyNA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '169062275554',
                                                                                         'out_octets': '439117581622356',
                                                                                         'out_ucast_pkts': '177638611557',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/1',
                                                                                                                 'id': '1/1',
                                                                                                                 'links': [       {       'connected_time': '8378181',
                                                                                                                                          'fiber_length': 22978,
                                                                                                                                          'poller_hash': 'MzIyNkYycGVlcnBvbnBvbiAxNzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '1993839',
                                                                                                                                          'fiber_length': 22664,
                                                                                                                                          'poller_hash': 'NEQ0MjM3cGVlcnBvbnBvbiA1Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535032',
                                                                                                                                          'fiber_length': 22649,
                                                                                                                                          'poller_hash': 'NjQwQjg0cGVlcnBvbnBvbiA1Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8529309',
                                                                                                                                          'fiber_length': 22770,
                                                                                                                                          'poller_hash': 'NDJEMUU1cGVlcnBvbnBvbiA1Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535295',
                                                                                                                                          'fiber_length': 23182,
                                                                                                                                          'poller_hash': 'NzNBQkUxcGVlcnBvbnBvbiAxMzc1NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535068',
                                                                                                                                          'fiber_length': 22271,
                                                                                                                                          'poller_hash': 'NjQwQjFBcGVlcnBvbnBvbiAxMzc2NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '67855892',
                                                                                                                                          'fiber_length': 23090,
                                                                                                                                          'poller_hash': 'NUE2QTUxcGVlcnBvbnBvbiAxMzgxMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8570933',
                                                                                                                                          'fiber_length': 22308,
                                                                                                                                          'poller_hash': 'NUE2QzgzcGVlcnBvbnBvbiAxMzg4MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '1957832',
                                                                                                                                          'fiber_length': 23505,
                                                                                                                                          'poller_hash': 'ODY5Mzk3cGVlcnBvbnBvbiAxNDAwMzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535333',
                                                                                                                                          'fiber_length': 22348,
                                                                                                                                          'poller_hash': 'ODY3OEFCcGVlcnBvbnBvbiAxNDA1MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8570820',
                                                                                                                                          'fiber_length': 22310,
                                                                                                                                          'poller_hash': 'ODcyNEVFcGVlcnBvbnBvbiAxNDM3MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535251',
                                                                                                                                          'fiber_length': 23613,
                                                                                                                                          'poller_hash': 'OTU2RjJGcGVlcnBvbnBvbiAxNDUxMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '21887065',
                                                                                                                                          'fiber_length': 22822,
                                                                                                                                          'poller_hash': 'OTU2RTQ5cGVlcnBvbnBvbiAxNDU3NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '1462678',
                                                                                                                                          'fiber_length': 22532,
                                                                                                                                          'poller_hash': 'QTlDMzZBcGVlcnBvbnBvbiAxNDY0NTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535141',
                                                                                                                                          'fiber_length': 23329,
                                                                                                                                          'poller_hash': 'QzgzQkMxcGVlcnBvbnBvbiAxNjI0ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '18323160',
                                                                                                                                          'fiber_length': 22984,
                                                                                                                                          'poller_hash': 'QzgzQkJFcGVlcnBvbnBvbiAxNjQyNTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '3204783',
                                                                                                                                          'fiber_length': 22826,
                                                                                                                                          'poller_hash': 'MTI3NDE5NXBlZXJwb25wb24gMTY3OTU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '11539619',
                                                                                                                                          'fiber_length': 22581,
                                                                                                                                          'poller_hash': 'Q0JFMDcxcGVlcnBvbnBvbiAxNjgzODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535146',
                                                                                                                                          'fiber_length': 23249,
                                                                                                                                          'poller_hash': 'QzgzQjE3cGVlcnBvbnBvbiAxODQ3MzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535213',
                                                                                                                                          'fiber_length': 23133,
                                                                                                                                          'poller_hash': 'REZGRTQ3cGVlcnBvbnBvbiAxODY1MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '2741850',
                                                                                                                                          'fiber_length': 22532,
                                                                                                                                          'poller_hash': 'QzgxREMzcGVlcnBvbnBvbiAxOTYyMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '13292304',
                                                                                                                                          'fiber_length': 22732,
                                                                                                                                          'poller_hash': 'MTE5NTE0MXBlZXJwb25wb24gMjEyOTE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '73164803',
                                                                                                                                          'fiber_length': 22730,
                                                                                                                                          'poller_hash': 'MTE5NTFCN3BlZXJwb25wb24gMjE2MTg3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535249',
                                                                                                                                          'fiber_length': 22151,
                                                                                                                                          'poller_hash': 'MTE5NTFCMnBlZXJwb25wb24gMjE3NjQ2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '508890',
                                                                                                                                          'fiber_length': 23248,
                                                                                                                                          'poller_hash': 'MTI3NDczRXBlZXJwb25wb24gMjMwNzE0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535176',
                                                                                                                                          'fiber_length': 22856,
                                                                                                                                          'poller_hash': 'MTI3NDE4RXBlZXJwb25wb24gMjQzNjgx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '68607123',
                                                                                                                                          'fiber_length': 23427,
                                                                                                                                          'poller_hash': 'MTU4RkI2NHBlZXJwb25wb24gMjQ2NDY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '8535125',
                                                                                                                                          'fiber_length': 22828,
                                                                                                                                          'poller_hash': 'MTI3NDgzRnBlZXJwb25wb24gMjYyNjg4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '4142146',
                                                                                                                                          'fiber_length': 22542,
                                                                                                                                          'poller_hash': 'MTI3NDczQnBlZXJwb25wb24gMjc0MDcx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735},
                                                                                                                                  {       'connected_time': '4235183',
                                                                                                                                          'fiber_length': 22543,
                                                                                                                                          'poller_hash': 'MTI5NUU4Q3BlZXJwb25wb24gMjgwNDI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.735}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/1',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/2',
                                                                                         'in_discards': '1517267',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '9363841',
                                                                                         'in_octets': '63050077081116',
                                                                                         'in_ucast_pkts': '99397024630',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '99924679246',
                                                                                                                  'in_ucast_pkts': '929818507',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76',
                                                                                                                  'out_octets': '2736161360369',
                                                                                                                  'out_ucast_pkts': '1974065566',
                                                                                                                  'poller_hash': 'MUQwNjgxcGVlcmludGZwb24gNTQw'},
                                                                                                          {       'in_nucast_pkts': '28',
                                                                                                                  'in_octets': '100241454627',
                                                                                                                  'in_ucast_pkts': '446269730',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12591428',
                                                                                                                  'out_octets': '2119914995184',
                                                                                                                  'out_ucast_pkts': '1481796854',
                                                                                                                  'poller_hash': 'ODY5M0U3cGVlcmludGZwb24gMTM4NjA0'},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '28559874395',
                                                                                                                  'in_ucast_pkts': '225402293',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878141',
                                                                                                                  'out_octets': '1718236441552',
                                                                                                                  'out_ucast_pkts': '1170540845',
                                                                                                                  'poller_hash': 'ODY5NDI5cGVlcmludGZwb24gMTM4NjQ1'},
                                                                                                          {       'in_nucast_pkts': '8211',
                                                                                                                  'in_octets': '28447001851',
                                                                                                                  'in_ucast_pkts': '74934545',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5341840',
                                                                                                                  'out_octets': '295670434221',
                                                                                                                  'out_ucast_pkts': '243438490',
                                                                                                                  'poller_hash': 'NUE2QkFCcGVlcmludGZwb24gMTM4NzIw'},
                                                                                                          {       'in_nucast_pkts': '2261981',
                                                                                                                  'in_octets': '175020005514',
                                                                                                                  'in_ucast_pkts': '265454169',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39388031',
                                                                                                                  'out_octets': '532085968884',
                                                                                                                  'out_ucast_pkts': '464656993',
                                                                                                                  'poller_hash': 'NUE2QjhEcGVlcmludGZwb24gMTM5Mzcy'},
                                                                                                          {       'in_nucast_pkts': '138883',
                                                                                                                  'in_octets': '68533297135',
                                                                                                                  'in_ucast_pkts': '192557822',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878221',
                                                                                                                  'out_octets': '693255198901',
                                                                                                                  'out_ucast_pkts': '527350078',
                                                                                                                  'poller_hash': 'ODJBNzMxcGVlcmludGZwb24gMTM5Mzky'},
                                                                                                          {       'in_nucast_pkts': '8231',
                                                                                                                  'in_octets': '146626809801',
                                                                                                                  'in_ucast_pkts': '830358802',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4108046',
                                                                                                                  'out_octets': '2304199635484',
                                                                                                                  'out_ucast_pkts': '1661201075',
                                                                                                                  'poller_hash': 'ODY5MkQ5cGVlcmludGZwb24gMTQwMzYx'},
                                                                                                          {       'in_nucast_pkts': '41563853',
                                                                                                                  'in_octets': '65091209116',
                                                                                                                  'in_ucast_pkts': '391690963',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26025772601',
                                                                                                                  'out_octets': '35834990181369',
                                                                                                                  'out_ucast_pkts': '1222387296',
                                                                                                                  'poller_hash': 'ODY5M0E4cGVlcmludGZwb24gMTQwNjA5'},
                                                                                                          {       'in_nucast_pkts': '419785',
                                                                                                                  'in_octets': '191567795710',
                                                                                                                  'in_ucast_pkts': '1320023592',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26250656',
                                                                                                                  'out_octets': '5503207594828',
                                                                                                                  'out_ucast_pkts': '3858877203',
                                                                                                                  'poller_hash': 'ODY5M0U5cGVlcmludGZwb24gMTQxNjkz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '372450311847',
                                                                                                                  'in_ucast_pkts': '996340988',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38',
                                                                                                                  'out_octets': '4197775308392',
                                                                                                                  'out_ucast_pkts': '3415928008',
                                                                                                                  'poller_hash': 'ODY5NDdDcGVlcmludGZwb24gMTQxNzAz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODcyNDNCcGVlcmludGZwb24gMTQzMjU0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODcyNTA0cGVlcmludGZwb24gMTQzNzgy'},
                                                                                                          {       'in_nucast_pkts': '418186',
                                                                                                                  'in_octets': '212334674186',
                                                                                                                  'in_ucast_pkts': '2160493823',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27538111',
                                                                                                                  'out_octets': '10598077154835',
                                                                                                                  'out_ucast_pkts': '7136619341',
                                                                                                                  'poller_hash': 'ODcyNEY0cGVlcmludGZwb24gMTQzODYy'},
                                                                                                          {       'in_nucast_pkts': '43398',
                                                                                                                  'in_octets': '16109324846',
                                                                                                                  'in_ucast_pkts': '55335851',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2594131',
                                                                                                                  'out_octets': '229447343250',
                                                                                                                  'out_ucast_pkts': '187563071',
                                                                                                                  'poller_hash': 'OTU2RjU4cGVlcmludGZwb24gMTQ0NDk3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45631819663',
                                                                                                                  'in_ucast_pkts': '282204782',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '1338327039744',
                                                                                                                  'out_ucast_pkts': '1007143456',
                                                                                                                  'poller_hash': 'OTU2RjgwcGVlcmludGZwb24gMTQ1MDc4'},
                                                                                                          {       'in_nucast_pkts': '12005518',
                                                                                                                  'in_octets': '190783911197',
                                                                                                                  'in_ucast_pkts': '753640468',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '16509356554',
                                                                                                                  'out_octets': '25377688791509',
                                                                                                                  'out_ucast_pkts': '2548817797',
                                                                                                                  'poller_hash': 'QTlDM0EycGVlcmludGZwb24gMTQ1ODcy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '518008115623',
                                                                                                                  'in_ucast_pkts': '84978936844',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76',
                                                                                                                  'out_octets': '12795237009973',
                                                                                                                  'out_ucast_pkts': '8785114391',
                                                                                                                  'poller_hash': 'OUMyRkUwcGVlcmludGZwb24gMTQ1OTE5'},
                                                                                                          {       'in_nucast_pkts': '18873024',
                                                                                                                  'in_octets': '79948801056110',
                                                                                                                  'in_ucast_pkts': '83238526526',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61748947',
                                                                                                                  'out_octets': '129963125544953',
                                                                                                                  'out_ucast_pkts': '111021986962',
                                                                                                                  'poller_hash': 'QTlDMkU3cGVlcmludGZwb24gMTQ1OTMw'},
                                                                                                          {       'in_nucast_pkts': '160819',
                                                                                                                  'in_octets': '197998098397',
                                                                                                                  'in_ucast_pkts': '480177746',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878131',
                                                                                                                  'out_octets': '4058375024795',
                                                                                                                  'out_ucast_pkts': '2838034073',
                                                                                                                  'poller_hash': 'QTlDMkE1cGVlcmludGZwb24gMTQ2OTA3'},
                                                                                                          {       'in_nucast_pkts': '166736',
                                                                                                                  'in_octets': '42995078300',
                                                                                                                  'in_ucast_pkts': '564694248',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878154',
                                                                                                                  'out_octets': '4207377821903',
                                                                                                                  'out_ucast_pkts': '2781392054',
                                                                                                                  'poller_hash': 'QjUxMEJCcGVlcmludGZwb24gMTU0NTk0'},
                                                                                                          {       'in_nucast_pkts': '452342',
                                                                                                                  'in_octets': '99769305882',
                                                                                                                  'in_ucast_pkts': '1192754322',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '29925432',
                                                                                                                  'out_octets': '3547385945600',
                                                                                                                  'out_ucast_pkts': '2399280080',
                                                                                                                  'poller_hash': 'QzgzQkEwcGVlcmludGZwb24gMTY2NjM4'},
                                                                                                          {       'in_nucast_pkts': '157053',
                                                                                                                  'in_octets': '140981254968',
                                                                                                                  'in_ucast_pkts': '303182354',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9877508',
                                                                                                                  'out_octets': '1250841397955',
                                                                                                                  'out_ucast_pkts': '1021028935',
                                                                                                                  'poller_hash': 'Q0JEMUREcGVlcmludGZwb24gMTY3MjM3'},
                                                                                                          {       'in_nucast_pkts': '473430',
                                                                                                                  'in_octets': '2700942599',
                                                                                                                  'in_ucast_pkts': '28256129',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '95603717',
                                                                                                                  'out_octets': '162977769578',
                                                                                                                  'out_ucast_pkts': '89951575',
                                                                                                                  'poller_hash': 'Q0JEMUQ1cGVlcmludGZwb24gMTY3OTA4'},
                                                                                                          {       'in_nucast_pkts': '151828',
                                                                                                                  'in_octets': '64158533927',
                                                                                                                  'in_ucast_pkts': '282584671',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9877979',
                                                                                                                  'out_octets': '1847512929831',
                                                                                                                  'out_ucast_pkts': '1327856785',
                                                                                                                  'poller_hash': 'MTI3NDAzOHBlZXJpbnRmcG9uIDI0NjQ0Nw=='},
                                                                                                          {       'in_nucast_pkts': '159307',
                                                                                                                  'in_octets': '40411010367',
                                                                                                                  'in_ucast_pkts': '338462182',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878022',
                                                                                                                  'out_octets': '2094792126570',
                                                                                                                  'out_ucast_pkts': '1435400185',
                                                                                                                  'poller_hash': 'MTE3REI2OXBlZXJpbnRmcG9uIDI1MTU4OQ=='},
                                                                                                          {       'in_nucast_pkts': '36302',
                                                                                                                  'in_octets': '9352817013',
                                                                                                                  'in_ucast_pkts': '35674433',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2005380',
                                                                                                                  'out_octets': '436172473560',
                                                                                                                  'out_ucast_pkts': '301464738',
                                                                                                                  'poller_hash': 'MTI5NUZERXBlZXJpbnRmcG9uIDI4NDQzMw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '166697146399',
                                                                                         'out_octets': '504090083372703',
                                                                                         'out_ucast_pkts': '213649699978',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/2',
                                                                                                                 'id': '1/2',
                                                                                                                 'links': [       {       'connected_time': '6865721',
                                                                                                                                          'fiber_length': 23173,
                                                                                                                                          'poller_hash': 'MUQwNjgxcGVlcnBvbnBvbiA1NDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '10881856',
                                                                                                                                          'fiber_length': 23031,
                                                                                                                                          'poller_hash': 'ODY5M0U3cGVlcnBvbnBvbiAxMzg2MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535218',
                                                                                                                                          'fiber_length': 22656,
                                                                                                                                          'poller_hash': 'ODY5NDI5cGVlcnBvbnBvbiAxMzg2NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '4591349',
                                                                                                                                          'fiber_length': 23566,
                                                                                                                                          'poller_hash': 'NUE2QkFCcGVlcnBvbnBvbiAxMzg3MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '67856227',
                                                                                                                                          'fiber_length': 23002,
                                                                                                                                          'poller_hash': 'NUE2QjhEcGVlcnBvbnBvbiAxMzkzNzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535457',
                                                                                                                                          'fiber_length': 22427,
                                                                                                                                          'poller_hash': 'ODJBNzMxcGVlcnBvbnBvbiAxMzkzOTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '3511700',
                                                                                                                                          'fiber_length': 22462,
                                                                                                                                          'poller_hash': 'ODY5MkQ5cGVlcnBvbnBvbiAxNDAzNjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '21887461',
                                                                                                                                          'fiber_length': 22813,
                                                                                                                                          'poller_hash': 'ODY5M0E4cGVlcnBvbnBvbiAxNDA2MDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '22473887',
                                                                                                                                          'fiber_length': 23169,
                                                                                                                                          'poller_hash': 'ODY5M0U5cGVlcnBvbnBvbiAxNDE2OTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535291',
                                                                                                                                          'fiber_length': 23735,
                                                                                                                                          'poller_hash': 'ODY5NDdDcGVlcnBvbnBvbiAxNDE3MDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535120',
                                                                                                                                          'fiber_length': 23642,
                                                                                                                                          'poller_hash': 'ODcyNDNCcGVlcnBvbnBvbiAxNDMyNTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535176',
                                                                                                                                          'fiber_length': 23634,
                                                                                                                                          'poller_hash': 'ODcyNTA0cGVlcnBvbnBvbiAxNDM3ODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '23752942',
                                                                                                                                          'fiber_length': 22584,
                                                                                                                                          'poller_hash': 'ODcyNEY0cGVlcnBvbnBvbiAxNDM4NjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '2212629',
                                                                                                                                          'fiber_length': 22138,
                                                                                                                                          'poller_hash': 'OTU2RjU4cGVlcnBvbnBvbiAxNDQ0OTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '4071451',
                                                                                                                                          'fiber_length': 23124,
                                                                                                                                          'poller_hash': 'OTU2RjgwcGVlcnBvbnBvbiAxNDUwNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '21886623',
                                                                                                                                          'fiber_length': 22901,
                                                                                                                                          'poller_hash': 'QTlDM0EycGVlcnBvbnBvbiAxNDU4NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535372',
                                                                                                                                          'fiber_length': 22442,
                                                                                                                                          'poller_hash': 'OUMyRkUwcGVlcnBvbnBvbiAxNDU5MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '18797163',
                                                                                                                                          'fiber_length': 22595,
                                                                                                                                          'poller_hash': 'QTlDMkU3cGVlcnBvbnBvbiAxNDU5MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535233',
                                                                                                                                          'fiber_length': 22376,
                                                                                                                                          'poller_hash': 'QTlDMkE1cGVlcnBvbnBvbiAxNDY5MDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535464',
                                                                                                                                          'fiber_length': 22900,
                                                                                                                                          'poller_hash': 'QjUxMEJCcGVlcnBvbnBvbiAxNTQ1OTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '26105093',
                                                                                                                                          'fiber_length': 23076,
                                                                                                                                          'poller_hash': 'QzgzQkEwcGVlcnBvbnBvbiAxNjY2Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535199',
                                                                                                                                          'fiber_length': 23827,
                                                                                                                                          'poller_hash': 'Q0JEMUREcGVlcnBvbnBvbiAxNjcyMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535136',
                                                                                                                                          'fiber_length': 23176,
                                                                                                                                          'poller_hash': 'Q0JEMUQ1cGVlcnBvbnBvbiAxNjc5MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535106',
                                                                                                                                          'fiber_length': 23636,
                                                                                                                                          'poller_hash': 'MTI3NDAzOHBlZXJwb25wb24gMjQ2NDQ3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '8535194',
                                                                                                                                          'fiber_length': 22908,
                                                                                                                                          'poller_hash': 'MTE3REI2OXBlZXJwb25wb24gMjUxNTg5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043},
                                                                                                                                  {       'connected_time': '1723414',
                                                                                                                                          'fiber_length': 23059,
                                                                                                                                          'poller_hash': 'MTI5NUZERXBlZXJwb25wb24gMjg0NDMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.043}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/2',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/3',
                                                                                         'in_discards': '2860029',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '14030660',
                                                                                         'in_octets': '34048569155511',
                                                                                         'in_ucast_pkts': '79625438131',
                                                                                         'links': [       {       'in_nucast_pkts': '160896',
                                                                                                                  'in_octets': '70798130485',
                                                                                                                  'in_ucast_pkts': '391720418',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878152',
                                                                                                                  'out_octets': '2483626511574',
                                                                                                                  'out_ucast_pkts': '1761111628',
                                                                                                                  'poller_hash': 'QTlDMzlCcGVlcmludGZwb24gNTQ3'},
                                                                                                          {       'in_nucast_pkts': '817',
                                                                                                                  'in_octets': '38353034534',
                                                                                                                  'in_ucast_pkts': '50001076',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1584472',
                                                                                                                  'out_octets': '155905052034',
                                                                                                                  'out_ucast_pkts': '127464849',
                                                                                                                  'poller_hash': 'NUE2QzkxcGVlcmludGZwb24gMTM4NzQy'},
                                                                                                          {       'in_nucast_pkts': '10436',
                                                                                                                  'in_octets': '3362289748087',
                                                                                                                  'in_ucast_pkts': '4375922892',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17807226',
                                                                                                                  'out_octets': '11214691231889',
                                                                                                                  'out_ucast_pkts': '8852804838',
                                                                                                                  'poller_hash': 'ODJBNUJCcGVlcmludGZwb24gMTM5Mzg0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODcyNTRDcGVlcmludGZwb24gMTQzNjMy'},
                                                                                                          {       'in_nucast_pkts': '161387',
                                                                                                                  'in_octets': '115010061378',
                                                                                                                  'in_ucast_pkts': '462884381',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878204',
                                                                                                                  'out_octets': '1618836687632',
                                                                                                                  'out_ucast_pkts': '1340783364',
                                                                                                                  'poller_hash': 'ODcyNDYxcGVlcmludGZwb24gMTQ0MjQ4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '25797027250',
                                                                                                                  'in_ucast_pkts': '94402764',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '96',
                                                                                                                  'out_octets': '312961961380',
                                                                                                                  'out_ucast_pkts': '275822104',
                                                                                                                  'poller_hash': 'OTU2RjcycGVlcmludGZwb24gMTQ0NzM2'},
                                                                                                          {       'in_nucast_pkts': '705091',
                                                                                                                  'in_octets': '208485139398',
                                                                                                                  'in_ucast_pkts': '1384153739',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63490020',
                                                                                                                  'out_octets': '5269134528414',
                                                                                                                  'out_ucast_pkts': '3927492952',
                                                                                                                  'poller_hash': 'OTU2RjhCcGVlcmludGZwb24gMTQ1NTI4'},
                                                                                                          {       'in_nucast_pkts': '119483',
                                                                                                                  'in_octets': '62721805619',
                                                                                                                  'in_ucast_pkts': '201994363',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7420838',
                                                                                                                  'out_octets': '730338398070',
                                                                                                                  'out_ucast_pkts': '605047082',
                                                                                                                  'poller_hash': 'QTlDOEJDcGVlcmludGZwb24gMTQ1ODMy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '395315463296',
                                                                                                                  'in_ucast_pkts': '5689261181',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37',
                                                                                                                  'out_octets': '5096451042607',
                                                                                                                  'out_ucast_pkts': '3812425713',
                                                                                                                  'poller_hash': 'QTlDM0UxcGVlcmludGZwb24gMTQ1OTQw'},
                                                                                                          {       'in_nucast_pkts': '74819',
                                                                                                                  'in_octets': '199270031750',
                                                                                                                  'in_ucast_pkts': '295457951',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4922862',
                                                                                                                  'out_octets': '633414614430',
                                                                                                                  'out_ucast_pkts': '603079358',
                                                                                                                  'poller_hash': 'OTU2RTQ4cGVlcmludGZwb24gMTQ2MDkz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '218143006952',
                                                                                                                  'in_ucast_pkts': '48074021927',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '4228165485230',
                                                                                                                  'out_ucast_pkts': '3325369341',
                                                                                                                  'poller_hash': 'OTU2RjE0cGVlcmludGZwb24gMTQ2MTY5'},
                                                                                                          {       'in_nucast_pkts': '927064',
                                                                                                                  'in_octets': '167547250740',
                                                                                                                  'in_ucast_pkts': '551982517',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63493237',
                                                                                                                  'out_octets': '1315179163791',
                                                                                                                  'out_ucast_pkts': '1157988098',
                                                                                                                  'poller_hash': 'QTlDM0M1cGVlcmludGZwb24gMTQ2OTc3'},
                                                                                                          {       'in_nucast_pkts': '67381409',
                                                                                                                  'in_octets': '318641638288',
                                                                                                                  'in_ucast_pkts': '2082121233',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '148172408217',
                                                                                                                  'out_octets': '212438091049524',
                                                                                                                  'out_ucast_pkts': '8387699341',
                                                                                                                  'poller_hash': 'QTlDMkY1cGVlcmludGZwb24gMTQ5NjMx'},
                                                                                                          {       'in_nucast_pkts': '9760006',
                                                                                                                  'in_octets': '221295763231',
                                                                                                                  'in_ucast_pkts': '587896456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878306',
                                                                                                                  'out_octets': '3250760851570',
                                                                                                                  'out_ucast_pkts': '2392140224',
                                                                                                                  'poller_hash': 'QzgwODdBcGVlcmludGZwb24gMTU3Mjk0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '685976031333',
                                                                                                                  'in_ucast_pkts': '45990137827',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '76',
                                                                                                                  'out_octets': '14068443686627',
                                                                                                                  'out_ucast_pkts': '10323879679',
                                                                                                                  'poller_hash': 'QzgxREVGcGVlcmludGZwb24gMTU3MzI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '488392480880',
                                                                                                                  'in_ucast_pkts': '1288916811',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '6882734545868',
                                                                                                                  'out_ucast_pkts': '5078137360',
                                                                                                                  'poller_hash': 'QzgxRDg4cGVlcmludGZwb24gMTU3NTMx'},
                                                                                                          {       'in_nucast_pkts': '4893891',
                                                                                                                  'in_octets': '1377028156673',
                                                                                                                  'in_ucast_pkts': '2393301385',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6728195517',
                                                                                                                  'out_octets': '12802565707024',
                                                                                                                  'out_ucast_pkts': '3202435844',
                                                                                                                  'poller_hash': 'QzgzQjY5cGVlcmludGZwb24gMTY0MTI2'},
                                                                                                          {       'in_nucast_pkts': '150004',
                                                                                                                  'in_octets': '148966826184',
                                                                                                                  'in_ucast_pkts': '447998595',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878225',
                                                                                                                  'out_octets': '1430438071545',
                                                                                                                  'out_ucast_pkts': '1116236778',
                                                                                                                  'poller_hash': 'Q0JEMjBEcGVlcmludGZwb24gMTY3Nzky'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '827565637526',
                                                                                                                  'in_ucast_pkts': '2088421011',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '131',
                                                                                                                  'out_octets': '6307374251140',
                                                                                                                  'out_ucast_pkts': '5019726148',
                                                                                                                  'poller_hash': 'QzgzQjlCcGVlcmludGZwb24gMTY4NTE0'},
                                                                                                          {       'in_nucast_pkts': '660',
                                                                                                                  'in_octets': '65273100733',
                                                                                                                  'in_ucast_pkts': '201815069',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878153',
                                                                                                                  'out_octets': '2228468029943',
                                                                                                                  'out_ucast_pkts': '1603512125',
                                                                                                                  'poller_hash': 'QzgzMUM5cGVlcmludGZwb24gMTcxNzU5'},
                                                                                                          {       'in_nucast_pkts': '149208',
                                                                                                                  'in_octets': '113271499726',
                                                                                                                  'in_ucast_pkts': '361966077',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878302',
                                                                                                                  'out_octets': '3459693388926',
                                                                                                                  'out_ucast_pkts': '2456125047',
                                                                                                                  'poller_hash': 'QzgzODIwcGVlcmludGZwb24gMTc0NTE0'},
                                                                                                          {       'in_nucast_pkts': '149446',
                                                                                                                  'in_octets': '55680950390',
                                                                                                                  'in_ucast_pkts': '106348544',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878248',
                                                                                                                  'out_octets': '491466332922',
                                                                                                                  'out_ucast_pkts': '370437533',
                                                                                                                  'poller_hash': 'REZGRTBFcGVlcmludGZwb24gMTc0OTMx'},
                                                                                                          {       'in_nucast_pkts': '158315',
                                                                                                                  'in_octets': '210489277399',
                                                                                                                  'in_ucast_pkts': '469777513',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9876305',
                                                                                                                  'out_octets': '1474793761268',
                                                                                                                  'out_ucast_pkts': '1229683318',
                                                                                                                  'poller_hash': 'QzgzQjk1cGVlcmludGZwb24gMTc0OTc5'},
                                                                                                          {       'in_nucast_pkts': '150440',
                                                                                                                  'in_octets': '408366261503',
                                                                                                                  'in_ucast_pkts': '1437720367',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878145',
                                                                                                                  'out_octets': '3900500422215',
                                                                                                                  'out_ucast_pkts': '3045343083',
                                                                                                                  'poller_hash': 'Q0JEMThGcGVlcmludGZwb24gMTc1MzUx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '99454265911',
                                                                                                                  'in_ucast_pkts': '870693132',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '3626593535230',
                                                                                                                  'out_ucast_pkts': '2604983567',
                                                                                                                  'poller_hash': 'RENGNEMxcGVlcmludGZwb24gMTg2Nzc2'},
                                                                                                          {       'in_nucast_pkts': '154306',
                                                                                                                  'in_octets': '37913603571',
                                                                                                                  'in_ucast_pkts': '207706616',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878275',
                                                                                                                  'out_octets': '1086547727912',
                                                                                                                  'out_ucast_pkts': '767868641',
                                                                                                                  'poller_hash': 'QzgzQzA2cGVlcmludGZwb24gMTg5ODk5'},
                                                                                                          {       'in_nucast_pkts': '2675572',
                                                                                                                  'in_octets': '122570986623',
                                                                                                                  'in_ucast_pkts': '445860606',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3008779',
                                                                                                                  'out_octets': '1355670044771',
                                                                                                                  'out_ucast_pkts': '1112158645',
                                                                                                                  'poller_hash': 'MTE3REJGM3BlZXJpbnRmcG9uIDI1MjY1MA=='},
                                                                                                          {       'in_nucast_pkts': '9613557',
                                                                                                                  'in_octets': '260172244636',
                                                                                                                  'in_ucast_pkts': '600574925',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878367',
                                                                                                                  'out_octets': '1589658642773',
                                                                                                                  'out_ucast_pkts': '1276775125',
                                                                                                                  'poller_hash': 'MTE3REMzMnBlZXJpbnRmcG9uIDI1MzQyNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '154711708667',
                                                                                         'out_octets': '490243278580684',
                                                                                         'out_ucast_pkts': '217824238665',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/3',
                                                                                                                 'id': '1/3',
                                                                                                                 'links': [       {       'connected_time': '8535197',
                                                                                                                                          'fiber_length': 24959,
                                                                                                                                          'poller_hash': 'QTlDMzlCcGVlcnBvbnBvbiA1NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '1357590',
                                                                                                                                          'fiber_length': 24801,
                                                                                                                                          'poller_hash': 'NUE2QzkxcGVlcnBvbnBvbiAxMzg3NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '15181507',
                                                                                                                                          'fiber_length': 24672,
                                                                                                                                          'poller_hash': 'ODJBNUJCcGVlcnBvbnBvbiAxMzkzODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535161',
                                                                                                                                          'fiber_length': 25137,
                                                                                                                                          'poller_hash': 'ODcyNTRDcGVlcnBvbnBvbiAxNDM2MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535169',
                                                                                                                                          'fiber_length': 25094,
                                                                                                                                          'poller_hash': 'ODcyNDYxcGVlcnBvbnBvbiAxNDQyNDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '21887300',
                                                                                                                                          'fiber_length': 24835,
                                                                                                                                          'poller_hash': 'OTU2RjcycGVlcnBvbnBvbiAxNDQ3MzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '21887001',
                                                                                                                                          'fiber_length': 24982,
                                                                                                                                          'poller_hash': 'OTU2RjhCcGVlcnBvbnBvbiAxNDU1Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '6396009',
                                                                                                                                          'fiber_length': 24932,
                                                                                                                                          'poller_hash': 'QTlDOEJDcGVlcnBvbnBvbiAxNDU4MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '10946538',
                                                                                                                                          'fiber_length': 25326,
                                                                                                                                          'poller_hash': 'QTlDM0UxcGVlcnBvbnBvbiAxNDU5NDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '4221509',
                                                                                                                                          'fiber_length': 24395,
                                                                                                                                          'poller_hash': 'OTU2RTQ4cGVlcnBvbnBvbiAxNDYwOTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535281',
                                                                                                                                          'fiber_length': 24788,
                                                                                                                                          'poller_hash': 'OTU2RjE0cGVlcnBvbnBvbiAxNDYxNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '21887572',
                                                                                                                                          'fiber_length': 25052,
                                                                                                                                          'poller_hash': 'QTlDM0M1cGVlcnBvbnBvbiAxNDY5Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '21887130',
                                                                                                                                          'fiber_length': 25051,
                                                                                                                                          'poller_hash': 'QTlDMkY1cGVlcnBvbnBvbiAxNDk2MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535197',
                                                                                                                                          'fiber_length': 25019,
                                                                                                                                          'poller_hash': 'QzgwODdBcGVlcnBvbnBvbiAxNTcyOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '7195080',
                                                                                                                                          'fiber_length': 25074,
                                                                                                                                          'poller_hash': 'QzgxREVGcGVlcnBvbnBvbiAxNTczMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535079',
                                                                                                                                          'fiber_length': 24368,
                                                                                                                                          'poller_hash': 'QzgxRDg4cGVlcnBvbnBvbiAxNTc1MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535231',
                                                                                                                                          'fiber_length': 24538,
                                                                                                                                          'poller_hash': 'QzgzQjY5cGVlcnBvbnBvbiAxNjQxMjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535206',
                                                                                                                                          'fiber_length': 25379,
                                                                                                                                          'poller_hash': 'Q0JEMjBEcGVlcnBvbnBvbiAxNjc3OTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '22786823',
                                                                                                                                          'fiber_length': 24813,
                                                                                                                                          'poller_hash': 'QzgzQjlCcGVlcnBvbnBvbiAxNjg1MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535158',
                                                                                                                                          'fiber_length': 24770,
                                                                                                                                          'poller_hash': 'QzgzMUM5cGVlcnBvbnBvbiAxNzE3NTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -28.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535156',
                                                                                                                                          'fiber_length': 24780,
                                                                                                                                          'poller_hash': 'QzgzODIwcGVlcnBvbnBvbiAxNzQ1MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535185',
                                                                                                                                          'fiber_length': 25053,
                                                                                                                                          'poller_hash': 'REZGRTBFcGVlcnBvbnBvbiAxNzQ5MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535223',
                                                                                                                                          'fiber_length': 24642,
                                                                                                                                          'poller_hash': 'QzgzQjk1cGVlcnBvbnBvbiAxNzQ5Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535108',
                                                                                                                                          'fiber_length': 24927,
                                                                                                                                          'poller_hash': 'Q0JEMThGcGVlcnBvbnBvbiAxNzUzNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535156',
                                                                                                                                          'fiber_length': 24426,
                                                                                                                                          'poller_hash': 'RENGNEMxcGVlcnBvbnBvbiAxODY3NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535142',
                                                                                                                                          'fiber_length': 25188,
                                                                                                                                          'poller_hash': 'QzgzQzA2cGVlcnBvbnBvbiAxODk4OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '2565938',
                                                                                                                                          'fiber_length': 24882,
                                                                                                                                          'poller_hash': 'MTE3REJGM3BlZXJwb25wb24gMjUyNjUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862},
                                                                                                                                  {       'connected_time': '8535221',
                                                                                                                                          'fiber_length': 25349,
                                                                                                                                          'poller_hash': 'MTE3REMzMnBlZXJwb25wb24gMjUzNDI2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.862}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/3',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/4',
                                                                                         'in_discards': '823124',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5037202',
                                                                                         'in_octets': '24298252778271',
                                                                                         'in_ucast_pkts': '56574757104',
                                                                                         'links': [       {       'in_nucast_pkts': '73346',
                                                                                                                  'in_octets': '15494382907',
                                                                                                                  'in_ucast_pkts': '135267901',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1285690',
                                                                                                                  'out_octets': '1290969282633',
                                                                                                                  'out_ucast_pkts': '868361148',
                                                                                                                  'poller_hash': 'NUE2QjZEcGVlcmludGZwb24gMTM4NzA4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NUE2Qjg4cGVlcmludGZwb24gMTM4ODE2'},
                                                                                                          {       'in_nucast_pkts': '25168',
                                                                                                                  'in_octets': '200479178105',
                                                                                                                  'in_ucast_pkts': '197198842',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1280566',
                                                                                                                  'out_octets': '294723015846',
                                                                                                                  'out_ucast_pkts': '280434420',
                                                                                                                  'poller_hash': 'ODJBNzREcGVlcmludGZwb24gMTM5Mzg4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '829876977823',
                                                                                                                  'in_ucast_pkts': '1703935141',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '133',
                                                                                                                  'out_octets': '5731676860218',
                                                                                                                  'out_ucast_pkts': '4593249660',
                                                                                                                  'poller_hash': 'ODY5MzlEcGVlcmludGZwb24gMTQwMTI5'},
                                                                                                          {       'in_nucast_pkts': '1215',
                                                                                                                  'in_octets': '4457363813',
                                                                                                                  'in_ucast_pkts': '25165602',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281565',
                                                                                                                  'out_octets': '143200174663',
                                                                                                                  'out_ucast_pkts': '108610249',
                                                                                                                  'poller_hash': 'ODQ0QkQwcGVlcmludGZwb24gMTQwMjEw'},
                                                                                                          {       'in_nucast_pkts': '23810',
                                                                                                                  'in_octets': '14530958379',
                                                                                                                  'in_ucast_pkts': '85788246',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281592',
                                                                                                                  'out_octets': '254814071063',
                                                                                                                  'out_ucast_pkts': '205930271',
                                                                                                                  'poller_hash': 'OTU2RjZFcGVlcmludGZwb24gMTQwMjE4'},
                                                                                                          {       'in_nucast_pkts': '956906',
                                                                                                                  'in_octets': '314223820053',
                                                                                                                  'in_ucast_pkts': '2728663129',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63494906',
                                                                                                                  'out_octets': '13584104736499',
                                                                                                                  'out_ucast_pkts': '9317655591',
                                                                                                                  'poller_hash': 'ODY5MzkycGVlcmludGZwb24gMTQwMjIy'},
                                                                                                          {       'in_nucast_pkts': '3279',
                                                                                                                  'in_octets': '22265003852',
                                                                                                                  'in_ucast_pkts': '110401755',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281513',
                                                                                                                  'out_octets': '513080923189',
                                                                                                                  'out_ucast_pkts': '360126770',
                                                                                                                  'poller_hash': 'ODQ0QzEwcGVlcmludGZwb24gMTQwMjY4'},
                                                                                                          {       'in_nucast_pkts': '3132',
                                                                                                                  'in_octets': '699317529801',
                                                                                                                  'in_ucast_pkts': '1378589442',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12380103',
                                                                                                                  'out_octets': '4263425818267',
                                                                                                                  'out_ucast_pkts': '3544884784',
                                                                                                                  'poller_hash': 'MTI3MzlGQnBlZXJpbnRmcG9uIDE0MDI3Mg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODQ0QkEzcGVlcmludGZwb24gMTQwMjg2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '350786853501',
                                                                                                                  'in_ucast_pkts': '1839441079',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '94',
                                                                                                                  'out_octets': '10531367458131',
                                                                                                                  'out_ucast_pkts': '7457703928',
                                                                                                                  'poller_hash': 'ODY5NDc4cGVlcmludGZwb24gMTQwMjkw'},
                                                                                                          {       'in_nucast_pkts': '19565',
                                                                                                                  'in_octets': '15669052751',
                                                                                                                  'in_ucast_pkts': '113045811',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281574',
                                                                                                                  'out_octets': '445470103349',
                                                                                                                  'out_ucast_pkts': '338398126',
                                                                                                                  'poller_hash': 'ODY5M0M5cGVlcmludGZwb24gMTQwMjk0'},
                                                                                                          {       'in_nucast_pkts': '28682',
                                                                                                                  'in_octets': '515225227',
                                                                                                                  'in_ucast_pkts': '4348812',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1610067',
                                                                                                                  'out_octets': '12008508010',
                                                                                                                  'out_ucast_pkts': '8858251',
                                                                                                                  'poller_hash': 'ODY5MzYxcGVlcmludGZwb24gMTQwMzQ3'},
                                                                                                          {       'in_nucast_pkts': '171',
                                                                                                                  'in_octets': '5412691680',
                                                                                                                  'in_ucast_pkts': '38344866',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281556',
                                                                                                                  'out_octets': '214581424236',
                                                                                                                  'out_ucast_pkts': '145998818',
                                                                                                                  'poller_hash': 'ODY5MzcycGVlcmludGZwb24gMTQwMzc5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '4213274133844',
                                                                                                                  'in_ucast_pkts': '1949276403630',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '84',
                                                                                                                  'out_octets': '3709163404360',
                                                                                                                  'out_ucast_pkts': '4161446371',
                                                                                                                  'poller_hash': 'QzgxRERGcGVlcmludGZwb24gMjA4OTMz'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '188894969989',
                                                                                         'out_octets': '455754024345234',
                                                                                         'out_ucast_pkts': '148492901198',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/4',
                                                                                                                 'id': '1/4',
                                                                                                                 'links': [       {       'connected_time': '1095976',
                                                                                                                                          'fiber_length': 31256,
                                                                                                                                          'poller_hash': 'NUE2QjZEcGVlcnBvbnBvbiAxMzg3MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1095976',
                                                                                                                                          'fiber_length': 31013,
                                                                                                                                          'poller_hash': 'NUE2Qjg4cGVlcnBvbnBvbiAxMzg4MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1095995',
                                                                                                                                          'fiber_length': 31110,
                                                                                                                                          'poller_hash': 'ODJBNzREcGVlcnBvbnBvbiAxMzkzODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1095959',
                                                                                                                                          'fiber_length': 31174,
                                                                                                                                          'poller_hash': 'ODY5MzlEcGVlcnBvbnBvbiAxNDAxMjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1096814',
                                                                                                                                          'fiber_length': 31430,
                                                                                                                                          'poller_hash': 'ODQ0QkQwcGVlcnBvbnBvbiAxNDAyMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1096792',
                                                                                                                                          'fiber_length': 31785,
                                                                                                                                          'poller_hash': 'OTU2RjZFcGVlcnBvbnBvbiAxNDAyMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '21886560',
                                                                                                                                          'fiber_length': 31211,
                                                                                                                                          'poller_hash': 'ODY5MzkycGVlcnBvbnBvbiAxNDAyMjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1096808',
                                                                                                                                          'fiber_length': 31268,
                                                                                                                                          'poller_hash': 'ODQ0QzEwcGVlcnBvbnBvbiAxNDAyNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '10709012',
                                                                                                                                          'fiber_length': 31146,
                                                                                                                                          'poller_hash': 'MTI3MzlGQnBlZXJwb25wb24gMTQwMjcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1096812',
                                                                                                                                          'fiber_length': 31974,
                                                                                                                                          'poller_hash': 'ODQ0QkEzcGVlcnBvbnBvbiAxNDAyODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '13190710',
                                                                                                                                          'fiber_length': 31223,
                                                                                                                                          'poller_hash': 'ODY5NDc4cGVlcnBvbnBvbiAxNDAyOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1096772',
                                                                                                                                          'fiber_length': 31244,
                                                                                                                                          'poller_hash': 'ODY5M0M5cGVlcnBvbnBvbiAxNDAyOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1379193',
                                                                                                                                          'fiber_length': 31465,
                                                                                                                                          'poller_hash': 'ODY5MzYxcGVlcnBvbnBvbiAxNDAzNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '1096772',
                                                                                                                                          'fiber_length': 31317,
                                                                                                                                          'poller_hash': 'ODY5MzcycGVlcnBvbnBvbiAxNDAzNzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134},
                                                                                                                                  {       'connected_time': '13271249',
                                                                                                                                          'fiber_length': 31458,
                                                                                                                                          'poller_hash': 'QzgxRERGcGVlcnBvbnBvbiAyMDg5MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.134}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/4',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/5',
                                                                                         'in_discards': '7954',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '12076728',
                                                                                         'in_octets': '14938362396946',
                                                                                         'in_ucast_pkts': '25646314777',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '622577640568',
                                                                                                                  'in_ucast_pkts': '1147782129',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '32137674620',
                                                                                                                  'out_octets': '47908664020544',
                                                                                                                  'out_ucast_pkts': '3386261288',
                                                                                                                  'poller_hash': 'ODY5MzlFcGVlcmludGZwb24gNTQy'},
                                                                                                          {       'in_nucast_pkts': '61961',
                                                                                                                  'in_octets': '148516955878',
                                                                                                                  'in_ucast_pkts': '230039591',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4003300',
                                                                                                                  'out_octets': '694327705555',
                                                                                                                  'out_ucast_pkts': '560619580',
                                                                                                                  'poller_hash': 'ODY5NDJDcGVlcmludGZwb24gNTQz'},
                                                                                                          {       'in_nucast_pkts': '7341919',
                                                                                                                  'in_octets': '7310193140953',
                                                                                                                  'in_ucast_pkts': '10224861181',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '65020200935',
                                                                                                                  'out_octets': '125985304580777',
                                                                                                                  'out_ucast_pkts': '29238826877',
                                                                                                                  'poller_hash': 'ODY5Mzk2cGVlcmludGZwb24gMTQyOTcz'},
                                                                                                          {       'in_nucast_pkts': '2557132',
                                                                                                                  'in_octets': '7053937086',
                                                                                                                  'in_ucast_pkts': '22207290',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2257874',
                                                                                                                  'out_octets': '35528952849',
                                                                                                                  'out_ucast_pkts': '38298722',
                                                                                                                  'poller_hash': 'NzZCODREcGVlcmludGZwb24gMTQ1NTU4'},
                                                                                                          {       'in_nucast_pkts': '2832',
                                                                                                                  'in_octets': '242685670473',
                                                                                                                  'in_ucast_pkts': '1061064055',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9977878',
                                                                                                                  'out_octets': '2453181458139',
                                                                                                                  'out_ucast_pkts': '2119964902',
                                                                                                                  'poller_hash': 'QjUxMURFcGVlcmludGZwb24gMTUyMjk5'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '141053158044',
                                                                                         'out_octets': '255081984294979',
                                                                                         'out_ucast_pkts': '52866746972',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/5',
                                                                                                                 'id': '1/5',
                                                                                                                 'links': [       {       'connected_time': '6471760',
                                                                                                                                          'fiber_length': 37658,
                                                                                                                                          'poller_hash': 'ODY5MzlFcGVlcnBvbnBvbiA1NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.176},
                                                                                                                                  {       'connected_time': '3422392',
                                                                                                                                          'fiber_length': 37809,
                                                                                                                                          'poller_hash': 'ODY5NDJDcGVlcnBvbnBvbiA1NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.176},
                                                                                                                                  {       'connected_time': '21887083',
                                                                                                                                          'fiber_length': 37683,
                                                                                                                                          'poller_hash': 'ODY5Mzk2cGVlcnBvbnBvbiAxNDI5NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.176},
                                                                                                                                  {       'connected_time': '3370045',
                                                                                                                                          'fiber_length': 36173,
                                                                                                                                          'poller_hash': 'NzZCODREcGVlcnBvbnBvbiAxNDU1NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.176},
                                                                                                                                  {       'connected_time': '8621422',
                                                                                                                                          'fiber_length': 35989,
                                                                                                                                          'poller_hash': 'QjUxMURFcGVlcnBvbnBvbiAxNTIyOTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.176}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/5',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/6',
                                                                                         'in_discards': '62462',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2232820',
                                                                                         'in_octets': '19414560348730',
                                                                                         'in_ucast_pkts': '35942492445',
                                                                                         'links': [       {       'in_nucast_pkts': '19303',
                                                                                                                  'in_octets': '17359525189',
                                                                                                                  'in_ucast_pkts': '40683432',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7100598765',
                                                                                                                  'out_octets': '9590194327098',
                                                                                                                  'out_ucast_pkts': '126600126',
                                                                                                                  'poller_hash': 'ODY5M0IwcGVlcmludGZwb24gMTQwMzA0'},
                                                                                                          {       'in_nucast_pkts': '13089',
                                                                                                                  'in_octets': '186116109048',
                                                                                                                  'in_ucast_pkts': '301820818',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281577',
                                                                                                                  'out_octets': '485487265313',
                                                                                                                  'out_ucast_pkts': '439549856',
                                                                                                                  'poller_hash': 'ODQ0QzA4cGVlcmludGZwb24gMTQwMzcx'},
                                                                                                          {       'in_nucast_pkts': '13',
                                                                                                                  'in_octets': '170219989',
                                                                                                                  'in_ucast_pkts': '174852',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '69388',
                                                                                                                  'out_octets': '61432048',
                                                                                                                  'out_ucast_pkts': '98161',
                                                                                                                  'poller_hash': 'ODQ0QkFEcGVlcmludGZwb24gMTQwNDMx'},
                                                                                                          {       'in_nucast_pkts': '23926',
                                                                                                                  'in_octets': '30174357233',
                                                                                                                  'in_ucast_pkts': '117950472',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4517086764',
                                                                                                                  'out_octets': '6407192615146',
                                                                                                                  'out_ucast_pkts': '341713722',
                                                                                                                  'poller_hash': 'ODY5MzY4cGVlcmludGZwb24gMTQwNDk5'},
                                                                                                          {       'in_nucast_pkts': '686006',
                                                                                                                  'in_octets': '256361278074',
                                                                                                                  'in_ucast_pkts': '486178765',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39420612',
                                                                                                                  'out_octets': '1656573708180',
                                                                                                                  'out_ucast_pkts': '1363946305',
                                                                                                                  'poller_hash': 'ODQ0QkM2cGVlcmludGZwb24gMTQwNTMx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '153342946876',
                                                                                                                  'in_ucast_pkts': '439091366',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '2073304076161',
                                                                                                                  'out_ucast_pkts': '1588103991',
                                                                                                                  'poller_hash': 'Q0JEMUFFcGVlcmludGZwb24gMTY1NDE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '580699373417',
                                                                                                                  'in_ucast_pkts': '40883946000',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '56',
                                                                                                                  'out_octets': '6513907509347',
                                                                                                                  'out_ucast_pkts': '5026822959',
                                                                                                                  'poller_hash': 'Q0JFMDJGcGVlcmludGZwb24gMTY4NDk0'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '47296020027',
                                                                                         'out_octets': '131616463022650',
                                                                                         'out_ucast_pkts': '60839727152',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/6',
                                                                                                                 'id': '1/6',
                                                                                                                 'links': [       {       'connected_time': '1096784',
                                                                                                                                          'fiber_length': 30629,
                                                                                                                                          'poller_hash': 'ODY5M0IwcGVlcnBvbnBvbiAxNDAzMDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.19},
                                                                                                                                  {       'connected_time': '1096808',
                                                                                                                                          'fiber_length': 30031,
                                                                                                                                          'poller_hash': 'ODQ0QzA4cGVlcnBvbnBvbiAxNDAzNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.19},
                                                                                                                                  {       'connected_time': '58684',
                                                                                                                                          'fiber_length': 29744,
                                                                                                                                          'poller_hash': 'ODQ0QkFEcGVlcnBvbnBvbiAxNDA0MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.19},
                                                                                                                                  {       'connected_time': '1096785',
                                                                                                                                          'fiber_length': 30109,
                                                                                                                                          'poller_hash': 'ODY5MzY4cGVlcnBvbnBvbiAxNDA0OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.19},
                                                                                                                                  {       'connected_time': '67856572',
                                                                                                                                          'fiber_length': 30611,
                                                                                                                                          'poller_hash': 'ODQ0QkM2cGVlcnBvbnBvbiAxNDA1MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.19},
                                                                                                                                  {       'connected_time': '1096783',
                                                                                                                                          'fiber_length': 29575,
                                                                                                                                          'poller_hash': 'Q0JEMUFFcGVlcnBvbnBvbiAxNjU0MTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.19},
                                                                                                                                  {       'connected_time': '1096769',
                                                                                                                                          'fiber_length': 28898,
                                                                                                                                          'poller_hash': 'Q0JFMDJGcGVlcnBvbnBvbiAxNjg0OTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.19}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/6',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/7',
                                                                                         'in_discards': '130586',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '4713747',
                                                                                         'in_octets': '15457128495587',
                                                                                         'in_ucast_pkts': '47604873097',
                                                                                         'links': [       {       'in_nucast_pkts': '54064',
                                                                                                                  'in_octets': '22510722159',
                                                                                                                  'in_ucast_pkts': '303682048',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3535919',
                                                                                                                  'out_octets': '1961984988155',
                                                                                                                  'out_ucast_pkts': '1298539853',
                                                                                                                  'poller_hash': 'ODY5M0RDcGVlcmludGZwb24gMTQwNTM1'},
                                                                                                          {       'in_nucast_pkts': '19474',
                                                                                                                  'in_octets': '17833422367',
                                                                                                                  'in_ucast_pkts': '94597598',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281609',
                                                                                                                  'out_octets': '241151493572',
                                                                                                                  'out_ucast_pkts': '200559929',
                                                                                                                  'poller_hash': 'ODY5M0VEcGVlcmludGZwb24gMTQwNTU5'},
                                                                                                          {       'in_nucast_pkts': '16317744',
                                                                                                                  'in_octets': '415440047427',
                                                                                                                  'in_ucast_pkts': '2012799458',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39415087',
                                                                                                                  'out_octets': '6492118150102',
                                                                                                                  'out_ucast_pkts': '5017537547',
                                                                                                                  'poller_hash': 'ODQ0QzBDcGVlcmludGZwb24gMTQwNjMz'},
                                                                                                          {       'in_nucast_pkts': '46',
                                                                                                                  'in_octets': '11821073280',
                                                                                                                  'in_ucast_pkts': '34255260',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281586',
                                                                                                                  'out_octets': '167063190417',
                                                                                                                  'out_ucast_pkts': '128668049',
                                                                                                                  'poller_hash': 'ODQ0QkY5cGVlcmludGZwb24gMTQwNjM3'},
                                                                                                          {       'in_nucast_pkts': '20154',
                                                                                                                  'in_octets': '12360324939',
                                                                                                                  'in_ucast_pkts': '28253058',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1281607',
                                                                                                                  'out_octets': '128642056250',
                                                                                                                  'out_ucast_pkts': '99305209',
                                                                                                                  'poller_hash': 'ODcyNDJFcGVlcmludGZwb24gMTQzOTEy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1816852732681',
                                                                                                                  'in_ucast_pkts': '37927309640',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '57',
                                                                                                                  'out_octets': '6994749673262',
                                                                                                                  'out_ucast_pkts': '7275309964',
                                                                                                                  'poller_hash': 'QTlDNDEzcGVlcmludGZwb24gMTQ2MDIx'},
                                                                                                          {       'in_nucast_pkts': '20374',
                                                                                                                  'in_octets': '4990193781',
                                                                                                                  'in_ucast_pkts': '15970766',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1280615',
                                                                                                                  'out_octets': '62273168872',
                                                                                                                  'out_ucast_pkts': '52091688',
                                                                                                                  'poller_hash': 'QjBBMjlBcGVlcmludGZwb24gMTUwNzg4'},
                                                                                                          {       'in_nucast_pkts': '24982',
                                                                                                                  'in_octets': '5027139070',
                                                                                                                  'in_ucast_pkts': '14402773',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1280614',
                                                                                                                  'out_octets': '48616884258',
                                                                                                                  'out_ucast_pkts': '39525506',
                                                                                                                  'poller_hash': 'QjUxMTkycGVlcmludGZwb24gMTU0NTU1'},
                                                                                                          {       'in_nucast_pkts': '18634264',
                                                                                                                  'in_octets': '473558169032',
                                                                                                                  'in_ucast_pkts': '3940663650',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '49972067448',
                                                                                                                  'out_octets': '81240553851644',
                                                                                                                  'out_ucast_pkts': '9605975034',
                                                                                                                  'poller_hash': 'QzgzOTQwcGVlcmludGZwb24gMTU2NDY2'},
                                                                                                          {       'in_nucast_pkts': '530849',
                                                                                                                  'in_octets': '5504365279030',
                                                                                                                  'in_ucast_pkts': '17045921995',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18782667',
                                                                                                                  'out_octets': '53575071730157',
                                                                                                                  'out_ucast_pkts': '42205412109',
                                                                                                                  'poller_hash': 'RkQ4NzNBcGVlcmludGZwb24gMTY0NzE1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '320387270029',
                                                                                                                  'in_ucast_pkts': '1776373869',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58',
                                                                                                                  'out_octets': '8891958179039',
                                                                                                                  'out_ucast_pkts': '6530057890',
                                                                                                                  'poller_hash': 'REZGRTIzcGVlcmludGZwb24gMTgzMTgy'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '89244918686',
                                                                                         'out_octets': '277406095283526',
                                                                                         'out_ucast_pkts': '123800009232',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/7',
                                                                                                                 'id': '1/7',
                                                                                                                 'links': [       {       'connected_time': '3015841',
                                                                                                                                          'fiber_length': 31751,
                                                                                                                                          'poller_hash': 'ODY5M0RDcGVlcnBvbnBvbiAxNDA1MzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '1096784',
                                                                                                                                          'fiber_length': 31560,
                                                                                                                                          'poller_hash': 'ODY5M0VEcGVlcnBvbnBvbiAxNDA1NTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '67856812',
                                                                                                                                          'fiber_length': 31540,
                                                                                                                                          'poller_hash': 'ODQ0QzBDcGVlcnBvbnBvbiAxNDA2MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '1096810',
                                                                                                                                          'fiber_length': 31364,
                                                                                                                                          'poller_hash': 'ODQ0QkY5cGVlcnBvbnBvbiAxNDA2Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '1096780',
                                                                                                                                          'fiber_length': 31474,
                                                                                                                                          'poller_hash': 'ODcyNDJFcGVlcnBvbnBvbiAxNDM5MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '1096784',
                                                                                                                                          'fiber_length': 32033,
                                                                                                                                          'poller_hash': 'QTlDNDEzcGVlcnBvbnBvbiAxNDYwMjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '1095970',
                                                                                                                                          'fiber_length': 31114,
                                                                                                                                          'poller_hash': 'QjBBMjlBcGVlcnBvbnBvbiAxNTA3ODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '1095990',
                                                                                                                                          'fiber_length': 31175,
                                                                                                                                          'poller_hash': 'QjUxMTkycGVlcnBvbnBvbiAxNTQ1NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '21886200',
                                                                                                                                          'fiber_length': 31284,
                                                                                                                                          'poller_hash': 'QzgzOTQwcGVlcnBvbnBvbiAxNTY0NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '30163838',
                                                                                                                                          'fiber_length': 32136,
                                                                                                                                          'poller_hash': 'RkQ4NzNBcGVlcnBvbnBvbiAxNjQ3MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896},
                                                                                                                                  {       'connected_time': '1096769',
                                                                                                                                          'fiber_length': 31925,
                                                                                                                                          'poller_hash': 'REZGRTIzcGVlcnBvbnBvbiAxODMxODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.896}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/7',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/8',
                                                                                         'in_discards': '14412',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5953884',
                                                                                         'in_octets': '16982892541646',
                                                                                         'in_ucast_pkts': '59850418780',
                                                                                         'links': [       {       'in_nucast_pkts': '32468',
                                                                                                                  'in_octets': '899893667459',
                                                                                                                  'in_ucast_pkts': '2867731036',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17533309',
                                                                                                                  'out_octets': '10298685359968',
                                                                                                                  'out_ucast_pkts': '7536882212',
                                                                                                                  'poller_hash': 'MTU4RkFBRnBlZXJpbnRmcG9uIDgw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '804542686016',
                                                                                                                  'in_ucast_pkts': '182641855850',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '6309080717687',
                                                                                                                  'out_ucast_pkts': '4748235885',
                                                                                                                  'poller_hash': 'QTlDMzlEcGVlcmludGZwb24gMTQxMDg2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '113705077156',
                                                                                                                  'in_ucast_pkts': '427944072',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38',
                                                                                                                  'out_octets': '3084307047575',
                                                                                                                  'out_ucast_pkts': '2275739139',
                                                                                                                  'poller_hash': 'OTU2RTQzcGVlcmludGZwb24gMTQxNjg5'},
                                                                                                          {       'in_nucast_pkts': '154036',
                                                                                                                  'in_octets': '206343381695',
                                                                                                                  'in_ucast_pkts': '681961428',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878387',
                                                                                                                  'out_octets': '4099782532724',
                                                                                                                  'out_ucast_pkts': '3000146836',
                                                                                                                  'poller_hash': 'ODcyNDI4cGVlcmludGZwb24gMTQzODcy'},
                                                                                                          {       'in_nucast_pkts': '161503',
                                                                                                                  'in_octets': '137323308469',
                                                                                                                  'in_ucast_pkts': '553717456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878489',
                                                                                                                  'out_octets': '2289950757169',
                                                                                                                  'out_ucast_pkts': '1848885558',
                                                                                                                  'poller_hash': 'OTU2Rjk3cGVlcmludGZwb24gMTQ0Nzk5'},
                                                                                                          {       'in_nucast_pkts': '909030',
                                                                                                                  'in_octets': '545790578420',
                                                                                                                  'in_ucast_pkts': '3089161428',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '57329872',
                                                                                                                  'out_octets': '14713296077088',
                                                                                                                  'out_ucast_pkts': '10686255915',
                                                                                                                  'poller_hash': 'OTU2RjhDcGVlcmludGZwb24gMTQ1MTI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '983604945228',
                                                                                                                  'in_ucast_pkts': '2764223582',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '11290719260810',
                                                                                                                  'out_ucast_pkts': '8689583679',
                                                                                                                  'poller_hash': 'OTU2RTMycGVlcmludGZwb24gMTQ1MTM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '201563588143',
                                                                                                                  'in_ucast_pkts': '294874766',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '86',
                                                                                                                  'out_octets': '325276687887',
                                                                                                                  'out_ucast_pkts': '359343412',
                                                                                                                  'poller_hash': 'OTU2RjAycGVlcmludGZwb24gMTQ1Mjc5'},
                                                                                                          {       'in_nucast_pkts': '159044',
                                                                                                                  'in_octets': '116121868024',
                                                                                                                  'in_ucast_pkts': '281284096',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9874974',
                                                                                                                  'out_octets': '1142879781283',
                                                                                                                  'out_ucast_pkts': '886597702',
                                                                                                                  'poller_hash': 'OTU2RkE0cGVlcmludGZwb24gMTQ1NTQ4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '118538017550',
                                                                                                                  'in_ucast_pkts': '431854538',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '67',
                                                                                                                  'out_octets': '1097047121425',
                                                                                                                  'out_ucast_pkts': '958463336',
                                                                                                                  'poller_hash': 'OTU2RjQ2cGVlcmludGZwb24gMTQ1ODEy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '408936004848',
                                                                                                                  'in_ucast_pkts': '1710641393',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '77',
                                                                                                                  'out_octets': '7039153217422',
                                                                                                                  'out_ucast_pkts': '5185671829',
                                                                                                                  'poller_hash': 'QTlDMkI0cGVlcmludGZwb24gMTQ2MjE2'},
                                                                                                          {       'in_nucast_pkts': '321285',
                                                                                                                  'in_octets': '896739626171',
                                                                                                                  'in_ucast_pkts': '1543577637',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '21121628',
                                                                                                                  'out_octets': '4538568365665',
                                                                                                                  'out_ucast_pkts': '3667120822',
                                                                                                                  'poller_hash': 'QTlDMzdDcGVlcmludGZwb24gMTQ2NjM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '473900533034',
                                                                                                                  'in_ucast_pkts': '1125373917',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '133',
                                                                                                                  'out_octets': '6266694550369',
                                                                                                                  'out_ucast_pkts': '4505661089',
                                                                                                                  'poller_hash': 'QzgxRUNBcGVlcmludGZwb24gMTU2NzQ2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3011182141924',
                                                                                                                  'in_ucast_pkts': '698784968',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37',
                                                                                                                  'out_octets': '22469748534087',
                                                                                                                  'out_ucast_pkts': '18118128684',
                                                                                                                  'poller_hash': 'QzgzOTA0cGVlcmludGZwb24gMTU3MzA0'},
                                                                                                          {       'in_nucast_pkts': '963199',
                                                                                                                  'in_octets': '407037654462',
                                                                                                                  'in_ucast_pkts': '2502038988',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63477020',
                                                                                                                  'out_octets': '12620894935242',
                                                                                                                  'out_ucast_pkts': '9247849619',
                                                                                                                  'poller_hash': 'QzgxRDcwcGVlcmludGZwb24gMTU3NDI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '256543344533',
                                                                                                                  'in_ucast_pkts': '1391881398',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '133',
                                                                                                                  'out_octets': '7093871491165',
                                                                                                                  'out_ucast_pkts': '5037474535',
                                                                                                                  'poller_hash': 'QzgxNTYzcGVlcmludGZwb24gMjA3NzMy'},
                                                                                                          {       'in_nucast_pkts': '26684107',
                                                                                                                  'in_octets': '244353092664',
                                                                                                                  'in_ucast_pkts': '2172832992',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '46273274895',
                                                                                                                  'out_octets': '68397272978120',
                                                                                                                  'out_ucast_pkts': '4210984802',
                                                                                                                  'poller_hash': 'MTI3NTBDNHBlZXJpbnRmcG9uIDIxODIzNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '215158064750',
                                                                                                                  'in_ucast_pkts': '2269997975',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18',
                                                                                                                  'out_octets': '6774036034251',
                                                                                                                  'out_ucast_pkts': '4602187384',
                                                                                                                  'poller_hash': 'MTI3NDMyNXBlZXJpbnRmcG9uIDI0Njk0MA=='},
                                                                                                          {       'in_nucast_pkts': '391855',
                                                                                                                  'in_octets': '124517618963',
                                                                                                                  'in_ucast_pkts': '608743215',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '24776325',
                                                                                                                  'out_octets': '5021722419935',
                                                                                                                  'out_ucast_pkts': '3484449320',
                                                                                                                  'poller_hash': 'MTI3NDYwM3BlZXJpbnRmcG9uIDI2MjcyNw=='},
                                                                                                          {       'in_nucast_pkts': '473152',
                                                                                                                  'in_octets': '53555235816',
                                                                                                                  'in_ucast_pkts': '723727997',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '97205163',
                                                                                                                  'out_octets': '4116142567471',
                                                                                                                  'out_ucast_pkts': '2902990414',
                                                                                                                  'poller_hash': 'MTI3NDY5N3BlZXJpbnRmcG9uIDI2NjEzNA=='},
                                                                                                          {       'in_nucast_pkts': '149039',
                                                                                                                  'in_octets': '101651505210',
                                                                                                                  'in_ucast_pkts': '807424406',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9878481',
                                                                                                                  'out_octets': '2920412619836',
                                                                                                                  'out_ucast_pkts': '2006293755',
                                                                                                                  'poller_hash': 'MTI3NDhBRHBlZXJpbnRmcG9uIDI3MjE5OQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '1/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '100245259918',
                                                                                         'out_octets': '352497016173165',
                                                                                         'out_ucast_pkts': '163874446524',
                                                                                         'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/8',
                                                                                                                 'id': '1/8',
                                                                                                                 'links': [       {       'connected_time': '27376190',
                                                                                                                                          'fiber_length': 23652,
                                                                                                                                          'poller_hash': 'MTU4RkFBRnBlZXJwb25wb24gODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '3195474',
                                                                                                                                          'fiber_length': 24063,
                                                                                                                                          'poller_hash': 'QTlDMzlEcGVlcnBvbnBvbiAxNDEwODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '8535227',
                                                                                                                                          'fiber_length': 23935,
                                                                                                                                          'poller_hash': 'OTU2RTQzcGVlcnBvbnBvbiAxNDE2ODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '8535187',
                                                                                                                                          'fiber_length': 23176,
                                                                                                                                          'poller_hash': 'ODcyNDI4cGVlcnBvbnBvbiAxNDM4NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '8535211',
                                                                                                                                          'fiber_length': 23217,
                                                                                                                                          'poller_hash': 'OTU2Rjk3cGVlcnBvbnBvbiAxNDQ3OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '12045338',
                                                                                                                                          'fiber_length': 23283,
                                                                                                                                          'poller_hash': 'OTU2RjhDcGVlcnBvbnBvbiAxNDUxMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '1965803',
                                                                                                                                          'fiber_length': 23039,
                                                                                                                                          'poller_hash': 'OTU2RTMycGVlcnBvbnBvbiAxNDUxMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '21886857',
                                                                                                                                          'fiber_length': 23913,
                                                                                                                                          'poller_hash': 'OTU2RjAycGVlcnBvbnBvbiAxNDUyNzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '8535185',
                                                                                                                                          'fiber_length': 23460,
                                                                                                                                          'poller_hash': 'OTU2RkE0cGVlcnBvbnBvbiAxNDU1NDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '4222606',
                                                                                                                                          'fiber_length': 23735,
                                                                                                                                          'poller_hash': 'OTU2RjQ2cGVlcnBvbnBvbiAxNDU4MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '2297172',
                                                                                                                                          'fiber_length': 23065,
                                                                                                                                          'poller_hash': 'QTlDMkI0cGVlcnBvbnBvbiAxNDYyMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '17942222',
                                                                                                                                          'fiber_length': 23188,
                                                                                                                                          'poller_hash': 'QTlDMzdDcGVlcnBvbnBvbiAxNDY2Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '4916795',
                                                                                                                                          'fiber_length': 23751,
                                                                                                                                          'poller_hash': 'QzgxRUNBcGVlcnBvbnBvbiAxNTY3NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '7316566',
                                                                                                                                          'fiber_length': 23011,
                                                                                                                                          'poller_hash': 'QzgzOTA0cGVlcnBvbnBvbiAxNTczMDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '21886272',
                                                                                                                                          'fiber_length': 23066,
                                                                                                                                          'poller_hash': 'QzgxRDcwcGVlcnBvbnBvbiAxNTc0MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '21886924',
                                                                                                                                          'fiber_length': 23260,
                                                                                                                                          'poller_hash': 'QzgxNTYzcGVlcnBvbnBvbiAyMDc3MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '15709246',
                                                                                                                                          'fiber_length': 23820,
                                                                                                                                          'poller_hash': 'MTI3NTBDNHBlZXJwb25wb24gMjE4MjM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '5782919',
                                                                                                                                          'fiber_length': 23081,
                                                                                                                                          'poller_hash': 'MTI3NDMyNXBlZXJwb25wb24gMjQ2OTQw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '20986130',
                                                                                                                                          'fiber_length': 23355,
                                                                                                                                          'poller_hash': 'MTI3NDYwM3BlZXJwb25wb24gMjYyNzI3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '8535154',
                                                                                                                                          'fiber_length': 23271,
                                                                                                                                          'poller_hash': 'MTI3NDY5N3BlZXJwb25wb24gMjY2MTM0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226},
                                                                                                                                  {       'connected_time': '8535166',
                                                                                                                                          'fiber_length': 23177,
                                                                                                                                          'poller_hash': 'MTI3NDhBRHBlZXJwb25wb24gMjcyMTk5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.226}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/8',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'NDcxOTA1MDEwMzU2',
                                                           'serial_number': '471905010356',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43502592'},
                                                   {       'id': '2',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '2/1',
                                                                                         'in_discards': '72934',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3010843',
                                                                                         'in_octets': '25941199461262',
                                                                                         'in_ucast_pkts': '46691286591',
                                                                                         'links': [       {       'in_nucast_pkts': '437504',
                                                                                                                  'in_octets': '201622042647',
                                                                                                                  'in_ucast_pkts': '529003547',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9925337',
                                                                                                                  'out_octets': '1423900731783',
                                                                                                                  'out_ucast_pkts': '1107384232',
                                                                                                                  'poller_hash': 'ODJBNTAzcGVlcmludGZwb24gMTMw'},
                                                                                                          {       'in_nucast_pkts': '18',
                                                                                                                  'in_octets': '65218191889',
                                                                                                                  'in_ucast_pkts': '214564732',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3754816',
                                                                                                                  'out_octets': '1027542914147',
                                                                                                                  'out_ucast_pkts': '813081171',
                                                                                                                  'poller_hash': 'ODY5NDg2cGVlcmludGZwb24gMTQxNzUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '871461322933',
                                                                                                                  'in_ucast_pkts': '4003059264',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '13766508254334',
                                                                                                                  'out_ucast_pkts': '10249686258',
                                                                                                                  'poller_hash': 'ODcyNDFDcGVlcmludGZwb24gMTQzODMy'},
                                                                                                          {       'in_nucast_pkts': '29714838',
                                                                                                                  'in_octets': '10051634793724',
                                                                                                                  'in_ucast_pkts': '14725841733',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66254244623',
                                                                                                                  'out_octets': '124634468626549',
                                                                                                                  'out_ucast_pkts': '27297610348',
                                                                                                                  'poller_hash': 'OTU2RjVBcGVlcmludGZwb24gMTQ0Nzc5'},
                                                                                                          {       'in_nucast_pkts': '38069',
                                                                                                                  'in_octets': '17860107366',
                                                                                                                  'in_ucast_pkts': '45377072',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1302894',
                                                                                                                  'out_octets': '270833296415',
                                                                                                                  'out_ucast_pkts': '216094353',
                                                                                                                  'poller_hash': 'QjUxMUVCcGVlcmludGZwb24gMTU1MDEx'},
                                                                                                          {       'in_nucast_pkts': '149868',
                                                                                                                  'in_octets': '50754844806',
                                                                                                                  'in_ucast_pkts': '256780881',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9925744',
                                                                                                                  'out_octets': '1234030289975',
                                                                                                                  'out_ucast_pkts': '970362070',
                                                                                                                  'poller_hash': 'QzgxRDY2cGVlcmludGZwb24gMTU3NDE0'},
                                                                                                          {       'in_nucast_pkts': '148451',
                                                                                                                  'in_octets': '123019137815',
                                                                                                                  'in_ucast_pkts': '535610090',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9925763',
                                                                                                                  'out_octets': '2026845143889',
                                                                                                                  'out_ucast_pkts': '1604495682',
                                                                                                                  'poller_hash': 'Q0JFMDlGcGVlcmludGZwb24gMTY4MTA1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '351129482327',
                                                                                                                  'in_ucast_pkts': '1782229673',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17',
                                                                                                                  'out_octets': '5798714029569',
                                                                                                                  'out_ucast_pkts': '4432117361',
                                                                                                                  'poller_hash': 'QzgzNjU0cGVlcmludGZwb24gMTY4NTQ0'},
                                                                                                          {       'in_nucast_pkts': '21428',
                                                                                                                  'in_octets': '423782856949',
                                                                                                                  'in_ucast_pkts': '433780937',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1188985',
                                                                                                                  'out_octets': '464155469959',
                                                                                                                  'out_ucast_pkts': '476829617',
                                                                                                                  'poller_hash': 'MTI3NDA1N3BlZXJpbnRmcG9uIDI0NzEyNQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '2/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '42494062042',
                                                                                         'out_octets': '188791791440747',
                                                                                         'out_ucast_pkts': '104735589548',
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '2/1',
                                                                                                                 'id': '2/1',
                                                                                                                 'links': [       {       'connected_time': '8535406',
                                                                                                                                          'fiber_length': 23220,
                                                                                                                                          'poller_hash': 'ODJBNTAzcGVlcnBvbnBvbiAxMzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '3179647',
                                                                                                                                          'fiber_length': 23801,
                                                                                                                                          'poller_hash': 'ODY5NDg2cGVlcnBvbnBvbiAxNDE3NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '8535196',
                                                                                                                                          'fiber_length': 23094,
                                                                                                                                          'poller_hash': 'ODcyNDFDcGVlcnBvbnBvbiAxNDM4MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '21886922',
                                                                                                                                          'fiber_length': 23586,
                                                                                                                                          'poller_hash': 'OTU2RjVBcGVlcnBvbnBvbiAxNDQ3Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '1112447',
                                                                                                                                          'fiber_length': 23817,
                                                                                                                                          'poller_hash': 'QjUxMUVCcGVlcnBvbnBvbiAxNTUwMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '8535138',
                                                                                                                                          'fiber_length': 23549,
                                                                                                                                          'poller_hash': 'QzgxRDY2cGVlcnBvbnBvbiAxNTc0MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '8535235',
                                                                                                                                          'fiber_length': 23061,
                                                                                                                                          'poller_hash': 'Q0JFMDlGcGVlcnBvbnBvbiAxNjgxMDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '4330025',
                                                                                                                                          'fiber_length': 23146,
                                                                                                                                          'poller_hash': 'QzgzNjU0cGVlcnBvbnBvbiAxNjg1NDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212},
                                                                                                                                  {       'connected_time': '1015547',
                                                                                                                                          'fiber_length': 23772,
                                                                                                                                          'poller_hash': 'MTI3NDA1N3BlZXJwb25wb24gMjQ3MTI1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.212}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '2/1',
                                                                                                                 'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '2/2',
                                                                                         'in_discards': '2275',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1178414',
                                                                                         'in_octets': '1827572833390',
                                                                                         'in_ucast_pkts': '4244256455',
                                                                                         'links': [       {       'in_nucast_pkts': '23817',
                                                                                                                  'in_octets': '17439765072',
                                                                                                                  'in_ucast_pkts': '42308720',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1284694',
                                                                                                                  'out_octets': '172830860635',
                                                                                                                  'out_ucast_pkts': '134723164',
                                                                                                                  'poller_hash': 'QjBBMkZBcGVlcmludGZwb24gMTUyMzYz'},
                                                                                                          {       'in_nucast_pkts': '15562',
                                                                                                                  'in_octets': '22105695454',
                                                                                                                  'in_ucast_pkts': '60766155',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17525873',
                                                                                                                  'out_octets': '136655581773',
                                                                                                                  'out_ucast_pkts': '117421896',
                                                                                                                  'poller_hash': 'QjUxMTgzcGVlcmludGZwb24gMTUyMzkx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '435847020448',
                                                                                                                  'in_ucast_pkts': '574198419',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31',
                                                                                                                  'out_octets': '1205044598618',
                                                                                                                  'out_ucast_pkts': '1191998714',
                                                                                                                  'poller_hash': 'QzgzRTE2cGVlcmludGZwb24gMTYxNjc3'},
                                                                                                          {       'in_nucast_pkts': '19667',
                                                                                                                  'in_octets': '9582782205',
                                                                                                                  'in_ucast_pkts': '23431773',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1284711',
                                                                                                                  'out_octets': '57950917193',
                                                                                                                  'out_ucast_pkts': '53794656',
                                                                                                                  'poller_hash': 'MTI3NDEwOHBlZXJpbnRmcG9uIDI0ODYwNQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '2/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '407637473',
                                                                                         'out_octets': '16997377686370',
                                                                                         'out_ucast_pkts': '13696515704',
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '2/2',
                                                                                                                 'id': '2/2',
                                                                                                                 'links': [       {       'connected_time': '1096785',
                                                                                                                                          'fiber_length': 28584,
                                                                                                                                          'poller_hash': 'QjBBMkZBcGVlcnBvbnBvbiAxNTIzNjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.276},
                                                                                                                                  {       'connected_time': '14920503',
                                                                                                                                          'fiber_length': 29479,
                                                                                                                                          'poller_hash': 'QjUxMTgzcGVlcnBvbnBvbiAxNTIzOTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.276},
                                                                                                                                  {       'connected_time': '107129',
                                                                                                                                          'fiber_length': 30331,
                                                                                                                                          'poller_hash': 'QzgzRTE2cGVlcnBvbnBvbiAxNjE2Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.276},
                                                                                                                                  {       'connected_time': '1096769',
                                                                                                                                          'fiber_length': 29772,
                                                                                                                                          'poller_hash': 'MTI3NDEwOHBlZXJwb25wb24gMjQ4NjA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.276}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '2/2',
                                                                                                                 'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '2/3',
                                                                                         'in_discards': '14214',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5627627',
                                                                                         'in_octets': '5985312051172',
                                                                                         'in_ucast_pkts': '20318008919',
                                                                                         'links': [       {       'in_nucast_pkts': '952934',
                                                                                                                  'in_octets': '275184349711',
                                                                                                                  'in_ucast_pkts': '2349447161',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63820781',
                                                                                                                  'out_octets': '7898817808653',
                                                                                                                  'out_ucast_pkts': '5703075585',
                                                                                                                  'poller_hash': 'QzgxRTAwcGVlcmludGZwb24gMjA3NjY3'},
                                                                                                          {       'in_nucast_pkts': '68761',
                                                                                                                  'in_octets': '21557579690',
                                                                                                                  'in_ucast_pkts': '135012854',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4741761',
                                                                                                                  'out_octets': '432485975138',
                                                                                                                  'out_ucast_pkts': '305460690',
                                                                                                                  'poller_hash': 'QzgxQkExcGVlcmludGZwb24gMjA4Njc4'},
                                                                                                          {       'in_nucast_pkts': '52251',
                                                                                                                  'in_octets': '12922386007',
                                                                                                                  'in_ucast_pkts': '78409878',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3185598',
                                                                                                                  'out_octets': '358285953315',
                                                                                                                  'out_ucast_pkts': '250089661',
                                                                                                                  'poller_hash': 'MTI3NTFDNHBlZXJpbnRmcG9uIDIxODMwNw=='},
                                                                                                          {       'in_nucast_pkts': '147145',
                                                                                                                  'in_octets': '126715413151',
                                                                                                                  'in_ucast_pkts': '629485532',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9925674',
                                                                                                                  'out_octets': '1955147794818',
                                                                                                                  'out_ucast_pkts': '1486489588',
                                                                                                                  'poller_hash': 'MTI3NTA4MXBlZXJpbnRmcG9uIDIyMTQ1NQ=='},
                                                                                                          {       'in_nucast_pkts': '156524',
                                                                                                                  'in_octets': '106486482961',
                                                                                                                  'in_ucast_pkts': '268087645',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9925938',
                                                                                                                  'out_octets': '1450411781218',
                                                                                                                  'out_ucast_pkts': '1144073127',
                                                                                                                  'poller_hash': 'MTI3NDA1RXBlZXJpbnRmcG9uIDI0MTg4NQ=='},
                                                                                                          {       'in_nucast_pkts': '81240',
                                                                                                                  'in_octets': '55804787048',
                                                                                                                  'in_ucast_pkts': '320590476',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4947346',
                                                                                                                  'out_octets': '1382732837058',
                                                                                                                  'out_ucast_pkts': '1030557206',
                                                                                                                  'poller_hash': 'MTI3NDMxNnBlZXJpbnRmcG9uIDI0NzAwMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '444009922373',
                                                                                                                  'in_ucast_pkts': '983044374',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3952282178779',
                                                                                                                  'out_ucast_pkts': '3267486745',
                                                                                                                  'poller_hash': 'MTE3REMyRHBlZXJpbnRmcG9uIDI1MjY2MA=='},
                                                                                                          {       'in_nucast_pkts': '465894',
                                                                                                                  'in_octets': '287876090658',
                                                                                                                  'in_ucast_pkts': '570888221',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '29819151',
                                                                                                                  'out_octets': '1944623996488',
                                                                                                                  'out_ucast_pkts': '1621166389',
                                                                                                                  'poller_hash': 'MTI5NzQ2M3BlZXJpbnRmcG9uIDI1NjM3Mw=='},
                                                                                                          {       'in_nucast_pkts': '113292',
                                                                                                                  'in_octets': '97286440669',
                                                                                                                  'in_ucast_pkts': '281517829',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7365605',
                                                                                                                  'out_octets': '1586391557925',
                                                                                                                  'out_ucast_pkts': '1197206085',
                                                                                                                  'poller_hash': 'MTI5NzQ5RHBlZXJpbnRmcG9uIDI1Njg3Nw=='},
                                                                                                          {       'in_nucast_pkts': '149978',
                                                                                                                  'in_octets': '31348798323',
                                                                                                                  'in_ucast_pkts': '258618480',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9925949',
                                                                                                                  'out_octets': '3097849369801',
                                                                                                                  'out_ucast_pkts': '2099248066',
                                                                                                                  'poller_hash': 'MTI3NDZBNnBlZXJpbnRmcG9uIDI2NTczNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1486956069316',
                                                                                                                  'in_ucast_pkts': '1676654054',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6206946734360',
                                                                                                                  'out_ucast_pkts': '4461339522',
                                                                                                                  'poller_hash': 'MTI3MEQ0QnBlZXJpbnRmcG9uIDI2OTg2OQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '2/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '52259818424',
                                                                                         'out_octets': '155436796981893',
                                                                                         'out_ucast_pkts': '62919342585',
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '2/3',
                                                                                                                 'id': '2/3',
                                                                                                                 'links': [       {       'connected_time': '21886946',
                                                                                                                                          'fiber_length': 24607,
                                                                                                                                          'poller_hash': 'QzgxRTAwcGVlcnBvbnBvbiAyMDc2Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '4031948',
                                                                                                                                          'fiber_length': 25593,
                                                                                                                                          'poller_hash': 'QzgxQkExcGVlcnBvbnBvbiAyMDg2Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '3772835',
                                                                                                                                          'fiber_length': 24796,
                                                                                                                                          'poller_hash': 'MTI3NTFDNHBlZXJwb25wb24gMjE4MzA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '8535150',
                                                                                                                                          'fiber_length': 24986,
                                                                                                                                          'poller_hash': 'MTI3NTA4MXBlZXJwb25wb24gMjIxNDU1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '8535318',
                                                                                                                                          'fiber_length': 24181,
                                                                                                                                          'poller_hash': 'MTI3NDA1RXBlZXJwb25wb24gMjQxODg1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '4213427',
                                                                                                                                          'fiber_length': 24979,
                                                                                                                                          'poller_hash': 'MTI3NDMxNnBlZXJwb25wb24gMjQ3MDAw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '10977854',
                                                                                                                                          'fiber_length': 24694,
                                                                                                                                          'poller_hash': 'MTE3REMyRHBlZXJwb25wb24gMjUyNjYw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '25827623',
                                                                                                                                          'fiber_length': 25191,
                                                                                                                                          'poller_hash': 'MTI5NzQ2M3BlZXJwb25wb24gMjU2Mzcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '6313818',
                                                                                                                                          'fiber_length': 25306,
                                                                                                                                          'poller_hash': 'MTI5NzQ5RHBlZXJwb25wb24gMjU2ODc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '8535196',
                                                                                                                                          'fiber_length': 24928,
                                                                                                                                          'poller_hash': 'MTI3NDZBNnBlZXJwb25wb24gMjY1NzM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003},
                                                                                                                                  {       'connected_time': '8535287',
                                                                                                                                          'fiber_length': 24753,
                                                                                                                                          'poller_hash': 'MTI3MEQ0QnBlZXJwb25wb24gMjY5ODY5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.003}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '2/3',
                                                                                                                 'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '2/4',
                                                                                         'in_discards': '765',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '358886',
                                                                                         'in_octets': '525175497503',
                                                                                         'in_ucast_pkts': '635537911',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '3337260957',
                                                                                                                  'in_ucast_pkts': '10606459',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7',
                                                                                                                  'out_octets': '27998385162',
                                                                                                                  'out_ucast_pkts': '39873219',
                                                                                                                  'poller_hash': 'MTE5MzAxM3BlZXJpbnRmcG9uIDI1Nzk4OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '301914527221',
                                                                                                                  'in_ucast_pkts': '8914642848',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '358623513396',
                                                                                                                  'out_ucast_pkts': '410008016',
                                                                                                                  'poller_hash': 'MTI3NDU5QXBlZXJpbnRmcG9uIDI2MjUzNw=='},
                                                                                                          {       'in_nucast_pkts': '76059',
                                                                                                                  'in_octets': '3264019972',
                                                                                                                  'in_ucast_pkts': '8575302',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4181556',
                                                                                                                  'out_octets': '24322106207',
                                                                                                                  'out_ucast_pkts': '22432106',
                                                                                                                  'poller_hash': 'MTI3NDdEQ3BlZXJpbnRmcG9uIDI2MjYxOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '9374293103',
                                                                                                                  'in_ucast_pkts': '30627598',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7',
                                                                                                                  'out_octets': '183004992559',
                                                                                                                  'out_ucast_pkts': '153021887',
                                                                                                                  'poller_hash': 'MTI3NDgxMHBlZXJpbnRmcG9uIDI2NDk2Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '36005028032',
                                                                                                                  'in_ucast_pkts': '74775443',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '182006938250',
                                                                                                                  'out_ucast_pkts': '171515576',
                                                                                                                  'poller_hash': 'MTI5NzREMHBlZXJpbnRmcG9uIDI3NDAyMQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '2/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '15420236',
                                                                                         'out_octets': '1060531083473',
                                                                                         'out_ucast_pkts': '1044489137',
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '2/4',
                                                                                                                 'id': '2/4',
                                                                                                                 'links': [       {       'connected_time': '8535281',
                                                                                                                                          'fiber_length': 22250,
                                                                                                                                          'poller_hash': 'MTE5MzAxM3BlZXJwb25wb24gMjU3OTg5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.99},
                                                                                                                                  {       'connected_time': '1367426',
                                                                                                                                          'fiber_length': 22285,
                                                                                                                                          'poller_hash': 'MTI3NDU5QXBlZXJwb25wb24gMjYyNTM3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.99},
                                                                                                                                  {       'connected_time': '3549183',
                                                                                                                                          'fiber_length': 22282,
                                                                                                                                          'poller_hash': 'MTI3NDdEQ3BlZXJwb25wb24gMjYyNjE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.99},
                                                                                                                                  {       'connected_time': '1370851',
                                                                                                                                          'fiber_length': 22272,
                                                                                                                                          'poller_hash': 'MTI3NDgxMHBlZXJwb25wb24gMjY0OTYz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.99},
                                                                                                                                  {       'connected_time': '4217748',
                                                                                                                                          'fiber_length': 22260,
                                                                                                                                          'poller_hash': 'MTI5NzREMHBlZXJwb25wb24gMjc0MDIx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 4.99}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '2/4',
                                                                                                                 'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': False,
                                                                                         'id': '2/5',
                                                                                         'name': 'gpon '
                                                                                                 '2/5',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzU='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '2/6',
                                                                                         'name': 'gpon '
                                                                                                 '2/6',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzY='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '2/7',
                                                                                         'name': 'gpon '
                                                                                                 '2/7',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzc='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '2/8',
                                                                                         'name': 'gpon '
                                                                                                 '2/8',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzg='}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'NDcxOTA0MDA0Mjgw',
                                                           'serial_number': '471904004280',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43502119'},
                                                   {       'id': '3',
                                                           'interfaces': [       {       'admin_up': False,
                                                                                         'id': '3/1',
                                                                                         'name': 'gpon '
                                                                                                 '3/1',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzE='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '3/2',
                                                                                         'name': 'gpon '
                                                                                                 '3/2',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzI='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '3/3',
                                                                                         'name': 'gpon '
                                                                                                 '3/3',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzM='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '3/4',
                                                                                         'name': 'gpon '
                                                                                                 '3/4',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzQ='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '3/5',
                                                                                         'name': 'gpon '
                                                                                                 '3/5',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzU='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '3/6',
                                                                                         'name': 'gpon '
                                                                                                 '3/6',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzY='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '3/7',
                                                                                         'name': 'gpon '
                                                                                                 '3/7',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzc='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '3/8',
                                                                                         'name': 'gpon '
                                                                                                 '3/8',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA0MDA0Mjk1Z3BvbiAzLzg='}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'NDcxOTA0MDA0Mjk1',
                                                           'serial_number': '471904004295',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43502614'},
                                                   {       'id': '5',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '5/1',
                                                                                         'in_discards': '1060408',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '10097909',
                                                                                         'in_octets': '19705152690933',
                                                                                         'in_ucast_pkts': '76404124200',
                                                                                         'links': [       {       'in_nucast_pkts': '16132',
                                                                                                                  'in_octets': '5217317501',
                                                                                                                  'in_ucast_pkts': '13265011',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '512745',
                                                                                                                  'out_octets': '59066722423',
                                                                                                                  'out_ucast_pkts': '48652320',
                                                                                                                  'poller_hash': 'MkU3RTY5cGVlcmludGZwb24gODU='},
                                                                                                          {       'in_nucast_pkts': '5025707',
                                                                                                                  'in_octets': '2882454264172',
                                                                                                                  'in_ucast_pkts': '6559021383',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22467087',
                                                                                                                  'out_octets': '20093335431437',
                                                                                                                  'out_ucast_pkts': '37371223866',
                                                                                                                  'poller_hash': 'MTUzQ0JwZWVyaW50ZnBvbiA4Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '488286698603',
                                                                                                                  'in_ucast_pkts': '2354089094',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '12907058635493',
                                                                                                                  'out_ucast_pkts': '8920050595',
                                                                                                                  'poller_hash': 'MzI5NjdDcGVlcmludGZwb24gODc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '703931399780',
                                                                                                                  'in_ucast_pkts': '2176207907',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '8983668520292',
                                                                                                                  'out_ucast_pkts': '2581477189',
                                                                                                                  'poller_hash': 'NEU2NDM0cGVlcmludGZwb24gODg='},
                                                                                                          {       'in_nucast_pkts': '119847',
                                                                                                                  'in_octets': '3285662554339',
                                                                                                                  'in_ucast_pkts': '13290553447',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22461463',
                                                                                                                  'out_octets': '42272320425183',
                                                                                                                  'out_ucast_pkts': '28699009384',
                                                                                                                  'poller_hash': 'MkZFOUY1cGVlcmludGZwb24gODk='},
                                                                                                          {       'in_nucast_pkts': '772628',
                                                                                                                  'in_octets': '256013793522',
                                                                                                                  'in_ucast_pkts': '1201570648',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22463981',
                                                                                                                  'out_octets': '5048842042384',
                                                                                                                  'out_ucast_pkts': '3902447418',
                                                                                                                  'poller_hash': 'ODZDMDdwZWVyaW50ZnBvbiA5MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '304734107530',
                                                                                                                  'in_ucast_pkts': '1376282561',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53',
                                                                                                                  'out_octets': '4355107613306',
                                                                                                                  'out_ucast_pkts': '3284430219',
                                                                                                                  'poller_hash': 'QjA4RUQ5cGVlcmludGZwb24gOTI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '500108880421',
                                                                                                                  'in_ucast_pkts': '4134327676',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '21131600484932',
                                                                                                                  'out_ucast_pkts': '14546673917',
                                                                                                                  'poller_hash': 'MkVEOTU3cGVlcmludGZwb24gOTU='},
                                                                                                          {       'in_nucast_pkts': '363',
                                                                                                                  'in_octets': '796664345340',
                                                                                                                  'in_ucast_pkts': '1525998582',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22467466',
                                                                                                                  'out_octets': '4519949468026',
                                                                                                                  'out_ucast_pkts': '3513563447',
                                                                                                                  'poller_hash': 'MkVCOTdFcGVlcmludGZwb24gOTc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '141651621819',
                                                                                                                  'in_ucast_pkts': '785664984',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3722903240236',
                                                                                                                  'out_ucast_pkts': '2681635815',
                                                                                                                  'poller_hash': 'NTYxNTE1cGVlcmludGZwb24gOTg='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '183220988886',
                                                                                                                  'in_ucast_pkts': '1360407416',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144',
                                                                                                                  'out_octets': '6834575735145',
                                                                                                                  'out_ucast_pkts': '4670534883',
                                                                                                                  'poller_hash': 'QTlDN0Y4cGVlcmludGZwb24gMTA0'},
                                                                                                          {       'in_nucast_pkts': '396449',
                                                                                                                  'in_octets': '14189285436',
                                                                                                                  'in_ucast_pkts': '50409698',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10719342',
                                                                                                                  'out_octets': '143182948410',
                                                                                                                  'out_ucast_pkts': '13005809172',
                                                                                                                  'poller_hash': 'NDE1NDhwZWVyaW50ZnBvbiAxMDU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '281575181664',
                                                                                                                  'in_ucast_pkts': '1074234314',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6055451795990',
                                                                                                                  'out_ucast_pkts': '167983691',
                                                                                                                  'poller_hash': 'NEQ0MjQwcGVlcmludGZwb24gMTA3'},
                                                                                                          {       'in_nucast_pkts': '2859421',
                                                                                                                  'in_octets': '17022924844',
                                                                                                                  'in_ucast_pkts': '201092487',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '137799884',
                                                                                                                  'out_octets': '941085938785',
                                                                                                                  'out_ucast_pkts': '632210027',
                                                                                                                  'poller_hash': 'NEQ0MjYzcGVlcmludGZwb24gMTA4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '288861334540',
                                                                                                                  'in_ucast_pkts': '2040680227',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3827217515639',
                                                                                                                  'out_ucast_pkts': '747869153',
                                                                                                                  'poller_hash': 'NEQxMTFBcGVlcmludGZwb24gMTEw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '317901599333',
                                                                                                                  'in_ucast_pkts': '1348604649',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7296536629219',
                                                                                                                  'out_ucast_pkts': '5213425229',
                                                                                                                  'poller_hash': 'MkZFQUY4cGVlcmludGZwb24gMTEx'},
                                                                                                          {       'in_nucast_pkts': '31848972',
                                                                                                                  'in_octets': '704352624574',
                                                                                                                  'in_ucast_pkts': '1540235433',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7998104185',
                                                                                                                  'out_octets': '16188451425278',
                                                                                                                  'out_ucast_pkts': '20285512084337',
                                                                                                                  'poller_hash': 'RjBGRnBlZXJpbnRmcG9uIDExMg=='},
                                                                                                          {       'in_nucast_pkts': '1041257',
                                                                                                                  'in_octets': '230273284542',
                                                                                                                  'in_ucast_pkts': '719725631',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4384041',
                                                                                                                  'out_octets': '3184777872853',
                                                                                                                  'out_ucast_pkts': '2467726237',
                                                                                                                  'poller_hash': 'MTk5QTNCcGVlcmludGZwb24gMTEz'},
                                                                                                          {       'in_nucast_pkts': '333078',
                                                                                                                  'in_octets': '220014967875',
                                                                                                                  'in_ucast_pkts': '402008295',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12074348',
                                                                                                                  'out_octets': '1847074398113',
                                                                                                                  'out_ucast_pkts': '1461736416',
                                                                                                                  'poller_hash': 'MTk5OUY1cGVlcmludGZwb24gMTE0'},
                                                                                                          {       'in_nucast_pkts': '342626',
                                                                                                                  'in_octets': '7062958895',
                                                                                                                  'in_ucast_pkts': '28952155',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '137078389',
                                                                                                                  'out_octets': '448868958515',
                                                                                                                  'out_ucast_pkts': '185816823',
                                                                                                                  'poller_hash': 'NTEyNEZFcGVlcmludGZwb24gMjE4'},
                                                                                                          {       'in_nucast_pkts': '928895',
                                                                                                                  'in_octets': '400555611335',
                                                                                                                  'in_ucast_pkts': '1688932269',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58763446',
                                                                                                                  'out_octets': '7557537191326',
                                                                                                                  'out_ucast_pkts': '5701787054',
                                                                                                                  'poller_hash': 'MzIyOTBGcGVlcmludGZwb24gMjgx'},
                                                                                                          {       'in_nucast_pkts': '222',
                                                                                                                  'in_octets': '243155437368',
                                                                                                                  'in_ucast_pkts': '2025242807',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22467581',
                                                                                                                  'out_octets': '11244046990977',
                                                                                                                  'out_ucast_pkts': '7683206093',
                                                                                                                  'poller_hash': 'NTYxOTUxcGVlcmludGZwb24gNDY3'},
                                                                                                          {       'in_nucast_pkts': '769216',
                                                                                                                  'in_octets': '565564191804',
                                                                                                                  'in_ucast_pkts': '1996567516',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22468751',
                                                                                                                  'out_octets': '9090163230280',
                                                                                                                  'out_ucast_pkts': '6749203937',
                                                                                                                  'poller_hash': 'MzhGRjU4cGVlcmludGZwb24gMTUwMTMx'},
                                                                                                          {       'in_nucast_pkts': '13158',
                                                                                                                  'in_octets': '4919680136',
                                                                                                                  'in_ucast_pkts': '14234630',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715408',
                                                                                                                  'out_octets': '48196197744',
                                                                                                                  'out_ucast_pkts': '37788154',
                                                                                                                  'poller_hash': 'QzgyMEJGcGVlcmludGZwb24gMTU4NTQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '182807811332',
                                                                                                                  'in_ucast_pkts': '908092578',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '107',
                                                                                                                  'out_octets': '4537439198505',
                                                                                                                  'out_ucast_pkts': '7500790876',
                                                                                                                  'poller_hash': 'RENGNDQwcGVlcmludGZwb24gMTY2NDM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '58832771952',
                                                                                                                  'in_ucast_pkts': '113999801',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53',
                                                                                                                  'out_octets': '415603916960',
                                                                                                                  'out_ucast_pkts': '357578338',
                                                                                                                  'poller_hash': 'QzgzQkQzcGVlcmludGZwb24gMTgwMzc3'},
                                                                                                          {       'in_nucast_pkts': '13421',
                                                                                                                  'in_octets': '6232943204',
                                                                                                                  'in_ucast_pkts': '20374189',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715434',
                                                                                                                  'out_octets': '106947127022',
                                                                                                                  'out_ucast_pkts': '87351713',
                                                                                                                  'poller_hash': 'QzgxREEwcGVlcmludGZwb24gMjA1NzYz'},
                                                                                                          {       'in_nucast_pkts': '11464',
                                                                                                                  'in_octets': '5975324839',
                                                                                                                  'in_ucast_pkts': '45859413',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715554',
                                                                                                                  'out_octets': '176442923723',
                                                                                                                  'out_ucast_pkts': '121411173',
                                                                                                                  'poller_hash': 'NDJCRUMwcGVlcmludGZwb24gMjMxMDI4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzI5NjQ2cGVlcmludGZwb24gMjMyNTc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '258099124702',
                                                                                                                  'in_ucast_pkts': '2106397237',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '165',
                                                                                                                  'out_octets': '12348352455368',
                                                                                                                  'out_ucast_pkts': '8363370646',
                                                                                                                  'poller_hash': 'MTI3NEIyOXBlZXJpbnRmcG9uIDIzNzc0Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '31362557418',
                                                                                                                  'in_ucast_pkts': '263983856',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1126834809077',
                                                                                                                  'out_ucast_pkts': '796015076',
                                                                                                                  'poller_hash': 'MTI3MEU1MnBlZXJpbnRmcG9uIDI3Njk3Ng=='},
                                                                                                          {       'in_nucast_pkts': '12324',
                                                                                                                  'in_octets': '2556046707',
                                                                                                                  'in_ucast_pkts': '10243465',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715543',
                                                                                                                  'out_octets': '45088981865',
                                                                                                                  'out_ucast_pkts': '34886336',
                                                                                                                  'poller_hash': 'MTI3NDg5OXBlZXJpbnRmcG9uIDI3ODIzMQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '132324884430',
                                                                                         'out_octets': '490559698057947',
                                                                                         'out_ucast_pkts': '225568394917',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/1',
                                                                                                                 'id': '5/1',
                                                                                                                 'links': [       {       'connected_time': '763757',
                                                                                                                                          'fiber_length': 19928,
                                                                                                                                          'poller_hash': 'MkU3RTY5cGVlcnBvbnBvbiA4NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '11964828',
                                                                                                                                          'fiber_length': 19792,
                                                                                                                                          'poller_hash': 'MTUzQ0JwZWVycG9ucG9uIDg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '28849411',
                                                                                                                                          'fiber_length': 19917,
                                                                                                                                          'poller_hash': 'MzI5NjdDcGVlcnBvbnBvbiA4Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611421',
                                                                                                                                          'fiber_length': 19985,
                                                                                                                                          'poller_hash': 'NEU2NDM0cGVlcnBvbnBvbiA4OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '552443',
                                                                                                                                          'fiber_length': 19961,
                                                                                                                                          'poller_hash': 'MkZFOUY1cGVlcnBvbnBvbiA4OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '552420',
                                                                                                                                          'fiber_length': 20035,
                                                                                                                                          'poller_hash': 'ODZDMDdwZWVycG9ucG9uIDkx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611443',
                                                                                                                                          'fiber_length': 19832,
                                                                                                                                          'poller_hash': 'QjA4RUQ5cGVlcnBvbnBvbiA5Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '639398',
                                                                                                                                          'fiber_length': 19745,
                                                                                                                                          'poller_hash': 'MkVEOTU3cGVlcnBvbnBvbiA5NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '552210',
                                                                                                                                          'fiber_length': 19803,
                                                                                                                                          'poller_hash': 'MkVCOTdFcGVlcnBvbnBvbiA5Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '639287',
                                                                                                                                          'fiber_length': 19554,
                                                                                                                                          'poller_hash': 'NTYxNTE1cGVlcnBvbnBvbiA5OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '67816077',
                                                                                                                                          'fiber_length': 19856,
                                                                                                                                          'poller_hash': 'QTlDN0Y4cGVlcnBvbnBvbiAxMDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '16338099',
                                                                                                                                          'fiber_length': 19821,
                                                                                                                                          'poller_hash': 'NDE1NDhwZWVycG9ucG9uIDEwNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '10841403',
                                                                                                                                          'fiber_length': 19860,
                                                                                                                                          'poller_hash': 'NEQ0MjQwcGVlcnBvbnBvbiAxMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '3210544',
                                                                                                                                          'fiber_length': 19925,
                                                                                                                                          'poller_hash': 'NEQ0MjYzcGVlcnBvbnBvbiAxMDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '639422',
                                                                                                                                          'fiber_length': 20176,
                                                                                                                                          'poller_hash': 'NEQxMTFBcGVlcnBvbnBvbiAxMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '639723',
                                                                                                                                          'fiber_length': 19680,
                                                                                                                                          'poller_hash': 'MkZFQUY4cGVlcnBvbnBvbiAxMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '11965356',
                                                                                                                                          'fiber_length': 19513,
                                                                                                                                          'poller_hash': 'RjBGRnBlZXJwb25wb24gMTEy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '6640172',
                                                                                                                                          'fiber_length': 20268,
                                                                                                                                          'poller_hash': 'MTk5QTNCcGVlcnBvbnBvbiAxMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '552423',
                                                                                                                                          'fiber_length': 20055,
                                                                                                                                          'poller_hash': 'MTk5OUY1cGVlcnBvbnBvbiAxMTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611456',
                                                                                                                                          'fiber_length': 19681,
                                                                                                                                          'poller_hash': 'NTEyNEZFcGVlcnBvbnBvbiAyMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '21886767',
                                                                                                                                          'fiber_length': 20045,
                                                                                                                                          'poller_hash': 'MzIyOTBGcGVlcnBvbnBvbiAyODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '552572',
                                                                                                                                          'fiber_length': 19727,
                                                                                                                                          'poller_hash': 'NTYxOTUxcGVlcnBvbnBvbiA0Njc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '551973',
                                                                                                                                          'fiber_length': 20054,
                                                                                                                                          'poller_hash': 'MzhGRjU4cGVlcnBvbnBvbiAxNTAxMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611445',
                                                                                                                                          'fiber_length': 19794,
                                                                                                                                          'poller_hash': 'QzgyMEJGcGVlcnBvbnBvbiAxNTg1NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '19938493',
                                                                                                                                          'fiber_length': 20098,
                                                                                                                                          'poller_hash': 'RENGNDQwcGVlcnBvbnBvbiAxNjY0Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611452',
                                                                                                                                          'fiber_length': 19825,
                                                                                                                                          'poller_hash': 'QzgzQkQzcGVlcnBvbnBvbiAxODAzNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611446',
                                                                                                                                          'fiber_length': 20071,
                                                                                                                                          'poller_hash': 'QzgxREEwcGVlcnBvbnBvbiAyMDU3NjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611450',
                                                                                                                                          'fiber_length': 19811,
                                                                                                                                          'poller_hash': 'NDJCRUMwcGVlcnBvbnBvbiAyMzEwMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '4926848',
                                                                                                                                          'fiber_length': 19905,
                                                                                                                                          'poller_hash': 'MzI5NjQ2cGVlcnBvbnBvbiAyMzI1Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '26451742',
                                                                                                                                          'fiber_length': 19837,
                                                                                                                                          'poller_hash': 'MTI3NEIyOXBlZXJwb25wb24gMjM3NzQz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '10362629',
                                                                                                                                          'fiber_length': 19890,
                                                                                                                                          'poller_hash': 'MTI3MEU1MnBlZXJwb25wb24gMjc2OTc2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72},
                                                                                                                                  {       'connected_time': '611450',
                                                                                                                                          'fiber_length': 19828,
                                                                                                                                          'poller_hash': 'MTI3NDg5OXBlZXJwb25wb24gMjc4MjMx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.72}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/1',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/2',
                                                                                         'in_discards': '95378',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '6097180',
                                                                                         'in_octets': '27623105774724',
                                                                                         'in_ucast_pkts': '94576080726',
                                                                                         'links': [       {       'in_nucast_pkts': '326',
                                                                                                                  'in_octets': '649413295360',
                                                                                                                  'in_ucast_pkts': '2655976580',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22489474',
                                                                                                                  'out_octets': '11602564460277',
                                                                                                                  'out_ucast_pkts': '19984268806507',
                                                                                                                  'poller_hash': 'RjQzQnBlZXJpbnRmcG9uIDExOA=='},
                                                                                                          {       'in_nucast_pkts': '245',
                                                                                                                  'in_octets': '719689433471',
                                                                                                                  'in_ucast_pkts': '5671585201',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22482292',
                                                                                                                  'out_octets': '18713806804077',
                                                                                                                  'out_ucast_pkts': '12874460983',
                                                                                                                  'poller_hash': 'NDAxMTUycGVlcmludGZwb24gMTE5'},
                                                                                                          {       'in_nucast_pkts': '10954',
                                                                                                                  'in_octets': '3386317179',
                                                                                                                  'in_ucast_pkts': '30440142',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715302',
                                                                                                                  'out_octets': '169988869151',
                                                                                                                  'out_ucast_pkts': '119230868',
                                                                                                                  'poller_hash': 'MUQwNEZGcGVlcmludGZwb24gMTIw'},
                                                                                                          {       'in_nucast_pkts': '100925',
                                                                                                                  'in_octets': '258567980858',
                                                                                                                  'in_ucast_pkts': '1424666635',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22477666',
                                                                                                                  'out_octets': '5143638288303',
                                                                                                                  'out_ucast_pkts': '23862338857506',
                                                                                                                  'poller_hash': 'MTMwNDhwZWVyaW50ZnBvbiAxMjE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '148115074921',
                                                                                                                  'in_ucast_pkts': '451738954',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '988018709002',
                                                                                                                  'out_ucast_pkts': '908737324',
                                                                                                                  'poller_hash': 'NDAwRDY4cGVlcmludGZwb24gMTIy'},
                                                                                                          {       'in_nucast_pkts': '13128',
                                                                                                                  'in_octets': '3714264057',
                                                                                                                  'in_ucast_pkts': '18091647',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715421',
                                                                                                                  'out_octets': '166804099469',
                                                                                                                  'out_ucast_pkts': '114974538',
                                                                                                                  'poller_hash': 'QTlDMkJCcGVlcmludGZwb24gMTIz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '250433045605',
                                                                                                                  'in_ucast_pkts': '1920666624',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '197',
                                                                                                                  'out_octets': '6104174274572',
                                                                                                                  'out_ucast_pkts': '4342379576',
                                                                                                                  'poller_hash': 'MzIxMTZDcGVlcmludGZwb24gMTI0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '409244878140',
                                                                                                                  'in_ucast_pkts': '23643709677',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '197',
                                                                                                                  'out_octets': '9327843038362',
                                                                                                                  'out_ucast_pkts': '6689844986',
                                                                                                                  'poller_hash': 'MzIxMjU0cGVlcmludGZwb24gMTI1'},
                                                                                                          {       'in_nucast_pkts': '423',
                                                                                                                  'in_octets': '477024289629',
                                                                                                                  'in_ucast_pkts': '1650841580',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22467434',
                                                                                                                  'out_octets': '7829916582812',
                                                                                                                  'out_ucast_pkts': '5866150340',
                                                                                                                  'poller_hash': 'NTVGNTFDcGVlcmludGZwb24gMTI2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '265617750148',
                                                                                                                  'in_ucast_pkts': '1446313533',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '6097911542002',
                                                                                                                  'out_ucast_pkts': '4489383414',
                                                                                                                  'poller_hash': 'NDZDRTRFcGVlcmludGZwb24gMTMz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2262266768472',
                                                                                                                  'in_ucast_pkts': '9615370358',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '30503958602473',
                                                                                                                  'out_ucast_pkts': '22010405001',
                                                                                                                  'poller_hash': 'MTY0MjdGcGVlcmludGZwb24gMTQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '995133012799',
                                                                                                                  'in_ucast_pkts': '2383219406',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '9072600269656',
                                                                                                                  'out_ucast_pkts': '7168989442',
                                                                                                                  'poller_hash': 'NDMzRTY3cGVlcmludGZwb24gMTQ3'},
                                                                                                          {       'in_nucast_pkts': '832663',
                                                                                                                  'in_octets': '349934076508',
                                                                                                                  'in_ucast_pkts': '2802837189',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22408168',
                                                                                                                  'out_octets': '26412036227387',
                                                                                                                  'out_ucast_pkts': '17652693398',
                                                                                                                  'poller_hash': 'NzI0QjIycGVlcmludGZwb24gMTUz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '767591782613',
                                                                                                                  'in_ucast_pkts': '2402468774',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11250857793076',
                                                                                                                  'out_ucast_pkts': '8307630310',
                                                                                                                  'poller_hash': 'QzU4MEZwZWVyaW50ZnBvbiAxNTc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1251926285014',
                                                                                                                  'in_ucast_pkts': '6520993690',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '29168065939078',
                                                                                                                  'out_ucast_pkts': '21366703317',
                                                                                                                  'poller_hash': 'M0ZFNDA5cGVlcmludGZwb24gMTU4'},
                                                                                                          {       'in_nucast_pkts': '855584',
                                                                                                                  'in_octets': '582752194551',
                                                                                                                  'in_ucast_pkts': '24524998591',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22483769',
                                                                                                                  'out_octets': '11160446571785',
                                                                                                                  'out_ucast_pkts': '22384920654304',
                                                                                                                  'poller_hash': 'RjBGNnBlZXJpbnRmcG9uIDE2MA=='},
                                                                                                          {       'in_nucast_pkts': '780095',
                                                                                                                  'in_octets': '250958971372',
                                                                                                                  'in_ucast_pkts': '1060787875',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22465208',
                                                                                                                  'out_octets': '4937471058361',
                                                                                                                  'out_ucast_pkts': '33980145082',
                                                                                                                  'poller_hash': 'NDE1QTdwZWVyaW50ZnBvbiAxNjI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '340212950557',
                                                                                                                  'in_ucast_pkts': '2426009229',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7480531738763',
                                                                                                                  'out_ucast_pkts': '5513313330',
                                                                                                                  'poller_hash': 'NzZEMDQ5cGVlcmludGZwb24gMTY0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '926514165303',
                                                                                                                  'in_ucast_pkts': '2402646055',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '12142890560884',
                                                                                                                  'out_ucast_pkts': '8918601323',
                                                                                                                  'poller_hash': 'NzI0M0ZEcGVlcmludGZwb24gMTY1'},
                                                                                                          {       'in_nucast_pkts': '21414',
                                                                                                                  'in_octets': '271835533402',
                                                                                                                  'in_ucast_pkts': '1310311464',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22490735',
                                                                                                                  'out_octets': '6076813659711',
                                                                                                                  'out_ucast_pkts': '22247984409181',
                                                                                                                  'poller_hash': 'MTMwM0VwZWVyaW50ZnBvbiAxNjY='},
                                                                                                          {       'in_nucast_pkts': '564',
                                                                                                                  'in_octets': '879707124536',
                                                                                                                  'in_ucast_pkts': '2338869588',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11126872',
                                                                                                                  'out_octets': '5643980433835',
                                                                                                                  'out_ucast_pkts': '4583377931',
                                                                                                                  'poller_hash': 'M0QwRkJwZWVyaW50ZnBvbiAxNjg='},
                                                                                                          {       'in_nucast_pkts': '3233865',
                                                                                                                  'in_octets': '118148671324',
                                                                                                                  'in_ucast_pkts': '1584259447',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '121625379',
                                                                                                                  'out_octets': '6518939503115',
                                                                                                                  'out_ucast_pkts': '11429017368876',
                                                                                                                  'poller_hash': 'RjBGRXBlZXJpbnRmcG9uIDE2OQ=='},
                                                                                                          {       'in_nucast_pkts': '5667',
                                                                                                                  'in_octets': '1316818830',
                                                                                                                  'in_ucast_pkts': '5348367',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715398',
                                                                                                                  'out_octets': '45309879000',
                                                                                                                  'out_ucast_pkts': '35712076',
                                                                                                                  'poller_hash': 'QjUxMTU0cGVlcmludGZwb24gNDk2'},
                                                                                                          {       'in_nucast_pkts': '774761',
                                                                                                                  'in_octets': '253124882609',
                                                                                                                  'in_ucast_pkts': '2929503305',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '42004007',
                                                                                                                  'out_octets': '17994161445509',
                                                                                                                  'out_ucast_pkts': '12030220936',
                                                                                                                  'poller_hash': 'MTI3NEE4NHBlZXJpbnRmcG9uIDIzMTQ1MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '504574604762',
                                                                                                                  'in_ucast_pkts': '57200178456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '79',
                                                                                                                  'out_octets': '4851671731213',
                                                                                                                  'out_ucast_pkts': '3757744075',
                                                                                                                  'poller_hash': 'MTI3NEIwQ3BlZXJpbnRmcG9uIDIzMzIzOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '218005782423',
                                                                                                                  'in_ucast_pkts': '2004674415',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '117',
                                                                                                                  'out_octets': '5988055257164',
                                                                                                                  'out_ucast_pkts': '4132675123',
                                                                                                                  'poller_hash': 'MTI3NEI3MXBlZXJpbnRmcG9uIDIzNTUwNg=='},
                                                                                                          {       'in_nucast_pkts': '381',
                                                                                                                  'in_octets': '174684152709',
                                                                                                                  'in_ucast_pkts': '234886406',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '714402',
                                                                                                                  'out_octets': '528766016046',
                                                                                                                  'out_ucast_pkts': '438421522',
                                                                                                                  'poller_hash': 'MTI3NDFCRHBlZXJpbnRmcG9uIDI0MjcwOA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '59152336085',
                                                                                                                  'in_ucast_pkts': '224825252',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '802614207670',
                                                                                                                  'out_ucast_pkts': '627376661',
                                                                                                                  'poller_hash': 'MTI3NDYyRnBlZXJpbnRmcG9uIDI1OTkwMg=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45577526008',
                                                                                                                  'in_ucast_pkts': '306572975',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1098654135436',
                                                                                                                  'out_ucast_pkts': '791088230',
                                                                                                                  'poller_hash': 'MTI5NUZGNXBlZXJpbnRmcG9uIDI3NzgzNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '24505239936',
                                                                                                                  'in_ucast_pkts': '121303891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '735195607554',
                                                                                                                  'out_ucast_pkts': '533955275',
                                                                                                                  'poller_hash': 'MTI3M0I0NnBlZXJpbnRmcG9uIDI3OTg2MQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '30639804575',
                                                                                                                  'in_ucast_pkts': '361664975',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1464971763666',
                                                                                                                  'out_ucast_pkts': '972359711',
                                                                                                                  'poller_hash': 'MTI5NUU5RHBlZXJpbnRmcG9uIDI4MjMyMw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '122214935616',
                                                                                         'out_octets': '530304771969305',
                                                                                         'out_ucast_pkts': '267014389211',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/2',
                                                                                                                 'id': '5/2',
                                                                                                                 'links': [       {       'connected_time': '11965741',
                                                                                                                                          'fiber_length': 19904,
                                                                                                                                          'poller_hash': 'RjQzQnBlZXJwb25wb24gMTE4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '552208',
                                                                                                                                          'fiber_length': 20202,
                                                                                                                                          'poller_hash': 'NDAxMTUycGVlcnBvbnBvbiAxMTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '611461',
                                                                                                                                          'fiber_length': 20196,
                                                                                                                                          'poller_hash': 'MUQwNEZGcGVlcnBvbnBvbiAxMjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '11964782',
                                                                                                                                          'fiber_length': 19694,
                                                                                                                                          'poller_hash': 'MTMwNDhwZWVycG9ucG9uIDEyMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '639520',
                                                                                                                                          'fiber_length': 20226,
                                                                                                                                          'poller_hash': 'NDAwRDY4cGVlcnBvbnBvbiAxMjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '611449',
                                                                                                                                          'fiber_length': 19782,
                                                                                                                                          'poller_hash': 'QTlDMkJCcGVlcnBvbnBvbiAxMjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '73884952',
                                                                                                                                          'fiber_length': 19564,
                                                                                                                                          'poller_hash': 'MzIxMTZDcGVlcnBvbnBvbiAxMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '30751422',
                                                                                                                                          'fiber_length': 19701,
                                                                                                                                          'poller_hash': 'MzIxMjU0cGVlcnBvbnBvbiAxMjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '552562',
                                                                                                                                          'fiber_length': 19967,
                                                                                                                                          'poller_hash': 'NTVGNTFDcGVlcnBvbnBvbiAxMjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '69151777',
                                                                                                                                          'fiber_length': 19812,
                                                                                                                                          'poller_hash': 'NDZDRTRFcGVlcnBvbnBvbiAxMzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '639606',
                                                                                                                                          'fiber_length': 19930,
                                                                                                                                          'poller_hash': 'MTY0MjdGcGVlcnBvbnBvbiAxNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '782108',
                                                                                                                                          'fiber_length': 19884,
                                                                                                                                          'poller_hash': 'NDMzRTY3cGVlcnBvbnBvbiAxNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '417331',
                                                                                                                                          'fiber_length': 20178,
                                                                                                                                          'poller_hash': 'NzI0QjIycGVlcnBvbnBvbiAxNTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '782087',
                                                                                                                                          'fiber_length': 19712,
                                                                                                                                          'poller_hash': 'QzU4MEZwZWVycG9ucG9uIDE1Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '13267352',
                                                                                                                                          'fiber_length': 19573,
                                                                                                                                          'poller_hash': 'M0ZFNDA5cGVlcnBvbnBvbiAxNTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '11965088',
                                                                                                                                          'fiber_length': 19434,
                                                                                                                                          'poller_hash': 'RjBGNnBlZXJwb25wb24gMTYw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '11965358',
                                                                                                                                          'fiber_length': 19892,
                                                                                                                                          'poller_hash': 'NDE1QTdwZWVycG9ucG9uIDE2Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '639182',
                                                                                                                                          'fiber_length': 20016,
                                                                                                                                          'poller_hash': 'NzZEMDQ5cGVlcnBvbnBvbiAxNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '639552',
                                                                                                                                          'fiber_length': 19742,
                                                                                                                                          'poller_hash': 'NzI0M0ZEcGVlcnBvbnBvbiAxNjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '11964494',
                                                                                                                                          'fiber_length': 19794,
                                                                                                                                          'poller_hash': 'MTMwM0VwZWVycG9ucG9uIDE2Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '16927246',
                                                                                                                                          'fiber_length': 19832,
                                                                                                                                          'poller_hash': 'M0QwRkJwZWVycG9ucG9uIDE2OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '30753817',
                                                                                                                                          'fiber_length': 19438,
                                                                                                                                          'poller_hash': 'RjBGRXBlZXJwb25wb24gMTY5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '611478',
                                                                                                                                          'fiber_length': 19986,
                                                                                                                                          'poller_hash': 'QjUxMTU0cGVlcnBvbnBvbiA0OTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '73010263',
                                                                                                                                          'fiber_length': 19626,
                                                                                                                                          'poller_hash': 'MTI3NEE4NHBlZXJwb25wb24gMjMxNDUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '10695148',
                                                                                                                                          'fiber_length': 19900,
                                                                                                                                          'poller_hash': 'MTI3NEIwQ3BlZXJwb25wb24gMjMzMjM4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '1877769',
                                                                                                                                          'fiber_length': 20016,
                                                                                                                                          'poller_hash': 'MTI3NEI3MXBlZXJwb25wb24gMjM1NTA2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '611466',
                                                                                                                                          'fiber_length': 19929,
                                                                                                                                          'poller_hash': 'MTI3NDFCRHBlZXJwb25wb24gMjQyNzA4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '22198070',
                                                                                                                                          'fiber_length': 19617,
                                                                                                                                          'poller_hash': 'MTI3NDYyRnBlZXJwb25wb24gMjU5OTAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '611464',
                                                                                                                                          'fiber_length': 19952,
                                                                                                                                          'poller_hash': 'MTI5NUZGNXBlZXJwb25wb24gMjc3ODM0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '611459',
                                                                                                                                          'fiber_length': 20385,
                                                                                                                                          'poller_hash': 'MTI3M0I0NnBlZXJwb25wb24gMjc5ODYx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527},
                                                                                                                                  {       'connected_time': '2930576',
                                                                                                                                          'fiber_length': 19633,
                                                                                                                                          'poller_hash': 'MTI5NUU5RHBlZXJwb25wb24gMjgyMzIz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.527}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/2',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/3',
                                                                                         'in_discards': '55976',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3910212',
                                                                                         'in_octets': '19205111759641',
                                                                                         'in_ucast_pkts': '75324864239',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '130198788517',
                                                                                                                  'in_ucast_pkts': '1287247308',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '161',
                                                                                                                  'out_octets': '7746673676690',
                                                                                                                  'out_ucast_pkts': '9434458105',
                                                                                                                  'poller_hash': 'QzgxRDUwcGVlcmludGZwb24gMTM2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '287870027893',
                                                                                                                  'in_ucast_pkts': '30950418098',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53',
                                                                                                                  'out_octets': '3488807425996',
                                                                                                                  'out_ucast_pkts': '2585574948',
                                                                                                                  'poller_hash': 'NDJDODFFcGVlcmludGZwb24gMTM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '398877910137',
                                                                                                                  'in_ucast_pkts': '2466779513',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7144546674994',
                                                                                                                  'out_ucast_pkts': '5203476959',
                                                                                                                  'poller_hash': 'NEQxMTE5cGVlcmludGZwb24gMTM4'},
                                                                                                          {       'in_nucast_pkts': '96178',
                                                                                                                  'in_octets': '659778656409',
                                                                                                                  'in_ucast_pkts': '6233908556',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22438159',
                                                                                                                  'out_octets': '5259919706063',
                                                                                                                  'out_ucast_pkts': '22299153829388',
                                                                                                                  'poller_hash': 'RjQzNnBlZXJpbnRmcG9uIDEzOQ=='},
                                                                                                          {       'in_nucast_pkts': '18175',
                                                                                                                  'in_octets': '47137398112',
                                                                                                                  'in_ucast_pkts': '4725304090',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22502491',
                                                                                                                  'out_octets': '1419351688769',
                                                                                                                  'out_ucast_pkts': '5293009612',
                                                                                                                  'poller_hash': 'M0NGQUVwZWVyaW50ZnBvbiAxNDE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '58393902762',
                                                                                                                  'in_ucast_pkts': '402833918',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19848474092',
                                                                                                                  'out_octets': '28482534603662',
                                                                                                                  'out_ucast_pkts': '1341707865',
                                                                                                                  'poller_hash': 'NDJDNjk3cGVlcmludGZwb24gMTQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1047036989361',
                                                                                                                  'in_ucast_pkts': '2798401287',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '16024678272395',
                                                                                                                  'out_ucast_pkts': '11677616392',
                                                                                                                  'poller_hash': 'NzFBNEM0cGVlcmludGZwb24gMTQ2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '745234633617',
                                                                                                                  'in_ucast_pkts': '6260466474',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '26308933953301',
                                                                                                                  'out_ucast_pkts': '1180635617',
                                                                                                                  'poller_hash': 'MjQ4QUJEcGVlcmludGZwb24gMTQ5'},
                                                                                                          {       'in_nucast_pkts': '2887140',
                                                                                                                  'in_octets': '271775972210',
                                                                                                                  'in_ucast_pkts': '846083288',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22512584',
                                                                                                                  'out_octets': '3504265650672',
                                                                                                                  'out_ucast_pkts': '20206293829158',
                                                                                                                  'poller_hash': 'RjBGNXBlZXJpbnRmcG9uIDE1MA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1002122980499',
                                                                                                                  'in_ucast_pkts': '7815130324',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '25848041822148',
                                                                                                                  'out_ucast_pkts': '18065445731',
                                                                                                                  'poller_hash': 'QkYyNEM2cGVlcmludGZwb24gMTUx'},
                                                                                                          {       'in_nucast_pkts': '30663605',
                                                                                                                  'in_octets': '2505669047856',
                                                                                                                  'in_ucast_pkts': '3496947896',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22489460',
                                                                                                                  'out_octets': '2070730835192',
                                                                                                                  'out_ucast_pkts': '3806536412',
                                                                                                                  'poller_hash': 'RjQzMXBlZXJpbnRmcG9uIDE1NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1520359047999',
                                                                                                                  'in_ucast_pkts': '2221170554',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '20617643973',
                                                                                                                  'out_octets': '33059394065835',
                                                                                                                  'out_ucast_pkts': '4225646687',
                                                                                                                  'poller_hash': 'ODY5MzY0cGVlcmludGZwb24gMjc4'},
                                                                                                          {       'in_nucast_pkts': '13',
                                                                                                                  'in_octets': '4966053618',
                                                                                                                  'in_ucast_pkts': '28926971',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6776688168',
                                                                                                                  'out_octets': '9124860230995',
                                                                                                                  'out_ucast_pkts': '43992822',
                                                                                                                  'poller_hash': 'NDJEMEM2cGVlcmludGZwb24gMzk2'},
                                                                                                          {       'in_nucast_pkts': '12932',
                                                                                                                  'in_octets': '2152365322',
                                                                                                                  'in_ucast_pkts': '11543084',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715414',
                                                                                                                  'out_octets': '40492887944',
                                                                                                                  'out_ucast_pkts': '32501933',
                                                                                                                  'poller_hash': 'NDJDRDkycGVlcmludGZwb24gNDEw'},
                                                                                                          {       'in_nucast_pkts': '21981',
                                                                                                                  'in_octets': '165103476318',
                                                                                                                  'in_ucast_pkts': '759747495',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19874760',
                                                                                                                  'out_octets': '2727374500396',
                                                                                                                  'out_ucast_pkts': '2033040096',
                                                                                                                  'poller_hash': 'NUE2QjkxcGVlcmludGZwb24gNDgy'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '6610508974',
                                                                                                                  'in_ucast_pkts': '36659849',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715457',
                                                                                                                  'out_octets': '184862946794',
                                                                                                                  'out_ucast_pkts': '131484156',
                                                                                                                  'poller_hash': 'NDJCRUQ1cGVlcmludGZwb24gNTI2'},
                                                                                                          {       'in_nucast_pkts': '13160',
                                                                                                                  'in_octets': '3912432542',
                                                                                                                  'in_ucast_pkts': '12249420',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715479',
                                                                                                                  'out_octets': '44982870288',
                                                                                                                  'out_ucast_pkts': '34313512',
                                                                                                                  'poller_hash': 'OTU2RjgycGVlcmludGZwb24gMTQ0OTc4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45896304270',
                                                                                                                  'in_ucast_pkts': '391906969',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '107',
                                                                                                                  'out_octets': '2583834022141',
                                                                                                                  'out_ucast_pkts': '5857676617',
                                                                                                                  'poller_hash': 'QjBBMjlGcGVlcmludGZwb24gMTQ5NjIx'},
                                                                                                          {       'in_nucast_pkts': '30711125',
                                                                                                                  'in_octets': '415956009613',
                                                                                                                  'in_ucast_pkts': '1748540448',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '68425976006',
                                                                                                                  'out_octets': '99755122707540',
                                                                                                                  'out_ucast_pkts': '5934949459',
                                                                                                                  'poller_hash': 'QzgzQzFCcGVlcmludGZwb24gMTk0MTI0'},
                                                                                                          {       'in_nucast_pkts': '10647',
                                                                                                                  'in_octets': '19860663988',
                                                                                                                  'in_ucast_pkts': '85409837',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715489',
                                                                                                                  'out_octets': '268837080250',
                                                                                                                  'out_ucast_pkts': '237178860',
                                                                                                                  'poller_hash': 'QzgxREU0cGVlcmludGZwb24gMjEwMDI2'},
                                                                                                          {       'in_nucast_pkts': '9572',
                                                                                                                  'in_octets': '11650831054',
                                                                                                                  'in_ucast_pkts': '47490585',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '536881',
                                                                                                                  'out_octets': '291248971664',
                                                                                                                  'out_ucast_pkts': '233648791',
                                                                                                                  'poller_hash': 'MTI3NTE0OHBlZXJpbnRmcG9uIDIyMzkzMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96078903339',
                                                                                                                  'in_ucast_pkts': '986579035',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '134',
                                                                                                                  'out_octets': '6062122637425',
                                                                                                                  'out_ucast_pkts': '4046278007',
                                                                                                                  'poller_hash': 'MTI3NDJCMHBlZXJpbnRmcG9uIDI0MjEzNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '247397203422',
                                                                                                                  'in_ucast_pkts': '1862054265',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '7847577947118',
                                                                                                                  'out_ucast_pkts': '5375807537',
                                                                                                                  'poller_hash': 'MTI3NEExQ3BlZXJpbnRmcG9uIDI0NTg2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '122945701136',
                                                                                                                  'in_ucast_pkts': '881008443',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3024469975679',
                                                                                                                  'out_ucast_pkts': '2134145664',
                                                                                                                  'poller_hash': 'MTI5NUU0RnBlZXJpbnRmcG9uIDI3Njk4Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '49789123368',
                                                                                                                  'in_ucast_pkts': '219138111',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1038238415146',
                                                                                                                  'out_ucast_pkts': '743021798',
                                                                                                                  'poller_hash': 'MTI5NUZEOXBlZXJpbnRmcG9uIDI3OTA1Nw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '174679028435',
                                                                                         'out_octets': '528419824015801',
                                                                                         'out_ucast_pkts': '214899644742',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/3',
                                                                                                                 'id': '5/3',
                                                                                                                 'links': [       {       'connected_time': '70857444',
                                                                                                                                          'fiber_length': 19899,
                                                                                                                                          'poller_hash': 'QzgxRDUwcGVlcnBvbnBvbiAxMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611449',
                                                                                                                                          'fiber_length': 19772,
                                                                                                                                          'poller_hash': 'NDJDODFFcGVlcnBvbnBvbiAxMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '639271',
                                                                                                                                          'fiber_length': 19676,
                                                                                                                                          'poller_hash': 'NEQxMTE5cGVlcnBvbnBvbiAxMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '11965976',
                                                                                                                                          'fiber_length': 19304,
                                                                                                                                          'poller_hash': 'RjQzNnBlZXJwb25wb24gMTM5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '11965226',
                                                                                                                                          'fiber_length': 19472,
                                                                                                                                          'poller_hash': 'M0NGQUVwZWVycG9ucG9uIDE0MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611451',
                                                                                                                                          'fiber_length': 20166,
                                                                                                                                          'poller_hash': 'NDJDNjk3cGVlcnBvbnBvbiAxNDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '782087',
                                                                                                                                          'fiber_length': 19347,
                                                                                                                                          'poller_hash': 'NzFBNEM0cGVlcnBvbnBvbiAxNDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '15842329',
                                                                                                                                          'fiber_length': 19846,
                                                                                                                                          'poller_hash': 'MjQ4QUJEcGVlcnBvbnBvbiAxNDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '11965030',
                                                                                                                                          'fiber_length': 20026,
                                                                                                                                          'poller_hash': 'RjBGNXBlZXJwb25wb24gMTUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '72549051',
                                                                                                                                          'fiber_length': 20194,
                                                                                                                                          'poller_hash': 'QkYyNEM2cGVlcnBvbnBvbiAxNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '11966034',
                                                                                                                                          'fiber_length': 19393,
                                                                                                                                          'poller_hash': 'RjQzMXBlZXJwb25wb24gMTU1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611454',
                                                                                                                                          'fiber_length': 19727,
                                                                                                                                          'poller_hash': 'ODY5MzY0cGVlcnBvbnBvbiAyNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611455',
                                                                                                                                          'fiber_length': 19728,
                                                                                                                                          'poller_hash': 'NDJEMEM2cGVlcnBvbnBvbiAzOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611447',
                                                                                                                                          'fiber_length': 19726,
                                                                                                                                          'poller_hash': 'NDJDRDkycGVlcnBvbnBvbiA0MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '17172868',
                                                                                                                                          'fiber_length': 19954,
                                                                                                                                          'poller_hash': 'NUE2QjkxcGVlcnBvbnBvbiA0ODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611454',
                                                                                                                                          'fiber_length': 19784,
                                                                                                                                          'poller_hash': 'NDJCRUQ1cGVlcnBvbnBvbiA1MjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611454',
                                                                                                                                          'fiber_length': 20037,
                                                                                                                                          'poller_hash': 'OTU2RjgycGVlcnBvbnBvbiAxNDQ5Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '12866516',
                                                                                                                                          'fiber_length': 20181,
                                                                                                                                          'poller_hash': 'QjBBMjlGcGVlcnBvbnBvbiAxNDk2MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '21717948',
                                                                                                                                          'fiber_length': 19434,
                                                                                                                                          'poller_hash': 'QzgzQzFCcGVlcnBvbnBvbiAxOTQxMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611459',
                                                                                                                                          'fiber_length': 20008,
                                                                                                                                          'poller_hash': 'QzgxREU0cGVlcnBvbnBvbiAyMTAwMjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611457',
                                                                                                                                          'fiber_length': 19745,
                                                                                                                                          'poller_hash': 'MTI3NTE0OHBlZXJwb25wb24gMjIzOTMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '71352148',
                                                                                                                                          'fiber_length': 20024,
                                                                                                                                          'poller_hash': 'MTI3NDJCMHBlZXJwb25wb24gMjQyMTM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611452',
                                                                                                                                          'fiber_length': 19711,
                                                                                                                                          'poller_hash': 'MTI3NEExQ3BlZXJwb25wb24gMjQ1ODY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '611456',
                                                                                                                                          'fiber_length': 20042,
                                                                                                                                          'poller_hash': 'MTI5NUU0RnBlZXJwb25wb24gMjc2OTg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799},
                                                                                                                                  {       'connected_time': '782079',
                                                                                                                                          'fiber_length': 19899,
                                                                                                                                          'poller_hash': 'MTI5NUZEOXBlZXJwb25wb24gMjc5MDU3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.799}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/3',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/4',
                                                                                         'in_discards': '28930352',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '32286229',
                                                                                         'in_octets': '113289646319612',
                                                                                         'in_ucast_pkts': '179273123846',
                                                                                         'links': [       {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '5428252594',
                                                                                                                  'in_ucast_pkts': '40304691',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394160',
                                                                                                                  'out_octets': '140668620094',
                                                                                                                  'out_ucast_pkts': '101032236',
                                                                                                                  'poller_hash': 'MzIxMUZCcGVlcmludGZwb24gMTk='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '83769762183',
                                                                                                                  'in_ucast_pkts': '548823285',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144',
                                                                                                                  'out_octets': '2300852304494',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzIxMERBcGVlcmludGZwb24gMjA='},
                                                                                                          {       'in_nucast_pkts': '3163',
                                                                                                                  'in_octets': '5046143765',
                                                                                                                  'in_ucast_pkts': '25659186',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394850',
                                                                                                                  'out_octets': '85396601164',
                                                                                                                  'out_ucast_pkts': '66346822',
                                                                                                                  'poller_hash': 'MzAwRTg2cGVlcmludGZwb24gMjE='},
                                                                                                          {       'in_nucast_pkts': '12864',
                                                                                                                  'in_octets': '9550878436',
                                                                                                                  'in_ucast_pkts': '39147914',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1225367',
                                                                                                                  'out_octets': '295572088761',
                                                                                                                  'out_ucast_pkts': '211569415',
                                                                                                                  'poller_hash': 'MzIxMEMycGVlcmludGZwb24gMjI='},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '2816633702',
                                                                                                                  'in_ucast_pkts': '16937606',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394171',
                                                                                                                  'out_octets': '63756685341',
                                                                                                                  'out_ucast_pkts': '46407077',
                                                                                                                  'poller_hash': 'NDJDOEFFcGVlcmludGZwb24gMjM='},
                                                                                                          {       'in_nucast_pkts': '226',
                                                                                                                  'in_octets': '2572587948',
                                                                                                                  'in_ucast_pkts': '7336541',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '161538',
                                                                                                                  'out_octets': '31721317878',
                                                                                                                  'out_ucast_pkts': '25214153',
                                                                                                                  'poller_hash': 'NDNDNDQ1cGVlcmludGZwb24gMjU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2462364261',
                                                                                                                  'in_ucast_pkts': '14831121',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17784466034',
                                                                                                                  'out_octets': '23594900507310',
                                                                                                                  'out_ucast_pkts': '15235034',
                                                                                                                  'poller_hash': 'MzIxNzA2cGVlcmludGZwb24gMjY='},
                                                                                                          {       'in_nucast_pkts': '1848',
                                                                                                                  'in_octets': '626862779058',
                                                                                                                  'in_ucast_pkts': '2215147263',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1603704',
                                                                                                                  'out_octets': '2624531316107',
                                                                                                                  'out_ucast_pkts': '4409498187',
                                                                                                                  'poller_hash': 'MzVFOTJDcGVlcmludGZwb24gMjc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '121809989253',
                                                                                                                  'in_ucast_pkts': '964999398',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144',
                                                                                                                  'out_octets': '3736476853982',
                                                                                                                  'out_ucast_pkts': '2690125197',
                                                                                                                  'poller_hash': 'MzIxMEM0cGVlcmludGZwb24gMjg='},
                                                                                                          {       'in_nucast_pkts': '7624',
                                                                                                                  'in_octets': '1043579514',
                                                                                                                  'in_ucast_pkts': '4179956',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394147',
                                                                                                                  'out_octets': '22456633619',
                                                                                                                  'out_ucast_pkts': '18803708',
                                                                                                                  'poller_hash': 'MzAwRDgxcGVlcmludGZwb24gMjk='},
                                                                                                          {       'in_nucast_pkts': '1612',
                                                                                                                  'in_octets': '1338475028',
                                                                                                                  'in_ucast_pkts': '8009403',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394168',
                                                                                                                  'out_octets': '31798162412',
                                                                                                                  'out_ucast_pkts': '23383513',
                                                                                                                  'poller_hash': 'MzAwRUZFcGVlcmludGZwb24gMzE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '136033870880',
                                                                                                                  'in_ucast_pkts': '727078052',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63',
                                                                                                                  'out_octets': '2891304678183',
                                                                                                                  'out_ucast_pkts': '2071443548',
                                                                                                                  'poller_hash': 'MjA1REQ2cGVlcmludGZwb24gMzI='},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '5714610919',
                                                                                                                  'in_ucast_pkts': '47371431',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394163',
                                                                                                                  'out_octets': '148101646356',
                                                                                                                  'out_ucast_pkts': '104917762',
                                                                                                                  'poller_hash': 'MzAwRUQzcGVlcmludGZwb24gMzM='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '355101486340',
                                                                                                                  'in_ucast_pkts': '538438420',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '107',
                                                                                                                  'out_octets': '2749409534493',
                                                                                                                  'out_ucast_pkts': '2195470762',
                                                                                                                  'poller_hash': 'MjA1REU1cGVlcmludGZwb24gMzQ='},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '3671492189',
                                                                                                                  'in_ucast_pkts': '21408494',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394158',
                                                                                                                  'out_octets': '98761079668',
                                                                                                                  'out_ucast_pkts': '68705410',
                                                                                                                  'poller_hash': 'NUE0N0YzcGVlcmludGZwb24gMzU='},
                                                                                                          {       'in_nucast_pkts': '5994',
                                                                                                                  'in_octets': '2808661113',
                                                                                                                  'in_ucast_pkts': '25260699',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394173',
                                                                                                                  'out_octets': '171188545072',
                                                                                                                  'out_ucast_pkts': '116547015',
                                                                                                                  'poller_hash': 'MzIxMjc4cGVlcmludGZwb24gMzc='},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '12691611695',
                                                                                                                  'in_ucast_pkts': '107658779',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394165',
                                                                                                                  'out_octets': '172278792600',
                                                                                                                  'out_ucast_pkts': '137900114',
                                                                                                                  'poller_hash': 'MzAwRUREcGVlcmludGZwb24gMzg='},
                                                                                                          {       'in_nucast_pkts': '141',
                                                                                                                  'in_octets': '765903816433',
                                                                                                                  'in_ucast_pkts': '1112933210',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11683981',
                                                                                                                  'out_octets': '2427758842636',
                                                                                                                  'out_ucast_pkts': '1958161678',
                                                                                                                  'poller_hash': 'MzVGOUQ3cGVlcmludGZwb24gMzk='},
                                                                                                          {       'in_nucast_pkts': '8580',
                                                                                                                  'in_octets': '2815073677438',
                                                                                                                  'in_ucast_pkts': '7916983952',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58895848',
                                                                                                                  'out_octets': '29316733848579',
                                                                                                                  'out_ucast_pkts': '21994817961',
                                                                                                                  'poller_hash': 'MzIxMzAzcGVlcmludGZwb24gNDA='},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '7385737658',
                                                                                                                  'in_ucast_pkts': '28934125',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394170',
                                                                                                                  'out_octets': '111625256325',
                                                                                                                  'out_ucast_pkts': '85349228',
                                                                                                                  'poller_hash': 'M0I2MUU1cGVlcmludGZwb24gNDE='},
                                                                                                          {       'in_nucast_pkts': '13083',
                                                                                                                  'in_octets': '1303428109638',
                                                                                                                  'in_ucast_pkts': '4485084075',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58901435',
                                                                                                                  'out_octets': '22243222925148',
                                                                                                                  'out_ucast_pkts': '16164417988',
                                                                                                                  'poller_hash': 'MkIwMzhBcGVlcmludGZwb24gNDQ='},
                                                                                                          {       'in_nucast_pkts': '8452988',
                                                                                                                  'in_octets': '29173026413535',
                                                                                                                  'in_ucast_pkts': '24020824783',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19109204',
                                                                                                                  'out_octets': '13971915569004',
                                                                                                                  'out_ucast_pkts': '20574354817',
                                                                                                                  'poller_hash': 'OTU2Rjg0cGVlcmludGZwb24gNjA='},
                                                                                                          {       'in_nucast_pkts': '30635',
                                                                                                                  'in_octets': '7878366543',
                                                                                                                  'in_ucast_pkts': '26421245',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394386',
                                                                                                                  'out_octets': '87155509900',
                                                                                                                  'out_ucast_pkts': '61604016',
                                                                                                                  'poller_hash': 'MzAwNzI4cGVlcmludGZwb24gNzE='},
                                                                                                          {       'in_nucast_pkts': '1646',
                                                                                                                  'in_octets': '63930674781',
                                                                                                                  'in_ucast_pkts': '97322056',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '917905',
                                                                                                                  'out_octets': '105028091895',
                                                                                                                  'out_ucast_pkts': '125032055',
                                                                                                                  'poller_hash': 'NDRBQUI2cGVlcmludGZwb24gMTcw'},
                                                                                                          {       'in_nucast_pkts': '3104',
                                                                                                                  'in_octets': '11962536215',
                                                                                                                  'in_ucast_pkts': '41588291',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '199736',
                                                                                                                  'out_octets': '185856584590',
                                                                                                                  'out_ucast_pkts': '134999405',
                                                                                                                  'poller_hash': 'NDJDNkU3cGVlcmludGZwb24gMTc5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '177076283056',
                                                                                                                  'in_ucast_pkts': '506066100',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '60',
                                                                                                                  'out_octets': '1234430340024',
                                                                                                                  'out_ucast_pkts': '1104603721',
                                                                                                                  'poller_hash': 'NDJEMjJGcGVlcmludGZwb24gMzk5'},
                                                                                                          {       'in_nucast_pkts': '418144',
                                                                                                                  'in_octets': '852766277944',
                                                                                                                  'in_ucast_pkts': '935173457',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6900840',
                                                                                                                  'out_octets': '2214694334799',
                                                                                                                  'out_ucast_pkts': '1856618422',
                                                                                                                  'poller_hash': 'NUE2NjdDcGVlcmludGZwb24gMTUwNTI4'},
                                                                                                          {       'in_nucast_pkts': '2953',
                                                                                                                  'in_octets': '2176107877',
                                                                                                                  'in_ucast_pkts': '14202127',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '213720',
                                                                                                                  'out_octets': '118992895322',
                                                                                                                  'out_ucast_pkts': '81858407',
                                                                                                                  'poller_hash': 'QzgzQjkycGVlcmludGZwb24gMTYzNjQz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '456468458995',
                                                                                                                  'in_ucast_pkts': '22832363227',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37',
                                                                                                                  'out_octets': '7126110654974',
                                                                                                                  'out_ucast_pkts': '5097479738',
                                                                                                                  'poller_hash': 'MTE5NTEyOXBlZXJpbnRmcG9uIDIxNjk3NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '113505885007',
                                                                                                                  'in_ucast_pkts': '64753346662',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '666129095175',
                                                                                                                  'out_ucast_pkts': '675605440',
                                                                                                                  'poller_hash': 'QTlDOEZFcGVlcmludGZwb24gMjc0Nzc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '43300221294',
                                                                                                                  'in_ucast_pkts': '372084931',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2150213528923',
                                                                                                                  'out_ucast_pkts': '1449025911',
                                                                                                                  'poller_hash': 'MTI5NUZFRXBlZXJpbnRmcG9uIDI3ODM1OA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '208369431135',
                                                                                         'out_octets': '685621268523905',
                                                                                         'out_ucast_pkts': '372966756764',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/4',
                                                                                                                 'id': '5/4',
                                                                                                                 'links': [       {       'connected_time': '335332',
                                                                                                                                          'fiber_length': 17678,
                                                                                                                                          'poller_hash': 'MzIxMUZCcGVlcnBvbnBvbiAxOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '19961094',
                                                                                                                                          'fiber_length': 17714,
                                                                                                                                          'poller_hash': 'MzIxMERBcGVlcnBvbnBvbiAyMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335330',
                                                                                                                                          'fiber_length': 17920,
                                                                                                                                          'poller_hash': 'MzAwRTg2cGVlcnBvbnBvbiAyMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '1050060',
                                                                                                                                          'fiber_length': 17129,
                                                                                                                                          'poller_hash': 'MzIxMEMycGVlcnBvbnBvbiAyMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335330',
                                                                                                                                          'fiber_length': 17518,
                                                                                                                                          'poller_hash': 'NDJDOEFFcGVlcnBvbnBvbiAyMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335329',
                                                                                                                                          'fiber_length': 17275,
                                                                                                                                          'poller_hash': 'NDNDNDQ1cGVlcnBvbnBvbiAyNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '21886983',
                                                                                                                                          'fiber_length': 17689,
                                                                                                                                          'poller_hash': 'MzIxNzA2cGVlcnBvbnBvbiAyNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '1371569',
                                                                                                                                          'fiber_length': 22188,
                                                                                                                                          'poller_hash': 'MzVFOTJDcGVlcnBvbnBvbiAyNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '8535127',
                                                                                                                                          'fiber_length': 17161,
                                                                                                                                          'poller_hash': 'MzIxMEM0cGVlcnBvbnBvbiAyOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335329',
                                                                                                                                          'fiber_length': 18002,
                                                                                                                                          'poller_hash': 'MzAwRDgxcGVlcnBvbnBvbiAyOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335330',
                                                                                                                                          'fiber_length': 17856,
                                                                                                                                          'poller_hash': 'MzAwRUZFcGVlcnBvbnBvbiAzMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335334',
                                                                                                                                          'fiber_length': 18033,
                                                                                                                                          'poller_hash': 'MjA1REQ2cGVlcnBvbnBvbiAzMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335331',
                                                                                                                                          'fiber_length': 17657,
                                                                                                                                          'poller_hash': 'MzAwRUQzcGVlcnBvbnBvbiAzMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335331',
                                                                                                                                          'fiber_length': 17110,
                                                                                                                                          'poller_hash': 'MjA1REU1cGVlcnBvbnBvbiAzNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335337',
                                                                                                                                          'fiber_length': 17358,
                                                                                                                                          'poller_hash': 'NUE0N0YzcGVlcnBvbnBvbiAzNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335333',
                                                                                                                                          'fiber_length': 18071,
                                                                                                                                          'poller_hash': 'MzIxMjc4cGVlcnBvbnBvbiAzNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335329',
                                                                                                                                          'fiber_length': 17020,
                                                                                                                                          'poller_hash': 'MzAwRUREcGVlcnBvbnBvbiAzOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '10094877',
                                                                                                                                          'fiber_length': 17762,
                                                                                                                                          'poller_hash': 'MzVGOUQ3cGVlcnBvbnBvbiAzOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '21886797',
                                                                                                                                          'fiber_length': 18017,
                                                                                                                                          'poller_hash': 'MzIxMzAzcGVlcnBvbnBvbiA0MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335333',
                                                                                                                                          'fiber_length': 17175,
                                                                                                                                          'poller_hash': 'M0I2MUU1cGVlcnBvbnBvbiA0MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '21886021',
                                                                                                                                          'fiber_length': 18049,
                                                                                                                                          'poller_hash': 'MkIwMzhBcGVlcnBvbnBvbiA0NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '16515557',
                                                                                                                                          'fiber_length': 21385,
                                                                                                                                          'poller_hash': 'OTU2Rjg0cGVlcnBvbnBvbiA2MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335330',
                                                                                                                                          'fiber_length': 17132,
                                                                                                                                          'poller_hash': 'MzAwNzI4cGVlcnBvbnBvbiA3MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '1372029',
                                                                                                                                          'fiber_length': 22072,
                                                                                                                                          'poller_hash': 'NDRBQUI2cGVlcnBvbnBvbiAxNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '169055',
                                                                                                                                          'fiber_length': 17375,
                                                                                                                                          'poller_hash': 'NDJDNkU3cGVlcnBvbnBvbiAxNzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '10763837',
                                                                                                                                          'fiber_length': 17877,
                                                                                                                                          'poller_hash': 'NDJEMjJGcGVlcnBvbnBvbiAzOTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '5921117',
                                                                                                                                          'fiber_length': 17857,
                                                                                                                                          'poller_hash': 'NUE2NjdDcGVlcnBvbnBvbiAxNTA1Mjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '180049',
                                                                                                                                          'fiber_length': 17839,
                                                                                                                                          'poller_hash': 'QzgzQjkycGVlcnBvbnBvbiAxNjM2NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '335336',
                                                                                                                                          'fiber_length': 17757,
                                                                                                                                          'poller_hash': 'MTE5NTEyOXBlZXJwb25wb24gMjE2OTc1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '8535295',
                                                                                                                                          'fiber_length': 21171,
                                                                                                                                          'poller_hash': 'QTlDOEZFcGVlcnBvbnBvbiAyNzQ3Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7},
                                                                                                                                  {       'connected_time': '8385735',
                                                                                                                                          'fiber_length': 17518,
                                                                                                                                          'poller_hash': 'MTI5NUZFRXBlZXJwb25wb24gMjc4MzU4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.7}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/4',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/5',
                                                                                         'in_discards': '150858',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5911319',
                                                                                         'in_octets': '16659077465362',
                                                                                         'in_ucast_pkts': '66253770607',
                                                                                         'links': [       {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '3583001168',
                                                                                                                  'in_ucast_pkts': '22861089',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394167',
                                                                                                                  'out_octets': '87165772110',
                                                                                                                  'out_ucast_pkts': '67620287',
                                                                                                                  'poller_hash': 'MkIwMThDcGVlcmludGZwb24gNA=='},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '2811906322',
                                                                                                                  'in_ucast_pkts': '26603221',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394166',
                                                                                                                  'out_octets': '71032092544',
                                                                                                                  'out_ucast_pkts': '60848582',
                                                                                                                  'poller_hash': 'MkIwMTg3cGVlcmludGZwb24gNQ=='},
                                                                                                          {       'in_nucast_pkts': '11033',
                                                                                                                  'in_octets': '10879847126',
                                                                                                                  'in_ucast_pkts': '31205796',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '969328',
                                                                                                                  'out_octets': '124066402700',
                                                                                                                  'out_ucast_pkts': '93124808',
                                                                                                                  'poller_hash': 'MkIwNDc3cGVlcmludGZwb24gNg=='},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '6300264069',
                                                                                                                  'in_ucast_pkts': '25585140',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394159',
                                                                                                                  'out_octets': '137967810460',
                                                                                                                  'out_ucast_pkts': '98987474',
                                                                                                                  'poller_hash': 'MzAwRDkxcGVlcmludGZwb24gNw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '101019491987',
                                                                                                                  'in_ucast_pkts': '326141022',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '197',
                                                                                                                  'out_octets': '1727253755109',
                                                                                                                  'out_ucast_pkts': '1272936274',
                                                                                                                  'poller_hash': 'MkFEMTIxcGVlcmludGZwb24gOQ=='},
                                                                                                          {       'in_nucast_pkts': '131527',
                                                                                                                  'in_octets': '1756122670',
                                                                                                                  'in_ucast_pkts': '7927715',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '438091',
                                                                                                                  'out_octets': '70434327589',
                                                                                                                  'out_ucast_pkts': '49701672',
                                                                                                                  'poller_hash': 'NTc0NTgzcGVlcmludGZwb24gMTA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '128264019518',
                                                                                                                  'in_ucast_pkts': '268585539',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '144',
                                                                                                                  'out_octets': '1175404602398',
                                                                                                                  'out_ucast_pkts': '963109639',
                                                                                                                  'poller_hash': 'MkFEMEZEcGVlcmludGZwb24gMTE='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MjA1REJBcGVlcmludGZwb24gMTM='},
                                                                                                          {       'in_nucast_pkts': '6728',
                                                                                                                  'in_octets': '2495422223',
                                                                                                                  'in_ucast_pkts': '15824239',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394140',
                                                                                                                  'out_octets': '97163421130',
                                                                                                                  'out_ucast_pkts': '69173496',
                                                                                                                  'poller_hash': 'MzAwN0JDcGVlcmludGZwb24gMTQ='},
                                                                                                          {       'in_nucast_pkts': '80300',
                                                                                                                  'in_octets': '90136103397',
                                                                                                                  'in_ucast_pkts': '595206266',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22546206',
                                                                                                                  'out_octets': '1964802213912',
                                                                                                                  'out_ucast_pkts': '1515954631',
                                                                                                                  'poller_hash': 'MzVGOUE2cGVlcmludGZwb24gMTU='},
                                                                                                          {       'in_nucast_pkts': '8317',
                                                                                                                  'in_octets': '153565240197',
                                                                                                                  'in_ucast_pkts': '460235051',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58877786',
                                                                                                                  'out_octets': '1430454393482',
                                                                                                                  'out_ucast_pkts': '1138812801',
                                                                                                                  'poller_hash': 'MzIxMzAxcGVlcmludGZwb24gMTY='},
                                                                                                          {       'in_nucast_pkts': '113',
                                                                                                                  'in_octets': '120074434937',
                                                                                                                  'in_ucast_pkts': '861214614',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '16616138',
                                                                                                                  'out_octets': '4289592711639',
                                                                                                                  'out_ucast_pkts': '3187698650',
                                                                                                                  'poller_hash': 'MzIxMEZCcGVlcmludGZwb24gMTc='},
                                                                                                          {       'in_nucast_pkts': '46',
                                                                                                                  'in_octets': '38407667396',
                                                                                                                  'in_ucast_pkts': '66683387',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '916121',
                                                                                                                  'out_octets': '115151047008',
                                                                                                                  'out_ucast_pkts': '101912491',
                                                                                                                  'poller_hash': 'M0I2MTI4cGVlcmludGZwb24gMTg='},
                                                                                                          {       'in_nucast_pkts': '14909',
                                                                                                                  'in_octets': '6848664887',
                                                                                                                  'in_ucast_pkts': '64676349',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '912792',
                                                                                                                  'out_octets': '333222240438',
                                                                                                                  'out_ucast_pkts': '224823375',
                                                                                                                  'poller_hash': 'MkFEMEM5cGVlcmludGZwb24gNDM='},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '729654976',
                                                                                                                  'in_ucast_pkts': '4830220',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394167',
                                                                                                                  'out_octets': '34525886895',
                                                                                                                  'out_ucast_pkts': '25191438',
                                                                                                                  'poller_hash': 'MzAwRThEcGVlcmludGZwb24gNDU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1058441259552',
                                                                                                                  'in_ucast_pkts': '1730330380',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '90',
                                                                                                                  'out_octets': '3264879733680',
                                                                                                                  'out_ucast_pkts': '3098020970',
                                                                                                                  'poller_hash': 'ODcyNTFGcGVlcmludGZwb24gNDY='},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '7221675814',
                                                                                                                  'in_ucast_pkts': '61108205',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7093901479',
                                                                                                                  'out_octets': '9799877089627',
                                                                                                                  'out_ucast_pkts': '169056819',
                                                                                                                  'poller_hash': 'MkFEMEQ3cGVlcmludGZwb24gNDc='},
                                                                                                          {       'in_nucast_pkts': '24',
                                                                                                                  'in_octets': '1269891017',
                                                                                                                  'in_ucast_pkts': '3718687',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394166',
                                                                                                                  'out_octets': '12082573167',
                                                                                                                  'out_ucast_pkts': '10444064',
                                                                                                                  'poller_hash': 'MzAwREQzcGVlcmludGZwb24gNDg='},
                                                                                                          {       'in_nucast_pkts': '12195',
                                                                                                                  'in_octets': '839381496',
                                                                                                                  'in_ucast_pkts': '5222318',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394366',
                                                                                                                  'out_octets': '25637680047',
                                                                                                                  'out_ucast_pkts': '17944552',
                                                                                                                  'poller_hash': 'MzAwRUEwcGVlcmludGZwb24gNDk='},
                                                                                                          {       'in_nucast_pkts': '82732',
                                                                                                                  'in_octets': '135334387416',
                                                                                                                  'in_ucast_pkts': '275983459',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4945372',
                                                                                                                  'out_octets': '714454014035',
                                                                                                                  'out_ucast_pkts': '662075810',
                                                                                                                  'poller_hash': 'MzIxMTVGcGVlcmludGZwb24gNTA='},
                                                                                                          {       'in_nucast_pkts': '8529',
                                                                                                                  'in_octets': '844743337963',
                                                                                                                  'in_ucast_pkts': '5737944460',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58901429',
                                                                                                                  'out_octets': '24931331451789',
                                                                                                                  'out_ucast_pkts': '17398577981',
                                                                                                                  'poller_hash': 'MzIxMTQ4cGVlcmludGZwb24gNTE='},
                                                                                                          {       'in_nucast_pkts': '258296',
                                                                                                                  'in_octets': '818547759766',
                                                                                                                  'in_ucast_pkts': '1968793687',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '41937474210',
                                                                                                                  'out_octets': '63058713573160',
                                                                                                                  'out_ucast_pkts': '5126203198',
                                                                                                                  'poller_hash': 'MzIxMjI2cGVlcmludGZwb24gNTI='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '251266900404',
                                                                                                                  'in_ucast_pkts': '1518680266',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15716324240',
                                                                                                                  'out_octets': '25893278806561',
                                                                                                                  'out_ucast_pkts': '3671440270',
                                                                                                                  'poller_hash': 'MzIxMTM4cGVlcmludGZwb24gNTM='},
                                                                                                          {       'in_nucast_pkts': '6210',
                                                                                                                  'in_octets': '5379421976',
                                                                                                                  'in_ucast_pkts': '21465088',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394165',
                                                                                                                  'out_octets': '86739811594',
                                                                                                                  'out_ucast_pkts': '62238211',
                                                                                                                  'poller_hash': 'MzIxMTkwcGVlcmludGZwb24gNTQ='},
                                                                                                          {       'in_nucast_pkts': '132',
                                                                                                                  'in_octets': '103583470052',
                                                                                                                  'in_ucast_pkts': '235215430',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '27648248',
                                                                                                                  'out_octets': '590102497405',
                                                                                                                  'out_ucast_pkts': '512780840',
                                                                                                                  'poller_hash': 'MzIxMTRDcGVlcmludGZwb24gNTU='},
                                                                                                          {       'in_nucast_pkts': '106227',
                                                                                                                  'in_octets': '27532153016',
                                                                                                                  'in_ucast_pkts': '162940047',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7099079',
                                                                                                                  'out_octets': '955255941446',
                                                                                                                  'out_ucast_pkts': '681646381',
                                                                                                                  'poller_hash': 'MzIxMTZGcGVlcmludGZwb24gNTY='},
                                                                                                          {       'in_nucast_pkts': '6454882',
                                                                                                                  'in_octets': '482855941819',
                                                                                                                  'in_ucast_pkts': '1050534021',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5516774313',
                                                                                                                  'out_octets': '10633040437161',
                                                                                                                  'out_ucast_pkts': '2957806147',
                                                                                                                  'poller_hash': 'MzIxMTcycGVlcmludGZwb24gNTc='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '206678523392',
                                                                                                                  'in_ucast_pkts': '1846654465',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '197',
                                                                                                                  'out_octets': '10996326448635',
                                                                                                                  'out_ucast_pkts': '7464424405',
                                                                                                                  'poller_hash': 'MzIxMTA4cGVlcmludGZwb24gNTg='},
                                                                                                          {       'in_nucast_pkts': '73',
                                                                                                                  'in_octets': '26616783',
                                                                                                                  'in_ucast_pkts': '216680',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394170',
                                                                                                                  'out_octets': '87454570',
                                                                                                                  'out_ucast_pkts': '243861',
                                                                                                                  'poller_hash': 'M0I1QURFcGVlcmludGZwb24gNTk='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '225182649028',
                                                                                                                  'in_ucast_pkts': '1966696189',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '79',
                                                                                                                  'out_octets': '6008347335132',
                                                                                                                  'out_ucast_pkts': '4192560610',
                                                                                                                  'poller_hash': 'QzgxQzc4cGVlcmludGZwb24gMjA1Nzcz'},
                                                                                                          {       'in_nucast_pkts': '674564',
                                                                                                                  'in_octets': '21114128215',
                                                                                                                  'in_ucast_pkts': '49347842',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38757385',
                                                                                                                  'out_octets': '82328627875',
                                                                                                                  'out_ucast_pkts': '92105554',
                                                                                                                  'poller_hash': 'MkFEMTBBcGVlcmludGZwb24gMjE1ODk1'},
                                                                                                          {       'in_nucast_pkts': '6060',
                                                                                                                  'in_octets': '7417714083',
                                                                                                                  'in_ucast_pkts': '15928577',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394324',
                                                                                                                  'out_octets': '79815742269',
                                                                                                                  'out_ucast_pkts': '61063874',
                                                                                                                  'poller_hash': 'MTI3NEE4MXBlZXJpbnRmcG9uIDIzMTU5MA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '347471868772',
                                                                                         'out_octets': '698007917061319',
                                                                                         'out_ucast_pkts': '175618404572',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/5',
                                                                                                                 'id': '5/5',
                                                                                                                 'links': [       {       'connected_time': '335336',
                                                                                                                                          'fiber_length': 16964,
                                                                                                                                          'poller_hash': 'MkIwMThDcGVlcnBvbnBvbiA0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335332',
                                                                                                                                          'fiber_length': 17181,
                                                                                                                                          'poller_hash': 'MkIwMTg3cGVlcnBvbnBvbiA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '598181',
                                                                                                                                          'fiber_length': 17009,
                                                                                                                                          'poller_hash': 'MkIwNDc3cGVlcnBvbnBvbiA2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335332',
                                                                                                                                          'fiber_length': 17157,
                                                                                                                                          'poller_hash': 'MzAwRDkxcGVlcnBvbnBvbiA3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '7443401',
                                                                                                                                          'fiber_length': 17193,
                                                                                                                                          'poller_hash': 'MkFEMTIxcGVlcnBvbnBvbiA5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335338',
                                                                                                                                          'fiber_length': 17405,
                                                                                                                                          'poller_hash': 'NTc0NTgzcGVlcnBvbnBvbiAxMA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '73026593',
                                                                                                                                          'fiber_length': 17336,
                                                                                                                                          'poller_hash': 'MkFEMEZEcGVlcnBvbnBvbiAxMQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '4650410',
                                                                                                                                          'fiber_length': 17572,
                                                                                                                                          'poller_hash': 'MjA1REJBcGVlcnBvbnBvbiAxMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335329',
                                                                                                                                          'fiber_length': 17202,
                                                                                                                                          'poller_hash': 'MzAwN0JDcGVlcnBvbnBvbiAxNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '19526429',
                                                                                                                                          'fiber_length': 17737,
                                                                                                                                          'poller_hash': 'MzVGOUE2cGVlcnBvbnBvbiAxNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '21886489',
                                                                                                                                          'fiber_length': 17533,
                                                                                                                                          'poller_hash': 'MzIxMzAxcGVlcnBvbnBvbiAxNg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '14334953',
                                                                                                                                          'fiber_length': 17754,
                                                                                                                                          'poller_hash': 'MzIxMEZCcGVlcnBvbnBvbiAxNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '782105',
                                                                                                                                          'fiber_length': 17136,
                                                                                                                                          'poller_hash': 'M0I2MTI4cGVlcnBvbnBvbiAxOA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '779562',
                                                                                                                                          'fiber_length': 17566,
                                                                                                                                          'poller_hash': 'MkFEMEM5cGVlcnBvbnBvbiA0Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335335',
                                                                                                                                          'fiber_length': 17396,
                                                                                                                                          'poller_hash': 'MzAwRThEcGVlcnBvbnBvbiA0NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335338',
                                                                                                                                          'fiber_length': 17099,
                                                                                                                                          'poller_hash': 'ODcyNTFGcGVlcnBvbnBvbiA0Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335337',
                                                                                                                                          'fiber_length': 17551,
                                                                                                                                          'poller_hash': 'MkFEMEQ3cGVlcnBvbnBvbiA0Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335335',
                                                                                                                                          'fiber_length': 17706,
                                                                                                                                          'poller_hash': 'MzAwREQzcGVlcnBvbnBvbiA0OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335334',
                                                                                                                                          'fiber_length': 17166,
                                                                                                                                          'poller_hash': 'MzAwRUEwcGVlcnBvbnBvbiA0OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '4218997',
                                                                                                                                          'fiber_length': 17701,
                                                                                                                                          'poller_hash': 'MzIxMTVGcGVlcnBvbnBvbiA1MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '21885988',
                                                                                                                                          'fiber_length': 17710,
                                                                                                                                          'poller_hash': 'MzIxMTQ4cGVlcnBvbnBvbiA1MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '21886170',
                                                                                                                                          'fiber_length': 17784,
                                                                                                                                          'poller_hash': 'MzIxMjI2cGVlcnBvbnBvbiA1Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '21886514',
                                                                                                                                          'fiber_length': 17498,
                                                                                                                                          'poller_hash': 'MzIxMTM4cGVlcnBvbnBvbiA1Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335336',
                                                                                                                                          'fiber_length': 17724,
                                                                                                                                          'poller_hash': 'MzIxMTkwcGVlcnBvbnBvbiA1NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '24896593',
                                                                                                                                          'fiber_length': 17234,
                                                                                                                                          'poller_hash': 'MzIxMTRDcGVlcnBvbnBvbiA1NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '7398406',
                                                                                                                                          'fiber_length': 17860,
                                                                                                                                          'poller_hash': 'MzIxMTZGcGVlcnBvbnBvbiA1Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '8384839',
                                                                                                                                          'fiber_length': 17626,
                                                                                                                                          'poller_hash': 'MzIxMTcycGVlcnBvbnBvbiA1Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335335',
                                                                                                                                          'fiber_length': 17149,
                                                                                                                                          'poller_hash': 'MzIxMTA4cGVlcnBvbnBvbiA1OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335336',
                                                                                                                                          'fiber_length': 17146,
                                                                                                                                          'poller_hash': 'M0I1QURFcGVlcnBvbnBvbiA1OQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335337',
                                                                                                                                          'fiber_length': 17508,
                                                                                                                                          'poller_hash': 'QzgxQzc4cGVlcnBvbnBvbiAyMDU3NzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '68775798',
                                                                                                                                          'fiber_length': 17507,
                                                                                                                                          'poller_hash': 'MkFEMTBBcGVlcnBvbnBvbiAyMTU4OTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767},
                                                                                                                                  {       'connected_time': '335336',
                                                                                                                                          'fiber_length': 17126,
                                                                                                                                          'poller_hash': 'MTI3NEE4MXBlZXJwb25wb24gMjMxNTkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.767}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/5',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/6',
                                                                                         'in_discards': '11006',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1883257',
                                                                                         'in_octets': '21596774939650',
                                                                                         'in_ucast_pkts': '50205783108',
                                                                                         'links': [       {       'in_nucast_pkts': '32795',
                                                                                                                  'in_octets': '6738665008241',
                                                                                                                  'in_ucast_pkts': '6954930528',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19103087',
                                                                                                                  'out_octets': '4538209192881',
                                                                                                                  'out_ucast_pkts': '6116501368',
                                                                                                                  'poller_hash': 'MjZBQzZDcGVlcmludGZwb24gNjE='},
                                                                                                          {       'in_nucast_pkts': '440',
                                                                                                                  'in_octets': '474118066278',
                                                                                                                  'in_ucast_pkts': '1100822703',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37965823',
                                                                                                                  'out_octets': '5287160512784',
                                                                                                                  'out_ucast_pkts': '4319478664',
                                                                                                                  'poller_hash': 'MzVGOUVEcGVlcmludGZwb24gNjM='},
                                                                                                          {       'in_nucast_pkts': '1470',
                                                                                                                  'in_octets': '150222209718',
                                                                                                                  'in_ucast_pkts': '887181915',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37928381',
                                                                                                                  'out_octets': '3561532359952',
                                                                                                                  'out_ucast_pkts': '2516767701',
                                                                                                                  'poller_hash': 'MzVFOTNBcGVlcmludGZwb24gNjU='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '585806984831',
                                                                                                                  'in_ucast_pkts': '138838003585',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6913875666',
                                                                                                                  'out_octets': '13523047597926',
                                                                                                                  'out_ucast_pkts': '3850063227',
                                                                                                                  'poller_hash': 'MzIxMEQ4cGVlcmludGZwb24gNjY='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '161023360506',
                                                                                                                  'in_ucast_pkts': '821212491',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4552173113606',
                                                                                                                  'out_ucast_pkts': '3412694003',
                                                                                                                  'poller_hash': 'ODQ1OEFDcGVlcmludGZwb24gNjc='},
                                                                                                          {       'in_nucast_pkts': '12005806',
                                                                                                                  'in_octets': '148617143181',
                                                                                                                  'in_ucast_pkts': '9346343743',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11516805720',
                                                                                                                  'out_octets': '18493231402649',
                                                                                                                  'out_ucast_pkts': '5458903433216',
                                                                                                                  'poller_hash': 'MTMwQTJwZWVyaW50ZnBvbiA2OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2066337506122',
                                                                                                                  'in_ucast_pkts': '2660860141',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7350362775930',
                                                                                                                  'out_ucast_pkts': '6072081251',
                                                                                                                  'poller_hash': 'MUM1QkEzcGVlcmludGZwb24gNzA='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTMwQURwZWVyaW50ZnBvbiA3Mg=='},
                                                                                                          {       'in_nucast_pkts': '7026',
                                                                                                                  'in_octets': '6616666640',
                                                                                                                  'in_ucast_pkts': '63105250',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '524584',
                                                                                                                  'out_octets': '271765576584',
                                                                                                                  'out_ucast_pkts': '191372284',
                                                                                                                  'poller_hash': 'RjMyOHBlZXJpbnRmcG9uIDcz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '279604597640',
                                                                                                                  'in_ucast_pkts': '1456097228',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '7428510553568',
                                                                                                                  'out_ucast_pkts': '5598281490',
                                                                                                                  'poller_hash': 'MkZFQTA5cGVlcmludGZwb24gNzQ='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '554305950457',
                                                                                                                  'in_ucast_pkts': '18589074820',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '11503261820844',
                                                                                                                  'out_ucast_pkts': '26930480423',
                                                                                                                  'poller_hash': 'MkVEOUNFcGVlcmludGZwb24gNzY='},
                                                                                                          {       'in_nucast_pkts': '96443',
                                                                                                                  'in_octets': '228845764487',
                                                                                                                  'in_ucast_pkts': '555905241',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '22465286',
                                                                                                                  'out_octets': '817618808412',
                                                                                                                  'out_ucast_pkts': '770345515',
                                                                                                                  'poller_hash': 'MTAwQTkzNHBlZXJpbnRmcG9uIDc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '310286584770',
                                                                                                                  'in_ucast_pkts': '1323830753',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '4600885394990',
                                                                                                                  'out_ucast_pkts': '3717440743',
                                                                                                                  'poller_hash': 'NzZCMDNEcGVlcmludGZwb24gNzg='},
                                                                                                          {       'in_nucast_pkts': '6729',
                                                                                                                  'in_octets': '568275362103',
                                                                                                                  'in_ucast_pkts': '957763640',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '58956450',
                                                                                                                  'out_octets': '5145598528907',
                                                                                                                  'out_ucast_pkts': '3729470169',
                                                                                                                  'poller_hash': 'MkIwMURBcGVlcmludGZwb24gODI='},
                                                                                                          {       'in_nucast_pkts': '12648',
                                                                                                                  'in_octets': '14628402724',
                                                                                                                  'in_ucast_pkts': '71591372',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '409718',
                                                                                                                  'out_octets': '224902052250',
                                                                                                                  'out_ucast_pkts': '187710089',
                                                                                                                  'poller_hash': 'NDNDMjI5cGVlcmludGZwb24gODM='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTEyNTQzcGVlcmludGZwb24gMjc0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45112092697',
                                                                                                                  'in_ucast_pkts': '136177023',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '117',
                                                                                                                  'out_octets': '444426574031',
                                                                                                                  'out_ucast_pkts': '402011697',
                                                                                                                  'poller_hash': 'NDJDRTM2cGVlcmludGZwb24gMzk3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '95918967921',
                                                                                                                  'in_ucast_pkts': '509508679',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14565507027',
                                                                                                                  'out_octets': '21622501022683',
                                                                                                                  'out_ucast_pkts': '1694838980',
                                                                                                                  'poller_hash': 'NzNBQzExcGVlcmludGZwb24gNTM4'},
                                                                                                          {       'in_nucast_pkts': '774',
                                                                                                                  'in_octets': '4786217950',
                                                                                                                  'in_ucast_pkts': '8663243',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715461',
                                                                                                                  'out_octets': '21356643371',
                                                                                                                  'out_ucast_pkts': '17496761',
                                                                                                                  'poller_hash': 'MkJENUIycGVlcmludGZwb24gMTE2MDIy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '26364584834',
                                                                                                                  'in_ucast_pkts': '213675042',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53',
                                                                                                                  'out_octets': '1012391049187',
                                                                                                                  'out_ucast_pkts': '751959371',
                                                                                                                  'poller_hash': 'QTlDMkFGcGVlcmludGZwb24gMTQ1ODUy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '244401041208',
                                                                                                                  'in_ucast_pkts': '441691511',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '136',
                                                                                                                  'out_octets': '7449992300546',
                                                                                                                  'out_ucast_pkts': '5836877918',
                                                                                                                  'poller_hash': 'QjBBMjhCcGVlcmludGZwb24gMTQ5NTAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '62268331893',
                                                                                                                  'in_ucast_pkts': '275381128',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '106',
                                                                                                                  'out_octets': '1911187515487',
                                                                                                                  'out_ucast_pkts': '1404163381',
                                                                                                                  'poller_hash': 'MzIxMjhFcGVlcmludGZwb24gMjA1MTMw'},
                                                                                                          {       'in_nucast_pkts': '240815',
                                                                                                                  'in_octets': '194360415892',
                                                                                                                  'in_ucast_pkts': '691031496',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17793542',
                                                                                                                  'out_octets': '2585665774200',
                                                                                                                  'out_ucast_pkts': '2058832485',
                                                                                                                  'poller_hash': 'QzgxRjE0cGVlcmludGZwb24gMjA1MTkw'},
                                                                                                          {       'in_nucast_pkts': '9729',
                                                                                                                  'in_octets': '30092396683',
                                                                                                                  'in_ucast_pkts': '36843945',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '715441',
                                                                                                                  'out_octets': '91459697650',
                                                                                                                  'out_ucast_pkts': '74609183',
                                                                                                                  'poller_hash': 'MTE5M0I5QnBlZXJpbnRmcG9uIDIyMTczMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '144496038553',
                                                                                                                  'in_ucast_pkts': '845833182',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '4106464704375',
                                                                                                                  'out_ucast_pkts': '2930755683',
                                                                                                                  'poller_hash': 'MTI3NEZDRnBlZXJpbnRmcG9uIDIyMzQ2Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '9889399558',
                                                                                                                  'in_ucast_pkts': '115502176',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '717242049405',
                                                                                                                  'out_ucast_pkts': '477766837',
                                                                                                                  'poller_hash': 'MTI3NTAwN3BlZXJpbnRmcG9uIDIyMzU2Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '814195273757',
                                                                                                                  'in_ucast_pkts': '1601192348',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '941471932974',
                                                                                                                  'out_ucast_pkts': '1708704776',
                                                                                                                  'poller_hash': 'MjNBMDBBcGVlcmludGZwb24gMjQxNjE0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '225088581',
                                                                                                                  'in_ucast_pkts': '974368',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '5222805121',
                                                                                                                  'out_ucast_pkts': '14618630',
                                                                                                                  'poller_hash': 'MTE3REI1RnBlZXJpbnRmcG9uIDI1NDc4NQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '45659829959',
                                                                                                                  'in_ucast_pkts': '8882063430',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26',
                                                                                                                  'out_octets': '1387475173529',
                                                                                                                  'out_ucast_pkts': '982262656',
                                                                                                                  'poller_hash': 'MTFCRTFFRXBlZXJpbnRmcG9uIDI1ODI3Mw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '48372488150',
                                                                                                                  'in_ucast_pkts': '48250192',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '72719485233',
                                                                                                                  'out_ucast_pkts': '71471276',
                                                                                                                  'poller_hash': 'MTI3NDgxNnBlZXJpbnRmcG9uIDI3NDUxNA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '74937254222',
                                                                                                                  'in_ucast_pkts': '633151254',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3110592237962',
                                                                                                                  'out_ucast_pkts': '2094930341',
                                                                                                                  'poller_hash': 'MTI3M0I0MnBlZXJpbnRmcG9uIDI3OTQxNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '184601471500',
                                                                                         'out_octets': '408506262652931',
                                                                                         'out_ucast_pkts': '124967566871',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/6',
                                                                                                                 'id': '5/6',
                                                                                                                 'links': [       {       'connected_time': '3036359',
                                                                                                                                          'fiber_length': 17665,
                                                                                                                                          'poller_hash': 'MjZBQzZDcGVlcnBvbnBvbiA2MQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '67856246',
                                                                                                                                          'fiber_length': 14339,
                                                                                                                                          'poller_hash': 'MzVGOUVEcGVlcnBvbnBvbiA2Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '67856448',
                                                                                                                                          'fiber_length': 14918,
                                                                                                                                          'poller_hash': 'MzVFOTNBcGVlcnBvbnBvbiA2NQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '21886975',
                                                                                                                                          'fiber_length': 14758,
                                                                                                                                          'poller_hash': 'MzIxMEQ4cGVlcnBvbnBvbiA2Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '639438',
                                                                                                                                          'fiber_length': 12520,
                                                                                                                                          'poller_hash': 'ODQ1OEFDcGVlcnBvbnBvbiA2Nw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '11964807',
                                                                                                                                          'fiber_length': 13226,
                                                                                                                                          'poller_hash': 'MTMwQTJwZWVycG9ucG9uIDY5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '611454',
                                                                                                                                          'fiber_length': 13539,
                                                                                                                                          'poller_hash': 'MUM1QkEzcGVlcnBvbnBvbiA3MA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '11965425',
                                                                                                                                          'fiber_length': 12286,
                                                                                                                                          'poller_hash': 'MTMwQURwZWVycG9ucG9uIDcy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '782079',
                                                                                                                                          'fiber_length': 11621,
                                                                                                                                          'poller_hash': 'RjMyOHBlZXJwb25wb24gNzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '639070',
                                                                                                                                          'fiber_length': 12264,
                                                                                                                                          'poller_hash': 'MkZFQTA5cGVlcnBvbnBvbiA3NA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '638647',
                                                                                                                                          'fiber_length': 12374,
                                                                                                                                          'poller_hash': 'MkVEOUNFcGVlcnBvbnBvbiA3Ng==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '552730',
                                                                                                                                          'fiber_length': 11760,
                                                                                                                                          'poller_hash': 'MTAwQTkzNHBlZXJwb25wb24gNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '639742',
                                                                                                                                          'fiber_length': 14067,
                                                                                                                                          'poller_hash': 'NzZCMDNEcGVlcnBvbnBvbiA3OA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '21885968',
                                                                                                                                          'fiber_length': 18203,
                                                                                                                                          'poller_hash': 'MkIwMURBcGVlcnBvbnBvbiA4Mg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '611458',
                                                                                                                                          'fiber_length': 17141,
                                                                                                                                          'poller_hash': 'NDNDMjI5cGVlcnBvbnBvbiA4Mw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '67856069',
                                                                                                                                          'fiber_length': 11859,
                                                                                                                                          'poller_hash': 'NTEyNTQzcGVlcnBvbnBvbiAyNzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '21886551',
                                                                                                                                          'fiber_length': 15734,
                                                                                                                                          'poller_hash': 'NDJDRTM2cGVlcnBvbnBvbiAzOTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '8535324',
                                                                                                                                          'fiber_length': 15730,
                                                                                                                                          'poller_hash': 'NzNBQzExcGVlcnBvbnBvbiA1Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '611462',
                                                                                                                                          'fiber_length': 11812,
                                                                                                                                          'poller_hash': 'MkJENUIycGVlcnBvbnBvbiAxMTYwMjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '611462',
                                                                                                                                          'fiber_length': 11880,
                                                                                                                                          'poller_hash': 'QTlDMkFGcGVlcnBvbnBvbiAxNDU4NTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '8462792',
                                                                                                                                          'fiber_length': 18754,
                                                                                                                                          'poller_hash': 'QjBBMjhCcGVlcnBvbnBvbiAxNDk1MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '21887149',
                                                                                                                                          'fiber_length': 11943,
                                                                                                                                          'poller_hash': 'MzIxMjhFcGVlcnBvbnBvbiAyMDUxMzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '15369455',
                                                                                                                                          'fiber_length': 13784,
                                                                                                                                          'poller_hash': 'QzgxRjE0cGVlcnBvbnBvbiAyMDUxOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '611474',
                                                                                                                                          'fiber_length': 18530,
                                                                                                                                          'poller_hash': 'MTE5M0I5QnBlZXJwb25wb24gMjIxNzMw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '68441994',
                                                                                                                                          'fiber_length': 17078,
                                                                                                                                          'poller_hash': 'MTI3NEZDRnBlZXJwb25wb24gMjIzNDYz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '6400538',
                                                                                                                                          'fiber_length': 17082,
                                                                                                                                          'poller_hash': 'MTI3NTAwN3BlZXJwb25wb24gMjIzNTY2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '72154142',
                                                                                                                                          'fiber_length': 14891,
                                                                                                                                          'poller_hash': 'MjNBMDBBcGVlcnBvbnBvbiAyNDE2MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '27382858',
                                                                                                                                          'fiber_length': 15504,
                                                                                                                                          'poller_hash': 'MTE3REI1RnBlZXJwb25wb24gMjU0Nzg1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '7833000',
                                                                                                                                          'fiber_length': 14283,
                                                                                                                                          'poller_hash': 'MTFCRTFFRXBlZXJwb25wb24gMjU4Mjcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '8535310',
                                                                                                                                          'fiber_length': 17474,
                                                                                                                                          'poller_hash': 'MTI3NDgxNnBlZXJwb25wb24gMjc0NTE0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598},
                                                                                                                                  {       'connected_time': '6043446',
                                                                                                                                          'fiber_length': 12400,
                                                                                                                                          'poller_hash': 'MTI3M0I0MnBlZXJwb25wb24gMjc5NDE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.598}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/6',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/7',
                                                                                         'in_discards': '264613',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5042329',
                                                                                         'in_octets': '16405605060521',
                                                                                         'in_ucast_pkts': '50414219403',
                                                                                         'links': [       {       'in_nucast_pkts': '7306',
                                                                                                                  'in_octets': '888275457',
                                                                                                                  'in_ucast_pkts': '2378254',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394166',
                                                                                                                  'out_octets': '28942931232',
                                                                                                                  'out_ucast_pkts': '21781686',
                                                                                                                  'poller_hash': 'Q0JEMTY1cGVlcmludGZwb24gMTA5'},
                                                                                                          {       'in_nucast_pkts': '2965',
                                                                                                                  'in_octets': '14295183843',
                                                                                                                  'in_ucast_pkts': '11570532',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '433401',
                                                                                                                  'out_octets': '6426091526',
                                                                                                                  'out_ucast_pkts': '7838387',
                                                                                                                  'poller_hash': 'ODY5MzkxcGVlcmludGZwb24gNDAw'},
                                                                                                          {       'in_nucast_pkts': '67054',
                                                                                                                  'in_octets': '56170455057',
                                                                                                                  'in_ucast_pkts': '72888821',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4043754',
                                                                                                                  'out_octets': '178326857515',
                                                                                                                  'out_ucast_pkts': '155673275',
                                                                                                                  'poller_hash': 'NDJEMUY5cGVlcmludGZwb24gNDAy'},
                                                                                                          {       'in_nucast_pkts': '111998',
                                                                                                                  'in_octets': '729713521839',
                                                                                                                  'in_ucast_pkts': '2589412157',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39607851',
                                                                                                                  'out_octets': '11460497878069',
                                                                                                                  'out_ucast_pkts': '8294976598',
                                                                                                                  'poller_hash': 'NDJCRjE1cGVlcmludGZwb24gNDQ2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '361125817747',
                                                                                                                  'in_ucast_pkts': '39608094675',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '139',
                                                                                                                  'out_octets': '3541624444641',
                                                                                                                  'out_ucast_pkts': '2725676103',
                                                                                                                  'poller_hash': 'NDJDNkJCcGVlcmludGZwb24gNDg1'},
                                                                                                          {       'in_nucast_pkts': '4354',
                                                                                                                  'in_octets': '418466235345',
                                                                                                                  'in_ucast_pkts': '1140003398',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37971536',
                                                                                                                  'out_octets': '3678948103827',
                                                                                                                  'out_ucast_pkts': '2975951431',
                                                                                                                  'poller_hash': 'NUE2NjZGcGVlcmludGZwb24gNDk1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '365598521716',
                                                                                                                  'in_ucast_pkts': '1389965126',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '117',
                                                                                                                  'out_octets': '6044509892618',
                                                                                                                  'out_ucast_pkts': '4468657270',
                                                                                                                  'poller_hash': 'MzQ4MTA4cGVlcmludGZwb24gNTM0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '260350626491',
                                                                                                                  'in_ucast_pkts': '878305827',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '79',
                                                                                                                  'out_octets': '6340607094072',
                                                                                                                  'out_ucast_pkts': '4591639392',
                                                                                                                  'poller_hash': 'OTU2RUEwcGVlcmludGZwb24gNTM2'},
                                                                                                          {       'in_nucast_pkts': '437',
                                                                                                                  'in_octets': '6706774658',
                                                                                                                  'in_ucast_pkts': '10723619',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '226581',
                                                                                                                  'out_octets': '33267903501',
                                                                                                                  'out_ucast_pkts': '28583706',
                                                                                                                  'poller_hash': 'NjQwQjExcGVlcmludGZwb24gMTM4MTE1'},
                                                                                                          {       'in_nucast_pkts': '612336',
                                                                                                                  'in_octets': '1304962235281',
                                                                                                                  'in_ucast_pkts': '3417673650',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '35732147',
                                                                                                                  'out_octets': '10985307440924',
                                                                                                                  'out_ucast_pkts': '8453397977',
                                                                                                                  'poller_hash': 'ODY5M0UycGVlcmludGZwb24gMTQwOTE5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '584477821774',
                                                                                                                  'in_ucast_pkts': '2873697244',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '117',
                                                                                                                  'out_octets': '13108082384200',
                                                                                                                  'out_ucast_pkts': '9343811698',
                                                                                                                  'poller_hash': 'ODY5M0FFcGVlcmludGZwb24gMTQxMjQ1'},
                                                                                                          {       'in_nucast_pkts': '6478',
                                                                                                                  'in_octets': '3061204072',
                                                                                                                  'in_ucast_pkts': '5901548',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394271',
                                                                                                                  'out_octets': '29166803831',
                                                                                                                  'out_ucast_pkts': '22652769',
                                                                                                                  'poller_hash': 'ODcyNDQxcGVlcmludGZwb24gMTQzMjY0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '479334883375',
                                                                                                                  'in_ucast_pkts': '2721514229',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '63',
                                                                                                                  'out_octets': '15019572381907',
                                                                                                                  'out_ucast_pkts': '11017526610',
                                                                                                                  'poller_hash': 'QTlDM0FCcGVlcmludGZwb24gMTQ3NjQ3'},
                                                                                                          {       'in_nucast_pkts': '5',
                                                                                                                  'in_octets': '808811528',
                                                                                                                  'in_ucast_pkts': '4282441',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1207500',
                                                                                                                  'out_octets': '1361573416',
                                                                                                                  'out_ucast_pkts': '4206468',
                                                                                                                  'poller_hash': 'QzgzQkM0cGVlcmludGZwb24gMTY0Nzk1'},
                                                                                                          {       'in_nucast_pkts': '7265',
                                                                                                                  'in_octets': '3358996896',
                                                                                                                  'in_ucast_pkts': '16431459',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '394296',
                                                                                                                  'out_octets': '124290105710',
                                                                                                                  'out_ucast_pkts': '96193084',
                                                                                                                  'poller_hash': 'Q0JEMjA2cGVlcmludGZwb24gMTY3Nzgy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '250929124239',
                                                                                                                  'in_ucast_pkts': '625337751',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '70',
                                                                                                                  'out_octets': '2110453430276',
                                                                                                                  'out_ucast_pkts': '1712366328',
                                                                                                                  'poller_hash': 'Q0JEMTkwcGVlcmludGZwb24gMTY4MjIy'},
                                                                                                          {       'in_nucast_pkts': '6741',
                                                                                                                  'in_octets': '841822473',
                                                                                                                  'in_ucast_pkts': '3282548',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '383993',
                                                                                                                  'out_octets': '22945721085',
                                                                                                                  'out_ucast_pkts': '17333934',
                                                                                                                  'poller_hash': 'MTE5NTE3N3BlZXJpbnRmcG9uIDIxNjIxNw=='},
                                                                                                          {       'in_nucast_pkts': '139118',
                                                                                                                  'in_octets': '2721474266135',
                                                                                                                  'in_ucast_pkts': '9843707430',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '53556606',
                                                                                                                  'out_octets': '22571405451020',
                                                                                                                  'out_ucast_pkts': '20293835713',
                                                                                                                  'poller_hash': 'MTI3NTBFN3BlZXJpbnRmcG9uIDIyMDM4OQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NEFDN3BlZXJpbnRmcG9uIDIzNzgzMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '13009538606',
                                                                                                                  'in_ucast_pkts': '71916281',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33',
                                                                                                                  'out_octets': '42795287222',
                                                                                                                  'out_ucast_pkts': '85375019',
                                                                                                                  'poller_hash': 'MTI5NzRFQnBlZXJpbnRmcG9uIDI1NjUwMg=='},
                                                                                                          {       'in_nucast_pkts': '85653',
                                                                                                                  'in_octets': '24611570126',
                                                                                                                  'in_ucast_pkts': '110193061',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5655172',
                                                                                                                  'out_octets': '457794689194',
                                                                                                                  'out_ucast_pkts': '341464169',
                                                                                                                  'poller_hash': 'MTI5NUVEM3BlZXJpbnRmcG9uIDI4MDI2NA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '2250123864',
                                                                                         'out_octets': '193680899037517',
                                                                                         'out_ucast_pkts': '148522298982',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/7',
                                                                                                                 'id': '5/7',
                                                                                                                 'links': [       {       'connected_time': '335343',
                                                                                                                                          'fiber_length': 18777,
                                                                                                                                          'poller_hash': 'Q0JEMTY1cGVlcnBvbnBvbiAxMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '498961',
                                                                                                                                          'fiber_length': 19979,
                                                                                                                                          'poller_hash': 'ODY5MzkxcGVlcnBvbnBvbiA0MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '3434334',
                                                                                                                                          'fiber_length': 20911,
                                                                                                                                          'poller_hash': 'NDJEMUY5cGVlcnBvbnBvbiA0MDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '72492405',
                                                                                                                                          'fiber_length': 18170,
                                                                                                                                          'poller_hash': 'NDJCRjE1cGVlcnBvbnBvbiA0NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '8535204',
                                                                                                                                          'fiber_length': 19544,
                                                                                                                                          'poller_hash': 'NDJDNkJCcGVlcnBvbnBvbiA0ODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '67856103',
                                                                                                                                          'fiber_length': 22032,
                                                                                                                                          'poller_hash': 'NUE2NjZGcGVlcnBvbnBvbiA0OTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335345',
                                                                                                                                          'fiber_length': 17679,
                                                                                                                                          'poller_hash': 'MzQ4MTA4cGVlcnBvbnBvbiA1MzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '7928912',
                                                                                                                                          'fiber_length': 17168,
                                                                                                                                          'poller_hash': 'OTU2RUEwcGVlcnBvbnBvbiA1MzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335344',
                                                                                                                                          'fiber_length': 17590,
                                                                                                                                          'poller_hash': 'NjQwQjExcGVlcnBvbnBvbiAxMzgxMTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '65416542',
                                                                                                                                          'fiber_length': 17234,
                                                                                                                                          'poller_hash': 'ODY5M0UycGVlcnBvbnBvbiAxNDA5MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335343',
                                                                                                                                          'fiber_length': 18005,
                                                                                                                                          'poller_hash': 'ODY5M0FFcGVlcnBvbnBvbiAxNDEyNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335347',
                                                                                                                                          'fiber_length': 17172,
                                                                                                                                          'poller_hash': 'ODcyNDQxcGVlcnBvbnBvbiAxNDMyNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335345',
                                                                                                                                          'fiber_length': 19174,
                                                                                                                                          'poller_hash': 'QTlDM0FCcGVlcnBvbnBvbiAxNDc2NDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '1033095',
                                                                                                                                          'fiber_length': 21660,
                                                                                                                                          'poller_hash': 'QzgzQkM0cGVlcnBvbnBvbiAxNjQ3OTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335343',
                                                                                                                                          'fiber_length': 18659,
                                                                                                                                          'poller_hash': 'Q0JEMjA2cGVlcnBvbnBvbiAxNjc3ODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335343',
                                                                                                                                          'fiber_length': 17602,
                                                                                                                                          'poller_hash': 'Q0JEMTkwcGVlcnBvbnBvbiAxNjgyMjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '335346',
                                                                                                                                          'fiber_length': 18064,
                                                                                                                                          'poller_hash': 'MTE5NTE3N3BlZXJwb25wb24gMjE2MjE3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '13895239',
                                                                                                                                          'fiber_length': 21791,
                                                                                                                                          'poller_hash': 'MTI3NTBFN3BlZXJwb25wb24gMjIwMzg5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '72734407',
                                                                                                                                          'fiber_length': 21904,
                                                                                                                                          'poller_hash': 'MTI3NEFDN3BlZXJwb25wb24gMjM3ODMz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '4232853',
                                                                                                                                          'fiber_length': 18036,
                                                                                                                                          'poller_hash': 'MTI5NzRFQnBlZXJwb25wb24gMjU2NTAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722},
                                                                                                                                  {       'connected_time': '4835297',
                                                                                                                                          'fiber_length': 19483,
                                                                                                                                          'poller_hash': 'MTI5NUVEM3BlZXJwb25wb24gMjgwMjY0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.722}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/7',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '5/8',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '638',
                                                                                         'in_octets': '14967131941',
                                                                                         'in_ucast_pkts': '118387855',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTI3NDhCNHBlZXJpbnRmcG9uIDI3OTM1Ng=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '8051226611',
                                                                                                                  'in_ucast_pkts': '24073121',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '108362666857',
                                                                                                                  'out_ucast_pkts': '91146553',
                                                                                                                  'poller_hash': 'MTI5NUVDNHBlZXJpbnRmcG9uIDI3OTQ4Ng=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '5/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '65173690',
                                                                                         'out_octets': '157154098326',
                                                                                         'out_ucast_pkts': '206278845',
                                                                                         'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '5/8',
                                                                                                                 'id': '5/8',
                                                                                                                 'links': [       {       'connected_time': '2503016',
                                                                                                                                          'fiber_length': 12020,
                                                                                                                                          'poller_hash': 'MTI3NDhCNHBlZXJwb25wb24gMjc5MzU2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791},
                                                                                                                                  {       'connected_time': '5870595',
                                                                                                                                          'fiber_length': 12372,
                                                                                                                                          'poller_hash': 'MTI5NUVDNHBlZXJwb25wb24gMjc5NDg2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.791}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '5/8',
                                                                                                                 'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNjExNjA0Mjgx',
                                                           'serial_number': '211611604281',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43502537'},
                                                   {       'id': '6',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '6/1',
                                                                                         'in_discards': '501295',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '8474275',
                                                                                         'in_octets': '23049418223520',
                                                                                         'in_ucast_pkts': '81259707331',
                                                                                         'links': [       {       'in_nucast_pkts': '10151',
                                                                                                                  'in_octets': '783390780130',
                                                                                                                  'in_ucast_pkts': '593886644',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '21816336',
                                                                                                                  'out_octets': '9129386004183',
                                                                                                                  'out_ucast_pkts': '2861820571',
                                                                                                                  'poller_hash': 'NEZBMjVDcGVlcmludGZwb24gMg=='},
                                                                                                          {       'in_nucast_pkts': '49202323',
                                                                                                                  'in_octets': '321645474912',
                                                                                                                  'in_ucast_pkts': '2573399717',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '44888533193',
                                                                                                                  'out_octets': '69286673895933',
                                                                                                                  'out_ucast_pkts': '6787917083',
                                                                                                                  'poller_hash': 'M0FERDBFcGVlcmludGZwb24gMTE1'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '43807249955',
                                                                                                                  'in_ucast_pkts': '197725654',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919752',
                                                                                                                  'out_octets': '1439329920585',
                                                                                                                  'out_ucast_pkts': '992161959',
                                                                                                                  'poller_hash': 'MzVFOUMzcGVlcmludGZwb24gMTI5'},
                                                                                                          {       'in_nucast_pkts': '23',
                                                                                                                  'in_octets': '76671518660',
                                                                                                                  'in_ucast_pkts': '479320748',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9753903',
                                                                                                                  'out_octets': '2444218624494',
                                                                                                                  'out_ucast_pkts': '1744741914',
                                                                                                                  'poller_hash': 'MzIyNzNCcGVlcmludGZwb24gMTcz'},
                                                                                                          {       'in_nucast_pkts': '17',
                                                                                                                  'in_octets': '24075029122',
                                                                                                                  'in_ucast_pkts': '182645458',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5084112',
                                                                                                                  'out_octets': '1138820894928',
                                                                                                                  'out_ucast_pkts': '790960215',
                                                                                                                  'poller_hash': 'MzIyNzI5cGVlcmludGZwb24gMTc0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '18717416172',
                                                                                                                  'in_ucast_pkts': '30505315',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '231',
                                                                                                                  'out_octets': '211518474830',
                                                                                                                  'out_ucast_pkts': '209976875',
                                                                                                                  'poller_hash': 'MzIyODcxcGVlcmludGZwb24gMTc2'},
                                                                                                          {       'in_nucast_pkts': '12471',
                                                                                                                  'in_octets': '140178648831',
                                                                                                                  'in_ucast_pkts': '868438891',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710247',
                                                                                                                  'out_octets': '6917769765984',
                                                                                                                  'out_ucast_pkts': '4829110153',
                                                                                                                  'poller_hash': 'MzIyM0MxcGVlcmludGZwb24gMTc3'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '30931131669',
                                                                                                                  'in_ucast_pkts': '173460800',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4682244',
                                                                                                                  'out_octets': '896628343833',
                                                                                                                  'out_ucast_pkts': '635746232',
                                                                                                                  'poller_hash': 'MzIyM0JFcGVlcmludGZwb24gMTc4'},
                                                                                                          {       'in_nucast_pkts': '213676',
                                                                                                                  'in_octets': '139800279181',
                                                                                                                  'in_ucast_pkts': '655816211',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61014332',
                                                                                                                  'out_octets': '1073213320231',
                                                                                                                  'out_ucast_pkts': '1176075356',
                                                                                                                  'poller_hash': 'OTU2RTVGcGVlcmludGZwb24gMTgw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '204866431802',
                                                                                                                  'in_ucast_pkts': '858918043',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '231',
                                                                                                                  'out_octets': '3102496969858',
                                                                                                                  'out_ucast_pkts': '2316335196',
                                                                                                                  'poller_hash': 'MzIyNkFEcGVlcmludGZwb24gMTgy'},
                                                                                                          {       'in_nucast_pkts': '197',
                                                                                                                  'in_octets': '196687169364',
                                                                                                                  'in_ucast_pkts': '642002235',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '44315403',
                                                                                                                  'out_octets': '3659640806126',
                                                                                                                  'out_ucast_pkts': '2781507231',
                                                                                                                  'poller_hash': 'MzIyODE4cGVlcmludGZwb24gMTgz'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '2813619178',
                                                                                                                  'in_ucast_pkts': '27340675',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '993968',
                                                                                                                  'out_octets': '159985846801',
                                                                                                                  'out_ucast_pkts': '108260387',
                                                                                                                  'poller_hash': 'MzIyODdBcGVlcmludGZwb24gMTg1'},
                                                                                                          {       'in_nucast_pkts': '8768',
                                                                                                                  'in_octets': '158047662204',
                                                                                                                  'in_ucast_pkts': '546301620',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39519357',
                                                                                                                  'out_octets': '1241963606079',
                                                                                                                  'out_ucast_pkts': '1118417924',
                                                                                                                  'poller_hash': 'MzVFOUM3cGVlcmludGZwb24gMTg2'},
                                                                                                          {       'in_nucast_pkts': '150561',
                                                                                                                  'in_octets': '1034414186643',
                                                                                                                  'in_ucast_pkts': '975813587',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652379',
                                                                                                                  'out_octets': '752923341203',
                                                                                                                  'out_ucast_pkts': '1006442707',
                                                                                                                  'poller_hash': 'NEY5MUU5cGVlcmludGZwb24gMTg3'},
                                                                                                          {       'in_nucast_pkts': '2129759',
                                                                                                                  'in_octets': '58707596057',
                                                                                                                  'in_ucast_pkts': '188018884',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4442741401',
                                                                                                                  'out_octets': '6462154229915',
                                                                                                                  'out_ucast_pkts': '381881567',
                                                                                                                  'poller_hash': 'MzIyOUFGcGVlcmludGZwb24gMTg4'},
                                                                                                          {       'in_nucast_pkts': '57997',
                                                                                                                  'in_octets': '1047449945795',
                                                                                                                  'in_ucast_pkts': '1985401029',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61708964',
                                                                                                                  'out_octets': '3576151707434',
                                                                                                                  'out_ucast_pkts': '3249002723',
                                                                                                                  'poller_hash': 'MzIyNDFFcGVlcmludGZwb24gMTg5'},
                                                                                                          {       'in_nucast_pkts': '62',
                                                                                                                  'in_octets': '87344010055',
                                                                                                                  'in_ucast_pkts': '758160614',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4879293',
                                                                                                                  'out_octets': '2502801864029',
                                                                                                                  'out_ucast_pkts': '1747748417',
                                                                                                                  'poller_hash': 'MzIyNDI0cGVlcmludGZwb24gMTk0'},
                                                                                                          {       'in_nucast_pkts': '5572919',
                                                                                                                  'in_octets': '275084848504',
                                                                                                                  'in_ucast_pkts': '1439959507',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9904841',
                                                                                                                  'out_octets': '6681588337177',
                                                                                                                  'out_ucast_pkts': '4758070447',
                                                                                                                  'poller_hash': 'MzIyODNCcGVlcmludGZwb24gMjA0'},
                                                                                                          {       'in_nucast_pkts': '6647',
                                                                                                                  'in_octets': '101301132901',
                                                                                                                  'in_ucast_pkts': '443159731',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61709911',
                                                                                                                  'out_octets': '1729394720975',
                                                                                                                  'out_ucast_pkts': '1268147794',
                                                                                                                  'poller_hash': 'MzIyOEMxcGVlcmludGZwb24gMjA2'},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '479586636223',
                                                                                                                  'in_ucast_pkts': '3861272606',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919966',
                                                                                                                  'out_octets': '15903468478317',
                                                                                                                  'out_ucast_pkts': '10938427492',
                                                                                                                  'poller_hash': 'MzIyNjAwcGVlcmludGZwb24gMjE1'},
                                                                                                          {       'in_nucast_pkts': '131663',
                                                                                                                  'in_octets': '87660242490',
                                                                                                                  'in_ucast_pkts': '448161439',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4197346',
                                                                                                                  'out_octets': '1715279236571',
                                                                                                                  'out_ucast_pkts': '1266424653',
                                                                                                                  'poller_hash': 'NjQwQjI3cGVlcmludGZwb24gNDYx'},
                                                                                                          {       'in_nucast_pkts': '2309732',
                                                                                                                  'in_octets': '2100555949462',
                                                                                                                  'in_ucast_pkts': '6012360135',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38392824',
                                                                                                                  'out_octets': '10086052639534',
                                                                                                                  'out_ucast_pkts': '9730720550',
                                                                                                                  'poller_hash': 'ODcyNEZFcGVlcmludGZwb24gMTQzNzQy'},
                                                                                                          {       'in_nucast_pkts': '154503',
                                                                                                                  'in_octets': '43242326409',
                                                                                                                  'in_ucast_pkts': '194951693',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920016',
                                                                                                                  'out_octets': '2283715735908',
                                                                                                                  'out_ucast_pkts': '1622573352',
                                                                                                                  'poller_hash': 'QzgzQjZGcGVlcmludGZwb24gMTY0MTY3'},
                                                                                                          {       'in_nucast_pkts': '153947',
                                                                                                                  'in_octets': '63334092710',
                                                                                                                  'in_ucast_pkts': '160454075',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919015',
                                                                                                                  'out_octets': '702779231618',
                                                                                                                  'out_ucast_pkts': '581370077',
                                                                                                                  'poller_hash': 'QzgzMTMxcGVlcmludGZwb24gMTkzMjk3'},
                                                                                                          {       'in_nucast_pkts': '57763904',
                                                                                                                  'in_octets': '104186266757',
                                                                                                                  'in_ucast_pkts': '622096314',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62552116191',
                                                                                                                  'out_octets': '88470814115337',
                                                                                                                  'out_ucast_pkts': '3226307595',
                                                                                                                  'poller_hash': 'QzgxN0ZFcGVlcmludGZwb24gMjA4OTc2'},
                                                                                                          {       'in_nucast_pkts': '163451',
                                                                                                                  'in_octets': '71608203986',
                                                                                                                  'in_ucast_pkts': '408059948',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919822',
                                                                                                                  'out_octets': '3442558103631',
                                                                                                                  'out_ucast_pkts': '2405154705',
                                                                                                                  'poller_hash': 'MTI3NDA2RnBlZXJpbnRmcG9uIDI0NzU5Mg=='},
                                                                                                          {       'in_nucast_pkts': '152125',
                                                                                                                  'in_octets': '196325902412',
                                                                                                                  'in_ucast_pkts': '534327456',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919930',
                                                                                                                  'out_octets': '2314263681628',
                                                                                                                  'out_ucast_pkts': '1759068029',
                                                                                                                  'poller_hash': 'MTE3REJBQ3BlZXJpbnRmcG9uIDI1MzU3Nw=='},
                                                                                                          {       'in_nucast_pkts': '471911',
                                                                                                                  'in_octets': '110916545356',
                                                                                                                  'in_ucast_pkts': '1105502636',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '29360794',
                                                                                                                  'out_octets': '7259752310644',
                                                                                                                  'out_ucast_pkts': '4948570877',
                                                                                                                  'poller_hash': 'MTE3REMxRnBlZXJpbnRmcG9uIDI1NjQzMg=='},
                                                                                                          {       'in_nucast_pkts': '155504',
                                                                                                                  'in_octets': '592813435683',
                                                                                                                  'in_ucast_pkts': '1316471687',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9511572',
                                                                                                                  'out_octets': '5291430138304',
                                                                                                                  'out_ucast_pkts': '4062124451',
                                                                                                                  'poller_hash': 'MTI3NDdEQXBlZXJpbnRmcG9uIDI3MjMyNA=='},
                                                                                                          {       'in_nucast_pkts': '20',
                                                                                                                  'in_octets': '27190528470',
                                                                                                                  'in_ucast_pkts': '287659063',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9635078',
                                                                                                                  'out_octets': '1764236051363',
                                                                                                                  'out_ucast_pkts': '1183986942',
                                                                                                                  'poller_hash': 'MTI3MEQ2RHBlZXJpbnRmcG9uIDI3ODQ2Nw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '250524680242',
                                                                                         'out_octets': '625464430629230',
                                                                                         'out_ucast_pkts': '212608639850',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/1',
                                                                                                                 'id': '6/1',
                                                                                                                 'links': [       {       'connected_time': '68952515',
                                                                                                                                          'fiber_length': 22900,
                                                                                                                                          'poller_hash': 'NEZBMjVDcGVlcnBvbnBvbiAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '552134',
                                                                                                                                          'fiber_length': 23075,
                                                                                                                                          'poller_hash': 'M0FERDBFcGVlcnBvbnBvbiAxMTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535469',
                                                                                                                                          'fiber_length': 23184,
                                                                                                                                          'poller_hash': 'MzVFOUMzcGVlcnBvbnBvbiAxMjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8394287',
                                                                                                                                          'fiber_length': 22625,
                                                                                                                                          'poller_hash': 'MzIyNzNCcGVlcnBvbnBvbiAxNzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '4332898',
                                                                                                                                          'fiber_length': 22226,
                                                                                                                                          'poller_hash': 'MzIyNzI5cGVlcnBvbnBvbiAxNzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '15630595',
                                                                                                                                          'fiber_length': 22859,
                                                                                                                                          'poller_hash': 'MzIyODcxcGVlcnBvbnBvbiAxNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '21887039',
                                                                                                                                          'fiber_length': 22262,
                                                                                                                                          'poller_hash': 'MzIyM0MxcGVlcnBvbnBvbiAxNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '3981837',
                                                                                                                                          'fiber_length': 22588,
                                                                                                                                          'poller_hash': 'MzIyM0JFcGVlcnBvbnBvbiAxNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '21887032',
                                                                                                                                          'fiber_length': 22750,
                                                                                                                                          'poller_hash': 'OTU2RTVGcGVlcnBvbnBvbiAxODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8027224',
                                                                                                                                          'fiber_length': 22236,
                                                                                                                                          'poller_hash': 'MzIyNkFEcGVlcnBvbnBvbiAxODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '73454773',
                                                                                                                                          'fiber_length': 22361,
                                                                                                                                          'poller_hash': 'MzIyODE4cGVlcnBvbnBvbiAxODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '847640',
                                                                                                                                          'fiber_length': 22988,
                                                                                                                                          'poller_hash': 'MzIyODdBcGVlcnBvbnBvbiAxODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '67856203',
                                                                                                                                          'fiber_length': 23155,
                                                                                                                                          'poller_hash': 'MzVFOUM3cGVlcnBvbnBvbiAxODY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535152',
                                                                                                                                          'fiber_length': 22674,
                                                                                                                                          'poller_hash': 'NEY5MUU5cGVlcnBvbnBvbiAxODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535254',
                                                                                                                                          'fiber_length': 22942,
                                                                                                                                          'poller_hash': 'MzIyOUFGcGVlcnBvbnBvbiAxODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '21886444',
                                                                                                                                          'fiber_length': 23339,
                                                                                                                                          'poller_hash': 'MzIyNDFFcGVlcnBvbnBvbiAxODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '4155956',
                                                                                                                                          'fiber_length': 22999,
                                                                                                                                          'poller_hash': 'MzIyNDI0cGVlcnBvbnBvbiAxOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535240',
                                                                                                                                          'fiber_length': 22626,
                                                                                                                                          'poller_hash': 'MzIyODNCcGVlcnBvbnBvbiAyMDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '21886877',
                                                                                                                                          'fiber_length': 22622,
                                                                                                                                          'poller_hash': 'MzIyOEMxcGVlcnBvbnBvbiAyMDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535239',
                                                                                                                                          'fiber_length': 22267,
                                                                                                                                          'poller_hash': 'MzIyNjAwcGVlcnBvbnBvbiAyMTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '6296984',
                                                                                                                                          'fiber_length': 22698,
                                                                                                                                          'poller_hash': 'NjQwQjI3cGVlcnBvbnBvbiA0NjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '66603086',
                                                                                                                                          'fiber_length': 23530,
                                                                                                                                          'poller_hash': 'ODcyNEZFcGVlcnBvbnBvbiAxNDM3NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535335',
                                                                                                                                          'fiber_length': 22507,
                                                                                                                                          'poller_hash': 'QzgzQjZGcGVlcnBvbnBvbiAxNjQxNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535226',
                                                                                                                                          'fiber_length': 22187,
                                                                                                                                          'poller_hash': 'QzgzMTMxcGVlcnBvbnBvbiAxOTMyOTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '21886400',
                                                                                                                                          'fiber_length': 23178,
                                                                                                                                          'poller_hash': 'QzgxN0ZFcGVlcnBvbnBvbiAyMDg5NzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535176',
                                                                                                                                          'fiber_length': 22569,
                                                                                                                                          'poller_hash': 'MTI3NDA2RnBlZXJwb25wb24gMjQ3NTky',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8535270',
                                                                                                                                          'fiber_length': 22465,
                                                                                                                                          'poller_hash': 'MTE3REJBQ3BlZXJwb25wb24gMjUzNTc3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '25393059',
                                                                                                                                          'fiber_length': 23190,
                                                                                                                                          'poller_hash': 'MTE3REMxRnBlZXJwb25wb24gMjU2NDMy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8182288',
                                                                                                                                          'fiber_length': 22444,
                                                                                                                                          'poller_hash': 'MTI3NDdEQXBlZXJwb25wb24gMjcyMzI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393},
                                                                                                                                  {       'connected_time': '8291191',
                                                                                                                                          'fiber_length': 23195,
                                                                                                                                          'poller_hash': 'MTI3MEQ2RHBlZXJwb25wb24gMjc4NDY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.393}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/1',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/2',
                                                                                         'in_discards': '597890',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '9813890',
                                                                                         'in_octets': '14474677857434',
                                                                                         'in_ucast_pkts': '52444716164',
                                                                                         'links': [       {       'in_nucast_pkts': '937512',
                                                                                                                  'in_octets': '1212745293173',
                                                                                                                  'in_ucast_pkts': '3713818693',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '55649463',
                                                                                                                  'out_octets': '17601237980155',
                                                                                                                  'out_ucast_pkts': '13217070020',
                                                                                                                  'poller_hash': 'QzgzQkE3cGVlcmludGZwb24gMTkw'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '39585049996',
                                                                                                                  'in_ucast_pkts': '157881046',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652498',
                                                                                                                  'out_octets': '617124826541',
                                                                                                                  'out_ucast_pkts': '480592289',
                                                                                                                  'poller_hash': 'NEY4REQycGVlcmludGZwb24gMTkx'},
                                                                                                          {       'in_nucast_pkts': '11286',
                                                                                                                  'in_octets': '516565464062',
                                                                                                                  'in_ucast_pkts': '1709342745',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61709813',
                                                                                                                  'out_octets': '4737596752077',
                                                                                                                  'out_ucast_pkts': '3592941855',
                                                                                                                  'poller_hash': 'MzIyNDJGcGVlcmludGZwb24gMTky'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '477957294090',
                                                                                                                  'in_ucast_pkts': '1292134835',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '2495748245162',
                                                                                                                  'out_ucast_pkts': '2238166978',
                                                                                                                  'poller_hash': 'NTU1Q0MycGVlcmludGZwb24gMTk1'},
                                                                                                          {       'in_nucast_pkts': '5139274',
                                                                                                                  'in_octets': '92289770842',
                                                                                                                  'in_ucast_pkts': '104891318',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '19363047799',
                                                                                                                  'out_octets': '26155427122172',
                                                                                                                  'out_ucast_pkts': '130330536',
                                                                                                                  'poller_hash': 'MjA1RERBcGVlcmludGZwb24gMTk2'},
                                                                                                          {       'in_nucast_pkts': '474871',
                                                                                                                  'in_octets': '284556994238',
                                                                                                                  'in_ucast_pkts': '763097918',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '30163648',
                                                                                                                  'out_octets': '1632962298784',
                                                                                                                  'out_ucast_pkts': '1489378667',
                                                                                                                  'poller_hash': 'MzIyNDE4cGVlcmludGZwb24gMTk3'},
                                                                                                          {       'in_nucast_pkts': '36414',
                                                                                                                  'in_octets': '78990745098',
                                                                                                                  'in_ucast_pkts': '311337624',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '24135150',
                                                                                                                  'out_octets': '662293673537',
                                                                                                                  'out_ucast_pkts': '581909552',
                                                                                                                  'poller_hash': 'NEQzQjRDcGVlcmludGZwb24gMTk4'},
                                                                                                          {       'in_nucast_pkts': '12127',
                                                                                                                  'in_octets': '258435090756',
                                                                                                                  'in_ucast_pkts': '1552263108',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710128',
                                                                                                                  'out_octets': '8091766302272',
                                                                                                                  'out_ucast_pkts': '5814867463',
                                                                                                                  'poller_hash': 'MzIyODIycGVlcmludGZwb24gMjAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzIyOTEycGVlcmludGZwb24gMjAx'},
                                                                                                          {       'in_nucast_pkts': '164074',
                                                                                                                  'in_octets': '88972112183',
                                                                                                                  'in_ucast_pkts': '445504369',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9910981',
                                                                                                                  'out_octets': '1579098041921',
                                                                                                                  'out_ucast_pkts': '1170780019',
                                                                                                                  'poller_hash': 'MzIyN0YycGVlcmludGZwb24gMjAy'},
                                                                                                          {       'in_nucast_pkts': '7996168',
                                                                                                                  'in_octets': '45767357846',
                                                                                                                  'in_ucast_pkts': '503561460',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '136512106',
                                                                                                                  'out_octets': '3390280760948',
                                                                                                                  'out_ucast_pkts': '2384726829',
                                                                                                                  'poller_hash': 'MzIyODY0cGVlcmludGZwb24gMjA3'},
                                                                                                          {       'in_nucast_pkts': '168202',
                                                                                                                  'in_octets': '420736113273',
                                                                                                                  'in_ucast_pkts': '620332570',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920032',
                                                                                                                  'out_octets': '1706197280222',
                                                                                                                  'out_ucast_pkts': '1456538770',
                                                                                                                  'poller_hash': 'NTEyNTE4cGVlcmludGZwb24gMjA4'},
                                                                                                          {       'in_nucast_pkts': '657037',
                                                                                                                  'in_octets': '196308862443',
                                                                                                                  'in_ucast_pkts': '731258084',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39262323',
                                                                                                                  'out_octets': '6674144852682',
                                                                                                                  'out_ucast_pkts': '4896864188',
                                                                                                                  'poller_hash': 'MzIyNkM4cGVlcmludGZwb24gMjA5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '165880523543',
                                                                                                                  'in_ucast_pkts': '1047111901',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '3738604999824',
                                                                                                                  'out_ucast_pkts': '2687510560',
                                                                                                                  'poller_hash': 'NDI4MUFEcGVlcmludGZwb24gMjU5'},
                                                                                                          {       'in_nucast_pkts': '243',
                                                                                                                  'in_octets': '324876426963',
                                                                                                                  'in_ucast_pkts': '1111161128',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39489278',
                                                                                                                  'out_octets': '5509505469287',
                                                                                                                  'out_ucast_pkts': '4064191472',
                                                                                                                  'poller_hash': 'NTEyNTM4cGVlcmludGZwb24gMjk4'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '174659753483',
                                                                                                                  'in_ucast_pkts': '492003005',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1647251',
                                                                                                                  'out_octets': '940506082993',
                                                                                                                  'out_ucast_pkts': '728204845',
                                                                                                                  'poller_hash': 'NDNCNDM3cGVlcmludGZwb24gMzAx'},
                                                                                                          {       'in_nucast_pkts': '20813',
                                                                                                                  'in_octets': '36290338839',
                                                                                                                  'in_ucast_pkts': '258014566',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5680964',
                                                                                                                  'out_octets': '1017718653859',
                                                                                                                  'out_ucast_pkts': '716734739',
                                                                                                                  'poller_hash': 'NDNCM0ZEcGVlcmludGZwb24gMzM3'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '44420951659',
                                                                                                                  'in_ucast_pkts': '132031779',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4414371',
                                                                                                                  'out_octets': '278514337798',
                                                                                                                  'out_ucast_pkts': '212332799',
                                                                                                                  'poller_hash': 'NDNBRjgzcGVlcmludGZwb24gMzg3'},
                                                                                                          {       'in_nucast_pkts': '21',
                                                                                                                  'in_octets': '61911372611',
                                                                                                                  'in_ucast_pkts': '283608202',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11414643587',
                                                                                                                  'out_octets': '17024903124278',
                                                                                                                  'out_ucast_pkts': '1279274354',
                                                                                                                  'poller_hash': 'NDJEMDZFcGVlcmludGZwb24gNDM3'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '42311768474',
                                                                                                                  'in_ucast_pkts': '138494639',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920062',
                                                                                                                  'out_octets': '1244764889734',
                                                                                                                  'out_ucast_pkts': '901837266',
                                                                                                                  'poller_hash': 'NDJEMDdEcGVlcmludGZwb24gNDY4'},
                                                                                                          {       'in_nucast_pkts': '87215821',
                                                                                                                  'in_octets': '828092551178',
                                                                                                                  'in_ucast_pkts': '2065911330',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '84820921895',
                                                                                                                  'out_octets': '120692158146893',
                                                                                                                  'out_ucast_pkts': '5296163715',
                                                                                                                  'poller_hash': 'NDJDNkFBcGVlcmludGZwb24gNDcx'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '64997662567',
                                                                                                                  'in_ucast_pkts': '442309040',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652574',
                                                                                                                  'out_octets': '1402947966173',
                                                                                                                  'out_ucast_pkts': '981701636',
                                                                                                                  'poller_hash': 'NjQwQjI0cGVlcmludGZwb24gNDcy'},
                                                                                                          {       'in_nucast_pkts': '32',
                                                                                                                  'in_octets': '15590539901',
                                                                                                                  'in_ucast_pkts': '110984397',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919650',
                                                                                                                  'out_octets': '577316807514',
                                                                                                                  'out_ucast_pkts': '394668307',
                                                                                                                  'poller_hash': 'NDJDMDhGcGVlcmludGZwb24gNDc5'},
                                                                                                          {       'in_nucast_pkts': '9742',
                                                                                                                  'in_octets': '167480799701',
                                                                                                                  'in_ucast_pkts': '650044499',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9937593',
                                                                                                                  'out_octets': '2526593287704',
                                                                                                                  'out_ucast_pkts': '1879884995',
                                                                                                                  'poller_hash': 'NDJDNEI4cGVlcmludGZwb24gNDkx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'Mzc4ODNFcGVlcmludGZwb24gNTIw'},
                                                                                                          {       'in_nucast_pkts': '35245',
                                                                                                                  'in_octets': '25146455229',
                                                                                                                  'in_ucast_pkts': '83223631',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1936295',
                                                                                                                  'out_octets': '259911461116',
                                                                                                                  'out_ucast_pkts': '209416686',
                                                                                                                  'poller_hash': 'QzgxRTRDcGVlcmludGZwb24gMTk3NzE2'},
                                                                                                          {       'in_nucast_pkts': '71834',
                                                                                                                  'in_octets': '87344712806',
                                                                                                                  'in_ucast_pkts': '418632755',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4686154',
                                                                                                                  'out_octets': '2215615976960',
                                                                                                                  'out_ucast_pkts': '1596773593',
                                                                                                                  'poller_hash': 'QzgxRTE5cGVlcmludGZwb24gMjA0OTE1'},
                                                                                                          {       'in_nucast_pkts': '159357',
                                                                                                                  'in_octets': '46587927477',
                                                                                                                  'in_ucast_pkts': '256367096',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919973',
                                                                                                                  'out_octets': '1290852796582',
                                                                                                                  'out_ucast_pkts': '905524667',
                                                                                                                  'poller_hash': 'MTI3NDJEMHBlZXJpbnRmcG9uIDI0NzYwMg=='},
                                                                                                          {       'in_nucast_pkts': '64948',
                                                                                                                  'in_octets': '8589656553',
                                                                                                                  'in_ucast_pkts': '17945799',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3749479',
                                                                                                                  'out_octets': '50654069692',
                                                                                                                  'out_ucast_pkts': '45087911',
                                                                                                                  'poller_hash': 'MTI5NUM4NHBlZXJpbnRmcG9uIDI4MjI0Mw=='},
                                                                                                          {       'in_nucast_pkts': '26141',
                                                                                                                  'in_octets': '51970229119',
                                                                                                                  'in_ucast_pkts': '83997726',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1426539',
                                                                                                                  'out_octets': '216266379728',
                                                                                                                  'out_ucast_pkts': '171263199',
                                                                                                                  'poller_hash': 'MTI5NjExMXBlZXJpbnRmcG9uIDI4NDU2Mg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '266279762553',
                                                                                         'out_octets': '556332980348503',
                                                                                         'out_ucast_pkts': '145344713824',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/2',
                                                                                                                 'id': '6/2',
                                                                                                                 'links': [       {       'connected_time': '12240433',
                                                                                                                                          'fiber_length': 22834,
                                                                                                                                          'poller_hash': 'QzgzQkE3cGVlcnBvbnBvbiAxOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535105',
                                                                                                                                          'fiber_length': 23553,
                                                                                                                                          'poller_hash': 'NEY4REQycGVlcnBvbnBvbiAxOTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '21886927',
                                                                                                                                          'fiber_length': 22601,
                                                                                                                                          'poller_hash': 'MzIyNDJGcGVlcnBvbnBvbiAxOTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '639563',
                                                                                                                                          'fiber_length': 22625,
                                                                                                                                          'poller_hash': 'NTU1Q0MycGVlcnBvbnBvbiAxOTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '927589',
                                                                                                                                          'fiber_length': 22350,
                                                                                                                                          'poller_hash': 'MjA1RERBcGVlcnBvbnBvbiAxOTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '26231005',
                                                                                                                                          'fiber_length': 23332,
                                                                                                                                          'poller_hash': 'MzIyNDE4cGVlcnBvbnBvbiAxOTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '551824',
                                                                                                                                          'fiber_length': 22290,
                                                                                                                                          'poller_hash': 'NEQzQjRDcGVlcnBvbnBvbiAxOTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '21886272',
                                                                                                                                          'fiber_length': 22860,
                                                                                                                                          'poller_hash': 'MzIyODIycGVlcnBvbnBvbiAyMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '21886755',
                                                                                                                                          'fiber_length': 22840,
                                                                                                                                          'poller_hash': 'MzIyOTEycGVlcnBvbnBvbiAyMDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8551983',
                                                                                                                                          'fiber_length': 22780,
                                                                                                                                          'poller_hash': 'MzIyN0YycGVlcnBvbnBvbiAyMDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '17360321',
                                                                                                                                          'fiber_length': 22958,
                                                                                                                                          'poller_hash': 'MzIyODY0cGVlcnBvbnBvbiAyMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535404',
                                                                                                                                          'fiber_length': 23661,
                                                                                                                                          'poller_hash': 'NTEyNTE4cGVlcnBvbnBvbiAyMDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '67900401',
                                                                                                                                          'fiber_length': 22858,
                                                                                                                                          'poller_hash': 'MzIyNkM4cGVlcnBvbnBvbiAyMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535258',
                                                                                                                                          'fiber_length': 22826,
                                                                                                                                          'poller_hash': 'NDI4MUFEcGVlcnBvbnBvbiAyNTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '67856134',
                                                                                                                                          'fiber_length': 22555,
                                                                                                                                          'poller_hash': 'NTEyNTM4cGVlcnBvbnBvbiAyOTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '2414745',
                                                                                                                                          'fiber_length': 22440,
                                                                                                                                          'poller_hash': 'NDNCNDM3cGVlcnBvbnBvbiAzMDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535071',
                                                                                                                                          'fiber_length': 22820,
                                                                                                                                          'poller_hash': 'NDNCM0ZEcGVlcnBvbnBvbiAzMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '6627411',
                                                                                                                                          'fiber_length': 22462,
                                                                                                                                          'poller_hash': 'NDNBRjgzcGVlcnBvbnBvbiAzODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535265',
                                                                                                                                          'fiber_length': 22528,
                                                                                                                                          'poller_hash': 'NDJEMDZFcGVlcnBvbnBvbiA0Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535208',
                                                                                                                                          'fiber_length': 22512,
                                                                                                                                          'poller_hash': 'NDJEMDdEcGVlcnBvbnBvbiA0Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '21886731',
                                                                                                                                          'fiber_length': 22809,
                                                                                                                                          'poller_hash': 'NDJDNkFBcGVlcnBvbnBvbiA0NzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535105',
                                                                                                                                          'fiber_length': 23238,
                                                                                                                                          'poller_hash': 'NjQwQjI0cGVlcnBvbnBvbiA0NzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535276',
                                                                                                                                          'fiber_length': 22854,
                                                                                                                                          'poller_hash': 'NDJDMDhGcGVlcnBvbnBvbiA0Nzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535291',
                                                                                                                                          'fiber_length': 22851,
                                                                                                                                          'poller_hash': 'NDJDNEI4cGVlcnBvbnBvbiA0OTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '551821',
                                                                                                                                          'fiber_length': 22447,
                                                                                                                                          'poller_hash': 'Mzc4ODNFcGVlcnBvbnBvbiA1MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '1648526',
                                                                                                                                          'fiber_length': 22817,
                                                                                                                                          'poller_hash': 'QzgxRTRDcGVlcnBvbnBvbiAxOTc3MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '3982702',
                                                                                                                                          'fiber_length': 23289,
                                                                                                                                          'poller_hash': 'QzgxRTE5cGVlcnBvbnBvbiAyMDQ5MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '8535212',
                                                                                                                                          'fiber_length': 22525,
                                                                                                                                          'poller_hash': 'MTI3NDJEMHBlZXJwb25wb24gMjQ3NjAy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '3176203',
                                                                                                                                          'fiber_length': 22957,
                                                                                                                                          'poller_hash': 'MTI5NUM4NHBlZXJwb25wb24gMjgyMjQz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188},
                                                                                                                                  {       'connected_time': '1217113',
                                                                                                                                          'fiber_length': 22866,
                                                                                                                                          'poller_hash': 'MTI5NjExMXBlZXJwb25wb24gMjg0NTYy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.188}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/2',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/3',
                                                                                         'in_discards': '12788',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2312790',
                                                                                         'in_octets': '47854411065470',
                                                                                         'in_ucast_pkts': '108042581761',
                                                                                         'links': [       {       'in_nucast_pkts': '100',
                                                                                                                  'in_octets': '17672146152989',
                                                                                                                  'in_ucast_pkts': '16795600894',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '36122332',
                                                                                                                  'out_octets': '22976172642282',
                                                                                                                  'out_ucast_pkts': '23506041948',
                                                                                                                  'poller_hash': 'MzIyNjBEcGVlcmludGZwb24gMjEx'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '69733650661',
                                                                                                                  'in_ucast_pkts': '516856609',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919812',
                                                                                                                  'out_octets': '1610547735454',
                                                                                                                  'out_ucast_pkts': '1174935723',
                                                                                                                  'poller_hash': 'NDI4MUVDcGVlcmludGZwb24gMjEy'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '59606836603',
                                                                                                                  'in_ucast_pkts': '438079941',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920282',
                                                                                                                  'out_octets': '1292521846620',
                                                                                                                  'out_ucast_pkts': '901295252',
                                                                                                                  'poller_hash': 'MzIyNDFEcGVlcmludGZwb24gMjEz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '380631709408',
                                                                                                                  'in_ucast_pkts': '1418862834',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66',
                                                                                                                  'out_octets': '7127115027448',
                                                                                                                  'out_ucast_pkts': '5285953768',
                                                                                                                  'poller_hash': 'MzIyNzk3cGVlcmludGZwb24gMjE2'},
                                                                                                          {       'in_nucast_pkts': '32543862',
                                                                                                                  'in_octets': '1690109241243',
                                                                                                                  'in_ucast_pkts': '3547371711',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '43997775660',
                                                                                                                  'out_octets': '71623326631909',
                                                                                                                  'out_ucast_pkts': '9691899840',
                                                                                                                  'poller_hash': 'MzIyNkMycGVlcmludGZwb24gMjE3'},
                                                                                                          {       'in_nucast_pkts': '32',
                                                                                                                  'in_octets': '71610447730',
                                                                                                                  'in_ucast_pkts': '390575465',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652693',
                                                                                                                  'out_octets': '946182885898',
                                                                                                                  'out_ucast_pkts': '729778510',
                                                                                                                  'poller_hash': 'NEY5MUYwcGVlcmludGZwb24gMjE5'},
                                                                                                          {       'in_nucast_pkts': '31',
                                                                                                                  'in_octets': '142340447741',
                                                                                                                  'in_ucast_pkts': '414291333',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919908',
                                                                                                                  'out_octets': '1314647134830',
                                                                                                                  'out_ucast_pkts': '1146789602',
                                                                                                                  'poller_hash': 'NDI4MUZEcGVlcmludGZwb24gMjIx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2294703571564',
                                                                                                                  'in_ucast_pkts': '3794345190',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66',
                                                                                                                  'out_octets': '9567791286550',
                                                                                                                  'out_ucast_pkts': '8261558851',
                                                                                                                  'poller_hash': 'ODcyNTIycGVlcmludGZwb24gMjIy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NEY5RTlFcGVlcmludGZwb24gMjI0'},
                                                                                                          {       'in_nucast_pkts': '6588',
                                                                                                                  'in_octets': '1410704960725',
                                                                                                                  'in_ucast_pkts': '4023819357',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61709696',
                                                                                                                  'out_octets': '14321293692865',
                                                                                                                  'out_ucast_pkts': '11100070060',
                                                                                                                  'poller_hash': 'NDI4MTk4cGVlcmludGZwb24gMjI4'},
                                                                                                          {       'in_nucast_pkts': '12195',
                                                                                                                  'in_octets': '241755634412',
                                                                                                                  'in_ucast_pkts': '832105427',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61709889',
                                                                                                                  'out_octets': '2952698482277',
                                                                                                                  'out_ucast_pkts': '2647295141',
                                                                                                                  'poller_hash': 'NDI4Mjg0cGVlcmludGZwb24gMjI5'},
                                                                                                          {       'in_nucast_pkts': '24',
                                                                                                                  'in_octets': '140787309423',
                                                                                                                  'in_ucast_pkts': '355078263',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3659494',
                                                                                                                  'out_octets': '752629715511',
                                                                                                                  'out_ucast_pkts': '616810835',
                                                                                                                  'poller_hash': 'NDI4MUQ5cGVlcmludGZwb24gMjMw'},
                                                                                                          {       'in_nucast_pkts': '12302',
                                                                                                                  'in_octets': '1704393423703',
                                                                                                                  'in_ucast_pkts': '1918465953',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710726',
                                                                                                                  'out_octets': '2856687786995',
                                                                                                                  'out_ucast_pkts': '2847212928',
                                                                                                                  'poller_hash': 'NDI4MUI3cGVlcmludGZwb24gMjMy'},
                                                                                                          {       'in_nucast_pkts': '4916179',
                                                                                                                  'in_octets': '187338555597',
                                                                                                                  'in_ucast_pkts': '440055046',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9005784789',
                                                                                                                  'out_octets': '13749762109280',
                                                                                                                  'out_ucast_pkts': '1194461337',
                                                                                                                  'poller_hash': 'MzIyNjE4cGVlcmludGZwb24gMjMz'},
                                                                                                          {       'in_nucast_pkts': '29177819',
                                                                                                                  'in_octets': '369761641545',
                                                                                                                  'in_ucast_pkts': '3281954368',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '30004254048',
                                                                                                                  'out_octets': '59761364999224',
                                                                                                                  'out_ucast_pkts': '13375755828',
                                                                                                                  'poller_hash': 'NDI4MURFcGVlcmludGZwb24gMjM1'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '152853690239',
                                                                                                                  'in_ucast_pkts': '879314810',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2331746',
                                                                                                                  'out_octets': '2414286715935',
                                                                                                                  'out_ucast_pkts': '1796920590',
                                                                                                                  'poller_hash': 'NEZBMjVFcGVlcmludGZwb24gMjQ4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '260376264036',
                                                                                                                  'in_ucast_pkts': '1369711905',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33',
                                                                                                                  'out_octets': '5991563604220',
                                                                                                                  'out_ucast_pkts': '4205903070',
                                                                                                                  'poller_hash': 'NDI4MTc2cGVlcmludGZwb24gMjU3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'ODcyNDQ2cGVlcmludGZwb24gMjYy'},
                                                                                                          {       'in_nucast_pkts': '118',
                                                                                                                  'in_octets': '1150170018469',
                                                                                                                  'in_ucast_pkts': '5427475028',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39527115',
                                                                                                                  'out_octets': '18704562055365',
                                                                                                                  'out_ucast_pkts': '13385448717',
                                                                                                                  'poller_hash': 'NTEyNERDcGVlcmludGZwb24gMjY0'},
                                                                                                          {       'in_nucast_pkts': '154273',
                                                                                                                  'in_octets': '260774408683',
                                                                                                                  'in_ucast_pkts': '1485736261',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652396',
                                                                                                                  'out_octets': '4619540025931',
                                                                                                                  'out_ucast_pkts': '3370338574',
                                                                                                                  'poller_hash': 'NEY5QjBFcGVlcmludGZwb24gMjY5'},
                                                                                                          {       'in_nucast_pkts': '30',
                                                                                                                  'in_octets': '121222678585',
                                                                                                                  'in_ucast_pkts': '503826213',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919654',
                                                                                                                  'out_octets': '1230070811571',
                                                                                                                  'out_ucast_pkts': '879613787',
                                                                                                                  'poller_hash': 'NTc0NTkycGVlcmludGZwb24gMzA5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '551630199953',
                                                                                                                  'in_ucast_pkts': '1175969188',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33',
                                                                                                                  'out_octets': '3094687160092',
                                                                                                                  'out_ucast_pkts': '2678878379',
                                                                                                                  'poller_hash': 'NDJCOUVFcGVlcmludGZwb24gMzI1'},
                                                                                                          {       'in_nucast_pkts': '119039',
                                                                                                                  'in_octets': '203990533648',
                                                                                                                  'in_ucast_pkts': '708153419',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7938827',
                                                                                                                  'out_octets': '3036464927314',
                                                                                                                  'out_ucast_pkts': '2365852718',
                                                                                                                  'poller_hash': 'REZGRTA5cGVlcmludGZwb24gMzc4'},
                                                                                                          {       'in_nucast_pkts': '18',
                                                                                                                  'in_octets': '18055108729',
                                                                                                                  'in_ucast_pkts': '79989374',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1181098',
                                                                                                                  'out_octets': '374206752229',
                                                                                                                  'out_ucast_pkts': '266056156',
                                                                                                                  'poller_hash': 'NDJEMEY3cGVlcmludGZwb24gNDE0'},
                                                                                                          {       'in_nucast_pkts': '49107598',
                                                                                                                  'in_octets': '517966686345',
                                                                                                                  'in_ucast_pkts': '3428958706',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '64910274738',
                                                                                                                  'out_octets': '99367977926098',
                                                                                                                  'out_ucast_pkts': '8922844018',
                                                                                                                  'poller_hash': 'NDJEMkM4cGVlcmludGZwb24gNDE1'},
                                                                                                          {       'in_nucast_pkts': '94994',
                                                                                                                  'in_octets': '232618794397',
                                                                                                                  'in_ucast_pkts': '532483494',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5860941',
                                                                                                                  'out_octets': '2036967657283',
                                                                                                                  'out_ucast_pkts': '1524228709',
                                                                                                                  'poller_hash': 'QzgzRTEwcGVlcmludGZwb24gMTYyNDAz'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '140902469545',
                                                                                                                  'in_ucast_pkts': '500255063',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920246',
                                                                                                                  'out_octets': '2172033896839',
                                                                                                                  'out_ucast_pkts': '1588960581',
                                                                                                                  'poller_hash': 'MTI3NEIwNHBlZXJpbnRmcG9uIDIzNzc3Mw=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '213521619987',
                                                                                         'out_octets': '624329701687956',
                                                                                         'out_ucast_pkts': '258825419140',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/3',
                                                                                                                 'id': '6/3',
                                                                                                                 'links': [       {       'connected_time': '64119657',
                                                                                                                                          'fiber_length': 23367,
                                                                                                                                          'poller_hash': 'MzIyNjBEcGVlcnBvbnBvbiAyMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535206',
                                                                                                                                          'fiber_length': 23348,
                                                                                                                                          'poller_hash': 'NDI4MUVDcGVlcnBvbnBvbiAyMTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535316',
                                                                                                                                          'fiber_length': 23391,
                                                                                                                                          'poller_hash': 'MzIyNDFEcGVlcnBvbnBvbiAyMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535151',
                                                                                                                                          'fiber_length': 23678,
                                                                                                                                          'poller_hash': 'MzIyNzk3cGVlcnBvbnBvbiAyMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '26482727',
                                                                                                                                          'fiber_length': 23311,
                                                                                                                                          'poller_hash': 'MzIyNkMycGVlcnBvbnBvbiAyMTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535135',
                                                                                                                                          'fiber_length': 23512,
                                                                                                                                          'poller_hash': 'NEY5MUYwcGVlcnBvbnBvbiAyMTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535183',
                                                                                                                                          'fiber_length': 23121,
                                                                                                                                          'poller_hash': 'NDI4MUZEcGVlcnBvbnBvbiAyMjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535216',
                                                                                                                                          'fiber_length': 23047,
                                                                                                                                          'poller_hash': 'ODcyNTIycGVlcnBvbnBvbiAyMjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '145751',
                                                                                                                                          'fiber_length': 23806,
                                                                                                                                          'poller_hash': 'NEY5RTlFcGVlcnBvbnBvbiAyMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '21887200',
                                                                                                                                          'fiber_length': 23028,
                                                                                                                                          'poller_hash': 'NDI4MTk4cGVlcnBvbnBvbiAyMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '21886553',
                                                                                                                                          'fiber_length': 23254,
                                                                                                                                          'poller_hash': 'NDI4Mjg0cGVlcnBvbnBvbiAyMjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '3096859',
                                                                                                                                          'fiber_length': 23671,
                                                                                                                                          'poller_hash': 'NDI4MUQ5cGVlcnBvbnBvbiAyMzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '21886585',
                                                                                                                                          'fiber_length': 23237,
                                                                                                                                          'poller_hash': 'NDI4MUI3cGVlcnBvbnBvbiAyMzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535314',
                                                                                                                                          'fiber_length': 23132,
                                                                                                                                          'poller_hash': 'MzIyNjE4cGVlcnBvbnBvbiAyMzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '21886825',
                                                                                                                                          'fiber_length': 23761,
                                                                                                                                          'poller_hash': 'NDI4MURFcGVlcnBvbnBvbiAyMzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '3434993',
                                                                                                                                          'fiber_length': 23706,
                                                                                                                                          'poller_hash': 'NEZBMjVFcGVlcnBvbnBvbiAyNDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535140',
                                                                                                                                          'fiber_length': 23489,
                                                                                                                                          'poller_hash': 'NDI4MTc2cGVlcnBvbnBvbiAyNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '15879229',
                                                                                                                                          'fiber_length': 23348,
                                                                                                                                          'poller_hash': 'ODcyNDQ2cGVlcnBvbnBvbiAyNjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '67856213',
                                                                                                                                          'fiber_length': 23453,
                                                                                                                                          'poller_hash': 'NTEyNERDcGVlcnBvbnBvbiAyNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535089',
                                                                                                                                          'fiber_length': 22943,
                                                                                                                                          'poller_hash': 'NEY5QjBFcGVlcnBvbnBvbiAyNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535199',
                                                                                                                                          'fiber_length': 23400,
                                                                                                                                          'poller_hash': 'NTc0NTkycGVlcnBvbnBvbiAzMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535206',
                                                                                                                                          'fiber_length': 23395,
                                                                                                                                          'poller_hash': 'NDJCOUVFcGVlcnBvbnBvbiAzMjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '6812304',
                                                                                                                                          'fiber_length': 23696,
                                                                                                                                          'poller_hash': 'REZGRTA5cGVlcnBvbnBvbiAzNzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '1008584',
                                                                                                                                          'fiber_length': 22989,
                                                                                                                                          'poller_hash': 'NDJEMEY3cGVlcnBvbnBvbiA0MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '21886552',
                                                                                                                                          'fiber_length': 23245,
                                                                                                                                          'poller_hash': 'NDJEMkM4cGVlcnBvbnBvbiA0MTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '5010038',
                                                                                                                                          'fiber_length': 23111,
                                                                                                                                          'poller_hash': 'QzgzRTEwcGVlcnBvbnBvbiAxNjI0MDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362},
                                                                                                                                  {       'connected_time': '8535205',
                                                                                                                                          'fiber_length': 23465,
                                                                                                                                          'poller_hash': 'MTI3NEIwNHBlZXJwb25wb24gMjM3Nzcz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.362}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/3',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/4',
                                                                                         'in_discards': '39335',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1770242',
                                                                                         'in_octets': '22160736120852',
                                                                                         'in_ucast_pkts': '68716593196',
                                                                                         'links': [       {       'in_nucast_pkts': '32532242',
                                                                                                                  'in_octets': '518154680935',
                                                                                                                  'in_ucast_pkts': '1595454894',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '67134885314',
                                                                                                                  'out_octets': '97367048211337',
                                                                                                                  'out_ucast_pkts': '5284000330',
                                                                                                                  'poller_hash': 'NDJDNkVDcGVlcmludGZwb24gMTAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '84423209899',
                                                                                                                  'in_ucast_pkts': '560638385',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '231',
                                                                                                                  'out_octets': '3530546725969',
                                                                                                                  'out_ucast_pkts': '2435119595',
                                                                                                                  'poller_hash': 'NDI4MUI5cGVlcmludGZwb24gMjIw'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '14441323347',
                                                                                                                  'in_ucast_pkts': '69893348',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2362997',
                                                                                                                  'out_octets': '368723722569',
                                                                                                                  'out_ucast_pkts': '256339920',
                                                                                                                  'poller_hash': 'NDI4MTk3cGVlcmludGZwb24gMjM2'},
                                                                                                          {       'in_nucast_pkts': '27',
                                                                                                                  'in_octets': '28887905294',
                                                                                                                  'in_ucast_pkts': '90969710',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652485',
                                                                                                                  'out_octets': '239211381295',
                                                                                                                  'out_ucast_pkts': '207123168',
                                                                                                                  'poller_hash': 'NEY4RUQzcGVlcmludGZwb24gMjM5'},
                                                                                                          {       'in_nucast_pkts': '167483',
                                                                                                                  'in_octets': '1226125540256',
                                                                                                                  'in_ucast_pkts': '1240259292',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920033',
                                                                                                                  'out_octets': '1621406201164',
                                                                                                                  'out_ucast_pkts': '1621790692',
                                                                                                                  'poller_hash': 'NTEyNTE1cGVlcmludGZwb24gMjY3'},
                                                                                                          {       'in_nucast_pkts': '319021',
                                                                                                                  'in_octets': '704037193535',
                                                                                                                  'in_ucast_pkts': '4737130178',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61700374',
                                                                                                                  'out_octets': '13994780115607',
                                                                                                                  'out_ucast_pkts': '9785113326',
                                                                                                                  'poller_hash': 'MzIyOTRCcGVlcmludGZwb24gMjY4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '86755719787',
                                                                                                                  'in_ucast_pkts': '8878415643',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '198',
                                                                                                                  'out_octets': '997730991304',
                                                                                                                  'out_ucast_pkts': '827793379',
                                                                                                                  'poller_hash': 'NDJDRTk2cGVlcmludGZwb24gMjcx'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '149450869401',
                                                                                                                  'in_ucast_pkts': '451698719',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920240',
                                                                                                                  'out_octets': '1469902913306',
                                                                                                                  'out_ucast_pkts': '1181921932',
                                                                                                                  'poller_hash': 'NTEyNTRFcGVlcmludGZwb24gMjc2'},
                                                                                                          {       'in_nucast_pkts': '7948',
                                                                                                                  'in_octets': '280164423310',
                                                                                                                  'in_ucast_pkts': '3225566026',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710415',
                                                                                                                  'out_octets': '16755554756970',
                                                                                                                  'out_ucast_pkts': '11175415129',
                                                                                                                  'poller_hash': 'NDJCOUQwcGVlcmludGZwb24gMzkw'},
                                                                                                          {       'in_nucast_pkts': '10857593',
                                                                                                                  'in_octets': '162043522087',
                                                                                                                  'in_ucast_pkts': '768868836',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8365952946',
                                                                                                                  'out_octets': '15218162727406',
                                                                                                                  'out_ucast_pkts': '3469436704',
                                                                                                                  'poller_hash': 'NDJDODVBcGVlcmludGZwb24gMzk4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '223104530637',
                                                                                                                  'in_ucast_pkts': '740595716',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '4068733618660',
                                                                                                                  'out_ucast_pkts': '3022332186',
                                                                                                                  'poller_hash': 'NDJDRTVBcGVlcmludGZwb24gNDA4'},
                                                                                                          {       'in_nucast_pkts': '27',
                                                                                                                  'in_octets': '62687032766',
                                                                                                                  'in_ucast_pkts': '326663815',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920208',
                                                                                                                  'out_octets': '941723265249',
                                                                                                                  'out_ucast_pkts': '700072206',
                                                                                                                  'poller_hash': 'NDJEMTQycGVlcmludGZwb24gNDEx'},
                                                                                                          {       'in_nucast_pkts': '332345',
                                                                                                                  'in_octets': '590589083222',
                                                                                                                  'in_ucast_pkts': '3036545916',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '761107233',
                                                                                                                  'out_octets': '10721732486915',
                                                                                                                  'out_ucast_pkts': '6873261514',
                                                                                                                  'poller_hash': 'NDJEMDkwcGVlcmludGZwb24gNDE5'},
                                                                                                          {       'in_nucast_pkts': '56438',
                                                                                                                  'in_octets': '27793109446',
                                                                                                                  'in_ucast_pkts': '112063475',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1989091',
                                                                                                                  'out_octets': '418017435811',
                                                                                                                  'out_ucast_pkts': '313473004',
                                                                                                                  'poller_hash': 'NDJDOERFcGVlcmludGZwb24gNDIz'},
                                                                                                          {       'in_nucast_pkts': '278',
                                                                                                                  'in_octets': '37540621',
                                                                                                                  'in_ucast_pkts': '185839',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11536',
                                                                                                                  'out_octets': '422157049',
                                                                                                                  'out_ucast_pkts': '378424',
                                                                                                                  'poller_hash': 'NDNCNUU3cGVlcmludGZwb24gNDM0'},
                                                                                                          {       'in_nucast_pkts': '34',
                                                                                                                  'in_octets': '114337537081',
                                                                                                                  'in_ucast_pkts': '750004277',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '29470200',
                                                                                                                  'out_octets': '4335874206543',
                                                                                                                  'out_ucast_pkts': '2926787992',
                                                                                                                  'poller_hash': 'NDJCRjExcGVlcmludGZwb24gNDQz'},
                                                                                                          {       'in_nucast_pkts': '227564',
                                                                                                                  'in_octets': '6198432952',
                                                                                                                  'in_ucast_pkts': '31170972',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '222406166',
                                                                                                                  'out_octets': '446897503343',
                                                                                                                  'out_ucast_pkts': '108734804',
                                                                                                                  'poller_hash': 'NDJCRjMzcGVlcmludGZwb24gNDYw'},
                                                                                                          {       'in_nucast_pkts': '141',
                                                                                                                  'in_octets': '311520490746',
                                                                                                                  'in_ucast_pkts': '2785836559',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39527005',
                                                                                                                  'out_octets': '12314308058590',
                                                                                                                  'out_ucast_pkts': '8567026125',
                                                                                                                  'poller_hash': 'NUE2NUU5cGVlcmludGZwb24gNDY0'},
                                                                                                          {       'in_nucast_pkts': '1859468',
                                                                                                                  'in_octets': '60033063662',
                                                                                                                  'in_ucast_pkts': '488256102',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2029152321',
                                                                                                                  'out_octets': '4879478972413',
                                                                                                                  'out_ucast_pkts': '1576363659',
                                                                                                                  'poller_hash': 'NDJDNjlFcGVlcmludGZwb24gNDc3'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '58033472775',
                                                                                                                  'in_ucast_pkts': '215657642',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652646',
                                                                                                                  'out_octets': '1023844095139',
                                                                                                                  'out_ucast_pkts': '825224856',
                                                                                                                  'poller_hash': 'NjQwQjQxcGVlcmludGZwb24gNTAx'},
                                                                                                          {       'in_nucast_pkts': '27',
                                                                                                                  'in_octets': '173045548569',
                                                                                                                  'in_ucast_pkts': '891313518',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920096',
                                                                                                                  'out_octets': '3221959349356',
                                                                                                                  'out_ucast_pkts': '2317988603',
                                                                                                                  'poller_hash': 'NDJDRTM1cGVlcmludGZwb24gNTA0'},
                                                                                                          {       'in_nucast_pkts': '27',
                                                                                                                  'in_octets': '307134767661',
                                                                                                                  'in_ucast_pkts': '2296599863',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '36499690',
                                                                                                                  'out_octets': '8954576515906',
                                                                                                                  'out_ucast_pkts': '6363324909',
                                                                                                                  'poller_hash': 'NDJDNzExcGVlcmludGZwb24gNTI0'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '555954172256',
                                                                                                                  'in_ucast_pkts': '1274257704',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '28814182182',
                                                                                                                  'out_octets': '43336881553978',
                                                                                                                  'out_ucast_pkts': '3600323664',
                                                                                                                  'poller_hash': 'ODcyNTYycGVlcmludGZwb24gMTM4MTI4'},
                                                                                                          {       'in_nucast_pkts': '46',
                                                                                                                  'in_octets': '142497291691',
                                                                                                                  'in_ucast_pkts': '1486248217',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31007740',
                                                                                                                  'out_octets': '5885428015901',
                                                                                                                  'out_ucast_pkts': '4042997512',
                                                                                                                  'poller_hash': 'QzgzQkE1cGVlcmludGZwb24gMTcwNDQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '257035188635',
                                                                                                                  'in_ucast_pkts': '664118991',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33',
                                                                                                                  'out_octets': '1229006334579',
                                                                                                                  'out_ucast_pkts': '1152840181',
                                                                                                                  'poller_hash': 'MTI5NzQ2RHBlZXJpbnRmcG9uIDI1NzQ2OA=='},
                                                                                                          {       'in_nucast_pkts': '431',
                                                                                                                  'in_octets': '541895068977',
                                                                                                                  'in_ucast_pkts': '2499228261',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '25452890',
                                                                                                                  'out_octets': '12284720149306',
                                                                                                                  'out_ucast_pkts': '8747823592',
                                                                                                                  'poller_hash': 'MTI3NDc0OXBlZXJpbnRmcG9uIDI2MjQ2Nw=='},
                                                                                                          {       'in_nucast_pkts': '165888',
                                                                                                                  'in_octets': '667678572271',
                                                                                                                  'in_ucast_pkts': '2032871673',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920132',
                                                                                                                  'out_octets': '7078236192010',
                                                                                                                  'out_ucast_pkts': '5352329993',
                                                                                                                  'poller_hash': 'MTI3NDY2OHBlZXJpbnRmcG9uIDI2Mjk5NQ=='},
                                                                                                          {       'in_nucast_pkts': '93567',
                                                                                                                  'in_octets': '29631380238',
                                                                                                                  'in_ucast_pkts': '104188615',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5434622',
                                                                                                                  'out_octets': '1212959020481',
                                                                                                                  'out_ucast_pkts': '894717682',
                                                                                                                  'poller_hash': 'MTE5MzEwNHBlZXJpbnRmcG9uIDI3MjA4OQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '268783312745',
                                                                                         'out_octets': '614757562932682',
                                                                                         'out_ucast_pkts': '184032793702',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/4',
                                                                                                                 'id': '6/4',
                                                                                                                 'links': [       {       'connected_time': '21887230',
                                                                                                                                          'fiber_length': 23018,
                                                                                                                                          'poller_hash': 'NDJDNkVDcGVlcnBvbnBvbiAxMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '21886550',
                                                                                                                                          'fiber_length': 22630,
                                                                                                                                          'poller_hash': 'NDI4MUI5cGVlcnBvbnBvbiAyMjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '2000646',
                                                                                                                                          'fiber_length': 22318,
                                                                                                                                          'poller_hash': 'NDI4MTk3cGVlcnBvbnBvbiAyMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535106',
                                                                                                                                          'fiber_length': 22597,
                                                                                                                                          'poller_hash': 'NEY4RUQzcGVlcnBvbnBvbiAyMzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535507',
                                                                                                                                          'fiber_length': 23072,
                                                                                                                                          'poller_hash': 'NTEyNTE1cGVlcnBvbnBvbiAyNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '21886761',
                                                                                                                                          'fiber_length': 22664,
                                                                                                                                          'poller_hash': 'MzIyOTRCcGVlcnBvbnBvbiAyNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '21886973',
                                                                                                                                          'fiber_length': 23003,
                                                                                                                                          'poller_hash': 'NDJDRTk2cGVlcnBvbnBvbiAyNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535434',
                                                                                                                                          'fiber_length': 22270,
                                                                                                                                          'poller_hash': 'NTEyNTRFcGVlcnBvbnBvbiAyNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '21886519',
                                                                                                                                          'fiber_length': 22674,
                                                                                                                                          'poller_hash': 'NDJCOUQwcGVlcnBvbnBvbiAzOTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '21886968',
                                                                                                                                          'fiber_length': 22829,
                                                                                                                                          'poller_hash': 'NDJDODVBcGVlcnBvbnBvbiAzOTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '5596335',
                                                                                                                                          'fiber_length': 23096,
                                                                                                                                          'poller_hash': 'NDJDRTVBcGVlcnBvbnBvbiA0MDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535309',
                                                                                                                                          'fiber_length': 22618,
                                                                                                                                          'poller_hash': 'NDJEMTQycGVlcnBvbnBvbiA0MTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '13281075',
                                                                                                                                          'fiber_length': 22524,
                                                                                                                                          'poller_hash': 'NDJEMDkwcGVlcnBvbnBvbiA0MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '1693604',
                                                                                                                                          'fiber_length': 23100,
                                                                                                                                          'poller_hash': 'NDJDOERFcGVlcnBvbnBvbiA0MjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '33279',
                                                                                                                                          'fiber_length': 23041,
                                                                                                                                          'poller_hash': 'NDNCNUU3cGVlcnBvbnBvbiA0MzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '25498581',
                                                                                                                                          'fiber_length': 22900,
                                                                                                                                          'poller_hash': 'NDJCRjExcGVlcnBvbnBvbiA0NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '613208',
                                                                                                                                          'fiber_length': 22855,
                                                                                                                                          'poller_hash': 'NDJCRjMzcGVlcnBvbnBvbiA0NjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '67856372',
                                                                                                                                          'fiber_length': 22279,
                                                                                                                                          'poller_hash': 'NUE2NUU5cGVlcnBvbnBvbiA0NjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535199',
                                                                                                                                          'fiber_length': 23021,
                                                                                                                                          'poller_hash': 'NDJDNjlFcGVlcnBvbnBvbiA0Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535122',
                                                                                                                                          'fiber_length': 22267,
                                                                                                                                          'poller_hash': 'NjQwQjQxcGVlcnBvbnBvbiA1MDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535182',
                                                                                                                                          'fiber_length': 23018,
                                                                                                                                          'poller_hash': 'NDJDRTM1cGVlcnBvbnBvbiA1MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '64529577',
                                                                                                                                          'fiber_length': 23566,
                                                                                                                                          'poller_hash': 'NDJDNzExcGVlcnBvbnBvbiA1MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535249',
                                                                                                                                          'fiber_length': 23084,
                                                                                                                                          'poller_hash': 'ODcyNTYycGVlcnBvbnBvbiAxMzgxMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '28348050',
                                                                                                                                          'fiber_length': 22822,
                                                                                                                                          'poller_hash': 'QzgzQkE1cGVlcnBvbnBvbiAxNzA0NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '4377689',
                                                                                                                                          'fiber_length': 22110,
                                                                                                                                          'poller_hash': 'MTI5NzQ2RHBlZXJwb25wb24gMjU3NDY4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '21588514',
                                                                                                                                          'fiber_length': 22741,
                                                                                                                                          'poller_hash': 'MTI3NDc0OXBlZXJwb25wb24gMjYyNDY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '8535225',
                                                                                                                                          'fiber_length': 22455,
                                                                                                                                          'poller_hash': 'MTI3NDY2OHBlZXJwb25wb24gMjYyOTk1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148},
                                                                                                                                  {       'connected_time': '4637845',
                                                                                                                                          'fiber_length': 23042,
                                                                                                                                          'poller_hash': 'MTE5MzEwNHBlZXJwb25wb24gMjcyMDg5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.148}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/4',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/5',
                                                                                         'in_discards': '412425',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '4479769',
                                                                                         'in_octets': '18777296940724',
                                                                                         'in_ucast_pkts': '73328012879',
                                                                                         'links': [       {       'in_nucast_pkts': '40',
                                                                                                                  'in_octets': '112278511276',
                                                                                                                  'in_ucast_pkts': '394295437',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919585',
                                                                                                                  'out_octets': '1033497913579',
                                                                                                                  'out_ucast_pkts': '935612746',
                                                                                                                  'poller_hash': 'NTc0NEE5cGVlcmludGZwb24gMzEw'},
                                                                                                          {       'in_nucast_pkts': '155891',
                                                                                                                  'in_octets': '192453694129',
                                                                                                                  'in_ucast_pkts': '421999659',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5648042',
                                                                                                                  'out_octets': '1056115836004',
                                                                                                                  'out_ucast_pkts': '848927460',
                                                                                                                  'poller_hash': 'NDNCMzdBcGVlcmludGZwb24gMzEx'},
                                                                                                          {       'in_nucast_pkts': '19',
                                                                                                                  'in_octets': '111794116405',
                                                                                                                  'in_ucast_pkts': '754612646',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919989',
                                                                                                                  'out_octets': '2637239676143',
                                                                                                                  'out_ucast_pkts': '1830911167',
                                                                                                                  'poller_hash': 'NDJCOUVBcGVlcmludGZwb24gMzE1'},
                                                                                                          {       'in_nucast_pkts': '421',
                                                                                                                  'in_octets': '403905662401',
                                                                                                                  'in_ucast_pkts': '1682819898',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39526176',
                                                                                                                  'out_octets': '9883649868712',
                                                                                                                  'out_ucast_pkts': '7137187591',
                                                                                                                  'poller_hash': 'NTEyNTJCcGVlcmludGZwb24gMzE3'},
                                                                                                          {       'in_nucast_pkts': '145',
                                                                                                                  'in_octets': '77341729906',
                                                                                                                  'in_ucast_pkts': '305669884',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39522617',
                                                                                                                  'out_octets': '1172412397993',
                                                                                                                  'out_ucast_pkts': '989047704',
                                                                                                                  'poller_hash': 'NTEyNTZCcGVlcmludGZwb24gMzM1'},
                                                                                                          {       'in_nucast_pkts': '3',
                                                                                                                  'in_octets': '14378242505',
                                                                                                                  'in_ucast_pkts': '118234395',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '402850',
                                                                                                                  'out_octets': '308388953085',
                                                                                                                  'out_ucast_pkts': '217223946',
                                                                                                                  'poller_hash': 'NDNCNUREcGVlcmludGZwb24gMzM2'},
                                                                                                          {       'in_nucast_pkts': '35',
                                                                                                                  'in_octets': '72183496777',
                                                                                                                  'in_ucast_pkts': '334503841',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '21816983',
                                                                                                                  'out_octets': '2053762045514',
                                                                                                                  'out_ucast_pkts': '1597597550',
                                                                                                                  'poller_hash': 'NTEyNjExcGVlcmludGZwb24gMzM4'},
                                                                                                          {       'in_nucast_pkts': '56',
                                                                                                                  'in_octets': '32293661762',
                                                                                                                  'in_ucast_pkts': '127742949',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39526334',
                                                                                                                  'out_octets': '440360682677',
                                                                                                                  'out_ucast_pkts': '349443717',
                                                                                                                  'poller_hash': 'NUE2NjI5cGVlcmludGZwb24gMzQy'},
                                                                                                          {       'in_nucast_pkts': '27',
                                                                                                                  'in_octets': '641245502744',
                                                                                                                  'in_ucast_pkts': '1409913915',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6379675',
                                                                                                                  'out_octets': '4233632149208',
                                                                                                                  'out_ucast_pkts': '3087353679',
                                                                                                                  'poller_hash': 'NDJCOUY4cGVlcmludGZwb24gMzQz'},
                                                                                                          {       'in_nucast_pkts': '53290',
                                                                                                                  'in_octets': '1033937114482',
                                                                                                                  'in_ucast_pkts': '3444698930',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17551937',
                                                                                                                  'out_octets': '7580397206458',
                                                                                                                  'out_ucast_pkts': '6389954299',
                                                                                                                  'poller_hash': 'NDJCOUFFcGVlcmludGZwb24gMzQ0'},
                                                                                                          {       'in_nucast_pkts': '11890',
                                                                                                                  'in_octets': '410389281366',
                                                                                                                  'in_ucast_pkts': '1032947145',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61709666',
                                                                                                                  'out_octets': '4549038430748',
                                                                                                                  'out_ucast_pkts': '3719187930',
                                                                                                                  'poller_hash': 'NDJCOUVDcGVlcmludGZwb24gMzQ3'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '56753754818',
                                                                                                                  'in_ucast_pkts': '247570171',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4209037',
                                                                                                                  'out_octets': '825584830359',
                                                                                                                  'out_ucast_pkts': '594420935',
                                                                                                                  'poller_hash': 'NDNCNUJCcGVlcmludGZwb24gMzUy'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '106545910595',
                                                                                                                  'in_ucast_pkts': '536315113',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652436',
                                                                                                                  'out_octets': '3579724867828',
                                                                                                                  'out_ucast_pkts': '2544064392',
                                                                                                                  'poller_hash': 'NDNCNTE5cGVlcmludGZwb24gMzUz'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '127920355253',
                                                                                                                  'in_ucast_pkts': '650137115',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919934',
                                                                                                                  'out_octets': '3056361881564',
                                                                                                                  'out_ucast_pkts': '2178845460',
                                                                                                                  'poller_hash': 'NDJCOURFcGVlcmludGZwb24gMzYy'},
                                                                                                          {       'in_nucast_pkts': '155801',
                                                                                                                  'in_octets': '131124952411',
                                                                                                                  'in_ucast_pkts': '749153453',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919591',
                                                                                                                  'out_octets': '3704855820064',
                                                                                                                  'out_ucast_pkts': '2632022680',
                                                                                                                  'poller_hash': 'NDJCOUYzcGVlcmludGZwb24gMzYz'},
                                                                                                          {       'in_nucast_pkts': '15',
                                                                                                                  'in_octets': '19454443839',
                                                                                                                  'in_ucast_pkts': '116342791',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4990205',
                                                                                                                  'out_octets': '576733692438',
                                                                                                                  'out_ucast_pkts': '393132281',
                                                                                                                  'poller_hash': 'NDJCOUM0cGVlcmludGZwb24gMzY0'},
                                                                                                          {       'in_nucast_pkts': '90391',
                                                                                                                  'in_octets': '202426836702',
                                                                                                                  'in_ucast_pkts': '847487228',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6693966',
                                                                                                                  'out_octets': '3163003219404',
                                                                                                                  'out_ucast_pkts': '2566890195',
                                                                                                                  'poller_hash': 'NTc0NTcxcGVlcmludGZwb24gMzY2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '182561562837',
                                                                                                                  'in_ucast_pkts': '22636051321',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '51',
                                                                                                                  'out_octets': '3335643128505',
                                                                                                                  'out_ucast_pkts': '2398806813',
                                                                                                                  'poller_hash': 'ODY5MzlBcGVlcmludGZwb24gMzcw'},
                                                                                                          {       'in_nucast_pkts': '160787',
                                                                                                                  'in_octets': '56718658465',
                                                                                                                  'in_ucast_pkts': '512241206',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919499',
                                                                                                                  'out_octets': '1437097013841',
                                                                                                                  'out_ucast_pkts': '1009556370',
                                                                                                                  'poller_hash': 'NTc0MUEwcGVlcmludGZwb24gMzcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '771733879395',
                                                                                                                  'in_ucast_pkts': '6214741325',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '3132818345813',
                                                                                                                  'out_ucast_pkts': '3292400491',
                                                                                                                  'poller_hash': 'NDJCOUQ1cGVlcmludGZwb24gMzcy'},
                                                                                                          {       'in_nucast_pkts': '299038',
                                                                                                                  'in_octets': '312613246088',
                                                                                                                  'in_ucast_pkts': '659634998',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9917494',
                                                                                                                  'out_octets': '1975604921149',
                                                                                                                  'out_ucast_pkts': '1522798086',
                                                                                                                  'poller_hash': 'NDJCOUM5cGVlcmludGZwb24gMzc0'},
                                                                                                          {       'in_nucast_pkts': '76304',
                                                                                                                  'in_octets': '150425570888',
                                                                                                                  'in_ucast_pkts': '560899145',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4981246',
                                                                                                                  'out_octets': '1741492827496',
                                                                                                                  'out_ucast_pkts': '1257300402',
                                                                                                                  'poller_hash': 'NDJCOUU3cGVlcmludGZwb24gMzc2'},
                                                                                                          {       'in_nucast_pkts': '151281',
                                                                                                                  'in_octets': '111210166656',
                                                                                                                  'in_ucast_pkts': '432610407',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919608',
                                                                                                                  'out_octets': '1810494385670',
                                                                                                                  'out_ucast_pkts': '1372271308',
                                                                                                                  'poller_hash': 'NTc0MjQ2cGVlcmludGZwb24gMzc5'},
                                                                                                          {       'in_nucast_pkts': '9618779',
                                                                                                                  'in_octets': '205036598685',
                                                                                                                  'in_ucast_pkts': '669598223',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919939',
                                                                                                                  'out_octets': '1803606031454',
                                                                                                                  'out_ucast_pkts': '1270319832',
                                                                                                                  'poller_hash': 'NTEyNTA5cGVlcmludGZwb24gMzg0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJDRTkzcGVlcmludGZwb24gMTU2NTM1'},
                                                                                                          {       'in_nucast_pkts': '155179',
                                                                                                                  'in_octets': '11327630304',
                                                                                                                  'in_ucast_pkts': '41524215',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919972',
                                                                                                                  'out_octets': '111189271329',
                                                                                                                  'out_ucast_pkts': '85033603',
                                                                                                                  'poller_hash': 'MTE5NTE1OHBlZXJpbnRmcG9uIDIxMjk3OQ=='},
                                                                                                          {       'in_nucast_pkts': '302122',
                                                                                                                  'in_octets': '143183121206',
                                                                                                                  'in_ucast_pkts': '881608705',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '20571531',
                                                                                                                  'out_octets': '4065877091367',
                                                                                                                  'out_ucast_pkts': '2782378938',
                                                                                                                  'poller_hash': 'MTI3NDdENXBlZXJpbnRmcG9uIDI2NjE2OA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '189516955743',
                                                                                         'out_octets': '508235246520436',
                                                                                         'out_ucast_pkts': '186223149583',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/5',
                                                                                                                 'id': '6/5',
                                                                                                                 'links': [       {       'connected_time': '8535200',
                                                                                                                                          'fiber_length': 25017,
                                                                                                                                          'poller_hash': 'NTc0NEE5cGVlcnBvbnBvbiAzMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535039',
                                                                                                                                          'fiber_length': 24421,
                                                                                                                                          'poller_hash': 'NDNCMzdBcGVlcnBvbnBvbiAzMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535212',
                                                                                                                                          'fiber_length': 24412,
                                                                                                                                          'poller_hash': 'NDJCOUVBcGVlcnBvbnBvbiAzMTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '67856370',
                                                                                                                                          'fiber_length': 24468,
                                                                                                                                          'poller_hash': 'NTEyNTJCcGVlcnBvbnBvbiAzMTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '67856271',
                                                                                                                                          'fiber_length': 24650,
                                                                                                                                          'poller_hash': 'NTEyNTZCcGVlcnBvbnBvbiAzMzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '600542',
                                                                                                                                          'fiber_length': 24269,
                                                                                                                                          'poller_hash': 'NDNCNUREcGVlcnBvbnBvbiAzMzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '18480391',
                                                                                                                                          'fiber_length': 24981,
                                                                                                                                          'poller_hash': 'NTEyNjExcGVlcnBvbnBvbiAzMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '67856405',
                                                                                                                                          'fiber_length': 24801,
                                                                                                                                          'poller_hash': 'NUE2NjI5cGVlcnBvbnBvbiAzNDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '5444113',
                                                                                                                                          'fiber_length': 24534,
                                                                                                                                          'poller_hash': 'NDJCOUY4cGVlcnBvbnBvbiAzNDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '27206082',
                                                                                                                                          'fiber_length': 25368,
                                                                                                                                          'poller_hash': 'NDJCOUFFcGVlcnBvbnBvbiAzNDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '21887048',
                                                                                                                                          'fiber_length': 24829,
                                                                                                                                          'poller_hash': 'NDJCOUVDcGVlcnBvbnBvbiAzNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '6314499',
                                                                                                                                          'fiber_length': 24149,
                                                                                                                                          'poller_hash': 'NDNCNUJCcGVlcnBvbnBvbiAzNTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535063',
                                                                                                                                          'fiber_length': 25343,
                                                                                                                                          'poller_hash': 'NDNCNTE5cGVlcnBvbnBvbiAzNTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535224',
                                                                                                                                          'fiber_length': 24586,
                                                                                                                                          'poller_hash': 'NDJCOURFcGVlcnBvbnBvbiAzNjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535184',
                                                                                                                                          'fiber_length': 24443,
                                                                                                                                          'poller_hash': 'NDJCOUYzcGVlcnBvbnBvbiAzNjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '4250988',
                                                                                                                                          'fiber_length': 25293,
                                                                                                                                          'poller_hash': 'NDJCOUM0cGVlcnBvbnBvbiAzNjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535249',
                                                                                                                                          'fiber_length': 24686,
                                                                                                                                          'poller_hash': 'NTc0NTcxcGVlcnBvbnBvbiAzNjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535305',
                                                                                                                                          'fiber_length': 25120,
                                                                                                                                          'poller_hash': 'ODY5MzlBcGVlcnBvbnBvbiAzNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535115',
                                                                                                                                          'fiber_length': 25180,
                                                                                                                                          'poller_hash': 'NTc0MUEwcGVlcnBvbnBvbiAzNzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '5498457',
                                                                                                                                          'fiber_length': 25373,
                                                                                                                                          'poller_hash': 'NDJCOUQ1cGVlcnBvbnBvbiAzNzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535251',
                                                                                                                                          'fiber_length': 24639,
                                                                                                                                          'poller_hash': 'NDJCOUM5cGVlcnBvbnBvbiAzNzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '4243188',
                                                                                                                                          'fiber_length': 24838,
                                                                                                                                          'poller_hash': 'NDJCOUU3cGVlcnBvbnBvbiAzNzY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535107',
                                                                                                                                          'fiber_length': 24177,
                                                                                                                                          'poller_hash': 'NTc0MjQ2cGVlcnBvbnBvbiAzNzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535385',
                                                                                                                                          'fiber_length': 24928,
                                                                                                                                          'poller_hash': 'NTEyNTA5cGVlcnBvbnBvbiAzODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '6026859',
                                                                                                                                          'fiber_length': 24517,
                                                                                                                                          'poller_hash': 'NDJDRTkzcGVlcnBvbnBvbiAxNTY1MzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '8535268',
                                                                                                                                          'fiber_length': 25386,
                                                                                                                                          'poller_hash': 'MTE5NTE1OHBlZXJwb25wb24gMjEyOTc5',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202},
                                                                                                                                  {       'connected_time': '17373731',
                                                                                                                                          'fiber_length': 25300,
                                                                                                                                          'poller_hash': 'MTI3NDdENXBlZXJwb25wb24gMjY2MTY4',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.202}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/5',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/6',
                                                                                         'in_discards': '31378',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '4537407',
                                                                                         'in_octets': '27789514374321',
                                                                                         'in_ucast_pkts': '74881125317',
                                                                                         'links': [       {       'in_nucast_pkts': '155944',
                                                                                                                  'in_octets': '178200524791',
                                                                                                                  'in_ucast_pkts': '638656477',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652551',
                                                                                                                  'out_octets': '2548673535317',
                                                                                                                  'out_ucast_pkts': '1934248442',
                                                                                                                  'poller_hash': 'NDNCM0E2cGVlcmludGZwb24gMzA3'},
                                                                                                          {       'in_nucast_pkts': '20462',
                                                                                                                  'in_octets': '895982562897',
                                                                                                                  'in_ucast_pkts': '1519472987',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '17859324',
                                                                                                                  'out_octets': '4186803094924',
                                                                                                                  'out_ucast_pkts': '3534011745',
                                                                                                                  'poller_hash': 'QzgxRTA3cGVlcmludGZwb24gMzE2'},
                                                                                                          {       'in_nucast_pkts': '17',
                                                                                                                  'in_octets': '37599419752',
                                                                                                                  'in_ucast_pkts': '133318278',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3388796',
                                                                                                                  'out_octets': '369222482907',
                                                                                                                  'out_ucast_pkts': '291334427',
                                                                                                                  'poller_hash': 'NDJCOUM3cGVlcmludGZwb24gMzE4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '235190663320',
                                                                                                                  'in_ucast_pkts': '658250763',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66',
                                                                                                                  'out_octets': '2434597226697',
                                                                                                                  'out_ucast_pkts': '1882412849',
                                                                                                                  'poller_hash': 'QjA4REE1cGVlcmludGZwb24gMzIx'},
                                                                                                          {       'in_nucast_pkts': '1250',
                                                                                                                  'in_octets': '40871036399',
                                                                                                                  'in_ucast_pkts': '177383935',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5252740',
                                                                                                                  'out_octets': '986242722940',
                                                                                                                  'out_ucast_pkts': '721239998',
                                                                                                                  'poller_hash': 'NDNCM0QxcGVlcmludGZwb24gMzIz'},
                                                                                                          {       'in_nucast_pkts': '10324',
                                                                                                                  'in_octets': '1078792504271',
                                                                                                                  'in_ucast_pkts': '2311393346',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652446',
                                                                                                                  'out_octets': '3942425473272',
                                                                                                                  'out_ucast_pkts': '3549730090',
                                                                                                                  'poller_hash': 'NDNCNUI1cGVlcmludGZwb24gMzM0'},
                                                                                                          {       'in_nucast_pkts': '47945',
                                                                                                                  'in_octets': '4080793336962',
                                                                                                                  'in_ucast_pkts': '9742499301',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61587601',
                                                                                                                  'out_octets': '34700162501208',
                                                                                                                  'out_ucast_pkts': '26595722428',
                                                                                                                  'poller_hash': 'QTlDODZBcGVlcmludGZwb24gMzQ2'},
                                                                                                          {       'in_nucast_pkts': '68',
                                                                                                                  'in_octets': '87603156686',
                                                                                                                  'in_ucast_pkts': '190399856',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39526646',
                                                                                                                  'out_octets': '567994523383',
                                                                                                                  'out_ucast_pkts': '471117693',
                                                                                                                  'poller_hash': 'NTEyM0NCcGVlcmludGZwb24gMzQ5'},
                                                                                                          {       'in_nucast_pkts': '6764',
                                                                                                                  'in_octets': '551556800415',
                                                                                                                  'in_ucast_pkts': '2457508550',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '37060612093',
                                                                                                                  'out_octets': '60598619850571',
                                                                                                                  'out_ucast_pkts': '7804794970',
                                                                                                                  'poller_hash': 'NDJCOUNBcGVlcmludGZwb24gMzU1'},
                                                                                                          {       'in_nucast_pkts': '13595633',
                                                                                                                  'in_octets': '90924287481',
                                                                                                                  'in_ucast_pkts': '587347236',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33710156433',
                                                                                                                  'out_octets': '48914961883857',
                                                                                                                  'out_ucast_pkts': '2923619037',
                                                                                                                  'poller_hash': 'NDJCOUQxcGVlcmludGZwb24gMzU4'},
                                                                                                          {       'in_nucast_pkts': '23',
                                                                                                                  'in_octets': '279864872167',
                                                                                                                  'in_ucast_pkts': '527670432',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919775',
                                                                                                                  'out_octets': '1656795594539',
                                                                                                                  'out_ucast_pkts': '1319955151',
                                                                                                                  'poller_hash': 'NDJDNEE3cGVlcmludGZwb24gMzY4'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '22540390309',
                                                                                                                  'in_ucast_pkts': '98175383',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '18003767403',
                                                                                                                  'out_octets': '24416436408345',
                                                                                                                  'out_ucast_pkts': '237235799',
                                                                                                                  'poller_hash': 'NDJCOUMxcGVlcmludGZwb24gMzc1'},
                                                                                                          {       'in_nucast_pkts': '65889636',
                                                                                                                  'in_octets': '1115208762127',
                                                                                                                  'in_ucast_pkts': '4352598238',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '91394784286',
                                                                                                                  'out_octets': '144144325086147',
                                                                                                                  'out_ucast_pkts': '16355846107',
                                                                                                                  'poller_hash': 'NDJCOURCcGVlcmludGZwb24gMzgw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJDRDdCcGVlcmludGZwb24gNDI1'},
                                                                                                          {       'in_nucast_pkts': '22',
                                                                                                                  'in_octets': '200970507143',
                                                                                                                  'in_ucast_pkts': '1512392458',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919768',
                                                                                                                  'out_octets': '5433905626616',
                                                                                                                  'out_ucast_pkts': '3816439181',
                                                                                                                  'poller_hash': 'NDJDRTVEcGVlcmludGZwb24gNDI5'},
                                                                                                          {       'in_nucast_pkts': '5300',
                                                                                                                  'in_octets': '649030630552',
                                                                                                                  'in_ucast_pkts': '807873636',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5301293',
                                                                                                                  'out_octets': '3815881492979',
                                                                                                                  'out_ucast_pkts': '2904689831',
                                                                                                                  'poller_hash': 'NDNCNTM0cGVlcmludGZwb24gNDMw'},
                                                                                                          {       'in_nucast_pkts': '153815',
                                                                                                                  'in_octets': '136248610904',
                                                                                                                  'in_ucast_pkts': '239703206',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919772',
                                                                                                                  'out_octets': '1634086026013',
                                                                                                                  'out_ucast_pkts': '1218883657',
                                                                                                                  'poller_hash': 'NDJEMDgycGVlcmludGZwb24gNDM4'},
                                                                                                          {       'in_nucast_pkts': '52',
                                                                                                                  'in_octets': '132985579019',
                                                                                                                  'in_ucast_pkts': '300417286',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5651745',
                                                                                                                  'out_octets': '1404222022465',
                                                                                                                  'out_ucast_pkts': '1135535621',
                                                                                                                  'poller_hash': 'NjQwQjEycGVlcmludGZwb24gMTM3NjE0'},
                                                                                                          {       'in_nucast_pkts': '18',
                                                                                                                  'in_octets': '279016136852',
                                                                                                                  'in_ucast_pkts': '1647090306',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1256',
                                                                                                                  'out_octets': '11311993584652',
                                                                                                                  'out_ucast_pkts': '7937262607',
                                                                                                                  'poller_hash': 'NDJDOEREcGVlcmludGZwb24gMTM3Nzkw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NjQwMEVDcGVlcmludGZwb24gMTM3ODEw'},
                                                                                                          {       'in_nucast_pkts': '22',
                                                                                                                  'in_octets': '136023182539',
                                                                                                                  'in_ucast_pkts': '955313947',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919992',
                                                                                                                  'out_octets': '3732609089484',
                                                                                                                  'out_ucast_pkts': '2664567270',
                                                                                                                  'poller_hash': 'ODY5NDBEcGVlcmludGZwb24gMTQwNzMy'},
                                                                                                          {       'in_nucast_pkts': '26682852',
                                                                                                                  'in_octets': '540786640806',
                                                                                                                  'in_ucast_pkts': '1557460295',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61923185955',
                                                                                                                  'out_octets': '92482202815400',
                                                                                                                  'out_ucast_pkts': '7023753504',
                                                                                                                  'poller_hash': 'OTU2RUZFcGVlcmludGZwb24gMTQ2MTYw'},
                                                                                                          {       'in_nucast_pkts': '468455',
                                                                                                                  'in_octets': '497622876606',
                                                                                                                  'in_ucast_pkts': '2114067570',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '29896962',
                                                                                                                  'out_octets': '12723172969411',
                                                                                                                  'out_ucast_pkts': '8973222816',
                                                                                                                  'poller_hash': 'REZGRDBDcGVlcmludGZwb24gMTg2NjM3'},
                                                                                                          {       'in_nucast_pkts': '76465',
                                                                                                                  'in_octets': '118857771353',
                                                                                                                  'in_ucast_pkts': '417007415',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920204',
                                                                                                                  'out_octets': '2672911696485',
                                                                                                                  'out_ucast_pkts': '1939229032',
                                                                                                                  'poller_hash': 'QzgxRUU3cGVlcmludGZwb24gMjAyNzc0'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '311721555585',
                                                                                         'out_octets': '699512433074207',
                                                                                         'out_ucast_pkts': '209917255972',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/6',
                                                                                                                 'id': '6/6',
                                                                                                                 'links': [       {       'connected_time': '8535106',
                                                                                                                                          'fiber_length': 24304,
                                                                                                                                          'poller_hash': 'NDNCM0E2cGVlcnBvbnBvbiAzMDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '15182577',
                                                                                                                                          'fiber_length': 24726,
                                                                                                                                          'poller_hash': 'QzgxRTA3cGVlcnBvbnBvbiAzMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '2866710',
                                                                                                                                          'fiber_length': 24565,
                                                                                                                                          'poller_hash': 'NDJCOUM3cGVlcnBvbnBvbiAzMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '1035063',
                                                                                                                                          'fiber_length': 24561,
                                                                                                                                          'poller_hash': 'QjA4REE1cGVlcnBvbnBvbiAzMjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535090',
                                                                                                                                          'fiber_length': 24572,
                                                                                                                                          'poller_hash': 'NDNCM0QxcGVlcnBvbnBvbiAzMjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535067',
                                                                                                                                          'fiber_length': 24469,
                                                                                                                                          'poller_hash': 'NDNCNUI1cGVlcnBvbnBvbiAzMzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '21677433',
                                                                                                                                          'fiber_length': 24889,
                                                                                                                                          'poller_hash': 'QTlDODZBcGVlcnBvbnBvbiAzNDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '67856140',
                                                                                                                                          'fiber_length': 24982,
                                                                                                                                          'poller_hash': 'NTEyM0NCcGVlcnBvbnBvbiAzNDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '21886454',
                                                                                                                                          'fiber_length': 24609,
                                                                                                                                          'poller_hash': 'NDJCOUNBcGVlcnBvbnBvbiAzNTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '21886715',
                                                                                                                                          'fiber_length': 24531,
                                                                                                                                          'poller_hash': 'NDJCOUQxcGVlcnBvbnBvbiAzNTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535238',
                                                                                                                                          'fiber_length': 24777,
                                                                                                                                          'poller_hash': 'NDJDNEE3cGVlcnBvbnBvbiAzNjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535235',
                                                                                                                                          'fiber_length': 24614,
                                                                                                                                          'poller_hash': 'NDJCOUMxcGVlcnBvbnBvbiAzNzU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '21886574',
                                                                                                                                          'fiber_length': 25054,
                                                                                                                                          'poller_hash': 'NDJCOURCcGVlcnBvbnBvbiAzODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '1018160',
                                                                                                                                          'fiber_length': 24541,
                                                                                                                                          'poller_hash': 'NDJDRDdCcGVlcnBvbnBvbiA0MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535270',
                                                                                                                                          'fiber_length': 24691,
                                                                                                                                          'poller_hash': 'NDJDRTVEcGVlcnBvbnBvbiA0Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '7997672',
                                                                                                                                          'fiber_length': 24716,
                                                                                                                                          'poller_hash': 'NDNCNTM0cGVlcnBvbnBvbiA0MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535231',
                                                                                                                                          'fiber_length': 24512,
                                                                                                                                          'poller_hash': 'NDJEMDgycGVlcnBvbnBvbiA0Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535105',
                                                                                                                                          'fiber_length': 24863,
                                                                                                                                          'poller_hash': 'NjQwQjEycGVlcnBvbnBvbiAxMzc2MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '31260441',
                                                                                                                                          'fiber_length': 25577,
                                                                                                                                          'poller_hash': 'NDJDOEREcGVlcnBvbnBvbiAxMzc3OTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535101',
                                                                                                                                          'fiber_length': 24922,
                                                                                                                                          'poller_hash': 'NjQwMEVDcGVlcnBvbnBvbiAxMzc4MTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535296',
                                                                                                                                          'fiber_length': 25071,
                                                                                                                                          'poller_hash': 'ODY5NDBEcGVlcnBvbnBvbiAxNDA3MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '21886847',
                                                                                                                                          'fiber_length': 24555,
                                                                                                                                          'poller_hash': 'OTU2RUZFcGVlcnBvbnBvbiAxNDYxNjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '25916541',
                                                                                                                                          'fiber_length': 24714,
                                                                                                                                          'poller_hash': 'REZGRDBDcGVlcnBvbnBvbiAxODY2Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232},
                                                                                                                                  {       'connected_time': '8535278',
                                                                                                                                          'fiber_length': 24476,
                                                                                                                                          'poller_hash': 'QzgxRUU3cGVlcnBvbnBvbiAyMDI3NzQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.232}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/6',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/7',
                                                                                         'in_discards': '26040',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2664165',
                                                                                         'in_octets': '40324605108281',
                                                                                         'in_ucast_pkts': '98158222539',
                                                                                         'links': [       {       'in_nucast_pkts': '7170',
                                                                                                                  'in_octets': '1049144975762',
                                                                                                                  'in_ucast_pkts': '4686368063',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710177',
                                                                                                                  'out_octets': '16061328771624',
                                                                                                                  'out_ucast_pkts': '12820667353',
                                                                                                                  'poller_hash': 'NDI4MUE5cGVlcmludGZwb24gMjQx'},
                                                                                                          {       'in_nucast_pkts': '21',
                                                                                                                  'in_octets': '193013881571',
                                                                                                                  'in_ucast_pkts': '411528763',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919916',
                                                                                                                  'out_octets': '1109254622584',
                                                                                                                  'out_ucast_pkts': '862253265',
                                                                                                                  'poller_hash': 'NDI4MjVCcGVlcmludGZwb24gMjQz'},
                                                                                                          {       'in_nucast_pkts': '22',
                                                                                                                  'in_octets': '276103108224',
                                                                                                                  'in_ucast_pkts': '1357566547',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920062',
                                                                                                                  'out_octets': '5050044210207',
                                                                                                                  'out_ucast_pkts': '3743928671',
                                                                                                                  'poller_hash': 'NDI4MUFDcGVlcmludGZwb24gMjQ1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '2592672738120',
                                                                                                                  'in_ucast_pkts': '16404090883',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '9406831688465',
                                                                                                                  'out_ucast_pkts': '8400366891',
                                                                                                                  'poller_hash': 'NDI4MUI2cGVlcmludGZwb24gMjQ5'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '19228690660',
                                                                                                                  'in_ucast_pkts': '146250278',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4018055',
                                                                                                                  'out_octets': '685910199226',
                                                                                                                  'out_ucast_pkts': '480939741',
                                                                                                                  'poller_hash': 'NEY4RENEcGVlcmludGZwb24gMjUx'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '125673038378',
                                                                                                                  'in_ucast_pkts': '808820704',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652649',
                                                                                                                  'out_octets': '2855127453646',
                                                                                                                  'out_ucast_pkts': '2083654024',
                                                                                                                  'poller_hash': 'NEY5MUNFcGVlcmludGZwb24gMjUy'},
                                                                                                          {       'in_nucast_pkts': '8006',
                                                                                                                  'in_octets': '3181406187200',
                                                                                                                  'in_ucast_pkts': '8856270689',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710072',
                                                                                                                  'out_octets': '25443060721841',
                                                                                                                  'out_ucast_pkts': '20231189307',
                                                                                                                  'poller_hash': 'NDI4MUQ4cGVlcmludGZwb24gMjU0'},
                                                                                                          {       'in_nucast_pkts': '5',
                                                                                                                  'in_octets': '443796890276',
                                                                                                                  'in_ucast_pkts': '1153506289',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4978630',
                                                                                                                  'out_octets': '3541351295562',
                                                                                                                  'out_ucast_pkts': '2787431191',
                                                                                                                  'poller_hash': 'NDJDNzBDcGVlcmludGZwb24gMjU1'},
                                                                                                          {       'in_nucast_pkts': '413175',
                                                                                                                  'in_octets': '271940403769',
                                                                                                                  'in_ucast_pkts': '1076875762',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10323303',
                                                                                                                  'out_octets': '4564708903532',
                                                                                                                  'out_ucast_pkts': '3375960212',
                                                                                                                  'poller_hash': 'NDI4MUMxcGVlcmludGZwb24gMjU2'},
                                                                                                          {       'in_nucast_pkts': '28',
                                                                                                                  'in_octets': '1344390113796',
                                                                                                                  'in_ucast_pkts': '2004960612',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919750',
                                                                                                                  'out_octets': '4050197410996',
                                                                                                                  'out_ucast_pkts': '3044233065',
                                                                                                                  'poller_hash': 'NDI4MUM3cGVlcmludGZwb24gMjYw'},
                                                                                                          {       'in_nucast_pkts': '76436',
                                                                                                                  'in_octets': '9424815391',
                                                                                                                  'in_ucast_pkts': '38123583',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '4587488',
                                                                                                                  'out_octets': '112217381435',
                                                                                                                  'out_ucast_pkts': '83266085',
                                                                                                                  'poller_hash': 'NDI4MTkzcGVlcmludGZwb24gMjYx'},
                                                                                                          {       'in_nucast_pkts': '38526355',
                                                                                                                  'in_octets': '698490571368',
                                                                                                                  'in_ucast_pkts': '1782748462',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8706060590',
                                                                                                                  'out_octets': '16837061832195',
                                                                                                                  'out_ucast_pkts': '4671973893',
                                                                                                                  'poller_hash': 'MzIyODc2cGVlcmludGZwb24gMjY1'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NEZBMTc3cGVlcmludGZwb24gMjcw'},
                                                                                                          {       'in_nucast_pkts': '100',
                                                                                                                  'in_octets': '276441147928',
                                                                                                                  'in_ucast_pkts': '771257156',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '39508027',
                                                                                                                  'out_octets': '3668292506795',
                                                                                                                  'out_ucast_pkts': '2952043956',
                                                                                                                  'poller_hash': 'NTEyNTk1cGVlcmludGZwb24gMjc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NTEyNEZBcGVlcmludGZwb24gMjc5'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '21321469025',
                                                                                                                  'in_ucast_pkts': '62612936',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652598',
                                                                                                                  'out_octets': '287702920975',
                                                                                                                  'out_ucast_pkts': '239475325',
                                                                                                                  'poller_hash': 'NEZBMTlDcGVlcmludGZwb24gMjkz'},
                                                                                                          {       'in_nucast_pkts': '7437',
                                                                                                                  'in_octets': '331576334480',
                                                                                                                  'in_ucast_pkts': '1725070538',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710363',
                                                                                                                  'out_octets': '7244642611165',
                                                                                                                  'out_ucast_pkts': '5343141845',
                                                                                                                  'poller_hash': 'NDJCOUEwcGVlcmludGZwb24gMzAz'},
                                                                                                          {       'in_nucast_pkts': '151',
                                                                                                                  'in_octets': '3351200541',
                                                                                                                  'in_ucast_pkts': '11816975',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '59782',
                                                                                                                  'out_octets': '66167538558',
                                                                                                                  'out_ucast_pkts': '47704451',
                                                                                                                  'poller_hash': 'NDNCNUZFcGVlcmludGZwb24gMzMx'},
                                                                                                          {       'in_nucast_pkts': '62870507',
                                                                                                                  'in_octets': '820229080988',
                                                                                                                  'in_ucast_pkts': '1898873729',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12338486013',
                                                                                                                  'out_octets': '22517709420466',
                                                                                                                  'out_ucast_pkts': '5920481254',
                                                                                                                  'poller_hash': 'NDJCOUJDcGVlcmludGZwb24gMzQx'},
                                                                                                          {       'in_nucast_pkts': '12117',
                                                                                                                  'in_octets': '272163044937',
                                                                                                                  'in_ucast_pkts': '1110896093',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710909',
                                                                                                                  'out_octets': '5460782842977',
                                                                                                                  'out_ucast_pkts': '4204060899',
                                                                                                                  'poller_hash': 'NDJCOUM4cGVlcmludGZwb24gMzcz'},
                                                                                                          {       'in_nucast_pkts': '382622',
                                                                                                                  'in_octets': '163567419730',
                                                                                                                  'in_ucast_pkts': '486046307',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5657731',
                                                                                                                  'out_octets': '1472339523740',
                                                                                                                  'out_ucast_pkts': '1177637716',
                                                                                                                  'poller_hash': 'NDNCNUQ3cGVlcmludGZwb24gMzkz'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '94993782314',
                                                                                                                  'in_ucast_pkts': '102717716',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '976234',
                                                                                                                  'out_octets': '103829153232',
                                                                                                                  'out_ucast_pkts': '115570354',
                                                                                                                  'poller_hash': 'NDNCM0RDcGVlcmludGZwb24gNDA0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1333255693325',
                                                                                                                  'in_ucast_pkts': '2587847511',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '99',
                                                                                                                  'out_octets': '9699658278209',
                                                                                                                  'out_ucast_pkts': '7626726221',
                                                                                                                  'poller_hash': 'NDJDRTI2cGVlcmludGZwb24gNDA1'},
                                                                                                          {       'in_nucast_pkts': '6473499',
                                                                                                                  'in_octets': '52854217759',
                                                                                                                  'in_ucast_pkts': '442236921',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10655081701',
                                                                                                                  'out_octets': '16033687491313',
                                                                                                                  'out_ucast_pkts': '1133637356',
                                                                                                                  'poller_hash': 'NDJDRTQycGVlcmludGZwb24gNDA2'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '15216400074',
                                                                                                                  'in_ucast_pkts': '147997824',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5373005',
                                                                                                                  'out_octets': '712592265621',
                                                                                                                  'out_ucast_pkts': '485482306',
                                                                                                                  'poller_hash': 'NDJDOEY5cGVlcmludGZwb24gNDIw'},
                                                                                                          {       'in_nucast_pkts': '181',
                                                                                                                  'in_octets': '23860531848',
                                                                                                                  'in_ucast_pkts': '222715740',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '13227391',
                                                                                                                  'out_octets': '1279780006649',
                                                                                                                  'out_ucast_pkts': '864212088',
                                                                                                                  'poller_hash': 'NDJDMUQ1cGVlcmludGZwb24gNDM5'},
                                                                                                          {       'in_nucast_pkts': '18602122',
                                                                                                                  'in_octets': '69111849648',
                                                                                                                  'in_ucast_pkts': '360544025',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '463539536',
                                                                                                                  'out_octets': '2534568400090',
                                                                                                                  'out_ucast_pkts': '1444009534',
                                                                                                                  'poller_hash': 'NDJCRjE3cGVlcmludGZwb24gNDQx'},
                                                                                                          {       'in_nucast_pkts': '57249',
                                                                                                                  'in_octets': '8551870898',
                                                                                                                  'in_ucast_pkts': '35188545',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2350724',
                                                                                                                  'out_octets': '141962704611',
                                                                                                                  'out_ucast_pkts': '108046327',
                                                                                                                  'poller_hash': 'MTI3NTBCOXBlZXJpbnRmcG9uIDIyMTQzNQ=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '306084419388',
                                                                                                                  'in_ucast_pkts': '997127458',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '66',
                                                                                                                  'out_octets': '4570686855971',
                                                                                                                  'out_ucast_pkts': '3455108804',
                                                                                                                  'poller_hash': 'NDI4MkEwcGVlcmludGZwb24gMjYyNDc3'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '253537177503',
                                                                                         'out_octets': '656950492563712',
                                                                                         'out_ucast_pkts': '241113058144',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/7',
                                                                                                                 'id': '6/7',
                                                                                                                 'links': [       {       'connected_time': '21886807',
                                                                                                                                          'fiber_length': 23074,
                                                                                                                                          'poller_hash': 'NDI4MUE5cGVlcnBvbnBvbiAyNDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535258',
                                                                                                                                          'fiber_length': 23044,
                                                                                                                                          'poller_hash': 'NDI4MjVCcGVlcnBvbnBvbiAyNDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535250',
                                                                                                                                          'fiber_length': 23444,
                                                                                                                                          'poller_hash': 'NDI4MUFDcGVlcnBvbnBvbiAyNDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '18146441',
                                                                                                                                          'fiber_length': 23440,
                                                                                                                                          'poller_hash': 'NDI4MUI2cGVlcnBvbnBvbiAyNDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '6011614',
                                                                                                                                          'fiber_length': 23256,
                                                                                                                                          'poller_hash': 'NEY4RENEcGVlcnBvbnBvbiAyNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535117',
                                                                                                                                          'fiber_length': 23087,
                                                                                                                                          'poller_hash': 'NEY5MUNFcGVlcnBvbnBvbiAyNTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '21886993',
                                                                                                                                          'fiber_length': 23271,
                                                                                                                                          'poller_hash': 'NDI4MUQ4cGVlcnBvbnBvbiAyNTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '4241042',
                                                                                                                                          'fiber_length': 24704,
                                                                                                                                          'poller_hash': 'NDJDNzBDcGVlcnBvbnBvbiAyNTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535183',
                                                                                                                                          'fiber_length': 23058,
                                                                                                                                          'poller_hash': 'NDI4MUMxcGVlcnBvbnBvbiAyNTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535227',
                                                                                                                                          'fiber_length': 23209,
                                                                                                                                          'poller_hash': 'NDI4MUM3cGVlcnBvbnBvbiAyNjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '3895185',
                                                                                                                                          'fiber_length': 23145,
                                                                                                                                          'poller_hash': 'NDI4MTkzcGVlcnBvbnBvbiAyNjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '21614921',
                                                                                                                                          'fiber_length': 23420,
                                                                                                                                          'poller_hash': 'MzIyODc2cGVlcnBvbnBvbiAyNjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535096',
                                                                                                                                          'fiber_length': 23167,
                                                                                                                                          'poller_hash': 'NEZBMTc3cGVlcnBvbnBvbiAyNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '67856183',
                                                                                                                                          'fiber_length': 23387,
                                                                                                                                          'poller_hash': 'NTEyNTk1cGVlcnBvbnBvbiAyNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '67856291',
                                                                                                                                          'fiber_length': 23505,
                                                                                                                                          'poller_hash': 'NTEyNEZBcGVlcnBvbnBvbiAyNzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535117',
                                                                                                                                          'fiber_length': 23968,
                                                                                                                                          'poller_hash': 'NEZBMTlDcGVlcnBvbnBvbiAyOTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '21886656',
                                                                                                                                          'fiber_length': 23528,
                                                                                                                                          'poller_hash': 'NDJCOUEwcGVlcnBvbnBvbiAzMDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '86255',
                                                                                                                                          'fiber_length': 23411,
                                                                                                                                          'poller_hash': 'NDNCNUZFcGVlcnBvbnBvbiAzMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '21886677',
                                                                                                                                          'fiber_length': 23514,
                                                                                                                                          'poller_hash': 'NDJCOUJDcGVlcnBvbnBvbiAzNDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '21887171',
                                                                                                                                          'fiber_length': 23880,
                                                                                                                                          'poller_hash': 'NDJCOUM4cGVlcnBvbnBvbiAzNzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535087',
                                                                                                                                          'fiber_length': 23381,
                                                                                                                                          'poller_hash': 'NDNCNUQ3cGVlcnBvbnBvbiAzOTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '1448873',
                                                                                                                                          'fiber_length': 24359,
                                                                                                                                          'poller_hash': 'NDNCM0RDcGVlcnBvbnBvbiA0MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '3961424',
                                                                                                                                          'fiber_length': 24639,
                                                                                                                                          'poller_hash': 'NDJDRTI2cGVlcnBvbnBvbiA0MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8535285',
                                                                                                                                          'fiber_length': 23290,
                                                                                                                                          'poller_hash': 'NDJDRTQycGVlcnBvbnBvbiA0MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '4584004',
                                                                                                                                          'fiber_length': 24487,
                                                                                                                                          'poller_hash': 'NDJDOEY5cGVlcnBvbnBvbiA0MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '11397939',
                                                                                                                                          'fiber_length': 23101,
                                                                                                                                          'poller_hash': 'NDJDMUQ1cGVlcnBvbnBvbiA0Mzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '8667364',
                                                                                                                                          'fiber_length': 23026,
                                                                                                                                          'poller_hash': 'NDJCRjE3cGVlcnBvbnBvbiA0NDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '2484326',
                                                                                                                                          'fiber_length': 23748,
                                                                                                                                          'poller_hash': 'MTI3NTBCOXBlZXJwb25wb24gMjIxNDM1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285},
                                                                                                                                  {       'connected_time': '5094976',
                                                                                                                                          'fiber_length': 23363,
                                                                                                                                          'poller_hash': 'NDI4MkEwcGVlcnBvbnBvbiAyNjI0Nzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.285}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/7',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '6/8',
                                                                                         'in_discards': '112511',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3711178',
                                                                                         'in_octets': '35366429273265',
                                                                                         'in_ucast_pkts': '102829917892',
                                                                                         'links': [       {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '55644595729',
                                                                                                                  'in_ucast_pkts': '153977163',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2712356',
                                                                                                                  'out_octets': '623032526113',
                                                                                                                  'out_ucast_pkts': '487165525',
                                                                                                                  'poller_hash': 'NDI4MUZDcGVlcmludGZwb24gMjM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1783079017117',
                                                                                                                  'in_ucast_pkts': '4141792164',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '132',
                                                                                                                  'out_octets': '17081818466478',
                                                                                                                  'out_ucast_pkts': '12801132377',
                                                                                                                  'poller_hash': 'NDJDREIzcGVlcmludGZwb24gMjM4'},
                                                                                                          {       'in_nucast_pkts': '11935',
                                                                                                                  'in_octets': '2265476802572',
                                                                                                                  'in_ucast_pkts': '5159153706',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '61710344',
                                                                                                                  'out_octets': '18461930371614',
                                                                                                                  'out_ucast_pkts': '14219520906',
                                                                                                                  'poller_hash': 'NDI4MThGcGVlcmludGZwb24gMjQ3'},
                                                                                                          {       'in_nucast_pkts': '160826',
                                                                                                                  'in_octets': '402442292792',
                                                                                                                  'in_ucast_pkts': '1222212224',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652586',
                                                                                                                  'out_octets': '5234931478684',
                                                                                                                  'out_ucast_pkts': '3898779840',
                                                                                                                  'poller_hash': 'NDNCNUNDcGVlcmludGZwb24gMjUz'},
                                                                                                          {       'in_nucast_pkts': '7',
                                                                                                                  'in_octets': '138296369929',
                                                                                                                  'in_ucast_pkts': '525079291',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652625',
                                                                                                                  'out_octets': '4213066103259',
                                                                                                                  'out_ucast_pkts': '2991732309',
                                                                                                                  'poller_hash': 'NEZBMUMxcGVlcmludGZwb24gMjYz'},
                                                                                                          {       'in_nucast_pkts': '1960',
                                                                                                                  'in_octets': '166876826130',
                                                                                                                  'in_ucast_pkts': '229189282',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1727640',
                                                                                                                  'out_octets': '408644712410',
                                                                                                                  'out_ucast_pkts': '409581960',
                                                                                                                  'poller_hash': 'MzVGOUI0cGVlcmludGZwb24gMjcy'},
                                                                                                          {       'in_nucast_pkts': '12552',
                                                                                                                  'in_octets': '72529259743',
                                                                                                                  'in_ucast_pkts': '410199904',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8065568',
                                                                                                                  'out_octets': '1298224061614',
                                                                                                                  'out_ucast_pkts': '982978963',
                                                                                                                  'poller_hash': 'NTEyNEY2cGVlcmludGZwb24gMjgw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '224297189099',
                                                                                                                  'in_ucast_pkts': '9415092543',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '33',
                                                                                                                  'out_octets': '3475246615241',
                                                                                                                  'out_ucast_pkts': '2621213464',
                                                                                                                  'poller_hash': 'NDJCOUU1cGVlcmludGZwb24gMzY3'},
                                                                                                          {       'in_nucast_pkts': '66',
                                                                                                                  'in_octets': '122551938421',
                                                                                                                  'in_ucast_pkts': '639647008',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919663',
                                                                                                                  'out_octets': '2530563934160',
                                                                                                                  'out_ucast_pkts': '1940280295',
                                                                                                                  'poller_hash': 'NTc0NERDcGVlcmludGZwb24gMzY5'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'QzgzMEUxcGVlcmludGZwb24gMzgy'},
                                                                                                          {       'in_nucast_pkts': '10899',
                                                                                                                  'in_octets': '236683277489',
                                                                                                                  'in_ucast_pkts': '905723935',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5652385',
                                                                                                                  'out_octets': '2304568453165',
                                                                                                                  'out_ucast_pkts': '1847734253',
                                                                                                                  'poller_hash': 'NDNCNThBcGVlcmludGZwb24gMzg4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJDNDU3cGVlcmludGZwb24gNDU5'},
                                                                                                          {       'in_nucast_pkts': '6537',
                                                                                                                  'in_octets': '73168032882',
                                                                                                                  'in_ucast_pkts': '160304247',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1436207',
                                                                                                                  'out_octets': '548737053497',
                                                                                                                  'out_ucast_pkts': '436322184',
                                                                                                                  'poller_hash': 'NUE2QjhBcGVlcmludGZwb24gNDYz'},
                                                                                                          {       'in_nucast_pkts': '163842',
                                                                                                                  'in_octets': '1324615033171',
                                                                                                                  'in_ucast_pkts': '2915460688',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919952',
                                                                                                                  'out_octets': '13671825187874',
                                                                                                                  'out_ucast_pkts': '10516245318',
                                                                                                                  'poller_hash': 'ODY5NDJCcGVlcmludGZwb24gNDg5'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '119159151689',
                                                                                                                  'in_ucast_pkts': '421968903',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920245',
                                                                                                                  'out_octets': '2937774076535',
                                                                                                                  'out_ucast_pkts': '2072964864',
                                                                                                                  'poller_hash': 'NDJDMUYxcGVlcmludGZwb24gNTA1'},
                                                                                                          {       'in_nucast_pkts': '78139',
                                                                                                                  'in_octets': '49635459944',
                                                                                                                  'in_ucast_pkts': '91629996',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2632897',
                                                                                                                  'out_octets': '341107588909',
                                                                                                                  'out_ucast_pkts': '272540685',
                                                                                                                  'poller_hash': 'NDJDRTUwcGVlcmludGZwb24gNTA5'},
                                                                                                          {       'in_nucast_pkts': '266844',
                                                                                                                  'in_octets': '82724726694',
                                                                                                                  'in_ucast_pkts': '617229582',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10268457',
                                                                                                                  'out_octets': '3625200318434',
                                                                                                                  'out_ucast_pkts': '2461714645',
                                                                                                                  'poller_hash': 'NzFBNEFBcGVlcmludGZwb24gNTMy'},
                                                                                                          {       'in_nucast_pkts': '2051',
                                                                                                                  'in_octets': '147722978918',
                                                                                                                  'in_ucast_pkts': '481431455',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2460629',
                                                                                                                  'out_octets': '2226184304524',
                                                                                                                  'out_ucast_pkts': '1709027558',
                                                                                                                  'poller_hash': 'NjQwQjIzcGVlcmludGZwb24gNTMz'},
                                                                                                          {       'in_nucast_pkts': '18',
                                                                                                                  'in_octets': '211024629564',
                                                                                                                  'in_ucast_pkts': '852342605',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9917745',
                                                                                                                  'out_octets': '2369931033824',
                                                                                                                  'out_ucast_pkts': '1886143351',
                                                                                                                  'poller_hash': 'NDJDNDk3cGVlcmludGZwb24gMTM4MTYw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJCOUYxcGVlcmludGZwb24gMTM4MjYz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NUE2QzkzcGVlcmludGZwb24gMTM5MzA5'},
                                                                                                          {       'in_nucast_pkts': '228857',
                                                                                                                  'in_octets': '34189912735',
                                                                                                                  'in_ucast_pkts': '204512582',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920163',
                                                                                                                  'out_octets': '830551159619',
                                                                                                                  'out_ucast_pkts': '621946806',
                                                                                                                  'poller_hash': 'ODQ0QzA2cGVlcmludGZwb24gMTQwMzU3'},
                                                                                                          {       'in_nucast_pkts': '15',
                                                                                                                  'in_octets': '188727843006',
                                                                                                                  'in_ucast_pkts': '1013487617',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '750',
                                                                                                                  'out_octets': '5817162045058',
                                                                                                                  'out_ucast_pkts': '4096675559',
                                                                                                                  'poller_hash': 'ODY5MzZCcGVlcmludGZwb24gMTQwNDgx'},
                                                                                                          {       'in_nucast_pkts': '163727',
                                                                                                                  'in_octets': '155101260533',
                                                                                                                  'in_ucast_pkts': '547763083',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920149',
                                                                                                                  'out_octets': '3622209551857',
                                                                                                                  'out_ucast_pkts': '2629295215',
                                                                                                                  'poller_hash': 'ODY5M0MycGVlcmludGZwb24gMTQwNjE5'},
                                                                                                          {       'in_nucast_pkts': '7404502',
                                                                                                                  'in_octets': '69432025952',
                                                                                                                  'in_ucast_pkts': '576668236',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '12454595',
                                                                                                                  'out_octets': '1818211347054',
                                                                                                                  'out_ucast_pkts': '1304205823',
                                                                                                                  'poller_hash': 'ODQ0QkU0cGVlcmludGZwb24gMTQwNjg5'},
                                                                                                          {       'in_nucast_pkts': '75',
                                                                                                                  'in_octets': '86336733416',
                                                                                                                  'in_ucast_pkts': '360023789',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9919989',
                                                                                                                  'out_octets': '1230444913445',
                                                                                                                  'out_ucast_pkts': '913913945',
                                                                                                                  'poller_hash': 'MzIxMDE4cGVlcmludGZwb24gMjMzNjQz'},
                                                                                                          {       'in_nucast_pkts': '149664',
                                                                                                                  'in_octets': '130254975536',
                                                                                                                  'in_ucast_pkts': '386207677',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9920035',
                                                                                                                  'out_octets': '1682372013687',
                                                                                                                  'out_ucast_pkts': '1305614801',
                                                                                                                  'poller_hash': 'MTI3NDdCM3BlZXJpbnRmcG9uIDI3NDUyNA=='},
                                                                                                          {       'in_nucast_pkts': '15',
                                                                                                                  'in_octets': '574758057468',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '508',
                                                                                                                  'out_octets': '2934394857591',
                                                                                                                  'out_ucast_pkts': '2265612906',
                                                                                                                  'poller_hash': 'MTI3M0EzMXBlZXJpbnRmcG9uIDI3OTMxNg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '6/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '88869134290',
                                                                                         'out_octets': '521224475575694',
                                                                                         'out_ucast_pkts': '303044323816',
                                                                                         'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '6/8',
                                                                                                                 'id': '6/8',
                                                                                                                 'links': [       {       'connected_time': '2296131',
                                                                                                                                          'fiber_length': 23150,
                                                                                                                                          'poller_hash': 'NDI4MUZDcGVlcnBvbnBvbiAyMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '7142486',
                                                                                                                                          'fiber_length': 23209,
                                                                                                                                          'poller_hash': 'NDJDREIzcGVlcnBvbnBvbiAyMzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '21886238',
                                                                                                                                          'fiber_length': 23181,
                                                                                                                                          'poller_hash': 'NDI4MThGcGVlcnBvbnBvbiAyNDc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535052',
                                                                                                                                          'fiber_length': 23180,
                                                                                                                                          'poller_hash': 'NDNCNUNDcGVlcnBvbnBvbiAyNTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535114',
                                                                                                                                          'fiber_length': 23327,
                                                                                                                                          'poller_hash': 'NEZBMUMxcGVlcnBvbnBvbiAyNjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '1472244',
                                                                                                                                          'fiber_length': 23330,
                                                                                                                                          'poller_hash': 'MzVGOUI0cGVlcnBvbnBvbiAyNzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '6732958',
                                                                                                                                          'fiber_length': 23457,
                                                                                                                                          'poller_hash': 'NTEyNEY2cGVlcnBvbnBvbiAyODA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '1741582',
                                                                                                                                          'fiber_length': 23273,
                                                                                                                                          'poller_hash': 'NDJCOUU1cGVlcnBvbnBvbiAzNjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535139',
                                                                                                                                          'fiber_length': 23712,
                                                                                                                                          'poller_hash': 'NTc0NERDcGVlcnBvbnBvbiAzNjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535289',
                                                                                                                                          'fiber_length': 23309,
                                                                                                                                          'poller_hash': 'QzgzMEUxcGVlcnBvbnBvbiAzODI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535083',
                                                                                                                                          'fiber_length': 23050,
                                                                                                                                          'poller_hash': 'NDNCNThBcGVlcnBvbnBvbiAzODg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535247',
                                                                                                                                          'fiber_length': 23737,
                                                                                                                                          'poller_hash': 'NDJDNDU3cGVlcnBvbnBvbiA0NTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '1224969',
                                                                                                                                          'fiber_length': 23400,
                                                                                                                                          'poller_hash': 'NUE2QjhBcGVlcnBvbnBvbiA0NjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535222',
                                                                                                                                          'fiber_length': 23956,
                                                                                                                                          'poller_hash': 'ODY5NDJCcGVlcnBvbnBvbiA0ODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535253',
                                                                                                                                          'fiber_length': 23696,
                                                                                                                                          'poller_hash': 'NDJDMUYxcGVlcnBvbnBvbiA1MDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '2227736',
                                                                                                                                          'fiber_length': 23298,
                                                                                                                                          'poller_hash': 'NDJDRTUwcGVlcnBvbnBvbiA1MDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '15182484',
                                                                                                                                          'fiber_length': 23860,
                                                                                                                                          'poller_hash': 'NzFBNEFBcGVlcnBvbnBvbiA1MzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '3620157',
                                                                                                                                          'fiber_length': 23608,
                                                                                                                                          'poller_hash': 'NjQwQjIzcGVlcnBvbnBvbiA1MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535261',
                                                                                                                                          'fiber_length': 23329,
                                                                                                                                          'poller_hash': 'NDJDNDk3cGVlcnBvbnBvbiAxMzgxNjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535242',
                                                                                                                                          'fiber_length': 23906,
                                                                                                                                          'poller_hash': 'NDJCOUYxcGVlcnBvbnBvbiAxMzgyNjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535397',
                                                                                                                                          'fiber_length': 23794,
                                                                                                                                          'poller_hash': 'NUE2QzkzcGVlcnBvbnBvbiAxMzkzMDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535529',
                                                                                                                                          'fiber_length': 23546,
                                                                                                                                          'poller_hash': 'ODQ0QzA2cGVlcnBvbnBvbiAxNDAzNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535236',
                                                                                                                                          'fiber_length': 23196,
                                                                                                                                          'poller_hash': 'ODY5MzZCcGVlcnBvbnBvbiAxNDA0ODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535273',
                                                                                                                                          'fiber_length': 23058,
                                                                                                                                          'poller_hash': 'ODY5M0MycGVlcnBvbnBvbiAxNDA2MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '10732600',
                                                                                                                                          'fiber_length': 23554,
                                                                                                                                          'poller_hash': 'ODQ0QkU0cGVlcnBvbnBvbiAxNDA2ODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535262',
                                                                                                                                          'fiber_length': 23036,
                                                                                                                                          'poller_hash': 'MzIxMDE4cGVlcnBvbnBvbiAyMzM2NDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '8535206',
                                                                                                                                          'fiber_length': 23783,
                                                                                                                                          'poller_hash': 'MTI3NDdCM3BlZXJwb25wb24gMjc0NTI0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599},
                                                                                                                                  {       'connected_time': '6648143',
                                                                                                                                          'fiber_length': 23798,
                                                                                                                                          'poller_hash': 'MTI3M0EzMXBlZXJwb25wb24gMjc5MzE2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.599}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '6/8',
                                                                                                                 'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNjEyNjA2NzMw',
                                                           'serial_number': '211612606730',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43502009'},
                                                   {       'id': '7',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '7/1',
                                                                                         'in_discards': '41018',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7061231',
                                                                                         'in_octets': '24970663152670',
                                                                                         'in_ucast_pkts': '66863610463',
                                                                                         'links': [       {       'in_nucast_pkts': '613',
                                                                                                                  'in_octets': '348367835580',
                                                                                                                  'in_ucast_pkts': '1148601839',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1352',
                                                                                                                  'out_octets': '5733089419131',
                                                                                                                  'out_ucast_pkts': '4196135622',
                                                                                                                  'poller_hash': 'QzgzOTNDcGVlcmludGZwb24gMTQy'},
                                                                                                          {       'in_nucast_pkts': '142',
                                                                                                                  'in_octets': '124891031684',
                                                                                                                  'in_ucast_pkts': '339538903',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10070347',
                                                                                                                  'out_octets': '1748459828314',
                                                                                                                  'out_ucast_pkts': '1425895022',
                                                                                                                  'poller_hash': 'NEY4Qzk5cGVlcmludGZwb24gMjE0'},
                                                                                                          {       'in_nucast_pkts': '673',
                                                                                                                  'in_octets': '130013133609',
                                                                                                                  'in_ucast_pkts': '618569053',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11888493',
                                                                                                                  'out_octets': '3053821624757',
                                                                                                                  'out_ucast_pkts': '2312905496',
                                                                                                                  'poller_hash': 'NDJDNDU2cGVlcmludGZwb24gNDU4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '505102367728',
                                                                                                                  'in_ucast_pkts': '1821306639',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '65',
                                                                                                                  'out_octets': '4810962640005',
                                                                                                                  'out_ucast_pkts': '3505372108',
                                                                                                                  'poller_hash': 'NDJDMjFGcGVlcmludGZwb24gNDY2'},
                                                                                                          {       'in_nucast_pkts': '4143805',
                                                                                                                  'in_octets': '166925286126',
                                                                                                                  'in_ucast_pkts': '716151228',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10115556677',
                                                                                                                  'out_octets': '16352884609525',
                                                                                                                  'out_ucast_pkts': '2056788808',
                                                                                                                  'poller_hash': 'QzgzREY3cGVlcmludGZwb24gMTU3MzE0'},
                                                                                                          {       'in_nucast_pkts': '9781406',
                                                                                                                  'in_octets': '820577320142',
                                                                                                                  'in_ucast_pkts': '1441284698',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797819',
                                                                                                                  'out_octets': '2649659498047',
                                                                                                                  'out_ucast_pkts': '2331225686',
                                                                                                                  'poller_hash': 'Q0JEMjNDcGVlcmludGZwb24gMTY2MzI4'},
                                                                                                          {       'in_nucast_pkts': '158068',
                                                                                                                  'in_octets': '722005898264',
                                                                                                                  'in_ucast_pkts': '881601907',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797962',
                                                                                                                  'out_octets': '2096718538878',
                                                                                                                  'out_ucast_pkts': '1653461257',
                                                                                                                  'poller_hash': 'Q0JEMUNCcGVlcmludGZwb24gMTY2ODU4'},
                                                                                                          {       'in_nucast_pkts': '155046',
                                                                                                                  'in_octets': '301518126624',
                                                                                                                  'in_ucast_pkts': '765660369',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797809',
                                                                                                                  'out_octets': '3216657779715',
                                                                                                                  'out_ucast_pkts': '2420052928',
                                                                                                                  'poller_hash': 'Q0JEMUJBcGVlcmludGZwb24gMTY3ODc4'},
                                                                                                          {       'in_nucast_pkts': '147406',
                                                                                                                  'in_octets': '200091345447',
                                                                                                                  'in_ucast_pkts': '567472555',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9516701',
                                                                                                                  'out_octets': '1935917451713',
                                                                                                                  'out_ucast_pkts': '1635418517',
                                                                                                                  'poller_hash': 'REZGRTAycGVlcmludGZwb24gMTgxMTEx'},
                                                                                                          {       'in_nucast_pkts': '112684',
                                                                                                                  'in_octets': '72461138623',
                                                                                                                  'in_ucast_pkts': '213030268',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7429023',
                                                                                                                  'out_octets': '835005650588',
                                                                                                                  'out_ucast_pkts': '696908710',
                                                                                                                  'poller_hash': 'QzgxREEzcGVlcmludGZwb24gMjA2MzM3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '935798595618',
                                                                                                                  'in_ucast_pkts': '4878812406',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '84',
                                                                                                                  'out_octets': '19929742495974',
                                                                                                                  'out_ucast_pkts': '25134346087',
                                                                                                                  'poller_hash': 'MTE5NTEyMnBlZXJpbnRmcG9uIDIxNjI2Nw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '426907319650',
                                                                                                                  'in_ucast_pkts': '1065860134',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '42',
                                                                                                                  'out_octets': '3106560711798',
                                                                                                                  'out_ucast_pkts': '2420295035',
                                                                                                                  'poller_hash': 'MTI3NDJFRnBlZXJpbnRmcG9uIDIyOTY5MQ=='},
                                                                                                          {       'in_nucast_pkts': '160987',
                                                                                                                  'in_octets': '419420874625',
                                                                                                                  'in_ucast_pkts': '1060192317',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797914',
                                                                                                                  'out_octets': '2848929861202',
                                                                                                                  'out_ucast_pkts': '2411132103',
                                                                                                                  'poller_hash': 'MTI3NEFEQXBlZXJpbnRmcG9uIDI0MjIwNQ=='},
                                                                                                          {       'in_nucast_pkts': '160555',
                                                                                                                  'in_octets': '144291648539',
                                                                                                                  'in_ucast_pkts': '267112540',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9798010',
                                                                                                                  'out_octets': '965629743923',
                                                                                                                  'out_ucast_pkts': '784376468',
                                                                                                                  'poller_hash': 'MTI3NDJCOHBlZXJpbnRmcG9uIDI0ODQ0Mw=='},
                                                                                                          {       'in_nucast_pkts': '149230',
                                                                                                                  'in_octets': '117338056572',
                                                                                                                  'in_ucast_pkts': '453410342',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797998',
                                                                                                                  'out_octets': '2063823211438',
                                                                                                                  'out_ucast_pkts': '1555444607',
                                                                                                                  'poller_hash': 'MTE3REI2QnBlZXJpbnRmcG9uIDI1MDA3NA=='},
                                                                                                          {       'in_nucast_pkts': '155108',
                                                                                                                  'in_octets': '60970917607',
                                                                                                                  'in_ucast_pkts': '336360467',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797748',
                                                                                                                  'out_octets': '1752589577869',
                                                                                                                  'out_ucast_pkts': '1297173095',
                                                                                                                  'poller_hash': 'MTI5NzRBM3BlZXJpbnRmcG9uIDI1Njc0Ng=='},
                                                                                                          {       'in_nucast_pkts': '307',
                                                                                                                  'in_octets': '1104261220985',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '564',
                                                                                                                  'out_octets': '8517556344612',
                                                                                                                  'out_ucast_pkts': '6695277736',
                                                                                                                  'poller_hash': 'MTI3NDY5MXBlZXJpbnRmcG9uIDI2NDkzMA=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '41541739749',
                                                                                                                  'in_ucast_pkts': '138037451',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '1101342216531',
                                                                                                                  'out_ucast_pkts': '772323692',
                                                                                                                  'poller_hash': 'MTI3MEQzMnBlZXJpbnRmcG9uIDI3OTg5MQ=='},
                                                                                                          {       'in_nucast_pkts': '49963',
                                                                                                                  'in_octets': '21507975923',
                                                                                                                  'in_ucast_pkts': '130455961',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2802214',
                                                                                                                  'out_octets': '219689138859',
                                                                                                                  'out_ucast_pkts': '196721849',
                                                                                                                  'poller_hash': 'MTI5NUU4MXBlZXJpbnRmcG9uIDI4Mjc1MA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '7/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '73088119751',
                                                                                         'out_octets': '287955254790394',
                                                                                         'out_ucast_pkts': '171447186598',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/1',
                                                                                                                 'id': '7/1',
                                                                                                                 'links': [       {       'connected_time': '8535178',
                                                                                                                                          'fiber_length': 23737,
                                                                                                                                          'poller_hash': 'QzgzOTNDcGVlcnBvbnBvbiAxNDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '15182469',
                                                                                                                                          'fiber_length': 23159,
                                                                                                                                          'poller_hash': 'NEY4Qzk5cGVlcnBvbnBvbiAyMTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '10372755',
                                                                                                                                          'fiber_length': 23224,
                                                                                                                                          'poller_hash': 'NDJDNDU2cGVlcnBvbnBvbiA0NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '4975882',
                                                                                                                                          'fiber_length': 24030,
                                                                                                                                          'poller_hash': 'NDJDMjFGcGVlcnBvbnBvbiA0NjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535231',
                                                                                                                                          'fiber_length': 23223,
                                                                                                                                          'poller_hash': 'QzgzREY3cGVlcnBvbnBvbiAxNTczMTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535159',
                                                                                                                                          'fiber_length': 23728,
                                                                                                                                          'poller_hash': 'Q0JEMjNDcGVlcnBvbnBvbiAxNjYzMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535180',
                                                                                                                                          'fiber_length': 23191,
                                                                                                                                          'poller_hash': 'Q0JEMUNCcGVlcnBvbnBvbiAxNjY4NTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535205',
                                                                                                                                          'fiber_length': 24054,
                                                                                                                                          'poller_hash': 'Q0JEMUJBcGVlcnBvbnBvbiAxNjc4Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8288457',
                                                                                                                                          'fiber_length': 23386,
                                                                                                                                          'poller_hash': 'REZGRTAycGVlcnBvbnBvbiAxODExMTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '6460052',
                                                                                                                                          'fiber_length': 24042,
                                                                                                                                          'poller_hash': 'QzgxREEzcGVlcnBvbnBvbiAyMDYzMzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '15182678',
                                                                                                                                          'fiber_length': 23588,
                                                                                                                                          'poller_hash': 'MTE5NTEyMnBlZXJwb25wb24gMjE2MjY3',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535228',
                                                                                                                                          'fiber_length': 24017,
                                                                                                                                          'poller_hash': 'MTI3NDJFRnBlZXJwb25wb24gMjI5Njkx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535272',
                                                                                                                                          'fiber_length': 23256,
                                                                                                                                          'poller_hash': 'MTI3NEFEQXBlZXJwb25wb24gMjQyMjA1',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535243',
                                                                                                                                          'fiber_length': 23345,
                                                                                                                                          'poller_hash': 'MTI3NDJCOHBlZXJwb25wb24gMjQ4NDQz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535293',
                                                                                                                                          'fiber_length': 23044,
                                                                                                                                          'poller_hash': 'MTE3REI2QnBlZXJwb25wb24gMjUwMDc0',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535330',
                                                                                                                                          'fiber_length': 23429,
                                                                                                                                          'poller_hash': 'MTI5NzRBM3BlZXJwb25wb24gMjU2NzQ2',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '8535190',
                                                                                                                                          'fiber_length': 23463,
                                                                                                                                          'poller_hash': 'MTI3NDY5MXBlZXJwb25wb24gMjY0OTMw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '585942',
                                                                                                                                          'fiber_length': 23291,
                                                                                                                                          'poller_hash': 'MTI3MEQzMnBlZXJwb25wb24gMjc5ODkx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336},
                                                                                                                                  {       'connected_time': '2413270',
                                                                                                                                          'fiber_length': 23386,
                                                                                                                                          'poller_hash': 'MTI5NUU4MXBlZXJwb25wb24gMjgyNzUw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.336}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/1',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/2',
                                                                                         'in_discards': '17833',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7401969',
                                                                                         'in_octets': '36373810928263',
                                                                                         'in_ucast_pkts': '103388527416',
                                                                                         'links': [       {       'in_nucast_pkts': '2427934',
                                                                                                                  'in_octets': '94457073622',
                                                                                                                  'in_ucast_pkts': '472054140',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '3551422255',
                                                                                                                  'out_octets': '6546242536242',
                                                                                                                  'out_ucast_pkts': '1362194228',
                                                                                                                  'poller_hash': 'QzgzQjY1cGVlcmludGZwb24gNDEy'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '137756127555',
                                                                                                                  'in_ucast_pkts': '648821837',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530414',
                                                                                                                  'out_octets': '2372935552957',
                                                                                                                  'out_ucast_pkts': '1785046835',
                                                                                                                  'poller_hash': 'NjQwQjdBcGVlcmludGZwb24gNDIy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '1211992282862',
                                                                                                                  'in_ucast_pkts': '51635139424',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '14968647035742',
                                                                                                                  'out_ucast_pkts': '11629311352',
                                                                                                                  'poller_hash': 'NDJDRTkycGVlcmludGZwb24gNDMz'},
                                                                                                          {       'in_nucast_pkts': '154',
                                                                                                                  'in_octets': '162915403927',
                                                                                                                  'in_ucast_pkts': '1058626955',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '790',
                                                                                                                  'out_octets': '5466292032678',
                                                                                                                  'out_ucast_pkts': '3923999564',
                                                                                                                  'poller_hash': 'NDJCRjIycGVlcmludGZwb24gNDQy'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '137084105653',
                                                                                                                  'in_ucast_pkts': '804791914',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530424',
                                                                                                                  'out_octets': '4567356494581',
                                                                                                                  'out_ucast_pkts': '3288538621',
                                                                                                                  'poller_hash': 'NDNCNDQ1cGVlcmludGZwb24gNDQ0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJCRjdCcGVlcmludGZwb24gNDU1'},
                                                                                                          {       'in_nucast_pkts': '329',
                                                                                                                  'in_octets': '110736534443',
                                                                                                                  'in_ucast_pkts': '1071260184',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '634',
                                                                                                                  'out_octets': '6221938724923',
                                                                                                                  'out_ucast_pkts': '4176102075',
                                                                                                                  'poller_hash': 'NDJDM0VDcGVlcmludGZwb24gNDU3'},
                                                                                                          {       'in_nucast_pkts': '101',
                                                                                                                  'in_octets': '236085084668',
                                                                                                                  'in_ucast_pkts': '1654527553',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797749',
                                                                                                                  'out_octets': '5425629628858',
                                                                                                                  'out_ucast_pkts': '3802461611',
                                                                                                                  'poller_hash': 'NDJCRTc3cGVlcmludGZwb24gNDYy'},
                                                                                                          {       'in_nucast_pkts': '161874',
                                                                                                                  'in_octets': '218635970141',
                                                                                                                  'in_ucast_pkts': '662643032',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9793291',
                                                                                                                  'out_octets': '2168619536862',
                                                                                                                  'out_ucast_pkts': '1623300579',
                                                                                                                  'poller_hash': 'NDJEMUZDcGVlcmludGZwb24gNDgz'},
                                                                                                          {       'in_nucast_pkts': '52703',
                                                                                                                  'in_octets': '37323514020',
                                                                                                                  'in_ucast_pkts': '145104816',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1308621',
                                                                                                                  'out_octets': '576806212622',
                                                                                                                  'out_ucast_pkts': '434314768',
                                                                                                                  'poller_hash': 'NDJDNTIwcGVlcmludGZwb24gNDkz'},
                                                                                                          {       'in_nucast_pkts': '159542',
                                                                                                                  'in_octets': '66807115068',
                                                                                                                  'in_ucast_pkts': '525393845',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797798',
                                                                                                                  'out_octets': '3900995309769',
                                                                                                                  'out_ucast_pkts': '2690602418',
                                                                                                                  'poller_hash': 'NDJCRTlGcGVlcmludGZwb24gNDk4'},
                                                                                                          {       'in_nucast_pkts': '9617476',
                                                                                                                  'in_octets': '62327785306',
                                                                                                                  'in_ucast_pkts': '257767862',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530412',
                                                                                                                  'out_octets': '2084064768837',
                                                                                                                  'out_ucast_pkts': '1456194812',
                                                                                                                  'poller_hash': 'NjQwQjY5cGVlcmludGZwb24gNTAy'},
                                                                                                          {       'in_nucast_pkts': '9',
                                                                                                                  'in_octets': '196633559224',
                                                                                                                  'in_ucast_pkts': '759786860',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797944',
                                                                                                                  'out_octets': '2433936600859',
                                                                                                                  'out_ucast_pkts': '1917048963',
                                                                                                                  'poller_hash': 'NDJDOUQ0cGVlcmludGZwb24gNTIz'},
                                                                                                          {       'in_nucast_pkts': '49770',
                                                                                                                  'in_octets': '1093595608',
                                                                                                                  'in_ucast_pkts': '6471007',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '211193',
                                                                                                                  'out_octets': '31626570773',
                                                                                                                  'out_ucast_pkts': '23503420',
                                                                                                                  'poller_hash': 'NjQwQjFDcGVlcmludGZwb24gMTM4MDky'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '900888007708',
                                                                                                                  'in_ucast_pkts': '3458094422',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '11469736959806',
                                                                                                                  'out_ucast_pkts': '8428211828',
                                                                                                                  'poller_hash': 'ODY3ODhCcGVlcmludGZwb24gMTQwNzQ2'},
                                                                                                          {       'in_nucast_pkts': '161566',
                                                                                                                  'in_octets': '176726041766',
                                                                                                                  'in_ucast_pkts': '529079260',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797834',
                                                                                                                  'out_octets': '1747579987269',
                                                                                                                  'out_ucast_pkts': '1337388843',
                                                                                                                  'poller_hash': 'QTlDM0Y2cGVlcmludGZwb24gMTQwNzcx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '457014585137',
                                                                                                                  'in_ucast_pkts': '1261365505',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '92',
                                                                                                                  'out_octets': '6670152845760',
                                                                                                                  'out_ucast_pkts': '4961636041',
                                                                                                                  'poller_hash': 'OTU2RjBCcGVlcmludGZwb24gMTQ1MTY1'},
                                                                                                          {       'in_nucast_pkts': '5565942',
                                                                                                                  'in_octets': '499993496335',
                                                                                                                  'in_ucast_pkts': '1434164150',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7686164651',
                                                                                                                  'out_octets': '15326734099184',
                                                                                                                  'out_ucast_pkts': '3751420060',
                                                                                                                  'poller_hash': 'OTU2RTQ2cGVlcmludGZwb24gMTQ1MjEy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '744161835304',
                                                                                                                  'in_ucast_pkts': '10209648228',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '84',
                                                                                                                  'out_octets': '3210153646976',
                                                                                                                  'out_ucast_pkts': '2844526005',
                                                                                                                  'poller_hash': 'OTU2RTU0cGVlcmludGZwb24gMTQ1MjQ5'},
                                                                                                          {       'in_nucast_pkts': '160647',
                                                                                                                  'in_octets': '112417894510',
                                                                                                                  'in_ucast_pkts': '357556129',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797735',
                                                                                                                  'out_octets': '1663596927123',
                                                                                                                  'out_ucast_pkts': '1255722599',
                                                                                                                  'poller_hash': 'OTU2RTFFcGVlcmludGZwb24gMTQ1Mjk5'},
                                                                                                          {       'in_nucast_pkts': '150495',
                                                                                                                  'in_octets': '67462213469',
                                                                                                                  'in_ucast_pkts': '251133008',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797736',
                                                                                                                  'out_octets': '1490684767664',
                                                                                                                  'out_ucast_pkts': '1113507393',
                                                                                                                  'poller_hash': 'OTU2RTMxcGVlcmludGZwb24gMTQ1NTY4'},
                                                                                                          {       'in_nucast_pkts': '152578',
                                                                                                                  'in_octets': '64603184283',
                                                                                                                  'in_ucast_pkts': '326836629',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797815',
                                                                                                                  'out_octets': '1329736349004',
                                                                                                                  'out_ucast_pkts': '986258214',
                                                                                                                  'poller_hash': 'OTU2RjkwcGVlcmludGZwb24gMTQ1NzMx'},
                                                                                                          {       'in_nucast_pkts': '3787867',
                                                                                                                  'in_octets': '66632331097',
                                                                                                                  'in_ucast_pkts': '201313874',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10547580733',
                                                                                                                  'out_octets': '15120671153763',
                                                                                                                  'out_ucast_pkts': '671146199',
                                                                                                                  'poller_hash': 'QTlDNDE5cGVlcmludGZwb24gMTQ2MDUx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '96796616546',
                                                                                                                  'in_ucast_pkts': '470772917',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38',
                                                                                                                  'out_octets': '3304017872402',
                                                                                                                  'out_ucast_pkts': '2365241708',
                                                                                                                  'poller_hash': 'QzgxRUFCcGVlcmludGZwb24gMjAyNjQ1'},
                                                                                                          {       'in_nucast_pkts': '148485',
                                                                                                                  'in_octets': '140049442117',
                                                                                                                  'in_ucast_pkts': '726021370',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797855',
                                                                                                                  'out_octets': '2952284841576',
                                                                                                                  'out_ucast_pkts': '2210150459',
                                                                                                                  'poller_hash': 'MTI3NDk2QnBlZXJpbnRmcG9uIDIzNTcxMw=='},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '7277027893047',
                                                                                                                  'in_ucast_pkts': '8828878048',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38',
                                                                                                                  'out_octets': '12658474929845',
                                                                                                                  'out_ucast_pkts': '11948342109',
                                                                                                                  'poller_hash': 'MTI3M0Y1OHBlZXJpbnRmcG9uIDI0NzAyMA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '7/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '230994563334',
                                                                                         'out_octets': '677955129195623',
                                                                                         'out_ucast_pkts': '273882452015',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/2',
                                                                                                                 'id': '7/2',
                                                                                                                 'links': [       {       'connected_time': '8535310',
                                                                                                                                          'fiber_length': 23326,
                                                                                                                                          'poller_hash': 'QzgzQjY1cGVlcnBvbnBvbiA0MTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535108',
                                                                                                                                          'fiber_length': 23323,
                                                                                                                                          'poller_hash': 'NjQwQjdBcGVlcnBvbnBvbiA0MjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '7383395',
                                                                                                                                          'fiber_length': 23562,
                                                                                                                                          'poller_hash': 'NDJDRTkycGVlcnBvbnBvbiA0MzM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535186',
                                                                                                                                          'fiber_length': 23663,
                                                                                                                                          'poller_hash': 'NDJCRjIycGVlcnBvbnBvbiA0NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535113',
                                                                                                                                          'fiber_length': 23132,
                                                                                                                                          'poller_hash': 'NDNCNDQ1cGVlcnBvbnBvbiA0NDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535203',
                                                                                                                                          'fiber_length': 23644,
                                                                                                                                          'poller_hash': 'NDJCRjdCcGVlcnBvbnBvbiA0NTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '29644083',
                                                                                                                                          'fiber_length': 23757,
                                                                                                                                          'poller_hash': 'NDJDM0VDcGVlcnBvbnBvbiA0NTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535246',
                                                                                                                                          'fiber_length': 23713,
                                                                                                                                          'poller_hash': 'NDJCRTc3cGVlcnBvbnBvbiA0NjI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535228',
                                                                                                                                          'fiber_length': 23473,
                                                                                                                                          'poller_hash': 'NDJEMUZDcGVlcnBvbnBvbiA0ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '1084853',
                                                                                                                                          'fiber_length': 23331,
                                                                                                                                          'poller_hash': 'NDJDNTIwcGVlcnBvbnBvbiA0OTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535351',
                                                                                                                                          'fiber_length': 23799,
                                                                                                                                          'poller_hash': 'NDJCRTlGcGVlcnBvbnBvbiA0OTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535110',
                                                                                                                                          'fiber_length': 23286,
                                                                                                                                          'poller_hash': 'NjQwQjY5cGVlcnBvbnBvbiA1MDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535201',
                                                                                                                                          'fiber_length': 23404,
                                                                                                                                          'poller_hash': 'NDJDOUQ0cGVlcnBvbnBvbiA1MjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '319455',
                                                                                                                                          'fiber_length': 23420,
                                                                                                                                          'poller_hash': 'NjQwQjFDcGVlcnBvbnBvbiAxMzgwOTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535393',
                                                                                                                                          'fiber_length': 23063,
                                                                                                                                          'poller_hash': 'ODY3ODhCcGVlcnBvbnBvbiAxNDA3NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535293',
                                                                                                                                          'fiber_length': 22987,
                                                                                                                                          'poller_hash': 'QTlDM0Y2cGVlcnBvbnBvbiAxNDA3NzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '73026156',
                                                                                                                                          'fiber_length': 23015,
                                                                                                                                          'poller_hash': 'OTU2RjBCcGVlcnBvbnBvbiAxNDUxNjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535334',
                                                                                                                                          'fiber_length': 23025,
                                                                                                                                          'poller_hash': 'OTU2RTQ2cGVlcnBvbnBvbiAxNDUyMTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '9748914',
                                                                                                                                          'fiber_length': 23287,
                                                                                                                                          'poller_hash': 'OTU2RTU0cGVlcnBvbnBvbiAxNDUyNDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535345',
                                                                                                                                          'fiber_length': 23590,
                                                                                                                                          'poller_hash': 'OTU2RTFFcGVlcnBvbnBvbiAxNDUyOTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535331',
                                                                                                                                          'fiber_length': 23377,
                                                                                                                                          'poller_hash': 'OTU2RTMxcGVlcnBvbnBvbiAxNDU1Njg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535310',
                                                                                                                                          'fiber_length': 23708,
                                                                                                                                          'poller_hash': 'OTU2RjkwcGVlcnBvbnBvbiAxNDU3MzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535286',
                                                                                                                                          'fiber_length': 23148,
                                                                                                                                          'poller_hash': 'QTlDNDE5cGVlcnBvbnBvbiAxNDYwNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535276',
                                                                                                                                          'fiber_length': 23449,
                                                                                                                                          'poller_hash': 'QzgxRUFCcGVlcnBvbnBvbiAyMDI2NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '8535234',
                                                                                                                                          'fiber_length': 23308,
                                                                                                                                          'poller_hash': 'MTI3NDk2QnBlZXJwb25wb24gMjM1NzEz',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -32.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203},
                                                                                                                                  {       'connected_time': '2994217',
                                                                                                                                          'fiber_length': 23449,
                                                                                                                                          'poller_hash': 'MTI3M0Y1OHBlZXJwb25wb24gMjQ3MDIw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.203}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/2',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/3',
                                                                                         'in_discards': '424817',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '4407159',
                                                                                         'in_octets': '43031383768495',
                                                                                         'in_ucast_pkts': '131666292275',
                                                                                         'links': [       {       'in_nucast_pkts': '42',
                                                                                                                  'in_octets': '238221952467',
                                                                                                                  'in_ucast_pkts': '734757595',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15247182',
                                                                                                                  'out_octets': '3721353188000',
                                                                                                                  'out_ucast_pkts': '2882463944',
                                                                                                                  'poller_hash': 'NTEyNzMycGVlcmludGZwb24gMjg5'},
                                                                                                          {       'in_nucast_pkts': '11820',
                                                                                                                  'in_octets': '1788954217519',
                                                                                                                  'in_ucast_pkts': '3046414785',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62537719',
                                                                                                                  'out_octets': '7495857184092',
                                                                                                                  'out_ucast_pkts': '6004428677',
                                                                                                                  'poller_hash': 'MzIyOTRDcGVlcmludGZwb24gMjk0'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '89297799324',
                                                                                                                  'in_ucast_pkts': '476447870',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797946',
                                                                                                                  'out_octets': '1599355253439',
                                                                                                                  'out_ucast_pkts': '1127670241',
                                                                                                                  'poller_hash': 'MzIyOTAzcGVlcmludGZwb24gMjk1'},
                                                                                                          {       'in_nucast_pkts': '2570',
                                                                                                                  'in_octets': '34428683739',
                                                                                                                  'in_ucast_pkts': '130026461',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530410',
                                                                                                                  'out_octets': '413523205286',
                                                                                                                  'out_ucast_pkts': '345177675',
                                                                                                                  'poller_hash': 'NEZBMjQ3cGVlcmludGZwb24gMjk3'},
                                                                                                          {       'in_nucast_pkts': '26972659',
                                                                                                                  'in_octets': '438832582488',
                                                                                                                  'in_ucast_pkts': '2394308559',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '26297412834',
                                                                                                                  'out_octets': '44515973023648',
                                                                                                                  'out_ucast_pkts': '7050970844',
                                                                                                                  'poller_hash': 'MzIxMThDcGVlcmludGZwb24gMjk5'},
                                                                                                          {       'in_nucast_pkts': '9554',
                                                                                                                  'in_octets': '192128548954',
                                                                                                                  'in_ucast_pkts': '858954285',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62541723',
                                                                                                                  'out_octets': '3325669541378',
                                                                                                                  'out_ucast_pkts': '2476221918',
                                                                                                                  'poller_hash': 'MzIyOTA3cGVlcmludGZwb24gMzAw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDJCOUI1cGVlcmludGZwb24gMzAy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NDI4MUZCcGVlcmludGZwb24gMzA1'},
                                                                                                          {       'in_nucast_pkts': '8',
                                                                                                                  'in_octets': '106429346226',
                                                                                                                  'in_ucast_pkts': '574896588',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530459',
                                                                                                                  'out_octets': '2867086059576',
                                                                                                                  'out_ucast_pkts': '2036788268',
                                                                                                                  'poller_hash': 'NDNCMzE5cGVlcmludGZwb24gMzA2'},
                                                                                                          {       'in_nucast_pkts': '550161',
                                                                                                                  'in_octets': '299652919442',
                                                                                                                  'in_ucast_pkts': '3114167925',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '32282471',
                                                                                                                  'out_octets': '9008061225855',
                                                                                                                  'out_ucast_pkts': '6205775920',
                                                                                                                  'poller_hash': 'MzIyODQ3cGVlcmludGZwb24gMzA4'},
                                                                                                          {       'in_nucast_pkts': '8672',
                                                                                                                  'in_octets': '153177722195',
                                                                                                                  'in_ucast_pkts': '352691183',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530310',
                                                                                                                  'out_octets': '920290385552',
                                                                                                                  'out_ucast_pkts': '780125252',
                                                                                                                  'poller_hash': 'NDNCM0I4cGVlcmludGZwb24gMzEz'},
                                                                                                          {       'in_nucast_pkts': '28901',
                                                                                                                  'in_octets': '13615991213969',
                                                                                                                  'in_ucast_pkts': '37502976656',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38469367',
                                                                                                                  'out_octets': '17558818234217',
                                                                                                                  'out_ucast_pkts': '45938305691',
                                                                                                                  'poller_hash': 'NTEyNUY4cGVlcmludGZwb24gMzE5'},
                                                                                                          {       'in_nucast_pkts': '11961',
                                                                                                                  'in_octets': '887448087193',
                                                                                                                  'in_ucast_pkts': '5167786253',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62542202',
                                                                                                                  'out_octets': '19184215617004',
                                                                                                                  'out_ucast_pkts': '14180667945',
                                                                                                                  'poller_hash': 'NDJCOUMwcGVlcmludGZwb24gMzI0'},
                                                                                                          {       'in_nucast_pkts': '694',
                                                                                                                  'in_octets': '1640300951901',
                                                                                                                  'in_ucast_pkts': '3969859818',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '6156076',
                                                                                                                  'out_octets': '7292589444029',
                                                                                                                  'out_ucast_pkts': '1141002173',
                                                                                                                  'poller_hash': 'NTc0NTdFcGVlcmludGZwb24gMzI3'},
                                                                                                          {       'in_nucast_pkts': '50',
                                                                                                                  'in_octets': '118373997083',
                                                                                                                  'in_ucast_pkts': '402921090',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797544',
                                                                                                                  'out_octets': '1623902777707',
                                                                                                                  'out_ucast_pkts': '1232016833',
                                                                                                                  'poller_hash': 'NTc0M0REcGVlcmludGZwb24gMzI5'},
                                                                                                          {       'in_nucast_pkts': '35089147',
                                                                                                                  'in_octets': '188957807820',
                                                                                                                  'in_ucast_pkts': '1857375764',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8963773107',
                                                                                                                  'out_octets': '21472063496709',
                                                                                                                  'out_ucast_pkts': '6765627918',
                                                                                                                  'poller_hash': 'NDJCOUY3cGVlcmludGZwb24gMzMy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '436663092689',
                                                                                                                  'in_ucast_pkts': '999982556',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '5299660405439',
                                                                                                                  'out_ucast_pkts': '3894911624',
                                                                                                                  'poller_hash': 'NDJCOUU5cGVlcmludGZwb24gMzM5'},
                                                                                                          {       'in_nucast_pkts': '44',
                                                                                                                  'in_octets': '363163613015',
                                                                                                                  'in_ucast_pkts': '4371166814',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38275452',
                                                                                                                  'out_octets': '23865169300902',
                                                                                                                  'out_ucast_pkts': '15878274045',
                                                                                                                  'poller_hash': 'NDJCOURDcGVlcmludGZwb24gMzUx'},
                                                                                                          {       'in_nucast_pkts': '68',
                                                                                                                  'in_octets': '588101596727',
                                                                                                                  'in_ucast_pkts': '817020341',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '45072938',
                                                                                                                  'out_octets': '2006741708667',
                                                                                                                  'out_ucast_pkts': '1850204835',
                                                                                                                  'poller_hash': 'NTEyNUMzcGVlcmludGZwb24gMzU2'},
                                                                                                          {       'in_nucast_pkts': '6834',
                                                                                                                  'in_octets': '38142530714',
                                                                                                                  'in_ucast_pkts': '289612497',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5499530',
                                                                                                                  'out_octets': '1258159581345',
                                                                                                                  'out_ucast_pkts': '892507840',
                                                                                                                  'poller_hash': 'NDNCNUI2cGVlcmludGZwb24gMzgz'},
                                                                                                          {       'in_nucast_pkts': '56',
                                                                                                                  'in_octets': '216585420505',
                                                                                                                  'in_ucast_pkts': '1177364562',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38464281',
                                                                                                                  'out_octets': '5441363382866',
                                                                                                                  'out_ucast_pkts': '4054651133',
                                                                                                                  'poller_hash': 'NTEyNUU5cGVlcmludGZwb24gMzg1'},
                                                                                                          {       'in_nucast_pkts': '118428',
                                                                                                                  'in_octets': '1434748618258',
                                                                                                                  'in_ucast_pkts': '3649198448',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '42407059',
                                                                                                                  'out_octets': '11176338817054',
                                                                                                                  'out_ucast_pkts': '8680444622',
                                                                                                                  'poller_hash': 'NDJCOUNFcGVlcmludGZwb24gMzg5'},
                                                                                                          {       'in_nucast_pkts': '235537',
                                                                                                                  'in_octets': '1098328894615',
                                                                                                                  'in_ucast_pkts': '1467437943',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '13895039',
                                                                                                                  'out_octets': '4956805661516',
                                                                                                                  'out_ucast_pkts': '3754894800',
                                                                                                                  'poller_hash': 'NUE2MkRFcGVlcmludGZwb24gNDA5'},
                                                                                                          {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '86058148593',
                                                                                                                  'in_ucast_pkts': '491071850',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797994',
                                                                                                                  'out_octets': '2280726982050',
                                                                                                                  'out_ucast_pkts': '1680265408',
                                                                                                                  'poller_hash': 'NDJCRTlDcGVlcmludGZwb24gNDE2'},
                                                                                                          {       'in_nucast_pkts': '10142',
                                                                                                                  'in_octets': '116428282081',
                                                                                                                  'in_ucast_pkts': '672618830',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530434',
                                                                                                                  'out_octets': '2257224188737',
                                                                                                                  'out_ucast_pkts': '1652284089',
                                                                                                                  'poller_hash': 'NjQwQjdEcGVlcmludGZwb24gNDI2'},
                                                                                                          {       'in_nucast_pkts': '7214920',
                                                                                                                  'in_octets': '132971512733',
                                                                                                                  'in_ucast_pkts': '575178936',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10711193821',
                                                                                                                  'out_octets': '16977517348736',
                                                                                                                  'out_ucast_pkts': '1822475515',
                                                                                                                  'poller_hash': 'NDJCRjFFcGVlcmludGZwb24gNDQ1'},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '91380112769',
                                                                                                                  'in_ucast_pkts': '666617100',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797967',
                                                                                                                  'out_octets': '2164060312464',
                                                                                                                  'out_ucast_pkts': '1586187745',
                                                                                                                  'poller_hash': 'NDJDOTE1cGVlcmludGZwb24gNDQ5'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '68192929767',
                                                                                                                  'in_ucast_pkts': '312899056',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797942',
                                                                                                                  'out_octets': '1745112863239',
                                                                                                                  'out_ucast_pkts': '1331293403',
                                                                                                                  'poller_hash': 'NDJDNDU5cGVlcmludGZwb24gNDU2'},
                                                                                                          {       'in_nucast_pkts': '160136',
                                                                                                                  'in_octets': '72612870151',
                                                                                                                  'in_ucast_pkts': '150723969',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797820',
                                                                                                                  'out_octets': '465348975786',
                                                                                                                  'out_ucast_pkts': '416091324',
                                                                                                                  'poller_hash': 'Q0JEMUZBcGVlcmludGZwb24gMTY0ODI1'},
                                                                                                          {       'in_nucast_pkts': '161781',
                                                                                                                  'in_octets': '57334070549',
                                                                                                                  'in_ucast_pkts': '240080486',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797949',
                                                                                                                  'out_octets': '953269362789',
                                                                                                                  'out_ucast_pkts': '713977350',
                                                                                                                  'poller_hash': 'QzgxRUVCcGVlcmludGZwb24gMTk4MDI4'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '7/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '223982703994',
                                                                                         'out_octets': '618002291343184',
                                                                                         'out_ucast_pkts': '273579054366',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/3',
                                                                                                                 'id': '7/3',
                                                                                                                 'links': [       {       'connected_time': '13227781',
                                                                                                                                          'fiber_length': 23940,
                                                                                                                                          'poller_hash': 'NTEyNzMycGVlcnBvbnBvbiAyODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '21886294',
                                                                                                                                          'fiber_length': 23947,
                                                                                                                                          'poller_hash': 'MzIyOTRDcGVlcnBvbnBvbiAyOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535242',
                                                                                                                                          'fiber_length': 24499,
                                                                                                                                          'poller_hash': 'MzIyOTAzcGVlcnBvbnBvbiAyOTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535107',
                                                                                                                                          'fiber_length': 24784,
                                                                                                                                          'poller_hash': 'NEZBMjQ3cGVlcnBvbnBvbiAyOTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '21886492',
                                                                                                                                          'fiber_length': 24693,
                                                                                                                                          'poller_hash': 'MzIxMThDcGVlcnBvbnBvbiAyOTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '21886928',
                                                                                                                                          'fiber_length': 24794,
                                                                                                                                          'poller_hash': 'MzIyOTA3cGVlcnBvbnBvbiAzMDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535243',
                                                                                                                                          'fiber_length': 24888,
                                                                                                                                          'poller_hash': 'NDJCOUI1cGVlcnBvbnBvbiAzMDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '21887218',
                                                                                                                                          'fiber_length': 23997,
                                                                                                                                          'poller_hash': 'NDI4MUZCcGVlcnBvbnBvbiAzMDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535079',
                                                                                                                                          'fiber_length': 24459,
                                                                                                                                          'poller_hash': 'NDNCMzE5cGVlcnBvbnBvbiAzMDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '29479976',
                                                                                                                                          'fiber_length': 24047,
                                                                                                                                          'poller_hash': 'MzIyODQ3cGVlcnBvbnBvbiAzMDg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535085',
                                                                                                                                          'fiber_length': 23875,
                                                                                                                                          'poller_hash': 'NDNCM0I4cGVlcnBvbnBvbiAzMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '67856266',
                                                                                                                                          'fiber_length': 23948,
                                                                                                                                          'poller_hash': 'NTEyNUY4cGVlcnBvbnBvbiAzMTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '21887381',
                                                                                                                                          'fiber_length': 23986,
                                                                                                                                          'poller_hash': 'NDJCOUMwcGVlcnBvbnBvbiAzMjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '5339178',
                                                                                                                                          'fiber_length': 24251,
                                                                                                                                          'poller_hash': 'NTc0NTdFcGVlcnBvbnBvbiAzMjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535216',
                                                                                                                                          'fiber_length': 24036,
                                                                                                                                          'poller_hash': 'NTc0M0REcGVlcnBvbnBvbiAzMjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '21886814',
                                                                                                                                          'fiber_length': 24818,
                                                                                                                                          'poller_hash': 'NDJCOUY3cGVlcnBvbnBvbiAzMzI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '2594924',
                                                                                                                                          'fiber_length': 24161,
                                                                                                                                          'poller_hash': 'NDJCOUU5cGVlcnBvbnBvbiAzMzk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '67648439',
                                                                                                                                          'fiber_length': 24687,
                                                                                                                                          'poller_hash': 'NDJCOURDcGVlcnBvbnBvbiAzNTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '67856116',
                                                                                                                                          'fiber_length': 24294,
                                                                                                                                          'poller_hash': 'NTEyNUMzcGVlcnBvbnBvbiAzNTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8485026',
                                                                                                                                          'fiber_length': 24184,
                                                                                                                                          'poller_hash': 'NDNCNUI2cGVlcnBvbnBvbiAzODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '67856209',
                                                                                                                                          'fiber_length': 24242,
                                                                                                                                          'poller_hash': 'NTEyNUU5cGVlcnBvbnBvbiAzODU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '72550404',
                                                                                                                                          'fiber_length': 23869,
                                                                                                                                          'poller_hash': 'NDJCOUNFcGVlcnBvbnBvbiAzODk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '12093658',
                                                                                                                                          'fiber_length': 23962,
                                                                                                                                          'poller_hash': 'NUE2MkRFcGVlcnBvbnBvbiA0MDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535226',
                                                                                                                                          'fiber_length': 24256,
                                                                                                                                          'poller_hash': 'NDJCRTlDcGVlcnBvbnBvbiA0MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535117',
                                                                                                                                          'fiber_length': 24708,
                                                                                                                                          'poller_hash': 'NjQwQjdEcGVlcnBvbnBvbiA0MjY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '5886982',
                                                                                                                                          'fiber_length': 24806,
                                                                                                                                          'poller_hash': 'NDJCRjFFcGVlcnBvbnBvbiA0NDU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535310',
                                                                                                                                          'fiber_length': 23889,
                                                                                                                                          'poller_hash': 'NDJDOTE1cGVlcnBvbnBvbiA0NDk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535256',
                                                                                                                                          'fiber_length': 24293,
                                                                                                                                          'poller_hash': 'NDJDNDU5cGVlcnBvbnBvbiA0NTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535213',
                                                                                                                                          'fiber_length': 24146,
                                                                                                                                          'poller_hash': 'Q0JEMUZBcGVlcnBvbnBvbiAxNjQ4MjU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613},
                                                                                                                                  {       'connected_time': '8535334',
                                                                                                                                          'fiber_length': 24973,
                                                                                                                                          'poller_hash': 'QzgxRUVCcGVlcnBvbnBvbiAxOTgwMjg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.613}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/3',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/4',
                                                                                         'in_discards': '806974',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '40701373',
                                                                                         'in_octets': '57748835692893',
                                                                                         'in_ucast_pkts': '110154287790',
                                                                                         'links': [       {       'in_nucast_pkts': '764612',
                                                                                                                  'in_octets': '36502713383',
                                                                                                                  'in_ucast_pkts': '220964306',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '8526883',
                                                                                                                  'out_octets': '631334332830',
                                                                                                                  'out_ucast_pkts': '467872460',
                                                                                                                  'poller_hash': 'ODY5M0I0cGVlcmludGZwb24gMTgx'},
                                                                                                          {       'in_nucast_pkts': '147816',
                                                                                                                  'in_octets': '156746241590',
                                                                                                                  'in_ucast_pkts': '495004379',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797846',
                                                                                                                  'out_octets': '3149479908137',
                                                                                                                  'out_ucast_pkts': '2427828354',
                                                                                                                  'poller_hash': 'QzgxRjEycGVlcmludGZwb24gNDE3'},
                                                                                                          {       'in_nucast_pkts': '13',
                                                                                                                  'in_octets': '151530127491',
                                                                                                                  'in_ucast_pkts': '811002677',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9732598',
                                                                                                                  'out_octets': '2976343922741',
                                                                                                                  'out_ucast_pkts': '2125463187',
                                                                                                                  'poller_hash': 'NDJDQTY2cGVlcmludGZwb24gNDY5'},
                                                                                                          {       'in_nucast_pkts': '23',
                                                                                                                  'in_octets': '221411946378',
                                                                                                                  'in_ucast_pkts': '1001226563',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '13088509',
                                                                                                                  'out_octets': '5456587605192',
                                                                                                                  'out_ucast_pkts': '3918998241',
                                                                                                                  'poller_hash': 'NDJCRjdGcGVlcmludGZwb24gNDcw'},
                                                                                                          {       'in_nucast_pkts': '30',
                                                                                                                  'in_octets': '134563693682',
                                                                                                                  'in_ucast_pkts': '341416440',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530468',
                                                                                                                  'out_octets': '1107274120843',
                                                                                                                  'out_ucast_pkts': '841041107',
                                                                                                                  'poller_hash': 'NjQwQjRFcGVlcmludGZwb24gNDc4'},
                                                                                                          {       'in_nucast_pkts': '15603216',
                                                                                                                  'in_octets': '378171460349',
                                                                                                                  'in_ucast_pkts': '1925862158',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '31294418370',
                                                                                                                  'out_octets': '54110912468042',
                                                                                                                  'out_ucast_pkts': '8743140080',
                                                                                                                  'poller_hash': 'NzIyMjE4cGVlcmludGZwb24gNDk0'},
                                                                                                          {       'in_nucast_pkts': '49933',
                                                                                                                  'in_octets': '767240745',
                                                                                                                  'in_ucast_pkts': '6200976',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9678537',
                                                                                                                  'out_octets': '5434528477',
                                                                                                                  'out_ucast_pkts': '12779776',
                                                                                                                  'poller_hash': 'NDNCNDFBcGVlcmludGZwb24gNTAz'},
                                                                                                          {       'in_nucast_pkts': '71503',
                                                                                                                  'in_octets': '212210761630',
                                                                                                                  'in_ucast_pkts': '759090838',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530445',
                                                                                                                  'out_octets': '3212157169876',
                                                                                                                  'out_ucast_pkts': '2395653777',
                                                                                                                  'poller_hash': 'NjQwQjUycGVlcmludGZwb24gNTA2'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NjQwQUZFcGVlcmludGZwb24gNTEz'},
                                                                                                          {       'in_nucast_pkts': '14793',
                                                                                                                  'in_octets': '29488289180',
                                                                                                                  'in_ucast_pkts': '91220101',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530351',
                                                                                                                  'out_octets': '573389722272',
                                                                                                                  'out_ucast_pkts': '418160510',
                                                                                                                  'poller_hash': 'NjQwQUY2cGVlcmludGZwb24gNTM3'},
                                                                                                          {       'in_nucast_pkts': '12223',
                                                                                                                  'in_octets': '216523051750',
                                                                                                                  'in_ucast_pkts': '1408814027',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '38463292',
                                                                                                                  'out_octets': '8447665564856',
                                                                                                                  'out_ucast_pkts': '5858540943',
                                                                                                                  'poller_hash': 'NUE2OUJDcGVlcmludGZwb24gMTM3Nzgz'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '40220557255',
                                                                                                                  'in_ucast_pkts': '77448230',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530454',
                                                                                                                  'out_octets': '148585922607',
                                                                                                                  'out_ucast_pkts': '130296789',
                                                                                                                  'poller_hash': 'NjQwMEYzcGVlcmludGZwb24gMTM4MTAy'},
                                                                                                          {       'in_nucast_pkts': '958335',
                                                                                                                  'in_octets': '151504799534',
                                                                                                                  'in_ucast_pkts': '495230176',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62544524',
                                                                                                                  'out_octets': '987205040885',
                                                                                                                  'out_ucast_pkts': '972006426',
                                                                                                                  'poller_hash': 'NkRENTI0cGVlcmludGZwb24gMTM4NDIz'},
                                                                                                          {       'in_nucast_pkts': '73599',
                                                                                                                  'in_octets': '312850740878',
                                                                                                                  'in_ucast_pkts': '339104162',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '2395957',
                                                                                                                  'out_octets': '950852963816',
                                                                                                                  'out_ucast_pkts': '816032177',
                                                                                                                  'poller_hash': 'NUE2Qjk0cGVlcmludGZwb24gMTM4NzE2'},
                                                                                                          {       'in_nucast_pkts': '73824',
                                                                                                                  'in_octets': '81764913770',
                                                                                                                  'in_ucast_pkts': '529600773',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9242228',
                                                                                                                  'out_octets': '3112411466480',
                                                                                                                  'out_ucast_pkts': '2180439438',
                                                                                                                  'poller_hash': 'ODJBNDBDcGVlcmludGZwb24gMTM5OTQ2'},
                                                                                                          {       'in_nucast_pkts': '4906',
                                                                                                                  'in_octets': '97157547235',
                                                                                                                  'in_ucast_pkts': '295034102',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9798032',
                                                                                                                  'out_octets': '1011248290260',
                                                                                                                  'out_ucast_pkts': '822671685',
                                                                                                                  'poller_hash': 'ODJBNzQ0cGVlcmludGZwb24gMTM5OTU0'},
                                                                                                          {       'in_nucast_pkts': '168044',
                                                                                                                  'in_octets': '59245524239',
                                                                                                                  'in_ucast_pkts': '206049500',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797903',
                                                                                                                  'out_octets': '951330844339',
                                                                                                                  'out_ucast_pkts': '715510713',
                                                                                                                  'poller_hash': 'ODJBNEVBcGVlcmludGZwb24gMTQwNjI5'},
                                                                                                          {       'in_nucast_pkts': '161085',
                                                                                                                  'in_octets': '266086606275',
                                                                                                                  'in_ucast_pkts': '965700022',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797872',
                                                                                                                  'out_octets': '3596175167022',
                                                                                                                  'out_ucast_pkts': '2740061179',
                                                                                                                  'poller_hash': 'ODQ0QkYycGVlcmludGZwb24gMTQwNjkz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '610259750851',
                                                                                                                  'in_ucast_pkts': '48330922313',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '69',
                                                                                                                  'out_octets': '2742689711556',
                                                                                                                  'out_ucast_pkts': '2327112673',
                                                                                                                  'poller_hash': 'ODcyNDE4cGVlcmludGZwb24gMTQzODQy'},
                                                                                                          {       'in_nucast_pkts': '202770',
                                                                                                                  'in_octets': '239625932782',
                                                                                                                  'in_ucast_pkts': '1392898948',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9798083',
                                                                                                                  'out_octets': '5511720542588',
                                                                                                                  'out_ucast_pkts': '4006265040',
                                                                                                                  'poller_hash': 'MzVFOTZFcGVlcmludGZwb24gMTQzODUy'},
                                                                                                          {       'in_nucast_pkts': '34',
                                                                                                                  'in_octets': '389340886831',
                                                                                                                  'in_ucast_pkts': '1842321151',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797397',
                                                                                                                  'out_octets': '4233152143546',
                                                                                                                  'out_ucast_pkts': '3780712355',
                                                                                                                  'poller_hash': 'QzgzQzIwcGVlcmludGZwb24gMTY2ODM4'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '495132749696',
                                                                                                                  'in_ucast_pkts': '4704141158',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '153',
                                                                                                                  'out_octets': '19247871353985',
                                                                                                                  'out_ucast_pkts': '13300642254',
                                                                                                                  'poller_hash': 'Q0JEMUI4cGVlcmludGZwb24gMTc1ODAy'},
                                                                                                          {       'in_nucast_pkts': '301',
                                                                                                                  'in_octets': '278225036949',
                                                                                                                  'in_ucast_pkts': '2118071520',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '552',
                                                                                                                  'out_octets': '6663471212577',
                                                                                                                  'out_ucast_pkts': '4972625411',
                                                                                                                  'poller_hash': 'QjBBQUY3cGVlcmludGZwb24gMTgwNjMw'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '16957581619',
                                                                                                                  'in_ucast_pkts': '47754057',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '103',
                                                                                                                  'out_octets': '124915731268',
                                                                                                                  'out_ucast_pkts': '127872760',
                                                                                                                  'poller_hash': 'QzgxRTA4cGVlcmludGZwb24gMjA4OTk2'},
                                                                                                          {       'in_nucast_pkts': '162246',
                                                                                                                  'in_octets': '118128095034',
                                                                                                                  'in_ucast_pkts': '539602273',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797732',
                                                                                                                  'out_octets': '4399793259749',
                                                                                                                  'out_ucast_pkts': '3270986244',
                                                                                                                  'poller_hash': 'MTI3NDA0OHBlZXJpbnRmcG9uIDI0MzczMQ=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05071',
                                                                                         'name': 'gpon '
                                                                                                 '7/4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '44353426304',
                                                                                         'out_octets': '390538306592971',
                                                                                         'out_ucast_pkts': '256289284278',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/4',
                                                                                                                 'id': '7/4',
                                                                                                                 'links': [       {       'connected_time': '7183310',
                                                                                                                                          'fiber_length': 23896,
                                                                                                                                          'poller_hash': 'ODY5M0I0cGVlcnBvbnBvbiAxODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535171',
                                                                                                                                          'fiber_length': 23392,
                                                                                                                                          'poller_hash': 'QzgxRjEycGVlcnBvbnBvbiA0MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8478876',
                                                                                                                                          'fiber_length': 23381,
                                                                                                                                          'poller_hash': 'NDJDQTY2cGVlcnBvbnBvbiA0Njk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '11398737',
                                                                                                                                          'fiber_length': 23080,
                                                                                                                                          'poller_hash': 'NDJCRjdGcGVlcnBvbnBvbiA0NzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535132',
                                                                                                                                          'fiber_length': 22958,
                                                                                                                                          'poller_hash': 'NjQwQjRFcGVlcnBvbnBvbiA0Nzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '552653',
                                                                                                                                          'fiber_length': 23440,
                                                                                                                                          'poller_hash': 'NzIyMjE4cGVlcnBvbnBvbiA0OTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '14604827',
                                                                                                                                          'fiber_length': 23804,
                                                                                                                                          'poller_hash': 'NDNCNDFBcGVlcnBvbnBvbiA1MDM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535099',
                                                                                                                                          'fiber_length': 23831,
                                                                                                                                          'poller_hash': 'NjQwQjUycGVlcnBvbnBvbiA1MDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535122',
                                                                                                                                          'fiber_length': 23448,
                                                                                                                                          'poller_hash': 'NjQwQUZFcGVlcnBvbnBvbiA1MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535118',
                                                                                                                                          'fiber_length': 23437,
                                                                                                                                          'poller_hash': 'NjQwQUY2cGVlcnBvbnBvbiA1Mzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '67856057',
                                                                                                                                          'fiber_length': 23020,
                                                                                                                                          'poller_hash': 'NUE2OUJDcGVlcnBvbnBvbiAxMzc3ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535108',
                                                                                                                                          'fiber_length': 23046,
                                                                                                                                          'poller_hash': 'NjQwMEYzcGVlcnBvbnBvbiAxMzgxMDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '21886650',
                                                                                                                                          'fiber_length': 24006,
                                                                                                                                          'poller_hash': 'NkRENTI0cGVlcnBvbnBvbiAxMzg0MjM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '2060312',
                                                                                                                                          'fiber_length': 23427,
                                                                                                                                          'poller_hash': 'NUE2Qjk0cGVlcnBvbnBvbiAxMzg3MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8046361',
                                                                                                                                          'fiber_length': 23411,
                                                                                                                                          'poller_hash': 'ODJBNDBDcGVlcnBvbnBvbiAxMzk5NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535477',
                                                                                                                                          'fiber_length': 23946,
                                                                                                                                          'poller_hash': 'ODJBNzQ0cGVlcnBvbnBvbiAxMzk5NTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535552',
                                                                                                                                          'fiber_length': 23304,
                                                                                                                                          'poller_hash': 'ODJBNEVBcGVlcnBvbnBvbiAxNDA2Mjk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535548',
                                                                                                                                          'fiber_length': 23029,
                                                                                                                                          'poller_hash': 'ODQ0QkYycGVlcnBvbnBvbiAxNDA2OTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535212',
                                                                                                                                          'fiber_length': 23212,
                                                                                                                                          'poller_hash': 'ODcyNDE4cGVlcnBvbnBvbiAxNDM4NDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535470',
                                                                                                                                          'fiber_length': 23546,
                                                                                                                                          'poller_hash': 'MzVFOTZFcGVlcnBvbnBvbiAxNDM4NTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535267',
                                                                                                                                          'fiber_length': 23900,
                                                                                                                                          'poller_hash': 'QzgzQzIwcGVlcnBvbnBvbiAxNjY4Mzg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535297',
                                                                                                                                          'fiber_length': 23499,
                                                                                                                                          'poller_hash': 'Q0JEMUI4cGVlcnBvbnBvbiAxNzU4MDI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535388',
                                                                                                                                          'fiber_length': 23501,
                                                                                                                                          'poller_hash': 'QjBBQUY3cGVlcnBvbnBvbiAxODA2MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535248',
                                                                                                                                          'fiber_length': 23892,
                                                                                                                                          'poller_hash': 'QzgxRTA4cGVlcnBvbnBvbiAyMDg5OTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152},
                                                                                                                                  {       'connected_time': '8535208',
                                                                                                                                          'fiber_length': 23083,
                                                                                                                                          'poller_hash': 'MTI3NDA0OHBlZXJwb25wb24gMjQzNzMx',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -21.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.152}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/4',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/5',
                                                                                         'in_discards': '61775',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7638364',
                                                                                         'in_octets': '131793244001674',
                                                                                         'in_ucast_pkts': '185748128382',
                                                                                         'links': [       {       'in_nucast_pkts': '35229145',
                                                                                                                  'in_octets': '525002300658',
                                                                                                                  'in_ucast_pkts': '2316500628',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62522896',
                                                                                                                  'out_octets': '7067492306618',
                                                                                                                  'out_ucast_pkts': '5507948556',
                                                                                                                  'poller_hash': 'NDJCOURGcGVlcmludGZwb24gMzc3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '293295793808',
                                                                                                                  'in_ucast_pkts': '1182716930',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '50',
                                                                                                                  'out_octets': '5930673199226',
                                                                                                                  'out_ucast_pkts': '4412364627',
                                                                                                                  'poller_hash': 'NDJEMzkxcGVlcmludGZwb24gMzk1'},
                                                                                                          {       'in_nucast_pkts': '40490209',
                                                                                                                  'in_octets': '1349087950404',
                                                                                                                  'in_ucast_pkts': '7744542290',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '102157373540',
                                                                                                                  'out_octets': '173138980398546',
                                                                                                                  'out_ucast_pkts': '25632041157',
                                                                                                                  'poller_hash': 'NDJEMkE2cGVlcmludGZwb24gNDEz'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'NjQwQjc0cGVlcmludGZwb24gNDIx'},
                                                                                                          {       'in_nucast_pkts': '158168',
                                                                                                                  'in_octets': '163308563220',
                                                                                                                  'in_ucast_pkts': '1200082490',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '10117648',
                                                                                                                  'out_octets': '4359800329038',
                                                                                                                  'out_ucast_pkts': '3098234753',
                                                                                                                  'poller_hash': 'NDJDOEMxcGVlcmludGZwb24gNDI0'},
                                                                                                          {       'in_nucast_pkts': '717',
                                                                                                                  'in_octets': '774739914353',
                                                                                                                  'in_ucast_pkts': '2029852838',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '42953708',
                                                                                                                  'out_octets': '8905954781741',
                                                                                                                  'out_ucast_pkts': '6620424412',
                                                                                                                  'poller_hash': 'NDJDOTBGcGVlcmludGZwb24gNDUz'},
                                                                                                          {       'in_nucast_pkts': '9620000',
                                                                                                                  'in_octets': '172879699572',
                                                                                                                  'in_ucast_pkts': '551992976',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797692',
                                                                                                                  'out_octets': '1477300921671',
                                                                                                                  'out_ucast_pkts': '1114185964',
                                                                                                                  'poller_hash': 'NDJDMjE4cGVlcmludGZwb24gNDU0'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '451138634334',
                                                                                                                  'in_ucast_pkts': '1669833178',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '88',
                                                                                                                  'out_octets': '11763458179578',
                                                                                                                  'out_ucast_pkts': '8239713937',
                                                                                                                  'poller_hash': 'NDJDREYzcGVlcmludGZwb24gNDgx'},
                                                                                                          {       'in_nucast_pkts': '6484052',
                                                                                                                  'in_octets': '294507922206',
                                                                                                                  'in_ucast_pkts': '1021256095',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '11571469093',
                                                                                                                  'out_octets': '22033046910435',
                                                                                                                  'out_ucast_pkts': '4524969456',
                                                                                                                  'poller_hash': 'NDJDNkY0cGVlcmludGZwb24gNDg3'},
                                                                                                          {       'in_nucast_pkts': '612268',
                                                                                                                  'in_octets': '3844195334413',
                                                                                                                  'in_ucast_pkts': '4862622772',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797335',
                                                                                                                  'out_octets': '6430012722685',
                                                                                                                  'out_ucast_pkts': '6317668199',
                                                                                                                  'poller_hash': 'NDJDNkE5cGVlcmludGZwb24gNDk5'},
                                                                                                          {       'in_nucast_pkts': '5654471',
                                                                                                                  'in_octets': '59743148778',
                                                                                                                  'in_ucast_pkts': '138971228',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '918941421',
                                                                                                                  'out_octets': '1788776419154',
                                                                                                                  'out_ucast_pkts': '485193924',
                                                                                                                  'poller_hash': 'NDJDNkRGcGVlcmludGZwb24gNTAw'},
                                                                                                          {       'in_nucast_pkts': '26',
                                                                                                                  'in_octets': '167193262018',
                                                                                                                  'in_ucast_pkts': '911443835',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14628996',
                                                                                                                  'out_octets': '3266445525640',
                                                                                                                  'out_ucast_pkts': '2328600785',
                                                                                                                  'poller_hash': 'NDJDQTA5cGVlcmludGZwb24gNTEx'},
                                                                                                          {       'in_nucast_pkts': '25152',
                                                                                                                  'in_octets': '27029650842',
                                                                                                                  'in_ucast_pkts': '125478927',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '1603040',
                                                                                                                  'out_octets': '401618939442',
                                                                                                                  'out_ucast_pkts': '338750864',
                                                                                                                  'poller_hash': 'NUE2NjVGcGVlcmludGZwb24gNTE0'},
                                                                                                          {       'in_nucast_pkts': '6397',
                                                                                                                  'in_octets': '22256473742',
                                                                                                                  'in_ucast_pkts': '93361267',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5370437',
                                                                                                                  'out_octets': '579828243568',
                                                                                                                  'out_ucast_pkts': '421201977',
                                                                                                                  'poller_hash': 'NDJDOTNDcGVlcmludGZwb24gNTE2'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '437804056239',
                                                                                                                  'in_ucast_pkts': '737270544',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5410189',
                                                                                                                  'out_octets': '649677100210',
                                                                                                                  'out_ucast_pkts': '990510536',
                                                                                                                  'poller_hash': 'NjQwQUM2cGVlcmludGZwb24gNTE3'},
                                                                                                          {       'in_nucast_pkts': '12',
                                                                                                                  'in_octets': '88317740425',
                                                                                                                  'in_ucast_pkts': '385626709',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797775',
                                                                                                                  'out_octets': '1345940157621',
                                                                                                                  'out_ucast_pkts': '1099181791',
                                                                                                                  'poller_hash': 'NUE2NUVEcGVlcmludGZwb24gNTE4'},
                                                                                                          {       'in_nucast_pkts': '4',
                                                                                                                  'in_octets': '186389502',
                                                                                                                  'in_ucast_pkts': '555692',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530209',
                                                                                                                  'out_octets': '617274363',
                                                                                                                  'out_ucast_pkts': '1825065',
                                                                                                                  'poller_hash': 'NjQwMTBFcGVlcmludGZwb24gNTIx'},
                                                                                                          {       'in_nucast_pkts': '924845',
                                                                                                                  'in_octets': '613796995284',
                                                                                                                  'in_ucast_pkts': '2746501605',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '62515985',
                                                                                                                  'out_octets': '10127946422436',
                                                                                                                  'out_ucast_pkts': '7517149569',
                                                                                                                  'poller_hash': 'NDJDOTAycGVlcmludGZwb24gMTM3NDEx'},
                                                                                                          {       'in_nucast_pkts': '19',
                                                                                                                  'in_octets': '78183118390',
                                                                                                                  'in_ucast_pkts': '219193972',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797190',
                                                                                                                  'out_octets': '752956627701',
                                                                                                                  'out_ucast_pkts': '584531484',
                                                                                                                  'poller_hash': 'NDJDODlBcGVlcmludGZwb24gMTM3NjMw'},
                                                                                                          {       'in_nucast_pkts': '9615226',
                                                                                                                  'in_octets': '184311563745',
                                                                                                                  'in_ucast_pkts': '345628142',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797719',
                                                                                                                  'out_octets': '1157913948289',
                                                                                                                  'out_ucast_pkts': '908512680',
                                                                                                                  'poller_hash': 'NDJCRUUxcGVlcmludGZwb24gMTM3Njgz'},
                                                                                                          {       'in_nucast_pkts': '12',
                                                                                                                  'in_octets': '50913849073',
                                                                                                                  'in_ucast_pkts': '338551163',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797613',
                                                                                                                  'out_octets': '2062282557647',
                                                                                                                  'out_ucast_pkts': '1430624477',
                                                                                                                  'poller_hash': 'NDJDODQ5cGVlcmludGZwb24gMTM3NzA0'},
                                                                                                          {       'in_nucast_pkts': '87443',
                                                                                                                  'in_octets': '280961030175',
                                                                                                                  'in_ucast_pkts': '663177102',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5122305665',
                                                                                                                  'out_octets': '8499157268053',
                                                                                                                  'out_ucast_pkts': '1557292062',
                                                                                                                  'poller_hash': 'NDJDOEU1cGVlcmludGZwb24gMTM3NzYw'},
                                                                                                          {       'in_nucast_pkts': '2409',
                                                                                                                  'in_octets': '119677772601',
                                                                                                                  'in_ucast_pkts': '461641908',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530078',
                                                                                                                  'out_octets': '1718438648332',
                                                                                                                  'out_ucast_pkts': '1278020711',
                                                                                                                  'poller_hash': 'NjQwQUJEcGVlcmludGZwb24gMTM3Nzg3'},
                                                                                                          {       'in_nucast_pkts': '14',
                                                                                                                  'in_octets': '175497202772',
                                                                                                                  'in_ucast_pkts': '974087372',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530171',
                                                                                                                  'out_octets': '2818988480943',
                                                                                                                  'out_ucast_pkts': '2030628296',
                                                                                                                  'poller_hash': 'NjQwQjBBcGVlcmludGZwb24gMTM4MDEw'},
                                                                                                          {       'in_nucast_pkts': '164479',
                                                                                                                  'in_octets': '5796600638205',
                                                                                                                  'in_ucast_pkts': '4890642412',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797590',
                                                                                                                  'out_octets': '6797544661027',
                                                                                                                  'out_ucast_pkts': '6466349730',
                                                                                                                  'poller_hash': 'ODcyNDQzcGVlcmludGZwb24gMTQzMTU3'},
                                                                                                          {       'in_nucast_pkts': '1832',
                                                                                                                  'in_octets': '212165581250',
                                                                                                                  'in_ucast_pkts': '693217410',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '15890294',
                                                                                                                  'out_octets': '3287226739040',
                                                                                                                  'out_ucast_pkts': '2518468212',
                                                                                                                  'poller_hash': 'NzFEODk5cGVlcmludGZwb24gMjQ5ODI3'},
                                                                                                          {       'in_nucast_pkts': '2154865',
                                                                                                                  'in_octets': '16348951335',
                                                                                                                  'in_ucast_pkts': '210940068',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '82833959',
                                                                                                                  'out_octets': '887048323587',
                                                                                                                  'out_ucast_pkts': '607966679',
                                                                                                                  'poller_hash': 'MTE3REFEN3BlZXJpbnRmcG9uIDI1NjIyMg=='},
                                                                                                          {       'in_nucast_pkts': '296146',
                                                                                                                  'in_octets': '24423615755826',
                                                                                                                  'in_ucast_pkts': '29910287731',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9770471',
                                                                                                                  'out_octets': '19311386210823',
                                                                                                                  'out_ucast_pkts': '26191647467',
                                                                                                                  'poller_hash': 'MTI5NkVGNXBlZXJpbnRmcG9uIDI1NjI0Mg=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '7/5',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '347877698554',
                                                                                         'out_octets': '838591430174641',
                                                                                         'out_ucast_pkts': '312952175890',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/5',
                                                                                                                 'id': '7/5',
                                                                                                                 'links': [       {       'connected_time': '21886917',
                                                                                                                                          'fiber_length': 24943,
                                                                                                                                          'poller_hash': 'NDJCOURGcGVlcnBvbnBvbiAzNzc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '3507322',
                                                                                                                                          'fiber_length': 25138,
                                                                                                                                          'poller_hash': 'NDJEMzkxcGVlcnBvbnBvbiAzOTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '7835280',
                                                                                                                                          'fiber_length': 25139,
                                                                                                                                          'poller_hash': 'NDJEMkE2cGVlcnBvbnBvbiA0MTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535140',
                                                                                                                                          'fiber_length': 24813,
                                                                                                                                          'poller_hash': 'NjQwQjc0cGVlcnBvbnBvbiA0MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535293',
                                                                                                                                          'fiber_length': 24773,
                                                                                                                                          'poller_hash': 'NDJDOEMxcGVlcnBvbnBvbiA0MjQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '73263774',
                                                                                                                                          'fiber_length': 24899,
                                                                                                                                          'poller_hash': 'NDJDOTBGcGVlcnBvbnBvbiA0NTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535330',
                                                                                                                                          'fiber_length': 24834,
                                                                                                                                          'poller_hash': 'NDJDMjE4cGVlcnBvbnBvbiA0NTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '1313567',
                                                                                                                                          'fiber_length': 24528,
                                                                                                                                          'poller_hash': 'NDJDREYzcGVlcnBvbnBvbiA0ODE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -29.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535258',
                                                                                                                                          'fiber_length': 24481,
                                                                                                                                          'poller_hash': 'NDJDNkY0cGVlcnBvbnBvbiA0ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535229',
                                                                                                                                          'fiber_length': 25588,
                                                                                                                                          'poller_hash': 'NDJDNkE5cGVlcnBvbnBvbiA0OTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535407',
                                                                                                                                          'fiber_length': 25594,
                                                                                                                                          'poller_hash': 'NDJDNkRGcGVlcnBvbnBvbiA1MDA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '12708337',
                                                                                                                                          'fiber_length': 24765,
                                                                                                                                          'poller_hash': 'NDJDQTA5cGVlcnBvbnBvbiA1MTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '1377493',
                                                                                                                                          'fiber_length': 24807,
                                                                                                                                          'poller_hash': 'NUE2NjVGcGVlcnBvbnBvbiA1MTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '4667581',
                                                                                                                                          'fiber_length': 25274,
                                                                                                                                          'poller_hash': 'NDJDOTNDcGVlcnBvbnBvbiA1MTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8348530',
                                                                                                                                          'fiber_length': 24745,
                                                                                                                                          'poller_hash': 'NjQwQUM2cGVlcnBvbnBvbiA1MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535402',
                                                                                                                                          'fiber_length': 25145,
                                                                                                                                          'poller_hash': 'NUE2NUVEcGVlcnBvbnBvbiA1MTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535063',
                                                                                                                                          'fiber_length': 25302,
                                                                                                                                          'poller_hash': 'NjQwMTBFcGVlcnBvbnBvbiA1MjE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -27.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '21886447',
                                                                                                                                          'fiber_length': 24240,
                                                                                                                                          'poller_hash': 'NDJDOTAycGVlcnBvbnBvbiAxMzc0MTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535271',
                                                                                                                                          'fiber_length': 24674,
                                                                                                                                          'poller_hash': 'NDJDODlBcGVlcnBvbnBvbiAxMzc2MzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535283',
                                                                                                                                          'fiber_length': 25373,
                                                                                                                                          'poller_hash': 'NDJCRUUxcGVlcnBvbnBvbiAxMzc2ODM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535266',
                                                                                                                                          'fiber_length': 25006,
                                                                                                                                          'poller_hash': 'NDJDODQ5cGVlcnBvbnBvbiAxMzc3MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '4377661',
                                                                                                                                          'fiber_length': 25590,
                                                                                                                                          'poller_hash': 'NDJDOEU1cGVlcnBvbnBvbiAxMzc3NjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535111',
                                                                                                                                          'fiber_length': 25051,
                                                                                                                                          'poller_hash': 'NjQwQUJEcGVlcnBvbnBvbiAxMzc3ODc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -28.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535127',
                                                                                                                                          'fiber_length': 24947,
                                                                                                                                          'poller_hash': 'NjQwQjBBcGVlcnBvbnBvbiAxMzgwMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535294',
                                                                                                                                          'fiber_length': 25588,
                                                                                                                                          'poller_hash': 'ODcyNDQzcGVlcnBvbnBvbiAxNDMxNTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '25803693',
                                                                                                                                          'fiber_length': 24692,
                                                                                                                                          'poller_hash': 'NzFEODk5cGVlcnBvbnBvbiAyNDk4Mjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '4933375',
                                                                                                                                          'fiber_length': 24852,
                                                                                                                                          'poller_hash': 'MTE3REFEN3BlZXJwb25wb24gMjU2MjIy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -26.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242},
                                                                                                                                  {       'connected_time': '8535449',
                                                                                                                                          'fiber_length': 25436,
                                                                                                                                          'poller_hash': 'MTI5NkVGNXBlZXJwb25wb24gMjU2MjQy',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.242}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/5',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/6',
                                                                                         'in_discards': '51736',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '5458594',
                                                                                         'in_octets': '7780642549579',
                                                                                         'in_ucast_pkts': '34994016677',
                                                                                         'links': [       {       'in_nucast_pkts': '2802',
                                                                                                                  'in_octets': '134284237971',
                                                                                                                  'in_ucast_pkts': '465137315',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530056',
                                                                                                                  'out_octets': '1436274227490',
                                                                                                                  'out_ucast_pkts': '1086021473',
                                                                                                                  'poller_hash': 'NjQwQUM1cGVlcmludGZwb24gNTE5'},
                                                                                                          {       'in_nucast_pkts': '150227',
                                                                                                                  'in_octets': '56748903794',
                                                                                                                  'in_ucast_pkts': '113862948',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797931',
                                                                                                                  'out_octets': '765751137741',
                                                                                                                  'out_ucast_pkts': '594122465',
                                                                                                                  'poller_hash': 'OTU2RTJGcGVlcmludGZwb24gNTQ2'},
                                                                                                          {       'in_nucast_pkts': '304',
                                                                                                                  'in_octets': '263002077123',
                                                                                                                  'in_ucast_pkts': '1468243754',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797587',
                                                                                                                  'out_octets': '8407183094963',
                                                                                                                  'out_ucast_pkts': '5933143119',
                                                                                                                  'poller_hash': 'NDJDOTA0cGVlcmludGZwb24gMTM3NDAx'},
                                                                                                          {       'in_nucast_pkts': '60',
                                                                                                                  'in_octets': '142370251648',
                                                                                                                  'in_ucast_pkts': '665394214',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5529949',
                                                                                                                  'out_octets': '2626074258277',
                                                                                                                  'out_ucast_pkts': '2019977505',
                                                                                                                  'poller_hash': 'NjQwQURDcGVlcmludGZwb24gMTM3NjE3'},
                                                                                                          {       'in_nucast_pkts': '2638',
                                                                                                                  'in_octets': '10689005622',
                                                                                                                  'in_ucast_pkts': '59123350',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '5530250',
                                                                                                                  'out_octets': '275715878756',
                                                                                                                  'out_ucast_pkts': '208053403',
                                                                                                                  'poller_hash': 'NjQwMTZFcGVlcmludGZwb24gMTM4Mjcw'},
                                                                                                          {       'in_nucast_pkts': '41701814',
                                                                                                                  'in_octets': '1136477356280',
                                                                                                                  'in_ucast_pkts': '4735292556',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '54024616670',
                                                                                                                  'out_octets': '93494576745751',
                                                                                                                  'out_ucast_pkts': '15151155147',
                                                                                                                  'poller_hash': 'ODY5MzUxcGVlcmludGZwb24gMTM5MzI3'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '463563210513',
                                                                                                                  'in_ucast_pkts': '1787644792',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '81',
                                                                                                                  'out_octets': '9152646432426',
                                                                                                                  'out_ucast_pkts': '7077470006',
                                                                                                                  'poller_hash': 'QTlDM0Y0cGVlcmludGZwb24gMTQ1OTkx'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MzVFOTJFcGVlcmludGZwb24gMTUwMzMx'},
                                                                                                          {       'in_nucast_pkts': '153752',
                                                                                                                  'in_octets': '192881093701',
                                                                                                                  'in_ucast_pkts': '438639543',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9795639',
                                                                                                                  'out_octets': '2116401877458',
                                                                                                                  'out_ucast_pkts': '1561024627',
                                                                                                                  'poller_hash': 'QzgxNzY5cGVlcmludGZwb24gMTU3Mzg0'},
                                                                                                          {       'in_nucast_pkts': '160383',
                                                                                                                  'in_octets': '23963681276',
                                                                                                                  'in_ucast_pkts': '116720490',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797984',
                                                                                                                  'out_octets': '658546344775',
                                                                                                                  'out_ucast_pkts': '490824684',
                                                                                                                  'poller_hash': 'QzgxRTU5cGVlcmludGZwb24gMTU3Mzk0'},
                                                                                                          {       'in_nucast_pkts': '151120',
                                                                                                                  'in_octets': '59315248009',
                                                                                                                  'in_ucast_pkts': '373671634',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797738',
                                                                                                                  'out_octets': '1276467665414',
                                                                                                                  'out_ucast_pkts': '979105777',
                                                                                                                  'poller_hash': 'QzgxNzM4cGVlcmludGZwb24gMTU3NDA0'},
                                                                                                          {       'in_nucast_pkts': '158177',
                                                                                                                  'in_octets': '170532885463',
                                                                                                                  'in_ucast_pkts': '816568366',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9797919',
                                                                                                                  'out_octets': '3910508214580',
                                                                                                                  'out_ucast_pkts': '2935288591',
                                                                                                                  'poller_hash': 'QzgxRTBEcGVlcmludGZwb24gMjA3ODIw'},
                                                                                                          {       'in_nucast_pkts': '149413',
                                                                                                                  'in_octets': '117971972671',
                                                                                                                  'in_ucast_pkts': '912857908',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '9795077',
                                                                                                                  'out_octets': '4279702127880',
                                                                                                                  'out_ucast_pkts': '3030736593',
                                                                                                                  'poller_hash': 'MTI3NTE5MnBlZXJpbnRmcG9uIDIyMTY5MA=='}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '7/6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '70700626415',
                                                                                         'out_octets': '255306844915263',
                                                                                         'out_ucast_pkts': '117020460697',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/6',
                                                                                                                 'id': '7/6',
                                                                                                                 'links': [       {       'connected_time': '8535149',
                                                                                                                                          'fiber_length': 23899,
                                                                                                                                          'poller_hash': 'NjQwQUM1cGVlcnBvbnBvbiA1MTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535325',
                                                                                                                                          'fiber_length': 24910,
                                                                                                                                          'poller_hash': 'OTU2RTJGcGVlcnBvbnBvbiA1NDY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535250',
                                                                                                                                          'fiber_length': 24658,
                                                                                                                                          'poller_hash': 'NDJDOTA0cGVlcnBvbnBvbiAxMzc0MDE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535117',
                                                                                                                                          'fiber_length': 24617,
                                                                                                                                          'poller_hash': 'NjQwQURDcGVlcnBvbnBvbiAxMzc2MTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.3,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535115',
                                                                                                                                          'fiber_length': 24802,
                                                                                                                                          'poller_hash': 'NjQwMTZFcGVlcnBvbnBvbiAxMzgyNzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '21887229',
                                                                                                                                          'fiber_length': 24945,
                                                                                                                                          'poller_hash': 'ODY5MzUxcGVlcnBvbnBvbiAxMzkzMjc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '3169492',
                                                                                                                                          'fiber_length': 24142,
                                                                                                                                          'poller_hash': 'QTlDM0Y0cGVlcnBvbnBvbiAxNDU5OTE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '67856432',
                                                                                                                                          'fiber_length': 25097,
                                                                                                                                          'poller_hash': 'MzVFOTJFcGVlcnBvbnBvbiAxNTAzMzE=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535271',
                                                                                                                                          'fiber_length': 24838,
                                                                                                                                          'poller_hash': 'QzgxNzY5cGVlcnBvbnBvbiAxNTczODQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -24.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535273',
                                                                                                                                          'fiber_length': 24982,
                                                                                                                                          'poller_hash': 'QzgxRTU5cGVlcnBvbnBvbiAxNTczOTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535180',
                                                                                                                                          'fiber_length': 24879,
                                                                                                                                          'poller_hash': 'QzgxNzM4cGVlcnBvbnBvbiAxNTc0MDQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -22.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535287',
                                                                                                                                          'fiber_length': 24757,
                                                                                                                                          'poller_hash': 'QzgxRTBEcGVlcnBvbnBvbiAyMDc4MjA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -23.1,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249},
                                                                                                                                  {       'connected_time': '8535219',
                                                                                                                                          'fiber_length': 24804,
                                                                                                                                          'poller_hash': 'MTI3NTE5MnBlZXJwb25wb24gMjIxNjkw',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -25.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 5.249}],
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -32.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/6',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/7',
                                                                                         'in_discards': '32099',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2136960',
                                                                                         'in_octets': '2196657014120',
                                                                                         'in_ucast_pkts': '13315348881',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-03417',
                                                                                         'name': 'gpon '
                                                                                                 '7/7',
                                                                                         'oper_up': False,
                                                                                         'out_nucast_pkts': '24150689629',
                                                                                         'out_octets': '89767399965169',
                                                                                         'out_ucast_pkts': '41167192634',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3Lzc=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/7',
                                                                                                                 'id': '7/7',
                                                                                                                 'max_distance': 60000,
                                                                                                                 'maximum_loss': -12.0,
                                                                                                                 'minimum_loss': -30.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/7',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3Lzc=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '7/8',
                                                                                         'in_discards': '16787',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '3446772',
                                                                                         'in_octets': '252054094539',
                                                                                         'in_ucast_pkts': '875947050',
                                                                                         'links': [       {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MUM4ODhDNnBlZXJpbnRmcG9uIDEy'},
                                                                                                          {       'in_nucast_pkts': '0',
                                                                                                                  'in_octets': '0',
                                                                                                                  'in_ucast_pkts': '0',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '0',
                                                                                                                  'out_ucast_pkts': '0',
                                                                                                                  'poller_hash': 'MTUxRTRCQ3BlZXJpbnRmcG9uIDMw'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01782',
                                                                                         'name': 'gpon '
                                                                                                 '7/8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '8564230502',
                                                                                         'out_octets': '13959831047239',
                                                                                         'out_ucast_pkts': '2704477174',
                                                                                         'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3Lzg=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '7/8',
                                                                                                                 'id': '7/8',
                                                                                                                 'links': [       {       'connected_time': '621884',
                                                                                                                                          'fiber_length': 255,
                                                                                                                                          'poller_hash': 'MUM4ODhDNnBlZXJwb25wb24gMTI=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -15.6,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.608},
                                                                                                                                  {       'connected_time': '761380',
                                                                                                                                          'fiber_length': 255,
                                                                                                                                          'poller_hash': 'MTUxRTRCQ3BlZXJwb25wb24gMzA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -14.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.608}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '7/8',
                                                                                                                 'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3Lzg=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03007',
                                                           'name': 'GPON-8x',
                                                           'poller_hash': 'MjExNTA1NjAwMzI1',
                                                           'serial_number': '211505600325',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43502577'},
                                                   {       'id': 'A',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': 'A/x1',
                                                                                         'in_discards': '1662445',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '70475197',
                                                                                         'in_octets': '7700334132495972',
                                                                                         'in_ucast_pkts': '6009566754282',
                                                                                         'mac_address': 'cc:be:59:47:ac:4d',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'MCCL-SUBS '
                                                                                                 'A/x1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '87676726',
                                                                                         'out_octets': '517002573054180',
                                                                                         'out_ucast_pkts': '1207086511383',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExTUNDTC1TVUJTIEEveDE=',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/x2',
                                                                                         'in_discards': '26187336',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '33901322',
                                                                                         'in_octets': '8426068134458',
                                                                                         'in_ucast_pkts': '41703806043',
                                                                                         'mac_address': 'cc:be:59:47:ac:4e',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01510',
                                                                                         'name': 'ethernet '
                                                                                                 'A/x2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '558173523565',
                                                                                         'out_octets': '922199895910959',
                                                                                         'out_ucast_pkts': '130182746782',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExZXRoZXJuZXQgQS94Mg==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': False,
                                                                                         'id': 'A/x3',
                                                                                         'mac_address': 'cc:be:59:47:ac:4f',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'ethernet '
                                                                                                 'A/x3',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExZXRoZXJuZXQgQS94Mw=='},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/x4',
                                                                                         'in_discards': '1567523',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '44758455',
                                                                                         'in_octets': '6204041709',
                                                                                         'in_ucast_pkts': '0',
                                                                                         'mac_address': 'cc:be:59:47:ac:50',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'Data_LAG '
                                                                                                 'A/x4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '47837741',
                                                                                         'out_octets': '6110184187',
                                                                                         'out_ucast_pkts': '0',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExRGF0YV9MQUcgQS94NA==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': False,
                                                                                         'id': 'A/x5',
                                                                                         'mac_address': 'cc:be:59:47:ac:51',
                                                                                         'name': 'ethernet '
                                                                                                 'A/x5',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExZXRoZXJuZXQgQS94NQ=='},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/x6',
                                                                                         'in_discards': '326800',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '1406873112',
                                                                                         'in_octets': '41699503814039',
                                                                                         'in_ucast_pkts': '29555485898',
                                                                                         'mac_address': 'cc:be:59:47:ac:52',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'ethernet '
                                                                                                 'A/x6',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '37941079',
                                                                                         'out_octets': '1517503798460',
                                                                                         'out_ucast_pkts': '8946411566',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExZXRoZXJuZXQgQS94Ng==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/x7',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '38620707',
                                                                                         'in_octets': '4943450496',
                                                                                         'in_ucast_pkts': '0',
                                                                                         'mac_address': 'cc:be:59:47:ac:53',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'IPTV_2469_LAG '
                                                                                                 'A/x7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '47837739',
                                                                                         'out_octets': '6062346174',
                                                                                         'out_ucast_pkts': '0',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExSVBUVl8yNDY5X0xBRyBBL3g3',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'A/x8',
                                                                                         'in_discards': '24327811',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '286900990',
                                                                                         'in_octets': '1813025789730',
                                                                                         'in_ucast_pkts': '14981074468',
                                                                                         'mac_address': 'cc:be:59:47:ac:54',
                                                                                         'manufacturer': 'Finisar '
                                                                                                         'corp.',
                                                                                         'model': 'FCMJ-8521-3',
                                                                                         'name': 'Voice_LAG '
                                                                                                 'A/x8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '56219189',
                                                                                         'out_octets': '1818056783465',
                                                                                         'out_ucast_pkts': '14849520836',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDExVm9pY2VfTEFHIEEveDg=',
                                                                                         'speed': '1000000000'}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03856',
                                                           'name': 'SCP2-10GE',
                                                           'poller_hash': 'MjExNzA2NjAwMDEx',
                                                           'serial_number': '211706600011',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43502689'},
                                                   {       'id': 'B',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': 'B/x1',
                                                                                         'in_discards': '1574671',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '44956536',
                                                                                         'in_octets': '53057526081',
                                                                                         'in_ucast_pkts': '37183241',
                                                                                         'mac_address': 'cc:be:59:47:ab:ad',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'MCCL-SUBS '
                                                                                                 'B/x1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '77943887',
                                                                                         'out_octets': '485359876880674',
                                                                                         'out_ucast_pkts': '1215517142704',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4TUNDTC1TVUJTIEIveDE=',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/x2',
                                                                                         'in_discards': '24906582',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '57290630',
                                                                                         'in_octets': '9222264420436',
                                                                                         'in_ucast_pkts': '16552142522',
                                                                                         'mac_address': 'cc:be:59:47:ab:ae',
                                                                                         'manufacturer': 'FS',
                                                                                         'model': 'SFPP-PC015',
                                                                                         'name': 'ethernet '
                                                                                                 'B/x2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '467458941504',
                                                                                         'out_octets': '665476884175917',
                                                                                         'out_ucast_pkts': '27671025840',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4ZXRoZXJuZXQgQi94Mg==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': False,
                                                                                         'id': 'B/x3',
                                                                                         'mac_address': 'cc:be:59:47:ab:af',
                                                                                         'name': 'ethernet '
                                                                                                 'B/x3',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4ZXRoZXJuZXQgQi94Mw=='},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/x4',
                                                                                         'in_discards': '963698362',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2063654882',
                                                                                         'in_octets': '187091680238686',
                                                                                         'in_ucast_pkts': '147518046274',
                                                                                         'mac_address': 'cc:be:59:47:ab:b0',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'Data_LAG '
                                                                                                 'B/x4',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '85643463',
                                                                                         'out_octets': '45467551225172',
                                                                                         'out_ucast_pkts': '67338340290',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4RGF0YV9MQUcgQi94NA==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': False,
                                                                                         'id': 'B/x5',
                                                                                         'mac_address': 'cc:be:59:47:ab:b1',
                                                                                         'name': 'ethernet '
                                                                                                 'B/x5',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4ZXRoZXJuZXQgQi94NQ=='},
                                                                                 {       'admin_up': False,
                                                                                         'id': 'B/x6',
                                                                                         'mac_address': 'cc:be:59:47:ab:b2',
                                                                                         'name': 'ethernet '
                                                                                                 'B/x6',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4ZXRoZXJuZXQgQi94Ng=='},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/x7',
                                                                                         'in_discards': '1756286',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '2147115912758',
                                                                                         'in_octets': '2934953662532181',
                                                                                         'in_ucast_pkts': '6086466477',
                                                                                         'mac_address': 'cc:be:59:47:ab:b3',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01514',
                                                                                         'name': 'IPTV_2469_LAG '
                                                                                                 'B/x7',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '85670876',
                                                                                         'out_octets': '950270018703',
                                                                                         'out_ucast_pkts': '5105435080',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4SVBUVl8yNDY5X0xBRyBCL3g3',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': True,
                                                                                         'id': 'B/x8',
                                                                                         'in_discards': '29492718',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '188737118',
                                                                                         'in_octets': '865389388582',
                                                                                         'in_ucast_pkts': '7153802481',
                                                                                         'mac_address': 'cc:be:59:47:ab:b4',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01661',
                                                                                         'name': 'Voice_LAG '
                                                                                                 'B/x8',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '50548485',
                                                                                         'out_octets': '861869159157',
                                                                                         'out_ucast_pkts': '7073403773',
                                                                                         'poller_hash': 'MjExNzA2NjAwMDA4Vm9pY2VfTEFHIEIveDg=',
                                                                                         'speed': '1000000000'}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03856',
                                                           'name': 'SCP2-10GE',
                                                           'poller_hash': 'MjExNzA2NjAwMDA4',
                                                           'serial_number': '211706600008',
                                                           'sw_version': '3.4.51.25',
                                                           'uptime': '43503005'}],
                                'serial_number': '71412913076',
                                'sw_version': '3.4.51.25',
                                'system_mac_address': '00:02:5d:e0:80:75'},
                'peers': [       {       'counter_source': '2/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:42:1a:10:43:f8'],
                                                                       'id': '2/g1',
                                                                       'in_nucast_pkts': '10151',
                                                                       'in_octets': '783390780130',
                                                                       'in_ucast_pkts': '593886644',
                                                                       'mac_address': 'cc:be:59:ab:88:ff',
                                                                       'mtu': 2000,
                                                                       'name': '2/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21816336',
                                                                       'out_octets': '9129386004183',
                                                                       'out_ucast_pkts': '2861820571',
                                                                       'poller_hash': 'NEZBMjVDcGVlcmludGYyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '2',
                                                                       'in_nucast_pkts': '21816336',
                                                                       'in_octets': '9129386004183',
                                                                       'in_ucast_pkts': '2861820571',
                                                                       'links': [       {       'in_nucast_pkts': '21816336',
                                                                                                'in_octets': '9129386004183',
                                                                                                'in_ucast_pkts': '2861820571',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10151',
                                                                                                'out_octets': '783390780130',
                                                                                                'out_ucast_pkts': '593886644',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10151',
                                                                       'out_octets': '783390780130',
                                                                       'out_ucast_pkts': '593886644',
                                                                       'poller_hash': 'NEZBMjVDcGVlcmludGZwb24gMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '2',
                                                                                               'links': [       {       'connected_time': '68952515',
                                                                                                                        'fiber_length': 22900,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.554,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.334}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMjVDcGVlcnBvbnBvbiAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '118 David St Wellesley (Daryl Kropf '
                                                 '10434350)',
                                         'poller_hash': 'NEZBMjVD',
                                         'serial_number': '4FA25C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:88:fe',
                                         'uptime': '68952515'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '3/g1',
                                                                       'name': '3/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2Rjc5cGVlcmludGYzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '3',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2Rjc5cGVlcmludGZwb24gMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '3',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2Rjc5cGVlcnBvbnBvbiAz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'MCCL CO Test (812G)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "3" (location: '
                                                                  'GPON port "7/8"), '
                                                                  '"ONT Dying Gasp"',
                                         'offline_time': '2024-10-10T12:26:27.040Z',
                                         'poller_hash': 'OTU2Rjc5',
                                         'serial_number': '956F79',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '4/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:f0:51'],
                                                                       'id': '4/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '3583001168',
                                                                       'in_ucast_pkts': '22861089',
                                                                       'mac_address': 'ec:4f:82:5b:70:ce',
                                                                       'mtu': 2000,
                                                                       'name': '4/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394167',
                                                                       'out_octets': '87165772110',
                                                                       'out_ucast_pkts': '67620287',
                                                                       'poller_hash': 'MkIwMThDcGVlcmludGY0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '4',
                                                                       'in_nucast_pkts': '394167',
                                                                       'in_octets': '87165772110',
                                                                       'in_ucast_pkts': '67620287',
                                                                       'links': [       {       'in_nucast_pkts': '394167',
                                                                                                'in_octets': '87165772110',
                                                                                                'in_ucast_pkts': '67620287',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '3583001168',
                                                                                                'out_ucast_pkts': '22861089',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '3583001168',
                                                                       'out_ucast_pkts': '22861089',
                                                                       'poller_hash': 'MkIwMThDcGVlcmludGZwb24gNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '4',
                                                                                               'links': [       {       'connected_time': '335336',
                                                                                                                        'fiber_length': 16964,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMThDcGVlcnBvbnBvbiA0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5577 King St. Nithburg (John '
                                                 'Vleeming 10433729)',
                                         'poller_hash': 'MkIwMThD',
                                         'serial_number': '2B018C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:cb',
                                         'uptime': '335336'},
                                 {       'counter_source': '5/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:f0:1b'],
                                                                       'id': '5/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '2811906322',
                                                                       'in_ucast_pkts': '26603221',
                                                                       'mac_address': 'ec:4f:82:5b:70:a6',
                                                                       'mtu': 2000,
                                                                       'name': '5/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394166',
                                                                       'out_octets': '71032092544',
                                                                       'out_ucast_pkts': '60848582',
                                                                       'poller_hash': 'MkIwMTg3cGVlcmludGY1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '5',
                                                                       'in_nucast_pkts': '394166',
                                                                       'in_octets': '71032092544',
                                                                       'in_ucast_pkts': '60848582',
                                                                       'links': [       {       'in_nucast_pkts': '394166',
                                                                                                'in_octets': '71032092544',
                                                                                                'in_ucast_pkts': '60848582',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '2811906322',
                                                                                                'out_ucast_pkts': '26603221',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '2811906322',
                                                                       'out_ucast_pkts': '26603221',
                                                                       'poller_hash': 'MkIwMTg3cGVlcmludGZwb24gNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '5',
                                                                                               'links': [       {       'connected_time': '335332',
                                                                                                                        'fiber_length': 17181,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMTg3cGVlcnBvbnBvbiA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5554 King St. Nithburg (Raymond '
                                                 'Robinson 10433731)',
                                         'poller_hash': 'MkIwMTg3',
                                         'serial_number': '2B0187',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:70:a3',
                                         'uptime': '335332'},
                                 {       'counter_source': '6/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1f:ab'],
                                                                       'id': '6/g1',
                                                                       'in_nucast_pkts': '11033',
                                                                       'in_octets': '10879847126',
                                                                       'in_ucast_pkts': '31205796',
                                                                       'mac_address': 'ec:4f:82:5b:88:26',
                                                                       'mtu': 2000,
                                                                       'name': '6/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '969328',
                                                                       'out_octets': '124066402700',
                                                                       'out_ucast_pkts': '93124808',
                                                                       'poller_hash': 'MkIwNDc3cGVlcmludGY2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '6',
                                                                       'in_nucast_pkts': '969328',
                                                                       'in_octets': '124066402700',
                                                                       'in_ucast_pkts': '93124808',
                                                                       'links': [       {       'in_nucast_pkts': '969328',
                                                                                                'in_octets': '124066402700',
                                                                                                'in_ucast_pkts': '93124808',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11033',
                                                                                                'out_octets': '10879847126',
                                                                                                'out_ucast_pkts': '31205796',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11033',
                                                                       'out_octets': '10879847126',
                                                                       'out_ucast_pkts': '31205796',
                                                                       'poller_hash': 'MkIwNDc3cGVlcmludGZwb24gNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '6',
                                                                                               'links': [       {       'connected_time': '598181',
                                                                                                                        'fiber_length': 17009,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwNDc3cGVlcnBvbnBvbiA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5587 King St. Nithburg (DREHER KEVIN '
                                                 '10433735)',
                                         'poller_hash': 'MkIwNDc3',
                                         'serial_number': '2B0477',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:88:23',
                                         'uptime': '598181'},
                                 {       'counter_source': '7/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fe:06'],
                                                                       'id': '7/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '6300264069',
                                                                       'in_ucast_pkts': '25585140',
                                                                       'mac_address': 'ec:4f:82:7d:c8:c9',
                                                                       'mtu': 2000,
                                                                       'name': '7/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394159',
                                                                       'out_octets': '137967810460',
                                                                       'out_ucast_pkts': '98987474',
                                                                       'poller_hash': 'MzAwRDkxcGVlcmludGY3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '7',
                                                                       'in_nucast_pkts': '394159',
                                                                       'in_octets': '137967810460',
                                                                       'in_ucast_pkts': '98987474',
                                                                       'links': [       {       'in_nucast_pkts': '394159',
                                                                                                'in_octets': '137967810460',
                                                                                                'in_ucast_pkts': '98987474',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '6300264069',
                                                                                                'out_ucast_pkts': '25585140',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '6300264069',
                                                                       'out_ucast_pkts': '25585140',
                                                                       'poller_hash': 'MzAwRDkxcGVlcmludGZwb24gNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '7',
                                                                                               'links': [       {       'connected_time': '335332',
                                                                                                                        'fiber_length': 17157,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRDkxcGVlcnBvbnBvbiA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5576 KING ST (WILLIAMS DAVE)',
                                         'poller_hash': 'MzAwRDkx',
                                         'serial_number': '300D91',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c8:c9',
                                         'uptime': '335332'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '8/g1',
                                                                       'name': '8/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTRGNEM4OHBlZXJpbnRmOC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '8',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTRGNEM4OHBlZXJpbnRmcG9uIDg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '8',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTRGNEM4OHBlZXJwb25wb24gOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'U6x Test',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "8" (location: '
                                                                  'GPON port "7/8"), '
                                                                  '"ONT went missing"',
                                         'offline_time': '2025-07-30T11:31:35.980Z',
                                         'poller_hash': 'MTRGNEM4OA==',
                                         'serial_number': '14F4C88',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '9/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '9/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '101019491987',
                                                                       'in_ucast_pkts': '326141022',
                                                                       'mac_address': 'ec:4f:82:59:ed:73',
                                                                       'name': '9/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '197',
                                                                       'out_octets': '1727253755109',
                                                                       'out_ucast_pkts': '1272936274',
                                                                       'poller_hash': 'MkFEMTIxcGVlcmludGY5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '9',
                                                                       'in_nucast_pkts': '197',
                                                                       'in_octets': '1727253755109',
                                                                       'in_ucast_pkts': '1272936274',
                                                                       'links': [       {       'in_nucast_pkts': '197',
                                                                                                'in_octets': '1727253755109',
                                                                                                'in_ucast_pkts': '1272936274',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '101019491987',
                                                                                                'out_ucast_pkts': '326141022',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '101019491987',
                                                                       'out_ucast_pkts': '326141022',
                                                                       'poller_hash': 'MkFEMTIxcGVlcmludGZwb24gOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '9',
                                                                                               'links': [       {       'connected_time': '7443401',
                                                                                                                        'fiber_length': 17193,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFEMTIxcGVlcnBvbnBvbiA5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5535 KING ST (WILSON CAROL D)',
                                         'poller_hash': 'MkFEMTIx',
                                         'serial_number': '2AD121',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:ed:73',
                                         'uptime': '7443401'},
                                 {       'counter_source': '10/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '28:87:ba:54:92:61'],
                                                                       'id': '10/g1',
                                                                       'in_nucast_pkts': '131527',
                                                                       'in_octets': '1756122670',
                                                                       'in_ucast_pkts': '7927715',
                                                                       'mac_address': 'cc:be:59:f8:64:85',
                                                                       'mtu': 2000,
                                                                       'name': '10/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '438091',
                                                                       'out_octets': '70434327589',
                                                                       'out_ucast_pkts': '49701672',
                                                                       'poller_hash': 'NTc0NTgzcGVlcmludGYxMC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '10',
                                                                       'in_nucast_pkts': '438091',
                                                                       'in_octets': '70434327589',
                                                                       'in_ucast_pkts': '49701672',
                                                                       'links': [       {       'in_nucast_pkts': '438091',
                                                                                                'in_octets': '70434327589',
                                                                                                'in_ucast_pkts': '49701672',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '131527',
                                                                                                'out_octets': '1756122670',
                                                                                                'out_ucast_pkts': '7927715',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '131527',
                                                                       'out_octets': '1756122670',
                                                                       'out_ucast_pkts': '7927715',
                                                                       'poller_hash': 'NTc0NTgzcGVlcmludGZwb24gMTA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '10',
                                                                                               'links': [       {       'connected_time': '335338',
                                                                                                                        'fiber_length': 17405,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.596,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.428}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0NTgzcGVlcnBvbnBvbiAxMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': '5498 King St. Nithburg (Linda&Ben '
                                                 'Veiledal 10433920)',
                                         'poller_hash': 'NTc0NTgz',
                                         'serial_number': '574583',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:64:7f',
                                         'uptime': '335338'},
                                 {       'counter_source': '11/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '11/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '128264019518',
                                                                       'in_ucast_pkts': '268585539',
                                                                       'mac_address': 'ec:4f:82:59:ec:53',
                                                                       'name': '11/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '144',
                                                                       'out_octets': '1175404602398',
                                                                       'out_ucast_pkts': '963109639',
                                                                       'poller_hash': 'MkFEMEZEcGVlcmludGYxMS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '11',
                                                                       'in_nucast_pkts': '144',
                                                                       'in_octets': '1175404602398',
                                                                       'in_ucast_pkts': '963109639',
                                                                       'links': [       {       'in_nucast_pkts': '144',
                                                                                                'in_octets': '1175404602398',
                                                                                                'in_ucast_pkts': '963109639',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '128264019518',
                                                                                                'out_ucast_pkts': '268585539',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '128264019518',
                                                                       'out_ucast_pkts': '268585539',
                                                                       'poller_hash': 'MkFEMEZEcGVlcmludGZwb24gMTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '11',
                                                                                               'links': [       {       'connected_time': '73026593',
                                                                                                                        'fiber_length': 17336,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFEMEZEcGVlcnBvbnBvbiAxMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5514 KING ST (COLLINS GARY)',
                                         'poller_hash': 'MkFEMEZE',
                                         'serial_number': '2AD0FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:ec:53',
                                         'uptime': '73026593'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '12/g1',
                                                                       'name': '12/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUM4ODhDNnBlZXJpbnRmMTIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '12',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MUM4ODhDNnBlZXJpbnRmcG9uIDEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '12',
                                                                                               'links': [       {       'connected_time': '621884',
                                                                                                                        'fiber_length': 255,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.674,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.71}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUM4ODhDNnBlZXJwb25wb24gMTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '4200 (Test)',
                                         'poller_hash': 'MUM4ODhDNg==',
                                         'serial_number': '1C888C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '25.1.500.244',
                                         'system_mac_address': '10:74:c5:fe:c0:54',
                                         'uptime': '621884'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '13/g1',
                                                                       'name': '13/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MjA1REJBcGVlcmludGYxMy9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '13',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MjA1REJBcGVlcmludGZwb24gMTM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '13',
                                                                                               'links': [       {       'connected_time': '4650410',
                                                                                                                        'fiber_length': 17572,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REJBcGVlcnBvbnBvbiAxMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '5484 KING ST (HOLMES MICHAEL)',
                                         'poller_hash': 'MjA1REJB',
                                         'serial_number': '205DBA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:68:5f',
                                         'uptime': '4650410'},
                                 {       'counter_source': '14/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e4:6c'],
                                                                       'id': '14/g1',
                                                                       'in_nucast_pkts': '6728',
                                                                       'in_octets': '2495422223',
                                                                       'in_ucast_pkts': '15824239',
                                                                       'mac_address': 'ec:4f:82:7d:bc:a7',
                                                                       'mtu': 2000,
                                                                       'name': '14/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394140',
                                                                       'out_octets': '97163421130',
                                                                       'out_ucast_pkts': '69173496',
                                                                       'poller_hash': 'MzAwN0JDcGVlcmludGYxNC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '14',
                                                                       'in_nucast_pkts': '394140',
                                                                       'in_octets': '97163421130',
                                                                       'in_ucast_pkts': '69173496',
                                                                       'links': [       {       'in_nucast_pkts': '394140',
                                                                                                'in_octets': '97163421130',
                                                                                                'in_ucast_pkts': '69173496',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6728',
                                                                                                'out_octets': '2495422223',
                                                                                                'out_ucast_pkts': '15824239',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6728',
                                                                       'out_octets': '2495422223',
                                                                       'out_ucast_pkts': '15824239',
                                                                       'poller_hash': 'MzAwN0JDcGVlcmludGZwb24gMTQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '14',
                                                                                               'links': [       {       'connected_time': '335329',
                                                                                                                        'fiber_length': 17202,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwN0JDcGVlcnBvbnBvbiAxNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '2091 Nithview Ct. Nithburg (Jeff '
                                                 'Foster)',
                                         'poller_hash': 'MzAwN0JD',
                                         'serial_number': '3007BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:bc:a7',
                                         'uptime': '335329'},
                                 {       'counter_source': '15/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:f8:b3:64:c8:d4'],
                                                                       'id': '15/g1',
                                                                       'in_nucast_pkts': '80300',
                                                                       'in_octets': '90136103397',
                                                                       'in_ucast_pkts': '595206266',
                                                                       'mac_address': 'ec:4f:82:7e:3a:36',
                                                                       'mtu': 2000,
                                                                       'name': '15/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22546206',
                                                                       'out_octets': '1964802213912',
                                                                       'out_ucast_pkts': '1515954631',
                                                                       'poller_hash': 'MzVGOUE2cGVlcmludGYxNS9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '15',
                                                                       'in_nucast_pkts': '22546206',
                                                                       'in_octets': '1964802213912',
                                                                       'in_ucast_pkts': '1515954631',
                                                                       'links': [       {       'in_nucast_pkts': '22546206',
                                                                                                'in_octets': '1964802213912',
                                                                                                'in_ucast_pkts': '1515954631',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '80300',
                                                                                                'out_octets': '90136103397',
                                                                                                'out_ucast_pkts': '595206266',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '80300',
                                                                       'out_octets': '90136103397',
                                                                       'out_ucast_pkts': '595206266',
                                                                       'poller_hash': 'MzVGOUE2cGVlcmludGZwb24gMTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '15',
                                                                                               'links': [       {       'connected_time': '19526429',
                                                                                                                        'fiber_length': 17737,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUE2cGVlcnBvbnBvbiAxNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5466 Brown St. Nithburg (Nancy '
                                                 'Firth)',
                                         'poller_hash': 'MzVGOUE2',
                                         'serial_number': '35F9A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3a:36',
                                         'uptime': '19526429'},
                                 {       'counter_source': '16/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a8:ed'],
                                                                       'id': '16/g1',
                                                                       'in_nucast_pkts': '8317',
                                                                       'in_octets': '153565240197',
                                                                       'in_ucast_pkts': '460235051',
                                                                       'mac_address': 'ec:4f:82:8d:55:d3',
                                                                       'mtu': 2000,
                                                                       'name': '16/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58877786',
                                                                       'out_octets': '1430454393482',
                                                                       'out_ucast_pkts': '1138812801',
                                                                       'poller_hash': 'MzIxMzAxcGVlcmludGYxNi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '16',
                                                                       'in_nucast_pkts': '58877786',
                                                                       'in_octets': '1430454393482',
                                                                       'in_ucast_pkts': '1138812801',
                                                                       'links': [       {       'in_nucast_pkts': '58877786',
                                                                                                'in_octets': '1430454393482',
                                                                                                'in_ucast_pkts': '1138812801',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8317',
                                                                                                'out_octets': '153565240197',
                                                                                                'out_ucast_pkts': '460235051',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8317',
                                                                       'out_octets': '153565240197',
                                                                       'out_ucast_pkts': '460235051',
                                                                       'poller_hash': 'MzIxMzAxcGVlcmludGZwb24gMTY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '16',
                                                                                               'links': [       {       'connected_time': '21886489',
                                                                                                                        'fiber_length': 17533,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMzAxcGVlcnBvbnBvbiAxNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2074 Albert St. Nithburg (Jim '
                                                 'Brubacher)',
                                         'poller_hash': 'MzIxMzAx',
                                         'serial_number': '321301',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:55:d0',
                                         'uptime': '21886489'},
                                 {       'counter_source': '17/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a7:a9'],
                                                                       'id': '17/g1',
                                                                       'in_nucast_pkts': '113',
                                                                       'in_octets': '120074434937',
                                                                       'in_ucast_pkts': '861214614',
                                                                       'mac_address': 'ec:4f:82:8d:45:a3',
                                                                       'mtu': 2000,
                                                                       'name': '17/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16616138',
                                                                       'out_octets': '4289592711639',
                                                                       'out_ucast_pkts': '3187698650',
                                                                       'poller_hash': 'MzIxMEZCcGVlcmludGYxNy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '17',
                                                                       'in_nucast_pkts': '16616138',
                                                                       'in_octets': '4289592711639',
                                                                       'in_ucast_pkts': '3187698650',
                                                                       'links': [       {       'in_nucast_pkts': '16616138',
                                                                                                'in_octets': '4289592711639',
                                                                                                'in_ucast_pkts': '3187698650',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '113',
                                                                                                'out_octets': '120074434937',
                                                                                                'out_ucast_pkts': '861214614',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '113',
                                                                       'out_octets': '120074434937',
                                                                       'out_ucast_pkts': '861214614',
                                                                       'poller_hash': 'MzIxMEZCcGVlcmludGZwb24gMTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '17',
                                                                                               'links': [       {       'connected_time': '14334953',
                                                                                                                        'fiber_length': 17754,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMEZCcGVlcnBvbnBvbiAxNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5469 BROWN ST (MCHUGH ROBERT)',
                                         'poller_hash': 'MzIxMEZC',
                                         'serial_number': '3210FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:45:a0',
                                         'uptime': '14334953'},
                                 {       'counter_source': '18/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '68:d7:9a:5c:bd:4f'],
                                                                       'id': '18/g1',
                                                                       'in_nucast_pkts': '46',
                                                                       'in_octets': '38407667396',
                                                                       'in_ucast_pkts': '66683387',
                                                                       'mac_address': 'ec:4f:82:ae:bb:93',
                                                                       'mtu': 2000,
                                                                       'name': '18/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '916121',
                                                                       'out_octets': '115151047008',
                                                                       'out_ucast_pkts': '101912491',
                                                                       'poller_hash': 'M0I2MTI4cGVlcmludGYxOC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '18',
                                                                       'in_nucast_pkts': '916121',
                                                                       'in_octets': '115151047008',
                                                                       'in_ucast_pkts': '101912491',
                                                                       'links': [       {       'in_nucast_pkts': '916121',
                                                                                                'in_octets': '115151047008',
                                                                                                'in_ucast_pkts': '101912491',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '46',
                                                                                                'out_octets': '38407667396',
                                                                                                'out_ucast_pkts': '66683387',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '46',
                                                                       'out_octets': '38407667396',
                                                                       'out_ucast_pkts': '66683387',
                                                                       'poller_hash': 'M0I2MTI4cGVlcmludGZwb24gMTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '18',
                                                                                               'links': [       {       'connected_time': '782105',
                                                                                                                        'fiber_length': 17136,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0I2MTI4cGVlcnBvbnBvbiAxOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5555 King St. Nithburg (Phillip '
                                                 'Garrett)',
                                         'poller_hash': 'M0I2MTI4',
                                         'serial_number': '3B6128',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:ae:bb:93',
                                         'uptime': '782105'},
                                 {       'counter_source': '19/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:9d:e8'],
                                                                       'id': '19/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '5428252594',
                                                                       'in_ucast_pkts': '40304691',
                                                                       'mac_address': 'ec:4f:82:8d:4d:a3',
                                                                       'mtu': 2000,
                                                                       'name': '19/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394160',
                                                                       'out_octets': '140668620094',
                                                                       'out_ucast_pkts': '101032236',
                                                                       'poller_hash': 'MzIxMUZCcGVlcmludGYxOS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '19',
                                                                       'in_nucast_pkts': '394160',
                                                                       'in_octets': '140668620094',
                                                                       'in_ucast_pkts': '101032236',
                                                                       'links': [       {       'in_nucast_pkts': '394160',
                                                                                                'in_octets': '140668620094',
                                                                                                'in_ucast_pkts': '101032236',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '5428252594',
                                                                                                'out_ucast_pkts': '40304691',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '5428252594',
                                                                       'out_ucast_pkts': '40304691',
                                                                       'poller_hash': 'MzIxMUZCcGVlcmludGZwb24gMTk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '19',
                                                                                               'links': [       {       'connected_time': '335332',
                                                                                                                        'fiber_length': 17678,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMUZCcGVlcnBvbnBvbiAxOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5483 BROWN ST (SHUSHACK PAUL)',
                                         'poller_hash': 'MzIxMUZC',
                                         'serial_number': '3211FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:4d:a0',
                                         'uptime': '335332'},
                                 {       'counter_source': '20/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '20/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '83769762183',
                                                                       'in_ucast_pkts': '548823285',
                                                                       'mac_address': 'ec:4f:82:8d:44:98',
                                                                       'name': '20/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '144',
                                                                       'out_octets': '2300852304494',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzIxMERBcGVlcmludGYyMC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '20',
                                                                       'in_nucast_pkts': '144',
                                                                       'in_octets': '2300852304494',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '144',
                                                                                                'in_octets': '2300852304494',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '83769762183',
                                                                                                'out_ucast_pkts': '548823285',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '83769762183',
                                                                       'out_ucast_pkts': '548823285',
                                                                       'poller_hash': 'MzIxMERBcGVlcmludGZwb24gMjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '20',
                                                                                               'links': [       {       'connected_time': '19961094',
                                                                                                                        'fiber_length': 17714,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMERBcGVlcnBvbnBvbiAyMA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5470 Brown St. Nithburg (John '
                                                 'Tabbert 10412107)',
                                         'poller_hash': 'MzIxMERB',
                                         'serial_number': '3210DA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:44:98',
                                         'uptime': '19961094'},
                                 {       'counter_source': '21/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:98:e8:33:66:2b'],
                                                                       'id': '21/g1',
                                                                       'in_nucast_pkts': '3163',
                                                                       'in_octets': '5046143765',
                                                                       'in_ucast_pkts': '25659186',
                                                                       'mac_address': 'ec:4f:82:7d:ca:b3',
                                                                       'mtu': 2000,
                                                                       'name': '21/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394850',
                                                                       'out_octets': '85396601164',
                                                                       'out_ucast_pkts': '66346822',
                                                                       'poller_hash': 'MzAwRTg2cGVlcmludGYyMS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '21',
                                                                       'in_nucast_pkts': '394850',
                                                                       'in_octets': '85396601164',
                                                                       'in_ucast_pkts': '66346822',
                                                                       'links': [       {       'in_nucast_pkts': '394850',
                                                                                                'in_octets': '85396601164',
                                                                                                'in_ucast_pkts': '66346822',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3163',
                                                                                                'out_octets': '5046143765',
                                                                                                'out_ucast_pkts': '25659186',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3163',
                                                                       'out_octets': '5046143765',
                                                                       'out_ucast_pkts': '25659186',
                                                                       'poller_hash': 'MzAwRTg2cGVlcmludGZwb24gMjE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '21',
                                                                                               'links': [       {       'connected_time': '335330',
                                                                                                                        'fiber_length': 17920,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRTg2cGVlcnBvbnBvbiAyMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5407 King St. Nithburg (Chantal '
                                                 'Valliere 10434016)',
                                         'poller_hash': 'MzAwRTg2',
                                         'serial_number': '300E86',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:ca:b3',
                                         'uptime': '335330'},
                                 {       'counter_source': '22/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:f1:b0'],
                                                                       'id': '22/g1',
                                                                       'in_nucast_pkts': '12864',
                                                                       'in_octets': '9550878436',
                                                                       'in_ucast_pkts': '39147914',
                                                                       'mac_address': 'ec:4f:82:8d:43:db',
                                                                       'mtu': 2000,
                                                                       'name': '22/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1225367',
                                                                       'out_octets': '295572088761',
                                                                       'out_ucast_pkts': '211569415',
                                                                       'poller_hash': 'MzIxMEMycGVlcmludGYyMi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '22',
                                                                       'in_nucast_pkts': '1225367',
                                                                       'in_octets': '295572088761',
                                                                       'in_ucast_pkts': '211569415',
                                                                       'links': [       {       'in_nucast_pkts': '1225367',
                                                                                                'in_octets': '295572088761',
                                                                                                'in_ucast_pkts': '211569415',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12864',
                                                                                                'out_octets': '9550878436',
                                                                                                'out_ucast_pkts': '39147914',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12864',
                                                                       'out_octets': '9550878436',
                                                                       'out_ucast_pkts': '39147914',
                                                                       'poller_hash': 'MzIxMEMycGVlcmludGZwb24gMjI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '22',
                                                                                               'links': [       {       'connected_time': '1050060',
                                                                                                                        'fiber_length': 17129,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMEMycGVlcnBvbnBvbiAyMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2087 Nithview Ct. Nithburg (Jeannine '
                                                 'Campbell 10434015)',
                                         'poller_hash': 'MzIxMEMy',
                                         'serial_number': '3210C2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:43:d8',
                                         'uptime': '1050060'},
                                 {       'counter_source': '23/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2f:31:80'],
                                                                       'id': '23/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '2816633702',
                                                                       'in_ucast_pkts': '16937606',
                                                                       'mac_address': 'cc:be:59:cb:ba:b5',
                                                                       'mtu': 2000,
                                                                       'name': '23/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394171',
                                                                       'out_octets': '63756685341',
                                                                       'out_ucast_pkts': '46407077',
                                                                       'poller_hash': 'NDJDOEFFcGVlcmludGYyMy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '23',
                                                                       'in_nucast_pkts': '394171',
                                                                       'in_octets': '63756685341',
                                                                       'in_ucast_pkts': '46407077',
                                                                       'links': [       {       'in_nucast_pkts': '394171',
                                                                                                'in_octets': '63756685341',
                                                                                                'in_ucast_pkts': '46407077',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '2816633702',
                                                                                                'out_ucast_pkts': '16937606',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '2816633702',
                                                                       'out_ucast_pkts': '16937606',
                                                                       'poller_hash': 'NDJDOEFFcGVlcmludGZwb24gMjM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '23',
                                                                                               'links': [       {       'connected_time': '335330',
                                                                                                                        'fiber_length': 17518,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOEFFcGVlcnBvbnBvbiAyMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5506 King St. Nithburg (Jennifer '
                                                 'Birmingham)',
                                         'poller_hash': 'NDJDOEFF',
                                         'serial_number': '42C8AE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ba:b2',
                                         'uptime': '335330'},
                                 {       'counter_source': '25/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b0:39:56:6b:cf:43'],
                                                                       'id': '25/g1',
                                                                       'in_nucast_pkts': '226',
                                                                       'in_octets': '2572587948',
                                                                       'in_ucast_pkts': '7336541',
                                                                       'mac_address': 'cc:be:59:50:1b:88',
                                                                       'mtu': 2000,
                                                                       'name': '25/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '161538',
                                                                       'out_octets': '31721317878',
                                                                       'out_ucast_pkts': '25214153',
                                                                       'poller_hash': 'NDNDNDQ1cGVlcmludGYyNS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '25',
                                                                       'in_nucast_pkts': '161538',
                                                                       'in_octets': '31721317878',
                                                                       'in_ucast_pkts': '25214153',
                                                                       'links': [       {       'in_nucast_pkts': '161538',
                                                                                                'in_octets': '31721317878',
                                                                                                'in_ucast_pkts': '25214153',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '226',
                                                                                                'out_octets': '2572587948',
                                                                                                'out_ucast_pkts': '7336541',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '226',
                                                                       'out_octets': '2572587948',
                                                                       'out_ucast_pkts': '7336541',
                                                                       'poller_hash': 'NDNDNDQ1cGVlcmludGZwb24gMjU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '25',
                                                                                               'links': [       {       'connected_time': '335329',
                                                                                                                        'fiber_length': 17275,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.728,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.312}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNDNDQ1cGVlcnBvbnBvbiAyNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5525 King St. Nithburg (Merlene '
                                                 'Austen)',
                                         'poller_hash': 'NDNDNDQ1',
                                         'serial_number': '43C445',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:1b:87',
                                         'uptime': '335329'},
                                 {       'counter_source': '26/F1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:70:70'],
                                                                       'id': '26/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2462364261',
                                                                       'in_ucast_pkts': '14831121',
                                                                       'name': '26/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '17784466034',
                                                                       'out_octets': '23594900507310',
                                                                       'out_ucast_pkts': '15235034',
                                                                       'poller_hash': 'MzIxNzA2cGVlcmludGYyNi9GMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '26',
                                                                       'in_nucast_pkts': '17784466034',
                                                                       'in_octets': '23594900507310',
                                                                       'in_ucast_pkts': '15235034',
                                                                       'links': [       {       'in_nucast_pkts': '17784466034',
                                                                                                'in_octets': '23594900507310',
                                                                                                'in_ucast_pkts': '15235034',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2462364261',
                                                                                                'out_ucast_pkts': '14831121',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2462364261',
                                                                       'out_ucast_pkts': '14831121',
                                                                       'poller_hash': 'MzIxNzA2cGVlcmludGZwb24gMjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '26',
                                                                                               'links': [       {       'connected_time': '21886983',
                                                                                                                        'fiber_length': 17689,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxNzA2cGVlcnBvbnBvbiAyNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5460 King St. Nithburg (Robert '
                                                 'Litke)',
                                         'poller_hash': 'MzIxNzA2',
                                         'serial_number': '321706',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:75:f8',
                                         'uptime': '21886983'},
                                 {       'counter_source': '27/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '90:ec:77:8e:7e:59'],
                                                                       'id': '27/g1',
                                                                       'in_nucast_pkts': '1848',
                                                                       'in_octets': '626862779058',
                                                                       'in_ucast_pkts': '2215147263',
                                                                       'mac_address': 'ec:4f:82:7e:08:a4',
                                                                       'mtu': 2000,
                                                                       'name': '27/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1603704',
                                                                       'out_octets': '2624531316107',
                                                                       'out_ucast_pkts': '4409498187',
                                                                       'poller_hash': 'MzVFOTJDcGVlcmludGYyNy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '27',
                                                                       'in_nucast_pkts': '1603704',
                                                                       'in_octets': '2624531316107',
                                                                       'in_ucast_pkts': '4409498187',
                                                                       'links': [       {       'in_nucast_pkts': '1603704',
                                                                                                'in_octets': '2624531316107',
                                                                                                'in_ucast_pkts': '4409498187',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1848',
                                                                                                'out_octets': '626862779058',
                                                                                                'out_ucast_pkts': '2215147263',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1848',
                                                                       'out_octets': '626862779058',
                                                                       'out_ucast_pkts': '2215147263',
                                                                       'poller_hash': 'MzVFOTJDcGVlcmludGZwb24gMjc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '27',
                                                                                               'links': [       {       'connected_time': '1371569',
                                                                                                                        'fiber_length': 22188,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTJDcGVlcnBvbnBvbiAyNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1430 Hutchison Rd Nithburg (T & T '
                                                 'Power Group 10431156)',
                                         'poller_hash': 'MzVFOTJD',
                                         'serial_number': '35E92C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:08:a4',
                                         'uptime': '1371569'},
                                 {       'counter_source': '28/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '28/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '121809989253',
                                                                       'in_ucast_pkts': '964999398',
                                                                       'mac_address': 'ec:4f:82:8d:43:e8',
                                                                       'name': '28/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '144',
                                                                       'out_octets': '3736476853982',
                                                                       'out_ucast_pkts': '2690125197',
                                                                       'poller_hash': 'MzIxMEM0cGVlcmludGYyOC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '28',
                                                                       'in_nucast_pkts': '144',
                                                                       'in_octets': '3736476853982',
                                                                       'in_ucast_pkts': '2690125197',
                                                                       'links': [       {       'in_nucast_pkts': '144',
                                                                                                'in_octets': '3736476853982',
                                                                                                'in_ucast_pkts': '2690125197',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '121809989253',
                                                                                                'out_ucast_pkts': '964999398',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '121809989253',
                                                                       'out_ucast_pkts': '964999398',
                                                                       'poller_hash': 'MzIxMEM0cGVlcmludGZwb24gMjg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '28',
                                                                                               'links': [       {       'connected_time': '8535127',
                                                                                                                        'fiber_length': 17161,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMEM0cGVlcnBvbnBvbiAyOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2050 Perth Line 56 Nithburg (Ken '
                                                 'Snider)',
                                         'poller_hash': 'MzIxMEM0',
                                         'serial_number': '3210C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:43:e8',
                                         'uptime': '8535127'},
                                 {       'counter_source': '29/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:ce:81'],
                                                                       'id': '29/g1',
                                                                       'in_nucast_pkts': '7624',
                                                                       'in_octets': '1043579514',
                                                                       'in_ucast_pkts': '4179956',
                                                                       'mac_address': 'ec:4f:82:7d:c8:a9',
                                                                       'mtu': 2000,
                                                                       'name': '29/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394147',
                                                                       'out_octets': '22456633619',
                                                                       'out_ucast_pkts': '18803708',
                                                                       'poller_hash': 'MzAwRDgxcGVlcmludGYyOS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '29',
                                                                       'in_nucast_pkts': '394147',
                                                                       'in_octets': '22456633619',
                                                                       'in_ucast_pkts': '18803708',
                                                                       'links': [       {       'in_nucast_pkts': '394147',
                                                                                                'in_octets': '22456633619',
                                                                                                'in_ucast_pkts': '18803708',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7624',
                                                                                                'out_octets': '1043579514',
                                                                                                'out_ucast_pkts': '4179956',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7624',
                                                                       'out_octets': '1043579514',
                                                                       'out_ucast_pkts': '4179956',
                                                                       'poller_hash': 'MzAwRDgxcGVlcmludGZwb24gMjk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '29',
                                                                                               'links': [       {       'connected_time': '335329',
                                                                                                                        'fiber_length': 18002,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRDgxcGVlcnBvbnBvbiAyOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '2067 Wendell Ct. Nithburg '
                                                 '(SUTHERLAND HEATHER)',
                                         'poller_hash': 'MzAwRDgx',
                                         'serial_number': '300D81',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c8:a9',
                                         'uptime': '335329'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '30/g1',
                                                                       'name': '30/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTUxRTRCQ3BlZXJpbnRmMzAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '30',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTUxRTRCQ3BlZXJpbnRmcG9uIDMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '30',
                                                                                               'links': [       {       'connected_time': '761380',
                                                                                                                        'fiber_length': 255,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.798,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.56}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUxRTRCQ3BlZXJwb25wb24gMzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GS2128G',
                                         'name': 'u4g-test',
                                         'poller_hash': 'MTUxRTRCQw==',
                                         'serial_number': '151E4BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.2.0.2.2',
                                         'system_mac_address': '04:bc:9f:22:fa:c3',
                                         'uptime': '761380'},
                                 {       'counter_source': '31/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '74:da:da:97:2c:18'],
                                                                       'id': '31/g1',
                                                                       'in_nucast_pkts': '1612',
                                                                       'in_octets': '1338475028',
                                                                       'in_ucast_pkts': '8009403',
                                                                       'mac_address': 'ec:4f:82:7d:cb:a3',
                                                                       'mtu': 2000,
                                                                       'name': '31/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394168',
                                                                       'out_octets': '31798162412',
                                                                       'out_ucast_pkts': '23383513',
                                                                       'poller_hash': 'MzAwRUZFcGVlcmludGYzMS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '31',
                                                                       'in_nucast_pkts': '394168',
                                                                       'in_octets': '31798162412',
                                                                       'in_ucast_pkts': '23383513',
                                                                       'links': [       {       'in_nucast_pkts': '394168',
                                                                                                'in_octets': '31798162412',
                                                                                                'in_ucast_pkts': '23383513',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1612',
                                                                                                'out_octets': '1338475028',
                                                                                                'out_ucast_pkts': '8009403',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1612',
                                                                       'out_octets': '1338475028',
                                                                       'out_ucast_pkts': '8009403',
                                                                       'poller_hash': 'MzAwRUZFcGVlcmludGZwb24gMzE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '31',
                                                                                               'links': [       {       'connected_time': '335330',
                                                                                                                        'fiber_length': 17856,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRUZFcGVlcnBvbnBvbiAzMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '2074 Wendell Ct. Nithburg '
                                                 '(Wilmer&Marjorie Gerber)',
                                         'poller_hash': 'MzAwRUZF',
                                         'serial_number': '300EFE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:cb:a3',
                                         'uptime': '335330'},
                                 {       'counter_source': '32/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '32/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '136033870880',
                                                                       'in_ucast_pkts': '727078052',
                                                                       'mac_address': 'ec:4f:82:0d:69:5b',
                                                                       'name': '32/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '63',
                                                                       'out_octets': '2891304678183',
                                                                       'out_ucast_pkts': '2071443548',
                                                                       'poller_hash': 'MjA1REQ2cGVlcmludGYzMi9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '32',
                                                                       'in_nucast_pkts': '63',
                                                                       'in_octets': '2891304678183',
                                                                       'in_ucast_pkts': '2071443548',
                                                                       'links': [       {       'in_nucast_pkts': '63',
                                                                                                'in_octets': '2891304678183',
                                                                                                'in_ucast_pkts': '2071443548',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '136033870880',
                                                                                                'out_ucast_pkts': '727078052',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '136033870880',
                                                                       'out_ucast_pkts': '727078052',
                                                                       'poller_hash': 'MjA1REQ2cGVlcmludGZwb24gMzI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '32',
                                                                                               'links': [       {       'connected_time': '335334',
                                                                                                                        'fiber_length': 18033,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REQ2cGVlcnBvbnBvbiAzMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '2073 WENDELL CT (DARCY MARTIN '
                                                 '10432882)',
                                         'poller_hash': 'MjA1REQ2',
                                         'serial_number': '205DD6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:5b',
                                         'uptime': '335334'},
                                 {       'counter_source': '33/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a6:92'],
                                                                       'id': '33/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '5714610919',
                                                                       'in_ucast_pkts': '47371431',
                                                                       'mac_address': 'ec:4f:82:7d:cb:4d',
                                                                       'mtu': 2000,
                                                                       'name': '33/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394163',
                                                                       'out_octets': '148101646356',
                                                                       'out_ucast_pkts': '104917762',
                                                                       'poller_hash': 'MzAwRUQzcGVlcmludGYzMy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '33',
                                                                       'in_nucast_pkts': '394163',
                                                                       'in_octets': '148101646356',
                                                                       'in_ucast_pkts': '104917762',
                                                                       'links': [       {       'in_nucast_pkts': '394163',
                                                                                                'in_octets': '148101646356',
                                                                                                'in_ucast_pkts': '104917762',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '5714610919',
                                                                                                'out_ucast_pkts': '47371431',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '5714610919',
                                                                       'out_ucast_pkts': '47371431',
                                                                       'poller_hash': 'MzAwRUQzcGVlcmludGZwb24gMzM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '33',
                                                                                               'links': [       {       'connected_time': '335331',
                                                                                                                        'fiber_length': 17657,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRUQzcGVlcnBvbnBvbiAzMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5455 King St. Nithburg (Steven '
                                                 'Careless)',
                                         'poller_hash': 'MzAwRUQz',
                                         'serial_number': '300ED3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:cb:4d',
                                         'uptime': '335331'},
                                 {       'counter_source': '34/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '34/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '355101486340',
                                                                       'in_ucast_pkts': '538438420',
                                                                       'mac_address': 'ec:4f:82:0d:69:e2',
                                                                       'name': '34/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '107',
                                                                       'out_octets': '2749409534493',
                                                                       'out_ucast_pkts': '2195470762',
                                                                       'poller_hash': 'MjA1REU1cGVlcmludGYzNC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '34',
                                                                       'in_nucast_pkts': '107',
                                                                       'in_octets': '2749409534493',
                                                                       'in_ucast_pkts': '2195470762',
                                                                       'links': [       {       'in_nucast_pkts': '107',
                                                                                                'in_octets': '2749409534493',
                                                                                                'in_ucast_pkts': '2195470762',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '355101486340',
                                                                                                'out_ucast_pkts': '538438420',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '355101486340',
                                                                       'out_ucast_pkts': '538438420',
                                                                       'poller_hash': 'MjA1REU1cGVlcmludGZwb24gMzQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '34',
                                                                                               'links': [       {       'connected_time': '335331',
                                                                                                                        'fiber_length': 17110,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1REU1cGVlcnBvbnBvbiAzNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '5556 KING ST (KURUCZ ROSE)',
                                         'poller_hash': 'MjA1REU1',
                                         'serial_number': '205DE5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:e2',
                                         'uptime': '335331'},
                                 {       'counter_source': '35/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:35:4a'],
                                                                       'id': '35/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '3671492189',
                                                                       'in_ucast_pkts': '21408494',
                                                                       'mac_address': 'cc:be:59:bb:0e:34',
                                                                       'mtu': 2000,
                                                                       'name': '35/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394158',
                                                                       'out_octets': '98761079668',
                                                                       'out_ucast_pkts': '68705410',
                                                                       'poller_hash': 'NUE0N0YzcGVlcmludGYzNS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '35',
                                                                       'in_nucast_pkts': '394158',
                                                                       'in_octets': '98761079668',
                                                                       'in_ucast_pkts': '68705410',
                                                                       'links': [       {       'in_nucast_pkts': '394158',
                                                                                                'in_octets': '98761079668',
                                                                                                'in_ucast_pkts': '68705410',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '3671492189',
                                                                                                'out_ucast_pkts': '21408494',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '3671492189',
                                                                       'out_ucast_pkts': '21408494',
                                                                       'poller_hash': 'NUE0N0YzcGVlcmludGZwb24gMzU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '35',
                                                                                               'links': [       {       'connected_time': '335337',
                                                                                                                        'fiber_length': 17358,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE0N0YzcGVlcnBvbnBvbiAzNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5507 King St. Nithburg (Kasey Hale)',
                                         'poller_hash': 'NUE0N0Yz',
                                         'serial_number': '5A47F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:bb:0e:34',
                                         'uptime': '335337'},
                                 {       'counter_source': '37/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:d6:e3'],
                                                                       'id': '37/g1',
                                                                       'in_nucast_pkts': '5994',
                                                                       'in_octets': '2808661113',
                                                                       'in_ucast_pkts': '25260699',
                                                                       'mac_address': 'ec:4f:82:8d:51:8b',
                                                                       'mtu': 2000,
                                                                       'name': '37/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394173',
                                                                       'out_octets': '171188545072',
                                                                       'out_ucast_pkts': '116547015',
                                                                       'poller_hash': 'MzIxMjc4cGVlcmludGYzNy9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '37',
                                                                       'in_nucast_pkts': '394173',
                                                                       'in_octets': '171188545072',
                                                                       'in_ucast_pkts': '116547015',
                                                                       'links': [       {       'in_nucast_pkts': '394173',
                                                                                                'in_octets': '171188545072',
                                                                                                'in_ucast_pkts': '116547015',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5994',
                                                                                                'out_octets': '2808661113',
                                                                                                'out_ucast_pkts': '25260699',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5994',
                                                                       'out_octets': '2808661113',
                                                                       'out_ucast_pkts': '25260699',
                                                                       'poller_hash': 'MzIxMjc4cGVlcmludGZwb24gMzc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '37',
                                                                                               'links': [       {       'connected_time': '335333',
                                                                                                                        'fiber_length': 18071,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMjc4cGVlcnBvbnBvbiAzNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2070 Wendell ct. Nithburg (D & R '
                                                 'Lantz)',
                                         'poller_hash': 'MzIxMjc4',
                                         'serial_number': '321278',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:51:88',
                                         'uptime': '335333'},
                                 {       'counter_source': '38/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:ef:ca'],
                                                                       'id': '38/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '12691611695',
                                                                       'in_ucast_pkts': '107658779',
                                                                       'mac_address': 'ec:4f:82:7d:cb:61',
                                                                       'mtu': 2000,
                                                                       'name': '38/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394165',
                                                                       'out_octets': '172278792600',
                                                                       'out_ucast_pkts': '137900114',
                                                                       'poller_hash': 'MzAwRUREcGVlcmludGYzOC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '38',
                                                                       'in_nucast_pkts': '394165',
                                                                       'in_octets': '172278792600',
                                                                       'in_ucast_pkts': '137900114',
                                                                       'links': [       {       'in_nucast_pkts': '394165',
                                                                                                'in_octets': '172278792600',
                                                                                                'in_ucast_pkts': '137900114',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '12691611695',
                                                                                                'out_ucast_pkts': '107658779',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '12691611695',
                                                                       'out_ucast_pkts': '107658779',
                                                                       'poller_hash': 'MzAwRUREcGVlcmludGZwb24gMzg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '38',
                                                                                               'links': [       {       'connected_time': '335329',
                                                                                                                        'fiber_length': 17020,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRUREcGVlcnBvbnBvbiAzOA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5573 King St. Nithburg (Sean & Tara '
                                                 'Haselden)',
                                         'poller_hash': 'MzAwRURE',
                                         'serial_number': '300EDD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:cb:61',
                                         'uptime': '335329'},
                                 {       'counter_source': '39/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a8:e4'],
                                                                       'id': '39/g1',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '765903816433',
                                                                       'in_ucast_pkts': '1112933210',
                                                                       'mac_address': 'ec:4f:82:7e:3a:c9',
                                                                       'mtu': 2000,
                                                                       'name': '39/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11683981',
                                                                       'out_octets': '2427758842636',
                                                                       'out_ucast_pkts': '1958161678',
                                                                       'poller_hash': 'MzVGOUQ3cGVlcmludGYzOS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '39',
                                                                       'in_nucast_pkts': '11683981',
                                                                       'in_octets': '2427758842636',
                                                                       'in_ucast_pkts': '1958161678',
                                                                       'links': [       {       'in_nucast_pkts': '11683981',
                                                                                                'in_octets': '2427758842636',
                                                                                                'in_ucast_pkts': '1958161678',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '141',
                                                                                                'out_octets': '765903816433',
                                                                                                'out_ucast_pkts': '1112933210',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '765903816433',
                                                                       'out_ucast_pkts': '1112933210',
                                                                       'poller_hash': 'MzVGOUQ3cGVlcmludGZwb24gMzk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '39',
                                                                                               'links': [       {       'connected_time': '10094877',
                                                                                                                        'fiber_length': 17762,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUQ3cGVlcnBvbnBvbiAzOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '5499 BROWN ST (ROES RICK)',
                                         'poller_hash': 'MzVGOUQ3',
                                         'serial_number': '35F9D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3a:c9',
                                         'uptime': '10094877'},
                                 {       'counter_source': '40/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ad:07:46'],
                                                                       'id': '40/g1',
                                                                       'in_nucast_pkts': '8580',
                                                                       'in_octets': '2815073677438',
                                                                       'in_ucast_pkts': '7916983952',
                                                                       'mac_address': 'ec:4f:82:8d:55:e3',
                                                                       'mtu': 2000,
                                                                       'name': '40/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58895848',
                                                                       'out_octets': '29316733848579',
                                                                       'out_ucast_pkts': '21994817961',
                                                                       'poller_hash': 'MzIxMzAzcGVlcmludGY0MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '40',
                                                                       'in_nucast_pkts': '58895848',
                                                                       'in_octets': '29316733848579',
                                                                       'in_ucast_pkts': '21994817961',
                                                                       'links': [       {       'in_nucast_pkts': '58895848',
                                                                                                'in_octets': '29316733848579',
                                                                                                'in_ucast_pkts': '21994817961',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8580',
                                                                                                'out_octets': '2815073677438',
                                                                                                'out_ucast_pkts': '7916983952',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8580',
                                                                       'out_octets': '2815073677438',
                                                                       'out_ucast_pkts': '7916983952',
                                                                       'poller_hash': 'MzIxMzAzcGVlcmludGZwb24gNDA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '40',
                                                                                               'links': [       {       'connected_time': '21886797',
                                                                                                                        'fiber_length': 18017,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMzAzcGVlcnBvbnBvbiA0MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2066 WENDELL CT (PRINCE JODY & '
                                                 'MICHELLE)',
                                         'poller_hash': 'MzIxMzAz',
                                         'serial_number': '321303',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:55:e0',
                                         'uptime': '21886797'},
                                 {       'counter_source': '41/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a4:fd'],
                                                                       'id': '41/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '7385737658',
                                                                       'in_ucast_pkts': '28934125',
                                                                       'mac_address': 'ec:4f:82:ae:bd:0d',
                                                                       'mtu': 2000,
                                                                       'name': '41/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394170',
                                                                       'out_octets': '111625256325',
                                                                       'out_ucast_pkts': '85349228',
                                                                       'poller_hash': 'M0I2MUU1cGVlcmludGY0MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '41',
                                                                       'in_nucast_pkts': '394170',
                                                                       'in_octets': '111625256325',
                                                                       'in_ucast_pkts': '85349228',
                                                                       'links': [       {       'in_nucast_pkts': '394170',
                                                                                                'in_octets': '111625256325',
                                                                                                'in_ucast_pkts': '85349228',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '7385737658',
                                                                                                'out_ucast_pkts': '28934125',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '7385737658',
                                                                       'out_ucast_pkts': '28934125',
                                                                       'poller_hash': 'M0I2MUU1cGVlcmludGZwb24gNDE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '41',
                                                                                               'links': [       {       'connected_time': '335333',
                                                                                                                        'fiber_length': 17175,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0I2MUU1cGVlcnBvbnBvbiA0MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '2039A Perth Line 56 Nithburg (Shawn '
                                                 'Conn)',
                                         'poller_hash': 'M0I2MUU1',
                                         'serial_number': '3B61E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:ae:bd:0d',
                                         'uptime': '335333'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '42/g1',
                                                                       'name': '42/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTU3Mzg0RXBlZXJpbnRmNDIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '42',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTU3Mzg0RXBlZXJpbnRmcG9uIDQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '42',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU3Mzg0RXBlZXJwb25wb24gNDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'MCCL Test 8802 (8802)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "42" (location: '
                                                                  'GPON port "7/8"), '
                                                                  '"ONT went missing"',
                                         'offline_time': '2025-02-12T18:36:15.290Z',
                                         'poller_hash': 'MTU3Mzg0RQ==',
                                         'serial_number': '157384E',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '43/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3c:3e'],
                                                                       'id': '43/g1',
                                                                       'in_nucast_pkts': '14909',
                                                                       'in_octets': '6848664887',
                                                                       'in_ucast_pkts': '64676349',
                                                                       'mac_address': 'ec:4f:82:59:ea:b6',
                                                                       'mtu': 2000,
                                                                       'name': '43/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '912792',
                                                                       'out_octets': '333222240438',
                                                                       'out_ucast_pkts': '224823375',
                                                                       'poller_hash': 'MkFEMEM5cGVlcmludGY0My9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '43',
                                                                       'in_nucast_pkts': '912792',
                                                                       'in_octets': '333222240438',
                                                                       'in_ucast_pkts': '224823375',
                                                                       'links': [       {       'in_nucast_pkts': '912792',
                                                                                                'in_octets': '333222240438',
                                                                                                'in_ucast_pkts': '224823375',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14909',
                                                                                                'out_octets': '6848664887',
                                                                                                'out_ucast_pkts': '64676349',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14909',
                                                                       'out_octets': '6848664887',
                                                                       'out_ucast_pkts': '64676349',
                                                                       'poller_hash': 'MkFEMEM5cGVlcmludGZwb24gNDM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '43',
                                                                                               'links': [       {       'connected_time': '779562',
                                                                                                                        'fiber_length': 17566,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFEMEM5cGVlcnBvbnBvbiA0Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5465 King St. Nithburg (James Green)',
                                         'poller_hash': 'MkFEMEM5',
                                         'serial_number': '2AD0C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:ea:b3',
                                         'uptime': '779562'},
                                 {       'counter_source': '44/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:f0:fc'],
                                                                       'id': '44/g1',
                                                                       'in_nucast_pkts': '13083',
                                                                       'in_octets': '1303428109638',
                                                                       'in_ucast_pkts': '4485084075',
                                                                       'mac_address': 'ec:4f:82:5b:80:be',
                                                                       'mtu': 2000,
                                                                       'name': '44/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58901435',
                                                                       'out_octets': '22243222925148',
                                                                       'out_ucast_pkts': '16164417988',
                                                                       'poller_hash': 'MkIwMzhBcGVlcmludGY0NC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '44',
                                                                       'in_nucast_pkts': '58901435',
                                                                       'in_octets': '22243222925148',
                                                                       'in_ucast_pkts': '16164417988',
                                                                       'links': [       {       'in_nucast_pkts': '58901435',
                                                                                                'in_octets': '22243222925148',
                                                                                                'in_ucast_pkts': '16164417988',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13083',
                                                                                                'out_octets': '1303428109638',
                                                                                                'out_ucast_pkts': '4485084075',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13083',
                                                                       'out_octets': '1303428109638',
                                                                       'out_ucast_pkts': '4485084075',
                                                                       'poller_hash': 'MkIwMzhBcGVlcmludGZwb24gNDQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '44',
                                                                                               'links': [       {       'connected_time': '21886021',
                                                                                                                        'fiber_length': 18049,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMzhBcGVlcnBvbnBvbiA0NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2068 Wendell Ct. Nithburg (Lisa Koch '
                                                 '10433828)',
                                         'poller_hash': 'MkIwMzhB',
                                         'serial_number': '2B038A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:80:bb',
                                         'uptime': '21886021'},
                                 {       'counter_source': '45/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c5:5f:6a'],
                                                                       'id': '45/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '729654976',
                                                                       'in_ucast_pkts': '4830220',
                                                                       'mac_address': 'ec:4f:82:7d:ca:c1',
                                                                       'mtu': 2000,
                                                                       'name': '45/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394167',
                                                                       'out_octets': '34525886895',
                                                                       'out_ucast_pkts': '25191438',
                                                                       'poller_hash': 'MzAwRThEcGVlcmludGY0NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '45',
                                                                       'in_nucast_pkts': '394167',
                                                                       'in_octets': '34525886895',
                                                                       'in_ucast_pkts': '25191438',
                                                                       'links': [       {       'in_nucast_pkts': '394167',
                                                                                                'in_octets': '34525886895',
                                                                                                'in_ucast_pkts': '25191438',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '729654976',
                                                                                                'out_ucast_pkts': '4830220',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '729654976',
                                                                       'out_ucast_pkts': '4830220',
                                                                       'poller_hash': 'MzAwRThEcGVlcmludGZwb24gNDU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '45',
                                                                                               'links': [       {       'connected_time': '335335',
                                                                                                                        'fiber_length': 17396,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRThEcGVlcnBvbnBvbiA0NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5500 King St. Nithburg (Sheila '
                                                 'Gustafson)',
                                         'poller_hash': 'MzAwRThE',
                                         'serial_number': '300E8D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:ca:c1',
                                         'uptime': '335335'},
                                 {       'counter_source': '46/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '46/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1058441259552',
                                                                       'in_ucast_pkts': '1730330380',
                                                                       'mac_address': 'd0:76:8f:1b:a6:8b',
                                                                       'name': '46/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '90',
                                                                       'out_octets': '3264879733680',
                                                                       'out_ucast_pkts': '3098020970',
                                                                       'poller_hash': 'ODcyNTFGcGVlcmludGY0Ni9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '46',
                                                                       'in_nucast_pkts': '90',
                                                                       'in_octets': '3264879733680',
                                                                       'in_ucast_pkts': '3098020970',
                                                                       'links': [       {       'in_nucast_pkts': '90',
                                                                                                'in_octets': '3264879733680',
                                                                                                'in_ucast_pkts': '3098020970',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1058441259552',
                                                                                                'out_ucast_pkts': '1730330380',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1058441259552',
                                                                       'out_ucast_pkts': '1730330380',
                                                                       'poller_hash': 'ODcyNTFGcGVlcmludGZwb24gNDY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '46',
                                                                                               'links': [       {       'connected_time': '335338',
                                                                                                                        'fiber_length': 17099,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTFGcGVlcnBvbnBvbiA0Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2086 NITHVIEW CT (BERARIU SIMONA)',
                                         'poller_hash': 'ODcyNTFG',
                                         'serial_number': '87251F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a6:8b',
                                         'uptime': '335338'},
                                 {       'counter_source': '47/F1+47/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a1:bf:84',
                                                                                                        '00:03:e6:a4:d9:c4'],
                                                                       'id': '47/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2035277693',
                                                                       'in_ucast_pkts': '10832988',
                                                                       'name': '47/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '7093507309',
                                                                       'out_octets': '9565381274575',
                                                                       'out_ucast_pkts': '11583035',
                                                                       'poller_hash': 'MkFEMEQ3cGVlcmludGY0Ny9GMQ=='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a9:a9:1a'],
                                                                       'id': '47/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '5186398121',
                                                                       'in_ucast_pkts': '50275217',
                                                                       'mac_address': 'ec:4f:82:59:eb:26',
                                                                       'mtu': 2000,
                                                                       'name': '47/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394170',
                                                                       'out_octets': '234495815052',
                                                                       'out_ucast_pkts': '157473784',
                                                                       'poller_hash': 'MkFEMEQ3cGVlcmludGY0Ny9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '47',
                                                                       'in_nucast_pkts': '7093901479',
                                                                       'in_octets': '9799877089627',
                                                                       'in_ucast_pkts': '169056819',
                                                                       'links': [       {       'in_nucast_pkts': '7093901479',
                                                                                                'in_octets': '9799877089627',
                                                                                                'in_ucast_pkts': '169056819',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '7221675814',
                                                                                                'out_ucast_pkts': '61108205',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '7221675814',
                                                                       'out_ucast_pkts': '61108205',
                                                                       'poller_hash': 'MkFEMEQ3cGVlcmludGZwb24gNDc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '47',
                                                                                               'links': [       {       'connected_time': '335337',
                                                                                                                        'fiber_length': 17551,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFEMEQ3cGVlcnBvbnBvbiA0Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5479 King St. Nithburg (Shannon '
                                                 'Gingerich)',
                                         'poller_hash': 'MkFEMEQ3',
                                         'serial_number': '2AD0D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:eb:23',
                                         'uptime': '335337'},
                                 {       'counter_source': '48/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd4:01:c3:0a:d7:1b'],
                                                                       'id': '48/g1',
                                                                       'in_nucast_pkts': '24',
                                                                       'in_octets': '1269891017',
                                                                       'in_ucast_pkts': '3718687',
                                                                       'mac_address': 'ec:4f:82:7d:c9:4d',
                                                                       'mtu': 2000,
                                                                       'name': '48/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394166',
                                                                       'out_octets': '12082573167',
                                                                       'out_ucast_pkts': '10444064',
                                                                       'poller_hash': 'MzAwREQzcGVlcmludGY0OC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '48',
                                                                       'in_nucast_pkts': '394166',
                                                                       'in_octets': '12082573167',
                                                                       'in_ucast_pkts': '10444064',
                                                                       'links': [       {       'in_nucast_pkts': '394166',
                                                                                                'in_octets': '12082573167',
                                                                                                'in_ucast_pkts': '10444064',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '24',
                                                                                                'out_octets': '1269891017',
                                                                                                'out_ucast_pkts': '3718687',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24',
                                                                       'out_octets': '1269891017',
                                                                       'out_ucast_pkts': '3718687',
                                                                       'poller_hash': 'MzAwREQzcGVlcmludGZwb24gNDg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '48',
                                                                                               'links': [       {       'connected_time': '335335',
                                                                                                                        'fiber_length': 17706,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwREQzcGVlcnBvbnBvbiA0OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5444 KING ST (KUEPFER PETER)',
                                         'poller_hash': 'MzAwREQz',
                                         'serial_number': '300DD3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:c9:4d',
                                         'uptime': '335335'},
                                 {       'counter_source': '49/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:f5:a2:ef:29:2a'],
                                                                       'id': '49/g1',
                                                                       'in_nucast_pkts': '12195',
                                                                       'in_octets': '839381496',
                                                                       'in_ucast_pkts': '5222318',
                                                                       'mac_address': 'ec:4f:82:7d:ca:e7',
                                                                       'mtu': 2000,
                                                                       'name': '49/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394366',
                                                                       'out_octets': '25637680047',
                                                                       'out_ucast_pkts': '17944552',
                                                                       'poller_hash': 'MzAwRUEwcGVlcmludGY0OS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '49',
                                                                       'in_nucast_pkts': '394366',
                                                                       'in_octets': '25637680047',
                                                                       'in_ucast_pkts': '17944552',
                                                                       'links': [       {       'in_nucast_pkts': '394366',
                                                                                                'in_octets': '25637680047',
                                                                                                'in_ucast_pkts': '17944552',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12195',
                                                                                                'out_octets': '839381496',
                                                                                                'out_ucast_pkts': '5222318',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12195',
                                                                       'out_octets': '839381496',
                                                                       'out_ucast_pkts': '5222318',
                                                                       'poller_hash': 'MzAwRUEwcGVlcmludGZwb24gNDk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '49',
                                                                                               'links': [       {       'connected_time': '335334',
                                                                                                                        'fiber_length': 17166,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwRUEwcGVlcnBvbnBvbiA0OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '2089 Nithview Ct. Nithburg '
                                                 '(HAWTHORNTHWAITE ROGER)',
                                         'poller_hash': 'MzAwRUEw',
                                         'serial_number': '300EA0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:ca:e7',
                                         'uptime': '335334'},
                                 {       'counter_source': '50/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:08:f8:b2'],
                                                                       'id': '50/g1',
                                                                       'in_nucast_pkts': '82732',
                                                                       'in_octets': '135334387416',
                                                                       'in_ucast_pkts': '275983459',
                                                                       'mac_address': 'ec:4f:82:8d:48:c3',
                                                                       'mtu': 2000,
                                                                       'name': '50/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4945372',
                                                                       'out_octets': '714454014035',
                                                                       'out_ucast_pkts': '662075810',
                                                                       'poller_hash': 'MzIxMTVGcGVlcmludGY1MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '50',
                                                                       'in_nucast_pkts': '4945372',
                                                                       'in_octets': '714454014035',
                                                                       'in_ucast_pkts': '662075810',
                                                                       'links': [       {       'in_nucast_pkts': '4945372',
                                                                                                'in_octets': '714454014035',
                                                                                                'in_ucast_pkts': '662075810',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '82732',
                                                                                                'out_octets': '135334387416',
                                                                                                'out_ucast_pkts': '275983459',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82732',
                                                                       'out_octets': '135334387416',
                                                                       'out_ucast_pkts': '275983459',
                                                                       'poller_hash': 'MzIxMTVGcGVlcmludGZwb24gNTA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '50',
                                                                                               'links': [       {       'connected_time': '4218997',
                                                                                                                        'fiber_length': 17701,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTVGcGVlcnBvbnBvbiA1MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5471 BROWN ST (ABBOTT TYLER)',
                                         'poller_hash': 'MzIxMTVG',
                                         'serial_number': '32115F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:48:c0',
                                         'uptime': '4218997'},
                                 {       'counter_source': '51/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:e3:2b'],
                                                                       'id': '51/g1',
                                                                       'in_nucast_pkts': '8529',
                                                                       'in_octets': '844743337963',
                                                                       'in_ucast_pkts': '5737944460',
                                                                       'mac_address': 'ec:4f:82:8d:48:0b',
                                                                       'mtu': 2000,
                                                                       'name': '51/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58901429',
                                                                       'out_octets': '24931331451789',
                                                                       'out_ucast_pkts': '17398577981',
                                                                       'poller_hash': 'MzIxMTQ4cGVlcmludGY1MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '51',
                                                                       'in_nucast_pkts': '58901429',
                                                                       'in_octets': '24931331451789',
                                                                       'in_ucast_pkts': '17398577981',
                                                                       'links': [       {       'in_nucast_pkts': '58901429',
                                                                                                'in_octets': '24931331451789',
                                                                                                'in_ucast_pkts': '17398577981',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8529',
                                                                                                'out_octets': '844743337963',
                                                                                                'out_ucast_pkts': '5737944460',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8529',
                                                                       'out_octets': '844743337963',
                                                                       'out_ucast_pkts': '5737944460',
                                                                       'poller_hash': 'MzIxMTQ4cGVlcmludGZwb24gNTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '51',
                                                                                               'links': [       {       'connected_time': '21885988',
                                                                                                                        'fiber_length': 17710,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTQ4cGVlcnBvbnBvbiA1MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5489 Brown St. Nithburg (Janice '
                                                 'Castell)',
                                         'poller_hash': 'MzIxMTQ4',
                                         'serial_number': '321148',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:48:08',
                                         'uptime': '21885988'},
                                 {       'counter_source': '52/F1+52/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a1:c1:ba',
                                                                                                        '5c:0f:fb:2f:ce:d4'],
                                                                       'id': '52/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '13962162788',
                                                                       'in_ucast_pkts': '29859036',
                                                                       'name': '52/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '41893706179',
                                                                       'out_octets': '56368214857533',
                                                                       'out_ucast_pkts': '35813916',
                                                                       'poller_hash': 'MzIxMjI2cGVlcmludGY1Mi9GMQ=='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5a:5e'],
                                                                       'id': '52/g1',
                                                                       'in_nucast_pkts': '258296',
                                                                       'in_octets': '804585596978',
                                                                       'in_ucast_pkts': '1938934651',
                                                                       'mac_address': 'ec:4f:82:8d:4e:fb',
                                                                       'mtu': 2000,
                                                                       'name': '52/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43768031',
                                                                       'out_octets': '6690498715627',
                                                                       'out_ucast_pkts': '5090389282',
                                                                       'poller_hash': 'MzIxMjI2cGVlcmludGY1Mi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '52',
                                                                       'in_nucast_pkts': '41937474210',
                                                                       'in_octets': '63058713573160',
                                                                       'in_ucast_pkts': '5126203198',
                                                                       'links': [       {       'in_nucast_pkts': '41937474210',
                                                                                                'in_octets': '63058713573160',
                                                                                                'in_ucast_pkts': '5126203198',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '258296',
                                                                                                'out_octets': '818547759766',
                                                                                                'out_ucast_pkts': '1968793687',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '258296',
                                                                       'out_octets': '818547759766',
                                                                       'out_ucast_pkts': '1968793687',
                                                                       'poller_hash': 'MzIxMjI2cGVlcmludGZwb24gNTI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '52',
                                                                                               'links': [       {       'connected_time': '21886170',
                                                                                                                        'fiber_length': 17784,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMjI2cGVlcnBvbnBvbiA1Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5521 Brown St. Nithburg (John '
                                                 'MacDonald)',
                                         'poller_hash': 'MzIxMjI2',
                                         'serial_number': '321226',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:4e:f8',
                                         'uptime': '21886170'},
                                 {       'counter_source': '53/F1+53/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a4:d9:84'],
                                                                       'id': '53/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1751417222',
                                                                       'in_ucast_pkts': '10008113',
                                                                       'name': '53/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '15716324161',
                                                                       'out_octets': '20982157135326',
                                                                       'out_ucast_pkts': '8822097',
                                                                       'poller_hash': 'MzIxMTM4cGVlcmludGY1My9GMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '53/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '249515483182',
                                                                       'in_ucast_pkts': '1508672153',
                                                                       'mac_address': 'ec:4f:82:8d:47:88',
                                                                       'name': '53/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '79',
                                                                       'out_octets': '4911121671235',
                                                                       'out_ucast_pkts': '3662618173',
                                                                       'poller_hash': 'MzIxMTM4cGVlcmludGY1My9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '53',
                                                                       'in_nucast_pkts': '15716324240',
                                                                       'in_octets': '25893278806561',
                                                                       'in_ucast_pkts': '3671440270',
                                                                       'links': [       {       'in_nucast_pkts': '15716324240',
                                                                                                'in_octets': '25893278806561',
                                                                                                'in_ucast_pkts': '3671440270',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '251266900404',
                                                                                                'out_ucast_pkts': '1518680266',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '251266900404',
                                                                       'out_ucast_pkts': '1518680266',
                                                                       'poller_hash': 'MzIxMTM4cGVlcmludGZwb24gNTM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '53',
                                                                                               'links': [       {       'connected_time': '21886514',
                                                                                                                        'fiber_length': 17498,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTM4cGVlcnBvbnBvbiA1Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5490 King St. Nithburg (Jody Broda)',
                                         'poller_hash': 'MzIxMTM4',
                                         'serial_number': '321138',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:47:88',
                                         'uptime': '21886514'},
                                 {       'counter_source': '54/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4f:b1:c2'],
                                                                       'id': '54/g1',
                                                                       'in_nucast_pkts': '6210',
                                                                       'in_octets': '5379421976',
                                                                       'in_ucast_pkts': '21465088',
                                                                       'mac_address': 'ec:4f:82:8d:4a:4b',
                                                                       'mtu': 2000,
                                                                       'name': '54/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394165',
                                                                       'out_octets': '86739811594',
                                                                       'out_ucast_pkts': '62238211',
                                                                       'poller_hash': 'MzIxMTkwcGVlcmludGY1NC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '54',
                                                                       'in_nucast_pkts': '394165',
                                                                       'in_octets': '86739811594',
                                                                       'in_ucast_pkts': '62238211',
                                                                       'links': [       {       'in_nucast_pkts': '394165',
                                                                                                'in_octets': '86739811594',
                                                                                                'in_ucast_pkts': '62238211',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6210',
                                                                                                'out_octets': '5379421976',
                                                                                                'out_ucast_pkts': '21465088',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6210',
                                                                       'out_octets': '5379421976',
                                                                       'out_ucast_pkts': '21465088',
                                                                       'poller_hash': 'MzIxMTkwcGVlcmludGZwb24gNTQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '54',
                                                                                               'links': [       {       'connected_time': '335336',
                                                                                                                        'fiber_length': 17724,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTkwcGVlcnBvbnBvbiA1NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5478 BROWN ST (LINGELBACH SARAH)',
                                         'poller_hash': 'MzIxMTkw',
                                         'serial_number': '321190',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:4a:48',
                                         'uptime': '335336'},
                                 {       'counter_source': '55/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:55:ce'],
                                                                       'id': '55/g1',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '103583470052',
                                                                       'in_ucast_pkts': '235215430',
                                                                       'mac_address': 'ec:4f:82:8d:48:2b',
                                                                       'mtu': 2000,
                                                                       'name': '55/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27648248',
                                                                       'out_octets': '590102497405',
                                                                       'out_ucast_pkts': '512780840',
                                                                       'poller_hash': 'MzIxMTRDcGVlcmludGY1NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '55',
                                                                       'in_nucast_pkts': '27648248',
                                                                       'in_octets': '590102497405',
                                                                       'in_ucast_pkts': '512780840',
                                                                       'links': [       {       'in_nucast_pkts': '27648248',
                                                                                                'in_octets': '590102497405',
                                                                                                'in_ucast_pkts': '512780840',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '132',
                                                                                                'out_octets': '103583470052',
                                                                                                'out_ucast_pkts': '235215430',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '103583470052',
                                                                       'out_ucast_pkts': '235215430',
                                                                       'poller_hash': 'MzIxMTRDcGVlcmludGZwb24gNTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '55',
                                                                                               'links': [       {       'connected_time': '24896593',
                                                                                                                        'fiber_length': 17234,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTRDcGVlcnBvbnBvbiA1NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2039B Perth Line 56 Nithburg (Delmer '
                                                 '& Mary Erb)',
                                         'poller_hash': 'MzIxMTRD',
                                         'serial_number': '32114C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:48:28',
                                         'uptime': '24896593'},
                                 {       'counter_source': '56/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e7:fd'],
                                                                       'id': '56/g1',
                                                                       'in_nucast_pkts': '106227',
                                                                       'in_octets': '27532153016',
                                                                       'in_ucast_pkts': '162940047',
                                                                       'mac_address': 'ec:4f:82:8d:49:43',
                                                                       'mtu': 2000,
                                                                       'name': '56/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7099079',
                                                                       'out_octets': '955255941446',
                                                                       'out_ucast_pkts': '681646381',
                                                                       'poller_hash': 'MzIxMTZGcGVlcmludGY1Ni9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '56',
                                                                       'in_nucast_pkts': '7099079',
                                                                       'in_octets': '955255941446',
                                                                       'in_ucast_pkts': '681646381',
                                                                       'links': [       {       'in_nucast_pkts': '7099079',
                                                                                                'in_octets': '955255941446',
                                                                                                'in_ucast_pkts': '681646381',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '106227',
                                                                                                'out_octets': '27532153016',
                                                                                                'out_ucast_pkts': '162940047',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '106227',
                                                                       'out_octets': '27532153016',
                                                                       'out_ucast_pkts': '162940047',
                                                                       'poller_hash': 'MzIxMTZGcGVlcmludGZwb24gNTY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '56',
                                                                                               'links': [       {       'connected_time': '7398406',
                                                                                                                        'fiber_length': 17860,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTZGcGVlcnBvbnBvbiA1Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5428 King St. Nithburg (Raymond '
                                                 'Lemieux)',
                                         'poller_hash': 'MzIxMTZG',
                                         'serial_number': '32116F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:49:40',
                                         'uptime': '7398406'},
                                 {       'counter_source': '57/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:f8:b1:89'],
                                                                       'id': '57/g1',
                                                                       'in_nucast_pkts': '6454882',
                                                                       'in_octets': '482855941819',
                                                                       'in_ucast_pkts': '1050534021',
                                                                       'mac_address': 'ec:4f:82:8d:49:5b',
                                                                       'mtu': 2000,
                                                                       'name': '57/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5516774313',
                                                                       'out_octets': '10633040437161',
                                                                       'out_ucast_pkts': '2957806147',
                                                                       'poller_hash': 'MzIxMTcycGVlcmludGY1Ny9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '57',
                                                                       'in_nucast_pkts': '5516774313',
                                                                       'in_octets': '10633040437161',
                                                                       'in_ucast_pkts': '2957806147',
                                                                       'links': [       {       'in_nucast_pkts': '5516774313',
                                                                                                'in_octets': '10633040437161',
                                                                                                'in_ucast_pkts': '2957806147',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6454882',
                                                                                                'out_octets': '482855941819',
                                                                                                'out_ucast_pkts': '1050534021',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6454882',
                                                                       'out_octets': '482855941819',
                                                                       'out_ucast_pkts': '1050534021',
                                                                       'poller_hash': 'MzIxMTcycGVlcmludGZwb24gNTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '57',
                                                                                               'links': [       {       'connected_time': '8384839',
                                                                                                                        'fiber_length': 17626,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTcycGVlcnBvbnBvbiA1Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5476 King St. Nithburg (Ed Jernigan)',
                                         'poller_hash': 'MzIxMTcy',
                                         'serial_number': '321172',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:49:58',
                                         'uptime': '8384839'},
                                 {       'counter_source': '58/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '58/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '206678523392',
                                                                       'in_ucast_pkts': '1846654465',
                                                                       'mac_address': 'ec:4f:82:8d:46:08',
                                                                       'name': '58/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '197',
                                                                       'out_octets': '10996326448635',
                                                                       'out_ucast_pkts': '7464424405',
                                                                       'poller_hash': 'MzIxMTA4cGVlcmludGY1OC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '58',
                                                                       'in_nucast_pkts': '197',
                                                                       'in_octets': '10996326448635',
                                                                       'in_ucast_pkts': '7464424405',
                                                                       'links': [       {       'in_nucast_pkts': '197',
                                                                                                'in_octets': '10996326448635',
                                                                                                'in_ucast_pkts': '7464424405',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '206678523392',
                                                                                                'out_ucast_pkts': '1846654465',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '206678523392',
                                                                       'out_ucast_pkts': '1846654465',
                                                                       'poller_hash': 'MzIxMTA4cGVlcmludGZwb24gNTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '58',
                                                                                               'links': [       {       'connected_time': '335335',
                                                                                                                        'fiber_length': 17149,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTA4cGVlcnBvbnBvbiA1OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5569 KING ST (TAYLOR KATHLEEN)',
                                         'poller_hash': 'MzIxMTA4',
                                         'serial_number': '321108',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:46:08',
                                         'uptime': '335335'},
                                 {       'counter_source': '59/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'bc:22:28:85:b9:b2'],
                                                                       'id': '59/g1',
                                                                       'in_nucast_pkts': '73',
                                                                       'in_octets': '26616783',
                                                                       'in_ucast_pkts': '216680',
                                                                       'mac_address': 'ec:4f:82:ae:ae:ff',
                                                                       'mtu': 2000,
                                                                       'name': '59/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394170',
                                                                       'out_octets': '87454570',
                                                                       'out_ucast_pkts': '243861',
                                                                       'poller_hash': 'M0I1QURFcGVlcmludGY1OS9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '59',
                                                                       'in_nucast_pkts': '394170',
                                                                       'in_octets': '87454570',
                                                                       'in_ucast_pkts': '243861',
                                                                       'links': [       {       'in_nucast_pkts': '394170',
                                                                                                'in_octets': '87454570',
                                                                                                'in_ucast_pkts': '243861',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '73',
                                                                                                'out_octets': '26616783',
                                                                                                'out_ucast_pkts': '216680',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '73',
                                                                       'out_octets': '26616783',
                                                                       'out_ucast_pkts': '216680',
                                                                       'poller_hash': 'M0I1QURFcGVlcmludGZwb24gNTk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '59',
                                                                                               'links': [       {       'connected_time': '335336',
                                                                                                                        'fiber_length': 17146,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0I1QURFcGVlcnBvbnBvbiA1OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5568 KING ST (MEADOWS BRETT)',
                                         'poller_hash': 'M0I1QURF',
                                         'serial_number': '3B5ADE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:ae:ae:ff',
                                         'uptime': '335336'},
                                 {       'counter_source': '60/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '7c:5a:1c:9b:15:0e'],
                                                                       'id': '60/g1',
                                                                       'in_nucast_pkts': '8452988',
                                                                       'in_octets': '29173026413535',
                                                                       'in_ucast_pkts': '24020824783',
                                                                       'mac_address': 'd0:76:8f:aa:52:f7',
                                                                       'mtu': 2000,
                                                                       'name': '60/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19109204',
                                                                       'out_octets': '13971915569004',
                                                                       'out_ucast_pkts': '20574354817',
                                                                       'poller_hash': 'OTU2Rjg0cGVlcmludGY2MC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '60',
                                                                       'in_nucast_pkts': '19109204',
                                                                       'in_octets': '13971915569004',
                                                                       'in_ucast_pkts': '20574354817',
                                                                       'links': [       {       'in_nucast_pkts': '19109204',
                                                                                                'in_octets': '13971915569004',
                                                                                                'in_ucast_pkts': '20574354817',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8452988',
                                                                                                'out_octets': '29173026413535',
                                                                                                'out_ucast_pkts': '24020824783',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8452988',
                                                                       'out_octets': '29173026413535',
                                                                       'out_ucast_pkts': '24020824783',
                                                                       'poller_hash': 'OTU2Rjg0cGVlcmludGZwb24gNjA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '60',
                                                                                               'links': [       {       'connected_time': '16515557',
                                                                                                                        'fiber_length': 21385,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2Rjg0cGVlcnBvbnBvbiA2MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1315  Hutchinson Rd Nithburg (Leis '
                                                 'Pet Distribution 10434255)',
                                         'poller_hash': 'OTU2Rjg0',
                                         'serial_number': '956F84',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:52:f4',
                                         'uptime': '16515557'},
                                 {       'counter_source': '61/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '94:83:c4:60:11:10'],
                                                                       'id': '61/g1',
                                                                       'in_nucast_pkts': '32795',
                                                                       'in_octets': '6738665008241',
                                                                       'in_ucast_pkts': '6954930528',
                                                                       'mac_address': 'ec:4f:82:28:82:25',
                                                                       'mtu': 2000,
                                                                       'name': '61/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19103087',
                                                                       'out_octets': '4538209192881',
                                                                       'out_ucast_pkts': '6116501368',
                                                                       'poller_hash': 'MjZBQzZDcGVlcmludGY2MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '61',
                                                                       'in_nucast_pkts': '19103087',
                                                                       'in_octets': '4538209192881',
                                                                       'in_ucast_pkts': '6116501368',
                                                                       'links': [       {       'in_nucast_pkts': '19103087',
                                                                                                'in_octets': '4538209192881',
                                                                                                'in_ucast_pkts': '6116501368',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '32795',
                                                                                                'out_octets': '6738665008241',
                                                                                                'out_ucast_pkts': '6954930528',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32795',
                                                                       'out_octets': '6738665008241',
                                                                       'out_ucast_pkts': '6954930528',
                                                                       'poller_hash': 'MjZBQzZDcGVlcmludGZwb24gNjE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '61',
                                                                                               'links': [       {       'connected_time': '3036359',
                                                                                                                        'fiber_length': 17665,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjZBQzZDcGVlcnBvbnBvbiA2MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '762GX',
                                         'name': '2076 Perth line 56 (DORTMAN BROS '
                                                 'BARN EQUIP)',
                                         'poller_hash': 'MjZBQzZD',
                                         'serial_number': '26AC6C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:28:82:2a',
                                         'uptime': '3036359'},
                                 {       'counter_source': '63/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a5:53'],
                                                                       'id': '63/g1',
                                                                       'in_nucast_pkts': '440',
                                                                       'in_octets': '474118066278',
                                                                       'in_ucast_pkts': '1100822703',
                                                                       'mac_address': 'ec:4f:82:7e:3b:0b',
                                                                       'mtu': 2000,
                                                                       'name': '63/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '37965823',
                                                                       'out_octets': '5287160512784',
                                                                       'out_ucast_pkts': '4319478664',
                                                                       'poller_hash': 'MzVGOUVEcGVlcmludGY2My9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '63',
                                                                       'in_nucast_pkts': '37965823',
                                                                       'in_octets': '5287160512784',
                                                                       'in_ucast_pkts': '4319478664',
                                                                       'links': [       {       'in_nucast_pkts': '37965823',
                                                                                                'in_octets': '5287160512784',
                                                                                                'in_ucast_pkts': '4319478664',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '440',
                                                                                                'out_octets': '474118066278',
                                                                                                'out_ucast_pkts': '1100822703',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '440',
                                                                       'out_octets': '474118066278',
                                                                       'out_ucast_pkts': '1100822703',
                                                                       'poller_hash': 'MzVGOUVEcGVlcmludGZwb24gNjM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '63',
                                                                                               'links': [       {       'connected_time': '67856246',
                                                                                                                        'fiber_length': 14339,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUVEcGVlcnBvbnBvbiA2Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '2291 Perth Line 56 (Ruthann Erb '
                                                 '10433944)',
                                         'poller_hash': 'MzVGOUVE',
                                         'serial_number': '35F9ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3b:0b',
                                         'uptime': '67856246'},
                                 {       'counter_source': '65/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:8b:cd:ae:8b:48'],
                                                                       'id': '65/g1',
                                                                       'in_nucast_pkts': '1470',
                                                                       'in_octets': '150222209718',
                                                                       'in_ucast_pkts': '887181915',
                                                                       'mac_address': 'ec:4f:82:7e:08:ce',
                                                                       'mtu': 2000,
                                                                       'name': '65/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '37928381',
                                                                       'out_octets': '3561532359952',
                                                                       'out_ucast_pkts': '2516767701',
                                                                       'poller_hash': 'MzVFOTNBcGVlcmludGY2NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '65',
                                                                       'in_nucast_pkts': '37928381',
                                                                       'in_octets': '3561532359952',
                                                                       'in_ucast_pkts': '2516767701',
                                                                       'links': [       {       'in_nucast_pkts': '37928381',
                                                                                                'in_octets': '3561532359952',
                                                                                                'in_ucast_pkts': '2516767701',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1470',
                                                                                                'out_octets': '150222209718',
                                                                                                'out_ucast_pkts': '887181915',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1470',
                                                                       'out_octets': '150222209718',
                                                                       'out_ucast_pkts': '887181915',
                                                                       'poller_hash': 'MzVFOTNBcGVlcmludGZwb24gNjU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '65',
                                                                                               'links': [       {       'connected_time': '67856448',
                                                                                                                        'fiber_length': 14918,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTNBcGVlcnBvbnBvbiA2NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '2249 PERTH LINE 56 (SCHULTZ KEITH & '
                                                 'FAYE)',
                                         'poller_hash': 'MzVFOTNB',
                                         'serial_number': '35E93A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:08:ce',
                                         'uptime': '67856448'},
                                 {       'counter_source': '66/F1+66/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a8:0e:9e'],
                                                                       'id': '66/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1619953302',
                                                                       'in_ucast_pkts': '9491443',
                                                                       'name': '66/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '6913875505',
                                                                       'out_octets': '8972260245053',
                                                                       'out_ucast_pkts': '5716690',
                                                                       'poller_hash': 'MzIxMEQ4cGVlcmludGY2Ni9GMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '66/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '584187031529',
                                                                       'in_ucast_pkts': '138828512142',
                                                                       'mac_address': 'ec:4f:82:8d:44:88',
                                                                       'name': '66/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '161',
                                                                       'out_octets': '4550787352873',
                                                                       'out_ucast_pkts': '3844346537',
                                                                       'poller_hash': 'MzIxMEQ4cGVlcmludGY2Ni9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '66',
                                                                       'in_nucast_pkts': '6913875666',
                                                                       'in_octets': '13523047597926',
                                                                       'in_ucast_pkts': '3850063227',
                                                                       'links': [       {       'in_nucast_pkts': '6913875666',
                                                                                                'in_octets': '13523047597926',
                                                                                                'in_ucast_pkts': '3850063227',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '585806984831',
                                                                                                'out_ucast_pkts': '138838003585',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '585806984831',
                                                                       'out_ucast_pkts': '138838003585',
                                                                       'poller_hash': 'MzIxMEQ4cGVlcmludGZwb24gNjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '66',
                                                                                               'links': [       {       'connected_time': '21886975',
                                                                                                                        'fiber_length': 14758,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMEQ4cGVlcnBvbnBvbiA2Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2268 PERTH LINE (JANTZI AARON & '
                                                 'MELISSA)',
                                         'poller_hash': 'MzIxMEQ4',
                                         'serial_number': '3210D8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:44:88',
                                         'uptime': '21886975'},
                                 {       'counter_source': '67/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '67/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '161023360506',
                                                                       'in_ucast_pkts': '821212491',
                                                                       'mac_address': '48:77:46:ce:a6:ce',
                                                                       'name': '67/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4552173113606',
                                                                       'out_ucast_pkts': '3412694003',
                                                                       'poller_hash': 'ODQ1OEFDcGVlcmludGY2Ny9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '67',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4552173113606',
                                                                       'in_ucast_pkts': '3412694003',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4552173113606',
                                                                                                'in_ucast_pkts': '3412694003',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '161023360506',
                                                                                                'out_ucast_pkts': '821212491',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '161023360506',
                                                                       'out_ucast_pkts': '821212491',
                                                                       'poller_hash': 'ODQ1OEFDcGVlcmludGZwb24gNjc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '67',
                                                                                               'links': [       {       'connected_time': '639438',
                                                                                                                        'fiber_length': 12520,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ1OEFDcGVlcnBvbnBvbiA2Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5525 Perth Rd. 119 (Heather Stever)',
                                         'poller_hash': 'ODQ1OEFD',
                                         'serial_number': '8458AC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:ce:a6:ce',
                                         'uptime': '639438'},
                                 {       'counter_source': '69/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:9c:25:5b:16:d8'],
                                                                       'id': '69/f1',
                                                                       'in_nucast_pkts': '12005806',
                                                                       'in_octets': '148617143181',
                                                                       'in_ucast_pkts': '9346343743',
                                                                       'mac_address': '00:06:31:11:ef:0a',
                                                                       'mtu': 1608,
                                                                       'name': '69/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11516805720',
                                                                       'out_octets': '18493231402649',
                                                                       'out_ucast_pkts': '5458903433216',
                                                                       'poller_hash': 'MTMwQTJwZWVyaW50ZjY5L2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '69',
                                                                       'in_nucast_pkts': '11516805720',
                                                                       'in_octets': '18493231402649',
                                                                       'in_ucast_pkts': '5458903433216',
                                                                       'links': [       {       'in_nucast_pkts': '11516805720',
                                                                                                'in_octets': '18493231402649',
                                                                                                'in_ucast_pkts': '5458903433216',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12005806',
                                                                                                'out_octets': '148617143181',
                                                                                                'out_ucast_pkts': '9346343743',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12005806',
                                                                       'out_octets': '148617143181',
                                                                       'out_ucast_pkts': '9346343743',
                                                                       'poller_hash': 'MTMwQTJwZWVyaW50ZnBvbiA2OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '69',
                                                                                               'links': [       {       'connected_time': '11964807',
                                                                                                                        'fiber_length': 13226,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwQTJwZWVycG9ucG9uIDY5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712',
                                         'name': '5451 Perth Rd. 119 B (Marvin Good)',
                                         'poller_hash': 'MTMwQTI=',
                                         'serial_number': '130A2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:11:ef:0a',
                                         'uptime': '11964807'},
                                 {       'counter_source': '70/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '70/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2066337506122',
                                                                       'in_ucast_pkts': '2660860141',
                                                                       'mac_address': 'ec:4f:82:08:1d:6d',
                                                                       'name': '70/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7350362775930',
                                                                       'out_ucast_pkts': '6072081251',
                                                                       'poller_hash': 'MUM1QkEzcGVlcmludGY3MC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '70',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7350362775930',
                                                                       'in_ucast_pkts': '6072081251',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7350362775930',
                                                                                                'in_ucast_pkts': '6072081251',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2066337506122',
                                                                                                'out_ucast_pkts': '2660860141',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2066337506122',
                                                                       'out_ucast_pkts': '2660860141',
                                                                       'poller_hash': 'MUM1QkEzcGVlcmludGZwb24gNzA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '70',
                                                                                               'links': [       {       'connected_time': '611454',
                                                                                                                        'fiber_length': 13539,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.206,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUM1QkEzcGVlcnBvbnBvbiA3MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '836GE',
                                         'name': '5432 PERTH ROAD 119 (MARTIN HARLAN & '
                                                 'MICHELLE)',
                                         'poller_hash': 'MUM1QkEz',
                                         'serial_number': '1C5BA3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:08:1d:6d',
                                         'uptime': '611454'},
                                 {       'counter_source': '71/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:f5:a2:7f:2b:32'],
                                                                       'id': '71/g1',
                                                                       'in_nucast_pkts': '30635',
                                                                       'in_octets': '7878366543',
                                                                       'in_ucast_pkts': '26421245',
                                                                       'mac_address': 'ec:4f:82:7d:bb:7f',
                                                                       'mtu': 2000,
                                                                       'name': '71/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394386',
                                                                       'out_octets': '87155509900',
                                                                       'out_ucast_pkts': '61604016',
                                                                       'poller_hash': 'MzAwNzI4cGVlcmludGY3MS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '71',
                                                                       'in_nucast_pkts': '394386',
                                                                       'in_octets': '87155509900',
                                                                       'in_ucast_pkts': '61604016',
                                                                       'links': [       {       'in_nucast_pkts': '394386',
                                                                                                'in_octets': '87155509900',
                                                                                                'in_ucast_pkts': '61604016',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '30635',
                                                                                                'out_octets': '7878366543',
                                                                                                'out_ucast_pkts': '26421245',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30635',
                                                                       'out_octets': '7878366543',
                                                                       'out_ucast_pkts': '26421245',
                                                                       'poller_hash': 'MzAwNzI4cGVlcmludGZwb24gNzE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '71',
                                                                                               'links': [       {       'connected_time': '335330',
                                                                                                                        'fiber_length': 17132,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwNzI4cGVlcnBvbnBvbiA3MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '2084 NithviewCt Nithview (Rola Scott '
                                                 '10434213)',
                                         'poller_hash': 'MzAwNzI4',
                                         'serial_number': '300728',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7d:bb:7f',
                                         'uptime': '335330'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '72/f1',
                                                                       'name': '72/f1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTMwQURwZWVyaW50ZjcyL2Yx'},
                                                               {       'admin_up': True,
                                                                       'id': '72',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTMwQURwZWVyaW50ZnBvbiA3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '72',
                                                                                               'links': [       {       'connected_time': '11965425',
                                                                                                                        'fiber_length': 12286,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwQURwZWVycG9ucG9uIDcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712',
                                         'name': '4017 Perth Line 56-1 (Kenneth Jantzi '
                                                 '10432211)',
                                         'poller_hash': 'MTMwQUQ=',
                                         'serial_number': '130AD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:11:ee:d7',
                                         'uptime': '11965425'},
                                 {       'counter_source': '73/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '9c:a2:f4:5f:23:41'],
                                                                       'id': '73/f1',
                                                                       'in_nucast_pkts': '7026',
                                                                       'in_octets': '6616666640',
                                                                       'in_ucast_pkts': '63105250',
                                                                       'mac_address': '00:06:31:08:4a:80',
                                                                       'mtu': 1608,
                                                                       'name': '73/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '524584',
                                                                       'out_octets': '271765576584',
                                                                       'out_ucast_pkts': '191372284',
                                                                       'poller_hash': 'RjMyOHBlZXJpbnRmNzMvZjE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '73',
                                                                       'in_nucast_pkts': '524584',
                                                                       'in_octets': '271765576584',
                                                                       'in_ucast_pkts': '191372284',
                                                                       'links': [       {       'in_nucast_pkts': '524584',
                                                                                                'in_octets': '271765576584',
                                                                                                'in_ucast_pkts': '191372284',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7026',
                                                                                                'out_octets': '6616666640',
                                                                                                'out_ucast_pkts': '63105250',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7026',
                                                                       'out_octets': '6616666640',
                                                                       'out_ucast_pkts': '63105250',
                                                                       'poller_hash': 'RjMyOHBlZXJpbnRmcG9uIDcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '73',
                                                                                               'links': [       {       'connected_time': '782079',
                                                                                                                        'fiber_length': 11621,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjMyOHBlZXJwb25wb24gNzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4002 PERTH ROAD 56 (WILLIAMSON '
                                                 'DOUGLAS & BETTY)',
                                         'poller_hash': 'RjMyOA==',
                                         'serial_number': 'F328',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:4a:80',
                                         'uptime': '782079'},
                                 {       'counter_source': '74/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '74/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '279604597640',
                                                                       'in_ucast_pkts': '1456097228',
                                                                       'mac_address': 'ec:4f:82:b9:1a:f8',
                                                                       'name': '74/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7428510553568',
                                                                       'out_ucast_pkts': '5598281490',
                                                                       'poller_hash': 'MkZFQTA5cGVlcmludGY3NC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '74',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7428510553568',
                                                                       'in_ucast_pkts': '5598281490',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7428510553568',
                                                                                                'in_ucast_pkts': '5598281490',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '279604597640',
                                                                                                'out_ucast_pkts': '1456097228',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '279604597640',
                                                                       'out_ucast_pkts': '1456097228',
                                                                       'poller_hash': 'MkZFQTA5cGVlcmludGZwb24gNzQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '74',
                                                                                               'links': [       {       'connected_time': '639070',
                                                                                                                        'fiber_length': 12264,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.56,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQTA5cGVlcnBvbnBvbiA3NA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4025 Perth Line 56 (Mark Nafziger)',
                                         'poller_hash': 'MkZFQTA5',
                                         'serial_number': '2FEA09',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:1a:f8',
                                         'uptime': '639070'},
                                 {       'counter_source': '76/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '76/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '554305950457',
                                                                       'in_ucast_pkts': '18589074820',
                                                                       'mac_address': 'ec:4f:82:dc:f1:a4',
                                                                       'name': '76/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11503261820844',
                                                                       'out_ucast_pkts': '26930480423',
                                                                       'poller_hash': 'MkVEOUNFcGVlcmludGY3Ni9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '76',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11503261820844',
                                                                       'in_ucast_pkts': '26930480423',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11503261820844',
                                                                                                'in_ucast_pkts': '26930480423',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '554305950457',
                                                                                                'out_ucast_pkts': '18589074820',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '554305950457',
                                                                       'out_ucast_pkts': '18589074820',
                                                                       'poller_hash': 'MkVEOUNFcGVlcmludGZwb24gNzY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '76',
                                                                                               'links': [       {       'connected_time': '638647',
                                                                                                                        'fiber_length': 12374,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVEOUNFcGVlcnBvbnBvbiA3Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5559 Perth Rd. 119 (Shane Stever '
                                                 '10420470)',
                                         'poller_hash': 'MkVEOUNF',
                                         'serial_number': '2ED9CE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:dc:f1:a4',
                                         'uptime': '638647'},
                                 {       'counter_source': '77/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '9c:05:d6:ab:d9:f7'],
                                                                       'id': '77/g1',
                                                                       'in_nucast_pkts': '96443',
                                                                       'in_octets': '228845764487',
                                                                       'in_ucast_pkts': '555905241',
                                                                       'mac_address': '04:bc:9f:36:53:38',
                                                                       'mtu': 2000,
                                                                       'name': '77/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22465286',
                                                                       'out_octets': '817618808412',
                                                                       'out_ucast_pkts': '770345515',
                                                                       'poller_hash': 'MTAwQTkzNHBlZXJpbnRmNzcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '77',
                                                                       'in_nucast_pkts': '22465286',
                                                                       'in_octets': '817618808412',
                                                                       'in_ucast_pkts': '770345515',
                                                                       'links': [       {       'in_nucast_pkts': '22465286',
                                                                                                'in_octets': '817618808412',
                                                                                                'in_ucast_pkts': '770345515',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '96443',
                                                                                                'out_octets': '228845764487',
                                                                                                'out_ucast_pkts': '555905241',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '96443',
                                                                       'out_octets': '228845764487',
                                                                       'out_ucast_pkts': '555905241',
                                                                       'poller_hash': 'MTAwQTkzNHBlZXJpbnRmcG9uIDc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '77',
                                                                                               'links': [       {       'connected_time': '552730',
                                                                                                                        'fiber_length': 11760,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.064,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTAwQTkzNHBlZXJwb25wb24gNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '4011 Perth Line 56 (Elgin Roth '
                                                 'Transport)',
                                         'poller_hash': 'MTAwQTkzNA==',
                                         'serial_number': '100A934',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '04:bc:9f:36:53:34',
                                         'uptime': '552730'},
                                 {       'counter_source': '78/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '78/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '310286584770',
                                                                       'in_ucast_pkts': '1323830753',
                                                                       'mac_address': '48:77:46:8c:2c:dc',
                                                                       'name': '78/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4600885394990',
                                                                       'out_ucast_pkts': '3717440743',
                                                                       'poller_hash': 'NzZCMDNEcGVlcmludGY3OC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '78',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4600885394990',
                                                                       'in_ucast_pkts': '3717440743',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '4600885394990',
                                                                                                'in_ucast_pkts': '3717440743',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '310286584770',
                                                                                                'out_ucast_pkts': '1323830753',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '310286584770',
                                                                       'out_ucast_pkts': '1323830753',
                                                                       'poller_hash': 'NzZCMDNEcGVlcmludGZwb24gNzg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '78',
                                                                                               'links': [       {       'connected_time': '639742',
                                                                                                                        'fiber_length': 14067,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzZCMDNEcGVlcnBvbnBvbiA3OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4009 Line 52 (Erik Vanaarsen)',
                                         'poller_hash': 'NzZCMDNE',
                                         'serial_number': '76B03D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:8c:2c:dc',
                                         'uptime': '639742'},
                                 {       'counter_source': '80/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd4:6a:91:b7:c6:72'],
                                                                       'id': '80/g1',
                                                                       'in_nucast_pkts': '32468',
                                                                       'in_octets': '899893667459',
                                                                       'in_ucast_pkts': '2867731036',
                                                                       'mac_address': '1c:8b:76:00:31:cd',
                                                                       'mtu': 2000,
                                                                       'name': '80/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17533309',
                                                                       'out_octets': '10298685359968',
                                                                       'out_ucast_pkts': '7536882212',
                                                                       'poller_hash': 'MTU4RkFBRnBlZXJpbnRmODAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '80',
                                                                       'in_nucast_pkts': '17533309',
                                                                       'in_octets': '10298685359968',
                                                                       'in_ucast_pkts': '7536882212',
                                                                       'links': [       {       'in_nucast_pkts': '17533309',
                                                                                                'in_octets': '10298685359968',
                                                                                                'in_ucast_pkts': '7536882212',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '32468',
                                                                                                'out_octets': '899893667459',
                                                                                                'out_ucast_pkts': '2867731036',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32468',
                                                                       'out_octets': '899893667459',
                                                                       'out_ucast_pkts': '2867731036',
                                                                       'poller_hash': 'MTU4RkFBRnBlZXJpbnRmcG9uIDgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '80',
                                                                                               'links': [       {       'connected_time': '27376190',
                                                                                                                        'fiber_length': 23652,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.046,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU4RkFBRnBlZXJwb25wb24gODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '3685 NAFZIGER RD (SCHMIDTSVILLE '
                                                 'RESTURANT)',
                                         'poller_hash': 'MTU4RkFBRg==',
                                         'serial_number': '158FAAF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '1c:8b:76:00:31:c9',
                                         'uptime': '27376190'},
                                 {       'counter_source': '82/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a8:87:6a'],
                                                                       'id': '82/g1',
                                                                       'in_nucast_pkts': '6729',
                                                                       'in_octets': '568275362103',
                                                                       'in_ucast_pkts': '957763640',
                                                                       'mac_address': 'ec:4f:82:5b:73:3e',
                                                                       'mtu': 2000,
                                                                       'name': '82/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58956450',
                                                                       'out_octets': '5145598528907',
                                                                       'out_ucast_pkts': '3729470169',
                                                                       'poller_hash': 'MkIwMURBcGVlcmludGY4Mi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '82',
                                                                       'in_nucast_pkts': '58956450',
                                                                       'in_octets': '5145598528907',
                                                                       'in_ucast_pkts': '3729470169',
                                                                       'links': [       {       'in_nucast_pkts': '58956450',
                                                                                                'in_octets': '5145598528907',
                                                                                                'in_ucast_pkts': '3729470169',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6729',
                                                                                                'out_octets': '568275362103',
                                                                                                'out_ucast_pkts': '957763640',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6729',
                                                                       'out_octets': '568275362103',
                                                                       'out_ucast_pkts': '957763640',
                                                                       'poller_hash': 'MkIwMURBcGVlcmludGZwb24gODI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '82',
                                                                                               'links': [       {       'connected_time': '21885968',
                                                                                                                        'fiber_length': 18203,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMURBcGVlcnBvbnBvbiA4Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2698 Line 47 (Sharon Schnarr)',
                                         'poller_hash': 'MkIwMURB',
                                         'serial_number': '2B01DA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:5b:73:3b',
                                         'uptime': '21885968'},
                                 {       'counter_source': '83/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:74:36'],
                                                                       'id': '83/g1',
                                                                       'in_nucast_pkts': '12648',
                                                                       'in_octets': '14628402724',
                                                                       'in_ucast_pkts': '71591372',
                                                                       'mac_address': 'cc:be:59:50:17:50',
                                                                       'mtu': 2000,
                                                                       'name': '83/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '409718',
                                                                       'out_octets': '224902052250',
                                                                       'out_ucast_pkts': '187710089',
                                                                       'poller_hash': 'NDNDMjI5cGVlcmludGY4My9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '83',
                                                                       'in_nucast_pkts': '409718',
                                                                       'in_octets': '224902052250',
                                                                       'in_ucast_pkts': '187710089',
                                                                       'links': [       {       'in_nucast_pkts': '409718',
                                                                                                'in_octets': '224902052250',
                                                                                                'in_ucast_pkts': '187710089',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12648',
                                                                                                'out_octets': '14628402724',
                                                                                                'out_ucast_pkts': '71591372',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12648',
                                                                       'out_octets': '14628402724',
                                                                       'out_ucast_pkts': '71591372',
                                                                       'poller_hash': 'NDNDMjI5cGVlcmludGZwb24gODM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '83',
                                                                                               'links': [       {       'connected_time': '611458',
                                                                                                                        'fiber_length': 17141,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.542,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.286}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNDMjI5cGVlcnBvbnBvbiA4Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5106 Perth Rd. 119 (Fallon Martin)',
                                         'poller_hash': 'NDNDMjI5',
                                         'serial_number': '43C229',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:17:4f',
                                         'uptime': '611458'},
                                 {       'counter_source': '85/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:b5:57'],
                                                                       'id': '85/g1',
                                                                       'in_nucast_pkts': '16132',
                                                                       'in_octets': '5217317501',
                                                                       'in_ucast_pkts': '13265011',
                                                                       'mac_address': 'ec:4f:82:da:16:38',
                                                                       'mtu': 2000,
                                                                       'name': '85/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '512745',
                                                                       'out_octets': '59066722423',
                                                                       'out_ucast_pkts': '48652320',
                                                                       'poller_hash': 'MkU3RTY5cGVlcmludGY4NS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '85',
                                                                       'in_nucast_pkts': '512745',
                                                                       'in_octets': '59066722423',
                                                                       'in_ucast_pkts': '48652320',
                                                                       'links': [       {       'in_nucast_pkts': '512745',
                                                                                                'in_octets': '59066722423',
                                                                                                'in_ucast_pkts': '48652320',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16132',
                                                                                                'out_octets': '5217317501',
                                                                                                'out_ucast_pkts': '13265011',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16132',
                                                                       'out_octets': '5217317501',
                                                                       'out_ucast_pkts': '13265011',
                                                                       'poller_hash': 'MkU3RTY5cGVlcmludGZwb24gODU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '85',
                                                                                               'links': [       {       'connected_time': '763757',
                                                                                                                        'fiber_length': 19928,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkU3RTY5cGVlcnBvbnBvbiA4NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2829 Line 45 Gadshill (Randy '
                                                 'Swartzentruber 10424337)',
                                         'poller_hash': 'MkU3RTY5',
                                         'serial_number': '2E7E69',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:da:16:34',
                                         'uptime': '763757'},
                                 {       'counter_source': '86/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '94:2a:6f:12:58:bc'],
                                                                       'id': '86/g1',
                                                                       'in_nucast_pkts': '5025707',
                                                                       'in_octets': '2882454264172',
                                                                       'in_ucast_pkts': '6559021383',
                                                                       'mac_address': '00:06:31:08:da:eb',
                                                                       'mtu': 1608,
                                                                       'name': '86/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22467087',
                                                                       'out_octets': '20093335431437',
                                                                       'out_ucast_pkts': '37371223866',
                                                                       'poller_hash': 'MTUzQ0JwZWVyaW50Zjg2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '86',
                                                                       'in_nucast_pkts': '22467087',
                                                                       'in_octets': '20093335431437',
                                                                       'in_ucast_pkts': '37371223866',
                                                                       'links': [       {       'in_nucast_pkts': '22467087',
                                                                                                'in_octets': '20093335431437',
                                                                                                'in_ucast_pkts': '37371223866',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5025707',
                                                                                                'out_octets': '2882454264172',
                                                                                                'out_ucast_pkts': '6559021383',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5025707',
                                                                       'out_octets': '2882454264172',
                                                                       'out_ucast_pkts': '6559021383',
                                                                       'poller_hash': 'MTUzQ0JwZWVyaW50ZnBvbiA4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '86',
                                                                                               'links': [       {       'connected_time': '11964828',
                                                                                                                        'fiber_length': 19792,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTUzQ0JwZWVycG9ucG9uIDg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '714G',
                                         'name': '2835 Line 45 (TMP International '
                                                 '10425548)',
                                         'poller_hash': 'MTUzQ0I=',
                                         'serial_number': '153CB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:da:eb',
                                         'uptime': '11964828'},
                                 {       'counter_source': '87/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '87/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '488286698603',
                                                                       'in_ucast_pkts': '2354089094',
                                                                       'mac_address': 'ec:4f:82:92:ff:80',
                                                                       'name': '87/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12907058635493',
                                                                       'out_ucast_pkts': '8920050595',
                                                                       'poller_hash': 'MzI5NjdDcGVlcmludGY4Ny9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '87',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12907058635493',
                                                                       'in_ucast_pkts': '8920050595',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '12907058635493',
                                                                                                'in_ucast_pkts': '8920050595',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '488286698603',
                                                                                                'out_ucast_pkts': '2354089094',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '488286698603',
                                                                       'out_ucast_pkts': '2354089094',
                                                                       'poller_hash': 'MzI5NjdDcGVlcmludGZwb24gODc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '87',
                                                                                               'links': [       {       'connected_time': '28849411',
                                                                                                                        'fiber_length': 19917,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.602,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzI5NjdDcGVlcnBvbnBvbiA4Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2837 Line 45 (Jamie Leach 10423704)',
                                         'poller_hash': 'MzI5NjdD',
                                         'serial_number': '32967C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:92:ff:80',
                                         'uptime': '28849411'},
                                 {       'counter_source': '88/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '88/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '703931399780',
                                                                       'in_ucast_pkts': '2176207907',
                                                                       'mac_address': 'cc:be:59:b1:12:4b',
                                                                       'name': '88/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8983668520292',
                                                                       'out_ucast_pkts': '2581477189',
                                                                       'poller_hash': 'NEU2NDM0cGVlcmludGY4OC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '88',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8983668520292',
                                                                       'in_ucast_pkts': '2581477189',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '8983668520292',
                                                                                                'in_ucast_pkts': '2581477189',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '703931399780',
                                                                                                'out_ucast_pkts': '2176207907',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '703931399780',
                                                                       'out_ucast_pkts': '2176207907',
                                                                       'poller_hash': 'NEU2NDM0cGVlcmludGZwb24gODg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '88',
                                                                                               'links': [       {       'connected_time': '611421',
                                                                                                                        'fiber_length': 19985,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEU2NDM0cGVlcnBvbnBvbiA4OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2845 Line 45 (John Bax 10427651)',
                                         'poller_hash': 'NEU2NDM0',
                                         'serial_number': '4E6434',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:b1:12:4b',
                                         'uptime': '611421'},
                                 {       'counter_source': '89/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '70:a7:41:a6:7e:b1'],
                                                                       'id': '89/g1',
                                                                       'in_nucast_pkts': '119847',
                                                                       'in_octets': '3285662554339',
                                                                       'in_ucast_pkts': '13290553447',
                                                                       'mac_address': 'ec:4f:82:b9:1a:5c',
                                                                       'mtu': 2000,
                                                                       'name': '89/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22461463',
                                                                       'out_octets': '42272320425183',
                                                                       'out_ucast_pkts': '28699009384',
                                                                       'poller_hash': 'MkZFOUY1cGVlcmludGY4OS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '89',
                                                                       'in_nucast_pkts': '22461463',
                                                                       'in_octets': '42272320425183',
                                                                       'in_ucast_pkts': '28699009384',
                                                                       'links': [       {       'in_nucast_pkts': '22461463',
                                                                                                'in_octets': '42272320425183',
                                                                                                'in_ucast_pkts': '28699009384',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '119847',
                                                                                                'out_octets': '3285662554339',
                                                                                                'out_ucast_pkts': '13290553447',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '119847',
                                                                       'out_octets': '3285662554339',
                                                                       'out_ucast_pkts': '13290553447',
                                                                       'poller_hash': 'MkZFOUY1cGVlcmludGZwb24gODk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '89',
                                                                                               'links': [       {       'connected_time': '552443',
                                                                                                                        'fiber_length': 19961,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.89,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFOUY1cGVlcnBvbnBvbiA4OQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2828 Line 45 (Paul&Linda Bogle '
                                                 '10433906)',
                                         'poller_hash': 'MkZFOUY1',
                                         'serial_number': '2FE9F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:1a:58',
                                         'uptime': '552443'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '90/g1',
                                                                       'name': '90/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUQwNzBDcGVlcmludGY5MC9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '90',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MUQwNzBDcGVlcmludGZwb24gOTA=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '90',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNzBDcGVlcnBvbnBvbiA5MA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '9 ST MARYS ST S (ROCK DOUG & KATHY)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'OLT PON port 7/7 is '
                                                                  'down',
                                         'offline_time': '2024-10-17T07:03:42.740Z',
                                         'poller_hash': 'MUQwNzBD',
                                         'serial_number': '1D070C',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '91/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1e:fb'],
                                                                       'id': '91/g1',
                                                                       'in_nucast_pkts': '772628',
                                                                       'in_octets': '256013793522',
                                                                       'in_ucast_pkts': '1201570648',
                                                                       'mac_address': '00:06:31:4c:0c:f4',
                                                                       'mtu': 2000,
                                                                       'name': '91/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22463981',
                                                                       'out_octets': '5048842042384',
                                                                       'out_ucast_pkts': '3902447418',
                                                                       'poller_hash': 'ODZDMDdwZWVyaW50ZjkxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '91',
                                                                       'in_nucast_pkts': '22463981',
                                                                       'in_octets': '5048842042384',
                                                                       'in_ucast_pkts': '3902447418',
                                                                       'links': [       {       'in_nucast_pkts': '22463981',
                                                                                                'in_octets': '5048842042384',
                                                                                                'in_ucast_pkts': '3902447418',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '772628',
                                                                                                'out_octets': '256013793522',
                                                                                                'out_ucast_pkts': '1201570648',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '772628',
                                                                       'out_octets': '256013793522',
                                                                       'out_ucast_pkts': '1201570648',
                                                                       'poller_hash': 'ODZDMDdwZWVyaW50ZnBvbiA5MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '91',
                                                                                               'links': [       {       'connected_time': '552420',
                                                                                                                        'fiber_length': 20035,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODZDMDdwZWVycG9ucG9uIDkx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2844 Line 45 (Gordon Hill)',
                                         'poller_hash': 'ODZDMDc=',
                                         'serial_number': '86C07',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:3c:cc:ef',
                                         'uptime': '552420'},
                                 {       'counter_source': '92/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '92/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '304734107530',
                                                                       'in_ucast_pkts': '1376282561',
                                                                       'mac_address': '60:db:98:72:f5:c7',
                                                                       'name': '92/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '53',
                                                                       'out_octets': '4355107613306',
                                                                       'out_ucast_pkts': '3284430219',
                                                                       'poller_hash': 'QjA4RUQ5cGVlcmludGY5Mi9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '92',
                                                                       'in_nucast_pkts': '53',
                                                                       'in_octets': '4355107613306',
                                                                       'in_ucast_pkts': '3284430219',
                                                                       'links': [       {       'in_nucast_pkts': '53',
                                                                                                'in_octets': '4355107613306',
                                                                                                'in_ucast_pkts': '3284430219',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '304734107530',
                                                                                                'out_ucast_pkts': '1376282561',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '304734107530',
                                                                       'out_ucast_pkts': '1376282561',
                                                                       'poller_hash': 'QjA4RUQ5cGVlcmludGZwb24gOTI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '92',
                                                                                               'links': [       {       'connected_time': '611443',
                                                                                                                        'fiber_length': 19832,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjA4RUQ5cGVlcnBvbnBvbiA5Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2847 Line 45 (Sean Bell)',
                                         'poller_hash': 'QjA4RUQ5',
                                         'serial_number': 'B08ED9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:72:f5:c7',
                                         'uptime': '611443'},
                                 {       'counter_source': '95/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '95/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '500108880421',
                                                                       'in_ucast_pkts': '4134327676',
                                                                       'mac_address': 'ec:4f:82:dc:ed:ec',
                                                                       'name': '95/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '21131600484932',
                                                                       'out_ucast_pkts': '14546673917',
                                                                       'poller_hash': 'MkVEOTU3cGVlcmludGY5NS9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '95',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '21131600484932',
                                                                       'in_ucast_pkts': '14546673917',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '21131600484932',
                                                                                                'in_ucast_pkts': '14546673917',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '500108880421',
                                                                                                'out_ucast_pkts': '4134327676',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '500108880421',
                                                                       'out_ucast_pkts': '4134327676',
                                                                       'poller_hash': 'MkVEOTU3cGVlcmludGZwb24gOTU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '95',
                                                                                               'links': [       {       'connected_time': '639398',
                                                                                                                        'fiber_length': 19745,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.456,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVEOTU3cGVlcnBvbnBvbiA5NQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2857 Line 45 (Hayward Martin)',
                                         'poller_hash': 'MkVEOTU3',
                                         'serial_number': '2ED957',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:dc:ed:ec',
                                         'uptime': '639398'},
                                 {       'counter_source': '97/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b6:5c'],
                                                                       'id': '97/g1',
                                                                       'in_nucast_pkts': '363',
                                                                       'in_octets': '796664345340',
                                                                       'in_ucast_pkts': '1525998582',
                                                                       'mac_address': 'ec:4f:82:db:ef:20',
                                                                       'mtu': 2000,
                                                                       'name': '97/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22467466',
                                                                       'out_octets': '4519949468026',
                                                                       'out_ucast_pkts': '3513563447',
                                                                       'poller_hash': 'MkVCOTdFcGVlcmludGY5Ny9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '97',
                                                                       'in_nucast_pkts': '22467466',
                                                                       'in_octets': '4519949468026',
                                                                       'in_ucast_pkts': '3513563447',
                                                                       'links': [       {       'in_nucast_pkts': '22467466',
                                                                                                'in_octets': '4519949468026',
                                                                                                'in_ucast_pkts': '3513563447',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '363',
                                                                                                'out_octets': '796664345340',
                                                                                                'out_ucast_pkts': '1525998582',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '363',
                                                                       'out_octets': '796664345340',
                                                                       'out_ucast_pkts': '1525998582',
                                                                       'poller_hash': 'MkVCOTdFcGVlcmludGZwb24gOTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '97',
                                                                                               'links': [       {       'connected_time': '552210',
                                                                                                                        'fiber_length': 19803,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.608,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVCOTdFcGVlcnBvbnBvbiA5Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2863 Line 45 (Greg Ganselves)',
                                         'poller_hash': 'MkVCOTdF',
                                         'serial_number': '2EB97E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:db:ef:1c',
                                         'uptime': '552210'},
                                 {       'counter_source': '98/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '98/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '141651621819',
                                                                       'in_ucast_pkts': '785664984',
                                                                       'mac_address': 'cc:be:59:70:c9:1a',
                                                                       'name': '98/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3722903240236',
                                                                       'out_ucast_pkts': '2681635815',
                                                                       'poller_hash': 'NTYxNTE1cGVlcmludGY5OC9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '98',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3722903240236',
                                                                       'in_ucast_pkts': '2681635815',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3722903240236',
                                                                                                'in_ucast_pkts': '2681635815',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '141651621819',
                                                                                                'out_ucast_pkts': '785664984',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '141651621819',
                                                                       'out_ucast_pkts': '785664984',
                                                                       'poller_hash': 'NTYxNTE1cGVlcmludGZwb24gOTg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '98',
                                                                                               'links': [       {       'connected_time': '639287',
                                                                                                                        'fiber_length': 19554,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.788,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYxNTE1cGVlcnBvbnBvbiA5OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2867 LINE 45 (WEBER PAYTON)',
                                         'poller_hash': 'NTYxNTE1',
                                         'serial_number': '561515',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:70:c9:1a',
                                         'uptime': '639287'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '99/f1',
                                                                       'name': '99/f1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'RjEwMHBlZXJpbnRmOTkvZjE='},
                                                               {       'admin_up': True,
                                                                       'id': '99',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'RjEwMHBlZXJpbnRmcG9uIDk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '99',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjEwMHBlZXJwb25wb24gOTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4903 Perth Rd. 119',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "99" (location: '
                                                                  'GPON port "5/1"), '
                                                                  '"ONT went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2025-04-30T19:25:37.850Z',
                                         'poller_hash': 'RjEwMA==',
                                         'serial_number': 'F100',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '100/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:47:2f'],
                                                                       'id': '100/g1',
                                                                       'in_nucast_pkts': '32532242',
                                                                       'in_octets': '518154680935',
                                                                       'in_ucast_pkts': '1595454894',
                                                                       'mac_address': 'cc:be:59:cb:ac:a5',
                                                                       'mtu': 2000,
                                                                       'name': '100/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67134885314',
                                                                       'out_octets': '97367048211337',
                                                                       'out_ucast_pkts': '5284000330',
                                                                       'poller_hash': 'NDJDNkVDcGVlcmludGYxMDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '100',
                                                                       'in_nucast_pkts': '67134885314',
                                                                       'in_octets': '97367048211337',
                                                                       'in_ucast_pkts': '5284000330',
                                                                       'links': [       {       'in_nucast_pkts': '67134885314',
                                                                                                'in_octets': '97367048211337',
                                                                                                'in_ucast_pkts': '5284000330',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '32532242',
                                                                                                'out_octets': '518154680935',
                                                                                                'out_ucast_pkts': '1595454894',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32532242',
                                                                       'out_octets': '518154680935',
                                                                       'out_ucast_pkts': '1595454894',
                                                                       'poller_hash': 'NDJDNkVDcGVlcmludGZwb24gMTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '100',
                                                                                               'links': [       {       'connected_time': '21887230',
                                                                                                                        'fiber_length': 23018,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkVDcGVlcnBvbnBvbiAxMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 Evelyn Dr Wellesley (Nancy Cook '
                                                 '10435554)',
                                         'poller_hash': 'NDJDNkVD',
                                         'serial_number': '42C6EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ac:a2',
                                         'uptime': '21887230'},
                                 {       'counter_source': '104/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '104/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '183220988886',
                                                                       'in_ucast_pkts': '1360407416',
                                                                       'mac_address': '60:db:98:4e:91:c4',
                                                                       'name': '104/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '144',
                                                                       'out_octets': '6834575735145',
                                                                       'out_ucast_pkts': '4670534883',
                                                                       'poller_hash': 'QTlDN0Y4cGVlcmludGYxMDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '104',
                                                                       'in_nucast_pkts': '144',
                                                                       'in_octets': '6834575735145',
                                                                       'in_ucast_pkts': '4670534883',
                                                                       'links': [       {       'in_nucast_pkts': '144',
                                                                                                'in_octets': '6834575735145',
                                                                                                'in_ucast_pkts': '4670534883',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '183220988886',
                                                                                                'out_ucast_pkts': '1360407416',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '183220988886',
                                                                       'out_ucast_pkts': '1360407416',
                                                                       'poller_hash': 'QTlDN0Y4cGVlcmludGZwb24gMTA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '104',
                                                                                               'links': [       {       'connected_time': '67816077',
                                                                                                                        'fiber_length': 19856,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDN0Y4cGVlcnBvbnBvbiAxMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2858 Line 45 (Mary VanStaveren)',
                                         'poller_hash': 'QTlDN0Y4',
                                         'serial_number': 'A9C7F8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:91:c4',
                                         'uptime': '67816077'},
                                 {       'counter_source': '105/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:76:1c:7a:84:d8'],
                                                                       'id': '105/g1',
                                                                       'in_nucast_pkts': '396449',
                                                                       'in_octets': '14189285436',
                                                                       'in_ucast_pkts': '50409698',
                                                                       'mac_address': '00:06:31:1f:08:e5',
                                                                       'mtu': 1608,
                                                                       'name': '105/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10719342',
                                                                       'out_octets': '143182948410',
                                                                       'out_ucast_pkts': '13005809172',
                                                                       'poller_hash': 'NDE1NDhwZWVyaW50ZjEwNS9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '105',
                                                                       'in_nucast_pkts': '10719342',
                                                                       'in_octets': '143182948410',
                                                                       'in_ucast_pkts': '13005809172',
                                                                       'links': [       {       'in_nucast_pkts': '10719342',
                                                                                                'in_octets': '143182948410',
                                                                                                'in_ucast_pkts': '13005809172',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '396449',
                                                                                                'out_octets': '14189285436',
                                                                                                'out_ucast_pkts': '50409698',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '396449',
                                                                       'out_octets': '14189285436',
                                                                       'out_ucast_pkts': '50409698',
                                                                       'poller_hash': 'NDE1NDhwZWVyaW50ZnBvbiAxMDU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '105',
                                                                                               'links': [       {       'connected_time': '16338099',
                                                                                                                        'fiber_length': 19821,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDE1NDhwZWVycG9ucG9uIDEwNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712GX',
                                         'name': '2860 Line 45 (Robert St. Pierre)',
                                         'poller_hash': 'NDE1NDg=',
                                         'serial_number': '41548',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:1f:08:e5',
                                         'uptime': '16338099'},
                                 {       'counter_source': '107/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '107/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '281575181664',
                                                                       'in_ucast_pkts': '1074234314',
                                                                       'mac_address': 'cc:be:59:ef:f7:6e',
                                                                       'name': '107/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6055451795990',
                                                                       'out_ucast_pkts': '167983691',
                                                                       'poller_hash': 'NEQ0MjQwcGVlcmludGYxMDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '107',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6055451795990',
                                                                       'in_ucast_pkts': '167983691',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6055451795990',
                                                                                                'in_ucast_pkts': '167983691',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '281575181664',
                                                                                                'out_ucast_pkts': '1074234314',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '281575181664',
                                                                       'out_ucast_pkts': '1074234314',
                                                                       'poller_hash': 'NEQ0MjQwcGVlcmludGZwb24gMTA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '107',
                                                                                               'links': [       {       'connected_time': '10841403',
                                                                                                                        'fiber_length': 19860,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.43,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjQwcGVlcnBvbnBvbiAxMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2868 Line 45 (Simon Kuepfer)',
                                         'poller_hash': 'NEQ0MjQw',
                                         'serial_number': '4D4240',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:6e',
                                         'uptime': '10841403'},
                                 {       'counter_source': '108/g2+108/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:80:18'],
                                                                       'id': '108/g2',
                                                                       'in_nucast_pkts': '1462385',
                                                                       'in_octets': '15677848648',
                                                                       'in_ucast_pkts': '195910641',
                                                                       'mac_address': 'cc:be:59:ef:f8:8b',
                                                                       'mtu': 2000,
                                                                       'name': '108/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '69288800',
                                                                       'out_octets': '894948274536',
                                                                       'out_ucast_pkts': '617835860',
                                                                       'poller_hash': 'NEQ0MjYzcGVlcmludGYxMDgvZzI=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:80:34'],
                                                                       'id': '108/g3',
                                                                       'in_nucast_pkts': '1397036',
                                                                       'in_octets': '1345076196',
                                                                       'in_ucast_pkts': '5181846',
                                                                       'mac_address': 'cc:be:59:ef:f8:8c',
                                                                       'mtu': 2000,
                                                                       'name': '108/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68511084',
                                                                       'out_octets': '46137664249',
                                                                       'out_ucast_pkts': '14374167',
                                                                       'poller_hash': 'NEQ0MjYzcGVlcmludGYxMDgvZzM=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '108',
                                                                       'in_nucast_pkts': '137799884',
                                                                       'in_octets': '941085938785',
                                                                       'in_ucast_pkts': '632210027',
                                                                       'links': [       {       'in_nucast_pkts': '137799884',
                                                                                                'in_octets': '941085938785',
                                                                                                'in_ucast_pkts': '632210027',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2859421',
                                                                                                'out_octets': '17022924844',
                                                                                                'out_ucast_pkts': '201092487',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2859421',
                                                                       'out_octets': '17022924844',
                                                                       'out_ucast_pkts': '201092487',
                                                                       'poller_hash': 'NEQ0MjYzcGVlcmludGZwb24gMTA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '108',
                                                                                               'links': [       {       'connected_time': '3210544',
                                                                                                                        'fiber_length': 19925,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.61,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjYzcGVlcnBvbnBvbiAxMDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2870 LINE 45 (HARNOCK PATRICIA)',
                                         'poller_hash': 'NEQ0MjYz',
                                         'serial_number': '4D4263',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f8:86',
                                         'uptime': '3210544'},
                                 {       'counter_source': '109/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e9:f3'],
                                                                       'id': '109/g1',
                                                                       'in_nucast_pkts': '7306',
                                                                       'in_octets': '888275457',
                                                                       'in_ucast_pkts': '2378254',
                                                                       'mac_address': '84:d3:43:95:8a:35',
                                                                       'mtu': 2000,
                                                                       'name': '109/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394166',
                                                                       'out_octets': '28942931232',
                                                                       'out_ucast_pkts': '21781686',
                                                                       'poller_hash': 'Q0JEMTY1cGVlcmludGYxMDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '109',
                                                                       'in_nucast_pkts': '394166',
                                                                       'in_octets': '28942931232',
                                                                       'in_ucast_pkts': '21781686',
                                                                       'links': [       {       'in_nucast_pkts': '394166',
                                                                                                'in_octets': '28942931232',
                                                                                                'in_ucast_pkts': '21781686',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7306',
                                                                                                'out_octets': '888275457',
                                                                                                'out_ucast_pkts': '2378254',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7306',
                                                                       'out_octets': '888275457',
                                                                       'out_ucast_pkts': '2378254',
                                                                       'poller_hash': 'Q0JEMTY1cGVlcmludGZwb24gMTA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '109',
                                                                                               'links': [       {       'connected_time': '335343',
                                                                                                                        'fiber_length': 18777,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTY1cGVlcnBvbnBvbiAxMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5324A Road 106 (LICHTI JASON)',
                                         'poller_hash': 'Q0JEMTY1',
                                         'serial_number': 'CBD165',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8a:32',
                                         'uptime': '335343'},
                                 {       'counter_source': '110/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '110/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '288861334540',
                                                                       'in_ucast_pkts': '2040680227',
                                                                       'mac_address': 'cc:be:59:b7:3b:3f',
                                                                       'name': '110/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3827217515639',
                                                                       'out_ucast_pkts': '747869153',
                                                                       'poller_hash': 'NEQxMTFBcGVlcmludGYxMTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '110',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3827217515639',
                                                                       'in_ucast_pkts': '747869153',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3827217515639',
                                                                                                'in_ucast_pkts': '747869153',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '288861334540',
                                                                                                'out_ucast_pkts': '2040680227',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '288861334540',
                                                                       'out_ucast_pkts': '2040680227',
                                                                       'poller_hash': 'NEQxMTFBcGVlcmludGZwb24gMTEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '110',
                                                                                               'links': [       {       'connected_time': '639422',
                                                                                                                        'fiber_length': 20176,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.684,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQxMTFBcGVlcnBvbnBvbiAxMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4893 LINHAVEN DR (SCHADE JOHN & '
                                                 'BRENDA)',
                                         'poller_hash': 'NEQxMTFB',
                                         'serial_number': '4D111A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:b7:3b:3f',
                                         'uptime': '639422'},
                                 {       'counter_source': '111/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '111/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '317901599333',
                                                                       'in_ucast_pkts': '1348604649',
                                                                       'mac_address': 'ec:4f:82:b9:22:70',
                                                                       'name': '111/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7296536629219',
                                                                       'out_ucast_pkts': '5213425229',
                                                                       'poller_hash': 'MkZFQUY4cGVlcmludGYxMTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '111',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7296536629219',
                                                                       'in_ucast_pkts': '5213425229',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7296536629219',
                                                                                                'in_ucast_pkts': '5213425229',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '317901599333',
                                                                                                'out_ucast_pkts': '1348604649',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '317901599333',
                                                                       'out_ucast_pkts': '1348604649',
                                                                       'poller_hash': 'MkZFQUY4cGVlcmludGZwb24gMTEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '111',
                                                                                               'links': [       {       'connected_time': '639723',
                                                                                                                        'fiber_length': 19680,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.314,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkZFQUY4cGVlcnBvbnBvbiAxMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4924 Linhaven Dr. Gadshill (Adam & '
                                                 'Jenny Riehl 10431415)',
                                         'poller_hash': 'MkZFQUY4',
                                         'serial_number': '2FEAF8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:b9:22:70',
                                         'uptime': '639723'},
                                 {       'counter_source': '112/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:47:fc'],
                                                                       'id': '112/f1',
                                                                       'in_nucast_pkts': '31848972',
                                                                       'in_octets': '704352624574',
                                                                       'in_ucast_pkts': '1540235433',
                                                                       'mac_address': '00:06:31:08:49:e7',
                                                                       'mtu': 1608,
                                                                       'name': '112/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7998104185',
                                                                       'out_octets': '16188451425278',
                                                                       'out_ucast_pkts': '20285512084337',
                                                                       'poller_hash': 'RjBGRnBlZXJpbnRmMTEyL2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '112',
                                                                       'in_nucast_pkts': '7998104185',
                                                                       'in_octets': '16188451425278',
                                                                       'in_ucast_pkts': '20285512084337',
                                                                       'links': [       {       'in_nucast_pkts': '7998104185',
                                                                                                'in_octets': '16188451425278',
                                                                                                'in_ucast_pkts': '20285512084337',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '31848972',
                                                                                                'out_octets': '704352624574',
                                                                                                'out_ucast_pkts': '1540235433',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31848972',
                                                                       'out_octets': '704352624574',
                                                                       'out_ucast_pkts': '1540235433',
                                                                       'poller_hash': 'RjBGRnBlZXJpbnRmcG9uIDExMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '112',
                                                                                               'links': [       {       'connected_time': '11965356',
                                                                                                                        'fiber_length': 19513,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjBGRnBlZXJwb25wb24gMTEy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4917 Linhaven Dr. (Shawn Gilhula)',
                                         'poller_hash': 'RjBGRg==',
                                         'serial_number': 'F0FF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:49:e7',
                                         'uptime': '11965356'},
                                 {       'counter_source': '113/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:94:90'],
                                                                       'id': '113/g1',
                                                                       'in_nucast_pkts': '1041257',
                                                                       'in_octets': '230273284542',
                                                                       'in_ucast_pkts': '719725631',
                                                                       'mac_address': '00:06:31:e3:fd:17',
                                                                       'mtu': 2000,
                                                                       'name': '113/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4384041',
                                                                       'out_octets': '3184777872853',
                                                                       'out_ucast_pkts': '2467726237',
                                                                       'poller_hash': 'MTk5QTNCcGVlcmludGYxMTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '113',
                                                                       'in_nucast_pkts': '4384041',
                                                                       'in_octets': '3184777872853',
                                                                       'in_ucast_pkts': '2467726237',
                                                                       'links': [       {       'in_nucast_pkts': '4384041',
                                                                                                'in_octets': '3184777872853',
                                                                                                'in_ucast_pkts': '2467726237',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1041257',
                                                                                                'out_octets': '230273284542',
                                                                                                'out_ucast_pkts': '719725631',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1041257',
                                                                       'out_octets': '230273284542',
                                                                       'out_ucast_pkts': '719725631',
                                                                       'poller_hash': 'MTk5QTNCcGVlcmludGZwb24gMTEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '113',
                                                                                               'links': [       {       'connected_time': '6640172',
                                                                                                                        'fiber_length': 20268,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.488,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTk5QTNCcGVlcnBvbnBvbiAxMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4984 Perth Rd. 119 Gadshill (Daniel '
                                                 'Kuepfer 10435144)',
                                         'poller_hash': 'MTk5QTNC',
                                         'serial_number': '199A3B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:e3:fd:13',
                                         'uptime': '6640172'},
                                 {       'counter_source': '114/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8b:22'],
                                                                       'id': '114/g1',
                                                                       'in_nucast_pkts': '333078',
                                                                       'in_octets': '220014967875',
                                                                       'in_ucast_pkts': '402008295',
                                                                       'mac_address': '00:06:31:e3:fa:e7',
                                                                       'mtu': 2000,
                                                                       'name': '114/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12074348',
                                                                       'out_octets': '1847074398113',
                                                                       'out_ucast_pkts': '1461736416',
                                                                       'poller_hash': 'MTk5OUY1cGVlcmludGYxMTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '114',
                                                                       'in_nucast_pkts': '12074348',
                                                                       'in_octets': '1847074398113',
                                                                       'in_ucast_pkts': '1461736416',
                                                                       'links': [       {       'in_nucast_pkts': '12074348',
                                                                                                'in_octets': '1847074398113',
                                                                                                'in_ucast_pkts': '1461736416',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '333078',
                                                                                                'out_octets': '220014967875',
                                                                                                'out_ucast_pkts': '402008295',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '333078',
                                                                       'out_octets': '220014967875',
                                                                       'out_ucast_pkts': '402008295',
                                                                       'poller_hash': 'MTk5OUY1cGVlcmludGZwb24gMTE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '114',
                                                                                               'links': [       {       'connected_time': '552423',
                                                                                                                        'fiber_length': 20055,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.848,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTk5OUY1cGVlcnBvbnBvbiAxMTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4959 PERTH ROAD 119 (FERNADES '
                                                 'ROMUALD)',
                                         'poller_hash': 'MTk5OUY1',
                                         'serial_number': '1999F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:e3:fa:e3',
                                         'uptime': '552423'},
                                 {       'counter_source': '115/g1+115/g2+115/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:64:6c'],
                                                                       'id': '115/g1',
                                                                       'in_nucast_pkts': '22785856',
                                                                       'in_octets': '13707642767',
                                                                       'in_ucast_pkts': '74626178',
                                                                       'mac_address': 'cc:be:59:0c:b2:b8',
                                                                       'mtu': 2000,
                                                                       'name': '115/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31072894363',
                                                                       'out_octets': '42079289837076',
                                                                       'out_ucast_pkts': '370872215',
                                                                       'poller_hash': 'M0FERDBFcGVlcmludGYxMTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f2:ca'],
                                                                       'id': '115/g2',
                                                                       'in_nucast_pkts': '822673',
                                                                       'in_octets': '171898654746',
                                                                       'in_ucast_pkts': '1724940305',
                                                                       'mac_address': 'cc:be:59:0c:b2:b9',
                                                                       'mtu': 2000,
                                                                       'name': '115/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24130195',
                                                                       'out_octets': '5266345587835',
                                                                       'out_ucast_pkts': '3627428029',
                                                                       'poller_hash': 'M0FERDBFcGVlcmludGYxMTUvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3a:44'],
                                                                       'id': '115/g3',
                                                                       'in_nucast_pkts': '25593794',
                                                                       'in_octets': '136039177399',
                                                                       'in_ucast_pkts': '773833234',
                                                                       'mac_address': 'cc:be:59:0c:b2:ba',
                                                                       'mtu': 2000,
                                                                       'name': '115/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13791508635',
                                                                       'out_octets': '21941038471022',
                                                                       'out_ucast_pkts': '2789616839',
                                                                       'poller_hash': 'M0FERDBFcGVlcmludGYxMTUvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '115',
                                                                       'in_nucast_pkts': '44888533193',
                                                                       'in_octets': '69286673895933',
                                                                       'in_ucast_pkts': '6787917083',
                                                                       'links': [       {       'in_nucast_pkts': '44888533193',
                                                                                                'in_octets': '69286673895933',
                                                                                                'in_ucast_pkts': '6787917083',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '49202323',
                                                                                                'out_octets': '321645474912',
                                                                                                'out_ucast_pkts': '2573399717',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49202323',
                                                                       'out_octets': '321645474912',
                                                                       'out_ucast_pkts': '2573399717',
                                                                       'poller_hash': 'M0FERDBFcGVlcmludGZwb24gMTE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '115',
                                                                                               'links': [       {       'connected_time': '552134',
                                                                                                                        'fiber_length': 23075,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.724,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0FERDBFcGVlcnBvbnBvbiAxMTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': 'Pond View MDU',
                                         'poller_hash': 'M0FERDBF',
                                         'serial_number': '3ADD0E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:0c:b2:b4',
                                         'uptime': '552134'},
                                 {       'counter_source': '118/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:12:f2:d4'],
                                                                       'id': '118/f1',
                                                                       'in_nucast_pkts': '326',
                                                                       'in_octets': '649413295360',
                                                                       'in_ucast_pkts': '2655976580',
                                                                       'mac_address': '00:06:31:08:4d:71',
                                                                       'mtu': 1608,
                                                                       'name': '118/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22489474',
                                                                       'out_octets': '11602564460277',
                                                                       'out_ucast_pkts': '19984268806507',
                                                                       'poller_hash': 'RjQzQnBlZXJpbnRmMTE4L2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '118',
                                                                       'in_nucast_pkts': '22489474',
                                                                       'in_octets': '11602564460277',
                                                                       'in_ucast_pkts': '19984268806507',
                                                                       'links': [       {       'in_nucast_pkts': '22489474',
                                                                                                'in_octets': '11602564460277',
                                                                                                'in_ucast_pkts': '19984268806507',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '326',
                                                                                                'out_octets': '649413295360',
                                                                                                'out_ucast_pkts': '2655976580',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '326',
                                                                       'out_octets': '649413295360',
                                                                       'out_ucast_pkts': '2655976580',
                                                                       'poller_hash': 'RjQzQnBlZXJpbnRmcG9uIDExOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '118',
                                                                                               'links': [       {       'connected_time': '11965741',
                                                                                                                        'fiber_length': 19904,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjQzQnBlZXJwb25wb24gMTE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4888 Linehaven Dr. (J. Bolhuis '
                                                 '10423884)',
                                         'poller_hash': 'RjQzQg==',
                                         'serial_number': 'F43B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:4d:71',
                                         'uptime': '11965741'},
                                 {       'counter_source': '119/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:13:26:16'],
                                                                       'id': '119/g1',
                                                                       'in_nucast_pkts': '245',
                                                                       'in_octets': '719689433471',
                                                                       'in_ucast_pkts': '5671585201',
                                                                       'mac_address': 'cc:be:59:6d:c9:32',
                                                                       'mtu': 2000,
                                                                       'name': '119/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22482292',
                                                                       'out_octets': '18713806804077',
                                                                       'out_ucast_pkts': '12874460983',
                                                                       'poller_hash': 'NDAxMTUycGVlcmludGYxMTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '119',
                                                                       'in_nucast_pkts': '22482292',
                                                                       'in_octets': '18713806804077',
                                                                       'in_ucast_pkts': '12874460983',
                                                                       'links': [       {       'in_nucast_pkts': '22482292',
                                                                                                'in_octets': '18713806804077',
                                                                                                'in_ucast_pkts': '12874460983',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '245',
                                                                                                'out_octets': '719689433471',
                                                                                                'out_ucast_pkts': '5671585201',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '245',
                                                                       'out_octets': '719689433471',
                                                                       'out_ucast_pkts': '5671585201',
                                                                       'poller_hash': 'NDAxMTUycGVlcmludGZwb24gMTE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '119',
                                                                                               'links': [       {       'connected_time': '552208',
                                                                                                                        'fiber_length': 20202,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDAxMTUycGVlcnBvbnBvbiAxMTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4887 Tiffany Crt. (Jonathan Spiers '
                                                 '10423980)',
                                         'poller_hash': 'NDAxMTUy',
                                         'serial_number': '401152',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6d:c9:2e',
                                         'uptime': '552208'},
                                 {       'counter_source': '120/g3',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:82:01:dc'],
                                                                       'id': '120/g3',
                                                                       'in_nucast_pkts': '10954',
                                                                       'in_octets': '3386317179',
                                                                       'in_ucast_pkts': '30440142',
                                                                       'mac_address': '00:06:31:e8:f3:b2',
                                                                       'mtu': 2000,
                                                                       'name': '120/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715302',
                                                                       'out_octets': '169988869151',
                                                                       'out_ucast_pkts': '119230868',
                                                                       'poller_hash': 'MUQwNEZGcGVlcmludGYxMjAvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '120',
                                                                       'in_nucast_pkts': '715302',
                                                                       'in_octets': '169988869151',
                                                                       'in_ucast_pkts': '119230868',
                                                                       'links': [       {       'in_nucast_pkts': '715302',
                                                                                                'in_octets': '169988869151',
                                                                                                'in_ucast_pkts': '119230868',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10954',
                                                                                                'out_octets': '3386317179',
                                                                                                'out_ucast_pkts': '30440142',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10954',
                                                                       'out_octets': '3386317179',
                                                                       'out_ucast_pkts': '30440142',
                                                                       'poller_hash': 'MUQwNEZGcGVlcmludGZwb24gMTIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '120',
                                                                                               'links': [       {       'connected_time': '611461',
                                                                                                                        'fiber_length': 20196,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNEZGcGVlcnBvbnBvbiAxMjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '4883 TIFFANY CT (WILLIG MIKE)',
                                         'poller_hash': 'MUQwNEZG',
                                         'serial_number': '1D04FF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e8:f3:ad',
                                         'uptime': '611461'},
                                 {       'counter_source': '121/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '74:da:88:25:b9:4b'],
                                                                       'id': '121/f1',
                                                                       'in_nucast_pkts': '100925',
                                                                       'in_octets': '258567980858',
                                                                       'in_ucast_pkts': '1424666635',
                                                                       'mac_address': '00:06:31:11:f0:12',
                                                                       'mtu': 1608,
                                                                       'name': '121/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22477666',
                                                                       'out_octets': '5143638288303',
                                                                       'out_ucast_pkts': '23862338857506',
                                                                       'poller_hash': 'MTMwNDhwZWVyaW50ZjEyMS9mMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '121',
                                                                       'in_nucast_pkts': '22477666',
                                                                       'in_octets': '5143638288303',
                                                                       'in_ucast_pkts': '23862338857506',
                                                                       'links': [       {       'in_nucast_pkts': '22477666',
                                                                                                'in_octets': '5143638288303',
                                                                                                'in_ucast_pkts': '23862338857506',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '100925',
                                                                                                'out_octets': '258567980858',
                                                                                                'out_ucast_pkts': '1424666635',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '100925',
                                                                       'out_octets': '258567980858',
                                                                       'out_ucast_pkts': '1424666635',
                                                                       'poller_hash': 'MTMwNDhwZWVyaW50ZnBvbiAxMjE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '121',
                                                                                               'links': [       {       'connected_time': '11964782',
                                                                                                                        'fiber_length': 19694,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwNDhwZWVycG9ucG9uIDEyMQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712',
                                         'name': '4906 Linhaven Dr. (Darly Iszczuk)',
                                         'poller_hash': 'MTMwNDg=',
                                         'serial_number': '13048',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:11:f0:12',
                                         'uptime': '11964782'},
                                 {       'counter_source': '122/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '122/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '148115074921',
                                                                       'in_ucast_pkts': '451738954',
                                                                       'mac_address': 'cc:be:59:6d:a9:86',
                                                                       'name': '122/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '988018709002',
                                                                       'out_ucast_pkts': '908737324',
                                                                       'poller_hash': 'NDAwRDY4cGVlcmludGYxMjIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '122',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '988018709002',
                                                                       'in_ucast_pkts': '908737324',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '988018709002',
                                                                                                'in_ucast_pkts': '908737324',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '148115074921',
                                                                                                'out_ucast_pkts': '451738954',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '148115074921',
                                                                       'out_ucast_pkts': '451738954',
                                                                       'poller_hash': 'NDAwRDY4cGVlcmludGZwb24gMTIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '122',
                                                                                               'links': [       {       'connected_time': '639520',
                                                                                                                        'fiber_length': 20226,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDAwRDY4cGVlcnBvbnBvbiAxMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4891 Linhaven Dr. (Gerry Schmidt)',
                                         'poller_hash': 'NDAwRDY4',
                                         'serial_number': '400D68',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6d:a9:86',
                                         'uptime': '639520'},
                                 {       'counter_source': '123/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b8:d2'],
                                                                       'id': '123/g1',
                                                                       'in_nucast_pkts': '13128',
                                                                       'in_octets': '3714264057',
                                                                       'in_ucast_pkts': '18091647',
                                                                       'mac_address': '60:db:98:4e:67:df',
                                                                       'mtu': 2000,
                                                                       'name': '123/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715421',
                                                                       'out_octets': '166804099469',
                                                                       'out_ucast_pkts': '114974538',
                                                                       'poller_hash': 'QTlDMkJCcGVlcmludGYxMjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '123',
                                                                       'in_nucast_pkts': '715421',
                                                                       'in_octets': '166804099469',
                                                                       'in_ucast_pkts': '114974538',
                                                                       'links': [       {       'in_nucast_pkts': '715421',
                                                                                                'in_octets': '166804099469',
                                                                                                'in_ucast_pkts': '114974538',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13128',
                                                                                                'out_octets': '3714264057',
                                                                                                'out_ucast_pkts': '18091647',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13128',
                                                                       'out_octets': '3714264057',
                                                                       'out_ucast_pkts': '18091647',
                                                                       'poller_hash': 'QTlDMkJCcGVlcmludGZwb24gMTIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '123',
                                                                                               'links': [       {       'connected_time': '611449',
                                                                                                                        'fiber_length': 19782,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkJCcGVlcnBvbnBvbiAxMjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4010 Fredrick St.',
                                         'poller_hash': 'QTlDMkJC',
                                         'serial_number': 'A9C2BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:67:dc',
                                         'uptime': '611449'},
                                 {       'counter_source': '124/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '124/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '250433045605',
                                                                       'in_ucast_pkts': '1920666624',
                                                                       'mac_address': 'ec:4f:82:8d:49:28',
                                                                       'name': '124/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '197',
                                                                       'out_octets': '6104174274572',
                                                                       'out_ucast_pkts': '4342379576',
                                                                       'poller_hash': 'MzIxMTZDcGVlcmludGYxMjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '124',
                                                                       'in_nucast_pkts': '197',
                                                                       'in_octets': '6104174274572',
                                                                       'in_ucast_pkts': '4342379576',
                                                                       'links': [       {       'in_nucast_pkts': '197',
                                                                                                'in_octets': '6104174274572',
                                                                                                'in_ucast_pkts': '4342379576',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '250433045605',
                                                                                                'out_ucast_pkts': '1920666624',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '250433045605',
                                                                       'out_ucast_pkts': '1920666624',
                                                                       'poller_hash': 'MzIxMTZDcGVlcmludGZwb24gMTI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '124',
                                                                                               'links': [       {       'connected_time': '73884952',
                                                                                                                        'fiber_length': 19564,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMTZDcGVlcnBvbnBvbiAxMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4912 PERTH ROAD 119 (DOWNEY BILL & '
                                                 'KAREN)',
                                         'poller_hash': 'MzIxMTZD',
                                         'serial_number': '32116C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:49:28',
                                         'uptime': '73884952'},
                                 {       'counter_source': '125/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '125/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '409244878140',
                                                                       'in_ucast_pkts': '23643709677',
                                                                       'mac_address': 'ec:4f:82:8d:50:68',
                                                                       'name': '125/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '197',
                                                                       'out_octets': '9327843038362',
                                                                       'out_ucast_pkts': '6689844986',
                                                                       'poller_hash': 'MzIxMjU0cGVlcmludGYxMjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '125',
                                                                       'in_nucast_pkts': '197',
                                                                       'in_octets': '9327843038362',
                                                                       'in_ucast_pkts': '6689844986',
                                                                       'links': [       {       'in_nucast_pkts': '197',
                                                                                                'in_octets': '9327843038362',
                                                                                                'in_ucast_pkts': '6689844986',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '409244878140',
                                                                                                'out_ucast_pkts': '23643709677',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '409244878140',
                                                                       'out_ucast_pkts': '23643709677',
                                                                       'poller_hash': 'MzIxMjU0cGVlcmludGZwb24gMTI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '125',
                                                                                               'links': [       {       'connected_time': '30751422',
                                                                                                                        'fiber_length': 19701,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMjU0cGVlcnBvbnBvbiAxMjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4920 LINHAVEN DR (FAULHAFER DON)',
                                         'poller_hash': 'MzIxMjU0',
                                         'serial_number': '321254',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:50:68',
                                         'uptime': '30751422'},
                                 {       'counter_source': '126/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:33:63:aa'],
                                                                       'id': '126/g1',
                                                                       'in_nucast_pkts': '423',
                                                                       'in_octets': '477024289629',
                                                                       'in_ucast_pkts': '1650841580',
                                                                       'mac_address': 'cc:be:59:6f:c9:4e',
                                                                       'mtu': 2000,
                                                                       'name': '126/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22467434',
                                                                       'out_octets': '7829916582812',
                                                                       'out_ucast_pkts': '5866150340',
                                                                       'poller_hash': 'NTVGNTFDcGVlcmludGYxMjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '126',
                                                                       'in_nucast_pkts': '22467434',
                                                                       'in_octets': '7829916582812',
                                                                       'in_ucast_pkts': '5866150340',
                                                                       'links': [       {       'in_nucast_pkts': '22467434',
                                                                                                'in_octets': '7829916582812',
                                                                                                'in_ucast_pkts': '5866150340',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '423',
                                                                                                'out_octets': '477024289629',
                                                                                                'out_ucast_pkts': '1650841580',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '423',
                                                                       'out_octets': '477024289629',
                                                                       'out_ucast_pkts': '1650841580',
                                                                       'poller_hash': 'NTVGNTFDcGVlcmludGZwb24gMTI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '126',
                                                                                               'links': [       {       'connected_time': '552562',
                                                                                                                        'fiber_length': 19967,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.226,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTVGNTFDcGVlcnBvbnBvbiAxMjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': "4955 Perth Rd. 119 (O'DRISCOLL "
                                                 'MICHELLE)',
                                         'poller_hash': 'NTVGNTFD',
                                         'serial_number': '55F51C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6f:c9:4a',
                                         'uptime': '552562'},
                                 {       'counter_source': '129/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:94:86'],
                                                                       'id': '129/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '43807249955',
                                                                       'in_ucast_pkts': '197725654',
                                                                       'mac_address': 'ec:4f:82:7e:0a:69',
                                                                       'mtu': 2000,
                                                                       'name': '129/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919752',
                                                                       'out_octets': '1439329920585',
                                                                       'out_ucast_pkts': '992161959',
                                                                       'poller_hash': 'MzVFOUMzcGVlcmludGYxMjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '129',
                                                                       'in_nucast_pkts': '9919752',
                                                                       'in_octets': '1439329920585',
                                                                       'in_ucast_pkts': '992161959',
                                                                       'links': [       {       'in_nucast_pkts': '9919752',
                                                                                                'in_octets': '1439329920585',
                                                                                                'in_ucast_pkts': '992161959',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '43807249955',
                                                                                                'out_ucast_pkts': '197725654',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '43807249955',
                                                                       'out_ucast_pkts': '197725654',
                                                                       'poller_hash': 'MzVFOUMzcGVlcmludGZwb24gMTI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '129',
                                                                                               'links': [       {       'connected_time': '8535469',
                                                                                                                        'fiber_length': 23184,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOUMzcGVlcnBvbnBvbiAxMjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1 Pond View Unit 303 (Mary Lou '
                                                 'Shantz 10434317)',
                                         'poller_hash': 'MzVFOUMz',
                                         'serial_number': '35E9C3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0a:69',
                                         'uptime': '8535469'},
                                 {       'counter_source': '130/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:ad:e0:4a:41:f2'],
                                                                       'id': '130/g1',
                                                                       'in_nucast_pkts': '437504',
                                                                       'in_octets': '201622042647',
                                                                       'in_ucast_pkts': '529003547',
                                                                       'mac_address': '48:77:46:f8:74:87',
                                                                       'mtu': 2000,
                                                                       'name': '130/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9925337',
                                                                       'out_octets': '1423900731783',
                                                                       'out_ucast_pkts': '1107384232',
                                                                       'poller_hash': 'ODJBNTAzcGVlcmludGYxMzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '130',
                                                                       'in_nucast_pkts': '9925337',
                                                                       'in_octets': '1423900731783',
                                                                       'in_ucast_pkts': '1107384232',
                                                                       'links': [       {       'in_nucast_pkts': '9925337',
                                                                                                'in_octets': '1423900731783',
                                                                                                'in_ucast_pkts': '1107384232',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '437504',
                                                                                                'out_octets': '201622042647',
                                                                                                'out_ucast_pkts': '529003547',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '437504',
                                                                       'out_octets': '201622042647',
                                                                       'out_ucast_pkts': '529003547',
                                                                       'poller_hash': 'ODJBNTAzcGVlcmludGZwb24gMTMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '130',
                                                                                               'links': [       {       'connected_time': '8535406',
                                                                                                                        'fiber_length': 23220,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNTAzcGVlcnBvbnBvbiAxMzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1125 MOLESWORTH ST (WHITSITT WARREN)',
                                         'poller_hash': 'ODJBNTAz',
                                         'serial_number': '82A503',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:74:87',
                                         'uptime': '8535406'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '132/g1',
                                                                       'name': '132/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyODFGcGVlcmludGYxMzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '132',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyODFGcGVlcmludGZwb24gMTMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '132',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODFGcGVlcnBvbnBvbiAxMzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '54 POND VIEW DR',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "132" '
                                                                  '(location: GPON '
                                                                  'port "6/1"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:12.890Z',
                                         'poller_hash': 'MzIyODFG',
                                         'serial_number': '32281F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '133/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '133/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '265617750148',
                                                                       'in_ucast_pkts': '1446313533',
                                                                       'mac_address': 'cc:be:59:6a:df:7e',
                                                                       'name': '133/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6097911542002',
                                                                       'out_ucast_pkts': '4489383414',
                                                                       'poller_hash': 'NDZDRTRFcGVlcmludGYxMzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '133',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6097911542002',
                                                                       'in_ucast_pkts': '4489383414',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6097911542002',
                                                                                                'in_ucast_pkts': '4489383414',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '265617750148',
                                                                                                'out_ucast_pkts': '1446313533',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '265617750148',
                                                                       'out_ucast_pkts': '1446313533',
                                                                       'poller_hash': 'NDZDRTRFcGVlcmludGZwb24gMTMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '133',
                                                                                               'links': [       {       'connected_time': '69151777',
                                                                                                                        'fiber_length': 19812,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.244,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDZDRTRFcGVlcnBvbnBvbiAxMzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4947 Perth Road 119 (Margaret Rieck)',
                                         'poller_hash': 'NDZDRTRF',
                                         'serial_number': '46CE4E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:6a:df:7e',
                                         'uptime': '69151777'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '134/g1',
                                                                       'name': '134/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkVEOUMycGVlcmludGYxMzQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '134',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkVEOUMycGVlcmludGZwb24gMTM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '134',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkVEOUMycGVlcnBvbnBvbiAxMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4913 Linhaven Dr.Gadshill (B & C '
                                                 'Player 10423825)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "134" '
                                                                  '(location: GPON '
                                                                  'port "5/3"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-05-30T17:00:56.060Z',
                                         'poller_hash': 'MkVEOUMy',
                                         'serial_number': '2ED9C2',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '136/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '136/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '130198788517',
                                                                       'in_ucast_pkts': '1287247308',
                                                                       'mac_address': '84:d3:43:7d:3f:37',
                                                                       'name': '136/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '161',
                                                                       'out_octets': '7746673676690',
                                                                       'out_ucast_pkts': '9434458105',
                                                                       'poller_hash': 'QzgxRDUwcGVlcmludGYxMzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '136',
                                                                       'in_nucast_pkts': '161',
                                                                       'in_octets': '7746673676690',
                                                                       'in_ucast_pkts': '9434458105',
                                                                       'links': [       {       'in_nucast_pkts': '161',
                                                                                                'in_octets': '7746673676690',
                                                                                                'in_ucast_pkts': '9434458105',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '130198788517',
                                                                                                'out_ucast_pkts': '1287247308',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '130198788517',
                                                                       'out_ucast_pkts': '1287247308',
                                                                       'poller_hash': 'QzgxRDUwcGVlcmludGZwb24gMTM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '136',
                                                                                               'links': [       {       'connected_time': '70857444',
                                                                                                                        'fiber_length': 19899,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDUwcGVlcnBvbnBvbiAxMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4900 Linhaven Dr. Gadshill (Ronald '
                                                 'Hundt 10423790)',
                                         'poller_hash': 'QzgxRDUw',
                                         'serial_number': 'C81D50',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:3f:37',
                                         'uptime': '70857444'},
                                 {       'counter_source': '137/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '137/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '287870027893',
                                                                       'in_ucast_pkts': '30950418098',
                                                                       'mac_address': 'cc:be:59:cb:b6:32',
                                                                       'name': '137/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '53',
                                                                       'out_octets': '3488807425996',
                                                                       'out_ucast_pkts': '2585574948',
                                                                       'poller_hash': 'NDJDODFFcGVlcmludGYxMzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '137',
                                                                       'in_nucast_pkts': '53',
                                                                       'in_octets': '3488807425996',
                                                                       'in_ucast_pkts': '2585574948',
                                                                       'links': [       {       'in_nucast_pkts': '53',
                                                                                                'in_octets': '3488807425996',
                                                                                                'in_ucast_pkts': '2585574948',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '287870027893',
                                                                                                'out_ucast_pkts': '30950418098',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '287870027893',
                                                                       'out_ucast_pkts': '30950418098',
                                                                       'poller_hash': 'NDJDODFFcGVlcmludGZwb24gMTM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137',
                                                                                               'links': [       {       'connected_time': '611449',
                                                                                                                        'fiber_length': 19772,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDODFFcGVlcnBvbnBvbiAxMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4922 Linhaven Dr. Gadshill (Protemp '
                                                 'Refridgeration 10436516)',
                                         'poller_hash': 'NDJDODFF',
                                         'serial_number': '42C81E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:b6:32',
                                         'uptime': '611449'},
                                 {       'counter_source': '138/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '138/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '398877910137',
                                                                       'in_ucast_pkts': '2466779513',
                                                                       'mac_address': 'cc:be:59:b7:3b:37',
                                                                       'name': '138/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7144546674994',
                                                                       'out_ucast_pkts': '5203476959',
                                                                       'poller_hash': 'NEQxMTE5cGVlcmludGYxMzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7144546674994',
                                                                       'in_ucast_pkts': '5203476959',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7144546674994',
                                                                                                'in_ucast_pkts': '5203476959',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '398877910137',
                                                                                                'out_ucast_pkts': '2466779513',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '398877910137',
                                                                       'out_ucast_pkts': '2466779513',
                                                                       'poller_hash': 'NEQxMTE5cGVlcmludGZwb24gMTM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138',
                                                                                               'links': [       {       'connected_time': '639271',
                                                                                                                        'fiber_length': 19676,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.21,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQxMTE5cGVlcnBvbnBvbiAxMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4914 Linhaven Dr. Gadshill (Richard '
                                                 'Schneider)',
                                         'poller_hash': 'NEQxMTE5',
                                         'serial_number': '4D1119',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:b7:3b:37',
                                         'uptime': '639271'},
                                 {       'counter_source': '139/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '0c:80:63:f6:b9:68'],
                                                                       'id': '139/f1',
                                                                       'in_nucast_pkts': '96178',
                                                                       'in_octets': '659778656409',
                                                                       'in_ucast_pkts': '6233908556',
                                                                       'mac_address': '00:06:31:08:4c:42',
                                                                       'mtu': 1608,
                                                                       'name': '139/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22438159',
                                                                       'out_octets': '5259919706063',
                                                                       'out_ucast_pkts': '22299153829388',
                                                                       'poller_hash': 'RjQzNnBlZXJpbnRmMTM5L2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139',
                                                                       'in_nucast_pkts': '22438159',
                                                                       'in_octets': '5259919706063',
                                                                       'in_ucast_pkts': '22299153829388',
                                                                       'links': [       {       'in_nucast_pkts': '22438159',
                                                                                                'in_octets': '5259919706063',
                                                                                                'in_ucast_pkts': '22299153829388',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '96178',
                                                                                                'out_octets': '659778656409',
                                                                                                'out_ucast_pkts': '6233908556',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '96178',
                                                                       'out_octets': '659778656409',
                                                                       'out_ucast_pkts': '6233908556',
                                                                       'poller_hash': 'RjQzNnBlZXJpbnRmcG9uIDEzOQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139',
                                                                                               'links': [       {       'connected_time': '11965976',
                                                                                                                        'fiber_length': 19304,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjQzNnBlZXJwb25wb24gMTM5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4902 Perth Road 119 (Steve Winkel)',
                                         'poller_hash': 'RjQzNg==',
                                         'serial_number': 'F436',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:4c:42',
                                         'uptime': '11965976'},
                                 {       'counter_source': '141/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:16:f9:4f:27:84'],
                                                                       'id': '141/g1',
                                                                       'in_nucast_pkts': '18175',
                                                                       'in_octets': '47137398112',
                                                                       'in_ucast_pkts': '4725304090',
                                                                       'mac_address': '00:06:31:18:c1:2e',
                                                                       'mtu': 1608,
                                                                       'name': '141/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22502491',
                                                                       'out_octets': '1419351688769',
                                                                       'out_ucast_pkts': '5293009612',
                                                                       'poller_hash': 'M0NGQUVwZWVyaW50ZjE0MS9nMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141',
                                                                       'in_nucast_pkts': '22502491',
                                                                       'in_octets': '1419351688769',
                                                                       'in_ucast_pkts': '5293009612',
                                                                       'links': [       {       'in_nucast_pkts': '22502491',
                                                                                                'in_octets': '1419351688769',
                                                                                                'in_ucast_pkts': '5293009612',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18175',
                                                                                                'out_octets': '47137398112',
                                                                                                'out_ucast_pkts': '4725304090',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18175',
                                                                       'out_octets': '47137398112',
                                                                       'out_ucast_pkts': '4725304090',
                                                                       'poller_hash': 'M0NGQUVwZWVyaW50ZnBvbiAxNDE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141',
                                                                                               'links': [       {       'connected_time': '11965226',
                                                                                                                        'fiber_length': 19472,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0NGQUVwZWVycG9ucG9uIDE0MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712GX',
                                         'name': '4919 Linhaven Dr. Gadshill (Ken '
                                                 'Huether)',
                                         'poller_hash': 'M0NGQUU=',
                                         'serial_number': '3CFAE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:18:c1:2e',
                                         'uptime': '11965226'},
                                 {       'counter_source': '142/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '142/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '613',
                                                                       'in_octets': '348367835580',
                                                                       'in_ucast_pkts': '1148601839',
                                                                       'mac_address': '84:d3:43:7e:1e:97',
                                                                       'name': '142/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1352',
                                                                       'out_octets': '5733089419131',
                                                                       'out_ucast_pkts': '4196135622',
                                                                       'poller_hash': 'QzgzOTNDcGVlcmludGYxNDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '142',
                                                                       'in_nucast_pkts': '1352',
                                                                       'in_octets': '5733089419131',
                                                                       'in_ucast_pkts': '4196135622',
                                                                       'links': [       {       'in_nucast_pkts': '1352',
                                                                                                'in_octets': '5733089419131',
                                                                                                'in_ucast_pkts': '4196135622',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '613',
                                                                                                'out_octets': '348367835580',
                                                                                                'out_ucast_pkts': '1148601839',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '613',
                                                                       'out_octets': '348367835580',
                                                                       'out_ucast_pkts': '1148601839',
                                                                       'poller_hash': 'QzgzOTNDcGVlcmludGZwb24gMTQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142',
                                                                                               'links': [       {       'connected_time': '8535178',
                                                                                                                        'fiber_length': 23737,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzOTNDcGVlcnBvbnBvbiAxNDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '60 REINER CR MILVERTON (WELLMAN '
                                                 'MICHELLE)',
                                         'poller_hash': 'QzgzOTND',
                                         'serial_number': 'C8393C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:1e:97',
                                         'uptime': '8535178'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '143/f1',
                                                                       'name': '143/f1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTJFODJwZWVyaW50ZjE0My9mMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '143',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTJFODJwZWVyaW50ZnBvbiAxNDM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTJFODJwZWVycG9ucG9uIDE0Mw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4898 Linhaven Dr. Gadshill (James '
                                                 'Smits)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "143" '
                                                                  '(location: GPON '
                                                                  'port "5/2"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2025-04-24T17:47:22.370Z',
                                         'poller_hash': 'MTJFODI=',
                                         'serial_number': '12E82',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '144/F1+144/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:aa:4d:57'],
                                                                       'id': '144/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1798379924',
                                                                       'in_ucast_pkts': '10540835',
                                                                       'name': '144/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '19848474029',
                                                                       'out_octets': '26622487272852',
                                                                       'out_ucast_pkts': '5504524',
                                                                       'poller_hash': 'NDJDNjk3cGVlcmludGYxNDQvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '144/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '56595522838',
                                                                       'in_ucast_pkts': '392293083',
                                                                       'mac_address': 'cc:be:59:cb:a9:fa',
                                                                       'name': '144/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '63',
                                                                       'out_octets': '1860047330810',
                                                                       'out_ucast_pkts': '1336203341',
                                                                       'poller_hash': 'NDJDNjk3cGVlcmludGYxNDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144',
                                                                       'in_nucast_pkts': '19848474092',
                                                                       'in_octets': '28482534603662',
                                                                       'in_ucast_pkts': '1341707865',
                                                                       'links': [       {       'in_nucast_pkts': '19848474092',
                                                                                                'in_octets': '28482534603662',
                                                                                                'in_ucast_pkts': '1341707865',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '58393902762',
                                                                                                'out_ucast_pkts': '402833918',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '58393902762',
                                                                       'out_ucast_pkts': '402833918',
                                                                       'poller_hash': 'NDJDNjk3cGVlcmludGZwb24gMTQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144',
                                                                                               'links': [       {       'connected_time': '611451',
                                                                                                                        'fiber_length': 20166,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNjk3cGVlcnBvbnBvbiAxNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4886 Linhaven Dr. Gadshill (Brian '
                                                 'Deleemans 10435582)',
                                         'poller_hash': 'NDJDNjk3',
                                         'serial_number': '42C697',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:a9:fa',
                                         'uptime': '611451'},
                                 {       'counter_source': '145/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2262266768472',
                                                                       'in_ucast_pkts': '9615370358',
                                                                       'mac_address': '00:06:31:c1:71:87',
                                                                       'name': '145/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '30503958602473',
                                                                       'out_ucast_pkts': '22010405001',
                                                                       'poller_hash': 'MTY0MjdGcGVlcmludGYxNDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '30503958602473',
                                                                       'in_ucast_pkts': '22010405001',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '30503958602473',
                                                                                                'in_ucast_pkts': '22010405001',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2262266768472',
                                                                                                'out_ucast_pkts': '9615370358',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2262266768472',
                                                                       'out_ucast_pkts': '9615370358',
                                                                       'poller_hash': 'MTY0MjdGcGVlcmludGZwb24gMTQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145',
                                                                                               'links': [       {       'connected_time': '639606',
                                                                                                                        'fiber_length': 19930,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.056,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY0MjdGcGVlcnBvbnBvbiAxNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4892 LINHAVEN DR (READINGS JOE)',
                                         'poller_hash': 'MTY0MjdG',
                                         'serial_number': '16427F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:b7:5c:46',
                                         'uptime': '639606'},
                                 {       'counter_source': '146/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1047036989361',
                                                                       'in_ucast_pkts': '2798401287',
                                                                       'mac_address': '48:77:46:a2:1c:ac',
                                                                       'name': '146/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '16024678272395',
                                                                       'out_ucast_pkts': '11677616392',
                                                                       'poller_hash': 'NzFBNEM0cGVlcmludGYxNDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '16024678272395',
                                                                       'in_ucast_pkts': '11677616392',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '16024678272395',
                                                                                                'in_ucast_pkts': '11677616392',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1047036989361',
                                                                                                'out_ucast_pkts': '2798401287',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1047036989361',
                                                                       'out_ucast_pkts': '2798401287',
                                                                       'poller_hash': 'NzFBNEM0cGVlcmludGZwb24gMTQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146',
                                                                                               'links': [       {       'connected_time': '782087',
                                                                                                                        'fiber_length': 19347,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.238,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFBNEM0cGVlcnBvbnBvbiAxNDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4011 Frederick St. Gadshill (Sheila '
                                                 'Loerzel)',
                                         'poller_hash': 'NzFBNEM0',
                                         'serial_number': '71A4C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a2:1c:ac',
                                         'uptime': '782087'},
                                 {       'counter_source': '147/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '995133012799',
                                                                       'in_ucast_pkts': '2383219406',
                                                                       'mac_address': 'cc:be:59:47:32:ff',
                                                                       'name': '147/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9072600269656',
                                                                       'out_ucast_pkts': '7168989442',
                                                                       'poller_hash': 'NDMzRTY3cGVlcmludGYxNDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9072600269656',
                                                                       'in_ucast_pkts': '7168989442',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '9072600269656',
                                                                                                'in_ucast_pkts': '7168989442',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '995133012799',
                                                                                                'out_ucast_pkts': '2383219406',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '995133012799',
                                                                       'out_ucast_pkts': '2383219406',
                                                                       'poller_hash': 'NDMzRTY3cGVlcmludGZwb24gMTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147',
                                                                                               'links': [       {       'connected_time': '782108',
                                                                                                                        'fiber_length': 19884,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDMzRTY3cGVlcnBvbnBvbiAxNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4896 Linhaven Dr. Gadshill (Trevor '
                                                 'Wraith)',
                                         'poller_hash': 'NDMzRTY3',
                                         'serial_number': '433E67',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:47:32:ff',
                                         'uptime': '782108'},
                                 {       'counter_source': '149/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '745234633617',
                                                                       'in_ucast_pkts': '6260466474',
                                                                       'mac_address': 'ec:4f:82:32:25:67',
                                                                       'name': '149/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '26308933953301',
                                                                       'out_ucast_pkts': '1180635617',
                                                                       'poller_hash': 'MjQ4QUJEcGVlcmludGYxNDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '26308933953301',
                                                                       'in_ucast_pkts': '1180635617',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '26308933953301',
                                                                                                'in_ucast_pkts': '1180635617',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '745234633617',
                                                                                                'out_ucast_pkts': '6260466474',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '745234633617',
                                                                       'out_ucast_pkts': '6260466474',
                                                                       'poller_hash': 'MjQ4QUJEcGVlcmludGZwb24gMTQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149',
                                                                                               'links': [       {       'connected_time': '15842329',
                                                                                                                        'fiber_length': 19846,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.95,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjQ4QUJEcGVlcnBvbnBvbiAxNDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4910 Linhaven Dr. Gadshill (Sheila '
                                                 'Hausleitner)',
                                         'poller_hash': 'MjQ4QUJE',
                                         'serial_number': '248ABD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:32:25:67',
                                         'uptime': '15842329'},
                                 {       'counter_source': '150/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:ec:5e:f6:14:31'],
                                                                       'id': '150/f1',
                                                                       'in_nucast_pkts': '2887140',
                                                                       'in_octets': '271775972210',
                                                                       'in_ucast_pkts': '846083288',
                                                                       'mac_address': '00:06:31:08:49:ae',
                                                                       'mtu': 1608,
                                                                       'name': '150/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22512584',
                                                                       'out_octets': '3504265650672',
                                                                       'out_ucast_pkts': '20206293829158',
                                                                       'poller_hash': 'RjBGNXBlZXJpbnRmMTUwL2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '150',
                                                                       'in_nucast_pkts': '22512584',
                                                                       'in_octets': '3504265650672',
                                                                       'in_ucast_pkts': '20206293829158',
                                                                       'links': [       {       'in_nucast_pkts': '22512584',
                                                                                                'in_octets': '3504265650672',
                                                                                                'in_ucast_pkts': '20206293829158',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2887140',
                                                                                                'out_octets': '271775972210',
                                                                                                'out_ucast_pkts': '846083288',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2887140',
                                                                       'out_octets': '271775972210',
                                                                       'out_ucast_pkts': '846083288',
                                                                       'poller_hash': 'RjBGNXBlZXJpbnRmcG9uIDE1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150',
                                                                                               'links': [       {       'connected_time': '11965030',
                                                                                                                        'fiber_length': 20026,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjBGNXBlZXJwb25wb24gMTUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4881 Tiffany Ct. Gadshill (Jim '
                                                 'Brubacher)',
                                         'poller_hash': 'RjBGNQ==',
                                         'serial_number': 'F0F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:49:ae',
                                         'uptime': '11965030'},
                                 {       'counter_source': '151/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '151/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1002122980499',
                                                                       'in_ucast_pkts': '7815130324',
                                                                       'mac_address': 'f8:85:f9:6f:20:b5',
                                                                       'name': '151/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25848041822148',
                                                                       'out_ucast_pkts': '18065445731',
                                                                       'poller_hash': 'QkYyNEM2cGVlcmludGYxNTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '151',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25848041822148',
                                                                       'in_ucast_pkts': '18065445731',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '25848041822148',
                                                                                                'in_ucast_pkts': '18065445731',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1002122980499',
                                                                                                'out_ucast_pkts': '7815130324',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1002122980499',
                                                                       'out_ucast_pkts': '7815130324',
                                                                       'poller_hash': 'QkYyNEM2cGVlcmludGZwb24gMTUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '151',
                                                                                               'links': [       {       'connected_time': '72549051',
                                                                                                                        'fiber_length': 20194,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QkYyNEM2cGVlcnBvbnBvbiAxNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4838 Perth Road 119 (Jean Langille)',
                                         'poller_hash': 'QkYyNEM2',
                                         'serial_number': 'BF24C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'f8:85:f9:6f:20:b5',
                                         'uptime': '72549051'},
                                 {       'counter_source': '153/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e4:a3'],
                                                                       'id': '153/g1',
                                                                       'in_nucast_pkts': '832663',
                                                                       'in_octets': '349934076508',
                                                                       'in_ucast_pkts': '2802837189',
                                                                       'mac_address': '48:77:46:3b:c8:3e',
                                                                       'mtu': 2000,
                                                                       'name': '153/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22408168',
                                                                       'out_octets': '26412036227387',
                                                                       'out_ucast_pkts': '17652693398',
                                                                       'poller_hash': 'NzI0QjIycGVlcmludGYxNTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '153',
                                                                       'in_nucast_pkts': '22408168',
                                                                       'in_octets': '26412036227387',
                                                                       'in_ucast_pkts': '17652693398',
                                                                       'links': [       {       'in_nucast_pkts': '22408168',
                                                                                                'in_octets': '26412036227387',
                                                                                                'in_ucast_pkts': '17652693398',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '832663',
                                                                                                'out_octets': '349934076508',
                                                                                                'out_ucast_pkts': '2802837189',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '832663',
                                                                       'out_octets': '349934076508',
                                                                       'out_ucast_pkts': '2802837189',
                                                                       'poller_hash': 'NzI0QjIycGVlcmludGZwb24gMTUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '153',
                                                                                               'links': [       {       'connected_time': '417331',
                                                                                                                        'fiber_length': 20178,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzI0QjIycGVlcnBvbnBvbiAxNTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4885 Tiffany Ct. Gadshill (Paul '
                                                 'Duden)',
                                         'poller_hash': 'NzI0QjIy',
                                         'serial_number': '724B22',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:3b:c8:3a',
                                         'uptime': '417331'},
                                 {       'counter_source': '155/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:84:6a:75:21:53'],
                                                                       'id': '155/f1',
                                                                       'in_nucast_pkts': '30663605',
                                                                       'in_octets': '2505669047856',
                                                                       'in_ucast_pkts': '3496947896',
                                                                       'mac_address': '00:06:31:08:4c:90',
                                                                       'mtu': 1608,
                                                                       'name': '155/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22489460',
                                                                       'out_octets': '2070730835192',
                                                                       'out_ucast_pkts': '3806536412',
                                                                       'poller_hash': 'RjQzMXBlZXJpbnRmMTU1L2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '155',
                                                                       'in_nucast_pkts': '22489460',
                                                                       'in_octets': '2070730835192',
                                                                       'in_ucast_pkts': '3806536412',
                                                                       'links': [       {       'in_nucast_pkts': '22489460',
                                                                                                'in_octets': '2070730835192',
                                                                                                'in_ucast_pkts': '3806536412',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '30663605',
                                                                                                'out_octets': '2505669047856',
                                                                                                'out_ucast_pkts': '3496947896',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30663605',
                                                                       'out_octets': '2505669047856',
                                                                       'out_ucast_pkts': '3496947896',
                                                                       'poller_hash': 'RjQzMXBlZXJpbnRmcG9uIDE1NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '155',
                                                                                               'links': [       {       'connected_time': '11966034',
                                                                                                                        'fiber_length': 19393,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjQzMXBlZXJwb25wb24gMTU1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4916 Perth Road 119 (Mark Walker)',
                                         'poller_hash': 'RjQzMQ==',
                                         'serial_number': 'F431',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:4c:90',
                                         'uptime': '11966034'},
                                 {       'counter_source': '157/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '157/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '767591782613',
                                                                       'in_ucast_pkts': '2402468774',
                                                                       'mac_address': '00:06:31:69:f5:d1',
                                                                       'name': '157/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '11250857793076',
                                                                       'out_ucast_pkts': '8307630310',
                                                                       'poller_hash': 'QzU4MEZwZWVyaW50ZjE1Ny9HMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '157',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '11250857793076',
                                                                       'in_ucast_pkts': '8307630310',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '11250857793076',
                                                                                                'in_ucast_pkts': '8307630310',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '767591782613',
                                                                                                'out_ucast_pkts': '2402468774',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '767591782613',
                                                                       'out_ucast_pkts': '2402468774',
                                                                       'poller_hash': 'QzU4MEZwZWVyaW50ZnBvbiAxNTc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157',
                                                                                               'links': [       {       'connected_time': '782087',
                                                                                                                        'fiber_length': 19712,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzU4MEZwZWVycG9ucG9uIDE1Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '4006 Andy St. Gadshill (Jen Rose)',
                                         'poller_hash': 'QzU4MEY=',
                                         'serial_number': 'C580F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:5f:68:cf',
                                         'uptime': '782087'},
                                 {       'counter_source': '158/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '158/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1251926285014',
                                                                       'in_ucast_pkts': '6520993690',
                                                                       'mac_address': 'cc:be:59:53:6f:6f',
                                                                       'name': '158/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '29168065939078',
                                                                       'out_ucast_pkts': '21366703317',
                                                                       'poller_hash': 'M0ZFNDA5cGVlcmludGYxNTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '158',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '29168065939078',
                                                                       'in_ucast_pkts': '21366703317',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '29168065939078',
                                                                                                'in_ucast_pkts': '21366703317',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1251926285014',
                                                                                                'out_ucast_pkts': '6520993690',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1251926285014',
                                                                       'out_ucast_pkts': '6520993690',
                                                                       'poller_hash': 'M0ZFNDA5cGVlcmludGZwb24gMTU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '158',
                                                                                               'links': [       {       'connected_time': '13267352',
                                                                                                                        'fiber_length': 19573,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.828,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0ZFNDA5cGVlcnBvbnBvbiAxNTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4005 Andy St. Gadshill (Michael '
                                                 'Bean)',
                                         'poller_hash': 'M0ZFNDA5',
                                         'serial_number': '3FE409',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:53:6f:6f',
                                         'uptime': '13267352'},
                                 {       'counter_source': '160/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:55:6d'],
                                                                       'id': '160/f1',
                                                                       'in_nucast_pkts': '855584',
                                                                       'in_octets': '582752194551',
                                                                       'in_ucast_pkts': '24524998591',
                                                                       'mac_address': '00:06:31:08:4a:05',
                                                                       'mtu': 1608,
                                                                       'name': '160/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22483769',
                                                                       'out_octets': '11160446571785',
                                                                       'out_ucast_pkts': '22384920654304',
                                                                       'poller_hash': 'RjBGNnBlZXJpbnRmMTYwL2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '160',
                                                                       'in_nucast_pkts': '22483769',
                                                                       'in_octets': '11160446571785',
                                                                       'in_ucast_pkts': '22384920654304',
                                                                       'links': [       {       'in_nucast_pkts': '22483769',
                                                                                                'in_octets': '11160446571785',
                                                                                                'in_ucast_pkts': '22384920654304',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '855584',
                                                                                                'out_octets': '582752194551',
                                                                                                'out_ucast_pkts': '24524998591',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '855584',
                                                                       'out_octets': '582752194551',
                                                                       'out_ucast_pkts': '24524998591',
                                                                       'poller_hash': 'RjBGNnBlZXJpbnRmcG9uIDE2MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '160',
                                                                                               'links': [       {       'connected_time': '11965088',
                                                                                                                        'fiber_length': 19434,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjBGNnBlZXJwb25wb24gMTYw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4900 Perth Road 119 (Jason&Trish '
                                                 'Bowles)',
                                         'poller_hash': 'RjBGNg==',
                                         'serial_number': 'F0F6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:4a:05',
                                         'uptime': '11965088'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '161/g1',
                                                                       'name': '161/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NkNFQkRwZWVyaW50ZjE2MS9nMQ=='},
                                                               {       'admin_up': False,
                                                                       'id': '161',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NkNFQkRwZWVyaW50ZnBvbiAxNjE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '161',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NkNFQkRwZWVycG9ucG9uIDE2MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4915 Perth Road 119',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "161" '
                                                                  '(location: GPON '
                                                                  'port "5/2"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:27:55.050Z',
                                         'poller_hash': 'NkNFQkQ=',
                                         'serial_number': '6CEBD',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '162/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:b7:88'],
                                                                       'id': '162/g1',
                                                                       'in_nucast_pkts': '780095',
                                                                       'in_octets': '250958971372',
                                                                       'in_ucast_pkts': '1060787875',
                                                                       'mac_address': '00:06:31:20:01:68',
                                                                       'mtu': 1608,
                                                                       'name': '162/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22465208',
                                                                       'out_octets': '4937471058361',
                                                                       'out_ucast_pkts': '33980145082',
                                                                       'poller_hash': 'NDE1QTdwZWVyaW50ZjE2Mi9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '162',
                                                                       'in_nucast_pkts': '22465208',
                                                                       'in_octets': '4937471058361',
                                                                       'in_ucast_pkts': '33980145082',
                                                                       'links': [       {       'in_nucast_pkts': '22465208',
                                                                                                'in_octets': '4937471058361',
                                                                                                'in_ucast_pkts': '33980145082',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '780095',
                                                                                                'out_octets': '250958971372',
                                                                                                'out_ucast_pkts': '1060787875',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '780095',
                                                                       'out_octets': '250958971372',
                                                                       'out_ucast_pkts': '1060787875',
                                                                       'poller_hash': 'NDE1QTdwZWVyaW50ZnBvbiAxNjI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '162',
                                                                                               'links': [       {       'connected_time': '11965358',
                                                                                                                        'fiber_length': 19892,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDE1QTdwZWVycG9ucG9uIDE2Mg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712GX',
                                         'name': '4872 Perth Road 119 (Jeremy&Dana '
                                                 'Satchell)',
                                         'poller_hash': 'NDE1QTc=',
                                         'serial_number': '415A7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:20:01:68',
                                         'uptime': '11965358'},
                                 {       'counter_source': '164/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '164/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '340212950557',
                                                                       'in_ucast_pkts': '2426009229',
                                                                       'mac_address': '48:77:46:cd:54:e6',
                                                                       'name': '164/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7480531738763',
                                                                       'out_ucast_pkts': '5513313330',
                                                                       'poller_hash': 'NzZEMDQ5cGVlcmludGYxNjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '164',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7480531738763',
                                                                       'in_ucast_pkts': '5513313330',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '7480531738763',
                                                                                                'in_ucast_pkts': '5513313330',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '340212950557',
                                                                                                'out_ucast_pkts': '2426009229',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '340212950557',
                                                                       'out_ucast_pkts': '2426009229',
                                                                       'poller_hash': 'NzZEMDQ5cGVlcmludGZwb24gMTY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164',
                                                                                               'links': [       {       'connected_time': '639182',
                                                                                                                        'fiber_length': 20016,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.584,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzZEMDQ5cGVlcnBvbnBvbiAxNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4949 Perth Road 119 (Elizabeth Berg '
                                                 '10431695)',
                                         'poller_hash': 'NzZEMDQ5',
                                         'serial_number': '76D049',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:cd:54:e6',
                                         'uptime': '639182'},
                                 {       'counter_source': '165/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '165/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '926514165303',
                                                                       'in_ucast_pkts': '2402646055',
                                                                       'mac_address': '48:77:46:3b:8f:12',
                                                                       'name': '165/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '12142890560884',
                                                                       'out_ucast_pkts': '8918601323',
                                                                       'poller_hash': 'NzI0M0ZEcGVlcmludGYxNjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '165',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '12142890560884',
                                                                       'in_ucast_pkts': '8918601323',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '12142890560884',
                                                                                                'in_ucast_pkts': '8918601323',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '926514165303',
                                                                                                'out_ucast_pkts': '2402646055',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '926514165303',
                                                                       'out_ucast_pkts': '2402646055',
                                                                       'poller_hash': 'NzI0M0ZEcGVlcmludGZwb24gMTY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '165',
                                                                                               'links': [       {       'connected_time': '639552',
                                                                                                                        'fiber_length': 19742,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzI0M0ZEcGVlcnBvbnBvbiAxNjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4937 PERTH ROAD 119 (BANNON AMY)',
                                         'poller_hash': 'NzI0M0ZE',
                                         'serial_number': '7243FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:3b:8f:12',
                                         'uptime': '639552'},
                                 {       'counter_source': '166/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:63:7d:4e:fe:c0'],
                                                                       'id': '166/f1',
                                                                       'in_nucast_pkts': '21414',
                                                                       'in_octets': '271835533402',
                                                                       'in_ucast_pkts': '1310311464',
                                                                       'mac_address': '00:06:31:11:ef:dc',
                                                                       'mtu': 1608,
                                                                       'name': '166/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22490735',
                                                                       'out_octets': '6076813659711',
                                                                       'out_ucast_pkts': '22247984409181',
                                                                       'poller_hash': 'MTMwM0VwZWVyaW50ZjE2Ni9mMQ==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '166',
                                                                       'in_nucast_pkts': '22490735',
                                                                       'in_octets': '6076813659711',
                                                                       'in_ucast_pkts': '22247984409181',
                                                                       'links': [       {       'in_nucast_pkts': '22490735',
                                                                                                'in_octets': '6076813659711',
                                                                                                'in_ucast_pkts': '22247984409181',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21414',
                                                                                                'out_octets': '271835533402',
                                                                                                'out_ucast_pkts': '1310311464',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21414',
                                                                       'out_octets': '271835533402',
                                                                       'out_ucast_pkts': '1310311464',
                                                                       'poller_hash': 'MTMwM0VwZWVyaW50ZnBvbiAxNjY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166',
                                                                                               'links': [       {       'connected_time': '11964494',
                                                                                                                        'fiber_length': 19794,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwM0VwZWVycG9ucG9uIDE2Ng==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712',
                                         'name': '4921 Perth Road 119 (Stephen '
                                                 'Weadick)',
                                         'poller_hash': 'MTMwM0U=',
                                         'serial_number': '1303E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:11:ef:dc',
                                         'uptime': '11964494'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '167/f1',
                                                                       'name': '167/f1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTMwNDBwZWVyaW50ZjE2Ny9mMQ=='},
                                                               {       'admin_up': False,
                                                                       'id': '167',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTMwNDBwZWVyaW50ZnBvbiAxNjc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTMwNDBwZWVycG9ucG9uIDE2Nw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '4894 Linhaven Dr. Gadshill (Larry '
                                                 'Commerford)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "167" '
                                                                  '(location: GPON '
                                                                  'port "5/2"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:27:55.100Z',
                                         'poller_hash': 'MTMwNDA=',
                                         'serial_number': '13040',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '168/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '0c:9d:92:ce:56:f0'],
                                                                       'id': '168/g1',
                                                                       'in_nucast_pkts': '564',
                                                                       'in_octets': '879707124536',
                                                                       'in_ucast_pkts': '2338869588',
                                                                       'mac_address': '00:06:31:15:a9:55',
                                                                       'mtu': 1608,
                                                                       'name': '168/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11126872',
                                                                       'out_octets': '5643980433835',
                                                                       'out_ucast_pkts': '4583377931',
                                                                       'poller_hash': 'M0QwRkJwZWVyaW50ZjE2OC9nMQ==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '168',
                                                                       'in_nucast_pkts': '11126872',
                                                                       'in_octets': '5643980433835',
                                                                       'in_ucast_pkts': '4583377931',
                                                                       'links': [       {       'in_nucast_pkts': '11126872',
                                                                                                'in_octets': '5643980433835',
                                                                                                'in_ucast_pkts': '4583377931',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '564',
                                                                                                'out_octets': '879707124536',
                                                                                                'out_ucast_pkts': '2338869588',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '564',
                                                                       'out_octets': '879707124536',
                                                                       'out_ucast_pkts': '2338869588',
                                                                       'poller_hash': 'M0QwRkJwZWVyaW50ZnBvbiAxNjg=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168',
                                                                                               'links': [       {       'connected_time': '16927246',
                                                                                                                        'fiber_length': 19832,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -28.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'M0QwRkJwZWVycG9ucG9uIDE2OA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '712GX',
                                         'name': '4927 Perth Road 119 (Ivan '
                                                 'Messerschmidt)',
                                         'poller_hash': 'M0QwRkI=',
                                         'serial_number': '3D0FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:15:a9:55',
                                         'uptime': '16927246'},
                                 {       'counter_source': '169/f1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:0a:08'],
                                                                       'id': '169/f1',
                                                                       'in_nucast_pkts': '3233865',
                                                                       'in_octets': '118148671324',
                                                                       'in_ucast_pkts': '1584259447',
                                                                       'mac_address': '00:06:31:08:49:ed',
                                                                       'mtu': 1608,
                                                                       'name': '169/f1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '121625379',
                                                                       'out_octets': '6518939503115',
                                                                       'out_ucast_pkts': '11429017368876',
                                                                       'poller_hash': 'RjBGRXBlZXJpbnRmMTY5L2Yx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '169',
                                                                       'in_nucast_pkts': '121625379',
                                                                       'in_octets': '6518939503115',
                                                                       'in_ucast_pkts': '11429017368876',
                                                                       'links': [       {       'in_nucast_pkts': '121625379',
                                                                                                'in_octets': '6518939503115',
                                                                                                'in_ucast_pkts': '11429017368876',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3233865',
                                                                                                'out_octets': '118148671324',
                                                                                                'out_ucast_pkts': '1584259447',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3233865',
                                                                       'out_octets': '118148671324',
                                                                       'out_ucast_pkts': '1584259447',
                                                                       'poller_hash': 'RjBGRXBlZXJpbnRmcG9uIDE2OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '169',
                                                                                               'links': [       {       'connected_time': '30753817',
                                                                                                                        'fiber_length': 19438,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RjBGRXBlZXJwb25wb24gMTY5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '710',
                                         'name': '4910 Perth Road 119 (John Dunsmore)',
                                         'poller_hash': 'RjBGRQ==',
                                         'serial_number': 'F0FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '00:06:31:08:49:ed',
                                         'uptime': '30753817'},
                                 {       'counter_source': '170/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f4:1e:57:18:67:ed'],
                                                                       'id': '170/g1',
                                                                       'in_nucast_pkts': '1646',
                                                                       'in_octets': '63930674781',
                                                                       'in_ucast_pkts': '97322056',
                                                                       'mac_address': 'cc:be:59:5a:ff:9e',
                                                                       'mtu': 2000,
                                                                       'name': '170/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '917905',
                                                                       'out_octets': '105028091895',
                                                                       'out_ucast_pkts': '125032055',
                                                                       'poller_hash': 'NDRBQUI2cGVlcmludGYxNzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '170',
                                                                       'in_nucast_pkts': '917905',
                                                                       'in_octets': '105028091895',
                                                                       'in_ucast_pkts': '125032055',
                                                                       'links': [       {       'in_nucast_pkts': '917905',
                                                                                                'in_octets': '105028091895',
                                                                                                'in_ucast_pkts': '125032055',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1646',
                                                                                                'out_octets': '63930674781',
                                                                                                'out_ucast_pkts': '97322056',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1646',
                                                                       'out_octets': '63930674781',
                                                                       'out_ucast_pkts': '97322056',
                                                                       'poller_hash': 'NDRBQUI2cGVlcmludGZwb24gMTcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '170',
                                                                                               'links': [       {       'connected_time': '1372029',
                                                                                                                        'fiber_length': 22072,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.822,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDRBQUI2cGVlcnBvbnBvbiAxNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '1426 Hutchison Rd Wellesley (Jantzi '
                                                 'Plumbing Inc 10430891)',
                                         'poller_hash': 'NDRBQUI2',
                                         'serial_number': '44AAB6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:5a:ff:9a',
                                         'uptime': '1372029'},
                                 {       'counter_source': '172/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e2:46'],
                                                                       'id': '172/g1',
                                                                       'in_nucast_pkts': '156787',
                                                                       'in_octets': '8183751538',
                                                                       'in_ucast_pkts': '19730132',
                                                                       'mac_address': 'ec:4f:82:8d:f5:5b',
                                                                       'mtu': 2000,
                                                                       'name': '172/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9693795',
                                                                       'out_octets': '54596088812',
                                                                       'out_ucast_pkts': '47878708',
                                                                       'poller_hash': 'MzIyNkYycGVlcmludGYxNzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '172',
                                                                       'in_nucast_pkts': '9693795',
                                                                       'in_octets': '54596088812',
                                                                       'in_ucast_pkts': '47878708',
                                                                       'links': [       {       'in_nucast_pkts': '9693795',
                                                                                                'in_octets': '54596088812',
                                                                                                'in_ucast_pkts': '47878708',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '156787',
                                                                                                'out_octets': '8183751538',
                                                                                                'out_ucast_pkts': '19730132',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '156787',
                                                                       'out_octets': '8183751538',
                                                                       'out_ucast_pkts': '19730132',
                                                                       'poller_hash': 'MzIyNkYycGVlcmludGZwb24gMTcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '172',
                                                                                               'links': [       {       'connected_time': '8378181',
                                                                                                                        'fiber_length': 22978,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNkYycGVlcnBvbnBvbiAxNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 POND VIEW DR (PETERS DIANE & '
                                                 'GARRY)',
                                         'poller_hash': 'MzIyNkYy',
                                         'serial_number': '3226F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:f5:58',
                                         'uptime': '8378181'},
                                 {       'counter_source': '173/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b5:4e'],
                                                                       'id': '173/g1',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '76671518660',
                                                                       'in_ucast_pkts': '479320748',
                                                                       'mac_address': 'ec:4f:82:8d:f7:a3',
                                                                       'mtu': 2000,
                                                                       'name': '173/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9753903',
                                                                       'out_octets': '2444218624494',
                                                                       'out_ucast_pkts': '1744741914',
                                                                       'poller_hash': 'MzIyNzNCcGVlcmludGYxNzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '173',
                                                                       'in_nucast_pkts': '9753903',
                                                                       'in_octets': '2444218624494',
                                                                       'in_ucast_pkts': '1744741914',
                                                                       'links': [       {       'in_nucast_pkts': '9753903',
                                                                                                'in_octets': '2444218624494',
                                                                                                'in_ucast_pkts': '1744741914',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23',
                                                                                                'out_octets': '76671518660',
                                                                                                'out_ucast_pkts': '479320748',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '76671518660',
                                                                       'out_ucast_pkts': '479320748',
                                                                       'poller_hash': 'MzIyNzNCcGVlcmludGZwb24gMTcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '173',
                                                                                               'links': [       {       'connected_time': '8394287',
                                                                                                                        'fiber_length': 22625,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNzNCcGVlcnBvbnBvbiAxNzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1352 Queen's Bush Rd Wellesley (Jeff "
                                                 'Moyer 10434297)',
                                         'poller_hash': 'MzIyNzNC',
                                         'serial_number': '32273B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:f7:a0',
                                         'uptime': '8394287'},
                                 {       'counter_source': '174/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b9:8f'],
                                                                       'id': '174/g1',
                                                                       'in_nucast_pkts': '17',
                                                                       'in_octets': '24075029122',
                                                                       'in_ucast_pkts': '182645458',
                                                                       'mac_address': 'ec:4f:82:8d:f7:13',
                                                                       'mtu': 2000,
                                                                       'name': '174/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5084112',
                                                                       'out_octets': '1138820894928',
                                                                       'out_ucast_pkts': '790960215',
                                                                       'poller_hash': 'MzIyNzI5cGVlcmludGYxNzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174',
                                                                       'in_nucast_pkts': '5084112',
                                                                       'in_octets': '1138820894928',
                                                                       'in_ucast_pkts': '790960215',
                                                                       'links': [       {       'in_nucast_pkts': '5084112',
                                                                                                'in_octets': '1138820894928',
                                                                                                'in_ucast_pkts': '790960215',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '17',
                                                                                                'out_octets': '24075029122',
                                                                                                'out_ucast_pkts': '182645458',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17',
                                                                       'out_octets': '24075029122',
                                                                       'out_ucast_pkts': '182645458',
                                                                       'poller_hash': 'MzIyNzI5cGVlcmludGZwb24gMTc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174',
                                                                                               'links': [       {       'connected_time': '4332898',
                                                                                                                        'fiber_length': 22226,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNzI5cGVlcnBvbnBvbiAxNzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1291 Queen's Bush Rd Wellesley (Dan "
                                                 'Gingrich 10434293)',
                                         'poller_hash': 'MzIyNzI5',
                                         'serial_number': '322729',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:f7:10',
                                         'uptime': '4332898'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '175/g1',
                                                                       'name': '175/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyNDExcGVlcmludGYxNzUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '175',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyNDExcGVlcmludGZwb24gMTc1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDExcGVlcnBvbnBvbiAxNzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '41 Pondview Dr',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "175" '
                                                                  '(location: GPON '
                                                                  'port "6/1"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-07-22T13:17:02.180Z',
                                         'poller_hash': 'MzIyNDEx',
                                         'serial_number': '322411',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '176/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '176/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '18717416172',
                                                                       'in_ucast_pkts': '30505315',
                                                                       'mac_address': 'ec:4f:82:8e:01:50',
                                                                       'name': '176/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '231',
                                                                       'out_octets': '211518474830',
                                                                       'out_ucast_pkts': '209976875',
                                                                       'poller_hash': 'MzIyODcxcGVlcmludGYxNzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '176',
                                                                       'in_nucast_pkts': '231',
                                                                       'in_octets': '211518474830',
                                                                       'in_ucast_pkts': '209976875',
                                                                       'links': [       {       'in_nucast_pkts': '231',
                                                                                                'in_octets': '211518474830',
                                                                                                'in_ucast_pkts': '209976875',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '18717416172',
                                                                                                'out_ucast_pkts': '30505315',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '18717416172',
                                                                       'out_ucast_pkts': '30505315',
                                                                       'poller_hash': 'MzIyODcxcGVlcmludGZwb24gMTc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '176',
                                                                                               'links': [       {       'connected_time': '15630595',
                                                                                                                        'fiber_length': 22859,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODcxcGVlcnBvbnBvbiAxNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '45 POND VIEW DR (DOTZERT DORIS)',
                                         'poller_hash': 'MzIyODcx',
                                         'serial_number': '322871',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:01:50',
                                         'uptime': '15630595'},
                                 {       'counter_source': '177/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:ba:dc'],
                                                                       'id': '177/g1',
                                                                       'in_nucast_pkts': '12471',
                                                                       'in_octets': '140178648831',
                                                                       'in_ucast_pkts': '868438891',
                                                                       'mac_address': 'ec:4f:82:8d:db:d3',
                                                                       'mtu': 2000,
                                                                       'name': '177/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710247',
                                                                       'out_octets': '6917769765984',
                                                                       'out_ucast_pkts': '4829110153',
                                                                       'poller_hash': 'MzIyM0MxcGVlcmludGYxNzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '177',
                                                                       'in_nucast_pkts': '61710247',
                                                                       'in_octets': '6917769765984',
                                                                       'in_ucast_pkts': '4829110153',
                                                                       'links': [       {       'in_nucast_pkts': '61710247',
                                                                                                'in_octets': '6917769765984',
                                                                                                'in_ucast_pkts': '4829110153',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12471',
                                                                                                'out_octets': '140178648831',
                                                                                                'out_ucast_pkts': '868438891',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12471',
                                                                       'out_octets': '140178648831',
                                                                       'out_ucast_pkts': '868438891',
                                                                       'poller_hash': 'MzIyM0MxcGVlcmludGZwb24gMTc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '177',
                                                                                               'links': [       {       'connected_time': '21887039',
                                                                                                                        'fiber_length': 22262,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyM0MxcGVlcnBvbnBvbiAxNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1303 Queen's Bush Rd Wellesley "
                                                 '(Karyn Shantz 10434342)',
                                         'poller_hash': 'MzIyM0Mx',
                                         'serial_number': '3223C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:db:d0',
                                         'uptime': '21887039'},
                                 {       'counter_source': '178/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b7:a0'],
                                                                       'id': '178/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '30931131669',
                                                                       'in_ucast_pkts': '173460800',
                                                                       'mac_address': 'ec:4f:82:8d:db:bb',
                                                                       'mtu': 2000,
                                                                       'name': '178/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4682244',
                                                                       'out_octets': '896628343833',
                                                                       'out_ucast_pkts': '635746232',
                                                                       'poller_hash': 'MzIyM0JFcGVlcmludGYxNzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '178',
                                                                       'in_nucast_pkts': '4682244',
                                                                       'in_octets': '896628343833',
                                                                       'in_ucast_pkts': '635746232',
                                                                       'links': [       {       'in_nucast_pkts': '4682244',
                                                                                                'in_octets': '896628343833',
                                                                                                'in_ucast_pkts': '635746232',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '30931131669',
                                                                                                'out_ucast_pkts': '173460800',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '30931131669',
                                                                       'out_ucast_pkts': '173460800',
                                                                       'poller_hash': 'MzIyM0JFcGVlcmludGZwb24gMTc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '178',
                                                                                               'links': [       {       'connected_time': '3981837',
                                                                                                                        'fiber_length': 22588,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -26.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyM0JFcGVlcnBvbnBvbiAxNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1358 Queen's Bush Rd Wellesley (Rick "
                                                 'Dietrich 10434312)',
                                         'poller_hash': 'MzIyM0JF',
                                         'serial_number': '3223BE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:db:b8',
                                         'uptime': '3981837'},
                                 {       'counter_source': '179/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:63:e8'],
                                                                       'id': '179/g1',
                                                                       'in_nucast_pkts': '3104',
                                                                       'in_octets': '11962536215',
                                                                       'in_ucast_pkts': '41588291',
                                                                       'mac_address': 'cc:be:59:cb:ac:7d',
                                                                       'mtu': 2000,
                                                                       'name': '179/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '199736',
                                                                       'out_octets': '185856584590',
                                                                       'out_ucast_pkts': '134999405',
                                                                       'poller_hash': 'NDJDNkU3cGVlcmludGYxNzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '179',
                                                                       'in_nucast_pkts': '199736',
                                                                       'in_octets': '185856584590',
                                                                       'in_ucast_pkts': '134999405',
                                                                       'links': [       {       'in_nucast_pkts': '199736',
                                                                                                'in_octets': '185856584590',
                                                                                                'in_ucast_pkts': '134999405',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3104',
                                                                                                'out_octets': '11962536215',
                                                                                                'out_ucast_pkts': '41588291',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3104',
                                                                       'out_octets': '11962536215',
                                                                       'out_ucast_pkts': '41588291',
                                                                       'poller_hash': 'NDJDNkU3cGVlcmludGZwb24gMTc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '179',
                                                                                               'links': [       {       'connected_time': '169055',
                                                                                                                        'fiber_length': 17375,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkU3cGVlcnBvbnBvbiAxNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5527 KING ST (KAUNTZ ROBERT & '
                                                 'PATRICIA)',
                                         'poller_hash': 'NDJDNkU3',
                                         'serial_number': '42C6E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ac:7a',
                                         'uptime': '169055'},
                                 {       'counter_source': '180/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:01:21:2a:eb:dd'],
                                                                       'id': '180/g1',
                                                                       'in_nucast_pkts': '213676',
                                                                       'in_octets': '139800279181',
                                                                       'in_ucast_pkts': '655816211',
                                                                       'mac_address': 'd0:76:8f:aa:49:cf',
                                                                       'mtu': 2000,
                                                                       'name': '180/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61014332',
                                                                       'out_octets': '1073213320231',
                                                                       'out_ucast_pkts': '1176075356',
                                                                       'poller_hash': 'OTU2RTVGcGVlcmludGYxODAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '180',
                                                                       'in_nucast_pkts': '61014332',
                                                                       'in_octets': '1073213320231',
                                                                       'in_ucast_pkts': '1176075356',
                                                                       'links': [       {       'in_nucast_pkts': '61014332',
                                                                                                'in_octets': '1073213320231',
                                                                                                'in_ucast_pkts': '1176075356',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '213676',
                                                                                                'out_octets': '139800279181',
                                                                                                'out_ucast_pkts': '655816211',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '213676',
                                                                       'out_octets': '139800279181',
                                                                       'out_ucast_pkts': '655816211',
                                                                       'poller_hash': 'OTU2RTVGcGVlcmludGZwb24gMTgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180',
                                                                                               'links': [       {       'connected_time': '21887032',
                                                                                                                        'fiber_length': 22750,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTVGcGVlcnBvbnBvbiAxODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1379 QUEEN'S BUSH RD (TOWNSHIP OF "
                                                 'WELLESLEY)',
                                         'poller_hash': 'OTU2RTVG',
                                         'serial_number': '956E5F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:49:cc',
                                         'uptime': '21887032'},
                                 {       'counter_source': '181/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:14:2a',
                                                                                                        'd0:76:8f:02:f4:09'],
                                                                       'id': '181/g1',
                                                                       'in_nucast_pkts': '764612',
                                                                       'in_octets': '36502713383',
                                                                       'in_ucast_pkts': '220964306',
                                                                       'mac_address': 'd0:76:8f:08:15:f6',
                                                                       'mtu': 2000,
                                                                       'name': '181/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8526883',
                                                                       'out_octets': '631334332830',
                                                                       'out_ucast_pkts': '467872460',
                                                                       'poller_hash': 'ODY5M0I0cGVlcmludGYxODEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '181',
                                                                       'in_nucast_pkts': '8526883',
                                                                       'in_octets': '631334332830',
                                                                       'in_ucast_pkts': '467872460',
                                                                       'links': [       {       'in_nucast_pkts': '8526883',
                                                                                                'in_octets': '631334332830',
                                                                                                'in_ucast_pkts': '467872460',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '764612',
                                                                                                'out_octets': '36502713383',
                                                                                                'out_ucast_pkts': '220964306',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '764612',
                                                                       'out_octets': '36502713383',
                                                                       'out_ucast_pkts': '220964306',
                                                                       'poller_hash': 'ODY5M0I0cGVlcmludGZwb24gMTgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '181',
                                                                                               'links': [       {       'connected_time': '7183310',
                                                                                                                        'fiber_length': 23896,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0I0cGVlcnBvbnBvbiAxODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3706 NAFZIGER UNIT 6 (MCCL WELLESLEY '
                                                 'STORE)',
                                         'poller_hash': 'ODY5M0I0',
                                         'serial_number': '8693B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:f3',
                                         'uptime': '7183310'},
                                 {       'counter_source': '182/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '182/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '204866431802',
                                                                       'in_ucast_pkts': '858918043',
                                                                       'mac_address': 'ec:4f:82:8d:f3:30',
                                                                       'name': '182/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '231',
                                                                       'out_octets': '3102496969858',
                                                                       'out_ucast_pkts': '2316335196',
                                                                       'poller_hash': 'MzIyNkFEcGVlcmludGYxODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '182',
                                                                       'in_nucast_pkts': '231',
                                                                       'in_octets': '3102496969858',
                                                                       'in_ucast_pkts': '2316335196',
                                                                       'links': [       {       'in_nucast_pkts': '231',
                                                                                                'in_octets': '3102496969858',
                                                                                                'in_ucast_pkts': '2316335196',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '204866431802',
                                                                                                'out_ucast_pkts': '858918043',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '204866431802',
                                                                       'out_ucast_pkts': '858918043',
                                                                       'poller_hash': 'MzIyNkFEcGVlcmludGZwb24gMTgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '182',
                                                                                               'links': [       {       'connected_time': '8027224',
                                                                                                                        'fiber_length': 22236,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNkFEcGVlcnBvbnBvbiAxODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '31 David St Wellesley (G & S Poole '
                                                 '10434325)',
                                         'poller_hash': 'MzIyNkFE',
                                         'serial_number': '3226AD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:f3:30',
                                         'uptime': '8027224'},
                                 {       'counter_source': '183/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:ba:16'],
                                                                       'id': '183/g1',
                                                                       'in_nucast_pkts': '197',
                                                                       'in_octets': '196687169364',
                                                                       'in_ucast_pkts': '642002235',
                                                                       'mac_address': 'ec:4f:82:8d:fe:8b',
                                                                       'mtu': 2000,
                                                                       'name': '183/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '44315403',
                                                                       'out_octets': '3659640806126',
                                                                       'out_ucast_pkts': '2781507231',
                                                                       'poller_hash': 'MzIyODE4cGVlcmludGYxODMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '183',
                                                                       'in_nucast_pkts': '44315403',
                                                                       'in_octets': '3659640806126',
                                                                       'in_ucast_pkts': '2781507231',
                                                                       'links': [       {       'in_nucast_pkts': '44315403',
                                                                                                'in_octets': '3659640806126',
                                                                                                'in_ucast_pkts': '2781507231',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '197',
                                                                                                'out_octets': '196687169364',
                                                                                                'out_ucast_pkts': '642002235',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '197',
                                                                       'out_octets': '196687169364',
                                                                       'out_ucast_pkts': '642002235',
                                                                       'poller_hash': 'MzIyODE4cGVlcmludGZwb24gMTgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '183',
                                                                                               'links': [       {       'connected_time': '73454773',
                                                                                                                        'fiber_length': 22361,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODE4cGVlcnBvbnBvbiAxODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1287 QUEEN'S BUSH RD (GLAUSER TED)",
                                         'poller_hash': 'MzIyODE4',
                                         'serial_number': '322818',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fe:88',
                                         'uptime': '73454773'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '184/g1',
                                                                       'name': '184/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyODM1cGVlcmludGYxODQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '184',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyODM1cGVlcmludGZwb24gMTg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODM1cGVlcnBvbnBvbiAxODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '40 David St Wellesley',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "184" '
                                                                  '(location: GPON '
                                                                  'port "6/1"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:12.940Z',
                                         'poller_hash': 'MzIyODM1',
                                         'serial_number': '322835',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '185/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b8:a5'],
                                                                       'id': '185/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '2813619178',
                                                                       'in_ucast_pkts': '27340675',
                                                                       'mac_address': 'ec:4f:82:8e:01:9b',
                                                                       'mtu': 2000,
                                                                       'name': '185/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '993968',
                                                                       'out_octets': '159985846801',
                                                                       'out_ucast_pkts': '108260387',
                                                                       'poller_hash': 'MzIyODdBcGVlcmludGYxODUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '185',
                                                                       'in_nucast_pkts': '993968',
                                                                       'in_octets': '159985846801',
                                                                       'in_ucast_pkts': '108260387',
                                                                       'links': [       {       'in_nucast_pkts': '993968',
                                                                                                'in_octets': '159985846801',
                                                                                                'in_ucast_pkts': '108260387',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '2813619178',
                                                                                                'out_ucast_pkts': '27340675',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '2813619178',
                                                                       'out_ucast_pkts': '27340675',
                                                                       'poller_hash': 'MzIyODdBcGVlcmludGZwb24gMTg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '185',
                                                                                               'links': [       {       'connected_time': '847640',
                                                                                                                        'fiber_length': 22988,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODdBcGVlcnBvbnBvbiAxODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '115 David St Wellesley (Arnold & '
                                                 'Arnetta Erb 10402642)',
                                         'poller_hash': 'MzIyODdB',
                                         'serial_number': '32287A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:01:98',
                                         'uptime': '847640'},
                                 {       'counter_source': '186/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:d4:c4:c3:1b:a8'],
                                                                       'id': '186/g1',
                                                                       'in_nucast_pkts': '8768',
                                                                       'in_octets': '158047662204',
                                                                       'in_ucast_pkts': '546301620',
                                                                       'mac_address': 'ec:4f:82:7e:0a:75',
                                                                       'mtu': 2000,
                                                                       'name': '186/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39519357',
                                                                       'out_octets': '1241963606079',
                                                                       'out_ucast_pkts': '1118417924',
                                                                       'poller_hash': 'MzVFOUM3cGVlcmludGYxODYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '186',
                                                                       'in_nucast_pkts': '39519357',
                                                                       'in_octets': '1241963606079',
                                                                       'in_ucast_pkts': '1118417924',
                                                                       'links': [       {       'in_nucast_pkts': '39519357',
                                                                                                'in_octets': '1241963606079',
                                                                                                'in_ucast_pkts': '1118417924',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8768',
                                                                                                'out_octets': '158047662204',
                                                                                                'out_ucast_pkts': '546301620',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8768',
                                                                       'out_octets': '158047662204',
                                                                       'out_ucast_pkts': '546301620',
                                                                       'poller_hash': 'MzVFOUM3cGVlcmludGZwb24gMTg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186',
                                                                                               'links': [       {       'connected_time': '67856203',
                                                                                                                        'fiber_length': 23155,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOUM3cGVlcnBvbnBvbiAxODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '120 David St Wellesley (Larry Roes, '
                                                 '10411398)',
                                         'poller_hash': 'MzVFOUM3',
                                         'serial_number': '35E9C7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:0a:75',
                                         'uptime': '67856203'},
                                 {       'counter_source': '187/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ef:25'],
                                                                       'id': '187/g1',
                                                                       'in_nucast_pkts': '150561',
                                                                       'in_octets': '1034414186643',
                                                                       'in_ucast_pkts': '975813587',
                                                                       'mac_address': 'cc:be:59:ab:33:a1',
                                                                       'mtu': 2000,
                                                                       'name': '187/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652379',
                                                                       'out_octets': '752923341203',
                                                                       'out_ucast_pkts': '1006442707',
                                                                       'poller_hash': 'NEY5MUU5cGVlcmludGYxODcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '187',
                                                                       'in_nucast_pkts': '5652379',
                                                                       'in_octets': '752923341203',
                                                                       'in_ucast_pkts': '1006442707',
                                                                       'links': [       {       'in_nucast_pkts': '5652379',
                                                                                                'in_octets': '752923341203',
                                                                                                'in_ucast_pkts': '1006442707',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '150561',
                                                                                                'out_octets': '1034414186643',
                                                                                                'out_ucast_pkts': '975813587',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150561',
                                                                       'out_octets': '1034414186643',
                                                                       'out_ucast_pkts': '975813587',
                                                                       'poller_hash': 'NEY5MUU5cGVlcmludGZwb24gMTg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '187',
                                                                                               'links': [       {       'connected_time': '8535152',
                                                                                                                        'fiber_length': 22674,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -24.262,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.304}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5MUU5cGVlcnBvbnBvbiAxODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '40 Pond View Dr (Sheri Zupko, '
                                                 '10434413)',
                                         'poller_hash': 'NEY5MUU5',
                                         'serial_number': '4F91E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:33:a0',
                                         'uptime': '8535152'},
                                 {       'counter_source': '188/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:be:7e'],
                                                                       'id': '188/g1',
                                                                       'in_nucast_pkts': '2129759',
                                                                       'in_octets': '58707596057',
                                                                       'in_ucast_pkts': '188018884',
                                                                       'mac_address': 'ec:4f:82:8e:0b:43',
                                                                       'mtu': 2000,
                                                                       'name': '188/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4442741401',
                                                                       'out_octets': '6462154229915',
                                                                       'out_ucast_pkts': '381881567',
                                                                       'poller_hash': 'MzIyOUFGcGVlcmludGYxODgvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '188',
                                                                       'in_nucast_pkts': '4442741401',
                                                                       'in_octets': '6462154229915',
                                                                       'in_ucast_pkts': '381881567',
                                                                       'links': [       {       'in_nucast_pkts': '4442741401',
                                                                                                'in_octets': '6462154229915',
                                                                                                'in_ucast_pkts': '381881567',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2129759',
                                                                                                'out_octets': '58707596057',
                                                                                                'out_ucast_pkts': '188018884',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2129759',
                                                                       'out_octets': '58707596057',
                                                                       'out_ucast_pkts': '188018884',
                                                                       'poller_hash': 'MzIyOUFGcGVlcmludGZwb24gMTg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '188',
                                                                                               'links': [       {       'connected_time': '8535254',
                                                                                                                        'fiber_length': 22942,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOUFGcGVlcnBvbnBvbiAxODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '18 POND VIEW DR (MULLEN THOMAS)',
                                         'poller_hash': 'MzIyOUFG',
                                         'serial_number': '3229AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:0b:40',
                                         'uptime': '8535254'},
                                 {       'counter_source': '189/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:7b:eb:ec:5e:b6'],
                                                                       'id': '189/g1',
                                                                       'in_nucast_pkts': '57997',
                                                                       'in_octets': '1047449945795',
                                                                       'in_ucast_pkts': '1985401029',
                                                                       'mac_address': 'ec:4f:82:8d:de:bb',
                                                                       'mtu': 2000,
                                                                       'name': '189/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61708964',
                                                                       'out_octets': '3576151707434',
                                                                       'out_ucast_pkts': '3249002723',
                                                                       'poller_hash': 'MzIyNDFFcGVlcmludGYxODkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '189',
                                                                       'in_nucast_pkts': '61708964',
                                                                       'in_octets': '3576151707434',
                                                                       'in_ucast_pkts': '3249002723',
                                                                       'links': [       {       'in_nucast_pkts': '61708964',
                                                                                                'in_octets': '3576151707434',
                                                                                                'in_ucast_pkts': '3249002723',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '57997',
                                                                                                'out_octets': '1047449945795',
                                                                                                'out_ucast_pkts': '1985401029',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '57997',
                                                                       'out_octets': '1047449945795',
                                                                       'out_ucast_pkts': '1985401029',
                                                                       'poller_hash': 'MzIyNDFFcGVlcmludGZwb24gMTg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '189',
                                                                                               'links': [       {       'connected_time': '21886444',
                                                                                                                        'fiber_length': 23339,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDFFcGVlcnBvbnBvbiAxODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '157 David St (Wellesley Men Chrch)',
                                         'poller_hash': 'MzIyNDFF',
                                         'serial_number': '32241E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:de:b8',
                                         'uptime': '21886444'},
                                 {       'counter_source': '190/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b2:b8'],
                                                                       'id': '190/g1',
                                                                       'in_nucast_pkts': '937512',
                                                                       'in_octets': '1212745293173',
                                                                       'in_ucast_pkts': '3713818693',
                                                                       'mac_address': '84:d3:43:7e:31:f2',
                                                                       'mtu': 2000,
                                                                       'name': '190/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '55649463',
                                                                       'out_octets': '17601237980155',
                                                                       'out_ucast_pkts': '13217070020',
                                                                       'poller_hash': 'QzgzQkE3cGVlcmludGYxOTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '190',
                                                                       'in_nucast_pkts': '55649463',
                                                                       'in_octets': '17601237980155',
                                                                       'in_ucast_pkts': '13217070020',
                                                                       'links': [       {       'in_nucast_pkts': '55649463',
                                                                                                'in_octets': '17601237980155',
                                                                                                'in_ucast_pkts': '13217070020',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '937512',
                                                                                                'out_octets': '1212745293173',
                                                                                                'out_ucast_pkts': '3713818693',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '937512',
                                                                       'out_octets': '1212745293173',
                                                                       'out_ucast_pkts': '3713818693',
                                                                       'poller_hash': 'QzgzQkE3cGVlcmludGZwb24gMTkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '190',
                                                                                               'links': [       {       'connected_time': '12240433',
                                                                                                                        'fiber_length': 22834,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkE3cGVlcnBvbnBvbiAxOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1006 Doering St (HURREN RICHARD & '
                                                 'KELLY)',
                                         'poller_hash': 'QzgzQkE3',
                                         'serial_number': 'C83BA7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:ef',
                                         'uptime': '12240433'},
                                 {       'counter_source': '191/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:d6:87'],
                                                                       'id': '191/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '39585049996',
                                                                       'in_ucast_pkts': '157881046',
                                                                       'mac_address': 'cc:be:59:ab:23:6d',
                                                                       'mtu': 2000,
                                                                       'name': '191/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652498',
                                                                       'out_octets': '617124826541',
                                                                       'out_ucast_pkts': '480592289',
                                                                       'poller_hash': 'NEY4REQycGVlcmludGYxOTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '191',
                                                                       'in_nucast_pkts': '5652498',
                                                                       'in_octets': '617124826541',
                                                                       'in_ucast_pkts': '480592289',
                                                                       'links': [       {       'in_nucast_pkts': '5652498',
                                                                                                'in_octets': '617124826541',
                                                                                                'in_ucast_pkts': '480592289',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '39585049996',
                                                                                                'out_ucast_pkts': '157881046',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '39585049996',
                                                                       'out_ucast_pkts': '157881046',
                                                                       'poller_hash': 'NEY4REQycGVlcmludGZwb24gMTkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '191',
                                                                                               'links': [       {       'connected_time': '8535105',
                                                                                                                        'fiber_length': 23553,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.694,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 0.982}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY4REQycGVlcnBvbnBvbiAxOTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '3889 Nafziger Rd (Kent Mccullough, '
                                                 '10434313)',
                                         'poller_hash': 'NEY4REQy',
                                         'serial_number': '4F8DD2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:23:6c',
                                         'uptime': '8535105'},
                                 {       'counter_source': '192/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b6:d1'],
                                                                       'id': '192/g1',
                                                                       'in_nucast_pkts': '11286',
                                                                       'in_octets': '516565464062',
                                                                       'in_ucast_pkts': '1709342745',
                                                                       'mac_address': 'ec:4f:82:8d:df:43',
                                                                       'mtu': 2000,
                                                                       'name': '192/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61709813',
                                                                       'out_octets': '4737596752077',
                                                                       'out_ucast_pkts': '3592941855',
                                                                       'poller_hash': 'MzIyNDJGcGVlcmludGYxOTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '192',
                                                                       'in_nucast_pkts': '61709813',
                                                                       'in_octets': '4737596752077',
                                                                       'in_ucast_pkts': '3592941855',
                                                                       'links': [       {       'in_nucast_pkts': '61709813',
                                                                                                'in_octets': '4737596752077',
                                                                                                'in_ucast_pkts': '3592941855',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11286',
                                                                                                'out_octets': '516565464062',
                                                                                                'out_ucast_pkts': '1709342745',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11286',
                                                                       'out_octets': '516565464062',
                                                                       'out_ucast_pkts': '1709342745',
                                                                       'poller_hash': 'MzIyNDJGcGVlcmludGZwb24gMTky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '192',
                                                                                               'links': [       {       'connected_time': '21886927',
                                                                                                                        'fiber_length': 22601,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDJGcGVlcnBvbnBvbiAxOTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1155 Henry St. Wellesley (Ron & '
                                                 'Sheila Cressman 10434337)',
                                         'poller_hash': 'MzIyNDJG',
                                         'serial_number': '32242F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:df:40',
                                         'uptime': '21886927'},
                                 {       'counter_source': '194/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b7:a9'],
                                                                       'id': '194/g1',
                                                                       'in_nucast_pkts': '62',
                                                                       'in_octets': '87344010055',
                                                                       'in_ucast_pkts': '758160614',
                                                                       'mac_address': 'ec:4f:82:8d:de:eb',
                                                                       'mtu': 2000,
                                                                       'name': '194/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4879293',
                                                                       'out_octets': '2502801864029',
                                                                       'out_ucast_pkts': '1747748417',
                                                                       'poller_hash': 'MzIyNDI0cGVlcmludGYxOTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '194',
                                                                       'in_nucast_pkts': '4879293',
                                                                       'in_octets': '2502801864029',
                                                                       'in_ucast_pkts': '1747748417',
                                                                       'links': [       {       'in_nucast_pkts': '4879293',
                                                                                                'in_octets': '2502801864029',
                                                                                                'in_ucast_pkts': '1747748417',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '62',
                                                                                                'out_octets': '87344010055',
                                                                                                'out_ucast_pkts': '758160614',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62',
                                                                       'out_octets': '87344010055',
                                                                       'out_ucast_pkts': '758160614',
                                                                       'poller_hash': 'MzIyNDI0cGVlcmludGZwb24gMTk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '194',
                                                                                               'links': [       {       'connected_time': '4155956',
                                                                                                                        'fiber_length': 22999,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDI0cGVlcnBvbnBvbiAxOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3609 Nafziger Rd Wellesley (Christie '
                                                 'Howes 10434371)',
                                         'poller_hash': 'MzIyNDI0',
                                         'serial_number': '322424',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:de:e8',
                                         'uptime': '4155956'},
                                 {       'counter_source': '195/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '195/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '477957294090',
                                                                       'in_ucast_pkts': '1292134835',
                                                                       'mac_address': 'cc:be:59:eb:02:66',
                                                                       'name': '195/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2495748245162',
                                                                       'out_ucast_pkts': '2238166978',
                                                                       'poller_hash': 'NTU1Q0MycGVlcmludGYxOTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '195',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2495748245162',
                                                                       'in_ucast_pkts': '2238166978',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2495748245162',
                                                                                                'in_ucast_pkts': '2238166978',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '477957294090',
                                                                                                'out_ucast_pkts': '1292134835',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '477957294090',
                                                                       'out_ucast_pkts': '1292134835',
                                                                       'poller_hash': 'NTU1Q0MycGVlcmludGZwb24gMTk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '195',
                                                                                               'links': [       {       'connected_time': '639563',
                                                                                                                        'fiber_length': 22625,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.776,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTU1Q0MycGVlcnBvbnBvbiAxOTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '3737 Nafziger Rd Wellesley '
                                                 '(Wellesley Auto Care 10434379)',
                                         'poller_hash': 'NTU1Q0My',
                                         'serial_number': '555CC2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:eb:02:66',
                                         'uptime': '639563'},
                                 {       'counter_source': '196/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '196/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '5139274',
                                                                       'in_octets': '92289770842',
                                                                       'in_ucast_pkts': '104891318',
                                                                       'mac_address': 'ec:4f:82:0d:69:7f',
                                                                       'name': '196/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '19363047799',
                                                                       'out_octets': '26155427122172',
                                                                       'out_ucast_pkts': '130330536',
                                                                       'poller_hash': 'MjA1RERBcGVlcmludGYxOTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '196',
                                                                       'in_nucast_pkts': '19363047799',
                                                                       'in_octets': '26155427122172',
                                                                       'in_ucast_pkts': '130330536',
                                                                       'links': [       {       'in_nucast_pkts': '19363047799',
                                                                                                'in_octets': '26155427122172',
                                                                                                'in_ucast_pkts': '130330536',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5139274',
                                                                                                'out_octets': '92289770842',
                                                                                                'out_ucast_pkts': '104891318',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5139274',
                                                                       'out_octets': '92289770842',
                                                                       'out_ucast_pkts': '104891318',
                                                                       'poller_hash': 'MjA1RERBcGVlcmludGZwb24gMTk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '196',
                                                                                               'links': [       {       'connected_time': '927589',
                                                                                                                        'fiber_length': 22350,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjA1RERBcGVlcnBvbnBvbiAxOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': "1235 QUEEN'S BUSH RD (JOHNSON VAL)",
                                         'poller_hash': 'MjA1RERB',
                                         'serial_number': '205DDA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:0d:69:7f',
                                         'uptime': '927589'},
                                 {       'counter_source': '197/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6a:1d'],
                                                                       'id': '197/g1',
                                                                       'in_nucast_pkts': '474871',
                                                                       'in_octets': '284556994238',
                                                                       'in_ucast_pkts': '763097918',
                                                                       'mac_address': 'ec:4f:82:8d:de:8b',
                                                                       'mtu': 2000,
                                                                       'name': '197/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30163648',
                                                                       'out_octets': '1632962298784',
                                                                       'out_ucast_pkts': '1489378667',
                                                                       'poller_hash': 'MzIyNDE4cGVlcmludGYxOTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '197',
                                                                       'in_nucast_pkts': '30163648',
                                                                       'in_octets': '1632962298784',
                                                                       'in_ucast_pkts': '1489378667',
                                                                       'links': [       {       'in_nucast_pkts': '30163648',
                                                                                                'in_octets': '1632962298784',
                                                                                                'in_ucast_pkts': '1489378667',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '474871',
                                                                                                'out_octets': '284556994238',
                                                                                                'out_ucast_pkts': '763097918',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '474871',
                                                                       'out_octets': '284556994238',
                                                                       'out_ucast_pkts': '763097918',
                                                                       'poller_hash': 'MzIyNDE4cGVlcmludGZwb24gMTk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '197',
                                                                                               'links': [       {       'connected_time': '26231005',
                                                                                                                        'fiber_length': 23332,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDE4cGVlcnBvbnBvbiAxOTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3840 Nafziger Rd-2 Wellesley (Barry '
                                                 '& Denise Bender 10434362)',
                                         'poller_hash': 'MzIyNDE4',
                                         'serial_number': '322418',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:de:88',
                                         'uptime': '26231005'},
                                 {       'counter_source': '198/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:d9:f5:5a:23:08'],
                                                                       'id': '198/g1',
                                                                       'in_nucast_pkts': '36414',
                                                                       'in_octets': '78990745098',
                                                                       'in_ucast_pkts': '311337624',
                                                                       'mac_address': 'cc:be:59:ef:bf:d2',
                                                                       'mtu': 2000,
                                                                       'name': '198/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24135150',
                                                                       'out_octets': '662293673537',
                                                                       'out_ucast_pkts': '581909552',
                                                                       'poller_hash': 'NEQzQjRDcGVlcmludGYxOTgvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '198',
                                                                       'in_nucast_pkts': '24135150',
                                                                       'in_octets': '662293673537',
                                                                       'in_ucast_pkts': '581909552',
                                                                       'links': [       {       'in_nucast_pkts': '24135150',
                                                                                                'in_octets': '662293673537',
                                                                                                'in_ucast_pkts': '581909552',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '36414',
                                                                                                'out_octets': '78990745098',
                                                                                                'out_ucast_pkts': '311337624',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36414',
                                                                       'out_octets': '78990745098',
                                                                       'out_ucast_pkts': '311337624',
                                                                       'poller_hash': 'NEQzQjRDcGVlcmludGZwb24gMTk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198',
                                                                                               'links': [       {       'connected_time': '551824',
                                                                                                                        'fiber_length': 22290,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.858,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQzQjRDcGVlcnBvbnBvbiAxOTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': "1235 Queen'sBush Rd (Erb Financial "
                                                 'Services 10434422)',
                                         'poller_hash': 'NEQzQjRD',
                                         'serial_number': '4D3B4C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:bf:ce',
                                         'uptime': '551824'},
                                 {       'counter_source': '200/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:ba:8b'],
                                                                       'id': '200/g1',
                                                                       'in_nucast_pkts': '12127',
                                                                       'in_octets': '258435090756',
                                                                       'in_ucast_pkts': '1552263108',
                                                                       'mac_address': 'ec:4f:82:8d:fe:db',
                                                                       'mtu': 2000,
                                                                       'name': '200/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710128',
                                                                       'out_octets': '8091766302272',
                                                                       'out_ucast_pkts': '5814867463',
                                                                       'poller_hash': 'MzIyODIycGVlcmludGYyMDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '200',
                                                                       'in_nucast_pkts': '61710128',
                                                                       'in_octets': '8091766302272',
                                                                       'in_ucast_pkts': '5814867463',
                                                                       'links': [       {       'in_nucast_pkts': '61710128',
                                                                                                'in_octets': '8091766302272',
                                                                                                'in_ucast_pkts': '5814867463',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12127',
                                                                                                'out_octets': '258435090756',
                                                                                                'out_ucast_pkts': '1552263108',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12127',
                                                                       'out_octets': '258435090756',
                                                                       'out_ucast_pkts': '1552263108',
                                                                       'poller_hash': 'MzIyODIycGVlcmludGZwb24gMjAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '200',
                                                                                               'links': [       {       'connected_time': '21886272',
                                                                                                                        'fiber_length': 22860,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODIycGVlcnBvbnBvbiAyMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3631 Nafziger Rd Wellesley (James & '
                                                 'Karen Gerber 10432639)',
                                         'poller_hash': 'MzIyODIy',
                                         'serial_number': '322822',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fe:d8',
                                         'uptime': '21886272'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '201/g1',
                                                                       'name': '201/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyOTEycGVlcmludGYyMDEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '201',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzIyOTEycGVlcmludGZwb24gMjAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '201',
                                                                                               'links': [       {       'connected_time': '21886755',
                                                                                                                        'fiber_length': 22840,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOTEycGVlcnBvbnBvbiAyMDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3640 NAFZIGER RD (MCCREA LIZ & JEFF)',
                                         'poller_hash': 'MzIyOTEy',
                                         'serial_number': '322912',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:06:58',
                                         'uptime': '21886755'},
                                 {       'counter_source': '202/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:47'],
                                                                       'id': '202/g1',
                                                                       'in_nucast_pkts': '164074',
                                                                       'in_octets': '88972112183',
                                                                       'in_ucast_pkts': '445504369',
                                                                       'mac_address': 'ec:4f:82:8d:fd:5b',
                                                                       'mtu': 2000,
                                                                       'name': '202/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9910981',
                                                                       'out_octets': '1579098041921',
                                                                       'out_ucast_pkts': '1170780019',
                                                                       'poller_hash': 'MzIyN0YycGVlcmludGYyMDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '202',
                                                                       'in_nucast_pkts': '9910981',
                                                                       'in_octets': '1579098041921',
                                                                       'in_ucast_pkts': '1170780019',
                                                                       'links': [       {       'in_nucast_pkts': '9910981',
                                                                                                'in_octets': '1579098041921',
                                                                                                'in_ucast_pkts': '1170780019',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '164074',
                                                                                                'out_octets': '88972112183',
                                                                                                'out_ucast_pkts': '445504369',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '164074',
                                                                       'out_octets': '88972112183',
                                                                       'out_ucast_pkts': '445504369',
                                                                       'poller_hash': 'MzIyN0YycGVlcmludGZwb24gMjAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202',
                                                                                               'links': [       {       'connected_time': '8551983',
                                                                                                                        'fiber_length': 22780,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyN0YycGVlcnBvbnBvbiAyMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1124 HENRY ST (BYER CAMERON & JAYNE)',
                                         'poller_hash': 'MzIyN0Yy',
                                         'serial_number': '3227F2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fd:58',
                                         'uptime': '8551983'},
                                 {       'counter_source': '204/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:70:1b'],
                                                                       'id': '204/g1',
                                                                       'in_nucast_pkts': '5572919',
                                                                       'in_octets': '275084848504',
                                                                       'in_ucast_pkts': '1439959507',
                                                                       'mac_address': 'ec:4f:82:8d:ff:a3',
                                                                       'mtu': 2000,
                                                                       'name': '204/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9904841',
                                                                       'out_octets': '6681588337177',
                                                                       'out_ucast_pkts': '4758070447',
                                                                       'poller_hash': 'MzIyODNCcGVlcmludGYyMDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '204',
                                                                       'in_nucast_pkts': '9904841',
                                                                       'in_octets': '6681588337177',
                                                                       'in_ucast_pkts': '4758070447',
                                                                       'links': [       {       'in_nucast_pkts': '9904841',
                                                                                                'in_octets': '6681588337177',
                                                                                                'in_ucast_pkts': '4758070447',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5572919',
                                                                                                'out_octets': '275084848504',
                                                                                                'out_ucast_pkts': '1439959507',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5572919',
                                                                       'out_octets': '275084848504',
                                                                       'out_ucast_pkts': '1439959507',
                                                                       'poller_hash': 'MzIyODNCcGVlcmludGZwb24gMjA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '204',
                                                                                               'links': [       {       'connected_time': '8535240',
                                                                                                                        'fiber_length': 22626,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODNCcGVlcnBvbnBvbiAyMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1346 Queen's Bush Rd Wellesley "
                                                 '(Sherry Cormier 10434322)',
                                         'poller_hash': 'MzIyODNC',
                                         'serial_number': '32283B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:ff:a0',
                                         'uptime': '8535240'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '205/g1',
                                                                       'name': '205/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyM0Q3cGVlcmludGYyMDUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzIyM0Q3cGVlcmludGZwb24gMjA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyM0Q3cGVlcnBvbnBvbiAyMDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '3775 Nafziger Rd (Ronald &Bonnie '
                                                 'Roth 10434363)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "205" '
                                                                  '(location: GPON '
                                                                  'port "6/2"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2025-06-28T01:20:52.390Z',
                                         'poller_hash': 'MzIyM0Q3',
                                         'serial_number': '3223D7',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '206/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b8:ff'],
                                                                       'id': '206/g1',
                                                                       'in_nucast_pkts': '6647',
                                                                       'in_octets': '101301132901',
                                                                       'in_ucast_pkts': '443159731',
                                                                       'mac_address': 'ec:4f:82:8e:03:d3',
                                                                       'mtu': 2000,
                                                                       'name': '206/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61709911',
                                                                       'out_octets': '1729394720975',
                                                                       'out_ucast_pkts': '1268147794',
                                                                       'poller_hash': 'MzIyOEMxcGVlcmludGYyMDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '206',
                                                                       'in_nucast_pkts': '61709911',
                                                                       'in_octets': '1729394720975',
                                                                       'in_ucast_pkts': '1268147794',
                                                                       'links': [       {       'in_nucast_pkts': '61709911',
                                                                                                'in_octets': '1729394720975',
                                                                                                'in_ucast_pkts': '1268147794',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6647',
                                                                                                'out_octets': '101301132901',
                                                                                                'out_ucast_pkts': '443159731',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6647',
                                                                       'out_octets': '101301132901',
                                                                       'out_ucast_pkts': '443159731',
                                                                       'poller_hash': 'MzIyOEMxcGVlcmludGZwb24gMjA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '206',
                                                                                               'links': [       {       'connected_time': '21886877',
                                                                                                                        'fiber_length': 22622,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOEMxcGVlcnBvbnBvbiAyMDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 LAWRENCE ST (ROBERTS ANN)',
                                         'poller_hash': 'MzIyOEMx',
                                         'serial_number': '3228C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:03:d0',
                                         'uptime': '21886877'},
                                 {       'counter_source': '207/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:33:c1:8d',
                                                                                                        '5c:0f:fb:33:c1:9a',
                                                                                                        'cc:be:59:7e:b6:80'],
                                                                       'id': '207/g1',
                                                                       'in_nucast_pkts': '7996168',
                                                                       'in_octets': '45767357846',
                                                                       'in_ucast_pkts': '503561460',
                                                                       'mac_address': 'ec:4f:82:8e:00:eb',
                                                                       'mtu': 2000,
                                                                       'name': '207/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '136512106',
                                                                       'out_octets': '3390280760948',
                                                                       'out_ucast_pkts': '2384726829',
                                                                       'poller_hash': 'MzIyODY0cGVlcmludGYyMDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '207',
                                                                       'in_nucast_pkts': '136512106',
                                                                       'in_octets': '3390280760948',
                                                                       'in_ucast_pkts': '2384726829',
                                                                       'links': [       {       'in_nucast_pkts': '136512106',
                                                                                                'in_octets': '3390280760948',
                                                                                                'in_ucast_pkts': '2384726829',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7996168',
                                                                                                'out_octets': '45767357846',
                                                                                                'out_ucast_pkts': '503561460',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7996168',
                                                                       'out_octets': '45767357846',
                                                                       'out_ucast_pkts': '503561460',
                                                                       'poller_hash': 'MzIyODY0cGVlcmludGZwb24gMjA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207',
                                                                                               'links': [       {       'connected_time': '17360321',
                                                                                                                        'fiber_length': 22958,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODY0cGVlcnBvbnBvbiAyMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1035 Doering St Wellesley (Paul Cook '
                                                 '10434450)',
                                         'poller_hash': 'MzIyODY0',
                                         'serial_number': '322864',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:00:e8',
                                         'uptime': '17360321'},
                                 {       'counter_source': '208/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:82:07:f6'],
                                                                       'id': '208/g1',
                                                                       'in_nucast_pkts': '168202',
                                                                       'in_octets': '420736113273',
                                                                       'in_ucast_pkts': '620332570',
                                                                       'mac_address': 'cc:be:59:b9:c3:b0',
                                                                       'mtu': 2000,
                                                                       'name': '208/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920032',
                                                                       'out_octets': '1706197280222',
                                                                       'out_ucast_pkts': '1456538770',
                                                                       'poller_hash': 'NTEyNTE4cGVlcmludGYyMDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '208',
                                                                       'in_nucast_pkts': '9920032',
                                                                       'in_octets': '1706197280222',
                                                                       'in_ucast_pkts': '1456538770',
                                                                       'links': [       {       'in_nucast_pkts': '9920032',
                                                                                                'in_octets': '1706197280222',
                                                                                                'in_ucast_pkts': '1456538770',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '168202',
                                                                                                'out_octets': '420736113273',
                                                                                                'out_ucast_pkts': '620332570',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '168202',
                                                                       'out_octets': '420736113273',
                                                                       'out_ucast_pkts': '620332570',
                                                                       'poller_hash': 'NTEyNTE4cGVlcmludGZwb24gMjA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208',
                                                                                               'links': [       {       'connected_time': '8535404',
                                                                                                                        'fiber_length': 23661,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTE4cGVlcnBvbnBvbiAyMDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '3894 Nafziger Rd Welleseley (Pete '
                                                 'Weber 10434296)',
                                         'poller_hash': 'NTEyNTE4',
                                         'serial_number': '512518',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:b0',
                                         'uptime': '8535404'},
                                 {       'counter_source': '209/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6c:59'],
                                                                       'id': '209/g1',
                                                                       'in_nucast_pkts': '657037',
                                                                       'in_octets': '196308862443',
                                                                       'in_ucast_pkts': '731258084',
                                                                       'mac_address': 'ec:4f:82:8d:f4:0b',
                                                                       'mtu': 2000,
                                                                       'name': '209/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39262323',
                                                                       'out_octets': '6674144852682',
                                                                       'out_ucast_pkts': '4896864188',
                                                                       'poller_hash': 'MzIyNkM4cGVlcmludGYyMDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '209',
                                                                       'in_nucast_pkts': '39262323',
                                                                       'in_octets': '6674144852682',
                                                                       'in_ucast_pkts': '4896864188',
                                                                       'links': [       {       'in_nucast_pkts': '39262323',
                                                                                                'in_octets': '6674144852682',
                                                                                                'in_ucast_pkts': '4896864188',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '657037',
                                                                                                'out_octets': '196308862443',
                                                                                                'out_ucast_pkts': '731258084',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '657037',
                                                                       'out_octets': '196308862443',
                                                                       'out_ucast_pkts': '731258084',
                                                                       'poller_hash': 'MzIyNkM4cGVlcmludGZwb24gMjA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '209',
                                                                                               'links': [       {       'connected_time': '67900401',
                                                                                                                        'fiber_length': 22858,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNkM4cGVlcnBvbnBvbiAyMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1147 HENRY ST (KRALL MELISSA)',
                                         'poller_hash': 'MzIyNkM4',
                                         'serial_number': '3226C8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:f4:08',
                                         'uptime': '67900401'},
                                 {       'counter_source': '211/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:e8:29:44:b5:13'],
                                                                       'id': '211/g1',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '17672146152989',
                                                                       'in_ucast_pkts': '16795600894',
                                                                       'mac_address': 'ec:4f:82:8d:ee:33',
                                                                       'mtu': 2000,
                                                                       'name': '211/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36122332',
                                                                       'out_octets': '22976172642282',
                                                                       'out_ucast_pkts': '23506041948',
                                                                       'poller_hash': 'MzIyNjBEcGVlcmludGYyMTEvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '211',
                                                                       'in_nucast_pkts': '36122332',
                                                                       'in_octets': '22976172642282',
                                                                       'in_ucast_pkts': '23506041948',
                                                                       'links': [       {       'in_nucast_pkts': '36122332',
                                                                                                'in_octets': '22976172642282',
                                                                                                'in_ucast_pkts': '23506041948',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '100',
                                                                                                'out_octets': '17672146152989',
                                                                                                'out_ucast_pkts': '16795600894',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '17672146152989',
                                                                       'out_ucast_pkts': '16795600894',
                                                                       'poller_hash': 'MzIyNjBEcGVlcmludGZwb24gMjEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '211',
                                                                                               'links': [       {       'connected_time': '64119657',
                                                                                                                        'fiber_length': 23367,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNjBEcGVlcnBvbnBvbiAyMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '25 Zoeger Ct Wellesley (David '
                                                 'Henderson 10434421)',
                                         'poller_hash': 'MzIyNjBE',
                                         'serial_number': '32260D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:ee:30',
                                         'uptime': '64119657'},
                                 {       'counter_source': '212/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b6:2f'],
                                                                       'id': '212/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '69733650661',
                                                                       'in_ucast_pkts': '516856609',
                                                                       'mac_address': 'cc:be:59:ca:df:6d',
                                                                       'mtu': 2000,
                                                                       'name': '212/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919812',
                                                                       'out_octets': '1610547735454',
                                                                       'out_ucast_pkts': '1174935723',
                                                                       'poller_hash': 'NDI4MUVDcGVlcmludGYyMTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '212',
                                                                       'in_nucast_pkts': '9919812',
                                                                       'in_octets': '1610547735454',
                                                                       'in_ucast_pkts': '1174935723',
                                                                       'links': [       {       'in_nucast_pkts': '9919812',
                                                                                                'in_octets': '1610547735454',
                                                                                                'in_ucast_pkts': '1174935723',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '69733650661',
                                                                                                'out_ucast_pkts': '516856609',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '69733650661',
                                                                       'out_ucast_pkts': '516856609',
                                                                       'poller_hash': 'NDI4MUVDcGVlcmludGZwb24gMjEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212',
                                                                                               'links': [       {       'connected_time': '8535206',
                                                                                                                        'fiber_length': 23348,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUVDcGVlcnBvbnBvbiAyMTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '49 Zoeger Cr Wellesley (Matthew '
                                                 'Harney 10434400)',
                                         'poller_hash': 'NDI4MUVD',
                                         'serial_number': '4281EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:df:6a',
                                         'uptime': '8535206'},
                                 {       'counter_source': '213/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b6:da'],
                                                                       'id': '213/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '59606836603',
                                                                       'in_ucast_pkts': '438079941',
                                                                       'mac_address': 'ec:4f:82:8d:de:b3',
                                                                       'mtu': 2000,
                                                                       'name': '213/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920282',
                                                                       'out_octets': '1292521846620',
                                                                       'out_ucast_pkts': '901295252',
                                                                       'poller_hash': 'MzIyNDFEcGVlcmludGYyMTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '213',
                                                                       'in_nucast_pkts': '9920282',
                                                                       'in_octets': '1292521846620',
                                                                       'in_ucast_pkts': '901295252',
                                                                       'links': [       {       'in_nucast_pkts': '9920282',
                                                                                                'in_octets': '1292521846620',
                                                                                                'in_ucast_pkts': '901295252',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '59606836603',
                                                                                                'out_ucast_pkts': '438079941',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '59606836603',
                                                                       'out_ucast_pkts': '438079941',
                                                                       'poller_hash': 'MzIyNDFEcGVlcmludGZwb24gMjEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '213',
                                                                                               'links': [       {       'connected_time': '8535316',
                                                                                                                        'fiber_length': 23391,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNDFEcGVlcnBvbnBvbiAyMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 ZOEGER CT (MARTIN RAY)',
                                         'poller_hash': 'MzIyNDFE',
                                         'serial_number': '32241D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:de:b0',
                                         'uptime': '8535316'},
                                 {       'counter_source': '214/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '74:83:c2:70:8d:f1'],
                                                                       'id': '214/g1',
                                                                       'in_nucast_pkts': '142',
                                                                       'in_octets': '124891031684',
                                                                       'in_ucast_pkts': '339538903',
                                                                       'mac_address': 'cc:be:59:ab:20:e1',
                                                                       'mtu': 2000,
                                                                       'name': '214/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10070347',
                                                                       'out_octets': '1748459828314',
                                                                       'out_ucast_pkts': '1425895022',
                                                                       'poller_hash': 'NEY4Qzk5cGVlcmludGYyMTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '214',
                                                                       'in_nucast_pkts': '10070347',
                                                                       'in_octets': '1748459828314',
                                                                       'in_ucast_pkts': '1425895022',
                                                                       'links': [       {       'in_nucast_pkts': '10070347',
                                                                                                'in_octets': '1748459828314',
                                                                                                'in_ucast_pkts': '1425895022',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '142',
                                                                                                'out_octets': '124891031684',
                                                                                                'out_ucast_pkts': '339538903',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '142',
                                                                       'out_octets': '124891031684',
                                                                       'out_ucast_pkts': '339538903',
                                                                       'poller_hash': 'NEY4Qzk5cGVlcmludGZwb24gMjE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '214',
                                                                                               'links': [       {       'connected_time': '15182469',
                                                                                                                        'fiber_length': 23159,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.626,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.252}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY4Qzk5cGVlcnBvbnBvbiAyMTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '41 Zoeger Ct Wellesley (David Green '
                                                 '10434462)',
                                         'poller_hash': 'NEY4Qzk5',
                                         'serial_number': '4F8C99',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:20:e0',
                                         'uptime': '15182469'},
                                 {       'counter_source': '215/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b5:9f'],
                                                                       'id': '215/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '479586636223',
                                                                       'in_ucast_pkts': '3861272606',
                                                                       'mac_address': 'ec:4f:82:8d:ed:cb',
                                                                       'mtu': 2000,
                                                                       'name': '215/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919966',
                                                                       'out_octets': '15903468478317',
                                                                       'out_ucast_pkts': '10938427492',
                                                                       'poller_hash': 'MzIyNjAwcGVlcmludGYyMTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '215',
                                                                       'in_nucast_pkts': '9919966',
                                                                       'in_octets': '15903468478317',
                                                                       'in_ucast_pkts': '10938427492',
                                                                       'links': [       {       'in_nucast_pkts': '9919966',
                                                                                                'in_octets': '15903468478317',
                                                                                                'in_ucast_pkts': '10938427492',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '479586636223',
                                                                                                'out_ucast_pkts': '3861272606',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '479586636223',
                                                                       'out_ucast_pkts': '3861272606',
                                                                       'poller_hash': 'MzIyNjAwcGVlcmludGZwb24gMjE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215',
                                                                                               'links': [       {       'connected_time': '8535239',
                                                                                                                        'fiber_length': 22267,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNjAwcGVlcnBvbnBvbiAyMTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1300 Queen's Bush Rd Wellesley (Bill "
                                                 '& Bonnie Love 10434529)',
                                         'poller_hash': 'MzIyNjAw',
                                         'serial_number': '322600',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:ed:c8',
                                         'uptime': '8535239'},
                                 {       'counter_source': '216/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '216/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '380631709408',
                                                                       'in_ucast_pkts': '1418862834',
                                                                       'mac_address': 'ec:4f:82:8d:fa:80',
                                                                       'name': '216/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '7127115027448',
                                                                       'out_ucast_pkts': '5285953768',
                                                                       'poller_hash': 'MzIyNzk3cGVlcmludGYyMTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '216',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '7127115027448',
                                                                       'in_ucast_pkts': '5285953768',
                                                                       'links': [       {       'in_nucast_pkts': '66',
                                                                                                'in_octets': '7127115027448',
                                                                                                'in_ucast_pkts': '5285953768',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '380631709408',
                                                                                                'out_ucast_pkts': '1418862834',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '380631709408',
                                                                       'out_ucast_pkts': '1418862834',
                                                                       'poller_hash': 'MzIyNzk3cGVlcmludGZwb24gMjE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216',
                                                                                               'links': [       {       'connected_time': '8535151',
                                                                                                                        'fiber_length': 23678,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNzk3cGVlcnBvbnBvbiAyMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '61 ZINKANN CR (RYERSE PATRICK)',
                                         'poller_hash': 'MzIyNzk3',
                                         'serial_number': '322797',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:fa:80',
                                         'uptime': '8535151'},
                                 {       'counter_source': '217/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:d2:a0'],
                                                                       'id': '217/g1',
                                                                       'in_nucast_pkts': '32543862',
                                                                       'in_octets': '1690109241243',
                                                                       'in_ucast_pkts': '3547371711',
                                                                       'mac_address': 'ec:4f:82:8d:f3:db',
                                                                       'mtu': 2000,
                                                                       'name': '217/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43997775660',
                                                                       'out_octets': '71623326631909',
                                                                       'out_ucast_pkts': '9691899840',
                                                                       'poller_hash': 'MzIyNkMycGVlcmludGYyMTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '217',
                                                                       'in_nucast_pkts': '43997775660',
                                                                       'in_octets': '71623326631909',
                                                                       'in_ucast_pkts': '9691899840',
                                                                       'links': [       {       'in_nucast_pkts': '43997775660',
                                                                                                'in_octets': '71623326631909',
                                                                                                'in_ucast_pkts': '9691899840',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '32543862',
                                                                                                'out_octets': '1690109241243',
                                                                                                'out_ucast_pkts': '3547371711',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32543862',
                                                                       'out_octets': '1690109241243',
                                                                       'out_ucast_pkts': '3547371711',
                                                                       'poller_hash': 'MzIyNkMycGVlcmludGZwb24gMjE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217',
                                                                                               'links': [       {       'connected_time': '26482727',
                                                                                                                        'fiber_length': 23311,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNkMycGVlcnBvbnBvbiAyMTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '10 Berdux Pl Wellesley (Calvin '
                                                 'Schmidt 10434396)',
                                         'poller_hash': 'MzIyNkMy',
                                         'serial_number': '3226C2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:f3:d8',
                                         'uptime': '26482727'},
                                 {       'counter_source': '218/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a2:74:22',
                                                                                                        '00:03:e6:da:37:d8',
                                                                                                        'cc:be:59:c3:c6:4d'],
                                                                       'id': '218/g1',
                                                                       'in_nucast_pkts': '342626',
                                                                       'in_octets': '7062958895',
                                                                       'in_ucast_pkts': '28952155',
                                                                       'mac_address': 'cc:be:59:b9:c3:62',
                                                                       'mtu': 2000,
                                                                       'name': '218/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '137078389',
                                                                       'out_octets': '448868958515',
                                                                       'out_ucast_pkts': '185816823',
                                                                       'poller_hash': 'NTEyNEZFcGVlcmludGYyMTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '218',
                                                                       'in_nucast_pkts': '137078389',
                                                                       'in_octets': '448868958515',
                                                                       'in_ucast_pkts': '185816823',
                                                                       'links': [       {       'in_nucast_pkts': '137078389',
                                                                                                'in_octets': '448868958515',
                                                                                                'in_ucast_pkts': '185816823',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '342626',
                                                                                                'out_octets': '7062958895',
                                                                                                'out_ucast_pkts': '28952155',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '342626',
                                                                       'out_octets': '7062958895',
                                                                       'out_ucast_pkts': '28952155',
                                                                       'poller_hash': 'NTEyNEZFcGVlcmludGZwb24gMjE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218',
                                                                                               'links': [       {       'connected_time': '611456',
                                                                                                                        'fiber_length': 19681,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNEZFcGVlcnBvbnBvbiAyMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '2869 Line 45 Gadshill (Neil Jenkins '
                                                 '10434799)',
                                         'poller_hash': 'NTEyNEZF',
                                         'serial_number': '5124FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:62',
                                         'uptime': '611456'},
                                 {       'counter_source': '219/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:18:61:34:f3:c7'],
                                                                       'id': '219/g1',
                                                                       'in_nucast_pkts': '32',
                                                                       'in_octets': '71610447730',
                                                                       'in_ucast_pkts': '390575465',
                                                                       'mac_address': 'cc:be:59:ab:33:c7',
                                                                       'mtu': 2000,
                                                                       'name': '219/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652693',
                                                                       'out_octets': '946182885898',
                                                                       'out_ucast_pkts': '729778510',
                                                                       'poller_hash': 'NEY5MUYwcGVlcmludGYyMTkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '219',
                                                                       'in_nucast_pkts': '5652693',
                                                                       'in_octets': '946182885898',
                                                                       'in_ucast_pkts': '729778510',
                                                                       'links': [       {       'in_nucast_pkts': '5652693',
                                                                                                'in_octets': '946182885898',
                                                                                                'in_ucast_pkts': '729778510',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '32',
                                                                                                'out_octets': '71610447730',
                                                                                                'out_ucast_pkts': '390575465',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32',
                                                                       'out_octets': '71610447730',
                                                                       'out_ucast_pkts': '390575465',
                                                                       'poller_hash': 'NEY5MUYwcGVlcmludGZwb24gMjE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '219',
                                                                                               'links': [       {       'connected_time': '8535135',
                                                                                                                        'fiber_length': 23512,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.536,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.04}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5MUYwcGVlcnBvbnBvbiAyMTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '40 Reiner Cr Wellesley (Len Glass '
                                                 '10434419)',
                                         'poller_hash': 'NEY5MUYw',
                                         'serial_number': '4F91F0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:33:c6',
                                         'uptime': '8535135'},
                                 {       'counter_source': '220/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '220/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '84423209899',
                                                                       'in_ucast_pkts': '560638385',
                                                                       'mac_address': 'cc:be:59:ca:dd:d2',
                                                                       'name': '220/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '231',
                                                                       'out_octets': '3530546725969',
                                                                       'out_ucast_pkts': '2435119595',
                                                                       'poller_hash': 'NDI4MUI5cGVlcmludGYyMjAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '220',
                                                                       'in_nucast_pkts': '231',
                                                                       'in_octets': '3530546725969',
                                                                       'in_ucast_pkts': '2435119595',
                                                                       'links': [       {       'in_nucast_pkts': '231',
                                                                                                'in_octets': '3530546725969',
                                                                                                'in_ucast_pkts': '2435119595',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '84423209899',
                                                                                                'out_ucast_pkts': '560638385',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '84423209899',
                                                                       'out_ucast_pkts': '560638385',
                                                                       'poller_hash': 'NDI4MUI5cGVlcmludGZwb24gMjIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '220',
                                                                                               'links': [       {       'connected_time': '21886550',
                                                                                                                        'fiber_length': 22630,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUI5cGVlcnBvbnBvbiAyMjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1313 Queen's Bush Rd Wellesley "
                                                 '(Sandra Erb 10434424)',
                                         'poller_hash': 'NDI4MUI5',
                                         'serial_number': '4281B9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dd:d2',
                                         'uptime': '21886550'},
                                 {       'counter_source': '221/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b5:c3'],
                                                                       'id': '221/g1',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '142340447741',
                                                                       'in_ucast_pkts': '414291333',
                                                                       'mac_address': 'cc:be:59:ca:df:f5',
                                                                       'mtu': 2000,
                                                                       'name': '221/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919908',
                                                                       'out_octets': '1314647134830',
                                                                       'out_ucast_pkts': '1146789602',
                                                                       'poller_hash': 'NDI4MUZEcGVlcmludGYyMjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '221',
                                                                       'in_nucast_pkts': '9919908',
                                                                       'in_octets': '1314647134830',
                                                                       'in_ucast_pkts': '1146789602',
                                                                       'links': [       {       'in_nucast_pkts': '9919908',
                                                                                                'in_octets': '1314647134830',
                                                                                                'in_ucast_pkts': '1146789602',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '31',
                                                                                                'out_octets': '142340447741',
                                                                                                'out_ucast_pkts': '414291333',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '142340447741',
                                                                       'out_ucast_pkts': '414291333',
                                                                       'poller_hash': 'NDI4MUZEcGVlcmludGZwb24gMjIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221',
                                                                                               'links': [       {       'connected_time': '8535183',
                                                                                                                        'fiber_length': 23121,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUZEcGVlcnBvbnBvbiAyMjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '121 Lawrence St Wellesley (Gregory '
                                                 'Parks 10434291)',
                                         'poller_hash': 'NDI4MUZE',
                                         'serial_number': '4281FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:df:f2',
                                         'uptime': '8535183'},
                                 {       'counter_source': '222/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '222/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2294703571564',
                                                                       'in_ucast_pkts': '3794345190',
                                                                       'mac_address': 'd0:76:8f:1b:a6:a3',
                                                                       'name': '222/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '9567791286550',
                                                                       'out_ucast_pkts': '8261558851',
                                                                       'poller_hash': 'ODcyNTIycGVlcmludGYyMjIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '222',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '9567791286550',
                                                                       'in_ucast_pkts': '8261558851',
                                                                       'links': [       {       'in_nucast_pkts': '66',
                                                                                                'in_octets': '9567791286550',
                                                                                                'in_ucast_pkts': '8261558851',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2294703571564',
                                                                                                'out_ucast_pkts': '3794345190',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2294703571564',
                                                                       'out_ucast_pkts': '3794345190',
                                                                       'poller_hash': 'ODcyNTIycGVlcmludGZwb24gMjIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '222',
                                                                                               'links': [       {       'connected_time': '8535216',
                                                                                                                        'fiber_length': 23047,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTIycGVlcnBvbnBvbiAyMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '105 LAWRENCE ST (WALKER CHRIS)',
                                         'poller_hash': 'ODcyNTIy',
                                         'serial_number': '872522',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a6:a3',
                                         'uptime': '8535216'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '224/g1',
                                                                       'name': '224/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEY5RTlFcGVlcmludGYyMjQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '224',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NEY5RTlFcGVlcmludGZwb24gMjI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '224',
                                                                                               'links': [       {       'connected_time': '145751',
                                                                                                                        'fiber_length': 23806,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.252,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.028}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5RTlFcGVlcnBvbnBvbiAyMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '92 Reiner Cr Wellesley',
                                         'poller_hash': 'NEY5RTlF',
                                         'serial_number': '4F9E9E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:77:86',
                                         'uptime': '145751'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '226/g1',
                                                                       'name': '226/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDI4MzMwcGVlcmludGYyMjYvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '226',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDI4MzMwcGVlcmludGZwb24gMjI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '226',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MzMwcGVlcnBvbnBvbiAyMjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '108 Reiner Cr Wellesley',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "226" '
                                                                  '(location: GPON '
                                                                  'port "6/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:11.310Z',
                                         'poller_hash': 'NDI4MzMw',
                                         'serial_number': '428330',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '227/g1',
                                                                       'name': '227/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwREVCcGVlcmludGYyMjcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '227',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzAwREVCcGVlcmludGZwb24gMjI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '227',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzAwREVCcGVlcnBvbnBvbiAyMjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '16 LAWRENCE ST (KOCH KIM)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "227" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2024-11-29T19:17:53.870Z',
                                         'poller_hash': 'MzAwREVC',
                                         'serial_number': '300DEB',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '228/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:52:eb'],
                                                                       'id': '228/g1',
                                                                       'in_nucast_pkts': '6588',
                                                                       'in_octets': '1410704960725',
                                                                       'in_ucast_pkts': '4023819357',
                                                                       'mac_address': 'cc:be:59:ca:dc:cd',
                                                                       'mtu': 2000,
                                                                       'name': '228/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61709696',
                                                                       'out_octets': '14321293692865',
                                                                       'out_ucast_pkts': '11100070060',
                                                                       'poller_hash': 'NDI4MTk4cGVlcmludGYyMjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '228',
                                                                       'in_nucast_pkts': '61709696',
                                                                       'in_octets': '14321293692865',
                                                                       'in_ucast_pkts': '11100070060',
                                                                       'links': [       {       'in_nucast_pkts': '61709696',
                                                                                                'in_octets': '14321293692865',
                                                                                                'in_ucast_pkts': '11100070060',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6588',
                                                                                                'out_octets': '1410704960725',
                                                                                                'out_ucast_pkts': '4023819357',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6588',
                                                                       'out_octets': '1410704960725',
                                                                       'out_ucast_pkts': '4023819357',
                                                                       'poller_hash': 'NDI4MTk4cGVlcmludGZwb24gMjI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '228',
                                                                                               'links': [       {       'connected_time': '21887200',
                                                                                                                        'fiber_length': 23028,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MTk4cGVlcnBvbnBvbiAyMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '65 Lawrence St Wellesley (Lee and '
                                                 'Cory Becker 10434327)',
                                         'poller_hash': 'NDI4MTk4',
                                         'serial_number': '428198',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dc:ca',
                                         'uptime': '21887200'},
                                 {       'counter_source': '229/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:d3:ae'],
                                                                       'id': '229/g1',
                                                                       'in_nucast_pkts': '12195',
                                                                       'in_octets': '241755634412',
                                                                       'in_ucast_pkts': '832105427',
                                                                       'mac_address': 'cc:be:59:ca:e4:2d',
                                                                       'mtu': 2000,
                                                                       'name': '229/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61709889',
                                                                       'out_octets': '2952698482277',
                                                                       'out_ucast_pkts': '2647295141',
                                                                       'poller_hash': 'NDI4Mjg0cGVlcmludGYyMjkvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '229',
                                                                       'in_nucast_pkts': '61709889',
                                                                       'in_octets': '2952698482277',
                                                                       'in_ucast_pkts': '2647295141',
                                                                       'links': [       {       'in_nucast_pkts': '61709889',
                                                                                                'in_octets': '2952698482277',
                                                                                                'in_ucast_pkts': '2647295141',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12195',
                                                                                                'out_octets': '241755634412',
                                                                                                'out_ucast_pkts': '832105427',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12195',
                                                                       'out_octets': '241755634412',
                                                                       'out_ucast_pkts': '832105427',
                                                                       'poller_hash': 'NDI4Mjg0cGVlcmludGZwb24gMjI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229',
                                                                                               'links': [       {       'connected_time': '21886553',
                                                                                                                        'fiber_length': 23254,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4Mjg0cGVlcnBvbnBvbiAyMjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '49 Lawrence St (Mcdonnell)',
                                         'poller_hash': 'NDI4Mjg0',
                                         'serial_number': '428284',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:e4:2a',
                                         'uptime': '21886553'},
                                 {       'counter_source': '230/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:46:94'],
                                                                       'id': '230/g1',
                                                                       'in_nucast_pkts': '24',
                                                                       'in_octets': '140787309423',
                                                                       'in_ucast_pkts': '355078263',
                                                                       'mac_address': 'cc:be:59:ca:de:d5',
                                                                       'mtu': 2000,
                                                                       'name': '230/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3659494',
                                                                       'out_octets': '752629715511',
                                                                       'out_ucast_pkts': '616810835',
                                                                       'poller_hash': 'NDI4MUQ5cGVlcmludGYyMzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '230',
                                                                       'in_nucast_pkts': '3659494',
                                                                       'in_octets': '752629715511',
                                                                       'in_ucast_pkts': '616810835',
                                                                       'links': [       {       'in_nucast_pkts': '3659494',
                                                                                                'in_octets': '752629715511',
                                                                                                'in_ucast_pkts': '616810835',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '24',
                                                                                                'out_octets': '140787309423',
                                                                                                'out_ucast_pkts': '355078263',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24',
                                                                       'out_octets': '140787309423',
                                                                       'out_ucast_pkts': '355078263',
                                                                       'poller_hash': 'NDI4MUQ5cGVlcmludGZwb24gMjMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230',
                                                                                               'links': [       {       'connected_time': '3096859',
                                                                                                                        'fiber_length': 23671,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUQ5cGVlcnBvbnBvbiAyMzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '51 Zinkann Cr Wellesley (PAUL KEVIN)',
                                         'poller_hash': 'NDI4MUQ5',
                                         'serial_number': '4281D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:d2',
                                         'uptime': '3096859'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '231/g1',
                                                                       'name': '231/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDI4NDJBcGVlcmludGYyMzEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '231',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDI4NDJBcGVlcmludGZwb24gMjMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4NDJBcGVlcnBvbnBvbiAyMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '19 ZINKANN CR',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "231" '
                                                                  '(location: GPON '
                                                                  'port "6/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:11.360Z',
                                         'poller_hash': 'NDI4NDJB',
                                         'serial_number': '42842A',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '232/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:52:be'],
                                                                       'id': '232/g1',
                                                                       'in_nucast_pkts': '12302',
                                                                       'in_octets': '1704393423703',
                                                                       'in_ucast_pkts': '1918465953',
                                                                       'mac_address': 'cc:be:59:ca:dd:c5',
                                                                       'mtu': 2000,
                                                                       'name': '232/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710726',
                                                                       'out_octets': '2856687786995',
                                                                       'out_ucast_pkts': '2847212928',
                                                                       'poller_hash': 'NDI4MUI3cGVlcmludGYyMzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '232',
                                                                       'in_nucast_pkts': '61710726',
                                                                       'in_octets': '2856687786995',
                                                                       'in_ucast_pkts': '2847212928',
                                                                       'links': [       {       'in_nucast_pkts': '61710726',
                                                                                                'in_octets': '2856687786995',
                                                                                                'in_ucast_pkts': '2847212928',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12302',
                                                                                                'out_octets': '1704393423703',
                                                                                                'out_ucast_pkts': '1918465953',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12302',
                                                                       'out_octets': '1704393423703',
                                                                       'out_ucast_pkts': '1918465953',
                                                                       'poller_hash': 'NDI4MUI3cGVlcmludGZwb24gMjMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232',
                                                                                               'links': [       {       'connected_time': '21886585',
                                                                                                                        'fiber_length': 23237,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUI3cGVlcnBvbnBvbiAyMzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '48 Lawrence St Wellesley (Brent & '
                                                 'Sarah Kipfer 10434373)',
                                         'poller_hash': 'NDI4MUI3',
                                         'serial_number': '4281B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dd:c2',
                                         'uptime': '21886585'},
                                 {       'counter_source': '233/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:a7:15',
                                                                                                        '5c:0f:fb:17:6e:95',
                                                                                                        'cc:be:59:2f:2f:b5'],
                                                                       'id': '233/g1',
                                                                       'in_nucast_pkts': '4916179',
                                                                       'in_octets': '187338555597',
                                                                       'in_ucast_pkts': '440055046',
                                                                       'mac_address': 'ec:4f:82:8d:ee:8b',
                                                                       'mtu': 2000,
                                                                       'name': '233/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9005784789',
                                                                       'out_octets': '13749762109280',
                                                                       'out_ucast_pkts': '1194461337',
                                                                       'poller_hash': 'MzIyNjE4cGVlcmludGYyMzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '233',
                                                                       'in_nucast_pkts': '9005784789',
                                                                       'in_octets': '13749762109280',
                                                                       'in_ucast_pkts': '1194461337',
                                                                       'links': [       {       'in_nucast_pkts': '9005784789',
                                                                                                'in_octets': '13749762109280',
                                                                                                'in_ucast_pkts': '1194461337',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4916179',
                                                                                                'out_octets': '187338555597',
                                                                                                'out_ucast_pkts': '440055046',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4916179',
                                                                       'out_octets': '187338555597',
                                                                       'out_ucast_pkts': '440055046',
                                                                       'poller_hash': 'MzIyNjE4cGVlcmludGZwb24gMjMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233',
                                                                                               'links': [       {       'connected_time': '8535314',
                                                                                                                        'fiber_length': 23132,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -23.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyNjE4cGVlcnBvbnBvbiAyMzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '113 Lawrence St Wellesley (Aimee '
                                                 'Higgins 10434513)',
                                         'poller_hash': 'MzIyNjE4',
                                         'serial_number': '322618',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:ee:88',
                                         'uptime': '8535314'},
                                 {       'counter_source': '235/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:53:84'],
                                                                       'id': '235/g1',
                                                                       'in_nucast_pkts': '29177819',
                                                                       'in_octets': '369761641545',
                                                                       'in_ucast_pkts': '3281954368',
                                                                       'mac_address': 'cc:be:59:ca:de:fd',
                                                                       'mtu': 2000,
                                                                       'name': '235/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30004254048',
                                                                       'out_octets': '59761364999224',
                                                                       'out_ucast_pkts': '13375755828',
                                                                       'poller_hash': 'NDI4MURFcGVlcmludGYyMzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '235',
                                                                       'in_nucast_pkts': '30004254048',
                                                                       'in_octets': '59761364999224',
                                                                       'in_ucast_pkts': '13375755828',
                                                                       'links': [       {       'in_nucast_pkts': '30004254048',
                                                                                                'in_octets': '59761364999224',
                                                                                                'in_ucast_pkts': '13375755828',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '29177819',
                                                                                                'out_octets': '369761641545',
                                                                                                'out_ucast_pkts': '3281954368',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29177819',
                                                                       'out_octets': '369761641545',
                                                                       'out_ucast_pkts': '3281954368',
                                                                       'poller_hash': 'NDI4MURFcGVlcmludGZwb24gMjM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235',
                                                                                               'links': [       {       'connected_time': '21886825',
                                                                                                                        'fiber_length': 23761,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MURFcGVlcnBvbnBvbiAyMzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 Reiner Cr Wellesley (Brad & Kim '
                                                 'Gerber 10434538)',
                                         'poller_hash': 'NDI4MURF',
                                         'serial_number': '4281DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:fa',
                                         'uptime': '21886825'},
                                 {       'counter_source': '236/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:48:4d'],
                                                                       'id': '236/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '14441323347',
                                                                       'in_ucast_pkts': '69893348',
                                                                       'mac_address': 'cc:be:59:ca:dc:c5',
                                                                       'mtu': 2000,
                                                                       'name': '236/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2362997',
                                                                       'out_octets': '368723722569',
                                                                       'out_ucast_pkts': '256339920',
                                                                       'poller_hash': 'NDI4MTk3cGVlcmludGYyMzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '236',
                                                                       'in_nucast_pkts': '2362997',
                                                                       'in_octets': '368723722569',
                                                                       'in_ucast_pkts': '256339920',
                                                                       'links': [       {       'in_nucast_pkts': '2362997',
                                                                                                'in_octets': '368723722569',
                                                                                                'in_ucast_pkts': '256339920',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '14441323347',
                                                                                                'out_ucast_pkts': '69893348',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '14441323347',
                                                                       'out_ucast_pkts': '69893348',
                                                                       'poller_hash': 'NDI4MTk3cGVlcmludGZwb24gMjM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '236',
                                                                                               'links': [       {       'connected_time': '2000646',
                                                                                                                        'fiber_length': 22318,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MTk3cGVlcnBvbnBvbiAyMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1285 Queen's Bush Rd Wellesley "
                                                 '(Brian & Felicity Finn 10434499)',
                                         'poller_hash': 'NDI4MTk3',
                                         'serial_number': '428197',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dc:c2',
                                         'uptime': '2000646'},
                                 {       'counter_source': '237/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c9:5c'],
                                                                       'id': '237/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '55644595729',
                                                                       'in_ucast_pkts': '153977163',
                                                                       'mac_address': 'cc:be:59:ca:df:ed',
                                                                       'mtu': 2000,
                                                                       'name': '237/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2712356',
                                                                       'out_octets': '623032526113',
                                                                       'out_ucast_pkts': '487165525',
                                                                       'poller_hash': 'NDI4MUZDcGVlcmludGYyMzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '237',
                                                                       'in_nucast_pkts': '2712356',
                                                                       'in_octets': '623032526113',
                                                                       'in_ucast_pkts': '487165525',
                                                                       'links': [       {       'in_nucast_pkts': '2712356',
                                                                                                'in_octets': '623032526113',
                                                                                                'in_ucast_pkts': '487165525',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '55644595729',
                                                                                                'out_ucast_pkts': '153977163',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '55644595729',
                                                                       'out_ucast_pkts': '153977163',
                                                                       'poller_hash': 'NDI4MUZDcGVlcmludGZwb24gMjM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '237',
                                                                                               'links': [       {       'connected_time': '2296131',
                                                                                                                        'fiber_length': 23150,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUZDcGVlcnBvbnBvbiAyMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1120 Molesworth St Wellesley (Brett '
                                                 'Moffett 10434484)',
                                         'poller_hash': 'NDI4MUZD',
                                         'serial_number': '4281FC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:df:ea',
                                         'uptime': '2296131'},
                                 {       'counter_source': '238/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '238/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1783079017117',
                                                                       'in_ucast_pkts': '4141792164',
                                                                       'mac_address': 'cc:be:59:cb:e2:da',
                                                                       'name': '238/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '17081818466478',
                                                                       'out_ucast_pkts': '12801132377',
                                                                       'poller_hash': 'NDJDREIzcGVlcmludGYyMzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '238',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '17081818466478',
                                                                       'in_ucast_pkts': '12801132377',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '17081818466478',
                                                                                                'in_ucast_pkts': '12801132377',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1783079017117',
                                                                                                'out_ucast_pkts': '4141792164',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1783079017117',
                                                                       'out_ucast_pkts': '4141792164',
                                                                       'poller_hash': 'NDJDREIzcGVlcmludGZwb24gMjM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '238',
                                                                                               'links': [       {       'connected_time': '7142486',
                                                                                                                        'fiber_length': 23209,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDREIzcGVlcnBvbnBvbiAyMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1129 MOLESWORTH ST (PRICE MATTHEW)',
                                         'poller_hash': 'NDJDREIz',
                                         'serial_number': '42CDB3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e2:da',
                                         'uptime': '7142486'},
                                 {       'counter_source': '239/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:47:75'],
                                                                       'id': '239/g1',
                                                                       'in_nucast_pkts': '27',
                                                                       'in_octets': '28887905294',
                                                                       'in_ucast_pkts': '90969710',
                                                                       'mac_address': 'cc:be:59:ab:25:a1',
                                                                       'mtu': 2000,
                                                                       'name': '239/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652485',
                                                                       'out_octets': '239211381295',
                                                                       'out_ucast_pkts': '207123168',
                                                                       'poller_hash': 'NEY4RUQzcGVlcmludGYyMzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '239',
                                                                       'in_nucast_pkts': '5652485',
                                                                       'in_octets': '239211381295',
                                                                       'in_ucast_pkts': '207123168',
                                                                       'links': [       {       'in_nucast_pkts': '5652485',
                                                                                                'in_octets': '239211381295',
                                                                                                'in_ucast_pkts': '207123168',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27',
                                                                                                'out_octets': '28887905294',
                                                                                                'out_ucast_pkts': '90969710',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27',
                                                                       'out_octets': '28887905294',
                                                                       'out_ucast_pkts': '90969710',
                                                                       'poller_hash': 'NEY4RUQzcGVlcmludGZwb24gMjM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '239',
                                                                                               'links': [       {       'connected_time': '8535106',
                                                                                                                        'fiber_length': 22597,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.898,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.132}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY4RUQzcGVlcnBvbnBvbiAyMzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '29 Lawrence St Wellesley (Fred '
                                                 'Lindner 10434643)',
                                         'poller_hash': 'NEY4RUQz',
                                         'serial_number': '4F8ED3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:25:a0',
                                         'uptime': '8535106'},
                                 {       'counter_source': '241/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:48:dd'],
                                                                       'id': '241/g1',
                                                                       'in_nucast_pkts': '7170',
                                                                       'in_octets': '1049144975762',
                                                                       'in_ucast_pkts': '4686368063',
                                                                       'mac_address': 'cc:be:59:ca:dd:55',
                                                                       'mtu': 2000,
                                                                       'name': '241/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710177',
                                                                       'out_octets': '16061328771624',
                                                                       'out_ucast_pkts': '12820667353',
                                                                       'poller_hash': 'NDI4MUE5cGVlcmludGYyNDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '241',
                                                                       'in_nucast_pkts': '61710177',
                                                                       'in_octets': '16061328771624',
                                                                       'in_ucast_pkts': '12820667353',
                                                                       'links': [       {       'in_nucast_pkts': '61710177',
                                                                                                'in_octets': '16061328771624',
                                                                                                'in_ucast_pkts': '12820667353',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7170',
                                                                                                'out_octets': '1049144975762',
                                                                                                'out_ucast_pkts': '4686368063',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7170',
                                                                       'out_octets': '1049144975762',
                                                                       'out_ucast_pkts': '4686368063',
                                                                       'poller_hash': 'NDI4MUE5cGVlcmludGZwb24gMjQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241',
                                                                                               'links': [       {       'connected_time': '21886807',
                                                                                                                        'fiber_length': 23074,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUE5cGVlcnBvbnBvbiAyNDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1079 MOLESWORTH ST (MUNRO JERED)',
                                         'poller_hash': 'NDI4MUE5',
                                         'serial_number': '4281A9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dd:52',
                                         'uptime': '21886807'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '242/g1',
                                                                       'name': '242/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGOUU3cGVlcmludGYyNDIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '242',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVGOUU3cGVlcmludGZwb24gMjQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUU3cGVlcnBvbnBvbiAyNDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '39 MILL ST E (FRIES GREGORY R)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'OLT PON port 7/7 is '
                                                                  'down',
                                         'offline_time': '2024-10-17T07:03:42.740Z',
                                         'poller_hash': 'MzVGOUU3',
                                         'serial_number': '35F9E7',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '243/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c9:0b'],
                                                                       'id': '243/g1',
                                                                       'in_nucast_pkts': '21',
                                                                       'in_octets': '193013881571',
                                                                       'in_ucast_pkts': '411528763',
                                                                       'mac_address': 'cc:be:59:ca:e2:e5',
                                                                       'mtu': 2000,
                                                                       'name': '243/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919916',
                                                                       'out_octets': '1109254622584',
                                                                       'out_ucast_pkts': '862253265',
                                                                       'poller_hash': 'NDI4MjVCcGVlcmludGYyNDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '243',
                                                                       'in_nucast_pkts': '9919916',
                                                                       'in_octets': '1109254622584',
                                                                       'in_ucast_pkts': '862253265',
                                                                       'links': [       {       'in_nucast_pkts': '9919916',
                                                                                                'in_octets': '1109254622584',
                                                                                                'in_ucast_pkts': '862253265',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21',
                                                                                                'out_octets': '193013881571',
                                                                                                'out_ucast_pkts': '411528763',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21',
                                                                       'out_octets': '193013881571',
                                                                       'out_ucast_pkts': '411528763',
                                                                       'poller_hash': 'NDI4MjVCcGVlcmludGZwb24gMjQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243',
                                                                                               'links': [       {       'connected_time': '8535258',
                                                                                                                        'fiber_length': 23044,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MjVCcGVlcnBvbnBvbiAyNDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1065 Molesworth St Wellesley '
                                                 '(Michelle Martin 10434298)',
                                         'poller_hash': 'NDI4MjVC',
                                         'serial_number': '42825B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:e2:e2',
                                         'uptime': '8535258'},
                                 {       'counter_source': '245/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:46:82'],
                                                                       'id': '245/g1',
                                                                       'in_nucast_pkts': '22',
                                                                       'in_octets': '276103108224',
                                                                       'in_ucast_pkts': '1357566547',
                                                                       'mac_address': 'cc:be:59:ca:dd:6d',
                                                                       'mtu': 2000,
                                                                       'name': '245/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920062',
                                                                       'out_octets': '5050044210207',
                                                                       'out_ucast_pkts': '3743928671',
                                                                       'poller_hash': 'NDI4MUFDcGVlcmludGYyNDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '245',
                                                                       'in_nucast_pkts': '9920062',
                                                                       'in_octets': '5050044210207',
                                                                       'in_ucast_pkts': '3743928671',
                                                                       'links': [       {       'in_nucast_pkts': '9920062',
                                                                                                'in_octets': '5050044210207',
                                                                                                'in_ucast_pkts': '3743928671',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22',
                                                                                                'out_octets': '276103108224',
                                                                                                'out_ucast_pkts': '1357566547',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22',
                                                                       'out_octets': '276103108224',
                                                                       'out_ucast_pkts': '1357566547',
                                                                       'poller_hash': 'NDI4MUFDcGVlcmludGZwb24gMjQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245',
                                                                                               'links': [       {       'connected_time': '8535250',
                                                                                                                        'fiber_length': 23444,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUFDcGVlcnBvbnBvbiAyNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '127 Ferris Dr (Mike Butler 10434478)',
                                         'poller_hash': 'NDI4MUFD',
                                         'serial_number': '4281AC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dd:6a',
                                         'uptime': '8535250'},
                                 {       'counter_source': '247/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c7:c7'],
                                                                       'id': '247/g1',
                                                                       'in_nucast_pkts': '11935',
                                                                       'in_octets': '2265476802572',
                                                                       'in_ucast_pkts': '5159153706',
                                                                       'mac_address': 'cc:be:59:ca:dc:85',
                                                                       'mtu': 2000,
                                                                       'name': '247/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710344',
                                                                       'out_octets': '18461930371614',
                                                                       'out_ucast_pkts': '14219520906',
                                                                       'poller_hash': 'NDI4MThGcGVlcmludGYyNDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '247',
                                                                       'in_nucast_pkts': '61710344',
                                                                       'in_octets': '18461930371614',
                                                                       'in_ucast_pkts': '14219520906',
                                                                       'links': [       {       'in_nucast_pkts': '61710344',
                                                                                                'in_octets': '18461930371614',
                                                                                                'in_ucast_pkts': '14219520906',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11935',
                                                                                                'out_octets': '2265476802572',
                                                                                                'out_ucast_pkts': '5159153706',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11935',
                                                                       'out_octets': '2265476802572',
                                                                       'out_ucast_pkts': '5159153706',
                                                                       'poller_hash': 'NDI4MThGcGVlcmludGZwb24gMjQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247',
                                                                                               'links': [       {       'connected_time': '21886238',
                                                                                                                        'fiber_length': 23181,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MThGcGVlcnBvbnBvbiAyNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1132 Molesworth St (Laura Lucchetti '
                                                 '10434476)',
                                         'poller_hash': 'NDI4MThG',
                                         'serial_number': '42818F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dc:82',
                                         'uptime': '21886238'},
                                 {       'counter_source': '248/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c6:3b'],
                                                                       'id': '248/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '152853690239',
                                                                       'in_ucast_pkts': '879314810',
                                                                       'mac_address': 'cc:be:59:ab:89:03',
                                                                       'mtu': 2000,
                                                                       'name': '248/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2331746',
                                                                       'out_octets': '2414286715935',
                                                                       'out_ucast_pkts': '1796920590',
                                                                       'poller_hash': 'NEZBMjVFcGVlcmludGYyNDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248',
                                                                       'in_nucast_pkts': '2331746',
                                                                       'in_octets': '2414286715935',
                                                                       'in_ucast_pkts': '1796920590',
                                                                       'links': [       {       'in_nucast_pkts': '2331746',
                                                                                                'in_octets': '2414286715935',
                                                                                                'in_ucast_pkts': '1796920590',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '152853690239',
                                                                                                'out_ucast_pkts': '879314810',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '152853690239',
                                                                       'out_ucast_pkts': '879314810',
                                                                       'poller_hash': 'NEZBMjVFcGVlcmludGZwb24gMjQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248',
                                                                                               'links': [       {       'connected_time': '3434993',
                                                                                                                        'fiber_length': 23706,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.724,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.068}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMjVFcGVlcnBvbnBvbiAyNDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '72 Reiner Cr Wellesley (Chris '
                                                 'Dietrich 10404238)',
                                         'poller_hash': 'NEZBMjVF',
                                         'serial_number': '4FA25E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:89:02',
                                         'uptime': '3434993'},
                                 {       'counter_source': '249/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '249/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2592672738120',
                                                                       'in_ucast_pkts': '16404090883',
                                                                       'mac_address': 'cc:be:59:ca:dd:ba',
                                                                       'name': '249/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '9406831688465',
                                                                       'out_ucast_pkts': '8400366891',
                                                                       'poller_hash': 'NDI4MUI2cGVlcmludGYyNDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '249',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '9406831688465',
                                                                       'in_ucast_pkts': '8400366891',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '9406831688465',
                                                                                                'in_ucast_pkts': '8400366891',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '2592672738120',
                                                                                                'out_ucast_pkts': '16404090883',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2592672738120',
                                                                       'out_ucast_pkts': '16404090883',
                                                                       'poller_hash': 'NDI4MUI2cGVlcmludGZwb24gMjQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249',
                                                                                               'links': [       {       'connected_time': '18146441',
                                                                                                                        'fiber_length': 23440,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUI2cGVlcnBvbnBvbiAyNDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '130 FERRIS DR (MCCUTCHEON DARRYL)',
                                         'poller_hash': 'NDI4MUI2',
                                         'serial_number': '4281B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dd:ba',
                                         'uptime': '18146441'},
                                 {       'counter_source': '251/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:46:ca'],
                                                                       'id': '251/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '19228690660',
                                                                       'in_ucast_pkts': '146250278',
                                                                       'mac_address': 'cc:be:59:ab:23:63',
                                                                       'mtu': 2000,
                                                                       'name': '251/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4018055',
                                                                       'out_octets': '685910199226',
                                                                       'out_ucast_pkts': '480939741',
                                                                       'poller_hash': 'NEY4RENEcGVlcmludGYyNTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '251',
                                                                       'in_nucast_pkts': '4018055',
                                                                       'in_octets': '685910199226',
                                                                       'in_ucast_pkts': '480939741',
                                                                       'links': [       {       'in_nucast_pkts': '4018055',
                                                                                                'in_octets': '685910199226',
                                                                                                'in_ucast_pkts': '480939741',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '19228690660',
                                                                                                'out_ucast_pkts': '146250278',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '19228690660',
                                                                       'out_ucast_pkts': '146250278',
                                                                       'poller_hash': 'NEY4RENEcGVlcmludGZwb24gMjUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251',
                                                                                               'links': [       {       'connected_time': '6011614',
                                                                                                                        'fiber_length': 23256,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.668,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.106}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY4RENEcGVlcnBvbnBvbiAyNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1032 Molesworth St Wellesley (Terry '
                                                 'Steckley 10434287)',
                                         'poller_hash': 'NEY4RENE',
                                         'serial_number': '4F8DCD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:23:62',
                                         'uptime': '6011614'},
                                 {       'counter_source': '252/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c8:0f'],
                                                                       'id': '252/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '125673038378',
                                                                       'in_ucast_pkts': '808820704',
                                                                       'mac_address': 'cc:be:59:ab:33:2f',
                                                                       'mtu': 2000,
                                                                       'name': '252/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652649',
                                                                       'out_octets': '2855127453646',
                                                                       'out_ucast_pkts': '2083654024',
                                                                       'poller_hash': 'NEY5MUNFcGVlcmludGYyNTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '252',
                                                                       'in_nucast_pkts': '5652649',
                                                                       'in_octets': '2855127453646',
                                                                       'in_ucast_pkts': '2083654024',
                                                                       'links': [       {       'in_nucast_pkts': '5652649',
                                                                                                'in_octets': '2855127453646',
                                                                                                'in_ucast_pkts': '2083654024',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '125673038378',
                                                                                                'out_ucast_pkts': '808820704',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '125673038378',
                                                                       'out_ucast_pkts': '808820704',
                                                                       'poller_hash': 'NEY5MUNFcGVlcmludGZwb24gMjUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252',
                                                                                               'links': [       {       'connected_time': '8535117',
                                                                                                                        'fiber_length': 23087,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.358,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.192}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5MUNFcGVlcnBvbnBvbiAyNTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '142 Ferris Drive Wellesley (William '
                                                 'Murray 10434537)',
                                         'poller_hash': 'NEY5MUNF',
                                         'serial_number': '4F91CE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:33:2e',
                                         'uptime': '8535117'},
                                 {       'counter_source': '253/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1e:2a'],
                                                                       'id': '253/g1',
                                                                       'in_nucast_pkts': '160826',
                                                                       'in_octets': '402442292792',
                                                                       'in_ucast_pkts': '1222212224',
                                                                       'mac_address': 'cc:be:59:50:04:0a',
                                                                       'mtu': 2000,
                                                                       'name': '253/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652586',
                                                                       'out_octets': '5234931478684',
                                                                       'out_ucast_pkts': '3898779840',
                                                                       'poller_hash': 'NDNCNUNDcGVlcmludGYyNTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253',
                                                                       'in_nucast_pkts': '5652586',
                                                                       'in_octets': '5234931478684',
                                                                       'in_ucast_pkts': '3898779840',
                                                                       'links': [       {       'in_nucast_pkts': '5652586',
                                                                                                'in_octets': '5234931478684',
                                                                                                'in_ucast_pkts': '3898779840',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160826',
                                                                                                'out_octets': '402442292792',
                                                                                                'out_ucast_pkts': '1222212224',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160826',
                                                                       'out_octets': '402442292792',
                                                                       'out_ucast_pkts': '1222212224',
                                                                       'poller_hash': 'NDNCNUNDcGVlcmludGZwb24gMjUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253',
                                                                                               'links': [       {       'connected_time': '8535052',
                                                                                                                        'fiber_length': 23180,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.472,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.332}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUNDcGVlcnBvbnBvbiAyNTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': 'Harris (Temp)',
                                         'poller_hash': 'NDNCNUND',
                                         'serial_number': '43B5CC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:04:09',
                                         'uptime': '8535052'},
                                 {       'counter_source': '254/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'dc:2c:6e:c7:55:a9'],
                                                                       'id': '254/g1',
                                                                       'in_nucast_pkts': '8006',
                                                                       'in_octets': '3181406187200',
                                                                       'in_ucast_pkts': '8856270689',
                                                                       'mac_address': 'cc:be:59:ca:de:cd',
                                                                       'mtu': 2000,
                                                                       'name': '254/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710072',
                                                                       'out_octets': '25443060721841',
                                                                       'out_ucast_pkts': '20231189307',
                                                                       'poller_hash': 'NDI4MUQ4cGVlcmludGYyNTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '254',
                                                                       'in_nucast_pkts': '61710072',
                                                                       'in_octets': '25443060721841',
                                                                       'in_ucast_pkts': '20231189307',
                                                                       'links': [       {       'in_nucast_pkts': '61710072',
                                                                                                'in_octets': '25443060721841',
                                                                                                'in_ucast_pkts': '20231189307',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8006',
                                                                                                'out_octets': '3181406187200',
                                                                                                'out_ucast_pkts': '8856270689',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8006',
                                                                       'out_octets': '3181406187200',
                                                                       'out_ucast_pkts': '8856270689',
                                                                       'poller_hash': 'NDI4MUQ4cGVlcmludGZwb24gMjU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254',
                                                                                               'links': [       {       'connected_time': '21886993',
                                                                                                                        'fiber_length': 23271,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUQ4cGVlcnBvbnBvbiAyNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '174 Ferris Drive Wellesley (Jeremy '
                                                 'Szabo 10434290)',
                                         'poller_hash': 'NDI4MUQ4',
                                         'serial_number': '4281D8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:ca',
                                         'uptime': '21886993'},
                                 {       'counter_source': '255/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'c8:4f:86:ac:ef:2f'],
                                                                       'id': '255/g1',
                                                                       'in_nucast_pkts': '5',
                                                                       'in_octets': '443796890276',
                                                                       'in_ucast_pkts': '1153506289',
                                                                       'mac_address': 'cc:be:59:cb:ad:a5',
                                                                       'mtu': 2000,
                                                                       'name': '255/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4978630',
                                                                       'out_octets': '3541351295562',
                                                                       'out_ucast_pkts': '2787431191',
                                                                       'poller_hash': 'NDJDNzBDcGVlcmludGYyNTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '255',
                                                                       'in_nucast_pkts': '4978630',
                                                                       'in_octets': '3541351295562',
                                                                       'in_ucast_pkts': '2787431191',
                                                                       'links': [       {       'in_nucast_pkts': '4978630',
                                                                                                'in_octets': '3541351295562',
                                                                                                'in_ucast_pkts': '2787431191',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5',
                                                                                                'out_octets': '443796890276',
                                                                                                'out_ucast_pkts': '1153506289',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5',
                                                                       'out_octets': '443796890276',
                                                                       'out_ucast_pkts': '1153506289',
                                                                       'poller_hash': 'NDJDNzBDcGVlcmludGZwb24gMjU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '255',
                                                                                               'links': [       {       'connected_time': '4241042',
                                                                                                                        'fiber_length': 24704,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNzBDcGVlcnBvbnBvbiAyNTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1400 Greenwood Hill Dr Wellesley '
                                                 '(Bio-Ag Consultants 10434635)',
                                         'poller_hash': 'NDJDNzBD',
                                         'serial_number': '42C70C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ad:a2',
                                         'uptime': '4241042'},
                                 {       'counter_source': '256/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:47:99'],
                                                                       'id': '256/g1',
                                                                       'in_nucast_pkts': '413175',
                                                                       'in_octets': '271940403769',
                                                                       'in_ucast_pkts': '1076875762',
                                                                       'mac_address': 'cc:be:59:ca:de:15',
                                                                       'mtu': 2000,
                                                                       'name': '256/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10323303',
                                                                       'out_octets': '4564708903532',
                                                                       'out_ucast_pkts': '3375960212',
                                                                       'poller_hash': 'NDI4MUMxcGVlcmludGYyNTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256',
                                                                       'in_nucast_pkts': '10323303',
                                                                       'in_octets': '4564708903532',
                                                                       'in_ucast_pkts': '3375960212',
                                                                       'links': [       {       'in_nucast_pkts': '10323303',
                                                                                                'in_octets': '4564708903532',
                                                                                                'in_ucast_pkts': '3375960212',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '413175',
                                                                                                'out_octets': '271940403769',
                                                                                                'out_ucast_pkts': '1076875762',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '413175',
                                                                       'out_octets': '271940403769',
                                                                       'out_ucast_pkts': '1076875762',
                                                                       'poller_hash': 'NDI4MUMxcGVlcmludGZwb24gMjU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256',
                                                                                               'links': [       {       'connected_time': '8535183',
                                                                                                                        'fiber_length': 23058,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUMxcGVlcnBvbnBvbiAyNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '183 FERRIS DR (GILBERT FIONA & MARK)',
                                         'poller_hash': 'NDI4MUMx',
                                         'serial_number': '4281C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:12',
                                         'uptime': '8535183'},
                                 {       'counter_source': '257/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '260376264036',
                                                                       'in_ucast_pkts': '1369711905',
                                                                       'mac_address': 'cc:be:59:ca:db:ba',
                                                                       'name': '257/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '33',
                                                                       'out_octets': '5991563604220',
                                                                       'out_ucast_pkts': '4205903070',
                                                                       'poller_hash': 'NDI4MTc2cGVlcmludGYyNTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '257',
                                                                       'in_nucast_pkts': '33',
                                                                       'in_octets': '5991563604220',
                                                                       'in_ucast_pkts': '4205903070',
                                                                       'links': [       {       'in_nucast_pkts': '33',
                                                                                                'in_octets': '5991563604220',
                                                                                                'in_ucast_pkts': '4205903070',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '260376264036',
                                                                                                'out_ucast_pkts': '1369711905',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '260376264036',
                                                                       'out_ucast_pkts': '1369711905',
                                                                       'poller_hash': 'NDI4MTc2cGVlcmludGZwb24gMjU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257',
                                                                                               'links': [       {       'connected_time': '8535140',
                                                                                                                        'fiber_length': 23489,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MTc2cGVlcnBvbnBvbiAyNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '39 ZINKANN CR (HOFFMAN KARI)',
                                         'poller_hash': 'NDI4MTc2',
                                         'serial_number': '428176',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:db:ba',
                                         'uptime': '8535140'},
                                 {       'counter_source': '259/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '259/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '165880523543',
                                                                       'in_ucast_pkts': '1047111901',
                                                                       'mac_address': 'cc:be:59:ca:dd:72',
                                                                       'name': '259/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3738604999824',
                                                                       'out_ucast_pkts': '2687510560',
                                                                       'poller_hash': 'NDI4MUFEcGVlcmludGYyNTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '259',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3738604999824',
                                                                       'in_ucast_pkts': '2687510560',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3738604999824',
                                                                                                'in_ucast_pkts': '2687510560',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '165880523543',
                                                                                                'out_ucast_pkts': '1047111901',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '165880523543',
                                                                       'out_ucast_pkts': '1047111901',
                                                                       'poller_hash': 'NDI4MUFEcGVlcmludGZwb24gMjU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '259',
                                                                                               'links': [       {       'connected_time': '8535258',
                                                                                                                        'fiber_length': 22826,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUFEcGVlcnBvbnBvbiAyNTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1128 Henry St Wellesley (Peter Brand '
                                                 '10434670)',
                                         'poller_hash': 'NDI4MUFE',
                                         'serial_number': '4281AD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dd:72',
                                         'uptime': '8535258'},
                                 {       'counter_source': '260/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:48:c2'],
                                                                       'id': '260/g1',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '1344390113796',
                                                                       'in_ucast_pkts': '2004960612',
                                                                       'mac_address': 'cc:be:59:ca:de:45',
                                                                       'mtu': 2000,
                                                                       'name': '260/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919750',
                                                                       'out_octets': '4050197410996',
                                                                       'out_ucast_pkts': '3044233065',
                                                                       'poller_hash': 'NDI4MUM3cGVlcmludGYyNjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '260',
                                                                       'in_nucast_pkts': '9919750',
                                                                       'in_octets': '4050197410996',
                                                                       'in_ucast_pkts': '3044233065',
                                                                       'links': [       {       'in_nucast_pkts': '9919750',
                                                                                                'in_octets': '4050197410996',
                                                                                                'in_ucast_pkts': '3044233065',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '28',
                                                                                                'out_octets': '1344390113796',
                                                                                                'out_ucast_pkts': '2004960612',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '1344390113796',
                                                                       'out_ucast_pkts': '2004960612',
                                                                       'poller_hash': 'NDI4MUM3cGVlcmludGZwb24gMjYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '260',
                                                                                               'links': [       {       'connected_time': '8535227',
                                                                                                                        'fiber_length': 23209,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUM3cGVlcnBvbnBvbiAyNjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '147 FERRIS DR (VONA STEFANO)',
                                         'poller_hash': 'NDI4MUM3',
                                         'serial_number': '4281C7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:de:42',
                                         'uptime': '8535227'},
                                 {       'counter_source': '261/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:bb:3a'],
                                                                       'id': '261/g1',
                                                                       'in_nucast_pkts': '76436',
                                                                       'in_octets': '9424815391',
                                                                       'in_ucast_pkts': '38123583',
                                                                       'mac_address': 'cc:be:59:ca:dc:a5',
                                                                       'mtu': 2000,
                                                                       'name': '261/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4587488',
                                                                       'out_octets': '112217381435',
                                                                       'out_ucast_pkts': '83266085',
                                                                       'poller_hash': 'NDI4MTkzcGVlcmludGYyNjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '261',
                                                                       'in_nucast_pkts': '4587488',
                                                                       'in_octets': '112217381435',
                                                                       'in_ucast_pkts': '83266085',
                                                                       'links': [       {       'in_nucast_pkts': '4587488',
                                                                                                'in_octets': '112217381435',
                                                                                                'in_ucast_pkts': '83266085',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '76436',
                                                                                                'out_octets': '9424815391',
                                                                                                'out_ucast_pkts': '38123583',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76436',
                                                                       'out_octets': '9424815391',
                                                                       'out_ucast_pkts': '38123583',
                                                                       'poller_hash': 'NDI4MTkzcGVlcmludGZwb24gMjYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '261',
                                                                                               'links': [       {       'connected_time': '3895185',
                                                                                                                        'fiber_length': 23145,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MTkzcGVlcnBvbnBvbiAyNjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1091 Henry St Wellesley (Ross Leis '
                                                 '10434674)',
                                         'poller_hash': 'NDI4MTkz',
                                         'serial_number': '428193',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:dc:a2',
                                         'uptime': '3895185'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '262/g1',
                                                                       'name': '262/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODcyNDQ2cGVlcmludGYyNjIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '262',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODcyNDQ2cGVlcmludGZwb24gMjYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262',
                                                                                               'links': [       {       'connected_time': '15879229',
                                                                                                                        'fiber_length': 23348,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDQ2cGVlcnBvbnBvbiAyNjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 Zoeger Ct Wellesley',
                                         'poller_hash': 'ODcyNDQ2',
                                         'serial_number': '872446',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:c3',
                                         'uptime': '15879229'},
                                 {       'counter_source': '263/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:48:d4'],
                                                                       'id': '263/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '138296369929',
                                                                       'in_ucast_pkts': '525079291',
                                                                       'mac_address': 'cc:be:59:ab:86:a7',
                                                                       'mtu': 2000,
                                                                       'name': '263/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652625',
                                                                       'out_octets': '4213066103259',
                                                                       'out_ucast_pkts': '2991732309',
                                                                       'poller_hash': 'NEZBMUMxcGVlcmludGYyNjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '263',
                                                                       'in_nucast_pkts': '5652625',
                                                                       'in_octets': '4213066103259',
                                                                       'in_ucast_pkts': '2991732309',
                                                                       'links': [       {       'in_nucast_pkts': '5652625',
                                                                                                'in_octets': '4213066103259',
                                                                                                'in_ucast_pkts': '2991732309',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '138296369929',
                                                                                                'out_ucast_pkts': '525079291',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '138296369929',
                                                                       'out_ucast_pkts': '525079291',
                                                                       'poller_hash': 'NEZBMUMxcGVlcmludGZwb24gMjYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '263',
                                                                                               'links': [       {       'connected_time': '8535114',
                                                                                                                        'fiber_length': 23327,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.502,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.208}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMUMxcGVlcnBvbnBvbiAyNjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '106 Schweitzer Cr Wellesley (Bob '
                                                 'Tracey 10434512)',
                                         'poller_hash': 'NEZBMUMx',
                                         'serial_number': '4FA1C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:86:a6',
                                         'uptime': '8535114'},
                                 {       'counter_source': '264/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:ff:c2:68'],
                                                                       'id': '264/g1',
                                                                       'in_nucast_pkts': '118',
                                                                       'in_octets': '1150170018469',
                                                                       'in_ucast_pkts': '5427475028',
                                                                       'mac_address': 'cc:be:59:b9:c2:fc',
                                                                       'mtu': 2000,
                                                                       'name': '264/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39527115',
                                                                       'out_octets': '18704562055365',
                                                                       'out_ucast_pkts': '13385448717',
                                                                       'poller_hash': 'NTEyNERDcGVlcmludGYyNjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '264',
                                                                       'in_nucast_pkts': '39527115',
                                                                       'in_octets': '18704562055365',
                                                                       'in_ucast_pkts': '13385448717',
                                                                       'links': [       {       'in_nucast_pkts': '39527115',
                                                                                                'in_octets': '18704562055365',
                                                                                                'in_ucast_pkts': '13385448717',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '118',
                                                                                                'out_octets': '1150170018469',
                                                                                                'out_ucast_pkts': '5427475028',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '118',
                                                                       'out_octets': '1150170018469',
                                                                       'out_ucast_pkts': '5427475028',
                                                                       'poller_hash': 'NTEyNERDcGVlcmludGZwb24gMjY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264',
                                                                                               'links': [       {       'connected_time': '67856213',
                                                                                                                        'fiber_length': 23453,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNERDcGVlcnBvbnBvbiAyNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '12 Reiner Cr Wellesley (Kim '
                                                 'Heinmiller 10434608)',
                                         'poller_hash': 'NTEyNERD',
                                         'serial_number': '5124DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c2:fc',
                                         'uptime': '67856213'},
                                 {       'counter_source': '265/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c5:5e:f5'],
                                                                       'id': '265/g1',
                                                                       'in_nucast_pkts': '38526355',
                                                                       'in_octets': '698490571368',
                                                                       'in_ucast_pkts': '1782748462',
                                                                       'mac_address': 'ec:4f:82:8e:01:7b',
                                                                       'mtu': 2000,
                                                                       'name': '265/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8706060590',
                                                                       'out_octets': '16837061832195',
                                                                       'out_ucast_pkts': '4671973893',
                                                                       'poller_hash': 'MzIyODc2cGVlcmludGYyNjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '265',
                                                                       'in_nucast_pkts': '8706060590',
                                                                       'in_octets': '16837061832195',
                                                                       'in_ucast_pkts': '4671973893',
                                                                       'links': [       {       'in_nucast_pkts': '8706060590',
                                                                                                'in_octets': '16837061832195',
                                                                                                'in_ucast_pkts': '4671973893',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '38526355',
                                                                                                'out_octets': '698490571368',
                                                                                                'out_ucast_pkts': '1782748462',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38526355',
                                                                       'out_octets': '698490571368',
                                                                       'out_ucast_pkts': '1782748462',
                                                                       'poller_hash': 'MzIyODc2cGVlcmludGZwb24gMjY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265',
                                                                                               'links': [       {       'connected_time': '21614921',
                                                                                                                        'fiber_length': 23420,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODc2cGVlcnBvbnBvbiAyNjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1026 Molesworth St Wellesley (Ron & '
                                                 'Brenda Poole 10411884)',
                                         'poller_hash': 'MzIyODc2',
                                         'serial_number': '322876',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:01:78',
                                         'uptime': '21614921'},
                                 {       'counter_source': '267/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ea:4b'],
                                                                       'id': '267/g1',
                                                                       'in_nucast_pkts': '167483',
                                                                       'in_octets': '1226125540256',
                                                                       'in_ucast_pkts': '1240259292',
                                                                       'mac_address': 'cc:be:59:b9:c3:a7',
                                                                       'mtu': 2000,
                                                                       'name': '267/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920033',
                                                                       'out_octets': '1621406201164',
                                                                       'out_ucast_pkts': '1621790692',
                                                                       'poller_hash': 'NTEyNTE1cGVlcmludGYyNjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '267',
                                                                       'in_nucast_pkts': '9920033',
                                                                       'in_octets': '1621406201164',
                                                                       'in_ucast_pkts': '1621790692',
                                                                       'links': [       {       'in_nucast_pkts': '9920033',
                                                                                                'in_octets': '1621406201164',
                                                                                                'in_ucast_pkts': '1621790692',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '167483',
                                                                                                'out_octets': '1226125540256',
                                                                                                'out_ucast_pkts': '1240259292',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '167483',
                                                                       'out_octets': '1226125540256',
                                                                       'out_ucast_pkts': '1240259292',
                                                                       'poller_hash': 'NTEyNTE1cGVlcmludGZwb24gMjY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '267',
                                                                                               'links': [       {       'connected_time': '8535507',
                                                                                                                        'fiber_length': 23072,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTE1cGVlcnBvbnBvbiAyNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '103 DAVID ST (STRYKER BRITTANY)',
                                         'poller_hash': 'NTEyNTE1',
                                         'serial_number': '512515',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:a7',
                                         'uptime': '8535507'},
                                 {       'counter_source': '268/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7d:2b'],
                                                                       'id': '268/g1',
                                                                       'in_nucast_pkts': '319021',
                                                                       'in_octets': '704037193535',
                                                                       'in_ucast_pkts': '4737130178',
                                                                       'mac_address': 'ec:4f:82:8e:08:23',
                                                                       'mtu': 2000,
                                                                       'name': '268/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61700374',
                                                                       'out_octets': '13994780115607',
                                                                       'out_ucast_pkts': '9785113326',
                                                                       'poller_hash': 'MzIyOTRCcGVlcmludGYyNjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '268',
                                                                       'in_nucast_pkts': '61700374',
                                                                       'in_octets': '13994780115607',
                                                                       'in_ucast_pkts': '9785113326',
                                                                       'links': [       {       'in_nucast_pkts': '61700374',
                                                                                                'in_octets': '13994780115607',
                                                                                                'in_ucast_pkts': '9785113326',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '319021',
                                                                                                'out_octets': '704037193535',
                                                                                                'out_ucast_pkts': '4737130178',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '319021',
                                                                       'out_octets': '704037193535',
                                                                       'out_ucast_pkts': '4737130178',
                                                                       'poller_hash': 'MzIyOTRCcGVlcmludGZwb24gMjY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '268',
                                                                                               'links': [       {       'connected_time': '21886761',
                                                                                                                        'fiber_length': 22664,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOTRCcGVlcnBvbnBvbiAyNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '81 David St Wellesley (Rob Greer '
                                                 '10434764)',
                                         'poller_hash': 'MzIyOTRC',
                                         'serial_number': '32294B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:08:20',
                                         'uptime': '21886761'},
                                 {       'counter_source': '269/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:bb:80'],
                                                                       'id': '269/g1',
                                                                       'in_nucast_pkts': '154273',
                                                                       'in_octets': '260774408683',
                                                                       'in_ucast_pkts': '1485736261',
                                                                       'mac_address': 'cc:be:59:ab:60:59',
                                                                       'mtu': 2000,
                                                                       'name': '269/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652396',
                                                                       'out_octets': '4619540025931',
                                                                       'out_ucast_pkts': '3370338574',
                                                                       'poller_hash': 'NEY5QjBFcGVlcmludGYyNjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '269',
                                                                       'in_nucast_pkts': '5652396',
                                                                       'in_octets': '4619540025931',
                                                                       'in_ucast_pkts': '3370338574',
                                                                       'links': [       {       'in_nucast_pkts': '5652396',
                                                                                                'in_octets': '4619540025931',
                                                                                                'in_ucast_pkts': '3370338574',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '154273',
                                                                                                'out_octets': '260774408683',
                                                                                                'out_ucast_pkts': '1485736261',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '154273',
                                                                       'out_octets': '260774408683',
                                                                       'out_ucast_pkts': '1485736261',
                                                                       'poller_hash': 'NEY5QjBFcGVlcmludGZwb24gMjY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '269',
                                                                                               'links': [       {       'connected_time': '8535089',
                                                                                                                        'fiber_length': 22943,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.012,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.374}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY5QjBFcGVlcnBvbnBvbiAyNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '98 Lawrence St Wellesley (Jason Neeb '
                                                 '10434715)',
                                         'poller_hash': 'NEY5QjBF',
                                         'serial_number': '4F9B0E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:60:58',
                                         'uptime': '8535089'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '270/g1',
                                                                       'name': '270/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEZBMTc3cGVlcmludGYyNzAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '270',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NEZBMTc3cGVlcmludGZwb24gMjcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '270',
                                                                                               'links': [       {       'connected_time': '8535096',
                                                                                                                        'fiber_length': 23167,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.114,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.278}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMTc3cGVlcnBvbnBvbiAyNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '163 Ferris Dr Wellesley',
                                         'poller_hash': 'NEZBMTc3',
                                         'serial_number': '4FA177',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:85:a6',
                                         'uptime': '8535096'},
                                 {       'counter_source': '271/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '271/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '86755719787',
                                                                       'in_ucast_pkts': '8878415643',
                                                                       'mac_address': 'cc:be:59:cb:e9:f2',
                                                                       'name': '271/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '198',
                                                                       'out_octets': '997730991304',
                                                                       'out_ucast_pkts': '827793379',
                                                                       'poller_hash': 'NDJDRTk2cGVlcmludGYyNzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '271',
                                                                       'in_nucast_pkts': '198',
                                                                       'in_octets': '997730991304',
                                                                       'in_ucast_pkts': '827793379',
                                                                       'links': [       {       'in_nucast_pkts': '198',
                                                                                                'in_octets': '997730991304',
                                                                                                'in_ucast_pkts': '827793379',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '86755719787',
                                                                                                'out_ucast_pkts': '8878415643',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '86755719787',
                                                                       'out_ucast_pkts': '8878415643',
                                                                       'poller_hash': 'NDJDRTk2cGVlcmludGZwb24gMjcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '271',
                                                                                               'links': [       {       'connected_time': '21886973',
                                                                                                                        'fiber_length': 23003,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTk2cGVlcnBvbnBvbiAyNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3625 NAFZIGER RD (KLASSEN FRAN)',
                                         'poller_hash': 'NDJDRTk2',
                                         'serial_number': '42CE96',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e9:f2',
                                         'uptime': '21886973'},
                                 {       'counter_source': '272/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:f4:11:7e:5c:77'],
                                                                       'id': '272/g1',
                                                                       'in_nucast_pkts': '1960',
                                                                       'in_octets': '166876826130',
                                                                       'in_ucast_pkts': '229189282',
                                                                       'mac_address': 'ec:4f:82:7e:3a:60',
                                                                       'mtu': 2000,
                                                                       'name': '272/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1727640',
                                                                       'out_octets': '408644712410',
                                                                       'out_ucast_pkts': '409581960',
                                                                       'poller_hash': 'MzVGOUI0cGVlcmludGYyNzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272',
                                                                       'in_nucast_pkts': '1727640',
                                                                       'in_octets': '408644712410',
                                                                       'in_ucast_pkts': '409581960',
                                                                       'links': [       {       'in_nucast_pkts': '1727640',
                                                                                                'in_octets': '408644712410',
                                                                                                'in_ucast_pkts': '409581960',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1960',
                                                                                                'out_octets': '166876826130',
                                                                                                'out_ucast_pkts': '229189282',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1960',
                                                                       'out_octets': '166876826130',
                                                                       'out_ucast_pkts': '229189282',
                                                                       'poller_hash': 'MzVGOUI0cGVlcmludGZwb24gMjcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272',
                                                                                               'links': [       {       'connected_time': '1472244',
                                                                                                                        'fiber_length': 23330,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVGOUI0cGVlcnBvbnBvbiAyNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '107 SCHWEITZER CR (MAGUIRE MARC)',
                                         'poller_hash': 'MzVGOUI0',
                                         'serial_number': '35F9B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:3a:60',
                                         'uptime': '1472244'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '274/g1',
                                                                       'name': '274/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNTQzcGVlcmludGYyNzQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '274',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTEyNTQzcGVlcmludGZwb24gMjc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274',
                                                                                               'links': [       {       'connected_time': '67856069',
                                                                                                                        'fiber_length': 11859,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTQzcGVlcnBvbnBvbiAyNzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4008 Perth Line 56 (Robert Wagler)',
                                         'poller_hash': 'NTEyNTQz',
                                         'serial_number': '512543',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c4:31',
                                         'uptime': '67856069'},
                                 {       'counter_source': '276/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c9:ad'],
                                                                       'id': '276/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '149450869401',
                                                                       'in_ucast_pkts': '451698719',
                                                                       'mac_address': 'cc:be:59:b9:c4:52',
                                                                       'mtu': 2000,
                                                                       'name': '276/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920240',
                                                                       'out_octets': '1469902913306',
                                                                       'out_ucast_pkts': '1181921932',
                                                                       'poller_hash': 'NTEyNTRFcGVlcmludGYyNzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '276',
                                                                       'in_nucast_pkts': '9920240',
                                                                       'in_octets': '1469902913306',
                                                                       'in_ucast_pkts': '1181921932',
                                                                       'links': [       {       'in_nucast_pkts': '9920240',
                                                                                                'in_octets': '1469902913306',
                                                                                                'in_ucast_pkts': '1181921932',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '149450869401',
                                                                                                'out_ucast_pkts': '451698719',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '149450869401',
                                                                       'out_ucast_pkts': '451698719',
                                                                       'poller_hash': 'NTEyNTRFcGVlcmludGZwb24gMjc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276',
                                                                                               'links': [       {       'connected_time': '8535434',
                                                                                                                        'fiber_length': 22270,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTRFcGVlcnBvbnBvbiAyNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '28 David St Wellesley (Ray & Carol '
                                                 'Rankin 10434730)',
                                         'poller_hash': 'NTEyNTRF',
                                         'serial_number': '51254E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c4:52',
                                         'uptime': '8535434'},
                                 {       'counter_source': '277/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:48:29'],
                                                                       'id': '277/g1',
                                                                       'in_nucast_pkts': '100',
                                                                       'in_octets': '276441147928',
                                                                       'in_ucast_pkts': '771257156',
                                                                       'mac_address': 'cc:be:59:b9:c5:27',
                                                                       'mtu': 2000,
                                                                       'name': '277/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39508027',
                                                                       'out_octets': '3668292506795',
                                                                       'out_ucast_pkts': '2952043956',
                                                                       'poller_hash': 'NTEyNTk1cGVlcmludGYyNzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '277',
                                                                       'in_nucast_pkts': '39508027',
                                                                       'in_octets': '3668292506795',
                                                                       'in_ucast_pkts': '2952043956',
                                                                       'links': [       {       'in_nucast_pkts': '39508027',
                                                                                                'in_octets': '3668292506795',
                                                                                                'in_ucast_pkts': '2952043956',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '100',
                                                                                                'out_octets': '276441147928',
                                                                                                'out_ucast_pkts': '771257156',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '100',
                                                                       'out_octets': '276441147928',
                                                                       'out_ucast_pkts': '771257156',
                                                                       'poller_hash': 'NTEyNTk1cGVlcmludGZwb24gMjc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277',
                                                                                               'links': [       {       'connected_time': '67856183',
                                                                                                                        'fiber_length': 23387,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTk1cGVlcnBvbnBvbiAyNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1052 Henry St Wellesley (Judy & '
                                                 'Murray Bender 10411529)',
                                         'poller_hash': 'NTEyNTk1',
                                         'serial_number': '512595',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c5:27',
                                         'uptime': '67856183'},
                                 {       'counter_source': '278/F1+278/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a6:7e:ce'],
                                                                       'id': '278/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '5528738114',
                                                                       'in_ucast_pkts': '30789055',
                                                                       'name': '278/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '20617643910',
                                                                       'out_octets': '27686255611812',
                                                                       'out_ucast_pkts': '13056822',
                                                                       'poller_hash': 'ODY5MzY0cGVlcmludGYyNzgvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '278/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1514830309885',
                                                                       'in_ucast_pkts': '2190381499',
                                                                       'mac_address': 'd0:76:8f:08:13:73',
                                                                       'name': '278/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '63',
                                                                       'out_octets': '5373138454023',
                                                                       'out_ucast_pkts': '4212589865',
                                                                       'poller_hash': 'ODY5MzY0cGVlcmludGYyNzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '278',
                                                                       'in_nucast_pkts': '20617643973',
                                                                       'in_octets': '33059394065835',
                                                                       'in_ucast_pkts': '4225646687',
                                                                       'links': [       {       'in_nucast_pkts': '20617643973',
                                                                                                'in_octets': '33059394065835',
                                                                                                'in_ucast_pkts': '4225646687',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1520359047999',
                                                                                                'out_ucast_pkts': '2221170554',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1520359047999',
                                                                       'out_ucast_pkts': '2221170554',
                                                                       'poller_hash': 'ODY5MzY0cGVlcmludGZwb24gMjc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278',
                                                                                               'links': [       {       'connected_time': '611454',
                                                                                                                        'fiber_length': 19727,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzY0cGVlcnBvbnBvbiAyNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4010 Andy St Gadshill (Katie Waters '
                                                 '10434612)',
                                         'poller_hash': 'ODY5MzY0',
                                         'serial_number': '869364',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:13:73',
                                         'uptime': '611454'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '279/g1',
                                                                       'name': '279/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNEZBcGVlcmludGYyNzkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '279',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NTEyNEZBcGVlcmludGZwb24gMjc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279',
                                                                                               'links': [       {       'connected_time': '67856291',
                                                                                                                        'fiber_length': 23505,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNEZBcGVlcnBvbnBvbiAyNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1019 Molesworth St Wellesley',
                                         'poller_hash': 'NTEyNEZB',
                                         'serial_number': '5124FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:56',
                                         'uptime': '67856291'},
                                 {       'counter_source': '280/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:42:a1:f5:ce:e9'],
                                                                       'id': '280/g1',
                                                                       'in_nucast_pkts': '12552',
                                                                       'in_octets': '72529259743',
                                                                       'in_ucast_pkts': '410199904',
                                                                       'mac_address': 'cc:be:59:b9:c3:4a',
                                                                       'mtu': 2000,
                                                                       'name': '280/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8065568',
                                                                       'out_octets': '1298224061614',
                                                                       'out_ucast_pkts': '982978963',
                                                                       'poller_hash': 'NTEyNEY2cGVlcmludGYyODAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '280',
                                                                       'in_nucast_pkts': '8065568',
                                                                       'in_octets': '1298224061614',
                                                                       'in_ucast_pkts': '982978963',
                                                                       'links': [       {       'in_nucast_pkts': '8065568',
                                                                                                'in_octets': '1298224061614',
                                                                                                'in_ucast_pkts': '982978963',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12552',
                                                                                                'out_octets': '72529259743',
                                                                                                'out_ucast_pkts': '410199904',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12552',
                                                                       'out_octets': '72529259743',
                                                                       'out_ucast_pkts': '410199904',
                                                                       'poller_hash': 'NTEyNEY2cGVlcmludGZwb24gMjgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280',
                                                                                               'links': [       {       'connected_time': '6732958',
                                                                                                                        'fiber_length': 23457,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNEY2cGVlcnBvbnBvbiAyODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '90 SCHWEITZER CR (NAGY DARYL)',
                                         'poller_hash': 'NTEyNEY2',
                                         'serial_number': '5124F6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:4a',
                                         'uptime': '6732958'},
                                 {       'counter_source': '281/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e2:67'],
                                                                       'id': '281/g1',
                                                                       'in_nucast_pkts': '928895',
                                                                       'in_octets': '400555611335',
                                                                       'in_ucast_pkts': '1688932269',
                                                                       'mac_address': 'ec:4f:82:8e:06:43',
                                                                       'mtu': 2000,
                                                                       'name': '281/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '58763446',
                                                                       'out_octets': '7557537191326',
                                                                       'out_ucast_pkts': '5701787054',
                                                                       'poller_hash': 'MzIyOTBGcGVlcmludGYyODEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '281',
                                                                       'in_nucast_pkts': '58763446',
                                                                       'in_octets': '7557537191326',
                                                                       'in_ucast_pkts': '5701787054',
                                                                       'links': [       {       'in_nucast_pkts': '58763446',
                                                                                                'in_octets': '7557537191326',
                                                                                                'in_ucast_pkts': '5701787054',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '928895',
                                                                                                'out_octets': '400555611335',
                                                                                                'out_ucast_pkts': '1688932269',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '928895',
                                                                       'out_octets': '400555611335',
                                                                       'out_ucast_pkts': '1688932269',
                                                                       'poller_hash': 'MzIyOTBGcGVlcmludGZwb24gMjgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '281',
                                                                                               'links': [       {       'connected_time': '21886767',
                                                                                                                        'fiber_length': 20045,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOTBGcGVlcnBvbnBvbiAyODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2838 LINE 45 (ADAIR MEGAN)',
                                         'poller_hash': 'MzIyOTBG',
                                         'serial_number': '32290F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:06:40',
                                         'uptime': '21886767'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '282/g1',
                                                                       'name': '282/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzMwM0EycGVlcmludGYyODIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '282',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzMwM0EycGVlcmludGZwb24gMjgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzMwM0EycGVlcnBvbnBvbiAyODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '24 Pugh St. Milverton (Matt McEwen '
                                                 '10433410)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "282" '
                                                                  '(location: GPON '
                                                                  'port "7/8"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2025-06-03T14:26:21.710Z',
                                         'poller_hash': 'MzMwM0Ey',
                                         'serial_number': '3303A2',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '283/g1',
                                                                       'name': '283/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5NDExcGVlcmludGYyODMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '283',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5NDExcGVlcmludGZwb24gMjgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '283',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDExcGVlcnBvbnBvbiAyODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '72 Main St. N. Milverton (Michelle '
                                                 'Ross)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'OLT PON port 7/7 is '
                                                                  'down',
                                         'offline_time': '2024-10-17T07:03:42.740Z',
                                         'poller_hash': 'ODY5NDEx',
                                         'serial_number': '869411',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '286/g1',
                                                                       'name': '286/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkIwMUMzcGVlcmludGYyODYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '286',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkIwMUMzcGVlcmludGZwb24gMjg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '286',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkIwMUMzcGVlcnBvbnBvbiAyODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '70 Main St. N. Milverton (Jim Yost)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'OLT PON port 7/7 is '
                                                                  'down',
                                         'offline_time': '2024-10-17T07:03:42.740Z',
                                         'poller_hash': 'MkIwMUMz',
                                         'serial_number': '2B01C3',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '287/g1',
                                                                       'name': '287/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk4MjFCcGVlcmludGYyODcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '287',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mjk4MjFCcGVlcmludGZwb24gMjg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '287',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mjk4MjFCcGVlcnBvbnBvbiAyODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '5 CPR DR (KIPFER RYAN & AMY)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'OLT PON port 7/7 is '
                                                                  'down',
                                         'offline_time': '2024-10-17T07:03:42.740Z',
                                         'poller_hash': 'Mjk4MjFC',
                                         'serial_number': '29821B',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '289/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c5:60:15'],
                                                                       'id': '289/g1',
                                                                       'in_nucast_pkts': '42',
                                                                       'in_octets': '238221952467',
                                                                       'in_ucast_pkts': '734757595',
                                                                       'mac_address': 'cc:be:59:b9:c9:fe',
                                                                       'mtu': 2000,
                                                                       'name': '289/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15247182',
                                                                       'out_octets': '3721353188000',
                                                                       'out_ucast_pkts': '2882463944',
                                                                       'poller_hash': 'NTEyNzMycGVlcmludGYyODkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '289',
                                                                       'in_nucast_pkts': '15247182',
                                                                       'in_octets': '3721353188000',
                                                                       'in_ucast_pkts': '2882463944',
                                                                       'links': [       {       'in_nucast_pkts': '15247182',
                                                                                                'in_octets': '3721353188000',
                                                                                                'in_ucast_pkts': '2882463944',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '42',
                                                                                                'out_octets': '238221952467',
                                                                                                'out_ucast_pkts': '734757595',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '42',
                                                                       'out_octets': '238221952467',
                                                                       'out_ucast_pkts': '734757595',
                                                                       'poller_hash': 'NTEyNzMycGVlcmludGZwb24gMjg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '289',
                                                                                               'links': [       {       'connected_time': '13227781',
                                                                                                                        'fiber_length': 23940,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNzMycGVlcnBvbnBvbiAyODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': "1014 Queen's Bush Rd Wellesley (Jeff "
                                                 'Weeks 10434549)',
                                         'poller_hash': 'NTEyNzMy',
                                         'serial_number': '512732',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c9:fe',
                                         'uptime': '13227781'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '291/g1',
                                                                       'name': '291/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Q0MwQ0ZwZWVyaW50ZjI5MS9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '291',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Q0MwQ0ZwZWVyaW50ZnBvbiAyOTE=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '291',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0MwQ0ZwZWVycG9ucG9uIDI5MQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '7 Tower Heights Milverton (Scott '
                                                 'Smith)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'OLT PON port 7/7 is '
                                                                  'down',
                                         'offline_time': '2024-10-17T07:03:42.740Z',
                                         'poller_hash': 'Q0MwQ0Y=',
                                         'serial_number': 'CC0CF',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '293/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fd:25'],
                                                                       'id': '293/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '21321469025',
                                                                       'in_ucast_pkts': '62612936',
                                                                       'mac_address': 'cc:be:59:ab:86:4b',
                                                                       'mtu': 2000,
                                                                       'name': '293/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652598',
                                                                       'out_octets': '287702920975',
                                                                       'out_ucast_pkts': '239475325',
                                                                       'poller_hash': 'NEZBMTlDcGVlcmludGYyOTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '293',
                                                                       'in_nucast_pkts': '5652598',
                                                                       'in_octets': '287702920975',
                                                                       'in_ucast_pkts': '239475325',
                                                                       'links': [       {       'in_nucast_pkts': '5652598',
                                                                                                'in_octets': '287702920975',
                                                                                                'in_ucast_pkts': '239475325',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '21321469025',
                                                                                                'out_ucast_pkts': '62612936',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '21321469025',
                                                                       'out_ucast_pkts': '62612936',
                                                                       'poller_hash': 'NEZBMTlDcGVlcmludGZwb24gMjkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '293',
                                                                                               'links': [       {       'connected_time': '8535117',
                                                                                                                        'fiber_length': 23968,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.036,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.534}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMTlDcGVlcnBvbnBvbiAyOTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1327 Greenwood Hill Rd Wellesley '
                                                 '(Kevin Bauman 10434641)',
                                         'poller_hash': 'NEZBMTlD',
                                         'serial_number': '4FA19C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:86:4a',
                                         'uptime': '8535117'},
                                 {       'counter_source': '294/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:35:53'],
                                                                       'id': '294/g1',
                                                                       'in_nucast_pkts': '11820',
                                                                       'in_octets': '1788954217519',
                                                                       'in_ucast_pkts': '3046414785',
                                                                       'mac_address': 'ec:4f:82:8e:08:2b',
                                                                       'mtu': 2000,
                                                                       'name': '294/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62537719',
                                                                       'out_octets': '7495857184092',
                                                                       'out_ucast_pkts': '6004428677',
                                                                       'poller_hash': 'MzIyOTRDcGVlcmludGYyOTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '294',
                                                                       'in_nucast_pkts': '62537719',
                                                                       'in_octets': '7495857184092',
                                                                       'in_ucast_pkts': '6004428677',
                                                                       'links': [       {       'in_nucast_pkts': '62537719',
                                                                                                'in_octets': '7495857184092',
                                                                                                'in_ucast_pkts': '6004428677',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11820',
                                                                                                'out_octets': '1788954217519',
                                                                                                'out_ucast_pkts': '3046414785',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11820',
                                                                       'out_octets': '1788954217519',
                                                                       'out_ucast_pkts': '3046414785',
                                                                       'poller_hash': 'MzIyOTRDcGVlcmludGZwb24gMjk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '294',
                                                                                               'links': [       {       'connected_time': '21886294',
                                                                                                                        'fiber_length': 23947,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOTRDcGVlcnBvbnBvbiAyOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1010 QUEEN'S BUSH RD (SPROULE BOB)",
                                         'poller_hash': 'MzIyOTRD',
                                         'serial_number': '32294C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:08:28',
                                         'uptime': '21886294'},
                                 {       'counter_source': '295/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fc:b9'],
                                                                       'id': '295/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '89297799324',
                                                                       'in_ucast_pkts': '476447870',
                                                                       'mac_address': 'ec:4f:82:8e:05:e3',
                                                                       'mtu': 2000,
                                                                       'name': '295/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797946',
                                                                       'out_octets': '1599355253439',
                                                                       'out_ucast_pkts': '1127670241',
                                                                       'poller_hash': 'MzIyOTAzcGVlcmludGYyOTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '295',
                                                                       'in_nucast_pkts': '9797946',
                                                                       'in_octets': '1599355253439',
                                                                       'in_ucast_pkts': '1127670241',
                                                                       'links': [       {       'in_nucast_pkts': '9797946',
                                                                                                'in_octets': '1599355253439',
                                                                                                'in_ucast_pkts': '1127670241',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '89297799324',
                                                                                                'out_ucast_pkts': '476447870',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '89297799324',
                                                                       'out_ucast_pkts': '476447870',
                                                                       'poller_hash': 'MzIyOTAzcGVlcmludGZwb24gMjk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '295',
                                                                                               'links': [       {       'connected_time': '8535242',
                                                                                                                        'fiber_length': 24499,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOTAzcGVlcnBvbnBvbiAyOTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1097 QUEEN'S BUSH RD (SEBBEN TED)",
                                         'poller_hash': 'MzIyOTAz',
                                         'serial_number': '322903',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:05:e0',
                                         'uptime': '8535242'},
                                 {       'counter_source': '297/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '38:d5:47:40:2b:f4'],
                                                                       'id': '297/g1',
                                                                       'in_nucast_pkts': '2570',
                                                                       'in_octets': '34428683739',
                                                                       'in_ucast_pkts': '130026461',
                                                                       'mac_address': 'cc:be:59:ab:88:d5',
                                                                       'mtu': 2000,
                                                                       'name': '297/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530410',
                                                                       'out_octets': '413523205286',
                                                                       'out_ucast_pkts': '345177675',
                                                                       'poller_hash': 'NEZBMjQ3cGVlcmludGYyOTcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '297',
                                                                       'in_nucast_pkts': '5530410',
                                                                       'in_octets': '413523205286',
                                                                       'in_ucast_pkts': '345177675',
                                                                       'links': [       {       'in_nucast_pkts': '5530410',
                                                                                                'in_octets': '413523205286',
                                                                                                'in_ucast_pkts': '345177675',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2570',
                                                                                                'out_octets': '34428683739',
                                                                                                'out_ucast_pkts': '130026461',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2570',
                                                                       'out_octets': '34428683739',
                                                                       'out_ucast_pkts': '130026461',
                                                                       'poller_hash': 'NEZBMjQ3cGVlcmludGZwb24gMjk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '297',
                                                                                               'links': [       {       'connected_time': '8535107',
                                                                                                                        'fiber_length': 24784,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.944,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.154}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMjQ3cGVlcnBvbnBvbiAyOTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '109 Brown St Wellesley (John Thaler '
                                                 '10434553)',
                                         'poller_hash': 'NEZBMjQ3',
                                         'serial_number': '4FA247',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:ab:88:d4',
                                         'uptime': '8535107'},
                                 {       'counter_source': '298/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:0f:ac'],
                                                                       'id': '298/g1',
                                                                       'in_nucast_pkts': '243',
                                                                       'in_octets': '324876426963',
                                                                       'in_ucast_pkts': '1111161128',
                                                                       'mac_address': 'cc:be:59:b9:c4:10',
                                                                       'mtu': 2000,
                                                                       'name': '298/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39489278',
                                                                       'out_octets': '5509505469287',
                                                                       'out_ucast_pkts': '4064191472',
                                                                       'poller_hash': 'NTEyNTM4cGVlcmludGYyOTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '298',
                                                                       'in_nucast_pkts': '39489278',
                                                                       'in_octets': '5509505469287',
                                                                       'in_ucast_pkts': '4064191472',
                                                                       'links': [       {       'in_nucast_pkts': '39489278',
                                                                                                'in_octets': '5509505469287',
                                                                                                'in_ucast_pkts': '4064191472',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '243',
                                                                                                'out_octets': '324876426963',
                                                                                                'out_ucast_pkts': '1111161128',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '243',
                                                                       'out_octets': '324876426963',
                                                                       'out_ucast_pkts': '1111161128',
                                                                       'poller_hash': 'NTEyNTM4cGVlcmludGZwb24gMjk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '298',
                                                                                               'links': [       {       'connected_time': '67856134',
                                                                                                                        'fiber_length': 22555,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTM4cGVlcnBvbnBvbiAyOTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1159 Henry St Wellelsey (Kirk Bergey '
                                                 '10434750)',
                                         'poller_hash': 'NTEyNTM4',
                                         'serial_number': '512538',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c4:10',
                                         'uptime': '67856134'},
                                 {       'counter_source': '299/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:10:ba'],
                                                                       'id': '299/g1',
                                                                       'in_nucast_pkts': '26972659',
                                                                       'in_octets': '438832582488',
                                                                       'in_ucast_pkts': '2394308559',
                                                                       'mac_address': 'ec:4f:82:8d:4a:2b',
                                                                       'mtu': 2000,
                                                                       'name': '299/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26297412834',
                                                                       'out_octets': '44515973023648',
                                                                       'out_ucast_pkts': '7050970844',
                                                                       'poller_hash': 'MzIxMThDcGVlcmludGYyOTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '299',
                                                                       'in_nucast_pkts': '26297412834',
                                                                       'in_octets': '44515973023648',
                                                                       'in_ucast_pkts': '7050970844',
                                                                       'links': [       {       'in_nucast_pkts': '26297412834',
                                                                                                'in_octets': '44515973023648',
                                                                                                'in_ucast_pkts': '7050970844',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '26972659',
                                                                                                'out_octets': '438832582488',
                                                                                                'out_ucast_pkts': '2394308559',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26972659',
                                                                       'out_octets': '438832582488',
                                                                       'out_ucast_pkts': '2394308559',
                                                                       'poller_hash': 'MzIxMThDcGVlcmludGZwb24gMjk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '299',
                                                                                               'links': [       {       'connected_time': '21886492',
                                                                                                                        'fiber_length': 24693,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMThDcGVlcnBvbnBvbiAyOTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1099 Queen's Bush Rd Wellesley "
                                                 '(Stephen Bender 10434690)',
                                         'poller_hash': 'MzIxMThD',
                                         'serial_number': '32118C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:4a:28',
                                         'uptime': '21886492'},
                                 {       'counter_source': '300/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fe:7b'],
                                                                       'id': '300/g1',
                                                                       'in_nucast_pkts': '9554',
                                                                       'in_octets': '192128548954',
                                                                       'in_ucast_pkts': '858954285',
                                                                       'mac_address': 'ec:4f:82:8e:06:03',
                                                                       'mtu': 2000,
                                                                       'name': '300/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62541723',
                                                                       'out_octets': '3325669541378',
                                                                       'out_ucast_pkts': '2476221918',
                                                                       'poller_hash': 'MzIyOTA3cGVlcmludGYzMDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '300',
                                                                       'in_nucast_pkts': '62541723',
                                                                       'in_octets': '3325669541378',
                                                                       'in_ucast_pkts': '2476221918',
                                                                       'links': [       {       'in_nucast_pkts': '62541723',
                                                                                                'in_octets': '3325669541378',
                                                                                                'in_ucast_pkts': '2476221918',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9554',
                                                                                                'out_octets': '192128548954',
                                                                                                'out_ucast_pkts': '858954285',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9554',
                                                                       'out_octets': '192128548954',
                                                                       'out_ucast_pkts': '858954285',
                                                                       'poller_hash': 'MzIyOTA3cGVlcmludGZwb24gMzAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '300',
                                                                                               'links': [       {       'connected_time': '21886928',
                                                                                                                        'fiber_length': 24794,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyOTA3cGVlcnBvbnBvbiAzMDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '101 Brown St Wellesley (Mary Jane '
                                                 'Falk 10434554)',
                                         'poller_hash': 'MzIyOTA3',
                                         'serial_number': '322907',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:06:00',
                                         'uptime': '21886928'},
                                 {       'counter_source': '301/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:00:58'],
                                                                       'id': '301/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '174659753483',
                                                                       'in_ucast_pkts': '492003005',
                                                                       'mac_address': 'cc:be:59:50:00:de',
                                                                       'mtu': 2000,
                                                                       'name': '301/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1647251',
                                                                       'out_octets': '940506082993',
                                                                       'out_ucast_pkts': '728204845',
                                                                       'poller_hash': 'NDNCNDM3cGVlcmludGYzMDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '301',
                                                                       'in_nucast_pkts': '1647251',
                                                                       'in_octets': '940506082993',
                                                                       'in_ucast_pkts': '728204845',
                                                                       'links': [       {       'in_nucast_pkts': '1647251',
                                                                                                'in_octets': '940506082993',
                                                                                                'in_ucast_pkts': '728204845',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '174659753483',
                                                                                                'out_ucast_pkts': '492003005',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '174659753483',
                                                                       'out_ucast_pkts': '492003005',
                                                                       'poller_hash': 'NDNCNDM3cGVlcmludGZwb24gMzAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '301',
                                                                                               'links': [       {       'connected_time': '2414745',
                                                                                                                        'fiber_length': 22440,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.974,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.224}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNDM3cGVlcnBvbnBvbiAzMDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': "1205 Queen's Bush Rd Wellesley "
                                                 '(Kristen Lobrutto 10434675)',
                                         'poller_hash': 'NDNCNDM3',
                                         'serial_number': '43B437',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:00:dd',
                                         'uptime': '2414745'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '302/g1',
                                                                       'name': '302/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCOUI1cGVlcmludGYzMDIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '302',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJCOUI1cGVlcmludGZwb24gMzAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '302',
                                                                                               'links': [       {       'connected_time': '8535243',
                                                                                                                        'fiber_length': 24888,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUI1cGVlcnBvbnBvbiAzMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '130 Brown St Wellesley',
                                         'poller_hash': 'NDJCOUI1',
                                         'serial_number': '42B9B5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:42:ea',
                                         'uptime': '8535243'},
                                 {       'counter_source': '303/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fe:72'],
                                                                       'id': '303/g1',
                                                                       'in_nucast_pkts': '7437',
                                                                       'in_octets': '331576334480',
                                                                       'in_ucast_pkts': '1725070538',
                                                                       'mac_address': 'cc:be:59:cb:42:45',
                                                                       'mtu': 2000,
                                                                       'name': '303/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710363',
                                                                       'out_octets': '7244642611165',
                                                                       'out_ucast_pkts': '5343141845',
                                                                       'poller_hash': 'NDJCOUEwcGVlcmludGYzMDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '303',
                                                                       'in_nucast_pkts': '61710363',
                                                                       'in_octets': '7244642611165',
                                                                       'in_ucast_pkts': '5343141845',
                                                                       'links': [       {       'in_nucast_pkts': '61710363',
                                                                                                'in_octets': '7244642611165',
                                                                                                'in_ucast_pkts': '5343141845',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7437',
                                                                                                'out_octets': '331576334480',
                                                                                                'out_ucast_pkts': '1725070538',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7437',
                                                                       'out_octets': '331576334480',
                                                                       'out_ucast_pkts': '1725070538',
                                                                       'poller_hash': 'NDJCOUEwcGVlcmludGZwb24gMzAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '303',
                                                                                               'links': [       {       'connected_time': '21886656',
                                                                                                                        'fiber_length': 23528,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUEwcGVlcnBvbnBvbiAzMDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1021 Maple Leaf St Wellesley (Ronald '
                                                 'Bisch 10434692)',
                                         'poller_hash': 'NDJCOUEw',
                                         'serial_number': '42B9A0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:42:42',
                                         'uptime': '21886656'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '304/g1',
                                                                       'name': '304/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCMzRGcGVlcmludGYzMDQvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '304',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCMzRGcGVlcmludGZwb24gMzA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '304',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMzRGcGVlcnBvbnBvbiAzMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '146 David St Wellesley',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "304" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:09.710Z',
                                         'poller_hash': 'NDNCMzRG',
                                         'serial_number': '43B34F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '305/g1',
                                                                       'name': '305/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDI4MUZCcGVlcmludGYzMDUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '305',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDI4MUZCcGVlcmludGZwb24gMzA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '305',
                                                                                               'links': [       {       'connected_time': '21887218',
                                                                                                                        'fiber_length': 23997,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MUZCcGVlcnBvbnBvbiAzMDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1019 QUEEN'S BUSH RD (LEIS DOLORES)",
                                         'poller_hash': 'NDI4MUZC',
                                         'serial_number': '4281FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:df:e2',
                                         'uptime': '21887218'},
                                 {       'counter_source': '306/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:04:51'],
                                                                       'id': '306/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '106429346226',
                                                                       'in_ucast_pkts': '574896588',
                                                                       'mac_address': 'cc:be:59:4f:fe:a0',
                                                                       'mtu': 2000,
                                                                       'name': '306/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530459',
                                                                       'out_octets': '2867086059576',
                                                                       'out_ucast_pkts': '2036788268',
                                                                       'poller_hash': 'NDNCMzE5cGVlcmludGYzMDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '306',
                                                                       'in_nucast_pkts': '5530459',
                                                                       'in_octets': '2867086059576',
                                                                       'in_ucast_pkts': '2036788268',
                                                                       'links': [       {       'in_nucast_pkts': '5530459',
                                                                                                'in_octets': '2867086059576',
                                                                                                'in_ucast_pkts': '2036788268',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '106429346226',
                                                                                                'out_ucast_pkts': '574896588',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '106429346226',
                                                                       'out_ucast_pkts': '574896588',
                                                                       'poller_hash': 'NDNCMzE5cGVlcmludGZwb24gMzA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '306',
                                                                                               'links': [       {       'connected_time': '8535079',
                                                                                                                        'fiber_length': 24459,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.428,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.378}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMzE5cGVlcnBvbnBvbiAzMDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': "1098 Queen's Bush Rd Wellesley (Dan "
                                                 'Koebel 10434542)',
                                         'poller_hash': 'NDNCMzE5',
                                         'serial_number': '43B319',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:fe:9f',
                                         'uptime': '8535079'},
                                 {       'counter_source': '307/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5a:f8'],
                                                                       'id': '307/g1',
                                                                       'in_nucast_pkts': '155944',
                                                                       'in_octets': '178200524791',
                                                                       'in_ucast_pkts': '638656477',
                                                                       'mac_address': 'cc:be:59:4f:ff:ba',
                                                                       'mtu': 2000,
                                                                       'name': '307/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652551',
                                                                       'out_octets': '2548673535317',
                                                                       'out_ucast_pkts': '1934248442',
                                                                       'poller_hash': 'NDNCM0E2cGVlcmludGYzMDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '307',
                                                                       'in_nucast_pkts': '5652551',
                                                                       'in_octets': '2548673535317',
                                                                       'in_ucast_pkts': '1934248442',
                                                                       'links': [       {       'in_nucast_pkts': '5652551',
                                                                                                'in_octets': '2548673535317',
                                                                                                'in_ucast_pkts': '1934248442',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '155944',
                                                                                                'out_octets': '178200524791',
                                                                                                'out_ucast_pkts': '638656477',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '155944',
                                                                       'out_octets': '178200524791',
                                                                       'out_ucast_pkts': '638656477',
                                                                       'poller_hash': 'NDNCM0E2cGVlcmludGZwb24gMzA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '307',
                                                                                               'links': [       {       'connected_time': '8535106',
                                                                                                                        'fiber_length': 24304,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.598,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.53}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0E2cGVlcnBvbnBvbiAzMDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '87 PARKVIEW DR (DAVIDSON IVAN)',
                                         'poller_hash': 'NDNCM0E2',
                                         'serial_number': '43B3A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:ff:b9',
                                         'uptime': '8535106'},
                                 {       'counter_source': '308/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:72:f7'],
                                                                       'id': '308/g1',
                                                                       'in_nucast_pkts': '550161',
                                                                       'in_octets': '299652919442',
                                                                       'in_ucast_pkts': '3114167925',
                                                                       'mac_address': 'ec:4f:82:8e:00:03',
                                                                       'mtu': 2000,
                                                                       'name': '308/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32282471',
                                                                       'out_octets': '9008061225855',
                                                                       'out_ucast_pkts': '6205775920',
                                                                       'poller_hash': 'MzIyODQ3cGVlcmludGYzMDgvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '308',
                                                                       'in_nucast_pkts': '32282471',
                                                                       'in_octets': '9008061225855',
                                                                       'in_ucast_pkts': '6205775920',
                                                                       'links': [       {       'in_nucast_pkts': '32282471',
                                                                                                'in_octets': '9008061225855',
                                                                                                'in_ucast_pkts': '6205775920',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '550161',
                                                                                                'out_octets': '299652919442',
                                                                                                'out_ucast_pkts': '3114167925',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '550161',
                                                                       'out_octets': '299652919442',
                                                                       'out_ucast_pkts': '3114167925',
                                                                       'poller_hash': 'MzIyODQ3cGVlcmludGZwb24gMzA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '308',
                                                                                               'links': [       {       'connected_time': '29479976',
                                                                                                                        'fiber_length': 24047,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIyODQ3cGVlcnBvbnBvbiAzMDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1300 Greenwood Hill Dr Wellesley (A '
                                                 '& B Detzler 10434514)',
                                         'poller_hash': 'MzIyODQ3',
                                         'serial_number': '322847',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8e:00:00',
                                         'uptime': '29479976'},
                                 {       'counter_source': '309/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fc:a7'],
                                                                       'id': '309/g1',
                                                                       'in_nucast_pkts': '30',
                                                                       'in_octets': '121222678585',
                                                                       'in_ucast_pkts': '503826213',
                                                                       'mac_address': 'cc:be:59:f8:65:05',
                                                                       'mtu': 2000,
                                                                       'name': '309/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919654',
                                                                       'out_octets': '1230070811571',
                                                                       'out_ucast_pkts': '879613787',
                                                                       'poller_hash': 'NTc0NTkycGVlcmludGYzMDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '309',
                                                                       'in_nucast_pkts': '9919654',
                                                                       'in_octets': '1230070811571',
                                                                       'in_ucast_pkts': '879613787',
                                                                       'links': [       {       'in_nucast_pkts': '9919654',
                                                                                                'in_octets': '1230070811571',
                                                                                                'in_ucast_pkts': '879613787',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '30',
                                                                                                'out_octets': '121222678585',
                                                                                                'out_ucast_pkts': '503826213',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30',
                                                                       'out_octets': '121222678585',
                                                                       'out_ucast_pkts': '503826213',
                                                                       'poller_hash': 'NTc0NTkycGVlcmludGZwb24gMzA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '309',
                                                                                               'links': [       {       'connected_time': '8535199',
                                                                                                                        'fiber_length': 23400,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.4,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.372}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0NTkycGVlcnBvbnBvbiAzMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': '19 REINER CR (BENDER TIFFANY)',
                                         'poller_hash': 'NTc0NTky',
                                         'serial_number': '574592',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:64:ff',
                                         'uptime': '8535199'},
                                 {       'counter_source': '310/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:36:3d'],
                                                                       'id': '310/g1',
                                                                       'in_nucast_pkts': '40',
                                                                       'in_octets': '112278511276',
                                                                       'in_ucast_pkts': '394295437',
                                                                       'mac_address': 'cc:be:59:f8:5d:a5',
                                                                       'mtu': 2000,
                                                                       'name': '310/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919585',
                                                                       'out_octets': '1033497913579',
                                                                       'out_ucast_pkts': '935612746',
                                                                       'poller_hash': 'NTc0NEE5cGVlcmludGYzMTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '310',
                                                                       'in_nucast_pkts': '9919585',
                                                                       'in_octets': '1033497913579',
                                                                       'in_ucast_pkts': '935612746',
                                                                       'links': [       {       'in_nucast_pkts': '9919585',
                                                                                                'in_octets': '1033497913579',
                                                                                                'in_ucast_pkts': '935612746',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '40',
                                                                                                'out_octets': '112278511276',
                                                                                                'out_ucast_pkts': '394295437',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '40',
                                                                       'out_octets': '112278511276',
                                                                       'out_ucast_pkts': '394295437',
                                                                       'poller_hash': 'NTc0NEE5cGVlcmludGZwb24gMzEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '310',
                                                                                               'links': [       {       'connected_time': '8535200',
                                                                                                                        'fiber_length': 25017,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.596,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.388}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0NEE5cGVlcnBvbnBvbiAzMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': '64 VILLAGE RD (BAIRD MICHAEL (MIKE))',
                                         'poller_hash': 'NTc0NEE5',
                                         'serial_number': '5744A9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:5d:9f',
                                         'uptime': '8535200'},
                                 {       'counter_source': '311/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:b7'],
                                                                       'id': '311/g1',
                                                                       'in_nucast_pkts': '155891',
                                                                       'in_octets': '192453694129',
                                                                       'in_ucast_pkts': '421999659',
                                                                       'mac_address': 'cc:be:59:4f:ff:62',
                                                                       'mtu': 2000,
                                                                       'name': '311/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5648042',
                                                                       'out_octets': '1056115836004',
                                                                       'out_ucast_pkts': '848927460',
                                                                       'poller_hash': 'NDNCMzdBcGVlcmludGYzMTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '311',
                                                                       'in_nucast_pkts': '5648042',
                                                                       'in_octets': '1056115836004',
                                                                       'in_ucast_pkts': '848927460',
                                                                       'links': [       {       'in_nucast_pkts': '5648042',
                                                                                                'in_octets': '1056115836004',
                                                                                                'in_ucast_pkts': '848927460',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '155891',
                                                                                                'out_octets': '192453694129',
                                                                                                'out_ucast_pkts': '421999659',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '155891',
                                                                       'out_octets': '192453694129',
                                                                       'out_ucast_pkts': '421999659',
                                                                       'poller_hash': 'NDNCMzdBcGVlcmludGZwb24gMzEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '311',
                                                                                               'links': [       {       'connected_time': '8535039',
                                                                                                                        'fiber_length': 24421,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -23.63,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.854}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMzdBcGVlcnBvbnBvbiAzMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '34 Parkview Dr Wellesley (Eric '
                                                 'Jeffray10434740)',
                                         'poller_hash': 'NDNCMzdB',
                                         'serial_number': '43B37A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:ff:61',
                                         'uptime': '8535039'},
                                 {       'counter_source': '313/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '74:fe:ce:31:9c:d1'],
                                                                       'id': '313/g1',
                                                                       'in_nucast_pkts': '8672',
                                                                       'in_octets': '153177722195',
                                                                       'in_ucast_pkts': '352691183',
                                                                       'mac_address': 'cc:be:59:4f:ff:de',
                                                                       'mtu': 2000,
                                                                       'name': '313/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530310',
                                                                       'out_octets': '920290385552',
                                                                       'out_ucast_pkts': '780125252',
                                                                       'poller_hash': 'NDNCM0I4cGVlcmludGYzMTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '313',
                                                                       'in_nucast_pkts': '5530310',
                                                                       'in_octets': '920290385552',
                                                                       'in_ucast_pkts': '780125252',
                                                                       'links': [       {       'in_nucast_pkts': '5530310',
                                                                                                'in_octets': '920290385552',
                                                                                                'in_ucast_pkts': '780125252',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8672',
                                                                                                'out_octets': '153177722195',
                                                                                                'out_ucast_pkts': '352691183',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8672',
                                                                       'out_octets': '153177722195',
                                                                       'out_ucast_pkts': '352691183',
                                                                       'poller_hash': 'NDNCM0I4cGVlcmludGZwb24gMzEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '313',
                                                                                               'links': [       {       'connected_time': '8535085',
                                                                                                                        'fiber_length': 23875,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.928,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.572}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0I4cGVlcnBvbnBvbiAzMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1260 GREENWOOD HILL RD (SLOTHOUBER '
                                                 'STEPHEN)',
                                         'poller_hash': 'NDNCM0I4',
                                         'serial_number': '43B3B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:ff:dd',
                                         'uptime': '8535085'},
                                 {       'counter_source': '315/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:08:4a'],
                                                                       'id': '315/g1',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '111794116405',
                                                                       'in_ucast_pkts': '754612646',
                                                                       'mac_address': 'cc:be:59:cb:44:95',
                                                                       'mtu': 2000,
                                                                       'name': '315/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919989',
                                                                       'out_octets': '2637239676143',
                                                                       'out_ucast_pkts': '1830911167',
                                                                       'poller_hash': 'NDJCOUVBcGVlcmludGYzMTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '315',
                                                                       'in_nucast_pkts': '9919989',
                                                                       'in_octets': '2637239676143',
                                                                       'in_ucast_pkts': '1830911167',
                                                                       'links': [       {       'in_nucast_pkts': '9919989',
                                                                                                'in_octets': '2637239676143',
                                                                                                'in_ucast_pkts': '1830911167',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19',
                                                                                                'out_octets': '111794116405',
                                                                                                'out_ucast_pkts': '754612646',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19',
                                                                       'out_octets': '111794116405',
                                                                       'out_ucast_pkts': '754612646',
                                                                       'poller_hash': 'NDJCOUVBcGVlcmludGZwb24gMzE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '315',
                                                                                               'links': [       {       'connected_time': '8535212',
                                                                                                                        'fiber_length': 24412,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUVBcGVlcnBvbnBvbiAzMTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '38 PARKVIEW DR (TABONE TESHIA)',
                                         'poller_hash': 'NDJCOUVB',
                                         'serial_number': '42B9EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:92',
                                         'uptime': '8535212'},
                                 {       'counter_source': '316/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '20:7c:14:a0:b8:fe'],
                                                                       'id': '316/g1',
                                                                       'in_nucast_pkts': '20462',
                                                                       'in_octets': '895982562897',
                                                                       'in_ucast_pkts': '1519472987',
                                                                       'mac_address': '84:d3:43:7d:44:f2',
                                                                       'mtu': 2000,
                                                                       'name': '316/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17859324',
                                                                       'out_octets': '4186803094924',
                                                                       'out_ucast_pkts': '3534011745',
                                                                       'poller_hash': 'QzgxRTA3cGVlcmludGYzMTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '316',
                                                                       'in_nucast_pkts': '17859324',
                                                                       'in_octets': '4186803094924',
                                                                       'in_ucast_pkts': '3534011745',
                                                                       'links': [       {       'in_nucast_pkts': '17859324',
                                                                                                'in_octets': '4186803094924',
                                                                                                'in_ucast_pkts': '3534011745',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20462',
                                                                                                'out_octets': '895982562897',
                                                                                                'out_ucast_pkts': '1519472987',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20462',
                                                                       'out_octets': '895982562897',
                                                                       'out_ucast_pkts': '1519472987',
                                                                       'poller_hash': 'QzgxRTA3cGVlcmludGZwb24gMzE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '316',
                                                                                               'links': [       {       'connected_time': '15182577',
                                                                                                                        'fiber_length': 24726,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTA3cGVlcnBvbnBvbiAzMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '58 Nithbank Rd Wellesley (CHOW ANDY)',
                                         'poller_hash': 'QzgxRTA3',
                                         'serial_number': 'C81E07',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:ef',
                                         'uptime': '15182577'},
                                 {       'counter_source': '317/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fd:c7'],
                                                                       'id': '317/g1',
                                                                       'in_nucast_pkts': '421',
                                                                       'in_octets': '403905662401',
                                                                       'in_ucast_pkts': '1682819898',
                                                                       'mac_address': 'cc:be:59:b9:c3:e9',
                                                                       'mtu': 2000,
                                                                       'name': '317/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39526176',
                                                                       'out_octets': '9883649868712',
                                                                       'out_ucast_pkts': '7137187591',
                                                                       'poller_hash': 'NTEyNTJCcGVlcmludGYzMTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '317',
                                                                       'in_nucast_pkts': '39526176',
                                                                       'in_octets': '9883649868712',
                                                                       'in_ucast_pkts': '7137187591',
                                                                       'links': [       {       'in_nucast_pkts': '39526176',
                                                                                                'in_octets': '9883649868712',
                                                                                                'in_ucast_pkts': '7137187591',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '421',
                                                                                                'out_octets': '403905662401',
                                                                                                'out_ucast_pkts': '1682819898',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '421',
                                                                       'out_octets': '403905662401',
                                                                       'out_ucast_pkts': '1682819898',
                                                                       'poller_hash': 'NTEyNTJCcGVlcmludGZwb24gMzE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '317',
                                                                                               'links': [       {       'connected_time': '67856370',
                                                                                                                        'fiber_length': 24468,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTJCcGVlcnBvbnBvbiAzMTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '45 PARKVIEW DR (GILCHRIST SHELDON)',
                                         'poller_hash': 'NTEyNTJC',
                                         'serial_number': '51252B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:e9',
                                         'uptime': '67856370'},
                                 {       'counter_source': '318/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fe:d5'],
                                                                       'id': '318/g1',
                                                                       'in_nucast_pkts': '17',
                                                                       'in_octets': '37599419752',
                                                                       'in_ucast_pkts': '133318278',
                                                                       'mac_address': 'cc:be:59:cb:43:7d',
                                                                       'mtu': 2000,
                                                                       'name': '318/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3388796',
                                                                       'out_octets': '369222482907',
                                                                       'out_ucast_pkts': '291334427',
                                                                       'poller_hash': 'NDJCOUM3cGVlcmludGYzMTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '318',
                                                                       'in_nucast_pkts': '3388796',
                                                                       'in_octets': '369222482907',
                                                                       'in_ucast_pkts': '291334427',
                                                                       'links': [       {       'in_nucast_pkts': '3388796',
                                                                                                'in_octets': '369222482907',
                                                                                                'in_ucast_pkts': '291334427',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '17',
                                                                                                'out_octets': '37599419752',
                                                                                                'out_ucast_pkts': '133318278',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17',
                                                                       'out_octets': '37599419752',
                                                                       'out_ucast_pkts': '133318278',
                                                                       'poller_hash': 'NDJCOUM3cGVlcmludGZwb24gMzE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '318',
                                                                                               'links': [       {       'connected_time': '2866710',
                                                                                                                        'fiber_length': 24565,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUM3cGVlcnBvbnBvbiAzMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '110 NITHBANK RD (SZIMANSKI JODI)',
                                         'poller_hash': 'NDJCOUM3',
                                         'serial_number': '42B9C7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:7a',
                                         'uptime': '2866710'},
                                 {       'counter_source': '319/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b0:6a:41:d0:c9:7c'],
                                                                       'id': '319/g1',
                                                                       'in_nucast_pkts': '28901',
                                                                       'in_octets': '13615991213969',
                                                                       'in_ucast_pkts': '37502976656',
                                                                       'mac_address': 'cc:be:59:b9:c6:50',
                                                                       'mtu': 2000,
                                                                       'name': '319/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38469367',
                                                                       'out_octets': '17558818234217',
                                                                       'out_ucast_pkts': '45938305691',
                                                                       'poller_hash': 'NTEyNUY4cGVlcmludGYzMTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '319',
                                                                       'in_nucast_pkts': '38469367',
                                                                       'in_octets': '17558818234217',
                                                                       'in_ucast_pkts': '45938305691',
                                                                       'links': [       {       'in_nucast_pkts': '38469367',
                                                                                                'in_octets': '17558818234217',
                                                                                                'in_ucast_pkts': '45938305691',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '28901',
                                                                                                'out_octets': '13615991213969',
                                                                                                'out_ucast_pkts': '37502976656',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '28901',
                                                                       'out_octets': '13615991213969',
                                                                       'out_ucast_pkts': '37502976656',
                                                                       'poller_hash': 'NTEyNUY4cGVlcmludGZwb24gMzE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '319',
                                                                                               'links': [       {       'connected_time': '67856266',
                                                                                                                        'fiber_length': 23948,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUY4cGVlcnBvbnBvbiAzMTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1256 Greenwoodhill Dr. (Anson Field)',
                                         'poller_hash': 'NTEyNUY4',
                                         'serial_number': '5125F8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c6:50',
                                         'uptime': '67856266'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '320/g1',
                                                                       'name': '320/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCOUNGcGVlcmludGYzMjAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '320',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCOUNGcGVlcmludGZwb24gMzIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '320',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUNGcGVlcnBvbnBvbiAzMjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '61 Welwood Av Wellesley',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "320" '
                                                                  '(location: GPON '
                                                                  'port "6/5"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:14.280Z',
                                         'poller_hash': 'NDJCOUNG',
                                         'serial_number': '42B9CF',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '321/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '321/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '235190663320',
                                                                       'in_ucast_pkts': '658250763',
                                                                       'mac_address': '60:db:98:72:ec:27',
                                                                       'name': '321/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '2434597226697',
                                                                       'out_ucast_pkts': '1882412849',
                                                                       'poller_hash': 'QjA4REE1cGVlcmludGYzMjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '321',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '2434597226697',
                                                                       'in_ucast_pkts': '1882412849',
                                                                       'links': [       {       'in_nucast_pkts': '66',
                                                                                                'in_octets': '2434597226697',
                                                                                                'in_ucast_pkts': '1882412849',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '235190663320',
                                                                                                'out_ucast_pkts': '658250763',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '235190663320',
                                                                       'out_ucast_pkts': '658250763',
                                                                       'poller_hash': 'QjA4REE1cGVlcmludGZwb24gMzIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '321',
                                                                                               'links': [       {       'connected_time': '1035063',
                                                                                                                        'fiber_length': 24561,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjA4REE1cGVlcnBvbnBvbiAzMjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '114 Nithbank Rd (Bryce Bramhill '
                                                 '10434766)',
                                         'poller_hash': 'QjA4REE1',
                                         'serial_number': 'B08DA5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:72:ec:27',
                                         'uptime': '1035063'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '322/g1',
                                                                       'name': '322/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCQTA1cGVlcmludGYzMjIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '322',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCQTA1cGVlcmludGZwb24gMzIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '322',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCQTA1cGVlcnBvbnBvbiAzMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': "1149 QUEEN'S BUSH RD",
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "322" '
                                                                  '(location: GPON '
                                                                  'port "7/3"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-07-30T20:36:31.770Z',
                                         'poller_hash': 'NDJCQTA1',
                                         'serial_number': '42BA05',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '323/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '38:2c:4a:c7:10:75'],
                                                                       'id': '323/g1',
                                                                       'in_nucast_pkts': '1250',
                                                                       'in_octets': '40871036399',
                                                                       'in_ucast_pkts': '177383935',
                                                                       'mac_address': 'cc:be:59:50:00:10',
                                                                       'mtu': 2000,
                                                                       'name': '323/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5252740',
                                                                       'out_octets': '986242722940',
                                                                       'out_ucast_pkts': '721239998',
                                                                       'poller_hash': 'NDNCM0QxcGVlcmludGYzMjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '323',
                                                                       'in_nucast_pkts': '5252740',
                                                                       'in_octets': '986242722940',
                                                                       'in_ucast_pkts': '721239998',
                                                                       'links': [       {       'in_nucast_pkts': '5252740',
                                                                                                'in_octets': '986242722940',
                                                                                                'in_ucast_pkts': '721239998',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1250',
                                                                                                'out_octets': '40871036399',
                                                                                                'out_ucast_pkts': '177383935',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1250',
                                                                       'out_octets': '40871036399',
                                                                       'out_ucast_pkts': '177383935',
                                                                       'poller_hash': 'NDNCM0QxcGVlcmludGZwb24gMzIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '323',
                                                                                               'links': [       {       'connected_time': '8535090',
                                                                                                                        'fiber_length': 24572,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.696,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.046}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0QxcGVlcnBvbnBvbiAzMjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '69 Parkview Dr (Jon Lambert '
                                                 '10434823)',
                                         'poller_hash': 'NDNCM0Qx',
                                         'serial_number': '43B3D1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:00:0f',
                                         'uptime': '8535090'},
                                 {       'counter_source': '324/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:35:92'],
                                                                       'id': '324/g1',
                                                                       'in_nucast_pkts': '11961',
                                                                       'in_octets': '887448087193',
                                                                       'in_ucast_pkts': '5167786253',
                                                                       'mac_address': 'cc:be:59:cb:43:45',
                                                                       'mtu': 2000,
                                                                       'name': '324/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62542202',
                                                                       'out_octets': '19184215617004',
                                                                       'out_ucast_pkts': '14180667945',
                                                                       'poller_hash': 'NDJCOUMwcGVlcmludGYzMjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '324',
                                                                       'in_nucast_pkts': '62542202',
                                                                       'in_octets': '19184215617004',
                                                                       'in_ucast_pkts': '14180667945',
                                                                       'links': [       {       'in_nucast_pkts': '62542202',
                                                                                                'in_octets': '19184215617004',
                                                                                                'in_ucast_pkts': '14180667945',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11961',
                                                                                                'out_octets': '887448087193',
                                                                                                'out_ucast_pkts': '5167786253',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11961',
                                                                       'out_octets': '887448087193',
                                                                       'out_ucast_pkts': '5167786253',
                                                                       'poller_hash': 'NDJCOUMwcGVlcmludGZwb24gMzI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '324',
                                                                                               'links': [       {       'connected_time': '21887381',
                                                                                                                        'fiber_length': 23986,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUMwcGVlcnBvbnBvbiAzMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1005 Queen's Bush Rd (Jeremy "
                                                 'Stemmler 10434548)',
                                         'poller_hash': 'NDJCOUMw',
                                         'serial_number': '42B9C0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:42',
                                         'uptime': '21887381'},
                                 {       'counter_source': '325/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '325/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '551630199953',
                                                                       'in_ucast_pkts': '1175969188',
                                                                       'mac_address': 'cc:be:59:cb:44:b2',
                                                                       'name': '325/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '33',
                                                                       'out_octets': '3094687160092',
                                                                       'out_ucast_pkts': '2678878379',
                                                                       'poller_hash': 'NDJCOUVFcGVlcmludGYzMjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '325',
                                                                       'in_nucast_pkts': '33',
                                                                       'in_octets': '3094687160092',
                                                                       'in_ucast_pkts': '2678878379',
                                                                       'links': [       {       'in_nucast_pkts': '33',
                                                                                                'in_octets': '3094687160092',
                                                                                                'in_ucast_pkts': '2678878379',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '551630199953',
                                                                                                'out_ucast_pkts': '1175969188',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '551630199953',
                                                                       'out_ucast_pkts': '1175969188',
                                                                       'poller_hash': 'NDJCOUVFcGVlcmludGZwb24gMzI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '325',
                                                                                               'links': [       {       'connected_time': '8535206',
                                                                                                                        'fiber_length': 23395,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUVFcGVlcnBvbnBvbiAzMjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '36 Zinkann Cr (Andrea Wiebe '
                                                 '10436178)',
                                         'poller_hash': 'NDJCOUVF',
                                         'serial_number': '42B9EE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:b2',
                                         'uptime': '8535206'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '326/g1',
                                                                       'name': '326/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNkI1cGVlcmludGYzMjYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '326',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTEyNkI1cGVlcmludGZwb24gMzI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '326',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNkI1cGVlcnBvbnBvbiAzMjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '7 FIRELLA PL',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "326" '
                                                                  '(location: GPON '
                                                                  'port "6/6"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-04-01T15:09:38.690Z',
                                         'poller_hash': 'NTEyNkI1',
                                         'serial_number': '5126B5',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '327/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'bc:ee:7b:f4:85:f0'],
                                                                       'id': '327/g1',
                                                                       'in_nucast_pkts': '694',
                                                                       'in_octets': '1640300951901',
                                                                       'in_ucast_pkts': '3969859818',
                                                                       'mac_address': 'cc:be:59:f8:64:5d',
                                                                       'mtu': 2000,
                                                                       'name': '327/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6156076',
                                                                       'out_octets': '7292589444029',
                                                                       'out_ucast_pkts': '1141002173',
                                                                       'poller_hash': 'NTc0NTdFcGVlcmludGYzMjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '327',
                                                                       'in_nucast_pkts': '6156076',
                                                                       'in_octets': '7292589444029',
                                                                       'in_ucast_pkts': '1141002173',
                                                                       'links': [       {       'in_nucast_pkts': '6156076',
                                                                                                'in_octets': '7292589444029',
                                                                                                'in_ucast_pkts': '1141002173',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '694',
                                                                                                'out_octets': '1640300951901',
                                                                                                'out_ucast_pkts': '3969859818',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '694',
                                                                       'out_octets': '1640300951901',
                                                                       'out_ucast_pkts': '3969859818',
                                                                       'poller_hash': 'NTc0NTdFcGVlcmludGZwb24gMzI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '327',
                                                                                               'links': [       {       'connected_time': '5339178',
                                                                                                                        'fiber_length': 24251,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.264,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.436}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0NTdFcGVlcnBvbnBvbiAzMjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': "1082 Queen's Bush Rd (Chris Dietz "
                                                 '10434525)',
                                         'poller_hash': 'NTc0NTdF',
                                         'serial_number': '57457E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:64:57',
                                         'uptime': '5339178'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '328/g1',
                                                                       'name': '328/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCNDZGcGVlcmludGYzMjgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '328',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCNDZGcGVlcmludGZwb24gMzI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '328',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNDZGcGVlcnBvbnBvbiAzMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '56 POND VIEW DR (HANN RACHAEL)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "328" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-11-28T18:54:07.580Z',
                                         'poller_hash': 'NDNCNDZG',
                                         'serial_number': '43B46F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '329/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fd:2e'],
                                                                       'id': '329/g1',
                                                                       'in_nucast_pkts': '50',
                                                                       'in_octets': '118373997083',
                                                                       'in_ucast_pkts': '402921090',
                                                                       'mac_address': 'cc:be:59:f8:57:1d',
                                                                       'mtu': 2000,
                                                                       'name': '329/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797544',
                                                                       'out_octets': '1623902777707',
                                                                       'out_ucast_pkts': '1232016833',
                                                                       'poller_hash': 'NTc0M0REcGVlcmludGYzMjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '329',
                                                                       'in_nucast_pkts': '9797544',
                                                                       'in_octets': '1623902777707',
                                                                       'in_ucast_pkts': '1232016833',
                                                                       'links': [       {       'in_nucast_pkts': '9797544',
                                                                                                'in_octets': '1623902777707',
                                                                                                'in_ucast_pkts': '1232016833',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '50',
                                                                                                'out_octets': '118373997083',
                                                                                                'out_ucast_pkts': '402921090',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '50',
                                                                       'out_octets': '118373997083',
                                                                       'out_ucast_pkts': '402921090',
                                                                       'poller_hash': 'NTc0M0REcGVlcmludGZwb24gMzI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '329',
                                                                                               'links': [       {       'connected_time': '8535216',
                                                                                                                        'fiber_length': 24036,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.982,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.54}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0M0REcGVlcnBvbnBvbiAzMjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': "1032 Queen's Bush Rd (Steve Tippin, "
                                                 '10434550)',
                                         'poller_hash': 'NTc0M0RE',
                                         'serial_number': '5743DD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:57:17',
                                         'uptime': '8535216'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '330/g1',
                                                                       'name': '330/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTc0NTJFcGVlcmludGYzMzAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '330',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTc0NTJFcGVlcmludGZwb24gMzMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '330',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0NTJFcGVlcnBvbnBvbiAzMzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '48 Pond View Dr',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "330" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:09.760Z',
                                         'poller_hash': 'NTc0NTJF',
                                         'serial_number': '57452E',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '331/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:a4:b7:a4:fc:75'],
                                                                       'id': '331/g1',
                                                                       'in_nucast_pkts': '151',
                                                                       'in_octets': '3351200541',
                                                                       'in_ucast_pkts': '11816975',
                                                                       'mac_address': 'cc:be:59:50:04:6e',
                                                                       'mtu': 2000,
                                                                       'name': '331/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '59782',
                                                                       'out_octets': '66167538558',
                                                                       'out_ucast_pkts': '47704451',
                                                                       'poller_hash': 'NDNCNUZFcGVlcmludGYzMzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '331',
                                                                       'in_nucast_pkts': '59782',
                                                                       'in_octets': '66167538558',
                                                                       'in_ucast_pkts': '47704451',
                                                                       'links': [       {       'in_nucast_pkts': '59782',
                                                                                                'in_octets': '66167538558',
                                                                                                'in_ucast_pkts': '47704451',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '151',
                                                                                                'out_octets': '3351200541',
                                                                                                'out_ucast_pkts': '11816975',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '151',
                                                                       'out_octets': '3351200541',
                                                                       'out_ucast_pkts': '11816975',
                                                                       'poller_hash': 'NDNCNUZFcGVlcmludGZwb24gMzMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '331',
                                                                                               'links': [       {       'connected_time': '86255',
                                                                                                                        'fiber_length': 23411,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.06,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.358}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUZFcGVlcnBvbnBvbiAzMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '88 Ferris Dr (Jordan Wisniewski, '
                                                 '10434389)',
                                         'poller_hash': 'NDNCNUZF',
                                         'serial_number': '43B5FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:04:6d',
                                         'uptime': '86255'},
                                 {       'counter_source': '332/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:59:a4:87'],
                                                                       'id': '332/g1',
                                                                       'in_nucast_pkts': '35089147',
                                                                       'in_octets': '188957807820',
                                                                       'in_ucast_pkts': '1857375764',
                                                                       'mac_address': 'cc:be:59:cb:44:fd',
                                                                       'mtu': 2000,
                                                                       'name': '332/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8963773107',
                                                                       'out_octets': '21472063496709',
                                                                       'out_ucast_pkts': '6765627918',
                                                                       'poller_hash': 'NDJCOUY3cGVlcmludGYzMzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '332',
                                                                       'in_nucast_pkts': '8963773107',
                                                                       'in_octets': '21472063496709',
                                                                       'in_ucast_pkts': '6765627918',
                                                                       'links': [       {       'in_nucast_pkts': '8963773107',
                                                                                                'in_octets': '21472063496709',
                                                                                                'in_ucast_pkts': '6765627918',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '35089147',
                                                                                                'out_octets': '188957807820',
                                                                                                'out_ucast_pkts': '1857375764',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35089147',
                                                                       'out_octets': '188957807820',
                                                                       'out_ucast_pkts': '1857375764',
                                                                       'poller_hash': 'NDJCOUY3cGVlcmludGZwb24gMzMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '332',
                                                                                               'links': [       {       'connected_time': '21886814',
                                                                                                                        'fiber_length': 24818,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUY3cGVlcnBvbnBvbiAzMzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1145 Queen's Bush Rd (Charles & "
                                                 'Kathleen Stemmler, 10434555)',
                                         'poller_hash': 'NDJCOUY3',
                                         'serial_number': '42B9F7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:fa',
                                         'uptime': '21886814'},
                                 {       'counter_source': '334/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '28:bd:89:d2:ca:2e'],
                                                                       'id': '334/g1',
                                                                       'in_nucast_pkts': '10324',
                                                                       'in_octets': '1078792504271',
                                                                       'in_ucast_pkts': '2311393346',
                                                                       'mac_address': 'cc:be:59:50:03:dc',
                                                                       'mtu': 2000,
                                                                       'name': '334/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652446',
                                                                       'out_octets': '3942425473272',
                                                                       'out_ucast_pkts': '3549730090',
                                                                       'poller_hash': 'NDNCNUI1cGVlcmludGYzMzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '334',
                                                                       'in_nucast_pkts': '5652446',
                                                                       'in_octets': '3942425473272',
                                                                       'in_ucast_pkts': '3549730090',
                                                                       'links': [       {       'in_nucast_pkts': '5652446',
                                                                                                'in_octets': '3942425473272',
                                                                                                'in_ucast_pkts': '3549730090',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10324',
                                                                                                'out_octets': '1078792504271',
                                                                                                'out_ucast_pkts': '2311393346',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10324',
                                                                       'out_octets': '1078792504271',
                                                                       'out_ucast_pkts': '2311393346',
                                                                       'poller_hash': 'NDNCNUI1cGVlcmludGZwb24gMzM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '334',
                                                                                               'links': [       {       'connected_time': '8535067',
                                                                                                                        'fiber_length': 24469,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.418,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.848}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUI1cGVlcnBvbnBvbiAzMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '71 Parkview Rd Wellesley (Mike '
                                                 'Masschaele 10434324)',
                                         'poller_hash': 'NDNCNUI1',
                                         'serial_number': '43B5B5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:03:db',
                                         'uptime': '8535067'},
                                 {       'counter_source': '335/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:34:a8'],
                                                                       'id': '335/g1',
                                                                       'in_nucast_pkts': '145',
                                                                       'in_octets': '77341729906',
                                                                       'in_ucast_pkts': '305669884',
                                                                       'mac_address': 'cc:be:59:b9:c4:a9',
                                                                       'mtu': 2000,
                                                                       'name': '335/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39522617',
                                                                       'out_octets': '1172412397993',
                                                                       'out_ucast_pkts': '989047704',
                                                                       'poller_hash': 'NTEyNTZCcGVlcmludGYzMzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '335',
                                                                       'in_nucast_pkts': '39522617',
                                                                       'in_octets': '1172412397993',
                                                                       'in_ucast_pkts': '989047704',
                                                                       'links': [       {       'in_nucast_pkts': '39522617',
                                                                                                'in_octets': '1172412397993',
                                                                                                'in_ucast_pkts': '989047704',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '145',
                                                                                                'out_octets': '77341729906',
                                                                                                'out_ucast_pkts': '305669884',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '145',
                                                                       'out_octets': '77341729906',
                                                                       'out_ucast_pkts': '305669884',
                                                                       'poller_hash': 'NTEyNTZCcGVlcmludGZwb24gMzM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '335',
                                                                                               'links': [       {       'connected_time': '67856271',
                                                                                                                        'fiber_length': 24650,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTZCcGVlcnBvbnBvbiAzMzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '31 Parkview Dr Wellesley (Wayne '
                                                 'Jantzi 10434738)',
                                         'poller_hash': 'NTEyNTZC',
                                         'serial_number': '51256B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c4:a9',
                                         'uptime': '67856271'},
                                 {       'counter_source': '336/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fc:d4'],
                                                                       'id': '336/g1',
                                                                       'in_nucast_pkts': '3',
                                                                       'in_octets': '14378242505',
                                                                       'in_ucast_pkts': '118234395',
                                                                       'mac_address': 'cc:be:59:50:04:2c',
                                                                       'mtu': 2000,
                                                                       'name': '336/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '402850',
                                                                       'out_octets': '308388953085',
                                                                       'out_ucast_pkts': '217223946',
                                                                       'poller_hash': 'NDNCNUREcGVlcmludGYzMzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '336',
                                                                       'in_nucast_pkts': '402850',
                                                                       'in_octets': '308388953085',
                                                                       'in_ucast_pkts': '217223946',
                                                                       'links': [       {       'in_nucast_pkts': '402850',
                                                                                                'in_octets': '308388953085',
                                                                                                'in_ucast_pkts': '217223946',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3',
                                                                                                'out_octets': '14378242505',
                                                                                                'out_ucast_pkts': '118234395',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3',
                                                                       'out_octets': '14378242505',
                                                                       'out_ucast_pkts': '118234395',
                                                                       'poller_hash': 'NDNCNUREcGVlcmludGZwb24gMzM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '336',
                                                                                               'links': [       {       'connected_time': '600542',
                                                                                                                        'fiber_length': 24269,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.896,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.884}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUREcGVlcnBvbnBvbiAzMzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '53 PARKVIEW DR (HOCHBAN ERIC)',
                                         'poller_hash': 'NDNCNURE',
                                         'serial_number': '43B5DD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:04:2b',
                                         'uptime': '600542'},
                                 {       'counter_source': '337/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '78:54:2e:f0:6c:03'],
                                                                       'id': '337/g1',
                                                                       'in_nucast_pkts': '20813',
                                                                       'in_octets': '36290338839',
                                                                       'in_ucast_pkts': '258014566',
                                                                       'mac_address': 'cc:be:59:50:00:68',
                                                                       'mtu': 2000,
                                                                       'name': '337/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5680964',
                                                                       'out_octets': '1017718653859',
                                                                       'out_ucast_pkts': '716734739',
                                                                       'poller_hash': 'NDNCM0ZEcGVlcmludGYzMzcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '337',
                                                                       'in_nucast_pkts': '5680964',
                                                                       'in_octets': '1017718653859',
                                                                       'in_ucast_pkts': '716734739',
                                                                       'links': [       {       'in_nucast_pkts': '5680964',
                                                                                                'in_octets': '1017718653859',
                                                                                                'in_ucast_pkts': '716734739',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20813',
                                                                                                'out_octets': '36290338839',
                                                                                                'out_ucast_pkts': '258014566',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20813',
                                                                       'out_octets': '36290338839',
                                                                       'out_ucast_pkts': '258014566',
                                                                       'poller_hash': 'NDNCM0ZEcGVlcmludGZwb24gMzM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '337',
                                                                                               'links': [       {       'connected_time': '8535071',
                                                                                                                        'fiber_length': 22820,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.78,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.05}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0ZEcGVlcnBvbnBvbiAzMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1031 Doering St Wellesley (Lawrence '
                                                 'Beach 10434628)',
                                         'poller_hash': 'NDNCM0ZE',
                                         'serial_number': '43B3FD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:00:67',
                                         'uptime': '8535071'},
                                 {       'counter_source': '338/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:9f:c2:6d:70:56'],
                                                                       'id': '338/g1',
                                                                       'in_nucast_pkts': '35',
                                                                       'in_octets': '72183496777',
                                                                       'in_ucast_pkts': '334503841',
                                                                       'mac_address': 'cc:be:59:b9:c6:9b',
                                                                       'mtu': 2000,
                                                                       'name': '338/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21816983',
                                                                       'out_octets': '2053762045514',
                                                                       'out_ucast_pkts': '1597597550',
                                                                       'poller_hash': 'NTEyNjExcGVlcmludGYzMzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '338',
                                                                       'in_nucast_pkts': '21816983',
                                                                       'in_octets': '2053762045514',
                                                                       'in_ucast_pkts': '1597597550',
                                                                       'links': [       {       'in_nucast_pkts': '21816983',
                                                                                                'in_octets': '2053762045514',
                                                                                                'in_ucast_pkts': '1597597550',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '35',
                                                                                                'out_octets': '72183496777',
                                                                                                'out_ucast_pkts': '334503841',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35',
                                                                       'out_octets': '72183496777',
                                                                       'out_ucast_pkts': '334503841',
                                                                       'poller_hash': 'NTEyNjExcGVlcmludGZwb24gMzM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '338',
                                                                                               'links': [       {       'connected_time': '18480391',
                                                                                                                        'fiber_length': 24981,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNjExcGVlcnBvbnBvbiAzMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '8 Parkview Dr Wellesley (D & D '
                                                 'Underwood 10434295)',
                                         'poller_hash': 'NTEyNjEx',
                                         'serial_number': '512611',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c6:9b',
                                         'uptime': '18480391'},
                                 {       'counter_source': '339/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '339/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '436663092689',
                                                                       'in_ucast_pkts': '999982556',
                                                                       'mac_address': 'cc:be:59:cb:44:8a',
                                                                       'name': '339/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '5299660405439',
                                                                       'out_ucast_pkts': '3894911624',
                                                                       'poller_hash': 'NDJCOUU5cGVlcmludGYzMzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '339',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '5299660405439',
                                                                       'in_ucast_pkts': '3894911624',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '5299660405439',
                                                                                                'in_ucast_pkts': '3894911624',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '436663092689',
                                                                                                'out_ucast_pkts': '999982556',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '436663092689',
                                                                       'out_ucast_pkts': '999982556',
                                                                       'poller_hash': 'NDJCOUU5cGVlcmludGZwb24gMzM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '339',
                                                                                               'links': [       {       'connected_time': '2594924',
                                                                                                                        'fiber_length': 24161,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUU5cGVlcnBvbnBvbiAzMzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1025 Queen's Bush Rd Wellesley "
                                                 '(Jacob Reed 10436059)',
                                         'poller_hash': 'NDJCOUU5',
                                         'serial_number': '42B9E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:8a',
                                         'uptime': '2594924'},
                                 {       'counter_source': '341/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:1b'],
                                                                       'id': '341/g1',
                                                                       'in_nucast_pkts': '62870507',
                                                                       'in_octets': '820229080988',
                                                                       'in_ucast_pkts': '1898873729',
                                                                       'mac_address': 'cc:be:59:cb:43:25',
                                                                       'mtu': 2000,
                                                                       'name': '341/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12338486013',
                                                                       'out_octets': '22517709420466',
                                                                       'out_ucast_pkts': '5920481254',
                                                                       'poller_hash': 'NDJCOUJDcGVlcmludGYzNDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '341',
                                                                       'in_nucast_pkts': '12338486013',
                                                                       'in_octets': '22517709420466',
                                                                       'in_ucast_pkts': '5920481254',
                                                                       'links': [       {       'in_nucast_pkts': '12338486013',
                                                                                                'in_octets': '22517709420466',
                                                                                                'in_ucast_pkts': '5920481254',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '62870507',
                                                                                                'out_octets': '820229080988',
                                                                                                'out_ucast_pkts': '1898873729',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62870507',
                                                                       'out_octets': '820229080988',
                                                                       'out_ucast_pkts': '1898873729',
                                                                       'poller_hash': 'NDJCOUJDcGVlcmludGZwb24gMzQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '341',
                                                                                               'links': [       {       'connected_time': '21886677',
                                                                                                                        'fiber_length': 23514,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUJDcGVlcnBvbnBvbiAzNDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '92 FERRIS DR (BOGOSLOWSKI PAUL)',
                                         'poller_hash': 'NDJCOUJD',
                                         'serial_number': '42B9BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:22',
                                         'uptime': '21886677'},
                                 {       'counter_source': '342/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:09:6a'],
                                                                       'id': '342/g1',
                                                                       'in_nucast_pkts': '56',
                                                                       'in_octets': '32293661762',
                                                                       'in_ucast_pkts': '127742949',
                                                                       'mac_address': '44:65:7f:7b:c3:aa',
                                                                       'mtu': 2000,
                                                                       'name': '342/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39526334',
                                                                       'out_octets': '440360682677',
                                                                       'out_ucast_pkts': '349443717',
                                                                       'poller_hash': 'NUE2NjI5cGVlcmludGYzNDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '342',
                                                                       'in_nucast_pkts': '39526334',
                                                                       'in_octets': '440360682677',
                                                                       'in_ucast_pkts': '349443717',
                                                                       'links': [       {       'in_nucast_pkts': '39526334',
                                                                                                'in_octets': '440360682677',
                                                                                                'in_ucast_pkts': '349443717',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '56',
                                                                                                'out_octets': '32293661762',
                                                                                                'out_ucast_pkts': '127742949',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '56',
                                                                       'out_octets': '32293661762',
                                                                       'out_ucast_pkts': '127742949',
                                                                       'poller_hash': 'NUE2NjI5cGVlcmludGZwb24gMzQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '342',
                                                                                               'links': [       {       'connected_time': '67856405',
                                                                                                                        'fiber_length': 24801,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjI5cGVlcnBvbnBvbiAzNDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '8 Edgewood Cr Wellesley (Roger Zehr '
                                                 '10434453)',
                                         'poller_hash': 'NUE2NjI5',
                                         'serial_number': '5A6629',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c3:aa',
                                         'uptime': '67856405'},
                                 {       'counter_source': '343/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fc:8c'],
                                                                       'id': '343/g1',
                                                                       'in_nucast_pkts': '27',
                                                                       'in_octets': '641245502744',
                                                                       'in_ucast_pkts': '1409913915',
                                                                       'mac_address': 'cc:be:59:cb:45:05',
                                                                       'mtu': 2000,
                                                                       'name': '343/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6379675',
                                                                       'out_octets': '4233632149208',
                                                                       'out_ucast_pkts': '3087353679',
                                                                       'poller_hash': 'NDJCOUY4cGVlcmludGYzNDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '343',
                                                                       'in_nucast_pkts': '6379675',
                                                                       'in_octets': '4233632149208',
                                                                       'in_ucast_pkts': '3087353679',
                                                                       'links': [       {       'in_nucast_pkts': '6379675',
                                                                                                'in_octets': '4233632149208',
                                                                                                'in_ucast_pkts': '3087353679',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27',
                                                                                                'out_octets': '641245502744',
                                                                                                'out_ucast_pkts': '1409913915',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27',
                                                                       'out_octets': '641245502744',
                                                                       'out_ucast_pkts': '1409913915',
                                                                       'poller_hash': 'NDJCOUY4cGVlcmludGZwb24gMzQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '343',
                                                                                               'links': [       {       'connected_time': '5444113',
                                                                                                                        'fiber_length': 24534,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUY4cGVlcnBvbnBvbiAzNDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '39 PARKVIEW DR (SIEWERT JASON)',
                                         'poller_hash': 'NDJCOUY4',
                                         'serial_number': '42B9F8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:45:02',
                                         'uptime': '5444113'},
                                 {       'counter_source': '344/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:9f:c2:15:09:9f'],
                                                                       'id': '344/g1',
                                                                       'in_nucast_pkts': '53290',
                                                                       'in_octets': '1033937114482',
                                                                       'in_ucast_pkts': '3444698930',
                                                                       'mac_address': 'cc:be:59:cb:42:b5',
                                                                       'mtu': 2000,
                                                                       'name': '344/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17551937',
                                                                       'out_octets': '7580397206458',
                                                                       'out_ucast_pkts': '6389954299',
                                                                       'poller_hash': 'NDJCOUFFcGVlcmludGYzNDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '344',
                                                                       'in_nucast_pkts': '17551937',
                                                                       'in_octets': '7580397206458',
                                                                       'in_ucast_pkts': '6389954299',
                                                                       'links': [       {       'in_nucast_pkts': '17551937',
                                                                                                'in_octets': '7580397206458',
                                                                                                'in_ucast_pkts': '6389954299',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '53290',
                                                                                                'out_octets': '1033937114482',
                                                                                                'out_ucast_pkts': '3444698930',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53290',
                                                                       'out_octets': '1033937114482',
                                                                       'out_ucast_pkts': '3444698930',
                                                                       'poller_hash': 'NDJCOUFFcGVlcmludGZwb24gMzQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '344',
                                                                                               'links': [       {       'connected_time': '27206082',
                                                                                                                        'fiber_length': 25368,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUFFcGVlcnBvbnBvbiAzNDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '30 JACK KOEHLER LN (DEWHURST ADRIAN)',
                                         'poller_hash': 'NDJCOUFF',
                                         'serial_number': '42B9AE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:42:b2',
                                         'uptime': '27206082'},
                                 {       'counter_source': '346/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '98:25:4a:ea:4d:dd'],
                                                                       'id': '346/g1',
                                                                       'in_nucast_pkts': '47945',
                                                                       'in_octets': '4080793336962',
                                                                       'in_ucast_pkts': '9742499301',
                                                                       'mac_address': '60:db:98:4e:95:57',
                                                                       'mtu': 2000,
                                                                       'name': '346/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61587601',
                                                                       'out_octets': '34700162501208',
                                                                       'out_ucast_pkts': '26595722428',
                                                                       'poller_hash': 'QTlDODZBcGVlcmludGYzNDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '346',
                                                                       'in_nucast_pkts': '61587601',
                                                                       'in_octets': '34700162501208',
                                                                       'in_ucast_pkts': '26595722428',
                                                                       'links': [       {       'in_nucast_pkts': '61587601',
                                                                                                'in_octets': '34700162501208',
                                                                                                'in_ucast_pkts': '26595722428',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '47945',
                                                                                                'out_octets': '4080793336962',
                                                                                                'out_ucast_pkts': '9742499301',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '47945',
                                                                       'out_octets': '4080793336962',
                                                                       'out_ucast_pkts': '9742499301',
                                                                       'poller_hash': 'QTlDODZBcGVlcmludGZwb24gMzQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '346',
                                                                                               'links': [       {       'connected_time': '21677433',
                                                                                                                        'fiber_length': 24889,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDODZBcGVlcnBvbnBvbiAzNDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '31 Nithbank Rd Wellesley (Chris '
                                                 'Tillery 10434314)',
                                         'poller_hash': 'QTlDODZB',
                                         'serial_number': 'A9C86A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:95:54',
                                         'uptime': '21677433'},
                                 {       'counter_source': '347/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:34:f0'],
                                                                       'id': '347/g1',
                                                                       'in_nucast_pkts': '11890',
                                                                       'in_octets': '410389281366',
                                                                       'in_ucast_pkts': '1032947145',
                                                                       'mac_address': 'cc:be:59:cb:44:a5',
                                                                       'mtu': 2000,
                                                                       'name': '347/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61709666',
                                                                       'out_octets': '4549038430748',
                                                                       'out_ucast_pkts': '3719187930',
                                                                       'poller_hash': 'NDJCOUVDcGVlcmludGYzNDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '347',
                                                                       'in_nucast_pkts': '61709666',
                                                                       'in_octets': '4549038430748',
                                                                       'in_ucast_pkts': '3719187930',
                                                                       'links': [       {       'in_nucast_pkts': '61709666',
                                                                                                'in_octets': '4549038430748',
                                                                                                'in_ucast_pkts': '3719187930',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11890',
                                                                                                'out_octets': '410389281366',
                                                                                                'out_ucast_pkts': '1032947145',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11890',
                                                                       'out_octets': '410389281366',
                                                                       'out_ucast_pkts': '1032947145',
                                                                       'poller_hash': 'NDJCOUVDcGVlcmludGZwb24gMzQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '347',
                                                                                               'links': [       {       'connected_time': '21887048',
                                                                                                                        'fiber_length': 24829,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUVDcGVlcnBvbnBvbiAzNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 Parkview Dr Wellesley (David Roth '
                                                 '10434761)',
                                         'poller_hash': 'NDJCOUVD',
                                         'serial_number': '42B9EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:a2',
                                         'uptime': '21887048'},
                                 {       'counter_source': '349/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:34:d5'],
                                                                       'id': '349/g1',
                                                                       'in_nucast_pkts': '68',
                                                                       'in_octets': '87603156686',
                                                                       'in_ucast_pkts': '190399856',
                                                                       'mac_address': 'cc:be:59:b9:bf:c9',
                                                                       'mtu': 2000,
                                                                       'name': '349/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39526646',
                                                                       'out_octets': '567994523383',
                                                                       'out_ucast_pkts': '471117693',
                                                                       'poller_hash': 'NTEyM0NCcGVlcmludGYzNDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '349',
                                                                       'in_nucast_pkts': '39526646',
                                                                       'in_octets': '567994523383',
                                                                       'in_ucast_pkts': '471117693',
                                                                       'links': [       {       'in_nucast_pkts': '39526646',
                                                                                                'in_octets': '567994523383',
                                                                                                'in_ucast_pkts': '471117693',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '68',
                                                                                                'out_octets': '87603156686',
                                                                                                'out_ucast_pkts': '190399856',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68',
                                                                       'out_octets': '87603156686',
                                                                       'out_ucast_pkts': '190399856',
                                                                       'poller_hash': 'NTEyM0NCcGVlcmludGZwb24gMzQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '349',
                                                                                               'links': [       {       'connected_time': '67856140',
                                                                                                                        'fiber_length': 24982,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyM0NCcGVlcnBvbnBvbiAzNDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '14 Nithbank Rd Wellesley (Marian '
                                                 'Weber 10434753)',
                                         'poller_hash': 'NTEyM0NC',
                                         'serial_number': '5123CB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:bf:c9',
                                         'uptime': '67856140'},
                                 {       'counter_source': '351/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:35:5c'],
                                                                       'id': '351/g1',
                                                                       'in_nucast_pkts': '44',
                                                                       'in_octets': '363163613015',
                                                                       'in_ucast_pkts': '4371166814',
                                                                       'mac_address': 'cc:be:59:cb:44:25',
                                                                       'mtu': 2000,
                                                                       'name': '351/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38275452',
                                                                       'out_octets': '23865169300902',
                                                                       'out_ucast_pkts': '15878274045',
                                                                       'poller_hash': 'NDJCOURDcGVlcmludGYzNTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '351',
                                                                       'in_nucast_pkts': '38275452',
                                                                       'in_octets': '23865169300902',
                                                                       'in_ucast_pkts': '15878274045',
                                                                       'links': [       {       'in_nucast_pkts': '38275452',
                                                                                                'in_octets': '23865169300902',
                                                                                                'in_ucast_pkts': '15878274045',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '44',
                                                                                                'out_octets': '363163613015',
                                                                                                'out_ucast_pkts': '4371166814',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '44',
                                                                       'out_octets': '363163613015',
                                                                       'out_ucast_pkts': '4371166814',
                                                                       'poller_hash': 'NDJCOURDcGVlcmludGZwb24gMzUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '351',
                                                                                               'links': [       {       'connected_time': '67648439',
                                                                                                                        'fiber_length': 24687,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOURDcGVlcnBvbnBvbiAzNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1127 Queen's Bush Rd (Keith Erb "
                                                 '10434601)',
                                         'poller_hash': 'NDJCOURD',
                                         'serial_number': '42B9DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:22',
                                         'uptime': '67648439'},
                                 {       'counter_source': '352/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:34:cc'],
                                                                       'id': '352/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '56753754818',
                                                                       'in_ucast_pkts': '247570171',
                                                                       'mac_address': 'cc:be:59:50:03:e8',
                                                                       'mtu': 2000,
                                                                       'name': '352/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4209037',
                                                                       'out_octets': '825584830359',
                                                                       'out_ucast_pkts': '594420935',
                                                                       'poller_hash': 'NDNCNUJCcGVlcmludGYzNTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '352',
                                                                       'in_nucast_pkts': '4209037',
                                                                       'in_octets': '825584830359',
                                                                       'in_ucast_pkts': '594420935',
                                                                       'links': [       {       'in_nucast_pkts': '4209037',
                                                                                                'in_octets': '825584830359',
                                                                                                'in_ucast_pkts': '594420935',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '56753754818',
                                                                                                'out_ucast_pkts': '247570171',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '56753754818',
                                                                       'out_ucast_pkts': '247570171',
                                                                       'poller_hash': 'NDNCNUJCcGVlcmludGZwb24gMzUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '352',
                                                                                               'links': [       {       'connected_time': '6314499',
                                                                                                                        'fiber_length': 24149,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -23.532,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.142}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUJCcGVlcnBvbnBvbiAzNTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '61 Parkview Dr (Scott Kelterborn '
                                                 '10434621)',
                                         'poller_hash': 'NDNCNUJC',
                                         'serial_number': '43B5BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:03:e7',
                                         'uptime': '6314499'},
                                 {       'counter_source': '353/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:06:40'],
                                                                       'id': '353/g1',
                                                                       'in_nucast_pkts': '8',
                                                                       'in_octets': '106545910595',
                                                                       'in_ucast_pkts': '536315113',
                                                                       'mac_address': 'cc:be:59:50:02:a4',
                                                                       'mtu': 2000,
                                                                       'name': '353/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652436',
                                                                       'out_octets': '3579724867828',
                                                                       'out_ucast_pkts': '2544064392',
                                                                       'poller_hash': 'NDNCNTE5cGVlcmludGYzNTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '353',
                                                                       'in_nucast_pkts': '5652436',
                                                                       'in_octets': '3579724867828',
                                                                       'in_ucast_pkts': '2544064392',
                                                                       'links': [       {       'in_nucast_pkts': '5652436',
                                                                                                'in_octets': '3579724867828',
                                                                                                'in_ucast_pkts': '2544064392',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8',
                                                                                                'out_octets': '106545910595',
                                                                                                'out_ucast_pkts': '536315113',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8',
                                                                       'out_octets': '106545910595',
                                                                       'out_ucast_pkts': '536315113',
                                                                       'poller_hash': 'NDNCNTE5cGVlcmludGZwb24gMzUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '353',
                                                                                               'links': [       {       'connected_time': '8535063',
                                                                                                                        'fiber_length': 25343,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.062}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNTE5cGVlcnBvbnBvbiAzNTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5 Jack Koehler LN (Dawn Gellert, '
                                                 '10434571)',
                                         'poller_hash': 'NDNCNTE5',
                                         'serial_number': '43B519',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:02:a3',
                                         'uptime': '8535063'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '354/g1',
                                                                       'name': '354/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCOUFEcGVlcmludGYzNTQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '354',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCOUFEcGVlcmludGZwb24gMzU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '354',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUFEcGVlcnBvbnBvbiAzNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '109 VILLAGE RD',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "354" '
                                                                  '(location: GPON '
                                                                  'port "6/5"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:37:14.330Z',
                                         'poller_hash': 'NDJCOUFE',
                                         'serial_number': '42B9AD',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '355/F1+355/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:37:cc'],
                                                                       'id': '355/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8328074540',
                                                                       'in_ucast_pkts': '47266128',
                                                                       'name': '355/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '36998907409',
                                                                       'out_octets': '49921227524814',
                                                                       'out_ucast_pkts': '19677702',
                                                                       'poller_hash': 'NDJCOUNBcGVlcmludGYzNTUvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fc:dd'],
                                                                       'id': '355/g1',
                                                                       'in_nucast_pkts': '6764',
                                                                       'in_octets': '543228725875',
                                                                       'in_ucast_pkts': '2410242422',
                                                                       'mac_address': 'cc:be:59:cb:43:95',
                                                                       'mtu': 2000,
                                                                       'name': '355/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61704684',
                                                                       'out_octets': '10677392325757',
                                                                       'out_ucast_pkts': '7785117268',
                                                                       'poller_hash': 'NDJCOUNBcGVlcmludGYzNTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '355',
                                                                       'in_nucast_pkts': '37060612093',
                                                                       'in_octets': '60598619850571',
                                                                       'in_ucast_pkts': '7804794970',
                                                                       'links': [       {       'in_nucast_pkts': '37060612093',
                                                                                                'in_octets': '60598619850571',
                                                                                                'in_ucast_pkts': '7804794970',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6764',
                                                                                                'out_octets': '551556800415',
                                                                                                'out_ucast_pkts': '2457508550',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6764',
                                                                       'out_octets': '551556800415',
                                                                       'out_ucast_pkts': '2457508550',
                                                                       'poller_hash': 'NDJCOUNBcGVlcmludGZwb24gMzU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '355',
                                                                                               'links': [       {       'connected_time': '21886454',
                                                                                                                        'fiber_length': 24609,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUNBcGVlcnBvbnBvbiAzNTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '18 FIRELLA PL (ANYAN GARY & JESSIE)',
                                         'poller_hash': 'NDJCOUNB',
                                         'serial_number': '42B9CA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:92',
                                         'uptime': '21886454'},
                                 {       'counter_source': '356/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c9:bf'],
                                                                       'id': '356/g1',
                                                                       'in_nucast_pkts': '68',
                                                                       'in_octets': '588101596727',
                                                                       'in_ucast_pkts': '817020341',
                                                                       'mac_address': 'cc:be:59:b9:c5:b1',
                                                                       'mtu': 2000,
                                                                       'name': '356/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '45072938',
                                                                       'out_octets': '2006741708667',
                                                                       'out_ucast_pkts': '1850204835',
                                                                       'poller_hash': 'NTEyNUMzcGVlcmludGYzNTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '356',
                                                                       'in_nucast_pkts': '45072938',
                                                                       'in_octets': '2006741708667',
                                                                       'in_ucast_pkts': '1850204835',
                                                                       'links': [       {       'in_nucast_pkts': '45072938',
                                                                                                'in_octets': '2006741708667',
                                                                                                'in_ucast_pkts': '1850204835',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '68',
                                                                                                'out_octets': '588101596727',
                                                                                                'out_ucast_pkts': '817020341',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68',
                                                                       'out_octets': '588101596727',
                                                                       'out_ucast_pkts': '817020341',
                                                                       'poller_hash': 'NTEyNUMzcGVlcmludGZwb24gMzU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '356',
                                                                                               'links': [       {       'connected_time': '67856116',
                                                                                                                        'fiber_length': 24294,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUMzcGVlcnBvbnBvbiAzNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': "1072 Queen's Bush Rd Wellesley (Neil "
                                                 '& Linda Lackey10434339)',
                                         'poller_hash': 'NTEyNUMz',
                                         'serial_number': '5125C3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c5:b1',
                                         'uptime': '67856116'},
                                 {       'counter_source': '358/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3a:7b'],
                                                                       'id': '358/g1',
                                                                       'in_nucast_pkts': '13595633',
                                                                       'in_octets': '90924287481',
                                                                       'in_ucast_pkts': '587347236',
                                                                       'mac_address': 'cc:be:59:cb:43:cd',
                                                                       'mtu': 2000,
                                                                       'name': '358/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '33710156433',
                                                                       'out_octets': '48914961883857',
                                                                       'out_ucast_pkts': '2923619037',
                                                                       'poller_hash': 'NDJCOUQxcGVlcmludGYzNTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '358',
                                                                       'in_nucast_pkts': '33710156433',
                                                                       'in_octets': '48914961883857',
                                                                       'in_ucast_pkts': '2923619037',
                                                                       'links': [       {       'in_nucast_pkts': '33710156433',
                                                                                                'in_octets': '48914961883857',
                                                                                                'in_ucast_pkts': '2923619037',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13595633',
                                                                                                'out_octets': '90924287481',
                                                                                                'out_ucast_pkts': '587347236',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13595633',
                                                                       'out_octets': '90924287481',
                                                                       'out_ucast_pkts': '587347236',
                                                                       'poller_hash': 'NDJCOUQxcGVlcmludGZwb24gMzU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '358',
                                                                                               'links': [       {       'connected_time': '21886715',
                                                                                                                        'fiber_length': 24531,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUQxcGVlcnBvbnBvbiAzNTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 Firella Pl Wellesley (Philip '
                                                 'Lepage 10434718)',
                                         'poller_hash': 'NDJCOUQx',
                                         'serial_number': '42B9D1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:ca',
                                         'uptime': '21886715'},
                                 {       'counter_source': '362/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c9:a4'],
                                                                       'id': '362/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '127920355253',
                                                                       'in_ucast_pkts': '650137115',
                                                                       'mac_address': 'cc:be:59:cb:44:35',
                                                                       'mtu': 2000,
                                                                       'name': '362/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919934',
                                                                       'out_octets': '3056361881564',
                                                                       'out_ucast_pkts': '2178845460',
                                                                       'poller_hash': 'NDJCOURFcGVlcmludGYzNjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '362',
                                                                       'in_nucast_pkts': '9919934',
                                                                       'in_octets': '3056361881564',
                                                                       'in_ucast_pkts': '2178845460',
                                                                       'links': [       {       'in_nucast_pkts': '9919934',
                                                                                                'in_octets': '3056361881564',
                                                                                                'in_ucast_pkts': '2178845460',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '127920355253',
                                                                                                'out_ucast_pkts': '650137115',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '127920355253',
                                                                       'out_ucast_pkts': '650137115',
                                                                       'poller_hash': 'NDJCOURFcGVlcmludGZwb24gMzYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '362',
                                                                                               'links': [       {       'connected_time': '8535224',
                                                                                                                        'fiber_length': 24586,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOURFcGVlcnBvbnBvbiAzNjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '167 Gerber Meadows Dr Wellesley '
                                                 '(Andrea Gebbie 10434736)',
                                         'poller_hash': 'NDJCOURF',
                                         'serial_number': '42B9DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:32',
                                         'uptime': '8535224'},
                                 {       'counter_source': '363/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8c:40'],
                                                                       'id': '363/g1',
                                                                       'in_nucast_pkts': '155801',
                                                                       'in_octets': '131124952411',
                                                                       'in_ucast_pkts': '749153453',
                                                                       'mac_address': 'cc:be:59:cb:44:dd',
                                                                       'mtu': 2000,
                                                                       'name': '363/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919591',
                                                                       'out_octets': '3704855820064',
                                                                       'out_ucast_pkts': '2632022680',
                                                                       'poller_hash': 'NDJCOUYzcGVlcmludGYzNjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '363',
                                                                       'in_nucast_pkts': '9919591',
                                                                       'in_octets': '3704855820064',
                                                                       'in_ucast_pkts': '2632022680',
                                                                       'links': [       {       'in_nucast_pkts': '9919591',
                                                                                                'in_octets': '3704855820064',
                                                                                                'in_ucast_pkts': '2632022680',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '155801',
                                                                                                'out_octets': '131124952411',
                                                                                                'out_ucast_pkts': '749153453',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '155801',
                                                                       'out_octets': '131124952411',
                                                                       'out_ucast_pkts': '749153453',
                                                                       'poller_hash': 'NDJCOUYzcGVlcmludGZwb24gMzYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '363',
                                                                                               'links': [       {       'connected_time': '8535184',
                                                                                                                        'fiber_length': 24443,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUYzcGVlcnBvbnBvbiAzNjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '41 Parkview Drive Wellesley '
                                                 '(Jennifer & Jeff Sommer 10434517)',
                                         'poller_hash': 'NDJCOUYz',
                                         'serial_number': '42B9F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:da',
                                         'uptime': '8535184'},
                                 {       'counter_source': '364/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:35:ad'],
                                                                       'id': '364/g1',
                                                                       'in_nucast_pkts': '15',
                                                                       'in_octets': '19454443839',
                                                                       'in_ucast_pkts': '116342791',
                                                                       'mac_address': 'cc:be:59:cb:43:65',
                                                                       'mtu': 2000,
                                                                       'name': '364/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4990205',
                                                                       'out_octets': '576733692438',
                                                                       'out_ucast_pkts': '393132281',
                                                                       'poller_hash': 'NDJCOUM0cGVlcmludGYzNjQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '364',
                                                                       'in_nucast_pkts': '4990205',
                                                                       'in_octets': '576733692438',
                                                                       'in_ucast_pkts': '393132281',
                                                                       'links': [       {       'in_nucast_pkts': '4990205',
                                                                                                'in_octets': '576733692438',
                                                                                                'in_ucast_pkts': '393132281',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15',
                                                                                                'out_octets': '19454443839',
                                                                                                'out_ucast_pkts': '116342791',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15',
                                                                       'out_octets': '19454443839',
                                                                       'out_ucast_pkts': '116342791',
                                                                       'poller_hash': 'NDJCOUM0cGVlcmludGZwb24gMzY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '364',
                                                                                               'links': [       {       'connected_time': '4250988',
                                                                                                                        'fiber_length': 25293,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUM0cGVlcnBvbnBvbiAzNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '75 Village Rd Wellesley (K & M '
                                                 'Gerber 10400460)',
                                         'poller_hash': 'NDJCOUM0',
                                         'serial_number': '42B9C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:62',
                                         'uptime': '4250988'},
                                 {       'counter_source': '366/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '366/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '90391',
                                                                       'in_octets': '202426836702',
                                                                       'in_ucast_pkts': '847487228',
                                                                       'mac_address': '40:a2:db:16:08:81',
                                                                       'name': '366/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '6693966',
                                                                       'out_octets': '3163003219404',
                                                                       'out_ucast_pkts': '2566890195',
                                                                       'poller_hash': 'NTc0NTcxcGVlcmludGYzNjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '366',
                                                                       'in_nucast_pkts': '6693966',
                                                                       'in_octets': '3163003219404',
                                                                       'in_ucast_pkts': '2566890195',
                                                                       'links': [       {       'in_nucast_pkts': '6693966',
                                                                                                'in_octets': '3163003219404',
                                                                                                'in_ucast_pkts': '2566890195',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '90391',
                                                                                                'out_octets': '202426836702',
                                                                                                'out_ucast_pkts': '847487228',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '90391',
                                                                       'out_octets': '202426836702',
                                                                       'out_ucast_pkts': '847487228',
                                                                       'poller_hash': 'NTc0NTcxcGVlcmludGZwb24gMzY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '366',
                                                                                               'links': [       {       'connected_time': '8535249',
                                                                                                                        'fiber_length': 24686,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.002,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.148}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0NTcxcGVlcnBvbnBvbiAzNjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': '7 Edgewood Ct Wellesley (D &S '
                                                 'Hergott 10429943)',
                                         'poller_hash': 'NTc0NTcx',
                                         'serial_number': '574571',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:63:ef',
                                         'uptime': '8535249'},
                                 {       'counter_source': '367/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '367/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '224297189099',
                                                                       'in_ucast_pkts': '9415092543',
                                                                       'mac_address': 'cc:be:59:cb:44:6a',
                                                                       'name': '367/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '33',
                                                                       'out_octets': '3475246615241',
                                                                       'out_ucast_pkts': '2621213464',
                                                                       'poller_hash': 'NDJCOUU1cGVlcmludGYzNjcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '367',
                                                                       'in_nucast_pkts': '33',
                                                                       'in_octets': '3475246615241',
                                                                       'in_ucast_pkts': '2621213464',
                                                                       'links': [       {       'in_nucast_pkts': '33',
                                                                                                'in_octets': '3475246615241',
                                                                                                'in_ucast_pkts': '2621213464',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '224297189099',
                                                                                                'out_ucast_pkts': '9415092543',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '224297189099',
                                                                       'out_ucast_pkts': '9415092543',
                                                                       'poller_hash': 'NDJCOUU1cGVlcmludGZwb24gMzY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '367',
                                                                                               'links': [       {       'connected_time': '1741582',
                                                                                                                        'fiber_length': 23273,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUU1cGVlcnBvbnBvbiAzNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '126 SCHWEITZER CR (LAROCCA CURTIS)',
                                         'poller_hash': 'NDJCOUU1',
                                         'serial_number': '42B9E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:6a',
                                         'uptime': '1741582'},
                                 {       'counter_source': '368/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c4:f7'],
                                                                       'id': '368/g1',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '279864872167',
                                                                       'in_ucast_pkts': '527670432',
                                                                       'mac_address': 'cc:be:59:cb:9a:7d',
                                                                       'mtu': 2000,
                                                                       'name': '368/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919775',
                                                                       'out_octets': '1656795594539',
                                                                       'out_ucast_pkts': '1319955151',
                                                                       'poller_hash': 'NDJDNEE3cGVlcmludGYzNjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '368',
                                                                       'in_nucast_pkts': '9919775',
                                                                       'in_octets': '1656795594539',
                                                                       'in_ucast_pkts': '1319955151',
                                                                       'links': [       {       'in_nucast_pkts': '9919775',
                                                                                                'in_octets': '1656795594539',
                                                                                                'in_ucast_pkts': '1319955151',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23',
                                                                                                'out_octets': '279864872167',
                                                                                                'out_ucast_pkts': '527670432',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '279864872167',
                                                                       'out_ucast_pkts': '527670432',
                                                                       'poller_hash': 'NDJDNEE3cGVlcmludGZwb24gMzY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '368',
                                                                                               'links': [       {       'connected_time': '8535238',
                                                                                                                        'fiber_length': 24777,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNEE3cGVlcnBvbnBvbiAzNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '70 Parkview Dr Wellesley (Kevin Zehr '
                                                 '10434638)',
                                         'poller_hash': 'NDJDNEE3',
                                         'serial_number': '42C4A7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:9a:7a',
                                         'uptime': '8535238'},
                                 {       'counter_source': '369/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:c9:b6'],
                                                                       'id': '369/g1',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '122551938421',
                                                                       'in_ucast_pkts': '639647008',
                                                                       'mac_address': 'cc:be:59:f8:5f:45',
                                                                       'mtu': 2000,
                                                                       'name': '369/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919663',
                                                                       'out_octets': '2530563934160',
                                                                       'out_ucast_pkts': '1940280295',
                                                                       'poller_hash': 'NTc0NERDcGVlcmludGYzNjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '369',
                                                                       'in_nucast_pkts': '9919663',
                                                                       'in_octets': '2530563934160',
                                                                       'in_ucast_pkts': '1940280295',
                                                                       'links': [       {       'in_nucast_pkts': '9919663',
                                                                                                'in_octets': '2530563934160',
                                                                                                'in_ucast_pkts': '1940280295',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '66',
                                                                                                'out_octets': '122551938421',
                                                                                                'out_ucast_pkts': '639647008',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '122551938421',
                                                                       'out_ucast_pkts': '639647008',
                                                                       'poller_hash': 'NTc0NERDcGVlcmludGZwb24gMzY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '369',
                                                                                               'links': [       {       'connected_time': '8535139',
                                                                                                                        'fiber_length': 23712,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.21,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.118}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0NERDcGVlcnBvbnBvbiAzNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': '54 SCHWEITZER CR (WILLIAMS JUSTIN & '
                                                 'ASHLEY)',
                                         'poller_hash': 'NTc0NERD',
                                         'serial_number': '5744DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:5f:3f',
                                         'uptime': '8535139'},
                                 {       'counter_source': '370/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '370/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '182561562837',
                                                                       'in_ucast_pkts': '22636051321',
                                                                       'mac_address': 'd0:76:8f:08:15:23',
                                                                       'name': '370/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '51',
                                                                       'out_octets': '3335643128505',
                                                                       'out_ucast_pkts': '2398806813',
                                                                       'poller_hash': 'ODY5MzlBcGVlcmludGYzNzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '370',
                                                                       'in_nucast_pkts': '51',
                                                                       'in_octets': '3335643128505',
                                                                       'in_ucast_pkts': '2398806813',
                                                                       'links': [       {       'in_nucast_pkts': '51',
                                                                                                'in_octets': '3335643128505',
                                                                                                'in_ucast_pkts': '2398806813',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '182561562837',
                                                                                                'out_ucast_pkts': '22636051321',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '182561562837',
                                                                       'out_ucast_pkts': '22636051321',
                                                                       'poller_hash': 'ODY5MzlBcGVlcmludGZwb24gMzcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '370',
                                                                                               'links': [       {       'connected_time': '8535305',
                                                                                                                        'fiber_length': 25120,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzlBcGVlcnBvbnBvbiAzNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '60 Village Rd Wellesley (James Doyle '
                                                 '10434763)',
                                         'poller_hash': 'ODY5MzlB',
                                         'serial_number': '86939A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:23',
                                         'uptime': '8535305'},
                                 {       'counter_source': '371/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d2:54'],
                                                                       'id': '371/g1',
                                                                       'in_nucast_pkts': '160787',
                                                                       'in_octets': '56718658465',
                                                                       'in_ucast_pkts': '512241206',
                                                                       'mac_address': 'cc:be:59:f8:44:fd',
                                                                       'mtu': 2000,
                                                                       'name': '371/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919499',
                                                                       'out_octets': '1437097013841',
                                                                       'out_ucast_pkts': '1009556370',
                                                                       'poller_hash': 'NTc0MUEwcGVlcmludGYzNzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '371',
                                                                       'in_nucast_pkts': '9919499',
                                                                       'in_octets': '1437097013841',
                                                                       'in_ucast_pkts': '1009556370',
                                                                       'links': [       {       'in_nucast_pkts': '9919499',
                                                                                                'in_octets': '1437097013841',
                                                                                                'in_ucast_pkts': '1009556370',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160787',
                                                                                                'out_octets': '56718658465',
                                                                                                'out_ucast_pkts': '512241206',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160787',
                                                                       'out_octets': '56718658465',
                                                                       'out_ucast_pkts': '512241206',
                                                                       'poller_hash': 'NTc0MUEwcGVlcmludGZwb24gMzcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '371',
                                                                                               'links': [       {       'connected_time': '8535115',
                                                                                                                        'fiber_length': 25180,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.046,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.21}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0MUEwcGVlcnBvbnBvbiAzNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': '28 Welwood Av Wellesley (David '
                                                 'Honderich 10434636)',
                                         'poller_hash': 'NTc0MUEw',
                                         'serial_number': '5741A0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:44:f7',
                                         'uptime': '8535115'},
                                 {       'counter_source': '372/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '372/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '771733879395',
                                                                       'in_ucast_pkts': '6214741325',
                                                                       'mac_address': 'cc:be:59:cb:43:ea',
                                                                       'name': '372/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '3132818345813',
                                                                       'out_ucast_pkts': '3292400491',
                                                                       'poller_hash': 'NDJCOUQ1cGVlcmludGYzNzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '372',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '3132818345813',
                                                                       'in_ucast_pkts': '3292400491',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '3132818345813',
                                                                                                'in_ucast_pkts': '3292400491',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '771733879395',
                                                                                                'out_ucast_pkts': '6214741325',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '771733879395',
                                                                       'out_ucast_pkts': '6214741325',
                                                                       'poller_hash': 'NDJCOUQ1cGVlcmludGZwb24gMzcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '372',
                                                                                               'links': [       {       'connected_time': '5498457',
                                                                                                                        'fiber_length': 25373,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUQ1cGVlcnBvbnBvbiAzNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '22 JACK KOEHLER LN (MCKAY MICHAEL)',
                                         'poller_hash': 'NDJCOUQ1',
                                         'serial_number': '42B9D5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:ea',
                                         'uptime': '5498457'},
                                 {       'counter_source': '373/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:25:6f'],
                                                                       'id': '373/g1',
                                                                       'in_nucast_pkts': '12117',
                                                                       'in_octets': '272163044937',
                                                                       'in_ucast_pkts': '1110896093',
                                                                       'mac_address': 'cc:be:59:cb:43:85',
                                                                       'mtu': 2000,
                                                                       'name': '373/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710909',
                                                                       'out_octets': '5460782842977',
                                                                       'out_ucast_pkts': '4204060899',
                                                                       'poller_hash': 'NDJCOUM4cGVlcmludGYzNzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '373',
                                                                       'in_nucast_pkts': '61710909',
                                                                       'in_octets': '5460782842977',
                                                                       'in_ucast_pkts': '4204060899',
                                                                       'links': [       {       'in_nucast_pkts': '61710909',
                                                                                                'in_octets': '5460782842977',
                                                                                                'in_ucast_pkts': '4204060899',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12117',
                                                                                                'out_octets': '272163044937',
                                                                                                'out_ucast_pkts': '1110896093',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12117',
                                                                       'out_octets': '272163044937',
                                                                       'out_ucast_pkts': '1110896093',
                                                                       'poller_hash': 'NDJCOUM4cGVlcmludGZwb24gMzcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '373',
                                                                                               'links': [       {       'connected_time': '21887171',
                                                                                                                        'fiber_length': 23880,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUM4cGVlcnBvbnBvbiAzNzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 Ferris Dr Wellesley (Kristin Good '
                                                 '10435391)',
                                         'poller_hash': 'NDJCOUM4',
                                         'serial_number': '42B9C8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:82',
                                         'uptime': '21887171'},
                                 {       'counter_source': '374/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:72:ba'],
                                                                       'id': '374/g2',
                                                                       'in_nucast_pkts': '299038',
                                                                       'in_octets': '312613246088',
                                                                       'in_ucast_pkts': '659634998',
                                                                       'mac_address': 'cc:be:59:cb:43:8e',
                                                                       'mtu': 2000,
                                                                       'name': '374/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9917494',
                                                                       'out_octets': '1975604921149',
                                                                       'out_ucast_pkts': '1522798086',
                                                                       'poller_hash': 'NDJCOUM5cGVlcmludGYzNzQvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '374',
                                                                       'in_nucast_pkts': '9917494',
                                                                       'in_octets': '1975604921149',
                                                                       'in_ucast_pkts': '1522798086',
                                                                       'links': [       {       'in_nucast_pkts': '9917494',
                                                                                                'in_octets': '1975604921149',
                                                                                                'in_ucast_pkts': '1522798086',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '299038',
                                                                                                'out_octets': '312613246088',
                                                                                                'out_ucast_pkts': '659634998',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '299038',
                                                                       'out_octets': '312613246088',
                                                                       'out_ucast_pkts': '659634998',
                                                                       'poller_hash': 'NDJCOUM5cGVlcmludGZwb24gMzc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '374',
                                                                                               'links': [       {       'connected_time': '8535251',
                                                                                                                        'fiber_length': 24639,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUM5cGVlcnBvbnBvbiAzNzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '146 Gerber Meadows Dr Wellesley (K&G '
                                                 'Kingsbury 10434610)',
                                         'poller_hash': 'NDJCOUM5',
                                         'serial_number': '42B9C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:8a',
                                         'uptime': '8535251'},
                                 {       'counter_source': '375/F1+375/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a2:64:fb',
                                                                                                        '00:03:e6:aa:4e:fd'],
                                                                       'id': '375/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3583433090',
                                                                       'in_ucast_pkts': '19979073',
                                                                       'name': '375/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '17993847562',
                                                                       'out_octets': '24139531632985',
                                                                       'out_ucast_pkts': '9361548',
                                                                       'poller_hash': 'NDJCOUMxcGVlcmludGYzNzUvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:47:5a'],
                                                                       'id': '375/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '18956957219',
                                                                       'in_ucast_pkts': '78196310',
                                                                       'mac_address': 'cc:be:59:cb:43:4d',
                                                                       'mtu': 2000,
                                                                       'name': '375/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919841',
                                                                       'out_octets': '276904775360',
                                                                       'out_ucast_pkts': '227874251',
                                                                       'poller_hash': 'NDJCOUMxcGVlcmludGYzNzUvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '375',
                                                                       'in_nucast_pkts': '18003767403',
                                                                       'in_octets': '24416436408345',
                                                                       'in_ucast_pkts': '237235799',
                                                                       'links': [       {       'in_nucast_pkts': '18003767403',
                                                                                                'in_octets': '24416436408345',
                                                                                                'in_ucast_pkts': '237235799',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '22540390309',
                                                                                                'out_ucast_pkts': '98175383',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '22540390309',
                                                                       'out_ucast_pkts': '98175383',
                                                                       'poller_hash': 'NDJCOUMxcGVlcmludGZwb24gMzc1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '375',
                                                                                               'links': [       {       'connected_time': '8535235',
                                                                                                                        'fiber_length': 24614,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUMxcGVlcnBvbnBvbiAzNzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '86 NITHBANK RD (KUEPFER GLENN)',
                                         'poller_hash': 'NDJCOUMx',
                                         'serial_number': '42B9C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:4a',
                                         'uptime': '8535235'},
                                 {       'counter_source': '376/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:7a:ef'],
                                                                       'id': '376/g1',
                                                                       'in_nucast_pkts': '76304',
                                                                       'in_octets': '150425570888',
                                                                       'in_ucast_pkts': '560899145',
                                                                       'mac_address': 'cc:be:59:cb:44:7d',
                                                                       'mtu': 2000,
                                                                       'name': '376/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4981246',
                                                                       'out_octets': '1741492827496',
                                                                       'out_ucast_pkts': '1257300402',
                                                                       'poller_hash': 'NDJCOUU3cGVlcmludGYzNzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '376',
                                                                       'in_nucast_pkts': '4981246',
                                                                       'in_octets': '1741492827496',
                                                                       'in_ucast_pkts': '1257300402',
                                                                       'links': [       {       'in_nucast_pkts': '4981246',
                                                                                                'in_octets': '1741492827496',
                                                                                                'in_ucast_pkts': '1257300402',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '76304',
                                                                                                'out_octets': '150425570888',
                                                                                                'out_ucast_pkts': '560899145',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76304',
                                                                       'out_octets': '150425570888',
                                                                       'out_ucast_pkts': '560899145',
                                                                       'poller_hash': 'NDJCOUU3cGVlcmludGZwb24gMzc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '376',
                                                                                               'links': [       {       'connected_time': '4243188',
                                                                                                                        'fiber_length': 24838,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUU3cGVlcnBvbnBvbiAzNzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '115 GERBER MEADOWS DR (JACOB EDWARD '
                                                 '(ED))',
                                         'poller_hash': 'NDJCOUU3',
                                         'serial_number': '42B9E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:7a',
                                         'uptime': '4243188'},
                                 {       'counter_source': '377/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c3:52:e2'],
                                                                       'id': '377/g1',
                                                                       'in_nucast_pkts': '35229145',
                                                                       'in_octets': '525002300658',
                                                                       'in_ucast_pkts': '2316500628',
                                                                       'mac_address': 'cc:be:59:cb:44:3d',
                                                                       'mtu': 2000,
                                                                       'name': '377/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62522896',
                                                                       'out_octets': '7067492306618',
                                                                       'out_ucast_pkts': '5507948556',
                                                                       'poller_hash': 'NDJCOURGcGVlcmludGYzNzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '377',
                                                                       'in_nucast_pkts': '62522896',
                                                                       'in_octets': '7067492306618',
                                                                       'in_ucast_pkts': '5507948556',
                                                                       'links': [       {       'in_nucast_pkts': '62522896',
                                                                                                'in_octets': '7067492306618',
                                                                                                'in_ucast_pkts': '5507948556',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '35229145',
                                                                                                'out_octets': '525002300658',
                                                                                                'out_ucast_pkts': '2316500628',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35229145',
                                                                       'out_octets': '525002300658',
                                                                       'out_ucast_pkts': '2316500628',
                                                                       'poller_hash': 'NDJCOURGcGVlcmludGZwb24gMzc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '377',
                                                                                               'links': [       {       'connected_time': '21886917',
                                                                                                                        'fiber_length': 24943,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOURGcGVlcnBvbnBvbiAzNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '9 Village Rd Wellesley (Mark Norris '
                                                 '10434645)',
                                         'poller_hash': 'NDJCOURG',
                                         'serial_number': '42B9DF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:3a',
                                         'uptime': '21886917'},
                                 {       'counter_source': '378/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:59:6e:00'],
                                                                       'id': '378/g1',
                                                                       'in_nucast_pkts': '119039',
                                                                       'in_octets': '203990533648',
                                                                       'in_ucast_pkts': '708153419',
                                                                       'mac_address': 'f8:85:f9:17:29:03',
                                                                       'mtu': 2000,
                                                                       'name': '378/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7938827',
                                                                       'out_octets': '3036464927314',
                                                                       'out_ucast_pkts': '2365852718',
                                                                       'poller_hash': 'REZGRTA5cGVlcmludGYzNzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '378',
                                                                       'in_nucast_pkts': '7938827',
                                                                       'in_octets': '3036464927314',
                                                                       'in_ucast_pkts': '2365852718',
                                                                       'links': [       {       'in_nucast_pkts': '7938827',
                                                                                                'in_octets': '3036464927314',
                                                                                                'in_ucast_pkts': '2365852718',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '119039',
                                                                                                'out_octets': '203990533648',
                                                                                                'out_ucast_pkts': '708153419',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '119039',
                                                                       'out_octets': '203990533648',
                                                                       'out_ucast_pkts': '708153419',
                                                                       'poller_hash': 'REZGRTA5cGVlcmludGZwb24gMzc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '378',
                                                                                               'links': [       {       'connected_time': '6812304',
                                                                                                                        'fiber_length': 23696,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTA5cGVlcnBvbnBvbiAzNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '54 Zinkann Cr Wellesley (Stephanie '
                                                 'Gohl 10434787)',
                                         'poller_hash': 'REZGRTA5',
                                         'serial_number': 'DFFE09',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:29:00',
                                         'uptime': '6812304'},
                                 {       'counter_source': '379/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:eb:d7'],
                                                                       'id': '379/g1',
                                                                       'in_nucast_pkts': '151281',
                                                                       'in_octets': '111210166656',
                                                                       'in_ucast_pkts': '432610407',
                                                                       'mac_address': 'cc:be:59:f8:4a:4d',
                                                                       'mtu': 2000,
                                                                       'name': '379/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919608',
                                                                       'out_octets': '1810494385670',
                                                                       'out_ucast_pkts': '1372271308',
                                                                       'poller_hash': 'NTc0MjQ2cGVlcmludGYzNzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '379',
                                                                       'in_nucast_pkts': '9919608',
                                                                       'in_octets': '1810494385670',
                                                                       'in_ucast_pkts': '1372271308',
                                                                       'links': [       {       'in_nucast_pkts': '9919608',
                                                                                                'in_octets': '1810494385670',
                                                                                                'in_ucast_pkts': '1372271308',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '151281',
                                                                                                'out_octets': '111210166656',
                                                                                                'out_ucast_pkts': '432610407',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '151281',
                                                                       'out_octets': '111210166656',
                                                                       'out_ucast_pkts': '432610407',
                                                                       'poller_hash': 'NTc0MjQ2cGVlcmludGZwb24gMzc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '379',
                                                                                               'links': [       {       'connected_time': '8535107',
                                                                                                                        'fiber_length': 24177,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.574,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.612}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTc0MjQ2cGVlcnBvbnBvbiAzNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812Gv2',
                                         'name': '59 Parkview Dr Wellesley (Bryan '
                                                 'Sansom 10436194)',
                                         'poller_hash': 'NTc0MjQ2',
                                         'serial_number': '574246',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.5.4.0.240',
                                         'system_mac_address': 'cc:be:59:f8:4a:47',
                                         'uptime': '8535107'},
                                 {       'counter_source': '380/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ed:ad'],
                                                                       'id': '380/g1',
                                                                       'in_nucast_pkts': '65889636',
                                                                       'in_octets': '1115208762127',
                                                                       'in_ucast_pkts': '4352598238',
                                                                       'mac_address': 'cc:be:59:cb:44:1d',
                                                                       'mtu': 2000,
                                                                       'name': '380/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '91394784286',
                                                                       'out_octets': '144144325086147',
                                                                       'out_ucast_pkts': '16355846107',
                                                                       'poller_hash': 'NDJCOURCcGVlcmludGYzODAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '380',
                                                                       'in_nucast_pkts': '91394784286',
                                                                       'in_octets': '144144325086147',
                                                                       'in_ucast_pkts': '16355846107',
                                                                       'links': [       {       'in_nucast_pkts': '91394784286',
                                                                                                'in_octets': '144144325086147',
                                                                                                'in_ucast_pkts': '16355846107',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '65889636',
                                                                                                'out_octets': '1115208762127',
                                                                                                'out_ucast_pkts': '4352598238',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '65889636',
                                                                       'out_octets': '1115208762127',
                                                                       'out_ucast_pkts': '4352598238',
                                                                       'poller_hash': 'NDJCOURCcGVlcmludGZwb24gMzgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '380',
                                                                                               'links': [       {       'connected_time': '21886574',
                                                                                                                        'fiber_length': 25054,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOURCcGVlcnBvbnBvbiAzODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '36 VILLAGE RD (MILLER AMANDA)',
                                         'poller_hash': 'NDJCOURC',
                                         'serial_number': '42B9DB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:1a',
                                         'uptime': '21886574'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '382/g1',
                                                                       'name': '382/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QzgzMEUxcGVlcmludGYzODIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '382',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QzgzMEUxcGVlcmludGZwb24gMzgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '382',
                                                                                               'links': [       {       'connected_time': '8535289',
                                                                                                                        'fiber_length': 23309,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMEUxcGVlcnBvbnBvbiAzODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '118 Schweitzer Cr Wellesley (Esteban '
                                                 'Giron 10434568)',
                                         'poller_hash': 'QzgzMEUx',
                                         'serial_number': 'C830E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:db:bf',
                                         'uptime': '8535289'},
                                 {       'counter_source': '383/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e8:48:b8:c0:bd:95'],
                                                                       'id': '383/g1',
                                                                       'in_nucast_pkts': '6834',
                                                                       'in_octets': '38142530714',
                                                                       'in_ucast_pkts': '289612497',
                                                                       'mac_address': 'cc:be:59:50:03:de',
                                                                       'mtu': 2000,
                                                                       'name': '383/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5499530',
                                                                       'out_octets': '1258159581345',
                                                                       'out_ucast_pkts': '892507840',
                                                                       'poller_hash': 'NDNCNUI2cGVlcmludGYzODMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '383',
                                                                       'in_nucast_pkts': '5499530',
                                                                       'in_octets': '1258159581345',
                                                                       'in_ucast_pkts': '892507840',
                                                                       'links': [       {       'in_nucast_pkts': '5499530',
                                                                                                'in_octets': '1258159581345',
                                                                                                'in_ucast_pkts': '892507840',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6834',
                                                                                                'out_octets': '38142530714',
                                                                                                'out_ucast_pkts': '289612497',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6834',
                                                                       'out_octets': '38142530714',
                                                                       'out_ucast_pkts': '289612497',
                                                                       'poller_hash': 'NDNCNUI2cGVlcmludGZwb24gMzgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '383',
                                                                                               'links': [       {       'connected_time': '8485026',
                                                                                                                        'fiber_length': 24184,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.044,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.272}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUI2cGVlcnBvbnBvbiAzODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '11 CAMPBELL PL (MOSER MICHAEL)',
                                         'poller_hash': 'NDNCNUI2',
                                         'serial_number': '43B5B6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:03:dd',
                                         'uptime': '8485026'},
                                 {       'counter_source': '384/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:8e:8d'],
                                                                       'id': '384/g1',
                                                                       'in_nucast_pkts': '9618779',
                                                                       'in_octets': '205036598685',
                                                                       'in_ucast_pkts': '669598223',
                                                                       'mac_address': 'cc:be:59:b9:c3:83',
                                                                       'mtu': 2000,
                                                                       'name': '384/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919939',
                                                                       'out_octets': '1803606031454',
                                                                       'out_ucast_pkts': '1270319832',
                                                                       'poller_hash': 'NTEyNTA5cGVlcmludGYzODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '384',
                                                                       'in_nucast_pkts': '9919939',
                                                                       'in_octets': '1803606031454',
                                                                       'in_ucast_pkts': '1270319832',
                                                                       'links': [       {       'in_nucast_pkts': '9919939',
                                                                                                'in_octets': '1803606031454',
                                                                                                'in_ucast_pkts': '1270319832',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9618779',
                                                                                                'out_octets': '205036598685',
                                                                                                'out_ucast_pkts': '669598223',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9618779',
                                                                       'out_octets': '205036598685',
                                                                       'out_ucast_pkts': '669598223',
                                                                       'poller_hash': 'NTEyNTA5cGVlcmludGZwb24gMzg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '384',
                                                                                               'links': [       {       'connected_time': '8535385',
                                                                                                                        'fiber_length': 24928,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNTA5cGVlcnBvbnBvbiAzODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '13 ParkviewDr Wellesley (Mary Lichty '
                                                 '10434481)',
                                         'poller_hash': 'NTEyNTA5',
                                         'serial_number': '512509',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c3:83',
                                         'uptime': '8535385'},
                                 {       'counter_source': '385/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:92:08'],
                                                                       'id': '385/g1',
                                                                       'in_nucast_pkts': '56',
                                                                       'in_octets': '216585420505',
                                                                       'in_ucast_pkts': '1177364562',
                                                                       'mac_address': 'cc:be:59:b9:c6:23',
                                                                       'mtu': 2000,
                                                                       'name': '385/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38464281',
                                                                       'out_octets': '5441363382866',
                                                                       'out_ucast_pkts': '4054651133',
                                                                       'poller_hash': 'NTEyNUU5cGVlcmludGYzODUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '385',
                                                                       'in_nucast_pkts': '38464281',
                                                                       'in_octets': '5441363382866',
                                                                       'in_ucast_pkts': '4054651133',
                                                                       'links': [       {       'in_nucast_pkts': '38464281',
                                                                                                'in_octets': '5441363382866',
                                                                                                'in_ucast_pkts': '4054651133',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '56',
                                                                                                'out_octets': '216585420505',
                                                                                                'out_ucast_pkts': '1177364562',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '56',
                                                                       'out_octets': '216585420505',
                                                                       'out_ucast_pkts': '1177364562',
                                                                       'poller_hash': 'NTEyNUU5cGVlcmludGZwb24gMzg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '385',
                                                                                               'links': [       {       'connected_time': '67856209',
                                                                                                                        'fiber_length': 24242,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTEyNUU5cGVlcnBvbnBvbiAzODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '3 CAMPBELL PL (PFOHL JEFF)',
                                         'poller_hash': 'NTEyNUU5',
                                         'serial_number': '5125E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:b9:c6:23',
                                         'uptime': '67856209'},
                                 {       'counter_source': '387/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:cb:29'],
                                                                       'id': '387/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '44420951659',
                                                                       'in_ucast_pkts': '132031779',
                                                                       'mac_address': 'cc:be:59:4f:f7:6e',
                                                                       'mtu': 2000,
                                                                       'name': '387/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4414371',
                                                                       'out_octets': '278514337798',
                                                                       'out_ucast_pkts': '212332799',
                                                                       'poller_hash': 'NDNBRjgzcGVlcmludGYzODcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '387',
                                                                       'in_nucast_pkts': '4414371',
                                                                       'in_octets': '278514337798',
                                                                       'in_ucast_pkts': '212332799',
                                                                       'links': [       {       'in_nucast_pkts': '4414371',
                                                                                                'in_octets': '278514337798',
                                                                                                'in_ucast_pkts': '212332799',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '44420951659',
                                                                                                'out_ucast_pkts': '132031779',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '44420951659',
                                                                       'out_ucast_pkts': '132031779',
                                                                       'poller_hash': 'NDNBRjgzcGVlcmludGZwb24gMzg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '387',
                                                                                               'links': [       {       'connected_time': '6627411',
                                                                                                                        'fiber_length': 22462,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.046,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.244}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNBRjgzcGVlcnBvbnBvbiAzODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1177 Henry St Wellesley (Murray '
                                                 'Bremner 10434791)',
                                         'poller_hash': 'NDNBRjgz',
                                         'serial_number': '43AF83',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:4f:f7:6d',
                                         'uptime': '6627411'},
                                 {       'counter_source': '388/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b0:a7:b9:3c:d7:92'],
                                                                       'id': '388/g1',
                                                                       'in_nucast_pkts': '10899',
                                                                       'in_octets': '236683277489',
                                                                       'in_ucast_pkts': '905723935',
                                                                       'mac_address': 'cc:be:59:50:03:86',
                                                                       'mtu': 2000,
                                                                       'name': '388/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652385',
                                                                       'out_octets': '2304568453165',
                                                                       'out_ucast_pkts': '1847734253',
                                                                       'poller_hash': 'NDNCNThBcGVlcmludGYzODgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '388',
                                                                       'in_nucast_pkts': '5652385',
                                                                       'in_octets': '2304568453165',
                                                                       'in_ucast_pkts': '1847734253',
                                                                       'links': [       {       'in_nucast_pkts': '5652385',
                                                                                                'in_octets': '2304568453165',
                                                                                                'in_ucast_pkts': '1847734253',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10899',
                                                                                                'out_octets': '236683277489',
                                                                                                'out_ucast_pkts': '905723935',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10899',
                                                                       'out_octets': '236683277489',
                                                                       'out_ucast_pkts': '905723935',
                                                                       'poller_hash': 'NDNCNThBcGVlcmludGZwb24gMzg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '388',
                                                                                               'links': [       {       'connected_time': '8535083',
                                                                                                                        'fiber_length': 23050,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.25,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.126}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNThBcGVlcnBvbnBvbiAzODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1119 Molesworth St Wellesley (Dean '
                                                 'Savory 10434599)',
                                         'poller_hash': 'NDNCNThB',
                                         'serial_number': '43B58A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:03:85',
                                         'uptime': '8535083'},
                                 {       'counter_source': '389/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ac:15:a2:c8:aa:75'],
                                                                       'id': '389/g1',
                                                                       'in_nucast_pkts': '118428',
                                                                       'in_octets': '1434748618258',
                                                                       'in_ucast_pkts': '3649198448',
                                                                       'mac_address': 'cc:be:59:cb:43:b5',
                                                                       'mtu': 2000,
                                                                       'name': '389/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '42407059',
                                                                       'out_octets': '11176338817054',
                                                                       'out_ucast_pkts': '8680444622',
                                                                       'poller_hash': 'NDJCOUNFcGVlcmludGYzODkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '389',
                                                                       'in_nucast_pkts': '42407059',
                                                                       'in_octets': '11176338817054',
                                                                       'in_ucast_pkts': '8680444622',
                                                                       'links': [       {       'in_nucast_pkts': '42407059',
                                                                                                'in_octets': '11176338817054',
                                                                                                'in_ucast_pkts': '8680444622',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '118428',
                                                                                                'out_octets': '1434748618258',
                                                                                                'out_ucast_pkts': '3649198448',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '118428',
                                                                       'out_octets': '1434748618258',
                                                                       'out_ucast_pkts': '3649198448',
                                                                       'poller_hash': 'NDJCOUNFcGVlcmludGZwb24gMzg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '389',
                                                                                               'links': [       {       'connected_time': '72550404',
                                                                                                                        'fiber_length': 23869,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUNFcGVlcnBvbnBvbiAzODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1205 GREENWOOD HILL RD (JOHNSTONE '
                                                 'TANYA)',
                                         'poller_hash': 'NDJCOUNF',
                                         'serial_number': '42B9CE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:b2',
                                         'uptime': '72550404'},
                                 {       'counter_source': '390/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:8f:02'],
                                                                       'id': '390/g1',
                                                                       'in_nucast_pkts': '7948',
                                                                       'in_octets': '280164423310',
                                                                       'in_ucast_pkts': '3225566026',
                                                                       'mac_address': 'cc:be:59:cb:43:c5',
                                                                       'mtu': 2000,
                                                                       'name': '390/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61710415',
                                                                       'out_octets': '16755554756970',
                                                                       'out_ucast_pkts': '11175415129',
                                                                       'poller_hash': 'NDJCOUQwcGVlcmludGYzOTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '390',
                                                                       'in_nucast_pkts': '61710415',
                                                                       'in_octets': '16755554756970',
                                                                       'in_ucast_pkts': '11175415129',
                                                                       'links': [       {       'in_nucast_pkts': '61710415',
                                                                                                'in_octets': '16755554756970',
                                                                                                'in_ucast_pkts': '11175415129',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7948',
                                                                                                'out_octets': '280164423310',
                                                                                                'out_ucast_pkts': '3225566026',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7948',
                                                                       'out_octets': '280164423310',
                                                                       'out_ucast_pkts': '3225566026',
                                                                       'poller_hash': 'NDJCOUQwcGVlcmludGZwb24gMzkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '390',
                                                                                               'links': [       {       'connected_time': '21886519',
                                                                                                                        'fiber_length': 22674,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUQwcGVlcnBvbnBvbiAzOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '22 Lawrence St Wellesley (Elizabeth '
                                                 'Eades-Kirby 10434655)',
                                         'poller_hash': 'NDJCOUQw',
                                         'serial_number': '42B9D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:43:c2',
                                         'uptime': '21886519'},
                                 {       'counter_source': '393/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '30:23:03:6c:1a:ae'],
                                                                       'id': '393/g1',
                                                                       'in_nucast_pkts': '382622',
                                                                       'in_octets': '163567419730',
                                                                       'in_ucast_pkts': '486046307',
                                                                       'mac_address': 'cc:be:59:50:04:20',
                                                                       'mtu': 2000,
                                                                       'name': '393/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5657731',
                                                                       'out_octets': '1472339523740',
                                                                       'out_ucast_pkts': '1177637716',
                                                                       'poller_hash': 'NDNCNUQ3cGVlcmludGYzOTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '393',
                                                                       'in_nucast_pkts': '5657731',
                                                                       'in_octets': '1472339523740',
                                                                       'in_ucast_pkts': '1177637716',
                                                                       'links': [       {       'in_nucast_pkts': '5657731',
                                                                                                'in_octets': '1472339523740',
                                                                                                'in_ucast_pkts': '1177637716',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '382622',
                                                                                                'out_octets': '163567419730',
                                                                                                'out_ucast_pkts': '486046307',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '382622',
                                                                       'out_octets': '163567419730',
                                                                       'out_ucast_pkts': '486046307',
                                                                       'poller_hash': 'NDNCNUQ3cGVlcmludGZwb24gMzkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '393',
                                                                                               'links': [       {       'connected_time': '8535087',
                                                                                                                        'fiber_length': 23381,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.324,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.018}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUQ3cGVlcnBvbnBvbiAzOTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '135 Ferris Dr Wellesley (Steven '
                                                 'Curliss 10434479)',
                                         'poller_hash': 'NDNCNUQ3',
                                         'serial_number': '43B5D7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:04:1f',
                                         'uptime': '8535087'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '394/g1',
                                                                       'name': '394/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJEMEI0cGVlcmludGYzOTQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '394',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJEMEI0cGVlcmludGZwb24gMzk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '394',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMEI0cGVlcnBvbnBvbiAzOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'MCCL Test 812 (8xPon)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "394" '
                                                                  '(location: GPON '
                                                                  'port "5/8"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2025-05-07T18:20:38.690Z',
                                         'poller_hash': 'NDJEMEI0',
                                         'serial_number': '42D0B4',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '395/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '395/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '293295793808',
                                                                       'in_ucast_pkts': '1182716930',
                                                                       'mac_address': 'cc:be:59:cc:11:ca',
                                                                       'name': '395/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '50',
                                                                       'out_octets': '5930673199226',
                                                                       'out_ucast_pkts': '4412364627',
                                                                       'poller_hash': 'NDJEMzkxcGVlcmludGYzOTUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '395',
                                                                       'in_nucast_pkts': '50',
                                                                       'in_octets': '5930673199226',
                                                                       'in_ucast_pkts': '4412364627',
                                                                       'links': [       {       'in_nucast_pkts': '50',
                                                                                                'in_octets': '5930673199226',
                                                                                                'in_ucast_pkts': '4412364627',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '293295793808',
                                                                                                'out_ucast_pkts': '1182716930',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '293295793808',
                                                                       'out_ucast_pkts': '1182716930',
                                                                       'poller_hash': 'NDJEMzkxcGVlcmludGZwb24gMzk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '395',
                                                                                               'links': [       {       'connected_time': '3507322',
                                                                                                                        'fiber_length': 25138,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMzkxcGVlcnBvbnBvbiAzOTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '56 VILLAGE RD (CROZIER SCOT)',
                                         'poller_hash': 'NDJEMzkx',
                                         'serial_number': '42D391',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:11:ca',
                                         'uptime': '3507322'},
                                 {       'counter_source': '396/F1+396/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:34:50',
                                                                                                        '00:03:e6:eb:3c:ce'],
                                                                       'id': '396/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2394690598',
                                                                       'in_ucast_pkts': '12790483',
                                                                       'name': '396/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '6775972824',
                                                                       'out_octets': '9065504088303',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJEMEM2cGVlcmludGYzOTYvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:8d:c7'],
                                                                       'id': '396/g1',
                                                                       'in_nucast_pkts': '13',
                                                                       'in_octets': '2571363020',
                                                                       'in_ucast_pkts': '16136488',
                                                                       'mac_address': 'cc:be:59:cb:fb:75',
                                                                       'mtu': 2000,
                                                                       'name': '396/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715344',
                                                                       'out_octets': '59356142692',
                                                                       'out_ucast_pkts': '43992822',
                                                                       'poller_hash': 'NDJEMEM2cGVlcmludGYzOTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '396',
                                                                       'in_nucast_pkts': '6776688168',
                                                                       'in_octets': '9124860230995',
                                                                       'in_ucast_pkts': '43992822',
                                                                       'links': [       {       'in_nucast_pkts': '6776688168',
                                                                                                'in_octets': '9124860230995',
                                                                                                'in_ucast_pkts': '43992822',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13',
                                                                                                'out_octets': '4966053618',
                                                                                                'out_ucast_pkts': '28926971',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13',
                                                                       'out_octets': '4966053618',
                                                                       'out_ucast_pkts': '28926971',
                                                                       'poller_hash': 'NDJEMEM2cGVlcmludGZwb24gMzk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '396',
                                                                                               'links': [       {       'connected_time': '611455',
                                                                                                                        'fiber_length': 19728,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMEM2cGVlcnBvbnBvbiAzOTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4012 Andy St Gadshill (April Rose '
                                                 '10435328)',
                                         'poller_hash': 'NDJEMEM2',
                                         'serial_number': '42D0C6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:fb:72',
                                         'uptime': '611455'},
                                 {       'counter_source': '397/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '397/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45112092697',
                                                                       'in_ucast_pkts': '136177023',
                                                                       'mac_address': 'cc:be:59:cb:e6:f2',
                                                                       'name': '397/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '444426574031',
                                                                       'out_ucast_pkts': '402011697',
                                                                       'poller_hash': 'NDJDRTM2cGVlcmludGYzOTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '397',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '444426574031',
                                                                       'in_ucast_pkts': '402011697',
                                                                       'links': [       {       'in_nucast_pkts': '117',
                                                                                                'in_octets': '444426574031',
                                                                                                'in_ucast_pkts': '402011697',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45112092697',
                                                                                                'out_ucast_pkts': '136177023',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45112092697',
                                                                       'out_ucast_pkts': '136177023',
                                                                       'poller_hash': 'NDJDRTM2cGVlcmludGZwb24gMzk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '397',
                                                                                               'links': [       {       'connected_time': '21886551',
                                                                                                                        'fiber_length': 15734,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTM2cGVlcnBvbnBvbiAzOTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2188 Perth Line Wellesley (P & B '
                                                 'Albrecht 10427546)',
                                         'poller_hash': 'NDJDRTM2',
                                         'serial_number': '42CE36',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e6:f2',
                                         'uptime': '21886551'},
                                 {       'counter_source': '398/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:41:94:1b'],
                                                                       'id': '398/g1',
                                                                       'in_nucast_pkts': '10857593',
                                                                       'in_octets': '162043522087',
                                                                       'in_ucast_pkts': '768868836',
                                                                       'mac_address': 'cc:be:59:cb:b8:15',
                                                                       'mtu': 2000,
                                                                       'name': '398/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8365952946',
                                                                       'out_octets': '15218162727406',
                                                                       'out_ucast_pkts': '3469436704',
                                                                       'poller_hash': 'NDJDODVBcGVlcmludGYzOTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '398',
                                                                       'in_nucast_pkts': '8365952946',
                                                                       'in_octets': '15218162727406',
                                                                       'in_ucast_pkts': '3469436704',
                                                                       'links': [       {       'in_nucast_pkts': '8365952946',
                                                                                                'in_octets': '15218162727406',
                                                                                                'in_ucast_pkts': '3469436704',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10857593',
                                                                                                'out_octets': '162043522087',
                                                                                                'out_ucast_pkts': '768868836',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10857593',
                                                                       'out_octets': '162043522087',
                                                                       'out_ucast_pkts': '768868836',
                                                                       'poller_hash': 'NDJDODVBcGVlcmludGZwb24gMzk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '398',
                                                                                               'links': [       {       'connected_time': '21886968',
                                                                                                                        'fiber_length': 22829,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDODVBcGVlcnBvbnBvbiAzOTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'CLARE WAGLER (30 PONDVIEW DR WELL)',
                                         'poller_hash': 'NDJDODVB',
                                         'serial_number': '42C85A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:b8:12',
                                         'uptime': '21886968'},
                                 {       'counter_source': '399/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '399/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '177076283056',
                                                                       'in_ucast_pkts': '506066100',
                                                                       'mac_address': 'cc:be:59:cc:06:ba',
                                                                       'name': '399/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '60',
                                                                       'out_octets': '1234430340024',
                                                                       'out_ucast_pkts': '1104603721',
                                                                       'poller_hash': 'NDJEMjJGcGVlcmludGYzOTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '399',
                                                                       'in_nucast_pkts': '60',
                                                                       'in_octets': '1234430340024',
                                                                       'in_ucast_pkts': '1104603721',
                                                                       'links': [       {       'in_nucast_pkts': '60',
                                                                                                'in_octets': '1234430340024',
                                                                                                'in_ucast_pkts': '1104603721',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '177076283056',
                                                                                                'out_ucast_pkts': '506066100',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '177076283056',
                                                                       'out_ucast_pkts': '506066100',
                                                                       'poller_hash': 'NDJEMjJGcGVlcmludGZwb24gMzk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '399',
                                                                                               'links': [       {       'connected_time': '10763837',
                                                                                                                        'fiber_length': 17877,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMjJGcGVlcnBvbnBvbiAzOTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2014 PERTH LINE (WAGLER DAVID)',
                                         'poller_hash': 'NDJEMjJG',
                                         'serial_number': '42D22F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:06:ba',
                                         'uptime': '10763837'},
                                 {       'counter_source': '400/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd4:5d:64:1b:57:00'],
                                                                       'id': '400/g1',
                                                                       'in_nucast_pkts': '2965',
                                                                       'in_octets': '14295183843',
                                                                       'in_ucast_pkts': '11570532',
                                                                       'mac_address': 'd0:76:8f:08:14:de',
                                                                       'mtu': 2000,
                                                                       'name': '400/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '433401',
                                                                       'out_octets': '6426091526',
                                                                       'out_ucast_pkts': '7838387',
                                                                       'poller_hash': 'ODY5MzkxcGVlcmludGY0MDAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '400',
                                                                       'in_nucast_pkts': '433401',
                                                                       'in_octets': '6426091526',
                                                                       'in_ucast_pkts': '7838387',
                                                                       'links': [       {       'in_nucast_pkts': '433401',
                                                                                                'in_octets': '6426091526',
                                                                                                'in_ucast_pkts': '7838387',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2965',
                                                                                                'out_octets': '14295183843',
                                                                                                'out_ucast_pkts': '11570532',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2965',
                                                                       'out_octets': '14295183843',
                                                                       'out_ucast_pkts': '11570532',
                                                                       'poller_hash': 'ODY5MzkxcGVlcmludGZwb24gNDAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '400',
                                                                                               'links': [       {       'connected_time': '498961',
                                                                                                                        'fiber_length': 19979,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzkxcGVlcnBvbnBvbiA0MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1010 Hutchison Rd Wellesley (HBEND '
                                                 'ENTERPRISES)',
                                         'poller_hash': 'ODY5Mzkx',
                                         'serial_number': '869391',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:14:db',
                                         'uptime': '498961'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '401/g1',
                                                                       'name': '401/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCMDgzcGVlcmludGY0MDEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '401',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNCMDgzcGVlcmludGZwb24gNDAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '401',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCMDgzcGVlcnBvbnBvbiA0MDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '1214 Hutchinson Rd Wellesley (Conrad '
                                                 'Jantzi 10434281)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "401" '
                                                                  '(location: GPON '
                                                                  'port "5/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-03-24T13:39:11.470Z',
                                         'poller_hash': 'NDNCMDgz',
                                         'serial_number': '43B083',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '402/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3c:80'],
                                                                       'id': '402/g1',
                                                                       'in_nucast_pkts': '67054',
                                                                       'in_octets': '56170455057',
                                                                       'in_ucast_pkts': '72888821',
                                                                       'mac_address': 'cc:be:59:cc:05:0d',
                                                                       'mtu': 2000,
                                                                       'name': '402/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4043754',
                                                                       'out_octets': '178326857515',
                                                                       'out_ucast_pkts': '155673275',
                                                                       'poller_hash': 'NDJEMUY5cGVlcmludGY0MDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '402',
                                                                       'in_nucast_pkts': '4043754',
                                                                       'in_octets': '178326857515',
                                                                       'in_ucast_pkts': '155673275',
                                                                       'links': [       {       'in_nucast_pkts': '4043754',
                                                                                                'in_octets': '178326857515',
                                                                                                'in_ucast_pkts': '155673275',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '67054',
                                                                                                'out_octets': '56170455057',
                                                                                                'out_ucast_pkts': '72888821',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67054',
                                                                       'out_octets': '56170455057',
                                                                       'out_ucast_pkts': '72888821',
                                                                       'poller_hash': 'NDJEMUY5cGVlcmludGZwb24gNDAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '402',
                                                                                               'links': [       {       'connected_time': '3434334',
                                                                                                                        'fiber_length': 20911,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMUY5cGVlcnBvbnBvbiA0MDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2480 GERBER RD (WAGLER PHILIP)',
                                         'poller_hash': 'NDJEMUY5',
                                         'serial_number': '42D1F9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:05:0a',
                                         'uptime': '3434334'},
                                 {       'counter_source': '404/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:31:4e:fa'],
                                                                       'id': '404/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '94993782314',
                                                                       'in_ucast_pkts': '102717716',
                                                                       'mac_address': 'cc:be:59:50:00:26',
                                                                       'mtu': 2000,
                                                                       'name': '404/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '976234',
                                                                       'out_octets': '103829153232',
                                                                       'out_ucast_pkts': '115570354',
                                                                       'poller_hash': 'NDNCM0RDcGVlcmludGY0MDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '404',
                                                                       'in_nucast_pkts': '976234',
                                                                       'in_octets': '103829153232',
                                                                       'in_ucast_pkts': '115570354',
                                                                       'links': [       {       'in_nucast_pkts': '976234',
                                                                                                'in_octets': '103829153232',
                                                                                                'in_ucast_pkts': '115570354',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '94993782314',
                                                                                                'out_ucast_pkts': '102717716',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '94993782314',
                                                                       'out_ucast_pkts': '102717716',
                                                                       'poller_hash': 'NDNCM0RDcGVlcmludGZwb24gNDA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '404',
                                                                                               'links': [       {       'connected_time': '1448873',
                                                                                                                        'fiber_length': 24359,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.108,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.224}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCM0RDcGVlcnBvbnBvbiA0MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1385 Greenwood Hill Rd Wellesley '
                                                 '(Nelson Hoover 10435286)',
                                         'poller_hash': 'NDNCM0RD',
                                         'serial_number': '43B3DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:00:25',
                                         'uptime': '1448873'},
                                 {       'counter_source': '405/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '405/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1333255693325',
                                                                       'in_ucast_pkts': '2587847511',
                                                                       'mac_address': 'cc:be:59:cb:e6:72',
                                                                       'name': '405/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '99',
                                                                       'out_octets': '9699658278209',
                                                                       'out_ucast_pkts': '7626726221',
                                                                       'poller_hash': 'NDJDRTI2cGVlcmludGY0MDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '405',
                                                                       'in_nucast_pkts': '99',
                                                                       'in_octets': '9699658278209',
                                                                       'in_ucast_pkts': '7626726221',
                                                                       'links': [       {       'in_nucast_pkts': '99',
                                                                                                'in_octets': '9699658278209',
                                                                                                'in_ucast_pkts': '7626726221',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1333255693325',
                                                                                                'out_ucast_pkts': '2587847511',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1333255693325',
                                                                       'out_ucast_pkts': '2587847511',
                                                                       'poller_hash': 'NDJDRTI2cGVlcmludGZwb24gNDA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '405',
                                                                                               'links': [       {       'connected_time': '3961424',
                                                                                                                        'fiber_length': 24639,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTI2cGVlcnBvbnBvbiA0MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1368 Greenwood Hill (Jerry Mailloux, '
                                                 '10435021)',
                                         'poller_hash': 'NDJDRTI2',
                                         'serial_number': '42CE26',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e6:72',
                                         'uptime': '3961424'},
                                 {       'counter_source': '406/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:a5:6e:93',
                                                                                                        '00:03:e6:a5:6e:a1',
                                                                                                        '00:03:e6:a5:6f:34',
                                                                                                        'cc:be:59:7e:b5:ba'],
                                                                       'id': '406/g1',
                                                                       'in_nucast_pkts': '6473499',
                                                                       'in_octets': '52854217759',
                                                                       'in_ucast_pkts': '442236921',
                                                                       'mac_address': 'cc:be:59:cb:e7:55',
                                                                       'mtu': 2000,
                                                                       'name': '406/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10655081701',
                                                                       'out_octets': '16033687491313',
                                                                       'out_ucast_pkts': '1133637356',
                                                                       'poller_hash': 'NDJDRTQycGVlcmludGY0MDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '406',
                                                                       'in_nucast_pkts': '10655081701',
                                                                       'in_octets': '16033687491313',
                                                                       'in_ucast_pkts': '1133637356',
                                                                       'links': [       {       'in_nucast_pkts': '10655081701',
                                                                                                'in_octets': '16033687491313',
                                                                                                'in_ucast_pkts': '1133637356',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6473499',
                                                                                                'out_octets': '52854217759',
                                                                                                'out_ucast_pkts': '442236921',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6473499',
                                                                       'out_octets': '52854217759',
                                                                       'out_ucast_pkts': '442236921',
                                                                       'poller_hash': 'NDJDRTQycGVlcmludGZwb24gNDA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '406',
                                                                                               'links': [       {       'connected_time': '8535285',
                                                                                                                        'fiber_length': 23290,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTQycGVlcnBvbnBvbiA0MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1080 Henry St (Brock Gerber '
                                                 '10434382)',
                                         'poller_hash': 'NDJDRTQy',
                                         'serial_number': '42CE42',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e7:52',
                                         'uptime': '8535285'},
                                 {       'counter_source': '408/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '408/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '223104530637',
                                                                       'in_ucast_pkts': '740595716',
                                                                       'mac_address': 'cc:be:59:cb:e8:12',
                                                                       'name': '408/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '132',
                                                                       'out_octets': '4068733618660',
                                                                       'out_ucast_pkts': '3022332186',
                                                                       'poller_hash': 'NDJDRTVBcGVlcmludGY0MDgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '408',
                                                                       'in_nucast_pkts': '132',
                                                                       'in_octets': '4068733618660',
                                                                       'in_ucast_pkts': '3022332186',
                                                                       'links': [       {       'in_nucast_pkts': '132',
                                                                                                'in_octets': '4068733618660',
                                                                                                'in_ucast_pkts': '3022332186',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '223104530637',
                                                                                                'out_ucast_pkts': '740595716',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '223104530637',
                                                                       'out_ucast_pkts': '740595716',
                                                                       'poller_hash': 'NDJDRTVBcGVlcmludGZwb24gNDA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '408',
                                                                                               'links': [       {       'connected_time': '5596335',
                                                                                                                        'fiber_length': 23096,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTVBcGVlcnBvbnBvbiA0MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3604 NAFZIGER RD (SCHNARR MEGHAN)',
                                         'poller_hash': 'NDJDRTVB',
                                         'serial_number': '42CE5A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e8:12',
                                         'uptime': '5596335'},
                                 {       'counter_source': '409/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f0:af'],
                                                                       'id': '409/g1',
                                                                       'in_nucast_pkts': '235537',
                                                                       'in_octets': '1098328894615',
                                                                       'in_ucast_pkts': '1467437943',
                                                                       'mac_address': '44:65:7f:7b:b9:c9',
                                                                       'mtu': 2000,
                                                                       'name': '409/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13895039',
                                                                       'out_octets': '4956805661516',
                                                                       'out_ucast_pkts': '3754894800',
                                                                       'poller_hash': 'NUE2MkRFcGVlcmludGY0MDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '409',
                                                                       'in_nucast_pkts': '13895039',
                                                                       'in_octets': '4956805661516',
                                                                       'in_ucast_pkts': '3754894800',
                                                                       'links': [       {       'in_nucast_pkts': '13895039',
                                                                                                'in_octets': '4956805661516',
                                                                                                'in_ucast_pkts': '3754894800',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '235537',
                                                                                                'out_octets': '1098328894615',
                                                                                                'out_ucast_pkts': '1467437943',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '235537',
                                                                       'out_octets': '1098328894615',
                                                                       'out_ucast_pkts': '1467437943',
                                                                       'poller_hash': 'NUE2MkRFcGVlcmludGZwb24gNDA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '409',
                                                                                               'links': [       {       'connected_time': '12093658',
                                                                                                                        'fiber_length': 23962,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2MkRFcGVlcnBvbnBvbiA0MDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': "1020 QUEEN'S BUSH RD (WESTMORLAND "
                                                 'BRENDAN)',
                                         'poller_hash': 'NUE2MkRF',
                                         'serial_number': '5A62DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:b9:c9',
                                         'uptime': '12093658'},
                                 {       'counter_source': '410/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:91:ca'],
                                                                       'id': '410/g1',
                                                                       'in_nucast_pkts': '12932',
                                                                       'in_octets': '2152365322',
                                                                       'in_ucast_pkts': '11543084',
                                                                       'mac_address': 'cc:be:59:cb:e1:d5',
                                                                       'mtu': 2000,
                                                                       'name': '410/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715414',
                                                                       'out_octets': '40492887944',
                                                                       'out_ucast_pkts': '32501933',
                                                                       'poller_hash': 'NDJDRDkycGVlcmludGY0MTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '410',
                                                                       'in_nucast_pkts': '715414',
                                                                       'in_octets': '40492887944',
                                                                       'in_ucast_pkts': '32501933',
                                                                       'links': [       {       'in_nucast_pkts': '715414',
                                                                                                'in_octets': '40492887944',
                                                                                                'in_ucast_pkts': '32501933',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12932',
                                                                                                'out_octets': '2152365322',
                                                                                                'out_ucast_pkts': '11543084',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12932',
                                                                       'out_octets': '2152365322',
                                                                       'out_ucast_pkts': '11543084',
                                                                       'poller_hash': 'NDJDRDkycGVlcmludGZwb24gNDEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '410',
                                                                                               'links': [       {       'connected_time': '611447',
                                                                                                                        'fiber_length': 19726,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -23.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRDkycGVlcnBvbnBvbiA0MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4016 ANDY ST (MARTIN JEFF)',
                                         'poller_hash': 'NDJDRDky',
                                         'serial_number': '42CD92',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e1:d2',
                                         'uptime': '611447'},
                                 {       'counter_source': '411/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:42:9d'],
                                                                       'id': '411/g1',
                                                                       'in_nucast_pkts': '27',
                                                                       'in_octets': '62687032766',
                                                                       'in_ucast_pkts': '326663815',
                                                                       'mac_address': 'cc:be:59:cb:ff:55',
                                                                       'mtu': 2000,
                                                                       'name': '411/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920208',
                                                                       'out_octets': '941723265249',
                                                                       'out_ucast_pkts': '700072206',
                                                                       'poller_hash': 'NDJEMTQycGVlcmludGY0MTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '411',
                                                                       'in_nucast_pkts': '9920208',
                                                                       'in_octets': '941723265249',
                                                                       'in_ucast_pkts': '700072206',
                                                                       'links': [       {       'in_nucast_pkts': '9920208',
                                                                                                'in_octets': '941723265249',
                                                                                                'in_ucast_pkts': '700072206',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27',
                                                                                                'out_octets': '62687032766',
                                                                                                'out_ucast_pkts': '326663815',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27',
                                                                       'out_octets': '62687032766',
                                                                       'out_ucast_pkts': '326663815',
                                                                       'poller_hash': 'NDJEMTQycGVlcmludGZwb24gNDEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '411',
                                                                                               'links': [       {       'connected_time': '8535309',
                                                                                                                        'fiber_length': 22618,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMTQycGVlcnBvbnBvbiA0MTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '73 David St Wellesley (Stephen Bader '
                                                 '10434843)',
                                         'poller_hash': 'NDJEMTQy',
                                         'serial_number': '42D142',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ff:52',
                                         'uptime': '8535309'},
                                 {       'counter_source': '412/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3b:0a'],
                                                                       'id': '412/g1',
                                                                       'in_nucast_pkts': '2427934',
                                                                       'in_octets': '94457073622',
                                                                       'in_ucast_pkts': '472054140',
                                                                       'mac_address': '84:d3:43:7e:2f:e2',
                                                                       'mtu': 2000,
                                                                       'name': '412/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3551422255',
                                                                       'out_octets': '6546242536242',
                                                                       'out_ucast_pkts': '1362194228',
                                                                       'poller_hash': 'QzgzQjY1cGVlcmludGY0MTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '412',
                                                                       'in_nucast_pkts': '3551422255',
                                                                       'in_octets': '6546242536242',
                                                                       'in_ucast_pkts': '1362194228',
                                                                       'links': [       {       'in_nucast_pkts': '3551422255',
                                                                                                'in_octets': '6546242536242',
                                                                                                'in_ucast_pkts': '1362194228',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2427934',
                                                                                                'out_octets': '94457073622',
                                                                                                'out_ucast_pkts': '472054140',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2427934',
                                                                       'out_octets': '94457073622',
                                                                       'out_ucast_pkts': '472054140',
                                                                       'poller_hash': 'QzgzQjY1cGVlcmludGZwb24gNDEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '412',
                                                                                               'links': [       {       'connected_time': '8535310',
                                                                                                                        'fiber_length': 23326,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjY1cGVlcnBvbnBvbiA0MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '157 Lawrence St Wellesley (Lisa '
                                                 'Snider 10435028)',
                                         'poller_hash': 'QzgzQjY1',
                                         'serial_number': 'C83B65',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:2f:df',
                                         'uptime': '8535310'},
                                 {       'counter_source': '413/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:b4:1a',
                                                                                                        '00:03:e6:da:bf:48',
                                                                                                        '00:03:e6:da:c0:aa',
                                                                                                        'b8:94:70:84:96:73'],
                                                                       'id': '413/g1',
                                                                       'in_nucast_pkts': '40490209',
                                                                       'in_octets': '1349087950404',
                                                                       'in_ucast_pkts': '7744542290',
                                                                       'mac_address': 'cc:be:59:cc:0a:75',
                                                                       'mtu': 2000,
                                                                       'name': '413/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '102157373540',
                                                                       'out_octets': '173138980398546',
                                                                       'out_ucast_pkts': '25632041157',
                                                                       'poller_hash': 'NDJEMkE2cGVlcmludGY0MTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '413',
                                                                       'in_nucast_pkts': '102157373540',
                                                                       'in_octets': '173138980398546',
                                                                       'in_ucast_pkts': '25632041157',
                                                                       'links': [       {       'in_nucast_pkts': '102157373540',
                                                                                                'in_octets': '173138980398546',
                                                                                                'in_ucast_pkts': '25632041157',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '40490209',
                                                                                                'out_octets': '1349087950404',
                                                                                                'out_ucast_pkts': '7744542290',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '40490209',
                                                                       'out_octets': '1349087950404',
                                                                       'out_ucast_pkts': '7744542290',
                                                                       'poller_hash': 'NDJEMkE2cGVlcmludGZwb24gNDEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '413',
                                                                                               'links': [       {       'connected_time': '7835280',
                                                                                                                        'fiber_length': 25139,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMkE2cGVlcnBvbnBvbiA0MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '75 GERBER MEADOWS DR (GERBER NICOLE)',
                                         'poller_hash': 'NDJEMkE2',
                                         'serial_number': '42D2A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:0a:72',
                                         'uptime': '7835280'},
                                 {       'counter_source': '414/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:bb'],
                                                                       'id': '414/g1',
                                                                       'in_nucast_pkts': '18',
                                                                       'in_octets': '18055108729',
                                                                       'in_ucast_pkts': '79989374',
                                                                       'mac_address': 'cc:be:59:cb:fc:fd',
                                                                       'mtu': 2000,
                                                                       'name': '414/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1181098',
                                                                       'out_octets': '374206752229',
                                                                       'out_ucast_pkts': '266056156',
                                                                       'poller_hash': 'NDJEMEY3cGVlcmludGY0MTQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '414',
                                                                       'in_nucast_pkts': '1181098',
                                                                       'in_octets': '374206752229',
                                                                       'in_ucast_pkts': '266056156',
                                                                       'links': [       {       'in_nucast_pkts': '1181098',
                                                                                                'in_octets': '374206752229',
                                                                                                'in_ucast_pkts': '266056156',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18',
                                                                                                'out_octets': '18055108729',
                                                                                                'out_ucast_pkts': '79989374',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18',
                                                                       'out_octets': '18055108729',
                                                                       'out_ucast_pkts': '79989374',
                                                                       'poller_hash': 'NDJEMEY3cGVlcmludGZwb24gNDE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '414',
                                                                                               'links': [       {       'connected_time': '1008584',
                                                                                                                        'fiber_length': 22989,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMEY3cGVlcnBvbnBvbiA0MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '97 Lawrence St Wellesley (Michel '
                                                 'Houle 10434807)',
                                         'poller_hash': 'NDJEMEY3',
                                         'serial_number': '42D0F7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:fc:fa',
                                         'uptime': '1008584'},
                                 {       'counter_source': '415/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:19'],
                                                                       'id': '415/g1',
                                                                       'in_nucast_pkts': '49107598',
                                                                       'in_octets': '517966686345',
                                                                       'in_ucast_pkts': '3428958706',
                                                                       'mac_address': 'cc:be:59:cc:0b:85',
                                                                       'mtu': 2000,
                                                                       'name': '415/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '64910274738',
                                                                       'out_octets': '99367977926098',
                                                                       'out_ucast_pkts': '8922844018',
                                                                       'poller_hash': 'NDJEMkM4cGVlcmludGY0MTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '415',
                                                                       'in_nucast_pkts': '64910274738',
                                                                       'in_octets': '99367977926098',
                                                                       'in_ucast_pkts': '8922844018',
                                                                       'links': [       {       'in_nucast_pkts': '64910274738',
                                                                                                'in_octets': '99367977926098',
                                                                                                'in_ucast_pkts': '8922844018',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '49107598',
                                                                                                'out_octets': '517966686345',
                                                                                                'out_ucast_pkts': '3428958706',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49107598',
                                                                       'out_octets': '517966686345',
                                                                       'out_ucast_pkts': '3428958706',
                                                                       'poller_hash': 'NDJEMkM4cGVlcmludGZwb24gNDE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '415',
                                                                                               'links': [       {       'connected_time': '21886552',
                                                                                                                        'fiber_length': 23245,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMkM4cGVlcnBvbnBvbiA0MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 Berdux Pl Wellesley (Bruce Rohr '
                                                 '10434970)',
                                         'poller_hash': 'NDJEMkM4',
                                         'serial_number': '42D2C8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:0b:82',
                                         'uptime': '21886552'},
                                 {       'counter_source': '416/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:2d:34'],
                                                                       'id': '416/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '86058148593',
                                                                       'in_ucast_pkts': '491071850',
                                                                       'mac_address': 'cc:be:59:cb:6a:25',
                                                                       'mtu': 2000,
                                                                       'name': '416/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797994',
                                                                       'out_octets': '2280726982050',
                                                                       'out_ucast_pkts': '1680265408',
                                                                       'poller_hash': 'NDJCRTlDcGVlcmludGY0MTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '416',
                                                                       'in_nucast_pkts': '9797994',
                                                                       'in_octets': '2280726982050',
                                                                       'in_ucast_pkts': '1680265408',
                                                                       'links': [       {       'in_nucast_pkts': '9797994',
                                                                                                'in_octets': '2280726982050',
                                                                                                'in_ucast_pkts': '1680265408',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '86058148593',
                                                                                                'out_ucast_pkts': '491071850',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '86058148593',
                                                                       'out_ucast_pkts': '491071850',
                                                                       'poller_hash': 'NDJCRTlDcGVlcmludGZwb24gNDE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '416',
                                                                                               'links': [       {       'connected_time': '8535226',
                                                                                                                        'fiber_length': 24256,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRTlDcGVlcnBvbnBvbiA0MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 Campbell Pl Wellesley (David '
                                                 'Redecop 10435159)',
                                         'poller_hash': 'NDJCRTlD',
                                         'serial_number': '42BE9C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6a:22',
                                         'uptime': '8535226'},
                                 {       'counter_source': '417/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e4:61'],
                                                                       'id': '417/g1',
                                                                       'in_nucast_pkts': '147816',
                                                                       'in_octets': '156746241590',
                                                                       'in_ucast_pkts': '495004379',
                                                                       'mac_address': '84:d3:43:7d:4d:4a',
                                                                       'mtu': 2000,
                                                                       'name': '417/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797846',
                                                                       'out_octets': '3149479908137',
                                                                       'out_ucast_pkts': '2427828354',
                                                                       'poller_hash': 'QzgxRjEycGVlcmludGY0MTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '417',
                                                                       'in_nucast_pkts': '9797846',
                                                                       'in_octets': '3149479908137',
                                                                       'in_ucast_pkts': '2427828354',
                                                                       'links': [       {       'in_nucast_pkts': '9797846',
                                                                                                'in_octets': '3149479908137',
                                                                                                'in_ucast_pkts': '2427828354',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '147816',
                                                                                                'out_octets': '156746241590',
                                                                                                'out_ucast_pkts': '495004379',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '147816',
                                                                       'out_octets': '156746241590',
                                                                       'out_ucast_pkts': '495004379',
                                                                       'poller_hash': 'QzgxRjEycGVlcmludGZwb24gNDE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '417',
                                                                                               'links': [       {       'connected_time': '8535171',
                                                                                                                        'fiber_length': 23392,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRjEycGVlcnBvbnBvbiA0MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '126 Ferris Dr Wellesley (Tara Proper '
                                                 '10434940)',
                                         'poller_hash': 'QzgxRjEy',
                                         'serial_number': 'C81F12',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4d:47',
                                         'uptime': '8535171'},
                                 {       'counter_source': '419/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:72:7d:df'],
                                                                       'id': '419/g1',
                                                                       'in_nucast_pkts': '332345',
                                                                       'in_octets': '590589083222',
                                                                       'in_ucast_pkts': '3036545916',
                                                                       'mac_address': 'cc:be:59:cb:f9:c5',
                                                                       'mtu': 2000,
                                                                       'name': '419/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '761107233',
                                                                       'out_octets': '10721732486915',
                                                                       'out_ucast_pkts': '6873261514',
                                                                       'poller_hash': 'NDJEMDkwcGVlcmludGY0MTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '419',
                                                                       'in_nucast_pkts': '761107233',
                                                                       'in_octets': '10721732486915',
                                                                       'in_ucast_pkts': '6873261514',
                                                                       'links': [       {       'in_nucast_pkts': '761107233',
                                                                                                'in_octets': '10721732486915',
                                                                                                'in_ucast_pkts': '6873261514',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '332345',
                                                                                                'out_octets': '590589083222',
                                                                                                'out_ucast_pkts': '3036545916',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '332345',
                                                                       'out_octets': '590589083222',
                                                                       'out_ucast_pkts': '3036545916',
                                                                       'poller_hash': 'NDJEMDkwcGVlcmludGZwb24gNDE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '419',
                                                                                               'links': [       {       'connected_time': '13281075',
                                                                                                                        'fiber_length': 22524,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMDkwcGVlcnBvbnBvbiA0MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '78 Pond View Dr Wellesley (Paul '
                                                 'Perkins 10434981)',
                                         'poller_hash': 'NDJEMDkw',
                                         'serial_number': '42D090',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:f9:c2',
                                         'uptime': '13281075'},
                                 {       'counter_source': '420/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0f:81'],
                                                                       'id': '420/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '15216400074',
                                                                       'in_ucast_pkts': '147997824',
                                                                       'mac_address': 'cc:be:59:cb:bd:0d',
                                                                       'mtu': 2000,
                                                                       'name': '420/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5373005',
                                                                       'out_octets': '712592265621',
                                                                       'out_ucast_pkts': '485482306',
                                                                       'poller_hash': 'NDJDOEY5cGVlcmludGY0MjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '420',
                                                                       'in_nucast_pkts': '5373005',
                                                                       'in_octets': '712592265621',
                                                                       'in_ucast_pkts': '485482306',
                                                                       'links': [       {       'in_nucast_pkts': '5373005',
                                                                                                'in_octets': '712592265621',
                                                                                                'in_ucast_pkts': '485482306',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '15216400074',
                                                                                                'out_ucast_pkts': '147997824',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '15216400074',
                                                                       'out_ucast_pkts': '147997824',
                                                                       'poller_hash': 'NDJDOEY5cGVlcmludGZwb24gNDIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '420',
                                                                                               'links': [       {       'connected_time': '4584004',
                                                                                                                        'fiber_length': 24487,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOEY5cGVlcnBvbnBvbiA0MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1380 Greenwood Hill Rd Wellesley '
                                                 '(Michael Lorentz 10435027)',
                                         'poller_hash': 'NDJDOEY5',
                                         'serial_number': '42C8F9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bd:0a',
                                         'uptime': '4584004'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '421/g1',
                                                                       'name': '421/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQjc0cGVlcmludGY0MjEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '421',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NjQwQjc0cGVlcmludGZwb24gNDIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '421',
                                                                                               'links': [       {       'connected_time': '8535140',
                                                                                                                        'fiber_length': 24813,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.658,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.074}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjc0cGVlcnBvbnBvbiA0MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '13 VILLAGE RD',
                                         'poller_hash': 'NjQwQjc0',
                                         'serial_number': '640B74',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:f8',
                                         'uptime': '8535140'},
                                 {       'counter_source': '422/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:72:7e:0c'],
                                                                       'id': '422/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '137756127555',
                                                                       'in_ucast_pkts': '648821837',
                                                                       'mac_address': '44:65:7f:33:22:07',
                                                                       'mtu': 2000,
                                                                       'name': '422/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530414',
                                                                       'out_octets': '2372935552957',
                                                                       'out_ucast_pkts': '1785046835',
                                                                       'poller_hash': 'NjQwQjdBcGVlcmludGY0MjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '422',
                                                                       'in_nucast_pkts': '5530414',
                                                                       'in_octets': '2372935552957',
                                                                       'in_ucast_pkts': '1785046835',
                                                                       'links': [       {       'in_nucast_pkts': '5530414',
                                                                                                'in_octets': '2372935552957',
                                                                                                'in_ucast_pkts': '1785046835',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '137756127555',
                                                                                                'out_ucast_pkts': '648821837',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '137756127555',
                                                                       'out_ucast_pkts': '648821837',
                                                                       'poller_hash': 'NjQwQjdBcGVlcmludGZwb24gNDIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '422',
                                                                                               'links': [       {       'connected_time': '8535108',
                                                                                                                        'fiber_length': 23323,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.618,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.38}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjdBcGVlcnBvbnBvbiA0MjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '11 Reiner Cr Wellesley (Andrew Scott '
                                                 '10434900)',
                                         'poller_hash': 'NjQwQjdB',
                                         'serial_number': '640B7A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:22:06',
                                         'uptime': '8535108'},
                                 {       'counter_source': '423/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '74:b6:b6:00:36:a0'],
                                                                       'id': '423/g1',
                                                                       'in_nucast_pkts': '56438',
                                                                       'in_octets': '27793109446',
                                                                       'in_ucast_pkts': '112063475',
                                                                       'mac_address': 'cc:be:59:cb:bc:35',
                                                                       'mtu': 2000,
                                                                       'name': '423/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1989091',
                                                                       'out_octets': '418017435811',
                                                                       'out_ucast_pkts': '313473004',
                                                                       'poller_hash': 'NDJDOERFcGVlcmludGY0MjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '423',
                                                                       'in_nucast_pkts': '1989091',
                                                                       'in_octets': '418017435811',
                                                                       'in_ucast_pkts': '313473004',
                                                                       'links': [       {       'in_nucast_pkts': '1989091',
                                                                                                'in_octets': '418017435811',
                                                                                                'in_ucast_pkts': '313473004',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '56438',
                                                                                                'out_octets': '27793109446',
                                                                                                'out_ucast_pkts': '112063475',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '56438',
                                                                       'out_octets': '27793109446',
                                                                       'out_ucast_pkts': '112063475',
                                                                       'poller_hash': 'NDJDOERFcGVlcmludGZwb24gNDIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '423',
                                                                                               'links': [       {       'connected_time': '1693604',
                                                                                                                        'fiber_length': 23100,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOERFcGVlcnBvbnBvbiA0MjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3600 Nafziger Rd Wellesley (Daniel '
                                                 'Predhomme 10435026)',
                                         'poller_hash': 'NDJDOERF',
                                         'serial_number': '42C8DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bc:32',
                                         'uptime': '1693604'},
                                 {       'counter_source': '424/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e0:a4'],
                                                                       'id': '424/g1',
                                                                       'in_nucast_pkts': '158168',
                                                                       'in_octets': '163308563220',
                                                                       'in_ucast_pkts': '1200082490',
                                                                       'mac_address': 'cc:be:59:cb:bb:4d',
                                                                       'mtu': 2000,
                                                                       'name': '424/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10117648',
                                                                       'out_octets': '4359800329038',
                                                                       'out_ucast_pkts': '3098234753',
                                                                       'poller_hash': 'NDJDOEMxcGVlcmludGY0MjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '424',
                                                                       'in_nucast_pkts': '10117648',
                                                                       'in_octets': '4359800329038',
                                                                       'in_ucast_pkts': '3098234753',
                                                                       'links': [       {       'in_nucast_pkts': '10117648',
                                                                                                'in_octets': '4359800329038',
                                                                                                'in_ucast_pkts': '3098234753',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '158168',
                                                                                                'out_octets': '163308563220',
                                                                                                'out_ucast_pkts': '1200082490',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '158168',
                                                                       'out_octets': '163308563220',
                                                                       'out_ucast_pkts': '1200082490',
                                                                       'poller_hash': 'NDJDOEMxcGVlcmludGZwb24gNDI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '424',
                                                                                               'links': [       {       'connected_time': '8535293',
                                                                                                                        'fiber_length': 24773,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOEMxcGVlcnBvbnBvbiA0MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 Edgewood Ct Wellelsey (Connie '
                                                 'Finlay 10434980)',
                                         'poller_hash': 'NDJDOEMx',
                                         'serial_number': '42C8C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bb:4a',
                                         'uptime': '8535293'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '425/g1',
                                                                       'name': '425/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDRDdCcGVlcmludGY0MjUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '425',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJDRDdCcGVlcmludGZwb24gNDI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '425',
                                                                                               'links': [       {       'connected_time': '1018160',
                                                                                                                        'fiber_length': 24541,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRDdCcGVlcnBvbnBvbiA0MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 FIRELLA PL (ROUBLE MARK)',
                                         'poller_hash': 'NDJDRDdC',
                                         'serial_number': '42CD7B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e1:1a',
                                         'uptime': '1018160'},
                                 {       'counter_source': '426/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b4:75:0e:4c:93:28'],
                                                                       'id': '426/g1',
                                                                       'in_nucast_pkts': '10142',
                                                                       'in_octets': '116428282081',
                                                                       'in_ucast_pkts': '672618830',
                                                                       'mac_address': '44:65:7f:33:22:0d',
                                                                       'mtu': 2000,
                                                                       'name': '426/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530434',
                                                                       'out_octets': '2257224188737',
                                                                       'out_ucast_pkts': '1652284089',
                                                                       'poller_hash': 'NjQwQjdEcGVlcmludGY0MjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '426',
                                                                       'in_nucast_pkts': '5530434',
                                                                       'in_octets': '2257224188737',
                                                                       'in_ucast_pkts': '1652284089',
                                                                       'links': [       {       'in_nucast_pkts': '5530434',
                                                                                                'in_octets': '2257224188737',
                                                                                                'in_ucast_pkts': '1652284089',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10142',
                                                                                                'out_octets': '116428282081',
                                                                                                'out_ucast_pkts': '672618830',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10142',
                                                                       'out_octets': '116428282081',
                                                                       'out_ucast_pkts': '672618830',
                                                                       'poller_hash': 'NjQwQjdEcGVlcmludGZwb24gNDI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '426',
                                                                                               'links': [       {       'connected_time': '8535117',
                                                                                                                        'fiber_length': 24708,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.566,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.274}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjdEcGVlcnBvbnBvbiA0MjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '105 Brown St Wellesley (Beverly '
                                                 'Meyer 10435259)',
                                         'poller_hash': 'NjQwQjdE',
                                         'serial_number': '640B7D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:22:0c',
                                         'uptime': '8535117'},
                                 {       'counter_source': '429/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:f9'],
                                                                       'id': '429/g1',
                                                                       'in_nucast_pkts': '22',
                                                                       'in_octets': '200970507143',
                                                                       'in_ucast_pkts': '1512392458',
                                                                       'mac_address': 'cc:be:59:cb:e8:2d',
                                                                       'mtu': 2000,
                                                                       'name': '429/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919768',
                                                                       'out_octets': '5433905626616',
                                                                       'out_ucast_pkts': '3816439181',
                                                                       'poller_hash': 'NDJDRTVEcGVlcmludGY0MjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '429',
                                                                       'in_nucast_pkts': '9919768',
                                                                       'in_octets': '5433905626616',
                                                                       'in_ucast_pkts': '3816439181',
                                                                       'links': [       {       'in_nucast_pkts': '9919768',
                                                                                                'in_octets': '5433905626616',
                                                                                                'in_ucast_pkts': '3816439181',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22',
                                                                                                'out_octets': '200970507143',
                                                                                                'out_ucast_pkts': '1512392458',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22',
                                                                       'out_octets': '200970507143',
                                                                       'out_ucast_pkts': '1512392458',
                                                                       'poller_hash': 'NDJDRTVEcGVlcmludGZwb24gNDI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '429',
                                                                                               'links': [       {       'connected_time': '8535270',
                                                                                                                        'fiber_length': 24691,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTVEcGVlcnBvbnBvbiA0Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '17 FIRELLA PL (AUSTIN KARLIE)',
                                         'poller_hash': 'NDJDRTVE',
                                         'serial_number': '42CE5D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e8:2a',
                                         'uptime': '8535270'},
                                 {       'counter_source': '430/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '38:86:f7:fd:46:8d'],
                                                                       'id': '430/g1',
                                                                       'in_nucast_pkts': '5300',
                                                                       'in_octets': '649030630552',
                                                                       'in_ucast_pkts': '807873636',
                                                                       'mac_address': 'cc:be:59:50:02:da',
                                                                       'mtu': 2000,
                                                                       'name': '430/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5301293',
                                                                       'out_octets': '3815881492979',
                                                                       'out_ucast_pkts': '2904689831',
                                                                       'poller_hash': 'NDNCNTM0cGVlcmludGY0MzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '430',
                                                                       'in_nucast_pkts': '5301293',
                                                                       'in_octets': '3815881492979',
                                                                       'in_ucast_pkts': '2904689831',
                                                                       'links': [       {       'in_nucast_pkts': '5301293',
                                                                                                'in_octets': '3815881492979',
                                                                                                'in_ucast_pkts': '2904689831',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5300',
                                                                                                'out_octets': '649030630552',
                                                                                                'out_ucast_pkts': '807873636',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5300',
                                                                       'out_octets': '649030630552',
                                                                       'out_ucast_pkts': '807873636',
                                                                       'poller_hash': 'NDNCNTM0cGVlcmludGZwb24gNDMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '430',
                                                                                               'links': [       {       'connected_time': '7997672',
                                                                                                                        'fiber_length': 24716,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.856,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.7}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNTM0cGVlcnBvbnBvbiA0MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '67 NITHBANK RD (PRSA STEVEN)',
                                         'poller_hash': 'NDNCNTM0',
                                         'serial_number': '43B534',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:02:d9',
                                         'uptime': '7997672'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '431/g1',
                                                                       'name': '431/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEY4QzlCcGVlcmludGY0MzEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '431',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEY4QzlCcGVlcmludGZwb24gNDMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '431',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEY4QzlCcGVlcnBvbnBvbiA0MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '26 Gerber Meadows Dr Wellesley',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "431" '
                                                                  '(location: GPON '
                                                                  'port "7/5"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:27:16.660Z',
                                         'poller_hash': 'NEY4QzlC',
                                         'serial_number': '4F8C9B',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '433/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '433/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1211992282862',
                                                                       'in_ucast_pkts': '51635139424',
                                                                       'mac_address': 'cc:be:59:cb:e9:d2',
                                                                       'name': '433/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '14968647035742',
                                                                       'out_ucast_pkts': '11629311352',
                                                                       'poller_hash': 'NDJDRTkycGVlcmludGY0MzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '433',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '14968647035742',
                                                                       'in_ucast_pkts': '11629311352',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '14968647035742',
                                                                                                'in_ucast_pkts': '11629311352',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1211992282862',
                                                                                                'out_ucast_pkts': '51635139424',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1211992282862',
                                                                       'out_ucast_pkts': '51635139424',
                                                                       'poller_hash': 'NDJDRTkycGVlcmludGZwb24gNDMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '433',
                                                                                               'links': [       {       'connected_time': '7383395',
                                                                                                                        'fiber_length': 23562,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTkycGVlcnBvbnBvbiA0MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '50 ZINKANN CR (BILODEAU MAURICE & '
                                                 'MONICA)',
                                         'poller_hash': 'NDJDRTky',
                                         'serial_number': '42CE92',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e9:d2',
                                         'uptime': '7383395'},
                                 {       'counter_source': '434/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:38:e0:0f:b6:99'],
                                                                       'id': '434/g1',
                                                                       'in_nucast_pkts': '278',
                                                                       'in_octets': '37540621',
                                                                       'in_ucast_pkts': '185839',
                                                                       'mac_address': 'cc:be:59:50:04:40',
                                                                       'mtu': 2000,
                                                                       'name': '434/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11536',
                                                                       'out_octets': '422157049',
                                                                       'out_ucast_pkts': '378424',
                                                                       'poller_hash': 'NDNCNUU3cGVlcmludGY0MzQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '434',
                                                                       'in_nucast_pkts': '11536',
                                                                       'in_octets': '422157049',
                                                                       'in_ucast_pkts': '378424',
                                                                       'links': [       {       'in_nucast_pkts': '11536',
                                                                                                'in_octets': '422157049',
                                                                                                'in_ucast_pkts': '378424',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '278',
                                                                                                'out_octets': '37540621',
                                                                                                'out_ucast_pkts': '185839',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '278',
                                                                       'out_octets': '37540621',
                                                                       'out_ucast_pkts': '185839',
                                                                       'poller_hash': 'NDNCNUU3cGVlcmludGZwb24gNDM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '434',
                                                                                               'links': [       {       'connected_time': '33279',
                                                                                                                        'fiber_length': 23041,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.39,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.146}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNUU3cGVlcnBvbnBvbiA0MzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '3603 Nafziger Rd Wellesley (Richard '
                                                 'Nafziger 10434874)',
                                         'poller_hash': 'NDNCNUU3',
                                         'serial_number': '43B5E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:04:3f',
                                         'uptime': '33279'},
                                 {       'counter_source': '437/F1+437/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:b0:e6'],
                                                                       'id': '437/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1417010320',
                                                                       'in_ucast_pkts': '7038451',
                                                                       'name': '437/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '11404723438',
                                                                       'out_octets': '15210869803376',
                                                                       'out_ucast_pkts': '8292489',
                                                                       'poller_hash': 'NDJEMDZFcGVlcmludGY0MzcvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:41'],
                                                                       'id': '437/g1',
                                                                       'in_nucast_pkts': '21',
                                                                       'in_octets': '60494362291',
                                                                       'in_ucast_pkts': '276569751',
                                                                       'mac_address': 'cc:be:59:cb:f8:b5',
                                                                       'mtu': 2000,
                                                                       'name': '437/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920149',
                                                                       'out_octets': '1814033320902',
                                                                       'out_ucast_pkts': '1270981865',
                                                                       'poller_hash': 'NDJEMDZFcGVlcmludGY0MzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '437',
                                                                       'in_nucast_pkts': '11414643587',
                                                                       'in_octets': '17024903124278',
                                                                       'in_ucast_pkts': '1279274354',
                                                                       'links': [       {       'in_nucast_pkts': '11414643587',
                                                                                                'in_octets': '17024903124278',
                                                                                                'in_ucast_pkts': '1279274354',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21',
                                                                                                'out_octets': '61911372611',
                                                                                                'out_ucast_pkts': '283608202',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21',
                                                                       'out_octets': '61911372611',
                                                                       'out_ucast_pkts': '283608202',
                                                                       'poller_hash': 'NDJEMDZFcGVlcmludGZwb24gNDM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '437',
                                                                                               'links': [       {       'connected_time': '8535265',
                                                                                                                        'fiber_length': 22528,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMDZFcGVlcnBvbnBvbiA0Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1169 Henry St Wellesley (Derek Brick '
                                                 '10435296)',
                                         'poller_hash': 'NDJEMDZF',
                                         'serial_number': '42D06E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:f8:b2',
                                         'uptime': '8535265'},
                                 {       'counter_source': '438/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b9:65'],
                                                                       'id': '438/g1',
                                                                       'in_nucast_pkts': '153815',
                                                                       'in_octets': '136248610904',
                                                                       'in_ucast_pkts': '239703206',
                                                                       'mac_address': 'cc:be:59:cb:f9:55',
                                                                       'mtu': 2000,
                                                                       'name': '438/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919772',
                                                                       'out_octets': '1634086026013',
                                                                       'out_ucast_pkts': '1218883657',
                                                                       'poller_hash': 'NDJEMDgycGVlcmludGY0MzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '438',
                                                                       'in_nucast_pkts': '9919772',
                                                                       'in_octets': '1634086026013',
                                                                       'in_ucast_pkts': '1218883657',
                                                                       'links': [       {       'in_nucast_pkts': '9919772',
                                                                                                'in_octets': '1634086026013',
                                                                                                'in_ucast_pkts': '1218883657',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '153815',
                                                                                                'out_octets': '136248610904',
                                                                                                'out_ucast_pkts': '239703206',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '153815',
                                                                       'out_octets': '136248610904',
                                                                       'out_ucast_pkts': '239703206',
                                                                       'poller_hash': 'NDJEMDgycGVlcmludGZwb24gNDM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '438',
                                                                                               'links': [       {       'connected_time': '8535231',
                                                                                                                        'fiber_length': 24512,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMDgycGVlcnBvbnBvbiA0Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '102 NITHBANK RD (HENNIGE JULIA)',
                                         'poller_hash': 'NDJEMDgy',
                                         'serial_number': '42D082',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:f9:52',
                                         'uptime': '8535231'},
                                 {       'counter_source': '439/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:fe'],
                                                                       'id': '439/g1',
                                                                       'in_nucast_pkts': '181',
                                                                       'in_octets': '23860531848',
                                                                       'in_ucast_pkts': '222715740',
                                                                       'mac_address': 'cc:be:59:cb:83:ed',
                                                                       'mtu': 2000,
                                                                       'name': '439/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13227391',
                                                                       'out_octets': '1279780006649',
                                                                       'out_ucast_pkts': '864212088',
                                                                       'poller_hash': 'NDJDMUQ1cGVlcmludGY0MzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '439',
                                                                       'in_nucast_pkts': '13227391',
                                                                       'in_octets': '1279780006649',
                                                                       'in_ucast_pkts': '864212088',
                                                                       'links': [       {       'in_nucast_pkts': '13227391',
                                                                                                'in_octets': '1279780006649',
                                                                                                'in_ucast_pkts': '864212088',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '181',
                                                                                                'out_octets': '23860531848',
                                                                                                'out_ucast_pkts': '222715740',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '181',
                                                                       'out_octets': '23860531848',
                                                                       'out_ucast_pkts': '222715740',
                                                                       'poller_hash': 'NDJDMUQ1cGVlcmludGZwb24gNDM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '439',
                                                                                               'links': [       {       'connected_time': '11397939',
                                                                                                                        'fiber_length': 23101,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMUQ1cGVlcnBvbnBvbiA0Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '171 FERRIS DR (MERLAU NELMA J)',
                                         'poller_hash': 'NDJDMUQ1',
                                         'serial_number': '42C1D5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:83:ea',
                                         'uptime': '11397939'},
                                 {       'counter_source': '441/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:4e'],
                                                                       'id': '441/g1',
                                                                       'in_nucast_pkts': '18602122',
                                                                       'in_octets': '69111849648',
                                                                       'in_ucast_pkts': '360544025',
                                                                       'mac_address': 'cc:be:59:cb:6d:fd',
                                                                       'mtu': 2000,
                                                                       'name': '441/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '463539536',
                                                                       'out_octets': '2534568400090',
                                                                       'out_ucast_pkts': '1444009534',
                                                                       'poller_hash': 'NDJCRjE3cGVlcmludGY0NDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '441',
                                                                       'in_nucast_pkts': '463539536',
                                                                       'in_octets': '2534568400090',
                                                                       'in_ucast_pkts': '1444009534',
                                                                       'links': [       {       'in_nucast_pkts': '463539536',
                                                                                                'in_octets': '2534568400090',
                                                                                                'in_ucast_pkts': '1444009534',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18602122',
                                                                                                'out_octets': '69111849648',
                                                                                                'out_ucast_pkts': '360544025',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18602122',
                                                                       'out_octets': '69111849648',
                                                                       'out_ucast_pkts': '360544025',
                                                                       'poller_hash': 'NDJCRjE3cGVlcmludGZwb24gNDQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '441',
                                                                                               'links': [       {       'connected_time': '8667364',
                                                                                                                        'fiber_length': 23026,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjE3cGVlcnBvbnBvbiA0NDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '182 Ferris Dr (Diane Wakem 10435238)',
                                         'poller_hash': 'NDJCRjE3',
                                         'serial_number': '42BF17',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6d:fa',
                                         'uptime': '8667364'},
                                 {       'counter_source': '442/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '442/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '154',
                                                                       'in_octets': '162915403927',
                                                                       'in_ucast_pkts': '1058626955',
                                                                       'mac_address': 'cc:be:59:cb:6e:52',
                                                                       'name': '442/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '790',
                                                                       'out_octets': '5466292032678',
                                                                       'out_ucast_pkts': '3923999564',
                                                                       'poller_hash': 'NDJCRjIycGVlcmludGY0NDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '442',
                                                                       'in_nucast_pkts': '790',
                                                                       'in_octets': '5466292032678',
                                                                       'in_ucast_pkts': '3923999564',
                                                                       'links': [       {       'in_nucast_pkts': '790',
                                                                                                'in_octets': '5466292032678',
                                                                                                'in_ucast_pkts': '3923999564',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '154',
                                                                                                'out_octets': '162915403927',
                                                                                                'out_ucast_pkts': '1058626955',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '154',
                                                                       'out_octets': '162915403927',
                                                                       'out_ucast_pkts': '1058626955',
                                                                       'poller_hash': 'NDJCRjIycGVlcmludGZwb24gNDQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '442',
                                                                                               'links': [       {       'connected_time': '8535186',
                                                                                                                        'fiber_length': 23663,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjIycGVlcnBvbnBvbiA0NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '56 ZINKANN CR (PARKINSON HOLLY)',
                                         'poller_hash': 'NDJCRjIy',
                                         'serial_number': '42BF22',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6e:52',
                                         'uptime': '8535186'},
                                 {       'counter_source': '443/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:2f'],
                                                                       'id': '443/g1',
                                                                       'in_nucast_pkts': '34',
                                                                       'in_octets': '114337537081',
                                                                       'in_ucast_pkts': '750004277',
                                                                       'mac_address': 'cc:be:59:cb:6d:cd',
                                                                       'mtu': 2000,
                                                                       'name': '443/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29470200',
                                                                       'out_octets': '4335874206543',
                                                                       'out_ucast_pkts': '2926787992',
                                                                       'poller_hash': 'NDJCRjExcGVlcmludGY0NDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '443',
                                                                       'in_nucast_pkts': '29470200',
                                                                       'in_octets': '4335874206543',
                                                                       'in_ucast_pkts': '2926787992',
                                                                       'links': [       {       'in_nucast_pkts': '29470200',
                                                                                                'in_octets': '4335874206543',
                                                                                                'in_ucast_pkts': '2926787992',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '34',
                                                                                                'out_octets': '114337537081',
                                                                                                'out_ucast_pkts': '750004277',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '34',
                                                                       'out_octets': '114337537081',
                                                                       'out_ucast_pkts': '750004277',
                                                                       'poller_hash': 'NDJCRjExcGVlcmludGZwb24gNDQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '443',
                                                                                               'links': [       {       'connected_time': '25498581',
                                                                                                                        'fiber_length': 22900,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjExcGVlcnBvbnBvbiA0NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '24 Pondview Dr (Eva Irwin, 10434918)',
                                         'poller_hash': 'NDJCRjEx',
                                         'serial_number': '42BF11',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6d:ca',
                                         'uptime': '25498581'},
                                 {       'counter_source': '444/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:09:48'],
                                                                       'id': '444/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '137084105653',
                                                                       'in_ucast_pkts': '804791914',
                                                                       'mac_address': 'cc:be:59:50:00:fa',
                                                                       'mtu': 2000,
                                                                       'name': '444/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530424',
                                                                       'out_octets': '4567356494581',
                                                                       'out_ucast_pkts': '3288538621',
                                                                       'poller_hash': 'NDNCNDQ1cGVlcmludGY0NDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '444',
                                                                       'in_nucast_pkts': '5530424',
                                                                       'in_octets': '4567356494581',
                                                                       'in_ucast_pkts': '3288538621',
                                                                       'links': [       {       'in_nucast_pkts': '5530424',
                                                                                                'in_octets': '4567356494581',
                                                                                                'in_ucast_pkts': '3288538621',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '137084105653',
                                                                                                'out_ucast_pkts': '804791914',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '137084105653',
                                                                       'out_ucast_pkts': '804791914',
                                                                       'poller_hash': 'NDNCNDQ1cGVlcmludGZwb24gNDQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '444',
                                                                                               'links': [       {       'connected_time': '8535113',
                                                                                                                        'fiber_length': 23132,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.846,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.426}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNDQ1cGVlcnBvbnBvbiA0NDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '16 Zinkann Cr Wellesley (Brigitte '
                                                 'Belcourt 10434818)',
                                         'poller_hash': 'NDNCNDQ1',
                                         'serial_number': '43B445',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:00:f9',
                                         'uptime': '8535113'},
                                 {       'counter_source': '445/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:9a:4a:a3',
                                                                                                        '00:03:e6:eb:46:da',
                                                                                                        'd0:76:8f:1f:57:d5'],
                                                                       'id': '445/g1',
                                                                       'in_nucast_pkts': '7214920',
                                                                       'in_octets': '132971512733',
                                                                       'in_ucast_pkts': '575178936',
                                                                       'mac_address': 'cc:be:59:cb:6e:35',
                                                                       'mtu': 2000,
                                                                       'name': '445/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10711193821',
                                                                       'out_octets': '16977517348736',
                                                                       'out_ucast_pkts': '1822475515',
                                                                       'poller_hash': 'NDJCRjFFcGVlcmludGY0NDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '445',
                                                                       'in_nucast_pkts': '10711193821',
                                                                       'in_octets': '16977517348736',
                                                                       'in_ucast_pkts': '1822475515',
                                                                       'links': [       {       'in_nucast_pkts': '10711193821',
                                                                                                'in_octets': '16977517348736',
                                                                                                'in_ucast_pkts': '1822475515',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7214920',
                                                                                                'out_octets': '132971512733',
                                                                                                'out_ucast_pkts': '575178936',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7214920',
                                                                       'out_octets': '132971512733',
                                                                       'out_ucast_pkts': '575178936',
                                                                       'poller_hash': 'NDJCRjFFcGVlcmludGZwb24gNDQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '445',
                                                                                               'links': [       {       'connected_time': '5886982',
                                                                                                                        'fiber_length': 24806,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjFFcGVlcnBvbnBvbiA0NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1005 CATHERINE ST (GARTLAND CASSADI)',
                                         'poller_hash': 'NDJCRjFF',
                                         'serial_number': '42BF1E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6e:32',
                                         'uptime': '5886982'},
                                 {       'counter_source': '446/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:d6'],
                                                                       'id': '446/g1',
                                                                       'in_nucast_pkts': '111998',
                                                                       'in_octets': '729713521839',
                                                                       'in_ucast_pkts': '2589412157',
                                                                       'mac_address': 'cc:be:59:cb:6d:ed',
                                                                       'mtu': 2000,
                                                                       'name': '446/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39607851',
                                                                       'out_octets': '11460497878069',
                                                                       'out_ucast_pkts': '8294976598',
                                                                       'poller_hash': 'NDJCRjE1cGVlcmludGY0NDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '446',
                                                                       'in_nucast_pkts': '39607851',
                                                                       'in_octets': '11460497878069',
                                                                       'in_ucast_pkts': '8294976598',
                                                                       'links': [       {       'in_nucast_pkts': '39607851',
                                                                                                'in_octets': '11460497878069',
                                                                                                'in_ucast_pkts': '8294976598',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '111998',
                                                                                                'out_octets': '729713521839',
                                                                                                'out_ucast_pkts': '2589412157',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '111998',
                                                                       'out_octets': '729713521839',
                                                                       'out_ucast_pkts': '2589412157',
                                                                       'poller_hash': 'NDJCRjE1cGVlcmludGZwb24gNDQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '446',
                                                                                               'links': [       {       'connected_time': '72492405',
                                                                                                                        'fiber_length': 18170,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjE1cGVlcnBvbnBvbiA0NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1969 PERTH LINE 56 (SCHIEDEL '
                                                 'ALLISON)',
                                         'poller_hash': 'NDJCRjE1',
                                         'serial_number': '42BF15',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6d:ea',
                                         'uptime': '72492405'},
                                 {       'counter_source': '449/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:e3'],
                                                                       'id': '449/g1',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '91380112769',
                                                                       'in_ucast_pkts': '666617100',
                                                                       'mac_address': 'cc:be:59:cb:bd:ed',
                                                                       'mtu': 2000,
                                                                       'name': '449/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797967',
                                                                       'out_octets': '2164060312464',
                                                                       'out_ucast_pkts': '1586187745',
                                                                       'poller_hash': 'NDJDOTE1cGVlcmludGY0NDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '449',
                                                                       'in_nucast_pkts': '9797967',
                                                                       'in_octets': '2164060312464',
                                                                       'in_ucast_pkts': '1586187745',
                                                                       'links': [       {       'in_nucast_pkts': '9797967',
                                                                                                'in_octets': '2164060312464',
                                                                                                'in_ucast_pkts': '1586187745',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '91380112769',
                                                                                                'out_ucast_pkts': '666617100',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '91380112769',
                                                                       'out_ucast_pkts': '666617100',
                                                                       'poller_hash': 'NDJDOTE1cGVlcmludGZwb24gNDQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '449',
                                                                                               'links': [       {       'connected_time': '8535310',
                                                                                                                        'fiber_length': 23889,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOTE1cGVlcnBvbnBvbiA0NDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1251 Greenwood Hill Rd Wellesley '
                                                 '(Steve & Ruth Hoffele 10434887)',
                                         'poller_hash': 'NDJDOTE1',
                                         'serial_number': '42C915',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bd:ea',
                                         'uptime': '8535310'},
                                 {       'counter_source': '453/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:72:7d:a9'],
                                                                       'id': '453/g1',
                                                                       'in_nucast_pkts': '717',
                                                                       'in_octets': '774739914353',
                                                                       'in_ucast_pkts': '2029852838',
                                                                       'mac_address': 'cc:be:59:cb:bd:bd',
                                                                       'mtu': 2000,
                                                                       'name': '453/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '42953708',
                                                                       'out_octets': '8905954781741',
                                                                       'out_ucast_pkts': '6620424412',
                                                                       'poller_hash': 'NDJDOTBGcGVlcmludGY0NTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '453',
                                                                       'in_nucast_pkts': '42953708',
                                                                       'in_octets': '8905954781741',
                                                                       'in_ucast_pkts': '6620424412',
                                                                       'links': [       {       'in_nucast_pkts': '42953708',
                                                                                                'in_octets': '8905954781741',
                                                                                                'in_ucast_pkts': '6620424412',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '717',
                                                                                                'out_octets': '774739914353',
                                                                                                'out_ucast_pkts': '2029852838',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '717',
                                                                       'out_octets': '774739914353',
                                                                       'out_ucast_pkts': '2029852838',
                                                                       'poller_hash': 'NDJDOTBGcGVlcmludGZwb24gNDUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '453',
                                                                                               'links': [       {       'connected_time': '73263774',
                                                                                                                        'fiber_length': 24899,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOTBGcGVlcnBvbnBvbiA0NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 Park ViewDr Wellesley (Paul Abell '
                                                 '10434965)',
                                         'poller_hash': 'NDJDOTBG',
                                         'serial_number': '42C90F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bd:ba',
                                         'uptime': '73263774'},
                                 {       'counter_source': '454/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d4:5f'],
                                                                       'id': '454/g1',
                                                                       'in_nucast_pkts': '9620000',
                                                                       'in_octets': '172879699572',
                                                                       'in_ucast_pkts': '551992976',
                                                                       'mac_address': 'cc:be:59:cb:86:05',
                                                                       'mtu': 2000,
                                                                       'name': '454/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797692',
                                                                       'out_octets': '1477300921671',
                                                                       'out_ucast_pkts': '1114185964',
                                                                       'poller_hash': 'NDJDMjE4cGVlcmludGY0NTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '454',
                                                                       'in_nucast_pkts': '9797692',
                                                                       'in_octets': '1477300921671',
                                                                       'in_ucast_pkts': '1114185964',
                                                                       'links': [       {       'in_nucast_pkts': '9797692',
                                                                                                'in_octets': '1477300921671',
                                                                                                'in_ucast_pkts': '1114185964',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9620000',
                                                                                                'out_octets': '172879699572',
                                                                                                'out_ucast_pkts': '551992976',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9620000',
                                                                       'out_octets': '172879699572',
                                                                       'out_ucast_pkts': '551992976',
                                                                       'poller_hash': 'NDJDMjE4cGVlcmludGZwb24gNDU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '454',
                                                                                               'links': [       {       'connected_time': '8535330',
                                                                                                                        'fiber_length': 24834,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMjE4cGVlcnBvbnBvbiA0NTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '110 Gerber Meadows Dr Wellesley '
                                                 '(Alan Bowman 10435225)',
                                         'poller_hash': 'NDJDMjE4',
                                         'serial_number': '42C218',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:86:02',
                                         'uptime': '8535330'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '455/g1',
                                                                       'name': '455/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCRjdCcGVlcmludGY0NTUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '455',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJCRjdCcGVlcmludGZwb24gNDU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '455',
                                                                                               'links': [       {       'connected_time': '8535203',
                                                                                                                        'fiber_length': 23644,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjdCcGVlcnBvbnBvbiA0NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '47 Reiner Cr Wellesley',
                                         'poller_hash': 'NDJCRjdC',
                                         'serial_number': '42BF7B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:71:1a',
                                         'uptime': '8535203'},
                                 {       'counter_source': '456/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:72:7d:3d'],
                                                                       'id': '456/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '68192929767',
                                                                       'in_ucast_pkts': '312899056',
                                                                       'mac_address': 'cc:be:59:cb:98:0d',
                                                                       'mtu': 2000,
                                                                       'name': '456/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797942',
                                                                       'out_octets': '1745112863239',
                                                                       'out_ucast_pkts': '1331293403',
                                                                       'poller_hash': 'NDJDNDU5cGVlcmludGY0NTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '456',
                                                                       'in_nucast_pkts': '9797942',
                                                                       'in_octets': '1745112863239',
                                                                       'in_ucast_pkts': '1331293403',
                                                                       'links': [       {       'in_nucast_pkts': '9797942',
                                                                                                'in_octets': '1745112863239',
                                                                                                'in_ucast_pkts': '1331293403',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '68192929767',
                                                                                                'out_ucast_pkts': '312899056',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '68192929767',
                                                                       'out_ucast_pkts': '312899056',
                                                                       'poller_hash': 'NDJDNDU5cGVlcmludGZwb24gNDU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '456',
                                                                                               'links': [       {       'connected_time': '8535256',
                                                                                                                        'fiber_length': 24293,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNDU5cGVlcnBvbnBvbiA0NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 Campbell Pl (Julie Whitney, '
                                                 '10435162)',
                                         'poller_hash': 'NDJDNDU5',
                                         'serial_number': '42C459',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:98:0a',
                                         'uptime': '8535256'},
                                 {       'counter_source': '457/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '457/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '329',
                                                                       'in_octets': '110736534443',
                                                                       'in_ucast_pkts': '1071260184',
                                                                       'mac_address': 'cc:be:59:cb:94:a2',
                                                                       'name': '457/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '634',
                                                                       'out_octets': '6221938724923',
                                                                       'out_ucast_pkts': '4176102075',
                                                                       'poller_hash': 'NDJDM0VDcGVlcmludGY0NTcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '457',
                                                                       'in_nucast_pkts': '634',
                                                                       'in_octets': '6221938724923',
                                                                       'in_ucast_pkts': '4176102075',
                                                                       'links': [       {       'in_nucast_pkts': '634',
                                                                                                'in_octets': '6221938724923',
                                                                                                'in_ucast_pkts': '4176102075',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '329',
                                                                                                'out_octets': '110736534443',
                                                                                                'out_ucast_pkts': '1071260184',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '329',
                                                                       'out_octets': '110736534443',
                                                                       'out_ucast_pkts': '1071260184',
                                                                       'poller_hash': 'NDJDM0VDcGVlcmludGZwb24gNDU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '457',
                                                                                               'links': [       {       'connected_time': '29644083',
                                                                                                                        'fiber_length': 23757,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDM0VDcGVlcnBvbnBvbiA0NTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '51 Reiner Cr (Ted Richards, '
                                                 '10410746)',
                                         'poller_hash': 'NDJDM0VD',
                                         'serial_number': '42C3EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:94:a2',
                                         'uptime': '29644083'},
                                 {       'counter_source': '458/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:f0'],
                                                                       'id': '458/g1',
                                                                       'in_nucast_pkts': '673',
                                                                       'in_octets': '130013133609',
                                                                       'in_ucast_pkts': '618569053',
                                                                       'mac_address': 'cc:be:59:cb:97:f5',
                                                                       'mtu': 2000,
                                                                       'name': '458/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11888493',
                                                                       'out_octets': '3053821624757',
                                                                       'out_ucast_pkts': '2312905496',
                                                                       'poller_hash': 'NDJDNDU2cGVlcmludGY0NTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '458',
                                                                       'in_nucast_pkts': '11888493',
                                                                       'in_octets': '3053821624757',
                                                                       'in_ucast_pkts': '2312905496',
                                                                       'links': [       {       'in_nucast_pkts': '11888493',
                                                                                                'in_octets': '3053821624757',
                                                                                                'in_ucast_pkts': '2312905496',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '673',
                                                                                                'out_octets': '130013133609',
                                                                                                'out_ucast_pkts': '618569053',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '673',
                                                                       'out_octets': '130013133609',
                                                                       'out_ucast_pkts': '618569053',
                                                                       'poller_hash': 'NDJDNDU2cGVlcmludGZwb24gNDU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '458',
                                                                                               'links': [       {       'connected_time': '10372755',
                                                                                                                        'fiber_length': 23224,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNDU2cGVlcnBvbnBvbiA0NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '17 ZINKANN CR (NAFZIGER HEIDI)',
                                         'poller_hash': 'NDJDNDU2',
                                         'serial_number': '42C456',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:97:f2',
                                         'uptime': '10372755'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '459/g1',
                                                                       'name': '459/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDNDU3cGVlcmludGY0NTkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '459',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJDNDU3cGVlcmludGZwb24gNDU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '459',
                                                                                               'links': [       {       'connected_time': '8535247',
                                                                                                                        'fiber_length': 23737,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNDU3cGVlcnBvbnBvbiA0NTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '42 Schweitzel Cr Wellesley',
                                         'poller_hash': 'NDJDNDU3',
                                         'serial_number': '42C457',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:97:fa',
                                         'uptime': '8535247'},
                                 {       'counter_source': '460/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:b1:10',
                                                                                                        '00:03:e6:da:ba:b5',
                                                                                                        '44:65:7f:72:7d:22'],
                                                                       'id': '460/g1',
                                                                       'in_nucast_pkts': '227564',
                                                                       'in_octets': '6198432952',
                                                                       'in_ucast_pkts': '31170972',
                                                                       'mac_address': 'cc:be:59:cb:6e:dd',
                                                                       'mtu': 2000,
                                                                       'name': '460/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '222406166',
                                                                       'out_octets': '446897503343',
                                                                       'out_ucast_pkts': '108734804',
                                                                       'poller_hash': 'NDJCRjMzcGVlcmludGY0NjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '460',
                                                                       'in_nucast_pkts': '222406166',
                                                                       'in_octets': '446897503343',
                                                                       'in_ucast_pkts': '108734804',
                                                                       'links': [       {       'in_nucast_pkts': '222406166',
                                                                                                'in_octets': '446897503343',
                                                                                                'in_ucast_pkts': '108734804',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '227564',
                                                                                                'out_octets': '6198432952',
                                                                                                'out_ucast_pkts': '31170972',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '227564',
                                                                       'out_octets': '6198432952',
                                                                       'out_ucast_pkts': '31170972',
                                                                       'poller_hash': 'NDJCRjMzcGVlcmludGZwb24gNDYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '460',
                                                                                               'links': [       {       'connected_time': '613208',
                                                                                                                        'fiber_length': 22855,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjMzcGVlcnBvbnBvbiA0NjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '43 Pond View Dr Wellesley (Peter Erb '
                                                 '10434856)',
                                         'poller_hash': 'NDJCRjMz',
                                         'serial_number': '42BF33',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6e:da',
                                         'uptime': '613208'},
                                 {       'counter_source': '461/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:06:3d:f6'],
                                                                       'id': '461/g1',
                                                                       'in_nucast_pkts': '131663',
                                                                       'in_octets': '87660242490',
                                                                       'in_ucast_pkts': '448161439',
                                                                       'mac_address': '44:65:7f:33:21:5f',
                                                                       'mtu': 2000,
                                                                       'name': '461/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4197346',
                                                                       'out_octets': '1715279236571',
                                                                       'out_ucast_pkts': '1266424653',
                                                                       'poller_hash': 'NjQwQjI3cGVlcmludGY0NjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '461',
                                                                       'in_nucast_pkts': '4197346',
                                                                       'in_octets': '1715279236571',
                                                                       'in_ucast_pkts': '1266424653',
                                                                       'links': [       {       'in_nucast_pkts': '4197346',
                                                                                                'in_octets': '1715279236571',
                                                                                                'in_ucast_pkts': '1266424653',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '131663',
                                                                                                'out_octets': '87660242490',
                                                                                                'out_ucast_pkts': '448161439',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '131663',
                                                                       'out_octets': '87660242490',
                                                                       'out_ucast_pkts': '448161439',
                                                                       'poller_hash': 'NjQwQjI3cGVlcmludGZwb24gNDYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '461',
                                                                                               'links': [       {       'connected_time': '6296984',
                                                                                                                        'fiber_length': 22698,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.7,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.744}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjI3cGVlcnBvbnBvbiA0NjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1126 HENRY ST (CALDWELL MICHAEL)',
                                         'poller_hash': 'NjQwQjI3',
                                         'serial_number': '640B27',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:5e',
                                         'uptime': '6296984'},
                                 {       'counter_source': '462/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d4:4d'],
                                                                       'id': '462/g1',
                                                                       'in_nucast_pkts': '101',
                                                                       'in_octets': '236085084668',
                                                                       'in_ucast_pkts': '1654527553',
                                                                       'mac_address': 'cc:be:59:cb:68:fd',
                                                                       'mtu': 2000,
                                                                       'name': '462/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797749',
                                                                       'out_octets': '5425629628858',
                                                                       'out_ucast_pkts': '3802461611',
                                                                       'poller_hash': 'NDJCRTc3cGVlcmludGY0NjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '462',
                                                                       'in_nucast_pkts': '9797749',
                                                                       'in_octets': '5425629628858',
                                                                       'in_ucast_pkts': '3802461611',
                                                                       'links': [       {       'in_nucast_pkts': '9797749',
                                                                                                'in_octets': '5425629628858',
                                                                                                'in_ucast_pkts': '3802461611',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '101',
                                                                                                'out_octets': '236085084668',
                                                                                                'out_ucast_pkts': '1654527553',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '101',
                                                                       'out_octets': '236085084668',
                                                                       'out_ucast_pkts': '1654527553',
                                                                       'poller_hash': 'NDJCRTc3cGVlcmludGZwb24gNDYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '462',
                                                                                               'links': [       {       'connected_time': '8535246',
                                                                                                                        'fiber_length': 23713,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRTc3cGVlcnBvbnBvbiA0NjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '64 Zinkann Cr Wellesley (E & M '
                                                 'Krueger 10435337)',
                                         'poller_hash': 'NDJCRTc3',
                                         'serial_number': '42BE77',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:68:fa',
                                         'uptime': '8535246'},
                                 {       'counter_source': '463/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:05:88:22:85:78'],
                                                                       'id': '463/g1',
                                                                       'in_nucast_pkts': '6537',
                                                                       'in_octets': '73168032882',
                                                                       'in_ucast_pkts': '160304247',
                                                                       'mac_address': '44:65:7f:7b:d3:cd',
                                                                       'mtu': 2000,
                                                                       'name': '463/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1436207',
                                                                       'out_octets': '548737053497',
                                                                       'out_ucast_pkts': '436322184',
                                                                       'poller_hash': 'NUE2QjhBcGVlcmludGY0NjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '463',
                                                                       'in_nucast_pkts': '1436207',
                                                                       'in_octets': '548737053497',
                                                                       'in_ucast_pkts': '436322184',
                                                                       'links': [       {       'in_nucast_pkts': '1436207',
                                                                                                'in_octets': '548737053497',
                                                                                                'in_ucast_pkts': '436322184',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6537',
                                                                                                'out_octets': '73168032882',
                                                                                                'out_ucast_pkts': '160304247',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6537',
                                                                       'out_octets': '73168032882',
                                                                       'out_ucast_pkts': '160304247',
                                                                       'poller_hash': 'NUE2QjhBcGVlcmludGZwb24gNDYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '463',
                                                                                               'links': [       {       'connected_time': '1224969',
                                                                                                                        'fiber_length': 23400,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjhBcGVlcnBvbnBvbiA0NjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '102 SCHWEITZER CR (OHLHAUSEN JEFF)',
                                         'poller_hash': 'NUE2QjhB',
                                         'serial_number': '5A6B8A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:cd',
                                         'uptime': '1224969'},
                                 {       'counter_source': '464/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d3:cf'],
                                                                       'id': '464/g1',
                                                                       'in_nucast_pkts': '141',
                                                                       'in_octets': '311520490746',
                                                                       'in_ucast_pkts': '2785836559',
                                                                       'mac_address': '44:65:7f:7b:c2:ea',
                                                                       'mtu': 2000,
                                                                       'name': '464/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39527005',
                                                                       'out_octets': '12314308058590',
                                                                       'out_ucast_pkts': '8567026125',
                                                                       'poller_hash': 'NUE2NUU5cGVlcmludGY0NjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '464',
                                                                       'in_nucast_pkts': '39527005',
                                                                       'in_octets': '12314308058590',
                                                                       'in_ucast_pkts': '8567026125',
                                                                       'links': [       {       'in_nucast_pkts': '39527005',
                                                                                                'in_octets': '12314308058590',
                                                                                                'in_ucast_pkts': '8567026125',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '141',
                                                                                                'out_octets': '311520490746',
                                                                                                'out_ucast_pkts': '2785836559',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '141',
                                                                       'out_octets': '311520490746',
                                                                       'out_ucast_pkts': '2785836559',
                                                                       'poller_hash': 'NUE2NUU5cGVlcmludGZwb24gNDY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '464',
                                                                                               'links': [       {       'connected_time': '67856372',
                                                                                                                        'fiber_length': 22279,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NUU5cGVlcnBvbnBvbiA0NjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': "1308 Queen's Bush Rd Wellesley "
                                                 '(Carrie Harman 10435242)',
                                         'poller_hash': 'NUE2NUU5',
                                         'serial_number': '5A65E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c2:ea',
                                         'uptime': '67856372'},
                                 {       'counter_source': '466/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '466/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '505102367728',
                                                                       'in_ucast_pkts': '1821306639',
                                                                       'mac_address': 'cc:be:59:cb:86:3a',
                                                                       'name': '466/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '65',
                                                                       'out_octets': '4810962640005',
                                                                       'out_ucast_pkts': '3505372108',
                                                                       'poller_hash': 'NDJDMjFGcGVlcmludGY0NjYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '466',
                                                                       'in_nucast_pkts': '65',
                                                                       'in_octets': '4810962640005',
                                                                       'in_ucast_pkts': '3505372108',
                                                                       'links': [       {       'in_nucast_pkts': '65',
                                                                                                'in_octets': '4810962640005',
                                                                                                'in_ucast_pkts': '3505372108',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '505102367728',
                                                                                                'out_ucast_pkts': '1821306639',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '505102367728',
                                                                       'out_ucast_pkts': '1821306639',
                                                                       'poller_hash': 'NDJDMjFGcGVlcmludGZwb24gNDY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '466',
                                                                                               'links': [       {       'connected_time': '4975882',
                                                                                                                        'fiber_length': 24030,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMjFGcGVlcnBvbnBvbiA0NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '91 Reiner Ct Wellesley (Justin '
                                                 'Martin 10435594)',
                                         'poller_hash': 'NDJDMjFG',
                                         'serial_number': '42C21F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:86:3a',
                                         'uptime': '4975882'},
                                 {       'counter_source': '467/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a8:83:7a'],
                                                                       'id': '467/g1',
                                                                       'in_nucast_pkts': '222',
                                                                       'in_octets': '243155437368',
                                                                       'in_ucast_pkts': '2025242807',
                                                                       'mac_address': 'cc:be:59:70:eb:02',
                                                                       'mtu': 2000,
                                                                       'name': '467/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22467581',
                                                                       'out_octets': '11244046990977',
                                                                       'out_ucast_pkts': '7683206093',
                                                                       'poller_hash': 'NTYxOTUxcGVlcmludGY0NjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '467',
                                                                       'in_nucast_pkts': '22467581',
                                                                       'in_octets': '11244046990977',
                                                                       'in_ucast_pkts': '7683206093',
                                                                       'links': [       {       'in_nucast_pkts': '22467581',
                                                                                                'in_octets': '11244046990977',
                                                                                                'in_ucast_pkts': '7683206093',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '222',
                                                                                                'out_octets': '243155437368',
                                                                                                'out_ucast_pkts': '2025242807',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '222',
                                                                       'out_octets': '243155437368',
                                                                       'out_ucast_pkts': '2025242807',
                                                                       'poller_hash': 'NTYxOTUxcGVlcmludGZwb24gNDY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '467',
                                                                                               'links': [       {       'connected_time': '552572',
                                                                                                                        'fiber_length': 19727,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYxOTUxcGVlcnBvbnBvbiA0Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '4897 Perth Rd 119 Gadshill (Neil '
                                                 'McCann 104323786)',
                                         'poller_hash': 'NTYxOTUx',
                                         'serial_number': '561951',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:70:ea:fe',
                                         'uptime': '552572'},
                                 {       'counter_source': '468/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d5:25'],
                                                                       'id': '468/g1',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '42311768474',
                                                                       'in_ucast_pkts': '138494639',
                                                                       'mac_address': 'cc:be:59:cb:f9:2d',
                                                                       'mtu': 2000,
                                                                       'name': '468/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920062',
                                                                       'out_octets': '1244764889734',
                                                                       'out_ucast_pkts': '901837266',
                                                                       'poller_hash': 'NDJEMDdEcGVlcmludGY0NjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '468',
                                                                       'in_nucast_pkts': '9920062',
                                                                       'in_octets': '1244764889734',
                                                                       'in_ucast_pkts': '901837266',
                                                                       'links': [       {       'in_nucast_pkts': '9920062',
                                                                                                'in_octets': '1244764889734',
                                                                                                'in_ucast_pkts': '901837266',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '42311768474',
                                                                                                'out_ucast_pkts': '138494639',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '42311768474',
                                                                       'out_ucast_pkts': '138494639',
                                                                       'poller_hash': 'NDJEMDdEcGVlcmludGZwb24gNDY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '468',
                                                                                               'links': [       {       'connected_time': '8535208',
                                                                                                                        'fiber_length': 22512,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMDdEcGVlcnBvbnBvbiA0Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1212 QUEEN'S BUSH RD (MARTIN "
                                                 'MARJORIE)',
                                         'poller_hash': 'NDJEMDdE',
                                         'serial_number': '42D07D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:f9:2a',
                                         'uptime': '8535208'},
                                 {       'counter_source': '469/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d5:b5'],
                                                                       'id': '469/g1',
                                                                       'in_nucast_pkts': '13',
                                                                       'in_octets': '151530127491',
                                                                       'in_ucast_pkts': '811002677',
                                                                       'mac_address': 'cc:be:59:cb:c8:75',
                                                                       'mtu': 2000,
                                                                       'name': '469/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9732598',
                                                                       'out_octets': '2976343922741',
                                                                       'out_ucast_pkts': '2125463187',
                                                                       'poller_hash': 'NDJDQTY2cGVlcmludGY0NjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '469',
                                                                       'in_nucast_pkts': '9732598',
                                                                       'in_octets': '2976343922741',
                                                                       'in_ucast_pkts': '2125463187',
                                                                       'links': [       {       'in_nucast_pkts': '9732598',
                                                                                                'in_octets': '2976343922741',
                                                                                                'in_ucast_pkts': '2125463187',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13',
                                                                                                'out_octets': '151530127491',
                                                                                                'out_ucast_pkts': '811002677',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13',
                                                                       'out_octets': '151530127491',
                                                                       'out_ucast_pkts': '811002677',
                                                                       'poller_hash': 'NDJDQTY2cGVlcmludGZwb24gNDY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '469',
                                                                                               'links': [       {       'connected_time': '8478876',
                                                                                                                        'fiber_length': 23381,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDQTY2cGVlcnBvbnBvbiA0Njk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1039 MOLESWORTH ST (LOGAN JULIE)',
                                         'poller_hash': 'NDJDQTY2',
                                         'serial_number': '42CA66',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:c8:72',
                                         'uptime': '8478876'},
                                 {       'counter_source': '470/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:47:ad'],
                                                                       'id': '470/g1',
                                                                       'in_nucast_pkts': '23',
                                                                       'in_octets': '221411946378',
                                                                       'in_ucast_pkts': '1001226563',
                                                                       'mac_address': 'cc:be:59:cb:71:3d',
                                                                       'mtu': 2000,
                                                                       'name': '470/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13088509',
                                                                       'out_octets': '5456587605192',
                                                                       'out_ucast_pkts': '3918998241',
                                                                       'poller_hash': 'NDJCRjdGcGVlcmludGY0NzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '470',
                                                                       'in_nucast_pkts': '13088509',
                                                                       'in_octets': '5456587605192',
                                                                       'in_ucast_pkts': '3918998241',
                                                                       'links': [       {       'in_nucast_pkts': '13088509',
                                                                                                'in_octets': '5456587605192',
                                                                                                'in_ucast_pkts': '3918998241',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23',
                                                                                                'out_octets': '221411946378',
                                                                                                'out_ucast_pkts': '1001226563',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23',
                                                                       'out_octets': '221411946378',
                                                                       'out_ucast_pkts': '1001226563',
                                                                       'poller_hash': 'NDJCRjdGcGVlcmludGZwb24gNDcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '470',
                                                                                               'links': [       {       'connected_time': '11398737',
                                                                                                                        'fiber_length': 23080,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRjdGcGVlcnBvbnBvbiA0NzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1116 Henry St Wellesley (Allen '
                                                 'Bowslaugh 10434962)',
                                         'poller_hash': 'NDJCRjdG',
                                         'serial_number': '42BF7F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:71:3a',
                                         'uptime': '11398737'},
                                 {       'counter_source': '471/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:da:b5:75',
                                                                                                        '00:03:e6:da:bf:bf',
                                                                                                        '44:65:7f:88:d3:99'],
                                                                       'id': '471/g1',
                                                                       'in_nucast_pkts': '87215821',
                                                                       'in_octets': '828092551178',
                                                                       'in_ucast_pkts': '2065911330',
                                                                       'mac_address': 'cc:be:59:cb:aa:95',
                                                                       'mtu': 2000,
                                                                       'name': '471/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '84820921895',
                                                                       'out_octets': '120692158146893',
                                                                       'out_ucast_pkts': '5296163715',
                                                                       'poller_hash': 'NDJDNkFBcGVlcmludGY0NzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '471',
                                                                       'in_nucast_pkts': '84820921895',
                                                                       'in_octets': '120692158146893',
                                                                       'in_ucast_pkts': '5296163715',
                                                                       'links': [       {       'in_nucast_pkts': '84820921895',
                                                                                                'in_octets': '120692158146893',
                                                                                                'in_ucast_pkts': '5296163715',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '87215821',
                                                                                                'out_octets': '828092551178',
                                                                                                'out_ucast_pkts': '2065911330',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '87215821',
                                                                       'out_octets': '828092551178',
                                                                       'out_ucast_pkts': '2065911330',
                                                                       'poller_hash': 'NDJDNkFBcGVlcmludGZwb24gNDcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '471',
                                                                                               'links': [       {       'connected_time': '21886731',
                                                                                                                        'fiber_length': 22809,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkFBcGVlcnBvbnBvbiA0NzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3767 Nafziger Rd Wellesley (Derek '
                                                 'Switzer 10435073)',
                                         'poller_hash': 'NDJDNkFB',
                                         'serial_number': '42C6AA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:aa:92',
                                         'uptime': '21886731'},
                                 {       'counter_source': '472/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:47:bf'],
                                                                       'id': '472/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '64997662567',
                                                                       'in_ucast_pkts': '442309040',
                                                                       'mac_address': '44:65:7f:33:21:59',
                                                                       'mtu': 2000,
                                                                       'name': '472/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652574',
                                                                       'out_octets': '1402947966173',
                                                                       'out_ucast_pkts': '981701636',
                                                                       'poller_hash': 'NjQwQjI0cGVlcmludGY0NzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '472',
                                                                       'in_nucast_pkts': '5652574',
                                                                       'in_octets': '1402947966173',
                                                                       'in_ucast_pkts': '981701636',
                                                                       'links': [       {       'in_nucast_pkts': '5652574',
                                                                                                'in_octets': '1402947966173',
                                                                                                'in_ucast_pkts': '981701636',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '64997662567',
                                                                                                'out_ucast_pkts': '442309040',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '64997662567',
                                                                       'out_ucast_pkts': '442309040',
                                                                       'poller_hash': 'NjQwQjI0cGVlcmludGZwb24gNDcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '472',
                                                                                               'links': [       {       'connected_time': '8535105',
                                                                                                                        'fiber_length': 23238,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.18,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.59}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjI0cGVlcnBvbnBvbiA0NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '3850 Nafziger Rd Wellesley (Irene '
                                                 'Cook 10434806)',
                                         'poller_hash': 'NjQwQjI0',
                                         'serial_number': '640B24',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:58',
                                         'uptime': '8535105'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '474/g1',
                                                                       'name': '474/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzQ5RTQ4cGVlcmludGY0NzQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '474',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzQ5RTQ4cGVlcmludGZwb24gNDc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '474',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ5RTQ4cGVlcnBvbnBvbiA0NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '140 DOWNIE ST (TEST MICHELLE)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "474" '
                                                                  '(location: GPON '
                                                                  'port "7/8"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-01T19:08:34.550Z',
                                         'poller_hash': 'MzQ5RTQ4',
                                         'serial_number': '349E48',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '476/g1',
                                                                       'name': '476/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTYwQkZEcGVlcmludGY0NzYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '476',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NTYwQkZEcGVlcmludGZwb24gNDc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '476',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NTYwQkZEcGVlcnBvbnBvbiA0NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '10 FIRELLA PL',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "476" '
                                                                  '(location: GPON '
                                                                  'port "6/6"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-10-09T15:01:19.790Z',
                                         'poller_hash': 'NTYwQkZE',
                                         'serial_number': '560BFD',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '477/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:47:65'],
                                                                       'id': '477/g1',
                                                                       'in_nucast_pkts': '1859468',
                                                                       'in_octets': '60033063662',
                                                                       'in_ucast_pkts': '488256102',
                                                                       'mac_address': 'cc:be:59:cb:aa:35',
                                                                       'mtu': 2000,
                                                                       'name': '477/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2029152321',
                                                                       'out_octets': '4879478972413',
                                                                       'out_ucast_pkts': '1576363659',
                                                                       'poller_hash': 'NDJDNjlFcGVlcmludGY0NzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '477',
                                                                       'in_nucast_pkts': '2029152321',
                                                                       'in_octets': '4879478972413',
                                                                       'in_ucast_pkts': '1576363659',
                                                                       'links': [       {       'in_nucast_pkts': '2029152321',
                                                                                                'in_octets': '4879478972413',
                                                                                                'in_ucast_pkts': '1576363659',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1859468',
                                                                                                'out_octets': '60033063662',
                                                                                                'out_ucast_pkts': '488256102',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1859468',
                                                                       'out_octets': '60033063662',
                                                                       'out_ucast_pkts': '488256102',
                                                                       'poller_hash': 'NDJDNjlFcGVlcmludGZwb24gNDc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '477',
                                                                                               'links': [       {       'connected_time': '8535199',
                                                                                                                        'fiber_length': 23021,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNjlFcGVlcnBvbnBvbiA0Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1018 Doering St Wellesley (Jim '
                                                 'Capling 10435327)',
                                         'poller_hash': 'NDJDNjlF',
                                         'serial_number': '42C69E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:aa:32',
                                         'uptime': '8535199'},
                                 {       'counter_source': '478/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0c:cc'],
                                                                       'id': '478/g1',
                                                                       'in_nucast_pkts': '30',
                                                                       'in_octets': '134563693682',
                                                                       'in_ucast_pkts': '341416440',
                                                                       'mac_address': '44:65:7f:33:21:ad',
                                                                       'mtu': 2000,
                                                                       'name': '478/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530468',
                                                                       'out_octets': '1107274120843',
                                                                       'out_ucast_pkts': '841041107',
                                                                       'poller_hash': 'NjQwQjRFcGVlcmludGY0NzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '478',
                                                                       'in_nucast_pkts': '5530468',
                                                                       'in_octets': '1107274120843',
                                                                       'in_ucast_pkts': '841041107',
                                                                       'links': [       {       'in_nucast_pkts': '5530468',
                                                                                                'in_octets': '1107274120843',
                                                                                                'in_ucast_pkts': '841041107',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '30',
                                                                                                'out_octets': '134563693682',
                                                                                                'out_ucast_pkts': '341416440',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30',
                                                                       'out_octets': '134563693682',
                                                                       'out_ucast_pkts': '341416440',
                                                                       'poller_hash': 'NjQwQjRFcGVlcmludGZwb24gNDc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '478',
                                                                                               'links': [       {       'connected_time': '8535132',
                                                                                                                        'fiber_length': 22958,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.986,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.204}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjRFcGVlcnBvbnBvbiA0Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1099 Henry St Wellesley (Jeff Hill '
                                                 '10435307)',
                                         'poller_hash': 'NjQwQjRF',
                                         'serial_number': '640B4E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:ac',
                                         'uptime': '8535132'},
                                 {       'counter_source': '479/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:46:72'],
                                                                       'id': '479/g1',
                                                                       'in_nucast_pkts': '32',
                                                                       'in_octets': '15590539901',
                                                                       'in_ucast_pkts': '110984397',
                                                                       'mac_address': 'cc:be:59:cb:79:bd',
                                                                       'mtu': 2000,
                                                                       'name': '479/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919650',
                                                                       'out_octets': '577316807514',
                                                                       'out_ucast_pkts': '394668307',
                                                                       'poller_hash': 'NDJDMDhGcGVlcmludGY0NzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '479',
                                                                       'in_nucast_pkts': '9919650',
                                                                       'in_octets': '577316807514',
                                                                       'in_ucast_pkts': '394668307',
                                                                       'links': [       {       'in_nucast_pkts': '9919650',
                                                                                                'in_octets': '577316807514',
                                                                                                'in_ucast_pkts': '394668307',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '32',
                                                                                                'out_octets': '15590539901',
                                                                                                'out_ucast_pkts': '110984397',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '32',
                                                                       'out_octets': '15590539901',
                                                                       'out_ucast_pkts': '110984397',
                                                                       'poller_hash': 'NDJDMDhGcGVlcmludGZwb24gNDc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '479',
                                                                                               'links': [       {       'connected_time': '8535276',
                                                                                                                        'fiber_length': 22854,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMDhGcGVlcnBvbnBvbiA0Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3758 Nafziger Rd (Rhonda Lee '
                                                 'Fetterly10434984)',
                                         'poller_hash': 'NDJDMDhG',
                                         'serial_number': '42C08F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:79:ba',
                                         'uptime': '8535276'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '480/g1',
                                                                       'name': '480/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEZBMjYycGVlcmludGY0ODAvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '480',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NEZBMjYycGVlcmludGZwb24gNDgw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '480',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEZBMjYycGVlcnBvbnBvbiA0ODA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '38 Pond View Dr',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "480" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'Dying Gasp", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-06-20T07:51:56.400Z',
                                         'poller_hash': 'NEZBMjYy',
                                         'serial_number': '4FA262',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '481/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '481/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '451138634334',
                                                                       'in_ucast_pkts': '1669833178',
                                                                       'mac_address': 'cc:be:59:cb:e4:da',
                                                                       'name': '481/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '11763458179578',
                                                                       'out_ucast_pkts': '8239713937',
                                                                       'poller_hash': 'NDJDREYzcGVlcmludGY0ODEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '481',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '11763458179578',
                                                                       'in_ucast_pkts': '8239713937',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '11763458179578',
                                                                                                'in_ucast_pkts': '8239713937',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '451138634334',
                                                                                                'out_ucast_pkts': '1669833178',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '451138634334',
                                                                       'out_ucast_pkts': '1669833178',
                                                                       'poller_hash': 'NDJDREYzcGVlcmludGZwb24gNDgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '481',
                                                                                               'links': [       {       'connected_time': '1313567',
                                                                                                                        'fiber_length': 24528,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDREYzcGVlcnBvbnBvbiA0ODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '162 Gerber Meadows Dr (Neil '
                                                 'MacDougall, 10435414)',
                                         'poller_hash': 'NDJDREYz',
                                         'serial_number': '42CDF3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e4:da',
                                         'uptime': '1313567'},
                                 {       'counter_source': '482/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:e9:31:8c:d8:64',
                                                                                                        '74:fe:ce:31:9f:89'],
                                                                       'id': '482/g1',
                                                                       'in_nucast_pkts': '21981',
                                                                       'in_octets': '165103476318',
                                                                       'in_ucast_pkts': '759747495',
                                                                       'mac_address': '44:65:7f:7b:d3:e2',
                                                                       'mtu': 2000,
                                                                       'name': '482/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19874760',
                                                                       'out_octets': '2727374500396',
                                                                       'out_ucast_pkts': '2033040096',
                                                                       'poller_hash': 'NUE2QjkxcGVlcmludGY0ODIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '482',
                                                                       'in_nucast_pkts': '19874760',
                                                                       'in_octets': '2727374500396',
                                                                       'in_ucast_pkts': '2033040096',
                                                                       'links': [       {       'in_nucast_pkts': '19874760',
                                                                                                'in_octets': '2727374500396',
                                                                                                'in_ucast_pkts': '2033040096',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21981',
                                                                                                'out_octets': '165103476318',
                                                                                                'out_ucast_pkts': '759747495',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21981',
                                                                       'out_octets': '165103476318',
                                                                       'out_ucast_pkts': '759747495',
                                                                       'poller_hash': 'NUE2QjkxcGVlcmludGZwb24gNDgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '482',
                                                                                               'links': [       {       'connected_time': '17172868',
                                                                                                                        'fiber_length': 19954,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjkxcGVlcnBvbnBvbiA0ODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4939 Perth Rd 119 Gadshill (J & M '
                                                 'Patterson 10423710)',
                                         'poller_hash': 'NUE2Qjkx',
                                         'serial_number': '5A6B91',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:e2',
                                         'uptime': '17172868'},
                                 {       'counter_source': '483/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f3:43'],
                                                                       'id': '483/g1',
                                                                       'in_nucast_pkts': '161874',
                                                                       'in_octets': '218635970141',
                                                                       'in_ucast_pkts': '662643032',
                                                                       'mac_address': 'cc:be:59:cc:05:25',
                                                                       'mtu': 2000,
                                                                       'name': '483/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9793291',
                                                                       'out_octets': '2168619536862',
                                                                       'out_ucast_pkts': '1623300579',
                                                                       'poller_hash': 'NDJEMUZDcGVlcmludGY0ODMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '483',
                                                                       'in_nucast_pkts': '9793291',
                                                                       'in_octets': '2168619536862',
                                                                       'in_ucast_pkts': '1623300579',
                                                                       'links': [       {       'in_nucast_pkts': '9793291',
                                                                                                'in_octets': '2168619536862',
                                                                                                'in_ucast_pkts': '1623300579',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '161874',
                                                                                                'out_octets': '218635970141',
                                                                                                'out_ucast_pkts': '662643032',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '161874',
                                                                       'out_octets': '218635970141',
                                                                       'out_ucast_pkts': '662643032',
                                                                       'poller_hash': 'NDJEMUZDcGVlcmludGZwb24gNDgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '483',
                                                                                               'links': [       {       'connected_time': '8535228',
                                                                                                                        'fiber_length': 23473,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMUZDcGVlcnBvbnBvbiA0ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '42 ZINKANN CR (FITZPATRICK SAMANTHA)',
                                         'poller_hash': 'NDJEMUZD',
                                         'serial_number': '42D1FC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:05:22',
                                         'uptime': '8535228'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '484/g1',
                                                                       'name': '484/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDNjkzcGVlcmludGY0ODQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '484',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDNjkzcGVlcmludGZwb24gNDg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '484',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNjkzcGVlcnBvbnBvbiA0ODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '51 ZOEGER CT',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "484" '
                                                                  '(location: GPON '
                                                                  'port "7/2"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:27:12.820Z',
                                         'poller_hash': 'NDJDNjkz',
                                         'serial_number': '42C693',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '485/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '485/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '361125817747',
                                                                       'in_ucast_pkts': '39608094675',
                                                                       'mac_address': 'cc:be:59:cb:ab:1a',
                                                                       'name': '485/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '139',
                                                                       'out_octets': '3541624444641',
                                                                       'out_ucast_pkts': '2725676103',
                                                                       'poller_hash': 'NDJDNkJCcGVlcmludGY0ODUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '485',
                                                                       'in_nucast_pkts': '139',
                                                                       'in_octets': '3541624444641',
                                                                       'in_ucast_pkts': '2725676103',
                                                                       'links': [       {       'in_nucast_pkts': '139',
                                                                                                'in_octets': '3541624444641',
                                                                                                'in_ucast_pkts': '2725676103',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '361125817747',
                                                                                                'out_ucast_pkts': '39608094675',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '361125817747',
                                                                       'out_ucast_pkts': '39608094675',
                                                                       'poller_hash': 'NDJDNkJCcGVlcmludGZwb24gNDg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '485',
                                                                                               'links': [       {       'connected_time': '8535204',
                                                                                                                        'fiber_length': 19544,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkJCcGVlcnBvbnBvbiA0ODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1890 Perth Line Wellesley (Wayne '
                                                 'Jantzi 10434423)',
                                         'poller_hash': 'NDJDNkJC',
                                         'serial_number': '42C6BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ab:1a',
                                         'uptime': '8535204'},
                                 {       'counter_source': '487/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0f:c9'],
                                                                       'id': '487/g1',
                                                                       'in_nucast_pkts': '6484052',
                                                                       'in_octets': '294507922206',
                                                                       'in_ucast_pkts': '1021256095',
                                                                       'mac_address': 'cc:be:59:cb:ac:e5',
                                                                       'mtu': 2000,
                                                                       'name': '487/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11571469093',
                                                                       'out_octets': '22033046910435',
                                                                       'out_ucast_pkts': '4524969456',
                                                                       'poller_hash': 'NDJDNkY0cGVlcmludGY0ODcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '487',
                                                                       'in_nucast_pkts': '11571469093',
                                                                       'in_octets': '22033046910435',
                                                                       'in_ucast_pkts': '4524969456',
                                                                       'links': [       {       'in_nucast_pkts': '11571469093',
                                                                                                'in_octets': '22033046910435',
                                                                                                'in_ucast_pkts': '4524969456',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6484052',
                                                                                                'out_octets': '294507922206',
                                                                                                'out_ucast_pkts': '1021256095',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6484052',
                                                                       'out_octets': '294507922206',
                                                                       'out_ucast_pkts': '1021256095',
                                                                       'poller_hash': 'NDJDNkY0cGVlcmludGZwb24gNDg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '487',
                                                                                               'links': [       {       'connected_time': '8535258',
                                                                                                                        'fiber_length': 24481,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkY0cGVlcnBvbnBvbiA0ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '47 ParkviewDr Wellesley (Sherri Mula '
                                                 '10435398)',
                                         'poller_hash': 'NDJDNkY0',
                                         'serial_number': '42C6F4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ac:e2',
                                         'uptime': '8535258'},
                                 {       'counter_source': '489/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:53'],
                                                                       'id': '489/g1',
                                                                       'in_nucast_pkts': '163842',
                                                                       'in_octets': '1324615033171',
                                                                       'in_ucast_pkts': '2915460688',
                                                                       'mac_address': 'd0:76:8f:08:19:ae',
                                                                       'mtu': 2000,
                                                                       'name': '489/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919952',
                                                                       'out_octets': '13671825187874',
                                                                       'out_ucast_pkts': '10516245318',
                                                                       'poller_hash': 'ODY5NDJCcGVlcmludGY0ODkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '489',
                                                                       'in_nucast_pkts': '9919952',
                                                                       'in_octets': '13671825187874',
                                                                       'in_ucast_pkts': '10516245318',
                                                                       'links': [       {       'in_nucast_pkts': '9919952',
                                                                                                'in_octets': '13671825187874',
                                                                                                'in_ucast_pkts': '10516245318',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '163842',
                                                                                                'out_octets': '1324615033171',
                                                                                                'out_ucast_pkts': '2915460688',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '163842',
                                                                       'out_octets': '1324615033171',
                                                                       'out_ucast_pkts': '2915460688',
                                                                       'poller_hash': 'ODY5NDJCcGVlcmludGZwb24gNDg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '489',
                                                                                               'links': [       {       'connected_time': '8535222',
                                                                                                                        'fiber_length': 23956,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDJCcGVlcnBvbnBvbiA0ODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 Schweitzer Cr Wellesley (Robin '
                                                 'Palermo 10435541)',
                                         'poller_hash': 'ODY5NDJC',
                                         'serial_number': '86942B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:ab',
                                         'uptime': '8535222'},
                                 {       'counter_source': '491/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '58:d5:6e:e7:49:ce'],
                                                                       'id': '491/g1',
                                                                       'in_nucast_pkts': '9742',
                                                                       'in_octets': '167480799701',
                                                                       'in_ucast_pkts': '650044499',
                                                                       'mac_address': 'cc:be:59:cb:9b:05',
                                                                       'mtu': 2000,
                                                                       'name': '491/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9937593',
                                                                       'out_octets': '2526593287704',
                                                                       'out_ucast_pkts': '1879884995',
                                                                       'poller_hash': 'NDJDNEI4cGVlcmludGY0OTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '491',
                                                                       'in_nucast_pkts': '9937593',
                                                                       'in_octets': '2526593287704',
                                                                       'in_ucast_pkts': '1879884995',
                                                                       'links': [       {       'in_nucast_pkts': '9937593',
                                                                                                'in_octets': '2526593287704',
                                                                                                'in_ucast_pkts': '1879884995',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9742',
                                                                                                'out_octets': '167480799701',
                                                                                                'out_ucast_pkts': '650044499',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9742',
                                                                       'out_octets': '167480799701',
                                                                       'out_ucast_pkts': '650044499',
                                                                       'poller_hash': 'NDJDNEI4cGVlcmludGZwb24gNDkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '491',
                                                                                               'links': [       {       'connected_time': '8535291',
                                                                                                                        'fiber_length': 22851,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNEI4cGVlcnBvbnBvbiA0OTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3766 Nafziger Rd (Marvin Martin, '
                                                 '10435662)',
                                         'poller_hash': 'NDJDNEI4',
                                         'serial_number': '42C4B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:9b:02',
                                         'uptime': '8535291'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '492/g1',
                                                                       'name': '492/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2NjMwcGVlcmludGY0OTIvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '492',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2NjMwcGVlcmludGZwb24gNDky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '492',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjMwcGVlcnBvbnBvbiA0OTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '803G ONT (Gigaspire Feed)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "492" '
                                                                  '(location: GPON '
                                                                  'port "7/8"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-01T19:08:34.510Z',
                                         'poller_hash': 'NUE2NjMw',
                                         'serial_number': '5A6630',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '493/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:45:b5'],
                                                                       'id': '493/g1',
                                                                       'in_nucast_pkts': '52703',
                                                                       'in_octets': '37323514020',
                                                                       'in_ucast_pkts': '145104816',
                                                                       'mac_address': 'cc:be:59:cb:9e:45',
                                                                       'mtu': 2000,
                                                                       'name': '493/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1308621',
                                                                       'out_octets': '576806212622',
                                                                       'out_ucast_pkts': '434314768',
                                                                       'poller_hash': 'NDJDNTIwcGVlcmludGY0OTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '493',
                                                                       'in_nucast_pkts': '1308621',
                                                                       'in_octets': '576806212622',
                                                                       'in_ucast_pkts': '434314768',
                                                                       'links': [       {       'in_nucast_pkts': '1308621',
                                                                                                'in_octets': '576806212622',
                                                                                                'in_ucast_pkts': '434314768',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '52703',
                                                                                                'out_octets': '37323514020',
                                                                                                'out_ucast_pkts': '145104816',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '52703',
                                                                       'out_octets': '37323514020',
                                                                       'out_ucast_pkts': '145104816',
                                                                       'poller_hash': 'NDJDNTIwcGVlcmludGZwb24gNDkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '493',
                                                                                               'links': [       {       'connected_time': '1084853',
                                                                                                                        'fiber_length': 23331,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNTIwcGVlcnBvbnBvbiA0OTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 Berdux Pl (Michael '
                                                 'Hayter,10434949)',
                                         'poller_hash': 'NDJDNTIw',
                                         'serial_number': '42C520',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:9e:42',
                                         'uptime': '1084853'},
                                 {       'counter_source': '494/g2+494/g3+494/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:ba:3d'],
                                                                       'id': '494/g2',
                                                                       'in_nucast_pkts': '2302222',
                                                                       'in_octets': '106782063819',
                                                                       'in_ucast_pkts': '641329670',
                                                                       'mac_address': '48:77:46:3a:7f:ef',
                                                                       'mtu': 2000,
                                                                       'name': '494/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19885971',
                                                                       'out_octets': '4124964460074',
                                                                       'out_ucast_pkts': '2853830666',
                                                                       'poller_hash': 'NzIyMjE4cGVlcmludGY0OTQvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1f:b6'],
                                                                       'id': '494/g3',
                                                                       'in_nucast_pkts': '12499514',
                                                                       'in_octets': '48409784627',
                                                                       'in_ucast_pkts': '436579019',
                                                                       'mac_address': '48:77:46:3a:7f:f0',
                                                                       'mtu': 2000,
                                                                       'name': '494/g3',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31251429619',
                                                                       'out_octets': '43972520459330',
                                                                       'out_ucast_pkts': '1445503043',
                                                                       'poller_hash': 'NzIyMjE4cGVlcmludGY0OTQvZzM=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d4:d2'],
                                                                       'id': '494/g4',
                                                                       'in_nucast_pkts': '801480',
                                                                       'in_octets': '222979611903',
                                                                       'in_ucast_pkts': '847953469',
                                                                       'mac_address': '48:77:46:3a:7f:f1',
                                                                       'mtu': 2000,
                                                                       'name': '494/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23102780',
                                                                       'out_octets': '6013427548638',
                                                                       'out_ucast_pkts': '4443806371',
                                                                       'poller_hash': 'NzIyMjE4cGVlcmludGY0OTQvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '494',
                                                                       'in_nucast_pkts': '31294418370',
                                                                       'in_octets': '54110912468042',
                                                                       'in_ucast_pkts': '8743140080',
                                                                       'links': [       {       'in_nucast_pkts': '31294418370',
                                                                                                'in_octets': '54110912468042',
                                                                                                'in_ucast_pkts': '8743140080',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15603216',
                                                                                                'out_octets': '378171460349',
                                                                                                'out_ucast_pkts': '1925862158',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15603216',
                                                                       'out_octets': '378171460349',
                                                                       'out_ucast_pkts': '1925862158',
                                                                       'poller_hash': 'NzIyMjE4cGVlcmludGZwb24gNDk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '494',
                                                                                               'links': [       {       'connected_time': '552653',
                                                                                                                        'fiber_length': 23440,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzIyMjE4cGVlcnBvbnBvbiA0OTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '1020 MAPLE LEAF ST (BRODRECHT '
                                                 'ANDREA)',
                                         'poller_hash': 'NzIyMjE4',
                                         'serial_number': '722218',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:3a:7f:ea',
                                         'uptime': '552653'},
                                 {       'counter_source': '495/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0d:da'],
                                                                       'id': '495/g1',
                                                                       'in_nucast_pkts': '4354',
                                                                       'in_octets': '418466235345',
                                                                       'in_ucast_pkts': '1140003398',
                                                                       'mac_address': '44:65:7f:7b:c4:7c',
                                                                       'mtu': 2000,
                                                                       'name': '495/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '37971536',
                                                                       'out_octets': '3678948103827',
                                                                       'out_ucast_pkts': '2975951431',
                                                                       'poller_hash': 'NUE2NjZGcGVlcmludGY0OTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '495',
                                                                       'in_nucast_pkts': '37971536',
                                                                       'in_octets': '3678948103827',
                                                                       'in_ucast_pkts': '2975951431',
                                                                       'links': [       {       'in_nucast_pkts': '37971536',
                                                                                                'in_octets': '3678948103827',
                                                                                                'in_ucast_pkts': '2975951431',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4354',
                                                                                                'out_octets': '418466235345',
                                                                                                'out_ucast_pkts': '1140003398',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4354',
                                                                       'out_octets': '418466235345',
                                                                       'out_ucast_pkts': '1140003398',
                                                                       'poller_hash': 'NUE2NjZGcGVlcmludGZwb24gNDk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '495',
                                                                                               'links': [       {       'connected_time': '67856103',
                                                                                                                        'fiber_length': 22032,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjZGcGVlcnBvbnBvbiA0OTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1420 Hutchinson Rd Wellesley (K & K '
                                                 'Liquidation 10435446)',
                                         'poller_hash': 'NUE2NjZG',
                                         'serial_number': '5A666F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:7c',
                                         'uptime': '67856103'},
                                 {       'counter_source': '496/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:81:f3:2a'],
                                                                       'id': '496/g1',
                                                                       'in_nucast_pkts': '5667',
                                                                       'in_octets': '1316818830',
                                                                       'in_ucast_pkts': '5348367',
                                                                       'mac_address': '60:db:98:8e:ff:b0',
                                                                       'mtu': 2000,
                                                                       'name': '496/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715398',
                                                                       'out_octets': '45309879000',
                                                                       'out_ucast_pkts': '35712076',
                                                                       'poller_hash': 'QjUxMTU0cGVlcmludGY0OTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '496',
                                                                       'in_nucast_pkts': '715398',
                                                                       'in_octets': '45309879000',
                                                                       'in_ucast_pkts': '35712076',
                                                                       'links': [       {       'in_nucast_pkts': '715398',
                                                                                                'in_octets': '45309879000',
                                                                                                'in_ucast_pkts': '35712076',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5667',
                                                                                                'out_octets': '1316818830',
                                                                                                'out_ucast_pkts': '5348367',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5667',
                                                                       'out_octets': '1316818830',
                                                                       'out_ucast_pkts': '5348367',
                                                                       'poller_hash': 'QjUxMTU0cGVlcmludGZwb24gNDk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '496',
                                                                                               'links': [       {       'connected_time': '611478',
                                                                                                                        'fiber_length': 19986,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMTU0cGVlcnBvbnBvbiA0OTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4894 Linhaven Dr Gadshill (Lucas '
                                                 'Commerford 10435700)',
                                         'poller_hash': 'QjUxMTU0',
                                         'serial_number': 'B51154',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8e:ff:b0',
                                         'uptime': '611478'},
                                 {       'counter_source': '498/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:b9:fb'],
                                                                       'id': '498/g1',
                                                                       'in_nucast_pkts': '159542',
                                                                       'in_octets': '66807115068',
                                                                       'in_ucast_pkts': '525393845',
                                                                       'mac_address': 'cc:be:59:cb:6a:3d',
                                                                       'mtu': 2000,
                                                                       'name': '498/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797798',
                                                                       'out_octets': '3900995309769',
                                                                       'out_ucast_pkts': '2690602418',
                                                                       'poller_hash': 'NDJCRTlGcGVlcmludGY0OTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '498',
                                                                       'in_nucast_pkts': '9797798',
                                                                       'in_octets': '3900995309769',
                                                                       'in_ucast_pkts': '2690602418',
                                                                       'links': [       {       'in_nucast_pkts': '9797798',
                                                                                                'in_octets': '3900995309769',
                                                                                                'in_ucast_pkts': '2690602418',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '159542',
                                                                                                'out_octets': '66807115068',
                                                                                                'out_ucast_pkts': '525393845',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '159542',
                                                                       'out_octets': '66807115068',
                                                                       'out_ucast_pkts': '525393845',
                                                                       'poller_hash': 'NDJCRTlGcGVlcmludGZwb24gNDk4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '498',
                                                                                               'links': [       {       'connected_time': '8535351',
                                                                                                                        'fiber_length': 23799,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRTlGcGVlcnBvbnBvbiA0OTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '67 Reiner Ct Wellesley (Jeff '
                                                 'Lawrence 10434859)',
                                         'poller_hash': 'NDJCRTlG',
                                         'serial_number': '42BE9F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6a:3a',
                                         'uptime': '8535351'},
                                 {       'counter_source': '499/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:38:83:18:98:11'],
                                                                       'id': '499/g1',
                                                                       'in_nucast_pkts': '612268',
                                                                       'in_octets': '3844195334413',
                                                                       'in_ucast_pkts': '4862622772',
                                                                       'mac_address': 'cc:be:59:cb:aa:8d',
                                                                       'mtu': 2000,
                                                                       'name': '499/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797335',
                                                                       'out_octets': '6430012722685',
                                                                       'out_ucast_pkts': '6317668199',
                                                                       'poller_hash': 'NDJDNkE5cGVlcmludGY0OTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '499',
                                                                       'in_nucast_pkts': '9797335',
                                                                       'in_octets': '6430012722685',
                                                                       'in_ucast_pkts': '6317668199',
                                                                       'links': [       {       'in_nucast_pkts': '9797335',
                                                                                                'in_octets': '6430012722685',
                                                                                                'in_ucast_pkts': '6317668199',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '612268',
                                                                                                'out_octets': '3844195334413',
                                                                                                'out_ucast_pkts': '4862622772',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '612268',
                                                                       'out_octets': '3844195334413',
                                                                       'out_ucast_pkts': '4862622772',
                                                                       'poller_hash': 'NDJDNkE5cGVlcmludGZwb24gNDk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '499',
                                                                                               'links': [       {       'connected_time': '8535229',
                                                                                                                        'fiber_length': 25588,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkE5cGVlcnBvbnBvbiA0OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '139 VILLAGE RD (TURNER PAUL)',
                                         'poller_hash': 'NDJDNkE5',
                                         'serial_number': '42C6A9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:aa:8a',
                                         'uptime': '8535229'},
                                 {       'counter_source': '500/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b5:ea'],
                                                                       'id': '500/g1',
                                                                       'in_nucast_pkts': '5654471',
                                                                       'in_octets': '59743148778',
                                                                       'in_ucast_pkts': '138971228',
                                                                       'mac_address': 'cc:be:59:cb:ac:3d',
                                                                       'mtu': 2000,
                                                                       'name': '500/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '918941421',
                                                                       'out_octets': '1788776419154',
                                                                       'out_ucast_pkts': '485193924',
                                                                       'poller_hash': 'NDJDNkRGcGVlcmludGY1MDAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '500',
                                                                       'in_nucast_pkts': '918941421',
                                                                       'in_octets': '1788776419154',
                                                                       'in_ucast_pkts': '485193924',
                                                                       'links': [       {       'in_nucast_pkts': '918941421',
                                                                                                'in_octets': '1788776419154',
                                                                                                'in_ucast_pkts': '485193924',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5654471',
                                                                                                'out_octets': '59743148778',
                                                                                                'out_ucast_pkts': '138971228',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5654471',
                                                                       'out_octets': '59743148778',
                                                                       'out_ucast_pkts': '138971228',
                                                                       'poller_hash': 'NDJDNkRGcGVlcmludGZwb24gNTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '500',
                                                                                               'links': [       {       'connected_time': '8535407',
                                                                                                                        'fiber_length': 25594,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNkRGcGVlcnBvbnBvbiA1MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '120 Village Rd Wellesley (Keith '
                                                 'Kuepfer 10435601)',
                                         'poller_hash': 'NDJDNkRG',
                                         'serial_number': '42C6DF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ac:3a',
                                         'uptime': '8535407'},
                                 {       'counter_source': '501/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:8e'],
                                                                       'id': '501/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '58033472775',
                                                                       'in_ucast_pkts': '215657642',
                                                                       'mac_address': '44:65:7f:33:21:93',
                                                                       'mtu': 2000,
                                                                       'name': '501/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5652646',
                                                                       'out_octets': '1023844095139',
                                                                       'out_ucast_pkts': '825224856',
                                                                       'poller_hash': 'NjQwQjQxcGVlcmludGY1MDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '501',
                                                                       'in_nucast_pkts': '5652646',
                                                                       'in_octets': '1023844095139',
                                                                       'in_ucast_pkts': '825224856',
                                                                       'links': [       {       'in_nucast_pkts': '5652646',
                                                                                                'in_octets': '1023844095139',
                                                                                                'in_ucast_pkts': '825224856',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '58033472775',
                                                                                                'out_ucast_pkts': '215657642',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '58033472775',
                                                                       'out_ucast_pkts': '215657642',
                                                                       'poller_hash': 'NjQwQjQxcGVlcmludGZwb24gNTAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '501',
                                                                                               'links': [       {       'connected_time': '8535122',
                                                                                                                        'fiber_length': 22267,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.27,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.458}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjQxcGVlcnBvbnBvbiA1MDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': "1316 Queen's Bush Rd Wellesley (Josh "
                                                 'Zehr 10434844)',
                                         'poller_hash': 'NjQwQjQx',
                                         'serial_number': '640B41',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:92',
                                         'uptime': '8535122'},
                                 {       'counter_source': '502/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:b7:3d'],
                                                                       'id': '502/g1',
                                                                       'in_nucast_pkts': '9617476',
                                                                       'in_octets': '62327785306',
                                                                       'in_ucast_pkts': '257767862',
                                                                       'mac_address': '44:65:7f:33:21:e3',
                                                                       'mtu': 2000,
                                                                       'name': '502/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530412',
                                                                       'out_octets': '2084064768837',
                                                                       'out_ucast_pkts': '1456194812',
                                                                       'poller_hash': 'NjQwQjY5cGVlcmludGY1MDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '502',
                                                                       'in_nucast_pkts': '5530412',
                                                                       'in_octets': '2084064768837',
                                                                       'in_ucast_pkts': '1456194812',
                                                                       'links': [       {       'in_nucast_pkts': '5530412',
                                                                                                'in_octets': '2084064768837',
                                                                                                'in_ucast_pkts': '1456194812',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9617476',
                                                                                                'out_octets': '62327785306',
                                                                                                'out_ucast_pkts': '257767862',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9617476',
                                                                       'out_octets': '62327785306',
                                                                       'out_ucast_pkts': '257767862',
                                                                       'poller_hash': 'NjQwQjY5cGVlcmludGZwb24gNTAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '502',
                                                                                               'links': [       {       'connected_time': '8535110',
                                                                                                                        'fiber_length': 23286,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.268,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.268}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjY5cGVlcnBvbnBvbiA1MDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '32 Zinkann Cr Wellesley (Jeff '
                                                 'Kneller 10435335)',
                                         'poller_hash': 'NjQwQjY5',
                                         'serial_number': '640B69',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:e2',
                                         'uptime': '8535110'},
                                 {       'counter_source': '503/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:90:7f:b9:86:b4'],
                                                                       'id': '503/g1',
                                                                       'in_nucast_pkts': '49933',
                                                                       'in_octets': '767240745',
                                                                       'in_ucast_pkts': '6200976',
                                                                       'mac_address': 'cc:be:59:50:00:a4',
                                                                       'mtu': 2000,
                                                                       'name': '503/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9678537',
                                                                       'out_octets': '5434528477',
                                                                       'out_ucast_pkts': '12779776',
                                                                       'poller_hash': 'NDNCNDFBcGVlcmludGY1MDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '503',
                                                                       'in_nucast_pkts': '9678537',
                                                                       'in_octets': '5434528477',
                                                                       'in_ucast_pkts': '12779776',
                                                                       'links': [       {       'in_nucast_pkts': '9678537',
                                                                                                'in_octets': '5434528477',
                                                                                                'in_ucast_pkts': '12779776',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '49933',
                                                                                                'out_octets': '767240745',
                                                                                                'out_ucast_pkts': '6200976',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49933',
                                                                       'out_octets': '767240745',
                                                                       'out_ucast_pkts': '6200976',
                                                                       'poller_hash': 'NDNCNDFBcGVlcmludGZwb24gNTAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '503',
                                                                                               'links': [       {       'connected_time': '14604827',
                                                                                                                        'fiber_length': 23804,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.862,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.69}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNCNDFBcGVlcnBvbnBvbiA1MDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '3706 Nafziger Rd Wellesley (Adds & '
                                                 'Ends 10435557)',
                                         'poller_hash': 'NDNCNDFB',
                                         'serial_number': '43B41A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': 'cc:be:59:50:00:a3',
                                         'uptime': '14604827'},
                                 {       'counter_source': '504/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0e:fa'],
                                                                       'id': '504/g2',
                                                                       'in_nucast_pkts': '27',
                                                                       'in_octets': '173045548569',
                                                                       'in_ucast_pkts': '891313518',
                                                                       'mac_address': 'cc:be:59:cb:e6:ee',
                                                                       'mtu': 2000,
                                                                       'name': '504/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920096',
                                                                       'out_octets': '3221959349356',
                                                                       'out_ucast_pkts': '2317988603',
                                                                       'poller_hash': 'NDJDRTM1cGVlcmludGY1MDQvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '504',
                                                                       'in_nucast_pkts': '9920096',
                                                                       'in_octets': '3221959349356',
                                                                       'in_ucast_pkts': '2317988603',
                                                                       'links': [       {       'in_nucast_pkts': '9920096',
                                                                                                'in_octets': '3221959349356',
                                                                                                'in_ucast_pkts': '2317988603',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27',
                                                                                                'out_octets': '173045548569',
                                                                                                'out_ucast_pkts': '891313518',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27',
                                                                       'out_octets': '173045548569',
                                                                       'out_ucast_pkts': '891313518',
                                                                       'poller_hash': 'NDJDRTM1cGVlcmludGZwb24gNTA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '504',
                                                                                               'links': [       {       'connected_time': '8535182',
                                                                                                                        'fiber_length': 23018,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTM1cGVlcnBvbnBvbiA1MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1025 Doering St Wellesley (Gretchen '
                                                 'Robertson 10435569)',
                                         'poller_hash': 'NDJDRTM1',
                                         'serial_number': '42CE35',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e6:ea',
                                         'uptime': '8535182'},
                                 {       'counter_source': '505/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:79:d7:02'],
                                                                       'id': '505/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '119159151689',
                                                                       'in_ucast_pkts': '421968903',
                                                                       'mac_address': 'cc:be:59:cb:84:cd',
                                                                       'mtu': 2000,
                                                                       'name': '505/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920245',
                                                                       'out_octets': '2937774076535',
                                                                       'out_ucast_pkts': '2072964864',
                                                                       'poller_hash': 'NDJDMUYxcGVlcmludGY1MDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '505',
                                                                       'in_nucast_pkts': '9920245',
                                                                       'in_octets': '2937774076535',
                                                                       'in_ucast_pkts': '2072964864',
                                                                       'links': [       {       'in_nucast_pkts': '9920245',
                                                                                                'in_octets': '2937774076535',
                                                                                                'in_ucast_pkts': '2072964864',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '119159151689',
                                                                                                'out_ucast_pkts': '421968903',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '119159151689',
                                                                       'out_ucast_pkts': '421968903',
                                                                       'poller_hash': 'NDJDMUYxcGVlcmludGZwb24gNTA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '505',
                                                                                               'links': [       {       'connected_time': '8535253',
                                                                                                                        'fiber_length': 23696,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDMUYxcGVlcnBvbnBvbiA1MDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '23 Galena St Wellesley (Andrew Moore '
                                                 '10435707)',
                                         'poller_hash': 'NDJDMUYx',
                                         'serial_number': '42C1F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:84:ca',
                                         'uptime': '8535253'},
                                 {       'counter_source': '506/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f0:ba'],
                                                                       'id': '506/g1',
                                                                       'in_nucast_pkts': '71503',
                                                                       'in_octets': '212210761630',
                                                                       'in_ucast_pkts': '759090838',
                                                                       'mac_address': '44:65:7f:33:21:b5',
                                                                       'mtu': 2000,
                                                                       'name': '506/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530445',
                                                                       'out_octets': '3212157169876',
                                                                       'out_ucast_pkts': '2395653777',
                                                                       'poller_hash': 'NjQwQjUycGVlcmludGY1MDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '506',
                                                                       'in_nucast_pkts': '5530445',
                                                                       'in_octets': '3212157169876',
                                                                       'in_ucast_pkts': '2395653777',
                                                                       'links': [       {       'in_nucast_pkts': '5530445',
                                                                                                'in_octets': '3212157169876',
                                                                                                'in_ucast_pkts': '2395653777',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '71503',
                                                                                                'out_octets': '212210761630',
                                                                                                'out_ucast_pkts': '759090838',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '71503',
                                                                       'out_octets': '212210761630',
                                                                       'out_ucast_pkts': '759090838',
                                                                       'poller_hash': 'NjQwQjUycGVlcmludGZwb24gNTA2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '506',
                                                                                               'links': [       {       'connected_time': '8535099',
                                                                                                                        'fiber_length': 23831,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.328,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.974}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjUycGVlcnBvbnBvbiA1MDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '3701 NAFZIGER RD (ROPP COURTNEY)',
                                         'poller_hash': 'NjQwQjUy',
                                         'serial_number': '640B52',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:b4',
                                         'uptime': '8535099'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '507/g1',
                                                                       'name': '507/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDRUNDcGVlcmludGY1MDcvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '507',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDRUNDcGVlcmludGZwb24gNTA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '507',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRUNDcGVlcnBvbnBvbiA1MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '2082 NITHVIEW CT',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "507" '
                                                                  '(location: GPON '
                                                                  'port "5/7"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:28:00.540Z',
                                         'poller_hash': 'NDJDRUND',
                                         'serial_number': '42CECC',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '508/g1',
                                                                       'name': '508/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQjVDcGVlcmludGY1MDgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '508',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQjVDcGVlcmludGZwb24gNTA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '508',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjVDcGVlcnBvbnBvbiA1MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '106 GERBER MEADOWS DR (EDWARDS CORY)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "508" '
                                                                  '(location: GPON '
                                                                  'port "7/5"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-01-31T13:21:39.770Z',
                                         'poller_hash': 'NjQwQjVD',
                                         'serial_number': '640B5C',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '509/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '7c:5e:98:d5:73:21'],
                                                                       'id': '509/g1',
                                                                       'in_nucast_pkts': '78139',
                                                                       'in_octets': '49635459944',
                                                                       'in_ucast_pkts': '91629996',
                                                                       'mac_address': 'cc:be:59:cb:e7:c5',
                                                                       'mtu': 2000,
                                                                       'name': '509/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2632897',
                                                                       'out_octets': '341107588909',
                                                                       'out_ucast_pkts': '272540685',
                                                                       'poller_hash': 'NDJDRTUwcGVlcmludGY1MDkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '509',
                                                                       'in_nucast_pkts': '2632897',
                                                                       'in_octets': '341107588909',
                                                                       'in_ucast_pkts': '272540685',
                                                                       'links': [       {       'in_nucast_pkts': '2632897',
                                                                                                'in_octets': '341107588909',
                                                                                                'in_ucast_pkts': '272540685',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '78139',
                                                                                                'out_octets': '49635459944',
                                                                                                'out_ucast_pkts': '91629996',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '78139',
                                                                       'out_octets': '49635459944',
                                                                       'out_ucast_pkts': '91629996',
                                                                       'poller_hash': 'NDJDRTUwcGVlcmludGZwb24gNTA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '509',
                                                                                               'links': [       {       'connected_time': '2227736',
                                                                                                                        'fiber_length': 23298,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTUwcGVlcnBvbnBvbiA1MDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '130 SCHWEITZER CR (HILL JONATHAN)',
                                         'poller_hash': 'NDJDRTUw',
                                         'serial_number': '42CE50',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e7:c2',
                                         'uptime': '2227736'},
                                 {       'counter_source': '511/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d3:e1'],
                                                                       'id': '511/g1',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '167193262018',
                                                                       'in_ucast_pkts': '911443835',
                                                                       'mac_address': 'cc:be:59:cb:c5:8d',
                                                                       'mtu': 2000,
                                                                       'name': '511/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14628996',
                                                                       'out_octets': '3266445525640',
                                                                       'out_ucast_pkts': '2328600785',
                                                                       'poller_hash': 'NDJDQTA5cGVlcmludGY1MTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '511',
                                                                       'in_nucast_pkts': '14628996',
                                                                       'in_octets': '3266445525640',
                                                                       'in_ucast_pkts': '2328600785',
                                                                       'links': [       {       'in_nucast_pkts': '14628996',
                                                                                                'in_octets': '3266445525640',
                                                                                                'in_ucast_pkts': '2328600785',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '26',
                                                                                                'out_octets': '167193262018',
                                                                                                'out_ucast_pkts': '911443835',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '167193262018',
                                                                       'out_ucast_pkts': '911443835',
                                                                       'poller_hash': 'NDJDQTA5cGVlcmludGZwb24gNTEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '511',
                                                                                               'links': [       {       'connected_time': '12708337',
                                                                                                                        'fiber_length': 24765,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDQTA5cGVlcnBvbnBvbiA1MTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 Village Rd Wellesley (Linda '
                                                 'Hergott 10435445)',
                                         'poller_hash': 'NDJDQTA5',
                                         'serial_number': '42CA09',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:c5:8a',
                                         'uptime': '12708337'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '513/g1',
                                                                       'name': '513/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwQUZFcGVlcmludGY1MTMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '513',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NjQwQUZFcGVlcmludGZwb24gNTEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '513',
                                                                                               'links': [       {       'connected_time': '8535122',
                                                                                                                        'fiber_length': 23448,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.104,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.538}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUZFcGVlcnBvbnBvbiA1MTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '87 FERRIS DR (MACDONALD CALUM)',
                                         'poller_hash': 'NjQwQUZF',
                                         'serial_number': '640AFE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:0c',
                                         'uptime': '8535122'},
                                 {       'counter_source': '514/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:73:a7'],
                                                                       'id': '514/g1',
                                                                       'in_nucast_pkts': '25152',
                                                                       'in_octets': '27029650842',
                                                                       'in_ucast_pkts': '125478927',
                                                                       'mac_address': '44:65:7f:7b:c4:4c',
                                                                       'mtu': 2000,
                                                                       'name': '514/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1603040',
                                                                       'out_octets': '401618939442',
                                                                       'out_ucast_pkts': '338750864',
                                                                       'poller_hash': 'NUE2NjVGcGVlcmludGY1MTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '514',
                                                                       'in_nucast_pkts': '1603040',
                                                                       'in_octets': '401618939442',
                                                                       'in_ucast_pkts': '338750864',
                                                                       'links': [       {       'in_nucast_pkts': '1603040',
                                                                                                'in_octets': '401618939442',
                                                                                                'in_ucast_pkts': '338750864',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '25152',
                                                                                                'out_octets': '27029650842',
                                                                                                'out_ucast_pkts': '125478927',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '25152',
                                                                       'out_octets': '27029650842',
                                                                       'out_ucast_pkts': '125478927',
                                                                       'poller_hash': 'NUE2NjVGcGVlcmludGZwb24gNTE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '514',
                                                                                               'links': [       {       'connected_time': '1377493',
                                                                                                                        'fiber_length': 24807,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjVGcGVlcnBvbnBvbiA1MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '18 Parkview Dr Wellesley (Matthew '
                                                 'Jones 10435652)',
                                         'poller_hash': 'NUE2NjVG',
                                         'serial_number': '5A665F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:4c',
                                         'uptime': '1377493'},
                                 {       'counter_source': '516/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d3:c6'],
                                                                       'id': '516/g1',
                                                                       'in_nucast_pkts': '6397',
                                                                       'in_octets': '22256473742',
                                                                       'in_ucast_pkts': '93361267',
                                                                       'mac_address': 'cc:be:59:cb:bf:25',
                                                                       'mtu': 2000,
                                                                       'name': '516/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5370437',
                                                                       'out_octets': '579828243568',
                                                                       'out_ucast_pkts': '421201977',
                                                                       'poller_hash': 'NDJDOTNDcGVlcmludGY1MTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '516',
                                                                       'in_nucast_pkts': '5370437',
                                                                       'in_octets': '579828243568',
                                                                       'in_ucast_pkts': '421201977',
                                                                       'links': [       {       'in_nucast_pkts': '5370437',
                                                                                                'in_octets': '579828243568',
                                                                                                'in_ucast_pkts': '421201977',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6397',
                                                                                                'out_octets': '22256473742',
                                                                                                'out_ucast_pkts': '93361267',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6397',
                                                                       'out_octets': '22256473742',
                                                                       'out_ucast_pkts': '93361267',
                                                                       'poller_hash': 'NDJDOTNDcGVlcmludGZwb24gNTE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '516',
                                                                                               'links': [       {       'connected_time': '4667581',
                                                                                                                        'fiber_length': 25274,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOTNDcGVlcnBvbnBvbiA1MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '77 VILLAGE RD (KANE THOMAS(TOM))',
                                         'poller_hash': 'NDJDOTND',
                                         'serial_number': '42C93C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bf:22',
                                         'uptime': '4667581'},
                                 {       'counter_source': '517/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d3:ea'],
                                                                       'id': '517/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '437804056239',
                                                                       'in_ucast_pkts': '737270544',
                                                                       'mac_address': '44:65:7f:33:20:99',
                                                                       'mtu': 2000,
                                                                       'name': '517/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5410189',
                                                                       'out_octets': '649677100210',
                                                                       'out_ucast_pkts': '990510536',
                                                                       'poller_hash': 'NjQwQUM2cGVlcmludGY1MTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '517',
                                                                       'in_nucast_pkts': '5410189',
                                                                       'in_octets': '649677100210',
                                                                       'in_ucast_pkts': '990510536',
                                                                       'links': [       {       'in_nucast_pkts': '5410189',
                                                                                                'in_octets': '649677100210',
                                                                                                'in_ucast_pkts': '990510536',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '437804056239',
                                                                                                'out_ucast_pkts': '737270544',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '437804056239',
                                                                       'out_ucast_pkts': '737270544',
                                                                       'poller_hash': 'NjQwQUM2cGVlcmludGZwb24gNTE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '517',
                                                                                               'links': [       {       'connected_time': '8348530',
                                                                                                                        'fiber_length': 24745,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.25,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.908}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUM2cGVlcnBvbnBvbiA1MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '17 Parkview Dr Wellesley (Jo-Anne '
                                                 'Kropf 10435560)',
                                         'poller_hash': 'NjQwQUM2',
                                         'serial_number': '640AC6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:20:98',
                                         'uptime': '8348530'},
                                 {       'counter_source': '518/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:60:8c:03:51:4c'],
                                                                       'id': '518/g1',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '88317740425',
                                                                       'in_ucast_pkts': '385626709',
                                                                       'mac_address': '44:65:7f:7b:c2:f6',
                                                                       'mtu': 2000,
                                                                       'name': '518/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797775',
                                                                       'out_octets': '1345940157621',
                                                                       'out_ucast_pkts': '1099181791',
                                                                       'poller_hash': 'NUE2NUVEcGVlcmludGY1MTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '518',
                                                                       'in_nucast_pkts': '9797775',
                                                                       'in_octets': '1345940157621',
                                                                       'in_ucast_pkts': '1099181791',
                                                                       'links': [       {       'in_nucast_pkts': '9797775',
                                                                                                'in_octets': '1345940157621',
                                                                                                'in_ucast_pkts': '1099181791',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12',
                                                                                                'out_octets': '88317740425',
                                                                                                'out_ucast_pkts': '385626709',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '88317740425',
                                                                       'out_ucast_pkts': '385626709',
                                                                       'poller_hash': 'NUE2NUVEcGVlcmludGZwb24gNTE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '518',
                                                                                               'links': [       {       'connected_time': '8535402',
                                                                                                                        'fiber_length': 25145,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NUVEcGVlcnBvbnBvbiA1MTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '63 Village Rd Wellesley (Xinbin Zhou '
                                                 '10434842)',
                                         'poller_hash': 'NUE2NUVE',
                                         'serial_number': '5A65ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c2:f6',
                                         'uptime': '8535402'},
                                 {       'counter_source': '519/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '6c:cd:d6:f8:ee:7a'],
                                                                       'id': '519/g1',
                                                                       'in_nucast_pkts': '2802',
                                                                       'in_octets': '134284237971',
                                                                       'in_ucast_pkts': '465137315',
                                                                       'mac_address': '44:65:7f:33:20:97',
                                                                       'mtu': 2000,
                                                                       'name': '519/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530056',
                                                                       'out_octets': '1436274227490',
                                                                       'out_ucast_pkts': '1086021473',
                                                                       'poller_hash': 'NjQwQUM1cGVlcmludGY1MTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '519',
                                                                       'in_nucast_pkts': '5530056',
                                                                       'in_octets': '1436274227490',
                                                                       'in_ucast_pkts': '1086021473',
                                                                       'links': [       {       'in_nucast_pkts': '5530056',
                                                                                                'in_octets': '1436274227490',
                                                                                                'in_ucast_pkts': '1086021473',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2802',
                                                                                                'out_octets': '134284237971',
                                                                                                'out_ucast_pkts': '465137315',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2802',
                                                                       'out_octets': '134284237971',
                                                                       'out_ucast_pkts': '465137315',
                                                                       'poller_hash': 'NjQwQUM1cGVlcmludGZwb24gNTE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '519',
                                                                                               'links': [       {       'connected_time': '8535149',
                                                                                                                        'fiber_length': 23899,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.442,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.972}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUM1cGVlcnBvbnBvbiA1MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': "1015 Queen's Bush Wellesley (Scott "
                                                 'Kittell 10435598)',
                                         'poller_hash': 'NjQwQUM1',
                                         'serial_number': '640AC5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:20:96',
                                         'uptime': '8535149'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '520/g1',
                                                                       'name': '520/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'Mzc4ODNFcGVlcmludGY1MjAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '520',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'Mzc4ODNFcGVlcmludGZwb24gNTIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '520',
                                                                                               'links': [       {       'connected_time': '551821',
                                                                                                                        'fiber_length': 22447,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.148,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Mzc4ODNFcGVlcnBvbnBvbiA1MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '1172 Henry St Wellesley '
                                                 '(Futher-Franklin Funeral 10435038)',
                                         'poller_hash': 'Mzc4ODNF',
                                         'serial_number': '37883E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:be:a8:09',
                                         'uptime': '551821'},
                                 {       'counter_source': '521/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:38:03:55'],
                                                                       'id': '521/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '186389502',
                                                                       'in_ucast_pkts': '555692',
                                                                       'mac_address': '44:65:7f:33:0d:0f',
                                                                       'mtu': 2000,
                                                                       'name': '521/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530209',
                                                                       'out_octets': '617274363',
                                                                       'out_ucast_pkts': '1825065',
                                                                       'poller_hash': 'NjQwMTBFcGVlcmludGY1MjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '521',
                                                                       'in_nucast_pkts': '5530209',
                                                                       'in_octets': '617274363',
                                                                       'in_ucast_pkts': '1825065',
                                                                       'links': [       {       'in_nucast_pkts': '5530209',
                                                                                                'in_octets': '617274363',
                                                                                                'in_ucast_pkts': '1825065',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '186389502',
                                                                                                'out_ucast_pkts': '555692',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '186389502',
                                                                       'out_ucast_pkts': '555692',
                                                                       'poller_hash': 'NjQwMTBFcGVlcmludGZwb24gNTIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '521',
                                                                                               'links': [       {       'connected_time': '8535063',
                                                                                                                        'fiber_length': 25302,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.296,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.424}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwMTBFcGVlcnBvbnBvbiA1MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '107 Village Rd Wellesley (SHEWAN '
                                                 'KELLY)',
                                         'poller_hash': 'NjQwMTBF',
                                         'serial_number': '64010E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:0d:0e',
                                         'uptime': '8535063'},
                                 {       'counter_source': '523/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d1:e0'],
                                                                       'id': '523/g1',
                                                                       'in_nucast_pkts': '9',
                                                                       'in_octets': '196633559224',
                                                                       'in_ucast_pkts': '759786860',
                                                                       'mac_address': 'cc:be:59:cb:c3:e5',
                                                                       'mtu': 2000,
                                                                       'name': '523/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797944',
                                                                       'out_octets': '2433936600859',
                                                                       'out_ucast_pkts': '1917048963',
                                                                       'poller_hash': 'NDJDOUQ0cGVlcmludGY1MjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '523',
                                                                       'in_nucast_pkts': '9797944',
                                                                       'in_octets': '2433936600859',
                                                                       'in_ucast_pkts': '1917048963',
                                                                       'links': [       {       'in_nucast_pkts': '9797944',
                                                                                                'in_octets': '2433936600859',
                                                                                                'in_ucast_pkts': '1917048963',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9',
                                                                                                'out_octets': '196633559224',
                                                                                                'out_ucast_pkts': '759786860',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9',
                                                                       'out_octets': '196633559224',
                                                                       'out_ucast_pkts': '759786860',
                                                                       'poller_hash': 'NDJDOUQ0cGVlcmludGZwb24gNTIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '523',
                                                                                               'links': [       {       'connected_time': '8535201',
                                                                                                                        'fiber_length': 23404,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOUQ0cGVlcnBvbnBvbiA1MjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '28 Zinkann Cr Wellesley (Scott '
                                                 'Hache)',
                                         'poller_hash': 'NDJDOUQ0',
                                         'serial_number': '42C9D4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:c3:e2',
                                         'uptime': '8535201'},
                                 {       'counter_source': '524/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d4:b0'],
                                                                       'id': '524/g1',
                                                                       'in_nucast_pkts': '27',
                                                                       'in_octets': '307134767661',
                                                                       'in_ucast_pkts': '2296599863',
                                                                       'mac_address': 'cc:be:59:cb:ad:cd',
                                                                       'mtu': 2000,
                                                                       'name': '524/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36499690',
                                                                       'out_octets': '8954576515906',
                                                                       'out_ucast_pkts': '6363324909',
                                                                       'poller_hash': 'NDJDNzExcGVlcmludGY1MjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '524',
                                                                       'in_nucast_pkts': '36499690',
                                                                       'in_octets': '8954576515906',
                                                                       'in_ucast_pkts': '6363324909',
                                                                       'links': [       {       'in_nucast_pkts': '36499690',
                                                                                                'in_octets': '8954576515906',
                                                                                                'in_ucast_pkts': '6363324909',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27',
                                                                                                'out_octets': '307134767661',
                                                                                                'out_ucast_pkts': '2296599863',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27',
                                                                       'out_octets': '307134767661',
                                                                       'out_ucast_pkts': '2296599863',
                                                                       'poller_hash': 'NDJDNzExcGVlcmludGZwb24gNTI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '524',
                                                                                               'links': [       {       'connected_time': '64529577',
                                                                                                                        'fiber_length': 23566,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNzExcGVlcnBvbnBvbiA1MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3524A Nafziger Rd Wellesley (Jodi '
                                                 'Kipp 10435639)',
                                         'poller_hash': 'NDJDNzEx',
                                         'serial_number': '42C711',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ad:ca',
                                         'uptime': '64529577'},
                                 {       'counter_source': '526/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:a4:90:a1'],
                                                                       'id': '526/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '6610508974',
                                                                       'in_ucast_pkts': '36659849',
                                                                       'mac_address': 'cc:be:59:cb:6b:ed',
                                                                       'mtu': 2000,
                                                                       'name': '526/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715457',
                                                                       'out_octets': '184862946794',
                                                                       'out_ucast_pkts': '131484156',
                                                                       'poller_hash': 'NDJCRUQ1cGVlcmludGY1MjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '526',
                                                                       'in_nucast_pkts': '715457',
                                                                       'in_octets': '184862946794',
                                                                       'in_ucast_pkts': '131484156',
                                                                       'links': [       {       'in_nucast_pkts': '715457',
                                                                                                'in_octets': '184862946794',
                                                                                                'in_ucast_pkts': '131484156',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '6610508974',
                                                                                                'out_ucast_pkts': '36659849',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '6610508974',
                                                                       'out_ucast_pkts': '36659849',
                                                                       'poller_hash': 'NDJCRUQ1cGVlcmludGZwb24gNTI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '526',
                                                                                               'links': [       {       'connected_time': '611454',
                                                                                                                        'fiber_length': 19784,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRUQ1cGVlcnBvbnBvbiA1MjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4014 Andy St GadsHill (Megan Lunn '
                                                 '10435814)',
                                         'poller_hash': 'NDJCRUQ1',
                                         'serial_number': '42BED5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6b:ea',
                                         'uptime': '611454'},
                                 {       'counter_source': '527/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:c9:55'],
                                                                       'id': '527/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '3760461523',
                                                                       'in_ucast_pkts': '18361801',
                                                                       'mac_address': 'cc:be:59:ef:f7:2a',
                                                                       'mtu': 2000,
                                                                       'name': '527/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1339071',
                                                                       'out_octets': '70549520952',
                                                                       'out_ucast_pkts': '55798354',
                                                                       'poller_hash': 'NEQ0MjM3cGVlcmludGY1MjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '527',
                                                                       'in_nucast_pkts': '1339071',
                                                                       'in_octets': '70549520952',
                                                                       'in_ucast_pkts': '55798354',
                                                                       'links': [       {       'in_nucast_pkts': '1339071',
                                                                                                'in_octets': '70549520952',
                                                                                                'in_ucast_pkts': '55798354',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '3760461523',
                                                                                                'out_ucast_pkts': '18361801',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '3760461523',
                                                                       'out_ucast_pkts': '18361801',
                                                                       'poller_hash': 'NEQ0MjM3cGVlcmludGZwb24gNTI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '527',
                                                                                               'links': [       {       'connected_time': '1993839',
                                                                                                                        'fiber_length': 22664,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.094,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NEQ0MjM3cGVlcnBvbnBvbiA1Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '3648 Nafziger Rd Unit A Wellesley '
                                                 '(Eleanore Watters 10435900)',
                                         'poller_hash': 'NEQ0MjM3',
                                         'serial_number': '4D4237',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'cc:be:59:ef:f7:26',
                                         'uptime': '1993839'},
                                 {       'counter_source': '528/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '30:fd:38:f4:5f:dc'],
                                                                       'id': '528/g1',
                                                                       'in_nucast_pkts': '2384',
                                                                       'in_octets': '145409401582',
                                                                       'in_ucast_pkts': '542846153',
                                                                       'mac_address': '44:65:7f:33:22:1b',
                                                                       'mtu': 2000,
                                                                       'name': '528/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5601482',
                                                                       'out_octets': '2202292228204',
                                                                       'out_ucast_pkts': '1601830828',
                                                                       'poller_hash': 'NjQwQjg0cGVlcmludGY1MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '528',
                                                                       'in_nucast_pkts': '5601482',
                                                                       'in_octets': '2202292228204',
                                                                       'in_ucast_pkts': '1601830828',
                                                                       'links': [       {       'in_nucast_pkts': '5601482',
                                                                                                'in_octets': '2202292228204',
                                                                                                'in_ucast_pkts': '1601830828',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2384',
                                                                                                'out_octets': '145409401582',
                                                                                                'out_ucast_pkts': '542846153',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2384',
                                                                       'out_octets': '145409401582',
                                                                       'out_ucast_pkts': '542846153',
                                                                       'poller_hash': 'NjQwQjg0cGVlcmludGZwb24gNTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '528',
                                                                                               'links': [       {       'connected_time': '8535032',
                                                                                                                        'fiber_length': 22649,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.822,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.374}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjg0cGVlcnBvbnBvbiA1Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '3648 Nafziger Rd Unit B Wellesley '
                                                 '(Shawn Beggs 10435918)',
                                         'poller_hash': 'NjQwQjg0',
                                         'serial_number': '640B84',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:22:1a',
                                         'uptime': '8535032'},
                                 {       'counter_source': '529/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:d3:6c'],
                                                                       'id': '529/g1',
                                                                       'in_nucast_pkts': '6',
                                                                       'in_octets': '75996546078',
                                                                       'in_ucast_pkts': '404755094',
                                                                       'mac_address': 'cc:be:59:cc:04:6d',
                                                                       'mtu': 2000,
                                                                       'name': '529/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9870781',
                                                                       'out_octets': '1267066622970',
                                                                       'out_ucast_pkts': '898055207',
                                                                       'poller_hash': 'NDJEMUU1cGVlcmludGY1MjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '529',
                                                                       'in_nucast_pkts': '9870781',
                                                                       'in_octets': '1267066622970',
                                                                       'in_ucast_pkts': '898055207',
                                                                       'links': [       {       'in_nucast_pkts': '9870781',
                                                                                                'in_octets': '1267066622970',
                                                                                                'in_ucast_pkts': '898055207',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6',
                                                                                                'out_octets': '75996546078',
                                                                                                'out_ucast_pkts': '404755094',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6',
                                                                       'out_octets': '75996546078',
                                                                       'out_ucast_pkts': '404755094',
                                                                       'poller_hash': 'NDJEMUU1cGVlcmludGZwb24gNTI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '529',
                                                                                               'links': [       {       'connected_time': '8529309',
                                                                                                                        'fiber_length': 22770,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJEMUU1cGVlcnBvbnBvbiA1Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3648 Nafziger Rd Wellesley (Linda '
                                                 'Vincent 10435908)',
                                         'poller_hash': 'NDJEMUU1',
                                         'serial_number': '42D1E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cc:04:6a',
                                         'uptime': '8529309'},
                                 {       'counter_source': '532/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:71:39'],
                                                                       'id': '532/g1',
                                                                       'in_nucast_pkts': '266844',
                                                                       'in_octets': '82724726694',
                                                                       'in_ucast_pkts': '617229582',
                                                                       'mac_address': '48:77:46:a2:1b:e0',
                                                                       'mtu': 2000,
                                                                       'name': '532/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10268457',
                                                                       'out_octets': '3625200318434',
                                                                       'out_ucast_pkts': '2461714645',
                                                                       'poller_hash': 'NzFBNEFBcGVlcmludGY1MzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '532',
                                                                       'in_nucast_pkts': '10268457',
                                                                       'in_octets': '3625200318434',
                                                                       'in_ucast_pkts': '2461714645',
                                                                       'links': [       {       'in_nucast_pkts': '10268457',
                                                                                                'in_octets': '3625200318434',
                                                                                                'in_ucast_pkts': '2461714645',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '266844',
                                                                                                'out_octets': '82724726694',
                                                                                                'out_ucast_pkts': '617229582',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '266844',
                                                                       'out_octets': '82724726694',
                                                                       'out_ucast_pkts': '617229582',
                                                                       'poller_hash': 'NzFBNEFBcGVlcmludGZwb24gNTMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '532',
                                                                                               'links': [       {       'connected_time': '15182484',
                                                                                                                        'fiber_length': 23860,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.332,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFBNEFBcGVlcnBvbnBvbiA1MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '83 GALENA ST (RUBY MYRON)',
                                         'poller_hash': 'NzFBNEFB',
                                         'serial_number': '71A4AA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a2:1b:dc',
                                         'uptime': '15182484'},
                                 {       'counter_source': '533/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '58:cb:52:de:3d:2e'],
                                                                       'id': '533/g1',
                                                                       'in_nucast_pkts': '2051',
                                                                       'in_octets': '147722978918',
                                                                       'in_ucast_pkts': '481431455',
                                                                       'mac_address': '44:65:7f:33:21:57',
                                                                       'mtu': 2000,
                                                                       'name': '533/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2460629',
                                                                       'out_octets': '2226184304524',
                                                                       'out_ucast_pkts': '1709027558',
                                                                       'poller_hash': 'NjQwQjIzcGVlcmludGY1MzMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '533',
                                                                       'in_nucast_pkts': '2460629',
                                                                       'in_octets': '2226184304524',
                                                                       'in_ucast_pkts': '1709027558',
                                                                       'links': [       {       'in_nucast_pkts': '2460629',
                                                                                                'in_octets': '2226184304524',
                                                                                                'in_ucast_pkts': '1709027558',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2051',
                                                                                                'out_octets': '147722978918',
                                                                                                'out_ucast_pkts': '481431455',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2051',
                                                                       'out_octets': '147722978918',
                                                                       'out_ucast_pkts': '481431455',
                                                                       'poller_hash': 'NjQwQjIzcGVlcmludGZwb24gNTMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '533',
                                                                                               'links': [       {       'connected_time': '3620157',
                                                                                                                        'fiber_length': 23608,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.71,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.232}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjIzcGVlcnBvbnBvbiA1MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '35 Ferris Dr Wellesley (Natalie '
                                                 'Parco 10436039)',
                                         'poller_hash': 'NjQwQjIz',
                                         'serial_number': '640B23',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:56',
                                         'uptime': '3620157'},
                                 {       'counter_source': '534/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '534/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '365598521716',
                                                                       'in_ucast_pkts': '1389965126',
                                                                       'mac_address': 'ec:4f:82:a1:ae:b3',
                                                                       'name': '534/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '6044509892618',
                                                                       'out_ucast_pkts': '4468657270',
                                                                       'poller_hash': 'MzQ4MTA4cGVlcmludGY1MzQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '534',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '6044509892618',
                                                                       'in_ucast_pkts': '4468657270',
                                                                       'links': [       {       'in_nucast_pkts': '117',
                                                                                                'in_octets': '6044509892618',
                                                                                                'in_ucast_pkts': '4468657270',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '365598521716',
                                                                                                'out_ucast_pkts': '1389965126',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '365598521716',
                                                                       'out_ucast_pkts': '1389965126',
                                                                       'poller_hash': 'MzQ4MTA4cGVlcmludGZwb24gNTM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '534',
                                                                                               'links': [       {       'connected_time': '335345',
                                                                                                                        'fiber_length': 17679,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzQ4MTA4cGVlcnBvbnBvbiA1MzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '2005A Perth Line 56 Wellesley (Emily '
                                                 'Lindner 10436103)',
                                         'poller_hash': 'MzQ4MTA4',
                                         'serial_number': '348108',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:a1:ae:b3',
                                         'uptime': '335345'},
                                 {       'counter_source': '536/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '536/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '260350626491',
                                                                       'in_ucast_pkts': '878305827',
                                                                       'mac_address': 'd0:76:8f:aa:4b:d4',
                                                                       'name': '536/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '79',
                                                                       'out_octets': '6340607094072',
                                                                       'out_ucast_pkts': '4591639392',
                                                                       'poller_hash': 'OTU2RUEwcGVlcmludGY1MzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '536',
                                                                       'in_nucast_pkts': '79',
                                                                       'in_octets': '6340607094072',
                                                                       'in_ucast_pkts': '4591639392',
                                                                       'links': [       {       'in_nucast_pkts': '79',
                                                                                                'in_octets': '6340607094072',
                                                                                                'in_ucast_pkts': '4591639392',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '260350626491',
                                                                                                'out_ucast_pkts': '878305827',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '260350626491',
                                                                       'out_ucast_pkts': '878305827',
                                                                       'poller_hash': 'OTU2RUEwcGVlcmludGZwb24gNTM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '536',
                                                                                               'links': [       {       'connected_time': '7928912',
                                                                                                                        'fiber_length': 17168,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RUEwcGVlcnBvbnBvbiA1MzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2080 Nithview Ct Nithberg (Robert '
                                                 'Witzel 10436127)',
                                         'poller_hash': 'OTU2RUEw',
                                         'serial_number': '956EA0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:4b:d4',
                                         'uptime': '7928912'},
                                 {       'counter_source': '537/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'bc:0f:9a:70:bc:6b'],
                                                                       'id': '537/g1',
                                                                       'in_nucast_pkts': '14793',
                                                                       'in_octets': '29488289180',
                                                                       'in_ucast_pkts': '91220101',
                                                                       'mac_address': '44:65:7f:33:20:fd',
                                                                       'mtu': 2000,
                                                                       'name': '537/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530351',
                                                                       'out_octets': '573389722272',
                                                                       'out_ucast_pkts': '418160510',
                                                                       'poller_hash': 'NjQwQUY2cGVlcmludGY1MzcvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '537',
                                                                       'in_nucast_pkts': '5530351',
                                                                       'in_octets': '573389722272',
                                                                       'in_ucast_pkts': '418160510',
                                                                       'links': [       {       'in_nucast_pkts': '5530351',
                                                                                                'in_octets': '573389722272',
                                                                                                'in_ucast_pkts': '418160510',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14793',
                                                                                                'out_octets': '29488289180',
                                                                                                'out_ucast_pkts': '91220101',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14793',
                                                                       'out_octets': '29488289180',
                                                                       'out_ucast_pkts': '91220101',
                                                                       'poller_hash': 'NjQwQUY2cGVlcmludGZwb24gNTM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '537',
                                                                                               'links': [       {       'connected_time': '8535118',
                                                                                                                        'fiber_length': 23437,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.31,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.236}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUY2cGVlcnBvbnBvbiA1Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1023 Maple Leaf Lane Wellesley '
                                                 '(MARTIN LUCILLE)',
                                         'poller_hash': 'NjQwQUY2',
                                         'serial_number': '640AF6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:20:fc',
                                         'uptime': '8535118'},
                                 {       'counter_source': '538/F1+538/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:30:b2'],
                                                                       'id': '538/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1925483355',
                                                                       'in_ucast_pkts': '11678940',
                                                                       'name': '538/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '14565506893',
                                                                       'out_octets': '19375798524181',
                                                                       'out_ucast_pkts': '9118291',
                                                                       'poller_hash': 'NzNBQzExcGVlcmludGY1MzgvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '538/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '93993484566',
                                                                       'in_ucast_pkts': '497829739',
                                                                       'mac_address': '48:77:46:47:07:f6',
                                                                       'name': '538/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '134',
                                                                       'out_octets': '2246702498502',
                                                                       'out_ucast_pkts': '1685720689',
                                                                       'poller_hash': 'NzNBQzExcGVlcmludGY1MzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '538',
                                                                       'in_nucast_pkts': '14565507027',
                                                                       'in_octets': '21622501022683',
                                                                       'in_ucast_pkts': '1694838980',
                                                                       'links': [       {       'in_nucast_pkts': '14565507027',
                                                                                                'in_octets': '21622501022683',
                                                                                                'in_ucast_pkts': '1694838980',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '95918967921',
                                                                                                'out_ucast_pkts': '509508679',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '95918967921',
                                                                       'out_ucast_pkts': '509508679',
                                                                       'poller_hash': 'NzNBQzExcGVlcmludGZwb24gNTM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '538',
                                                                                               'links': [       {       'connected_time': '8535324',
                                                                                                                        'fiber_length': 15730,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzNBQzExcGVlcnBvbnBvbiA1Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '2184 Perth Line Wellesley (Alex '
                                                 'Albrecht 10436172)',
                                         'poller_hash': 'NzNBQzEx',
                                         'serial_number': '73AC11',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:47:07:f6',
                                         'uptime': '8535324'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '539/g1',
                                                                       'name': '539/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2NUVCcGVlcmludGY1MzkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '539',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2NUVCcGVlcmludGZwb24gNTM5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '539',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NUVCcGVlcnBvbnBvbiA1Mzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'MCCL U6 Test',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "539" '
                                                                  '(location: GPON '
                                                                  'port "7/8"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-10-01T19:08:34.460Z',
                                         'poller_hash': 'NUE2NUVC',
                                         'serial_number': '5A65EB',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '540/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '540/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '99924679246',
                                                                       'in_ucast_pkts': '929818507',
                                                                       'mac_address': '00:06:31:e9:01:3f',
                                                                       'name': '540/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '76',
                                                                       'out_octets': '2736161360369',
                                                                       'out_ucast_pkts': '1974065566',
                                                                       'poller_hash': 'MUQwNjgxcGVlcmludGY1NDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '540',
                                                                       'in_nucast_pkts': '76',
                                                                       'in_octets': '2736161360369',
                                                                       'in_ucast_pkts': '1974065566',
                                                                       'links': [       {       'in_nucast_pkts': '76',
                                                                                                'in_octets': '2736161360369',
                                                                                                'in_ucast_pkts': '1974065566',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '99924679246',
                                                                                                'out_ucast_pkts': '929818507',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '99924679246',
                                                                       'out_ucast_pkts': '929818507',
                                                                       'poller_hash': 'MUQwNjgxcGVlcmludGZwb24gNTQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '540',
                                                                                               'links': [       {       'connected_time': '6865721',
                                                                                                                        'fiber_length': 23173,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MUQwNjgxcGVlcnBvbnBvbiA1NDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844G-1',
                                         'name': '1 PONDVIEW UNIT 102 (SHANTZ ROSS '
                                                 '10436359)',
                                         'poller_hash': 'MUQwNjgx',
                                         'serial_number': '1D0681',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '00:06:31:e9:01:3f',
                                         'uptime': '6865721'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '541/g1',
                                                                       'name': '541/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkFDNDhEcGVlcmludGY1NDEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '541',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MkFDNDhEcGVlcmludGZwb24gNTQx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '541',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFDNDhEcGVlcnBvbnBvbiA1NDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'test (Ryan)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "541" '
                                                                  '(location: GPON '
                                                                  'port "1/5"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:26:04.520Z',
                                         'poller_hash': 'MkFDNDhE',
                                         'serial_number': '2AC48D',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '542/F1+542/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:2e:e4',
                                                                                                        '00:03:e6:dd:31:60',
                                                                                                        '00:03:e6:eb:49:1c'],
                                                                       'id': '542/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6583004101',
                                                                       'in_ucast_pkts': '33353110',
                                                                       'name': '542/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '32137639723',
                                                                       'out_octets': '43393568448662',
                                                                       'out_ucast_pkts': '2909715',
                                                                       'poller_hash': 'ODY5MzlFcGVlcmludGY1NDIvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '542/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '615994636467',
                                                                       'in_ucast_pkts': '1114429019',
                                                                       'mac_address': 'd0:76:8f:08:15:43',
                                                                       'name': '542/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '34897',
                                                                       'out_octets': '4515095571882',
                                                                       'out_ucast_pkts': '3383351573',
                                                                       'poller_hash': 'ODY5MzlFcGVlcmludGY1NDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '542',
                                                                       'in_nucast_pkts': '32137674620',
                                                                       'in_octets': '47908664020544',
                                                                       'in_ucast_pkts': '3386261288',
                                                                       'links': [       {       'in_nucast_pkts': '32137674620',
                                                                                                'in_octets': '47908664020544',
                                                                                                'in_ucast_pkts': '3386261288',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '622577640568',
                                                                                                'out_ucast_pkts': '1147782129',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '622577640568',
                                                                       'out_ucast_pkts': '1147782129',
                                                                       'poller_hash': 'ODY5MzlFcGVlcmludGZwb24gNTQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '542',
                                                                                               'links': [       {       'connected_time': '6471760',
                                                                                                                        'fiber_length': 37658,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzlFcGVlcnBvbnBvbiA1NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2400 Notre Dame Dr. (Russ& Linda '
                                                 'Straus)',
                                         'poller_hash': 'ODY5MzlF',
                                         'serial_number': '86939E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:43',
                                         'uptime': '6471760'},
                                 {       'counter_source': '543/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ed:4a'],
                                                                       'id': '543/g1',
                                                                       'in_nucast_pkts': '61961',
                                                                       'in_octets': '148516955878',
                                                                       'in_ucast_pkts': '230039591',
                                                                       'mac_address': 'd0:76:8f:08:19:b6',
                                                                       'mtu': 2000,
                                                                       'name': '543/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4003300',
                                                                       'out_octets': '694327705555',
                                                                       'out_ucast_pkts': '560619580',
                                                                       'poller_hash': 'ODY5NDJDcGVlcmludGY1NDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '543',
                                                                       'in_nucast_pkts': '4003300',
                                                                       'in_octets': '694327705555',
                                                                       'in_ucast_pkts': '560619580',
                                                                       'links': [       {       'in_nucast_pkts': '4003300',
                                                                                                'in_octets': '694327705555',
                                                                                                'in_ucast_pkts': '560619580',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '61961',
                                                                                                'out_octets': '148516955878',
                                                                                                'out_ucast_pkts': '230039591',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61961',
                                                                       'out_octets': '148516955878',
                                                                       'out_ucast_pkts': '230039591',
                                                                       'poller_hash': 'ODY5NDJDcGVlcmludGZwb24gNTQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '543',
                                                                                               'links': [       {       'connected_time': '3422392',
                                                                                                                        'fiber_length': 37809,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDJDcGVlcnBvbnBvbiA1NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1836 CARMEL-KOCH RD (STRAUS THOMAS & '
                                                 'STACY 10436566)',
                                         'poller_hash': 'ODY5NDJD',
                                         'serial_number': '86942C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:b3',
                                         'uptime': '3422392'},
                                 {       'counter_source': '546/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d5:fb'],
                                                                       'id': '546/g1',
                                                                       'in_nucast_pkts': '150227',
                                                                       'in_octets': '56748903794',
                                                                       'in_ucast_pkts': '113862948',
                                                                       'mac_address': 'd0:76:8f:aa:48:4f',
                                                                       'mtu': 2000,
                                                                       'name': '546/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797931',
                                                                       'out_octets': '765751137741',
                                                                       'out_ucast_pkts': '594122465',
                                                                       'poller_hash': 'OTU2RTJGcGVlcmludGY1NDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '546',
                                                                       'in_nucast_pkts': '9797931',
                                                                       'in_octets': '765751137741',
                                                                       'in_ucast_pkts': '594122465',
                                                                       'links': [       {       'in_nucast_pkts': '9797931',
                                                                                                'in_octets': '765751137741',
                                                                                                'in_ucast_pkts': '594122465',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '150227',
                                                                                                'out_octets': '56748903794',
                                                                                                'out_ucast_pkts': '113862948',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150227',
                                                                       'out_octets': '56748903794',
                                                                       'out_ucast_pkts': '113862948',
                                                                       'poller_hash': 'OTU2RTJGcGVlcmludGZwb24gNTQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '546',
                                                                                               'links': [       {       'connected_time': '8535325',
                                                                                                                        'fiber_length': 24910,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTJGcGVlcnBvbnBvbiA1NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '126 Brown St (Will Bender)',
                                         'poller_hash': 'OTU2RTJG',
                                         'serial_number': '956E2F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:48:4c',
                                         'uptime': '8535325'},
                                 {       'counter_source': '547/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:75:6f'],
                                                                       'id': '547/g1',
                                                                       'in_nucast_pkts': '160896',
                                                                       'in_octets': '70798130485',
                                                                       'in_ucast_pkts': '391720418',
                                                                       'mac_address': '60:db:98:4e:6e:df',
                                                                       'mtu': 2000,
                                                                       'name': '547/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878152',
                                                                       'out_octets': '2483626511574',
                                                                       'out_ucast_pkts': '1761111628',
                                                                       'poller_hash': 'QTlDMzlCcGVlcmludGY1NDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '547',
                                                                       'in_nucast_pkts': '9878152',
                                                                       'in_octets': '2483626511574',
                                                                       'in_ucast_pkts': '1761111628',
                                                                       'links': [       {       'in_nucast_pkts': '9878152',
                                                                                                'in_octets': '2483626511574',
                                                                                                'in_ucast_pkts': '1761111628',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160896',
                                                                                                'out_octets': '70798130485',
                                                                                                'out_ucast_pkts': '391720418',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160896',
                                                                       'out_octets': '70798130485',
                                                                       'out_ucast_pkts': '391720418',
                                                                       'poller_hash': 'QTlDMzlCcGVlcmludGZwb24gNTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '547',
                                                                                               'links': [       {       'connected_time': '8535197',
                                                                                                                        'fiber_length': 24959,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMzlCcGVlcnBvbnBvbiA1NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 SHADYWOOD CT WELLESLEY (BROWN '
                                                 'SHANE)',
                                         'poller_hash': 'QTlDMzlC',
                                         'serial_number': 'A9C39B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6e:dc',
                                         'uptime': '8535197'},
                                 {       'counter_source': '116022/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:f8:b3:f5:9c:29'],
                                                                       'id': '116022/g1',
                                                                       'in_nucast_pkts': '774',
                                                                       'in_octets': '4786217950',
                                                                       'in_ucast_pkts': '8663243',
                                                                       'mac_address': 'ec:4f:82:63:5a:e7',
                                                                       'mtu': 2000,
                                                                       'name': '116022/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715461',
                                                                       'out_octets': '21356643371',
                                                                       'out_ucast_pkts': '17496761',
                                                                       'poller_hash': 'MkJENUIycGVlcmludGYxMTYwMjIvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '116022',
                                                                       'in_nucast_pkts': '715461',
                                                                       'in_octets': '21356643371',
                                                                       'in_ucast_pkts': '17496761',
                                                                       'links': [       {       'in_nucast_pkts': '715461',
                                                                                                'in_octets': '21356643371',
                                                                                                'in_ucast_pkts': '17496761',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '774',
                                                                                                'out_octets': '4786217950',
                                                                                                'out_ucast_pkts': '8663243',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '774',
                                                                       'out_octets': '4786217950',
                                                                       'out_ucast_pkts': '8663243',
                                                                       'poller_hash': 'MkJENUIycGVlcmludGZwb24gMTE2MDIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '116022',
                                                                                               'links': [       {       'connected_time': '611462',
                                                                                                                        'fiber_length': 11812,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkJENUIycGVlcnBvbnBvbiAxMTYwMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801G',
                                         'name': '5597 PERTH LINE 56A (CHOWN '
                                                 'CHRISTOPHER)',
                                         'poller_hash': 'MkJENUIy',
                                         'serial_number': '2BD5B2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:63:5a:e7',
                                         'uptime': '611462'},
                                 {       'counter_source': '137401/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '58:d9:d5:f9:2f:61'],
                                                                       'id': '137401/g1',
                                                                       'in_nucast_pkts': '304',
                                                                       'in_octets': '263002077123',
                                                                       'in_ucast_pkts': '1468243754',
                                                                       'mac_address': 'cc:be:59:cb:bd:65',
                                                                       'mtu': 2000,
                                                                       'name': '137401/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797587',
                                                                       'out_octets': '8407183094963',
                                                                       'out_ucast_pkts': '5933143119',
                                                                       'poller_hash': 'NDJDOTA0cGVlcmludGYxMzc0MDEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137401',
                                                                       'in_nucast_pkts': '9797587',
                                                                       'in_octets': '8407183094963',
                                                                       'in_ucast_pkts': '5933143119',
                                                                       'links': [       {       'in_nucast_pkts': '9797587',
                                                                                                'in_octets': '8407183094963',
                                                                                                'in_ucast_pkts': '5933143119',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '304',
                                                                                                'out_octets': '263002077123',
                                                                                                'out_ucast_pkts': '1468243754',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '304',
                                                                       'out_octets': '263002077123',
                                                                       'out_ucast_pkts': '1468243754',
                                                                       'poller_hash': 'NDJDOTA0cGVlcmludGZwb24gMTM3NDAx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137401',
                                                                                               'links': [       {       'connected_time': '8535250',
                                                                                                                        'fiber_length': 24658,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOTA0cGVlcnBvbnBvbiAxMzc0MDE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1116 QUEEN'S BUSH RD (COOK DANIEL)",
                                         'poller_hash': 'NDJDOTA0',
                                         'serial_number': '42C904',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bd:62',
                                         'uptime': '8535250'},
                                 {       'counter_source': '137411/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e2:eb'],
                                                                       'id': '137411/g1',
                                                                       'in_nucast_pkts': '924845',
                                                                       'in_octets': '613796995284',
                                                                       'in_ucast_pkts': '2746501605',
                                                                       'mac_address': 'cc:be:59:cb:bd:55',
                                                                       'mtu': 2000,
                                                                       'name': '137411/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62515985',
                                                                       'out_octets': '10127946422436',
                                                                       'out_ucast_pkts': '7517149569',
                                                                       'poller_hash': 'NDJDOTAycGVlcmludGYxMzc0MTEvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137411',
                                                                       'in_nucast_pkts': '62515985',
                                                                       'in_octets': '10127946422436',
                                                                       'in_ucast_pkts': '7517149569',
                                                                       'links': [       {       'in_nucast_pkts': '62515985',
                                                                                                'in_octets': '10127946422436',
                                                                                                'in_ucast_pkts': '7517149569',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '924845',
                                                                                                'out_octets': '613796995284',
                                                                                                'out_ucast_pkts': '2746501605',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '924845',
                                                                       'out_octets': '613796995284',
                                                                       'out_ucast_pkts': '2746501605',
                                                                       'poller_hash': 'NDJDOTAycGVlcmludGZwb24gMTM3NDEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137411',
                                                                                               'links': [       {       'connected_time': '21886447',
                                                                                                                        'fiber_length': 24240,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOTAycGVlcnBvbnBvbiAxMzc0MTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '54 PARKVIEW DR (FROKLAGE HARRY)',
                                         'poller_hash': 'NDJDOTAy',
                                         'serial_number': '42C902',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bd:52',
                                         'uptime': '21886447'},
                                 {       'counter_source': '137547/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '137547/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '89321075993',
                                                                       'in_ucast_pkts': '331453889',
                                                                       'mac_address': '48:77:46:47:06:46',
                                                                       'name': '137547/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '57',
                                                                       'out_octets': '1207739454853',
                                                                       'out_ucast_pkts': '1027257800',
                                                                       'poller_hash': 'NzNBQkUxcGVlcmludGYxMzc1NDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '137547',
                                                                       'in_nucast_pkts': '57',
                                                                       'in_octets': '1207739454853',
                                                                       'in_ucast_pkts': '1027257800',
                                                                       'links': [       {       'in_nucast_pkts': '57',
                                                                                                'in_octets': '1207739454853',
                                                                                                'in_ucast_pkts': '1027257800',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '89321075993',
                                                                                                'out_ucast_pkts': '331453889',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '89321075993',
                                                                       'out_ucast_pkts': '331453889',
                                                                       'poller_hash': 'NzNBQkUxcGVlcmludGZwb24gMTM3NTQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137547',
                                                                                               'links': [       {       'connected_time': '8535295',
                                                                                                                        'fiber_length': 23182,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzNBQkUxcGVlcnBvbnBvbiAxMzc1NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '1 POND VIEW DR (PENTELBURY LAURENE)',
                                         'poller_hash': 'NzNBQkUx',
                                         'serial_number': '73ABE1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:47:06:46',
                                         'uptime': '8535295'},
                                 {       'counter_source': '137614/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:62:25'],
                                                                       'id': '137614/g1',
                                                                       'in_nucast_pkts': '52',
                                                                       'in_octets': '132985579019',
                                                                       'in_ucast_pkts': '300417286',
                                                                       'mac_address': '44:65:7f:33:21:35',
                                                                       'mtu': 2000,
                                                                       'name': '137614/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5651745',
                                                                       'out_octets': '1404222022465',
                                                                       'out_ucast_pkts': '1135535621',
                                                                       'poller_hash': 'NjQwQjEycGVlcmludGYxMzc2MTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137614',
                                                                       'in_nucast_pkts': '5651745',
                                                                       'in_octets': '1404222022465',
                                                                       'in_ucast_pkts': '1135535621',
                                                                       'links': [       {       'in_nucast_pkts': '5651745',
                                                                                                'in_octets': '1404222022465',
                                                                                                'in_ucast_pkts': '1135535621',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '52',
                                                                                                'out_octets': '132985579019',
                                                                                                'out_ucast_pkts': '300417286',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '52',
                                                                       'out_octets': '132985579019',
                                                                       'out_ucast_pkts': '300417286',
                                                                       'poller_hash': 'NjQwQjEycGVlcmludGZwb24gMTM3NjE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137614',
                                                                                               'links': [       {       'connected_time': '8535105',
                                                                                                                        'fiber_length': 24863,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.304,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.89}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjEycGVlcnBvbnBvbiAxMzc2MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '15 NITHBANK RD (ELLIOTT ADRIENNE)',
                                         'poller_hash': 'NjQwQjEy',
                                         'serial_number': '640B12',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:34',
                                         'uptime': '8535105'},
                                 {       'counter_source': '137617/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:52:80'],
                                                                       'id': '137617/g1',
                                                                       'in_nucast_pkts': '60',
                                                                       'in_octets': '142370251648',
                                                                       'in_ucast_pkts': '665394214',
                                                                       'mac_address': '44:65:7f:33:20:c5',
                                                                       'mtu': 2000,
                                                                       'name': '137617/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5529949',
                                                                       'out_octets': '2626074258277',
                                                                       'out_ucast_pkts': '2019977505',
                                                                       'poller_hash': 'NjQwQURDcGVlcmludGYxMzc2MTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137617',
                                                                       'in_nucast_pkts': '5529949',
                                                                       'in_octets': '2626074258277',
                                                                       'in_ucast_pkts': '2019977505',
                                                                       'links': [       {       'in_nucast_pkts': '5529949',
                                                                                                'in_octets': '2626074258277',
                                                                                                'in_ucast_pkts': '2019977505',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '60',
                                                                                                'out_octets': '142370251648',
                                                                                                'out_ucast_pkts': '665394214',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '60',
                                                                       'out_octets': '142370251648',
                                                                       'out_ucast_pkts': '665394214',
                                                                       'poller_hash': 'NjQwQURDcGVlcmludGZwb24gMTM3NjE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137617',
                                                                                               'links': [       {       'connected_time': '8535117',
                                                                                                                        'fiber_length': 24617,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.142,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.26}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQURDcGVlcnBvbnBvbiAxMzc2MTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': "1122 QUEEN'S BUSH RD (DIETZ JON W)",
                                         'poller_hash': 'NjQwQURD',
                                         'serial_number': '640ADC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:20:c4',
                                         'uptime': '8535117'},
                                 {       'counter_source': '137630/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:37:fe:60'],
                                                                       'id': '137630/g1',
                                                                       'in_nucast_pkts': '19',
                                                                       'in_octets': '78183118390',
                                                                       'in_ucast_pkts': '219193972',
                                                                       'mac_address': 'cc:be:59:cb:ba:15',
                                                                       'mtu': 2000,
                                                                       'name': '137630/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797190',
                                                                       'out_octets': '752956627701',
                                                                       'out_ucast_pkts': '584531484',
                                                                       'poller_hash': 'NDJDODlBcGVlcmludGYxMzc2MzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137630',
                                                                       'in_nucast_pkts': '9797190',
                                                                       'in_octets': '752956627701',
                                                                       'in_ucast_pkts': '584531484',
                                                                       'links': [       {       'in_nucast_pkts': '9797190',
                                                                                                'in_octets': '752956627701',
                                                                                                'in_ucast_pkts': '584531484',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19',
                                                                                                'out_octets': '78183118390',
                                                                                                'out_ucast_pkts': '219193972',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19',
                                                                       'out_octets': '78183118390',
                                                                       'out_ucast_pkts': '219193972',
                                                                       'poller_hash': 'NDJDODlBcGVlcmludGZwb24gMTM3NjMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137630',
                                                                                               'links': [       {       'connected_time': '8535271',
                                                                                                                        'fiber_length': 24674,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDODlBcGVlcnBvbnBvbiAxMzc2MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '139 GERBER MEADOWS DR (DIETRICH '
                                                 'VALERIE A)',
                                         'poller_hash': 'NDJDODlB',
                                         'serial_number': '42C89A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:ba:12',
                                         'uptime': '8535271'},
                                 {       'counter_source': '137666/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b0:95:75:b3:56:8e'],
                                                                       'id': '137666/g1',
                                                                       'in_nucast_pkts': '36',
                                                                       'in_octets': '288978149469',
                                                                       'in_ucast_pkts': '540920784',
                                                                       'mac_address': '44:65:7f:33:21:45',
                                                                       'mtu': 2000,
                                                                       'name': '137666/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5611055',
                                                                       'out_octets': '1282082640440',
                                                                       'out_ucast_pkts': '1073003968',
                                                                       'poller_hash': 'NjQwQjFBcGVlcmludGYxMzc2NjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137666',
                                                                       'in_nucast_pkts': '5611055',
                                                                       'in_octets': '1282082640440',
                                                                       'in_ucast_pkts': '1073003968',
                                                                       'links': [       {       'in_nucast_pkts': '5611055',
                                                                                                'in_octets': '1282082640440',
                                                                                                'in_ucast_pkts': '1073003968',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '36',
                                                                                                'out_octets': '288978149469',
                                                                                                'out_ucast_pkts': '540920784',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36',
                                                                       'out_octets': '288978149469',
                                                                       'out_ucast_pkts': '540920784',
                                                                       'poller_hash': 'NjQwQjFBcGVlcmludGZwb24gMTM3NjY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137666',
                                                                                               'links': [       {       'connected_time': '8535068',
                                                                                                                        'fiber_length': 22271,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.956,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.41}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjFBcGVlcnBvbnBvbiAxMzc2NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '24 WATER ST (PETERS BRYAN)',
                                         'poller_hash': 'NjQwQjFB',
                                         'serial_number': '640B1A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:44',
                                         'uptime': '8535068'},
                                 {       'counter_source': '137683/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:54:66'],
                                                                       'id': '137683/g1',
                                                                       'in_nucast_pkts': '9615226',
                                                                       'in_octets': '184311563745',
                                                                       'in_ucast_pkts': '345628142',
                                                                       'mac_address': 'cc:be:59:cb:6c:4d',
                                                                       'mtu': 2000,
                                                                       'name': '137683/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797719',
                                                                       'out_octets': '1157913948289',
                                                                       'out_ucast_pkts': '908512680',
                                                                       'poller_hash': 'NDJCRUUxcGVlcmludGYxMzc2ODMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137683',
                                                                       'in_nucast_pkts': '9797719',
                                                                       'in_octets': '1157913948289',
                                                                       'in_ucast_pkts': '908512680',
                                                                       'links': [       {       'in_nucast_pkts': '9797719',
                                                                                                'in_octets': '1157913948289',
                                                                                                'in_ucast_pkts': '908512680',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9615226',
                                                                                                'out_octets': '184311563745',
                                                                                                'out_ucast_pkts': '345628142',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9615226',
                                                                       'out_octets': '184311563745',
                                                                       'out_ucast_pkts': '345628142',
                                                                       'poller_hash': 'NDJCRUUxcGVlcmludGZwb24gMTM3Njgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137683',
                                                                                               'links': [       {       'connected_time': '8535283',
                                                                                                                        'fiber_length': 25373,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRUUxcGVlcnBvbnBvbiAxMzc2ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '40 WELWOOD AV (PENNY TOM)',
                                         'poller_hash': 'NDJCRUUx',
                                         'serial_number': '42BEE1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6c:4a',
                                         'uptime': '8535283'},
                                 {       'counter_source': '137704/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:4b:e4'],
                                                                       'id': '137704/g1',
                                                                       'in_nucast_pkts': '12',
                                                                       'in_octets': '50913849073',
                                                                       'in_ucast_pkts': '338551163',
                                                                       'mac_address': 'cc:be:59:cb:b7:8d',
                                                                       'mtu': 2000,
                                                                       'name': '137704/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797613',
                                                                       'out_octets': '2062282557647',
                                                                       'out_ucast_pkts': '1430624477',
                                                                       'poller_hash': 'NDJDODQ5cGVlcmludGYxMzc3MDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137704',
                                                                       'in_nucast_pkts': '9797613',
                                                                       'in_octets': '2062282557647',
                                                                       'in_ucast_pkts': '1430624477',
                                                                       'links': [       {       'in_nucast_pkts': '9797613',
                                                                                                'in_octets': '2062282557647',
                                                                                                'in_ucast_pkts': '1430624477',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12',
                                                                                                'out_octets': '50913849073',
                                                                                                'out_ucast_pkts': '338551163',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12',
                                                                       'out_octets': '50913849073',
                                                                       'out_ucast_pkts': '338551163',
                                                                       'poller_hash': 'NDJDODQ5cGVlcmludGZwb24gMTM3NzA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137704',
                                                                                               'links': [       {       'connected_time': '8535266',
                                                                                                                        'fiber_length': 25006,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDODQ5cGVlcnBvbnBvbiAxMzc3MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 PARKVIEW DR (NAFZIGER JUANITA)',
                                         'poller_hash': 'NDJDODQ5',
                                         'serial_number': '42C849',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:b7:8a',
                                         'uptime': '8535266'},
                                 {       'counter_source': '137760/F1+137760/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:eb:41:08'],
                                                                       'id': '137760/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3937421288',
                                                                       'in_ucast_pkts': '20949567',
                                                                       'name': '137760/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '5117266655',
                                                                       'out_octets': '6643008233335',
                                                                       'out_ucast_pkts': '9824702',
                                                                       'poller_hash': 'NDJDOEU1cGVlcmludGYxMzc3NjAvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:df:65'],
                                                                       'id': '137760/g1',
                                                                       'in_nucast_pkts': '87443',
                                                                       'in_octets': '277023608887',
                                                                       'in_ucast_pkts': '642227535',
                                                                       'mac_address': 'cc:be:59:cb:bc:6d',
                                                                       'mtu': 2000,
                                                                       'name': '137760/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5039010',
                                                                       'out_octets': '1856149034718',
                                                                       'out_ucast_pkts': '1547467360',
                                                                       'poller_hash': 'NDJDOEU1cGVlcmludGYxMzc3NjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137760',
                                                                       'in_nucast_pkts': '5122305665',
                                                                       'in_octets': '8499157268053',
                                                                       'in_ucast_pkts': '1557292062',
                                                                       'links': [       {       'in_nucast_pkts': '5122305665',
                                                                                                'in_octets': '8499157268053',
                                                                                                'in_ucast_pkts': '1557292062',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '87443',
                                                                                                'out_octets': '280961030175',
                                                                                                'out_ucast_pkts': '663177102',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '87443',
                                                                       'out_octets': '280961030175',
                                                                       'out_ucast_pkts': '663177102',
                                                                       'poller_hash': 'NDJDOEU1cGVlcmludGZwb24gMTM3NzYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137760',
                                                                                               'links': [       {       'connected_time': '4377661',
                                                                                                                        'fiber_length': 25590,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOEU1cGVlcnBvbnBvbiAxMzc3NjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '81 WELWOOD AV (DIENESCH MICHELLE)',
                                         'poller_hash': 'NDJDOEU1',
                                         'serial_number': '42C8E5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bc:6a',
                                         'uptime': '4377661'},
                                 {       'counter_source': '137783/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'bc:a5:11:00:fd:5a'],
                                                                       'id': '137783/g1',
                                                                       'in_nucast_pkts': '12223',
                                                                       'in_octets': '216523051750',
                                                                       'in_ucast_pkts': '1408814027',
                                                                       'mac_address': '44:65:7f:7b:ce:63',
                                                                       'mtu': 2000,
                                                                       'name': '137783/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38463292',
                                                                       'out_octets': '8447665564856',
                                                                       'out_ucast_pkts': '5858540943',
                                                                       'poller_hash': 'NUE2OUJDcGVlcmludGYxMzc3ODMvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137783',
                                                                       'in_nucast_pkts': '38463292',
                                                                       'in_octets': '8447665564856',
                                                                       'in_ucast_pkts': '5858540943',
                                                                       'links': [       {       'in_nucast_pkts': '38463292',
                                                                                                'in_octets': '8447665564856',
                                                                                                'in_ucast_pkts': '5858540943',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12223',
                                                                                                'out_octets': '216523051750',
                                                                                                'out_ucast_pkts': '1408814027',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12223',
                                                                       'out_octets': '216523051750',
                                                                       'out_ucast_pkts': '1408814027',
                                                                       'poller_hash': 'NUE2OUJDcGVlcmludGZwb24gMTM3Nzgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137783',
                                                                                               'links': [       {       'connected_time': '67856057',
                                                                                                                        'fiber_length': 23020,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2OUJDcGVlcnBvbnBvbiAxMzc3ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1107 HENRY ST (SOULLIERE CINDY)',
                                         'poller_hash': 'NUE2OUJD',
                                         'serial_number': '5A69BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:ce:63',
                                         'uptime': '67856057'},
                                 {       'counter_source': '137787/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '9c:3d:cf:4c:02:9c'],
                                                                       'id': '137787/g1',
                                                                       'in_nucast_pkts': '2409',
                                                                       'in_octets': '119677772601',
                                                                       'in_ucast_pkts': '461641908',
                                                                       'mac_address': '44:65:7f:33:20:87',
                                                                       'mtu': 2000,
                                                                       'name': '137787/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530078',
                                                                       'out_octets': '1718438648332',
                                                                       'out_ucast_pkts': '1278020711',
                                                                       'poller_hash': 'NjQwQUJEcGVlcmludGYxMzc3ODcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '137787',
                                                                       'in_nucast_pkts': '5530078',
                                                                       'in_octets': '1718438648332',
                                                                       'in_ucast_pkts': '1278020711',
                                                                       'links': [       {       'in_nucast_pkts': '5530078',
                                                                                                'in_octets': '1718438648332',
                                                                                                'in_ucast_pkts': '1278020711',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2409',
                                                                                                'out_octets': '119677772601',
                                                                                                'out_ucast_pkts': '461641908',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2409',
                                                                       'out_octets': '119677772601',
                                                                       'out_ucast_pkts': '461641908',
                                                                       'poller_hash': 'NjQwQUJEcGVlcmludGZwb24gMTM3Nzg3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137787',
                                                                                               'links': [       {       'connected_time': '8535111',
                                                                                                                        'fiber_length': 25051,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.872,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.304}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQUJEcGVlcnBvbnBvbiAxMzc3ODc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '71 VILLAGE RD (LEBOLD STEVE)',
                                         'poller_hash': 'NjQwQUJE',
                                         'serial_number': '640ABD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:20:86',
                                         'uptime': '8535111'},
                                 {       'counter_source': '137790/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '137790/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '18',
                                                                       'in_octets': '279016136852',
                                                                       'in_ucast_pkts': '1647090306',
                                                                       'mac_address': 'cc:be:59:cb:bc:2a',
                                                                       'name': '137790/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '1256',
                                                                       'out_octets': '11311993584652',
                                                                       'out_ucast_pkts': '7937262607',
                                                                       'poller_hash': 'NDJDOEREcGVlcmludGYxMzc3OTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '137790',
                                                                       'in_nucast_pkts': '1256',
                                                                       'in_octets': '11311993584652',
                                                                       'in_ucast_pkts': '7937262607',
                                                                       'links': [       {       'in_nucast_pkts': '1256',
                                                                                                'in_octets': '11311993584652',
                                                                                                'in_ucast_pkts': '7937262607',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18',
                                                                                                'out_octets': '279016136852',
                                                                                                'out_ucast_pkts': '1647090306',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18',
                                                                       'out_octets': '279016136852',
                                                                       'out_ucast_pkts': '1647090306',
                                                                       'poller_hash': 'NDJDOEREcGVlcmludGZwb24gMTM3Nzkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137790',
                                                                                               'links': [       {       'connected_time': '31260441',
                                                                                                                        'fiber_length': 25577,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDOEREcGVlcnBvbnBvbiAxMzc3OTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '73 WELWOOD AV (MCLEOD ANGUS & '
                                                 'HARTMANN GERT)',
                                         'poller_hash': 'NDJDOERE',
                                         'serial_number': '42C8DD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:bc:2a',
                                         'uptime': '31260441'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '137800/g1',
                                                                       'name': '137800/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDODQ2cGVlcmludGYxMzc4MDAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '137800',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDODQ2cGVlcmludGZwb24gMTM3ODAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137800',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDODQ2cGVlcnBvbnBvbiAxMzc4MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '74 POND VIEW DR (MACMILLAN BRAD)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "137800" '
                                                                  '(location: GPON '
                                                                  'port "6/4"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2024-11-28T18:54:44.460Z',
                                         'poller_hash': 'NDJDODQ2',
                                         'serial_number': '42C846',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '137810/g1',
                                                                       'name': '137810/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwMEVDcGVlcmludGYxMzc4MTAvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '137810',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NjQwMEVDcGVlcmludGZwb24gMTM3ODEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '137810',
                                                                                               'links': [       {       'connected_time': '8535101',
                                                                                                                        'fiber_length': 24922,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.28,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.136}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwMEVDcGVlcnBvbnBvbiAxMzc4MTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '44 VILLAGE RD',
                                         'poller_hash': 'NjQwMEVD',
                                         'serial_number': '6400EC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:0c:c8',
                                         'uptime': '8535101'},
                                 {       'counter_source': '138010/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:50:fd'],
                                                                       'id': '138010/g1',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '175497202772',
                                                                       'in_ucast_pkts': '974087372',
                                                                       'mac_address': '44:65:7f:33:21:25',
                                                                       'mtu': 2000,
                                                                       'name': '138010/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530171',
                                                                       'out_octets': '2818988480943',
                                                                       'out_ucast_pkts': '2030628296',
                                                                       'poller_hash': 'NjQwQjBBcGVlcmludGYxMzgwMTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138010',
                                                                       'in_nucast_pkts': '5530171',
                                                                       'in_octets': '2818988480943',
                                                                       'in_ucast_pkts': '2030628296',
                                                                       'links': [       {       'in_nucast_pkts': '5530171',
                                                                                                'in_octets': '2818988480943',
                                                                                                'in_ucast_pkts': '2030628296',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '14',
                                                                                                'out_octets': '175497202772',
                                                                                                'out_ucast_pkts': '974087372',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '175497202772',
                                                                       'out_ucast_pkts': '974087372',
                                                                       'poller_hash': 'NjQwQjBBcGVlcmludGZwb24gMTM4MDEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138010',
                                                                                               'links': [       {       'connected_time': '8535127',
                                                                                                                        'fiber_length': 24947,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.61,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.348}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjBBcGVlcnBvbnBvbiAxMzgwMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '4 SHADYWOOD CT (BAUMAN CHARLES)',
                                         'poller_hash': 'NjQwQjBB',
                                         'serial_number': '640B0A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:24',
                                         'uptime': '8535127'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '138016/g1',
                                                                       'name': '138016/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NzNBNjc5cGVlcmludGYxMzgwMTYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138016',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NzNBNjc5cGVlcmludGZwb24gMTM4MDE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138016',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzNBNjc5cGVlcnBvbnBvbiAxMzgwMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '3706 NAFZIGER RD UNIT 5',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "138016" '
                                                                  '(location: GPON '
                                                                  'port "7/4"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:27:12.580Z',
                                         'poller_hash': 'NzNBNjc5',
                                         'serial_number': '73A679',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '138092/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:07:b6:a8:fa:99'],
                                                                       'id': '138092/g1',
                                                                       'in_nucast_pkts': '49770',
                                                                       'in_octets': '1093595608',
                                                                       'in_ucast_pkts': '6471007',
                                                                       'mac_address': '44:65:7f:33:21:49',
                                                                       'mtu': 2000,
                                                                       'name': '138092/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '211193',
                                                                       'out_octets': '31626570773',
                                                                       'out_ucast_pkts': '23503420',
                                                                       'poller_hash': 'NjQwQjFDcGVlcmludGYxMzgwOTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138092',
                                                                       'in_nucast_pkts': '211193',
                                                                       'in_octets': '31626570773',
                                                                       'in_ucast_pkts': '23503420',
                                                                       'links': [       {       'in_nucast_pkts': '211193',
                                                                                                'in_octets': '31626570773',
                                                                                                'in_ucast_pkts': '23503420',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '49770',
                                                                                                'out_octets': '1093595608',
                                                                                                'out_ucast_pkts': '6471007',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49770',
                                                                       'out_octets': '1093595608',
                                                                       'out_ucast_pkts': '6471007',
                                                                       'poller_hash': 'NjQwQjFDcGVlcmludGZwb24gMTM4MDky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138092',
                                                                                               'links': [       {       'connected_time': '319455',
                                                                                                                        'fiber_length': 23420,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.19,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.488}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjFDcGVlcnBvbnBvbiAxMzgwOTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '28 REINER CR (SNYDER SCOTT)',
                                         'poller_hash': 'NjQwQjFD',
                                         'serial_number': '640B1C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:48',
                                         'uptime': '319455'},
                                 {       'counter_source': '138102/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:9a:4e:87'],
                                                                       'id': '138102/g1',
                                                                       'in_nucast_pkts': '4',
                                                                       'in_octets': '40220557255',
                                                                       'in_ucast_pkts': '77448230',
                                                                       'mac_address': '44:65:7f:33:0c:d7',
                                                                       'mtu': 2000,
                                                                       'name': '138102/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530454',
                                                                       'out_octets': '148585922607',
                                                                       'out_ucast_pkts': '130296789',
                                                                       'poller_hash': 'NjQwMEYzcGVlcmludGYxMzgxMDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138102',
                                                                       'in_nucast_pkts': '5530454',
                                                                       'in_octets': '148585922607',
                                                                       'in_ucast_pkts': '130296789',
                                                                       'links': [       {       'in_nucast_pkts': '5530454',
                                                                                                'in_octets': '148585922607',
                                                                                                'in_ucast_pkts': '130296789',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4',
                                                                                                'out_octets': '40220557255',
                                                                                                'out_ucast_pkts': '77448230',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4',
                                                                       'out_octets': '40220557255',
                                                                       'out_ucast_pkts': '77448230',
                                                                       'poller_hash': 'NjQwMEYzcGVlcmludGZwb24gMTM4MTAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138102',
                                                                                               'links': [       {       'connected_time': '8535108',
                                                                                                                        'fiber_length': 23046,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.666,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.158}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwMEYzcGVlcnBvbnBvbiAxMzgxMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '1087 HENRY ST (Hugh Cameron)',
                                         'poller_hash': 'NjQwMEYz',
                                         'serial_number': '6400F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:0c:d6',
                                         'uptime': '8535108'},
                                 {       'counter_source': '138111/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd4:be:d9:85:3f:5d'],
                                                                       'id': '138111/g1',
                                                                       'in_nucast_pkts': '826836',
                                                                       'in_octets': '7638134032',
                                                                       'in_ucast_pkts': '32792287',
                                                                       'mac_address': '44:65:7f:7b:d0:22',
                                                                       'mtu': 2000,
                                                                       'name': '138111/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39418315',
                                                                       'out_octets': '83131146754',
                                                                       'out_ucast_pkts': '75114098',
                                                                       'poller_hash': 'NUE2QTUxcGVlcmludGYxMzgxMTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138111',
                                                                       'in_nucast_pkts': '39418315',
                                                                       'in_octets': '83131146754',
                                                                       'in_ucast_pkts': '75114098',
                                                                       'links': [       {       'in_nucast_pkts': '39418315',
                                                                                                'in_octets': '83131146754',
                                                                                                'in_ucast_pkts': '75114098',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '826836',
                                                                                                'out_octets': '7638134032',
                                                                                                'out_ucast_pkts': '32792287',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '826836',
                                                                       'out_octets': '7638134032',
                                                                       'out_ucast_pkts': '32792287',
                                                                       'poller_hash': 'NUE2QTUxcGVlcmludGZwb24gMTM4MTEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138111',
                                                                                               'links': [       {       'connected_time': '67855892',
                                                                                                                        'fiber_length': 23090,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QTUxcGVlcnBvbnBvbiAxMzgxMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '125 DAVID ST (LUDINGTON GORDON)',
                                         'poller_hash': 'NUE2QTUx',
                                         'serial_number': '5A6A51',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d0:22',
                                         'uptime': '67855892'},
                                 {       'counter_source': '138115/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:f4:11:45:6e:c1'],
                                                                       'id': '138115/g1',
                                                                       'in_nucast_pkts': '437',
                                                                       'in_octets': '6706774658',
                                                                       'in_ucast_pkts': '10723619',
                                                                       'mac_address': '44:65:7f:33:21:33',
                                                                       'mtu': 2000,
                                                                       'name': '138115/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '226581',
                                                                       'out_octets': '33267903501',
                                                                       'out_ucast_pkts': '28583706',
                                                                       'poller_hash': 'NjQwQjExcGVlcmludGYxMzgxMTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138115',
                                                                       'in_nucast_pkts': '226581',
                                                                       'in_octets': '33267903501',
                                                                       'in_ucast_pkts': '28583706',
                                                                       'links': [       {       'in_nucast_pkts': '226581',
                                                                                                'in_octets': '33267903501',
                                                                                                'in_ucast_pkts': '28583706',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '437',
                                                                                                'out_octets': '6706774658',
                                                                                                'out_ucast_pkts': '10723619',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '437',
                                                                       'out_octets': '6706774658',
                                                                       'out_ucast_pkts': '10723619',
                                                                       'poller_hash': 'NjQwQjExcGVlcmludGZwb24gMTM4MTE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138115',
                                                                                               'links': [       {       'connected_time': '335344',
                                                                                                                        'fiber_length': 17590,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.802,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.032}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwQjExcGVlcnBvbnBvbiAxMzgxMTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': '5447 KING ST (WURDELL PAUL)',
                                         'poller_hash': 'NjQwQjEx',
                                         'serial_number': '640B11',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:21:32',
                                         'uptime': '335344'},
                                 {       'counter_source': '138128/F1+138128/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:31:2e',
                                                                                                        '00:03:e6:eb:3c:d0'],
                                                                       'id': '138128/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6138521020',
                                                                       'in_ucast_pkts': '33689059',
                                                                       'name': '138128/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '28814181558',
                                                                       'out_octets': '38844548442246',
                                                                       'out_ucast_pkts': '6850567',
                                                                       'poller_hash': 'ODcyNTYycGVlcmludGYxMzgxMjgvRjE='},
                                                               {       'admin_up': True,
                                                                       'id': '138128/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '549815651236',
                                                                       'in_ucast_pkts': '1240568645',
                                                                       'mac_address': 'd0:76:8f:1b:a8:a3',
                                                                       'name': '138128/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '624',
                                                                       'out_octets': '4492333111732',
                                                                       'out_ucast_pkts': '3593473097',
                                                                       'poller_hash': 'ODcyNTYycGVlcmludGYxMzgxMjgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138128',
                                                                       'in_nucast_pkts': '28814182182',
                                                                       'in_octets': '43336881553978',
                                                                       'in_ucast_pkts': '3600323664',
                                                                       'links': [       {       'in_nucast_pkts': '28814182182',
                                                                                                'in_octets': '43336881553978',
                                                                                                'in_ucast_pkts': '3600323664',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '555954172256',
                                                                                                'out_ucast_pkts': '1274257704',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '555954172256',
                                                                       'out_ucast_pkts': '1274257704',
                                                                       'poller_hash': 'ODcyNTYycGVlcmludGZwb24gMTM4MTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138128',
                                                                                               'links': [       {       'connected_time': '8535249',
                                                                                                                        'fiber_length': 23084,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTYycGVlcnBvbnBvbiAxMzgxMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '107 DAVID ST (SCHROTH ANDRE)',
                                         'poller_hash': 'ODcyNTYy',
                                         'serial_number': '872562',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a8:a3',
                                         'uptime': '8535249'},
                                 {       'counter_source': '138160/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:df:05'],
                                                                       'id': '138160/g1',
                                                                       'in_nucast_pkts': '18',
                                                                       'in_octets': '211024629564',
                                                                       'in_ucast_pkts': '852342605',
                                                                       'mac_address': 'cc:be:59:cb:99:fd',
                                                                       'mtu': 2000,
                                                                       'name': '138160/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9917745',
                                                                       'out_octets': '2369931033824',
                                                                       'out_ucast_pkts': '1886143351',
                                                                       'poller_hash': 'NDJDNDk3cGVlcmludGYxMzgxNjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138160',
                                                                       'in_nucast_pkts': '9917745',
                                                                       'in_octets': '2369931033824',
                                                                       'in_ucast_pkts': '1886143351',
                                                                       'links': [       {       'in_nucast_pkts': '9917745',
                                                                                                'in_octets': '2369931033824',
                                                                                                'in_ucast_pkts': '1886143351',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18',
                                                                                                'out_octets': '211024629564',
                                                                                                'out_ucast_pkts': '852342605',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18',
                                                                       'out_octets': '211024629564',
                                                                       'out_ucast_pkts': '852342605',
                                                                       'poller_hash': 'NDJDNDk3cGVlcmludGZwb24gMTM4MTYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138160',
                                                                                               'links': [       {       'connected_time': '8535261',
                                                                                                                        'fiber_length': 23329,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDNDk3cGVlcnBvbnBvbiAxMzgxNjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '122 SCHWEITZER CR (ANTONIOU LOUIE)',
                                         'poller_hash': 'NDJDNDk3',
                                         'serial_number': '42C497',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:99:fa',
                                         'uptime': '8535261'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '138185/g1',
                                                                       'name': '138185/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2Qzk3cGVlcmludGYxMzgxODUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138185',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2Qzk3cGVlcmludGZwb24gMTM4MTg1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138185',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2Qzk3cGVlcnBvbnBvbiAxMzgxODU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '97 POND VIEW DR',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "138185" '
                                                                  '(location: GPON '
                                                                  'port "1/1"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:26:45.190Z',
                                         'poller_hash': 'NUE2Qzk3',
                                         'serial_number': '5A6C97',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '138189/g1',
                                                                       'name': '138189/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwMDJBcGVlcmludGYxMzgxODkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138189',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NjQwMDJBcGVlcmludGZwb24gMTM4MTg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138189',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwMDJBcGVlcnBvbnBvbiAxMzgxODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': "1109 QUEEN'S BUSH RD (LOTZ CAL)",
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "138189" '
                                                                  '(location: GPON '
                                                                  'port "7/6"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-04-24T15:01:53.750Z',
                                         'poller_hash': 'NjQwMDJB',
                                         'serial_number': '64002A',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '138263/g1',
                                                                       'name': '138263/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJCOUYxcGVlcmludGYxMzgyNjMvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138263',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJCOUYxcGVlcmludGZwb24gMTM4MjYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138263',
                                                                                               'links': [       {       'connected_time': '8535242',
                                                                                                                        'fiber_length': 23906,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCOUYxcGVlcnBvbnBvbiAxMzgyNjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '82 GALENA ST',
                                         'poller_hash': 'NDJCOUYx',
                                         'serial_number': '42B9F1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:44:ca',
                                         'uptime': '8535242'},
                                 {       'counter_source': '138270/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:32:e5:54:05:0a'],
                                                                       'id': '138270/g1',
                                                                       'in_nucast_pkts': '2638',
                                                                       'in_octets': '10689005622',
                                                                       'in_ucast_pkts': '59123350',
                                                                       'mac_address': '44:65:7f:33:0d:d1',
                                                                       'mtu': 2000,
                                                                       'name': '138270/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5530250',
                                                                       'out_octets': '275715878756',
                                                                       'out_ucast_pkts': '208053403',
                                                                       'poller_hash': 'NjQwMTZFcGVlcmludGYxMzgyNzAvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138270',
                                                                       'in_nucast_pkts': '5530250',
                                                                       'in_octets': '275715878756',
                                                                       'in_ucast_pkts': '208053403',
                                                                       'links': [       {       'in_nucast_pkts': '5530250',
                                                                                                'in_octets': '275715878756',
                                                                                                'in_ucast_pkts': '208053403',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2638',
                                                                                                'out_octets': '10689005622',
                                                                                                'out_ucast_pkts': '59123350',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2638',
                                                                       'out_octets': '10689005622',
                                                                       'out_ucast_pkts': '59123350',
                                                                       'poller_hash': 'NjQwMTZFcGVlcmludGZwb24gMTM4Mjcw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138270',
                                                                                               'links': [       {       'connected_time': '8535115',
                                                                                                                        'fiber_length': 24802,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.08,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.078}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NjQwMTZFcGVlcnBvbnBvbiAxMzgyNzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '801Gv2',
                                         'name': "1144 QUEEN'S BUSH RD (JACOBS CORY)",
                                         'poller_hash': 'NjQwMTZF',
                                         'serial_number': '64016E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '1.0.5.0.120',
                                         'system_mac_address': '44:65:7f:33:0d:d0',
                                         'uptime': '8535115'},
                                 {       'counter_source': '138423/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e6:50'],
                                                                       'id': '138423/g1',
                                                                       'in_nucast_pkts': '958335',
                                                                       'in_octets': '151504799534',
                                                                       'in_ucast_pkts': '495230176',
                                                                       'mac_address': '48:77:46:0c:04:61',
                                                                       'mtu': 2000,
                                                                       'name': '138423/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62544524',
                                                                       'out_octets': '987205040885',
                                                                       'out_ucast_pkts': '972006426',
                                                                       'poller_hash': 'NkRENTI0cGVlcmludGYxMzg0MjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138423',
                                                                       'in_nucast_pkts': '62544524',
                                                                       'in_octets': '987205040885',
                                                                       'in_ucast_pkts': '972006426',
                                                                       'links': [       {       'in_nucast_pkts': '62544524',
                                                                                                'in_octets': '987205040885',
                                                                                                'in_ucast_pkts': '972006426',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '958335',
                                                                                                'out_octets': '151504799534',
                                                                                                'out_ucast_pkts': '495230176',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '958335',
                                                                       'out_octets': '151504799534',
                                                                       'out_ucast_pkts': '495230176',
                                                                       'poller_hash': 'NkRENTI0cGVlcmludGZwb24gMTM4NDIz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138423',
                                                                                               'links': [       {       'connected_time': '21886650',
                                                                                                                        'fiber_length': 24006,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NkRENTI0cGVlcnBvbnBvbiAxMzg0MjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3691 NAFZIGER RD (G/A REPAIR '
                                                 'MACHINERY INC)',
                                         'poller_hash': 'NkRENTI0',
                                         'serial_number': '6DD524',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:0c:04:5e',
                                         'uptime': '21886650'},
                                 {       'counter_source': '138604/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:cb:32'],
                                                                       'id': '138604/g1',
                                                                       'in_nucast_pkts': '28',
                                                                       'in_octets': '100241454627',
                                                                       'in_ucast_pkts': '446269730',
                                                                       'mac_address': 'd0:76:8f:08:17:8e',
                                                                       'mtu': 2000,
                                                                       'name': '138604/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12591428',
                                                                       'out_octets': '2119914995184',
                                                                       'out_ucast_pkts': '1481796854',
                                                                       'poller_hash': 'ODY5M0U3cGVlcmludGYxMzg2MDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138604',
                                                                       'in_nucast_pkts': '12591428',
                                                                       'in_octets': '2119914995184',
                                                                       'in_ucast_pkts': '1481796854',
                                                                       'links': [       {       'in_nucast_pkts': '12591428',
                                                                                                'in_octets': '2119914995184',
                                                                                                'in_ucast_pkts': '1481796854',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '28',
                                                                                                'out_octets': '100241454627',
                                                                                                'out_ucast_pkts': '446269730',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '28',
                                                                       'out_octets': '100241454627',
                                                                       'out_ucast_pkts': '446269730',
                                                                       'poller_hash': 'ODY5M0U3cGVlcmludGZwb24gMTM4NjA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138604',
                                                                                               'links': [       {       'connected_time': '10881856',
                                                                                                                        'fiber_length': 23031,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0U3cGVlcnBvbnBvbiAxMzg2MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6 POND VIEW DR (JANTZI SANDRA)',
                                         'poller_hash': 'ODY5M0U3',
                                         'serial_number': '8693E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:8b',
                                         'uptime': '10881856'},
                                 {       'counter_source': '138645/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:ac:f1:05'],
                                                                       'id': '138645/g1',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '28559874395',
                                                                       'in_ucast_pkts': '225402293',
                                                                       'mac_address': 'd0:76:8f:08:19:9e',
                                                                       'mtu': 2000,
                                                                       'name': '138645/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878141',
                                                                       'out_octets': '1718236441552',
                                                                       'out_ucast_pkts': '1170540845',
                                                                       'poller_hash': 'ODY5NDI5cGVlcmludGYxMzg2NDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138645',
                                                                       'in_nucast_pkts': '9878141',
                                                                       'in_octets': '1718236441552',
                                                                       'in_ucast_pkts': '1170540845',
                                                                       'links': [       {       'in_nucast_pkts': '9878141',
                                                                                                'in_octets': '1718236441552',
                                                                                                'in_ucast_pkts': '1170540845',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7',
                                                                                                'out_octets': '28559874395',
                                                                                                'out_ucast_pkts': '225402293',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '28559874395',
                                                                       'out_ucast_pkts': '225402293',
                                                                       'poller_hash': 'ODY5NDI5cGVlcmludGZwb24gMTM4NjQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138645',
                                                                                               'links': [       {       'connected_time': '8535218',
                                                                                                                        'fiber_length': 22656,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDI5cGVlcnBvbnBvbiAxMzg2NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '17 LAWRENCE ST (BISCH DORIS & '
                                                 'MURRAY)',
                                         'poller_hash': 'ODY5NDI5',
                                         'serial_number': '869429',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:19:9b',
                                         'uptime': '8535218'},
                                 {       'counter_source': '138708/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:38:e0:a8:30:8d'],
                                                                       'id': '138708/g1',
                                                                       'in_nucast_pkts': '73346',
                                                                       'in_octets': '15494382907',
                                                                       'in_ucast_pkts': '135267901',
                                                                       'mac_address': '44:65:7f:7b:d3:76',
                                                                       'mtu': 2000,
                                                                       'name': '138708/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1285690',
                                                                       'out_octets': '1290969282633',
                                                                       'out_ucast_pkts': '868361148',
                                                                       'poller_hash': 'NUE2QjZEcGVlcmludGYxMzg3MDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138708',
                                                                       'in_nucast_pkts': '1285690',
                                                                       'in_octets': '1290969282633',
                                                                       'in_ucast_pkts': '868361148',
                                                                       'links': [       {       'in_nucast_pkts': '1285690',
                                                                                                'in_octets': '1290969282633',
                                                                                                'in_ucast_pkts': '868361148',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '73346',
                                                                                                'out_octets': '15494382907',
                                                                                                'out_ucast_pkts': '135267901',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '73346',
                                                                       'out_octets': '15494382907',
                                                                       'out_ucast_pkts': '135267901',
                                                                       'poller_hash': 'NUE2QjZEcGVlcmludGZwb24gMTM4NzA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138708',
                                                                                               'links': [       {       'connected_time': '1095976',
                                                                                                                        'fiber_length': 31256,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjZEcGVlcnBvbnBvbiAxMzg3MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1834 MOSER-YOUNG RD (MAINPRIZE PAUL)',
                                         'poller_hash': 'NUE2QjZE',
                                         'serial_number': '5A6B6D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:76',
                                         'uptime': '1095976'},
                                 {       'counter_source': '138716/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:2d:6c:40:58:01'],
                                                                       'id': '138716/g1',
                                                                       'in_nucast_pkts': '73599',
                                                                       'in_octets': '312850740878',
                                                                       'in_ucast_pkts': '339104162',
                                                                       'mac_address': '44:65:7f:7b:d3:eb',
                                                                       'mtu': 2000,
                                                                       'name': '138716/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2395957',
                                                                       'out_octets': '950852963816',
                                                                       'out_ucast_pkts': '816032177',
                                                                       'poller_hash': 'NUE2Qjk0cGVlcmludGYxMzg3MTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138716',
                                                                       'in_nucast_pkts': '2395957',
                                                                       'in_octets': '950852963816',
                                                                       'in_ucast_pkts': '816032177',
                                                                       'links': [       {       'in_nucast_pkts': '2395957',
                                                                                                'in_octets': '950852963816',
                                                                                                'in_ucast_pkts': '816032177',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '73599',
                                                                                                'out_octets': '312850740878',
                                                                                                'out_ucast_pkts': '339104162',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '73599',
                                                                       'out_octets': '312850740878',
                                                                       'out_ucast_pkts': '339104162',
                                                                       'poller_hash': 'NUE2Qjk0cGVlcmludGZwb24gMTM4NzE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138716',
                                                                                               'links': [       {       'connected_time': '2060312',
                                                                                                                        'fiber_length': 23427,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2Qjk0cGVlcnBvbnBvbiAxMzg3MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '119 FERRIS DR (CHAKRABURTTY SAIBAL)',
                                         'poller_hash': 'NUE2Qjk0',
                                         'serial_number': '5A6B94',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:eb',
                                         'uptime': '2060312'},
                                 {       'counter_source': '138720/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd8:07:b6:c5:34:07'],
                                                                       'id': '138720/g1',
                                                                       'in_nucast_pkts': '8211',
                                                                       'in_octets': '28447001851',
                                                                       'in_ucast_pkts': '74934545',
                                                                       'mac_address': '44:65:7f:7b:d4:30',
                                                                       'mtu': 2000,
                                                                       'name': '138720/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5341840',
                                                                       'out_octets': '295670434221',
                                                                       'out_ucast_pkts': '243438490',
                                                                       'poller_hash': 'NUE2QkFCcGVlcmludGYxMzg3MjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138720',
                                                                       'in_nucast_pkts': '5341840',
                                                                       'in_octets': '295670434221',
                                                                       'in_ucast_pkts': '243438490',
                                                                       'links': [       {       'in_nucast_pkts': '5341840',
                                                                                                'in_octets': '295670434221',
                                                                                                'in_ucast_pkts': '243438490',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8211',
                                                                                                'out_octets': '28447001851',
                                                                                                'out_ucast_pkts': '74934545',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8211',
                                                                       'out_octets': '28447001851',
                                                                       'out_ucast_pkts': '74934545',
                                                                       'poller_hash': 'NUE2QkFCcGVlcmludGZwb24gMTM4NzIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138720',
                                                                                               'links': [       {       'connected_time': '4591349',
                                                                                                                        'fiber_length': 23566,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QkFCcGVlcnBvbnBvbiAxMzg3MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '3536 NAFZIGER RD (DANIEL BISCH '
                                                 'GENERAL CONTRACTO)',
                                         'poller_hash': 'NUE2QkFC',
                                         'serial_number': '5A6BAB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d4:30',
                                         'uptime': '4591349'},
                                 {       'counter_source': '138742/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '24:05:88:0f:65:ff'],
                                                                       'id': '138742/g1',
                                                                       'in_nucast_pkts': '817',
                                                                       'in_octets': '38353034534',
                                                                       'in_ucast_pkts': '50001076',
                                                                       'mac_address': '44:65:7f:7b:d6:e2',
                                                                       'mtu': 2000,
                                                                       'name': '138742/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1584472',
                                                                       'out_octets': '155905052034',
                                                                       'out_ucast_pkts': '127464849',
                                                                       'poller_hash': 'NUE2QzkxcGVlcmludGYxMzg3NDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138742',
                                                                       'in_nucast_pkts': '1584472',
                                                                       'in_octets': '155905052034',
                                                                       'in_ucast_pkts': '127464849',
                                                                       'links': [       {       'in_nucast_pkts': '1584472',
                                                                                                'in_octets': '155905052034',
                                                                                                'in_ucast_pkts': '127464849',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '817',
                                                                                                'out_octets': '38353034534',
                                                                                                'out_ucast_pkts': '50001076',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '817',
                                                                       'out_octets': '38353034534',
                                                                       'out_ucast_pkts': '50001076',
                                                                       'poller_hash': 'NUE2QzkxcGVlcmludGZwb24gMTM4NzQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138742',
                                                                                               'links': [       {       'connected_time': '1357590',
                                                                                                                        'fiber_length': 24801,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzkxcGVlcnBvbnBvbiAxMzg3NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '16 VILLAGE RD (EGGERT MIKE)',
                                         'poller_hash': 'NUE2Qzkx',
                                         'serial_number': '5A6C91',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d6:e2',
                                         'uptime': '1357590'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '138808/g1',
                                                                       'name': '138808/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2QzZBcGVlcmludGYxMzg4MDgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138808',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2QzZBcGVlcmludGZwb24gMTM4ODA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138808',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzZBcGVlcnBvbnBvbiAxMzg4MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '1020 MOLESWORTH ST (LALONDE KAREN)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "138808" '
                                                                  '(location: GPON '
                                                                  'port "7/4"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-07-01T16:11:10.010Z',
                                         'poller_hash': 'NUE2QzZB',
                                         'serial_number': '5A6C6A',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '138812/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:fd:74:7a:b8:52'],
                                                                       'id': '138812/g1',
                                                                       'in_nucast_pkts': '469',
                                                                       'in_octets': '4680914934648',
                                                                       'in_ucast_pkts': '8871026434',
                                                                       'mac_address': '44:65:7f:7b:d6:b8',
                                                                       'mtu': 2000,
                                                                       'name': '138812/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14071240',
                                                                       'out_octets': '8058128001026',
                                                                       'out_ucast_pkts': '8896216037',
                                                                       'poller_hash': 'NUE2QzgzcGVlcmludGYxMzg4MTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '138812',
                                                                       'in_nucast_pkts': '14071240',
                                                                       'in_octets': '8058128001026',
                                                                       'in_ucast_pkts': '8896216037',
                                                                       'links': [       {       'in_nucast_pkts': '14071240',
                                                                                                'in_octets': '8058128001026',
                                                                                                'in_ucast_pkts': '8896216037',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '469',
                                                                                                'out_octets': '4680914934648',
                                                                                                'out_ucast_pkts': '8871026434',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '469',
                                                                       'out_octets': '4680914934648',
                                                                       'out_ucast_pkts': '8871026434',
                                                                       'poller_hash': 'NUE2QzgzcGVlcmludGZwb24gMTM4ODEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138812',
                                                                                               'links': [       {       'connected_time': '8570933',
                                                                                                                        'fiber_length': 22308,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzgzcGVlcnBvbnBvbiAxMzg4MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': "1245 QUEEN'S BUSH RD (GUENTHER TECH "
                                                 'INC)',
                                         'poller_hash': 'NUE2Qzgz',
                                         'serial_number': '5A6C83',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d6:b8',
                                         'uptime': '8570933'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '138816/g1',
                                                                       'name': '138816/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2Qjg4cGVlcmludGYxMzg4MTYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '138816',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NUE2Qjg4cGVlcmludGZwb24gMTM4ODE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '138816',
                                                                                               'links': [       {       'connected_time': '1095976',
                                                                                                                        'fiber_length': 31013,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2Qjg4cGVlcnBvbnBvbiAxMzg4MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1810 MOSER-YOUNG RD (SPURR BRENT)',
                                         'poller_hash': 'NUE2Qjg4',
                                         'serial_number': '5A6B88',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:c7',
                                         'uptime': '1095976'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '139309/g1',
                                                                       'name': '139309/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NUE2QzkzcGVlcmludGYxMzkzMDkvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '139309',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NUE2QzkzcGVlcmludGZwb24gMTM5MzA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139309',
                                                                                               'links': [       {       'connected_time': '8535397',
                                                                                                                        'fiber_length': 23794,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QzkzcGVlcnBvbnBvbiAxMzkzMDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '16 SCHWEITZER CR (AZIZ SHAHZAD)',
                                         'poller_hash': 'NUE2Qzkz',
                                         'serial_number': '5A6C93',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d6:e8',
                                         'uptime': '8535397'},
                                 {       'counter_source': '139327/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f2:3b'],
                                                                       'id': '139327/g1',
                                                                       'in_nucast_pkts': '41701814',
                                                                       'in_octets': '1136477356280',
                                                                       'in_ucast_pkts': '4735292556',
                                                                       'mac_address': 'd0:76:8f:08:12:de',
                                                                       'mtu': 2000,
                                                                       'name': '139327/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '54024616670',
                                                                       'out_octets': '93494576745751',
                                                                       'out_ucast_pkts': '15151155147',
                                                                       'poller_hash': 'ODY5MzUxcGVlcmludGYxMzkzMjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139327',
                                                                       'in_nucast_pkts': '54024616670',
                                                                       'in_octets': '93494576745751',
                                                                       'in_ucast_pkts': '15151155147',
                                                                       'links': [       {       'in_nucast_pkts': '54024616670',
                                                                                                'in_octets': '93494576745751',
                                                                                                'in_ucast_pkts': '15151155147',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '41701814',
                                                                                                'out_octets': '1136477356280',
                                                                                                'out_ucast_pkts': '4735292556',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41701814',
                                                                       'out_octets': '1136477356280',
                                                                       'out_ucast_pkts': '4735292556',
                                                                       'poller_hash': 'ODY5MzUxcGVlcmludGZwb24gMTM5MzI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139327',
                                                                                               'links': [       {       'connected_time': '21887229',
                                                                                                                        'fiber_length': 24945,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzUxcGVlcnBvbnBvbiAxMzkzMjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1152 QUEEN'S BUSH RD (QUAST "
                                                 'LORRAINE)',
                                         'poller_hash': 'ODY5MzUx',
                                         'serial_number': '869351',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:12:db',
                                         'uptime': '21887229'},
                                 {       'counter_source': '139372/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a4:2b:8c:fd:6d:a1'],
                                                                       'id': '139372/g1',
                                                                       'in_nucast_pkts': '2261981',
                                                                       'in_octets': '175020005514',
                                                                       'in_ucast_pkts': '265454169',
                                                                       'mac_address': '44:65:7f:7b:d3:d6',
                                                                       'mtu': 2000,
                                                                       'name': '139372/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39388031',
                                                                       'out_octets': '532085968884',
                                                                       'out_ucast_pkts': '464656993',
                                                                       'poller_hash': 'NUE2QjhEcGVlcmludGYxMzkzNzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139372',
                                                                       'in_nucast_pkts': '39388031',
                                                                       'in_octets': '532085968884',
                                                                       'in_ucast_pkts': '464656993',
                                                                       'links': [       {       'in_nucast_pkts': '39388031',
                                                                                                'in_octets': '532085968884',
                                                                                                'in_ucast_pkts': '464656993',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2261981',
                                                                                                'out_octets': '175020005514',
                                                                                                'out_ucast_pkts': '265454169',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2261981',
                                                                       'out_octets': '175020005514',
                                                                       'out_ucast_pkts': '265454169',
                                                                       'poller_hash': 'NUE2QjhEcGVlcmludGZwb24gMTM5Mzcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139372',
                                                                                               'links': [       {       'connected_time': '67856227',
                                                                                                                        'fiber_length': 23002,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2QjhEcGVlcnBvbnBvbiAxMzkzNzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '3620 NAFZIGER RD (FIRST ST PAULS '
                                                 'LUTHERAN CHURCH)',
                                         'poller_hash': 'NUE2QjhE',
                                         'serial_number': '5A6B8D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:d3:d6',
                                         'uptime': '67856227'},
                                 {       'counter_source': '139384/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:7c:3f:d1:b2:50'],
                                                                       'id': '139384/g1',
                                                                       'in_nucast_pkts': '10436',
                                                                       'in_octets': '3362289748087',
                                                                       'in_ucast_pkts': '4375922892',
                                                                       'mac_address': '48:77:46:f8:76:af',
                                                                       'mtu': 2000,
                                                                       'name': '139384/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17807226',
                                                                       'out_octets': '11214691231889',
                                                                       'out_ucast_pkts': '8852804838',
                                                                       'poller_hash': 'ODJBNUJCcGVlcmludGYxMzkzODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139384',
                                                                       'in_nucast_pkts': '17807226',
                                                                       'in_octets': '11214691231889',
                                                                       'in_ucast_pkts': '8852804838',
                                                                       'links': [       {       'in_nucast_pkts': '17807226',
                                                                                                'in_octets': '11214691231889',
                                                                                                'in_ucast_pkts': '8852804838',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10436',
                                                                                                'out_octets': '3362289748087',
                                                                                                'out_ucast_pkts': '4375922892',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10436',
                                                                       'out_octets': '3362289748087',
                                                                       'out_ucast_pkts': '4375922892',
                                                                       'poller_hash': 'ODJBNUJCcGVlcmludGZwb24gMTM5Mzg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139384',
                                                                                               'links': [       {       'connected_time': '15181507',
                                                                                                                        'fiber_length': 24672,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNUJCcGVlcnBvbnBvbiAxMzkzODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '134 GERBER MEADOWS DR (CAUCEAN '
                                                 'FLORIN)',
                                         'poller_hash': 'ODJBNUJC',
                                         'serial_number': '82A5BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:76:af',
                                         'uptime': '15181507'},
                                 {       'counter_source': '139388/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:39:cb'],
                                                                       'id': '139388/g1',
                                                                       'in_nucast_pkts': '25168',
                                                                       'in_octets': '200479178105',
                                                                       'in_ucast_pkts': '197198842',
                                                                       'mac_address': '48:77:46:f8:7b:65',
                                                                       'mtu': 2000,
                                                                       'name': '139388/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1280566',
                                                                       'out_octets': '294723015846',
                                                                       'out_ucast_pkts': '280434420',
                                                                       'poller_hash': 'ODJBNzREcGVlcmludGYxMzkzODgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139388',
                                                                       'in_nucast_pkts': '1280566',
                                                                       'in_octets': '294723015846',
                                                                       'in_ucast_pkts': '280434420',
                                                                       'links': [       {       'in_nucast_pkts': '1280566',
                                                                                                'in_octets': '294723015846',
                                                                                                'in_ucast_pkts': '280434420',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '25168',
                                                                                                'out_octets': '200479178105',
                                                                                                'out_ucast_pkts': '197198842',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '25168',
                                                                       'out_octets': '200479178105',
                                                                       'out_ucast_pkts': '197198842',
                                                                       'poller_hash': 'ODJBNzREcGVlcmludGZwb24gMTM5Mzg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139388',
                                                                                               'links': [       {       'connected_time': '1095995',
                                                                                                                        'fiber_length': 31110,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNzREcGVlcnBvbnBvbiAxMzkzODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1831 MOSER-YOUNG RD (ESBOUGH '
                                                 'MALLORY)',
                                         'poller_hash': 'ODJBNzRE',
                                         'serial_number': '82A74D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:7b:65',
                                         'uptime': '1095995'},
                                 {       'counter_source': '139392/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '10:27:f5:b3:77:79'],
                                                                       'id': '139392/g1',
                                                                       'in_nucast_pkts': '138883',
                                                                       'in_octets': '68533297135',
                                                                       'in_ucast_pkts': '192557822',
                                                                       'mac_address': '48:77:46:f8:7b:11',
                                                                       'mtu': 2000,
                                                                       'name': '139392/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878221',
                                                                       'out_octets': '693255198901',
                                                                       'out_ucast_pkts': '527350078',
                                                                       'poller_hash': 'ODJBNzMxcGVlcmludGYxMzkzOTIvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139392',
                                                                       'in_nucast_pkts': '9878221',
                                                                       'in_octets': '693255198901',
                                                                       'in_ucast_pkts': '527350078',
                                                                       'links': [       {       'in_nucast_pkts': '9878221',
                                                                                                'in_octets': '693255198901',
                                                                                                'in_ucast_pkts': '527350078',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '138883',
                                                                                                'out_octets': '68533297135',
                                                                                                'out_ucast_pkts': '192557822',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '138883',
                                                                       'out_octets': '68533297135',
                                                                       'out_ucast_pkts': '192557822',
                                                                       'poller_hash': 'ODJBNzMxcGVlcmludGZwb24gMTM5Mzky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139392',
                                                                                               'links': [       {       'connected_time': '8535457',
                                                                                                                        'fiber_length': 22427,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNzMxcGVlcnBvbnBvbiAxMzkzOTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '91 POND VIEW DR (BEVERS PAULINE)',
                                         'poller_hash': 'ODJBNzMx',
                                         'serial_number': '82A731',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:7b:11',
                                         'uptime': '8535457'},
                                 {       'counter_source': '139946/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:49'],
                                                                       'id': '139946/g1',
                                                                       'in_nucast_pkts': '73824',
                                                                       'in_octets': '81764913770',
                                                                       'in_ucast_pkts': '529600773',
                                                                       'mac_address': '48:77:46:f8:71:a2',
                                                                       'mtu': 2000,
                                                                       'name': '139946/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9242228',
                                                                       'out_octets': '3112411466480',
                                                                       'out_ucast_pkts': '2180439438',
                                                                       'poller_hash': 'ODJBNDBDcGVlcmludGYxMzk5NDYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139946',
                                                                       'in_nucast_pkts': '9242228',
                                                                       'in_octets': '3112411466480',
                                                                       'in_ucast_pkts': '2180439438',
                                                                       'links': [       {       'in_nucast_pkts': '9242228',
                                                                                                'in_octets': '3112411466480',
                                                                                                'in_ucast_pkts': '2180439438',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '73824',
                                                                                                'out_octets': '81764913770',
                                                                                                'out_ucast_pkts': '529600773',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '73824',
                                                                       'out_octets': '81764913770',
                                                                       'out_ucast_pkts': '529600773',
                                                                       'poller_hash': 'ODJBNDBDcGVlcmludGZwb24gMTM5OTQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139946',
                                                                                               'links': [       {       'connected_time': '8046361',
                                                                                                                        'fiber_length': 23411,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNDBDcGVlcnBvbnBvbiAxMzk5NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1022 MOLESWORTH ST (ROTH STEPHENIE)',
                                         'poller_hash': 'ODJBNDBD',
                                         'serial_number': '82A40C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:71:a2',
                                         'uptime': '8046361'},
                                 {       'counter_source': '139954/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:a7:31:21:45:55'],
                                                                       'id': '139954/g1',
                                                                       'in_nucast_pkts': '4906',
                                                                       'in_octets': '97157547235',
                                                                       'in_ucast_pkts': '295034102',
                                                                       'mac_address': '48:77:46:f8:7b:4a',
                                                                       'mtu': 2000,
                                                                       'name': '139954/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9798032',
                                                                       'out_octets': '1011248290260',
                                                                       'out_ucast_pkts': '822671685',
                                                                       'poller_hash': 'ODJBNzQ0cGVlcmludGYxMzk5NTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '139954',
                                                                       'in_nucast_pkts': '9798032',
                                                                       'in_octets': '1011248290260',
                                                                       'in_ucast_pkts': '822671685',
                                                                       'links': [       {       'in_nucast_pkts': '9798032',
                                                                                                'in_octets': '1011248290260',
                                                                                                'in_ucast_pkts': '822671685',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4906',
                                                                                                'out_octets': '97157547235',
                                                                                                'out_ucast_pkts': '295034102',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4906',
                                                                       'out_octets': '97157547235',
                                                                       'out_ucast_pkts': '295034102',
                                                                       'poller_hash': 'ODJBNzQ0cGVlcmludGZwb24gMTM5OTU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '139954',
                                                                                               'links': [       {       'connected_time': '8535477',
                                                                                                                        'fiber_length': 23946,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNzQ0cGVlcnBvbnBvbiAxMzk5NTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '3697 NAFZIGER RD (DAMMEIER RYAN)',
                                         'poller_hash': 'ODJBNzQ0',
                                         'serial_number': '82A744',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:7b:4a',
                                         'uptime': '8535477'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '140011/g1',
                                                                       'name': '140011/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODQ0QkFCcGVlcmludGYxNDAwMTEvZzE='},
                                                               {       'admin_up': False,
                                                                       'id': '140011',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODQ0QkFCcGVlcmludGZwb24gMTQwMDEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140011',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkFCcGVlcnBvbnBvbiAxNDAwMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '85 VILLAGE RD',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "140011" '
                                                                  '(location: GPON '
                                                                  'port "1/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:26:48.480Z',
                                         'poller_hash': 'ODQ0QkFC',
                                         'serial_number': '844BAB',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '140035/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '140035/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1861246674171',
                                                                       'in_ucast_pkts': '2745302225',
                                                                       'mac_address': 'd0:76:8f:08:15:0b',
                                                                       'name': '140035/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '75',
                                                                       'out_octets': '3810103788860',
                                                                       'out_ucast_pkts': '3897870947',
                                                                       'poller_hash': 'ODY5Mzk3cGVlcmludGYxNDAwMzUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140035',
                                                                       'in_nucast_pkts': '75',
                                                                       'in_octets': '3810103788860',
                                                                       'in_ucast_pkts': '3897870947',
                                                                       'links': [       {       'in_nucast_pkts': '75',
                                                                                                'in_octets': '3810103788860',
                                                                                                'in_ucast_pkts': '3897870947',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1861246674171',
                                                                                                'out_ucast_pkts': '2745302225',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1861246674171',
                                                                       'out_ucast_pkts': '2745302225',
                                                                       'poller_hash': 'ODY5Mzk3cGVlcmludGZwb24gMTQwMDM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140035',
                                                                                               'links': [       {       'connected_time': '1957832',
                                                                                                                        'fiber_length': 23505,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5Mzk3cGVlcnBvbnBvbiAxNDAwMzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3875 NAFZIGER RD (MORDEN WANDA)',
                                         'poller_hash': 'ODY5Mzk3',
                                         'serial_number': '869397',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:0b',
                                         'uptime': '1957832'},
                                 {       'counter_source': '140129/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '140129/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '829876977823',
                                                                       'in_ucast_pkts': '1703935141',
                                                                       'mac_address': 'd0:76:8f:08:15:3b',
                                                                       'name': '140129/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '133',
                                                                       'out_octets': '5731676860218',
                                                                       'out_ucast_pkts': '4593249660',
                                                                       'poller_hash': 'ODY5MzlEcGVlcmludGYxNDAxMjkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140129',
                                                                       'in_nucast_pkts': '133',
                                                                       'in_octets': '5731676860218',
                                                                       'in_ucast_pkts': '4593249660',
                                                                       'links': [       {       'in_nucast_pkts': '133',
                                                                                                'in_octets': '5731676860218',
                                                                                                'in_ucast_pkts': '4593249660',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '829876977823',
                                                                                                'out_ucast_pkts': '1703935141',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '829876977823',
                                                                       'out_ucast_pkts': '1703935141',
                                                                       'poller_hash': 'ODY5MzlEcGVlcmludGZwb24gMTQwMTI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140129',
                                                                                               'links': [       {       'connected_time': '1095959',
                                                                                                                        'fiber_length': 31174,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzlEcGVlcnBvbnBvbiAxNDAxMjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1826 MOSER-YOUNG RD (ENGLISH SCOTT)',
                                         'poller_hash': 'ODY5MzlE',
                                         'serial_number': '86939D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:3b',
                                         'uptime': '1095959'},
                                 {       'counter_source': '140210/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'a8:63:7d:7e:06:20'],
                                                                       'id': '140210/g1',
                                                                       'in_nucast_pkts': '1215',
                                                                       'in_octets': '4457363813',
                                                                       'in_ucast_pkts': '25165602',
                                                                       'mac_address': 'd0:76:8f:06:e5:bf',
                                                                       'mtu': 2000,
                                                                       'name': '140210/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281565',
                                                                       'out_octets': '143200174663',
                                                                       'out_ucast_pkts': '108610249',
                                                                       'poller_hash': 'ODQ0QkQwcGVlcmludGYxNDAyMTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140210',
                                                                       'in_nucast_pkts': '1281565',
                                                                       'in_octets': '143200174663',
                                                                       'in_ucast_pkts': '108610249',
                                                                       'links': [       {       'in_nucast_pkts': '1281565',
                                                                                                'in_octets': '143200174663',
                                                                                                'in_ucast_pkts': '108610249',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1215',
                                                                                                'out_octets': '4457363813',
                                                                                                'out_ucast_pkts': '25165602',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1215',
                                                                       'out_octets': '4457363813',
                                                                       'out_ucast_pkts': '25165602',
                                                                       'poller_hash': 'ODQ0QkQwcGVlcmludGZwb24gMTQwMjEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140210',
                                                                                               'links': [       {       'connected_time': '1096814',
                                                                                                                        'fiber_length': 31430,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkQwcGVlcnBvbnBvbiAxNDAyMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1765 MOSER-YOUNG RD (DEVRIES '
                                                 'CLARENCE & KATIE)',
                                         'poller_hash': 'ODQ0QkQw',
                                         'serial_number': '844BD0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:bf',
                                         'uptime': '1096814'},
                                 {       'counter_source': '140218/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f4:b9'],
                                                                       'id': '140218/g1',
                                                                       'in_nucast_pkts': '23810',
                                                                       'in_octets': '14530958379',
                                                                       'in_ucast_pkts': '85788246',
                                                                       'mac_address': 'd0:76:8f:aa:52:47',
                                                                       'mtu': 2000,
                                                                       'name': '140218/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281592',
                                                                       'out_octets': '254814071063',
                                                                       'out_ucast_pkts': '205930271',
                                                                       'poller_hash': 'OTU2RjZFcGVlcmludGYxNDAyMTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140218',
                                                                       'in_nucast_pkts': '1281592',
                                                                       'in_octets': '254814071063',
                                                                       'in_ucast_pkts': '205930271',
                                                                       'links': [       {       'in_nucast_pkts': '1281592',
                                                                                                'in_octets': '254814071063',
                                                                                                'in_ucast_pkts': '205930271',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23810',
                                                                                                'out_octets': '14530958379',
                                                                                                'out_ucast_pkts': '85788246',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23810',
                                                                       'out_octets': '14530958379',
                                                                       'out_ucast_pkts': '85788246',
                                                                       'poller_hash': 'OTU2RjZFcGVlcmludGZwb24gMTQwMjE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140218',
                                                                                               'links': [       {       'connected_time': '1096792',
                                                                                                                        'fiber_length': 31785,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjZFcGVlcnBvbnBvbiAxNDAyMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3974 WEIMAR LINE (VOGEL BARB & '
                                                 'KLAUS)',
                                         'poller_hash': 'OTU2RjZF',
                                         'serial_number': '956F6E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:52:44',
                                         'uptime': '1096792'},
                                 {       'counter_source': '140222/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ef:b2'],
                                                                       'id': '140222/g1',
                                                                       'in_nucast_pkts': '956906',
                                                                       'in_octets': '314223820053',
                                                                       'in_ucast_pkts': '2728663129',
                                                                       'mac_address': 'd0:76:8f:08:14:e6',
                                                                       'mtu': 2000,
                                                                       'name': '140222/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63494906',
                                                                       'out_octets': '13584104736499',
                                                                       'out_ucast_pkts': '9317655591',
                                                                       'poller_hash': 'ODY5MzkycGVlcmludGYxNDAyMjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140222',
                                                                       'in_nucast_pkts': '63494906',
                                                                       'in_octets': '13584104736499',
                                                                       'in_ucast_pkts': '9317655591',
                                                                       'links': [       {       'in_nucast_pkts': '63494906',
                                                                                                'in_octets': '13584104736499',
                                                                                                'in_ucast_pkts': '9317655591',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '956906',
                                                                                                'out_octets': '314223820053',
                                                                                                'out_ucast_pkts': '2728663129',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '956906',
                                                                       'out_octets': '314223820053',
                                                                       'out_ucast_pkts': '2728663129',
                                                                       'poller_hash': 'ODY5MzkycGVlcmludGZwb24gMTQwMjIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140222',
                                                                                               'links': [       {       'connected_time': '21886560',
                                                                                                                        'fiber_length': 31211,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzkycGVlcnBvbnBvbiAxNDAyMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4107 WEIMAR LINE (BRUBACHER MENNO)',
                                         'poller_hash': 'ODY5Mzky',
                                         'serial_number': '869392',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:14:e3',
                                         'uptime': '21886560'},
                                 {       'counter_source': '140268/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '0c:ea:14:5a:b9:a4'],
                                                                       'id': '140268/g1',
                                                                       'in_nucast_pkts': '3279',
                                                                       'in_octets': '22265003852',
                                                                       'in_ucast_pkts': '110401755',
                                                                       'mac_address': 'd0:76:8f:06:e6:7f',
                                                                       'mtu': 2000,
                                                                       'name': '140268/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281513',
                                                                       'out_octets': '513080923189',
                                                                       'out_ucast_pkts': '360126770',
                                                                       'poller_hash': 'ODQ0QzEwcGVlcmludGYxNDAyNjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140268',
                                                                       'in_nucast_pkts': '1281513',
                                                                       'in_octets': '513080923189',
                                                                       'in_ucast_pkts': '360126770',
                                                                       'links': [       {       'in_nucast_pkts': '1281513',
                                                                                                'in_octets': '513080923189',
                                                                                                'in_ucast_pkts': '360126770',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3279',
                                                                                                'out_octets': '22265003852',
                                                                                                'out_ucast_pkts': '110401755',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3279',
                                                                       'out_octets': '22265003852',
                                                                       'out_ucast_pkts': '110401755',
                                                                       'poller_hash': 'ODQ0QzEwcGVlcmludGZwb24gMTQwMjY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140268',
                                                                                               'links': [       {       'connected_time': '1096808',
                                                                                                                        'fiber_length': 31268,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QzEwcGVlcnBvbnBvbiAxNDAyNjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4124 WEIMAR LINE (BUSCH BRAD)',
                                         'poller_hash': 'ODQ0QzEw',
                                         'serial_number': '844C10',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:7f',
                                         'uptime': '1096808'},
                                 {       'counter_source': '140272/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e0:46:ee:1e:4a:4f'],
                                                                       'id': '140272/g1',
                                                                       'in_nucast_pkts': '3132',
                                                                       'in_octets': '699317529801',
                                                                       'in_ucast_pkts': '1378589442',
                                                                       'mac_address': '04:bc:9f:5b:9b:dc',
                                                                       'mtu': 2000,
                                                                       'name': '140272/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12380103',
                                                                       'out_octets': '4263425818267',
                                                                       'out_ucast_pkts': '3544884784',
                                                                       'poller_hash': 'MTI3MzlGQnBlZXJpbnRmMTQwMjcyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140272',
                                                                       'in_nucast_pkts': '12380103',
                                                                       'in_octets': '4263425818267',
                                                                       'in_ucast_pkts': '3544884784',
                                                                       'links': [       {       'in_nucast_pkts': '12380103',
                                                                                                'in_octets': '4263425818267',
                                                                                                'in_ucast_pkts': '3544884784',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3132',
                                                                                                'out_octets': '699317529801',
                                                                                                'out_ucast_pkts': '1378589442',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3132',
                                                                       'out_octets': '699317529801',
                                                                       'out_ucast_pkts': '1378589442',
                                                                       'poller_hash': 'MTI3MzlGQnBlZXJpbnRmcG9uIDE0MDI3Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140272',
                                                                                               'links': [       {       'connected_time': '10709012',
                                                                                                                        'fiber_length': 31146,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MzlGQnBlZXJwb25wb24gMTQwMjcy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1818 MOSER-YOUNG RD (DOYLE MATTHEW & '
                                                 'ALYSSA)',
                                         'poller_hash': 'MTI3MzlGQg==',
                                         'serial_number': '12739FB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:9b:d9',
                                         'uptime': '10709012'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '140286/g1',
                                                                       'name': '140286/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODQ0QkEzcGVlcmludGYxNDAyODYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140286',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODQ0QkEzcGVlcmludGZwb24gMTQwMjg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140286',
                                                                                               'links': [       {       'connected_time': '1096812',
                                                                                                                        'fiber_length': 31974,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkEzcGVlcnBvbnBvbiAxNDAyODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '3952 WEIMAR LINE',
                                         'poller_hash': 'ODQ0QkEz',
                                         'serial_number': '844BA3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:38',
                                         'uptime': '1096812'},
                                 {       'counter_source': '140290/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '140290/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '350786853501',
                                                                       'in_ucast_pkts': '1839441079',
                                                                       'mac_address': 'd0:76:8f:08:1c:13',
                                                                       'name': '140290/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '94',
                                                                       'out_octets': '10531367458131',
                                                                       'out_ucast_pkts': '7457703928',
                                                                       'poller_hash': 'ODY5NDc4cGVlcmludGYxNDAyOTAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140290',
                                                                       'in_nucast_pkts': '94',
                                                                       'in_octets': '10531367458131',
                                                                       'in_ucast_pkts': '7457703928',
                                                                       'links': [       {       'in_nucast_pkts': '94',
                                                                                                'in_octets': '10531367458131',
                                                                                                'in_ucast_pkts': '7457703928',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '350786853501',
                                                                                                'out_ucast_pkts': '1839441079',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '350786853501',
                                                                       'out_ucast_pkts': '1839441079',
                                                                       'poller_hash': 'ODY5NDc4cGVlcmludGZwb24gMTQwMjkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140290',
                                                                                               'links': [       {       'connected_time': '13190710',
                                                                                                                        'fiber_length': 31223,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDc4cGVlcnBvbnBvbiAxNDAyOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4080 WEIMAR LINE (THOMPSON DALE)',
                                         'poller_hash': 'ODY5NDc4',
                                         'serial_number': '869478',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1c:13',
                                         'uptime': '13190710'},
                                 {       'counter_source': '140294/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:e4:c8'],
                                                                       'id': '140294/g1',
                                                                       'in_nucast_pkts': '19565',
                                                                       'in_octets': '15669052751',
                                                                       'in_ucast_pkts': '113045811',
                                                                       'mac_address': 'd0:76:8f:08:16:9e',
                                                                       'mtu': 2000,
                                                                       'name': '140294/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281574',
                                                                       'out_octets': '445470103349',
                                                                       'out_ucast_pkts': '338398126',
                                                                       'poller_hash': 'ODY5M0M5cGVlcmludGYxNDAyOTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140294',
                                                                       'in_nucast_pkts': '1281574',
                                                                       'in_octets': '445470103349',
                                                                       'in_ucast_pkts': '338398126',
                                                                       'links': [       {       'in_nucast_pkts': '1281574',
                                                                                                'in_octets': '445470103349',
                                                                                                'in_ucast_pkts': '338398126',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19565',
                                                                                                'out_octets': '15669052751',
                                                                                                'out_ucast_pkts': '113045811',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19565',
                                                                       'out_octets': '15669052751',
                                                                       'out_ucast_pkts': '113045811',
                                                                       'poller_hash': 'ODY5M0M5cGVlcmludGZwb24gMTQwMjk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140294',
                                                                                               'links': [       {       'connected_time': '1096772',
                                                                                                                        'fiber_length': 31244,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0M5cGVlcnBvbnBvbiAxNDAyOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4090 WEIMAR LINE (ENGLER PAULA)',
                                         'poller_hash': 'ODY5M0M5',
                                         'serial_number': '8693C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:9b',
                                         'uptime': '1096772'},
                                 {       'counter_source': '140304/F1+140304/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:17:74:cc'],
                                                                       'id': '140304/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '812256915',
                                                                       'in_ucast_pkts': '4687679',
                                                                       'name': '140304/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '7099317202',
                                                                       'out_octets': '9433598645360',
                                                                       'out_ucast_pkts': '386910',
                                                                       'poller_hash': 'ODY5M0IwcGVlcmludGYxNDAzMDQvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e8:7d'],
                                                                       'id': '140304/g1',
                                                                       'in_nucast_pkts': '19303',
                                                                       'in_octets': '16547268274',
                                                                       'in_ucast_pkts': '35995753',
                                                                       'mac_address': 'd0:76:8f:08:15:d6',
                                                                       'mtu': 2000,
                                                                       'name': '140304/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281563',
                                                                       'out_octets': '156595681738',
                                                                       'out_ucast_pkts': '126213216',
                                                                       'poller_hash': 'ODY5M0IwcGVlcmludGYxNDAzMDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140304',
                                                                       'in_nucast_pkts': '7100598765',
                                                                       'in_octets': '9590194327098',
                                                                       'in_ucast_pkts': '126600126',
                                                                       'links': [       {       'in_nucast_pkts': '7100598765',
                                                                                                'in_octets': '9590194327098',
                                                                                                'in_ucast_pkts': '126600126',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19303',
                                                                                                'out_octets': '17359525189',
                                                                                                'out_ucast_pkts': '40683432',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19303',
                                                                       'out_octets': '17359525189',
                                                                       'out_ucast_pkts': '40683432',
                                                                       'poller_hash': 'ODY5M0IwcGVlcmludGZwb24gMTQwMzA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140304',
                                                                                               'links': [       {       'connected_time': '1096784',
                                                                                                                        'fiber_length': 30629,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0IwcGVlcnBvbnBvbiAxNDAzMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4175 WEIMAR LINE (MARSHALL DAVID)',
                                         'poller_hash': 'ODY5M0Iw',
                                         'serial_number': '8693B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:d3',
                                         'uptime': '1096784'},
                                 {       'counter_source': '140347/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:cc'],
                                                                       'id': '140347/g1',
                                                                       'in_nucast_pkts': '28682',
                                                                       'in_octets': '515225227',
                                                                       'in_ucast_pkts': '4348812',
                                                                       'mac_address': 'd0:76:8f:08:13:5e',
                                                                       'mtu': 2000,
                                                                       'name': '140347/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1610067',
                                                                       'out_octets': '12008508010',
                                                                       'out_ucast_pkts': '8858251',
                                                                       'poller_hash': 'ODY5MzYxcGVlcmludGYxNDAzNDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140347',
                                                                       'in_nucast_pkts': '1610067',
                                                                       'in_octets': '12008508010',
                                                                       'in_ucast_pkts': '8858251',
                                                                       'links': [       {       'in_nucast_pkts': '1610067',
                                                                                                'in_octets': '12008508010',
                                                                                                'in_ucast_pkts': '8858251',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '28682',
                                                                                                'out_octets': '515225227',
                                                                                                'out_ucast_pkts': '4348812',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '28682',
                                                                       'out_octets': '515225227',
                                                                       'out_ucast_pkts': '4348812',
                                                                       'poller_hash': 'ODY5MzYxcGVlcmludGZwb24gMTQwMzQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140347',
                                                                                               'links': [       {       'connected_time': '1379193',
                                                                                                                        'fiber_length': 31465,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzYxcGVlcnBvbnBvbiAxNDAzNDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4019 WEIMAR LINE (ZIEGLER CAROL)',
                                         'poller_hash': 'ODY5MzYx',
                                         'serial_number': '869361',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:13:5b',
                                         'uptime': '1379193'},
                                 {       'counter_source': '140357/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f3:dd'],
                                                                       'id': '140357/g1',
                                                                       'in_nucast_pkts': '228857',
                                                                       'in_octets': '34189912735',
                                                                       'in_ucast_pkts': '204512582',
                                                                       'mac_address': 'd0:76:8f:06:e6:61',
                                                                       'mtu': 2000,
                                                                       'name': '140357/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920163',
                                                                       'out_octets': '830551159619',
                                                                       'out_ucast_pkts': '621946806',
                                                                       'poller_hash': 'ODQ0QzA2cGVlcmludGYxNDAzNTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140357',
                                                                       'in_nucast_pkts': '9920163',
                                                                       'in_octets': '830551159619',
                                                                       'in_ucast_pkts': '621946806',
                                                                       'links': [       {       'in_nucast_pkts': '9920163',
                                                                                                'in_octets': '830551159619',
                                                                                                'in_ucast_pkts': '621946806',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '228857',
                                                                                                'out_octets': '34189912735',
                                                                                                'out_ucast_pkts': '204512582',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '228857',
                                                                       'out_octets': '34189912735',
                                                                       'out_ucast_pkts': '204512582',
                                                                       'poller_hash': 'ODQ0QzA2cGVlcmludGZwb24gMTQwMzU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140357',
                                                                                               'links': [       {       'connected_time': '8535529',
                                                                                                                        'fiber_length': 23546,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QzA2cGVlcnBvbnBvbiAxNDAzNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '82 SCHWEITZER CR (BOESE JEFF)',
                                         'poller_hash': 'ODQ0QzA2',
                                         'serial_number': '844C06',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:61',
                                         'uptime': '8535529'},
                                 {       'counter_source': '140361/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:09:0d:25:0d:4f'],
                                                                       'id': '140361/g1',
                                                                       'in_nucast_pkts': '8231',
                                                                       'in_octets': '146626809801',
                                                                       'in_ucast_pkts': '830358802',
                                                                       'mac_address': 'd0:76:8f:08:0f:1e',
                                                                       'mtu': 2000,
                                                                       'name': '140361/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4108046',
                                                                       'out_octets': '2304199635484',
                                                                       'out_ucast_pkts': '1661201075',
                                                                       'poller_hash': 'ODY5MkQ5cGVlcmludGYxNDAzNjEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140361',
                                                                       'in_nucast_pkts': '4108046',
                                                                       'in_octets': '2304199635484',
                                                                       'in_ucast_pkts': '1661201075',
                                                                       'links': [       {       'in_nucast_pkts': '4108046',
                                                                                                'in_octets': '2304199635484',
                                                                                                'in_ucast_pkts': '1661201075',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '8231',
                                                                                                'out_octets': '146626809801',
                                                                                                'out_ucast_pkts': '830358802',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '8231',
                                                                       'out_octets': '146626809801',
                                                                       'out_ucast_pkts': '830358802',
                                                                       'poller_hash': 'ODY5MkQ5cGVlcmludGZwb24gMTQwMzYx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140361',
                                                                                               'links': [       {       'connected_time': '3511700',
                                                                                                                        'fiber_length': 22462,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MkQ5cGVlcnBvbnBvbiAxNDAzNjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1317 QUEEN'S BUSH RD (MACARTHUR "
                                                 'JAMES)',
                                         'poller_hash': 'ODY5MkQ5',
                                         'serial_number': '8692D9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:0f:1b',
                                         'uptime': '3511700'},
                                 {       'counter_source': '140371/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '28:bd:89:ed:93:51'],
                                                                       'id': '140371/g1',
                                                                       'in_nucast_pkts': '13089',
                                                                       'in_octets': '186116109048',
                                                                       'in_ucast_pkts': '301820818',
                                                                       'mac_address': 'd0:76:8f:06:e6:67',
                                                                       'mtu': 2000,
                                                                       'name': '140371/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281577',
                                                                       'out_octets': '485487265313',
                                                                       'out_ucast_pkts': '439549856',
                                                                       'poller_hash': 'ODQ0QzA4cGVlcmludGYxNDAzNzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140371',
                                                                       'in_nucast_pkts': '1281577',
                                                                       'in_octets': '485487265313',
                                                                       'in_ucast_pkts': '439549856',
                                                                       'links': [       {       'in_nucast_pkts': '1281577',
                                                                                                'in_octets': '485487265313',
                                                                                                'in_ucast_pkts': '439549856',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13089',
                                                                                                'out_octets': '186116109048',
                                                                                                'out_ucast_pkts': '301820818',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13089',
                                                                       'out_octets': '186116109048',
                                                                       'out_ucast_pkts': '301820818',
                                                                       'poller_hash': 'ODQ0QzA4cGVlcmludGZwb24gMTQwMzcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140371',
                                                                                               'links': [       {       'connected_time': '1096808',
                                                                                                                        'fiber_length': 30031,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QzA4cGVlcnBvbnBvbiAxNDAzNzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4187 WEIMAR LINE (JANZEN TERRY)',
                                         'poller_hash': 'ODQ0QzA4',
                                         'serial_number': '844C08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:67',
                                         'uptime': '1096808'},
                                 {       'counter_source': '140379/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f7:9a'],
                                                                       'id': '140379/g1',
                                                                       'in_nucast_pkts': '171',
                                                                       'in_octets': '5412691680',
                                                                       'in_ucast_pkts': '38344866',
                                                                       'mac_address': 'd0:76:8f:08:13:e6',
                                                                       'mtu': 2000,
                                                                       'name': '140379/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281556',
                                                                       'out_octets': '214581424236',
                                                                       'out_ucast_pkts': '145998818',
                                                                       'poller_hash': 'ODY5MzcycGVlcmludGYxNDAzNzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140379',
                                                                       'in_nucast_pkts': '1281556',
                                                                       'in_octets': '214581424236',
                                                                       'in_ucast_pkts': '145998818',
                                                                       'links': [       {       'in_nucast_pkts': '1281556',
                                                                                                'in_octets': '214581424236',
                                                                                                'in_ucast_pkts': '145998818',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '171',
                                                                                                'out_octets': '5412691680',
                                                                                                'out_ucast_pkts': '38344866',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '171',
                                                                       'out_octets': '5412691680',
                                                                       'out_ucast_pkts': '38344866',
                                                                       'poller_hash': 'ODY5MzcycGVlcmludGZwb24gMTQwMzc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140379',
                                                                                               'links': [       {       'connected_time': '1096772',
                                                                                                                        'fiber_length': 31317,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzcycGVlcnBvbnBvbiAxNDAzNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4140 WEIMAR LINE (FINK DOUGLAS)',
                                         'poller_hash': 'ODY5Mzcy',
                                         'serial_number': '869372',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:13:e3',
                                         'uptime': '1096772'},
                                 {       'counter_source': '140431/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:d9:f5:a3:af:d0'],
                                                                       'id': '140431/g1',
                                                                       'in_nucast_pkts': '13',
                                                                       'in_octets': '170219989',
                                                                       'in_ucast_pkts': '174852',
                                                                       'mac_address': 'd0:76:8f:06:e5:56',
                                                                       'mtu': 2000,
                                                                       'name': '140431/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '69388',
                                                                       'out_octets': '61432048',
                                                                       'out_ucast_pkts': '98161',
                                                                       'poller_hash': 'ODQ0QkFEcGVlcmludGYxNDA0MzEvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140431',
                                                                       'in_nucast_pkts': '69388',
                                                                       'in_octets': '61432048',
                                                                       'in_ucast_pkts': '98161',
                                                                       'links': [       {       'in_nucast_pkts': '69388',
                                                                                                'in_octets': '61432048',
                                                                                                'in_ucast_pkts': '98161',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13',
                                                                                                'out_octets': '170219989',
                                                                                                'out_ucast_pkts': '174852',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13',
                                                                       'out_octets': '170219989',
                                                                       'out_ucast_pkts': '174852',
                                                                       'poller_hash': 'ODQ0QkFEcGVlcmludGZwb24gMTQwNDMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140431',
                                                                                               'links': [       {       'connected_time': '58684',
                                                                                                                        'fiber_length': 29744,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkFEcGVlcnBvbnBvbiAxNDA0MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4575 WEIMAR LINE (SHERK AMOS)',
                                         'poller_hash': 'ODQ0QkFE',
                                         'serial_number': '844BAD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:56',
                                         'uptime': '58684'},
                                 {       'counter_source': '140481/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '140481/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '15',
                                                                       'in_octets': '188727843006',
                                                                       'in_ucast_pkts': '1013487617',
                                                                       'mac_address': 'd0:76:8f:08:13:ab',
                                                                       'name': '140481/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '750',
                                                                       'out_octets': '5817162045058',
                                                                       'out_ucast_pkts': '4096675559',
                                                                       'poller_hash': 'ODY5MzZCcGVlcmludGYxNDA0ODEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140481',
                                                                       'in_nucast_pkts': '750',
                                                                       'in_octets': '5817162045058',
                                                                       'in_ucast_pkts': '4096675559',
                                                                       'links': [       {       'in_nucast_pkts': '750',
                                                                                                'in_octets': '5817162045058',
                                                                                                'in_ucast_pkts': '4096675559',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15',
                                                                                                'out_octets': '188727843006',
                                                                                                'out_ucast_pkts': '1013487617',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15',
                                                                       'out_octets': '188727843006',
                                                                       'out_ucast_pkts': '1013487617',
                                                                       'poller_hash': 'ODY5MzZCcGVlcmludGZwb24gMTQwNDgx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140481',
                                                                                               'links': [       {       'connected_time': '8535236',
                                                                                                                        'fiber_length': 23196,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzZCcGVlcnBvbnBvbiAxNDA0ODE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1133 MOLESWORTH ST (BEDELL MEGAN)',
                                         'poller_hash': 'ODY5MzZC',
                                         'serial_number': '86936B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:13:ab',
                                         'uptime': '8535236'},
                                 {       'counter_source': '140499/F1+140499/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:dd:31:a8'],
                                                                       'id': '140499/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '559442610',
                                                                       'in_ucast_pkts': '3314095',
                                                                       'name': '140499/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '4515805187',
                                                                       'out_octets': '6019676826744',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODY5MzY4cGVlcmludGYxNDA0OTkvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:6a'],
                                                                       'id': '140499/g1',
                                                                       'in_nucast_pkts': '23926',
                                                                       'in_octets': '29614914623',
                                                                       'in_ucast_pkts': '114636377',
                                                                       'mac_address': 'd0:76:8f:08:13:96',
                                                                       'mtu': 2000,
                                                                       'name': '140499/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281577',
                                                                       'out_octets': '387515788402',
                                                                       'out_ucast_pkts': '341713722',
                                                                       'poller_hash': 'ODY5MzY4cGVlcmludGYxNDA0OTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140499',
                                                                       'in_nucast_pkts': '4517086764',
                                                                       'in_octets': '6407192615146',
                                                                       'in_ucast_pkts': '341713722',
                                                                       'links': [       {       'in_nucast_pkts': '4517086764',
                                                                                                'in_octets': '6407192615146',
                                                                                                'in_ucast_pkts': '341713722',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23926',
                                                                                                'out_octets': '30174357233',
                                                                                                'out_ucast_pkts': '117950472',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23926',
                                                                       'out_octets': '30174357233',
                                                                       'out_ucast_pkts': '117950472',
                                                                       'poller_hash': 'ODY5MzY4cGVlcmludGZwb24gMTQwNDk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140499',
                                                                                               'links': [       {       'connected_time': '1096785',
                                                                                                                        'fiber_length': 30109,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzY4cGVlcnBvbnBvbiAxNDA0OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4211 WEIMAR LINE (MARSHALL DAVID)',
                                         'poller_hash': 'ODY5MzY4',
                                         'serial_number': '869368',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:13:93',
                                         'uptime': '1096785'},
                                 {       'counter_source': '140516/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '140516/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '16525172990',
                                                                       'in_ucast_pkts': '65487729',
                                                                       'mac_address': 'd0:76:8f:07:19:2e',
                                                                       'name': '140516/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '90122780074',
                                                                       'out_ucast_pkts': '135168013',
                                                                       'poller_hash': 'ODY3OEFCcGVlcmludGYxNDA1MTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140516',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '90122780074',
                                                                       'in_ucast_pkts': '135168013',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '90122780074',
                                                                                                'in_ucast_pkts': '135168013',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '16525172990',
                                                                                                'out_ucast_pkts': '65487729',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '16525172990',
                                                                       'out_ucast_pkts': '65487729',
                                                                       'poller_hash': 'ODY3OEFCcGVlcmludGZwb24gMTQwNTE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140516',
                                                                                               'links': [       {       'connected_time': '8535333',
                                                                                                                        'fiber_length': 22348,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY3OEFCcGVlcnBvbnBvbiAxNDA1MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': "1235 QUEEN'S BUSH RD (HORST LOIS)",
                                         'poller_hash': 'ODY3OEFC',
                                         'serial_number': '8678AB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:07:19:2e',
                                         'uptime': '8535333'},
                                 {       'counter_source': '140531/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e2:5c'],
                                                                       'id': '140531/g1',
                                                                       'in_nucast_pkts': '686006',
                                                                       'in_octets': '256361278074',
                                                                       'in_ucast_pkts': '486178765',
                                                                       'mac_address': 'd0:76:8f:06:e5:a1',
                                                                       'mtu': 2000,
                                                                       'name': '140531/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39420612',
                                                                       'out_octets': '1656573708180',
                                                                       'out_ucast_pkts': '1363946305',
                                                                       'poller_hash': 'ODQ0QkM2cGVlcmludGYxNDA1MzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140531',
                                                                       'in_nucast_pkts': '39420612',
                                                                       'in_octets': '1656573708180',
                                                                       'in_ucast_pkts': '1363946305',
                                                                       'links': [       {       'in_nucast_pkts': '39420612',
                                                                                                'in_octets': '1656573708180',
                                                                                                'in_ucast_pkts': '1363946305',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '686006',
                                                                                                'out_octets': '256361278074',
                                                                                                'out_ucast_pkts': '486178765',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '686006',
                                                                       'out_octets': '256361278074',
                                                                       'out_ucast_pkts': '486178765',
                                                                       'poller_hash': 'ODQ0QkM2cGVlcmludGZwb24gMTQwNTMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140531',
                                                                                               'links': [       {       'connected_time': '67856572',
                                                                                                                        'fiber_length': 30611,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkM2cGVlcnBvbnBvbiAxNDA1MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4161 WEIMAR LINE (KAYS ROSALIND)',
                                         'poller_hash': 'ODQ0QkM2',
                                         'serial_number': '844BC6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:a1',
                                         'uptime': '67856572'},
                                 {       'counter_source': '140535/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:eb:ea'],
                                                                       'id': '140535/g1',
                                                                       'in_nucast_pkts': '54064',
                                                                       'in_octets': '22510722159',
                                                                       'in_ucast_pkts': '303682048',
                                                                       'mac_address': 'd0:76:8f:08:17:36',
                                                                       'mtu': 2000,
                                                                       'name': '140535/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3535919',
                                                                       'out_octets': '1961984988155',
                                                                       'out_ucast_pkts': '1298539853',
                                                                       'poller_hash': 'ODY5M0RDcGVlcmludGYxNDA1MzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140535',
                                                                       'in_nucast_pkts': '3535919',
                                                                       'in_octets': '1961984988155',
                                                                       'in_ucast_pkts': '1298539853',
                                                                       'links': [       {       'in_nucast_pkts': '3535919',
                                                                                                'in_octets': '1961984988155',
                                                                                                'in_ucast_pkts': '1298539853',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '54064',
                                                                                                'out_octets': '22510722159',
                                                                                                'out_ucast_pkts': '303682048',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '54064',
                                                                       'out_octets': '22510722159',
                                                                       'out_ucast_pkts': '303682048',
                                                                       'poller_hash': 'ODY5M0RDcGVlcmludGZwb24gMTQwNTM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140535',
                                                                                               'links': [       {       'connected_time': '3015841',
                                                                                                                        'fiber_length': 31751,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0RDcGVlcnBvbnBvbiAxNDA1MzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3998 WEIMAR LINE (YOUNG DENNIS '
                                                 '10436227)',
                                         'poller_hash': 'ODY5M0RD',
                                         'serial_number': '8693DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:33',
                                         'uptime': '3015841'},
                                 {       'counter_source': '140559/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:c2'],
                                                                       'id': '140559/g1',
                                                                       'in_nucast_pkts': '19474',
                                                                       'in_octets': '17833422367',
                                                                       'in_ucast_pkts': '94597598',
                                                                       'mac_address': 'd0:76:8f:08:17:be',
                                                                       'mtu': 2000,
                                                                       'name': '140559/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281609',
                                                                       'out_octets': '241151493572',
                                                                       'out_ucast_pkts': '200559929',
                                                                       'poller_hash': 'ODY5M0VEcGVlcmludGYxNDA1NTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140559',
                                                                       'in_nucast_pkts': '1281609',
                                                                       'in_octets': '241151493572',
                                                                       'in_ucast_pkts': '200559929',
                                                                       'links': [       {       'in_nucast_pkts': '1281609',
                                                                                                'in_octets': '241151493572',
                                                                                                'in_ucast_pkts': '200559929',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19474',
                                                                                                'out_octets': '17833422367',
                                                                                                'out_ucast_pkts': '94597598',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19474',
                                                                       'out_octets': '17833422367',
                                                                       'out_ucast_pkts': '94597598',
                                                                       'poller_hash': 'ODY5M0VEcGVlcmludGZwb24gMTQwNTU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140559',
                                                                                               'links': [       {       'connected_time': '1096784',
                                                                                                                        'fiber_length': 31560,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0VEcGVlcnBvbnBvbiAxNDA1NTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4018 WEIMAR LINE (JACOBS LISA '
                                                 '10436609)',
                                         'poller_hash': 'ODY5M0VE',
                                         'serial_number': '8693ED',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:bb',
                                         'uptime': '1096784'},
                                 {       'counter_source': '140609/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:74'],
                                                                       'id': '140609/g4',
                                                                       'in_nucast_pkts': '41563853',
                                                                       'in_octets': '65091209116',
                                                                       'in_ucast_pkts': '391690963',
                                                                       'mac_address': 'd0:76:8f:08:15:99',
                                                                       'mtu': 2000,
                                                                       'name': '140609/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26025772601',
                                                                       'out_octets': '35834990181369',
                                                                       'out_ucast_pkts': '1222387296',
                                                                       'poller_hash': 'ODY5M0E4cGVlcmludGYxNDA2MDkvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140609',
                                                                       'in_nucast_pkts': '26025772601',
                                                                       'in_octets': '35834990181369',
                                                                       'in_ucast_pkts': '1222387296',
                                                                       'links': [       {       'in_nucast_pkts': '26025772601',
                                                                                                'in_octets': '35834990181369',
                                                                                                'in_ucast_pkts': '1222387296',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '41563853',
                                                                                                'out_octets': '65091209116',
                                                                                                'out_ucast_pkts': '391690963',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '41563853',
                                                                       'out_octets': '65091209116',
                                                                       'out_ucast_pkts': '391690963',
                                                                       'poller_hash': 'ODY5M0E4cGVlcmludGZwb24gMTQwNjA5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140609',
                                                                                               'links': [       {       'connected_time': '21887461',
                                                                                                                        'fiber_length': 22813,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0E4cGVlcnBvbnBvbiAxNDA2MDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1064 DOERING ST (BAST KIM)',
                                         'poller_hash': 'ODY5M0E4',
                                         'serial_number': '8693A8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:93',
                                         'uptime': '21887461'},
                                 {       'counter_source': '140619/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f4:82'],
                                                                       'id': '140619/g1',
                                                                       'in_nucast_pkts': '163727',
                                                                       'in_octets': '155101260533',
                                                                       'in_ucast_pkts': '547763083',
                                                                       'mac_address': 'd0:76:8f:08:16:66',
                                                                       'mtu': 2000,
                                                                       'name': '140619/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920149',
                                                                       'out_octets': '3622209551857',
                                                                       'out_ucast_pkts': '2629295215',
                                                                       'poller_hash': 'ODY5M0MycGVlcmludGYxNDA2MTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140619',
                                                                       'in_nucast_pkts': '9920149',
                                                                       'in_octets': '3622209551857',
                                                                       'in_ucast_pkts': '2629295215',
                                                                       'links': [       {       'in_nucast_pkts': '9920149',
                                                                                                'in_octets': '3622209551857',
                                                                                                'in_ucast_pkts': '2629295215',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '163727',
                                                                                                'out_octets': '155101260533',
                                                                                                'out_ucast_pkts': '547763083',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '163727',
                                                                       'out_octets': '155101260533',
                                                                       'out_ucast_pkts': '547763083',
                                                                       'poller_hash': 'ODY5M0MycGVlcmludGZwb24gMTQwNjE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140619',
                                                                                               'links': [       {       'connected_time': '8535273',
                                                                                                                        'fiber_length': 23058,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0MycGVlcnBvbnBvbiAxNDA2MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1112 MOLESWORTH ST (MILLER CHERIE)',
                                         'poller_hash': 'ODY5M0My',
                                         'serial_number': '8693C2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:16:63',
                                         'uptime': '8535273'},
                                 {       'counter_source': '140629/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f2:eb'],
                                                                       'id': '140629/g1',
                                                                       'in_nucast_pkts': '168044',
                                                                       'in_octets': '59245524239',
                                                                       'in_ucast_pkts': '206049500',
                                                                       'mac_address': '48:77:46:f8:74:3c',
                                                                       'mtu': 2000,
                                                                       'name': '140629/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797903',
                                                                       'out_octets': '951330844339',
                                                                       'out_ucast_pkts': '715510713',
                                                                       'poller_hash': 'ODJBNEVBcGVlcmludGYxNDA2MjkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140629',
                                                                       'in_nucast_pkts': '9797903',
                                                                       'in_octets': '951330844339',
                                                                       'in_ucast_pkts': '715510713',
                                                                       'links': [       {       'in_nucast_pkts': '9797903',
                                                                                                'in_octets': '951330844339',
                                                                                                'in_ucast_pkts': '715510713',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '168044',
                                                                                                'out_octets': '59245524239',
                                                                                                'out_ucast_pkts': '206049500',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '168044',
                                                                       'out_octets': '59245524239',
                                                                       'out_ucast_pkts': '206049500',
                                                                       'poller_hash': 'ODJBNEVBcGVlcmludGZwb24gMTQwNjI5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140629',
                                                                                               'links': [       {       'connected_time': '8535552',
                                                                                                                        'fiber_length': 23304,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODJBNEVBcGVlcnBvbnBvbiAxNDA2Mjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': "1185 QUEEN'S BUSH RD (SAUDER "
                                                 'JONATHAN)',
                                         'poller_hash': 'ODJBNEVB',
                                         'serial_number': '82A4EA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '48:77:46:f8:74:3c',
                                         'uptime': '8535552'},
                                 {       'counter_source': '140633/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:84:6a:75:20:33'],
                                                                       'id': '140633/g1',
                                                                       'in_nucast_pkts': '16317744',
                                                                       'in_octets': '415440047427',
                                                                       'in_ucast_pkts': '2012799458',
                                                                       'mac_address': 'd0:76:8f:06:e6:73',
                                                                       'mtu': 2000,
                                                                       'name': '140633/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '39415087',
                                                                       'out_octets': '6492118150102',
                                                                       'out_ucast_pkts': '5017537547',
                                                                       'poller_hash': 'ODQ0QzBDcGVlcmludGYxNDA2MzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140633',
                                                                       'in_nucast_pkts': '39415087',
                                                                       'in_octets': '6492118150102',
                                                                       'in_ucast_pkts': '5017537547',
                                                                       'links': [       {       'in_nucast_pkts': '39415087',
                                                                                                'in_octets': '6492118150102',
                                                                                                'in_ucast_pkts': '5017537547',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16317744',
                                                                                                'out_octets': '415440047427',
                                                                                                'out_ucast_pkts': '2012799458',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16317744',
                                                                       'out_octets': '415440047427',
                                                                       'out_ucast_pkts': '2012799458',
                                                                       'poller_hash': 'ODQ0QzBDcGVlcmludGZwb24gMTQwNjMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140633',
                                                                                               'links': [       {       'connected_time': '67856812',
                                                                                                                        'fiber_length': 31540,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QzBDcGVlcnBvbnBvbiAxNDA2MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1732 MOSER-YOUNG RD (MARTIN MELANIE)',
                                         'poller_hash': 'ODQ0QzBD',
                                         'serial_number': '844C0C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:73',
                                         'uptime': '67856812'},
                                 {       'counter_source': '140637/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '88:de:7c:2b:cf:10'],
                                                                       'id': '140637/g1',
                                                                       'in_nucast_pkts': '46',
                                                                       'in_octets': '11821073280',
                                                                       'in_ucast_pkts': '34255260',
                                                                       'mac_address': 'd0:76:8f:06:e6:3a',
                                                                       'mtu': 2000,
                                                                       'name': '140637/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281586',
                                                                       'out_octets': '167063190417',
                                                                       'out_ucast_pkts': '128668049',
                                                                       'poller_hash': 'ODQ0QkY5cGVlcmludGYxNDA2MzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140637',
                                                                       'in_nucast_pkts': '1281586',
                                                                       'in_octets': '167063190417',
                                                                       'in_ucast_pkts': '128668049',
                                                                       'links': [       {       'in_nucast_pkts': '1281586',
                                                                                                'in_octets': '167063190417',
                                                                                                'in_ucast_pkts': '128668049',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '46',
                                                                                                'out_octets': '11821073280',
                                                                                                'out_ucast_pkts': '34255260',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '46',
                                                                       'out_octets': '11821073280',
                                                                       'out_ucast_pkts': '34255260',
                                                                       'poller_hash': 'ODQ0QkY5cGVlcmludGZwb24gMTQwNjM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140637',
                                                                                               'links': [       {       'connected_time': '1096810',
                                                                                                                        'fiber_length': 31364,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkY5cGVlcnBvbnBvbiAxNDA2Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4025 WEIMAR LINE (NENADOV DAVID)',
                                         'poller_hash': 'ODQ0QkY5',
                                         'serial_number': '844BF9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:3a',
                                         'uptime': '1096810'},
                                 {       'counter_source': '140689/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:84:6a:0e:82:1b'],
                                                                       'id': '140689/g1',
                                                                       'in_nucast_pkts': '7404502',
                                                                       'in_octets': '69432025952',
                                                                       'in_ucast_pkts': '576668236',
                                                                       'mac_address': 'd0:76:8f:06:e5:fb',
                                                                       'mtu': 2000,
                                                                       'name': '140689/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12454595',
                                                                       'out_octets': '1818211347054',
                                                                       'out_ucast_pkts': '1304205823',
                                                                       'poller_hash': 'ODQ0QkU0cGVlcmludGYxNDA2ODkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140689',
                                                                       'in_nucast_pkts': '12454595',
                                                                       'in_octets': '1818211347054',
                                                                       'in_ucast_pkts': '1304205823',
                                                                       'links': [       {       'in_nucast_pkts': '12454595',
                                                                                                'in_octets': '1818211347054',
                                                                                                'in_ucast_pkts': '1304205823',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7404502',
                                                                                                'out_octets': '69432025952',
                                                                                                'out_ucast_pkts': '576668236',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7404502',
                                                                       'out_octets': '69432025952',
                                                                       'out_ucast_pkts': '576668236',
                                                                       'poller_hash': 'ODQ0QkU0cGVlcmludGZwb24gMTQwNjg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140689',
                                                                                               'links': [       {       'connected_time': '10732600',
                                                                                                                        'fiber_length': 23554,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkU0cGVlcnBvbnBvbiAxNDA2ODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '78 SCHWEITZER CR (WEBBER DAVID)',
                                         'poller_hash': 'ODQ0QkU0',
                                         'serial_number': '844BE4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e5:fb',
                                         'uptime': '10732600'},
                                 {       'counter_source': '140693/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ea:53'],
                                                                       'id': '140693/g1',
                                                                       'in_nucast_pkts': '161085',
                                                                       'in_octets': '266086606275',
                                                                       'in_ucast_pkts': '965700022',
                                                                       'mac_address': 'd0:76:8f:06:e6:25',
                                                                       'mtu': 2000,
                                                                       'name': '140693/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797872',
                                                                       'out_octets': '3596175167022',
                                                                       'out_ucast_pkts': '2740061179',
                                                                       'poller_hash': 'ODQ0QkYycGVlcmludGYxNDA2OTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140693',
                                                                       'in_nucast_pkts': '9797872',
                                                                       'in_octets': '3596175167022',
                                                                       'in_ucast_pkts': '2740061179',
                                                                       'links': [       {       'in_nucast_pkts': '9797872',
                                                                                                'in_octets': '3596175167022',
                                                                                                'in_ucast_pkts': '2740061179',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '161085',
                                                                                                'out_octets': '266086606275',
                                                                                                'out_ucast_pkts': '965700022',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '161085',
                                                                       'out_octets': '266086606275',
                                                                       'out_ucast_pkts': '965700022',
                                                                       'poller_hash': 'ODQ0QkYycGVlcmludGZwb24gMTQwNjkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140693',
                                                                                               'links': [       {       'connected_time': '8535548',
                                                                                                                        'fiber_length': 23029,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODQ0QkYycGVlcnBvbnBvbiAxNDA2OTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '202 FERRIS DR (FOSTER KEDRIC)',
                                         'poller_hash': 'ODQ0QkYy',
                                         'serial_number': '844BF2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:06:e6:25',
                                         'uptime': '8535548'},
                                 {       'counter_source': '140732/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:2d:a4:96'],
                                                                       'id': '140732/g1',
                                                                       'in_nucast_pkts': '22',
                                                                       'in_octets': '136023182539',
                                                                       'in_ucast_pkts': '955313947',
                                                                       'mac_address': 'd0:76:8f:08:18:be',
                                                                       'mtu': 2000,
                                                                       'name': '140732/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919992',
                                                                       'out_octets': '3732609089484',
                                                                       'out_ucast_pkts': '2664567270',
                                                                       'poller_hash': 'ODY5NDBEcGVlcmludGYxNDA3MzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140732',
                                                                       'in_nucast_pkts': '9919992',
                                                                       'in_octets': '3732609089484',
                                                                       'in_ucast_pkts': '2664567270',
                                                                       'links': [       {       'in_nucast_pkts': '9919992',
                                                                                                'in_octets': '3732609089484',
                                                                                                'in_ucast_pkts': '2664567270',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '22',
                                                                                                'out_octets': '136023182539',
                                                                                                'out_ucast_pkts': '955313947',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22',
                                                                       'out_octets': '136023182539',
                                                                       'out_ucast_pkts': '955313947',
                                                                       'poller_hash': 'ODY5NDBEcGVlcmludGZwb24gMTQwNzMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140732',
                                                                                               'links': [       {       'connected_time': '8535296',
                                                                                                                        'fiber_length': 25071,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDBEcGVlcnBvbnBvbiAxNDA3MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '32 VILLAGE RD (BIGNEY KRISTOFER)',
                                         'poller_hash': 'ODY5NDBE',
                                         'serial_number': '86940D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:18:bb',
                                         'uptime': '8535296'},
                                 {       'counter_source': '140746/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '140746/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '900888007708',
                                                                       'in_ucast_pkts': '3458094422',
                                                                       'mac_address': 'd0:76:8f:07:18:0e',
                                                                       'name': '140746/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '88',
                                                                       'out_octets': '11469736959806',
                                                                       'out_ucast_pkts': '8428211828',
                                                                       'poller_hash': 'ODY3ODhCcGVlcmludGYxNDA3NDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '140746',
                                                                       'in_nucast_pkts': '88',
                                                                       'in_octets': '11469736959806',
                                                                       'in_ucast_pkts': '8428211828',
                                                                       'links': [       {       'in_nucast_pkts': '88',
                                                                                                'in_octets': '11469736959806',
                                                                                                'in_ucast_pkts': '8428211828',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '900888007708',
                                                                                                'out_ucast_pkts': '3458094422',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '900888007708',
                                                                       'out_ucast_pkts': '3458094422',
                                                                       'poller_hash': 'ODY3ODhCcGVlcmludGZwb24gMTQwNzQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140746',
                                                                                               'links': [       {       'connected_time': '8535393',
                                                                                                                        'fiber_length': 23063,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY3ODhCcGVlcnBvbnBvbiAxNDA3NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '106 LAWRENCE ST (MOL RHONDDA)',
                                         'poller_hash': 'ODY3ODhC',
                                         'serial_number': '86788B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:07:18:0e',
                                         'uptime': '8535393'},
                                 {       'counter_source': '140771/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:ac'],
                                                                       'id': '140771/g1',
                                                                       'in_nucast_pkts': '161566',
                                                                       'in_octets': '176726041766',
                                                                       'in_ucast_pkts': '529079260',
                                                                       'mac_address': '60:db:98:4e:71:b7',
                                                                       'mtu': 2000,
                                                                       'name': '140771/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797834',
                                                                       'out_octets': '1747579987269',
                                                                       'out_ucast_pkts': '1337388843',
                                                                       'poller_hash': 'QTlDM0Y2cGVlcmludGYxNDA3NzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140771',
                                                                       'in_nucast_pkts': '9797834',
                                                                       'in_octets': '1747579987269',
                                                                       'in_ucast_pkts': '1337388843',
                                                                       'links': [       {       'in_nucast_pkts': '9797834',
                                                                                                'in_octets': '1747579987269',
                                                                                                'in_ucast_pkts': '1337388843',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '161566',
                                                                                                'out_octets': '176726041766',
                                                                                                'out_ucast_pkts': '529079260',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '161566',
                                                                       'out_octets': '176726041766',
                                                                       'out_ucast_pkts': '529079260',
                                                                       'poller_hash': 'QTlDM0Y2cGVlcmludGZwb24gMTQwNzcx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140771',
                                                                                               'links': [       {       'connected_time': '8535293',
                                                                                                                        'fiber_length': 22987,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0Y2cGVlcnBvbnBvbiAxNDA3NzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '82 LAWRENCE ST (MACKINNON MATTHEW)',
                                         'poller_hash': 'QTlDM0Y2',
                                         'serial_number': 'A9C3F6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:71:b4',
                                         'uptime': '8535293'},
                                 {       'counter_source': '140919/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:62:0f'],
                                                                       'id': '140919/g1',
                                                                       'in_nucast_pkts': '612336',
                                                                       'in_octets': '1304962235281',
                                                                       'in_ucast_pkts': '3417673650',
                                                                       'mac_address': 'd0:76:8f:08:17:66',
                                                                       'mtu': 2000,
                                                                       'name': '140919/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35732147',
                                                                       'out_octets': '10985307440924',
                                                                       'out_ucast_pkts': '8453397977',
                                                                       'poller_hash': 'ODY5M0UycGVlcmludGYxNDA5MTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '140919',
                                                                       'in_nucast_pkts': '35732147',
                                                                       'in_octets': '10985307440924',
                                                                       'in_ucast_pkts': '8453397977',
                                                                       'links': [       {       'in_nucast_pkts': '35732147',
                                                                                                'in_octets': '10985307440924',
                                                                                                'in_ucast_pkts': '8453397977',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '612336',
                                                                                                'out_octets': '1304962235281',
                                                                                                'out_ucast_pkts': '3417673650',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '612336',
                                                                       'out_octets': '1304962235281',
                                                                       'out_ucast_pkts': '3417673650',
                                                                       'poller_hash': 'ODY5M0UycGVlcmludGZwb24gMTQwOTE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '140919',
                                                                                               'links': [       {       'connected_time': '65416542',
                                                                                                                        'fiber_length': 17234,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0UycGVlcnBvbnBvbiAxNDA5MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2090 NITHVIEW CT (FREUND DENNIS)',
                                         'poller_hash': 'ODY5M0Uy',
                                         'serial_number': '8693E2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:63',
                                         'uptime': '65416542'},
                                 {       'counter_source': '141086/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141086/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '804542686016',
                                                                       'in_ucast_pkts': '182641855850',
                                                                       'mac_address': '60:db:98:4e:6e:ec',
                                                                       'name': '141086/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '6309080717687',
                                                                       'out_ucast_pkts': '4748235885',
                                                                       'poller_hash': 'QTlDMzlEcGVlcmludGYxNDEwODYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141086',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '6309080717687',
                                                                       'in_ucast_pkts': '4748235885',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '6309080717687',
                                                                                                'in_ucast_pkts': '4748235885',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '804542686016',
                                                                                                'out_ucast_pkts': '182641855850',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '804542686016',
                                                                       'out_ucast_pkts': '182641855850',
                                                                       'poller_hash': 'QTlDMzlEcGVlcmludGZwb24gMTQxMDg2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141086',
                                                                                               'links': [       {       'connected_time': '3195474',
                                                                                                                        'fiber_length': 24063,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMzlEcGVlcnBvbnBvbiAxNDEwODY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1317 GREENWOOD HILL RD (JANTZI '
                                                 'RANDELL 10436707)',
                                         'poller_hash': 'QTlDMzlE',
                                         'serial_number': 'A9C39D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6e:ec',
                                         'uptime': '3195474'},
                                 {       'counter_source': '141245/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141245/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '584477821774',
                                                                       'in_ucast_pkts': '2873697244',
                                                                       'mac_address': 'd0:76:8f:08:15:c3',
                                                                       'name': '141245/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '13108082384200',
                                                                       'out_ucast_pkts': '9343811698',
                                                                       'poller_hash': 'ODY5M0FFcGVlcmludGYxNDEyNDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141245',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '13108082384200',
                                                                       'in_ucast_pkts': '9343811698',
                                                                       'links': [       {       'in_nucast_pkts': '117',
                                                                                                'in_octets': '13108082384200',
                                                                                                'in_ucast_pkts': '9343811698',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '584477821774',
                                                                                                'out_ucast_pkts': '2873697244',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '584477821774',
                                                                       'out_ucast_pkts': '2873697244',
                                                                       'poller_hash': 'ODY5M0FFcGVlcmludGZwb24gMTQxMjQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141245',
                                                                                               'links': [       {       'connected_time': '335343',
                                                                                                                        'fiber_length': 18005,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0FFcGVlcnBvbnBvbiAxNDEyNDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2069 WENDELL CT (EAST SEAN)',
                                         'poller_hash': 'ODY5M0FF',
                                         'serial_number': '8693AE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:c3',
                                         'uptime': '335343'},
                                 {       'counter_source': '141689/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141689/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '113705077156',
                                                                       'in_ucast_pkts': '427944072',
                                                                       'mac_address': 'd0:76:8f:aa:48:ec',
                                                                       'name': '141689/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '38',
                                                                       'out_octets': '3084307047575',
                                                                       'out_ucast_pkts': '2275739139',
                                                                       'poller_hash': 'OTU2RTQzcGVlcmludGYxNDE2ODkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141689',
                                                                       'in_nucast_pkts': '38',
                                                                       'in_octets': '3084307047575',
                                                                       'in_ucast_pkts': '2275739139',
                                                                       'links': [       {       'in_nucast_pkts': '38',
                                                                                                'in_octets': '3084307047575',
                                                                                                'in_ucast_pkts': '2275739139',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '113705077156',
                                                                                                'out_ucast_pkts': '427944072',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '113705077156',
                                                                       'out_ucast_pkts': '427944072',
                                                                       'poller_hash': 'OTU2RTQzcGVlcmludGZwb24gMTQxNjg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141689',
                                                                                               'links': [       {       'connected_time': '8535227',
                                                                                                                        'fiber_length': 23935,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTQzcGVlcnBvbnBvbiAxNDE2ODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3705 NAFZIGER RD (HORST CARL)',
                                         'poller_hash': 'OTU2RTQz',
                                         'serial_number': '956E43',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:48:ec',
                                         'uptime': '8535227'},
                                 {       'counter_source': '141693/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:6d:7c'],
                                                                       'id': '141693/g1',
                                                                       'in_nucast_pkts': '419785',
                                                                       'in_octets': '191567795710',
                                                                       'in_ucast_pkts': '1320023592',
                                                                       'mac_address': 'd0:76:8f:08:17:9e',
                                                                       'mtu': 2000,
                                                                       'name': '141693/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26250656',
                                                                       'out_octets': '5503207594828',
                                                                       'out_ucast_pkts': '3858877203',
                                                                       'poller_hash': 'ODY5M0U5cGVlcmludGYxNDE2OTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141693',
                                                                       'in_nucast_pkts': '26250656',
                                                                       'in_octets': '5503207594828',
                                                                       'in_ucast_pkts': '3858877203',
                                                                       'links': [       {       'in_nucast_pkts': '26250656',
                                                                                                'in_octets': '5503207594828',
                                                                                                'in_ucast_pkts': '3858877203',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '419785',
                                                                                                'out_octets': '191567795710',
                                                                                                'out_ucast_pkts': '1320023592',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '419785',
                                                                       'out_octets': '191567795710',
                                                                       'out_ucast_pkts': '1320023592',
                                                                       'poller_hash': 'ODY5M0U5cGVlcmludGZwb24gMTQxNjkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141693',
                                                                                               'links': [       {       'connected_time': '22473887',
                                                                                                                        'fiber_length': 23169,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5M0U5cGVlcnBvbnBvbiAxNDE2OTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 POND VIEW DR (HORST MAHLON)',
                                         'poller_hash': 'ODY5M0U5',
                                         'serial_number': '8693E9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:17:9b',
                                         'uptime': '22473887'},
                                 {       'counter_source': '141703/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '141703/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '372450311847',
                                                                       'in_ucast_pkts': '996340988',
                                                                       'mac_address': 'd0:76:8f:08:1c:33',
                                                                       'name': '141703/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '38',
                                                                       'out_octets': '4197775308392',
                                                                       'out_ucast_pkts': '3415928008',
                                                                       'poller_hash': 'ODY5NDdDcGVlcmludGYxNDE3MDMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '141703',
                                                                       'in_nucast_pkts': '38',
                                                                       'in_octets': '4197775308392',
                                                                       'in_ucast_pkts': '3415928008',
                                                                       'links': [       {       'in_nucast_pkts': '38',
                                                                                                'in_octets': '4197775308392',
                                                                                                'in_ucast_pkts': '3415928008',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '372450311847',
                                                                                                'out_ucast_pkts': '996340988',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '372450311847',
                                                                       'out_ucast_pkts': '996340988',
                                                                       'poller_hash': 'ODY5NDdDcGVlcmludGZwb24gMTQxNzAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141703',
                                                                                               'links': [       {       'connected_time': '8535291',
                                                                                                                        'fiber_length': 23735,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDdDcGVlcnBvbnBvbiAxNDE3MDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '67 CRAB APPLE CT (NYMAN ROBERT)',
                                         'poller_hash': 'ODY5NDdD',
                                         'serial_number': '86947C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1c:33',
                                         'uptime': '8535291'},
                                 {       'counter_source': '141753/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:d9:f5:74:47:10'],
                                                                       'id': '141753/g1',
                                                                       'in_nucast_pkts': '18',
                                                                       'in_octets': '65218191889',
                                                                       'in_ucast_pkts': '214564732',
                                                                       'mac_address': 'd0:76:8f:08:1c:86',
                                                                       'mtu': 2000,
                                                                       'name': '141753/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3754816',
                                                                       'out_octets': '1027542914147',
                                                                       'out_ucast_pkts': '813081171',
                                                                       'poller_hash': 'ODY5NDg2cGVlcmludGYxNDE3NTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '141753',
                                                                       'in_nucast_pkts': '3754816',
                                                                       'in_octets': '1027542914147',
                                                                       'in_ucast_pkts': '813081171',
                                                                       'links': [       {       'in_nucast_pkts': '3754816',
                                                                                                'in_octets': '1027542914147',
                                                                                                'in_ucast_pkts': '813081171',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18',
                                                                                                'out_octets': '65218191889',
                                                                                                'out_ucast_pkts': '214564732',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18',
                                                                       'out_octets': '65218191889',
                                                                       'out_ucast_pkts': '214564732',
                                                                       'poller_hash': 'ODY5NDg2cGVlcmludGZwb24gMTQxNzUz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '141753',
                                                                                               'links': [       {       'connected_time': '3179647',
                                                                                                                        'fiber_length': 23801,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5NDg2cGVlcnBvbnBvbiAxNDE3NTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '38 GALENA ST (GRANT JACOB)',
                                         'poller_hash': 'ODY5NDg2',
                                         'serial_number': '869486',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:1c:83',
                                         'uptime': '3179647'},
                                 {       'counter_source': '142973/F1+142973/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:db:4d:e0',
                                                                                                        '00:03:e6:db:50:94',
                                                                                                        '00:03:e6:ec:14:10'],
                                                                       'id': '142973/F1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9982269561',
                                                                       'in_ucast_pkts': '50441728',
                                                                       'name': '142973/F1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '64956701765',
                                                                       'out_octets': '87969641663067',
                                                                       'out_ucast_pkts': '15718346',
                                                                       'poller_hash': 'ODY5Mzk2cGVlcmludGYxNDI5NzMvRjE='},
                                                               {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '68:d7:9a:66:cb:50'],
                                                                       'id': '142973/g1',
                                                                       'in_nucast_pkts': '7341919',
                                                                       'in_octets': '7300210871392',
                                                                       'in_ucast_pkts': '10174419453',
                                                                       'mac_address': 'd0:76:8f:08:15:06',
                                                                       'mtu': 2000,
                                                                       'name': '142973/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63499170',
                                                                       'out_octets': '38015662917710',
                                                                       'out_ucast_pkts': '29223108531',
                                                                       'poller_hash': 'ODY5Mzk2cGVlcmludGYxNDI5NzMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '142973',
                                                                       'in_nucast_pkts': '65020200935',
                                                                       'in_octets': '125985304580777',
                                                                       'in_ucast_pkts': '29238826877',
                                                                       'links': [       {       'in_nucast_pkts': '65020200935',
                                                                                                'in_octets': '125985304580777',
                                                                                                'in_ucast_pkts': '29238826877',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7341919',
                                                                                                'out_octets': '7310193140953',
                                                                                                'out_ucast_pkts': '10224861181',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7341919',
                                                                       'out_octets': '7310193140953',
                                                                       'out_ucast_pkts': '10224861181',
                                                                       'poller_hash': 'ODY5Mzk2cGVlcmludGZwb24gMTQyOTcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '142973',
                                                                                               'links': [       {       'connected_time': '21887083',
                                                                                                                        'fiber_length': 37683,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5Mzk2cGVlcnBvbnBvbiAxNDI5NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1690 CARMEL-KOCH RD (STRAUS DAN)',
                                         'poller_hash': 'ODY5Mzk2',
                                         'serial_number': '869396',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:08:15:03',
                                         'uptime': '21887083'},
                                 {       'counter_source': '143157/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:52:4a'],
                                                                       'id': '143157/g1',
                                                                       'in_nucast_pkts': '164479',
                                                                       'in_octets': '5796600638205',
                                                                       'in_ucast_pkts': '4890642412',
                                                                       'mac_address': 'd0:76:8f:1b:9f:ae',
                                                                       'mtu': 2000,
                                                                       'name': '143157/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797590',
                                                                       'out_octets': '6797544661027',
                                                                       'out_ucast_pkts': '6466349730',
                                                                       'poller_hash': 'ODcyNDQzcGVlcmludGYxNDMxNTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143157',
                                                                       'in_nucast_pkts': '9797590',
                                                                       'in_octets': '6797544661027',
                                                                       'in_ucast_pkts': '6466349730',
                                                                       'links': [       {       'in_nucast_pkts': '9797590',
                                                                                                'in_octets': '6797544661027',
                                                                                                'in_ucast_pkts': '6466349730',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '164479',
                                                                                                'out_octets': '5796600638205',
                                                                                                'out_ucast_pkts': '4890642412',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '164479',
                                                                       'out_octets': '5796600638205',
                                                                       'out_ucast_pkts': '4890642412',
                                                                       'poller_hash': 'ODcyNDQzcGVlcmludGZwb24gMTQzMTU3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143157',
                                                                                               'links': [       {       'connected_time': '8535294',
                                                                                                                        'fiber_length': 25588,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDQzcGVlcnBvbnBvbiAxNDMxNTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '143 VILLAGE RD (EVANS GARY)',
                                         'poller_hash': 'ODcyNDQz',
                                         'serial_number': '872443',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:ab',
                                         'uptime': '8535294'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '143254/g1',
                                                                       'name': '143254/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODcyNDNCcGVlcmludGYxNDMyNTQvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143254',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODcyNDNCcGVlcmludGZwb24gMTQzMjU0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143254',
                                                                                               'links': [       {       'connected_time': '8535120',
                                                                                                                        'fiber_length': 23642,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDNCcGVlcnBvbnBvbiAxNDMyNTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '47 CRAB APPLE CT (BILLINGS KANDICE)',
                                         'poller_hash': 'ODcyNDNC',
                                         'serial_number': '87243B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:6b',
                                         'uptime': '8535120'},
                                 {       'counter_source': '143264/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:a5:63:cb'],
                                                                       'id': '143264/g1',
                                                                       'in_nucast_pkts': '6478',
                                                                       'in_octets': '3061204072',
                                                                       'in_ucast_pkts': '5901548',
                                                                       'mac_address': 'd0:76:8f:1b:9f:9e',
                                                                       'mtu': 2000,
                                                                       'name': '143264/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394271',
                                                                       'out_octets': '29166803831',
                                                                       'out_ucast_pkts': '22652769',
                                                                       'poller_hash': 'ODcyNDQxcGVlcmludGYxNDMyNjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143264',
                                                                       'in_nucast_pkts': '394271',
                                                                       'in_octets': '29166803831',
                                                                       'in_ucast_pkts': '22652769',
                                                                       'links': [       {       'in_nucast_pkts': '394271',
                                                                                                'in_octets': '29166803831',
                                                                                                'in_ucast_pkts': '22652769',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6478',
                                                                                                'out_octets': '3061204072',
                                                                                                'out_ucast_pkts': '5901548',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6478',
                                                                       'out_octets': '3061204072',
                                                                       'out_ucast_pkts': '5901548',
                                                                       'poller_hash': 'ODcyNDQxcGVlcmludGZwb24gMTQzMjY0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143264',
                                                                                               'links': [       {       'connected_time': '335347',
                                                                                                                        'fiber_length': 17172,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDQxcGVlcnBvbnBvbiAxNDMyNjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2082 NITHVIEW CT (HACKBART DAN)',
                                         'poller_hash': 'ODcyNDQx',
                                         'serial_number': '872441',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:9b',
                                         'uptime': '335347'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '143632/g1',
                                                                       'name': '143632/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODcyNTRDcGVlcmludGYxNDM2MzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143632',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODcyNTRDcGVlcmludGZwb24gMTQzNjMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143632',
                                                                                               'links': [       {       'connected_time': '8535161',
                                                                                                                        'fiber_length': 25137,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTRDcGVlcnBvbnBvbiAxNDM2MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '79 GERBER MEADOWS DR',
                                         'poller_hash': 'ODcyNTRD',
                                         'serial_number': '87254C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a7:f3',
                                         'uptime': '8535161'},
                                 {       'counter_source': '143732/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:c2:41:82:a6:01'],
                                                                       'id': '143732/g1',
                                                                       'in_nucast_pkts': '29621',
                                                                       'in_octets': '44442529034',
                                                                       'in_ucast_pkts': '153916249',
                                                                       'mac_address': 'd0:76:8f:1b:a5:06',
                                                                       'mtu': 2000,
                                                                       'name': '143732/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9918791',
                                                                       'out_octets': '100857311262',
                                                                       'out_ucast_pkts': '130528996',
                                                                       'poller_hash': 'ODcyNEVFcGVlcmludGYxNDM3MzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143732',
                                                                       'in_nucast_pkts': '9918791',
                                                                       'in_octets': '100857311262',
                                                                       'in_ucast_pkts': '130528996',
                                                                       'links': [       {       'in_nucast_pkts': '9918791',
                                                                                                'in_octets': '100857311262',
                                                                                                'in_ucast_pkts': '130528996',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '29621',
                                                                                                'out_octets': '44442529034',
                                                                                                'out_ucast_pkts': '153916249',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29621',
                                                                       'out_octets': '44442529034',
                                                                       'out_ucast_pkts': '153916249',
                                                                       'poller_hash': 'ODcyNEVFcGVlcmludGZwb24gMTQzNzMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143732',
                                                                                               'links': [       {       'connected_time': '8570820',
                                                                                                                        'fiber_length': 22310,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNEVFcGVlcnBvbnBvbiAxNDM3MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1245 QUEEN'S BUSH RD",
                                         'poller_hash': 'ODcyNEVF',
                                         'serial_number': '8724EE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:03',
                                         'uptime': '8570820'},
                                 {       'counter_source': '143742/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '2c:3f:0b:ac:78:a5'],
                                                                       'id': '143742/g1',
                                                                       'in_nucast_pkts': '2309732',
                                                                       'in_octets': '2100555949462',
                                                                       'in_ucast_pkts': '6012360135',
                                                                       'mac_address': 'd0:76:8f:1b:a5:86',
                                                                       'mtu': 2000,
                                                                       'name': '143742/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38392824',
                                                                       'out_octets': '10086052639534',
                                                                       'out_ucast_pkts': '9730720550',
                                                                       'poller_hash': 'ODcyNEZFcGVlcmludGYxNDM3NDIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143742',
                                                                       'in_nucast_pkts': '38392824',
                                                                       'in_octets': '10086052639534',
                                                                       'in_ucast_pkts': '9730720550',
                                                                       'links': [       {       'in_nucast_pkts': '38392824',
                                                                                                'in_octets': '10086052639534',
                                                                                                'in_ucast_pkts': '9730720550',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2309732',
                                                                                                'out_octets': '2100555949462',
                                                                                                'out_ucast_pkts': '6012360135',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2309732',
                                                                       'out_octets': '2100555949462',
                                                                       'out_ucast_pkts': '6012360135',
                                                                       'poller_hash': 'ODcyNEZFcGVlcmludGZwb24gMTQzNzQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143742',
                                                                                               'links': [       {       'connected_time': '66603086',
                                                                                                                        'fiber_length': 23530,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNEZFcGVlcnBvbnBvbiAxNDM3NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2196 GERBER RD (WELLESLEY HOME '
                                                 'CENTRE)',
                                         'poller_hash': 'ODcyNEZF',
                                         'serial_number': '8724FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:83',
                                         'uptime': '66603086'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '143782/g1',
                                                                       'name': '143782/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODcyNTA0cGVlcmludGYxNDM3ODIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143782',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'ODcyNTA0cGVlcmludGZwb24gMTQzNzgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143782',
                                                                                               'links': [       {       'connected_time': '8535176',
                                                                                                                        'fiber_length': 23634,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNTA0cGVlcnBvbnBvbiAxNDM3ODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '43 CRAB APPLE CT',
                                         'poller_hash': 'ODcyNTA0',
                                         'serial_number': '872504',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:b3',
                                         'uptime': '8535176'},
                                 {       'counter_source': '143832/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143832/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '871461322933',
                                                                       'in_ucast_pkts': '4003059264',
                                                                       'mac_address': 'd0:76:8f:1b:9e:73',
                                                                       'name': '143832/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '13766508254334',
                                                                       'out_ucast_pkts': '10249686258',
                                                                       'poller_hash': 'ODcyNDFDcGVlcmludGYxNDM4MzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143832',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '13766508254334',
                                                                       'in_ucast_pkts': '10249686258',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '13766508254334',
                                                                                                'in_ucast_pkts': '10249686258',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '871461322933',
                                                                                                'out_ucast_pkts': '4003059264',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '871461322933',
                                                                       'out_ucast_pkts': '4003059264',
                                                                       'poller_hash': 'ODcyNDFDcGVlcmludGZwb24gMTQzODMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143832',
                                                                                               'links': [       {       'connected_time': '8535196',
                                                                                                                        'fiber_length': 23094,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDFDcGVlcnBvbnBvbiAxNDM4MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1116 MOLESWORTH ST (PIASECKI LUKE)',
                                         'poller_hash': 'ODcyNDFD',
                                         'serial_number': '87241C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9e:73',
                                         'uptime': '8535196'},
                                 {       'counter_source': '143842/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '143842/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '610259750851',
                                                                       'in_ucast_pkts': '48330922313',
                                                                       'mac_address': 'd0:76:8f:1b:9e:53',
                                                                       'name': '143842/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '69',
                                                                       'out_octets': '2742689711556',
                                                                       'out_ucast_pkts': '2327112673',
                                                                       'poller_hash': 'ODcyNDE4cGVlcmludGYxNDM4NDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143842',
                                                                       'in_nucast_pkts': '69',
                                                                       'in_octets': '2742689711556',
                                                                       'in_ucast_pkts': '2327112673',
                                                                       'links': [       {       'in_nucast_pkts': '69',
                                                                                                'in_octets': '2742689711556',
                                                                                                'in_ucast_pkts': '2327112673',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '610259750851',
                                                                                                'out_ucast_pkts': '48330922313',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '610259750851',
                                                                       'out_ucast_pkts': '48330922313',
                                                                       'poller_hash': 'ODcyNDE4cGVlcmludGZwb24gMTQzODQy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143842',
                                                                                               'links': [       {       'connected_time': '8535212',
                                                                                                                        'fiber_length': 23212,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDE4cGVlcnBvbnBvbiAxNDM4NDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '151 FERRIS DR (WETTLAUFER MATTHEW)',
                                         'poller_hash': 'ODcyNDE4',
                                         'serial_number': '872418',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9e:53',
                                         'uptime': '8535212'},
                                 {       'counter_source': '143852/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '38:94:ed:bf:72:bf'],
                                                                       'id': '143852/g1',
                                                                       'in_nucast_pkts': '202770',
                                                                       'in_octets': '239625932782',
                                                                       'in_ucast_pkts': '1392898948',
                                                                       'mac_address': 'ec:4f:82:7e:09:6a',
                                                                       'mtu': 2000,
                                                                       'name': '143852/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9798083',
                                                                       'out_octets': '5511720542588',
                                                                       'out_ucast_pkts': '4006265040',
                                                                       'poller_hash': 'MzVFOTZFcGVlcmludGYxNDM4NTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143852',
                                                                       'in_nucast_pkts': '9798083',
                                                                       'in_octets': '5511720542588',
                                                                       'in_ucast_pkts': '4006265040',
                                                                       'links': [       {       'in_nucast_pkts': '9798083',
                                                                                                'in_octets': '5511720542588',
                                                                                                'in_ucast_pkts': '4006265040',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '202770',
                                                                                                'out_octets': '239625932782',
                                                                                                'out_ucast_pkts': '1392898948',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '202770',
                                                                       'out_octets': '239625932782',
                                                                       'out_ucast_pkts': '1392898948',
                                                                       'poller_hash': 'MzVFOTZFcGVlcmludGZwb24gMTQzODUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143852',
                                                                                               'links': [       {       'connected_time': '8535470',
                                                                                                                        'fiber_length': 23546,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTZFcGVlcnBvbnBvbiAxNDM4NTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '64 FERRIS DR (FANE PAMELA)',
                                         'poller_hash': 'MzVFOTZF',
                                         'serial_number': '35E96E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:09:6a',
                                         'uptime': '8535470'},
                                 {       'counter_source': '143862/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3c:1d'],
                                                                       'id': '143862/g1',
                                                                       'in_nucast_pkts': '418186',
                                                                       'in_octets': '212334674186',
                                                                       'in_ucast_pkts': '2160493823',
                                                                       'mac_address': 'd0:76:8f:1b:a5:36',
                                                                       'mtu': 2000,
                                                                       'name': '143862/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27538111',
                                                                       'out_octets': '10598077154835',
                                                                       'out_ucast_pkts': '7136619341',
                                                                       'poller_hash': 'ODcyNEY0cGVlcmludGYxNDM4NjIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143862',
                                                                       'in_nucast_pkts': '27538111',
                                                                       'in_octets': '10598077154835',
                                                                       'in_ucast_pkts': '7136619341',
                                                                       'links': [       {       'in_nucast_pkts': '27538111',
                                                                                                'in_octets': '10598077154835',
                                                                                                'in_ucast_pkts': '7136619341',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '418186',
                                                                                                'out_octets': '212334674186',
                                                                                                'out_ucast_pkts': '2160493823',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '418186',
                                                                       'out_octets': '212334674186',
                                                                       'out_ucast_pkts': '2160493823',
                                                                       'poller_hash': 'ODcyNEY0cGVlcmludGZwb24gMTQzODYy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143862',
                                                                                               'links': [       {       'connected_time': '23752942',
                                                                                                                        'fiber_length': 22584,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNEY0cGVlcnBvbnBvbiAxNDM4NjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '53 DAVID ST (STEINMANN HAROLD)',
                                         'poller_hash': 'ODcyNEY0',
                                         'serial_number': '8724F4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a5:33',
                                         'uptime': '23752942'},
                                 {       'counter_source': '143872/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5b:3a'],
                                                                       'id': '143872/g1',
                                                                       'in_nucast_pkts': '154036',
                                                                       'in_octets': '206343381695',
                                                                       'in_ucast_pkts': '681961428',
                                                                       'mac_address': 'd0:76:8f:1b:9e:d6',
                                                                       'mtu': 2000,
                                                                       'name': '143872/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878387',
                                                                       'out_octets': '4099782532724',
                                                                       'out_ucast_pkts': '3000146836',
                                                                       'poller_hash': 'ODcyNDI4cGVlcmludGYxNDM4NzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143872',
                                                                       'in_nucast_pkts': '9878387',
                                                                       'in_octets': '4099782532724',
                                                                       'in_ucast_pkts': '3000146836',
                                                                       'links': [       {       'in_nucast_pkts': '9878387',
                                                                                                'in_octets': '4099782532724',
                                                                                                'in_ucast_pkts': '3000146836',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '154036',
                                                                                                'out_octets': '206343381695',
                                                                                                'out_ucast_pkts': '681961428',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '154036',
                                                                       'out_octets': '206343381695',
                                                                       'out_ucast_pkts': '681961428',
                                                                       'poller_hash': 'ODcyNDI4cGVlcmludGZwb24gMTQzODcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143872',
                                                                                               'links': [       {       'connected_time': '8535187',
                                                                                                                        'fiber_length': 23176,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDI4cGVlcnBvbnBvbiAxNDM4NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1095 HENRY ST (JANZEN KARL)',
                                         'poller_hash': 'ODcyNDI4',
                                         'serial_number': '872428',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9e:d3',
                                         'uptime': '8535187'},
                                 {       'counter_source': '143912/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:5a:a0'],
                                                                       'id': '143912/g1',
                                                                       'in_nucast_pkts': '20154',
                                                                       'in_octets': '12360324939',
                                                                       'in_ucast_pkts': '28253058',
                                                                       'mac_address': 'd0:76:8f:1b:9f:06',
                                                                       'mtu': 2000,
                                                                       'name': '143912/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1281607',
                                                                       'out_octets': '128642056250',
                                                                       'out_ucast_pkts': '99305209',
                                                                       'poller_hash': 'ODcyNDJFcGVlcmludGYxNDM5MTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '143912',
                                                                       'in_nucast_pkts': '1281607',
                                                                       'in_octets': '128642056250',
                                                                       'in_ucast_pkts': '99305209',
                                                                       'links': [       {       'in_nucast_pkts': '1281607',
                                                                                                'in_octets': '128642056250',
                                                                                                'in_ucast_pkts': '99305209',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20154',
                                                                                                'out_octets': '12360324939',
                                                                                                'out_ucast_pkts': '28253058',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20154',
                                                                       'out_octets': '12360324939',
                                                                       'out_ucast_pkts': '28253058',
                                                                       'poller_hash': 'ODcyNDJFcGVlcmludGZwb24gMTQzOTEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143912',
                                                                                               'links': [       {       'connected_time': '1096780',
                                                                                                                        'fiber_length': 31474,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDJFcGVlcnBvbnBvbiAxNDM5MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4054 WEIMAR LINE (DANIS-HARKNESS '
                                                 'SUZANNE)',
                                         'poller_hash': 'ODcyNDJF',
                                         'serial_number': '87242E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:9f:03',
                                         'uptime': '1096780'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '143972/g1',
                                                                       'name': '143972/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODcyNDIycGVlcmludGYxNDM5NzIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '143972',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODcyNDIycGVlcmludGZwb24gMTQzOTcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '143972',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDIycGVlcnBvbnBvbiAxNDM5NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '34 GERBER MEADOWS DR',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "143972" '
                                                                  '(location: GPON '
                                                                  'port "1/3"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:26:48.520Z',
                                         'poller_hash': 'ODcyNDIy',
                                         'serial_number': '872422',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '144248/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '48:77:46:f8:b1:c0'],
                                                                       'id': '144248/g1',
                                                                       'in_nucast_pkts': '161387',
                                                                       'in_octets': '115010061378',
                                                                       'in_ucast_pkts': '462884381',
                                                                       'mac_address': 'd0:76:8f:1b:a0:9e',
                                                                       'mtu': 2000,
                                                                       'name': '144248/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878204',
                                                                       'out_octets': '1618836687632',
                                                                       'out_ucast_pkts': '1340783364',
                                                                       'poller_hash': 'ODcyNDYxcGVlcmludGYxNDQyNDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '144248',
                                                                       'in_nucast_pkts': '9878204',
                                                                       'in_octets': '1618836687632',
                                                                       'in_ucast_pkts': '1340783364',
                                                                       'links': [       {       'in_nucast_pkts': '9878204',
                                                                                                'in_octets': '1618836687632',
                                                                                                'in_ucast_pkts': '1340783364',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '161387',
                                                                                                'out_octets': '115010061378',
                                                                                                'out_ucast_pkts': '462884381',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '161387',
                                                                       'out_octets': '115010061378',
                                                                       'out_ucast_pkts': '462884381',
                                                                       'poller_hash': 'ODcyNDYxcGVlcmludGZwb24gMTQ0MjQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144248',
                                                                                               'links': [       {       'connected_time': '8535169',
                                                                                                                        'fiber_length': 25094,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODcyNDYxcGVlcnBvbnBvbiAxNDQyNDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '63 GERBER MEADOWS DR (CLARK CATRINA)',
                                         'poller_hash': 'ODcyNDYx',
                                         'serial_number': '872461',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:1b:a0:9b',
                                         'uptime': '8535169'},
                                 {       'counter_source': '144497/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:11'],
                                                                       'id': '144497/g1',
                                                                       'in_nucast_pkts': '43398',
                                                                       'in_octets': '16109324846',
                                                                       'in_ucast_pkts': '55335851',
                                                                       'mac_address': 'd0:76:8f:aa:51:97',
                                                                       'mtu': 2000,
                                                                       'name': '144497/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2594131',
                                                                       'out_octets': '229447343250',
                                                                       'out_ucast_pkts': '187563071',
                                                                       'poller_hash': 'OTU2RjU4cGVlcmludGYxNDQ0OTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '144497',
                                                                       'in_nucast_pkts': '2594131',
                                                                       'in_octets': '229447343250',
                                                                       'in_ucast_pkts': '187563071',
                                                                       'links': [       {       'in_nucast_pkts': '2594131',
                                                                                                'in_octets': '229447343250',
                                                                                                'in_ucast_pkts': '187563071',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '43398',
                                                                                                'out_octets': '16109324846',
                                                                                                'out_ucast_pkts': '55335851',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43398',
                                                                       'out_octets': '16109324846',
                                                                       'out_ucast_pkts': '55335851',
                                                                       'poller_hash': 'OTU2RjU4cGVlcmludGZwb24gMTQ0NDk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144497',
                                                                                               'links': [       {       'connected_time': '2212629',
                                                                                                                        'fiber_length': 22138,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjU4cGVlcnBvbnBvbiAxNDQ0OTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 DAVID ST (BLACKBURN LISA)',
                                         'poller_hash': 'OTU2RjU4',
                                         'serial_number': '956F58',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:51:94',
                                         'uptime': '2212629'},
                                 {       'counter_source': '144736/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '144736/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '25797027250',
                                                                       'in_ucast_pkts': '94402764',
                                                                       'mac_address': 'd0:76:8f:aa:52:64',
                                                                       'name': '144736/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '96',
                                                                       'out_octets': '312961961380',
                                                                       'out_ucast_pkts': '275822104',
                                                                       'poller_hash': 'OTU2RjcycGVlcmludGYxNDQ3MzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144736',
                                                                       'in_nucast_pkts': '96',
                                                                       'in_octets': '312961961380',
                                                                       'in_ucast_pkts': '275822104',
                                                                       'links': [       {       'in_nucast_pkts': '96',
                                                                                                'in_octets': '312961961380',
                                                                                                'in_ucast_pkts': '275822104',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '25797027250',
                                                                                                'out_ucast_pkts': '94402764',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '25797027250',
                                                                       'out_ucast_pkts': '94402764',
                                                                       'poller_hash': 'OTU2RjcycGVlcmludGZwb24gMTQ0NzM2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144736',
                                                                                               'links': [       {       'connected_time': '21887300',
                                                                                                                        'fiber_length': 24835,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjcycGVlcnBvbnBvbiAxNDQ3MzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 VILLAGE RD (HUBER PAUL)',
                                         'poller_hash': 'OTU2Rjcy',
                                         'serial_number': '956F72',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:52:64',
                                         'uptime': '21887300'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '144756/g1',
                                                                       'name': '144756/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2RjYxcGVlcmludGYxNDQ3NTYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '144756',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'OTU2RjYxcGVlcmludGZwb24gMTQ0NzU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144756',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjYxcGVlcnBvbnBvbiAxNDQ3NTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '110 SCHWEITZER CR',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "144756" '
                                                                  '(location: GPON '
                                                                  'port "2/1"), "ONT '
                                                                  'went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2024-03-21T04:33:17.100Z',
                                         'poller_hash': 'OTU2RjYx',
                                         'serial_number': '956F61',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '144779/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d6:1c'],
                                                                       'id': '144779/g1',
                                                                       'in_nucast_pkts': '29714838',
                                                                       'in_octets': '10051634793724',
                                                                       'in_ucast_pkts': '14725841733',
                                                                       'mac_address': 'd0:76:8f:aa:51:a7',
                                                                       'mtu': 2000,
                                                                       'name': '144779/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '66254244623',
                                                                       'out_octets': '124634468626549',
                                                                       'out_ucast_pkts': '27297610348',
                                                                       'poller_hash': 'OTU2RjVBcGVlcmludGYxNDQ3NzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '144779',
                                                                       'in_nucast_pkts': '66254244623',
                                                                       'in_octets': '124634468626549',
                                                                       'in_ucast_pkts': '27297610348',
                                                                       'links': [       {       'in_nucast_pkts': '66254244623',
                                                                                                'in_octets': '124634468626549',
                                                                                                'in_ucast_pkts': '27297610348',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '29714838',
                                                                                                'out_octets': '10051634793724',
                                                                                                'out_ucast_pkts': '14725841733',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29714838',
                                                                       'out_octets': '10051634793724',
                                                                       'out_ucast_pkts': '14725841733',
                                                                       'poller_hash': 'OTU2RjVBcGVlcmludGZwb24gMTQ0Nzc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144779',
                                                                                               'links': [       {       'connected_time': '21886922',
                                                                                                                        'fiber_length': 23586,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjVBcGVlcnBvbnBvbiAxNDQ3Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '86 SCHWEITZER CR (LIVINGSTONE CHAD)',
                                         'poller_hash': 'OTU2RjVB',
                                         'serial_number': '956F5A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:51:a4',
                                         'uptime': '21886922'},
                                 {       'counter_source': '144799/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d4:7a'],
                                                                       'id': '144799/g1',
                                                                       'in_nucast_pkts': '161503',
                                                                       'in_octets': '137323308469',
                                                                       'in_ucast_pkts': '553717456',
                                                                       'mac_address': 'd0:76:8f:aa:53:8f',
                                                                       'mtu': 2000,
                                                                       'name': '144799/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878489',
                                                                       'out_octets': '2289950757169',
                                                                       'out_ucast_pkts': '1848885558',
                                                                       'poller_hash': 'OTU2Rjk3cGVlcmludGYxNDQ3OTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '144799',
                                                                       'in_nucast_pkts': '9878489',
                                                                       'in_octets': '2289950757169',
                                                                       'in_ucast_pkts': '1848885558',
                                                                       'links': [       {       'in_nucast_pkts': '9878489',
                                                                                                'in_octets': '2289950757169',
                                                                                                'in_ucast_pkts': '1848885558',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '161503',
                                                                                                'out_octets': '137323308469',
                                                                                                'out_ucast_pkts': '553717456',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '161503',
                                                                       'out_octets': '137323308469',
                                                                       'out_ucast_pkts': '553717456',
                                                                       'poller_hash': 'OTU2Rjk3cGVlcmludGZwb24gMTQ0Nzk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144799',
                                                                                               'links': [       {       'connected_time': '8535211',
                                                                                                                        'fiber_length': 23217,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2Rjk3cGVlcnBvbnBvbiAxNDQ3OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '158 FERRIS DR (TOMPKINS TANYA)',
                                         'poller_hash': 'OTU2Rjk3',
                                         'serial_number': '956F97',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:53:8c',
                                         'uptime': '8535211'},
                                 {       'counter_source': '144978/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d4:85'],
                                                                       'id': '144978/g1',
                                                                       'in_nucast_pkts': '13160',
                                                                       'in_octets': '3912432542',
                                                                       'in_ucast_pkts': '12249420',
                                                                       'mac_address': 'd0:76:8f:aa:52:e7',
                                                                       'mtu': 2000,
                                                                       'name': '144978/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715479',
                                                                       'out_octets': '44982870288',
                                                                       'out_ucast_pkts': '34313512',
                                                                       'poller_hash': 'OTU2RjgycGVlcmludGYxNDQ5NzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '144978',
                                                                       'in_nucast_pkts': '715479',
                                                                       'in_octets': '44982870288',
                                                                       'in_ucast_pkts': '34313512',
                                                                       'links': [       {       'in_nucast_pkts': '715479',
                                                                                                'in_octets': '44982870288',
                                                                                                'in_ucast_pkts': '34313512',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13160',
                                                                                                'out_octets': '3912432542',
                                                                                                'out_ucast_pkts': '12249420',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13160',
                                                                       'out_octets': '3912432542',
                                                                       'out_ucast_pkts': '12249420',
                                                                       'poller_hash': 'OTU2RjgycGVlcmludGZwb24gMTQ0OTc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '144978',
                                                                                               'links': [       {       'connected_time': '611454',
                                                                                                                        'fiber_length': 20037,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjgycGVlcnBvbnBvbiAxNDQ5Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2839 LINE 45 (MCDOWELL CAROLINE)',
                                         'poller_hash': 'OTU2Rjgy',
                                         'serial_number': '956F82',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:52:e4',
                                         'uptime': '611454'},
                                 {       'counter_source': '145078/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145078/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45631819663',
                                                                       'in_ucast_pkts': '282204782',
                                                                       'mac_address': 'd0:76:8f:aa:52:d4',
                                                                       'name': '145078/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '1338327039744',
                                                                       'out_ucast_pkts': '1007143456',
                                                                       'poller_hash': 'OTU2RjgwcGVlcmludGYxNDUwNzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145078',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '1338327039744',
                                                                       'in_ucast_pkts': '1007143456',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '1338327039744',
                                                                                                'in_ucast_pkts': '1007143456',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45631819663',
                                                                                                'out_ucast_pkts': '282204782',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45631819663',
                                                                       'out_ucast_pkts': '282204782',
                                                                       'poller_hash': 'OTU2RjgwcGVlcmludGZwb24gMTQ1MDc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145078',
                                                                                               'links': [       {       'connected_time': '4071451',
                                                                                                                        'fiber_length': 23124,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjgwcGVlcnBvbnBvbiAxNDUwNzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '121 DAVID ST (COBER BRUCE)',
                                         'poller_hash': 'OTU2Rjgw',
                                         'serial_number': '956F80',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:52:d4',
                                         'uptime': '4071451'},
                                 {       'counter_source': '145118/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:ca:da'],
                                                                       'id': '145118/g1',
                                                                       'in_nucast_pkts': '7947516',
                                                                       'in_octets': '347613346840',
                                                                       'in_ucast_pkts': '1194881380',
                                                                       'mac_address': 'd0:76:8f:aa:50:4f',
                                                                       'mtu': 2000,
                                                                       'name': '145118/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10903553127',
                                                                       'out_octets': '18621243132098',
                                                                       'out_ucast_pkts': '3106591191',
                                                                       'poller_hash': 'OTU2RjJGcGVlcmludGYxNDUxMTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145118',
                                                                       'in_nucast_pkts': '10903553127',
                                                                       'in_octets': '18621243132098',
                                                                       'in_ucast_pkts': '3106591191',
                                                                       'links': [       {       'in_nucast_pkts': '10903553127',
                                                                                                'in_octets': '18621243132098',
                                                                                                'in_ucast_pkts': '3106591191',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7947516',
                                                                                                'out_octets': '347613346840',
                                                                                                'out_ucast_pkts': '1194881380',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7947516',
                                                                       'out_octets': '347613346840',
                                                                       'out_ucast_pkts': '1194881380',
                                                                       'poller_hash': 'OTU2RjJGcGVlcmludGZwb24gMTQ1MTE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145118',
                                                                                               'links': [       {       'connected_time': '8535251',
                                                                                                                        'fiber_length': 23613,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjJGcGVlcnBvbnBvbiAxNDUxMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3890 NAFZIGER RD (GOOD JEFF)',
                                         'poller_hash': 'OTU2RjJG',
                                         'serial_number': '956F2F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:50:4c',
                                         'uptime': '8535251'},
                                 {       'counter_source': '145128/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d3:88'],
                                                                       'id': '145128/g1',
                                                                       'in_nucast_pkts': '909030',
                                                                       'in_octets': '545790578420',
                                                                       'in_ucast_pkts': '3089161428',
                                                                       'mac_address': 'd0:76:8f:aa:53:37',
                                                                       'mtu': 2000,
                                                                       'name': '145128/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '57329872',
                                                                       'out_octets': '14713296077088',
                                                                       'out_ucast_pkts': '10686255915',
                                                                       'poller_hash': 'OTU2RjhDcGVlcmludGYxNDUxMjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145128',
                                                                       'in_nucast_pkts': '57329872',
                                                                       'in_octets': '14713296077088',
                                                                       'in_ucast_pkts': '10686255915',
                                                                       'links': [       {       'in_nucast_pkts': '57329872',
                                                                                                'in_octets': '14713296077088',
                                                                                                'in_ucast_pkts': '10686255915',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '909030',
                                                                                                'out_octets': '545790578420',
                                                                                                'out_ucast_pkts': '3089161428',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '909030',
                                                                       'out_octets': '545790578420',
                                                                       'out_ucast_pkts': '3089161428',
                                                                       'poller_hash': 'OTU2RjhDcGVlcmludGZwb24gMTQ1MTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145128',
                                                                                               'links': [       {       'connected_time': '12045338',
                                                                                                                        'fiber_length': 23283,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjhDcGVlcnBvbnBvbiAxNDUxMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1038 MOLESWORTH ST (SHAW TRICIA)',
                                         'poller_hash': 'OTU2RjhD',
                                         'serial_number': '956F8C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:53:34',
                                         'uptime': '12045338'},
                                 {       'counter_source': '145138/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145138/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '983604945228',
                                                                       'in_ucast_pkts': '2764223582',
                                                                       'mac_address': 'd0:76:8f:aa:48:64',
                                                                       'name': '145138/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '11290719260810',
                                                                       'out_ucast_pkts': '8689583679',
                                                                       'poller_hash': 'OTU2RTMycGVlcmludGYxNDUxMzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145138',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '11290719260810',
                                                                       'in_ucast_pkts': '8689583679',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '11290719260810',
                                                                                                'in_ucast_pkts': '8689583679',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '983604945228',
                                                                                                'out_ucast_pkts': '2764223582',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '983604945228',
                                                                       'out_ucast_pkts': '2764223582',
                                                                       'poller_hash': 'OTU2RTMycGVlcmludGZwb24gMTQ1MTM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145138',
                                                                                               'links': [       {       'connected_time': '1965803',
                                                                                                                        'fiber_length': 23039,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTMycGVlcnBvbnBvbiAxNDUxMzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '187 FERRIS DR (TRUSSELL JASON)',
                                         'poller_hash': 'OTU2RTMy',
                                         'serial_number': '956E32',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:48:64',
                                         'uptime': '1965803'},
                                 {       'counter_source': '145165/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145165/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '457014585137',
                                                                       'in_ucast_pkts': '1261365505',
                                                                       'mac_address': 'd0:76:8f:aa:4f:2c',
                                                                       'name': '145165/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '92',
                                                                       'out_octets': '6670152845760',
                                                                       'out_ucast_pkts': '4961636041',
                                                                       'poller_hash': 'OTU2RjBCcGVlcmludGYxNDUxNjUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145165',
                                                                       'in_nucast_pkts': '92',
                                                                       'in_octets': '6670152845760',
                                                                       'in_ucast_pkts': '4961636041',
                                                                       'links': [       {       'in_nucast_pkts': '92',
                                                                                                'in_octets': '6670152845760',
                                                                                                'in_ucast_pkts': '4961636041',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '457014585137',
                                                                                                'out_ucast_pkts': '1261365505',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '457014585137',
                                                                       'out_ucast_pkts': '1261365505',
                                                                       'poller_hash': 'OTU2RjBCcGVlcmludGZwb24gMTQ1MTY1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145165',
                                                                                               'links': [       {       'connected_time': '73026156',
                                                                                                                        'fiber_length': 23015,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjBCcGVlcnBvbnBvbiAxNDUxNjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '73 LAWRENCE ST (SINDALL ASHLEY)',
                                         'poller_hash': 'OTU2RjBC',
                                         'serial_number': '956F0B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:4f:2c',
                                         'uptime': '73026156'},
                                 {       'counter_source': '145212/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d0:0d'],
                                                                       'id': '145212/g1',
                                                                       'in_nucast_pkts': '5565942',
                                                                       'in_octets': '499993496335',
                                                                       'in_ucast_pkts': '1434164150',
                                                                       'mac_address': 'd0:76:8f:aa:49:07',
                                                                       'mtu': 2000,
                                                                       'name': '145212/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7686164651',
                                                                       'out_octets': '15326734099184',
                                                                       'out_ucast_pkts': '3751420060',
                                                                       'poller_hash': 'OTU2RTQ2cGVlcmludGYxNDUyMTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145212',
                                                                       'in_nucast_pkts': '7686164651',
                                                                       'in_octets': '15326734099184',
                                                                       'in_ucast_pkts': '3751420060',
                                                                       'links': [       {       'in_nucast_pkts': '7686164651',
                                                                                                'in_octets': '15326734099184',
                                                                                                'in_ucast_pkts': '3751420060',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5565942',
                                                                                                'out_octets': '499993496335',
                                                                                                'out_ucast_pkts': '1434164150',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5565942',
                                                                       'out_octets': '499993496335',
                                                                       'out_ucast_pkts': '1434164150',
                                                                       'poller_hash': 'OTU2RTQ2cGVlcmludGZwb24gMTQ1MjEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145212',
                                                                                               'links': [       {       'connected_time': '8535334',
                                                                                                                        'fiber_length': 23025,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTQ2cGVlcnBvbnBvbiAxNDUyMTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '69 LAWRENCE ST (ANDERSON PILAR)',
                                         'poller_hash': 'OTU2RTQ2',
                                         'serial_number': '956E46',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:49:04',
                                         'uptime': '8535334'},
                                 {       'counter_source': '145249/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145249/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '744161835304',
                                                                       'in_ucast_pkts': '10209648228',
                                                                       'mac_address': 'd0:76:8f:aa:49:74',
                                                                       'name': '145249/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '84',
                                                                       'out_octets': '3210153646976',
                                                                       'out_ucast_pkts': '2844526005',
                                                                       'poller_hash': 'OTU2RTU0cGVlcmludGYxNDUyNDkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145249',
                                                                       'in_nucast_pkts': '84',
                                                                       'in_octets': '3210153646976',
                                                                       'in_ucast_pkts': '2844526005',
                                                                       'links': [       {       'in_nucast_pkts': '84',
                                                                                                'in_octets': '3210153646976',
                                                                                                'in_ucast_pkts': '2844526005',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '744161835304',
                                                                                                'out_ucast_pkts': '10209648228',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '744161835304',
                                                                       'out_ucast_pkts': '10209648228',
                                                                       'poller_hash': 'OTU2RTU0cGVlcmludGZwb24gMTQ1MjQ5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145249',
                                                                                               'links': [       {       'connected_time': '9748914',
                                                                                                                        'fiber_length': 23287,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTU0cGVlcnBvbnBvbiAxNDUyNDk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '35 ZOEGER CT (BUERKLE HEATHER)',
                                         'poller_hash': 'OTU2RTU0',
                                         'serial_number': '956E54',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:49:74',
                                         'uptime': '9748914'},
                                 {       'counter_source': '145279/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145279/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '201563588143',
                                                                       'in_ucast_pkts': '294874766',
                                                                       'mac_address': 'd0:76:8f:aa:4e:e4',
                                                                       'name': '145279/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '86',
                                                                       'out_octets': '325276687887',
                                                                       'out_ucast_pkts': '359343412',
                                                                       'poller_hash': 'OTU2RjAycGVlcmludGYxNDUyNzkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145279',
                                                                       'in_nucast_pkts': '86',
                                                                       'in_octets': '325276687887',
                                                                       'in_ucast_pkts': '359343412',
                                                                       'links': [       {       'in_nucast_pkts': '86',
                                                                                                'in_octets': '325276687887',
                                                                                                'in_ucast_pkts': '359343412',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '201563588143',
                                                                                                'out_ucast_pkts': '294874766',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '201563588143',
                                                                       'out_ucast_pkts': '294874766',
                                                                       'poller_hash': 'OTU2RjAycGVlcmludGZwb24gMTQ1Mjc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145279',
                                                                                               'links': [       {       'connected_time': '21886857',
                                                                                                                        'fiber_length': 23913,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -25.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjAycGVlcnBvbnBvbiAxNDUyNzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3706 NAFZIGER RD (KATES KINETICS)',
                                         'poller_hash': 'OTU2RjAy',
                                         'serial_number': '956F02',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:4e:e4',
                                         'uptime': '21886857'},
                                 {       'counter_source': '145299/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d6:27'],
                                                                       'id': '145299/g1',
                                                                       'in_nucast_pkts': '160647',
                                                                       'in_octets': '112417894510',
                                                                       'in_ucast_pkts': '357556129',
                                                                       'mac_address': 'd0:76:8f:aa:47:c7',
                                                                       'mtu': 2000,
                                                                       'name': '145299/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797735',
                                                                       'out_octets': '1663596927123',
                                                                       'out_ucast_pkts': '1255722599',
                                                                       'poller_hash': 'OTU2RTFFcGVlcmludGYxNDUyOTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145299',
                                                                       'in_nucast_pkts': '9797735',
                                                                       'in_octets': '1663596927123',
                                                                       'in_ucast_pkts': '1255722599',
                                                                       'links': [       {       'in_nucast_pkts': '9797735',
                                                                                                'in_octets': '1663596927123',
                                                                                                'in_ucast_pkts': '1255722599',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160647',
                                                                                                'out_octets': '112417894510',
                                                                                                'out_ucast_pkts': '357556129',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160647',
                                                                       'out_octets': '112417894510',
                                                                       'out_ucast_pkts': '357556129',
                                                                       'poller_hash': 'OTU2RTFFcGVlcmludGZwb24gMTQ1Mjk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145299',
                                                                                               'links': [       {       'connected_time': '8535345',
                                                                                                                        'fiber_length': 23590,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTFFcGVlcnBvbnBvbiAxNDUyOTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '44 ZINKANN CR (QUINT MELISSA)',
                                         'poller_hash': 'OTU2RTFF',
                                         'serial_number': '956E1E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:47:c4',
                                         'uptime': '8535345'},
                                 {       'counter_source': '145528/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:d0:06'],
                                                                       'id': '145528/g1',
                                                                       'in_nucast_pkts': '705091',
                                                                       'in_octets': '208485139398',
                                                                       'in_ucast_pkts': '1384153739',
                                                                       'mac_address': 'd0:76:8f:aa:53:2f',
                                                                       'mtu': 2000,
                                                                       'name': '145528/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63490020',
                                                                       'out_octets': '5269134528414',
                                                                       'out_ucast_pkts': '3927492952',
                                                                       'poller_hash': 'OTU2RjhCcGVlcmludGYxNDU1MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145528',
                                                                       'in_nucast_pkts': '63490020',
                                                                       'in_octets': '5269134528414',
                                                                       'in_ucast_pkts': '3927492952',
                                                                       'links': [       {       'in_nucast_pkts': '63490020',
                                                                                                'in_octets': '5269134528414',
                                                                                                'in_ucast_pkts': '3927492952',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '705091',
                                                                                                'out_octets': '208485139398',
                                                                                                'out_ucast_pkts': '1384153739',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '705091',
                                                                       'out_octets': '208485139398',
                                                                       'out_ucast_pkts': '1384153739',
                                                                       'poller_hash': 'OTU2RjhCcGVlcmludGZwb24gMTQ1NTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145528',
                                                                                               'links': [       {       'connected_time': '21887001',
                                                                                                                        'fiber_length': 24982,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjhCcGVlcnBvbnBvbiAxNDU1Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 SHADYWOOD CT (ENS GERHARD)',
                                         'poller_hash': 'OTU2RjhC',
                                         'serial_number': '956F8B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:53:2c',
                                         'uptime': '21887001'},
                                 {       'counter_source': '145548/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d5:1f'],
                                                                       'id': '145548/g1',
                                                                       'in_nucast_pkts': '159044',
                                                                       'in_octets': '116121868024',
                                                                       'in_ucast_pkts': '281284096',
                                                                       'mac_address': 'd0:76:8f:aa:53:f7',
                                                                       'mtu': 2000,
                                                                       'name': '145548/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9874974',
                                                                       'out_octets': '1142879781283',
                                                                       'out_ucast_pkts': '886597702',
                                                                       'poller_hash': 'OTU2RkE0cGVlcmludGYxNDU1NDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145548',
                                                                       'in_nucast_pkts': '9874974',
                                                                       'in_octets': '1142879781283',
                                                                       'in_ucast_pkts': '886597702',
                                                                       'links': [       {       'in_nucast_pkts': '9874974',
                                                                                                'in_octets': '1142879781283',
                                                                                                'in_ucast_pkts': '886597702',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '159044',
                                                                                                'out_octets': '116121868024',
                                                                                                'out_ucast_pkts': '281284096',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '159044',
                                                                       'out_octets': '116121868024',
                                                                       'out_ucast_pkts': '281284096',
                                                                       'poller_hash': 'OTU2RkE0cGVlcmludGZwb24gMTQ1NTQ4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145548',
                                                                                               'links': [       {       'connected_time': '8535185',
                                                                                                                        'fiber_length': 23460,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RkE0cGVlcnBvbnBvbiAxNDU1NDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '138 FERRIS DR (LANTEIGNE CONNIE)',
                                         'poller_hash': 'OTU2RkE0',
                                         'serial_number': '956FA4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:53:f4',
                                         'uptime': '8535185'},
                                 {       'counter_source': '145558/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:59:c0:ac:a6:ee'],
                                                                       'id': '145558/g1',
                                                                       'in_nucast_pkts': '2557132',
                                                                       'in_octets': '7053937086',
                                                                       'in_ucast_pkts': '22207290',
                                                                       'mac_address': '48:77:46:8c:6d:60',
                                                                       'mtu': 2000,
                                                                       'name': '145558/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2257874',
                                                                       'out_octets': '35528952849',
                                                                       'out_ucast_pkts': '38298722',
                                                                       'poller_hash': 'NzZCODREcGVlcmludGYxNDU1NTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145558',
                                                                       'in_nucast_pkts': '2257874',
                                                                       'in_octets': '35528952849',
                                                                       'in_ucast_pkts': '38298722',
                                                                       'links': [       {       'in_nucast_pkts': '2257874',
                                                                                                'in_octets': '35528952849',
                                                                                                'in_ucast_pkts': '38298722',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2557132',
                                                                                                'out_octets': '7053937086',
                                                                                                'out_ucast_pkts': '22207290',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2557132',
                                                                       'out_octets': '7053937086',
                                                                       'out_ucast_pkts': '22207290',
                                                                       'poller_hash': 'NzZCODREcGVlcmludGZwb24gMTQ1NTU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145558',
                                                                                               'links': [       {       'connected_time': '3370045',
                                                                                                                        'fiber_length': 36173,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzZCODREcGVlcnBvbnBvbiAxNDU1NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2632 NOTRE DAME DR (BROWN WADE)',
                                         'poller_hash': 'NzZCODRE',
                                         'serial_number': '76B84D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:8c:6d:5c',
                                         'uptime': '3370045'},
                                 {       'counter_source': '145568/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d3:30'],
                                                                       'id': '145568/g1',
                                                                       'in_nucast_pkts': '150495',
                                                                       'in_octets': '67462213469',
                                                                       'in_ucast_pkts': '251133008',
                                                                       'mac_address': 'd0:76:8f:aa:48:5f',
                                                                       'mtu': 2000,
                                                                       'name': '145568/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797736',
                                                                       'out_octets': '1490684767664',
                                                                       'out_ucast_pkts': '1113507393',
                                                                       'poller_hash': 'OTU2RTMxcGVlcmludGYxNDU1NjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145568',
                                                                       'in_nucast_pkts': '9797736',
                                                                       'in_octets': '1490684767664',
                                                                       'in_ucast_pkts': '1113507393',
                                                                       'links': [       {       'in_nucast_pkts': '9797736',
                                                                                                'in_octets': '1490684767664',
                                                                                                'in_ucast_pkts': '1113507393',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '150495',
                                                                                                'out_octets': '67462213469',
                                                                                                'out_ucast_pkts': '251133008',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150495',
                                                                       'out_octets': '67462213469',
                                                                       'out_ucast_pkts': '251133008',
                                                                       'poller_hash': 'OTU2RTMxcGVlcmludGZwb24gMTQ1NTY4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145568',
                                                                                               'links': [       {       'connected_time': '8535331',
                                                                                                                        'fiber_length': 23377,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTMxcGVlcnBvbnBvbiAxNDU1Njg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '41 LAWRENCE ST (PETRIE CHERYL)',
                                         'poller_hash': 'OTU2RTMx',
                                         'serial_number': '956E31',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:48:5c',
                                         'uptime': '8535331'},
                                 {       'counter_source': '145731/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d4:22'],
                                                                       'id': '145731/g1',
                                                                       'in_nucast_pkts': '152578',
                                                                       'in_octets': '64603184283',
                                                                       'in_ucast_pkts': '326836629',
                                                                       'mac_address': 'd0:76:8f:aa:53:57',
                                                                       'mtu': 2000,
                                                                       'name': '145731/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797815',
                                                                       'out_octets': '1329736349004',
                                                                       'out_ucast_pkts': '986258214',
                                                                       'poller_hash': 'OTU2RjkwcGVlcmludGYxNDU3MzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145731',
                                                                       'in_nucast_pkts': '9797815',
                                                                       'in_octets': '1329736349004',
                                                                       'in_ucast_pkts': '986258214',
                                                                       'links': [       {       'in_nucast_pkts': '9797815',
                                                                                                'in_octets': '1329736349004',
                                                                                                'in_ucast_pkts': '986258214',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '152578',
                                                                                                'out_octets': '64603184283',
                                                                                                'out_ucast_pkts': '326836629',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '152578',
                                                                       'out_octets': '64603184283',
                                                                       'out_ucast_pkts': '326836629',
                                                                       'poller_hash': 'OTU2RjkwcGVlcmludGZwb24gMTQ1NzMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145731',
                                                                                               'links': [       {       'connected_time': '8535310',
                                                                                                                        'fiber_length': 23708,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjkwcGVlcnBvbnBvbiAxNDU3MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '66 ZINKANN CR (GOLDEN SARAH)',
                                         'poller_hash': 'OTU2Rjkw',
                                         'serial_number': '956F90',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:53:54',
                                         'uptime': '8535310'},
                                 {       'counter_source': '145772/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145772/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '957741445244',
                                                                       'in_ucast_pkts': '116913296200',
                                                                       'mac_address': 'd0:76:8f:aa:49:1c',
                                                                       'name': '145772/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '1046097147561',
                                                                       'out_ucast_pkts': '1313849486',
                                                                       'poller_hash': 'OTU2RTQ5cGVlcmludGYxNDU3NzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145772',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '1046097147561',
                                                                       'in_ucast_pkts': '1313849486',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '1046097147561',
                                                                                                'in_ucast_pkts': '1313849486',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '957741445244',
                                                                                                'out_ucast_pkts': '116913296200',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '957741445244',
                                                                       'out_ucast_pkts': '116913296200',
                                                                       'poller_hash': 'OTU2RTQ5cGVlcmludGZwb24gMTQ1Nzcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145772',
                                                                                               'links': [       {       'connected_time': '21887065',
                                                                                                                        'fiber_length': 22822,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTQ5cGVlcnBvbnBvbiAxNDU3NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3 LAWRENCE ST (POOLE ELECTRIC LTD.)',
                                         'poller_hash': 'OTU2RTQ5',
                                         'serial_number': '956E49',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:49:1c',
                                         'uptime': '21887065'},
                                 {       'counter_source': '145812/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145812/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '118538017550',
                                                                       'in_ucast_pkts': '431854538',
                                                                       'mac_address': 'd0:76:8f:aa:51:04',
                                                                       'name': '145812/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '67',
                                                                       'out_octets': '1097047121425',
                                                                       'out_ucast_pkts': '958463336',
                                                                       'poller_hash': 'OTU2RjQ2cGVlcmludGYxNDU4MTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145812',
                                                                       'in_nucast_pkts': '67',
                                                                       'in_octets': '1097047121425',
                                                                       'in_ucast_pkts': '958463336',
                                                                       'links': [       {       'in_nucast_pkts': '67',
                                                                                                'in_octets': '1097047121425',
                                                                                                'in_ucast_pkts': '958463336',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '118538017550',
                                                                                                'out_ucast_pkts': '431854538',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '118538017550',
                                                                       'out_ucast_pkts': '431854538',
                                                                       'poller_hash': 'OTU2RjQ2cGVlcmludGZwb24gMTQ1ODEy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145812',
                                                                                               'links': [       {       'connected_time': '4222606',
                                                                                                                        'fiber_length': 23735,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjQ2cGVlcnBvbnBvbiAxNDU4MTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3700 NAFZIGER RD (RESOLUTE WELLNESS '
                                                 'INC)',
                                         'poller_hash': 'OTU2RjQ2',
                                         'serial_number': '956F46',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:51:04',
                                         'uptime': '4222606'},
                                 {       'counter_source': '145832/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f5:48'],
                                                                       'id': '145832/g1',
                                                                       'in_nucast_pkts': '119483',
                                                                       'in_octets': '62721805619',
                                                                       'in_ucast_pkts': '201994363',
                                                                       'mac_address': '60:db:98:4e:97:e7',
                                                                       'mtu': 2000,
                                                                       'name': '145832/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7420838',
                                                                       'out_octets': '730338398070',
                                                                       'out_ucast_pkts': '605047082',
                                                                       'poller_hash': 'QTlDOEJDcGVlcmludGYxNDU4MzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145832',
                                                                       'in_nucast_pkts': '7420838',
                                                                       'in_octets': '730338398070',
                                                                       'in_ucast_pkts': '605047082',
                                                                       'links': [       {       'in_nucast_pkts': '7420838',
                                                                                                'in_octets': '730338398070',
                                                                                                'in_ucast_pkts': '605047082',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '119483',
                                                                                                'out_octets': '62721805619',
                                                                                                'out_ucast_pkts': '201994363',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '119483',
                                                                       'out_octets': '62721805619',
                                                                       'out_ucast_pkts': '201994363',
                                                                       'poller_hash': 'QTlDOEJDcGVlcmludGZwb24gMTQ1ODMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145832',
                                                                                               'links': [       {       'connected_time': '6396009',
                                                                                                                        'fiber_length': 24932,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOEJDcGVlcnBvbnBvbiAxNDU4MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '35 NITHBANK RD (ERB CAROLYN)',
                                         'poller_hash': 'QTlDOEJD',
                                         'serial_number': 'A9C8BC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:97:e4',
                                         'uptime': '6396009'},
                                 {       'counter_source': '145852/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145852/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '26364584834',
                                                                       'in_ucast_pkts': '213675042',
                                                                       'mac_address': '60:db:98:4e:67:7c',
                                                                       'name': '145852/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '53',
                                                                       'out_octets': '1012391049187',
                                                                       'out_ucast_pkts': '751959371',
                                                                       'poller_hash': 'QTlDMkFGcGVlcmludGYxNDU4NTIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145852',
                                                                       'in_nucast_pkts': '53',
                                                                       'in_octets': '1012391049187',
                                                                       'in_ucast_pkts': '751959371',
                                                                       'links': [       {       'in_nucast_pkts': '53',
                                                                                                'in_octets': '1012391049187',
                                                                                                'in_ucast_pkts': '751959371',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '26364584834',
                                                                                                'out_ucast_pkts': '213675042',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '26364584834',
                                                                       'out_ucast_pkts': '213675042',
                                                                       'poller_hash': 'QTlDMkFGcGVlcmludGZwb24gMTQ1ODUy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145852',
                                                                                               'links': [       {       'connected_time': '611462',
                                                                                                                        'fiber_length': 11880,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkFGcGVlcnBvbnBvbiAxNDU4NTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5596 PERTH LINE 56A (KUEPFER '
                                                 'BARBARA)',
                                         'poller_hash': 'QTlDMkFG',
                                         'serial_number': 'A9C2AF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:67:7c',
                                         'uptime': '611462'},
                                 {       'counter_source': '145872/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d6:7f'],
                                                                       'id': '145872/g1',
                                                                       'in_nucast_pkts': '12005518',
                                                                       'in_octets': '190783911197',
                                                                       'in_ucast_pkts': '753640468',
                                                                       'mac_address': '60:db:98:4e:6f:17',
                                                                       'mtu': 2000,
                                                                       'name': '145872/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16509356554',
                                                                       'out_octets': '25377688791509',
                                                                       'out_ucast_pkts': '2548817797',
                                                                       'poller_hash': 'QTlDM0EycGVlcmludGYxNDU4NzIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145872',
                                                                       'in_nucast_pkts': '16509356554',
                                                                       'in_octets': '25377688791509',
                                                                       'in_ucast_pkts': '2548817797',
                                                                       'links': [       {       'in_nucast_pkts': '16509356554',
                                                                                                'in_octets': '25377688791509',
                                                                                                'in_ucast_pkts': '2548817797',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12005518',
                                                                                                'out_octets': '190783911197',
                                                                                                'out_ucast_pkts': '753640468',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12005518',
                                                                       'out_octets': '190783911197',
                                                                       'out_ucast_pkts': '753640468',
                                                                       'poller_hash': 'QTlDM0EycGVlcmludGZwb24gMTQ1ODcy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145872',
                                                                                               'links': [       {       'connected_time': '21886623',
                                                                                                                        'fiber_length': 22901,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0EycGVlcnBvbnBvbiAxNDU4NzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '26 POND VIEW DR (ALBRECHT ORIE)',
                                         'poller_hash': 'QTlDM0Ey',
                                         'serial_number': 'A9C3A2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6f:14',
                                         'uptime': '21886623'},
                                 {       'counter_source': '145919/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145919/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '518008115623',
                                                                       'in_ucast_pkts': '84978936844',
                                                                       'mac_address': 'd0:76:8f:e7:24:43',
                                                                       'name': '145919/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '76',
                                                                       'out_octets': '12795237009973',
                                                                       'out_ucast_pkts': '8785114391',
                                                                       'poller_hash': 'OUMyRkUwcGVlcmludGYxNDU5MTkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145919',
                                                                       'in_nucast_pkts': '76',
                                                                       'in_octets': '12795237009973',
                                                                       'in_ucast_pkts': '8785114391',
                                                                       'links': [       {       'in_nucast_pkts': '76',
                                                                                                'in_octets': '12795237009973',
                                                                                                'in_ucast_pkts': '8785114391',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '518008115623',
                                                                                                'out_ucast_pkts': '84978936844',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '518008115623',
                                                                       'out_ucast_pkts': '84978936844',
                                                                       'poller_hash': 'OUMyRkUwcGVlcmludGZwb24gMTQ1OTE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145919',
                                                                                               'links': [       {       'connected_time': '8535372',
                                                                                                                        'fiber_length': 22442,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OUMyRkUwcGVlcnBvbnBvbiAxNDU5MTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '81 POND VIEW DR (FUTHER KELSEY)',
                                         'poller_hash': 'OUMyRkUw',
                                         'serial_number': '9C2FE0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:e7:24:43',
                                         'uptime': '8535372'},
                                 {       'counter_source': '145930/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '40:f2:e9:dc:d0:64'],
                                                                       'id': '145930/g1',
                                                                       'in_nucast_pkts': '18873024',
                                                                       'in_octets': '79948801056110',
                                                                       'in_ucast_pkts': '83238526526',
                                                                       'mac_address': '60:db:98:4e:69:3f',
                                                                       'mtu': 2000,
                                                                       'name': '145930/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61748947',
                                                                       'out_octets': '129963125544953',
                                                                       'out_ucast_pkts': '111021986962',
                                                                       'poller_hash': 'QTlDMkU3cGVlcmludGYxNDU5MzAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '145930',
                                                                       'in_nucast_pkts': '61748947',
                                                                       'in_octets': '129963125544953',
                                                                       'in_ucast_pkts': '111021986962',
                                                                       'links': [       {       'in_nucast_pkts': '61748947',
                                                                                                'in_octets': '129963125544953',
                                                                                                'in_ucast_pkts': '111021986962',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18873024',
                                                                                                'out_octets': '79948801056110',
                                                                                                'out_ucast_pkts': '83238526526',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18873024',
                                                                       'out_octets': '79948801056110',
                                                                       'out_ucast_pkts': '83238526526',
                                                                       'poller_hash': 'QTlDMkU3cGVlcmludGZwb24gMTQ1OTMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145930',
                                                                                               'links': [       {       'connected_time': '18797163',
                                                                                                                        'fiber_length': 22595,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkU3cGVlcnBvbnBvbiAxNDU5MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '68 POND VIEW DR (MARTIN BRETT)',
                                         'poller_hash': 'QTlDMkU3',
                                         'serial_number': 'A9C2E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:69:3c',
                                         'uptime': '18797163'},
                                 {       'counter_source': '145940/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145940/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '395315463296',
                                                                       'in_ucast_pkts': '5689261181',
                                                                       'mac_address': '60:db:98:4e:71:0c',
                                                                       'name': '145940/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '37',
                                                                       'out_octets': '5096451042607',
                                                                       'out_ucast_pkts': '3812425713',
                                                                       'poller_hash': 'QTlDM0UxcGVlcmludGYxNDU5NDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145940',
                                                                       'in_nucast_pkts': '37',
                                                                       'in_octets': '5096451042607',
                                                                       'in_ucast_pkts': '3812425713',
                                                                       'links': [       {       'in_nucast_pkts': '37',
                                                                                                'in_octets': '5096451042607',
                                                                                                'in_ucast_pkts': '3812425713',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '395315463296',
                                                                                                'out_ucast_pkts': '5689261181',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '395315463296',
                                                                       'out_ucast_pkts': '5689261181',
                                                                       'poller_hash': 'QTlDM0UxcGVlcmludGZwb24gMTQ1OTQw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145940',
                                                                                               'links': [       {       'connected_time': '10946538',
                                                                                                                        'fiber_length': 25326,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0UxcGVlcnBvbnBvbiAxNDU5NDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 JACK KOEHLER LN (WENDE JASON)',
                                         'poller_hash': 'QTlDM0Ux',
                                         'serial_number': 'A9C3E1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:71:0c',
                                         'uptime': '10946538'},
                                 {       'counter_source': '145991/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '145991/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '463563210513',
                                                                       'in_ucast_pkts': '1787644792',
                                                                       'mac_address': '60:db:98:4e:71:a4',
                                                                       'name': '145991/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '81',
                                                                       'out_octets': '9152646432426',
                                                                       'out_ucast_pkts': '7077470006',
                                                                       'poller_hash': 'QTlDM0Y0cGVlcmludGYxNDU5OTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '145991',
                                                                       'in_nucast_pkts': '81',
                                                                       'in_octets': '9152646432426',
                                                                       'in_ucast_pkts': '7077470006',
                                                                       'links': [       {       'in_nucast_pkts': '81',
                                                                                                'in_octets': '9152646432426',
                                                                                                'in_ucast_pkts': '7077470006',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '463563210513',
                                                                                                'out_ucast_pkts': '1787644792',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '463563210513',
                                                                       'out_ucast_pkts': '1787644792',
                                                                       'poller_hash': 'QTlDM0Y0cGVlcmludGZwb24gMTQ1OTkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '145991',
                                                                                               'links': [       {       'connected_time': '3169492',
                                                                                                                        'fiber_length': 24142,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0Y0cGVlcnBvbnBvbiAxNDU5OTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1044 QUEEN'S BUSH RD (PANKIW ANDRE)",
                                         'poller_hash': 'QTlDM0Y0',
                                         'serial_number': 'A9C3F4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:71:a4',
                                         'uptime': '3169492'},
                                 {       'counter_source': '146021/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146021/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1816852732681',
                                                                       'in_ucast_pkts': '37927309640',
                                                                       'mac_address': '60:db:98:4e:72:9c',
                                                                       'name': '146021/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '57',
                                                                       'out_octets': '6994749673262',
                                                                       'out_ucast_pkts': '7275309964',
                                                                       'poller_hash': 'QTlDNDEzcGVlcmludGYxNDYwMjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146021',
                                                                       'in_nucast_pkts': '57',
                                                                       'in_octets': '6994749673262',
                                                                       'in_ucast_pkts': '7275309964',
                                                                       'links': [       {       'in_nucast_pkts': '57',
                                                                                                'in_octets': '6994749673262',
                                                                                                'in_ucast_pkts': '7275309964',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1816852732681',
                                                                                                'out_ucast_pkts': '37927309640',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1816852732681',
                                                                       'out_ucast_pkts': '37927309640',
                                                                       'poller_hash': 'QTlDNDEzcGVlcmludGZwb24gMTQ2MDIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146021',
                                                                                               'links': [       {       'connected_time': '1096784',
                                                                                                                        'fiber_length': 32033,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNDEzcGVlcnBvbnBvbiAxNDYwMjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1561 MOSER-YOUNG RD (OPENSPACE '
                                                 'SOLUTIONS INC)',
                                         'poller_hash': 'QTlDNDEz',
                                         'serial_number': 'A9C413',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:72:9c',
                                         'uptime': '1096784'},
                                 {       'counter_source': '146051/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:db:49:4d',
                                                                                                        '00:03:e6:db:4f:8d',
                                                                                                        'd0:76:8f:93:c8:1b'],
                                                                       'id': '146051/g1',
                                                                       'in_nucast_pkts': '3787867',
                                                                       'in_octets': '66632331097',
                                                                       'in_ucast_pkts': '201313874',
                                                                       'mac_address': '60:db:98:4e:72:cf',
                                                                       'mtu': 2000,
                                                                       'name': '146051/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10547580733',
                                                                       'out_octets': '15120671153763',
                                                                       'out_ucast_pkts': '671146199',
                                                                       'poller_hash': 'QTlDNDE5cGVlcmludGYxNDYwNTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '146051',
                                                                       'in_nucast_pkts': '10547580733',
                                                                       'in_octets': '15120671153763',
                                                                       'in_ucast_pkts': '671146199',
                                                                       'links': [       {       'in_nucast_pkts': '10547580733',
                                                                                                'in_octets': '15120671153763',
                                                                                                'in_ucast_pkts': '671146199',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3787867',
                                                                                                'out_octets': '66632331097',
                                                                                                'out_ucast_pkts': '201313874',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3787867',
                                                                       'out_octets': '66632331097',
                                                                       'out_ucast_pkts': '201313874',
                                                                       'poller_hash': 'QTlDNDE5cGVlcmludGZwb24gMTQ2MDUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146051',
                                                                                               'links': [       {       'connected_time': '8535286',
                                                                                                                        'fiber_length': 23148,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDNDE5cGVlcnBvbnBvbiAxNDYwNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '55 ZOEGER CT (MORRISON LISA)',
                                         'poller_hash': 'QTlDNDE5',
                                         'serial_number': 'A9C419',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:72:cc',
                                         'uptime': '8535286'},
                                 {       'counter_source': '146093/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:8d:0d:cc'],
                                                                       'id': '146093/g1',
                                                                       'in_nucast_pkts': '74819',
                                                                       'in_octets': '199270031750',
                                                                       'in_ucast_pkts': '295457951',
                                                                       'mac_address': 'd0:76:8f:aa:49:17',
                                                                       'mtu': 2000,
                                                                       'name': '146093/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4922862',
                                                                       'out_octets': '633414614430',
                                                                       'out_ucast_pkts': '603079358',
                                                                       'poller_hash': 'OTU2RTQ4cGVlcmludGYxNDYwOTMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '146093',
                                                                       'in_nucast_pkts': '4922862',
                                                                       'in_octets': '633414614430',
                                                                       'in_ucast_pkts': '603079358',
                                                                       'links': [       {       'in_nucast_pkts': '4922862',
                                                                                                'in_octets': '633414614430',
                                                                                                'in_ucast_pkts': '603079358',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '74819',
                                                                                                'out_octets': '199270031750',
                                                                                                'out_ucast_pkts': '295457951',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '74819',
                                                                       'out_octets': '199270031750',
                                                                       'out_ucast_pkts': '295457951',
                                                                       'poller_hash': 'OTU2RTQ4cGVlcmludGZwb24gMTQ2MDkz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146093',
                                                                                               'links': [       {       'connected_time': '4221509',
                                                                                                                        'fiber_length': 24395,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RTQ4cGVlcnBvbnBvbiAxNDYwOTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '89 PARKVIEW DR (HARCOURT CHRIS)',
                                         'poller_hash': 'OTU2RTQ4',
                                         'serial_number': '956E48',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:49:14',
                                         'uptime': '4221509'},
                                 {       'counter_source': '146160/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:8d:0f:42'],
                                                                       'id': '146160/g1',
                                                                       'in_nucast_pkts': '26682852',
                                                                       'in_octets': '540786640806',
                                                                       'in_ucast_pkts': '1557460295',
                                                                       'mac_address': 'd0:76:8f:aa:4e:c7',
                                                                       'mtu': 2000,
                                                                       'name': '146160/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61923185955',
                                                                       'out_octets': '92482202815400',
                                                                       'out_ucast_pkts': '7023753504',
                                                                       'poller_hash': 'OTU2RUZFcGVlcmludGYxNDYxNjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '146160',
                                                                       'in_nucast_pkts': '61923185955',
                                                                       'in_octets': '92482202815400',
                                                                       'in_ucast_pkts': '7023753504',
                                                                       'links': [       {       'in_nucast_pkts': '61923185955',
                                                                                                'in_octets': '92482202815400',
                                                                                                'in_ucast_pkts': '7023753504',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '26682852',
                                                                                                'out_octets': '540786640806',
                                                                                                'out_ucast_pkts': '1557460295',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26682852',
                                                                       'out_octets': '540786640806',
                                                                       'out_ucast_pkts': '1557460295',
                                                                       'poller_hash': 'OTU2RUZFcGVlcmludGZwb24gMTQ2MTYw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146160',
                                                                                               'links': [       {       'connected_time': '21886847',
                                                                                                                        'fiber_length': 24555,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RUZFcGVlcnBvbnBvbiAxNDYxNjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '95 Parkview Dr (Bonnie Roth)',
                                         'poller_hash': 'OTU2RUZF',
                                         'serial_number': '956EFE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:4e:c4',
                                         'uptime': '21886847'},
                                 {       'counter_source': '146169/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146169/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '218143006952',
                                                                       'in_ucast_pkts': '48074021927',
                                                                       'mac_address': 'd0:76:8f:aa:4f:74',
                                                                       'name': '146169/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '4228165485230',
                                                                       'out_ucast_pkts': '3325369341',
                                                                       'poller_hash': 'OTU2RjE0cGVlcmludGYxNDYxNjkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146169',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '4228165485230',
                                                                       'in_ucast_pkts': '3325369341',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '4228165485230',
                                                                                                'in_ucast_pkts': '3325369341',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '218143006952',
                                                                                                'out_ucast_pkts': '48074021927',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '218143006952',
                                                                       'out_ucast_pkts': '48074021927',
                                                                       'poller_hash': 'OTU2RjE0cGVlcmludGZwb24gMTQ2MTY5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146169',
                                                                                               'links': [       {       'connected_time': '8535281',
                                                                                                                        'fiber_length': 24788,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'OTU2RjE0cGVlcnBvbnBvbiAxNDYxNjk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '131 GERBER MEADOWS DR (ROBERTSON '
                                                 'AMY)',
                                         'poller_hash': 'OTU2RjE0',
                                         'serial_number': '956F14',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'd0:76:8f:aa:4f:74',
                                         'uptime': '8535281'},
                                 {       'counter_source': '146216/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146216/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '408936004848',
                                                                       'in_ucast_pkts': '1710641393',
                                                                       'mac_address': '60:db:98:4e:67:a4',
                                                                       'name': '146216/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '7039153217422',
                                                                       'out_ucast_pkts': '5185671829',
                                                                       'poller_hash': 'QTlDMkI0cGVlcmludGYxNDYyMTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146216',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '7039153217422',
                                                                       'in_ucast_pkts': '5185671829',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '7039153217422',
                                                                                                'in_ucast_pkts': '5185671829',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '408936004848',
                                                                                                'out_ucast_pkts': '1710641393',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '408936004848',
                                                                       'out_ucast_pkts': '1710641393',
                                                                       'poller_hash': 'QTlDMkI0cGVlcmludGZwb24gMTQ2MjE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146216',
                                                                                               'links': [       {       'connected_time': '2297172',
                                                                                                                        'fiber_length': 23065,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkI0cGVlcnBvbnBvbiAxNDYyMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '195 FERRIS DRIVE WELLESLEY (KUNG '
                                                 'KIMBERLEY)',
                                         'poller_hash': 'QTlDMkI0',
                                         'serial_number': 'A9C2B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:67:a4',
                                         'uptime': '2297172'},
                                 {       'counter_source': '146451/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '146451/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '476070975611',
                                                                       'in_ucast_pkts': '69745778124',
                                                                       'mac_address': '60:db:98:4e:6d:54',
                                                                       'name': '146451/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '169',
                                                                       'out_octets': '817716268781',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'QTlDMzZBcGVlcmludGYxNDY0NTEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '146451',
                                                                       'in_nucast_pkts': '169',
                                                                       'in_octets': '817716268781',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '169',
                                                                                                'in_octets': '817716268781',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '476070975611',
                                                                                                'out_ucast_pkts': '69745778124',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '476070975611',
                                                                       'out_ucast_pkts': '69745778124',
                                                                       'poller_hash': 'QTlDMzZBcGVlcmludGZwb24gMTQ2NDUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146451',
                                                                                               'links': [       {       'connected_time': '1462678',
                                                                                                                        'fiber_length': 22532,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMzZBcGVlcnBvbnBvbiAxNDY0NTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1201 QUEEN'S BUSH RD (COOKS "
                                                 'PHARMACY)',
                                         'poller_hash': 'QTlDMzZB',
                                         'serial_number': 'A9C36A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6d:54',
                                         'uptime': '1462678'},
                                 {       'counter_source': '146638/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:64:cf'],
                                                                       'id': '146638/g1',
                                                                       'in_nucast_pkts': '321285',
                                                                       'in_octets': '896739626171',
                                                                       'in_ucast_pkts': '1543577637',
                                                                       'mac_address': '60:db:98:4e:6d:e7',
                                                                       'mtu': 2000,
                                                                       'name': '146638/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21121628',
                                                                       'out_octets': '4538568365665',
                                                                       'out_ucast_pkts': '3667120822',
                                                                       'poller_hash': 'QTlDMzdDcGVlcmludGYxNDY2MzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '146638',
                                                                       'in_nucast_pkts': '21121628',
                                                                       'in_octets': '4538568365665',
                                                                       'in_ucast_pkts': '3667120822',
                                                                       'links': [       {       'in_nucast_pkts': '21121628',
                                                                                                'in_octets': '4538568365665',
                                                                                                'in_ucast_pkts': '3667120822',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '321285',
                                                                                                'out_octets': '896739626171',
                                                                                                'out_ucast_pkts': '1543577637',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '321285',
                                                                       'out_octets': '896739626171',
                                                                       'out_ucast_pkts': '1543577637',
                                                                       'poller_hash': 'QTlDMzdDcGVlcmludGZwb24gMTQ2NjM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146638',
                                                                                               'links': [       {       'connected_time': '17942222',
                                                                                                                        'fiber_length': 23188,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMzdDcGVlcnBvbnBvbiAxNDY2Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1155 QUEEN'S BUSH RD (BERTRAM "
                                                 'ALLISTER)',
                                         'poller_hash': 'QTlDMzdD',
                                         'serial_number': 'A9C37C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6d:e4',
                                         'uptime': '17942222'},
                                 {       'counter_source': '146907/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:ed:57'],
                                                                       'id': '146907/g1',
                                                                       'in_nucast_pkts': '160819',
                                                                       'in_octets': '197998098397',
                                                                       'in_ucast_pkts': '480177746',
                                                                       'mac_address': '60:db:98:4e:67:2f',
                                                                       'mtu': 2000,
                                                                       'name': '146907/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878131',
                                                                       'out_octets': '4058375024795',
                                                                       'out_ucast_pkts': '2838034073',
                                                                       'poller_hash': 'QTlDMkE1cGVlcmludGYxNDY5MDcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '146907',
                                                                       'in_nucast_pkts': '9878131',
                                                                       'in_octets': '4058375024795',
                                                                       'in_ucast_pkts': '2838034073',
                                                                       'links': [       {       'in_nucast_pkts': '9878131',
                                                                                                'in_octets': '4058375024795',
                                                                                                'in_ucast_pkts': '2838034073',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160819',
                                                                                                'out_octets': '197998098397',
                                                                                                'out_ucast_pkts': '480177746',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160819',
                                                                       'out_octets': '197998098397',
                                                                       'out_ucast_pkts': '480177746',
                                                                       'poller_hash': 'QTlDMkE1cGVlcmludGZwb24gMTQ2OTA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146907',
                                                                                               'links': [       {       'connected_time': '8535233',
                                                                                                                        'fiber_length': 22376,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkE1cGVlcnBvbnBvbiAxNDY5MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '49 DAVID ST (BUONO TERESA)',
                                         'poller_hash': 'QTlDMkE1',
                                         'serial_number': 'A9C2A5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:67:2c',
                                         'uptime': '8535233'},
                                 {       'counter_source': '146977/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d5:98'],
                                                                       'id': '146977/g1',
                                                                       'in_nucast_pkts': '927064',
                                                                       'in_octets': '167547250740',
                                                                       'in_ucast_pkts': '551982517',
                                                                       'mac_address': '60:db:98:4e:70:2f',
                                                                       'mtu': 2000,
                                                                       'name': '146977/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63493237',
                                                                       'out_octets': '1315179163791',
                                                                       'out_ucast_pkts': '1157988098',
                                                                       'poller_hash': 'QTlDM0M1cGVlcmludGYxNDY5NzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '146977',
                                                                       'in_nucast_pkts': '63493237',
                                                                       'in_octets': '1315179163791',
                                                                       'in_ucast_pkts': '1157988098',
                                                                       'links': [       {       'in_nucast_pkts': '63493237',
                                                                                                'in_octets': '1315179163791',
                                                                                                'in_ucast_pkts': '1157988098',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '927064',
                                                                                                'out_octets': '167547250740',
                                                                                                'out_ucast_pkts': '551982517',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '927064',
                                                                       'out_octets': '167547250740',
                                                                       'out_ucast_pkts': '551982517',
                                                                       'poller_hash': 'QTlDM0M1cGVlcmludGZwb24gMTQ2OTc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '146977',
                                                                                               'links': [       {       'connected_time': '21887572',
                                                                                                                        'fiber_length': 25052,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0M1cGVlcnBvbnBvbiAxNDY5Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 SHADYWOOD CT (LEIS GARY)',
                                         'poller_hash': 'QTlDM0M1',
                                         'serial_number': 'A9C3C5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:70:2c',
                                         'uptime': '21887572'},
                                 {       'counter_source': '147647/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '147647/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '479334883375',
                                                                       'in_ucast_pkts': '2721514229',
                                                                       'mac_address': '60:db:98:4e:6f:5c',
                                                                       'name': '147647/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '63',
                                                                       'out_octets': '15019572381907',
                                                                       'out_ucast_pkts': '11017526610',
                                                                       'poller_hash': 'QTlDM0FCcGVlcmludGYxNDc2NDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '147647',
                                                                       'in_nucast_pkts': '63',
                                                                       'in_octets': '15019572381907',
                                                                       'in_ucast_pkts': '11017526610',
                                                                       'links': [       {       'in_nucast_pkts': '63',
                                                                                                'in_octets': '15019572381907',
                                                                                                'in_ucast_pkts': '11017526610',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '479334883375',
                                                                                                'out_ucast_pkts': '2721514229',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '479334883375',
                                                                       'out_ucast_pkts': '2721514229',
                                                                       'poller_hash': 'QTlDM0FCcGVlcmludGZwb24gMTQ3NjQ3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '147647',
                                                                                               'links': [       {       'connected_time': '335345',
                                                                                                                        'fiber_length': 19174,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDM0FCcGVlcnBvbnBvbiAxNDc2NDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1907 PERTH LINE 56 (MANNHARDT '
                                                 'JOSHUA)',
                                         'poller_hash': 'QTlDM0FC',
                                         'serial_number': 'A9C3AB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:6f:5c',
                                         'uptime': '335345'},
                                 {       'counter_source': '149500/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149500/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '244401041208',
                                                                       'in_ucast_pkts': '441691511',
                                                                       'mac_address': '60:db:98:73:9b:cc',
                                                                       'name': '149500/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '136',
                                                                       'out_octets': '7449992300546',
                                                                       'out_ucast_pkts': '5836877918',
                                                                       'poller_hash': 'QjBBMjhCcGVlcmludGYxNDk1MDAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149500',
                                                                       'in_nucast_pkts': '136',
                                                                       'in_octets': '7449992300546',
                                                                       'in_ucast_pkts': '5836877918',
                                                                       'links': [       {       'in_nucast_pkts': '136',
                                                                                                'in_octets': '7449992300546',
                                                                                                'in_ucast_pkts': '5836877918',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '244401041208',
                                                                                                'out_ucast_pkts': '441691511',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '244401041208',
                                                                       'out_ucast_pkts': '441691511',
                                                                       'poller_hash': 'QjBBMjhCcGVlcmludGZwb24gMTQ5NTAw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149500',
                                                                                               'links': [       {       'connected_time': '8462792',
                                                                                                                        'fiber_length': 18754,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMjhCcGVlcnBvbnBvbiAxNDk1MDA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5039 PERTH ROAD 119 (RIEHL ROBIN)',
                                         'poller_hash': 'QjBBMjhC',
                                         'serial_number': 'B0A28B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:9b:cc',
                                         'uptime': '8462792'},
                                 {       'counter_source': '149621/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '149621/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45896304270',
                                                                       'in_ucast_pkts': '391906969',
                                                                       'mac_address': '60:db:98:73:9c:6c',
                                                                       'name': '149621/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '107',
                                                                       'out_octets': '2583834022141',
                                                                       'out_ucast_pkts': '5857676617',
                                                                       'poller_hash': 'QjBBMjlGcGVlcmludGYxNDk2MjEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '149621',
                                                                       'in_nucast_pkts': '107',
                                                                       'in_octets': '2583834022141',
                                                                       'in_ucast_pkts': '5857676617',
                                                                       'links': [       {       'in_nucast_pkts': '107',
                                                                                                'in_octets': '2583834022141',
                                                                                                'in_ucast_pkts': '5857676617',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45896304270',
                                                                                                'out_ucast_pkts': '391906969',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45896304270',
                                                                       'out_ucast_pkts': '391906969',
                                                                       'poller_hash': 'QjBBMjlGcGVlcmludGZwb24gMTQ5NjIx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149621',
                                                                                               'links': [       {       'connected_time': '12866516',
                                                                                                                        'fiber_length': 20181,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMjlGcGVlcnBvbnBvbiAxNDk2MjE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4879 TIFFANY CT (ROULSTON ROD & '
                                                 'GLORIA)',
                                         'poller_hash': 'QjBBMjlG',
                                         'serial_number': 'B0A29F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:9c:6c',
                                         'uptime': '12866516'},
                                 {       'counter_source': '149631/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:03:e6:db:2e:b7',
                                                                                                        '00:03:e6:db:48:37',
                                                                                                        '00:03:e6:db:4b:65',
                                                                                                        '00:03:e6:db:51:db',
                                                                                                        '60:db:98:55:4e:9a'],
                                                                       'id': '149631/g1',
                                                                       'in_nucast_pkts': '67381409',
                                                                       'in_octets': '318641638288',
                                                                       'in_ucast_pkts': '2082121233',
                                                                       'mac_address': '60:db:98:4e:69:af',
                                                                       'mtu': 2000,
                                                                       'name': '149631/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '148172408217',
                                                                       'out_octets': '212438091049524',
                                                                       'out_ucast_pkts': '8387699341',
                                                                       'poller_hash': 'QTlDMkY1cGVlcmludGYxNDk2MzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '149631',
                                                                       'in_nucast_pkts': '148172408217',
                                                                       'in_octets': '212438091049524',
                                                                       'in_ucast_pkts': '8387699341',
                                                                       'links': [       {       'in_nucast_pkts': '148172408217',
                                                                                                'in_octets': '212438091049524',
                                                                                                'in_ucast_pkts': '8387699341',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '67381409',
                                                                                                'out_octets': '318641638288',
                                                                                                'out_ucast_pkts': '2082121233',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67381409',
                                                                       'out_octets': '318641638288',
                                                                       'out_ucast_pkts': '2082121233',
                                                                       'poller_hash': 'QTlDMkY1cGVlcmludGZwb24gMTQ5NjMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '149631',
                                                                                               'links': [       {       'connected_time': '21887130',
                                                                                                                        'fiber_length': 25051,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDMkY1cGVlcnBvbnBvbiAxNDk2MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 PARKVIEW DR (YOST BRIAN)',
                                         'poller_hash': 'QTlDMkY1',
                                         'serial_number': 'A9C2F5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:69:ac',
                                         'uptime': '21887130'},
                                 {       'counter_source': '150131/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:ac:bb'],
                                                                       'id': '150131/g1',
                                                                       'in_nucast_pkts': '769216',
                                                                       'in_octets': '565564191804',
                                                                       'in_ucast_pkts': '1996567516',
                                                                       'mac_address': 'ec:4f:82:cc:11:00',
                                                                       'mtu': 2000,
                                                                       'name': '150131/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '22468751',
                                                                       'out_octets': '9090163230280',
                                                                       'out_ucast_pkts': '6749203937',
                                                                       'poller_hash': 'MzhGRjU4cGVlcmludGYxNTAxMzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '150131',
                                                                       'in_nucast_pkts': '22468751',
                                                                       'in_octets': '9090163230280',
                                                                       'in_ucast_pkts': '6749203937',
                                                                       'links': [       {       'in_nucast_pkts': '22468751',
                                                                                                'in_octets': '9090163230280',
                                                                                                'in_ucast_pkts': '6749203937',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '769216',
                                                                                                'out_octets': '565564191804',
                                                                                                'out_ucast_pkts': '1996567516',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '769216',
                                                                       'out_octets': '565564191804',
                                                                       'out_ucast_pkts': '1996567516',
                                                                       'poller_hash': 'MzhGRjU4cGVlcmludGZwb24gMTUwMTMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150131',
                                                                                               'links': [       {       'connected_time': '551973',
                                                                                                                        'fiber_length': 20054,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzhGRjU4cGVlcnBvbnBvbiAxNTAxMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2846 LINE 45 (WALSH VICTORIA)',
                                         'poller_hash': 'MzhGRjU4',
                                         'serial_number': '38FF58',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:cc:10:fc',
                                         'uptime': '551973'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '150331/g1',
                                                                       'name': '150331/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzVFOTJFcGVlcmludGYxNTAzMzEvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '150331',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzVFOTJFcGVlcmludGZwb24gMTUwMzMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150331',
                                                                                               'links': [       {       'connected_time': '67856432',
                                                                                                                        'fiber_length': 25097,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzVFOTJFcGVlcnBvbnBvbiAxNTAzMzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1000 MAPLE LEAF ST (TOWNSHIP OF '
                                                 'WELLESLEY)',
                                         'poller_hash': 'MzVFOTJF',
                                         'serial_number': '35E92E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:7e:08:aa',
                                         'uptime': '67856432'},
                                 {       'counter_source': '150528/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:62:ec:3b:6c:57'],
                                                                       'id': '150528/g1',
                                                                       'in_nucast_pkts': '418144',
                                                                       'in_octets': '852766277944',
                                                                       'in_ucast_pkts': '935173457',
                                                                       'mac_address': '44:65:7f:7b:c4:a3',
                                                                       'mtu': 2000,
                                                                       'name': '150528/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6900840',
                                                                       'out_octets': '2214694334799',
                                                                       'out_ucast_pkts': '1856618422',
                                                                       'poller_hash': 'NUE2NjdDcGVlcmludGYxNTA1MjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '150528',
                                                                       'in_nucast_pkts': '6900840',
                                                                       'in_octets': '2214694334799',
                                                                       'in_ucast_pkts': '1856618422',
                                                                       'links': [       {       'in_nucast_pkts': '6900840',
                                                                                                'in_octets': '2214694334799',
                                                                                                'in_ucast_pkts': '1856618422',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '418144',
                                                                                                'out_octets': '852766277944',
                                                                                                'out_ucast_pkts': '935173457',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '418144',
                                                                       'out_octets': '852766277944',
                                                                       'out_ucast_pkts': '935173457',
                                                                       'poller_hash': 'NUE2NjdDcGVlcmludGZwb24gMTUwNTI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150528',
                                                                                               'links': [       {       'connected_time': '5921117',
                                                                                                                        'fiber_length': 17857,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NUE2NjdDcGVlcnBvbnBvbiAxNTA1Mjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '2076 WENDELL CT (BOGART SAM)',
                                         'poller_hash': 'NUE2NjdD',
                                         'serial_number': '5A667C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '44:65:7f:7b:c4:a3',
                                         'uptime': '5921117'},
                                 {       'counter_source': '150788/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:82:0a:8a'],
                                                                       'id': '150788/g1',
                                                                       'in_nucast_pkts': '20374',
                                                                       'in_octets': '4990193781',
                                                                       'in_ucast_pkts': '15970766',
                                                                       'mac_address': '60:db:98:73:9c:47',
                                                                       'mtu': 2000,
                                                                       'name': '150788/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1280615',
                                                                       'out_octets': '62273168872',
                                                                       'out_ucast_pkts': '52091688',
                                                                       'poller_hash': 'QjBBMjlBcGVlcmludGYxNTA3ODgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '150788',
                                                                       'in_nucast_pkts': '1280615',
                                                                       'in_octets': '62273168872',
                                                                       'in_ucast_pkts': '52091688',
                                                                       'links': [       {       'in_nucast_pkts': '1280615',
                                                                                                'in_octets': '62273168872',
                                                                                                'in_ucast_pkts': '52091688',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20374',
                                                                                                'out_octets': '4990193781',
                                                                                                'out_ucast_pkts': '15970766',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20374',
                                                                       'out_octets': '4990193781',
                                                                       'out_ucast_pkts': '15970766',
                                                                       'poller_hash': 'QjBBMjlBcGVlcmludGZwb24gMTUwNzg4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '150788',
                                                                                               'links': [       {       'connected_time': '1095970',
                                                                                                                        'fiber_length': 31114,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMjlBcGVlcnBvbnBvbiAxNTA3ODg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1825 MOSER-YOUNG RD (DOERING '
                                                 'CHARLES)',
                                         'poller_hash': 'QjBBMjlB',
                                         'serial_number': 'B0A29A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:9c:44',
                                         'uptime': '1095970'},
                                 {       'counter_source': '152299/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '10:27:f5:a8:23:15'],
                                                                       'id': '152299/g1',
                                                                       'in_nucast_pkts': '2832',
                                                                       'in_octets': '242685670473',
                                                                       'in_ucast_pkts': '1061064055',
                                                                       'mac_address': '60:db:98:8f:01:4e',
                                                                       'mtu': 2000,
                                                                       'name': '152299/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9977878',
                                                                       'out_octets': '2453181458139',
                                                                       'out_ucast_pkts': '2119964902',
                                                                       'poller_hash': 'QjUxMURFcGVlcmludGYxNTIyOTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '152299',
                                                                       'in_nucast_pkts': '9977878',
                                                                       'in_octets': '2453181458139',
                                                                       'in_ucast_pkts': '2119964902',
                                                                       'links': [       {       'in_nucast_pkts': '9977878',
                                                                                                'in_octets': '2453181458139',
                                                                                                'in_ucast_pkts': '2119964902',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2832',
                                                                                                'out_octets': '242685670473',
                                                                                                'out_ucast_pkts': '1061064055',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2832',
                                                                       'out_octets': '242685670473',
                                                                       'out_ucast_pkts': '1061064055',
                                                                       'poller_hash': 'QjUxMURFcGVlcmludGZwb24gMTUyMjk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '152299',
                                                                                               'links': [       {       'connected_time': '8621422',
                                                                                                                        'fiber_length': 35989,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMURFcGVlcnBvbnBvbiAxNTIyOTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '2832 NOTRE DAME DR (LAM EDMUND)',
                                         'poller_hash': 'QjUxMURF',
                                         'serial_number': 'B511DE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:01:4e',
                                         'uptime': '8621422'},
                                 {       'counter_source': '152363/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1a:e6'],
                                                                       'id': '152363/g1',
                                                                       'in_nucast_pkts': '23817',
                                                                       'in_octets': '17439765072',
                                                                       'in_ucast_pkts': '42308720',
                                                                       'mac_address': '60:db:98:73:9f:47',
                                                                       'mtu': 2000,
                                                                       'name': '152363/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1284694',
                                                                       'out_octets': '172830860635',
                                                                       'out_ucast_pkts': '134723164',
                                                                       'poller_hash': 'QjBBMkZBcGVlcmludGYxNTIzNjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '152363',
                                                                       'in_nucast_pkts': '1284694',
                                                                       'in_octets': '172830860635',
                                                                       'in_ucast_pkts': '134723164',
                                                                       'links': [       {       'in_nucast_pkts': '1284694',
                                                                                                'in_octets': '172830860635',
                                                                                                'in_ucast_pkts': '134723164',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '23817',
                                                                                                'out_octets': '17439765072',
                                                                                                'out_ucast_pkts': '42308720',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '23817',
                                                                       'out_octets': '17439765072',
                                                                       'out_ucast_pkts': '42308720',
                                                                       'poller_hash': 'QjBBMkZBcGVlcmludGZwb24gMTUyMzYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '152363',
                                                                                               'links': [       {       'connected_time': '1096785',
                                                                                                                        'fiber_length': 28584,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBMkZBcGVlcnBvbnBvbiAxNTIzNjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4424 WEIMAR LINE (COBER VALERIE)',
                                         'poller_hash': 'QjBBMkZB',
                                         'serial_number': 'B0A2FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:9f:44',
                                         'uptime': '1096785'},
                                 {       'counter_source': '152391/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e0:b2:60:5f:70:f0'],
                                                                       'id': '152391/g1',
                                                                       'in_nucast_pkts': '15562',
                                                                       'in_octets': '22105695454',
                                                                       'in_ucast_pkts': '60766155',
                                                                       'mac_address': '60:db:98:8f:00:3d',
                                                                       'mtu': 2000,
                                                                       'name': '152391/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17525873',
                                                                       'out_octets': '136655581773',
                                                                       'out_ucast_pkts': '117421896',
                                                                       'poller_hash': 'QjUxMTgzcGVlcmludGYxNTIzOTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '152391',
                                                                       'in_nucast_pkts': '17525873',
                                                                       'in_octets': '136655581773',
                                                                       'in_ucast_pkts': '117421896',
                                                                       'links': [       {       'in_nucast_pkts': '17525873',
                                                                                                'in_octets': '136655581773',
                                                                                                'in_ucast_pkts': '117421896',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15562',
                                                                                                'out_octets': '22105695454',
                                                                                                'out_ucast_pkts': '60766155',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15562',
                                                                       'out_octets': '22105695454',
                                                                       'out_ucast_pkts': '60766155',
                                                                       'poller_hash': 'QjUxMTgzcGVlcmludGZwb24gMTUyMzkx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '152391',
                                                                                               'links': [       {       'connected_time': '14920503',
                                                                                                                        'fiber_length': 29479,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMTgzcGVlcnBvbnBvbiAxNTIzOTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '4525 WEIMAR LINE (WEIMAR '
                                                 'MANUFACTURING INC)',
                                         'poller_hash': 'QjUxMTgz',
                                         'serial_number': 'B51183',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:00:3d',
                                         'uptime': '14920503'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '154246/g1',
                                                                       'name': '154246/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QjUxMUIzcGVlcmludGYxNTQyNDYvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '154246',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'QjUxMUIzcGVlcmludGZwb24gMTU0MjQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '154246',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMUIzcGVlcnBvbnBvbiAxNTQyNDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '1010 DOERING ST (HILL JEFF)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "154246" '
                                                                  '(location: GPON '
                                                                  'port "1/2"), "ONT '
                                                                  'Dying Gasp"',
                                         'offline_time': '2025-07-29T22:37:35.920Z',
                                         'poller_hash': 'QjUxMUIz',
                                         'serial_number': 'B511B3',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '154555/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:aa'],
                                                                       'id': '154555/g1',
                                                                       'in_nucast_pkts': '24982',
                                                                       'in_octets': '5027139070',
                                                                       'in_ucast_pkts': '14402773',
                                                                       'mac_address': '60:db:98:8f:00:6a',
                                                                       'mtu': 2000,
                                                                       'name': '154555/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1280614',
                                                                       'out_octets': '48616884258',
                                                                       'out_ucast_pkts': '39525506',
                                                                       'poller_hash': 'QjUxMTkycGVlcmludGYxNTQ1NTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '154555',
                                                                       'in_nucast_pkts': '1280614',
                                                                       'in_octets': '48616884258',
                                                                       'in_ucast_pkts': '39525506',
                                                                       'links': [       {       'in_nucast_pkts': '1280614',
                                                                                                'in_octets': '48616884258',
                                                                                                'in_ucast_pkts': '39525506',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '24982',
                                                                                                'out_octets': '5027139070',
                                                                                                'out_ucast_pkts': '14402773',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24982',
                                                                       'out_octets': '5027139070',
                                                                       'out_ucast_pkts': '14402773',
                                                                       'poller_hash': 'QjUxMTkycGVlcmludGZwb24gMTU0NTU1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '154555',
                                                                                               'links': [       {       'connected_time': '1095990',
                                                                                                                        'fiber_length': 31175,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMTkycGVlcnBvbnBvbiAxNTQ1NTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1821 MOSER-YOUNG RD (KROETSCH PAUL)',
                                         'poller_hash': 'QjUxMTky',
                                         'serial_number': 'B51192',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:00:6a',
                                         'uptime': '1095990'},
                                 {       'counter_source': '154594/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:38:ad'],
                                                                       'id': '154594/g1',
                                                                       'in_nucast_pkts': '166736',
                                                                       'in_octets': '42995078300',
                                                                       'in_ucast_pkts': '564694248',
                                                                       'mac_address': '60:db:98:8e:fd:e5',
                                                                       'mtu': 2000,
                                                                       'name': '154594/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878154',
                                                                       'out_octets': '4207377821903',
                                                                       'out_ucast_pkts': '2781392054',
                                                                       'poller_hash': 'QjUxMEJCcGVlcmludGYxNTQ1OTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '154594',
                                                                       'in_nucast_pkts': '9878154',
                                                                       'in_octets': '4207377821903',
                                                                       'in_ucast_pkts': '2781392054',
                                                                       'links': [       {       'in_nucast_pkts': '9878154',
                                                                                                'in_octets': '4207377821903',
                                                                                                'in_ucast_pkts': '2781392054',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '166736',
                                                                                                'out_octets': '42995078300',
                                                                                                'out_ucast_pkts': '564694248',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '166736',
                                                                       'out_octets': '42995078300',
                                                                       'out_ucast_pkts': '564694248',
                                                                       'poller_hash': 'QjUxMEJCcGVlcmludGZwb24gMTU0NTk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '154594',
                                                                                               'links': [       {       'connected_time': '8535464',
                                                                                                                        'fiber_length': 22900,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMEJCcGVlcnBvbnBvbiAxNTQ1OTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '1010 DOERING ST (FRIES  ROBERT)',
                                         'poller_hash': 'QjUxMEJC',
                                         'serial_number': 'B510BB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8e:fd:e5',
                                         'uptime': '8535464'},
                                 {       'counter_source': '155011/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '3c:5c:f1:63:9e:c1'],
                                                                       'id': '155011/g1',
                                                                       'in_nucast_pkts': '38069',
                                                                       'in_octets': '17860107366',
                                                                       'in_ucast_pkts': '45377072',
                                                                       'mac_address': '60:db:98:8f:01:75',
                                                                       'mtu': 2000,
                                                                       'name': '155011/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1302894',
                                                                       'out_octets': '270833296415',
                                                                       'out_ucast_pkts': '216094353',
                                                                       'poller_hash': 'QjUxMUVCcGVlcmludGYxNTUwMTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '155011',
                                                                       'in_nucast_pkts': '1302894',
                                                                       'in_octets': '270833296415',
                                                                       'in_ucast_pkts': '216094353',
                                                                       'links': [       {       'in_nucast_pkts': '1302894',
                                                                                                'in_octets': '270833296415',
                                                                                                'in_ucast_pkts': '216094353',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '38069',
                                                                                                'out_octets': '17860107366',
                                                                                                'out_ucast_pkts': '45377072',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38069',
                                                                       'out_octets': '17860107366',
                                                                       'out_ucast_pkts': '45377072',
                                                                       'poller_hash': 'QjUxMUVCcGVlcmludGZwb24gMTU1MDEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '155011',
                                                                                               'links': [       {       'connected_time': '1112447',
                                                                                                                        'fiber_length': 23817,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjUxMUVCcGVlcnBvbnBvbiAxNTUwMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '803G',
                                         'name': '62 GALENA ST (HARFORD MATTHEW)',
                                         'poller_hash': 'QjUxMUVC',
                                         'serial_number': 'B511EB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:8f:01:75',
                                         'uptime': '1112447'},
                                 {       'counter_source': '156466/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:63:6a'],
                                                                       'id': '156466/g1',
                                                                       'in_nucast_pkts': '18634264',
                                                                       'in_octets': '473558169032',
                                                                       'in_ucast_pkts': '3940663650',
                                                                       'mac_address': '84:d3:43:7e:1e:ba',
                                                                       'mtu': 2000,
                                                                       'name': '156466/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49972067448',
                                                                       'out_octets': '81240553851644',
                                                                       'out_ucast_pkts': '9605975034',
                                                                       'poller_hash': 'QzgzOTQwcGVlcmludGYxNTY0NjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '156466',
                                                                       'in_nucast_pkts': '49972067448',
                                                                       'in_octets': '81240553851644',
                                                                       'in_ucast_pkts': '9605975034',
                                                                       'links': [       {       'in_nucast_pkts': '49972067448',
                                                                                                'in_octets': '81240553851644',
                                                                                                'in_ucast_pkts': '9605975034',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '18634264',
                                                                                                'out_octets': '473558169032',
                                                                                                'out_ucast_pkts': '3940663650',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18634264',
                                                                       'out_octets': '473558169032',
                                                                       'out_ucast_pkts': '3940663650',
                                                                       'poller_hash': 'QzgzOTQwcGVlcmludGZwb24gMTU2NDY2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156466',
                                                                                               'links': [       {       'connected_time': '21886200',
                                                                                                                        'fiber_length': 31284,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzOTQwcGVlcnBvbnBvbiAxNTY0NjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4147 WEIMAR LINE (KREUTZKAMP JOHN)',
                                         'poller_hash': 'QzgzOTQw',
                                         'serial_number': 'C83940',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:1e:b7',
                                         'uptime': '21886200'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '156535/g1',
                                                                       'name': '156535/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDJDRTkzcGVlcmludGYxNTY1MzUvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '156535',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'NDJDRTkzcGVlcmludGZwb24gMTU2NTM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156535',
                                                                                               'links': [       {       'connected_time': '6026859',
                                                                                                                        'fiber_length': 24517,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJDRTkzcGVlcnBvbnBvbiAxNTY1MzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '32 PARKVIEW DR (GOLDRICK TINA)',
                                         'poller_hash': 'NDJDRTkz',
                                         'serial_number': '42CE93',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:e9:da',
                                         'uptime': '6026859'},
                                 {       'counter_source': '156746/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '156746/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '473900533034',
                                                                       'in_ucast_pkts': '1125373917',
                                                                       'mac_address': '84:d3:43:7d:4b:07',
                                                                       'name': '156746/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '133',
                                                                       'out_octets': '6266694550369',
                                                                       'out_ucast_pkts': '4505661089',
                                                                       'poller_hash': 'QzgxRUNBcGVlcmludGYxNTY3NDYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '156746',
                                                                       'in_nucast_pkts': '133',
                                                                       'in_octets': '6266694550369',
                                                                       'in_ucast_pkts': '4505661089',
                                                                       'links': [       {       'in_nucast_pkts': '133',
                                                                                                'in_octets': '6266694550369',
                                                                                                'in_ucast_pkts': '4505661089',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '473900533034',
                                                                                                'out_ucast_pkts': '1125373917',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '473900533034',
                                                                       'out_ucast_pkts': '1125373917',
                                                                       'poller_hash': 'QzgxRUNBcGVlcmludGZwb24gMTU2NzQ2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '156746',
                                                                                               'links': [       {       'connected_time': '4916795',
                                                                                                                        'fiber_length': 23751,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUNBcGVlcnBvbnBvbiAxNTY3NDY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3685 NAFZIGER RD (KUEPFER SHANE)',
                                         'poller_hash': 'QzgxRUNB',
                                         'serial_number': 'C81ECA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4b:07',
                                         'uptime': '4916795'},
                                 {       'counter_source': '157294/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b7:fa'],
                                                                       'id': '157294/g1',
                                                                       'in_nucast_pkts': '9760006',
                                                                       'in_octets': '221295763231',
                                                                       'in_ucast_pkts': '587896456',
                                                                       'mac_address': '84:d3:43:7c:98:8a',
                                                                       'mtu': 2000,
                                                                       'name': '157294/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878306',
                                                                       'out_octets': '3250760851570',
                                                                       'out_ucast_pkts': '2392140224',
                                                                       'poller_hash': 'QzgwODdBcGVlcmludGYxNTcyOTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157294',
                                                                       'in_nucast_pkts': '9878306',
                                                                       'in_octets': '3250760851570',
                                                                       'in_ucast_pkts': '2392140224',
                                                                       'links': [       {       'in_nucast_pkts': '9878306',
                                                                                                'in_octets': '3250760851570',
                                                                                                'in_ucast_pkts': '2392140224',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9760006',
                                                                                                'out_octets': '221295763231',
                                                                                                'out_ucast_pkts': '587896456',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9760006',
                                                                       'out_octets': '221295763231',
                                                                       'out_ucast_pkts': '587896456',
                                                                       'poller_hash': 'QzgwODdBcGVlcmludGZwb24gMTU3Mjk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157294',
                                                                                               'links': [       {       'connected_time': '8535197',
                                                                                                                        'fiber_length': 25019,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgwODdBcGVlcnBvbnBvbiAxNTcyOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 PARKVIEW DR (FRANKLIN PAUL)',
                                         'poller_hash': 'QzgwODdB',
                                         'serial_number': 'C8087A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7c:98:87',
                                         'uptime': '8535197'},
                                 {       'counter_source': '157304/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '157304/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3011182141924',
                                                                       'in_ucast_pkts': '698784968',
                                                                       'mac_address': '84:d3:43:7e:1c:d7',
                                                                       'name': '157304/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '37',
                                                                       'out_octets': '22469748534087',
                                                                       'out_ucast_pkts': '18118128684',
                                                                       'poller_hash': 'QzgzOTA0cGVlcmludGYxNTczMDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '157304',
                                                                       'in_nucast_pkts': '37',
                                                                       'in_octets': '22469748534087',
                                                                       'in_ucast_pkts': '18118128684',
                                                                       'links': [       {       'in_nucast_pkts': '37',
                                                                                                'in_octets': '22469748534087',
                                                                                                'in_ucast_pkts': '18118128684',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3011182141924',
                                                                                                'out_ucast_pkts': '698784968',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3011182141924',
                                                                       'out_ucast_pkts': '698784968',
                                                                       'poller_hash': 'QzgzOTA0cGVlcmludGZwb24gMTU3MzA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157304',
                                                                                               'links': [       {       'connected_time': '7316566',
                                                                                                                        'fiber_length': 23011,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzOTA0cGVlcnBvbnBvbiAxNTczMDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '196 FERRIS DR (FIFE PAUL)',
                                         'poller_hash': 'QzgzOTA0',
                                         'serial_number': 'C83904',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:1c:d7',
                                         'uptime': '7316566'},
                                 {       'counter_source': '157314/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ed:21'],
                                                                       'id': '157314/g1',
                                                                       'in_nucast_pkts': '4143805',
                                                                       'in_octets': '166925286126',
                                                                       'in_ucast_pkts': '716151228',
                                                                       'mac_address': '84:d3:43:7e:44:72',
                                                                       'mtu': 2000,
                                                                       'name': '157314/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10115556677',
                                                                       'out_octets': '16352884609525',
                                                                       'out_ucast_pkts': '2056788808',
                                                                       'poller_hash': 'QzgzREY3cGVlcmludGYxNTczMTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157314',
                                                                       'in_nucast_pkts': '10115556677',
                                                                       'in_octets': '16352884609525',
                                                                       'in_ucast_pkts': '2056788808',
                                                                       'links': [       {       'in_nucast_pkts': '10115556677',
                                                                                                'in_octets': '16352884609525',
                                                                                                'in_ucast_pkts': '2056788808',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4143805',
                                                                                                'out_octets': '166925286126',
                                                                                                'out_ucast_pkts': '716151228',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4143805',
                                                                       'out_octets': '166925286126',
                                                                       'out_ucast_pkts': '716151228',
                                                                       'poller_hash': 'QzgzREY3cGVlcmludGZwb24gMTU3MzE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157314',
                                                                                               'links': [       {       'connected_time': '8535231',
                                                                                                                        'fiber_length': 23223,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzREY3cGVlcnBvbnBvbiAxNTczMTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '15 ZINKANN CR (MASSEL SHELLY)',
                                         'poller_hash': 'QzgzREY3',
                                         'serial_number': 'C83DF7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:44:6f',
                                         'uptime': '8535231'},
                                 {       'counter_source': '157324/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '157324/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '685976031333',
                                                                       'in_ucast_pkts': '45990137827',
                                                                       'mac_address': '84:d3:43:7d:44:2f',
                                                                       'name': '157324/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '76',
                                                                       'out_octets': '14068443686627',
                                                                       'out_ucast_pkts': '10323879679',
                                                                       'poller_hash': 'QzgxREVGcGVlcmludGYxNTczMjQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '157324',
                                                                       'in_nucast_pkts': '76',
                                                                       'in_octets': '14068443686627',
                                                                       'in_ucast_pkts': '10323879679',
                                                                       'links': [       {       'in_nucast_pkts': '76',
                                                                                                'in_octets': '14068443686627',
                                                                                                'in_ucast_pkts': '10323879679',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '685976031333',
                                                                                                'out_ucast_pkts': '45990137827',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '685976031333',
                                                                       'out_ucast_pkts': '45990137827',
                                                                       'poller_hash': 'QzgxREVGcGVlcmludGZwb24gMTU3MzI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157324',
                                                                                               'links': [       {       'connected_time': '7195080',
                                                                                                                        'fiber_length': 25074,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREVGcGVlcnBvbnBvbiAxNTczMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 SHADYWOOD CT (HARKINS RYAN)',
                                         'poller_hash': 'QzgxREVG',
                                         'serial_number': 'C81DEF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:2f',
                                         'uptime': '7195080'},
                                 {       'counter_source': '157384/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:62:20'],
                                                                       'id': '157384/g1',
                                                                       'in_nucast_pkts': '153752',
                                                                       'in_octets': '192881093701',
                                                                       'in_ucast_pkts': '438639543',
                                                                       'mac_address': '84:d3:43:7d:10:02',
                                                                       'mtu': 2000,
                                                                       'name': '157384/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9795639',
                                                                       'out_octets': '2116401877458',
                                                                       'out_ucast_pkts': '1561024627',
                                                                       'poller_hash': 'QzgxNzY5cGVlcmludGYxNTczODQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157384',
                                                                       'in_nucast_pkts': '9795639',
                                                                       'in_octets': '2116401877458',
                                                                       'in_ucast_pkts': '1561024627',
                                                                       'links': [       {       'in_nucast_pkts': '9795639',
                                                                                                'in_octets': '2116401877458',
                                                                                                'in_ucast_pkts': '1561024627',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '153752',
                                                                                                'out_octets': '192881093701',
                                                                                                'out_ucast_pkts': '438639543',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '153752',
                                                                       'out_octets': '192881093701',
                                                                       'out_ucast_pkts': '438639543',
                                                                       'poller_hash': 'QzgxNzY5cGVlcmludGZwb24gMTU3Mzg0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157384',
                                                                                               'links': [       {       'connected_time': '8535271',
                                                                                                                        'fiber_length': 24838,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxNzY5cGVlcnBvbnBvbiAxNTczODQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1140 QUEEN'S BUSH RD (Chad "
                                                 'KIESWETTER 10438397)',
                                         'poller_hash': 'QzgxNzY5',
                                         'serial_number': 'C81769',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:0f:ff',
                                         'uptime': '8535271'},
                                 {       'counter_source': '157394/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:63:a1'],
                                                                       'id': '157394/g1',
                                                                       'in_nucast_pkts': '160383',
                                                                       'in_octets': '23963681276',
                                                                       'in_ucast_pkts': '116720490',
                                                                       'mac_address': '84:d3:43:7d:47:82',
                                                                       'mtu': 2000,
                                                                       'name': '157394/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797984',
                                                                       'out_octets': '658546344775',
                                                                       'out_ucast_pkts': '490824684',
                                                                       'poller_hash': 'QzgxRTU5cGVlcmludGYxNTczOTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157394',
                                                                       'in_nucast_pkts': '9797984',
                                                                       'in_octets': '658546344775',
                                                                       'in_ucast_pkts': '490824684',
                                                                       'links': [       {       'in_nucast_pkts': '9797984',
                                                                                                'in_octets': '658546344775',
                                                                                                'in_ucast_pkts': '490824684',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160383',
                                                                                                'out_octets': '23963681276',
                                                                                                'out_ucast_pkts': '116720490',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160383',
                                                                       'out_octets': '23963681276',
                                                                       'out_ucast_pkts': '116720490',
                                                                       'poller_hash': 'QzgxRTU5cGVlcmludGZwb24gMTU3Mzk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157394',
                                                                                               'links': [       {       'connected_time': '8535273',
                                                                                                                        'fiber_length': 24982,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTU5cGVlcnBvbnBvbiAxNTczOTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '92 BROWN ST (GERBER CONNOR)',
                                         'poller_hash': 'QzgxRTU5',
                                         'serial_number': 'C81E59',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:7f',
                                         'uptime': '8535273'},
                                 {       'counter_source': '157404/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:bb:f9'],
                                                                       'id': '157404/g1',
                                                                       'in_nucast_pkts': '151120',
                                                                       'in_octets': '59315248009',
                                                                       'in_ucast_pkts': '373671634',
                                                                       'mac_address': '84:d3:43:7d:0e:7a',
                                                                       'mtu': 2000,
                                                                       'name': '157404/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797738',
                                                                       'out_octets': '1276467665414',
                                                                       'out_ucast_pkts': '979105777',
                                                                       'poller_hash': 'QzgxNzM4cGVlcmludGYxNTc0MDQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157404',
                                                                       'in_nucast_pkts': '9797738',
                                                                       'in_octets': '1276467665414',
                                                                       'in_ucast_pkts': '979105777',
                                                                       'links': [       {       'in_nucast_pkts': '9797738',
                                                                                                'in_octets': '1276467665414',
                                                                                                'in_ucast_pkts': '979105777',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '151120',
                                                                                                'out_octets': '59315248009',
                                                                                                'out_ucast_pkts': '373671634',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '151120',
                                                                       'out_octets': '59315248009',
                                                                       'out_ucast_pkts': '373671634',
                                                                       'poller_hash': 'QzgxNzM4cGVlcmludGZwb24gMTU3NDA0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157404',
                                                                                               'links': [       {       'connected_time': '8535180',
                                                                                                                        'fiber_length': 24879,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxNzM4cGVlcnBvbnBvbiAxNTc0MDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1147 QUEEN'S BUSH RD (POOLE MATTHEW)",
                                         'poller_hash': 'QzgxNzM4',
                                         'serial_number': 'C81738',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:0e:77',
                                         'uptime': '8535180'},
                                 {       'counter_source': '157414/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:bb:28'],
                                                                       'id': '157414/g1',
                                                                       'in_nucast_pkts': '149868',
                                                                       'in_octets': '50754844806',
                                                                       'in_ucast_pkts': '256780881',
                                                                       'mac_address': '84:d3:43:7d:3f:ea',
                                                                       'mtu': 2000,
                                                                       'name': '157414/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9925744',
                                                                       'out_octets': '1234030289975',
                                                                       'out_ucast_pkts': '970362070',
                                                                       'poller_hash': 'QzgxRDY2cGVlcmludGYxNTc0MTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157414',
                                                                       'in_nucast_pkts': '9925744',
                                                                       'in_octets': '1234030289975',
                                                                       'in_ucast_pkts': '970362070',
                                                                       'links': [       {       'in_nucast_pkts': '9925744',
                                                                                                'in_octets': '1234030289975',
                                                                                                'in_ucast_pkts': '970362070',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149868',
                                                                                                'out_octets': '50754844806',
                                                                                                'out_ucast_pkts': '256780881',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149868',
                                                                       'out_octets': '50754844806',
                                                                       'out_ucast_pkts': '256780881',
                                                                       'poller_hash': 'QzgxRDY2cGVlcmludGZwb24gMTU3NDE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157414',
                                                                                               'links': [       {       'connected_time': '8535138',
                                                                                                                        'fiber_length': 23549,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDY2cGVlcnBvbnBvbiAxNTc0MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '74 SCHWEITZER CR (ENNS JAMIE)',
                                         'poller_hash': 'QzgxRDY2',
                                         'serial_number': 'C81D66',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:3f:e7',
                                         'uptime': '8535138'},
                                 {       'counter_source': '157424/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b6:a5'],
                                                                       'id': '157424/g1',
                                                                       'in_nucast_pkts': '963199',
                                                                       'in_octets': '407037654462',
                                                                       'in_ucast_pkts': '2502038988',
                                                                       'mac_address': '84:d3:43:7d:40:3a',
                                                                       'mtu': 2000,
                                                                       'name': '157424/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63477020',
                                                                       'out_octets': '12620894935242',
                                                                       'out_ucast_pkts': '9247849619',
                                                                       'poller_hash': 'QzgxRDcwcGVlcmludGYxNTc0MjQvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '157424',
                                                                       'in_nucast_pkts': '63477020',
                                                                       'in_octets': '12620894935242',
                                                                       'in_ucast_pkts': '9247849619',
                                                                       'links': [       {       'in_nucast_pkts': '63477020',
                                                                                                'in_octets': '12620894935242',
                                                                                                'in_ucast_pkts': '9247849619',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '963199',
                                                                                                'out_octets': '407037654462',
                                                                                                'out_ucast_pkts': '2502038988',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '963199',
                                                                       'out_octets': '407037654462',
                                                                       'out_ucast_pkts': '2502038988',
                                                                       'poller_hash': 'QzgxRDcwcGVlcmludGZwb24gMTU3NDI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157424',
                                                                                               'links': [       {       'connected_time': '21886272',
                                                                                                                        'fiber_length': 23066,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDcwcGVlcnBvbnBvbiAxNTc0MjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '206 FERRIS DR (KUBE AMY & CHRIS)',
                                         'poller_hash': 'QzgxRDcw',
                                         'serial_number': 'C81D70',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:40:37',
                                         'uptime': '21886272'},
                                 {       'counter_source': '157531/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '157531/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '488392480880',
                                                                       'in_ucast_pkts': '1288916811',
                                                                       'mac_address': '84:d3:43:7d:40:f7',
                                                                       'name': '157531/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '6882734545868',
                                                                       'out_ucast_pkts': '5078137360',
                                                                       'poller_hash': 'QzgxRDg4cGVlcmludGYxNTc1MzEvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '157531',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '6882734545868',
                                                                       'in_ucast_pkts': '5078137360',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '6882734545868',
                                                                                                'in_ucast_pkts': '5078137360',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '488392480880',
                                                                                                'out_ucast_pkts': '1288916811',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '488392480880',
                                                                       'out_ucast_pkts': '1288916811',
                                                                       'poller_hash': 'QzgxRDg4cGVlcmludGZwb24gMTU3NTMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '157531',
                                                                                               'links': [       {       'connected_time': '8535079',
                                                                                                                        'fiber_length': 24368,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRDg4cGVlcnBvbnBvbiAxNTc1MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '49 PARKVIEW DR (KIENITZ JASVEEN)',
                                         'poller_hash': 'QzgxRDg4',
                                         'serial_number': 'C81D88',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:40:f7',
                                         'uptime': '8535079'},
                                 {       'counter_source': '158545/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1c:f6'],
                                                                       'id': '158545/g1',
                                                                       'in_nucast_pkts': '13158',
                                                                       'in_octets': '4919680136',
                                                                       'in_ucast_pkts': '14234630',
                                                                       'mac_address': '84:d3:43:7d:5a:b2',
                                                                       'mtu': 2000,
                                                                       'name': '158545/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715408',
                                                                       'out_octets': '48196197744',
                                                                       'out_ucast_pkts': '37788154',
                                                                       'poller_hash': 'QzgyMEJGcGVlcmludGYxNTg1NDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '158545',
                                                                       'in_nucast_pkts': '715408',
                                                                       'in_octets': '48196197744',
                                                                       'in_ucast_pkts': '37788154',
                                                                       'links': [       {       'in_nucast_pkts': '715408',
                                                                                                'in_octets': '48196197744',
                                                                                                'in_ucast_pkts': '37788154',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13158',
                                                                                                'out_octets': '4919680136',
                                                                                                'out_ucast_pkts': '14234630',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13158',
                                                                       'out_octets': '4919680136',
                                                                       'out_ucast_pkts': '14234630',
                                                                       'poller_hash': 'QzgyMEJGcGVlcmludGZwb24gMTU4NTQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '158545',
                                                                                               'links': [       {       'connected_time': '611445',
                                                                                                                        'fiber_length': 19794,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgyMEJGcGVlcnBvbnBvbiAxNTg1NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2853 LINE 45 (LECOMTE OLIVIA)',
                                         'poller_hash': 'QzgyMEJG',
                                         'serial_number': 'C820BF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:5a:af',
                                         'uptime': '611445'},
                                 {       'counter_source': '161677/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '161677/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '435847020448',
                                                                       'in_ucast_pkts': '574198419',
                                                                       'mac_address': '84:d3:43:7e:45:67',
                                                                       'name': '161677/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '31',
                                                                       'out_octets': '1205044598618',
                                                                       'out_ucast_pkts': '1191998714',
                                                                       'poller_hash': 'QzgzRTE2cGVlcmludGYxNjE2NzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '161677',
                                                                       'in_nucast_pkts': '31',
                                                                       'in_octets': '1205044598618',
                                                                       'in_ucast_pkts': '1191998714',
                                                                       'links': [       {       'in_nucast_pkts': '31',
                                                                                                'in_octets': '1205044598618',
                                                                                                'in_ucast_pkts': '1191998714',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '435847020448',
                                                                                                'out_ucast_pkts': '574198419',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '435847020448',
                                                                       'out_ucast_pkts': '574198419',
                                                                       'poller_hash': 'QzgzRTE2cGVlcmludGZwb24gMTYxNjc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '161677',
                                                                                               'links': [       {       'connected_time': '107129',
                                                                                                                        'fiber_length': 30331,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzRTE2cGVlcnBvbnBvbiAxNjE2Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4639 WEIMAR LINE (HOOVER EDWIN)',
                                         'poller_hash': 'QzgzRTE2',
                                         'serial_number': 'C83E16',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:45:67',
                                         'uptime': '107129'},
                                 {       'counter_source': '162403/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b6:fd'],
                                                                       'id': '162403/g1',
                                                                       'in_nucast_pkts': '94994',
                                                                       'in_octets': '232618794397',
                                                                       'in_ucast_pkts': '532483494',
                                                                       'mac_address': '84:d3:43:7e:45:3a',
                                                                       'mtu': 2000,
                                                                       'name': '162403/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5860941',
                                                                       'out_octets': '2036967657283',
                                                                       'out_ucast_pkts': '1524228709',
                                                                       'poller_hash': 'QzgzRTEwcGVlcmludGYxNjI0MDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '162403',
                                                                       'in_nucast_pkts': '5860941',
                                                                       'in_octets': '2036967657283',
                                                                       'in_ucast_pkts': '1524228709',
                                                                       'links': [       {       'in_nucast_pkts': '5860941',
                                                                                                'in_octets': '2036967657283',
                                                                                                'in_ucast_pkts': '1524228709',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '94994',
                                                                                                'out_octets': '232618794397',
                                                                                                'out_ucast_pkts': '532483494',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '94994',
                                                                       'out_octets': '232618794397',
                                                                       'out_ucast_pkts': '532483494',
                                                                       'poller_hash': 'QzgzRTEwcGVlcmludGZwb24gMTYyNDAz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '162403',
                                                                                               'links': [       {       'connected_time': '5010038',
                                                                                                                        'fiber_length': 23111,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzRTEwcGVlcnBvbnBvbiAxNjI0MDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '117 LAWRENCE ST (BUTLER STACI)',
                                         'poller_hash': 'QzgzRTEw',
                                         'serial_number': 'C83E10',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:45:37',
                                         'uptime': '5010038'},
                                 {       'counter_source': '162483/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '162483/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '187111430646',
                                                                       'in_ucast_pkts': '1038889776',
                                                                       'mac_address': '84:d3:43:7e:32:bf',
                                                                       'name': '162483/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '20',
                                                                       'out_octets': '3909870071172',
                                                                       'out_ucast_pkts': '2828664552',
                                                                       'poller_hash': 'QzgzQkMxcGVlcmludGYxNjI0ODMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '162483',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '3909870071172',
                                                                       'in_ucast_pkts': '2828664552',
                                                                       'links': [       {       'in_nucast_pkts': '20',
                                                                                                'in_octets': '3909870071172',
                                                                                                'in_ucast_pkts': '2828664552',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '187111430646',
                                                                                                'out_ucast_pkts': '1038889776',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '187111430646',
                                                                       'out_ucast_pkts': '1038889776',
                                                                       'poller_hash': 'QzgzQkMxcGVlcmludGZwb24gMTYyNDgz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '162483',
                                                                                               'links': [       {       'connected_time': '8535141',
                                                                                                                        'fiber_length': 23329,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkMxcGVlcnBvbnBvbiAxNjI0ODM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3832 NAFZIGER RD (ADAM KATRINA)',
                                         'poller_hash': 'QzgzQkMx',
                                         'serial_number': 'C83BC1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:bf',
                                         'uptime': '8535141'},
                                 {       'counter_source': '163643/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e7:d1'],
                                                                       'id': '163643/g1',
                                                                       'in_nucast_pkts': '2953',
                                                                       'in_octets': '2176107877',
                                                                       'in_ucast_pkts': '14202127',
                                                                       'mac_address': '84:d3:43:7e:31:4a',
                                                                       'mtu': 2000,
                                                                       'name': '163643/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '213720',
                                                                       'out_octets': '118992895322',
                                                                       'out_ucast_pkts': '81858407',
                                                                       'poller_hash': 'QzgzQjkycGVlcmludGYxNjM2NDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '163643',
                                                                       'in_nucast_pkts': '213720',
                                                                       'in_octets': '118992895322',
                                                                       'in_ucast_pkts': '81858407',
                                                                       'links': [       {       'in_nucast_pkts': '213720',
                                                                                                'in_octets': '118992895322',
                                                                                                'in_ucast_pkts': '81858407',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2953',
                                                                                                'out_octets': '2176107877',
                                                                                                'out_ucast_pkts': '14202127',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2953',
                                                                       'out_octets': '2176107877',
                                                                       'out_ucast_pkts': '14202127',
                                                                       'poller_hash': 'QzgzQjkycGVlcmludGZwb24gMTYzNjQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '163643',
                                                                                               'links': [       {       'connected_time': '180049',
                                                                                                                        'fiber_length': 17839,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjkycGVlcnBvbnBvbiAxNjM2NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5421 KING ST (NICKEL KIM)',
                                         'poller_hash': 'QzgzQjky',
                                         'serial_number': 'C83B92',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:47',
                                         'uptime': '180049'},
                                 {       'counter_source': '164126/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b8:f7'],
                                                                       'id': '164126/g1',
                                                                       'in_nucast_pkts': '4893891',
                                                                       'in_octets': '1377028156673',
                                                                       'in_ucast_pkts': '2393301385',
                                                                       'mac_address': '84:d3:43:7e:30:02',
                                                                       'mtu': 2000,
                                                                       'name': '164126/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6728195517',
                                                                       'out_octets': '12802565707024',
                                                                       'out_ucast_pkts': '3202435844',
                                                                       'poller_hash': 'QzgzQjY5cGVlcmludGYxNjQxMjYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '164126',
                                                                       'in_nucast_pkts': '6728195517',
                                                                       'in_octets': '12802565707024',
                                                                       'in_ucast_pkts': '3202435844',
                                                                       'links': [       {       'in_nucast_pkts': '6728195517',
                                                                                                'in_octets': '12802565707024',
                                                                                                'in_ucast_pkts': '3202435844',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '4893891',
                                                                                                'out_octets': '1377028156673',
                                                                                                'out_ucast_pkts': '2393301385',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4893891',
                                                                       'out_octets': '1377028156673',
                                                                       'out_ucast_pkts': '2393301385',
                                                                       'poller_hash': 'QzgzQjY5cGVlcmludGZwb24gMTY0MTI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164126',
                                                                                               'links': [       {       'connected_time': '8535231',
                                                                                                                        'fiber_length': 24538,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjY5cGVlcnBvbnBvbiAxNjQxMjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '37 PARKVIEW DR (KORCZYNSKI JENNA)',
                                         'poller_hash': 'QzgzQjY5',
                                         'serial_number': 'C83B69',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:2f:ff',
                                         'uptime': '8535231'},
                                 {       'counter_source': '164167/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f7:6e'],
                                                                       'id': '164167/g1',
                                                                       'in_nucast_pkts': '154503',
                                                                       'in_octets': '43242326409',
                                                                       'in_ucast_pkts': '194951693',
                                                                       'mac_address': '84:d3:43:7e:30:32',
                                                                       'mtu': 2000,
                                                                       'name': '164167/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920016',
                                                                       'out_octets': '2283715735908',
                                                                       'out_ucast_pkts': '1622573352',
                                                                       'poller_hash': 'QzgzQjZGcGVlcmludGYxNjQxNjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '164167',
                                                                       'in_nucast_pkts': '9920016',
                                                                       'in_octets': '2283715735908',
                                                                       'in_ucast_pkts': '1622573352',
                                                                       'links': [       {       'in_nucast_pkts': '9920016',
                                                                                                'in_octets': '2283715735908',
                                                                                                'in_ucast_pkts': '1622573352',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '154503',
                                                                                                'out_octets': '43242326409',
                                                                                                'out_ucast_pkts': '194951693',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '154503',
                                                                       'out_octets': '43242326409',
                                                                       'out_ucast_pkts': '194951693',
                                                                       'poller_hash': 'QzgzQjZGcGVlcmludGZwb24gMTY0MTY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164167',
                                                                                               'links': [       {       'connected_time': '8535335',
                                                                                                                        'fiber_length': 22507,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjZGcGVlcnBvbnBvbiAxNjQxNjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'ont 164167',
                                         'poller_hash': 'QzgzQjZG',
                                         'serial_number': 'C83B6F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:30:2f',
                                         'uptime': '8535335'},
                                 {       'counter_source': '164256/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:fa:5d'],
                                                                       'id': '164256/g1',
                                                                       'in_nucast_pkts': '948815',
                                                                       'in_octets': '49910692484',
                                                                       'in_ucast_pkts': '260000959',
                                                                       'mac_address': '84:d3:43:7e:32:aa',
                                                                       'mtu': 2000,
                                                                       'name': '164256/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '61546449',
                                                                       'out_octets': '1567682492530',
                                                                       'out_ucast_pkts': '1286149536',
                                                                       'poller_hash': 'QzgzQkJFcGVlcmludGYxNjQyNTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '164256',
                                                                       'in_nucast_pkts': '61546449',
                                                                       'in_octets': '1567682492530',
                                                                       'in_ucast_pkts': '1286149536',
                                                                       'links': [       {       'in_nucast_pkts': '61546449',
                                                                                                'in_octets': '1567682492530',
                                                                                                'in_ucast_pkts': '1286149536',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '948815',
                                                                                                'out_octets': '49910692484',
                                                                                                'out_ucast_pkts': '260000959',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '948815',
                                                                       'out_octets': '49910692484',
                                                                       'out_ucast_pkts': '260000959',
                                                                       'poller_hash': 'QzgzQkJFcGVlcmludGZwb24gMTY0MjU2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164256',
                                                                                               'links': [       {       'connected_time': '18323160',
                                                                                                                        'fiber_length': 22984,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkJFcGVlcnBvbnBvbiAxNjQyNTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '11 POND VIEW DR (LICHTY SAM S)',
                                         'poller_hash': 'QzgzQkJF',
                                         'serial_number': 'C83BBE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:a7',
                                         'uptime': '18323160'},
                                 {       'counter_source': '164715/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:2b:fb'],
                                                                       'id': '164715/g1',
                                                                       'in_nucast_pkts': '530849',
                                                                       'in_octets': '5504365279030',
                                                                       'in_ucast_pkts': '17045921995',
                                                                       'mac_address': 'b8:94:70:1f:09:f5',
                                                                       'mtu': 2000,
                                                                       'name': '164715/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '18782667',
                                                                       'out_octets': '53575071730157',
                                                                       'out_ucast_pkts': '42205412109',
                                                                       'poller_hash': 'RkQ4NzNBcGVlcmludGYxNjQ3MTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '164715',
                                                                       'in_nucast_pkts': '18782667',
                                                                       'in_octets': '53575071730157',
                                                                       'in_ucast_pkts': '42205412109',
                                                                       'links': [       {       'in_nucast_pkts': '18782667',
                                                                                                'in_octets': '53575071730157',
                                                                                                'in_ucast_pkts': '42205412109',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '530849',
                                                                                                'out_octets': '5504365279030',
                                                                                                'out_ucast_pkts': '17045921995',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '530849',
                                                                       'out_octets': '5504365279030',
                                                                       'out_ucast_pkts': '17045921995',
                                                                       'poller_hash': 'RkQ4NzNBcGVlcmludGZwb24gMTY0NzE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164715',
                                                                                               'links': [       {       'connected_time': '30163838',
                                                                                                                        'fiber_length': 32136,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.138,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RkQ4NzNBcGVlcnBvbnBvbiAxNjQ3MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE-I',
                                         'name': '3937 WEIMAR LINE (BORONKA PETER)',
                                         'poller_hash': 'RkQ4NzNB',
                                         'serial_number': 'FD873A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'b8:94:70:1f:09:f1',
                                         'uptime': '30163838'},
                                 {       'counter_source': '164795/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f0:2f:74:10:95:c0'],
                                                                       'id': '164795/g1',
                                                                       'in_nucast_pkts': '5',
                                                                       'in_octets': '808811528',
                                                                       'in_ucast_pkts': '4282441',
                                                                       'mac_address': '84:d3:43:7e:32:da',
                                                                       'mtu': 2000,
                                                                       'name': '164795/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1207500',
                                                                       'out_octets': '1361573416',
                                                                       'out_ucast_pkts': '4206468',
                                                                       'poller_hash': 'QzgzQkM0cGVlcmludGYxNjQ3OTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '164795',
                                                                       'in_nucast_pkts': '1207500',
                                                                       'in_octets': '1361573416',
                                                                       'in_ucast_pkts': '4206468',
                                                                       'links': [       {       'in_nucast_pkts': '1207500',
                                                                                                'in_octets': '1361573416',
                                                                                                'in_ucast_pkts': '4206468',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5',
                                                                                                'out_octets': '808811528',
                                                                                                'out_ucast_pkts': '4282441',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5',
                                                                       'out_octets': '808811528',
                                                                       'out_ucast_pkts': '4282441',
                                                                       'poller_hash': 'QzgzQkM0cGVlcmludGZwb24gMTY0Nzk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164795',
                                                                                               'links': [       {       'connected_time': '1033095',
                                                                                                                        'fiber_length': 21660,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkM0cGVlcnBvbnBvbiAxNjQ3OTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1287 HUTCHISON RD (LEIS FEED & '
                                                 'SUPPLY LTD)',
                                         'poller_hash': 'QzgzQkM0',
                                         'serial_number': 'C83BC4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:32:d7',
                                         'uptime': '1033095'},
                                 {       'counter_source': '164825/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:f9:c3'],
                                                                       'id': '164825/g1',
                                                                       'in_nucast_pkts': '160136',
                                                                       'in_octets': '72612870151',
                                                                       'in_ucast_pkts': '150723969',
                                                                       'mac_address': '84:d3:43:95:8e:dd',
                                                                       'mtu': 2000,
                                                                       'name': '164825/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797820',
                                                                       'out_octets': '465348975786',
                                                                       'out_ucast_pkts': '416091324',
                                                                       'poller_hash': 'Q0JEMUZBcGVlcmludGYxNjQ4MjUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '164825',
                                                                       'in_nucast_pkts': '9797820',
                                                                       'in_octets': '465348975786',
                                                                       'in_ucast_pkts': '416091324',
                                                                       'links': [       {       'in_nucast_pkts': '9797820',
                                                                                                'in_octets': '465348975786',
                                                                                                'in_ucast_pkts': '416091324',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160136',
                                                                                                'out_octets': '72612870151',
                                                                                                'out_ucast_pkts': '150723969',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160136',
                                                                       'out_octets': '72612870151',
                                                                       'out_ucast_pkts': '150723969',
                                                                       'poller_hash': 'Q0JEMUZBcGVlcmludGZwb24gMTY0ODI1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '164825',
                                                                                               'links': [       {       'connected_time': '8535213',
                                                                                                                        'fiber_length': 24146,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUZBcGVlcnBvbnBvbiAxNjQ4MjU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1031 QUEEN'S BUSH RD (PITT BRENDA)",
                                         'poller_hash': 'Q0JEMUZB',
                                         'serial_number': 'CBD1FA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8e:da',
                                         'uptime': '8535213'},
                                 {       'counter_source': '165418/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '165418/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '153342946876',
                                                                       'in_ucast_pkts': '439091366',
                                                                       'mac_address': '84:d3:43:95:8c:7a',
                                                                       'name': '165418/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '2073304076161',
                                                                       'out_ucast_pkts': '1588103991',
                                                                       'poller_hash': 'Q0JEMUFFcGVlcmludGYxNjU0MTgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '165418',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '2073304076161',
                                                                       'in_ucast_pkts': '1588103991',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '2073304076161',
                                                                                                'in_ucast_pkts': '1588103991',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '153342946876',
                                                                                                'out_ucast_pkts': '439091366',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '153342946876',
                                                                       'out_ucast_pkts': '439091366',
                                                                       'poller_hash': 'Q0JEMUFFcGVlcmludGZwb24gMTY1NDE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '165418',
                                                                                               'links': [       {       'connected_time': '1096783',
                                                                                                                        'fiber_length': 29575,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUFFcGVlcnBvbnBvbiAxNjU0MTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4225 WEIMAR LINE (MARTIN MARY)',
                                         'poller_hash': 'Q0JEMUFF',
                                         'serial_number': 'CBD1AE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8c:7a',
                                         'uptime': '1096783'},
                                 {       'counter_source': '166328/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b7:60'],
                                                                       'id': '166328/g1',
                                                                       'in_nucast_pkts': '9781406',
                                                                       'in_octets': '820577320142',
                                                                       'in_ucast_pkts': '1441284698',
                                                                       'mac_address': '84:d3:43:95:90:ed',
                                                                       'mtu': 2000,
                                                                       'name': '166328/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797819',
                                                                       'out_octets': '2649659498047',
                                                                       'out_ucast_pkts': '2331225686',
                                                                       'poller_hash': 'Q0JEMjNDcGVlcmludGYxNjYzMjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '166328',
                                                                       'in_nucast_pkts': '9797819',
                                                                       'in_octets': '2649659498047',
                                                                       'in_ucast_pkts': '2331225686',
                                                                       'links': [       {       'in_nucast_pkts': '9797819',
                                                                                                'in_octets': '2649659498047',
                                                                                                'in_ucast_pkts': '2331225686',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9781406',
                                                                                                'out_octets': '820577320142',
                                                                                                'out_ucast_pkts': '1441284698',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9781406',
                                                                       'out_octets': '820577320142',
                                                                       'out_ucast_pkts': '1441284698',
                                                                       'poller_hash': 'Q0JEMjNDcGVlcmludGZwb24gMTY2MzI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166328',
                                                                                               'links': [       {       'connected_time': '8535159',
                                                                                                                        'fiber_length': 23728,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjNDcGVlcnBvbnBvbiAxNjYzMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '64 REINER CR (GOREL SHANTELLE)',
                                         'poller_hash': 'Q0JEMjND',
                                         'serial_number': 'CBD23C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:90:ea',
                                         'uptime': '8535159'},
                                 {       'counter_source': '166438/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '166438/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '182807811332',
                                                                       'in_ucast_pkts': '908092578',
                                                                       'mac_address': '84:d3:43:b3:79:6d',
                                                                       'name': '166438/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '107',
                                                                       'out_octets': '4537439198505',
                                                                       'out_ucast_pkts': '7500790876',
                                                                       'poller_hash': 'RENGNDQwcGVlcmludGYxNjY0MzgvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '166438',
                                                                       'in_nucast_pkts': '107',
                                                                       'in_octets': '4537439198505',
                                                                       'in_ucast_pkts': '7500790876',
                                                                       'links': [       {       'in_nucast_pkts': '107',
                                                                                                'in_octets': '4537439198505',
                                                                                                'in_ucast_pkts': '7500790876',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '182807811332',
                                                                                                'out_ucast_pkts': '908092578',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '182807811332',
                                                                       'out_ucast_pkts': '908092578',
                                                                       'poller_hash': 'RENGNDQwcGVlcmludGZwb24gMTY2NDM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166438',
                                                                                               'links': [       {       'connected_time': '19938493',
                                                                                                                        'fiber_length': 20098,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RENGNDQwcGVlcnBvbnBvbiAxNjY0Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2840 LINE 45 (TAYLOR BRIAN)',
                                         'poller_hash': 'RENGNDQw',
                                         'serial_number': 'DCF440',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:b3:79:6d',
                                         'uptime': '19938493'},
                                 {       'counter_source': '166638/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:70:84'],
                                                                       'id': '166638/g1',
                                                                       'in_nucast_pkts': '452342',
                                                                       'in_octets': '99769305882',
                                                                       'in_ucast_pkts': '1192754322',
                                                                       'mac_address': '84:d3:43:7e:31:ba',
                                                                       'mtu': 2000,
                                                                       'name': '166638/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29925432',
                                                                       'out_octets': '3547385945600',
                                                                       'out_ucast_pkts': '2399280080',
                                                                       'poller_hash': 'QzgzQkEwcGVlcmludGYxNjY2MzgvZzE=',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '166638',
                                                                       'in_nucast_pkts': '29925432',
                                                                       'in_octets': '3547385945600',
                                                                       'in_ucast_pkts': '2399280080',
                                                                       'links': [       {       'in_nucast_pkts': '29925432',
                                                                                                'in_octets': '3547385945600',
                                                                                                'in_ucast_pkts': '2399280080',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '452342',
                                                                                                'out_octets': '99769305882',
                                                                                                'out_ucast_pkts': '1192754322',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '452342',
                                                                       'out_octets': '99769305882',
                                                                       'out_ucast_pkts': '1192754322',
                                                                       'poller_hash': 'QzgzQkEwcGVlcmludGZwb24gMTY2NjM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166638',
                                                                                               'links': [       {       'connected_time': '26105093',
                                                                                                                        'fiber_length': 23076,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkEwcGVlcnBvbnBvbiAxNjY2Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 POND VIEW DR (FOWLER ANNE)',
                                         'poller_hash': 'QzgzQkEw',
                                         'serial_number': 'C83BA0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:b7',
                                         'uptime': '26105093'},
                                 {       'counter_source': '166838/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:88:45:a3'],
                                                                       'id': '166838/g1',
                                                                       'in_nucast_pkts': '34',
                                                                       'in_octets': '389340886831',
                                                                       'in_ucast_pkts': '1842321151',
                                                                       'mac_address': '84:d3:43:7e:35:ba',
                                                                       'mtu': 2000,
                                                                       'name': '166838/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797397',
                                                                       'out_octets': '4233152143546',
                                                                       'out_ucast_pkts': '3780712355',
                                                                       'poller_hash': 'QzgzQzIwcGVlcmludGYxNjY4MzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '166838',
                                                                       'in_nucast_pkts': '9797397',
                                                                       'in_octets': '4233152143546',
                                                                       'in_ucast_pkts': '3780712355',
                                                                       'links': [       {       'in_nucast_pkts': '9797397',
                                                                                                'in_octets': '4233152143546',
                                                                                                'in_ucast_pkts': '3780712355',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '34',
                                                                                                'out_octets': '389340886831',
                                                                                                'out_ucast_pkts': '1842321151',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '34',
                                                                       'out_octets': '389340886831',
                                                                       'out_ucast_pkts': '1842321151',
                                                                       'poller_hash': 'QzgzQzIwcGVlcmludGZwb24gMTY2ODM4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166838',
                                                                                               'links': [       {       'connected_time': '8535267',
                                                                                                                        'fiber_length': 23900,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQzIwcGVlcnBvbnBvbiAxNjY4Mzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3709 NAFZIGER ROAD (HANER ARDEN)',
                                         'poller_hash': 'QzgzQzIw',
                                         'serial_number': 'C83C20',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:35:b7',
                                         'uptime': '8535267'},
                                 {       'counter_source': '166858/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:b4:70'],
                                                                       'id': '166858/g1',
                                                                       'in_nucast_pkts': '158068',
                                                                       'in_octets': '722005898264',
                                                                       'in_ucast_pkts': '881601907',
                                                                       'mac_address': '84:d3:43:95:8d:65',
                                                                       'mtu': 2000,
                                                                       'name': '166858/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797962',
                                                                       'out_octets': '2096718538878',
                                                                       'out_ucast_pkts': '1653461257',
                                                                       'poller_hash': 'Q0JEMUNCcGVlcmludGYxNjY4NTgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '166858',
                                                                       'in_nucast_pkts': '9797962',
                                                                       'in_octets': '2096718538878',
                                                                       'in_ucast_pkts': '1653461257',
                                                                       'links': [       {       'in_nucast_pkts': '9797962',
                                                                                                'in_octets': '2096718538878',
                                                                                                'in_ucast_pkts': '1653461257',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '158068',
                                                                                                'out_octets': '722005898264',
                                                                                                'out_ucast_pkts': '881601907',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '158068',
                                                                       'out_octets': '722005898264',
                                                                       'out_ucast_pkts': '881601907',
                                                                       'poller_hash': 'Q0JEMUNCcGVlcmludGZwb24gMTY2ODU4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '166858',
                                                                                               'links': [       {       'connected_time': '8535180',
                                                                                                                        'fiber_length': 23191,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUNCcGVlcnBvbnBvbiAxNjY4NTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '126 LAWRENCE ST (VOTEARY JIM)',
                                         'poller_hash': 'Q0JEMUNC',
                                         'serial_number': 'CBD1CB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8d:62',
                                         'uptime': '8535180'},
                                 {       'counter_source': '167237/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1f:e2'],
                                                                       'id': '167237/g1',
                                                                       'in_nucast_pkts': '157053',
                                                                       'in_octets': '140981254968',
                                                                       'in_ucast_pkts': '303182354',
                                                                       'mac_address': '84:d3:43:95:8d:f5',
                                                                       'mtu': 2000,
                                                                       'name': '167237/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9877508',
                                                                       'out_octets': '1250841397955',
                                                                       'out_ucast_pkts': '1021028935',
                                                                       'poller_hash': 'Q0JEMUREcGVlcmludGYxNjcyMzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '167237',
                                                                       'in_nucast_pkts': '9877508',
                                                                       'in_octets': '1250841397955',
                                                                       'in_ucast_pkts': '1021028935',
                                                                       'links': [       {       'in_nucast_pkts': '9877508',
                                                                                                'in_octets': '1250841397955',
                                                                                                'in_ucast_pkts': '1021028935',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '157053',
                                                                                                'out_octets': '140981254968',
                                                                                                'out_ucast_pkts': '303182354',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '157053',
                                                                       'out_octets': '140981254968',
                                                                       'out_ucast_pkts': '303182354',
                                                                       'poller_hash': 'Q0JEMUREcGVlcmludGZwb24gMTY3MjM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167237',
                                                                                               'links': [       {       'connected_time': '8535199',
                                                                                                                        'fiber_length': 23827,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUREcGVlcnBvbnBvbiAxNjcyMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3524B NAFZIGER RD (BARTLETT BRETT)',
                                         'poller_hash': 'Q0JEMURE',
                                         'serial_number': 'CBD1DD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8d:f2',
                                         'uptime': '8535199'},
                                 {       'counter_source': '167782/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:55:ba:27'],
                                                                       'id': '167782/g1',
                                                                       'in_nucast_pkts': '7265',
                                                                       'in_octets': '3358996896',
                                                                       'in_ucast_pkts': '16431459',
                                                                       'mac_address': '84:d3:43:95:8f:3d',
                                                                       'mtu': 2000,
                                                                       'name': '167782/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394296',
                                                                       'out_octets': '124290105710',
                                                                       'out_ucast_pkts': '96193084',
                                                                       'poller_hash': 'Q0JEMjA2cGVlcmludGYxNjc3ODIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '167782',
                                                                       'in_nucast_pkts': '394296',
                                                                       'in_octets': '124290105710',
                                                                       'in_ucast_pkts': '96193084',
                                                                       'links': [       {       'in_nucast_pkts': '394296',
                                                                                                'in_octets': '124290105710',
                                                                                                'in_ucast_pkts': '96193084',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '7265',
                                                                                                'out_octets': '3358996896',
                                                                                                'out_ucast_pkts': '16431459',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7265',
                                                                       'out_octets': '3358996896',
                                                                       'out_ucast_pkts': '16431459',
                                                                       'poller_hash': 'Q0JEMjA2cGVlcmludGZwb24gMTY3Nzgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167782',
                                                                                               'links': [       {       'connected_time': '335343',
                                                                                                                        'fiber_length': 18659,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjA2cGVlcnBvbnBvbiAxNjc3ODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5329 Road 106 (Scott Rupert)',
                                         'poller_hash': 'Q0JEMjA2',
                                         'serial_number': 'CBD206',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8f:3a',
                                         'uptime': '335343'},
                                 {       'counter_source': '167792/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:13:56'],
                                                                       'id': '167792/g1',
                                                                       'in_nucast_pkts': '150004',
                                                                       'in_octets': '148966826184',
                                                                       'in_ucast_pkts': '447998595',
                                                                       'mac_address': '84:d3:43:95:8f:75',
                                                                       'mtu': 2000,
                                                                       'name': '167792/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878225',
                                                                       'out_octets': '1430438071545',
                                                                       'out_ucast_pkts': '1116236778',
                                                                       'poller_hash': 'Q0JEMjBEcGVlcmludGYxNjc3OTIvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '167792',
                                                                       'in_nucast_pkts': '9878225',
                                                                       'in_octets': '1430438071545',
                                                                       'in_ucast_pkts': '1116236778',
                                                                       'links': [       {       'in_nucast_pkts': '9878225',
                                                                                                'in_octets': '1430438071545',
                                                                                                'in_ucast_pkts': '1116236778',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '150004',
                                                                                                'out_octets': '148966826184',
                                                                                                'out_ucast_pkts': '447998595',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150004',
                                                                       'out_octets': '148966826184',
                                                                       'out_ucast_pkts': '447998595',
                                                                       'poller_hash': 'Q0JEMjBEcGVlcmludGZwb24gMTY3Nzky',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167792',
                                                                                               'links': [       {       'connected_time': '8535206',
                                                                                                                        'fiber_length': 25379,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMjBEcGVlcnBvbnBvbiAxNjc3OTI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '23 GERBER MEADOWS DR (SHANTZ BRENDA)',
                                         'poller_hash': 'Q0JEMjBE',
                                         'serial_number': 'CBD20D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8f:72',
                                         'uptime': '8535206'},
                                 {       'counter_source': '167878/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b9:7b'],
                                                                       'id': '167878/g1',
                                                                       'in_nucast_pkts': '155046',
                                                                       'in_octets': '301518126624',
                                                                       'in_ucast_pkts': '765660369',
                                                                       'mac_address': '84:d3:43:95:8c:dd',
                                                                       'mtu': 2000,
                                                                       'name': '167878/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797809',
                                                                       'out_octets': '3216657779715',
                                                                       'out_ucast_pkts': '2420052928',
                                                                       'poller_hash': 'Q0JEMUJBcGVlcmludGYxNjc4NzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '167878',
                                                                       'in_nucast_pkts': '9797809',
                                                                       'in_octets': '3216657779715',
                                                                       'in_ucast_pkts': '2420052928',
                                                                       'links': [       {       'in_nucast_pkts': '9797809',
                                                                                                'in_octets': '3216657779715',
                                                                                                'in_ucast_pkts': '2420052928',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '155046',
                                                                                                'out_octets': '301518126624',
                                                                                                'out_ucast_pkts': '765660369',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '155046',
                                                                       'out_octets': '301518126624',
                                                                       'out_ucast_pkts': '765660369',
                                                                       'poller_hash': 'Q0JEMUJBcGVlcmludGZwb24gMTY3ODc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167878',
                                                                                               'links': [       {       'connected_time': '8535205',
                                                                                                                        'fiber_length': 24054,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUJBcGVlcnBvbnBvbiAxNjc4Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '87 REINER CR (PATTERSON JENNIFER)',
                                         'poller_hash': 'Q0JEMUJB',
                                         'serial_number': 'CBD1BA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8c:da',
                                         'uptime': '8535205'},
                                 {       'counter_source': '167908/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:2a:c7:9e'],
                                                                       'id': '167908/g1',
                                                                       'in_nucast_pkts': '473430',
                                                                       'in_octets': '2700942599',
                                                                       'in_ucast_pkts': '28256129',
                                                                       'mac_address': '84:d3:43:95:8d:b5',
                                                                       'mtu': 2000,
                                                                       'name': '167908/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '95603717',
                                                                       'out_octets': '162977769578',
                                                                       'out_ucast_pkts': '89951575',
                                                                       'poller_hash': 'Q0JEMUQ1cGVlcmludGYxNjc5MDgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '167908',
                                                                       'in_nucast_pkts': '95603717',
                                                                       'in_octets': '162977769578',
                                                                       'in_ucast_pkts': '89951575',
                                                                       'links': [       {       'in_nucast_pkts': '95603717',
                                                                                                'in_octets': '162977769578',
                                                                                                'in_ucast_pkts': '89951575',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '473430',
                                                                                                'out_octets': '2700942599',
                                                                                                'out_ucast_pkts': '28256129',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '473430',
                                                                       'out_octets': '2700942599',
                                                                       'out_ucast_pkts': '28256129',
                                                                       'poller_hash': 'Q0JEMUQ1cGVlcmludGZwb24gMTY3OTA4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167908',
                                                                                               'links': [       {       'connected_time': '8535136',
                                                                                                                        'fiber_length': 23176,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUQ1cGVlcnBvbnBvbiAxNjc5MDg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 POND VIEW DR (GAJDOS ANITA)',
                                         'poller_hash': 'Q0JEMUQ1',
                                         'serial_number': 'CBD1D5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8d:b2',
                                         'uptime': '8535136'},
                                 {       'counter_source': '167958/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '167958/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3160681212',
                                                                       'in_ucast_pkts': '10878071',
                                                                       'mac_address': '04:bc:9f:5b:d8:a9',
                                                                       'name': '167958/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '39187841749',
                                                                       'out_ucast_pkts': '35116132',
                                                                       'poller_hash': 'MTI3NDE5NXBlZXJpbnRmMTY3OTU4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '167958',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '39187841749',
                                                                       'in_ucast_pkts': '35116132',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '39187841749',
                                                                                                'in_ucast_pkts': '35116132',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3160681212',
                                                                                                'out_ucast_pkts': '10878071',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3160681212',
                                                                       'out_ucast_pkts': '10878071',
                                                                       'poller_hash': 'MTI3NDE5NXBlZXJpbnRmcG9uIDE2Nzk1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '167958',
                                                                                               'links': [       {       'connected_time': '3204783',
                                                                                                                        'fiber_length': 22826,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE5NXBlZXJwb25wb24gMTY3OTU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3771 NAFZIGER RD (FALK RON)',
                                         'poller_hash': 'MTI3NDE5NQ==',
                                         'serial_number': '1274195',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:a9',
                                         'uptime': '3204783'},
                                 {       'counter_source': '168105/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1d:a6'],
                                                                       'id': '168105/g1',
                                                                       'in_nucast_pkts': '148451',
                                                                       'in_octets': '123019137815',
                                                                       'in_ucast_pkts': '535610090',
                                                                       'mac_address': '84:d3:43:96:04:05',
                                                                       'mtu': 2000,
                                                                       'name': '168105/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9925763',
                                                                       'out_octets': '2026845143889',
                                                                       'out_ucast_pkts': '1604495682',
                                                                       'poller_hash': 'Q0JFMDlGcGVlcmludGYxNjgxMDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '168105',
                                                                       'in_nucast_pkts': '9925763',
                                                                       'in_octets': '2026845143889',
                                                                       'in_ucast_pkts': '1604495682',
                                                                       'links': [       {       'in_nucast_pkts': '9925763',
                                                                                                'in_octets': '2026845143889',
                                                                                                'in_ucast_pkts': '1604495682',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '148451',
                                                                                                'out_octets': '123019137815',
                                                                                                'out_ucast_pkts': '535610090',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '148451',
                                                                       'out_octets': '123019137815',
                                                                       'out_ucast_pkts': '535610090',
                                                                       'poller_hash': 'Q0JFMDlGcGVlcmludGZwb24gMTY4MTA1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168105',
                                                                                               'links': [       {       'connected_time': '8535235',
                                                                                                                        'fiber_length': 23061,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JFMDlGcGVlcnBvbnBvbiAxNjgxMDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1096 MOLESWORTH ST (MCGOVERN MATT)',
                                         'poller_hash': 'Q0JFMDlG',
                                         'serial_number': 'CBE09F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:96:04:02',
                                         'uptime': '8535235'},
                                 {       'counter_source': '168222/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '168222/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '250929124239',
                                                                       'in_ucast_pkts': '625337751',
                                                                       'mac_address': '84:d3:43:95:8b:8a',
                                                                       'name': '168222/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '70',
                                                                       'out_octets': '2110453430276',
                                                                       'out_ucast_pkts': '1712366328',
                                                                       'poller_hash': 'Q0JEMTkwcGVlcmludGYxNjgyMjIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '168222',
                                                                       'in_nucast_pkts': '70',
                                                                       'in_octets': '2110453430276',
                                                                       'in_ucast_pkts': '1712366328',
                                                                       'links': [       {       'in_nucast_pkts': '70',
                                                                                                'in_octets': '2110453430276',
                                                                                                'in_ucast_pkts': '1712366328',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '250929124239',
                                                                                                'out_ucast_pkts': '625337751',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '250929124239',
                                                                       'out_ucast_pkts': '625337751',
                                                                       'poller_hash': 'Q0JEMTkwcGVlcmludGZwb24gMTY4MjIy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168222',
                                                                                               'links': [       {       'connected_time': '335343',
                                                                                                                        'fiber_length': 17602,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMTkwcGVlcnBvbnBvbiAxNjgyMjI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5470 KING ST (BAILLIE BRAD)',
                                         'poller_hash': 'Q0JEMTkw',
                                         'serial_number': 'CBD190',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8b:8a',
                                         'uptime': '335343'},
                                 {       'counter_source': '168389/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '168389/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '580888822682',
                                                                       'in_ucast_pkts': '23049604361',
                                                                       'mac_address': '84:d3:43:96:02:92',
                                                                       'name': '168389/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '16602389041182',
                                                                       'out_ucast_pkts': '12057190323',
                                                                       'poller_hash': 'Q0JFMDcxcGVlcmludGYxNjgzODkvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '168389',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '16602389041182',
                                                                       'in_ucast_pkts': '12057190323',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '16602389041182',
                                                                                                'in_ucast_pkts': '12057190323',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '580888822682',
                                                                                                'out_ucast_pkts': '23049604361',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '580888822682',
                                                                       'out_ucast_pkts': '23049604361',
                                                                       'poller_hash': 'Q0JFMDcxcGVlcmludGZwb24gMTY4Mzg5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168389',
                                                                                               'links': [       {       'connected_time': '11539619',
                                                                                                                        'fiber_length': 22581,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JFMDcxcGVlcnBvbnBvbiAxNjgzODk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1202 QUEEN'S BUSH RD (BORSATO LARRY)",
                                         'poller_hash': 'Q0JFMDcx',
                                         'serial_number': 'CBE071',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:96:02:92',
                                         'uptime': '11539619'},
                                 {       'counter_source': '168494/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '168494/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '580699373417',
                                                                       'in_ucast_pkts': '40883946000',
                                                                       'mac_address': '84:d3:43:96:00:82',
                                                                       'name': '168494/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '56',
                                                                       'out_octets': '6513907509347',
                                                                       'out_ucast_pkts': '5026822959',
                                                                       'poller_hash': 'Q0JFMDJGcGVlcmludGYxNjg0OTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '168494',
                                                                       'in_nucast_pkts': '56',
                                                                       'in_octets': '6513907509347',
                                                                       'in_ucast_pkts': '5026822959',
                                                                       'links': [       {       'in_nucast_pkts': '56',
                                                                                                'in_octets': '6513907509347',
                                                                                                'in_ucast_pkts': '5026822959',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '580699373417',
                                                                                                'out_ucast_pkts': '40883946000',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '580699373417',
                                                                       'out_ucast_pkts': '40883946000',
                                                                       'poller_hash': 'Q0JFMDJGcGVlcmludGZwb24gMTY4NDk0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168494',
                                                                                               'links': [       {       'connected_time': '1096769',
                                                                                                                        'fiber_length': 28898,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JFMDJGcGVlcnBvbnBvbiAxNjg0OTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4332 WEIMAR LINE (ROSS JEFF)',
                                         'poller_hash': 'Q0JFMDJG',
                                         'serial_number': 'CBE02F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:96:00:82',
                                         'uptime': '1096769'},
                                 {       'counter_source': '168514/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '168514/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '827565637526',
                                                                       'in_ucast_pkts': '2088421011',
                                                                       'mac_address': '84:d3:43:7e:31:8f',
                                                                       'name': '168514/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '131',
                                                                       'out_octets': '6307374251140',
                                                                       'out_ucast_pkts': '5019726148',
                                                                       'poller_hash': 'QzgzQjlCcGVlcmludGYxNjg1MTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '168514',
                                                                       'in_nucast_pkts': '131',
                                                                       'in_octets': '6307374251140',
                                                                       'in_ucast_pkts': '5019726148',
                                                                       'links': [       {       'in_nucast_pkts': '131',
                                                                                                'in_octets': '6307374251140',
                                                                                                'in_ucast_pkts': '5019726148',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '827565637526',
                                                                                                'out_ucast_pkts': '2088421011',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '827565637526',
                                                                       'out_ucast_pkts': '2088421011',
                                                                       'poller_hash': 'QzgzQjlCcGVlcmludGZwb24gMTY4NTE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168514',
                                                                                               'links': [       {       'connected_time': '22786823',
                                                                                                                        'fiber_length': 24813,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjlCcGVlcnBvbnBvbiAxNjg1MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '13 FIRELLA PL (ARENDOQUE LORD)',
                                         'poller_hash': 'QzgzQjlC',
                                         'serial_number': 'C83B9B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:8f',
                                         'uptime': '22786823'},
                                 {       'counter_source': '168544/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '168544/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '351129482327',
                                                                       'in_ucast_pkts': '1782229673',
                                                                       'mac_address': '84:d3:43:7e:07:57',
                                                                       'name': '168544/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '17',
                                                                       'out_octets': '5798714029569',
                                                                       'out_ucast_pkts': '4432117361',
                                                                       'poller_hash': 'QzgzNjU0cGVlcmludGYxNjg1NDQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '168544',
                                                                       'in_nucast_pkts': '17',
                                                                       'in_octets': '5798714029569',
                                                                       'in_ucast_pkts': '4432117361',
                                                                       'links': [       {       'in_nucast_pkts': '17',
                                                                                                'in_octets': '5798714029569',
                                                                                                'in_ucast_pkts': '4432117361',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '351129482327',
                                                                                                'out_ucast_pkts': '1782229673',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '351129482327',
                                                                       'out_ucast_pkts': '1782229673',
                                                                       'poller_hash': 'QzgzNjU0cGVlcmludGZwb24gMTY4NTQ0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '168544',
                                                                                               'links': [       {       'connected_time': '4330025',
                                                                                                                        'fiber_length': 23146,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzNjU0cGVlcnBvbnBvbiAxNjg1NDQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '145 SCHWEITZER CR (MCCLEMENTS TIM)',
                                         'poller_hash': 'QzgzNjU0',
                                         'serial_number': 'C83654',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:07:57',
                                         'uptime': '4330025'},
                                 {       'counter_source': '170445/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:7e:d4:8f'],
                                                                       'id': '170445/g1',
                                                                       'in_nucast_pkts': '46',
                                                                       'in_octets': '142497291691',
                                                                       'in_ucast_pkts': '1486248217',
                                                                       'mac_address': '84:d3:43:7e:31:e2',
                                                                       'mtu': 2000,
                                                                       'name': '170445/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '31007740',
                                                                       'out_octets': '5885428015901',
                                                                       'out_ucast_pkts': '4042997512',
                                                                       'poller_hash': 'QzgzQkE1cGVlcmludGYxNzA0NDUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '170445',
                                                                       'in_nucast_pkts': '31007740',
                                                                       'in_octets': '5885428015901',
                                                                       'in_ucast_pkts': '4042997512',
                                                                       'links': [       {       'in_nucast_pkts': '31007740',
                                                                                                'in_octets': '5885428015901',
                                                                                                'in_ucast_pkts': '4042997512',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '46',
                                                                                                'out_octets': '142497291691',
                                                                                                'out_ucast_pkts': '1486248217',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '46',
                                                                       'out_octets': '142497291691',
                                                                       'out_ucast_pkts': '1486248217',
                                                                       'poller_hash': 'QzgzQkE1cGVlcmludGZwb24gMTcwNDQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '170445',
                                                                                               'links': [       {       'connected_time': '28348050',
                                                                                                                        'fiber_length': 22822,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkE1cGVlcnBvbnBvbiAxNzA0NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '32 POND VIEW DR (ERB FRANKLIN & '
                                                 'JEAN)',
                                         'poller_hash': 'QzgzQkE1',
                                         'serial_number': 'C83BA5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:df',
                                         'uptime': '28348050'},
                                 {       'counter_source': '171759/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:de:de:92'],
                                                                       'id': '171759/g1',
                                                                       'in_nucast_pkts': '660',
                                                                       'in_octets': '65273100733',
                                                                       'in_ucast_pkts': '201815069',
                                                                       'mac_address': '84:d3:43:7d:e3:02',
                                                                       'mtu': 2000,
                                                                       'name': '171759/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878153',
                                                                       'out_octets': '2228468029943',
                                                                       'out_ucast_pkts': '1603512125',
                                                                       'poller_hash': 'QzgzMUM5cGVlcmludGYxNzE3NTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '171759',
                                                                       'in_nucast_pkts': '9878153',
                                                                       'in_octets': '2228468029943',
                                                                       'in_ucast_pkts': '1603512125',
                                                                       'links': [       {       'in_nucast_pkts': '9878153',
                                                                                                'in_octets': '2228468029943',
                                                                                                'in_ucast_pkts': '1603512125',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '660',
                                                                                                'out_octets': '65273100733',
                                                                                                'out_ucast_pkts': '201815069',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '660',
                                                                       'out_octets': '65273100733',
                                                                       'out_ucast_pkts': '201815069',
                                                                       'poller_hash': 'QzgzMUM5cGVlcmludGZwb24gMTcxNzU5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '171759',
                                                                                               'links': [       {       'connected_time': '8535158',
                                                                                                                        'fiber_length': 24770,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -22.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMUM5cGVlcnBvbnBvbiAxNzE3NTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '63 NITHBANK RD (DAY RYAN)',
                                         'poller_hash': 'QzgzMUM5',
                                         'serial_number': 'C831C9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:e2:ff',
                                         'uptime': '8535158'},
                                 {       'counter_source': '174514/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ea:fb'],
                                                                       'id': '174514/g1',
                                                                       'in_nucast_pkts': '149208',
                                                                       'in_octets': '113271499726',
                                                                       'in_ucast_pkts': '361966077',
                                                                       'mac_address': '84:d3:43:7e:15:ba',
                                                                       'mtu': 2000,
                                                                       'name': '174514/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878302',
                                                                       'out_octets': '3459693388926',
                                                                       'out_ucast_pkts': '2456125047',
                                                                       'poller_hash': 'QzgzODIwcGVlcmludGYxNzQ1MTQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174514',
                                                                       'in_nucast_pkts': '9878302',
                                                                       'in_octets': '3459693388926',
                                                                       'in_ucast_pkts': '2456125047',
                                                                       'links': [       {       'in_nucast_pkts': '9878302',
                                                                                                'in_octets': '3459693388926',
                                                                                                'in_ucast_pkts': '2456125047',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149208',
                                                                                                'out_octets': '113271499726',
                                                                                                'out_ucast_pkts': '361966077',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149208',
                                                                       'out_octets': '113271499726',
                                                                       'out_ucast_pkts': '361966077',
                                                                       'poller_hash': 'QzgzODIwcGVlcmludGZwb24gMTc0NTE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174514',
                                                                                               'links': [       {       'connected_time': '8535156',
                                                                                                                        'fiber_length': 24780,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzODIwcGVlcnBvbnBvbiAxNzQ1MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 VILLAGE RD (SCHMIDT RICHARD)',
                                         'poller_hash': 'QzgzODIw',
                                         'serial_number': 'C83820',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:15:b7',
                                         'uptime': '8535156'},
                                 {       'counter_source': '174931/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ef:f7'],
                                                                       'id': '174931/g1',
                                                                       'in_nucast_pkts': '149446',
                                                                       'in_octets': '55680950390',
                                                                       'in_ucast_pkts': '106348544',
                                                                       'mac_address': 'f8:85:f9:17:29:2b',
                                                                       'mtu': 2000,
                                                                       'name': '174931/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878248',
                                                                       'out_octets': '491466332922',
                                                                       'out_ucast_pkts': '370437533',
                                                                       'poller_hash': 'REZGRTBFcGVlcmludGYxNzQ5MzEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174931',
                                                                       'in_nucast_pkts': '9878248',
                                                                       'in_octets': '491466332922',
                                                                       'in_ucast_pkts': '370437533',
                                                                       'links': [       {       'in_nucast_pkts': '9878248',
                                                                                                'in_octets': '491466332922',
                                                                                                'in_ucast_pkts': '370437533',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149446',
                                                                                                'out_octets': '55680950390',
                                                                                                'out_ucast_pkts': '106348544',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149446',
                                                                       'out_octets': '55680950390',
                                                                       'out_ucast_pkts': '106348544',
                                                                       'poller_hash': 'REZGRTBFcGVlcmludGZwb24gMTc0OTMx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174931',
                                                                                               'links': [       {       'connected_time': '8535185',
                                                                                                                        'fiber_length': 25053,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTBFcGVlcnBvbnBvbiAxNzQ5MzE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '40 VILLAGE RD (ERB WENDY)',
                                         'poller_hash': 'REZGRTBF',
                                         'serial_number': 'DFFE0E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:29:28',
                                         'uptime': '8535185'},
                                 {       'counter_source': '174979/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e9:4e'],
                                                                       'id': '174979/g1',
                                                                       'in_nucast_pkts': '158315',
                                                                       'in_octets': '210489277399',
                                                                       'in_ucast_pkts': '469777513',
                                                                       'mac_address': '84:d3:43:7e:31:62',
                                                                       'mtu': 2000,
                                                                       'name': '174979/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9876305',
                                                                       'out_octets': '1474793761268',
                                                                       'out_ucast_pkts': '1229683318',
                                                                       'poller_hash': 'QzgzQjk1cGVlcmludGYxNzQ5NzkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '174979',
                                                                       'in_nucast_pkts': '9876305',
                                                                       'in_octets': '1474793761268',
                                                                       'in_ucast_pkts': '1229683318',
                                                                       'links': [       {       'in_nucast_pkts': '9876305',
                                                                                                'in_octets': '1474793761268',
                                                                                                'in_ucast_pkts': '1229683318',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '158315',
                                                                                                'out_octets': '210489277399',
                                                                                                'out_ucast_pkts': '469777513',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '158315',
                                                                       'out_octets': '210489277399',
                                                                       'out_ucast_pkts': '469777513',
                                                                       'poller_hash': 'QzgzQjk1cGVlcmludGZwb24gMTc0OTc5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '174979',
                                                                                               'links': [       {       'connected_time': '8535223',
                                                                                                                        'fiber_length': 24642,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjk1cGVlcnBvbnBvbiAxNzQ5Nzk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 FIRELLA PL (ROBINSON MARTIN)',
                                         'poller_hash': 'QzgzQjk1',
                                         'serial_number': 'C83B95',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:31:5f',
                                         'uptime': '8535223'},
                                 {       'counter_source': '175351/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1d:7a'],
                                                                       'id': '175351/g1',
                                                                       'in_nucast_pkts': '150440',
                                                                       'in_octets': '408366261503',
                                                                       'in_ucast_pkts': '1437720367',
                                                                       'mac_address': '84:d3:43:95:8b:85',
                                                                       'mtu': 2000,
                                                                       'name': '175351/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878145',
                                                                       'out_octets': '3900500422215',
                                                                       'out_ucast_pkts': '3045343083',
                                                                       'poller_hash': 'Q0JEMThGcGVlcmludGYxNzUzNTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '175351',
                                                                       'in_nucast_pkts': '9878145',
                                                                       'in_octets': '3900500422215',
                                                                       'in_ucast_pkts': '3045343083',
                                                                       'links': [       {       'in_nucast_pkts': '9878145',
                                                                                                'in_octets': '3900500422215',
                                                                                                'in_ucast_pkts': '3045343083',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '150440',
                                                                                                'out_octets': '408366261503',
                                                                                                'out_ucast_pkts': '1437720367',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150440',
                                                                       'out_octets': '408366261503',
                                                                       'out_ucast_pkts': '1437720367',
                                                                       'poller_hash': 'Q0JEMThGcGVlcmludGZwb24gMTc1MzUx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175351',
                                                                                               'links': [       {       'connected_time': '8535108',
                                                                                                                        'fiber_length': 24927,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMThGcGVlcnBvbnBvbiAxNzUzNTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '102 GERBER MEADOWS DR (LADD ANITA)',
                                         'poller_hash': 'Q0JEMThG',
                                         'serial_number': 'CBD18F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8b:82',
                                         'uptime': '8535108'},
                                 {       'counter_source': '175802/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '175802/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '495132749696',
                                                                       'in_ucast_pkts': '4704141158',
                                                                       'mac_address': '84:d3:43:95:8c:ca',
                                                                       'name': '175802/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '153',
                                                                       'out_octets': '19247871353985',
                                                                       'out_ucast_pkts': '13300642254',
                                                                       'poller_hash': 'Q0JEMUI4cGVlcmludGYxNzU4MDIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '175802',
                                                                       'in_nucast_pkts': '153',
                                                                       'in_octets': '19247871353985',
                                                                       'in_ucast_pkts': '13300642254',
                                                                       'links': [       {       'in_nucast_pkts': '153',
                                                                                                'in_octets': '19247871353985',
                                                                                                'in_ucast_pkts': '13300642254',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '495132749696',
                                                                                                'out_ucast_pkts': '4704141158',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '495132749696',
                                                                       'out_ucast_pkts': '4704141158',
                                                                       'poller_hash': 'Q0JEMUI4cGVlcmludGZwb24gMTc1ODAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '175802',
                                                                                               'links': [       {       'connected_time': '8535297',
                                                                                                                        'fiber_length': 23499,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'Q0JEMUI4cGVlcnBvbnBvbiAxNzU4MDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1010 MAPLE LEAF ST (GERVAIS STEVE)',
                                         'poller_hash': 'Q0JEMUI4',
                                         'serial_number': 'CBD1B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:95:8c:ca',
                                         'uptime': '8535297'},
                                 {       'counter_source': '180377/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '180377/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '58832771952',
                                                                       'in_ucast_pkts': '113999801',
                                                                       'mac_address': '84:d3:43:7e:33:4f',
                                                                       'name': '180377/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '53',
                                                                       'out_octets': '415603916960',
                                                                       'out_ucast_pkts': '357578338',
                                                                       'poller_hash': 'QzgzQkQzcGVlcmludGYxODAzNzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180377',
                                                                       'in_nucast_pkts': '53',
                                                                       'in_octets': '415603916960',
                                                                       'in_ucast_pkts': '357578338',
                                                                       'links': [       {       'in_nucast_pkts': '53',
                                                                                                'in_octets': '415603916960',
                                                                                                'in_ucast_pkts': '357578338',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '58832771952',
                                                                                                'out_ucast_pkts': '113999801',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '58832771952',
                                                                       'out_ucast_pkts': '113999801',
                                                                       'poller_hash': 'QzgzQkQzcGVlcmludGZwb24gMTgwMzc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180377',
                                                                                               'links': [       {       'connected_time': '611452',
                                                                                                                        'fiber_length': 19825,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQkQzcGVlcnBvbnBvbiAxODAzNzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2854 LINE 45 (JANZEN AARON & MARIA)',
                                         'poller_hash': 'QzgzQkQz',
                                         'serial_number': 'C83BD3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:33:4f',
                                         'uptime': '611452'},
                                 {       'counter_source': '180630/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '180630/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '301',
                                                                       'in_octets': '278225036949',
                                                                       'in_ucast_pkts': '2118071520',
                                                                       'mac_address': '60:db:98:73:e2:e3',
                                                                       'name': '180630/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '552',
                                                                       'out_octets': '6663471212577',
                                                                       'out_ucast_pkts': '4972625411',
                                                                       'poller_hash': 'QjBBQUY3cGVlcmludGYxODA2MzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '180630',
                                                                       'in_nucast_pkts': '552',
                                                                       'in_octets': '6663471212577',
                                                                       'in_ucast_pkts': '4972625411',
                                                                       'links': [       {       'in_nucast_pkts': '552',
                                                                                                'in_octets': '6663471212577',
                                                                                                'in_ucast_pkts': '4972625411',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '301',
                                                                                                'out_octets': '278225036949',
                                                                                                'out_ucast_pkts': '2118071520',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '301',
                                                                       'out_octets': '278225036949',
                                                                       'out_ucast_pkts': '2118071520',
                                                                       'poller_hash': 'QjBBQUY3cGVlcmludGZwb24gMTgwNjMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '180630',
                                                                                               'links': [       {       'connected_time': '8535388',
                                                                                                                        'fiber_length': 23501,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 1.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QjBBQUY3cGVlcnBvbnBvbiAxODA2MzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '844GE-1',
                                         'name': '1020 MOLESWORTH ST (BENDER WYATT)',
                                         'poller_hash': 'QjBBQUY3',
                                         'serial_number': 'B0AAF7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:73:e2:e3',
                                         'uptime': '8535388'},
                                 {       'counter_source': '181111/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:73:91'],
                                                                       'id': '181111/g1',
                                                                       'in_nucast_pkts': '147406',
                                                                       'in_octets': '200091345447',
                                                                       'in_ucast_pkts': '567472555',
                                                                       'mac_address': 'f8:85:f9:17:28:cb',
                                                                       'mtu': 2000,
                                                                       'name': '181111/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9516701',
                                                                       'out_octets': '1935917451713',
                                                                       'out_ucast_pkts': '1635418517',
                                                                       'poller_hash': 'REZGRTAycGVlcmludGYxODExMTEvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '181111',
                                                                       'in_nucast_pkts': '9516701',
                                                                       'in_octets': '1935917451713',
                                                                       'in_ucast_pkts': '1635418517',
                                                                       'links': [       {       'in_nucast_pkts': '9516701',
                                                                                                'in_octets': '1935917451713',
                                                                                                'in_ucast_pkts': '1635418517',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '147406',
                                                                                                'out_octets': '200091345447',
                                                                                                'out_ucast_pkts': '567472555',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '147406',
                                                                       'out_octets': '200091345447',
                                                                       'out_ucast_pkts': '567472555',
                                                                       'poller_hash': 'REZGRTAycGVlcmludGZwb24gMTgxMTEx',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '181111',
                                                                                               'links': [       {       'connected_time': '8288457',
                                                                                                                        'fiber_length': 23386,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTAycGVlcnBvbnBvbiAxODExMTE=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '142 LAWRENCE ST (REED GEOFF)',
                                         'poller_hash': 'REZGRTAy',
                                         'serial_number': 'DFFE02',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:28:c8',
                                         'uptime': '8288457'},
                                 {       'counter_source': '183182/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '183182/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '320387270029',
                                                                       'in_ucast_pkts': '1776373869',
                                                                       'mac_address': 'f8:85:f9:17:29:d0',
                                                                       'name': '183182/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '58',
                                                                       'out_octets': '8891958179039',
                                                                       'out_ucast_pkts': '6530057890',
                                                                       'poller_hash': 'REZGRTIzcGVlcmludGYxODMxODIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '183182',
                                                                       'in_nucast_pkts': '58',
                                                                       'in_octets': '8891958179039',
                                                                       'in_ucast_pkts': '6530057890',
                                                                       'links': [       {       'in_nucast_pkts': '58',
                                                                                                'in_octets': '8891958179039',
                                                                                                'in_ucast_pkts': '6530057890',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '320387270029',
                                                                                                'out_ucast_pkts': '1776373869',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '320387270029',
                                                                       'out_ucast_pkts': '1776373869',
                                                                       'poller_hash': 'REZGRTIzcGVlcmludGZwb24gMTgzMTgy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '183182',
                                                                                               'links': [       {       'connected_time': '1096769',
                                                                                                                        'fiber_length': 31925,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTIzcGVlcnBvbnBvbiAxODMxODI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3956 WEIMAR LINE (ARNOTT TAMARA)',
                                         'poller_hash': 'REZGRTIz',
                                         'serial_number': 'DFFE23',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:29:d0',
                                         'uptime': '1096769'},
                                 {       'counter_source': '184735/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:fd:74:1d:05:ce'],
                                                                       'id': '184735/g1',
                                                                       'in_nucast_pkts': '63',
                                                                       'in_octets': '2412648999213',
                                                                       'in_ucast_pkts': '2157430160',
                                                                       'mac_address': '84:d3:43:7e:2d:72',
                                                                       'mtu': 2000,
                                                                       'name': '184735/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878567',
                                                                       'out_octets': '1986554904429',
                                                                       'out_ucast_pkts': '2127243194',
                                                                       'poller_hash': 'QzgzQjE3cGVlcmludGYxODQ3MzUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '184735',
                                                                       'in_nucast_pkts': '9878567',
                                                                       'in_octets': '1986554904429',
                                                                       'in_ucast_pkts': '2127243194',
                                                                       'links': [       {       'in_nucast_pkts': '9878567',
                                                                                                'in_octets': '1986554904429',
                                                                                                'in_ucast_pkts': '2127243194',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '63',
                                                                                                'out_octets': '2412648999213',
                                                                                                'out_ucast_pkts': '2157430160',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63',
                                                                       'out_octets': '2412648999213',
                                                                       'out_ucast_pkts': '2157430160',
                                                                       'poller_hash': 'QzgzQjE3cGVlcmludGZwb24gMTg0NzM1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '184735',
                                                                                               'links': [       {       'connected_time': '8535146',
                                                                                                                        'fiber_length': 23249,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQjE3cGVlcnBvbnBvbiAxODQ3MzU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3580 NAFZIGER RD WELLESLEY '
                                                 '(WELLESLEY VET CLINIC)',
                                         'poller_hash': 'QzgzQjE3',
                                         'serial_number': 'C83B17',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:2d:6f',
                                         'uptime': '8535146'},
                                 {       'counter_source': '186507/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186507/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '623082883234',
                                                                       'in_ucast_pkts': '25204925642',
                                                                       'mac_address': 'f8:85:f9:17:2a:f0',
                                                                       'name': '186507/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '16304780654239',
                                                                       'out_ucast_pkts': '12160570696',
                                                                       'poller_hash': 'REZGRTQ3cGVlcmludGYxODY1MDcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186507',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '16304780654239',
                                                                       'in_ucast_pkts': '12160570696',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '16304780654239',
                                                                                                'in_ucast_pkts': '12160570696',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '623082883234',
                                                                                                'out_ucast_pkts': '25204925642',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '623082883234',
                                                                       'out_ucast_pkts': '25204925642',
                                                                       'poller_hash': 'REZGRTQ3cGVlcmludGZwb24gMTg2NTA3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186507',
                                                                                               'links': [       {       'connected_time': '8535213',
                                                                                                                        'fiber_length': 23133,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRTQ3cGVlcnBvbnBvbiAxODY1MDc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3610 NAFZIGER RD (BILLSON JOANNA)',
                                         'poller_hash': 'REZGRTQ3',
                                         'serial_number': 'DFFE47',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:2a:f0',
                                         'uptime': '8535213'},
                                 {       'counter_source': '186637/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1a:a4'],
                                                                       'id': '186637/g1',
                                                                       'in_nucast_pkts': '468455',
                                                                       'in_octets': '497622876606',
                                                                       'in_ucast_pkts': '2114067570',
                                                                       'mac_address': 'f8:85:f9:17:21:1b',
                                                                       'mtu': 2000,
                                                                       'name': '186637/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29896962',
                                                                       'out_octets': '12723172969411',
                                                                       'out_ucast_pkts': '8973222816',
                                                                       'poller_hash': 'REZGRDBDcGVlcmludGYxODY2MzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '186637',
                                                                       'in_nucast_pkts': '29896962',
                                                                       'in_octets': '12723172969411',
                                                                       'in_ucast_pkts': '8973222816',
                                                                       'links': [       {       'in_nucast_pkts': '29896962',
                                                                                                'in_octets': '12723172969411',
                                                                                                'in_ucast_pkts': '8973222816',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '468455',
                                                                                                'out_octets': '497622876606',
                                                                                                'out_ucast_pkts': '2114067570',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '468455',
                                                                       'out_octets': '497622876606',
                                                                       'out_ucast_pkts': '2114067570',
                                                                       'poller_hash': 'REZGRDBDcGVlcmludGZwb24gMTg2NjM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186637',
                                                                                               'links': [       {       'connected_time': '25916541',
                                                                                                                        'fiber_length': 24714,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'REZGRDBDcGVlcnBvbnBvbiAxODY2Mzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '70 NITHBANK RD (NOLAN DENNIS)',
                                         'poller_hash': 'REZGRDBD',
                                         'serial_number': 'DFFD0C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'f8:85:f9:17:21:18',
                                         'uptime': '25916541'},
                                 {       'counter_source': '186776/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '186776/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '99454265911',
                                                                       'in_ucast_pkts': '870693132',
                                                                       'mac_address': '84:d3:43:b3:7d:75',
                                                                       'name': '186776/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '77',
                                                                       'out_octets': '3626593535230',
                                                                       'out_ucast_pkts': '2604983567',
                                                                       'poller_hash': 'RENGNEMxcGVlcmludGYxODY3NzYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '186776',
                                                                       'in_nucast_pkts': '77',
                                                                       'in_octets': '3626593535230',
                                                                       'in_ucast_pkts': '2604983567',
                                                                       'links': [       {       'in_nucast_pkts': '77',
                                                                                                'in_octets': '3626593535230',
                                                                                                'in_ucast_pkts': '2604983567',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '99454265911',
                                                                                                'out_ucast_pkts': '870693132',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '99454265911',
                                                                       'out_ucast_pkts': '870693132',
                                                                       'poller_hash': 'RENGNEMxcGVlcmludGZwb24gMTg2Nzc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '186776',
                                                                                               'links': [       {       'connected_time': '8535156',
                                                                                                                        'fiber_length': 24426,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'RENGNEMxcGVlcnBvbnBvbiAxODY3NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '90 PARKVIEW DR WELLESLEY (TEMPLETON '
                                                 'JOANN)',
                                         'poller_hash': 'RENGNEMx',
                                         'serial_number': 'DCF4C1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:b3:7d:75',
                                         'uptime': '8535156'},
                                 {       'counter_source': '189899/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1d:2d'],
                                                                       'id': '189899/g1',
                                                                       'in_nucast_pkts': '154306',
                                                                       'in_octets': '37913603571',
                                                                       'in_ucast_pkts': '207706616',
                                                                       'mac_address': '84:d3:43:7e:34:ea',
                                                                       'mtu': 2000,
                                                                       'name': '189899/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878275',
                                                                       'out_octets': '1086547727912',
                                                                       'out_ucast_pkts': '767868641',
                                                                       'poller_hash': 'QzgzQzA2cGVlcmludGYxODk4OTkvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '189899',
                                                                       'in_nucast_pkts': '9878275',
                                                                       'in_octets': '1086547727912',
                                                                       'in_ucast_pkts': '767868641',
                                                                       'links': [       {       'in_nucast_pkts': '9878275',
                                                                                                'in_octets': '1086547727912',
                                                                                                'in_ucast_pkts': '767868641',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '154306',
                                                                                                'out_octets': '37913603571',
                                                                                                'out_ucast_pkts': '207706616',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '154306',
                                                                       'out_octets': '37913603571',
                                                                       'out_ucast_pkts': '207706616',
                                                                       'poller_hash': 'QzgzQzA2cGVlcmludGZwb24gMTg5ODk5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '189899',
                                                                                               'links': [       {       'connected_time': '8535142',
                                                                                                                        'fiber_length': 25188,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQzA2cGVlcnBvbnBvbiAxODk4OTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '20 WELWOOD AV (SEIP TERRY)',
                                         'poller_hash': 'QzgzQzA2',
                                         'serial_number': 'C83C06',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:34:e7',
                                         'uptime': '8535142'},
                                 {       'counter_source': '193297/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1e:6c'],
                                                                       'id': '193297/g1',
                                                                       'in_nucast_pkts': '153947',
                                                                       'in_octets': '63334092710',
                                                                       'in_ucast_pkts': '160454075',
                                                                       'mac_address': '84:d3:43:7d:de:42',
                                                                       'mtu': 2000,
                                                                       'name': '193297/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919015',
                                                                       'out_octets': '702779231618',
                                                                       'out_ucast_pkts': '581370077',
                                                                       'poller_hash': 'QzgzMTMxcGVlcmludGYxOTMyOTcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '193297',
                                                                       'in_nucast_pkts': '9919015',
                                                                       'in_octets': '702779231618',
                                                                       'in_ucast_pkts': '581370077',
                                                                       'links': [       {       'in_nucast_pkts': '9919015',
                                                                                                'in_octets': '702779231618',
                                                                                                'in_ucast_pkts': '581370077',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '153947',
                                                                                                'out_octets': '63334092710',
                                                                                                'out_ucast_pkts': '160454075',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '153947',
                                                                       'out_octets': '63334092710',
                                                                       'out_ucast_pkts': '160454075',
                                                                       'poller_hash': 'QzgzMTMxcGVlcmludGZwb24gMTkzMjk3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '193297',
                                                                                               'links': [       {       'connected_time': '8535226',
                                                                                                                        'fiber_length': 22187,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzMTMxcGVlcnBvbnBvbiAxOTMyOTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1294 QUEEN'S BUSH ROAD (WINKLER "
                                                 'MESLISSA)',
                                         'poller_hash': 'QzgzMTMx',
                                         'serial_number': 'C83131',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:de:3f',
                                         'uptime': '8535226'},
                                 {       'counter_source': '194124/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:38:d9'],
                                                                       'id': '194124/g1',
                                                                       'in_nucast_pkts': '30711125',
                                                                       'in_octets': '415956009613',
                                                                       'in_ucast_pkts': '1748540448',
                                                                       'mac_address': '84:d3:43:7e:35:92',
                                                                       'mtu': 2000,
                                                                       'name': '194124/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68425976006',
                                                                       'out_octets': '99755122707540',
                                                                       'out_ucast_pkts': '5934949459',
                                                                       'poller_hash': 'QzgzQzFCcGVlcmludGYxOTQxMjQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '194124',
                                                                       'in_nucast_pkts': '68425976006',
                                                                       'in_octets': '99755122707540',
                                                                       'in_ucast_pkts': '5934949459',
                                                                       'links': [       {       'in_nucast_pkts': '68425976006',
                                                                                                'in_octets': '99755122707540',
                                                                                                'in_ucast_pkts': '5934949459',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '30711125',
                                                                                                'out_octets': '415956009613',
                                                                                                'out_ucast_pkts': '1748540448',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '30711125',
                                                                       'out_octets': '415956009613',
                                                                       'out_ucast_pkts': '1748540448',
                                                                       'poller_hash': 'QzgzQzFCcGVlcmludGZwb24gMTk0MTI0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '194124',
                                                                                               'links': [       {       'connected_time': '21717948',
                                                                                                                        'fiber_length': 19434,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgzQzFCcGVlcnBvbnBvbiAxOTQxMjQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4005 FREDERICK ST (SKINNER SCOTT)',
                                         'poller_hash': 'QzgzQzFC',
                                         'serial_number': 'C83C1B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7e:35:8f',
                                         'uptime': '21717948'},
                                 {       'counter_source': '196234/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:0f:b9:b2'],
                                                                       'id': '196234/g1',
                                                                       'in_nucast_pkts': '54874',
                                                                       'in_octets': '58492491491',
                                                                       'in_ucast_pkts': '148063402',
                                                                       'mac_address': '84:d3:43:7d:42:d2',
                                                                       'mtu': 2000,
                                                                       'name': '196234/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3212341',
                                                                       'out_octets': '467658485197',
                                                                       'out_ucast_pkts': '353470255',
                                                                       'poller_hash': 'QzgxREMzcGVlcmludGYxOTYyMzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '196234',
                                                                       'in_nucast_pkts': '3212341',
                                                                       'in_octets': '467658485197',
                                                                       'in_ucast_pkts': '353470255',
                                                                       'links': [       {       'in_nucast_pkts': '3212341',
                                                                                                'in_octets': '467658485197',
                                                                                                'in_ucast_pkts': '353470255',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '54874',
                                                                                                'out_octets': '58492491491',
                                                                                                'out_ucast_pkts': '148063402',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '54874',
                                                                       'out_octets': '58492491491',
                                                                       'out_ucast_pkts': '148063402',
                                                                       'poller_hash': 'QzgxREMzcGVlcmludGZwb24gMTk2MjM0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '196234',
                                                                                               'links': [       {       'connected_time': '2741850',
                                                                                                                        'fiber_length': 22532,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREMzcGVlcnBvbnBvbiAxOTYyMzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1220 QUEEN'S BUSH RD (KUEPFER CRAIG)",
                                         'poller_hash': 'QzgxREMz',
                                         'serial_number': 'C81DC3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:42:cf',
                                         'uptime': '2741850'},
                                 {       'counter_source': '197716/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:e9:bc'],
                                                                       'id': '197716/g1',
                                                                       'in_nucast_pkts': '35245',
                                                                       'in_octets': '25146455229',
                                                                       'in_ucast_pkts': '83223631',
                                                                       'mac_address': '84:d3:43:7d:47:1a',
                                                                       'mtu': 2000,
                                                                       'name': '197716/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1936295',
                                                                       'out_octets': '259911461116',
                                                                       'out_ucast_pkts': '209416686',
                                                                       'poller_hash': 'QzgxRTRDcGVlcmludGYxOTc3MTYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '197716',
                                                                       'in_nucast_pkts': '1936295',
                                                                       'in_octets': '259911461116',
                                                                       'in_ucast_pkts': '209416686',
                                                                       'links': [       {       'in_nucast_pkts': '1936295',
                                                                                                'in_octets': '259911461116',
                                                                                                'in_ucast_pkts': '209416686',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '35245',
                                                                                                'out_octets': '25146455229',
                                                                                                'out_ucast_pkts': '83223631',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '35245',
                                                                       'out_octets': '25146455229',
                                                                       'out_ucast_pkts': '83223631',
                                                                       'poller_hash': 'QzgxRTRDcGVlcmludGZwb24gMTk3NzE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '197716',
                                                                                               'links': [       {       'connected_time': '1648526',
                                                                                                                        'fiber_length': 22817,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTRDcGVlcnBvbnBvbiAxOTc3MTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2 LAWRENCE STREET (BARKER ANGELA)',
                                         'poller_hash': 'QzgxRTRD',
                                         'serial_number': 'C81E4C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:47:17',
                                         'uptime': '1648526'},
                                 {       'counter_source': '198028/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:08:ff:d4'],
                                                                       'id': '198028/g1',
                                                                       'in_nucast_pkts': '161781',
                                                                       'in_octets': '57334070549',
                                                                       'in_ucast_pkts': '240080486',
                                                                       'mac_address': '84:d3:43:7d:4c:12',
                                                                       'mtu': 2000,
                                                                       'name': '198028/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797949',
                                                                       'out_octets': '953269362789',
                                                                       'out_ucast_pkts': '713977350',
                                                                       'poller_hash': 'QzgxRUVCcGVlcmludGYxOTgwMjgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '198028',
                                                                       'in_nucast_pkts': '9797949',
                                                                       'in_octets': '953269362789',
                                                                       'in_ucast_pkts': '713977350',
                                                                       'links': [       {       'in_nucast_pkts': '9797949',
                                                                                                'in_octets': '953269362789',
                                                                                                'in_ucast_pkts': '713977350',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '161781',
                                                                                                'out_octets': '57334070549',
                                                                                                'out_ucast_pkts': '240080486',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '161781',
                                                                       'out_octets': '57334070549',
                                                                       'out_ucast_pkts': '240080486',
                                                                       'poller_hash': 'QzgxRUVCcGVlcmludGZwb24gMTk4MDI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '198028',
                                                                                               'links': [       {       'connected_time': '8535334',
                                                                                                                        'fiber_length': 24973,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUVCcGVlcnBvbnBvbiAxOTgwMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '116 BROWN STREET (LICHTY SARAH)',
                                         'poller_hash': 'QzgxRUVC',
                                         'serial_number': 'C81EEB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4c:0f',
                                         'uptime': '8535334'},
                                 {       'counter_source': '202645/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '202645/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96796616546',
                                                                       'in_ucast_pkts': '470772917',
                                                                       'mac_address': '84:d3:43:7d:4a:0f',
                                                                       'name': '202645/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '38',
                                                                       'out_octets': '3304017872402',
                                                                       'out_ucast_pkts': '2365241708',
                                                                       'poller_hash': 'QzgxRUFCcGVlcmludGYyMDI2NDUvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '202645',
                                                                       'in_nucast_pkts': '38',
                                                                       'in_octets': '3304017872402',
                                                                       'in_ucast_pkts': '2365241708',
                                                                       'links': [       {       'in_nucast_pkts': '38',
                                                                                                'in_octets': '3304017872402',
                                                                                                'in_ucast_pkts': '2365241708',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96796616546',
                                                                                                'out_ucast_pkts': '470772917',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96796616546',
                                                                       'out_ucast_pkts': '470772917',
                                                                       'poller_hash': 'QzgxRUFCcGVlcmludGZwb24gMjAyNjQ1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202645',
                                                                                               'links': [       {       'connected_time': '8535276',
                                                                                                                        'fiber_length': 23449,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUFCcGVlcnBvbnBvbiAyMDI2NDU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '162 LAWRENCE ST (CAMPBELL PAYNE '
                                                 'MORGAN)',
                                         'poller_hash': 'QzgxRUFC',
                                         'serial_number': 'C81EAB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4a:0f',
                                         'uptime': '8535276'},
                                 {       'counter_source': '202774/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:22:32:96:32:86'],
                                                                       'id': '202774/g1',
                                                                       'in_nucast_pkts': '76465',
                                                                       'in_octets': '118857771353',
                                                                       'in_ucast_pkts': '417007415',
                                                                       'mac_address': '84:d3:43:7d:4b:f2',
                                                                       'mtu': 2000,
                                                                       'name': '202774/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920204',
                                                                       'out_octets': '2672911696485',
                                                                       'out_ucast_pkts': '1939229032',
                                                                       'poller_hash': 'QzgxRUU3cGVlcmludGYyMDI3NzQvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '202774',
                                                                       'in_nucast_pkts': '9920204',
                                                                       'in_octets': '2672911696485',
                                                                       'in_ucast_pkts': '1939229032',
                                                                       'links': [       {       'in_nucast_pkts': '9920204',
                                                                                                'in_octets': '2672911696485',
                                                                                                'in_ucast_pkts': '1939229032',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '76465',
                                                                                                'out_octets': '118857771353',
                                                                                                'out_ucast_pkts': '417007415',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76465',
                                                                       'out_octets': '118857771353',
                                                                       'out_ucast_pkts': '417007415',
                                                                       'poller_hash': 'QzgxRUU3cGVlcmludGZwb24gMjAyNzc0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '202774',
                                                                                               'links': [       {       'connected_time': '8535278',
                                                                                                                        'fiber_length': 24476,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRUU3cGVlcnBvbnBvbiAyMDI3NzQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '83 Parkview Dr Wellesley (Dave '
                                                 'Wintrip 10434356)',
                                         'poller_hash': 'QzgxRUU3',
                                         'serial_number': 'C81EE7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4b:ef',
                                         'uptime': '8535278'},
                                 {       'counter_source': '204915/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9d:ae:9d'],
                                                                       'id': '204915/g1',
                                                                       'in_nucast_pkts': '71834',
                                                                       'in_octets': '87344712806',
                                                                       'in_ucast_pkts': '418632755',
                                                                       'mac_address': '84:d3:43:7d:45:82',
                                                                       'mtu': 2000,
                                                                       'name': '204915/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4686154',
                                                                       'out_octets': '2215615976960',
                                                                       'out_ucast_pkts': '1596773593',
                                                                       'poller_hash': 'QzgxRTE5cGVlcmludGYyMDQ5MTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '204915',
                                                                       'in_nucast_pkts': '4686154',
                                                                       'in_octets': '2215615976960',
                                                                       'in_ucast_pkts': '1596773593',
                                                                       'links': [       {       'in_nucast_pkts': '4686154',
                                                                                                'in_octets': '2215615976960',
                                                                                                'in_ucast_pkts': '1596773593',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '71834',
                                                                                                'out_octets': '87344712806',
                                                                                                'out_ucast_pkts': '418632755',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '71834',
                                                                       'out_octets': '87344712806',
                                                                       'out_ucast_pkts': '418632755',
                                                                       'poller_hash': 'QzgxRTE5cGVlcmludGZwb24gMjA0OTE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '204915',
                                                                                               'links': [       {       'connected_time': '3982702',
                                                                                                                        'fiber_length': 23289,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTE5cGVlcnBvbnBvbiAyMDQ5MTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3835 NAFZIGER RD (JANZEN PETER)',
                                         'poller_hash': 'QzgxRTE5',
                                         'serial_number': 'C81E19',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:45:7f',
                                         'uptime': '3982702'},
                                 {       'counter_source': '205130/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '205130/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '62268331893',
                                                                       'in_ucast_pkts': '275381128',
                                                                       'mac_address': 'ec:4f:82:8d:52:38',
                                                                       'name': '205130/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '106',
                                                                       'out_octets': '1911187515487',
                                                                       'out_ucast_pkts': '1404163381',
                                                                       'poller_hash': 'MzIxMjhFcGVlcmludGYyMDUxMzAvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205130',
                                                                       'in_nucast_pkts': '106',
                                                                       'in_octets': '1911187515487',
                                                                       'in_ucast_pkts': '1404163381',
                                                                       'links': [       {       'in_nucast_pkts': '106',
                                                                                                'in_octets': '1911187515487',
                                                                                                'in_ucast_pkts': '1404163381',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '62268331893',
                                                                                                'out_ucast_pkts': '275381128',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '62268331893',
                                                                       'out_ucast_pkts': '275381128',
                                                                       'poller_hash': 'MzIxMjhFcGVlcmludGZwb24gMjA1MTMw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205130',
                                                                                               'links': [       {       'connected_time': '21887149',
                                                                                                                        'fiber_length': 11943,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMjhFcGVlcnBvbnBvbiAyMDUxMzA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5586 PERTH LINE 56A (DAN ZEHR '
                                                 '10400068)',
                                         'poller_hash': 'MzIxMjhF',
                                         'serial_number': '32128E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:52:38',
                                         'uptime': '21887149'},
                                 {       'counter_source': '205190/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:7f:5c'],
                                                                       'id': '205190/g1',
                                                                       'in_nucast_pkts': '240815',
                                                                       'in_octets': '194360415892',
                                                                       'in_ucast_pkts': '691031496',
                                                                       'mac_address': '84:d3:43:7d:4d:5a',
                                                                       'mtu': 2000,
                                                                       'name': '205190/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '17793542',
                                                                       'out_octets': '2585665774200',
                                                                       'out_ucast_pkts': '2058832485',
                                                                       'poller_hash': 'QzgxRjE0cGVlcmludGYyMDUxOTAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205190',
                                                                       'in_nucast_pkts': '17793542',
                                                                       'in_octets': '2585665774200',
                                                                       'in_ucast_pkts': '2058832485',
                                                                       'links': [       {       'in_nucast_pkts': '17793542',
                                                                                                'in_octets': '2585665774200',
                                                                                                'in_ucast_pkts': '2058832485',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '240815',
                                                                                                'out_octets': '194360415892',
                                                                                                'out_ucast_pkts': '691031496',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '240815',
                                                                       'out_octets': '194360415892',
                                                                       'out_ucast_pkts': '691031496',
                                                                       'poller_hash': 'QzgxRjE0cGVlcmludGZwb24gMjA1MTkw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205190',
                                                                                               'links': [       {       'connected_time': '15369455',
                                                                                                                        'fiber_length': 13784,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRjE0cGVlcnBvbnBvbiAyMDUxOTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2343 PERTH LINE 56 (R&G DEEN FARMS '
                                                 'LTD)',
                                         'poller_hash': 'QzgxRjE0',
                                         'serial_number': 'C81F14',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:4d:57',
                                         'uptime': '15369455'},
                                 {       'counter_source': '205763/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9d:ae:3a'],
                                                                       'id': '205763/g1',
                                                                       'in_nucast_pkts': '13421',
                                                                       'in_octets': '6232943204',
                                                                       'in_ucast_pkts': '20374189',
                                                                       'mac_address': '84:d3:43:7d:41:ba',
                                                                       'mtu': 2000,
                                                                       'name': '205763/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715434',
                                                                       'out_octets': '106947127022',
                                                                       'out_ucast_pkts': '87351713',
                                                                       'poller_hash': 'QzgxREEwcGVlcmludGYyMDU3NjMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '205763',
                                                                       'in_nucast_pkts': '715434',
                                                                       'in_octets': '106947127022',
                                                                       'in_ucast_pkts': '87351713',
                                                                       'links': [       {       'in_nucast_pkts': '715434',
                                                                                                'in_octets': '106947127022',
                                                                                                'in_ucast_pkts': '87351713',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '13421',
                                                                                                'out_octets': '6232943204',
                                                                                                'out_ucast_pkts': '20374189',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '13421',
                                                                       'out_octets': '6232943204',
                                                                       'out_ucast_pkts': '20374189',
                                                                       'poller_hash': 'QzgxREEwcGVlcmludGZwb24gMjA1NzYz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205763',
                                                                                               'links': [       {       'connected_time': '611446',
                                                                                                                        'fiber_length': 20071,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREEwcGVlcnBvbnBvbiAyMDU3NjM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2827 LINE 45 (VANZANDWIJK GARY)',
                                         'poller_hash': 'QzgxREEw',
                                         'serial_number': 'C81DA0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:41:b7',
                                         'uptime': '611446'},
                                 {       'counter_source': '205773/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '205773/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '225182649028',
                                                                       'in_ucast_pkts': '1966696189',
                                                                       'mac_address': '84:d3:43:7d:38:77',
                                                                       'name': '205773/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '79',
                                                                       'out_octets': '6008347335132',
                                                                       'out_ucast_pkts': '4192560610',
                                                                       'poller_hash': 'QzgxQzc4cGVlcmludGYyMDU3NzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '205773',
                                                                       'in_nucast_pkts': '79',
                                                                       'in_octets': '6008347335132',
                                                                       'in_ucast_pkts': '4192560610',
                                                                       'links': [       {       'in_nucast_pkts': '79',
                                                                                                'in_octets': '6008347335132',
                                                                                                'in_ucast_pkts': '4192560610',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '225182649028',
                                                                                                'out_ucast_pkts': '1966696189',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '225182649028',
                                                                       'out_ucast_pkts': '1966696189',
                                                                       'poller_hash': 'QzgxQzc4cGVlcmludGZwb24gMjA1Nzcz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '205773',
                                                                                               'links': [       {       'connected_time': '335337',
                                                                                                                        'fiber_length': 17508,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxQzc4cGVlcnBvbnBvbiAyMDU3NzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2073 ALBERT ST (ERICKSON KAREN)',
                                         'poller_hash': 'QzgxQzc4',
                                         'serial_number': 'C81C78',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:38:77',
                                         'uptime': '335337'},
                                 {       'counter_source': '206337/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:b2:60'],
                                                                       'id': '206337/g1',
                                                                       'in_nucast_pkts': '112684',
                                                                       'in_octets': '72461138623',
                                                                       'in_ucast_pkts': '213030268',
                                                                       'mac_address': '84:d3:43:7d:41:d2',
                                                                       'mtu': 2000,
                                                                       'name': '206337/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7429023',
                                                                       'out_octets': '835005650588',
                                                                       'out_ucast_pkts': '696908710',
                                                                       'poller_hash': 'QzgxREEzcGVlcmludGYyMDYzMzcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '206337',
                                                                       'in_nucast_pkts': '7429023',
                                                                       'in_octets': '835005650588',
                                                                       'in_ucast_pkts': '696908710',
                                                                       'links': [       {       'in_nucast_pkts': '7429023',
                                                                                                'in_octets': '835005650588',
                                                                                                'in_ucast_pkts': '696908710',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '112684',
                                                                                                'out_octets': '72461138623',
                                                                                                'out_ucast_pkts': '213030268',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '112684',
                                                                       'out_octets': '72461138623',
                                                                       'out_ucast_pkts': '213030268',
                                                                       'poller_hash': 'QzgxREEzcGVlcmludGZwb24gMjA2MzM3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '206337',
                                                                                               'links': [       {       'connected_time': '6460052',
                                                                                                                        'fiber_length': 24042,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREEzcGVlcnBvbnBvbiAyMDYzMzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '112 REINER CR (DA SILVA HUGO)',
                                         'poller_hash': 'QzgxREEz',
                                         'serial_number': 'C81DA3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:41:cf',
                                         'uptime': '6460052'},
                                 {       'counter_source': '207667/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:bc:ba:7f'],
                                                                       'id': '207667/g1',
                                                                       'in_nucast_pkts': '952934',
                                                                       'in_octets': '275184349711',
                                                                       'in_ucast_pkts': '2349447161',
                                                                       'mac_address': '84:d3:43:7d:44:ba',
                                                                       'mtu': 2000,
                                                                       'name': '207667/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '63820781',
                                                                       'out_octets': '7898817808653',
                                                                       'out_ucast_pkts': '5703075585',
                                                                       'poller_hash': 'QzgxRTAwcGVlcmludGYyMDc2NjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '207667',
                                                                       'in_nucast_pkts': '63820781',
                                                                       'in_octets': '7898817808653',
                                                                       'in_ucast_pkts': '5703075585',
                                                                       'links': [       {       'in_nucast_pkts': '63820781',
                                                                                                'in_octets': '7898817808653',
                                                                                                'in_ucast_pkts': '5703075585',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '952934',
                                                                                                'out_octets': '275184349711',
                                                                                                'out_ucast_pkts': '2349447161',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '952934',
                                                                       'out_octets': '275184349711',
                                                                       'out_ucast_pkts': '2349447161',
                                                                       'poller_hash': 'QzgxRTAwcGVlcmludGZwb24gMjA3NjY3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207667',
                                                                                               'links': [       {       'connected_time': '21886946',
                                                                                                                        'fiber_length': 24607,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTAwcGVlcnBvbnBvbiAyMDc2Njc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '26 PARKVIEW DR (MYERS BOB)',
                                         'poller_hash': 'QzgxRTAw',
                                         'serial_number': 'C81E00',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:b7',
                                         'uptime': '21886946'},
                                 {       'counter_source': '207732/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '207732/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '256543344533',
                                                                       'in_ucast_pkts': '1391881398',
                                                                       'mac_address': '84:d3:43:7c:ff:cf',
                                                                       'name': '207732/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '133',
                                                                       'out_octets': '7093871491165',
                                                                       'out_ucast_pkts': '5037474535',
                                                                       'poller_hash': 'QzgxNTYzcGVlcmludGYyMDc3MzIvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '207732',
                                                                       'in_nucast_pkts': '133',
                                                                       'in_octets': '7093871491165',
                                                                       'in_ucast_pkts': '5037474535',
                                                                       'links': [       {       'in_nucast_pkts': '133',
                                                                                                'in_octets': '7093871491165',
                                                                                                'in_ucast_pkts': '5037474535',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '256543344533',
                                                                                                'out_ucast_pkts': '1391881398',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '256543344533',
                                                                       'out_ucast_pkts': '1391881398',
                                                                       'poller_hash': 'QzgxNTYzcGVlcmludGZwb24gMjA3NzMy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207732',
                                                                                               'links': [       {       'connected_time': '21886924',
                                                                                                                        'fiber_length': 23260,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxNTYzcGVlcnBvbnBvbiAyMDc3MzI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '167 FERRIS DR (MITCHELL DAVID)',
                                         'poller_hash': 'QzgxNTYz',
                                         'serial_number': 'C81563',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7c:ff:cf',
                                         'uptime': '21886924'},
                                 {       'counter_source': '207820/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:1e:98'],
                                                                       'id': '207820/g1',
                                                                       'in_nucast_pkts': '158177',
                                                                       'in_octets': '170532885463',
                                                                       'in_ucast_pkts': '816568366',
                                                                       'mac_address': '84:d3:43:7d:45:22',
                                                                       'mtu': 2000,
                                                                       'name': '207820/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797919',
                                                                       'out_octets': '3910508214580',
                                                                       'out_ucast_pkts': '2935288591',
                                                                       'poller_hash': 'QzgxRTBEcGVlcmludGYyMDc4MjAvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '207820',
                                                                       'in_nucast_pkts': '9797919',
                                                                       'in_octets': '3910508214580',
                                                                       'in_ucast_pkts': '2935288591',
                                                                       'links': [       {       'in_nucast_pkts': '9797919',
                                                                                                'in_octets': '3910508214580',
                                                                                                'in_ucast_pkts': '2935288591',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '158177',
                                                                                                'out_octets': '170532885463',
                                                                                                'out_ucast_pkts': '816568366',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '158177',
                                                                       'out_octets': '170532885463',
                                                                       'out_ucast_pkts': '816568366',
                                                                       'poller_hash': 'QzgxRTBEcGVlcmludGZwb24gMjA3ODIw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '207820',
                                                                                               'links': [       {       'connected_time': '8535287',
                                                                                                                        'fiber_length': 24757,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTBEcGVlcnBvbnBvbiAyMDc4MjA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1134 QUEEN'S BUSH RD (SCHOTT JEFF)",
                                         'poller_hash': 'QzgxRTBE',
                                         'serial_number': 'C81E0D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:45:1f',
                                         'uptime': '8535287'},
                                 {       'counter_source': '208678/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:b7:bf'],
                                                                       'id': '208678/g1',
                                                                       'in_nucast_pkts': '68761',
                                                                       'in_octets': '21557579690',
                                                                       'in_ucast_pkts': '135012854',
                                                                       'mac_address': '84:d3:43:7d:31:c2',
                                                                       'mtu': 2000,
                                                                       'name': '208678/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4741761',
                                                                       'out_octets': '432485975138',
                                                                       'out_ucast_pkts': '305460690',
                                                                       'poller_hash': 'QzgxQkExcGVlcmludGYyMDg2NzgvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '208678',
                                                                       'in_nucast_pkts': '4741761',
                                                                       'in_octets': '432485975138',
                                                                       'in_ucast_pkts': '305460690',
                                                                       'links': [       {       'in_nucast_pkts': '4741761',
                                                                                                'in_octets': '432485975138',
                                                                                                'in_ucast_pkts': '305460690',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '68761',
                                                                                                'out_octets': '21557579690',
                                                                                                'out_ucast_pkts': '135012854',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '68761',
                                                                       'out_octets': '21557579690',
                                                                       'out_ucast_pkts': '135012854',
                                                                       'poller_hash': 'QzgxQkExcGVlcmludGZwb24gMjA4Njc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208678',
                                                                                               'links': [       {       'connected_time': '4031948',
                                                                                                                        'fiber_length': 25593,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxQkExcGVlcnBvbnBvbiAyMDg2Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '68 WELWOOD AV (SCHMIDT DARLENE)',
                                         'poller_hash': 'QzgxQkEx',
                                         'serial_number': 'C81BA1',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:31:bf',
                                         'uptime': '4031948'},
                                 {       'counter_source': '208933/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '208933/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '4213274133844',
                                                                       'in_ucast_pkts': '1949276403630',
                                                                       'mac_address': '84:d3:43:7d:43:af',
                                                                       'name': '208933/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '84',
                                                                       'out_octets': '3709163404360',
                                                                       'out_ucast_pkts': '4161446371',
                                                                       'poller_hash': 'QzgxRERGcGVlcmludGYyMDg5MzMvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '208933',
                                                                       'in_nucast_pkts': '84',
                                                                       'in_octets': '3709163404360',
                                                                       'in_ucast_pkts': '4161446371',
                                                                       'links': [       {       'in_nucast_pkts': '84',
                                                                                                'in_octets': '3709163404360',
                                                                                                'in_ucast_pkts': '4161446371',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '4213274133844',
                                                                                                'out_ucast_pkts': '1949276403630',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '4213274133844',
                                                                       'out_ucast_pkts': '1949276403630',
                                                                       'poller_hash': 'QzgxRERGcGVlcmludGZwb24gMjA4OTMz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208933',
                                                                                               'links': [       {       'connected_time': '13271249',
                                                                                                                        'fiber_length': 31458,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRERGcGVlcnBvbnBvbiAyMDg5MzM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4026 WEIMAR LINE (BROWN ALEXANDER)',
                                         'poller_hash': 'QzgxRERG',
                                         'serial_number': 'C81DDF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:43:af',
                                         'uptime': '13271249'},
                                 {       'counter_source': '208976/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cb:8e'],
                                                                       'id': '208976/g1',
                                                                       'in_nucast_pkts': '57763904',
                                                                       'in_octets': '104186266757',
                                                                       'in_ucast_pkts': '622096314',
                                                                       'mac_address': '84:d3:43:7d:14:aa',
                                                                       'mtu': 2000,
                                                                       'name': '208976/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '62552116191',
                                                                       'out_octets': '88470814115337',
                                                                       'out_ucast_pkts': '3226307595',
                                                                       'poller_hash': 'QzgxN0ZFcGVlcmludGYyMDg5NzYvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '208976',
                                                                       'in_nucast_pkts': '62552116191',
                                                                       'in_octets': '88470814115337',
                                                                       'in_ucast_pkts': '3226307595',
                                                                       'links': [       {       'in_nucast_pkts': '62552116191',
                                                                                                'in_octets': '88470814115337',
                                                                                                'in_ucast_pkts': '3226307595',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '57763904',
                                                                                                'out_octets': '104186266757',
                                                                                                'out_ucast_pkts': '622096314',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '57763904',
                                                                       'out_octets': '104186266757',
                                                                       'out_ucast_pkts': '622096314',
                                                                       'poller_hash': 'QzgxN0ZFcGVlcmludGZwb24gMjA4OTc2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208976',
                                                                                               'links': [       {       'connected_time': '21886400',
                                                                                                                        'fiber_length': 23178,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxN0ZFcGVlcnBvbnBvbiAyMDg5NzY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 POND VIEW DR (LICHTY NORMA)',
                                         'poller_hash': 'QzgxN0ZF',
                                         'serial_number': 'C817FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:14:a7',
                                         'uptime': '21886400'},
                                 {       'counter_source': '208996/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '208996/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '16957581619',
                                                                       'in_ucast_pkts': '47754057',
                                                                       'mac_address': '84:d3:43:7d:44:f7',
                                                                       'name': '208996/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '103',
                                                                       'out_octets': '124915731268',
                                                                       'out_ucast_pkts': '127872760',
                                                                       'poller_hash': 'QzgxRTA4cGVlcmludGYyMDg5OTYvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '208996',
                                                                       'in_nucast_pkts': '103',
                                                                       'in_octets': '124915731268',
                                                                       'in_ucast_pkts': '127872760',
                                                                       'links': [       {       'in_nucast_pkts': '103',
                                                                                                'in_octets': '124915731268',
                                                                                                'in_ucast_pkts': '127872760',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '16957581619',
                                                                                                'out_ucast_pkts': '47754057',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '16957581619',
                                                                       'out_ucast_pkts': '47754057',
                                                                       'poller_hash': 'QzgxRTA4cGVlcmludGZwb24gMjA4OTk2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '208996',
                                                                                               'links': [       {       'connected_time': '8535248',
                                                                                                                        'fiber_length': 23892,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxRTA4cGVlcnBvbnBvbiAyMDg5OTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3706 NAFZIGER RD (JENNIFER PITSCH '
                                                 'RMT)',
                                         'poller_hash': 'QzgxRTA4',
                                         'serial_number': 'C81E08',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:44:f7',
                                         'uptime': '8535248'},
                                 {       'counter_source': '210026/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cb:f1'],
                                                                       'id': '210026/g2',
                                                                       'in_nucast_pkts': '10647',
                                                                       'in_octets': '19860663988',
                                                                       'in_ucast_pkts': '85409837',
                                                                       'mac_address': '84:d3:43:7d:43:db',
                                                                       'mtu': 2000,
                                                                       'name': '210026/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715489',
                                                                       'out_octets': '268837080250',
                                                                       'out_ucast_pkts': '237178860',
                                                                       'poller_hash': 'QzgxREU0cGVlcmludGYyMTAwMjYvZzI=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '210026',
                                                                       'in_nucast_pkts': '715489',
                                                                       'in_octets': '268837080250',
                                                                       'in_ucast_pkts': '237178860',
                                                                       'links': [       {       'in_nucast_pkts': '715489',
                                                                                                'in_octets': '268837080250',
                                                                                                'in_ucast_pkts': '237178860',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10647',
                                                                                                'out_octets': '19860663988',
                                                                                                'out_ucast_pkts': '85409837',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10647',
                                                                       'out_octets': '19860663988',
                                                                       'out_ucast_pkts': '85409837',
                                                                       'poller_hash': 'QzgxREU0cGVlcmludGZwb24gMjEwMDI2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '210026',
                                                                                               'links': [       {       'connected_time': '611459',
                                                                                                                        'fiber_length': 20008,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QzgxREU0cGVlcnBvbnBvbiAyMTAwMjY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2830 LINE 45 (SZUSZ RYAN)',
                                         'poller_hash': 'QzgxREU0',
                                         'serial_number': 'C81DE4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '84:d3:43:7d:43:d7',
                                         'uptime': '611459'},
                                 {       'counter_source': '212916/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '212916/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '58472040479',
                                                                       'in_ucast_pkts': '327474330',
                                                                       'mac_address': 'b8:94:70:ba:e1:65',
                                                                       'name': '212916/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1087824903634',
                                                                       'out_ucast_pkts': '812411350',
                                                                       'poller_hash': 'MTE5NTE0MXBlZXJpbnRmMjEyOTE2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '212916',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1087824903634',
                                                                       'in_ucast_pkts': '812411350',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1087824903634',
                                                                                                'in_ucast_pkts': '812411350',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '58472040479',
                                                                                                'out_ucast_pkts': '327474330',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '58472040479',
                                                                       'out_ucast_pkts': '327474330',
                                                                       'poller_hash': 'MTE5NTE0MXBlZXJpbnRmcG9uIDIxMjkxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212916',
                                                                                               'links': [       {       'connected_time': '13292304',
                                                                                                                        'fiber_length': 22732,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE0MXBlZXJwb25wb24gMjEyOTE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1193 QUEEN'S BUSH RD (1001090591 "
                                                 'ONTARIO INC)',
                                         'poller_hash': 'MTE5NTE0MQ==',
                                         'serial_number': '1195141',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e1:65',
                                         'uptime': '13292304'},
                                 {       'counter_source': '212979/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:91:30'],
                                                                       'id': '212979/g1',
                                                                       'in_nucast_pkts': '155179',
                                                                       'in_octets': '11327630304',
                                                                       'in_ucast_pkts': '41524215',
                                                                       'mac_address': 'b8:94:70:ba:e2:20',
                                                                       'mtu': 2000,
                                                                       'name': '212979/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919972',
                                                                       'out_octets': '111189271329',
                                                                       'out_ucast_pkts': '85033603',
                                                                       'poller_hash': 'MTE5NTE1OHBlZXJpbnRmMjEyOTc5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '212979',
                                                                       'in_nucast_pkts': '9919972',
                                                                       'in_octets': '111189271329',
                                                                       'in_ucast_pkts': '85033603',
                                                                       'links': [       {       'in_nucast_pkts': '9919972',
                                                                                                'in_octets': '111189271329',
                                                                                                'in_ucast_pkts': '85033603',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '155179',
                                                                                                'out_octets': '11327630304',
                                                                                                'out_ucast_pkts': '41524215',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '155179',
                                                                       'out_octets': '11327630304',
                                                                       'out_ucast_pkts': '41524215',
                                                                       'poller_hash': 'MTE5NTE1OHBlZXJpbnRmcG9uIDIxMjk3OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '212979',
                                                                                               'links': [       {       'connected_time': '8535268',
                                                                                                                        'fiber_length': 25386,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE1OHBlZXJwb25wb24gMjEyOTc5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '111 VILLAGE RD (BELL NANCY)',
                                         'poller_hash': 'MTE5NTE1OA==',
                                         'serial_number': '1195158',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e2:1d',
                                         'uptime': '8535268'},
                                 {       'counter_source': '215895/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:90:ac'],
                                                                       'id': '215895/g1',
                                                                       'in_nucast_pkts': '674564',
                                                                       'in_octets': '21114128215',
                                                                       'in_ucast_pkts': '49347842',
                                                                       'mac_address': 'ec:4f:82:59:ec:be',
                                                                       'mtu': 2000,
                                                                       'name': '215895/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '38757385',
                                                                       'out_octets': '82328627875',
                                                                       'out_ucast_pkts': '92105554',
                                                                       'poller_hash': 'MkFEMTBBcGVlcmludGYyMTU4OTUvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '215895',
                                                                       'in_nucast_pkts': '38757385',
                                                                       'in_octets': '82328627875',
                                                                       'in_ucast_pkts': '92105554',
                                                                       'links': [       {       'in_nucast_pkts': '38757385',
                                                                                                'in_octets': '82328627875',
                                                                                                'in_ucast_pkts': '92105554',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '674564',
                                                                                                'out_octets': '21114128215',
                                                                                                'out_ucast_pkts': '49347842',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '674564',
                                                                       'out_octets': '21114128215',
                                                                       'out_ucast_pkts': '49347842',
                                                                       'poller_hash': 'MkFEMTBBcGVlcmludGZwb24gMjE1ODk1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '215895',
                                                                                               'links': [       {       'connected_time': '68775798',
                                                                                                                        'fiber_length': 17507,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MkFEMTBBcGVlcnBvbnBvbiAyMTU4OTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5485 KING ST (MARTIN KEITH)',
                                         'poller_hash': 'MkFEMTBB',
                                         'serial_number': '2AD10A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:59:ec:bb',
                                         'uptime': '68775798'},
                                 {       'counter_source': '216187/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:91:d2:f9'],
                                                                       'id': '216187/g1',
                                                                       'in_nucast_pkts': '832139',
                                                                       'in_octets': '560997896276',
                                                                       'in_ucast_pkts': '6180244033',
                                                                       'mac_address': 'b8:94:70:ba:e5:18',
                                                                       'mtu': 2000,
                                                                       'name': '216187/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '43750319',
                                                                       'out_octets': '14964221716612',
                                                                       'out_ucast_pkts': '11110998925',
                                                                       'poller_hash': 'MTE5NTFCN3BlZXJpbnRmMjE2MTg3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216187',
                                                                       'in_nucast_pkts': '43750319',
                                                                       'in_octets': '14964221716612',
                                                                       'in_ucast_pkts': '11110998925',
                                                                       'links': [       {       'in_nucast_pkts': '43750319',
                                                                                                'in_octets': '14964221716612',
                                                                                                'in_ucast_pkts': '11110998925',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '832139',
                                                                                                'out_octets': '560997896276',
                                                                                                'out_ucast_pkts': '6180244033',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '832139',
                                                                       'out_octets': '560997896276',
                                                                       'out_ucast_pkts': '6180244033',
                                                                       'poller_hash': 'MTE5NTFCN3BlZXJpbnRmcG9uIDIxNjE4Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216187',
                                                                                               'links': [       {       'connected_time': '73164803',
                                                                                                                        'fiber_length': 22730,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTFCN3BlZXJwb25wb24gMjE2MTg3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '12 LAWRENCE ST (GIVLIN FRANCIS)',
                                         'poller_hash': 'MTE5NTFCNw==',
                                         'serial_number': '11951B7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e5:15',
                                         'uptime': '73164803'},
                                 {       'counter_source': '216217/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:ac:e7'],
                                                                       'id': '216217/g1',
                                                                       'in_nucast_pkts': '6741',
                                                                       'in_octets': '841822473',
                                                                       'in_ucast_pkts': '3282548',
                                                                       'mac_address': 'b8:94:70:ba:e3:18',
                                                                       'mtu': 2000,
                                                                       'name': '216217/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '383993',
                                                                       'out_octets': '22945721085',
                                                                       'out_ucast_pkts': '17333934',
                                                                       'poller_hash': 'MTE5NTE3N3BlZXJpbnRmMjE2MjE3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '216217',
                                                                       'in_nucast_pkts': '383993',
                                                                       'in_octets': '22945721085',
                                                                       'in_ucast_pkts': '17333934',
                                                                       'links': [       {       'in_nucast_pkts': '383993',
                                                                                                'in_octets': '22945721085',
                                                                                                'in_ucast_pkts': '17333934',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6741',
                                                                                                'out_octets': '841822473',
                                                                                                'out_ucast_pkts': '3282548',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6741',
                                                                       'out_octets': '841822473',
                                                                       'out_ucast_pkts': '3282548',
                                                                       'poller_hash': 'MTE5NTE3N3BlZXJpbnRmcG9uIDIxNjIxNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216217',
                                                                                               'links': [       {       'connected_time': '335346',
                                                                                                                        'fiber_length': 18064,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTE3N3BlZXJwb25wb24gMjE2MjE3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5446 ROAD 106 (FORAN GERALD)',
                                         'poller_hash': 'MTE5NTE3Nw==',
                                         'serial_number': '1195177',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e3:15',
                                         'uptime': '335346'},
                                 {       'counter_source': '216267/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '216267/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '935798595618',
                                                                       'in_ucast_pkts': '4878812406',
                                                                       'mac_address': 'b8:94:70:ba:e0:6d',
                                                                       'name': '216267/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '84',
                                                                       'out_octets': '19929742495974',
                                                                       'out_ucast_pkts': '25134346087',
                                                                       'poller_hash': 'MTE5NTEyMnBlZXJpbnRmMjE2MjY3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '216267',
                                                                       'in_nucast_pkts': '84',
                                                                       'in_octets': '19929742495974',
                                                                       'in_ucast_pkts': '25134346087',
                                                                       'links': [       {       'in_nucast_pkts': '84',
                                                                                                'in_octets': '19929742495974',
                                                                                                'in_ucast_pkts': '25134346087',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '935798595618',
                                                                                                'out_ucast_pkts': '4878812406',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '935798595618',
                                                                       'out_ucast_pkts': '4878812406',
                                                                       'poller_hash': 'MTE5NTEyMnBlZXJpbnRmcG9uIDIxNjI2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216267',
                                                                                               'links': [       {       'connected_time': '15182678',
                                                                                                                        'fiber_length': 23588,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTEyMnBlZXJwb25wb24gMjE2MjY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '39 REINER CR (SUI DANIEL)',
                                         'poller_hash': 'MTE5NTEyMg==',
                                         'serial_number': '1195122',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e0:6d',
                                         'uptime': '15182678'},
                                 {       'counter_source': '216975/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '216975/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '456468458995',
                                                                       'in_ucast_pkts': '22832363227',
                                                                       'mac_address': 'b8:94:70:ba:e0:a5',
                                                                       'name': '216975/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '37',
                                                                       'out_octets': '7126110654974',
                                                                       'out_ucast_pkts': '5097479738',
                                                                       'poller_hash': 'MTE5NTEyOXBlZXJpbnRmMjE2OTc1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '216975',
                                                                       'in_nucast_pkts': '37',
                                                                       'in_octets': '7126110654974',
                                                                       'in_ucast_pkts': '5097479738',
                                                                       'links': [       {       'in_nucast_pkts': '37',
                                                                                                'in_octets': '7126110654974',
                                                                                                'in_ucast_pkts': '5097479738',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '456468458995',
                                                                                                'out_ucast_pkts': '22832363227',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '456468458995',
                                                                       'out_ucast_pkts': '22832363227',
                                                                       'poller_hash': 'MTE5NTEyOXBlZXJpbnRmcG9uIDIxNjk3NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '216975',
                                                                                               'links': [       {       'connected_time': '335336',
                                                                                                                        'fiber_length': 17757,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTEyOXBlZXJwb25wb24gMjE2OTc1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5517 BROWN ST (ELLISON WHITE MIKE & '
                                                 'JULIE)',
                                         'poller_hash': 'MTE5NTEyOQ==',
                                         'serial_number': '1195129',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e0:a5',
                                         'uptime': '335336'},
                                 {       'counter_source': '217646/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e1:5f'],
                                                                       'id': '217646/g1',
                                                                       'in_nucast_pkts': '147771',
                                                                       'in_octets': '50596051326',
                                                                       'in_ucast_pkts': '335151754',
                                                                       'mac_address': 'b8:94:70:ba:e4:f0',
                                                                       'mtu': 2000,
                                                                       'name': '217646/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878607',
                                                                       'out_octets': '2538778560317',
                                                                       'out_ucast_pkts': '1754066759',
                                                                       'poller_hash': 'MTE5NTFCMnBlZXJpbnRmMjE3NjQ2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '217646',
                                                                       'in_nucast_pkts': '9878607',
                                                                       'in_octets': '2538778560317',
                                                                       'in_ucast_pkts': '1754066759',
                                                                       'links': [       {       'in_nucast_pkts': '9878607',
                                                                                                'in_octets': '2538778560317',
                                                                                                'in_ucast_pkts': '1754066759',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '147771',
                                                                                                'out_octets': '50596051326',
                                                                                                'out_ucast_pkts': '335151754',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '147771',
                                                                       'out_octets': '50596051326',
                                                                       'out_ucast_pkts': '335151754',
                                                                       'poller_hash': 'MTE5NTFCMnBlZXJpbnRmcG9uIDIxNzY0Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '217646',
                                                                                               'links': [       {       'connected_time': '8535249',
                                                                                                                        'fiber_length': 22151,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5NTFCMnBlZXJwb25wb24gMjE3NjQ2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 DAVID ST (BRENNAN KARALEE)',
                                         'poller_hash': 'MTE5NTFCMg==',
                                         'serial_number': '11951B2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:e4:ed',
                                         'uptime': '8535249'},
                                 {       'counter_source': '218237/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e0:99'],
                                                                       'id': '218237/g1',
                                                                       'in_nucast_pkts': '26684107',
                                                                       'in_octets': '244353092664',
                                                                       'in_ucast_pkts': '2172832992',
                                                                       'mac_address': '04:bc:9f:5c:52:24',
                                                                       'mtu': 2000,
                                                                       'name': '218237/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '46273274895',
                                                                       'out_octets': '68397272978120',
                                                                       'out_ucast_pkts': '4210984802',
                                                                       'poller_hash': 'MTI3NTBDNHBlZXJpbnRmMjE4MjM3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '218237',
                                                                       'in_nucast_pkts': '46273274895',
                                                                       'in_octets': '68397272978120',
                                                                       'in_ucast_pkts': '4210984802',
                                                                       'links': [       {       'in_nucast_pkts': '46273274895',
                                                                                                'in_octets': '68397272978120',
                                                                                                'in_ucast_pkts': '4210984802',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '26684107',
                                                                                                'out_octets': '244353092664',
                                                                                                'out_ucast_pkts': '2172832992',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26684107',
                                                                       'out_octets': '244353092664',
                                                                       'out_ucast_pkts': '2172832992',
                                                                       'poller_hash': 'MTI3NTBDNHBlZXJpbnRmcG9uIDIxODIzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218237',
                                                                                               'links': [       {       'connected_time': '15709246',
                                                                                                                        'fiber_length': 23820,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTBDNHBlZXJwb25wb24gMjE4MjM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3710 NAFZIGER RD (LENG MILDRED)',
                                         'poller_hash': 'MTI3NTBDNA==',
                                         'serial_number': '12750C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:52:21',
                                         'uptime': '15709246'},
                                 {       'counter_source': '218307/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e5:b6'],
                                                                       'id': '218307/g1',
                                                                       'in_nucast_pkts': '52251',
                                                                       'in_octets': '12922386007',
                                                                       'in_ucast_pkts': '78409878',
                                                                       'mac_address': '04:bc:9f:5c:5a:24',
                                                                       'mtu': 2000,
                                                                       'name': '218307/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3185598',
                                                                       'out_octets': '358285953315',
                                                                       'out_ucast_pkts': '250089661',
                                                                       'poller_hash': 'MTI3NTFDNHBlZXJpbnRmMjE4MzA3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '218307',
                                                                       'in_nucast_pkts': '3185598',
                                                                       'in_octets': '358285953315',
                                                                       'in_ucast_pkts': '250089661',
                                                                       'links': [       {       'in_nucast_pkts': '3185598',
                                                                                                'in_octets': '358285953315',
                                                                                                'in_ucast_pkts': '250089661',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '52251',
                                                                                                'out_octets': '12922386007',
                                                                                                'out_ucast_pkts': '78409878',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '52251',
                                                                       'out_octets': '12922386007',
                                                                       'out_ucast_pkts': '78409878',
                                                                       'poller_hash': 'MTI3NTFDNHBlZXJpbnRmcG9uIDIxODMwNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '218307',
                                                                                               'links': [       {       'connected_time': '3772835',
                                                                                                                        'fiber_length': 24796,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 3.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTFDNHBlZXJwb25wb24gMjE4MzA3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '6 EDGEWOOD CT (GUENTHER JOANNE)',
                                         'poller_hash': 'MTI3NTFDNA==',
                                         'serial_number': '12751C4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:5a:21',
                                         'uptime': '3772835'},
                                 {       'counter_source': '220389/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:01:21:06:75:ef'],
                                                                       'id': '220389/g2',
                                                                       'in_nucast_pkts': '139118',
                                                                       'in_octets': '2721474266135',
                                                                       'in_ucast_pkts': '9843707430',
                                                                       'mac_address': '04:bc:9f:5c:53:3d',
                                                                       'mtu': 2000,
                                                                       'name': '220389/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '53556606',
                                                                       'out_octets': '22571405451020',
                                                                       'out_ucast_pkts': '20293835713',
                                                                       'poller_hash': 'MTI3NTBFN3BlZXJpbnRmMjIwMzg5L2cy',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '220389',
                                                                       'in_nucast_pkts': '53556606',
                                                                       'in_octets': '22571405451020',
                                                                       'in_ucast_pkts': '20293835713',
                                                                       'links': [       {       'in_nucast_pkts': '53556606',
                                                                                                'in_octets': '22571405451020',
                                                                                                'in_ucast_pkts': '20293835713',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '139118',
                                                                                                'out_octets': '2721474266135',
                                                                                                'out_ucast_pkts': '9843707430',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '139118',
                                                                       'out_octets': '2721474266135',
                                                                       'out_ucast_pkts': '9843707430',
                                                                       'poller_hash': 'MTI3NTBFN3BlZXJpbnRmcG9uIDIyMDM4OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '220389',
                                                                                               'links': [       {       'connected_time': '13895239',
                                                                                                                        'fiber_length': 21791,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTBFN3BlZXJwb25wb24gMjIwMzg5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1401 QUEEN'S BUSH ROAD (TOWNSHIP OF "
                                                 'WELLESLEY)',
                                         'poller_hash': 'MTI3NTBFNw==',
                                         'serial_number': '12750E7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:53:39',
                                         'uptime': '13895239'},
                                 {       'counter_source': '221435/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4f:ac:9a'],
                                                                       'id': '221435/g1',
                                                                       'in_nucast_pkts': '57249',
                                                                       'in_octets': '8551870898',
                                                                       'in_ucast_pkts': '35188545',
                                                                       'mac_address': '04:bc:9f:5c:51:cc',
                                                                       'mtu': 2000,
                                                                       'name': '221435/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2350724',
                                                                       'out_octets': '141962704611',
                                                                       'out_ucast_pkts': '108046327',
                                                                       'poller_hash': 'MTI3NTBCOXBlZXJpbnRmMjIxNDM1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '221435',
                                                                       'in_nucast_pkts': '2350724',
                                                                       'in_octets': '141962704611',
                                                                       'in_ucast_pkts': '108046327',
                                                                       'links': [       {       'in_nucast_pkts': '2350724',
                                                                                                'in_octets': '141962704611',
                                                                                                'in_ucast_pkts': '108046327',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '57249',
                                                                                                'out_octets': '8551870898',
                                                                                                'out_ucast_pkts': '35188545',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '57249',
                                                                       'out_octets': '8551870898',
                                                                       'out_ucast_pkts': '35188545',
                                                                       'poller_hash': 'MTI3NTBCOXBlZXJpbnRmcG9uIDIyMTQzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221435',
                                                                                               'links': [       {       'connected_time': '2484326',
                                                                                                                        'fiber_length': 23748,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTBCOXBlZXJwb25wb24gMjIxNDM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '26 FERRIS DR (NAFZIGER JAMES)',
                                         'poller_hash': 'MTI3NTBCOQ==',
                                         'serial_number': '12750B9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:51:c9',
                                         'uptime': '2484326'},
                                 {       'counter_source': '221455/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e0:83'],
                                                                       'id': '221455/g1',
                                                                       'in_nucast_pkts': '147145',
                                                                       'in_octets': '126715413151',
                                                                       'in_ucast_pkts': '629485532',
                                                                       'mac_address': '04:bc:9f:5c:50:0c',
                                                                       'mtu': 2000,
                                                                       'name': '221455/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9925674',
                                                                       'out_octets': '1955147794818',
                                                                       'out_ucast_pkts': '1486489588',
                                                                       'poller_hash': 'MTI3NTA4MXBlZXJpbnRmMjIxNDU1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '221455',
                                                                       'in_nucast_pkts': '9925674',
                                                                       'in_octets': '1955147794818',
                                                                       'in_ucast_pkts': '1486489588',
                                                                       'links': [       {       'in_nucast_pkts': '9925674',
                                                                                                'in_octets': '1955147794818',
                                                                                                'in_ucast_pkts': '1486489588',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '147145',
                                                                                                'out_octets': '126715413151',
                                                                                                'out_ucast_pkts': '629485532',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '147145',
                                                                       'out_octets': '126715413151',
                                                                       'out_ucast_pkts': '629485532',
                                                                       'poller_hash': 'MTI3NTA4MXBlZXJpbnRmcG9uIDIyMTQ1NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221455',
                                                                                               'links': [       {       'connected_time': '8535150',
                                                                                                                        'fiber_length': 24986,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTA4MXBlZXJwb25wb24gMjIxNDU1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 SHADYWOOD CT (FRASER SHEILA)',
                                         'poller_hash': 'MTI3NTA4MQ==',
                                         'serial_number': '1275081',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:50:09',
                                         'uptime': '8535150'},
                                 {       'counter_source': '221690/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:64:1f'],
                                                                       'id': '221690/g1',
                                                                       'in_nucast_pkts': '149413',
                                                                       'in_octets': '117971972671',
                                                                       'in_ucast_pkts': '912857908',
                                                                       'mac_address': '04:bc:9f:5c:58:94',
                                                                       'mtu': 2000,
                                                                       'name': '221690/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9795077',
                                                                       'out_octets': '4279702127880',
                                                                       'out_ucast_pkts': '3030736593',
                                                                       'poller_hash': 'MTI3NTE5MnBlZXJpbnRmMjIxNjkwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '221690',
                                                                       'in_nucast_pkts': '9795077',
                                                                       'in_octets': '4279702127880',
                                                                       'in_ucast_pkts': '3030736593',
                                                                       'links': [       {       'in_nucast_pkts': '9795077',
                                                                                                'in_octets': '4279702127880',
                                                                                                'in_ucast_pkts': '3030736593',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149413',
                                                                                                'out_octets': '117971972671',
                                                                                                'out_ucast_pkts': '912857908',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149413',
                                                                       'out_octets': '117971972671',
                                                                       'out_ucast_pkts': '912857908',
                                                                       'poller_hash': 'MTI3NTE5MnBlZXJpbnRmcG9uIDIyMTY5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221690',
                                                                                               'links': [       {       'connected_time': '8535219',
                                                                                                                        'fiber_length': 24804,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTE5MnBlZXJwb25wb24gMjIxNjkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1139 QUEEN'S BUSH RD (HARDY FRANK)",
                                         'poller_hash': 'MTI3NTE5Mg==',
                                         'serial_number': '1275192',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:58:91',
                                         'uptime': '8535219'},
                                 {       'counter_source': '221730/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e0:57'],
                                                                       'id': '221730/g1',
                                                                       'in_nucast_pkts': '9729',
                                                                       'in_octets': '30092396683',
                                                                       'in_ucast_pkts': '36843945',
                                                                       'mac_address': 'b8:94:70:ba:34:38',
                                                                       'mtu': 2000,
                                                                       'name': '221730/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715441',
                                                                       'out_octets': '91459697650',
                                                                       'out_ucast_pkts': '74609183',
                                                                       'poller_hash': 'MTE5M0I5QnBlZXJpbnRmMjIxNzMwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '221730',
                                                                       'in_nucast_pkts': '715441',
                                                                       'in_octets': '91459697650',
                                                                       'in_ucast_pkts': '74609183',
                                                                       'links': [       {       'in_nucast_pkts': '715441',
                                                                                                'in_octets': '91459697650',
                                                                                                'in_ucast_pkts': '74609183',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9729',
                                                                                                'out_octets': '30092396683',
                                                                                                'out_ucast_pkts': '36843945',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9729',
                                                                       'out_octets': '30092396683',
                                                                       'out_ucast_pkts': '36843945',
                                                                       'poller_hash': 'MTE5M0I5QnBlZXJpbnRmcG9uIDIyMTczMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '221730',
                                                                                               'links': [       {       'connected_time': '611474',
                                                                                                                        'fiber_length': 18530,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5M0I5QnBlZXJwb25wb24gMjIxNzMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2660 LINE 47 GADSHILL (JAIME NEEB)',
                                         'poller_hash': 'MTE5M0I5Qg==',
                                         'serial_number': '1193B9B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ba:34:35',
                                         'uptime': '611474'},
                                 {       'counter_source': '223463/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '223463/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '144496038553',
                                                                       'in_ucast_pkts': '845833182',
                                                                       'mac_address': '04:bc:9f:5c:4a:79',
                                                                       'name': '223463/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '4106464704375',
                                                                       'out_ucast_pkts': '2930755683',
                                                                       'poller_hash': 'MTI3NEZDRnBlZXJpbnRmMjIzNDYzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '223463',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '4106464704375',
                                                                       'in_ucast_pkts': '2930755683',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '4106464704375',
                                                                                                'in_ucast_pkts': '2930755683',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '144496038553',
                                                                                                'out_ucast_pkts': '845833182',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '144496038553',
                                                                       'out_ucast_pkts': '845833182',
                                                                       'poller_hash': 'MTI3NEZDRnBlZXJpbnRmcG9uIDIyMzQ2Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '223463',
                                                                                               'links': [       {       'connected_time': '68441994',
                                                                                                                        'fiber_length': 17078,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEZDRnBlZXJwb25wb24gMjIzNDYz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2088 PERTH LINE (LANTZ LORNE & '
                                                 'IRENE)',
                                         'poller_hash': 'MTI3NEZDRg==',
                                         'serial_number': '1274FCF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4a:79',
                                         'uptime': '68441994'},
                                 {       'counter_source': '223566/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '223566/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9889399558',
                                                                       'in_ucast_pkts': '115502176',
                                                                       'mac_address': '04:bc:9f:5c:4c:39',
                                                                       'name': '223566/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '717242049405',
                                                                       'out_ucast_pkts': '477766837',
                                                                       'poller_hash': 'MTI3NTAwN3BlZXJpbnRmMjIzNTY2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '223566',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '717242049405',
                                                                       'in_ucast_pkts': '477766837',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '717242049405',
                                                                                                'in_ucast_pkts': '477766837',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '9889399558',
                                                                                                'out_ucast_pkts': '115502176',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9889399558',
                                                                       'out_ucast_pkts': '115502176',
                                                                       'poller_hash': 'MTI3NTAwN3BlZXJpbnRmcG9uIDIyMzU2Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '223566',
                                                                                               'links': [       {       'connected_time': '6400538',
                                                                                                                        'fiber_length': 17082,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTAwN3BlZXJwb25wb24gMjIzNTY2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2086 PERTH LINE 56',
                                         'poller_hash': 'MTI3NTAwNw==',
                                         'serial_number': '1275007',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:4c:39',
                                         'uptime': '6400538'},
                                 {       'counter_source': '223933/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:ec:96'],
                                                                       'id': '223933/g1',
                                                                       'in_nucast_pkts': '9572',
                                                                       'in_octets': '11650831054',
                                                                       'in_ucast_pkts': '47490585',
                                                                       'mac_address': '04:bc:9f:5c:56:44',
                                                                       'mtu': 2000,
                                                                       'name': '223933/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '536881',
                                                                       'out_octets': '291248971664',
                                                                       'out_ucast_pkts': '233648791',
                                                                       'poller_hash': 'MTI3NTE0OHBlZXJpbnRmMjIzOTMzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '223933',
                                                                       'in_nucast_pkts': '536881',
                                                                       'in_octets': '291248971664',
                                                                       'in_ucast_pkts': '233648791',
                                                                       'links': [       {       'in_nucast_pkts': '536881',
                                                                                                'in_octets': '291248971664',
                                                                                                'in_ucast_pkts': '233648791',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9572',
                                                                                                'out_octets': '11650831054',
                                                                                                'out_ucast_pkts': '47490585',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9572',
                                                                       'out_octets': '11650831054',
                                                                       'out_ucast_pkts': '47490585',
                                                                       'poller_hash': 'MTI3NTE0OHBlZXJpbnRmcG9uIDIyMzkzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '223933',
                                                                                               'links': [       {       'connected_time': '611457',
                                                                                                                        'fiber_length': 19745,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTE0OHBlZXJwb25wb24gMjIzOTMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4918 LINHAVEN DR (SCHWARTZ TREVOR)',
                                         'poller_hash': 'MTI3NTE0OA==',
                                         'serial_number': '1275148',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:56:41',
                                         'uptime': '611457'},
                                 {       'counter_source': '229691/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '229691/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '426907319650',
                                                                       'in_ucast_pkts': '1065860134',
                                                                       'mac_address': '04:bc:9f:5b:e3:79',
                                                                       'name': '229691/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '42',
                                                                       'out_octets': '3106560711798',
                                                                       'out_ucast_pkts': '2420295035',
                                                                       'poller_hash': 'MTI3NDJFRnBlZXJpbnRmMjI5NjkxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '229691',
                                                                       'in_nucast_pkts': '42',
                                                                       'in_octets': '3106560711798',
                                                                       'in_ucast_pkts': '2420295035',
                                                                       'links': [       {       'in_nucast_pkts': '42',
                                                                                                'in_octets': '3106560711798',
                                                                                                'in_ucast_pkts': '2420295035',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '426907319650',
                                                                                                'out_ucast_pkts': '1065860134',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '426907319650',
                                                                       'out_ucast_pkts': '1065860134',
                                                                       'poller_hash': 'MTI3NDJFRnBlZXJpbnRmcG9uIDIyOTY5MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '229691',
                                                                                               'links': [       {       'connected_time': '8535228',
                                                                                                                        'fiber_length': 24017,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJFRnBlZXJwb25wb24gMjI5Njkx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '116 REINER CR (ROSE ANNA)',
                                         'poller_hash': 'MTI3NDJFRg==',
                                         'serial_number': '12742EF',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e3:79',
                                         'uptime': '8535228'},
                                 {       'counter_source': '230714/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '230714/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '87465860317',
                                                                       'in_ucast_pkts': '71390913',
                                                                       'mac_address': '04:bc:9f:5c:05:f1',
                                                                       'name': '230714/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '20',
                                                                       'out_octets': '36935808451',
                                                                       'out_ucast_pkts': '82425144',
                                                                       'poller_hash': 'MTI3NDczRXBlZXJpbnRmMjMwNzE0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '230714',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '36935808451',
                                                                       'in_ucast_pkts': '82425144',
                                                                       'links': [       {       'in_nucast_pkts': '20',
                                                                                                'in_octets': '36935808451',
                                                                                                'in_ucast_pkts': '82425144',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '87465860317',
                                                                                                'out_ucast_pkts': '71390913',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '87465860317',
                                                                       'out_ucast_pkts': '71390913',
                                                                       'poller_hash': 'MTI3NDczRXBlZXJpbnRmcG9uIDIzMDcxNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '230714',
                                                                                               'links': [       {       'connected_time': '508890',
                                                                                                                        'fiber_length': 23248,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDczRXBlZXJwb25wb24gMjMwNzE0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '146 DAVID ST (COMMUNITY FOOD '
                                                 'CUPBOARD)',
                                         'poller_hash': 'MTI3NDczRQ==',
                                         'serial_number': '127473E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:05:f1',
                                         'uptime': '508890'},
                                 {       'counter_source': '231028/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:ee:ec'],
                                                                       'id': '231028/g4',
                                                                       'in_nucast_pkts': '11464',
                                                                       'in_octets': '5975324839',
                                                                       'in_ucast_pkts': '45859413',
                                                                       'mac_address': 'cc:be:59:cb:6b:48',
                                                                       'mtu': 2000,
                                                                       'name': '231028/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715554',
                                                                       'out_octets': '176442923723',
                                                                       'out_ucast_pkts': '121411173',
                                                                       'poller_hash': 'NDJCRUMwcGVlcmludGYyMzEwMjgvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231028',
                                                                       'in_nucast_pkts': '715554',
                                                                       'in_octets': '176442923723',
                                                                       'in_ucast_pkts': '121411173',
                                                                       'links': [       {       'in_nucast_pkts': '715554',
                                                                                                'in_octets': '176442923723',
                                                                                                'in_ucast_pkts': '121411173',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11464',
                                                                                                'out_octets': '5975324839',
                                                                                                'out_ucast_pkts': '45859413',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11464',
                                                                       'out_octets': '5975324839',
                                                                       'out_ucast_pkts': '45859413',
                                                                       'poller_hash': 'NDJCRUMwcGVlcmludGZwb24gMjMxMDI4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231028',
                                                                                               'links': [       {       'connected_time': '611450',
                                                                                                                        'fiber_length': 19811,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDJCRUMwcGVlcnBvbnBvbiAyMzEwMjg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2849 Line 45 (Jason Rose)',
                                         'poller_hash': 'NDJCRUMw',
                                         'serial_number': '42BEC0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:cb:6b:42',
                                         'uptime': '611450'},
                                 {       'counter_source': '231450/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:ca:2b'],
                                                                       'id': '231450/g1',
                                                                       'in_nucast_pkts': '774761',
                                                                       'in_octets': '253124882609',
                                                                       'in_ucast_pkts': '2929503305',
                                                                       'mac_address': '04:bc:9f:5c:20:24',
                                                                       'mtu': 2000,
                                                                       'name': '231450/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '42004007',
                                                                       'out_octets': '17994161445509',
                                                                       'out_ucast_pkts': '12030220936',
                                                                       'poller_hash': 'MTI3NEE4NHBlZXJpbnRmMjMxNDUwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231450',
                                                                       'in_nucast_pkts': '42004007',
                                                                       'in_octets': '17994161445509',
                                                                       'in_ucast_pkts': '12030220936',
                                                                       'links': [       {       'in_nucast_pkts': '42004007',
                                                                                                'in_octets': '17994161445509',
                                                                                                'in_ucast_pkts': '12030220936',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '774761',
                                                                                                'out_octets': '253124882609',
                                                                                                'out_ucast_pkts': '2929503305',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '774761',
                                                                       'out_octets': '253124882609',
                                                                       'out_ucast_pkts': '2929503305',
                                                                       'poller_hash': 'MTI3NEE4NHBlZXJpbnRmcG9uIDIzMTQ1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231450',
                                                                                               'links': [       {       'connected_time': '73010263',
                                                                                                                        'fiber_length': 19626,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE4NHBlZXJwb25wb24gMjMxNDUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4896 PERTH ROAD 119 (MOHR DAVID & '
                                                 'LYNDA)',
                                         'poller_hash': 'MTI3NEE4NA==',
                                         'serial_number': '1274A84',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:21',
                                         'uptime': '73010263'},
                                 {       'counter_source': '231590/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e3:c7'],
                                                                       'id': '231590/g1',
                                                                       'in_nucast_pkts': '6060',
                                                                       'in_octets': '7417714083',
                                                                       'in_ucast_pkts': '15928577',
                                                                       'mac_address': '04:bc:9f:5c:20:0c',
                                                                       'mtu': 2000,
                                                                       'name': '231590/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '394324',
                                                                       'out_octets': '79815742269',
                                                                       'out_ucast_pkts': '61063874',
                                                                       'poller_hash': 'MTI3NEE4MXBlZXJpbnRmMjMxNTkwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '231590',
                                                                       'in_nucast_pkts': '394324',
                                                                       'in_octets': '79815742269',
                                                                       'in_ucast_pkts': '61063874',
                                                                       'links': [       {       'in_nucast_pkts': '394324',
                                                                                                'in_octets': '79815742269',
                                                                                                'in_ucast_pkts': '61063874',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '6060',
                                                                                                'out_octets': '7417714083',
                                                                                                'out_ucast_pkts': '15928577',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '6060',
                                                                       'out_octets': '7417714083',
                                                                       'out_ucast_pkts': '15928577',
                                                                       'poller_hash': 'MTI3NEE4MXBlZXJpbnRmcG9uIDIzMTU5MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '231590',
                                                                                               'links': [       {       'connected_time': '335336',
                                                                                                                        'fiber_length': 17126,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEE4MXBlZXJwb25wb24gMjMxNTkw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5565 KING ST (WILSON DAVE)',
                                         'poller_hash': 'MTI3NEE4MQ==',
                                         'serial_number': '1274A81',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:20:09',
                                         'uptime': '335336'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '232578/g1',
                                                                       'name': '232578/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MzI5NjQ2cGVlcmludGYyMzI1NzgvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '232578',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MzI5NjQ2cGVlcmludGZwb24gMjMyNTc4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '232578',
                                                                                               'links': [       {       'connected_time': '4926848',
                                                                                                                        'fiber_length': 19905,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.902,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzI5NjQ2cGVlcnBvbnBvbiAyMzI1Nzg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '2864 LINE 45 (ROTH HAROLD)',
                                         'poller_hash': 'MzI5NjQ2',
                                         'serial_number': '329646',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:92:fd:d0',
                                         'uptime': '4926848'},
                                 {       'counter_source': '233238/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '233238/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '504574604762',
                                                                       'in_ucast_pkts': '57200178456',
                                                                       'mac_address': '04:bc:9f:5c:24:61',
                                                                       'name': '233238/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '79',
                                                                       'out_octets': '4851671731213',
                                                                       'out_ucast_pkts': '3757744075',
                                                                       'poller_hash': 'MTI3NEIwQ3BlZXJpbnRmMjMzMjM4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '233238',
                                                                       'in_nucast_pkts': '79',
                                                                       'in_octets': '4851671731213',
                                                                       'in_ucast_pkts': '3757744075',
                                                                       'links': [       {       'in_nucast_pkts': '79',
                                                                                                'in_octets': '4851671731213',
                                                                                                'in_ucast_pkts': '3757744075',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '504574604762',
                                                                                                'out_ucast_pkts': '57200178456',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '504574604762',
                                                                       'out_ucast_pkts': '57200178456',
                                                                       'poller_hash': 'MTI3NEIwQ3BlZXJpbnRmcG9uIDIzMzIzOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233238',
                                                                                               'links': [       {       'connected_time': '10695148',
                                                                                                                        'fiber_length': 19900,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIwQ3BlZXJwb25wb24gMjMzMjM4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4876 PERTH ROAD 119 (ROESNER NANCY)',
                                         'poller_hash': 'MTI3NEIwQw==',
                                         'serial_number': '1274B0C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:61',
                                         'uptime': '10695148'},
                                 {       'counter_source': '233643/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c5:5f:e8'],
                                                                       'id': '233643/g1',
                                                                       'in_nucast_pkts': '75',
                                                                       'in_octets': '86336733416',
                                                                       'in_ucast_pkts': '360023789',
                                                                       'mac_address': 'ec:4f:82:8d:3e:8b',
                                                                       'mtu': 2000,
                                                                       'name': '233643/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919989',
                                                                       'out_octets': '1230444913445',
                                                                       'out_ucast_pkts': '913913945',
                                                                       'poller_hash': 'MzIxMDE4cGVlcmludGYyMzM2NDMvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '233643',
                                                                       'in_nucast_pkts': '9919989',
                                                                       'in_octets': '1230444913445',
                                                                       'in_ucast_pkts': '913913945',
                                                                       'links': [       {       'in_nucast_pkts': '9919989',
                                                                                                'in_octets': '1230444913445',
                                                                                                'in_ucast_pkts': '913913945',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '75',
                                                                                                'out_octets': '86336733416',
                                                                                                'out_ucast_pkts': '360023789',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '75',
                                                                       'out_octets': '86336733416',
                                                                       'out_ucast_pkts': '360023789',
                                                                       'poller_hash': 'MzIxMDE4cGVlcmludGZwb24gMjMzNjQz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '233643',
                                                                                               'links': [       {       'connected_time': '8535262',
                                                                                                                        'fiber_length': 23036,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MzIxMDE4cGVlcnBvbnBvbiAyMzM2NDM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1100 MOLESWORTH STREET (FRITZ '
                                                 'MICHAEL)',
                                         'poller_hash': 'MzIxMDE4',
                                         'serial_number': '321018',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'ec:4f:82:8d:3e:88',
                                         'uptime': '8535262'},
                                 {       'counter_source': '235506/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '235506/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '218005782423',
                                                                       'in_ucast_pkts': '2004674415',
                                                                       'mac_address': '04:bc:9f:5c:27:89',
                                                                       'name': '235506/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '117',
                                                                       'out_octets': '5988055257164',
                                                                       'out_ucast_pkts': '4132675123',
                                                                       'poller_hash': 'MTI3NEI3MXBlZXJpbnRmMjM1NTA2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '235506',
                                                                       'in_nucast_pkts': '117',
                                                                       'in_octets': '5988055257164',
                                                                       'in_ucast_pkts': '4132675123',
                                                                       'links': [       {       'in_nucast_pkts': '117',
                                                                                                'in_octets': '5988055257164',
                                                                                                'in_ucast_pkts': '4132675123',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '218005782423',
                                                                                                'out_ucast_pkts': '2004674415',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '218005782423',
                                                                       'out_ucast_pkts': '2004674415',
                                                                       'poller_hash': 'MTI3NEI3MXBlZXJpbnRmcG9uIDIzNTUwNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235506',
                                                                                               'links': [       {       'connected_time': '1877769',
                                                                                                                        'fiber_length': 20016,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEI3MXBlZXJwb25wb24gMjM1NTA2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4917 PERTH ROAD 119 (BROWN JOHN T)',
                                         'poller_hash': 'MTI3NEI3MQ==',
                                         'serial_number': '1274B71',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:27:89',
                                         'uptime': '1877769'},
                                 {       'counter_source': '235713/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8c:6c'],
                                                                       'id': '235713/g1',
                                                                       'in_nucast_pkts': '148485',
                                                                       'in_octets': '140049442117',
                                                                       'in_ucast_pkts': '726021370',
                                                                       'mac_address': '04:bc:9f:5c:17:5c',
                                                                       'mtu': 2000,
                                                                       'name': '235713/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797855',
                                                                       'out_octets': '2952284841576',
                                                                       'out_ucast_pkts': '2210150459',
                                                                       'poller_hash': 'MTI3NDk2QnBlZXJpbnRmMjM1NzEzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '235713',
                                                                       'in_nucast_pkts': '9797855',
                                                                       'in_octets': '2952284841576',
                                                                       'in_ucast_pkts': '2210150459',
                                                                       'links': [       {       'in_nucast_pkts': '9797855',
                                                                                                'in_octets': '2952284841576',
                                                                                                'in_ucast_pkts': '2210150459',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '148485',
                                                                                                'out_octets': '140049442117',
                                                                                                'out_ucast_pkts': '726021370',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '148485',
                                                                       'out_octets': '140049442117',
                                                                       'out_ucast_pkts': '726021370',
                                                                       'poller_hash': 'MTI3NDk2QnBlZXJpbnRmcG9uIDIzNTcxMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '235713',
                                                                                               'links': [       {       'connected_time': '8535234',
                                                                                                                        'fiber_length': 23308,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -25.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDk2QnBlZXJwb25wb24gMjM1NzEz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5 BERDUX PL (SCHWARTZ TREVOR)',
                                         'poller_hash': 'MTI3NDk2Qg==',
                                         'serial_number': '127496B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:17:59',
                                         'uptime': '8535234'},
                                 {       'counter_source': '237743/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '237743/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '258099124702',
                                                                       'in_ucast_pkts': '2106397237',
                                                                       'mac_address': '04:bc:9f:5c:25:49',
                                                                       'name': '237743/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '165',
                                                                       'out_octets': '12348352455368',
                                                                       'out_ucast_pkts': '8363370646',
                                                                       'poller_hash': 'MTI3NEIyOXBlZXJpbnRmMjM3NzQzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '237743',
                                                                       'in_nucast_pkts': '165',
                                                                       'in_octets': '12348352455368',
                                                                       'in_ucast_pkts': '8363370646',
                                                                       'links': [       {       'in_nucast_pkts': '165',
                                                                                                'in_octets': '12348352455368',
                                                                                                'in_ucast_pkts': '8363370646',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '258099124702',
                                                                                                'out_ucast_pkts': '2106397237',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '258099124702',
                                                                       'out_ucast_pkts': '2106397237',
                                                                       'poller_hash': 'MTI3NEIyOXBlZXJpbnRmcG9uIDIzNzc0Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '237743',
                                                                                               'links': [       {       'connected_time': '26451742',
                                                                                                                        'fiber_length': 19837,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIyOXBlZXJwb25wb24gMjM3NzQz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2848 LINE 45 GADSHILL (WALKER LARRY)',
                                         'poller_hash': 'MTI3NEIyOQ==',
                                         'serial_number': '1274B29',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:25:49',
                                         'uptime': '26451742'},
                                 {       'counter_source': '237773/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:c5:5d:18'],
                                                                       'id': '237773/g1',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '140902469545',
                                                                       'in_ucast_pkts': '500255063',
                                                                       'mac_address': '04:bc:9f:5c:24:24',
                                                                       'mtu': 2000,
                                                                       'name': '237773/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920246',
                                                                       'out_octets': '2172033896839',
                                                                       'out_ucast_pkts': '1588960581',
                                                                       'poller_hash': 'MTI3NEIwNHBlZXJpbnRmMjM3NzczL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '237773',
                                                                       'in_nucast_pkts': '9920246',
                                                                       'in_octets': '2172033896839',
                                                                       'in_ucast_pkts': '1588960581',
                                                                       'links': [       {       'in_nucast_pkts': '9920246',
                                                                                                'in_octets': '2172033896839',
                                                                                                'in_ucast_pkts': '1588960581',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '140902469545',
                                                                                                'out_ucast_pkts': '500255063',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '140902469545',
                                                                       'out_ucast_pkts': '500255063',
                                                                       'poller_hash': 'MTI3NEIwNHBlZXJpbnRmcG9uIDIzNzc3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '237773',
                                                                                               'links': [       {       'connected_time': '8535205',
                                                                                                                        'fiber_length': 23465,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEIwNHBlZXJwb25wb24gMjM3Nzcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '37 ZINKANN CR WELLESLEY (DYSON '
                                                 'CATHERINE)',
                                         'poller_hash': 'MTI3NEIwNA==',
                                         'serial_number': '1274B04',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:24:21',
                                         'uptime': '8535205'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '237833/g1',
                                                                       'name': '237833/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NEFDN3BlZXJpbnRmMjM3ODMzL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '237833',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NEFDN3BlZXJpbnRmcG9uIDIzNzgzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '237833',
                                                                                               'links': [       {       'connected_time': '72734407',
                                                                                                                        'fiber_length': 21904,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFDN3BlZXJwb25wb24gMjM3ODMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1401 QUEEN'S BUSH RD (WOOLWICH "
                                                 'COMMUNITY HEALTH CENT)',
                                         'poller_hash': 'MTI3NEFDNw==',
                                         'serial_number': '1274AC7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:39',
                                         'uptime': '72734407'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '241435/g1',
                                                                       'name': '241435/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NTE1OHBlZXJpbnRmMjQxNDM1L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '241435',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NTE1OHBlZXJpbnRmcG9uIDI0MTQzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241435',
                                                                                               'max_distance': 60000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NTE1OHBlZXJwb25wb24gMjQxNDM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '5 Temperance St. Milverton (Woody '
                                                 'Gilbert)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'OLT PON port 7/7 is '
                                                                  'down',
                                         'offline_time': '2024-10-17T07:03:42.740Z',
                                         'poller_hash': 'MTI3NTE1OA==',
                                         'serial_number': '1275158',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '241614/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '241614/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '814195273757',
                                                                       'in_ucast_pkts': '1601192348',
                                                                       'mac_address': 'ec:4f:82:44:18:57',
                                                                       'name': '241614/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '941471932974',
                                                                       'out_ucast_pkts': '1708704776',
                                                                       'poller_hash': 'MjNBMDBBcGVlcmludGYyNDE2MTQvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '241614',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '941471932974',
                                                                       'in_ucast_pkts': '1708704776',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '941471932974',
                                                                                                'in_ucast_pkts': '1708704776',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '814195273757',
                                                                                                'out_ucast_pkts': '1601192348',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '814195273757',
                                                                       'out_ucast_pkts': '1601192348',
                                                                       'poller_hash': 'MjNBMDBBcGVlcmludGZwb24gMjQxNjE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241614',
                                                                                               'links': [       {       'connected_time': '72154142',
                                                                                                                        'fiber_length': 14891,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.848,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MjNBMDBBcGVlcnBvbnBvbiAyNDE2MTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '5321 PERTH ROAD 119 (CEDAR CREST '
                                                 'ACRES INC)',
                                         'poller_hash': 'MjNBMDBB',
                                         'serial_number': '23A00A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': 'ec:4f:82:44:18:57',
                                         'uptime': '72154142'},
                                 {       'counter_source': '241885/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:f3:e9:e1'],
                                                                       'id': '241885/g1',
                                                                       'in_nucast_pkts': '156524',
                                                                       'in_octets': '106486482961',
                                                                       'in_ucast_pkts': '268087645',
                                                                       'mac_address': '04:bc:9f:5b:ce:f4',
                                                                       'mtu': 2000,
                                                                       'name': '241885/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9925938',
                                                                       'out_octets': '1450411781218',
                                                                       'out_ucast_pkts': '1144073127',
                                                                       'poller_hash': 'MTI3NDA1RXBlZXJpbnRmMjQxODg1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '241885',
                                                                       'in_nucast_pkts': '9925938',
                                                                       'in_octets': '1450411781218',
                                                                       'in_ucast_pkts': '1144073127',
                                                                       'links': [       {       'in_nucast_pkts': '9925938',
                                                                                                'in_octets': '1450411781218',
                                                                                                'in_ucast_pkts': '1144073127',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '156524',
                                                                                                'out_octets': '106486482961',
                                                                                                'out_ucast_pkts': '268087645',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '156524',
                                                                       'out_octets': '106486482961',
                                                                       'out_ucast_pkts': '268087645',
                                                                       'poller_hash': 'MTI3NDA1RXBlZXJpbnRmcG9uIDI0MTg4NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '241885',
                                                                                               'links': [       {       'connected_time': '8535318',
                                                                                                                        'fiber_length': 24181,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA1RXBlZXJwb25wb24gMjQxODg1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '60 PARKVIEW DR (CHATFIELD RICHARD)',
                                         'poller_hash': 'MTI3NDA1RQ==',
                                         'serial_number': '127405E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:f1',
                                         'uptime': '8535318'},
                                 {       'counter_source': '242135/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '242135/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96078903339',
                                                                       'in_ucast_pkts': '986579035',
                                                                       'mac_address': '04:bc:9f:5b:e1:81',
                                                                       'name': '242135/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '134',
                                                                       'out_octets': '6062122637425',
                                                                       'out_ucast_pkts': '4046278007',
                                                                       'poller_hash': 'MTI3NDJCMHBlZXJpbnRmMjQyMTM1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '242135',
                                                                       'in_nucast_pkts': '134',
                                                                       'in_octets': '6062122637425',
                                                                       'in_ucast_pkts': '4046278007',
                                                                       'links': [       {       'in_nucast_pkts': '134',
                                                                                                'in_octets': '6062122637425',
                                                                                                'in_ucast_pkts': '4046278007',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '96078903339',
                                                                                                'out_ucast_pkts': '986579035',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96078903339',
                                                                       'out_ucast_pkts': '986579035',
                                                                       'poller_hash': 'MTI3NDJCMHBlZXJpbnRmcG9uIDI0MjEzNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242135',
                                                                                               'links': [       {       'connected_time': '71352148',
                                                                                                                        'fiber_length': 20024,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJCMHBlZXJwb25wb24gMjQyMTM1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4957 PERTH ROAD 119 (KOCHER DON & '
                                                 'MARY)',
                                         'poller_hash': 'MTI3NDJCMA==',
                                         'serial_number': '12742B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e1:81',
                                         'uptime': '71352148'},
                                 {       'counter_source': '242205/g2',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'b8:94:70:9e:cd:d5'],
                                                                       'id': '242205/g2',
                                                                       'in_nucast_pkts': '160987',
                                                                       'in_octets': '419420874625',
                                                                       'in_ucast_pkts': '1060192317',
                                                                       'mac_address': '04:bc:9f:5c:22:d5',
                                                                       'mtu': 2000,
                                                                       'name': '242205/g2',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797914',
                                                                       'out_octets': '2848929861202',
                                                                       'out_ucast_pkts': '2411132103',
                                                                       'poller_hash': 'MTI3NEFEQXBlZXJpbnRmMjQyMjA1L2cy',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '242205',
                                                                       'in_nucast_pkts': '9797914',
                                                                       'in_octets': '2848929861202',
                                                                       'in_ucast_pkts': '2411132103',
                                                                       'links': [       {       'in_nucast_pkts': '9797914',
                                                                                                'in_octets': '2848929861202',
                                                                                                'in_ucast_pkts': '2411132103',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160987',
                                                                                                'out_octets': '419420874625',
                                                                                                'out_ucast_pkts': '1060192317',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160987',
                                                                       'out_octets': '419420874625',
                                                                       'out_ucast_pkts': '1060192317',
                                                                       'poller_hash': 'MTI3NEFEQXBlZXJpbnRmcG9uIDI0MjIwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242205',
                                                                                               'links': [       {       'connected_time': '8535272',
                                                                                                                        'fiber_length': 23256,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEFEQXBlZXJwb25wb24gMjQyMjA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '61 LAWRENCE STREET (DOPKO MATT)',
                                         'poller_hash': 'MTI3NEFEQQ==',
                                         'serial_number': '1274ADA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:22:d1',
                                         'uptime': '8535272'},
                                 {       'counter_source': '242708/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '7c:10:c9:02:8d:70'],
                                                                       'id': '242708/g1',
                                                                       'in_nucast_pkts': '381',
                                                                       'in_octets': '174684152709',
                                                                       'in_ucast_pkts': '234886406',
                                                                       'mac_address': '04:bc:9f:5b:d9:ec',
                                                                       'mtu': 2000,
                                                                       'name': '242708/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '714402',
                                                                       'out_octets': '528766016046',
                                                                       'out_ucast_pkts': '438421522',
                                                                       'poller_hash': 'MTI3NDFCRHBlZXJpbnRmMjQyNzA4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '242708',
                                                                       'in_nucast_pkts': '714402',
                                                                       'in_octets': '528766016046',
                                                                       'in_ucast_pkts': '438421522',
                                                                       'links': [       {       'in_nucast_pkts': '714402',
                                                                                                'in_octets': '528766016046',
                                                                                                'in_ucast_pkts': '438421522',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '381',
                                                                                                'out_octets': '174684152709',
                                                                                                'out_ucast_pkts': '234886406',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '381',
                                                                       'out_octets': '174684152709',
                                                                       'out_ucast_pkts': '234886406',
                                                                       'poller_hash': 'MTI3NDFCRHBlZXJpbnRmcG9uIDI0MjcwOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '242708',
                                                                                               'links': [       {       'connected_time': '611466',
                                                                                                                        'fiber_length': 19929,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -20.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDFCRHBlZXJwb25wb24gMjQyNzA4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4931 PERTH ROAD 119 (STECKLEY TERRY)',
                                         'poller_hash': 'MTI3NDFCRA==',
                                         'serial_number': '12741BD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d9:e9',
                                         'uptime': '611466'},
                                 {       'counter_source': '243681/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '243681/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '454964383673',
                                                                       'in_ucast_pkts': '182468851289',
                                                                       'mac_address': '04:bc:9f:5b:d8:71',
                                                                       'name': '243681/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '48',
                                                                       'out_octets': '6050323459431',
                                                                       'out_ucast_pkts': '4786241060',
                                                                       'poller_hash': 'MTI3NDE4RXBlZXJpbnRmMjQzNjgxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '243681',
                                                                       'in_nucast_pkts': '48',
                                                                       'in_octets': '6050323459431',
                                                                       'in_ucast_pkts': '4786241060',
                                                                       'links': [       {       'in_nucast_pkts': '48',
                                                                                                'in_octets': '6050323459431',
                                                                                                'in_ucast_pkts': '4786241060',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '454964383673',
                                                                                                'out_ucast_pkts': '182468851289',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '454964383673',
                                                                       'out_ucast_pkts': '182468851289',
                                                                       'poller_hash': 'MTI3NDE4RXBlZXJpbnRmcG9uIDI0MzY4MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243681',
                                                                                               'links': [       {       'connected_time': '8535176',
                                                                                                                        'fiber_length': 22856,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDE4RXBlZXJwb25wb24gMjQzNjgx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3637 NAFZIGER RD (HUNT KEVIN)',
                                         'poller_hash': 'MTI3NDE4RQ==',
                                         'serial_number': '127418E',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d8:71',
                                         'uptime': '8535176'},
                                 {       'counter_source': '243731/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:6b:2b'],
                                                                       'id': '243731/g1',
                                                                       'in_nucast_pkts': '162246',
                                                                       'in_octets': '118128095034',
                                                                       'in_ucast_pkts': '539602273',
                                                                       'mac_address': '04:bc:9f:5b:ce:44',
                                                                       'mtu': 2000,
                                                                       'name': '243731/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797732',
                                                                       'out_octets': '4399793259749',
                                                                       'out_ucast_pkts': '3270986244',
                                                                       'poller_hash': 'MTI3NDA0OHBlZXJpbnRmMjQzNzMxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '243731',
                                                                       'in_nucast_pkts': '9797732',
                                                                       'in_octets': '4399793259749',
                                                                       'in_ucast_pkts': '3270986244',
                                                                       'links': [       {       'in_nucast_pkts': '9797732',
                                                                                                'in_octets': '4399793259749',
                                                                                                'in_ucast_pkts': '3270986244',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '162246',
                                                                                                'out_octets': '118128095034',
                                                                                                'out_ucast_pkts': '539602273',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '162246',
                                                                       'out_octets': '118128095034',
                                                                       'out_ucast_pkts': '539602273',
                                                                       'poller_hash': 'MTI3NDA0OHBlZXJpbnRmcG9uIDI0MzczMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '243731',
                                                                                               'links': [       {       'connected_time': '8535208',
                                                                                                                        'fiber_length': 23083,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA0OHBlZXJwb25wb24gMjQzNzMx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '175 FERRIS DRIVE (WILSON GEOFF)',
                                         'poller_hash': 'MTI3NDA0OA==',
                                         'serial_number': '1274048',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:41',
                                         'uptime': '8535208'},
                                 {       'counter_source': '245867/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '245867/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '247397203422',
                                                                       'in_ucast_pkts': '1862054265',
                                                                       'mac_address': '04:bc:9f:5c:1c:e1',
                                                                       'name': '245867/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '7847577947118',
                                                                       'out_ucast_pkts': '5375807537',
                                                                       'poller_hash': 'MTI3NEExQ3BlZXJpbnRmMjQ1ODY3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '245867',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '7847577947118',
                                                                       'in_ucast_pkts': '5375807537',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '7847577947118',
                                                                                                'in_ucast_pkts': '5375807537',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '247397203422',
                                                                                                'out_ucast_pkts': '1862054265',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '247397203422',
                                                                       'out_ucast_pkts': '1862054265',
                                                                       'poller_hash': 'MTI3NEExQ3BlZXJpbnRmcG9uIDI0NTg2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '245867',
                                                                                               'links': [       {       'connected_time': '611452',
                                                                                                                        'fiber_length': 19711,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NEExQ3BlZXJwb25wb24gMjQ1ODY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4913 LINHAVEN DR (PLAYER BILL & '
                                                 'CINDIE)',
                                         'poller_hash': 'MTI3NEExQw==',
                                         'serial_number': '1274A1C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:1c:e1',
                                         'uptime': '611452'},
                                 {       'counter_source': '246447/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:86:d0:a8'],
                                                                       'id': '246447/g1',
                                                                       'in_nucast_pkts': '151828',
                                                                       'in_octets': '64158533927',
                                                                       'in_ucast_pkts': '282584671',
                                                                       'mac_address': '04:bc:9f:5b:cd:c4',
                                                                       'mtu': 2000,
                                                                       'name': '246447/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9877979',
                                                                       'out_octets': '1847512929831',
                                                                       'out_ucast_pkts': '1327856785',
                                                                       'poller_hash': 'MTI3NDAzOHBlZXJpbnRmMjQ2NDQ3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '246447',
                                                                       'in_nucast_pkts': '9877979',
                                                                       'in_octets': '1847512929831',
                                                                       'in_ucast_pkts': '1327856785',
                                                                       'links': [       {       'in_nucast_pkts': '9877979',
                                                                                                'in_octets': '1847512929831',
                                                                                                'in_ucast_pkts': '1327856785',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '151828',
                                                                                                'out_octets': '64158533927',
                                                                                                'out_ucast_pkts': '282584671',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '151828',
                                                                       'out_octets': '64158533927',
                                                                       'out_ucast_pkts': '282584671',
                                                                       'poller_hash': 'MTI3NDAzOHBlZXJpbnRmcG9uIDI0NjQ0Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246447',
                                                                                               'links': [       {       'connected_time': '8535106',
                                                                                                                        'fiber_length': 23636,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDAzOHBlZXJwb25wb24gMjQ2NDQ3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 CRAB APPLE CT (DIETRICH LINDSEY)',
                                         'poller_hash': 'MTI3NDAzOA==',
                                         'serial_number': '1274038',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cd:c1',
                                         'uptime': '8535106'},
                                 {       'counter_source': '246467/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '246467/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '290551080274',
                                                                       'in_ucast_pkts': '304899716',
                                                                       'mac_address': '1c:8b:76:00:37:71',
                                                                       'name': '246467/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '528399700125',
                                                                       'out_ucast_pkts': '488640813',
                                                                       'poller_hash': 'MTU4RkI2NHBlZXJpbnRmMjQ2NDY3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '246467',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '528399700125',
                                                                       'in_ucast_pkts': '488640813',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '528399700125',
                                                                                                'in_ucast_pkts': '488640813',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '290551080274',
                                                                                                'out_ucast_pkts': '304899716',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '290551080274',
                                                                       'out_ucast_pkts': '304899716',
                                                                       'poller_hash': 'MTU4RkI2NHBlZXJpbnRmcG9uIDI0NjQ2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246467',
                                                                                               'links': [       {       'connected_time': '68607123',
                                                                                                                        'fiber_length': 23427,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.81,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTU4RkI2NHBlZXJwb25wb24gMjQ2NDY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '717GE',
                                         'name': '3546 NAFZIGER RD (WELLESLEY '
                                                 'WOODWORKS)',
                                         'poller_hash': 'MTU4RkI2NA==',
                                         'serial_number': '158FB64',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '1c:8b:76:00:37:71',
                                         'uptime': '68607123'},
                                 {       'counter_source': '246940/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '246940/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '215158064750',
                                                                       'in_ucast_pkts': '2269997975',
                                                                       'mac_address': '04:bc:9f:5b:e5:29',
                                                                       'name': '246940/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '18',
                                                                       'out_octets': '6774036034251',
                                                                       'out_ucast_pkts': '4602187384',
                                                                       'poller_hash': 'MTI3NDMyNXBlZXJpbnRmMjQ2OTQwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '246940',
                                                                       'in_nucast_pkts': '18',
                                                                       'in_octets': '6774036034251',
                                                                       'in_ucast_pkts': '4602187384',
                                                                       'links': [       {       'in_nucast_pkts': '18',
                                                                                                'in_octets': '6774036034251',
                                                                                                'in_ucast_pkts': '4602187384',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '215158064750',
                                                                                                'out_ucast_pkts': '2269997975',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '215158064750',
                                                                       'out_ucast_pkts': '2269997975',
                                                                       'poller_hash': 'MTI3NDMyNXBlZXJpbnRmcG9uIDI0Njk0MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '246940',
                                                                                               'links': [       {       'connected_time': '5782919',
                                                                                                                        'fiber_length': 23081,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMyNXBlZXJwb25wb24gMjQ2OTQw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1096 HENRY ST (GERBER EUNICE)',
                                         'poller_hash': 'MTI3NDMyNQ==',
                                         'serial_number': '1274325',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e5:29',
                                         'uptime': '5782919'},
                                 {       'counter_source': '247000/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:6a:28'],
                                                                       'id': '247000/g1',
                                                                       'in_nucast_pkts': '81240',
                                                                       'in_octets': '55804787048',
                                                                       'in_ucast_pkts': '320590476',
                                                                       'mac_address': '04:bc:9f:5b:e4:b4',
                                                                       'mtu': 2000,
                                                                       'name': '247000/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4947346',
                                                                       'out_octets': '1382732837058',
                                                                       'out_ucast_pkts': '1030557206',
                                                                       'poller_hash': 'MTI3NDMxNnBlZXJpbnRmMjQ3MDAwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '247000',
                                                                       'in_nucast_pkts': '4947346',
                                                                       'in_octets': '1382732837058',
                                                                       'in_ucast_pkts': '1030557206',
                                                                       'links': [       {       'in_nucast_pkts': '4947346',
                                                                                                'in_octets': '1382732837058',
                                                                                                'in_ucast_pkts': '1030557206',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '81240',
                                                                                                'out_octets': '55804787048',
                                                                                                'out_ucast_pkts': '320590476',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '81240',
                                                                       'out_octets': '55804787048',
                                                                       'out_ucast_pkts': '320590476',
                                                                       'poller_hash': 'MTI3NDMxNnBlZXJpbnRmcG9uIDI0NzAwMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247000',
                                                                                               'links': [       {       'connected_time': '4213427',
                                                                                                                        'fiber_length': 24979,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDMxNnBlZXJwb25wb24gMjQ3MDAw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '83 GERBER MEADOWS DR (MARTIN ANDREW '
                                                 'J)',
                                         'poller_hash': 'MTI3NDMxNg==',
                                         'serial_number': '1274316',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e4:b1',
                                         'uptime': '4213427'},
                                 {       'counter_source': '247020/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '247020/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '7277027893047',
                                                                       'in_ucast_pkts': '8828878048',
                                                                       'mac_address': '04:bc:9f:5b:c6:c1',
                                                                       'name': '247020/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '38',
                                                                       'out_octets': '12658474929845',
                                                                       'out_ucast_pkts': '11948342109',
                                                                       'poller_hash': 'MTI3M0Y1OHBlZXJpbnRmMjQ3MDIwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '247020',
                                                                       'in_nucast_pkts': '38',
                                                                       'in_octets': '12658474929845',
                                                                       'in_ucast_pkts': '11948342109',
                                                                       'links': [       {       'in_nucast_pkts': '38',
                                                                                                'in_octets': '12658474929845',
                                                                                                'in_ucast_pkts': '11948342109',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '7277027893047',
                                                                                                'out_ucast_pkts': '8828878048',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '7277027893047',
                                                                       'out_ucast_pkts': '8828878048',
                                                                       'poller_hash': 'MTI3M0Y1OHBlZXJpbnRmcG9uIDI0NzAyMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247020',
                                                                                               'links': [       {       'connected_time': '2994217',
                                                                                                                        'fiber_length': 23449,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0Y1OHBlZXJwb25wb24gMjQ3MDIw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '162 LAWRENCE ST WELLESLEY (STENDEL '
                                                 'TROY)',
                                         'poller_hash': 'MTI3M0Y1OA==',
                                         'serial_number': '1273F58',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:c6:c1',
                                         'uptime': '2994217'},
                                 {       'counter_source': '247125/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:73:86'],
                                                                       'id': '247125/g1',
                                                                       'in_nucast_pkts': '21428',
                                                                       'in_octets': '423782856949',
                                                                       'in_ucast_pkts': '433780937',
                                                                       'mac_address': '04:bc:9f:5b:ce:bc',
                                                                       'mtu': 2000,
                                                                       'name': '247125/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1188985',
                                                                       'out_octets': '464155469959',
                                                                       'out_ucast_pkts': '476829617',
                                                                       'poller_hash': 'MTI3NDA1N3BlZXJpbnRmMjQ3MTI1L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '247125',
                                                                       'in_nucast_pkts': '1188985',
                                                                       'in_octets': '464155469959',
                                                                       'in_ucast_pkts': '476829617',
                                                                       'links': [       {       'in_nucast_pkts': '1188985',
                                                                                                'in_octets': '464155469959',
                                                                                                'in_ucast_pkts': '476829617',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '21428',
                                                                                                'out_octets': '423782856949',
                                                                                                'out_ucast_pkts': '433780937',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '21428',
                                                                       'out_octets': '423782856949',
                                                                       'out_ucast_pkts': '433780937',
                                                                       'poller_hash': 'MTI3NDA1N3BlZXJpbnRmcG9uIDI0NzEyNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247125',
                                                                                               'links': [       {       'connected_time': '1015547',
                                                                                                                        'fiber_length': 23772,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA1N3BlZXJwb25wb24gMjQ3MTI1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '50 SCHWEITZER CR (MOORE DARRYL W)',
                                         'poller_hash': 'MTI3NDA1Nw==',
                                         'serial_number': '1274057',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ce:b9',
                                         'uptime': '1015547'},
                                 {       'counter_source': '247592/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:78:56'],
                                                                       'id': '247592/g1',
                                                                       'in_nucast_pkts': '163451',
                                                                       'in_octets': '71608203986',
                                                                       'in_ucast_pkts': '408059948',
                                                                       'mac_address': '04:bc:9f:5b:cf:7c',
                                                                       'mtu': 2000,
                                                                       'name': '247592/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919822',
                                                                       'out_octets': '3442558103631',
                                                                       'out_ucast_pkts': '2405154705',
                                                                       'poller_hash': 'MTI3NDA2RnBlZXJpbnRmMjQ3NTkyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '247592',
                                                                       'in_nucast_pkts': '9919822',
                                                                       'in_octets': '3442558103631',
                                                                       'in_ucast_pkts': '2405154705',
                                                                       'links': [       {       'in_nucast_pkts': '9919822',
                                                                                                'in_octets': '3442558103631',
                                                                                                'in_ucast_pkts': '2405154705',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '163451',
                                                                                                'out_octets': '71608203986',
                                                                                                'out_ucast_pkts': '408059948',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '163451',
                                                                       'out_octets': '71608203986',
                                                                       'out_ucast_pkts': '408059948',
                                                                       'poller_hash': 'MTI3NDA2RnBlZXJpbnRmcG9uIDI0NzU5Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247592',
                                                                                               'links': [       {       'connected_time': '8535176',
                                                                                                                        'fiber_length': 22569,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDA2RnBlZXJwb25wb24gMjQ3NTky',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '64 POND VIEW DR (DELUTIS GREG)',
                                         'poller_hash': 'MTI3NDA2Rg==',
                                         'serial_number': '127406F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:cf:79',
                                         'uptime': '8535176'},
                                 {       'counter_source': '247602/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:71:13'],
                                                                       'id': '247602/g1',
                                                                       'in_nucast_pkts': '159357',
                                                                       'in_octets': '46587927477',
                                                                       'in_ucast_pkts': '256367096',
                                                                       'mac_address': '04:bc:9f:5b:e2:84',
                                                                       'mtu': 2000,
                                                                       'name': '247602/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919973',
                                                                       'out_octets': '1290852796582',
                                                                       'out_ucast_pkts': '905524667',
                                                                       'poller_hash': 'MTI3NDJEMHBlZXJpbnRmMjQ3NjAyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '247602',
                                                                       'in_nucast_pkts': '9919973',
                                                                       'in_octets': '1290852796582',
                                                                       'in_ucast_pkts': '905524667',
                                                                       'links': [       {       'in_nucast_pkts': '9919973',
                                                                                                'in_octets': '1290852796582',
                                                                                                'in_ucast_pkts': '905524667',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '159357',
                                                                                                'out_octets': '46587927477',
                                                                                                'out_ucast_pkts': '256367096',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '159357',
                                                                       'out_octets': '46587927477',
                                                                       'out_ucast_pkts': '256367096',
                                                                       'poller_hash': 'MTI3NDJEMHBlZXJpbnRmcG9uIDI0NzYwMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '247602',
                                                                                               'links': [       {       'connected_time': '8535212',
                                                                                                                        'fiber_length': 22525,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJEMHBlZXJwb25wb24gMjQ3NjAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 WATER ST (FUTHER DAVID)',
                                         'poller_hash': 'MTI3NDJEMA==',
                                         'serial_number': '12742D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e2:81',
                                         'uptime': '8535212'},
                                 {       'counter_source': '248443/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:79:11'],
                                                                       'id': '248443/g1',
                                                                       'in_nucast_pkts': '160555',
                                                                       'in_octets': '144291648539',
                                                                       'in_ucast_pkts': '267112540',
                                                                       'mac_address': '04:bc:9f:5b:e1:c4',
                                                                       'mtu': 2000,
                                                                       'name': '248443/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9798010',
                                                                       'out_octets': '965629743923',
                                                                       'out_ucast_pkts': '784376468',
                                                                       'poller_hash': 'MTI3NDJCOHBlZXJpbnRmMjQ4NDQzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248443',
                                                                       'in_nucast_pkts': '9798010',
                                                                       'in_octets': '965629743923',
                                                                       'in_ucast_pkts': '784376468',
                                                                       'links': [       {       'in_nucast_pkts': '9798010',
                                                                                                'in_octets': '965629743923',
                                                                                                'in_ucast_pkts': '784376468',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '160555',
                                                                                                'out_octets': '144291648539',
                                                                                                'out_ucast_pkts': '267112540',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '160555',
                                                                       'out_octets': '144291648539',
                                                                       'out_ucast_pkts': '267112540',
                                                                       'poller_hash': 'MTI3NDJCOHBlZXJpbnRmcG9uIDI0ODQ0Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248443',
                                                                                               'links': [       {       'connected_time': '8535243',
                                                                                                                        'fiber_length': 23345,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDJCOHBlZXJwb25wb24gMjQ4NDQz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '146 LAWRENCE ST (HEAD MIKE)',
                                         'poller_hash': 'MTI3NDJCOA==',
                                         'serial_number': '12742B8',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:e1:c1',
                                         'uptime': '8535243'},
                                 {       'counter_source': '248605/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:8b:68:c8'],
                                                                       'id': '248605/g1',
                                                                       'in_nucast_pkts': '19667',
                                                                       'in_octets': '9582782205',
                                                                       'in_ucast_pkts': '23431773',
                                                                       'mac_address': '04:bc:9f:5b:d4:44',
                                                                       'mtu': 2000,
                                                                       'name': '248605/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1284711',
                                                                       'out_octets': '57950917193',
                                                                       'out_ucast_pkts': '53794656',
                                                                       'poller_hash': 'MTI3NDEwOHBlZXJpbnRmMjQ4NjA1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '248605',
                                                                       'in_nucast_pkts': '1284711',
                                                                       'in_octets': '57950917193',
                                                                       'in_ucast_pkts': '53794656',
                                                                       'links': [       {       'in_nucast_pkts': '1284711',
                                                                                                'in_octets': '57950917193',
                                                                                                'in_ucast_pkts': '53794656',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '19667',
                                                                                                'out_octets': '9582782205',
                                                                                                'out_ucast_pkts': '23431773',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '19667',
                                                                       'out_octets': '9582782205',
                                                                       'out_ucast_pkts': '23431773',
                                                                       'poller_hash': 'MTI3NDEwOHBlZXJpbnRmcG9uIDI0ODYwNQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '248605',
                                                                                               'links': [       {       'connected_time': '1096769',
                                                                                                                        'fiber_length': 29772,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDEwOHBlZXJwb25wb24gMjQ4NjA1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4576 WEIMAR LINE (WILZ HERMANN & '
                                                 'LISA)',
                                         'poller_hash': 'MTI3NDEwOA==',
                                         'serial_number': '1274108',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:d4:41',
                                         'uptime': '1096769'},
                                 {       'counter_source': '249827/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:89:0b:b5'],
                                                                       'id': '249827/g1',
                                                                       'in_nucast_pkts': '1832',
                                                                       'in_octets': '212165581250',
                                                                       'in_ucast_pkts': '693217410',
                                                                       'mac_address': '48:77:46:a3:bb:58',
                                                                       'mtu': 2000,
                                                                       'name': '249827/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15890294',
                                                                       'out_octets': '3287226739040',
                                                                       'out_ucast_pkts': '2518468212',
                                                                       'poller_hash': 'NzFEODk5cGVlcmludGYyNDk4MjcvZzE=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '249827',
                                                                       'in_nucast_pkts': '15890294',
                                                                       'in_octets': '3287226739040',
                                                                       'in_ucast_pkts': '2518468212',
                                                                       'links': [       {       'in_nucast_pkts': '15890294',
                                                                                                'in_octets': '3287226739040',
                                                                                                'in_ucast_pkts': '2518468212',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '1832',
                                                                                                'out_octets': '212165581250',
                                                                                                'out_ucast_pkts': '693217410',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1832',
                                                                       'out_octets': '212165581250',
                                                                       'out_ucast_pkts': '693217410',
                                                                       'poller_hash': 'NzFEODk5cGVlcmludGZwb24gMjQ5ODI3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '249827',
                                                                                               'links': [       {       'connected_time': '25803693',
                                                                                                                        'fiber_length': 24692,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.92,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.478}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -3.0,
                                                                                               'minimum_loss': -23.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NzFEODk5cGVlcnBvbnBvbiAyNDk4Mjc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '716GE',
                                         'name': '135 GERBER MEADOWS DRIVE (LOBRUTTO '
                                                 'JOANNE)',
                                         'poller_hash': 'NzFEODk5',
                                         'serial_number': '71D899',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '10.8.110.25',
                                         'system_mac_address': '48:77:46:a3:bb:54',
                                         'uptime': '25803693'},
                                 {       'counter_source': '250074/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:eb:73'],
                                                                       'id': '250074/g1',
                                                                       'in_nucast_pkts': '149230',
                                                                       'in_octets': '117338056572',
                                                                       'in_ucast_pkts': '453410342',
                                                                       'mac_address': 'b8:94:70:a1:4b:00',
                                                                       'mtu': 2000,
                                                                       'name': '250074/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797998',
                                                                       'out_octets': '2063823211438',
                                                                       'out_ucast_pkts': '1555444607',
                                                                       'poller_hash': 'MTE3REI2QnBlZXJpbnRmMjUwMDc0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '250074',
                                                                       'in_nucast_pkts': '9797998',
                                                                       'in_octets': '2063823211438',
                                                                       'in_ucast_pkts': '1555444607',
                                                                       'links': [       {       'in_nucast_pkts': '9797998',
                                                                                                'in_octets': '2063823211438',
                                                                                                'in_ucast_pkts': '1555444607',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149230',
                                                                                                'out_octets': '117338056572',
                                                                                                'out_ucast_pkts': '453410342',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149230',
                                                                       'out_octets': '117338056572',
                                                                       'out_ucast_pkts': '453410342',
                                                                       'poller_hash': 'MTE3REI2QnBlZXJpbnRmcG9uIDI1MDA3NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '250074',
                                                                                               'links': [       {       'connected_time': '8535293',
                                                                                                                        'fiber_length': 23044,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI2QnBlZXJwb25wb24gMjUwMDc0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '102 LAWRENCE ST (WELCHES MARGARET T)',
                                         'poller_hash': 'MTE3REI2Qg==',
                                         'serial_number': '117DB6B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:fd',
                                         'uptime': '8535293'},
                                 {       'counter_source': '251589/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:1f:63:c7'],
                                                                       'id': '251589/g1',
                                                                       'in_nucast_pkts': '159307',
                                                                       'in_octets': '40411010367',
                                                                       'in_ucast_pkts': '338462182',
                                                                       'mac_address': 'b8:94:70:a1:4a:f0',
                                                                       'mtu': 2000,
                                                                       'name': '251589/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878022',
                                                                       'out_octets': '2094792126570',
                                                                       'out_ucast_pkts': '1435400185',
                                                                       'poller_hash': 'MTE3REI2OXBlZXJpbnRmMjUxNTg5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '251589',
                                                                       'in_nucast_pkts': '9878022',
                                                                       'in_octets': '2094792126570',
                                                                       'in_ucast_pkts': '1435400185',
                                                                       'links': [       {       'in_nucast_pkts': '9878022',
                                                                                                'in_octets': '2094792126570',
                                                                                                'in_ucast_pkts': '1435400185',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '159307',
                                                                                                'out_octets': '40411010367',
                                                                                                'out_ucast_pkts': '338462182',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '159307',
                                                                       'out_octets': '40411010367',
                                                                       'out_ucast_pkts': '338462182',
                                                                       'poller_hash': 'MTE3REI2OXBlZXJpbnRmcG9uIDI1MTU4OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '251589',
                                                                                               'links': [       {       'connected_time': '8535194',
                                                                                                                        'fiber_length': 22908,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI2OXBlZXJwb25wb24gMjUxNTg5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1010 DOERING STREET (GUENTHER ROY)',
                                         'poller_hash': 'MTE3REI2OQ==',
                                         'serial_number': '117DB69',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:ed',
                                         'uptime': '8535194'},
                                 {       'counter_source': '252650/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:92:c8:fd'],
                                                                       'id': '252650/g1',
                                                                       'in_nucast_pkts': '2675572',
                                                                       'in_octets': '122570986623',
                                                                       'in_ucast_pkts': '445860606',
                                                                       'mac_address': 'b8:94:70:a1:4f:40',
                                                                       'mtu': 2000,
                                                                       'name': '252650/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3008779',
                                                                       'out_octets': '1355670044771',
                                                                       'out_ucast_pkts': '1112158645',
                                                                       'poller_hash': 'MTE3REJGM3BlZXJpbnRmMjUyNjUwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '252650',
                                                                       'in_nucast_pkts': '3008779',
                                                                       'in_octets': '1355670044771',
                                                                       'in_ucast_pkts': '1112158645',
                                                                       'links': [       {       'in_nucast_pkts': '3008779',
                                                                                                'in_octets': '1355670044771',
                                                                                                'in_ucast_pkts': '1112158645',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2675572',
                                                                                                'out_octets': '122570986623',
                                                                                                'out_ucast_pkts': '445860606',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2675572',
                                                                       'out_octets': '122570986623',
                                                                       'out_ucast_pkts': '445860606',
                                                                       'poller_hash': 'MTE3REJGM3BlZXJpbnRmcG9uIDI1MjY1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252650',
                                                                                               'links': [       {       'connected_time': '2565938',
                                                                                                                        'fiber_length': 24882,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJGM3BlZXJwb25wb24gMjUyNjUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '21 PARKVIEW DR (LUKAS MATTHEW)',
                                         'poller_hash': 'MTE3REJGMw==',
                                         'serial_number': '117DBF3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4f:3d',
                                         'uptime': '2565938'},
                                 {       'counter_source': '252660/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '252660/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '444009922373',
                                                                       'in_ucast_pkts': '983044374',
                                                                       'mac_address': 'b8:94:70:a1:51:0d',
                                                                       'name': '252660/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3952282178779',
                                                                       'out_ucast_pkts': '3267486745',
                                                                       'poller_hash': 'MTE3REMyRHBlZXJpbnRmMjUyNjYwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '252660',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3952282178779',
                                                                       'in_ucast_pkts': '3267486745',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3952282178779',
                                                                                                'in_ucast_pkts': '3267486745',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '444009922373',
                                                                                                'out_ucast_pkts': '983044374',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '444009922373',
                                                                       'out_ucast_pkts': '983044374',
                                                                       'poller_hash': 'MTE3REMyRHBlZXJpbnRmcG9uIDI1MjY2MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '252660',
                                                                                               'links': [       {       'connected_time': '10977854',
                                                                                                                        'fiber_length': 24694,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMyRHBlZXJwb25wb24gMjUyNjYw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '138 GERBER MEADOWS DR (SMITH '
                                                 'GREGORY)',
                                         'poller_hash': 'MTE3REMyRA==',
                                         'serial_number': '117DC2D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:51:0d',
                                         'uptime': '10977854'},
                                 {       'counter_source': '253426/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '44:65:7f:79:dc:fc'],
                                                                       'id': '253426/g1',
                                                                       'in_nucast_pkts': '9613557',
                                                                       'in_octets': '260172244636',
                                                                       'in_ucast_pkts': '600574925',
                                                                       'mac_address': 'b8:94:70:a1:51:38',
                                                                       'mtu': 2000,
                                                                       'name': '253426/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878367',
                                                                       'out_octets': '1589658642773',
                                                                       'out_ucast_pkts': '1276775125',
                                                                       'poller_hash': 'MTE3REMzMnBlZXJpbnRmMjUzNDI2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253426',
                                                                       'in_nucast_pkts': '9878367',
                                                                       'in_octets': '1589658642773',
                                                                       'in_ucast_pkts': '1276775125',
                                                                       'links': [       {       'in_nucast_pkts': '9878367',
                                                                                                'in_octets': '1589658642773',
                                                                                                'in_ucast_pkts': '1276775125',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '9613557',
                                                                                                'out_octets': '260172244636',
                                                                                                'out_ucast_pkts': '600574925',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9613557',
                                                                       'out_octets': '260172244636',
                                                                       'out_ucast_pkts': '600574925',
                                                                       'poller_hash': 'MTE3REMzMnBlZXJpbnRmcG9uIDI1MzQyNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253426',
                                                                                               'links': [       {       'connected_time': '8535221',
                                                                                                                        'fiber_length': 25349,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMzMnBlZXJwb25wb24gMjUzNDI2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '26 JACK KOEHLER LINE (NG ALBERT)',
                                         'poller_hash': 'MTE3REMzMg==',
                                         'serial_number': '117DC32',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:51:35',
                                         'uptime': '8535221'},
                                 {       'counter_source': '253577/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:74:ca'],
                                                                       'id': '253577/g1',
                                                                       'in_nucast_pkts': '152125',
                                                                       'in_octets': '196325902412',
                                                                       'in_ucast_pkts': '534327456',
                                                                       'mac_address': 'b8:94:70:a1:4d:08',
                                                                       'mtu': 2000,
                                                                       'name': '253577/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9919930',
                                                                       'out_octets': '2314263681628',
                                                                       'out_ucast_pkts': '1759068029',
                                                                       'poller_hash': 'MTE3REJBQ3BlZXJpbnRmMjUzNTc3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '253577',
                                                                       'in_nucast_pkts': '9919930',
                                                                       'in_octets': '2314263681628',
                                                                       'in_ucast_pkts': '1759068029',
                                                                       'links': [       {       'in_nucast_pkts': '9919930',
                                                                                                'in_octets': '2314263681628',
                                                                                                'in_ucast_pkts': '1759068029',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '152125',
                                                                                                'out_octets': '196325902412',
                                                                                                'out_ucast_pkts': '534327456',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '152125',
                                                                       'out_octets': '196325902412',
                                                                       'out_ucast_pkts': '534327456',
                                                                       'poller_hash': 'MTE3REJBQ3BlZXJpbnRmcG9uIDI1MzU3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '253577',
                                                                                               'links': [       {       'connected_time': '8535270',
                                                                                                                        'fiber_length': 22465,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REJBQ3BlZXJwb25wb24gMjUzNTc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1328 QUEEN'S BUSH RD (DE ARTOLA "
                                                 'THOMAS)',
                                         'poller_hash': 'MTE3REJBQw==',
                                         'serial_number': '117DBAC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4d:05',
                                         'uptime': '8535270'},
                                 {       'counter_source': '254785/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '254785/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '225088581',
                                                                       'in_ucast_pkts': '974368',
                                                                       'mac_address': 'b8:94:70:a1:4a:9d',
                                                                       'name': '254785/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '5222805121',
                                                                       'out_ucast_pkts': '14618630',
                                                                       'poller_hash': 'MTE3REI1RnBlZXJpbnRmMjU0Nzg1L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '254785',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '5222805121',
                                                                       'in_ucast_pkts': '14618630',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '5222805121',
                                                                                                'in_ucast_pkts': '14618630',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '225088581',
                                                                                                'out_ucast_pkts': '974368',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '225088581',
                                                                       'out_ucast_pkts': '974368',
                                                                       'poller_hash': 'MTE3REI1RnBlZXJpbnRmcG9uIDI1NDc4NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '254785',
                                                                                               'links': [       {       'connected_time': '27382858',
                                                                                                                        'fiber_length': 15504,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REI1RnBlZXJwb25wb24gMjU0Nzg1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5277 PERTH ROAD 119 (SUN RISE '
                                                 'SEAMLESS EAVESTROUGHI)',
                                         'poller_hash': 'MTE3REI1Rg==',
                                         'serial_number': '117DB5F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:4a:9d',
                                         'uptime': '27382858'},
                                 {       'counter_source': '256222/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:3d:51:70'],
                                                                       'id': '256222/g1',
                                                                       'in_nucast_pkts': '2154865',
                                                                       'in_octets': '16348951335',
                                                                       'in_ucast_pkts': '210940068',
                                                                       'mac_address': 'b8:94:70:a1:46:60',
                                                                       'mtu': 2000,
                                                                       'name': '256222/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '82833959',
                                                                       'out_octets': '887048323587',
                                                                       'out_ucast_pkts': '607966679',
                                                                       'poller_hash': 'MTE3REFEN3BlZXJpbnRmMjU2MjIyL2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256222',
                                                                       'in_nucast_pkts': '82833959',
                                                                       'in_octets': '887048323587',
                                                                       'in_ucast_pkts': '607966679',
                                                                       'links': [       {       'in_nucast_pkts': '82833959',
                                                                                                'in_octets': '887048323587',
                                                                                                'in_ucast_pkts': '607966679',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '2154865',
                                                                                                'out_octets': '16348951335',
                                                                                                'out_ucast_pkts': '210940068',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2154865',
                                                                       'out_octets': '16348951335',
                                                                       'out_ucast_pkts': '210940068',
                                                                       'poller_hash': 'MTE3REFEN3BlZXJpbnRmcG9uIDI1NjIyMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256222',
                                                                                               'links': [       {       'connected_time': '4933375',
                                                                                                                        'fiber_length': 24852,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REFEN3BlZXJwb25wb24gMjU2MjIy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '19 PARKVIEW DR WELLESLEY (SCHUMACHER '
                                                 'RICHARD)',
                                         'poller_hash': 'MTE3REFENw==',
                                         'serial_number': '117DAD7',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:46:5d',
                                         'uptime': '4933375'},
                                 {       'counter_source': '256242/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:0c:29:e3:f7:44'],
                                                                       'id': '256242/g1',
                                                                       'in_nucast_pkts': '296146',
                                                                       'in_octets': '24423615755826',
                                                                       'in_ucast_pkts': '29910287731',
                                                                       'mac_address': '04:bc:9f:6e:db:91',
                                                                       'mtu': 2000,
                                                                       'name': '256242/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9770471',
                                                                       'out_octets': '19311386210823',
                                                                       'out_ucast_pkts': '26191647467',
                                                                       'poller_hash': 'MTI5NkVGNXBlZXJpbnRmMjU2MjQyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256242',
                                                                       'in_nucast_pkts': '9770471',
                                                                       'in_octets': '19311386210823',
                                                                       'in_ucast_pkts': '26191647467',
                                                                       'links': [       {       'in_nucast_pkts': '9770471',
                                                                                                'in_octets': '19311386210823',
                                                                                                'in_ucast_pkts': '26191647467',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '296146',
                                                                                                'out_octets': '24423615755826',
                                                                                                'out_ucast_pkts': '29910287731',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '296146',
                                                                       'out_octets': '24423615755826',
                                                                       'out_ucast_pkts': '29910287731',
                                                                       'poller_hash': 'MTI5NkVGNXBlZXJpbnRmcG9uIDI1NjI0Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256242',
                                                                                               'links': [       {       'connected_time': '8535449',
                                                                                                                        'fiber_length': 25436,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NkVGNXBlZXJwb25wb24gMjU2MjQy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 JACK KOEHLER LN (MACMILLAN BRAD)',
                                         'poller_hash': 'MTI5NkVGNQ==',
                                         'serial_number': '1296EF5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:db:8e',
                                         'uptime': '8535449'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '256302/g1',
                                                                       'name': '256302/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5MzdGcGVlcmludGYyNTYzMDIvZzE='},
                                                               {       'admin_up': True,
                                                                       'id': '256302',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'ODY5MzdGcGVlcmludGZwb24gMjU2MzAy',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256302',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'ODY5MzdGcGVlcnBvbnBvbiAyNTYzMDI=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '16 MILL ST E (MORNINGTON '
                                                 'COMMUNICATIONS)',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "256302" '
                                                                  '(location: GPON '
                                                                  'port "7/8"), "ONT '
                                                                  'went missing"',
                                         'offline_time': '2025-01-27T16:15:52.130Z',
                                         'poller_hash': 'ODY5MzdG',
                                         'serial_number': '86937F',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '256373/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:72:d0'],
                                                                       'id': '256373/g1',
                                                                       'in_nucast_pkts': '465894',
                                                                       'in_octets': '287876090658',
                                                                       'in_ucast_pkts': '570888221',
                                                                       'mac_address': '04:bc:9f:6f:07:01',
                                                                       'mtu': 2000,
                                                                       'name': '256373/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29819151',
                                                                       'out_octets': '1944623996488',
                                                                       'out_ucast_pkts': '1621166389',
                                                                       'poller_hash': 'MTI5NzQ2M3BlZXJpbnRmMjU2MzczL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256373',
                                                                       'in_nucast_pkts': '29819151',
                                                                       'in_octets': '1944623996488',
                                                                       'in_ucast_pkts': '1621166389',
                                                                       'links': [       {       'in_nucast_pkts': '29819151',
                                                                                                'in_octets': '1944623996488',
                                                                                                'in_ucast_pkts': '1621166389',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '465894',
                                                                                                'out_octets': '287876090658',
                                                                                                'out_ucast_pkts': '570888221',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '465894',
                                                                       'out_octets': '287876090658',
                                                                       'out_ucast_pkts': '570888221',
                                                                       'poller_hash': 'MTI5NzQ2M3BlZXJpbnRmcG9uIDI1NjM3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256373',
                                                                                               'links': [       {       'connected_time': '25827623',
                                                                                                                        'fiber_length': 25191,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ2M3BlZXJwb25wb24gMjU2Mzcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 WELWOOD AV (HARNACK ROBERT)',
                                         'poller_hash': 'MTI5NzQ2Mw==',
                                         'serial_number': '1297463',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:06:fe',
                                         'uptime': '25827623'},
                                 {       'counter_source': '256432/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:77:8a'],
                                                                       'id': '256432/g1',
                                                                       'in_nucast_pkts': '471911',
                                                                       'in_octets': '110916545356',
                                                                       'in_ucast_pkts': '1105502636',
                                                                       'mac_address': 'b8:94:70:a1:50:a0',
                                                                       'mtu': 2000,
                                                                       'name': '256432/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '29360794',
                                                                       'out_octets': '7259752310644',
                                                                       'out_ucast_pkts': '4948570877',
                                                                       'poller_hash': 'MTE3REMxRnBlZXJpbnRmMjU2NDMyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256432',
                                                                       'in_nucast_pkts': '29360794',
                                                                       'in_octets': '7259752310644',
                                                                       'in_ucast_pkts': '4948570877',
                                                                       'links': [       {       'in_nucast_pkts': '29360794',
                                                                                                'in_octets': '7259752310644',
                                                                                                'in_ucast_pkts': '4948570877',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '471911',
                                                                                                'out_octets': '110916545356',
                                                                                                'out_ucast_pkts': '1105502636',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '471911',
                                                                       'out_octets': '110916545356',
                                                                       'out_ucast_pkts': '1105502636',
                                                                       'poller_hash': 'MTE3REMxRnBlZXJpbnRmcG9uIDI1NjQzMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256432',
                                                                                               'links': [       {       'connected_time': '25393059',
                                                                                                                        'fiber_length': 23190,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE3REMxRnBlZXJwb25wb24gMjU2NDMy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 POND VIEW DRIVE UNIT 201 (BAST '
                                                 'DARLENE)',
                                         'poller_hash': 'MTE3REMxRg==',
                                         'serial_number': '117DC1F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:a1:50:9d',
                                         'uptime': '25393059'},
                                 {       'counter_source': '256502/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '256502/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '13009538606',
                                                                       'in_ucast_pkts': '71916281',
                                                                       'mac_address': '04:bc:9f:6f:0b:3e',
                                                                       'name': '256502/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '33',
                                                                       'out_octets': '42795287222',
                                                                       'out_ucast_pkts': '85375019',
                                                                       'poller_hash': 'MTI5NzRFQnBlZXJpbnRmMjU2NTAyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '256502',
                                                                       'in_nucast_pkts': '33',
                                                                       'in_octets': '42795287222',
                                                                       'in_ucast_pkts': '85375019',
                                                                       'links': [       {       'in_nucast_pkts': '33',
                                                                                                'in_octets': '42795287222',
                                                                                                'in_ucast_pkts': '85375019',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '13009538606',
                                                                                                'out_ucast_pkts': '71916281',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '13009538606',
                                                                       'out_ucast_pkts': '71916281',
                                                                       'poller_hash': 'MTI5NzRFQnBlZXJpbnRmcG9uIDI1NjUwMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256502',
                                                                                               'links': [       {       'connected_time': '4232853',
                                                                                                                        'fiber_length': 18036,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRFQnBlZXJwb25wb24gMjU2NTAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2018 PERTH LINE (SNELL EMILY)',
                                         'poller_hash': 'MTI5NzRFQg==',
                                         'serial_number': '12974EB',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:0b:3e',
                                         'uptime': '4232853'},
                                 {       'counter_source': '256746/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:71:86'],
                                                                       'id': '256746/g1',
                                                                       'in_nucast_pkts': '155108',
                                                                       'in_octets': '60970917607',
                                                                       'in_ucast_pkts': '336360467',
                                                                       'mac_address': '04:bc:9f:6f:09:01',
                                                                       'mtu': 2000,
                                                                       'name': '256746/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9797748',
                                                                       'out_octets': '1752589577869',
                                                                       'out_ucast_pkts': '1297173095',
                                                                       'poller_hash': 'MTI5NzRBM3BlZXJpbnRmMjU2NzQ2L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256746',
                                                                       'in_nucast_pkts': '9797748',
                                                                       'in_octets': '1752589577869',
                                                                       'in_ucast_pkts': '1297173095',
                                                                       'links': [       {       'in_nucast_pkts': '9797748',
                                                                                                'in_octets': '1752589577869',
                                                                                                'in_ucast_pkts': '1297173095',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '155108',
                                                                                                'out_octets': '60970917607',
                                                                                                'out_ucast_pkts': '336360467',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '155108',
                                                                       'out_octets': '60970917607',
                                                                       'out_ucast_pkts': '336360467',
                                                                       'poller_hash': 'MTI5NzRBM3BlZXJpbnRmcG9uIDI1Njc0Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256746',
                                                                                               'links': [       {       'connected_time': '8535330',
                                                                                                                        'fiber_length': 23429,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzRBM3BlZXJwb25wb24gMjU2NzQ2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '37 LAWRENCE ST (VAN DER MAAS PETER)',
                                         'poller_hash': 'MTI5NzRBMw==',
                                         'serial_number': '12974A3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:08:fe',
                                         'uptime': '8535330'},
                                 {       'counter_source': '256877/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:6d:c9'],
                                                                       'id': '256877/g1',
                                                                       'in_nucast_pkts': '113292',
                                                                       'in_octets': '97286440669',
                                                                       'in_ucast_pkts': '281517829',
                                                                       'mac_address': '04:bc:9f:6f:08:d1',
                                                                       'mtu': 2000,
                                                                       'name': '256877/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7365605',
                                                                       'out_octets': '1586391557925',
                                                                       'out_ucast_pkts': '1197206085',
                                                                       'poller_hash': 'MTI5NzQ5RHBlZXJpbnRmMjU2ODc3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '256877',
                                                                       'in_nucast_pkts': '7365605',
                                                                       'in_octets': '1586391557925',
                                                                       'in_ucast_pkts': '1197206085',
                                                                       'links': [       {       'in_nucast_pkts': '7365605',
                                                                                                'in_octets': '1586391557925',
                                                                                                'in_ucast_pkts': '1197206085',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '113292',
                                                                                                'out_octets': '97286440669',
                                                                                                'out_ucast_pkts': '281517829',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '113292',
                                                                       'out_octets': '97286440669',
                                                                       'out_ucast_pkts': '281517829',
                                                                       'poller_hash': 'MTI5NzQ5RHBlZXJpbnRmcG9uIDI1Njg3Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '256877',
                                                                                               'links': [       {       'connected_time': '6313818',
                                                                                                                        'fiber_length': 25306,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ5RHBlZXJwb25wb24gMjU2ODc3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '36 WELWOOD AV (GRUBER JEFF)',
                                         'poller_hash': 'MTI5NzQ5RA==',
                                         'serial_number': '129749D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:08:ce',
                                         'uptime': '6313818'},
                                 {       'counter_source': '257468/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257468/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '257035188635',
                                                                       'in_ucast_pkts': '664118991',
                                                                       'mac_address': '04:bc:9f:6f:07:4e',
                                                                       'name': '257468/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '33',
                                                                       'out_octets': '1229006334579',
                                                                       'out_ucast_pkts': '1152840181',
                                                                       'poller_hash': 'MTI5NzQ2RHBlZXJpbnRmMjU3NDY4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257468',
                                                                       'in_nucast_pkts': '33',
                                                                       'in_octets': '1229006334579',
                                                                       'in_ucast_pkts': '1152840181',
                                                                       'links': [       {       'in_nucast_pkts': '33',
                                                                                                'in_octets': '1229006334579',
                                                                                                'in_ucast_pkts': '1152840181',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '257035188635',
                                                                                                'out_ucast_pkts': '664118991',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '257035188635',
                                                                       'out_ucast_pkts': '664118991',
                                                                       'poller_hash': 'MTI5NzQ2RHBlZXJpbnRmcG9uIDI1NzQ2OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257468',
                                                                                               'links': [       {       'connected_time': '4377689',
                                                                                                                        'fiber_length': 22110,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzQ2RHBlZXJwb25wb24gMjU3NDY4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '7 DAVID ST (JUKES CALEB)',
                                         'poller_hash': 'MTI5NzQ2RA==',
                                         'serial_number': '129746D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:07:4e',
                                         'uptime': '4377689'},
                                 {       'counter_source': '257989/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '257989/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3337260957',
                                                                       'in_ucast_pkts': '10606459',
                                                                       'mac_address': 'b8:94:70:b9:d7:f5',
                                                                       'name': '257989/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '27998385162',
                                                                       'out_ucast_pkts': '39873219',
                                                                       'poller_hash': 'MTE5MzAxM3BlZXJpbnRmMjU3OTg5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '257989',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '27998385162',
                                                                       'in_ucast_pkts': '39873219',
                                                                       'links': [       {       'in_nucast_pkts': '7',
                                                                                                'in_octets': '27998385162',
                                                                                                'in_ucast_pkts': '39873219',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '3337260957',
                                                                                                'out_ucast_pkts': '10606459',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3337260957',
                                                                       'out_ucast_pkts': '10606459',
                                                                       'poller_hash': 'MTE5MzAxM3BlZXJpbnRmcG9uIDI1Nzk4OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '257989',
                                                                                               'links': [       {       'connected_time': '8535281',
                                                                                                                        'fiber_length': 22250,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MzAxM3BlZXJwb25wb24gMjU3OTg5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1416 HUTCHISON RD (RON C ESBAUGH '
                                                 'CONST LTD)',
                                         'poller_hash': 'MTE5MzAxMw==',
                                         'serial_number': '1193013',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:d7:f5',
                                         'uptime': '8535281'},
                                 {       'counter_source': '258273/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '258273/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45659829959',
                                                                       'in_ucast_pkts': '8882063430',
                                                                       'mac_address': 'b8:94:70:ce:5c:e4',
                                                                       'name': '258273/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '1387475173529',
                                                                       'out_ucast_pkts': '982262656',
                                                                       'poller_hash': 'MTFCRTFFRXBlZXJpbnRmMjU4MjczL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '258273',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '1387475173529',
                                                                       'in_ucast_pkts': '982262656',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '1387475173529',
                                                                                                'in_ucast_pkts': '982262656',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45659829959',
                                                                                                'out_ucast_pkts': '8882063430',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45659829959',
                                                                       'out_ucast_pkts': '8882063430',
                                                                       'poller_hash': 'MTFCRTFFRXBlZXJpbnRmcG9uIDI1ODI3Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '258273',
                                                                                               'links': [       {       'connected_time': '7833000',
                                                                                                                        'fiber_length': 14283,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTFCRTFFRXBlZXJwb25wb24gMjU4Mjcz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2312 PERTH LINE (SCHULTZ NICHOLAS)',
                                         'poller_hash': 'MTFCRTFFRQ==',
                                         'serial_number': '11BE1EE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:ce:5c:e4',
                                         'uptime': '7833000'},
                                 {       'counter_source': '259902/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '259902/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '59152336085',
                                                                       'in_ucast_pkts': '224825252',
                                                                       'mac_address': '04:bc:9f:5b:fd:79',
                                                                       'name': '259902/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '26',
                                                                       'out_octets': '802614207670',
                                                                       'out_ucast_pkts': '627376661',
                                                                       'poller_hash': 'MTI3NDYyRnBlZXJpbnRmMjU5OTAyL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '259902',
                                                                       'in_nucast_pkts': '26',
                                                                       'in_octets': '802614207670',
                                                                       'in_ucast_pkts': '627376661',
                                                                       'links': [       {       'in_nucast_pkts': '26',
                                                                                                'in_octets': '802614207670',
                                                                                                'in_ucast_pkts': '627376661',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '59152336085',
                                                                                                'out_ucast_pkts': '224825252',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '59152336085',
                                                                       'out_ucast_pkts': '224825252',
                                                                       'poller_hash': 'MTI3NDYyRnBlZXJpbnRmcG9uIDI1OTkwMg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '259902',
                                                                                               'links': [       {       'connected_time': '22198070',
                                                                                                                        'fiber_length': 19617,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDYyRnBlZXJwb25wb24gMjU5OTAy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4920 PERTH ROAD 119 (BRANDON ROB)',
                                         'poller_hash': 'MTI3NDYyRg==',
                                         'serial_number': '127462F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:fd:79',
                                         'uptime': '22198070'},
                                 {       'counter_source': '262467/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'ec:4f:82:2d:7d:dd'],
                                                                       'id': '262467/g1',
                                                                       'in_nucast_pkts': '431',
                                                                       'in_octets': '541895068977',
                                                                       'in_ucast_pkts': '2499228261',
                                                                       'mac_address': '04:bc:9f:5c:06:4c',
                                                                       'mtu': 2000,
                                                                       'name': '262467/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '25452890',
                                                                       'out_octets': '12284720149306',
                                                                       'out_ucast_pkts': '8747823592',
                                                                       'poller_hash': 'MTI3NDc0OXBlZXJpbnRmMjYyNDY3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '262467',
                                                                       'in_nucast_pkts': '25452890',
                                                                       'in_octets': '12284720149306',
                                                                       'in_ucast_pkts': '8747823592',
                                                                       'links': [       {       'in_nucast_pkts': '25452890',
                                                                                                'in_octets': '12284720149306',
                                                                                                'in_ucast_pkts': '8747823592',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '431',
                                                                                                'out_octets': '541895068977',
                                                                                                'out_ucast_pkts': '2499228261',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '431',
                                                                       'out_octets': '541895068977',
                                                                       'out_ucast_pkts': '2499228261',
                                                                       'poller_hash': 'MTI3NDc0OXBlZXJpbnRmcG9uIDI2MjQ2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262467',
                                                                                               'links': [       {       'connected_time': '21588514',
                                                                                                                        'fiber_length': 22741,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDc0OXBlZXJwb25wb24gMjYyNDY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '16 LAWRENCE ST (KOCH KIM)',
                                         'poller_hash': 'MTI3NDc0OQ==',
                                         'serial_number': '1274749',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:06:49',
                                         'uptime': '21588514'},
                                 {       'counter_source': '262477/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262477/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '306084419388',
                                                                       'in_ucast_pkts': '997127458',
                                                                       'mac_address': 'cc:be:59:ca:e5:0a',
                                                                       'name': '262477/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '66',
                                                                       'out_octets': '4570686855971',
                                                                       'out_ucast_pkts': '3455108804',
                                                                       'poller_hash': 'NDI4MkEwcGVlcmludGYyNjI0NzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '262477',
                                                                       'in_nucast_pkts': '66',
                                                                       'in_octets': '4570686855971',
                                                                       'in_ucast_pkts': '3455108804',
                                                                       'links': [       {       'in_nucast_pkts': '66',
                                                                                                'in_octets': '4570686855971',
                                                                                                'in_ucast_pkts': '3455108804',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '306084419388',
                                                                                                'out_ucast_pkts': '997127458',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '306084419388',
                                                                       'out_ucast_pkts': '997127458',
                                                                       'poller_hash': 'NDI4MkEwcGVlcmludGZwb24gMjYyNDc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262477',
                                                                                               'links': [       {       'connected_time': '5094976',
                                                                                                                        'fiber_length': 23363,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDI4MkEwcGVlcnBvbnBvbiAyNjI0Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '104 FERRIS DR (LOUTER SANDY)',
                                         'poller_hash': 'NDI4MkEw',
                                         'serial_number': '4282A0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'cc:be:59:ca:e5:0a',
                                         'uptime': '5094976'},
                                 {       'counter_source': '262537/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262537/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '301914527221',
                                                                       'in_ucast_pkts': '8914642848',
                                                                       'mac_address': '04:bc:9f:5b:f8:d1',
                                                                       'name': '262537/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '358623513396',
                                                                       'out_ucast_pkts': '410008016',
                                                                       'poller_hash': 'MTI3NDU5QXBlZXJpbnRmMjYyNTM3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262537',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '358623513396',
                                                                       'in_ucast_pkts': '410008016',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '358623513396',
                                                                                                'in_ucast_pkts': '410008016',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '301914527221',
                                                                                                'out_ucast_pkts': '8914642848',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '301914527221',
                                                                       'out_ucast_pkts': '8914642848',
                                                                       'poller_hash': 'MTI3NDU5QXBlZXJpbnRmcG9uIDI2MjUzNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262537',
                                                                                               'links': [       {       'connected_time': '1367426',
                                                                                                                        'fiber_length': 22285,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDU5QXBlZXJwb25wb24gMjYyNTM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1416 HUTCHISON RD (ROHR ELECTRIC)',
                                         'poller_hash': 'MTI3NDU5QQ==',
                                         'serial_number': '127459A',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:f8:d1',
                                         'uptime': '1367426'},
                                 {       'counter_source': '262618/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7e:ee'],
                                                                       'id': '262618/g1',
                                                                       'in_nucast_pkts': '76059',
                                                                       'in_octets': '3264019972',
                                                                       'in_ucast_pkts': '8575302',
                                                                       'mac_address': '04:bc:9f:5c:0a:e4',
                                                                       'mtu': 2000,
                                                                       'name': '262618/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4181556',
                                                                       'out_octets': '24322106207',
                                                                       'out_ucast_pkts': '22432106',
                                                                       'poller_hash': 'MTI3NDdEQ3BlZXJpbnRmMjYyNjE4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '262618',
                                                                       'in_nucast_pkts': '4181556',
                                                                       'in_octets': '24322106207',
                                                                       'in_ucast_pkts': '22432106',
                                                                       'links': [       {       'in_nucast_pkts': '4181556',
                                                                                                'in_octets': '24322106207',
                                                                                                'in_ucast_pkts': '22432106',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '76059',
                                                                                                'out_octets': '3264019972',
                                                                                                'out_ucast_pkts': '8575302',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '76059',
                                                                       'out_octets': '3264019972',
                                                                       'out_ucast_pkts': '8575302',
                                                                       'poller_hash': 'MTI3NDdEQ3BlZXJpbnRmcG9uIDI2MjYxOA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262618',
                                                                                               'links': [       {       'connected_time': '3549183',
                                                                                                                        'fiber_length': 22282,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdEQ3BlZXJwb25wb24gMjYyNjE4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1416 HUTCHISON RD (HURREN RICHARD)',
                                         'poller_hash': 'MTI3NDdEQw==',
                                         'serial_number': '12747DC',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0a:e1',
                                         'uptime': '3549183'},
                                 {       'counter_source': '262688/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '262688/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '114311097246',
                                                                       'in_ucast_pkts': '47532727073',
                                                                       'mac_address': '04:bc:9f:5c:0d:f9',
                                                                       'name': '262688/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '18',
                                                                       'out_octets': '890434808114',
                                                                       'out_ucast_pkts': '770707866',
                                                                       'poller_hash': 'MTI3NDgzRnBlZXJpbnRmMjYyNjg4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '262688',
                                                                       'in_nucast_pkts': '18',
                                                                       'in_octets': '890434808114',
                                                                       'in_ucast_pkts': '770707866',
                                                                       'links': [       {       'in_nucast_pkts': '18',
                                                                                                'in_octets': '890434808114',
                                                                                                'in_ucast_pkts': '770707866',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '114311097246',
                                                                                                'out_ucast_pkts': '47532727073',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '114311097246',
                                                                       'out_ucast_pkts': '47532727073',
                                                                       'poller_hash': 'MTI3NDgzRnBlZXJpbnRmcG9uIDI2MjY4OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262688',
                                                                                               'links': [       {       'connected_time': '8535125',
                                                                                                                        'fiber_length': 22828,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDgzRnBlZXJwb25wb24gMjYyNjg4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1129 HENRY ST (INSPIRING MINDS EARLY '
                                                 'LEARNING)',
                                         'poller_hash': 'MTI3NDgzRg==',
                                         'serial_number': '127483F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0d:f9',
                                         'uptime': '8535125'},
                                 {       'counter_source': '262727/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8a:f6'],
                                                                       'id': '262727/g1',
                                                                       'in_nucast_pkts': '391855',
                                                                       'in_octets': '124517618963',
                                                                       'in_ucast_pkts': '608743215',
                                                                       'mac_address': '04:bc:9f:5b:fc:1c',
                                                                       'mtu': 2000,
                                                                       'name': '262727/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '24776325',
                                                                       'out_octets': '5021722419935',
                                                                       'out_ucast_pkts': '3484449320',
                                                                       'poller_hash': 'MTI3NDYwM3BlZXJpbnRmMjYyNzI3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '262727',
                                                                       'in_nucast_pkts': '24776325',
                                                                       'in_octets': '5021722419935',
                                                                       'in_ucast_pkts': '3484449320',
                                                                       'links': [       {       'in_nucast_pkts': '24776325',
                                                                                                'in_octets': '5021722419935',
                                                                                                'in_ucast_pkts': '3484449320',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '391855',
                                                                                                'out_octets': '124517618963',
                                                                                                'out_ucast_pkts': '608743215',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '391855',
                                                                       'out_octets': '124517618963',
                                                                       'out_ucast_pkts': '608743215',
                                                                       'poller_hash': 'MTI3NDYwM3BlZXJpbnRmcG9uIDI2MjcyNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262727',
                                                                                               'links': [       {       'connected_time': '20986130',
                                                                                                                        'fiber_length': 23355,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDYwM3BlZXJwb25wb24gMjYyNzI3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1189 QUEEN'S BUSH RD (WAGNER CARRIE)",
                                         'poller_hash': 'MTI3NDYwMw==',
                                         'serial_number': '1274603',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:fc:19',
                                         'uptime': '20986130'},
                                 {       'counter_source': '262995/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:89:ac'],
                                                                       'id': '262995/g1',
                                                                       'in_nucast_pkts': '165888',
                                                                       'in_octets': '667678572271',
                                                                       'in_ucast_pkts': '2032871673',
                                                                       'mac_address': '04:bc:9f:5b:ff:44',
                                                                       'mtu': 2000,
                                                                       'name': '262995/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920132',
                                                                       'out_octets': '7078236192010',
                                                                       'out_ucast_pkts': '5352329993',
                                                                       'poller_hash': 'MTI3NDY2OHBlZXJpbnRmMjYyOTk1L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '262995',
                                                                       'in_nucast_pkts': '9920132',
                                                                       'in_octets': '7078236192010',
                                                                       'in_ucast_pkts': '5352329993',
                                                                       'links': [       {       'in_nucast_pkts': '9920132',
                                                                                                'in_octets': '7078236192010',
                                                                                                'in_ucast_pkts': '5352329993',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '165888',
                                                                                                'out_octets': '667678572271',
                                                                                                'out_ucast_pkts': '2032871673',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '165888',
                                                                       'out_octets': '667678572271',
                                                                       'out_ucast_pkts': '2032871673',
                                                                       'poller_hash': 'MTI3NDY2OHBlZXJpbnRmcG9uIDI2Mjk5NQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '262995',
                                                                                               'links': [       {       'connected_time': '8535225',
                                                                                                                        'fiber_length': 22455,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDY2OHBlZXJwb25wb24gMjYyOTk1',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1221 QUEEN'S BUSH RD (DICASA "
                                                 'MICHAEL)',
                                         'poller_hash': 'MTI3NDY2OA==',
                                         'serial_number': '1274668',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:ff:41',
                                         'uptime': '8535225'},
                                 {       'counter_source': '264930/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '264930/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '307',
                                                                       'in_octets': '1104261220985',
                                                                       'in_ucast_pkts': '0',
                                                                       'mac_address': '04:bc:9f:5c:00:89',
                                                                       'name': '264930/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '564',
                                                                       'out_octets': '8517556344612',
                                                                       'out_ucast_pkts': '6695277736',
                                                                       'poller_hash': 'MTI3NDY5MXBlZXJpbnRmMjY0OTMwL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264930',
                                                                       'in_nucast_pkts': '564',
                                                                       'in_octets': '8517556344612',
                                                                       'in_ucast_pkts': '6695277736',
                                                                       'links': [       {       'in_nucast_pkts': '564',
                                                                                                'in_octets': '8517556344612',
                                                                                                'in_ucast_pkts': '6695277736',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '307',
                                                                                                'out_octets': '1104261220985',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '307',
                                                                       'out_octets': '1104261220985',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDY5MXBlZXJpbnRmcG9uIDI2NDkzMA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264930',
                                                                                               'links': [       {       'connected_time': '8535190',
                                                                                                                        'fiber_length': 23463,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDY5MXBlZXJwb25wb24gMjY0OTMw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '40 ZINKANN CR (LANG TODD)',
                                         'poller_hash': 'MTI3NDY5MQ==',
                                         'serial_number': '1274691',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:00:89',
                                         'uptime': '8535190'},
                                 {       'counter_source': '264963/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '264963/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '9374293103',
                                                                       'in_ucast_pkts': '30627598',
                                                                       'mac_address': '04:bc:9f:5c:0c:81',
                                                                       'name': '264963/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '183004992559',
                                                                       'out_ucast_pkts': '153021887',
                                                                       'poller_hash': 'MTI3NDgxMHBlZXJpbnRmMjY0OTYzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '264963',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '183004992559',
                                                                       'in_ucast_pkts': '153021887',
                                                                       'links': [       {       'in_nucast_pkts': '7',
                                                                                                'in_octets': '183004992559',
                                                                                                'in_ucast_pkts': '153021887',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '9374293103',
                                                                                                'out_ucast_pkts': '30627598',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '9374293103',
                                                                       'out_ucast_pkts': '30627598',
                                                                       'poller_hash': 'MTI3NDgxMHBlZXJpbnRmcG9uIDI2NDk2Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '264963',
                                                                                               'links': [       {       'connected_time': '1370851',
                                                                                                                        'fiber_length': 22272,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDgxMHBlZXJwb25wb24gMjY0OTYz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1416 HUTCHISON RD (CENTRE LINE '
                                                 'CONSTRUCTION LTD)',
                                         'poller_hash': 'MTI3NDgxMA==',
                                         'serial_number': '1274810',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0c:81',
                                         'uptime': '1370851'},
                                 {       'counter_source': '265737/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7f:1a'],
                                                                       'id': '265737/g1',
                                                                       'in_nucast_pkts': '149978',
                                                                       'in_octets': '31348798323',
                                                                       'in_ucast_pkts': '258618480',
                                                                       'mac_address': '04:bc:9f:5c:01:34',
                                                                       'mtu': 2000,
                                                                       'name': '265737/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9925949',
                                                                       'out_octets': '3097849369801',
                                                                       'out_ucast_pkts': '2099248066',
                                                                       'poller_hash': 'MTI3NDZBNnBlZXJpbnRmMjY1NzM3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '265737',
                                                                       'in_nucast_pkts': '9925949',
                                                                       'in_octets': '3097849369801',
                                                                       'in_ucast_pkts': '2099248066',
                                                                       'links': [       {       'in_nucast_pkts': '9925949',
                                                                                                'in_octets': '3097849369801',
                                                                                                'in_ucast_pkts': '2099248066',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149978',
                                                                                                'out_octets': '31348798323',
                                                                                                'out_ucast_pkts': '258618480',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149978',
                                                                       'out_octets': '31348798323',
                                                                       'out_ucast_pkts': '258618480',
                                                                       'poller_hash': 'MTI3NDZBNnBlZXJpbnRmcG9uIDI2NTczNw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '265737',
                                                                                               'links': [       {       'connected_time': '8535196',
                                                                                                                        'fiber_length': 24928,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDZBNnBlZXJwb25wb24gMjY1NzM3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '107 GERBER MEADOWS DR (JANTZI '
                                                 'CHARLOTTE)',
                                         'poller_hash': 'MTI3NDZBNg==',
                                         'serial_number': '12746A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:01:31',
                                         'uptime': '8535196'},
                                 {       'counter_source': '266134/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '5c:0f:fb:33:c1:92'],
                                                                       'id': '266134/g1',
                                                                       'in_nucast_pkts': '473152',
                                                                       'in_octets': '53555235816',
                                                                       'in_ucast_pkts': '723727997',
                                                                       'mac_address': '04:bc:9f:5c:00:bc',
                                                                       'mtu': 2000,
                                                                       'name': '266134/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '97205163',
                                                                       'out_octets': '4116142567471',
                                                                       'out_ucast_pkts': '2902990414',
                                                                       'poller_hash': 'MTI3NDY5N3BlZXJpbnRmMjY2MTM0L2cx',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '266134',
                                                                       'in_nucast_pkts': '97205163',
                                                                       'in_octets': '4116142567471',
                                                                       'in_ucast_pkts': '2902990414',
                                                                       'links': [       {       'in_nucast_pkts': '97205163',
                                                                                                'in_octets': '4116142567471',
                                                                                                'in_ucast_pkts': '2902990414',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '473152',
                                                                                                'out_octets': '53555235816',
                                                                                                'out_ucast_pkts': '723727997',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '473152',
                                                                       'out_octets': '53555235816',
                                                                       'out_ucast_pkts': '723727997',
                                                                       'poller_hash': 'MTI3NDY5N3BlZXJpbnRmcG9uIDI2NjEzNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266134',
                                                                                               'links': [       {       'connected_time': '8535154',
                                                                                                                        'fiber_length': 23271,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDY5N3BlZXJwb25wb24gMjY2MTM0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1060 MOLESWORTH ST (STROH TOM)',
                                         'poller_hash': 'MTI3NDY5Nw==',
                                         'serial_number': '1274697',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:00:b9',
                                         'uptime': '8535154'},
                                 {       'counter_source': '266168/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:7c:86'],
                                                                       'id': '266168/g1',
                                                                       'in_nucast_pkts': '302122',
                                                                       'in_octets': '143183121206',
                                                                       'in_ucast_pkts': '881608705',
                                                                       'mac_address': '04:bc:9f:5c:0a:ac',
                                                                       'mtu': 2000,
                                                                       'name': '266168/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20571531',
                                                                       'out_octets': '4065877091367',
                                                                       'out_ucast_pkts': '2782378938',
                                                                       'poller_hash': 'MTI3NDdENXBlZXJpbnRmMjY2MTY4L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '266168',
                                                                       'in_nucast_pkts': '20571531',
                                                                       'in_octets': '4065877091367',
                                                                       'in_ucast_pkts': '2782378938',
                                                                       'links': [       {       'in_nucast_pkts': '20571531',
                                                                                                'in_octets': '4065877091367',
                                                                                                'in_ucast_pkts': '2782378938',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '302122',
                                                                                                'out_octets': '143183121206',
                                                                                                'out_ucast_pkts': '881608705',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzU='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '302122',
                                                                       'out_octets': '143183121206',
                                                                       'out_ucast_pkts': '881608705',
                                                                       'poller_hash': 'MTI3NDdENXBlZXJpbnRmcG9uIDI2NjE2OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '266168',
                                                                                               'links': [       {       'connected_time': '17373731',
                                                                                                                        'fiber_length': 25300,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzU=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdENXBlZXJwb25wb24gMjY2MTY4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '29 WELWOOD AV (SCHLUETER BETH)',
                                         'poller_hash': 'MTI3NDdENQ==',
                                         'serial_number': '12747D5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0a:a9',
                                         'uptime': '17373731'},
                                 {       'counter_source': '269869/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '269869/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1486956069316',
                                                                       'in_ucast_pkts': '1676654054',
                                                                       'mac_address': '04:bc:9f:5a:36:59',
                                                                       'name': '269869/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '6206946734360',
                                                                       'out_ucast_pkts': '4461339522',
                                                                       'poller_hash': 'MTI3MEQ0QnBlZXJpbnRmMjY5ODY5L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '269869',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '6206946734360',
                                                                       'in_ucast_pkts': '4461339522',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '6206946734360',
                                                                                                'in_ucast_pkts': '4461339522',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '1486956069316',
                                                                                                'out_ucast_pkts': '1676654054',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1486956069316',
                                                                       'out_ucast_pkts': '1676654054',
                                                                       'poller_hash': 'MTI3MEQ0QnBlZXJpbnRmcG9uIDI2OTg2OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '269869',
                                                                                               'links': [       {       'connected_time': '8535287',
                                                                                                                        'fiber_length': 24753,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ0QnBlZXJwb25wb24gMjY5ODY5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '14 FIRELLA PL (LICHTY TODD)',
                                         'poller_hash': 'MTI3MEQ0Qg==',
                                         'serial_number': '1270D4B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:36:59',
                                         'uptime': '8535287'},
                                 {       'counter_source': '272089/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:09:3c:d8'],
                                                                       'id': '272089/g1',
                                                                       'in_nucast_pkts': '93567',
                                                                       'in_octets': '29631380238',
                                                                       'in_ucast_pkts': '104188615',
                                                                       'mac_address': 'b8:94:70:b9:df:80',
                                                                       'mtu': 2000,
                                                                       'name': '272089/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5434622',
                                                                       'out_octets': '1212959020481',
                                                                       'out_ucast_pkts': '894717682',
                                                                       'poller_hash': 'MTE5MzEwNHBlZXJpbnRmMjcyMDg5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272089',
                                                                       'in_nucast_pkts': '5434622',
                                                                       'in_octets': '1212959020481',
                                                                       'in_ucast_pkts': '894717682',
                                                                       'links': [       {       'in_nucast_pkts': '5434622',
                                                                                                'in_octets': '1212959020481',
                                                                                                'in_ucast_pkts': '894717682',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '93567',
                                                                                                'out_octets': '29631380238',
                                                                                                'out_ucast_pkts': '104188615',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '93567',
                                                                       'out_octets': '29631380238',
                                                                       'out_ucast_pkts': '104188615',
                                                                       'poller_hash': 'MTE5MzEwNHBlZXJpbnRmcG9uIDI3MjA4OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272089',
                                                                                               'links': [       {       'connected_time': '4637845',
                                                                                                                        'fiber_length': 23042,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTE5MzEwNHBlZXJwb25wb24gMjcyMDg5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '48 REINER CR (ALBRECHT ALLISON)',
                                         'poller_hash': 'MTE5MzEwNA==',
                                         'serial_number': '1193104',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': 'b8:94:70:b9:df:7d',
                                         'uptime': '4637845'},
                                 {       'counter_source': '272199/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '84:d3:43:6d:ea:da'],
                                                                       'id': '272199/g1',
                                                                       'in_nucast_pkts': '149039',
                                                                       'in_octets': '101651505210',
                                                                       'in_ucast_pkts': '807424406',
                                                                       'mac_address': '04:bc:9f:5c:11:6c',
                                                                       'mtu': 2000,
                                                                       'name': '272199/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9878481',
                                                                       'out_octets': '2920412619836',
                                                                       'out_ucast_pkts': '2006293755',
                                                                       'poller_hash': 'MTI3NDhBRHBlZXJpbnRmMjcyMTk5L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272199',
                                                                       'in_nucast_pkts': '9878481',
                                                                       'in_octets': '2920412619836',
                                                                       'in_ucast_pkts': '2006293755',
                                                                       'links': [       {       'in_nucast_pkts': '9878481',
                                                                                                'in_octets': '2920412619836',
                                                                                                'in_ucast_pkts': '2006293755',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149039',
                                                                                                'out_octets': '101651505210',
                                                                                                'out_ucast_pkts': '807424406',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149039',
                                                                       'out_octets': '101651505210',
                                                                       'out_ucast_pkts': '807424406',
                                                                       'poller_hash': 'MTI3NDhBRHBlZXJpbnRmcG9uIDI3MjE5OQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272199',
                                                                                               'links': [       {       'connected_time': '8535166',
                                                                                                                        'fiber_length': 23177,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDhBRHBlZXJwb25wb24gMjcyMTk5',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1086 HENRY ST (KOCH PAUL R)',
                                         'poller_hash': 'MTI3NDhBRA==',
                                         'serial_number': '12748AD',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:11:69',
                                         'uptime': '8535166'},
                                 {       'counter_source': '272324/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'd0:76:8f:02:f1:d8'],
                                                                       'id': '272324/g1',
                                                                       'in_nucast_pkts': '155504',
                                                                       'in_octets': '592813435683',
                                                                       'in_ucast_pkts': '1316471687',
                                                                       'mac_address': '04:bc:9f:5c:0a:d4',
                                                                       'mtu': 2000,
                                                                       'name': '272324/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9511572',
                                                                       'out_octets': '5291430138304',
                                                                       'out_ucast_pkts': '4062124451',
                                                                       'poller_hash': 'MTI3NDdEQXBlZXJpbnRmMjcyMzI0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '272324',
                                                                       'in_nucast_pkts': '9511572',
                                                                       'in_octets': '5291430138304',
                                                                       'in_ucast_pkts': '4062124451',
                                                                       'links': [       {       'in_nucast_pkts': '9511572',
                                                                                                'in_octets': '5291430138304',
                                                                                                'in_ucast_pkts': '4062124451',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '155504',
                                                                                                'out_octets': '592813435683',
                                                                                                'out_ucast_pkts': '1316471687',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '155504',
                                                                       'out_octets': '592813435683',
                                                                       'out_ucast_pkts': '1316471687',
                                                                       'poller_hash': 'MTI3NDdEQXBlZXJpbnRmcG9uIDI3MjMyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '272324',
                                                                                               'links': [       {       'connected_time': '8182288',
                                                                                                                        'fiber_length': 22444,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdEQXBlZXJwb25wb24gMjcyMzI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '84 POND VIEW DR (BOLLER PATRICK)',
                                         'poller_hash': 'MTI3NDdEQQ==',
                                         'serial_number': '12747DA',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0a:d1',
                                         'uptime': '8182288'},
                                 {       'counter_source': '274021/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274021/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '36005028032',
                                                                       'in_ucast_pkts': '74775443',
                                                                       'mac_address': '04:bc:9f:6f:0a:66',
                                                                       'name': '274021/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '182006938250',
                                                                       'out_ucast_pkts': '171515576',
                                                                       'poller_hash': 'MTI5NzREMHBlZXJpbnRmMjc0MDIxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274021',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '182006938250',
                                                                       'in_ucast_pkts': '171515576',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '182006938250',
                                                                                                'in_ucast_pkts': '171515576',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '36005028032',
                                                                                                'out_ucast_pkts': '74775443',
                                                                                                'poller_hash': 'NDcxOTA0MDA0MjgwZ3BvbiAyLzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '36005028032',
                                                                       'out_ucast_pkts': '74775443',
                                                                       'poller_hash': 'MTI5NzREMHBlZXJpbnRmcG9uIDI3NDAyMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274021',
                                                                                               'links': [       {       'connected_time': '4217748',
                                                                                                                        'fiber_length': 22260,
                                                                                                                        'poller_hash': 'NDcxOTA0MDA0MjgwcG9uZ3BvbiAyLzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NzREMHBlZXJwb25wb24gMjc0MDIx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1416 HUTCHISON RD (DOERING ELECTRIC '
                                                 'INC)',
                                         'poller_hash': 'MTI5NzREMA==',
                                         'serial_number': '12974D0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6f:0a:66',
                                         'uptime': '4217748'},
                                 {       'counter_source': '274071/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '00:0d:b9:5d:9a:05'],
                                                                       'id': '274071/g1',
                                                                       'in_nucast_pkts': '3876',
                                                                       'in_octets': '158377159985',
                                                                       'in_ucast_pkts': '396630105',
                                                                       'mac_address': '04:bc:9f:5c:05:dc',
                                                                       'mtu': 2000,
                                                                       'name': '274071/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '4829510',
                                                                       'out_octets': '434188533211',
                                                                       'out_ucast_pkts': '508224534',
                                                                       'poller_hash': 'MTI3NDczQnBlZXJpbnRmMjc0MDcxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274071',
                                                                       'in_nucast_pkts': '4829510',
                                                                       'in_octets': '434188533211',
                                                                       'in_ucast_pkts': '508224534',
                                                                       'links': [       {       'in_nucast_pkts': '4829510',
                                                                                                'in_octets': '434188533211',
                                                                                                'in_ucast_pkts': '508224534',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '3876',
                                                                                                'out_octets': '158377159985',
                                                                                                'out_ucast_pkts': '396630105',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3876',
                                                                       'out_octets': '158377159985',
                                                                       'out_ucast_pkts': '396630105',
                                                                       'poller_hash': 'MTI3NDczQnBlZXJpbnRmcG9uIDI3NDA3MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274071',
                                                                                               'links': [       {       'connected_time': '4142146',
                                                                                                                        'fiber_length': 22542,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDczQnBlZXJwb25wb24gMjc0MDcx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1214 QUEEN'S BUSH RD (LEIS FEED & "
                                                 'SUPPLY LTD)',
                                         'poller_hash': 'MTI3NDczQg==',
                                         'serial_number': '127473B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:05:d9',
                                         'uptime': '4142146'},
                                 {       'counter_source': '274514/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274514/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '48372488150',
                                                                       'in_ucast_pkts': '48250192',
                                                                       'mac_address': '04:bc:9f:5c:0c:b1',
                                                                       'name': '274514/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '72719485233',
                                                                       'out_ucast_pkts': '71471276',
                                                                       'poller_hash': 'MTI3NDgxNnBlZXJpbnRmMjc0NTE0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '274514',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '72719485233',
                                                                       'in_ucast_pkts': '71471276',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '72719485233',
                                                                                                'in_ucast_pkts': '71471276',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '48372488150',
                                                                                                'out_ucast_pkts': '48250192',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '48372488150',
                                                                       'out_ucast_pkts': '48250192',
                                                                       'poller_hash': 'MTI3NDgxNnBlZXJpbnRmcG9uIDI3NDUxNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274514',
                                                                                               'links': [       {       'connected_time': '8535310',
                                                                                                                        'fiber_length': 17474,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDgxNnBlZXJwb25wb24gMjc0NTE0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2060 PERTH LINE (JANTZI GABRIEL)',
                                         'poller_hash': 'MTI3NDgxNg==',
                                         'serial_number': '1274816',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:0c:b1',
                                         'uptime': '8535310'},
                                 {       'counter_source': '274524/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:d1:8c:1f'],
                                                                       'id': '274524/g1',
                                                                       'in_nucast_pkts': '149664',
                                                                       'in_octets': '130254975536',
                                                                       'in_ucast_pkts': '386207677',
                                                                       'mac_address': '04:bc:9f:5c:09:9c',
                                                                       'mtu': 2000,
                                                                       'name': '274524/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9920035',
                                                                       'out_octets': '1682372013687',
                                                                       'out_ucast_pkts': '1305614801',
                                                                       'poller_hash': 'MTI3NDdCM3BlZXJpbnRmMjc0NTI0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '274524',
                                                                       'in_nucast_pkts': '9920035',
                                                                       'in_octets': '1682372013687',
                                                                       'in_ucast_pkts': '1305614801',
                                                                       'links': [       {       'in_nucast_pkts': '9920035',
                                                                                                'in_octets': '1682372013687',
                                                                                                'in_ucast_pkts': '1305614801',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '149664',
                                                                                                'out_octets': '130254975536',
                                                                                                'out_ucast_pkts': '386207677',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '149664',
                                                                       'out_octets': '130254975536',
                                                                       'out_ucast_pkts': '386207677',
                                                                       'poller_hash': 'MTI3NDdCM3BlZXJpbnRmcG9uIDI3NDUyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274524',
                                                                                               'links': [       {       'connected_time': '8535206',
                                                                                                                        'fiber_length': 23783,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDdCM3BlZXJwb25wb24gMjc0NTI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '42 GALENA ST WELLESLEY (ROMANOVA '
                                                 'ELENA)',
                                         'poller_hash': 'MTI3NDdCMw==',
                                         'serial_number': '12747B3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:09:99',
                                         'uptime': '8535206'},
                                 {       'counter_source': '274777/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '274777/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '113505885007',
                                                                       'in_ucast_pkts': '64753346662',
                                                                       'mac_address': '60:db:98:4e:99:f4',
                                                                       'name': '274777/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '666129095175',
                                                                       'out_ucast_pkts': '675605440',
                                                                       'poller_hash': 'QTlDOEZFcGVlcmludGYyNzQ3NzcvRzE='},
                                                               {       'admin_up': True,
                                                                       'id': '274777',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '666129095175',
                                                                       'in_ucast_pkts': '675605440',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '666129095175',
                                                                                                'in_ucast_pkts': '675605440',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '113505885007',
                                                                                                'out_ucast_pkts': '64753346662',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '113505885007',
                                                                       'out_ucast_pkts': '64753346662',
                                                                       'poller_hash': 'QTlDOEZFcGVlcmludGZwb24gMjc0Nzc3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '274777',
                                                                                               'links': [       {       'connected_time': '8535295',
                                                                                                                        'fiber_length': 21171,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'QTlDOEZFcGVlcnBvbnBvbiAyNzQ3Nzc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1214 HUTCHISON RD (JANTZI CONRAD)',
                                         'poller_hash': 'QTlDOEZF',
                                         'serial_number': 'A9C8FE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '60:db:98:4e:99:f4',
                                         'uptime': '8535295'},
                                 {       'counter_source': '276976/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '276976/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '31362557418',
                                                                       'in_ucast_pkts': '263983856',
                                                                       'mac_address': '04:bc:9f:5a:3e:91',
                                                                       'name': '276976/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1126834809077',
                                                                       'out_ucast_pkts': '796015076',
                                                                       'poller_hash': 'MTI3MEU1MnBlZXJpbnRmMjc2OTc2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '276976',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1126834809077',
                                                                       'in_ucast_pkts': '796015076',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1126834809077',
                                                                                                'in_ucast_pkts': '796015076',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '31362557418',
                                                                                                'out_ucast_pkts': '263983856',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '31362557418',
                                                                       'out_ucast_pkts': '263983856',
                                                                       'poller_hash': 'MTI3MEU1MnBlZXJpbnRmcG9uIDI3Njk3Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276976',
                                                                                               'links': [       {       'connected_time': '10362629',
                                                                                                                        'fiber_length': 19890,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEU1MnBlZXJwb25wb24gMjc2OTc2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2859 LINE 45 (FAULHAFER ROBERT & '
                                                 'DOROTHEA)',
                                         'poller_hash': 'MTI3MEU1Mg==',
                                         'serial_number': '1270E52',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:3e:91',
                                         'uptime': '10362629'},
                                 {       'counter_source': '276986/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '276986/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '122945701136',
                                                                       'in_ucast_pkts': '881008443',
                                                                       'mac_address': '04:bc:9f:6e:56:5e',
                                                                       'name': '276986/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3024469975679',
                                                                       'out_ucast_pkts': '2134145664',
                                                                       'poller_hash': 'MTI5NUU0RnBlZXJpbnRmMjc2OTg2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '276986',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3024469975679',
                                                                       'in_ucast_pkts': '2134145664',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3024469975679',
                                                                                                'in_ucast_pkts': '2134145664',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '122945701136',
                                                                                                'out_ucast_pkts': '881008443',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '122945701136',
                                                                       'out_ucast_pkts': '881008443',
                                                                       'poller_hash': 'MTI5NUU0RnBlZXJpbnRmcG9uIDI3Njk4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '276986',
                                                                                               'links': [       {       'connected_time': '611456',
                                                                                                                        'fiber_length': 20042,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU0RnBlZXJwb25wb24gMjc2OTg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4940 LINHAVEN DR GADSHILL (RIEHL TIM '
                                                 '& CASEY)',
                                         'poller_hash': 'MTI5NUU0Rg==',
                                         'serial_number': '1295E4F',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:56:5e',
                                         'uptime': '611456'},
                                 {       'counter_source': '277834/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '277834/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '45577526008',
                                                                       'in_ucast_pkts': '306572975',
                                                                       'mac_address': '04:bc:9f:6e:63:8e',
                                                                       'name': '277834/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1098654135436',
                                                                       'out_ucast_pkts': '791088230',
                                                                       'poller_hash': 'MTI5NUZGNXBlZXJpbnRmMjc3ODM0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '277834',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1098654135436',
                                                                       'in_ucast_pkts': '791088230',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1098654135436',
                                                                                                'in_ucast_pkts': '791088230',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '45577526008',
                                                                                                'out_ucast_pkts': '306572975',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '45577526008',
                                                                       'out_ucast_pkts': '306572975',
                                                                       'poller_hash': 'MTI5NUZGNXBlZXJpbnRmcG9uIDI3NzgzNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '277834',
                                                                                               'links': [       {       'connected_time': '611464',
                                                                                                                        'fiber_length': 19952,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUZGNXBlZXJwb25wb24gMjc3ODM0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4898 LINHAVEN DR (SMITS JAMES)',
                                         'poller_hash': 'MTI5NUZGNQ==',
                                         'serial_number': '1295FF5',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:63:8e',
                                         'uptime': '611464'},
                                 {       'counter_source': '278231/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:75:22'],
                                                                       'id': '278231/g1',
                                                                       'in_nucast_pkts': '12324',
                                                                       'in_octets': '2556046707',
                                                                       'in_ucast_pkts': '10243465',
                                                                       'mac_address': '04:bc:9f:5c:10:cc',
                                                                       'mtu': 2000,
                                                                       'name': '278231/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '715543',
                                                                       'out_octets': '45088981865',
                                                                       'out_ucast_pkts': '34886336',
                                                                       'poller_hash': 'MTI3NDg5OXBlZXJpbnRmMjc4MjMxL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '278231',
                                                                       'in_nucast_pkts': '715543',
                                                                       'in_octets': '45088981865',
                                                                       'in_ucast_pkts': '34886336',
                                                                       'links': [       {       'in_nucast_pkts': '715543',
                                                                                                'in_octets': '45088981865',
                                                                                                'in_ucast_pkts': '34886336',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12324',
                                                                                                'out_octets': '2556046707',
                                                                                                'out_ucast_pkts': '10243465',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12324',
                                                                       'out_octets': '2556046707',
                                                                       'out_ucast_pkts': '10243465',
                                                                       'poller_hash': 'MTI3NDg5OXBlZXJpbnRmcG9uIDI3ODIzMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278231',
                                                                                               'links': [       {       'connected_time': '611450',
                                                                                                                        'fiber_length': 19828,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDg5OXBlZXJwb25wb24gMjc4MjMx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4903 PERTH ROAD 119 (ALLIN JODI)',
                                         'poller_hash': 'MTI3NDg5OQ==',
                                         'serial_number': '1274899',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:10:c9',
                                         'uptime': '611450'},
                                 {       'counter_source': '278358/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '278358/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '43300221294',
                                                                       'in_ucast_pkts': '372084931',
                                                                       'mac_address': '04:bc:9f:6e:63:56',
                                                                       'name': '278358/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '2150213528923',
                                                                       'out_ucast_pkts': '1449025911',
                                                                       'poller_hash': 'MTI5NUZFRXBlZXJpbnRmMjc4MzU4L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '278358',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '2150213528923',
                                                                       'in_ucast_pkts': '1449025911',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '2150213528923',
                                                                                                'in_ucast_pkts': '1449025911',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '43300221294',
                                                                                                'out_ucast_pkts': '372084931',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzQ='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '43300221294',
                                                                       'out_ucast_pkts': '372084931',
                                                                       'poller_hash': 'MTI5NUZFRXBlZXJpbnRmcG9uIDI3ODM1OA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278358',
                                                                                               'links': [       {       'connected_time': '8385735',
                                                                                                                        'fiber_length': 17518,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzQ=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -23.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUZFRXBlZXJwb25wb24gMjc4MzU4',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '2076 ALBERT ST (DETZLER A & SMITH C)',
                                         'poller_hash': 'MTI5NUZFRQ==',
                                         'serial_number': '1295FEE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:63:56',
                                         'uptime': '8385735'},
                                 {       'counter_source': '278467/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'cc:be:59:32:95:16'],
                                                                       'id': '278467/g1',
                                                                       'in_nucast_pkts': '20',
                                                                       'in_octets': '27190528470',
                                                                       'in_ucast_pkts': '287659063',
                                                                       'mac_address': '04:bc:9f:5a:37:6c',
                                                                       'mtu': 2000,
                                                                       'name': '278467/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '9635078',
                                                                       'out_octets': '1764236051363',
                                                                       'out_ucast_pkts': '1183986942',
                                                                       'poller_hash': 'MTI3MEQ2RHBlZXJpbnRmMjc4NDY3L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '278467',
                                                                       'in_nucast_pkts': '9635078',
                                                                       'in_octets': '1764236051363',
                                                                       'in_ucast_pkts': '1183986942',
                                                                       'links': [       {       'in_nucast_pkts': '9635078',
                                                                                                'in_octets': '1764236051363',
                                                                                                'in_ucast_pkts': '1183986942',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '20',
                                                                                                'out_octets': '27190528470',
                                                                                                'out_ucast_pkts': '287659063',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '20',
                                                                       'out_octets': '27190528470',
                                                                       'out_ucast_pkts': '287659063',
                                                                       'poller_hash': 'MTI3MEQ2RHBlZXJpbnRmcG9uIDI3ODQ2Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '278467',
                                                                                               'links': [       {       'connected_time': '8291191',
                                                                                                                        'fiber_length': 23195,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQ2RHBlZXJwb25wb24gMjc4NDY3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1 POND VIEW DR APT 301 (STEINMAN '
                                                 'ANNA MAE)',
                                         'poller_hash': 'MTI3MEQ2RA==',
                                         'serial_number': '1270D6D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:37:69',
                                         'uptime': '8291191'},
                                 {       'counter_source': '279057/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279057/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '49789123368',
                                                                       'in_ucast_pkts': '219138111',
                                                                       'mac_address': '04:bc:9f:6e:62:ae',
                                                                       'name': '279057/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1038238415146',
                                                                       'out_ucast_pkts': '743021798',
                                                                       'poller_hash': 'MTI5NUZEOXBlZXJpbnRmMjc5MDU3L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279057',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1038238415146',
                                                                       'in_ucast_pkts': '743021798',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1038238415146',
                                                                                                'in_ucast_pkts': '743021798',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '49789123368',
                                                                                                'out_ucast_pkts': '219138111',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzM='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '49789123368',
                                                                       'out_ucast_pkts': '219138111',
                                                                       'poller_hash': 'MTI5NUZEOXBlZXJpbnRmcG9uIDI3OTA1Nw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279057',
                                                                                               'links': [       {       'connected_time': '782079',
                                                                                                                        'fiber_length': 19899,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzM=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUZEOXBlZXJwb25wb24gMjc5MDU3',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4908 LINHAVEN DR GADSHILL (PINKNEY '
                                                 'BRANDON)',
                                         'poller_hash': 'MTI5NUZEOQ==',
                                         'serial_number': '1295FD9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:62:ae',
                                         'uptime': '782079'},
                                 {       'counter_source': '279316/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279316/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '15',
                                                                       'in_octets': '574758057468',
                                                                       'in_ucast_pkts': '0',
                                                                       'mac_address': '04:bc:9f:5b:9d:89',
                                                                       'name': '279316/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '508',
                                                                       'out_octets': '2934394857591',
                                                                       'out_ucast_pkts': '2265612906',
                                                                       'poller_hash': 'MTI3M0EzMXBlZXJpbnRmMjc5MzE2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279316',
                                                                       'in_nucast_pkts': '508',
                                                                       'in_octets': '2934394857591',
                                                                       'in_ucast_pkts': '2265612906',
                                                                       'links': [       {       'in_nucast_pkts': '508',
                                                                                                'in_octets': '2934394857591',
                                                                                                'in_ucast_pkts': '2265612906',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '15',
                                                                                                'out_octets': '574758057468',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '15',
                                                                       'out_octets': '574758057468',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3M0EzMXBlZXJpbnRmcG9uIDI3OTMxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279316',
                                                                                               'links': [       {       'connected_time': '6648143',
                                                                                                                        'fiber_length': 23798,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0EzMXBlZXJwb25wb24gMjc5MzE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '34 GALENA ST (SEHER STEFAN)',
                                         'poller_hash': 'MTI3M0EzMQ==',
                                         'serial_number': '1273A31',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:9d:89',
                                         'uptime': '6648143'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '279356/g1',
                                                                       'name': '279356/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTI3NDhCNHBlZXJpbnRmMjc5MzU2L2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279356',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '0',
                                                                       'in_ucast_pkts': '0',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '0',
                                                                                                'in_ucast_pkts': '0',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '0',
                                                                                                'out_ucast_pkts': '0',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '0',
                                                                       'out_ucast_pkts': '0',
                                                                       'poller_hash': 'MTI3NDhCNHBlZXJpbnRmcG9uIDI3OTM1Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279356',
                                                                                               'links': [       {       'connected_time': '2503016',
                                                                                                                        'fiber_length': 12020,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3NDhCNHBlZXJwb25wb24gMjc5MzU2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '5590 PERTH LINE 56A (KIPFER JESSICA)',
                                         'poller_hash': 'MTI3NDhCNA==',
                                         'serial_number': '12748B4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5c:11:a1',
                                         'uptime': '2503016'},
                                 {       'counter_source': '279416/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279416/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '74937254222',
                                                                       'in_ucast_pkts': '633151254',
                                                                       'mac_address': '04:bc:9f:5b:a6:11',
                                                                       'name': '279416/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '3110592237962',
                                                                       'out_ucast_pkts': '2094930341',
                                                                       'poller_hash': 'MTI3M0I0MnBlZXJpbnRmMjc5NDE2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279416',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '3110592237962',
                                                                       'in_ucast_pkts': '2094930341',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '3110592237962',
                                                                                                'in_ucast_pkts': '2094930341',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '74937254222',
                                                                                                'out_ucast_pkts': '633151254',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzY='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '74937254222',
                                                                       'out_ucast_pkts': '633151254',
                                                                       'poller_hash': 'MTI3M0I0MnBlZXJpbnRmcG9uIDI3OTQxNg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279416',
                                                                                               'links': [       {       'connected_time': '6043446',
                                                                                                                        'fiber_length': 12400,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzY=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0I0MnBlZXJwb25wb24gMjc5NDE2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4021 PERTH LINE 56 (KIPP DEBBIE & '
                                                 'NATHAN)',
                                         'poller_hash': 'MTI3M0I0Mg==',
                                         'serial_number': '1273B42',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:a6:11',
                                         'uptime': '6043446'},
                                 {       'counter_source': '279486/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279486/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '8051226611',
                                                                       'in_ucast_pkts': '24073121',
                                                                       'mac_address': '04:bc:9f:6e:5a:06',
                                                                       'name': '279486/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '108362666857',
                                                                       'out_ucast_pkts': '91146553',
                                                                       'poller_hash': 'MTI5NUVDNHBlZXJpbnRmMjc5NDg2L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279486',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '108362666857',
                                                                       'in_ucast_pkts': '91146553',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '108362666857',
                                                                                                'in_ucast_pkts': '91146553',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '8051226611',
                                                                                                'out_ucast_pkts': '24073121',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzg='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '8051226611',
                                                                       'out_ucast_pkts': '24073121',
                                                                       'poller_hash': 'MTI5NUVDNHBlZXJpbnRmcG9uIDI3OTQ4Ng==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279486',
                                                                                               'links': [       {       'connected_time': '5870595',
                                                                                                                        'fiber_length': 12372,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzg=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -13.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVDNHBlZXJwb25wb24gMjc5NDg2',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4030 PERTH LINE 56 (GERBER PHILIP & '
                                                 'ESTHER)',
                                         'poller_hash': 'MTI5NUVDNA==',
                                         'serial_number': '1295EC4',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:5a:06',
                                         'uptime': '5870595'},
                                 {       'counter_source': '279861/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279861/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '24505239936',
                                                                       'in_ucast_pkts': '121303891',
                                                                       'mac_address': '04:bc:9f:5b:a6:31',
                                                                       'name': '279861/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '735195607554',
                                                                       'out_ucast_pkts': '533955275',
                                                                       'poller_hash': 'MTI3M0I0NnBlZXJpbnRmMjc5ODYxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279861',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '735195607554',
                                                                       'in_ucast_pkts': '533955275',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '735195607554',
                                                                                                'in_ucast_pkts': '533955275',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '24505239936',
                                                                                                'out_ucast_pkts': '121303891',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '24505239936',
                                                                       'out_ucast_pkts': '121303891',
                                                                       'poller_hash': 'MTI3M0I0NnBlZXJpbnRmcG9uIDI3OTg2MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279861',
                                                                                               'links': [       {       'connected_time': '611459',
                                                                                                                        'fiber_length': 20385,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3M0I0NnBlZXJwb25wb24gMjc5ODYx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4980 PERTH ROAD 119 (MCMURRAY STEVE '
                                                 '& WENDY)',
                                         'poller_hash': 'MTI3M0I0Ng==',
                                         'serial_number': '1273B46',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5b:a6:31',
                                         'uptime': '611459'},
                                 {       'counter_source': '279891/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '279891/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '41541739749',
                                                                       'in_ucast_pkts': '138037451',
                                                                       'mac_address': '04:bc:9f:5a:35:91',
                                                                       'name': '279891/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1101342216531',
                                                                       'out_ucast_pkts': '772323692',
                                                                       'poller_hash': 'MTI3MEQzMnBlZXJpbnRmMjc5ODkxL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '279891',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1101342216531',
                                                                       'in_ucast_pkts': '772323692',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1101342216531',
                                                                                                'in_ucast_pkts': '772323692',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '41541739749',
                                                                                                'out_ucast_pkts': '138037451',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '41541739749',
                                                                       'out_ucast_pkts': '138037451',
                                                                       'poller_hash': 'MTI3MEQzMnBlZXJpbnRmcG9uIDI3OTg5MQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '279891',
                                                                                               'links': [       {       'connected_time': '585942',
                                                                                                                        'fiber_length': 23291,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI3MEQzMnBlZXJwb25wb24gMjc5ODkx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4 BERDUX PL (GRILLI JOHN)',
                                         'poller_hash': 'MTI3MEQzMg==',
                                         'serial_number': '1270D32',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:5a:35:91',
                                         'uptime': '585942'},
                                 {       'counter_source': '280264/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '04:bc:9f:46:8f:99'],
                                                                       'id': '280264/g1',
                                                                       'in_nucast_pkts': '85653',
                                                                       'in_octets': '24611570126',
                                                                       'in_ucast_pkts': '110193061',
                                                                       'mac_address': '04:bc:9f:6e:5a:81',
                                                                       'mtu': 2000,
                                                                       'name': '280264/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5655172',
                                                                       'out_octets': '457794689194',
                                                                       'out_ucast_pkts': '341464169',
                                                                       'poller_hash': 'MTI5NUVEM3BlZXJpbnRmMjgwMjY0L2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '280264',
                                                                       'in_nucast_pkts': '5655172',
                                                                       'in_octets': '457794689194',
                                                                       'in_ucast_pkts': '341464169',
                                                                       'links': [       {       'in_nucast_pkts': '5655172',
                                                                                                'in_octets': '457794689194',
                                                                                                'in_ucast_pkts': '341464169',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '85653',
                                                                                                'out_octets': '24611570126',
                                                                                                'out_ucast_pkts': '110193061',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1Lzc='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '85653',
                                                                       'out_octets': '24611570126',
                                                                       'out_ucast_pkts': '110193061',
                                                                       'poller_hash': 'MTI5NUVEM3BlZXJpbnRmcG9uIDI4MDI2NA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280264',
                                                                                               'links': [       {       'connected_time': '4835297',
                                                                                                                        'fiber_length': 19483,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1Lzc=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -23.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUVEM3BlZXJwb25wb24gMjgwMjY0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '1870 PERTH LINE (BRENNAN ANNE)',
                                         'poller_hash': 'MTI5NUVEMw==',
                                         'serial_number': '1295ED3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:5a:7e',
                                         'uptime': '4835297'},
                                 {       'counter_source': '280424/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '280424/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '169726796686',
                                                                       'in_ucast_pkts': '184720273',
                                                                       'mac_address': '04:bc:9f:6e:58:46',
                                                                       'name': '280424/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '53653861395',
                                                                       'out_ucast_pkts': '168717362',
                                                                       'poller_hash': 'MTI5NUU4Q3BlZXJpbnRmMjgwNDI0L0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '280424',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '53653861395',
                                                                       'in_ucast_pkts': '168717362',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '53653861395',
                                                                                                'in_ucast_pkts': '168717362',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '169726796686',
                                                                                                'out_ucast_pkts': '184720273',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '169726796686',
                                                                       'out_ucast_pkts': '184720273',
                                                                       'poller_hash': 'MTI5NUU4Q3BlZXJpbnRmcG9uIDI4MDQyNA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '280424',
                                                                                               'links': [       {       'connected_time': '4235183',
                                                                                                                        'fiber_length': 22543,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU4Q3BlZXJwb25wb24gMjgwNDI0',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': "1220 QUEEN'S BUSH RD (WELLESLEY "
                                                 'SERVICE CENTRE LTD)',
                                         'poller_hash': 'MTI5NUU4Qw==',
                                         'serial_number': '1295E8C',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:58:46',
                                         'uptime': '4235183'},
                                 {       'counter_source': '282243/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4e:e9:4f'],
                                                                       'id': '282243/g1',
                                                                       'in_nucast_pkts': '64948',
                                                                       'in_octets': '8589656553',
                                                                       'in_ucast_pkts': '17945799',
                                                                       'mac_address': '04:bc:9f:6e:48:09',
                                                                       'mtu': 2000,
                                                                       'name': '282243/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '3749479',
                                                                       'out_octets': '50654069692',
                                                                       'out_ucast_pkts': '45087911',
                                                                       'poller_hash': 'MTI5NUM4NHBlZXJpbnRmMjgyMjQzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '282243',
                                                                       'in_nucast_pkts': '3749479',
                                                                       'in_octets': '50654069692',
                                                                       'in_ucast_pkts': '45087911',
                                                                       'links': [       {       'in_nucast_pkts': '3749479',
                                                                                                'in_octets': '50654069692',
                                                                                                'in_ucast_pkts': '45087911',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '64948',
                                                                                                'out_octets': '8589656553',
                                                                                                'out_ucast_pkts': '17945799',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '64948',
                                                                       'out_octets': '8589656553',
                                                                       'out_ucast_pkts': '17945799',
                                                                       'poller_hash': 'MTI5NUM4NHBlZXJpbnRmcG9uIDI4MjI0Mw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282243',
                                                                                               'links': [       {       'connected_time': '3176203',
                                                                                                                        'fiber_length': 22957,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUM4NHBlZXJwb25wb24gMjgyMjQz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': 'ont 282243',
                                         'poller_hash': 'MTI5NUM4NA==',
                                         'serial_number': '1295C84',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:48:06',
                                         'uptime': '3176203'},
                                 {       'counter_source': '282323/G1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'id': '282323/G1',
                                                                       'in_errors': '0',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '30639804575',
                                                                       'in_ucast_pkts': '361664975',
                                                                       'mac_address': '04:bc:9f:6e:58:ce',
                                                                       'name': '282323/G1',
                                                                       'oper_up': True,
                                                                       'out_errors': '0',
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '1464971763666',
                                                                       'out_ucast_pkts': '972359711',
                                                                       'poller_hash': 'MTI5NUU5RHBlZXJpbnRmMjgyMzIzL0cx'},
                                                               {       'admin_up': True,
                                                                       'id': '282323',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '1464971763666',
                                                                       'in_ucast_pkts': '972359711',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '1464971763666',
                                                                                                'in_ucast_pkts': '972359711',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '0',
                                                                                                'out_octets': '30639804575',
                                                                                                'out_ucast_pkts': '361664975',
                                                                                                'poller_hash': 'MjExNjExNjA0MjgxZ3BvbiA1LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '30639804575',
                                                                       'out_ucast_pkts': '361664975',
                                                                       'poller_hash': 'MTI5NUU5RHBlZXJpbnRmcG9uIDI4MjMyMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282323',
                                                                                               'links': [       {       'connected_time': '2930576',
                                                                                                                        'fiber_length': 19633,
                                                                                                                        'poller_hash': 'MjExNjExNjA0MjgxcG9uZ3BvbiA1LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 20000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -27.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU5RHBlZXJwb25wb24gMjgyMzIz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '4908 PERTH ROAD 119 (GARY PATTERSON)',
                                         'poller_hash': 'MTI5NUU5RA==',
                                         'serial_number': '1295E9D',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:58:ce',
                                         'uptime': '2930576'},
                                 {       'counter_source': '282750/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'e4:6c:d1:4f:b7:79'],
                                                                       'id': '282750/g1',
                                                                       'in_nucast_pkts': '49963',
                                                                       'in_octets': '21507975923',
                                                                       'in_ucast_pkts': '130455961',
                                                                       'mac_address': '04:bc:9f:6e:57:f1',
                                                                       'mtu': 2000,
                                                                       'name': '282750/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2802214',
                                                                       'out_octets': '219689138859',
                                                                       'out_ucast_pkts': '196721849',
                                                                       'poller_hash': 'MTI5NUU4MXBlZXJpbnRmMjgyNzUwL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '282750',
                                                                       'in_nucast_pkts': '2802214',
                                                                       'in_octets': '219689138859',
                                                                       'in_ucast_pkts': '196721849',
                                                                       'links': [       {       'in_nucast_pkts': '2802214',
                                                                                                'in_octets': '219689138859',
                                                                                                'in_ucast_pkts': '196721849',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '49963',
                                                                                                'out_octets': '21507975923',
                                                                                                'out_ucast_pkts': '130455961',
                                                                                                'poller_hash': 'MjExNTA1NjAwMzI1Z3BvbiA3LzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '49963',
                                                                       'out_octets': '21507975923',
                                                                       'out_ucast_pkts': '130455961',
                                                                       'poller_hash': 'MTI5NUU4MXBlZXJpbnRmcG9uIDI4Mjc1MA==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '282750',
                                                                                               'links': [       {       'connected_time': '2413270',
                                                                                                                        'fiber_length': 23386,
                                                                                                                        'poller_hash': 'MjExNTA1NjAwMzI1cG9uZ3BvbiA3LzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -17.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUU4MXBlZXJwb25wb24gMjgyNzUw',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '138 LAWRENCE ST (MORAN JOHN)',
                                         'poller_hash': 'MTI5NUU4MQ==',
                                         'serial_number': '1295E81',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:57:ee',
                                         'uptime': '2413270'},
                                 {       'counter_source': '284433/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '60:db:98:08:ff:7c'],
                                                                       'id': '284433/g1',
                                                                       'in_nucast_pkts': '36302',
                                                                       'in_octets': '9352817013',
                                                                       'in_ucast_pkts': '35674433',
                                                                       'mac_address': '04:bc:9f:6e:62:d9',
                                                                       'mtu': 2000,
                                                                       'name': '284433/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '2005380',
                                                                       'out_octets': '436172473560',
                                                                       'out_ucast_pkts': '301464738',
                                                                       'poller_hash': 'MTI5NUZERXBlZXJpbnRmMjg0NDMzL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '284433',
                                                                       'in_nucast_pkts': '2005380',
                                                                       'in_octets': '436172473560',
                                                                       'in_ucast_pkts': '301464738',
                                                                       'links': [       {       'in_nucast_pkts': '2005380',
                                                                                                'in_octets': '436172473560',
                                                                                                'in_ucast_pkts': '301464738',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '36302',
                                                                                                'out_octets': '9352817013',
                                                                                                'out_ucast_pkts': '35674433',
                                                                                                'poller_hash': 'NDcxOTA1MDEwMzU2Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '36302',
                                                                       'out_octets': '9352817013',
                                                                       'out_ucast_pkts': '35674433',
                                                                       'poller_hash': 'MTI5NUZERXBlZXJpbnRmcG9uIDI4NDQzMw==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '284433',
                                                                                               'links': [       {       'connected_time': '1723414',
                                                                                                                        'fiber_length': 23059,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwMzU2cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NUZERXBlZXJwb25wb24gMjg0NDMz',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '8 POND VIEW DR (SCHELTER-BROWN '
                                                 'BONNE)',
                                         'poller_hash': 'MTI5NUZERQ==',
                                         'serial_number': '1295FDE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:62:d6',
                                         'uptime': '1723414'},
                                 {       'counter_source': '284562/g1',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '14:21:03:a9:74:a9'],
                                                                       'id': '284562/g1',
                                                                       'in_nucast_pkts': '26141',
                                                                       'in_octets': '51970229119',
                                                                       'in_ucast_pkts': '83997726',
                                                                       'mac_address': '04:bc:9f:6e:6c:71',
                                                                       'mtu': 2000,
                                                                       'name': '284562/g1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '1426539',
                                                                       'out_octets': '216266379728',
                                                                       'out_ucast_pkts': '171263199',
                                                                       'poller_hash': 'MTI5NjExMXBlZXJpbnRmMjg0NTYyL2cx',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '284562',
                                                                       'in_nucast_pkts': '1426539',
                                                                       'in_octets': '216266379728',
                                                                       'in_ucast_pkts': '171263199',
                                                                       'links': [       {       'in_nucast_pkts': '1426539',
                                                                                                'in_octets': '216266379728',
                                                                                                'in_ucast_pkts': '171263199',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '26141',
                                                                                                'out_octets': '51970229119',
                                                                                                'out_ucast_pkts': '83997726',
                                                                                                'poller_hash': 'MjExNjEyNjA2NzMwZ3BvbiA2LzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '26141',
                                                                       'out_octets': '51970229119',
                                                                       'out_ucast_pkts': '83997726',
                                                                       'poller_hash': 'MTI5NjExMXBlZXJpbnRmcG9uIDI4NDU2Mg==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '284562',
                                                                                               'links': [       {       'connected_time': '1217113',
                                                                                                                        'fiber_length': 22866,
                                                                                                                        'poller_hash': 'MjExNjEyNjA2NzMwcG9uZ3BvbiA2LzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -18.0,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.0}],
                                                                                               'max_distance': 60000,
                                                                                               'maximum_loss': -8.0,
                                                                                               'minimum_loss': -30.0,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTI5NjExMXBlZXJwb25wb24gMjg0NTYy',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': '812G-1',
                                         'name': '3634 NAFZIGER RD (SCHULTZ EMILY)',
                                         'poller_hash': 'MTI5NjExMQ==',
                                         'serial_number': '1296111',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '12.2.13.0.49',
                                         'system_mac_address': '04:bc:9f:6e:6c:6e',
                                         'uptime': '1217113'}]
        }
        id_field = 'mac_address' if not obj.stations else ('mac_address' if hasattr(obj.stations[0], 'mac_address') and getattr(obj.stations[0], 'mac_address') else ('imsi' if hasattr(obj.stations[0], 'imsi') else 'serial'))
        cur_info = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('info', 'ap', obj.model) + [('vendor', 'str'), ('model', 'str'), ('sw_version', 'str')]}
        if 'rf_freq' in cur_info and cur_info.get('rf_freq') and (cur_info['rf_freq'] <= 0 or int.bit_length(cur_info['rf_freq']) >= 32):
            cur_info.pop('rf_freq')
        cur_stats = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'ap', obj.model)}
        cur_sta_stats = {getattr(sta, id_field): {x: getattr(sta, x, None) for (x, data_type) in sorted(self.apdd.get_metric_fields('stats', 'cpe', obj.model)) + [(id_field, 'str'), ('ip_address', 'str'), ('name', 'str'), ('dev_macs', 'str')]} for sta in sorted(obj.stations)}
        cur_ne_msg = getattr(obj, 'network_element_update_msg', None)
        cur_ne_msg = cur_ne_msg.pb if cur_ne_msg else None
        self.assertEqual(cur_info, exp_info)
        self.assertEqual(cur_stats, exp_stats)
        self.assertEqual(cur_sta_stats, exp_sta_stats)
        self.assertTrue(self.compare_pb_dict(cur_ne_msg, exp_ne_msg))

    def test_calix_exa_E7_2_EXA_3_4_60_118_01(self):
        obj = self.run_data("calix.exa.E7-2 (EXA).3.4.60.118.01")
        exp_info = {
                'model': 'E7-2 (EXA)',
                'sta_count': 16,
                'sw_version': '3.4.60.118',
                'vendor': 'Calix'
        }
        exp_stats = {
         'serial_number': '71904927309', 'system_mac_address': None
        }
        exp_sta_stats = {
                '166DD36': {       'dev_macs': ['f8:27:2e:18:40:21'],
                                   'ip_address': None,
                                   'name': '141582_AJJonker (1 Ashgrove Lane)',
                                   'serial': '166DD36',
                                   'serial_number': '166DD36',
                                   'system_mac_address': None},
                '166DDCE': {       'dev_macs': ['f8:27:2e:18:40:d5'],
                                   'ip_address': None,
                                   'name': '141825_PeterSimpson (138 Grove Road)',
                                   'serial': '166DDCE',
                                   'serial_number': '166DDCE',
                                   'system_mac_address': None},
                '166DDE2': {       'dev_macs': ['f8:ca:59:4c:99:cd'],
                                   'ip_address': None,
                                   'name': '140301_JacobAckerman (95 Grove Road)',
                                   'serial': '166DDE2',
                                   'serial_number': '166DDE2',
                                   'system_mac_address': None},
                '166DE2B': {       'dev_macs': ['f8:27:2e:18:41:6b'],
                                   'ip_address': None,
                                   'name': '141113_TheHerbFarm (86 Grove Road)',
                                   'serial': '166DE2B',
                                   'serial_number': '166DE2B',
                                   'system_mac_address': None},
                '166DFD6': {       'dev_macs': ['f8:27:2e:18:39:f5'],
                                   'ip_address': None,
                                   'name': '141084_BillKirkland (72 Grove Road)',
                                   'serial': '166DFD6',
                                   'serial_number': '166DFD6',
                                   'system_mac_address': None},
                '166E1B9': {       'dev_macs': ['f8:ca:59:ad:34:7e'],
                                   'ip_address': None,
                                   'name': '140906 Michaela Tamblyn (730 Colyton Road)',
                                   'serial': '166E1B9',
                                   'serial_number': '166E1B9',
                                   'system_mac_address': None},
                '166E244': {       'dev_macs': ['f8:ca:59:48:fa:ac'],
                                   'ip_address': None,
                                   'name': '140320_LizzieAsh (122 Grove Road)',
                                   'serial': '166E244',
                                   'serial_number': '166E244',
                                   'system_mac_address': None},
                '166E2F3': {       'dev_macs': ['18:f1:45:2a:bf:f9'],
                                   'ip_address': None,
                                   'name': '139906 Donna MacMillan (694 Colyton Road)',
                                   'serial': '166E2F3',
                                   'serial_number': '166E2F3',
                                   'system_mac_address': None},
                '166E34D': {       'dev_macs': [],
                                   'ip_address': None,
                                   'name': '142100 Jonty Cahill (670 Colyton Road)',
                                   'serial': '166E34D',
                                   'serial_number': '166E34D',
                                   'system_mac_address': None},
                '166E358': {       'dev_macs': ['f8:ca:59:59:7f:75'],
                                   'ip_address': None,
                                   'name': '109378_MattHermann (18 Ashgrove Lane)',
                                   'serial': '166E358',
                                   'serial_number': '166E358',
                                   'system_mac_address': None},
                '166E5A6': {       'dev_macs': ['f8:27:2e:18:41:f7'],
                                   'ip_address': None,
                                   'name': '140645 Bevan Heap (634 Colyton Road)',
                                   'serial': '166E5A6',
                                   'serial_number': '166E5A6',
                                   'system_mac_address': None},
                '166E651': {       'dev_macs': ['18:f1:45:cc:bb:be'],
                                   'ip_address': None,
                                   'name': '140237 Jamie McNabb (707 Colyton Road)',
                                   'serial': '166E651',
                                   'serial_number': '166E651',
                                   'system_mac_address': None},
                '166E827': {       'dev_macs': ['f8:27:2e:18:40:17'],
                                   'ip_address': None,
                                   'name': '141575_ArnoldCraw (81 Grove Road)',
                                   'serial': '166E827',
                                   'serial_number': '166E827',
                                   'system_mac_address': None},
                '166E82B': {       'dev_macs': ['f8:27:2e:18:45:99'],
                                   'ip_address': None,
                                   'name': '141311_RussellandAnnetteHughes (19 '
                                           'Ashgrove Lane)',
                                   'serial': '166E82B',
                                   'serial_number': '166E82B',
                                   'system_mac_address': None},
                '166E8B0': {       'dev_macs': ['f8:ca:59:0a:aa:cd'],
                                   'ip_address': None,
                                   'name': '140512_MichaelSchnell (130 Harrisons Line)',
                                   'serial': '166E8B0',
                                   'serial_number': '166E8B0',
                                   'system_mac_address': None},
                '43DF26': {       'dev_macs': [],
                                  'ip_address': None,
                                  'name': 'ont 1',
                                  'serial': '43DF26',
                                  'serial_number': '43DF26',
                                  'system_mac_address': None}
        }
        exp_ne_msg = {
                'data': {       'interfaces': [       {       'admin_up': True,
                                                              'id': 'front',
                                                              'mac_address': '48:77:46:05:e9:30',
                                                              'name': 'Craft-FE: front',
                                                              'oper_up': True,
                                                              'poller_hash': 'NzE5MDQ5MjczMDlmcm9udA=='},
                                                      {       'admin_up': False,
                                                              'id': 'rear',
                                                              'name': 'Craft-FE: rear',
                                                              'oper_up': False,
                                                              'poller_hash': 'NzE5MDQ5MjczMDlyZWFy'}],
                                'manufacturer': 'Calix',
                                'model': 'E7-2 (EXA)',
                                'modules': [       {       'id': '1',
                                                           'interfaces': [       {       'admin_up': True,
                                                                                         'id': '1/x3',
                                                                                         'in_discards': '123989737',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '199330877',
                                                                                         'in_octets': '12091229830603',
                                                                                         'in_ucast_pkts': '9156776137',
                                                                                         'mac_address': '48:77:46:05:e9:3f',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-01793',
                                                                                         'name': 'ethernet '
                                                                                                 '1/x3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '1078881',
                                                                                         'out_octets': '1137335927335',
                                                                                         'out_ucast_pkts': '3469057735',
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1ZXRoZXJuZXQgMS94Mw==',
                                                                                         'speed': '10000000000'},
                                                                                 {       'admin_up': False,
                                                                                         'id': '1/x4',
                                                                                         'mac_address': '48:77:46:05:e9:40',
                                                                                         'name': 'ethernet '
                                                                                                 '1/x4',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1ZXRoZXJuZXQgMS94NA=='},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/g1',
                                                                                         'mac_address': '48:77:46:05:e9:31',
                                                                                         'name': 'ethernet '
                                                                                                 '1/g1',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1ZXRoZXJuZXQgMS9nMQ=='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '1/g2',
                                                                                         'mac_address': '48:77:46:05:e9:32',
                                                                                         'name': 'ethernet '
                                                                                                 '1/g2',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1ZXRoZXJuZXQgMS9nMg=='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '1/g3',
                                                                                         'mac_address': '48:77:46:05:e9:33',
                                                                                         'name': 'ethernet '
                                                                                                 '1/g3',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1ZXRoZXJuZXQgMS9nMw=='},
                                                                                 {       'admin_up': False,
                                                                                         'id': '1/g4',
                                                                                         'mac_address': '48:77:46:05:e9:34',
                                                                                         'name': 'ethernet '
                                                                                                 '1/g4',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1ZXRoZXJuZXQgMS9nNA=='},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/1',
                                                                                         'in_discards': '7304648',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '7304623',
                                                                                         'in_octets': '441566744303',
                                                                                         'in_ucast_pkts': '1492602835',
                                                                                         'links': [       {       'in_nucast_pkts': '11',
                                                                                                                  'in_octets': '10505522748',
                                                                                                                  'in_ucast_pkts': '36064250',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '143894240373',
                                                                                                                  'out_ucast_pkts': '118894319',
                                                                                                                  'poller_hash': 'MTY2RTJGM3BlZXJpbnRmcG9uIDU='},
                                                                                                          {       'in_nucast_pkts': '150',
                                                                                                                  'in_octets': '127968213925',
                                                                                                                  'in_ucast_pkts': '187208789',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '392491677073',
                                                                                                                  'out_ucast_pkts': '319960616',
                                                                                                                  'poller_hash': 'MTY2RTFCOXBlZXJpbnRmcG9uIDc='},
                                                                                                          {       'in_nucast_pkts': '67201',
                                                                                                                  'in_octets': '59987939528',
                                                                                                                  'in_ucast_pkts': '204932933',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7',
                                                                                                                  'out_octets': '1568997571201',
                                                                                                                  'out_ucast_pkts': '1117332527',
                                                                                                                  'poller_hash': 'MTY2RTVBNnBlZXJpbnRmcG9uIDk='},
                                                                                                          {       'in_nucast_pkts': '85',
                                                                                                                  'in_octets': '49925474171',
                                                                                                                  'in_ucast_pkts': '105757716',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '232016143753',
                                                                                                                  'out_ucast_pkts': '192346402',
                                                                                                                  'poller_hash': 'MTY2RTY1MXBlZXJpbnRmcG9uIDEw'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '1/1',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '5088692',
                                                                                         'out_octets': '6344874012475',
                                                                                         'out_ucast_pkts': '4580013990',
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzE=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/1',
                                                                                                                 'id': '1/1',
                                                                                                                 'links': [       {       'connected_time': '4397172',
                                                                                                                                          'fiber_length': 6593,
                                                                                                                                          'poller_hash': 'MTY2RTJGM3BlZXJwb25wb24gNQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.264},
                                                                                                                                  {       'connected_time': '4386698',
                                                                                                                                          'fiber_length': 6581,
                                                                                                                                          'poller_hash': 'MTY2RTFCOXBlZXJwb25wb24gNw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.0,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.264},
                                                                                                                                  {       'connected_time': '3696088',
                                                                                                                                          'fiber_length': 7210,
                                                                                                                                          'poller_hash': 'MTY2RTVBNnBlZXJwb25wb24gOQ==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.264},
                                                                                                                                  {       'connected_time': '3687286',
                                                                                                                                          'fiber_length': 6442,
                                                                                                                                          'poller_hash': 'MTY2RTY1MXBlZXJwb25wb24gMTA=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -17.9,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.264}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/1',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzE=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/2',
                                                                                         'in_discards': '126903',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '245679',
                                                                                         'in_octets': '679611122477',
                                                                                         'in_ucast_pkts': '1901803041',
                                                                                         'links': [       {       'in_nucast_pkts': '121826',
                                                                                                                  'in_octets': '52674780452',
                                                                                                                  'in_ucast_pkts': '134220087',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14',
                                                                                                                  'out_octets': '484962338039',
                                                                                                                  'out_ucast_pkts': '355558712',
                                                                                                                  'poller_hash': 'MTY2RERFMnBlZXJpbnRmcG9uIDI='},
                                                                                                          {       'in_nucast_pkts': '16',
                                                                                                                  'in_octets': '8007111241',
                                                                                                                  'in_ucast_pkts': '26138878',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '96584184989',
                                                                                                                  'out_ucast_pkts': '79866788',
                                                                                                                  'poller_hash': 'MTY2RThCMHBlZXJpbnRmcG9uIDM='},
                                                                                                          {       'in_nucast_pkts': '5',
                                                                                                                  'in_octets': '3303219759',
                                                                                                                  'in_ucast_pkts': '13960628',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '0',
                                                                                                                  'out_octets': '51068345808',
                                                                                                                  'out_ucast_pkts': '38041407',
                                                                                                                  'poller_hash': 'MTY2RTM1OHBlZXJpbnRmcG9uIDQ='},
                                                                                                          {       'in_nucast_pkts': '27644',
                                                                                                                  'in_octets': '128817160764',
                                                                                                                  'in_ucast_pkts': '439720206',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7',
                                                                                                                  'out_octets': '1072996760052',
                                                                                                                  'out_ucast_pkts': '836806660',
                                                                                                                  'poller_hash': 'MTY2RERDRXBlZXJpbnRmcG9uIDEz'},
                                                                                                          {       'in_nucast_pkts': '27643',
                                                                                                                  'in_octets': '41581374486',
                                                                                                                  'in_ucast_pkts': '92785301',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7',
                                                                                                                  'out_octets': '344681570190',
                                                                                                                  'out_ucast_pkts': '287058533',
                                                                                                                  'poller_hash': 'MTY2REUyQnBlZXJpbnRmcG9uIDE0'},
                                                                                                          {       'in_nucast_pkts': '10',
                                                                                                                  'in_octets': '25352610052',
                                                                                                                  'in_ucast_pkts': '108304178',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14',
                                                                                                                  'out_octets': '370763996083',
                                                                                                                  'out_ucast_pkts': '295912483',
                                                                                                                  'poller_hash': 'MTY2RTI0NHBlZXJpbnRmcG9uIDE1'},
                                                                                                          {       'in_nucast_pkts': '27625',
                                                                                                                  'in_octets': '18879918486',
                                                                                                                  'in_ucast_pkts': '86040396',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7',
                                                                                                                  'out_octets': '495745453822',
                                                                                                                  'out_ucast_pkts': '374457374',
                                                                                                                  'poller_hash': 'MTY2RTgyN3BlZXJpbnRmcG9uIDE2'},
                                                                                                          {       'in_nucast_pkts': '109973',
                                                                                                                  'in_octets': '14834051241',
                                                                                                                  'in_ucast_pkts': '56632797',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '7',
                                                                                                                  'out_octets': '245236369864',
                                                                                                                  'out_ucast_pkts': '180593220',
                                                                                                                  'poller_hash': 'MTY2REQzNnBlZXJpbnRmcG9uIDE3'},
                                                                                                          {       'in_nucast_pkts': '27643',
                                                                                                                  'in_octets': '4367043755',
                                                                                                                  'in_ucast_pkts': '18880744',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14',
                                                                                                                  'out_octets': '53620654324',
                                                                                                                  'out_ucast_pkts': '46552388',
                                                                                                                  'poller_hash': 'MTY2RTgyQnBlZXJpbnRmcG9uIDE4'},
                                                                                                          {       'in_nucast_pkts': '12816',
                                                                                                                  'in_octets': '136653847820',
                                                                                                                  'in_ucast_pkts': '165388692',
                                                                                                                  'oper_up': True,
                                                                                                                  'out_nucast_pkts': '14',
                                                                                                                  'out_octets': '230595532402',
                                                                                                                  'out_ucast_pkts': '264515272',
                                                                                                                  'poller_hash': 'MTY2REZENnBlZXJpbnRmcG9uIDE5'}],
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '1/2',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '0',
                                                                                         'out_octets': '5727264099108',
                                                                                         'out_ucast_pkts': '4514178376',
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/2',
                                                                                                                 'id': '1/2',
                                                                                                                 'links': [       {       'connected_time': '1520324',
                                                                                                                                          'fiber_length': 6688,
                                                                                                                                          'poller_hash': 'MTY2RERFMnBlZXJwb25wb24gMg==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '1372995',
                                                                                                                                          'fiber_length': 7197,
                                                                                                                                          'poller_hash': 'MTY2RThCMHBlZXJwb25wb24gMw==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -20.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '493999',
                                                                                                                                          'fiber_length': 6786,
                                                                                                                                          'poller_hash': 'MTY2RTM1OHBlZXJwb25wb24gNA==',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.7,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '1520331',
                                                                                                                                          'fiber_length': 7087,
                                                                                                                                          'poller_hash': 'MTY2RERDRXBlZXJwb25wb24gMTM=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '1520331',
                                                                                                                                          'fiber_length': 6730,
                                                                                                                                          'poller_hash': 'MTY2REUyQnBlZXJwb25wb24gMTQ=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.5,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '1520331',
                                                                                                                                          'fiber_length': 6770,
                                                                                                                                          'poller_hash': 'MTY2RTI0NHBlZXJwb25wb24gMTU=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '1520325',
                                                                                                                                          'fiber_length': 6626,
                                                                                                                                          'poller_hash': 'MTY2RTgyN3BlZXJwb25wb24gMTY=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '1520326',
                                                                                                                                          'fiber_length': 6612,
                                                                                                                                          'poller_hash': 'MTY2REQzNnBlZXJwb25wb24gMTc=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.4,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '1520324',
                                                                                                                                          'fiber_length': 6803,
                                                                                                                                          'poller_hash': 'MTY2RTgyQnBlZXJwb25wb24gMTg=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -18.2,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439},
                                                                                                                                  {       'connected_time': '339718',
                                                                                                                                          'fiber_length': 6893,
                                                                                                                                          'poller_hash': 'MTY2REZENnBlZXJwb25wb24gMTk=',
                                                                                                                                          'rx_link_rate': '1244160000',
                                                                                                                                          'rx_power': -19.8,
                                                                                                                                          'tx_link_rate': '2488320000',
                                                                                                                                          'tx_power': 3.439}],
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/2',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/3',
                                                                                         'in_discards': '0',
                                                                                         'in_errors': '0',
                                                                                         'in_nucast_pkts': '0',
                                                                                         'in_octets': '0',
                                                                                         'in_ucast_pkts': '0',
                                                                                         'manufacturer': 'Calix',
                                                                                         'model': '100-05121',
                                                                                         'name': 'gpon '
                                                                                                 '1/3',
                                                                                         'oper_up': True,
                                                                                         'out_nucast_pkts': '0',
                                                                                         'out_octets': '0',
                                                                                         'out_ucast_pkts': '0',
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzM=',
                                                                                         'pons': [       {       'description': 'gpon '
                                                                                                                                '1/3',
                                                                                                                 'id': '1/3',
                                                                                                                 'max_distance': 20000,
                                                                                                                 'maximum_loss': -8.0,
                                                                                                                 'minimum_loss': -28.0,
                                                                                                                 'name': 'gpon '
                                                                                                                         '1/3',
                                                                                                                 'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzM=',
                                                                                                                 'pon_technology': 'GPON',
                                                                                                                 'rx_link_rate': '1244160000',
                                                                                                                 'split_ratio': 64,
                                                                                                                 'tx_link_rate': '2488320000'}]},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/4',
                                                                                         'name': 'gpon '
                                                                                                 '1/4',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzQ='},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/5',
                                                                                         'name': 'gpon '
                                                                                                 '1/5',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzU='},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/6',
                                                                                         'name': 'gpon '
                                                                                                 '1/6',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzY='},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/7',
                                                                                         'name': 'gpon '
                                                                                                 '1/7',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzc='},
                                                                                 {       'admin_up': True,
                                                                                         'id': '1/8',
                                                                                         'name': 'gpon '
                                                                                                 '1/8',
                                                                                         'oper_up': False,
                                                                                         'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzg='}],
                                                           'manufacturer': 'Calix',
                                                           'model': '100-03006',
                                                           'name': 'GPON-8',
                                                           'poller_hash': 'NDcxOTA1MDEwNDE1',
                                                           'serial_number': '471905010415',
                                                           'sw_version': '3.4.60.118',
                                                           'uptime': '9576134'}],
                                'serial_number': '71904927309',
                                'sw_version': '3.4.60.118',
                                'system_mac_address': '48:77:46:05:e9:2f'},
                'peers': [       {       'counter_source': '',
                                         'interfaces': [       {       'id': '1/g1',
                                                                       'name': '1/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNERjI2cGVlcmludGYxL2cx'},
                                                               {       'admin_up': True,
                                                                       'id': '1',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'NDNERjI2cGVlcmludGZwb24gMQ==',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '1',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'NDNERjI2cGVlcnBvbnBvbiAx',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': 'ont 1',
                                         'offline_reason': 'OFFLINE_REASON_LOSS_OF_SIGNAL',
                                         'offline_reason_detail': 'ONT "1" (location: '
                                                                  'GPON port "1/1"), '
                                                                  '"ONT went missing", '
                                                                  '(suppressed)',
                                         'offline_time': '2025-05-29T01:28:27.970Z',
                                         'poller_hash': 'NDNERjI2',
                                         'serial_number': '43DF26',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '2/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:ca:59:4c:99:cd'],
                                                                       'id': '2/g4',
                                                                       'in_nucast_pkts': '121826',
                                                                       'in_octets': '52674780452',
                                                                       'in_ucast_pkts': '134220087',
                                                                       'mac_address': '1c:8b:76:b5:07:6b',
                                                                       'mtu': 9974,
                                                                       'name': '2/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '484962338039',
                                                                       'out_ucast_pkts': '355558712',
                                                                       'poller_hash': 'MTY2RERFMnBlZXJpbnRmMi9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '2',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '484962338039',
                                                                       'in_ucast_pkts': '355558712',
                                                                       'links': [       {       'in_nucast_pkts': '14',
                                                                                                'in_octets': '484962338039',
                                                                                                'in_ucast_pkts': '355558712',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '121826',
                                                                                                'out_octets': '52674780452',
                                                                                                'out_ucast_pkts': '134220087',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '121826',
                                                                       'out_octets': '52674780452',
                                                                       'out_ucast_pkts': '134220087',
                                                                       'poller_hash': 'MTY2RERFMnBlZXJpbnRmcG9uIDI=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '2',
                                                                                               'links': [       {       'connected_time': '1520324',
                                                                                                                        'fiber_length': 6688,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.542,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.302}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RERFMnBlZXJwb25wb24gMg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '140301_JacobAckerman (95 Grove Road)',
                                         'poller_hash': 'MTY2RERFMg==',
                                         'serial_number': '166DDE2',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:07:65',
                                         'uptime': '1520324'},
                                 {       'counter_source': '3/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:ca:59:0a:aa:cd'],
                                                                       'id': '3/g4',
                                                                       'in_nucast_pkts': '16',
                                                                       'in_octets': '8007111241',
                                                                       'in_ucast_pkts': '26138878',
                                                                       'mac_address': '1c:8b:76:b5:5d:db',
                                                                       'mtu': 9974,
                                                                       'name': '3/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '96584184989',
                                                                       'out_ucast_pkts': '79866788',
                                                                       'poller_hash': 'MTY2RThCMHBlZXJpbnRmMy9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '3',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '96584184989',
                                                                       'in_ucast_pkts': '79866788',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '96584184989',
                                                                                                'in_ucast_pkts': '79866788',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '16',
                                                                                                'out_octets': '8007111241',
                                                                                                'out_ucast_pkts': '26138878',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '16',
                                                                       'out_octets': '8007111241',
                                                                       'out_ucast_pkts': '26138878',
                                                                       'poller_hash': 'MTY2RThCMHBlZXJpbnRmcG9uIDM=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '3',
                                                                                               'links': [       {       'connected_time': '1372995',
                                                                                                                        'fiber_length': 7197,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -21.366,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.454}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RThCMHBlZXJwb25wb24gMw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '140512_MichaelSchnell (130 Harrisons '
                                                 'Line)',
                                         'poller_hash': 'MTY2RThCMA==',
                                         'serial_number': '166E8B0',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:5d:d5',
                                         'uptime': '1372995'},
                                 {       'counter_source': '4/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:ca:59:59:7f:75'],
                                                                       'id': '4/g4',
                                                                       'in_nucast_pkts': '5',
                                                                       'in_octets': '3303219759',
                                                                       'in_ucast_pkts': '13960628',
                                                                       'mac_address': '1c:8b:76:b5:33:1b',
                                                                       'mtu': 9974,
                                                                       'name': '4/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '51068345808',
                                                                       'out_ucast_pkts': '38041407',
                                                                       'poller_hash': 'MTY2RTM1OHBlZXJpbnRmNC9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '4',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '51068345808',
                                                                       'in_ucast_pkts': '38041407',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '51068345808',
                                                                                                'in_ucast_pkts': '38041407',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '5',
                                                                                                'out_octets': '3303219759',
                                                                                                'out_ucast_pkts': '13960628',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '5',
                                                                       'out_octets': '3303219759',
                                                                       'out_ucast_pkts': '13960628',
                                                                       'poller_hash': 'MTY2RTM1OHBlZXJpbnRmcG9uIDQ=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '4',
                                                                                               'links': [       {       'connected_time': '493999',
                                                                                                                        'fiber_length': 6786,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.766,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.3}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTM1OHBlZXJwb25wb24gNA==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '109378_MattHermann (18 Ashgrove '
                                                 'Lane)',
                                         'poller_hash': 'MTY2RTM1OA==',
                                         'serial_number': '166E358',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:33:15',
                                         'uptime': '493999'},
                                 {       'counter_source': '5/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:f1:45:2a:bf:f9'],
                                                                       'id': '5/g4',
                                                                       'in_nucast_pkts': '11',
                                                                       'in_octets': '10505522748',
                                                                       'in_ucast_pkts': '36064250',
                                                                       'mac_address': '1c:8b:76:b5:2f:f3',
                                                                       'mtu': 9974,
                                                                       'name': '5/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '143894240373',
                                                                       'out_ucast_pkts': '118894319',
                                                                       'poller_hash': 'MTY2RTJGM3BlZXJpbnRmNS9nNA==',
                                                                       'speed': '100000000'},
                                                               {       'admin_up': True,
                                                                       'id': '5',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '143894240373',
                                                                       'in_ucast_pkts': '118894319',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '143894240373',
                                                                                                'in_ucast_pkts': '118894319',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '11',
                                                                                                'out_octets': '10505522748',
                                                                                                'out_ucast_pkts': '36064250',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '11',
                                                                       'out_octets': '10505522748',
                                                                       'out_ucast_pkts': '36064250',
                                                                       'poller_hash': 'MTY2RTJGM3BlZXJpbnRmcG9uIDU=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '5',
                                                                                               'links': [       {       'connected_time': '4397172',
                                                                                                                        'fiber_length': 6593,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.722,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.868}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTJGM3BlZXJwb25wb24gNQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '139906 Donna MacMillan (694 Colyton '
                                                 'Road)',
                                         'poller_hash': 'MTY2RTJGMw==',
                                         'serial_number': '166E2F3',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:2f:ed',
                                         'uptime': '4397172'},
                                 {       'counter_source': '',
                                         'interfaces': [       {       'id': '6/g1',
                                                                       'name': '6/g1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTY2RTM0RHBlZXJpbnRmNi9nMQ=='},
                                                               {       'admin_up': True,
                                                                       'id': '6',
                                                                       'name': 'pon 1',
                                                                       'oper_up': False,
                                                                       'poller_hash': 'MTY2RTM0RHBlZXJpbnRmcG9uIDY=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '6',
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTM0RHBlZXJwb25wb24gNg==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'name': '142100 Jonty Cahill (670 Colyton '
                                                 'Road)',
                                         'offline_reason': 'OFFLINE_REASON_POWER_OFF',
                                         'offline_reason_detail': 'ONT "6" (location: '
                                                                  'GPON port "1/1"), '
                                                                  '"ONT Dying Gasp"',
                                         'offline_time': '2025-07-12T23:51:35.230Z',
                                         'poller_hash': 'MTY2RTM0RA==',
                                         'serial_number': '166E34D',
                                         'status': 'ELEMENT_STATUS_OFFLINE'},
                                 {       'counter_source': '7/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:ca:59:ad:34:7e'],
                                                                       'id': '7/g4',
                                                                       'in_nucast_pkts': '150',
                                                                       'in_octets': '127968213925',
                                                                       'in_ucast_pkts': '187208789',
                                                                       'mac_address': '1c:8b:76:b5:26:23',
                                                                       'mtu': 9974,
                                                                       'name': '7/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '392491677073',
                                                                       'out_ucast_pkts': '319960616',
                                                                       'poller_hash': 'MTY2RTFCOXBlZXJpbnRmNy9nNA==',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '7',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '392491677073',
                                                                       'in_ucast_pkts': '319960616',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '392491677073',
                                                                                                'in_ucast_pkts': '319960616',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '150',
                                                                                                'out_octets': '127968213925',
                                                                                                'out_ucast_pkts': '187208789',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '150',
                                                                       'out_octets': '127968213925',
                                                                       'out_ucast_pkts': '187208789',
                                                                       'poller_hash': 'MTY2RTFCOXBlZXJpbnRmcG9uIDc=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '7',
                                                                                               'links': [       {       'connected_time': '4386698',
                                                                                                                        'fiber_length': 6581,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.1,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.486}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTFCOXBlZXJwb25wb24gNw==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '140906 Michaela Tamblyn (730 Colyton '
                                                 'Road)',
                                         'poller_hash': 'MTY2RTFCOQ==',
                                         'serial_number': '166E1B9',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:26:1d',
                                         'uptime': '4386698'},
                                 {       'counter_source': '9/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:18:41:f7'],
                                                                       'id': '9/g4',
                                                                       'in_nucast_pkts': '67201',
                                                                       'in_octets': '59987939528',
                                                                       'in_ucast_pkts': '204932933',
                                                                       'mac_address': '1c:8b:76:b5:45:8b',
                                                                       'mtu': 9974,
                                                                       'name': '9/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '1568997571201',
                                                                       'out_ucast_pkts': '1117332527',
                                                                       'poller_hash': 'MTY2RTVBNnBlZXJpbnRmOS9nNA==',
                                                                       'speed': '2500000000'},
                                                               {       'admin_up': True,
                                                                       'id': '9',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '1568997571201',
                                                                       'in_ucast_pkts': '1117332527',
                                                                       'links': [       {       'in_nucast_pkts': '7',
                                                                                                'in_octets': '1568997571201',
                                                                                                'in_ucast_pkts': '1117332527',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '67201',
                                                                                                'out_octets': '59987939528',
                                                                                                'out_ucast_pkts': '204932933',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '67201',
                                                                       'out_octets': '59987939528',
                                                                       'out_ucast_pkts': '204932933',
                                                                       'poller_hash': 'MTY2RTVBNnBlZXJpbnRmcG9uIDk=',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '9',
                                                                                               'links': [       {       'connected_time': '3696088',
                                                                                                                        'fiber_length': 7210,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.002,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.19}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTVBNnBlZXJwb25wb24gOQ==',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '140645 Bevan Heap (634 Colyton Road)',
                                         'poller_hash': 'MTY2RTVBNg==',
                                         'serial_number': '166E5A6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:45:85',
                                         'uptime': '3696088'},
                                 {       'counter_source': '10/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       '18:f1:45:cc:bb:be'],
                                                                       'id': '10/g4',
                                                                       'in_nucast_pkts': '85',
                                                                       'in_octets': '49925474171',
                                                                       'in_ucast_pkts': '105757716',
                                                                       'mac_address': '1c:8b:76:b5:4a:e3',
                                                                       'mtu': 9974,
                                                                       'name': '10/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '0',
                                                                       'out_octets': '232016143753',
                                                                       'out_ucast_pkts': '192346402',
                                                                       'poller_hash': 'MTY2RTY1MXBlZXJpbnRmMTAvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '10',
                                                                       'in_nucast_pkts': '0',
                                                                       'in_octets': '232016143753',
                                                                       'in_ucast_pkts': '192346402',
                                                                       'links': [       {       'in_nucast_pkts': '0',
                                                                                                'in_octets': '232016143753',
                                                                                                'in_ucast_pkts': '192346402',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '85',
                                                                                                'out_octets': '49925474171',
                                                                                                'out_ucast_pkts': '105757716',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzE='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '85',
                                                                       'out_octets': '49925474171',
                                                                       'out_ucast_pkts': '105757716',
                                                                       'poller_hash': 'MTY2RTY1MXBlZXJpbnRmcG9uIDEw',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '10',
                                                                                               'links': [       {       'connected_time': '3687286',
                                                                                                                        'fiber_length': 6442,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzE=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -14.57,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.524}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTY1MXBlZXJwb25wb24gMTA=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '140237 Jamie McNabb (707 Colyton '
                                                 'Road)',
                                         'poller_hash': 'MTY2RTY1MQ==',
                                         'serial_number': '166E651',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:4a:dd',
                                         'uptime': '3687286'},
                                 {       'counter_source': '13/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:18:40:d5'],
                                                                       'id': '13/g4',
                                                                       'in_nucast_pkts': '27644',
                                                                       'in_octets': '128817160764',
                                                                       'in_ucast_pkts': '439720206',
                                                                       'mac_address': '1c:8b:76:b5:06:cb',
                                                                       'mtu': 9974,
                                                                       'name': '13/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '1072996760052',
                                                                       'out_ucast_pkts': '836806660',
                                                                       'poller_hash': 'MTY2RERDRXBlZXJpbnRmMTMvZzQ=',
                                                                       'speed': '2500000000'},
                                                               {       'admin_up': True,
                                                                       'id': '13',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '1072996760052',
                                                                       'in_ucast_pkts': '836806660',
                                                                       'links': [       {       'in_nucast_pkts': '7',
                                                                                                'in_octets': '1072996760052',
                                                                                                'in_ucast_pkts': '836806660',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27644',
                                                                                                'out_octets': '128817160764',
                                                                                                'out_ucast_pkts': '439720206',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27644',
                                                                       'out_octets': '128817160764',
                                                                       'out_ucast_pkts': '439720206',
                                                                       'poller_hash': 'MTY2RERDRXBlZXJpbnRmcG9uIDEz',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '13',
                                                                                               'links': [       {       'connected_time': '1520331',
                                                                                                                        'fiber_length': 7087,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -19.828,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.912}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RERDRXBlZXJwb25wb24gMTM=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '141825_PeterSimpson (138 Grove Road)',
                                         'poller_hash': 'MTY2RERDRQ==',
                                         'serial_number': '166DDCE',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:06:c5',
                                         'uptime': '1520331'},
                                 {       'counter_source': '14/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:18:41:6b'],
                                                                       'id': '14/g4',
                                                                       'in_nucast_pkts': '27643',
                                                                       'in_octets': '41581374486',
                                                                       'in_ucast_pkts': '92785301',
                                                                       'mac_address': '1c:8b:76:b5:09:b3',
                                                                       'mtu': 9974,
                                                                       'name': '14/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '344681570190',
                                                                       'out_ucast_pkts': '287058533',
                                                                       'poller_hash': 'MTY2REUyQnBlZXJpbnRmMTQvZzQ=',
                                                                       'speed': '2500000000'},
                                                               {       'admin_up': True,
                                                                       'id': '14',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '344681570190',
                                                                       'in_ucast_pkts': '287058533',
                                                                       'links': [       {       'in_nucast_pkts': '7',
                                                                                                'in_octets': '344681570190',
                                                                                                'in_ucast_pkts': '287058533',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27643',
                                                                                                'out_octets': '41581374486',
                                                                                                'out_ucast_pkts': '92785301',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27643',
                                                                       'out_octets': '41581374486',
                                                                       'out_ucast_pkts': '92785301',
                                                                       'poller_hash': 'MTY2REUyQnBlZXJpbnRmcG9uIDE0',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '14',
                                                                                               'links': [       {       'connected_time': '1520331',
                                                                                                                        'fiber_length': 6730,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.326,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.578}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2REUyQnBlZXJwb25wb24gMTQ=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '141113_TheHerbFarm (86 Grove Road)',
                                         'poller_hash': 'MTY2REUyQg==',
                                         'serial_number': '166DE2B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:09:ad',
                                         'uptime': '1520331'},
                                 {       'counter_source': '15/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:ca:59:48:fa:ac'],
                                                                       'id': '15/g4',
                                                                       'in_nucast_pkts': '10',
                                                                       'in_octets': '25352610052',
                                                                       'in_ucast_pkts': '108304178',
                                                                       'mac_address': '1c:8b:76:b5:2a:7b',
                                                                       'mtu': 9974,
                                                                       'name': '15/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '370763996083',
                                                                       'out_ucast_pkts': '295912483',
                                                                       'poller_hash': 'MTY2RTI0NHBlZXJpbnRmMTUvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '15',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '370763996083',
                                                                       'in_ucast_pkts': '295912483',
                                                                       'links': [       {       'in_nucast_pkts': '14',
                                                                                                'in_octets': '370763996083',
                                                                                                'in_ucast_pkts': '295912483',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '10',
                                                                                                'out_octets': '25352610052',
                                                                                                'out_ucast_pkts': '108304178',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '10',
                                                                       'out_octets': '25352610052',
                                                                       'out_ucast_pkts': '108304178',
                                                                       'poller_hash': 'MTY2RTI0NHBlZXJpbnRmcG9uIDE1',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '15',
                                                                                               'links': [       {       'connected_time': '1520331',
                                                                                                                        'fiber_length': 6770,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.622,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.85}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTI0NHBlZXJwb25wb24gMTU=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '140320_LizzieAsh (122 Grove Road)',
                                         'poller_hash': 'MTY2RTI0NA==',
                                         'serial_number': '166E244',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:2a:75',
                                         'uptime': '1520331'},
                                 {       'counter_source': '16/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:18:40:17'],
                                                                       'id': '16/g4',
                                                                       'in_nucast_pkts': '27625',
                                                                       'in_octets': '18879918486',
                                                                       'in_ucast_pkts': '86040396',
                                                                       'mac_address': '1c:8b:76:b5:59:93',
                                                                       'mtu': 9974,
                                                                       'name': '16/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '495745453822',
                                                                       'out_ucast_pkts': '374457374',
                                                                       'poller_hash': 'MTY2RTgyN3BlZXJpbnRmMTYvZzQ=',
                                                                       'speed': '2500000000'},
                                                               {       'admin_up': True,
                                                                       'id': '16',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '495745453822',
                                                                       'in_ucast_pkts': '374457374',
                                                                       'links': [       {       'in_nucast_pkts': '7',
                                                                                                'in_octets': '495745453822',
                                                                                                'in_ucast_pkts': '374457374',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27625',
                                                                                                'out_octets': '18879918486',
                                                                                                'out_ucast_pkts': '86040396',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27625',
                                                                       'out_octets': '18879918486',
                                                                       'out_ucast_pkts': '86040396',
                                                                       'poller_hash': 'MTY2RTgyN3BlZXJpbnRmcG9uIDE2',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '16',
                                                                                               'links': [       {       'connected_time': '1520325',
                                                                                                                        'fiber_length': 6626,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.142,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.25}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTgyN3BlZXJwb25wb24gMTY=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '141575_ArnoldCraw (81 Grove Road)',
                                         'poller_hash': 'MTY2RTgyNw==',
                                         'serial_number': '166E827',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:59:8d',
                                         'uptime': '1520325'},
                                 {       'counter_source': '17/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:18:40:21'],
                                                                       'id': '17/g4',
                                                                       'in_nucast_pkts': '109973',
                                                                       'in_octets': '14834051241',
                                                                       'in_ucast_pkts': '56632797',
                                                                       'mac_address': '1c:8b:76:b5:02:0b',
                                                                       'mtu': 9974,
                                                                       'name': '17/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '7',
                                                                       'out_octets': '245236369864',
                                                                       'out_ucast_pkts': '180593220',
                                                                       'poller_hash': 'MTY2REQzNnBlZXJpbnRmMTcvZzQ=',
                                                                       'speed': '1000000000'},
                                                               {       'admin_up': True,
                                                                       'id': '17',
                                                                       'in_nucast_pkts': '7',
                                                                       'in_octets': '245236369864',
                                                                       'in_ucast_pkts': '180593220',
                                                                       'links': [       {       'in_nucast_pkts': '7',
                                                                                                'in_octets': '245236369864',
                                                                                                'in_ucast_pkts': '180593220',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '109973',
                                                                                                'out_octets': '14834051241',
                                                                                                'out_ucast_pkts': '56632797',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '109973',
                                                                       'out_octets': '14834051241',
                                                                       'out_ucast_pkts': '56632797',
                                                                       'poller_hash': 'MTY2REQzNnBlZXJpbnRmcG9uIDE3',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '17',
                                                                                               'links': [       {       'connected_time': '1520326',
                                                                                                                        'fiber_length': 6612,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.16,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.322}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2REQzNnBlZXJwb25wb24gMTc=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '141582_AJJonker (1 Ashgrove Lane)',
                                         'poller_hash': 'MTY2REQzNg==',
                                         'serial_number': '166DD36',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:02:05',
                                         'uptime': '1520326'},
                                 {       'counter_source': '18/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:18:45:99'],
                                                                       'id': '18/g4',
                                                                       'in_nucast_pkts': '27643',
                                                                       'in_octets': '4367043755',
                                                                       'in_ucast_pkts': '18880744',
                                                                       'mac_address': '1c:8b:76:b5:59:b3',
                                                                       'mtu': 9974,
                                                                       'name': '18/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '53620654324',
                                                                       'out_ucast_pkts': '46552388',
                                                                       'poller_hash': 'MTY2RTgyQnBlZXJpbnRmMTgvZzQ=',
                                                                       'speed': '2500000000'},
                                                               {       'admin_up': True,
                                                                       'id': '18',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '53620654324',
                                                                       'in_ucast_pkts': '46552388',
                                                                       'links': [       {       'in_nucast_pkts': '14',
                                                                                                'in_octets': '53620654324',
                                                                                                'in_ucast_pkts': '46552388',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '27643',
                                                                                                'out_octets': '4367043755',
                                                                                                'out_ucast_pkts': '18880744',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '27643',
                                                                       'out_octets': '4367043755',
                                                                       'out_ucast_pkts': '18880744',
                                                                       'poller_hash': 'MTY2RTgyQnBlZXJpbnRmcG9uIDE4',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '18',
                                                                                               'links': [       {       'connected_time': '1520324',
                                                                                                                        'fiber_length': 6803,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -15.316,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.818}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2RTgyQnBlZXJwb25wb24gMTg=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '141311_RussellandAnnetteHughes (19 '
                                                 'Ashgrove Lane)',
                                         'poller_hash': 'MTY2RTgyQg==',
                                         'serial_number': '166E82B',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:59:ad',
                                         'uptime': '1520324'},
                                 {       'counter_source': '19/g4',
                                         'interfaces': [       {       'admin_up': True,
                                                                       'bridged_mac_addresses': [       'f8:27:2e:18:39:f5'],
                                                                       'id': '19/g4',
                                                                       'in_nucast_pkts': '12816',
                                                                       'in_octets': '136653847820',
                                                                       'in_ucast_pkts': '165388692',
                                                                       'mac_address': '1c:8b:76:b5:17:0b',
                                                                       'mtu': 9974,
                                                                       'name': '19/g4',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '14',
                                                                       'out_octets': '230595532402',
                                                                       'out_ucast_pkts': '264515272',
                                                                       'poller_hash': 'MTY2REZENnBlZXJpbnRmMTkvZzQ=',
                                                                       'speed': '2500000000'},
                                                               {       'admin_up': True,
                                                                       'id': '19',
                                                                       'in_nucast_pkts': '14',
                                                                       'in_octets': '230595532402',
                                                                       'in_ucast_pkts': '264515272',
                                                                       'links': [       {       'in_nucast_pkts': '14',
                                                                                                'in_octets': '230595532402',
                                                                                                'in_ucast_pkts': '264515272',
                                                                                                'oper_up': True,
                                                                                                'out_nucast_pkts': '12816',
                                                                                                'out_octets': '136653847820',
                                                                                                'out_ucast_pkts': '165388692',
                                                                                                'poller_hash': 'NDcxOTA1MDEwNDE1Z3BvbiAxLzI='}],
                                                                       'name': 'pon 1',
                                                                       'oper_up': True,
                                                                       'out_nucast_pkts': '12816',
                                                                       'out_octets': '136653847820',
                                                                       'out_ucast_pkts': '165388692',
                                                                       'poller_hash': 'MTY2REZENnBlZXJpbnRmcG9uIDE5',
                                                                       'pons': [       {       'description': 'pon '
                                                                                                              '1',
                                                                                               'id': '19',
                                                                                               'links': [       {       'connected_time': '339718',
                                                                                                                        'fiber_length': 6893,
                                                                                                                        'poller_hash': 'NDcxOTA1MDEwNDE1cG9uZ3BvbiAxLzI=',
                                                                                                                        'rx_link_rate': '2488320000',
                                                                                                                        'rx_power': -16.108,
                                                                                                                        'tx_link_rate': '1244160000',
                                                                                                                        'tx_power': 2.188}],
                                                                                               'max_distance': 20000,
                                                                                               'name': 'pon '
                                                                                                       '1',
                                                                                               'poller_hash': 'MTY2REZENnBlZXJwb25wb24gMTk=',
                                                                                               'pon_technology': 'GPON',
                                                                                               'rx_link_rate': '2488320000',
                                                                                               'split_ratio': 64,
                                                                                               'tx_link_rate': '1244160000'}]}],
                                         'manufacturer': 'Calix',
                                         'model': 'GP4200A',
                                         'name': '141084_BillKirkland (72 Grove Road)',
                                         'poller_hash': 'MTY2REZENg==',
                                         'serial_number': '166DFD6',
                                         'status': 'ELEMENT_STATUS_ONLINE',
                                         'sw_version': '24.1.0.0.39',
                                         'system_mac_address': '1c:8b:76:b5:17:05',
                                         'uptime': '339718'}]
        }
        id_field = 'mac_address' if not obj.stations else ('mac_address' if hasattr(obj.stations[0], 'mac_address') and getattr(obj.stations[0], 'mac_address') else ('imsi' if hasattr(obj.stations[0], 'imsi') else 'serial'))
        cur_info = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('info', 'ap', obj.model) + [('vendor', 'str'), ('model', 'str'), ('sw_version', 'str')]}
        if 'rf_freq' in cur_info and cur_info.get('rf_freq') and (cur_info['rf_freq'] <= 0 or int.bit_length(cur_info['rf_freq']) >= 32):
            cur_info.pop('rf_freq')
        cur_stats = {x: getattr(obj, x, None) for (x, data_type) in self.apdd.get_metric_fields('stats', 'ap', obj.model)}
        cur_sta_stats = {getattr(sta, id_field): {x: getattr(sta, x, None) for (x, data_type) in sorted(self.apdd.get_metric_fields('stats', 'cpe', obj.model)) + [(id_field, 'str'), ('ip_address', 'str'), ('name', 'str'), ('dev_macs', 'str')]} for sta in sorted(obj.stations)}
        cur_ne_msg = getattr(obj, 'network_element_update_msg', None)
        cur_ne_msg = cur_ne_msg.pb if cur_ne_msg else None
        self.assertEqual(cur_info, exp_info)
        self.assertEqual(cur_stats, exp_stats)
        self.assertEqual(cur_sta_stats, exp_sta_stats)
        self.assertTrue(self.compare_pb_dict(cur_ne_msg, exp_ne_msg))
